/*
** Module   : Code2Name.java
** Abstract : tool which provides rudimentary conversion services for locale scripts.
*             This particular app will replace the char names with their codes.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20160602 Created initial version. Using a naive approach, when a pattern like <abc>
**                  is encountered a lookup in the internal map is performed to identify the
**                  charcode and in affirmative case it is replaced in in the output file. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

import java.io.*;
import java.util.*;

/** 
 * Utility class that replaces any occurrence of a pattern like {@code <abc..k>} with its code 
 *
 * Note that this helper class is not compiled by default into the main P2J jar file. It is a 
 * standalone class.
 */
public class Name2Code
{
   /** The character used in a escape sequence. */
   private static final char ESCAPE_CHAR = '/';
   
   /** The character used as comment. Lines started with it are simply copied to output file. */
   private static final char COMMENT_CHAR = '%';
   
   /** Internal mapping. */ 
   private static final Map<String, Integer> TO_CODE = new HashMap<>(256);

   /** Character name mapping. */
   private static final String[] CHAR_CODES =
   {
      "NUL",
      "SOH",
      "STX",
      "ETX",
      "EOT",
      "ENQ",
      "ACK",
      "BEL",
      "backspace",
      "tab",
      "newline",
      "vertical-tab",
      "form-feed",
      "carriage-return",
      "SO",
      "SI",
      "DLE",
      "DC1",
      "DC2",
      "DC3",
      "DC4",
      "NAK",
      "SYN",
      "ETB",
      "CAN",
      "EM",
      "SUB",
      "ESC",
      "IS4",
      "IS3",
      "IS2",
      "IS1",
      "SP",  // also found as 'space'
      "exclamation",
      "quotation-mark",
      "number-sign",
      "dollar-sign",
      "percent-sign",
      "ampersand",
      "apostrophe",
      "left-parenthesis",
      "right-parenthesis",
      "asterisk",
      "plus-sign",
      "comma",
      "hyphen",
      "period",
      "slash",
      "zero",
      "one",
      "two",
      "three",
      "four",
      "five",
      "six",
      "seven",
      "eight",
      "nine",
      "colon",
      "semicolon",
      "less-than",
      "equals",
      "greater-than",
      "question",
      "atsym",
      "A",
      "B",
      "C",
      "D",
      "E",
      "F",
      "G",
      "H",
      "I",
      "J",
      "K",
      "L",
      "M",
      "N",
      "O",
      "P",
      "Q",
      "R",
      "S",
      "T",
      "U",
      "V",
      "W",
      "X",
      "Y",
      "Z",
      "left-sq-br",
      "rev-solidus",
      "right-sq-br",
      "circumflex",
      "underscore",
      "grave",
      "a",
      "b",
      "c",
      "d",
      "e",
      "f",
      "g",
      "h",
      "i",
      "j",
      "k",
      "l",
      "m",
      "n",
      "o",
      "p",
      "q",
      "r",
      "s",
      "t",
      "u",
      "v",
      "w",
      "x",
      "y",
      "z",
      "left-brace",
      "vertical-line",
      "right-brace",
      "tilde",
      "DEL",
      "PAD",
      "HOP",
      "BPH",
      "NBH",
      "IND",
      "NEL",
      "SSA",
      "ESA",
      "HTS",
      "HTJ",
      "VTS",
      "PLD",
      "PLU",
      "RI",
      "SS2",
      "SS3",
      "DCS",
      "PU1",
      "PU2",
      "STS",
      "CCH",
      "MW",
      "SPS",
      "EPA",
      "SOS",
      "SGCI",
      "SCI",
      "CSI",
      "ST",
      "OSC",
      "PM",
      "APC",
      "nobreak-space",
      "exclamation-down",
      "cent",
      "sterling",
      "currency",
      "yen",
      "broken-bar",
      "section",
      "diaresis",
      "copyright",
      "masculine",
      "guillemot-left",
      "not",
      "dash",
      "registered",
      "macron",
      "degree",
      "plus-minus",
      "two-supperior",
      "three-supperior",
      "acute",
      "mu",
      "paragraph",
      "dot",
      "cedilla",
      "one-supperior",
      "feminine",
      "guillemot-right",
      "quarter",
      "half",
      "three-quarters",
      "question-down",
      "A-grave",
      "A-acute",
      "A-circumflex",
      "A-tilde",
      "A-diaresis",
      "A-ring",
      "AE",
      "C-cedilla",
      "E-grave",
      "E-acute",
      "E-circumflex",
      "E-diaresis",
      "I-grave",
      "I-acute",
      "I-circumflex",
      "I-diaresis",
      "Eth",
      "N-tilde",
      "O-grave",
      "O-acute",
      "O-circumflex",
      "O-tilde",
      "O-diaresis",
      "multiply",
      "O-slash",
      "U-grave",
      "U-acute",
      "U-circumflex",
      "U-diaresis",
      "Y-acute",
      "Thorn",
      "s-sharp",
      "a-grave",
      "a-acute",
      "a-circumflex",
      "a-tilde",
      "a-diaresis",
      "a-ring",
      "ae",
      "c-cedilla",
      "e-grave",
      "e-acute",
      "e-circumflex",
      "e-diaresis",
      "i-grave",
      "i-acute",
      "i-circumflex",
      "i-diaresis",
      "eth",
      "n-tilde",
      "o-grave",
      "o-acute",
      "o-circumflex",
      "o-tilde",
      "o-diaresis",
      "divide",
      "o-slash",
      "u-grave",
      "u-acute",
      "u-circumflex",
      "u-diaresis",
      "y-acute",
      "thorn",
      "y-diaresis"
   };
   
   static 
   {
      // prepare internal data
      for (int k = 0; k < CHAR_CODES.length; k++)
      {
         TO_CODE.put(CHAR_CODES[k], k);
      }
   }
   
   /**
    * Application entry point. Requires two parameters, input and output file. 
    *
    * @param   args
    *          The command line arguments. 
    */
   public static void main(String[] args)
   {
      if (args.length != 2)
      {
         System.err.println("Name2Code: Converts a locale from char-name declaration to char-code.");
         System.err.println("Syntax is 'Name2Code <name-locale> <code-locale>'.");
         System.exit(-1);
      }
   
      String inLine;
      StringBuilder outLine;
      int cnt = 0;
      int wrn = 0;
      
      BufferedReader br;
      try
      {
         br = new BufferedReader(new InputStreamReader(new FileInputStream(args[0])));
      }
      catch (FileNotFoundException fnfe)
      {
         System.err.println("Input locale script not found.");
         System.exit(-2);
         return;
      }
   
      BufferedWriter bw;
      try
      {
         bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(args[1])));
      }
      catch (FileNotFoundException fnfe)
      {
         System.err.println("Output locale script file cannot be created.");
         System.exit(-3);
         return;
      }
   
      try
      {
         int line = 0;
         while ((inLine = br.readLine()) != null)
         {
            line++;
            if (inLine.trim().startsWith(String.valueOf(COMMENT_CHAR)))
            {
               bw.write(inLine); // keep comments unchanged
            }
            else 
            {
               outLine = new StringBuilder(inLine.length()); // just a hint
               int k = 0;
               while (k < inLine.length())
               {
                  String charNameStr = null;
                  if (inLine.charAt(k) == '<')
                  {
                     // search for closing '>'
                     for (int i = k + 2; i < inLine.length(); i++)
                     {
                        if (inLine.charAt(i) == '>')
                        {
                           charNameStr = inLine.substring(k + 1, i);
                           break;
                        }
                     }
                  }
                  if (charNameStr != null) 
                  {
                     Integer code = TO_CODE.get(charNameStr);
                     if (code == null)
                     {
                        System.err.println("Invalid char name '" + charNameStr + 
                                           "' at " + line + ":" + k + ".");
                        outLine.append("<").append(charNameStr).append('>'); // leave unchanged
                        ++ wrn;
                     }
                     else
                     {
                        outLine.append("<U").append(hexDigits(code)).append('>'); // replace
                        ++ cnt;
                     }
                     k += charNameStr.length() + 2; 
                  }
                  else
                  {
                     outLine.append(inLine.charAt(k)); // leave char unchanged
                     k++;
                  }
               }
               bw.write(outLine.toString());
            }
            bw.newLine();
         }
      }
      catch (IOException ioe)
      {
         System.err.println("Failed to read from locale script file.");
         System.exit(-10);
      }
   
      try
      {
         br.close();
      }
      catch (IOException e) { /*ignored*/ }
      try
      {
         bw.close();
      }
      catch (IOException e) { /*ignored*/ }
   
      System.out.println("Done. \n" + 
                         cnt + " characters codes names replaced.\n" +
                         wrn + " recoverable errors were encountered.");
   }
   
   /**
    * Utility method that converts an integer to its hexadecimal representation on 4 digits.
    * 
    * @param   code
    *          An integer value.
    *
    * @return  The hexadecimal representation on 4 digits of the input parameter.
    */
   private static String hexDigits(int code)
   {
      String ret = "000" + Integer.toHexString(code).toUpperCase();
      return ret.substring(ret.length() - 4); // the last 4 digits is what we need
   }
   
}