/*
** Module   : AsmClassLoader.java
** Abstract : class loader for custom assembled interfaces and classes
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20150407 Adapted from ProxyClassLoader, formerly in the com.goldencode.proxy package.
** 002 ECF 20150828 Clean up javadoc.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.asm;

import java.util.*;

/**
 * A custom class loader which loads dynamically generated, in-memory classes. Each is
 * represented as an array of bytes which conforms to the Java class file format as specified by
 * <u>The Java Virtual Machine Specification, Second Edition</u> by Tim Lindholm and Frank Yellin.
 * <p>
 * A class is loaded by invoking the <code>loadClass(String, byte[])</code> method, passing it
 * the fully qualified name of the class, and the byte code array representing its class file.
 * <p>
 * TODO: enable class unloading by delegating to individual instances of this class, organized by
 * key, which can be discarded when their loaded classes are no longer needed.
 */
public final class AsmClassLoader
extends ClassLoader
{
   /** Singleton instance */
   private static AsmClassLoader instance = null;
   
   /** Thread-local, temporary association of a class' name to its data */
   private final ThreadLocal<AsmClassLoader.Data> local = new ThreadLocal<>();
   
   /** Cache of previously loaded, dynamic classes */
   private final Map<String, Class<?>> loaded = new HashMap<>();
   
   /**
    * Default constructor. Used if all class loading is managed by the caller.
    */
   public AsmClassLoader()
   {
      super();
   }
   
   /**
    * Constructor which accepts a parent class loader. Used if classes must be loaded through
    * standard means (e.g., <code>Class.forName</code>.
    * 
    * @param   parent
    *          Parent class loader.
    */
   public AsmClassLoader(ClassLoader parent)
   {
      super(parent);
      if (instance != null)
      {
         throw new IllegalStateException("There can be only one AsmClassLoader");
      }
      
      instance = this;
   }
   
   /**
    * Get the singleton instance of this class loader. This method is not synchronized because
    * it should only be called after the singleton class loader has been created, which occurs
    * early in the life of the JVM.
    * 
    * @return  Singleton instance of the ASM class loader.
    */
   public static AsmClassLoader getInstance()
   {
      return instance;
   }
   
   /**
    * Overrides the parent's implementation to find the class file data to first check the cache
    * of previously loaded classes, and if not found there, to load the dynamically generated
    * class with the given name.
    * 
    * @param   name
    *          Name of the class to be loaded.
    * 
    * @return  Dynamically loaded class.
    * 
    * @throws  ClassNotFoundException
    *          if no thread-local class file data is found, or if its name does not match the
    *          given class name.
    */
   @Override
   public Class<?> findClass(String name)
   throws ClassNotFoundException
   {
      Class<?> existing = loaded.get(name);
      if (existing != null)
      {
         return existing;
      }
      
      AsmClassLoader.Data data = local.get();
      if (data == null || !name.equals(data.name))
      {
         throw new ClassNotFoundException(name);
      }
      
      byte[] code = data.code;
      Class<?> clazz = defineClass(name, code, 0, code.length);
      
      loaded.put(name, clazz);
      
      return clazz;
   }
   
   /**
    * Store a piece of thread-local data which associates the given class name with the byte
    * array representing its class file.  This information is stored temporarily, while the work
    * of loading the class is delegated to the <code>loadClass(String, boolean)</code> method,
    * then the thread-local data is immediately removed.
    * 
    * @param   name
    *          Name of the proxy class to be loaded.
    * @param   code
    *          Byte array representing the proxy's class file.
    * 
    * @return  The proxy class.
    * 
    * @throws  ClassNotFoundException
    *          if something unexpected goes horribly wrong with the
    *          thread-local data mechanism.
    */
   public Class<?> loadClass(String name, byte[] code)
   throws ClassNotFoundException
   {
      try
      {
         AsmClassLoader.Data data = new Data();
         data.name = name;
         data.code = code;
         local.set(data);
         
         return loadClass(name, true);
      }
      finally
      {
         local.remove();
      }
   }
   
   /**
    * Association between an assembled class' name and its class file data. Used transiently
    * during class loading.
    */
   private static class Data
   {
      /** Class name */
      private String name;
      
      /** Byte array representing class file */
      private byte[] code;
   }
}