/*
** Module   : AsmUtils.java
** Abstract : helper methods for ASM bytecode assembly
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 ECF 20150407 Adapted from the former class com.goldencode.proxy.Utils.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.asm;

import org.objectweb.asm.*;

/**
 * Static helper methods for ASM bytecode assembly.
 */
public final class AsmUtils
{
   /**
    * Given a type name, return a string representation which is suitable for use as an internal
    * type name; that is, replace all dot separators in the name with forward slashes.
    * 
    * @param   name
    *          External class name which uses dot separators.
    * 
    * @return  Internal class name which uses slash separators, derived from
    *          the given class name.
    */
   public static String commonToInternalTypeName(String name)
   {
      return name.replaceAll("\\.", "/");
   }
   
   /**
    * Given a class, return a string representation of its name which is
    * suitable for use as an internal class name.  That is, replace all dot
    * separators in the class' name with forward slashes.
    * 
    * @param   type
    *          A class.
    * 
    * @return  Internal class name which uses slash separators, derived from
    *          the given class' name.
    */
   public static String commonToInternalTypeName(Class<?> type)
   {
      return commonToInternalTypeName(type.getName());
   }
   
   /**
    * Helper method to choose the most appropriate instruction with which to
    * push an integer value onto the stack.  The instruction used differs,
    * based upon the value of the integer.
    * 
    * @param   mv
    *          <code>CodeVisitor</code> used to write the instruction.
    * @param   value
    *          Integer value to be pushed.
    */
   public static void pushInt(MethodVisitor mv, int value)
   {
      switch (value)
      {
         case 0:
            mv.visitInsn(Opcodes.ICONST_0);
            break;
         case 1:
            mv.visitInsn(Opcodes.ICONST_1);
            break;
         case 2:
            mv.visitInsn(Opcodes.ICONST_2);
            break;
         case 3:
            mv.visitInsn(Opcodes.ICONST_3);
            break;
         case 4:
            mv.visitInsn(Opcodes.ICONST_4);
            break;
         case 5:
            mv.visitInsn(Opcodes.ICONST_5);
            break;
         case -1:
            mv.visitInsn(Opcodes.ICONST_M1);
            break;
         default:
            if (value >= Byte.MIN_VALUE && value <= Byte.MAX_VALUE)
            {
               mv.visitIntInsn(Opcodes.BIPUSH, value);
            }
            else if (value >= Short.MIN_VALUE && value <= Short.MAX_VALUE)
            {
               mv.visitIntInsn(Opcodes.SIPUSH, value);
            }
            else
            {
               mv.visitLdcInsn(value);
            }
            break;
      }
   }
}