/*
** Module   : BasePlugin.java
** Abstract : maintains a project-specific master bidirectional indexing
**            facility for tree names to unique IDs
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20090513   @42200 Refactored code from AstRegistry to implement the first version.
** 002 ECF 20140918          Added usesMemoryStorage method, to help with memory management.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import java.util.*;
import java.util.concurrent.locks.*;

/**
 * Abstract base class for plugins which provides a common implementation for
 * features which are not related to the state persistence mechanism.
 */
public abstract class BasePlugin
implements AstManagerPlugin
{
   /** Allows easy removal of the node ID bits in an AST ID. */
   protected static final long TREE_MASK = 0xFFFFFFFF00000000L;
   
   /** Number of bits to shift to make a file ID a normal int. */
   protected static final int HIGH_BITS = 32;
   
   /** Specifies the maximum possible number of files in the registry. */
   protected static final long MAX_POSSIBLE = Integer.MAX_VALUE;
   
   /** Stores the tree name to ID mapping information. */
   protected Map<String, Long> fnames = new HashMap<>();
      
   /** Stores the ID to tree name mapping information. */
   protected Map<Long, Pair> fids = new HashMap<>();
   
   /** Stores the tree name to lock mapping information. */
   protected Map<Long, ReentrantLock> locklist = new HashMap<>();
      
   /** Records if any changes have been made since the registry was read. */
   protected boolean dirty = false;
   
   /**
    * Does this plug-in use memory as its primary storage medium? This method returns
    * <code>false</code> by default; subclasses should override as necessary.
    * 
    * @return  <code>false</code>.
    */
   public boolean usesMemoryStorage()
   {
      return false;
   }
   
   /**
    * Lookup the name associated with the tree specified by this ID. The ID can
    * be a tree ID (an ID that represents an entire AST) or it can be the ID
    * an individual AST node ID. Node IDs encode the tree ID and a portion that
    * is node-specific. The node-specific portion will be ignored.
    *
    * @param    id
    *           The ID to use as a key in the lookup. This ID may be a tree ID
    *           or a specific node ID.
    *
    * @return   The name associated with the tree or <code>null</code> if no
    *           such tree ID exists.
    */
   public String getTreeName(long id)
   {
      Pair   pair = null;
      String name = null;
      
      if (id != AstManager.INVALID_ID)
      {
         synchronized (this)
         {
            pair = fids.get(getTreeId(id));
            
            if (pair != null)
            {
               name = pair.name;
            }
         }
      }            
      
      return name;
   }
   
   /**
    * Reverse lookup of the ID that corresponds to a given AST name.
    *
    * @param    name
    *           Tree name whose unique ID should be returned.
    *
    * @return   The tree ID associated with the name or
    *           {@link AstManager#INVALID_ID} if the tree name doesn't exist or
    *           is <code>null</code>.
    */
   public long getTreeId(String name)
   {
      long id = AstManager.INVALID_ID;
      
      if (name != null)
      {
         synchronized (this)
         {
            Long possible = fnames.get(name);
            
            if (possible != null)
            {
               id = possible;
            }
         }
      }
      
      return id;
   }
   
   /**
    * Returns the tree ID portion of a node ID.
    *
    * @return   The ID which corresponds to the tree itself, with all
    *           node ID portions removed (masked off) or
    *           {@link AstManager#INVALID_ID} if the ID is invalid.
    */
   public long getTreeId(long id)
   {
      return (id != AstManager.INVALID_ID) ? id & TREE_MASK
                                           : AstManager.INVALID_ID;
   }
   
   /**
    * Obtains a process-wide lock on any processing related to the given tree.
    * If the lock is already held, this method will block until it obtains
    * the lock. When this method returns normally, the calling thread is
    * guaranteed to be holding the lock. This may be called recursively by the
    * same thread and if the current thread is already holding the lock, then
    * this method will only update the reference count for the lock but
    * otherwise it will immediately return. As a consequence, the lock will not
    * be released unless {@link #unlockTree} is called the same number of times
    * as this lock method. Thus the calls to this lock method and to
    * <code>unlock</code> must always be balanced.
    * <p>
    * This locking has nothing to do with file system locks nor is there any
    * non-Java locking backing this lock. Rather, this lock is a cooperative
    * agreement between all callers (within the same JVM) that no use of the
    * given tree will be made except when the caller holds this exclusive lock
    * to the resource.
    *
    * @param    id
    *           Tree ID that is associated with the lock.
    *
    * @throws   InterruptedException
    *           When an interruption occurs during blocking for the lock. If
    *           this is thrown, the caller will NOT be holding the lock.
    * @throws   IllegalArgumentException
    *           If the ID is not a valid tree ID.
    */
   public void lockTree(long id)
   throws InterruptedException,
          IllegalArgumentException
   {
      ReentrantLock lock = null;
      
      if (isInvalid(id))
      {
         throw new IllegalArgumentException("Invalid tree ID " + id);
      }
      
      // protect our core data structures
      synchronized (this)
      {
         lock = locklist.get(id);
         
         if (lock == null)
         {
            lock = new ReentrantLock();
            locklist.put(id, lock);
         }
      }
      
      // this blocks so it must be done outside the synch block (above)
      lock.lockInterruptibly();
   }
   
   /**
    * Releases the process-wide lock on any processing related to the given
    * tree. If the lock is already held by the calling thread, this method 
    * will decrement the reference count. If the resulting reference count is
    * 0, the lock will be released. If the calling thread does not hold the
    * lock, this method will exit abnormally. This may be called recursively by
    * the same thread and must be called the same number of times as the
    * {@link #lockTree} method. Thus the calls to this unlock method must always
    * be balanced with calls to this method.
    * <p>
    * This locking has nothing to do with file system locks nor is there any
    * non-Java locking backing this lock. Rather, this lock is a cooperative
    * agreement between all callers (within the same JVM) that no use of the
    * given tree will be made except when the caller holds this exclusive lock
    * to the resource.
    *
    * @param    id
    *           Tree ID that is associated with the lock.
    *
    * @throws   IllegalMonitorStateException
    *           If the calling thread does not own the lock.
    * @throws   IllegalArgumentException
    *           If the ID is not a valid tree ID.
    */
   public void unlockTree(long id)
   throws IllegalMonitorStateException,
          IllegalArgumentException
   {
      ReentrantLock lock = null;
      
      if (isInvalid(id))
      {
         throw new IllegalArgumentException("Invalid tree ID " + id);
      }
      
      // protect our core data structures
      synchronized (this)
      {
         lock = locklist.get(id);
         
         if (lock == null)
         {
            throw new IllegalMonitorStateException("No such lock!");
         }
      }
      
      // may also throw the exception if the current thread is not the owner
      lock.unlock();
   }
   
   /**
    * Access the next unassigned node ID for the specified tree. The
    * value is looked up and returned, just before returning the internal
    * cache is incremented and stored back for next use. Once this value is
    * returned, it is considered "used" and will not be returned again.
    *
    * @param    id
    *           The ID that specifies a tree.
    *
    * @return   The newly assigned node ID or {@link AstManager#INVALID_ID} if
    *           no such tree exists.
    *
    * @throws   AstException
    *           If too many AST nodes exist.
    */
   public long getNextNodeId(long id)
   throws AstException
   {
      Pair pair = null;
      long next = AstManager.INVALID_ID;
      
      if (id != AstManager.INVALID_ID)
      {
         synchronized (this)
         {
            pair = fids.get(getTreeId(id));
            
            if (pair != null)
            {
               next = pair.next;
               pair.next++;
               
               // mark the registry as needing a save
               dirty = true;
               
               if ((pair.next & ~TREE_MASK) > MAX_POSSIBLE)
               {
                  throw new AstException("Too many AST nodes in file.");
               }
            }
         }         
      }
            
      return next;
   }
      
   /**
    * Resets the next unassigned AST node ID for the specified tree to the
    * first possible value (tree ID + 1).
    * <p>
    * <b>WARNING: there is no check on the current usage of node IDs for this
    * file, so if AST nodes exist already they will have IDs at or above this
    * value and duplicate IDs may (probably will) result!</b>
    *
    * @param    treeId
    *           The ID that specifies a valid tree.
    *
    * @throws   AstException
    *           If the file ID is invalid.
    */
   public void resetNextNodeId(long treeId)
   throws AstException
   {  
      if (!isInvalid(treeId))
      {
         setNextNodeId(treeId, treeId + 1);
      }
      else
      {
         throw new AstException("Invalid tree ID " + treeId);
      }
   }
            
   /**
    * Adds a mapping from the specified AST to a new ID. The next available
    * tree ID is assigned and this mapping is added to the registry. If the
    * tree name already exists, the existing tree ID is returned.
    *
    * @param    name
    *           The tree name. Often this will be the name of a file.
    *
    * @return   The assigned tree ID.
    *
    * @throws   AstException
    *           If too many ASTs exist in the registry or if the input
    *           name is <code>null</code>.
    */
   public long addTree(String name)
   throws AstException
   {
      if (name == null)
      {
         throw new AstException("Cannot add null source name to the registry.");
      }
      
      long result = getTreeId(name);
      
      if (result == AstManager.INVALID_ID)
      {
         synchronized (this)
         {
            // this is the only place a new filename to ID mapping can occur
            // take the size of the mapping table (the number of entries),
            // this is our starting number
            long possible = fids.size();
            
            do
            {
               // increment the number to hopefully make one that is unique
               possible++;
               
               // test for overflow (the compliment of the TREE_MASK is the
               // largest possible file ID
               if (possible > MAX_POSSIBLE)
               {
                  throw new AstException("Too many files in registry.");
               }
               
               // shift our number right by the number of bits assigned to
               // node IDs to get the next possible file ID
               result = possible << HIGH_BITS;
               
            } while (getTreeName(result) != null);
            
            
            // the first node ID is stored in the next member of the inner
            // class Pair, this value is set to the file ID + 1 which is the
            // first unique AST node ID
            Pair newFile = new Pair(name, result + 1);
            fids.put(result, newFile);
            fnames.put(name, result);  
            
            // mark the registry as needing a save
            dirty = true;
         }
      }
      
      return result;
   }
   
   /**
    * Removes all mapping data referencing the specified tree.
    *
    * @param    id
    *           The ID the tree to remove.
    *
    * @return   <code>true</code> if the tree existed in the registry (and 
    *           was removed), <code>false</code> if the tree didn't exist.
    */
   public boolean removeTree(long id)
   throws AstException
   {  
      if (id == AstManager.INVALID_ID)
      {
         // nothing to do
         return false;
      }
      
      synchronized (this)
      {
         Pair pair = fids.get(id);
         
         if (pair == null)
         {
            // invalid ID
            return false;
         }
         
         fnames.remove(pair.name);
         fids.remove(id);
         
         // mark the registry as needing a save
         dirty = true;
      }
      
      return true;
   }
      
   /**
    * Set the next unassigned AST node ID for the specified tree.
    * <p>
    * <b>WARNING: there is no check on the validity of this new value, if AST
    * nodes exist with IDs at or above this value, duplicate IDs may
    * result!</b>
    *
    * @param    treeId
    *           The ID that specifies a tree.
    * @param    nodeId
    *           The ID used for the next node ID.
    *
    * @throws   AstException
    *           If the any input is invalid.
    */
   protected void setNextNodeId(long treeId, long nodeId)
   throws AstException
   {  
      if (treeId != AstManager.INVALID_ID && nodeId != AstManager.INVALID_ID)
      {
         int simple = (int) (nodeId & ~TREE_MASK);
         
         if (simple < 1 || simple > MAX_POSSIBLE)
         {
            throw new AstException("Invalid node ID ("+ nodeId + ").");
         }         
         
         synchronized (this)
         {
            Pair pair = fids.get(getTreeId(treeId));
            
            if (pair != null)
            {
               pair.next = nodeId;
               
               // mark the registry as needing a save
               dirty = true;            
            }
         }         
      }
      else
      {
         throw new AstException("Invalid input(s) for treeId (" + treeId + 
                                ") and/or nodeId (" + nodeId + ").");
      }
   }
            
   /**
    * Detects if the given ID is invalid. An ID is invalid if it is equivalent
    * to the {@link AstManager#INVALID_ID} or if the tree portion does not match
    * a registered tree.
    *
    * @param    id
    *           The tree or node ID to test.
    *
    * @return   <code>true</code> if the ID is not valid.
    */
   protected boolean isInvalid(long id)
   {
      if (id == AstManager.INVALID_ID)
         return true;
      
      boolean registered = false;
      
      synchronized (this)
      {
         long tid = getTreeId(id);
         
         if (fids.containsKey(tid))
            registered = true;
      }
      
      return !registered;
   }
      
   /**
    * Simple container class to store the next available ID for a tree and
    * the associated tree name.
    */
   protected static class Pair
   {
      /** A tree name. */
      protected String name;
      
      /** The next available ID. */
      protected long next;
      
      /**
       * Default constructor.
       *
       * @param    name
       *           A tree name.
       * @param    next
       *           The next available ID associated with this tree.
       */ 
      protected Pair(String name, long next)
      {
         this.name = name;
         this.next = next;
      }
   }
}