/*
** Module   : XmlFilePlugin.java
** Abstract : implements an XML file storage option for an AstManagerPlugin
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20090513   @42201 Refactored code from AstRegistry and AstPersister
**                           to implement the first version.
** 002 GES 20090518   @42372 Import change.
** 003 GES 20090602   @42543 Optionally ignore the .ast postfix in tree name to
**                           ID resolution.
** 004 GES 20100924          Added support to persist the tree-local data. At
**                           this time, only shadow nodes are tree-local.
** 005 GES 20101001          Shifted ShadowNode to private members. Forced
**                           the parent of shadow nodes to be the root of the
**                           AST which is required for certain lookups to work.
** 006 SVL 20130624          Do not try to store a file if its name is null.
** 007 CA  20131023          Implemented the isExistingAst. When saving an AST, make sure all the
**                           parent folders exist.
** 008 SVL 20140320          Added mode in which plugin reads data from resources rather than
**                           files.
** 009 ECF 20140808          Changed access of ELEM_AST and ATTR_ID to public.
** 010 ECF 20150715          Replace StringBuffer with StringBuilder.
** 011 EVL 20160222          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import com.goldencode.util.*;
import antlr.*;

/**
 * Provides an {@link AstManagerPlugin} implementation which is backed by an
 * XML file. 
 */
public class XmlFilePlugin
extends BasePlugin
{
   /** Convention for an XML AST filename extension. */
   public static final String AST_POSTFIX = ".ast";
   
   /** Root element containing the entire AST. */
   public static final String ELEM_ROOT = "ast-root";
   
   /** 
    * AST-specific element, containing all data as attributes, all
    * annotations as child elements and any number of child AST elements.
    */
   public static final String ELEM_AST = "ast";
   
   /**
    * The project-unique 64-bit identifier of each AST. The upper DWORD is
    * the file-specific ID and the lower DWORD is associated with the node
    * ID.  If an ID refers to a file the lower DWORD is 0x00000000.
    */
   public static final String ATTR_ID = "id";
   
   /** Calculated postfix length. */
   private static final int AST_POSTFIX_LEN = AST_POSTFIX.length();
   
   /** A single annotation as a child element of an AST element. */
   private static final String ELEM_ANN = "annotation";
   
   /** One element of an array annotation. */
   private static final String ELEM_LISTITEM = "listitem";
   
   /** Element which contains all tree-local data. */
   private static final String ELEM_LOCAL = "tree-local";
   
   /** Element containing a list of shadow nodes. */
   private static final String ELEM_SHADOW_LIST = "shadow-node-list";
   
   /** Element containing a special shadow node AST. */
   private static final String ELEM_SHADOW = "shadow-node";
   
   /** Root element containing the entire mapping database. */
   private static final String ELEM_REGISTRY_ROOT  = "name-root";
   
   /** Represents a single name = ID pair. */
   private static final String ELEM_MAP   = "mapping";
   
   /** The class name of the AnnotatedAst subclass used for all instances. */
   private static final String ATTR_CLASS = "ast_class";
   
   /** Indicator that this XML document uses terse AST format */
   private static final String ATTR_TERSE = "terse";
   
   /** The source text as parsed from the original input file. */
   private static final String ATTR_TEXT = "text";
   
   /** The token type of the AST. */
   private static final String ATTR_TTYPE = "type";
   
   /** The hidden member of the AST. */
   private static final String ATTR_HIDDEN = "hidden";
   
   /** The input file line number on which the AST's text starts. */
   private static final String ATTR_LINE = "line";
   
   /** The input file column number at which the AST's text starts. */
   private static final String ATTR_COL = "col";
   
   /** The data type of an annotation, as a Java class name. */
   private static final String ATTR_DTYPE = "datatype";
   
   /** The key with which a unique annotation can be stored or retrieved. */
   private static final String ATTR_KEY = "key";
   
   /** The value of an annotation. */
   private static final String ATTR_VAL = "value";
   
   /** Left-side AST identifier for a shadow node. */
   private static final String ATTR_LEFT_ID = "left-id";
   
   /** Right-side AST identifier for a shadow node. */
   private static final String ATTR_RIGHT_ID = "right-id";
   
   /** The filename that is associated with a unique ID. */
   private static final String ATTR_TREE = "treename";
      
   /** The next available AST node ID for that filename. */
   private static final String ATTR_NEXT = "next";
      
   /** XML file for persisting the registry. */ 
   private File file = null;

   /** If <code>true</code> then plugin reads data from resources rather than files. */
   private boolean useResources = false;
      
   /**
    * Constructor which looks up the registry file or resource name and initializes its data
    * structures based on reading that file or resource, if it exists.
    *
    * @param    name
    *           The registry will be loaded from this file or resource. Can be <code>null</code>.
    * @param    useResources
    *           If <code>true</code> then plugin reads data from resources rather than files.
    */
   public XmlFilePlugin(String name, boolean useResources)
   throws AstException
   {
      this.useResources = useResources;

      // configure the project's load and save point
      if (name != null)
      {
         if (useResources)
         {
            // TODO implement loading registry from the resource
         }
         else
         {
            file = new File(name);
      
            // try to load the persisted registry
            if (file.exists() && file.isFile())
            {
               load();
            }
         }
      }
   }

   /**
    * Constructor which looks up the registry filename and initializes its data structures based
    * on reading that file, if it exists.
    *
    * @param    name
    *           The registry will be loaded from this filename. Can be
    *           <code>null</code>.
    */
   public XmlFilePlugin(String name)
   throws AstException
   {
      this(name, false);
   }
   
   /**
    * Reverse lookup of the ID that corresponds to a given AST name.
    *
    * @param    name
    *           Tree name whose unique ID should be returned.
    *
    * @return   The tree ID associated with the name or
    *           {@link AstManager#INVALID_ID} if the tree name doesn't exist or
    *           is <code>null</code>.
    */
   public long getTreeId(String name)
   {
      long id = super.getTreeId(name);
      
      if (id == AstManager.INVALID_ID)
      {
         // try removing a well known suffix
         if (name.endsWith(AST_POSTFIX))
         {
            int len = name.length() - AST_POSTFIX_LEN;
            id = super.getTreeId(name.substring(0, len));
         }
      }
      
      return id;
   }
   
   /**
    * Load a complete AST by parsing the source XML tree and to obtain the
    * stored AST and annotation elements.  The XML tree is processed using a
    * top-down recursive procedure that creates an AST with identical
    * structure and data.
    * <p>
    * This method will synchronize on the object monitor of the given
    * <code>root</code> node.
    *
    * @param   root
    *          The root element of an AST document to be read and converted
    *          into an AST.
    *
    * @return   The loaded AST.
    *
    * @throws  AstException
    *          if the XML parser has a failure during parsing.
    */
   public static Aast loadTree(Element root)
   throws AstException,
          IllegalArgumentException
   {
      AnnotatedAst result = null;      
      Class        astCls = null;      
      boolean      terse  = false;
      
      synchronized (root)
      {
         try
         {         
            // read the AST class name out of the XML root and obtain the
            // class object
            String cname = root.getAttribute(ATTR_CLASS);
            astCls = Class.forName(cname);
            
            // determine whether we are in terse mode
            String terseAttr = root.getAttribute(ATTR_TERSE);
            terse = Boolean.TRUE.toString().equals(terseAttr);
         }
         
         catch (Exception excpt)
         {
            throw new AstException("Error reading AST XML tree.", excpt);
         }      
                           
         // the list of children will include #Text and other DOM defined
         // trash... get the first of our children
         Node ast = root.getFirstChild();
   
         // search for the child that has a tagName of ELEM_AST
         while (ast != null && ast.getNodeName().equals(ELEM_AST) == false)
         {
            ast = ast.getNextSibling();
         }
         
         // make sure something was found
         if (ast == null)
         {
            throw new AstException("Failure reading root element " +
                                     root.getNodeName());
         }
         
         // there must always be ONLY a single AST element under this
         // ast-root element!  recursively read the ASTs and build the tree...
         result = (AnnotatedAst) readAst((Element)ast, astCls, terse);
         
         // find any tree-local data container element, starting back at the
         // root node
         ast = root.getFirstChild();
         
         // search for the child that has a tagName of ELEM_LOCAL
         while (ast != null && ast.getNodeName().equals(ELEM_LOCAL) == false)
         {
            ast = ast.getNextSibling();
         }
         
         // read the tree-local data, IF it exists (it is OK to have none)
         if (ast != null)
         {
            result.setLocal(readTreeLocal((Element)ast, astCls, result, terse));
         }
         
         Long id = result.getId();
         
         if (id != null)
         {
            // this file *should* have already been registered once, so we try
            // to get the file ID using the stored node ID of the root node
            id = AstManager.get().getTreeId(id);
         }
         
         // makes sure that parent references (and IDs) are properly assigned,
         // note that the file ID may be null if the ID is null... this is OK 
         // though the IDs just won't ever be set
         result.fixups((Aast)null, id);
      }
      
      return result;  
   }
   
   /**
    * Load a named AST from persistent storage.
    *
    * @param    name
    *           The name of the tree. Often this is a file or resource name.
    *
    * @return   The loaded AST.
    *
    * @throws   AstException
    *           On any error during the loading process.
    * @throws   IllegalArgumentException
    *           If the name is invalid or the specified resource doesn't exist.
    */
   public Aast loadTree(String name)
   throws AstException,
          IllegalArgumentException
   {
      Element  root = null;
      Document dom  = null;
      
      try
      {
         if (useResources)
         {
            InputStream stream = ClassLoader.getSystemClassLoader().getResourceAsStream(name);
            if (stream != null)
            {
               dom = XmlHelper.parse(stream);
            }
         }
         else
         {
            File file = new File(name);
            if (file.exists() && file.isFile())
            {
               dom  = XmlHelper.parse(file);
            }
         }

         if (dom == null)
            throw new IllegalArgumentException("Missing/bogus input file/resource " + name);

         root = dom.getDocumentElement();
      }
      catch (Exception excpt)
      {
         throw new AstException(
            "Error loading AST XML file/resource: " + name, excpt);
      }      
      
      return loadTree(root);
   }
   
   /**
    * Save the given AST to persistent storage under the given name.
    *
    * @param    ast
    *           The tree to save.
    * @param    name
    *           The name of the tree. Often this is a file name.
    * @param    terse
    *           Do not store line and column data when the node data is saved.
    * 
    * @throws   AstException
    *           On any error during the saving process.
    * @throws   IllegalArgumentException
    *           If the name is invalid or the specified resource cannot be
    *           used as a target to store the AST.
    */
   public void saveTree(Aast ast, String name, boolean terse)
   throws AstException,
          IllegalArgumentException
   {
      if (ast == null)
         throw new IllegalArgumentException("Cannot persist null AST.");
      
      File file = new File(name);
      
      // create path if necessary
      File dir = file.getParentFile();
      if (!dir.exists())
      {
         dir.mkdirs();
      }
      
      Document dom = null;
      
      try
      {
         dom = XmlHelper.newDocument();
      }
      catch (ParserConfigurationException exc)
      {
         throw new AstException("Failure to create new XML document.", exc);
      }
      
      // add comments
      Comment comment = dom.createComment("Persisted AST");
      dom.appendChild(comment);
      
      // create root element
      Element elem = dom.createElement(ELEM_ROOT);
      dom.appendChild(elem);
      
      // write the AST class name into an attribute of the root node
      elem.setAttribute(ATTR_CLASS, ast.getClass().getName());
      
      // if condensed format, note this as an attribute of the root element.
      if (terse)
      {
         elem.setAttribute(ATTR_TERSE, Boolean.TRUE.toString());
      }
      
      // recursively write the AST out to this document, as children of 
      // this root element
      writeAst(ast, dom, (Node)elem, ELEM_AST, terse);
      
      // write out the tree-local data
      writeTreeLocal(ast, dom, (Node)elem, terse);
      
      // persist the XML document as a file
      persist(dom, new File(name));
   }
   
   /**
    * Check if the AST specified by the given name exists already in the plugin's storage.
    * 
    * @param    name
    *           The name (including path) of the AST file to be checked.
    * 
    * @return   <code>true</code> if the specified AST is in the plugin's storage.
    */
   public boolean isExistingAst(String name)
   {
      return new File(name).exists();
   }
   
   /**
    * Create the target XML document, add comments and a timestamp, create
    * a map element for every file and persist the XML file to the file
    * system.
    *
    * @throws   AstException
    *           If it is not possible to create an empty XML document or in    
    *           the case that the XML document cannot be stored in the file
    *           system or if the contents of the maps to persist are 
    *           <code>null</code>.
    */
   public synchronized void save()
   throws AstException
   {      
      Document dom = null;
      
      // check if there is anything to do
      if (fids.size() == 0 || !dirty || file == null)
         return;
      
      try
      {
         dom = XmlHelper.newDocument();
      }
      catch (ParserConfigurationException exc)
      {
         throw new AstException("Failure to create new XML document.", exc);
      }
      
      // add comments
      Comment comment = dom.createComment("AST Registry");
      dom.appendChild(comment);
      
      // generate timestamp
      comment = dom.createComment("Generated " + (new Date()));
      dom.appendChild(comment);
      
      // create root element
      Element root = dom.createElement(ELEM_REGISTRY_ROOT);
      dom.appendChild(root);
      
      // write map elements out to this document, as children of this root 
      // element
      Iterator<Long> ids = fids.keySet().iterator();
      
      while (ids.hasNext())
      {
         Long current = ids.next();
         Pair pair    = fids.get(current);
         Long next    = new Long(pair.next);
         
         Element map = dom.createElement(ELEM_MAP);
         
         map.setAttribute(ATTR_ID,   current.toString());
         map.setAttribute(ATTR_TREE, pair.name);
         map.setAttribute(ATTR_NEXT, next.toString());
         
         root.appendChild(map);
      }
      
      // persist the XML document as a file
      persist(dom, file);
      
      // turn off dirty flag
      dirty = false;
   }
            
   /**
    * Read the source XML document and build the mapping database from
    * the stored elements.
    *
    * @throws   AstException
    *           If the XML parser has a failure during parsing or in its    
    *           file system operations.
    * @throws   IllegalArgumentException
    *           If the input file is invalid or doesn't exist.
    */
   private synchronized void load()
   throws AstException
   {      
      Document dom    = null;
      Element  root   = null;
      
      // TODO: if dirty, log that unsaved modifications exist
      
      try
      {  
         if (file.exists() && file.isFile())
         {
            dom  = XmlHelper.parse(file);
            root = dom.getDocumentElement();            
         }
         else
         {
            throw new AstException("Bogus registry file " + file.toString());
         }
      }
      catch (Exception excpt)
      {
         throw new AstException("Error loading registry XML file: " +
                                file.toString(),
                                excpt);
      }      
      
      NodeList pairs = root.getElementsByTagName(ELEM_MAP);

      int len = pairs.getLength();
      
      // if we are reloading, we need to reset state
      fnames.clear();
      fids.clear();
      dirty = false;
      
      for (int i = 0; i < len; i++)
      {
         Element elem = (Element) pairs.item(i);
         
         Long   id    = new Long(elem.getAttribute(ATTR_ID));
         Long   next  = new Long(elem.getAttribute(ATTR_NEXT));         
         String fname = elem.getAttribute(ATTR_TREE);
         Pair   pair  = new Pair(fname, next.longValue());
         
         fnames.put(fname, id);
         fids.put(id, pair);
      }
   }
   
   /**
    * Processes an AST element in an XML document to create an associated AST
    * instance and then recursively calls itself for each AST element that
    * is a child.  The returned AST is then added as a child of the current
    * AST and when all children are processed, the current AST is returned.
    *
    * @param    current
    *           Represents an element in the XML document that is an AST
    *           element and may contain 0 or more AST elements.           
    * @param    astCls
    *           The class of the AnnotatedAst subclass used for the instance.
    * @param    terse
    *           <code>true</code> if line and column data should not be read
    *           from XML elements (because it doesn't exist).
    *
    * @return   An Aast sub-tree that has the identical structure and data
    *           as the XML element passed as a parameter.
    */
   private static Aast readAst(Element current, Class astCls, boolean terse)
   throws AstException
   {
      // create an AST node for the current XML node being processed
      Aast parent = createAst(current, astCls, terse);
      
      // read first child
      Node ast = current.getFirstChild();
      
      // for each child, create a subtree and then attach it to ourselves
      while (ast != null)
      {
         if (ast.getNodeName().equals(ELEM_AST))
         {
            parent.addChild(readAst((Element)ast, astCls, terse));         
         }
         
         ast = ast.getNextSibling();
      }
      
      return parent;
   }
   
   /**
    * Reads an AST associated with the passed-in XML element and creates a
    * new AST based on the associated attributes and annotations.
    *
    * @param    current
    *           The XML AST element.
    * @param    astCls
    *           The class of the AnnotatedAst subclass used for the instance.
    * @param    terse
    *           <code>true</code> if line and column data should not be read
    *           from XML elements (because it doesn't exist).
    *
    * @return   The created AST.
    *
    * @throws   AstException
    *           if an unknown data type is encountered.
    */
   private static Aast createAst(Element current, Class astCls, boolean terse)
   throws AstException
   {
      AnnotatedAst ast = null;
      
      try
      {
         // create the AST
         ast = (AnnotatedAst) astCls.newInstance();         
         
         // read the associated data
         String type  = current.getAttribute(ATTR_TTYPE);
         int    ttype = ast.lookupTokenType(type);
         
         // special processing for the text, since this can have embedded
         // control chars that can't be directly represented in XML; this        
         // helper transparently handles this special case, where the entire
         // string is encoded in a safe manner, with each char as a 2 char
         // hexidecimal string and this is detected and coverted for us
         String text = XmlHelper.getAttribute(current, ATTR_TEXT);
         
         // create a token
         CommonToken tok = new CommonToken(ttype, text);
         
         // determine whether we are a hidden node (if this attribute doesn't
         // exist, this resolves to false)
         String  hiddenAttr = current.getAttribute(ATTR_HIDDEN);
         boolean hidden     = Boolean.TRUE.toString().equals(hiddenAttr);         
         
         if (!terse)
         {
            tok.setLine( Integer.parseInt(current.getAttribute(ATTR_LINE)) );
            tok.setColumn( Integer.parseInt(current.getAttribute(ATTR_COL)) );
         }
         
         // "install" our token inside the AST (this inserts the line/col
         // nums and the token type...
         ast.initialize(tok);
         
         // set the ID
         Long id = new Long(current.getAttribute(ATTR_ID));
         
         // the ID is persisted as 0 if it was null, only set the ID if it
         // is non-zero
         if (id.longValue() != 0)
         {
            ast.setId(id);
         }
         
         // set our hidden attribute
         ast.setHidden(hidden);
         
         // read annotations and store them in the AST
         readAnnotations(current, ast);
      }
      
      catch (Exception exc)
      {
         throw new AstException("Failure to create AST for " +
                                 current.getTagName(),
                                 exc);
      }
      
      return ast;
   }
   
   /**
    * Reads any annotations associated with the passed-in XML element and
    * stores these in the AST.  Each valid annotation will be converted into
    * the proper object wrapper based on the class name stored in the data 
    * type attribute.
    *
    * @param    current
    *           The XML element which may have annotations to read and
    *           store.
    * @param    ast
    *           The AST in which to store the annotations.
    *
    * @throws   AstException
    *           if an unknown data type is encountered.
    */
   private static void readAnnotations(Element current, Aast ast)
   throws AstException
   {
      // read first child
      Node ann = current.getFirstChild();

      // for each child, create a subtree and then attach it to ourselves
      while (ann != null)
      {
         if (ann.getNodeName().equals(ELEM_ANN))
         {         
            Element elem = (Element)ann;
            
            // read the annotation's data and type
            String  cls  = elem.getAttribute(ATTR_DTYPE);
            String  key  = elem.getAttribute(ATTR_KEY);
            
            if ("java.util.ArrayList".equals(cls))
            {
               readListElements(elem, ast, key);
            }
            else
            {
               // handle decoding binary data transparently               
               String val = XmlHelper.getAttribute(elem, ATTR_VAL);
               
               // instantiate the Object wrapper for the annotation based on the
               // stored class name, then store the annotation in the AST
               if ("java.lang.String".equals(cls))
               {
                  ast.putAnnotation(key, val);
               }
               else if ("java.lang.Long".equals(cls))
               {
                  ast.putAnnotation(key, new Long(val));
               }
               else if ("java.lang.Boolean".equals(cls))
               {
                  Boolean b = new Boolean(val);
                  ast.putAnnotation(key, (b ? Boolean.TRUE : Boolean.FALSE));
               }
               else if ("java.lang.Double".equals(cls))
               {
                  ast.putAnnotation(key, new Double(val));
               }
               else
               {
                  throw new AstException("Invalid annotation data type " + cls);
               }
            }            
         }      
         
         ann = ann.getNextSibling();
      }         
   }
   
   /**
    * Reads the elements of an array annotation associated with the passed-in
    * XML element and stores these in the AST.  Each valid annotation will be
    * converted into the proper object wrapper based on the class name stored
    * in the data type attribute. Annotation lists support <code>null</code>
    * values.
    *
    * @param    list
    *           The XML element which may have annotations to read and
    *           store.
    * @param    ast
    *           The AST in which to store the annotations.
    * @param    key
    *           The annotation's name.
    *
    * @throws   AstException
    *           if an unknown data type is encountered.
    */
   private static void readListElements(Element list, Aast ast, String key)
   throws AstException
   {
      Node listitem = list.getFirstChild();
      
      while (listitem != null)
      {
         if (listitem.getNodeName().equals(ELEM_LISTITEM))
         {         
            Element listelem = (Element)listitem;
            
            // read the annotation's data and type
            String  icls = listelem.getAttribute(ATTR_DTYPE);
            
            // handle decoding binary data transparently
            String  ival = XmlHelper.getAttribute(listelem, ATTR_VAL);
            
            // instantiate the Object wrapper for the annotation based on the
            // stored class name, then store the annotation in the AST
            if ("java.lang.String".equals(icls))
            {
               ast.putAnnotation(key, ival, -1);
            }
            else if ("java.lang.Long".equals(icls))
            {
               ast.putAnnotation(key, new Long(ival), -1);
            }
            else if ("java.lang.Boolean".equals(icls))
            {
               ast.putAnnotation(key, new Boolean(ival), -1);
            }
            else if ("java.lang.Double".equals(icls))
            {
               ast.putAnnotation(key, new Double(ival), -1);
            }
            else if ("null".equals(icls))
            {
               // it doesn't really matter what class it is and we have
               // to pick one...
               ast.putAnnotation(key, (Long) null, -1);
            }
            else
            {
               throw new AstException("Invalid annotation data type " +
                                       icls);
            }
         }
         
         listitem = listitem.getNextSibling();
      }
   }
   
   /**
    * Processes a tree-local element in an XML document to create an associated
    * shadow node instances from the contained list.
    *
    * @param    data
    *           Represents an element in the XML document that is the
    *           tree-local element.           
    * @param    astCls
    *           The class of the AnnotatedAst subclass used for the instance.
    * @param    root
    *           The root node for the tree, which will be set as the parent
    *           of each shadow node.
    * @param    terse
    *           <code>true</code> if line and column data should not be read
    *           from XML elements (because it doesn't exist).
    *
    * @return   A tree-local instance.
    */
   private static TreeLocal readTreeLocal(Element data,
                                          Class   astCls,
                                          Aast    root,
                                          boolean terse)
   throws AstException
   {
      TreeLocal local = new TreeLocal();
      
      // read first child
      Node ast = data.getFirstChild();
      
      // find the shadow node list
      while (ast != null && !ast.getNodeName().equals(ELEM_SHADOW_LIST))
      {
         ast = ast.getNextSibling();
      }
      
      if (ast != null)
      {
         // read the first shadow node list element
         ast = ast.getFirstChild();
         
         // process the list of shadow nodes
         while (ast != null)
         {
            if (ast.getNodeName().equals(ELEM_SHADOW))
            {
               ShadowNode sn = new ShadowNode();
               
               Element elem = (Element) ast; 
               
               Aast newNode = createAst(elem, astCls, terse);
               newNode.setParent(root);
               sn.setNode(newNode);
               
               if (elem.hasAttribute(ATTR_LEFT_ID))
               {
                  sn.setLeft(new Long(elem.getAttribute(ATTR_LEFT_ID)));
               }
               
               if (elem.hasAttribute(ATTR_RIGHT_ID))
               {
                  sn.setRight(new Long(elem.getAttribute(ATTR_RIGHT_ID)));
               }
               
               local.addShadowNode(sn);
            }
               
            ast = ast.getNextSibling();
         }
      }
      
      return local;
   }
   
   /**                    
    * Write an AST node to XML, recursively calling itself to write all
    * children.  All data is stored as attributes and any annotations are
    * stored as contained annotation elements.
    *
    * @param    ast
    *           AST node to persist.
    * @param    dom
    *           The document being written.
    * @param    parent
    *           Parent node in the XML document, to which the associated
    *           element should be appended as a child.
    * @param    elemName
    *           Element name.
    * @param    terse
    *           Do not store line and column data when the node data is saved.
    *
    * @return   The root element of the AST subtree created.
    */
   private Element writeAst(Aast     ast,
                            Document dom,
                            Node     parent,
                            String   elemName,
                            boolean  terse)
   {
      // create an empty node of the proper element type 
      Element elem = dom.createElement(elemName);
      
      // read unique ID
      Long id = ast.getId();
      
      // if null, set to 0 which is an invalid value
      if (id == null)
      {
         id = new Long(0);
      }      
      
      // obtain token type as a human readable string
      String type = ast.getSymbolicTokenType();
      
      // store the AST's data as attributes
      elem.setAttribute(ATTR_ID, id.toString());
      elem.setAttribute(ATTR_TTYPE, type);
      
      // store the hidden attribute (if it is true, otherwise don't store it
      // since we will default it to false if it doesn't exist)
      if (ast.isHidden())
      {
         elem.setAttribute(ATTR_HIDDEN, "true");
      }
      
      if (!terse)
      {
         elem.setAttribute(ATTR_LINE, Integer.toString(ast.getLine()));
         elem.setAttribute(ATTR_COL, Integer.toString(ast.getColumn()));
      }
      
      // special processing for the text, since this can have embedded
      // control chars that can't be directly represented in XML
      String text = ast.getText();
            
      // special processing for the text, since this can have embedded
      // control chars that can't be directly represented in XML; this        
      // helper transparently handles this special case, where the entire
      // string is encoded in a safe manner, with each char as a 2 char
      // hexidecimal string and this is detected and coverted for us
      XmlHelper.setAttribute(elem, ATTR_TEXT, text);
      
      // set the AST as a child of the parent node
      parent.appendChild(elem);
      
      // create and write annotation nodes 
      writeAnnotations(ast, dom, (Node)elem);
      
      // recursively process all children
      Aast child = (Aast) ast.getFirstChild();
      while (child != null)
      {
         writeAst(child, dom, (Node)elem, elemName, terse);
         child = (Aast) child.getNextSibling();
      }
      
      return elem;
   }   
   
   /**
    * Process the list of annotations from an AST and write these into the
    * current XML document as child nodes.
    *
    * @param    ast
    *           AST containing the annotations to write.
    * @param    dom
    *           The document being written.
    * @param    parent
    *           Parent node in the XML document, to which the associated
    *           element should be appended as a child.
    */
   private void writeAnnotations(Aast ast, Document dom, Node parent)
   {
      Iterator list = ast.annotationKeys();
      String   key  = null;
      
      while (list.hasNext())
      {
         key = (String)list.next();
         
         // safety code to eliminate the "special" annotations that are used
         // internally at AST creation but which should never be persisted
         if ("hidden-before".equals(key) || "hidden-after".equals(key))
            continue;
         
         writeSingleAnnotation(ast, dom, parent, key);
      }      
   }
   
   /**
    * Write a single annotation (as identified by a single key) into the
    * current XML document as child nodes of the passed parent node. In the
    * case of array annotations, the list elements are written as elements
    * of the annotation element itself and no value attribute will occur
    * in the annotation.
    *
    * @param    ast
    *           AST containing the annotation to write.
    * @param    dom
    *           The document being written.
    * @param    parent
    *           Parent node in the XML document, to which the associated
    *           element should be appended as a child.
    * @param    key
    *           Specifies the annotation to write.
    */
   private void writeSingleAnnotation(Aast     ast,
                                      Document dom,
                                      Node     parent,
                                      String   key)
   {
      // create an annotation element
      Element elem = dom.createElement(ELEM_ANN);
      
      // read the annotation from the AST
      Object a = ast.getAnnotation(key);
      
      // throw NPE with some diagnostic info if annotation comes back null
      if (a == null)
      {
         StringBuilder buf = new StringBuilder();
         buf.append("null value for annotation '" + key + "':");
         buf.append(ast.dumpTree());
         
         throw new NullPointerException(buf.toString());
      }
      
      // WARNING: not thread safe, if another thread modified the annotation
      // list while we are running, we could get a null pointer exception...
      String cls = a.getClass().getName();
      
      // set its data as attributes
      elem.setAttribute(ATTR_DTYPE, cls);
      elem.setAttribute(ATTR_KEY, key);
      
      // handle arrays and scalar annotation values differently
      if (a instanceof List)
      {
         // write a list as child elements of the current annotation element
         writeListElements((List)a, dom, (Node)elem);
      }
      else
      {
         // write a scalar value using an attribute and let the helper
         // handle encoding binary data transparently
         XmlHelper.setAttribute(elem, ATTR_VAL, a.toString());
      }
      
      // set the annotation as a child of the parent node
      parent.appendChild(elem);
   }
      
   /**
    * Write a the list elements of an arry annotation (as identified by a
    * single key) into the current XML document as child nodes of the passed
    * parent node. Annotation lists support <code>null</code> values.
    *
    * @param    list
    *           Array annotation to write.
    * @param    dom
    *           The document being written.
    * @param    parent
    *           Parent node in the XML document, to which the associated
    *           element should be appended as a child.
    */
   private void writeListElements(List list, Document dom, Node parent)
   {
      Iterator items = list.iterator();
      
      while (items.hasNext())
      {
         // create an list item element
         Element elem = dom.createElement(ELEM_LISTITEM);
      
         // read the annotation element
         Object a = items.next();
         
         // WARNING: not thread safe, if another thread modified the
         // annotation while we are running, we could get a null pointer
         // exception...
         String cls   = null;
         String value = null;
         
         if (a == null)
         {
            // allow null values to be written out
            value = cls = "null";
         }
         else
         {
            cls = a.getClass().getName();
            value = a.toString();
         }
         
         // set its data as attributes
         elem.setAttribute(ATTR_DTYPE, cls);
         
         // handle encoding binary data transparently
         XmlHelper.setAttribute(elem, ATTR_VAL, value);
         
         // set the annotation as a child of the parent node
         parent.appendChild(elem);
      }      
   }
   
   /**                    
    * Write the tree-local data to XML.  Nothing is written if there is no
    * local data.
    *
    * @param    ast
    *           AST root node from which the tree-local data can be obtained.
    * @param    dom
    *           The document being written.
    * @param    parent
    *           Parent node in the XML document, to which the associated
    *           element should be appended as a child.
    * @param    terse
    *           Do not store line and column data when the node data is saved.
    */
   private void writeTreeLocal(Aast     ast, 
                               Document dom,
                               Node     parent,
                               boolean  terse)
   {
      TreeLocal local = ast.getLocal();
      
      if (local != null)
      {
         // create container element for local data as a child of the root node
         Element localElem = dom.createElement(ELEM_LOCAL);
         parent.appendChild(localElem);
         
         // create shadow node list element as a child of the local container
         Element elem = dom.createElement(ELEM_SHADOW_LIST);
         localElem.appendChild(elem);
         
         Iterator<ShadowNode> iter = local.getShadowNodeList();
         
         // write each shadow node into the list
         while (iter.hasNext())
         {
            ShadowNode sn = iter.next();
            
            // primary AST writing uses a common worker, it is already
            // parented properly when it is returned
            Element shadow = writeAst(sn.getNode(),
                                      dom,
                                      (Node) elem,
                                      ELEM_SHADOW,
                                      terse);
            
            // shadow specific data is written separately
            if (sn.getLeft() != null)
               shadow.setAttribute(ATTR_LEFT_ID, sn.getLeft().toString());
            if (sn.getRight() != null)
               shadow.setAttribute(ATTR_RIGHT_ID, sn.getRight().toString());
         }
      }
   }
   
   /**
    * Complete the processing by serializing the XML document to the output
    * file.
    *
    * @param    dom
    *           The XML document to persist.
    * @param    target
    *           The file in which to persist the XML document.
    *
    * @throws   AstException
    *           If any error occurs during XML serialization.
    */
   private void persist(Document dom, File target)
   throws AstException
   {
      try
      {
         // delete any previous version that may exist
         if (target.exists() && target.isFile())
         {
            target.delete();
         }
         
         XmlHelper.write(dom, target);
      }
      catch (Exception exc)
      {
         throw new AstException("Error serializing XML file " + 
                                target.toString(),
                                exc);
      }
   }
}