/*
** Module   : CacheExpiryEvent.java
** Abstract : Event indicating one or more element was expired from a cache
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140922 Created initial version.
** 002 ECF 20150715 Use LinkedHashMap for the expiredEntries map to ensure stable iteration
**                  order.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.cache;

import java.util.*;

/**
 * An event indicating one or more element was expired from a cache. Contains a reference to the
 * cache from which the event originated, and a map of the expired elements' keys to their values.
 */
public class CacheExpiryEvent<K, V>
{
   /** Cache object originating this event */
   private final ExpiryCache<K, V> source;
   
   /** Map of expired entries' keys to their values */
   private final LinkedHashMap<K, V> expiredEntries;
   
   /**
    * Constructor.
    * 
    * @param   source
    *          Cache object originating this event.
    * @param   expiredEntries
    *          Map of expired entries' keys to their values.
    */
   CacheExpiryEvent(ExpiryCache<K, V> source, LinkedHashMap<K, V> expiredEntries)
   {
      this.source = source;
      this.expiredEntries = expiredEntries;
   }
   
   /**
    * Get the source cache object which originated this event.
    * 
    * @return  See above.
    */
   public Cache<K, V> getSource()
   {
      return source;
   }
   
   /**
    * Get the map of expired entries' keys to their values.
    * 
    * @return  See above.
    */
   public LinkedHashMap<K, V> getExpiredEntries()
   {
      return expiredEntries;
   }
}