/*
** Module   : ExpiryCache.java
** Abstract : Abstract implementation of a fixed-size cache which expires less useful elements
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140925 Created initial version.
** 002 ECF 20150715 Minor change to internal map initialization.
** 003 ECF 20160519 Set initial capacity of hash map based on limit.
** 003 ECF 20160616 Javadoc clarification regarding external synchronization.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.cache;

import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.util.*;

/**
 * An abstract implementation of a cache with a fixed capacity, which expires one or more of its
 * elements when its capacity is reached and a new element is added to the cache. This class
 * provides the framework for such a cache in the form of an API to access the cache, but it is
 * up to concrete subclasses to implement the policy of a particular cache type.
 * <p>
 * Entities interested in expiry events can register themselves as {@link CacheExpiryListener
 * listeners} and they will receive an {@link CacheExpiryEvent event notification} when such
 * events occur.
 * <p>
 * <strong>This implementation is not synchronized</strong>. Access to any of the public methods
 * of this class (including {@link #get(Object)}) modify the internal state of the cache, and must
 * be synchronized externally.
 * 
 * @param   K
 *          Type of cache element key.
 * @param   V
 *          Type of cache element value.
 */
public abstract class ExpiryCache<K, V>
implements Cache<K, V>
{
   /** Logger */
   protected static final Logger log = LogHelper.getLogger(ExpiryCache.class);
   
   /** Cache name, for debug and logging purposes */
   protected final String name;
   
   /** Maximum number of elements allowed in the cache */
   protected final int limit;
   
   /** Map of keys to sortable, key-value nodes */
   protected final Map<K, Node<K, V>> cache;
   
   /** Sorted set of key-value nodes */
   protected final TreeSet<Node<K, V>> sorted = new TreeSet<>();
   
   /** Number of elements to clear from cache when it is full and capacity is needed */
   private final int cleanLimit;
   
   /** List of cache expiry listeners */
   private List<CacheExpiryListener<K, V>> listeners = null;
   
   /**
    * Constructor.
    * 
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   protected ExpiryCache(int limit)
   {
      this("", limit);
   }
   
   /**
    * Constructor.
    * 
    * @param   name
    *          Cache name, for debug and logging purposes.
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   protected ExpiryCache(String name, int limit)
   {
      if (limit < 1)
      {
         throw new IllegalArgumentException("Cannot create a cache with a limit < 1");
      }
      
      int capacity = (int) (((double) limit) / 0.75);
      this.cache = new HashMap<>(capacity);
      this.name = name;
      this.limit = limit;
      this.cleanLimit = getCleanLimit();
   }
   
   /**
    * Retrieve a value, given its key.
    * <p>
    * Note for multi-threaded access: this method makes changes to the state of the cache. A call
    * to this method should therefore be treated as write access for concurrency purposes.
    * 
    * @param   key
    *          A hashable key.
    * 
    * @return  Value associated with the given key, or <code>null</code> if there is no mapping
    *          for the given key.
    */
   public V get(K key)
   {
      Node<K, V> n = cache.get(key);
      
      // if we found a node, remove it from the sorted set, update its access state, then add it
      // back, to ensure the sort order is maintained
      if (n != null)
      {
         sorted.remove(n);
         V value = n.accessValue();
         sorted.add(n);
         
         return value;
      }
      
      return null;
   }
   
   /**
    * Add an entry to the cache, pushing out the least desirable element(s) if the cache is full,
    * and this entry does not replace an existing entry with the same key. This may result in a
    * {@link CacheExpiryEvent} being sent to registered listeners before this method returns.
    * 
    * @param   key
    *          A hashable key to associate with the value.
    * @param   value
    *          Value to store.
    * 
    * @return  The old value stored with the given key, or <code>null</code> if there was no
    *          mapping for the given key.
    */
   public Object put(K key, V value)
   {
      // create a new node and put it into the hash map
      Node<K, V> n = createNode(key, value);
      Node<K, V> old = cache.put(key, n);
      
      if (old == null)
      {
         // clear some room in the cache if it's full and we are not replacing an existing entry
         if (cache.size() == limit)
         {
            ensureCapacity();
         }
      }
      else
      {
         // we had an existing mapping in the hash map; remove its companion in the sorted set,
         // to keep the two in sync
         sorted.remove(old);
      }
      
      // add the new node to the sorted set
      sorted.add(n);
      
      // return the old value, if any
      return (old != null ? old.getValue() : null);
   }
   
   /**
    * Remove the entry associated with the given key from the cache.
    * 
    * @param   key
    *          A hashable key identifying the entry to remove.
    * 
    * @return  The value removed, or <code>null</code> if there was no mapping for the given key.
    */
   public V remove(K key)
   {
      Node<K, V> n = cache.remove(key);
      if (n != null)
      {
         // remove its companion entry in the sorted set, to keep the two in sync
         sorted.remove(n);
         
         return n.getValue();
      }
      
      return null;
   }
   
   /**
    * Add the given object to the list of listeners interested in receiving cache expiry events
    * from this cache. If the listener was added previously, it will not be added again.
    * Listeners will receive events in the order in which they were added.
    * <p>
    * Note that adding listeners (any number) to the cache will add a slight amount of overhead
    * to cache expiry processing, in that the expired elements are collected in a linked hash
    * map, to be sent with a cache expiry event. If no listeners are registered, this collection
    * does not take place.
    * 
    * @param   listener
    *          Listener to register.
    */
   public void addCacheExpiryListener(CacheExpiryListener<K, V> listener)
   {
      if (listeners ==  null)
      {
         listeners = new LinkedList<>();
      }
      
      if (!listeners.contains(listener))
      {
         listeners.add(listener);
      }
   }
   
   /**
    * Remove the given object from the list of listeners interested in receiving cache expiry
    * events from this cache. If the listener was not added previously, the request effects no
    * change and does not raise an error.
    * 
    * @param   listener
    *          Listener to deregister.
    */
   public void removeCacheExpiryListener(CacheExpiryListener<K, V> listener)
   {
      if (listeners != null)
      {
         listeners.remove(listener);
         if (listeners.isEmpty())
         {
            listeners = null;
         }
      }
   }
   
   /**
    * Get the number of elements which will be expired from the cache each time it becomes full,
    * and capacity is required for a new element.
    * 
    * @return  Number of elements to expire on each clearing pass.
    */
   protected abstract int getCleanLimit();
   
   /**
    * Create a node of the appropriate type for the given key and value.
    * 
    * @param   key
    *          Hashable key.
    * @param   value
    *          Cached value.
    * 
    * @return  A sortable node to store in the cache.
    */
   protected abstract Node<K, V> createNode(K key, V value);
   
   /**
    * Ensure there is sufficient capacity to add a new record to the cache, by expiring the least
    * desirable element(s), according to the policy of the cache.
    * <p>
    * When elements are expired from the cache, a cache expiry event is sent to all registered
    * listeners.
    */
   protected void ensureCapacity()
   {
      LinkedHashMap<K, V> expiredEntries = (listeners != null
                                            ? new LinkedHashMap<K, V>(((int) cleanLimit * 4 / 3))
                                            : null);
      
      Iterator<Node<K, V>> iter = sorted.iterator();
      for (int i = 0; iter.hasNext() && i < cleanLimit; i++)
      {
         Node<K, V> n = iter.next();
         K key = n.getKey();
         iter.remove();
         cache.remove(key);
         
         if (expiredEntries != null)
         {
            V value = n.getValue();
            expiredEntries.put(key, value);
         }
      }
      
      if (listeners != null)
      {
         CacheExpiryEvent<K, V> event = new CacheExpiryEvent<>(this, expiredEntries);
         for (CacheExpiryListener<K, V> l : listeners)
         {
            l.cacheExpired(event);
         }
      }
   }
   
   /**
    * A sortable node containing the hashable key and value of a cache entry. When an element of
    * the cache is accessed (created or retrieved), the abstract method {@link #access()} is
    * invoked. Subclasses of the enclosing class must subclass this inner class as part of
    * defining the policy of the cache. In particular {@link #compareTo(Object)} must be
    * implemented to sort less desirable elements (those to be expired first) ahead of more
    * desirable elements (those to be retained longer).
    * 
    * @param <K>
    *        Hashable key
    * @param <V>
    *        Cached value
    */
   protected abstract static class Node<K, V>
   implements Comparable<Node<K, V>>
   {
      /** Hashable key */
      private final K key;
      
      /** Cached value */
      private final V value;
      
      /**
       * Constructor. Sets initial access time to now, and hit count to 0.
       * 
       * @param   key
       *          Hashable key.
       * @param   value
       *          Cached value.
       */
      protected Node(K key, V value)
      {
         this.key = key;
         this.value = value;
         access();
      }
      
      /**
       * This method should be implemented to provide equivalence checking that is consistent
       * with a subclass' implementation of <code>compareTo(Node)</code>.
       * 
       * @param   o
       *          Object with which to compare this instance.
       * 
       * @return  <code>true</code> if <code>o</code> is equivalent to this instance, else
       *          <code>false</code>.
       */
      public abstract boolean equals(Object o);
      
      /**
       * Get the key.
       * 
       * @return  Key.
       */
      K getKey()
      {
         return key;
      }
      
      /**
       * Get the value.
       * 
       * @return  Value.
       */
      V getValue()
      {
         return value;
      }
      
      /**
       * Access the value, which updates the node with hit count and last access information.
       * 
       * @return  Value.
       */
      V accessValue()
      {
         access();
         
         return value;
      }
      
      /**
       * Increment the hit count for this node and record the timestamp of this access.
       */
      protected abstract void access();
   }
}