/*
** Module   : Attribute.java
** Abstract : 
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @xxxxx Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;


import java.io.*;


/**
 * This is the base class for all classes which represent attribute info
 * structures found within a class file.
 */
abstract class Attribute
{

   /** ConstantValue attribute name. */
   protected static final String CONSTANTVALUE = "ConstantValue";

   /** Code attribute name. */
   protected static final String CODE = "Code";

   /** Exceptions attribute name. */
   protected static final String EXCEPTIONS = "Exceptions";

   /** InnerClass attribute name. */
   protected static final String INNERCLASS = "InnerClass";

   /** Synthetic attribute name. */
   protected static final String SYNTHETIC = "Synthetic";

   /** SourceFile attribute name. */
   protected static final String SOURCEFILE = "SourceFile";

   /** LineNumberTable attribute name. */
   protected static final String LINENUMBERTABLE = "LineNumberTable";

   /** LocalVariableTable attribute name. */
   protected static final String LOCALVARIABLETABLE = "LocalVariableTable";

   /** Deprecated attribute name. */
   protected static final String DEPRECATED = "Deprecated";

   /** Type name of the current attribute. */
   private Utf8Constant nameConstant;


   /** 
    * Create a new instance of this class, given a string indicating the
    * attribute type.
    *
    * @param   pool
    *          Constant pool from which to retrieve the Utf8Constant which
    *          represents the name of the attribute.
    * @param   type
    *          One of the attribute type name constants defined by this
    *          class.
    */
   protected Attribute(ConstantPool pool, String type)
   {
      nameConstant = Utf8Constant.create(pool, type);
   }


   /** 
    * This method must be implemented to return the total number of bytes
    * in the subclass attribute structure.
    *
    * @return  Length of the subclass' attribute structure.
    */
   protected abstract int getLength();


   /** 
    * This method writes the data for this attribute to the specified
    * output stream.  The attribute type and length are written by this
    * method, then the abstract <code>writeContent()</code> is invoked to
    * write the subclass-specific data.
    *
    * @param   out
    *          Output stream to which data is written.
    *
    * @exception  IOException
    *             If any error is encountered while writing data.
    */
   void write(DataOutput out) throws IOException {

      out.writeShort(nameConstant.getCpIndex());
      out.writeInt(getLength());
      writeContent(out);

   }


   /** 
    * This method must be implemented to write the data content (not
    * including the header info - type and length) for the subclassed
    * attribute type.
    *
    * @param   out
    *          Output stream to which data content is written.
    *
    * @exception  IOException
    *             If any error is encountered while writing data.
    */
   protected abstract void writeContent(DataOutput out) throws IOException;


}