/*
** Module   : ExpressionClassLoader.java
** Abstract : 
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @xxxxx Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;


import java.io.*;
import java.util.*;


/**
 * This class extends <code>ClassLoader</code> to enable the loading of
 * a class file stored as a byte array in memory.  When a class file is
 * generated by the expression compiler, it should be stored under its
 * fully qualified class name using the <code>putClassFile()</code>
 * method.  When the class is subsequently loaded, the <code>ClassFile
 * </code> is retrieved and removed from this temporary cache, and the
 * byte array representing the class file is used to load the class
 * into the JVM.
 */
final class ExpressionClassLoader extends ClassLoader
{

   /** Cache of classes already loaded by this class loader. */
   private Map cache = new HashMap();

   /** Cache of class files created by expression compiler. */
   private Map dataMap = new HashMap();


   /** 
    * This method should be called by the class file compiler to
    * temporarily cache a compiled class in the form of a <code>ClassFile
    * </code> instance.  This object will be retrieved subsequently and
    * removed from this cache when it is necessary to load the class into
    * the JVM.
    *
    * @param   name
    *          Fully qualified name of the class to be cached.
    * @param   file
    *          ClassFile object to cache temporarily.
    */
   synchronized void putClassFile(String name, ClassFile file)
   {
      dataMap.put(name, file);
   }


   /** 
    * This method is used to retrieve the byte array containing the class
    * file information for the class with the specified name.
    *
    * @param   name
    *          Fully qualified name of class for which the class file
    *          information byte array is required.
    *
    * @return  Byte array containing class file information, or null if
    *          no class file is found for <code>name</code>.
    */
   private synchronized byte[] getClassData(String name)
   {

      ClassFile file = (ClassFile) dataMap.remove(name);
      if (file == null)
         return null;

      try
      {
         ByteArrayOutputStream baos = new ByteArrayOutputStream();
         BufferedOutputStream bos = new BufferedOutputStream(baos);
         DataOutputStream out = new DataOutputStream(bos);

         file.write(out);
         out.flush();
         byte[] data = baos.toByteArray();

         out.close();
         bos.close();
         baos.close();

         return data;
      }
      catch (IOException exc)
      {
         exc.printStackTrace();
      }

      return null;

   }


   /** 
    * This method is called by the JVM when the class with the specified
    * name is referenced and needs to be loaded.  The primordial class
    * loader is first given a chance to load the class, in case <code>
    * name</code> refers to a system class.  If this load fails, an attempt
    * is then made to find the class in the cache maintained by this class
    * loader.  If the class is not found in the cache, a lookup is
    * performed to find the byte array containing the class file information
    * for the class identified by <code>name</code>, and this byte array is
    * used to define the class.
    *
    * @param   name
    *          Fully qualified name of the class to be loaded.
    * @param   resolve
    *          True if class should be resolved, else false.
    *
    * @return  The <code>Class</code> object identified by <code>name</code>.
    *
    * @exception  ClassNotFoundException
    *             If no class can be found for the specified name.
    */
   protected synchronized Class loadClass(String name, boolean resolve)
   throws ClassNotFoundException
   {

      Class c = null;

      try
      {
         c = findSystemClass(name);
      }
      catch (Exception exc)
      {
         try
         {
            c = (Class) cache.get(name);
            if (c == null)
            {
               byte[] data = getClassData(name);
               c = defineClass(name, data, 0, data.length);
               cache.put(name, c);
            }
         }
         catch (Exception e)
         {
            exc.printStackTrace();
            throw new ClassNotFoundException(
               "[" + e.getClass().getName() + "]  " + e.getMessage());
         }
      }

      if (resolve)
         resolveClass(c);

      return c;

   }


}