/*
** Module   : Instruction.java
** Abstract : Represents a single opcode instruction and its arguments
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 ECF 20041130   @19660 Production version from JTA, reworked into a new package and with 
**                           minor cleanups in javadoc.
** 002 ECF 20050207   @19661 Reorganized static initializer for improved readability. Added 
**                           several opcodes to support compiler overhaul. Added debug code to
**                           print opcode name during fixupStack method. Since this method is 
**                           called for every instruction in order, just before the compiled
**                           expression class is loaded, this gives us a simple listing of
**                           instructions to stdout. Arguments are not included in this listing. 
**                           This code is enabled/disabled via the static debug variable.
** 003 CA  20140630          Always use ldc_w instead of ldc, otherwise the constant pool index 
**                           is limited to a max of 255.
** 004 ECF 20150715          Replace StringBuffer with StringBuilder.
**/

package com.goldencode.expr;

import java.io.*;
import java.util.*;

/**
 * Instances of this class represent individual bytecode instructions.
 * Bytecode instructions are contained within {@link CodeUnit}s, which
 * are created, populated, and organized by the {@link Compiler}.
 * <p>
 * Each instruction may affect the depth of the current JVM operand stack.
 * The impact of each instruction type on the stack is recorded in the
 * {@link #stackMap} variable as an array of integer values.  Each value
 * represents a single stack change caused by the instruction.  A negative
 * number indicates that a value is popped from the runtime stack;  a
 * positive number indicates that a value is pushed onto the runtime stack.
 * The magnitude of the number indicates the width of the value being
 * pushed or popped, with longs and doubles having a width of 2 and every
 * other type having a width of 1.  A single instruction may conduct
 * multiple operations on the runtime stack.  Each such operation is
 * represented by a separate number.  Some instructions vary in their
 * interaction with the stack, such as those that invoke methods, where
 * the method's signature dictates the size and number of arguments popped
 * from the stack and the return type (if any) pushed onto the stack. For
 * these types of instructions, a custom stack change array must be
 * calculated by the compiler and provided via the specialized {@link
 * #Instruction(int, int[])} constructor.
 * <p>
 * An instruction may have zero or more arguments. These are
 * recorded as a list. The length of the instruction is one (for the
 * bytecode instruction itself) plus the width of each of the arguments.
 */
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
public class Instruction
extends BytecodeContainer
{
   /** Opcode to instruction name map */
   private static Map OPCODE_NAMES = null;
   
   /** Map of stack depth change values, indexed by op code */
   private static Map stackMap = new HashMap();
   
   /** Op code for the instruction */
   private int opCode = -1;
   
   /** List of arguments to follow */
   private List args = null;
   
   /** Ordered array of stack depth changes caused by this instruction */
   private int[] stackChange = null;
   
   // Initialize the static stack depth map. Only record values for those
   // instructions which may be involved in compiled expressions.
   static
   {
      if (Compiler.debug)
      {
         OPCODE_NAMES = new HashMap();
         OPCODE_NAMES.put(new Integer(_aconst_null)    , "aconst_null    ");
         OPCODE_NAMES.put(new Integer(_iconst_0)       , "iconst_0       ");
         OPCODE_NAMES.put(new Integer(_iconst_1)       , "iconst_1       ");
         OPCODE_NAMES.put(new Integer(_iconst_2)       , "iconst_2       ");
         OPCODE_NAMES.put(new Integer(_iconst_3)       , "iconst_3       ");
         OPCODE_NAMES.put(new Integer(_iconst_4)       , "iconst_4       ");
         OPCODE_NAMES.put(new Integer(_iconst_5)       , "iconst_5       ");
         OPCODE_NAMES.put(new Integer(_lconst_0)       , "lconst_0       ");
         OPCODE_NAMES.put(new Integer(_lconst_1)       , "lconst_1       ");
         OPCODE_NAMES.put(new Integer(_dconst_0)       , "dconst_0       ");
         OPCODE_NAMES.put(new Integer(_dconst_1)       , "dconst_1       ");
         OPCODE_NAMES.put(new Integer(_ldc_w)          , "ldc_w          ");
         OPCODE_NAMES.put(new Integer(_ldc2_w)         , "ldc2_w         ");
         OPCODE_NAMES.put(new Integer(_aload_0)        , "aload_0        ");
         OPCODE_NAMES.put(new Integer(_aload_1)        , "aload_1        ");
         OPCODE_NAMES.put(new Integer(_aload_2)        , "aload_2        ");
         OPCODE_NAMES.put(new Integer(_aload_3)        , "aload_3        ");
         OPCODE_NAMES.put(new Integer(_astore_1)       , "astore_1       ");
         OPCODE_NAMES.put(new Integer(_astore_2)       , "astore_2       ");
         OPCODE_NAMES.put(new Integer(_astore_3)       , "astore_3       ");
         OPCODE_NAMES.put(new Integer(_aastore)        , "aastore        ");
         OPCODE_NAMES.put(new Integer(_pop)            , "pop            ");
         OPCODE_NAMES.put(new Integer(_pop2)           , "pop2           ");
         OPCODE_NAMES.put(new Integer(_dup)            , "dup            ");
         OPCODE_NAMES.put(new Integer(_swap)           , "swap           ");
         OPCODE_NAMES.put(new Integer(_iadd)           , "iadd           ");
         OPCODE_NAMES.put(new Integer(_ladd)           , "ladd           ");
         OPCODE_NAMES.put(new Integer(_dadd)           , "dadd           ");
         OPCODE_NAMES.put(new Integer(_isub)           , "isub           ");
         OPCODE_NAMES.put(new Integer(_lsub)           , "lsub           ");
         OPCODE_NAMES.put(new Integer(_dsub)           , "dsub           ");
         OPCODE_NAMES.put(new Integer(_imul)           , "imul           ");
         OPCODE_NAMES.put(new Integer(_lmul)           , "lmul           ");
         OPCODE_NAMES.put(new Integer(_dmul)           , "dmul           ");
         OPCODE_NAMES.put(new Integer(_idiv)           , "idiv           ");
         OPCODE_NAMES.put(new Integer(_ldiv)           , "ldiv           ");
         OPCODE_NAMES.put(new Integer(_ddiv)           , "ddiv           ");
         OPCODE_NAMES.put(new Integer(_ineg)           , "ineg           ");
         OPCODE_NAMES.put(new Integer(_lneg)           , "lneg           ");
         OPCODE_NAMES.put(new Integer(_dneg)           , "dneg           ");
         OPCODE_NAMES.put(new Integer(_irem)           , "irem           ");
         OPCODE_NAMES.put(new Integer(_lrem)           , "lrem           ");
         OPCODE_NAMES.put(new Integer(_drem)           , "drem           ");
         OPCODE_NAMES.put(new Integer(_ishl)           , "ishl           ");
         OPCODE_NAMES.put(new Integer(_lshl)           , "lshl           ");
         OPCODE_NAMES.put(new Integer(_ishr)           , "ishr           ");
         OPCODE_NAMES.put(new Integer(_lshr)           , "lshr           ");
         OPCODE_NAMES.put(new Integer(_iushr)          , "iushr          ");
         OPCODE_NAMES.put(new Integer(_lushr)          , "lushr          ");
         OPCODE_NAMES.put(new Integer(_iand)           , "iand           ");
         OPCODE_NAMES.put(new Integer(_land)           , "land           ");
         OPCODE_NAMES.put(new Integer(_ior)            , "ior            ");
         OPCODE_NAMES.put(new Integer(_lor)            , "lor            ");
         OPCODE_NAMES.put(new Integer(_ixor)           , "ixor           ");
         OPCODE_NAMES.put(new Integer(_lxor)           , "lxor           ");
         OPCODE_NAMES.put(new Integer(_i2l)            , "i2l            ");
         OPCODE_NAMES.put(new Integer(_i2d)            , "i2d            ");
         OPCODE_NAMES.put(new Integer(_l2i)            , "l2i            ");
         OPCODE_NAMES.put(new Integer(_l2d)            , "l2d            ");
         OPCODE_NAMES.put(new Integer(_d2i)            , "d2i            ");
         OPCODE_NAMES.put(new Integer(_d2l)            , "d2l            ");
         OPCODE_NAMES.put(new Integer(_lcmp)           , "lcmp           ");
         OPCODE_NAMES.put(new Integer(_dcmpg)          , "dcmpg          ");
         OPCODE_NAMES.put(new Integer(_ifeq)           , "ifeq           ");
         OPCODE_NAMES.put(new Integer(_ifne)           , "ifne           ");
         OPCODE_NAMES.put(new Integer(_iflt)           , "iflt           ");
         OPCODE_NAMES.put(new Integer(_ifge)           , "ifge           ");
         OPCODE_NAMES.put(new Integer(_ifgt)           , "ifgt           ");
         OPCODE_NAMES.put(new Integer(_ifle)           , "ifle           ");
         OPCODE_NAMES.put(new Integer(_if_icmpeq)      , "if_icmpeq      ");
         OPCODE_NAMES.put(new Integer(_if_icmpne)      , "if_icmpne      ");
         OPCODE_NAMES.put(new Integer(_if_icmplt)      , "if_icmplt      ");
         OPCODE_NAMES.put(new Integer(_if_icmpge)      , "if_icmpge      ");
         OPCODE_NAMES.put(new Integer(_if_icmpgt)      , "if_icmpgt      ");
         OPCODE_NAMES.put(new Integer(_if_icmple)      , "if_icmple      ");
         OPCODE_NAMES.put(new Integer(_goto)           , "goto           ");
         OPCODE_NAMES.put(new Integer(_ireturn)        , "ireturn        ");
         OPCODE_NAMES.put(new Integer(_dreturn)        , "dreturn        ");
         OPCODE_NAMES.put(new Integer(_areturn)        , "areturn        ");
         OPCODE_NAMES.put(new Integer(_return)         , "return         ");
         OPCODE_NAMES.put(new Integer(_getfield)       , "getfield       ");
         OPCODE_NAMES.put(new Integer(_invokevirtual)  , "invokevirtual  ");
         OPCODE_NAMES.put(new Integer(_invokespecial)  , "invokespecial  ");
         OPCODE_NAMES.put(new Integer(_invokestatic)   , "invokestatic   ");
         OPCODE_NAMES.put(new Integer(_invokeinterface), "invokeinterface");
         OPCODE_NAMES.put(new Integer(_new)            , "new            ");
         OPCODE_NAMES.put(new Integer(_anewarray)      , "anewarray      ");
         OPCODE_NAMES.put(new Integer(_checkcast)      , "checkcast      ");
         OPCODE_NAMES.put(new Integer(_ifnull)         , "ifnull         ");
         OPCODE_NAMES.put(new Integer(_ifnonnull)      , "ifnonnull      ");
      }
      
      // Stack depth changes.
      int[] __aconst_null = {  1                     };
      int[] __iconst_0    = {  1                     };
      int[] __iconst_1    = {  1                     };
      int[] __iconst_2    = {  1                     };
      int[] __iconst_3    = {  1                     };
      int[] __iconst_4    = {  1                     };
      int[] __iconst_5    = {  1                     };
      int[] __lconst_0    = {  2                     };
      int[] __lconst_1    = {  2                     };
      int[] __dconst_0    = {  2                     };
      int[] __dconst_1    = {  2                     };
      int[] __ldc_w       = {  1                     };
      int[] __ldc2_w      = {  2                     };
      int[] __aload_0     = {  1                     };
      int[] __aload_1     = {  1                     };
      int[] __aload_2     = {  1                     };
      int[] __aload_3     = {  1                     };
      int[] __astore_1    = { -1                     };
      int[] __astore_2    = { -1                     };
      int[] __astore_3    = { -1                     };
      int[] __aastore     = { -1 , -1 , -1           };
      int[] __pop         = { -1                     };
      int[] __pop2        = { -2                     };
      int[] __dup         = { -1 ,  1 ,  1           };
      int[] __swap        = { -1 , -1 ,  1 ,  1      };
      int[] __iadd        = { -1 , -1 ,  1           };
      int[] __ladd        = { -2 , -2 ,  2           };
      int[] __dadd        = { -2 , -2 ,  2           };
      int[] __isub        = { -1 , -1 ,  1           };
      int[] __lsub        = { -2 , -2 ,  2           };
      int[] __dsub        = { -2 , -2 ,  2           };
      int[] __imul        = { -1 , -1 ,  1           };
      int[] __lmul        = { -2 , -2 ,  2           };
      int[] __dmul        = { -2 , -2 ,  2           };
      int[] __idiv        = { -1 , -1 ,  1           };
      int[] __ldiv        = { -2 , -2 ,  2           };
      int[] __ddiv        = { -2 , -2 ,  2           };
      int[] __ineg        = { -1 ,  1                };
      int[] __lneg        = { -2 ,  2                };
      int[] __dneg        = { -2 ,  2                };
      int[] __irem        = { -1 , -1 ,  1           };
      int[] __lrem        = { -2 , -2 ,  2           };
      int[] __drem        = { -2 , -2 ,  2           };
      int[] __ishl        = { -1 , -1 ,  1           };
      int[] __lshl        = { -1 , -2 ,  2           };
      int[] __ishr        = { -1 , -1 ,  1           };
      int[] __lshr        = { -1 , -2 ,  2           };
      int[] __iushr       = { -1 , -1 ,  1           };
      int[] __lushr       = { -1 , -2 ,  2           };
      int[] __iand        = { -1 , -1 ,  1           };
      int[] __land        = { -2 , -2 ,  2           };
      int[] __ior         = { -1 , -1 ,  1           };
      int[] __lor         = { -2 , -2 ,  2           };
      int[] __ixor        = { -1 , -1 ,  1           };
      int[] __lxor        = { -2 , -2 ,  2           };
      int[] __i2l         = { -1 ,  2                };
      int[] __i2d         = { -1 ,  2                };
      int[] __l2i         = { -2 ,  1                };
      int[] __l2d         = { -2 ,  2                };
      int[] __d2i         = { -2 ,  1                };
      int[] __d2l         = { -2 ,  2                };
      int[] __lcmp        = { -2 , -2 ,  1           };
      int[] __dcmpg       = { -2 , -2 ,  1           };
      int[] __ifeq        = { -1                     };
      int[] __ifne        = { -1                     };
      int[] __iflt        = { -1                     };
      int[] __ifge        = { -1                     };
      int[] __ifgt        = { -1                     };
      int[] __ifle        = { -1                     };
      int[] __if_icmpeq   = { -1 , -1                };
      int[] __if_icmpne   = { -1 , -1                };
      int[] __if_icmplt   = { -1 , -1                };
      int[] __if_icmpge   = { -1 , -1                };
      int[] __if_icmpgt   = { -1 , -1                };
      int[] __if_icmple   = { -1 , -1                };
      int[] __goto        = {  0                     };
      int[] __ireturn     = { -1                     };
      int[] __dreturn     = { -2                     };
      int[] __areturn     = { -1                     };
      int[] __return      = {  0                     };
      int[] __getfield    = { -1 , 1                 };
      int[] __new         = {  1                     };
      int[] __anewarray   = { -1 , 1                 };
      int[] __checkcast   = { -1 , 1                 };
      int[] __ifnull      = { -1                     };
      int[] __ifnonnull   = { -1                     };
      
      // Mappings of bytecode constants to stack depth changes.
      stackMap.put(new Integer(_aconst_null) , __aconst_null);
      stackMap.put(new Integer(_iconst_0)    , __iconst_0   );
      stackMap.put(new Integer(_iconst_1)    , __iconst_1   );
      stackMap.put(new Integer(_iconst_2)    , __iconst_2   );
      stackMap.put(new Integer(_iconst_3)    , __iconst_3   );
      stackMap.put(new Integer(_iconst_4)    , __iconst_4   );
      stackMap.put(new Integer(_iconst_5)    , __iconst_5   );
      stackMap.put(new Integer(_lconst_0)    , __lconst_0   );
      stackMap.put(new Integer(_lconst_1)    , __lconst_1   );
      stackMap.put(new Integer(_dconst_0)    , __dconst_0   );
      stackMap.put(new Integer(_dconst_1)    , __dconst_1   );
      stackMap.put(new Integer(_ldc_w)       , __ldc_w      );
      stackMap.put(new Integer(_ldc2_w)      , __ldc2_w     );
      stackMap.put(new Integer(_aload_0)     , __aload_0    );
      stackMap.put(new Integer(_aload_1)     , __aload_1    );
      stackMap.put(new Integer(_aload_2)     , __aload_2    );
      stackMap.put(new Integer(_aload_3)     , __aload_3    );
      stackMap.put(new Integer(_astore_1)    , __astore_1   );
      stackMap.put(new Integer(_astore_2)    , __astore_2   );
      stackMap.put(new Integer(_astore_3)    , __astore_3   );
      stackMap.put(new Integer(_aastore)     , __aastore    );
      stackMap.put(new Integer(_pop)         , __pop        );
      stackMap.put(new Integer(_pop2)        , __pop2       );
      stackMap.put(new Integer(_dup)         , __dup        );
      stackMap.put(new Integer(_swap)        , __swap       );
      stackMap.put(new Integer(_iadd)        , __iadd       );
      stackMap.put(new Integer(_ladd)        , __ladd       );
      stackMap.put(new Integer(_dadd)        , __dadd       );
      stackMap.put(new Integer(_isub)        , __isub       );
      stackMap.put(new Integer(_lsub)        , __lsub       );
      stackMap.put(new Integer(_dsub)        , __dsub       );
      stackMap.put(new Integer(_imul)        , __imul       );
      stackMap.put(new Integer(_lmul)        , __lmul       );
      stackMap.put(new Integer(_dmul)        , __dmul       );
      stackMap.put(new Integer(_idiv)        , __idiv       );
      stackMap.put(new Integer(_ldiv)        , __ldiv       );
      stackMap.put(new Integer(_ddiv)        , __ddiv       );
      stackMap.put(new Integer(_ineg)        , __ineg       );
      stackMap.put(new Integer(_lneg)        , __lneg       );
      stackMap.put(new Integer(_dneg)        , __dneg       );
      stackMap.put(new Integer(_irem)        , __irem       );
      stackMap.put(new Integer(_lrem)        , __lrem       );
      stackMap.put(new Integer(_drem)        , __drem       );
      stackMap.put(new Integer(_ishl)        , __ishl       );
      stackMap.put(new Integer(_lshl)        , __lshl       );
      stackMap.put(new Integer(_ishr)        , __ishr       );
      stackMap.put(new Integer(_lshr)        , __lshr       );
      stackMap.put(new Integer(_iushr)       , __iushr      );
      stackMap.put(new Integer(_lushr)       , __lushr      );
      stackMap.put(new Integer(_iand)        , __iand       );
      stackMap.put(new Integer(_land)        , __land       );
      stackMap.put(new Integer(_ior)         , __ior        );
      stackMap.put(new Integer(_lor)         , __lor        );
      stackMap.put(new Integer(_ixor)        , __ixor       );
      stackMap.put(new Integer(_lxor)        , __lxor       );
      stackMap.put(new Integer(_i2l)         , __i2l        );
      stackMap.put(new Integer(_i2d)         , __i2d        );
      stackMap.put(new Integer(_l2i)         , __l2i        );
      stackMap.put(new Integer(_l2d)         , __l2d        );
      stackMap.put(new Integer(_d2i)         , __d2i        );
      stackMap.put(new Integer(_d2l)         , __d2l        );
      stackMap.put(new Integer(_lcmp)        , __lcmp       );
      stackMap.put(new Integer(_dcmpg)       , __dcmpg      );
      stackMap.put(new Integer(_ifeq)        , __ifeq       );
      stackMap.put(new Integer(_ifne)        , __ifne       );
      stackMap.put(new Integer(_iflt)        , __iflt       );
      stackMap.put(new Integer(_ifge)        , __ifge       );
      stackMap.put(new Integer(_ifgt)        , __ifgt       );
      stackMap.put(new Integer(_ifle)        , __ifle       );
      stackMap.put(new Integer(_if_icmpeq)   , __if_icmpeq  );
      stackMap.put(new Integer(_if_icmpne)   , __if_icmpne  );
      stackMap.put(new Integer(_if_icmplt)   , __if_icmplt  );
      stackMap.put(new Integer(_if_icmpge)   , __if_icmpge  );
      stackMap.put(new Integer(_if_icmpgt)   , __if_icmpgt  );
      stackMap.put(new Integer(_if_icmple)   , __if_icmple  );
      stackMap.put(new Integer(_goto)        , __goto       );
      stackMap.put(new Integer(_ireturn)     , __ireturn    );
      stackMap.put(new Integer(_dreturn)     , __dreturn    );
      stackMap.put(new Integer(_areturn)     , __areturn    );
      stackMap.put(new Integer(_return)      , __return     );
      stackMap.put(new Integer(_getfield)    , __getfield   );
      stackMap.put(new Integer(_new)         , __new        );
      stackMap.put(new Integer(_anewarray)   , __anewarray  );
      stackMap.put(new Integer(_checkcast)   , __checkcast  );
      stackMap.put(new Integer(_ifnull)      , __ifnull     );
      stackMap.put(new Integer(_ifnonnull)   , __ifnonnull  );
   }
   
   /**
    * Constructor.
    *
    * @param   opCode
    *          Op code for this instruction.
    */
   Instruction(int opCode)
   {
      this.opCode = opCode;
      stackChange = (int[]) stackMap.get(new Integer(opCode));
      length = 1;
   }
   
   /**
    * This constructor should be used for any instruction which has a
    * variable impact on the depth of the stack.  For instance, invocation
    * of a method call has an impact on the stack depth, depending upon
    * the method's signature.
    * <p>
    * The stack change is represented as an array of positive or negative
    * widths (in 32-bit units) related to the popping of operands, if any,
    * and the subsequent pushing of a result, if any. Each array entry
    * represents one operand or result, listed in the order in which the
    * instruction will pop/push to/from the stack as it executes. All data
    * types take up one slot, except for <code>double</code> and
    * <code>long</code>, which each take up two.
    *
    * @param   opCode
    *          Op code for this instruction.
    * @param   stackChange
    *          Incremental change to the stack depth which will be caused
    *          by this instruction at runtime. This array is not stored
    *          directly; it is copied.
    */
   Instruction(int opCode, int[] stackChange)
   {
      this.opCode = opCode;
      int sLen = stackChange.length;
      this.stackChange = new int[sLen];
      System.arraycopy(stackChange, 0, this.stackChange, 0, sLen);
      length = 1;
   }
   
   /**
    * This method updates the current stack depth maintained by the specified
    * expression compiler, based upon the stack depth change required by this
    * instruction.
    *
    * @param   compiler
    *          Expression compiler whose current stack depth must be changed.
    */
   public void modifyStackDepth(Compiler compiler)
   {
      if (Compiler.debug)
      {
         System.out.println(this.toString());
      }
      
      int len = stackChange.length;
      for (int i = 0; i < len; i++)
      {
         compiler.incrementStackDepth(stackChange[i]);
      }
   }
   
   /**
    * This method is called to set the offset position of this instruction
    * when laying out the bytecode instructions in a code unit.
    *
    * @param   offset
    *          New offset of this instruction in its code array.
    *
    * @return  Offset to be applied to the bytecode container immediately
    *          following this instruction.
    */
   public int layout(int offset)
   {
      this.offset = offset;
      
      return (offset + length);
   }
   
   /**
    * Generate a string representation of this instruction.
    *
    * @return  When compiler is in debug mode, the name of the instruction
    *          is returned, else the hex string representation of the op
    *          code is returned.
    *
    * @see     Compiler#debug
    */
   public String toString()
   {
      if (Compiler.debug)
      {
         StringBuilder buf = new StringBuilder();
         buf.append(OPCODE_NAMES.get(new Integer(opCode)));
         
         if (args != null)
         {
            Iterator iter = args.iterator();
            for (int i = 0; iter.hasNext(); i++)
            {
               buf.append(' ')
                  .append(iter.next());
            }
         }
         
         return buf.toString();
      }
      
      return Integer.toHexString(opCode);
   }
   
   /**
    * Write the data represented by this instruction to the specified
    * output stream.  This data includes the bytecode instruction itself,
    * as well as any data represented by the arguments, if any, of this
    * instruction.
    *
    * @param   out
    *          Output stream to which the data will be written.
    *
    * @throws  IOException
    *          if any error occurs while writing.
    */
   public void write(DataOutput out)
   throws IOException
   {
      out.writeByte(opCode);
      
      // Write arguments, if any.
      if (length == 1)
      {
         return;
      }
      
      int size = args.size();
      for (int i = 0; i < size; i++)
      {
         Argument next = (Argument) args.get(i);
         next.write(out);
      }
   }
   
   /**
    * This method returns the op code for this instruction.
    *
    * @return  Op code for this instruction.
    */
   int getOpCode()
   {
      return opCode;
   }
   
   /**
    * Add an individual argument to the list of this instruction's arguments.
    * Update the length of this instruction accordingly.
    *
    * @param   arg
    *          Argument to be added.
    */
   void addArgument(Argument arg)
   {
      if (args == null)
      {
         args = new ArrayList();
      }
      args.add(arg);
      length += arg.getWidth();
   }
   
   /**
    * This method updates the state of the specified stack to reflect the
    * changes this instruction will make to the runtime stack when it is
    * executed.  The ordered array of stack depth changes is used to
    * manipulate the stack.  The array is traversed in order:  for each
    * negative number encountered, the top entry is popped from the
    * stack;  for each positive number encountered, that number is pushed
    * onto the stack.
    *
    * @param   stack
    *          Stack object which represents the state the runtime stack
    *          will be in at the time the instructions in this bytecode
    *          container are executed.
    *
    * @throws  EmptyStackException
    *          if stack is empty when popped.  This indicates an incorrect
    *          instruction was used when compiling the current expression.
    */
   void fixupStack(Stack stack)
   {
      int len = stackChange.length;
      for (int i = 0; i < len; i++)
      {
         int next = stackChange[i];
         
         if (next < 0)
         {
            stack.pop();
         }
         else if (next > 0)
         {
            stack.push(new Integer(next));
         }
      }
   }
}