/*
** Module   : Verifier.java
** Abstract : expression engine abstract syntax tree data type and operator
**            syntax checking helpers
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050509 NEW   @21143 Created initial version which supports
**                               querying the data type of a sub-tree as
**                               well as the validity of an operator's
**                               operands.
** 002 GES 20050527 CHG   @21321 Added support for a unary cast operator.
** 003 GES 20051003 CHG   @22922 Changed anomalous code for binary logical
**                               operator processing to cache the common
**                               operand type without abusing the concept of
**                               the resolveDataType() method which is
**                               supposed to return the expression type,
**                               rather than any unrelated info about the
**                               operands. This was the only such anomaly.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

/**
 * Provides inspection methods that search down the expression tree to
 * determine operator/operand types.  Methods are provided to confirm the
 * validity of the tree as well as to return the data type of the resulting
 * evaluated
 *
 * @author   GES
 * @version  1.0.2
 */
public final class Verifier
implements ExpressionParserTokenTypes
{
   /**
    * Walks the tree and determines the data type of the expression's result.
    * This method recursively processes the tree to verify the correctness
    * of the entire subtree.  Note that the structural makeup of the tree is
    * assumed to be correct and enforced by the parser so no structural tests
    * are done in this method.  Only data type tests are done.
    * <p>
    * The operator tests that are implemented:
    * <pre>
    * Operator         Operand 1 Type      Operand 2 Type      Return Type
    * -------------    -----------------   -----------------   --------------
    * NOT              Boolean             n/a                 Boolean
    * AND              Boolean             Boolean             Boolean
    * OR               Boolean             Boolean             Boolean
    * GT               same as op2         same as op1         Boolean
    * LT               same as op2         same as op1         Boolean
    * GTE              same as op2         same as op1         Boolean
    * LTE              same as op2         same as op1         Boolean
    * EQUALS           same as op2         same as op1         Boolean
    * NOT_EQ           same as op2         same as op1         Boolean
    * RSHIFT           Integral            Integral            same as op1
    * LSHIFT           Integral            Integral            same as op1
    * ZRSHIFT          Integral            Integral            same as op1
    * BIT_AND          Integral            Integral            same as op1
    * BIT_OR           Integral            Integral            same as op1
    * BIT_XOR          Integral            Integral            same as op1
    * BIT_COMP         Integral            Integral            same as op1
    * LPARENS          n/a                 n/a                 n/a
    * IS_NULL          n/a                 n/a                 Boolean
    * NOT_NULL         n/a                 n/a                 Boolean
    * UN_MINUS         Numeric             n/a                 same as op1
    * MULTIPLY         Numeric             Numeric             widened Numeric
    * DIVIDE           Numeric             Numeric             widened Numeric
    * MODULO           Numeric             Numeric             widened Numeric
    * PLUS             Numeric             Numeric             widened Numeric
    * MINUS            Numeric             Numeric             widened Numeric
    * CAST             any                 n/a                 user-specified class
    * </pre>
    * The entries specified as 'same as op1' or 'same as op2' indicate that
    * operands must meet one of the following conditions:
    * <ul>
    *     <li> both operands are Boolean
    *     <li> both operands are of any numeric type (see below)
    *     <li> both operands implement the <code>Comparable</code> interface
    * </ul>
    * <p>
    * Numeric types refer to the following: Double, Long, Float, Integer,
    * Character, Short and Byte.  
    * <p>
    * Integral types are Long, Integer, Character, Short and Byte.
    * <p>
    * <code>LPARENS</code> doesn't exist in the final tree (and even if it
    * did it would just return the data type of its contained expression).
    * For this reason, no <code>LPARENS</code> processing is implemented.
    * <p>
    * <code>IS_NULL</code> and <code>NOT_NULL</code> are artificial
    * operators created by the parser whenever an equality or inequality
    * test against the <code>null</code> literal is encountered. For this
    * reason, no verification is done since the result is always a
    * Boolean and the type of the operand is always some valid class.
    * <p>
    * The arithmetic operators do force a widening conversion on the output 
    * type that is based on the types of the input.  Please see
    * {@link #widestNumericType} for more details.
    * <p>
    * The binary logical comparison operators will have the common operand
    * data type cached in the operator node via the method
    * {@link ExtraAst#setOperandType}.  This will be done for GT, LT, GTE,
    * LTE, EQUALS and NOT_EQ.  The resulting common operand type will be
    * set to the wider of operand1 or operand2.  The resulting 
    * <code>Class</code> instance will be stored in the AST.
    * <p>
    * The following literal processing occurs:
    * <pre>
    * Literal Type                   Class
    * -----------------              ----------------------
    * NUM_LITERAL                    java.lang.Long.TYPE or
    *                                java.lang.Integer.TYPE
    * HEX_LITERAL                    java.lang.Long.TYPE or
    *                                java.lang.Integer.TYPE
    * DEC_LITERAL                    java.lang.Double.TYPE
    * STRING                         java.lang.String.class
    * BOOL_TRUE                      java.lang.Boolean.TYPE
    * BOOL_FALSE                     java.lang.Boolean.TYPE
    * NULL                           null
    * </pre>
    * The following method call processing occurs:
    * <pre>
    * Method Call                    Class
    * -----------------              ----------------------
    * METH_BOOLEAN                   java.lang.Boolean (.TYPE or .class)
    * METH_DOUBLE                    java.lang.Double (.TYPE or .class)
    * METH_LONG                      java.lang.Long (.TYPE or .class)
    * METH_INT                       java.lang.Integer (.TYPE or .class)
    * METH_OBJ                       return type from the embedded Function
    * </pre>
    * Any method object that exists in the tree has already been resolved
    * to a <code>Function</code> object that represents all necessary
    * information about the method including its return type. This is used
    * as the return value for <code>METH_OBJ</code> nodes.
    * <p>
    * At the end of all processing, if an exception is not thrown and if the
    * {@link ExtraAst#getExtra} property is <code>null</code>, the node's
    * class (data type) will be cached using {@link ExtraAst#setExtra}. 
    *
    * @param    ast
    *           The root AST node of an expression tree (or sub-tree).
    *
    * @return   The class of the result that would be generated when the
    *           expression is evaluated.
    *
    * @throws   IllegalArgumentException
    *           If the AST argument is <code>null</code>.
    * @throws   UnsupportedOperationException
    *           On any verification error.
    */
   public static Class resolveDataType(ExtraAst ast)
   throws IllegalArgumentException,
          UnsupportedOperationException
   {
      if (ast == null)
         throw new IllegalArgumentException("Expression AST cannot be null!"); 
      
      int      type   = ast.getType();
      int      radix  = 10;
      ExtraAst first  = null;
      ExtraAst second = null;
      String   errmsg = null;
      Class    cls    = null;
      Class    op1    = null;
      Class    op2    = null;
      
      switch (type)
      {
         // unary logical operator
         case NOT:
            first = (ExtraAst) ast.getFirstChild();
            if (isLogicalOperator(first.getType()))
            {
               op1 = Boolean.TYPE;
               resolveDataType(first);
            }
            else
            {
               op1 = resolveDataType(first);
            }
            cls = Boolean.TYPE;
            
            if (op1 != java.lang.Boolean.class &&
                op1 != java.lang.Boolean.TYPE)
               errmsg = "NOT operator requires a child of type Boolean (" +
                        op1 + " was found).";
            break;
         // binary logical (conjunction) operators
         case AND:
         case OR:
            first  = (ExtraAst) ast.getFirstChild();
            second = (ExtraAst) first.getNextSibling();
            if (isLogicalOperator(first.getType()))
            {
               op1 = Boolean.TYPE;
               resolveDataType(first);
            }
            else
            {
               op1 = resolveDataType(first);
            }
            if (isLogicalOperator(second.getType()))
            {
               op2 = Boolean.TYPE;
               resolveDataType(second);
            }
            else
            {
               op2 = resolveDataType(second);
            }
            cls = java.lang.Boolean.TYPE;
            Class wrap = java.lang.Boolean.class;
            
            if ((op1 != cls && op1 != wrap) || (op2 != cls && op2 != wrap))
            {
               errmsg = " requires 2 Boolean children (" +
                        op1 + " and " + op2 + " were found).";
            }
            break;
         // binary logical operators
         case GT:
         case LT:
         case GTE:
         case LTE:
         case EQUALS:
         case NOT_EQ:
            first  = (ExtraAst) ast.getFirstChild();
            second = (ExtraAst) first.getNextSibling();
            
            if (type == EQUALS || type == NOT_EQ)
            {
               if (isLogicalOperator(first.getType()))
               {
                  op1 = Boolean.TYPE;
                  resolveDataType(first);
               }
               if (isLogicalOperator(second.getType()))
               {
                  op2 = Boolean.TYPE;
                  resolveDataType(second);
               }
            }
            
            if (op1 == null)
            {
               op1 = resolveDataType(first);
            }
            if (op2 == null)
            {
               op2 = resolveDataType(second);
            }
            
            boolean compatible = false;
            
            cls = java.lang.Boolean.TYPE;
            
            if ((op1 == java.lang.Boolean.class ||
                 op1 == java.lang.Boolean.TYPE) &&
                (op2 == java.lang.Boolean.class ||
                 op2 == java.lang.Boolean.TYPE))
            {
               compatible = true;
               ast.setOperandType(java.lang.Boolean.TYPE);
            }
            else if (isNumeric(op1) && isNumeric(op2))
            {
               compatible = true;
               ast.setOperandType(widestNumericType(op1, op2));
            }
            else if (isComparable(op1) && isComparable(op2))
            {
               compatible = true;
               ast.setOperandType(java.lang.Comparable.class);
            }
            
            if (!compatible)
            {
               errmsg = ExpressionParser._tokenNames[type] +
                        " operator requires 2 comparable children (" + op1 +
                        " and " + op2 + " were found).";
            }
            break;
         case BIT_COMP:
            first = (ExtraAst) ast.getFirstChild();
            op1 = resolveDataType(first);
            if (!isIntegral(op1))
            {
               errmsg = ExpressionParser._tokenNames[type] +
                        " operator requires an integral child (" + op1 +
                        " was found).";
            }
            // the result is always the same type as the operand 
            cls = op1;
            break;
         case RSHIFT:
         case LSHIFT:
         case ZRSHIFT:
         case BIT_AND:
         case BIT_OR:
         case BIT_XOR:
            first = (ExtraAst) ast.getFirstChild();
            second = (ExtraAst) first.getNextSibling();
            op1 = resolveDataType(first);
            op2 = resolveDataType(second);
            
            if (!isIntegral(op1) || !isIntegral(op2))
            {
               errmsg = ExpressionParser._tokenNames[type] +
                        " operator requires 2 integral children (" + op1 +
                        " and " + op2 + " were found).";
            }
            else
            {
               if (type == RSHIFT || type == LSHIFT || type == ZRSHIFT)
               {
                  // the result is always the same type as the left operand 
                  cls = op1;
               }
               else
               { 
                  // the result is widened to the largest of the two types
                  // converted from integral to floating point if either of
                  // the operands are floating point
                  cls = widestNumericType(op1, op2);
               }
            }
            break;
         case UN_MINUS:
            first = (ExtraAst) ast.getFirstChild();
            cls = resolveDataType(first);
            
            if (!isNumeric(cls))
            {
               errmsg = "UN_MINUS operator requires 1 numeric child (" +
                        cls + " was found).";
            }
            break;
         case CAST:
            // make sure we verify the child expression
            resolveDataType((ExtraAst) ast.getFirstChild());
            
            // determine the class that matches the specified override type
            try
            {
               cls = ExpressionParser.resolveClass(ast.getText());
            }
            
            catch (IllegalArgumentException iae)
            {
               errmsg = "Invalid class name in #(" + ast.getText() + ").";
            }            
            break;
         case MULTIPLY:
         case DIVIDE:
         case MODULO:
         case PLUS:
         case MINUS:
            first = (ExtraAst) ast.getFirstChild();
            second = (ExtraAst) first.getNextSibling();
            op1 = resolveDataType(first);
            op2 = resolveDataType(second);
            
            if (!isNumeric(op1) || !isNumeric(op2))
            {
               errmsg = ExpressionParser._tokenNames[type] +
                        " operator requires 2 numeric children (" + op1 +
                        " and " + op2 + " were found).";
            }
            else
            {
               cls = widestNumericType(op1, op2);
            }
            break;
         case STRING:
            cls = java.lang.String.class;
            break;
         case NULL:
            // there isn't anything better to return here...
            cls = null;
            break;
         case IS_NULL:
         case NOT_NULL:
            // this processing is only done to drive verification on the
            // child, the results are ignored
            resolveDataType((ExtraAst) ast.getFirstChild());
            cls = java.lang.Boolean.TYPE;
            break;         
         case BOOL_TRUE:
         case BOOL_FALSE:
            cls = java.lang.Boolean.TYPE;
            break;
         case DEC_LITERAL:
            cls = java.lang.Double.TYPE;
            break;
         case HEX_LITERAL:
            radix = 16;
         case NUM_LITERAL:
            cls = java.lang.Integer.TYPE;
            try
            {
               long l = Long.parseLong(ast.getText(), radix);
               if (l > Integer.MAX_VALUE || l < Integer.MIN_VALUE)
               {
                  cls = java.lang.Long.TYPE;
               }
            }
            catch (NumberFormatException exc)
            {
               // should not be possible, since parser pre-determines that
               // we have only safe characters in AST's text
               exc.printStackTrace();
            }
            break;
         case METH_BOOLEAN:
         case METH_INT:
         case METH_LONG:
         case METH_DOUBLE:
         case METH_OBJ:
            Function func = (Function) ast.getExtra(); 
            
            if (func == null)
            {
               errmsg = "Method call must have an embedded Function! " +
                        type + " (" + ExpressionParser._tokenNames[type] +
                        ") with text \"" + ast.getText() + "\".";  
            }
            else
            {
               cls = func.getReturnType(); 
               
               if ((type == METH_BOOLEAN           &&
                    cls != java.lang.Boolean.class &&
                    cls != java.lang.Boolean.TYPE) ||
                   (type == METH_INT               &&
                    cls != java.lang.Integer.class &&
                    cls != java.lang.Integer.TYPE) ||
                   (type == METH_LONG              &&
                    cls != java.lang.Long.class    &&
                    cls != java.lang.Long.TYPE)    ||
                   (type == METH_DOUBLE            &&
                    cls != java.lang.Double.class  &&
                    cls != java.lang.Double.TYPE))
               {
                  errmsg = "Type mismatch for " + type + " (" +
                            ExpressionParser._tokenNames[type] +
                            ") with text \"" + ast.getText() + "\".";  
               }
            }
            break;
         default:
            errmsg = "Unrecognized node " + type + " (" +
                     ExpressionParser._tokenNames[type] + ") with text \"" +
                     ast.getText() + "\".";
      }
      
      // fail the verification on any error!
      if (errmsg != null)
         throw new UnsupportedOperationException(errmsg);
      
      // cache the class object if the "extra" is not in use (this should
      // only be in a case where the AST is a method call OR for some reason
      // this is the second or greater time checking this node)
      if (ast.getExtra() == null)
         ast.setExtra(cls);
      
      return cls;
   }
   
   /**
    * Returns the widest representation of either of the two numeric operands 
    * and will return a floating point value in preference to an integral 
    * value if either of the operands are floating point.  The only anomoly
    * is the case where one operand is a <code>long</code> and one is a 
    * <code>float</code>.  In this case, the widest safe representation
    * is a <code>double</code>.  All other cases will always return a
    * type that is the same as one or the other operand.
    * <p>
    * The following precedence order is followed:
    * <pre>
    * java.lang.Double,        64-bits
    * java.lang.Long,          64-bits
    * java.lang.Float,         32-bits
    * java.lang.Integer,       32-bits
    * java.lang.Short,         16-bits
    * java.lang.Character,     16-bits
    * java.lang.Byte            8-bits    
    * </pre>
    *
    * @param    op1
    *           The first operand class.
    * @param    op2
    *           The second operand class.
    *
    * @return   The widest class found or <code>null</code> if neither
    *           class is numeric.
    */
   public static Class widestNumericType(Class op1, Class op2)
   {
      Class[] widestToSmallest =
      {
         java.lang.Double.TYPE,
         java.lang.Double.class,
         java.lang.Long.TYPE,
         java.lang.Long.class,
         java.lang.Float.TYPE,
         java.lang.Float.class,
         java.lang.Integer.TYPE,
         java.lang.Integer.class,
         java.lang.Short.TYPE,
         java.lang.Short.class,
         java.lang.Character.TYPE,
         java.lang.Character.class,
         java.lang.Byte.TYPE,
         java.lang.Byte.class,
      };
      
      // special case where the simple precedence list approach doesn't
      // work properly since the result is not one of the operand types
      if (((op1 == java.lang.Long.class  || op1 == java.lang.Long.TYPE)   &&
           (op2 == java.lang.Float.class || op2 == java.lang.Float.TYPE)) ||
          ((op2 == java.lang.Long.class  || op2 == java.lang.Long.TYPE)   &&
           (op1 == java.lang.Float.class || op1 == java.lang.Float.TYPE)))
      {
         return java.lang.Double.TYPE;
      }
      
      // general case
      for (int i = 0; i < widestToSmallest.length; i++)
      {
         if (op1 == widestToSmallest[i] || op2 == widestToSmallest[i])
            return widestToSmallest[i];
      }
      
      return null;
   }
   
   /**
    * Tests whether the token type represents a logical operator.
    *
    * @param   type
    *          Token type to be tested.
    *
    * @return  <code>true</code> if the token type represents a logical
    *          operator, else <code>false</code>.
    */
   private static boolean isLogicalOperator(int type)
   {
      switch (type)
      {
         case GT:
         case LT:
         case GTE:
         case LTE:
         case EQUALS:
         case NOT_EQ:
         case AND:
         case OR:
         case NOT:
            return true;
      }
      
      return false;
   }
   
   /**
    * Tests if the class implements the <code>Comparable</code> interface. 
    *
    * @param    cls
    *           The class to test.
    *
    * @return   <code>true</code> if the class implements the
    *           <code>Comparable</code> interface.
    */
   private static boolean isComparable(Class cls)
   {
      return java.lang.Comparable.class.isAssignableFrom(cls);
   }
   
   /**
    * Tests if the class represents one of the integral types. 
    *
    * @param    cls
    *           The class to test.
    *
    * @return   <code>true</code> if the class is an integral type.
    */
   private static boolean isIntegral(Class cls)
   {
      if (cls == java.lang.Long.class        ||
          cls == java.lang.Long.TYPE         ||
          cls == java.lang.Integer.class     ||
          cls == java.lang.Integer.TYPE      ||
          cls == java.lang.Short.class       ||
          cls == java.lang.Short.TYPE        ||
          cls == java.lang.Character.class   ||
          cls == java.lang.Character.TYPE    ||
          cls == java.lang.Byte.class        ||
          cls == java.lang.Byte.TYPE)
      {
         return true;
      }
      
      return false;
   }
   
   /**
    * Tests if the class represents one of the floating point types. 
    *
    * @param    cls
    *           The class to test.
    *
    * @return   <code>true</code> if the class is an floating point type.
    */
   private static boolean isFloat(Class cls)
   {
      if (cls == java.lang.Double.class   ||
          cls == java.lang.Double.TYPE    ||
          cls == java.lang.Float.class    ||
          cls == java.lang.Float.TYPE)
      {
         return true;
      }
      
      return false;
   }
   
   /**
    * Tests if the class represents one of the numeric types (either
    * integral or floating point). 
    *
    * @param    cls
    *           The class to test.
    *
    * @return   <code>true</code> if the class is a numeric type.
    */
   private static boolean isNumeric(Class cls)
   {
      if (isIntegral(cls) || isFloat(cls))
      {
         return true;
      }
      
      return false;
   }
}