/*
** Module   : GraphDB.java
** Abstract : Provides generic access to a tinkerpop-compatible graph DB.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140314 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.graphdb;

import org.apache.commons.configuration.*;

import com.tinkerpop.blueprints.*;

/**
 * An abstract adapter to obtain connections and create indexes in a tinkerpop-compatible 
 * graph DB.
 */
public abstract class GraphDB
{
   /** The singleton implementation of our graph DB. */
   private static final GraphDB instance = new TitanGraphDB();
   
   /**
    * Obtain a graph DB with the connection details specified in the given {@link Configuration}.
    * 
    * @param    configuration
    *           The connection details.
    *           
    * @return   A connected graph DB.
    */
   public static Graph obtainGraphDB(Configuration configuration)
   {
      Graph db = instance.connectGraphDBImpl(configuration);

      final Graph graph = db;
      
      // ensure the DB is disconnected on JVM shutdown
      Runtime.getRuntime().addShutdownHook(new Thread()
      {
         @Override
         public void run()
         {
            graph.shutdown();
         }
      });
      
      return graph;
   }

   /**
    * Create an unique index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    key
    *           The property key name to be indexed.
    * @param    datatype
    *           The key's type.
    * @param    indexed
    *           The graph element to be indexed.
    * @param    text
    *           {@code true} if this index allows text-search.
    * @param    indexNames
    *           List of additional index provider names, to which an index will be created.
    */
   public static void createUniqueIndex(Graph                    db,
                                        String                   key,
                                        Class<?>                 datatype,
                                        Class<? extends Element> indexed,
                                        boolean                  text,
                                        String...                indexNames)
   {
      instance.createUniqueIndexImpl(db, key, datatype, indexed, text, indexNames);
   }

   /**
    * Create a sort index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    key
    *           The property key name to be indexed.
    * @param    datatype
    *           The key's type.
    * @param    indexed
    *           The graph element to be indexed.
    * @param    text
    *           {@code true} if this index allows text-search.
    * @param    indexNames
    *           List of additional index provider names, to which an index will be created.
    */
   public static void createIndex(Graph                    db,
                                  String                   key,
                                  Class<?>                 datatype,
                                  Class<? extends Element> indexed,
                                  boolean                  text,
                                  String...                indexNames)
   {
      instance.createIndexImpl(db, key, datatype, indexed, text, indexNames);
   }

   /**
    * Obtain a graph DB with the connection details specified in the given {@link Configuration}.
    * 
    * @param    configuration
    *           The connection details.
    *           
    * @return   A connected graph DB.
    */
   protected abstract Graph connectGraphDBImpl(Configuration configuration);

   /**
    * Create an unique index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    key
    *           The property key name to be indexed.
    * @param    datatype
    *           The key's type.
    * @param    indexed
    *           The graph element to be indexed.
    * @param    text
    *           {@code true} if this index allows text-search.
    * @param    indexNames
    *           List of additional index provider names, to which an index will be created.
    */
   protected abstract void createUniqueIndexImpl(Graph                    db,
                                                 String                   key,
                                                 Class<?>                 datatype,
                                                 Class<? extends Element> indexed,
                                                 boolean                  text,
                                                 String...                indexNames);

   /**
    * Create a sort index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    key
    *           The property key name to be indexed.
    * @param    datatype
    *           The key's type.
    * @param    indexed
    *           The graph element to be indexed.
    * @param    text
    *           {@code true} if this index allows text-search.
    * @param    indexNames
    *           List of additional index provider names, to which an index will be created.
    */
   protected abstract void createIndexImpl(Graph                    db,
                                           String                   key,
                                           Class<?>                 datatype,
                                           Class<? extends Element> indexed,
                                           boolean                  text,
                                           String...                indexNames);
}