/*
** Module   : TitanGraphDB.java
** Abstract : Provides access to a Titan graph DB.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140314 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.graphdb;

import org.apache.commons.configuration.*;

import com.thinkaurelius.titan.core.*;
import com.thinkaurelius.titan.core.Parameter;
import com.tinkerpop.blueprints.*;

/**
 * An adapter to obtain connections and create indexes in a Titan graph DB
 */
class TitanGraphDB
extends GraphDB
{
   /**
    * Obtain a graph DB with the connection details specified in the given {@link Configuration}.
    * 
    * @param    configuration
    *           The connection details.
    *           
    * @return   A connected graph DB.
    */
   @Override
   protected Graph connectGraphDBImpl(Configuration configuration)
   {
      Graph graph = TitanFactory.open(configuration);

      return graph;
   }

   /**
    * Create an unique index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    key
    *           The property key name to be indexed.
    * @param    datatype
    *           The key's type.
    * @param    indexed
    *           The graph element to be indexed.
    * @param    text
    *           {@code true} if this index allows text-search.
    * @param    indexNames
    *           List of additional index provider names, to which an index will be created.
    */
   @Override
   protected void createUniqueIndexImpl(Graph                    db,
                                        String                   key,
                                        Class<?>                 datatype,
                                        Class<? extends Element> indexed,
                                        boolean                  text,
                                        String...                indexNames)
   {
      TitanGraph tg = (TitanGraph) db;
      if (!tg.getIndexedKeys(indexed).contains(key))
      {
         KeyMaker km = tg.makeKey(key).dataType(datatype).indexed(indexed);

         for (int i = 0; i < indexNames.length; i++)
         {
            if (text)
            {
               km.indexed(indexNames[i], 
                          indexed, Parameter.of(Mapping.MAPPING_PREFIX, Mapping.STRING));
            }
            else
            {
               km.indexed(indexNames[i],indexed);
            }
         }
         
         km.unique().make();
         
         tg.commit();
      }
   }

   /**
    * Create a sort index in the given graph DB.
    * 
    * @param    db
    *           The graph DB.
    * @param    key
    *           The property key name to be indexed.
    * @param    datatype
    *           The key's type.
    * @param    indexed
    *           The graph element to be indexed.
    * @param    text
    *           {@code true} if this index allows text-search.
    * @param    indexNames
    *           List of additional index provider names, to which an index will be created.
    */
   @Override
   protected void createIndexImpl(Graph                    db,
                                  String                   key,
                                  Class<?>                 datatype,
                                  Class<? extends Element> indexed,
                                  boolean                  text,
                                  String...                indexNames)
   {
      TitanGraph tg = (TitanGraph) db;
      if (!tg.getIndexedKeys(indexed).contains(key))
      {
         KeyMaker km = tg.makeKey(key).dataType(datatype).indexed(indexed);
         for (int i = 0; i < indexNames.length; i++)
         {
            if (text)
            {
               km.indexed(indexNames[i], indexed, 
                          Parameter.of(Mapping.MAPPING_PREFIX, Mapping.STRING));
            }
            else
            {
               km.indexed(indexNames[i], indexed);
            }
         }
         
         km.make();
         
         tg.commit();
      }
   }
}