/*
** Module   : AdminAccountExtension.java
** Abstract : Defines a custom account administration extension interface
**            for the admin server.
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090522   @42442 Created initial version.
** 002 NVS 20090528   @42499 Declared extension objects as Serializable.
** 003 CA  20090604   @42570 Modified add,set and clone extension methods to
**                           throw an AccountExtValidationException if 
**                           validation fails or an extension field maintained
**                           by server-side can't be computed.
** 004 NVS 20090604   @42602 Moved AccountExtValidationException class to
**                           admin package.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin;

import com.goldencode.p2j.directory.DirectoryService;

import java.io.Serializable;

/**
 * Defines a set of methods that enable communication between the standard
 * admin server and a custom admin server plugin, responsible for the
 * maintaining the custom account extension data in a sound state. 
 */
public interface AdminAccountExtension
{
   /**
    * Inspects the given account directory node and tells whether this account
    * has custom extension data associated with it.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. Typically, an editing batch would be open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @return   <code>true</code> if this account possesses any extension data
    */
   public boolean isExtended(DirectoryService ds, String node);
   
   /**
    * Adds custom account extension data to the newly created account
    * directory node.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @param    ext
    *           a container with extension data as received from the admin
    *           client extension plugin
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    *           
    * @throws   AccountExtValidationException
    *           if any exception is encountered while validating fields or 
    *           computing default field values, an exception will be thrown. 
    */
   public boolean addExtension(DirectoryService ds, String node, 
                               Serializable ext)
   throws AccountExtValidationException;
   
   /**
    * Changes custom account extension data associated with the exsiting
    * account directory node.
    * Changes may include synchronizaion, addition or deletion of the whole
    * set of extension data, based on the current state of the node and the
    * provided data input.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @param    ext
    *           a container with extension data as received from the admin
    *           client extension plugin
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    *           
    * @throws   AccountExtValidationException
    *           if any exception is encountered while validating fields or 
    *           computing default field values, an exception will be thrown. 
    */
   public boolean setExtension(DirectoryService ds, String node,
                               Serializable ext)
   throws AccountExtValidationException;
   
   /**
    * Clones existing custom account extension data to the newly cloned 
    * account directory node.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    fromNode
    *           full account directory node path and ID of the source account
    *
    * @param    toNode
    *           full account directory node path and ID of the target account
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    *           
    * @throws   AccountExtValidationException
    *           if any exception is encountered while validating fields or 
    *           computing default field values, an exception will be thrown. 
    */
   public boolean cloneExtension(DirectoryService ds, String fromNode, 
                                 String toNode)
   throws AccountExtValidationException;

   /**
    * Deletes custom account extension data associated with the account
    * directory node being deleted.
    * <p>
    * An instance of <code>DirectoryService</code> is provided for the method,
    * which may be used to access the directory, should there be a need. The
    * calling thread is bound to the service already and should not unbind
    * itself. A directory editing batch is open at the time of
    * the method call, so the implementation should not try opening a batch
    * internally.
    *
    * @param    ds
    *           instance of <code>DirectoryService</code> that may be used 
    *           for all directory access needs
    *
    * @param    node
    *           full account directory node path and ID
    *
    * @return   <code>true</code> if operation was successful; otherwise
    *           <code>false</code>, which will cause the batch rollback
    */
   public boolean deleteExtension(DirectoryService ds, String node);
}