/*
** Module   : RecordLockInfo.java
** Abstract : Information about a single record lock held by the lock manager
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20090623   @42788 Created initial version. Information about a
**                           single record lock held by the P2J lock manager.
** 002 ECF 20090801   @43503 Added physical database name.
** 003 CA  20091119   @44435 Record lockers are identified by session tokens
**                           instead of session IDs, to provide the subject 
**                           name too.
** 004 SVL 20091126   @44446 Make this class extend RecordInfo.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin;

import java.io.Serializable;
import java.util.Arrays;

import com.goldencode.p2j.security.SessionToken;

/**
 * Information about a single record lock held by the P2J lock manager.
 * Represents a transient snapshot of lock information at the moment this
 * information is requested.
 * <p>
 * The information is read-only, once the object is constructed.  It includes
 * the type of lock, the time the lock was acquired, the table name and
 * primary key of the lock, and an array of session IDs of those contexts
 * which currently hold the lock (must be at least one).
 */
public final class RecordLockInfo
extends RecordInfo
implements Serializable
{
   /** Possible lock types */
   public static enum Type
   {
      /** Exclusive lock - there can be only one locker */
      EXCLUSIVE,
      
      /** Shared lock - multiple lockers are possible, one minimum */
      SHARE,
      
      /** No lock */
      NONE,
   }
   
   /** Lock type */
   private final Type type;
   
   /** Timestamp (time of epoch) at which lock was acquired */
   private final long timeAcquired;
   
   /** Session tokens of one or more contexts which hold the lock */
   private final SessionToken[] lockers;
   
   /**
    * Constructor.
    * 
    * @param   type
    *          Lock type.
    * @param   timeAcquired
    *          Timestamp (time of epoch) at which lock was acquired.
    * @param   databaseName
    *          Physical database name.
    * @param   tableName
    *          Name of table.
    * @param   recordID
    *          Primary key ID of locked record.
    * @param   lockers
    *          Session tokens of one or more contexts which hold the lock.
    */
   public RecordLockInfo(Type type,
                         long timeAcquired,
                         String databaseName,
                         String tableName,
                         long recordID,
                         SessionToken[] lockers)
   {
      super(databaseName, tableName, recordID);
      this.type = type;
      this.timeAcquired = timeAcquired;
      this.lockers = lockers;
   }
   
   /**
    * Retrieve the lock type.
    * 
    * @return  Lock type.
    */
   public Type getType()
   {
      return type;
   }
   
   /**
    * Retrieve the timestamp (time of epoch) at which the lock was acquired.
    * 
    * @return  Acquisition timestamp.
    */
   public long getTimeAcquired()
   {
      return timeAcquired;
   }
   
   /**
    * Retrieve an array of session tokens which hold this lock.  This method
    * actually returns a copy to ensure the stored value is immutable.
    * 
    * @return  Copy of lockers' session tokens.  Must not be <code>null</code>
    *          and must contain at least one entry (possibly more for lock
    *          type <code>SHARE</code>.
    */
   public SessionToken[] getLockers()
   {
      return Arrays.copyOf(lockers, lockers.length);
   }
}