/*
** Module   : AdminACLMenu.java
** Abstract : Access Control menu support for the admin session
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 CA  20090714  @43203  Created the first version of the file.
** 002 CA  20090720  @43241  The "Manage Access Control by Server" menu is 
**                           disabled and needs to be implemented. Added 
**                           method getAclAdmin.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.awt.event.*;
import javax.swing.*;

/**
 * This is a Access Control menu management class for the admin session. 
 * It provides lazy Access Control screens creation and dispatches menu 
 * actions to the specialized handlers.
 * <p>
 * At logoff, these actions are inversed. The menu is reattached back
 * to this class to help release allocated resources and give the next session
 * a fresh start.
 * 
 * @author CA
 */
public class AdminACLMenu
implements AdminNotify
{
   /** Related instance of AdminClient */
   private AdminClient client = null;

   /** Access Control menu */
   private JMenu acl = null;
   
   /** Access Control --&gt; Show by Account menu item */
   private JMenuItem menuByAccount = null; 
   
   /** Access Control --&gt; Show by Resource Type menu */
   private JMenu menuByResource = null; 
   
   /** Access Control --&gt; Show by Resource Type -&gt; menu items*/
   private JMenuItem[] menuByResourceItems = null; 
   
   /** Access Control --&gt; Show All menu item */
   private JMenuItem menuAll = null; 
   
   /** Access Control --&gt; Manage Access Control by Server menu */
   private JMenu menuByServer = null; 
   
   /** Access Control --&gt; Manage Access Control by Server -&gt; menu items */
   private JMenuItem[] menuByServerItems = null; 
   
   /** The current ACL management instance. */
   private AdminACL aAcl = null; 

   /**
    * Constructor.
    */
   AdminACLMenu()
   {
   }

   /** 
    * Initializes useful references.
    *
    * @param  client
    *         instance of AdminClient
    *
    * @param  aclMenu
    *         reference to Access Control menu
    */
   void init(AdminClient client, JMenu aclMenu)
   {
      // store useful references
      this.client = client;
      this.acl    = aclMenu;
      
      // register for logoff notifications (safe to make multiple calls)
      client.registerLogoffListener(this);

      // register for target notifications (safe to make multiple calls)
      client.registerTargetListener(this);
      
      menuByAccount  = aclMenu.getItem(0);
      menuByResource = (JMenu) aclMenu.getItem(1);
      menuAll        = aclMenu.getItem(2);
      menuByServer   = (JMenu) aclMenu.getItem(3);
      // disable this menu (no impl provided)
      menuByServer.setEnabled(false);
      
      menuByResourceItems = AdminUtils.getMenuItems(menuByResource);
      menuByServerItems = AdminUtils.getMenuItems(menuByServer);
      
      AdminUtils.removeActionListeners(menuByAccount);
      AdminUtils.removeActionListeners(menuAll);
      AdminUtils.removeActionListeners(menuByResourceItems);
      AdminUtils.removeActionListeners(menuByServerItems);

      menuByAccount.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               lazyByAccount();
            }
         });
      
      menuAll.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               lazyAll();
            }
         });
      
      for (int i = 0; i < menuByResourceItems.length; i++)
      {
         final String resourceType = menuByResourceItems[i].getText();
         final JMenuItem menuItem = menuByResourceItems[i];
         
         menuByResourceItems[i].addActionListener(
            // this class provides lazy screen init                               
            new ActionListener()
            {
               public void actionPerformed(ActionEvent evt) 
               {
                  lazyByResourceType(menuItem, resourceType);
               }
            });
      }
      
      for (int i = 0; i < menuByServerItems.length; i++)
      {
         final String server = menuByServerItems[i].getText();
         final JMenuItem menuItem = menuByServerItems[i];
         
         menuByServerItems[i].addActionListener(
               // this class provides lazy screen init                               
               new ActionListener()
               {
                  public void actionPerformed(ActionEvent evt) 
                  {
                     lazyByServer(menuItem, server);
                  }
               });
      }
   }

   /** 
    * Notifies about successful logoff or target change and gives chance 
    * to release resources.
    * <p>In both cases the reaction is the full cleanup.
    *
    * @param eventCode
    *        event code, one of the NOTIFY_* constants
    */
   public void notifyEvent(int eventCode)
   {
      // filter out unwanted events
      if (eventCode != AdminNotify.NOTIFY_LOGOFF &&
          eventCode != AdminNotify.NOTIFY_TARGET)
      {
         return;
      }
      
      // release references
      aAcl = null;
      menuByServerItems = null;
      menuByResourceItems = null;
      
      // reinit menu
      init(client, acl);
   }
   
   /**
    * Get the current ACL management instance.
    * 
    * @return the current {@link #aAcl ACL management} instance.
    */
   AdminACL getAclAdmin()
   {
      if (aAcl == null)
      {
         aAcl = new AdminACL(client);
      }

      return aAcl;
   }
   
   /** 
    * Handles lazy Show by Account menu initialization.
    */
   private void lazyByAccount() 
   {
      // initialize ACL screen
      if (aAcl == null)
      {
         aAcl = new AdminACL(client);
      }
      
      AdminUtils.removeActionListeners(menuByAccount);
      
      menuByAccount.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               aAcl.showByAccount();
            }
         });

      // handle this action once
      aAcl.showByAccount();
   }

   /** 
    * Handles lazy Show All menu initialization.
    */
   private void lazyAll() 
   {
      // initialize ACL screen
      if (aAcl == null)
      {
         aAcl = new AdminACL(client);
      }
      
      AdminUtils.removeActionListeners(menuAll);
      
      menuAll.addActionListener(
            // this class provides lazy screen init                               
            new ActionListener()
            {
               public void actionPerformed(ActionEvent evt) 
               {
                  aAcl.showAll();
               }
            });
      
      // handle this action once
      aAcl.showAll();
   }

   /** 
    * Handles lazy Show by Resource Type menu initialization.
    */
   private void lazyByResourceType(JMenuItem menu, final String resourceType) 
   {
      // initialize ACL screen
      if (aAcl == null)
      {
         aAcl = new AdminACL(client);
      }
      
      AdminUtils.removeActionListeners(menu);
      
      menu.addActionListener(
            // this class provides lazy screen init                               
            new ActionListener()
            {
               public void actionPerformed(ActionEvent evt) 
               {
                  aAcl.showByResourceType(resourceType);
               }
            });
      
      // handle this action once
      aAcl.showByResourceType(resourceType);
   }
   
   /** 
    * Handles lazy Manage Access Control by Server menu initialization.
    */
   private void lazyByServer(JMenuItem menu, final String server) 
   {
      // initialize ACL screen
      if (aAcl == null)
      {
         aAcl = new AdminACL(client);
      }
      
      AdminUtils.removeActionListeners(menu);
      
      menu.addActionListener(
            // this class provides lazy screen init                               
            new ActionListener()
            {
               public void actionPerformed(ActionEvent evt) 
               {
                  aAcl.showByServer(server);
               }
            });
      
      // handle this action once
      aAcl.showByServer(server);
   }
}