/*
** Module   : AdminAccounts.java
** Abstract : Accounts menu support for the admin session
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090325  @41686  Created the first version of the file.
** 002 NVS 20090428  @42035  Extended dynamic menu creation for multiple
**                           sessions and all type of items.
** 003 CA  20090513  @42174  Added getter for the AdminCertificates instance.
** 004 CA  20090714  @43155  Modified lazy loading of menuitems - use an API 
**                           in AdminUtils to remove the action listeners.
** 005 NVS 20091007  @44110  Implemented LOGON event notification handler which
**                           selectively disables the menu items in the Accounts
**                           screen according to the admin permissions profile.
** 006 SVL 20100920          Added Accounts -> Bulk Auth Mode Update menu
**                           item.
** 007 EVL 20160222          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.AdminExports;
import com.goldencode.p2j.admin.AdminDef;

import java.awt.event.*;
import javax.swing.*;

/**
 * This is a Accounts menu management class for the admin session. 
 * It provides lazy Accounts screens creation and dispatches menu actions
 * to the specialized handlers by the account type.
 * <p>
 * At logoff, these actions are inversed. Accouns menu is reattached back
 * to this class to help release allocated resources and give the next session
 * a fresh start.
 * 
 * @author NVS
 */
public class AdminAccounts
implements AdminNotify
{
   /** Related instance of AdminClient */
   private AdminClient client = null;

   /** Accounts menu */
   private JMenu acc = null;
   
   /** Accounts --&gt; groups menu item */
   private JMenuItem groups = null; 

   /** Accounts --&gt; groups menu item */
   private JMenuItem users = null; 

   /** Accounts --&gt; groups menu item */
   private JMenuItem procs = null; 

   /** Accounts --&gt; groups menu item */
   private JMenuItem certs = null;

   /** instance of group account management class */
   private AdminGroupAccounts ga = null;
   
   /** instance of user account management class */
   private AdminUserAccounts ua = null;
   
   /** instance of process account management class */
   private AdminProcessAccounts pa = null;
   
   /** instance of certificate management class */
   private AdminCertificates ca = null;
   
   
   /**
    * Constructor.
    */
   AdminAccounts()
   {
   }

   /** 
    * Initializes useful references.
    *
    * @param  client
    *         instance of AdminClient
    *
    * @param  accMenu
    *         reference to Accounts menu
    */
   void init(AdminClient client, JMenu accMenu)
   {
      // store useful references
      this.client = client;
      this.acc    = accMenu;
      
      // register for logon notifications (safe to make multiple calls)
      client.registerLogonListener(this);
      
      // register for logoff notifications (safe to make multiple calls)
      client.registerLogoffListener(this);

      // register for target notifications (safe to make multiple calls)
      client.registerTargetListener(this);
      
      // create Account menu items
      groups = new JMenuItem("Groups");
      acc.add(groups);
      groups.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               lazyGroups();
            }
         });
      
      users = new JMenuItem("Users");
      acc.add(users);
      users.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               lazyUsers();
            }
         });
      
      procs = new JMenuItem("Processes");
      acc.add(procs);
      procs.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               lazyProcs();
            }
         });
      
      certs = new JMenuItem("Certificates");
      acc.add(certs);
      certs.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               lazyCerts();
            }
         });

      // adding Accounts -> Bulk Auth Mode Update menu item 
      JMenuItem itemBulkAuth = new JMenuItem(AdminBulkAuthForm.DIALOG_TITLE);
      acc.add(itemBulkAuth);
      final AdminBulkAuthForm bulkAuthForm = new AdminBulkAuthForm(client);

      itemBulkAuth.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            bulkAuthForm.showDialog();
         }
      });
   }
   
   /** 
    * Notifies about successful logoff or target change and gives chance 
    * to release resources.
    * <p>In both cases the reaction is the full cleanup.
    *
    * @param eventCode
    *        event code, one of the NOTIFY_* constants
    */
   public void notifyEvent(int eventCode)
   {
      // filter out unwanted events
      if (eventCode != AdminNotify.NOTIFY_LOGON  &&
          eventCode != AdminNotify.NOTIFY_LOGOFF &&
          eventCode != AdminNotify.NOTIFY_TARGET)
      {
         return;
      }
      
      // process logon notification - adjust menu permissions
      if (eventCode == AdminNotify.NOTIFY_LOGON)
      {
         AdminExports adm = client.getExports();
         AdminDef.Profile[] prof = adm.getProfile(new String[]
         {
            "/accounts/groups",
            "/accounts/users",
            "/accounts/processes",
            "/accounts/certificates",
         });
         
         if (prof == null)
         {
            prof = new AdminDef.Profile[]
                   {
                      new AdminDef.Profile("", false),
                      new AdminDef.Profile("", false),  
                      new AdminDef.Profile("", false),  
                      new AdminDef.Profile("", false),
                   };
         }
         
         // adjust the menu items
         groups.setEnabled(prof[0].permission);
         users.setEnabled( prof[1].permission);
         procs.setEnabled( prof[2].permission);
         certs.setEnabled( prof[3].permission);
         
         return;
      }
      
      // destroy the existing Accounts menu items
      acc.removeAll();
      
      // release references
      ga = null;
      ua = null;
      pa = null;
      ca = null;
      
      // reinit menu
      init(client, acc);
   }
   
   /**
    * Get the current {@link AdminCertificates} menu handler. If it was not
    * already initialized, the {@link #lazyCerts()} method will be called.
    * 
    * @return current {@link AdminCertificates} menu handler.
    */
   AdminCertificates getCA()
   {
      if (ca == null)
      {
         lazyCerts();
      }

      return ca;
   }
   
   /** 
    * Handles lazy groups menu initialization.
    */
   private void lazyGroups() 
   {
      // initialize groups screen
      ga = new AdminGroupAccounts(client); 
      
      AdminUtils.removeActionListeners(groups);

      // allow direct action handler invocation
      groups.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               ga.manageGroups();
            }
         });

      // handle this action once
      ga.manageGroups();
   }

   /** 
    * Handles lazy users menu initialization.
    */
   private void lazyUsers() 
   {
      // initialize users screen
      ua = new AdminUserAccounts(client); 
      
      AdminUtils.removeActionListeners(users);
      
      // allow direct action handler invocation
      users.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               ua.manageUsers();
            }
         });

      // handle this action once
      ua.manageUsers();
   }

   /** 
    * Handles lazy processes menu initialization.
    */
   private void lazyProcs() 
   {
      // initialize processes screen
      pa = new AdminProcessAccounts(client); 
      
      AdminUtils.removeActionListeners(procs);
      
      // allow direct action handler invocation      
      procs.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               pa.manageProcesses();
            }
         });

      // handle this action once
      pa.manageProcesses();
   }

   /** 
    * Handles lazy certificates menu initialization.
    */
   private void lazyCerts() 
   {
      // initialize certificates screen
      ca = new AdminCertificates(client); 
      
      AdminUtils.removeActionListeners(certs);
      
      // allow direct action handler invocation
      certs.addActionListener(
         // this class provides lazy screen init                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               ca.manageCertificates();
            }
         });

      // handle this action once
      ca.manageCertificates();
   }
}