/*
** Module   : AdminAcquireLocks.java
** Abstract : screen that allows to lock specific database records
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description--------------------
** 001 SVL 20091126   @44453 Created the first version to provide ability to
**                           lock records from the admin UI.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.awt.*;
import java.awt.event.*;
import java.awt.print.PrinterException;
import java.text.MessageFormat;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

import com.goldencode.p2j.admin.RecordInfo;

/**
 * Provides a screen that allows to lock specific database records.
 */
public class AdminAcquireLocks
extends AdminScreen
implements PrintRequest
{
   /** Screen switcher - acquire record locks screen ID. */
   public static final String SCREEN_ACQUIRE_LOCKS = "adm_lock_records";
   
   /** Screen title. */
   private static final String TITLE_ACQUIRE_LOCKS = "Records to lock";
   
   /** List of records to lock. */
   private JTable recordTable = null;

   /** Record list data model. */
   private RecordInfoModel dataModel = null;

   /** Lock button. */
   private JButton lockBtn = null;

   /** Unlock button. */
   private JButton unlockBtn = null;

   /** Add button. */
   private JButton addBtn = null;

   /** Delete button. */
   private JButton deleteBtn = null;

   /** Delete all button. */
   private JButton deleteAllBtn = null;

   /** Panel for the "add record" dialog. */
   private JPanel addRecordPanel = null;

   /** Input for the record ID in the "add record" dialog. */
   private JTextField recordIdField = null;

   /** Combo-box for the database name in the "add record" dialog. */
   JComboBox databaseCombo = null;

   /** Combo-box for the table name in the "add record" dialog. */
   JComboBox tableCombo = null;

   /**
    * Mapping of the database names to the combo-box models which contain
    * names of the tables of these databases. 
    */
   private Map<String, DefaultComboBoxModel> databaseTables = null;

   /**
    * Constructor.
    *
    * @param    client
    *           Core UI module of which we are a part.
    */
   AdminAcquireLocks(AdminClient client)
   {
      // store useful references
      this.client = client;
      adm = client.getExports();
      
      // create the main screen and add it to the swapper
      mainPanel = new JPanel();
      client.add(mainPanel, SCREEN_ACQUIRE_LOCKS, this);
      mainPanel.setLayout(new BorderLayout());
      
      recordTable = new JTable();
      recordTable.getTableHeader().setReorderingAllowed(false);
      
      JScrollPane scrollPane = new JScrollPane(recordTable);
      scrollPane.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
      recordTable.setFillsViewportHeight(true);
      recordTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      
      mainPanel.add(scrollPane, BorderLayout.CENTER);

      // create control widgets
      JPanel ctrlPanel = new JPanel(new GridLayout(1, 0));
      ctrlPanel.setBorder(BorderFactory.createEmptyBorder(0,10,10,10));
      
      lockBtn = new JButton("Lock records");
      lockBtn.setEnabled(false);
      unlockBtn = new JButton("Unlock records");
      unlockBtn.setEnabled(false);
      addBtn = new JButton("Add record");
      deleteBtn = new JButton("Delete selected");
      deleteBtn.setEnabled(false);
      deleteAllBtn = new JButton("Delete all");
      deleteAllBtn.setEnabled(false);

      ctrlPanel.add(lockBtn);
      ctrlPanel.add(unlockBtn);
      ctrlPanel.add(addBtn);
      ctrlPanel.add(deleteBtn);
      ctrlPanel.add(deleteAllBtn);
      
      mainPanel.add(ctrlPanel, BorderLayout.SOUTH);
      
      lockBtn.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent actionEvent)
         {
            lock();
         }
      });

      unlockBtn.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent actionEvent)
         {
            unlock();
         }
      });

      addBtn.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            add();
         }
      });

      deleteBtn.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt)
         {
            delete();
         }
      });

      deleteAllBtn.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt)
         {
            deleteAll();
         }
      });

      // create the array of components for the state machine
      state = new JComponent[]
      {
         recordTable,
         ctrlPanel,
         lockBtn,
         addBtn,
         deleteBtn,
         deleteAllBtn
      };

      // create the data model for the table of records
      dataModel = new RecordInfoModel();
      recordTable.setModel(dataModel);
      recordTable.setSelectionMode(
                              ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
      ListSelectionModel selectionModel = recordTable.getSelectionModel();
      selectionModel.addListSelectionListener(new ListSelectionListener()
      {
         public void valueChanged(ListSelectionEvent listSelectionEvent)
         {
            // disable delete button if there is no selected records
            deleteBtn.setEnabled(recordTable.getSelectedRowCount() != 0);
         }
      });

      int num = dataModel.getColumnCount();
      for (int i = 0; i < num; i++)
      {
         TableColumn column = recordTable.getColumnModel().getColumn(i);
         column.setPreferredWidth(dataModel.getColumnWidth(i));
      }

      // create "add record" dialog
      JLabel databaseLabel = new JLabel("Database");
      JLabel tableLabel    = new JLabel("Table");
      JLabel recordIdLabel = new JLabel("Record ID");
      databaseCombo  = new JComboBox();
      tableCombo     = new JComboBox();
      recordIdField  = new JTextField(20);
      JButton addBtn = new JButton("Add");
      JButton cancelBtn = new JButton("Cancel");

      databaseCombo.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent actionEvent)
         {
            reloadDatabaseTables();
         }
      });

      addBtn.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent actionEvent)
         {
            handleAddRecord();
         }
      });

      cancelBtn.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent actionEvent)
         {
            handleCancelRecord();
         }
      });

      addRecordPanel = AdminUtils.groupComponents(new Component[][]
      {
         {databaseLabel, databaseCombo},
         {tableLabel, tableCombo},
         {recordIdLabel, recordIdField},
         {cancelBtn, addBtn}
      });
   }

   /**
    * Performs printing of the current screen data.
    * This screen prints the current view of the record table.
    */
   public void print()
   {
      MessageFormat header = new MessageFormat(TITLE_ACQUIRE_LOCKS);
      MessageFormat footer = new MessageFormat("page {0, number}");

      try
      {
         recordTable.print(JTable.PrintMode.FIT_WIDTH, header, footer);
      }
      catch (PrinterException pe)
      {
         client.beep("Printer exception " + pe);
      }
      catch (SecurityException se)
      {
         // nothing to do
      }
   }

   /**
    * Queries the extended printing capability. If the exteneded capability
    * is not provided, the printExtended() method should not be called.
    *
    * @return <code>false</code> to tell the caller the extended printing
    *         is not implemented.
    */
   public boolean canPrintExtended()
   {
      return false;
   }

   /**
    * Does nothing here.
    */
   public void printExtended()
   {
   }

   /**
    * Presents the "acquire locks" screen.
    */
   void manageLocks()
   {
      // switch to this screen without changing state of components
      client.show(SCREEN_ACQUIRE_LOCKS, TITLE_ACQUIRE_LOCKS);
   }

   /**
    * Lock records specified in the table.
    */
   private void lock()
   {
      client.ok();
      
      if (dataModel.getRowCount() == 0)
      {
         return;
      }

      RecordInfo[] records = dataModel.getRecords();

      AdminUtils.busy(client);

      boolean res = adm.lockRecords(records);
      if (res)
      {
         recordTable.setEnabled(false);
         unlockBtn.setEnabled(true);
         lockBtn.setEnabled(false);
         addBtn.setEnabled(false);
         deleteBtn.setEnabled(false);
         deleteAllBtn.setEnabled(false);
      }
      else
      {
         client.beep(adm.peekMessage());
      }

      AdminUtils.idle(client);
   }

   /**
    * Unlock records previously locked by admin UI.
    */
   private void unlock()
   {
      AdminUtils.busy(client);

      adm.unlockRecords();

      recordTable.setEnabled(true);
      unlockBtn.setEnabled(false);
      lockBtn.setEnabled(true);
      addBtn.setEnabled(true);
      deleteBtn.setEnabled(false);
      deleteAllBtn.setEnabled(true);
      recordTable.clearSelection();

      AdminUtils.idle(client);
   }

   /**
    * Load the tables which correspond the selected database into the
    * combo-box.  
    */
   private void reloadDatabaseTables()
   {
      Object item = databaseCombo.getSelectedItem();
      if (item == null)
      {
         return;
      }

      String database = (String) item;
      DefaultComboBoxModel model = databaseTables.get(database);

      if (model == null)
      {
         AdminUtils.busy(client);

         // create combo-box model containing tables
         String[] tables = adm.getDatabaseTables(database);
         model = new DefaultComboBoxModel();
         if (tables == null)
         {
            client.beep("Cannot load tables of the database " + database);
         }
         else
         {
            for (String table : tables)
            {
               model.addElement(table);
            }
         }
         // even if we have an error, put an empty model to the map in order
         // to not interrogate the sever about this database anymore
         databaseTables.put(database, model);

         AdminUtils.idle(client);
      }

      tableCombo.setModel(model);
   }

   /**
    * Handles "add" button in the "add record" dialog. Adds a record to the
    * record table.
    */
   private void handleAddRecord()
   {
      // validate the record ID
      Long id = validateRecordId(recordIdField, "Not a valid record ID");
      if (id == null)
      {
         return;
      }

      RecordInfo info = new RecordInfo(
                                     (String) databaseCombo.getSelectedItem(),
                                     (String) tableCombo.getSelectedItem(),
                                     id);
      dataModel.addRecord(info);

      recordTable.repaint();
      deleteAllBtn.setEnabled(true);
      lockBtn.setEnabled(true);

      // close the dialog
      handleCancelRecord();
   }

   /**
    * Handles "cancel" button in the "add record" dialog.
    */
   private void handleCancelRecord()
   {
      client.ok();

      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
   }

   /**
    * Validate the given text. Its content are expected to be a valid positive
    * long value. If the field has no content, it will return
    * <code>null</code>; also, the specified error will be displayed.
    * <p>
    * Else, it will return the field content.
    *
    * @param    jtf
    *           The source text field
    * @param    err
    *           The error to be displayed if the field has no content.
    *
    * @return   converted value or <code>null</code> if field is invalid.
    */
   private Long validateRecordId(JTextField jtf, String err)
   {
      Long lv = validateLongField(jtf, err);

      if (lv == null)
      {
         return null;
      }

      // ensure the value is positive
      if (lv <= 0)
      {
         // set initial focus on the field
         jtf.requestFocusInWindow();

         // alarm
         client.beep(err);
         return null;
      }

      return lv;
   }

   /**
    * Call the "add record" dialog.
    */
   private void add()
   {
      if (databaseTables == null)
      {
         AdminUtils.busy(client);

         // load the list of databases
         String[] databases = adm.getDatabases();
         databaseTables = new LinkedHashMap<String, DefaultComboBoxModel>();
         for (String database : databases)
         {
            databaseTables.put(database, null);
         }
         databaseCombo.setModel(new DefaultComboBoxModel(databases));
         reloadDatabaseTables();

         AdminUtils.idle(client);
      }


      // display modal dialog
      dialog = AdminUtils.modalDialog(client,
                                      "Add record to lock",
                                      addRecordPanel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancelRecord();
      }
   }

   /**
    * Delete the selected record(s) from the record table.
    */
   private void delete()
   {
      int[] rows = recordTable.getSelectedRows();
      dataModel.deleteRecords(rows);

      recordTable.clearSelection();
      recordTable.repaint();

      boolean anyRecords = dataModel.getRowCount() != 0;
      deleteAllBtn.setEnabled(anyRecords);
      lockBtn.setEnabled(anyRecords);
   }

   /**
    * Delete all record(s) from the record table.
    */
   private void deleteAll()
   {
      dataModel.deleteAllRecords();
      recordTable.clearSelection();
      recordTable.repaint();
      deleteAllBtn.setEnabled(false);
      lockBtn.setEnabled(false);
   }

   /**
    * Provides a table data model to represent session descriptors.
    */
   private class RecordInfoModel
   extends AbstractTableModel
   {
      /** Column names. */
      private final String[] columnNames =
      {
         "Record ID",
         "Database",
         "Table Name"
      };
      
      /** Default column sizes. */
      private final int columnSizes[] =
      {
         50,
         50,
         200
      };
      
      /** Column data types. */
      private final Class[] columnTypes =
      {
         Long.class,
         String.class,
         String.class
      };
      
      /**
       * Construct an instance.
       */
      public RecordInfoModel()
      {
         this.data = new ArrayList<RecordInfo>();
      }
      
      /** Core data for the table. */
      private java.util.List<RecordInfo> data = null;
   
      /**
       * Gets the number of columns in the table model.
       *
       * @return   The number of columns.
       */
      public int getColumnCount()
      {
         return columnNames.length;
      }
   
      /**
       * Gets the width of the given column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The width in pixels.
       */
      public int getColumnWidth(int col)
      {
         return columnSizes[col];
      }
   
      /**
       * Gets the number of rows in the table model.
       *
       * @return   The number of rows.
       */
      public int getRowCount()
      {
         return data.size();
      }

      /**
       * Gets the name of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column name.
       */
      public String getColumnName(int col)
      {
         return columnNames[col];
      }

      /**
       * Get the array of records in the table model.
       *
       * @return   See above.
       */
      public RecordInfo[] getRecords()
      {
         return data.toArray(new RecordInfo[data.size()]);
      }

      /**
       * Add a record to the table model (if it does not already present). 
       *
       * @param record
       *        Record to add.
       */
      public void addRecord(RecordInfo record)
      {
         if (!data.contains(record))
         {
            data.add(record);
         }
      }

      /**
       * Remove the specified records from the table model.
       *
       * @param indexes
       *        Indexes of the records in the table to delete.
       */
      public void deleteRecords(int[] indexes)
      {
         for (int i = indexes.length - 1; i >= 0; i--)
         {
            data.remove(indexes[i]);
         }
      }

      /**
       * Remove all records from the table model.
       */
      public void deleteAllRecords()
      {
         data.clear();
      }
   
      /**
       * Get the cell value to be displayed.
       *
       * @param    col
       *           The column index (0-based).
       * @param    row
       *           The row index (0-based).
       */
      public Object getValueAt(int row, int col)
      {
         Object         datum = null;
         RecordInfo info  = data.get(row);
         
         switch (col)
         {
            case 0:
               datum = info.getRecordID();
               break;
            case 1:
               datum = info.getDatabaseName();
               break;
            case 2:
               datum = info.getTableName();
               break;
         }
         
         return datum;
      }
   
      /**
       * Gets the data type of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column type.
       */
      public Class getColumnClass(int col)
      {
         return columnTypes[col];
      }
   }
}