/*
** Module   : AdminBulkAuthForm.java
** Abstract : The "Bulk Auth Mode Update" dialog.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description-------------------------------
** 001 SVL 20100920 Created the initial version which allows bulk update of 
**                  user and group account settings.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.*;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.border.EtchedBorder;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import com.goldencode.p2j.security.*;

/**
 * The "Bulk Auth Mode Update" dialog.
 */
public class AdminBulkAuthForm
{
   /** Combo box for selecting the target type of accounts. */
   private JComboBox comboTargetAccounts = null;

   /** Auth settings combo boxes. */
   private AuthModeComboBoxes authElements = null;

   /** Main panel for the dialog. */
   private JPanel mainPanel = null;

   /** The main dialog. */ 
   private JDialog dialog = null;

   /** Admin Server APIs. */
   private final AdminExports adm;

   /** Related instance of AdminClient. */
   private final AdminClient client;

   /** The title of the dialog. */
   public static final String DIALOG_TITLE = "Bulk Auth Mode Update";

   /** The title of the message box for displaying results. */
   public static final String UPDATE_RESULT_TITLE = "Update result";

   /**
    * Constructor.
    *
    * @param client
    *        Related instance of AdminClient.
    */
   public AdminBulkAuthForm(AdminClient client)
   {
      this.client = client;
      adm = client.getExports();
   }

   /**
    * Show the dialog.
    */
   public void showDialog()
   {
      if (mainPanel == null)
      {
         buildMainPanel();
      }

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, DIALOG_TITLE, mainPanel);
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleClose();
      }
   }

   /**
    * Handle Close button press in the dialog.
    */
   private void handleClose()
   {
      // clear error messages
      client.ok();
      AdminUtils.idle(client);

      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
   }

   /**
    * Build the main dialog panel.
    */
   private void buildMainPanel()
   {
      JLabel lblTargetAccounts = new JLabel(" Type of Accounts to Modify ");
      comboTargetAccounts = new JComboBox();
      comboTargetAccounts.addItem(
                      "All User Accounts (except the one you are using now)");
      comboTargetAccounts.addItem("All Group Accounts");

      authElements = new AuthModeComboBoxes(adm,
                      AuthModeComboBoxes.TargetScreen.BULK_AUTH_MODE_UPDATE);
      authElements.reset();

      comboTargetAccounts.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent e)
         {
            authElements.updateInheritString(
                  comboTargetAccounts.getSelectedIndex() == 0 ?
                  AuthModeComboBoxes.TargetScreen.USER_ACCOUNT_SCREEN :
                  AuthModeComboBoxes.TargetScreen.GROUP_ACCOUNT_SCREEN);
         }
      });

      JButton btnClose = new JButton("Close");
      JButton btnUpdate = new JButton("Update Auth Settings");

      btnClose.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent e)
         {
            handleClose();
         }
      });

      btnUpdate.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent e)
         {
            handleUpdateAccounts();
         }
      });

      // do layout
      JPanel pnlBottom = AdminUtils.groupComponents(
                                   new Component[][]{{btnClose, btnUpdate}});
      GroupLayout layout = (GroupLayout) pnlBottom.getLayout();
      layout.setAutoCreateContainerGaps(false);

      JPanel pnlTop = AdminUtils.groupComponents(new Component[][]
      {
         {lblTargetAccounts, comboTargetAccounts},
         {authElements.getAuthModeLabel(), authElements.getAuthModeCombo()},
         {authElements.getAuthPluginsLabel(),
          authElements.getAuthPluginsCombo()}
      });
      pnlTop.setBorder(new TitledBorder
            (new EtchedBorder(EtchedBorder.RAISED), DIALOG_TITLE));

      mainPanel = AdminUtils.groupComponents(
                                    new Component[][]{{pnlTop}, {pnlBottom}});
   }

   /**
    * Handle Update button press in the dialog.
    */
   private void handleUpdateAccounts()
   {
      boolean res = AdminUtils.confirmDialog(client,
            "Confirm Bulk Update",
            "WARNING, authentication mode will be updated for ALL " +
            "accounts of the specified type.\nDo you want to continue?");

      if (!res)
         return;

      AdminUtils.busy(client);
      boolean group = comboTargetAccounts.getSelectedIndex() == 1;
      int updatedAccounts = adm.bulkAuthModeUpdate(
            group ? Account.ACC_GROUP : Account.ACC_USER,
            group ? null : new String[]{AdminLogon.getLoggedInUser()},
            authElements.getAuthMode(),
            authElements.getAuthMode() == SecurityConstants.AUTH_MODE_CUSTOM ?
               authElements.getAuthPlugin() : null);
      AdminUtils.idle(client);

      if (updatedAccounts < 0)
      {
         JOptionPane.showMessageDialog(dialog,
               "There was an error while updating authentication mode.",
               UPDATE_RESULT_TITLE,
               JOptionPane.ERROR_MESSAGE);
      }
      else
      {
         JOptionPane.showMessageDialog(dialog,
               String.format("%d accounts were updated.", updatedAccounts),
               UPDATE_RESULT_TITLE,
               JOptionPane.INFORMATION_MESSAGE);
      }
   }
}