/*
** Module   : AdminConsole.java
** Abstract : Console menu support for the admin session
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090609   @42624 Created the first version of the file.
** 002 GES 20090617   @42757 Added session screen support.
** 003 GES 20090721   @43301 Added record locks screen support.
** 004 NVS 20090816   @43648 Added shutdown menu item support.
** 005 NVS 20090820   @43706 Shutdown operation allows the server to severe
**                           the connection. Version in About is 21.02.
** 006 NVS 20091117   @44415 Safe directory reload support added.
** 007 NVS 20091118   @44427 Safe directory backup support added.
** 008 SVL 20091126   @44451 Added "acquire locks" screen support.
** 009 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * This is a Console menu management class for the admin session.
 * <p> 
 * It provides lazy Console screens creation and dispatches menu actions
 * to the specialized handlers by the action.
 */
public class AdminConsole
implements AdminNotify
{
   /** Related instance of AdminClient */
   private AdminClient client = null;

   /** Console menu */
   private JMenu cns = null;
   
   /** Console --&gt; Backup Directory menu item. */
   private JMenuItem cnsBck = null; 

   /** Console --&gt; Reload Directory menu item. */
   private JMenuItem cnsRld = null; 

   /** Console --&gt; Sessions menu item. */
   private JMenuItem cnsSess = null;
   
   /** Console --&gt; Shutdown menu item. */
   private JMenuItem cnsShtd = null;
   
   /** Console --&gt; Record Locks menu item. */
   private JMenuItem cnsLocks = null;

   /** Console --&gt; Acquire locks menu item. */
   private JMenuItem cnsAcquireLocks = null;
   
   /** Session management screen. */
   private AdminSessions sessScr = null;
   
   /** Record lock management screen. */
   private AdminLocks locksScr = null;

   /** Screen that allows to lock specific records. */
   private AdminAcquireLocks acquireLocksScr = null;
   
   /** Admin Server APIs */
   protected AdminExports adm = null;
   
   // dialog specific widgets
   
   /** modal dialog panel for Reload Directory function */
   private JPanel rdPanel = null;
   
   /** information label */
   private JLabel rdlInfo = null;
   
   /** information label 2 */
   private JLabel rdlInfo2 = null;

   /** lock duration input field */
   private JTextField rdfTime = null;
   
   /** Lock button */
   private JButton rdbLock = null;

   /** Reload button */
   private JButton rdbReload = null;
   
   /** Cancel button */
   private JButton rdbCancel = null;

   /** modal dialog panel for Backup Directory function */
   private JPanel bdPanel = null;
   
   /** information label */
   private JLabel bdlInfo = null;
   
   /** information label 2 */
   private JLabel bdlInfo2 = null;
   
   /** file name input field */
   private JTextField bdfName = null;
   
   /** Backup button */
   private JButton bdbBackup = null;
   
   /** holds the most recent modal dialog instance */
   private JDialog dialog = null;

   /**
    * Constructor.
    */
   AdminConsole()
   {
   }

   /** 
    * Initializes useful references.
    *
    * @param  client
    *         instance of AdminClient
    *
    * @param  cnsMenu
    *         reference to Console menu
    */
   void init(AdminClient client, JMenu cnsMenu)
   {
      // store useful references
      this.client = client;
      this.cns    = cnsMenu;
      
      // register for logoff notifications (safe to make multiple calls)
      client.registerLogoffListener(this);

      // register for target notifications (safe to make multiple calls)
      client.registerTargetListener(this);
      
      // create Console menu items
      
      JMenuItem tmp = null;
      
      cnsSess = new JMenuItem("Sessions");
      cnsSess.setEnabled(true);
      cns.add(cnsSess);
      
      cnsLocks = new JMenuItem("Record Locks");
      cnsLocks.setEnabled(true);
      cns.add(cnsLocks);

      cnsAcquireLocks = new JMenuItem("Acquire locks");
      cnsAcquireLocks.setEnabled(true);
      cns.add(cnsAcquireLocks);
      
      tmp = new JMenuItem("Services");
      tmp.setEnabled(false);
      cns.add(tmp);
      
      tmp = new JMenuItem("Logging");
      tmp.setEnabled(false);
      cns.add(tmp);
      
      tmp = new JMenuItem("Audit");
      tmp.setEnabled(false);
      cns.add(tmp);
      
      cnsBck = new JMenuItem("Backup Directory");
      cnsBck.setEnabled(true);
      cns.add(cnsBck);
      
      cnsRld = new JMenuItem("Reload Directory");
      cnsRld.setEnabled(true);
      cns.add(cnsRld);
      
      cnsShtd = new JMenuItem("Shutdown");
      cnsShtd.setEnabled(true);
      cns.add(cnsShtd);
      
      cnsSess.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            lazySessions();
         }
      });
      
      cnsLocks.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            lazyLocks();
         }
      });

      cnsAcquireLocks.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt)
         {
            lazyAcquireLocks();
         }
      });
      
      cnsBck.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            handleBackupDirectory();
         }
      });
      
      cnsRld.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            handleReloadDirectory();
         }
      });
      
      cnsShtd.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            handleShutdown();
         }
      });
   }
   
   /** 
    * Notifies about successful logoff or target change and gives chance 
    * to release resources.
    * <p>
    * In both cases the reaction is the full cleanup.
    *
    * @param    eventCode
    *           The event code, one of the NOTIFY_* constants.
    */
   public void notifyEvent(int eventCode)
   {
      // filter out unwanted events
      if (eventCode != AdminNotify.NOTIFY_LOGOFF &&
          eventCode != AdminNotify.NOTIFY_TARGET)
      {
         return;
      }
      
      // destroy the existing Console menu items
      cns.removeAll();
      
      // release references
      sessScr = null;
      
      // reinit menu
      init(client, cns);
   }
   
   /** 
    * Handles lazy sessions list menu initialization.
    */
   private void lazySessions() 
   {
      // initialize groups screen
      sessScr = new AdminSessions(client); 
      
      // allow direct action handler invocation
      ActionListener[] lst = cnsSess.getActionListeners();
      
      for (int i = 0; i < lst.length; i++)
      {
         cnsSess.removeActionListener(lst[i]);
      }
      
      cnsSess.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            sessScr.manageSessions();
         }
      });

      // handle this action once
      sessScr.manageSessions();
   }

   /** 
    * Handles lazy record locks list menu initialization.
    */
   private void lazyLocks() 
   {
      // initialize groups screen
      locksScr = new AdminLocks(client);
      
      // allow direct action handler invocation
      ActionListener[] lst = cnsLocks.getActionListeners();
      
      for (int i = 0; i < lst.length; i++)
      {
         cnsLocks.removeActionListener(lst[i]);
      }
      
      cnsLocks.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            locksScr.manageLocks();
         }
      });

      // handle this action once
      locksScr.manageLocks();
   }

   /**
    * Handles lazy "acquire locks" list menu initialization.
    */
   private void lazyAcquireLocks()
   {
      // initialize groups screen
      acquireLocksScr = new AdminAcquireLocks(client);

      // allow direct action handler invocation
      ActionListener[] lst = cnsAcquireLocks.getActionListeners();

      for (int i = 0; i < lst.length; i++)
      {
         cnsAcquireLocks.removeActionListener(lst[i]);
      }

      cnsAcquireLocks.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt)
         {
            acquireLocksScr.manageLocks();
         }
      });

      // handle this action once
      acquireLocksScr.manageLocks();
   }

   /**
    * Handles Backup Directory menu selection.
    */
   private void handleBackupDirectory() 
   {
      // create "Backup Directory" modal dialog panel with Group layout
      bdlInfo  = new JLabel
         (" Enter the backup file name and press the Backup button ");
      bdlInfo2 = new JLabel("");
      bdfName = new JTextField(32);
      bdbBackup = new JButton("Backup");

      bdPanel = AdminUtils.groupComponents(new Component[][] {
                  { bdlInfo,   bdlInfo2 },
                  { bdfName,   bdbBackup },
                });
      
      // Backup button press handling
      bdbBackup.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")          
         public void actionPerformed(ActionEvent evt) 
         {
            handleBackup();
         }
      });
      
      // focus on backup file name field
      bdfName.requestFocusInWindow();

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "Backing up Directory", 
                                      bdPanel);
      
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         dialog = null;
      }
   }
   
   /** 
    * Handles Reload Directory menu selection.
    */
   private void handleReloadDirectory() 
   {
      // create "Reload Directory" modal dialog panel with Group layout
      rdlInfo  = new JLabel(" Enter the lock duration and press the Lock button ");
      rdlInfo2 = new JLabel("");
      rdfTime = new JTextField(4);
      rdbLock = new JButton("Lock");
      rdbReload = new JButton("Reload");
      rdbCancel = new JButton("Cancel");

      rdPanel = AdminUtils.groupComponents(new Component[][] {
                  { rdlInfo,   rdlInfo2 },
                  { rdfTime,   rdbLock },
                  { rdbReload, rdbCancel },
                });
      
      // Lock button press handling
      rdbLock.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")          
         public void actionPerformed(ActionEvent evt) 
         {
            handleReloadLock();
         }
      });
      
      // Reload button press handling
      rdbReload.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")          
         public void actionPerformed(ActionEvent evt) 
         {
            handleReload();
         }
      });
      
      // Cancel button press handling
      rdbCancel.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")          
         public void actionPerformed(ActionEvent evt) 
         {
            handleReloadCancel();
         }
      });
      
      // focus on duration field
      rdfTime.requestFocusInWindow();

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "Reloading Directory", 
                                      rdPanel);
      
      rdbReload.setEnabled(false);
      rdbCancel.setEnabled(false);
      
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleReloadCancel();
      }
   }
   
   /** 
    * Handles Shutdown menu selection.
    */
   private void handleShutdown() 
   {
      if (!AdminUtils.confirmDialog(client, "Server Shutdown Confirmation",
         "You are about to shut the server down. " + "Continue?"))
                                    
      {
         // declined
         return;
      }
      
      adm = client.getExports();

      // busy
      AdminUtils.busy(client);
      
      boolean rc = adm.shutdown();
      if (!rc)
      {
         client.beep("no rights to initiate shutdown");
         
         // idle
         AdminUtils.idle(client);
         
         return;
      }
      
      // idle
      AdminUtils.idle(client);

      // the session disconnection notification is coming
   }

   /** 
    * Handles Reload Directory dialog's Lock button press.
    */
   private void handleReloadLock() 
   {
      // focus on duration field
      rdfTime.requestFocusInWindow();

      // get the lock duration
      String dur = rdfTime.getText();
      if (dur.trim().length() == 0)
      {
         client.beep("");
         return;
      }
      
      int sec = 0;
      try
      {
         sec = Integer.parseInt(dur);
      }
      catch (NumberFormatException nfe)
      {
         client.beep("");
         return;
      }
      
      if (sec <=1)
      {
         client.beep("");
         return;
      }

      // change widgets sta
      rdlInfo.setText(String.
              format("Directory is temporarily locked for %d seconds.", sec));
      rdfTime.setEnabled(false);
      rdbLock.setEnabled(false);
      rdbReload.setEnabled(true);
      rdbCancel.setEnabled(true);
      
      // lock the directory for reloading      
      adm = client.getExports();

      // busy
      AdminUtils.busy(client);
      
      boolean rc = adm.lockForRereadDir(sec);
      
      // idle
      AdminUtils.idle(client);
      
      // signal errors
      if (!rc)
      {
         client.alarm();
      }
      else
      {
         client.ok();
      }
   }
   
   /** 
    * Handles Reload Directory dialog's Reload button press.
    */
   private void handleReload() 
   {
      // clear error messages
      client.ok();
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;

      // reread the directory      
      adm = client.getExports();

      // busy
      AdminUtils.busy(client);
      
      int rc = adm.rereadDirectory();
      
      // idle
      AdminUtils.idle(client);
      
      // signal errors
      if (rc != 0)
      {
         client.alarm();
      }
      else
      {
         client.ok();
      }
   }
   
   /** 
    * Handles Reload Directory dialog's Cancel button press.
    */
   private void handleReloadCancel() 
   {
      // clear error messages
      client.ok();
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;

      // lock the directory for reloading      
      adm = client.getExports();

      // busy
      AdminUtils.busy(client);
      
      adm.cancelRereadDir();
      
      // idle
      AdminUtils.idle(client);
      client.ok();
   }

   /** 
    * Handles Backup Directory dialog's Backup button press.
    */
   private void handleBackup() 
   {
      // focus on filename field
      bdfName.requestFocusInWindow();

      // get the file name duration
      String backup = bdfName.getText();
      if (backup.trim().length() == 0)
      {
         client.beep("");
         return;
      }
      
      // clear error messages
      client.ok();
      
      // issue the relevant API call
      adm = client.getExports();

      // busy
      AdminUtils.busy(client);
      
      boolean rc = adm.backupDirectory(backup);
      
      // signal errors
      if (!rc)
      {
         client.alarm();
      }
      else
      {
         client.ok();
      }
      
      // idle
      AdminUtils.idle(client);
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
   }
}
