/*
** Module   : AdminLocks.java
** Abstract : record locks list screen
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description--------------------
** 001 GES 20090721   @43304 Created the first version to provide a list of
**                           record locks.
** 002 NVS 20090723   @43352 Tables don't allow moving columns.
** 003 NVS 20090729   @43429 Implemented locks printing.
** 004 NVS 20090803   @43482 Using a better approach to printing with extended
**                           detail choice.
** 005 ECF 20090801   @43504 Added database name to table model.
** 006 CA  20091119   @44436 Record lockers are retrieved as session tokens 
**                           instead of session ID's; also, the lockers are 
**                           displayed in the format subject:sessionID.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.awt.*;
import java.awt.event.*;
import java.awt.print.PrinterException;
import java.text.MessageFormat;
import javax.swing.*;
import javax.swing.table.*;

import com.goldencode.p2j.admin.RecordLockInfo;
import com.goldencode.p2j.security.SessionToken;

/**
 * Provides a screen that lists the record locks for the current server. 
 */
public class AdminLocks 
extends AdminScreen
implements PrintRequest
{
   /** Screen switcher - record locks screen ID. */
   public static final String SCREEN_LOCKS = "adm_locks";
   
   /** Screen title. */
   private static final String TITLE_LOCKS = "Record Locks";
   
   /** Record locks list. */
   private JTable table = null;

   /** Record lock list data model. */
   private LockInfoModel list = null;

   /** Subpanel that holds buttons and other controls. */
   private JPanel ctrlPanel = null;
   
   /** Refresh button. */
   private JButton refresh = null;

   /** Filter button. */
   private JButton filter = null;
   
   /** Locks table sorter */
   private TableRowSorter sorter = null;
   
   /** Cache the last filter string. */
   private String lastFilter = null;
   
   /** 
    * Constructor.
    *
    * @param    client
    *           Core UI module of which we are a part.
    */
   AdminLocks(AdminClient client)
   {
      // store useful references
      this.client = client;
      adm = client.getExports();
      
      // create the main screen and add it to the swapper
      mainPanel = new JPanel();
      client.add(mainPanel, SCREEN_LOCKS, this);
      mainPanel.setLayout(new BorderLayout());
      
      table = new JTable();
      table.getTableHeader().setReorderingAllowed(false);
      
      JScrollPane scrollPane = new JScrollPane(table);
      scrollPane.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
      table.setFillsViewportHeight(true);
      table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      
      mainPanel.add(scrollPane, BorderLayout.CENTER);

      // create control widgets
      ctrlPanel = new JPanel(new GridLayout(1, 0));
      ctrlPanel.setBorder(BorderFactory.createEmptyBorder(0,10,10,10));
      
      refresh = new JButton("Refresh");
      filter  = new JButton("Filter");
      
      ctrlPanel.add(refresh);
      ctrlPanel.add(filter);
      
      mainPanel.add(ctrlPanel, BorderLayout.SOUTH);
      
      // filter button
      filter.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent evt) 
         {
            filter();
         }
      });
      
      // refresh button
      refresh.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent evt) 
         {
            refresh();
         }
      });
      
      // create the array of components for the state machine
      state = new JComponent[]
      {
         table,
         ctrlPanel,
         refresh,
         filter
      };
      
      // populate the record lock list upfront
      refresh();
   }
   
   /**
    * Performs printing of the current screen data.
    * This screen prints the current view of the locks table.
    */
   public void print()
   {
      MessageFormat header = new MessageFormat("Active Record Locks");
      MessageFormat footer = new MessageFormat("page {0, number}");
      
      try
      {
         table.print(JTable.PrintMode.FIT_WIDTH, header, footer);
      }
      catch (PrinterException pe)
      {
         client.beep("printer exception" + pe);
      }
      catch (SecurityException se)
      {
         // nothing to do
      }
   }
   
   /**
    * Queries the extended printing capability. If the exteneded capability
    * is not provided, the printExtended() method should not be called.
    *
    * @return <code>false</code> to tell the caller the extended printing 
    *         is not implemented.
    */
   public boolean canPrintExtended()
   {
      return false;
   }

   /**
    * Does nothing here.
    */
   public void printExtended()
   {
   }
   
   /** 
    * Refresh the display with the latest list of sessions.
    */
   void refresh()
   {
      // not connected yet
      if (adm == null)
      {
         client.beep("Not connected!");
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      RecordLockInfo[] info = adm.getRecordLockList();
      
      String msg = "Cannot obtain list of record locks.";
      
      if (info != null)
      {
         msg  = "";
         list = new LockInfoModel(info);
         table.setModel(list);
         
         int num = list.getColumnCount();
         
         for (int i = 0; i < num; i++)
         {
            TableColumn column = table.getColumnModel().getColumn(i);
            column.setPreferredWidth(list.getColumnWidth(i));
         }
         
         sorter = new TableRowSorter(list);
         sorter.setRowFilter(null);
         table.setRowSorter(sorter);
         
         lastFilter = null;
         
         // force a default sort on the table name column
         sorter.toggleSortOrder(2);
      }
      
      client.setTargetInfo(AdminClient.TARGET_ERROR, msg);
      
      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Process the filter dialog and create/install the resulting filter into
    * the table sorter.
    */
   private void filter() 
   {
      int    type = JOptionPane.QUESTION_MESSAGE;
      String txt  = "Define the filtering specification below. Press the\n" +
                    "'OK' button without any specification to remove any\n" +
                    "previous filter.";
      String spec = (String) JOptionPane.showInputDialog(client,
                                                         txt,
                                                         "Filter",
                                                         type,
                                                         null,
                                                         null,
                                                         lastFilter);
      
      if (spec == null || spec.trim().length() == 0)
      {
         // remove filtering
         sorter.setRowFilter(null);
         lastFilter = null;
      }
      else
      {
         // install a filter
         sorter.setRowFilter(AdminUtils.createFilter(spec));
         lastFilter = spec;
      }
   }
   
   /** 
    * Presents the refreshed record locks screen.
    */
   void manageLocks() 
   {
      // switch to this screen without changing state of components
      client.show(SCREEN_LOCKS, TITLE_LOCKS);
   }
   
   /**
    * Provides a table data model to represent session descriptors.
    */
   private class LockInfoModel
   extends AbstractTableModel
   {
      /** Column names. */
      private final String[] columnNames =
      {
         "Record ID",
         "Database",
         "Table Name",
         "Session ID(s)",
         "Lock Type",
         "Acquisition Timestamp"
      };
      
      /** Default column sizes. */
      private final int columnSizes[] =
      {
         50,
         50,
         200,
         200,
         50,
         150
      };
      
      /** Column data types. */
      private final Class[] columnTypes =
      {
         Long.class,
         String.class,
         String.class,
         String.class,
         String.class,
         String.class
      };
      
      /**
       * Construct an instance.
       *
       * @param    data
       *           The data to provide to the table.
       */
      public LockInfoModel(RecordLockInfo[] data)
      {
         this.data = data;
      }
      
      /** Core data for the table. */
      private RecordLockInfo[] data = null;
   
      /**
       * Gets the number of columns in the table model.
       *
       * @return   The number of columns.
       */
      public int getColumnCount()
      {
         return columnNames.length;
      }
   
      /**
       * Gets the width of the given column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The width in pixels.
       */
      public int getColumnWidth(int col)
      {
         return columnSizes[col];
      }
   
      /**
       * Gets the number of rows in the table model.
       *
       * @return   The number of rows.
       */
      public int getRowCount()
      {
         return data.length;
      }
   
      /**
       * Gets the name of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column name.
       */
      public String getColumnName(int col)
      {
         return columnNames[col];
      }
   
      /**
       * Get the cell value to be displayed.
       *
       * @param    col
       *           The column index (0-based).
       * @param    row
       *           The row index (0-based).
       */
      public Object getValueAt(int row, int col)
      {
         Object         datum = null;
         RecordLockInfo info  = data[row];
         
         switch (col)
         {
            case 0:
               datum = new Long(info.getRecordID());
               break;
            case 1:
               datum = info.getDatabaseName();
               break;
            case 2:
               datum = info.getTableName();
               break;
            case 3:
               SessionToken[] lockers = info.getLockers();
               
               StringBuilder sb = new StringBuilder();
               
               for (int i = 0; i < lockers.length; i++)
               {
                  if (i > 0)
                     sb.append(", ");
                  
                  sb.append(lockers[i]);
               }
               
               datum = sb.toString();
               break;
            case 4:
               datum = info.getType().toString();
               break;
            case 5:
               datum = String.format("%1$Tm/%1$Td/%1$TY %1$TT %1$Tz",
                                     info.getTimeAcquired());
               break;
         }
         
         return datum;
      }
   
      /**
       * Gets the data type of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column type.
       */
      public Class getColumnClass(int col)
      {
         return columnTypes[col];
      }
   }
}