/*
** Module   : AdminMenu.java
** Abstract : Admin client menu management 
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090116  @41620  Created the first version of the file.
** 002 NVS 20090318  @41632  Added Target --> refresh menu item and
**                           supporting listener.
** 003 NVS 20090323  @41666  Refactored one more time to get rid of static
**                           members.
** 004 NVS 20090326  @41690  Added Help --> Show errors menu item creation
**                           and state management. Javadoc changes.
** 005 NVS 20090428  @42039  Reworked logoff notification.
** 006 NVS 20090429  @42081  Capitalized menu items.
** 007 CA  20090513  @42175  Added getter for the AdminAccounts instance.
** 008 NVS 20090609  @42625  Separated Console menu into AdminConsole.java
** 009 NVS 20090714  @43161  Added temporary test code
** 010 CA  20090714  @43156  Build the ACL menu accordingly to new specs.
**                           Separated ACL menu into AdminACLMenu.java.
** 011 NVS 20090717  @43233  Removed temporary test code
** 012 CA  20090720  @43244  Renamed getACLAdmin to getACL.
** 013 NVS 20090729  @43430  Added Print menu.
** 014 NVS 20090803  @43483  Extended detail printing provides a better menu  
**                           control.
** 015 NVS 20090807  @43550  Refactored in support of custom menu extentions.
** 016 NVS 20091007  @44113  Implemented admin permissions profile handling.
**                           Major menu items can be selectively disabled.
** 017 SVL 20111209          "Runtime Settings" menu initialization was
**                           extracted to a separate class.
** 018 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.awt.event.*;
import java.util.List;
import java.util.LinkedList;
import javax.swing.*;

import com.goldencode.p2j.admin.*;

/**
 * This class builds and subsequently manages the admin client menu.
 *
 * @author NVS
 */
public class AdminMenu
implements AdminNotify
{
   /** Related instance of AdminClient */
   private AdminClient client = null;

   /** Related instance of AdminLogon */
   private AdminLogon logon = null;

   /** Related instance of AdminTarget */
   private AdminTarget target = null;

   /** menu bar */
   private JMenuBar mainMenu = null;
   
   /** Console menu */
   private JMenu cns = null;
   
   /** Accounts menu */
   private JMenu acc = null;

   /** Access Control menu */
   private JMenu acl = null;

   /** Runtime Settings menu */
   private JMenu rts = null;
   
   /** General Configuration menu */
   private JMenu gen = null;

   /** Print menu */
   private JMenuItem prt = null;

   /** Logoff menu */
   private JMenuItem off = null;

   /** Filler to confine the Logoff button */
   private JMenuItem flr = null;
   
   /** Target menu */
   private JMenu tgt = null;

   /** Help menu */
   private JMenu hlp = null;

   /** Target --&gt; live menu item */
   private JRadioButtonMenuItem targetLive = null;
   
   /** Target --&gt; open menu item */
   private JRadioButtonMenuItem targetOpen = null;
   
   /** Target --&gt; new menu item */
   private JRadioButtonMenuItem targetNew = null;
   
   /** Target --&gt; refresh menu item */
   private JMenuItem targetRefresh  = null; 

   /** groups radio buttons of Target menu */
   private ButtonGroup targetGroup = null;
   
   /** Help --&gt; show messages menu item */
   private JMenuItem hlpMsg = null; 
   
   /** Help --&gt; clear messages menu item */
   private JMenuItem hlpClr = null; 
   
   /** Help --&gt; About menu item */
   private JMenuItem hlpAbo = null;
   
   /** The handler of Console menu. */
   private AdminConsole aCns = null;
   
   /** The handler of Accounts menu. */
   private AdminAccounts aAcc = null;
   
   /** The handler of Access Control menu. */
   private AdminACLMenu aAcl = null;

   /** The handler of Runtime Settings menu. */
   private AdminRuntimeSettings aRts = null;
   
   /** storage for saving state of the menu */
   private boolean[] state = null;
   
   /** major menu items as a list */
   List<JMenuItem> menuList = new LinkedList<JMenuItem>();
   
   /** 
    *  Constructs the menu hierarchy.
    */
   AdminMenu() 
   {
      // Console menu
      cns = new JMenu("Console");

      // Accounts menu
      acc = new JMenu("Accounts", false);
      
      // Access Control menu
      acl = new JMenu("Access Control", false);
      acl.add(new JMenuItem("Show by Account"));
      acl.add(new JMenu("Show by Resource Type"));
      acl.add(new JMenuItem("Show All"));
      acl.add(new JMenu("Manage ACL by Server"));

      // Runtime Settings menu
      rts = new JMenu("Runtime Settings", false);
      
      // General Configuration menu
      gen = new JMenu("Configuration");
      gen.add(new JMenuItem("Authentication"));
      gen.add(new JMenuItem("Resources"));
      gen.add(new JMenuItem("Audit"));
      gen.add(new JMenuItem("Holidays"));
      
      // Target menu
      tgt = new JMenu("Target");
      targetLive = new JRadioButtonMenuItem("Live");
      tgt.add(targetLive);
      targetOpen = new JRadioButtonMenuItem("Open ...");
      tgt.add(targetOpen);
      targetNew  = new JRadioButtonMenuItem("New  ...");
      tgt.add(targetNew );
      targetGroup = new ButtonGroup();
      targetGroup.add(targetLive);
      targetGroup.add(targetOpen);
      targetGroup.add(targetNew );

      tgt.addSeparator();
      targetRefresh = new JMenuItem("Refresh");
      targetRefresh.setEnabled(false);
      tgt.add(targetRefresh);

      // Print menu
      prt = new JMenuItem("Print");
      
      // Connect menu
      off = new JMenuItem("Log Off");
      
      // Filler
      flr = new JMenuItem("");
      
      // Help menu
      hlp = new JMenu("Help");
      hlpMsg = new JMenuItem("Show Messages");
      hlp.add(hlpMsg);
      hlpClr = new JMenuItem("Clear Messages");
      hlp.add(hlpClr);
      hlp.addSeparator();
      hlpAbo = new JMenuItem("About AdminClient");
      hlp.add(hlpAbo);
   }

   /** 
    * Initializes useful references.
    *
    * @param  client_
    *         instance of AdminClient
    *
    * @param  logon_
    *         instance of AdminLogon
    *
    * @param  target_
    *         instance of AdminTarget
    *
    */
   void init(AdminClient client_, AdminLogon logon_, AdminTarget target_)
   {
      // this section is for some session control related classes
      client = client_;
      logon = logon_;
      target = target_;
      
      prt.addActionListener(
         // this class handles print requests                 
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               client.printScreen();
            }
         });
      
      off.addActionListener(
         // this class handles logoff choice                               
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               logon.logoff();
            }
         });

      targetLive.addActionListener(
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               target.goLive();
            }
         });
      
      targetOpen.addActionListener(
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               target.goOpen();
            }
         });
      
      targetNew .addActionListener(
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               target.goNew();
            }
         });

      targetRefresh.addActionListener(
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               target.refresh();
            }
         });
      
      hlpMsg.addActionListener(
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               client.showMessages();
            }
         });

      hlpClr.addActionListener(
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               client.clearMessages();
            }
         });
      
      hlpAbo.addActionListener(
         new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               AdminUtils.aboutDialog(client);
            }
         });
      
      // this section is for the rest of the menu that provides on-demand
      // (lazy) screen creation
      aCns = new AdminConsole();
      aCns.init(client, cns);
      
      aAcc = new AdminAccounts();
      aAcc.init(client, acc);
      
      aAcl = new AdminACLMenu();
      aAcl.init(client, acl);

      aRts = new AdminRuntimeSettings();
      aRts.init(client, rts);
      
      // create the menu list
      menuList.add(cns);
      menuList.add(acc);
      menuList.add(acl);
      menuList.add(rts);
      menuList.add(gen);
      menuList.add(tgt);
      menuList.add(prt);
      menuList.add(off);
      menuList.add(flr);
      menuList.add(hlp);
      
      // construct the menu bar
      buildMenu(null);
      
      // register for logoff notifications
      client.registerLogoffListener(this);
   }

   /** 
    * Constructs the menu bar based on the current contents of the menu list,
    * which may have been modified by the custom extensions plugin.
    *
    * @param profile
    *        admin rights profile or <code>null</code>
    */
   void buildMenu(AdminDef.Profile[] profile)
   {
      // construct the menu bar
      mainMenu = new JMenuBar();
      
      // list of menu extensions
      List<JMenuItem> ext = new LinkedList<JMenuItem>();
      
      for (JMenuItem m : menuList)
      {
         mainMenu.add(m);
         
         if (!m.equals(cns) &&
             !m.equals(acc) &&
             !m.equals(acl) &&
             !m.equals(rts) &&
             !m.equals(gen) &&
             !m.equals(tgt) &&
             !m.equals(prt) &&
             !m.equals(off) &&
             !m.equals(flr) &&
             !m.equals(hlp))
         {
            ext.add(m);
         }
      }
      
      // process profile and enable/disab;e the menu items
      if (profile != null)
      {
         for (int i = 0; i < profile.length; i++)
         {
            if (profile[i].name.equals("/console"))
            {
               cns.setEnabled(profile[i].permission);
            }
            
            if (profile[i].name.equals("/accounts"))
            {
               acc.setEnabled(profile[i].permission);
            }
            
            if (profile[i].name.equals("/access_control"))
            {
               acl.setEnabled(profile[i].permission);
            }
            
            if (profile[i].name.equals("/extension"))
            {
               for (JMenuItem m : ext)
               {
                  m.setEnabled(profile[i].permission);
               }
            }
            
            if (profile[i].name.equals("/runtime"))
            {
               rts.setEnabled(profile[i].permission);
            }
            
            if (profile[i].name.equals("/configuration"))
            {
               gen.setEnabled(profile[i].permission);
            }
            
            if (profile[i].name.equals("/target"))
            {
               tgt.setEnabled(profile[i].permission);
            }
         }
      }
   }
   
   /** 
    * Notifies about successful logoff and gives chance to release resources.
    *
    * @param eventCode
    *        event code, one of the NOTIFY_* constants
    */
   public void notifyEvent(int eventCode)
   {
      // filter out unwanted events
      if (eventCode != AdminNotify.NOTIFY_LOGOFF)
      {
         return;
      }
      
      disconnectedState();
   }
   
   /**
    * Get the current {@link AdminAccounts} instance.
    * 
    * @return the current {@link AdminAccounts} instance.
    */
   AdminAccounts getAccounts()
   {
      return aAcc;
   }
   
   /**
    * Get the current {@link AdminACLMenu} instance.
    * 
    * @return the current {@link AdminACLMenu} instance.
    */
   AdminACLMenu getACL()
   {
      return aAcl;
   }
   
   /** 
    * Returns the instance of the menu bar. 
    *
    * @return instance of menu bar to use as the master menu in the applet 
    */
   JMenuBar getMenu()
   {
      return mainMenu;
   }

   /** 
    * Returns the menu list. 
    *
    * @return the menu list which is modifiable 
    */
   List<JMenuItem> getMenuList()
   {
      return menuList;
   }

   /** 
    * Sets the menu items in a state appropriate for the disconnected 
    * state of the applet.
    */
   void disconnectedState()
   {
      cns.setEnabled(false);
      acc.setEnabled(false);
      acl.setEnabled(false);
      rts.setEnabled(false);
      gen.setEnabled(false);
      targetGroup.clearSelection();
      tgt.setEnabled(false);
      prt.setEnabled(false);
      off.setEnabled(false);
      hlpMsg.setEnabled(false);
      hlpClr.setEnabled(false);
   }
   
   /** 
    * Sets the menu items in a state appropriate for the connected 
    * state of the applet.
    */
   void connectedState()
   {
      cns.setEnabled(false);
      acc.setEnabled(false);
      acl.setEnabled(false);
      rts.setEnabled(false);
      gen.setEnabled(false);
      targetGroup.clearSelection();
      tgt.setEnabled(true);
      prt.setEnabled(false);
      off.setEnabled(true);
      hlpMsg.setEnabled(true);
      hlpClr.setEnabled(false);
   }

   /** 
    * Sets the menu items in a state appropriate for the targeted 
    * state of the applet.
    */
   void targetedState()
   {
      cns.setEnabled(target != null && target.isLive());
      acc.setEnabled(true);
      acl.setEnabled(true);
      rts.setEnabled(true);
      gen.setEnabled(false);        // temporary - menu not implemented  
      tgt.setEnabled(true);
      prt.setEnabled(true);
      off.setEnabled(true);
      hlpMsg.setEnabled(true);
      hlpClr.setEnabled(false);
   }
   
   /** 
    * Gets the Target menu. 
    *
    * @return  Target menu
    */
   JMenu getTarget()
   {
      return tgt;
   }
   
   /** 
    * Gets the Target --&gt; live menu item. 
    *
    * @return  Target --&gt; live menu item
    */
   JMenuItem getTargetLive()
   {
      return targetLive;
   }
   
   /** 
    * Gets the Access Control menu. 
    *
    * @return  Access Control menu
    */
   JMenu getAccessControl()
   {
      return acl;
   }
   
   /** 
    * Enables or disables Target --&gt; refresh menu item. 
    *
    * @param   enabled
    *          <code>true</code> enables it
    */
   void setTargetRefresh(boolean enabled)
   {
      targetRefresh.setEnabled(enabled);
   }
   
   /** 
    * Enables or disables Console menu. 
    *
    * @param   enabled
    *          <code>true</code> enables it
    */
   void setConsole(boolean enabled)
   {
      cns.setEnabled(enabled);
   }
   
   /** 
    * Enables or disables Help --&gt; clear messages menu item. 
    *
    * @param   enabled
    *          <code>true</code> enables it
    */
   void setClearMessages(boolean enabled)
   {
      hlpClr.setEnabled(enabled);
   }
   
   /** 
    * Saves the state of the menu and disables it. 
    */
   void saveAndDisableMenu()
   {
      MenuElement[] items = mainMenu.getSubElements();
      int n = items.length;
      state = new boolean[n];
      
      for (int i = 0; i < n; i++)
      {
         state[i] = items[i].getComponent().isEnabled();
         items[i].getComponent().setEnabled(false);
      }
   }
   
   /** 
    * Restores the state of the menu from the saved info. 
    */
   void restoreMenu()
   {
      MenuElement[] items = mainMenu.getSubElements();
      
      for (int i = 0; i < state.length; i++)
      {
         items[i].getComponent().setEnabled(state[i]);
      }
   }
}