/*
** Module   : AdminProcessAccounts.java
** Abstract : Process Accounts management client
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090507  @42117  Created initial version.
** 002 CA  20090507  @42120  Changed texts for Add/Edit/Switch buttons to be
**                           in a standard way.
** 003 CA  20090512  @42145  Display the DN Common Name property as the 
**                           certificate description.
** 004 CA  20090519  @42330  When switching to Certificates screen, set the 
**                           title of the screen too.
** 005 NVS 20090525  @42448  Added temporary code to support custom account
**                           extensions. To be reimplemented.
** 006 CA  20090603  @42553  Added code for custom account extension plugin.
** 007 CA  20090604  @42563  Moved plugin loading to AdminTarget. After save is 
**                           pressed in add dialog, all widgets will be reset.
** 008 CA  20090005  @42608  Added 'account is disabled' check-box.
** 009 CA  20090609  @42615  Moved the 'account is disabled' check under the 
**                           account name field.
** 010 CA  20090611  @42655  Fixed a NPE when no extension plugin is set.
** 011 CA  20090612  @42664  Added Disable Process button. When checked, the
**                           color for 'Disabled Account' is set to red.
** 012 CA  20090720  @43245  Added code for the Switch to ACLs button.
** 013 NVS 20090723  @43353  Tables don't allow moving columns.
** 014 NVS 20090729  @43431  Implemented processes printing.
** 015 NVS 20090803  @43484  Using a better approach to printing with extended
**                           detail choice.
** 016 NVS 20090804  @43514  Fixed some texts in dialogs.
** 017 NVS 20090804  @43526  Plugin initializes as soon as this instance.
**                           Extended view printing is implemented.
** 018 NVS 20090805  @43533  Added busy/idle feedback to printing methods.
** 019 NVS 20090807  @43551  Refactored in support of custom menu extensions.
** 020 NVS 20091007  @44115  Changed to comply with the new ClientAccountExt.
**                           Changes are formal.
** 021 SIY 20091116   @44335 Refactoring of things related to TaggedName. 
**                           Minor cleanups.
** 022 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.awt.*;
import java.awt.event.*;
import java.io.Serializable;
import java.text.MessageFormat;
import java.util.*;
import java.util.List;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.TableModel;

import com.goldencode.p2j.admin.*;

/**
 * This is a Process Accounts management class for the admin session. 
 * 
 * @author NVS
 */
public class AdminProcessAccounts
extends AdminScreen
implements IAccountScreen, PrintRequest
{
   /** id of the account-name widget in account-def dialog */
   public static final String WIDGET_NAME = "NAME";
   
   /** id of the description widget in account-def dialog */
   public static final String WIDGET_DESCRIPTION = "DESCRIPTION";
   
   /** id of the certificate combo-box in account-def dialog */
   public static final String WIDGET_CERTIFICATE = "CERTIFICATE";
   
   /** id of the server check-box in account-def dialog */
   public static final String WIDGET_SERVER = "SERVER";
   
   /** id of the master check-box in account-def dialog */
   public static final String WIDGET_MASTER = "MASTER";
   
   /** id of the disabled account check-box in account-def dialog */
   public static final String WIDGET_ACC_DISABLED = "ACC_DISABLED";

   /** Screen switcher - user accounts screen ID */
   public static final String SCREEN_ACC_PROCS = "adm_acc_procs";
   
   /** Screen title */
   private static final String TITLE_ACC_PROCS = "Process Accounts Management";

   // west panel widgets, all named wp*

   /** west panel - works with the list of processes */
   private JPanel west = null;
   
   /** scroller for list of processes */
   private JScrollPane wpScroller = null;
   
   /** table to display all processes */
   private JTable wpProcs = null;

   /** related table model */
   private TaggedDataTableModel wpPtm = null;

   /** related table selection model */
   private ListSelectionModel wpPsm = null;

   /** sub-panel that holds all controls for list of processes */
   private JPanel wpCtlPanel = null;
   
   /** Populate button */
   private JButton wpbR = null;

   /** Add Process button */
   private JButton wpbA = null;
   
   /** Edit Process button */
   private JButton wpbE = null;

   /** Delete Selected button */
   private JButton wpbD = null;
      
   /** Switch to ACLs button */
   private JButton wpbS = null;
   
   /** Clone Process button */
   private JButton wpbC = null;
   
   /** Disable process button */
   private JButton wpbP = null;

   // east panel widgets, all named ep*
   
   /** east panel - works with the list of all certificates */
   private JPanel east = null;

   /** scroller for list of all certificates */
   private JScrollPane epScroller = null;
   
   /** table to display certificates */
   private JTable epCerts = null;

   /** related table model */
   private TaggedDataTableModel epCtm = null;

   /** related table selection model */
   private ListSelectionModel epCsm = null;
   
   /** sub-panel that holds all controls for list of certificates */
   private JPanel epCtlPanel = null;

   /** refresh button */
   private JButton epbR = null;

   /** assign button */
   private JButton epbA = null;
   
   /** switch to certificates button */
   private JButton epbS = null;
   
   // dialog for clone process account specific widgets, all named wp*

   /** Name field label */
   private JLabel wplN = null;
   
   /** Name field */
   private JTextField wpfN = null;
   
   /** With ACLs label */
   private JLabel wplW = null;

   /** With ACLs check box */
   private JCheckBox wpcW = null;
   
   /** Clone button  in modal dialog*/
   private JButton wpbG = null;
   
   /** modal dialog panel for Clone As function */
   private JPanel wpClonePanel = null;

   // dialog for process account specific widgets, all named wpd*
   
   /** modal dialog panel for Process Definition function */
   private JPanel wpdpProcessPanel = null;

   /** process account name label */
   private JLabel wpdlN = null;
   
   /** process account name field */
   private JTextField wpdfN = null;

   /** process account description label */
   private JLabel wpdlD = null;
   
   /** process account description field */
   private JTextField wpdfD = null;
   
   /** certificate alias label */
   private JLabel wpdlC = null;
   
   /** certificate combo-box */
   private JComboBox wpdcbC = null;

   /** certificate combo-box model */
   private TaggedDataComboModel wpdcbCm = null;
   
   /** process account 'server' flag label */
   private JLabel wpdlS = null;
   
   /** process account 'server' flag check-box */
   private JCheckBox wpdchkS = null;
   
   /** process account 'master' flag label */
   private JLabel wpdlM = null;
   
   /** process account 'master' flag check-box */
   private JCheckBox wpdchkM = null;
   
   /** process account disabled label */
   private JLabel wpdlDs = null;
   
   /** process account disabled check-box */
   private JCheckBox wpdchkDs = null;

   /** Cancel button */
   private JButton wpdbC = null;
   
   /** Save button */
   private JButton wpdbS = null;
   

   /** state machine's array of components */
   /* west,                    */
   /*    wpScroller,           */
   /*       wpProcs,           */
   /*    wpCtlPanel,           */
   /*       wpbR,              */
   /*       wpbA,              */
   /*       wpbD,              */
   /*       wpbC,              */
   /*       wpbE,              */
   /*       wpbS,              */
   /*  east,                   */
   /*     epScroller,          */
   /*        epCerts,          */
   /*     epCtlPanel,          */
   /*        epbR,             */
   /*        epbA,             */
   /*        epbS              */
   
   /** state definition - initial empty */
   private final boolean[] STATE_INITIAL_EMPTY =
   {
   /* west,                    */       true,
   /*    wpScroller,           */       true,
   /*       wpProcs,           */       true,
   /*    wpCtlPanel,           */       true,
   /*       wpbR,              */       true,
   /*       wpbA,              */       true,
   /*       wpbD,              */ false,
   /*       wpbC,              */ false,
   /*       wpbE,              */ false,
   /*       wpbS,              */ false,
   /*       wpbP,              */ false,
   /*  east,                   */       true,
   /*     epScroller,          */       true,
   /*        epCerts,          */       true,
   /*     epCtlPanel,          */       true,
   /*        epbR,             */       true,
   /*        epbA,             */ false,
   /*        epbS              */ false
   };

   /** state definition - single selection in process accounts list */
   private final boolean[] STATE_PROCS_SINGLE =
   {
   /* west,                    */       true,
   /*    wpScroller,           */       true,
   /*       wpProcs,           */       true,
   /*    wpCtlPanel,           */       true,
   /*       wpbR,              */       true,
   /*       wpbA,              */       true,
   /*       wpbD,              */       true,
   /*       wpbC,              */       true,
   /*       wpbE,              */       true,
   /*       wpbS,              */       true,
   /*       wpbP,              */       true,
   /*  east,                   */ false,
   /*     epScroller,          */ false,
   /*        epCerts,          */ false,
   /*     epCtlPanel,          */ false,
   /*        epbR,             */ false,
   /*        epbA,             */ false,
   /*        epbS              */ false
   };

   /** state definition - single selection in certificate list */
   private final boolean[] STATE_CERTS_SINGLE =
   {
   /* west,                    */ false,
   /*    wpScroller,           */ false,
   /*       wpProcs,           */ false,
   /*    wpCtlPanel,           */ false,
   /*       wpbR,              */ false,
   /*       wpbA,              */ false,
   /*       wpbD,              */ false,
   /*       wpbC,              */ false,
   /*       wpbE,              */ false,
   /*       wpbS,              */ false,
   /*       wpbP,              */ false,
   /*  east,                   */       true,
   /*     epScroller,          */       true,
   /*        epCerts,          */       true,
   /*     epCtlPanel,          */       true,
   /*        epbR,             */       true,
   /*        epbA,             */ false,
   /*        epbS              */       true
   };

   /** state definition - multiple selection in process accounts list */
   private final boolean[] STATE_PROCS_MULTI =
   {
   /* west,                    */       true,
   /*    wpScroller,           */       true,
   /*       wpProcs,           */       true,
   /*    wpCtlPanel,           */       true,
   /*       wpbR,              */       true,
   /*       wpbA,              */       true,
   /*       wpbD,              */       true,
   /*       wpbC,              */ false,
   /*       wpbE,              */ false,
   /*       wpbS,              */ false,
   /*       wpbP,              */ false,
   /*  east,                   */       true,
   /*     epScroller,          */       true,
   /*        epCerts,          */       true,
   /*     epCtlPanel,          */       true,
   /*        epbR,             */       true,
   /*        epbA,             */ false,
   /*        epbS              */ false
   };

   /** mask definition - west */
   private final boolean[] MASK_WEST =
   {
   /* west,                    */ true,
   /*    wpScroller,           */ true,
   /*       wpProcs,           */ true,
   /*    wpCtlPanel,           */ true,
   /*       wpbR,              */ true,
   /*       wpbA,              */ true,
   /*       wpbD,              */ true,
   /*       wpbC,              */ true,
   /*       wpbE,              */ true,
   /*       wpbS,              */ true,
   /*       wpbP,              */ true,
   /*  east,                   */ false,
   /*     epScroller,          */ false,
   /*        epCerts,          */ false,
   /*     epCtlPanel,          */ false,
   /*        epbR,             */ false,
   /*        epbA,             */ false,
   /*        epbS              */ false
   };
   
   /** mask definition - east */
   private final boolean[] MASK_EAST =
   {
   /* west,                    */ false,
   /*    wpScroller,           */ false,
   /*       wpProcs,           */ false,
   /*    wpCtlPanel,           */ false,
   /*       wpbR,              */ false,
   /*       wpbA,              */ false,
   /*       wpbD,              */ false,
   /*       wpbC,              */ false,
   /*       wpbE,              */ false,
   /*       wpbS,              */ false,
   /*       wpbP,              */ false,
   /*  east,                   */ true,
   /*     epScroller,          */ true,
   /*        epCerts,          */ true,
   /*     epCtlPanel,          */ true,
   /*        epbR,             */ true,
   /*        epbA,             */ true,
   /*        epbS              */ true
   };

   /** indicates single process selection state */
   private boolean singleProcess = false;

   /** the selected process definition, for single process selection */
   private ProcessDef process = null;

   /** the customer-specific extension plugin */
   private ClientAccountExtension extPlugin = null;

   /** container for user-extension fields */
   private Container extContainer = null;
   
   /** map of default user dialog widgets, kept by their name */
   private HashMap<String, Component> processDefWidgets = null; 

   /** print: sortable table model */
   private TableModel backTm = null;
   
   /** print: sortable table */
   private JTable backTable = null;
   
   /** print: extended table model */
   private TableModel frontTm = null;
   
   /** print: extended table */
   private JTable frontTable = null;
   
   /** print header */
   private MessageFormat header = null;
   
   /** print footer */
   private MessageFormat footer = null;
   
   /**
    * Constructor.
    * 
    * @param    client
    *           instance of AdminClient
    */
   AdminProcessAccounts(AdminClient client)
   {
      // store useful references
      this.client = client;
      adm = client.getExports();
      
      extPlugin = client.getAccountExtPlugin();
      if (extPlugin != null)
      {
         extPlugin.initialize(this);
      }

      // create the main screen and add it to the swapper
      mainPanel = new JPanel();
      client.add(mainPanel, SCREEN_ACC_PROCS, this);

         GroupLayout mainLayout = new GroupLayout(mainPanel);
      mainPanel.setLayout(mainLayout);
      mainLayout.setAutoCreateGaps(true);
      mainLayout.setAutoCreateContainerGaps(true);
      
      // top panel with group layout manages
      // west and east sub-panels
      //  +-------screen-----+
      //  |+------+  +------+|
      //  ||      |  |      ||
      //  ||      |  |      ||
      //  || west |  | east ||
      //  ||      |  |      ||
      //  |+------+  +------+|
      //  +------------------+
      
      west = new JPanel();
      east = new JPanel();

      // Create a sequential group for the horizontal axis:
      {
         GroupLayout.SequentialGroup hGroup = 
            mainLayout.createSequentialGroup();
         
         // three panels 
         hGroup.
            addComponent(west).
            addComponent(east)
         ;
         
         mainLayout.setHorizontalGroup(hGroup);
      }
      
      // Create a parallel group for the vertical axis
      {
         GroupLayout.ParallelGroup vGroup = 
            mainLayout.createParallelGroup(GroupLayout.Alignment.BASELINE);
            
         vGroup.
            addComponent(west).
            addComponent(east)
         ;
            
         mainLayout.setVerticalGroup(vGroup);
      }

      // stuffing west panel with components:
      // west panel is controlled with BorderLayout
      west.setLayout(new BorderLayout());
      
      //  +---------west-----------+
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    wpScroller      | |
      //  | |      CENTER        | |
      //  | |                    | |
      //  | |                    | |
      //  | +--------------------+ |
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    wpCtlPanel      | |
      //  | |      SOUTH         | |
      //  | +--------------------+ |
      //  +------------------------+
      
      wpPtm = new TaggedDataTableModel("Process account", "Description");
      wpProcs = new JTable(wpPtm);
      wpProcs.getTableHeader().setReorderingAllowed(false);
      wpPsm = wpProcs.getSelectionModel();
      
      wpScroller = new JScrollPane();
      wpScroller.getViewport().add(wpProcs);
      wpScroller.setBorder(new TitledBorder
       (new EtchedBorder(EtchedBorder.RAISED), " Defined Process Accounts "));
      
      wpScroller.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      west.add(wpScroller, BorderLayout.CENTER);

      // wpCtlPanel panel is controlled with GroupLayout
      //  +------wpCtlPanel--------+
      //  |                        |
      //  |                        |
      //  |     buttons here       |   
      //  |                        |
      //  |                        |
      //  +------------------------+
      
      // creating control widgets
      wpbR = new JButton("Refresh");
      wpbA = new JButton("Add Process");
      wpbD = new JButton("Delete Selected");
      wpbC = new JButton("Clone Process");
      wpbE = new JButton("Edit Process");
      wpbS = new JButton("Switch to ACLs");
      wpbP = new JButton("Disable Process");

      // create layout of the wpCtlPanel panel with all controls
      wpCtlPanel = AdminUtils.groupComponents(new Component[][] 
         { 
            { wpbR, wpbA, wpbD },
            { wpbC, wpbE, wpbP },
            { new JLabel(), new JLabel(), wpbS }
         },
         true
      );
      west.add(wpCtlPanel, BorderLayout.SOUTH);
    

      // stuffing east panel with components:
      // east panel is controlled with BorderLayout
      east.setLayout(new BorderLayout());
      
      //  +---------east-----------+
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    epScroller      | |
      //  | |      CENTER        | |
      //  | |                    | |
      //  | |                    | |
      //  | +--------------------+ |
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    epCtlPanel      | |
      //  | |      SOUTH         | |
      //  | +--------------------+ |
      //  +------------------------+
      
      epCtm = new TaggedDataTableModel("Alias", "Description");
      epCerts = new JTable(epCtm);
      epCerts.getTableHeader().setReorderingAllowed(false);
      epCerts.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      epCsm = epCerts.getSelectionModel();
      
      epScroller = new JScrollPane();
      epScroller.getViewport().add(epCerts);
      epScroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Defined Certificates "));
      
      epScroller.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      east.add(epScroller, BorderLayout.CENTER);

      epCtlPanel = new JPanel();
      east.add(epCtlPanel, BorderLayout.SOUTH);

      // epCtlPanel panel is controlled with GroupLayout
      //  +------epCtlPanel--------+
      //  |                        |
      //  |                        |
      //  |     buttons here       |   
      //  |                        |
      //  |                        |
      //  +------------------------+
      
      // creating control widgets
      epbR = new JButton("Refresh");
      epbA = new JButton("Assign");
      epbS = new JButton("Switch to Certificates");

      // create layout of the epCtlPanel panel with all controls
      GroupLayout epCtlLayout = new GroupLayout(epCtlPanel);
      epCtlPanel.setLayout(epCtlLayout);
    
      epCtlLayout.setAutoCreateGaps(true);
      epCtlLayout.setAutoCreateContainerGaps(true);
    
      // Create a sequential group for the horizontal axis:
      {
         GroupLayout.SequentialGroup hGroup = 
            epCtlLayout.createSequentialGroup();
         
         // Refresh button
         hGroup.addComponent(epbR);
         // Assign button
         hGroup.addComponent(epbA);
         // Switch button
         hGroup.addComponent(epbS);

         epCtlLayout.setHorizontalGroup(hGroup);
      }
      
      // Create a sequential group for the vertical axis
      {
         GroupLayout.SequentialGroup vGroup = epCtlLayout.
                                              createSequentialGroup();
         vGroup.
            // parallel group of three buttons
            addGroup(epCtlLayout.createParallelGroup(
                                              GroupLayout.Alignment.BASELINE).
               // Refresh button
               addComponent(epbR).
               // Assign button
               addComponent(epbA).
               // Switch button
               addComponent(epbS)
               )
            ;
            
         epCtlLayout.setVerticalGroup(vGroup);
      }
      
      // create "Process Def" modal dialog panel with Group layout
      wpdlN = new JLabel(" Account Name ");
      wpdfN = new JTextField(12);
      wpdlD  = new JLabel(" Description ");
      wpdfD  = new JTextField(32);
      wpdlC  = new JLabel(" Certificate ");
      
      wpdcbCm = new TaggedDataComboModel();
      wpdcbC = new JComboBox(wpdcbCm);
      wpdcbC.setRenderer(new TaggedDataComboRenderer());
      wpdcbC.setMaximumRowCount(10);

      wpdlS   = new JLabel(" Server ");
      wpdchkS = new JCheckBox();
      wpdlM   = new JLabel(" Master ");
      wpdchkM = new JCheckBox();

      wpdlDs   = new JLabel(" Account is Disabled ");
      wpdchkDs = new JCheckBox();
      
      wpdbC = new JButton("Cancel");
      wpdbS = new JButton("Save");
      
      layoutDefaultProcessDefWidgets(null);

      // create "Clone as" modal dialog panel with Group layout
      wpClonePanel = new JPanel();
      
      wplN = new JLabel("New Name");
      wpfN = new JTextField(12);
      wplW = new JLabel("With ACLs");
      wpcW = new JCheckBox();
      wpbG = new JButton("Clone Process");

      wpClonePanel.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), 
                           " Clone Selected Process As "));
      GroupLayout wpCloneLayout = new GroupLayout(wpClonePanel);
      wpClonePanel.setLayout(wpCloneLayout);
    
      wpCloneLayout.setAutoCreateGaps(true);
      wpCloneLayout.setAutoCreateContainerGaps(true);
    
      // Create a sequential group for the horizontal axis.
      {
         GroupLayout.SequentialGroup hGroup = 
            wpCloneLayout.createSequentialGroup();
       
         hGroup.
            addComponent(wplN).
            addComponent(wpfN).
            addGroup(wpCloneLayout.createParallelGroup().
               addGroup(wpCloneLayout.createSequentialGroup().
                  addComponent(wplW).
                  addComponent(wpcW)
                  ).
               addComponent(wpbG)
               );
         wpCloneLayout.setHorizontalGroup(hGroup);
      }
      
      // Create a sequential group for the vertical axis.
      {
         GroupLayout.SequentialGroup vGroup = wpCloneLayout.
                                              createSequentialGroup();
       
         vGroup.addGroup(wpCloneLayout.createParallelGroup(
                                       GroupLayout.Alignment.BASELINE).
            addComponent(wplN).
            addComponent(wpfN).
            addComponent(wplW).
            addComponent(wpcW)
            ).
            addComponent(wpbG);
         wpCloneLayout.setVerticalGroup(vGroup);
      }

      // create event handlers: west panel

      // Process accounts table selection events
      wpPsm.addListSelectionListener(new ListSelectionListener() 
      {
         public void valueChanged(ListSelectionEvent evt) 
         {
            handleProcessAccountsSelection(evt);
         }
      });

      // Populate button
      wpbR.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handlePopulateProcessAccounts();
         }
      });

      // Add button
      wpbA.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleAddProcessAccount();
         }
      });

      // Delete Selected button
      wpbD.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleDeleteSelectedProcessAcounts();
         }
      });

      // Edit process button
      wpbE.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleEditProcessAccount();
         }
      });

      // Switch to ACLs button
      wpbS.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleSwitchToACLs();
         }
      });

      // Clone process button
      wpbC.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleClone();
         }
      });
      
      // Disable process  button
      wpbP.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleDisableProcess();
         }
      });

      // create event handlers: east panel

      // Certificates table selection events
      epCsm.addListSelectionListener(new ListSelectionListener() 
      {
         public void valueChanged(ListSelectionEvent evt) 
         {
            handleCertificatesSelection(evt);
         }
      });
    
      // Populate button
      epbR.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handlePopulateCertificates();
         }
      });

      // Assign button
      epbA.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleAssignCertificate();
         }
      });

      // Switch to certificates button
      epbS.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleSwitchToCertificates();
         }
      });

      //  create event handlers: west panel process definition dialog

      // Server check-box
      wpdchkS.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleProcessServerCheckbox();
         }
      });

      // Cancel button
      wpdbC.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // Save button
      wpdbS.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleSave();
         }
      });

      // CloneAs button
      wpbG.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCloneAs();
         }
      });

      // Disabled Account check-box
      wpdchkDs.addActionListener(new ActionListener() 
      {
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleDisabledCheckbox();
         }
      });

      // create the array of components for the state machine
      state = new JComponent[]
      {
         west,
            wpScroller,
               wpProcs,
            wpCtlPanel,
               wpbR, wpbA, wpbD,
               wpbC, wpbE, wpbP,
                           wpbS,
          east,
             epScroller,
                epCerts,
             epCtlPanel,
                epbR, epbA, epbS
      };

      // build the map with the default widgets
      processDefWidgets = new HashMap<String, Component>();
      processDefWidgets.put(WIDGET_NAME, wpdfN);
      processDefWidgets.put(WIDGET_DESCRIPTION, wpdfD);
      processDefWidgets.put(WIDGET_CERTIFICATE, wpdcbC);
      processDefWidgets.put(WIDGET_SERVER, wpdchkS);
      processDefWidgets.put(WIDGET_MASTER, wpdchkM);
      processDefWidgets.put(WIDGET_ACC_DISABLED, wpdchkDs);

      // populate the process accounts list upfront
      handlePopulateProcessAccounts();

      // populate the certificates list upfront
      handlePopulateCertificates();
   }

   /**
    * Performs printing of the current screen data.
    * This screen prints the current view of the processes table.
    */
   public void print()
   {
      // check the processes table selection
      int[] sel = wpProcs.getSelectedRows();
      int nsel = sel.length;
      int nall = wpPtm.getRowCount();
      final boolean selectOnly;
      
      // no selection means the whole table, otherwise there is a choice
      if (nsel == 0 || nsel == nall)
      {
         // printing the whole table confirmation
         String quest = String.format(
                        "%d process accounts being printed. Continue?",
                        nall);
         if (!AdminUtils.confirmDialog(client, "Printing Processes", quest))
         {
            // declined
            return;
         }
         
         // print all
         selectOnly = false;
      }
      else
      {
         // offering a choice 
         String quest = "Please refine your request";
         String[] choices = new String[2];
         choices[0] = nsel == 1 ? "the selected process account" :
                      String.format("%d selected process accounts", nsel);
         choices[1] = nall == 2 ? "both process accounts" :
                      String.format("all %d process accounts", nall);
         
         int n = AdminUtils.choiceDialog(client, "Printing Processes", quest,
                                         choices, "Print");
         if (n < 0)
         {
            // declined
            return;
         }
         
         // print mode choice
         selectOnly = n == 0;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // build preview main table
      List<AdminUtils.AccountInfo> backUl = new LinkedList
                                                <AdminUtils.AccountInfo>();
      String[] extHeaders = null;
      
      if (extPlugin != null)
      {
         extHeaders = extPlugin.getPreviewHeaders();
      }
      else
      {
         extHeaders = new String[] {};
      }
                                          
      for (int i = 0; i < nall; i++)
      {
         // translate model to view
         int iv = wpProcs.convertRowIndexToView(i);
         
         // honor selection if asked
         if (selectOnly && !wpPsm.isSelectedIndex(iv))
         {
            continue;
         }
         
         // stuff the back list with the process info
         AdminUtils.AccountInfo backUi = new AdminUtils.AccountInfo();
         backUi.aid    = (String)wpPtm.getValueAt(i, 0);
         backUi.descr  = (String)wpPtm.getValueAt(i, 1);
         backUi.disabled = "";
         backUi.alias = "";
         backUi.attr1 = "";
         backUi.attr2 = "";
         
         // load process from server
         ProcessDef proc = adm.getProcess(backUi.aid);
         if (proc == null)
         {
            // process may have been deleted
            backUi.descr = "*** possibly deleted ***";
         }
         else
         {
            // gotten process info
            backUi.disabled = proc.enabled ? "" : "yes"; 
            backUi.alias = proc.alias; 
            backUi.attr1 = proc.server ? "server" : "app";  
            backUi.attr2 = proc.server && proc.master ? "master" : "";
         }
         
         if (extPlugin != null)
         {
            backUi.ext = extPlugin.getPreviewRow(backUi.aid);
         }
         else
         {
            backUi.ext = new String[] {};
         }
         
         // all details are collected
         backUl.add(backUi);
      }
      
      header = new MessageFormat(
         selectOnly ? "Selected Process Accounts" : "All Process Accounts");
      footer = new MessageFormat("page {0, number}");
               
      backTm = new AdminUtils.PrintAccountModel(extHeaders,
                              backUl.toArray(new AdminUtils.AccountInfo[] {}));
      backTable = new JTable(backTm);
      
      // idle
      AdminUtils.idle(client);
      
      client.printOn(backTable, backTm, header, footer);
   }
   
   /**
    * Queries the extended printing capability. If the extended capability is
    * not provided, the printExtended() method should not be called.
    * 
    * @return   <code>true</code> to tell the caller the extended printing is
    *           implemented
    */
   public boolean canPrintExtended()
   {
      return true;
   }

   /**
    * Creates extended print preview table off the current screen data.
    * This screen prints the current view of the users table.
    */
   public void printExtended()
   {
      // busy
      AdminUtils.busy(client);
      
      // standard part
      String[] stdHeaders = new String[] {"Account",
                                          "Disabled",
                                          "Description",
                                          "Status"
                                         };
         
      int stdRows = 1;          // all std info takes one row
      int stdCols = 4;          // all std info takes four columns
      
      // extended part
      String[] extHeaders = null;
      int extRows = 0;
      int extCols = 0;
      
      if (extPlugin != null)
      {
         extHeaders = extPlugin.getExtendedHeaders();
         extCols = extHeaders != null ? extHeaders.length : 0;
         extRows = extPlugin.getExtendedRows();
      }
      else
      {
         extHeaders = new String[] {};
      }
      
      // extended view table model and table
      frontTm = new AdminUtils.PrintExtendedAccountModel(
                               stdHeaders, extHeaders, new String[][] {{}});
      frontTable = new JTable(frontTm);
      
      // create the extended view
      int nRec = backTable.getRowCount();
      int step = extRows > stdRows ? extRows : stdRows;
      int nExt = nRec * step;
      int nCol = stdCols + extCols;
      
      String[][] extView = new String[nExt][nCol];
      
      // pre-fill the extended view array
      for (int i = 0; i < nExt; i++)
      {
         for (int j = 0; j < nCol; j++)
         {
            extView[i][j] = "";
         }
      }
      ((AdminUtils.PrintExtendedAccountModel)frontTm).reload(extView);
      
      // fill the extended view with real data
      for (int n = 0; n < nRec; n++)
      {
         // convert the preview table index into preview model
         int nd = backTable.convertRowIndexToModel(n);
         
         // starting row
         int nr = n * step;
         
         // get the standard info
         String subj = (String)backTm.getValueAt(nd, 0);
         frontTm.setValueAt(subj, nr, 0);                         // pid
         frontTm.setValueAt(backTm.getValueAt(nd, 1), nr, 2);     // descr 
         frontTm.setValueAt(backTm.getValueAt(nd, 2), nr, 1);     // disabled
         frontTm.setValueAt(backTm.getValueAt(nd, 4), nr, 3);     // status
         
         // get the extended info
         if (extCols > 0)
         {
            extPlugin.fillExtendedRow(subj, frontTm, nr, stdCols);
         }
      }
      
      // idle
      AdminUtils.idle(client);
      
      client.printExtendedOn(frontTable, header, footer);
   }
   
   /**
    * Get the close button associated with the account dialog.
    * 
    * @return   the close button
    */
   public JButton getCloseButton()
   {
      return wpdbC;
   }

   /**
    * Get the save button for the account dialog.
    * 
    * @return   the save button
    */
    public JButton getSaveButton()
   {
      return wpdbS;
   }

   /**
    * Get the panel where default account widgets are kept.
    * 
    * @return   the panel which contains default widgets.
    */
   public JPanel getDefaultWidgetPanel()
   {
      return wpdpProcessPanel;
   }

   /**
    * Get the widget with the specified name.
    * 
    * @param    name
    *           The widget's name
    * 
    * @return   the widget with the specified name or <code>null</code> if the
    *           widget does not exist.
    */
   public Component getWidgetByName(String name)
   {
      return processDefWidgets.get(name);
   }
   
   /**
    * Layout the extension fields for this process.
    * 
    * @param    add
    *           <code>true</code> if this is for adding a new process.
    * 
    * @param    writable
    *           <code>true</code> if this is for enabled inputs.
    *           <code>false</code> for read only panel
    * 
    * @return   the container which holds process def fields.
    */
   private JPanel checkExtFieldsLayout(boolean add, boolean writable)
   {
      if (extPlugin == null)
         return wpdpProcessPanel;
      
      JPanel c = wpdpProcessPanel;
      
      String proc = add ? null : process.subjectId;
      
      switch (extPlugin.getLayoutMode())
      {
         case ClientAccountExtension.LAYOUT_ALL_WIDGETS:
         {
            // relayout the default widgets, just in case the plugin doesn't
            // need to change their order (it just moves the panel around)
            layoutDefaultProcessDefWidgets(null);

            // in this case, the ext plugin must be given a chance to re-layout
            // the default widgets too
            c = extPlugin.layoutAllWidgets(proc, writable);
            break;
         }
         case ClientAccountExtension.LAYOUT_EXT_WIDGETS:
         {
            extContainer = extPlugin.getExtContainer();
            
            // this container must be added to the dialog ...
            layoutDefaultProcessDefWidgets(extContainer);
            
            // allow the plugin to layout its data
            extPlugin.layoutExtWidgets(extContainer, proc, writable);
            break;
         }
      }
      
      return c;
   }

   /**
    * Layout the default widgets in the process-def dialog.
    * 
    * @param    extContainer
    */
   private void layoutDefaultProcessDefWidgets(Container extContainer)
   {
      // Cancel/Save buttons will be added only if the plugin doesn't need
      // to layout all widgets
      boolean addButtons = extPlugin == null || 
       extPlugin.getLayoutMode() == ClientAccountExtension.LAYOUT_EXT_WIDGETS;
      
      wpdpProcessPanel = new JPanel();

      wpdpProcessPanel.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Process Definition "));
      GroupLayout wpdGroupLayout = new GroupLayout(wpdpProcessPanel);
      wpdpProcessPanel.setLayout(wpdGroupLayout);

      // Turn on automatically adding gaps
      wpdGroupLayout.setAutoCreateGaps(true);
      wpdGroupLayout.setAutoCreateContainerGaps(true);
    
      // Create a sequential group for the horizontal axis.
      {
         GroupLayout.SequentialGroup hGroup = wpdGroupLayout.
                                              createSequentialGroup();
       
         GroupLayout.ParallelGroup pGroup1 = 
            wpdGroupLayout.createParallelGroup();

         hGroup.addGroup(pGroup1.
            addComponent(wpdlN).
            addComponent(wpdlDs).
            addComponent(wpdlD).
            addComponent(wpdlC).
            addComponent(wpdlS).
            addComponent(wpdlM)
            );
         
         if (addButtons)
         {
            pGroup1.addComponent(wpdbC);
         }

         GroupLayout.ParallelGroup pGroup2 = 
            wpdGroupLayout.createParallelGroup();

         hGroup.addGroup(pGroup2.
            addComponent(wpdfN).
            addComponent(wpdchkDs).
            addComponent(wpdfD).
            addComponent(wpdcbC).
            addComponent(wpdchkS).
            addComponent(wpdchkM)
            );
         
         if (addButtons)
         {
            pGroup2.addComponent(wpdbS);
         }
         
         wpdGroupLayout.setHorizontalGroup(hGroup);
      }
      
      // Create a sequential group for the vertical axis.
      {
         GroupLayout.SequentialGroup vGroup = wpdGroupLayout.
                                              createSequentialGroup();
       
         vGroup.addGroup(wpdGroupLayout.createParallelGroup(GroupLayout.
                                                          Alignment.BASELINE).
            addComponent(wpdlN).
            addComponent(wpdfN)
            );
         vGroup.addGroup(wpdGroupLayout.createParallelGroup(GroupLayout.
               Alignment.BASELINE).
               addComponent(wpdlDs).
               addComponent(wpdchkDs)
         );
         vGroup.addGroup(wpdGroupLayout.createParallelGroup(GroupLayout.
                                                          Alignment.BASELINE).
            addComponent(wpdlD).
            addComponent(wpdfD)
            );
         vGroup.addGroup(wpdGroupLayout.createParallelGroup(GroupLayout.
                                                          Alignment.BASELINE).
            addComponent(wpdlC).
            addComponent(wpdcbC)
            );
         vGroup.addGroup(wpdGroupLayout.createParallelGroup(GroupLayout.
               Alignment.BASELINE).
               addComponent(wpdlS).
               addComponent(wpdchkS)
         );
         vGroup.addGroup(wpdGroupLayout.createParallelGroup(GroupLayout.
               Alignment.BASELINE).
               addComponent(wpdlM).
               addComponent(wpdchkM)
         );
         
         if (addButtons)
         {
            vGroup.addGroup(wpdGroupLayout.createParallelGroup(GroupLayout.
                  Alignment.BASELINE).
                  addComponent(wpdbC).
                  addComponent(wpdbS)
            );
         }
         
         wpdGroupLayout.setVerticalGroup(vGroup);
      }
      
   }
   
   /** 
    * Presents the refreshed process accounts management screen.
    */
   void manageProcesses() 
   {
      // switch to this screen without changing state of components
      client.show(SCREEN_ACC_PROCS, TITLE_ACC_PROCS);
   }
   
   /**
    * Populate the list of process accounts with fresh data from the server.
    */
   private void populateProcessAccountsList() 
   {
      // clear selection if any
      wpPsm.clearSelection();
      
      // clear the old list
      wpPtm.clear();
      
      // get the list of processes from the server
      AdminUtils.busy(client);
      TaggedName[] list = adm.listProcesses();
      AdminUtils.idle(client);
      
      if (list != null)
      {
         if (list.length > 0)
         {
            wpPtm.loadAll(list);
         }
      }
      else
      {
         client.alarm();
      }

      // enable the process list explicitly
      wpProcs.setEnabled(true);
   }

   /**
    * Populate the list of certificates with fresh data from the server.
    */
   private void populateCertificatesList() 
   {
      // clear selection if any
      epCsm.clearSelection();
      
      // clear the old list
      epCtm.clear();
      
      // clear the process def certificates combo-box
      wpdcbCm.clear();

      // get the list of certificates from the server
      AdminUtils.busy(client);
      TaggedName[] list = adm.listPeerCerts(false);
      AdminUtils.idle(client);

      if (list != null)
      {
         if (list.length > 0)
         {
            // set certificate description to Common Name
            setCertificateDescription(list, CERT_DESCRIPTION_PROP);

            epCtm.loadAll(list);
            
            // add the certificates to the dialog combo too
            wpdcbCm.loadAll(list);
         }
      }
      else
      {
         client.alarm();
      }

      // enable the certificate list explicitly
      epCerts.setEnabled(true);
   }
      
   /**
    * Sets state of components to no selection in the Process Accounts list
    * <p>
    * No selection means that the state should be STATE_INITIAL_EMPTY.
    * Only the "Populate" and "Add process" buttons are enabled.
    */
   private void setProcessAcountsNoSelection()
   {
      // reset flags
      singleProcess = false;

      // set enabled state of the list accordingly
      stateMachine(MASK_WEST, STATE_INITIAL_EMPTY);
   }

   /**
    * Sets state of components to no selection in the Certificates list
    * <p>
    * No selection means that the state should be STATE_INITIAL_EMPTY.
    * Only the "Populate" and "Add process" buttons are enabled.
    */
   private void setCertificatesNoSelection()
   {
      // set enabled state of the list accordingly
      stateMachine(MASK_EAST, STATE_INITIAL_EMPTY);
   }

   /**
    * Populate the list of process accounts with fresh data from the server,
    * when the Refresh button is pressed.
    */
   private void handlePopulateProcessAccounts()
   {
      // populate the process list
      populateProcessAccountsList();

      // set the state accordingly
      setProcessAcountsNoSelection();
   }

   /**
    * Populate the list of certificates with fresh data from the server, when
    * the Refresh button is pressed.
    * <p>
    * If there is only one selected process account, it will auto-select the
    * certificate alias assigned to the process.
    */
   private void handlePopulateCertificates()
   {
      // populate the certificate list
      populateCertificatesList();

      // set the state accordingly
      setCertificatesNoSelection();
      
      // select the certificate associated with the selected process
      selectProcessAcountCertificate();
   }
   
   /** 
    * Changes state of the components in response to process list selections.
    * There are three distinctive states:
    * <ul>
    *   <li>no selection or empty list
    *   <li>single entry selected
    *   <li>multiple entries selected
    * </ul>
    * This method determines the selection state of the list and changes
    * the state of the components accordingly.
    * <p>
    * No selection means that the state should be STATE_INITIAL_EMPTY.
    * Only the "Populate" and "Add Process" buttons are enabled.
    * <p>
    * Single selection means that all buttons of the west panel are enabled.
    * Also, the east panel (certificate list) will auto-select the certificate
    * assigned for this process. The Assign button is disabled, only Refresh
    * and Switch to Certificates are enabled.
    * <p>
    * Multiple selection means that buttons, specific to the single selection,
    * are disabled. The enabled buttons on the west panel are Populate,
    * Add Process and Delete selected. In this case, the east-panel Assign 
    * button will always be disabled.
    *
    * @param    evt
    *           event caused this method call
    */
   private void handleProcessAccountsSelection(ListSelectionEvent evt)
   {
      // do not bother until the dust is fully settled
      if (evt.getValueIsAdjusting())
      {
         return;
      }

      // figure out the selection state
      if (wpPsm.isSelectionEmpty())
      {
         // no selection
         setProcessAcountsNoSelection();
      }
      else if (wpProcs.getSelectedRows().length == 1)
      {
         // single selection
         setProcessAcountsSingleSelection();
      }
      else
      {
         // multiple selection
         setProcessAcountsMultiSelection();
      }

      // auto-select process certificate, if in single-selection mode; else,
      // disable Assign button
      selectProcessAcountCertificate();      
   }
   
   /**
    * Changes state of the components in response to certificates list
    * selections. This table allows only single selections.
    * <p>
    * No selection means only the Refresh button is enabled.
    * <p>
    * Single selection means the Refresh and Switch to certificates buttons
    * are enabled. Also, if the process account list has a only one selected
    * process, the Assign button will be enabled only if the selected
    * certificate is not the one currently assign with the selected process.
    * 
    * @param    evt
    *           event caused this method call
    */
   private void handleCertificatesSelection(ListSelectionEvent evt)
   {      
      // do not bother until the dust is fully settled
      if (evt.getValueIsAdjusting())
      {
         return;
      }
      
      // figure out the selection state
      if (epCsm.isSelectionEmpty())
      {
         // no selection
         setCertificatesNoSelection();
      }
      else if (epCerts.getSelectedRows().length == 1)
      {
         // single selection
         setCertificatesSingleSelection();
      }      
   }
   
   /**
    * Sets state of components to single selection in the Process list 
    */
   private void setProcessAcountsSingleSelection()
   {
      // busy
      AdminUtils.busy(client);
      
      // set flags
      singleProcess = true;
      String pName = (String) wpProcs.getValueAt(wpProcs.getSelectedRow(), 0);

      // load and keep process definition from the server
      process = adm.getProcess(pName);
      if (process == null)
      {
         setProcessAcountsNoSelection();

         client.alarm();
      }
      else
      {
         // set enabled state of the components
         stateMachine(MASK_WEST, STATE_PROCS_SINGLE);
         
         boolean enabled = process.enabled;
         wpbP.setText(enabled ? " Disable Process " : " Enable Process ");
         wpbP.repaint();

         client.ok();
      }

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Sets state of components to single selection in the Certificate list 
    */
   private void setCertificatesSingleSelection()
   {
      // busy
      AdminUtils.busy(client);
      
      // set enabled state of the components
      stateMachine(MASK_EAST, STATE_CERTS_SINGLE);

      // verify if selected certificate is the one assigned for the currently
      // selected process, if any
      verifySelectedCertificate();
      
      // idle
      AdminUtils.idle(client);
   }

   /**
    * Check if the selected certificate is the one assigned to the selected 
    * process, if any.
    * <p>
    * If not, and there one selected process, enable the Assign certificate 
    * button.
    */
   private void verifySelectedCertificate()
   {
      // assign button may be enabled only if a single process is selected
      boolean enabled = singleProcess;
      
      if (enabled && singleProcess && epCerts.getSelectedRowCount() == 1)
      {
         String certname = 
            (String) epCerts.getValueAt(epCerts.getSelectedRow(), 0);
         
         if (process != null       && 
             process.alias != null && 
             process.alias.equals(certname))
         {
            // the certificate is the one for the current process, disable
            // Assign button
            enabled = false;
         }
      }
      
      epbA.setEnabled(enabled);
   }

   /**
    * If there is only one selected process account, auto select its 
    * certificate in the east-panel table. 
    */
   private void selectProcessAcountCertificate()
   {
      epCsm.clearSelection();

      if (singleProcess)
      {
         // get the selected certificate index
         int certidx = epCtm.getRowIndex(process.alias);
         if (certidx >= 0)
         {
            epCsm.setSelectionInterval(certidx, certidx);
         }
      }
   }

   /**
    * Sets state of components to multiple selection in the Process list 
    */
   private void setProcessAcountsMultiSelection()
   {
      // reset flags
      singleProcess = false;
      process = null;
      
      // set enabled state of the components
      stateMachine(STATE_PROCS_MULTI);
   }
   
   /**
    * Handle Add button press in the west panel.
    * <p>
    * This will reset all fields in the Add Process dialog and will display it.
    */
   private void handleAddProcessAccount()
   {
      // focus on process name
      wpdfN.requestFocusInWindow();

      // for now, the panel is writable
      JPanel panel = checkExtFieldsLayout(true, true);

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "Process Acount Definition", 
                                      panel);

      resetProcessDialog();

      wpdfN.setEnabled(true); // enable Name field
      wpdcbC.setEditable(true); // enable cert alias combo-box for editing
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }
   
   /**
    * Reset all fields from the dialog.
    */
   private void resetProcessDialog()
   {
      // reset dialog fields         
      wpdfN.setText("");
      wpdfD.setText("");
      wpdcbCm.setSelectedItem(null);
      wpdchkS.setSelected(false);
      wpdchkM.setSelected(false);
      wpdchkM.setEnabled(false);
      wpdchkDs.setSelected(false);
      
      wpdlDs.setForeground(Color.BLACK);
      
      if (extPlugin != null)
      {
         extPlugin.resetDialog();
      }
   }

   /**
    * Handle Edit button press in the west panel.
    * <p>
    * This will load the process definition for the currently selected row.
    * If the data can not be loaded, it will re
    */
   private void handleEditProcessAccount()
   {
      // reload process from server
      process = adm.getProcess(process.subjectId);
      if (process == null)
      {
         // process may have been deleted
         client.alarm();
         
         // refresh the data from the server
         handlePopulateProcessAccounts();
         return;
      }
      
      // initialize dialog fields
      wpdfN.setText(process.subjectId);
      wpdfD.setText(process.description);
      wpdchkS.setSelected(process.server);
      wpdcbCm.setSelectedItem(process.alias);
      wpdchkM.setSelected(process.server && process.master);
      wpdchkM.setEnabled(process.server);
      
      wpdchkDs.setSelected(!process.enabled);

      // focus on description
      wpdfD.requestFocusInWindow();
      
      // for now, the panel is writable
      JPanel panel = checkExtFieldsLayout(false, true);

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "Process Acount Definition", 
                                      panel);
      wpdlDs.setForeground(process.enabled ? Color.BLACK : Color.RED);
      wpdfN.setEnabled(false); // disable Name field
      wpdcbC.setEditable(true); // enable cert alias combo-box for editing
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }
   
   /**
    * Handle the Clone As button, in the Clone As dialog.
    */
   private void handleCloneAs()
   {
      // get the source info from selection
      String sourceName = (String) wpPtm.
                                   getValueAt(wpProcs.getSelectedRow(), 0);
      String sourceDesc = (String) wpPtm.
                                   getValueAt(wpProcs.getSelectedRow(), 1);
      
      // get and check the target process name
      String targetName = validateStringField(wpfN, 
                                              "type clone's process name");
      // no name specified, return
      if (targetName == null)
      {
         return;
      }
      
      // get the ACL option
      boolean withACLs = wpcW.isSelected();
      
      // clone the process
      AdminUtils.busy(client);
      boolean rc = adm.cloneProcess(sourceName, targetName, withACLs);
      AdminUtils.idle(client);
      
      // check the result
      if (!rc)
      {
         // set initial focus on the process name field
         wpfN.requestFocusInWindow();

         // alarm
         client.alarm();
         return;
      }
      
      // cloned; clear panel
      client.ok();
      dialog.dispose();
      dialog = null;
      
      // added to the server; clear selection and add to the list
      wpPsm.clearSelection();
      wpPtm.addRow(new TaggedName(targetName, sourceDesc));
   }
   
   /**
    * Handle Disabled Account check-box in process def dialog.
    */
   private void handleDisabledCheckbox()
   {
      Color color = wpdchkDs.isSelected() ? Color.red : Color.black;
      
      wpdlDs.setForeground(color);
   }

   /**
    * Handle Disable Process button press in west-panel.
    */
   private void handleDisableProcess()
   {
      String pName = process.subjectId;
      
      if (!AdminUtils.confirmDialog(client, 
                                   wpbP.getText(), 
                                   wpbP.getText() + " " + pName + " ?"))
      {
         return;
      }
      
      AdminUtils.busy(client);
      
      process = adm.getProcess(pName);
      if (process != null)
      {
         // toggle the disabled flag
         process.enabled = !process.enabled;
         adm.setProcess(process);
         
         setProcessAcountsSingleSelection();
      }
      else
      {
         client.alarm();
      }
      
      AdminUtils.idle(client);
   }

   /**
    * Handle Clone Process button press in west-panel.
    */
   private void handleClone()
   {
      // reset fields
      wpfN.setText("");
      wpcW.setSelected(false);
      
      // display modal dialog
      dialog = AdminUtils.modalDialog(client, "Clone Process", wpClonePanel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         dialog = null;
      }
   }
   
   /**
    * Handle Delete Selected button press in the west panel.
    */
   private void handleDeleteSelectedProcessAcounts()
   {
      // get the selection
      int[] sel = wpProcs.getSelectedRows();
      
      // safety check
      if (sel.length == 0)
      {
         return;
      }
      
      // ask confirmation
      String quest = sel.length == 1 ?
                     "1 process account being deleted. Continue?" :
                     Integer.toString(sel.length) + 
                     " process accounts being deleted. Continue?";
      
      if (!AdminUtils.confirmDialog(client, 
                                    "Deleting Process Accounts", 
                                    quest))
      {
         // declined
         return;
      }
      
      // busy
      AdminUtils.busy(client);

      // sort the array of selected indices to walk backwards
      Arrays.sort(sel);

      // walk the array backwards and delete selected processes
      for (int i = sel.length - 1; i >= 0; i--)
      {
         // pick up the name of process being deleted
         String process = (String) wpProcs.getValueAt((sel[i]), 0);
         
         // attempt deletion
         boolean rc = adm.deleteProcess(process);
         if (!rc)
         {
            // attempt has failed
            client.alarm();
         }
         else
         {
            // attempt has succeeded; delete this row
            client.ok();
            wpPtm.deleteRow(sel[i]);
         }
      }

      // idle
      AdminUtils.idle(client);
   }

   /**
    * This handles the Server check-box, displayed in the Process Def dialog.
    * <p>
    * When the Server check-box is selected, it will enable the Master check-box
    * too. Else, the Master check-box gets disabled and deselected. 
    */
   private void handleProcessServerCheckbox()
   {
      if (wpdchkS.isSelected() == false)
      {
         wpdchkM.setSelected(false);
         wpdchkM.setEnabled(false);
      }
      else
      {
         wpdchkM.setEnabled(true);
         wpdchkM.setSelected(false);
      }
   }
   
   /**
    * Handle Cancel button press in the Process Def dialog.
    */
   private void handleCancel()
   {
      // clear error messages
      client.ok();
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
   }
   
   /**
    * Handle Save button press in the Process Def panel.
    * Save button tries and adds or update the process account definition.
    * When adding, if the server responds OK, it will add the new row to the
    * west-side table. When editing, it will update the currently selected
    * row.
    * <p>
    * The following fields are mandatory: name and description. The Certificate
    * combo-box will display certificate aliases and description.
    */
   private void handleSave()
   {

      // determine the mode: add vs edit
      boolean add = wpdfN.isEnabled();
      
      // get and check the process name
      String name = validateStringField(wpdfN, "type a process name");
      if (name == null)
      {
         return;
      }
      
      // get and check the process description
      String desc = validateStringField(wpdfD, "type a process description");
      if (desc == null)
      {
         return;
      }

      // get the certificate alias
      TaggedName certTag= ((TaggedName) wpdcbCm.getSelectedItem());
      String certAlias = null;
      if (certTag == null)
      {
         // alarm
         client.beep("select a certificate");
         return;
      }
      
      certAlias = certTag.getName();
      
      // get the server flag
      boolean server = wpdchkS.isSelected();
      
      // get the master flag - ensure is ON only if server is ON too
      boolean master = server && wpdchkM.isSelected();
      
      // get the disabled account state
      boolean disabled = wpdchkDs.isSelected();
      
      // dispatch the action
      if (add)
      {
         // add the definition to the server and iterate
         // list remains with no selection
         if (saveProcessDef(name, desc, certAlias, server, master, disabled))
         {
            // successfully added to the server; add to the list
            client.ok();
            wpPtm.addRow(new TaggedName(name, desc));

            // refocus name field
            wpdfN.requestFocusInWindow();
            
            // clear selection now when the table is modified
            wpPsm.clearSelection();
            
            resetProcessDialog();
            return;
         }
         
         // adding definition failed; stay in the editing
         wpdfN.requestFocusInWindow();
         client.alarm();
      }
      else
      {
         // change the existing definition
         // list retains the selection
         if (setProcessDef(name, desc, certAlias, server, master, disabled))
         {
            // successfully saved changes
            client.ok();
            wpPtm.refreshRow(new TaggedName(name, desc),
                             wpProcs.getSelectedRow());
            
            // hide the dialog
            dialog.setVisible(false);
            dialog = null;

            setProcessAcountsSingleSelection();
            
            // re-select process certificate
            selectProcessAcountCertificate();
            
            return;
         }
         
         // saving definition failed; stay in the editing 
         wpdfN.requestFocusInWindow();
         client.alarm();
      }
   }

   /**
    * Adds new process account definition to the server.
    * 
    * @param    name
    *           new process name
    * @param    desc
    *           new process description
    * @param    alias
    *           new certificate alias
    * @param    server
    *           new server state
    * @param    master
    *           new master state
    * @param    disabled
    *           <code>true</code> if account is disabled
    * 
    * @return   <code>true</code> if added successfully
    */
   private boolean saveProcessDef(String name, 
                                  String desc, 
                                  String alias, 
                                  boolean server, 
                                  boolean master,
                                  boolean disabled)
   {
      ProcessDef pd = new ProcessDef();
      pd.subjectId   = name;
      pd.description = desc;
      pd.alias       = alias;
      pd.server      = server;
      pd.master      = master;
      pd.enabled     = !disabled;
      
      Serializable ext = null;
      if (extPlugin != null)
      {
         // validate the extensions, if a plugin was provided
         try
         {
            ext = extPlugin.validateExtensions(name);
         } 
         catch (AccountExtValidationException e)
         {
            client.beep(e.getMessage());
            AdminUtils.idle(client);
            return false;
         }         
      }
      
      AdminUtils.busy(client);
      boolean rc = adm.addProcess(pd, ext);
      if (rc)
      {
         process = pd;
      }
      AdminUtils.idle(client);
      
      return rc;
   }

   /**
    * Changes the existing process account definition on the server.
    * 
    * @param    name
    *           new process name
    * @param    desc
    *           new process description
    * @param    alias
    *           new certificate alias
    * @param    server
    *           new server state
    * @param    master
    *           new master state
    * @param    disabled
    *           <code>true</code> if account is disabled
    * 
    * @return   <code>true</code> if changed successfully
    */
   private boolean setProcessDef(String name, 
                                 String desc, 
                                 String alias, 
                                 boolean server, 
                                 boolean master,
                                 boolean disabled)
   {
      ProcessDef pd = new ProcessDef();
      pd.subjectId   = name;
      pd.description = desc;
      pd.alias       = alias;
      pd.server      = server;
      pd.master      = master;
      pd.enabled     = !disabled;
      
      Serializable ext = null;
      if (extPlugin != null)
      {
         // validate the extensions, if a plugin was provided
         try
         {
            ext = extPlugin.validateExtensions(name);
         } 
         catch (AccountExtValidationException e)
         {
            client.beep(e.getMessage());
            AdminUtils.idle(client);
            return false;
         }         
      }

      AdminUtils.busy(client);
      boolean rc = adm.setProcess(pd, ext);
      if (rc)
      {
         process = pd;
      }
      AdminUtils.idle(client);
      
      return rc;
   }

   /**
    * Handle the Assign button in the east-panel. 
    */
   private void handleAssignCertificate()
   {
      // ensure a single process is selected
      if (!singleProcess)
      {
         return;
      }
      
      String certname = (String) epCerts.getValueAt(epCerts.getSelectedRow(), 
                                                    0);

      String quest = "Assign certificate <" + certname +  "> " +
      		         " to process <" + process.subjectId + "> ?";

      if (!AdminUtils.confirmDialog(client, 
                                    "Assign Certificate", 
                                    quest))
      {
         // declined
         return;
      }
      
      // busy
      AdminUtils.busy(client);

      try
      {
         ProcessDef pdef = adm.getProcess(process.subjectId);
         if (pdef == null)
         {
            // process may have been deleted
            client.alarm();
            
            // refresh the data from the server
            handlePopulateProcessAccounts();
            return;
         }

         pdef.alias = certname;

         // send new definition to the server
         if (adm.setProcess(pdef))
         {
            process = pdef;

            // re-verify the selected certificate
            verifySelectedCertificate();

            client.ok();
         }
         else
         {
            client.alarm();
         }
      }
      finally
      {
         // idle
         AdminUtils.idle(client);         
      }
   }

   /**
    * Handle Switch to ACLs button press in the west panel.
    * <p>
    * This method should passes the process name to the ACLs screen and 
    * initiate Access Control --&gt; Show By Account menu choice
    */
   private void handleSwitchToACLs()
   {
      client.show(AdminACL.SCREEN_ACLS, AdminACL.TITLE_ACLS);
      AdminACL adminAcl = client.getMenu().getACL().getAclAdmin();
      
      adminAcl.showByAccount(process.subjectId);
   }

   /**
    * Handle Switch to Certificates button press in the west panel.
    */
   private void handleSwitchToCertificates()
   {
      String cert = (String) epCerts.getValueAt(epCerts.getSelectedRow(), 0);
      client.show(AdminCertificates.SCREEN_ACC_CERTS, 
                  AdminCertificates.TITLE_ACC_CERTS);
      AdminCertificates ca = client.getMenu().getAccounts().getCA();
      
      ca.init(cert);
   }
}