/*
** Module   : AdminScreen.java
** Abstract : Defines common functionalities for admin client screens.
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 CA  20090507   @42115 Created initial version. 
** 002 CA  20090512   @42146 Added another string validation method. Modified
**                           TaggedDataTableModel to allow non-sorted mode.
** 003 CA  20090522   @42428 Extracted TaggedDataTableModel into a separate
**                           file. Added methods to layout components in a 
**                           'matrix' mode, using GroupLayout.
** 004 CA  20090529   @42490 Added borderComponents, to layout components using 
**                           border layout, horizontally or vertically.
** 005 CA  20090714   @43157 Added methods to validate an integer field.
** 006 NVS 20090803   @43485 Refactored the code and relocated panel creation  
**                           methods using border and group layouts from here.
** 007 SIY 20091116   @44338 Refactoring of things related to TaggedName. 
**                           Minor cleanups.
** 008 SVL 20091126   @44450 Added validateLongField* functions.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;

import com.goldencode.p2j.admin.*;

/**
 * Define common functionalities for admin client screens.
 * <p>
 * This also defines the following inner classes:
 * <ul>
 *    <li> 
 *       {@link TaggedDataComboModel}
 *       <p>
 *       This class defines a model to display a two-column combo-box, with 
 *       each row backed by a {@link TaggedName} instance.
 *    </li>
 *    <li> 
 *       {@link TaggedDataComboRenderer}
 *       <p>
 *       This class defines a model to render a two-column combo-box, with
 *       each row backed by a {@link TaggedName} instance.
 *    </li>
 * </ul>
 * 
 * @author CA
 */
public class AdminScreen
{
   /** 
    * Property to be displayed for a certificate, in any sections where
    * certificate description is needed. 
    */
   protected final String CERT_DESCRIPTION_PROP = "CN";
   
   /** Related instance of AdminClient */
   protected AdminClient client = null;

   /** Admin Server APIs */
   protected AdminExports adm = null;
   
   /** main swappable screen */
   protected JPanel mainPanel = null;
   
   /** state machine's array of components */
   protected JComponent[] state = null;
   
   /** holds the most recent modal dialog instance */
   JDialog dialog = null;
   
   /**
    * Sets the state of the components according to the encoded array.
    * 
    * @param    newState
    *           array of boolean states, one per component, in a predefined
    *           order. <code>true</code> means enabled.
    */
   protected void stateMachine(boolean[] newState) 
   {
      if (newState.length != state.length)
      {
         throw new RuntimeException("Invalid state machine state definition");
      }
      
      for (int i = 0; i < state.length; i++)
      {
         state[i].setEnabled(newState[i]);
      }
   }
   
   /**
    * Selectively sets the state of the components according to the encoded
    * array.
    * 
    * @param    which
    *           array of booleans acting as a mask,one boolean per component.
    *           <code>true</code> means component changes state as specified
    *           in newState array.
    * 
    * @param    newState
    *           array of boolean states, one per component, in a predefined
    *           order. <code>true</code> means enabled.
    */
   protected void stateMachine(boolean[] which, boolean[] newState) 
   {
      if (which.length    != state.length ||
          newState.length != state.length)
      {
         throw new RuntimeException("Invalid state machine state definition");
      }
      
      for (int i = 0; i < state.length; i++)
      {
         if (which[i])
         {
            state[i].setEnabled(newState[i]);
         }
      }
   }
   
   /**
    * Selectively sets the state of the components according to the encoded
    * array and returns the array of the saved current state. The saved state
    * can be used as input to the stateMachine(boolean[]) method call.
    * 
    * @param    which
    *           array of booleans acting as a mask,one boolean per component.
    *           <code>true</code> means component changes state as specified
    *           in newState array.
    * 
    * @param    newState
    *           array of boolean states, one per component, in a predefined
    *           order. <code>true</code> means enabled.
    * 
    * @return   saved state of all components
    */
   protected boolean[] stateMachineModal(boolean[] which, boolean[] newState) 
   {
      if (which.length    != state.length ||
          newState.length != state.length)
      {
         throw new RuntimeException("Invalid state machine state definition");
      }
      
      boolean[] saved = new boolean[state.length];
      
      for (int i = 0; i < state.length; i++)
      {
         saved[i] = state[i].isEnabled();
         
         if (which[i])
         {
            state[i].setEnabled(newState[i]);
         }
      }
      
      return saved;
   }

   /**
    * Validate the given text field. Its content are expected to be a valid
    * integer value. If the field has no content, it will return
    * <code>null</code>; also, the specified error will be displayed.
    * <p>
    * Else, it will return the field content.
    * 
    * @param    jtf
    *           The text field which needs to be validated.
    * @param    err
    *           The error to be displayed if the field has no content.
    * 
    * @return   field content or <code>null</code> if field is empty.
    */
   protected Integer validateIntField(JTextField jtf, String err)
   {
      return validateIntField(jtf.getText(), err, jtf);
   }
   
   /**
    * Validate the given text. If the text is a valid integer value, it will
    * return it.
    * <p>
    * Else, it will return <code>null</code> and the specified error will be
    * displayed.
    * 
    * @param    text
    *           The text which needs to be validated.
    * @param    err
    *           The error to be displayed if the text is not a valid integer.
    * @param    comp
    *           The component to which the text belongs to
    * 
    * @return   original text or <code>null</code> if text is empty.
    */
   protected Integer validateIntField(String text,
                                      String err, 
                                      Component comp)
   {
      if (text == null)
      {
         return null;
      }

      try
      {
         return Integer.parseInt(text);
      }
      catch (NumberFormatException nfe)
      {
         // set initial focus on the field
         comp.requestFocusInWindow();
         
         // alarm
         client.beep(err);
         return null;
      }
   }

   /**
    * Validate the given text field. Its content are expected to be a valid
    * long value. If the field has no content, it will return
    * <code>null</code>; also, the specified error will be displayed.
    * <p>
    * Else, it will return the field content.
    *
    * @param    jtf
    *           The text field which needs to be validated.
    * @param    err
    *           The error to be displayed if the field has no content.
    *
    * @return   field content or <code>null</code> if field is empty.
    */
   protected Long validateLongField(JTextField jtf, String err)
   {
      return validateLongField(jtf.getText(), err, jtf);
   }

   /**
    * Validate the given text. If the text is a valid long value, it will
    * return it.
    * <p>
    * Else, it will return <code>null</code> and the specified error will be
    * displayed.
    *
    * @param    text
    *           The text which needs to be validated.
    * @param    err
    *           The error to be displayed if the text is not a valid long
    *           value.
    * @param    comp
    *           The component to which the text belongs to.
    *
    * @return   original text or <code>null</code> if text is empty.
    */
   protected Long validateLongField(String text,
                                    String err,
                                    Component comp)
   {
      if (text == null)
      {
         return null;
      }

      try
      {
         return Long.parseLong(text);
      }
      catch (NumberFormatException nfe)
      {
         // set initial focus on the field
         comp.requestFocusInWindow();

         // alarm
         client.beep(err);
         return null;
      }
   }

   /**
    * Validate the given text field. Its content are expected to be a
    * non-empty string. If the field has no content, it will return
    * <code>null</code>; also, the specified error will be displayed.
    * <p>
    * Else, it will return the field content.
    * 
    * @param    jtf
    *           The text field which needs to be validated.
    * @param    err
    *           The error to be displayed if the field has no content.
    * 
    * @return   field content or <code>null</code> if field is empty.
    */
   protected String validateStringField(JTextField jtf, String err)
   {
      return validateStringField(jtf.getText(), err, jtf);
   }
   
   /**
    * Validate the given text area. Its content are expected to be a non-empty
    * string. If the field has no content, it will return <code>null</code>;
    * also, the specified error will be displayed.
    * <p>
    * Else, it will return the field content.
    * 
    * @param    jta
    *           The text area which needs to be validated.
    * @param    err
    *           The error to be displayed if the field has no content.
    * 
    * @return   field content or <code>null</code> if field is empty.
    */
   protected String validateStringField(JTextArea jta, String err)
   {
      return validateStringField(jta.getText(), err, jta);
   }
   
   /**
    * Validate the given text. If the text is not empty, it will return it.
    * <p>
    * Else, it will return the <code>null</code> and the specified error will
    * be displayed.
    * 
    * @param    text
    *           The text which needs to be validated.
    * @param    err
    *           The error to be displayed if the text is empty.
    * @param    comp
    *           The component to which the text belongs to
    * 
    * @return   original text or <code>null</code> if text is empty.
    */
   protected String validateStringField(String text, 
                                        String err, 
                                        Component comp)
   {
      if (text.length() == 0)
      {
         // set initial focus on the field
         comp.requestFocusInWindow();
         
         // alarm
         client.beep(err);
         return null;
      }

      return text;
   }
   
   /**
    * For the given certificates, set their description to the specified
    * property.
    * 
    * @param    list
    *           List of certificate data
    * @param    prop
    *           The owner property which should be extracted from the CN
    *           entry.
    */
   protected void setCertificateDescription(TaggedName[] list, 
                                            String prop)
   {
      for (int i = 0; i < list.length; i++)
      {
         list[i].setTag(CertificateUtils.getCertificateDetail(list[i].getTag(), 
                                                             prop));
      }
   }

   /**
    * Creates a titled group of related controls with input and button.
    * 
    * @param    title
    *           group title
    * 
    * @param    input
    *           input field to add to the group
    * 
    * @param    control
    *           button to add to the group
    * 
    * @return   new panel with the group of controls
    */
   protected JPanel createGroup(String title, JTextField input, JButton control) 
   {
      JPanel panel = new JPanel();
      
      Border empty = new EmptyBorder(/*t*/0, /*l*/3, /*b*/3, /*r*/3);
      Border etchd = new EtchedBorder(EtchedBorder.LOWERED);
      Border compn = new CompoundBorder(etchd, empty);
      Border bordr = new TitledBorder(compn, title);
      
      panel.setBorder(bordr);
      GroupLayout layout = new GroupLayout(panel);
      panel.setLayout(layout);

      // Turn on automatically adding gaps between components
      layout.setAutoCreateGaps(true);
    
      // Create a sequential group for the horizontal axis.
      GroupLayout.SequentialGroup hGroup = layout.createSequentialGroup();
    
      hGroup.
         addComponent(input).
         addComponent(control)
         ;
      layout.setHorizontalGroup(hGroup);
      
      // Create a sequential group for the vertical axis.
      GroupLayout.ParallelGroup vGroup = layout.createParallelGroup
                                         (GroupLayout.Alignment.BASELINE);
    
      vGroup.
         addComponent(input).
         addComponent(control)
         ;
      layout.setVerticalGroup(vGroup);
      
      return panel;
   }

   /**
    * A custom ComboboxModel to feed JCombobox from an array of TaggedNames.
    */
   class TaggedDataComboModel
   extends AbstractListModel
   implements javax.swing.ComboBoxModel
   {
      /** Storage for tagged data */
      ArrayList<TaggedName> data = new ArrayList<TaggedName>();
      
      /** Stores the currently selected index, or -1 if no item selected. */
      private int selectedIndex = -1;
      
      /**
       * Returns the length of the list.
       * 
       * @return   the length of the list
       */
      public int getSize()
      {
         return data.size();
      }
   
      /**
       * Returns the value at the specified index.
       * 
       * @param    index
       *           the requested index
       * 
       * @return   the value at <code>index</code> or <code>null</code> if index
       *           is outside list.
       */
      public Object getElementAt(int index)
      {
         if (index < 0 || index >= data.size())
         {
            return null;
         }
   
         return data.get(index);
      }
   
      /**
       * Set the selected item. The implementation of this method should
       * notify all registered <code>ListDataListener</code>s that the
       * contents have changed.
       * 
       * @param    anItem
       *           the list object to select or <code>null</code> to clear the
       *           selection
       */
      public void setSelectedItem(Object anItem)
      {
         if (anItem == null)
         {
            selectedIndex = -1;
         }
         else
         {
            TaggedName tag = null;
            if (anItem instanceof TaggedName)
            {
               tag = (TaggedName) anItem;
            }
            else
            {
               tag = new TaggedName((String) anItem, "");
            }
            selectedIndex = data.indexOf(tag);
         }
         
         fireContentsChanged(this, -1, -1);
      }
   
      /**
       * Returns the selected item.
       * 
       * @return   The selected item or <code>null</code> if there is no
       *           selection
       */
      public Object getSelectedItem()
      {
         if (selectedIndex < 0 || selectedIndex >= data.size())
         {
            return null;
         }
   
         return data.get(selectedIndex);
      }
   
      /**
       * Deletes all items.
       */
      public void clear()
      {
         int rows = data.size();
         if (rows > 0)
         {
            // notify listeners
            fireIntervalRemoved(this, 0, rows - 1);
            
            // clear the storage
            data.clear();
         }
      }
      
      /**
       * Clears and then reloads the table from a data source.
       * 
       * @param    list
       *           array of tagged names to load into the model
       */
      public void loadAll(TaggedName[] list)
      {
         // clear first
         int rows = data.size();
         if (rows > 0)
         {
            clear();
         }
         
         // deal with the input
         if (list.length > 0)
         {
            // sort the input
            Arrays.sort(list);
            
            // stuff the table next
            for (int i = 0; i < list.length; i++)
            {
               data.add(list[i]);
            }
            
            // notify listeners
            fireIntervalAdded(this, 0, list.length - 1);
         }
      }
   }

   /**
    * A custom ListCellRenderer to render a two-column combo-box.
    */
   class TaggedDataComboRenderer
   implements ListCellRenderer
   {
      /** The panel used to render a combo item. */
      private JPanel renderer;
      
      /** The label which renders data for first column (data name). */
      private JLabel nameLabel;
   
      /** The label which renders data for the second column (data tag). */
      private JLabel tabLabel;
    
      public TaggedDataComboRenderer()
      {
         // create a panel to hold two columns
         renderer = new JPanel();
         renderer.setLayout(new GridLayout(0, 2));
   
         // create the label which holds data for the NAME value
         nameLabel = new JLabel(" ");
         Dimension d1 = nameLabel.getPreferredSize();
         nameLabel.setMaximumSize(d1);
         nameLabel.setPreferredSize(d1);
   
         // create the label which holds data for the TAG value
         tabLabel = new JLabel(" ");
         Dimension d2 = tabLabel.getPreferredSize();
         tabLabel.setMaximumSize(d2);
         tabLabel.setPreferredSize(d2);
         
         renderer.add(nameLabel);
         renderer.add(tabLabel);
      }
    
      /**
       * Render the specified value.
       * 
       * @param    list
       *           The JList we're painting.
       * @param    value
       *           The value returned by list.getModel().getElementAt(index).
       * @param    index
       *           The cells index.
       * @param    isSelected
       *           True if the specified cell was selected.
       * @param    cellHasFocus
       *           True if the specified cell has the focus.
       * 
       * @return   A component whose paint() method will render the specified
       *           value.
       */
      public Component getListCellRendererComponent(JList list, 
                                                    Object value,
                                                    int index, 
                                                    boolean isSelected, 
                                                    boolean cellHasFocus)
      {
         String s1 = "";
         String s2 = "";
         if (value instanceof TaggedName)
         {
            TaggedName item = (TaggedName) value;
            s1 = item.getName();
            s2 = item.getTag();
         }
         else
         {
            // this is needed for editable combo's
            s1 = value == null ? "" : value.toString();
         }
         
         nameLabel.setText(s1);
         tabLabel.setText(s2);
         
         renderer.setBackground(isSelected ? list.getSelectionBackground() 
                                           : null);
         renderer.setForeground(isSelected ? list.getSelectionForeground() 
                                           : null);

         return renderer;
      }
   }
}