/*
** Module   : AdminSelectionListDialog.java
** Abstract : Dialog which allow to select a single item from a list.
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SVL 20111209 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * Dialog which allow to select a single item from a list.
 */
public class AdminSelectionListDialog
{
   /** Main panel of the dialog. */
   protected JPanel mainPanel;

   /** Related AdminClient instance. */
   protected AdminClient client;

   /** Dialog title. */
   protected String dialogTitle;

   /** OK button. */
   protected JButton btnOk;

   /** Cancel button. */
   protected JButton btnCancel;

   /** Table which contains selectable items. */
   protected JTable table;

   /** Table model for selectable items. */
   protected TaggedDataTableModel tableModel;

   /** Text of the message which is displayed if an item is not selected. */
   protected String selectOneErrorText;

   /** Title of the message which is displayed if an item is not selected. */
   protected String selectOneErrorTitle;

   /** Scroller of the table which contains selectable items. */
   protected JScrollPane tableScroller;

   /** Text area for displaying help text above the table. */
   protected JTextArea txtHelp;

   /** Panel which contains buttons. */
   protected JPanel controlPanel;

   /** TaggedName object which represent the selected item. */
   protected TaggedName result;

   /**
    * Create a dialog with "OK" and "Cancel" buttons.
    *
    * @param client
    *        Related instance of AdminClient.
    * @param dialogTitle
    *        Dialog title.
    * @param tableTitle
    *        Table title.
    * @param columns
    *        Array of column names.
    * @param helpText
    *        Help text to be displayed above the table or <code>null</code>
    *        if it is not needed.
    * @param selectOneErrorText
    *        Text of the message which is displayed if an item is not
    *        selected.
    * @param selectOneErrorTitle
    *        Title of the message which is displayed if an item is not
    *        selected.
    */
   public AdminSelectionListDialog(AdminClient client,
                                   String dialogTitle,
                                   String tableTitle,
                                   String[] columns,
                                   String helpText,
                                   String selectOneErrorText,
                                   String selectOneErrorTitle)
   {
      this(client, dialogTitle, tableTitle, columns, helpText,
           selectOneErrorText, selectOneErrorTitle, "OK", "Cancel");
   }

   /**
    * Create a dialog.
    *
    * @param client
    *        Related instance of AdminClient.
    * @param dialogTitle
    *        Dialog title.
    * @param tableTitle
    *        Table title.
    * @param columns
    *        Array of column names.
    * @param helpText
    *        Help text to be displayed above the table or <code>null</code>
    *        if it is not needed.
    * @param selectOneErrorText
    *        Text of the message which is displayed if an item is not
    *        selected.
    * @param selectOneErrorTitle
    *        Title of the message which is displayed if an item is not
    *        selected.
    * @param okButtonCaption
    *        Custom caption for the "OK" button.
    * @param cancelButtonCaption
    *        Custom caption for the "Cancel" button.
    */
   public AdminSelectionListDialog(AdminClient client,
                                   String dialogTitle,
                                   String tableTitle,
                                   String[] columns,
                                   String helpText,
                                   String selectOneErrorText,
                                   String selectOneErrorTitle,
                                   String okButtonCaption,
                                   String cancelButtonCaption)
   {
      this.client = client;
      this.dialogTitle = dialogTitle;
      this.selectOneErrorText = selectOneErrorText;
      this.selectOneErrorTitle = selectOneErrorTitle;

      tableModel = new TaggedDataTableModel(columns);
      tableScroller = new JScrollPane();
      table = AdminUtils.createBorderedScrolledTable(tableTitle,
                        tableModel, tableScroller);

      if (helpText != null)
         txtHelp = AdminUtils.createReadOnlyTextArea(helpText);

      btnOk = new JButton(okButtonCaption);
      btnCancel = new JButton(cancelButtonCaption);

      buildControlPanel();
      buildMainPanel();
   }

   /**
    * Display the dialog.
    *
    * @param  data
    *         Data to be displayed in the selection table. Can be
    *         <code>null</code> if data was previously loaded into the table
    *         using other means.
    *
    * @return TaggedName object which represents the selected item or
    *         <code>null</code> if the dialog was closed using the "Cancel"
    *         button.
    */
   public TaggedName runDialog(TaggedName[] data)
   {
      if (data != null)
         tableModel.loadAll(data);
      final JDialog dialog = AdminUtils.modalDialog(client,
                                                    dialogTitle,
                                                    mainPanel);

      result = null;

      btnCancel.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            dialog.setVisible(false);
         }
      });

      btnOk.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            int selectedRow = table.getSelectedRow();

            if (selectedRow >= 0)
            {
               result = tableModel.data.get(selectedRow);
               dialog.setVisible(false);
            }
            else
            {
               JOptionPane.showMessageDialog(client,
                        selectOneErrorText,
                        selectOneErrorTitle,
                        JOptionPane.ERROR_MESSAGE);
            }
         }
      });

      dialog.setVisible(true);
      dialog.dispose();
      return result;
   }

   /**
    * Build control panel with buttons.
    */
   protected void buildControlPanel()
   {
      controlPanel = AdminUtils.groupComponents(
                        new Component[][]{{btnCancel, btnOk}});
   }

   /**
    * Build main panel of the dialog.
    */
   protected void buildMainPanel()
   {
      mainPanel = AdminUtils.groupComponents(
            txtHelp == null ?
            new Component[][]{{tableScroller}, {controlPanel}} :
            new Component[][]{{txtHelp}, {tableScroller}, {controlPanel}});
   }
}