/*
** Module   : AdminSessions.java
** Abstract : sessions list screen
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description--------------------
** 001 GES 20090611   @42759 Created the first version to provide a list of
**                           sessions and termination of a session.
** 002 GES 20090622   @42781 Finished support for session termination.
** 003 GES 20090721   @43302 Minor source code cleanup.
** 004 NVS 20090723   @43354 Tables don't allow moving columns.
** 005 NVS 20090729   @43432 Implemented sessions printing.
** 006 NVS 20090803   @43486 Using a better approach to printing with extended
**                           detail choice.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.print.PrinterException;
import java.text.MessageFormat;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.*;

import com.goldencode.p2j.admin.AdminExports;
import com.goldencode.p2j.admin.SessionInfo;

/**
 * Provides a screen that shows the list of sessions for the current server
 * as well as some basic management of those sessions. 
 */
public class AdminSessions 
extends AdminScreen
implements PrintRequest
{
   /** Screen switcher - session list screen ID. */
   public static final String SCREEN_SESSIONS = "adm_sessions";
   
   /** Screen title. */
   private static final String TITLE_SESSIONS = "Session Management";
   
   /** Session termination confirmation text. */
   private static final String CONFIRM_SPEC = "Are you sure you want to "   +
                                              "forcibly terminate session " +
                                              "%d?";

   /** Session list. */
   private JTable table = null;

   /** Session list data model. */
   private SessionInfoModel list = null;

   /** Subpanel that holds buttons and other controls. */
   private JPanel ctrlPanel = null;
   
   /** Refresh button. */
   private JButton refresh = null;

   /** Terminate Session button. */
   private JButton term = null;
   
   /** Filter button. */
   private JButton filter = null;
   
   /** Sessions table sorter */
   private TableRowSorter sorter = null;
   
   /** Cache the last filter string. */
   private String lastFilter = null;
   
   /** 
    * Constructor.
    *
    * @param    client
    *           Core UI module of which we are a part.
    */
   AdminSessions(AdminClient client)
   {
      // store useful references
      this.client = client;
      adm = client.getExports();
      
      // create the main screen and add it to the swapper
      mainPanel = new JPanel();
      client.add(mainPanel, SCREEN_SESSIONS, this);
      mainPanel.setLayout(new BorderLayout());
      
      table = new JTable();
      table.getTableHeader().setReorderingAllowed(false);
      
      JScrollPane scrollPane = new JScrollPane(table);
      scrollPane.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
      table.setFillsViewportHeight(true);
      table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      
      mainPanel.add(scrollPane, BorderLayout.CENTER);

      // create control widgets
      ctrlPanel = new JPanel(new GridLayout(1, 0));
      ctrlPanel.setBorder(BorderFactory.createEmptyBorder(0,10,10,10));
      
      refresh = new JButton("Refresh");
      filter  = new JButton("Filter");
      term    = new JButton("Terminate Session");
      term.setEnabled(false);
      
      ctrlPanel.add(refresh);
      ctrlPanel.add(filter);
      ctrlPanel.add(term);
      
      mainPanel.add(ctrlPanel, BorderLayout.SOUTH);
      
      // filter button
      filter.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent evt) 
         {
            filter();
         }
      });
      
      // refresh button
      refresh.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent evt) 
         {
            refresh(true);
         }
      });
      
      // terminate button
      term.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent evt) 
         {
            terminate();
         }
      });
      
      // create the array of components for the state machine
      state = new JComponent[]
      {
         table,
         ctrlPanel,
         refresh,
         filter,
         term
      };
      
      // populate the session list upfront
      refresh(true);
   }
   
   /**
    * Performs printing of the current screen data.
    * This screen prints the current view of the sessions table.
    */
   public void print()
   {
      MessageFormat header = new MessageFormat("Active Sessions");
      MessageFormat footer = new MessageFormat("page {0, number}");
      
      try
      {
         table.print(JTable.PrintMode.FIT_WIDTH, header, footer);
      }
      catch (PrinterException pe)
      {
         client.beep("printer exception" + pe);
      }
      catch (SecurityException se)
      {
         // nothing to do
      }
   }
   
   /**
    * Queries the extended printing capability. If the exteneded capability
    * is not provided, the printExtended() method should not be called.
    *
    * @return <code>false</code> to tell the caller the extended printing 
    *         is not implemented.
    */
   public boolean canPrintExtended()
   {
      return false;
   }

   /**
    * Does nothing here.
    */
   public void printExtended()
   {
   }
   
   /** 
    * Refresh the display with the latest list of sessions.
    *
    * @param    standalone
    *           <code>true</code> if this is called as a standalone function.
    *           In such a case, the cursor is managed as part of the processing.
    */
   void refresh(boolean standalone)
   {
      // not connected yet
      if (adm == null)
      {
         client.beep("Not connected!");
         return;
      }
      
      if (standalone)
      {
         // busy
         AdminUtils.busy(client);
      }
      
      SessionInfo[] info = adm.getSessionList();
      
      String msg = "Cannot obtain list of sessions.";
      
      if (info != null)
      {
         msg  = "";
         list = new SessionInfoModel(info);
         table.setModel(list);
         
         int num = list.getColumnCount();
         
         for (int i = 0; i < num; i++)
         {
            TableColumn column = table.getColumnModel().getColumn(i);
            column.setPreferredWidth(list.getColumnWidth(i));
         }
         
         sorter = new TableRowSorter(list);
         sorter.setRowFilter(null);
         table.setRowSorter(sorter);
         
         lastFilter = null;
         
         // force a default sort on the account name column
         sorter.toggleSortOrder(0);
         
         // enable/disable the terminate button based on selection changes
         ListSelectionModel selectionModel = table.getSelectionModel();
         selectionModel.addListSelectionListener(new ListSelectionListener()
         {
            public void valueChanged(ListSelectionEvent e)
            {
               ListSelectionModel lsm = (ListSelectionModel) e.getSource();
               term.setEnabled(!lsm.isSelectionEmpty());
            }
         });
      }
      
      if (standalone)
      {
         client.setTargetInfo(AdminClient.TARGET_ERROR, msg);
         
         // idle
         AdminUtils.idle(client);
      }
   }
   
   /**
    * Process the filter dialog and create/install the resulting filter into
    * the table sorter.
    */
   private void filter() 
   {
      int    type = JOptionPane.QUESTION_MESSAGE;
      String txt  = "Define the filtering specification below. Press the\n" +
                    "'OK' button without any specification to remove any\n" +
                    "previous filter.";
      String spec = (String) JOptionPane.showInputDialog(client,
                                                         txt,
                                                         "Filter",
                                                         type,
                                                         null,
                                                         null,
                                                         lastFilter);
      
      if (spec == null || spec.trim().length() == 0)
      {
         // remove filtering
         sorter.setRowFilter(null);
         lastFilter = null;
      }
      else
      {
         // install a filter
         sorter.setRowFilter(AdminUtils.createFilter(spec));
         lastFilter = spec;
      }
   }
   
   /** 
    * Terminate the selected session.
    */
   void terminate()
   {
      // not connected yet
      if (adm == null)
      {
         client.beep("Not connected!");
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      int row = table.getSelectedRow();
      
      String msg = null;
      
      if (row != -1)
      {
         int    sid = -1;
         Object obj = list.getValueAt(table.convertRowIndexToModel(row), 3);
         
         if (obj instanceof Integer)
         {
            Integer id = (Integer) obj;
            sid = id.intValue();
            
            String ques = String.format(CONFIRM_SPEC, sid);
            
            if (AdminUtils.confirmDialog(client,
                                         "Session Termination Confirmation",
                                         ques))
            {
               try
               {
                  adm.terminateSession(sid);
                  
                  msg = String.format("Session %d was terminated.", sid);
               }
               
               catch (Exception exc)
               {
                  String err = exc.getMessage();
                  
                  if (err == null)
                  {
                     err = exc.toString();
                  }
                  
                  msg = String.format("Failure terminating session %d (%s).",
                                      sid,
                                      err);
               }
            }
            else
            {
               msg = "Session termination canceled at user's request.";
            }
         }
         else
         {
            msg = String.format("Invalid session ID in row %d, column 3.", row);
         }
      }
      else
      {
         msg = "No session selected.";
      }
      
      client.setTargetInfo(AdminClient.TARGET_ERROR, msg);
      
      // reset the table's data
      refresh(false);
      
      // idle
      AdminUtils.idle(client);
   }
   
   /** 
    * Presents the refreshed sessions list screen.
    */
   void manageSessions() 
   {
      // switch to this screen without changing state of components
      client.show(SCREEN_SESSIONS, TITLE_SESSIONS);
   }
   
   /**
    * Provides a table data model to represent session descriptors.
    */
   private class SessionInfoModel
   extends AbstractTableModel
   {
      /** Column names. */
      private final String[] columnNames =
      {
         "Account",
         "Description",
         "Type",
         "Session ID",
         "Remote Node",
         "Local Node",
         "Secure",
         "Logon Timestamp",
         "Generation #"
      };
      
      /** Default column sizes. */
      private final int columnSizes[] =
      {
         50,
         100,
         15,
         25,
         70,
         70,
         15,
         150,
         35
      };
      
      /** Column data types. */
      private final Class[] columnTypes =
      {
         String.class,
         String.class,
         String.class,
         Integer.class,
         String.class,
         String.class,
         Boolean.class,
         String.class,
         Integer.class
      };
      
      /**
       * Construct an instance.
       *
       * @param    data
       *           The data to provide to the table.
       */
      public SessionInfoModel(SessionInfo[] data)
      {
         this.data = data;
      }
      
      /** Core data for the table. */
      private SessionInfo[] data = null;
   
      /**
       * Gets the number of columns in the table model.
       *
       * @return   The number of columns.
       */
      public int getColumnCount()
      {
         return columnNames.length;
      }
   
      /**
       * Gets the width of the given column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The width in pixels.
       */
      public int getColumnWidth(int col)
      {
         return columnSizes[col];
      }
   
      /**
       * Gets the number of rows in the table model.
       *
       * @return   The number of rows.
       */
      public int getRowCount()
      {
         return data.length;
      }
   
      /**
       * Gets the name of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column name.
       */
      public String getColumnName(int col)
      {
         return columnNames[col];
      }
   
      /**
       * Get the cell value to be displayed.
       *
       * @param    col
       *           The column index (0-based).
       * @param    row
       *           The row index (0-based).
       */
      public Object getValueAt(int row, int col)
      {
         Object      datum = null;
         SessionInfo info  = data[row];
         
         switch (col)
         {
            case 0:
               datum = info.name;
               break;
            case 1:
               datum = info.descr;
               break;
            case 2:
               datum = info.user ? "User" : "Process";
               break;
            case 3:
               datum = new Integer(info.id);
               break;
            case 4:
               datum = info.remoteNet;
               break;
            case 5:
               datum = info.localNet;
               break;
            case 6:
               datum = new Boolean(info.secure);
               break;
            case 7:
               datum = String.format("%1$Tm/%1$Td/%1$TY %1$TT %1$Tz",
                                     info.logonTime);
               break;
            case 8:
               datum = new Integer(info.cacheGen);
               break;
         }
         
         return datum;
      }
   
      /**
       * Gets the data type of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column type.
       */
      public Class getColumnClass(int col)
      {
         return columnTypes[col];
      }
   }
}