/*
** Module   : AdminTarget.java
** Abstract : Target management for the admin session
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090116  @41621  Created the first version of the file.
** 002 NVS 20090318  @41633  Added cache serial number display for live
**                           target and refresh action support.
** 003 NVS 20090323  @41667  Refactored one more time to get rid of static
**                           members.
** 004 NVS 20090428  @42042  Added target change notification framework.
** 005 NVS 20090429  @42082  Added bracketing of server calls with busy/idle
**                           mouse cursors.
** 006 CA  20090604  @42562  When going to live target, loads the custom 
**                           account extension plugin; if this fails, it will
**                           logoff and display an error dialog.
** 007 NVS 20090611  @42659  Fixed error message.
** 008 CA  20090616  @42699  Fixed NPE - load account extension plugin after 
**                           the target is set to live. Also, cache the admin
**                           definitions.
** 009 CA  20090714  @43158  Build the ACL menu accordingly to new specs.
** 010 CA  20090720  @43246  When going to live target, load and cache the 
**                           custom ACL editors.
** 011 NVS 20090724  @43391  Added getResourceTypes() method.
** 012 NVS 20090807  @43552  Refactored in support of custom menu extensions.
** 013 SIY 20091217  @44517  "Refresh" menu item is enabled only if user has
**                           enough rights.
** 014 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import javax.swing.*;

import com.goldencode.p2j.admin.*;

/**
 * This is a target management class for the admin session. It offers multiple
 * choices of targets:
 * <ul>
 *   <li>live - the active directory on the target server will be edited
 *   <li>open - another existing directory on the target server will be edited
 *   <li>new - another new directory on the target server will be created
 * </ul>
 * For the live target it also offers an option of refreshing the active
 * directory.
 * 
 * @author NVS
 */
public class AdminTarget 
implements AdminNotify
{
   /** Related instance of AdminClient */
   private AdminClient client = null;

   /** Related instance of AdminMenu */
   private AdminMenu menu = null;

   /** Related instance of AdminLogon */
   private AdminLogon logon = null;

   /** Admin Server APIs */
   private AdminExports adm = null;

   /** Last used serial number for security cache creation. */
   private int cacheSerial = -1;

   /** target type none definition */
   private static final int TARGET_TYPE_NONE = 0;
   
   /** target type live definition */
   private static final int TARGET_TYPE_LIVE = 1;
   
   /** target type file definition */
   private static final int TARGET_TYPE_FILE = 2;
   
   /** current target type */
   private int targetType = TARGET_TYPE_NONE;

   /** 
    * Constructor.
    */
   AdminTarget()
   {
   }

   /** 
    * Initializes useful references.
    *
    * @param  client
    *         instance of AdminClient
    *
    * @param  menu
    *         instance of AdminMenu
    *
    * @param  logon
    *         instance of AdminLogon
    */
   void init(AdminClient client, AdminMenu menu, AdminLogon logon)
   {
      this.client = client;
      this.menu = menu;
      this.logon = logon;

      // register for logoff notifications
      client.registerLogoffListener(this);
   }

   /** 
    * Returns <code>true</code> if the target is set to LIVE.
    *
    * @return <code>true</code> if the target is set to LIVE.
    */
   boolean isLive()
   {
      return targetType == TARGET_TYPE_LIVE;
   }
      
   /** 
    * Returns <code>true</code> if the target is set to NONE.
    *
    * @return <code>true</code> if the target is set to NONE.
    */
   boolean isNone()
   {
      return targetType == TARGET_TYPE_NONE;
   }
      
   /** 
    * Notifies about successful logoff and gives chance to release resources.
    *
    * @param eventCode
    *        event code, one of the NOTIFY_* constants
    */
   public void notifyEvent(int eventCode)
   {
      // filter out unwanted events
      if (eventCode != AdminNotify.NOTIFY_LOGOFF)
      {
         return;
      }
      
      // reset target type
      targetType = TARGET_TYPE_NONE;
      
      // reset security cache serial number
      resetCacheSerial();
      
      // reset the ACL menu
      resetACLMenu();
   }
   
   /** 
    * Receives the reference to the server admin exports.
    *
    * @param  exports
    *         instance of the remote admin exports to use
    */
   void initExports(AdminExports exports)
   {
      adm = exports;
   }
   
   /** 
    * Handles target --&gt; live choice.
    */
   void goLive()
   {
      // check if the same
      if (targetType == TARGET_TYPE_LIVE)
      {
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // notify the client about the going target
      if (targetType != TARGET_TYPE_NONE)
      {
         targetType = TARGET_TYPE_NONE;
         client.show(AdminClient.SCREEN_BLANK, "");
         client.notifyEvent(AdminNotify.NOTIFY_TARGET);
      }
      
      // reset the ACL menu
      resetACLMenu();
      
      // tell the server we're targeting the live directory
      cacheSerial = -1;
      int newCacheSerial = adm.setTargetLive(true);

      if (newCacheSerial == cacheSerial)
      {
         // idle
         AdminUtils.idle(client);
         
         client.alarm();
         menu.connectedState();
         return;
      }
      
      // reflect the new state
      client.ok();
      cacheSerial = newCacheSerial; 
      targetType = TARGET_TYPE_LIVE;
      menu.targetedState();
      
      // set the target description
      client.setTargetInfo(AdminClient.TARGET_TARGET, "live[" + 
                           Integer.toString(cacheSerial) + "]");

      // enable refresh action
      menu.setTargetRefresh(adm.canRefresh());
      
      // enable Console menu
      menu.setConsole(true);
      
      // notify the client about the new target
      client.show(AdminClient.SCREEN_BLANK, "");
      client.notifyEvent(AdminNotify.NOTIFY_TARGET);
      
      // create ACL menu items by resource type
      // 1. get array of resource plugins and acl instances
      String[] plugins = adm.listResourceNames();
      if (plugins == null)
      {
         // idle
         AdminUtils.idle(client);
         
         client.alarm();
         return;
      }
      
      String[] servers = adm.listAclInstances();
      if (servers == null)
      {
         // idle
         AdminUtils.idle(client);
         
         client.alarm();
         return;
      }
      
      // 2. Add menu items one per resource type
      JMenu acl = menu.getAccessControl();

      JMenuItem aclMenuByResource = acl.getItem(1);
      JMenuItem aclMenuByServer = acl.getItem(3);
      
      for (int i = 0; i < plugins.length; i++)
      {
         String rsn = adm.getResourceType(plugins[i]);
         aclMenuByResource.add(new JMenuItem(rsn));
      }

      // 3. Add menu items one per ACL instance
      for (int i = 0; i < servers.length; i++)
      {
         aclMenuByServer.add(new JMenuItem(servers[i]));
      }
      
      menu.getACL().init(client, acl);
      
      // idle
      AdminUtils.idle(client);
   }

   /**
    * Handles target --&gt; open choice.
    */
   void goOpen()
   {
      // notify the client about target change
      targetType = TARGET_TYPE_NONE;
      client.show(AdminClient.SCREEN_BLANK, "");
      client.notifyEvent(AdminNotify.NOTIFY_TARGET);
      
      // reset the ACL menu
      resetACLMenu();

      // not implemented yet
      client.beep("not implemented");
      menu.connectedState();
      
      client.setTargetInfo(AdminClient.TARGET_TARGET, "");
      
      // disable refresh action
      menu.setTargetRefresh(false);
   }
   
   /** 
    * Handles target --&gt; new choice.
    */
   void goNew()
   {
      // notify the client about target change
      targetType = TARGET_TYPE_NONE;
      client.show(AdminClient.SCREEN_BLANK, "");
      client.notifyEvent(AdminNotify.NOTIFY_TARGET);
      
      // reset the ACL menu
      resetACLMenu();

      // not implemented yet
      client.beep("not implemented");
      menu.connectedState();
      client.setTargetInfo(AdminClient.TARGET_TARGET, "");

      // disable refresh action
      menu.setTargetRefresh(false);
   }

   boolean canRefresh()
   {
      return adm.canRefresh();
   }
   
   /** 
    * Handles target --&gt; refresh action.
    */
   void refresh()
   {
      // busy
      AdminUtils.busy(client);
      
      // tell the server to refresh the live directory
      int newCacheSerial = adm.targetRefresh();

      if (newCacheSerial <= cacheSerial)
      {
         // idle
         AdminUtils.idle(client);
         
         // not refreshed
         client.alarm();
         return;
      }

      // reflect the new state
      client.ok();
      cacheSerial = newCacheSerial; 

      // set the target description
      client.setTargetInfo(AdminClient.TARGET_TARGET, "live[" + 
                           Integer.toString(cacheSerial) + "]");
      
      // idle
      AdminUtils.idle(client);
   }
   
   /** 
    * Resets cache serial number at logoff.
    */
   void resetCacheSerial()
   {
      cacheSerial = -1;
   }
   
   /** 
    * Resets modified ACL menu
    */
   private void resetACLMenu()
   {
      JMenu acl = menu.getAccessControl();

      JMenuItem aclMenuByResource = acl.getItem(1);
      JMenuItem aclMenuByServer = acl.getItem(3);

      aclMenuByResource.removeAll();
      aclMenuByServer.removeAll();
   }
}