/*
** Module   : AdminTerminateSessions.java
** Abstract : Terminate sessions screen.
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SVL 20111209 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.*;

import javax.swing.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * "Terminate sessions" screen.
 */
public class AdminTerminateSessions
{
   /** Main panel for dialog. */
   protected JPanel mainPanel;

   /** Related instance of AdminClient. */
   protected AdminClient client;

   /** Cancel button. */
   protected JButton btnCancel;

   /** Refresh sessions button. */
   protected JButton btnRefresh;

   /** Perform action and, optionally, forcibly terminate sessions button. */
   protected JButton btnTerminate;

   /** Sessions table. */
   protected JTable table;

   /** Session table model. */
   protected SessionInfoModel tableModel;

   /** Sessions table scroller. */
   protected JScrollPane tableScroller;

   /** Admin Server APIs. */
   protected AdminExports adm;

   /** Target JAR. */
   protected final String targetJar;

   /** Interface which defines target API. */
   protected final String apiInterfaceClass;

   /** Action which we are trying to perform. */
   protected final JarActionType actionType;

   /**
    * Creates the "terminate sessions" dialog.
    *
    * @param client
    *        Related instance of AdminClient.
    * @param adm
    *        Admin Server APIs.
    * @param actionType
    *        Action which we are trying to perform.
    * @param targetJar
    *        Target JAR.
    * @param apiInterfaceClass
    *        Interface which defines target API.
    */
   public AdminTerminateSessions(AdminClient client,
                                 AdminExports adm,
                                 JarActionType actionType,
                                 String targetJar,
                                 String apiInterfaceClass)
   {
      this.client = client;
      this.adm = adm;
      this.targetJar = targetJar;
      this.apiInterfaceClass = apiInterfaceClass;
      this.actionType = actionType;

      // initialize table
      tableModel = new SessionInfoModel();
      tableScroller = new JScrollPane();
      table = AdminUtils.createBorderedScrolledTable(
                      actionType.equals(JarActionType.API_DEREGISTER)  ?
                        "Sessions which have executed the selected API" :
                        "Sessions which have executed API(s) associated to the selected customer library",
                     tableModel, tableScroller);

      // set columns width
      int tableWidth = 0;
      for (int i = 0; i < tableModel.columnSizes.length; i++)
      {
         TableColumn column = table.getColumnModel().getColumn(i);
         int size = tableModel.columnSizes[i];
         column.setWidth(size);
         column.setPreferredWidth(size);
         tableWidth += size;
      }

      // set table width
      Dimension d = table.getPreferredScrollableViewportSize();
      table.setPreferredScrollableViewportSize(new Dimension(tableWidth, d.height));

      // create help text
      String replacement1 = null;
      String replacement2 = null;
      String replacement3 = null;
      String terminateCaption = null;

      switch (actionType)
      {
         case JAR_DEREGISTER:
            replacement1 = "API(s) associated to the selected customer library";
            replacement2 = "customer library";
            replacement3 = "deregistered";
            terminateCaption = "Deregister Customer Library";
            break;
         case JAR_RELOAD:
            replacement1 = "API(s) associated to the selected customer library";
            replacement2 = "customer library";
            replacement3 = "reloaded";
            terminateCaption = "Reload Customer Library";
            break;
         case API_DEREGISTER:
            replacement1 = "the selected API";
            replacement2 = "API";
            replacement3 = "deregistered";
            terminateCaption = "Deregister API";
            break;
      }

      StringBuilder sb = new StringBuilder();
      sb.append("The sessions displayed in this table have executed ");
      sb.append(replacement1);
      sb.append(".  The ");
      sb.append(replacement2);
      sb.append(" cannot be ");
      sb.append(replacement3);
      sb.append(" while these sessions are active.\n");
      sb.append("You can finish them manually or forcibly terminate them.\n");
      sb.append("If you decide to finish them manually, then after you finish them, you should refresh the session list,");
      sb.append("check that there are no more active sessions and use the \"");
      sb.append(terminateCaption);
      sb.append("\" button.\n");
      sb.append("If you decide to forcibly terminated them, just use the \"");
      sb.append(terminateCaption);
      sb.append("\" button, you will be promted whether you really want to terminate them.");

      JTextArea txtHelp = AdminUtils.createReadOnlyTextArea(sb.toString());

      // initialize control components
      btnCancel = new JButton("Cancel");
      btnRefresh = new JButton("Refresh");
      btnTerminate = new JButton(terminateCaption);

      JPanel controlPanel = AdminUtils.groupComponents(
                     new Component[][]{{btnCancel, btnRefresh, btnTerminate}});

      mainPanel = AdminUtils.groupComponents(
            new Component[][]{{txtHelp}, {tableScroller}, {controlPanel}});
   }

   /**
    * Display the "terminate sessions" button.
    *
    * @return <code>true</code> if the desired action was successfully
    *         performed.
    */
   public boolean runDialog()
   {
      if (!refreshSessions())
         return false;

      String dialogTitle = null;

      switch (actionType)
      {
         case JAR_DEREGISTER:
            dialogTitle = "Customer Library Deregistration";
            break;
         case JAR_RELOAD:
            dialogTitle = "Customer Library Reload";
            break;
         case API_DEREGISTER:
            dialogTitle = "API Deregistration";
            break;
      }

      final boolean[] result = new boolean[]{false};
      final JDialog dialog = AdminUtils.modalDialog(client,
                                                    dialogTitle,
                                                    mainPanel);

      btnCancel.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            dialog.setVisible(false);
         }
      });

      btnRefresh.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            refreshSessions();
         }
      });

      btnTerminate.addActionListener(new ActionListener()
      {
         @Override
         public void actionPerformed(ActionEvent e)
         {
            result[0] = handleTerminateSessions();
            if (result[0])
               dialog.setVisible(false);
         }
      });

      dialog.setVisible(true);
      dialog.dispose();
      return result[0];
   }

   /**
    * Refresh the list of active sessions. If the list cannot be refreshed
    * then an error message is displayed.
    *
    * @return <code>true</code> if the list was successfully refreshed.
    */
   protected boolean refreshSessions()
   {
      AdminUtils.busy(client);
      SessionInfo[] sessions = adm.getActiveSessions(targetJar,
                                                     apiInterfaceClass);
      AdminUtils.idle(client);

      if (sessions == null)
      {
         AdminUtils.errorDialog(client,
                                "Cannot load the list of sessions.",
                                "Error");
         return false;
      }
      else
      {
         tableModel.loadAll(sessions);
         return true;
      }
   }

   /**
    * Perform the desired action, optionally, forcibly terminating sessions.
    *
    * @return <code>true</code> if the action was successful.
    */
   protected boolean handleTerminateSessions()
   {
      boolean force = tableModel.size() > 0;

      if (force)
      {
         // confirm force termination of sessions
         StringBuilder sb = new StringBuilder();
         sb.append("Are you sure you want to ");
         switch (actionType)
         {
            case JAR_DEREGISTER:
               sb.append("deregister customer library");
               break;
            case JAR_RELOAD:
               sb.append("reload customer library");
               break;
            case API_DEREGISTER:
               sb.append("deregister API");
               break;
         }
         sb.append(" and forcibly terminate all related sessions?");

         String confirmationTitle = null;
         switch (actionType)
         {
            case JAR_DEREGISTER:
               confirmationTitle = "Force Deregister Customer Library";
               break;
            case JAR_RELOAD:
               confirmationTitle = "Force Reload Customer Library";
               break;
            case API_DEREGISTER:
               confirmationTitle = "Force Deregister API";
               break;
         }

         boolean res = AdminUtils.confirmDialog(client,
                                                confirmationTitle,
                                                sb.toString());
         if (!res)
            return false;
      }

      // perform desired action
      JarOpErrorCode result = null;
      switch (actionType)
      {
         case JAR_DEREGISTER:
            AdminUtils.busy(client);
            result = adm.deregisterJar(targetJar, force, true);
            AdminUtils.idle(client);
            break;
         case JAR_RELOAD:
            AdminUtils.busy(client);
            result = adm.deregisterJar(targetJar, force, false);
            AdminUtils.idle(client);
            break;
         case API_DEREGISTER:
            AdminUtils.busy(client);
            result = adm.deregisterJarAPI(targetJar,
                                          apiInterfaceClass,
                                          force);
            AdminUtils.idle(client);
            break;
      }

      if (!force && result == JarOpErrorCode.ACTIVE_SESSIONS_PRESENT)
      {
         // some new related sessions have appeared since the last time the
         // list of session has been refreshed, unable to perform desired
         // action in non-forcible way

         refreshSessions();

         StringBuilder sb = new StringBuilder();
         sb.append("Some new sessions have appeared. Unable to perform non-forcible\n");
         switch (actionType)
         {
            case JAR_DEREGISTER:
               sb.append("customer library deregistration");
               break;
            case JAR_RELOAD:
               sb.append("customer library reloading");
               break;
            case API_DEREGISTER:
               sb.append("API deregistration");
               break;
         }
         sb.append(". Session list has been refreshed.");
         JOptionPane.showMessageDialog(client,
               sb.toString(),
               "Operation Error",
               JOptionPane.ERROR_MESSAGE);
         return false;
      }
      else if (result != JarOpErrorCode.SUCCESS)
      {
         // an error has occurred
         String errorText = null;
         switch (actionType)
         {
            case JAR_DEREGISTER:
            case JAR_RELOAD:
               errorText = "Cannot deregister customer library.";
               break;
            case API_DEREGISTER:
               errorText = "Cannot deregister API.";
               break;
         }
         AdminJars.displayErrorMessage(client, errorText, result);
         return false;
      }
      else
      {
         // success
         return true;
      }
   }

   /**
    * Provides a table data model to represent session descriptors.
    */
   private class SessionInfoModel
   extends AbstractTableModel
   {
      /** Column names. */
      private final String[] columnNames =
      {
         "Account",
         "Description",
         "Type",
         "Session ID",
         "Remote Node",
         "Logon Timestamp",
      };

      /** Default column sizes. */
      private final int columnSizes[] =
      {
         75,
         100,
         40,
         75,
         175,
         200,
      };

      /** Column data types. */
      private final Class[] columnTypes =
      {
         String.class,
         String.class,
         String.class,
         Integer.class,
         String.class,
         String.class,
      };

      /**
       * Construct an instance.
       */
      public SessionInfoModel()
      {
      }

      /** Core data for the table. */
      private final ArrayList<SessionInfo> data = new ArrayList<SessionInfo>();

      /**
       * Gets the number of columns in the table model.
       *
       * @return   The number of columns.
       */
      public int getColumnCount()
      {
         return columnNames.length;
      }

      /**
       * Gets the number of rows in the table model.
       *
       * @return   The number of rows.
       */
      public int getRowCount()
      {
         return data.size();
      }

      /**
       * Gets the name of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column name.
       */
      public String getColumnName(int col)
      {
         return columnNames[col];
      }

      /**
       * Get the cell value to be displayed.
       *
       * @param    col
       *           The column index (0-based).
       * @param    row
       *           The row index (0-based).
       */
      public Object getValueAt(int row, int col)
      {
         Object      datum = null;
         SessionInfo info  = data.get(row);

         switch (col)
         {
            case 0:
               datum = info.name;
               break;
            case 1:
               datum = info.descr;
               break;
            case 2:
               datum = info.user ? "User" : "Process";
               break;
            case 3:
               datum = info.id;
               break;
            case 4:
               datum = info.remoteNet;
               break;
            case 5:
               datum = String.format("%1$Tm/%1$Td/%1$TY %1$TT %1$Tz",
                                     info.logonTime);
               break;
         }

         return datum;
      }

      /**
       * Get the number of data rows.
       *
       * @return See above.
       */
      public int size()
      {
         return data.size();
      }

      /**
       * Gets the data type of a specific column.
       *
       * @param    col
       *           The column index (0-based).
       *
       * @return   The column type.
       */
      public Class getColumnClass(int col)
      {
         return columnTypes[col];
      }

      /**
       * Clears and then reloads the table from a data source.
       *
       * @param    list
       *           array of session descriptors to load into the model
       */
      public void loadAll(SessionInfo[] list)
      {
         // clear first
         int rows = data.size();
         if (rows > 0)
         {
            // notify listeners
            fireTableRowsDeleted(0, rows - 1);

            // clear the storage
            data.clear();
         }

         // stuff the table next
         data.addAll(Arrays.asList(list));

         // notify listeners about the data change
         fireTableDataChanged();
      }
   }

   /**
    * Types of actions supported by this screen.
    */
   public enum JarActionType
   {
      JAR_DEREGISTER, JAR_RELOAD, API_DEREGISTER
   }
}