/*
** Module   : AdminUserAccounts.java
** Abstract : User Accounts management client
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090522  @42450  Created initial version.
** 002 CA  20090529  @42493  Clone dialog provides possibility to define new 
**                           password or to clone the password. Added changes
**                           related to client account extension plugin.
** 003 NVS 20090602  @42550  Client extension plugin is gotten according to
**                           the new AdminDef.
** 004 CA  20090605  @42609  Added 'account is disabled' check-box. Added a
**                           combo with explicit account mode selection. Added
**                           check-boxes related to password management. Moved
**                           custom plugin load to AdminTarget.
** 005 NVS 20090608  @42613  Fixed 'account disabled' check-box reading yielding
**                           the inverse meaning.
** 006 CA  20090609  @42616  Added Switch to ACLs button. Changed layout for 
**                           some widgets.
** 007 CA  20090610  @42629  If user has no password specified, but the protect
**                           flag is off, select the 'no password protection' 
**                           check-box.
** 008 CA  20090611  @42656  Fixed a NPE when no extension plugin is set.
** 009 CA  20090612  @42665  The color for 'Disabled Account' is set to red 
**                           when checked. Removed the 'hash with all zero 
**                           means no password' check in edit user.
** 010 CA  20090720  @43247  Added code for the Switch to ACLs button.
** 011 NVS 20090723  @43355  Tables don't allow moving columns.
** 012 NVS 20090729  @43433  Implemented users printing.
** 013 NVS 20090803  @43487  Using a better approach to printing with extended
**                           detail choice.
** 014 NVS 20090804  @43515  First pass of detailed printing implementation.
** 015 NVS 20090804  @43525  Plugin initializes as soon as this instance.
**                           Extended view printing is implemented.
** 016 NVS 20090805  @43532  Added busy/idle feedback to printing methods.
** 017 NVS 20090807  @43553  Refactored in support of custom menu extensions.
** 018 NVS 20091007  @44116  Changed to comply with the new ClientAccountExt.
**                           Changes allow the extension fields in the add/edit
**                           user dialog to be enabled or disabled according to
**                           granular admin permissions.
** 019 SIY 20091116   @44340 Refactoring of things related to TaggedName. 
**                           Minor cleanups.
** 020 SIY 20100118   @44531 Fixed incorrect message for case when extension
**                           plugin throws an exception.
** 021 SVL 20100211   @44577 Added user search capabilities.
** 022 SVL 20100916          Added authPlugin combo. Auth-related elements are
**                           grouped using AuthModeComboBoxes.
** 023 SVL 20100920          Changed to conform new AuthModeComboBoxes API.
** 024 SVL 20110619          During user cloning user is considered to have a
**                           password if user.protected is true (earlier this
**                           decision was based on the user auth mode).
** 025 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.io.*;
import java.util.*;
import java.util.List;
import java.awt.*;
import java.awt.event.*;
import java.text.MessageFormat;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.*;

import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.security.*;

/**
 * This is a User Accounts management class for the admin session. 
 * 
 * @author NVS
 */
public class AdminUserAccounts
extends AdminScreen
implements IAccountScreen, PrintRequest
{
   /** id of the user-name widget in user-def dialog */
   public static final String WIDGET_NAME = "NAME";
   
   /** id of the person widget in user-def dialog */
   public static final String WIDGET_PERSON = "PERSON";
   
   /** id of the certificate combo-box in user-def dialog */
   public static final String WIDGET_CERTIFICATE = "CERTIFICATE";
   
   /** id of the no password authentication check-box in user-def dialog */
   public static final String WIDGET_NO_PW_AUTH = "NO_PASSWORD_AUTH";
   
   /** id of the enter password check-box in user-def dialog */
   public static final String WIDGET_ENTER_PW = "ENTER_PASSWORD";
   
   /** id of the leave password unchanged check-box in user-def dialog */
   public static final String WIDGET_PW_UNCHAGED = "PASSWORD_UNCHAGED";

   /** id of the password widget in user-def dialog */
   public static final String WIDGET_PASSWORD = "PASSWORD";
   
   /** id of the confirm password widget in user-def dialog */
   public static final String WIDGET_CONFIRM_PASSWORD = "CONFIRM_PASSWORD";

   /** id of the authentication mode combo in user-def dialog */
   public static final String WIDGET_AUTH_COMBO = "CUSTOM_AUTH";

   /** id of the authentication plugin combo in user-def dialog */
   public static final String WIDGET_AUTH_PLUGIN = "AUTH_PLUGIN";
   
   /** id of the disabled account check-box in user-def dialog */
   public static final String WIDGET_ACC_DISABLED = "ACC_DISABLED";
   
   /** Screen switcher - user accounts screen ID */
   public static final String SCREEN_ACC_USERS = "adm_acc_users";
   
   /** Screen title */
   private static final String TITLE_ACC_USERS = "User Accounts Management";
   
   // west panel widgets, all named wp*
   
   /** west panel - works with the list of users */
   private JPanel west = null;
   
   /** scroller for list of users */
   private JScrollPane wpScroller = null;
   
   /** table to display all users */
   private JTable wpUsers = null;

   /** related table model */
   private TaggedDataTableModel wpUtm = null;

   /** related table selection model */
   private ListSelectionModel wpUsm = null;

   /** sub-panel that holds all controls for list of users */
   private JPanel wpCtlPanel = null;
   
   /** Refresh button */
   private JButton wpbR = null;

   /** Add User button */
   private JButton wpbA = null;
   
   /** Edit User button */
   private JButton wpbE = null;

   /** Delete Selected button */
   private JButton wpbD = null;

   /** Clone User button */
   private JButton wpbC = null;

   /** Find User button */
   private JButton wpbF = null;
   
   /** Disable User button */
   private JButton wpbP = null;
   
   /** Switch to ACL's button */
   private JButton wpbS = null;
   
   // center panel widgets, all named cp* 

   /** center panel - works with the selected user */
   private JPanel center = null;
   
   /** scroller for list of assigned groups */
   private JScrollPane cpScroller = null;

   /** table to display assigned groups */
   private JTable cpGroups = null;

   /** related table model */
   private TaggedDataTableModel cpGtm = null;

   /** related table selection model */
   private ListSelectionModel cpGsm = null;
   
   /** Name field label */
   @SuppressWarnings("unused")
   private JLabel cplN = null;
   
   /** Name field */
   private JTextField cpfN = null;
   
   /** Add Group button */
   private JButton cpbA = null;
   
   /** Delete Selected button */
   private JButton cpbD = null;
   
   /** sub-panel that holds all controls for list of assigned groups */
   private JPanel cpCtlPanel = null;

   /** pane that holds all controls for Add group */
   private JPanel cpAddPanel = null;
   
   // east panel widgets, all named ep*
   
   /** east panel - works with the list of all groups */
   private JPanel east = null;

   /** scroller for list of all groups */
   private JScrollPane epScroller = null;
   
   /** scroller's titled border */
   private TitledBorder epScrollTitle = null; 
 
   /** table to display groups */
   private JTable epGroups = null;

   /** related table model */
   private TaggedDataTableModel epGtm = null;

   /** related table selection model */
   private ListSelectionModel epGsm = null;
   
   /** sub-panel that holds all controls for list of all groups */
   private JPanel epCtlPanel = null;

   /** sub-panel that holds Search related controls */
   private JPanel epSearchPanel = null;
   
   /** Populate button */
   private JButton epbP = null;

   /** Add Selected button */
   private JButton epbA = null;
   
   /** search argument field */
   private JTextField epfS = null;
   
   /** Filter button */
   private JButton epbF = null;
   
   /** All users table sorter */
   private TableRowSorter<TaggedDataTableModel> epSorter = null;    
   
   // dialog for clone user account specific widgets, all named wpcd*

   /** Name field label */
   private JLabel wpcdlN = null;
   
   /** Name field */
   private JTextField wpcdfN = null;
   
   /** With ACLs label */
   private JLabel wpcdlW = null;

   /** password panel */
   private JPanel wpcdpnlPw = null;
   
   /** Clone password radio */
   private JRadioButton wpcdrCp = null;
   
   /** Input password radio */
   private JRadioButton wpcdrNp = null;
   
   /** New password label. */
   private JLabel wpcdlP1 = null;

   /** New password field */
   private JPasswordField wpcdfP1 = null;
   
   /** Confirm password label */
   private JLabel wpcdlP2 = null;
   
   /** Confirm password field */
   private JPasswordField wpcdfP2 = null;

   /** With ACLs check box */
   private JCheckBox wpcdcW = null;
   
   /** Clone button in modal dialog*/
   private JButton wpcdbG = null;
   
   /** modal dialog panel for Clone As function */
   private JPanel wpcdClonePanel = null;

   // dialog for user account specific widgets, all named wpud*
   
   /** modal dialog panel for User Definition function */
   private JPanel wpudpUserPanel = null;

   /** user account name label */
   private JLabel wpudlN = null;
   
   /** user account name field */
   private JTextField wpudfN = null;

   /** Person label */
   private JLabel wpudlP = null;
   
   /** Person field */
   private JTextField wpudfP = null;
   
   /** certificate alias label */
   private JLabel wpudlC = null;
   
   /** certificate combo-box */
   private JComboBox wpudcbC = null;

   /** certificate combo-box model */
   private TaggedDataComboModel wpudcbCm = null;
   
   /** leave password unchaged check-box */
   private JRadioButton wpudchkPWUnchaged = null;
   
   /** enter password check-box */
   private JRadioButton wpudchkEnterPW = null;
   
   /** no password auth check-box */
   private JRadioButton wpudchkNoPW = null;
   
   /** Password label. */
   private JLabel wpudlP1 = null;

   /** Password field */
   private JPasswordField wpudfP1 = null;
   
   /** Confirm password label */
   private JLabel wpudlP2 = null;
   
   /** Confirm password field */
   private JPasswordField wpudfP2 = null;

   /** Auth mode and auth plugin combo-boxes. */
   private AuthModeComboBoxes authElements = null;
   
   /** account disabled label */
   private JLabel wpudlD = null;
   
   /** account disabled check-box */
   private JCheckBox wpudchkD = null;
   
   /** Cancel button */
   private JButton wpudbC = null;
   
   /** Save button */
   private JButton wpudbS = null;

   /** Default input field for search requests. */
   private JTextField findUserDefInput = null;

   /** Panel for search elements. */
   private JPanel findUserAccountPanel = null;

   /** "Find" button. */
   private JButton findUserFindBtn = null;

   /** "Cancel" button for find screen. */
   private JButton findUserCancelBtn = null;

   /** Mapping of user name to row number (in the table). */
   private final Map<String, Integer> userSearchIndex =
                                             new HashMap<String, Integer>();
   
   /** state definition - initial empty */
   private final static boolean[] STATE_INITIAL_EMPTY =
   {
   /* west,                  00*/       true,
   /*    wpScroller,         01*/       true,
   /*       wpUsers,         02*/       true,
   /*    wpCtlPanel,         03*/       true,
   /*       wpbR,            04*/       true,
   /*       wpbA,            05*/       true,
   /*       wpbD,            06*/ false,
   /*       wpbC,            07*/ false,
   /*       wpbE,            08*/ false,
   /*       wpbP,            09*/ false,
   /*       wpbS,            10*/ false,
   /* center,                11*/ false,
   /*    cpScroller,         12*/ false,
   /*       cpGroups,        13*/ false,
   /*    cpCtlPanel,         14*/ false,
   /*       cpfN,            15*/ false,
   /*       cpbA,            16*/ false,
   /*       cpbD,            17*/ false,
   /*  east,                 18*/       true,
   /*     epScroller,        19*/       true,
   /*        epGroups,       20*/       true,
   /*     epCtlPanel,        21*/       true,
   /*        epbP,           22*/       true,
   /*        epbA,           23*/ false,
   /*        epSearchPanel,  24*/ false,
   /*           epfS,        25*/ false,
   /*           epbF         26*/ false
   };

   /** indices of various controls */
   private final static int I_WPBR = 4;
   private final static int I_WPBA = 5;
   private final static int I_WPBD = 6;
   private final static int I_WPBC = 7;
   private final static int I_WPBE = 8;
   private final static int I_WPBP = 9;
   private final static int I_WPBS = 10;
   private final static int I_CPFN = 15;
   private final static int I_CPBA = 16;
   private final static int I_CPBD = 17;
   private final static int I_EPBP = 22;
   private final static int I_EPBA = 23;
   private final static int I_EPBF = 26;
   
   /** state definition - single selection in users list */
   private final static boolean[] STATE_USERS_SINGLE =
   {
   /* west,                    */       true,
   /*    wpScroller,           */       true,
   /*       wpUsers,           */       true,
   /*    wpCtlPanel,           */       true,
   /*       wpbR,              */       true,
   /*       wpbA,              */       true,
   /*       wpbD,              */       true,
   /*       wpbC,              */       true,
   /*       wpbE,              */       true,
   /*       wpbP,              */       true,
   /*       wpbS,              */       true,
   /* center,                  */       true,
   /*    cpScroller,           */       true,
   /*       cpGroups,          */       true,
   /*    cpCtlPanel,           */       true,
   /*       cpfN,              */       true,
   /*       cpbA,              */       true,
   /*       cpbD,              */ false,
   /*  east,                   */       true,
   /*     epScroller,          */ false,
   /*        epGroups,         */ false,
   /*     epCtlPanel,          */       true,
   /*        epbP,             */       true,
   /*        epbA,             */ false,
   /*        epSearchPanel,    */ false,
   /*           epfS,          */ false,
   /*           epbF           */ false
   };
   
   /** state definition - multiple selection in users list */
   private final static boolean[] STATE_USERS_MULTI =
   {
   /* west,                    */       true,
   /*    wpScroller,           */       true,
   /*       wpUsers,           */       true,
   /*    wpCtlPanel,           */       true,
   /*       wpbR,              */       true,
   /*       wpbA,              */       true,
   /*       wpbD,              */       true,
   /*       wpbC,              */ false,
   /*       wpbE,              */ false,
   /*       wpbP,              */ false,
   /*       wpbS,              */ false,
   /* center,                  */ false,
   /*    cpScroller,           */ false,
   /*       cpGroups,          */ false,
   /*    cpCtlPanel,           */ false,
   /*       cpfN,              */ false,
   /*       cpbA,              */ false,
   /*       cpbD,              */ false,
   /*  east,                   */       true,
   /*     epScroller,          */ false,
   /*        epGroups,         */ false,
   /*     epCtlPanel,          */       true,
   /*        epbP,             */       true,
   /*        epbA,             */ false,
   /*        epSearchPanel,    */ false,
   /*           epfS,          */ false,
   /*           epbF           */ false
   };

   /** state definition - all groups list populated */
   private final static boolean[] STATE_ALL_GROUPS =
   {
   /* west,                    */ false,
   /*    wpScroller,           */ false,
   /*       wpUsers,           */ false,
   /*    wpCtlPanel,           */ false,
   /*       wpbR,              */ false,
   /*       wpbA,              */ false,
   /*       wpbD,              */ false,
   /*       wpbC,              */ false,
   /*       wpbE,              */ false,
   /*       wpbP,              */ false,
   /*       wpbS,              */ false,
   /* center,                  */ false,
   /*    cpScroller,           */ false,
   /*       cpGroups,          */ false,
   /*    cpCtlPanel,           */ false,
   /*       cpfN,              */ false,
   /*       cpbA,              */ false,
   /*       cpbD,              */ false,
   /*  east,                   */       true,
   /*     epScroller,          */       true,
   /*        epGroups,         */       true,
   /*     epCtlPanel,          */       true,
   /*        epbP,             */       true,
   /*        epbA,             */ false,
   /*        epSearchPanel,    */       true,
   /*           epfS,          */       true,
   /*           epbF           */       true,
   };
   
   /** mask definition - west-center */
   private final static boolean[] MASK_WEST_CENTER =
   {
   /* west,                    */ true,
   /*    wpScroller,           */ true,
   /*       wpUsers,           */ true,
   /*    wpCtlPanel,           */ true,
   /*       wpbR,              */ true,
   /*       wpbA,              */ true,
   /*       wpbD,              */ true,
   /*       wpbC,              */ true,
   /*       wpbE,              */ true,
   /*       wpbP,              */ true,
   /*       wpbS,              */ true,
   /* center,                  */ true,
   /*    cpScroller,           */ true,
   /*       cpGroups,          */ true,
   /*    cpCtlPanel,           */ true,
   /*       cpfN,              */ true,
   /*       cpbA,              */ true,
   /*       cpbD,              */ true,
   /*  east,                   */ false,
   /*     epScroller,          */ false,
   /*        epGroups,         */ false,
   /*     epCtlPanel,          */ false,
   /*        epbP,             */ false,
   /*        epbA,             */ false,
   /*        epSearchPanel,    */ false,
   /*           epfS,          */ false,
   /*           epbF           */ false
   };

   /** mask definition - east */
   private final static boolean[] MASK_EAST =
   {
   /* west,                    */ false,
   /*    wpScroller,           */ false,
   /*       wpUsers,           */ false,
   /*    wpCtlPanel,           */ false,
   /*       wpbR,              */ false,
   /*       wpbA,              */ false,
   /*       wpbD,              */ false,
   /*       wpbC,              */ false,
   /*       wpbE,              */ false,
   /*       wpbP,              */ false,
   /*       wpbS,              */ false,
   /* center,                  */ false,
   /*    cpScroller,           */ false,
   /*       cpGroups,          */ false,
   /*    cpCtlPanel,           */ false,
   /*       cpfN,              */ false,
   /*       cpbA,              */ false,
   /*       cpbD,              */ false,
   /*  east,                   */ true,
   /*     epScroller,          */ true,
   /*        epGroups,         */ true,
   /*     epCtlPanel,          */ true,
   /*        epbP,             */ true,
   /*        epbA,             */ true,
   /*        epSearchPanel,    */ true,
   /*           epfS,          */ true,
   /*           epbF           */ true
   };
   
   /** size of the state array */
   private final static int STATE_SIZE = STATE_INITIAL_EMPTY.length;
   
   /** state definition - adjusted initial empty */                
   private boolean[] state_initial_empty = null;
      
   /** state definition - adjusted users single */                
   private boolean[] state_users_single = null;
   
   /** state definition - adjusted users multiple */                
   private boolean[] state_users_multi = null;
   
   /** state definition - adjusted all groups */                
   private boolean[] state_all_groups = null;
   
   /** indicates single user selection state */
   private boolean singleUser = false;

   /** name of the selected user for single user selection */
   private String userName = null;

   /** recursion prevention for all users list */
   private boolean selecting = false;

   /** the customer-specific extension plugin */
   private ClientAccountExtension extPlugin = null;

   /** container for user-extension fields */
   private Container extContainer = null;
   
   /** map of default user dialog widgets, kept by their name */
   private HashMap<String, Component> userDefWidgets = null; 
   
   /** print: sortable table model */
   private TableModel backTm = null;
   
   /** print: sortable table */
   private JTable backTable = null;
   
   /** print: extended table model */
   private TableModel frontTm = null;
   
   /** print: extended table */
   private JTable frontTable = null;
   
   /** print header */
   private MessageFormat header = null;
   
   /** print footer */
   private MessageFormat footer = null;
   
   /** permission to enumerate */
   private boolean canEnum = false;
   
   /** permission to change password */
   private boolean canPw = false;
   
   /** permission to group */
   private boolean canGroup = false;

   /** permission to read */
   private boolean canRead = false;

   /** permission to write */
   private boolean canWrite = false;

   /** permission to create */
   private boolean canCreate = false;

   /** permission to delete */
   private boolean canDelete = false;
   
   /** permission to switch to acls */
   private boolean canSwitch = false;

   /**
    * Constructor.
    * 
    * @param    client
    *           instance of AdminClient
    */
   AdminUserAccounts(AdminClient client)
   {
      // store useful references
      this.client = client;
      adm = client.getExports();
      
      // query the permissions profile
      AdminDef.Profile[] prof = adm.getProfile("/accounts/users", new int[]
      {
         AdminRights.ADLU_ENUMERATE_ACCESS,
         AdminRights.ADLU_PASSWORD_ACCESS,
         AdminRights.ADLU_GROUP_ACCESS,
         AdminRights.ADLU_READ_ACCESS,
         AdminRights.ADLU_WRITE_ACCESS,
         AdminRights.ADLU_CREATE_ACCESS,
         AdminRights.ADLU_DELETE_ACCESS
      });
      
      canEnum   = prof[0].permission;
      canPw     = prof[1].permission;
      canGroup  = prof[2].permission;
      canRead   = prof[3].permission;
      canWrite  = prof[4].permission;
      canCreate = prof[5].permission;
      canDelete = prof[6].permission;
      
      prof = adm.getProfile(new String[] {"/access_control"});
      
      canSwitch = prof[0].permission;
      
      // adjust the state machines states according to the profile
      state_initial_empty = Arrays.copyOf(STATE_INITIAL_EMPTY, STATE_SIZE);
      state_users_single = Arrays.copyOf(STATE_USERS_SINGLE, STATE_SIZE);
      state_users_multi = Arrays.copyOf(STATE_USERS_MULTI, STATE_SIZE);
      state_all_groups = Arrays.copyOf(STATE_ALL_GROUPS, STATE_SIZE);
  
      if (!canEnum)
      {
         // no permission to enumerate
         // refresh button disabled
         state_initial_empty[I_WPBR] = false;
         state_users_single [I_WPBR] = false;
         state_users_multi  [I_WPBR] = false;
         state_all_groups   [I_WPBR] = false;
      }
      
      if (!canGroup)
      {
         // no permission to group users
         // center and east panel buttons disabled
         state_initial_empty[I_CPFN] = false;
         state_initial_empty[I_CPBA] = false;
         state_initial_empty[I_CPBD] = false;
         state_initial_empty[I_EPBP] = false;
         state_initial_empty[I_EPBA] = false;
         state_initial_empty[I_EPBF] = false;
         state_users_single [I_CPFN] = false;
         state_users_single [I_CPBA] = false;
         state_users_single [I_CPBD] = false;
         state_users_single [I_EPBP] = false;
         state_users_single [I_EPBA] = false;
         state_users_single [I_EPBF] = false;
         state_users_multi  [I_CPFN] = false;
         state_users_multi  [I_CPBA] = false;
         state_users_multi  [I_CPBD] = false;
         state_users_multi  [I_EPBP] = false;
         state_users_multi  [I_EPBA] = false;
         state_users_multi  [I_EPBF] = false;
         state_all_groups   [I_CPFN] = false;
         state_all_groups   [I_CPBA] = false;
         state_all_groups   [I_CPBD] = false;
         state_all_groups   [I_EPBP] = false;
         state_all_groups   [I_EPBA] = false;
         state_all_groups   [I_EPBF] = false;
      }
      
      if (!canRead && !canPw && !canWrite)
      {
         // no permission to see user account details
         // Edit User button disabled
         state_initial_empty[I_WPBE] = false;
         state_users_single [I_WPBE] = false;
         state_users_multi  [I_WPBE] = false;
         state_all_groups   [I_WPBE] = false;
      }

      if (!canWrite)
      {
         // no permission to write to user account
         // Disable button disabled
         state_initial_empty[I_WPBP] = false;
         state_users_single [I_WPBP] = false;
         state_users_multi  [I_WPBP] = false;
         state_all_groups   [I_WPBP] = false;
      }

      if (!canCreate)
      {
         // no permission to create users
         // Clone and Add buttons disabled
         state_initial_empty[I_WPBA] = false;
         state_initial_empty[I_WPBC] = false;
         state_users_single [I_WPBA] = false;
         state_users_single [I_WPBC] = false;
         state_users_multi  [I_WPBA] = false;
         state_users_multi  [I_WPBC] = false;
         state_all_groups   [I_WPBA] = false;
         state_all_groups   [I_WPBC] = false;
      }
      
      if (!canDelete)
      {
         // no permission to create users
         // Delete button disabled
         state_initial_empty[I_WPBD] = false;
         state_users_single [I_WPBD] = false;
         state_users_multi  [I_WPBD] = false;
         state_all_groups   [I_WPBD] = false;
      }
      
      if (!canSwitch)
      {
         // no permission to use ACL screen
         // Switch to ACLs button disabled
         state_initial_empty[I_WPBS] = false;
         state_users_single [I_WPBS] = false;
         state_users_multi  [I_WPBS] = false;
         state_all_groups   [I_WPBS] = false;
      }
      
      // initialize the extensions plugin
      extPlugin = client.getAccountExtPlugin();
      if (extPlugin != null)
      {
         extPlugin.initialize(this);
      }

      // create the main screen and add it to the swapper
      mainPanel = new JPanel();
      mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.X_AXIS));
      client.add(mainPanel, SCREEN_ACC_USERS, this);

         GroupLayout mainLayout = new GroupLayout(mainPanel);
      mainPanel.setLayout(mainLayout);
      mainLayout.setAutoCreateGaps(true);
      mainLayout.setAutoCreateContainerGaps(true);

      // top panel with group layout manages
      // west, center and east sub-panels
      //  +-------screen---------------+
      //  |+------+  +------+  +------+|
      //  ||      |  |      |  |      ||
      //  ||      |  |      |  |      ||
      //  || west |  |center|  | east ||
      //  ||      |  |      |  |      ||
      //  |+------+  +------+  +------+|
      //  +----------------------------+
      
      west = new JPanel();
      center = new JPanel();
      east = new JPanel();

      // Create a sequential group for the horizontal axis:
      {
         GroupLayout.SequentialGroup hGroup = 
            mainLayout.createSequentialGroup();
         
         // three panels 
         hGroup.
            addComponent(west).
            addComponent(center).
            addComponent(east)
         ;
         
         mainLayout.setHorizontalGroup(hGroup);
      }
      
      // Create a parallel group for the vertical axis
      {
         GroupLayout.ParallelGroup vGroup = 
            mainLayout.createParallelGroup(GroupLayout.Alignment.BASELINE);
            
         vGroup.
            addComponent(west).
            addComponent(center).
            addComponent(east)
         ;
            
         mainLayout.setVerticalGroup(vGroup);
      }

      // stuffing west panel with components:
      // west panel is controlled with BorderLayout
      west.setLayout(new BorderLayout());
      
      //  +---------west-----------+
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    wpScroller      | |
      //  | |      CENTER        | |
      //  | |                    | |
      //  | |                    | |
      //  | +--------------------+ |
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    wpCtlPanel      | |
      //  | |      SOUTH         | |
      //  | +--------------------+ |
      //  +------------------------+
      
      wpUtm = new TaggedDataTableModel("User ID", "Person");
      wpUsers = new JTable(wpUtm);
      wpUsers.getTableHeader().setReorderingAllowed(false);
      wpUsm = wpUsers.getSelectionModel();
      
      wpScroller = new JScrollPane();
      wpScroller.getViewport().add(wpUsers);
      wpScroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Defined Users "));
      
      wpScroller.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      west.add(wpScroller, BorderLayout.CENTER);
      
      // wpCtlPanel panel is controlled with GroupLayout
      //  +------wpCtlPanel--------+
      //  |                        |
      //  |                        |
      //  |     buttons here       |   
      //  |                        |
      //  |                        |
      //  +------------------------+
      
      // creating control widgets
      wpbR = new JButton("Refresh");
      wpbA = new JButton("Add User");
      wpbD = new JButton("Delete Selected");
      wpbC = new JButton("Clone User");
      wpbF = new JButton("Find User");
      wpbE = new JButton(canWrite | canPw ? "Edit User" : "Details");
      wpbP = new JButton("Disable User");
      wpbS = new JButton("Switch to ACLs");

      wpCtlPanel = AdminUtils.groupComponents(new Component[][] 
         { 
            { wpbR, wpbA, wpbD },
            { wpbC, wpbE, wpbP },
            { wpbF, new JLabel(), wpbS }
         },
         true
      );
      west.add(wpCtlPanel, BorderLayout.SOUTH);

      // stuffing center panel with components
      // center panel is controlled with BorderLayout
      center.setLayout(new BorderLayout());

      //  +--------center----------+
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    cpScroller      | |
      //  | |      CENTER        | |
      //  | |                    | |
      //  | |                    | |
      //  | +--------------------+ |
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    cpCtlPanel      | |
      //  | |      SOUTH         | |
      //  | +--------------------+ |
      //  +------------------------+
      
      cpGtm = new TaggedDataTableModel("Group ID", "Description");
      cpGroups = new JTable(cpGtm);
      cpGroups.getTableHeader().setReorderingAllowed(false);
      cpGsm = cpGroups.getSelectionModel();

      cpScroller = new JScrollPane();
      cpScroller.getViewport().add(cpGroups);
      cpScroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Assigned Groups "));
      cpScroller.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      center.add(cpScroller, BorderLayout.CENTER);

      cpCtlPanel = new JPanel();
      center.add(cpCtlPanel, BorderLayout.SOUTH);
      
      // create control widgets
      cpfN = new JTextField(12);
      cpbA = new JButton("Add Group");
      cpbD = new JButton("Delete Selected");
      
      // create "Add Group" pane
      cpAddPanel = createGroup(" Add Group ", cpfN, cpbA);
      
      // create layout of the cpCtlPanel panel with all controls
      GroupLayout cpCtlLayout = new GroupLayout(cpCtlPanel);
      cpCtlPanel.setLayout(cpCtlLayout);
    
      cpCtlLayout.setAutoCreateGaps(true);
      cpCtlLayout.setAutoCreateContainerGaps(true);
  
      // Create a parallel group for the horizontal axis
      {
         GroupLayout.ParallelGroup hGroup = 
            cpCtlLayout.createParallelGroup(GroupLayout.Alignment.TRAILING);
       
         hGroup.
            // Delete Selected button
            addComponent(cpbD).
            // Add group pane
            addComponent(cpAddPanel)
            ;
         cpCtlLayout.setHorizontalGroup(hGroup);
      }
    
      // Create a sequential group for the vertical axis:
      {
         GroupLayout.SequentialGroup vGroup = 
            cpCtlLayout.createSequentialGroup();
       
         vGroup.
            // Delete Selected button
            addComponent(cpbD).
            // Add group pane
            addComponent(cpAddPanel)
            ;
         cpCtlLayout.setVerticalGroup(vGroup);
      }

      // stuffing east panel with components
      // east panel is controlled with BorderLayout
      east.setLayout(new BorderLayout());
      
      //  +---------east-----------+
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    epScroller      | |
      //  | |      CENTER        | |
      //  | |                    | |
      //  | |                    | |
      //  | +--------------------+ |
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    epCtlPanel      | |
      //  | |      SOUTH         | |
      //  | +--------------------+ |
      //  +------------------------+
      
      epGtm = new TaggedDataTableModel("Group ID", "Description");
      epGroups = new JTable(epGtm);
      epGroups.getTableHeader().setReorderingAllowed(false);
      epGsm = epGroups.getSelectionModel();
      epSorter = new TableRowSorter<TaggedDataTableModel>(epGtm);
      epSorter.setRowFilter(null);
      epGroups.setRowSorter(epSorter);
      
      epScroller = new JScrollPane();
      epScroller.getViewport().add(epGroups);
      epScrollTitle = new TitledBorder(new EtchedBorder(EtchedBorder.RAISED), 
                                       " All Groups ");
      epScroller.setBorder(epScrollTitle);
      epScroller.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      east.add(epScroller, BorderLayout.CENTER);
      
      epCtlPanel = new JPanel();
      east.add(epCtlPanel, BorderLayout.SOUTH);

      // create control widgets
      epbP = new JButton("Populate");
      epbA = new JButton("Add Selected");
      
      epfS = new JTextField(12);
      epbF = new JButton("Filter");

      // create upper pane
      JPanel epUpPane = new JPanel();
      epUpPane.setLayout(new BoxLayout(epUpPane, BoxLayout.LINE_AXIS));
      epUpPane.add(epbP);
      epUpPane.add(Box.createHorizontalGlue());
      epUpPane.add(epbA);
      
      // create "Search" pane
      epSearchPanel = createGroup(" Search ", epfS, epbF);

      // create layout of the epCtlPanel panel with all controls:
      // buttons and the "search" sub-panel
      GroupLayout epCtlLayout = new GroupLayout(epCtlPanel);
      epCtlPanel.setLayout(epCtlLayout);
    
      epCtlLayout.setAutoCreateGaps(true);
      epCtlLayout.setAutoCreateContainerGaps(true);
    
      // Create a parallel group for the horizontal axis
      {
         GroupLayout.ParallelGroup hGroup = 
            epCtlLayout.createParallelGroup();
       
         hGroup.
            addComponent(epUpPane).
            addComponent(epSearchPanel)
            ;
         epCtlLayout.setHorizontalGroup(hGroup);
      }
      
      // Create a sequential group for the vertical axis
      {
         GroupLayout.SequentialGroup vGroup = 
            epCtlLayout.createSequentialGroup();
       
         vGroup.
            addComponent(epUpPane).
            addComponent(epSearchPanel)
            ;
         epCtlLayout.setVerticalGroup(vGroup);
      }
      
      // create "Clone as" modal dialog panel with Group layout
      wpcdClonePanel = new JPanel();
      
      wpcdlN = new JLabel(" New Name ");
      wpcdfN = new JTextField(12);
      
      wpcdrCp = new JRadioButton(" Clone Password ");      
      wpcdrNp = new JRadioButton(" Input New Password ");
      ButtonGroup bg = new ButtonGroup();
      bg.add(wpcdrCp);
      bg.add(wpcdrNp);
      
      wpcdlP1 = new JLabel(" New Password ");
      wpcdlP2 = new JLabel(" Confirm Password ");
      
      wpcdfP1 = new JPasswordField(12);
      wpcdfP2 = new JPasswordField(12);
      
      // put all widgets related to password in a titled panel
      wpcdpnlPw = AdminUtils.groupComponents(
            new Component[][]
                {
                  { wpcdrCp, wpcdrNp },
                  { wpcdlP1, wpcdfP1 },
                  { wpcdlP2, wpcdfP2 }
                }
            ,
            " Clone Password ? "
      );

      wpcdlW = new JLabel(" With ACLs ");
      wpcdcW = new JCheckBox();
      wpcdbG = new JButton("Clone User");

      wpcdClonePanel.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), 
                           " Clone Selected User As "));
      GroupLayout wpcdCloneLayout = new GroupLayout(wpcdClonePanel);
      wpcdClonePanel.setLayout(wpcdCloneLayout);
    
      wpcdCloneLayout.setAutoCreateGaps(true);
      wpcdCloneLayout.setAutoCreateContainerGaps(true);
    
      // Create a parallel group for the horizontal axis.
      {
         GroupLayout.ParallelGroup hGroup = 
            wpcdCloneLayout.createParallelGroup();
       
         hGroup.addGroup(wpcdCloneLayout.createSequentialGroup().
               addComponent(wpcdlN).
               addComponent(wpcdfN).
               addComponent(wpcdlW).
               addComponent(wpcdcW)
               );
         
         hGroup.addComponent(wpcdpnlPw);
         
         hGroup.addComponent(wpcdbG);

         wpcdCloneLayout.setHorizontalGroup(hGroup);
      }
      
      // Create a sequential group for the vertical axis.
      {
         GroupLayout.SequentialGroup vGroup = wpcdCloneLayout.
                                              createSequentialGroup();
       
         vGroup.addGroup(wpcdCloneLayout.createParallelGroup(
                                       GroupLayout.Alignment.BASELINE).
            addComponent(wpcdlN).
            addComponent(wpcdfN).
            addComponent(wpcdlW).
            addComponent(wpcdcW)
            ).
            addComponent(wpcdpnlPw).
            addComponent(wpcdbG);
         wpcdCloneLayout.setVerticalGroup(vGroup);
      }

      // create "User Def" modal dialog widgets
      wpudlN  = new JLabel(" Account Name ");
      wpudfN  = new JTextField(12);
      wpudlP  = new JLabel(" Person ");
      wpudfP  = new JTextField(32);
      wpudlC  = new JLabel(" Certificate ");
      
      wpudcbCm = new TaggedDataComboModel();
      wpudcbC  = new JComboBox(wpudcbCm);
      wpudcbC.setEditable(true);
      wpudcbC.setRenderer(new TaggedDataComboRenderer());
      wpudcbC.setMaximumRowCount(10);

      wpudchkPWUnchaged = new JRadioButton(" Leave Password Unchaged ");
      wpudchkEnterPW = new JRadioButton(" Enter Password ");
      wpudchkNoPW = new JRadioButton(" No Password Protection ");
      
      ButtonGroup bgPw = new ButtonGroup();
      bgPw.add(wpudchkEnterPW);
      bgPw.add(wpudchkNoPW);
      bgPw.add(wpudchkPWUnchaged);

      wpudlP1 = new JLabel(" Password ");
      wpudfP1 = new JPasswordField(16);
      wpudlP2 = new JLabel(" Confirm Password ");
      wpudfP2 = new JPasswordField(16);

      AdminUtils.busy(client);
      authElements = new AuthModeComboBoxes(adm,
                         AuthModeComboBoxes.TargetScreen.USER_ACCOUNT_SCREEN);
      AdminUtils.idle(client);

      wpudlD   = new JLabel(" Account is Disabled ");
      wpudchkD = new JCheckBox();

      wpudbC = new JButton("Cancel");
      wpudbS = new JButton("Save");
      
      wpudfP1.setMaximumSize(wpudfP1.getPreferredSize());
      wpudfP2.setMaximumSize(wpudfP2.getPreferredSize());

      layoutDefaultUserDefWidgets(null);

      // create find user account panel
      findUserFindBtn = new JButton("Find");
      findUserCancelBtn = new JButton("Cancel");

      if (extPlugin != null)
      {
         IFindAccountScreen findScreen = new IFindAccountScreen()
         {
            public JButton getCancelButton()
            {
               return findUserCancelBtn;
            }

            public JButton getFindButton()
            {
               return findUserFindBtn;
            }

            public void findAccountByName(String name)
            {
               handlePerformFind(name);
            }

            public void handleCancel()
            {
               AdminUserAccounts.this.handleCancel();
            }
         };

         findUserAccountPanel = extPlugin.buildUserFindPanel(findScreen);
      }

      if (findUserAccountPanel == null)
      {
         findUserAccountPanel = buildDefaultFindPanel();
      }

      // create event handlers: west panel

      // Groups table selection events
      wpUsm.addListSelectionListener(new ListSelectionListener() 
      {
         public void valueChanged(ListSelectionEvent evt) 
         {
            handleUsersSelection(evt);
         }
      });

      // Populate button
      wpbR.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handlePopulateUsers();
         }
      });

      // Add button
      wpbA.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleAddUser();
         }
      });

      // Delete Selected button
      wpbD.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleDeleteSelectedUsers();
         }
      });

      // Edit group button
      wpbE.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleEditUser();
         }
      });

      // Clone user button
      wpbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleClone();
         }
      });

      // Find user button
      wpbF.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            handleFind();
         }
      });

      // Disable user button
      wpbP.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleDisableUser();
         }
      });

      // Switch to ACL's button
      wpbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleSwitchToACLs();
         }
      });

      // create event handlers: center panel

      // users table selection events
      cpGsm.addListSelectionListener(new ListSelectionListener() 
      {
         @Override
         public void valueChanged(ListSelectionEvent evt) 
         {
            handleGroupsSelection(evt);
         }
      });

      // Add user button
      cpbA.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleAddGroup();
         }
      });

      // Delete selected button
      cpbD.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleDeleteSelectedGroups();
         }
      });

      // create event handlers: east panel

      // all users table selection events
      epGsm.addListSelectionListener(new ListSelectionListener() 
      {
         @Override
         public void valueChanged(ListSelectionEvent evt) 
         {
            handleAllGroupsSelection(evt);
         }
      });

      // Populate button
      epbP.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handlePopulateAllGroups();
         }
      });

      // Add selected button
      epbA.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleAddSelectedGroups();
         }
      });

      // Filter button
      epbF.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleFilter();
         }
      });

      // create event handlers: clone user dialog

      // Clone user As button
      wpcdbG.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleCloneAs();
         }
      });

      // clone password radio
      wpcdrCp.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleClonePassword();
         }
      });

      // new password radio
      wpcdrNp.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleClonePassword();
         }
      });

      //  create event handlers: west panel user definition dialog

      // no password authentication check-box
      wpudchkNoPW.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handlePasswordCheckbox();
         }
      });

      // enter password check-box
      wpudchkEnterPW.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handlePasswordCheckbox();
         }
      });

      // leave password unchanged check-box
      wpudchkPWUnchaged.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handlePasswordCheckbox();
         }
      });

      // Cancel button
      wpudbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // Save button
      wpudbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleSave();
         }
      });

      // Disabled Account check-box
      wpudchkD.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt) 
         {
            handleDisabledCheckbox();
         }
      });

      // create the array of components for the state machine
      state = new JComponent[]
      {
         west,
            wpScroller,
               wpUsers,
            wpCtlPanel,
               wpbR, wpbA, wpbD,
               wpbC, wpbE, wpbP,
                           wpbS,
         center,
            cpScroller,
               cpGroups,
            cpCtlPanel,
               cpfN, cpbA, cpbD,
          east,
             epScroller,
                epGroups,
             epCtlPanel,
                epbP, epbA,
                epSearchPanel,
                   epfS, epbF
      };

      // build the map with the default widgets
      userDefWidgets = new HashMap<String, Component>();
      userDefWidgets.put(WIDGET_NAME, wpudfN);
      userDefWidgets.put(WIDGET_PERSON, wpudfP);
      userDefWidgets.put(WIDGET_CERTIFICATE, wpudcbC);
      userDefWidgets.put(WIDGET_NO_PW_AUTH, wpudchkNoPW);
      userDefWidgets.put(WIDGET_ENTER_PW, wpudchkEnterPW);
      userDefWidgets.put(WIDGET_PW_UNCHAGED, wpudchkPWUnchaged);
      userDefWidgets.put(WIDGET_PASSWORD, wpudfP1);
      userDefWidgets.put(WIDGET_CONFIRM_PASSWORD, wpudfP2);
      userDefWidgets.put(WIDGET_AUTH_COMBO,
                         authElements.getAuthModeCombo());
      userDefWidgets.put(WIDGET_AUTH_PLUGIN,
                         authElements.getAuthPluginsCombo());
      userDefWidgets.put(WIDGET_ACC_DISABLED, wpudchkD);

      // set initial state of all components
      stateMachine(state_initial_empty);
      
      // populate the groups list upfront
      handlePopulateUsers();
   }

   /**
    * Performs printing of the current screen data.
    * This screen prints the current view of the users table.
    */
   public void print()
   {
      // check the users table selection
      int[] sel = wpUsers.getSelectedRows();
      int nsel = sel.length;
      int nall = wpUtm.getRowCount();
      final boolean selectOnly;
      
      // no selection means the whole table, otherwise there is a choice
      if (nsel == 0 || nsel == nall)
      {
         // printing the whole table confirmation
         String quest = String.format(
                        "%d user accounts being printed. Continue?",
                        nall);
         if (!AdminUtils.confirmDialog(client, "Printing Users", quest))
         {
            // declined
            return;
         }
         
         // print all
         selectOnly = false;
      }
      else
      {
         // offering a choice 
         String quest = "Please refine your request";
         String[] choices = new String[2];
         choices[0] = nsel == 1 ? "the selected user account" :
                      String.format("%d selected user accounts", nsel);
         choices[1] = nall == 2 ? "both user accounts" :
                      String.format("all %d user accounts", nall);
         
         int n = AdminUtils.choiceDialog(client, "Printing Users", quest,
                                         choices, "Print");
         if (n < 0)
         {
            // declined
            return;
         }
         
         // print mode choice
         selectOnly = n == 0;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // build preview main table
      List<AdminUtils.AccountInfo> backUl = new LinkedList
                                                <AdminUtils.AccountInfo>();
      String[] extHeaders = null;
      
      if (extPlugin != null)
      {
         extHeaders = extPlugin.getPreviewHeaders();
      }
      else
      {
         extHeaders = new String[] {};
      }
      
      for (int i = 0; i < nall; i++)
      {
         // traslate model to view
         int iv = wpUsers.convertRowIndexToView(i);
         
         // honor selection if asked
         if (selectOnly && !wpUsm.isSelectedIndex(iv))
         {
            continue;
         }
         
         // stuff the back list with the user info
         AdminUtils.AccountInfo backUi = new AdminUtils.AccountInfo();
         backUi.aid    = (String)wpUtm.getValueAt(i, 0);
         backUi.descr  = (String)wpUtm.getValueAt(i, 1);
         backUi.disabled = "";
         backUi.alias = "";
         backUi.attr1 = "";
         backUi.attr2 = "";
         
         // load user from server
         UserDef user = adm.getUser(backUi.aid);
         if (user == null)
         {
            // user may have been deleted
            backUi.descr = "*** possibly deleted ***";
         }
         else
         {
            // gotten user info
            backUi.disabled = user.enabled ? "" : "yes"; 
            backUi.alias = user.alias; 
            backUi.attr1 = user.protect ? "password" : "unprotected";  

            StringBuilder sb = new StringBuilder();
            for (int j = 0; j < user.groups.length; j++)
            {
               sb.append(user.groups[j]);
               if (j < user.groups.length - 1)
               {
                  sb.append(",");
               }
            }
            backUi.attr2 = sb.toString();
         }
         
         if (extPlugin != null)
         {
            backUi.ext = extPlugin.getPreviewRow(backUi.aid);
         }
         else
         {
            backUi.ext = new String[] {};
         }
         
         // all details are collected
         backUl.add(backUi);
      }
      
      header = new MessageFormat(
             selectOnly ? "Selected User Accounts" : "All User Accounts");
      footer = new MessageFormat("page {0, number}");
               
      backTm = new AdminUtils.PrintAccountModel(extHeaders,
                              backUl.toArray(new AdminUtils.AccountInfo[] {}));
      backTable = new JTable(backTm);
      
      // idle
      AdminUtils.idle(client);
      
      client.printOn(backTable, backTm, header, footer);
   }
   
   /**
    * Queries the extended printing capability. If the extended capability is
    * not provided, the printExtended() method should not be called.
    * 
    * @return   <code>true</code> to tell the caller the extended printing is
    *           implemented
    */
   public boolean canPrintExtended()
   {
      return true;
   }

   /**
    * Creates extended print preview table off the current screen data.
    * This screen prints the current view of the users table.
    */
   public void printExtended()
   {
      // busy
      AdminUtils.busy(client);
      
      // standard part
      String[] stdHeaders = new String[] {"Account",
                                          "Disabled",
                                          "Person",
                                          "Status"
                                         };
         
      int stdRows = 1;          // all std info takes one row
      int stdCols = 4;          // all std info takes four columns
      
      // extended part
      String[] extHeaders = null;
      int extRows = 0;
      int extCols = 0;
      
      if (extPlugin != null)
      {
         extHeaders = extPlugin.getExtendedHeaders();
         extCols = extHeaders != null ? extHeaders.length : 0;
         extRows = extPlugin.getExtendedRows();
      }
      else
      {
         extHeaders = new String[] {};
      }
      
      // extended view table model and table
      frontTm = new AdminUtils.PrintExtendedAccountModel(
                               stdHeaders, extHeaders, new String[][] {{}});
      frontTable = new JTable(frontTm);
      
      // create the extended view
      int nRec = backTable.getRowCount();
      int step = extRows > stdRows ? extRows : stdRows;
      int nExt = nRec * step;
      int nCol = stdCols + extCols;
      
      String[][] extView = new String[nExt][nCol];
      
      // pre-fill the extended view array
      for (int i = 0; i < nExt; i++)
      {
         for (int j = 0; j < nCol; j++)
         {
            extView[i][j] = "";
         }
      }
      ((AdminUtils.PrintExtendedAccountModel)frontTm).reload(extView);
      
      // fill the extended view with real data
      for (int n = 0; n < nRec; n++)
      {
         // convert the preview table index into preview model
         int nd = backTable.convertRowIndexToModel(n);
         
         // starting row
         int nr = n * step;
         
         // get the standard info
         String subj = (String)backTm.getValueAt(nd, 0);
         frontTm.setValueAt(subj, nr, 0);                         // uid
         frontTm.setValueAt(backTm.getValueAt(nd, 1), nr, 2);     // person
         frontTm.setValueAt(backTm.getValueAt(nd, 2), nr, 1);     // disabled
         frontTm.setValueAt(backTm.getValueAt(nd, 4), nr, 3);     // status
         
         // get the extended info
         if (extCols > 0)
         {
            extPlugin.fillExtendedRow(subj, frontTm, nr, stdCols);
         }
      }
      
      // idle
      AdminUtils.idle(client);
      
      client.printExtendedOn(frontTable, header, footer);
   }
   
   /**
    * Get the close button associated with the account dialog.
    * 
    * @return   the close button
    */
   public JButton getCloseButton()
   {
      return wpudbC;
   }

   /**
    * Get the save button for the account dialog.
    * 
    * @return   the save button
    */
    public JButton getSaveButton()
   {
      return wpudbS;
   }

   /**
    * Get the panel where default account widgets are kept.
    * 
    * @return   the panel which contains default widgets.
    */
   public JPanel getDefaultWidgetPanel()
   {
      return wpudpUserPanel;
   }

   /**
    * Get the widget with the specified name.
    * 
    * @param    name
    *           The widget's name
    * 
    * @return   the widget with the specified name or <code>null</code> if the
    *           widget does not exist.
    */
   public Component getWidgetByName(String name)
   {
      return userDefWidgets.get(name);
   }

   /** 
    * Presents the refreshed user accounts management screen.
    */
   void manageUsers() 
   {
      // switch to this screen without changing state of components
      client.show(SCREEN_ACC_USERS, TITLE_ACC_USERS);
   }
   
   /**
    * Handle radio selections related to password in clone-user dialog.
    */
   private void handleClonePassword()
   {
      boolean newPw = wpcdrNp.isSelected();
      
      // enable the password fields only if a new password is computed
      wpcdfP1.setEnabled(newPw);
      wpcdfP2.setEnabled(newPw);
      
      if (!newPw)
      {
         // if no password is needed, reset the fields
         wpcdfP1.setText("");
         wpcdfP2.setText("");
      }
   }

   /** 
    * Changes state of the components in response to users list selections.
    * There are three distinctive states:
    * <ul>
    *   <li>no selection or empty list
    *   <li>single entry selected
    *   <li>multiple entries selected
    * </ul>
    * This method determines the selection state of the list and changes
    * the state of the components accordingly.
    * <p>
    * No selection means that the state should be STATE_INITIAL_EMPTY.
    * Only the "Populate" and "Add user" buttons are enabled.
    * <p>
    * Single selection means that all buttons of the west panel are enabled, 
    * since all these operations make sense for a single selected user.<br>
    * Additionally, the center panel's list of groups gets populated with 
    * the assigned groups for the selected user. The related "Add group" 
    * button is enabled, too.<br>
    * At last, the east panel, if it is populated, shows the selection
    * of groups from the center panel.
    * <p>
    * Multiple selection means that buttons, specific to the single selection,
    * are disabled. The enabled buttons on the west panel are Populate,
    * Add user and Delete selected.<br>
    * The center panel is fully disabled and the east panel's list of all
    * groups has no selection and is disabled.
    *
    * @param    evt
    *           event caused this method call
    */
   private void handleUsersSelection(ListSelectionEvent evt)
   {
      // do not bother until the dust is fully settled
      if (evt.getValueIsAdjusting())
      {
         return;
      }

      // figure out the selection state
      if (wpUsm.isSelectionEmpty())
      {
         // no selection
         setUsersNoSelection();
      }
      else if (wpUsers.getSelectedRows().length == 1)
      {
         // single selection
         setUsersSingleSelection();
      }
      else
      {
         // multiple selection
         setUsersMultiSelection();
      }
   }

   /** 
    * Changes state of the components in response to groups list selections.
    * There are two distinctive states:
    * <ul>
    *   <li>no selection or empty list
    *   <li>single or multiple selection
    * </ul>
    * This method determines the selection state of the list and changes
    * the state of the components accordingly.
    * <p>
    * No selection means that the Delete Selected button should be disabled.
    * Otherwise it should be enabled.
    *
    * @param    evt
    *           event caused this method call
    */
   private void handleGroupsSelection(ListSelectionEvent evt) 
   {
      // do not bother until the dust is fully settled
      if (evt.getValueIsAdjusting())
      {
         return;
      }

      // figure out the selection state
      if (cpGsm.isSelectionEmpty())
      {
         // no selection
         setGroupsNoSelection();
      }
      else
      {
         // some selection
         setGroupsAnySelection();
      }
   }

   /**
    * Changes state of the components in response to all groups list
    * selections. The idea is to keep all assigned groups of the selected user
    * also in the selected state in the all groups list. If extra lines are
    * selected, the Add selected button gets enabled. If some pre-selected
    * lines are deselected, they are reselected to maintain the status.
    * 
    * @param    evt
    *           event caused this method call
    */
   private void handleAllGroupsSelection(ListSelectionEvent evt)
   {
      // do not bother until the dust is fully settled
      if (evt.getValueIsAdjusting())
      {
         return;
      }

      if (selecting)
      {
         return;
      }
      selecting = true;
      
      // check the user selection type
      if (!singleUser)
      {
         epGsm.clearSelection();
         selecting = false;
         
         // disable Add selected button
         epbA.setEnabled(false);
         return;
      }

      // get the changed selection range
      int fi = evt.getFirstIndex();
      int ti = evt.getLastIndex();
      
      // inspect the change in selection line by line
      boolean extra = false;
      
      for (int i = (fi < 0 ? 0 : fi); i <= ti; i++)
      {
         // check the selection status
         boolean sel = epGsm.isSelectedIndex(i);
         String user = (String) epGroups.getValueAt(i, 0);
         
         if (sel)
         {
            // this line may have been selected
            if (!cpGtm.isListed(user))
            {
               // extra selection
               extra = true;
            }
            // else do nothing
         }
         else
         {
            // this line may have been deselected
            if (cpGtm.isListed(user))
            {
               // should be reselected
               epGroups.addRowSelectionInterval(i, i);
            }
            else
            {
               // an extra line was deselected
               // rescan the whole selection to detect extra lines
               int[] all = epGroups.getSelectedRows();
               extra = false;
               
               for (int j = 0; j < all.length; j++)
               {
                  if (!cpGtm.isListed((String)epGroups.getValueAt(all[j], 0)))
                  {
                     extra = true;
                  }
               }
            }
         }
      }
      
      // enable or disable Add selected button
      epbA.setEnabled(extra);

      selecting = false;
   }
   
   /**
    * Handle REfresh button press in the west panel.
    * <p>
    * Refresh button fills the list of users with the information from
    * the server.
    * It also clears the group list of the center panel and any selection in
    * the all groups list of the east panel.
    */
   private void handlePopulateUsers()
   {
      // populate the users list
      populateUsersList();

      // set the state accordingly
      setUsersNoSelection();
   }

   /**
    * Handle Populate button press in the east panel.
    * <p>
    * Populate button fills the list of all groups with the information from
    * the server.
    * <p>
    * If the group list is non-empty, the search sub-panel gets enabled as well.
    * <p>
    * If there is a single selected user and the group list is non-empty,
    * the list selection is set to reflect the assigned groups for the user.
    */
   private void handlePopulateAllGroups()
   {
      // populate the all users list
      populateAllGroupsList();
      epbP.setText("Refresh");

      // set the state accordingly
      if (epGtm.data.size() > 0)
      {
         // set enabled state of the components
         stateMachine(MASK_EAST, state_all_groups);
       
         // check for the single selection in Users
         if (wpUsers.getSelectedRows().length == 1)
         {
            setAllGroupsSelection();
         }
      }
   }
   
   /**
    * Populate the list of users with fresh data from the server.
    */
   private void populateUsersList() 
   {
      if (!canEnum)
      {
         return;
      }
      
      // clear selection if any
      wpUsm.clearSelection();
      
      // clear the old list
      wpUtm.clear();
      rebuildSearchIndex();

      // get the list of users from the server
      AdminUtils.busy(client);
      TaggedName[] list = adm.listUsers();
      AdminUtils.idle(client);
      
      if (list != null)
      {
         if (list.length > 0)
         {
            wpUtm.loadAll(list);
            rebuildSearchIndex();
         }
      }
      else
      {
         client.alarm();
      }
   }
   
   /**
    * Populate the list of all defined groups. 
    */
   private void populateAllGroupsList() 
   {
      // clear existing selection if any
      epGsm.clearSelection();

      // clear existing records
      epGtm.clear();
      
      // populate the list of all users
      AdminUtils.busy(client);
      TaggedName[] list = adm.listGroups();
      AdminUtils.idle(client);
      
      if (list != null)
      {
         if (list.length > 0)
         {
            epGtm.loadAll(list);
         }
      }
      else
      {
         client.alarm();
      }
   }
   
   /**
    * Populate the list of groups for a given user with fresh data from the
    * server.
    * 
    * @param    userName
    *           name of the target user
    */
   private void populateGroupsList(String userName) 
   {
      // clear existing records
      cpGtm.clear();
      
      // populate the list of group users
      AdminUtils.busy(client);
      TaggedName[] list = adm.listUserGroups(userName);
      AdminUtils.idle(client);
      
      if (list != null)
      {
         if (list.length > 0)
         {
            cpGtm.loadAll(list);
         }
      }
      else
      {
         client.alarm();
      }
   }

   /**
    * Sets state of components to no selection in the Users list
    * <p>
    * No selection means that the state should be STATE_INITIAL_EMPTY.
    * Only the "Populate" and "Add User" buttons are enabled.
    */
   private void setUsersNoSelection()
   {
      // reset flags
      singleUser = false;

      // set enabled state of the list accordingly
      stateMachine(MASK_WEST_CENTER, state_initial_empty);
      
      // clear the center panel
      resetCenterPanel();
      
      // clear selection in east panel, if any
      resetEastPanel();
   }

   /**
    * Sets state of components to single selection in the Users list 
    * <p>
    * Single selection means that all buttons of the west panel are enabled, 
    * since all these operations make sense for a single selected user.<br>
    * Additionally, the center panel's list of groups gets populated with 
    * the assigned groups for the selected user. The related "Add group" 
    * button is enabled, too.<br>
    * At last, the east panel, if it is populated, shows the selection
    * of groups from the center panel.
    */
   private void setUsersSingleSelection()
   {
      // busy
      AdminUtils.busy(client);
      
      // set flags
      singleUser = true;

      // set enabled state of the components
      stateMachine(MASK_WEST_CENTER, state_users_single);
      
      // populate center panel's groups list
      userName = (String) wpUtm.getValueAt(wpUsers.getSelectedRow(), 0);
      populateGroupsList(userName);
      
      // set proper selection in east panel's all groups list
      setAllGroupsSelection();
      
      UserDef user = adm.getUser(userName);
      if (user != null)
      {
         boolean enabled = user.enabled;
         wpbP.setText(enabled ? " Disable User " : " Enable User ");
         wpbP.repaint();
      }
      else
      {
         client.alarm();
      }

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Sets state of components to no selection in the Groups list 
    * <p>
    * Multiple selection means that buttons, specific to the single selection,
    * are disabled. The enabled buttons on the west panel are Populate,
    * Add group and Delete selected.<br>
    * The center panel is fully disabled and the east panel's list of all
    * users has no selection and is disabled.
    */
   private void setUsersMultiSelection()
   {
      // reset flags
      singleUser = false;
      
      // set enabled state of the components
      stateMachine(MASK_WEST_CENTER, state_users_multi);
      
      // clear the center panel
      resetCenterPanel();
      
      // clear selection in east panel, if any
      resetEastPanel();
   }
   
   /**
    * Sets state of components to no selection in the Groups list
    * No selection means that the Delete Selected button should be disabled.
    */
   private void setGroupsNoSelection()
   {
      cpbD.setEnabled(false);
   }
   
   /**
    * Sets state of components to any selection in the Groups list
    * Any selection means that the Delete Selected button should be enabled.
    */
   private void setGroupsAnySelection()
   {
      cpbD.setEnabled(true & canGroup);
   }

   /**
    * Sets selection of the all groups list so that all groups of the
    * currently selected single user are selected in the all groups list.
    */
   private void setAllGroupsSelection()
   {
      // clear existing selection if any
      epGsm.clearSelection();
      
      // get all assigned groups of the selected user and select them one by 
      // one
      int nusers = cpGtm.getRowCount();
      
      nextGroup:
      for (int i = 0; i < nusers; i++)
      {
         // get group at this position
         String group = (String)cpGtm.getValueAt(i, 0);
         
         // locate the user in the all users list
         for (int j = 0; j < epGtm.getRowCount(); j++)
         {
            if (epGtm.getValueAt(j, 0).equals(group))
            {
               // convert the model index to view index
               int iv = epGroups.convertRowIndexToView(j);
               
               // select this index
               if (iv != -1)
               {
                  epGroups.addRowSelectionInterval(iv, iv);
               }
               continue nextGroup;
            }
         }
      }

      // disable Add selected button
      epbA.setEnabled(false);
   }   

   /**
    * Resets the center panel after reloading the users list. 
    * Center panel's list of groups is cleared as well as two text fields. 
    */
   private void resetCenterPanel()
   {
      // clear existing records
      cpGtm.clear();
      
      // clear text input
      cpfN.setText("");
   }
   
   /**
    * Resets the east panel after reloading the users list. 
    * East panel's list of all groups clears the current selection. 
    */
   private void resetEastPanel()
   {
      // clear existing selection
      epGsm.clearSelection();

      // disable Add selected button
      epbA.setEnabled(false);
   }

   /**
    * Handle Switch to ACLs button press in the west panel.
    * <p>
    * This method passes the process name to the ACLs screen and initiate 
    * Access Control --&gt; Show By Account menu choice
    */
   private void handleSwitchToACLs()
   {
      client.show(AdminACL.SCREEN_ACLS, AdminACL.TITLE_ACLS);
      AdminACL adminAcl = client.getMenu().getACL().getAclAdmin();
      
      adminAcl.showByAccount(userName);
   }

   /**
    * Handle Add button press in the west panel.
    * <p>
    * This will reset all fields in the Add User dialog and will display it.
    */
   private void handleAddUser()
   {
      loadCertificatesInCombo();
      
      // focus on user name
      wpudfN.requestFocusInWindow();

      JPanel panel = checkExtFieldsLayout(true, canWrite);

      // (re-)enable some controls that may have been disabled
      wpudfP.setEnabled(true);
      wpudcbC.setEnabled(true);
      authElements.setEnabled(true);
      wpudchkD.setEnabled(true);
      wpudchkPWUnchaged.setEnabled(true);
      wpudchkEnterPW.setEnabled(true);
      wpudchkNoPW.setEnabled(true);
      wpudfP1.setEnabled(true);
      wpudfP2.setEnabled(true);
      wpudbS.setEnabled(true);
      
      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "User Account Definition", 
                                      panel);
      wpudfN.setEnabled(true); // enable Name field

      resetUserDialog();
      
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }

   /**
    * Reset all fields from the dialog.
    */
   private void resetUserDialog()
   {
      wpudfN.setText("");
      wpudfP.setText("");
      wpudcbCm.setSelectedItem(null);
      wpudfP1.setText("");
      wpudfP2.setText("");
      authElements.reset();
      
      wpudchkNoPW.setSelected(false);
      wpudchkPWUnchaged.setSelected(false);
      wpudchkEnterPW.setSelected(true);
      
      boolean add = wpudfN.isEnabled();
      wpudchkPWUnchaged.setEnabled(!add);

      wpudchkD.setSelected(false);
      wpudlD.setForeground(Color.DARK_GRAY);
      
      if (extPlugin != null)
      {
         extPlugin.resetDialog();
      }
   }
   
   /**
    * Load certificate definitions in combo.
    */
   private void loadCertificatesInCombo()
   {
      AdminUtils.busy(client);
      TaggedName[] list = adm.listPeerCerts(false);
      AdminUtils.idle(client);

      if (list != null)
      {
         if (list.length > 0)
         {
            // set certificate description to Common Name
            setCertificateDescription(list, CERT_DESCRIPTION_PROP);

            // add the certificates to the user dialog combo
            wpudcbCm.loadAll(list);
         }
      }
   }

   /**
    * Handle Edit button press in the west panel.
    * <p>
    * This will load the user definition for the currently selected row.
    * If the data can not be loaded, it will display an error.
    */
   private void handleEditUser()
   {
      // load user from server
      UserDef user = adm.getUser(userName);
      if (user == null)
      {
         // user may have been deleted
         client.alarm();
         
         // refresh the data from the server
         handlePopulateUsers();
         return;
      }

      loadCertificatesInCombo();
      
      // initialize dialog fields
      wpudfN.setText(user.subjectId);
      wpudfP.setText(user.person);
      wpudcbCm.setSelectedItem(user.alias);
      authElements.applyUserSettings(user);

      // first pass for password check-boxes - select using the 'protect' flag
      wpudchkNoPW.setSelected(!user.protect);
      wpudchkEnterPW.setSelected(false);
      wpudchkPWUnchaged.setSelected(user.protect);
      
      boolean hasPass = user.password != null;

      wpudchkPWUnchaged.setEnabled(hasPass);
      // second pass for password check-boxes - select using the password 
      // state, if there is no password set
      if (!hasPass)
      {
         wpudchkNoPW.setSelected(!user.protect);
         wpudchkEnterPW.setSelected(user.protect);
         wpudchkPWUnchaged.setSelected(false);
         
      }

      handlePasswordCheckbox();
      
      // reset password text
      wpudfP1.setText("");
      wpudfP2.setText("");

      // set account disabled check-box
      wpudchkD.setSelected(!user.enabled);
      
      // focus on person
      wpudfP.requestFocusInWindow();
      
      wpudfN.setEnabled(false); // disable user name field

      JPanel panel = checkExtFieldsLayout(false, canWrite);

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "User Acount Definition", 
                                      panel);
      
      wpudlD.setForeground(user.enabled ? Color.DARK_GRAY: Color.RED);

      // selectively disable inputs depending on permissions profile
      if (!canWrite)
      {
         wpudfP.setEnabled(false);
         wpudcbC.setEnabled(false);
         authElements.setEnabled(false);
         wpudchkD.setEnabled(false);
      }
      
      if (!canWrite && ! canPw)
      {
         wpudchkPWUnchaged.setEnabled(false);
         wpudchkEnterPW.setEnabled(false);
         wpudchkNoPW.setEnabled(false);
         wpudfP1.setEnabled(false);
         wpudfP2.setEnabled(false);
      }
      
      // allow Save button conditionally
      wpudbS.setEnabled(canWrite | canPw);
      
      dialog.setVisible(true);

      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }
   
   /**
    * Layout the extension fields for this user.
    * 
    * @param    add
    *           <code>true</code> if this is for adding a new user.
    * 
    * @param    writable
    *           <code>true</code> if this is for enabled inputs.
    *           <code>false</code> for read only panel
    * 
    * @return   the container which holds user def fields.
    */
   private JPanel checkExtFieldsLayout(boolean add, boolean writable)
   {
      if (extPlugin == null)
         return wpudpUserPanel;
      
      JPanel c = wpudpUserPanel;
      
      String user = add ? null : userName;
      
      switch (extPlugin.getLayoutMode())
      {
         case ClientAccountExtension.LAYOUT_ALL_WIDGETS:
         {
            // relayout the default widgets, just in case the plugin doesn't
            // need to change their order (it just moves the panel around)
            layoutDefaultUserDefWidgets(null);

            // in this case, the ext plugin must be given a chance to re-layout
            // the default widgets too
            c = extPlugin.layoutAllWidgets(user, writable);
            break;
         }
         case ClientAccountExtension.LAYOUT_EXT_WIDGETS:
         {
            extContainer = extPlugin.getExtContainer();
            
            // this container must be added to the dialog ...
            layoutDefaultUserDefWidgets(extContainer);
            
            // allow the plugin to layout its data
            extPlugin.layoutExtWidgets(extContainer, user, writable);
            break;
         }
      }
      
      return c;
   }

   /**
    * Layout the default widgets in the user-def dialog.
    */
   private void layoutDefaultUserDefWidgets(Container extContainer)
   {
      // Cancel/Save buttons will be added only if the plugin doesn't need
      // to layout all widgets
      boolean addButtons = extPlugin == null || 
       extPlugin.getLayoutMode() == ClientAccountExtension.LAYOUT_EXT_WIDGETS;

      JPanel wpudpPW = AdminUtils.groupComponents(
            new Component[][] 
            { 
                  { wpudchkPWUnchaged, new JLabel(""), new JLabel("") }, 
                  { wpudchkEnterPW, wpudlP1, wpudfP1 }, 
                  { wpudchkNoPW, wpudlP2, wpudfP2 }, 
            }
         );
      wpudfP1.setMaximumSize(wpudfP1.getPreferredSize());
      wpudfP2.setMaximumSize(wpudfP2.getPreferredSize());
      
      wpudpPW.setBorder(new EtchedBorder(EtchedBorder.RAISED));
      
      GroupLayout layout = (GroupLayout) wpudpPW.getLayout();
      layout.setAutoCreateContainerGaps(false);

      // add components
      JPanel pnl1 = AdminUtils.groupComponents(new Component[][] {
            { wpudlN, wpudfN },
            { wpudlD, wpudchkD },
            { wpudlP, wpudfP },
            { authElements.getAuthModeLabel(),
              authElements.getAuthModeCombo() },
            { authElements.getAuthPluginsLabel(),
              authElements.getAuthPluginsCombo() },
            { wpudlC, wpudcbC }
      });
      
      layout = (GroupLayout) pnl1.getLayout();
      layout.setAutoCreateContainerGaps(false);
      

      if (addButtons)
      {
         JPanel pnl2 = null;

         pnl2 = AdminUtils.groupComponents(new Component[][] {
               { wpudbC, wpudbS },
         });
         layout = (GroupLayout) pnl2.getLayout();
         layout.setAutoCreateContainerGaps(false);

         wpudpUserPanel = AdminUtils.groupComponents(
               new Component[][] { { pnl1 }, { wpudpPW }, { pnl2 } } );
      }
      else
      {
         wpudpUserPanel = AdminUtils.groupComponents(
               new Component[][] { { pnl1 }, { wpudpPW } } );         
      }

      wpudpUserPanel.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " User Definition "));
      
      if (extContainer != null)
      {
         wpudpUserPanel.add(extContainer);
      }
   }

   /**
    * Handle Save button press in the User Def panel.
    * Save button tries and adds or update the user account definition.
    * When adding, if the server responds OK, it will add the new row to the
    * west-side table. When editing, it will update the currently selected
    * row.
    */
   private void handleSave()
   {
      // determine the mode: add vs edit
      boolean add = wpudfN.isEnabled();
      
      // get and check the user name
      String name = validateStringField(wpudfN, "type a user name");
      if (name == null)
      {
         return;
      }
      
      // get and check the person field
      String person = validateStringField(wpudfP, "type a person name");
      if (person == null)
      {
         return;
      }

      // get the certificate alias
      TaggedName certTag= ((TaggedName) wpudcbCm.getSelectedItem());
      String certAlias = certTag == null ? null : certTag.getName();

      String password = null;
      // validate password, if specified
      if (wpudchkEnterPW.isSelected())
      {
         password = validatePassword(wpudfP1, wpudfP2);
         // return if provided passwords do not match  
         if (password == null)
         {
            return;
         }
      }
      
      // if no pw protection is required, password must be null
      boolean noPwProtection = wpudchkNoPW.isSelected();
      if (noPwProtection)
      {
         password = null;
      }
      
      // determine the authentication mode
      int authMode = authElements.getAuthMode();
      String authPlugin = authElements.getAuthPlugin();
      
      // determine account disabled check-box
      boolean disabled = wpudchkD.isSelected();
      
      byte[] hashPass = null;
      if (password != null)
      {
         hashPass = HashPassword.hashPassword(password);
      }
      else
      {
         if (!add && wpudchkPWUnchaged.isSelected())
         {
             // the old password must be kept
            AdminUtils.busy(client);

            UserDef ud = adm.getUser(name);
            
            if (ud != null)
            {
               hashPass = ud.password;
               noPwProtection = !ud.protect;
            }
            
            AdminUtils.idle(client);
         }
      }
      
      // dispatch the action
      if (add)
      {
         try
         {
            // add the definition to the server and iterate
            // list remains with no selection         
            if (saveUserDef(name, person, certAlias, hashPass, noPwProtection, 
                            authMode, authPlugin, disabled))
            {
               // successfully added to the server; add to the list
               client.ok();
               wpUtm.addRow(new TaggedName(name, person));
               rebuildSearchIndex();

               // refocus name field
               wpudfN.requestFocusInWindow();

               // clear selection now when the table is modified
               wpUsm.clearSelection();

               // reset the dialog
               resetUserDialog();
               return;
            }

            // adding definition failed; stay in the editing
            wpudfN.requestFocusInWindow();
            client.alarm();
         }
         catch (AccountExtValidationException e)
         {
            client.beep(e.getMessage());
            AdminUtils.idle(client);
            return;
         }         
      }
      else
      {
         // change the existing definition
         // list retains the selection
         if (setUserDef(name, person, hashPass, noPwProtection, certAlias, 
               authMode, authPlugin, disabled))
         {
            // successfully saved changes
            client.ok();
            wpUtm.refreshRow(new TaggedName(name, person),
                             wpUsers.getSelectedRow());
            //(no need to rebuild search index because user name cannot be
            // changed)
            
            // hide the dialog
            dialog.setVisible(false);
            dialog = null;
            
            // re-select user
            setUsersSingleSelection();
            
            return;
         }
         
         // saving definition failed; stay in the editing 
         wpudfN.requestFocusInWindow();
         client.alarm();
      }
   }

   /**
    * Validate the password in the specified pass-fields.
    * 
    * @param    wp1
    *           the new password widget
    * @param    wp2
    *           the confirm password widget
    * 
    * @return   the specified password or null if password is not valid.
    */
   private String validatePassword(JPasswordField wp1, JPasswordField wp2) 
   {
      // verify the passwords
      String p1 = new String(wp1.getPassword());
      String p2 = new String(wp2.getPassword());
      
      if (!p1.equals(p2))
      {
         client.beep("passwords do not match");
         return null;
      }
      
      if (p1.length() == 0)
      {
         client.beep("password must be provided");
         return null;
      }

      return p1;
   }

   /**
    * Adds new user account definition to the server.
    * 
    * @param    name
    *           new user name
    * @param    person
    *           assigned person
    * @param    alias
    *           new certificate alias
    * @param    password
    *           new password
    * @param    noPwProtection
    *           <code>true</code> if account is not protected with a password
    * @param    mode
    *           new authentication mode
    * @param    authPlugin
    *           new custom authentication plugin class
    * @param    disabled
    *           <code>true</code> if account is disabled
    * 
    * @return   <code>true</code> if added successfully
    */
   private boolean saveUserDef(String  name, 
                               String  person, 
                               String  alias, 
                               byte[]  password,
                               boolean noPwProtection,
                               int     mode,
                               String authPlugin,
                               boolean disabled)
   throws AccountExtValidationException
   {
      UserDef ud    = new UserDef();
      ud.subjectId  = name;
      ud.person     = person;
      ud.alias      = alias;
      ud.password   = password;
      ud.protect    = !noPwProtection;
      ud.mode       = mode;
      ud.groups     = null;
      ud.enabled    = !disabled;
      ud.authPlugin = authPlugin;
      
      // date and time fields are set by the server
      
      Serializable ext = null;
      if (extPlugin != null)
      {
         // validate the extensions, if a plugin was provided
         ext = extPlugin.validateExtensions(name);
      }
      
      AdminUtils.busy(client);
      boolean rc = adm.addUser(ud, ext);
      AdminUtils.idle(client);
      
      return rc;
   }

   /**
    * Change user account definition.
    * 
    * @param    name
    *           user name
    * @param    person
    *           new assigned person
    * @param    password
    *           new assigned password
    * @param    noPwProtection
    *           <code>true</code> if account is not protected with a password
    * @param    alias
    *           certificate alias
    * @param    mode
    *           authentication mode
    * @param    authPlugin
    *           new custom authentication plugin class
    * @param    disabled
    *           <code>true</code> if account is disabled
    * 
    * @return   <code>true</code> if changed successfully
    */
   private boolean setUserDef(String  name, 
                              String  person,
                              byte[]  password,
                              boolean noPwProtection,
                              String  alias, 
                              int     mode,
                              String  authPlugin,
                              boolean disabled)
   {
      AdminUtils.busy(client);

      UserDef ud = adm.getUser(name);
      // check if the user exists on the server
      if (ud == null)
      {
         AdminUtils.idle(client);
         return false;
      }

      // set the new field values
      ud.person     = person;
      if (noPwProtection)
      {
         // set the protect flag to OFF if 'no pw protection' check-box is set
         ud.protect = false;
         password   = null;
      }

      if (password != null)
      {
         // set the protect flag to ON only if password is provided
         ud.protect    = true;
      }

      ud.password   = password;

      // if password is not set or the 'no pw protection' check-box is off, 
      // leave the flag untouched

      ud.alias      = alias;
      ud.mode       = mode;
      ud.authPlugin = authPlugin;
      ud.enabled    = !disabled;
      
      Serializable ext = null;
      if (extPlugin != null)
      {
         // validate the extensions, if a plugin was provided
         try
         {
            ext = extPlugin.validateExtensions(name);
         } 
         catch (AccountExtValidationException e)
         {
            client.beep(e.getMessage());
            AdminUtils.idle(client);
            return false;
         }         
      }
      
      boolean rc = adm.setUser(ud, ext);
      AdminUtils.idle(client);
      
      return rc;
   }
   
   /**
    * This handles the password check-boxes, displayed in the User Def dialog.
    * <p>
    * When the No Password Authentication or Leave Password Unchanged
    * check-box is selected, it will disable and reset the password fields
    */
   private void handlePasswordCheckbox()
   {
      if (wpudchkNoPW.isSelected() || wpudchkPWUnchaged.isSelected())
      {
         wpudfP1.setEnabled(false);
         wpudfP2.setEnabled(false);

         wpudfP1.setText("");
         wpudfP2.setText("");
      }
      
      if (wpudchkEnterPW.isSelected())
      {
         wpudfP1.setEnabled(true);
         wpudfP2.setEnabled(true);
      }
   }

   /**
    * Handle Delete Selected button press in the west panel.
    */
   private void handleDeleteSelectedUsers() 
   {
      // get the selection
      int[] sel = wpUsers.getSelectedRows();
      
      // safety check
      if (sel.length == 0)
      {
         return;
      }
      
      // ask confirmation
      String quest = sel.length == 1 ?
                     "1 user account being deleted. Continue?" :
                     Integer.toString(sel.length) + 
                     " user accounts being deleted. Continue?";
      
      if (!AdminUtils.confirmDialog(client, "Deleting Users", quest))
      {
         // declined
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // sort the array of selected indices to walk backwards
      Arrays.sort(sel);
      
      // walk the array backwards and delete selected groups
      for (int i = sel.length - 1; i >= 0; i--)
      {
         // pick up the name of group being deleted
         String user = (String) wpUtm.getValueAt((sel[i]), 0);
         
         // attempt deletion
         boolean rc = adm.deleteUser(user);
         if (!rc)
         {
            // attempt has failed
            client.alarm();
         }
         else
         {
            // attempt has succeeded; delete this row
            client.ok();
            wpUtm.deleteRow(sel[i]);
            rebuildSearchIndex();
         }
      }

      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Handle Disabled Account check-box in user def dialog.
    */
   private void handleDisabledCheckbox()
   {
      Color color = wpudchkD.isSelected() ? Color.red : Color.black;
      
      wpudlD.setForeground(color);
   }
   
   /**
    * Handle Disable User button press in west-panel.
    */
   private void handleDisableUser()
   {
      if (!AdminUtils.confirmDialog(client, 
                                   wpbP.getText(), 
                                   wpbP.getText() + " " + userName + " ?"))
      {
         return;
      }
      
      AdminUtils.busy(client);
      
      UserDef user = adm.getUser(userName);
      if (user != null)
      {
         // toggle the disabled flag
         user.enabled = !user.enabled;
         adm.setUser(user);
         
         setUsersSingleSelection();
      }
      else
      {
         client.alarm();
      }
      
      AdminUtils.idle(client);
   }
   
   /**
    * Handle Clone User button press in west-panel.
    */
   private void handleClone()
   {
      // reset fields
      wpcdfN.setText("");
      wpcdfP1.setText("");
      wpcdfP2.setText("");
      wpcdrCp.setSelected(false);
      wpcdrNp.setSelected(true);
      wpcdcW.setSelected(false);
      
      UserDef user = null;
      
      AdminUtils.busy(client);
      user = adm.getUser(userName);
      AdminUtils.idle(client);
      if (user == null)
      {
         client.alarm();
         return;
      }
      
      boolean hasPw = user.protect;
      
      // enable all widgets related to password only if needed
      wpcdrCp.setEnabled(hasPw);
      wpcdrNp.setEnabled(hasPw);
      wpcdfP1.setEnabled(hasPw);
      wpcdfP2.setEnabled(hasPw);

      wpcdrCp.setSelected(false);
      wpcdrNp.setSelected(hasPw);

      if (hasPw)
      {
         handleClonePassword();
      }
      
      /*
      // hide or show the password panel, depending on user mode (pw auth)
      wpcdpnlPw.setVisible(hasPw);
      wpcdClonePanel.invalidate();
      wpcdClonePanel.repaint();
      */

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, "Clone User", wpcdClonePanel);
      dialog.setVisible(true);
      
      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         dialog = null;
      }
   }
   
   /**
    * Handle the Clone As button, in the Clone As dialog.
    */
   private void handleCloneAs()
   {
      // get the source info from selection
      String sourceName = (String) wpUtm.
                                   getValueAt(wpUsers.getSelectedRow(), 0);
      String sourceDesc = (String) wpUtm.
                                   getValueAt(wpUsers.getSelectedRow(), 1);
      
      // get and check the target user name
      String targetName = validateStringField(wpcdfN, 
                                              "type clone's user name");
      // no name specified, return
      if (targetName == null)
      {
         return;
      }
      
      // check if the user must be cloned along with its password
      String password = null;
      boolean hasPass = wpcdrCp.isEnabled();
      boolean newPass = hasPass && wpcdrNp.isSelected();
      boolean clonePass = hasPass && wpcdrCp.isSelected();

      if (newPass)
      {
         password = validatePassword(wpcdfP1, wpcdfP2);
         // if password validation failed, return
         if (password == null)
         {
            return;
         }
      }
      
      if (clonePass)
      {
         // password must be null if passwords needs to be cloned
         password = null;
      }
      
      byte[] hashPass = null;
      if (password != null)
      {
         hashPass = HashPassword.hashPassword(password);
      }

      // get the ACL option
      boolean withACLs = wpcdcW.isSelected();
      
      // clone the user
      AdminUtils.busy(client);
      // when cloning, if a password is specified, then the pass is set
      // to the passed value; else, password is not altered
      boolean rc = adm.cloneUser(sourceName,
                                 targetName,
                                 withACLs,
                                 hashPass
                                 ); 
      
      AdminUtils.idle(client);
      
      // check the result
      if (!rc)
      {
         // set initial focus on the user name field
         wpcdfN.requestFocusInWindow();

         // alarm
         client.alarm();
         return;
      }
      
      // cloned; clear panel
      client.ok();
      dialog.dispose();
      dialog = null;
      
      // added to the server; clear selection and add to the list
      wpUsm.clearSelection();
      wpUtm.addRow(new TaggedName(targetName, sourceDesc));
      rebuildSearchIndex();
   }

   /**
    * Handle Cancel button press in any of the west-panel dialogs.
    */
   private void handleCancel() 
   {
      // clear error messages
      client.ok();
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
      
      if (extContainer != null)
      {
         extContainer.getParent().remove(extContainer);
      }
   }

   /**
    * Handle Add group button press in the center panel.
    */
   private void handleAddGroup() 
   {
      // clear selection if any
      cpGsm.clearSelection();

      // check the group name field
      String newGroup = cpfN.getText().trim();
      
      if (newGroup.length() == 0)
      {
         // set initial focus on the group name field
         cpfN.requestFocusInWindow();
         
         // alarm
         client.beep("type name of a group");
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // assign the user to this group
      boolean rc = adm.changeGroupUsers(newGroup, new String[] {userName}, 
                                        new String[0]);
      
      if (!rc)
      {
         client.alarm();
         
         // idle
         AdminUtils.idle(client);
         return;
      }
      
      // added; get the group info from the server and add it to the table
      client.ok();
      TaggedName ui = adm.checkGroup(newGroup);
      cpGtm.addRow(ui);
      
      // clear input field
      cpfN.setText("");
      
      // position the cursor for possible continuation
      cpfN.requestFocusInWindow();
      
      // set selection in all users accordingly
      setAllGroupsSelection();

      // idle
      AdminUtils.idle(client);
   }
               
   /**
    * Handle Delete selected button press in the center panel.
    */
   private void handleDeleteSelectedGroups()
   {

      // get the selection
      int[] sel = cpGroups.getSelectedRows();
      
      // safety check
      if (sel.length == 0)
      {
         return;
      }
      
      // ask confirmation
      String quest = sel.length == 1 ?
                     "User being removed from 1 group. Continue?" : 
                     "User being removed from " + 
                     Integer.toString(sel.length) + " groups. Continue?";
      
      if (!AdminUtils.confirmDialog(client, "Removing groups", quest))
      {
         // declined
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // sort the array of selected indices to walk backwards
      Arrays.sort(sel);
      
      String[] groups = new String[sel.length];
      
      // walk the array backwards and remove the groups from the user def 
      for (int i = sel.length - 1; i >= 0; i--)
      {
         groups[i] = (String) cpGtm.getValueAt((sel[i]), 0);
         
         // delete this row
         cpGtm.deleteRow(sel[i]);
      }
            
      UserDef user = adm.getUser(userName);
      ArrayList<String> ugroups = 
         new ArrayList<String>(Arrays.asList(user.groups));
      ugroups.removeAll(Arrays.asList(groups));
      user.groups = ugroups.toArray(new String[0]);
      
      // unassign the user from the groups on the server
      boolean rc = adm.setUser(user);
      
      if (!rc)
      {
         // attempt failed; restore the table
         client.alarm();
         populateGroupsList(userName);
         
         // idle
         AdminUtils.idle(client);
         return;
      }
      
      // refresh the selection in the all users list
      client.ok();
      setAllGroupsSelection();

      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Handle Add selected button press in the east panel.
    */
   private void handleAddSelectedGroups() 
   {
      // scan the whole selection to detect extra lines
      int[] all = epGroups.getSelectedRows();
      
      java.util.List<String> extra = new LinkedList<String>();
      
      for (int j = 0; j < all.length; j++)
      {
         int k = all[j];
         
         String user = (String) epGroups.getValueAt(k, 0);
         if (!cpGtm.isListed(user))
         {
            // add to the list for the server
            extra.add(user);
            
            // add to the table
            TaggedName tn = new TaggedName(user, 
                                           (String)epGroups.getValueAt(k, 1));
            cpGtm.addRow(tn);
         }
      }
      
      // process the additions, if any
      if (!extra.isEmpty())
      {
         String[] additions = extra.toArray(new String[0]);
         
         UserDef user = adm.getUser(userName);
         ArrayList<String> ugroups = 
            new ArrayList<String>(Arrays.asList(user.groups));
         ugroups.addAll(Arrays.asList(additions));
         user.groups = ugroups.toArray(new String[0]);
         
         AdminUtils.busy(client);
         boolean rc = adm.setUser(user);
         AdminUtils.idle(client);
      
         if (!rc)
         {
            client.alarm();
            
            // refresh the users list of center panel
            populateGroupsList(userName);
            return;
         }
         
      }
      
      // disable Add selected button
      client.ok();
      epbA.setEnabled(false);
   }

   /**
    * Handle Filter button press in the east panel.
    * When pressed with some contents in the Search text field, the text
    * becomes the filter. Otherwise the existing filter is removed.
    */
   private void handleFilter()
   {

      // clear selection first
      selecting = true;
      epGsm.clearSelection();
      
      // get and check the filter value
      String filter = epfS.getText().trim();
      
      epScrollTitle.setTitle(filter.length() == 0 ? " All Groups " 
                                                  : " Groups with Filter");
      epScroller.repaint();

      if (filter.length() == 0)
      {
         // remove filtering
         epSorter.setRowFilter(null);
      }
      else
      {
         StringBuilder sb = new StringBuilder(".*");
         for (int i = 0; i < filter.length(); i++)
         {
            char c = filter.charAt(i);
            sb.append("[");
            if (Character.isUpperCase(c))
            {
               sb.append(Character.toLowerCase(c));
               sb.append(c);
            }
            else
            {
               sb.append(c);
               sb.append(Character.toUpperCase(c));
            }
            sb.append("]");
         }
         sb.append(".*");
         
         epSorter.setRowFilter(RowFilter.regexFilter(sb.toString()));
      }
      
      // reshow selection
      selecting = false;
      setAllGroupsSelection();
   }

   /**
    * Create and return default panel containing search elements.
    *
    * @return See above.
    */
   private JPanel buildDefaultFindPanel()
   {
      findUserDefInput = new JTextField();
      findUserDefInput.setToolTipText(
                                  "You can press ENTER to perform search.");

      JPanel optionsPanel = new JPanel();
      optionsPanel.setLayout(new BoxLayout(optionsPanel, BoxLayout.Y_AXIS));
      optionsPanel.add(new JLabel("Enter User Name:"));
      optionsPanel.add(Box.createRigidArea(new Dimension(0,5)));
      optionsPanel.add(findUserDefInput);

      // add Find and Cancel buttons
      JPanel buttonsPanel = new JPanel();
      buttonsPanel.add(findUserCancelBtn);
      buttonsPanel.add(findUserFindBtn);

      JPanel targetPanel = AdminUtils.groupComponents(new Component[][]
      {
         { optionsPanel },
         { buttonsPanel }
      });

      findUserDefInput.addKeyListener(new KeyListener()
      {
         public void keyTyped(KeyEvent keyEvent)
         {
         }

         public void keyPressed(KeyEvent keyEvent)
         {
            if (keyEvent.getKeyCode() == KeyEvent.VK_ENTER)
            {
               String username = findUserDefInput.getText().trim();
               handlePerformFind(username);
            }
         }

         public void keyReleased(KeyEvent keyEvent)
         {
         }
      });

      findUserFindBtn.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            String username = findUserDefInput.getText().trim();
            handlePerformFind(username);
         }
      });

      findUserCancelBtn.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused")
         public void actionPerformed(ActionEvent evt)
         {
            handleCancel();
         }
      });

      return targetPanel;
   }

   /**
    * Display find window. 
    */
   private void handleFind()
   {
      // display modal dialog
      dialog = AdminUtils.modalDialog(client,
                                      "Find User",
                                      findUserAccountPanel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }

   /**
    * Find user by the given username and highlight it into the user list.
    *
    * @param username
    *        User to find.
    */
   private void handlePerformFind(String username)
   {
      Integer idx = userSearchIndex.get(username);
      if (idx == null)
      {
         AdminUtils.errorDialog(client,
                                "Find results",
                                "Cannot find such user.");
      }
      else
      {
         wpUsm.setSelectionInterval(idx, idx);

         JViewport viewport = (JViewport) wpUsers.getParent();

         // This rectangle is relative to the table where the
         // northwest corner of cell (0,0) is always (0,0).
         Rectangle rect = wpUsers.getCellRect(idx, 0, true);

         // The location of the viewport relative to the table
         Point pt = viewport.getViewPosition();

         // Translate the cell location so that it is relative
         // to the view, assuming the northwest corner of the
         // view is (0,0)
         rect.setLocation(rect.x-pt.x, rect.y-pt.y);

         // Scroll the area into view
         viewport.scrollRectToVisible(rect);

         handleCancel();
      }
   }

   /**
    * Rebuild the search index (which helps to find user by username).
    */
   private void rebuildSearchIndex()
   {
      userSearchIndex.clear();
      int count = wpUtm.getRowCount();
      for (int i = 0; i < count; i++)
      {
         userSearchIndex.put((String) wpUtm.getValueAt(i, 0), i);
      }
   }
}