/*
** Module   : AuthModeComboBoxes.java
** Abstract : Class which keeps authentication mode and custom authentication
**            plugin selection combo boxes and provides an interface to them.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description-------------------------------
** 001 SVL 20100916 Created the initial version which keeps auth mode and auth
**                  plugin selection combo boxes.
** 002 SVL 20100920 Add support for the Bulk Auth Mode Update screen.
** 003 EVL 20160223 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.security.*;

import javax.swing.*;
import java.util.*;
import java.awt.event.*;

/**
 * Class which keeps authentication mode and custom authentication plugin
 * selection combo boxes and provides an interface to them. Used into user and
 * group account setting.
 */
public class AuthModeComboBoxes
{
   /** Authentication mode combo box. */
   private JComboBox authModeCombo;

   /** Custom authentication plugin combo box. */
   private JComboBox authPluginsCombo;

   /** Authentication mode label. */
   private JLabel authModeLabel;

   /** Custom authentication plugin label. */
   private JLabel authPluginsLabel;

   /** Map of plugin classes to plugin item objects. */
   private Map<String, AuthPluginItem> pluginsMap;

   /**
    * The string that corresponds "inherit auth mode" option for the user
    * account screen.
    */
   private final static String USER_INHERIT_STRING =
                                             "Use Group or Server Auth Mode";

   /**
    * The string that corresponds "inherit auth mode" option for the group
    * account screen.
    */
   private final static String GROUP_INHERIT_STRING = "Use Server Auth Mode";

   /**
    * Creates and initializes combo boxes.
    *
    * @param adm
    *        Instance of the remote admin exports to use. 
    * @param targetScreen
    *        The target screen for which these combo boxes should be
    *        initialized. 
    */
   public AuthModeComboBoxes(AdminExports adm, TargetScreen targetScreen)
   {
      authModeCombo = new JComboBox();
      authModeCombo.addItem(
            targetScreen == TargetScreen.GROUP_ACCOUNT_SCREEN ?
            GROUP_INHERIT_STRING :
            USER_INHERIT_STRING);
      authModeCombo.addItem("Password Authentication");
      authModeCombo.addItem("Certificate Authentication");
      authModeCombo.addItem("Certificate + Password Authentication");
      authModeCombo.addItem("Custom Authentication");

      pluginsMap = new HashMap<String, AuthPluginItem>();
      authPluginsCombo = new JComboBox();
      TaggedName[] taggedNames = adm.listAuthPlugins();

      if (taggedNames != null)
      {
         for(TaggedName taggedName : taggedNames)
         {
            AuthPlugin plugin = new AuthPlugin(taggedName);
            AuthPluginItem item = new AuthPluginItem(plugin);
            authPluginsCombo.addItem(item);
            pluginsMap.put(plugin.getPluginClass(), item);
         }
      }

      // enable/disable auth mode combo depending on selected auth mode
      authModeCombo.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent e)
         {
            boolean enablePluginsCombo = authModeCombo.getSelectedIndex() ==
                                          SecurityConstants.AUTH_MODE_CUSTOM;
            authPluginsCombo.setEnabled(enablePluginsCombo);

            if (enablePluginsCombo &&
                getAuthPlugin() == null &&
                pluginsMap.size() > 0)
            {
               authPluginsCombo.setSelectedIndex(0);
            }
         }
      });

      if (targetScreen == TargetScreen.BULK_AUTH_MODE_UPDATE)
      {
         authModeLabel    = new JLabel(" Authentication Mode To Set ");
         authPluginsLabel =
                          new JLabel(" Custom Authentication Plugin To Set ");
      }
      else
      {
         authModeLabel    = new JLabel(" Authentication Mode ");
         authPluginsLabel = new JLabel(" Custom Authentication Plugin ");
      }
   }

   /**
    * Get authentication mode combo box.
    *
    * @return  See above.
    */
   public JComboBox getAuthModeCombo()
   {
      return authModeCombo;
   }

   /**
    * Get custom authentication plugin combo box.
    *
    * @return  See above.
    */
   public JComboBox getAuthPluginsCombo()
   {
      return authPluginsCombo;
   }

   /**
    * Get authentication mode label.
    *
    * @return  See above.
    */
   public JLabel getAuthModeLabel()
   {
      return authModeLabel;
   }

   /**
    * Get custom authentication plugin label.
    *
    * @return  See above.
    */
   public JLabel getAuthPluginsLabel()
   {
      return authPluginsLabel;
   }

   /**
    * Enable/disable combo boxes.
    *
    * @param enabled
    *        <code>true</code> to enable combo boxes, <code>false</code> to
    *        disable them.  
    */
   public void setEnabled(boolean enabled)
   {
      authModeCombo.setEnabled(enabled);

      authPluginsCombo.setEnabled(enabled &&
                        getAuthMode() == SecurityConstants.AUTH_MODE_CUSTOM);
   }

   /**
    * Reset the state of combo boxes (set the initial state).
    */
   public void reset()
   {
      authModeCombo.setSelectedIndex(0);
      authPluginsCombo.setSelectedItem(null);
      authPluginsCombo.setEnabled(false);
   }

   /**
    * Get selected authentication mode.
    *
    * @return  See above.
    */
   public int getAuthMode()
   {
      return authModeCombo.getSelectedIndex();
   }

   /**
    * Get selected custom authentication plugin.
    *
    * @return  See above.
    */
   public String getAuthPlugin()
   {
      AuthPluginItem item =
               (AuthPluginItem) authPluginsCombo.getSelectedItem();
      return item != null ? item.authPlugin.getPluginClass() : null;
   }

   /**
    * Set the specified authentication mode as selected into the appropriate
    * combo box.
    *
    * @param mode
    *        Mode to be selected.
    */
   public void setAuthMode(int mode)
   {
      authModeCombo.setSelectedIndex(mode);
      authPluginsCombo.setEnabled(mode == SecurityConstants.AUTH_MODE_CUSTOM);
   }

   /**
    * Set the specified custom authentication plugin as selected into the
    * appropriate combo box.
    *
    * @param authPlugin
    *        Auth plugin to be selected.
    */
   public void setAuthPlugin(String authPlugin)
   {
      authPluginsCombo.setSelectedItem(pluginsMap.get(authPlugin));
   }

   /**
    * Set the selected combo box values in conformity with auth settings of
    * the given user account.
    *
    * @param def
    *        User account, which auth settings should be applied to the combo
    *        boxes.
    */
   public void applyUserSettings(UserDef def)
   {
      authModeCombo.setSelectedIndex(def.mode);
      authPluginsCombo.setSelectedItem(pluginsMap.get(def.authPlugin));
   }

   /**
    * Set the selected combo box values in conformity with auth settings of
    * the given group account.
    *
    * @param def
    *        Group account, which auth settings should be applied to the combo
    *        boxes.
    */
   public void applyGroupSettings(GroupDef def)
   {
      authModeCombo.setSelectedIndex(def.mode);
      authPluginsCombo.setSelectedItem(pluginsMap.get(def.authPlugin));
   }

   /**
    * Updates the "inherit auth mode" option text according to the given mode.
    *
    * @param targetMode
    *        Target mode to set. Specify
    *        <code>TargetScreen.GROUP_ACCOUNT_SCREEN</code> if the "inherit
    *        auth mode" option string should correspond group account settings
    *        or <code>TargetScreen.USER_ACCOUNT_SCREEN</code> if it should
    *        correspond user account settings.
    */
   public void updateInheritString(TargetScreen targetMode)
   {
      authModeCombo.insertItemAt(
            targetMode == TargetScreen.GROUP_ACCOUNT_SCREEN ?
            GROUP_INHERIT_STRING :
            USER_INHERIT_STRING,
            1);
      authModeCombo.removeItemAt(0);
   }

   /**
    * Represents the screen on which combo boxes are displayed.
    */
   public enum TargetScreen
   {
      /** Accounts -&gt; Users -&gt; Add / Edit User screen */
      USER_ACCOUNT_SCREEN,

      /** Accounts -&gt; Groups -&gt; Add / Edit Group screen */
      GROUP_ACCOUNT_SCREEN,

      /** Accounts -&gt; Bulk Auth Mode Update screen */
      BULK_AUTH_MODE_UPDATE
   }

   /**
    * Represents custom authentication plugin as a combo box item. 
    */
   private class AuthPluginItem
   {
      /** Custom authentication plugin class. */
      private AuthPlugin authPlugin;

      /**
       * Creates an item for the given plugin.
       *
       * @param authPlugin
       *        Custom authentication plugin class.
       */
      public AuthPluginItem(AuthPlugin authPlugin)
      {
         this.authPlugin = authPlugin;
      }

      /**
       * Used for rendering into combo box.
       *
       * @return  string representation of the auth plugin item.
       */
      public String toString()
      {
         return authPlugin.getDescription();
      }
   }
}