/*
** Module   : ClientAccountExtension.java
** Abstract : Plugin definition for handling customer-specific extensions.
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 CA  20090515  @42444  Created initial version.
** 002 CA  20090604  @42559  Added resetDialog, which is will reset all widgets
**                           to a default state.
** 003 NVS 20090604  @42601  Moved AccountExtValidationException class to
**                           admin package.
** 004 NVS 20090803  @43517  Added account extension data printing related
**                           methods: getPreviewHeaders(), fillPreviewRow(),
**                           getExtendedHeaders(), fillExtendedRow().
** 005 NVS 20090804  @43518  Simplified extension data printing related methods.
** 006 NVS 20090806  @43555  Simplified plugin interconnections.
** 007 NVS 20091007  @44117  Changed to allow the extension fields in the add
**                           or edit user dialog to be enabled or disabled 
**                           according to granular admin permissions.
** 008 SVL 20100211  @44576  Added buildUserFindPanel() function.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.io.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.table.*;

import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.admin.client.*;
import com.goldencode.p2j.security.*;

/**
 * Plugin-definition for handling customer-specific extensions.
 * 
 * @author CA
 */
public interface ClientAccountExtension
{
   /** Constant indicating the plugin can re-layout the default widgets too.*/
   public static final int LAYOUT_ALL_WIDGETS = 0;
   
   /** 
    * Constant indicating the plugin is responsible only for extension
    * field layout.
    */
   public static final int LAYOUT_EXT_WIDGETS = 1;
   
   /**
    * Get the layout mode of the plugin:
    * <p>
    * {@link #LAYOUT_ALL_WIDGETS}: plugin is responsible to layout all widgets
    * <p>
    * {@link #LAYOUT_EXT_WIDGETS}: plugin is responsible to layout only ext
    * widgets
    *  
    * @return   plugin layout mode.
    */
   public int getLayoutMode();
   
   /**
    * Layout the extension fields; this method can re-layout the default 
    * widgets too, if needed. If <code>userName</code> is provided, it
    * will populate the extension fields with its extension data, if any. 
    * 
    * @param    userName
    *           The user for which extension field layout is needed. If 
    *           <code>null</code>, it means a new user is added.
    * 
    * @param    writable
    *           <code>true</code> if this is for enabled inputs.
    *           <code>false</code> for readonly panel
    *
    * @return   the root panel which will be shown as a modal dialog.
    */
   public JPanel layoutAllWidgets(String userName, boolean writable);
   
   /**
    * If the plugin is in {@link #LAYOUT_EXT_WIDGETS} mode, it must return
    * the container to which extension fields will be added.
    * 
    * @return   a container to which extension fields will be added.
    */
   public Container getExtContainer();
   
   /**
    * Layout the extension fields in the specified container. If 
    * <code>userName</code> is provided, it will populate the extension fields 
    * with its extension data, if any. 
    * 
    * @param    c
    *           the container to which widgets will be added.
    * 
    * @param    userName
    *           The user for which extension field layout is needed. If 
    *           <code>null</code>, it means a new user is added.
    * 
    * @param    writable
    *           <code>true</code> if this is for enabled inputs.
    *           <code>false</code> for readonly panel
    */
   public void layoutExtWidgets(Container c, String userName, boolean writable);
   
   /**
    * Client-side validation of the extension data. This means only following
    * validation is done:
    * <ul>
    *    <li>check if the mandatory widgets have data</li>
    *    <li>check if the numeric, date, etc fields are valid</li>
    *    <li>any other special validation which can be done on client-side</li>
    * </ul>
    * 
    * @param    userName
    *           The user for which extension field validation is needed. If 
    *           <code>null</code>, it means a new user is added.
    * 
    * @return   the extension object which holds extension data
    * 
    * @throws   AccountExtValidationException if client-side validation fails
    */
   public Serializable validateExtensions(String userName)
   throws AccountExtValidationException;
   
   /**
    * Set the current {@link IAccountScreen} and {@link AdminClient} instances 
    * to be used in the plugin.
    * 
    * @param    screen
    *           the current {@link IAccountScreen} instance.
    */
   public void initialize(IAccountScreen screen);
   
   /**
    * Reset all extension widgets to a predefined state.
    */
   public void resetDialog();
   
   /**
    * Returns the print preview table headers for the account extension data.
    * An ampty array disables extension data printout.
    */
   public String[] getPreviewHeaders();
   
   /**
    * Gathers and returns all extra columns info for the specific account
    * as strings. The number and the order of values should correspond to
    * the headers returned by {@link #getPreviewHeaders}.
    *
    * @param subjectId
    *        account name
    */
   public String[] getPreviewRow(String subjectId);
   
   /**
    * Returns the extended print preview table headers for the account
    * extension data. An ampty array disables extended print for the
    * extension data printout.
    */
   public String[] getExtendedHeaders();
   
   /**
    * Returns the extended print preview table rows required for the account
    * extension data per account.
    */
   public int getExtendedRows();
   
   /**
    * Fills the extended print preview table row columns with the account 
    * extension data.
    *
    * @param subjectId
    *        account name
    *
    * @param tm
    *        table model to be used
    *
    * @param row
    *        index of the table model row to start filling from
    *
    * @param column
    *        index of the table model column to start filling from
    */
   public void fillExtendedRow(String subjectId, TableModel tm, int row,
                               int column);

   /**
    * Create and return the find panel for the given find screen handler.
    *
    * @param  findScreen
    *         Target find screen handler. 
    *
    * @return created find panel.
    */
   public JPanel buildUserFindPanel(IFindAccountScreen findScreen);
}