/*
** Module   : TaggedDataTableModel.java
** Abstract : Defines a custom table model. 
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 CA  20090522  @42434  Extracted initial version from AdminScreen. Also 
**                           added setColumnName, to allow column name change.
** 002 CA  20090529  @42492  Added setter for the sort field.
** 003 CA  20090714  @43160  Refactored to work with more than two columns.
** 004 NVS 20090807  @43564  Full table reload needed a better notification.
** 005 SIY 20091211  @44501  Changed access permission to protected for data
**                           member.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client;

import java.util.*;

import com.goldencode.p2j.admin.*;

/**
 * A custom TableModel to feed JTable from an array of TaggedNames.
 */
public class TaggedDataTableModel
extends javax.swing.table.AbstractTableModel
{
   /** Storage for tagged data */
   protected ArrayList<TaggedName> data = new ArrayList<TaggedName>();
   
   /** column names */
   String[] columns = null;
   
   /** flag which indicates if rows should be sorted or not */
   boolean sort = true;

   /**
    * Constructor
    *
    * @param columns
    *        the column names
    */
   public TaggedDataTableModel(String... columns)
   {
      super();
      this.columns = columns;
   }

   /**
    * Returns the number of columns in the model.
    *
    * @return length of {@link #columns} array.
    */
   public int getColumnCount()
   {
      return columns.length;
   }
   
   /**
    * Returns the number of rows in the model.
    *
    * @return the number of rows in the array of tagged names
    */
   public int getRowCount()
   {
      return data.size();
   }
   
   /**
    * Fetches a string from the tagged name instance from the model.
    *
    * @param row
    *        model row index
    *
    * @param col
    *        model column index
    *
    * @return
    *        string which is either name or its tag from the given row 
    */
   public Object getValueAt(int row, int col)
   {
      if (row < 0 || row >= data.size())
      {
         return null;
      }
      
      TaggedName tn = data.get(row);

      return tn.getValueAt(col);
   }
   
   /**
    * Returns the column names for this model.
    *
    * @param col
    *        model column index
    *
    * @return column name
    */
   public String getColumnName(int col)
   {
      return col < 0 || col >= columns.length ? null : columns[col];
   }
   
   /**
    * Change the specified column's name.
    *
    * @param col
    *        model column index
    * @param name
    *        new model column name
    */
   public void setColumnName(int col, String name)
   {
      if (col < 0 || col >= columns.length)
      {
         return;
      }
      
      columns[col] = name;
   }
   
   /**
    * Returns the column class for this model.
    *
    * @param col
    *        model column index
    *
    * @return column class, which is always <code>String.class</code>
    */
   public Class<?> getColumnClass(int col)
   {
      return String.class;
   }
   
   /**
    * Deletes all rows.
    */
   public void clear()
   {
      int rows = data.size();
      if (rows > 0)
      {
         // notify listeners
         fireTableRowsDeleted(0, rows - 1);
         
         // clear the storage
         data.clear();
      }
   }
   
   /**
    * Clears and then reloads the table from a data source.
    *
    * @param list
    *        array of tagged names to load into the model
    */
   public void loadAll(TaggedName[] list)
   {
      // clear first
      int rows = data.size();
      if (rows > 0)
      {
         clear();
      }
      
      // deal with the input
      if (list.length > 0)
      {
         // sort the input
         if (sort)
         {
            Arrays.sort(list);
         }
         
         // stuff the table next
         for (int i = 0; i < list.length; i++)
         {
            data.add(list[i]);
         }
      }
      
      // notify listeners about the data change
      fireTableDataChanged();
   }
   
   /**
    * Checks to see if a name is in the list.
    *
    * @param   name
    *          search argument
    *
    * @return  <code>true</code> if name is in the list
    */
   public boolean isListed(String name)
   {
      if (name == null)
      {
         return false;
      }
      
      return data.contains(new TaggedName(name, ""));
   }
   
   /**
    * Gets the index of the specified name, if contained by the list. Else, it
    * will return -1.
    *
    * @param   name
    *          search argument
    *
    * @return  row index if found; else, -1.
    */
   public int getRowIndex(String name)
   {
      if (name == null)
      {
         return -1;
      }
      
      return data.indexOf(new TaggedName(name, ""));
   }

   /**
    * Inserts a row into the table maintaining the sort order.
    *
    * @param item
    *        instance of a tagged name to add to the model
    */
   public void addRow(TaggedName item)
   {
      // add the item to the storage
      data.add(item);
      
      // resort
      if (sort)
      {
         Collections.sort(data);
      }
      
      // notify listeners about new row
      fireTableRowsInserted(data.size() - 1, data.size() - 1);

      // notify listeners about rows shuffle
      fireTableRowsUpdated(0, data.size() - 1);
   }
   
   /**
    * Deletes a row from the table.
    *
    * @param index
    *        model row index for the row being deleted
    */
   public void deleteRow(int index)
   {
      // add the item to the storage
      data.remove(index);
      
      // notify listeners about deleted row
      fireTableRowsDeleted(index, index);
   }
   
   /**
    * Refreshes a row in the table.
    *
    * @param item
    *        instance of a replacement tagged name
    *
    * @param index
    *        model row index for the row being replaced
    */
   public void refreshRow(TaggedName item, int index)
   {
      // replace the item
      data.set(index, item);
      
      // notify listeners about updated row
      fireTableRowsUpdated(index, index);
   }

   /**
    * Set the new sort type of this table model.
    * 
    * @param sort
    *        <code>true</code> if the records should be sorted.
    */
   public void setSort(boolean sort)
   {
      this.sort = sort;
   }

}