/*
** Module   : ConfigFieldSetterAspect.java
** Abstract : An aspect which intercepts field assignments in widget config classes.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140926 Created initial version.
** 002 CA  20141207 Transient fields must not be tracked for dirty changes.
** 003 HC  20141207 Implemented synchronization of linked widget coordinate attributes.
** 004 CA  20150722 Added a new exclusion in beforeSetField: LabelConfig.setDefaults is called 
**                  only from c'tors.
** 005 EVL 20160223 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.aspects.ui;

import org.aspectj.lang.*;
import org.aspectj.lang.annotation.*;

import com.goldencode.p2j.ui.*;

/**
 * Aspect which intercepts the assignment of public fields from the 
 * {@code com.goldencode.p2j.ui.*Config} classes.
 * <p>
 * Only the client-side uses this aspect.  All calls from the server-side are ignored.  The field
 * assignments during {@link WidgetConfig} instantiation are also ignored.
 * <p>
 * AspectJ Point Cuts are called during custom setters and only when the field is set outside of 
 * this code:
 * <ol>
 *   <li>{@link WidgetConfig#WidgetConfig constructors}</li>
 *   <li>{@link WidgetConfig#readExternal} and {@link WidgetConfig#writeExternal}</li>
 *   <li>{@link WidgetConfig#applyConfig}</li>
 *   <li>serve-side code</li>
 *   <li><code>com.goldencode.p2j.ui</code> package</li>
 * </ol>
 * While these methods are executed, changes must not be tracked; field changes can be tracked
 * only when the client-side UI runtime changes it explicitly.
 * <p>
 * When collecting the changes, {@link ClientConfigManager#getConfigUpdates} will use the IDs of
 * the collected {@link WidgetConfig} instances, and only the instance with an associated active
 * config will be used.
 */
@Aspect()
public class ConfigFieldSetterAspect
{
   /**
    * Aspect which intercepts the assignment of public fields from the 
    * {@code com.goldencode.p2j.ui.*Config} classes.
    * 
    * @param    cfg
    *           The widget configuration.
    * @param    sp
    *           The details about the static part of this join point.
    */
   @Before("set(public !transient * com.goldencode.p2j.ui.*Config.*) " +
           "&& target(cfg) " +
           "&& (within(com.goldencode.p2j.ui..*)" +
           "    || within(com.goldencode.p2j.aspects.ui.SyncCoordinatesAspect))" +
           "&& !withincode(void readExternal(..)) " +
           "&& !withincode(void applyConfig(..)) " +
           "&& !withincode(private void com.goldencode.p2j.ui.LabelConfig.setDefaults(..)) " +
           "&& !withincode(com.goldencode.p2j.ui.*Config.new(..)) " +
           "&& !within(com.goldencode.p2j.ui.GenericFrame) " +
           "&& !within(com.goldencode.p2j.ui.*Widget) " +
           "&& !within(com.goldencode.p2j.ui.*Entity) ")
   public void beforeSetField(WidgetConfig cfg, JoinPoint.StaticPart sp)
   throws Throwable
   {
      ConfigManager mgr = ConfigManager.getInstance();
      String field = sp.getSignature().getName();

      mgr.addDirtyConfig(cfg, field);
   }
}