/*
** Module   : Configuration.java
** Abstract : Master configuration object
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041217   @19157 Created initial version.
** 002 ECF 20050120   @19359 Added support for individual parameters with
**                           variable substitution. New public methods:
**                           - getParameter (two flavors)
**                           - putParameter
** 003 GES 20050223   @19931 Added a public utility method to normalize
**                           a passed in filename to the project root so
**                           that tools can refer to the same file from
**                           different current directories without having
**                           to use absolute filenames.
** 004 ECF 20050225   @19963 Modified toFile method to canonicalize file.
** 005 GES 20050617   @21507 Moved sweep() to StringHelper.
** 006 GES 20070411   @32924 Forced a default value of "." for P2J_HOME
**                           since this is essentially the only value
**                           we ever use anyway. This is just safer and
**                           easier than forcing everyone to always
**                           specify a value.
** 007 GES 20070717   @34580 Exposed the ${SYMBOL} style parameter
**                           substitution as an external service. Any of
**                           the parameters in the parmMap will be matched
**                           and replaced as is normally done during
**                           putParameter(). This allows other config
**                           files to act like and reference the main
**                           system config.
** 008 GES 20090424   @41927 Import change.
** 009 LMR 20110105          Replaced Apache Digester usage with direct SAX
**                           XML processing using and the tools in 
**                           GenericSAXHandler and Mapper; performed other
**                           refactoring: 
**                             -renamed load() to getSchemaConfig()
**                             -removed the 'key' parameter and throws clause
**                              of getSchemaConfig()
**                             -inlined code in initialize() in the static
**                              initializer block and removed initialize()
**                             -added main() method to test the parsing of
**                              the XML file received in args[0]
** 010 GES 20110712          Added a helper to return back parameter values
**                           already converted to the proper type.
** 011 ECF 20130125          Added schema metadata configuration. Added generics.
** 012 SVL 20130624          Use singleton instead of static initialization. Added server run time
**                           mode.
** 013 CA  20131024          The configuration file must be loaded at runtime mode too.
** 014 SVL 20140320          Load configuration from resource in runtime mode.
** 015 IAS 20160331          Fix the instance field initialization
** 015 IAS 20160509          Fixed DDL dialects processing in case of more than one 
**                           default namespace.
** 016 OM  20170113          Unify settings for windows and unix/linux in p2j.cfg.xml.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.cfg;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import com.goldencode.util.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.schema.SchemaConfig.*;
import com.goldencode.p2j.schema.SchemaConfig.NsConfig.*;

/**
 * Master configuration object which loads subsystem configurations. The
 * main configuration and all subsystem configurations are stored as XML
 * documents, which are expected to be found in a <code>cfg</code>
 * subdirectory of the P2J home path. The home path is defined by the
 * required <code>P2J_HOME</code> system property.
 * <p>
 * Subsystem configurations are java objects returned by the {@link 
 * #getSchemaConfig()} method. They are created from XML configuration 
 * information by a SAX parser using the <code>GenericSAXHandler</code> and 
 * <code>Mapper</code> classes in package <code>com.goldencode.p2j.util
 * </code>. XML configuration data is stored in a file named <code>p2j.cfg.xml
 * </code>.
 * <p>
 * This class also provides services to get and store individual parameter
 * entries which are read from the main configuration file upon initialization
 * (from cfg/global/parameter XML elements). 
 * <p>Variable substitution is supported in these parameters, in the form 
 * <code>${key_name}</code>, where <code>key_name</code> is the name of a 
 * parameter previously defined by this mechanism. The replacement value 
 * applied to the substitution variable is the value of the parameter 
 * identified by <code>key_name</code>. For example,
 * <pre>
 *    &lt;cfg&gt;
 *      &lt;global&gt;
 *         &lt;parameter name="user.home" value="/home/user" /&gt;
 *         &lt;parameter name="subdir"    value="${user.home}/mysubdir" /&gt;
 *      &lt;/global&gt;
 *      ...
 *    &lt;/cfg&gt;
 * </pre>
 * results in the following parameter definitions after substitution:
 * <pre>
 *    user.home = /home/user
 *    subdir    = /home/user/mysubdir
 * </pre>
 * <strong>Note:</strong> the following restrictions apply to this
 * mechanism:
 * <ul>
 * <li><code>P2J_HOME</code> is a reserved parameter key. It may not be 
 *     defined in the configuration file, but it is available for use as a
 *     substitution variable.</li>
 * <li>The variable name <code>key_name</code> must not include any
 *     characters which have special syntax meaning in regular expressions.
 *     Regular expressions are used internally for the replacement; special
 *     characters in <code>key_name</code> are not explicitly escaped,
 *     which could cause unexpected results.</li>
 * </ul>
 */
public final class Configuration
{
   /** Constant denoting the <code>&lt;cfg&gt;</code> XML tag */
   public static final String TAG_CFG = "cfg";

   /** Constant denoting the <code>&lt;global&gt;</code> XML tag */
   public static final String TAG_GLOBAL = "global";

   /** Constant denoting the <code>&lt;parameter&gt;</code> XML tag */
   public static final String TAG_PARAMETER = "parameter";
   
   /** Constant denoting the <code>&lt;schema&gt;</code> XML tag */
   public static final String TAG_SCHEMA = "schema";
   
   /** Constant denoting the <code>&lt;namespace&gt;</code> XML tag */
   public static final String TAG_NAMESPACE = "namespace";
   
   /** Constant denoting the <code>&lt;dialect-specific&gt;</code> XML tag */
   public static final String TAG_DIALECTSPECIFIC = "dialect-specific";
   
   /** Constant denoting the <code>&lt;metadata&gt;</code> XML tag */
   public static final String TAG_METADATA = "metadata";
   
   /** Constant denoting the <code>&lt;table&gt;</code> XML tag */
   public static final String TAG_TABLE = "table";
   
   /** System property name for P2J home path */
   private static final String PROP_HOME_DIR = "P2J_HOME";
   
   /** 
    * System property name for P2J file separator used internally. We have chosen '/', same as
    * for Linux/UNIX, but it could have been any other char as long as it does not interfere with
    * UNIX and Windows file systems. 
    */
   public static final char P2J_FILE_SEP = '/';
   
   /** 
    * System property name for P2J path separator used internally. We have chosen ':', same as
    * for Linux/UNIX, but it could have been any other char as long as it does not interfere with
    * UNIX and Windows file systems. 
    */
   public static final char P2J_PATH_SEP = ':';
   
   /** Configuration subdirectory name */
   private static final String CFG_DIR = "cfg";
   
   /** Configuration file name extension */
   private static final String CFG_EXTENSION = ".cfg.xml";
   
   /** Default configuration filename */
   private static final String DEF_CFG_FILE = "p2j.cfg.xml";
   
   /** Singleton instance. */
   private static volatile Configuration instance;
   
   /**
    * If <code>true</code> then it is a server run time mode when configuration is NOT read from
    * the configuration directory.
    */
   private static boolean runtimeConfig = false;
   
   /** Map of parameters */
   private final Map<String, String> parmMap = new LinkedHashMap<>();
   
   /** Factory providing the SAX parser instance that will be used to parse the config file. */
   private SAXParserFactory saxParserFactory;
   
   /** The parser instance. */
   private SAXParser saxParser;
   
   /**
    * The generic SAX handler that will be used by the parser after it's
    * configured to work with P2J config files.
    */
   private GenericSAXHandler saxHandler;
   
   /**
    * Holds a set of configuration parameter names whose values are OS dependent. These are
    * related to the system we run now, and are invisible for original P4GL code.
    * 
    * The values from {@code p2j.cfg.xml} are encoded with internal {@code P2J_FILE_SEP} and
    * {@code P2J_PATH_SEP} separators, but they need to be switched when configuration file is
    * loaded, on the fly, to actual separators for the OS we are running on.
    * 
    * We exclude the {@code PROP_HOME_DIR} from this list because it is internally computed and
    * the result uses the correct path separator.
    */
   private static final Set<String> OS_DEPENDENT_VALUES = new HashSet<>();
   static {
      String[] depVal =
      {
         "basepath",                // path used for computing relative paths for procedure lookup 
         "callgraph-db-folder",     // the storage directory for DB graph
         "codenames",               // the matching list for source code name conversion file
         "datanames",               // the matching list for database name conversion file
         "dmo_index_merge",         // dmo_index_merge.xml file for merging DMO definitions
         "dynamic-append",          // set of path to dynamic append to [propath]
         "dynamic-prepend",         // set of path to dynamic prepend to [propath]
         "error-dest",              // file used for preprocessor error messages
         "keyword-ignore",          // ignore keywords file (the -k parameter on startup of P4GL)
         "matchlist",               // project file holding the match list parameter keys
         "merge-dmo-root",          // location of generated .hbm files 
         "name_map_merge",          // file storing the name map to be merged
         "oo-skeleton-path",        // location for the non-project class/interface 4GL code
         "output-root",             // destination of generated files from current project
         "p2j_rules", "P2J_RULES",  // location of TRPL rules in current file-system
         "patpath",                 // set of paths where configuration files can be found
         "propath",                 // set of paths where source procedure files can be found 
         "registry",                // location of the registry within the project
         "rootlist",                // filename for the XML registry
         "warn-dest",               // file used for preprocessor warning messages
      };
      // do not add these:
      //    P2J_HOME - the root of the P2J project, it is automatically configured and added 
      //    [all boolean values]
      //    pkgroot        - java package format
      //    merge-point    - FWD directory path
      //    path-separator - legacy path separator, must be kept as original
      //    file-separator - legacy file separator, must be kept as original
      //    date-order     - date order (Eg. DMY or MDY) from legacy machine
      //    windowing-year - windowing year for 2 digits year
      //    opsys          - the legacy OS
      //    winsys         - WINDOW-SYSTEM preprocessor built-in variable
      //    wsoptions      - list of options used by WebSpeed Embedded SpeedScript Preprocessor
      //    wsmode         - WebSpeed's compatibility mode name
      //    marker         - marker character. ('-marker n' P4GL command line parameter)
      //    include-spec   - the shell pattern used to filter the list of files
      //    number-group-sep     - character used as group separator in numbers
      //    number-decimal-sep   - character used as decimal separator in numbers
      //    source-code-version  - OpenEdge compatibility version
      //    graph-node-filter    - the fully qualified name of the callgraph filter class
      Collections.addAll(OS_DEPENDENT_VALUES, depVal);
   }

   /**
    * Private default constructor to prevent creation of instances of this class.
    */
   private Configuration()
   {
   }

   /**
    * Determines whether the configuration is in conversion mode (configuration is read from the
    * configuration directory) or server run time mode.
    *
    * @return <code>true</code> if it is a server run time mode (when configuration is NOT read
    *         from the configuration directory).
    */
   public static boolean isRuntimeConfig()
   {
      return runtimeConfig;
   }

   /**
    * Set whether the configuration should be in conversion mode (configuration is read from the
    * configuration directory) or server run time mode.
    *
    * @param isRuntime
    *        <code>true</code> if it is a server run time mode (when configuration is NOT read
    *        from the configuration directory).
    *
    * @throws IllegalStateException
    *         If the configuration instance was already initialized.
    */
   public static void setRuntimeConfig(boolean isRuntime)
   {
      synchronized (Configuration.class)
      {
         if (instance != null)
         {
            throw new IllegalStateException("Configuration was already created!");
         }
         runtimeConfig = isRuntime;
      }
   }

   /**
    * Return a schema configuration object. The configuration data 
    * itself is in the default XML configuration file <code>p2j.cfg.xml</code>
    * that was parsed when this class was loaded.
    *
    * @return  The configuration object loaded from XML.
    */
   public static SchemaConfig getSchemaConfig()
   {
      return (SchemaConfig) getInstance().saxHandler.getGlobalObjects(TAG_SCHEMA).get(0);
   }
   
   /**
    * Get the configuration parameter with the specified key. This may have
    * been set in the main configuration file, or added later using the
    * {@link #putParameter} method.
    *
    * @param   key
    *          Name of parameter to retrieve.
    *
    * @return  The value of the specified parameter, or <code>null</code> if
    *          <code>key</code> is an invalid identifier.
    */
   public static String getParameter(String key)
   {
      return getParameter(key, null);
   }
   
   /**
    * Get the configuration parameter with the specified key. This may have been set in the main
    * configuration file, or added later using the {@link #putParameter} method. Values for
    * parameters known to be OS filesystem dependent (files, paths and their lists) will be
    * returned already converted to current OS.
    *
    * @param   key
    *          Name of parameter to retrieve.
    * @param   defaultValue
    *          Value to be returned if {@code key} cannot be found in the map of parameters.
    *          May be {@code null}.
    *
    * @return  The value of the specified parameter, or {@code defaultValue} if {@code key} is not
    *          found.
    */
   public static String getParameter(String key, String defaultValue)
   {
      String value = getInstance().parmMap.get(key);
      
      return (value != null ? value : defaultValue);
   }
   
   /**
    * Get the configuration parameter with the specified key as a boolean value. This may have
    * been set in the main configuration file, or added later using the {@link #putParameter}
    * method.  Values for parameters known to be OS filesystem dependent (files, paths and their
    * lists) will be returned already converted to current OS. 
    *
    * @param   key
    *          Name of parameter to retrieve.
    * @param   defaultValue
    *          Value to be returned if {@code key} cannot be found in the map of parameters.
    *
    * @return  The value of the specified parameter, or {@code defaultValue} if {@code key} is not
    *          found.
    */
   public static boolean getParameter(String key, boolean defaultValue)
   {
      String value = getInstance().parmMap.get(key);
      
      return (value != null ? "true".equals(value) : defaultValue);
   }
   
   /**
    * Associate a configuration parameter value with {@code key}. Variable substitution is
    * performed on {@code value} before it is added to the map.
    * <p>
    * In the case of OS dependent values, they are adjusted to match the current OS (operation
    * only needed for Windows OS. For Linux/Unix this operation is a no-op because the neutral
    * separators were chosen to match the separators for them).
    *
    * @param   key
    *          A name which uniquely identifies the parameter.
    * @param   value
    *          The parameter's value.
    *
    * @return  The text of {@code value} after variable substitution has been performed.
    *
    * @throws  ConfigurationException
    *          if {@code key} already identifies a configuration parameter.
    */
   private String putParameter(String key, String value)
   throws ConfigurationException
   {
      if (key == null || value == null)
      {
         throw new ConfigurationException("Null parameter key or value");
      }
      
      if (parmMap.containsKey(key))
      {
         throw new ConfigurationException("Duplicate parameter: " + key);
      }
      
      // for parameters known to be OS dependent, switch from (neutral) to current OS separators
      if (OS_DEPENDENT_VALUES.contains(key))
      {
         value = value.replace(P2J_FILE_SEP, File.separatorChar);
         value = value.replace(P2J_PATH_SEP, File.pathSeparatorChar);
      }
      
      // handle variable substitution
      value = substituteParms(value);
      
      parmMap.put(key, value);
      
      return value;
   }
   
   /**
    * Computes missing derived parameter values based on previously set ones. 
    */
   private void computeDerivedValues()
   {
      // default is Unix OS, if not set, add it now
      String value = parmMap.computeIfAbsent("opsys", x -> "UNIX");
      boolean isWin = value.equalsIgnoreCase("WIN32");
      
      // compute defaults for derived parameter values
      parmMap.computeIfAbsent("file-separator", x -> String.valueOf(isWin ? '\\' : P2J_FILE_SEP));
      parmMap.computeIfAbsent("path-separator", x -> String.valueOf(isWin ? ';' : P2J_PATH_SEP));
      parmMap.computeIfAbsent("case-sensitive", x -> String.valueOf(!isWin));
      parmMap.computeIfAbsent("unix-escapes",   x -> String.valueOf(!isWin));
      parmMap.computeIfAbsent("winsys",         x -> (isWin ? "MS-WIN95" : "TTY"));
   }
   
   /**
    * Variable substitution is performed on <code>value</code> before it is
    * returned.  Any portions of the string matching a ${SYMBOL} syntax will
    * have the parameter map searched for SYMBOL. If it is found, the value
    * of that parameter will be inserted in place of ${SYMBOL}.
    * <p>
    * Replacements occur in the same order they were defined in the
    * main configuration file, and then in the order in which they were
    * added by external callers of this method.
    *
    * @param   value
    *          The value on which to operate.
    *
    * @return  The text of <code>value</code> after variable substitution
    *          has been performed.
    */
   public String substituteParms(String value)
   {
      return StringHelper.sweep(value, parmMap);
   }
   
   /**
    * Get the P2J home path defined for the application or "." if no such
    * path is defined.
    *
    * @return   P2J home path or "." if the system property has not been 
    *           defined.
    *
    * @throws   ConfigurationException
    *           If there is any problem with normalizing the specified path.
    */
   public static String home()
   throws ConfigurationException
   {
      // Check for the P2J_HOME system property.
      String home = System.getProperty(PROP_HOME_DIR, ".");
      
      // Normalize and verify the path.
      File dir = new File(home);
      home = normalizePath(dir, true, true);
      
      return home;
   }
   
   /**
    * If <code>path</code> represents a relative path, generate a
    * <code>File</code> object for this path, relative to the home directory
    * defined for the application. If <code>filename</code> is an absolute
    * path, simply use it to construct the <code>File</code>.
    *
    * @param   path
    *          A relative or absolute file or directory path.
    *
    * @throws  ConfigurationException
    *          if <code>path</code> is relative, but no home path has been
    *          defined for the application.
    */
   public static File toFile(String path)
   throws ConfigurationException
   {
      File file = new File(path);
      if (!file.isAbsolute())
      {
         file = new File(home(), path);
         try
         {
            file = file.getCanonicalFile();
         }
         catch (IOException exc)
         {
         }
      }
      
      return file;
   }
   
   /**
    * Converts an absolute or relative filename (relative to the current
    * directory) into a filename that is relative to the project home. This 
    * allows tools referencing the same file from different current
    * directories to reference the same file, even though the pathing may 
    * not appear to be an exact match.
    *
    * @param    filename
    *           Absolute or current directory relative filename.
    *
    * @return   Filename normalized to the project root OR the original
    *           filename if the file exists outside of the project directory
    *           tree (and thus cannot be normalized).
    */
   public static String normalizeFilename(String filename)
   {
      File   abnormal = new File(filename); 
      String absolute;
      String home;
      
      try
      {
         home = home();
      }
      catch (ConfigurationException cfg)
      {
         // we are done, there is no normalization to be done if we can't find the project root
         return filename;
      }
      
      try
      {
         // make best effort to clean up the path
         absolute = abnormal.getCanonicalPath();
      }
      catch (IOException exc)
      {
         // fallback plan if canonical doesn't work
         absolute = abnormal.getAbsolutePath();
      }
      
      // make sure this filename is in the project directory tree
      if (absolute.startsWith(home))
      {
         // all is well, we can normalize this path
         return "." + absolute.substring(home.length());
      }
      
      // the path can't be normalized (because it exists outside of the
      // project directory tree, so we can only return the original filename
      return filename;
   }

   /**
    * Get the singleton instance of this class.
    *
    * @return  singleton instance of this class.
    */
   public static Configuration getInstance()
   {
      if (instance == null)
      {
         synchronized (Configuration.class)
         {
            if (instance == null)
               instance = createConfiguration();
         }
      }
      return instance;
   }

   /**
    * Create a Configuration instance.
    * 
    * @return Configuration instance
    */
   private static Configuration createConfiguration()
   {
      Configuration config = new Configuration();

      try
      {
         // always start with P2J_HOME parameter; this is reserved
         config.parmMap.put(PROP_HOME_DIR, home());

         InputStream cfgStream;

         if (isRuntimeConfig())
         {
            StringBuilder path = new StringBuilder(CFG_DIR);
            path.append("/");
            path.append(DEF_CFG_FILE);

            ClassLoader cl = ClassLoader.getSystemClassLoader();
            cfgStream = cl.getResourceAsStream(path.toString());

            if (cfgStream == null)
            {
               throw new ConfigurationException("Cannot find resource " + path);
            }
         }
         else
         {
            File cfgFile = getConfigFile(getConfigDirectory(), DEF_CFG_FILE);
            cfgStream = new FileInputStream(cfgFile);
         }
         
         config.saxParserFactory = SAXParserFactory.newInstance();
         config.saxParser = config.saxParserFactory.newSAXParser();
         config.saxHandler = new GenericSAXHandler();
         
         // Prepare the generic handler.
         setupHandler(config.saxHandler);
         
         // Parse the file.
         config.loadXml(cfgStream);
         
         // map parameters by their names, which should be unique
         List<?> params = config.saxHandler.getGlobalObjects(TAG_PARAMETER);
         for (Object o : params)
         {
            Parameter p = (Parameter) o;
            config.putParameter(p.getName(), p.getValue());
         }
         
         // if needed, compute derived values
         config.computeDerivedValues();
      }
      catch (Exception exc)
      {
         throw new RuntimeException(new ConfigurationException(
               "Error loading " + DEF_CFG_FILE + " configuration", exc));
      }
      return config;
   }

   /**
    * Get the configuration directory, which must be a <code>cfg</code>
    * subdirectory beneath the P2J home path.
    *
    * @return  Configuration subdirectory.
    *
    * @throws  ConfigurationException
    *          if configuration directory does not exist or is not a
    *          directory.
    */
   private static File getConfigDirectory()
   throws ConfigurationException
   {
      File dir = new File(home(), CFG_DIR);
      normalizePath(dir, true, true);
      
      return dir;
   }
   
   /**
    * Get the configuration file for the subsystem associated with
    * <code>root</code>, which is either <code>&lt;root&gt;.cfg.xml</code>
    * or, if this file is missing, is the default configuration file,
    * <code>p2j.cfg.xml</code>.
    *
    * @param   dir
    *          Configuration directory.
    * @param   root
    *          Name associated with the target subsystem.
    *
    * @return  Configuration file.
    *
    * @throws  ConfigurationException
    *          if configuration file cannot be found.
    */
   private static File getConfigFile(File dir, String root)
   throws ConfigurationException
   {
      File file = new File(dir, root + CFG_EXTENSION);
      
      if (!file.exists())
      {
         file = new File(dir, DEF_CFG_FILE);
      }
      
      if (!file.exists())
      {
         throw new ConfigurationException("No data file found for " + root + " configuration");
      }
      
      return file;
   }
   
   /**
    * Normalize and optionally verify the specified file system object.
    * <p>
    * Normalizing the path converts relative paths to absolute paths and attempts to remove any
    * redundant, relative constructs, such as references to the current directory ({@code .}) and
    * to the parent directory ({@code ..}).
    * <p>
    * Verifying the path checks that the object exists in the file system and that it is the
    * expected type (file vs. directory).
    *
    * @param   file
    *          File whose path is to be manipulated.
    * @param   verify
    *          {@code true} to verify the path, else {@code false}.
    * @param   isDirectory
    *          {@code true} if {@code file} is supposed to represent a directory, {@code false}
    *          if it should represent a file. This argument is only relevant if {@code verify} is
    *          set to {@code true}.
    *
    * @throws  ConfigurationException
    *          if verification fails.
    */
   private static String normalizePath(File file, boolean verify, boolean isDirectory)
   throws ConfigurationException
   {
      String path = null;
      
      try
      {
         path = file.getCanonicalPath();
      }
      catch (IOException exc)
      {
         path = file.getAbsolutePath();
      }
      
      if (verify)
      {
         if (!file.exists())
         {
            throw new ConfigurationException("'" + path + "' does not exist");
         }
         
         if (file.isDirectory() != isDirectory)
         {
            String type = (isDirectory ? "directory" : "file");
            throw new ConfigurationException("'" + path + "' is not a " + type);
         }
      }
      
      return path;
   }
   
   /**
    * Prepare the general purpose SAX handler for working with P2J XML config 
    * file syntax.  
    * 
    * @param saxHandler
    *        The SAX handler to be configured.
    */
   private static void setupHandler(GenericSAXHandler saxHandler)
   {
      // XML/object mappers:
      saxHandler.installMapper(TAG_PARAMETER, new Mapper<>(Parameter.class));
      saxHandler.installMapper(TAG_SCHEMA, new Mapper<SchemaConfig>(SchemaConfig.class)
         {
            @Override
            public void mergeSubObj(SchemaConfig s, Object c)
            {
               checkSubObjTypes(c, new Class<?>[]{NsConfig.class, Metadata.class});
               if (c instanceof NsConfig)
               {
                  s.addNamespace((NsConfig)c);
               }
               else
               {
                  s.setMetadata((Metadata) c);
               }
            }
         });
      
      saxHandler.installMapper(TAG_NAMESPACE, 
         new Mapper<NsConfig>(NsConfig.class)
         {
            @Override
            public void mergeSubObj(NsConfig ns, Object c)
            {
               checkSubObjTypes(c, new Class<?>[]{Parameter.class, DialectParam.class});
               if (c instanceof Parameter)
               {
                  ns.addParameter((Parameter)c);
               }
               else
               {
                  ns.addDialectParameter((DialectParam)c);
               }
            }
         });
  
      saxHandler.installMapper(TAG_DIALECTSPECIFIC, new Mapper<DialectParam>(DialectParam.class)
         {
            @Override
            public void mergeSubObj(DialectParam d, Object c)
            {
               checkSubObjType(c, Parameter.class);
               d.addParameter((Parameter)c);
            }
         });
      
      // throwaway class used just to map the <table> tag so that we can
      // add a table name to the Metadata object in the Mapper below
      final class Table
      {
         private String name = null;
         
         public String getName()
         {
            return name;
         }
         
         public void setName(String name)
         {
            this.name = name;
         }
      }
      
      saxHandler.installMapper(TAG_METADATA, new Mapper<Metadata>(Metadata.class)
         {
            @Override
            public void mergeSubObj(Metadata m, Object c)
            {
               checkSubObjType(c, Table.class);
               m.addTable(((Table) c).getName());
            }
         });
      
      saxHandler.installMapper(TAG_TABLE, new Mapper<>(Table.class));
      
      // tags that don't map to objects:
      saxHandler.setUnhandledTags(new String[] {TAG_CFG, TAG_GLOBAL});

      // constraints:
      saxHandler.setConstraints(TAG_PARAMETER, 
                                new String[]{"name", "value"}, 
                                null, 
                                new String[]{}, 
                                new String[]{});
      
      saxHandler.setConstraints(TAG_NAMESPACE, 
         new String[]{"name", "importFile", "xmlFile"}, 
         new String[]{"default", "primary"}, 
         null, 
         new String[]{
                      TAG_PARAMETER, 
                      TAG_DIALECTSPECIFIC
                     });
      
      saxHandler.setConstraints(TAG_DIALECTSPECIFIC, 
         new String[]{"name"}, 
         new String[]{}, 
         new String[]{TAG_PARAMETER}, 
         new String[]{});
      
      saxHandler.setConstraints(TAG_METADATA,
         new String[]{"name"},
         null,
         null,
         new String[]{"table"});
      
      saxHandler.setConstraints(TAG_TABLE,
         new String[]{"name"},
         null,
         null,
         null);
   }
   
   /**
    * Parses a given XML data stream using the SAX parser and handler that were
    * already prepared. The objects built in the process are held in the 
    * handler and can be fetched by XML tag name with {@link 
    * GenericSAXHandler#getGlobalObjects(String)}.  
    * 
    * @param  cfgStream
    *         Input stream with data to be parsed.
    *         
    * @throws IOException 
    *         If any IO errors occur in reading the file.
    * @throws SAXException 
    *         If any SAX errors occur during processing.
    * @throws RuntimeException
    *         wrapping a <code>ConfigurationException</code> if any error 
    *         occurs reading the XML data or creating the configuration 
    *         object. This includes missing configuration file or file format
    *         constraint violations detected during parsing.
    *              
    * @see    GenericSAXHandler#getGlobalObjects(String)
    */
   private void loadXml(InputStream cfgStream)
   throws SAXException, 
          IOException,
          RuntimeException
   {
      try
      {
         this.saxParser.reset();
         this.saxHandler.resetState();
         this.saxParser.parse(new InputSource(cfgStream), this.saxHandler);
      }
      finally
      {
         cfgStream.close();
      }
   }
   
   /**
    * Simple schema loader test: the XML file passed as an argument is parsed
    * and the resulting <code>SchemaConfig</code> together with all global
    * parameters are printed to standard output. Variable substitution is not
    * performed on parameters, their values are shown as they were read from 
    * XML. 
    * 
    * @param  args
    *         The first element is the absolute path of the XML file to parse.
    *         
    * @throws IOException
    *         If IO errors occur. 
    * @throws SAXException
    *         If the SAX parser encounters problems while processing the XML
    *         content.
    * @throws RuntimeException
    *         wrapping a <code>ConfigurationException</code> if any error 
    *         occurs reading the XML data or creating the configuration 
    *         objects. This includes missing configuration file or file format
    *         constraint violations detected during parsing.
    */
   public static void main(String[] args) 
   throws SAXException, 
          IOException,
          RuntimeException
   {
      //parse the file
      getInstance().loadXml(new FileInputStream(new File(args[0])));
      
      //print the constructed objects:
      
      SchemaConfig sc = (SchemaConfig) getInstance().saxHandler.getGlobalObjects(
         TAG_SCHEMA).get(0);
      System.out.println("Global schema object:\n\n" + sc);
      
      System.out.println("\nGlobal parameters:\n");
      List<?> params = getInstance().saxHandler.getGlobalObjects(TAG_PARAMETER);
      for (Object o : params)
      {
         Parameter p = (Parameter)o;
         System.out.println("name = " + p.getName() + ", value = " + p.getValue());
      }
   }
}