/*
** Module   : Parameter.java
** Abstract : Simple key/value configuration parameter
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20050120 NEW   @19361 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.cfg;

import java.io.*;

/**
 * A simple parameter object with <code>name</code> and <code>value</code>
 * properties and a bean-like interface to access them. This class is used by
 * the {@link Configuration} to load arbitrary configuration parameters using
 * the Apache digester.
 * <p>
 * Parameter values are always treated as strings and are returned as such
 * from the {@link Configuration#getParameter} method. Callers must further
 * process/parse this data if different types are required.
 */
public class Parameter
implements Serializable
{
   /** Name which uniquely identifies this parameter within configuration */
   private String name = null;
   
   /** Value of the parameter */
   private String value = null;
   
   /**
    * Default constructor.
    */
   public Parameter()
   {
   }
   
   /**
    * Get name which uniquely identifies this parameter within the
    * configuration.
    *
    * @return  Parameter name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Set name which uniquely identifies this parameter within the
    * configuration. This generally will only be called by the digester;
    * it is public by necessity so that the digester can access it.
    *
    * @param   name
    *          Parameter name.
    */
   public void setName(String name)
   {
      this.name = name;
   }
   
   /**
    * Get parameter's value.
    *
    * @return  Parameter value.
    */
   public String getValue()
   {
      return value;
   }
   
   /**
    * Set this parameter's value. This generally will only be called by the
    * digester; it is public by necessity so that the digester can access it.
    *
    * @param   value
    *          Parameter value as read from configuration document.
    */
   public void setValue(String value)
   {
      this.value = value;
   }
}