/*
** Module   : ConversionDriver.java
** Abstract : runs the entire conversion process for a file or set of files   
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20060131   @24171 First version.  Supports:
**                           - schema loader/schema fixups
**                           - scan driver (preprocess/lex/parse/persist)
**                           - post parse fixups
**                           - call graph generation
**                           - unreachable code processing
**                           - annotations
**                           - frame generator
**                           - base structure
**                           - core conversion
**                           - brew
**                           - the above steps are separated into a
**                             front end and back end that are separately
**                             or jointly runnable
**                           - scanning results are fully or partially
**                             cleared when running the front end
**                           - quick mode - reset with cached front end
**                             results before running back end
** 002 GES 20060131   @24172 Added middle processing.
** 003 ECF 20060203   @24343 Enabled temp table P2O processing. Had been
**                           commented out pending testing.
** 004 GES 20060217   @24612 Ensure that hints files don't get deleted.
** 005 ECF 20060210   @24504 Make temp table processing conditional. Only
**                           attempted if source-level temp table schema
**                           files exist.
** 006 ECF 20060220   @24618 Fixed metadata schema file handling. Needed
**                           additional logic to correctly omit this file
**                           from the list of fixed-up schema files.
** 007 GES 20060308   @24932 Only cleanup generated schema files when
**                           schema processing is being run (in front
**                           end).
** 008 GES 20060311   @24973 Force different match lists into the name
**                           converter for different types of pattern
**                           engine runs.
** 009 ECF 20060405   @25372 Made archival of key files (with .pristine
**                           extension) optional. This option is enabled
**                           with the '-A' command line option.
** 010 GES 20071015   @35527 Filenames including regex characters like
**                           square brackets, need to be escaped in
**                           certain cases.
** 011 GES 20090105   @41010 Added file-based filename list as an option.
**                           The filespec option is VERY slow when there
**                           are large numbers of files. This allows a
**                           very large file list to be passed and/or
**                           a custom list to be passed which cannot
**                           easily be duplicated by a single filespec.
**                           Eliminated use of the FileSpec class during
**                           cleanup of prior scan processing results.
**                           This is significantly faster.
** 012 GES 20090424   @41928 Import change.
** 013 GES 20090429   @42046 Match package and class name changes.
** 014 GES 20090515   @42204 Removed quick mode and archive support. These
**                           did not work properly and used unnecessary
**                           features in the backing infrastructure.
** 015 GES 20091117   @44417 Modified the name converter loading to be aware
**                           of database defaults when necessary.
** 016 ECF 20101006          Allow conversion to proceed if no database metadata
**                           file has been configured.
** 017 CA  20101102          Call the generate_ddl.xml as part of M0, for each
**                           schema and registered target DB.
** 018 LMR 20110106          Fixes to reflect renaming of load() method in 
**                           Configuration class to getSchemaConfig() and 
**                           removal of its 'key' parameter.
** 019 GES 20110824          cleanScanResults() was updated to clear from some
**                           "fakeout" directories. These directories hold
**                           class/interface definitions of the Progress
**                           built-ins and .NET classes/interfaces that are in
**                           use. In particular, the ASTs cannot be reused if
**                           there has been parser/lexer changes.
** 020 OM  20121121          Add support for writing DDL schema for sequences.
** 021 ECF 20130125          Added database metadata support.
** 022 ECF 20130624          Additional metadata support to generate DDL and export data.
** 023 CA  20140313          Changed callgraph generation to use a callgraph DB.
** 024 HC  20140613          Logging added.
** 025 ECF 20141218          Made metadata-related steps conditional; some projects do not use
**                           metadata.
** 026 VIG 20141121          Added menus generation.
** 027 ECF 20150327          Refactored middle conversion to support cross-schema natural joins.
** 028 ECF 20150804          Exit with System.exit(1) on fatal conversion error.
** 029 OM  20150824          Added pre-processing for P2O Generation for temp-table schema files.
** 030 GES 20160224          Cleanup file list inputs (trim whitespace and ignore blank lines).
** 031 IAS 20160331          Add input streams' close.
** 032 ECF 20160302          Added "abort-on-error" option for the scan portion of the front end
**                           and implemented multithreaded front end scan.
** 033 GES 20160311          Added gap analysis marking to the front end processing.
** 034 IAS 20160509          Fixed DDL dialects processing in case of more than one 
**                           default namespace.
** 035 OM  20161111          Implemented SHARED FRAMEs hierarchy.
** 036 ECF 20170213          Fixed typo in syntax output.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/*
 TODOS:
 - back end for the data side (create database(s), run import)
 - separate modes for various cleanup (deletion of different sets of output)
 - more granular debug options (step specific instead of global)
 - call graph mode for back-end processing
*/

package com.goldencode.p2j.convert;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import java.text.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.schema.*;

/**
 * Provides a harness to drive the end-to-end conversion process for a
 * user-defined list of files.
 * <p>
 * This program provides a command line interface.  See {@link #main} for the
 * syntax. 
 */
public class ConversionDriver
{
   /** No message output - quiet mode. */
   public static final int MSG_NONE = 0;
   
   /** Output status messages only. */
   public static final int MSG_STATUS = 1;
   
   /** Output debug and status messages.  */
   public static final int MSG_DEBUG = 2;
   
   /** Full debug trace output. */
   public static final int MSG_TRACE = 3;   
   
   /** Spacer for section headers and output. */
   private static final String spacer = StringHelper.repeatChar('-', 78);
   
   /** Match list for source code name conversion. */
   private static final String codenames = Configuration.getParameter("codenames");
   
   /** Match list for database name conversion. */
   private static final String datanames = Configuration.getParameter("datanames");
   
   /** Associates mode names with detailed mode definitions. */
   private static final Map<String, RunMode> modes = new HashMap<>();
   
   /** Chars that must be escaped to eliminate regular expression meaning. */
   private static final char[] DANGER_LIST = { '{', '}', '[', ']', '(', ')' };
   
   /** H2 dialect name */
   private static final String DIALECT_H2 = "h2";
   
   /** Form in which to obtain the list of files. */
   private static enum ListType { CMD_LINE, INPUT_FILE, FILESPEC };
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ConversionDriver.class);
   
   static
   {
      modes.put("F0", new RunMode(true , false, false, false, false, false, false));
      modes.put("F1", new RunMode(true , true , false, false, false, false, false));
      modes.put("F2", new RunMode(true , true , true , false, false, false, false));
      modes.put("F3", new RunMode(true , true , true , true , false, false, false));
      modes.put("M0", new RunMode(false, false, false, false, true , false, false));
      modes.put("M1", new RunMode(false, false, false, false, true , true , false));
      modes.put("CB", new RunMode(false, false, false, false, false, false, true));
      
      try
      {
         String name = Configuration.getParameter("registry");
         AstManager.initialize(new XmlFilePlugin(name));
      }
      catch (AstException ae)
      {
         LOG.log(Level.SEVERE, "Conversion initialization failed!", ae);
      }
   }
   
   /** The list of Progress source files (command line input). */
   private FileList source = null;
   
   /** The list of Progress AST (abstract syntax tree) files. */
   private FileList asts = null;
   
   /** The list of Java AST (abstract syntax tree) files. */
   private FileList jasts = null;
   
   /** Java ASTs (abstract syntax trees) created during frame generation. */
   private String[] frames = null;
   
   /** Java ASTs (abstract syntax trees) created during menus generation. */
   private String[] menus = null;
   
   /** The list of database dictionary files. */
   private FileList dicts = null;
   
   /** The list of Progress source file specific dictionary files. */
   private FileList sdicts = null;
   
   /** The list of database schema files. */
   private FileList dschemas = null;
   
   /** The list of Progress source file specific schema files. */
   private FileList sschemas = null;
   
   /** The list of database P2O files. */
   private FileList dp2os = null;
   
   /** The list of Progress source file specific P2O files. */
   private FileList sp2os = null;
   
   /** The list of DMO files. */
   private FileList dmos = null;
   
   /** The debug level to honor throughout the process. */
   private int debug = MSG_STATUS;
   
   /**
    * Creates an instance with a specific configuration.
    *
    * @param    source
    *           The list of Progress source files to process.
    * @param    debug
    *           The debug level (must be one of the MSG_ constants in this
    *           class).
    */
   ConversionDriver(FileList source, int debug)
   {
      this.source = source;
      
      if (debug < MSG_NONE || debug > MSG_TRACE)
      {
         throw new IllegalArgumentException("Invalid debug level: " + debug);
      }
      else
      {
         this.debug = debug;
      }      
   }
   
   /**
    * Drive the front end of the conversion process using the file list
    * of Progress source files that is already defined in this instance.
    * <p>
    * Once the scanning of the source files is complete, this processing will
    * modify the files list to create one that exposes AST files instead of
    * Progress source files.
    * <p>
    * This runs the following:
    * <ul>
    *    <li> schema loader (optional)
    *    <li> preprocessor (optional) 
    *    <li> lexer
    *    <li> parser
    *    <li> AST persister
    *    <li> schema fixups for the dictionary and for source-specific schema
    *         files (optional)
    *    <li> call graph generator (optional)
    *    <li> saves off a set results for enabling a future restore to this
    *         point
    * </ul>
    *
    * @param    preproc
    *           <code>true</code> if the <code>Preprocessor</code> should
    *           should be run.  <code>false</code> if previously cached
    *           preprocessor results should be used.
    * @param    schema
    *           <code>true</code> if the <code>SchemaLoader</code> and
    *           <code>schema/fixups.rules</code> should be run.
    * @param    callgraph
    *           <code>true</code> if the <code>CallGraphGenerator</code>
    *           should be run.
    * @param    abortScanOnError
    *           <code>true</code> to end the scanning run if an error is encountered,
    *           else <code>false</code>.
    * @param    threads
    *           Number of threads to use for the scan. A non-positive value will enable
    *           the default behavior of a work-stealing thread pool with one thread per
    *           per available CPU. A positive value will create a fixed size thread pool.
    */
   private void front(boolean preproc,
                      boolean schema,
                      boolean callgraph,
                      boolean abortScanOnError,
                      int     threads)
   throws ConfigurationException,
          AstException,
          SchemaException,
          IOException,
          Exception
   {
      if (schema)
      {
         // run schema loader to generate new dictionary files
         runSchemaLoader();
      }
      
      // optionally preprocess, lex, parse and save the resulting ASTs as
      // persisted XML files with the name sourcename.[pPwW].ast
      runScanDriver(preproc, schema, abortScanOnError, threads);
      
      // ensure that downstream processing has the proper list of files
      convertSourceNamesToAstNames();
      
      processTrees("Post-Parse Fixups",
                   "fixups/post_parse_fixups",
                   null,
                   false,
                   asts,
                   debug);
      
      processTrees("Gap Analysis Marking",
                   "gaps/gap_analysis_marking",
                   null,
                   false,
                   asts,
                   debug);
      
      if (schema)
      {
         // clean up previous run - build schema list and delete those files
         FileList delDicts = createDatabaseList(".schema", false);
         deleteFiles(delDicts.listFilenames());
         
         // build dictionary list
         dicts = createDatabaseList(".dict", omitSchemaMetadata());
         
         // fixup dictionary files
         processTrees("Schema Fixups (data dictionary)", 
                      "schema/fixups",
                      datanames,
                      true,
                      dicts,
                      debug);
         
         // build schema list
         sdicts = convertSourceNamesToDictNames(".dict");
         
         // fixup source file schema files (if any exist for these specific
         // source files)
         if (sdicts.size() > 0)
         {
            processTrees("Schema Fixups (Progress source file schemas)",
                         "schema/fixups",
                         datanames,
                         true,
                         sdicts,
                         debug);         
         }
      }
      
      if (callgraph)
      {
         runCallGraphGenerator();
      }
   }
   
   /**
    * Runs the <code>SchemaLoader</code> class to convert <code>.df</code>
    * files into an AST form which can be read by the
    * <code>SchemaDictionary</code>.  Files left by previous runs, if any,
    * are removed first.
    */
   private void runSchemaLoader()
   throws ConfigurationException,
          SchemaException,
          IOException
   {
      printHeader("SchemaLoader", debug);
      
      // clean up previous run - build dictionary list and delete those files
      FileList delDicts = createDatabaseList(".dict", false);
      deleteFiles(delDicts.listFilenames());
      
      SchemaLoader loader = new SchemaLoader();
      
      if (debug == MSG_NONE)
      {
         loader.setQuietMode(true);
      }
      
      loader.importAll();
   }
   
   /**
    * Runs the <code>ScanDriver</code> class to convert Progress source
    * files into an AST form.  This corresponds to preprocessing, lexing,
    * parsing and AST persistence.
    *
    * @param    preproc
    *           <code>true</code> if the <code>Preprocessor</code> should
    *           should be forced to run.  <code>false</code> if previously
    *           cached preprocessor results should be honored.
    * @param    schema
    *           <code>true</code> if the <code>SchemaLoader</code> and
    *           <code>schema/fixups.rules</code> should be run.  Note this
    *           is only to assist with cleaning files as the schema 
    *           processing itself is not handled here.
    * @param    abortOnError
    *           <code>true</code> to end the scanning run if an error is encountered,
    *           else <code>false</code>.
    * @param    threads
    *           Number of threads to use for the scan. A non-positive value will enable
    *           the default behavior of a work-stealing thread pool with one thread per
    *           per available CPU. A positive value will create a fixed size thread pool.
    */
   private void runScanDriver(boolean preproc, boolean schema, boolean abortOnError, int threads)
   throws Exception
   {
      printHeader("Scanning Progress Source (" +
                  (preproc ? "preprocessor, " : "") +
                  "lexer, parser, persist ASTs)",
                  debug);
      
      long start = System.currentTimeMillis();
      
      cleanScanResults(preproc, schema);
      
      ScanDriver.scan(source, (debug == MSG_NONE), abortOnError, threads);
      
      printElapsed(System.currentTimeMillis() - start);
   }
   
   /**
    * Remove some or all scan driver output files associated with the
    * current Progress source file list AND in the list of fakeout files
    * that represent the built-in 4GL classes/interfaces or .NET classes
    * that can be accessed.  All output except preprocessor and schema output
    * is unconditionally deleted.
    *
    * @param    preproc
    *           <code>true</code> means that cached preprocessor files should
    *           be deleted.
    * @param    schema
    *           <code>true</code> means that schema related generated files
    *           should be deleted.
    */
   private void cleanScanResults(boolean preproc, boolean schema)
   throws IOException
   {
      ArrayList<String> exts = new ArrayList<>();
      
      exts.add(".lexer");
      exts.add(".parser");
      exts.add(".ast");
      exts.add(".jast");
      
      if (preproc)
      {
         exts.add(".cache");
         exts.add(".pphints");
      }
      
      if (schema)
      {
         exts.add(".dict");
         exts.add(".schema");
         exts.add(".schema.original");
         exts.add(".p2o");
      }
      
      // we never delete ".schema.hints" or ".hints" as those are both
      // hand-coded configuration files
      
      String[] extslist = exts.toArray(new String[0]);
      
      // do the work
      zapFiles(source.listFilenames(), extslist);
      zapFiles(SymbolResolver.listFakeoutFiles(), extslist);
   }
   
   /**
    * Concatenate all extensions with all of the files and if any resulting
    * names exist, delete the matching file from the file system.
    *
    * @param    files
    *           The list of file base names.
    * @param    extslist
    *           The list of extensions to zap.
    */
   private void zapFiles(String[] files, String[] extslist)
   {
      
      for (int i = 0; i < files.length; i++)
      {
         for (int j = 0; j < extslist.length; j++)
         {
            StringBuilder sb = new StringBuilder();
            sb.append(files[i]).append(extslist[j]);
            
            File check = new File(sb.toString());
            
            if (check.exists() && check.isFile())
            {
               // delete the file
               check.delete();
            }
         }
      }
   }
   
   /**
    * Process each input element and escape characters that would be
    * interpreted as part of a regular expression to convert them all into
    * a literal match.  At this time, only a limited subset of characters
    * is escaped (curly braces, square brackets and parenthesis). This allows
    * valid file name characters to be passed literally in a larger file
    * specification that will be interpreted as a regular expression.
    *
    * @param    list
    *           List of literal strings that may need to be escaped.
    *
    * @return   Safe version of the list.
    */
   private String[] escapeRegex(String[] list)
   {
      String[] result = new String[list.length];
      
      for (int i = 0; i < list.length; i++)
      {
         StringBuilder sb = new StringBuilder();
         
         int len = list[i].length();
         
         for (int j = 0; j < len; j++)
         {
            char next = list[i].charAt(j);
            
            for (int x = 0; x < DANGER_LIST.length; x++)
            {
               if (DANGER_LIST[x] == next)
               {
                  sb.append('\\');
                  break;
               }
            }
            
            sb.append(next);
         }
         
         result[i] = sb.toString();
      }
      
      return result;
   }
      
   /**
    * Runs the <code>CallGraphGenerator</code> using the project's defined
    * root nodes list to create a call graph for each root node.
    */
   private void runCallGraphGenerator()
   throws ConfigurationException,
          IOException,
          AstException
   {
      printHeader("Generating Call Graphs", debug);
      
      try
      {
         CallGraphGenerator.generateGraphs(null, debug);
      }
      finally
      {
         CallGraphWorker.shutdownGraph();
      }
   }

   /**
    * Drive the middle of the conversion process using the schema dictionary
    * lists generated from the output of the <code>SchemaLoader</code> and
    * the file list of Progress AST files that is already defined in this
    * instance.
    * <p>
    * This runs the following:
    * <ul>
    *    <li> schema annotations (analyze Progress source code)
    *    <li> P2O generation for each database schema and the schema files
    *         associated with all temp-tables (stored in Progress source
    *         file specific filenames)
    *    <li> length statistics (optional)
    *    <li> generate Hibernate mapping documents
    *    <li> DMO generation for each database schema and the schema files
    *         associated with all temp-tables (stored in Progress source
    *         file specific filenames)
    *    <li> brew (converts from JAST to <code>.java</code>)
    * </ul>
    *
    * @param    length
    *           <code>true</code> if the length statistics step must be run
    *           (it takes a VERY long time). 
    */
   private void middle(boolean length)
   throws ConfigurationException,
          AstException,
          SchemaException,
          IOException
   {      
      // ensure that downstream processing has the proper list of files
      convertSourceNamesToAstNames();
      
      processTrees("Schema Annotations (scan Progress source code)", 
                   "schema/annotations",
                   datanames,
                   true,
                   asts,
                   debug);
      
      boolean omitMeta = omitSchemaMetadata();
      
      // build database schema list
      dschemas = createDatabaseList(".schema", omitMeta);
      
      // build Progress source schema list
      sschemas = convertSourceNamesToDictNames(".schema");
      
      processTrees("P2O Generation (database schema files)", 
                   "schema/p2o",
                   datanames,
                   true,
                   dschemas,
                   debug);
      
      // convert source file schema files to P2O (if any exist for these
      // specific source files)
      if (sschemas.size() > 0)
      {
         processTrees("Pre-processing P2O Generation (temp-table schema files)", 
                      "schema/p2o_pre",
                      datanames,
                      true,
                      sschemas,
                      debug);
         
         processTrees("P2O Generation (temp-table schema files)", 
                      "schema/p2o",
                      datanames,
                      true,
                      sschemas,
                      debug);
      }
      
      processTrees("P2O Post-processing (database schema files)", 
                   "schema/p2o_post",
                   datanames,
                   true,
                   dschemas,
                   debug);
      
      // convert source file schema files to P2O (if any exist for these
      // specific source files)
      if (sschemas.size() > 0)
      {
         processTrees("P2O Post-processing (temp-table schema files)", 
                      "schema/p2o_post",
                      datanames,
                      true,
                      sschemas,
                      debug);
      }
      
      // build database p2o list
      dp2os = createDatabaseList(".p2o", omitMeta);
      
      if (length)
      {
         processTrees("Database Field Length Statistics",
                      "schema/data_analyzer",
                      datanames,
                      true,
                      dp2os,
                      debug);
      }
      
      processTrees("Generate Hibernate Mappings (database P2Os)",
                   "schema/hibernate",
                   datanames,
                   true,
                   dp2os,
                   debug);
                   
      // build Progress source P2O list
      sp2os = convertSourceNamesToDictNames(".p2o");
      
      // generate temp table Hibernate mappings (if any exist for these
      // specific source files)
      if (sp2os.size() > 0)
      {
         processTrees("Generate Hibernate Mappings (temp-table P2Os)", 
                      "schema/hibernate",
                      datanames,
                      true,
                      sp2os,
                      debug);
      }
      
      dmos = generateDMOs();
      
      brew(dmos, "Data Model Objects (DMOs)");
      
      // get application schema names (no metadata)
      SchemaConfig config = Configuration.getSchemaConfig();
      List<String> schemas = config.getSchemas(true);
      
      // generate DDL for application schemas
      generateDDL("base/table", schemas, null, "schema/generate_ddl");
      generateDDL("sequence", schemas, null, "schema/generate_seq_ddl");
      
      if (!omitMeta)
      {
         // generate metadata DDL
         schemas = new ArrayList<String>(1);
         schemas.add(config.getMetadata().getName());
         generateDDL("metadata", schemas, DIALECT_H2, "schema/generate_ddl");
         
         // rebuild database p2o list, without metadata
         dp2os = createDatabaseList(".p2o", true);
         
         // generate metaschema XML which will be used to load metadata at server startup
         processTrees("Metadata XML Generation",
                      "schema/metaschema",
                      datanames,
                      true,
                      dp2os,
                      debug);
      }
   }
   
   /**
    * For each .p2o file associated with a schema and for each specified dialect, generate the
    * specified type of DDL.
    * <p>
    * This is done by calling the ruleset specified and setting the <code>dbName</code> and
    * <code>targetDb</code> parameters accordingly.
    * 
    * @param   ddlType
    *          Type of DDL being generated (used only for console message).
    * @param   schemas
    *          List of names of schemas to process.
    * @param   dialectList
    *          List of database dialects for which to generate DDL or <code>null</code> if this
    *          information is to be read from the configuration.
    * @param   ruleset
    *          Name of ruleset to execute (without file extension suffix).
    * 
    * @throws  ConfigurationException
    *          if the schema configuration can't be loaded.
    */
   private void generateDDL(String ddlType,
                            List<String> schemas,
                            String dialectList,
                            String ruleset)
   throws ConfigurationException
   {
      SchemaConfig config = Configuration.getSchemaConfig();
      
      // for each .p2o file and for each read dialect, generate the DDL
      Map<String, String> vars = new HashMap<String, String>();
      
      for (String schema: schemas)
      {
         String dialects_ = dialectList;
         File schemaFile = config.getSchemaFile(schema);
         String p2oName = replaceFilenameExtension(schemaFile.getPath(), ".p2o");
         ExplicitFileList files = new ExplicitFileList(new String[] { p2oName });
         
         vars.put("dbName", "\"" + schema + "\"") ;
         
         // for each dialect
         if (dialects_ == null)
         {
            dialects_ = config.getNamespaceParameter(schema, "ddl-dialects");
         }
         
         if (dialects_ == null)
         {
            System.out.println("WARNING: no DDL dialect specified for schema " + schema);
            
            continue;
         }
         
         String[] dialects = dialects_.split(",");
         for (String dialect : dialects)
         {
            dialect = dialect.trim().toLowerCase();
            vars.put("targetDb", "\"" + dialect + "\"");
            
            // generate the DDL
            processTrees("Generate " + ddlType +
                         " DDL scripts for schema '" + schema + 
                         "' and target DB '" + dialect + "'",
                         ruleset,
                         datanames,
                         true,
                         files,
                         debug,
                         vars);
         }
      }
   }
   
   /**
    * Drives the execution of the DMO creation rule-set using the P2O lists
    * as input and capturing the list of Java AST files generated.
    *
    * @return   The list of JAST files to be "brewed".
    */
   private FileList generateDMOs()
   throws AstException,
          ConfigurationException
   {
      // enable new file tracking mode
      JavaPatternWorker.enableNewFileTracking();
      
      processTrees("Generate Data Model Objects (database DMOs)", 
                   "schema/java_dmo",
                   datanames,
                   true,
                   dp2os,
                   debug);
      
      if (sp2os.size() > 0)
      {
         processTrees("Generate Data Model Objects (temp-table DMOs)", 
                      "schema/java_dmo",
                      datanames,
                      true,
                      sp2os,
                      debug);
      }
      
      // add generated JASTs to the overall list (used with brew later)
      JavaPatternWorker.disableNewFileTracking();      
      String[] dmolist = JavaPatternWorker.getPersistedFiles();
      
      return new ExplicitFileList(dmolist);
   }
   
   /**
    * Drive the back end of the conversion process using the file list
    * of Progress AST files that is already defined in this instance.
    * <p>
    * This runs the following:
    * <ul>
    *    <li> unreachable code analysis
    *    <li> code conversion annotations
    *    <li> frame generator
    *    <li> business logic base structure
    *    <li> core conversion
    *    <li> brew (converts from JAST to <code>.java</code>)
    * </ul>
    */
   private void back()
   throws ConfigurationException,
          AstException,
          SchemaException,
          IOException
   {      
      // ensure that downstream processing has the proper list of files
      convertSourceNamesToAstNames();
      
      processTrees("Unreachable Code Analysis",
                   "unreachable/unreachable",
                   null,
                   false,
                   asts,
                   debug);
      
      processTrees("Code Conversion Annotations",
                   "annotations/annotations",
                   codenames,
                   false,
                   asts,
                   debug);
                   
      generateFrames();
      
      generateMenus();
                   
      processTrees("Business Logic Base Structure",
                   "convert/base_structure",
                   codenames,
                   false,
                   asts,
                   debug);
                   
      processTrees("Core Code Conversion",
                   "convert/core_conversion",
                   codenames,
                   false,
                   asts,
                   debug);
                   
      // ensure that downstream processing has the proper list of files
      convertSourceNamesToJavaAstNames();
      
      brew(jasts, "Business Logic and Frames Classes");
   }
   
   /**
    * Drives the execution of the frame generator rule-set using the Progress
    * ASTs as input and capturing the list of Java AST files generated. These
    * generated files are all added to the list of Java AST files to be
    * processed downstream.
    */
   private void generateFrames()
   throws AstException,
          ConfigurationException
   {
      // TODO: change to a more adequate naming 
      processTrees("Detecting Frame Interfaces",
            "annotations/frame_generator_pre",
            null,
            false,
            asts,
            debug);
      
      // enable new file tracking mode
      JavaPatternWorker.enableNewFileTracking();
      
      processTrees("Frame Generator",
                   "convert/frame_generator",
                   codenames,
                   false,
                   asts,
                   debug);
                   
      // add generated JASTs to the overall list (used with brew later)
      JavaPatternWorker.disableNewFileTracking();
      frames = JavaPatternWorker.getPersistedFiles();
   }
   
   /**
    * Drives the execution of the frame generator rule-set using the Progress
    * ASTs as input and capturing the list of Java AST files generated. These
    * generated files are all added to the list of Java AST files to be
    * processed downstream.
    */
   private void generateMenus()
   throws AstException,
          ConfigurationException
   {
      // enable new file tracking mode
      JavaPatternWorker.enableNewFileTracking();
      
      processTrees("Menu Generator",
                   "convert/menu_generator",
                   codenames,
                   false,
                   asts,
                   debug);
                   
      // add generated JASTs to the overall list (used with brew later)
      JavaPatternWorker.disableNewFileTracking();
      menus = JavaPatternWorker.getPersistedFiles();
   }
   
   /**
    * Convert Java AST (JAST) files into Java source code (<code>.java</code>
    * files).  The output is written into the configured project's output
    * path.
    *
    * @param    files
    *           The list of JAST files to process.
    * @param    extra
    *           Text to append to the displayed header.
    */
   private void brew(FileList files, String extra)
   throws ConfigurationException,
          AstException
   {
      processTrees("Generate Java Source " + extra,
                   "convert/brew",
                   null,
                   false,
                   files,
                   debug);
   }
   
   /**
    * Run the pattern engine using the given rules filename as the program 
    * and the list of files are the ASTs (abstract syntax trees) to process.
    *
    * @param    section
    *           The section name (for the header);
    * @param    rules
    *           The pattern engine program (known as a rules file) to run.
    * @param    matches
    *           The match list name to use or <code>null</code> to use the
    *           project default.
    * @param    db
    *           Load default database name mappings.
    * @param    files
    *           The list of ASTs to process.
    * @param    debug
    *           The debug level to use with the pattern engine.
    */
   private void processTrees(String   section,
                             String   rules,
                             String   matches,
                             boolean  db,
                             FileList files,
                             int      debug)
   throws ConfigurationException,
          AstException
   {
      printHeader(section, debug);
      
      // load/reload a known match list for name conversion
      NameConverter.load(matches, db);
      
      // create a new pattern engine, configure it and run
      PatternEngine engine = new PatternEngine();
      engine.addAstSpec(files);
      engine.setDebugLevel(debug);
      engine.run(rules);
   }

   /**
    * Run the pattern engine using the given rules filename as the program 
    * and the list of files are the ASTs (abstract syntax trees) to process.
    * <p>
    * Also, set the variable initializers using the given map.
    *
    * @param    section
    *           The section name (for the header);
    * @param    rules
    *           The pattern engine program (known as a rules file) to run.
    * @param    matches
    *           The match list name to use or <code>null</code> to use the
    *           project default.
    * @param    db
    *           Load default database name mappings.
    * @param    files
    *           The list of ASTs to process.
    * @param    debug
    *           The debug level to use with the pattern engine.
    * @param    vars
    *           The variable initializers.
    */
   private void processTrees(String   section,
                             String   rules,
                             String   matches,
                             boolean  db,
                             FileList files,
                             int      debug,
                             Map      vars)
   throws ConfigurationException,
          AstException
   {
      printHeader(section, debug);
      
      // load/reload a known match list for name conversion
      NameConverter.load(matches, db);
      
      // create a new pattern engine, configure it and run
      PatternEngine engine = new PatternEngine();
      engine.addAstSpec(files);
      engine.setDebugLevel(debug);
      engine.setVariableInitializers(vars);
      engine.run(rules);
   }
   
   /**
    * Delete the file associated with each name in the list.
    *
    * @param    list
    *           List of explicit filenames to delete.
    */
   private void deleteFiles(String[] list)
   throws IOException
   {
      for (int i = 0; i < list.length; i++)
      {
         File next = new File(list[i]);
         next.delete();
      }
   }   
   
   /**
    * For each element in the spec list, generate a list of matching files
    * (each spec may contain traditional file system wildcard characters)
    * and delete each of these files.  The given <code>ExcludeCheck</code> 
    * instance is used to remove results from the list on a callback basis.
    *
    * @param    speclist
    *           List of file specifications.
    * @param    check
    *           If <code>null</code>, no callback occurs.  Otherwise, the 
    *           <code>exclude</code> method will be called on the passed
    *           instance.
    */
   private void deleteFileSpecs(String[] speclist, ExcludeCheck check)
   throws IOException
   {
      for (int i = 0; i < speclist.length; i++)
      {
         int    idx  = speclist[i].lastIndexOf(File.separator);
         String dir  = (idx < 0) ? "." : speclist[i].substring(0, idx);         
         String spec = speclist[i].substring(idx + 1);
         
         // create a list of actual files matching this spec
         FileList list = new FileSpecList(new File(dir), spec, false);
         File[]   fset = list.list();
         
         for (int j = 0; j < fset.length; j++)
         {
            // prune out the preprocessor cache files
            if (check != null && check.exclude(fset[j].getCanonicalPath()))
            {
               continue;
            }
            
            // delete the file
            fset[j].delete();
         }
      }
   }
   
   /**
    * Print a section header and optionally spacers, depending on debug
    * level (if level &gt; <code>MSG_NONE</code> then spacers are printed).
    *
    * @param    section
    *           The section name.
    * @param    debug
    *           The debug level.
    */
   private static void printHeader(String section, int debug)
   {
      if (debug > MSG_NONE)
      {
         System.out.println();
         System.out.println(spacer);
      }
      
      System.out.println(section);
      
      if (debug > MSG_NONE)
      {
         System.out.println(spacer);
         System.out.println();
      }
   }
   
   /**
    * Using the source name list as input, this method creates duplicate
    * lists with the proper <code>.ast</code> or <code>.jast</code> file
    * suffixes on each entry. The results are saved into the Progress AST
    * name list.
    */
   private void convertSourceNamesToAstNames()
   {
      String[] filenames = source.listFilenames();      
      asts  = new ExplicitFileList(appendText(filenames, ".ast"));
   }
   
   /**
    * Using the source name list as input, this method creates duplicate
    * lists with the proper <code>.ast</code> or <code>.jast</code> file
    * suffixes on each entry.  Appended to this list is the list of Java ASTs
    * created during frame generation.  The results are saved into the
    * Java AST name list.
    */
   private void convertSourceNamesToJavaAstNames()
   {
      String[] filenames = source.listFilenames();      
      String[] oldjasts  = appendText(filenames, ".jast");
      
      // merge the lists
      ArrayList list = new ArrayList(Arrays.asList(oldjasts));
      list.addAll(Arrays.asList(frames));
      list.addAll(Arrays.asList(menus));
      
      // save the results
      jasts = new ExplicitFileList((String[]) list.toArray(new String[0]));
   }
   
   /**
    * Creates a list of all schema files in the project's 
    * <code>data/namespace</code> directory by the given extension that are
    * not associated with the special meta-data dictionary.
    *
    * @param    append
    *           The file extension to search for (usually <code>.dict</code>,
    *           <code>.schema</code>, or <code>.p2o</code>).
    * @param    omitMeta
    *           <code>true</code> to omit the metadata filename from the
    *           resulting list;  <code>false</code> to retain it.
    *
    * @return   The list of files found.
    */
   private FileList createDatabaseList(String append, boolean omitMeta)
   throws ConfigurationException
   {
      File dir = Configuration.toFile("data/namespace");
      FileList files = new FileSpecList(dir, ("*" + append), false);
      List<String> list = new ArrayList<String>();
      String[] filelist = files.listFilenames();
      File metafile = null;
      
      if (omitMeta)
      {
         SchemaConfig config = Configuration.getSchemaConfig();
         metafile = config.getSchemaFile(config.getMetadata().getName());
         if (metafile != null)
         {
            String name = replaceFilenameExtension(metafile.toString(), append);
            metafile = new File(name);
         }
      }
      
      // omit metadata entry if any
      for (int i = 0; i < filelist.length; i++)
      {
         if ((metafile == null || !metafile.equals(new File(filelist[i]))))
         {
            list.add(filelist[i]);
         }
      }
      
      String[] result = (String[]) list.toArray(new String[0]);
      
      return new ExplicitFileList(result);
   }
   
   /**
    * Determine whether we should omit the metadata schema when processing and persisting schema
    * information.
    * 
    * @return  <code>true</code> if no metadata tables are specified by the project configuration,
    *          else <code>false</code>.
    */
   private boolean omitSchemaMetadata()
   {
      return Configuration.getSchemaConfig().getMetadata().getTables().isEmpty();
   }
   
   /**
    * Using the source name list as input, this method creates a duplicate
    * list with the given file suffix on each entry.
    *
    * @param    append
    *           The file extension to search for (usually <code>.dict</code>
    *           or <code>.schema</code>).
    *
    * @return   The list of files found.
    */
   private FileList convertSourceNamesToDictNames(String append)
   {
      String[] schemanames = appendText(source.listFilenames(), append);
      return new ExplicitFileList(schemanames);
   }
   
   /**
    * Appends the given text to every entry in the array and returns
    * a new array with the result.
    *
    * @param    input
    *           The array of strings to modify.
    * @param    text
    *           The text to append to each element.
    *
    * @return   A new array with each element modified.
    */
   private String[] appendText(String[] input, String text)
   {
      String[] array = new String[input.length];
      
      StringBuilder sb;
      
      for (int i = 0; i < input.length; i++)
      {
         sb = new StringBuilder(input[i]).append(text);
         array[i] = sb.toString();
      }
      
      return array;
   }
   
   /**
    * Replace the filename extension with the given extension and return the new string. If the
    * given filename already has the given extension, simply return the existing filename.
    * 
    * @param   filename
    *          File name;  may not be <code>null</code>.
    * @param   append
    *          Extension to append after the filename's existing extension is removed. Must
    *          begin with the dot (<code>.</code>) character.
    * 
    * @return  Updated filename.
    */
   private String replaceFilenameExtension(String filename, String append)
   {
      if (filename.endsWith(append))
      {
         return filename;
      }
      
      StringBuilder buf = new StringBuilder();
      int pos = filename.lastIndexOf(".");
      if (pos < 0)
      {
         buf.append(filename);
      }
      else
      {
         buf.append(filename.substring(0, pos));
      }
      
      buf.append(append);
      
      return buf.toString();
   }
   
   /**
    * Enables delegation of list processing to another method while providing 
    * a callback to determine if a given list entry needs to be excluded from
    * processing.
    */
   private abstract class ExcludeCheck
   {
      /**
       * Determines if a given name should be excluded from list processing. 
       *
       * @param    name
       *           The list element to check.
       *
       * @return   <code>true</code> if the element should be excluded from
       *           list processing.
       */
      abstract boolean exclude(String name); 
   }
   
   /**
    * Helper to only exclude the generated preprocessor files (cache and 
    * preprocessor hints) conditionally and to always exclude the manually 
    * coded hints files.
    */
   private class ScanCheck
   extends ExcludeCheck
   {
      /** <code>true</code> if preprocessor files should be deleted. */
      private boolean preproc = false;
      
      /** <code>true</code> if schema files should be deleted. */
      private boolean schema = false;
      
      /**
       * Construct and instance and store the preprocessor delete setting.
       *
       * @param    preproc
       *           <code>true</code> if preprocessor files should be deleted.
       * @param    schema
       *           <code>true</code> if schema files should be deleted.
       */
      ScanCheck(boolean preproc, boolean schema)
      {
         this.preproc = preproc;
         this.schema  = schema;
      }
      
      /**
       * Determines if a given name should be excluded from list processing. 
       *
       * @param    name
       *           The list element to check.
       *
       * @return   <code>true</code> if the element should be excluded from
       *           list processing.
       */
      boolean exclude(String name)
      {
         return (!preproc && name.endsWith(".cache"))           ||
                (!preproc && name.endsWith(".pphints"))         ||
                (!schema  && name.endsWith(".dict"))            ||
                (!schema  && name.endsWith(".schema"))          ||                
                (!schema  && name.endsWith(".schema.original")) ||
                (!schema  && name.endsWith(".p2o"))             ||
                name.endsWith(".schema.hints")                  ||
                name.endsWith(".hints");
      }
   }
   
   /**
    * Container class that aggregates a set of options to control a run of  
    * the conversion driver.
    */
   private static class RunMode
   {
      /** Run the front end. */
      boolean front = false;
      
      /** The front end should run the schema-related components. */
      boolean fschema = false;
      
      /** The front end should run the call graph generator. */
      boolean callgraph = false;
      
      /** The front end should run the preprocessor. */
      boolean preproc = false;
      
      /** Run the middle. */
      boolean middle = false;
      
      /** Run the length statistics during the middle. */
      boolean length = false;      
      
      /** Run the back end. */
      boolean back = false;
      
      /**
       * Builds an instance and initializes all data.
       */
      RunMode(boolean front,
              boolean preproc,
              boolean fschema,
              boolean callgraph,
              boolean middle,
              boolean length,
              boolean back)
      {
         this.front     = front;
         this.preproc   = preproc;
         this.fschema   = fschema;
         this.callgraph = callgraph;
         this.middle    = middle;
         this.length    = length;
         this.back      = back;
      }
      
      /**
       * Logically OR the given mode flags with the current mode flags and 
       * return an instance that has all flags active that are active in
       * either of the 2 original input modes.  No changes are made to the
       * original instances.
       *
       * @param    run
       *           The mode to logically OR with this instance's data.
       *
       * @return   The logically OR'd result.
       */
      RunMode merge(RunMode run)
      {
         return new RunMode((this.front     || run.front),
                            (this.preproc   || run.preproc),
                            (this.fschema   || run.fschema),
                            (this.callgraph || run.callgraph),
                            (this.middle    || run.middle),
                            (this.length    || run.length),
                            (this.back      || run.back));
      }
   }
   
   /**
    * Interpret the mode command line argument and convert this into the
    * effective run mode object.  Multiple "+" delimited mode values can
    * be included in the same argument.  In this case, the result is the
    * logical OR of all the activated flags in each specified mode.
    *
    * @param    mode
    *           The mode command line argument.
    *
    * @return   The run mode to use for controlling program flow. 
    */
   private static RunMode interpretRunMode(String mode)
   {
      String[] cand   = mode.split("\\+");
      RunMode  result = null;
      
      if (mode.length() == 0 || cand.length == 0)
      {
         syntax("Missing mode parameter.", -4);
      }
      
      for (int i = 0; i < cand.length; i++)
      {
         RunMode next = modes.get(cand[i]);
         
         if (next == null)
         {
            syntax("Unrecognized mode '" + mode + "'.", -5);         
         }
         
         if (result == null)
         {
            // our first match sets the starting result
            result = next;
         }
         else
         {
            // logically OR these together so that the result is the
            // combination of all true flags
            result = result.merge(next);
         }
      }
      
      return result;
   }
   
   
   /**
    * Print out a simple elapsed time counter.
    *
    * @param    elapsed
    *           Number of milliseconds of time elapsed.
    */
   private static void printElapsed(long elapsed)
   {
      Date date = new Date(elapsed - TimeZone.getDefault().getRawOffset());
      SimpleDateFormat format = new SimpleDateFormat("HH:mm:ss");
      
      StringBuilder buf = new StringBuilder("Elapsed job time:  ");
      buf.append(format.format(date));
      
      System.out.println();
      System.out.println(spacer);
      System.out.println(buf.toString());
      System.out.println();
   }
   
   /**
    * Emit a syntax statement and optional error message to
    * <code>stderr</code>, then exit the process with the specified return
    * code. This method is called from {@link #main} when the user passes
    * command line parameters which are not viable.
    *
    * @param    msg
    *           Optional error message; may be <code>null</code>.
    * @param    rc
    *           Process return code used with <code>System.exit()</code>.
    */
   private static void syntax(String msg, int rc)
   {
      if (msg != null)
      {
         System.err.println(msg);
         System.err.println();
      }
      
      String[] staticText =
      {
         "",
         "----------------------P2J Conversion Driver----------------------",
         "",
         "Syntax:",
         "",
         "   java -DP2J_HOME=<home> ConversionDriver [options] <mode> " +
         "<filelist>",
         "   java -DP2J_HOME=<home> ConversionDriver -F[options] <mode> " +
         "<filename>",
         "   java -DP2J_HOME=<home> ConversionDriver -S[options] <mode> " +
         "<directory> \"<filespec>\"",
         "",
         "Where",
         "",
         "   Mode (1 or more of the following values, use '+' to delimit",
         "         if there are 2 or more values; don't insert spaces!)", 
         "",
         "      -------------------------Front End-------------------------",
         "      F0 = preprocessor (in honor cache mode), lexer, parser, ",
         "           AST persistence and post-parse fixups",
         "      F1 = F0 + forced preprocessor execution",
         "      F2 = F1 + schema loader/fixups",
         "      F3 = F2 + call graph generation",
         "",
         "      --------------------------Middle---------------------------",
         "      M0 = schema annotations, P2O generation, Hibernate mapping",
         "           docs, DMO generation, brew for DMO classes",
         "      M1 = M1 + length statistics",
         "",
         /*
         "      ----------------------Data Back End------------------------",
         "      DB = generate DDL, create the database(s) and import data",
         "",
         */
         "      ----------------------Code Back End------------------------",
         "      CB = unreachable code, annotations, frame generator,",
         "           base structure, core conversion and brew",
         "",
         "   Options",
         "",
         "      Dn = set debug level 'n' (must be a numeric digit between",
         "           " + MSG_NONE + " and " + MSG_TRACE + " inclusive)",
         "      F  = pass the filename of a custom file list instead of ",
         "           using the explicit command line file list",
         "      S  = use filespecs instead of an explicit command line file",
         "           list",
         "      N  = no recursion in filespec mode",
         "      A  = abort scan part of front end on an error",
         "      Tn = specify number of threads 'n' to use for front end",
         "           file scan ('n' must be >= 0; defaults to 1; set to 0 to",
         "           use 1 thread per available CPU)",
         "",
         "   filelist   = arbitrary list of absolute and/or relative file", 
         "                names to scan (default approach)",
         "   directory  = a relative or absolute path (only used when using",
         "                filespecs)",
         "   filespec   = file filter specification of the filenames in the",      
         "                given directory to scan, should be enclosed in",
         "                double quotes if any of the wildcard characters '*'",
         "                or '?' are used (only used when using filespecs)",
         "   filename   = is a single absolute or relative filename of a ",
         "                text file that contains a custom list of absolute ",
         "                and/or relative file names of files to scan (i.e. ",
         "                the file list will be read from a file instead of ",
         "                being hard coded on the command line); there must ",
         "                be one filename per line in the file",
         ""
      };
      
      for (int i = 0; i < staticText.length; i++)
      {
         System.err.println(staticText[i]);
      }
      
      System.exit(rc);
   }
   
   /**
    * Provides a command line interface for an end user to drive the scanning harness.
    * <p>
    * The program will return a non-zero code if there is a fatal error running the conversion.
    * <p>
    * Syntax:
    * <pre>
    *   java -DP2J_HOME=&lt;home&gt; ConversionDriver &lt;mode&gt; [options] &lt;filelist&gt;
    *   java -DP2J_HOME=&lt;home&gt; ConversionDriver &lt;mode&gt; -F[options] &lt;filename&gt;
    *   java -DP2J_HOME=&lt;home&gt; ConversionDriver &lt;mode&gt; -S[options] &lt;directory&gt; &lt;&quot;filespec&quot;&gt;
    * </pre>
    * Where:
    * <ul>
    *   <li> Mode (1 or more of the following values, use '+' to delimit
    *        if there are 2 or more values; don't insert spaces!)
    *   <ul>
    *      <li> Front End
    *      <ul>
    *         <li> F0 = preprocessor (in honor cache mode), lexer, parser,
    *              AST persistence and post-parse fixups
    *         <li> F1 = F0 + forced preprocessor execution
    *         <li> F2 = F1 + schema loader/fixups
    *         <li> F3 = F2 + call graph generation
    *      </ul>
    *      <li> Middle
    *      <ul>
    *         <li> M0 = schema annotations, P2O generation, Hibernate mapping
    *              docs, DMO generation, brew for DMO classes
    *         <li> M1 = M1 + length statistics
    *      </ul>
    *      <li> Code Back End
    *      <ul>
    *         <li> CB = unreachable code, annotations, frame generator,
    *              base structure, core conversion and brew
    *      </ul>
    *   </ul>
    *   <li> Options
    *   <ul>
    *      <li> Dn = set debug level 'n' (must be a numeric digit between
    *           MSG_NONE (0) and MSG_TRACE (3) inclusive
    *      <li> F  = pass the filename of a custom file list instead of using
    *           the explicit command line file list
    *      <li> S  = use filespecs instead of an explicit command line file list
    *      <li> N  = no recursion in filespec mode
    *      <li> A  = abort scan part of front end on an error
    *      <li> Tn = specify number of threads 'n' to use for front end file scan
    *           ('n' must be &gt;= 0; defaults to 1; set to 0 to use 1 thread per
    *           available CPU)
    *   </ul>
    *   <li> explicit command line list (default approach)
    *   <ul>
    *      <li> &lt;filelist&gt; is arbitrary list of absolute and/or
    *           relative file names to scan, this list is hard coded in the
    *           command line itself
    *      <li> any number of files may be listed on the command line, subject
    *           to command shell limits
    *   </ul>
    *   <li> custom file list
    *   <ul>
    *      <li> &lt;filename&gt; is a single absolute or relative filename
    *           of a text file that contains a custom list of absolute and/or
    *           relative file names of files to scan
    *      <li> in other words, the file list will be read from a file instead
    *           of being hard coded on the command line
    *      <li> there must be one filename per line in the file
    *      <li> there is no limit of the number of files in the list
    *   </ul>
    *   <li> filespec mode
    *   <ul>
    *      <li> &lt;directory&gt; is the directory to search for files
    *      <li> &quot;filespec&quot; is the file specification that will be
    *           used to create a list of files to process, should be enclosed
    *           in double quotes if any of the wildcard characters '*' or '?'
    *           are used
    *   </ul>
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args) 
   {
      // option defaults
      ListType type         = ListType.CMD_LINE;
      boolean  recursion    = true;
      boolean  abortOnError = false;
      int      debug        = MSG_STATUS;
      int      idx          = 0;
      int      threads      = 1;
      
      // requires (mode + directory + filespec) OR (mode and a filelist)
      if (args.length < 2)
      {
         syntax(null, -1);
      }
      
      // process options
      if (args[0].startsWith("-"))
      {
         idx++;
         String opts = args[0].toLowerCase();
         
         if (opts.indexOf('s') != -1) type = ListType.FILESPEC;
         if (opts.indexOf('f') != -1) type = ListType.INPUT_FILE;
         if (opts.indexOf('n') != -1) recursion = false;
         if (opts.indexOf('a') != -1) abortOnError = true;
         
         int dbg = opts.indexOf('d');
         
         if (dbg != -1)
         {
            // get level (set to space char if of out-of-bounds to avoid exception)
            char level = (dbg + 1 < opts.length() ? opts.charAt(dbg + 1) : ' ');
            
            if (level == ' ' || !Character.isDigit(level))
            {
               syntax("Missing numeric debug level character following D option.", -2);
            }
            else
            {
               debug = Character.digit(level, 10);
               
               if (debug < MSG_NONE || debug > MSG_TRACE)
               {
                  syntax("Debug level must be between " + MSG_NONE + " and " + MSG_TRACE + ".",
                         -3);
               }
            }
         }
         
         int tpos = opts.indexOf('t');
         
         if (tpos != -1)
         {
            int start = tpos + 1;
            int next = start;
            int len = opts.length();
            
            while (next < len && Character.isDigit(opts.charAt(next)))
            {
               next++;
            }
            
            boolean fail = next == start;
            if (!fail)
            {
               try
               {
                  threads = Integer.parseInt(opts.substring(start, next));
               }
               catch (NumberFormatException exc)
               {
                  // should not be thrown, since we pre-validated characters as digits
                  fail = true;
               }
            }
            
            if (fail)
            {
               syntax("Missing number of threads following T option.", -4);
            }
         }
      }
      
      // process the mode argument
      String  mode = args[idx].toUpperCase();
      RunMode run  = interpretRunMode(mode);
      idx++;
      
      FileList files = null;
      
      // at this point idx points to the next arg after the options, this
      // must be one of our various forms of filename input
      switch (type)
      {
         case CMD_LINE:
            // there must be at least 1 arg left
            if (args.length <= idx)
            {
               syntax("Missing file list parameter.", -6);
            }
            
            int      len      = args.length;
            String[] filelist = new String[len - idx];
            
            for (int j = idx; j < len; j++)
            {
               filelist[j - idx] = args[j];
            }
            
            files = new ExplicitFileList(filelist);
            break;
         case INPUT_FILE:
         
            // there must be exactly 1 arg left
            if ((args.length - idx) != 1)
            {
               syntax("Missing custom file list parameter.", -7);
            }
            
            String[] textList = null;
            
            try (BufferedReader in  = new BufferedReader(new FileReader(args[idx])))
            {
               ArrayList<String> list = new ArrayList<>();
               String            next = null;
               
               while ((next = in.readLine()) != null)
               {
                  next = next.trim();
                  
                  if (!next.isEmpty())
                  {
                     list.add(next);
                  }
               }
               
               textList = (String[]) list.toArray(new String[0]);
            }
            
            catch (Exception exc)
            {
               System.err.println("ERROR:");
               exc.printStackTrace();
               return;
            }
            
            files = new ExplicitFileList(textList);
            break;
         case FILESPEC:
            // there must be exactly 2 args left
            if ((args.length - idx) != 2)
            {
               syntax("Missing/incomplete filespec parameter.", -8);
            }
            
            files = new FileSpecList(new File(args[idx]), args[idx+1], recursion);
            break;
      }
      
      ConversionDriver driver = new ConversionDriver(files, debug);
      
      try
      {
         // print the main header
         printHeader("P2J Conversion Driver", MSG_STATUS);
         
         // capture the start time
         long jobStartTime = System.currentTimeMillis();
         
         // front-end
         if (run.front)
         {
            driver.front(run.preproc, run.fschema, run.callgraph, abortOnError, threads);
         }
         
         // middle
         if (run.middle)
         {
            driver.middle(run.length);
         }
         
         // back-end
         if (run.back)
         {
            driver.back();
         }
         
         // print out total time
         printElapsed(System.currentTimeMillis() - jobStartTime);
      }
      
      catch (Exception exc)
      {
         System.err.println("ERROR:");
         exc.printStackTrace();
         System.exit(1);
      }
   }
}