/*
** Module   :ObjectClass.java
** Abstract :Base Object Class
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 SIY 20050222 ADD  @20032  Created initial version
** 002 SIY 20050429 CHG  @21023  Fixed comments and formatting.
** 003 SIY 20050501 CHG  @21198  Added quick attribute definition lookup, 
**                               updated documentation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.directory;

import java.util.HashMap;

/**
 * This class is used to hold P2J object class definition.
 * <p>
 * P2J Object Class definition contains following data: <br>
 * <ul>
 * <li>Object class name.</li>
 * <li>Boolean value which determines of object of this class can be updated
 * or they are read-only.</li>
 * <li>Boolean value which determines if nodes of this class can have
 * children nodes.</li>
 * <li>Definition of the attributes.</li>
 * </ul>
 * <p>
 * For performance reasons class uses shadow map which speedups attribute
 * definition lookup.
 * 
 * @author  SIY
 * @version 1.0
 */ 
class ObjectClass
{
   /** Object class name */
   private String name;
   
   /** Is object class terminal (leaf)? */
   private boolean isLeaf;
   
   /** Is object class read-only (immutable)? */
   private boolean isImmutable;
    
   /** Object class attribute definition */
   private AttributeDefinition[] attributes;

   /** Map for the quick attribute definition search */
   private HashMap attrMap;
   
   /**
    * Construct an instance of the class. Note that constructor is protected,
    * so instances can be created only by derived classes.
    * 
    * @param   name
    *          Object class name.
    * @param   isLeaf
    *          If <code>true</code> then class can have child nodes.
    * @param   isImmutable
    *          If <code>true</code> then class attributes can be changed.
    * @param   attributes
    *          List of class attributes.
    */
   protected ObjectClass(String name, boolean isLeaf, boolean isImmutable,
                         AttributeDefinition[] attributes)
   {
      this.name = name;
      this.isLeaf = isLeaf;
      this.isImmutable = isImmutable;
      this.attributes = attributes;
      attrMap = new HashMap();
      
      for (int i = 0; attributes != null && i < attributes.length; i++)
      {
         attrMap.put(attributes[i].getName(), attributes[i]);
      }
   }
                  
   /**
    * Class is leaf if no other object can be created under objects of the
    * class. Also known as terminal.
    * 
    * @return  <code>true</code> if class is leaf.
    */
   boolean isClassLeaf()
   {
      return isLeaf;
   }

   /**
    * Class is immutable if the back-end does not allow any change to its
    * state.
    * 
    * @return  <code>true</code> if class is read-only.
    */
   boolean isClassImmutable()
   {
      return isImmutable;
   }
   
   /**
    * Get definitions for all class attributes.
    * 
    * @return  An array of objects describing class attributes.
    */
   AttributeDefinition[] getClassDefinition()
   {
      return attributes;
   }

   /**
    * Get definitions for specified attribute.
    * 
    * @param   attribute
    *          Attribute name.
    * @return  Definition of specified attribute.
    */
   AttributeDefinition getAttributeDefinition(String attribute)
   {
      return (AttributeDefinition)attrMap.get(attribute);
   }
   
   /**
    * Returns Object class name.
    * 
    * @return  Name of the Object class.
    */
   String getName()
   {
      return name;
   }
}
