/*
** Module   :Remapper.java
** Abstract :Back-end interface declaration.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20050217   @20048 Created initial version
** 002 SIY 20050310   @20369 Reordered members.
** 003 SIY 20050328   @20585 Fixed comments.
** 004 SIY 20050429   @21027 Formatting cleanups.
** 005 SIY 20050505   @21201 Changed visibility of interface to public to 
**                           allow implementation of this interface in 
**                           other packages. Updated documentation.
** 006 SIY 20090514   @42181 Added refresh() method. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.directory;

/**
 * Directory back-end interface. All classes which implement back-end
 * functionality implement it.
 * 
 * @author  SIY
 * @version 1.0
 */
public interface Remapper
{
   /**
    * Return class name for the specified node.
    * 
    * @param   nodeId
    *          Node for which class name is requested.
    * @return  Node class name.
    */
   public String getNodeClassName(String nodeId);

   /**
    * Add new node of specified class with specified ID. For node
    * initialization list of attributes and corresponding list of attribute
    * values should be provided. The presence of the mandatory attributes in
    * the lists is obligatory.
    * 
    * @param   nodeId
    *          Node name.
    * @param   nodeClass
    *          Node class name.
    * @param   names
    *          A list of names of the attributes.
    * @param   values
    *          A list of values of the attributes.
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   boolean addNode(String nodeId, String nodeClass, String[] names,
                   Object[] values);

   /**
    * Add new <code>BitField</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeBitField(String nodeId, String name, BitField value);

   /**
    * Add new <code>BitSelector</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeBitSelector(String nodeId, String name, BitSelector value);

   /**
    * Add new <code>boolean</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeBoolean(String nodeId, String name, boolean value);

   /**
    * Add new <code>byte array</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeByteArray(String nodeId, String name, byte[] value);

   /**
    * Add new <code>DateValue</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeDate(String nodeId, String name, DateValue value);

   /**
    * Add new <code>double</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeDouble(String nodeId, String name, double value);

   /**
    * Add new <code>int</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */ 
   boolean addNodeInteger(String nodeId, String name, int value);

   /**
    * Add new <code>String</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeString(String nodeId, String name, String value);

   /**
    * Add new <code>TimeValue</code> value to the attribute.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean addNodeTime(String nodeId, String name, TimeValue value);

   /**
    * Remove specified node from the directory (unless node has children).
    * 
    * @param   nodeId
    *          Node identifier.
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   boolean deleteNode(String nodeId);

   /**
    * Remove whole node attribute. Note that call will fail if attribute is
    * marked mandatory.
    * 
    * @param   nodeId
    *          Node name.
    * @param   name
    *          Attribute name.
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   boolean deleteNodeAttribute(String nodeId, String name);

   /**
    * Remove particular value from the node attribute. Note that removing last
    * value from the attribute deletes entire attribute unless attribute is
    * marked mandatory.
    * 
    * @param   nodeId
    *          Node name.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Attribute value index.
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   boolean deleteNodeAttributeValue(String nodeId, String name, int index);

   /**
    * Return list of all node attribute definitions.
    * 
    * @param   nodeId
    *          Node for which list of attributes is requested.
    * @return  Array of attribute definitions.
    */
   AttributeDefinition[] enumerateNodeAttributes(String nodeId);

   /**
    * Get direct list of children nodes for current node.
    * 
    * @param   nodeId
    *          Node for which list of nodes is requested.
    * @return  An array of children node names.
    */
   String[] enumerateNodes(String nodeId);

   /**
    * Get definitions for all class attributes.
    * 
    * @param   className
    *          Name of the class.
    * @return  An array of objects describing class attributes.
    */
   AttributeDefinition[] getClassDefinition(String className);

   /**
    * Get list of names of all defined classes.
    * 
    * @return  Array of names of the classes.
    */
   String[] getClassNames();

   /**
    * Get <code>BitField</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>BitField</code> then <code>null</code>
    * is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   BitField getNodeBitField(String nodeId, String name, int index);

   /**
    * Get <code>BitSelector</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>BitSelector</code> then
    * <code>null</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   BitSelector getNodeBitSelector(String nodeId, String name, int index);

   /**
    * Get <code>Boolean</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>Boolean</code> then <code>null</code>
    * is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   Boolean getNodeBoolean(String nodeId, String name, int index);

   /**
    * Get <code>byte array</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>byte array</code> then
    * <code>null</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   byte[] getNodeByteArray(String nodeId, String name, int index);

   /**
    * Get <code>DateValue</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>DateValue</code> then <code>null</code>
    * is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   DateValue getNodeDate(String nodeId, String name, int index);

   /**
    * Get <code>Double</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>Double</code> then <code>null</code>
    * is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   Double getNodeDouble(String nodeId, String name, int index);

   /**
    * Get <code>Integer</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>Integer</code> then <code>null</code>
    * is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   Integer getNodeInteger(String nodeId, String name, int index);

   /**
    * Get <code>String</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>String</code> then <code>null</code>
    * is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   String getNodeString(String nodeId, String name, int index);

   /**
    * Get <code>TimeValue</code> attribute value for specified node ID and
    * index. If wrong parameter (node ID, attribute name or index) are passed
    * or attribute type is not <code>TimeValue</code> then <code>null</code>
    * is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @return  Value of the attribute for the specified attribute name and
    *          index.
    */
   TimeValue getNodeTime(String nodeId, String name, int index);

   /**
    * Class is immutable if the back-end does not allow any change to its
    * state.
    * 
    * @param   className
    *          Name of the class.
    * @return  <code>true</code> if class is read-only.
    */
   boolean isClassImmutable(String className);

   /**
    * Class is leaf if no other object can be created under objects of the
    * class. Also known as terminal.
    * 
    * @param   className
    *          Name of the class.
    * @return  <code>true</code> if class is leaf.
    */
   boolean isClassLeaf(String className);

   /**
    * Moves or renames the existing object, specified by nodeId to different
    * location. The parent object of the newId should exist and allow for
    * children creation.
    * 
    * @param   nodeId
    *          An ID of the node to move.
    * @param   newId
    *          The desired new location of the node.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   public boolean moveNode(String nodeId, String newId);

   /**
    * Set <code>BitField</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeBitField(String nodeId, String name, int index,
                           BitField value);

   /**
    * Set <code>BitSelector</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeBitSelector(String nodeId, String name, int index,
                              BitSelector value);

   /**
    * Set <code>boolean</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeBoolean(String nodeId, String name, int index, boolean value);

   /**
    * Set <code>byte array</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeByteArray(String nodeId, String name, int index,
                            byte[] value);

   /**
    * Set <code>DateValue</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeDate(String nodeId, String name, int index, DateValue value);

   /**
    * Set <code>double</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeDouble(String nodeId, String name, int index, double value);

   /**
    * Set <code>int</code> attribute value for specified node ID, attribute
    * name and index. If wrong parameter is passed or attribute read-only then
    * <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeInteger(String nodeId, String name, int index, int value);

   /**
    * Set <code>String</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeString(String nodeId, String name, int index, String value);

   /**
    * Set <code>TimeValue</code> attribute value for specified node ID,
    * attribute name and index. If wrong parameter is passed or attribute
    * read-only then <code>false</code> is returned.
    * 
    * @param   nodeId
    *          Node identifier.
    * @param   name
    *          Attribute name.
    * @param   index
    *          Index of the value to retrieve.
    * @param   value
    *          New value for the attribute variable.
    * @return  <code>true</code> is operation is successful and
    *          <code>false</code> otherwise.
    */
   boolean setNodeTime(String nodeId, String name, int index, TimeValue value);

   /**
    * Initialise back-end and open session.
    * 
    * @return  <code>true</code> if operation was successful.
    */
   boolean bind();

   /**
    * Close session and shutdown back-end.
    * 
    * @return  <code>true</code> if operation was successful.
    */
   boolean unbind();

   /**
    * Update external storage. This method is invoked by the DirectoryService
    * to commit the changes made by batch editing into external storage.
    * 
    * @return  <code>true</code> if operation was successful.
    */
   boolean update();

   /**
    * Reload directory content from external source.
    * <p>
    * Note that it is backend responsibility to preserve its state when 
    * refresh operation was not successful.
    * 
    * @return   <code>true</code> if operation was successful.
    */
   boolean refresh();
}