/*
** Module   : ByteCountOutputStream.java
** Abstract : provides an output filter that counts the number of bytes
**            written
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050430 NEW   @20947 First version which tracks the number of
**                               bytes written as a filter implementation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.io;

import java.io.*;

/**
 * Provides an output filter that counts the number of bytes written.  All
 * other function is redirected directly to the encapsulated output stream.
 *
 * @author    GES
 * @version   1.0.0
 */
public class ByteCountOutputStream extends FilterOutputStream
{   
   /** Tracks the total number of bytes written. */
   private long bytes = 0;
   
   /**
    * Creates an instance with the associated output stream as the base.
    */
   public ByteCountOutputStream(OutputStream out)
   {
      super(out);
   }

   /**
    * Accesses the current counter of the number bytes written to this
    * stream.
    *
    * @return   The current bytes written counter.
    */
   public long getBytesWritten()
   {
      return bytes;
   }

   /**
    * Writes the specified byte to this output stream (only the low order 8
    * bits are written, the high order 24 bits are ignored).
    * 
    * @param    b
    *           The byte to write. 
    *
    * @throws   IOException
    *           if an I/O error occurs. In particular, an IOException may be
    *           thrown if the output stream has been closed.
    */
   public void write(int b)
   throws IOException
   {
      // update our counter
      bytes++;
      
      // OK, actually write the byte
      out.write(b);
   }
      
   /**
    * Writes the entire contents of the specified byte array to this output
    * stream.  
    * 
    * @param    b
    *           The byte array to write. 
    *
    * @throws   IOException
    *           if an I/O error occurs. In particular, an IOException may be
    *           thrown if the output stream has been closed.
    * @throws   NullPointerException   
    *           if the array is <code>null</code>. 
    */
   public void write(byte[] b)
   throws IOException,
          NullPointerException
   {
      write(b, 0, b.length);
   }
   
   /**
    * Writes <code>len</code> bytes from the specified byte array starting
    * at offset <code>off</code> in this output stream. Element b[off] is
    * the first byte written and b[off+len-1] is the last byte written.
    * 
    * @param    b
    *           The array of bytes to write. 
    * @param    off
    *           the start offset in the array.
    * @param    len
    *           the number of bytes to write.
    *
    * @throws   IOException
    *           if an I/O error occurs. In particular, an IOException may be
    *           thrown if the output stream has been closed.
    * @throws   NullPointerException   
    *           if the array is <code>null</code>. 
    * @throws   IndexOutOfBoundsException   
    *           if the offset or length are negative, or if offset + length
    *           is larger than the length of the array.
    */
   public void write(byte[] b, int off, int len)
   throws IOException,
          NullPointerException,
          IndexOutOfBoundsException
   {
      // update our counter
      bytes += len;
      
      // delegate the real work
      out.write(b, off, len);
   }   
}
