/*
** Module   : IndentHandler.java
** Abstract : allows caller-controlled indenting management 
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20080309 NEW   @37538 Created initial version which allows
**                               caller-controlled indenting management.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.io;

import java.io.*;

/**
 * Provides management of indention state.
 * <p>
 * The <code>0</code> indent level is equivalent to no indents. For each
 * positive indent level, the number of space characters written will be
 * equivalent to the indent level times the width of a single indent level
 * (which defaults to 3 spaces).
 * <p>
 * So at the <code>0</code> indent level, no spaces are written.  At the
 * <code>1</code> indent level, 3 spaces (by default) are written.  At the
 * <code>2</code> indent level, 6 spaces (by default) are written and so on.
 * By changing the width and indent level, the exact indent results can be
 * completely controlled.
 * <p>
 * This class is not thread-safe.
 *
 * @author    GES
 */
public class IndentHandler
implements Indentable
{
   /** Current indent level (defaults to 0). */
   private int indent = 0;
   
   /** Width of each indent level (defaults to 3 space characters). */
   private int width = 3;
   
   /** Current pad inserted between newlines and subsequent bytes. */
   private StringBuilder pad = new StringBuilder();
   
   /**
    * Accesses the current indent level.
    *
    * @return   Indent level.
    */
   public int getIndent()
   {
      return indent;
   }

   /**
    * Sets the current indent level.
    *
    * @param    indent
    *           New indent level.
    */
   public void setIndent(int indent)
   {
      this.indent = indent;
      pad(indent * width);
   }

   /**
    * Increments the indent level by 1.
    */
   public void increment()
   {
      pad(++indent * width);
   }

   /**
    * Decrements the indent level by 1.
    */
   public void decrement()
   {
      pad(--indent * width);
   }

   /**
    * Accesses the number of spaces in a single indent level.
    *
    * @return   Width of each indent level.
    */
   public int getWidth()
   {
      return width;
   }

   /**
    * Sets the number of spaces in a single indent level.
    *
    * @param    width
    *           New indent width.
    */
   public void setWidth(int width)
   {
      this.width = width;
      pad(indent * width);
   }

   /**
    * Sets both the indent level and the number of spaces in a single indent
    * level. This is a convenience function to avoid 2 separate calls and
    * the small amount of shared overhead that would have been incurred in
    * those calls.
    *
    * @param    indent
    *           New indent level.
    * @param    width
    *           New indent width.
    */
   public void setPad(int indent, int width)
   {
      this.indent = indent;
      this.width = width;
      pad(indent * width);
   }
   
   /**
    * Get the current pad.
    *
    * @return   The pad string for the next line of output.
    */
   public String getPad()
   {
      return pad.toString();
   }
   
   /**
    * Create a pad string containing <code>sz</code> space characters.
    *
    * @param    sz
    *           The number of spaces to put into the pad string.
    */
   private void pad(int sz)
   {
      int len = pad.length();
      
      if (len != sz)
      {
         pad.setLength(sz);
         
         // if we are truncating, this code won't execute, but if we are
         // extending, it will add the needed spaces 
         for (int i = len; i < sz; i++)
         {
            pad.setCharAt(i, ' ');
         }
      }  
   }
}