/*
** Module   : IndentingOutputStream.java
** Abstract : allows caller-controlled indenting of output 
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20070711 NEW   @34476 Created initial version which allows
**                               caller-controlled indenting of output.
** 002 GES 20080318 CHG   @37539 Moved pad handling to a separate class.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.io;

import java.io.*;

/**
 * Provides <code>OutputStream</code> semantics with caller-controlled
 * indenting.  An indent is written (inserted) between each newline and any
 * subsequent (non-newline) bytes that are written.
 * <p>
 * This class is not thread-safe.
 *
 * @author    GES
 */
public class IndentingOutputStream
extends FilterOutputStream
implements Indentable
{
   /** Indent state manager. */
   private Indentable indent = new IndentHandler();
   
   /** The last byte written to the stream. */
   private byte last = (byte) '\n';
   
   /**
    * Creates an instance with the associated output stream as the base and
    * a default 0 indent level.
    *
    * @param    out
    *           The stream on which we filter.
    */
   public IndentingOutputStream(OutputStream out)
   {
      super(out);
   }

   /**
    * Creates an instance with the associated output stream as the base.
    *
    * @param    out
    *           The stream on which we filter.
    * @param    indent
    *           The initial indent level.
    */
   public IndentingOutputStream(OutputStream out, int indent)
   {
      this(out);
      setIndent(indent);
   }

   /**
    * Accesses the current indent level.
    *
    * @return   Indent level.
    */
   public int getIndent()
   {
      return indent.getIndent();
   }

   /**
    * Sets the current indent level.
    *
    * @param    indent
    *           New indent level.
    */
   public void setIndent(int indent)
   {
      this.indent.setIndent(indent);
   }

   /**
    * Increments the indent level by 1.
    */
   public void increment()
   {
      indent.increment();
   }

   /**
    * Decrements the indent level by 1.
    */
   public void decrement()
   {
      indent.decrement();
   }

   /**
    * Accesses the number of spaces in a single indent level.
    *
    * @return   Width of each indent level.
    */
   public int getWidth()
   {
      return indent.getWidth();
   }

   /**
    * Sets the number of spaces in a single indent level.
    *
    * @param    width
    *           New indent width.
    */
   public void setWidth(int width)
   {
      indent.setWidth(width);
   }

   /**
    * Get the current pad.
    *
    * @return   The pad string for the next line of output.
    */
   public String getPad()
   {
      return indent.getPad();
   }
   
   /**
    * Sets both the indent level and the number of spaces in a single indent
    * level. This is a convenience function to avoid 2 separate calls and
    * the small amount of shared overhead that would have been incurred in
    * those calls.
    *
    * @param    indent
    *           New indent level.
    * @param    width
    *           New indent width.
    */
   public void setPad(int indent, int width)
   {
      this.indent.setPad(indent, width);
   }
   
   /**
    * Writes the specified byte to this output stream (only the low order 8
    * bits are written, the high order 24 bits are ignored).
    * <p>
    * The current pad is inserted into the stream between all newline
    * characters and any subsequent non-newline bytes.
    * 
    * @param    b
    *           The byte to write. 
    *
    * @throws   IOException
    *           If an I/O error occurs. In particular, an IOException may be
    *           thrown if the output stream has been closed.
    */
   public void write(int b)
   throws IOException
   {
      write(new byte[] { (byte) b });
   }
      
   /**
    * Writes the entire contents of the specified byte array to this output
    * stream.  
    * <p>
    * The current pad is inserted into the stream between all newline
    * characters and any subsequent non-newline bytes.
    * 
    * @param    b
    *           The byte array to write. 
    *
    * @throws   IOException
    *           If an I/O error occurs. In particular, an IOException may be
    *           thrown if the output stream has been closed.
    * @throws   NullPointerException   
    *           If the array is <code>null</code>. 
    */
   public void write(byte[] b)
   throws IOException,
          NullPointerException
   {
      write(b, 0, b.length);
   }
   
   /**
    * Writes <code>len</code> bytes from the specified byte array starting
    * at offset <code>off</code> in this output stream. Element b[off] is
    * the first byte written and b[off+len-1] is the last byte written.
    * <p>
    * The current pad is inserted into the stream between all newline
    * characters and any subsequent non-newline bytes.
    * 
    * @param    b
    *           The array of bytes to write. 
    * @param    off
    *           The start offset in the array.
    * @param    len
    *           The number of bytes to write.
    *
    * @throws   IOException
    *           If an I/O error occurs. In particular, an IOException may be
    *           thrown if the output stream has been closed.
    *
    * @throws   NullPointerException   
    *           If the array is <code>null</code>. 
    * @throws   IndexOutOfBoundsException   
    *           If the offset or length are negative, or if offset + length
    *           is larger than the length of the array.
    */
   public void write(byte[] b, int off, int len)
   throws IOException,
          NullPointerException,
          IndexOutOfBoundsException
   {
      boolean empty = (indent.getIndent() == 0 && indent.getWidth() == 0);
      
      if (empty)
      {
         // no identing needed, use faster approach
         out.write(b, off, len);
         
         // remember the last char
         last = b[off + len - 1];
      }
      else
      {
         // indent needed, process one byte at a time
         for (int i = 0; i < len; i++)
         {
            byte next = b[i + off];
            
            // look for an output of a non-newline that follows a newline
            if (last == '\n' && next != '\n')
            {
               // insert the indent
               out.write(indent.getPad().getBytes());
            }
            
            // write the current character
            out.write(next);
            
            // update the last character
            last = next;
         }
      }
   }
}