/*
** Module   : NativePrimitive.java
** Abstract : descriptor base class for scalar data as an argument or return value to a native
**            function 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131212 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * A descriptor base class for scalar data as an argument or return value to a native function.
 */
public abstract class NativePrimitive
extends BaseNativeType
{
   /**
    * Default constructor.
    */
   public NativePrimitive()
   {
   }
   
   /**
    * Render the data into native memory, which will be allocated if the address is not
    * given.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address to which the data should be rendered, if the caller
    *           is supplying a buffer or 0 if the buffer should be allocated.
    *
    * @return   The memory address of the rendered data or 0 if there is no data to
    *           output.
    */
   public long render(long addr)
   {
      int sz = size();
      
      memptr local = new memptr();
      
      if (addr == 0)
      {
         // save off for restore use later
         ref = local;
         
         // allocate the block which must be big enough to store the data
         local.setLength(sz);
         
         addr = local.getPointerValue().longValue();
      }
      else
      {
         local.setPointerValue(addr);
      }
      
      if (copyIn)
      {
         writeNative(local);
      }
      
      return addr;
   }
   
   /**
    * Copy back (if necessary) data from the native memory and deallocate the memory
    * if this instance allocated it in the first place.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address from which the data should be restored, if the caller
    *           is supplying a buffer or 0 if the buffer should already exist.
    */
   public void restore(long addr)
   {
      memptr local = ref;
      
      if (local == null && addr != 0)
      {
         local = new memptr();
         local.setPointerValue(addr);
      }
      
      if (copyOut && local != null)
      {
         readNative(local);
      }
      
      // deallocate memory if we allocated it
      if (ref != null)
         ref.setLength(0);
   }
   
   /**
    * Worker method to read the primitive value from the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected abstract void readNative(memptr ptr);
   
   /**
    * Worker method to write the primitive value to the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected abstract void writeNative(memptr ptr);
}