/*
** Module   : AppServerInvocationResult.java
** Abstract : Container for the result of a remote appserver procedure call. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130829 Return the resolved parameter modes to the caller too.
** 003 CA  20131220 Extracted data common with async web service calls to RemoteInvocationResult.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.*;

import com.goldencode.p2j.util.*;

/**
 * This class acts as a container for the various data which is returned by or updated during 
 * remote appserver calls.  This includes the function/procedure return value, updated 
 * OUTPUT/INPUT-OUTPUT parameters or the persistent procedure handle for a RUN ... PERSISTENT
 * call.
 */
public class AppServerInvocationResult
extends RemoteInvocationResult
implements Externalizable
{
   /** The remote procedure handle, used to build a {@link ProxyProcedureWrapper} instance. */
   private String procedureId;
   
   /** The connection ID, returned during the connect request. */
   private String connectionID;
   
   /**
    * Serialize this object, to be picked up by the other side.
    *  
    * @param    out
    *           The output destination to which the data will be sent.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(arguments.length);
      for (int i = 0; i < arguments.length; i++)
      {
         out.writeObject(arguments[i]);
      }
      
      out.writeObject(result);
      out.writeObject(procedureId);
      out.writeObject(error);
      out.writeObject(connectionID);
      out.writeObject(modes);
   }

   /**
    * Used the specified input source and initialize this instance. 
    *  
    * @param    in
    *           The input source from which the data will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      int parameterCount = in.readInt();
      arguments = new Object[parameterCount];
      for (int i = 0; i < parameterCount; i++)
      {
         arguments[i] = in.readObject();
      }

      result = (BaseDataType) in.readObject();
      procedureId = (String) in.readObject();
      error = (Throwable) in.readObject();
      connectionID = (String) in.readObject();
      modes = (String) in.readObject();
   }

   /**
    * Get the handle code for the remote persistent procedure.
    * 
    * @return   The {@link #procedureId} field.
    */
   public String getProcedureId()
   {
      return procedureId;
   }

   /**
    * Save the remote persistent procedure handle's code.
    * 
    * @param    procedureId
    *           The code for the remote persistent procedure handle.
    */
   public void setProcedureId(String procedureId)
   {
      this.procedureId = procedureId;
   }

   /**
    * Get the established connection ID, for a connect request.
    * 
    * @return   See above.
    */
   public String getConnectionID()
   {
      return connectionID;
   }

   /**
    * Set the new connection ID, after a connect request.
    * 
    * @param    connectionID
    *           The connection ID, which will be unknown if there are any {@link #error errors}.
    */
   public void setConnectionID(String connectionID)
   {
      this.connectionID = connectionID;
   }
   
   /**
    * Check if the RETURN-VALUE will be set.
    * 
    * @return   In case of appserver results, it returns <code>true</code>.
    */
   public boolean setReturnValue()
   {
      return true;
   }
}