/*
** Module   : AppletHandler.java
** Abstract : delivers applet pages and the jar resources needed
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 GES 20110929 First version based on code from AdminGate.
** 002 MAG 20131101 Upgrade to jetty 9.1 embedded server
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.server.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.util.*;

/**
 * Delivers applet pages and the jar resources needed as a Jetty handler.
 */
public class AppletHandler
extends AbstractHandler
{
   /** Logger. */
   private static final Logger LOG = LogHelper.
                                     getLogger(AppletHandler.class.getName());
   
   /** The root portion of the target request name (part of the URL). */
   private String targetRoot = null;
   
   /** The applet's html skeleton page. */
   private String skeletonPage = null;

   /** The server's hostname. */
   private String hostname = null;

   /** The server's secure port. */
   private String securePort = null;

   /** List of virtual JAR names to be delivered as the archive attribute. */
   private String jarList = null;
   
   /** Handler for processing JAR resource requests. */
   private JarResourceHandler jarMgr = null;
   
   /**
    * Constructor.
    *
    * @param    targetRoot
    *           Prefix for all URLs relating to this applet.
    * @param    skeletonPage
    *           Fully qualified resource name of the HTML file which should be
    *           used to load the applet.
    * @param    classList
    *           List of class names which must be exposed via JAR resources.
    *           If this is <code>null</code> or a single element array with
    *           no valid element, then the current class' JAR file will be
    *           found.
    */
   public AppletHandler(String   targetRoot,
                        String   skeletonPage,
                        String[] classList)
   {
      if (!targetRoot.startsWith("/"))
         targetRoot = "/" + targetRoot;
      
      if (targetRoot.endsWith("/"))
         targetRoot = targetRoot.substring(0, targetRoot.length() - 1);
      
      this.targetRoot   = targetRoot;
      this.skeletonPage = skeletonPage;
      
      HashMap<String, String> jarMap = setupJarMappings(targetRoot, classList);
      jarMgr = new JarResourceHandler(jarMap);
   }
   
   /**
    * Configure JAR file mappings that are necessary to allow the given class
    * list to be served to the applet.
    *
    * @param    targetRoot
    *           Prefix for all URLs relating to this applet.
    * @param    list
    *           List of class names which must be exposed via JAR resources.
    *           If this is <code>null</code> or a single element array with
    *           no valid element, then the current class' JAR file will be
    *           found.
    */
   private HashMap<String, String> setupJarMappings(String   targetRoot,
                                                    String[] list)
   {
      HashMap<String, String> jarMap = new HashMap<String, String>();
      
      if (list == null ||
          (list.length == 1 && (list[0] == null || list[0].length() == 0)))
      {
         list = new String[1];
         list[0] = getClass().getName();
      }
      
      StringBuilder allJars = new StringBuilder();

      for (int i = 0; i < list.length; i++)
      {
         if (list[i] != null)
         {
            String realJarName = getJarName(list[i]);
            
            if (realJarName != null)
            {
               if (allJars.length() > 0)
                  allJars.append(',');
               
               String baseName    = getBaseName(realJarName);
               String virtJarName = targetRoot + "/" + baseName; 
               
               allJars.append(virtJarName);
               jarMap.put(virtJarName, realJarName);
            }
         }
      }
      
      jarList = allJars.toString();
      
      return jarMap;
   }
   
   /**
    * Return either the main applet HTML page for the configured applet OR 
    * a valid JAR file referenced from that page.
    *    
    * @param    target 
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * @param    response 
    *           The object or a wrapper of the response. 
    */
   public void handle(String              target, 
                      Request             base,
                      HttpServletRequest  request, 
                      HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      String  lower = target.toLowerCase();
         
      if (base.isHandled() || !lower.startsWith(targetRoot.toLowerCase()))
      {
         return;
      }
      
      if (target.equalsIgnoreCase(targetRoot))
      {
         generateAppletPage(base, response);
      }
      else
      {
         // may return HttpServletResponse.SC_FORBIDDEN if the target is
         // not valid, otherwise respond with the jar file
         jarMgr.handle(target, base, request, response);
      }
   }
   
   /**
    * Process the given string and make any replacements of parameter values
    * as needed.  Generally, the format of the replacement placeholder is
    * a % character followed by a placeholder name, but this is just a
    * convention.
    *
    * @param    base
    *           The details and state the request.
    * @param    text
    *           The text line that may need modifications.
    *
    * @return   The new text if a modification occurred or the original text
    *           if no modification was needed.
    */
   protected String handleReplacements(Request base, String text)
   {
      if (text.contains("p2jservername"))
      {
         text = text.replace("%hostname", hostname);
      }
      else if (text.contains("p2jserverport"))
      {
         text = text.replace("%portnumber", getSecurePort());
      }
      else if (text.contains("java_arguments"))
      {
         // support JVM debug parameters
         String jvmOptions = "";
         
         if (base.getParameter("debug") != null)
         {
            jvmOptions = "-Xdebug -Xnoagent -Djava.compiler=NONE "     +
                         "-Xrunjdwp:transport=dt_socket,address=2999," +
                         "server=y,suspend=y";
         }
         
         text = text.replace("%jvmparms", jvmOptions);
      } 
      else if (text.contains("%archives"))
      {
         text = text.replace("%archives", jarList);
      }
      
      return text;
   }
   
   /**
    * Use the applet's skeleton page to generate a response that is a valid
    * HTML page which includes a runtime-customized applet tag.
    *
    * @param    base
    *           The details and state the request.
    * @param    response 
    *           The details and state of the response. 
    */
   private void generateAppletPage(Request base, HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      // mark the request as handled
      base.setHandled(true);
      
      // save off the server's hostname (one time only)
      synchronized (this)
      {
         if (hostname == null)
         {
            hostname = base.getLocalName();
         }
      }
      
      // produce the page contents
      response.setContentType("text/html");
      response.setStatus(HttpServletResponse.SC_OK);
      
      PrintWriter writer = response.getWriter();
      
      try
      {
         InputStream is = this.getClass().getResourceAsStream(skeletonPage);
         BufferedReader reader = new BufferedReader(new InputStreamReader(is));
         
         String str = null;
         
         // read the skeleton page and translate %substitutions
         // on the fly
         while (null != (str = reader.readLine()))
         {
            writer.println(handleReplacements(base, str));
         }
         
         reader.close();
      }
      
      catch (IOException ioe)
      {
         LOG.logp(Level.SEVERE,
                  "AppletHandler.generateAppletPage()",
                  "",
                  LogHelper.generate("IOException!"),
                  ioe);
      }
   }
      
   /**
    * Convert the given fully qualified class name into the real name of the
    * JAR file in which that class can be found.
    *
    * @param    cls
    *           The fully qualified class name. It can have the ".class" at the
    *           end but it is not required.
    *
    * @return   The JAR name in which the class can be found or
    *           <code>null</code> if the class cannot be found or no JAR exists
    *           for that class.
    */
   private String getJarName(String cls)
   {
      StringBuilder sb = new StringBuilder();
      
      if (cls.charAt(0) != '/')
         sb.append('/');
      
      sb.append(cls.replace('.', '/'));
      
      if (!cls.endsWith(".class"))
         sb.append(".class");
      
      URL url = getClass().getResource(sb.toString());
      
      cls = (url == null) ? null : url.getPath();
      
      if (cls != null)
      {
         int j = cls.indexOf(".jar!");
         
         if (j > 0)
         {
            cls = cls.substring(5, j + 4);
         }
         else
         {
            cls = null;
         }
      }
      
      return cls;
   }
   
   /**
    * Return the portion of the name after the last path separator OR the
    * original parameter if there is no path separator.
    *
    * @param    full
    *           Fully qualified resource name.
    *
    * @return   The calculated base name.
    */
   private String getBaseName(String full)
   {
      int idx = full.lastIndexOf("/");
      return idx >= 0 ? full.substring(idx) : full;
   }
   
   /**
    * Obtains the server's secure port for P2J communications.
    *
    * @return   The secure port.
    */
   private synchronized String getSecurePort()
   {
      if (securePort == null)
      {
         RouterSessionManager rsm = (RouterSessionManager)
                                    RouterSessionManager.get(); 
         
         securePort = Integer.toString(rsm.getSSLSocket().getLocalPort());
      }
      
      return securePort;
   }
   
   /**
    * Respond to JAR resource requests for an applet, which the browser issues
    * automatically based on applet tag's ARCHIVE attribute.
    * <p>
    * The virtual resource names are just the jar file names, relative to the
    * target (virtual) root for the given applet. Only those requests will be
    * honored, that match one of the known virtual jar names. These requests
    * are remapped to the real jar files.
    * <p>
    * All other static resource requests will fail.
    */
   public static class JarResourceHandler 
   extends ResourceHandler
   {
      /** Virtual to real JAR name mapping. */
      private HashMap<String, String> jarMap = new HashMap<String, String>();
      
      /**
       * Constructor.
       *
       * @param    jarMap
       *           Map of virtual jar names to real names.
       */
      public JarResourceHandler(HashMap<String, String> jarMap)
      {
         this.jarMap = jarMap;
      }
      
      /**
       * Returns a JAR resource if it is one of the JAR files associated with
       * the current applet.
       *    
       * @param    target 
       *           The target of the request - either a URI or a name.
       * @param    base
       *           The base request.
       * @param    request
       *           The object or a wrapper of the request.
       * @param    response 
       *           The object or a wrapper of the response. 
       */
      public void handle(String              target, 
                         Request             base,
                         HttpServletRequest  request, 
                         HttpServletResponse response) 
      throws IOException, 
             ServletException
      {

         if (base.isHandled())
         {
            return;
         }
         
         // check to see if it's a request for one of the allowed jar files
         String real = jarMap.get(target);
         
         // return error if no match
         if (real == null)
         {
            base.setHandled(true);
            response.sendError(HttpServletResponse.SC_FORBIDDEN);
            return;
         }
         
         synchronized (JarResourceHandler.class)
         {
            int idx = real.lastIndexOf("/");
            setResourceBase(real.substring(0, idx));
            base.setPathInfo(real.substring(idx));
            
            // delegate the work to the base resource handler
            super.handle(real, base, request, response);
         }
      }
   }
}