/*
** Module   : ClientBuilderParameters.java
** Abstract : store the process parameters.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131119 First version.
** 002 MAG 20140206 Add referrer property.
** 003 CA  20140206 Refactored to make it unaware of the type of the P2J client being spawned.
** 004 MAG 20140224 Add support for outputToFile option.
** 005 GES 20150211 Removed application arguments support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.*;

import com.goldencode.p2j.security.SecurityManager;

/**
 * This structure holds default parameters as well specific parameters used to create and spawn a
 * P2J client process. 
 */
public class ClientBuilderParameters
{   
   /** Default options from directory */
   protected final ClientBuilderOptions options;

   /** Client assigned uuid */
   private String uuid;

   /** User name */
   private String userName;
   
   /** Password */
   private String password;

   /**
    * Constructor. Assign a unique UUID to this process.
    */
   public ClientBuilderParameters(ClientBuilderOptions options)
   {
      this.options = options;
      uuid = UUID.randomUUID().toString();
   }

   /**
    * Get the authentication mode for this spawner.
    * 
    * @return   {@code true} indicating that the user's password is required for authentication.
    */
   public boolean isPasswordAuthentication()
   {
      return options.isPasswordAuthentication();
   }

   /**
    * Get process assigned UUID.
    * 
    * @return  The assigned UUID. 
    */
   public String getUuid()
   {
      return uuid;
   }
   
   /**
    * Get user name.
    * 
    * @return  User name.
    */
   public String getUserName()
   {
      return userName;
   }

   /**
    * Set user name.
    * 
    * @param   userName
    *          User name.
    */
   public void setUserName(String userName)
   {
      this.userName = userName;
   }

   /**
    * Get user password;
    * 
    * @return  User password.
    */
   public String getPassword()
   {
      return password;
   }

   /**
    * Set user password.
    * 
    * @param   password
    *          User password;
    */

   public void setPassword(String password)
   {
      this.password = password;
   }

   /**
    * Get the security mode..
    * 
    * @return  <code>true</code> use a secured connection <code>false</code> otherwise.
    */
   public boolean isSecure()
   {
      return options.isSecure();
   }

   /**
    * Get command
    * 
    * @return  Command name string.
    */
   public String getCommand()
   {
      return options.getCommand();
   }

   /**
    * Get classpath
    * 
    * @return  Class path string.
    */
   public String getClassPath()
   {
      return options.getClassPath();
   }

   /**
    * Get native library path.
    * 
    * @return  Native library path string.
    */
   public String getLibPath()
   {
      return options.getLibPath();
   }

   /**
    * Get user working directory.
    * 
    * @return  User working directory.
    */
   public String getWorkingDirectory()
   {
      return options.getWorkingDirectory();
   }

   /**
    * Get configuration file name.
    * 
    * @return  Configuration file name.
    */
   public String getConfigFile()
   {
      return options.getConfigFile();
   }
   
   /**
    * Get JVM arguments
    * 
    * @return  A List with JVM arguments.
    */
   public List<String> getJvmArguments()
   {
      return options.getJvmArguments();
   }
   
   /**
    * Get P2J client's bootstrap config.
    * 
    * @return  A List with bootstrap config settings.
    */
   public List<String> getBootstrapConfig()
   {
      return options.getBootstrapConfig();
   }

   /**
    * Get spawner command.
    * 
    * @return  The spawner command;
    */
   public String getSpawner()
   {
      return options.getSpawner();
   }

   /**
    * Get the Java process main class.
    * 
    * @return  The main class.
    */
   public Class<?> getMainClass()
   {
      return options.getMainClass();
   }
   
   /**
    * Get user id by default.
    * Should be override by subclasses to provide other implementation. 
    * 
    * @return  P2J subject id.
    */
   public String getUser()
   {
      return SecurityManager.getInstance().getUserId();
   }

   /**
    * Get output file name. This is used only by processes.
    * Should be override by subclasses to provide other values.
    * 
    * @return  The name of the output file used tom redirect STDOUT
    */
   public String getOutputToFile()
   {
      return null;
   }
}