/*
** Module   : ProcessClientSpawner.java
** Abstract : define the process-based client spawner
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20140206 First version.
** 002 MAG 20140224 Use ProcessBuilderParameters for processes.
** 003 MAG 20140310 On Windows OS we must use also a password.
** 004 OM  20150108 Passed null map as additional environment to ProcessClientSpawner.spawn().
** 005 GES 20150505 Made the inner class private.
** 006 ECF 20150828 Enable Java 8.
** 007 GES 20160301 Added some javadoc.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.Base64;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.client.chui.driver.web.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Class used to spawn a new P2J client which will authenticate as a P2J process.
 * <p>
 * This class is tricky to understand because the main part of it runs on the server while the
 * <code>TemporaryClientTask</code> inner class is serialized and sent down to execute on the
 * client side.  That client side code then calls methods which are implemented as a remote
 * object with the {@link Spawner} interface, which the server portions of this class implement. 
 * <p>
 * The spawner process will connect to the server using a secure socket and temporary credentials
 * to establish a secure session.  If that works, it will lookup the credentials to use for
 * spawning the client based on a unique UUID for this spawning session.  Using those credentials
 * it will start the client and notify the server about success or failure.
 * <p>
 * This notification must not be treated as the client process authenticated successfully; it
 * needs to be interpreted as a signal that the temporary connection has finished successfully.
 */
public class ProcessClientSpawner
extends ClientSpawner
{
   /** The P2J process ID being spawned by this instance.*/
   private String process;

   /**
    * Create a new spawner for P2J processes.
    * 
    * @param    process
    *           The P2J process ID.
    */
   public ProcessClientSpawner(String process)
   {
      super(new ProcessBuilderParameters(new ProcessBuilderOptions(process)));
      
      this.process = process;
   }

   /**
    * Launch the specified batch process.
    * <p>
    * If the P2J process is linked with an appserver, it will delegate the launching to
    * {@link AppServerLauncher#launch}.  Else, it will use the {@link ProcessClientSpawner 
    * spawner} infrastructure to start a P2J client.
    * 
    * @param    batchProcess
    *           The name of the batch process to be launched.
    */
   public static void launch(String batchProcess)
   {
      SecurityManager sm = SecurityManager.getInstance();
      
      String appserver = sm.getAppserverForProcess(batchProcess);
      if (appserver != null)
      {
         // appserver's are special, so treat them now
         AppServerLauncher.launch(appserver);
      }
      else
      {
         // normal process, launch it
         ProcessClientSpawner spawner = new ProcessClientSpawner(batchProcess);
         
         // get the user name from the directory
         spawner.spawn(sm.getSystemUser(batchProcess), sm.getSystemPassword(batchProcess), null);
      }
   }
   
   /**
    * Get a map containing the P2J process secure credentials: the serialized process keystore and
    * trust store and other details, like target aliases in the keystore and truststore. 
    */
   @Override
   public Object getServerData()
   {
      // return a map with the process defaults
      Map<String, Object> defaults = new HashMap<>();
      
      ProcessBuilderOptions opts = (ProcessBuilderOptions) cb.config.options;
      
      if (opts.getProcessAlias() != null)
         defaults.put("security:keystore:processalias", opts.getProcessAlias());
      if (opts.getTruststoreAlias() != null)
         defaults.put("security:truststore:alias", opts.getTruststoreAlias());
      if (opts.getKeyentryPassword() != null)
         defaults.put("access:password:keystore", opts.getKeystorePassword());
      if (opts.getKeyentryPassword() != null)
         defaults.put("access:password:keyentry", opts.getKeyentryPassword());
      if (opts.getTruststorePassword() != null)
         defaults.put("access:password:truststore", opts.getTruststorePassword());
      if (opts.isValidateCertificate())
         defaults.put("security:certificate:validate", opts.isValidateCertificate());
      if (opts.getProcessKeystore() != null)
         defaults.put("security:keystore:bytes", opts.getProcessKeystore());
      if (opts.getTruststore() != null)
         defaults.put("security:truststore:bytes", opts.getTruststore());
      
      return defaults;
   }

   /**
    * Get a {@link ProcessClientBuilder} used to spawn P2J clients for the target {@link #process}.
    * 
    * @param    cfg
    *           The P2J web client configuration.
    *           
    * @return   See above.
    */
   @Override
   protected ClientBuilder getBuilder(ClientBuilderParameters cfg)
   {
      return new ProcessClientBuilder(cfg);
   }

   /**
    * Get a {@link TemporaryClient} worker which will do the work after authenticating on the P2J
    * server using the temporary credentials.
    * 
    * @return   See above.
    */
   @Override
   public TemporaryClient getTemporaryClient()
   {
      return new TemporaryClientTask();
   }
   
   /**
    * On P2J client side, it will do custom work after authenticating using the temporary 
    * credentials.
    * <p>
    * For process clients case, it will retrieve from the P2J server the secure credentials
    * and save them in the {@link BootstrapConfig configuration} used to start the P2J process.
    */
   private static class TemporaryClientTask
   implements TemporaryClient
   {
      /**
       * On P2J client side, it will do custom work after authenticating using the temporary 
       * credentials.
       * <p>
       * For web chui clients case, it will create a {@link ChuiWebDriver} and start the web 
       * embedded server which will serve this client's requests.
       * <p>
       * Any custom configuration already existing in the {@code config} parameter will not
       * be overwritten.
       * 
       * @param    spawner
       *           The spawner remote access interface. 
       * @param    uuid
       *           Remote client identifier uuid.
       * @param    config
       *           The configuration to use for setup of the client and for the server connection.
       *
       * @return   Always {@code null}.
       * 
       * @throws   RuntimeException
       *           If there are problems accessing the secure credentials.
       */
      @Override
      public ScreenDriver doWork(Spawner spawner, String uuid, BootstrapConfig config)
      {
         // get the process key store and trust stores from the server
         @SuppressWarnings("unchecked")
         Map<String, Object> defaults = (Map<String, Object>) spawner.exportData(uuid);

         String  alias       = (String)  defaults.get("security:keystore:processalias");
         String  tsAlias     = (String)  defaults.get("security:truststore:alias");
         String  ksPass      = (String)  defaults.get("access:password:keystore");
         String  kePass      = (String)  defaults.get("access:password:keyentry");
         String  tsPass      = (String)  defaults.get("access:password:truststore");
         Boolean validate    = (Boolean) defaults.get("security:certificate:validate");
         byte[]  keystore    = (byte[])  defaults.get("security:keystore:bytes");
         byte[]  truststore  = (byte[])  defaults.get("security:truststore:bytes");

         try
         {
            // save them at the bootstrap config. any client-side config has priority
            if (alias != null &&
                config.getConfigItem("security", "keystore", "processalias") == null)
            {
               config.setConfigItem("security", "keystore", "processalias", alias);
            }
            
            if (tsAlias != null &&
                config.getConfigItem("security", "truststore", "alias") == null)
            {
               config.setConfigItem("security", "truststore", "alias", tsAlias);
            }
            
            if (ksPass != null && 
                config.getConfigItem("access", "password", "keystore") == null)
            {
               config.setConfigItem("access", "password", "keystore", ksPass);
            }
            
            if (kePass != null &&
                config.getConfigItem("access", "password", "keyentry") == null)
            {
               config.setConfigItem("access", "password", "keyentry", kePass);
            }
            
            if (tsPass != null &&
               config.getConfigItem("access", "password", "truststore") == null)
            {
               config.setConfigItem("access", "password", "truststore", tsPass);
            }
            
            if (validate != null &&
                config.getConfigItem("security", "certificate", "validate") == null)
            {
               config.setConfigItem("security", "certificate", "validate", validate.toString());
            }
            
            if (keystore != null &&
                config.getConfigItem("security", "keystore", "filename") == null)
            {
               config.setConfigItem("security", "keystore", "bytes",
                                    Base64.byteArrayToBase64(keystore));
            }

            if (truststore != null &&
               config.getConfigItem("security", "truststore", "filename") == null)
            {
               config.setConfigItem("security", "truststore", "bytes",
                                    Base64.byteArrayToBase64(truststore));
            }
            
            // invalidate the security transport...
            config.setConfigItem("security", "transport", "refresh", "true");

            spawner.clientIsReady(uuid, null);
         }
         catch (ConfigurationException e)
         {
            throw new RuntimeException(e);
         }
         
         return null;
      }
   }
}