/*
** Module   : WebClientBuilder.java
** Abstract : build and start a process for the web client
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version, extracted from ClientBuilder.
** 002 MAG 20140711 Enable/Disable clipboard.
** 003 MAG 20140722 Implement remote launcher.
** 004 GES 20150312 Added support for the GUI web client.
** 005 GES 20150710 Modified/added configuration processing for GUI, clipboard flag is ChUI-only.
** 006 CA  20160205 Changed to use ServerKeyStore.getStore().
** 007 GES 20160203 Added trusted mode support.
**     SBI 20160321 Added maxBinaryMessage, maxTextMessage, maxIdleTime websocket's parameters
**                  and set the web client embedded mode "client:web:embedded".
** 008 CA  20160428 Added command-line configuration options for web-clients.
** 009 SBI 20160427 Added graphicsCached parameter to turn off the graphics cache for testing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.*;

// we explicitly import this class to avoid other references to the UI classes 
import com.goldencode.p2j.ui.client.OutputManager;

/**
 * This class is used to create and spawn a web client process using a {@link ProcessBuilder}
 * instance.  A special small tools are designed for Linux / Windows OS in order to spawn the 
 * process on the OS user account.  Both tools are written in C using specific OS API and are 
 * build as executable using ant native target.  The source code and the make file for this tools 
 * are available in the /src/native process directory.  The password is provided vis stdin 
 * redirection on both platforms.  The command line parameters for the spawned process are 
 * constructed from default parameters stored in directory and some specific parameters stored 
 * inside a {@link WebClientBuilderParameters} structure.
 */
public class WebClientBuilder
extends ClientBuilder
{   
   /** Server key store */
   private ServerKeyStore serverKeyStore;
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The specific parameters for this client.
    */
   public WebClientBuilder(ClientBuilderParameters config)
   {
      super(config);
      
      // load server key store
      serverKeyStore = ServerKeyStore.getStore();
   }
 
   /**
    * Get server key store
    * 
    * @return  The loaded server key store.
    */
   public ServerKeyStore getServerKeyStore()
   {
      return (serverKeyStore != null && serverKeyStore.isValid()) ? serverKeyStore : null;
   }

   /**
    * Add terminal options.
    */
   protected void addClientOptions()
   {
      WebClientBuilderParameters webConfig = (WebClientBuilderParameters) config;
      Map<String, String> options = webConfig.getClientOptions();
      
      boolean isGui = webConfig.isGui();

      // add to command
      if (isGui)
      {
         // GUI driver options
         command.add("client:driver:type=" + OutputManager.GUI_DRIVER_WEB);
         command.add("client:gui:desktopHeight=" + options.get("desktopHeight"));
         command.add("client:gui:desktopWidth=" + options.get("desktopWidth"));
         command.add("client:gui:taskbar=" + options.get("taskbar"));
         command.add("client:gui:graphicsCached=" + options.get("graphicsCached"));
      }
      else
      {
         // ChUI driver options
         command.add("client:driver:type=" + OutputManager.CHUI_DRIVER_WEB);
         command.add("client:chui:rows=" + options.get("rows"));
         command.add("client:chui:columns=" + options.get("columns"));
         command.add("client:chui:background=" + options.get("background"));
         command.add("client:chui:foreground=" + options.get("foreground"));
         command.add("client:chui:selection=" + options.get("selection"));
         command.add("client:chui:fontname=" + options.get("fontname"));
         command.add("client:chui:fontsize=" + options.get("fontsize"));
         command.add("client:web:clipboard=" + options.get("clipboard"));
      }
      // Common Websocket parameters:
      command.add("client:web:socketTimeout=" + options.get("webSocketTimeout"));
      command.add("client:web:watchdogTimeout=" + options.get("watchdogTimeout"));
      
      command.add("client:web:maxBinaryMessage=" + options.get("maxBinaryMessage"));
      command.add("client:web:maxTextMessage="   + options.get("maxTextMessage"));
      command.add("client:web:maxIdleTime="      + options.get("maxIdleTime"));
      
      // Sets the web client embedded mode
      command.add("client:web:embedded=" + options.get("embedded"));
      
      if (options.get("port") != null)
      {
         command.add("client:web:port=" + options.get("port"));
      }
      
      if (options.get("host") != null)
      {
         // do not send this parameter on remote spawn, let the remote web server select the IP
         if (!remote)
         {
            command.add("client:web:host=" + options.get("host"));            
         }
      }
      
      // add referrer address
      String referrer = webConfig.getReferrer();
      
      if (remote && referrer != null)
      {
         referrer = referrer.replace("localhost", BrokerCore.PARAM_HOST);
      }
      
      command.add("web:referrer:url=" + referrer);
      
      // add any override configuration options
      String[] cfgOptions = webConfig.getConfigOptions();
      if (cfgOptions != null)
      {
         command.addAll(Arrays.asList(cfgOptions));
      }
   }

   /**
    * Get the spawn arguments.  Web chui clients call the {@code spawn} tool using the
    * {@code spawn 0 <user> <workdir> <command> [args]} syntax.
    * <p>
    * As these arguments are already built into the {@link ClientBuilder#command} list, it just
    * returns the already build list with a "1" prefix. 
    * 
    * @return   See above.
    */
   @Override
   protected List<String> getSpawnArguments()
   {
      WebClientBuilderParameters webConfig = (WebClientBuilderParameters) config;
      
      List<String> cmd = new ArrayList<>();
      
      if (webConfig.isTrusted())
      {
         Map<String, String> options = webConfig.getClientOptions();
         
         String port  = options.get("securePort");
         String alias = options.get("serverAlias");
         
         // mode 0 (passwordless/trusted, uses the NativeSecureConnection)
         initNativeSecureConnection(cmd, port, alias);
      }
      else
      {
         // add mode 1 (userid/password authentication at OS level)
         cmd.add("1");
      }
      
      // the entire command
      cmd.addAll(command);
      
      return cmd;
   }
}