/*
** Module   : WebHandler.java
** Abstract : starts a web client and redirects to the web client's embedded server URI
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20131113 Created initial version
** 002 MAG 20131129 Add authentication page. Handle GET and POST requests.
** 003 MAG 20140206 Send URL as referrer address via configuration file.
** 004 CA  20140206 Changes related to refactoring the spawner infrastructure (allow P2J process
**                  clients to be started, beside web chui clients).
** 005 MAG 20140303 Add authorization token.
** 006 MAG 20140723 Send authorization token as a query parameter instead a cookie.
** 007 OM  20150108 Passed null map as additional environment to WebClientSpawner.spawn().
** 008 GES 20150127 Cleanup code formatting.
** 009 GES 20150311 Modified this to support both chui and gui web clients.
** 010 SBI 20160205 Changed the redirect HTTP response to the "text/html" response containing
**                  the redirect path in its body and changed the error page response to
**                  the error message to the ajax client.
**     GES 20160121 Added static implementation of WebClientLauncher interface for remote web
**                  client spawning. Reworked spawning implementation for embedded mode.
**                  Eliminated use of PlatformHelper, since that was testing the server OS and
**                  had no relation to the platform on which spawning was occurring.
** 011 CA  20160428 Added command-line configuration options for web-clients.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.http.*;
import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.handler.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;

/**                                                              
 * Starts a web client and redirects the requestor to the embedded server URI.
 * <p>
 * On GET method a web page containing an authentication form is returned to caller.
 * The user should enter the user name and the password used to authenticate on
 * the target client platform (Linux / Windows) and press the "Login" button to submit the
 * form (which will submit/POST the form).
 * <p>
 * On POST method using the user submitted credentials a new web client process is spawned
 * with the user's operating-system login environment. On web client URI notification the user's
 * browser is redirected to this URI. If no response from the spawned client occurs within a
 * specified amount of time, the authentication page is reloaded with an error message.
 */
public class WebHandler 
extends AbstractHandler 
{      
   /** Error message prefix. */
   public static final String ERROR_PREFIX = "ERROR: ";

   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(WebHandler.class.getName());
   
   /** The root portion of the ChUI client request name (part of the URL). */
   private static final String TARGET_CHUI_ROOT = "/chui";
   
   /** The root portion of the GUI client request name (part of the URL). */
   private static final String TARGET_GUI_ROOT = "/gui";
   
   /** Error message to display, if any. */
   private String errorMessage = "";
   
   /** Flag that denotes client type (<code>true</code> for GUI, <code>false</code> for ChUI). */
   private boolean isGui = false;
   
   /**
    * Spawn a new web client process, returning the redirect URL if successful.
    * <p>
    * The following modes are possible:
    * <p>
    * <pre>
    *                                                   Default
    *                           Override    Trusted     Account
    *   User       Password    Configured  Configured  Configured                        Result
    * ---------  ------------  ----------  ----------  ----------  --------------------------------------------------
    * non-null   non-null      no          n/a         n/a         Spawn using the given credentials.
    * non-null   non-null      yes         n/a         no          Error (override mode requires a default account).
    * non-null   non-null      yes         n/a         yes         Spawn using the default account and the given pw.
    * non-null   null          n/a         no          n/a         Error (pw can't be null unless trusted mode is allowed).
    * non-null   null          no          yes         n/a         Spawn using the given user.
    * non-null   null          yes         yes         no          Error (override mode requires a default account).
    * non-null   null          yes         yes         yes         Spawn using the default account.
    * null       non-null      no          n/a         no          Error (null user requires a default account).
    * null       non-null      yes         n/a         no          Error (override mode requires a default account).
    * null       null          no          n/a         no          Error (null user requires a default account).
    * null       null          yes         n/a         no          Error (override mode requires a default account).
    * null       null          yes         no          yes         Error (pw can't be null unless trusted mode is allowed).
    * null       null          yes         yes         yes         Spawn using the default account.
    * </pre>
    * <p>
    * Trusted mode is attempted when the password is passed as <code>null</code>, but it will
    * only work if the mode is also enabled in the server's configuration.
    * <p>
    * In all cases above, the security context must have the sufficient rights to spawn a web
    * client for the given user name.  Even if all other values are configured properly, this
    * will still cause an error to occur.
    *  
    * @param    user
    *           The OS userid to use for the new process. If <code>null</code>, the default
    *           account will be used (if such an account has been configured).  If the
    *           default account is not configured, then a <code>null</code> will cause an
    *           error.  If override mode is configured and a default account is configured,
    *           then the default account will be used regardless of the value passed.  
    * @param    pw
    *           The OS password to authenticate the user.  If <code>null</code>, then the
    *           spawn will be attempted in trusted mode.  Trusted mode must be allowed for the
    *           current security context otherwise this will result in a failure.
    * @param    gui
    *           This specifies the UI mode (<code>true</code> for GUI, <code>false</code> for
    *           ChUI).
    * @param    options
    *           List of caller-specified options to be passed on the client command line or
    *           <code>null</code> if no values are specified. The options will be checked to
    *           ensure that they are allowed by the security context.  If not allowed, this
    *           will cause an error.
    * @param    referrer
    *           The referrer string to provide the web client. If the client gets disconnected
    *           it will be redirected back to this URL. 
    *           
    * @return   The redirect URL for the new web client on success.  If a failure occurred, the
    *           returned string will be prefixed by {@link #ERROR_PREFIX} and followed by
    *           descriptive error text.  
    */
   public static String spawn(String   user,
                              String   pw,
                              boolean  gui,
                              String[] options,
                              String   referrer)
   {
      Directory dir = DirectoryManager.getInstance();
      
      boolean override = Utils.getDirectoryNodeBoolean(null, "webClient/override", false, false);
      String  defuser  = Utils.getDirectoryNodeString(null,
                                                      "webClient/defaultAccount",
                                                      null,
                                                      false);
      
      // the override flag allows the server to force the OS user to a specific value
      if (override)
      {
         if (defuser == null)
         {
            String msg = "Override mode is active, missing required defaultAccount.";
            
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.logp(Level.SEVERE, "WebHandler.spawn", "", LogHelper.generate(msg));
            }
            
            return ERROR_PREFIX + msg;
         }
         
         if (LOG.isLoggable(Level.INFO))
         {
            LOG.logp(Level.INFO, 
                     "WebHandler.spawn", 
                     "",
                     LogHelper.generate("Overriding user account %s with %s.", user, defuser));
         }
         
         user = defuser;
      }
      
      // honor default account if user was not specified
      if (user == null)
      {
         if (defuser == null)
         {
            String msg = "User name cannot be null without configuring defaultAccount.";
            
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.logp(Level.SEVERE, "WebHandler.spawn", "", LogHelper.generate(msg));
            }
            
            return ERROR_PREFIX + msg;
         }
         
         if (LOG.isLoggable(Level.INFO))
         {
            LOG.logp(Level.INFO, 
                     "WebHandler.spawn", 
                     "",
                     LogHelper.generate("Using defaultAccount %s.", defuser));
         }
         
         user = defuser;
      }
      
      boolean trusted = (pw == null);                            
      
      // no password specified, we need to check for trusted mode access 
      if (trusted)
      {
         SecurityManager sm = SecurityManager.getInstance();
         
         TrustedSpawnerResource tsta =
            (TrustedSpawnerResource) sm.getPluginInstance("trustedspawner");
         
         if (tsta == null)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               String spec1 = "Trusted mode spawn of user '%s' cannot occur because the " +
                              "TrustedSpawnerResource is NOT enabled.";
               String msg = String.format(spec1, user);
                  
               LOG.logp(Level.SEVERE, "WebHandler.spawn", "", LogHelper.generate(msg));
               
               return ERROR_PREFIX + msg;
            }
         }
         else
         {
            // check that this account is allowed to launch a web client for this user
            if (!tsta.isAllowed(user))
            {
               String msg = String.format("User %s cannot be spawned by this account.", user);
               
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.logp(Level.SEVERE, "WebHandler.spawn", "", LogHelper.generate(msg));
               }
               
               return ERROR_PREFIX + msg;
            }
         }
      }
      
      if (LOG.isLoggable(Level.INFO))
      {
         LOG.logp(Level.INFO, 
                  "WebHandler.spawn", 
                  "",
                  LogHelper.generate("Attempting spawn for user %s.", user));
      }
      
      SpawnerResult result = spawnWorker(user, pw, gui, options, trusted, referrer);
      
      String retval = null;
      
      if (result.uri != null)
      {
         retval = result.uri;
         
         if (LOG.isLoggable(Level.INFO))
         {
            // strip off the auth token, since that is security sensitive and should not
            // appear in any logs
            int    idx = retval.indexOf("?");
            String url = retval.substring(0, idx < 0 ? retval.length() : idx);
            
            LOG.logp(Level.INFO, 
                     "WebHandler.spawn", 
                     "",
                     LogHelper.generate("Successful spawn for user %s at %s.", user, url));
         }
      }
      else
      {
         retval = ERROR_PREFIX + errorMessage(result.rc, user);
         
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.logp(Level.WARNING, 
                     "WebHandler.spawn", 
                     "",
                     LogHelper.generate("Spawn failed for user %s (%s).", user, retval));
         }
      }
      
      return retval;
   }                                                            
                                                                
   /**
    * Handle HTTP requests.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * @param    response
    *           The object or a wrapper of the response.
    */
   public void handle(String              target,               
                      Request             base, 
                      HttpServletRequest  request,
                      HttpServletResponse response) 
   throws IOException, 
          ServletException
   {
      if (!base.isHandled())                                    
      {                                                         
         boolean chui = TARGET_CHUI_ROOT.equalsIgnoreCase(target);
         boolean gui  = TARGET_GUI_ROOT.equalsIgnoreCase(target);
         
         // only process for our two known targets 
         if (chui || gui)
         {
            this.isGui = gui;
            
            if ("GET".equalsIgnoreCase(base.getMethod()))
            {
               doGet(target, base, request, response);
            }
            else if ("POST".equalsIgnoreCase(base.getMethod()))
            {
               doPost(target, base, request, response);
            }
         }
      }
   }

   /**
    * Handle GET requests. Send a Login dialog page back to user.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * @param    response
    *           The object or a wrapper of the response.
    */
   public void doGet(String              target, 
                     Request             base, 
                     HttpServletRequest  request,
                     HttpServletResponse response) 
   throws IOException,                                          
          ServletException
   {
      // produce the page contents
      response.setContentType(MimeTypes.Type.TEXT_HTML.asString());
      response.setStatus(HttpServletResponse.SC_OK);
      
      PrintWriter writer = response.getWriter();
      
      try
      {
         StringBuilder sb = new StringBuilder("/");
         String packageName = WebHandler.class.getPackage().getName();
         sb.append(packageName.replace('.', '/')).append("/web_client.html");
         
         InputStream is = WebHandler.class.getResourceAsStream(sb.toString());
         BufferedReader reader = new BufferedReader(new InputStreamReader(is));
         
         String str = null;
                                                      
         while (null != (str = reader.readLine()))
         {
            writer.println(handleReplacements(str));
         }
         
         writer.flush();
         
         reader.close();
      }
      
      catch (IOException ioe)
      {
         LOG.logp(Level.SEVERE,
                  "WebHandler.doGet()",
                  "",
                  LogHelper.generate("IOException!"),
                  ioe);
      }
      
      finally
      {
         // mark the request as handled
         base.setHandled(true);                                 
      }
   }
   
   /**
    * Handle POST request. Spawn web client and redirect the request to the embedded server.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * @param    response
    *           The object or a wrapper of the response.
    */
   public void doPost(String              target, 
                      Request             base, 
                      HttpServletRequest  request,
                      HttpServletResponse response) 
   throws IOException, 
          ServletException
   {
      // mark as handled
      base.setHandled(true);
      
      // get referrer URL
      String referrer = base.getHeader(HttpHeader.REFERER.asString());
      
      // get form parameters;
      String user = base.getParameter("usr");
      String pw   = base.getParameter("psw");
      
      // spawn a new web client (trusted mode should not be enabled here unless the proper
      // security plugin checks are added (see spawn() above)
      SpawnerResult result = spawnWorker(user, pw, isGui, null, false, referrer);
                                  
      if (result.uri != null)
      {
         String type = isGui ? "GUI" : "ChUI";
         
         LOG.logp(Level.INFO, 
                  "WebHandler.handle", 
                  "",
                  LogHelper.generate("Redirect %s client to %s", type, result.uri));
                  
         sendRedirectPath(result.uri, base, request, response);
      }
      else
      {
         // build error message 
         errorMessage = errorMessage(result.rc, user);
         
         // return a page having containing error message
         sendError(errorMessage, base, request, response);
         
         // reset error message
         errorMessage = "";
      }
   }
   
   /**
    * Sends the redirect to the main application site that requires users to be authorized.
    * 
    * @param    remoteUri
    *           The redirect url provided with the authorization parameter.
    * @param    base
    *           The base request.
    * @param    request
    *           The http request.
    * @param    response
    *           The http response.
    */             
   private void sendRedirectPath(String              remoteUri,
                                 Request             base, 
                                 HttpServletRequest  request,
                                 HttpServletResponse response)
   {
      response.setContentType(MimeTypes.Type.TEXT_HTML.asString());
      response.setStatus(HttpServletResponse.SC_OK);
      response.setHeader(HttpHeader.CACHE_CONTROL.asString(), 
                         "no-cache, no-store, must-revalidate");
      response.setDateHeader(HttpHeader.EXPIRES.asString(), 0);
      
      try
      {
         PrintWriter writer = response.getWriter();
         writer.println(remoteUri);
         writer.flush();
      }
      catch (IOException ioe)
      {
         LOG.logp(Level.SEVERE,
                  "WebHandler.sendRedirectPath()",
                  "",
                  LogHelper.generate("IOException!"),
                  ioe);
      }
      finally
      {
         // mark the request as handled
         base.setHandled(true);
      }
   }
   
   /**
    * Sends the error message to the client.
    * 
    * @param    error
    *           The error message to the client.
    * @param    base
    *           The base request.
    * @param    request
    *           The http request.
    * @param    response
    *           The http response.
    */
   private void sendError(
                      String              error,
                      Request             base, 
                      HttpServletRequest  request,
                      HttpServletResponse response)
   {
      response.setContentType(MimeTypes.Type.TEXT_HTML.asString());
      response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
      response.setHeader(HttpHeader.CACHE_CONTROL.asString(), 
                         "no-cache, no-store, must-revalidate");
      response.setDateHeader(HttpHeader.EXPIRES.asString(), 0);
      
      try
      {
         PrintWriter writer = response.getWriter();
         writer.println(error);
         writer.flush();
      }
      catch (IOException ioe)
      {
         LOG.logp(Level.SEVERE,
                  "WebHandler.sendRedirectPath()",
                  "",
                  LogHelper.generate("IOException!"),
                  ioe);
      }
      finally
      {
         // mark the request as handled
         base.setHandled(true);
      }
   }
   
   /**
    * Process the given string and make any replacements of parameter
    * values as needed. The replacement placeholder follow the JEE 
    * expression language (EL) format for immediate evaluation. 
    * <p>
    * Example: ${expression} 
    *
    * @param    text
    *           The text line that may need modifications.
    *
    * @return   The new text if a modification occurred or the original text
    *           if no modification was needed.
    */
   protected String handleReplacements(String text)
   {
      if (text.contains("${error}"))
      {
         text = text.replace("${error}", errorMessage);
      }
      
      if (text.contains("${client_ui_type}"))
      {
         text = text.replace("${client_ui_type}", isGui ? "GUI" : "ChUI");
      }
      
      return text;
   }
   
   /**
    * Spawn a client with the given account credentials and UI mode.
    *
    * @param    user
    *           The operating system userid.
    * @param    pw
    *           The password for authenticating the operating system account.
    * @param    gui
    *           <code>true</code> for GUI mode, <code>false</code> for ChUI mode.
    * @param    options
    *           List of caller-specified options to be passed on the client command line or
    *           <code>null</code> if no values are specified.
    * @param    trusted
    *           <code>true</code> if trusted spawning mode should be used.
    * @param    referrer
    *           The referrer string to provide the web client. If the client gets disconnected
    *           it will be redirected back to this URL. 
    *
    * @return   The redirect URI will be populated if successful, otherwise the URI will be
    *           <code>null</code> and the return code will describe the error. 
    */
   private static SpawnerResult spawnWorker(String   user,
                                            String   pw,
                                            boolean  gui,
                                            String[] options,
                                            boolean  trusted,
                                            String   referrer)
   {
      if (!verifySpawnOptions(user, options))
      {
         SpawnerResult sr = new SpawnerResult();
         sr.rc = -100;
         
         return sr;
      }
      
      WebClientSpawner spawner = new WebClientSpawner(gui, trusted, referrer, options);
      
      spawner.spawn(user, pw, null);
      
      SpawnerResult result = new SpawnerResult();
      
      result.uri = spawner.getRemoteUri();
      
      if (result.uri == null)
      {
         result.rc = spawner.getExitCode();
      }
      
      return result;
   }
   
   /**
    * Verify the spawn options passed by the remote side.  If access is not authorized to them,
    * then do not spawn the client.
    * 
    * @param    user
    *           The operating system userid.
    * @param    options
    *           List of caller-specified options to be passed on the client command line or
    *           <code>null</code> if no values are specified.
    *           
    * @return   <code>true</code> if all the options are authorized for the spawning account. 
    */
   private static boolean verifySpawnOptions(String user, String[] options)
   {
      if (options == null)
      {
         return true;
      }
      
      String msg;
      
      SecurityManager sm = SecurityManager.getInstance();

      RemoteLaunchOptionResource rlor =
         (RemoteLaunchOptionResource) sm.getPluginInstance("remotelaunchoption");
      if (rlor == null)
      {
         msg = "Attempting to use configuration options %s for user %s while " +
               "the RemoteLaunchOptionResource is NOT enabled.";
         msg = LogHelper.generate(msg, Arrays.asList(options).toString(), user);

         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.logp(Level.SEVERE, "WebHandler.verifySpawnOptions", "", msg);
         }
         
         return false;
      }

      msg = "Attempting to use configuration option %s for user %s, which is not authorized";
      
      for (int i = 0; i < options.length; i++)
      {
         String option = options[i];
         
         // use only the key, otherwise use the full option, if no '=' char
         String key = option;

         int idx = option.indexOf('=');
         if (idx > 0)
         {
            key = option.substring(0,  idx);
         }
         
         // verify this key
         if (!rlor.isAllowed(key))
         {
            msg = LogHelper.generate(msg, key, user);
            
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.logp(Level.SEVERE, "WebHandler.verifySpawnOptions", "", msg);
            }
            
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Build a text string containing a message error based on the exit code.
    * 
    * @param    rc
    *           Exit code.
    * @param    user
    *           Account user name.
    *           
    * @return   The matching error message.
    */
   private static String errorMessage(int rc, String user)
   {
      String err = "Cannot start the embedded server.";
      
      switch (rc)
      {
         // UNIX or Linux client platforms
         case 1:
            err = String.format("Invalid user name '%s'.", user);
            break;
         case 2:
            err = "Password does not match.";
            break;
         case 3:
            err = "User root not allowed.";
            break;
         case 4:
            err = "PAM authentication failed.";
            break;
         case -1:
            err = "Spawner has no root access permissions.";
            break;
            
         // Windows client platforms
         case -2:
            err = "Invalid user name or password does not match.";
            break;
         case -3:
            err = "Cannot create user environment.";
            break;
         case -4:
            err = "Cannot get user profile.";
            break;
         case -5:
            err = "Cannot set working directory.";
            break;
         case -6:
            err = String.format("Cannot create process for user '%s'.", user);
            break;
         case -7:                                               
            err = "Cannot load dynamic libraries.";
            break;
         
         // misc
         case -100:
            err = "The client is not authorized to use explicit configuration options";
            break;
      }
      
      return err;
   }
   
   /**
    * Simple container for the spawn result.
    */
   private static class SpawnerResult
   {
      /** Redirect URI in case of a successful spawn. <code>null</code> on failure. */
      public String uri = null;
      
      /** Error code on failure, 0 on success. */
      public int rc = 0;
   }
}