/*
** Module   : NetRightsEditor.java
** Abstract : Rights editor for the "net" type resources.
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090707  @43093  Created initial version.
** 002 NVS 20090713  @43165  Fully functional version.
** 003 NVS 20090723  @43364  Added createDefaultRights() method implementation.
** 004 NVS 20090803  @43491  Refactored the code and relocated panel creation  
**                           methods using border and group layouts.
** 005 NVS 20090930  @44099  Extended the edit() method to take the resource 
**                           instance name associated with this instance of 
**                           rights.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.admin.client.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/**
 * Edits the "net" rights for the admin client.
 * 
 * @author NVS
 */
public class NetRightsEditor
implements RightsEditor
{
   /** instance of AdminClient */
   private AdminClient client = null;
   
   /** editor panel */
   private JPanel edit = null;
   
   /** Read permission check box */
   private JCheckBox pRead = null;
   
   /** Write permission check box */
   private JCheckBox pWrite = null;
   
   /** Execute permission check box */
   private JCheckBox pExecute = null;
   
   /** Denied access check box */
   private JCheckBox pDenied = null;
   
   /** Save button */
   private JButton bsave = null;
   
   /** Cancel button */
   private JButton bcanc = null;
   
   /** dialog */
   private JDialog dialog = null;
   
   /** save or cancel flag */
   private boolean saving = false;
   
   /**
    * Gives chance to editors to initialize their data structures.
    * <p>
    * This is a one time call for this kind of resource for the life time 
    * of the admin session. 
    *
    * @param ac
    *        instance of AdminClient the editor is called from
    *
    * @param description
    *        rights structure description
    */
   public void initialize(AdminClient ac, Description[] description)
   {
      // save the useful reference
      client = ac;
      
      // create the edit panel
      Component[][] comps = new Component[5][2];
      
      String[] bitNames = description[0].getBitNames();
      
      // 1st row - Read permission
      pRead = new JCheckBox("R");
      pRead.setToolTipText("allows seeing a remote method");
      comps[0][0] = new JLabel(bitNames[0]);
      comps[0][1] = pRead;
      
      // 2nd row - Write permission
      pWrite = new JCheckBox("W");
      pWrite.setToolTipText("allows exporting a method for remote calls");
      comps[1][0] = new JLabel(bitNames[1]);
      comps[1][1] = pWrite;
      
      // 3rd row - Execute permission
      pExecute = new JCheckBox("X");
      pExecute.setToolTipText("allows calling a remote method");
      comps[2][0] = new JLabel(bitNames[2]);
      comps[2][1] = pExecute;
      
      // 4th row - Denied access
      pDenied = new JCheckBox("N");
      pDenied.setToolTipText("denies any and all accesses");
      comps[3][0] = new JLabel(bitNames[3]);
      comps[3][1] = pDenied;
      
      // last row - buttons
      bsave = new JButton("Save");
      bcanc = new JButton("Cancel");
      comps[4][0] = bcanc;
      comps[4][1] = bsave;
      
      // layout
      edit = AdminUtils.groupComponents(comps, "net rights", false);
      
      // event handlers:
      
      // denied access check box
      pDenied.addItemListener(new ItemListener()
         {
            public void itemStateChanged(ItemEvent evt) 
            {
               handleDenied();
            }
         });
      
      // Cancel button
      bcanc.addActionListener(new ActionListener() 
         {
            public void actionPerformed(ActionEvent evt) 
            {
               handleCancel();
            }
         });
      
      // Save button
      bsave.addActionListener(new ActionListener() 
         {
            public void actionPerformed(ActionEvent evt) 
            {
               handleSave();
            }
         });
   }

   /**
    * Creates an instance of the <code>Rights</code>, which does not take
    * any input and is considered a default starting point in editing.
    *
    * @return an instance of <code>Rights</code>
    */
   public Rights createDefaultRights()
   {
      return new NetRights();
   }
   
   /**
    * Edits the given instance of the <code>Rights</code>. 
    * The implementation should prepare its modal dialog using the initial state
    * from the passed instance of the <code>Rights</code>, show the dialog, 
    * provide event processing until the Save or Cancel action is applied, then
    * hide the dialog and return the appropriate result.
    *
    * @param name
    *        name of the resource instance being edited
    *
    * @param exact
    *        nature of the name
    
    * @param rights
    *        instance of <code>Rights</code> to be edited
    *
    * @return the resulting instance of <code>Rights</code> if the user saved
    *         the results, otherwise <code>null</code>
    */
   public Rights edit(String name, boolean exact, Rights rights)
   {
      // prefill the edit panel with the current rights
      BitField perms = ((NetRights)rights).getPermissions();
      pRead.setSelected(perms.get(NetRights.BIT_READ_ACCESS));
      pWrite.setSelected(perms.get(NetRights.BIT_WRITE_ACCESS));
      pExecute.setSelected(perms.get(NetRights.BIT_EXECUTE_ACCESS));
      pDenied.setSelected(perms.get(NetRights.BIT_DENIED_ACCESS));
      if (perms.get(NetRights.BIT_DENIED_ACCESS))
      {
         pRead.setEnabled(false);
         pWrite.setEnabled(false);
         pExecute.setEnabled(false);
      }
      
      // display the dialog
      dialog = AdminUtils.modalDialog(client, "Net Rights", edit);
      dialog.setVisible(true);
      
      // process the results
      dialog = null;
      
      if (!saving)
      {
         return null;
      }
      
      perms.set(NetRights.BIT_READ_ACCESS, pRead.isSelected());
      perms.set(NetRights.BIT_WRITE_ACCESS, pWrite.isSelected());
      perms.set(NetRights.BIT_EXECUTE_ACCESS, pExecute.isSelected());
      perms.set(NetRights.BIT_DENIED_ACCESS, pDenied.isSelected());
      if (pDenied.isSelected())
      {
         perms.clear(NetRights.BIT_READ_ACCESS);
         perms.clear(NetRights.BIT_WRITE_ACCESS);
         perms.clear(NetRights.BIT_EXECUTE_ACCESS);
      }
      
      NetRights nr = new NetRights(perms);
      return nr;
   }
   
   /**
    * Handle Denied check box selection.
    */
   private void handleDenied()
   {
      boolean denied = pDenied.isSelected();
      pRead.setEnabled(!denied);
      pWrite.setEnabled(!denied);
      pExecute.setEnabled(!denied);
   }
   
   /**
    * Handle Save button press.
    */
   private void handleSave()
   {
      saving = true;      
      dialog.setVisible(false);
   }
   
   /**
    * Handle Cancel button press.
    */
   private void handleCancel()
   {
      saving = false;
      dialog.setVisible(false);
   }
}   