/*
** Module   : SessionManagerFactory.java
** Abstract : Factory to create singleton SessionManager instance
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20071112 ADD  @35869  Created initial version. Factory to create
**                               singleton SessionManager instance.
** 002 ECF 20071121 CHG  @35938  Fixed initialization sequence. On subsequent
**                               calls to createXxxxxNode(), the session
**                               manager's initialize() method was not being
**                               called.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

import com.goldencode.p2j.cfg.*;

/**
 * Factory for creating a singleton instance of the network {@link
 * SessionManager} for this process.  Either a {@link
 * #createLeafNode(BootstrapConfig) leaf} or a {@link
 * #createRouterNode(BootstrapConfig) router} session manager can be created.
 * <p>
 * This class may not be constructed.  All functionality is provided by static
 * methods.
 */
public final class SessionManagerFactory
{
   /**
    * Private constructor to prevent this class from being instantiated.
    */
   private SessionManagerFactory()
   {
   }
   
   /**
    * Create the singleton instance of the <code>SessionManager</code> used
    * for a leaf network node, or simply return this instance if it already
    * exists.
    * <p>
    * The preferred way to retrieve the existing singleton session manager,
    * once it has been created, is with {@link SessionManager#get()}.
    * 
    * @param   config
    *          Bootstrap configuration for this network node.
    *          
    * @return  Leaf node session manager.
    * 
    * @throws  ConfigurationException
    *          if there is any configuration error during the session
    *          manager's initialization.
    */
   public static SessionManager createLeafNode(BootstrapConfig config)
   throws ConfigurationException
   {
      SessionManager sessMgr;
      synchronized (SessionManager.class)
      {
         sessMgr = SessionManager.get();
         if (sessMgr != null)
         {
            assert (sessMgr.isLeaf());
         }
         else
         {
            sessMgr = new LeafSessionManager();
         }
      }
      
      sessMgr.initialize(config);
      
      return sessMgr;
   }
   
   /**
    * Create the singleton instance of the <code>SessionManager</code> used
    * for a router network node, or simply return this instance if it already
    * exists.
    * <p>
    * The preferred way to retrieve the existing singleton session manager,
    * once it has been created, is with {@link SessionManager#get()}.
    * 
    * @param   config
    *          Bootstrap configuration for this network node.
    *          
    * @return  Router node session manager.
    * 
    * @throws  ConfigurationException
    *          if there is any configuration error during the session
    *          manager's initialization.
    */
   public static SessionManager createRouterNode(BootstrapConfig config)
   throws ConfigurationException
   {
      SessionManager sessMgr;
      synchronized (SessionManager.class)
      {
         sessMgr = SessionManager.get();
         if (sessMgr != null)
         {
            assert (sessMgr.isRouter());
         }
         else
         {
            sessMgr = new RouterSessionManager();
         }
      }
      
      sessMgr.initialize(config);
      
      return sessMgr;
   }
}