/*
** Module   : MatchCategory.java
** Abstract : counts or accumulates matches to a specific condition in some list of files
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20050225   @20123 Created initial version supporting maintaining a map of files and 
**                           for each file a set of matches.  Counters are kept for the total 
**                           number of matches, the number of files with matches and all of the 
**                           backing data is made accessible.
** 002 GES 20050318   @20440 Added support to save off the condition during match processing and
**                           to provide an iterator to these strings for future reporting use.
** 003 ECF 20111206          Added ID for database support.
** 004 GES 20130131          Cut down to work as little more than a container.Renamed the class
**                           to be a more accurate description.
** 005 GES 20160311          Added support level.
** 006 GES 20170307          Added getUrl()/setUrl().
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.pattern;

import java.util.*;

/**
 * Counts the number of matches of a specific named condition by file. 
 */
public class MatchCategory
{
   /** Match category ID. */
   private int cid;
   
   /** The verbose descriptive name of this category, used in reports. */
   private String descr = null;
   
   /** Counter that records the total number of matches across all files. */
   private int aggregate = 0;

   /** Counter that records the total number files with matches. */
   private int matchFiles = 0;

   /** Support level bitmask. */
   private int supportLvl = ReportWorker.LVL_UNKNOWN;
   
   /** The list of file IDs for each of which an associated set of matches is stored. */
   private Map<Integer, Integer> fids = null;
   
   /** URL for the detailed report for this category. */
   private String url = null;
   
   /**
    * Constructs an instance of a cross-file category with a specific
    * descriptive name.
    *
    * @param    cid
    *           MatchCategory ID.
    * @param    descr
    *           Verbose description of this category.
    * @param    aggregate
    *           The total number of matches to this category.
    * @param    matchFiles
    *           The total number of files with at least 1 match.
    * @param    fids
    *           The map of all file IDs and the number of matches for that file.
    * @param    supportLvl
    *           The support level bitmask.
    */
   public MatchCategory(int                   cid,
                        String                descr,
                        int                   aggregate,
                        int                   matchFiles,
                        Map<Integer, Integer> fids,
                        int                   supportLvl)
   {
      this.cid        = cid;
      this.descr      = descr;
      this.aggregate  = aggregate;
      this.matchFiles = matchFiles;
      this.fids       = fids;
      this.supportLvl = supportLvl;
   }
   
   /**
    * Accesses the database ID for this category
    * 
    * @return   The database ID for this category.
    */
   public int getID()
   {
      return cid;
   }
   
   /**
    * Accesses the verbose description of this category.
    *
    * @return   The verbose description of this category.
    */
   public String getDescr()
   {
      return descr;
   }
   
   /**
    * Accesses the total number of matches across all files.
    *
    * @return   The total number of matches across all files.
    */
   public int getAggregateCounter()
   {
      return aggregate;
   }
   
   /**
    * Accesses the number of files which have matches.  Some files in
    * the list may have no matches.
    *
    * @return   The number of files in the list which have matches.
    */
   public int getMatchFiles()
   {
      return matchFiles;
   }
   
   /**
    * Accesses the support level bitmask for this category.
    *
    * @return   The support level bitmask.
    */
   public int getSupportLvl()
   {
      return supportLvl;
   }
   
   /**
    * Obtain the URL for the detailed report for this category.
    *
    * @return   The URL or <code>null</code> if it was never explicitly assigned.
    */
   public String getUrl()
   {
      return url;
   }
   
   /**
    * Set the URL for the detailed report for this category.
    *
    * @param    url
    *           The new value.
    */
   public void setUrl(String url)
   {
      this.url = url;
   }
   
   /**
    * Accesses the iterator to the list of file-specific match sets.  This
    * allows direct access to the <code>Set</code> objects that each contain
    * all matches for each given file. 
    *
    * @return   The iterator to the list of match sets.
    */
   public Iterator<Integer> getFileIterator()
   {
      return fids.keySet().iterator();
   }
   
   /**
    * Returns the number of matches for a given file.
    *
    * @return   The number of matches for a given file or 0 if the file is
    *           not in the file list.
    */
   public int getCounterFor(int fid)
   {
      Integer matchCount = fids.get(fid);
      
      // safety first
      if (matchCount == null)
      {
         return 0;
      }
      
      return matchCount;
   }
}