/*
** Module   : ReportDriver.java
** Abstract : drives the pattern engine based on a list of report definitions
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20070514   @33498 Created initial version that provides
**                           support for command line processing based
**                           on a list of report definitions.
** 002 GES 20070626   @34286 Added support for naming the master report
**                           index and source file index.
** 003 GES 20090424   @41934 Import change.
** 004 GES 20090515   @42212 Another import change.
** 005 GES 20130203          Rewrite to process all reports in a single pattern
**                           engine run. This cuts the total time of a reporting
**                           run down by 80%! As part of this, the input format
**                           (for defining the reports) has been streamlined and
**                           the reports are parsed into a list of report
**                           definition objects.
** 006 ECF 20150715          Replace StringBuffer with StringBuilder.
** 007 GES 20160311          Added supportLvlExpr parsing and assignment into the report def.
** 008 CA  20170118          Added -T option, which is like -S but with a trailing of explicit 
**                           schema AST file names.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.pattern;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

import com.goldencode.util.*;
import com.goldencode.p2j.util.*;

/**
 * Special-purpose driver for the pattern engine to process a list of report
 * definitions against a given set of ASTs.  This provides a command line
 * interface, please see {@link #main} for the syntax.
 * <p>
 * The report definitions are stored in an XML file with the following
 * structure:
 * <p>
 * <ul>
 *    <li>list (required root element, contains a list of all reports to
 *        create)
 *    <ul> 
 *       <li>idxname (optional attribute, contains the base file name of the
 *           master report index to be generated, defaults to "index" which
 *           would result in the master report filename being "index.html",
 *           when overridden this value will also be prepended to the
 *           source file cross-reference default filename 
 *           "source_file_index.html")
 *       <li>report (element, 1 or more, represents a single report)
 *       <ul>
 *          <li>name (required attribute, contains the rule-set file name
 *              that creates the associated report)
 *          <li>descr (optional attribute, contains the description of the
 *              report)
 *          <li>variable (element, 0 or more, represents a user-replaceable
 *              input for the report being run)
 *          <ul>
 *             <li>name (required attribute, the name of the variable being
 *                 set)
 *             <li>value (required attribute, the value to assign to the
 *                 variable)
 *          </ul>
 *       </ul>
 *    </ul>
 * </ul>
 * <p>
 * Report definition files have a <code>.rpt</code> filename suffix and must
 * reside in the pattern path.
 */
public class ReportDriver
implements DebugLevels
{
   /** Report definition file suffix. */
   public static final String RPT_SUFFIX = ".rpt";
   
   /** list element text. */
   public static final String ELEM_LIST = "list";
   
   /** report element text. */
   public static final String ELEM_RPT = "report";
   
   /** database attribute text. */
   public static final String ATTR_DB = "database";
   
   /** Index report name. */
   public static final String ATTR_IDXNAME = "idxname";
   
   /** Index report title. */
   public static final String ATTR_IDXTITLE = "idxtitle";
   
   /** title attribute text. */
   public static final String ATTR_TITLE = "title";
   
   /** prefix attribute text. */
   public static final String ATTR_PREFIX = "prefix";
   
   /** category attribute text. */
   public static final String ATTR_CATEGORY = "categories";
   
   /** condition attribute text. */
   public static final String ATTR_COND = "condition";
   
   /** sort attribute text. */
   public static final String ATTR_SORT = "sort";
   
   /** mutex attribute text. */
   public static final String ATTR_MUTEX = "mutex";
   
   /** dumpType attribute text. */
   public static final String ATTR_DUMPTYPE = "dumpType";
   
   /** dumpExpr attribute text. */
   public static final String ATTR_DUMPEXPR = "dumpExpr";
   
   /** dumpLevel attribute text. */
   public static final String ATTR_DUMPLVL = "dumpLevel";
   
   /** multiplexExpr attribute text. */
   public static final String ATTR_MULTIPLEX = "multiplexExpr";
   
   /** supportLvlExpr attribute text. */
   public static final String ATTR_SUPPORT = "supportLvlExpr";
   
   /** Database mode flag. */
   private static boolean database = false;
   
   /** Master report index name. */
   private static String idxname = "index.html";
   
   /** Master report title. */
   private static String idxtitle = null;
   
   /** Source file index name. */
   private static String xrefname = "source_file_index.html";
   
   /** Reporting rule-set. */
   private static String ruleset = "reports/consolidated_reports";
   
   /**
    * Read the given report definition and store the list of reports for
    * subsequent processing.
    *
    * @param    rptdef
    *           File name of the XML file storing the list of report
    *           definitions.
    *
    * @return   The list of report definitions.
    */
   private static ArrayList<ReportDefinition> loadReports(String rptdef)
   {
      ArrayList<ReportDefinition> reports = new ArrayList<ReportDefinition>();
      
      try
      {
         File file = ConfigLoader.findConfigurationFile(rptdef, RPT_SUFFIX);
         
         // is this a valid file?
         if (!file.exists() || !file.isFile())
            return reports;
         
         StringBuilder sb = new StringBuilder();
         
         Document dom  = XmlHelper.parse(file);
         Element  root = dom.getDocumentElement();
         
         // read the database mode 
         String db = root.getAttribute(ATTR_DB);
         
         if (db != null && db.length() > 0)
         {
            database = Boolean.parseBoolean(db);
         }
         
         // read the master index report name if it exists
         String masterName = root.getAttribute(ATTR_IDXNAME);
         
         if (masterName != null && masterName.length() > 0)
         {
            sb = new StringBuilder(masterName).append(".html");
            idxname = sb.toString();
            
            sb = new StringBuilder(masterName).append("_").append(xrefname);
            xrefname = sb.toString();
         }
         
         // read the master index report title
         String masterTitle = root.getAttribute(ATTR_IDXTITLE);
         
         if (masterTitle != null && masterTitle.length() > 0)
         {
            sb = new StringBuilder("'").append(masterTitle).append("'");
            idxtitle = sb.toString();
         }
         
         // fixup the variable values where needed
         sb = new StringBuilder("'").append(idxname).append("'");
         idxname = sb.toString();
         sb = new StringBuilder("'").append(xrefname).append("'");
         xrefname = sb.toString();
         
         // get the list of reports
         NodeList rpts = root.getElementsByTagName(ELEM_RPT);
         
         int rlen = rpts.getLength();
         
         // process each report
         for (int i = 0; i < rlen; i++)
         {
            ReportDefinition rpt = new ReportDefinition();
            Element elem = (Element) rpts.item(i);
            
            // mandatory attributes
            
            rpt.id         = i;
            rpt.title      = elem.getAttribute(ATTR_TITLE);
            rpt.prefix     = elem.getAttribute(ATTR_PREFIX);
            rpt.condition  = elem.getAttribute(ATTR_COND);
            
            rpt.setCategoryList(elem.getAttribute(ATTR_CATEGORY));
            
            // optional attributes
            
            String sort = elem.getAttribute(ATTR_SORT);
            
            if (sort != null && sort.length() > 0)
            {
               rpt.setSortOrder(sort);
            }
            
            String mutex = elem.getAttribute(ATTR_MUTEX);
            
            if (mutex != null && mutex.length() > 0)
            {
               rpt.mutex = Boolean.parseBoolean(mutex);
            }
            
            String dumpType = elem.getAttribute(ATTR_DUMPTYPE);
            
            if (dumpType != null && dumpType.length() > 0)
            {
               rpt.dumpType = dumpType;
            }
            
            String dumpExpr = elem.getAttribute(ATTR_DUMPEXPR);
            
            if (dumpExpr != null && dumpExpr.length() > 0)
            {
               rpt.dumpExpr = dumpExpr;
            }
            
            String dumpLevel = elem.getAttribute(ATTR_DUMPLVL);
            
            if (dumpLevel != null && dumpLevel.length() > 0)
            {
               rpt.dumpLevel = Integer.parseInt(dumpLevel);
            }
            
            String multiplexExpr = elem.getAttribute(ATTR_MULTIPLEX);
            
            if (multiplexExpr != null && multiplexExpr.length() > 0)
            {
               rpt.multiplexExpr = multiplexExpr;
            }
            
            String supportLvlExpr = elem.getAttribute(ATTR_SUPPORT);
            
            if (supportLvlExpr != null && supportLvlExpr.length() > 0)
            {
               rpt.supportLvlExpr = supportLvlExpr;
            }
            
            reports.add(rpt);
         }
      }
      
      catch (Exception exc)
      {
         exc.printStackTrace();
      }
      
      return reports;
   }
   
   /**
    * Emit a syntax statement and optional error message to
    * <code>stderr</code>, then exit the process with the specified return
    * code. This method is called from {@link #main} when the user passes
    * command line parameters which are not viable.
    *
    * @param    msg
    *           Optional error message; may be <code>null</code>.
    * @param    rc
    *           Process return code used with <code>System.exit()</code>.
    */
   private static void syntax(String msg, int rc)
   {
      if (msg != null)
      {
         System.err.println(msg);
         System.err.println();
      }
      
      String[] staticText =
      {
         "",
         "Syntax:",
         "",
         "   java -DP2J_HOME=<home> ReportDriver [options] <rptdef> " +
         "<outdir> <filelist>",
         "   java -DP2J_HOME=<home> ReportDriver -S[options] <rptdef> " +
         "<outdir> <directory> \"<filespec>\"",
         "   java -DP2J_HOME=<home> ReportDriver -T[options] <rptdef> " +
         "<outdir> <directory> \"<filespec>\" <filelist>",
         "",
         "Where",
         "",
         "   Options",
         "",
         "      Dn = set debug level 'n' (must be a numeric digit between",
         "           " + MSG_NONE + " and " + MSG_TRACE + " inclusive)",
         "      S  = use filespecs instead of an explicit file list",
         "      T  = use filespecs and a trail of explicit files",
         "      N  = no recursion in filespec mode or trail mode",
         "",
         "   rptdef     = the report definition file to use, must be found",
         "                somewhere in the pattern path", 
         "   outdir     = the directory in which all report output will be",
         "                written", 
         "   filelist   = arbitrary list of absolute and/or relative file", 
         "                names to scan (default approach)",
         "   directory  = a relative or absolute path (only used when using",
         "                filespecs)",
         "   filespec   = file filter specification of the filenames in the",      
         "                given directory to scan, should be enclosed in",
         "                double quotes if any of the wildcard characters '*'",
         "                or '?' are used (only used when using filespecs)",
         ""
      };
      
      for (int i = 0; i < staticText.length; i++)
      {
         System.err.println(staticText[i]);
      }
      
      System.exit(rc);
   }
   
   /**
    * Run the pattern engine with the given input.
    *
    * @param    files
    *           List of ASTs to process.
    * @param    dbg
    *           Debug level.
    * @param    init
    *           Variable initializers to enable.
    * @param    rset
    *           Rule-set to execute.
    */
   private static void patternEngine(FileList[] files,
                                     int        dbg,
                                     Map        init,
                                     String     rset)
   throws Exception
   {
      PatternEngine engine = new PatternEngine();
      for (FileList fl : files)
      {
         if (fl != null)
         {
            engine.addAstSpec(fl);
         }
      }
      engine.setDebugLevel(dbg);      
      engine.setReadOnly(true);
      engine.setVariableInitializers(init);
      engine.run(rset);
   }
   
   /**
    * Provides a command line interface for generating a list of reports
    * using a list of ASTs and a report definition.
    * <p>
    * Syntax:
    * <pre>
    *   java -DP2J_HOME=&lt;home&gt; ReportDriver [options] &lt;rptdef&gt; &lt;outdir&gt; &lt;filelist&gt;
    *   java -DP2J_HOME=&lt;home&gt; ReportDriver -S[options] &lt;rptdef&gt; &lt;outdir&gt; &lt;directory&gt; &lt;filespec&gt;
    *   java -DP2J_HOME=&lt;home&gt; ReportDriver -T[options] &lt;rptdef&gt; &lt;outdir&gt; &lt;directory&gt; &lt;filespec&gt; &lt;filelist&gt;
    * </pre>
    * Where:
    * <ul>
    *   <li> Options
    *   <ul>
    *      <li> Dn = set debug level 'n' (must be a numeric digit between
    *           MSG_NONE (0) and MSG_TRACE (3) inclusive
    *      <li> S  = use filespecs instead of an explicit file list
    *      <li> T  = use filespecs with a trailing explicit file list (optional)
    *      <li> N  = no recursion in filespec mode
    *      <li> A  = archive key files (with a .pristine extenstion)
    *   </ul>
    *   <li> <code>rptdef</code> is the name of a report definition file 
    *        found via the pattern path
    *   <li> <code>outdir</code> is the directory in which output will be 
    *        written
    *   <li> explicit file list mode (default approach)
    *   <ul>
    *      <li> &lt;filelist&gt; is arbitrary list of absolute and/or
    *           relative file names to scan
    *   </ul> 
    *   <li> filespec mode
    *   <ul>
    *      <li> &lt;directory&gt; is the directory to search for files
    *      <li> &quot;filespec&quot; is the file specification that will be
    *           used to create a list of files to process, should be enclosed
    *           in double quotes if any of the wildcard characters '*' or '?'
    *           are used
    *   </ul>
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args) 
   {
      // option defaults
      boolean explicit  = true;
      boolean filespec  = false;
      boolean recursion = true;
      int     debug     = MSG_STATUS;
      int     idx       = 0;
      
      // requires (mode + directory + filespec) OR (mode and a filelist)
      if (args.length < 2)
      {
         syntax(null, -1);
      }
      
      // process options
      if (args[0].startsWith("-"))
      {
         idx++;
         String opts = args[0].toLowerCase();
         
         if (opts.indexOf('s') != -1)
         {
            explicit = false;
            filespec = true;
         }
         if (opts.indexOf('t') != -1)
         {
            explicit = true;
            filespec = true;
         }
         
         
         if (opts.indexOf('n') != -1) recursion = false;
         else recursion = true;
         
         int dbg = opts.indexOf('d');
         
         if (dbg != -1)
         {
            char level = opts.charAt(dbg + 1);
            
            if (!Character.isDigit(level))
            {
               syntax("Missing numeric debug level character following D" + 
                      " option.",
                      -2);
            }
            else
            {
               debug = Character.digit(level, 10);
               
               if (debug < MSG_NONE || debug > MSG_TRACE)
               {
                  syntax("Debug level must be between " + MSG_NONE +
                         " and " + MSG_TRACE + ".",
                         -3);
               }
            }
         }
      }
      
      // process the report definition
      if (idx >= args.length) 
      {
         syntax("Missing report definition parameter.", -4);
      }
      
      ArrayList<ReportDefinition> reports = loadReports(args[idx]);
      idx++;
      
      // process the output directory
      if (idx >= args.length) 
      {
         syntax("Missing output directory parameter.", -5);
      }
      
      StringBuilder sb = new StringBuilder("'");
      String outdir = sb.append(args[idx]).append("'").toString();
      idx++;
      
      FileList[] files = new FileList[2];
      
      // at this point idx points to the next arg after the options, this
      // must be one of our various forms of filename input

      if (filespec)
      {
         // there must be exactly 2 args left, unless explicit mode is used, too
         int remArgs = args.length - idx;
         if ((!explicit && remArgs != 2) || (explicit && remArgs < 2))
         {
            syntax("Missing/incomplete filespec parameter.", -7);
         }
         
         files[0] = new FileSpecList(new File(args[idx]), args[idx+1], recursion);
         idx = idx + 2;
      }

      if (explicit)
      {
         // there must be at least 1 arg left
         if (args.length <= idx)
         {
            if (!filespec)
            {
               // mandatory only if filespec was not included
               syntax("Missing file list parameter.", -6);
            }
         }
         else
         {
            int      len      = args.length;
            String[] filelist = new String[len - idx];
            
            for (int j = idx; j < len; j++)
            {
               filelist[j - idx] = args[j];
            }
            
            files[1] = new ExplicitFileList(filelist);
         }
      }
      
      try
      {
         HashMap vars = new HashMap();
         
         // add a definition for our output directory
         vars.put("outputDir", outdir);
         
         // add a definition for our output directory
         vars.put("databaseMode", database);
         
         // force source file cross-referencing
         vars.put("xref", "true");
         vars.put("xrefname", xrefname);
         
         // configure the master index
         vars.put("idxname", idxname);
         vars.put("idxtitle", idxtitle);
         
         // add our reports
         vars.put("reports", reports);
            
         // create a new pattern engine, configure it and run
         patternEngine(files, debug, vars, ruleset);
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }
}