/*
** Module   : AdaptiveFind.java
** Abstract : Query which adapts its retrieval type from dynamic to preselect
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20060801   @28343 Created initial version. Query which adapts its
**                           retrieval type from dynamic to preselect, and
**                           which starts its search from the current buffer
**                           record(s).
** 002 ECF 20060803   @28386 Fixed multiple flaws. Set lenientOffEnd to true
**                           in constructor and use a FindQuery as the dynamic
**                           replacement query delegate. Improved result set
**                           invalidation strategy.
** 003 ECF 20060901   @29153 Fixed NPE. Cannot safely use invalidBuffer to
**                           extract DMO proxy in createSimpleQuery().
** 004 ECF 20060925   @29961 Check for a pending error before retrieving a
**                           record. Retrieval is not attempted if an error is
**                           pending in the ErrorManager.
** 005 ECF 20061003   @30151 Call prepareFetch() instead of obsolete method
**                           checkPendingError(). Implemented releaseBuffers()
**                           to override do-nothing AbstractQuery
**                           implementation.
** 006 ECF 20061020   @30564 Simplified constructors. Removed constructors
**                           which contained client-side where expression,
**                           inverse DMO, or query substitution parameters
**                           arguments, since none of these are supported by
**                           the adaptive find model.
** 007 ECF 20061027   @30747 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required for
**                           compile-time type safety.
** 008 ECF 20061115   @31207 Override ancestor's implementation of
**                           isResetSnapshot() method. A converted FIND must
**                           clear the placeholder snapshot when a record is
**                           not found.
** 009 ECF 20070225   @32233 Added isGlobalCleanup() method. Used by
**                           PreselectQuery to determine that this object
**                           registers its results resources cleaner to the
**                           global (i.e., external proc) scope.
** 010 ECF 20070308   @32333 Fixed behavior for first(). Upon failure to find
**                           a record, if not in looping mode, this method
**                           must report a 'FIND FIRST/LAST failed...'
**                           message. Note that this covers the FIND
**                           LAST/PREVIOUS case, because AdaptiveFind inverts
**                           the sort direction for these queries and always
**                           invokes first().
** 011 ECF 20070412   @32973 Fixed createSimpleQuery(). Always use lock type
**                           of NONE for the dynamic query. This will be
**                           overridden by PreselectQuery.fetch(), if a
**                           stricter lock is required.
** 012 ECF 20070705   @34399 Fixed maybeInvalidate(). Invalidation is required
**                           if record buffer releases a record AdaptiveFind
**                           has loaded.
** 013 ECF 20071130   @36117 Removed isRefreshSnapshot() method. The
**                           RecordBuffer placeholder snapshot is no longer
**                           cleared by any query type when the query does not
**                           find a record to place in the buffer.
** 014 SVL 20080418   @38042 Added new convenience constructor.
** 015 ECF 20080813   @39552 Support invalidation of results based on DMO
**                           change events in other sessions. Currently, this
**                           support is limited to changes which affect an
**                           index on the backing table.
** 016 ECF 20080826   @39572 Fixed isGlobalValidation(). Can only return true
**                           for non-temp tables.
** 017 ECF 20081031   @40304 Modified methods due to changes in parent class.
** 018 ECF 20090603   @42582 Made maybeInvalidate() invalidation approach more
**                           conservative. We now invalidate if there is any
**                           uncommitted change to the backing table, in any
**                           session.
** 019 ECF 20090716   @43211 Implemented lazy initialization. Backing buffer
**                           cannot be initialized and flushed until the first
**                           attempt to find a record, once we are sure the
**                           buffer's scope has been opened.
** 020 CA  20090731   @43463 Removed lenientOffEnd activation in c'tor. The 
**                           query is allowed to throw QueryOffEndException's
**                           (to be consistent with FIND NEXT/PREV behavior).
** 021 CA  20090806   @43535 Fixed regression related to H020 (@43463) - a
**                           FIND NEXT/PREV with a NO-ERROR clause must not
**                           end the block (a QOEE must not be thrown, so
**                           lenientOffEnd must be true if we are in silent
**                           error mode).
** 022 ECF 20131028          Import change.
** 023 SVL 20140604          Resolved ambiguity with the newly added constructor of the
**                           superclass.
** 024 ECF 20140814          Replaced RecordBuffer.flushAll with flush in initialize method.
** 025 ECF 20150701          Changed addComponent to accommodate query component refactoring.
** 026 IAS 20160701          Fixed legacy name in the [565] error message.
** 027 ECF 20160728          Added optional, external buffers to constructors.
** 028 GES 20160919          Changes to match parent class constructor/initialization changes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of {@link AdaptiveQuery} which may begin in dynamic
 * record retrieval mode.  If so, it switches to preselect mode as soon as
 * possible.  It is intended as the backing implementation for converted FIND
 * statements which can be executed in succession without requiring a new
 * query to be executed for each record retrieved.  Candidates are
 * combinations of FIND FIRST/NEXT and FIND LAST/PREV, or instances of FIND
 * NEXT and FIND PREV which operate in loops.
 * <p>
 * This type of query by its nature operates only on a single table and
 * therefore on a single record buffer.  If the record buffer already is in
 * use, its latest record is used as a reference point for the next record to
 * be retrieved and the query is initialized in dynamic mode.  Otherwise, the
 * query is initialized in preselect mode.
 */
public class AdaptiveFind
extends AdaptiveQuery
{
   /** Buffer updated by this query */
   private final RecordBuffer buffer;
   
   /** Record most recently found by this query */
   private Persistable lastFound = null;
   
   /** Flag indicating whether this object has been initialized */
   private boolean initialized = false;
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   dmos
    *          Zero or more external buffers.
    */
   public AdaptiveFind(DataModelObject dmo, String where, String sort, DataModelObject... dmos)
   {
      this(dmo, where, sort, LockType.SHARE, dmos);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    * @param   dmos
    *          Zero or more external buffers.
    */
   public AdaptiveFind(DataModelObject dmo,
                       String where,
                       String sort,
                       LockType lockType,
                       DataModelObject... dmos)
   {
      super();
      super.initialize(dmo, where, null, sort, (DataModelObject) null, null, lockType);
      
      // if the NO-ERROR clause is specified, then a QOEE must not be thrown
      setLenientOffEnd(ErrorManager.isSilentError());
      
      // set external buffers, if any
      if (dmos != null && dmos.length > 0)
      {
         setExternalBuffers(dmos);
      }
      
      // start off in dynamic mode if the buffer already is active
      buffer = ((BufferReference) dmo).buffer();
      if (buffer.isActive() && buffer.getSnapshot() != null)
      {
         AdaptiveComponent comp = (AdaptiveComponent) components().next();
         invalidate(comp);
      }
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverseSorting
    *          Whether this query has the inverse sorting direction.
    * @param   dmos
    *          Zero or more external buffers.
    */
   public AdaptiveFind(DataModelObject dmo,
                       String where,
                       String sort,
                       boolean inverseSorting,
                       DataModelObject... dmos)
   {
      this(dmo, where, sort, LockType.SHARE, dmos);
      this.inverseSorting = inverseSorting;
   }
   
   /**
    * Retrieve the first composite row of results for the query.  The
    * underlying buffer is updated with the appropriate record.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void first(LockType lockType)
   {
      init();
      
      if (!prepareFetch())
      {
         return;
      }
      
      maybeInvalidate();
      
      try
      {
         super.first(lockType);
      }
      catch (QueryOffEndException exc)
      {
         ErrorManager.recordOrThrowError(
            new PersistenceException(
               "FIND FIRST/LAST failed for table " + buffer.getLegacyName(),
               565));
      }
      
      lastFound = buffer.getSnapshot();
   }
   
   /**
    * Retrieve the next composite row of results for the query.  The
    * underlying buffer is updated with the appropriate record.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void next(LockType lockType)
   {
      init();
      
      if (!prepareFetch())
      {
         return;
      }
      
      maybeInvalidate();
      
      super.next(lockType);
      
      lastFound = buffer.getSnapshot();
   }

   /**
    * If in silent error mode, this method will always return <code>true</code>.
    * Else, it will let the {@link PreselectQuery#isLenientOffEnd()} decide 
    * the mode.
    * 
    * @return  <code>true</code> if an exception will be thrown, else
    *          <code>false</code>.
    */
   public boolean isLenientOffEnd()
   {
      return ErrorManager.isSilentError() || super.isLenientOffEnd();
   }
   
   /**
    * Report whether this query must check for changes across all sessions and
    * possibly invalidate its current results if any are found.  If not, this
    * query will only honor local changes to invalidate results.
    * 
    * @return  <code>true</code> if query is against a permanent table;
    *          <code>false</code> if query is against a temporary table.
    */
   protected boolean isGlobalInvalidation()
   {
      return !buffer.isTemporary();
   }
   
   /**
    * Add a component to the query.  This implementation overrides the parent
    * method only to disallow multiple tables.  Because it is called from a
    * superclass constructor, instance variables must not be initialized in
    * this method.
    *
    * @param   comp
    *          Query component to be added.
    *
    * @throws  UnsupportedOperationException
    *          if invoked more than once per instance;  also:
    *          if attempting an outer join (temporary);
    *          if iteration type is not <code>NEXT</code> (temporary);
    */
   protected void addComponent(QueryComponent comp)
   {
      if (getTableCount() > 0)
      {
         throw new UnsupportedOperationException(
            "Multi-table adaptive finds are not supported");
      }
      
      super.addComponent(comp);
   }
   
   /**
    * Create a single-table, dynamic query to be used as the worker object
    * for a <code>DynamicResults</code> object when this query switches from
    * preselect to dynamic mode.  Overrides the parent's implementation to
    * create a <code>FindQuery</code> for its lenient-off-end behavior.
    *
    * @param   invalidBuffer
    *          Record buffer which caused this query to be invalidated.
    *
    * @return  A <code>FindQuery</code> based upon this query's single
    *          component.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which setup data is required.
    */
   protected DynamicQuery createSimpleQuery(RecordBuffer invalidBuffer)
   throws PersistenceException
   {
      // Always do the initial fetch with lock type of NONE.  It will be
      // overridden as necessary in PreselectQuery.fetch().
      LockType lockType = LockType.NONE;
      
      AdaptiveComponent comp = (AdaptiveComponent) components().next();
      
      return (new FindQuery(comp.getBuffer().getDMOProxy(),
                            comp.getOriginalWhere(),
                            null,
                            comp.getOriginalSort(),
                            comp.getInverse(),
                            comp.getArgs(),
                            lockType));
   }
   
   /**
    * Overrides the parent to disallow creation of a compound query, since
    * this class supports only single-table find queries.
    *
    * @param   invalidBuffer
    *          Record buffer which caused this query to be invalidated.
    *
    * @return  This method does not return normally.
    *
    * @throws  UnsupportedOperationException
    *          always.
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which setup data is required.  Not thrown by this
    *          implementation.
    */
   protected DynamicQuery createCompoundQuery(RecordBuffer invalidBuffer)
   throws PersistenceException
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Release the current record, if any, from the buffer which is used by
    * this query.
    *
    * @throws  QueryOffEndException
    *          if the buffer contained a record and it was released.
    */
   protected void releaseBuffers()
   {
      buffer.release();
   }
   
   /**
    * Indicate whether results resources should be cleaned up upon exiting
    * the current program's top-level scope or upon exiting the nearest
    * enclosing scope in which the results were created.
    *
    * @return  <code>true</code>.
    */
   protected boolean isTopLevelCleanup()
   {
      return true;
   }
   
   /**
    * Initialize this object once and only once.
    */
   private void init()
   {
      if (initialized)
      {
         return;
      }
      
      buffer.initialize();
      
      try
      {
         buffer.flush();
      }
      catch (ValidationException exc)
      {
         ErrorManager.throwError(exc);
      }
   }
   
   /**
    * Invalidate the current result set associated with this query if the
    * record buffer's contents have changed since the last time this query
    * updated the buffer.  If in preselect mode, this will switch to dynamic
    * mode.  If already in dynamic mode, a new dynamic query will be created
    * to reflect the new buffer contents.  This method will also take a
    * conservative approach and invalidate if there is any uncommitted change
    * in the backing table.
    */
   private void maybeInvalidate()
   {
      Persistable snapshot = buffer.getSnapshot();
      String entity = buffer.getEntityName();
      DirtyShareContext dirtyContext = buffer.getDirtyContext();
      if ((snapshot == null && lastFound != null)         ||
          (snapshot != null &&
           (lastFound == null ||
            !lastFound.getId().equals(snapshot.getId()))) ||
            dirtyContext.isEntityDirty(entity))
      {
         AdaptiveComponent comp = (AdaptiveComponent) components().next();
         invalidate(comp);
      }
   }
}