/*
** Module   : BufferFieldImpl.java
** Abstract : Implementation of the BufferField interface
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 ECF 20150315 Created initial version by refactoring all support specific to BUFFER-FIELD
**                  4GL compatibility from the FieldReference class. See that class for a full
**                  history. This refactoring was done to eliminate the memory footprint and
**                  overhead of the BUFFER-FIELD implementation from the many simpler use cases
**                  for the FieldReference class.
** 002 ECF 20150320 Javadoc fix.
** 003 OM  20150415 Added LITERAL-QUESTION attribute implementation.
** 004 SVL 20150503 Added bufferHandle.
** 005 ECF 20151208 Fixed getDataType, which was reporting the data type of extent fields with
**                  trailing brackets.
** 006 ECF 20160106 Fixed extent method, which would fail when the buffer was empty.
** 007 ECF 20160227 Reworked extent field implementation to support denormalized extent fields.
** 008 OM  20160613 Fixed LITERAL-QUESTION attribute implementation based on new researches.
** 009 IAS 20160714 Use legacy name instead of DMO alias.
** 010 EVL 20161003 Fix for initial() character return value.  If the backing variable is unknown
**                  the returned character value should be unknown as well.
** 011 SVL 20170126 Added caseSensitive.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.logging.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * Implementation of Progress 4GL BUFFER-FIELD methods and attributes, using a {@link
 * FieldReference} as a delegate worker.
 */
public final class BufferFieldImpl
extends HandleChain
implements BufferField,
           Deletable
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(BufferField.class.getName());
   
   /** {@link FieldReference} delegate object */
   private final FieldReference fieldRef;
   
   /** Corresponds to ADM-DATA attribute. */
   private final character admData = new character();
   
   /** Corresponds to UNIQUE-ID attribute. */
   private final integer uniqueID =
      new integer(UniqueIdGenerator.getUniqueId(IdKind.BUFFER_FIELD));
   
   /**
    * Create a new field reference given a DMO proxy and property name.  A non-indexed getter
    * method will be used to resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    */
   public BufferFieldImpl(DataModelObject dmo, String property)
   {
      this(dmo, property, false, null);
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property. An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value to be
    *          uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    *
    * @throws  IllegalArgumentException
    *          if field reference is not recognized.
    */
   private BufferFieldImpl(DataModelObject dmo,
                           String property,
                           boolean uppercase,
                           NumberType index)
   {
      this.fieldRef = new FieldReference(dmo, property, uppercase, index);
   }
   
   /**
    * Create an instance of this class which accesses a specific element of an extent field,
    * given a field reference for the same DMO property, which does not reference a specific
    * element of that field.
    * 
    * @param   fieldRef
    *          A reference to an extent field, which does not have an index into a specific
    *          element.
    * @param   index
    *          A one-based index into an extent field.
    */
   private BufferFieldImpl(FieldReference fieldRef, int index)
   {
      this.fieldRef = new FieldReference(fieldRef, index - 1);
   }
   
   /**
    * Get the <code>name</code> attribute of a BUFFER-FIELD, which is the legacy name of the
    * field referenced by this object.
    * 
    * @return  See above.
    */
   @Override
   public character name()
   {
      if (name.isUnknown())
      {
         // assign directly to the variable, rather than through the API, to avoid a 4052 error
         String property = fieldRef.getProperty();
         Object dmo = fieldRef.getDMO();
         this.name.assign(TableMapper.getLegacyFieldName((Buffer) dmo, property));
      }
      
      return super.name();
   }
   
   /**
    * Conversion of EXTENT attribute (KW_EXTENT).
    *
    * Getter of EXTENT attribute
    *
    */
   public integer extent()
   {
      Integer extent = fieldRef.getExtent();
      
      return new integer(extent == null ? 0 : extent.intValue());
   }
   
   /**
    * Conversion of INITIAL attribute (KW_INIT).
    *
    * Getter of INITIAL attribute.
    *
    * The value of the INITIAL schema field (which is always CHARACTER), formatted with the
    * buffer-field's format. If the INITIAL schema field has the Unknown value (?), the
    * value of the INITIAL attribute is the null string ("").
    *
    * @return  Current value of INITIAL attribute.
    */
   public character initial()
   {
      String property = fieldRef.getProperty();
      String init = TableMapper.getLegacyFieldInitial(getDMOProxy(), property);
      character initial = character.valueOf(init);
      Class<?> cls = fieldRef.getType();
      BaseDataType baseDataType;
      
      try
      {
         Class[] parmTypes = { character.class };
         Constructor ctor = cls.getConstructor(parmTypes);
         baseDataType = (BaseDataType) ctor.newInstance(initial);
      }
      catch (Exception exc)
      {
         throw new RuntimeException(exc);
      }
      
      return baseDataType.isUnknown()
                ? new character()
                : character.valueOf(baseDataType.toString(getFormat().getValue()));
   }
   
   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue()
   {
      Integer extent = fieldRef.getExtent();
      if (extent == null)
      {
         return character.valueOf(value().toString(getFormat().getValue()));
      }
      else
      {
         ErrorManager.recordOrShowError(26, "Array subscript 0 is out of range", false);
         
         String msg = "Unable to extract BUFFER-VALUE for field %s";
         String property = fieldRef.getProperty();
         String err = String.format(msg, property);
         ErrorManager.recordOrShowError(7366, err, false);
         
         return new character();
      }
   }
   
   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @param   index
    *          An integer expression representing a subscript, for fields that
    *          have extents.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue(int index)
   {
      return stringValue(new integer(index));
   }
   
   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @param    index
    *           An integer expression representing a subscript, for fields that
    *           have extents.
    *
    * @return   Current value of STRING-VALUE attribute.
    */
   public character stringValue(NumberType index)
   {
      return character.valueOf(value(index).toString(getFormat().getValue()));
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    *
    * @return  Current value of a buffer-field object.
    */
   public BaseDataType value(int index)
   {
      return value(index, true);
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    *
    * @return  Current value of a buffer-field object if the object is an extent field or
    *          <code>unknown</code> if it is a scalar one.
    */
   public BaseDataType value(NumberType index)
   {
      if (!validateBufferFieldIndex(index))
      {
         return fieldRef.unknownValue();
      }
      
      return value(index.intValue(), false);
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(int index, Object newValue)
   {
      changeValue(index, newValue, true);
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(NumberType index, Object newValue)
   {
      if (!this.validateBufferFieldIndex(index))
      {
         return;
      }
      
      changeValue(index.intValue(), newValue, false);
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the new value.
    *
    * @return  Current value of a buffer-field object. If the field is an extent field,
    *          <code>unknown</code> value is returned.
    */
   public BaseDataType value()
   {
      return fieldRef.getExtent() == null ? fieldRef.get() : fieldRef.unknownValue();
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    * <p>
    * Setter of current value of a buffer-field object. If you modify the BUFFER-VALUE attribute,
    * Progress sets the buffer-field object to the new value.
    * <p> 
    * Special care is taken when the field is a CHARACTER and the LITERAL-QUESTION is the default
    * value {@code false}; in this case the
    * <ul>
    *    <li>"?" is understand as the unknown literal (?) and</li>
    *    <li>trailing blanks (spaces) for strings with total length less than 9 are removed.
    *       Note that the {@code BUFFER-VALUE} specification does not mention any thing about the
    *       string length, and also upholds that enclosing quotes are removed. The current
    *       implementation reflects the 4GL's behaviour obtained by manual testing.</li>
    * </ul> 
    * If LITERAL-QUESTION is set to {@code yes} the string are not affected.
    *
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(Object newValue)
   {
      BaseDataType bdt = null;
      
      if (!(newValue instanceof BaseDataType))
      {
         // try some simple conversions
         if (newValue instanceof Integer)
         {
            bdt = new integer((Integer) newValue);
         }
         else if (newValue instanceof Long)
         {
            bdt = new integer((Long) newValue);
         }
         else if (newValue instanceof Double)
         {
            bdt = new decimal((Double) newValue);
         }
         else if (newValue instanceof Boolean)
         {
            bdt = new logical((Boolean) newValue);
         }
         else if (newValue instanceof String)
         {
            bdt = new character((String) newValue);
         }
         else
         {
            throw new ClassCastException("Can't convert " +
                                         newValue.getClass().getName() +
                                         " to BaseDataType!");
         }
      }
      else if (newValue instanceof unknown)
      {
         bdt = fieldRef.unknownValue();
      }
      else if (!fieldRef.getType().isAssignableFrom(newValue.getClass()))
      {
         try
         {
            // try to invoke a polymorphic constructor of the field's type
            Constructor<?> ctor = fieldRef.getType().getConstructor(BaseDataType.class);
            bdt = (BaseDataType) ctor.newInstance(newValue);
         }
         catch (InvocationTargetException exc)
         {
            // the constructor could not manage to convert the passed value into a meaningful
            // object of the field's type
            Throwable cause = exc.getCause();
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            
            LOG.log(Level.SEVERE, "Error performing runtime type conversion", exc);
            
            return;
         }
         catch (NoSuchMethodException  |
                InstantiationException |
                IllegalAccessException |
                IllegalArgumentException exc)
         {
            ErrorManager.recordOrThrowError(
               5729,
               "Incompatible datatypes found during runtime conversion");
            return;
         }
      }
      else
      {
         // finally... the simple case
         bdt = (BaseDataType) newValue;
         
         // special handling for character fields when LITERAL-QUESTION attribute is false
         if (bdt instanceof character              &&
             fieldRef.getType() == character.class &&
             !literalQuestion().booleanValue())
         {
            character newChar = ((character) bdt);
            if ("?".equals(newChar.toStringMessage()))
            {
               bdt = newChar.instantiateUnknown();
            }
            else if (newChar.toStringMessage().length() < 9) 
            {
               // remove trailing spaces only if total length is at most 8 characters long
               bdt = TextOps.rightTrim(newChar, " ");
            }
            // otherwise use the character as it was passed in as [bdt]
         }
      }
      
      fieldRef.set(bdt);
   }
   
   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.
    *
    * @return  current value of COLUMN-LABEL attribute.
    */
   public character columnLabel()
   {
      String property = fieldRef.getProperty();
      
      return character.valueOf(TableMapper.getLegacyFieldColumnLabel(getDMOProxy(), property));
   }
   
   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.
    *
    * @param   columnLabel
    *          New value of a attribute.
    */
   public void changeColumnLabel(character columnLabel)
   {
      setColumnLabel(columnLabel);
   }
   
   /**
    * Conversion of BUFFER-NAME attribute (KW_BUF_NAME).
    *
    * Gets the name of the buffer object to which the buffer-field object belongs.
    *
    * @return  name of the buffer object to which the buffer-field object belongs.
    */
   public character bufferName()
   {
      return character.valueOf(fieldRef.getParentBuffer().getLegacyName());
   }

   /**
    * Returns the handle of the buffer object to which the buffer-field belongs.
    *
    * @return the handle of the buffer object to which the buffer-field belongs.
    */
   public handle bufferHandle()
   {
      return new handle((Buffer) fieldRef.getParentBuffer().getDMOProxy());
   }
   
   /**
    * Conversion of LITERAL-QUESTION attribute (KW_LIT_QSTN).
    *
    * Getter of LITERAL-QUESTION attribute.
    *
    * @return  current value of LITERAL-QUESTION attribute.
    */
   public logical literalQuestion()
   {
      String property = fieldRef.getProperty();
      
      return new logical(TableMapper.getLegacyFieldLiteralQuestion(getDMOProxy(), property));
   }
   
   /**
    * Conversion of LITERAL-QUESTION attribute (KW_LIT_QSTN).
    *
    * Setter of LITERAL-QUESTION attribute.
    *
    * @param   literalQuestion
    *          New value of LITERAL-QUESTION attribute.
    */
   public void changeLiteralQuestion(logical literalQuestion)
   {
      String property = fieldRef.getProperty();
      TableMapper.setLegacyFieldLiteralQuestion(getDMOProxy(), property, literalQuestion);
   }
   
   /**
    * Conversion of VALIDATE-MESSAGE attribute (KW_VAL_MSG).
    *
    * Getter of VALIDATE-MESSAGE attribute.
    *
    * @return  current value of VALIDATE-MESSAGE attribute.
    */
   public character validateMessage()
   {
      String property = fieldRef.getProperty();
      
      return new character(TableMapper.getLegacyFieldValidateMessage(getDMOProxy(), property));
   }
   
   /**
    * Conversion of VALIDATE-MESSAGE attribute (KW_VAL_MSG).
    *
    * Setter of VALIDATE-MESSAGE attribute.
    *
    * @param   validateMessage
    *          New value of VALIDATE-MESSAGE attribute.
    */
   public void changeValidateMessage(character validateMessage)
   {
      String property = fieldRef.getProperty();
      TableMapper.setLegacyFieldValidateMessage(getDMOProxy(),
                                                property,
                                                validateMessage.getValue());
   }
   
   /**
    * Conversion of MANDATORY attribute (KW_MAND).
    *
    * Getter of KW_MAND attribute.
    *
    * @return current value of KW_MAND attribute.
    */
   public logical mandatory()
   {
      String property = fieldRef.getProperty();
      
      return new logical(TableMapper.isLegacyFieldMandatory(getDMOProxy(), property));
   }
   
   /**
    * Conversion of VALIDATE-EXPRESSION attribute (KW_VAL_EXPR).
    *
    * Getter of VALIDATE-EXPRESSION attribute.
    *
    * @return current value of VALIDATE-EXPRESSION attribute.
    */
   public character validateExpression()
   {
      String property = fieldRef.getProperty();
      
      return new character(TableMapper.getLegacyFieldValidateExpression(getDMOProxy(), property));
   }
   
   /**
    * Conversion of VALIDATE-EXPRESSION attribute (KW_VAL_EXPR).
    *
    * Setter of VALIDATE-EXPRESSION attribute.
    *
    * @param  validateExpression
    *         New value of VALIDATE-EXPRESSION attribute.
    */
   public void changeValidateExpression(character validateExpression)
   {
      String property = fieldRef.getProperty();
      TableMapper.setLegacyFieldValidateExpression(getDMOProxy(),
                                                   property,
                                                   validateExpression.getValue());
   }
   
   /**
    * Reports if this object is valid for use.
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return getDMOProxy().valid();
   }
   
   /**
    * Conversion of POSITION attribute (KW_POS).
    *
    * Getter of POSITION attribute.
    *
    * @return current value of POSITION attribute.
    */
   public integer position()
   {
      String property = fieldRef.getProperty();
      integer position = TableMapper.getLegacyFieldId(getDMOProxy(), property);
      position.increment();
      
      return position;
   }
   
   /**
    * Returns the state of the DBNAME attribute
    *
    * @return    Current value of the DBNAME attribute.
    */
   public character getDbName()
   {
      return new character(fieldRef.getParentBuffer().getDatabase().getName());
   }
   
   /**
    * Returns the state of the TABLE attribute
    *
    * @return    Current value of the TABLE attribute.
    */
   public character getTable()
   {
      return new character(fieldRef.getParentBuffer().getTable());
   }
   
   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public decimal getHeightChars()
   {
      invalidAttribute("HEIGHT-CHARS");
      
      return new decimal();
   }
   
   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    * 
    * @param   h
    *          Not used.
    */
   public void setHeightChars(NumberType h)
   {
      readOnlyError("HEIGHT");
   }
   
   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    * 
    * @param   h
    *          Not used.
    */
   public void setHeightChars(double h)
   {
      readOnlyError("HEIGHT");
   }
   
   /**
    * Get the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * For buffer-field objects, the WIDTH-CHARS attribute is the number of characters in the
    * STRING-VALUE, which the AVM calculates using the FORMAT attribute.
    *
    * @return  See above.
    */
   public decimal getWidthChars()
   {
      String strValue = stringValue().toStringMessage();
      
      return new decimal(strValue == null ? 0 : strValue.length());
   }
   
   /**
    * Set the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * The WIDTH-CHARS attribute of a buffer-field is readable but not writeable.
    *
    * @param   widthChars
    *          The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(NumberType widthChars)
   {
      readOnlyError("WIDTH");
   }
   
   /**
    * Set the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * The WIDTH-CHARS attribute of a buffer-field is readable but not writeable.
    * 
    * @param   widthChars
    *          The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(double widthChars)
   {
      readOnlyError("WIDTH");
   }
   
   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public integer getWidthPixels()
   {
      invalidAttribute("WIDTH-PIXELS");
      
      return new integer();
   }
   
   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    widthPixels
    *           Not used.
    */
   public void setWidthPixels(NumberType widthPixels)
   {
      readOnlyError("WIDTH-PIXELS");
   }
   
   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    widthPixels
    *           Not used.
    */
   public void setWidthPixels(int widthPixels)
   {
      readOnlyError("WIDTH-PIXELS");
   }
   
   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public integer getHeightPixels()
   {
      invalidAttribute("HEIGHT-PIXELS");
      
      return new integer();
   }
   
   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    heightPixels
    *           Not used.
    */
   public void setHeightPixels(NumberType heightPixels)
   {
      readOnlyError("HEIGHT-PIXELS");
   }
   
   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    heightPixels
    *           Not used.
    */
   public void setHeightPixels(int heightPixels)
   {
      readOnlyError("HEIGHT-PIXELS");
   }
   
   /**
    * Getter of FORMAT attribute.
    *
    * @return   current value of FORMAT attribute.
    */
   @Override
   public character getFormat()
   {
      String property = fieldRef.getProperty();
      return character.valueOf(TableMapper.getLegacyFieldFormat(getDMOProxy(), property));
   }
   
   /**
    * Setter of FORMAt attribute.
    *
    * @param   format
    *          The new value.
    */
   @Override
   public void setFormat(String format)
   {
      setFormat(new character(format));
   }
   
   /**
    * Setter of FORMAt attribute.
    *
    * @param   format
    *          The new value.
    */
   @Override
   public void setFormat(character format)
   {
      String property = fieldRef.getProperty();
      TableMapper.setLegacyFieldFormat(getDMOProxy(), property, format.getValue());
   }
   
   /**
    * Getter of COLUMN-LABEL attribute.
    *
    * @return   current value of COLUMN-LABEL attribute.
    */
   public character getColumnLabel()
   {
      return columnLabel();
   }
   
   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   public void setColumnLabel(character columnLabel)
   {
      setColumnLabel(columnLabel.getValue());
   }
   
   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   public void setColumnLabel(String columnLabel)
   {
      String property = fieldRef.getProperty();
      TableMapper.setLegacyFieldColumnLabel(getDMOProxy(), property, columnLabel);
   }
   
   /**
    * Returns the value of the DATA-TYPE attribute
    *
    * @return   Current value of the DATA-TYPE attribute.
    */
   public character getDataType()
   {
      Class<?> type = fieldRef.getType();
      if (type.isArray())
      {
         type = type.getComponentType();
      }
      
      String name = null;
      
      if (type == datetimetz.class)
      {
         // the only wrapper type currently whose simple Java name doesn't match 4GL name
         // TODO: this will change with 4GL OO support
         name = "datetime-tz";
      }
      else
      {
         name = type.getSimpleName();
      }
      
      return character.valueOf(name);
   }
   
   /**
    * Set the <code>DATA-TYPE</code> attribute..
    *
    * @param    dataType
    *           The DATA-TYPE to be set.
    */
   public void setDataType(String dataType)
   {
      readOnlyError("DATA-TYPE");
   }
   
   /**
    * Set the <code>DATA-TYPE</code> attribute..
    *
    * @param    dataType
    *           The DATA-TYPE to be set.
    */
   public void setDataType(character dataType)
   {
      readOnlyError("DATA-TYPE");
   }
   
   /**
    * Returns the value of the LABEL attribute
    *
    * @return    Current value of the LABEL attribute.
    */
   public character getLabel()
   {
      String property = fieldRef.getProperty();
      
      return character.valueOf(TableMapper.getLegacyFieldLabel(getDMOProxy(), property));
   }
   
   /**
    * Set the <code>LABEL</code> attribute of handle.
    *
    * @param    label
    *           The LABEL to be set.
    */
   public void setLabel(String label)
   {
      String property = fieldRef.getProperty();
      TableMapper.setLegacyFieldLabel(getDMOProxy(), property, label);
   }
   
   /**
    * Set the <code>LABEL</code> attribute of handle.
    *
    * @param    label
    *           The LABEL to be set.
    */
   public void setLabel(character label)
   {
      setLabel(label.getValue());
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueID);
   }

   /**
    * Getter of CASE-SENSITIVE attribute. Indicates whether the buffer-field is case-sensitive.
    *
    * @return current value of CASE-SENSITIVE attribute.
    */
   @Override
   public logical caseSensitive()
   {
      UnimplementedFeature.missing("CASE-SENSITIVE attribute is not implemented");
      return new logical();
   }
   
   /**
    * Delete the resource.
    *
    * BUFFER-FIELD will be deleted only when the dynamic buffer is deleted. No resource-specific
    * delete is required.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }
   
   /**
    * Retrieve the value of a specific element of an extent field, optionally validating the
    * supplied subscript.
    * 
    * @param   index
    *          One-based subscript of the extent field to be changed.
    * @param   validate
    *          <code>true</code> to validate <code>index</code>, else <code>false</code>.
    */
   private BaseDataType value(int index, boolean validate)
   {
      if (validate && !validateBufferFieldIndex(index))
      {
         return fieldRef.unknownValue();
      }
      
      if (index == 0)
      {
         return fieldRef.get();
      }
      
      return new FieldReference(fieldRef, index - 1).get();
   }
   
   /**
    * Change the value of a specific element of an extent field, optionally validating the
    * supplied subscript.
    * 
    * @param   index
    *          One-based subscript of the extent field to be changed.
    * @param   newValue
    *          New value to set into the extent field.
    * @param   validate
    *          <code>true</code> to validate <code>index</code>, else <code>false</code>.
    */
   private void changeValue(int index, Object newValue, boolean validate)
   {
      if (validate && !validateBufferFieldIndex(index))
      {
         return;
      }
      
      if (index == 0)
      {
         changeValue(newValue);
      }
      else
      {
         new BufferFieldImpl(fieldRef, index).changeValue(newValue);
      }
   }
   
   /**
    * Get the DMO proxy which references parent buffer.
    *
    * @return  DMO proxy object.
    */
   private Buffer getDMOProxy()
   {
      return (Buffer) fieldRef.getParentBuffer().getDMOProxy();
   }
   
   /**
    * Validate the index passed to a BUFFER-VALUE method call. For scalar fields, this value must
    * be zero. For extent fields, this value must be a positive number less than or equal to the
    * size of the extent field. Unknown value is invalid in either case.
    * 
    * @param   index
    *          Index value (unadjusted) provided by converted business logic.
    * 
    * @return  <code>true</code> if <code>index</code> is valid, else <code>false</code>.
    * 
    * @throws  ErrorConditionException
    *          if <code>index</code> is invalid and session is not in silent error mode.
    */
   private boolean validateBufferFieldIndex(NumberType index)
   {
      return validateBufferFieldIndex(index.isUnknown() ? -1 : index.intValue());
   }
   
   /**
    * Validate the index passed to a BUFFER-VALUE method call. For scalar fields, this value must
    * be zero. For extent fields, this value must be a positive number less than or equal to the
    * size of the extent field. Unknown value is invalid in either case.
    * 
    * @param   index
    *          Index value (unadjusted) provided by converted business logic, or <code>-1</code>
    *          if that value was unknown.
    * 
    * @return  <code>true</code> if <code>index</code> is valid, else <code>false</code>.
    * 
    * @throws  ErrorConditionException
    *          if <code>index</code> is invalid and session is not in silent error mode.
    */
   private boolean validateBufferFieldIndex(int index)
   {
      int max = 0;
      
      if (index > 0)
      {
         Integer extent = fieldRef.getExtent();
         if (extent != null)
         {
            max = extent.intValue();
         }
      }
      
      if (index < 0 || index > max || (max == 0 && index != 0))
      {
         String property = fieldRef.getProperty();
         String fieldName = TableMapper.getLegacyFieldName(getDMOProxy(), property);
         String msg = String.format("%s BUFFER-VALUE argument must be valid array index, less " +
                                    "than or equal to field extent %d",
                                    fieldName,
                                    max);
         new Throwable().printStackTrace();
         ErrorManager.recordOrThrowError(7365, msg, false);
         
         return false;
      }
      
      return true;
   }
}