/*
** Module   : CompoundComponent.java
** Abstract : A joinable, single-table query component used in a compound query.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20150801 Refactored from CompoundQuery, which previously contained this as an inner
**                  class.
** 002 ECF 20161014 Set errorIfNull to false on query if it is an outer join.
** 003 ECF 20161108 Avoided altering errorIfNull flag when it was not cleared in constructor.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

/**
 * A helper container for a query component. A query component defines the <code>Joinable</code>
 * object to which record retrieval tasks are delegated, the natural iteration type of this
 * component query, and the type of join to perform with the previous component (if any).
 */
final class CompoundComponent
{
   /** Component query */
   private final Joinable query;
   
   /** Iteration behavior (from {@link QueryConstants}) */
   private final int iteration;
   
   /** <code>true</code> for outer join;  <code>false</code> for inner */
   private boolean outer;
   
   /**
    * Constructor.
    *
    * @param   query
    *          Component query.
    * @param   iteration
    *          Iteration behavior (from {@link QueryConstants}).
    * @param   outer
    *          <code>true</code> for outer join;  <code>false</code> for
    *          inner.
    */
   CompoundComponent(Joinable query, int iteration, boolean outer)
   {
      this.query = query;
      this.iteration = iteration;
      this.outer = outer;
      
      if (outer)
      {
         query.setErrorIfNull(false);
      }
      
      for (RecordBuffer buffer : query.getRecordBuffers())
      {
         buffer.initialize();
      }
   }
   
   /**
    * Get the delegate query.
    *
    * @return  Query.
    */
   Joinable getQuery()
   {
      return query;
   }
   
   /**
    * Get the natural iteration type.
    *
    * @return  Iteration constant.
    */
   int getIteration()
   {
      return iteration;
   }
   
   /**
    * Does this query join to the previous query component using a left
    * outer join?
    *
    * @return  <code>true</code> if left outer join;  <code>false</code>
    *          if inner join.
    */
   boolean isOuter()
   {
      return outer;
   }
   
   /**
    * Disable the left outer join designation of this query component. Exposed to allow outer join
    * components to be reset to inner join components, if an inner join component is added after
    * an outer join component has been added.
    */
   void setInner()
   {
      if (!outer)
      {
         return;
      }
      
      this.outer = false;
      query.setErrorIfNull(true);
   }
}