/*
** Module   : P2JIndexComponent.java
** Abstract : Index helper class; represents a Progress or SQL index component
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20051010   @23179 Created initial version. Represents a single component of a database
**                           index.
** 002 ECF 20070328   @32646 Partial fix for text components of an index. The getNameCaseAware()
**                           method was not embedding the rtrim() function inside the upper()
**                           function.
** 003 ECF 20070328   @32647 Need a more complete solution for the getNameCaseAware() method. The
**                           previous fix does not properly wrap a component with the rtrim()
**                           function for case-sensitive text components. We need more type
**                           information for an index component in order to do that.
** 004 ECF 20070629   @35300 Minor optimization. Replaced StringBuffer with StringBuilder.
** 005 ECF 20080311   @37479 Added flag to indicate whether component is a character data type.
**                           Also optimized flags into a single bitfield.
** 006 ECF 20140318          Added copy constructor.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

/**
 * Represents a single index component; stores information about the
 * component's name, sort direction, whether it contains text data, its case
 * sensitivity, and abbreviated status. The purpose of the name attribute is
 * left intentionally vague, such that it could represent a Progress field
 * name, or an SQL database table column name, or a Java DMO property name.
 */
public final class P2JIndexComponent
{
   /** Flag indicating descending (non-default) sort direction */
   private static final int DESCENDING = 0x01;
   
   /** Flag indicating component represents character data */
   private static final int CHARACTER = 0x02;
   
   /** Flag indicating component ignores case (character data only) */
   private static final int IGNORE_CASE = 0x04;
   
   /** Flag indicating abbreviated status (Progress semantic) */
   private static final int ABBREVIATED = 0x08;
   
   /** Name of index field or column */
   private final String name;
   
   /** Property flags */
   private final int flags;
   
   /**
    * Constructor.
    *
    * @param   name
    *          Field or column name.
    * @param   descending
    *          <code>true</code> if component has non-default, descending
    *          sort direction;  <code>false</code> if ascending.
    * @param   isChar
    *          <code>true</code> if this component represents a character
    *          field.
    * @param   ignoreCase
    *          <code>true</code> if this component represents a character
    *          field AND case is ignored when sorting index, else
    *          <code>false</code>.
    * @param   abbreviated
    *          <code>true</code> if index component supports the Progress
    *          semantic of using abbreviated data for a sort or selection
    *          match. Valid only if this is the last component of the
    *          index (not enforced here).
    */
   public P2JIndexComponent(String name,
                            boolean descending,
                            boolean isChar,
                            boolean ignoreCase,
                            boolean abbreviated)
   {
      int flags = 0;
      if (descending)
      {
         flags |= DESCENDING;
      }
      if (isChar)
      {
         flags |= CHARACTER;
      }
      if (ignoreCase)
      {
         if (!isChar)
         {
            throw new IllegalArgumentException(
               "Non-character index component cannot be case-insensitive");
         }
         flags |= IGNORE_CASE;
      }
      if (abbreviated)
      {
         flags |= ABBREVIATED;
      }
      
      this.name = name;
      this.flags = flags;
   }
   
   /**
    * Copy constructor.
    * 
    * @param   component
    *          Instance to copy.
    */
   public P2JIndexComponent(P2JIndexComponent component)
   {
      this.name = component.name;
      this.flags = component.flags;
   }
   
   /**
    * Get index component's name.
    *
    * @return  Index field or column name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Indicate if index component uses descending (non-default) sort
    * direction.
    *
    * @return  <code>true</code> if sort direction is descending, else
    *          <code>false</code> if ascending.
    */
   public boolean isDescending()
   {
      return ((flags & DESCENDING) != 0);
   }
   
   /**
    * Indicate if index component represents a character field.
    *
    * @return  <code>true</code> if case is ignored;  else <code>false</code>.
    */
   public boolean isCharType()
   {
      return ((flags & CHARACTER) != 0);
   }
   
   /**
    * Indicate if index component represents a character field and ignores
    * case in data comparisons.
    *
    * @return  <code>true</code> if case is ignored;  else <code>false</code>.
    */
   public boolean isIgnoreCase()
   {
      return ((flags & IGNORE_CASE) != 0);
   }
   
   /**
    * Indicate if index component supports Progress semantic of using
    * abbreviated data for a sort or selection match.
    *
    * @return  <code>true</code> if abbreviated match is supported, else
    *          <code>false</code>.
    */
   public boolean isAbbreviated()
   {
      return ((flags & ABBREVIATED) != 0);
   }
   
   /**
    * Test for equivalence; <code>o</code> is considered equivalent if it
    * a <code>Component</code> instance with the same name, sort
    * direction, and abbreviation support as this component instance.
    *
    * @param   o
    *          An object instance, presumably a <code>Component</code>
    *          object, to test for equivalence with this object.
    *
    * @return  <code>true</code> if <code>o</code> is equivalent to this
    *          instance, else <code>false</code>.
    */
   public boolean equals(Object o)
   {
      if (!(o instanceof P2JIndexComponent))
      {
         return false;
      }
      
      P2JIndexComponent c = (P2JIndexComponent) o;
      
      return (name.equals(c.name) && flags == c.flags);
   }
   
   /**
    * Hash code implementation which is consistent with this class'
    * implementation of {@link #equals}.
    *
    * @return  This object's hash code.
    */
   public int hashCode()
   {
      int result = 17;
      result = 37 * result + name.hashCode();
      result = 37 * result + flags;
      
      return result;
   }
}