/*
** Module   : Persistence.java
** Abstract : Provides persistence services using Hibernate ORM
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 ECF 20050830   @22419 Created initial version. Provides bootstrap
**                           services to configure a Hibernate session
**                           factory per database.
** 002 ECF 20050902   @22480 Added essential query services; refactored.
**                           This includes scrollable cursor queries,
**                           individual record loads into a re-used data
**                           model object, and enforced, unique queries.
**                           Refactored all bootstrap and database
**                           initialization logic into external class
**                           (DatabaseManager).
** 003 ECF 20051104   @23254 Changes to support locking and transactions.
**                           Added methods to lock records and query lock
**                           status, and to begin, commit, and roll back
**                           transactions.
** 004 ECF 20051109   @23382 Changed use paradigm from all static methods
**                           to getting an instance by factory. Previously
**                           all methods were static and required a
**                           physical database name to be specified. Now
**                           one instance is created and cached per
**                           physical database. Client code retrieves an
**                           instance via a factory method and operates on
**                           that instance. Instances are immutable and
**                           shared across multiple contexts. Added method
**                           to retrieve the next unique primary key using
**                           a database sequence.
** 005 ECF 20051215   @23751 Added support for ambiguous check. Also made
**                           primary key (phase 1) searches read-only.
** 006 ECF 20060120   @24044 Integrated with P2J server. Replaced
**                           ThreadLocal with ContextLocal.
** 007 ECF 20060130   @24354 Added temp table support. Required a new
**                           list() method, disabling the lock manager for
**                           temp table cases, and enabling arbitrary SQL
**                           statements to be executed (for temp table
**                           creation and dropping).
** 008 ECF 20060218   @24720 Created uniqueResultViolation method. This
**                           processing was needed from multiple places,
**                           including outside this class.
** 009 ECF 20060306   @25016 Added loadReadOnly method. Supports internal
**                           queries used for persistence framework
**                           housekeeping, including resolving foreign
**                           records using legacy keys (for natural join
**                           reverse resolution).
** 010 ECF 20060329   @25255 Fixed executeSQL method. Was passing the SQL
**                           string to PreparedStatement's execute method.
** 011 ECF 20060602   @26943 Optimization. HQL queries now will always
**                           attempt to load the full record, rather than
**                           just the primary key ID. This avoids the
**                           overhead of a separate load step.
** 012 ECF 20060608   @27001 Changed list() method signature and updated
**                           implementation. Number of results can now be
**                           limited and results can be set read-only.
** 013 ECF 20060610   @27098 Consolidated lock processing. Internal lock
**                           processing now uses public lock() method.
** 014 ECF 20060621   @27505 Changed session management and transaction
**                           implementation. Sessions are now always
**                           created upon demand, and detached records
**                           from a previous session are immediately
**                           associated with the new session. The begin
**                           transaction API is more lenient, reporting if
**                           a new transaction could not be created,
**                           rather than throwing an exception.
** 015 ECF 20060627   @27590 Improved temp table support. Always flush
**                           Hibernate session before executing JDBC
**                           statements directly. Ensure save and delete
**                           operations are executed within transactions.
** 016 ECF 20060712   @28050 Added trimTextArgs. Trims trailing whitespace
**                           from any query substitution parameter which
**                           is of type com.goldencode.p2j.util.character.
** 017 ECF 20060719   @28100 Fixed memory leak. Implemented cleanup()
**                           method for Hibernate Session's ContextLocal
**                           variable to ensure Session is explicitly
**                           closed when context ends. Although the P2J
**                           server would not pin the Session's memory
**                           directly, there was still a reference within
**                           Hibernate, such that the Session data could
**                           not be garbage collected after the client
**                           context ended.
** 018 ECF 20060719   @28101 Enable result set 'paging'. The list() method
**                           now accepts a startOffset parameter.
** 019 ECF 20060720   @28115 Integrated ChangeBroker. Enables optimization
**                           of avoiding unnecessary auto-flushing of the
**                           active Hibernate session. This dramatically
**                           improves the performance of RandomAccessQuery
**                           when the session contains a large number of
**                           entities.
** 020 ECF 20060725   @28187 Consolidated context local variables within
**                           the Context inner class. Also added a
**                           mechanism to prevent sessions from being
**                           closed while external entities are dependent
**                           upon them.
** 021 ECF 20060728   @28269 Made queries cacheable; improved reliability
**                           of locking in dynamic retrieval mode.
** 022 ECF 20060828   @28930 Better error handling. In some cases, session
**                           was not being closed after catching an
**                           unrecoverable exception.
** 023 ECF 20060830   @29005 Fixed getQuery(). In case where we were not
**                           explicitly paging results, the query would
**                           return an empty result set, due to incorrect
**                           maxResults setting.
** 024 ECF 20060901   @29136 Fixed memory leak. decrementSessionsUsers()
**                           was actually incrementing the session user
**                           count, due to a typo.
** 025 ECF 20060903   @29203 Fixed Hibernate assertion failure. Ensure
**                           session is flushed before an object is
**                           evicted.
** 026 ECF 20060904   @29219 Better error reporting. scroll() reports HQL
**                           statement and context info on query error.
** 027 GES 20060905   @29230 Updated trimTextArgs() to truncate any 
**                           character vars that have embedded null bytes.
**                           This ensures that null bytes can never be
**                           passed into queries as substitution
**                           arguments.
** 028 ECF 20060911   @29460 Fixed query defect. Restore default query
**                           settings when checking out a cached query.
** 029 ECF 20060919   @29697 Made mods required by changes to ChangeBroker
**                           calls. Also added trace debug output during
**                           record eviction.
** 030 ECF 20060929   @30035 Fixed out of range query substitution date
**                           parameters. We now 'clip' those dates to the
**                           minimum and maximum date values supported by
**                           the backing database and log a debug message.
**                           Renamed trimTextArgs() method to
**                           preprocessQueryParameters() to reflect
**                           broader purpose.
** 031 ECF 20061010   @30309 Made changes to accommodate auto-commit temp
**                           tables. These were primarily side effects of
**                           changes made to TemporaryBuffer and
**                           RecordBuffer. A new variant of the static
**                           factory method to retrieve instances of this
**                           class was added. The new version of get()
**                           will create a secondary instance of this
**                           class, backed by the same database as the
**                           primary instance. This allows auto-commit and
**                           non-auto-commit temp tables to share the same
**                           physical backend, but maintain separate
**                           connections and Hibernate sessions.
** 032 ECF 20061012   @30390 Fixed nonthreadsafe access error. Flush
**                           session to ensure consistency, when a request
**                           to close the session cannot be honored
**                           immediately (because a query is pinning it
**                           open).
** 033 ECF 20061014   @30527 Integrated SessionListener architecture.
**                           Objects which register as SessionListeners
**                           are notified when a Hibernate session is
**                           about to close. Deprecated session user count
**                           mechanism which would pin open a session as
**                           long as a reference count of dependents was
**                           non-zero.
** 034 ECF 20061110   @31094 Improved logging. Errors reassociating
**                           detached records are logged in detail to
**                           prevent these errors from being masked in
**                           silent error mode.
** 035 ECF 20061113   @31140 Added new variant of load(). Needed for cases
**                           where a RecordBuffer instance is not readily
**                           available.
** 036 ECF 20061116   @31236 Minor adjustments to exception handling.
**                           Ensure session gets closed when a database
**                           error occurs.
** 037 ECF 20061120   @31325 Changed session closing behavior. Do not
**                           close temp table sessions if the temp table
**                           count exceeds zero (with no database errors).
** 038 ECF 20061212   @31667 Fixed context cleanup. Inner class Context's
**                           cleanup was not properly connected to end of
**                           context cleanup, so it was never being
**                           invoked.
** 039 ECF 20070111   @31779 Added flushBatch method. Flushes a batch of
**                           records to the database and clears the active
**                           session.
** 040 ECF 20070130   @32039 Various cleanup. Ensure session is flushed if
**                           necessary after attaching a record. Ensure
**                           session is closed for all Hibernate exception
**                           processing. Ensure session is closed if a
**                           SessionListener throws an exception during
**                           close notification. Deprecated method which
**                           creates secondary Persistence instance for a
**                           single database.
** 041 ECF 20070212   @32155 Integrated DirtyReadManager. This allows DMO
**                           validation to check uncommitted updates
**                           across client contexts. Also added the
**                           ability to temporarily disable flushing,
**                           which was necessary to correct a flaw in
**                           validation processing.
** 042 ECF 20070228   @32246 Modified HQL-based load() method. It can now
**                           handle either an ID-only projection query or
**                           a query which fetches and returns a DMO
**                           directly. In the latter case, the additional
**                           lock() and get() steps are omitted.
** 043 ECF 20070326   @32611 Added bulk delete method. Deletes all records
**                           which meet the specified criteria for a given
**                           DMO type.
** 044 ECF 20070329   @32688 Fixed attach/detach. Flush check was not
**                           correct.
** 045 ECF 20070412   @32983 Retrofitted locking to use RecordLockContext.
**                           An instance of this class is created for each
**                           context, one per database. Simplified context
**                           cleanup.
** 046 ECF 20070416   @33025 Integrated user interrupt handling.
** 047 ECF 20070508   @33458 Added queryIndexData(). Uses database
**                           metadata to inspect indices on a given table.
** 048 ECF 20070514   @33485 Fixed queryIndexData(). In some cases, root
**                           column name must be extracted from string
**                           queried from index metadata in a dialect
**                           specific way.
** 049 ECF 20070516   @33669 Fixed abend caused by an orphaned transaction
**                           at session close. When a session is closed
**                           due to an error, roll back any open
**                           transaction before closing the session. Also
**                           made error logging less noisy at this layer,
**                           since a severe error usually will be logged
**                           in error handling provided by users of the
**                           persistence framework. Details are still
**                           logged at FINE logging level.
** 050 ECF 20070519   @33682 Fixed load() method (variant which loads by
**                           primary key). Ensure session is flushed if
**                           necessary and use Session.get() instead of
**                           Session.load(), in case record was deleted.
** 051 ECF 20070602   @33926 Integrated new scope options of
**                           SessionListener. GLOBAL_SCOPE registers a
**                           session listener cleanup routine with the
**                           transaction manager global scope. NO_SCOPE
**                           does not register a cleanup method at all.
** 052 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 053 ECF 20070703   @34360 Removed IllegalStateException from flush().
**                           Now throws recoverable PersistenceException
**                           instead when no session is available.
** 054 ECF 20070706   @34401 Implemented SessionListener deregistration
**                           notification.
** 055 ECF 20070708   @34414 Optimization. Cache ChangeBroker in context
**                           local state, rather than retrieving it every
**                           time flushing is checked.
** 056 ECF 20070709   @34419 Optimization. Consolidated access to Context
**                           instance to prevent multiple ContextLocal
**                           lookups within the same methods. Changed some
**                           signatures and added/removed several private
**                           methods to accomodate this.
** 057 ECF 20070720   @34620 Modified implementation of attach(). It now
**                           associates a detached DMO with the current
**                           Hibernate session.
** 058 ECF 20070812   @34857 Changed error message on non-unique result
**                           exception. Report buffer name rather than
**                           DMO name.
** 059 ECF 20070817   @34876 Fixed all variants of load() to prevent stale
**                           data retrieval. There were cases where a DMO
**                           instance in the Hibernate session cache would
**                           become stale based on an update in another
**                           user context. We now force a refresh from the
**                           database when reading a record with a lock.
** 060 ECF 20070901   @35302 Changed executeSQL() method. Connection is
**                           now set to autocommit=true before the SQL
**                           statement is executed. Added new method
**                           executeSQLQuery() to allow execution of
**                           arbitrary SQL queries with no change to
**                           connection's autocommit setting. Replaced use
**                           of database name string with database info
**                           object. Added trace logging of execution time
**                           of queries and other database accesses.
** 061 ECF 20071002   @35331 Implemented dialect-specific callback hooks.
**                           Registered P2JDialect is given a chance to
**                           perform post-transaction processing with
**                           dialect-specific data stored in the current
**                           client context.
** 062 ECF 20071008   @35399 Added back convenience method get(String).
** 063 ECF 20071008   @35445 Fixed regression introduced by #060 (@35302).
**                           Incorrect LockManager and DirtyReadManager
**                           instances were being created for the temp
**                           table database.
** 064 ECF 20071018   @35575 Refactored to enable creation by factory.
**                           Moved static factory methods to new class
**                           PersistenceFactory. Also, lock manager and
**                           dirty read manager classes are now read from
**                           the directory instead of hard-coded.
** 065 ECF 20071221   @36545 Fixed load() method implementations. We now
**                           catch UnresolvableObjectException, which is a
**                           recoverable condition which can occur when
**                           invoking Session.refresh(). This will be
**                           thrown by Hibernate when we have a DMO in the
**                           first level cache, which has been deleted
**                           from the database.
** 066 SVL 20071231   @36626 Identity manager is initialized to support
**                           primary key generation on a per-table basis.
** 067 ECF 20080102   @36644 Removed deprecated nextPrimaryKey() method.
**                           Replacement is nextPrimaryKey(String) added
**                           in #066 (@36626).
** 068 ECF 20080108   @36733 Fixed bulk delete. Removed optimization of
**                           query flush, which was causing a
**                           StaleStateException in cases of deleted DMOs
**                           with collection associations.
** 069 SVL 20080111   @36772 getIdentityManager() function has been added.
** 070 ECF 20080129   @37004 Added debug logging. Log DMO persist event.
** 071 ECF 20080312   @37480 Integrated support for embedded H2 database.
**                           Also changed return type of list() method to
**                           be less specific.
** 072 ECF 20080403   @38136 Use entity name instead of DMO class to test
**                           if flush is required. Added load() variant
**                           which accepts entity name instead of DMO
**                           class.
** 073 SVL 20080421   @38097 Added foreignKeysEnabled property.
** 074 CA  20080430   @38160 Added a static method retrieveIndexData which
**                           contains the queryIndexData's code related to
**                           building the index list from a result set.
** 075 ECF 20080606   @38642 Removed ValidationContext references.
** 076 ECF 20080611   @38697 Major refactoring. Index metadata query
**                           methods were rationalized to be accessible in
**                           more situations. Added methods to temporarily
**                           "borrow" a connection from a current session.
**                           Removed all references to the obsolete
**                           DirtyReadManager class hierarchy.
** 077 ECF 20080624   @38918 Optimized column type metadata methods used
**                           primarily at startup. This approach avoids
**                           many round trips to the database to query
**                           column metadata, but may not work with all
**                           JDBC drivers. In case not, the original,
**                           unoptimized approach is used.
** 078 CA  20080605   @38566 Renamed method delete to deleteOrUpdate.
** 079 ECF 20080806   @39324 Added rowid support. Changed default identity
**                           manager to SequenceIdentityManager, which
**                           generates database-wide unique IDs. Primary
**                           keys are now 64-bit integers.
** 080 CA  20080815   @39462 Support API change in DBUtils. Fixed NPE in
**                           getColumnDataType.
** 081 SVL 20080901   @39652 Identity pool is initialized together with
**                           identity manager.
** 082 ECF 20080923   @39936 Memory tuning. Do not cache queries used for
**                           deletes/updates.
** 083 ECF 20080924   @39946 Added load(RecordIdentifier). Loads a single
**                           record, given its unique identifier.
** 084 ECF 20081015   @40091 Refactored query substitution parameter
**                           preprocessing. Abstracted out a static method
**                           to process a single parameter, so it is
**                           accessible to other classes in the package.
** 085 ECF 20081021   @40208 Force refresh of DMOs in active buffers when
**                           getting a new Session after a rollback. This
**                           ensures active buffers have current versions
**                           of these records.
** 086 ECF 20081024   @40224 Fixed regression caused by #085 (@40208). If
**                           a record in a buffer was deleted during a
**                           rollback, we must handle this stale state
**                           when creating the new session and refreshing
**                           records left in active buffers.
** 087 ECF 20081106   @40354 Modified directory lookups for lock manager
**                           and identity manager. Changed path to allow
**                           for other, related properties to be set in
**                           the directory.
** 088 ECF 20081118   @40549 Minor change to error message text.
** 089 ECF 20081204   @40803 Added performance warning logging. Queries
**                           which take longer than certain thresholds
**                           (configured in the directory) trigger a
**                           warning log entry.
** 090 ECF 20090114   @41162 Fixed load() (HQL variant). Was not correctly
**                           unlocking when passed a lock type of NONE.
** 091 ECF 20090128   @41239 Improved Context.getSession(). When
**                           associating the current records of active
**                           buffers with a new session after a rollback
**                           of the previous session, we only refresh
**                           those we know to be stale, rather than all
**                           of them. This reduces the number of trips
**                           to the database. The rolledBack parameter to
**                           Context.closeSession() was removed as a
**                           result of this change.
** 092 ECF 20090204   @41257 Improved getQuery() performance. We now cache
**                           Query objects more aggressively, which avoids
**                           the overhead of Hibernate's initialization of
**                           these objects when possible.
** 093 ECF 20090212   @41295 Added bulk eviction method. Exposed Hibernate
**                           session via bind() method.
** 094 ECF 20090311   @41502 Added support for limit/max value inlining in
**                           a SELECT statement. We do this only in very
**                           limited cases: currently only for simple load
**                           queries.
** 095 ECF 20090601   @42576 API change. Renamed load(RecordIdentifier) to
**                           quickLoad(RecordIdentifier). Also made method
**                           public, as it is needed from dirty package.
** 096 ECF 20090619   @42771 Added scroll() variant which accepts ScrollMode.
**                           Existing variant now calls this variant with a
**                           default ScrollMode of SCROLL_INSENSITIVE.
** 097 CA  20090625   @42945 Transaction roll back is executed at database 
**                           level, so session needs to be flushed before the
**                           actual roll back.
** 098 ECF 20090702   @43036 Fixed session closing/cleanup logic. Removed
**                           deprecated code.
** 099 ECF 20090711   @43188 Implemented implicit transaction support. To
**                           prevent Hibernate from beginning open-ended
**                           transactions for query operations, we now open an
**                           implicit transaction for each such operation.
**                           These are closed either when an explicit
**                           transaction is opened, or after an idle timeout.
** 100 ECF 20090716   @43219 Fixed regression in #099 (@43188). In the event a
**                           session was marked as broken in the implicit
**                           transaction timer thread, this state was never
**                           being reset after a new session has been opened.
** 101 ECF 20090719   @43305 Added safety code to implicit commit task. It is
**                           more resilient to unexpected errors (such as
**                           OOME), to avoid taking down the timer thread.
** 102 ECF 20090722   @43325 Disabled implicit transaction commit thread. It
**                           is the prime suspect in a "possible nonthreadsafe
**                           access to session" error. May be re-enabled at a
**                           later time.
** 103 ECF 20090723   @43339 Made preprocessQueryParameter() public. This
**                           static method is needed by the dirty subpackage.
** 104 CA  20090724   @43381 Rolled back H091 (#41239) - removed the
**                           buffer.isStale() dependency in
**                           Context.getSession().
** 105 ECF 20090723   @43418 Rewrote implicit transaction support. Removed
**                           secondary thread and push/pop calls from low
**                           level query services. These must be implemented
**                           by callers instead to enable coarser, implicit
**                           transaction granularity (i.e., fewer transactions
**                           when possible). If a session is being used only
**                           for an implicit transaction, it is made read-only
**                           to prevent unnecessary session flushing.
** 106 ECF 20090810   @43587 Fixed SessionListener notifications. Modified
**                           notification mechanism to allow more event types.
**                           Notify listeners when a transaction is about to
**                           be committed, in addition to session closing
**                           events.
** 107 ECF 20090811   @43599 Fixed regressions in #105 (@43418). Implicit
**                           transaction use count was not being properly set
**                           and used in some places. Fixed endTransaction(),
**                           which was nulling out active transaction too
**                           early.
** 108 CA  20090901   @43818 In initializeInstance, catch any StopCondition's
**                           and re-throw them (do not pack it in a generic
**                           runtime exception).
** 109 ECF 20090925   @44029 Temporarily disabled identity pool. There is a
**                           defect in the implementation of IdentityPool that
**                           can prevent primary keys from being distributed.
**                           When that is fixed, this should be re-enabled.
**                           Affected code was commented out from
**                           createIdentityManager().
** 110 CA  20091001   @44062 Modified executeSQLQuery() to leave the statement
**                           open; else, it will not be able to iterate 
**                           through the returned result set.
** 111 ECF 20091002   @44082 Re-enabled identity pool. Also made additional
**                           change to executeSQLQuery() to require the caller
**                           to push/pop the implicit transaction counter. We
**                           cannot do this in the method, as this could cause
**                           the transaction to be closed before the caller
**                           could process results.
** 112 SVL 20091019   @44145 Added support for the trigger manager.
** 113 SVL 20091126   @44442 Added quickLoad(RecordIdentifier, boolean)
**                           function.
** 114 CA  20100115   @44527 Added executeSQL and executeSQLQuery versions
**                           which accept query parameters.
** 115 CA  20101028          All metadata string query parameters and results 
**                           will be processed using dialect-specific case.
**                           Do not prepare permanent H2 database.
** 116 CA  20101202          The permanent H2 database needs to be prepared
**                           in a special way - the custom functions need to 
**                           be registered with the HQL preprocessor. 
** 117 AIL 20121206          Updated from character to TextOps.
** 118 SVL 20120331          Upgraded to Hibernate 4.
** 119 CA  20130620          Added APIs to check if a given class is one of P2J's Hibernate
**                           compatible types and to disambiguate a BDT or Java native type to a
**                           Hibernate compatible type.
** 120 OM  20130531          Added preprocessing of datetime/-tz parameters.
**                           Added small patch for binding composed-type parameters.
**                           Maybe this should be used for all parameters when switching
**                           from positional to named parameters 
**                           Do not close sql connection when 4GL error occurs in CAN-FINDs.
** 121 VMN 20130830          Added lockTable method.
** 122 OM  20130829          Enriched the parameter list of load() methods with a boolean for
**                           specifying if the record should be forced to be read from database.
**                           Added support for -rereadnolock P4GL startup parameter emulation.
** 123 ECF 20130804          Added base metadata support.
** 124 ECF 20131028          Added lock metadata support.
** 125 CA  20131013          Added no-op deleted() method, required by the changes in Finalizable
**                           interface.
** 126 OM  20131206          Added initialization call for DatabaseTriggerManager.
** 127 SVL 20140210          Set JPA-style HQL parameters instead of positional.
** 128 GES 20140206          Added isActive() to allow callers to determine if the persistence
**                           layer is or is not disabled.
** 129 ECF 20140312          Replaced Apache commons logging with J2SE standard logging.
** 130 ECF 20140318          Minor memory optimization when gathering JDBC index metadata.
** 131 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 132 ECF 20140624          Rolled back 119. See persist.type.TypeHelper.
** 133 OM  20140711          Changed initialization location for database triggers to pkgroot/dmo/.
** 134 ECF 20140930          Wrap IllegalStateException in PersistenceException in rollback, to
**                           be consistent with the declared API and to be less catastrophic for
**                           callers.
** 135 ECF 20141204          Fixes to transaction and context cleanup.
** 136 OM  20150112          Replaced calls to deprecated method. Additional code cleanup.
** 137 ECF 20150221          Raise STOP condition instead of ERROR for JDBC-level errors, to avoid
**                           getting caught in RETRY loops for unrecoverable errors.
** 138 ECF 20150326          Optionally log diagnostic information on begin transaction conflicts.
** 139 ECF 20150504          Enable explicit deregistration of session listeners.
** 140 ECF 20150811          Fixed database connection leak during context cleanup. Added some
**                           debug logging and adjusted level of existing log statements.
** 141 ECF 20150821          Minor format fix and import change.
** 142 ECF 20150906          Change code to match BufferManager API changes; override hashCode
**                           and equals implementations.
** 143 ECF 20150928          Reimplemented query cache to reduce memory use. Refactored certain
**                           methods used frequently within the persist package to reduce
**                           unnecessary context local lookups.
** 144 OM  20151028          Added comment/javadoc.
** 145 OM  20140721          Called preparePermanentDatabase() for MS SQL database server in order
**                           to prepare HQL for support UDFs from loaded assemblies.
** 146 ECF 20160305          Collect index metadata at initialization in multiple threads to
**                           reduce server initialization time.
** 147 ECF 20160429          Do not make queries against _temp database cacheable. Reimplemented
**                           executeSQL variants to work directly with JDBC, rather than using
**                           Hibernate's SQLQuery, which was too expensive. Fixed load method
**                           variants to not rely upon DMO reference counting in their refresh
**                           logic. Refined rereadnolock implementation, at least when used by
**                           RandomAccessQuery.
** 148 ECF 20160512          Fixed deregisterSessionListener.
** 149 ECF 20160610          Optimized getSingleSQLResult.
** 150 OM  20160603          Fixed rtrimming to space character only.
** 151 IAS 20160621          Added support for read-only tables
** 152 ECF 20160614          Reduce unnecessary flushing. Expanded use of read-only metadata.
**                           Further refined rereadnolock implementation.
** 153 IAS 20160701          Fixed [3166] error message.
** 154 IAS 20160714          Use legacy name instead of DMO alias.
** 155 OM  20160715          Double guarded saving of template records to database. Prevented
**                           ContextLocalCleanupException to be thrown when client disconnects.
** 156 ECF 20160820          Fixes to support better concurrency and better detect records
**                           deleted in other sessions. Includes new lock timeout feature and
**                           better integration with record lock context.
** 157 OM  20160905          Small optimizations.
** 158 ECF 20160831          Performance enhancements. Improved use of read-only table
**                           information. Reworked session flushing to allow better batching of
**                           database statements. Initialize P2OLookup instances at server
**                           startup. Logging changes.
** 159 ECF 20160929          Removed obsolete enableFlush feature. Use implicit transactions
**                           where possible to reduce flushing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.lang.InstantiationException;
import java.io.*;
import java.sql.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.Date;
import java.util.logging.*;
import com.goldencode.cache.*;
import com.goldencode.p2j.directory.*;
import org.hibernate.*;
import org.hibernate.dialect.*;
import org.hibernate.engine.spi.*;
import org.hibernate.internal.*;
import org.hibernate.jdbc.*;
import org.hibernate.proxy.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.hql.*;
import com.goldencode.p2j.persist.id.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.persist.meta.*;
import com.goldencode.p2j.persist.trigger.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

/**
 * The central point for fundamental persistence services in the P2J
 * environment.  This implementation is backed by the Hibernate ORM framework.
 * This class provides basic CRUD (create, read, update, delete) services,
 * as well as methods to begin/commit/rollback database-level transactions.
 * <p>
 * An instance of this class is retrieved from the factory method {@link
 * #get}.  One instance of this class is associated with a single, physical
 * database, per user context.
 * <p>
 * An instance of a concrete {@link com.goldencode.p2j.persist.lock.LockManager} implementation
 * is created for each each instance of this class.  The implementation to use is
 * queried from the directory service, and defaults to the {@link
 * InMemoryLockManager} if no other implementation is specified in the
 * directory.
 * <p>
 * Transactions are explicitly managed with {@link #beginTransaction}, {@link
 * #commit} and {@link #rollback}.  Only one transaction per client context
 * may be active at a time.  These are database-level transactions, which may
 * or may not coincide with application-level transactions.
 * <p>
 * Most methods which perform some sort of query will open an implicit
 * transaction before the query is executed, but only if no transaction
 * already is active.  Implicit transactions are closed in one of two ways:
 * <ul>
 *   <li>a new explicit transaction is begun (in this case the implicit
 *       transaction is first committed);
 *   <li>a timeout period elapses with no additional query activity within the
 *       implicit transaction.
 * </ul>
 * In the latter case, the transaction is closed on a different thread than
 * that on which it was opened.  This timeout mechanism is used to avoid a
 * very noticeable performance penalty which is incurred when performing many
 * consecutive queries, if a transaction is committed immediately after each
 * one.
 * <p>
 * The exception to this rule are the variants of the <code>scroll</code>
 * method.  Because of the nature of <code>ScrollableResults</code>, a commit
 * should not take place until after calling code is finished using this
 * object.  This is necessary because fetching results may require additional
 * trips to the database within the transaction.  In this case, the
 * <code>scroll</code> methods themselves do not trigger implicit
 * transactions.  Rather, it is up to calling code (see {@link
 * ScrollingResults}) to bracket its use of the result set with calls to
 * {@link #pushImplicitTransaction()} and {@link #popImplicitTransaction(Context)}.
 * <p>
 * The implicit transaction mechanism is intended only to support legacy use
 * cases.  <strong>External code which directly accesses query methods within
 * the <code>Persistence</code> class should always properly bracket its use
 * of these methods with explicit transactions</strong>.  Otherwise,
 * unpredictable behavior may result as implicit transactions are opened and
 * closed automatically.  Take special care to initialize any necessary
 * collection associations (for example, converted extent fields within DMOs
 * are implemented as lazily initialized, associated collections), before
 * ending a transaction.
 * <p>
 * Hibernate sessions are managed implicitly, transparently to application
 * code and even to other classes within the persistence framework.  No more
 * than one session per user context may be active at a time.  Sessions are
 * opened lazily as they are needed to fulfill database operations:  executing
 * a query, persisting a record, creating a temp table, etc.  They are closed
 * upon committing or rolling back a transaction, or if any exception is
 * caught during the processing of a database operation.  Upon opening a new
 * session, the current context's {@link BufferManager} instance is
 * interrogated for all active records.  These are the current records stored
 * in all open record buffers for the database matching the current
 * <code>Persistence</code> instance;  they represent detached records
 * previously associated with the last open session (if any).
 */
public class Persistence
{
   /** Well-known name of ID generator sequence */
   public static final String ID_GEN_SEQUENCE = "p2j_id_generator_sequence";
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(Persistence.class.getName());
   
   /** Default lock manager implementation */
   private static final String DEFAULT_LOCK_MANAGER =
      "com.goldencode.p2j.persist.lock.InMemoryLockManager";
   
   /** Default identity manager implementation */
   private static final String DEFAULT_IDENTITY_MANAGER =
      "com.goldencode.p2j.persist.id.SequenceIdentityManager";
   
   /** A transient mapping of JDBC connections to column data type maps */
   private static final Map<Connection, Map<ColumnKey, Integer>> columnTypes =
      new WeakHashMap<>(4);
   
   /** Flag indicating whether records are joined using foreign keys */
   private static boolean foreignKeysEnabled =
      Utils.getDirectoryNodeBoolean(null, "persistence/foreign-keys", false, false);
   
   /** <code>true</code> if the persistence layer can be used. */
   private static volatile boolean active = false;
   
   /** Context local object to manage persistence variables */
   private final ContextLocal<Context> context = new ContextLocal<Context>()
   {
      @Override
      public WeightFactor getWeight() { return WeightFactor.LEVEL_1; }
      protected Context initialValue() { return (new Context()); }
      protected void cleanup(Context value) { value.cleanup(); }
   };
   
   /** Database served by this object */
   private final Database database;
   
   /** Hibernate database dialect helper */
   private final Dialect dialect;
   
   /** Does this object represent the temporary database? */
   private final boolean temporary;
   
   /** Cached hash code */
   private final int hashCode;
   
   /** Milliseconds a scrolling query may take before being reported */
   private final long scrollThreshold;
   
   /** Milliseconds a list query may take before being reported */
   private final long listThreshold;
   
   /** Milliseconds a load query may take before being reported */
   private final long loadThreshold;
   
   /** Milliseconds a load-by-id query may take before being reported */
   private final long loadByIDThreshold;
   
   /** Pessimistic lock manager */
   private LockManager lockManager;
   
   /** Object which keeps lock metadata in sync with record lock activity */
   private LockTableUpdater lockTableUpdater = null;
   
   /** Identity manager */
   private IdentityManager identityManager;
   
   /** Sessions in use for special connection requests */
   private Map<Connection, Session> borrowedSessions = new HashMap<>();
   
   /** Mappings of table names to index lists, organized by schema */
   private static Map<String, Map<String, List<P2JIndex>>> indexesByTable = new HashMap<>();
   
   /**
    * Construct an instance for a particular, physical database, storing the
    * dialect class which Hibernate uses to create a portability layer.
    *
    * @param   database
    *          Database with which this instance is permanently associated.
    * @param   dialect
    *          Portability object used by Hibernate to issue commands to the database.
    */
   protected Persistence(Database database, Dialect dialect)
   {
      this.hashCode = super.hashCode();
      this.database = database;
      this.dialect = dialect;
      this.temporary = DatabaseManager.TEMP_TABLE_DB.equals(database);
      
      String dbName = database.getName();
      String path = "database/" + dbName + "/p2j/warn_threshold/enable";
      boolean warn = Utils.getDirectoryNodeBoolean(null, path, false, false);
      
      if (warn)
      {
         // Configure various warning thresholds to trigger reporting of slow data access.
         path = "database/" + dbName + "/p2j/warn_threshold/data_access/scroll_query";
         scrollThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         path = "database/" + dbName + "/p2j/warn_threshold/data_access/list_query";
         listThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         path = "database/" + dbName + "/p2j/warn_threshold/data_access/load_query";
         loadThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         path = "database/" + dbName + "/p2j/warn_threshold/data_access/load_by_id_query";
         loadByIDThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         
         if (LOG.isLoggable(Level.CONFIG))
         {
            LOG.config("Warn thresholds for " + database +
                       ": scroll_query=" + scrollThreshold +
                       "; list_query=" + listThreshold +
                       ": load_query=" + loadThreshold +
                       ": load_by_id_query=" + loadByIDThreshold);
         }
      }
      else
      {
         scrollThreshold = -1L;
         listThreshold = -1L;
         loadThreshold = -1L;
         loadByIDThreshold = -1L;
      }
   }
   
   /**
    * Get the instance of this class associated with the given, physical
    * database name, creating it first if necessary.  Newly created instances
    * are stored in a static cache.
    *
    * @param   name
    *          Name of physical database with which the instance is
    *          permanently associated.
    *          
    * @return  Persistence object associated with the given database.
    * 
    * @deprecated Use {@link PersistenceFactory#getInstance(String)} instead.
    */
   public static Persistence get(String name)
   {
      return PersistenceFactory.getInstance(name);
   }
   
   /**
    * Initialize various infrastructure components of the persistence framework:
    * <ul>
    *   <li>{@code DataTypeHelper}, which is used when preprocessing HQL where clauses.
    *   <li>{@code BufferManager}, which tracks open buffers and transactions.
    *   <li>{@code ChangeBroker}, which dispatches notifications about changes to DMO properties.
    *   <li>{@code DatabaseManager}, which configures the Hibernate environment for the databases
    *       to which the framework connects.
    *   <li>{@code P2OLookup}, instances of which are used for runtime, on-the-fly query and
    *       temp-table conversion.
    *   <li>{@code DatabaseTriggerManager}, which tracks and fires database triggers at the
    *       appropriate times.
    * </ul>
    *
    * @throws  IllegalStateException
    *          if invoked more than once.
    * @throws  PersistenceException
    *          if any error is encountered initializing the persistence framework.
    *
    * @see     DataTypeHelper#initialize()
    * @see     BufferManager#initialize()
    * @see     ChangeBroker#initialize()
    * @see     DatabaseManager#initialize()
    * @see     P2OLookup#initialize()
    * @see     DatabaseTriggerManager#initialize(String)
    */
   public static void initialize()
   throws PersistenceException
   {
      DataTypeHelper.initialize();
      BufferManager.initialize();
      ChangeBroker.initialize();
      DatabaseManager.initialize();
      
      // if runtime conversion services are required, initialize schema lookup resources
      if (ConversionPool.isInitialized())
      {
         try
         {
            P2OLookup.initialize();
         }
         catch (SchemaException exc)
         {
            throw new PersistenceException(exc);
         }
      }
      
      String pkgroot = Utils.getDirectoryNodeString(null, "file-system/pkgroot", ".");
      DatabaseTriggerManager.initialize(pkgroot + "/dmo/");
      
      // this is volatile and only assigned here, so it should be safe without synchronization
      active = true;
      
      if (LOG.isLoggable(Level.INFO))
      {
         LOG.info("Persistence services ready");
      }
   }
   
   /**
    * Reports if the persistence layer can be used or not.
    *
    * @return   <code>true</code> if the persistence layer is initialized and can be used.
    */
   public static boolean isActive()
   {
      return active;
   }
   
   /**
    * Perform special handling on a query substitution parameter to avoid
    * errors during query execution.  This entails:
    * <ul>
    *   <li>Trimming all trailing spaces, tabs, line feeds, and carriage
    *       returns from any parameter of type <code>character</code>.  Any
    *       parameter affected is replaced with a new <code>character</code>
    *       instance at the same index position within <code>args</code>.
    *   <li>Clipping out of bounds date values to the range supported by the
    *       backing database's implementation of a date data type.  Any
    *       parameter affected is replaced with a new <code>date</code>
    *       instance at the same index position within <code>args</code>.
    * </ul>
    * <p>
    * Note:  this method has special processing for <code>null</code> bytes
    * embedded in <code>character</code> variables.  The text of the variable
    * is truncated starting at the first embedded <code>null</code> byte.
    * This means that the resulting substitution parameters are guaranteed
    * to never include <code>null</code> bytes.
    *
    * @param   arg
    *          A single substitution parameter.
    * @param   dialect
    *          Database dialect.
    * 
    * @return  The given parameter, processed as described above.
    */
   public static Object preprocessQueryParameter(Object arg, P2JDialect dialect)
   {
      Object processed = arg;
      
      // Trim text arguments.
      if (arg instanceof character)
      {
         character c = (character) arg;
         
         if (!c.isUnknown())
         {
            // this obtains a clone of the character variable,
            // truncating it's text at any null byte that may be
            // embedded
            c = c.safeDuplicate();
            
            // this is safe because it doesn't alter the original
            // reference's state but rather replaces that reference
            processed = TextOps.rightTrim(c, " ");
         }
      }
      
      // "Clip" dates which are out of range.  This may result in
      // incorrect query behavior (hence the logged warning), but will
      // avoid catastrophic failure.
      else if (arg instanceof date)
      {
         date d = (date) arg;
         Date when = d.dateValue();
         
         Date low = dialect.getLowDate();
         if (low != null && when.before(low))
         {
            if (d instanceof datetimetz)
            {
               datetimetz dtz = (datetimetz)d;
               processed = new datetimetz(new datetime(low), dtz.getOffset());
            }
            else if (d instanceof datetime)
            {
               processed = new datetime(low);
            }
            else 
            {
               processed = new date(low);
            }
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, 
                       "Query substitution date parameter " + d.toStringMessage() + 
                       " was reset to the earliest date (" + low + 
                       ") supported by the backing database");
            }
         }
         
         Date high = dialect.getHighDate();
         if (high != null && when.after(high))
         {
            if (d instanceof datetimetz)
            {
               datetimetz dtz = (datetimetz)d;
               processed = new datetimetz(new datetime(high), dtz.getOffset());
            }
            else if (d instanceof datetime)
            {
               processed = new datetime(high);
            }
            else
            {
               processed = new date(high);
            }
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, 
                       "Query substitution date parameter " + d.toStringMessage() + 
                       " was reset to the latest date (" + high + 
                       ") supported by the backing database");
            }
         }
      }
      
      return processed;
   }
   
   /**
    * Gather a list of <code>P2JIndex</code> objects which describe the
    * indices on the given table.  This method uses the database metadata,
    * if any, exposed by the database's JDBC driver.
    *
    * @param   database
    *          Database which contains <code>table</code>.
    * @param   table
    *          Name of table to inspect.
    *
    * @return  A list of <code>P2JIndex</code> objects, one per index found,
    *          or <code>null</code> if the JDBC metadata query throws
    *          <code>SQLException</code>, indicating a database error, or
    *          that the JDBC driver does not support the requested feature.
    *          An empty list is returned if the driver supported the query,
    *          but no indices were found.
    *
    * @throws  PersistenceException
    *          if there is an error getting a Hibernate session or determining
    *          an indexed column's data type.
    * @throws  UnsupportedOperationException
    *          if the JDBC driver cannot query database metadata.
    */
   static List<P2JIndex> queryIndexData(Database database, String table)
   throws PersistenceException
   {
      String schema = DatabaseManager.getSchema(database);
      Map<String, List<P2JIndex>> map = indexesByTable.get(schema);
      if (map != null)
      {
         List<P2JIndex> list = map.get(table);
         if (list != null)
         {
            return list;
         }
      }
      
      Connection conn = null;
      P2JDialect dialect = (P2JDialect) DatabaseManager.getDialect(database);
      String _table = dialect.prepareMetadataParameter(table);
      
      try
      {
         conn = DatabaseManager.checkOutConnection(database);
         DatabaseMetaData meta = conn.getMetaData();
         
         // the parameters will be prepared using dialect specific semantics
         ResultSet rs = meta.getIndexInfo(null, null, _table, false, false);
         
         Map<String, P2JIndex> indices = retrieveIndexData(rs, database, table);
         
         return new ArrayList<>(indices.values());
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);
         
         throw new UnsupportedOperationException("Error querying index metadata for " + database,
                                                 exc);
      }
      finally
      {
         if (conn != null)
         {
            DatabaseManager.checkInConnection(conn, database);
         }
      }
   }
   
   /**
    * Query all index metadata from the given database, using a thread pool with 1 thread per
    * available CPU. Store this information for faster access later.
    * 
    * @param   database
    *          Database from which to query index metadata.
    * @param   schema
    *          Schema containing index information.
    * 
    * @throws  PersistenceException
    *          if there is an error accessing the database or querying metadata.
    */
   static void queryAllIndexData(Database database, String schema)
   throws PersistenceException
   {
      Connection conn = null;
      Map<String, List<P2JIndex>> map = indexesByTable.get(schema);
      if (map != null)
      {
         return;
      }
      
      map = new ConcurrentHashMap<>();
      
      Map<String, List<P2JIndex>> innerMap = map;
      
      try
      {
         conn = DatabaseManager.checkOutConnection(database);
         DatabaseMetaData meta = conn.getMetaData();
         
         int threads = Runtime.getRuntime().availableProcessors();
         ExecutorService exec = Executors.newFixedThreadPool(threads);
         List<Future<PersistenceException>> futures = new LinkedList<>();
         
         // TODO: make dialect-aware schema and table type arguments
         ResultSet rs = meta.getTables(null, "public", "%", new String[] { "TABLE" });
         
         while (rs.next())
         {
            String table = rs.getString("TABLE_NAME");
            
            Callable<PersistenceException> task = () ->
            {
               try
               {
                  innerMap.put(table, queryIndexData(database, table));
               }
               catch (PersistenceException exc)
               {
                  return exc;
               }
               
               return null;
            };
            
            futures.add(exec.submit(task));
         }
         
         for (Future<PersistenceException> future : futures)
         {
            try
            {
               PersistenceException exc = future.get();
               if (exc != null)
               {
                  throw exc;
               }
            }
            catch (ExecutionException | InterruptedException exc)
            {
               throw new PersistenceException(exc);
            }
         }
         
         exec.shutdown();
         
         indexesByTable.put(schema, map);
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);
         
         throw new UnsupportedOperationException("Error querying table metadata for " + database,
                                                 exc);
      }
      finally
      {
         if (conn != null)
         {
            DatabaseManager.checkInConnection(conn, database);
         }
      }
   }
   
   /**
    * Given a the result set of an index metadata query (for a certain table),
    * this method is responsible for creating a map which will hold, for each
    * index, a {@link P2JIndex} object.
    * 
    * @param   rs
    *          The {@link ResultSet} instance used to retrieve the index 
    *          metadata.
    * @param   database
    *          Database which contains <code>table</code>.
    * @param   table
    *          The table for which the indexes are retrieved.
    * 
    * @return  A map with {@link P2JIndex} instances for each existing table
    *          index.
    * 
    * @throws  PersistenceException
    *          if any database error is encountered.
    */
   static Map<String, P2JIndex> retrieveIndexData(ResultSet rs, Database database, String table)
   throws PersistenceException
   {
      P2JDialect dialect = (P2JDialect) DatabaseManager.getDialect(database);
      Map<String, P2JIndex> indices = new LinkedHashMap<>();
      
      try
      {
         P2JIndex index = null;
         
         // all results must be processed using the dialect
         while (rs.next())
         {
            if (rs.getShort("TYPE") == DatabaseMetaData.tableIndexStatistic)
            {
               continue;
            }
            
            String indexName = rs.getString("INDEX_NAME");
            indexName = dialect.processMetadataResult(indexName);
            if (indexName == null)
            {
               continue;
            }
            
            index = indices.get(indexName);
            if (index == null)
            {
               boolean nonUniq = rs.getBoolean("NON_UNIQUE");
               index = new P2JIndex(table, indexName, !nonUniq, false, false);
               indices.put(indexName, index);
            }
            
            String colName = rs.getString("COLUMN_NAME");
            colName = dialect.processMetadataResult(colName);
            
            // Determine case insensitivity in a dialect-specific way.
            boolean ignoreCase = dialect.isCaseInsensitiveColumn(colName);
            
            // Extract root column name in a dialect-specific way.
            colName = dialect.extractColumnName(colName);
            
            // Determine the data type of the indexed column.
            int dataType = getColumnDataType(database, table, colName);
            boolean isChar = (dataType == Types.VARCHAR);
            
            String sort = rs.getString("ASC_OR_DESC");
            sort = dialect.processMetadataResult(sort);
            
            boolean descending = dialect.isMetadataSortDesc(sort);
            
            index.addComponent(colName, descending, isChar, isChar && ignoreCase, false);
         }
         
         if (index != null)
         {
            index.compact();
         }
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);
         
         throw new PersistenceException(exc);
      }
      
      return indices;
   }
   
   /**
    * Detemine whether records are joined using foreign keys.
    *
    * @return  <code>true</code> if records are joined using foreign keys.
    *          <code>false</code> if records are joined using legacy key
    *          fields.
    */
   static boolean isForeignKeysEnabled()
   {
      return foreignKeysEnabled;
   }
   
   /**
    * Check if the given object is associated with the given Hibernate session. Works in the same
    * way as {@link org.hibernate.Session#contains(Object)} in Hibernate 3 (in Hibernate 4
    * {@link org.hibernate.Session#contains(Object)} returns <code>false</code> if the entity is
    * kept in the session context, but enqueued for deletion or was actually deleted from the
    * database).
    *
    * @param   object
    *          An instance of a persistent class.
    * @param   session
    *          Hibernate session presumably containing target object.
    *
    * @return  <code>true</code> if the given object is associated with the given Hibernate
    *          session.
    */
   static boolean containedInSession(Object object, Session session)
   {
      PersistenceContext persistenceContext = ((SessionImpl) session).getPersistenceContext();
      
      if (object instanceof HibernateProxy)
      {
         // do not use proxiesByKey, since not all proxies that point to this session's
         // instances are in that collection!
         LazyInitializer li = ((HibernateProxy)object).getHibernateLazyInitializer();
         if (li.isUninitialized())
         {
            // if it is an uninitialized proxy, pointing with this session, then when it is
            // accessed, the underlying instance will be "contained"
            return li.getSession() == session;
         }
         else
         {
            // if it is initialized, see if the underlying instance is contained, since we need
            // to account for the fact that it might have been evicted
            object = li.getImplementation();
         }
      }
      
      EntityEntry entry = persistenceContext.getEntry(object);
      
      return entry != null;
   }

   /**
    * Logs the FWD version returned by the getFWDVersion user-defined function (UDF/p2jpl) for
    * this database. When this database is a temporary or not primary this method is a no-op.
    * If the method fails to retrieve the version string, "undefined" version is reported in
    * the log output.
    */
   static void reportUDFVersion(Database database)
   throws PersistenceException
   {
      String udfVersion = "undefined";

      Connection conn = null;
      try
      {
         String sqlFunc = HQLPreprocessor.getRegisteredFunction(database, "getFWDVersion", null);
         if (sqlFunc != null)
         {
            conn = DatabaseManager.checkOutConnection(database);
            Statement st = conn.createStatement();

            try
            {
               ResultSet rs = st.executeQuery("SELECT " + sqlFunc + "()");
               if (rs.next())
               {
                  String val = rs.getString(1);
                  if (val != null)
                  {
                     udfVersion = val;
                  }
               }
            }
            catch(SQLException exc)
            {
               // do not propagate the exception, in case the UDF is not defined we want to
               // report "undefined" version
               LOG.log(Level.FINEST, "Error querying UDF version for " + database, exc);
            }
         }
      }
      catch(SQLException ex)
      {
         DBUtils.handleException(database, ex);
         throw new RuntimeException("Error querying UDF version for " + database, ex);
      }
      finally
      {
         if (conn != null)
         {
            DatabaseManager.checkInConnection(conn, database);
         }
      }

      LOG.info(String.format("UDF version for db %s is %s", database, udfVersion));
   }
   
   /**
    * Use database metadata to determine the data type of a database column.
    * 
    * @param   database
    *          Target database.
    * @param   table
    *          Target table.
    * @param   column
    *          Target column.
    * 
    * @return  Data type of the given column, as one of the <code>java.sql.Types</code> constants.
    * 
    * @throws  PersistenceException
    *          if there is any database error, or if the metadata query fails
    *          to produce a result for the given column.
    */
   private static int getColumnDataType(Database database, String table, String column)
   throws PersistenceException
   {
      P2JDialect dialect = (P2JDialect) DatabaseManager.getDialect(database);
      Connection conn = null;
      
      // the parameters will be prepared using dialect specific
      // semantics
      String _public = dialect.prepareMetadataParameter("public");
      String _table  = dialect.prepareMetadataParameter(table);
      String _column = dialect.prepareMetadataParameter(column);
      
      try
      {
         conn = DatabaseManager.checkOutConnection(database);
         DatabaseMetaData meta = null;
         
         if (DatabaseManager.isInitializing())
         {
            // Initialization optimization:
            // Try to gather up and cache all public schema column data types
            // in a single metadata query, rather than with many individual
            // queries.  The cache only lasts as long as the connection, which
            // is assumed to be shared across many of these data type
            // requests.
            Map<ColumnKey, Integer> types = columnTypes.get(conn);
            if (types == null)
            {
               types = new HashMap<>();
               columnTypes.put(conn, types);
            }
            
            ColumnKey key = new ColumnKey(table, column);
            Integer dataType = types.get(key);
            
            if (dataType == null)
            {
               meta = conn.getMetaData();
               // TODO:  is "public" schema portable?
               ResultSet rs = meta.getColumns(null, _public, "", "");
               
               while (rs.next())
               {
                  // all results must be processed using the dialect
                  String nextTab = rs.getString("TABLE_NAME");
                  nextTab = dialect.processMetadataResult(nextTab);
                  
                  String nextCol = rs.getString("COLUMN_NAME");
                  nextCol = dialect.processMetadataResult(nextCol);
                  
                  int nextType = rs.getInt("DATA_TYPE");
                  ColumnKey nextKey = new ColumnKey(nextTab, nextCol);
                  if (key.equals(nextKey))
                  {
                     dataType = nextType;
                  }
                  types.put(nextKey, nextType);
               }
            }
            
            if (dataType != null)
            {
               return dataType;
            }
         }
         
         // Either we are not in the persistence initialization phase or the
         // all-in-one strategy failed.  The latter could happen if a JDBC
         // driver does not honor an empty string for the column name pattern
         // in the meta.getColumns(...) call above.  Try an individual
         // metadata query for the exact table and column name instead.  This
         // is less efficient, but may be necessary with some drivers.
         // TODO:  is "public" schema portable?
         if (meta == null)
         {
            meta = conn.getMetaData();
         }
         
         // the parameters will be prepared using dialect specific semantics
         ResultSet rs = meta.getColumns(null, _public, _table, _column);
         if (rs.next())
         {
            return rs.getInt("DATA_TYPE");
         }
         
         throw new PersistenceException(
            "Unable to determine data type for column '" +
            table +
            ":" +
            column +
            "'");
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         if (conn != null)
         {
            DatabaseManager.checkInConnection(conn, database);
         }
      }
   }
   
   /**
    * Override the parent's implementation to return a hash code cached at construction.
    */
   @Override
   public final int hashCode()
   {
      return hashCode;
   }
   
   /**
    * Override the parent's implementation to ensure it is consistent with {@link #hashCode}.
    */
   @Override
   public final boolean equals(Object o)
   {
      return super.equals(o);
   }
   
   /**
    * Get the database dialect helper object associated with the database
    * backing this object.
    *
    * @return  Database dialect helper.
    */
   public Dialect getDialect()
   {
      return dialect;
   }
   
   /**
    * Get the database information object associated with this persistence
    * service object.
    *
    * @return  Database information object.
    */
   public Database getDatabase()
   {
      return database;
   }
   
   /**
    * Get the lock manager associated with this <code>Persistence</code>
    * instance.  For temp tables, this will be <code>null</code>, since temp
    * table data is naturally compartmentalized and record locking is
    * meaningless in this case.  For permanent databases, this method will
    * return the lock manager created when this class was instantiated.
    *
    * @return  Lock manager instance associated with this persistence service
    *          object.
    */
   public LockManager getLockManager()
   {
      return lockManager;
   }
   
   /**
    * Force the lock table updater to process all record lock events which are pending at the time
    * this method is invoked. Additional events may be added by other threads while this is
    * happening, but any events pending at the start of this method will be processed before this
    * method returns.
    */
   public void flushMetaLockEvents()
   {
      if (lockTableUpdater != null)
      {
         lockTableUpdater.flushQueue();
      }
   }
   
   /**
    * Get the identity manager associated with this <code>Persistence</code>
    * instance.  For temp tables, this will be <code>null</code>, since for
    * temp tables unique primary keys generated using {@link
    * TemporaryBuffer#nextPrimaryKey()}. For permanent databases, this
    * method will return the identity manager created when this class was
    * instantiated.
    *
    * @return  Identity manager instance associated with this persistence
    *          service object.
    */
   public IdentityManager getIdentityManager()
   {
      return identityManager;
   }
   
   /**
    * Execute an HQL query and get a scrollable cursor on the result set.
    * Records represented by the result set are not locked by this method.
    * This method creates a new, context local Hibernate session if necessary.
    * <p>
    * <em>Note:</em> while this method attempts to use a server-side database
    * cursor to fetch results, some JDBC driver implementations (e.g.,
    * PostgreSQL) may store an intermediate form of the <em>entire</em> result
    * set in memory instead, effectively doubling the amount of memory
    * required.  To avoid this, consider using {@link
    * #scroll(String, Object[], Type[], int, int, ScrollMode)} instead.  Using
    * that method with a <code>ScrollMode</code> of <code>FORWARD_ONLY</code>
    * and setting a JDBC fetch size to a reasonable value may allow a JDBC
    * driver to use a server-side cursor to fetch smaller batches of results
    * at a time.  Consult your JDBC driver documentation and/or source code to
    * determine whether this is an issue for your JDBC implementation.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    *
    * @return  An object which provides access to a scrollable cursor on the
    *          result set.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    * 
    * @see     DBUtils#makeTypeArray(Object[])
    * @see     DBUtils#makeTypeArray(Object[], Type[])
    */
   public ScrollableResults scroll(String hql, Object[] values, Type[] types)
   throws PersistenceException
   {
      return scroll(hql, values, types, 0, 0);
   }
   
   /**
    * Execute an HQL query and get a scrollable cursor on the result set.
    * Records represented by the result set are not locked by this method.
    * This method creates a new, context local Hibernate session if necessary.
    * <p>
    * <em>Note:</em> while this method attempts to use a server-side database
    * cursor to fetch results, some JDBC driver implementations (e.g.,
    * PostgreSQL) may store an intermediate form of the <em>entire</em> result
    * set in memory instead, effectively doubling the amount of memory
    * required.  To avoid this, consider using {@link
    * #scroll(String, Object[], Type[], int, int, ScrollMode)} instead.  Using
    * that method with a <code>ScrollMode</code> of <code>FORWARD_ONLY</code>
    * and setting a JDBC fetch size to a reasonable value may allow a JDBC
    * driver to use a server-side cursor to fetch smaller batches of results
    * at a time.  Consult your JDBC driver documentation and/or source code to
    * determine whether this is an issue for your JDBC implementation.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    *
    * @return  An object which provides access to a scrollable cursor on the
    *          result set.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    * 
    * @see     DBUtils#makeTypeArray(Object[])
    * @see     DBUtils#makeTypeArray(Object[], Type[])
    */
   public ScrollableResults scroll(String hql,
                                   Object[] values,
                                   Type[] types,
                                   int maxResults,
                                   int startOffset)
   throws PersistenceException
   {
      return scroll(hql,
                    values,
                    types,
                    maxResults,
                    startOffset,
                    ScrollMode.SCROLL_INSENSITIVE);
   }
   
   /**
    * Execute an HQL query and get a scrollable cursor on the result set.
    * Records represented by the result set are not locked by this method.
    * This method creates a new, context local Hibernate session if necessary.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   scrollMode
    *          Scroll mode which should apply to the result set.
    *
    * @return  An object which provides access to a scrollable cursor on the
    *          result set.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public ScrollableResults scroll(String hql,
                                   Object[] values,
                                   Type[] types,
                                   int maxResults,
                                   int startOffset,
                                   ScrollMode scrollMode)
   throws PersistenceException
   {
      return scroll(null, hql, values, types, maxResults, startOffset, scrollMode);
   }
   
   /**
    * Execute an HQL query and get a scrollable cursor on the result set.
    * Records represented by the result set are not locked by this method.
    * This method creates a new, context local Hibernate session if necessary.
    *
    * @param   entities
    *          Names of the DMO entities associated with this query statement.
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   scrollMode
    *          Scroll mode which should apply to the result set.
    *
    * @return  An object which provides access to a scrollable cursor on the
    *          result set.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public ScrollableResults scroll(String[] entities,
                                   String hql,
                                   Object[] values,
                                   Type[] types,
                                   int maxResults,
                                   int startOffset,
                                   ScrollMode scrollMode)
   throws PersistenceException
   {
      boolean trace = LOG.isLoggable(Level.FINEST);
      boolean warn = (trace || LOG.isLoggable(Level.WARNING));
      long start = (warn ? System.currentTimeMillis() : 0L);
      Context local = context.get();
      ScrollableResults results = null;
      
      boolean readOnly = checkReadOnly(entities);
      
      try
      {
         Query query = getQuery(local, hql, maxResults, startOffset, values, types, readOnly);
         
         if (readOnly)
         {
            query.setFlushMode(FlushMode.MANUAL);
         }
         else if (!local.isFlushRequired(entities))
         {
            query.setFlushMode(FlushMode.MANUAL);
         }
         else if (local.isAutoCommit())
         {
            local.flushSession();
         }
         
         results = query.scroll(scrollMode);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = message("error executing query [" + hql + "]");
         
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, msg);
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Details:", exc);
            }
         }
         
         handleException(msg, exc);
      }
      finally
      {
         if (warn && scrollThreshold >= 0L)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > scrollThreshold);
            if (trace || slow)
            {
               String msg = message("TIMING:SCROLL [" + elapsed + "] {" + 
                                    startOffset + ":" + maxResults + "} " + hql);
               
               if (slow)
               {
                  LOG.log(Level.WARNING, msg);
               }
               else
               {
                  LOG.log(Level.FINEST, msg);
               }
            }
         }
      }
      
      return results;
   }
   
   /**
    * Execute an HQL query and return the results as a list.  If the query
    * returns multiple results per row, each element in the list will be an
    * array of <code>Object</code>s.
    * <p>
    * No locking is attempted.
    * <p>
    * <em>Note:</em> this method should be used only for queries that are
    * expected to return a relatively small result set, since the entire
    * result set is stored in the returned list.  Additionally, some JDBC
    * driver implementations (e.g., PostgreSQL) may store an intermediate form
    * of the result set in memory, effectively doubling the amount of memory
    * required.  To avoid this, consider using {@link
    * #scroll(String, Object[], Type[], int, int, ScrollMode)} instead.  Using
    * that method with a <code>ScrollMode</code> of <code>FORWARD_ONLY</code>
    * and setting a JDBC fetch size to a reasonable value may allow a JDBC
    * driver to use a server-side cursor to fetch smaller batches of results
    * at a time.  Consult your JDBC driver documentation and/or source code to
    * determine whether this is an issue for your JDBC implementation.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.  This method makes no
    *          assumptions as to the types of object(s) returned.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   readOnlyQuery
    *          <code>true</code> to execute the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  A list of results, or <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    * 
    * @see     DBUtils#makeTypeArray(Object[])
    * @see     DBUtils#makeTypeArray(Object[], Type[])
    */
   public <T> List<T> list(String hql,
                           Object[] values,
                           Type[] types,
                           int maxResults,
                           int startOffset,
                           boolean readOnlyQuery)
   throws PersistenceException
   {
      return list(null, hql, values, types, maxResults, startOffset, readOnlyQuery, false);
   }
   
   /**
    * Execute an HQL query and return the results as a list.  If the query
    * returns multiple results per row, each element in the list will be an
    * array of <code>Object</code>s.
    * <p>
    * No locking is attempted.
    * <p>
    * <em>Note:</em> this method should be used only for queries that are
    * expected to return a relatively small result set, since the entire
    * result set is stored in the returned list.  Additionally, some JDBC
    * driver implementations (e.g., PostgreSQL) may store an intermediate form
    * of the result set in memory, effectively doubling the amount of memory
    * required.  To avoid this, consider using {@link
    * #scroll(String, Object[], Type[], int, int, ScrollMode)} instead.  Using
    * that method with a <code>ScrollMode</code> of <code>FORWARD_ONLY</code>
    * and setting a JDBC fetch size to a reasonable value may allow a JDBC
    * driver to use a server-side cursor to fetch smaller batches of results
    * at a time.  Consult your JDBC driver documentation and/or source code to
    * determine whether this is an issue for your JDBC implementation.
    * 
    * @param   entities
    *          Names of the DMO entities associated with this query statement.
    * @param   hql
    *          HQL query statement passed to Hibernate.  This method makes no
    *          assumptions as to the types of object(s) returned.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   readOnlyQuery
    *          <code>true</code> to execute the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  A list of results, or <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    * 
    * @see     DBUtils#makeTypeArray(Object[])
    * @see     DBUtils#makeTypeArray(Object[], Type[])
    */
   public <T> List<T> list(String[] entities,
                           String hql,
                           Object[] values,
                           Type[] types,
                           int maxResults,
                           int startOffset,
                           boolean readOnlyQuery,
                           boolean noFlush)
   throws PersistenceException
   {
      boolean trace = LOG.isLoggable(Level.FINEST);
      boolean warn = (trace || LOG.isLoggable(Level.WARNING));
      long start = (warn ? System.currentTimeMillis() : 0L);
      Context local = context.get();
      List<T> results = null;
      
      boolean readOnlyEntities = checkReadOnly(entities);
      if (readOnlyEntities)
      {
         readOnlyQuery = true;
      }
      
      try
      {
         Query query =
            getQuery(local, hql, maxResults, startOffset, values, types, readOnlyQuery);
         
         if (noFlush || readOnlyEntities)
         {
            query.setFlushMode(FlushMode.MANUAL);
         }
         else if (!local.isFlushRequired(entities))
         {
            query.setFlushMode(FlushMode.MANUAL);
         }
         else if (local.isAutoCommit())
         {
            local.flushSession();
         }
         
         results = (List<T>) query.list();
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = message("error executing query [" + hql + "]");
         
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, msg);
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Details:", exc);
            }
         }
         
         handleException(msg, exc);
      }
      finally
      {
         if (warn && listThreshold >= 0L)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > listThreshold);
            if (trace || slow)
            {
               String msg = message("TIMING:LIST [" + elapsed + "] {" + 
                                    startOffset + ":" + maxResults + "} " + hql);
               
               if (slow)
               {
                  LOG.log(Level.WARNING, msg);
               }
               else
               {
                  LOG.log(Level.FINEST, msg);
               }
            }
         }
      }
      
      return (results == null || results.isEmpty() ? null : results);
   }
   
   /**
    * Execute an HQL query and load a single record from the session-level cache or, optionally,
    * refresh it from the database. The record will be locked with the specified lock type upon
    * return from this method. Note that the lock refers to a runtime-level lock, not a
    * database-level lock. The record is <i>not</i> automatically loaded into the record buffer
    * provided. The <code>buffer</code> argument is used only to extract information needed
    * to perform the query.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO class type
    *          and name of the table being queried.
    * @param   hql
    *          HQL query statement passed to Hibernate. This may be either an
    *          ID-only projection query or a full query (see above).
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   timeout
    *          Number of milliseconds to wait to acquire the requested lock type.
    *          Throws {@link LockTimeoutException} if this time elapses before the
    *          lock is acquired. Specify 0 to wait indefinitely for the lock. Ignored
    *          if {@code lockType} is {@code NONE}.
    * @param   unique
    *          {@code true} if no more than one result is expected from this query;
    *          {@code false} if the query could result in
    *          more than one record.
    * @param   updateLock
    *          {@code true} to update the current lock state;
    *          {@code false} to leave the current lock state unchanged.
    * @param   forceRefresh
    *          If {@code true} forces the record to be refreshed from database. Otherwise
    *          it will be refreshed only if the lockType has been upgraded from NONE.
    * @param   findByKey
    *          If {@code true}, this overrides the current {@code rereadnolock} setting;
    *          if {@code false}, honor {@code rereadnolock}.
    *
    * @return  A single DMO instance representing the query result, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockTimeoutException
    *          if a record was found, but its lock could not be acquired within the positive
    *          {@code timeout} period specified.
    * @throws  LockUnavailableException
    *          if a record was found and a no-wait lock type was specified, but the lock could
    *          not be acquired immediately.
    * @throws  PersistenceException
    *          if there was an error executing the query or if a unique query
    *          (i.e., <code>unique == true</code>) returned multiple results,
    */
   public Persistable load(RecordBuffer buffer,
                           String hql,
                           Object[] values,
                           Type[] types,
                           LockType lockType,
                           long timeout,
                           boolean unique,
                           boolean updateLock,
                           boolean forceRefresh,
                           boolean findByKey)
   throws PersistenceException
   {
      boolean trace = LOG.isLoggable(Level.FINEST);
      boolean warn = (trace || LOG.isLoggable(Level.WARNING));
      boolean trackSlow = warn && loadThreshold > 0L;
      long start = (trackSlow ? System.currentTimeMillis() : 0L);
      
      Context local = context.get();
      
      boolean inlineLimit = (values == null || values.length == 0);
      Serializable id = null;
      Persistable initialResult = null;
      Persistable result = null;
      String table = buffer.getTable();
      LockType oldLock = null;
      RecordLockContext lockContext = local.getRecordLockContext();
      P2JDialect p2jDialect = (P2JDialect) dialect;
      
      if (temporary)
      {
         updateLock = false;
      }
      
      try
      {
         // The inlineLimit flag is a crude test to detect a "simple" query:
         // no substitution parameters.  We want to avoid inlining the limit
         // on more complex statements, as the intended performance benefit
         // may backfire in these cases.
         p2jDialect.inlineLimit(inlineLimit);
         
         // We have to give queries which are supposed to return a unique
         // result enough rope to hang themselves, so set max results to 2
         // for these.  All other queries are limited to 1 result.
         String entity = buffer.getEntityName();
         boolean readOnly = !temporary && DatabaseManager.isReadOnly(entity);
         Query query = getQuery(local, hql, unique ? 2 : 1, 0, values, types, readOnly);
         
         // disable flushing if possible, for better performance
         if (readOnly)
         {
            query.setFlushMode(FlushMode.MANUAL);
         }
         else if (!local.isFlushRequired(entity))
         {
            query.setFlushMode(FlushMode.MANUAL);
         }
         else if (local.isAutoCommit())
         {
            local.flushSession();
         }
         
         // Reset ambiguous flag in buffer.
         buffer.setAmbiguous(false);
         
         // Find the record.
         Object o = query.uniqueResult();
         
         // Reset limit inlining if necessary.
         if (inlineLimit)
         {
            p2jDialect.inlineLimit(false);
            inlineLimit = false;
         }
         
         if (o == null)
         {
            // No hit.
            return null;
         }
         
         if (o instanceof Persistable)
         {
            // This is the fast path where we assume we retrieved the DMO
            // directly instead of the ID because we don't have to lock
            // (we might need to unlock, however).
            // This is the case for temp-tables and no-lock queries.
            initialResult = result = (Persistable) o;
            
            // Lock record if required, remembering old lock type in case we
            // need to roll back on error.
            if (updateLock)
            {
               id = result.getId();
               RecordIdentifier ident = new RecordIdentifier(table, id);
               oldLock = queryLock(ident);
               
               if (oldLock.compareTo(lockType) != 0)
               {
                  lockContext.lock(ident, lockType, timeout, null);
                  
                  // ensure that if we have acquired a new lock, we have the latest version of
                  // the record from the database
                  if (!readOnly && !forceRefresh && oldLock == LockType.NONE)
                  {
//                     LOG.info("LOAD refresh (lock acquisition 1) [" + hql + "]");
                     forceRefresh = true;
                  }
               }
            }
            
            // If we found the record with a non-primary-key query and it already exists in any
            // buffer, it is possible it came from the session cache and needs to be refreshed
            if (!forceRefresh &&
                !temporary &&
                !readOnly &&
                oldLock == LockType.NONE &&
                lockType == LockType.NONE &&
                !findByKey &&
                local.isRereadNoLock())
            {
               // if any other buffer references this DMO already, it means we got it from
               // the session cache and it might be stale, so force a refresh
               int useCount = local.bufferManager.getDMOUseCount(result);
               if (useCount > 0)
               {
//                  LOG.info("LOAD refresh (-rereadnolock 1) [" + hql + "]");
                  forceRefresh = true;
               }
            }
         }
         else
         {
            // If we got here, we're dealing with a primary key query.
            id = (Serializable) o;
            RecordIdentifier ident = new RecordIdentifier(table, id);
            
            // Lock record if required, remembering old lock type in case we
            // need to roll back on error.
            if (updateLock)
            {
               oldLock = queryLock(ident);
               lockContext.lock(ident, lockType, timeout, null);
            }
            else
            {
               lock(lockType, ident, false);
            }
            
            // We use get() instead of load() here because we don't want to return a proxy and
            // have to take another hit retrieving the data almost immediately after this method
            // returns, which is the common case for dynamic record retrieval.
            Session session = local.getSession();
            initialResult = result = (Persistable) session.get(entity, id);
            
            if (result == null)
            {
               return null;
            }
            
            // Ensure that if we are acquiring a lock, we have the latest version of the record
            // from the database. The Hibernate API advertises that Session.get() with a LockMode
            // of READ should bypass both levels of cache, but debugging indicates that the
            // record is always taken from the session cache if present.
            // So, we brute force a refresh instead.
            if (!forceRefresh && updateLock && oldLock == LockType.NONE && !readOnly)
            {
               if (lockType != LockType.NONE)
               {
//                  LOG.info("LOAD refresh (lock acquisition 2) [" + hql + "]");
                  
                  // forced refresh if lock is upgraded
                  forceRefresh = true;
               }
               // possibly force refresh if rereadnolock is enabled, and query is not by
               // recid/rowid
               else if (lockType == LockType.NONE && !findByKey && local.isRereadNoLock())
               {
                  // if any other buffer references this DMO already, it means we got it from
                  // the session cache and it might be stale (since we have no lock on it), so
                  // force a refresh
                  int useCount = local.bufferManager.getDMOUseCount(result);
                  if (useCount > 0)
                  {
//                     LOG.info("LOAD refresh (-rereadnolock 2) [" + hql + "]");
                     forceRefresh = true;
                  }
               }
            }
         }
         
         // The Hibernate API advertises that Session.get() with a LockMode of READ should
         // bypass both levels of cache, but debugging indicates that the record is always
         // taken from the session cache if present. So, we brute force a refresh instead.
         if (!readOnly && forceRefresh)
         {
            Session session = local.getSession();
            session.refresh(result);
         }
      }
      catch (NonUniqueResultException exc)
      {
         // This is recoverable, no need to close session.
         uniqueResultViolation(buffer, exc);
      }
      catch (UnresolvableObjectException exc)
      {
         // the initial result we found must be evicted from the session cache
         if (initialResult != null)
         {
            Session session = local.getSessionNoCreate();
            if (session != null)
            {
               session.evict(initialResult);
            }
         }
         
         // Catching this exception indicates we were holding a stale copy of
         // a record that was deleted in the meantime.  This can happen in
         // bulk delete cases, or when a preselect query is holding an old
         // reference.  This is not a fatal error, it just means that the
         // refresh failed and the result is null.
         result = null;
         RecordIdentifier ident = new RecordIdentifier(table, id);
         throw new MissingRecordException(ident);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         // if exc was caused by ConditionException and 
         //    ErrorManager is in CAN-FIND mode, let the session alive
         Throwable cause = exc;
         if (cause != null && !(cause instanceof ConditionException))
         {
            cause = cause.getCause();
         }
         if (cause == null && !ErrorManager.isInCanFindBracket())
         {
            local.closeSession(true);
         }
         
         String msg = message("error loading record");
         
         if (warn)
         {
            LOG.log(Level.WARNING, msg + System.getProperty("line.separator") + hql);
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Details:", exc);
            }
         }
         
         handleException(msg, exc);
      }
      finally
      {
         // Reset limit inlining if (still) necessary.
         if (inlineLimit)
         {
            p2jDialect.inlineLimit(false);
         }
         
         if (updateLock      &&
             id != null      &&
             result == null  &&
             oldLock != null &&
             oldLock.compareTo(lockType) != 0)
         {
            // Reset lock to its previous state.  This is required if another
            // thread deleted the record after we got its ID but before we
            // retrieved its data.
            try
            {
               RecordIdentifier ident = new RecordIdentifier(table, id);
               if (oldLock == LockType.NONE)
               {
                  lockContext.forceRelease(ident, null);
               }
               else
               {
                  lockContext.lock(ident, oldLock.toNoWaitVariant());
               }
            }
            catch (LockUnavailableException exc)
            {
               // OK to ignore.
            }
         }
         
         if (trackSlow)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > loadThreshold);
            if (trace || slow)
            {
               String msg = message("TIMING:LOAD " + (result != null ? "+" : "-") + 
                                    "[" + elapsed + "] " + hql);
               
               if (slow)
               {
                  LOG.log(Level.WARNING, msg);
               }
               else
               {
                  LOG.log(Level.FINEST, msg);
               }
            }
         }
      }
      
      return result;
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * The record will be locked with the specified lock type upon return from
    * this method.  Note that the lock refers to a runtime-level lock, not a
    * database-level lock.  The record is <i>not</i> automatically loaded into
    * the record buffer provided.  The <code>buffer</code> argument is used
    * only to extract information needed to perform the locking and loading.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO class type
    *          and name of the table being queried.
    * @param   id
    *          Primary key of the record to be loaded into the buffer.  May
    *          not be <code>null</code>.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   timeout
    *          Number of milliseconds to wait to acquire the requested lock type.
    *          Throws {@link LockTimeoutException} if this time elapses before the
    *          lock is acquired. Specify 0 to wait indefinitely for the lock. Ignored
    *          if {@code lockType} is {@code NONE}.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    * @param   forceRefresh
    *          If <code>true</code> forces the record to be refreshed from database. Otherwise
    *          it will be refreshed only if the lockType has been upgraded from NONE.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    * 
    * @throws  LockTimeoutException
    *          if a record was found, but its lock could not be acquired within the positive
    *          {@code timeout} period specified.
    * @throws  LockUnavailableException
    *          if a record was found and a no-wait lock type was specified, but the lock could
    *          not be acquired immediately.
    * @throws  PersistenceException
    *          if there was an error loading the DMO from the Hibernate
    *          session and/or database.
    */
   public Persistable load(RecordBuffer buffer,
                           Serializable id,
                           LockType lockType,
                           long timeout,
                           boolean updateLock,
                           boolean forceRefresh)
   throws PersistenceException
   {
      String table = buffer.getTable();
      String entity = buffer.getEntityName();
      
      return load(table, entity, id, lockType, timeout, updateLock, forceRefresh);
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * The record will be locked with the specified lock type upon return from
    * this method.  Note that the lock refers to a runtime-level lock, not a
    * database-level lock.
    *
    * @param   table
    *          Table from which record is to be read.
    * @param   dmoClass
    *          DMO implementation class of record being read.
    * @param   id
    *          Primary key of the record to be loaded into the buffer.  May
    *          not be <code>null</code>.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockUnavailableException
    *          if a record was found but its lock could not be acquired.
    * @throws  PersistenceException
    *          if there was an error loading the DMO from the Hibernate
    *          session and/or database.
    */
   public Persistable load(String table,
                           Class<?> dmoClass,
                           Serializable id,
                           LockType lockType,
                           boolean updateLock)
   throws PersistenceException
   {
      return load(table, dmoClass.getName(), id, lockType, updateLock, false);
   }
   
   /**
    * Load a single record from the session-level cache or, optionally, refresh it from the
    * database. The record will be locked with the specified lock type upon return from this
    * method. Note that the lock refers to a runtime-level lock, not a database-level lock.
    *
    * @param   table
    *          Table from which record is to be read.
    * @param   entityName
    *          Name of DMO implementation class of record being read.
    * @param   id
    *          Primary key of the record to be loaded into the buffer.  May
    *          not be <code>null</code>.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    * @param   forceRefresh
    *          If <code>true</code> forces the record to be refreshed from database. Otherwise
    *          it will be refreshed only if the lockType has been upgraded from NONE.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockUnavailableException
    *          if a record was found and a no-wait lock type was specified, but the lock could
    *          not be acquired immediately.
    * @throws  PersistenceException
    *          if there was an error loading the DMO from the Hibernate
    *          session and/or database.
    */
   public Persistable load(String table,
                           String entityName,
                           Serializable id,
                           LockType lockType,
                           boolean updateLock,
                           boolean forceRefresh)
   throws PersistenceException
   {
      return load(table, entityName, id, lockType, 0L, updateLock, forceRefresh);
   }
   
   /**
    * Load a single record from the session-level cache or, optionally, refresh it from the
    * database. The record will be locked with the specified lock type upon return from this
    * method. Note that the lock refers to a runtime-level lock, not a database-level lock.
    *
    * @param   table
    *          Table from which record is to be read.
    * @param   entityName
    *          Name of DMO implementation class of record being read.
    * @param   id
    *          Primary key of the record to be loaded into the buffer.  May
    *          not be <code>null</code>.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   timeout
    *          Number of milliseconds to wait to acquire the requested lock type.
    *          Throws {@link LockTimeoutException} if this time elapses before the
    *          lock is acquired. Specify 0 to wait indefinitely for the lock. Ignored
    *          if {@code lockType} is {@code NONE}.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    * @param   forceRefresh
    *          If <code>true</code> forces the record to be refreshed from database. Otherwise
    *          it will be refreshed only if the lockType has been upgraded from NONE.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockTimeoutException
    *          if a record was found, but its lock could not be acquired within the positive
    *          {@code timeout} period specified.
    * @throws  LockUnavailableException
    *          if a record was found and a no-wait lock type was specified, but the lock could
    *          not be acquired immediately.
    * @throws  PersistenceException
    *          if there was an error loading the DMO from the Hibernate
    *          session and/or database.
    */
   public Persistable load(String table,
                           String entityName,
                           Serializable id,
                           LockType lockType,
                           long timeout,
                           boolean updateLock,
                           boolean forceRefresh)
   throws PersistenceException,
          LockTimeoutException
   {
      boolean trace = LOG.isLoggable(Level.FINEST);
      boolean warn = (trace || LOG.isLoggable(Level.WARNING));
      boolean trackSlow = warn && loadThreshold > 0L;
      long start = (trackSlow ? System.currentTimeMillis() : 0L);
      
      Persistable initialResult = null;
      Persistable result = null;
      Context local = context.get();
      LockType oldLock = LockType.NONE;
      RecordIdentifier ident = new RecordIdentifier(table, id);
      RecordLockContext lockContext = local.getRecordLockContext();
      boolean resetLock = false;
      boolean readOnly = !temporary && DatabaseManager.isReadOnly(entityName);
      
      if (temporary)
      {
         updateLock = false;
      }
      
      try
      {
         // Lock record if required, remembering old lock type in case we
         // need to roll back on error.
         if (updateLock)
         {
            oldLock = queryLock(ident);
            lockContext.lock(ident, lockType, timeout, null);
         }
         else
         {
            lock(lockType, ident, false);
         }
         
         Session session = local.getSession();
         if (!readOnly && local.isFlushRequired(entityName))
         {
            local.flushSession();
         }
         
         // Use get() instead of load() here, to ensure that if the record
         // has been deleted, we get a null back, rather than an invalid proxy.
         initialResult = result = (Persistable) session.get(entityName, id);
         
         if (result == null)
         {
            resetLock = true;
            
            return null;
         }
         
         if (!readOnly && !temporary && oldLock == LockType.NONE)
         {
            // ensure that if we are acquiring a lock, we have the latest version of the record
            // from the database
            if (updateLock && lockType != LockType.NONE)
            {
//               LOG.info("LOADBYID refresh (lock acquisition) [" + entityName + ":" + id + "]");
               
               // forced refresh if lock is upgraded
               forceRefresh = true;
            }
            else if (lockType == LockType.NONE && local.isRereadNoLock())
            {
               // if any other buffer references this DMO already, it means we got it from
               // the session cache and it might be stale, so force a refresh
               int useCount = local.bufferManager.getDMOUseCount(result);
               if (useCount > 0)
               {
//                  LOG.info("LOADBYID refresh (-rereadnolock) [" + entityName + ":" + id + "]");
                  forceRefresh = true;
               }
               
               // TODO: there is one exception here: if the records are retrieved through
               // some kind of where expression having RECID or ROWID functions the 4GL
               // will NOT re-fetch the records even if the rereadnolock parameter is active
            }
         }
         
         // The Hibernate API advertises that Session.get() with a LockMode of READ should bypass
         // both levels of cache, but debugging indicates that the record is always taken from
         // the session cache if present. So, we brute force a refresh instead.
         if (forceRefresh)
         {
            session.refresh(result);
         }
      }
      catch (UnresolvableObjectException exc)
      {
         resetLock = true;
         
         // the initial result we found must be evicted from the session cache
         if (initialResult != null)
         {
            Session session = local.getSessionNoCreate();
            if (session != null)
            {
               session.evict(initialResult);
            }
         }
         
         // Catching this exception indicates we were holding a stale copy of
         // a record that was deleted in the meantime.  This can happen in
         // bulk delete cases, or when a preselect query is holding an old
         // reference.  This is not a fatal error, it just means that the
         // refresh failed and the result is null.
         throw new MissingRecordException(ident);
      }
      catch (HibernateException exc)
      {
         resetLock = true;
         
         try
         {
            local.closeSession(true);
         }
         finally
         {
            DBUtils.handleException(database, exc);
         }
         
         String msg = message("error loading record " + table + ":" + id);
         
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, msg);
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Details:", exc);
            }
         }
         
         handleException(msg, exc);
      }
      finally
      {
         if (resetLock && updateLock && oldLock.compareTo(lockType) != 0)
         {
            // reset lock to its previous state in case another thread deleted the record after
            // we got its ID but before we retrieved its data
            try
            {
               if (oldLock == LockType.NONE && result == null)
               {
                  lockContext.forceRelease(ident, null);
               }
               else
               {
                  lockContext.lock(ident, oldLock.toNoWaitVariant());
               }
            }
            catch (LockUnavailableException luExc)
            {
               // OK to ignore.
            }
         }
         
         if (trackSlow)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > loadByIDThreshold);
            if (trace || slow)
            {
               String msg = message("TIMING:LOAD_BY_ID " + (result != null ? "+" : "-") + 
                                    "[" + elapsed + "] " + table + ":" + id);
               
               if (slow)
               {
                  LOG.log(Level.WARNING, msg);
               }
               else
               {
                  LOG.log(Level.FINEST, msg);
               }
            }
         }
      }
      
      return result;
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * No locking is attempted, the assumption being that any required lock
    * previously was acquired.
    *
    * @param   ident
    *          Record identifier of the record to be loaded.  Consists of the
    *          primary key and the fully qualified DMO entity name.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error creating a Hibernate session or getting
    *          the requested record.
    */
   public Persistable quickLoad(RecordIdentifier ident)
   throws PersistenceException
   {
      return quickLoad(ident, false);
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * No locking is attempted, the assumption being that any required lock
    * previously was acquired.
    *
    * @param   ident
    *          Record identifier of the record to be loaded.  Consists of the
    *          primary key and the fully qualified DMO entity name.
    * @param   refresh
    *          If <code>true</code> then re-read the state of the record from
    *          the underlying database. This is useful when record has been
    *          changed bypassing Hibernate.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error creating a Hibernate session or getting
    *          the requested record.
    */
   public Persistable quickLoad(RecordIdentifier ident, boolean refresh)
   throws PersistenceException
   {
      Context local = context.get();
      String entity = ident.getTable();
      Serializable id = ident.getRecordID();
      Session session = local.getSession();
      Persistable result = null;
      boolean readOnly = !temporary && DatabaseManager.isReadOnly(entity);
      if (readOnly)
      {
         refresh = false;
      }
      
      try
      {
         if (!readOnly && local.isFlushRequired(entity))
         {
            local.flushSession();
         }
         
         // Use get() instead of load() here, to ensure that if the record
         // has been deleted, we get a null back, rather than an invalid proxy.
         result = (Persistable) session.get(entity, id);
         
         // The Hibernate API advertises that Session.get() with a LockMode of READ should bypass
         // both levels of cache, but debugging indicates that the record is always taken from
         // the session cache if present. So, we brute force a refresh instead.
         if (refresh && result != null)
         {
//            LOG.info("QUICKLOAD refresh [" + ident + "]");
            session.refresh(result);
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = message("error executing quick load [" + entity + ":" + id + "]");
         
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, msg);
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Details:", exc);
            }
         }
         
         handleException(msg, exc);
      }
      
      return result;
   }
   
   /**
    * Execute an arbitrary SQL statement which does not produce a result set.
    * This method relies on Hibernate to obtain the JDBC connection, but
    * otherwise bypasses Hibernate and executes the SQL directly, using JDBC.
    * The Hibernate session is flushed before this statement is executed, to
    * ensure any pending changes are synchronized with the database first.
    *
    * @param   sql
    *          SQL statement text.
    * 
    * @return  The number of entities updated or deleted, or {@code -1} if the statement
    *          did not complete normally.
    *
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   public int executeSQL(String sql)
   throws PersistenceException
   {
      return executeSQL(sql, null);
   }
   
   /**
    * Execute an arbitrary SQL statement which does not produce a result set.
    * This method relies on Hibernate to obtain the JDBC connection, but
    * otherwise bypasses Hibernate and executes the SQL directly, using JDBC.
    * The Hibernate session is flushed before this statement is executed, to
    * ensure any pending changes are synchronized with the database first.
    *
    * @param   sql
    *          SQL statement text.
    * @param   args
    *          The statement arguments.
    * 
    * @return  The number of entities updated or deleted, or {@code -1} if the statement
    *          did not complete normally.
    *          
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   public int executeSQL(String sql, Object[] args)
   throws PersistenceException
   {
      return executeSQL(sql, args, null, false);
   }
   
   /**
    * Execute an arbitrary SQL query.  This method relies on Hibernate to
    * obtain the JDBC connection, but otherwise bypasses Hibernate and
    * executes the SQL directly, using JDBC.  The Hibernate session is flushed
    * before this statement is executed, to ensure any pending changes are
    * synchronized with the database first.
    * <p>
    * It is the responsibility of the calling code to close the returned
    * result set.  In addition, if this method is called outside of a database
    * transaction, the caller must invoke {@link #pushImplicitTransaction()}
    * before calling this method and {@link #popImplicitTransaction(Context)} after
    * fully processing the result set returned by this method.  This will
    * ensure a transaction is opened and properly closed if necessary.
    *
    * @param   sql
    *          SQL query text.
    * 
    * @return  Result set produced by query.
    *
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   public ScrollableResults executeSQLQuery(String sql)
   throws PersistenceException
   {
      return executeSQLQuery(sql, null);
   }
   
   /**
    * Execute an arbitrary SQL query.  This method relies on Hibernate to
    * obtain the JDBC connection, but otherwise bypasses Hibernate and
    * executes the SQL directly, using JDBC.  The Hibernate session is flushed
    * before this statement is executed, to ensure any pending changes are
    * synchronized with the database first.
    * <p>
    * It is the responsibility of the calling code to close the returned
    * result set.  In addition, if this method is called outside of a database
    * transaction, the caller must invoke {@link #pushImplicitTransaction()}
    * before calling this method and {@link #popImplicitTransaction(Context)} after
    * fully processing the result set returned by this method.  This will
    * ensure a transaction is opened and properly closed if necessary.
    *
    * @param   sql
    *          SQL query text.
    * @param   args
    *          The query parameters.
    *          
    * @return  Result set produced by query.
    *
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   public ScrollableResults executeSQLQuery(String sql, Object[] args)
   throws PersistenceException
   {
      return executeSQLQuery(sql, args, null, false);
   }
   
   /**
    * Execute an arbitrary SQL query.  This method relies on Hibernate to
    * obtain the JDBC connection, but otherwise bypasses Hibernate and
    * executes the SQL directly, using JDBC.  The Hibernate session is flushed
    * before this statement is executed, to ensure any pending changes are
    * synchronized with the database first.
    * <p>
    * It is the responsibility of the calling code to close the returned
    * result set.  In addition, if this method is called outside of a database
    * transaction, the caller must invoke {@link #pushImplicitTransaction()}
    * before calling this method and {@link #popImplicitTransaction(Context)} after
    * fully processing the result set returned by this method.  This will
    * ensure a transaction is opened and properly closed if necessary.
    *
    * @param   sql
    *          SQL query text.
    * @param   args
    *          The query parameters.
    * @param   entities
    *          Array of entity names involved in the SQL operation represented by this
    *          statement, or {@code null} to check whether a flush is required for
    *          <i>any</i> entity type.
    * @param   noFlush
    *          If {@code true}, flushing is bypassed.
    *          
    * @return  Result set produced by query.
    *
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   public ScrollableResults executeSQLQuery(String sql,
                                            Object[] args,
                                            String[] entities,
                                            boolean noFlush)
   throws PersistenceException
   {
      long start = 0L;
      boolean trace = LOG.isLoggable(Level.FINEST);
      if (trace)
      {
         start = System.currentTimeMillis();
      }
      
      Context local = context.get();
      
      ScrollableResults results = null;
      
      try
      {
         Session session = local.getSession();
         
         if (!noFlush && local.isFlushRequired(entities))
         {
            local.flushSession();
         }
         
         SQLQuery query = session.createSQLQuery(sql);
         
         if (args != null)
         {
            for (int i = 0; i < args.length; i++)
            {
               query.setParameter(i, args[i]);
            }
         }
         
         results = query.scroll();
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = "Error executing SQL statement:  " + sql;
         handleException(msg, exc);
      }
      finally
      {
         if (trace)
         {
            long elapsed = System.currentTimeMillis() - start;
            LOG.log(Level.FINEST, message("TIMING:SQLQ [" + elapsed + "] " + sql));
         }
      }
      
      return results;
   }
   
   /**
    * Convenience method to retrieve a single result from a SQL statement. This is intended to
    * be used for SQL statements that return only one result. Only the first result is checked;
    * if there is more than one in the result set, all but the first are ignored.
    * <p>
    * An implicit transaction will be opened, if an explicit one does not already exist.
    * The result set and statement are closed immediately; no additional cleanup is required by
    * the caller.
    * 
    * @param   sql
    *          SQL statement to execute.
    * @param   args
    *          Query substitution parameters, if any. If present, these should match the number
    *          of placeholders in the SQL statement. If not present, this argument may either be
    *          {@code null} or an empty array.
    * @param   entity
    *          Entity associated with the SQL statement, if any. Used to check whether a session
    *          flush is needed before execution.
    * @param   noFlush
    *          If {@code true}, flushing is bypassed.
    * 
    * @return  The first (and presumably only) result returned by the SQL statement.
    * 
    * @throws  PersistenceException
    *          if any error occurs, either during the statement's execution, or when pushing or
    *          popping an implicit transaction.
    */
   public Object getSingleSQLResult(String sql, Object[] args, String entity, boolean noFlush)
   throws PersistenceException
   {
      Context local = context.get();
      Object result = null;
      
      try
      {
         local.pushImplicitTransaction();
         
         Session session = local.getSession();
         
         if (!noFlush && local.isFlushRequired(entity))
         {
            local.flushSession();
         }
         
         ReturningWork<Object> work = (Connection) ->
         {
            try (PreparedStatement stmt = Connection.prepareStatement(sql))
            {
               if (args != null)
               {
                  int len = args.length;
                  for (int i = 0; i < len; i++)
                  {
                     stmt.setObject(i + 1, args[i]);
                  }
               }
               
               ResultSet rs = stmt.executeQuery();
               
               return rs.next() ? rs.getObject(1) : null;
            }
         };
         
         result = session.doReturningWork(work);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = "Error executing SQL statement:  " + sql;
         handleException(msg, exc);
      }
      finally
      {
         local.popImplicitTransaction();
      }
      
      return result;
   }
   
   /**
    * Execute an arbitrary list of SQL statements in batch.  This method relies on Hibernate to
    * obtain the JDBC connection, but otherwise bypasses Hibernate and executes the SQL directly,
    * using JDBC.  The Hibernate session is flushed before the statements are executed, to ensure
    * any pending changes are synchronized with the database first.
    *
    * @param   sql
    *          List of SQL statements to execute in batch.
    * @param   entities
    *          Array of entity names involved in the SQL operation represented by the batched
    *          statements, or {@code null} to check whether a flush is required for
    *          <i>any</i> entity type.
    * @param   noFlush
    *          If {@code true}, flushing is bypassed.
    *
    * @throws  PersistenceException
    *          if any error occurs executing the statements.
    */
   public void executeSQLBatch(List<String> sql, String[] entities, boolean noFlush)
   throws PersistenceException
   {
      if (sql.isEmpty())
      {
         return;
      }
      
      long start = 0L;
      boolean trace = LOG.isLoggable(Level.FINEST);
      if (trace)
      {
         start = System.currentTimeMillis();
      }
      
      Context local = context.get();
      
      try
      {
         local.pushImplicitTransaction();
         Session session = local.getSession();
         
         if (!noFlush && local.isFlushRequired(entities))
         {
            local.flushSession();
         }
         
         Work work = (Connection) ->
         {
            Statement stmt = Connection.createStatement();
            int batchSize = 50;
            
            try
            {
               int i = 0;
               for (String s : sql)
               {
                  stmt.addBatch(s);
                  if (++i % batchSize == 0)
                  {
                     stmt.executeBatch();
                  }
               }
               
               if (i % batchSize != 0)
               {
                  stmt.executeBatch();
               }
            }
            finally
            {
               if (stmt != null)
               {
                  stmt.close();
               }
            }
         };
         
         session.doWork(work);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = "Error executing SQL statement:  " + sql;
         handleException(msg, exc);
      }
      finally
      {
         local.popImplicitTransaction();
         
         if (trace)
         {
            long elapsed = System.currentTimeMillis() - start;
            LOG.log(Level.FINEST, message("TIMING:SQLB [" + elapsed + "] " + sql));
         }
      }
   }
   
   /**
    * Attempt to obtain the specified lock type on a particular record.  If the lock cannot be
    * obtained immediately, the current thread will block until the lock becomes available and
    * it is obtained, unless the requested lock type is a "no-wait" lock.
    * <p>
    * In the no-wait case, an exception is raised indicating that the lock is not available.
    * <p>
    * The current thread continues normally once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying {@code LockType.NONE} for the
    * {@code lockType} parameter.
    * <p>
    * The normal return of this method indicates that the lock request (or lack thereof)
    * completed successfully, and that the current context now holds a lock of the requested
    * type.
    * <p>
    * NOTE: legacy persistence framework code should access this method only via a {@link
    * RecordLockContext} method, unless simply testing whether a lock is available (i.e.,
    * {@code update == false}). This ensures that record locks shared across record buffers
    * within a client context are managed properly.  Non-legacy code may access this API
    * directly.
    *
    * @param   lockType
    *          Type of lock to be obtained. {@code LockType.NONE} is used to release an existing
    *          lock, and to continue with no lock.
    * @param   ident
    *          ID which uniquely identifies the record being locked.
    * @param   update
    *          {@code true} if the lock state for the current record should be updated;
    *          else {@code false}.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    *
    * @see     #getRecordLockContext
    */
   public void lock(LockType lockType, RecordIdentifier ident, boolean update)
   throws LockUnavailableException
   {
      if (lockManager != null)
      {
         lockManager.lock(lockType, ident, update);
      }
   }
   
   /**
    * Attempt to obtain the specified lock type on a particular record.  If the lock cannot be
    * obtained immediately, the current thread will block until the lock becomes available and
    * it is obtained, unless the requested lock type is a "no-wait" lock, or a positive timeout
    * value is specified and that time elapses before the lock is acquired.
    * <p>
    * In the no-wait case, an exception is raised indicating that the lock is not available.
    * In the timeout case, an exception is raised indicating that the timeout period elapsed.
    * <p>
    * The current thread continues normally once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying {@code LockType.NONE} for the
    * {@code lockType} parameter.
    * <p>
    * The normal return of this method indicates that the lock request (or lack thereof)
    * completed successfully, and that the current context now holds a lock of the requested
    * type.
    * <p>
    * NOTE: legacy persistence framework code should access this method only via a {@link
    * RecordLockContext} method, unless simply testing whether a lock is available (i.e.,
    * {@code update == false}). This ensures that record locks shared across record buffers
    * within a client context are managed properly.  Non-legacy code may access this API
    * directly.
    *
    * @param   lockType
    *          Type of lock to be obtained. {@code LockType.NONE} is used to release an existing
    *          lock, and to continue with no lock.
    * @param   ident
    *          ID which uniquely identifies the record being locked.
    * @param   update
    *          {@code true} if the lock state for the current record should be updated;
    *          else {@code false}.
    * @param   timeout
    *          Number of milliseconds to wait to acquire the requested lock type.
    *          Throws {@link LockTimeoutException} if this time elapses before the
    *          lock is acquired. Specify 0 to wait indefinitely for the lock. Ignored
    *          if {@code lockType} is {@code NONE}.
    *
    * @throws  LockTimeoutException
    *          if a lock cannot be acquired within the positive {@code timeout} period specified.
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    *
    * @see     #getRecordLockContext
    */
   public void lock(LockType lockType, RecordIdentifier ident, boolean update, long timeout)
   throws LockUnavailableException
   {
      if (lockManager != null)
      {
         lockManager.lock(lockType, ident, update, timeout);
      }
   }
   
   /**
    * Attempt to obtain the specified lock type on a table. If the lock cannot be obtained
    * immediately, the current thread will block until the lock becomes available and is
    * obtained, unless the requested lock type is a "no-wait" lock.  In the latter case,
    * an exception is raised, indicating that the lock is not available. The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * Method returns previous lock type (so that calling code knows what to set it back to after
    * bulk delete).
    * The normal return of this method indicates that the lock request (or lack thereof)
    * completed successfully, and that the current context now holds a lock of the requested
    * type.
    *
    * @param   table
    *          Table name.
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is used to release an
    *          existing lock, and to continue with no lock.
    * @param   update
    *          <code>true</code> if the lock state for the current table should be updated;
    *          else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    *
    * @return  previous lock type.
    */
   public LockType lockTable(String table, LockType lockType, boolean update)
   throws LockUnavailableException
   {
      if (lockManager != null)
      {
         return lockManager.lockTable(lockType, table, update);
      }
      
      return LockType.NONE;
   }
   
   /**
    * Query the lock type currently held by the current context for the
    * specified database record.
    * <p>
    * This method will always return the <b>non-</b><code>NO_WAIT</code>
    * variants of a lock type.  That is, even if the actual lock type is
    * <code>LockType.EXCLUSIVE_NO_WAIT</code> or
    * <code>LockType.SHARE_NO_WAIT</code>, the simpler types of
    * <code>LockType.EXCLUSIVE</code> and <code>LockType.SHARE</code>,
    * respectively, are returned.  The lack of a lock returns
    * <code>LockType.NONE</code> rather than <code>null</code>.
    *
    * @param   ident
    *          ID which uniquely identifies the record being queried.
    *
    * @return  The lock type currently held by the current context, never
    *          <code>null</code>.
    */
   public LockType queryLock(RecordIdentifier ident)
   {
      if (lockManager == null)
      {
         return LockType.NONE;
      }
      
      return lockManager.queryLock(ident);
   }
   
   /**
    * Obtain the next available primary key ID on a per-table basis.
    *
    * @param   table
    *          Table for which next primary key will be returned. 
    *
    * @return  The next available ID for the given table.
    * 
    * @throws  PersistenceException
    *          if there is an error determining the next primary key ID.
    * @throws  IllegalStateException
    *          if identity manager was not initialized.
    */
   public Serializable nextPrimaryKey(String table)
   throws PersistenceException
   {
      if (identityManager != null)
      {
         return identityManager.nextPrimaryKey(table);
      }
      else
      {
         throw new IllegalStateException("identity manager should be initialized");
      }
   }
   
   /**
    * Persist the specified object to the backing database session, using the
    * given ID.  It is assumed this method will only be invoked within a
    * database transaction.
    *
    * @param   dmo
    *          DMO to be persisted.
    * @param   id
    *          Primary key to associate with the persisted DMO.
    *
    * @throws  PersistenceException
    *          if a problem occurs saving the object in the current session.
    */
   public void save(Persistable dmo, Serializable id)
   throws PersistenceException
   {
      if (dmo.getId() != null && dmo.getId() < 0)
      {
         // this should not happen: we have already tested this in RecordBuffer in the event of
         // a field change. Since the template record cannot be altered, it should also not need
         // to be flushed/saved to persistence.
         final String msg = "Cannot update a template record without a schema lock";
         ErrorManager.recordOrThrowError(12376, msg, false, false);
         
         // print the error as SEVERE
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, 
                    "Template records MUST not be flushed/saved to database!",
                    new PersistenceException("Saving template records to database is forbidden"));
         }
         
         return;
      }
      
      Context local = context.get();
      
      try
      {
         local.pushImplicitTransaction();
         
         if (dmo.getId() == null)
         {
            dmo.setId((Long) id);
         }
         
         local.getSession().save(dmo);
         
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, message("Persisted DMO (" + dmo + ") with ID " + id));
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         handleException("error persisting object", exc);
      }
      finally
      {
         local.popImplicitTransaction();
      }
   }
   
   /**
    * Merges the state of the given object with the persistent object with the same identifier and
    * returns the merged object. See {@link org.hibernate.Session#merge(Object)} for more
    * information.
    *
    * @param  dmo
    *         DMO to merge.
    *
    * @return Merged DMO.
    */
   public Persistable merge(Persistable dmo)
   throws PersistenceException
   {
      Context local = context.get();
      
      Object merged = null;
      
      try
      {
         merged = local.getSession().merge(dmo);
         // Hibernate's Session.merge() saves the record if not done already.
         // The caller must mark the dmo as persisted.
         
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, message("Merged DMO (" + dmo + ")"));
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         handleException("error merging object", exc);
      }
      
      return (Persistable) merged;
   }
   
   /**
    * Get the Hibernate session associated with the current context.
    *
    * @return See above.
    */
   public Session getSession()
   {
      try
      {
         return context.get().getSession();
      }
      catch (PersistenceException e)
      {
         throw new IllegalStateException("Cannot get Hibernate session");
      }
   }
   
   /**
    * Ensure the specified object is being managed by the session associated
    * with the backing database in the current context.
    *
    * @param   object
    *          DMO to be persisted.
    *
    * @throws  PersistenceException
    *          if a problem occurs attaching the object to the current
    *          session.
    */
   public void attach(Object object)
   throws PersistenceException
   {
      Context local = context.get();
      
      try
      {
         Session session = local.getSession();
         if (!session.contains(object))
         {
            session.buildLockRequest(LockOptions.NONE).lock(object);
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         handleException("error attaching record to database session", exc);
      }
   }
   
   /**
    * Evict the specified object from the session associated with the backing
    * database in the current context.  This is a no-op if there is no active
    * session in the context.
    * 
    * @param   object
    *          Persistent entity to be evicted from the session.
    * 
    * @throws  PersistenceException
    *          if a problem occurs evicting the object from the current
    *          session.
    */
   public void detach(Object object)
   throws PersistenceException
   {
      Context local = context.get();
      evict(local, (Persistable) object);
   }
   
   /**
    * Clean up the current context.
    */
   public void cleanup()
   {
      Context ctxt = context.get(false);
      if (ctxt != null)
      {
         ctxt.cleanup();
      }
   }
   
   /**
    * Delete the persisted object from the specified database.
    * 
    * @param   object
    *          Persistent entity to be deleted.
    * 
    * @throws  PersistenceException
    *          if a problem occurs deleting the object.
    */
   public void delete(Object object)
   throws PersistenceException
   {
      Context local = context.get();
      
      try
      {
         Session session = local.getSession();
         session.delete(object);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         handleException("error deleting persistent object", exc);
      }
   }
   
   /**
    * Perform a bulk delete or update of all objects which meet the specified 
    * criteria. The criteria must indicate a single DMO type only.
    * <p>
    * This action must take place within a database transaction, so this
    * method attempts to begin one and commit it, if one is not already
    * active.
    * <p>
    * This method should only be used for temp-tables (where an exclusive
    * lock is irrelevant) or in cases where the current context already holds
    * locks on all records which meet the specified criteria.
    *
    * @param   hql
    *          Query string which defines the criteria which determine which
    *          records are deleted or updated.
    * @param   values
    *          Query substitution parameters, if any.
    * @param   types
    *          Types of the query substitution parameters, if any.
    * @param   entity
    *          Entity associated with the HQL statement, if any. Used to check whether a session
    *          flush is needed before execution.
    *
    * @return  The number of entities deleted or updated.
    *
    * @throws  PersistenceException
    *          if a problem occurs performing the bulk delete or update.
    */
   public int deleteOrUpdate(String hql, Object[] values, Type[] types, String entity)
   throws PersistenceException
   {
      boolean trace = LOG.isLoggable(Level.FINEST);
      long start = System.currentTimeMillis();
      
      Context local = context.get();
      int count = 0;
      
      local.pushImplicitTransaction();
      try
      {
         Session session = local.getSessionNoCreate();
         if (session != null && local.isFlushRequired(entity))
         {
            local.flushSession();
         }
         
         Query query = getQuery(local, hql, 0, 0, values, types, false);
         count = query.executeUpdate();
         
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, "[" + hql + "] count = " + count);
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = "error performing bulk delete [" + hql + "]";
         handleException(msg, exc);
      }
      finally
      {
         local.popImplicitTransaction();
         
         if (trace)
         {
            long elapsed = System.currentTimeMillis() - start;
            LOG.log(Level.FINEST, message("TIMING:DELETE [" + elapsed + "] " + hql));
         }
      }
      
      return count;
   }
   
   /**
    * Begin a new Hibernate transaction and store the instance in the local
    * context.  A new database session is created if one does not already
    * exist in the current context.
    *
    * @return  <code>true</code> if a new transaction was begun;
    *          <code>false</code> if there was already an active transaction.
    *
    * @throws  PersistenceException
    *          if there is an error beginning the new transaction.
    */
   public boolean beginTransaction()
   throws PersistenceException
   {
      Context local = context.get();
      
      return beginTransaction(local);
   }
   
   /**
    * Commit the current Hibernate transaction and close the current session.
    * If the transaction cannot be committed, roll it back and raise an error.
    * 
    * @throws  IllegalStateException
    *          if no transaction is active in the current session;  this
    *          represents a programming error.
    * @throws  PersistenceException
    *          if there is an error committing the transaction.
    */
   public void commit()
   throws PersistenceException
   {
      Context local = context.get();
      commit(local);
   }
   
   /**
    * Roll back the current Hibernate transaction and close the current
    * session.  If the transaction cannot be rolled back, raise an error.
    * 
    * @throws  IllegalStateException
    *          if no transaction is active in the current session;  this
    *          represents a programming error.
    * @throws  PersistenceException
    *          if there is an error rolling back the transaction.
    */
   public void rollback()
   throws PersistenceException
   {
      Context local = context.get();
      rollback(local);
   }
   
   /**
    * Increment the implicit transaction use counter and begin an implicit
    * transaction if none is currently active, and the use count currently
    * is at zero.
    * 
    * @return  The implicit transaction use count after incrementing it.
    * 
    * @throws  PersistenceException
    *          if there is an error beginning a new transaction.
    */
   public int pushImplicitTransaction()
   throws PersistenceException
   {
      Context local = context.get();
      
      return local.pushImplicitTransaction();
   }
   
   /**
    * Decrement the implicit transaction use counter.  If it reaches zero as a result, remember
    * the current system time, so that the implicit transaction commit timer can commit the
    * current, implicit transaction when necessary.
    *
    * @param   local
    *          Cached context. If {@code null}, will use {code get()} to obtain it.
    * 
    * @return  The implicit transaction use count after decrementing it.
    */
   public int popImplicitTransaction(Context local)
   {
      if (local == null)
      {
         local = context.get();
      }
      
      return local.popImplicitTransaction();
   }
   
   /**
    * Flush a batch of records to the database, typically while inserting
    * records in bulk.  After the flush, the session is cleared of all
    * records, to prevent a progressive memory leak.
    * <p>
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * <p>
    * It is expected that this method will only be called within an active
    * transaction.  If no session is active, this method will raise an
    * exception.
    *
    * @throws  IllegalStateException
    *          if no session currently is active.
    * @throws  PersistenceException
    *          if an error occurs during session flush.
    */
   public void flushBatch()
   throws PersistenceException
   {
      flush(true);
   }
   
   /**
    * Flush the current session, which synchronizes the in-memory state of
    * DMOs to the database.
    * <p>
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * <p>
    * It is expected that this method will only be called within an active
    * transaction.  If no session is active, this method will raise an
    * exception.
    *
    * @throws  IllegalStateException
    *          if no session currently is active.
    * @throws  PersistenceException
    *          if an error occurs during session flush.
    */
   public void flush()
   throws PersistenceException
   {
      flush(false);
   }
   
   /**
    * Indicate whether an explicit, database-level transaction currently is open.
    *
    * @return  <code>true</code> if open, else <code>false</code>.
    */
   public boolean isTransactionOpen()
   {
      return context.get().isTransactionOpen(true);
   }
   
   /**
    * Format the given message text by prepending a descriptor including the
    * user's current context and the physical name of the database.
    *
    * @param   message
    *          Core message
    *
    * @return  Descriptor.
    */
   public String message(String message)
   {
      return "[" + Utils.describeContext() + "-->" + database + "] " + message;
   }
   
   /**
    * Get the record lock context associated with the current client session. Although is method
    * is public to allow access from other internal packages, it is not meant for use outside the
    * <code>persist</code> package and its sub-packages.
    *
    * @return  Record lock context for the current session.
    */
   public RecordLockContext getRecordLockContext()
   {
      return context.get().getRecordLockContext();
   }
   
   /**
    * Initialize the helper objects to which this object delegates certain service requests.
    */
   protected void initializeInstance()
   {
      try
      {
         lockManager = createLockManager();
         
         // If using embedded H2 dialect and not a permanent database, prepare it.
         if (dialect instanceof P2JH2Dialect)
         {
            if (temporary)
            {
               H2Helper.prepareDatabase(database);
            }
            else
            {
               H2Helper.preparePermanentDatabase(database);
            }
         }
         else if (!temporary && dialect instanceof P2JSQLServer2008Dialect)
         {
            SQLServerHelper.preparePermanentDatabase(database);
         }
         
         identityManager = createIdentityManager();
      }
      catch (StopConditionException sce)
      {
         throw sce;
      }
      catch (Exception exc)
      {
         throw new RuntimeException("Error initializing persistence services", exc);
      }
   }
   
   /**
    * Create and initialize the lock manager to be used with this instance of
    * persistence services for the given database.  The lock manager
    * implementation to use is read from the directory.  The default if no
    * entry is found is the {@link InMemoryLockManager}.
    *
    * @return  Lock manager instance.
    *
    * @throws  IllegalAccessException
    *          if the <code>LockManager</code> concrete implementation's default constructor
    *          cannot be accessed.
    * @throws  java.lang.InstantiationException
    *          if the <code>LockManager</code> concrete implementation cannot be instantiated.
    * @throws  ClassNotFoundException
    *          if the <code>LockManager</code> concrete implementation class cannot be located.
    * @throws  IllegalArgumentException
    *          if lock metadata is used by the current application, but no lock table updater is
    *          found for the current database.
    */
   protected LockManager createLockManager()
   throws IllegalAccessException,
          InstantiationException,
          ClassNotFoundException
   {
      if (DatabaseManager.TEMP_TABLE_DB.equals(database))
      {
         return null;
      }
      
      // retrieve fully qualified name of LockManager implementation class and instantiate it
      String path = "database/" + database.getName() + "/p2j/lock_manager/class";
      String className = Utils.getDirectoryNodeString(null, path, DEFAULT_LOCK_MANAGER, false);
      LockManager manager = (LockManager) Class.forName(className).newInstance();
      manager.setDatabase(database);
      
      // register lock table updater with lock manager if the former exists
      if (database.isPrimary())
      {
         // if a lock updater is found for the database, register it as a listener with the
         // lock manager
         lockTableUpdater = DatabaseManager.getLockTableUpdater(database);
         if (lockTableUpdater != null)
         {
            manager.setLockListener(lockTableUpdater);
         }
      }
      
      if (LOG.isLoggable(Level.CONFIG))
      {
         LOG.config(message("lock manager initialized [" + className + "]"));
      }
      
      return manager;
   }
   
   /**
    * Create and initialize the identity manager to be used with this
    * instance of persistence services for the given database.  The identity
    * manager implementation to use is read from the directory. The default
    * if no entry is found is {@link #DEFAULT_IDENTITY_MANAGER}.
    *
    * @return  Identity manager instance or <code>null</code> if given
    *          database is temporary.
    *
    * @throws  IllegalAccessException
    *          if the <code>IdentityManager</code> concrete implementation's
    *          default constructor cannot be accessed.
    * @throws  java.lang.InstantiationException
    *          if the <code>IdentityManager</code> concrete implementation
    *          cannot be instantiated.
    * @throws  ClassNotFoundException
    *          if the <code>IdentityManager</code> concrete implementation
    *          class cannot be located.
    */
   protected IdentityManager createIdentityManager()
   throws ClassNotFoundException,
          IllegalAccessException,
          InstantiationException
   {
      if (DatabaseManager.TEMP_TABLE_DB.equals(database))
      {
         return null;
      }
      
      String className = DEFAULT_IDENTITY_MANAGER;
      boolean usePool = false;
      
      if (database.isPrimary())
      {
         // retrieve fully qualified name of IdentityManager implementation class and instantiate
         String dbName = database.getName();
         String path = "database/" + dbName + "/p2j/identity_manager/class";
         className = Utils.getDirectoryNodeString(null, path, DEFAULT_IDENTITY_MANAGER, false);
         
         // check whether use of identity pool is forced by directory
         path = String.format("database/%s/p2j/identity_manager/pool", dbName);
         usePool = Utils.getDirectoryNodeBoolean(null, path, false, false);
      }
      
      IdentityPoolManager mgr = (IdentityPoolManager) Class.forName(className).newInstance();
      
      mgr.setPersistence(this);
      
      IdentityPool identityPool = usePool ? new IdentityPool(database) : null;
      mgr.setIdentityPool(identityPool);
      
      if (LOG.isLoggable(Level.CONFIG))
      {
         LOG.config(message("identity manager initialized [" + className + "]"));
      }
      
      return mgr;
   }
   
   /**
    * Determine whether this persistence object is for a temporary table database.
    * 
    * @return  {@code true} if temporary, else {@code false}.
    */
   boolean isTemporary()
   {
      return temporary;
   }
   
   /**
    * Begin a new Hibernate transaction and store the instance in the local
    * context.  A new database session is created if one does not already
    * exist in the current context.
    * 
    * @param   local
    *          User context.
    * 
    * @return  <code>true</code> if a new transaction was begun;
    *          <code>false</code> if there was already an active transaction.
    * 
    * @throws  PersistenceException
    *          if there is an error beginning the new transaction.
    */
   boolean beginTransaction(Context local)
   throws PersistenceException
   {
      if (local.beginTransaction(true) == null)
      {
         return false;
      }
      
      if (LOG.isLoggable(Level.FINER))
      {
         LOG.log(Level.FINER, message("beginning transaction"));
      }
      
      return true;
   }
   
   /**
    * Commit the current Hibernate transaction and close the current session.
    * If the transaction cannot be committed, roll it back and raise an error.
    * 
    * @param   local
    *          User context.
    * 
    * @throws  IllegalStateException
    *          if no transaction is active in the current session;  this
    *          represents a programming error.
    * @throws  PersistenceException
    *          if there is an error committing the transaction.
    */
   void commit(Context local)
   throws PersistenceException
   {
      local.commit(true);
   }
   
   /**
    * Roll back the current Hibernate transaction and close the current
    * session.  If the transaction cannot be rolled back, raise an error.
    * 
    * @param   local
    *          User context.
    * 
    * @throws  IllegalStateException
    *          if no transaction is active in the current session;  this
    *          represents a programming error.
    * @throws  PersistenceException
    *          if there is an error rolling back the transaction.
    */
   void rollback(Context local)
   throws PersistenceException
   {
      try
      {
         local.rollback(true, false);
      }
      catch (IllegalStateException exc)
      {
         throw new PersistenceException(exc);
      }
   }
   
   /**
    * Execute an arbitrary SQL statement which does not produce a result set.
    * This method relies on Hibernate to obtain the JDBC connection, but
    * otherwise bypasses Hibernate and executes the SQL directly, using JDBC.
    * The Hibernate session is flushed before this statement is executed, to
    * ensure any pending changes are synchronized with the database first.
    *
    * @param   sql
    *          SQL statement text.
    * @param   args
    *          The statement arguments.
    * @param   entities
    *          Array of entity names involved in the SQL operation represented by this
    *          statement, or {@code null} to check whether a flush is required for
    *          <i>any</i> entity type.
    * @param   noFlush
    *          If {@code true}, flushing is bypassed.
    * 
    * @return  The number of entities updated or deleted, or {@code -1} if the statement
    *          did not complete normally.
    * 
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   int executeSQL(String sql, Object[] args, String[] entities, boolean noFlush)
   throws PersistenceException
   {
      long start = 0L;
      boolean trace = LOG.isLoggable(Level.FINEST);
      if (trace)
      {
         start = System.currentTimeMillis();
      }
      
      Context local = context.get();
      int rc = -1;
      
      try
      {
         local.pushImplicitTransaction();
         
         Session session = local.getSession();
         
         if (!noFlush && local.isFlushRequired(entities))
         {
            local.flushSession();
         }
         
         ReturningWork<Integer> work = (Connection) ->
         {
            try (PreparedStatement stmt = Connection.prepareStatement(sql))
            {
               if (args != null)
               {
                  int len = args.length;
                  for (int i = 0; i < len; i++)
                  {
                     stmt.setObject(i + 1, args[i]);
                  }
               }
               
               return stmt.executeUpdate();
            }
         };
         
         rc = session.doReturningWork(work);
         
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, "[" + sql + "] rc = " + rc);
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = "Error executing SQL statement:  " + sql;
         handleException(msg, exc);
      }
      finally
      {
         local.popImplicitTransaction();
         
         if (trace)
         {
            long elapsed = System.currentTimeMillis() - start;
            LOG.log(Level.FINEST, message("TIMING:SQL [" + elapsed + "] " + sql));
         }
      }
      
      return rc;
   }
   
   /**
    * Set enable mode for transaction tracing in the current context.
    * 
    * @param   enable
    *          <code>true</code> to enable; <code>false</code> to disable.
    */
   void enableTransactionTrace(boolean enable)
   {
      context.get().enableTransactionTrace(enable);
   }
   
   /**
    * Retrieve trace information about where the current transaction, if any, was started.
    * This mode must be enabled with {@link #enableTransactionTrace(boolean)}, or this method
    * always will return <code>null</code>.
    * 
    * @return  <code>Throwable</code> containing stack trace from last transaction start, or
    *          <code>null</code>.
    */
   Throwable getBeginTxTrace()
   {
      return context.get().getBeginTxTrace();
   }
   
   /**
    * If this object is associated with a metadata database, set this object's lock table updater
    * to be the same as that of the <code>Persistence</code> instance associated with the given,
    * primary database.
    * 
    * @param   database
    *          Primary database presumably associated with the metadata database represented by
    *          this object.
    * 
    * @throws  PersistenceException
    *          if this object is not associated with a metadata database or the given database is
    *          not a primary database.
    */
   void shareLockTableUpdater(Database database)
   throws PersistenceException
   {
      if (!this.database.isMeta() && database.isPrimary())
      {
         throw new PersistenceException("Invalid lock updater share request");
      }
      
      Persistence pPrim = PersistenceFactory.getInstance(database);
      this.lockTableUpdater = pPrim.lockTableUpdater;
   }
   
   /**
    * Get a JDBC connection from the current, context local session, creating
    * the session first if necessary.  The session associated with the
    * connection is remembered so that it can be cleaned up properly when the
    * connection is later checked back in.
    * <p>
    * Note:  the caller must not call <code>close()</code> on the returned
    * connection, since the connection may be "borrowed" from the current,
    * active session;  rather, {@link #checkInConnection(Connection)} must be
    * used.
    * 
    * @return  JDBC connection.
    * 
    * @throws  PersistenceException
    *          if a session cannot be created.
    * 
    * @see     #checkInConnection(Connection)
    */
   Connection checkOutConnection()
   throws PersistenceException
   {
      Context local = context.get();
      Session session = local.getSession();
      Connection connection = ((SessionImpl) session).connection();
      borrowedSessions.put(connection, session);
      
      return connection;
   }
   
   /**
    * Check in a JDBC connection previously obtained with {@link
    * #checkOutConnection()}.  This will close the connection if it is not
    * otherwise in use in the current, context local session.
    * 
    * @param   connection
    *          JDBC connection which is no longer needed by the caller.
    *          
    * @throws  PersistenceException
    *          if there is an error closing the session associated with the
    *          connection.
    */
   void checkInConnection(Connection connection)
   throws PersistenceException
   {
      Session session = borrowedSessions.get(connection);
      if (session != null)
      {
         Session localSession = context.get().getSessionNoCreate();
         if (session != localSession && session.isOpen())
         {
            try
            {
               session.close();
            }
            catch (HibernateException exc)
            {
               throw new PersistenceException(exc);
            }
         }
      }
   }
   
   /**
    * Get the context for the current user session.
    * <p>
    * NOTE: this method is package private and should remain so. It is meant to allow optimized
    * calls to certain frequently used methods of this class, in order to avoid the overhead of
    * an unnecessary context local lookup.
    * 
    * @return  User context.
    */
   Context getContext()
   {
      return context.get();
   }
   
   /**
    * Bind this instance to a backing, Hibernate session, in case it is not
    * currently backed by one.
    * <p>
    * This is the only method in this class which exposes the Hibernate
    * session to the outside world.  Because of this, this method is
    * intentionally package private.  Use with care.
    * 
    * @param   local
    *          User context.
    * 
    * @throws  PersistenceException
    *          if an error occurs creating a session.
    * 
    * @return  Hibernate session.
    */
   Session bind(Context local)
   throws PersistenceException
   {
      return local.getSession();
   }
   
   /**
    * Evict the given DMO from the current session immediately. If there is no existing session,
    * this method is a no-op.
    * 
    * @param   local
    *          Persistence context.
    * @param   dmo
    *          Data model object to be evicted.
    */
   void evict(Persistence.Context local, Persistable dmo)
   {
      Session session = local.getSessionNoCreate();
      if (session != null)
      {
         session.evict(dmo);
      }
   }
   
   /**
    * Evict a collection of DMOs from the current Hibernate session. If there is no existing
    * session, this method is a no-op.
    * 
    * @param   evictees
    *          Collection of DMOs to be evicted.
    */
   void evict(Collection<Persistable> evictees)
   {
      Context local = context.get();
      Session session = local.getSessionNoCreate();
      if (session != null)
      {
         for (Persistable dmo : evictees)
         {
            session.evict(dmo);
         }
      }
   }
   
   /**
    * Flush the current session (if any), if there are any pending changes.
    * If there is no active Hibernate session, or no pending changes, no
    * action is taken.
    * 
    * @throws  PersistenceException
    *          if there is any error flushing the session.
    */
   void maybeFlush()
   throws PersistenceException
   {
      Context local = context.get();
      Session session = local.getSessionNoCreate();
      if (session != null && local.isFlushRequired((String) null))
      {
         flush(local, session, false);
      }
   }
   
   /**
    * Register a <code>SessionListener</code> to receive session life
    * cycle event notifications.  The listener will be deregistered
    * according to the following rules at the earlier of:
    * <ul>
    *   <li>the point at which it is collected by the garbage collector
    *       (once business logic no longer strongly references it);  or
    *   <li>the end of the scope indicated by the <code>scope</code>
    *       parameter:
    *       <ul>
    *         <li><code>ENCLOSING_EXTERNAL_SCOPE</code>
    *         <li><code>CURRENT_SCOPE</code>
    *         <li><code>NEXT_EXTERNAL_SCOPE</code>
    *       </ul>
    * </ul>
    *
    * @param   listener
    *          Session listener to be registered.
    * @param   scope
    *          <code>SessionListener.Scope</code> enum indicating the scope to
    *          which this listener should be registered.  When the scope ends,
    *          the listener is deregistered as described above.
    */
   void registerSessionListener(SessionListener listener, SessionListener.Scope scope)
   {
      Context local = context.get();
      local.registerSessionListener(listener, scope);
   }
   
   /**
    * Deregister a session listener when it no longer needs to receive session life cycle events.
    * 
    * @param   listener
    *          Listener to deregister.
    */
   void deregisterSessionListener(SessionListener listener)
   {
      Context local = context.get();
      local.deregisterSessionListener(listener);
   }
   
   /**
    * Execute an HQL query and load a single record from the database or from
    * the session-level cache, in read-only mode.  The record, if present,
    * must be unique.  No locking is attempted.
    * <p>
    * This method is provided to support housekeeping queries internal to the
    * persistence framework.  It is expected that <code>hql</code> is the
    * minimal query necessary to return the primary key of the DMO which
    * matches the query criteria. The Hibernate session is then used to
    * create and return an uninitialized proxy instance of the DMO.  This
    * implementation is used to avoid the overhead of:
    * <ul>
    *   <li>retrieving all data for the matching record from the database;
    *   <li>hydrating the DMO with that data.
    * </ul>
    *
    * @param   dmoClass
    *          Expected return type;  the DMO implementation class being
    *          queried.
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    *
    * @return  A single DMO instance representing the query result, or
    *          <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query or if the query
    *          returned multiple results.
    */
   /*
   Persistable loadReadOnly(Class<?> dmoClass,
                            String hql,
                            Object[] values,
                            Type[] types)
   throws PersistenceException
   {
      Context local = context.get();
      Persistable result = null;
      
      try
      {
         Session session = local.getSession();
         Query query = getQuery(local, hql, 2, 0, values, types);
         query.setReadOnly(true);
         if (!local.isFlushRequired(dmoClass))
         {
            query.setFlushMode(FlushMode.NEVER);
         }
         
         Serializable id = (Serializable) query.uniqueResult();
         if (id == null)
         {
            return null;
         }
         
         result = (Persistable) session.load(dmoClass, id);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleInterrupt(exc);
         
         local.closeSession(true);
         
         String msg = message("error loading record");
         
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, msg + System.getProperty("line.separator") + hql);
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Details:", exc);
            }
         }
         
         throw new PersistenceException(msg, exc);
      }
      
      return result;
   }
   */
   
   /**
    * Process an error condition in which a unique result should have been
    * found, but multiple results occurred.
    *
    * @param   buffer
    *          Record buffer associated with the failed query.
    * @param   exc
    *          Optional root cause exception.
    *
    * @throws  PersistenceException
    *          always;  indicates the nature of the error.
    */
   void uniqueResultViolation(RecordBuffer buffer, Exception exc)
   throws PersistenceException
   {
      int errnum = 3166;
      buffer.setAmbiguous(true);
      
      String msg = ErrorManager.buildErrorText(errnum,
                                               "More than one " +
                                               buffer.getLegacyName() +
                                               " records found by a unique FIND",
                                               false);
      
      if (LOG.isLoggable(Level.WARNING))
      {
         LOG.log(Level.WARNING, message(msg));
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Details:", exc);
         }
      }
      
      if (exc != null)
      {
         throw new PersistenceException(msg, errnum, exc);
      }
      else
      {
         throw new PersistenceException(msg, errnum);
      }
   }
   
   /**
    * Flush the current session, which synchronizes the in-memory state of
    * DMOs to the database.  Optionally, the session is cleared as well.
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * <p>
    * It is expected that this method will only be called within an active
    * transaction.  If no session is active, this method will raise an
    * exception.
    *
    * @param   clear
    *          <code>true</code> to clear the session (used during bulk
    *          inserts or updates to prevent a progressive memory leak) after
    *          the flush.
    *
    * @throws  PersistenceException
    *          if no session currently is active or if an error occurs during
    *          session flush.
    */
   private void flush(boolean clear)
   throws PersistenceException
   {
      Context local = context.get();
      Session session = local.getSessionNoCreate();
      
      // Sanity.
      if (session == null)
      {
         throw new PersistenceException("No active session to flush");
      }
      
      flush(local, session, clear);
   }
   
   /**
    * Flush the current session, which synchronizes the in-memory state of
    * DMOs to the database.  Optionally, the session is cleared as well.
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * 
    * @param   local
    *          Context-local state.
    * @param   session
    *          Hibernate session
    * @param   clear
    *          <code>true</code> to clear the session (used during bulk
    *          inserts or updates to prevent a progressive memory leak) after
    *          the flush.
    *
    * @throws  PersistenceException
    *          if an error occurs during session flush.
    */
   private void flush(Context local, Session session, boolean clear)
   throws PersistenceException
   {
      try
      {
         local.flushSession();
         
         if (clear)
         {
            session.clear();
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = "error flushing " + (clear ? "and clearing " : "") + "session";
         handleException(msg, exc);
      }
   }
   
   /**
    * Perform special handling on query substitution parameters to avoid
    * errors during query execution.  This entails:
    * <ul>
    *   <li>Trimming all trailing spaces, tabs, line feeds, and carriage
    *       returns from any parameter of type <code>character</code>.  Any
    *       parameter affected is replaced with a new <code>character</code>
    *       instance at the same index position within <code>args</code>.
    *   <li>Clipping out of bounds date values to the range supported by the
    *       backing database's implementation of a date data type.  Any
    *       parameter affected is replaced with a new <code>date</code>
    *       instance at the same index position within <code>args</code>.
    * </ul>
    * <p>
    * Note:  this method has special processing for <code>null</code> bytes
    * embedded in <code>character</code> variables.  The text of the variable
    * is truncated starting at the first embedded <code>null</code> byte.
    * This means that the resulting substitution parameters are guaranteed
    * to never include <code>null</code> bytes.
    *
    * @param   args
    *          Array of query substitution parameters or <code>null</code> if
    *          none.
    */
   private void preprocessQueryParameters(Object[] args)
   {
      if (args != null)
      {
         int len = args.length;
         for (int i = 0; i < len; i++)
         {
            args[i] = preprocessQueryParameter(args[i], (P2JDialect) dialect);
         }
      }
   }
   
   /**
    * Determine whether all of the given DMO entity names represent read-only tables.
    * 
    * @param   entities
    *          DMO entity names to check, or {@code null}.
    * 
    * @return  {@code false} if {@code entities} is {@code null} or if any entity represents a
    *          non-read-only table, else {@code true}.
    */
   private boolean checkReadOnly(String[] entities)
   {
      if (entities == null || temporary)
      {
         return false;
      }
      
      for (String entity : entities)
      {
         if (!DatabaseManager.isReadOnly(entity))
         {
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Get or create a query for the given HQL query string. A cached query is used if it exists.
    *
    * @param   local
    *          Context-local state.
    * @param   hql
    *          HQL query string.
    * @param   maxResults
    *          The maximum number of results to be returned by the query.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   readOnly
    *          <code>true</code> to prepare the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  Hibernate query object.
    *
    * @throws  PersistenceException
    *          if there was an error creating a Hibernate session.
    */
   private Query getQuery(Context local,
                          String hql,
                          int maxResults,
                          int startOffset,
                          Object[] values,
                          Type[] types,
                          boolean readOnly)
   throws PersistenceException
   {
      Query query = local.getQuery(hql, maxResults, startOffset, readOnly);
      
      // Bind parameters, if any.  It is OK to do this with cached queries, because each use is
      // guaranteed to reset the parameters.
      if (values != null && values.length > 0)
      {
         preprocessQueryParameters(values);
         // query.setParameters(values, types);
         
         // This is an workaround for setting up named parameters. It was needed because 
         // Hibernate cannot handle multi-column values with positional parameters. 
         // Even this could work for all cases, the implementation was only tested for
         // simple cases CompositeCustomType that were needed for datetime-tz.
         // The code will be eliminated anyway when named parameters will replace 
         // positional parameters globally.
         int posIndex = 0;
         for (int k = 0; k < values.length; k++)
         {
            if (types[k] instanceof org.hibernate.type.CompositeCustomType)
            {
               query.setParameter("px" + (++posIndex), values[k], types[k]);
            }
            else
            {
               query.setParameter(Integer.toString(k), values[k], types[k]);
            }
         }
      }
      
      // allow second level cache to store query results
      query.setCacheable(!temporary);
      
      return query;
   }
   
   /**
    * Raise a STOP condition, if the given exception represents a catastrophic, JDBC failure;
    * else throw a <code>PersistenceException</code>.
    * 
    * @param   message
    *          Error text for the exception.
    * @param   exc
    *          A Hibernate exception which may or may not represent an unrecoverable database
    *          error.
    * 
    * @throws  StopConditionException
    *          if the error is unrecoverable.
    * @throws  PersistenceException
    *          if the error is likely recoverable.
    */
   private void handleException(String message, HibernateException exc)
   throws PersistenceException
   {
      if (exc instanceof JDBCException)   // TODO: too broad a category?
      {
         throw new StopConditionException(message(message), exc);
      }
      
      throw new PersistenceException(message(message), exc);
   }
   
   /**
    * A hash key for a unique combination of table and column name.
    */
   private static class ColumnKey
   {
      /** Table name */
      private final String table;
      
      /** Column name */
      private final String column;
      
      /**
       * Constructor.
       * 
       * @param   table
       *          Table name.
       * @param   column
       *          Column name.
       */
      ColumnKey(String table, String column)
      {
         this.table = table;
         this.column = column;
      }
      
      /**
       * Test for equivalence with another object.
       * 
       * @param   o
       *          Another <code>ColumnKey</code> instance.
       * 
       * @return  <code>true</code> if <code>o</code> has same table and
       *          column name as this object;  else <code>false</code>.
       */
      @Override
      public boolean equals(Object o)
      {
         if (!(o instanceof ColumnKey))
         {
            return false;
         }
         ColumnKey that = (ColumnKey) o;
         
         return this.table.equals(that.table) && this.column.equals(that.column);
      }
      
      /**
       * A hash code implementation consisent with our overridden {@link
       * #equals(Object)} method.
       * 
       * @return  Hash code.
       */
      @Override
      public int hashCode()
      {
         int result = 17;
         
         result = 37 * result + table.hashCode();
         result = 37 * result + column.hashCode();
         
         return result;
      }
      
      /**
       * Get string representation of this object.
       * 
       * @return  String representation.
       */
      @Override
      public String toString()
      {
         return table + ":" + column;
      }
   }
   
   /**
    * A hash key for a unique combination of HQL, maximum results, and starting row offset.
    */
   private static class QueryKey
   {
      /** HQL statement */
      private final String hql;
      
      /** Maximum results */
      private final int maxResults;
      
      /** Starting row offset */
      private final int startOffset;
      
      /** Precalculated hash code */
      private final int hash;
      
      /**
       * Constructor.
       * 
       * @param   hql
       *          HQL statement
       * @param   maxResults
       *          Maximum results.
       * @param   startOffset
       *          Starting row offset.
       */
      QueryKey(String hql, int maxResults, int startOffset)
      {
         this.hql = hql;
         this.maxResults = maxResults;
         this.startOffset = startOffset;
         
         int result = 17;
         result = 37 * result + hql.hashCode();
         result = 37 * result + maxResults;
         result = 37 * result + startOffset;
         this.hash = result;
      }
      
      /**
       * Test for equivalence with another object.
       * 
       * @param   o
       *          Another <code>ColumnKey</code> instance.
       * 
       * @return  <code>true</code> if <code>o</code> has same internal
       *          state as this object;  else <code>false</code>.
       */
      @Override
      public boolean equals(Object o)
      {
         if (!(o instanceof QueryKey))
         {
            return false;
         }
         QueryKey that = (QueryKey) o;
         
         return this.hash == that.hash               &&
                this.maxResults == that.maxResults   &&
                this.startOffset == that.startOffset &&
                this.hql.equals(that.hql);
      }
      
      /**
       * A hash code implementation consistent with our overridden {@link #equals(Object)} method.
       * 
       * @return  Hash code.
       */
      @Override
      public int hashCode()
      {
         return hash;
      }
      
      /**
       * Get string representation of this object.
       * 
       * @return  String representation.
       */
      @Override
      public String toString()
      {
         return hql + "[" + startOffset + ":" + maxResults + "]";
      }
   }
   
   /**
    * A storage and work area for context-local variables used by the
    * enclosing class.  One instance of this class is created for each
    * client context.  It is used to:
    * <ul>
    *   <li>create and cache Hibernate query objects;
    *   <li>create, store and close the active Hibernate session;
    *   <li>manage the active Hibernate transaction.
    * </ul>
    * <p>
    * Special care is taken to not close a Hibernate session while other
    * entities are dependent upon it for services.
    */
   class Context
   {
      /** Context-local buffer manager */
      private final BufferManager bufferManager = BufferManager.get();
      
      /** Context-local record lock context */
      private final RecordLockContext lockContext =
         RecordLockContext.get(Persistence.this, bufferManager);
      
      /** Context-local ChangeBroker */
      private final ChangeBroker changeBroker = ChangeBroker.get();
      
      /** Session listener objects */
      private final WeakHashMap<SessionListener, Finalizable> sessionListeners =
         new WeakHashMap<>();
      
      /** Cache of HQL strings, max results, and start offsets to Hibernate queries */
      private ExpiryCache<QueryKey, Query> queryCache = null;
      
      /** Active Hibernate session, if any */
      private Session session = null;
      
      /** Active Hibernate transaction, if any */
      private Transaction transaction = null;
      
      /** Implicit transaction use count */
      private int implicitTxUseCount = 0;
      
      /** Flag to prevent nested session close attempts */
      private boolean closingSession = false;
      
      /** Count of open Hibernate sessions in this context */
      private volatile int openSessionCount = 0;
      
      /** Optional, dialect-specific, context-local data */
      private Object dialectData = ((P2JDialect) dialect).createContext(Persistence.this);
      
      /** Is current transaction explicit? */
      private boolean explicitTx = false;
      
      /** rereadnolock command-line parameter, by default deactivated */
      private boolean rereadNoLock = false;
      
      /** Enable tracing transaction starts */
      private boolean traceTx = false;
      
      /** Throwable which holds stack trace where last transaction began */
      private Throwable beginTxTrace = null;
      
      /** The default constructor reads rereadNoLock from directory. */
      private Context()
      {
         // the '-rereadnolock' startup parameter was introduced in Progress 8.3B'
         // only needed for permanent tables, not temp-tables
         if (!temporary)
         {
            Directory dir = DirectoryManager.getInstance();
            rereadNoLock = dir.getBoolean(Directory.ID_RELATIVE, "rereadnolock", false);
         }
      }
      
      /**
       * Get the record lock context associated with the current client
       * session.
       *
       * @return  Record lock context for the current session.
       */
      RecordLockContext getRecordLockContext()
      {
         return lockContext;
      }
      
      /**
       * Get a cached query for the given combination of HQL query string, maximum result rows,
       * and starting row offset, creating and caching a new query if no such query already is
       * cached. The cache is fixed size. When it is full and a new query needs to be added, the
       * least recently used query is removed.
       * 
       * @param   hql
       *          HQL query string.
       * @param   maxResults
       *          The maximum number of results to be returned.
       * @param   startOffset
       *          Zero-based offset of first result to be returned (undefined results unless
       *          <code>hql</code> includes an <code>order by</code> phrase).
       * @param   readOnly
       *          <code>true</code> to execute the query in read-only mode, else
       *          <code>false</code>.
       * 
       * @return  Hibernate query object.
       * 
       * @throws  PersistenceException
       *          if there was an error creating a Hibernate session.
       */
      Query getQuery(String hql, int maxResults, int startOffset, boolean readOnly)
      throws PersistenceException
      {
         QueryKey key = new QueryKey(hql, maxResults, startOffset);
         Query query = (queryCache != null ? queryCache.get(key) : null);
         if (query == null)
         {
            Session session = getSession();
            query = session.createQuery(hql);
            
            if (maxResults > 0)
            {
               query.setMaxResults(maxResults);
            }
            else
            {
               // negative value indicates no limit
               query.setMaxResults(-1);
            }
            
            if (startOffset > 0)
            {
               query.setFirstResult(startOffset);
            }
            else
            {
               query.setFirstResult(0);
            }
            
            queryCache.put(key, query);
         }
         else
         {
            query.setFlushMode(FlushMode.AUTO);
         }
         
         query.setReadOnly(readOnly);
         
         return query;
      }
      
      /**
       * Commit the current Hibernate transaction and possibly close the
       * current session.  If the transaction cannot be committed, roll it
       * back and raise an error.
       *
       * @param   explicit
       *          <code>true</code> if the current transaction is expected to
       *          be explicit, <code>false</code> if implicit.
       *
       * @throws  IllegalStateException
       *          if no transaction is active in the current session;  this
       *          represents a programming error.
       * @throws  PersistenceException
       *          if there is an error committing the transaction.
       */
      void commit(boolean explicit)
      throws PersistenceException
      {
         if (transaction == null)
         {
            throw new IllegalStateException(
               message("no current transaction available to commit"));
         }
         
         try
         {
            try
            {
               // increment use count so SessionListeners do not commit implicit transaction
               implicitTxUseCount++;
               
               if (implicitTxUseCount > 1 && !sessionListeners.isEmpty())
               {
                  notifySessionEvent(SessionListener.Event.TRANSACTION_COMMITTING);
               }
            }
            finally
            {
               // Decrement use count back to balance increment above.
               implicitTxUseCount--;
               
               transaction.commit();
            }
            
            if (LOG.isLoggable(Level.FINER))
            {
               LOG.log(Level.FINER, message((explicit ? "explicit" : "implicit") +
                                            " transaction committed"));
            }
            
            endTransaction(explicit, false, false);
         }
         catch (HibernateException exc)
         {
            DBUtils.handleException(database, exc);
            
            try
            {
               rollback(explicit, true);
            }
            catch (PersistenceException exc1)
            {
               // Don't need to throw an exception here;  this would hide the
               // commit exception, so just log it.
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.log(Level.SEVERE, message("rollback error"), exc1);
               }
            }
            
            String msg = "unable to commit " +
                         (explicit ? "explicit" : "implicit") +
                         " transaction";
            handleException(msg, exc);
         }
      }
      
      /**
       * Roll back the current Hibernate transaction and possibly close the
       * current session.  If the transaction cannot be rolled back, raise an
       * error.
       *
       * @param   explicit
       *          <code>true</code> if the current transaction is expected to
       *          be explicit, <code>false</code> if implicit.
       * @param   error
       *          Indicates whether the rollback is due to an error.
       *
       * @throws  IllegalStateException
       *          if no transaction is active in the current session;  this
       *          represents a programming error.
       * @throws  PersistenceException
       *          if there is an error rolling back the transaction.
       */
      void rollback(boolean explicit, boolean error)
      throws PersistenceException
      {
         if (transaction == null)
         {
            throw new IllegalStateException(
               message("no current transaction available to rollback"));
         }
         
         try
         {
            try
            {
               if (!error)
               {
                  maybeFlushSession();
               }
            }
            finally
            {
               transaction.rollback();
               
               if (LOG.isLoggable(Level.FINER))
               {
                  LOG.log(Level.FINER, message("transaction rolled back"));
               }
            }
         }
         catch (HibernateException exc)
         {
            DBUtils.handleException(database, exc);
            
            error = true;
            closeSession(true);
            
            handleException("unable to rollback current transaction", exc);
         }
         finally
         {
            endTransaction(explicit, error, true);
         }
      }
      
      /**
       * Set enable mode for transaction tracing.
       * 
       * @param   enable
       *          <code>true</code> to enable; <code>false</code> to disable.
       */
      void enableTransactionTrace(boolean enable)
      {
         this.traceTx = enable;
      }
      
      /**
       * Begin a new Hibernate transaction and store the instance for future
       * use.  If an explicit transaction already exists, it is returned, and
       * a new one is not created.  If an implicit transaction already exists,
       * and a new, explicit transaction is being requested, the implicit
       * transaction is first committed, then the new, explicit transaction is
       * begun.  A new database session is created if one does not already
       * exist in the current context.
       *
       * @param   explicit
       *          <code>true</code> if an explicit transaction is being
       *          requested;  <code>false</code> if the request is for an
       *          implicit transaction.
       *
       * @return  The active transaction, whether an existing explicit
       *          transaction, or a new implicit or explicit transaction.
       *
       * @throws  PersistenceException
       *          if there is an error beginning the new transaction or
       *          committing an existing one.
       */
      Transaction beginTransaction(boolean explicit)
      throws PersistenceException
      {
         if (transaction != null)
         {
            if (explicitTx)
            {
               // already an explicit transaction active
               return null;
            }
            
            // an active implicit transaction must be committed before we can begin a new
            // explicit transaction; if requested transaction also is implicit, disallow it
            if (transaction.isActive())
            {
               if (explicit)
               {
                  commit(false);
               }
               else
               {
                  return null;
               }
            }
         }
         
         try
         {
            Session session = getSession();
            transaction = session.beginTransaction();
            
            if (traceTx)
            {
               beginTxTrace = new Throwable();
            }
         }
         catch (HibernateException exc)
         {
            DBUtils.handleException(database, exc);
            
            closeSession(true);
            
            // TODO:  use proper error message/number
            handleException("error beginning database transaction", exc);
         }
         finally
         {
            if (explicit && transaction != null)
            {
               explicitTx = true;
            }
         }
         
         return transaction;
      }
      
      /**
       * Discard the current transaction, which was previously committed or
       * rolled back.
       * <p>
       * For an implicit transaction, the currently scheduled, implicit commit
       * task is cancelled and nulled out.  In case of an error, the current
       * session is marked broken, so it will be closed and replaced upon the
       * next attempt to use it on the user sesion thread (the current thread
       * must be the implicit commit timer if we get here for an implicit
       * transaction).
       * <p>
       * For an explicit transaction, the current session is closed and the
       * active dialect is allowed to do some post-transaction processing.
       * Also, the implicit transaction use count is reset.
       * 
       * @param   explicit
       *          Indicates whether the transaction being ended is explicit
       *          or implicit.
       * @param   error
       *          Indicates whether the transaction ended with an error.
       * @param   rollback
       *          Indicates whether the transaction is being ended after a
       *          rollback (as opposed to a commit).
       * 
       * @throws  PersistenceException
       *          if an error occurs closing the session or performing
       *          post-transaction processing.
       */
      void endTransaction(boolean explicit, boolean error, boolean rollback)
      throws PersistenceException
      {
         try
         {
            try
            {
               if (explicit)
               {
                  closeSession(error);
               }
            }
            finally
            {
               transaction = null;
            }
            
            // Give the dialect a chance to perform post-transaction processing.
            P2JDialect p2jDialect = (P2JDialect) dialect;
            p2jDialect.postTransaction(Persistence.this, dialectData, rollback);
            
            // If an implicit transaction previously was interrupted to begin
            // an explicit transaction, begin a new implicit transaction now.
            if (explicitTx && implicitTxUseCount > 0)
            {
               beginTransaction(false);
            }
         }
         finally
         {
            explicitTx = false;
            beginTxTrace = null;
         }
      }
      
      /**
       * Retrieve trace information about where the current transaction, if any, was started.
       * FINER level logging must be enabled, or this method always will return <code>null</code>.
       * 
       * @return  <code>Throwable</code> containing stack trace from last transaction start, or
       *          <code>null</code>.
       */
      Throwable getBeginTxTrace()
      {
         return beginTxTrace;
      }
      
      /**
       * Increment the implicit transaction use counter and begin an implicit
       * transaction if none is currently active, and the use count currently
       * is at zero.
       * 
       * @return  The implicit transaction use count after incrementing it.
       * 
       * @throws  PersistenceException
       *          if there is an error beginning a new transaction.
       */
      int pushImplicitTransaction()
      throws PersistenceException
      {
         if (implicitTxUseCount >= 0)
         {
            if (implicitTxUseCount == 0 && transaction == null)
            {
               beginTransaction(false);
            }
            
            implicitTxUseCount++;
         }
         
         return implicitTxUseCount;
      }
      
      /**
       * Decrement the implicit transaction use counter.  If it reaches zero
       * as a result, remember the current system time, so that the implicit
       * transaction commit timer can commit the current, implicit transaction
       * when necessary.
       * 
       * @return  The implicit transaction use count after decrementing it.
       */
      int popImplicitTransaction()
      {
         if (implicitTxUseCount > 0)
         {
            implicitTxUseCount--;
         }
         
         if (!explicitTx &&
             implicitTxUseCount == 0 &&
             transaction != null &&
             transaction.isActive())
         {
            try
            {
               if (LOG.isLoggable(Level.FINER))
               {
                  LOG.log(Level.FINER, "Performing implicit transaction commit");
               }
               
               commit(false);
            }
            catch (PersistenceException exc)
            {
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.log(Level.SEVERE, "Error committing implicit transaction", exc);
               }
            }
         }
         
         return implicitTxUseCount;
      }
      
      /**
       * Determine whether a session flush is required for the given entity types. A flush is
       * required if flushing currently is enabled and the {@link ChangeBroker} has tracked any
       * change to a DMO of any of the given types since the last flush.
       *
       * @param   entities
       *          Array of entity names, or {@code null} to check whether a flush is required for
       *          <i>any</i> entity type.
       *
       * @return  {@code true} if a flush is required, else {@code false}.
       */
      boolean isFlushRequired(String[] entities)
      {
         if (entities == null)
         {
            return changeBroker.requiresFlush(Persistence.this, (String) null);
         }
         
         for (String entity : entities)
         {
            if (changeBroker.requiresFlush(Persistence.this, entity))
            {
               return true;
            }
         }
         
         return false;
      }
      
      /**
       * Determine whether a session flush is required for the given entity type. A flush is
       * required if flushing currently is enabled and the {@link ChangeBroker} has tracked any
       * change to a DMO of the given type since the last flush.
       *
       * @param   entity
       *          Name of an entity, or {@code null} to check whether a flush is required for
       *          <i>any</i> entity type.
       *
       * @return  {@code true} if a flush is required, else {@code false}.
       */
      boolean isFlushRequired(String entity)
      {
         return changeBroker.requiresFlush(Persistence.this, entity);
      }
      
      /**
       * Register a <code>SessionListener</code> to receive session life
       * cycle event notifications.  The listener will be deregistered
       * according to the following rules at the earlier of:
       * <ul>
       *   <li>the point at which it is collected by the garbage collector
       *       (once business logic no longer strongly references it);  or
       *   <li>the end of the scope indicated by the <code>scope</code>
       *       parameter:
       *       <ul>
       *         <li><code>ENCLOSING_EXTERNAL_SCOPE</code>
       *         <li><code>CURRENT_SCOPE</code>
       *         <li><code>NEXT_EXTERNAL_SCOPE</code>
       *       </ul>
       * </ul>
       *
       * @param   listener
       *          Session listener to be registered.
       * @param   scope
       *          <code>SessionListener.Scope</code> enum indicating the scope
       *          to which this listener should be registered.  When the
       *          scope ends, the listener is deregistered as described above.
       */
      void registerSessionListener(final SessionListener listener, SessionListener.Scope scope)
      {
         if (SessionListener.Scope.NONE.equals(scope))
         {
            // Store listener as a key in the map, so it may be released when no
            // longer referenced elsewhere.
            sessionListeners.put(listener, null);
            
            // Never deregister the listener by scope.
            return;
         }
         
         // Finalizable which will remove listener at a safe time.  The
         // listener may be removed earlier if not referenced in business
         // logic by the time this Finalizable executes its finished method,
         // and the GC collects it.  In the latter case, the listener will not
         // be notified, but it doesn't matter, since it is no longer
         // referenced by business logic at that point.
         final Finalizable cleaner = new Finalizable()
         {
            public void finished()
            {
               sessionListeners.remove(listener);
               listener.deregisteredSessionListener();
            }
            public void deleted() {}
            public void iterate() {}
            public void retry() {}
         };
         
         // store the finalizable so we can deregister it explicitly if needed
         sessionListeners.put(listener, cleaner);
         
         switch (scope)
         {
            case GLOBAL:
               TransactionManager.registerFinalizable(cleaner, true);
               break;
            case ENCLOSING_EXTERNAL:
               TransactionManager.registerTopLevelFinalizable(cleaner, true);
               break;
            case CURRENT:
               TransactionManager.registerFinalizable(cleaner, false);
               break;
            case NEXT_EXTERNAL:
               TransactionManager.registerNextExternal(cleaner);
               break;
         }
      }
      
      /**
       * Explicitly deregister a global session listener, which represents a dynamic temp-table
       * resource which has been deleted explicitly.
       * 
       * @param   listener
       *          Listener to deregister.
       */
      void deregisterSessionListener(SessionListener listener)
      {
         sessionListeners.remove(listener);
      }
      
      /**
       * Retrieve the Hibernate <code>Session</code> object associated with
       * this context, if any.
       * <p>
       * If the current session is broken, return <code>null</code>.
       *
       * @return  The session object associated with the current context, or
       *          <code>null</code> if none has been created.
       */
      Session getSessionNoCreate()
      {
         return session;
      }
      
      /**
       * Retrieve the Hibernate <code>Session</code> object associated with
       * this context, creating it if necessary.  The session can be closed
       * by {@link #closeSession}.
       *
       * @return  The session object associated with the current context.
       *
       * @throws  PersistenceException
       *          if an error occurs associating a detached record with the
       *          new session.
       * @throws  IllegalArgumentException
       *          if <code>database</code> has not been registered with the
       *          persistence service.
       */
      Session getSession()
      throws PersistenceException
      {
         if (session == null)
         {
            SessionFactory factory = DatabaseManager.getSessionFactory(database);
            if (factory == null)
            {
               throw new IllegalArgumentException(
                  message("database not registered with persistence service"));
            }
            
            if (openSessionCount > 0 && LOG.isLoggable(Level.WARNING))
            {
               LOG.warning("Opening new Hibernate session with " +
                           openSessionCount +
                           " already open!");
            }
            
            Interceptor interceptor = changeBroker.getInterceptor(Persistence.this);
            session = factory.withOptions().interceptor(interceptor).openSession();
            openSessionCount++;
            
            // a new session requires a new query cache
            queryCache = new LRUCache<>(1024);
            
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE,
                       message("opened Hibernate session; " + openSessionCount + " open"));
               /*
               if (LOG.isLoggable(Level.FINEST))
               {
                  LOG.log(Level.FINEST, "Stack trace:  ", new Throwable());
               }
               */
            }
            
            // Associate all detached records in the current context with the new session.
            Iterator<RecordBuffer> iter = bufferManager.activeBuffers(Persistence.this);
            RecordBuffer buffer = null;
            Persistable record = null;
            try
            {
               boolean finer = LOG.isLoggable(Level.FINER);
               while (iter.hasNext())
               {
                  buffer = iter.next();
                  record = buffer.getCurrentRecord();
                  
                  if (finer)
                  {
                     LOG.log(Level.FINER, message("associating record with session:  " +
                                                  record.getClass().getName() + "#" +
                                                  record.getId()));
                  }
                  
                  session.buildLockRequest(LockOptions.NONE).lock(record);
               }
            }
            catch (HibernateException exc)
            {
               DBUtils.handleException(database, exc);
               
               closeSession(true);
               
               String msg = "Error reassociating detached DMO with new session";
               if (record != null)
               {
                  msg += (" [" + record.getClass().getName() + "#" + record.getId() + "]");
               }
               
               LOG.log(Level.SEVERE, message(msg), exc);
               
               handleException(msg, exc);
            }
         }
         
         return session;
      }
      
      /**
       * Indicate whether a database-level transaction currently is open.
       * 
       * @param   explicit
       *          <code>true</code> to limit the check to an explicit
       *          transaction;  <code>false</code> to check for either an
       *          implicit or explicit transaction.
       *
       * @return  <code>true</code> if open, else <code>false</code>.
       */
      boolean isTransactionOpen(boolean explicit)
      {
         return ((!explicit || explicitTx) && transaction != null && transaction.isActive());
      }
      
      /**
       * Close the Hibernate <code>Session</code> object associated with this
       * context.
       * <p>
       * This implementation is tolerant of multiple closes on the same
       * session. If the session was already closed or could not be found,
       * the method simply returns.
       *
       * @param   error
       *          <code>true</code> to indicate the session is closing due to
       *          an error;  <code>false</code> to indicate the session is
       *          closing in the course of normal processing.
       * 
       * @throws  PersistenceException
       *          if an open session was found but the attempt to close it
       *          failed.
       */
      void closeSession(boolean error)
      throws PersistenceException
      {
         // Prevent nested close attempts.
         if (closingSession)
         {
            return;
         }
         
         try
         {
            closingSession = true;
            
            if (session != null && session.isOpen() && isTransactionOpen(false))
            {
               if (LOG.isLoggable(Level.WARNING))
               {
                  LOG.log(Level.WARNING, message("rolling back orphaned transaction"));
               }
               rollback(explicitTx, error);
            }
            
            if (!error)
            {
               // If session is not being closed due to an error, attempt a flush to ensure
               // consistency in case we just ended a transaction with a rollback.
               maybeFlushSession();
               
               if (temporary && TemporaryBuffer.getTableCount() > 0)
               {
                  // allow unused tables to be cleaned up
                  if (!sessionListeners.isEmpty())
                  {
                     notifySessionEvent(SessionListener.Event.SESSION_CLEANUP);
                  }
                  
                  // Special treatment for temp table sessions.  These must
                  // remain open as long as any table exists, otherwise the
                  // table and its data will be lost.
                  return;
               }
            }
            
            closeSessionImpl(error);
         }
         finally
         {
            closingSession = false;
         }
      }
      
      /**
       * Get the context-local, dialect-specific data object, if any.
       * 
       * @return  Dialect specific data object, or <code>null</code> if no
       *          such data exists.
       */
      Object getDialectData()
      {
         return dialectData;
      }
      
      /**
       * Indicate whether auto-commit is needed; this is the case for temp-tables being modified
       * outside of an explicit transaction.
       * 
       * @return  {@code true} if auto-commit is needed, else {@code false}.
       */
      boolean isAutoCommit()
      {
         return temporary && (transaction == null || !explicitTx);
      }
      
      /**
       * If an open session exists and the change broker indicates a flush is required, force a
       * session flush now. Bracket the flush with an implicit transaction.
       *
       * @throws  PersistenceException
       *          if an error occurs during session flush.
       */
      void flushSession()
      throws PersistenceException
      {
         // wrap this flush in a transaction if necessary (temp tables only)
         boolean needCommit = isAutoCommit();
         
         if (needCommit)
         {
            beginTransaction(true);
         }
         
         try
         {
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.fine(message("Explicit session flush"));
            }
            maybeFlushSession();
         }
         finally
         {
            if (needCommit)
            {
               commit(true);
            }
         }
      }
      
      /**
       * If an open session exists and the change broker indicates a flush is required, force a
       * session flush now.
       *
       * @throws  PersistenceException
       *          if an error occurs during session flush.
       */
      private void maybeFlushSession()
      throws PersistenceException
      {
         if (session == null || !session.isOpen() || !isFlushRequired((String) null))
         {
            return;
         }
         
         try
         {
            session.flush();
         }
         catch (HibernateException exc)
         {
            DBUtils.handleException(database, exc);
            
            closeSession(true);
            
            handleException("error flushing session", exc);
         }
      }
      
      /**
       * Close the Hibernate <code>Session</code> object associated with this
       * context.
       * <p>
       * This implementation is tolerant of multiple closes on the same
       * session. If the session was already closed or could not be found,
       * the method simply returns.
       * <p>
       * All registered session listeners are notified before the session is
       * closed.
       *
       * @param   error
       *          <code>true</code> to indicate the session is closing due to
       *          an error;  <code>false</code> to indicate the session is
       *          closing in the course of normal processing.
       *
       * @throws  PersistenceException
       *          if an open session was found but the attempt to close it
       *          failed, or if any error occurs during a registered session
       *          listener's session closing processing.
       */
      private void closeSessionImpl(boolean error)
      throws PersistenceException
      {
         // all cached queries are dependent upon the current session, so when that goes away,
         // the cache is no longer relevant
         queryCache = null;
         
         if (session != null)
         {
            try
            {
               if (session.isOpen())
               {
                  try
                  {
                     if (!sessionListeners.isEmpty())
                     {
                        SessionListener.Event event =
                           (error
                            ? SessionListener.Event.SESSION_CLOSING_WITH_ERROR
                            : SessionListener.Event.SESSION_CLOSING_NORMALLY);
                        notifySessionEvent(event);
                     }
                  }
                  finally
                  {
                     session.close();
                  }
               }
            }
            catch (HibernateException exc)
            {
               DBUtils.handleException(database, exc);
               
               handleException("error closing Hibernate session", exc);
            }
            finally
            {
               session = null;
               openSessionCount--;
               
               if (LOG.isLoggable(Level.FINE))
               {
                  LOG.log(Level.FINE,
                          message("closed Hibernate session; " + openSessionCount + " open"));
                  /*
                  if (LOG.isLoggable(Level.FINEST))
                  {
                     LOG.log(Level.FINEST, "Stack trace:  ", new Throwable());
                  }
                  */
               }
            }
         }
      }
      
      /**
       * Notify all registered session listeners that the current Hibernate session is about to
       * close or the current, implicit transaction is being committed.  Listeners which were set
       * to be deregistered upon an event are deregistered after they are notified.  There is no
       * guaranteed order in which listeners are notified.
       * <p>
       * All registered listeners are given a chance to process this event, even if one or more
       * raise a <code>PersistenceException</code> (there is no such guarantee for unchecked
       * exceptions). Each exception caught is logged.
       *
       * @param   event
       *          Session event type.
       */
      private void notifySessionEvent(SessionListener.Event event)
      {
         Iterator<SessionListener> iter = sessionListeners.keySet().iterator();
         while (iter.hasNext())
         {
            SessionListener listener = iter.next();
            boolean deregister = false;
            
            try
            {
               // allow listener to process session event
               deregister = listener.sessionEvent(event);
            }
            catch (PersistenceException exc)
            {
               // deregister listener if there was an error
               deregister = true;
               
               LOG.log(Level.SEVERE, "Error processing session event", exc);
            }
            finally
            {
               // remove listeners if there was an error, or if the listener indicated it should
               // be removed in response to the session closing notification
               if (deregister)
               {
                  iter.remove();
                  listener.deregisteredSessionListener();
               }
            }
         }
      }
      
      /**
       * Cleanup hook invoked at the end of the current client context.  We
       * close the currently open Hibernate session, if any.  This is the
       * only context local object which pins memory outside of this class
       * (within the Hibernate framework), so it requires explicit closure.
       * All other variables are garbage collected naturally when the context
       * ends.
       */
      void cleanup()
      {
         if (session != null && session.isOpen())
         {
            if (isTransactionOpen(false))
            {
               // Under normal circumstances, there should not be any tx open
               // at the context's end of life, but just in case...
               
               Level level = Level.FINER;
               String txType = "implicit";
               if (explicitTx)
               {
                  level = Level.WARNING;
                  txType = "explicit";
               }
               if (LOG.isLoggable(level))
               {
                  LOG.log(level, message("rolling back open, " + txType + 
                                         " transaction during context cleanup"));
               }
               
               try
               {
                  rollback(explicitTx, false);
               }
               catch (PersistenceException exc)
               {
                  LOG.log(Level.SEVERE,
                          "Error rolling back open transaction during context cleanup",
                          exc);
               }
            }
            
            try
            {
               closeSession(false);
            }
            catch (PersistenceException exc)
            {
               LOG.log(Level.SEVERE,
                       "Error closing Hibernate session during context cleanup",
                       exc);
            }
         }
      }
      
      /**
       * Checks if the rereadnolock 4GL command-line option is active for this client context.
       *
       * @return  {@code true} if rereadnolock is specified for this client context
       */
      boolean isRereadNoLock()
      {
         return rereadNoLock;
      }
   }
}