/*
** Module   : Reversible.java
** Abstract : Abstract base class for reversible database actions
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description------------------------------
** 001 ECF 20070130   @32042 Created initial version. Moved implementation
**                           from inner class of RecordBuffer to a top
**                           level class.
** 002 ECF 20071221   @36547 Modified constructor. Throw NPE if dmo
**                           parameter is null.
** 003 CA  20080604   @38538 Allow the DMO to be changed (when a DELETE or
**                           CREATE action is passed to the parent block,
**                           the DMO must be set to the chain's DMO).
**                           Added new c'tor variant.
** 004 CA  20080606   @38598 Added duplicate() method.
** 005 ECF 20080628   @39071 Track whether DMO was newly created.
** 006 ECF 20080925   @39944 Refactored to reduce memory use. Some
**                           functionality moved to AbstractReversible and
**                           concrete subclasses.
** 007 ECF 20081020   @40210 Added methods necessary to fix rollback
**                           problems at the full transaction level.
** 008 CA  20091202   @44468 Removed the newlyCreated c'tor parameter - the 
**                           flag will be set on rollback, depending on the 
**                           existence of Create operations; currently is set
**                           only for undoable tables.
** 009 ECF 20140305          Replaced Apache commons logging with J2SE logging.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.logging.*;
import com.goldencode.p2j.util.LogHelper;

/**
 * Abstract base class of various types of database actions that must be
 * reversible.  Invocation of the {@link #rollback(boolean, boolean)} method
 * reverses a specific database action on that record.  The action taken to
 * roll back the original action depends upon the subclass' implementation of
 * the <code>rollback</code> method.
 */
abstract class Reversible
{
   /** Logger */
   protected static final Logger LOG = LogHelper.getLogger(Reversible.class);
   
   /** Was DMO record newly created at the time Reversible was created? */
   private boolean newlyCreated;
   
   /**
    * Get a string representation of the internal state of this object,
    * primarily for debugging purposes.
    *
    * @return  String representing this object's state.
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder(getAction());
      buf.append(" of ");
      buf.append(getDMOName());
      buf.append(" [ID ");
      buf.append(getId());
      buf.append("]");
      
      return buf.toString();
   }
   
   /**
    * Get a string which briefly describes the action which this object
    * will reverse upon rollback.
    * 
    * @return  Action string.
    */
   protected String getAction()
   {
      return null;
   }
   
   /**
    * Indicate whether this <code>Reversible</code> implementation stores its
    * DMO locally, as opposed to storing an identifier and having to retrieve
    * the DMO from the backing database, or not making a DMO available at all.
    * 
    * @return  <code>true</code> if the DMO is stored locally, else
    *          <code>false</code>.
    */
   protected abstract boolean storesDMO();
   
   /**
    * Get the unqualified interface name of the DMO being managed by this
    * reversible.
    * 
    * @return  DMO interface name.
    */
   protected abstract String getDMOName();
   
   /**
    * Roll back the database change represented by this object.
    * <p>
    * Subclasses must implement this method in a manner appropriate to
    * the database actions they represent.
    * 
    * @param   rollForward
    *          If <code>true</code>, this method is being invoked in order to
    *          re-roll forward a non-undoable change, after a database-level
    *          rollback.  This indicates the modification is made within a
    *          new transaction invoked after the database-level rollback.
    *          If <code>false</code>, this method is being invoked in order to
    *          roll back an undoable change within the context of an
    *          application-level, full transaction or sub-transaction
    *          rollback.
    * @param   transaction
    *          <code>true</code> if this rollback occurs at a full transaction
    *          boundary, else <code>false</code>.
    *
    * @throws  PersistenceException
    *          if an error occurs at the database or in the persistence
    *          layer when reversing the action.
    */
   protected abstract void rollback(boolean rollForward, boolean transaction)
   throws PersistenceException;
   
   /**
    * Get the underlying DMO.
    *
    * @return  DMO.
    */
   protected abstract Persistable getDMO();
   
   /**
    * Set the underlying DMO for this reversible action.
    * 
    * @param   dmo
    *          The underlying DMO; may not be null.
    * 
    * @throws  NullPointerException
    *          if <code>dmo</code> is <code>null</code>.
    */
   protected abstract void setDMO(Persistable dmo);
   
   /**
    * Get the ID of the underlying DMO.
    *
    * @return  DMO ID.
    */
   protected abstract Serializable getId();
   
   /**
    * Prepare for this reversible to be rolled back.  Generally, this will
    * involve readying the DMO which will be acted upon by the rollback
    * actions.
    * 
    * @param   dmo
    *          If non-<code>null</code>, the caller is providing a DMO which
    *          it thinks is the appropriate object to act upon during rollback
    *          processing.
    */
   protected void prepareRollback(Persistable dmo)
   {
   }
   
   /**
    * Indicate whether the DMO which is the target of a rollback was newly
    * created by the current context.
    * 
    * @return  <code>true</code> to indicate target DMO was newly created by
    *          the current context;  else <code>false</code>.
    */
   protected boolean isNewlyCreated()
   {
      return newlyCreated;
   }
   
   /**
    * Indicate whether the DMO which is the target of a rollback was newly
    * created by the current context.
    * 
    * @param   isNew
    *          <code>true</code> to indicate <code>dmo</code> is newly created
    *          by the current context;  else <code>false</code>.
    */
   protected void setNewlyCreated(boolean isNew)
   {
      this.newlyCreated = isNew;
   }
   
   /**
    * Duplicate this reversible action. Implementations which require 
    * this behavior must override this method.
    * 
    * @param   key
    *          The key to set for the new reversible action.
    * 
    * @return  This implementation does not return normally;  override to
    *          return a duplicate of this reversible.
    * 
    * @throws  UnsupportedOperationException
    *          If this operation is not supported.
    */
   Reversible duplicate(Serializable key)
   {
      throw new UnsupportedOperationException(
            "Duplicate is not supported for reversible " + 
            this.toString());
   }
}