/*
** Module   : ScrollingResults.java
** Abstract : Implementation of Results which wrappers ScrollableResults
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20070711  @34451  Created initial version. Extracted inner class from PreselectQuery.
** 002 ECF 20080709  @39089  Modified sessionClosing(). We now clean up resources when the session
**                           is closing, since the results cannot be accessed afterward.
** 003 CA  20080815  @39467  Support API change in DBUtils.
** 004 CA  20090707  @43070  On cleanup, do not re-close the results if they are already closed.
** 005 ECF 20090711  @43190  Added implicit transaction support. Call {push/pop}
**                           ImplicitTransaction() in Persistence in c'tor and cleanup().
** 006 ECF 20150715          Replaced Apache commons logging with J2SE logging.
** 007 OM  20160720          Prevented ContextLocalCleanupException to be thrown when client
**                           disconnects and the context is not accessible.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.logging.*;
import org.hibernate.*;
import com.goldencode.p2j.util.*;

/**
 * A simple, scrolling implementation of the <code>Results</code>
 * interface, which offers a thin layer around the internal,
 * <code>ScrollableResults</code> object provided by the query.
 */
final class ScrollingResults
implements Results
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ScrollingResults.class.getName());
   
   /** Scrollable result set;  primary keys or <code>Persistable</code>s */
   private final ScrollableResults results;
   
   /** Persistence object which provided the underlying ScrollableResults */
   private final Persistence persistence;
   
   /** 
    * The {@link Persistence} context. Only used at cleanup time when the getter is not accessible
    * any more.
    */
   private final Persistence.Context persistenceContext;
   
   /** Track closed state of the {@link #results}. */
   private boolean closed = false;
   
   /**
    * Construct an instance of this object from a scrollable result set
    * provided by query execution.
    *
    * @param   persistence
    *          The associated persistence service object.
    * @param   results
    *          Scrollable result set generated by the query.
    */
   ScrollingResults(Persistence persistence, ScrollableResults results)
   throws PersistenceException
   {
      this.results = results;
      this.persistence = persistence;
      this.persistenceContext = persistence.getContext();
      
      persistence.pushImplicitTransaction();
   }
   
   /**
    * Move cursor to the first result row.
    *
    * @return  <code>true</code> if there are any results.
    */
   public boolean first()
   {
      return results.first();
   }
   
   /**
    * Move cursor to the last result row.
    *
    * @return  <code>true</code> if there are any results.
    */
   public boolean last()
   {
      return results.last();
   }
   
   /**
    * Move cursor to the next result row.
    *
    * @return  <code>true</code> if there is a result under the cursor
    *          under the move.
    */
   public boolean next()
   {
      return results.next();
   }
   
   /**
    * Move cursor to the previous result row.
    *
    * @return  <code>true</code> if there is a result under the cursor
    *          under the move.
    */
   public boolean previous()
   {
      return results.previous();
   }
   
   /**
    * Is the cursor on the first row in the result set?
    *
    * @return  <code>true</code> if the cursor is on the first row.
    */
   public boolean isFirst()
   {
      return results.isFirst();
   }
   
   /**
    * Is the cursor on the last row in the result set?
    *
    * @return  <code>true</code> if the cursor is on the first row.
    */
   public boolean isLast()
   {
      return results.isLast();
   }
   
   /**
    * Get the array of objects at the current result row.
    *
    * @return  Object array or <code>null</code>.
    */
   public Object[] get()
   {
      return results.get();
   }
   
   /**
    * Get the object at the current result row, at the specified column.
    *
    * @param   column
    *          Zero-based index column of the desired object.
    *
    * @return  Object at <code>column</code> or <code>null</code>.
    */
   public Object get(int column)
   {
      Object[] row = get();
      
      return (row != null ? row[column] : null);
   }
   
   /**
    * Get the primary key ID at the current result row, at the specified
    * column.
    *
    * @param   column
    *          Zero-based index column of the desired ID.
    *
    * @return  ID of the record or <code>null</code>.
    */
   public Serializable getID(int column)
   {
      Object result = get(column);
      if (result instanceof Persistable)
      {
         return ((Persistable) result).getId();
      }
      
      return (Serializable) result;
   }
   
   /**
    * Get the row number currently under the cursor.
    *
    * @return  Zero-based index of the current row, or <code>-1</code>
    *          if the cursor is not currently on a result.
    */
   public int getRowNumber()
   {
      return results.getRowNumber();
   }
   
   /**
    * Set the row number currently under the cursor.
    *
    * @param   row
    *          Zero-based index of the row to be set as the current row.
    *
    * @return  <code>true</code> if there is a row at the specified row
    *          number;  else <code>false</code>.
    */
   public boolean setRowNumber(int row)
   {
      return results.setRowNumber(row);
   }
   
   /**
    * Scroll the cursor ahead by the specified number of rows.  If the
    * number is negative, the cursor is moved backward.  This may put the
    * cursor off the end (either end) of the query's results.
    *
    * @param   rows
    *          Number of rows to jump ahead or back.
    *
    * @return  <code>true</code> if there is a row at the new location;
    *          else <code>false</code>.
    */
   public boolean scroll(int rows)
   {
      return results.scroll(rows);
   }
   
   /**
    * Reset the cursor to its natural starting position, before the first
    * result row.
    */
   public void reset()
   {
      results.beforeFirst();
   }
   
   /**
    * Invoked when the current Hibernate session is about to close.  Gives
    * the implementation an opportunity to perform appropriate processing in
    * response to this event.
    */
   public void sessionClosing()
   {
      cleanup();
   }
   
   /**
    * Close the wrapped {@code ScrollableResults} object.
    */
   public void cleanup()
   {
      if (closed)
      {
         return;
      }
      
      try
      {
         results.close();
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(persistence.getDatabase(), exc);
         
         // Non-critical operation;  logging the error is sufficient.
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, exc.getMessage(), exc);
         }
         else if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, exc.getMessage());
         }
      }
      finally
      {
         closed = true;
         persistence.popImplicitTransaction(persistenceContext);
      }
   }
}