/*
** Module   : UnpooledConnectionProvider.java
** Abstract : Simplified implementation of connection provider which does no connection pooling
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060619 @27501 Created initial version. Simple connection
**                         provider implementation designed to reduce
**                         complexity for debugging purposes.
** 002 SVL 20080725 @39175 Javadoc and logging output was modified
**                         in order to indicate that this class is
**                         used in production.
** 003 SVL 20120331        Upgraded to Hibernate 4.
** 004 ECF 20160202        Replaced Apache commons logging with J2SE logging.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.sql.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.service.*;
import org.hibernate.service.jdbc.connections.spi.*;
import org.hibernate.service.spi.*;
import com.goldencode.p2j.util.*;

/**
 * A simplified implementation of Hibernate's <code>ConnectionProvider</code>
 * interface which does no connection pooling, but instead fetches a new JDBC
 * connection each time {@link #getConnection} is invoked, and closes that
 * connection each time {@link #closeConnection} is invoked.
 * <p>
 * This implementation is used for in-memory databases, when connection
 * pooling actually works slower than opening a new connection each time. Also
 * it can be used for debugging purposes, it allows one to debug
 * connection-related issues while avoiding the behavioral side effects which
 * occur when a pooling connection provider recycles, proxies, caches, or
 * otherwise obfuscates JDBC connections.
 */
public final class UnpooledConnectionProvider
implements ConnectionProvider,
           Configurable
{
   /** Logger */
   private static final Logger LOG =
      LogHelper.getLogger(UnpooledConnectionProvider.class.getName());
   
   /** URL to connect to the backing datastore */
   private String jdbcURL = null;
   
   /** User name used to connect */
   private String user = null;
   
   /** Password used to connect */
   private String password = null;
   
   /**
    * Default constructor required by Hibernate's connection provider
    * contract.
    */
   public UnpooledConnectionProvider()
   {
   }
   
   /**
    * Configure connection provider, extracting a URL, user name, and
    * password from the specified properties.
    *
    * @param   props
    *          Hibernate configuration properties.
    */
   public void configure(Map props)
   {
      jdbcURL = (String) props.get("hibernate.connection.url");
      user = (String) props.get("hibernate.connection.username");
      password = (String) props.get("hibernate.connection.password");
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.fine("Simple connection provider configured for " + jdbcURL +
                  ";  user:  " + user + ";  password:  " +
                  (password == null ? "N/A" : "********"));
      }
   }
   
   /**
    * Obtain a new JDBC connection from the backing datastore.
    *
    * @return  New JDBC connection.
    */
   public Connection getConnection()
   throws SQLException
   {
      Connection conn = DriverManager.getConnection(jdbcURL, user, password);
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.fine("Getting JDBC connection:  " + conn);
      }
      
      return conn;
   }
   
   /**
    * Close the given JDBC connection.
    *
    * @param   conn
    *          Connection to close.
    */
   public void closeConnection(Connection conn)
   throws SQLException
   {
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.fine("Closing JDBC connection:  " + conn);
      }
      
      conn.close();
   }
   
   /**
    * No-op implementation of required method.
    */
   public void close()
   {
   }
   
   /**
    * Returns <code>false</code> to indicate JDBC connections will not be
    * released aggressively.
    *
    * @return  <code>false</code>.
    */
   public boolean supportsAggressiveRelease()
   {
      return false;
   }
   
   /**
    * Allows to determine if this wrapped service can be unwrapped as the indicated type.
    *
    * @param  unwrapType
    *         The type to check.
    *
    * @return <code>true</code> if this wrapped service can be unwrapped as the indicated type.
    */
   public boolean isUnwrappableAs(Class unwrapType)
   {
      return ConnectionProvider.class.equals(unwrapType) ||
             UnpooledConnectionProvider.class.isAssignableFrom(unwrapType);
   }
   
   /**
    * Unproxy the service proxy.
    *
    * @param  unwrapType
    *         The java type as which to unwrap this instance.
    *
    * @return The unwrapped reference.
    *
    * @throws UnknownUnwrapTypeException
    *         if the service cannot be unwrapped as the indicated type.
    */
   public <T> T unwrap(Class<T> unwrapType)
   {
      if (isUnwrappableAs(unwrapType))
      {
         return (T) this;
      }
      
      throw new UnknownUnwrapTypeException(unwrapType);
   }
}