/*
** Module   : ValidationData.java
** Abstract : Validation parameters
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20090331   @42122 Created initial version. Validation parameters.
** 002 ECF 20090511   @42156 Fixed default constructor. Last two arguments
**                           were transposed.
** 003 ECF 20090623   @43224 Added isTransient flag and getter.
** 004 ECF 20140826          Added deferred flag and getter/setter.
** 005 OM  20160316          Added cumulativeDirtyNonUnique.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.util.*;

/**
 * This object stores validation parameters to assist the {@link DMOValidator}
 * during record validation.  Depending upon the type of validation requested
 * (e.g., full, remaining, dirty), the parameters will vary.  This object
 * references the names of all dirty properties, the names of all unique and
 * non-unique indexes that should be checked, whether dirty updates should be
 * shared via the dirty share manager, and whether an insert into the dirty
 * share manager is required.
 */
final class ValidationData
{
   /** All dirty properties requiring validation */
   private final Set<String> dirtyProps;
   
   /** All unique indexes requiring validation/sharing */
   private final Collection<String> dirtyUnique;
   
   /** All non-unique indexes requiring sharing */
   private final Collection<String> dirtyNonUnique;
   
   /** All non-unique indexes requiring sharing */
   private final Collection<String> cumulativeDirtyNonUnique;
   
   /** Is an insert required in the dirty share manager? */
   private final boolean insertRequired;
   
   /** Should insert/update information be shared with other contexts? */
   private final boolean shareDirty;
   
   /** Is the affected record newly created and still transient? */
   private final boolean isTransient;
   
   /** Was some validation work and dirty database sharing deferred? */
   private boolean deferred = false;
   
   /**
    * Constructor.
    * 
    * @param   dirtyProps
    *          All dirty properties requiring validation.
    * @param   dirtyUnique
    *          All unique indexes requiring validation/sharing.
    * @param   dirtyNonUnique
    *          All non-unique indexes requiring sharing.
    * @param   cumulativeDirtyNonUnique
    *          All non-unique indexes matched since the record was created/loaded.
    * @param   insertRequired
    *          Is an insert required in the dirty share manager?
    * @param   shareDirty
    *          Should insert/update information be shared with other contexts?
    * @param   isTransient
    *          Is the affected record newly created and still transient?
    */
   ValidationData(Set<String> dirtyProps,
                  Collection<String> dirtyUnique,
                  Collection<String> dirtyNonUnique,
                  Collection<String> cumulativeDirtyNonUnique,
                  boolean insertRequired,
                  boolean shareDirty,
                  boolean isTransient)
   {
      this.dirtyProps = Collections.unmodifiableSet(dirtyProps);
      this.dirtyUnique = dirtyUnique;
      this.dirtyNonUnique = dirtyNonUnique;
      this.cumulativeDirtyNonUnique = cumulativeDirtyNonUnique;
      this.insertRequired = insertRequired;
      this.shareDirty = shareDirty;
      this.isTransient = isTransient;
   }
   
   /**
    * Get the set of dirty property names.
    * 
    * @return  Dirty property names.
    */
   Set<String> getDirtyProperties()
   {
      return dirtyProps;
   }
   
   /**
    * Get the collection of dirty, unique, index names.
    * 
    * @return  Collection of dirty, unique, index names, in ascending
    *          alphabetical order.
    */
   Collection<String> getDirtyUnique()
   {
      return dirtyUnique;
   }
   
   /**
    * Get the collection of dirty, non-unique, index names since the last validation/share.
    * 
    * @return  Collection of dirty, non-unique, index names, in ascending
    *          alphabetical order.
    */
   Collection<String> getDirtyNonUnique()
   {
      return dirtyNonUnique;
   }
   
   /**
    * Get the collection of dirty, non-unique, index names since the record was loaded/created.
    * 
    * @return  Collection of dirty, non-unique, index names, in ascending alphabetical order.
    */
   Collection<String> getCumulativeDirtyNonUnique()
   {
      return cumulativeDirtyNonUnique;
   }
   
   /**
    * Indicate whether we need to insert a record into the dirty share
    * manager to reflect a new record.
    * 
    * @return  <code>true</code> to make the insert, else <code>false</code>.
    */
   boolean isInsertRequired()
   {
      return insertRequired;
   }
   
   /**
    * Indicate whether DMO insert/update information should be shared with
    * other contexts via the dirty share manager.
    * 
    * @return  <code>true</code> to share updates, else <code>false</code>.
    */
   boolean isShareDirty()
   {
      return shareDirty;
   }
   
   /**
    * Indicate whether the affected record is newly created and is still
    * transient.
    * 
    * @return  <code>true</code> if transient, else <code>false</code>.
    */
   boolean isTransient()
   {
      return isTransient;
   }
   
   /**
    * Indicate whether possibly uniqueness validation and dirty sharing were deferred because not
    * all properties in an index were dirty for a transient record. This will require the index
    * to be validated when the record is flushed.
    * 
    * @return  <code>true</code> if deferred, else <code>false</code>.
    */
   boolean wasDeferred()
   {
      return deferred;
   }
   
   /**
    * Set the deferred flag to indicate dirty sharing and possibly uniqueness validation of an
    * updated property were deferred.
    * 
    * @see     #wasDeferred()
    */
   void setDeferred()
   {
      deferred = true;
   }
}