/*
** Module   : DialectHelper.java
** Abstract : a helper to P2J dialect classes
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 VMN 20131012 Created initial version.
** 002 OM  20140410 Fixed support case-sensitive fields in indexes.
** 003 OM  20150525 Added static method getDialectClass().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.dialect;

import java.net.*;
import com.goldencode.p2j.persist.*;
import org.hibernate.dialect.*;

/**
 * A helper to P2J dialect classes. This class provides static helper methods and constants used for
 * P2J dialects.
 */
public class DialectHelper
{
   /** 
    * Common prefix prepended to computed column/property names. Additional decoration will be added
    * in order to identify columns that maps case sensitive fields.
    */
   static final String CC_PREFIX = "__";
   
   /** Signature for case-insensitive character column/property names */
   public static final String INSENSITIVE_CHAR_FIELD = CC_PREFIX + "i";
   
   /** Signature for case-sensitive character column/property names */
   public static final String SENSITIVE_CHAR_FIELD   = CC_PREFIX + "s";
   
   /** Double quote character */
   public static final char DOUBLE_QUOTE = '"';
   
   /**
    * Build the URL to be used for a remote connection, based on the given URL and database.
    *
    * @param    url
    *           The URL as received from the remote server.
    * @param    database
    *           The database to which this URL belongs.
    *
    * @return   The built URL as to be used for a remote connection.
    *
    * @throws com.goldencode.p2j.persist.PersistenceException
    *           if the specified connection URL is invalid.
    */
   public static String buildRemoteURL(String url, Database database)
   throws PersistenceException
   {
      String origUrl = url;
      try
      {
         url = url.substring("jdbc:".length());
         URI uri = (new URI(url)).normalize();
         
         // check for localhost references.
         String host = uri.getHost().toLowerCase();
         if (InetAddress.getByName(host).isLoopbackAddress())
         {
            // use hostname stored in database object.
            host = database.getSocketAddress().getHostName();
            
            // generate new URI.
            uri = new URI(uri.getScheme(),
                          uri.getUserInfo(),
                          host,
                          uri.getPort(),
                          uri.getPath(),
                          uri.getQuery(),
                          uri.getFragment());
            
            // prepend 'jdbc:' prefix.
            url = "jdbc:" + uri;
         }
         else
         {
            // no preparation is needed
            url = origUrl;
         }
      }
      catch (Exception exc)
      {
         throw new PersistenceException("Invalid database connection URL", exc);
      }
      
      return url;
   }
   
   /**
    * Return the Java class for a dialect using Hibernate identifier.
    * 
    * @param   dialect
    *          The dialect name. Case insensitive.
    * 
    * @return  The dialect class or {@code null} if the parameter does not match any known
    *          dialect identifier.
    */
   public static Class getDialectClass(String dialect)
   {
      switch (dialect.toLowerCase())
      {
         case "h2":
            return P2JH2Dialect.class;
         case "postgresql":
            return P2JPostgreSQLDialect.class;
         case "sqlserver2012":
            return P2JSQLServer2012Dialect.class;
         default:
            return null;
      }
   }
}