/*
** Module   : P2JSQLServer2008Dialect.java
** Abstract : Implementation of P2JDialect for a MS SQL Server 2008 backend
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 VMN 20131012 Created initial version. Adds P2JDialect method implementations to extend
**                  Hibernate's SQLServer2008Dialect.
** 002 OM  20140106 Added support for mixed asc/desc indexes.
** 003 OM  20140410 Fixed support case-sensitive fields in indexes. Replaced Apache logging.
**                  Added Override method annotations.
** 004 VMN 20140427 Added enhance schema name conversion support for hint "escape" attribute.
** 005 OM  20140508 Added support for computing the index-key size.
**                  Fixed character computed columns expression.
**                  Added implementation for some method stubs required for p2j server startup.
**                  Forced adding ID to each index (including already unique) because SQL issue.
** 006 OM  20140529 Lowest date set to 01/01/001. Added rtrim() dialect specific implementation.
**                  Added query max parameters count support. Changed generation of computed
**                  character columns.
** 007 OM  20140407 Added support for optional SQL length specified in hints.
** 008 OM  20140924 Removed usesSingleBackslash.
** 009 OM  20151130 rtrim() dialect specific implementation returns the number of occurrences
**                  of the expression processed. Returned the list of ddl for creating support
**                  UDFs from p2jpl assembly. Implemented requiresExplicitCastInsideTernary() and
**                  supportsBooleanDatatype().
** 010 EVL 20160217 Clean up comments from symbols invalid for Solaris to compile.
** 011 OM  20160608 Greatly simplified rtrimming.
** 012 ECF 20160912 Added getSequencePrefetchString.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.dialect;

import java.sql.Date;
import java.sql.SQLException;
import java.sql.Types;
import java.text.NumberFormat;
import java.util.*;
import java.util.List;
import java.util.logging.*;
import com.goldencode.p2j.util.*;
import org.hibernate.*;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.SQLServer2008Dialect;
import org.hibernate.dialect.function.SQLFunction;
import org.hibernate.engine.spi.*;
import org.hibernate.mapping.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.sequence.SequenceHandler;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.util.date;
import com.goldencode.util.StringHelper;

import static com.goldencode.p2j.persist.dialect.DialectHelper.*;

/**
 * Extends Hibernate's <code>SQLServer2008Dialect</code> class to add method implementations
 * required by the {@link P2JDialect} interface.
 * <p>
 * Note:  this implementation expects a directory to be available at the time
 * the class is loaded.
 */
public class P2JSQLServer2008Dialect
extends SQLServer2008Dialect
implements P2JDialect
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(P2JSQLServer2008Dialect.class);

   /** Flag indicating whether to inline a record limit in an SQL SELECT */
   private static final ContextLocal<Boolean> inlineLimit =
           new ContextLocal<Boolean>()
           {
              protected Boolean initialValue() { return Boolean.FALSE; }
           };

   /**
    * <code>datetime</code> is deprecated and it does not align with the SQL Standard. 
    * We convert p4gl <code>date</code>-related datatypes using the MSSQL <code>date</code> and 
    * <code>datetime2</code>. For both the lower limit is January 1, 001.
    * <p>
    * date: <a href='http://msdn.microsoft.com/en-us/library/bb630352(v=sql.100).aspx'>
    *    msdn.microsoft.com/en-us/library/bb630352(v=sql.100).aspx</a><br>
    * datetime2: <a href='http://msdn.microsoft.com/en-us/library/bb677335(v=sql.100).aspx'>
    *    msdn.microsoft.com/en-us/library/bb677335(v=sql.100).aspx</a>
    */
   private static final Date lowDate;
   
   /**
    * Highest date supported by SQL Server 2008 (December 31, 9999)
    * http://msdn.microsoft.com/en-US/library/ms187819%28v=sql.100%29.aspx
    */
   private static final Date highDate;
   
   /**
    * Private constant from SQLServer2005Dialect used in sql types mapping with
    * registerColumnType(). We need to overwrite mapping to varchar/varbinary exactly with this
    * constant.
    */
   private static final int MAX_LENGTH = 8000;
   
   /** The maximum size of all fields of an index, in bytes. */
   private static int MAX_INDEX_SIZE = 900;
   
   /** The maximum number of parameters allowed by the SQL Server is a single query. */ 
   private static int MAX_PARAMETER_COUNT = 2100;
   
   static
   {
      // Prepare supported date values.  We use the GMT timezone to define
      // our date range, since this is how com.goldencode.p2j.util.date handles its dates.
      
      TimeZone tz = TimeZone.getTimeZone("GMT");
      GregorianCalendar cal = new GregorianCalendar(tz);
      
      cal.set(Calendar.ERA, GregorianCalendar.AD);
      cal.set(Calendar.YEAR, 1);
      cal.set(Calendar.MONTH, Calendar.JANUARY);
      cal.set(Calendar.DATE, 1);
      lowDate = new Date(cal.getTime().getTime());
      
      cal.clear();
      
      cal.set(Calendar.ERA, GregorianCalendar.AD);
      cal.set(Calendar.YEAR, 9999);
      cal.set(Calendar.MONTH, Calendar.DECEMBER);
      cal.set(Calendar.DATE, 31);
      highDate = new Date(cal.getTime().getTime());
   }
   
   /**
    * Default constructor.  This constructor is for Hibernate's use.  Use code
    * should not construct instances of this class.
    * <p>
    * We provide this constructor to override the superclass' mapping of
    * <code>Types.VARCHAR</code> and <code>Types.VARBINARY</code>.
    */
   public P2JSQLServer2008Dialect()
   {
      super();
      
      // Override superclass' VARCHAR and VARBINARY mappings, since we always
      // use the unbounded variants of these data types.
      registerColumnType(Types.VARBINARY, MAX_LENGTH, "varbinary");
      
      // the varchar (set in CharacterUserType.sqlTypes()) will be specially handled in
      // getTypeName(int code, long length, int precision, int scale), see javadoc for details
      registerColumnType(Types.VARCHAR, MAX_LENGTH, "varchar($l)");
      
      // even if SQL Servers support much larger datatypes here (2GB), we use the P4GL limitation
      // of 32K instead, however, these columns CANNOT be part of an index
      registerColumnType(Types.LONGNVARCHAR, 32*1024, "varchar(MAX)");
      
      // Maximum SQL Server 2008-2012 NUMERIC column precision is 38.
      registerColumnType(Types.NUMERIC, "numeric(38,$s)");
   }
   
   /**
    * Create a data object, the purpose of which is specific to a particular
    * dialect implementation, to be stored as a context-local object.  This
    * allows a dialect to establish context-local data which is stored by the
    * persistence runtime framework, without the overhead of a separate,
    * context-local variable.  This object will be handed back to the dialect
    * in callback methods when the services of certain hooks are required.
    * <p>
    * Since the context-local data is specific to a particular dialect
    * implementation, it is optional.  Dialect implementations which do not
    * have a need for this data should simply return <code>null</code>.
    *
    * @param   persistence
    *          Provider of persistence services.
    *
    * @return Dialect-specific, context-local data, or <code>null</code>.
    */
   @Override
   public Object createContext(Persistence persistence)
   {
      return null;
   }
   
   /**
    * A dialect-specific callback hook which is invoked by the persistence
    * runtime framework after a transaction is complete.  The Hibernate
    * session will have been closed by the time this method is invoked.
    *
    * @param   persistence
    *          Provider of persistence services.
    * @param   context
    *          Optional, context-local data created previously by {@link #createContext}.
    * @param   rollback
    *          <code>true</code> if transaction ended with a rollback;
    *          <code>false</code> if transaction ended with a commit.
    *
    * @throws com.goldencode.p2j.persist.PersistenceException
    *          if any persistence-related error occurs.
    */
   @Override
   public void postTransaction(Persistence persistence, Object context, boolean rollback)
   throws PersistenceException
   {
      // do nothing
   }
   
   /**
    * Get the name of the database type associated with the given {@link java.sql.Types} typecode
    * with the given storage specification parameters.
    * <p>
    * The <code>Types.VARCHAR</code> is handled differently because Hibernate cannot handle the
    * optional sql size hint (see also the class javadoc for
    * {@link org.hibernate.dialect.TypeNames}):
    * <ul>
    *    <li>if <i>varchar</i> would be register for this type, maximum one character would be
    *          allowed because this is equivalent to <i>varchar(1)</i>;
    *    <li>if <i>varchar(8000)</i> would be register for this type, then the
    *          {@link org.hibernate.dialect.TypeNames#get(int, long, int, int)} cannot set
    *          the optional size parameter;
    *    <li>otherwise, if <i>varchar($l)</i> would be register for this type then, if the
    *          sql size is not specified, the default value (0) would lead to <b>varchar(0)</b>
    *          to be generated by method above mentioned, which is obviously invalid.
    * </ul>
    * As a result, we manually handle the character columns by using the sql size value if the
    * hint was specified, otherwise using the maximum allowable of 8000 for non-large data-type.
    * At the same time we do the validation do the size hint (for this dialect) is restricted
    * to a positive number and not greater than 8K bytes.
    *<p>
    * When generating the *.hbm file all properties must have the 'length' attribute set.
    * Otherwise, <code>SchemaExport</code> will use the default column length of 255 when 
    * calling this method so the output will not use the expected MAX_LENGTH size.
    *
    * @param   code
    *          The {@link java.sql.Types} typecode
    * @param   length
    *          The datatype length
    * @param   precision
    *          The datatype precision
    * @param   scale
    *          The datatype scale
    *
    * @return  the database type name
    *
    * @throws  HibernateException If no mapping was specified for that type.
    */
   @Override
   public String getTypeName(int code, long length, int precision, int scale)
   throws HibernateException
   {
      if (code == Types.VARCHAR && (length <= 0 || length > MAX_LENGTH))
      {
         // SQL size not specified by a hint or invalid size ?
         length = MAX_LENGTH;
      }
      return super.getTypeName(code, length, precision, scale);
   }
   
   /**
    * Retrieve the earliest possible date value which is valid for the
    * backing database.
    *
    * @return Earliest date supported by the backing database.
    */
   @Override
   public Date getLowDate()
   {
      return lowDate;
   }

   /**
    * Retrieve the latest possible date value which is valid for the backing
    * database.
    *
    * @return Latest date supported by the backing database.
    */
   @Override
   public Date getHighDate()
   {
      return highDate;
   }

   /**
    * Format a date object as a string which is appropriate for use in an SQL statement.
    *
    * @param   d
    *          Date to be formatted as a string.
    *
    * @return  Formatted representation of the given date.
    */
   @Override
   public String formatDate(date d)
   {
      NumberFormat nf = NumberFormat.getInstance();
      
      nf.setMinimumIntegerDigits(4);
      nf.setGroupingUsed(false);
      
      StringBuilder buf = new StringBuilder();
      int year = d.getYear();
      
      buf.append(nf.format(Math.abs(year)));
      buf.append('-');
      
      nf.setMinimumIntegerDigits(2);
      buf.append(nf.format(d.getMonth()));
      buf.append('-');
      buf.append(nf.format(d.getDay()));
      
      return buf.toString();
   }
   
   /**
    * Given a native column name as queried from database metadata, extract the root name of a
    * column. For example, a column name as found within an index in metadata might be represented
    * as:
    * <pre>
    *    __my_column
    * </pre>
    * We are interested only in the {@code my_column} root name, so we strip away the leading
    * {@code __} prefix.
    *
    * @param   expr
    *          Column name as natively represented by the database dialect.
    *
    * @return  Root column name.
    */
   @Override
   public String extractColumnName(String expr)
   {
      if (expr.startsWith(INSENSITIVE_CHAR_FIELD))
      {
         return expr.substring(INSENSITIVE_CHAR_FIELD.length());
      }
      else if (expr.startsWith(SENSITIVE_CHAR_FIELD))
      {
         return expr.substring(SENSITIVE_CHAR_FIELD.length());
      }
      else if (expr.startsWith(CC_PREFIX))
      {
         // has the computed column marker but not case-sensitivity
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Invalid column name: [" + expr + "]." +
                  "Failed to extract the root name of the column.");
         }
      }
      // Strip enclosing double quotes or backticks.
      expr = StringHelper.stripEnclosing(expr, '"');
      expr = StringHelper.stripEnclosing(expr, '`');
      
      return expr;
   }
   
   /**
    * Adjust a character column name appropriate to this dialect, such that it
    * can be inserted into an index.  This may involve wrapping it within one
    * or more functions, prepending a special prefix to designate a computed
    * column, or any other transformation appropriate to the database syntax.
    *
    * @param   name
    *          Base column name.
    * @param   ignoreCase
    *          <code>true</code> to cause case to be ignored, else <code>false</code>.
    *
    * @return  Processed column name.
    */
   @Override
   public String getProcessedCharacterColumnName(String name, boolean ignoreCase)
   {
      // not sure implementation is correct, TODO investigate issue
      name = DBUtils.getPrefixedParameter(
                  name,
                  ignoreCase ? INSENSITIVE_CHAR_FIELD : SENSITIVE_CHAR_FIELD);
      
      StringBuilder buf = new StringBuilder();
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      buf.append("rtrim(").append(name).append(")");
      if (ignoreCase)
      {
         buf.append(")");
      }
      
      return buf.toString();
   }

   /**
    * Report whether the given column name represents an alias or an
    * expression which indicates the column is case insensitive.  Assume the
    * column represented by <code>name</code> is already determined to be the
    * appropriate data type for textual data.
    *
    * @param   name
    *          Column name, alias, or expression.
    *
    * @return  <code>true</code> if <code>name</code> represents case-insensitive data.
    */
   @Override
   public boolean isCaseInsensitiveColumn(String name)
   {
      name = StringHelper.stripEnclosing(name, DOUBLE_QUOTE);
      return name.startsWith(INSENSITIVE_CHAR_FIELD);
   }
   
   /**
    * Report whether the given column name represents an alias or an expression which indicates
    * the column is computed.  Assume the column represented by <code>name</code> is
    * already determined to be the appropriate data type for textual data.
    *
    * @param   name
    *          Column name, alias, or expression.
    *
    * @return  <code>true</code> if <code>name</code> represents a computed column.
    */
   public boolean isComputedColumn(String name)
   {
      // we rely on the fact that both INSENSITIVE_CHAR_FIELD and SENSITIVE_CHAR_FIELD
      // have the same prefix: CC_PREFIX
      return name.startsWith(CC_PREFIX);
   }

   /**
    * Indicate whether this dialect requires computed columns to participate
    * within index definitions, instead of embedded expressions.
    *
    * @return  <code>true</code> to indicate computed columns are required.
    */
   @Override
   public boolean needsComputedColumns()
   {
      return true;
   }

   /**
    * Get the prefix, if any, which should be prepended to computed column and/or property names.
    *
    * @param   caseSensitive
    *          <code>true</code> if this is a character field was declared with
    *          <code>CASE-SENSITIVE</code> attribute.
    *
    * @return  <code>null</code> to indicate the backing database does not use computed columns.
    *
    * @see     #needsComputedColumns()
    */
   @Override
   public String getComputedColumnPrefix(boolean caseSensitive)
   {
      return caseSensitive ? SENSITIVE_CHAR_FIELD : INSENSITIVE_CHAR_FIELD;
   }
   
   /**
    * Indicate whether this dialect natively supports the overloading of user
    * defined functions by parameter number and data type.
    *
    * @return  false because Server 2008 does not support functions overloading
    *          ("Function names ... must be unique within the database and to its schema.")
    */
   @Override
   public boolean supportsFunctionOverloading()
   {
      return false;
   }
   
   /**
    * Indicates whether this dialect supports boolean datatype. Support includes:
    * <ul>
    *    <li>usage of {@code true} and {@code false} literals;
    *    <li>operations with boolean values;
    *    <li>allowing functions to have booleans as parameters and return datatype.
    * </ul>
    * Since SQL Server uses BIT (integer datatype with {@code 0} and {@code 1} values) as boolean
    * replacement, this method always returns {@code false}.
    *
    * @return  Always {@code false}.
    */
   public boolean supportsBooleanDatatype()
   {
      return false;
   }
   
   /**
    * Generate a DDL string which will create a global sequence with the given
    * starting value.
    *
    * @param   name
    *          Sequence name.
    * @param   start
    *          Sequence starting value.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getCreateSequenceString(String name, long start)
   {
      return null;
   }

   /**
    * Generate a DDL string which will create a sequence with given properties.
    *
    * @param   name
    *          Sequence name. Mandatory not null.
    * @param   init
    *          Sequence starting value. Mandatory(not null).
    * @param   increment
    *          Sequence incrementing value. Mandatory(not null).
    * @param   minVal
    *          Sequence minimum value. Optional (may be null).
    * @param   maxVal
    *          Sequence maximum value. Optional (may be null).
    * @param   cycle
    *          Sequence cycling property. Optional, if missing assumed not cycling.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getCreateSequenceString(String name,
                                         Long init,
                                         Long increment,
                                         Long minVal,
                                         Long maxVal,
                                         Boolean cycle)
   {
      return null;
   }

   /**
    * Generate a DDL string which will drop a sequence with given properties.
    *
    * @param   name
    *          The name of the sequence to be dropped.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getDropSequenceString(String name)
   {
      return null;
   }

   /**
    * Generate the appropriate select statement to retrieve the next value of a sequence.
    *
    * @param   sequenceName
    *          The name of the sequence.
    * @param   newVal
    *          The new value to be set.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getSequenceSetValString(String sequenceName, long newVal)
   {
      return null;
   }

   /**
    * Generate the appropriate select statement to reset the value of a sequence.
    *
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public String getSequenceCurrValString(String sequenceName)
   {
      return null;
   }

   /**
    * Returns <code>true</code> if this dialect fully support the sequences as
    * 4GL language (most important: bounded values, cycle). If not fully
    * supported, additional code is needed in P2J sequence class implementation to fix this.
    *
    * @return  <code>false</code> because Server 2008 does not have the concept of a SEQUENCE.
    *          It appeared since SQL Server 2012.
    */
   @Override
   public boolean hasFullSequenceSupport()
   {
      return false;
   }
   
   /**
    * Generate the appropriate select statement to pre-fetch multiple, ascending values from a
    * sequence. The values need not be contiguous.
    * 
    * @param   sequenceName
    *          The name of the sequence from which the values are to be retrieved.
    * 
    * @return  Dialect-specific SQL expression, or {@code null} if this feature is not supported
    *          by the dialect.
    */
   @Override
   public String getSequencePrefetchString(String sequenceName)
   {
      return null;
   }
   
   /**
    * Given an exception, this method will check if it indicates a database
    * connectivity error.
    * <p>
    * For the SQLServer dialect, 08S01 - common state for disconnect and cancel.
    * <p>
    *
    * @param   exc
    *          The exception to be checked.
    *
    * @return  <code>true</code> if the exception indicates a database
    *          connectivity error.
    */
   @Override
   public boolean isConnectionError(Exception exc)
   {
      boolean error = false;

      Throwable next = exc;
      while (next != null && !error)
      {
         if (next instanceof SQLException)
         {
            String state = ((SQLException) next).getSQLState();

            error = "08S01".equals(state);
         }
         next = next.getCause();
      }

      return error;
   }

   /**
    * Indicate whether a query parameter which is part of a range check should
    * be inlined into the where clause, or whether it should be substituted
    * into a prepared statement.  This will result in an "inlined"
    * substitution parameter within a where clause.  For example, given a
    * where clause of:
    * <pre>
    *    where value &ge; ? and id = ?
    * </pre>
    * along with respective substitution parameters of 45 and 100, the
    * inlined version would read:
    * <pre>
    *    where value &ge; 45 and id = ?
    * </pre>
    * Only the <code>id</code> substitution parameter of 100 would be
    * substituted into the prepared statement.
    * <p>
    * Inlining in such a way can give some database dialects the opportunity
    * to choose a better query plan, than if the parameter being tested with
    * a range check were to be substituted at statement execution time.
    * <p>
    * Note:  this serves as a hint which may be overridden by runtime logic or
    * other configuration settings.
    *
    * @return  <code>true</code> to inline range check parameters;
    *          <code>false</code> to use normal parameter substitution.
    */
   @Override
   public boolean isQueryRangeParameterInlined()
   {
      return false;
   }

   /**
    * Enable/disable limit clause inlining.  Inlining the limit clause will
    * cause the limit/max value to be hard-coded into the SQL statement,
    * rather than being bound as a substitution parameter.
    * <p>
    * Dialects are free to treat this method as a hint.
    *
    * @param   on
    *          <code>true</code> to enable inlining;  <code>false</code> to disable it.
    */
   @Override
   public void inlineLimit(boolean on)
   {
      // TODO
   }

   /**
    * Determines whether substitution parameters should be explicitly casted to their datatypes
    * inside "then ? else ?" statement.
    *
    * @return  always {@code true} for SQL Server dialects.
    */
   @Override
   public boolean requiresExplicitCastInsideTernary()
   {
      return true;
   }

   /**
    * Build the DDL to drop an index based upon the specified definition.
    * The SQL is generated by using the database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be dropped.
    *
    * @return  The DDL to drop the given index.
    */
   @Override
   public String getDropIndexString(P2JIndex index)
   {
      // SQL Server 2008-2012 requires fully qualified index name for drop index.
      return "if exists (select * from sys.indexes where name = '" + index.getName() + "') " +
             "drop index " + index.getTable() + '.' + index.getName();
   }

   /**
    * Build the DDL to create an index based upon the specified definition.
    * The SQL is generated by using a database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be created.
    * @param   unique
    *          If <code>true</code>, create the index as a unique index,
    *          <code>false</code> to omit the unique constraint.
    * @param   toFile
    *          <code>true</code> if the string will be output to a file or
    *          <code>false</code> if it will be directly submitted via JDBC.
    *
    * @return  The DDL which was created.
    */
   @Override
   public String getCreateIndexString(P2JIndex index, boolean unique, boolean toFile)
   {
      // Add primary key column to any index which is not explicitly or implicitly unique.
      // TODO: move this to import ruleset so we don't hardcode primary key column name.
      
      if (true /* !index.isUnique() */) 
      /* because MSSQL does not allow multiple nulls in unique indexes, we are forced to add the
       * [id] field as index component. This way, all records became unique, but the application
       * is validating the records (using 4GL constraints) before saving to persistence. 
       * This workaround is used until we find some solution to the problem. Solutions exist
       * but we need to investigate which one is suitable in our environment.
       */
      {
         // the direction of the pk column name is the same with the last component before it.
         // this make this index much easier to fit using the sort clause of the FindQuery, 
         // especially in reverse sorting mode (at runtime I cannot tell is the sorting direction
         // was already reversed during conversion for FIND LAST/FIND PREV structures so I
         // cannot assume the direction of pk id is always ascending; however, if the direction
         // is linked to one of the components of the index, reversing it's sorting direction
         // will also reverse the sort order of the pk id so the query planner will be able to
         // identify the correct index at runtime).
         boolean idDesc = false;
         if (index.size() > 0) // should normally be
         {
            P2JIndexComponent last = index.getComponent(index.size() - 1);
            idDesc = last.isDescending(); 
         }
         index.addComponent(DatabaseManager.PRIMARY_KEY, idDesc, false, false, false);
      }

      // Extract index information into a Hibernate-friendly form.
      String tabName = index.getTable();
      Table table = new Table(tabName);
      List<Column> columns = new ArrayList<>();
      Iterator<P2JIndexComponent> iter = index.components();
      while (iter.hasNext())
      {
         P2JIndexComponent next = iter.next();
         String colName = next.getName();
         if (next.isCharType())
         {
            colName = DBUtils.getPrefixedParameter(
               colName,
               getComputedColumnPrefix(!next.isIgnoreCase()));
         }
         
         // to allow asc/desc clauses for the index, we manipulate the column name to add the 
         // "desc" direction, if needed; this is safe as the H2 dialect doesn't have quoted 
         // columns, so the output DDL will be without quotes
         if (next.isDescending())
         {
            colName = colName + " desc";
         }

         columns.add(new Column(colName));
      }

      // Compose the string.
      String createString =
              Index.buildSqlCreateIndexString(this,
                      quote(index.getName()),
                      table,
                      columns.iterator(),
                      unique,
                      null,
                      null);

      return createString;
   }

   /**
    * Build the DDL to set the formula for the computed columns associated with an indexed,
    * character column. Because the computed columns have already been injected and SQL Server
    * only support minimal <code>ALTER COLUMN</code> operations, we drop the column and add it
    * back as a computed column with the correct definition formula.
    * <p>
    * The syntax to set the formula for the computed column is:
    * <pre>alter table [table] alter column [column] as [formula]</pre>
    *
    * @param   tableName
    *          Table name possible containing backticks.
    * @param   config
    *          The hibernate's configuration object.
    * @param   implClassName
    *          The full name of the implementing class for the DMO.
    * @param   colName
    *          The column name which needs to have a computed version. This is not the computed
    *          column name, but the name of its "parent".
    * @param   ignoreCase
    *          <code>true</code> if this is a character field which has the
    *          "ignore case" flag set.
    *
    * @return  The DDL to alter the computed column, so that its formula is specified.
    */
   public String getAlterComputedColumnString(String tableName,
                                              Configuration config,
                                              String implClassName,
                                              String colName,
                                              boolean ignoreCase)
   {
      Column computedCol = new Column(
            DBUtils.getPrefixedParameter(colName, getComputedColumnPrefix(!ignoreCase)));
      
      final String quotedTableName = quote(tableName);
      final String quotedComputedColumnName = quote(computedCol.getName());
      String drop = String.format("alter table %s drop column %s;",
                                  quotedTableName,
                                  quotedComputedColumnName);
      
      // the syntax to set the formula for adding a computed column is:
      // alter table <table> add column <column> as <formula>
      String create = String.format("alter table %s add %s as %s;",
                                    quotedTableName,
                                    quotedComputedColumnName,
                                    getComputedColumnFormula(quote(colName), ignoreCase));
      
      return drop + "\n" + create;
   }
   
   /**
    * Returns the formula for character computed columns. This will be called in two cases: when
    * generating the ddl for permanent databases schema (this will be written to a .sql file) and
    * when creating temp-tables (the code will be sent directly to SQL).
    *
    * @param   colName
    *          The base column name.
    * @param   ignoreCase
    *          {@code true} if this is a char field which has the "ignore case" flag set.
    *
    * @return  The formula for computed columns or {@code null} if the dialect does not
    *          support computed columns.
    */
   @Override
   public String getComputedColumnFormula(String colName, boolean ignoreCase)
   {
      StringBuilder buf = new StringBuilder();
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      buf.append("rtrim(").append(colName).append(")");
      if (ignoreCase)
      {
         buf.append(")");
      }
      return buf.toString();
   }
   
   /**
    * Prepare the given metadata query parameter.
    *
    * @param   param
    *          The metadata query parameter.
    *
    * @return the prepared parameter.
    */
   @Override
   public String prepareMetadataParameter(String param)
   {
      switch (param)
      {
         case "public" :
            // at this moment we are using the [dbo] public/default schema in SQL Server.
            // Later we might define out own schema to allow you to better manage your namespace.
            return "dbo";
         // other 'parameters' go here
         default: 
            return param;
      }
   }
   
   /**
    * Process the given metadata query result, (by making it lowercase ?).
    *
    * @param   value
    *          The metadata query result.
    *
    * @return  the  result.
    */
   @Override
   public String processMetadataResult(String value)
   {
      return value;
   }
   
   /**
    * Check if the given value is a descending clause for a column index.
    *
    * @param   value
    *          The value to be checked.
    *
    * @return  <code>true</code> if the value is the "D" or "d" string.
    */
   @Override
   public boolean isMetadataSortDesc(String value)
   {
      return "D".equalsIgnoreCase(value);
   }

   /**
    * Check if the given value is an ascending clause for a column index.
    *
    * @param   value
    *          The value to be checked.
    *
    * @return  <code>true</code> if the value is the "A" or "a" string.
    */
   @Override
   public boolean isMetadataSortAsc(String value)
   {
      return "A".equalsIgnoreCase(value);
   }

   /**
    * Get a list of database prepare DDL statements.
    *
    * @param    database
    *           The database which needs to be prepared.
    *
    * @return   an empty list.
    */
   @Override
   @SuppressWarnings("unchecked")
   public List<String> getDatabasePrepareStatements(Database database)
   {
      return SQLServerHelper.getPrepareStatements(database);
   }
   
   /**
    * Check if the dialect needs to set an explicit collation during schema
    * DDL generation.
    *
    * @return  always <code>true</code>.
    */
   @Override
   public boolean explicitSetCollation()
   {
      // collation is a property of a database or a column, not of a connection. But we can
      // override the collation on the statement level using the COLLATE statement:
      // SELECT ... FROM ... COLLATE ...
      return false;
   }
   
   /**
    * Get the statement delimiter for this dialect.
    *
    * @return  the statement delimiter.
    */
   @Override
   public String getDelimiter()
   {
      return ";";
   }
   
   /**
    * Set any database parameters before, to be available when the database
    * driver is loaded.
    * <p>
    * This implementation does nothing.
    */
   @Override
   public void beforeDriverLoad()
   {
      // no-op
   }
   
   /**
    * Build the URL to be used for a remote connection, based on the given URL and database.
    *
    * @param   url
    *          The URL as received from the remote server.
    * @param   database
    *          The database to which this URL belongs.
    *
    * @return  The built URL as to be used for a remote connection.
    *
    * @throws  PersistenceException
    *          if the specified connection URL is invalid.
    */
   @Override
   public String buildRemoteURL(String url, Database database)
   throws PersistenceException
   {
      return DialectHelper.buildRemoteURL(url, database);
   }
   
   /**
    * Creates and returns a <code>SequenceHandler</code> for a SQL Server 2008 database.
    *
    * @param   ldbName
    *          The logical name of the database to create the <code>SequenceHandler</code> for.
    *
    * @return  null because Server 2008 does not have the concept of a SEQUENCE.
    *          It is appeared since SQL Server 2012.
    */
   @Override
   public SequenceHandler createSequenceHandler(String ldbName)
   {
      return null;
   }

   /**
    * Report whether the computed columns should be injected on table creating.
    *
    * @return  always <code>true</code> for SQL Server.
    */
   @Override
   public boolean injectComputedColumns()
   {
      return true;
   }
   
   /**
    * MS SQL Server has constraints about the maximum size of fields that compose an index. All 
    * combined fields of any index must not exceed 900 bytes. 
    *
    * @return  always 900.
    */
   @Override
   public int getIndexLengthLimit()
   {
      return MAX_INDEX_SIZE;
   }
   
   /**
    * Computes the size (in bytes) occupied by this value when stored as index-key in an index.
    * It only make sense to call this method if the <code>getIndexLengthLimit()</code> returns
    * positive values. 
    * If this dialect does not have index-size constraints then this method should return 0. 
    * If the value or value type is unknown or cannot be handled by the dialect -1 will be 
    * returned. 
    *
    * @param   value
    *          The value to be analyzed. Except for the char datatype, only the type of value 
    *          is used, the actual value is not needed to determine the size occupied by the 
    *          value as index-key.
    *
    * @return  the size (in bytes) occupied by this value when stored as index-key in an index.  
    */
   public int computeIndexKeySize(BaseDataType value)
   {
      if (value instanceof logical)
      {
         return 1; // bit
      } 
      else if (value instanceof integer)
      {
         return 4; // int
      }
      else if (value instanceof int64)
      {
         return 8; // bigint
      }
      else if (value instanceof datetimetz)
      {
         return 10; // datetimeoffset
      }
      else if (value instanceof datetime)
      {
         return 8; // datetime2
      }
      else if (value instanceof date)
      {
         return 3; // date
      }
      else if (value instanceof decimal)
      {
         return 17; // numeral(38, n)
      }
      else if (value instanceof Text)
      {
         if (value.isUnknown())
         {
            return 0; // null values do not occupy space in SQL index key.
         }
         return ((Text) value).getValue().length(); // varchar
      }
      return -1;
   }
   
   /**
    * Returns the size of the synthetic part of the index. At this moment only the <code>id</code>
    * of the record may be additionally added to an index to add custom properties (eg uniqueness)
    * but other fields could be added. This part of the index is not visible using index 
    * reflection routines from <code>IndexHelper</code>. The synthetic components may differ 
    * between unique and non-unique indexes.
    * <p>
    * See {@link #getCreateIndexString(P2JIndex, boolean, boolean)}
    *
    * @param   unique
    *          <code>true</code> if the index is unique, unused.
    *
    * @return  Always 8 because the workaround of the multiple NULLS in UNIQUE indexes described 
    *          in <code>getCreateIndexString()</code>.
    */
   public int getSyntheticIndexSize(boolean unique)
   {
      return unique ? 8 : 8;
   }
   
   /**
    * Composes a rtrim equivalent call for a specified expression. This is dialect specific.
    * The returned value looks like:
    * <pre>
    *    rtrim(expr)
    * </pre>
    * The rtrimmed expression is appended to the string builder.
    *
    * @param   expr
    *          The expression to be rtrim -med. It's up to the caller if it is 
    *          a qualified field name or not or any other kind of character-type expression.
    * @param   sb
    *          The {@code StringBuilder} to append the trimmed filed to.
    */
   @Override
   public void addRtrimmedExpression(String expr, StringBuilder sb)
   {
      sb.append("rtrim(").append(expr).append(")");
   }
   
   /**
    * Obtain the number of parameters a query of this dialect is allowed by the SQL Server. 
    * For 2008 and 2012 the maximum number of parameters is 2100.
    * If the query need more parameter it have to be rewritten or split and the result 
    * reassembled in memory.
    *
    * @return  always 2100.
    */
   @Override
   public int getMaxParameterCount()
   {
      return MAX_PARAMETER_COUNT;
   }
}