/*
** Module   : LockUnavailableException.java
** Abstract : Exception thrown when a lock cannot be acquired for a no-wait lock request
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20051207   @23682 Created initial version. An exception thrown
**                           to indicate that a lock could not be acquired
**                           for a database record.
** 002 ECF 20131028          Moved to lock sub-package.
** 003 ECF 20160130          Optimized to not collect stack trace information unless performing
**                           FINE (or more granular) logging for the lock manager.
** 004 ECF 20160812          Added lock type and record identifier information to exception.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.lock;

import java.util.logging.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * This exception is thrown by the persistence service layer to indicate that
 * a request to lock a record, issued with the no-wait option, could not be
 * fulfilled, because another user holds the lock.
 * <p>
 * Supports the assignment of an error number at construction.
 */
public class LockUnavailableException
extends PersistenceException
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(LockManager.class.getName());
   
   /** Flag indicating whether we capture stack trace data */
   private static final boolean captureStack = LOG.isLoggable(Level.FINE);
   
   /** Requested lock type which was not acquired */
   private final LockType lockType;
   
   /** Identifier of record which could not be locked */
   private final RecordIdentifier identifier;
   
   /**
    * Constructor which accepts an error message only.
    *
    * @param   message
    *          Error message.
    */
   public LockUnavailableException(String message, LockType lockType, RecordIdentifier identifier)
   {
      super(message);
      
      this.lockType = lockType;
      this.identifier = identifier;
   }
   
   /**
    * Get the requested lock type.
    * 
    * @return  Requested lock type.
    */
   public LockType getLockType()
   {
      return lockType;
   }
   
   /**
    * Get the record identifier.
    * 
    * @return  Record identifier.
    */
   public RecordIdentifier getIdentifier()
   {
      return identifier;
   }
   
   /**
    * Override the ancestor class' implementation to capture stack trace data only if the lock
    * manager is configured for debug logging. In most cases, we do not actually use the stack
    * trace information when we catch this exception; however, we retain the capability for
    * debugging purposes. For normal production use, where a common use case is to poll for a
    * record lock using the NO-WAIT option in a loop, this implementation does not pay the
    * performance penalty of collecting the data, which is an expensive operation.
    * 
    * @return  This exception instance.
    */
   @Override
   public final Throwable fillInStackTrace()
   {
      if (captureStack)
      {
         return super.fillInStackTrace();
      }
      
      return this;
   }
}