/*
** Module   : LockTableUpdater.java
** Abstract : Manages metadata lock records for a particular, primary database
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20131101 Created initial version.
** 002 ECF 20140116 Use session ID as user number (id) for lock events.
** 003 ECF 20140213 Change query strings to use JPA-style query substitution placeholders.
** 004 ECF 20140509 Fix HQL_FILENUM query string to use new computed column prefix.
** 005 ECF 20140509 Encoded more maintainable version of previous fix 004 at OM's suggestion.
** 006 ECF 20140918 Only roll back in process event if in a transaction.
** 007 ECF 20151203 Fixed HQL_FILENUM query, which was searching on historical table name, while
**                  the substitution parameter provided by the record lock event is the converted
**                  table name.
** 008 ECF 20160228 Name the update worker thread.
** 009 ECF 20160421 Replaced expensive, sequence-based primary key request with simple counter.
** 010 ECF 20160514 Replaced expensive HQL queries with more efficient SQL queries.
** 011 ECF 20160609 SQL performance improvement.
** 012 ECF 20160822 Fixed failing lock metadata updates.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.meta;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;
import com.goldencode.proxy.*;

/**
 * A {@link com.goldencode.p2j.persist.lock.LockListener lock listener implementation} which
 * receives record lock events and updates the metadata lock table accordingly. This involves the
 * creation, lookup, and deletion of metadata lock records. To avoid impacting the performance of
 * the current session's database operations, most of this work is done within a separate,
 * dedicated, server thread.
 */
public final class LockTableUpdater
implements LockListener
{
   /** Logger */
   private static final Logger log = LogHelper.getLogger(LockTableUpdater.class.getName());
   
   /** Worker which processes record lock events in a dedicated server thread */
   private final UpdateWorker worker;
   
   /** Numeric ID of the user in the current session */
   private final ContextLocal<integer> sessionID = new ContextLocal<integer>()
   {
      /** Security manager */
      private final SecurityManager securityManager;
      
      /** List of reclaimed user numbers */
      private final List<integer> reclaimedSessionIDs;
      
      /** Highest allocated session ID to date (may no longer be active) */
      private int highSessionID = 0;
      
      // Constructor for this anonymous inner class
      {
         securityManager = SecurityManager.getInstance();
         reclaimedSessionIDs = (securityManager == null ? new LinkedList<>() : null);
      }
      
      @Override
      protected integer initialValue()
      {
         integer sessionID;
         
         if (securityManager != null)
         {
            sessionID = new integer(securityManager.getSessionId());
         }
         else
         {
            // fallback method, though we should always have a security manager in server mode
            synchronized (this)
            {
               if (reclaimedSessionIDs.size() > 0)
               {
                  sessionID = reclaimedSessionIDs.remove(0);
               }
               else
               {
                  sessionID = new integer(++highSessionID);
               }
            }
         }
         
         return sessionID;
      }
      
      protected void cleanup(integer value)
      {
         if (securityManager == null)
         {
            synchronized (this)
            {
               reclaimedSessionIDs.add(value);
            }
         }
      }
   };
   
   /**
    * Constructor.
    * 
    * @param   metaDatabase
    *          Metadata database.
    */
   public LockTableUpdater(Database metaDatabase)
   {
      this.worker = new UpdateWorker(PersistenceFactory.getInstance(metaDatabase));
   }
   
   /**
    * This method is called within a user context to force any pending lock events to be
    * processed. It guarantees that any events that are pending in the queue at the time this
    * method is invoked will have been processed by the time this method returns. However,
    * other events may be added by other user contexts in the meantime, and these are not
    * guaranteed to have been processed.
    */
   public void flushQueue()
   {
      worker.forceFlush();
   }
   
   /**
    * Stop processing record lock events for the associated, primary database. This will terminate
    * the update worker and its processing thread.
    */
   public void terminate()
   {
      worker.terminate();
   }
   
   /**
    * Method which is called by a {@link LockManager} in response to a lock event. Does some
    * preliminary processing and sanity checking, then queues the event for further processing on
    * a dedicated thread.
    * 
    * @param   event
    *          Object which describes the lock event.
    */
   public void lockChanged(RecordLockEvent event)
   {
      LockType oldType = event.getOldType();
      LockType newType = event.getNewType();
      
      // if old and new lock types are both NONE, nothing to do, but log a warning, because this
      // suggests a problem
      if (LockType.NONE.equals(oldType) && LockType.NONE.equals(newType))
      {
         if (log.isLoggable(Level.WARNING))
         {
            log.log(Level.WARNING, "Unexpected lock event:  " + event);
            
            return;
         }
      }
      
      event.setUserid(sessionID.get());
      
      worker.addEvent(event);
   }
   
   /**
    * A worker containing a queue to which record lock events are written from multiple user
    * threads/contexts, and from which those events are read in a dedicated, daemon thread. The
    * idea is to offload the potentially more time-consuming work of querying and updating the
    * metadata database from a dedicated thread, to minimize any performance penalty of this work
    * on user sessions. Since this naturally introduces the possibility of the lock metadata to
    * get out of sync with the actual lock states in the lock manager, a synchronization mechanism
    * is used to ensure any pending updates are flushed to the database before any query against
    * this table is executed. This way, only those user sessions actually querying the lock
    * metadata are affected by any lag in updating the metadata from record lock events.
    * <p>
    * Note that some analysis of the lock event has to occur in the user session in which the
    * lock event occurred, but as much of the work as possible is deferred until the separate
    * thread processes the events.
    */
   private static class UpdateWorker
   implements Runnable
   {
      /** Metadata lock table name */
      private static final String LOCK_TABLE = "meta_lock";
      
      /** SQL query to find a lock record by lock recid and user */
      private static final String SQL_LOCK =
         "select m.id from " + LOCK_TABLE + " m where m.lock_recid = ?1 and m.lock_usr = ?2";
      
      /** SQL query to find a file number by converted table name (used by lock manager) */
      private static final String SQL_FILENUM =
         "select m.file_number from meta_file m where m.user_misc = ?1";
      
      /** MetaLockImpl DMO implementation class */
      private static final Class<?> dmoClass;
      
      /** MetaLockImpl DMO entity name */
      private static final String dmoEntity;
      
      /** Map of <code>MinimalLock</code> methods to <code>MetaLockImpl</code> methods */
      private static final Map<Method, Method> methodMap = new HashMap<>();
      
      static
      {
         dmoEntity = String.format("%s._meta.impl.MetaLockImpl", DMOIndex.getBasePackage());
         try
         {
            dmoClass = Class.forName(dmoEntity);
            
            // map MinimalLock methods to corresponding dmoClass methods; if we try to use
            // MinimalLock methods with a MetaLockImpl object, we will raise
            // IllegalArgumentException during Method.invoke(), because our object will not be an
            // instance of MinimalLock
            for (Method ifcMeth : MinimalLock.class.getMethods())
            {
               Method dmoMeth = dmoClass.getMethod(ifcMeth.getName(),
                                                   ifcMeth.getParameterTypes());
               methodMap.put(ifcMeth, dmoMeth);
            }
         }
         catch (ClassNotFoundException | NoSuchMethodException exc)
         {
            throw new RuntimeException("Error preparing MetaLock implementation class", exc);
         }
      }
      
      /** Event queue to which many threads can write and from which only one thread reads */
      private final List<RecordLockEvent> eventQueue = new LinkedList<>();
      
      /** Persistence helper object for metadata database */
      private final Persistence persistence;
      
      /** Proxy object for the current metadata lock DMO */
      private final MinimalLock proxy;
      
      /** Invocation handler which maps proxy method calls to underlying DMO methods */
      private final Handler handler = new Handler();
      
      /** Update worker thread */
      private final AssociatedThread workThread;
      
      /** Cache of converted table names to file numbers */
      private final Map<String, Integer> fileNums = new HashMap<>();
      
      /** Termination flag */
      private boolean done = false;
      
      /** Lock needed to read from queue; must be obtained first if write lock is needed also */
      private Object readLock = new Object();
      
      /** Lock needed to write to queue; must be obtained second if read lock is needed also */
      private Object writeLock = new Object();
      
      /** Unique and sequential ID of the last event added to the queue */
      private long lastAddedID = -1L;
      
      /** Unique and sequential ID of the last event read from the queue and processed */
      private long lastProcessedID = -1L;
      
      /** Next available primary key for a new lock metadata record */
      private long nextPrimaryKey = 1L;
      
      /**
       * Constructor.
       * 
       * @param   persistence
       *          Persistence helper object for metadata database.
       */
      UpdateWorker(Persistence persistence)
      {
         super();
         
         this.persistence = persistence;
         this.proxy = ProxyFactory.getProxy(dmoClass,
                                            new Class<?>[] { MinimalLock.class },
                                            handler);
         
         workThread = new AssociatedThread(this);
         workThread.setDaemon(true);
         workThread.setName("Lock metadata update thread");
         workThread.start();
      }
      
      /**
       * This method executes in the worker thread, reading record lock events from the queue when
       * they become available, and updating the metadata database accordingly.
       */
      @Override
      public void run()
      {
         while (true)
         {
            // check if we should terminate event processing
            synchronized (writeLock)
            {
               if (done)
               {
                  break;
               }
            }
            
            // process any pending events
            processPendingEvents();
            
            // wait for new events to be added to the queue if it is still empty
            synchronized (writeLock)
            {
               while (!done && eventQueue.isEmpty())
               {
                  try
                  {
                     writeLock.wait();
                  }
                  catch (InterruptedException exc)
                  {
                     // may be spurious, ignore;  if interrupted due to termination, we will
                     // detect this in the termination check above
                  }
               }
            }
         }
      }
      
      /**
       * Add a record lock event to the queue, locking out other write and read operations until
       * the queue is updated. Notify a the worker thread when the new event has been added.
       * 
       * @param   event
       *          Record lock event to be added.
       */
      void addEvent(RecordLockEvent event)
      {
         synchronized (writeLock)
         {
            eventQueue.add(event);
            // DBG: System.out.println("-> " + event);
            lastAddedID = event.getEventID();
            writeLock.notify();
         }
      }
      
      /**
       * This method is called within a user context to force any pending lock events to be
       * processed. It guarantees that any events that are pending in the queue at the time this
       * method is invoked will have been processed by the time this method returns. However,
       * other events may be added by other user contexts in the meantime, and these are not
       * guaranteed to have been processed.
       */
      void forceFlush()
      {
         long markerID = -1L;
         long lastProcessed = -1L;
         
         synchronized (writeLock)
         {
            markerID = lastAddedID;
            lastProcessed = lastProcessedID;
         }
         
         synchronized (readLock)
         {
            while (lastProcessed < markerID)
            {
               try
               {
                  readLock.wait();
               }
               catch (InterruptedException exc)
               {
                  if (done)
                  {
                     break;
                  }
               }
               
               synchronized (writeLock)
               {
                  lastProcessed = lastProcessedID;
               }
            }
         }
      }
      
      /**
       * Set a flag to terminate the queue, removing any pending events which have not yet been
       * processed. The worker thread is interrupted.
       * <p>
       * This method should invoked when the primary database associated with the enclosing lock
       * table updater is deregistered, as part of cleaning up the associated metadata database.
       */
      void terminate()
      {
         synchronized (writeLock)
         {
            done = true;
            eventQueue.clear();
            workThread.interrupt();
         }
      }
      
      /**
       * Process all pending record lock events in the queue. Events are read and removed from the
       * queue synchronously, such that no new events may be added during the read/remove
       * operation. However, they are then processed asynchronously, so new events may be added
       * while lock metadata is being updated.
       */
      private void processPendingEvents()
      {
         RecordLockEvent[] pending = null;
         
         synchronized (readLock)
         {
            // collect all pending events (if any)
            synchronized (writeLock)
            {
               int size = eventQueue.size();
               if (size > 0)
               {
                  pending = new RecordLockEvent[size];
                  Iterator<RecordLockEvent> iter = eventQueue.iterator();
                  for (int i = 0; iter.hasNext(); i++)
                  {
                     pending[i] = iter.next();
                     iter.remove();
                  }
               }
            }
            
            // process pending events (if any); note that new events can be added to the queue
            // while we are doing this
            if (pending != null)
            {
               int len = pending.length;
               for (int i = 0; i < len; i++)
               {
                  processEvent(pending[i]);
               }
               
               synchronized (writeLock)
               {
                  lastProcessedID = pending[len - 1].getEventID();
               }
            }
            
            // give any threads waiting in forceFlush a chance to proceed
            readLock.notifyAll();
         }
      }
      
      /**
       * Process an individual record lock event, updating the metadata lock table as appropriate.
       * If the lock event represents a newly acquired lock, create a new lock record, otherwise
       * find the existing lock record associated with the event, which must represent a change of
       * the state of an existing lock. If that change of state represents a lock release, delete
       * the associated lock record from the lock table.
       * 
       * @param   event
       *          Record lock event describing a new lock or a change in state of an existing
       *          lock.
       */
      private void processEvent(RecordLockEvent event)
      {
         Persistable dmo = null;
         boolean commit = false;
         
         try
         {
            // DBG: System.out.println("<- " + event);
            commit = persistence.beginTransaction();
            
            LockType oldType = event.getOldType();
            LockType newType = event.getNewType();
            
            if (LockType.NONE.equals(oldType))
            {
               dmo = createRecord(event);
            }
            else
            {
               dmo = findRecord(event);
            }
            
            if (LockType.NONE.equals(newType))
            {
               persistence.delete(dmo);
            }
            else
            {
               String code = (newType.isExclusive() ? "X" : "S"); // TODO:  get this right
               proxy.setLockFlags(new character(code));
               boolean isTable = (event.getIdentifier() instanceof TableIdentifier);
               proxy.setLockType(new character(isTable ? "TAB" : "REC"));
               persistence.save(dmo, dmo.getId());
            }
            
            persistence.flush();
            
            if (commit)
            {
               commit = false;   // to avoid double-rollback in catch block if commit fails
               persistence.commit();
            }
            
            if (log.isLoggable(Level.FINEST))
            {
               log.log(Level.FINEST, event.toString());
            }
         }
         catch (Exception exc)
         {
            try
            {
               if (commit)
               {
                  persistence.rollback();
               }
            }
            catch (PersistenceException pe)
            {
               // error already logged in finally block
            }
            finally
            {
               if (log.isLoggable(Level.SEVERE))
               {
                  String msg = String.format("Error updating _lock table for %s (%s)",
                                             persistence.getDatabase().toString(),
                                             event.toString());
                  log.log(Level.SEVERE, msg, exc);
               }
            }
         }
         finally
         {
            if (dmo != null)
            {
               RecordIdentifier metaIdent = new RecordIdentifier(dmoClass.getName(), dmo.getId());
               try
               {
                  persistence.lock(LockType.NONE, metaIdent, true);
               }
               catch (LockUnavailableException exc)
               {
                  // won't be thrown from lock release
               }
            }
         }
      }
      
      /**
       * Extract an integral lock record ID from the lock event. This will be a positive value
       * for records and a negative value for tables.
       * 
       * @param   event
       *          A record lock event.
       * 
       * @return  See above.
       */
      private long getLockRecordID(RecordLockEvent event)
      throws PersistenceException
      {
         Serializable recordID = event.getIdentifier().getRecordID();
         long recID = (recordID instanceof Long ? (Long) recordID : -getFileNum(event));
         
         return recID;
      }
      
      /**
       * Find the metadata lock record associated with the user ID and lock record ID specified by
       * the given record lock event.
       * 
       * @param   event
       *          Record lock event.
       * 
       * @return  Metadata lock record, if found.
       * 
       * @throws  PersistenceException
       *          if a lock record cannot be found for the given event, or if there is an error
       *          executing the query.
       */
      private Persistable findRecord(RecordLockEvent event)
      throws PersistenceException
      {
         Long lockRec = getLockRecordID(event);
         Integer lockUsr = event.getUserid().intValue();
         Object[] args = new Object[] { lockRec, lockUsr };
         
         Long id = (Long) persistence.getSingleSQLResult(SQL_LOCK, args, dmoEntity, false);
         Persistable dmo = null;
         
         if (id != null)
         {
            RecordIdentifier metaIdent = new RecordIdentifier(dmoClass.getName(), id);
            persistence.lock(LockType.EXCLUSIVE, metaIdent, true);
            dmo = persistence.quickLoad(metaIdent);
         }
         
         if (dmo == null)
         {
            throw new PersistenceException("No lock record found for lock event:  " + event);
         }
         
         handler.setDelegate(dmo);
         
         return dmo;
      }
      
      /**
       * Create a metadata lock record for the given record lock event.
       * 
       * @param   event
       *          Event which specifies information to be stored in the lock record.
       * 
       * @return  New lock record.
       * 
       * @throws  PersistenceException
       *          if there is any error querying related metadata, or if there is a reflection
       *          error instantiating the DMO.
       */
      private Persistable createRecord(RecordLockEvent event)
      throws PersistenceException
      {
         Persistable dmo = null;
         
         try
         {
            // create new DMO instance and back the proxy with it
            dmo = (Persistable) dmoClass.newInstance();
            handler.setDelegate(dmo);
            
            // assign primary key
            Long key = nextPrimaryKey++;
            dmo.setId(key);
            
            // get file number for table
            Integer fileNum = getFileNum(event);
            
            // lock meta record for writing
            RecordIdentifier metaIdent = new RecordIdentifier(dmoClass.getName(), dmo.getId());
            persistence.lock(LockType.EXCLUSIVE, metaIdent, true);
            
            // set user name and user id
            String subject = event.getLocker().getSubject();
            character userName = new character(subject);
            proxy.setLockName(userName);
            proxy.setLockUsr(event.getUserid());
            
            // assign other properties that will not change with new events
            proxy.setLockId(new int64(key));    // re-use primary key
            proxy.setLockTable(new integer(fileNum));
            proxy.setLockRecid(new int64(getLockRecordID(event)));
            proxy.setLockName(new character(event.getLocker().getSubject()));
         }
         catch (InstantiationException | IllegalAccessException exc)
         {
            String msg = String.format("Error creating meta lock record for event %s",
                                       event.toString());
            
            throw new PersistenceException(msg, exc);
         }
         
         return dmo;
      }
      
      /**
       * Get the file number associated with a lock event.
       * 
       * @param   event
       *          Record lock event.
       * 
       * @return  File number for the table associated with the event.
       */
      private Integer getFileNum(RecordLockEvent event)
      throws PersistenceException
      {
         RecordIdentifier ident = event.getIdentifier();
         String table = ident.getTable();
         Integer fileNum = fileNums.get(table);
         if (fileNum == null)
         {
            // converted table name is stored in MetaFile.userMisc
            Object[] args = new Object[] { table };
            fileNum = (Integer) persistence.getSingleSQLResult(SQL_FILENUM, args, null, true);
            
            if (fileNum == null)
            {
               String msg = String.format("Table '%s' not found in metadata [%s]",
                                          table,
                                          event.toString());
               throw new PersistenceException(msg);
            }
            
            // cache this to avoid future query by userMisc, which is not indexed
            fileNums.put(table, fileNum);
         }
         
         return fileNum;
      }
      
      /**
       * Invocation handler which delegates calls to the {@link MinimalLock} proxy methods to the
       * backing DMO instance.
       */
      private static class Handler
      implements InvocationHandler
      {
         /** DMO delegate whose methods ultimately are invoked */
         private Object delegate = null;
         
         /**
          * Process method invocations on the {@link MinimalLock} proxy by delegating them to the
          * backing DMO instance.
          * 
          * @param   proxy
          *          Proxy on which the enclosing class invokes methods.
          * @param   method
          *          Method which was invoked.
          * @param   args
          *          Method arguments.
          */
         @Override
         public Object invoke(Object proxy, Method method, Object[] args)
         throws Throwable
         {
            Method dmoMethod = methodMap.get(method);
            
            return dmoMethod.invoke(delegate, args);
         }
         
         /**
          * Set the backing DMO delegate instance.
          * 
          * @param   delegate
          *          DMO whose methods will be invoked.
          */
         void setDelegate(Object delegate)
         {
            this.delegate = delegate;
         }
      }
   }
   
   /**
    * Interface which defines the minimally required methods of the <code>MetaLock</code>
    * interface. This interface defines the proxy used to create and manipulate metadata lock
    * DMO instances. The DMO's methods cannot be invoked directly from code here, since the
    * <code>MetaLock</code> interface resides in an application-specific package, which is not
    * known at compile time.
    */
   public interface MinimalLock
   extends Persistable
   {
      /**
       * Getter: lockId
       * 
       * @return  lockId
       */
      public int64 getLockId();
      
      /**
       * Setter: lockId
       * 
       * @param   lockId
       *          lockId
       */
      public void setLockId(NumberType lockId);
      
      /**
       * Getter: Usr
       * 
       * @return  Usr
       */
      public integer getLockUsr();
      
      /**
       * Setter: Usr
       * 
       * @param   lockUsr
       *          Usr
       */
      public void setLockUsr(NumberType lockUsr);
      
      /**
       * Getter: Name
       * 
       * @return  Name
       */
      public character getLockName();
      
      /**
       * Setter: Name
       * 
       * @param   lockName
       *          Name
       */
      public void setLockName(Text lockName);
      
      /**
       * Getter: Type
       * 
       * @return  Type
       */
      public character getLockType();
      
      /**
       * Setter: Type
       * 
       * @param   lockType
       *          Type
       */
      public void setLockType(Text lockType);
      
      /**
       * Getter: Table
       * 
       * @return  Table
       */
      public integer getLockTable();
      
      /**
       * Setter: Table
       * 
       * @param   lockTable
       *          Table
       */
      public void setLockTable(NumberType lockTable);
      
      /**
       * Getter: RECID
       * 
       * @return  RECID
       */
      public int64 getLockRecid();
      
      /**
       * Setter: RECID
       * 
       * @param   lockRecid
       *          RECID
       */
      public void setLockRecid(NumberType lockRecid);
      
      /**
       * Getter: Flags
       * 
       * @return  Flags
       */
      public character getLockFlags();
      
      /**
       * Setter: Flags
       * 
       * @param   lockFlags
       *          Flags
       */
      public void setLockFlags(Text lockFlags);
   }
}