/*
** Module   : Functions.cs
** Abstract : Delegates the CLR decorated methods to overloaded functions from assembly converted
** from java p2jpl.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 OM  20140313 Created initial version with full support of existing functions from
**                  Functions.java at this time.
** 002 OM  20140725 Fixed toInt_s method name. Return type nullable.
** 003 OM  20160908 Added trimWS_ss function.
** 004 HC  20170119 Added getFWDVersion() function.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

using System;

/**
 * The role of this class is to resolve the SQL Server method overload constraint and simplify
 * the name of the functions.
 *
 * Each method of this class has as correspondent method in the original
 * <code>com.goldencode.p2j.persist.pl.Functions</code> that will call with the parameters
 * converted for wrapped in IKVM converted datatypes. The method names of the
 * original class is decorated with as list of characters, one for each parameter, each such
 * character represents the type of the parameters in the original method signature. This way,
 * there are no overloaded methods so SQL Server cannot complain about this.
 *
 * The decorations are as follows:
 * <ul>
 *    <li>s - for character / String / (string / String) / NVARCHAR(4000)
 *    <li>b - for logical / (boolean / Boolean) / (bool / Boolean) / BIT
 *    <li>i - for integer / (int / Integer) / (int / Int32) / INT
 *    <li>l - for int64 / (long / Long) / (long / Int64) / BIGINT
 *    <li>r - for decimal / (double / BigDecimal) / (double / Double) / (DOUBLE PRECISION / DECIMAL(38))
 *    <li>d - for date / Date / DateTime / DATE
 *    <li>t - for datetime / Timestamp / DateTime / DATETIME2
 * </ul>
 *
 * For each function an UDF will be defined in SQL having the same name as the method of this
 * class.
 *
 * Note that, if the original interface of the <code>Functions</code> is changed (new methods are
 * added or existing ones change their signature) the current class must be adjusted to fit those
 * changes (add/remove method or signature change).
 */
public class Functions
{
   /**
    * Private constructor. This class is never instantiated. All access is done using static
    * public methods.
    */
   private Functions()
   {
   }

   /**
    * Test whether one string begins with another case insensitively.
    *
    * @see com.goldencode.p2j.util.TextOps#begins(Text, character)
    */
   public static Boolean? begins_ss(string a, string b)
   {
      return Common.fromBoolean(com.goldencode.p2j.persist.pl.Functions.begins(a, b));
   }

   /**
    * Test whether one string begins with another, optionally taking case into account.
    *
    * @see com.goldencode.p2j.util.TextOps#begins(String, String, boolean)
    */
   public static Boolean? begins_ssb(string a, string b, Boolean caseSensitive)
   {
      return Common.fromBoolean(com.goldencode.p2j.persist.pl.Functions.begins(
                  a, b, caseSensitive));
   }

   /**
    * Returns the character value of a numeric expression which must be between 1 and 65534
    * (inclusive).
    *
    * @see com.goldencode.p2j.util.character#chr(long)
    */
   public static string chr_l(Int64 value)
   {
      return com.goldencode.p2j.persist.pl.Functions.chr(Common.toLong(value));
   }

   /**
    * Returns the character value of a numeric expression which must be between 1 and 65534
    * (inclusive).
    *
    * @see com.goldencode.p2j.util.character#chr(long)
    */
   public static string chr_i(Int32 value)
   {
      return com.goldencode.p2j.persist.pl.Functions.chr(Common.toInteger(value));
   }

   /**
    * Test for a pattern match within a string case insensitively.
    *
    * @see com.goldencode.p2j.util.TextOps#matches(Text, character, boolean)
    */
   public static Boolean? matches_ssb(string a, string b, Boolean windows)
   {
      return Common.fromBoolean(com.goldencode.p2j.persist.pl.Functions.matches(
                  a, b, Common.toBoolean(windows)));
   }
   
   /**
    * Test for a pattern match within a string, optionally taking case into account.
    *
    * @see com.goldencode.p2j.util.TextOps#matches(String, String, boolean, boolean)
    */
   public static Boolean? matches_ssbb(string a, string b, Boolean caseSensitive, Boolean windows)
   {
      return Common.fromBoolean(com.goldencode.p2j.persist.pl.Functions.matches(
                  a, b, Common.toBoolean(caseSensitive), Common.toBoolean(windows)));
   }

   /**
    * Retrieve the 1-based index at which one string is found within another, or 0 if the string
    * is not found.
    *
    * @see com.goldencode.p2j.util.TextOps#indexOf(Text, Text)
    */
   public static Int64? indexOf_ss(string source, string target)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.indexOf(source, target));
   }

   /**
    * Retrieve the 1-based index at which one string is found within another, or 0 if the string
    * is not found, starting at a particular index.
    *
    * @see com.goldencode.p2j.util.TextOps#indexOf(Text, Text, NumberType)
    */
   public static Int64? indexOf_ssl(string source, string target, Int64 start)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.indexOf(
                  source, target, Common.toLong(start)));
   }

   /**
    * Retrieve the 1-based index at which one string is found within another, or 0 if the string
    * is not found, starting at a particular index.
    *
    * @see com.goldencode.p2j.util.TextOps#indexOf(Text, Text, NumberType)
    */
   public static Int64? indexOf_ssi(string source, string target, Int32 start)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.indexOf(
                  source, target, Common.toInteger(start)));
   }

   /**
    * Retrieve the 1-based index at which one string is found within another, or 0 if the string
    * is not found, starting at a particular index.
    *
    * @see com.goldencode.p2j.util.TextOps#indexOf(Text, Text, NumberType)
    */
   public static Int64? indexOf_ssr(string source, string target, Double start)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.indexOf(
                  source, target, Common.toBigDecimal(start)));
   }

   /**
    * Retrieve the 1-based index at which one string is found within another, or 0 if the string
    * is not found, starting at a particular index, optionally taking case into account.
    *
    * @see com.goldencode.p2j.util.TextOps#indexOf(java.lang.String, java.lang.String, double,
    * boolean)
    */
   public static Int64? indexOf_sslb(string source,
                                    string target,
                                    Int64 start,
                                    Boolean caseSensitive)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.indexOf(
                  source, target, Common.toLong(start), caseSensitive));
   }

   /**
    * Retrieve the 1-based index at which one string is found within another, or 0 if the string
    * is not found, starting at a particular index, optionally taking case into account.
    *
    * @see com.goldencode.p2j.util.TextOps#indexOf(java.lang.String, java.lang.String, double,
    * boolean)
    */
   public static Int64? indexOf_ssib(string source,
                                    string target,
                                    Int32 start,
                                    Boolean caseSensitive)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.indexOf(
                  source, target, Common.toInteger(start), caseSensitive));
   }

   /**
    * Retrieve the 1-based index at which one string is found within another, or 0 if the string
    * is not found, starting at a particular index, optionally taking case into account.
    *
    * @see com.goldencode.p2j.util.TextOps#indexOf(java.lang.String, java.lang.String, double,
    * boolean)
    */
   public static Int64? indexOf_ssrb(string source,
                                    string target,
                                    Double start,
                                    Boolean caseSensitive)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.indexOf(
                  source, target, Common.toBigDecimal(start), caseSensitive));
   }

   /**
    * Transform the given value into an <code>integer</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.integer#integer(java.lang.Boolean)
    */
   public static Int64? toInt_b(Boolean value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt(
                  Common.toBoolean(value)));
   }

   /**
    * Transform the given value into an <code>integer</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.integer#integer(java.lang.Number)
    */
   public static Int64? toInt_l(Int64 value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt(
                  Common.toLong(value)));
   }

   /**
    * Transform the given value into an <code>integer</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.integer#integer(java.lang.Number)
    */
   public static Int64? toInt_i(Int32 value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt(
                  Common.toInteger(value)));
   }

   /**
    * Transform the given value into an <code>integer</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.integer#integer(java.lang.Number)
    */
   public static Int64? toInt_r(Double value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt(
                  Common.toBigDecimal(value)));
   }
   
   /**
    * Transform the given value into an <code>integer</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.integer#integer(java.lang.String)
    */
   public static Int64? toInt_s(string value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt(value));
   }
   
   /**
    * Transform the given value into an <code>integer</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.date#longValue()
    */
   public static Int64? toInt_d(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt(Common.toDate(value)));
   }

   /**
    * Transform the given value into an <code>int64</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.int64#int64(java.lang.Boolean)
    */
   public static Int64? toInt64_b(Boolean value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt64(
                  Common.toBoolean(value)));
   }

   /**
    * Transform the given value into an <code>int64</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.int64#int64(java.lang.Number)
    */
   public static Int64? toInt64_l(Int64 value)
   {
      return value;
      // return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt64(
      //             Common.toLong(value)));
   }

   /**
    * Transform the given value into an <code>integer</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.int64#int64(java.lang.Number)
    */
   public static Int64? toInt64_i(Int32 value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt64(
                  Common.toInteger(value)));
   }

   /**
    * Transform the given value into an <code>int64</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.int64#int64(java.lang.Number)
    */
   public static Int64? toInt64_r(Double value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt64(
                  Common.toBigDecimal(value)));
   }

   /**
    * Transform the given value into an <code>int64</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.int64#int64(java.lang.String)
    */
   public static Int64? toInt64_s(string value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt64(value));
   }

   /**
    * Transform the given value into an <code>int64</code>, then return it as a J2SE
    * <code>Long</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.date#longValue()
    */
   public static Int64? toInt64_d(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toInt64(
                  Common.toDate(value)));
   }

   /**
    * Transform the given value into a <code>decimal</code>, then return it as a J2SE
    * <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.decimal#decimal(com.goldencode.p2j.util.logical)
    */
   public static Double? toDec_b(Boolean value)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(
                  Common.toBoolean(value)));
   }

   /**
    * Transform the given value into a <code>decimal</code>, then return it as a J2SE
    * <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.decimal#decimal(java.lang.Number)
    */
   public static Double? toDec_l(Int64 value)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(
                  Common.toLong(value)));
   }

   /**
    * Transform the given value into a <code>decimal</code>, then return it as a J2SE
    * <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.decimal#decimal(java.lang.Number)
    */
   public static Double? toDec_i(Int32 value)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(
                  Common.toInteger(value)));
   }

   /**
    * Return the given <code>BigDecimal</code> value.  Do not alter the precision.
    *
    * @param value
    *       decimal numeric value.
    */
   public static Double? toDec_r(Double value)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(
                  Common.toBigDecimal(value)));
   }

   /**
    * Return the given <code>BigDecimal</code> value after applying the given precision.  This
    * function is provided for the special case of a literal value, where the default precision
    * derived from that literal is not appropriate.
    *
    * @param value
    *       Decimal numeric value.
    * @param precision
    *       Precision to apply.
    */
   public static Double? toDec_rl(Double value, Int64 precision)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(
                  Common.toBigDecimal(value), Common.toLong(precision)));
   }

   /**
    * Return the given <code>BigDecimal</code> value after applying the given precision.  This
    * function is provided for the special case of a literal value, where the default precision
    * derived from that literal is not appropriate.
    *
    * @param value
    *       Decimal numeric value.
    * @param precision
    *       Precision to apply.
    */
   public static Double? toDec_ri(Double value, Int32 precision)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(
                  Common.toBigDecimal(value), Common.toInteger(precision)));
   }

   /**
    * Transform the given value into a <code>decimal</code>, then return it as a J2SE
    * <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.decimal#decimal(java.lang.String)
    */
   public static Double? toDec_s(string value)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(value));
   }

   /**
    * Transform the given value into a <code>decimal</code>, then return it as a J2SE
    * <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.decimal#decimal(com.goldencode.p2j.util.date)
    */
   public static Double? toDec_d(DateTime value)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.toDec(
                  Common.toDate(value)));
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(boolean)
    */
   public static string toString_b(Boolean value)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toBoolean(value));
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(boolean, String)
    */
   public static string toString_bs(Boolean value, string fmt)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toBoolean(value), fmt);
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(long)
    */
   public static string toString_l(Int64 value)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toLong(value));
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(long)
    */
   public static string toString_i(Int32 value)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toInteger(value));
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(long, String)
    */
   public static string toString_ls(Int64 value, string fmt)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toLong(value), fmt);
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(long, String)
    */
   public static string toString_is(Int32 value, string fmt)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toInteger(value), fmt);
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(BaseDataType)
    */
   public static string toString_d(DateTime value)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toDate(value));
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(BaseDataType, String)
    */
   public static string toString_ds(DateTime value, string fmt)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toDate(value), fmt);
   }

   /**
    * Transform the given datetime value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(BaseDataType)
    */
   public static string toString_t(DateTime value)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toTimestamp(value));
   }

   /**
    * Transform the given datetime value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(BaseDataType, String)
    */
   public static string toString_ts(DateTime value, string fmt)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toTimestamp(value), fmt);
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(double)
    */
   public static string toString_r(Double value)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toBigDecimal(value));
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(double, String)
    */
   public static string toString_rs(Double value, string fmt)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(Common.toBigDecimal(value), fmt);
   }

   /**
    * Simply return the given string.  Normally, the value would be converted to a
    * <code>character</code>, then back to a J2SE <code>String</code>, but this method
    * short-circuits that unnecessary, round trip.
    *
    * @param value
    *       string to be returned.
    * @return <code>value</code> parameter, unchanged.
    */
   public static string toString_s(string value)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(value);
   }

   /**
    * Transform the given value into an <code>character</code>, then return it as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.character#valueOf(String, String)
    */
   public static string toString_ss(string value, string fmt)
   {
      return com.goldencode.p2j.persist.pl.Functions.toString(value, fmt);
   }

   /**
    * Returns the Rowid representation of a string value
    *
    * @see com.goldencode.p2j.util.rowid#parseRowidString
    */
   public static Int64? toRowid_s(string text)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.toRowid(text));
   }

   /**
    * Concatenate two strings using P2J character's default behavior.  Return the result as a J2SE
    * <code>String</code>, or <code>null</code> if unknown.
    *
    * @see com.goldencode.p2j.util.TextOps#concat
    */
   public static string concat_ss(string a, string b)
   {
      return com.goldencode.p2j.persist.pl.Functions.concat(a, b);
   }

   /**
    * Extract a substring from a string, starting at a specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_sl(string text, Int64 pos)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text, Common.toLong(pos));
   }

   /**
    * Extract a substring from a string, starting at a specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_si(string text, Int32 pos)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text, Common.toInteger(pos));
   }

   /**
    * Extract a substring of a specific length from a string, starting at a specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_sll(string text, Int64 pos, Int64 len)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toLong(pos),
                                                                 Common.toLong(len));
   }

   /**
    * Extract a substring of a specific length from a string, starting at a specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_sli(string text, Int64 pos, Int32 len)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toLong(pos),
                                                                 Common.toInteger(len));
   }

   /**
    * Extract a substring of a specific length from a string, starting at a specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_sil(string text, Int32 pos, Int64 len)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toInteger(pos),
                                                                 Common.toLong(len));
   }

   /**
    * Extract a substring of a specific length from a string, starting at a specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_sii(string text, Int32 pos, Int32 len)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toInteger(pos),
                                                                 Common.toInteger(len));
   }

   /**
    * Extract a substring of a specific length and Progress type from a string, starting at a
    * specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_slls(string text, Int64 pos, Int64 len, string type)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toLong(pos),
                                                                 Common.toLong(len),
                                                                 type);
   }

   /**
    * Extract a substring of a specific length and Progress type from a string, starting at a
    * specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_slis(string text, Int64 pos, Int32 len, string type)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toLong(pos),
                                                                 Common.toInteger(len),
                                                                 type);
   }

   /**
    * Extract a substring of a specific length and Progress type from a string, starting at a
    * specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_sils(string text, Int32 pos, Int64 len, string type)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toInteger(pos),
                                                                 Common.toLong(len),
                                                                 type);
   }

   /**
    * Extract a substring of a specific length and Progress type from a string, starting at a
    * specific index.
    *
    * @see com.goldencode.p2j.util.TextOps#substring
    */
   public static string substringOf_siis(string text, Int32 pos, Int32 len, string type)
   {
      return com.goldencode.p2j.persist.pl.Functions.substringOf(text,
                                                                 Common.toInteger(pos),
                                                                 Common.toInteger(len),
                                                                 type);
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Double? maximum_rr(Double var1, Double var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toBigDecimal(var1), Common.toBigDecimal(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Double? maximum_lr(Int64 var1, Double var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toLong(var1), Common.toBigDecimal(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Double? maximum_ir(Int32 var1, Double var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toInteger(var1), Common.toBigDecimal(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Double? maximum_rl(Double var1, Int64 var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toBigDecimal(var1), Common.toLong(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Double? maximum_ri(Double var1, Int32 var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toBigDecimal(var1), Common.toInteger(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Int64? maximum_ll(Int64 var1, Int64 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toLong(var1), Common.toLong(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Int64? maximum_li(Int64 var1, Int32 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toLong(var1), Common.toInteger(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Int64? maximum_il(Int32 var1, Int64 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toInteger(var1), Common.toLong(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static Int64? maximum_ii(Int32 var1, Int32 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.maximum(
                  Common.toInteger(var1), Common.toInteger(var2)));
   }

   /**
    * Returns the maximum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static string maximum_ss(string var1, string var2)
   {
      return com.goldencode.p2j.persist.pl.Functions.maximum(var1, var2);
   }

   /**
    * Returns the maximum from a list of date values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static DateTime? maximum_dd(DateTime var1, DateTime var2)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.maximum(Common.toDate(var1),
                                                                             Common.toDate(var2)));
   }

   /**
    * Returns the maximum from a list of datetime values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The maximum of the given values.
    */
   public static DateTime? maximum_tt(DateTime var1, DateTime var2)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.maximum(Common.toTimestamp(
            var1), Common.toTimestamp(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Double? minimum_rr(Double var1, Double var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toBigDecimal(var1), Common.toBigDecimal(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Double? minimum_lr(Int64 var1, Double var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toLong(var1), Common.toBigDecimal(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Double? minimum_ir(Int32 var1, Double var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toInteger(var1), Common.toBigDecimal(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Double? minimum_rl(Double var1, Int64 var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toBigDecimal(var1), Common.toLong(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Double? minimum_ri(Double var1, Int32 var2)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toBigDecimal(var1), Common.toInteger(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Int64? minimum_ll(Int64 var1, Int64 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toLong(var1), Common.toLong(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Int64? minimum_li(Int64 var1, Int32 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toLong(var1), Common.toInteger(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Int64? minimum_il(Int32 var1, Int64 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toInteger(var1), Common.toLong(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static Int64? minimum_ii(Int32 var1, Int32 var2)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toInteger(var1), Common.toInteger(var2)));
   }

   /**
    * Returns the minimum from a list of values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static string minimum_ss(string var1, string var2)
   {
      return com.goldencode.p2j.persist.pl.Functions.minimum(var1, var2);
   }

   /**
    * Returns the minimum from a list of date values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static DateTime? minimum_dd(DateTime var1, DateTime var2)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toDate(var1), Common.toDate(var2)));
   }

   /**
    * Returns the minimum from a list of datetime values.
    *
    * @param var1
    *       Value 1.
    * @param var2
    *       Value 2.
    * 
    * @return The minimum of the given values.
    */
   public static DateTime? minimum_tt(DateTime var1, DateTime var2)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.minimum(
                  Common.toTimestamp(var1), Common.toTimestamp(var2)));
   }

   /**
    * Returns logical representation of a char variable from a format of type "true/false".
    *
    * @param var
    *       The variable we want to match against the logical format
    *
    * @return The logical representation of a char variable from a format of type "true/false"
    */
   public static Boolean? toLogical_s(string var)
   {
      return Common.fromBoolean(com.goldencode.p2j.persist.pl.Functions.toLogical(var));
   }
   
   /**
    * Returns logical representation of a char variable from a format of type "true/false".
    *
    * @param var
    *       The variable we want to match against the logical format
    * @param format
    *       The format of form "leftLogical/rightLogical" in which leftLogical will represent
    *       <code>true</code> and rightLogical represents <code>false</code>
    *
    * @return The logical representation of a char variable from a format of type "true/false"
    */
   public static Boolean? toLogical_ss(string var, string format)
   {
      return Common.fromBoolean(com.goldencode.p2j.persist.pl.Functions.toLogical(var, format));
   }

   /**
    * Replaces the given values into the destination.
    *
    * @see com.goldencode.p2j.util.TextOps#replaceAll
    */
   public static string replace_sss(string source, string toReplace, string replacement)
   {
      return com.goldencode.p2j.persist.pl.Functions.replace(source, toReplace, replacement);
   }

   /**
    * Trim white space from the left side of text.
    *
    * @see com.goldencode.p2j.util.TextOps#leftTrim
    */
   public static string ltrimWS_s(string text)
   {
      return com.goldencode.p2j.persist.pl.Functions.ltrimWS(text);
   }
   
   /**
    * Trim the specified characters from the left side of text.
    *
    * @see  com.goldencode.p2j.util.TextOps#leftTrim(String, String)
    */
   public static string ltrimWS_ss(string text, string list)
   {
      return com.goldencode.p2j.persist.pl.Functions.ltrimWS(text, list);
   }

   /**
    * Trim white space from the left and right side of text.
    *
    * @see com.goldencode.p2j.util.TextOps#trim
    */
   public static string trimWS_s(string text)
   {
      return com.goldencode.p2j.persist.pl.Functions.trimWS(text);
   }
   
   /**
    * Trim special characters from the left and right side of text.
    *
    * @see com.goldencode.p2j.util.TextOps#trim
    */
   public static string trimWS_ss(string text, string trimChars)
   {
      return com.goldencode.p2j.persist.pl.Functions.trimWS(text, trimChars);
   }
   
   /**
    * Returns the uppercase representation of the string
    * Deprecated! This UDF was replaced by native 'upper' function. 
    *
    * @see com.goldencode.p2j.util.TextOps#toUpperCase(String, boolean)
    */
   public static string caps_s(string text)
   {
      return com.goldencode.p2j.persist.pl.Functions.caps(text);
   }

   /**
    * Retrieve the number of entries in the variable.
    *
    * @see com.goldencode.p2j.util.TextOps#numEntries(Text)
    */
   public static Int64? numEntries_s(string value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.numEntries(value));
   }

   /**
    * Retrieve the number of entries in the variable.
    *
    * @see com.goldencode.p2j.util.TextOps#numEntries(Text, String)
    */
   public static Int64? numEntries_ss(string value, string separator)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.numEntries(
                  value, separator));
   }

   /**
    * Get the length of a string.
    *
    * @see com.goldencode.p2j.util.TextOps#length(Text)
    */
   public static Int64? lengthOf_s(string text)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.lengthOf(text));
   }

   /**
    * Get the length of a string.
    *
    * @see com.goldencode.p2j.util.TextOps#length(Text)
    */
   public static Int64? lengthOf_ss(string text, string type)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.lengthOf(text, type));
   }

   /**
    * Extract a string from a comma-delimited list represented by another string, given its index
    * within the list.
    *
    * @see com.goldencode.p2j.util.TextOps#entry
    */
   public static string entry_ls(Int64 i, string list)
   {
      return com.goldencode.p2j.persist.pl.Functions.entry(Common.toLong(i), list);
   }

   /**
    * Extract a string from a comma-delimited list represented by another string, given its index
    * within the list.
    *
    * @see com.goldencode.p2j.util.TextOps#entry
    */
   public static string entry_is(Int32 i, string list)
   {
      return com.goldencode.p2j.persist.pl.Functions.entry(Common.toInteger(i), list);
   }

   /**
    * Extract a string from a delimited list represented by another string, given its index within
    * the list and the delimiter.
    *
    * @see com.goldencode.p2j.util.TextOps#entry
    */
   public static string entry_lss(Int64 i, string list, string delimiter)
   {
      return com.goldencode.p2j.persist.pl.Functions.entry(Common.toLong(i), list, delimiter);
   }

   /**
    * Extract a string from a delimited list represented by another string, given its index within
    * the list and the delimiter.
    *
    * @see com.goldencode.p2j.util.TextOps#entry
    */
   public static string entry_iss(Int32 i, string list, string delimiter)
   {
      return com.goldencode.p2j.persist.pl.Functions.entry(Common.toInteger(i), list, delimiter);
   }

   /**
    * Look up the index of a string which represents an element in a comma-delimited list
    * represented by another string.
    *
    * @see com.goldencode.p2j.util.TextOps#lookup(character, Text)
    */
   public static Int64? lookup_ss(string entry, string list)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.lookup(entry, list));
   }

   /**
    * Look up the index of a string which represents an element in a delimited list represented by
    * another string, using a specific delimiter.
    *
    * @see com.goldencode.p2j.util.TextOps#lookup(com.goldencode.p2j.util.character,
    * com.goldencode.p2j.util.Text, com.goldencode.p2j.util.character)
    */
   public static Int64? lookup_sss(string entry, string list, string delimiter)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.lookup(
                  entry, list, delimiter));
   }

   /**
    * Look up the index of a string which represents an element in a delimited list represented by
    * another string, using a specific delimiter, and optionally taking case into account.
    *
    * @see com.goldencode.p2j.util.TextOps#lookup(java.lang.String, java.lang.String, char,
    * boolean)
    */
   public static Int64? lookup_sssb(string entry,
                                   string list,
                                   string delimiter,
                                   Boolean caseSensitive)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.lookup(
                  entry, list, delimiter, Common.toBoolean(caseSensitive)));
   }

   /**
    * Determine whether a comma-separated list contains a specific item.
    *
    * @see com.goldencode.p2j.util.TextOps#matchesList(Text, character)
    */
   public static Boolean? matchesList_ss(string list, string item)
   {
      return Common.fromBoolean(com.goldencode.p2j.persist.pl.Functions.matchesList(list, item));
   }

   /**
    * Round a floating point number to 10 decimal places.
    *
    * @see com.goldencode.p2j.util.MathOps#round(com.goldencode.p2j.util.decimal)
    */
   public static Double? roundDec_r(Double value)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.roundDec(
                  Common.toBigDecimal(value)));
   }

   /**
    * Round a floating point number to a particular precision.
    *
    * @see com.goldencode.p2j.util.MathOps#round(com.goldencode.p2j.util.decimal, int)
    */
   public static Double? roundDec_rl(Double value, Int64 precision)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.roundDec(
                  Common.toBigDecimal(value), Common.toLong(precision)));
   }

   /**
    * Round a floating point number to a particular precision.
    *
    * @see com.goldencode.p2j.util.MathOps#round(com.goldencode.p2j.util.decimal, int)
    */
   public static Double? roundDec_ri(Double value, Int32 precision)
   {
      return Common.fromBigDecimal(com.goldencode.p2j.persist.pl.Functions.roundDec(
                  Common.toBigDecimal(value), Common.toInteger(precision)));
   }

   /**
    * Retrieve the date from a string variable.
    *
    * @param  text
    *         The text value from which we return the new date.
    *
    * @return The calculated date value.
    */
   [Obsolete]

   public static DateTime? toDate_s(string text)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(text));
   }

   /**
    * Retrieve the date from a string variable. The date format and windowing year must be
    * provided by the query.
    *
    * @param text
    *       The text value from which we return the new date.
    * @param fmt
    *       The format string used for date conversion.
    * @param wYear
    *       The start of the 100 year window used for conversion when year is in 2-digits.
    *
    * @return The calculated date value.
    */
   public static DateTime? toDate_ssi(string text, string fmt, Int32 wYear)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  text, fmt, Common.toInteger(wYear)));
   }

   /**
    * Retrieve the date from a string variable.
    *
    * @param value
    *       The integer representing the number of days from the Progress 4GL base data time.
    *         
    * @return The calculated date value.
    */
   public static DateTime? toDate_l(Int64 value)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toLong(value)));
   }

   /**
    * Retrieve the date from a string variable.
    *
    * @param value
    *       The integer representing the number of days from the Progress 4GL base data time.
    *         
    * @return The calculated date value.
    */
   public static DateTime? toDate_i(Int32 value)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toInteger(value)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_lll(Int64 month, Int64 day, Int64 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
            Common.toLong(month), Common.toLong(day), Common.toLong(year)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_lli(Int64 month, Int64 day, Int32 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
            Common.toLong(month), Common.toLong(day), Common.toInteger(year)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_lil(Int64 month, Int32 day, Int64 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toLong(month), Common.toInteger(day), Common.toLong(year)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_lii(Int64 month, Int32 day, Int32 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toLong(month), Common.toInteger(day), Common.toInteger(year)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_ill(Int32 month, Int64 day, Int64 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toInteger(month), Common.toLong(day), Common.toLong(year)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_ili(Int32 month, Int64 day, Int32 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toInteger(month), Common.toLong(day), Common.toInteger(year)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_iil(Int32 month, Int32 day, Int64 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toInteger(month), Common.toInteger(day), Common.toLong(year)));
   }

   /**
    * Retrieve the date from a month/day/year combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDate_iii(Int32 month, Int32 day, Int32 year)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toInteger(month), Common.toInteger(day), Common.toInteger(year)));
   }

   /**
    * Extract the date from a timestamp variable.
    *
    * @param t
    *       The timestamp.
    *
    * @return The calculated date value
    */
   public static DateTime? toDate_t(DateTime dt)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.toDate(
                  Common.toTimestamp(dt)));
   }

   /**
    * Construct a timestamp from a date variable.
    *
    * @param d
    *       The date.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDatetime_d(DateTime d)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
                  Common.toDate(d)));
   }

   /**
    * Retrieve the timestamp from a date/time combination.
    *
    * @param d
    *       The date.
    * @param mtime
    *       Times since midnight in milliseconds.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDatetime_dl(DateTime d, Int64 mtime)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
                  Common.toDate(d), Common.toLong(mtime)));
   }

   /**
    * Retrieve the timestamp from a date/time combination.
    *
    * @param d
    *       The date.
    * @param mtime
    *       Times since midnight in milliseconds.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDatetime_di(DateTime d, Int32 mtime)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
                  Common.toDate(d), Common.toInteger(mtime)));
   }

   /**
    * Retrieve the date from a month/day/year/hour/minutes combination .
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    * @param hour
    *       The hour of the day.
    * @param minutes
    *       The minutes of the hour.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDatetime_lllll(Int64 month,
                                           Int64 day,
                                           Int64 year,
                                           Int64 hour,
                                           Int64 minutes)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
               Common.toLong(month), Common.toLong(day), Common.toLong(year),
               Common.toLong(hour), Common.toLong(minutes)));
   }

   /**
    * Retrieve the date from a month/day/year/hour/minutes combination.
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    * @param hour
    *       The hour of the day.
    * @param minutes
    *       The minutes of the hour.
    *         
    * @return The calculated date value
    */
   public static DateTime? toDatetime_iiiii(Int32 month,
                                           Int32 day,
                                           Int32 year,
                                           Int32 hour,
                                           Int32 minutes)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
               Common.toInteger(month), Common.toInteger(day), Common.toInteger(year),
               Common.toInteger(hour), Common.toInteger(minutes)));
   }

   /**
    * Retrieve the timestamp from a month/day/year/hour/minutes./seconds combination.
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    * @param hour
    *       The hour of the day.
    * @param minutes
    *       The minutes of the hour.
    * @param seconds
    *       The seconds in the minute.
    *
    * @return The calculated date value
    */
   public static DateTime? toDatetime_llllll(Int64 month, Int64 day, Int64 year,
                                            Int64 hour, Int64 minutes, Int64 seconds)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
               Common.toLong(month), Common.toLong(day), Common.toLong(year),
               Common.toLong(hour), Common.toLong(minutes), Common.toLong(seconds)));
   }

   /**
    * Retrieve the timestamp from a month/day/year/hour/minutes./seconds combination.
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    * @param hour
    *       The hour of the day.
    * @param minutes
    *       The minutes of the hour.
    * @param seconds
    *       The seconds in the minute.
    *
    * @return The calculated date value
    */
   public static DateTime? toDatetime_iiiiii(Int32 month, Int32 day, Int32 year,
                                            Int32 hour, Int32 minutes, Int32 seconds)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
            Common.toInteger(month), Common.toInteger(day), Common.toInteger(year),
            Common.toInteger(hour), Common.toInteger(minutes), Common.toInteger(seconds)));
   }

   /**
    * Retrieve the timestamp from a month/day/year/hour/minutes/seconds/millis combination.
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    * @param hour
    *       The hour of the day.
    * @param minutes
    *       The minutes of the hour.
    * @param seconds
    *       The seconds in the minute.
    * @param millis
    *       The milliseconds of the second.
    *
    * @return The calculated date value
    */
   public static DateTime? toDatetime_lllllll(Int64 month,
                                             Int64 day,
                                             Int64 year,
                                             Int64 hour,
                                             Int64 minutes,
                                             Int64 seconds,
                                             Int64 millis)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
            Common.toLong(month),
            Common.toLong(day),
            Common.toLong(year),
            Common.toLong(hour),
            Common.toLong(minutes),
            Common.toLong(seconds),
            Common.toLong(millis)));
   }

   /**
    * Retrieve the timestamp from a month/day/year/hour/minutes/seconds/millis combination.
    *
    * @param month
    *       The month of the date.
    * @param day
    *       The day of the date.
    * @param year
    *       The year of the date.
    * @param hour
    *       The hour of the day.
    * @param minutes
    *       The minutes of the hour.
    * @param seconds
    *       The seconds in the minute.
    * @param millis
    *       The milliseconds of the second.
    *
    * @return The calculated date value
    */
   public static DateTime? toDatetime_iiiiiii(Int32 month,
                                             Int32 day,
                                             Int32 year,
                                             Int32 hour,
                                             Int32 minutes,
                                             Int32 seconds,
                                             Int32 millis)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
            Common.toInteger(month),
            Common.toInteger(day),
            Common.toInteger(year),
            Common.toInteger(hour),
            Common.toInteger(minutes),
            Common.toInteger(seconds),
            Common.toInteger(millis)));
   }

   /**
    * Retrieve the timestamp from a string variable.
    *
    * @param  text
    *         The text value from which we return the new date.
    *
    * @return The calculated date value.
    */
   [Obsolete]

   public static DateTime? toDatetime_s(string text)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(text));
   }

   /**
    * Retrieve the timestamp from a string variable. The date format and windowing year must be
    * provided by the query.
    *
    * @param text
    *       The text value from which we return the new date.
    * @param fmt
    *       The format string used for date conversion.
    * @param wYear
    *       The start of the 100 year window used for conversion when year is in 2-digits.
    *
    * @return The calculated date value.
    */
   public static DateTime? toDatetime_ssi(string text, string fmt, Int32 wYear)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.toDatetime(
                  text, fmt, Common.toInteger(wYear)));
   }

   /**
    * Retrieve the week day number from a datetime.
    *
    * @see com.goldencode.p2j.util.date#weekday(com.goldencode.p2j.util.date)
    */
   public static Int64? getWeekday_t(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getWeekday(
                  Common.toTimestamp(value)));
   }

   /**
    * Retrieve the year number from a datetime.
    *
    * @see com.goldencode.p2j.util.date#year(com.goldencode.p2j.util.date)
    */
   public static Int64? getYear_t(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getYear(
                  Common.toTimestamp(value)));
   }

   /**
    * Retrieve the month number (1-based) from a datetime.
    *
    * @see com.goldencode.p2j.util.date#month(com.goldencode.p2j.util.date)
    */
   public static Int64? getMonth_t(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getMonth(
                  Common.toTimestamp(value)));
   }

   /**
    * Retrieve the day of the month (1-based) from a datetime.
    *
    * @see com.goldencode.p2j.util.date#year(com.goldencode.p2j.util.date)
    */
   public static Int64? getDay_t(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getDay(
                  Common.toTimestamp(value)));
   }

   /**
    * Retrieve the week day number from a date.
    *
    * @see com.goldencode.p2j.util.date#weekday(com.goldencode.p2j.util.date)
    */
   public static Int64? getWeekday_d(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getWeekday(
                  Common.toDate(value)));
   }

   /**
    * Retrieve the year number from a date.
    *
    * @see com.goldencode.p2j.util.date#year(com.goldencode.p2j.util.date)
    */
   public static Int64? getYear_d(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getYear(
                  Common.toDate(value)));
   }

   /**
    * Retrieve the month number (1-based) from a date.
    *
    * @see com.goldencode.p2j.util.date#month(com.goldencode.p2j.util.date)
    */
   public static Int64? getMonth_d(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getMonth(
                  Common.toDate(value)));
   }

   /**
    * Retrieve the day of the month (1-based) from a date.
    *
    * @see com.goldencode.p2j.util.date#year(com.goldencode.p2j.util.date)
    */
   public static Int64? getDay_d(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getDay(
                  Common.toDate(value)));
   }

   /**
    * Retrieve the time since midnight in milliseconds.
    * <p><strong>Note:</strong>
    * <br>This function is deprecated and should not exist as it should not be invoked on
    * sql-server side where the system clock is most likely different.
    *
    * @see  com.goldencode.p2j.util.datetime#millisecondsSinceMidnight()
    *
    * @return current timezone.
    */
   [Obsolete]
   public static Int64 getMtime_()
   {
      return com.goldencode.p2j.persist.pl.Functions.getMtime().longValue();
   }

   /**
    * Retrieve the time since midnight in milliseconds.
    *
    * @see com.goldencode.p2j.util.datetime#getTime()
    */
   public static Int64? getMtime_t(DateTime value)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getMtime(
                  Common.toTimestamp(value)));
   }

   /**
    * Computes the interval between two datetimes. The value is expressed in <code>unit</code>
    * units.
    *
    * @param   d1
    *          The first date.
    * @param   d2
    *          The second date.
    * @param   unit
    *          The unit to measure the interval.
    *
    * @return The interval between the two dates.
    * @see com.goldencode.p2j.util.DateOps#interval(date, date, String)
    */
   public static Int64? getInterval_tts(DateTime d1, DateTime d2, string unit)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getInterval(
                  Common.toTimestamp(d1), Common.toTimestamp(d1), unit));
   }

   /**
    * Computes the interval between a date and a datetime. The value is expressed in
    * <code>unit</code> units.
    *
    * @param   d1
    *          The date is the 1st reference.
    * @param   d2
    *          The datetime is the 2nd reference.
    * @param   unit
    *          The unit to measure the interval.
    *
    * @return The interval between the two dates.
    * @see com.goldencode.p2j.util.DateOps#interval(date, date, String)
    */
   public static Int64? getInterval_dts(DateTime d1, DateTime d2, string unit)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getInterval(
                  Common.toDate(d1), Common.toTimestamp(d1), unit));
   }

   /**
    * Computes the interval between a datetime and a date. The value is expressed in
    * <code>unit</code> units.
    *
    * @param   d1
    *          The datetime is the 1st reference.
    * @param   d2
    *          The date is the 2nd reference.
    * @param   unit
    *          The unit to measure the interval.
    *
    * @return The interval between the two dates.
    * @see com.goldencode.p2j.util.DateOps#interval(date, date, String)
    */
   public static Int64? getInterval_tds(DateTime d1, DateTime d2, string unit)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getInterval(
                  Common.toTimestamp(d1), Common.toDate(d1), unit));
   }

   /**
    * Computes the interval between two dates. The value is expressed in <code>unit</code> units.
    *
    * @param   d1
    *          The first date.
    * @param   d2
    *          The second date.
    * @param   unit
    *          The unit to measure the interval.
    *
    * @return The interval between the two dates.
    * @see com.goldencode.p2j.util.DateOps#interval(date, date, String)
    */
   public static Int64? getInterval_dds(DateTime d1, DateTime d2, string unit)
   {
      return Common.fromLong(com.goldencode.p2j.persist.pl.Functions.getInterval(
                  Common.toDate(d1), Common.toDate(d1), unit));
   }

   /**
    * Computes a new date by adding respective number of time units.
    *
    * @param initial
    *       The first date.
    * @param amount
    *       The count of units to be added.
    * @param unit
    *       The unit to measure the interval.
    *
    * @return The date that is amount units further in time.
    * @see com.goldencode.p2j.util.DateOps#addInterval(date, long, String)
    */
   public static DateTime? addInterval_dls(DateTime initial, Int64 amount, string unit)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.addInterval(
            Common.toDate(initial), Common.toLong(amount), unit));
   }

   /**
    * Computes a new date by adding respective number of time units.
    *
    * @param initial
    *       The first date.
    * @param amount
    *       The count of units to be added.
    * @param unit
    *       The unit to measure the interval.
    *
    * @return The date that is amount units further in time.
    * @see com.goldencode.p2j.util.DateOps#addInterval(date, long, String)
    */
   public static DateTime? addInterval_dis(DateTime initial, Int32 amount, string unit)
   {
      return Common.fromDate(com.goldencode.p2j.persist.pl.Functions.addInterval(
            Common.toDate(initial), Common.toInteger(amount), unit));
   }

   /**
    * Computes a new datetime by adding respective number of time units.
    *
    * @param initial
    *       The first datetime.
    * @param amount
    *       The count of units to be added.
    * @param unit
    *       The unit to measure the interval.
    *
    * @return The datetime that is amount units further in time.
    * @see com.goldencode.p2j.util.DateOps#addInterval(date, long, String)
    */
   public static DateTime? addInterval_tls(DateTime initial, Int64 amount, string unit)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.addInterval(
            Common.toTimestamp(initial), Common.toLong(amount), unit));
   }

   /**
    * Computes a new datetime by adding respective number of time units.
    *
    * @param   initial
    *          The first datetime.
    * @param   amount
    *          The count of units to be added.
    * @param   unit
    *          The unit to measure the interval.
    *
    * @return The datetime that is amount units further in time.
    * @see com.goldencode.p2j.util.DateOps#addInterval(date, long, String)
    */
   public static DateTime? addInterval_tis(DateTime initial, Int32 amount, string unit)
   {
      return Common.fromTimestamp(com.goldencode.p2j.persist.pl.Functions.addInterval(
               Common.toTimestamp(initial), Common.toInteger(amount), unit));
   }
   
   /**
    * Returns the current timezone.
    * <p><strong>Note:</strong>
    * <br>This function is deprecated and should not exist as it should not be invoked on 
    * sql-server side where the value of <code.SESSION:TIMEZONE</code> is not accessible.
    * 
    * @see date#getDefaultOffset()
    * 
    * @return  current timezone.
    */
   public static Int64 getTimezone_()
   {
      return com.goldencode.p2j.persist.pl.Functions.getTimezone().longValue();
   }
   
   /**
    * Helper function to report the precision and scale of a numeric value.
    * 
    * @param   num
    *          Numeric value.
    * 
    * @return  String representing the precision <i>p</i> and the scale
    *          <i>s</i> in the form <code>(<i>p</i>,<i>s</i>)</code>.
    */
   public static String reportPrecisionScale_r(Double num)
   {
      return com.goldencode.p2j.persist.pl.Functions.reportPrecisionScale(
               Common.toBigDecimal(num));
   }

   /**
    * Helper function to return the FWD version string.
    *
    * @return  see above
    */
   public static String getFWDVersion_()
   {
      return com.goldencode.p2j.persist.pl.Functions.getFWDVersion();
   }
}