/*
** Module   : IdentityManagerMultiplexerImpl.java
** Abstract : Handles server-to-server requests for record identity
**            management
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 SVL 20080111 ADD  @36794  Created initial version. Handles server to
**                               server requests for record identity
**                               management.
** 002 SVL 20080901 CHG  @39650  Added reclaimKeys() function. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.remote;

import java.io.Serializable;
import com.goldencode.p2j.persist.id.IdentityManager;
import com.goldencode.p2j.persist.Database;
import com.goldencode.p2j.persist.PersistenceFactory;
import com.goldencode.p2j.persist.Persistence;
import com.goldencode.p2j.persist.PersistenceException;
import com.goldencode.p2j.net.RemoteObject;

/**
 * A singleton instance of this class services server-to-server requests for
 * database identity management.  This instance is registered as a remote
 * object and serves identity management requests for one or more databases
 * which the server authoritatively manages.
 * <p>
 * Service requests may originate from more than one external server, so this
 * implementation is threadsafe.
 *
 * @author SVL
 */
public final class IdentityManagerMultiplexerImpl
extends DatabaseMultiplexer<IdentityManager>
implements IdentityManagerMultiplexer
{
   /** Singleton instance of this class */
   private static IdentityManagerMultiplexer instance = null;

   /**
    * Initialize the identity multiplexer for this server instance. This
    * method is to be called during server bootstrap, but not before the local
    * database manager has been initialized.
    * <p>
    * This method enforces the singleton pattern.
    *
    * @throws  IllegalStateException
    *          if this method is invoked more than once.
    */
   public static synchronized void initialize()
   {
      if (instance != null)
      {
         throw new IllegalStateException(
            "Identity multiplexer already initialized for this process");
      }

      instance = new IdentityManagerMultiplexerImpl();
      RemoteObject.registerNetworkServer(IdentityManagerMultiplexer.class,
                                         instance);
   }

   /**
    * Default constructor which initializes the resources needed for this
    * multiplexer to service identity manager requests.
    */
   private IdentityManagerMultiplexerImpl()
   {
      super();
   }

   /**
    * Return the next primary key for the specified table.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    *
    * @param   table
    *          Table for which the next primary key will be returned.
    *
    * @return  The next primary key for the given table.
    *
    * @throws  PersistenceException
    *          if there is an error determining the next primary key ID.
    */
   public Serializable nextPrimaryKey(int database, String table)
   throws PersistenceException
   {
      return lookupWorker(database).nextPrimaryKey(table);
   }

   /**
    * Inform the identity manager that the given primary keys can be reused.
    *
    * @param database
    *        Value that uniquely identifies the target database on the
    *        remote server.
    *
    * @param keys
    *        Keys for reuse.
    */
   public void reclaimKeys(int database, Serializable[] keys)
   {
      lookupWorker(database).reclaimKeys(keys);
   }

   /**
    * Get the identity manager which is associated with the given database.
    *
    * @param   database
    *          Physical database with which the returned worker is associated.
    *
    * @return  Worker object to which service requests associated with the
    *          given database must be delegated.
    */
   protected IdentityManager getWorker(Database database)
   {
      Persistence p = PersistenceFactory.getInstance(database);
      return p.getIdentityManager();
   }
}