/*
** Module   : TriggerData.java
** Abstract : This is a container that the database trigger manager uses for handling trigger
**             classes.
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 OM  20131206 Initial commit.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.trigger;

import com.goldencode.p2j.util.DatabaseTrigger;

/**
 * This is a container for all information needed for a trigger callback. This class is
 * not mutable, the values passed to constructor can be later obtained using the individual
 * getters.
 * This class and all his methods are package-protected so it is not visible outside the trigger
 * package, in fact it can only be accessed by <code>TriggerBlockLayer</code> and
 * <code>DatabaseTriggerManager</code>.
 */
class TriggerData
{
   /** Trigger class that will be instantiated and executed. */
   private Class<? extends DatabaseTrigger> triggerClass;

   /** The container in which the trigger is declared. */
   private Object triggerContainer;

   /** Flag for disabling the schema trigger. */
   private boolean override;

   /** Flag that the trigger needs the old value of the buffer (WRITE only) */
   private boolean oldBuffer;

   /**
    * Builds an immutable object with data needed for a trigger to instantiate/execute.
    *
    * @param   triggerClass
    *          The class of the trigger.
    * @param   triggerContainer
    *          The container of the trigger.
    * @param   override
    *          Allows to block the schema trigger or execute before it.
    * @param   oldBuffer
    *          Some triggers need a buffer with old value.
    */
   TriggerData(Class<? extends DatabaseTrigger> triggerClass,
               Object triggerContainer,
               boolean override,
               boolean oldBuffer)
   {
      this.triggerClass = triggerClass;
      this.triggerContainer = triggerContainer;
      this.override = override;
      this.oldBuffer = oldBuffer;
   }

   /**
    * Obtain the trigger class.
    *
    * @return  The trigger class.
    */
   Class<? extends DatabaseTrigger> getTriggerClass()
   {
      return triggerClass;
   }

   /**
    * Obtain the trigger container.
    *
    * @return  The container object of the trigger.
    */
   Object getTriggerContainer()
   {
      return triggerContainer;
   }

   /**
    * Check if this trigger replaces the schema trigger (if any) or if it executes before.
    *
    * @return  The override flag.
    */
   boolean isOverride()
   {
      return override;
   }

   /**
    * Check if this trigger additionally need the old value of the buffer.
    *
    * @return  true if this trigger additionally need the old value of the buffer.
    */
   boolean hasOldBuffer()
   {
      return oldBuffer;
   }
}