/*
** Module   : AbstractWrapperUserType.java
** Abstract : Abstract base class for wrapper-related Hibernate user type
**            implementations
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050630   @21589 Created initial version. Base class for
**                           Hibernate custom user types which manage P2J
**                           data wrapper classes.
** 002 ECF 20050729   @21951 Refactored to move nullSafeSet method
**                           implementation up from subclasses. This was
**                           necessary to add centralized logging for
**                           prepared statement parameter binding.
** 003 ECF 20050809   @22071 Refactored to move nullSafeGet method
**                           implementation up from subclasses. This was
**                           necessary to add centralized logging and null
**                           checking logic.
** 004 ECF 20051025   @23168 Changed isMutable method implementation. Now
**                           returns false instead of true to reflect the
**                           access strategy implemented by DMOs for most
**                           user types inheriting from this class.
** 005 ECF 20051127   @23536 Minor change to logging. Use toStringMessage
**                           instead of toString on wrapper objects.
** 006 ECF 20060327   @25254 Implement Serializable. Required for caching.
** 007 ECF 20070306   @32285 Made equals() and hashCode() non-final. It
**                           is necessary to allow subclasses to override
**                           these methods.
** 008 ECF 20070531   @33907 Modified isMutable(). Always returns true
**                           (previously returned false), indicating that
**                           Hibernate must dirty check DMOs against its
**                           snapshots even if no explicit write was
**                           tracked (all P2J data wrappers are mutable).
** 009 ECF 20080806   @39294 Moved to type sub-package of persist.
** 010 GES 20090424   @41945 Import change.
** 011 ECF 20090605   @42648 Rolled back #010 (@41945). StringHelper in use is
**                           Hibernate's implementation, not P2J's. Import of
**                           package com.goldencode.util is not necessary.
** 012 SVL 20120331          Upgraded to Hibernate 4.
** 013 ECF 20140220          Replaced Apache commons logging with J2SE standard logging.
** 014 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import java.io.*;
import java.sql.*;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.engine.spi.*;
import org.hibernate.type.*;
import org.hibernate.usertype.*;
import com.goldencode.p2j.util.*;

/**
 * The abstract base class for custom, Hibernate user type implementations
 * which manage the P2J data wrapper classes.
 */
abstract class AbstractWrapperUserType
implements UserType,
           Serializable
{
   /** Logger */
   protected static final Logger LOG = LogHelper.getLogger(Type.class);
   
   /** Is trace level logging enabled? */
   protected static final boolean LOG_TRACE = LOG.isLoggable(Level.FINEST);
   
   /**
    * Returns <code>true</code> to indicate that the P2J data wrapper object
    * is mutable.
    *
    * @return  <code>true</code>.
    */
   public boolean isMutable()
   {
      return true;
   }
   
   /**
    * Reconstruct an object from the cacheable representation. This minimal
    * implementation simply returns a deep copy of <code>cached</code>.
    *
    * @param   cached
    *          A data wrapper instance.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  Copy of <code>cached</code>
    */
   public final Object assemble(Serializable cached, Object owner)
   {
      return deepCopy(cached);
   }
   
   /**
    * Transform an object into its cacheable representation. This minimal
    * implementation simply returns a deep copy of <code>value</code>.
    *
    * @param   value
    *          A data wrapper instance.
    *
    * @return  Copy of <code>value</code>.
    */
   public final Serializable disassemble(Object value)
   {
      return (Serializable) deepCopy(value);
   }
   
   /**
    * During merge, replace the existing (target) value in the entity we
    * are merging to with a new (original) value from the detached entity
    * we are merging. This minimal implementation simply returns a deep
    * copy of <code>original</code>.
    *
    * @param   original
    *          The value from the detached entity being merged.
    * @param   target
    *          The value in the managed entity.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  Copy of <code>original</code>.
    */
   public final Object replace(Object original, Object target, Object owner)
   {
      return deepCopy(original);
   }
   
   /**
    * Tests for equality between two data wrapper instances. The objects are
    * considered the same if they meet any of the following conditions:
    * <ul>
    * <li>they represent the same object instance in memory; OR
    * <li>they are both <code>null</code>; OR
    * <li><code>x.equals(y)</code> evaluates to <code>true</code>.
    * </ul>
    * Otherwise, they are considered different.
    * <p>
    * This method is used by Hibernate for dirty checking.
    *
    * @param   x
    *          A data wrapper type instance.
    * @param   y
    *          A data wrapper type instance.
    */
   public boolean equals(Object x, Object y)
   {
      if (x == y)
      {
         return true;
      }
      
      if (x == null || y == null)
      {
         return false;
      }
      
      return x.equals(y);
   }
   
   /**
    * Get the hash code of the given object in a manner that is consistent
    * with the equality check performed by {@link #equals}. This
    * implementation simply returns the result of invoking
    * <code>x.hashCode()</code>.
    *
    * @param   x
    *          A data wrapper type instance.
    *
    * @return  The given object's hash code.
    */
   public int hashCode(Object x)
   {
      return x.hashCode();
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon
    * a data value in the specified result set. If the corresponding data
    * value is NULL in the result set, return a data wrapper instance which
    * is initialized to the unknown value. Performs logging if trace is
    * enabled.
    * <p>
    * <b>Note:</b> this method assumes a mapping of JDBC data type to a
    * <i>Java object</i>. Subclasses which deal in primitives must override
    * this implementation to avoid attempting to extract an object reference
    * from the result set.
    *
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          our data.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   session
    *          Not used.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  A <code>BaseDataType</code> instance, possibly initialized to
    *          the unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs,
                             String[] names,
                             SessionImplementor session,
                             Object owner)
   throws HibernateException, SQLException
   {
      Object value = getBaseValue(rs, names[0]);
      
      if (value == null || rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.log(Level.FINEST, "returning unknown value as column: " + names[0]);
         }
         return instantiateUnknown();
      }
      
      BaseDataType bdt = instantiateRegular(value);
      if (LOG_TRACE)
      {
         LOG.log(Level.FINEST, "returning '" + toString(bdt) + "' as column: " + names[0]);
      }
      
      return bdt;
   }
   
   /**
    * Set a value or <code>null</code> into the given prepared statement,
    * based upon the <code>BaseDataType</code> instance <code>value</code>.
    * Performs logging if trace is enabled.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          A <code>BaseDataType</code> instance or <code>null</code>. If
    *          this parameter is <code>null</code> or represents the unknown
    *          value, NULL is written.
    * @param   index
    *          Index in the statement at which to write the data.
    * @param   session
    *          Not used.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   public void nullSafeSet(PreparedStatement ps,
                           Object value,
                           int index,
                           SessionImplementor session)
   throws HibernateException, SQLException
   {
      BaseDataType bdt = (BaseDataType) value;
      if (bdt == null || bdt.isUnknown())
      {
         if (LOG_TRACE)
         {
            LOG.log(Level.FINEST, "binding null to parameter: " + index);
         }
         ps.setNull(index, sqlTypes()[0]);
      }
      else
      {
         if (LOG_TRACE)
         {
            LOG.log(Level.FINEST, "binding '" + toString(bdt) + "' to parameter: " + index);
         }
         bindNonNull(ps, bdt, index);
      }
   }
   
   /**
    * Subclasses should override this method to return a string
    * representation of the given value, if it must differ from the default
    * value supplied by invoking <code>value.toString()</code>.
    *
    * @param   value
    *          The value to be converted to a string. It is safe to assume
    *          this will not be <code>null</code>.
    *
    * @return  The result of <code>value.toString()</code>.
    */
   protected String toString(BaseDataType value)
   {
      return value.toStringMessage();
   }
   
   /**
    * Extract a data value (not wrapped by a <code>BaseDataType</code>
    * implementation) object from a result set at a particular column.
    * <p>
    * This must be implemented by subclasses which wrap object references
    * inside P2J wrapper objects. This operation is not required to be
    * supported by subclasses which wrap primitive values inside P2J wrapper
    * objects.
    *
    * @param   rs
    *          Result set containing data to be extracted.
    * @param   column
    *          Name of the column at which data will be extracted.
    *
    * @return  An object which will be wrapped inside a P2J wrapper object,
    *          or <code>null</code> if the SQL value was NULL at the target
    *          column.
    *
    * @throws  UnsupportedOperationException
    *          always; must be overridden to avoid this.
    * @throws  SQLException
    *          if a database access error occurs.
    */
   protected Object getBaseValue(ResultSet rs, String column)
   throws SQLException
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which is initialized to the unknown value.
    * <p>
    * This must be implemented by subclasses which can instantiate wrapper
    * instances from an object reference. This operation is not required to
    * be supported by subclasses which can only instantiate wrapper instances
    * from primitive values.
    *
    * @return  Data wrapper object initialized to unknown value.
    *
    * @throws  UnsupportedOperationException
    *          always; must be overridden to avoid this.
    */
   protected BaseDataType instantiateUnknown()
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Construct and return an instance of the appropriate subclass of
    * <code>BaseDataType</code>, which wraps <code>value</code>.
    * <p>
    * This must be implemented by subclasses which can instantiate wrapper
    * instances from an object reference. This operation is not supported by
    * subclasses which can only instantiate wrapper instances from primitive
    * values.
    *
    * @param   value
    *          Value to be wrapped by a concrete subclass of
    *          <code>BaseDataType</code>; will not be <code>null</code>.
    *
    * @return  Instance of the appropriate <code>BaseDataType</code> concrete
    *          implementation.
    *
    * @throws  UnsupportedOperationException
    *          always; must be overridden to avoid this.
    */
   protected BaseDataType instantiateRegular(Object value)
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Subclasses must implement this method to set a non-null value into the
    * given prepared statement, based upon the <code>BaseDataType</code>
    * instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          A <code>BaseDataType</code> instance, guaranteed to be
    *          non-<code>null</code>.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected abstract void bindNonNull(PreparedStatement ps, BaseDataType value, int index)
   throws SQLException;
}