/*
** Module   : CharacterUserType.java
** Abstract : Custom user type to wrap P2J character for Hibernate access
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050630   @21593 Created initial version. Maps P2J character
**                           to SQL varchar.
** 002 ECF 20050729   @21952 Refactored to move nullSafeSet method
**                           implementation up to superclass. This was
**                           necessary to add centralized logging for
**                           prepared statement parameter binding.
** 003 ECF 20050809   @22072 Refactored to move nullSafeGet method
**                           implementation up to superclass. This was
**                           necessary to add centralized logging and null
**                           checking logic.
** 004 ECF 20060831   @29083 Use toStringMessage() instead of getValue()
**                           on character object.
** 005 ECF 20070306   @32287 Override superclass' implementation of equals
**                           and hashCode. We require these methods to
**                           operate on the underlying character objects
**                           in a case sensitive manner always.
** 006 ECF 20080311   @37472 Fixed deepCopy(). Did not handle null safely.
** 007 ECF 20080806   @39295 Moved to type sub-package of persist.
** 008 ECF 20101001          Removed toString method, which was a duplicate of
**                           the same method in the superclass.
** 009 ECF 20140624          Refactored SQL_TYPES to use TypeHelper.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.character P2J character}. This permits Hibernate
 * to convert this data type to and from the SQL VARCHAR data type. The
 * unknown value is mapped to SQL NULL.
 */
public class CharacterUserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { TypeHelper.getSqlType(character.class) };
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map a character to/from a
    * database record.
    *
    * @return  Array containing SQL type constant.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.character P2J character}
    *          class.
    */
   public final Class returnedClass()
   {
      return character.class;
   }
   
   /**
    * Tests for equality between two <code>character</code> instances in a
    * case-sensitive manner, regardless of the case-sensitivity setting of
    * either.  The objects are considered the same if they meet any of the
    * following conditions:
    * <ul>
    * <li>they represent the same object instance in memory; OR
    * <li>they are both <code>null</code>; OR
    * <li><code>x.equalsCaseSensitively(y)</code> evaluates to
    * <code>true</code>.
    * </ul>
    * Otherwise, they are considered different.
    * <p>
    * This method is used by Hibernate for dirty checking.
    *
    * @param   x
    *          A <code>character</code> data wrapper type instance.
    * @param   y
    *          A <code>character</code> data wrapper type instance.
    */
   public boolean equals(Object x, Object y)
   {
      if (x == y)
      {
         return true;
      }
      
      if (x == null || y == null)
      {
         return false;
      }
      
      return ((character) x).equalsCaseSensitively((character) y);
   }
   
   /**
    * Get the hash code of the given object in a manner that is consistent
    * with the equality check performed by {@link #equals}. This
    * implementation simply returns the result of invoking
    * <code>x.caseSensitiveHashCode()</code>.
    *
    * @param   x
    *          A <code>character</code> data wrapper type instance.
    *
    * @return  The given object's hash code.
    */
   public int hashCode(Object x)
   {
      return ((character) x).caseSensitiveHashCode();
   }
   
   /**
    * Create and return a deep copy of the given <code>character</code>
    * instance.
    *
    * @param   value
    *          The <code>character</code> instance to be copied.
    *
    * @return  A new instance of <code>character</code> which contains the
    *          same internal value as <code>value</code>.
    */
   public final Object deepCopy(Object value)
   {
      if (value == null)
      {
         return null;
      }
      
      return new character((character) value);
   }
   
   /**
    * Extract a <code>String</code> value from a result set at a particular
    * column.
    *
    * @param   rs
    *          Result set containing data to be extracted.
    * @param   column
    *          Name of the column at which data will be extracted.
    *
    * @return  String data extracted from the result set; may be
    *          <code>null</code>.
    *
    * @throws  SQLException
    *          if a database access error occurs.
    */
   protected final Object getBaseValue(ResultSet rs, String column)
   throws SQLException
   {
      return rs.getString(column);
   }
   
   /**
    * Construct and return an instance of <code>character</code> which is
    * initialized to the unknown value.
    *
    * @return  A <code>character</code> instance which represents the unknown
    *          value.
    */
   protected final BaseDataType instantiateUnknown()
   {
      return new character();
   }
   
   /**
    * Construct and return an instance of <code>character</code> which wraps
    * <code>value</code>.
    *
    * @param   value
    *          String value to be wrapped by a <code>character</code>;
    *          will not be <code>null</code>.
    *
    * @return  <code>value</code> wrapped in a <code>character</code>.
    */
   protected final BaseDataType instantiateRegular(Object value)
   {
      return new character((String) value);
   }
   
   /**
    * Set a <code>String</code> value into the given prepared statement,
    * based upon the <code>character</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          A <code>character</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected final void bindNonNull(PreparedStatement ps,
                                    BaseDataType value,
                                    int index)
   throws SQLException
   {
      ps.setString(index, ((character) value).getValue());
   }
}