/*
** Module   : HandleFieldRef.java
** Abstract : Data wrapper which associates with a DMO property and provides direct access to
**            this property.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140103 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * Data wrapper which associates with a DMO property and provides direct access to this property.
 * <p>
 * All writing is done to both the DMO and the local {@link handle}'s private {@code value}.
 * On instantiation, the current field value is saved in the {@link handle}'s private 
 * {@code value}, and this instance will be used for reading until {@link #set} or 
 * {@link #setUnknown} is called again.
 * <p>
 * This usage is OK 
 */
public class HandleFieldRef
extends handle
{
   /** The field referenced by this instance. */
   private final FieldReference fieldRef;
   
   /**
    * Constructor.
    * 
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    */
   public HandleFieldRef(DataModelObject dmo, String property)
   {
      this.fieldRef = new FieldReference(dmo, property);
      
      RecordBuffer rb = ((BufferReference) dmo).buffer();
      // initialize this instance based on the current buffer state; if no record is loaded in the
      // buffer, then set it to unknown. there will be a 142 - Unable to update field error raised
      // when we try to assign a new value to this instance.
      if (rb.getCurrentRecord() != null)
      {
         super.set(((handle) fieldRef.get()).getResource());
      }
      else
      {
         super.setUnknown();
      }
   }

   /**
    * Set the value of this instance as a {@link WrappedResource}.
    * <p>
    * Also updates the {@link #fieldRef} with the new state of this handle instance.
    * 
    * @param    value
    *           The value to be represented in this instance.
    */
   @Override
   public void set(WrappedResource value)
   {
      super.set(value);
      
      fieldRef.set(this);
   }
   
   /**
    * Sets the state of this instance's <code>unknown value</code> flag or state to <code>true</code>.
    * <p>
    * <b>Warning: the data stored in this instance may be invalid after calling this method.</b>
    * <p>
    * Also updates the {@link #fieldRef} with the new state of this handle instance.
    */
   public void setUnknown()
   {
      super.setUnknown();
      
      fieldRef.set(this);
   }
}