/*
** Module   : HintsTreeNode.java
** Abstract : tree node for preprocessor hints tree
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 NVS 20050526 NEW   @21297 Created initial version which provides a
**                               storage for the list of references and file
**                               arguments.
** 002 NVS 20050601 CHG   @21349 Arguments and reference hints are returned as
**                               lists instead of arrays. A boolean parameter
**                               "deep" tells to return either node's hints 
**                               (false) or the subtree's hints (true).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.preproc;

import java.util.*;
import com.goldencode.p2j.util.TreeNode;

/**
 * A preprocessor hints tree node which is an extenstion of TreeNode class.
 * This class extends the TreeNode by providing the storage for the list of
 * references and arguments, which are attached to this node. The node itself
 * represents either the top level source file or an include file.
 *
 * In addition to TreeNode, each instance of this class provides:
 * <ul>
 *    <li> An embedded instance of the <code>IncludeHint</code> class.
 *    <li> An array of <code>ReferenceHint</code>s.
 *    <li> An array of <code>ArgumentHint</code>s.
 * </ul>
 */
public class HintsTreeNode
extends TreeNode
{
   /** Embedded IncludeHint instance */
   private IncludeHint include = null;
   
   /** Array of ReferenceHints */
   private ReferenceHint[] references = null;
   
   /** Array of ArgumentHints */
   private ArgumentHint[] arguments = null;
   
   /**
    * Constructor.
    *
    * @param include
    *        IncludeHint instance to embed into this HintsTreeNode
    * @param references
    *        array of ReferenceHints to embed into this HintsTreeNode
    * @param arguments
    *        array of ArgumentHints to embed into this HintsTreeNode
    */
   public HintsTreeNode(IncludeHint include, ReferenceHint[] references,
                        ArgumentHint[] arguments)
   {
      this.include    = include;
      this.references = references;
      this.arguments  = arguments;
   }

   /**
    * Gets the embedded IncludeHint instance.
    *
    * @return the embedded IncludeHint instance.
    */
   public IncludeHint getIncludeHint()
   {
      return include;
   }

   /**
    * Gets the embedded ReferenceHints.
    *
    * @param deep
    *        tells to get only this node's hints (<code>false</code>) or
    *        everything
    * @return list of ReferenceHints.
    */
   public List getReferenceHints(boolean deep)
   {
      List list = new LinkedList();
      getReferences(this, list, deep);

      return list;
   }

   /**
    * Gets the embedded ArgumentHints.
    *
    * @param deep
    *        tells to get only this node's hints (<code>false</code>) or
    *        everything
    * @return list of ArgumentHints.
    */
   public List getArgumentHints(boolean deep)
   {
      List list = new LinkedList();
      getArguments(this, list, deep);

      return list;
   }

   /**
    * Gets the embedded ReferenceHints.
    *
    * @param node
    *        HintsTreeNode to get information from
    * @param list
    *        the target list to store hints
    * @param deep
    *        tells to get only this node's hints (<code>false</code>) or
    *        everything
    */
   private static void getReferences(HintsTreeNode node, List list, boolean deep)
   {
      // get this node's hints
      if (node.references != null)
         for (int i = 0; i < node.references.length; i ++)
            list.add(node.references[i]);
      else
         return;

      if (!deep)
         return; 

      // walk the subtree
      Iterator iter = node.childIterator();
      HintsTreeNode child = null;

      while (iter.hasNext())
      {
         child = (HintsTreeNode)iter.next();
         getReferences(child, list, deep);
      }
   }

   /**
    * Gets the embedded ArgumentHints.
    *
    * @param node
    *        HintsTreeNode to get information from
    * @param list
    *        the target list to store hints
    * @param deep
    *        tells to get only this node's hints (<code>false</code>) or
    *        everything
    */
   private static void getArguments(HintsTreeNode node, List list, boolean deep)
   {
      // get this node's hints
      if (node.arguments != null)
         for (int i = 0; i < node.arguments.length; i ++)
            list.add(node.arguments[i]);
      else
         return;

      if (!deep)
         return; 

      // walk the subtree
      Iterator iter = node.childIterator();
      HintsTreeNode child = null;

      while (iter.hasNext())
      {
         child = (HintsTreeNode)iter.next();
         getArguments(child, list, deep);
      }
   }
}