/*
** Module   : PreprocesorHintsWorker.java
** Abstract : provides a pattern engine service to lookup preprocessor hints
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20050525   @21280 Created initial version based on UastHints-
**                           Worker code.
** 002 GES 20090518   @42393 Import change.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.preproc;

import java.io.*;
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;

/**
 * Provides a pattern engine service to query the preprocessor hints linked to
 * an AST.
 */
public class PreprocessorHintsWorker
extends AbstractPatternWorker
{
   /** Stores the current AST's hints object. */
   private PreprocessorHints hints = null;
   
   /**
    * Default constructor which calls the super-class constructor, registers
    * its libraries and initializes its instance members.
    */
   public PreprocessorHintsWorker()
   throws ConfigurationException
   {
      super();
      setLibrary(new HintsReader());
   }
      
   /**
    * Resets the current hints instance to <code>null</code> each time a new
    * AST is visited.  This enables the lazy loading of hints files only when
    * needed.
    *
    * @param   ast
    *          The root node of the source AST about to be processed by the
    *          pattern engine.
    */
   public void visitAst(Aast ast)
   {
      hints = null;
   }   
   
   /**
    * Provides a service for reading preprocessor hints associated with the
    * current AST.  This class is a <code>PatternEngine</code> worker.
    */
   public class HintsReader
   {
      /**
       * Gets the PreprocessorHints instance for this file.
       *
       * @return <code>preprocessorHints</code> instance.
       */
      public PreprocessorHints getHints()
      {
         if (!loadHints())
            return null;         
         
         return hints;
      }      

      /**
       * Loads the hints file for an AST if it exists and hasn't already been
       * loaded.
       *
       * @return   <code>true</code> if there are hints for this file and the
       *           hints are loaded, <code>false</code> if no hints are
       *           available.
       */
      private boolean loadHints()
      {
         if (hints == null)
         {
            try
            {
               Aast src  = resolver.getSourceAst();
               File file = Configuration.toFile(src.getFilename()); 
               
               hints = new PreprocessorHints(file.getAbsolutePath());
            }
            
            catch (Exception exc)
            {
               // any failure drops us out
               return false;
            }
         }
         
         return true;
      }
   }
}