/*
** Module   : JobMode.java
** Abstract : enum defining the possible job modes.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.scheduler;

/**
 * Definition of possible job modes:
 * <ul>
 * <li>{@link #NOW} - execute immediately.</li>
 * <li>{@link #ONE_TIME} - schedule for execution at a certain time/date.</li>
 * <li>{@link #RECURRENT} - schedule for execution at a certain time/date, recurrent.</li>
 * </ul>
 */
public enum JobMode
{
   NOW       (false, "NOW"),
   RECURRENT (true,  "RECURRENT"),
   ONE_TIME  (true,  "ONE-TIME");
   
   /** Flag indicating if this job mode is timed. */
   private final boolean timed;
   
   /** The job mode. */
   private final String mode;
   
   /**
    * Initialize a new job type.
    *
    * @param    timed
    *           Flag indicating if this is a timed job or not.
    * @param    mode
    *           The job mode.
    */
   JobMode(boolean timed, String mode)
   {
      this.timed = timed;
      this.mode = mode;
   }

   /**
    * Check if this job mode is timed (i.e. it must be scheduled for execution at a certain 
    * time/date).
    * 
    * @return   See above.
    */
   public boolean isTimed()
   {
      return this.timed;
   }

   /**
    * Resolve the job mode from the given string.
    * 
    * @param    mode
    *           The job mode.
    *           
    * @return   One of the {@link JobMode} values or {@code null} if it can not be resolved.
    */
   public static JobMode resolveJobMode(String mode)
   {
      for (JobMode jobMode : JobMode.values())
      {
         if (jobMode.mode.equalsIgnoreCase(mode))
         {
            return jobMode;
         }
      }

      return null;
   }
}