/*
** Module   : SchemaStatsHelper.java
** Abstract : helper to copy and link to schema source files
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20070716   @34522 Helper to copy and link to schema source
**                           files.
** 002 GES 20130206          Moved parent class to the pattern package.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.uast.*;

/**
 * Helper to copy and link to Progress 4GL schema source files.
 */
public class SchemaStatsHelper
implements SourceFileHelper
{
   /** Column headings. */
   private static final String[] headings = new String[]
   {
      "Original Source File",
      "Schema AST"
   };
   
   /** Store a reverse lookup from the schema name to the source file. */
   private HashMap reverse = new HashMap();
   
   /**
    * Obtain the list of source file descriptions which is suitable for use
    * as the column headings in the source file index table.
    *
    * @return   Column headings.
    */
   public String[] getHeadings()
   {
      return headings;
   }
   
   /**
    * Provide the list of copy specifications for the given source file and
    * output directory.
    *
    * @param    schema
    *           Schema AST filename.
    * @param    outDir
    *           Root of the output directory where all copies must be made.
    *
    * @return   The list of source and target copy operations to be made.
    */
   public CopySpec[] getSpecifications(String schema, String outDir)
   {
      // safety first
      if (schema == null)
         return null;
      
      CopySpec[] specs = new CopySpec[2];
      
      SourceSpec ss = (SourceSpec) reverse.get(schema);
      
      String src = null;
      String tar = null;
      String scm = null;
      
      try
      {
         // create the master source filename
         src = Configuration.toFile(ss.source).getAbsolutePath();
         scm = Configuration.toFile(schema).getAbsolutePath();
      }
      catch (ConfigurationException ce)
      {
         return null;
      }
      
      if (!ss.df)
      {
         src = String.format("%s%s", src, AstGenerator.CACHE_POSTFIX);
         tar = String.format("%s%s%s.html",
                             outDir,
                             ss.source,
                             AstGenerator.CACHE_POSTFIX);
      }
      else
      {
         tar = String.format("%s%s.html",
                             outDir,
                             ss.source);
      }
      
      specs[0] = new CopySpec(src, tar, CopySpec.Type.TXT_TO_HTML);
      specs[1] = new CopySpec(scm, String.format("%s%s", outDir, schema));
      
      return specs;
   }
   
   /**
    * Create a URL to link to the given source file and line number.
    *
    * @param    relPath
    *           Relative path to the root of the output directory where the
    *           source files copies were made.
    * @param    schema
    *           Schema AST filename.
    * @param    line
    *           The line number to which to link.
    *
    * @return   A URL to link to the copied file.
    */
   public String getURL(String relPath, String schema, String line)
   {
      StringBuilder sb = new StringBuilder();
      SourceSpec spec = (SourceSpec) reverse.get(schema);
      
      sb.append(relPath)
        .append(spec.source);
        
      if (!spec.df)
         sb.append(AstGenerator.CACHE_POSTFIX);
      
      sb.append(".html#")
        .append(line);
      
      return sb.toString();
   }
   
   /**
    * Add a mapping definition from the given schema name to the original
    * source file that defined that schema. This must be obtained from the
    * schema AST files during any report processing.  This is relied upon
    * in order to properly determine the files for copying.
    *
    * @param    schema
    *           The schema AST file name.
    * @param    source
    *           The source file.
    * @param    df
    *           <code>true</code> if this source file is a .df file.
    */
   public void addMapping(String schema, String source, boolean df)
   {
      reverse.put(schema, new SourceSpec(source, df));
   }
   
   /**
    * Simple data structure to store key data for reverse mapping purposes.
    */
   private static class SourceSpec
   {
      /** Source file name. */
      private String source = null;
      
      /** <code>true</code> if the source file is a .df file. */
      private boolean df = false;
      
      /**
       * Create an instance.
       *
       * @param    source
       *           The source file.
       * @param    df
       *           <code>true</code> if this source file is a .df file.
       */
      SourceSpec(String source, boolean df)
      {
         this.source = source;
         this.df     = df;
      }
   }
}