/*
** Module   : AccessControlList.java
** Abstract : Represents associations between an abstract resource instance 
**            and a list of related bindings.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 NVS 20050318 NEW   @20446 Created the initial version. This class 
**                               represents associations between an abstract 
**                               resource instance and a list of related 
**                               bindings. 
** 002 NVS 20060412 CHG   @25492 Made the class and some getters public. Made
**                               the instance protected from the indirect
**                               changes.
** 003 NVS 20060425 CHG   @25728 Fixed a regression. getBindings() should
**                               return null if no bindings are set.
** 004 NVS 20090415 CHG   @41803 ACL now keeps directory object IDs of the
**                               objects it was constructed from.
** 005 NVS 20090624 CHG   @42842 Simplified due to move to new ACLs.
** 006 EVL 20160224              Javadoc fixes to make compatible with Oracle Java 8 for
**                               Solaris 10.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import java.util.*;

/**
 * Associates a named instance of abstract resource with a list of bindings. 
 */
public class AccessControlList
{
   /** abstract resource instance name */
   private String instanceName = null;

   /** instance or match flag */
   private boolean exact = true;

   /** directory object ID the ACL came from */
   private String refoid = null;

   /** list of associated bindings (now always one binding in the list) */
   private List bindings = null;

   /**
    * Package private constructor. 
    *
    * @param refoid
    *        resource instance directory object ID
    *
    * @param instanceName
    *        an instance of the resource 
    *
    * @param exact
    *        <code>true</code> if name is an exact instance name, otherwise it
    *        is a regular expression
    * @param bindings
    *        list of associated bindings
    */
   AccessControlList(String refoid, String instanceName, boolean exact, 
                     List bindings)
   {
      this.refoid = refoid;
      this.instanceName = instanceName;
      this.exact = exact;
      this.bindings = bindings;
   }

   /**
    * Gets the embedded resource instance name.
    *
    * @return embedded resource instance name
    */
   public String getInstanceName()
   {
      return instanceName;
   }

   /**
    * Gets the directory node id where this ACL originated.
    *
    * @return directory node id
    */
   public String getOrigin()
   {
      return refoid;
   }

   /**
    * Gets the embedded bindings list.
    *
    * @return embedded bindings list
    */
   public List getBindings()
   {
      if (bindings == null)
         return null;

      List ret = new LinkedList();
      ret.addAll(bindings);
      
      return ret;
   }
   
   /**
    * Sets the bindings list.
    *
    * @param bindings
    *        list of bindings to be saved 
    */
   void setBindings(List bindings)
   {
      this.bindings = bindings;
   }

   /**
    * Checks whether the instance name is exact or not.
    *
    * @return <code>true</code> if the instance name is an exact name
    */
   public boolean isExact()
   {
      return exact;
   }
}