/*
** Module   : StringConditionRights.java
** Abstract : implements the Rights interface for any resource which only needs a simple
**            string condition
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------------
** 001 GES 20160222 Created the first version, using nearly the complete implementation from
**                  EntryPointRights.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import com.goldencode.p2j.directory.*;

import java.io.Serializable;

/**
 * Implements the <code>StringConditionRights</code> rights objects.  This is meant to be
 * common code for implementing a simple string expression as a condition with almost no
 * code in the child class.
 */
public class StringConditionRights
implements Rights,
           Serializable
{
   /** Access condition (<code>boolean</code> expression). */
   private String condition = null;
   
   /**
    * Default constructor which creates an instance that always evaluates to no access.
    */
   public StringConditionRights()
   {
      condition = "false";
   }

   /**
    * Protected constructor. 
    *
    * @param    condition
    *           <code>boolean</code> expression that allows access when it evaluates to 
    *           <code>true</code>.
    */
   protected StringConditionRights(String condition)
   {
      this.condition = condition;
   }

   /**
    * Reports this class' rights name. 
    *
    * @return   The name of this rights class.
    */
   public String getRightsName()
   {
      return "StringConditionRights";
   }
   
   /**
    * Converts this object to string. 
    *
    * @return   The text representation of this object.
    */
   public String toString()
   {
      return "{" + condition + "}";
   }

   /**
    * Gets the condition. 
    *
    * @return the boolean expression that determines the access condition
    */
   String getCondition()
   {
      return condition;
   }
   
   /**
    * Creates an external representation of this instance as a directory node.
    *
    * @param    ds
    *           The instance of <code>DirectoryService</code> to use. 
    * @param    node
    *           The full directory node name for the rights node being created.
    *           
    * @return   <code>true</code> if the directory node is successfully created. 
    */
   public boolean toDirectoryNode(DirectoryService ds, String node)
   {
      String name = getRightsName();
      
      Attribute[] data = new Attribute[]
      {
         new Attribute(ds.getClassNodeAttribute(name, "allow"), new Object[] { condition }),
      };
      
      return ds.addNode(node, name, data);
   }
}