/*
** Module   : StringConditionRightsEditor.java
** Abstract : admin console rights editor for the StringConditionResource
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------------
** 001 GES 20160222 Created the first version, using nearly the complete implementation from
**                  StringConditionRights.
 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import com.goldencode.p2j.admin.client.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/**
 * Edits <code>StringConditionRights</code> for the admin client.
 */
public class StringConditionRightsEditor
implements RightsEditor
{
   /** The instance of <code>AdminClient</code> from which we were called. */
   private AdminClient client = null;
   
   /** The editor panel instance. */
   private JPanel edit = null;
   
   /** The condition expression text field. */
   private JTextField cond = null;
   
   /** Save button. */
   private JButton bsave = null;
   
   /** Cancel button. */
   private JButton bcanc = null;
   
   /** The dialog instance. */
   private JDialog dialog = null;
   
   /** Save or cancel flag. */
   private boolean saving = false;
   
   /**
    * Gives chance to editors to initialize their data structures.
    * <p>
    * This is a one time call for this kind of resource for the life time of the admin session. 
    *
    * @param    ac
    *           The instance of <code>AdminClient</code> from which the editor is called.
    * @param    description
    *           Rights structure description.
    */
   public void initialize(AdminClient ac, Description[] description)
   {
      // save the useful reference
      client = ac;
      
      // create the edit panel
      Component[][] comps = new Component[2][2];
      
      // first row - label and text input field
      cond = new JTextField(24);
      cond.setToolTipText(description[0].getDescription());
      comps[0][0] = new JLabel(description[0].getLabel());
      comps[0][1] = cond;
      
      // second row - buttons
      bsave = new JButton("Save");
      bcanc = new JButton("Cancel");
      comps[1][0] = bcanc;
      comps[1][1] = bsave;
      
      // layout
      edit = AdminUtils.groupComponents(comps, getTitle(), false);
      
      // event handlers:
      
      // cancel button
      bcanc.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });
      
      // save button
      bsave.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent evt) 
         {
            handleSave();
         }
      });
   }

   /**
    * Creates an instance of the <code>Rights</code>, which does not take any input and is
    * considered a default starting point in editing.
    *
    * @return   An instance of <code>Rights</code>.
    */
   public Rights createDefaultRights()
   {
      return new StringConditionRights();
   }
   
   /**
    * Edits the given instance of the <code>Rights</code>.
    * <p>
    * The implementation should prepare its modal dialog using the initial state from the passed
    * instance of the <code>Rights</code>, show the dialog, provide event processing until the 
    * <code>Save</code> or <code>Cancel</code> action is applied, then hide the dialog and return
    * the appropriate result.
    *
    * @param    name
    *           The name of the resource instance being edited.
    * @param    exact
    *           If <code>true</code> this is an exact name match otherwise it is a regex.
    * @param    rights
    *           The instance of <code>Rights</code> to be edited.
    *          
    * @return   The resulting instance of <code>Rights</code> if the user saved the results,
    *           otherwise <code>null</code> on cancel.
    */
   public Rights edit(String name, boolean exact, Rights rights)
   {
      // fill the edit panel with the current rights
      cond.setText(((StringConditionRights)rights).getCondition());
      
      // display the dialog
      dialog = AdminUtils.modalDialog(client, getTitle(), edit);
      dialog.setVisible(true);
      
      // process the results
      dialog = null;
      
      if (!saving)
      {
         return null;
      }
      
      return createPopulatedRights(cond.getText());
   }
   
   /**
    * Creates an instance of the <code>Rights</code> configured using the given input.
    *
    * @param    expr
    *           The condition expression.
    *
    * @return   An instance of <code>Rights</code>.
    */
   protected Rights createPopulatedRights(String expr)
   {
      return new StringConditionRights(expr);
   }
   
   /**
    * Obtains the window title for the editor.
    *
    * @return   The title.
    */
   protected String getTitle()
   {
      return "String Condition Rights";
   }
   
   /**
    * Handle <code>Save</code> button press.
    */
   private void handleSave()
   {
      saving = true;      
      dialog.setVisible(false);
   }
   
   /**
    * Handle <code>Cancel</code> button press.
    */
   private void handleCancel()
   {
      saving = false;
      dialog.setVisible(false);
   }
}