/*
** Module   : CallGraphNode.java
** Abstract : self-referential node structure that supports creating/using a
**            call graph in tree form
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050310 NEW   @20281 Created initial version which stores a
**                               filename and inherits from TreeNode which
**                               provides all tree structure and processing. 
**                               Helper functions exist to turn the resulting
**                               tree into a degenerate tree (an array) of
**                               filename strings or File objects.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.util.*;

/**
 * A tree node which provides a simple mechanism to represent and use a
 * call graph of external filenames.  All tree features are implemented in
 * the parent class and this class is really just a simple container to
 * store and access call graph related data.
 */
public class CallGraphNode
extends TreeNode
{
   /** The file to which this node refers. */
   private String filename = null;
   
   /**
    * Creates an instance with a specific filename reference.  The filename
    * will be converted to a project relative form if necessary.
    *
    * @param    filename
    *           The filename of the source file to which the node refers.
    * @param    projectRelative
    *           Specifies if the name is relative to the project home
    *           directory.  If <code>false</code>, the name is either an
    *           absolute filename or it is relative to the current directory.
    */
   public CallGraphNode(String filename, boolean projectRelative)
   throws ConfigurationException
   {
      if (projectRelative)
      {
         // common case, already normalized
         this.filename = filename;
      }
      else
      {
         // convert to project-relative
         this.filename = Configuration.normalizeFilename(filename);
      }
   }
      
   /**
    * Get filename associated with the current node.
    *
    * @return  Filename associated with this node.
    */
   public String getFilename()
   {
      return filename;  
   }
   
   /**
    * Sets the filename associated with the current node.
    *
    * @param    filename
    *           The filename associated with this node.
    */
   public void setFilename(String filename)
   {
      this.filename = filename;  
   }
   
   /**
    * Get a <code>File</code> object based on the filename associated with
    * the current node.
    *
    * @return  The file associated with this node.
    */
   public File getFile()
   throws ConfigurationException
   {
      return Configuration.toFile(filename);  
   }   
   
   /**
    * Creates an array of all filenames in a depth-first traversal order.
    *
    * @return   The array of all filenames in this tree.
    */
   public String[] getFilenameList()
   {
      Iterator  iter  = descendantsIterator();
      ArrayList names = new ArrayList();
      
      while (iter.hasNext())
      {
         CallGraphNode cgn = (CallGraphNode) iter.next();
         names.add(cgn.getFilename());
      }
      
      return (String[])names.toArray(new String[0]);
   }
   
   /**
    * Creates an array of <code>File</code> objects corresponding to each
    * call graph node in a depth-first traversal order.
    *
    * @return   The array of all files in this tree.
    */
   public File[] getFileList()
   throws ConfigurationException
   {
      Iterator  iter  = descendantsIterator();
      ArrayList files = new ArrayList();
      
      while (iter.hasNext())
      {
         CallGraphNode cgn = (CallGraphNode) iter.next();
         files.add(cgn.getFile());
      }
      
      return (File[])files.toArray(new File[0]);
   }
}