/*
** Module   : CallGraphPersister.java
** Abstract : Reads/writes a CallGraph from/to an XML file.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050311   @20289 Created initial version which has a simple
**                           XML grammar that mirrors the call graph   
**                           tree structure and reads/writes such trees.
**                           The grammar is not made accessible to the 
**                           outside world as it is private to this class.
** 002 ECF 20050316   @20421 Normalize filename in constructor to always
**                           have the required extension.
** 003 ECF 20050421   @20808 Import p2j.xml package to use XmlHelper.
** 004 GES 20090429   @42060 Match package and class name changes.
** 005 GES 20090515   @42230 Import change.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;

/**
 * Visits each node of a call graph to persist the tree into an XML document
 * OR reads an XML document and instantiates the corresponding call graph.
 * <p>
 * Warning: this class can and will only read/write a single call graph tree
 * from/to the XML file.  Multiple call graphs are not supported in a single 
 * file.
 */
public class CallGraphPersister
{
   /** The commonly used postfix for all call graph persistence files. */
   public static final String CALL_GRAPH_POSTFIX = ".graph";
   
   /** XML target output filename */
   private File file = null;
   
   /** XML document which will be assembled during inspection */
   private Document dom = null;
   
   /** Root element containing the entire call graph. */
   private static final String ELEM_ROOT  = "graph-root";
   
   /** 
    * CallGraphNode element, containing all data as attributes and any number
    * of child CallGraphNode elements.
    */
   private static final String ELEM_CGN   = "call-graph-node";
   
   /** The source text as parsed from the original input file. */
   private static final String ATTR_FILE  = "filename";
      
   /**
    * Constructs an instance using a specific filename for the XML
    * persistence file.
    *
    * @param    filename
    *           Absolute or relative filename of the XML file to be written
    *           or read from.
    */
   public CallGraphPersister(String filename)
   {
      this(new File(normalizeFileSpec(filename)));
   }
   
   /**
    * Constructs an instance using a specific <code>File</code> instance.
    *
    * @param    file
    *           XML file to be written or read from.
    */
   public CallGraphPersister(File file)
   {
      this.file = file;
   }
   
   /**
    * Tack the necessary filename extension onto the end of the specified
    * fileSpec, if it is not already present.
    *
    * @param   fileSpec
    *          Specification of persistence file, with our without the
    *          required extension.
    *
    * @return  A filename spec which is guaranteed to end with the required
    *          extension.
    */
   public static String normalizeFileSpec(String fileSpec)
   {
      if (!fileSpec.toLowerCase().endsWith(CALL_GRAPH_POSTFIX))
      {
         StringBuffer buf = new StringBuffer(fileSpec);
         buf.append(CALL_GRAPH_POSTFIX);
         
         fileSpec = buf.toString();
      }
      
      return fileSpec;
   }
   
   /**
    * Read the source XML document and build a complete call graph from the 
    * stored call graph node elements.  The XML document is processed using a
    * top-down recursive procedure that creates a call graph with identical
    * structure and data.
    *
    * @throws  AstException
    *          if the XML parser has a failure during parsing or in its    
    *          file system operations.
    * @throws  IllegalArgumentException
    *          if the input file is invalid or doesn't exist.
    */
   public CallGraphNode read()
   throws AstException,
          IllegalArgumentException
   {
      Element root = null;
      
      try
      {
         if (file.exists() && file.isFile())
         {
            dom  = XmlHelper.parse(file);
            root = dom.getDocumentElement();            
         }
         else
         {
            throw new IllegalArgumentException("Missing/bogus input file " +
                                               file.toString());
         }
      }
      catch (Exception excpt)
      {
         throw new AstException(
            "Error loading CallGraph XML file: " + file.toString(), excpt);
      }      
      
      // the list of children will include #Text and other DOM defined
      // trash... get the first of our children
      Node cgn = root.getFirstChild();

      // search for the child that has a tagName of ELEM_CGN
      while (cgn != null && cgn.getNodeName().equals(ELEM_CGN) == false)
      {
         cgn = cgn.getNextSibling();
      }
      
      // make sure something was found
      if (cgn == null)
      {
         throw new AstException("Failure reading root element " +
                                  root.getNodeName());
      }
      
      // there must always be ONLY a single call graph element under this
      // root element!  recursively read the call graph nodes and build the
      // tree...
      CallGraphNode result = (CallGraphNode)readCgn((Element)cgn);
      
      return result;  
   }
   
   /**
    * Create the target XML document, add comments and a timestamp, then
    * persist the entire call graph to the file through a top-down recursive
    * processing of the tree.
    *
    * @param   cgn
    *          Root node of the tree to write to the XML file.
    *
    * @throws  AstException
    *          if it is not possible to create an empty XML document or in    
    *          the case that the XML document cannot be stored in the file
    *          system.
    * @throws  IllegalArgumentException
    *          if the CallGraphNode to persist is <code>null</code>.
    */
   public void write(CallGraphNode cgn)
   throws AstException,
          IllegalArgumentException
   {
      if (cgn == null)
      {
         throw new IllegalArgumentException("Cannot persist null " +
                                            "CallGraphNode.");
      }
      
      try
      {
         dom = XmlHelper.newDocument();
      }
      catch (ParserConfigurationException exc)
      {
         throw new AstException("Failure to create new XML document.", exc);
      }
      
      // add comments
      Comment comment = dom.createComment("Persisted CallGraph");
      dom.appendChild(comment);
      
      // create root element
      Element elem = dom.createElement(ELEM_ROOT);
      dom.appendChild(elem);
      
      // recursively write the call graph out to this document, as children 
      // of this root element
      writeCgn(cgn, (Node)elem);
      
      // persist the XML document as a file
      persist();
   }
   
   /**
    * Write a CallGraphNode node to XML, recursively calling itself to write 
    * all children.  All data is stored as attributes and child nodes are
    * written as contained elements.
    *
    * @param   cgn
    *          The node to persist.
    * @param   parent
    *          Parent node in the XML document, to which the associated
    *          element should be appended as a child.
    */
   private void writeCgn(CallGraphNode cgn, Node parent)
   {
      // create an empty node of the proper element type 
      Element elem = dom.createElement(ELEM_CGN);
      
      // store the AST's data as attributes
      elem.setAttribute(ATTR_FILE, cgn.getFilename());      
      
      // set the CallGraphNode as a child of the parent node
      parent.appendChild(elem);
      
      // recursively process all children
      CallGraphNode child = (CallGraphNode) cgn.getFirstChild();
      while (child != null)
      {
         writeCgn(child, (Node)elem);
         child = (CallGraphNode) child.getNextSibling();
      }      
   }   
   
   /**
    * Complete the processing by serializing the XML document to the output
    * file.
    *
    * @throws  AstException
    *          if any error occurs during XML serialization.
    */
   private void persist()
   throws AstException
   {
      try
      {
         // delete any previous version that may exist
         if (file.exists() && file.isFile())
         {
            file.delete();
         }
         
         XmlHelper.write(dom, file);
      }
      catch (Exception exc)
      {
         throw new AstException(
            "Error serializing XML file " + file.toString(), exc);
      }
   }   
   
   /**
    * Processes a call graph node element in an XML document to create an
    * associated call graph node instance and then recursively calls itself 
    * for each node element that is a child.  The returned node is then added
    * as a child of the current node and when all children are processed, the
    * current CallGraphNode is returned.
    *
    * @param    current
    *           Represents an element in the XML document that is a call
    *           graph node element and may contain 0 or more call graph node
    *           elements.           
    *
    * @return   A call graph that has the identical structure and data
    *           as the XML element passed as a parameter.

    */
   private CallGraphNode readCgn(Element current)
   throws AstException
   {      
      // create an call graph node for the current XML node being processed
      CallGraphNode parent = createCallGraphNode(current);
      
      // read first child
      Node cgn = current.getFirstChild();

      // for each child, create a subtree and then attach it to ourselves
      while (cgn != null)
      {
         if (cgn.getNodeName().equals(ELEM_CGN))
         {
            parent.addChild(readCgn((Element)cgn));         
         }
         
         cgn = cgn.getNextSibling();
      }
      
      return parent;
   }   
   
   /**
    * Reads a call graph node associated with the passed-in XML element and 
    * creates a new CallGraphNode based on the associated attributes.
    *
    * @param    current
    *           The XML call graph node element.
    *
    * @return   The created CallGraphNode.
    *
    * @throws   AstException
    *           if an unknown data type is encountered.
    */
   private CallGraphNode createCallGraphNode(Element current)
   throws AstException
   {
      CallGraphNode cgn = null;
      
      try
      {
         // read the associated data
         String filename  = current.getAttribute(ATTR_FILE);
                  
         // create the CallGraphNode
         cgn = new CallGraphNode(filename, true);         
      }
      
      catch (Exception exc)
      {
         throw new AstException("Failure to create CallGraphNode for " +
                                 current.getTagName());
      }
      
      return cgn;
   }
}