/*
** Module   : CallGraphWorker.java
** Abstract : provides a pattern engine service to create a new call graph
**            based on matches found during AST processing
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050311   @20310 Created initial version that provides
**                           support for creating the call graph in
**                           memory during pattern engine processing. It
**                           provides a dynamic iterator that walks the
**                           call graph and returns File objects for
**                           each node until the entire graph is complete.
**                           This iterator changes its return value
**                           dynamically based on inline changes to the
**                           graph. This is the key to driving the
**                           pattern engine from a single root entry
**                           point and being able to create an entire
**                           graph of any number of ASTs since the next
**                           AST for the pattern engine is read from the
**                           call graph that was just created by pattern
**                           engine processing.
** 002 GES 20050314   @20321 Added support for RUN VALUE() based call
**                           graph nodes using hints.
** 003 GES 20050316   @20397 Converted this worker to use the visitAst()
**                           interface in PatternWorker instead of using
**                           a rule driven approach.  
** 004 ECF 20050526   @21290 Changes to support new expression engine
**                           implementation. Changed library registration
**                           mechanism based on superclass' modifications
**                           to support single library per pattern worker
**                           limit.
** 005 SIY 20050531   @21341 Minimal changes to enable CallGraphWorker 
**                           use .ast files instead of .p.
** 006 SIY 20050606   @21453 Removed redundant methods for the call graph
**                           processing. Added methods for the "dead" 
**                           files processing. Minor cleanups.  
** 007 SIY 20050622   @21544 Added absolute application path handling.
** 008 SIY 20051011   @23018 Fixed null pointer exception during loading of 
**                           the hints file. 
** 009 GES 20060129   @24131 Minor name change for FileList.
** 010 GES 20060130   @24140 Safety code added to ensure that this worker
**                           could be used > 1 time per JVM process.
** 011 GES 20060130   @24157 Honor debug settings before writing output
**                           to the terminal.
** 012 SIY 20060312   @24981 Improved include files path handling to avoid
**                           incorrect marking them as not referenced.
** 013 GES 20090429   @42061 Match package and class name changes.
** 014 GES 20090514   @42233 Removed dead code.
** 015 CA  20140313          Moved to a graph DB implementation.  Added support for various 
**                           call-sites, mos
** 016 CA  20140408          Fixed filename matching for source-code originating from a different
**                           OS than the one where conversion is ran.
** 017 CA  20140424          Fixed disambiguation of call sites for which the hints provide no 
**                           targets.
** 018 ECF 20160407          Added TODO in loadIncludeFiles to check hints for an include-spec
**                           setting.
** 019 OM  20170123          Removed dead code.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;

import org.apache.commons.configuration.*;

import com.goldencode.ast.*;
import com.goldencode.graphdb.*;
import com.goldencode.p2j.cfg.Configuration;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.preproc.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;
import com.thinkaurelius.titan.core.*;
import com.thinkaurelius.titan.core.attribute.Text;
import com.tinkerpop.blueprints.*;

/**
 * Provides a pattern engine service to create a new call graph based on matches found during AST
 * processing.  Only one call graph can be created at a time due to use of static members.
 * <p>
 * This call graph contains external file references, schema tables and external link sites (as
 * native processes or web services) AND all duplications have been removed.
 * <p>
 * This worker provides generic routines, so other call-sites can be added in the future.
 * <p>
 * Ambiguous call-sites can be disambiguated using UAST hints to specify 1 or more targets (see
 * {@link UastHints}); note that such hints are named using the call-site-key as prefix,
 * then an instance number specifying the 0-based index of this match in the given source file.
 */
public class CallGraphWorker
extends AbstractPatternWorker
{
   /** Default folder for the callgraph DB. */
   public static final String DEFAULT_CALLGRAPH_DB_FOLDER = "./callgraph/";
   
   /**
    * List of targets to be processed. This starts with the initial entry-points and acts
    * as a queue to which called external programs are added.
    */
   private static final LinkedList<Vertex> targets = new LinkedList<>();

   /** Id of the node to which ambiguous call-sites link. */
   private static final long AMBIGUOUS_NODE_ID = -1;

   /** Case sensitive state of the file names. */
   private static Boolean caseSensitive;
   
   /** The connected graph DB. */
   private static Graph db = null;

   /** Stores the current AST's hints object. */
   private UastHints hints = null;

   /** Stores the current AST's preprocessor hints object. */
   private PreprocessorHints pphints = null;

   /** <code>true</code> if no output should be written to the terminal. */
   private boolean silent = false;
   
   /** Flag indicating if we are starting to collect targets. */
   private boolean collectTargets = false;
   
   /**
    * Default constructor which calls the super-class constructor, registers
    * its libraries and initializes its instance members.
    * 
    * @throws  AstException
    *          Forwarded from underlying methods.
    */
   public CallGraphWorker() throws AstException
   {
      super();
      setLibrary(new CallGraph());
      
      // honor debug settings
      silent = (PatternEngine.getDebugLevel() == PatternEngine.MSG_NONE);
      
      // initialize the graph DB
      getGraph();
   }
   
   /**
    * Check if we are in a case-sensitive OS, by reading the {@code case-sensitive} configuration
    * parameter.
    * 
    * @return   {@code true} if the OS filenames are case-sensitive.
    */
   public static boolean isCaseSensitive()
   {
      if (caseSensitive == null)
      {
         caseSensitive = 
            "true".equalsIgnoreCase(Configuration.getParameter("case-sensitive", "false"));
      }
      
      return caseSensitive;
   }
   
   /**
    * Find an unique node having a property with the given key and value.
    * 
    * @param    key
    *           The property key.
    * @param    val
    *           The property value.
    *           
    * @return   See above.
    * 
    * @throws   RuntimeException
    *           If more than one node is found.
    */
   public static Vertex findUniqueNode(String key, Object val)
   {
      Vertex res = null;
      
      for (Vertex v : db.query().has(key, val).vertices())
      {
         if (res == null)
         {
            res = v;
         }
         else
         {
            final String msg = "More than one node found with key [%s] and value [%s]!";
            throw new RuntimeException(String.format(msg, key, val));
         }
      }

      return res;
   }

   /**
    * Find a node by its type and ID.
    * 
    * @param    nodeType
    *           The node's type.
    * @param    id
    *           The node's ID.
    * 
    * @return   The found node or {@code null} if it does not exist.
    */
   public static Vertex findNodeById(int nodeType, Object id)
   {
      Iterator<Vertex> iter = db.query().has("node-type", nodeType).has("node-id", id)
                                .vertices().iterator();
      
      return iter.hasNext() ? iter.next() : null;
   }
   
   /**
    * Initialize the {@link #db graph DB} as necessary and return it.
    * <p>
    * The graph DB will connect to the folder specified via the {@code callgraph-db-folder} 
    * parameter and it will default to {@link #DEFAULT_CALLGRAPH_DB_FOLDER}, if not set.
    * <p>
    * The graph DB is configured for a Titan implementation, backed by {@code persistit} storage 
    * and with {@code lucene} indexes.
    * 
    * @return   The graph DB instance associated with the call-graph.
    */
   public static Graph getGraph()
   {
      if (db == null)
      {
         String folder = Configuration.getParameter("callgraph-db-folder",
                                                    DEFAULT_CALLGRAPH_DB_FOLDER);

         org.apache.commons.configuration.Configuration config = new BaseConfiguration();
         config.setProperty("storage.directory", folder);
         config.setProperty("storage.backend", "persistit");
         config.setProperty("storage.index.search.backend", "lucene");
         config.setProperty("storage.index.search.directory",
                            folder + File.separator + "searchindex");

         db = GraphDB.obtainGraphDB(config);
         
         // unique indexes
         GraphDB.createUniqueIndex(db, "os-filename",String.class, Vertex.class, true, "search");
         GraphDB.createUniqueIndex(db, "filename",   String.class, Vertex.class, true, "search");
         GraphDB.createUniqueIndex(db, "reverse-filename", String.class, Vertex.class, true, 
                                   "search");
         GraphDB.createUniqueIndex(db, "port-type",  String.class, Vertex.class, true, "search");
         GraphDB.createUniqueIndex(db, "schemaname", String.class, Vertex.class, true, "search");
         GraphDB.createUniqueIndex(db, "command",    String.class, Vertex.class, true, "search");
         GraphDB.createUniqueIndex(db, "connect-string", String.class, Vertex.class, true,
                                   "search");
         GraphDB.createUniqueIndex(db, "com-target", String.class, Vertex.class, true, "search");
         GraphDB.createUniqueIndex(db, "dde-target", String.class, Vertex.class, true, "search");
          
         // search indexes TODO: (node-type, node-id) make an unique index
         GraphDB.createIndex(db, "node-type",    Integer.class, Vertex.class, false, "search");
         GraphDB.createIndex(db, "node-id",      Long.class,    Vertex.class, false, "search");
         GraphDB.createIndex(db, "call-site-id", Long.class,    Edge.class,   false, "search");
      }
      
      return db;
   }

   /**
    * Shutdown the graph DB.
    */
   public static void shutdownGraph()
   {
      if (db != null)
      {
         db.shutdown();
         
         db = null;
      }
   }
   
   /**
    * Prepare the procedure file name by replacing all windows-style separators with linux-style 
    * separators and making the name lowercase, if we are on a case-insensitive OS.
    * 
    * @param   filename
    *          Source file name.
    * 
    * @return  Prepared file name.
    */
   public static String prepareFilename(String filename)
   {
      if (filename.trim().equals(""))
         return null;
   
      // TODO: if we have been provided an absolute path, these kind of files need to be marked in 
      //  the graph, because an absolute path hard-coded in the 4GL code will not resolve well at 
      // runtime.  all paths need to be relative to basepath
   
      // transform to unix-like separators
      if (PlatformHelper.isUnderWindowsFamily())
      {
         filename = filename.replaceAll("\\\\", "/");
      }
      
      if (!isCaseSensitive())
      {
         filename = filename.toLowerCase();
      }
      
      return filename;
   }
   
   /**
    * Creates a new iterator that uses as initial nodes the explicit list of programs passed at
    * rootNodes parameter and also all the external programs with ambiguous call-sites.
    * <p>
    * The explicit targets in the rootNodes array will not be marked as entry-points.
    * <p>
    * The call graph is walked dynamically, using the nodes as the tree is being built to properly
    * direct the pattern engine in its processing, and avoiding already-processed nodes.
    *
    * @param    rootNodes
    *           List of root AST names to use as entry points.
    * @param    projectRelative
    *           Specifies if the name is relative to the project home directory.  If 
    *           <code>false</code>, the name is either an absolute filename or it is relative to
    *           the current directory.
    * 
    * @return   <code>Iterator</code> instance suitable for the node list traversal.
   */
   static Iterator<Vertex> newUpdateCallGraphIterator(String[] rootNodes, boolean projectRelative)
   {
      targets.clear();
      
      loadRootNodes(rootNodes, projectRelative, false);
      
      // ensure the AST Manager is initialized
      new PatternEngine();
      AstManager mgr = AstManager.get();
      
      // load the ambiguous nodes
      Vertex ambiguous = findNodeById(ProgressParserTokenTypes.AMBIGUOUS, AMBIGUOUS_NODE_ID);
      Iterator<Vertex> itr = ambiguous.query().direction(Direction.IN).vertices().iterator();
      Set<String> ambigExtProgs = new HashSet<>();
      while (itr.hasNext())
      {
         Vertex node = itr.next();
         
         String filename = mgr.getTreeName((long) node.getProperty("node-id"));
         if (filename != null)
         {
            filename = prepareFilename(filename);
            Vertex extProg = findUniqueNode("filename", filename);
            if (!ambigExtProgs.contains(filename))
            {
               ambigExtProgs.add(filename);

               extProg.setProperty("processing", true);
               targets.add(extProg);
            }
         }
      }
      
      return targetsIterator();
   }
   
   /**
    * Creates a new iterator that walks a call graph dynamically, using the nodes as the tree is
    * being built to properly direct the pattern engine in its processing.  Advancing to the next
    * node is done when the tree is changed rather than when the next() method is called.
    * <p>
    * The explicit targets in the rootNodes array will be marked as entry-points.
    *
    * @param    rootNodes
    *           List of root AST names to use as entry points.
    * @param    projectRelative
    *           Specifies if the name is relative to the project home directory.  If 
    *           <code>false</code>, the name is either an absolute filename or it is relative to
    *           the current directory.
    * 
    * @return   <code>Iterator</code> instance suitable for the node list traversal.
   */
   static Iterator<Vertex> newCallGraphIterator(String[] rootNodes, boolean projectRelative)
   {
      // clear the targets and load the root nodes
      targets.clear();

      loadRootNodes(rootNodes, projectRelative, true);

      return targetsIterator();
   }
   
   /**
    * Create a new iterator over the {@link #targets} queue, by polling elements from the
    * head of the queue until the queue is empty.
    * 
    * @return   See above.
    */
   private static Iterator<Vertex> targetsIterator()
   {
      return new Iterator<Vertex>()
      {
         @Override
         public void remove()
         {
            throw new UnsupportedOperationException();
         }

         @Override
         public Vertex next()
         {
            return targets.pollFirst();
         }

         @Override
         public boolean hasNext()
         {
            return !targets.isEmpty();
         }
      };
   }
   
   /**
    * Load the specified root nodes into the {@link #targets}.
    * 
    * @param    rootNodes
    *           List of root AST names to use as entry points.
    * @param    projectRelative
    *           Specifies if the name is relative to the project home directory.  If 
    *           <code>false</code>, the name is either an absolute filename or it is relative to
    *           the current directory.
    * @param    entryPoint
    *           Flag to mark or not the nodes as entry-points.
    */
   private static void loadRootNodes(String[] rootNodes, 
                                     boolean  projectRelative, 
                                     boolean  entryPoint)
   {
      if (rootNodes == null)
      {
         return;
      }
      
      for (int i = 0; i < rootNodes.length; i++)
      {
         String filename = rootNodes[i];

         if (!projectRelative)
         {
            // convert to project-relative
            filename = Configuration.normalizeFilename(filename);
         }
         
         // discard the .ast suffix
         if (filename.endsWith(".ast"))
         {
            filename = filename.substring(0, filename.length() - ".ast".length());
         }
         else
         {
            System.err.println("Non-AST entry point received:" + filename);
            continue;
         }
         
         // when using physical files, search by OS filename
         Iterator<Vertex> itrv = db.query().has("os-filename", filename).vertices().iterator();
         Vertex v = null;
         if (itrv.hasNext())
         {
            v = itrv.next();

            if (itrv.hasNext())
            {
               throw new RuntimeException("More than one file named " + filename + " found!");
            }
         }
         else
         {
            System.err.println("Entry point not found:" + filename);
            continue;
         }
         
         if (entryPoint)
         {
            v.setProperty("entry-point", true);
         }

         v.setProperty("processing", true);

         targets.add(v);
      }
   }
   
   /**
    * Resets the current hints object and the counters for all call-sites and each time a new AST
    * is visited.  This enables the lazy loading of hints files only when needed and provides the
    * ability to calculate the hint name corresponding to a specific call-site in a given file.
    *
    * @param   ast
    *          The root node of the source AST about to be processed by the pattern engine.
    */
   public void visitAst(Aast ast)
   {
      hints = null;
      pphints = null;
   }

   /**
    * Provides a service for creating call graph nodes in a tree starting at some root node.
    * This class is a <code>PatternEngine</code> worker.
    */
   public class CallGraph
   {
      /**
       * Set the new state of the {@link CallGraphWorker#collectTargets} flag.
       * 
       * @param    state
       *           {@code true} to start collecting new targets, {@code false} otherwise.
       */
      public void setCollectTargets(boolean state)
      {
         collectTargets = state;
      }
      
      /**
       * Get the include source file for this AST.  If the source is not from an include file,
       * return {@code null}.
       * 
       * @param    ast
       *           The AST for which the include source is needed.
       *           
       * @return   See above.
       * 
       * @throws   PreprocessorException
       *           An unknown or unexpected failure.
       */
      public String getIncludeSourceFor(Aast ast)
      throws PreprocessorException
      {
         PreprocessorHints hints = loadPreprocessorHints(ast);
         
         int line = ast.getLine();
         int column = ast.getColumn();
         
         if (line == 0 || column == 0)
         {
            final String msg = "ERROR: can not determine the include file for synthetic ASTs, " +
                               "with no line/column information: %d \n %s";
            System.err.println(String.format(msg, ast.getId(), ast.dumpTree()));
            return null;
         }
         
         String includeSource = hints.getSourceFor(line, column);
         includeSource = prepareFilename(includeSource);
         String filename = prepareFilename(ast.getFilename());

         return (includeSource.equals(filename)) ? null : includeSource;
      }
      
      /**
       * Prepare the procedure file name by replacing all windows-style separators with linux-style 
       * separators and making the name lowercase, if we are on a case-insensitive OS.
       * 
       * @param   filename
       *          Source file name.
       * 
       * @return  Prepared file name.
       */
      public String prepareFilename(String filename)
      {
         return CallGraphWorker.prepareFilename(filename);
      }
      
      
      /**
       * Check if we are in a case-sensitive OS, by reading the {@code case-sensitive} configuration
       * parameter.
       * 
       * @return   {@code true} if the OS filenames are case-sensitive.
       */
      public boolean isCaseSensitive()
      {
         return CallGraphWorker.isCaseSensitive();
      }

      /**
       * Get the {@link Direction#OUT} for an edge.
       * 
       * @return   See above.
       */
      public Direction out()
      {
         return Direction.OUT;
      }
      
      /**
       * Get the {@link Direction#IN} for an edge.
       * 
       * @return   See above.
       */
      public Direction in()
      {
         return Direction.IN;
      }
      
      /**
       * Get the graph node to which all ambiguous call-sites link. This is temporary until the
       * call-site ASTs are added into the callgraph.
       * 
       * @return   See above.
       */
      public Vertex ambiguousNode()
      {
         Vertex node = findNodeById(ProgressParserTokenTypes.AMBIGUOUS, AMBIGUOUS_NODE_ID);
         if (node == null)
         {
            node = createNode(AMBIGUOUS_NODE_ID, ProgressParserTokenTypes.AMBIGUOUS);
         }
         
         return node;
      }
      
      /**
       * For the specified node, remove all the edges in the given direction which match the
       * label, property key and property value.
       * 
       * @param    node
       *           A graph node.
       * @param    dir
       *           The edge's direction.
       * @param    label
       *           The edge's label.
       * @param    key
       *           The property's key.
       * @param    value
       *           The property's value.
       */
      public void removeEdges(Vertex node, Direction dir, String label, String key, Object value)
      {
         for (Edge edge : node.query().labels(label).direction(dir)
                              .has(key, value).edges())
         {
            edge.remove();
         }
      }
      
      /**
       * For the specified node, remove all the edges in the given direction which match the 
       * property key and property value.
       * 
       * @param    node
       *           A graph node.
       * @param    dir
       *           The edge's direction.
       * @param    key
       *           The property's key.
       * @param    value
       *           The property's value.
       */
      public void removeEdges(Vertex node, Direction dir, String key, Object value)
      {
         for (Edge edge : node.query().direction(dir).has(key, value).edges())
         {
            edge.remove();
         }
      }
      
      /**
       * Create a new graph DB node associated with the given AST, representing a permanent schema
       * table/
       * 
       * @param    ast
       *           The schema table AST.
       *           
       * @return   The created {@link Vertex node}.
       */
      public Vertex addTable(Aast ast, String schemaname)
      {
         Vertex node = createAstNode(ast, ProgressParserTokenTypes.TABLE);
         node.setProperty("schemaname", schemaname);
         node.setProperty("node-key", "schemaname");
         
         return node;
      }
      
      /**
       * Create a new graph DB node associated with the given AST, representing an external
       * procedure.
       * 
       * @param    ast
       *           The external-procedure AST.
       *           
       * @return   The created {@link Vertex node}.
       */
      public Vertex addExternalProcedure(Aast ast)
      {
         Vertex node = createAstNode(ast, ProgressParserTokenTypes.EXTERNAL_PROCEDURE);

         // need to use linux-style separators
         String fullName = prepareFilename(ast.getFilename());
         node.setProperty("os-filename", ast.getFilename());
         node.setProperty("filename", fullName);
         node.setProperty("reverse-filename", reverseString(fullName));
         node.setProperty("node-key", "filename");

         return node;
      }
      
      /**
       * Find an edge between the two nodes, which has the specified label and a property with
       * the given key and value.
       * 
       * @param    out
       *           The source node.
       * @param    in
       *           The target node.
       * @param    label
       *           The edge's label.
       * @param    key
       *           The property's key.
       * @param    value
       *           The property's value.
       *           
       * @return   The first found edge or {@code null} if not found.
       */
      public Edge findEdge(Vertex out, Vertex in, String label, String key, Object value)
      {
         // check if we already have an edge or not between these two
         for (Edge edge : out.query().labels(label).has(key, value).edges())
         {
            if (edge.getVertex(Direction.IN).equals(in))
            {
               return edge;
            }
         }
         
         return null;
      }
      
      /**
       * Create a graph DB node associated with the given AST.
       * 
       * @param    ast
       *           The associated AST.
       * @param    type
       *           The node's type, depending on the AST type.
       *           
       * @return   The new graph DB node.
       */
      public Vertex createAstNode(Aast ast, int type)
      {
         // create a new node
         Vertex node = createNode(ast.getId(), type);
         // load all annotations from this AST
         node.setProperty("column", ast.getColumn());
         node.setProperty("line",   ast.getLine());
         node.setProperty("text",   ast.getText());
         node.setProperty("type",   ast.getType());
      
         @SuppressWarnings("unchecked")
         Iterator<String> itr = ast.annotationKeys();
         while (itr.hasNext())
         {
            String key = itr.next();
            Object val = ast.getAnnotation(key);
      
            node.setProperty(key, val);
         }
      
         return node;
      }

      /**
       * Create an external node with the given type.
       * 
       * @param    type
       *           The node's type.
       *           
       * @return   The created {@link Vertex node}.
       */
      public Vertex createExternalNode(int type)
      {
         Vertex node = createNode(nextExternalId(type), type);
         node.setProperty("external", true);
         
         return node;
      }
      
      /**
       * Create a new {@link Vertex node} with the specified ID and type.
       * 
       * @param    id
       *           The node's ID.
       * @param    type
       *           The node's type.
       *           
       * @return   The created {@link Vertex node}.
       */
      public Vertex createNode(long id, int type)
      {
         Vertex node = db.addVertex(null);
         node.setProperty("node-type", type);
         node.setProperty("node-id", id);
         
         return node;
      }
      
      /**
       * Create a new edge between the specified nodes, with the given label.
       * 
       * @param    out
       *           The source node.
       * @param    in
       *           The target node.
       * @param    label
       *           The edge label.
       *           
       * @return   The created {@link Edge}.
       */
      public Edge createEdge(Vertex out, Vertex in, String label)
      {
         Edge edge = db.addEdge(null, out, in, label);
         // this is useful only for the GraphML - as the edge's label is not serialized
         edge.setProperty("edge-type", label);
         
         return edge;
      }
      
      /**
       * Create an iterator for all the nodes linked with an outbound edge which has a property 
       * with the specified key and value.
       *
       * @param    node
       *           The graph node.
       * @param    key
       *           The property key.
       * @param    value
       *           The property value.
       * 
       * @return   See above.
       */
      public Iterator<Vertex> nodeIterator(Vertex node, String key, Object value)
      {
         return node.query().labels().has(key, value).vertices().iterator();
      }
      
      /**
       * Create an iterator for all the nodes having a property with the specified key, regardless
       * of value.
       * 
       * @param    key
       *           The property key.
       * 
       * @return   See above.
       */
      public Iterator<Vertex> nodeIterator(String key)
      {
         return db.query().has(key).vertices().iterator();
      }

      /**
       * Create a BFS node iterator starting with the given root, in the given direction.
       * 
       * @param    root
       *           The root node.
       * @param    dir
       *           The edge direction.
       *           
       * @return   See above.
       */
      public Iterator<Vertex> traversalIterator(Vertex root, final Direction dir)
      {
         // start with our root
         final LinkedList<Vertex> targets = new LinkedList<>();
         targets.add(root);
         final Set<Vertex> traversed = new HashSet<Vertex>();
         traversed.add(root);
         
         return new Iterator<Vertex>()
         {
            @Override
            public void remove()
            {
               throw new UnsupportedOperationException();
            }
            
            @Override
            public Vertex next()
            {
               Vertex next = targets.removeFirst();
               // collect all edges in the given direction
               for (Vertex node : next.query().direction(dir).vertices())
               {
                  if (!traversed.contains(node))
                  {
                     targets.addLast(node);
                     traversed.add(node);
                  }
               }
               
               return next;
            }
            
            @Override
            public boolean hasNext()
            {
               return !targets.isEmpty();
            }
         };
      }

      /**
       * Create a BFS node iterator starting with the given root.
       * 
       * @param    root
       *           The root node.
       * @param    label
       *           The labels for the edges which will be followed.
       *           
       * @return   See above.
       */
      public Iterator<Vertex> traversalIterator(Vertex root, final String label)
      {
         // start with our root
         final LinkedList<Vertex> targets = new LinkedList<>();
         targets.add(root);
         final Set<Vertex> traversed = new HashSet<Vertex>();
         traversed.add(root);
         
         return new Iterator<Vertex>()
         {
            @Override
            public void remove()
            {
               throw new UnsupportedOperationException();
            }
            
            @Override
            public Vertex next()
            {
               Vertex next = targets.removeFirst();
               // collect all outbound edges with the given label
               for (Vertex node : next.query().labels(label).direction(Direction.OUT).vertices())
               {
                  if (!traversed.contains(node))
                  {
                     targets.addLast(node);
                     traversed.add(node);
                  }
               }
               
               return next;
            }
            
            @Override
            public boolean hasNext()
            {
               return !targets.isEmpty();
            }
         };
      }
      
      /**
       * Create an iterator for all the nodes having a property with the specified key and value.
       * 
       * @param    key
       *           The property key.
       * @param    value
       *           The property value.
       * 
       * @return   See above.
       */
      public Iterator<Vertex> nodeIterator(String key, Object value)
      {
         return db.query().has(key, value).vertices().iterator();
      }

      /**
       * Sort the given elements by the specified properties.
       * 
       * @param    elements
       *           A query result.
       * @param    props
       *           The property names, each element a {@link String}.
       *           
       * @return   The sorted elements.
       */
      @SuppressWarnings("unchecked")
      public Iterator<Element> sortedResults(Iterable<Element> elements, final Object[] props)
      {
         if (!elements.iterator().hasNext())
         {
            return Collections.EMPTY_LIST.iterator();
         }

         List<Element> sorted = new ArrayList<>();
         for (Element e : elements)
         {
            sorted.add(e);
         }
         
         Collections.sort(sorted, new Comparator<Element>()
         {
            @Override
            public int compare(Element o1, Element o2)
            {
               for (int i = 0; i < props.length; i++)
               {
                  Object v1 = o1.getProperty((String) props[i]);
                  Object v2 = o2.getProperty((String) props[i]);
                  
                  if (v1 != null && v1.equals(v2))
                  {
                     continue;
                  }
                  else if (v1 instanceof Comparable)
                  {
                     return ((Comparable<Object>) v1).compareTo(v2);
                  }
                  else
                  {
                     return 0;
                  }
               }
               
               return 0;
            }
         });
         
         return sorted.iterator();
      }
      
      /**
       * Create an iterator for all outbound edges with the given label.
       * 
       * @param    node
       *           The node for which the outbound edges are needed.
       * @param    label
       *           The edge's label.
       *           
       * @return   See above.
       */
      public Iterator<Edge> edgeIterator(Vertex node, String label)
      {
         return node.query().labels(label).direction(Direction.OUT).edges().iterator();
      }

      /**
       * Create an iterator for all outbound edges with the given label and property key and value.
       * 
       * @param    node
       *           The node for which the outbound edges are needed.
       * @param    label
       *           The edge's label.
       * @param    key
       *           Only edges with the having a property with this key.
       * @param    value
       *           Only edges with the having a property with this value.
       *           
       * @return   See above.
       */
      public Iterator<Edge> edgeIterator(Vertex node, String label, String key, Object value)
      {
         return node.query().labels(label).has(key, value)
                    .direction(Direction.OUT).edges().iterator();
      }

      /**
       * Create an iterator for all inbound edges with the given label.
       * 
       * @param    node
       *           The node for which the inbound edges are needed.
       * @param    label
       *           The edge's label.
       *           
       * @return   See above.
       */
      public Iterator<Edge> edgeInIterator(Vertex node, String label)
      {
         return node.query().labels(label).direction(Direction.IN).edges().iterator();
      }
      
      /**
       * Find a node by its type and ID.
       * 
       * @param    nodeType
       *           The node's type.
       * @param    id
       *           The node's ID.
       *
       * @return   The found node or {@code null} if it does not exist.
       */
      public Vertex findNodeById(int nodeType, Object id)
      {
         return CallGraphWorker.findNodeById(nodeType, id);
      }

      /**
       * Add a new target, to be processed.
       * <p>
       * Only {@code external-procedure} targets with no {@code finished} or {@code processing}
       * annotations are added.
       * 
       * @param    node
       *           The node to be added.
       */
      public void addTarget(Vertex node)
      {
         if (!collectTargets)
         {
            return;
         }
         
         // add only AST targets and skip finished IDs and WIP IDs
         if (node.getProperty("node-type").equals(ProgressParserTokenTypes.EXTERNAL_PROCEDURE) &&
             node.getProperty("finished") == null                                              && 
             node.getProperty("processing") == null)
         {
            targets.addLast(node);
            node.setProperty("processing", true);
         }
      }
      
      /**
       * Commit the current transaction.
       */
      public void commit()
      {
         ((TransactionalGraph) db).commit();
      }
      
      /**
       * Find an unique node having a property with the given key and value.
       * 
       * @param    key
       *           The property key.
       * @param    val
       *           The property value.
       *           
       * @return   See above.
       * 
       * @throws   RuntimeException
       *           If more than one node is found.
       */
      public Vertex findUniqueNode(String key, Object val)
      {
         return CallGraphWorker.findUniqueNode(key, val);
      }

      /**
       * Load all the include files (physical files) into the graph DB, as distinct nodes. This
       * ensures that unreferenced/dead include files can be determined.
       */
      public void loadIncludeFiles()
      {
         String basepath = Configuration.getParameter("basepath", "./");
         String includeSpec = Configuration.getParameter("include-spec", "*.[iI]");
         // TODO: use UAST hints to override
         System.out.println(String.format("Using %s for include-file specifications.", 
                                          includeSpec));
         
         FileSpecList list = new FileSpecList(new File(basepath), includeSpec, true);
         File[] includes = list.list();
         int counter = 0;
         for (File fi : includes)
         {
            String filename = fi.getPath();
            System.out.println(filename);
            Vertex vnode = findUniqueNode("os-filename", filename);
            if (vnode == null)
            {
               createIncludeNode(filename);
               if (counter % 100 == 0)
               {
                  // do not allow the changes to build up in the transaction, but do not commit 
                  // after each operation either, as this is expensive
                  commit();
               }
               counter++;
            }
         }
         // commit the remainder transaction
         commit();
         
         System.out.println("Done adding include-files.");
      }

      /**
       * Create the include graph for the given external program.
       * 
       * @param    ast
       *           The external program AST.
       *           
       * @throws   PreprocessorException
       *           An unknown or unexpected failure.
       */
      public void createIncludeGraph(Aast ast)
      throws PreprocessorException
      {
         PreprocessorHints ppHints = loadPreprocessorHints(ast);

         HintsTreeNode root = ppHints.getRoot();
         if (root != null)
         {
            long astId = ast.getId();

            // the root at this call is for the external program, not for the include file
            Vertex parent = findNodeById(ProgressParserTokenTypes.EXTERNAL_PROCEDURE, astId);
            
            createIncludeGraph(root, parent);
         }
      }

      /**
       * Get the targets for a specific call-site (which is determined to be ambiguous), 
       * from the hints file.
       * 
       * @param    callSiteKey
       *           The key of this call-site.
       * @param    instance
       *           The instance number for this call-site.
       * 
       * @return   A map of targets, where the key represents the hint no (and the index in the
       *           string array) for the target.
       */
      public Map<String, String> getCallSiteHintTargets(int callSiteKey, int instance)
      {
         StringBuilder sb = new StringBuilder(ProgressParser.lookupTokenName(callSiteKey));
         sb.append("_");
         sb.append(Integer.toString(instance));
         String key = sb.toString();
         
         String cls = hints.getUastType(key);
         
         if (cls == null)
         {
            // bypass other logic and return null
            return null;
         }
         
         Map<String, String> res = new HashMap<>();
         
         if (String.class.getName().equals(cls))
         {
            String target = hints.getUastString(key);
            
            List<String> extProgs = resolveExternalPrograms(target);
            for (int j = 0; j < extProgs.size(); j++)
            {
               String extProg = extProgs.get(j);
               if (extProgs.size() == 1)
               {
                  res.put(Integer.toString(instance), target.trim());
               }
               else
               {
                  res.put(String.format("%d %d", instance, j), extProg);
               }
            }
         }
         else
         {         
            // second try, maybe the hint is an array (we will get an exception if this isn't 
            // the case... this is OK)
            String[] targets = hints.getUastStringArray(key);
            
            // process each entry in the array
            for (int i = 0; i < targets.length; i++)
            {
               String target = targets[i];
               
               List<String> extProgs = resolveExternalPrograms(target);
               for (int j = 0; j < extProgs.size(); j++)
               {
                  String extProg = extProgs.get(j);
                  if (extProgs.size() == 1)
                  {
                     res.put(String.format("%d_%d", instance, i), extProg);
                  }
                  else
                  {
                     res.put(String.format("%d_%d %d", instance, i, j), extProg);
                  }
               }
            }
         }
         
         return res;
      }

      /**
       * Resolve all the external programs which match the given file name.
       * 
       * @param    filename
       *           The filename.
       *           
       * @return   See above.
       */
      public List<String> resolveExternalPrograms(String filename)
      {
         filename = prepareFilename(filename);
         
         // append a "/" prefix so that the name will match in any folder
         String search = "/" + filename;
         
         // we can't search in the propath and maybe resolve an unique name... the runtime propath
         // is unknown at this time, thus resolve all matches
         
         List<String> res = new ArrayList<>();
         // instead of looking for a suffix, use the "reverse-filename" property to look for a 
         // prefix, which is indexed properly by then graph DB.
         search = reverseString(search);
         for (Vertex v : db.query().has("reverse-filename", Text.PREFIX, search).vertices())
         {
            res.add((String) v.getProperty("filename"));
         }
         
         if (res.size() == 0)
         {
            // if not found, add it, as it will be marked as missing
            res.add(filename);
         }
         
         return res;
      }
      
      /**
       * Print output to the terminal if debug settings allow.
       *
       * @param    text
       *           Text to print.
       */
      public void println(String text)
      {
         if (!silent)
         {
            System.out.println(text);
         }
      }

      /**
       * Loads the hints file for an AST if it exists and hasn't already been loaded.
       *
       * @return   <code>true</code> if there are hints for this file and the hints are loaded, 
       *           <code>false</code> if no hints are available.
       */
      public boolean loadHints()
      {
         if (hints == null)
         {
            try
            {
               Aast src = resolver.getSourceAst();
               File file = Configuration.toFile(src.getFilename());

               hints = new UastHints(file.getAbsolutePath());
            }

            catch (Exception exc)
            {
               // any failure drops us out
               return false;
            }
         }

         return true;
      }


      /**
       * Loads the preprocessor hints file for an AST if it exists and hasn't already been loaded.
       *
       * @return   The loaded {@link PreprocessorHints} instance.
       *           
       * @throws   PreprocessorException
       *           An unknown or unexpected failure.
       */
      private PreprocessorHints loadPreprocessorHints(Aast ast)
      throws PreprocessorException
      {
         if (pphints == null)
         {
            pphints = new PreprocessorHints(ast.getFilename());
         }
         
         return pphints;
      }
      
      /**
       * Reverse the given string.
       * 
       * @param    s
       *           The string to reverse.
       *           
       * @return   The reversed string.
       */
      private String reverseString(String s)
      {
         char[] chars = s.toCharArray();
         for (int i = 0; i < chars.length / 2; i++)
         {
            // switch chars in-place
            int idx2 = chars.length - (i + 1);
            char aux = chars[i];
            chars[i] = chars[idx2];
            chars[idx2] = aux;
         }
         
         return new String(chars);
      }

      /**
       * Create an include node in the graph DB, associated the given physical file.
       * 
       * @param    filename
       *           The name of this include file.
       *           
       * @return   The created node.
       */
      private Vertex createIncludeNode(String filename)
      {
         long id = nextExternalId(ProgressParserTokenTypes.INCLUDE);
         Vertex node = createNode(id, ProgressParserTokenTypes.INCLUDE);
         // un-prepared filename here
         node.setProperty("os-filename", filename);
         filename = prepareFilename(filename);
         node.setProperty("filename", filename);
         node.setProperty("reverse-filename", reverseString(filename));
         node.setProperty("node-key", "filename");
         
         return node;
      }
      
      /**
       * Create the include graph by walking the include tree specified by the root, recursively.
       * 
       * @param    root
       *           The root include node.
       * @param    parent
       *           The parent node in the graph DB, to which the root's children will be attached.
       */
      private void createIncludeGraph(HintsTreeNode root, Vertex parent)
      {
         @SuppressWarnings("unchecked")
         Iterator<HintsTreeNode> iter = (Iterator<HintsTreeNode>) root.childIterator();
         while (iter.hasNext())
         {
            HintsTreeNode node = iter.next();
            IncludeHint hint = node.getIncludeHint();
            
            String hintfile = hint.getFilename();

            Vertex vnode = findUniqueNode("os-filename", hintfile);
            
            if (vnode == null)
            {
               final String msg = "WARNING: creating a new node for the include file %s, which " +
                                  "was not found in initial include-file list.";
               System.out.println(String.format(msg, hint.getFilename()));

               // create it
               vnode = createIncludeNode(hintfile);
            }
            
            Edge edge = createEdge(parent, vnode, "includes");
      
            // copy the info: this needs to be set at the edge, as the same include file may
            // appear multiple times
            edge.setProperty("start-column", hint.getStartColumn());
            edge.setProperty("end-column",   hint.getEndColumn());
            edge.setProperty("start-line",   hint.getStartLine());
            edge.setProperty("end-line",     hint.getEndLine());
            
            edge.setProperty("column", hint.getStartColumn());
            edge.setProperty("line", hint.getStartLine());
            edge.setProperty("call-site-id", parent.getProperty("node-id"));
            edge.setProperty("call-site-key", ProgressParserTokenTypes.INCLUDES);

            createIncludeGraph(node, vnode);
         }
      }
      
      /**
       * Compute the next available ID for an external node. External nodes are targets which
       * are not part of the code set ASTs, like native processes, OCX controls, web service 
       * connections, etc.
       * 
       * @param    type
       *           The external node type.
       *           
       * @return   The next available ID.
       */
      private long nextExternalId(int type)
      {
         // all external IDs are negative
         TitanGraphQuery query = (TitanGraphQuery) db.query().has("node-type", type)
                                                     .has("node-id");
         Iterator<Vertex> iter = query.orderBy("node-id", Order.DESC)
                                      .limit(1).vertices().iterator();
         if (iter.hasNext())
         {
            Vertex v = (Vertex) iter.next();
            long id = (long) v.getProperty("node-id");
            
            return id + 1;
         }
         
         return 0;
      }
   }
}