/*
** Module   : ClassDefinition.java
** Abstract : defines the 4GL API for a class or interface definition
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20070911   @35134 First version, with support for resolving member variables,
**                           properties and methods. Honors the public/protected/private
**                           accessor. Manages the inheritance and interface hierarchy.
** 002 GES 20090518   @42405 Import change.
** 003 GES 20110822          Added support for static members. Added flag to differentiate 4GL
**                           classes from .NET classes.
** 004 GES 20140319          Store static and instance methods/vars in different maps since
**                           overloads can occur. Store the least restrictive method/var (from
**                           an access mode perspective) since overloads can occur.
** 005 GES 20160312          Added support for buffers, temp-tables, queries, datasets and
**                           data-sources as members of a class. Protected members can be
**                           accessed from child classes so the lookup processing was enhanced.
**                           Rewrite for common code instead of separate workers. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.schema.*;

/**
 * Contains a dictionary for each type of class resource and provides
 * methods to maintain and lookup in the respective namespaces.
 * <p>
 * Provides support for the following namespaces:
 * <ul>
 *    <li> member variables
 *    <li> properties
 *    <li> methods
 *    <li> temp-tables
 *    <li> buffers
 *    <li> queries
 *    <li> datasets
 *    <li> data-sources
 * </ul>
 */
public class ClassDefinition
implements ProgressParserTokenTypes
{
   /** Data store identifier. */
   private static enum DataStoreType { VAR, METHOD, TABLE, QRY, DATASET, DATASRC };
   
   /** Instance data stores. */
   private Map<DataStoreType, Map<String, MemberData>> istores = null; 
   
   /** Static data stores. */
   private Map<DataStoreType, Map<String, MemberData>> sstores = null;
   
   /** Internal lock object. */
   private Object lock = new Object();
   
   /** Provides schema lookup for this class' scope. */
   private SchemaDictionary schemaDict = null;
   
   /** Fully qualified name of the class or interface. */
   private final String name;
   
   /** Super class definition. */
   private final ClassDefinition parent;
   
   /** List of implemented interfaces. */
   private final Set<String> interfaces;
   
   /** Flag marking if this is a built-in Progress class. */
   private final boolean builtin;
   
   /** Flag marking if this is a class or interface. */
   private final boolean iface;
   
   /** Flag to differentiate between 4GL classes and .NET classes. */
   private final boolean dotnet;
   
   /**
    * Construct an instance and load any superclass or interface hierarchy
    * that is needed.  For each superclass or interface, a separate class
    * definition will be loaded. These definitions will be linked back by
    * reference from this instance. This allows lookups to be done based on
    * the object hierarchy.
    *
    * @param    name
    *           Fully qualified name of the class or interface.
    * @param    parent
    *           Fully qualified name of the super class or <code>null</code>
    *           if there is no explicit super class.
    * @param    interfaces
    *           Collection of fully qualified interface names that are
    *           implemented by this class.
    * @param    builtin
    *           <code>true</code> if this definition is built-in to Progress.
    * @param    iface
    *           <code>true</code> if this is an interface definition,
    *           <code>false</code> if this is a class definition.
    * @param    dotnet
    *           <code>true</code> if this is a .NET definition,
    *           <code>false</code> if this is a 4GL definition.
    */
   public ClassDefinition(String          name,
                          ClassDefinition parent,
                          Set<String>     interfaces,
                          boolean         builtin,
                          boolean         iface,
                          boolean         dotnet)
   {
      this.name       = name;
      this.parent     = parent;
      this.interfaces = interfaces;
      this.builtin    = builtin;
      this.iface      = iface;
      this.dotnet     = dotnet;
   }
   
   /**
    * Obtains the name of this class.
    *
    * @return   Fully qualified class name.  
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Obtains the parent class definition for this class.
    *
    * @return   Parent class definition or <code>null</code> if there is no
    *           explicit parent.
    */
   public ClassDefinition getParent()
   {
      return parent;
   }
   
   /**
    * Obtains the list of interfaces implemented by this class.
    *
    * @return   Implemented interfaces.
    */
   public Set<String> getInterfaces()
   {
      return new HashSet<String>(interfaces);
   }
   
   /**
    * Access the flag that denotes if this is a built-in class definition.
    *
    * @return   <code>true</code> if this is a built-in class.
    */
   public boolean isBuiltIn()
   {
      return builtin;
   }
   
   /**
    * Access the flag that denotes if this is an interface or a class
    * definition.
    *
    * @return   <code>true</code> if this is an interface definition.
    */
   public boolean isInterface()
   {
      return iface;
   }
   
   /**
    * Access the flag that denotes if this is a .NET class definition.
    *
    * @return   <code>true</code> if this is a .NET class.
    */
   public boolean isDotNet()
   {
      return dotnet;
   }
   
   /**
    * Add the named method to this class.
    *
    * @param    name
    *           Method name.
    * @param    type
    *           Return type for the method.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the method is static.
    * @param    node
    *           The AST to be annotated or <code>null</code> if no node needs
    *           annotations.
    * @param    qname
    *           Fully qualified class name where the resource represents an
    *           object instance, otherwise <code>null</code>.
    */
   public void addMethod(String  name,
                         int     type,
                         int     access,
                         boolean isStatic,
                         Aast    node,
                         String  qname)
   {
      addWorker(DataStoreType.METHOD, name, type, access, isStatic, qname);
      
      // annotate the given node
      if (node != null)
      {
         node.putAnnotation("name", name);
         node.putAnnotation("return_type", new Long(type));
         node.putAnnotation("access", new Long(access));
         node.putAnnotation("static", new Boolean(isStatic));
         
         if (qname != null)
            node.putAnnotation("qualified", qname);
      }
   }
      
   /**
    * Obtain the return token type of the named method given the name and
    * access mode. If the method does not exist in this class definition
    * the parent (if one exists) will be checked.
    *
    * @param    name
    *           Method name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   Return token type or -1 if no such method exists. 
    */
   public int lookupMethod(String name, int access, boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.METHOD, name, access, isStatic);
      return (mdat == null) ? -1 : mdat.type;
   }
   
   /**
    * Annotate the given method invocation reference with details for the specified method
    * in this class.  This is a no-operation if the method doesn't exist or if the node is
    * <code>null</code>.
    *
    * @param    mname
    *           Method name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the method is static.
    * @param    node
    *           The AST to be annotated.
    */
   public void annotateMethodCall(String mname, int access, boolean isStatic, Aast node)
   {
      if (node != null)
      {
         MemberData mdat = lookupHierarchy(DataStoreType.METHOD, mname, access, isStatic);
         
         if (mdat != null)
         {
            // the access mode won't necessarily be the same as that passed in
            node.putAnnotation("access", new Long(mdat.access));
            node.putAnnotation("static", new Boolean(mdat.isStatic));
            node.putAnnotation("found-in-cls", this.name);
         }
      }
   }
      
   /**
    * Obtain the fully qualified class name of the object instance 
    * represented by the method given the name and access mode.
    * <p>
    * If the resource does not exist in this class definition the parent
    * (if one exists) will be checked. This means a recursive call can
    * occur here.
    *
    * @param    name
    *           Variable or property name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic    
    *           <code>true</code> if this is a static reference.
    *
    * @return   Fully qualified class name of the the object instance
    *           represented by the named method, or <code>null</code> if no
    *           such method exists. 
    */
   public String lookupMethodClassName(String name, int access, boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.METHOD, name, access, isStatic);
      return (mdat == null) ? null : mdat.qname;
   }
   
   /**
    * Add a data member (variable) or property to this class.
    *
    * @param    name
    *           Variable or property name.
    * @param    type
    *           Token type.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the variable/property is static.
    * @param    qname
    *           Fully qualified class name where the resource represents an
    *           object instance, otherwise <code>null</code>.
    */
   public void addVariable(String  name,
                           int     type,
                           int     access,
                           boolean isStatic,
                           String  qname)
   {
      addWorker(DataStoreType.VAR, name, type, access, isStatic, qname);
   }
   
   /**
    * Obtain the token type of the data member (variable) or property given
    * the name and access mode.
    * <p>
    * If the resource does not exist in this class definition the parent
    * (if one exists) will be checked. This means a recursive call can
    * occur here.
    *
    * @param    name
    *           Variable or property name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   Return token type or -1 if no such variable exists. 
    */
   public int lookupVariable(String name, int access, boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.VAR, name, access, isStatic);
      return (mdat == null) ? -1 : mdat.type;
   }
   
   /**
    * Obtain the fully qualified class name of the object instance 
    * represented by the data member (variable) or property given
    * the name and access mode.
    * <p>
    * If the resource does not exist in this class definition the parent
    * (if one exists) will be checked. This means a recursive call can
    * occur here.
    *
    * @param    name
    *           Variable or property name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic    
    *           <code>true</code> if this is a static reference.
    *
    * @return   Fully qualified class name of the the object instance
    *           represented by this variable or data member, or 
    *           <code>null</code> if no such variable exists. 
    */
   public String lookupVariableClassName(String  name,
                                         int     access,
                                         boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.VAR, name, access, isStatic);
      return (mdat == null) ? null : mdat.qname;
   }
   
   /**
    * Cleanup any object resources which should not remain in memory.  The only resources
    * that will be used after this point are ones which can be looked up by other classes.
    */
   public void cleanupObjectResources()
   {
      synchronized (lock)
      {
         if (schemaDict != null)
         {
            ArrayList<String> saveList = new ArrayList<String>();
            
            defineExceptions(saveList, false);
            defineExceptions(saveList, true);
            
            String[] save = saveList.toArray(new String[0]);
            
            schemaDict.removeAllExcept(save);
         }
      }
   }
   
   /**
    * Obtain the backing schema dictionary, if it exists.
    *
    * @return   The schema dictionary or <code>null</code> if it was never set.
    */
   public SchemaDictionary getSchemaDict()
   {
      synchronized (lock)
      {
         return schemaDict;
      }
   }
   
   /**
    * Add the named temp-table or buffer to this class.
    *
    * @param    name
    *           Temp-table or buffer name.
    * @param    type
    *           The integer token type representing the type of table. 
    * @param    access
    *           Access mode (<code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the method is static.
    * @param    schemaDict
    *           The schema dictionary that manages the temp-tables and buffers for this
    *           class.
    */
   public void addTable(String           name,
                        int              type,
                        int              access,
                        boolean          isStatic,
                        SchemaDictionary schemaDict)
   {
      synchronized (lock)
      {
         if (this.schemaDict == null)
         {
            this.schemaDict = schemaDict;
         }
      }
      
      addWorker(DataStoreType.TABLE, name, type, access, isStatic, null);
   }
      
   /**
    * Obtain the token type of the temp-table or buffer given the name and access mode.
    * <p>
    * If the resource does not exist in this class definition the parent
    * (if one exists) will be checked. This means a recursive call can
    * occur here.
    *
    * @param    name
    *           Temp-table or buffer name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   Return token type or -1 if no such variable exists. 
    */
   public int lookupTable(String name, int access, boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.TABLE, name, access, isStatic);
      return (mdat == null) ? -1 : mdat.type;
   }
   
   /**
    * Add the named query to this class.
    *
    * @param    name
    *           Query name.
    * @param    type
    *           The integer token type representing the type of table. 
    * @param    access
    *           Access mode (<code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the method is static.
    */
   public void addQuery(String name, int type, int access, boolean isStatic)
   {
      addWorker(DataStoreType.QRY, name, type, access, isStatic, null);
   }
      
   /**
    * Obtain the token type of the query member given the name and access mode.
    * <p>
    * If the resource does not exist in this class definition the parent
    * (if one exists) will be checked. This means a recursive call can
    * occur here.
    *
    * @param    name
    *           Query name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   Return token type or -1 if no such variable exists. 
    */
   public int lookupQuery(String name, int access, boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.QRY, name, access, isStatic);
      return (mdat == null) ? -1 : mdat.type;
   }
   
   /**
    * Add the named dataset to this class.
    *
    * @param    name
    *           Dataset name.
    * @param    type
    *           The integer token type representing the type of table. 
    * @param    access
    *           Access mode (<code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the method is static.
    */
   public void addDataSet(String name, int type, int access, boolean isStatic)
   {
      addWorker(DataStoreType.DATASET, name, type, access, isStatic, null);
   }
      
   /**
    * Obtain the token type of the dataset member given the name and access mode.
    * <p>
    * If the resource does not exist in this class definition the parent
    * (if one exists) will be checked. This means a recursive call can
    * occur here.
    *
    * @param    name
    *           Dataset name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   Return token type or -1 if no such dataset exists. 
    */
   public int lookupDataSet(String name, int access, boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.DATASET, name, access, isStatic);
      return (mdat == null) ? -1 : mdat.type;
   }
   
   /**
    * Add the named data source to this class.
    *
    * @param    name
    *           Data source name.
    * @param    type
    *           The integer token type representing the type of table. 
    * @param    access
    *           Access mode (<code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the method is static.
    */
   public void addDataSource(String name, int type, int access, boolean isStatic)
   {
      addWorker(DataStoreType.DATASRC, name, type, access, isStatic, null);
   }
      
   /**
    * Obtain the token type of the data source member given the name and access mode.
    * <p>
    * If the resource does not exist in this class definition the parent
    * (if one exists) will be checked. This means a recursive call can
    * occur here.
    *
    * @param    name
    *           Data source name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   Return token type or -1 if no such data source exists. 
    */
   public int lookupDataSource(String name, int access, boolean isStatic)
   {
      MemberData mdat = lookupHierarchy(DataStoreType.DATASRC, name, access, isStatic);
      return (mdat == null) ? -1 : mdat.type;
   }
   
   /**
    * Obtain the static or instance member map or create (and return) it if it doesn't
    * already exist.
    *
    * @param    type
    *           The data store type to lookup.
    * @param    isStatic
    *           <code>true</code> if the static map is to be returned, <code>false</code> for
    *           the instance version.
    *
    * @return   The specified map. It will not be <code>null</code>.
    */
   private Map<String, MemberData> getMemberStore(DataStoreType type, boolean isStatic)
   {
      Map<String, MemberData> data = null;
      
      synchronized (lock)
      {
         if (isStatic && sstores == null)
         {
            sstores = new HashMap<DataStoreType, Map<String, MemberData>>();
         }
         if (!isStatic && istores == null)
         {
            istores = new HashMap<DataStoreType, Map<String, MemberData>>();
         }
         
         data = isStatic ? sstores.get(type) : istores.get(type);
         
         if (data == null)
         {
            data = new HashMap<String, MemberData>();
            
            if (isStatic)
            {
               sstores.put(type, data);
            }
            else
            {
               istores.put(type, data);
            }
         }
      }
      
      return data;
   }
   
   /**
    * Obtain the member given the name and access mode. If the member does not exist in
    * this class definition the parent (if one exists) will be checked.
    *
    * @param    type
    *           Type of data store to check.
    * @param    name
    *           Variable/data member name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   Member data or <code>null</code> if no such variable or data
    *           member exists. 
    */
   private MemberData lookupHierarchy(DataStoreType type,
                                      String        name,
                                      int           access,
                                      boolean       isStatic)
   {
      MemberData mdat = null;
      
      synchronized (lock)
      {
         mdat = lookupWorker(getMemberStore(type, isStatic), name, access);
         
         if (mdat == null && parent != null)
         {
            // change access mode since we aren't allowed to see private stuff
            // in parent
            if (access == KW_PRIVATE)
               access = KW_PROTECTD;
            
            // recurse up inheritence hierarchy
            mdat = parent.lookupHierarchy(type, name, access, isStatic);
         }
         
         // if only static members will match
         if (mdat != null && isStatic && !mdat.isStatic)
         {
            // clear our result
            mdat = null;
         }
      }
      
      return mdat;
   }
   
   /**
    * Add the resource to the given maps, if it is not already present OR if it is present
    * under a more restrictive access type (we must store the least restrictive of the types
    * to allow operator overloading to work properly).
    *
    * @param    stype
    *           Type of data store to check.
    * @param    name
    *           Resource name.
    * @param    type
    *           Token type.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the variable/property is static.
    * @param    qname
    *           Fully qualified class name where the resource represents an
    *           object instance, otherwise <code>null</code>.
    */
   private void addWorker(DataStoreType stype,
                          String        name,
                          int           type,
                          int           access,
                          boolean       isStatic,
                          String        qname)
   {
      synchronized (lock)
      {
         Map<String, MemberData> map   = getMemberStore(stype, isStatic);
         String                  lower = name.toLowerCase();
         MemberData              odat  = map.get(lower);
         
         boolean insert = true;
         
         if (odat != null)
         {
            // a record already exists, only replace it if we are making it more accessible
            if ((odat.access == KW_PRIVATE && (access == KW_PUBLIC || access == KW_PROTECTD)) ||
                (odat.access == KW_PROTECTD && access == KW_PUBLIC))
            {
               // leave insert as true
            }
            else
            {
               insert = false;
            }
         }
         
         if (insert)
         {
            MemberData mdat = new MemberData(name, type, access, isStatic, qname);
         
            // add/replace in the full method list
            map.put(lower, mdat);
         }
      }
   }
   
   /**
    * Find the named resource in the given map.
    *
    * @param    map
    *           Dictionary for all resources regardless of access mode.
    * @param    name
    *           Resource name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    *
    * @return   Resource found or <code>null</code> if no match exists.
    */
   private MemberData lookupWorker(Map<String, MemberData> map, String name, int access)
   {
      MemberData mdat = null;
      
      synchronized (lock)
      {
         if (map != null)
         {
            mdat = map.get(name.toLowerCase());
            
            if (mdat != null)
            {
               // check for a match to the access mode:
               // 1. searching for private methods match everything
               // 2. a direct match is always OK
               // 3. a public method matches a protected search 
               boolean ok = (access == KW_PRIVATE)  ||
                            (access == mdat.access) ||
                            (access == KW_PROTECTD && mdat.access == KW_PUBLIC);
               
               // pretend no match if access modes are not right
               if (!ok)
               {
                  mdat = null;
               }
            }
         }
      }
      
      return mdat;
   }
   
   /**
    * Pull out our lists of temp-tables and buffers so they can be used for
    * cleanup.
    *
    * @param    save
    *           Set of temp-tables and buffers.
    * @param    isStatic
    *           <code>true</code> to process static resources, otherwise instance resources.
    */
   private void defineExceptions(ArrayList<String> save, boolean isStatic)
   {
      Map<String, MemberData> map = getMemberStore(DataStoreType.TABLE, isStatic);
      
      if (map != null && !map.isEmpty())
      {
         Iterator<String> iter = map.keySet().iterator();
         
         while (iter.hasNext())
         {
            MemberData mdat = map.get(iter.next());
            
            if (mdat != null)
            {
               if ((mdat.type == BUFFER || mdat.type == TEMP_TABLE) && mdat.access == KW_PROTECTD)
               {
                  save.add(mdat.name);
               }
            }
         }
      }
   }
   
   /**
    * Stores a member's data.
    */
   private static class MemberData
   {
      /** Text name. */
      private String name = null;
      
      /** Token type. */
      private int type = -1;
      
      /** Access mode. */
      private int access = -1;
      
      /** Static or instance member. */
      private boolean isStatic = false;
      
      /** Class type name for members that return an object instance. */
      private String qname = null;
      
      /**
       * Create an instance, setting the data members.
       *
       * @param    name
       *           The name of the member.
       * @param    type
       *           The token type associated with the member.
       * @param    access
       *           Access mode (<code>KW_PUBLIC</code>,
       *           <code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>).
       * @param    isStatic
       *           <code>true</code> if the variable/property is static.
       */
      public MemberData(String name, int type, int access, boolean isStatic)
      {
         this.name     = name;
         this.type     = type;
         this.access   = access;
         this.isStatic = isStatic;
      }
      
      /**
       * Create an instance, setting the data members.
       *
       * @param    name
       *           The name of the member.
       * @param    type
       *           The token type associated with the member.
       * @param    access
       *           Access mode (<code>KW_PUBLIC</code>,
       *           <code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>).
       * @param    isStatic
       *           <code>true</code> if the variable/property is static.
       * @param    qname
       *           Fully qualified class name where <code>type</code> equals
       *           <code>METH_CLASS</code>. <code>null</code> if this method
       *           does not return an object instance.
       */
      public MemberData(String  name,
                        int     type,
                        int     access,
                        boolean isStatic,
                        String  qname)
      {
         this(name, type, access, isStatic);
         this.qname = qname;
      }
   }
}