/*
** Module   : ProgressAst.java
** Abstract : A Progress 4GL-specific AST.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20050323   @20486 Created initial version.
** 002 GES 20070720   @34649 Added getDescriptiveTokenText().
** 003 GES 20090518   @42409 Import change.
** 004 GES 20110712          Added built-in functions KEYWORD and KEYWORD-ALL,
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import antlr.*;
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.util.*;

/**
 * Provides symbolic token name lookups that are specific to the Progress
 * AST, all other function is implemented in the super class.
 * <p>
 * This also provides static implementations of the 4GL built-in functions
 * KEYWORD and KEYWORD-ALL,
 */
public class ProgressAst
extends AnnotatedAst
{
   /** Provide a reverse lookup of token types to keyword text. */
   private static HashMap reverse = new HashMap();
   
   /** Progress 4GL keyword forward lookup services. */
   private static KeywordDictionary kwDict = new KeywordDictionary();
   
   /**
    * Default constructor.
    */
   public ProgressAst()
   {
      super();
   }
   
   /**
    * Constructor which accepts and wrappers token.
    *
    * @param   token
    *          Token upon which this node is based.
    */
   public ProgressAst(Token token)
   {
      super(token);
   }
   
   /**
    * Duplicates the Progress 4GL KEYWORD built-in function, which allows
    * code to lookup if given text is a case-insensitive match (abbreviated or
    * not) for any valid 4GL keyword.
    *
    * @param    text
    *           The text to lookup.
    *
    * @return   The full text of the matching keyword if it is found or
    *           <code>unknown</code> if there is no match.
    */
   public static character keyword(character text)
   {
      return keyword(text.toStringMessage());
   }
   
   /**
    * Duplicates the Progress 4GL KEYWORD built-in function, which allows
    * code to lookup if given text is a case-insensitive match (abbreviated or
    * not) for any valid 4GL keyword.
    *
    * @param    text
    *           The text to lookup.
    *
    * @return   The full text of the matching keyword if it is found or
    *           <code>unknown</code> if there is no match.
    */
   public static character keyword(String text)
   {
      Keyword match = lookupWorker(text);
      return match == null ? new character()
                           : new character(match.getFullText());
   }
   
   /**
    * Duplicates the Progress 4GL KEYWORD-ALL built-in function, which allows
    * code to lookup if given text is a case-insensitive match (abbreviated or
    * not) for any valid 4GL RESERVED keyword.
    *
    * @param    text
    *           The text to lookup.
    *
    * @return   The full text of the matching RESERVED keyword if it is found
    *           or <code>unknown</code> if there is no match.
    */
   public static character keywordAll(character text)
   {
      return keywordAll(text.toStringMessage());
   }
   
   /**
    * Duplicates the Progress 4GL KEYWORD-ALL built-in function, which allows
    * code to lookup if given text is a case-insensitive match (abbreviated or
    * not) for any valid 4GL RESERVED keyword.
    *
    * @param    text
    *           The text to lookup.
    *
    * @return   The full text of the matching RESERVED keyword if it is found
    *           or <code>unknown</code> if there is no match.
    */
   public static character keywordAll(String text)
   {
      Keyword match = lookupWorker(text);
      boolean found = (match != null && match.isReserved());
      
      return found ? new character(match.getFullText())
                   : new character();
   }
   
   /**
    * Returns the "canonical" keyword text that is associated with a given
    * token type if that type represents a known Progress 4GL language
    * keyword.
    *
    * @param    type
    *           The token type for which to reverse lookup the text.
    *
    * @return   The keyword text or <code>null</code> if this type is not
    *           associated with a language keyword.
    */
   public String keywordText(int type)
   {
      // this method would/should be static but the pattern engine rules
      // can't call it in that case... when they are fixed this should be
      // switched
      
      synchronized (reverse)
      {
         if (reverse.isEmpty())
         {
            Keyword[] keywords = ProgressLexer.getKeywords();
            
            for (int i = 0; i < keywords.length; i++)
            {
               String ttxt  = keywords[i].getFullText();
               int    ttype = keywords[i].getTokenType();
               
               reverse.put(new Integer(ttype), ttxt);
            }
         }
      }
      
      return (String) reverse.get(new Integer(type));
   }
   
   /**
    * Translates a text representation of a token type into the
    * actual integer value.
    *
    * @param    tokenName
    *           The text representation of the token type.
    *
    * @return   A valid integer token type value or -1 if no match was found.
    */
   public int lookupTokenType(String tokenName)
   {   
      return ProgressParser.lookupTokenType(tokenName);
   }
   
   /**
    * Translates an integer value of a token type into the associated 
    * human readable text representation.
    *
    * @param    type
    *           The integer token type.
    *
    * @return   A valid symbolic token type name or <code>null</code> if
    *           no match was found.
    */
   public String lookupTokenName(int type)  
   {   
      return ProgressParser.lookupTokenName(type);
   }
   
   /**
    * Returns the symbolic representation of this AST's token type.
    *
    * @return  Symbolic token name for this AST's token type.
    */
   public String getSymbolicTokenType()
   {
      return ProgressParser.lookupTokenName(getType());
   }
   
   /**
    * Returns a descriptive string for the current node.  This string will
    * normally contain both the text and the symbolic representation of this
    * AST's token type.  If the token type is not associated with a Progress
    * 4GL language keyword then this will result in the same text as the
    * {@link #getSymbolicTokenType}.  Otherwise this will result in the
    * keyword text followed by the symbolic token type in quare brackets.
    *
    * @return   A descriptive string for the token type of this node.
    */
   public String getDescriptiveTokenText()
   {
      String ktxt = keywordText(getType());
      String stxt = getSymbolicTokenType();
      
      return ktxt == null ? stxt : String.format("%s [%s]", ktxt, stxt);
   }
   
   /**
    * Lookup a keyword given partial (abbreviated) or full keyword text.
    *
    * @param    text
    *           The keyword text to lookup.
    *
    * @return   The keyword object found or <code>null</code> if no match is
    *           found.
    */
   private static Keyword lookupWorker(String text)
   {
      Keyword match = null;
      
      synchronized (kwDict)
      {
         if (kwDict.size() == 0)
         {
            Keyword[] keywords = ProgressLexer.getKeywords();
            
            for (int i = 0; i < keywords.length; i++)
            {
               kwDict.addKeyword(keywords[i]);
            }
         }
         
         match = kwDict.lookupKeyword(text);
      }
      
      return match;
   }
}