/*
** Module   : BrowseColumnWidget.java
** Abstract : server side browse column widget implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 NVS 20051026   @23480 Created initial version.
** 002 SIY 20060227   @25122 Changes for the browse controller 
**                           implementation.
** 003 SIY 20060330   @25278 Added methods for LABEL attribute.
** 004 SIY 20060417   @26218 Added setColumnLabel() method.
** 005 GES 20060803   @28365 Major cleanup and refactoring.
** 006 GES 20080320   @37580 Added TYPE attribute.
** 007 SIY 20080603   @38474 setColumn() now operates on separate
**                           COLUMN-LABEL attribute. 
** 008 SIY 20090206   @41262 Changed setLabel() logic to make it properly
**                           handling setLabel/setColumnLabel/setLavel 
**                           sequences (this can happen when label is
**                           dynamically set after browse construction).
** 009 SIY 20090218   @41328 Added support for AT base field format width.
** 010 SIY 20111101          Updated imports.
** 011 CA  20130221          Refactored usage of PRIVATE-DATA and TYPE attribute.
** 012 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 013 CA  20130927          Added LegacyResource annotation. TODO: implement the resource type 
**                           based actual column type and "-browcoltype" parameter.
** 014 CA  20131115          Changes related to properly instantiate dynamic resources.
** 015 CA  20140711          Enhanced font support, for GUI clients.
** 016 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 017 CA  20141104          Renamed setWidth to setWidthChars.
** 018 HC  20141021          Implemented window move/position support.
** 019 SVL 20141102          Added some attributes.
** 020 SVL 20141209          Implemented getNextColumn.
** 021 GES 20141124          Added inner-lines support. Added several combo-box attrs/meths. 
** 022 IAS 20150123          delete(int) replaced with delete(long).
** 023 IAS 20150211          RemoveMove methods specific for list widgets
** 024 HC  20150323          Removed business logic from config classes.
** 025 SVL 20150503          Changed getNextColumn to work properly with dynamic columns.
** 026 SVL 20150605          Added editingFillIn.
** 027 IAS 20150606          Added support for the DISABLE-AUTO-ZAP
** 028 SVL 20150722          Implemented MODIFIED attribute.
** 029 SVL 20151013          Added colors and fonts accessors.
** 030 OM  20150923          Switch to long/int64 datatypes.
** 031 SVL 20151019          Moved ChUI-specific column width calculation logic to BrowseColumn.
** 032 SVL 20151126          NO-LABELS attribute is not inherited form the parent browse.
** 033 SVL 20151202          Implemented getWidthChars.
** 034 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 035 IAS 20160229          LOAD-MOUSE-POINTER support.
** 036 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 037 HC  20160714          Fixed runtime support of BROWSE-COLUMN:NAME.
** 038 HC  20160716          Implemented runtime support for BROWSE-COLUMN:RESIZABLE attribute.
** 039 SVL 20160811          Support for reassigned browsed queries with another backing buffers.
** 040 SBI 20160909          Changed to override getFormat() to return the default data type
**                           format if the column data format is not specified explicitly, added
**                           new constructor to create a dynamic widget.
** 041 SVL 20170120          Added setFgColor(NumberType), getters/setters for SORT-NUMBER,
**                           SORT-ASCENDING and WIDTH-PIXELS attributes
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * Server side representation of a browse column.
 * <p>
 * The {@link LegacyResource} annotation targets the {@link LegacyResource#FILL_IN} type.
 * <p>
 * The 4GL documentation states that {@link LegacyResource#BROWSE_COLUMN} type is used only when 
 * "-browcoltype" parameter is * at program startup.  Else, the type can be either 
 * {@link LegacyResource#FILL_IN}, {@link LegacyResource#COMBO_BOX} or 
 * {@link LegacyResource#TOGGLE_BOX} (depending on the column's widget).
 * 
 * TODO: fix this to report the correct type. 
 * TODO: check if the browse columns can be dynamic; if so, change the c'tor to receive a boolean
 *       flag.
 */
@LegacyResource(resource = LegacyResource.BROWSE_COLUMN)
public class BrowseColumnWidget
extends GenericWidget<BrowseColumnConfig>
implements AutoZapElement,
           BrowseColumnInterface,
           BrowseElement,
           MaxCharsElement,
           NumItems,
           SortedElement,
           InnerLines,
           Hoverable
{
   /** Back reference to the variable accessor. */
   private FrameElement element = null;

   /**
    * Fill-in widget which is used for in-browse editing for this column (if the column is
    * editable).
    */
   private FillInWidget editingFillIn = null;

   /**
    * If the query uses a field reference as the data source, this is DMO interface of the
    * related backing buffer in the browsed query. Otherwise it is <code>null</code>.
    */
   private Class<?> linkedDMOInterface = null;

   /**
    * If the query uses a field reference as the data source, this is the 0-based index of the
    * related backing buffer in the browsed query. Otherwise it is <code>-1</code>.
    */
   private int bufferIndexInQuery = -1;

   /**
    * Default constructor creates a static browse column
    */
   public BrowseColumnWidget()
   {
      this(false);
   }

   /**
    * Parameterized constructor creates a dynamic browse column.
    * 
    * @param    dynamic
    *           The true value indicates a dynamic widget will be created, otherwise a static one.
    */
   public BrowseColumnWidget(boolean dynamic)
   {
      super(dynamic, new BrowseColumnConfig());
   }

   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   public logical isReadOnly() 
   {
      return new logical(config.readOnly);
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   public void setReadOnly(logical r) 
   {
      if (config.readOnly == r.booleanValue())
      {
         return;
      }
      
      config.readOnly = r.booleanValue();
      pushScreenDefinition();
   }
   
   /**
    * Gets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @return   The number of lines of text.
    */
   @Override
   public integer getInnerLines()
   {
      return new integer(config.innerLines);
   }
   
   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget. For the INNER-LINES clause in the VIEW-AS phrase
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInitialInnerLines(double lines)
   {
      setInnerLines(lines);
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInnerLines(double lines)
   {
      config.innerLines = (int) lines;
      pushScreenDefinition();
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInnerLines(NumberType lines)
   {
      if (lines == null || lines.isUnknown())
      {
         UnimplementedFeature.todo("INNER-LINES attribute unknown behavior is not known.");
         return;
      }
   
      setInnerLines(lines.intValue());
   }
   
   /**
    * Queries if the SCREEN-VALUE attribute for this widget was changed.
    * 
    * @return  If <code>true</code>, the widget's screen value was changed.
    */
   public logical isModified()
   {
      if (editingFillIn == null)
      {
         return new logical(false);
      }

      return editingFillIn.isModified();
   }

   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    *
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(logical modified)
   {
      if (modified == null || modified.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute MODIFIED on FILL-IN " +
                     getFrame().getOriginalName(this),
               false);
         return;
      }

      FillInWidget fillInWidget = getEditingFillIn();
      if (fillInWidget != null)
      {
         fillInWidget.setModified(modified);
      }
   }

   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    *
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(boolean modified)
   {
      if (config.modified == modified)
      {
         return;
      }
      
      setModified(new logical(modified));
   }
   
   /**
    * Sets the AUTO-RETURN option.
    * 
    * @param    autoReturn
    *           <code>true</code> if the option is set to ON.
    */
   public void setAutoReturn(boolean autoReturn)
   {
      if (config.autoReturn == autoReturn)
      {
         return;
      }
      
      config.autoReturn = autoReturn;
      pushScreenDefinition();
   }
   
   /**
    * Set the COLUMN-LABEL attribute.
    * 
    * @param    label
    *           The new value for the attribute.
    */
   public void setColumnLabel(String label)
   {
      if (Utils.equals(config.clabel, label))
      {
         return;
      }

      config.clabel = label;
      pushScreenDefinition();
   }
   
   /**
    * Sets the label text, removing any '&amp;' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   public void setLabel(final String label)
   {
      // TODO: why is this not removing any "&" char first, as column-label ends up with with them,
      // if they exist?

      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            if (config.label != null)
            {
               config.clabel = label;
            }
         }
      });
      
      super.setLabel(label);
   }

   /**
    * Set the DATA-TYPE attribute value.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setDataType(final String value)
   {
      if (Utils.equals(config.dataType, value))
      {
         return;
      }
      
      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.dataType = value;
         }
      });
      
      pushScreenDefinition();
   }

   /**
    * Queries the DATA-TYPE attribute of this widget.
    *
    * @return A character value that represents the data type of the field
    *         associated with the widget.
    */
   @Override
   public character getDataType()
   {
      return new character(config.dataType);
   }

   /**
    * Sets new value of DATA-TYPE attribute.
    *
    * @param dataType
    *        new value of DATA-TYPE attribute
    */
   @Override
   public void setDataType(character dataType)
   {
      UnimplementedFeature.missing("DATA-TYPE attribute not implemented");
      // TODO: implement this
   }

   /**
    * Set the WIDTH attribute value.
    *
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setWidthChars(NumberType value)
   {
      if (!value.isUnknown())
      {
         setWidthChars(value.doubleValue());
      }
   }

   /**
    * Set the WIDTH attribute value.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setWidthChars(final double value)
   {
      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.autoResize = false;
            config.widthChars = Coordinate.scale(value);
         }
      });
      
      pushScreenDefinition();
   }

   /**
    * Get the WIDTH attribute value.
    *
    * @return  WIDTH attribute value (width of the column's data part), in characters.
    */
   @Override
   public decimal getWidthChars()
   {
      return new decimal(config.widthChars);
   }

   /**
    * Queries the WIDTH-PIXELS attribute.
    *
    * @return   WIDTH-PIXELS attribute value (width of the column's data part), in pixels.
    */
   @Override
   public integer getWidthPixels()
   {
      CoordinatesConversion cc = LogicalTerminal.coordinates();
      return new integer(cc.columnToPixels(config.widthChars));
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setWidthPixels(NumberType widthPixels)
   {
      if (widthPixels.isUnknown() || widthPixels.intValue() < 0)
      {
         ErrorManager.recordOrShowError(4084,
               "**Attribute WIDTH-PIXELS must be greater than zero", false, false);
         return;
      }

      CoordinatesConversion cc = LogicalTerminal.coordinates();
      setWidthChars(cc.columnFromPixels(widthPixels.intValue()));
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setWidthPixels(int widthPixels)
   {
      setWidthPixels(new integer(widthPixels));
   }

   /**
    * Set format width of AT base field.
    * 
    * @param   value
    *          New format width.
    */
   public void setFormatWidth(final int value)
   {
      config.formatWidth = value;
      
      // We do not push definition here because
      // this is not a run-time changeable value, 
      // it set only once during startup.
   }
   
   /**
    * Enable editing of the column.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setEditable(boolean value)
   {
      if (config.readOnly == !value)
      {
         return;
      }

      config.readOnly = !value;
      pushScreenDefinition();
   }

   /**
    * Set AUTO-ZAP attribute. Applies only to fill-in and combo-box browse columns.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setAutoZap(boolean value)
   {
      setAutoZap(new logical(value));
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(final boolean labels)
   {
      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.labels = labels;
            if (config.autoResize)   // trigger size recalculation.
               config.widthChars = -1;
         }
      });
      
      pushScreenDefinition();
   }

   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    nolabels
    *           New value for the attribute.
    */
   public void setNoLabels(boolean nolabels)
   {
      setLabels(!nolabels);
   }
   
   /**
    * Force presence of the label.
    * 
    * @param    force
    *           New value for the flag.
    */
   public void setForceLabel(final boolean force)
   {
      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.forceLabel = force;
            if (config.autoResize)   // trigger size recalculation.
               config.widthChars = -1;
         }
      });
         
      pushScreenDefinition();
   }
   
   /**
    * Provide link to the backing accessor.
    *
    * @param    element
    *           The accessor for this column.
    * @param    query
    *           Backing browsed query.
    *
    * @throws   IllegalArgumentException
    *           if the browsed query is <code>null</code>.
    */
   void setElement(FrameElement element, P2JQuery query)
   {
      this.element = element;

      // Check if the element uses a field reference. If it is, store information which identifies
      // this buffer: position in the query and its DMO interface. If another query is assigned,
      // it should have the same kind of buffer at the specific position.
      if (element instanceof Element)
      {
         Accessor accessor = ((Element) element).getDataAccessor();
         if (accessor instanceof FieldReference)
         {
            if (query == null)
            {
               throw new IllegalArgumentException("Browsed query should be provided.");
            }

            FieldReference ref = (FieldReference) accessor;
            RecordBuffer refBuffer = ref.getParentBuffer();

            bufferIndexInQuery = -1;
            linkedDMOInterface = null;
            int numBuffers = query.numBuffers().intValue();
            for (int i = 1; i <= numBuffers; i++)
            {
               Buffer buffer = query.bufferHandle(i).unwrapBuffer();
               RecordBuffer recordBuffer = ((BufferReference) buffer).buffer();
               if (recordBuffer.equals(refBuffer))
               {
                  bufferIndexInQuery = i - 1;
                  linkedDMOInterface = ref.getDMOInterface();
                  break;
               }
            }

            // linkedDMOInterface can be null at this point if the column references a buffer
            // outside of the browsed query.
         }
      }
   }

   /**
    * If this column uses a field reference as the data source, and a new query was assigned (this
    * query may have different backing buffers which, however, have the same DMO interfaces as in
    * the original query), update the field reference so that it points to the new buffer.
    *
    * @param  query
    *         New assigned query.
    *
    * @return <code>true</code> if the field reference was successfully updated or if this column
    *          doesn't use a field reference as the data source. <code>false</code> if the
    *          reference cannot be updated because the new query users buffer of a different
    *          DMO interface.
    */
   boolean updateBufferReferences(P2JQuery query)
   {
      if (bufferIndexInQuery < 0)
      {
         return true;
      }

      Buffer buffer = query.bufferHandle(bufferIndexInQuery + 1).unwrapBuffer();
      RecordBuffer recordBuffer = ((BufferReference) buffer).buffer();
      if (!linkedDMOInterface.isAssignableFrom(recordBuffer.getDMOInterface()))
      {
         //TODO in 4GL browse becomes unusable
         ErrorManager.recordOrShowError(0, String.format("Backing buffer #%d in assigned " +
               "query has different dmo class (was %s, while assigned query has %s)",
               bufferIndexInQuery, linkedDMOInterface, recordBuffer.getDMOInterface()), false);
         return false;
      }

      FieldReference oldRef = (FieldReference) ((Element) element).getDataAccessor();
      NumberType index = oldRef.getIndex() == -1 ? null : new integer(oldRef.getIndex());
      FieldReference newRef = new FieldReference((DataModelObject) buffer,
                                                 oldRef.getProperty(),
                                                 oldRef.isUpperCase(),
                                                 index);
      element = new Element(newRef, element.getFormat(), this);

      return true;
   }

   /**
    * Retrieve value of the associated variable.
    * 
    * @return   Current value of the variable.
    */
   public BaseDataType getValue()
   {
      return element.get();
   }

   /**
    * Assign value to associated variable.
    * 
    * @param    value
    *           New value for the variable.
    */
   public void setValue(BaseDataType value)
   {
      element.set(value);
   }

   /**
    * Set ID of the browse widget to which this column belongs.
    * 
    * @param    id
    *           Browse ID.
    */
   public void setBrowseId(final int id)
   {
      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.browseId = id;         
         }
      });
   }

   /**
    * Get the COLUMN-FONT attribute of this browse column widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getColumnFont()
   {
      int font = config.columnFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the COLUMN-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setColumnFont(int64 fontNum)
   {
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized

      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.columnFont = font;
      pushScreenDefinition();
   }

   /**
    * Get the LABEL-FONT attribute of this browse column widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getLabelFont()
   {
      int font = config.labelFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the LABEL-FONT attribute of this browse widget.
    *
    * @param    font
    *           An entry in the font-table or unknown to refer the default font.
    */
   public void setLabelFont(long font)
   {
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized

      config.labelFont = (int) font;
      pushScreenDefinition();
   }

   /**
    * Set the LABEL-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setLabelFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      setLabelFont(fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());
   }

   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The color to use.
    */
   public void setPfColor(NumberType param)
   {
      config.columnPfColor = new ColorSpec(param.intValue()).convert();
      pushScreenDefinition();
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    inputColor
    *           The color to use for when in "input" mode.
    */
   public void setPfColor(Color inputColor)
   {
      config.columnPfColor = inputColor;
      pushScreenDefinition();
   }
   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   public void setDColor(NumberType param)
   {
      config.columnDColor = new ColorSpec(param.intValue()).convert();
      pushScreenDefinition();
   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setDColor(Color displayColor)
   {
      config.columnDColor = displayColor;
      pushScreenDefinition();
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    param
    *           The new value for the COLUMN-BGCOLOR attribute.
    */
   public void setBgColor(NumberType param)
   {
      setBgColor(param.isUnknown() ? -1 : param.intValue());
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    param
    *           The new value for the COLUMN-BGCOLOR attribute.
    */
   public void setBgColor(long param)
   {
      if (config.columnBgColor == param)
      {
         return;
      }

      config.columnBgColor = (int) param;
      pushScreenDefinition();
   }

   /**
    * Sets the COLUMN-FGCOLOR writable attribute.
    *
    * @param    param
    *           The new value for the COLUMN-FGCOLOR attribute.
    */
   public void setFgColor(NumberType param)
   {
      setFgColor(param.isUnknown() ? -1 : param.longValue());
   }

   /**
    * Sets the COLUMN-FGCOLOR writable attribute.
    *
    * @param    param
    *           The new value for the COLUMN-FGCOLOR attribute.
    */
   public void setFgColor(long param)
   {
      if (config.columnFgColor == param)
      {
         return;
      }

      config.columnFgColor = (int) param;
      pushScreenDefinition();
   }

   /**
    * Get the NUM-ITEMS attribute.
    *
    * @return   The numebr of items in the list.
    */
   public integer getNumItems()
   {
      UnimplementedFeature.missing("NUM-ITEMS attribute getter");
      return new integer();
   }
   
   /**
    * Implements the NEXT-COLUMN attribute getter.<p>
    *
    * @return the widget handle of the next sibling, in physical order, of the current browse
    *         column whether or not the column is visible.
    */
   public handle getNextColumn()
   {
      BrowseWidget browseWidget =
            (BrowseWidget) LogicalTerminal.getWidgetForId(config().browseId);
      return browseWidget.getBrowseColumn(config().ordinal + 2, false);
   }

   /**
    * Gets the COLUMN-BGCOLOR writable attribute.
    *
    * @return   The current value of the COLUMN-BGCOLOR attribute.
    */
   public integer getBgColor()
   {
      return (config.columnBgColor == -1) ? new integer() : new integer(config.columnBgColor);
   }

   /**
    * Gets the COLUMN-FGCOLOR writable attribute.
    *
    * @return   The current value of the COLUMN-FGCOLOR attribute.
    */
   public integer getFgColor()
   {
      return (config.columnFgColor == -1) ? new integer() : new integer(config.columnFgColor);
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   public void setBgColor(integer bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setBgColor(color);
   }

   /**
    * Gets the LABEL-BGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-BGCOLOR attribute.
    */
   public integer getLabelBgColor()
   {
      int color = config.labelBgColor;
      return (color == -1) ? new integer() : new integer(color);
   }

   /**
    * Gets the LABEL-FGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-FGCOLOR attribute.
    */
   public integer getLabelFgColor()
   {
      int color = config.labelBgColor;
      return (color == -1) ? new integer() : new integer(color);
   }

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   public void setLabelBgColor(int64 bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setLabelBgColor(color);
   }

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   public void setLabelBgColor(long bgcolor)
   {
      if (config.labelBgColor == bgcolor)
      {
         return;
      }
      
      config.labelBgColor = (int) bgcolor;
      pushScreenDefinition();
   }

   /**
    * Sets the LABEL-FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the LABEL-FGCOLOR attribute.
    */
   public void setLabelFgColor(int64 fgcolor)
   {
      int color = (fgcolor == null || fgcolor.isUnknown()) ? -1 : fgcolor.intValue();
      setLabelFgColor(color);
   }

   /**
    * Sets the LABEL-FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the LABEL-FGCOLOR attribute.
    */
   public void setLabelFgColor(long fgcolor)
   {
      if (config.labelFgColor == fgcolor)
      {
         return;
      }
      
      config.labelFgColor = (int) fgcolor;
      pushScreenDefinition();
   }

   /**
    * Set value of the SORT attribute.
    *
    * @param    sort
    *           <code>true</code> if the combo-box browse columns has sorted items.
    */
   public void setSort(boolean sort)
   {
      setSort(new logical(sort));
   }

   /**
    * Set value of the SORT attribute.
    *
    * @param    sort
    *           <code>true</code> if the combo-box browse columns has sorted items.
    */
   public void setSort(logical sort)
   {
      if (sort.isUnknown())
         return;

      if (config.sort == sort.booleanValue())
      {
         return;
      }
      
      config.sort = sort.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the SORT attribute.
    *
    * @return   <code>true</code> if the combo-box browse columns has sorted items.
    */
   public logical isSort()
   {
      return new logical(config.sort);
   }

   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int64 start, int64 end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int64 start, long end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(long start, int64 end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(long start, long end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Obtains the index position of the cursor in the edited cell in this column.
    *
    * @return   The character index at which the cursor is placed.
    */
   public integer getCursorOffset()
   {
      if (editingFillIn == null)
         return new integer();
      else
         return editingFillIn.getCursorOffset();
   }

   /**
    * Sets the index position of the cursor in the edited cell in this column.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(double offset)
   {
      setCursorOffset(new decimal(offset));
   }

   /**
    * Sets the index position of the cursor in the edited cell in this column.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(NumberType offset)
   {
      if (editingFillIn != null)
         editingFillIn.setCursorOffset(offset, frame.getOriginalName(this));
   }

   /**
    * Set value of the MAX-CHARS attribute. Applies only to combo-box browse columns.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   public void setMaxChars(double maxChars)
   {
      setMaxChars(new decimal(maxChars));
   }

   /**
    * Set value of the MAX-CHARS attribute. Applies only to combo-box browse columns.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   public void setMaxChars(NumberType maxChars)
   {
      if (maxChars.isUnknown())
         return;

      config.maxChars = maxChars.intValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the MAX-CHARS attribute. Applies only to combo-box browse columns.
    *
    * @return   value of the attribute.
    */
   public integer getMaxChars()
   {
      return new integer(config.maxChars);
   }

   /**
    * Set value of the AUTO-ZAP attribute. Applies only to fill-in and combo-box browse columns.
    *
    * @param    autoZap
    *           Value of the attribute.
    */
   public void setAutoZap(logical autoZap)
   {
      if (editingFillIn != null)
         editingFillIn.setAutoZap(autoZap);
   }

   /**
    * Get value of the AUTO-ZAP attribute. Applies only to fill-in and combo-box browse columns.
    *
    * @return   value of the attribute.
    */
   public logical isAutoZap()
   {
      return new logical(true); //FillInWidget.isFocusedAutoZap();
   }

   /**
    * Set DISABLE-AUTO-ZAP attribute. Applies only to fill-in and combo-box browse columns.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setDisableAutoZap(boolean value)
   {
      config.disableAutoZap = value;
   }


   /**
    * Set DISABLE-AUTO-ZAP attribute. 
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setDisableAutoZap(logical value)
   {
      setDisableAutoZap(value == null || value.isUnknown() ? false : value.booleanValue());
   }

   /**
    * Get value of the DISABLE-AUTO-ZAP attribute
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isDisableAutoZap()
   {
      return new logical(config.disableAutoZap);
   }

   /**
    * Get the VISIBLE writable attribute.
    *
    * @return   The current value of VISIBLE attribute.
    */
   public boolean _isVisible()
   {
      return true;
   }

   /**
    * Gets the HIDDEN writable attribute.
    *
    * @return   The current value of the HIDDEN attribute.
    */
   public boolean _isHidden() 
   {
      return false;
   }
   
   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }
   
   /**
    * Gets the widget name.
    * 
    * @return   The widget name or unknown value if not configured.
    */
   @Override
   public character name()
   {
      character result = super.name();
      if (result != null && !result.isUnknown())
      {
         // NAME explicitly assigned
         return result;
      }
      
      // NAME not explicitly assigned, use the field name
      if (element instanceof Element)
      {
         Accessor acc = ((Element) element).getDataAccessor();
         if (acc instanceof FieldReference)
         {
            result = new character(((FieldReference) acc).getProperty());
         }
      }
      
      if (result == null)
      {
         result = new character();
      }
      
      return result;
   }
   
   /**
    * Getter for RESIZABLE attribute.
    *
    * @return  {@code true} if widget is resizable.
    */
   @Override
   public logical isResizable()
   {
      return new logical(config.resizable);
   }

   /**
    * Setter for RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} to make the widget resizable
    */
   @Override
   public void setResizable(boolean on)
   {
      boolean dirty = on != config.resizable;
      config.resizable = on;
      
      if (dirty)
      {
         pushScreenDefinition();
      }
   }

   /**
    * Setter for RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} to make the widget resizable
    */
   @Override
   public void setResizable(logical on)
   {
      // TODO: respond with error message when unknown 
      if (on != null && !on.isUnknown())
      {
         setResizable(on.booleanValue());
      }
   }

   /**
    * Gets the format specification.
    * 
    * @return   The format specification.
    */
   @Override
   public character getFormat()
   {
      character format = super.getFormat();
      
      if (!format.isUnknown())
      {
         return format;
      }
      
      if (element != null)
      {
         if (element.hasFormat())
         {
            return new character(element.getFormat());
         }
         
         BaseDataType data = element.get();
         
         if (data != null)
         {
            return new character(data.defaultFormatString());
         }
      }
      
      return new character();
   }

   /**
    * Sets the SORT-ASCENDING attribute, which defines the state of the sort arrow in the column
    * header.
    *
    * @param    ascending
    *           The new value for the SORT-ASCENDING attribute. Unknown value to hide the arrow,
    *           <code>true</code> for ascending sorting (arrow up). <code>false</code> for
    *           descending sorting (arrow down).
    */
   @Override
   public void setSortAscending(logical ascending)
   {
      config.sortAscending = ascending.toJavaType();
      pushScreenDefinition();
   }

   /**
    * Sets the SORT-ASCENDING attribute, which defines the state of the sort arrow in the column
    * header.
    *
    * @param    ascending
    *           The new value for the SORT-ASCENDING attribute. Unknown value to hide the arrow,
    *           <code>true</code> for ascending sorting (arrow up). <code>false</code> for
    *           descending sorting (arrow down).
    */
   @Override
   public void setSortAscending(boolean ascending)
   {
      setSortAscending(new logical(ascending));
   }

   /**
    * Gets the SORT-ASCENDING attribute, which defines the state of the sort arrow in the column
    * header.
    *
    * @return   The current value of the SORT-ASCENDING attribute. Unknown value if the arrow is
    *           hidden, <code>true</code> for ascending sorting (arrow up). <code>false</code> for
    *           descending sorting (arrow down).
    */
   @Override
   public logical isSortAscending()
   {
      return new logical(config.sortAscending);
   }

   /**
    * Sets the SORT-NUMBER attribute, which indicates the sort-level number for the browse column.
    * The number is displayed next to the sorting arrow.
    *
    * @param    sortNumber
    *           The new value for the SORT-NUMBER attribute. Unknown or <code>0</code> to hide the
    *           sort-level number, otherwise numbers from <code>1</code> to <code>9</code> are
    *           allowed.
    */
   @Override
   public void setSortNumber(integer sortNumber)
   {
      Integer num = sortNumber.toJavaIntegerType();
      if (num != null && (num < 0 || num > 9))
      {
         ErrorManager.recordOrShowError(13686,
               "Sort number must be between 0 and 9 or the unknown value", false);
         return;
      }

      config.sortNumber = sortNumber.toJavaIntegerType();
      pushScreenDefinition();
   }

   /**
    * Sets the SORT-NUMBER attribute, which indicates the sort-level number for the browse column.
    * The number is displayed next to the sorting arrow.
    *
    * @param    sortNumber
    *           The new value for the SORT-NUMBER attribute. Unknown or <code>0</code> to hide the
    *           sort-level number, otherwise numbers from <code>1</code> to <code>9</code> are
    *           allowed.
    */
   @Override
   public void setSortNumber(int sortNumber)
   {
      setSortNumber(new integer(sortNumber));
   }

   /**
    * Get the SORT-NUMBER attribute, which indicates the sort-level number for the browse column.
    * The number is displayed next to the sorting arrow.
    *
    * @return   The current value of the SORT-NUMBER attribute.
    */
   @Override
   public integer getSortNumber()
   {
      return new integer(config.sortNumber);
   }

   /**
    * Set fill-in widget which is used for in-browse editing for this column.
    *
    * @param editingFillIn
    *        Fill-in widget which is used for in-browse editing for this column.
    */
   void setEditingFillIn(FillInWidget editingFillIn)
   {
      this.editingFillIn = editingFillIn;
   }

   /**
    * Get fill-in widget which is used for in-browse editing for this column.
    *
    * @return  fill-in widget which is used for in-browse editing for this column or
    *          <code>null</code> if the column is not editable.
    */
   FillInWidget getEditingFillIn()
   {
      return editingFillIn;
   }
}