/*
** Module   : BrowseConfig.java
** Abstract : browse widget attribute container.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SIY 20050909   @22824 Created initial version.
** 002 NVS 20051116   @23346 Constructor access changed to public. 
** 003 SIY 20051130   @23588 Added get/setQuery(). 
** 004 SIY 20060209   @24490 Major refactoring.
** 005 SIY 20060315   @25123 Changes for complete BROWSE implementation.
** 006 SIY 20060329   @25277 Improved encapsulation, added BOX attribute,
**                           added WIDTH attribute.
** 007 SIY 20060406   @25401 Added centered attribute.
** 008 GES 20060712   @27995 Added help attribute.
** 009 GES 20060803   @28392 Made externalizable for better performance.
**                           Many code standard cleanups.
** 010 SIY 20070319   @32460 Avoid setting DOWN if size is specified 
**                           in different way.
** 011 NVS 20070426   @33248 Removed checks of SENSITIVE attribute as
**                           redundant.
** 012 ECF 20080208   @37033 Added getSelectedRowIndex(). Gets the row
**                           index of a selected row in the browse widget.
** 013 SIY 20100619          Switching to own implementation of character UI.
** 014 SIY 20100804          Changes in imports, some refactorings.
** 015 SIY 20110112          Changes related to new color handling scheme.
** 016 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 017 SIY 20111202          Updated imports.
** 018 GES 20140627          Changed color attribute method names to match naming standards.
** 019 CA  20140711          Enhanced font support, for GUI clients.
** 020 HC  20140911          GUI coordinate support - data types refactoring.
** 021 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 022 CA  20141104          Removed width - it collides with BaseConfig.widthChars.
** 023 HC  20141021          Implemented window move/position support.
** 024 SVL 20141102          Added some attributes.
** 025 GES 20141124          Javadoc fix.
** 026 HC  20150323          Removed business logic from config classes.
** 027 SVL 20150605          Added editPossible.
** 028 CA  20150722          The BROWSE transfers to the client-side only the IDs of its 
**                           associated BrowseColumn widgets.  The actual column definitions are
**                           transferred via pushScreenDefinition.
** 029 SVL 20150722          Added repositionedRow, conditionalReposition.
** 030 SVL 20151013          Added colors and fonts variables, changed some defaults.
** 031 IAS 20151028          Added NO-VALIDATE attribute.
** 032 SVL 20151027          ROW-MARKERS attribute now has "default" value.
**     SVL 20151101          Added maxDataReached.
** 033 SVL 20151230          Javadoc fix.
** 034 IAS 20151230          MOUSE-POINTER attribute support.
** 035 SVL 20160418          Added minColumnWidthPixels and minHeightChars.
** 036 SVL 20160922          Switched from BrowseConfig.queryId to BrowseConfig.queryAssigned.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import java.io.*;

/**
 * This class holds all necessary attributes for the Browse widget.
 */
public class BrowseConfig
extends BaseConfig
{
   /** Default value for MAX-DATA-GUESS attribute. */
   public static final int DATA_GUESS_DEFAULT_VALUE = 100;

   /** NUM-ITERATIONS attribute variable. */
   public int numIterations = 0;            

   /** SEPARATORS attribute variable. */
   public boolean separators = false;      

   /** 
    * ALLOW-COLUMN-SEARCHING attribute variable.  
    * <p>
    * ISSUE: not implemented.
    */
   public boolean allowColumnSearching = false;

   /** 
    * COLUMN-SCROLLING attribute variable.
    * <p>
    * ISSUE: non-default behavior is not implemented.
    */
   public boolean columnScrolling = true;             

   /** DOWN attribute variable */
   public int down = 0;

   /** LABELS attribute variable. */
   public boolean labels = true;            

   /** 
    * MAX-DATA-GUESS attribute variable. 
    */
   public int maxDataGuess = DATA_GUESS_DEFAULT_VALUE;

   /** If the end of the result set has been reached. */
   public boolean maxDataReached = false;

   /** 
    * NUM-LOCKED-COLUMNS attribute variable. 
    */
   public int numLockedColumns = 0;                

   /** NUM-VISIBLE-COLUMNS attribute variable. */
   public int numVisibleColumns = 1;        

   /** READ-ONLY attribute variable. */
   public boolean readOnly = false;        

   /** REFRESHABLE attribute variable. */
   public boolean refreshable = true;      

   /** 
    * ROW-HEIGHT-CHARS attribute variable. 
    * <p>
    * ISSUE: non-default values are not implemented.
    */
   public double rowHeightChars = 0;

   /**
    * ROW-MARKERS attribute variable. <code>null</code> for default value (depends on GUI/ChUI,
    * static/dynamic, editable/read-only).
    */
   public Boolean rowMarkers = null;

   /** SCROLLBAR-VERTICAL attribute variable. */
   public boolean scrollbarVertical = true;

   /** HELP attribute variable. */
   public String help = null;             

   /** Support multiple selection. */
   public boolean multiple = false;             

   /** <code>true</code> if a query was assigned to the browse. */
   public boolean queryAssigned = false;

   /** BOX attribute. */
   public boolean box = true;

   // TODO: not clear if it has any meaning or used at all
   /** OVERLAY attribute. */
   public boolean overlay = false;

   /** CENTERED attribute. */
   public boolean centered = false;

   /** NO-ASSIGN attribute. */
   public boolean noAssign = false;

   /** FIT-LAST-COLUMN attribute. */
   public boolean fitLastColumn = false;

   /** NO-UNDERLINE attribute. */
   public boolean noUnderline = false;

   /** SCROLLABLE attribute. */
   public boolean scrollable = false;

   /** COLUMN-RESIZABLE attribute. */
   public boolean columnResizable = false;

   /** COLUMN-MOVABLE attribute. */
   public boolean columnMovable = false;

   /** CREATE-ON-ADD attribute. */
   public boolean createOnAdd = false; //TODO check if the default value is correct

   /** Determine if in-browse editing is possible. */
   public boolean editPossible = false;

   /** 0-based repositioned row (see SET-REPOSITIONED-ROW). */
   public int repositionedRow = 0;

   /** NO-VALIDATE option. */
   public boolean noValidate = false;

   /**
    * <code>true</code> if reposition mode is "CONDITIONAL", <code>false</code> if reposition mode
    * is "ALWAYS" (see SET-REPOSITIONED-ROW).
    */
   public boolean conditionalReposition = false;

   /** The IDs of the widgets corresponding with browse's columns. */
   public int[] columnIds = new int[0];

   /** Client-side only flag indicating that layout needs to be performed or not. */
   public transient boolean performLayout = true;

   /** SEPARATOR-FGCOLOR option */
   public int separatorFgColor = -1;

   /** COLUMN-FGCOLOR option */
   public int columnFgColor = -1;

   /** COLUMN-BGCOLOR option */
   public int columnBgColor = -1;

   /** LABEL-FGCOLOR option */
   public int labelFgColor = -1;

   /** LABEL-BGCOLOR option */
   public int labelBgColor = -1;

   /** LABEL-FONT option */
   public int labelFont = FontManager.UNDEFINED_FONT;
   
   /** MOUSE-POINTER attribute */
   public String mousePointer = null;

   /** Minimum width of a browse column in pixels. */
   public int minColumnWidthPixels = 1;

   /** Minimum height of the browse, in character units. */
   public Double minHeightChars = null;

   /**
    * Default constructor (only used in de-serialization).
    */
   public BrowseConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected BrowseConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected BrowseConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      BrowseConfig cfg = (BrowseConfig) config;
      
      numIterations = cfg.numIterations;
      separators = cfg.separators;
      allowColumnSearching = cfg.allowColumnSearching;
      columnScrolling = cfg.columnScrolling;
      down = cfg.down;
      labels = cfg.labels;
      maxDataGuess = cfg.maxDataGuess;
      maxDataReached = cfg.maxDataReached;
      numLockedColumns = cfg.numLockedColumns;
      numVisibleColumns = cfg.numVisibleColumns;
      readOnly = cfg.readOnly;
      refreshable = cfg.refreshable;
      rowHeightChars = cfg.rowHeightChars;
      rowMarkers = cfg.rowMarkers;
      scrollbarVertical = cfg.scrollbarVertical;
      help = cfg.help;
      multiple = cfg.multiple;
      queryAssigned = cfg.queryAssigned;
      box = cfg.box;
      overlay = cfg.overlay;
      centered = cfg.centered;
      noAssign = cfg.noAssign;
      fitLastColumn = cfg.fitLastColumn;
      noUnderline = cfg.noUnderline;
      scrollable = cfg.scrollable;
      columnResizable = cfg.columnResizable;
      columnMovable = cfg.columnMovable;
      createOnAdd = cfg.createOnAdd;
      editPossible = cfg.editPossible;
      repositionedRow = cfg.repositionedRow;
      conditionalReposition = cfg.conditionalReposition;
      separatorFgColor = cfg.separatorFgColor;
      columnFgColor = cfg.columnFgColor;
      columnBgColor = cfg.columnBgColor;
      labelFgColor = cfg.labelFgColor;
      labelBgColor = cfg.labelBgColor;
      labelFont = cfg.labelFont;
      noValidate = cfg.noValidate;
      mousePointer = cfg.mousePointer;
      minColumnWidthPixels = cfg.minColumnWidthPixels;
      minHeightChars = cfg.minHeightChars;
      
      columnIds = new int[cfg.columnIds.length];
      System.arraycopy(cfg.columnIds, 0, columnIds, 0, cfg.columnIds.length);
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      numIterations         = in.readInt();
      down                  = in.readInt();
      numLockedColumns      = in.readInt();
      rowHeightChars        = in.readDouble();
      queryAssigned         = in.readBoolean();
      separators            = in.readBoolean();
      allowColumnSearching  = in.readBoolean();
      columnScrolling       = in.readBoolean();
      labels                = in.readBoolean();
      readOnly              = in.readBoolean();
      refreshable           = in.readBoolean();
      rowMarkers            = (Boolean) in.readObject();
      scrollbarVertical     = in.readBoolean();
      multiple              = in.readBoolean();
      box                   = in.readBoolean();
      overlay               = in.readBoolean();
      centered              = in.readBoolean();
      help                  = (String) in.readObject();
      maxDataGuess          = in.readInt();
      maxDataReached        = in.readBoolean();
      numVisibleColumns     = in.readInt();
      noAssign              = in.readBoolean();
      fitLastColumn         = in.readBoolean();
      noUnderline           = in.readBoolean();
      scrollable            = in.readBoolean();
      columnResizable       = in.readBoolean();
      columnMovable         = in.readBoolean();
      createOnAdd           = in.readBoolean();
      editPossible          = in.readBoolean();
      repositionedRow       = in.readInt();
      conditionalReposition = in.readBoolean();
      separatorFgColor      = in.readInt();
      columnFgColor         = in.readInt();
      columnBgColor         = in.readInt();
      labelFgColor          = in.readInt();
      labelBgColor          = in.readInt();
      labelFont             = in.readInt();
      noValidate            = in.readBoolean();
      mousePointer          = (String) in.readObject();
      minColumnWidthPixels  = in.readInt();
      minHeightChars        = (Double) in.readObject();
      
      int colNo = in.readInt();
      columnIds = new int[colNo];
      for (int i = 0; i < colNo; i++)
      {
         columnIds[i] = in.readInt();
      }
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(numIterations);
      out.writeInt(down);
      out.writeInt(numLockedColumns);
      out.writeDouble(rowHeightChars);
      out.writeBoolean(queryAssigned);
      out.writeBoolean(separators);
      out.writeBoolean(allowColumnSearching);
      out.writeBoolean(columnScrolling);
      out.writeBoolean(labels);
      out.writeBoolean(readOnly);
      out.writeBoolean(refreshable);
      out.writeObject(rowMarkers);
      out.writeBoolean(scrollbarVertical);
      out.writeBoolean(multiple);
      out.writeBoolean(box);
      out.writeBoolean(overlay);
      out.writeBoolean(centered);
      out.writeObject(help);
      out.writeInt(maxDataGuess);
      out.writeBoolean(maxDataReached);
      out.writeInt(numVisibleColumns);
      out.writeBoolean(noAssign);
      out.writeBoolean(fitLastColumn);
      out.writeBoolean(noUnderline);
      out.writeBoolean(scrollable);
      out.writeBoolean(columnResizable);
      out.writeBoolean(columnMovable);
      out.writeBoolean(createOnAdd);
      out.writeBoolean(editPossible);
      out.writeInt(repositionedRow);
      out.writeBoolean(conditionalReposition);
      out.writeInt(separatorFgColor);
      out.writeInt(columnFgColor);
      out.writeInt(columnBgColor);
      out.writeInt(labelFgColor);
      out.writeInt(labelBgColor);
      out.writeInt(labelFont);
      out.writeBoolean(noValidate);
      out.writeObject(mousePointer);
      out.writeInt(minColumnWidthPixels);
      out.writeObject(minHeightChars);

      // write column IDs
      out.writeInt(columnIds.length);
      for (int i = 0; i < columnIds.length; i++)
      {
         out.writeInt(columnIds[i]);
      }
   }
}