/*
** Module   : BrowseInterface.java
** Abstract : Define all BROWSE-specific widget attributes and methods available via a handle
**            instance.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SVL 20141106 Created initial version.
** 002 SVL 20141209 Updated javadoc.
** 003 GES 20141124 Moved scrollbar-vertical to a separate interface since it is not browse-only.
** 004 SVL 20150503 Added addLikeColumn(String, integer), getBrowseColumn.
** 005 SVL 20150804 Added ROW-MARKERS attribute.
** 006 SVL 20151013 Added some colors and fonts accessors.
** 007 OM  20150922 Overloaded MOVE-COLUMN with missing signatures. Added NUM-ITERATIONS
**                  attribute. Switch to long/int64 datatypes. Added missing attributes/methods.
** 008 SVL 20151019 Added COLUMN-SCROLLING attribute.
** 009 IAS 20151028 Added NO-VALIDATE attribute.
** 010 EVL 20160217 Clean up comments from symbols invalid for Solaris to compile.
** 011 HC  20160714 Fixed the data type of the converted GET-BROWSE-COLUMN in the Java APIs.
** 012 SVL 20170126 Added getCurrentColumn.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Define all BROWSE-specific widget attributes and methods which can be accessed via a handle
 * instance.  Attributes and methods shared by multiple widget types will be defined elsewhere.
 */
public interface BrowseInterface
{
   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   @LegacyMethod(name = "FETCH-SELECTED-ROW")
   public logical fetchSelectedRow(int64 n);
   
   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   @LegacyMethod(name = "FETCH-SELECTED-ROW")
   public logical fetchSelectedRow(long n);
   
   /**
    * Get value of the FIT-LAST-COLUMN attribute.
    *
    * @return  value of the attribute.
    */
   @LegacyAttribute(name = "FIT-LAST-COLUMN")
   public logical isFitLastColumn();

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "FIT-LAST-COLUMN", setter = true)
   public void setFitLastColumn(boolean fitLastColumn);

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "FIT-LAST-COLUMN", setter = true)
   public void setFitLastColumn(logical fitLastColumn);

   /**
    * Set value of the COLUMN-SCROLLING attribute.
    *
    * @param    columnScrolling
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "COLUMN-SCROLLING", setter = true)
   public void setColumnScrolling(boolean columnScrolling);

   /**
    * Set value of the COLUMN-SCROLLING attribute.
    *
    * @param    columnScrolling
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "COLUMN-SCROLLING", setter = true)
   public void setColumnScrolling(logical columnScrolling);

   /**
    * Get value of the COLUMN-SCROLLING attribute.
    *
    * @return   <code>true</code> if browse is scrolled horizontally per column.
    *           <code>false</code> for usual scrolling.
    */
   @LegacyAttribute(name = "COLUMN-SCROLLING")
   public logical isColumnScrolling();

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "EXPANDABLE", setter = true)
   public void setExpandable(boolean expandable);

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "EXPANDABLE", setter = true)
   public void setExpandable(logical expandable);

   /**
    * Get value of the EXPANDABLE attribute.
    *
    * @return    value of the attribute.
    */
   @LegacyAttribute(name = "EXPANDABLE")
   public logical isExpandable();

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-RESIZABLE", setter = true)
   public void setColumnResizable(boolean resizable);

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-RESIZABLE", setter = true)
   public void setColumnResizable(logical resizable);

   /**
    * Get value of the COLUMN-RESIZABLE attribute.
    *
    * @return   <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-RESIZABLE")
   public logical isColumnResizable();

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-MOVABLE", setter = true)
   public void setColumnMovable(boolean movable);

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-MOVABLE", setter = true)
   public void setColumnMovable(logical movable);

   /**
    * Get value of the COLUMN-MOVABLE attribute.
    *
    * @return   <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   @LegacyAttribute(name = "COLUMN-MOVABLE")
   public logical isColumnMovable();

   /**
    * Get value of the CREATE-ON-ADD attribute.
    *
    * @return   value of the attribute.
    */
   @LegacyAttribute(name = "CREATE-ON-ADD")
   public logical isCreateOnAdd();

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "CREATE-ON-ADD", setter = true)
   public void setCreateOnAdd(boolean createOnAdd);

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "CREATE-ON-ADD", setter = true)
   public void setCreateOnAdd(logical createOnAdd);

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   @LegacyAttribute(name = "REFRESHABLE", setter = true)
   public void setRefreshable(boolean refreshable);

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   @LegacyAttribute(name = "REFRESHABLE", setter = true)
   public void setRefreshable(logical refreshable);

   /**
    * Get value of the REFRESHABLE attribute.
    *
    * @return   <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   @LegacyAttribute(name = "REFRESHABLE")
   public logical isRefreshable();

   /**
    * Get value of the CURRENT-ROW-MODIFIED attribute.
    *
    * @return   <code>true</code> if any cells in the current row have been changed.
    */
   @LegacyAttribute(name = "CURRENT-ROW-MODIFIED")
   public logical isCurrentRowModified();

   /**
    * Get value of the NEW-ROW attribute.
    *
    * @return   <code>true</code> if the row in focus was added to the browse using the
    *           INSERT-ROW() method and has not been added to the database.
    */
   @LegacyAttribute(name = "NEW-ROW")
   public logical isNewRow();

   /**
    * Get value of the ROW-MARKERS attribute.
    *
    * @return   <code>true</code> if the browse has row markers.
    */
   @LegacyAttribute(name = "ROW-MARKERS")
   public logical hasRowMarkers();

   /**
    * Set value of the ROW-MARKERS attribute.
    *
    * @param    markers
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "ROW-MARKERS", setter = true)
   public void setRowMarkers(boolean markers);

   /**
    * Set value of the ROW-MARKERS attribute.
    *
    * @param    markers
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "ROW-MARKERS", setter = true)
   public void setRowMarkers(logical markers);

   /**
    * Gets the LABEL-BGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-BGCOLOR attribute.
    */
   @LegacyAttribute(name = "LABEL-BGCOLOR")
   public integer getLabelBgColor();

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   @LegacyAttribute(name = "LABEL-BGCOLOR", setter = true)
   public void setLabelBgColor(int64 bgcolor);

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   @LegacyAttribute(name = "LABEL-BGCOLOR", setter = true)
   public void setLabelBgColor(long bgcolor);

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "ROW-HEIGHT-CHARS", setter = true)
   public void setRowHeightChars(double rowHeightChars);

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   @LegacyAttribute(name = "ROW-HEIGHT-CHARS", setter = true)
   public void setRowHeightChars(NumberType rowHeightChars);

   /**
    * Implements the ROW-HEIGHT-CHARS attribute getter.<p>
    *
    * @return row height, in characters, of the browse.
    */
   @LegacyAttribute(name = "ROW-HEIGHT-CHARS")
   public decimal getRowHeightChars();

   /**
    * Implements the SELECT-FOCUSED-ROW() widget method.<p>
    * Selects the row that currently has focus in a browse widget, even if it is not currently
    * displayed.
    *
    * @return   Always <code>true</code>.
    */
   @LegacyMethod(name = "SELECT-FOCUSED-ROW")
   public logical selectFocusedRow();

   /**
    * Implements the DESELECT-FOCUSED-ROW() widget method.<p>
    * Deselects the row with current focus.
    *
    * @return   <code>true</code> if the browse has at least one record.
    */
   @LegacyMethod(name = "DESELECT-FOCUSED-ROW")
   public logical deselectFocusedRow();

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   @LegacyMethod(name = "SELECT-ROW")
   public logical selectRow(int64 row);

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   @LegacyMethod(name = "SELECT-ROW")
   public logical selectRow(long row);

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @return  TODO
    */
   @LegacyMethod(name = "SELECT-NEXT-ROW")
   public logical selectNextRow();

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @return  TODO
    */
   @LegacyMethod(name = "SELECT-PREV-ROW")
   public logical selectPreviousRow();

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   @LegacyMethod(name = "DESELECT-SELECTED-ROW")
   public logical deselectSelectedRow(int64 row);

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   @LegacyMethod(name = "DESELECT-SELECTED-ROW")
   public logical deselectSelectedRow(int row);

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   @LegacyMethod(name = "DELETE-SELECTED-ROW")
   public logical deleteSelectedRow(int64 row);

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   @LegacyMethod(name = "DELETE-SELECTED-ROW")
   public logical deleteSelectedRow(int row);

   /**
    * Implements the DELETE-SELECTED-ROWS() widget method.<p>
    * Deletes all currently selected rows from a browse and the associated results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if rows were successfully deleted.
    */
   @LegacyMethod(name = "DELETE-SELECTED-ROWS")
   public logical deleteSelectedRows();

   /**
    * Implements the DELETE-CURRENT-ROW() widget method.<p>
    * Deletes the most recently selected row from a browse and the results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   @LegacyMethod(name = "DELETE-CURRENT-ROW")
   public logical deleteCurrentRow();

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   @LegacyMethod(name = "IS-ROW-SELECTED")
   public logical isRowSelected(int64 row);

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   @LegacyMethod(name = "IS-ROW-SELECTED")
   public logical isRowSelected(int row);

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   @LegacyMethod(name = "INSERT-ROW")
   public logical insertRow(character mode);

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   @LegacyMethod(name = "INSERT-ROW")
   public logical insertRow(String mode);

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   @LegacyMethod(name = "INSERT-ROW")
   public logical insertRow();

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(String fieldName, int64 position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(character fieldName, int64 position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(String fieldName, long position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse's query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(handle fieldHandle, int64 position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse's query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(handle fieldHandle, long position);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(character fieldName);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse's query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(String fieldName);

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse's query.
    *
    * @return handle to the created column.
    */
   @LegacyMethod(name = "ADD-LIKE-COLUMN")
   public handle addLikeColumn(handle fieldHandle);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   @LegacyMethod(name = "MOVE-COLUMN")
   public logical moveColumn(int64 src, int64 dest);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   @LegacyMethod(name = "MOVE-COLUMN")
   public logical moveColumn(long src, int64 dest);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   @LegacyMethod(name = "MOVE-COLUMN")
   public logical moveColumn(int64 src, long dest);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   @LegacyMethod(name = "MOVE-COLUMN")
   public logical moveColumn(long src, long dest);

   /**
    * Implements the GET-REPOSITIONED-ROW() widget method.<p>
    * Returns the row index of the browse viewport where the REPOSITION TO ROWID (or
    * RECID) statement displays a repositioned record.
    *
    * @return the row index of the browse viewport where the REPOSITION TO ROWID (or
    *         RECID) statement displays a repositioned record
    */
   @LegacyMethod(name = "GET-REPOSITIONED-ROW")
   public integer getRepositionedRow();

   /**
    * Implements the FIRST-COLUMN attribute getter.<p>
    *
    * @return a handle to the first column in a browse widget, regardless of the value of its
    *         READ-ONLY attribute or its VISIBLE attribute.
    */
   @LegacyAttribute(name = "FIRST-COLUMN")
   public handle getFirstColumn();

   /**
    * Implements the FOCUSED-ROW attribute getter.<p>
    *
    * @return the 1-based index of the focused row in the viewport.
    */
   @LegacyAttribute(name = "FOCUSED-ROW")
   public integer getFocusedRow();

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute getter.<p>
    *
    * @return the number of visible leading columns locked in a browse widget.
    */
   @LegacyAttribute(name = "NUM-LOCKED-COLUMNS")
   public integer getNumLockedColumns();

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   @LegacyAttribute(name = "NUM-LOCKED-COLUMNS", setter = true)
   public void setNumLockedColumns(double num);

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   @LegacyAttribute(name = "NUM-LOCKED-COLUMNS", setter = true)
   public void setNumLockedColumns(NumberType num);

   /**
    * Implements the MIN-HEIGHT-CHARS attribute getter.<p>
    *
    * @return  minimum height of the browse, in character units.
    */
   @LegacyAttribute(name = "MIN-HEIGHT-CHARS")
   public decimal getMinHeightChars();

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute getter.<p>
    *
    * @return <code>true</code> if column searching is allowed for the browse.
    */
   @LegacyAttribute(name = "ALLOW-COLUMN-SEARCHING")
   public logical isAllowColumnSearching();

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   @LegacyAttribute(name = "ALLOW-COLUMN-SEARCHING", setter = true)
   public void setAllowColumnSearching(boolean allow);

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   @LegacyAttribute(name = "ALLOW-COLUMN-SEARCHING", setter = true)
   public void setAllowColumnSearching(logical allow);

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   @LegacyAttribute(name = "MAX-DATA-GUESS", setter = true)
   public void setMaxDataGuess(double maxDataGuess);

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   @LegacyAttribute(name = "MAX-DATA-GUESS", setter = true)
   public void setMaxDataGuess(NumberType maxDataGuess);

   /**
    * Get value of the MAX-DATA-GUESS attribute.
    *
    * @return   the estimated number of records in a browse query.
    */
   @LegacyAttribute(name = "MAX-DATA-GUESS")
   public integer getMaxDataGuess();

   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @LegacyMethod(name = "SET-REPOSITIONED-ROW")
   public logical setRepositionedRow(int64 row);
   
   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @LegacyMethod(name = "SET-REPOSITIONED-ROW")
   public logical setRepositionedRow(long row);

   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @LegacyMethod(name = "SET-REPOSITIONED-ROW")
   public logical setRepositionedRow(int64 row, String cond);
   
   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   @LegacyMethod(name = "SET-REPOSITIONED-ROW")
   public logical setRepositionedRow(long row, String cond);
   
   /**
    * Gets the SCROLLABLE writeable attribute.
    *
    * @return   The current value of the SCROLLABLE attribute.
    */
   @LegacyAttribute(name = "SCROLLABLE")
   public logical isScrollable();

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   @LegacyAttribute(name = "SCROLLABLE", setter = true)
   public void setScrollable(logical scroll);

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   @LegacyAttribute(name = "SCROLLABLE", setter = true)
   public void setScrollable(boolean scroll);

   /**
    * Set SEPARATORS attribute.
    *
    * @param    separators
    *           New value for the attribute.
    */
   @LegacyAttribute(name = "SEPARATORS", setter = true)
   public void setSeparators(boolean separators);

   /**
    * Set SEPARATORS attribute.
    *
    * @param    separators
    *           New value for the attribute.
    */
   @LegacyAttribute(name = "SEPARATORS", setter = true)
   public void setSeparators(logical separators);

   /**
    * Implements the SEPARATORS attribute getter.<p>
    *
    * @return <code>true</code> if row and column separators of the browse widget are displayed.
    */
   @LegacyAttribute(name = "SEPARATORS")
   public logical isSeparators();

   /**
    * Get specified browse column.
    *
    * @param    index
    *           1-based column index.
    *
    * @return   reference to column widget.
    */
   @LegacyMethod(name = "GET-BROWSE-COLUMN")
   public handle getBrowseColumn(int index);

   /**
    * Get specified browse column.
    *
    * @param    index
    *           1-based column index.
    *
    * @return   reference to column widget.
    */
   @LegacyMethod(name = "GET-BROWSE-COLUMN")
   public handle getBrowseColumn(NumberType index);

   /**
    * Get the SEPARATOR-FGCOLOR attribute value.
    *
    * @return SEPARATOR-FGCOLOR attribute value.
    */
   @LegacyAttribute(name = "SEPARATOR-FGCOLOR")
   public integer getSeparatorFgColor();

   /**
    * Get the COLUMN-FGCOLOR attribute value.
    *
    * @return COLUMN-FGCOLOR attribute value.
    */
   @LegacyAttribute(name = "COLUMN-FGCOLOR")
   public integer getColumnFgColor();


   /**
    * Get the COLUMN-BGCOLOR attribute value.
    *
    * @return COLUMN-BGCOLOR attribute value.
    */
   @LegacyAttribute(name = "COLUMN-BGCOLOR")
   public integer getColumnBgColor();

   /**
    * Sets the SEPARATOR-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "SEPARATOR-FGCOLOR", setter = true)
   public void setSeparatorFgColor(long color);

   /**
    * Sets the SEPARATOR-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "SEPARATOR-FGCOLOR", setter = true)
   public void setSeparatorFgColor(int64 color);

   /**
    * Sets the COLUMN-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "COLUMN-FGCOLOR", setter = true)
   public void setColumnFgColor(long color);

   /**
    * Sets the COLUMN-FGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "COLUMN-FGCOLOR", setter = true)
   public void setColumnFgColor(int64 color);

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "COLUMN-BGCOLOR", setter = true)
   public void setColumnBgColor(long color);

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    color
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "COLUMN-BGCOLOR", setter = true)
   public void setColumnBgColor(int64 color);
   
   /**
    * Get the number of rows currently visible in a browse widget.
    * 
    * @return  the number of rows currently visible in a browse widget.
    */
   @LegacyMethod(name = "NUM-ITERATIONS")
   public integer visibleIterations();
   
   /**
    * Get the handle of the browse column's BUFFER-FIELD.
    * 
    * @return  the handle of the browse column's buffer-field.
    */
   @LegacyMethod(name = "BUFFER-FIELD")
   public handle bufferField();

   /**
    * Sets the NO-VALIDATE writable attribute.
    *
    * @param    value
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "NO-VALIDATE", setter = true)
   public void setNoValidate(boolean value);

   /**
    * Sets the NO-VALIDATE writable attribute.
    *
    * @param    value
    *           The new value for the attribute.
    */
   @LegacyAttribute(name = "NO-VALIDATE", setter = true)
   public void setNoValidate(logical value);

   /**
    * Gets the NO-VALIDATE attribute.
    *
    * @return   the value of the attribute.
    */
   @LegacyAttribute(name = "NO-VALIDATE")
   public logical isNoValidate();

   /**
    * Get the current column (a column can be selected by clicking on its header if
    * ALLOW-COLUMN-SEARCHING is <code>true</code>).
    *
    * @return   handle to the current column .
    */
   @LegacyAttribute(name = "CURRENT-COLUMN")
   public handle getCurrentColumn();

   /**
    * Set the current column (a column can be selected by clicking on its header if
    * ALLOW-COLUMN-SEARCHING is <code>true</code>).
    *
    * @param    columnHandle
    *           Handle to the new current column.
    */
   @LegacyAttribute(name = "CURRENT-COLUMN", setter = true)
   public void setCurrentColumn(handle columnHandle);
}