/*
** Module   : ButtonWidget.java
** Abstract : server side button widget implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23482 Created initial version.
** 002 GES 20060114   @23907 Removed totally useless screen-value attr.
** 003 GES 20060803   @28367 Some cleanup and refactoring.
** 004 GES 20061116   @31247 Exposed button-specific flags on the server
**                           side.
** 005 GES 20080320   @37582 Added TYPE attribute.
** 006 SIY 20101001          Added more button options.
** 007 SIY 20101004          Added IMAGE property.
** 008 SIY 20101006          Added LOAD-IMAGE method.
** 009 SIY 20111101          Updated imports.
** 010 CA  20130221          Refactored usage of TYPE attribute.
** 011 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 012 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 013 CA  20131115          Changes related to properly instantiate dynamic resources.
** 014 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 015 EVL 20141222          Moving tooltip related methods to the base class to be used for every
**                           widget that needs tooltip feature.
** 016 EVL 20141226          Adding support for methods LOAD-IMAGE-UP, LOAD-IMAGE-DOWN,
**                           LOAD-IMAGE-INSENSITIVE.
** 017 EVL 20150115          Rework for image support implementation.
** 018 EVL 20150128          Button widget now implements image loading interface.
** 019 EVL 20150130          Image phrase handling support added.
** 020 EVL 20150222          Adding boolean option getter/setter for CONVERT-3D-COLORS.
**                           NO-CONVERT-3D-COLORS option is now processing with CONVERT-3D-COLORS
**                           attribute that is really the same(with boolean inversion).
** 021 EVL 20150227          Adding support for FLAT-BUTTON option/attribute.
** 022 EVL 20150312          Adding DEFAULT attribute missing setters and getters.  Error message
**                           should be displayed when trying to change DEFAULT, NO-FOCUS and
**                           FLAT-BUTTON attributes when the button has already been realised.
** 023 EVL 20150316          Adding AUTO-GO and AUTO-ENDKEY attributes support.
** 024 HC  20150323          WidgetId.getId() renamed to asInt().
** 025 EVL 20150429          In prepackaged images case we can ony have up image. The required
**                           disabled state images are loading on the client side when up images
**                           are processing so save server-client rounds.  Fix for button
**                           resizing issue for very long labels.
** 026 IAS 20150924          Added AUTO-RESIZE attribute support
** 027 IAS 20151011          Fixed NPE in case of the late FRAME assignment
** 028 CA  20151016          Fixed NPE when LABEL attr for dyn widget is set before FRAME.
** 029 CA  20151026          Fixed labelSizeCheck to use driver-resolved label width.
**                           TODO: the cut suffix needs to be computed on client-side.
**                           labelSizeCheck checks width only if in auto-resize mode.
** 030 IAS 20151214          LOAD-MOUSE-POINTER support.
** 031 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 032 EVL 20160412          We need to take into account possible '&' accelerator symbol in
** 033 VIG 20160215          Overridden setLabel().
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.ImageHelper;

/**
 * Server side button widget implementation.
 */
@LegacyResource(resource = LegacyResource.BUTTON)
public class ButtonWidget
extends ControlEntity<ButtonConfig>
implements ButtonInterface,
           AutoResizableWidget,
           Hoverable
{
   /**
    * Default constructor.
    */
   public ButtonWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public ButtonWidget(boolean dynamic)
   {
      super(dynamic, new ButtonConfig());
   }
   
   /**
    * Set the flag for whether or not this is a default button.
    * 
    * @param    isDefault
    *           <code>true</code> to make it default or <code>false</code>
    *           to make it an ordinary button.
    */
   @Override
   public void setDefault(boolean isDefault)
   {
      if (!_isRealized())
      {
         if (config.isDefault == isDefault)
         {
            return;
         }
         
         config.isDefault = isDefault;
         pushScreenDefinition();
      }
      else
      {
         ErrorManager.displayError(4053, String.format(
               "Unable to set DEFAULT because the %s widget has been realised.", type()));
      }
   }

   /**
    * Set the flag for whether or not this is a default button.
    * 
    * @param    isDefault
    *           <code>true</code> to make it default or <code>false</code>
    *           to make it an ordinary button.
    */
   @Override
   public void setDefault(logical isDefault)
   {
      if (isDefault != null && !isDefault.isUnknown())
      {
         setDefault(isDefault.booleanValue());
      }
   }

   /**
    * Gets DEFAULT attribute.
    * 
    * @return  DEFAULT attribute current value.
    */
   @Override
   public logical getDefault()
   {
      return new logical(config.isDefault);
   }

   /**
    * Set the flag for the AUTO-GO attribute.
    * 
    * @param    autoGo
    *           The new value for the attribute.
    */
   @Override
   public void setAutoGo(boolean autoGo)
   {
      if (config.autoGo == autoGo)
      {
         return;
      }
      
      config.autoGo = autoGo;
      pushScreenDefinition();
   }

   /**
    * This is the special version of the AUTO-GO attribute setter to be used inside DEFINE BUTON
    * statement.  It disables AUTO_ENDKEY if it is already defined when AUTO-GO is true.
    * 
    * @param    autoGo
    *           The new value for the attribute.
    */
   public void setAutoGoOption(boolean autoGo)
   {
      // the both option can not exist simultaneously AUTO-GO has precedence in DEFINE BUTON
      if (autoGo && config.autoEndKey)
      {
         config.autoEndKey = false;
      }
      setAutoGo(autoGo);
   }

   /**
    * Set the flag for the AUTO-GO attribute.
    * 
    * @param    autoGo
    *           The new value for the attribute.
    */
   @Override
   public void setAutoGo(logical autoGo)
   {
      if (autoGo != null && !autoGo.isUnknown())
      {
         setAutoGo(autoGo.booleanValue());
      }
   }

   /**
    * Gets AUTO-GO attribute.
    * 
    * @return  AUTO-GO attribute current value.
    */
   @Override
   public logical isAutoGo()
   {
      return new logical(config.autoGo);
   }

   /**
    * Set the flag for the AUTO-ENDKEY attribute.
    * 
    * @param    autoEndKey
    *           The new value for the attribute.
    */
   @Override
   public void setAutoEndKey(boolean autoEndKey)
   {
      if (config.autoEndKey == autoEndKey)
      {
         return;
      }
      
      config.autoEndKey = autoEndKey;
      pushScreenDefinition();
   }

   /**
    * This is the special version of the AUTO-ENDKEY attribute setter to be used inside
    * DEFINE BUTON statement.  It does nothing when AUTO-GO is already defined as true.
    * 
    * @param    autoEndKey
    *           The new value for the attribute.
    */
   public void setAutoEndKeyOption(boolean autoEndKey)
   {
      // the both option can not exist simultaneously AUTO-GO has precedence in DEFINE BUTON
      if (autoEndKey && config.autoGo)
      {
         return;
      }
      setAutoEndKey(autoEndKey);
   }

   /**
    * Set the flag for the AUTO-ENDKEY attribute.
    * 
    * @param    autoEndKey
    *           The new value for the attribute.
    */
   @Override
   public void setAutoEndKey(logical autoEndKey)
   {
      if (autoEndKey != null && !autoEndKey.isUnknown())
      {
         setAutoEndKey(autoEndKey.booleanValue());
      }
   }

   /**
    * Gets AUTO-ENDKEY attribute.
    * 
    * @return  AUTO-ENDKEY attribute current value.
    */
   @Override
   public logical isAutoEndKey()
   {
      return new logical(config.autoEndKey);
   }

   /**
    * Returns the name of the image loaded by LOAD-IMAGE. Corresponds Progress
    * IMAGE attribute
    *
    * @return the name of the image loaded by LOAD-IMAGE. 
    */
   public character getImage()
   {
      return new character(config.imgDef.name);
   }

   /**
    * Set button image file name.
    * 
    * @param   name
    *          Image file name.
    */
   public void setImage(String name)
   {
      loadImage(name);
   }
   
   /**
    * Sets the width for image used to draw button in character units.
    *
    * @param   width
    *          The new value for the image width in characters.
    */
   public void setImageWidthChars(NumberType width)
   {
      if (!width.isUnknown())
      {
         setImageWidthChars(width.doubleValue());
      }
   }
   
   /**
    * Sets the width for image used to draw button in character units.
    *
    * @param   width
    *          The new value for the image width in characters.
    */
   public void setImageWidthChars(double width)
   {
      config.imgDef.widthChars = width;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for image used to draw button in character units.
    *
    * @param   height
    *          The new value for the image height in characters.
    */
   public void setImageHeightChars(NumberType height)
   {
      if (!height.isUnknown())
      {
         setImageHeightChars(height.doubleValue());
      }
   }
   
   /**
    * Sets the height for image used to draw button in character units.
    *
    * @param   height
    *          The new value for the image height in characters.
    */
   public void setImageHeightChars(double height)
   {
      config.imgDef.heightChars = height;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the width for image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the image width in pixels.
    */
   public void setImageWidthPixels(NumberType widthPixels)
   {
      if (!widthPixels.isUnknown())
      {
         setImageWidthPixels(widthPixels.intValue());
      }
   }
   
   /**
    * Sets the width for image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the image width in pixels.
    */
   public void setImageWidthPixels(int widthPixels)
   {
      config.imgDef.widthPixels = widthPixels;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for image used to draw button in pixel units.
    *
    * @param   heightPixels
    *          The new value for the image height in pixels.
    */
   public void setImageHeightPixels(NumberType heightPixels)
   {
      if (!heightPixels.isUnknown())
      {
         setImageHeightPixels(heightPixels.intValue());
      }
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param   heightPixels
    *          The new value for the image height in pixels.
    */
   public void setImageHeightPixels(int heightPixels)
   {
      config.imgDef.heightPixels = heightPixels;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of X-offset image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageOffsetX(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetX(offset.intValue());
      }
   }

   /**
    * Sets new value of X-offset image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageOffsetX(int offset)
   {
      config.imgDef.offsetX = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of Y-offset image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageOffsetY(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetY(offset.intValue());
      }
   }

   /**
    * Sets new value of Y-offset image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageOffsetY(int offset)
   {
      config.imgDef.offsetY = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of ROW-offset image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageOffsetRow(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetRow(offset.doubleValue());
      }
   }

   /**
    * Sets new value of ROW-offset image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageOffsetRow(double offset)
   {
      config.imgDef.offsetRow = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of COLUMN-offset image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageOffsetColumn(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetColumn(offset.doubleValue());
      }
   }

   /**
    * Sets new value of COLUMN-offset image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageOffsetColumn(double offset)
   {
      config.imgDef.offsetCol = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Set button up image file name.
    * 
    * @param   name
    *          Image file name.
    */
   public void setImageUp(String name)
   {
      loadImageUp(name);
   }

   /**
    * Sets the width for up image used to draw button in character units.
    *
    * @param   width
    *          The new value for the up image width in characters.
    */
   public void setImageUpWidthChars(NumberType width)
   {
      if (!width.isUnknown())
      {
         setImageUpWidthChars(width.doubleValue());
      }
   }
   
   /**
    * Sets the width for up image used to draw button in character units.
    *
    * @param   width
    *          The new value for the up image width in characters.
    */
   public void setImageUpWidthChars(double width)
   {
      config.imgUpDef.widthChars = width;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for up image used to draw button in character units.
    *
    * @param   height
    *          The new value for the up image height in characters.
    */
   public void setImageUpHeightChars(NumberType height)
   {
      if (!height.isUnknown())
      {
         setImageUpHeightChars(height.doubleValue());
      }
   }
   
   /**
    * Sets the height for up image used to draw button in character units.
    *
    * @param   height
    *          The new value for the up image height in characters.
    */
   public void setImageUpHeightChars(double height)
   {
      config.imgUpDef.heightChars = height;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the width for up image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the up image width in pixels.
    */
   public void setImageUpWidthPixels(NumberType widthPixels)
   {
      if (!widthPixels.isUnknown())
      {
         setImageUpWidthPixels(widthPixels.intValue());
      }
   }
   
   /**
    * Sets the width for up image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the up image width in pixels.
    */
   public void setImageUpWidthPixels(int widthPixels)
   {
      config.imgUpDef.widthPixels = widthPixels;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for up image used to draw button in pixel units.
    *
    * @param   heightPixels
    *          The new value for the up image height in pixels.
    */
   public void setImageUpHeightPixels(NumberType heightPixels)
   {
      if (!heightPixels.isUnknown())
      {
         setImageUpHeightPixels(heightPixels.intValue());
      }
   }

   /**
    * Sets the height for up image used to draw button in pixel units.
    *
    * @param   heightPixels
    *          The new value for the up image height in pixels.
    */
   public void setImageUpHeightPixels(int heightPixels)
   {
      config.imgUpDef.heightPixels = heightPixels;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of X-offset up image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageUpOffsetX(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageUpOffsetX(offset.intValue());
      }
   }

   /**
    * Sets new value of X-offset up image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageUpOffsetX(int offset)
   {
      config.imgUpDef.offsetX = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of Y-offset up image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageUpOffsetY(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageUpOffsetY(offset.intValue());
      }
   }

   /**
    * Sets new value of Y-offset up image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageUpOffsetY(int offset)
   {
      config.imgUpDef.offsetY = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of ROW-offset up image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageUpOffsetRow(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageUpOffsetRow(offset.doubleValue());
      }
   }

   /**
    * Sets new value of ROW-offset up image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageUpOffsetRow(double offset)
   {
      config.imgUpDef.offsetRow = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of COLUMN-offset up image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageUpOffsetColumn(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageUpOffsetColumn(offset.doubleValue());
      }
   }

   /**
    * Sets new value of COLUMN-offset up image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageUpOffsetColumn(double offset)
   {
      config.imgUpDef.offsetCol = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Set button down image file name.
    * 
    * @param   name
    *          Image file name.
    */
   public void setImageDown(String name)
   {
      loadImageDown(name);
   }
   
   /**
    * Sets the width for down image used to draw button in character units.
    *
    * @param   width
    *          The new value for the down image width in characters.
    */
   public void setImageDownWidthChars(NumberType width)
   {
      if (!width.isUnknown())
      {
         setImageDownWidthChars(width.doubleValue());
      }
   }
   
   /**
    * Sets the width for down image used to draw button in character units.
    *
    * @param   width
    *          The new value for the down image width in characters.
    */
   public void setImageDownWidthChars(double width)
   {
      config.imgDownDef.widthChars = width;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for down image used to draw button in character units.
    *
    * @param   height
    *          The new value for the down image height in characters.
    */
   public void setImageDownHeightChars(NumberType height)
   {
      if (!height.isUnknown())
      {
         setImageDownHeightChars(height.doubleValue());
      }
   }
   
   /**
    * Sets the height for down image used to draw button in character units.
    *
    * @param   height
    *          The new value for the down image height in characters.
    */
   public void setImageDownHeightChars(double height)
   {
      config.imgDownDef.heightChars = height;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the width for down image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the down image width in pixels.
    */
   public void setImageDownWidthPixels(NumberType widthPixels)
   {
      if (!widthPixels.isUnknown())
      {
         setImageDownWidthPixels(widthPixels.intValue());
      }
   }
   
   /**
    * Sets the width for down image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the down image width in pixels.
    */
   public void setImageDownWidthPixels(int widthPixels)
   {
      config.imgDownDef.widthPixels = widthPixels;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for down image used to draw button in pixel units.
    *
    * @param   heightPixels
    *          The new value for the down image height in pixels.
    */
   public void setImageDownHeightPixels(NumberType heightPixels)
   {
      if (!heightPixels.isUnknown())
      {
         setImageDownHeightPixels(heightPixels.intValue());
      }
   }

   /**
    * Sets the height for down image used to draw button in pixel units.
    *
    * @param   heightPixels
    *          The new value for the down image height in pixels.
    */
   public void setImageDownHeightPixels(int heightPixels)
   {
      config.imgDownDef.heightPixels = heightPixels;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of X-offset down image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageDownOffsetX(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageDownOffsetX(offset.intValue());
      }
   }

   /**
    * Sets new value of X-offset down image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageDownOffsetX(int offset)
   {
      config.imgDownDef.offsetX = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of Y-offset down image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageDownOffsetY(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageDownOffsetY(offset.intValue());
      }
   }

   /**
    * Sets new value of Y-offset down image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageDownOffsetY(int offset)
   {
      config.imgDownDef.offsetY = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of ROW-offset down image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageDownOffsetRow(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageDownOffsetRow(offset.doubleValue());
      }
   }

   /**
    * Sets new value of ROW-offset down image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageDownOffsetRow(double offset)
   {
      config.imgDownDef.offsetRow = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of COLUMN-offset down image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageDownOffsetColumn(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageDownOffsetColumn(offset.doubleValue());
      }
   }

   /**
    * Sets new value of COLUMN-offset down image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageDownOffsetColumn(double offset)
   {
      config.imgDownDef.offsetCol = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Set button insensitive image file name.
    * 
    * @param   name
    *          Image file name.
    */
   public void setImageIns(String name)
   {
      loadImageInsensitive(name);
   }
   
   /**
    * Sets the width for insensitive image used to draw button in character units.
    *
    * @param   width
    *          The new value for the insensitive image width in characters.
    */
   public void setImageInsWidthChars(NumberType width)
   {
      if (!width.isUnknown())
      {
         setImageInsWidthChars(width.doubleValue());
      }
   }
   
   /**
    * Sets the width for insensitive image used to draw button in character units.
    *
    * @param   width
    *          The new value for the insensitive image width in characters.
    */
   public void setImageInsWidthChars(double width)
   {
      config.imgInsDef.widthChars = width;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for insensitive image used to draw button in character units.
    *
    * @param   height
    *          The new value for the insensitive image height in characters.
    */
   public void setImageInsHeightChars(NumberType height)
   {
      if (!height.isUnknown())
      {
         setImageInsHeightChars(height.doubleValue());
      }
   }
   
   /**
    * Sets the height for insensitive image used to draw button in character units.
    *
    * @param   height
    *          The new value for the insensitive image height in characters.
    */
   public void setImageInsHeightChars(double height)
   {
      config.imgInsDef.heightChars = height;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the width for insensitive image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the insensitive image width in pixels.
    */
   public void setImageInsWidthPixels(NumberType widthPixels)
   {
      if (!widthPixels.isUnknown())
      {
         setImageInsWidthPixels(widthPixels.intValue());
      }
   }
   
   /**
    * Sets the width for insensitive image used to draw button in pixel units.
    *
    * @param   widthPixels
    *          The new value for the insensitive image width in pixels.
    */
   public void setImageInsWidthPixels(int widthPixels)
   {
      config.imgInsDef.widthPixels = widthPixels;
      pushScreenDefinition();
      return;
   }
   
   /**
    * Sets the height for insensitive image used to draw button in pixel units.
    *
    * @param   heightPixels
    *          The new value for the insensitive image height in pixels.
    */
   public void setImageInsHeightPixels(NumberType heightPixels)
   {
      if (!heightPixels.isUnknown())
      {
         setImageInsHeightPixels(heightPixels.intValue());
      }
   }

   /**
    * Sets the height for insensitive image used to draw button in pixel units.
    *
    * @param   heightPixels
    *          The new value for the insensitive image height in pixels.
    */
   public void setImageInsHeightPixels(int heightPixels)
   {
      config.imgInsDef.heightPixels = heightPixels;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of X-offset insensitive image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageInsOffsetX(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageInsOffsetX(offset.intValue());
      }
   }

   /**
    * Sets new value of X-offset insensitive image phrase option in pixel units.
    *
    * @param   offset
    *          new value of X-offset in pixels.
    */
   public void setImageInsOffsetX(int offset)
   {
      config.imgInsDef.offsetX = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of Y-offset insensitive image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageInsOffsetY(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageInsOffsetY(offset.intValue());
      }
   }

   /**
    * Sets new value of Y-offset insensitive image phrase option in pixel units.
    *
    * @param   offset
    *          new value of Y-offset in pixels.
    */
   public void setImageInsOffsetY(int offset)
   {
      config.imgInsDef.offsetY = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of ROW-offset insensitive image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageInsOffsetRow(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageInsOffsetRow(offset.doubleValue());
      }
   }

   /**
    * Sets new value of ROW-offset insensitive image phrase option in character units.
    *
    * @param   offset
    *          new value of ROW-offset in characters.
    */
   public void setImageInsOffsetRow(double offset)
   {
      config.imgInsDef.offsetRow = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of COLUMN-offset insensitive image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageInsOffsetColumn(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageInsOffsetColumn(offset.doubleValue());
      }
   }

   /**
    * Sets new value of COLUMN-offset insensitive image phrase option in character units.
    *
    * @param   offset
    *          new value of COLUMN-offset in characters.
    */
   public void setImageInsOffsetColumn(double offset)
   {
      config.imgInsDef.offsetCol = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Gets NO-FOCUS attribute.
    * 
    * @return  NO-FOCUS current value.
    */
   @Override
   public logical getNoFocus()
   {
      return new logical(config.noFocus);
   }

   /**
    * Set NO-FOCUS option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setNoFocus(boolean flag)
   {
      if (!_isRealized())
      {
         if (config.noFocus == flag)
         {
            return;
         }
         
         config.noFocus = flag;
         pushScreenDefinition();
      }
      else
      {
         ErrorManager.displayError(4053, String.format(
               "Unable to set NO-FOCUS because the %s %s has been realised.", type(),
                  frame.getName(this)));
      }
   }

   /**
    * Set NO-FOCUS option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setNoFocus(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setNoFocus(flag.booleanValue());
      }
   }

   /**
    * Gets FLAT-BUTTON attribute.
    * 
    * @return  FLAT-BUTTON current value.
    */
   @Override
   public logical getFlatButton()
   {
      return new logical(config.flatButton);
   }

   /**
    * Set FLAT-BUTTON attribute.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setFlatButton(boolean flag)
   {
      if (!_isRealized())
      {
         if (config.flatButton == flag)
         {
            return;
         }
         
         config.flatButton = flag;
         pushScreenDefinition();
      }
      else
      {
         ErrorManager.displayError(4053, String.format(
               "Unable to set FLAT-BUTTON because the %s %s has been realised.", type(),
                  frame.getName(this)));
      }
   }

   /**
    * Set FLAT-BUTTON attribute.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setFlatButton(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setFlatButton(flag.booleanValue());
      }
   }

   /**
    * Set NO-CONVERT-3D-COLORS option.
    * 
    * @param   flag
    *          New value for the option.
    */
   public void setNoConvert3D(boolean flag)
   {
      setConvert3D(!flag);
   }

   /**
    * Set CONVERT-3D-COLORS attribute.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setConvert3D(boolean flag)
   {
      config.convert3D = flag;
      pushScreenDefinition();
   }

   /**
    * Set CONVERT-3D-COLORS attribute.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setConvert3D(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setConvert3D(flag.booleanValue());
      }
   }

   /**
    * Gets CONVERT-3D-COLORS option.
    * 
    * @return  CONVERT-3D-COLORS current value.
    */
   @Override
   public logical getConvert3D()
   {
      return new logical(config.convert3D);
   }

   /**
    * Sets the label text, removing any '&amp;' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   @Override
   public void setLabel(String label)
   {
      setLabelInt(label);
   }
   
   /**
    * Load button image from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(String name)
   {
      if (loadImageInt(name, ButtonConfig.IMAGE_UP))
      {
         // only if image was loaded we need to change it's name in config
         config.imgDef.name = name;
         // for prepackaged images load the insensitive alternative too
         if (ImageHelper.isPrepackaged(name))
         {
            config.imgInsDef.name = name.concat(ButtonConfig.PREPACKAGED_INS_SUFFIX);
         }
         pushScreenDefinition();
         return new logical(true);
      }
      else
      {
         return new logical(false);
      }
   }

   /**
    * Load button image from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(character name)
   {
      return name != null && !name.isUnknown() ? loadImage(name.toStringMessage())
                                               : new logical(false);
   }

   /**
    * Load button image for normal state from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImageUp(String name)
   {
      if (loadImageInt(name, ButtonConfig.IMAGE_UP))
      {
         // only if image was loaded we need to change it's name in config
         config.imgUpDef.name = name;
         // for prepackaged images load the insensitive alternative too
         if (ImageHelper.isPrepackaged(name))
         {
            config.imgInsDef.name = name.concat(ButtonConfig.PREPACKAGED_INS_SUFFIX);
         }
         pushScreenDefinition();
         return new logical(true);
      }
      else
      {
         return new logical(false);
      }
   }

   /**
    * Load button image for normal state from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImageUp(character name)
   {
      return name != null && !name.isUnknown() ? loadImageUp(name.toStringMessage())
                                               : new logical(false);
   }

   /**
    * Load button image for pushed state from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImageDown(String name)
   {
      // for prepackaged images it is not possible to have down or insensitive images
      if (ImageHelper.isPrepackaged(config.imgDef.name) ||
          ImageHelper.isPrepackaged(config.imgUpDef.name))
      {
         return new logical(false);
      }
      
      if (loadImageInt(name, ButtonConfig.IMAGE_DOWN))
      {
         // only if image was loaded we need to change it's name in config
         config.imgDownDef.name = name;
         pushScreenDefinition();
         return new logical(true);
      }
      else
      {
         return new logical(false);
      }
   }

   /**
    * Load button image for pushed state from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImageDown(character name)
   {
      return name != null && !name.isUnknown() ? loadImageDown(name.toStringMessage())
                                               : new logical(false); 
   }

   /**
    * Load button image for disabled state from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImageInsensitive(String name)
   {
      // for prepackaged images it is not possible to have down or insensitive images
      if (ImageHelper.isPrepackaged(config.imgDef.name) ||
          ImageHelper.isPrepackaged(config.imgUpDef.name))
      {
         return new logical(false);
      }
      
      if (loadImageInt(name, ButtonConfig.IMAGE_INS))
      {
         // only if image was loaded we need to change it's name in config
         config.imgInsDef.name = name;
         pushScreenDefinition();
         return new logical(true);
      }
      else
      {
         return new logical(false);
      }
   }

   /**
    * Load button image for disabled state from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImageInsensitive(character name)
   {
      return name != null && !name.isUnknown() ? loadImageInsensitive(name.toStringMessage())
                                               : new logical(false); 
   }

   /**
    * Set AUTO-RESIZE to false;
    * 
    */
   public void resetAutoResize()
   {
      config.autoResize = false;
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Check it the requested new label size is OK for button widget.
    * 
    * @param label
    *        the new label to check
    * 
    * @return  the possibly modified label string that fits frame size it belongs to.
    */
   @Override
   protected String labelSizeCheck(String label)
   {
      if (frame == null)
      {
         return label;
      }
      
      // check if the label can fit the frame
      if (config.autoResize && frame._isVisible() && label != null && !label.isEmpty())
      {
         // new button width, ChUI mode needs two more chars for <> braces
         double lblLength = FontTable.getTextWidthChars(label).doubleValue();
         // check for possible accelerator
         int accelPos = label == null || label.isEmpty() ? -1 : label.indexOf("&");
         // if we have accelerator char - decrease the total label length 
         if (accelPos != -1 && accelPos < label.length() - 1)
         {
            lblLength -= 1;
         }
         // the button width to fit the new label
         double newBtnWidth = LogicalTerminal.isChui() ? lblLength + 2 : lblLength;
         
         // compute the expected delta to be removed
         // the formula is: (new button right edge) - (frame width) - 1,
         //   button right edge is (getColumn().intValue() + newBtnWidth - 1) for 0 based coord
         //   frame width is frame.getWidthChars().intValue() - 2, 2 chars are for frame border
         // so we can decrease some constants in final calculation
         double toCut = getColumn().doubleValue() + 
                        newBtnWidth - 
                        frame.getWidthChars().doubleValue();
         // do we have anything to remove?
         if (toCut > 0)
         {
            StringBuilder err = new StringBuilder();
            // form the message
            err.append("New width for ")
               .append(type())
               .append(" ")
               .append(frame.getName(this))
               .append(" is too large to fit in frame");
            // and display it
            ErrorManager.displayError(4475, err.toString());
            // truncate the label to fit the frame
            int cut = (int) Math.floor(lblLength - toCut);
            // TODO: this works only for ChUI, for GUI the suffix to cut is more complex to compute
            if (cut > 0 && LogicalTerminal.isChui())
            {
               label = label.substring(0, cut);
            }
            
            // disable auto resizing too
            config.autoResize = false;
         }
      }
      
      return label;
   }

   /**
    * Load image for button.  Internal version.
    * 
    * @param    imageSrc
    *           Image file name to load.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   private boolean loadImageInt(String imageSrc, int type)
   {
      // logical terminal will try to load image from application jar, then ask the
      // client to load from local file system in case of failure
      return LogicalTerminal.setImage(config.id.asInt(), type, imageSrc); 
   }
}