/*
** Module   : ComboBoxWidget.java
** Abstract : server side combo box widget implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23483 Created initial version.
** 002 SIY 20050209   @24433 Major refactoring.
** 003 SIY 20050209   @24494 Added missing setInnerLines method.
** 004 GES 20060803   @28368 Some cleanup and refactoring.
** 005 GES 20080320   @37583 Added TYPE attribute.
** 006 SIY 20111101          Updated imports.
** 007 CA  20130221          Refactored usage of TYPE attribute.
** 008 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 009 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 010 CA  20131115          Changes related to properly instantiate dynamic resources.
** 011 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 012 SVL 20141102          Added MAX-CHARS and AUTO-ZAP attributes.
** 013 GES 20141124          Added INNER-LINES interface, NUM-ITEMS attribute and the
**                           SET-SELECTION() method support.
** 014 IAS 20150123          Implemented DELETE() method
** 015 IAS 20150209          Fixed behavior to be compatible with 4GL.
** 016 IAS 20150225          Fixed support for the LIST-ITEMS and LIST-ITEM-PAIRS attributes
** 017 IAS 20150604          Added support for the REPLACE method, text selection, 
**                           AUTO-ZAP, DISABLE-AUTO-ZAP, CURSOR-OFFSET, NUM-ITEMS 
**                           and INNER-LINES attribute, realize() operation .
** 018 EVL 20150528          Adding mode setters implementation.  Adding MAX-CHARS attrubute
**                           setter to be compatible with current 4GL behavior.
** 019 EVL 20150727          Widget height can not be explicitly changed after realization.
** 020 HC  20151103          Improved GUI COMBO-BOX layout.
** 021 EVL 20151124          Fixed the inner lines calculation in finishSetup().
** 022 EVL 20151210          Combo-box ignores format while getting screen value. 
** 023 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 024 CA  20161005          Fixed issues related to SCREEN-VALUE attribute (initial value or 
**                           explicitly set, when the value is empty string or unknown, or the 
**                           widget is uninitialized).
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Server side combo box widget implementation.
 */
@LegacyResource(resource = LegacyResource.COMBO_BOX)
public class ComboBoxWidget
extends ControlSetEntity<ComboBoxConfig>
implements AutoZapElement,
           MaxCharsElement,
           NumItems,
           SortedElement,
           InnerLines,
           SelectableText
{
   /**
    * Default constructor.
    */
   public ComboBoxWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public ComboBoxWidget(boolean dynamic)
   {
      super(dynamic, new ComboBoxConfig());
   }
   
   /**
    * Set DROP-DOWN-LIST mode for the widget. Has no effect in character mode
    * interface.
    * 
    * @param   b
    *          <code>true</code> enables mode, <code>false</code> is
    *          ignored.
    */
   public void setModeDropDownList(boolean b)
   {
      // this mode should work in all UI cases
      if (b && config.mode != ComboBoxConfig.Mode.DROP_DOWN_LIST)
      {
         config.mode = ComboBoxConfig.Mode.DROP_DOWN_LIST;
         pushScreenDefinition();
      }
   }

   /**
    * Set DROP-DOWN mode for the widget. Has no effect in character mode
    * interface.
    * 
    * @param   b
    *          <code>true</code> enables mode, <code>false</code> is
    *          ignored.
    */
   public void setModeDropDown(boolean b)
   {
      // this is valid only in GUI
      if (b && !LogicalTerminal.isChui() && config.mode != ComboBoxConfig.Mode.DROP_DOWN)
      {
         config.mode = ComboBoxConfig.Mode.DROP_DOWN;
         pushScreenDefinition();
      }
   }
   
   /**
    * Set SIMPLE mode for the widget. Has no effect in character mode
    * interface.
    * 
    * @param   b
    *          <code>true</code> enables mode, <code>false</code> is
    *          ignored.
    */
   public void setModeSimple(boolean b)
   {
      // this is valid only in GUI
      if (b && !LogicalTerminal.isChui() && config.mode != ComboBoxConfig.Mode.SIMPLE)
      {
         config.mode = ComboBoxConfig.Mode.SIMPLE;
         pushScreenDefinition();
      }
   }

   /**
    * Get the NUM-ITEMS attribute.
    *
    * @return   The numebr of items in the list.
    */
   @Override
   public integer getNumItems()
   {
      return new integer(config.items == null ? 0 : config.items.length);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int64 end)
   {
      return setSelection(start == null || start.isUnknown() ? 0 : start.intValue(), 
            end == null || end.isUnknown() ? 0 : end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int end)
   {
      return setSelection(start == null || start.isUnknown() ? 0 : start.intValue(), end);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int64 end)
   {
      return setSelection(start, end == null || end.isUnknown() ? 0 : end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int end)
   {
      if (!canAccess("SET-SELECTION"))
      {
         return new logical();
      }

      if (start == 0 || end == 0)
      {
         ErrorManager.recordOrShowError(4057, 
               String.format(
               "Attribute SET-SELECTION for the %s %s has an invalid value of 0", 
               type(), widgetName()),
               false);
         return new logical();
      }
      int s1 = Math.min(start, end);
      int s2 = Math.max(start,  end);
      if (s1 < 0 || s2 > config.widthChars - 4)
      {
         return new logical(false);
      }

      config.selectionStart = s1;
      config.selectionEnd = s2;
      config.selectionActive = true;
      pushScreenDefinition();
      return new logical(true);
   }
   
   /**
    * Get SELECTION-TEXT attribute.
    *
    * @return  SELECTION-TEXT value   
    */
   public character getSelectionText()
   {
      if (!canAccess("SELECTION-TEXT"))
      {
         return new character();
      }

      if (!config.selectionActive)
      {
         return new character("");
      }
      String s = "";
      if (config.items != null && 
            config.selectedIndex >= 0 && config.selectedIndex < config.items.length)
      {
         ControlSetItem item = config.items[config.selectedIndex];
         s = item.getLabel().toStringMessage();
      }
      
      s += StringHelper.repeatChar(' ', (int)config.widthChars -4 - s.length());
      return new character(
            config.selectionEnd + 1 >= (int)config.widthChars - 4 ?
               s.substring(config.selectionStart) :
               s.substring(config.selectionStart, config.selectionEnd + 1)
            );
   }

   /**
    * Get SELECTION-END attribute.
    *
    * @return  SELECTION-END value   
    */
   public integer getSelectionEnd()
   {
      if (!canAccess("SELECTION-END"))
      {
         return new integer();
      }

      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionEnd);
   }

   /**
    * Get SELECTION-START attribute.
    *
    * @return  SELECTION-START value   
    */
   public integer getSelectionStart()
   {
      if (!canAccess("SELECTION-START"))
      {
         return new integer();
      }

      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionStart);
   }

   /**
    * Get TEXT-SELECTED attribute.
    *
    * @return  TEXT-SELECTED value   
    */
   @Override
   public logical isTextSelected()
   {
      if (!canAccess("TEXT-SELECTION"))
      {
         return new logical();
      }
      return new logical(config.selectionActive);
   }

   /**
    * Clears text selection
    */
   @Override
   public logical clearSelection()
   {
      boolean wasSelection = config.selectionActive;
      config.selectionActive = false;
      pushScreenDefinition();
      return new logical(wasSelection);
   }

   /**
    * Gets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @return   The number of lines of text.
    */
   @Override
   public integer getInnerLines()
   {
      if (!config.realized)
      {
         return new integer(0);
      }
      return new integer(config.innerLines);
   }
   
   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget. For the INNER-LINES clause in the VIEW-AS phrase
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInitialInnerLines(double lines)
   {
      setInnerLines(lines);
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInnerLines(double lines)
   {
      // WARNING: even if the lines are the same, this can't be a no-op. 
      // also, if the initial assign value is zero, it defaults to 3

      int value = (int) lines;
      if (value < 0)
      {
         ErrorManager.recordOrShowError(4082, 
               String.format(
               "Attribute INNER-LINES must be greater than or equal to 0 on %s %s", 
               type(), widgetName()),
               false);
         return;
      }

      config.innerLinesAssignments++;

      // dynamic widgets always accept the inner line value directly
      if (!config.realized || config.dynamic)
      {
         config.innerLines = value;
      }
      else 
      {
         config.innerLines = Math.min(value, 
               config.innerLinesLimit - config.innerLinesAssignments);
      }
      pushScreenDefinition();
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInnerLines(NumberType lines)
   {
      if (lines == null || lines.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute INNER-LINES on %s %s", 
               type(), widgetName()), false);
         return;
      }
   
      setInnerLines(lines.intValue());
   }

   /**
    * Gets the HEIGHT-CHARS writable attribute.
    *
    * @return   The current value of the HEIGHT-CHARS attribute.
    */
   @Override
   public decimal getHeightChars()
   {
      if (!canAccess("HEIGHT"))
      {
         return new decimal(1);
      }
      return frame == null ? new decimal(1) : super.getHeightChars();
   }

   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDisableAutoZap(logical disableAutoZap)
   {
      notSettable("DISABLE-AUTO-ZAP");
   }

   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDisableAutoZap(boolean disableAutoZap)
   {
      notSettable("DISABLE-AUTO-ZAP");
   }

   /**
    * Get value of the DISABLE-AUTO-ZAP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isDisableAutoZap()
   {
      notQueryable("DISABLE-AUTO-ZAP");
      return new logical();
   }

   /**
    * Set AUTO-ZAP attribute.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setAutoZap(logical value)
   {
      notSettable("AUTO-ZAP");
   }

   /**
    * Set AUTO-ZAP attribute.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setAutoZap(boolean value)
   {
      notSettable("AUTO-ZAP");
   }

   /**
    * Get value of the AUTO-ZAP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isAutoZap()
   {
      notQueryable("AUTO-ZAP");
      return new logical();
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute 
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(double offset)
   {
      notSettable("CURSOR-OFFSET");
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute 
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(NumberType offset)
   {
      notSettable("CURSOR-OFFSET");
   }

   /**
    * Obtains the index position of the cursor in a text editing widget. This corresponds to
    * the CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @return   The character index at which the cursor is placed.
    */
   @Override
   public integer getCursorOffset()
   {
      notQueryable("CURSOR-OFFSET");
      return new integer();
   }

   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   @Override
   public void setHeightChars(NumberType height)
   {
      if (_isRealized())
      {
         ErrorManager.recordOrShowError(4078, 
            String.format("Unable to set attribute HEIGHT for %s %s",
                          type(), widgetName()), false);
      }
      else
      {
         super.setHeightChars(height);
      }
   }
   
   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   @Override
   public void setHeightChars(double height)
   {
      if (_isRealized())
      {
         ErrorManager.recordOrShowError(4078, 
            String.format("Unable to set attribute HEIGHT for %s %s",
                          type(), widgetName()), false);
      }
      else
      {
         super.setHeightChars(height);
      }
   }
   
   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   @Override
   public void setHeightPixels(NumberType heightPixels)
   {
      if (_isRealized())
      {
         ErrorManager.recordOrShowError(4078, 
            String.format("Unable to set attribute HEIGHT-PIXELS for %s %s",
                          type(), widgetName()), false);
      }
      else
      {
         super.setHeightPixels(heightPixels);
      }
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   @Override
   public void setHeightPixels(int heightPixels)
   {
      if (_isRealized())
      {
         ErrorManager.recordOrShowError(4078, 
            String.format("Unable to set attribute HEIGHT-PIXELS for %s %s",
                          type(), widgetName()), false);
      }
      else
      {
         super.setHeightPixels(heightPixels);
      }
   }

   /**
    * Set value of the MAX-CHARS attribute.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   @Override
   public void setMaxChars(double maxChars)
   {
      setMaxChars(new decimal(maxChars));
   }

   /**
    * Set value of the MAX-CHARS attribute.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   @Override
   public void setMaxChars(NumberType maxChars)
   {
      if (maxChars.isUnknown())
         return;
      
      short mc = (short)(maxChars.intValue() & 0x0000FFFF);
      if (config.maxChars == mc)
      {
         return;
      }

      config.maxChars = mc;
      pushScreenDefinition();
   }

   /**
    * Get value of the MAX-CHARS attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public integer getMaxChars()
   {
      return new integer(config.maxChars);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(character itemList)
   {
      return delete(itemList, COMPARE_VALUE_IGNORE_CASE, true);
   }

   /**
    * Finish initialization of the combo box widget. 
    */
   @Override
   public void finishSetup()
   {
      if (frame == null)
         return;
 
      FrameConfig frameConfig = frame.getFrameWidget().config();
      config.innerLinesLimit = LogicalTerminal.getScreenLines().intValue() - 
            ((int)frameConfig.row + 1) + 2;
      if (frameConfig.box)
         config.innerLinesLimit--;

      /*
       *  This value of innerLinesLimitDelta is valid in the case when there is only
       *  one frame F containing the widget W.
       *  When there are n frames
       *  Fn -> ... -> F -> W
       *  then the value of innerLinesLimitDelta is
       *  SUM(Fi:ROW + (if Fi:BOX then 1 else 0) - 1 | i = 1,...,n) + W:ROW.
       *  
       *  At this moment nested frames are not supported.   
       */
      config.innerLinesLimitDelta = (int)config.row;
      
      // inner line was not set, compute from the item list
      if (config.innerLines == 0)
      {
         config.innerLines = config.items == null ? 0 : config.items.length;
      }
      // if the are restrictions from up find it
      if (config.innerLinesLimit > 0 && config.innerLines > 0)
      {
         config.innerLines = Math.min(config.innerLines, config.innerLinesLimit);
      }
      // inner lines value can not be less than 3 anyway
      if (config.innerLines < 3)
      {
         config.innerLines = 3;
      }
      
      config.innerLinesAssignments = 0;
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   public logical delete(String itemList)
   {
      return delete(itemList, COMPARE_VALUE_IGNORE_CASE, true);
   }
   
   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   @Override
   public void setFrame(FrameWidget frame)
   {
      super.setFrame(frame);
      if (config.innerLines < 0)
      {
         config.innerLines = 2; 
      }
      config.heightChars = 1;
   }

   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @Override
   public character getScreenValue()
   {
      // in GUI and SIMPLE and DROP-DOWN modes we need to ignore format string
      if (!LogicalTerminal.isChui() && config.mode != ComboBoxConfig.Mode.DROP_DOWN_LIST)
      {
         return frame.getScreenValue(this, true);
      }
      else
      {
         return super.getScreenValue();
      }
   }
   
   /**
    * Remove trailing whitespaces
    * 
    * @param value
    *    the string to be truncated
    *    
    * @return the truncated argument
    */
   @Override
   protected String truncateValue(String value)
   {
      return RTRIM.matcher(value).replaceAll("");
   }

   /**
    * Remove trailing whitespaces if applicable
    * 
    * @param value
    *    the BaseDataType to be truncated
    *    
    * @return
    *    the truncated argument if the argument is of type character
    *    (nothing is required for non-character types)
    */
   @Override
   protected BaseDataType truncateValue(BaseDataType value)
   {
      if (value instanceof character)
      {
         return new character(truncateValue(value.toStringMessage()));
      }
      return value;
   }

   /**
    * Create new ControlSetItem
    * 
    * @param label
    *       item' label
    * @param value
    *       item's value
    *       
    * @return 
    *    new instance of the ControlSetItem
    */
   @Override
   protected ControlSetItem controlSetItem(character label, BaseDataType value) 
   {
      return new ControlSetItem(config.itemId++, 
            label.isUnknown() ? new character(" ") : (character)label.duplicate(), 
            truncateValue(value.duplicate())); 
   }
   
   /**
    * Truncate values of the provided items and remove invalid ones.
    * 
    * @param items
    *    items to be processed
    *    
    * @return
    *    valid items with truncated values
    */
   @Override
   protected ControlSetItem[] truncateValues(ControlSetItem[] items)
   {
      List<ControlSetItem> list = new ArrayList<>(items.length);

      // reset item ID counter, required to avoid wrong item numbers setting
      config.itemId = 1;
      for (ControlSetItem item : items)
      {
         if (invalid(item.getLabel(), item.getValue()))
         {
            continue;
         }
         list.add(controlSetItem(item.getLabel(), item.getValue()));
      }
      return list.toArray(new ControlSetItem[list.size()]);
   }
   
   /**
    * 
    * Checks if a value should be ignored (UNKNOWN for COMBO-BOX)
    *  
    * @param 
    *    value - value to be checked
    *    
    * @return <code>true</code> if the value should be ignored
    */
   @Override
   protected boolean ignore(BaseDataType value)
   {
      return value != null && "\"?\"".equals(value.toStringExport()); 
   }

   /**
    * Checks if a value should be ignored ("?" for the COMBO-BOX)
    *  
    * @param value
    *          value to be checked
    *    
    * @return <Code>true</code> if the value should be ignored
    */
   @Override
   protected boolean ignore(String value)
   {
      return "?".equals(value); 
   }

   /**
    * Internal worker for getting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   initialized
    *          Flag indicating if the widget is realized or not.
    * @param   bdt
    *          The current value in the frame's buffer.
    * @param   fmt
    *          The format to be applied, in case this is a formatted value.
    * @param   ignoreFormat
    *          Flag indicating if the format needs to be ignored.
    *          
    * @return  This widget's screen-value
    */
   @Override
   character getScreenValue(boolean initialized, BaseDataType bdt, String fmt, boolean ignoreFormat)
   {
      if (!initialized)
      {
         return new character();
      }
      
      return super.getScreenValue(initialized, bdt, fmt, ignoreFormat);
   }

   /**
    * Internal worker for setting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   frameBuf
    *          The frame buffer where to save the value.
    * @param   value
    *          The value to be set via SCREEN-VALUE attribute.
    * @param   inUIStmt
    *          Flag indicating this call originates from a UI statement.
    *          
    * @return  <code>true</code> if the caller can proceed, as the screen-value can be set.
    */
   @Override
   boolean setScreenValue(ScreenBuffer frameBuf, Object value, boolean inUIStmt)
   {
      if (frameBuf.getWidgetValue(getId()) != null)
      {
         if (value == null || (value instanceof BaseDataType && ((BaseDataType) value).isUnknown()))
         {
            return false;
         }
         
         if (value instanceof character && ((character) value).toJavaType().length() == 0)
         {
            return false;
         }
      }
      
      return super.setScreenValue(frameBuf, value, inUIStmt);
   }
}