/*
** Module   : CommonWidget.java
** Abstract : Define all widget-related APIs available via a widget handle instance.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20130112 Created initial version.
** 002 CS  20130204 Added conversion support for socket and server-socket. Moved sensitive 
**                  related methods from CommonWidget to the Sensitive interface.
** 003 CA  20130221 Refactored usage of attributes defined by the CommonHandleChain, CommonField
**                  and Sizeable interfaces.
** 004 EVL 20130222 Adding UI related attributes and methods.
** 005 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 006 CA  20130323 The PARENT attribute must be a CommonWidget, not Object.
**                  Added moveAfterTAb(handle) and moveBeforeTabItem(handle) APIs.
** 007 CA  20130323 The PARENT and FRAME attributes need to be set/get via handles.
** 008 OM  20130606 Added int64 support.
** 009 CA  20130927 Fixed getFirst/LastChild - they must return a handle. Removed unknown() API.
**                  Added setFrameHandle(CommonFrame).
** 010 EVL 20131002 Adding legacy attributes and methods annotation.
** 011 GES 20130925 Added CURSOR-OFFSET, FOCUSED-ROW-SELECTED and SIDE-LABEL-HANDLE attributes.
** 012 VIG 20140514 Changed moveBeforeTabItem method signature.
** 013 EVL 20140609 Added CHECKED attribute getter and setter.  Also new method added to check
**                  if the widget can be realized preRealizeCheck().
** 014 MAG 20140512 Added color attributes and changed some method names to match naming
**                  standards.
** 015 CA  20140701 Added support for font-related attributes: FONT, COLUMN-FONT, LABEL-FONT, 
**                  TITLE-FONT.
** 016 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 017 EVL 20141105 Adding rectangle related attriutes setters for: EDGE-CHARS, EDGE-PIXELS,
**                  FILLED, GRAPHIC-EDGE, GROUP-BOX, ROUNDED.
** 018 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 019 GES 20141124 Added many common widget attrs and methods.
** 020 VIG 20141210 Added setPopupMenu methods.
** 021 IAS 20150123 delete(int) replaced with delete(long).
** 022 IAS 20150210 Move methods specific for list widgets to a separate interface.
** 023 CA  20150323 Added API to set the format length from the widget's @ phrase(s).
** 024 IAS 20150325 Added support for the TEXT-SELECTED attribute.
** 025 CA  20150416 Moved FIRST/LAST-CHILD APIs to separate interface (CommonHandleTree).
** 026 VIG 20150414 Added setParentHandle for CommonWidget attribute type.
** 027 SVL 20150503 getBrowseColumn moved to BrowseWidget.
** 088 IAS 20150605 Removed support for the text selection, added stub implementation of the 
**                  realize() method.
** 089 CA  20150725 Added FRAME-COL, FRAME-ROW, FRAME-X, FRAME-Y support.
** 090 HC  20150803 Improved runtime support for WINDOW size attributes.
** 091 IAS 20150804 Added support for the WIDGET-ID, CONTEXT-HELP-ID, TAB-STOP, MANUAL-HIGHLIGHT
**                  and WINDOW attribute.
** 092 IAS 20150922 Added support for the SELECTABLE, MENU-MOUSE, MENU-KEY attributes.
** 093 OM  20150915 Added support for the LOAD-MOUSE-POINTER, VALIDATE method and RESIZABLE
**                  INPUT-VALUE, ENABLED and PREV-TAB-ITEM attributes.
** 094 IAS 20151015 Added setter for the THREE-D attribute
** 095 CA  20151027 Added support for MOVE-AFTER-TAB-ITEM(), TABLE and DBNAME.
** 096 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 097 IAS 20160229 MOUSE-POINTER attribute support.
** 098 CA  2016327  Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 099 GES 20160311 Removed getHeight() and getWidth() which were dead code.
** 100 GES 20160525 Reworked validation expression APIs to use lambdas.
** 101 GES 20160630 Javadoc improvement.
** 102 GES 20160909 Moved rectangle attribute setters to the RectangleInterface.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.function.*;

import com.goldencode.p2j.persist.P2JQuery;
import com.goldencode.p2j.util.*;

/**
 * Provide access to all widget-related APIs which can be accessed via a 
 * handle instance.
 */
public interface CommonWidget
extends CommonHandleChain,
        CommonField,
        DatabaseInfo,
        Sizeable,
        XYCoordinates
{
   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    event
    *           The name of the event.
    */
   public void apply(character event);

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    eventCode
    *           The event code.
    */
   public void apply(double eventCode);

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    eventCode
    *           The event code.
    */
   public void apply(int64 eventCode);

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    event
    *           The name of the event.
    */
   public void apply(String event);

   /**
    * Get this widget wrapped in a handle.
    * 
    * @return  handle to this frame.
    */
   public handle asWidgetHandle();

   /**
    * COLOR DISPLAY
    * 
    * @param   cs
    *          Color specification.
    */
   public void colorDisplay(ColorSpec cs);

   /**
    * COLOR PROMPT
    * 
    * @param   cs
    *          Color specification.
    */
   public void colorPrompt(ColorSpec cs);

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   @LegacyMethod(name = "DESELECT-ROWS")
   public logical deselectRows();

   /**
    * Disables this widget.
    * <p>
    * This method corresponds to the Progress statement: DISABLE widget
    */
   public void disable();

   /**
    * Enables this widget.
    * <p>
    * This method corresponds to the Progress statement: ENABLE widget
    */
   public void enable();

   /**
    * Get the ENABLE attribute of this widget.
    * 
    * @return  logical value of ENABLED attribute
    */
   @LegacyAttribute(name = "ENABLED")
   public logical isEnabled();

   /**
    * Get the PREV-TAB-ITEM attribute of this widget.
    * 
    * @return  logical value of PREV-TAB-ITEM attribute
    */
   @LegacyAttribute(name = "PREV-TAB-ITEM")
   public handle getPreviousTabItem();

   /**
    * Get the FONT attribute of this widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @LegacyAttribute(name = "FONT")
   public integer getFont();
   
   /**
    * Set the FONT attribute of this widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "FONT", setter = true)
   public void setFont(long fontNum);
   
   /**
    * Set the FONT attribute of this widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "FONT", setter = true)
   public void setFont(int64 fontNum);
   
   /**
    * Get the COLUMN-FONT attribute of this browse widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @LegacyAttribute(name = "COLUMN-FONT")
   public integer getColumnFont();
   
   /**
    * Set the COLUMN-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "COLUMN-FONT", setter = true)
   public void setColumnFont(long fontNum);
   
   /**
    * Set the COLUMN-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "COLUMN-FONT", setter = true)
   public void setColumnFont(int64 fontNum);
   
   /**
    * Get the LABEL-FONT attribute of this browse widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @LegacyAttribute(name = "LABEL-FONT")
   public integer getLabelFont();
   
   /**
    * Set the LABEL-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "LABEL-FONT", setter = true)
   public void setLabelFont(long fontNum);
   
   /**
    * Set the LABEL-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "LABEL-FONT", setter = true)
   public void setLabelFont(int64 fontNum);
   
   /**
    * Get the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @LegacyAttribute(name = "TITLE-FONT")
   public integer getTitleFont();
   
   /**
    * Set the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "TITLE-FONT", setter = true)
   public void setTitleFont(long fontNum);
   
   /**
    * Set the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @LegacyAttribute(name = "TITLE-FONT", setter = true)
   public void setTitleFont(int64 fontNum);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "COLUMN")
   public decimal getColumn();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "FRAME-COL")
   public decimal getFrameColumn();
   
   /**
    * Obtains the index position of the cursor in a text editing widget. This corresponds to
    * the CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @return   The character index at which the cursor is placed.
    */
   @LegacyAttribute(name = "CURSOR-OFFSET")
   public integer getCursorOffset();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "DCOLOR")
   public integer getDColor();

   /**
    * Gets the containing frame instance.
    * 
    * @return   The frame that contains this widget.
    */
   public GenericFrame getFrame();

   /**
    * Gets the FRAME attribute, which is the containing frame instance, as a handle.
    * 
    * @return   The frame that contains this widget.
    */
   @LegacyAttribute(name = "FRAME")
   public handle getFrameHandle();

   /**
    * Get the HELP writable attribute.
    * 
    * @return  current value of HELP attribute
    */
   @LegacyAttribute(name = "HELP")
   public character getHelp();

   /**
    * Gets the numeric ID of this widget (this is used as an index into the
    * screen-buffer among other things).
    * 
    * @return   The numeric ID of this widget.
    */
   public int getId();

   /**
    * Get the label text for this widget.
    *
    * @return   The label text.
    */
   public String getLabelStr();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getMaxHeightChars();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getMaxWidthChars();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "NEXT-TAB-ITEM")
   public CommonWidget getNextTabItem();

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "NUM-COLUMNS")
   public integer getNumColumns();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "NUM-SELECTED-ROWS")
   public integer getNumSelectedRows();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public CommonWidget getParent();

   /**
    * Gets the PARENT attribute, as a handle.
    * 
    * @return  the PARENT attribute.
    */
   @LegacyAttribute(name = "PARENT")
   public handle getParentHandle();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "PFCOLOR")
   public integer getPfColor();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public P2JQuery getQuery();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "ROW")
   public decimal getRow();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "FRAME-ROW")
   public decimal getFrameRow();
   
   /**
    * Obtains the text widget that is used as the side label for this widget. This is the
    * equivalent of the SIDE-LABEL-HANDLE attribute of COMBO-BOX, EDITOR, FILL-IN, RADIO-SET,
    * SELECTION-LIST, SLIDER and TEXT widgets.
    *
    * @return   The handle to the text widget that is the side label of this widget.
    */
   @LegacyAttribute(name = "SIDE-LABEL-HANDLE")
   public handle getSideLabelHandle();

   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @LegacyAttribute(name = "SCREEN-VALUE")
   public character getScreenValue();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "TITLE")
   public character getTitle();

   /**
    * Queries the TOOLTIP attribute of this widget.
    *
    * @return A help text message for a text field or text variable.
    */
   @LegacyAttribute(name = "TOOLTIP")
   public character getTooltip();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "FRAME-X")
   public integer getFrameX();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "FRAME-Y")
   public integer getFrameY();

   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    */
   public void hide();

   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    * 
    * @param   noPause
    *          <code>true</code> causes omission of any pauses which would
    *          be necessary otherwise.
    */
   public void hide(boolean noPause);

   /**
    * Gets the CHECKED writable attribute.
    *
    * @return    The current value of the CHECKED attribute.
    */
   @LegacyAttribute(name = "CHECKED")
   public logical isChecked();

   /**
    * Checks whether the widget is changed during the most recent operation.
    * 
    * @return   <code>true</code> if this widget has changed its value.
    */
   public logical isEntered();

   /**
    * Checks whether the row that has the focus is selected.  This is the equivalent of the
    * FOCUSED-ROW-SELECTED attribute of the browse widget.
    * 
    * @return   <code>true</code> if the focused row is selected.
    */
   @LegacyAttribute(name = "FOCUSED-ROW-SELECTED")
   public logical isFocusedRowSelected();

   /**
    * Gets the HIDDEN writable attribute.
    *
    * @return   The current value of the HIDDEN attribute.
    */
   @LegacyAttribute(name = "HIDDEN")
   public logical isHidden();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public logical isKeepFrameZOrder();

   /**
    * Queries if the SCREEN-VALUE attribute for this widget was changed.
    * 
    * @return  If <code>true</code>, the widget's screen value was changed.
    */
   @LegacyAttribute(name = "MODIFIED")
   public logical isModified();
   
   /**
    * Queries if the user is allowed to move this widget using direct manipulation.
    * 
    * @return  If <code>true</code>, the widget is movable via direct manipulation.
    */
   @LegacyAttribute(name = "MOVABLE")
   public logical isMovable();
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "MULTIPLE")
   public logical isMultiple();

   /**
    * Checks whether the widget is not changed during the most recent
    * operation.
    * 
    * @return   <code>true</code> if this widget has not changed its value.
    */
   public logical isNotEntered();

   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   @LegacyAttribute(name = "READ-ONLY")
   public logical isReadOnly();

   /**
    * Reports if this widget object is valid to realize.
    * 
    * @return   <code>true</code> if the widget can be relized.
    */
   public boolean preRealizeCheck();
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "SCROLLABLE")
   public logical isScrollable();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "SELECTED")
   public logical isSelected();

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   item
    *          Not used.
    * 
    * @return  nothing.
    */
   @LegacyMethod(name = "IS-SELECTED")
   public logical isSelected(character item);

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   index
    *          Not used.
    * 
    * @return  nothing.
    */
   @LegacyMethod(name = "IS-SELECTED")
   public logical isSelected(int64 index);

   /**
    * Reports if this widget object is valid for use.
    * <p>
    * The validity of the widget object is determined solely by the scope of
    * the containing frame.
    * 
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid();

   /**
    * Checks if the widget has a validation expression registered.
    * 
    * @return   <code>true</code> if the widget will have a validation expression executed
    *           during editing.
    */
   public boolean isValidatable();

   /**
    * Get the VISIBLE writable attribute.
    *
    * @return   The current value of VISIBLE attribute.
    */
   @LegacyAttribute(name = "VISIBLE")
   public logical isVisible();
   
   /**
    * Get the SELECTABLE writable attribute.
    *
    * @return   The current value of SELECTABLE attribute.
    */
   @LegacyAttribute(name = "SELECTABLE")
   public logical isSelectable();
   
   /**
    * Set the SELECTABLE writable attribute.
    *
    * @param value
    *        the new value of the attribute
    */
   @LegacyAttribute(name = "SELECTABLE", setter = true)
   public void setSelectable(logical value);
   
   /**
    * Get the SELECTABLE writable attribute.
    *
    * @param value
    *        the new value of the attribute
    */
   @LegacyAttribute(name = "SELECTABLE", setter = true)
   public void setSelectable(boolean value);
   
   /**
    * This method implements MOVE-AFTER-TAB function.
    *
    * @param   widget
    *          Widget reference.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "MOVE-AFTER-TAB-ITEM")
   public logical moveAfterTab(CommonWidget widget);

   /**
    * This method implements MOVE-AFTER-TAB function.
    *
    * @param   widget
    *          The handle of the widget after whose tab position this will be moved.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "MOVE-AFTER-TAB-ITEM")
   public logical moveAfterTab(handle widget);

   /**
    * This method implements MOVE-TO-BOTTOM function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "MOVE-TO-BOTTOM")
   public logical moveToBottom();

   /**
    * This method implements MOVE-TO-TOP function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "MOVE-TO-TOP")
   public logical moveToTop();

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   @LegacyMethod(name = "REFRESH")
   public logical refresh();

   /**
    * Restore configured colors.
    */
   public void restoreColors();

   /**
    * Save configured colors.
    */
   public void saveColors();

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   @LegacyMethod(name = "SELECT-ALL")
   public logical selectAll();

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   @LegacyMethod(name = "SELECT-FOCUSED-ROW")
   public logical selectFocusedRow();

   /**
    * Sets the AUTO-RETURN option.
    * 
    * @param    autoReturn
    *           <code>true</code> if the option is set to ON.
    */
   public void setAutoReturn(boolean autoReturn);

   /**
    * Set AUTO-RESIZE attribute value.
    * 
    * @param    autoResize
    *           New value for the attribute.
    */
   @LegacyAttribute(name = "AUTO-RESIZE", setter = true)
   public void setAutoResize(logical autoResize);

   /**
    * Set AUTO-RESIZE attribute value.
    * 
    * @param    autoResize
    *           New value for the attribute.
    */
   @LegacyAttribute(name = "AUTO-RESIZE", setter = true)
   public void setAutoResize(boolean autoResize);

   /**
    * Get AUTO-RESIZE attribute value.
    * 
    * @return
    *           The value of the attribute.
    */
   @LegacyAttribute(name = "AUTO-RESIZE")
   public logical isAutoResize();

   /**
    * Sets the BLANK option.
    * 
    * @param    blank
    *           <code>true</code> if the option is set to ON.
    */
   public void setBlank(boolean blank);

   /**
    * Sets the CHECKED writable attribute.
    *
    * @param    checked
    *           The new value for the CHECKED attribute.
    */
   @LegacyAttribute(name = "CHECKED", setter = true)
   public void setChecked(boolean checked);

   /**
    * Sets the CHECKED writable attribute.
    *
    * @param    checked
    *           The new value for the CHECKED attribute.
    */
   @LegacyAttribute(name = "CHECKED", setter = true)
   public void setChecked(logical checked);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   column
    *          Not used.
    */
   @LegacyAttribute(name = "COLUMN", setter = true)
   public void setColumn(double column);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   column
    *          Not used.
    */
   @LegacyAttribute(name = "COLUMN", setter = true)
   public void setColumn(NumberType column);

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @LegacyAttribute(name = "CURSOR-OFFSET", setter = true)
   public void setCursorOffset(double offset);

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @LegacyAttribute(name = "CURSOR-OFFSET", setter = true)
   public void setCursorOffset(NumberType offset);

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   @LegacyAttribute(name = "DCOLOR", setter = true)
   public void setDColor(Color displayColor);

   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   @LegacyAttribute(name = "DCOLOR", setter = true)
   public void setDColor(NumberType param);

   /**
    * Sets the DEBLANK option.
    * 
    * @param    deblank
    *           <code>true</code> if the option is set to ON.
    */
   public void setDeblank(boolean deblank);

   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   public void setDisableAutoZap(boolean disableAutoZap);

   /**
    * Sets the format specification.
    * 
    * @param    format
    *           The format specification.
    */
   public void setFormat(String format);

   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   public void setFrame(FrameWidget frame);

   /**
    * Sets the FRAME attribute, containing frame instance.
    * 
    * @param    h
    *           a handle containing the <code>GenericFrame</code> instance that contains
    *           this widget
    */
   @LegacyAttribute(name = "FRAME", setter = true)
   public void setFrameHandle(handle h);

   /**
    * Sets the FRAME attribute, containing frame instance.
    * 
    * @param    frame
    *           the frame instance that will contain this widget.
    */
   @LegacyAttribute(name = "FRAME", setter = true)
   public void setFrameHandle(CommonFrame frame);

   /**
    * Sets the FRAME attribute, containing frame instance.
    * 
    * @param    frame
    *           The frame instance that will contain this widget.
    */
   @LegacyAttribute(name = "FRAME", setter = true)
   public void setFrameHandle(FrameWidget frame);
   
   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   public void setFrame(GenericFrame frame);

   /**
    * Sets the containing frame instance.
    * 
    * @param    hFrame
    *           handle which is expected to contain a <code>FrameWidget</code>
    *           instance
    */
   public void setFrame(handle hFrame);

   /**
    * Establishes a relationship between the container and the elements in
    * a given widget group.
    *
    * @param    group
    *           The container of the widget.
    */
   public void setGroup(Group group);

   /**
    * Mark widget as a header.
    * 
    * @param    header
    *           New value of the flag.
    */
   public void setHeader(boolean header);

   /**
    * Sets the help text.
    * 
    * @param    help
    *           The new help text.
    */
   @LegacyAttribute(name = "HELP", setter = true)
   public void setHelp(character help);

   /**
    * Sets the help text.
    * 
    * @param    help
    *           The new help text.
    */
   @LegacyAttribute(name = "HELP", setter = true)
   public void setHelp(String help);

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @LegacyAttribute(name = "HIDDEN", setter = true)
   public void setHidden(boolean hidden);

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @LegacyAttribute(name = "HIDDEN", setter = true)
   public void setHidden(logical hidden);

   /**
    * Sets the numeric ID of this widget (this is used as an index into the
    * screen-buffer among other things).
    * 
    * @param    id
    *           The numeric ID of this widget.
    */
   public void setId(int id);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   k
    *          Not used.
    */
   public void setKeepFrameZOrder(boolean k);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   k
    *          Not used.
    */
   public void setKeepFrameZOrder(logical k);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   m
    *          Not used.
    */
   public void setMaxHeightChars(decimal m);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   m
    *          Not used.
    */
   public void setMaxWidthChars(decimal m);

   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   @LegacyAttribute(name = "MODIFIED", setter = true)
   public void setModified(boolean modified);
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   @LegacyAttribute(name = "MODIFIED", setter = true)
   public void setModified(logical modified);
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   @LegacyAttribute(name = "MOVABLE", setter = true)
   public void setMovable(boolean movable);
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   @LegacyAttribute(name = "MOVABLE", setter = true)
   public void setMovable(logical movable);
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   multiple
    *          Not used.
    */
   @LegacyAttribute(name = "MULTIPLE", setter = true)
   public void setMultiple(boolean multiple);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   multiple
    *          Not used.
    */
   @LegacyAttribute(name = "MULTIPLE", setter = true)
   public void setMultiple(logical multiple);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   n
    *          Not used.
    */
   public void setNextTabItem(GenericWidget<?> n);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   r
    *          Not used.
    */
   public void setNumSelectedRows(int64 r);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   p
    *          Not used.
    */
   public void setParent(GenericWidget<?> p);

   /**
    * Sets the PARENT attribute to the resource specified in the given handle.
    * 
    * @param   h
    *          The handle with the parent resource.
    */
   @LegacyAttribute(name = "PARENT", setter = true)
   public void setParentHandle(handle h);

   /**
    * Sets the PARENT attribute to the specified widget.
    * 
    * @param   w
    *          The parent widget
    */
   @LegacyAttribute(name = "PARENT", setter = true)
   public void setParentHandle(CommonWidget w);
   
   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    inputColor
    *           The color to use for when in "input" mode.
    */
   @LegacyAttribute(name = "PFCOLOR", setter = true)
   public void setPfColor(Color inputColor);

   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The color to use.
    */
   @LegacyAttribute(name = "PFCOLOR", setter = true)
   public void setPfColor(NumberType param);

   /**
    * Get widget POPUP-MENU.
    * 
    * @return  The pop-up menu handle.
    */
   @LegacyAttribute(name = "POPUP-MENU")
   public handle getPopupMenu();
   
   /**
    * Set widget POPUP-MENU.
    * 
    * @param  menu
    *         The pop-up menu handle.
    */
   @LegacyAttribute(name = "POPUP-MENU", setter = true)
   public void setPopupMenu(handle menu);
   
   /**
    * Set widget POPUP-MENU.
    * 
    * @param  menu
    *         The pop-up menu widget.
    */
   @LegacyAttribute(name = "POPUP-MENU", setter = true)
   public void setPopupMenu(MenuWidget menu);
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   q
    *          Not used.
    */
   public void setQuery(P2JQuery q);

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   @LegacyAttribute(name = "READ-ONLY", setter = true)
   public void setReadOnly(logical r);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   row
    *          Not used.
    */
   @LegacyAttribute(name = "ROW", setter = true)
   public void setRow(double row);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   row
    *          Not used.
    */
   @LegacyAttribute(name = "ROW", setter = true)
   public void setRow(NumberType row);

   /**
    * Sets the text widget that is used as the side label for this widget. This is the
    * equivalent of the SIDE-LABEL-HANDLE attribute of COMBO-BOX, EDITOR, FILL-IN, RADIO-SET,
    * SELECTION-LIST, SLIDER and TEXT widgets.
    *
    * @param    label
    *           The handle to the text widget to be used as the side label of this widget.
    */
   @LegacyAttribute(name = "SIDE-LABEL-HANDLE", setter = true)
   public void setSideLabelHandle(handle label);
   
   /**
    * Set the current value in the screen buffer of the backing data for
    * this widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   @LegacyAttribute(name = "SCREEN-VALUE", setter = true)
   public void setScreenValue(character value);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @LegacyAttribute(name = "SCROLLABLE", setter = true)
   public void setScrollable(boolean s);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @LegacyAttribute(name = "SCROLLABLE", setter = true)
   public void setScrollable(logical s);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @LegacyAttribute(name = "SELECTED", setter = true)
   public void setSelected(boolean s);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @LegacyAttribute(name = "SELECTED", setter = true)
   public void setSelected(logical s);

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   public void setSensitive(logical s);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   t
    *          Not used.
    */
   @LegacyAttribute(name = "TITLE", setter = true)
   public void setTitle(character t);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   t
    *          Not used.
    */
   @LegacyAttribute(name = "TITLE", setter = true)
   public void setTitle(String t);

   /**
    * Sets new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   @LegacyAttribute(name = "TOOLTIP", setter = true)
   public void setTooltip(character tooltip);

   /**
    * Sets new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   @LegacyAttribute(name = "TOOLTIP", setter = true)
   public void setTooltip(String tooltip);

   /**
    * Registers callbacks for validation processing (during editing) for any input into this
    * widget.
    * 
    * @param    valexp
    *           The <code>Function</code> to call to determine if the edits are allowed.
    * @param    valmsg
    *           The <code>Function</code> used to generate the error message on a validation
    *           failure. 
    */
   public void setValidation(ValidationExpr valexp, ValidationMsg valmsg);

   /**
    * Registers callbacks for validation processing (during editing) for any input into this
    * widget.  Use this version of the method when the message can be determined in advance.
    * 
    * @param    valexp
    *           The <code>Function</code> to call to determine if the edits are allowed.
    * @param    valmsg
    *           The error message on a validation failure.
    */
   public void setValidation(ValidationExpr valexp, String valmsg);

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   direction
    *          Not used.
    */
   public void setVertical(boolean direction);

   /**
    * Set the VISIBLE writable attribute.
    *
    * @param    visible
    *           The new value for the VISIBLE attribute.
    */
   @LegacyAttribute(name = "VISIBLE", setter = true)
   public void setVisible(boolean visible);

   /**
    * Set the VISIBLE writable attribute.
    *
    * @param    visible
    *           The new value for the VISIBLE attribute.
    */
   @LegacyAttribute(name = "VISIBLE", setter = true)
   public void setVisible(logical visible);

   /**
    * Underlines this widget on the screen.
    * <p>
    * This method corresponds to the Progress statement: UNDERLINE
    */
   public void underline();

   /**
    * Underlines this widget on the screen.
    * <p>
    * This method corresponds to the Progress statement: UNDERLINE
    *
    * @param    stream
    *           An explicit stream which will be used for the output.
    */
   public void underline(Stream stream);

   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @return   {@code true} if the validation is successful.
    */
   @LegacyMethod(name = "VALIDATE")
   public logical validateFields();
   
   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    val
    *           The only correct argument is "ENABLED-FIELDS" string. If specified, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   @LegacyMethod(name = "VALIDATE")
   public logical validateFields(character val);

   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    val
    *           The only correct argument is "ENABLED-FIELDS" string. If specified, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   @LegacyMethod(name = "VALIDATE")
   public logical validateFields(String val);

   /**
    * Shows this widget.
    * <p>
    * This method corresponds to the Progress statement: VIEW widget
    */
   public void view();

   /**
    * Gets the integer handle value of the window containing the widget.
    * 
    * @return   The integer handle value.
    */
   @LegacyAttribute(name = "HWND")
   public integer getHWND();

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @LegacyAttribute(name = "FIRST-TAB-ITEM")
   public handle getFirstTabItem();
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   h
    *          Not used.
    */
   @LegacyAttribute(name = "FIRST-TAB-ITEM", setter = true)
   public void setFirstTabItem(handle h);

   /**
    * Assigns the current widget to the tab position before the specified in the same field group.
    *
    * @param   widget
    *          The handle of the widget before whose we need to make new assignment.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "MOVE-BEFORE-TAB-ITEM")
   public logical moveBeforeTabItem(CommonWidget widget);

   /**
    * Assigns the current widget to the tab position before the specified in the same field group.
    *
    * @param   widget
    *          The handle of the widget before whose we need to make new assignment.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "MOVE-BEFORE-TAB-ITEM")
   public logical moveBeforeTabItem(handle widget);

   /**
    * Check if window has THREE-D option set.
    * 
    * @return <code>true</code> if window has THREE-D option set.
    */
   @LegacyAttribute(name = "THREE-D")
   public logical isThreeD();

   /**
    * Enable/disable 3D look.
    * 
    * @param   threeD
    *          New flag value.
    */
   @LegacyAttribute(name = "THREE-D", setter = true)
   public void setThreeD(boolean threeD);

   /**
    * Enable/disable 3D look.
    * 
    * @param   threeD
    *          New flag value.
    */
   @LegacyAttribute(name = "THREE-D", setter = true)
   public void setThreeD(logical threeD);

   /**
    * Get VIRTUAL-HEIGHT-CHARS attribute. 
    */
   @LegacyAttribute(name = "VIRTUAL-HEIGHT-CHARS")
   public decimal getVirtHeight();

   /**
    * Set VIRTUAL-HEIGHT-CHARS attribute. 
    *
    * @param   height
    *          VIRTUAL-HEIGHT-CHAR value
    */   
   @LegacyAttribute(name = "VIRTUAL-HEIGHT-CHARS", setter = true)
   public void setVirtHeight(NumberType height);
   
   /**
    * Get VIRTUAL-WIDTH-CHARS attribute. 
    */
   @LegacyAttribute(name = "VIRTUAL-WIDTH-CHARS")
   public decimal getVirtWidth();

   /**
    * Set VIRTUAL-WIDTH-CHARS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-CHAR value
    */
   @LegacyAttribute(name = "VIRTUAL-WIDTH-CHARS", setter = true)
   public void setVirtWidth(NumberType width);

   /**
    * Get VIRTUAL-HEIGHT-PIXELS attribute. 
    */
   @LegacyAttribute(name = "VIRTUAL-HEIGHT-PIXELS")
   public integer getVirtHeightPixels();

   /**
    * Set VIRTUAL-HEIGHT-PIXELS attribute. 
    *
    * @param   height
    *          VIRTUAL-HEIGHT-PIXELS value
    */   
   @LegacyAttribute(name = "VIRTUAL-HEIGHT-PIXELS", setter = true)
   public void setVirtHeightPixels(NumberType height);
   
   /**
    * Get VIRTUAL-WIDTH-PIXELS attribute. 
    */
   @LegacyAttribute(name = "VIRTUAL-WIDTH-PIXELS")
   public integer getVirtWidthPixels();

   /**
    * Set VIRTUAL-WIDTH-CHARS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-CHAR value
    */
   @LegacyAttribute(name = "VIRTUAL-WIDTH-PIXELS", setter = true)
   public void setVirtWidthPixels(NumberType width);

   /**
    * Get BGCOLOR attribute.
    *
    * @return  BGCOLOR value   
    */
   @LegacyAttribute(name = "BGCOLOR")
   public integer getBgColor();

   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the bGCOLOR attribute.
    */
   @LegacyAttribute(name = "BGCOLOR", setter = true)
   public void setBgColor(integer bgcolor);

   /**
    * Get FGCOLOR attribute.
    *
    * @return  FGCOLOR value   
    */
   @LegacyAttribute(name = "FGCOLOR")
   public integer getFgColor();

   /**
    * Sets the FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the FGCOLOR attribute.
    */
   @LegacyAttribute(name = "FGCOLOR", setter = true)
   public void setFgColor(integer fgcolor);

   /**
    * Sets the WIDGET-ID writable attribute.
    *
    * @param    id
    *           The new value for the WIDGET-ID attribute.
    */
   @LegacyAttribute(name = "WIDGET-ID", setter = true)
   public void setWidgetId(integer id);
   
   /**
    * Sets the WIDGET-ID writable attribute.
    *
    * @param    id
    *           The new value for the WIDGET-ID attribute.
    */
   @LegacyAttribute(name = "WIDGET-ID", setter = true)
   public void setWidgetId(int id);

   /**
    * Gets the WIDGET-ID writable attribute.
    *
    * @return
    *          the widget id
    */
   @LegacyAttribute(name = "WIDGET-ID")
   public integer getWidgetId();

   /**
    * Sets the CONTEXT-HELP-ID writable attribute.
    *
    * @param    id
    *           The new value for the CONTEXT-HELP-ID attribute.
    */
   @LegacyAttribute(name = "CONTEXT-HELP-ID", setter = true)
   public void setContextHelpId(integer id);
   
   /**
    * Sets the WIDGET-ID writable attribute.
    *
    * @param    id
    *           The new value for the CONTEXT-HELP-ID attribute.
    */
   @LegacyAttribute(name = "CONTEXT-HELP-ID", setter = true)
   public void setContextHelpId(int id);

   /**
    * Gets the CONTEXT-HELP-ID writable attribute.
    *
    * @return
    *          the widget context help id
    */
   @LegacyAttribute(name = "CONTEXT-HELP-ID")
   public integer getContextHelpId();

   /**
    * Set value of the TAB-STOP attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @LegacyAttribute(name = "TAB-STOP", setter = true)
   public void setTabStop(boolean value);

   /**
    * Set value of the TAB-STOP attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @LegacyAttribute(name = "TAB-STOP", setter = true)
   public void setTabStop(logical value);

   /**
    * Get value of the TAB-STOP attribute.
    *
    * @return   value of the attribute.
    */
   @LegacyAttribute(name = "TAB-STOP")
   public logical isTabStop();

   /**
    * Set value of the MANUAL-HIGHLIGHT attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @LegacyAttribute(name = "MANUAL-HIGHLIGHT", setter = true)
   public void setManualHighlight(boolean value);

   /**
    * Set value of the BLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @LegacyAttribute(name = "MANUAL-HIGHLIGHT", setter = true)
   public void setManualHighlight(logical value);

   /**
    * Get value of the MANUAL-HIGHLIGHT attribute.
    *
    * @return   value of the attribute.
    */
   @LegacyAttribute(name = "MANUAL-HIGHLIGHT")
   public logical isManualHighlight();
   
   /**
    * Gets the WINDOW attribute.
    *
    * @return  the value of the attribute
    */
   @LegacyAttribute(name = "WINDOW")
   public handle getWindow();
   
   /**
    * Sets the MENU-MOUSE writable attribute.
    *
    * @param    value
    *           The new value for the MENU-MOUSE attribute  (Left 1, Middle 2, Right 3).
    */
   @LegacyAttribute(name = "MENU-MOUSE", setter = true)
   public void setMenuMouse(int value);
   
   /**
    * Sets the MENU-MOUSE writable attribute.
    *
    * @param    value
    *           The new value for the MENU-MOUSE attribute ( (Left 1, Middle 2, Right 3).
    */
   @LegacyAttribute(name = "MENU-MOUSE", setter = true)
   public void setMenuMouse(int64 value);
   
   /**
    * Gets the MENU-MOUSE attribute.
    *
    * @return  The mouse button id currently associated (Left 1, Middle 2, Right 3, or unknown 
    *          if not set).
    */
   @LegacyAttribute(name = "MENU-MOUSE")
   public integer getMenuMouse();
   
   /**
    * Sets the MENU-KEY writable attribute.
    *
    * @param    value
    *           The new value for the MENU-KEY attribute.
    */
   @LegacyAttribute(name = "MENU-KEY", setter = true)
   public void setMenuKey(String value);
   
   /**
    * Sets the MENU-KEY writable attribute.
    *
    * @param    value
    *           The new value for the MENU-KEY attribute.
    */
   @LegacyAttribute(name = "MENU-KEY", setter = true)
   public void setMenuKey(character value);
   
   /**
    * Gets the MENU-KEY attribute.
    *
    * @return  the value of the attribute
    */
   @LegacyAttribute(name = "MENU-KEY")
   public character getMenuKey();
   
   /**
    * The getter side of RESIZABLE attribute.
    *
    * @return  {@code true} is widget is resizable.
    */
   @LegacyAttribute(name = "RESIZABLE")
   public logical isResizable();
   
   /**
    * The setter side of RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} is widget will be resizable.
    */
   @LegacyAttribute(name = "RESIZABLE", setter = true)
   public void setResizable(boolean on);
   
   /**
    * The setter side of RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} is widget will be resizable.
    */
   @LegacyAttribute(name = "RESIZABLE", setter = true)
   public void setResizable(logical on);
   
   /**
    * The implementation of LOAD-MOUSE-POINTER() method. Specifies the mouse pointer to display
    * when the pointer is moved over the widget.
    *
    * @param   pointerName
    *          The name of the mouse cursor. Here is the list of all cursor names:
    *        <table border='1' summary="">
    *          <tr><td>APPSTARTING</td><td>Arrow with an hourglass beside it</td></tr>
    *          <tr><td>ARROW</td><td>Standard arrow cursor</td></tr>
    *          <tr><td>CROSS</td><td>Cross hairs</td></tr>
    *          <tr><td>HELP</td><td>Arrow with a question mark beside it</td></tr>
    *          <tr><td>IBEAM</td><td>I-beam text cursor</td></tr>
    *          <tr><td>NO</td><td>Circle with a slash through it</td></tr>
    *          <tr><td>RECTANGLE (NT 3.51 only)</td><td>White rectangle</td></tr>
    *          <tr><td>SIZE</td><td>Sizing rectangle</td></tr>
    *          <tr><td>SIZE-E</td><td>Size to right</td></tr>
    *          <tr><td>SIZE-N</td><td>Size to top</td></tr>
    *          <tr><td>SIZE-NE</td><td>Size to top right</td></tr>
    *          <tr><td>SIZE-NW</td><td>Size to top left</td></tr>
    *          <tr><td>SIZE-S</td><td>Size to bottom</td></tr>
    *          <tr><td>SIZE-SE</td><td>Size to bottom right</td></tr>
    *          <tr><td>SIZE-SW</td><td>Size to bottom left</td></tr>
    *          <tr><td>SIZE-W</td><td>Size to left</td></tr>
    *          <tr><td>UPARROW</td><td>Up arrow</td></tr>
    *          <tr><td>WAIT</td><td>System busy</td></tr>
    *          <tr><td>GLOVE</td><td>Glove/finger</td></tr>
    *          <tr><td>COMPILER-WAIT</td><td>Compiler busy</td></tr>
    *        </table>
    * @return {@code true} if the mouse pointer is loaded successfully.
    */
   @LegacyMethod(name = "LOAD-MOUSE-POINTER")
   public logical loadMousePointer(String pointerName);
   
   /**
    * The implementation of LOAD-MOUSE-POINTER() method. Specifies the mouse pointer to display
    * when the pointer is moved over the widget.
    *
    * @param   pointerName
    *          The name of the mouse cursor. Here is the list of all cursor names:
    *        <table border='1' summary="">
    *          <tr><td>APPSTARTING</td><td>Arrow with an hourglass beside it</td></tr>
    *          <tr><td>ARROW</td><td>Standard arrow cursor</td></tr>
    *          <tr><td>CROSS</td><td>Cross hairs</td></tr>
    *          <tr><td>HELP</td><td>Arrow with a question mark beside it</td></tr>
    *          <tr><td>IBEAM</td><td>I-beam text cursor</td></tr>
    *          <tr><td>NO</td><td>Circle with a slash through it</td></tr>
    *          <tr><td>RECTANGLE (NT 3.51 only)</td><td>White rectangle</td></tr>
    *          <tr><td>SIZE</td><td>Sizing rectangle</td></tr>
    *          <tr><td>SIZE-E</td><td>Size to right</td></tr>
    *          <tr><td>SIZE-N</td><td>Size to top</td></tr>
    *          <tr><td>SIZE-NE</td><td>Size to top right</td></tr>
    *          <tr><td>SIZE-NW</td><td>Size to top left</td></tr>
    *          <tr><td>SIZE-S</td><td>Size to bottom</td></tr>
    *          <tr><td>SIZE-SE</td><td>Size to bottom right</td></tr>
    *          <tr><td>SIZE-SW</td><td>Size to bottom left</td></tr>
    *          <tr><td>SIZE-W</td><td>Size to left</td></tr>
    *          <tr><td>UPARROW</td><td>Up arrow</td></tr>
    *          <tr><td>WAIT</td><td>System busy</td></tr>
    *          <tr><td>GLOVE</td><td>Glove/finger</td></tr>
    *          <tr><td>COMPILER-WAIT</td><td>Compiler busy</td></tr>
    *        </table>
    * @return {@code true} if the mouse pointer is loaded successfully.
    */
   @LegacyMethod(name = "LOAD-MOUSE-POINTER")
   public logical loadMousePointer(character pointerName);
   
   /**
    * Reads the INPUT-VALUE attribute (value of the widget currently in the screen-buffer).
    *
    * @return   The input value of the widget of the same data type as the field or variable 
    *           associated with the widget.
    */
   @LegacyMethod(name = "INPUT-VALUE")
   public BaseDataType inputValue();
   
   /**
    * Get the value of the MOUSE-POINTER attribute for a widget
    *
    * @return   the value of the attribute 
    */
   @LegacyAttribute(name = "MOUSE-POINTER")
   public character getMousePointer();
   
   /**
    * Set the length of the format for the widget's @ phrases.
    * 
    * @param   length
    *          The maximum format length from the widget's @ phrases.
    */
   public void setAtFormatLength(int length);
   
   /**
    * Stub implementation of the 'realize' method.
    * 4GL can realize widgets on access to some of its attributes
    */
   public void realize();
}