/*
** Module   : ConfigHelper.java
** Abstract : Implements logic around config fields. 
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 HC  20150323 Created initial version
** 002 VIG 20150430 Changed setModified signature: WidgetConfig instead of BaseConfig.
** 003 HC  20150803 Improved runtime support for WINDOW size attributes.
** 004 CA  20151020 setModified sets the config.state flag only if the widget has changed.
** 005 HC  20160320 Improved frame virtual size handling.
** 006 HC  20160922 Fixed positioning of nested frames.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;

/**
 * The class contains methods implementing elementary logic around the
 * {@link BaseConfig} fields.
 * 
 */
public class ConfigHelper
{
   /**
    * Returns the effective COLUMN value from the passed config instance.
    * The returned value represents the actual widget location as rendered
    * on the client.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    *          
    * @return  COLUMN character value. 
    */
   public static double getEffectiveColumn(BaseConfig cfg)
   {
      double col = cfg.clientColumn;
      
      if (col == BaseConfig.INV_COORD)
      {
         col = getAssignedColumn(cfg);
      }
      
      if (col == BaseConfig.INV_COORD)
      {
         col = 0;
      }
      
      return col;
   }
   
   /**
    * Returns the effective ROW value from the passed config instance.
    * The returned value represents the actual widget location as rendered
    * on the client.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    *          
    * @return  ROW character value. 
    */
   public static double getEffectiveRow(BaseConfig cfg)
   {
      double row = cfg.clientRow;
      
      if (row == BaseConfig.INV_COORD)
      {
         row = getAssignedRow(cfg);
      }
      
      if (row == BaseConfig.INV_COORD)
      {
         row = 0;
      }
      
      return row;
   }
   
   /**
    * The returned value represents the actual widget location as rendered
    * on the client.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    *          
    * @return  A valid {@link Point} reference. 
    */
   public static Point getEffectiveLocation(BaseConfig cfg)
   {
      return new Point(getEffectiveColumn(cfg), getEffectiveRow(cfg));
   }
   
   /**
    * Returns the effective X value from the passed config instance.
    * The returned value represents the actual widget location as rendered
    * on the client.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    *          
    * @return  X coordinate value. 
    */
   public static int getEffectiveX(BaseConfig cfg)
   {
      int x = cfg.clientX;
      
      if (x == BaseConfig.INV_COORD)
      {
         x = getAssignedX(cfg);
      }
      
      if (x == BaseConfig.INV_COORD)
      {
         x = 0;
      }
      
      return x;
   }
   
   /**
    * Returns the effective Y value from the passed config instance.
    * The returned value represents the actual widget location as rendered
    * on the client.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    *          
    * @return  Y coordinate value. 
    */
   public static int getEffectiveY(BaseConfig cfg)
   {
      int y = cfg.clientY;
      
      if (y == BaseConfig.INV_COORD)
      {
         y = getAssignedY(cfg);
      }
      
      if (y == BaseConfig.INV_COORD)
      {
         y = 0;
      }
      
      return y;
   }
   
   /**
    * The returned value represents the actual widget location as rendered
    * on the client.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    *          
    * @return  A valid {@link NativePoint} reference. 
    */
   public static NativePoint getEffectivePixelLocation(BaseConfig cfg)
   {
      return new NativePoint(getEffectiveX(cfg), getEffectiveY(cfg));
   }
   
   /**
    * Returns the initial COLUMN value from the passed config
    * instance. The returned value represents the value assigned in
    * frame definition.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  COLUMN character value.
    */
   public static double getInitialColumn(BaseConfig cfg)
   {
      if (!isLeftAligned(cfg))
      {
         return BaseConfig.INV_COORD;
      }
      
      return cfg.initColumn;
   }
   
   /**
    * Returns the initial ROW value from the passed config
    * instance. The returned value represents the value assigned in
    * frame definition.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  ROW character value.
    */
   public static double getInitialRow(BaseConfig cfg)
   {
      return cfg.initRow;
   }
   
   /**
    * The returned value represents the location value assigned in
    * frame definition.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  A valid {@link Point} reference.
    */
   public static Point getInitialLocation(BaseConfig cfg)
   {
      return new Point(getInitialColumn(cfg), getInitialRow(cfg));
   }
   
   /**
    * Returns the initial X value from the passed config instance. The
    * returned value represents the value assigned in frame
    * definition.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  X coordinate value.
    */
   public static int getInitialX(BaseConfig cfg)
   {
      if (!isLeftAligned(cfg))
      {
         return BaseConfig.INV_COORD;
      }
      
      return cfg.initX;
   }
   
   /**
    * Returns the initial Y value from the passed config instance. The
    * returned value represents the value assigned in frame
    * definition.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  Y coordinate value.
    */
   public static int getInitialY(BaseConfig cfg)
   {
      return cfg.initY;
   }
   
   /**
    * The returned value represents the location value assigned in
    * frame definition.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  A valid {@link NativePoint} reference.
    */
   public static NativePoint getInitialPixelLocation(BaseConfig cfg)
   {
      return new NativePoint(getInitialX(cfg), getInitialY(cfg));
   }
   
   /**
    * Returns the initial or set COLUMN value from the passed config
    * instance. The returned value represents the value assigned in
    * frame definition or as set by COLUMN attribute.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  COLUMN character value.
    */
   public static double getAssignedColumn(BaseConfig cfg)
   {
      if (!isLeftAligned(cfg))
      {
         return BaseConfig.INV_COORD;
      }
      
      return cfg.column != BaseConfig.INV_COORD ? cfg.column : cfg.initColumn;
   }
   
   /**
    * Returns the initial or set ROW value from the passed config
    * instance. The returned value represents the value assigned in
    * frame definition or as set by ROW attribute.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  ROW character value.
    */
   public static double getAssignedRow(BaseConfig cfg)
   {
      return cfg.row != BaseConfig.INV_COORD ? cfg.row : cfg.initRow;
   }
   
   /**
    * The returned value represents the location value assigned in
    * frame definition or as set by COLUMN/ROW attributes.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  A valid {@link Point} reference.
    */
   public static Point getAssignedLocation(BaseConfig cfg)
   {
      return new Point(getAssignedColumn(cfg), getAssignedRow(cfg));
   }
   
   /**
    * Returns the initial or set X value from the passed config
    * instance. The returned value represents the value assigned in
    * frame definition or as set by X attribute.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  X pixel value.
    */
   public static int getAssignedX(BaseConfig cfg)
   {
      if (!isLeftAligned(cfg))
      {
         return BaseConfig.INV_COORD;
      }
      
      return cfg.x != BaseConfig.INV_COORD ? cfg.x : cfg.initX;
   }
   
   /**
    * Returns the initial or set Y value from the passed config
    * instance. The returned value represents the value assigned in
    * frame definition or as set by Y attribute.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  Y pixel value.
    */
   public static int getAssignedY(BaseConfig cfg)
   {
      return cfg.y != BaseConfig.INV_COORD ? cfg.y : cfg.initY;
   }
   
   /**
    * The returned value represents the location value assigned in
    * frame definition or as set by X/Y attributes.
    * 
    * @param   cfg
    *          A valid {@link BaseConfig} reference.
    * 
    * @return  A valid {@link NativePoint} reference.
    */
   public static NativePoint getAssignedPixelLocation(BaseConfig cfg)
   {
      return new NativePoint(getAssignedX(cfg), getAssignedY(cfg));
   }
   
   /**
    * Check if widget is left aligned.
    * 
    * @return  <code>true</code> if widget is left aligned.
    */
   public static boolean isLeftAligned(BaseConfig cfg)
   {
      return cfg.align == ControlEntity.ALIGN_LEFT;
   }
   
   /**
    * Get COLON attribute value from the passed configuration. Note 
    * that Progress does not have corresponding attribute accessible 
    * to program.
    * 
    * @param    cfg
    *           A valid {@link BaseConfig} instance.
    * 
    * @return   Current attribute value.
    */
   public static double getTo(BaseConfig cfg)
   {
      if (cfg.align == ControlEntity.ALIGN_RIGHT)
      {
         return cfg.column != BaseConfig.INV_COORD ? cfg.column : cfg.initColumn;
      }
      
      return BaseConfig.INV_COORD;
   }

   /**
    * Sets the rightmost column of the widget in the passed configuration.
    * This method must be called only on the server.
    * 
    * @param    cfg
    *           A valid {@link BaseConfig} instance.
    * @param    value
    *           The 0-based column number.
    */
   public static void setTo(BaseConfig cfg, double value)
   {
      value = Coordinate.scale(value);
      
      if (value == BaseConfig.INV_COORD)
      {
         return;
      }
      
      cfg.initColumn = value;
      cfg.align = ControlEntity.ALIGN_RIGHT;
   }

   /**
    * Get COLON attribute value from the passed configuration. Note
    * that Progress does not have corresponding attribute accessible
    * to program.
    * 
    * @param    cfg
    *           A valid {@link BaseConfig} instance.
    * 
    * @return   Current attribute value.
    */
   public static double getColon(BaseConfig cfg)
   {
      if (cfg.align == ControlEntity.ALIGN_COLON)
      {
         return cfg.column != BaseConfig.INV_COORD ? cfg.column : cfg.initColumn;
      }
      
      return BaseConfig.INV_COORD;
   }

   /**
    * Set COLON attribute value in the passed configuration. Note that
    * Progress does not have corresponding attribute accessible to
    * program. This method must be called only on the server. 
    * 
    * @param    cfg
    *           A valid {@link BaseConfig} instance.
    * 
    * @param    colon
    *           New value for the attribute.
    */
   public static void setColon(BaseConfig cfg, double colon)
   {
      colon = Coordinate.scale(colon);
      
      if (colon == BaseConfig.INV_COORD)
         return;
      
      cfg.initColumn = colon;
      cfg.align = ControlEntity.ALIGN_COLON;
   }

   /**
    * Set new value to the MODIFIED attribute in the passed
    * configuration.
    * 
    * @param    cfg
    *           A valid {@link BaseConfig} instance.
    * @param    modified
    *           New value for the attribute.
    */
   public static void setModified(WidgetConfig cfg, boolean modified)
   {
      cfg.modified = modified;
      
      // TODO: should this always get unilaterally set or should this only get set when
      //       modified changes from false to true?
      if (modified)
      {
         cfg.state = ScreenBuffer.CHANGED;
      }
   }
   
   /**
    * An accessor method to get the value of MAX-WIDTH-CHARS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static double getMaxWidthChars(WindowConfig cfg)
   {
      if (cfg.maxWidthChars == BaseConfig.INV_COORD)
      {
         return cfg.widthChars;
      }
      
      return cfg.maxWidthChars;
   }
   
   /**
    * An accessor method to get the value of MAX-HEIGHT-CHARS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static double getMaxHeightChars(WindowConfig cfg)
   {
      if (cfg.maxHeightChars == BaseConfig.INV_COORD)
      {
         return cfg.heightChars;
      }
      
      return cfg.maxHeightChars;
   }
   
   /**
    * An accessor method to get the value of MAX-WIDTH-PIXELS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static int getMaxWidthPixels(WindowConfig cfg)
   {
      if (cfg.maxWidthPixels == BaseConfig.INV_COORD)
      {
         return cfg.widthPixels;
      }
      
      return cfg.maxWidthPixels;
   }

   /**
    * An accessor method to get the value of MAX-HEIGHT-PIXELS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static int getMaxHeightPixels(WindowConfig cfg)
   {
      if (cfg.maxHeightPixels == BaseConfig.INV_COORD)
      {
         return cfg.heightPixels;
      }
      
      return cfg.maxHeightPixels;
   }

   /**
    * An accessor method to get the value of MIN-WIDTH-CHARS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static double getMinWidthChars(WindowConfig cfg)
   {
      if (cfg.minWidthChars == BaseConfig.INV_COORD)
      {
         return cfg.widthChars;
      }
      
      return cfg.minWidthChars;
   }
   
   /**
    * An accessor method to get the value of MIN-HEIGHT-CHARS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static double getMinHeightChars(WindowConfig cfg)
   {
      if (cfg.minHeightChars == BaseConfig.INV_COORD)
      {
         return cfg.heightChars;
      }
      
      return cfg.minHeightChars;
   }
   
   /**
    * An accessor method to get the value of MIN-WIDTH-PIXELS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static int getMinWidthPixels(WindowConfig cfg)
   {
      if (cfg.minWidthPixels == BaseConfig.INV_COORD)
      {
         return cfg.widthPixels;
      }
      
      return cfg.minWidthPixels;
   }

   /**
    * An accessor method to get the value of MIN-HEIGHT-PIXELS attribute from
    * the supplied config.
    * 
    * @param    cfg
    *           A reference to {@link WindowConfig}.
    * 
    * @return   See above.
    */
   public static int getMinHeightPixels(WindowConfig cfg)
   {
      if (cfg.minHeightPixels == BaseConfig.INV_COORD)
      {
         return cfg.heightPixels;
      }
      
      return cfg.minHeightPixels;
   }
   
   /**
    * Returns the effective frame virtual size from the supplied pane config. 
    * 
    * @param    cfg
    *           A reference to {@link PaneConfig}.
    * 
    * @return   See above.
    */
   public static double getEffectiveVirtualWidth(PaneConfig cfg)
   {
      return cfg.virtualWidthChars == BaseConfig.INV_COORD ? cfg.widthChars : 
                                                             cfg.virtualWidthChars;
   }
   
   /**
    * Returns the effective frame virtual size from the supplied pane config. 
    * 
    * @param    cfg
    *           A reference to {@link PaneConfig}.
    * 
    * @return   See above.
    */
   public static double getEffectiveVirtualHeight(PaneConfig cfg)
   {
      return cfg.virtualHeightChars == BaseConfig.INV_COORD ? cfg.heightChars : 
                                                              cfg.virtualHeightChars;
   }
   
   /**
    * Returns the effective frame virtual size from the supplied pane config. 
    * 
    * @param    cfg
    *           A reference to {@link PaneConfig}.
    * 
    * @return   See above.
    */
   public static int getEffectiveVirtualWidthPixels(PaneConfig cfg)
   {
      return cfg.virtualWidthPixels == BaseConfig.INV_COORD ? cfg.widthPixels : 
                                                              cfg.virtualWidthPixels;
   }
   
   /**
    * Returns the effective frame virtual size from the supplied pane config. 
    * 
    * @param    cfg
    *           A reference to {@link PaneConfig}.
    * 
    * @return   See above.
    */
   public static int getEffectiveVirtualHeightPixels(PaneConfig cfg)
   {
      return cfg.virtualHeightPixels == BaseConfig.INV_COORD ? cfg.heightPixels : 
                                                               cfg.virtualHeightPixels;
   }
}