/*
** Module   : CoordinatesConversion.java
** Abstract : Provides character-native and character-pixel conversion support.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20140911 Created initial version.
** 002 HC  20141021 Implemented window move/position support.
** 003 HC  20150405 Method parameters representing character coordinates are now scaled.
** 004 CA  20150429 Added baseUnits.  Removed {column|Row}{To|From}Native APIs which receive a 
**                  font as parameter, as the coordinates in GUI are always dependent on the 
**                  DEFAULT-FONT, not on a specific font in use by the window or widget.
** 005 HC  20150803 Added support for multiple rounding modes for coordinates conversion.
** 006 HC  20150831 Added conversion methods for Insets.
** 007 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import java.math.BigDecimal;

import com.goldencode.p2j.ui.Coordinate.RoundingMode;
import com.goldencode.p2j.ui.client.*;

/**
 * The class implements coordinates conversion, that is conversion of character coordinate 
 * values to native coordinate values and vice versa.
 * <p>
 * The conversion is implemented with the help of conversion parameters which are injected as
 * {@link CoordinatesConversionParams} instance during class instantiation.
 * <p>
 * Note that the class considers the character coordinate values as 0-based. This is contrary
 * to the 1-based nature of ROW/COLUMN 4GL attributes.
 */
public class CoordinatesConversion
{
   /**
    * The static instance suitable for ChUI environments.
    */
   private static final CoordinatesConversion CHUI = 
         new CoordinatesConversion(new CoordinatesConversionParams()
   {
      @Override
      public boolean isChui()
      {
         return true;
      }

      @Override
      public int getPixelsPerColumn()
      {
         return CHUI_PIXELS_PER_COLUMN;
      }

      @Override
      public int getPixelsPerRow()
      {
         return CHUI_PIXELS_PER_ROW;
      }

      @Override
      public int getFontWidth(WidgetId windowId, int font)
      {
         return CHUI_PIXELS_PER_COLUMN;
      }

      @Override
      public int getFontHeight(WidgetId windowId, int font)
      {
         return CHUI_PIXELS_PER_ROW;
      }
   });

   /**
    * The instance responsible for providing the conversion parameters.
    */
   private CoordinatesConversionParams params;
   
   /**
    * Constructs new instance of this class.
    *  
    * @param   params
    *          A valid reference to {@link CoordinatesConversionParams} instance.
    */
   public CoordinatesConversion(CoordinatesConversionParams params)
   {
      super();
      this.params = params;
   } 

   /**
    * Returns the singleton instance suitable for Chui environments.
    * 
    * @return  The valid instance of this class.  
    */
   public static CoordinatesConversion chuiInstance()
   {
      return CHUI;
   }

   /**
    * Converts the passed-in instance from character to native units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font.
    *  
    * @param   point
    *          A {@link Point} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    */
   public NativePoint pointToNative(Point point)
   {
      int x = columnToNative(point.x);
      int y = rowToNative(point.y);
      return new NativePoint(x, y);
   }

   /**
    * Determine the base units for the current client mode.
    * <p>
    * In ChUI, the base units is 1 character for row or column coordinates.
    * <p>
    * In GUI, the base units is computed by converting 1 character to 1 pixel, depending on the 
    * PIXELS-PER-COLUMN and PIXELS-PER-ROW values.
    *  
    * @return    See above.
    */
   public Point baseUnits()
   {
      if (params.isChui())
      {
         return new Point(1, 1);
      }
      
      return pointFromNative(new NativePoint(1, 1));
   }

   /**
    * Converts the passed-in instance from native to character units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font.
    * 
    * @param   point
    *          A {@link NativePoint} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    */
   public Point pointFromNative(NativePoint point)
   {
      double x = columnFromNative(point.x);
      double y = rowFromNative(point.y);
      return new Point(x, y);
   }
   
   /**
    * Converts the passed-in instance from character to native units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font.
    * 
    * @param   dimension
    *          A {@link Dimension} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    */
   public NativeDimension dimensionToNative(Dimension dimension)
   {
      int width = widthToNative(dimension.width);
      int height = heightToNative(dimension.height);
      return new NativeDimension(width, height);
   }
   
   /**
    * Converts the passed-in instance from native to character units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font.
    * 
    * @param   dimension
    *          A {@link NativeDimension} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    *          
    */
   public Dimension dimensionFromNative(NativeDimension dimension)
   {
      double width = widthFromNative(dimension.width);
      double height = heightFromNative(dimension.height);
      return new Dimension(width, height);
   }
   
   /**
    * Converts the passed-in instance from character to native units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font. 
    * 
    * @param   rect
    *          A {@link Rectangle} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    *          
    */
   public NativeRectangle rectangleToNative(Rectangle rect)
   {
      int top = rowToNative(rect.top());
      int left = columnToNative(rect.left());
      int bottom = rowToNative(rect.bottom());
      int right = columnToNative(rect.right());
      return new NativeRectangle(top, left, bottom, right);
   }
   
   /**
    * Converts the passed-in instance from native to character units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font.
    * 
    * @param   rect
    *          A {@link NativeRectangle} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    *          
    */
   public Rectangle rectangleFromNative(NativeRectangle rect)
   {
      double top = rowFromNative(rect.top());
      double left = columnFromNative(rect.left());
      double bottom = rowFromNative(rect.bottom());
      double right = columnFromNative(rect.right());
      return new Rectangle(top, left, bottom, right);
   }
   
   /**
    * Converts the passed-in instance from character to native units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font. 
    * 
    * @param   insets
    *          An {@link Insets} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    *          
    */
   public NativeInsets insetsToNative(Insets insets)
   {
      int top = rowToNative(insets.top);
      int left = columnToNative(insets.left);
      int bottom = rowToNative(insets.bottom);
      int right = columnToNative(insets.right);
      return new NativeInsets(top, left, bottom, right);
   }
   
   /**
    * Converts the passed-in instance from native to character units.
    * The method doesn't modify the passed-in instance.
    * The conversion is based on the font metrics defined by the current system font.
    * 
    * @param   insets
    *          An {@link NativeInsets} instance.
    * 
    * @return  An instance holding the result of the coordinate conversion. 
    *          
    */
   public Insets insetsFromNative(NativeInsets insets)
   {
      double top = rowFromNative(insets.top);
      double left = columnFromNative(insets.left);
      double bottom = rowFromNative(insets.bottom);
      double right = columnFromNative(insets.right);
      return new Insets(top, left, bottom, right);
   }
   
   /**
    * Converts a column ChUI coordinate value to native coordinate value.
    * The conversion is based on the font metrics defined by the current system font.
    * 
    * @param   column
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public int columnToNative(double column)
   {
      if (params.isChui())
      {
         return (int) Coordinate.scale(column);
      }

      return toPixels(column, params.getPixelsPerColumn());
   }
   
   /**
    * Converts a native X coordinate value to a column ChUI coordinate value.
    * The conversion is based on the font metrics defined by the current system font.
    *  
    * @param   x
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public double columnFromNative(int x)
   {
      if (params.isChui())
      {
         return x;
      }
      
      return toChar(x, params.getPixelsPerColumn());
   }
   
   /**
    * Converts a row ChUI coordinate value to native coordinate value.
    * The conversion is based on the font metrics defined by the passed-in 
    * <code>window</code> and <code>font</code> parameters.
    * 
    * @param   row
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public int rowToNative(double row)
   {
      if (params.isChui())
      {
         return (int) Coordinate.scale(row);
      }

      return toPixels(row, params.getPixelsPerRow());
   }
   
   /**
    * Converts a native Y coordinate value to a row ChUI coordinate value.
    * The conversion is based on the font metrics defined by the current system font.
    *  
    * @param   y
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public double rowFromNative(int y)
   {
      if (params.isChui())
      {
         return y;
      }

      return toChar(y, params.getPixelsPerRow());
   }
   
   /**
    * Converts ChUI width value to native width value.
    * The conversion is based on the font metrics defined by the current system font.
    *  
    * @param   width
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public int widthToNative(double width)
   {
      if (params.isChui())
      {
         return (int) Coordinate.scale(width);
      }

      return toPixels(width, params.getPixelsPerColumn());
   }
   
   /**
    * Converts ChUI width value to native width value.
    * The conversion is based on the font metrics defined by the passed-in 
    * <code>window</code> and <code>font</code> parameters.
    * 
    * @param   width
    *          The value being converted.
    * @param   windowId
    *          ID of the window representing the current UI environment.
    * @param   font
    *          The font number defined in the window represented by the <code>window</code> 
    *          parameter.
    * 
    * @return  The converted value. 
    */
   public int widthToNative(double width, WidgetId windowId, int font)
   {
      if (params.isChui())
      {
         return (int) Coordinate.scale(width);
      }

      return toPixels(width, params.getFontWidth(windowId, font));
   }

   /**
    * Converts a native width value to a ChUI width value.
    * The conversion is based on the font metrics defined by the current system font.
    *  
    * @param   width
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public double widthFromNative(int width)
   {
      if (params.isChui())
      {
         return width;
      }
      
      return toChar(width, params.getPixelsPerColumn());
   }

   /**
    * Converts a native width value to a ChUI width value.
    * The conversion is based on font metrics defined by the passed-in 
    * <code>window</code> and <code>font</code> parameters.
    * 
    * @param   width
    *          The value being converted.
    * @param   windowId
    *          ID of the window representing the current UI environment.
    * @param   font
    *          The font number defined in the window represented by the <code>window</code> 
    *          parameter.
    * 
    * @return  The converted value. 
    */
   public double widthFromNative(int width, WidgetId windowId, int font)
   {
      if (params.isChui())
      {
         return width;
      }
      
      return toChar(width, params.getFontWidth(windowId, font));
   }

   /**
    * Converts ChUI height value to native height value.
    * The conversion is based on the font metrics defined by the current system font.
    *  
    * @param   height
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public int heightToNative(double height)
   {
      if (params.isChui())
      {
         return (int) height;
      }

      return toPixels(height, params.getPixelsPerRow()); 
   }
   
   /**
    * Converts ChUI height value to native height value.
    * The conversion is based on the font metrics defined by the passed-in 
    * <code>window</code> and <code>font</code> parameters.
    * 
    * @param   height
    *          The value being converted.
    * @param   windowId
    *          ID of the window representing the current UI environment.
    * @param   font
    *          The font number defined in the window represented by the <code>window</code> 
    *          parameter.
    * 
    * @return  The converted value. 
    */
   public int heightToNative(double height, WidgetId windowId, int font)
   {
      if (params.isChui())
      {
         return (int) height;
      }
      
      return toPixels(height, params.getFontHeight(windowId, font));
   }

   /**
    * Converts a native height value to a ChUI height value.
    * The conversion is based on the font metrics defined by the current system font.
    *  
    * @param   height
    *          The value being converted.
    * 
    * @return  The converted value. 
    */
   public double heightFromNative(int height)
   {
      if (params.isChui())
      {
         return height;
      }
      
      return toChar(height, params.getPixelsPerRow());
   }
   
   /**
    * Converts a native height value to a ChUI height value.
    * The conversion is based on font metrics defined by the passed-in 
    * <code>window</code> and <code>font</code> parameters.
    * 
    * @param   height
    *          The value being converted.
    * @param   windowId
    *          ID of the window representing the current UI environment.
    * @param   font
    *          The font number defined in the window represented by the <code>window</code> 
    *          parameter.
    * 
    * @return  The converted value. 
    */
   public double heightFromNative(int height, WidgetId windowId, int font)
   {
      if (params.isChui())
      {
         return height;
      }
      
      return toChar(height, params.getFontHeight(windowId, font));
   }
   
   /**
    * Converts character columns to a value measured in pixel units.
    * The conversion is based on the value of 
    * {@link CoordinatesConversionParams#getPixelsPerColumn()}.
    * 
    * @param   columns
    *          The value being converted.
    * 
    * @return  The converted value in pixel units. 
    */
   public int columnToPixels(double columns)
   {
      return toPixels(Coordinate.scale(columns), params.getPixelsPerColumn());
   }
   
   /**
    * Converts a value in pixel units to character columns.
    * The conversion is based on the value of 
    * {@link CoordinatesConversionParams#getPixelsPerColumn()}.
    * 
    * @param   pixels
    *          The value in pixel units.
    * 
    * @return  The converted character columns. 
    */
   public double columnFromPixels(int pixels)
   {
      return toChar(pixels, params.getPixelsPerColumn());
   }
   
   /**
    * Converts a value in pixel units to character columns.
    * The conversion is based on the value of 
    * {@link CoordinatesConversionParams#getPixelsPerColumn()}.
    * <p>
    * See {@link #toChar(int, int)} for more information about rounding modes.
    * 
    * @param   pixels
    *          The value in pixel units.
    * @param   mode
    *          The desired rounding mode.
    * 
    * @return  The converted character columns. 
    */
   public double columnFromPixels(int pixels, RoundingMode mode)
   {
      return toChar(pixels, params.getPixelsPerColumn(), mode);
   }
   
   /**
    * Converts character rows to a value measured in pixel units.
    * The conversion is based on the value of 
    * {@link CoordinatesConversionParams#getPixelsPerRow()}.
    * 
    * @param   rows
    *          The value being converted.
    * 
    * @return  The converted value in pixel units. 
    */
   public int rowToPixels(double rows)
   {
      return toPixels(Coordinate.scale(rows), params.getPixelsPerRow());
   }
   
   /**
    * Converts a value in pixel units to character rows.
    * The conversion is based on the value of 
    * {@link CoordinatesConversionParams#getPixelsPerRow()}.
    * 
    * @param   pixels
    *          The value in pixel units.
    * 
    * @return  The converted character rows. 
    */
   public double rowFromPixels(int pixels)
   {
      return toChar(pixels, params.getPixelsPerRow());
   }
   
   /**
    * Converts a value in pixel units to character rows.
    * The conversion is based on the value of 
    * {@link CoordinatesConversionParams#getPixelsPerRow()}.
    * <p>
    * See {@link #toChar(int, int)} for more information about rounding modes.
    * 
    * @param   pixels
    *          The value in pixel units.
    * @param   mode
    *          The desired rounding mode.
    * 
    * @return  The converted character rows. 
    */
   public double rowFromPixels(int pixels, RoundingMode mode)
   {
      return toChar(pixels, params.getPixelsPerRow(), mode);
   }
   
   /**
    * Converts pixels to character units.
    * The method rounds the result half-up to two decimal places.  
    * 
    * @param   pixels
    *          The pixel value being converted.
    * @param   pixelsPerChar
    *          Number of pixels per one character unit. 
    * 
    * @return  The converted value.
    */
   private double toChar(int pixels, int pixelsPerChar)
   {
      return toChar(pixels, pixelsPerChar, RoundingMode.ROUND);
   }
   
   /**
    * Converts pixels to character units.
    * The method rounds the result half-up to two decimal places.
    * <p>
    * Two rounding modes are supported, {@link RoundingMode#ROUND}
    * and {@link RoundingMode#TRUNCATE}.
    * {@link RoundingMode#ROUND} performs two rounds operations. First 
    * a half-up round with a scale of three and second a half-up round with
    * a scale of two.
    * {@link RoundingMode#TRUNCATE} simply discards the decimal part and 
    * effectively makes the method return a whole number.
    * 
    * @param   pixels
    *          The pixel value being converted.
    * @param   pixelsPerChar
    *          Number of pixels per one character unit. 
    * @param   mode
    *          The desired rounding mode.
    * 
    * @return  The converted value.
    */
   private double toChar(int pixels, int pixelsPerChar, RoundingMode mode)
   {
      // 2 levels of rounding here, force rounding to 3 digits first
      BigDecimal rnd =  BigDecimal.valueOf((double)pixels/pixelsPerChar);
      
      if (mode == RoundingMode.ROUND)
      {
         rnd = rnd.setScale(3, BigDecimal.ROUND_HALF_UP);
         rnd = rnd.setScale(2, BigDecimal.ROUND_HALF_UP);
      }
      else if (mode == RoundingMode.TRUNCATE)
      {
         rnd = rnd.setScale(2, BigDecimal.ROUND_DOWN);
      }
      else
      {
         throw new RuntimeException("Unsupported rouding mode!");
      }
      
      // then round to 2 digits
      return rnd.doubleValue();
   }
   
   /**
    * Converts character units to pixels.
    * 
    * @param   chars
    *          The value in character units being converted.
    * @param   pixelsPerChar
    *          Number of pixels per one character unit. 
    * 
    * @return  The converted value.
    */
   private int toPixels(double chars, int pixelsPerChar)
   {
      // do not use Math.round, we need half-up rounding (even for negative values)
      return BigDecimal.valueOf(Coordinate.scale(chars)*pixelsPerChar)
            .setScale(0, BigDecimal.ROUND_HALF_UP)
            .intValue();
   }
}