/*
** Module   : WidgetFactory.java
** Abstract : Provides factory APIs to create widgets using the CREATE WIDGET statement. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130918 Added widget-pool support. The pool name must be validated before anything 
**                  else, and a null name means the default pool is used.
** 003 CA  20130927 Widget resource type are determined from LegacyResource constants.
** 004 CA  20131021 Mark the resources as being created dynamic.
** 005 CA  20131115 Changes related to properly instantiate dynamic resources.
** 006 VIG 20131127 Added method to check possibility to create new window
** 007 EVL 20140918 Adding support for dynamic widget.
** 008 CA  20140929 Replaced WidgetId.allocateId with the WidgetId.nextId API, which returns a
**                  WidgetId instance.
** 009 HC  20141021 Fixed window config ID assignment.
** 010 VIG 20141104 Added MENU widgets support.
** 011 EVL 20150112 Adding image creation support.
** 012 HC  20150323 WidgetId.getId() renamed to asInt().
** 013 CA  20150728 Adding screen definition pushing lock/unlock feature to defer data transfer
**                  to client side until all attributes of the dynamic widget are properly set.
** 014 SVL 20150804 Added BROWSE widget.
** 015 EVL 20160211 Adding SLIDER widget creation.
** 016 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 017 HC  20170131 Added runtime support for dynamic MENU* widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * Provides APIs to create the following widgets:
 * <table summary="">
 * <tr>
 *    <td>Widget Type</td>
 *    <td>API</td>
 * <tr>
 * <tr>
 *    <td>CREATE VALUE(...)</td>
 *    <td>{@link #create}</td>
 * </tr>
 * <tr>
 *    <td>CREATE BUTTON</td>
 *    <td>{@link #createButton}</td>
 * </tr>
 * <tr>
 *    <td>CREATE COMBO-BOX</td>
 *    <td>{@link #createComboBox}</td>
 * </tr>
 * <tr>
 *    <td>CREATE CONTROL-FRAME</td>
 *    <td>{@link #createControlFrame}</td>
 * </tr>
 * <tr>
 *    <td>CREATE DIALOG-BOX</td>
 *    <td>{@link #createDialogBox}</td>
 * </tr>
 * <tr>
 *    <td>CREATE EDITOR</td>
 *    <td>{@link #createEditor}</td>
 * </tr>
 * <tr>
 *    <td>CREATE FILL-IN</td>
 *    <td>{@link #createFillIn}</td>
 * </tr>
 * <tr>
 *    <td>CREATE FRAME</td>
 *    <td>{@link #createFrame}</td>
 * </tr>
 * <tr>
 *    <td>CREATE IMAGE</td>
 *    <td>{@link #createImage}</td>
 * </tr>
 * <tr>
 *    <td>CREATE MENU</td>
 *    <td>{@link #createMenu}</td>
 * </tr>
 * <tr>
 *    <td>CREATE MENU-ITEM</td>
 *    <td>{@link #createMenuItem}</td>
 * </tr>
 * <tr>
 *    <td>CREATE RADIO-SET</td>
 *    <td>{@link #createRadioSet}</td>
 * </tr>
 * <tr>
 *    <td>CREATE RECTANGLE</td>
 *    <td>{@link #createRectangle}</td>
 * </tr>
 * <tr>
 *    <td>CREATE SELECTION-LIST</td>
 *    <td>{@link #createSelectionList}</td>
 * </tr>
 * <tr>
 *    <td>CREATE SLIDER</td>
 *    <td>{@link #createSlider}</td>
 * </tr>
 * <tr>
 *    <td>CREATE SUB-MENU</td>
 *    <td>{@link #createSubMenu}</td>
 * </tr>
 * <tr>
 *    <td>CREATE TEXT</td>
 *    <td>{@link #createText}</td>
 * </tr>
 * <tr>
 *    <td>CREATE TOGGLE-BOX</td>
 *    <td>{@link #createToggleBox}</td>
 * </tr>
 * <tr>
 *    <td>CREATE WINDOW</td>
 *    <td>{@link #createWindow}</td>
 * </tr>
 * </table>
 * 
 * TODO: once the widget has been created, its attributes can be set regardless if is attached to
 * a frame or not. Currently, P2J allows the attributes to be set only if the widget is attached
 * to a frame, as the widget is not valid until is attached to a frame.
 */
public class DynamicWidgetFactory
{
   /** A map of widget names to their widget classes. */
   private static final Map<String, Class<? extends GenericWidget>> widgetFactory = 
      new HashMap<String, Class<? extends GenericWidget>>();
   
   static
   {
      widgetFactory.put(LegacyResource.BROWSE,         BrowseWidget.class);
      widgetFactory.put(LegacyResource.BUTTON,         ButtonWidget.class);
      widgetFactory.put(LegacyResource.COMBO_BOX,      ComboBoxWidget.class);
      widgetFactory.put(LegacyResource.CONTROL_FRAME,  null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.DIALOG_BOX,     null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.EDITOR,         EditorWidget.class);
      widgetFactory.put(LegacyResource.FILL_IN,        FillInWidget.class);
      widgetFactory.put(LegacyResource.FRAME,          FrameWidget.class);
      widgetFactory.put(LegacyResource.IMAGE,          ImageWidget.class);
      widgetFactory.put(LegacyResource.MENU,           MenuWidget.class); 
      widgetFactory.put(LegacyResource.MENU_ITEM,      MenuItemWidget.class); 
      widgetFactory.put(LegacyResource.RADIO_SET,      RadioSetWidget.class);
      widgetFactory.put(LegacyResource.RECTANGLE,      RectangleWidget.class);
      widgetFactory.put(LegacyResource.SELECTION_LIST, SelectionListWidget.class);
      widgetFactory.put(LegacyResource.SLIDER,         SliderWidget.class);
      widgetFactory.put(LegacyResource.SUB_MENU,       SubMenuWidget.class); 
      widgetFactory.put(LegacyResource.TEXT,           TextWidget.class); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.TOGGLE_BOX,     ToggleBoxWidget.class);
      widgetFactory.put(LegacyResource.WINDOW,         WindowWidget.class);
   }
   
   /**
    * Create a widget, using the given widget type. This must be a value from the following list
    * (case insensitive): BUTTON, COMBO-BOX, CONTROL-FRAME, DIALOG-BOX, EDITOR, FILL-IN, FRAME,
    * IMAGE, MENU, MENU-ITEM, RADIO-SET, RECTANGLE, SELECTION-LIST, SLIDER, SUB-MENU, TEXT,
    * TOGGLE-BOX, WINDOW.
    * <p>
    * If the <code>widgetType</code> is one of the above strings, the appropriate factory API to
    * create the widget will be called.
    * <p>
    * If available, the widget will be saved in the unnamed widget pool.
    *  
    * @param    widgetType
    *           The widget type. Must be not-null.
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void create(String widgetType, handle h)
   {
      create(widgetType, h, (character) null);
   }

   /**
    * Create a widget, using the given widget type. This must be a value from the following list
    * (case insensitive): BUTTON, COMBO-BOX, CONTROL-FRAME, DIALOG-BOX, EDITOR, FILL-IN, FRAME,
    * IMAGE, MENU, MENU-ITEM, RADIO-SET, RECTANGLE, SELECTION-LIST, SLIDER, SUB-MENU, TEXT,
    * TOGGLE-BOX, WINDOW.
    * <p>
    * If the <code>widgetType</code> is one of the above strings, the appropriate factory API to
    * create the widget will be called.
    * <p>
    * If available, the widget will be saved in the unnamed widget pool.
    *  
    * @param    widgetType
    *           The widget type. Must be not-null.
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void create(String widgetType, handle h, String widgetPool)
   {
      create(widgetType, h, (widgetPool == null ? null : new character(widgetPool)));
   }
   
   /**
    * Create a widget, using the given widget type. This must be a value from the following list
    * (case insensitive): BUTTON, COMBO-BOX, CONTROL-FRAME, DIALOG-BOX, EDITOR, FILL-IN, FRAME,
    * IMAGE, MENU, MENU-ITEM, RADIO-SET, RECTANGLE, SELECTION-LIST, SLIDER, SUB-MENU, TEXT,
    * TOGGLE-BOX, WINDOW.
    * <p>
    * If the <code>widgetType</code> is one of the above strings, the appropriate factory API to
    * create the widget will be called.
    * <p>
    * If available, the widget will be saved in the unnamed widget pool.
    *  
    * @param    widgetType
    *           The widget type. Must be not-null.
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void create(String widgetType, handle h, character widgetPool)
   {
      // validate widget pool before anything else!!!
      if (!WidgetPool.validWidgetPool(widgetPool))
      {
         return;
      }
      
      Class<? extends GenericWidget> widCls = widgetFactory.get(widgetType.toUpperCase());
      
      if (widCls == null)
      {
         String msg = String.format("Invalid widget type %s", widgetType);
         ErrorManager.recordOrThrowError(3180, msg, false);
         
         // if the widget type is not known, do not touch the handle.
         return;
      }

      if (LegacyResource.WINDOW.equals(widgetType) && !isAbleToCreateWindow())
      {
         if (!EnvironmentOps.isBatchMode().booleanValue())
         {
            h.assign(LogicalTerminal.currentWindow());
         }  
         return;
      }

      try
      {
         Constructor<? extends GenericWidget> ctor = widCls.getDeclaredConstructor(boolean.class);
         WrappedResource res = ctor.newInstance(true);

         h.assign(res);
         
         // assign ID for new widget
         // the ID may have been assigned during widget instance construction
         CommonWidget w = h.unwrapWidget();
         int id = w.getId();
         if (id == -1)
         {
            id = WidgetId.nextID().asInt();
            h.unwrapWidget().setId(id);
         }
         
         WidgetPool.addResource(res, widgetPool);
         LogicalTerminal.registerWidget((GenericWidget<?>) h.unwrapWidget(), id);
      }
      catch (Exception e)
      {
         // any exception here converts in a runtime exception
         throw new RuntimeException("Could not create widget of type " + widgetType, e);
      }
   }

   /**
    * Create a BUTTON widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createButton(handle h)
   {
      createButton(h, (character) null);
   }

   /**
    * Create a BUTTON widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createButton(handle h, String widgetPool)
   {
      create(LegacyResource.BUTTON, h, widgetPool);
   }

   /**
    * Create a BUTTON widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createButton(handle h, character widgetPool)
   {
      create(LegacyResource.BUTTON, h, widgetPool);
   }

   /**
    * Create a COMBO-BOX widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createComboBox(handle h)
   {
      createComboBox(h, (character) null);
   }

   /**
    * Create a COMBO-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createComboBox(handle h, String widgetPool)
   {
      create(LegacyResource.COMBO_BOX, h, widgetPool);
   }

   /**
    * Create a COMBO-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createComboBox(handle h, character widgetPool)
   {
      create(LegacyResource.COMBO_BOX, h, widgetPool);
   }

   /**
    * Create a CONTROL-FRAME widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createControlFrame(handle h)
   {
      createControlFrame(h, (character) null);
   }

   /**
    * Create a CONTROL-FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createControlFrame(handle h, String widgetPool)
   {
      create(LegacyResource.CONTROL_FRAME, h, widgetPool);
   }

   /**
    * Create a CONTROL-FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createControlFrame(handle h, character widgetPool)
   {
      create(LegacyResource.CONTROL_FRAME, h, widgetPool);
   }

   /**
    * Create a DIALOG-BOX widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createDialogBox(handle h)
   {
      createDialogBox(h, (character) null);
   }

   /**
    * Create a DIALOG-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createDialogBox(handle h, String widgetPool)
   {
      create(LegacyResource.DIALOG_BOX, h, widgetPool);
   }

   /**
    * Create a DIALOG-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createDialogBox(handle h, character widgetPool)
   {
      create(LegacyResource.DIALOG_BOX, h, widgetPool);
   }

   /**
    * Create a EDITOR widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createEditor(handle h)
   {
      createEditor(h, (character) null);
   }

   /**
    * Create a EDITOR widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createEditor(handle h, String widgetPool)
   {
      create(LegacyResource.EDITOR, h, widgetPool);
   }

   /**
    * Create a EDITOR widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createEditor(handle h, character widgetPool)
   {
      create(LegacyResource.EDITOR, h, widgetPool);
   }

   /**
    * Create a FILL-IN widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createFillIn(handle h)
   {
      createFillIn(h, (character) null);
   }

   /**
    * Create a FILL-IN widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFillIn(handle h, String widgetPool)
   {
      create(LegacyResource.FILL_IN, h, widgetPool);
   }

   /**
    * Create a FILL-IN widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFillIn(handle h, character widgetPool)
   {
      create(LegacyResource.FILL_IN, h, widgetPool);
   }

   /**
    * Create a FRAME widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createFrame(handle h)
   {
      createFrame(h, (character) null);
   }

   /**
    * Create a FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFrame(handle h, String widgetPool)
   {
      create(LegacyResource.FRAME, h, widgetPool);
   }

   /**
    * Create a FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFrame(handle h, character widgetPool)
   {
      create(LegacyResource.FRAME, h, widgetPool);
   }

   /**
    * Create a IMAGE widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createImage(handle h)
   {
      createImage(h, (character) null);
   }

   /**
    * Create a IMAGE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createImage(handle h, String widgetPool)
   {
      create(LegacyResource.IMAGE, h, widgetPool);
   }

   /**
    * Create a IMAGE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createImage(handle h, character widgetPool)
   {
      create(LegacyResource.IMAGE, h, widgetPool);
   }

   /**
    * Create a MENU widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createMenu(handle h)
   {
      createMenu(h, (character) null);
   }

   /**
    * Create a MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenu(handle h, String widgetPool)
   {
      create(LegacyResource.MENU, h, widgetPool);
   }

   /**
    * Create a MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenu(handle h, character widgetPool)
   {
      create(LegacyResource.MENU, h, widgetPool);
   }

   /**
    * Create a MENU-ITEM widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createMenuItem(handle h)
   {
      createMenuItem(h, (character) null);
   }

   /**
    * Create a MENU-ITEM widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenuItem(handle h, String widgetPool)
   {
      create(LegacyResource.MENU_ITEM, h, widgetPool);
   }

   /**
    * Create a MENU-ITEM widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenuItem(handle h, character widgetPool)
   {
      create(LegacyResource.MENU_ITEM, h, widgetPool);
   }

   /**
    * Create a RADIO-SET widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createRadioSet(handle h)
   {
      createRadioSet(h, (character) null);
   }

   /**
    * Create a RADIO-SET widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRadioSet(handle h, String widgetPool)
   {
      create(LegacyResource.RADIO_SET, h, widgetPool);
   }

   /**
    * Create a RADIO-SET widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRadioSet(handle h, character widgetPool)
   {
      create(LegacyResource.RADIO_SET, h, widgetPool);
   }

   /**
    * Create a RECTANGLE widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createRectangle(handle h)
   {
      createRectangle(h, (character) null);
   }

   /**
    * Create a RECTANGLE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRectangle(handle h, String widgetPool)
   {
      create(LegacyResource.RECTANGLE, h, widgetPool);
   }

   /**
    * Create a RECTANGLE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRectangle(handle h, character widgetPool)
   {
      create(LegacyResource.RECTANGLE, h, widgetPool);
   }

   /**
    * Create a SELECTION-LIST widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createSelectionList(handle h)
   {
      createSelectionList(h, (character) null);
   }

   /**
    * Create a SELECTION-LIST widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSelectionList(handle h, String widgetPool)
   {
      create(LegacyResource.SELECTION_LIST, h, widgetPool);
   }

   /**
    * Create a SELECTION-LIST widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSelectionList(handle h, character widgetPool)
   {
      create(LegacyResource.SELECTION_LIST, h, widgetPool);
   }

   /**
    * Create a SLIDER widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createSlider(handle h)
   {
      createSlider(h, (character) null);
   }

   /**
    * Create a SLIDER widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSlider(handle h, String widgetPool)
   {
      create(LegacyResource.SLIDER, h, widgetPool);
   }

   /**
    * Create a SLIDER widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSlider(handle h, character widgetPool)
   {
      create(LegacyResource.SLIDER, h, widgetPool);
   }

   /**
    * Create a SUB-MENU widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createSubMenu(handle h)
   {
      createSubMenu(h, (character) null);
   }

   /**
    * Create a SUB-MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSubMenu(handle h, String widgetPool)
   {
      create(LegacyResource.SUB_MENU, h, widgetPool);
   }

   /**
    * Create a SUB-MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSubMenu(handle h, character widgetPool)
   {
      create(LegacyResource.SUB_MENU, h, widgetPool);
   }

   /**
    * Create a TEXT widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createText(handle h)
   {
      createText(h, (character) null);
   }

   /**
    * Create a TEXT widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createText(handle h, String widgetPool)
   {
      create(LegacyResource.TEXT, h, widgetPool);
   }

   /**
    * Create a TEXT widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createText(handle h, character widgetPool)
   {
      create(LegacyResource.TEXT, h, widgetPool);
   }

   /**
    * Create a TOGGLE-BOX widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createToggleBox(handle h)
   {
      createToggleBox(h, (character) null);
   }

   /**
    * Create a TOGGLE-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createToggleBox(handle h, String widgetPool)
   {
      create(LegacyResource.TOGGLE_BOX, h, widgetPool);
   }

   /**
    * Create a TOGGLE-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createToggleBox(handle h, character widgetPool)
   {
      create(LegacyResource.TOGGLE_BOX, h, widgetPool);
   }

   /**
    * Create a WINDOW widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createWindow(handle h)
   {
      createWindow(h, (character) null);
   }

   /**
    * Create a WINDOW widget and save it in the given handle and in the specified pool (which can
    * be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createWindow(handle h, String widgetPool)
   {
      create(LegacyResource.WINDOW, h, widgetPool);
   }

   /**
    * Create a WINDOW widget and save it in the given handle and in the specified pool (which can
    * be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createWindow(handle h, character widgetPool)
   {
      create(LegacyResource.WINDOW, h, widgetPool);
   }
   
   /**
    * Unlock the specified handle and push any pending changes to the client-side.
    * 
    * @param    h
    *           The widget handle to unlock.
    *           
    * @throws   IllegalArgumentException
    *           If the handle is not a valid, widget handle.  This is a P2J-specific API so any
    *           miss-usage is not allowed.
    */
   public static void pushScreenUnlock(handle h)
   {
      if (h.isUnknown())
      {
         throw new IllegalArgumentException("The handle is unknown!");
      }
      WrappedResource res = h.getResource();
      if (!(res instanceof GenericWidget))
      {
         throw new IllegalArgumentException("The handle must reference a widget!");
      }

      GenericWidget gw = (GenericWidget) res;
      gw.setScreenLock(false);
   }
   
   /**
    * Lock the specified handle and any attribute assignments are cached until
    * {@link #pushScreenUnlock lock} is released.
    * 
    * @param    h
    *           The widget handle to lock.
    *           
    * @throws   IllegalArgumentException
    *           If the handle is not a valid, widget handle.  This is a P2J-specific API so any
    *           miss-usage is not allowed.
    */
   public static void pushScreenLock(handle h)
   {
      if (h.isUnknown())
      {
         throw new IllegalArgumentException("The handle is unknown!");
      }
      WrappedResource res = h.getResource();
      if (!(res instanceof GenericWidget))
      {
         throw new IllegalArgumentException("The handle must reference a widget!");
      }

      GenericWidget gw = (GenericWidget) res;
      gw.setScreenLock(true);
   }

   /**
    * Create a BROWSE widget and save it in the given handle and in the unnamed pool, if
    * available.
    *
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createBrowse(handle h)
   {
      createBrowse(h, (character) null);
   }

   /**
    * Create a BROWSE widget and save it in the given handle and in the specified pool
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    *
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createBrowse(handle h, String widgetPool)
   {
      create(LegacyResource.BROWSE, h, widgetPool);
   }

   /**
    * Create a BROWSE widget and save it in the given handle and in the specified pool
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    *
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createBrowse(handle h, character widgetPool)
   {
      create(LegacyResource.BROWSE, h, widgetPool);
   }
   
   /**
    * Checks if it is possible to create new window. Character interfaces 
    * do not support new window creation.
    *
    * @return  <code>TRUE</code> - creation is possible, 
    *          <code>FALSE</code> - otherwise  
    */
   private static boolean isAbleToCreateWindow()
   {
      String msg = "CREATE WINDOW in character mode is not supported";
      boolean batchMode = EnvironmentOps.isBatchMode().booleanValue();
      if (LogicalTerminal.isChui() && batchMode)
      {
         int[] codes = new int[] { 4139, 3181 };
         String[] messages = new String[] 
                  { msg, "Cannot create widget of type WINDOW" };
         ErrorManager.recordOrThrowError(codes, messages, false, false);
         return false;
      }
      
      if (LogicalTerminal.isChui() && !batchMode)
      {
         ErrorManager.recordOrShowError(4139, msg, false, false, false);
         return false;
      }

      return true;
   }
}
