/*
** Module   : ErrorWriterBatch.java
** Abstract : Implementation for error writer for batch mode.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131119 Created initial version.
** 002 EVL 20140115 Adding background and foreground differences support.
** 003 EVL 20140210 In Windows we need to use println() to display message in background mode.
** 004 EVL 20140221 If STDOUT is redirected by OS command we need to use println() even in
**                  foreground batch mode. In Linux we have to also use println().
** 005 VIG 20140922 Changed message method call.
** 006 IAS 20160328 Client log support.
** 007 OM  20161018 Error messages are printed to batch virtual screen before being logged to
**                  (redirected) console.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Implements display outputs for batch mode.
 */
public class ErrorWriterBatch
implements ErrorWriter
{
   /** Client exports access. */
   private ClientExports client = null; 
   
   /**
    * The only constructor used to create error writer.
    *
    * @param    client
    *           The client instance associated with writer.
    */
   public ErrorWriterBatch(ClientExports client)
   {
      this.client = client;
   }
   
   /**
    * Displays error message in bach mode.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void displayError(String errmsg)
   {
      // if the terminal is redirected to a stream, print the error to stream's screen buffer
      if (client != null && client.isRedirected())
      {
         client.displayErrorMessage(errmsg);
      }
      
      // then log it to redirected console
      if (PlatformHelper.isUnderWindowsFamily())
      {
         if (isBatchInBackground() || FileChecker.isConsoleRedirected())
         {
            System.out.println(errmsg);
         }
         else
         {
            // TODO: In Windows the native message box should be displayed here.
            System.out.print(errmsg);
         }
      }
      else
      {
         System.out.println(errmsg);
      }
   }
   
   /**
    * Displays error message using currently installed writer implementation.  This is the
    * redirected version of the displayError.  If the output is redirected to file the message
    * is writing to the file.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void displayErrorRedirected(String errmsg)
   {
      if (client != null && client.isRedirected())
      {
         // we are running on the client
         ColorSpec cs = new ColorSpec(ColorSpec.COLOR_MESSAGES);
         
         client.message(errmsg, false, null, false, null, cs.convert(), WindowConfig.RESOLVE_WINDOW);
      }
      else
      {
         displayError(errmsg);
      }
   }
    
   /**
    * Checks if the client is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public boolean isInBatchMode()
   {
      return (client != null) ? client.isInBatchMode() : false;
   }

   /**
    * Checks if the client is running in batch background mode.
    * 
    * @return  <code>true</code> for background mode, <code>false</code> otherwise.
    * 
    */
   public boolean isBatchInBackground()
   {
      return (client != null) ? client.isBatchInBackground() : false;
   }

   /**
    * Displays error message in bach mode.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void message(String errmsg)
   {
      if (PlatformHelper.isUnderWindowsFamily())
      {
         if (isBatchInBackground() || FileChecker.isConsoleRedirected())
         {
            System.out.println(errmsg);
         }
         else
         {
            // TODO: In Windows the native message box should be displayed here.
            System.out.print(errmsg);
         }
      }
      else
      {
         System.out.println(errmsg);
      }
   }
   
   /**
    * Displays error message box with title in bach mode.
    *
    * @param    masterMsg
    *           Text of the message to display.
    * @param    title
    *           The title for the message box to display.
    */
   public void messageBox(String masterMsg, String title)
   {
      if (PlatformHelper.isUnderWindowsFamily())
      {
         if (isBatchInBackground() || FileChecker.isConsoleRedirected())
         {
            System.out.println(title + ":" + masterMsg);
         }
         else
         {
            // TODO: In Windows the native message box should be displayed here.
            System.out.print(title + ":" + masterMsg);
         }
      }
      else
      {
         System.out.println(title + ":" + masterMsg);
      }
   }
   
   /**
    * Implements pause processing in batch mode.
    */
   public void pause()
   {
      // In batch mode do noting.
      return;
   }

   /**
    * Stops the associated client session by generating UnstoppableExitException.
    */
   public void stopBatchSession()
   {
      if (client != null)
      {
         client.stopBatchSession();
      }
   }

   /**
    * Write message and stack trace to the client log
    * 
    * @param message
    *        the message
    */
   @Override
   public void logStackTrace(String message)
   {
      client.logStackTrace(message);
   }
}