/*
** Module   : FrameConfig.java
** Abstract : stores frame-specific attributes
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 EVL 20050713   @21692 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050721   @21801 Adding layout specific constants.
** 003 SIY 20050831   @22504 Refactoring (see @@22394-22409) 
** 004 SIY 20050907   @22637 Fixed use of @inheritDoc tag in javadoc,
**                           moved methods to ContainerConfig.
** 005 SIY 20050909   @22667 Fixed frame centering.
** 006 SIY 20050926   @22840 Minor cleanup.
** 007 NVS 20051118   @23417 Added setCentered().
** 008 NVS 20051121   @23418 Added setColumns(int) method and columns var.
**                           Added KEEP-TAB-ORDER related methods and var.
**                           Added NO-HELP related methods and var.
**                           Added NO-HIDE related methods and var.
**                           Added NO-UNDERLINE related methods and var.
**                           Added NO-VALIDATE related methods and var.
**                           Added SCROLL n related methods and var.
**                           Added TOP_ONLY related methods and var.
**                           Added missing setBox() method.
** 009 SIY 20060128   @24119 Added support code for the frame manager.
** 010 SIY 20060209   @24469 Major refactoring.
** 011 SIY 20060213   @24576 Changes for the DialogBox support.
** 012 GES 20060227   @24768 pageTop/pageBottom MUST default to false.
** 013 SIY 20060607   @26963 Fixed applying of some attributes.
** 014 SIY 20060627   @27657 Changed handling of the 'title' attribute.
** 015 GES 20060802   @28408 Made externalizable for better performance.
**                           Many code standard cleanups.
** 016 SIY 20060825   @28890 Encapsulated and removed unused fields.
** 017 SIY 20070501   @33337 Get rid of P2JFrame.
** 018 CA  20071122   @35951 Added RETAIN n related methods and var.                              
** 019 SIY 20080218   @37115 Removed setTitle() because it is not necessary  
**                           and triggers redundant re-layouts while frame 
**                           layout is recalculated anyway after applying  
**                           config.
**                           Ref: #37107
** 020 SIY 20080228   @37201 Javadoc cleanup.
** 021 EVL 20080326   @37691 Fix for incorrect numeric fill-in alignment
**                           for COLUMN Progress statement without explicit
**                           SIDE-LABELS parameter set. Must be left
**                           aligned in this case.
** 022 CA  20080529   @38434 Save the configured no-hide value, so it can
**                           be restored when the frame scope is entered 
**                           again. Ref: #38431
** 023 CA  20080905   @39746 Fixed FRAME:ROW/COL handling.
** 024 SIY 20080923   @39907 Limit FRAME:ROW/COL error reporting to cases
**                           when frame is moved at run time by application
**                           so initial frame placement remain silent. 
** 025 CA  20080930   @39983 Added sizeChanged flag, used to notify the
**                           client if the frame size was changed. Fixed
**                           WIDTH/HEIGHT-CHARS behavior.
** 026 SIY 20100619          Switching to own implementation of character UI.
** 027 SIY 20100804          Changes in imports, some refactorings.
** 028 SIY 20100831          Changes related to refactoring of FrameHelper.
** 029 SIY 20110112          Changes related to new color handling scheme.
** 030 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 031 SIY 20111202          Updated imports.
** 032 CA  20131121          If the frame is no longer visible, attempt to delete it.
** 033 GES 20140514          Inherit from PaneConfig and move all of ContainerConfig into
**                           this class.
** 034 CA  20140711          Enhanced font support, for GUI clients.
** 035 HC  20140911          GUI coordinate support - data types refactoring.
** 036 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 037 VIG 20141021          Added windowID.
** 038 HC  20141021          Implemented window move/position support.
** 039 GES 20141124          Added scrollbar-vertical, v6frame, selected and movable.
** 040 HC  20150323          Removed business logic from config classes.
** 041 HC  20150913          Changed the default value of scrollable field to true.
** 042 IAS 20151015          Added support for SCREEN-IO/STREAM-IO/USE-TEXT options
** 043 CA  20151017          Save the frame's top-label header height, so it can be set when other 
**                           dynamic widgets are added after the frame realization.
** 044 IAS 20151027          Added support for NO-AUTO-VALIDATE option, 
**                           removed  NO-HELP and NO-VALIDATE flags
** 045 CA  20151112          Added wasVisible flag.
** 046 CA  20160105          Added runtime support for BORDER- attributes.
** 047 HC  20160203          Improved runtime support of FRAME:SCROLLABLE attribute.
** 048 IAS 20160130          MOUSE-POINTER attribute support.
** 049 CA  20161020          Track when configuration is pushed because the HIDDEN attribute was
**                           set.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * This class is used to hold frame and child frame specific attributes.
 */
public class FrameConfig
extends PaneConfig
{
   /** RETAIN n option, where 0 means paging frame */
   public int retain = 0;

   /** SCROLL n option, where 0 means paging frame */
   public int scroll = 0;

   /** The Progress DOWN attribute. */
   public int down = 1;

   /** Mark DialogBox. */
   public boolean dialog = false;

   /** The Progress CAREFUL-PAINT attribute. */
   public boolean carefulPaint = false;

   /** Means whether the frame must be centered within the parent container. */
   public boolean centered = false;

   /** The Progress LABELS attribute. */
   public boolean labels = true;

   /** The Progress MANUAL-HIGHLIGHT attribute. */
   public boolean manualHighlight = false;

   /** The Progress OVERLAY attribute. */
   public boolean overlay = false;

   /** The Progress PAGE-BOTTOM attribute. */
   public boolean pageBottom = false;

   /** The Progress PAGE-TOP attribute. */
   public boolean pageTop = false;

   /** The Progress SIDE-LABELS attribute. */
   public boolean sideLabels = false;

   /** The Progress BOX (R/O) attribute. */
   public boolean box = true;

   /** KEEP-TAB-ORDER option. */
   public boolean keepTabOrder = false;

   /** 
    * NO-HIDE option - used to track the NO-HIDE status of a frame at runtime. 
    * This will be set to false when the frame goes out of scope.
    */
   public boolean noHide = false;

   /** 
    * The NO-HIDE option as configured at frame definition - used to restore 
    * the NO-HIDE status when the frame scope is opened again.
    */
   public boolean noHideConfig = false;

   /** NO-UNDERLINE option. */
   public boolean noUnderline = false;

   /** NO-AUTO-VALIDATE option. */
   public boolean noAutoValidate = false;

   /** ATTR-SPACE attribute. */
   public boolean attr_space = false;

   /** Flag indicating position change. */
   public boolean positionChanged = false;

   /** Flag indicating size change. */
   public boolean sizeChanged = false;

   /** The amount of the columns to present the widgets */
   public int columns = 0;

   /** The Progress cancel button ID. */
   public int cancelButtonId = -1;

   /** The Progress default button ID. */
   public int defaultButtonId = -1;

   /** Whether this container is over all other non-top only containers. */
   public boolean topOnly = false;

   /** The container has scrolling ability. */
   public boolean scrollable;
   
   /** SCROLLBAR-VERTICAL attribute variable. */
   public boolean scrollbarVertical = true;
   
   /** V6FRAME option. */
   public boolean v6frame = false;
   
   /** SCREEN-IO option. */
   public boolean screenIO = true;

   /** USE-TEXT option. */
   public boolean useText = false;

   /** Id of window owning frame or -1 if frame is not owned by a window */
   public int windowID = WindowConfig.RESOLVE_WINDOW;

   /** Flag to save the previous frame visibility state (when changed by server-side). */
   public boolean wasVisible = false;
   
   /** The title height (in pixels), for a titled frame or dialog. */
   public int titleHeightPixels = 0;
   
   /** Client-side only: saves the frame's top-label header height. */
   public transient double headerHeightChars = 0;
   
   /** MOUSE-POINTER attribute */
   public String mousePointer = null;

   /** Flag indicating if the HIDDEN flag was set. */
   public boolean hiddenSet = false;
   
   /**
    * Default constructor (only used in deserialization).
    */
   public FrameConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected FrameConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   public FrameConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      FrameConfig cfg = (FrameConfig) config;
      
      retain = cfg.retain;
      scroll = cfg.scroll;
      down = cfg.down;
      dialog = cfg.dialog;
      carefulPaint = cfg.carefulPaint;
      centered = cfg.centered;
      labels = cfg.labels;
      manualHighlight = cfg.manualHighlight;
      overlay = cfg.overlay;
      pageBottom = cfg.pageBottom;
      pageTop = cfg.pageTop;
      sideLabels = cfg.sideLabels;
      box = cfg.box;
      keepTabOrder = cfg.keepTabOrder;
      noHide = cfg.noHide;
      noHideConfig = cfg.noHideConfig;
      noUnderline = cfg.noUnderline;
      noAutoValidate = cfg.noAutoValidate;
      attr_space = cfg.attr_space;
      positionChanged = cfg.positionChanged;
      sizeChanged = cfg.sizeChanged;
      columns = cfg.columns;
      cancelButtonId = cfg.cancelButtonId;
      defaultButtonId = cfg.defaultButtonId;
      topOnly = cfg.topOnly;
      scrollable = cfg.scrollable;
      scrollbarVertical = cfg.scrollbarVertical;
      v6frame = cfg.v6frame;
      windowID = cfg.windowID;
      screenIO = cfg.screenIO;
      useText = cfg.useText;
      wasVisible = cfg.wasVisible;
      titleHeightPixels = cfg.titleHeightPixels;
      mousePointer = cfg.mousePointer;
      hiddenSet = cfg.hiddenSet;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      retain             = in.readInt();
      scroll             = in.readInt();
      down               = in.readInt();
      dialog             = in.readBoolean();
      carefulPaint       = in.readBoolean();
      centered           = in.readBoolean();
      labels             = in.readBoolean();
      manualHighlight    = in.readBoolean();
      overlay            = in.readBoolean();
      pageBottom         = in.readBoolean();
      pageTop            = in.readBoolean();
      sideLabels         = in.readBoolean();
      box                = in.readBoolean();
      keepTabOrder       = in.readBoolean();
      noHide             = in.readBoolean();
      noHideConfig       = in.readBoolean();
      noUnderline        = in.readBoolean();
      noAutoValidate     = in.readBoolean();
      attr_space         = in.readBoolean();
      positionChanged    = in.readBoolean();
      sizeChanged        = in.readBoolean();
      columns            = in.readInt();
      cancelButtonId     = in.readInt();
      defaultButtonId    = in.readInt();
      topOnly            = in.readBoolean();
      scrollable         = in.readBoolean();
      scrollbarVertical  = in.readBoolean();
      v6frame            = in.readBoolean();
      windowID           = in.readInt();
      screenIO           = in.readBoolean();
      useText            = in.readBoolean();
      wasVisible         = in.readBoolean();
      titleHeightPixels    = in.readInt();
      mousePointer       = (String) in.readObject();
      hiddenSet          = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(retain);
      out.writeInt(scroll);
      out.writeInt(down);
      out.writeBoolean(dialog);
      out.writeBoolean(carefulPaint);
      out.writeBoolean(centered);
      out.writeBoolean(labels);
      out.writeBoolean(manualHighlight);
      out.writeBoolean(overlay);
      out.writeBoolean(pageBottom);
      out.writeBoolean(pageTop);
      out.writeBoolean(sideLabels);
      out.writeBoolean(box);
      out.writeBoolean(keepTabOrder);
      out.writeBoolean(noHide);
      out.writeBoolean(noHideConfig);
      out.writeBoolean(noUnderline);
      out.writeBoolean(noAutoValidate);
      out.writeBoolean(attr_space);
      out.writeBoolean(positionChanged);
      out.writeBoolean(sizeChanged);
      out.writeInt(columns);
      out.writeInt(cancelButtonId);
      out.writeInt(defaultButtonId);
      out.writeBoolean(topOnly);
      out.writeBoolean(scrollable);
      out.writeBoolean(scrollbarVertical);
      out.writeBoolean(v6frame);
      out.writeInt(windowID);
      out.writeBoolean(screenIO);
      out.writeBoolean(useText);
      out.writeBoolean(wasVisible);
      out.writeInt(titleHeightPixels);
      out.writeObject(mousePointer);
      out.writeBoolean(hiddenSet);
   }
}