/*
** Module   : GenericWidget.java
** Abstract : common superclass for all server side widgets
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23492 Created initial version.
** 002 GES 20060114   @23910 Removed totally useless screen-value attr.
** 003 NVS 20060114   @24094 Implemented hide() and view() methods.
** 004 NVS 20060116   @24095 hide() and view() methods change visibility flag in widget
**                           config instance.
** 005 GES 20060203   @24261 Cache the widget ID in an instance var and require this to be set at
**                           the moment that the widget is added to the frame.  This is much
**                           faster than scanning the map (the old way).
** 006 GES 20060208   @24384 Added getFormat() and switched internal storage to be a string.
** 007 SIY 20060209   @24426 Minor cleanups and fixes.
** 008 NVS 20060310   @24964 Tracking ENTERED state using the ScreenBuffer
** 009 GES 20060315   @25080 Added missing setMultiple() signatures.
** 010 SIY 20060307   @25135 Changed signatures in get/setQuery.
** 011 SIY 20060307   @25136 Use ID declared in ComponentConfig.
** 012 NVS 20060321   @25178 Added isValidatable() method.
** 013 SIY 20060401   @25307 Added setHeader() method.
** 014 GES 20060413   @25515 Added getScreenValue/setScreenValue methods.
** 015 GES 20060426   @25733 Changed underline() implementation.
** 016 SIY 20060426   @25741 Changed signatures of get/setNextTabItem.
** 017 SIY 20060515   @26153 Moved COLUMN-LABEL attribute to ControlEntity.
** 018 GES 20060803   @28379 Massive code cleanups to remove unneeded vars (less memory),
**                           unneeded processing (better performance), duplicated code and to
**                           meet coding standards. The only data left in this class is server-
**                           side data.
** 019 SIY 20060825   @28895 Mark frame as viewed if widget is enabled.
** 020 GES 20060925   @29869 Added removal of ampersand chars from label text to duplicate an
**                           undocumented 4GL feature.
** 021 GES 20061108   @31027 Added missing APIs for the list-item-pairs attribute.
** 022 NVS 20070426   @33244 Reimplemented SENSITIVE attribute support using new methods
**                           of LogicalTerminal.
** 023 GES 20080319   @37592 Added NEXT-SIBLING, PREV-SIBLING and TYPE.
** 024 GES 20080421   @38071 setFormat() doesn't call pushScreenDefinition but rather, it forces
**                           the format string change directly down to the client. This helps
**                           performance and allows the widget to redraw immediately which is
**                           needed to duplicate Progress behavior.
** 025 SIY 20080429   @38142 Added saveColors()/restoreColors() methods.
** 026 SIY 20080604   @38549 Added generation of error on any attempt to set too long label.
**                           Fixed comments.
** 027 SIY 20080609   @38649 Do not check label length until frame is  viewed.
** 028 SIY 20090210   @41287 Removed call to GenericFrame.setHadView(). Now this method is
**                           invoked by LogicalTerminal.
** 029 CA  20090227   @41387 Added WrappedResources.isUnknown method; this
**                           implementation returns false.
** 030 CA  20090413   @41776 Fixed handling of VISIBLE/HIDDEN flag. When set to ON, it turns
**                           its counterpart to OFF and vice versa.
** 031 GES 20090424   @41959 Import change.
** 032 SIY 20090701   @43023 Added getLabelStr() method.
** 033 SVL 20101004          Added support for DATA-TYPE, TABLE, TOOLTIP and
**                           WIDTH-PIXELS widget attributes.
** 034 SIY 20101004          Added more stubs.
** 035 SIY 20101005          Added stubs for HEIGHT-PIXELS getter/setter.
** 036 SVL 20101006          Added support for WIDTH, HEIGHT, TOOLTIP(String) attributes.
** 037 SVL 20101006          Added moveAfterTab, moveToTop, moveToBottom.
** 038 ECF 20101006          Changed signature of getName to return character rather than String.
** 039 SIY 20101014          Added asWidgetHandle() method.
** 040 SIY 20110112          Changes related to new color scheme.
** 041 SIY 20111101          Updated imports.
** 042 CA  20130112          getNextSibling and getPrevSibling must return a handle instance.
**                           Extracted into CommonWidget interface all widget APIs.
** 043 CS  20130204          Added conversion support for socket and 
**                           server-socket. Moved sensitive related methods from 
**                           CommonWidget in the Sensitive interface.
** 044 CA  20130221          Refactored usage of TYPE, FORMAT, COlUMN-LABEL, DATA-TYPE attrs.
**                           Completed support for WIDTH/HEIGHT-CHARS/PIXELS. 
** 045 EVL 20130222          Adding UI related attributes and methods.
** 046 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
**                           A widget is considered valid if not attached to a frame.
** 047 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 048 CA  20130323          The PARENT attribute must be a CommonWidget, not Object.
**                           Added stubs for moveBeforeTabItem(handle) and moveAfterTab(handle)
**                           methods.
** 049 CA  20130323          The PARENT and FRAME attributes need to be set/get via handles.
** 050 CA  20130603          resourceDelete must return a boolean value.
** 051 OM  20130527          Added int64 support for APPLY events and other accessors.
** 052 CA  20130919          Added a flag to track the deletion of a widget handle.
** 053 CA  20130927          Removed unknown(); fixed getFirst/LastChild to return a handle 
**                           instance; added setFrameHandle(CommonFrame).
** 054 CA  20131021          Added DYNAMIC attribute support. Dynamic widgets can not be deleted.
** 055 SVL 20131203          Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 056 CA  20131115          Changes related to properly instantiate dynamic resources and to
**                           implicit deletion of static resources.
** 057 HC  20131215          Improved error handling of the COLUMN-LABEL attribute.
** 058 GES 20130925          Added stubs for CURSOR-OFFSET, FOCUSED-ROW-SELECTED and
**                           SIDE-LABEL-HANDLE attributes.
** 059 VIG 20140314          Name implementation for field-groups.
** 060 VIG 20140510          moveBeforeTabItem methods implementation, 
**                           first-tab-item and next-tab-item management methods.
** 061 EVL 20140609          Added CHECKED attribute getter and setter.  Also new method added to
**                           check if the widget can be realized preRealizeCheck().
** 062 MAG 20140512          Added WINDOW-related attributes. Changed color attribute method
**                           names to match naming standards.
** 063 CA  20140701          Added stubbed support for font-related attributes: FONT, COLUMN-FONT,  
**                           LABEL-FONT, TITLE-FONT.
** 064 CA  20140711          Enhanced font support, for GUI clients.
** 065 EVL 20140918          Adding support for dynamic widget to add an remove.
** 066 EVL 20140926          Dynamic frames processing should be separated from dynamic widgets.
**                           More processing for dynamic widgets implementation.
** 067 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 068 HC  20141021          Implemented window move/position support.
** 069 EVL 20141105          Adding rectangle related attriutes setters for: EDGE-CHARS,
**                           EDGE-PIXELS, FILLED, GRAPHIC-EDGE, GROUP-BOX, ROUNDED.
** 070 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 071 GES 20141124          Added many shared attrs/meths. Shifted most exceptions to the
**                           UnimplementedFeature.unsupported() method.
** 072 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 073 VIG 20141210          Added setPopupMenu methods.
** 074 CA  20150123          Fixed handling of side-labels in shared frames.
** 075 IAS 20150123          delete(int) replaced with delete(long).
** 076 HC  20150214          Improved positioning of side labels.
** 077 VIG 20150128          Added new field menu (associated popup-menu). 
** 078 CA  20150223          Fixed label's COLUMN and WIDTH-CHARS attributes (in ChUI). 
** 079 CA  20150224          Setting the SENSITIVE attribute must not make the widget visible.  
** 080 IAS 20150225          Move methods specific for list widgets to a separate interface.
** 081 CA  20150323          Added API to set the format length from the widget's @ phrase(s).
** 082 IAS 20150325          Added support for the TEXT-SELECTED attribute.
** 083 HC  20150323          Removed business logic from config classes.
** 084 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 085 VIG 20150225          Fixed name() method: added MenuItemWidget support, setParentHandle.
** 086 EVL 20150429          Fix for button resizing issue for very long labels.
** 087 SVL 20150503          getBrowseColumn moved to BrowseWidget.
** 088 IAS 20150605          Added support for the text selection, stub implementation of the
**                           realize() method, misc. utility methods
** 089 CA  20150723          Added finishConfigProcessing.
** 090 CA  20150725          Added FRAME-COL, FRAME-ROW, FRAME-X, FRAME-Y support.
**                           Added batching of widget assignments in CREATE ... ASSIGN case.
** 091 HC  20150803          Improved runtime support for WINDOW size attributes.
** 092 IAS 20150804          Added support for the WIDGET-ID, CONTEXT-HELP-ID, TAB-STOP,  
**                           MANUAL-HIGHLIGHT, and WINDOW attribute
** 093 SVL 20150821          Fixes related to pushScreenLock/pushScreenUnlock. Added cleanup().
**                           setVisible is no-op if frame is not set.
** 094 IAS 20150922          Added support for SELECTABLE, MENU-MOUSE, MENU-KEY, WINDOW.
**                           attributes and MOVE-TO-TOP()/MOVE-TO-BOTTOM() methods
** 095 OM  20150915          Added support for the LOAD-MOUSE-POINTER, VALIDATE methods and
**                           RESIZABLE, INPUT-VALUE, ENABLED and PREV-TAB-ITEM attributes.
**                           Fixed NPE in case of the late FRAME assignment
** 096 IAS 20151015          Added setter for the THREE-D attribute
** 097 CA  20151016          Fixed NPE when LABEL attr for dyn widget is set before FRAME.
** 098 IAS 20151020          setParentHandle() for the case when parent is a FieldGroup
**                           (BACKGROUND attribute support)
**     IAS 20151029          VALIDATE() method implementation
** 099 CA  20151024          Ignore WIDGET-ID option at the frame setup, for now.
**                           Allow FRAME attribute to be set for frames, too, to enable nested
**                           frames.  Added support for MOVE-AFTER-TAB-ITEM, TABLE and DBNAME.
**                           Fixed move-to-top/bottom for nested frames.
**                           A fix for HIDDEN attribute: distinguish between HIDE statement and 
**                           setting the HIDDEN attribute to true.  Also, if the frame's window is 
**                           marked as HIDDEN, do not show the window.
** 100 CA  20151125          Fixed SENSITIVE=TRUE attribute set when the parent (and root) frame 
**                           is hidden.
** 101 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 102 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 103 IAS 20160229          LOAD-MOUSE-POINTER support, fixed name() method
** 104 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 105 CA  20160304          resourceDelete() must allow deletion of widgets which belon solely to
**                           the slave shared frame.
** 106 EVL 20160411          Keep the accelerator '&' symbol inside the widget label but only for
**                           not redirected case.
** 107 GES 20160311          Removed getHeight() and getWidth() which were dead code.
** 108 VIG 20160407          Extracted setLabelInt method from setLabel.
** 109 GES 20160525          Reworked validation expression processing to use lambdas.
** 110 GES 20160616          Added boolean versions of entered/not entered (to reduce unwrapping)
**                           and reworked the entered code to directly access the screen buffer
**                           avoiding a more expensive call.
** 111 HC  20160628          Fixed name() to report the legacy 4GL name and not the converted P2J
**                           name.
** 112 SVL 20160630          Fixed assignment of SCREEN-VALUE when the parent frame is not set.
** 113 HC  20160714          Fixed frame widgets being properly added and linked with the frame's
**                           primary field group.
** 114 HC  20160716          Implemented runtime support for BROWSE-COLUMN:RESIZABLE attribute.
**     SBI 20160720          Don't log UnimplementedFeature for RESIZABLE attribute.
** 115 CA  20160902          Fixed canDelete - implicit deletion is checked only for static widgets.
** 116 GES 20160909          Moved rectangle attribute setters to the RectangleInterface.
** 117 SBI 20160914          A browse column always has a label.
** 118 CA  20161005          Fixed issues related to SCREEN-VALUE attribute (initial value or 
**                           explicitly set, when the value is empty string or unknown, or the 
**                           widget is uninitialized).
** 119 CA  20160928          Setting the HIDDEN attribute for a not-realized widget attached to a
**                           non-visible frame does not realize the widget, and it does not set
**                           the VISIBLE attribute to true.
** 120 VIG 20161019          Added modified flag setting to setScreenValue().
** 121 OM  20161125          Added setHelpStatic() and default help.
** 122 OM  20161206          Dropped dead code.
** 123 GES 20170129          Added a getter for dbName and tableName raw values.
**     GES 20170130          Don't set the pending error flag for a 4052 "error".
**     HC  20170131          Added runtime support for dynamic MENU* widgets. Fixed FRAME
**                           assignment not adding the target dynamic widget to the frame's field
**                           group.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * This class represents server side widgets. Every widget is associated with
 * its configuration object, which is embedded into the widget. The
 * widget constructors create and initialize their backing configuration
 * objects.
 */
public class GenericWidget<T extends WidgetConfig>
extends HandleChain
implements WrappedResource,
           CommonWidget,
           Sensitive,
           DynamicResource,
           ConfigOwner<T>
{
   /** The containing frame object. */
   protected GenericFrame frame = null;
   
   /** The configuration container. */
   protected T config = null;

   /** Flag indicating the resource was deleted. */
   protected boolean deleted = false;

   /** The validation expression to use to test proposed edits for this widget. */
   private ValidationExpr valexp = null;
   
   /** The expression to generate an error message when validation fails. */
   private ValidationMsg valmsg = null;
   
   /** MenuWidget handle owned by this widget */
   private MenuWidget menu;
   
   /** When <code>true</code>, marks this resource as being dynamically created. */
   private boolean dynamic = false;

   /**
    * When set, all attribute assignments are cached on server-side and pushed to client-side
    * only when the lock is set to <code>false</code> via {@link #setScreenLock}.  When this flag
    * is <code>true</code>, any {@link #pushScreenDefinition()} calls are ignored
    */
   private boolean frameLock = false;
   
   /** The value of the DBNAME attribute. */
   private String dbName = null;
   
   /** The value of the TABLE attribute. */
   private String tableName = null;

   /** Holds the SCREEN-VALUE until the parent frame is set. */
   private character pendingScreenValue = null;
   
   /** Holds default (schema) help message if any is defined in database. */
   private String defaultHelp = null;
   
   /** {@code true} if the currently displayed help is the default(oevrrideable) schema help. */
   private boolean isDefaultHelp = true;
   
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    config
    *           The container for all state to be transferred to the client.
    */
   public GenericWidget(boolean dynamic, T config)
   {
      super(dynamic);
      
      this.dynamic = dynamic;
      this.config = config;
   }
   
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    config
    *           The container for all state to be transferred to the client.
    * @param    before
    *           The right neighbor of the BaseEntity been created or null if new BaseEntity
    *           will be the last in the chain. If <b>before</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    * @param    firstAsLast
    *           The flag indicating that this new entity must be inserted as last element 
    *           in the chain in the case when <b>before</b> - is the first element of the chain.
    *           When <b>before</b> is the first element of the chain, default behavior (expected
    *           from the name of this parameter is to insert new element as the first. 
    *           This flag allows to change the expected behavior of <b>before</b> parameter when 
    *           it needs and makes sense only if <b>before</b> is the first element.
    */
   public GenericWidget(boolean dynamic, T config, GenericWidget<T> before, boolean firstAsLast)
   {
      super(dynamic, true, before, firstAsLast);
      
      this.dynamic = dynamic;
      this.config = config;
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    hFrame
    *           handle which is expected to contain a <code>FrameWidget</code>
    *           instance
    */
   public void setFrame(handle hFrame)
   {
      if (!hFrame._isValid())
      {
         String msg = 
            "**Unable to assign UNKNOWN value to attribute FRAME on " + type() + " widget";
         ErrorManager.recordOrShowError(4083, msg, false, false, false);
         
         return;
      }

      HandleResource res = (HandleResource) hFrame.get();
      
      if (!(res instanceof FrameWidget))
      {
         String msg = "**Attribute FRAME for the FRAME %s was passed an invalid %s widget";
         
         msg = String.format(msg, name(), res.type());
         ErrorManager.recordOrShowError(4062, msg, false, false, false);
         
         return;
      }
      
      setFrame((FrameWidget) hFrame.get());
   }
   
   /**
    * Sets the FRAME attribute, containing frame instance.
    * 
    * @param    h
    *           a handle containing the <code>GenericFrame</code> instance that contains
    *           this widget
    */
   public void setFrameHandle(handle h)
   {
      setFrame(h);
   }
   
   /**
    * Sets the FRAME attribute, containing frame instance.
    * 
    * @param    frame
    *           The frame instance that will contain this widget.
    */
   public void setFrameHandle(CommonFrame frame)
   {
      setFrame(frame.asWidget());
   }

   /**
    * Sets the FRAME attribute, containing frame instance.
    * 
    * @param    frame
    *           The frame instance that will contain this widget.
    */
   public void setFrameHandle(FrameWidget frame)
   {
      setFrame(frame);
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   public void setFrame(GenericFrame frame)
   {
      this.frame = frame;
      this.config.frameId = frame.getFrameId();
      
      if (this instanceof BaseEntity && !(this instanceof FieldGroup))
      {
         FieldGroup fg = frame.getFieldGroup();

         // fg == null when this frame widget is being initialized,
         // during the initialization this method may be called to link
         // this widget with the corresponding GenericFrame part and we don't
         // want it to end up in the field group
         if (fg != null)
         {
            ((BaseEntity) this).group = fg;
            if (dynamic)
            {
               fg.addWidget(this);
            }
         }
      }

      // dor dynamic widgets this is the place where we need to initialize all frame internals
      // related to the wiget integration into the frame functionality
      if (dynamic)
      {
         if (!(this instanceof FrameWidget))
         {
            if (frameLock)
            {
               // continue collecting changes if frame lock was set
               frame.batch(true);
            }
            
            frame.addDynamicWidget(this);
         }
      }

      if (pendingScreenValue != null)
      {
         setScreenValue(pendingScreenValue);
         pendingScreenValue = null;
      }
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   public void setFrame(FrameWidget frame)
   {
      setFrame(frame.getFrame());
   }
   
   /**
    * Gets the containing frame instance.
    * 
    * @return   The frame that contains this widget.
    */
   public GenericFrame getFrame()
   {
      return frame;
   }

   /**
    * Gets the containing frame instance, as a handle.
    * 
    * @return   The frame that contains this widget.
    */
   public handle getFrameHandle()
   {
      return new handle(getFrame());
   }

   /**
    * Gets the numeric ID of this widget (this is used as an index into the
    * screen-buffer among other things).
    * 
    * @return   The numeric ID of this widget.
    */
   public int getId()
   {
      return config.id == null ? -1 : config.id.asInt();
   }

   /**
    * Sets the numeric ID of this widget (this is used as an index into the
    * screen-buffer among other things).
    * 
    * @param    id
    *           The numeric ID of this widget.
    */
   public void setId(int id)
   {
      // TODO: for now, ID's are not allowed to be set if the widget's ID is already set
      if (config.id != null)
      {
         return;
      }
      
      config.id = new WidgetId(id);
      
      // when creating dynamic frames add them to the widget registry
      if (dynamic && this instanceof FrameWidget)
      {
         // create new generic frame for this frame widget
         CommonFrame gDynFrame = GenericFrame.createDynamicFrame(DynamicFrameTemplate.class,
                                                                 (FrameWidget)this);
      }
   }

   /**
    * Checks whether the widget was changed during the most recent editing operation.
    * 
    * @return   <code>true</code> if this widget has changed its value.
    */
   public boolean _isEntered()
   {
      // it is safe to use the frame buffer here without any update to dynamic content
      return frame.getFrameBufferRaw().isEntered(getId());
   }

   /**
    * Checks whether the widget was changed during the most recent editing operation.
    * 
    * @return   <code>true</code> if this widget has changed its value.
    */
   public logical isEntered()
   {
      return new logical(_isEntered());
   }

   /**
    * Checks whether the widget was changed during the most recent editing operation.
    * 
    * @return   <code>true</code> if this widget has changed its value.
    */
   public boolean _isNotEntered()
   {
      return !_isEntered();
   }

   /**
    * Checks whether the widget was NOT changed during the most recent editing operation.
    * 
    * @return   <code>true</code> if this widget has not changed its value.
    */
   public logical isNotEntered()
   {
      return new logical(_isNotEntered());
   }

   /**
    * Checks whether the row that has the focus is selected.  This is the equivalent of the
    * FOCUSED-ROW-SELECTED attribute of the browse widget.
    * 
    * @return   <code>true</code> if the focused row is selected.
    */
   @Override
   public logical isFocusedRowSelected()
   {
      UnimplementedFeature.unsupported("FOCUSED-ROW-SELECTED attribute is undefined");
      return new logical();
   }
   
   /**
    * Get the ENABLE attribute of this widget.
    *
    * @return  logical value of ENABLED attribute
    */
   @Override
   public logical isEnabled()
   {
      UnimplementedFeature.missing("ENABLED attribute not implemented.");
      return new logical();
   }
   
   /**
    * Get the PREV-TAB-ITEM attribute of this widget.
    *
    * @return  logical value of PREV-TAB-ITEM attribute
    */
   @Override
   public handle getPreviousTabItem()
   {
      UnimplementedFeature.missing("PREV-TAB-ITEM attribute not implemented.");
      return new handle();
   }
   
   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   public logical isReadOnly() 
   {
      UnimplementedFeature.unsupported("READ-ONLY attribute is undefined");
      return new logical();
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   public void setReadOnly(logical r) 
   {
      UnimplementedFeature.unsupported("READ-ONLY attribute is undefined");
   }

   /**
    * Sets the AUTO-RETURN option.
    * 
    * @param    autoReturn
    *           <code>true</code> if the option is set to ON.
    */
   public void setAutoReturn(boolean autoReturn)
   {
      UnimplementedFeature.unsupported("AUTO-RETURN attribute is undefined");
   }

   /**
    * Set AUTO-RESIZE attribute value.
    * 
    * @param    autoResize
    *           New value for the attribute.
    */
   public void setAutoResize(logical autoResize)
   {
      UnimplementedFeature.unsupported("AUTO-RESIZE attribute is undefined");
   }

   /**
    * Set AUTO-RESIZE attribute value.
    * 
    * @param    autoResize
    *           New value for the attribute.
    */
   @Override
   public void setAutoResize(boolean autoResize)
   {
      UnimplementedFeature.unsupported("AUTO-RESIZE attribute is undefined");
   }

   /**
    * Get AUTO-RESIZE attribute value.
    * 
    * @return
    *           The value of the attribute.
    */
   @Override
   public logical isAutoResize()
   {
      UnimplementedFeature.unsupported("AUTO-RESIZE attribute is undefined");
      return new logical();
   }

   /**
    * Sets the BLANK option.
    * 
    * @param    blank
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setBlank(boolean blank)
   {
      UnimplementedFeature.unsupported("BLANK attribute is undefined");
   }

   /**
    * Gets the CHECKED writable attribute.
    *
    * @return    The current value of the CHECKED attribute.
    */
   @Override
   public logical isChecked()
   {
      UnimplementedFeature.unsupported("CHECKED attribute is undefined");
      return new logical();
   }
   
   /**
    * Set CHECKED attribute value.
    * 
    * @param   checked
    *          New value for the attribute.
    */
   @Override
   public void setChecked(boolean checked)
   {
      UnimplementedFeature.unsupported("CHECKED attribute is undefined");
   }

   /**
    * Set CHECKED attribute value.
    * 
    * @param   checked
    *          New value for the attribute.
    */
   @Override
   public void setChecked(logical checked)
   {
      UnimplementedFeature.unsupported("CHECKED attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public decimal getColumn()
   {
      UnimplementedFeature.unsupported("COLUMN attribute is undefined");
      return new decimal();
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getFrameColumn()
   {
      UnimplementedFeature.unsupported("FRAME-COL attribute is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   column
    *          Not used.
    */
   @Override
   public void setColumn(NumberType column)
   {
      UnimplementedFeature.unsupported("COLUMN attribute is undefined");
   }

   /**
    * Getter of COLUMN-LABEL attribute.
    *
    * @return   current value of COLUMN-LABEL attribute.
    */
   @Override
   public character getColumnLabel()
   {
      invalidAttribute("COLUMN-LABEL");
      return new character();
   }

   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   @Override
   public void setColumnLabel(character columnLabel)
   {
      readOnlyError("COLUMN-LABEL");
   }

   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   @Override
   public void setColumnLabel(String columnLabel)
   {
      readOnlyError("COLUMN-LABEL");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   column
    *          Not used.
    */
   @Override
   public void setColumn(double column)
   {
      UnimplementedFeature.unsupported("COLUMN attribute is undefined");
   }

   /**
    * Obtains the index position of the cursor in a text editing widget. This corresponds to
    * the CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @return   The character index at which the cursor is placed.
    */
   @Override
   public integer getCursorOffset()
   {
      UnimplementedFeature.unsupported("CURSOR-OFFSET attribute is undefined");
      return new integer();
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(double offset)
   {
      UnimplementedFeature.unsupported("CURSOR-OFFSET attribute is undefined");
   }

   /**
    * Sets the index position of the cursor in a text editing widget. This corresponds to the
    * CURSOR-OFFSET attribute in BROWSE columns (only when the cell is FILL-IN or COMBO-BOX),
    * COMBO-BOX, EDITOR or FILL-IN widgets.
    * 
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   @Override
   public void setCursorOffset(NumberType offset)
   {
      UnimplementedFeature.unsupported("CURSOR-OFFSET attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public integer getDColor()
   {
      UnimplementedFeature.unsupported("DCOLOR attribute is undefined");
      return new integer();
   }

   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   @Override
   public void setDColor(NumberType param)
   {
      UnimplementedFeature.unsupported("DCOLOR attribute is undefined");
   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   @Override
   public void setDColor(Color displayColor)
   {
      UnimplementedFeature.unsupported("DCOLOR attribute is undefined");
   }

   /**
    * Sets the DEBLANK option.
    * 
    * @param    deblank
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDeblank(boolean deblank)
   {
      UnimplementedFeature.unsupported("DEBLANK attribute is undefined");
   }

   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDisableAutoZap(boolean disableAutoZap)
   {
      UnimplementedFeature.unsupported("DISABLE-AUTO-ZAP attribute is undefined");
   }

   /**
    * Sets the format specification.
    * 
    * @param   format
    *          format specification
    */
   @Override
   public void setFormat(character format)
   {
      setFormat(format.isUnknown() ? "" : format.getValue());
   }

   /**
    * Sets the format specification.
    * 
    * @param    format
    *           The format specification.
    */
   @Override
   public void setFormat(String format)
   {
      if (Utils.equals(config.getDynamicFormat(), format))
      {
         return;
      }

      config.setDynamicFormat(format);
      
      if (frame != null && !frame.isBatch())
         LogicalTerminal.setFormat(getId(), format);
   }

   /**
    * Gets the format specification.
    * 
    * @return   The format specification.
    */
   public String _getFormat()
   {
      return config.getDynamicFormat();
   }

   /**
    * Gets the format specification.
    * 
    * @return   The format specification.
    */
   public character getFormat()
   {
      return new character(_getFormat());
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public decimal getHeightChars()
   {
      UnimplementedFeature.unsupported("HEIGHT-CHARS attribute is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   h
    *          Not used.
    */
   @Override
   public void setHeightChars(NumberType h)
   {
      UnimplementedFeature.unsupported("HEIGHT-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   h
    *          Not used.
    */
   @Override
   public void setHeightChars(double h)
   {
      UnimplementedFeature.unsupported("HEIGHT-CHARS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public decimal getWidthChars()
   {
      UnimplementedFeature.unsupported("WIDTH-CHARS attribute is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   h
    *          Not used.
    */
   @Override
   public void setWidthChars(NumberType h)
   {
      UnimplementedFeature.unsupported("WIDTH-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   h
    *          Not used.
    */
   @Override
   public void setWidthChars(double h)
   {
      UnimplementedFeature.unsupported("WIDTH-CHARS attribute is undefined");
   }

   /**
    * Get the HELP writable attribute. It it was not set, the default value is returned.
    * 
    * @return  current value of HELP attribute
    */
   @Override
   public character getHelp()
   {
      String help = config.getDynamicHelp();
      if (help == null)
      {
         help = defaultHelp;
      }
      return (help == null) ? new character() : new character(help);
   }

   /**
    * Dynamically sets the help text, by setting the attribute.
    * 
    * @param    help
    *           The new help text.
    */
   @Override
   public void setHelp(character help)
   {
      setHelp(help.isUnknown() ? "" : help.getValue());
   }

   /**
    * Dynamically sets the help text, by setting the attribute.
    * 
    * @param    help
    *           The new help text.
    */
   @Override
   public void setHelp(String help)
   {
      isDefaultHelp = false;
      if (Utils.equals(config.getDynamicHelp(), help))
      {
         return;
      }
      
      config.setDynamicHelp(help);
      pushScreenDefinition();
   }
   
   /**
    * Sets the help text statically, when the frame is designed.
    * For {@code SHARED FRAMES}, this attribute cannot be overwritten once it was set.
    *
    * @param   help
    *          The help text.
    */
   public void setHelpStatic(String help)
   {
      if (!isDefaultHelp)
      {
         // already set?
         return;
      }
      
      setHelp(help);
   }
   
   /**
    * Sets the default help text statically, when the frame is designed using value from database
    * schema.
    *
    * @param   help
    *          The default help text.
    */
   public void setHelpDefault(String help)
   {
      defaultHelp = help; // kept locally only
      isDefaultHelp = true;
      
      config.setDynamicHelp(help);
      pushScreenDefinition();
   }

   /**
    * Gets the HIDDEN writable attribute.
    *
    * @return   The current value of the HIDDEN attribute.
    */
   @Override
   public logical isHidden() 
   {
      return new logical(_isHidden());
   }

   /**
    * Gets the HIDDEN writable attribute.
    *
    * @return   The current value of the HIDDEN attribute.
    */
   public boolean _isHidden() 
   {
      return config.hidden;
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @Override
   public void setHidden(logical hidden) 
   {
      setHidden(hidden.booleanValue());
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @Override
   public void setHidden(boolean hidden) 
   {
      if (config.hidden == hidden && (config.visible || config.hidden))
      {
         // be a no-op only if the widget has been realized.
         return;
      }

      // check if the parent frame is set and not visible
      if (!config.realized && (frame == null || !frame.getFrameWidget()._isVisible()))
      {
         config.hidden = hidden;

         // not yet attached or the frame is not realized, just set the attribute and do not push 
         // the configuration
         return;
      }

      _setVisible(!hidden);
      config.hidden = hidden;
      pushScreenDefinition();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public logical isKeepFrameZOrder()
   {
      UnimplementedFeature.unsupported("KEEP-FRAME-Z-ORDER attribute is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   k
    *          Not used.
    */
   @Override
   public void setKeepFrameZOrder(logical k)
   {
      UnimplementedFeature.unsupported("KEEP-FRAME-Z-ORDER attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   k
    *          Not used.
    */
   @Override
   public void setKeepFrameZOrder(boolean k)
   {
      UnimplementedFeature.unsupported("KEEP-FRAME-Z-ORDER attribute is undefined");
   }

   /**
    * Get the label text for this widget.
    *
    * @return   The label text.
    */
   @Override
   public character getLabel()
   {
      return new character(config.getDynamicLabel());
   }

   /**
    * Get the label text for this widget.
    *
    * @return   The label text.
    */
   @Override
   public String getLabelStr()
   {
      String clabel = config.getDynamicColumnLabel();
      
      return  clabel != null ? clabel : config.getDynamicLabel();
   }
   
   /**
    * Sets the label text, removing any '&amp;' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   @Override
   public void setLabel(character label)
   {
      setLabel((label.isUnknown() ? "" : label.getValue()));
   }

   /**
    * Sets the label text, removing any '&amp;' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   @Override
   public void setLabel(String label)
   {
      // TODO: Some refactoring is required here to avoid instance check and class casting
      //       the code below is not very good from the architecture perspective. Ideally we need
      //       to provide some protected method here allowing the particular widget to override it
      //       and perform widget specific actions, for example ButtonWidget or ToggleBox
      //       in this case.

      // dynamic widgets can not accept labels, except BUTTON and TOGGLE-BX, BROWSE-COLUMN
      if (_dynamic() &&
          !(this instanceof ButtonWidget)    &&
          !(this instanceof ToggleBoxWidget) &&
          !(this instanceof BrowseColumnWidget))
      {
         if (this instanceof FrameWidget)
         {
            // form the message and display it
            ErrorManager.displayError(4052, "LABEL is not a settable attribute for FRAME widget");
         }
         else
         {
            StringBuilder err = new StringBuilder();
            // form the message
            err.append("Unable to set LABEL because no LABEL-HANDLE on the ")
               .append(type())
               .append(" widget");
            // and display it
            ErrorManager.displayError(4072, err.toString());
         }
         return;
      }

      setLabelInt(label);
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public decimal getMaxHeightChars()
   {
      UnimplementedFeature.unsupported("MAX-HEIGHT-CHARS attribute is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   m
    *          Not used.
    */
   @Override
   public void setMaxHeightChars(decimal m)
   {
      UnimplementedFeature.unsupported("MAX-HEIGHT-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public decimal getMaxWidthChars()
   {
      UnimplementedFeature.unsupported("MAX-WIDTH-CHARS attribute is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   m
    *          Not used.
    */
   @Override
   public void setMaxWidthChars(decimal m)
   {
      UnimplementedFeature.unsupported("MAX-WIDTH-CHARS attribute is undefined");
   }


   /**
    * Queries if the SCREEN-VALUE attribute for this widget was changed.
    * 
    * @return  If <code>true</code>, the widget's screen value was changed.
    */
   @Override
   public logical isModified()
   {
      UnimplementedFeature.unsupported("MODIFIED attribute getter");
      return new logical();
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   @Override
   public void setModified(boolean modified)
   {
      UnimplementedFeature.unsupported("MODIFIED attribute setter");
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   @Override
   public void setModified(logical modified)
   {
      UnimplementedFeature.unsupported("MODIFIED attribute setter"); 
   }
   
   /**
    * Queries if the user is allowed to move this widget using direct manipulation.
    * 
    * @return  If <code>true</code>, the widget is movable via direct manipulation.
    */
   @Override
   public logical isMovable()
   {
      UnimplementedFeature.unsupported("MOVABLE attribute getter");
      return new logical();
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   @Override
   public void setMovable(boolean movable)
   {
      UnimplementedFeature.unsupported("MOVABLE attribute setter");
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   @Override
   public void setMovable(logical movable)
   {
      UnimplementedFeature.unsupported("MOVABLE attribute setter");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public logical isMultiple()
   {
      UnimplementedFeature.unsupported("MULTIPLE attribute is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   multiple
    *          Not used.
    */
   @Override
   public void setMultiple(boolean multiple)
   {
      UnimplementedFeature.unsupported("MULTIPLE attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   multiple
    *          Not used.
    */
   @Override
   public void setMultiple(logical multiple)
   {
      UnimplementedFeature.unsupported("MULTIPLE attribute is undefined");
   }

   /**
    * Calculates the next widget in the tab order.
    * 
    * @return  The next widget or null if this widget is the last.
    */
   @Override
   public CommonWidget getNextTabItem()
   {
      int widgetId = LogicalTerminal.getNextTabItem(this);
      return getFrame().getWidgetForId(widgetId);
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   n
    *          Not used.
    */
   @Override
   public void setNextTabItem(GenericWidget<?> n)
   {
      UnimplementedFeature.unsupported("NEXT-TAB-ITEM attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public integer getNumSelectedRows()
   {
      UnimplementedFeature.unsupported("NUM-SELECTED-ROWS attribute is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   r
    *          Not used.
    */
   @Override
   public void setNumSelectedRows(int64 r)
   {
      UnimplementedFeature.unsupported("NUM-SELECTED-ROWS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public CommonWidget getParent()
   {
      UnimplementedFeature.unsupported("PARENT attribute is undefined");
      return null;
   }

   /**
    * Gets the PARENT attribute, as a handle.
    * 
    * @return  the PARENT attribute.
    */
   @Override
   public handle getParentHandle()
   {
      return new handle(getParent());
   }

   /**
    * Sets the PARENT attribute to the resource specified in the given handle.
    * 
    * @param   h
    *          The handle with the parent resource.
    */
   @Override
   public void setParentHandle(handle h)
   {
      if (!h._isValid())
      {
         String msg = 
            "**Unable to assign UNKNOWN value to attribute PARENT on " + type() + " widget";
         ErrorManager.recordOrShowError(4083, msg, false, false, false);
         
         return;
      }
      
      Object parent = h.get();
      if (dynamic && (parent instanceof FieldGroup))
      {
         ((FieldGroup)parent).addDynamicWidget(this); 
      }
      else if (parent instanceof GenericWidget)
      {      
         setParent((GenericWidget<?>) parent);
      }
      else
      {
         UnimplementedFeature.unsupported("Parent must be an instance of GenericWidget.");
      }
   }

   /**
    * Sets the PARENT attribute to the specified widget.
    * 
    * @param   w
    *          The parent widget
    */
   @Override
   public void setParentHandle(CommonWidget w)
   {
      if (w instanceof GenericWidget)
      {      
         setParent((GenericWidget<?>) w);
      }
      else
      {
         UnimplementedFeature.unsupported("Parent must be an instance of GenericWidget.");
      }
   }
   
   /**
    * Establishes a relationship between the container and the elements in
    * a given widget group.
    *
    * @param    group
    *           The container of the widget.
    */
   @Override
   public void setGroup(Group group)
   {
      UnimplementedFeature.unsupported("Group support not available.");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public integer getPfColor()
   {
      UnimplementedFeature.unsupported("PFCOLOR attribute is undefined");
      return new integer();
   }

   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The color to use.
    */
   @Override
   public void setPfColor(NumberType param)
   {
      UnimplementedFeature.unsupported("PFCOLOR attribute is undefined");
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    inputColor
    *           The color to use for when in "input" mode.
    */
   @Override
   public void setPfColor(Color inputColor)
   {
      UnimplementedFeature.unsupported("PFCOLOR attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public P2JQuery getQuery()
   {
      UnimplementedFeature.unsupported("QUERY attribute is undefined");
      return null;
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   q
    *          Not used.
    */
   @Override
   public void setQuery(P2JQuery q)
   {
      UnimplementedFeature.unsupported("QUERY attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public decimal getRow()
   {
      UnimplementedFeature.unsupported("ROW attribute is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   row
    *          Not used.
    */
   @Override
   public void setRow(NumberType row)
   {
      UnimplementedFeature.unsupported("ROW attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   row
    *          Not used.
    */
   @Override
   public void setRow(double row)
   {
      UnimplementedFeature.unsupported("ROW attribute is undefined");
   }

   /**
    * Obtains the text widget that is used as the side label for this widget. This is the
    * equivalent of the SIDE-LABEL-HANDLE attribute of COMBO-BOX, EDITOR, FILL-IN, RADIO-SET,
    * SELECTION-LIST, SLIDER and TEXT widgets.
    *
    * @return   The handle to the text widget that is the side label of this widget.
    */
   @Override
   public handle getSideLabelHandle()
   {
      UnimplementedFeature.unsupported("SIDE-LABEL-HANDLE attribute is undefined");
      return new handle();
   }
   
   /**
    * Sets the text widget that is used as the side label for this widget. This is the
    * equivalent of the SIDE-LABEL-HANDLE attribute of COMBO-BOX, EDITOR, FILL-IN, RADIO-SET,
    * SELECTION-LIST, SLIDER and TEXT widgets.
    *
    * @param    label
    *           The handle to the text widget to be used as the side label of this widget.
    */
   @Override
   public void setSideLabelHandle(handle label)
   {
      UnimplementedFeature.unsupported("SIDE-LABEL-HANDLE attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public logical isScrollable()
   {
      UnimplementedFeature.unsupported("SCROLLABLE attribute is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @Override
   public void setScrollable(logical s)
   {
      UnimplementedFeature.unsupported("SCROLLABLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @Override
   public void setScrollable(boolean s)
   {
      UnimplementedFeature.unsupported("SCROLLABLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   @Override
   public logical isSelected()
   {
      UnimplementedFeature.unsupported("SELECTED attribute is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public logical isTextSelected()
   {
      UnimplementedFeature.unsupported("TEXT-SELECTED attribute is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @Override
   public void setSelected(logical s)
   {
      UnimplementedFeature.unsupported("SELECTED attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   @Override
   public void setSelected(boolean s)
   {
      UnimplementedFeature.unsupported("SELECTED attribute is undefined");
   }
   
   /**
    * Queries the SENSITIVE attribute of this widget.
    *
    * @return <code>true</code> if SENSITIVE is set.
    */
   @Override
   public logical isSensitive()
   {
      return new logical(config.enabled);
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(logical s)
   {
      setSensitive(s.booleanValue());
   }
   
   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(String s)
   {
      setSensitive(new logical(s).booleanValue());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(character s)
   {
      setSensitive(new logical(s).booleanValue());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(boolean s)
   {
      // dynamic widgets can not set sensitive attribute if not yet in a frame
      if (_dynamic() && frame == null)
      {
         StringBuilder err = new StringBuilder();
         // form the message
         err.append("Unable to process SENSITIVE attribute.  ")
            .append(type())
            .append(" widget is not in a frame");
         // and display it
         ErrorManager.displayError(4073, err.toString());
         return;
      }
      
      if (!frameLock     &&
          config.visible && 
          frame != null  && 
          frame.getFrameWidget().allFramesVisible())
      {
         // if the widget is already visible, let it draw properly
         LogicalTerminal.setSensitive(this, s);
      }
      else
      {
         if (config.enabled == s)
         {
            return;
         }

         // changing the SENSITIVE attribute does not affect the widget's visibility
         config.enabled = s;

         pushScreenDefinition();
      }
   }

   /**
    * Queries the DATA-TYPE attribute of this widget.
    *
    * @return A character value that represents the data type of the field
    *         associated with the widget.
    */
   @Override
   public character getDataType()
   {
      UnimplementedFeature.missing("DATA-TYPE attribute not implemented");
      return new character();
   }

   /**
    * Sets new value of DATA-TYPE attribute.
    *
    * @param dataType
    *        new value of DATA-TYPE attribute
    */
   @Override
   public void setDataType(character dataType)
   {
      UnimplementedFeature.missing("DATA-TYPE attribute setter not implemented");
   }

   /**
    * Sets new value of DATA-TYPE attribute.
    *
    * @param dataType
    *        new value of DATA-TYPE attribute
    */
   @Override
   public void setDataType(String dataType)
   {
      UnimplementedFeature.missing("DATA-TYPE attribute not implemented");
   }

   /**
    * Queries the TABLE attribute of this widget.
    *
    * @return The name of the database table containing the field associated
    *         with the widget.
    */
   @Override
   public character getTable()
   {
      // TODO: this returns unknown value when no name has been set, is that correct?
      //       (it probably isn't becuase the FRAME-FILE function returns "" when none if
      //       set and this seems like it should have the same backing code)
      return new character(tableName);
   }

   /**
    * Returns the state of the TABLE attribute in non-wrapped form.
    *
    * @return   Current value of the TABLE attribute or empty string if no value is set.
    */
   public String _getTable()
   {
      return (tableName == null) ? "" : tableName;
   }

   /**
    * Returns the state of the DBNAME attribute.
    *
    * @return   Current value of the DBNAME attribute.
    */
   @Override
   public character getDbName()
   {
      // TODO: this returns unknown value when no name has been set, is that correct?
      //       (it probably isn't becuase the FRAME-DB function returns "" when none if
      //       set and this seems like it should have the same backing code)
      return new character(dbName);
   }

   /**
    * Returns the state of the DBNAME attribute in non-wrapped form.
    *
    * @return   Current value of the DBNAME attribute or empty string if no value is set.
    */
   public String _getDbName()
   {
      return (dbName == null) ? "" : dbName;
   }

   /**
    * Set the value of the widget's TABLE attribute.
    * 
    * @param    table
    *           The table name (not the buffer name).
    */
   public void setTable(String table)
   {
      this.tableName = table;
   }

   /**
    * Set the value of the widget's DBNAME attribute.
    * 
    * @param    dbname
    *           The database name.
    */
   public void setDbname(String dbname)
   {
      this.dbName = dbname;
   }
   
   /**
    * Queries the TOOLTIP attribute of this widget.
    *
    * @return A help text message for a text field or text variable.
    */
   @Override
   public character getTooltip()
   {
      UnimplementedFeature.missing("TOOLTIP attribute not implemented");
      return new character();
   }

   /**
    * Sets new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   @Override
   public void setTooltip(character tooltip)
   {
      UnimplementedFeature.missing("TOOLTIP attribute not implemented");
   }

   /**
    * Sets new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   @Override
   public void setTooltip(String tooltip)
   {
      UnimplementedFeature.missing("TOOLTIP attribute not implemented");
   }

   /**
    * Queries the WIDTH-PIXELS attribute of this widget.
    *
    * @return  The width of the widget in pixels.
    */
   @Override
   public integer getWidthPixels()
   {
      UnimplementedFeature.missing("WIDTH-PIXELS attribute not implemented");
      return new integer();
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setWidthPixels(NumberType widthPixels)
   {
      UnimplementedFeature.missing("WIDTH-PIXELS attribute setter not implemented");
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setWidthPixels(int widthPixels)
   {
      UnimplementedFeature.missing("WIDTH-PIXELS attribute setter not implemented");
   }

   /**
    * Set the length of the format for the widget's @ phrases.
    * 
    * @param   length
    *          The maximum format length from the widget's @ phrases.
    */
   @Override
   public void setAtFormatLength(int length)
   {
      UnimplementedFeature.missing("@ attribute not implemented");
   }
   
   /**
    * Queries the HEIGHT-PIXELS attribute of this widget.
    *
    * @return  The height of the widget in pixels.
    */
   @Override
   public integer getHeightPixels()
   {
      UnimplementedFeature.missing("HEIGHT-PIXELS attribute not implemented");
      return new integer();
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   @Override
   public void setHeightPixels(NumberType heightPixels)
   {
      UnimplementedFeature.missing("HEIGHT-PIXELS attribute setter not implemented");
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   @Override
   public void setHeightPixels(int heightPixels)
   {
      UnimplementedFeature.missing("HEIGHT-PIXELS attribute setter not implemented");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public character getTitle() 
   {
      UnimplementedFeature.unsupported("TITLE attribute is undefined");
      return new character();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   t
    *          Not used.
    */
   public void setTitle(character t) 
   {
      UnimplementedFeature.unsupported("TITLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   t
    *          Not used.
    */
   public void setTitle(String t)
   {
      UnimplementedFeature.unsupported("TITLE attribute is undefined");
   }

   /**
    * Registers callbacks for validation processing (during editing) for any input into this
    * widget.
    * 
    * @param    valexp
    *           The <code>Function</code> to call to determine if the edits are allowed.
    * @param    valmsg
    *           The <code>Function</code> used to generate the error message on a validation
    *           failure. 
    */
   public void setValidation(ValidationExpr valexp, ValidationMsg valmsg)
   {
      this.valexp = valexp;
      this.valmsg = valmsg;
   }

   /**
    * Registers callbacks for validation processing (during editing) for any input into this
    * widget.  Use this version of the method when the message can be determined in advance.
    * 
    * @param    valexp
    *           The <code>Function</code> to call to determine if the edits are allowed.
    * @param    valmsg
    *           The error message on a validation failure.
    */
   public void setValidation(ValidationExpr valexp, String valmsg)
   {
      this.valexp = valexp;
      this.valmsg = (BaseDataType v) -> new character(valmsg);
   }

   /**
    * Checks if the widget has a validation expression registered.
    * 
    * @return   <code>true</code> if the widget will have a validation expression executed
    *           during editing.
    */
   public boolean isValidatable()
   {
      return valexp != null;
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   direction
    *          Not used.
    */
   public void setVertical(boolean direction)
   {
      UnimplementedFeature.unsupported("setVertical attribute is undefined");
   }

   /**
    * Get the VISIBLE writable attribute.
    *
    * @return   The current value of VISIBLE attribute.
    */
   public logical isVisible()
   {
      return new logical(_isVisible());
   }

   /**
    * Get the VISIBLE writable attribute.
    *
    * @return   The current value of VISIBLE attribute.
    */
   public boolean _isVisible()
   {
      return config.visible;
   }
   
   /**
    * Set the VISIBLE writable attribute.
    *
    * @param    visible
    *           The new value for the VISIBLE attribute.
    */
   public void setVisible(logical visible)
   {
      setVisible(visible.booleanValue());
   }

   /**
    * Set the VISIBLE writable attribute.
    *
    * @param    visible
    *           The new value for the VISIBLE attribute.
    */
   public void setVisible(boolean visible)
   {
      if (frame == null)
      {
         // no-op if frame is not set
         return;
      }

      // if this frame (or any of its ancestors) is not visible, force a VIEW on the widget's frame
      if (visible)
      {
         boolean forceView = !frame.getFrameWidget().allFramesVisible();
         
         if (forceView)
         {
            // the direct frame is not visible, force a view on it
            if (!frame.getFrameWidget()._isVisible())
            {
               frame.view();
            }
            else
            {
               // the direct frame is visible, do a view on the widget
               view();
            }
         }
      }
      
      if (config.visible == visible && (config.visible || config.hidden))
      {
         // be a no-op only if the widget has been realized.
         return;
      }

      _setVisible(visible);
      config.hidden = !visible;
      pushScreenDefinition();
   }
   
   /**
    * Get the SELECTABLE writable attribute.
    *
    * @return   The current value of SELECTABLE attribute.
    */
   @LegacyAttribute(name = "SELECTABLE")
   public logical isSelectable()
   {
      return new logical(config.selectable);
   }

   /**
    * Set the SELECTABLE writable attribute.
    *
    * @param value
    *        the new value of the attribute
    */
   @LegacyAttribute(name = "SELECTABLE", setter = true)
   public void setSelectable(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute SELECTABE on %s %s", 
               type(), widgetName()),
               false);
        return;
      }
      setSelectable(value.booleanValue());
   }

   /**
    * Get the SELECTABLE writable attribute.
    *
    * @param value
    *        the new value of the attribute
    */
   @LegacyAttribute(name = "SELECTABLE", setter = true)
   public void setSelectable(boolean value) 
   {
      if (config.selectable == value)
      {
         return;
      }

      config.selectable = value;
      pushScreenDefinition();
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getX()
   {
      UnimplementedFeature.unsupported("X attribute is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getFrameX()
   {
      UnimplementedFeature.unsupported("FRAME-X attribute is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    *
    * @param   x
    *          Not used.
    */
   public void setX(NumberType x)
   {
      UnimplementedFeature.unsupported("X attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getY()
   {
      UnimplementedFeature.unsupported("Y attribute is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getFrameY()
   {
      UnimplementedFeature.unsupported("FRAME-Y attribute is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    *
    * @param   y
    *          Not used.
    */
   public void setY(NumberType y)
   {
      UnimplementedFeature.unsupported("Y attribute is undefined");
   }
   
    /**
    * Blocks attempts to access this widget method.
    * 
    * @param   item
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical isSelected(character item)
   {
      UnimplementedFeature.unsupported("IS-SELECTED() method is undefined");
      return new logical();
   }
   
   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   index
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical isSelected(int64 index)
   {
      UnimplementedFeature.unsupported("IS-SELECTED() method is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public integer getNumColumns()
   {
      UnimplementedFeature.unsupported("NUM-COLUMNS attribute is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical selectAll()
   {
      UnimplementedFeature.unsupported("SELECT-ALL() method is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical deselectRows()
   {
      UnimplementedFeature.unsupported("DESELECT-ROWS() method is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical refresh()
   {
      UnimplementedFeature.unsupported("REFRESH() method is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical selectFocusedRow()
   {
      UnimplementedFeature.unsupported("SELECT-FOCUSED-ROW() method is undefined");
      return new logical();
   }

   /**
    * Gets the widget name.
    * 
    * @return   The widget name or unknown value if not configured.
    */
   public character name()
   {
      if (config == null)
      {
         return new character();
      }
      
      if (this instanceof FieldGroup || 
          this instanceof MenuItemWidget || 
          this instanceof SubMenuWidget || 
          this instanceof MenuWidget)
      {
         return super.name();
      }
      
      return name != null && !name.isUnknown() ? new character(name) : 
         frame != null ? new character(frame.getOriginalName(this)) : new character();
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    event
    *           The name of the event.
    */
   public void apply(character event)
   {
      if (event.isUnknown())
         return;
      
      apply(event.toStringMessage());
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    event
    *           The name of the event.
    */
   public void apply(String event)
   {
      LogicalTerminal.apply(frame, getId(), event);
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    eventCode
    *           The event code.
    */
   public void apply(int64 eventCode)
   {
      if (eventCode.isUnknown())
         return;
      
      apply(eventCode.longValue());
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    eventCode
    *           The event code.
    */
   public void apply(double eventCode)
   {
      LogicalTerminal.apply(frame, getId(), (int) eventCode);
   }

   /**
    * COLOR DISPLAY
    * 
    * @param   cs
    *          Color specification.
    */
   public void colorDisplay(ColorSpec cs)
   {
      setDColor(cs.convert());
   }

   /**
    * COLOR PROMPT
    * 
    * @param   cs
    *          Color specification.
    */
   public void colorPrompt(ColorSpec cs)
   {
      setPfColor(cs.convert());
   }

   /**
    * Save configured colors.
    */
   public void saveColors()
   {
      // do nothing by default
   }
   
   /**
    * Restore configured colors.
    */
   public void restoreColors()
   {
      // do nothing by default
   }
   
   /**
    * Disables this widget.
    * <p>
    * This method corresponds to the Progress statement: DISABLE widget
    */
   public void disable()
   {
      LogicalTerminal.disable(getFrame(), new int[] {getId()});
   }

   /**
    * Enables this widget.
    * <p>
    * This method corresponds to the Progress statement: ENABLE widget
    */
   public void enable()
   {
      _setVisible(true);
      LogicalTerminal.enable(getFrame(), new int[] {getId()});
   }

   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    */
   public void hide()
   {
      hide(false);
   }

   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    * 
    * @param   noPause
    *          <code>true</code> causes omission of any pauses which would
    *          be necessary otherwise.
    */
   public void hide(boolean noPause)
   {
      _setVisible(false);
      LogicalTerminal.hide(getId(), noPause, false);
   }

   /**
    * Reports if this widget object is valid to realize.
    * 
    * @return   <code>true</code> if the widget can be relized.
    */
   @Override
   public boolean preRealizeCheck()
   {
      return true;
   }
   
   /**
    * Underlines this widget on the screen.
    * <p>
    * This method corresponds to the Progress statement: UNDERLINE
    */
   public void underline()
   {
      getFrame().underline(new int[] {getId()});
   }

   /**
    * Underlines this widget on the screen.
    * <p>
    * This method corresponds to the Progress statement: UNDERLINE
    *
    * @param    stream
    *           An explicit stream which will be used for the output.
    */
   public void underline(Stream stream)
   {
      getFrame().underline(stream, new int[] {getId()});
   }

   /**
    * Shows this widget.
    * <p>
    * This method corresponds to the Progress statement: VIEW widget
    */
   public void view()
   {
      _setVisible(true);
      LogicalTerminal.view(getId());
   }
   
   /**
    * Reports if this widget object is valid for use.
    * <p>
    * The validity of the widget object is determined solely by the scope of
    * the containing frame.
    * 
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      if (deleted)
      {
         return false;
      }

      // TODO: improve this, add "deleted widget" support
      if (frame == null)
         return true;
      
      return !deleted;
   }

   /**
    * Processes validation rules (if they exist), error checking and anyother constraint
    * processing that needs to be applied.  This is used to check if a proposed update to a
    * widget will pass validation.
    *
    * @param    val
    *           The new value that is proposed for assignment to the widget.
    * @param    ignore
    *           If <code>true</code>, do not return an error when the validation expression is
    *           not defined.
    *
    * @return   <code>null</code> if validation succeeds, error message otherwise.
    */
   public String validate(BaseDataType val, boolean ignore)
   {
      String errmsg = null;
      
      if (valexp == null)
      {
         return ignore ? null : "No validation method defined";
      }

      try
      {
         errmsg = validateWorker(val);
      }
      catch (UnsupportedOperationException uoe)
      {
         errmsg = "This widget is non-editable";
      }

      return errmsg;
   }
   
   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @return   {@code true} if the validation is successful.
    */
   @Override
   public logical validateFields()
   {
      return validateFields((character) null);
   }
   
   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    val
    *           The only correct argument is "ENABLED-FIELDS" string. If specified, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   @Override
   public logical validateFields(String val)
   {
      return validateFields(new character(val)); 
   }
   
   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    val
    *           The only correct argument is "ENABLED-FIELDS" string. If specified, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   @Override
   public logical validateFields(character val)
   {
      return validateFields((val != null) && !val.isUnknown() && 
            "ENABLED-FIELDS".equalsIgnoreCase(val.toStringMessage()));
   }
   
   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    enabledOnly
    *           If <code>true</code>, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   protected logical validateFields(boolean enabledOnly)
   {
      return new logical(true);
   }
   
   /**
    * Gets the configuration object that defines the appearance and
    * functioning of this widget.
    * 
    * @return   <code>WidgetConfig</code> instance.
    */
   @Override
   public T config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(T beforeUpdate)
   {
      // no-op
   }

   /**
    * Mark widget as a header.
    * 
    * @param    header
    *           New value of the flag.
    */
   public void setHeader(boolean header)
   {
      config.setDynamicHeader(header);
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   public character getScreenValue()
   {
      // TODO: in the pending case, are there behaviors such as an uninitialized value that
      //       needs to be supported?
      return frame != null ? frame.getScreenValue(this) : pendingScreenValue;
   }
   
   /**
    * Set the current value in the screen buffer of the backing data for
    * this widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   public void setScreenValue(character value)   
   {
      // EDITOR/SELECTION-LIST/TOGGLE-BOX: setter changes the MODIFIED to TRUE, even if the
      // value doesn't change

      if (frame != null)
      {
         frame.setScreenValue(this, value);
      }
      else
      {
         pendingScreenValue = (character) value.duplicate();
      }
   }
   
   /**
    * Internal worker for getting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   initialized
    *          Flag indicating if the widget is realized or not.
    * @param   bdt
    *          The current value in the frame's buffer.
    * @param   fmt
    *          The format to be applied, in case this is a formatted value.
    * @param   ignoreFormat
    *          Flag indicating if the format needs to be ignored.
    *          
    * @return  This widget's screen-value
    */
   character getScreenValue(boolean initialized, BaseDataType bdt, String fmt, boolean ignoreFormat)
   {
      if (!initialized)
      {
         return new character("");
      }
      
      if (bdt.isUnknown())
      {
         return new character(bdt.getClass().isAssignableFrom(date.class) ? "" : "?");
      }
      
      // format the value
      // TODO: is this always the same as the way the display would format it?
      // some widgtes like selection-list can have screen value outside the format in multiple
      // selection case so we need the option to get this value too
      String formatted = ignoreFormat ? bdt.toStringMessage() : bdt.toString(fmt);
      
      // return the right-trimmed value, unless is a character
      return bdt instanceof Text ? TextOps.rightTrim(formatted) : new character(formatted.trim());

   }
   
   /**
    * Internal worker for setting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   frameBuf
    *          The frame buffer where to save the value.
    * @param   value
    *          The value to be set via SCREEN-VALUE attribute.
    * @param   inUIStmt
    *          Flag indicating this call originates from a UI statement.
    *          
    * @return  <code>true</code> if the caller can proceed, as the screen-value can be set.
    */
   boolean setScreenValue(ScreenBuffer frameBuf, Object value, boolean inUIStmt)
   {
      int widgetID = getId();
      
      frameBuf.putWidgetValue(widgetID, value);
      
      if (inUIStmt && frameBuf.isIdValid(widgetID))
      {
         config.modified = true;
      }
      
      return true;
   }

   /**
    * This method implements MOVE-TO-TOP function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveToTop()
   {
      if (LogicalTerminal.isChui() && !(this instanceof FrameWidget))
      {
         ErrorManager.recordOrShowError(4088, 
               String.format(
                  "Progress does not support the %s attribute" + 
                  " for the  %s %s in this display environment", 
                  "MOVE-TO-TOP", type(), widgetName()
               ), 
               false);
         return new logical(false);
      }

      return new logical(moveToWorker(true));
   }

   /**
    * This method implements MOVE-TO-BOTTOM function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveToBottom()
   {
      if (LogicalTerminal.isChui() && !(this instanceof FrameWidget))
      {
         ErrorManager.recordOrShowError(4088, 
               String.format(
                  "Progress does not support the %s attribute" + 
                  " for the  %s %s in this display environment", 
                  "MOVE-TO-BOTTOM", type(), widgetName()
               ), 
               false);
         return new logical(false);
      }

      return new logical(moveToWorker(false));
   }

   /**
    * This method implements MOVE-AFTER-TAB function.
    *
    * @param   widget
    *          Widget reference.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveAfterTab(CommonWidget widget)
   {
      if (!getParent().equals(widget.getParent()))
      {
         String ms = "**Unable to query attribute MOVE-AFTER-TAB-ITEM for %s %s";

         ErrorManager.recordOrThrowError(4077, String.format(ms, type(), name().getValue()), false);
         
         return new logical(false);
      }
      
      // Is the parent always field-group here?
      FieldGroup fieldGroup = (FieldGroup) getParent();
      
      return LogicalTerminal.moveAfterTabItem(fieldGroup, this, widget);
   }
   
   /**
    * This method implements MOVE-AFTER-TAB function.
    *
    * @param   widget
    *          The handle of the widget after whose tab position this will be moved.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveAfterTab(handle widget)
   {
      if (widget == null || !widget._isValid() || widget.isUnknown())
      {
         String ms = "**Attribute MOVE-BEFORE-TAB-ITEM for the %s %s was passed an invalid widget";

         ErrorManager.recordOrThrowError(4062, String.format(ms, type(), name().getValue()), false);
         
         return new logical(false);
      }
      
      return moveAfterTab(widget.unwrapWidget());
   }

   /**
    * Get this widget wrapped in a handle.
    * 
    * @return  handle to this frame.
    */
   public handle asWidgetHandle()
   {
      return new handle(this);
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    */
   public void readOnlyError(String attribute)
   {
      handle.readOnlyError(asWidgetHandle(), attribute);
   }

   /**
    * Gets the integer handle value of the window containing the widget.
    * 
    * @return   The integer handle value.
    */
   public integer getHWND()
   {
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public handle getFirstTabItem()
   {
      UnimplementedFeature.unsupported("FIRST-TAB-ITEM attribute is undefined");
      return new handle();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   h
    *          Not used.
    */
   public void setFirstTabItem(handle h)
   {
      UnimplementedFeature.unsupported("FISRT-TAB-ITEM attribute is undefined");
   }

   /**
    * Assigns the current widget to the tab position before the specified in the same field group.
    *
    * @param   widget
    *          The handle of the widget before whose we need to make new assignment.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical moveBeforeTabItem(CommonWidget widget)
   {
      if (!getParent().equals(widget.getParent()))
      {
         String ms = "**Unable to query attribute MOVE-BEFORE-TAB-ITEM for %s %s";

         ErrorManager.recordOrThrowError(4077, String.format(ms, type(), name().getValue()), false);
         
         return new logical(false);
      }
      
      // Is the parent always field-group here?
      FieldGroup fieldGroup = (FieldGroup) getParent();
      
      return LogicalTerminal.moveBeforeTabItem(fieldGroup, this, widget);
   }

   /**
    * Assigns the current widget to the tab position before the specified in the same field group.
    *
    * @param   widget
    *          The handle of the widget before whose we need to make new assignment.
    *
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical moveBeforeTabItem(handle widget)
   {
      if (widget == null || !widget._isValid() || widget.isUnknown())
      {
         String ms = "**Attribute MOVE-BEFORE-TAB-ITEM for the %s %s was passed an invalid widget";

         ErrorManager.recordOrThrowError(4062, String.format(ms, type(), name().getValue()), false);
         
         return new logical(false);
      }
      
      return moveBeforeTabItem(widget.unwrapWidget());
   }
   
   /**
    * Checks  whether the resource is dynamic or static.
    * 
    * @return   <code>true</code> if the resource is dynamic, <code>false</code> otherwise. 
    */
   @Override
   public logical dynamic()
   {
      return new logical(dynamic);
   }
   
   /**
    * Checks  whether the resource is dynamic or static.
    * 
    * @return   <code>true</code> if the resource is dynamic, <code>false</code> otherwise. 
    */
   @Override
   public boolean _dynamic()
   {
      return dynamic;
   }
   
   /**
    * Set the dynamic status of this widget.
    * 
    * @param    dynFlag
    *           New dynamic status for widget.
    */
   public void _setDynamic(boolean dynFlag)
   {
      dynamic = dynFlag;
   }
   
   /**
    * Get the FONT attribute of this widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getFont()
   {
      UnimplementedFeature.unsupported("FONT attribute is undefined");
      return new integer();
   }

   /**
    * Set the FONT attribute of this widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setFont(long fontNum)
   {
      setFont(new int64(fontNum));
   }

   /**
    * Set the FONT attribute of this widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setFont(int64 fontNum)
   {
      UnimplementedFeature.unsupported("FONT attribute is undefined");
   }

   /**
    * Get the COLUMN-FONT attribute of this browse column widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getColumnFont()
   {
      UnimplementedFeature.unsupported("COLUMN-FONT attribute is undefined");
      return new integer();
   }
   
   /**
    * Sets the NO-AUTO-VALIDATE option.
    *
    * @param noAutoValidate
    *        <code>true</code> if the option is set to ON.
    */
   public void setNoAutoValidate(boolean noAutoValidate)
   {
      UnimplementedFeature.missing("NO-AUTO-VALIDATE option not implemented.");
   }   
   
   /**
    * Set the COLUMN-FONT attribute of this browse column widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setColumnFont(long fontNum)
   {
      setColumnFont(new int64(fontNum));
   }

   /**
    * Set the COLUMN-FONT attribute of this browse column widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setColumnFont(int64 fontNum)
   {
      UnimplementedFeature.unsupported("COLUMN-FONT attribute is undefined");
   }

   /**
    * Get the LABEL-FONT attribute of this browse column widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getLabelFont()
   {
      UnimplementedFeature.unsupported("LABEL-FONT attribute is undefined");
      return new integer();
   }

   /**
    * Set the LABEL-FONT attribute of this browse column widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setLabelFont(long fontNum)
   {
      setLabelFont(new int64(fontNum));
   }

   /**
    * Set the LABEL-FONT attribute of this browse column widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setLabelFont(int64 fontNum)
   {
      UnimplementedFeature.unsupported("LABEL-FONT attribute is undefined");
   }

   /**
    * Get the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getTitleFont()
   {
      UnimplementedFeature.unsupported("TITLE-FONT attribute is undefined");
      return new integer();
   }

   /**
    * Set the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setTitleFont(long fontNum)
   {
      setTitleFont(new int64(fontNum));
   }

   /**
    * Set the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setTitleFont(int64 fontNum)
   {
      UnimplementedFeature.unsupported("TITLE-FONT attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   p
    *          Not used.
    */
   @Override
   public void setParent(GenericWidget<?> p)
   {
      UnimplementedFeature.unsupported("PARENT method is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    */
   @Override
   public logical isThreeD()
   {
      UnimplementedFeature.unsupported("THREE-D method is undefined");
      return new logical();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   threeD
    *          Not used.
    */
   @Override
   public void setThreeD(logical threeD)
   {
      UnimplementedFeature.unsupported("THREE-D method is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   threeD
    *          Not used.
    */
   @Override
   public void setThreeD(boolean threeD)
   {
      UnimplementedFeature.unsupported("THREE-D method is undefined");
   }
   /**
    * Blocks attempts to access this attribute.
    */
   @Override
   public decimal getVirtHeight()
   {
      UnimplementedFeature.unsupported("VIRTUAL-HEIGHT-CHARS method is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   height
    *          Not used.
    */
   @Override
   public void setVirtHeight(NumberType height)
   {
      UnimplementedFeature.unsupported("VIRTUAL-HEIGHT-CHARS method is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    */
   @Override
   public decimal getVirtWidth()
   {
      UnimplementedFeature.unsupported("VIRTUAL-WIDTH-CHARS method is undefined");
      return new decimal();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   width
    *          Not used.
    */
   @Override
   public void setVirtWidth(NumberType width)
   {
      UnimplementedFeature.unsupported("VIRTUAL-WIDTH-CHARS method is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    */
   @Override
   public integer getVirtHeightPixels()
   {
      UnimplementedFeature.unsupported("VIRTUAL-HEIGHT-PIXELS method is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   height
    *          Not used.
    */
   @Override
   public void setVirtHeightPixels(NumberType height)
   {
      UnimplementedFeature.unsupported("VIRTUAL-HEIGHT-PIXELS method is undefined");
   }
   
   /**
    * Get widget POPUP-MENU.
    *
    * @return  The pop-up menu handle.
    */
   @Override
   public handle getPopupMenu()
   {
      UnimplementedFeature.missing("POPUP-MENU getter is not implemented");
      return new handle();
   }
   
   /**
    * Set widget's POPUP-MENU.
    * 
    * @param  menu
    *         The menu handle.
    */
   @Override
   public void setPopupMenu(handle menu)
   {
      if (menu != null && !menu.isUnknown())
      {
         setPopupMenu((MenuWidget) menu.getResource());
      }
      else 
      {
         setPopupMenu((MenuWidget)null);
      }
   }
   
   /**
    * Set widget's POPUP-MENU.
    * 
    * @param  menu
    *         The menu widget
    */
   @Override
   public void setPopupMenu(MenuWidget menu)
   {
      if (!menu.isPopupOnly())
      {
         // TODO check
         //String msg = "The widget type assigned to WINDOW widget must be a MENUBAR";
         //ErrorManager.recordOrShowError(4092, msg, false);
      }
      
      this.config.popupMenuId = menu.getId();
      this.menu = menu;
      this.menu.setOwner(this);
      pushScreenDefinition();
   }
   
   /**
    * Blocks attempts to access this attribute.
    */
   @Override
   public integer getVirtWidthPixels()
   {
      UnimplementedFeature.unsupported("VIRTUAL-WIDTH-PIXELS method is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   width
    *          Not used.
    */
   @Override
   public void setVirtWidthPixels(NumberType width)
   {
      UnimplementedFeature.unsupported("VIRTUAL-WIDTH-PIXELS method is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    */
   @Override
   public integer getBgColor()
   {
      UnimplementedFeature.unsupported("BGCOLOR method is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   bgcolor
    *          Not used.
    */
   @Override
   public void setBgColor(integer bgcolor)
   {
      UnimplementedFeature.unsupported("BGCOLOR method is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    */
   @Override
   public integer getFgColor()
   {
      UnimplementedFeature.unsupported("FGCOLOR method is undefined");
      return new integer();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   fgcolor
    *          Not used.
    */
   @Override
   public void setFgColor(integer fgcolor)
   {
      UnimplementedFeature.unsupported("FGCOLOR method is undefined");
   }
   
   /**
    * Set value of the TAB-STOP attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setTabStop(boolean value)
   {
      UnimplementedFeature.unsupported("TAB-STOP attribute is undefined");
   }

   /**
    * Set value of the TAB-STOP attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setTabStop(logical value)
   {
      UnimplementedFeature.unsupported("TAB-STOP attribute is undefined");
   }

   /**
    * Get value of the TAB-STOP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isTabStop()
   {
      UnimplementedFeature.unsupported("TAB-STOP attribute is undefined");
      return new logical();
   }

   /**
    * Returns a flag indicating whether the widget has been realized.
    * 
    * @return  See above.
    */
   public boolean _isRealized()
   {
      return config.realized;
   }

   /**
    * Sets the WIDGET-ID writable attribute.
    *
    * @param    id
    *           The new value for the WIDGET-ID attribute.
    */
   @Override
   public void setWidgetId(integer id)
   {
      if (id == null || id.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute WIDGET-ID on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setWidgetId(id.intValue());
   }
   
   /**
    * Check if the widget's layout is valid
    *   
    * @return
    *       <code>true</code> if the layout is OK.
    *
    */
   public boolean validateLayout()
   {
      return true;
   }
   
   /**
    * Sets the WIDGET-ID writable attribute.
    *
    * @param    id
    *           The new value for the WIDGET-ID attribute.
    */
   @Override
   public void setWidgetId(int id)
   {
      // NOOP, 4GL doesn't validate the value
      // if -usewidgetid is NOT used (e.g. in ChUI), it is just ignored
   }

   /**
    * Gets the WIDGET-ID writable attribute.
    *
    * @return
    *          the widget id
    */
   @Override
   public integer getWidgetId() 
   {
      return new integer();
   }

   /**
    * Sets the CONTEXT-HELP-ID writable attribute.
    *
    * @param    id
    *           The new value for the CONTEXT-HELP-ID attribute.
    */
   @Override
   public void setContextHelpId(integer id)
   {
      if (id == null || id.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute CONTEXT-HELP-ID on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setContextHelpId(id.intValue());
   }
   
   /**
    * Sets the WIDGET-ID writable attribute.
    *
    * @param    id
    *           The new value for the CONTEXT-HELP-ID attribute.
    */
   @Override
   public void setContextHelpId(int id)
   {
      if (config.contextHelpId == id)
      {
         return;
      }
      
      config.contextHelpId = id;
      pushScreenDefinition();
   }

   /**
    * Gets the CONTEXT-HELP-ID writable attribute.
    *
    * @return
    *          the widget context help id
    */
   @Override
   public integer getContextHelpId()
   {
      return new integer(config.contextHelpId);
   }

   /**
    * Gets the FRAME-ROW attribute.
    *
    * @return
    *          the value of the attribute
    */
   @Override
   public decimal getFrameRow()
   {
      UnimplementedFeature.unsupported("FRAME-ROW attribute is undefined");
      return new decimal();
   }

   /**
    * Set value of the MANUAL-HIGHLIGHT attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setManualHighlight(boolean value)
   {
      if (config.manualHightlight == value)
      {
         return;
      }
      
      config.manualHightlight = value;
      pushScreenDefinition();
   }

   /**
    * Set value of the BLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setManualHighlight(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute MANUAL-HIGHLIGHT on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setManualHighlight(value.booleanValue());
   }

   /**
    * Get value of the MANUAL-HIGHLIGHT attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isManualHighlight()
   {
      return new logical(config.manualHightlight);
   }

   /**
    * Sets the MENU-MOUSE writable attribute.
    *
    * @param    value
    *           The new value for the MENU-MOUSE attribute.
    */
   @LegacyAttribute(name = "MENU-MOUSE", setter = true)
   public void setMenuMouse(int value)
   {
      if (value < 0)
      {
         ErrorManager.recordOrShowError(4082, 
               String.format(
               "Attribute MENU-MOUSE must be greater than or equal to 0 on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      if (value > 3)
      {
         ErrorManager.recordOrShowError(4103, 
               String.format(
               "Unknown error code 1 for attribute MENU-MOUSE on %s %s", 
               type(), widgetName()),
               false);
         ErrorManager.recordOrShowError(3131, 
               String.format(
               "Unable to set attribute MENU-MOUSE on widget %s of type %s", 
               widgetName(), type()),
               false);
         return;
         
      }
      
      Integer menuMouse = (value == 0) ? null : Integer.valueOf(value);
      if (Utils.equals(config.menuMouse, menuMouse))
      {
         return;
      }
      
      config.menuMouse = menuMouse;
      pushScreenDefinition();
   }

   /**
    * Sets the MENU-MOUSE writable attribute.
    *
    * @param    value
    *           The new value for the MENU-MOUSE attribute.
    */
   @LegacyAttribute(name = "MENU-MOUSE", setter = true)
   public void setMenuMouse(int64 value)
   {
      if (value == null || value.isUnknown())
      {
         if (config.menuMouse != null)
         {
            config.menuMouse = null;
            pushScreenDefinition();
            return;
         }
      }
      setMenuMouse(value.intValue());
   }

   /**
    * Gets the MENU-MOUSE attribute.
    *
    * @return
    *          the value of the attribute
    */
   @LegacyAttribute(name = "MENU-MOUSE")
   public integer getMenuMouse()
   {
      return new integer(config.menuMouse);
   }

   /**
    * Sets the MENU-KEY writable attribute.
    *
    * @param    value
    *           The new value for the MENU-KEY attribute.
    */
   @LegacyAttribute(name = "MENU-KEY", setter = true)
   public void setMenuKey(String value)
   {
      if (Utils.equals(config.menuKey, value))
      {
         return;
      }
      
      config.menuKey = value;
      pushScreenDefinition();
   }

   /**
    * Sets the MENU-KEY writable attribute.
    *
    * @param    value
    *           The new value for the MENU-KEY attribute.
    */
   @LegacyAttribute(name = "MENU-KEY", setter = true)
   public void setMenuKey(character value)
   {
      setMenuKey(value == null || value.isUnknown() ? null : value.toStringMessage());
   }

   /**
    * Gets the MENU-KEY attribute.
    *
    * @return
    *          the value of the attribute
    */
   @LegacyAttribute(name = "MENU-KEY")
   public character getMenuKey()
   {
      return new character(config.menuKey);
   }

   /**
    * Gets the WINDOW attribute.
    *
    * @return
    *          the value of the attribute
    */
   @Override
   public handle getWindow()
   {
      return frame == null ? new handle() : 
         LogicalTerminal.getWidgetHandleById(frame.asWidget().config().windowID);
   }
   
   /**
    * Getter for RESIZABLE attribute.
    *
    * @return  {@code true} if widget is resizable.
    */
   @Override
   public logical isResizable()
   {
      //UnimplementedFeature.missing("RESIZABLE attribute is not implemented");
      return new logical();
   }
   
   /**
    * Setter for RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} to make the widget resizable
    */
   @Override
   public void setResizable(boolean on)
   {
      setResizable(new logical(on));
   }
   
   /**
    * Setter for RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} to make the widget resizable
    */
   @Override
   public void setResizable(logical on)
   {
      //UnimplementedFeature.missing("RESIZABLE attribute is not implemented");
   }
   
   /**
    * The implementation of LOAD-MOUSE-POINTER() method. Specifies the mouse pointer to display
    * when the pointer is moved over the widget.
    *
    * @param pointerName
    *        The name of the mouse cursor. Here is the list of all cursor names:
    *        <table border='1' summary="">
    *          <tr><td>APPSTARTING</td><td>Arrow with an hourglass beside it</td></tr>
    *          <tr><td>ARROW</td><td>Standard arrow cursor</td></tr>
    *          <tr><td>CROSS</td><td>Cross hairs</td></tr>
    *          <tr><td>HELP</td><td>Arrow with a question mark beside it</td></tr>
    *          <tr><td>IBEAM</td><td>I-beam text cursor</td></tr>
    *          <tr><td>NO</td><td>Circle with a slash through it</td></tr>
    *          <tr><td>RECTANGLE (NT 3.51 only)</td><td>White rectangle</td></tr>
    *          <tr><td>SIZE</td><td>Sizing rectangle</td></tr>
    *          <tr><td>SIZE-E</td><td>Size to right</td></tr>
    *          <tr><td>SIZE-N</td><td>Size to top</td></tr>
    *          <tr><td>SIZE-NE</td><td>Size to top right</td></tr>
    *          <tr><td>SIZE-NW</td><td>Size to top left</td></tr>
    *          <tr><td>SIZE-S</td><td>Size to bottom</td></tr>
    *          <tr><td>SIZE-SE</td><td>Size to bottom right</td></tr>
    *          <tr><td>SIZE-SW</td><td>Size to bottom left</td></tr>
    *          <tr><td>SIZE-W</td><td>Size to left</td></tr>
    *          <tr><td>UPARROW</td><td>Up arrow</td></tr>
    *          <tr><td>WAIT</td><td>System busy</td></tr>
    *          <tr><td>GLOVE</td><td>Glove/finger</td></tr>
    *          <tr><td>COMPILER-WAIT</td><td>Compiler busy</td></tr>
    *        </table>
    *
    * @return {@code true} if the mouse pointer is loaded successfully.
    */
   @Override
   public logical loadMousePointer(String pointerName)
   {
      if (!(this instanceof Hoverable))
      {
         ErrorManager.recordOrShowError(4082, String.format(
               "LOAD-MOUSE-POINTER is not a queryable attribute for %s %s", 
               type(), widgetName()),
               false);
         return new logical();
      }

      Hoverable h = (Hoverable)this;

      // Check is the cursor name is a pre-defined one or null
      String ptrName = pointerName == null || "".equals(pointerName.trim()) ? null : 
         StringHelper.safeTrimTrailing(pointerName);
      
      if (!LogicalTerminal.loadMousePointer(getId(), ptrName))
      {
         ErrorManager.recordOrShowError(4447, String.format(
               "Unable to open file for %s %s", 
               type(), widgetName()),
               false);
         return new logical(false);
      }
      h.setMousePointerName(ptrName);
      return new logical(true);
   }
   
   /**
    * The implementation of LOAD-MOUSE-POINTER() method. Specifies the mouse pointer to display
    * when the pointer is moved over the widget.
    *
    * @param pointerName
    *        The name of the mouse cursor. Here is the list of all cursor names:
    *        <table border='1' summary="">
    *          <tr><td>APPSTARTING</td><td>Arrow with an hourglass beside it</td></tr>
    *          <tr><td>ARROW</td><td>Standard arrow cursor</td></tr>
    *          <tr><td>CROSS</td><td>Cross hairs</td></tr>
    *          <tr><td>HELP</td><td>Arrow with a question mark beside it</td></tr>
    *          <tr><td>IBEAM</td><td>I-beam text cursor</td></tr>
    *          <tr><td>NO</td><td>Circle with a slash through it</td></tr>
    *          <tr><td>RECTANGLE (NT 3.51 only)</td><td>White rectangle</td></tr>
    *          <tr><td>SIZE</td><td>Sizing rectangle</td></tr>
    *          <tr><td>SIZE-E</td><td>Size to right</td></tr>
    *          <tr><td>SIZE-N</td><td>Size to top</td></tr>
    *          <tr><td>SIZE-NE</td><td>Size to top right</td></tr>
    *          <tr><td>SIZE-NW</td><td>Size to top left</td></tr>
    *          <tr><td>SIZE-S</td><td>Size to bottom</td></tr>
    *          <tr><td>SIZE-SE</td><td>Size to bottom right</td></tr>
    *          <tr><td>SIZE-SW</td><td>Size to bottom left</td></tr>
    *          <tr><td>SIZE-W</td><td>Size to left</td></tr>
    *          <tr><td>UPARROW</td><td>Up arrow</td></tr>
    *          <tr><td>WAIT</td><td>System busy</td></tr>
    *          <tr><td>GLOVE</td><td>Glove/finger</td></tr>
    *          <tr><td>COMPILER-WAIT</td><td>Compiler busy</td></tr>
    *        </table>
    *
    * @return {@code true} if the mouse pointer is loaded successfully.
    */
   @Override
   public logical loadMousePointer(character pointerName)
   {
      return loadMousePointer(pointerName == null || pointerName.isUnknown() ? "" : 
            pointerName.toStringMessage());
   }
   

   /**
    * Get the value of the MOUSE-POINTER attribute for a widget
    *
    * @return   the value of the attribute 
    */
   @Override
   public character getMousePointer()
   {
      if (!(this instanceof Hoverable))
      {
         ErrorManager.recordOrShowError(4082, String.format(
               "LOAD-MOUSE-POINTER is not a queryable attribute for %s %s", 
               type(), widgetName()),
               false);
         return new character();
      }
      String ptrName = ((Hoverable)this).getMousePointerImpl();
      return new character(ptrName == null ? "" : ptrName);
   }
   
   /**
    * Reads the INPUT-VALUE attribute (value of the widget currently in the screen-buffer).
    * TODO: should this implementation should be moved/overridden in ControlTextWidget ? 
    *
    * @return   The input value of the widget of the same data type as the field or variable 
    *           associated with the widget.
    */
   @Override
   public BaseDataType inputValue()
   {
      // TODO: this is a basic implementation that maps directly to the widget's getter, we need
      //       to fully explore if this is correct and resolve any compatibility issues
      
      Class<?> type = null;
      
      // TODO: does this check cover exactly the set of widgets that support this attribute?
      if (this instanceof ControlEntity<?>)
      {
         type = ((ControlEntity<?>) this).getDataClass();
      }
      
      // TODO: type == null is probably a scenario where there should be a 4GL error because 
      //       input-value won't work on widgets that don't represent data
      
      if (frame != null && type != null)
      {
         // TODO: not sure if the last 2 parameters are correct
         return (BaseDataType) frame.getter(getId(), type, true, false);
      }
      
      // TODO: without a frame attached, it is still possible to assign the screen-value attribute,
      //       that caching implementation will be added in the future, when it is available
      //       check if that value needs to be returned 
      return new unknown();
   }
   
   /**
    * Check it the requested new label size is OK for given widget.
    * 
    * @param label
    *        the new label to check
    * 
    * @return  the possibly modified label string that fits particular widget constraints.
    */
   protected String labelSizeCheck(String label)
   {
      // by default it does nothing
      return label;
   }

   /**
    * Process the widget's side-label (if it can have one).
    *
    * @param    explicit
    *           <code>true</code> if this label is explicitly set via the LABEL attribute.
    * 
    * @return   The side-label widget's ID or <code>-1</code> if none exists.
    */
   protected int processSideLabel(boolean explicit)
   {
      return -1;
   }
   
   /**
    * Check if this resource supports the NAME attribute in read-only mode.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      // TODO: not all widgets have writable name
      return false;
   }
   
   /**
    * Check if this resource has as parent the specified resource.
    *
    * @param    parent
    *           The parent for which the first child is needed.  If <code>null</code>, the first
    *           resource with no parent set is returned.
    *           
    * @return   <code>true</code> if the parent is <code>null</code> or if the parent has the ID 
    *           as set in {@link BaseConfig#parentId}.
    */
   @Override
   protected boolean hasParent(HandleChain parent)
   {
      if (parent != null && !(parent instanceof GenericWidget))
      {
         return false;
      }
      
      GenericWidget<?> wp = (GenericWidget<?>) parent;
      
      if (config instanceof BaseConfig)
      {
         int parentId = ((BaseConfig) config).parentId;
         
         // check if the parent matches
         return (parentId == -1 && wp == null) || (wp != null && parentId == wp.getId());
      }
      
      // no parenting is required
      return true;
   }

   /**
    * Check if this resource can be reported as chained.
    * 
    * @return   <code>true</code> if the widget is attached to a frame.
    */
   @Override
   protected boolean isChained()
   {
      return config.frameId != -1;
   }
   
   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!canDelete())
      {
         return false;
      }

      // a static widget can be deleted ONLY if the parent frame is not visible, regardless if 
      // forced or not
      if (!_dynamic() && frame != null  && 
          frame.asWidget()._isVisible() && 
          !frame.asWidget().deleted)
      {
         // widgets in a shared frame inherited from master can not be deleted here. from a shared  
         // frame, only widgets which are defined only at the shared frame are allowed to pass.
         return false;
      }
      
      // removing dynamic widget from associated frame
      if (_dynamic())
      {
         if (frame != null)
         {
            frame.deleteDynamicWidget(this);
         }
         else
         {
            LogicalTerminal.deregisterWidget(getId());
         }
      }
      
      // remove the config for this widget
      ConfigManager.getInstance().removeWidgetConfig(config.id);
      
      deleted = true;
      
      return true;
   }
   
   /**
    * Check if the resource may be implicitly deleted. 
    * 
    * @return   <code>true</code> if the INSTANTIATING-PROCEDURE attribute is unknown (i.e. the 
    *           instantiating procedure has been deleted) and we are during finalizable or
    *           scopeable notification.
    *           <code>true</code> if the frame can be deleted.
    */
   @Override
   protected boolean implicitDeletion()
   {
      return super.implicitDeletion() || (frame != null && frame.asWidget().canDelete());
   }
   
   /**
    * Check if this resource can be deleted (implicitly or not).
    * 
    * @return   See above.
    */
   protected boolean canDelete()
   {
      // the !implicitDeletion() check must be made only for static widgets, so the !dynamic check
      // must be executed before !implicitDeletion()
      if (!dynamic && !implicitDeletion())
      {
         // static widgets can't be deleted
         final String errmsg = 
            "Static handle type %s given to DELETE PROCEDURE or DELETE OBJECT statement";
         ErrorManager.recordOrThrowError(5428, String.format(errmsg, type()), false);

         return false;
      }
      
      return true;
   }
   
   /**
    * Conditionally call {@link GenericFrame#pushScreenDefinition()}, depending if the frame was
    * set or not for this widget.
    */
   protected void pushScreenDefinition()
   {
      // For window widgets we should push window configuration 
      if (this instanceof WindowWidget && config instanceof WindowConfig)
      {
         LogicalTerminal.pushWindow((WindowConfig) config);
      }
      else
      {
         if (frame != null && !frameLock)
         {
            // defer this call if the frame is locked
            frame.pushScreenDefinition();
         }
      }
   }
   
   /**
    * Stub implementation of the 'realize' method.
    * 4GL can realize widgets on access to some of its attributes
    */
   @Override
   public void realize()
   {
      if (config.wasRealized)
      {
         return;
      }
      if (frame == null)
      {
         ErrorManager.recordOrShowError(4040, String.format(
               "Unable to realise %s widget because it is not in a frame", 
               type()),
               false);
         
      }
      /** 
       * Set the auxiliary flag to indicate whether the widget's has been realized.
       *  A temporary solution. Most likely should be revised
       *  after a proper realization of the realize() method, 
       */
      config.wasRealized = true;
      // TODO: implement
   }
   
   /**
    * Finish initialization of the widget. 
    */
   public void finishSetup()
   {
      // Default implementations does nothing
   }

   /**
    * Cleanup the dynamic widget before deletion.
    */
   public void cleanup()
   {
   }

   /**
    * Check if can access attribute
    * 
    * @param attr
    *        name of the attribute
    *        
    * @return <code>true</code> if the attrubute can be accessed
    */
   protected boolean canAccess(String attr)
   {
      realize();
      if (!config.wasRealized)
      {
         ErrorManager.recordOrShowError(4104, 
               String.format(
               "Unknown error code 1 for attribute ? on the %s %s", 
               type(), widgetName()),
               false);
         if (attr != null)
         {
            ErrorManager.recordOrShowError(4077, 
                  String.format(
                  "Unable to query attribute %s for %s %s", 
                  attr, type(), widgetName()),
                  false);
         }
         return false;
      }
      return true;
   }

   /**
    * Get widget name for the error message
    * 
    * @return
    *       widget name or 'widget' string for dynamic widgets
    */
   protected String widgetName()
   {
      return dynamic ? "widget" : frame.getName(this);
   }

   /**
    * Report that the attribute is not settable
    * 
    * @param attr
    *        the name of the attribute
    */
   protected void notSettable(String attr)
   {
      // not all cases have been checked, but those that have been checked do not
      // set the error flag
      notWorker(false, attr);
   }

   /**
    * Report that the attribute is not queryable
    * 
    * @param attr
    *        the name of the attribute
    */
   protected void notQueryable(String attr)
   {
      // not all cases have been checked, but those that have been checked do not
      // set the error flag
      notWorker(true, attr);
   }

   /**
    * Report that the attribute is not queryable or settable.
    *
    * @param   qry
    *          <code>true</code> for "queryable", else "setable" (yes, that is the
    *          spelling in Progress).
    * @param   attr
    *          The name of the attribute for the error message.
    */
   protected void notWorker(boolean qry, String attr)
   {
      String[] msg = new String[] { String.format("%s is not a %s attribute for %s %s", 
                                                  attr,
                                                  qry ? "queryable" : "setable",
                                                  type(),
                                                  widgetName()) };
      
      ErrorManager.recordOrShowError(new int[] { 4052 }, msg, false, true, false, false, true);
   }

   /**
    * Finish the config processing for this widget.
    * <p>
    * The default implementation is a no-op.
    */
   protected void finishConfigProcessing()
   {
      // no-op
   }
   
   /**
    * Lock or the the specified handle and any attribute assignments are cached until the
    * {@link DynamicWidgetFactory#pushScreenUnlock lock} is released.
    * 
    * @param    lock
    *           <code>true</code> to set the lock or <code>false</code> to unset it and push any
    *           changes to the client-side (if and only if the frame has been set).
    */
   protected void setScreenLock(boolean lock)
   {
      frameLock = lock;

      if (!lock && frame != null)
      {
         // if the lock is unset, push any changes to the client-side
         if (frame.isBatch())
         {
            // if frame was set in batch mode, remove it and push everything to client-side
            frame.batch(false, true);
         }
         else
         {
            // otherwise, just perform a simple call
            pushScreenDefinition();
         }
      }

      // nothing to do if the lock is set
   }
   
   /**
    * Worker for the {@link #moveToTop()} and {@link #moveToBottom()} code.
    * 
    * @param    top
    *           <code>true</code> if moving to top, <code>false</code> for bottom.
    * 
    * @return   <code>true</code> if move succeeded.
    */
   protected boolean moveToWorker(boolean top)
   {
      return top ? LogicalTerminal.moveToTop(config.id.asInt())
                 : LogicalTerminal.moveToBottom(config.id.asInt());
   }
   
   /**
    * Sets the widget's config {@link WidgetConfig#visible visibility flag}.
    * 
    * @param    visible
    *           The widget's visibility state.
    */
   protected void _setVisible(boolean visible)
   {
      config.visible = visible;
   }

   /**
    * Sets the label text.
    * 
    * @param    label
    *           The new label text.
    */
   protected void setLabelInt(String label)
   {
      // check if the label can fit the frame
      label = labelSizeCheck(label);
      
      if (frame == null || !frame.isViewed())
      {
         // force setting of frame
         config.setDynamicLabel(null);
      }

      config.setDynamicLabel(label);
      
      processSideLabel(true);
      
      pushScreenDefinition();
   }
   
   /**
    * Processes validation rules, error checking and any other user-defined constraint processing
    * that needs to be applied.  This is used to check if a proposed update to an object will
    * pass validation. If the method silently returns then all tests have passed, otherwise an
    * exception will be thrown on any validation failure. Any error condition raised during the
    * processing of the validation expression will be ignored (but will be considered a
    * validation failure).  It is expected that any error message output should already have been
    * handled by the code that generated it.  The exception is ignored here so as to not bypass
    * the processing of the validation message evaluation AND so that this code doesn't change
    * the control flow of the application.
    *
    * @param    newVal
    *           The new value that is proposed for assignment to the widget.
    *
    * @return   <code>null</code> if validation succeeded or an error message on any failure.
    */
   private String validateWorker(BaseDataType newVal)
   {
      String errmsg = null;
      
      boolean success = true;
      
      try
      {
         success = valexp.validate(newVal).booleanValue();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      
      if (!success)
      {
         errmsg = valmsg.message(newVal).toStringMessage();
      }
      
      return errmsg;
   }
}