/*
** Module   : ImageWidget.java
** Abstract : server side image widget implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20141208 Created initial version.
** 002 EVL 20150115 Rework for image support implementation.
** 003 EVL 20150128 Image widget now implements image loading interface.
** 004 EVL 20150130 Adding image phrase conversion and runtime support.
** 005 EVL 20150216 Adding TRANSPARENT, STRETCH-TO-FIT and RETAIN-SHAPE options support.
** 006 EVL 20150222 Adding boolean option getters for CONVERT-3D-COLORS, TRANSPARENT,
**                  STRETCH-TO-FIT and RETAIN-SHAPE.
** 007 HC  20150323 WidgetId.getId() renamed to asInt().
** 008 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Server side image widget implementation.
 */
@LegacyResource(resource = LegacyResource.IMAGE)
public class ImageWidget
extends BaseEntity<ImageConfig>
implements ImageWidgetSupport
{
   /**
    * Default constructor.
    */
   public ImageWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public ImageWidget(boolean dynamic)
   {
      super(dynamic, new ImageConfig());
   }
   
   /**
    * Returns the name of the image loaded by LOAD-IMAGE. Corresponds Progress
    * IMAGE attribute
    *
    * @return the name of the image loaded by LOAD-IMAGE. 
    */
   public character getImage()
   {
      return new character(config.image);
   }

   /**
    * Set button image file name.
    * 
    * @param   name
    *          Image file name.
    */
   public void setImage(String name)
   {
      loadImage(name);
   }
   
   /**
    * Set image tooltip.
    * 
    * @param   text
    *          Tooltip text
    */
   public void setTooltip(String text)
   {
      if (Utils.equals(config.tooltip, text))
      {
         return;
      }
               
      config.tooltip = text;
      pushScreenDefinition();
   }

   /**
    * Get image tooltip.
    * 
    * @return  image tooltip.
    */
   public character getTooltip()
   {
      return new character(config.tooltip);
   }
   
   /**
    * Set CONVERT-3D-COLORS option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setConvert3D(boolean flag)
   {
      if (config.convert3D == flag)
      {
         return;
      }
      
      config.convert3D = flag;
      pushScreenDefinition();
   }

   /**
    * Set CONVERT-3D-COLORS option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setConvert3D(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setConvert3D(flag.booleanValue());
      }
   }

   /**
    * Gets CONVERT-3D-COLORS option.
    * 
    * @return  CONVERT-3D-COLORS current value.
    */
   @Override
   public logical getConvert3D()
   {
      return new logical(config.convert3D);
   }

   /**
    * Set TRANSPARENT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setTransparent(boolean flag)
   {
      if (config.transparent == flag)
      {
         return;
      }
      
      config.transparent = flag;
      pushScreenDefinition();
   }

   /**
    * Set TRANSPARENT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setTransparent(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setTransparent(flag.booleanValue());
      }
   }

   /**
    * Gets TRANSPARENT option.
    * 
    * @return  TRANSPARENT current value.
    */
   @Override
   public logical getTransparent()
   {
      return new logical(config.transparent);
   }

   /**
    * Set STRETCH-TO-FIT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setStretchToFit(boolean flag)
   {
      if (config.stretchToFit == flag)
      {
         return;
      }
      
      config.stretchToFit = flag;
      pushScreenDefinition();
   }

   /**
    * Set STRETCH-TO-FIT option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setStretchToFit(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setStretchToFit(flag.booleanValue());
      }
   }

   /**
    * Gets STRETCH-TO-FIT option.
    * 
    * @return  STRETCH-TO-FIT current value.
    */
   @Override
   public logical getStretchToFit()
   {
      return new logical(config.stretchToFit);
   }

   /**
    * Set RETAIN-SHAPE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setRetainShape(boolean flag)
   {
      if (config.retainShape == flag)
      {
         return;
      }
      
      config.retainShape = flag;
      pushScreenDefinition();
   }

   /**
    * Set RETAIN-SHAPE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setRetainShape(logical flag)
   {
      if (flag != null && !flag.isUnknown())
      {
         setRetainShape(flag.booleanValue());
      }
   }

   /**
    * Gets RETAIN-SHAPE option.
    * 
    * @return  RETAIN-SHAPE current value.
    */
   @Override
   public logical getRetainShape()
   {
      return new logical(config.retainShape);
   }

   /**
    * Sets new value of X-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of X-offset in pixels.
    */
   public void setImageOffsetX(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetX(offset.intValue());
      }
   }

   /**
    * Sets new value of X-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of X-offset in pixels.
    */
   public void setImageOffsetX(int offset)
   {
      config.offsetX = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of Y-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of Y-offset in pixels.
    */
   public void setImageOffsetY(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetY(offset.intValue());
      }
   }

   /**
    * Sets new value of Y-offset image phrase option in pixel units.
    *
    * @param offset
    *        new value of Y-offset in pixels.
    */
   public void setImageOffsetY(int offset)
   {
      config.offsetY = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of ROW-offset image phrase option in character units.
    *
    * @param offset
    *        new value of ROW-offset in characters.
    */
   public void setImageOffsetRow(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetRow(offset.doubleValue());
      }
   }

   /**
    * Sets new value of ROW-offset image phrase option in character units.
    *
    * @param offset
    *        new value of ROW-offset in characters.
    */
   public void setImageOffsetRow(double offset)
   {
      config.offsetRow = offset;
      pushScreenDefinition();
      return;
   }

   /**
    * Sets new value of COLUMN-offset image phrase option in character units.
    *
    * @param offset
    *        new value of COLUMN-offset in characters.
    */
   public void setImageOffsetColumn(NumberType offset)
   {
      if (!offset.isUnknown())
      {
         setImageOffsetColumn(offset.doubleValue());
      }
   }

   /**
    * Sets new value of COLUMN-offset image phrase option in character units.
    *
    * @param offset
    *        new value of COLUMN-offset in characters.
    */
   public void setImageOffsetColumn(double offset)
   {
      config.offsetCol = offset;
      pushScreenDefinition();
   }

   /**
    * Load image from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(String name)
   {
      return new logical(loadImageInt(name));
   }

   /**
    * Load image from specified file name.
    * 
    * @param   name
    *          Image file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadImage(character name)
   {
      return name != null && !name.isUnknown() ? loadImage(name.toStringMessage())
                                               : new logical(false);
   }

   /**
    * Load image for widget.  Internal version.
    * 
    * @param    imageSrc
    *           Image file name to load.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   private boolean loadImageInt(String imageSrc)
   {
      // save current values
      String imageNameSaved = config.image;
      
      // this must be done before sending image bytes because the image should be on client side
      if (frame != null)
      {
         config.image = imageSrc;
         frame.pushScreenDefinition(true);
      }
      
      // logical terminal will try to load image from application jar, then ask the
      // client to load from local file system in cas of failure
      boolean result = LogicalTerminal.setImage(
                          config.id.asInt(), ButtonConfig.IMAGE_IMG, imageSrc);
      
      // if we are here - there is no image to load we need to undo image name change
      if (!result && frame != null)
      {
         config.image = imageNameSaved;
         frame.pushScreenDefinition(true);
      }
      
      return result;
   }
}