/*
** Module   : MenuItemWidget.java
** Abstract : server side menu item widget implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 VIG 20141118 Created initial version.
** 002 VIG 20150115 Added pushScreenDefinition method, fixed javadoc.
** 003 VIG 20150227 Added findMenuStatic method implementation and pushScreenDefinition calls.
**                  Added setSubTypeInt method.
** 004 VIG 20150525 Added validParent() method.
** 005 VIG 20150813 Added additional setAccelerator method and implemented MenuItemInterface.
**                  Removed mnemonic processing.
** 006 GES 20150923 Javadoc fix.
** 007 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 008 SVL 20170120 Changes related to addition of the TOGGLE-BOX attribute to the menu-item
**                  interface.
**                  Fixed setSubType - it must fail if the parent is realized.
**     HC  20170131 Added runtime support for dynamic MENU* widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This class represents menu-item widget.
 */
@LegacyResource(resource = LegacyResource.MENU_ITEM)
public class MenuItemWidget
extends BaseEntity<MenuItemConfig>
implements SubTypeAttribute, 
           MenuItemInterface
{
   /**
    * Default constructor for dynamic menu-items  
    */
   public MenuItemWidget(boolean dynamic)
   {
      super(dynamic, new MenuItemConfig());
   }

   /**
    * Default constructor for static menu-item
    * 
    * @param    name
    *           The name of menu-item.
    * @param    label
    *           MENU-ITEM label.          
    */
   public MenuItemWidget(String name, String label)
   {
      super(false, new MenuItemConfig());
      name(name);
      config.title = label;
   }

   /**
    * constructor for static menu-item of non standard sub-type: RULE or SKIP.
    * 
    * @param    subtype
    *           MENU-ITEM sub-type. One of 3 constants: {@link MenuItemConfig#SBTP_NORMAL},
    *           {@link MenuItemConfig#SBTP_RULE} or {@link MenuItemConfig#SBTP_SKIP}                    
    */
   public MenuItemWidget(String subtype)
   {
      super(false, new MenuItemConfig());
      setSubTypeInt(subtype);
   }
   
   /**
    * Sets the label text, parsing mnemonic.
    * 
    * TODO: find out what is going on with undocumented feature of very long label
    * See parent's javadoc.
    * 
    * @param    label
    *           The new label text.
    */
   @Override
   public void setLabel(String label)
   {
      config.title = label;
      
      pushScreenDefinition();
   }

   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   @Override
   public logical isReadOnly()
   {
      return new logical(config.readOnly);
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   @Override
   public void setReadOnly(logical r)
   {
      if (r != null)
      {
         setReadOnly(r.booleanValue());
      }
   }
   
   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   public void setReadOnly(boolean r)
   {
      config.readOnly = r;
      pushScreenDefinition();
   }
   
   /**
    * Sets disabled flag during construction in menu definition class.
    * Intended only on for construction step and must no be called elsewhere.
    * 
    * @param   disabled
    *          <code>true</code> means menu-item is disabled.
    *          
    * @throws  IllegalStateException
    *          thrown if this method is called somewhere else except menu definition class.
    */
   public void setInitDisabled(boolean disabled)
   {
      if (config.initDisabled != null)
      {
         throw new IllegalStateException("initDisabled already has a value.");
      }
      
      config.initDisabled = disabled;
      setSensitive(!disabled);
      pushScreenDefinition();
   }

   /**
    * Gets the TOGGLE-BOX attribute value.
    *
    * @return TOGGLE-BOX attribute value: <code>true</code> if the menu-item should have a
    *         checkbox.
    */
   @Override
   public boolean isToggleBox()
   {
      return config.toggleBox;
   }

   /**
    * Sets the TOGGLE-BOX attribute of the menu-item.
    * 
    * @param    toggleBox
    *           <code>true</code> if the menu-item should have a checkbox.
    */
   @Override
   public void setToggleBox(boolean toggleBox)
   {
      setToggleBox(new logical(toggleBox));
   }

   /**
    * Sets the TOGGLE-BOX attribute of the menu-item.
    *
    * @param    toggleBox
    *           <code>true</code> if the menu-item should have a checkbox.
    */
   @Override
   public void setToggleBox(logical toggleBox)
   {
      config.toggleBox = toggleBox.booleanValue();
      pushScreenDefinition();
   }
   
   /**
    * Get this menu-item accelerator - KEYLABEL of the short key assigned to this menu-item.
    * 
    * @return  See above.
    */
   @Override
   public String getAccelerator()
   {
      return config.accelerator;
   }

   /**
    * Set this menu-item accelerator - KEYLABEL of the short key assigned to this menu-item.
    * 
    * @param  accelerator
    *         The valid key-label, that will be target accelerator.
    */
   @Override
   public void setAccelerator(String accelerator)
   {
      if (Keyboard.keyCode(accelerator) == -1)
      {
         String msg = "Warning: Invalid ACCELERATOR given in string: " + accelerator;
         if (LogicalTerminal.isChui())
         {
            ErrorManager.recordOrThrowError(4130, msg, false, false);
         }
         else
         {
            ErrorManager.displayError(4130, msg, false);
         }
         
         return;
      }
      
      config.accelerator = accelerator;
      pushScreenDefinition();
   }

   /**
    * Sets this menu-item accelerator.
    * 
    * @param  accel
    *         The valid key-label, that will be target accelerator.
    */
   @Override
   public void setAccelerator(character accel)
   {
      String accelerator = (accel == null || accel.isUnknown()) ? null : accel.getValue();
      setAccelerator(accelerator);
   }

   /**
    * Returns the sub-type of this menu-item.
    * 
    * @return   See above.
    */
   @Override
   @LegacyAttribute(name = "SUBTYPE")
   public character getSubType()
   {
      if (!_dynamic())
      {
         String[] msg = new String[]{"**SUBTYPE is not a queryable attribute for MENU m"};
         ErrorManager.recordOrShowError(new int[]{4052}, msg, false, false, false);
         return new character();
      }
      
      return new character(config.subtype);
   }
   
   /**
    * Sets the the sub-type of the object.
    * 
    * @param    value
    *           The new value of the sub-type object attribute.
    */
   @LegacyAttribute(name = "SUBTYPE", setter = true)
   public void setSubType(character value)
   {
      if (value == null || value.isUnknown())
      {
         String templ = "**Unable to assign UNKNOWN value to attribute SUBTYPE on MENU-ITEM %s.";
         String[] msg = new String[] { String.format(templ, name()) };
         ErrorManager.recordOrShowError(new int[] {4083}, msg, false, false, false);
      }
      else
      {
         setSubType(value.getValue());
      }
   }
   
   /**
    * Sets the the sub-type of the object.
    * 
    * @param    value
    *           The new value of the sub-type object attribute.
    */
   @LegacyAttribute(name = "SUBTYPE", setter = true)
   public void setSubType(String value)
   {
      if (!_dynamic())
      {
         String templ = "**Unable to set attribute SUBTYPE for MENU-ITEM %s";
         String[] msg = new String[] { String.format(templ, name()) };
         ErrorManager.recordOrShowError(new int[] {4078}, msg, false, false, false);
         return;
      }
      
      if (getParent() != null)
      {
         MenuContainerWidget p = (MenuContainerWidget) getParent();
         
         if (p._isRealized())
         {
            String msg = "**Unable to set SUBTYPE because the MENU-ITEM widget has been realised";
            ErrorManager.recordOrShowError(new int[] {4053}, new String[] {msg}, false, false, false);
            return;
         }
      }

      setSubTypeInt(value);
   }

   /**
    * Sets new value of SENSITIVE attribute. Overrides {@link GenericWidget} logic 
    * (that enables widget in its frame).
    *
    * @param    s
    *           The new value of the SENSITIVE attribute.
    */
   @Override
   public void setSensitive(boolean s)
   {
      // changing the SENSITIVE attribute does not affect the widget's visibility
      config.enabled = s;
      pushScreenDefinition();
   }
   
   /**
    * Gets the CHECKED writable attribute.
    *
    * @return    The current value of the CHECKED attribute.
    */
   public logical isChecked()
   {
      return new logical(config.checked);
   }

   /**
    * Set CHECKED attribute value.
    * 
    * @param   checked
    *          New value for the attribute.
    */
   public void setChecked(boolean checked)
   {
      config.checked = checked;
      pushScreenDefinition();
   }

   /**
    * Set CHECKED attribute value.
    * 
    * @param   checked
    *          New value for the attribute.
    */
   public void setChecked(logical checked)
   {
      if (checked.isUnknown())
      {
         String msg = "Unable to assign UNKNOWN value to attribute CHECKED on MENU-ITEM " + name();
         ErrorManager.displayError(4083, msg);
         return;
      }         

      if (!config.toggleBox)
      {
         String msg = "**The CHECKED attribute is valid only for TOGGLE-BOX MENU-ITEMS.";
         ErrorManager.displayError(4091, msg);
         return;
      }

      setChecked(checked.booleanValue());
   }
   
   /**
    * Finds the first MENU-ITEM with specified name in all menus.
    * 
    * @param   name
    *          The name of searched MENU-ITEM
    *          
    * @return  The first found MenuItemWidget or null.
    */
   public static MenuItemWidget findMenuItemStatic(String name)
   {
      for (MenuWidget menu : LogicalTerminal.getAllMenus())
      {
         MenuItemWidget found = menu.findMenuItem(name);
         if (found != null)
         {
            return found;
         }
      }
      
      return null;
   }

   /**
    * Conditionally call {@link GenericFrame#pushScreenDefinition()}, depending if the frame was
    * set or not for this widget.
    */
   @Override
   protected void pushScreenDefinition()
   {
      CommonWidget parent = getParent();
      if (parent != null && parent instanceof MenuContainerWidget<?>)
      {
         ((MenuContainerWidget<?>)parent).pushScreenDefinition();
      }
   }

   /**
    * Returns the widget types that can act as parent for this widget.
    *
    * @return   see above.
    */
   @Override
   protected String[] getValidParentTypes()
   {
      // parent must be a menu or sub-menu
      return new String[] {LegacyResource.MENU, LegacyResource.SUB_MENU};
   }

   /**
    * Sets the the sub-type of the object.
    * 
    * @param    value
    *           The new value of the sub-type object attribute.
    */
   private void setSubTypeInt(String value)
   {
      config.subtype = value;
      pushScreenDefinition();
   }
}