/*
** Module   : PaneConfig.java
** Abstract : container for the DIALOG, FRAME, WINDOS widget state
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140514 Created initial version.
** 002 HC  20140911 GUI coordinate support - data types refactoring.
** 003 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 004 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 005 HC  20150803 Improved runtime support for WINDOW size attributes.
** 006 EVL 20151111 THREE-D option has false as default value.
** 007 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 008 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 009 HC  20160320 Improved frame virtual size handling.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * The container for common widget state variables shared by 
 * DIALOG, FRAME, WINDOW widgets.
 */
public abstract class PaneConfig
extends BaseConfig
{
   /** THREE-D options */
   public boolean threeD = false;
   
   /**   
    * VIRTUAL-WIDTH-CHARS option, 
    * to get the effective value see @link {@link ConfigHelper}. 
    */
   public double virtualWidthChars = BaseConfig.INV_COORD; 

   /**   
    * VIRTUAL-HEIGHT-CHARS option, 
    * to get the effective value see @link {@link ConfigHelper}. 
    */
   public double virtualHeightChars = BaseConfig.INV_COORD; 
   
   /**   
    * VIRTUAL-WIDTH-PIXELS option, 
    * to get the effective value see @link {@link ConfigHelper}. 
    */
   public int virtualWidthPixels = BaseConfig.INV_COORD; 

   /**   
    * VIRTUAL-HEIGHT-PXIELS option, 
    * to get the effective value see @link {@link ConfigHelper}. 
    */
   public int virtualHeightPixels = BaseConfig.INV_COORD;
   
   /** 
    * This is the server-logic assigned visible state.
    * <p> 
    * It is needed in cases where the widget is effectively invisible, 
    * but the requested state is visible. An example is a DIALOG-BOX - 
    * when its parent window is made invisible the box must be made 
    * unconditionally invisible as well, when the parent window is made 
    * visible again the box is made visible only when the last assigned 
    * value in VISIBLE was TRUE (or ENABLE called, etc.). 
    */
   public boolean assignedVisible = false;
   
   /**
    * Default constructor (only used in deserialization).
    */
   public PaneConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected PaneConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected PaneConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      PaneConfig cfg = (PaneConfig) config;
      
      threeD = cfg.threeD;
      virtualWidthChars = cfg.virtualWidthChars;
      virtualHeightChars = cfg.virtualHeightChars;
      virtualWidthPixels = cfg.virtualWidthPixels;
      virtualHeightPixels = cfg.virtualHeightPixels;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      threeD             = in.readBoolean();
      virtualHeightChars = in.readDouble();
      virtualWidthChars  = in.readDouble();
      virtualHeightPixels = in.readInt();
      virtualWidthPixels  = in.readInt();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeBoolean(threeD);
      out.writeDouble(virtualHeightChars);
      out.writeDouble(virtualWidthChars);      
      out.writeInt(virtualHeightPixels);
      out.writeInt(virtualWidthPixels);      
   }
}