/*
** Module   : PaneEntity.java
** Abstract : server side base class for DIALOG, FRAME, WINDOW widgets
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140514 Created initial version.      
** 002 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 003 HC  20141021 Implemented window move/position support.
** 004 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 005 HC  20150803 Improved runtime support for WINDOW size attributes.
** 006 IAS 20151015 Implemented setters  for the THREE-D attribute
** 007 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 008 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
** 009 HC  20160320 Improved frame virtual size handling.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * The server side base class for DIALOG, FRAME, WINDOW widgets. 
 */
public class PaneEntity<T extends PaneConfig>
extends BaseEntity<T>
{
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    paneConfig
    *           The container for all state to be transferred to the client.
    */
   public PaneEntity(boolean dynamic, T paneConfig)
   {
      super(dynamic, paneConfig);
   }
   
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    paneConfig
    *           The container for all state to be transferred to the client.
    * @param    before
    *           The right neighbor of the BaseEntity been created or null if new BaseEntity
    *           will be the last in the chain. If <b>before</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    * @param    firstAsLast
    *           The flag indicating that this new entity must be inserted as last element 
    *           in the chain in the case when <b>before</b> - is the first element of the chain.
    *           When <b>before</b> is the first element of the chain, default behavior (expected
    *           from the name of this parameter is to insert new element as the first. 
    *           This flag allows to change the expected behavior of <b>before</b> parameter when 
    *           it needs and makes sense only if <b>before</b> is the first element.
    */
   public PaneEntity(boolean dynamic, T paneConfig, PaneEntity<T> before, boolean firstAsLast)
   {
      super(dynamic, paneConfig, before, firstAsLast);
   }
   
   /**
    * Check if window has THREE-D option set.
    * 
    * @return <code>true</code> if window has THREE-D option set.
    */
   @Override
   public logical isThreeD()
   {
      return new logical(config.threeD);
   }

   /**
    * Enable/disable 3D look.
    * 
    * @param   threeD
    *          New flag value.
    */
   @Override
   public void setThreeD(logical threeD)
   {
      if (threeD == null || threeD.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute THREE-D on %s %s", 
               type(), widgetName()
               ),
               false);
         return;
      }
      setThreeD(threeD.booleanValue());
   }

   /**
    * Enable/disable 3D look.
    * 
    * @param   threeD
    *          New flag value.
    */
   @Override
   public void setThreeD(boolean threeD)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set THREE-D because the %s %s has been realised", 
                  type(), widgetName()
               ), 
               false);
         return;
      }
      
      if (config.threeD == threeD)
      {
         return;
      }
      
      config.threeD = threeD;
      pushScreenDefinition();
   }

   /**
    * Get VIRTUAL-HEIGHT-CHARS attribute. 
    */
   @Override
   public decimal getVirtHeight()
   {
      return new decimal(config.virtualHeightChars == BaseConfig.INV_COORD ? 
                                                      config.heightChars : 
                                                      config.virtualHeightChars);
   }

   /**
    * Set VIRTUAL-HEIGHT-CHARS attribute. 
    *
    * @param   height
    *          VIRTUAL-HEIGHT-CHARS value
    */
   @Override
   public void setVirtHeight(NumberType height)
   {
      setSizeChars(height.isUnknown() ? null : height.doubleValue(), 
                   SizeAttribute.VIRT_HEIGHT_CHARS);
   }

   /**
    * Get VIRTUAL-WIDTH-CHARS attribute. 
    */
   @Override
   public decimal getVirtWidth()
   {
      return new decimal(config.virtualWidthChars == BaseConfig.INV_COORD ? 
                                                     config.widthChars : 
                                                     config.virtualWidthChars);
   }

   /**
    * Set VIRTUAL-WIDTH-CHARS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-CHAR value
    */
   @Override
   public void setVirtWidth(NumberType width)
   {
      setSizeChars(width.isUnknown() ? null : width.doubleValue(), 
                   SizeAttribute.VIRT_WIDTH_CHARS);
   }
   
   /**
    * Get VIRTUAL-HEIGHT-PIXELS attribute. 
    */
   @Override
   public integer getVirtHeightPixels()
   {
      return new integer(config.virtualHeightPixels == BaseConfig.INV_COORD ? 
                                                       config.heightPixels : 
                                                       config.virtualHeightPixels);
   }

   /**
    * Set VIRTUAL-HEIGHT-PIXELS attribute. 
    *
    * @param   height
    *          VIRTUAL-HEIGHT-PIXELS value
    */
   @Override
   public void setVirtHeightPixels(NumberType height)
   {
      setSizePixels(height.isUnknown() ? null : height.intValue(), 
                    SizeAttribute.VIRT_HEIGHT_PIXELS);
   }

   /**
    * Get VIRTUAL-WIDTH-PIXELS attribute. 
    */
   @Override
   public integer getVirtWidthPixels()
   {
      return new integer(config.virtualWidthPixels == BaseConfig.INV_COORD ? 
                                                      config.widthPixels : 
                                                      config.virtualWidthPixels);
   }

   /**
    * Set VIRTUAL-WIDTH-PIXELS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-PIXELS value
    */
   @Override
   public void setVirtWidthPixels(NumberType width)
   {
      setSizePixels(width.isUnknown() ? null : width.intValue(), 
                    SizeAttribute.VIRT_WIDTH_PIXELS);
   }
   
   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthCharsWorker(double width)
   {
      config.virtualWidthChars = width;
      
      // do not use width here as the chars value may be adjusted due to
      // chars->pixels->chars conversion
      if (config.virtualWidthChars < config.widthChars)
         config.widthChars = config.virtualWidthChars;
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightCharsWorker(double height)
   {
      config.virtualHeightChars = height;

      // do not use height here as the chars value may be adjusted due to
      // chars->pixels->chars conversion
      if (config.virtualHeightChars < config.heightChars)
         config.heightChars = config.virtualHeightChars;
   }

   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthPixelsWorker(int width)
   {
      config.virtualWidthPixels= width;
      
      if (width < config.widthPixels)
         config.widthPixels = width;
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightPixelsWorker(int height)
   {
      config.virtualHeightPixels = height;

      if (height < config.heightPixels)
         config.heightPixels = height;
   }
}