/*
** Module   : RadioSetWidget.java
** Abstract : server side radioset widget implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23494 Created initial version.
** 002 SIY 20050209   @24428 Major refactoring.
** 003 GES 20060803   @28382 Some cleanup and refactoring.
** 004 GES 20080320   @37593 Added TYPE attribute.
** 018 SIY 20111101          Updated imports.
** 019 CA  20130221          Refactored usage of TYPE attribute.
** 020 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 021 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 022 CA  20131115          Changes related to properly instantiate dynamic resources.
** 023 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 025 IAS 20150123          Implemented DELETE() method
** 026 IAS 20150202          Implemented ADD_LAST() method, fixed ADD-FIRST()
** 027 IAS 20150304          Fixed getRadioButtons() & setItems() methods
** 028 CA  20150310          Changes related to ControlSetItems.setItems API change.
** 029 IAS 20150312          Implemented ENABLE/DISABLE method and NUM-BUTTONS attribute.
** 030 HC  20150323          Removed business logic from config classes
** 031 IAS 20150604          Added support for the REPLACE method and realize() operation.
** 032 EVL 20150911          Adding handlers for vertical orientation option.
** 033 EVL 20150930          Adding enabled state propagation to the internal radio-buttons.
**                           Overloading getScreenValue() method to ignore the format.  Fix for
**                           reset size for horizontal widget height.
** 034 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 035 HC  20160628          Fixed SENSITIVE value not propagated to radio buttons.
** 036 SBI 20160826          Commented getItemIndex() to return an index enumerated from 1.
** 037 CA  20161005          Fixed issues related to SCREEN-VALUE attribute (initial value or 
**                           explicitly set, when the value is empty string or unknown, or the 
**                           widget is uninitialized).
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;

/**
 * Server side radioset widget implementation.
 */
@LegacyResource(resource = LegacyResource.RADIO_SET)
public class RadioSetWidget
extends ControlSetEntity<RadioSetConfig>
implements RadioSetInterface
{
   /**
    * Default constructor.
    */
   public RadioSetWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public RadioSetWidget(boolean dynamic)
   {
      super(dynamic, new RadioSetConfig());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setHorizontal(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute HORIZONTAL on %s %s", type(),
                  widgetName()), false);
      }
      setHorizontal(value.booleanValue());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setHorizontal(boolean value)
   {
      if (!config.realized)
      {
         config.horizontal = value;
         resetSize(config.items);
      }
      else {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set HORIZONTAL because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   /**
    * Get the value of the HORIZONTAL attribute for radio-set.
    * 
    * @return
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public logical isHorizontal()
   {
      return new logical(config.horizontal);
   }
   
   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setVertical(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute HORIZONTAL on %s %s", type(),
                  widgetName()), false);
      }
      setHorizontal(!value.booleanValue());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   value
    *          The new value of the VERTICAL option.
    */
   @Override
   public void setVertical(boolean value)
   {
      setHorizontal(!value);
   }

   /**
    * Get the value of the EXPAND attribute for radio-set.
    * 
    * @return
    *          The value of the EXPAND attribute.
    */
   @Override
   public logical isExpand()
   {
      return new logical(config.expand);
   }

   /**
    * Resets the value of the EXPAND attribute for radio-set.
    * 
    * @param   value
    *          The new value of the EXPAND attribute.
    */
   @Override
   public void setExpand(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute EXPAND on %s %s", type(),
                  widgetName()), false);
      }
      setExpand(value.booleanValue());
   }

   /**
    * Resets the value of the EXPAND attribute for radio-set.
    * 
    * @param   value
    *          The new value of the EXPAND attribute.
    */
   @Override
   public void setExpand(boolean value)
   {
      if (!config.realized)
      {
         if (config.expand == value)
         {
            return;
         }
         
         config.expand = value;
         pushScreenDefinition();
      }
      else
      {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set EXPAND because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   
   /**
    * Get RADIO-BUTTONS attribute.
    * 
    * @return  Current value of the attribute.
    */
   @Override
   public character getRadioButtons()
   {
      if (config.items == null || config.items.length == 0)
      {
         ErrorManager.recordOrShowError(4094, 
               String.format(
                  "Attribute RADIO-BUTTONS is not queryable on %s %s.  It has no buttons", 
                  type(), widgetName()), 
               false);
         return new character();
      }
      return getListItemPairs();
   }
   
   /**
    * Set RADIO-BUTTONS attribute. The parameter contains string
    * representation of all radio button labels and values separated by
    * delimiter (comma is the default value).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,..., LabelN,valueN</code>
    * 
    * @param   s
    *          New value of the attribute.
    */
   @Override
   public void setRadioButtons(String s)
   {
      setRadioButtons(new character(s));
   }

   /**
    * Set RADIO-BUTTONS attribute. The parameter contains string
    * representation of all radio button labels and values separated by
    * delimiter (comma is the default value).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,..., LabelN,valueN</code>
    * 
    * @param   c
    *          New value of the attribute.
    */
   @Override
   public void setRadioButtons(character c)
   {
      setListItemPairs(c);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param item
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(character item)
   {
      if (!canAccess("DELETE"))
      {
         return new logical(false);
      }
    
      if (item == null || item.isUnknown())
      {
         invalidDeleteArg();
         return new logical(false);
      }
      return delete(item.toStringMessage());
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param item
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(String item)
   {
      if (!canAccess("DELETE"))
      {
         return new logical(false);
      }
      List<String> items = new ArrayList<>();
      items.add(item);
      /*
       * Progress 10.2B compares items for the RADIO-SET by label (case sensitive) and
       * if one tries to delete an item with label which doesn't exist in a case-sensitive sense 
       * but does exist in a case-insensitive one 
       * it crashes with "SYSTEM ERROR: Memory violation. (49)"
       */
      logical rc = delete(items, COMPARE_LABEL, true);
      if (!rc.booleanValue())
      {
         invalidDeleteArg();
      }
      return rc;
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(int64 idx)
   {
      UnimplementedFeature.unsupported("DELETE() method is undefined");
      return new logical();
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(long idx)
   {
      UnimplementedFeature.unsupported("DELETE() method is undefined");
      return new logical();
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a single label
    * and value pair to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character label, BaseDataType value)
   {
      if (!canAccess("ADD-LAST"))
      {
         return new logical(false);
      }
      ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];
      int currentHeight = items.length;
      int currentWidth = width(items);

      ControlSetItem[] nitems = Arrays.copyOf(items, items.length + 1);
      ControlSetItem item = controlSetItem(label, value);
      nitems[items.length] = item;
      item.setEnabled(!config.horizontal || resetSize(nitems));

      super.setItems(nitems, true);

      return new logical(true);
   }
   

   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   Always reports warning 4052 and returns "unknown".
    */
   @Override
   public logical addFirst(character label, BaseDataType value)
   {
      ErrorManager.recordOrShowError(4052, String.format(
            "ADD-FIRST is not a queryable attribute for %s %s", type(),
               widgetName()), false);
      return new logical();
   }

   /**
    * Store array of items into config.
    * 
    * @param    items
    *           Array of new items.
    */
   @Override
   public void setItems(ControlSetItem[] items, boolean pairs)
   {
      // for dynamic widgets the order below is important
      resetSize(items);
      super.setItems(items, true);
   }


   /**
    * Enable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be enabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical enable(String label)
   {
      return setEnabled(new character(label), true);
   }

   /**
    * Enable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be enabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical enable(character label)
   {
      return setEnabled(label, true);
   }

   /**
    * Disable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be disabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical disable(String label)
   {
      return setEnabled(new character(label), false);
   }

   /**
    * Disable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be disabled.
    *  
    * @return
    *          <code>true</code> if successful
    */
   @Override
   public logical disable(character label)
   {
      return setEnabled(label, false);
   }

   /**
    * Get the value of the NUM-BUTTONS attribute which is the number of buttions in the
    * RADIO-SET.
    *
    * @return   The number of buttons in the widget.
    */
   @Override
   public integer getNumButtons()
   {
      return config.items == null ? new integer(0) : new integer(config.items.length);
   }


   /**
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character from, character to, integer flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
.    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, character to, integer flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
.    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character from, character to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
.    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, character to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character from, String to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }


   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, String to, integer flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }
   
   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String from, String to, int flag)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character newItemList, character listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
     return missingArgs();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newItemList, String listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
     return missingArgs();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, character listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE(REPLACE (new-item-list, list-item )) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, String listItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character newItemList, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(character newItemList, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-item-list, list-index ) method
    * (for COMBO-BOX and SELECTION-LIST)
    * not applicable to RADIO-SET
    */
   @Override
   public logical replace(String newItemList, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return missingArgs();
   }

   /**
    * REPLACE (new-label, new-value, list-item) method for COMBO-BOX/SELECTION-LIST,
    * REPLACE (new-label, new-value, old-label) method for RADIO-SET
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param oldItem
    *          oldItem (old-label for the READIO-SET, and list-item for the COMBO-BOX/SELECTION-LIST)
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType newValue, character oldItem)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      // get zero based index
      int idx = getItemIndex(oldItem) - 1;
      if (idx < 0)
      {
         invalidReplaceArg();
         return new logical(false);
      }
      ControlSetItem item = config.items[idx];
      
      character empty = new character("");
      if (empty.equals(newLabel))
      {
         newLabel = item.getLabel();
      }
      
      if (empty.equals(newValue))
      {
         newValue = item.getValue();
      }

      ControlSetItem nitem = controlSetItem(newLabel, newValue);
      config.items[idx] = nitem;
      resetSize(config.items);
      item.setEnabled(item.isEnabled());
      super.setItems(config.items, true);
      
      return new logical(true);
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType  newValue, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(String newLabel, BaseDataType newValue, integer listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * REPLACE (old-string, new-string, flag ) method for EDITOR
    * 
    * @param arg0
    *          new item's label for COMBO-BOX/SELECTION-LIST,
    *          the source text for EDITOR
    * @param arg1         
    *          new item's value for COMBO-BOX/SELECTION-LIST,
    *          replacement test for editor
    * @param arg2
    *          index of the item to be replaced for COMBO-BOX/SELECTION-LIST,
    *          the search mode for EDITOR
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character arg0, String arg1, integer arg2)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(character newLabel, BaseDataType newValue, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * REPLACE (new-label, new-value, list-index) method for COMBO-BOX/SELECTION-LIST
    * 
    * @param newLabel
    *          new item's label
    * @param newValue         
    *          new item's value
    * @param listIndex
    *          index of the item to be replaced
    * @return   <code>true</code> if the operation was successful.
    */
   @Override
   public logical replace(String newLabel, BaseDataType  newValue, int listIndex)
   {
      if (!canAccess("REPLACE"))
      {
         return new logical(false);
      }
      return invalid3RdArg();
   }

   /**
    * Enable/Disable RADIO-BUTTON in a RADIO-SET widget. 
    * 
    * @param   label
    *          New label of the button to be enabled.
    * @param   enabled 
    *          new 'enabled' status
    *          
    * @return
    *          <code>true</code> if successful
    */
   protected logical setEnabled(character label, boolean enabled)
   {
      if (label != null && !label.isUnknown() && config.items != null)
      {
         String l = label.toStringMessage();
         for (int i = 0; i < config.items.length; i++)
         {
            ControlSetItem item = config.items[i];
            if (COMPARE_LABEL_IGNORE_CASE.test(l, item))
            {
               item.setEnabled(enabled);
               pushScreenDefinition();
               return new logical(true);
            }
         }
      }
      ErrorManager.recordOrShowError(4095, String.format(
                  "The radio button in the ENABLE attribute is invalid for %s %s", 
                  type(), widgetName()
               ), 
               false);
      return new logical(false);
   }
 
   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(boolean s)
   {
      for (int i = 0; i < config.items.length; i++)
      {
         config.items[i].setEnabled(s);
      }

      // then call superclass
      super.setSensitive(s);
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @Override
   public character getScreenValue()
   {
      // need to override this method because radio-set ignores format when getting
      // the screen value
      return frame.getScreenValue(this, true);
   }
   
   /**
    * Calculate widget width for items' list
    * @param items
    *        list of items 
    * @return
    *        the total width of widget associated with the items 
    */
   protected int width(ControlSetItem[] items)
   {
      if (items == null)
      {
         return 0;
      } 
      int maxButtonWidth = 0;
      int totalWidth = 0;
      for (ControlSetItem item: items)
      {
         character label = item.getLabel();
         int lw = label== null ? 0 : label.toStringMessage().length();
         maxButtonWidth = Math.max(maxButtonWidth, lw);
         totalWidth += lw + 3 + 1;
      }
      if (config.expand)
      {
         totalWidth = items.length * (maxButtonWidth + 3 + 1);
      } 
      return config.horizontal ? Math.max(0,totalWidth - 1) : maxButtonWidth;
   }

   /**
    * Replace array of items in the config.
    * 
    * 
    * @param    items
    *           Array of new items.
    * @return
    *           true if size changed OK (or not changed(
    */
   protected boolean resetSize(ControlSetItem[] items)
   {
      if (!config.autoResize)
      {
         return true;
      }

      // TODO: why is this assuming we are in ChUI, when setting it to 1 char height?
      int height = items == null ? 0 : config.horizontal ? 1 : items.length;
      int width = width(items);
      boolean overflow = false;
      if (dynamic().booleanValue() ? frame != null : config.realized)
      {
         FrameConfig frameConfig = frame.getFrameWidget().config();
         int maxSize = (int)(config.horizontal ? frameConfig.widthChars 
                                               :frameConfig.heightChars);

         if (frameConfig.box) 
         {
            // TODO: why is this assuming we are in ChUI?
            maxSize -= 2;
         }
         
         overflow = maxSize < (config.horizontal ? width : height); 
      }

      if (overflow)
      {
         ErrorManager.recordOrShowError(config.horizontal ? 4475 : 4474, 
               String.format("New %s for %s %s is too large to fit in frame",
                           config.horizontal ? "width" : "height",
                           type(), widgetName()
               ), 
               false);
      }
      else 
      {
         config.listHeightChars = Coordinate.scale(height);
         config.listWidthChars = Coordinate.scale(width);
      }
      return !overflow;
   }

   /**
    * Report invalid argument for delete().
    * 
    */
   private void invalidDeleteArg()
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The radio button in the DELETE attribute is invalid for %s %s", 
               type(), widgetName()), 
            false);
   }

   /**
    * Report invalid argument for replace().
    * 
    */
   private void invalidReplaceArg()
   {
      ErrorManager.recordOrShowError(4095, 
            String.format(
               "The radio button in the REPLACE attribute is invalid for %s %s", 
               type(), widgetName()), 
            false);
   }

   /**
    * Get an item index, enumerated from 1.
    *  
    * @param    itemId
    *           Item's label or value
    * 
    * @return   Item's index enumerated from number 1 or -1 if not found
    */
   @Override
   protected int getItemIndex(character itemId)
   {
      if (itemId == null || itemId.isUnknown())
      {
         return -1;
      }
      
      int itemIndex = itemIndex(itemId.toStringMessage(), COMPARE_LABEL);
      
      return itemIndex >= 0 ? itemIndex + 1 : itemIndex;
   }

   /**
    * Internal worker for getting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   initialized
    *          Flag indicating if the widget is realized or not.
    * @param   bdt
    *          The current value in the frame's buffer.
    * @param   fmt
    *          The format to be applied, in case this is a formatted value.
    * @param   ignoreFormat
    *          Flag indicating if the format needs to be ignored.
    *          
    * @return  This widget's screen-value
    */
   @Override
   character getScreenValue(boolean initialized, BaseDataType bdt, String fmt, boolean ignoreFormat)
   {
      if (!initialized)
      {
         return new character();
      }
      
      return super.getScreenValue(initialized, bdt, fmt, ignoreFormat);
   }
   
   /**
    * Internal worker for setting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   frameBuf
    *          The frame buffer where to save the value.
    * @param   value
    *          The value to be set via SCREEN-VALUE attribute.
    * @param   inUIStmt
    *          Flag indicating this call originates from a UI statement.
    *          
    * @return  <code>true</code> if the caller can proceed, as the screen-value can be set.
    */
   @Override
   boolean setScreenValue(ScreenBuffer frameBuf, Object value, boolean inUIStmt)
   {
      if (frameBuf.getWidgetValue(getId()) != null)
      {
         if (value instanceof character && "".equals(((character) value).toJavaType()))
         {
            return false;
         }
      }
      
      return super.setScreenValue(frameBuf, value, inUIStmt);
   }

   /**
    * Report missing argument for replace().
    * 
    */
   private logical missingArgs()
   {
         ErrorManager.recordOrShowError(4064, 
               String.format(
                  "Missing argument when setting attribute REPLACE for %s %s", 
                  type(), widgetName()), 
               false);
         ErrorManager.recordOrShowError(4065, 
               String.format(
                  "The REPLACE attribute on the %s %s has invalid arguments", 
                  type(), widgetName()), 
               false);
         return new logical(false);
   }
      
   /**
    * Report invalid third argument for replace().
    * 
    */
   private logical invalid3RdArg()
   {
      ErrorManager.recordOrShowError(4068, 
            String.format(
               "The third argument for REPLACE on the %s %s should not be of type %s", 
               type(), widgetName(), "INTEGER"), 
            false);
      return new logical(false);
   }
}