/*
** Module   : ServerImageWorker.java
** Abstract : image helper methods to be used on the server side.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20151028 Created initial version.  Adding logic to handle images defined without
**                  extension.  In this case we need to search BMP, ICO and CUR files.
** 002 IAS 20150217 Minor refactoring of the getResourceName method
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.classloader.*;
import com.goldencode.p2j.util.*;

/**
 * Image helper methods to handle image loading from application jar files on the server side.
 */
public class ServerImageWorker
{
   /** Jar classes cache for images sources. */
   private static Map<String, JarClassLoader> jarImgCache = new HashMap<String, JarClassLoader>();
   
   /**
    * Gets jar file class loader to obtaing image resources.  If not in current cache - the new
    * loader is creating and storing into cache.
    * 
    * @param    jarFullName
    *           The fully qualified jar file name to load.  It must be correct refined name.
    * @param    parentCl
    *           The class loader to be used as parent class loaded when new class loader should be
    *           created. 
    *
    * @return   The instance of the JarClassLoader to use in image loading.
    */
   public static JarClassLoader getJarClassLoader(String jarFullName, ClassLoader parentCl)
   {
      // first try to reuse the cache 
      JarClassLoader jclRes = jarImgCache.get(jarFullName);

      // not in cache, create and store for further usage
      if (jclRes == null)
      {
         jclRes = new JarClassLoader(jarFullName, parentCl);
         // loader created, initialize and store into the cache
         if (jclRes != null)
         {
            jclRes.init();
            jarImgCache.put(jarFullName, jclRes);
         }
      }
      
      return jclRes;
   }
   
   /**
    * Gets the resource name inside jar file for given image name.  Handles images without
    * extensions if the jar has 4gl compatible image files.  The BMP, ICO and CUR are possible.
    * 
    * @param    jcl
    *           The jar class loader that controls the jar file to search for.
    * @param    name
    *           The image name to find inside jar resource. 
    * @param    caseSens
    *           The flag to define current case sensitivity mode to use while searching. 
    *
    * @return   The image name inside jar file if can be found or <CODE>null</CODE> if the
    *           resource does not exist.
    */
   public static String getResourceName(JarClassLoader jcl, String name, boolean caseSens)
   {
      if (!"".equals(Utils.getExt(name)))
      {
         return jcl.getResourceName(name, caseSens);
      }

     return Utils.tryWhileNull(caseSens ? ImageExt.EXT_VARIANTS : ImageExt.EXT_BASES, 
                               ext -> jcl.getResourceName(name.concat(ext), caseSens));
   }
}