/*
** Module   : WidgetIdHelper.java
** Abstract : represents service for widget identifiers
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20140813 Initial version.
** 002 HC  20161110 Added getParentFrames().
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.security.*;

import java.util.*;

/**
 * This class performs useful helper calls related to widget identifiers.
 */
public class WidgetIdHelper
{
   /** Stores session context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
   
   /**
    * Initializes the proper widget ID worker on the client or server side.
    * 
    * @param    idWorkerInitial
    *           ID worker implementation for client side or server side.
    */
   public static void initWorker(WidgetIdWorker idWorkerInitial)
   {
      work.obtain().idWorker = idWorkerInitial;
   }

   /**
    * Retrieve the ID of the frame containing the widget with a given ID.  Public version for
    * usage on the client side.
    *
    * @param    widgetId
    *           The widget ID.
    *
    * @return   The frame ID if widget is found or -1 if not.
    */
   public static int frameIdFromWidgetId(int widgetId)
   {
      WorkArea wa = work.obtain();
      
      if (wa.idWorker != null)
      {
         return wa.idWorker.frameIdFromWidgetId(widgetId);
      }
      else
      {
         return -1;
      }
   }

   /**
    * Returns all parent frames up the root frame starting from the provided widget.
    * The frames are sorted from the most inner up to the root frame. If no frames
    * are found, an empty list is returned. If the given widget is itself a frame,
    * then it will be included in the resulting list.
    *
    * @param    widgetId
    *           A widget id.
    *
    * @return   See above, if no frames are found, returns an empty list.
    */
   public static List<Integer> getParentFrames(int widgetId)
   {
      WorkArea wa = work.obtain();

      if (wa.idWorker != null)
      {
         return wa.idWorker.getParentFrames(widgetId);
      }
      else
      {
         return Collections.emptyList();
      }
   }

   /** 
    * Stores global data relating to the state of the current session context.
    */
   private static class WorkArea
   {  
      /** Server or client side registered instance to perform actual calls. */
      private WidgetIdWorker idWorker = null;
   }
   
   /**
    * Simple container that stores and returns a context-local instance of the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work area for the session.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a new session context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();

         return wa;
      }   
   }   
}
