/*
** Module   : WindowWidget.java
** Abstract : server side window widget implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23498 Created initial version.
** 002 GES 20060803   @28385 Major cleanup and refactoring.
** 003 SIY 20061201   @31528 Ignore attempts to change window visibility
**                           for CHUI.
** 004 GES 20080320   @37595 Added TYPE attribute.
** 005 SVL 20081202   @40751 setVisible(false) should have the same effect
**                           as LogicalTerminal.hideAll(true).
** 006 SIY 20081224   @40990 Replaced hideAll() in setVisible() with call 
**                           to dedicated client API.
** 007 SIY 20101004          Added necessary method stubs.
** 008 SIY 20101005          Icon-related method stubs.
** 009 SIY 20101007          Added asWidget() method.
** 010 SIY 20111101          Updated imports.
** 011 CA  20130112          Extracted into CommonWindow interface all window
**                           APIs.
** 012 CA  20130221          Refactored usage of TYPE attribute.
** 013 CA  20130223          Removed assign, as it window assignment is done using
**                           LogicalTerminal APIs.
** 014 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 015 CA  20130323          The PARENT attribute must be a CommonWidget, not Object.
** 016 CA  20130927          Resource type is determined from LegacyResource annotation.
** 017 CA  20131115          Changes related to properly instantiate dynamic resources.
** 018 MAG 20140512          Implemented new attributes and support for CREATE WINDOW.
** 019 CA  20140701          Added stubbed support for font-related attributes:  MESSAGE-AREA-FONT, 
**                           STATUS-AREA-FONT.
** 020 CA  20140711          Enhanced font support, for GUI clients.
** 021 CA  20140805          Push the window config to the client, if the title gets set.
**                           resourceDelete needs to call LT.deregisterWindow, to cleanup both
**                           server and client-side.
** 022 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 023 HC  20141021          Implemented window move/position support.
** 024 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 025 CA  20141123          Added first pass at dynamic window dimension/location.
** 026 EVL 20141204          Added LOAD-SMALL-ICON implementation.
** 027 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 028 VIG 20141105          Added MENUBAR attribute stubs.
** 029 EVL 20150108          Differentiate small icon from large one flag.  Rework for icon/image
**                           implementation.  Fixed some javadoc warnings for menubars.
** 030 VIG 20150119          Added LogicalTerminal.pushWindow call to transfer menubar in config.
** 031 CA  20150408          Fixed HIDE/SHOW for windows.  Fixed window delete.
** 032 HC  20150323          WidgetId.getId() renamed to asInt().
** 033 CA  20150429          Static windows (the DEFAULT-WINDOW) is visible as soon as it is 
**                           created.
** 034 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 035 HC  20150508          Improvements to GUI window scrolling support.
** 037 CA  20150725          Added canSetPixelPositionInChui to return false - WINDOW can't be 
**                           moved via pixel coords in ChUI.
** 038 HC  20150803          Improved runtime support for WINDOW size attributes.
** 039 SVL 20150827          Override setVisible in parent GenericWidget.
** 040 HC  20150831          MIN/MAX size attribute values must be fixed after window is
**                           realized.
** 041 IAS 20150926          Added support for the MOVE-TO-TOP()/MOVE-TO-BOTTOM() methods.
** 042 IAS 20151007          Added support for the TOP-ONLY attribute.
** 043 VIG 20151021          Fixed NPE in setMenubar, when specified menubar is null.   
** 044 CA  20151024          Added support for WINDOW:SENSITIVE attribute.  Implicit window state
**                           is "not visible".
**     HC  20151027          Implemented WINDOW MESSAGE-AREA and STATUS-AREA attributes.
** 045 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 046 HC  20160203          Improved runtime support of WINDOW:SCROLL-BARS attribute.
** 047 IAS 20160217          LOAD-MOUSE-POINTER support.
** 048 HC  20160304          Fixed various window sizing/positioning issues. Window layout
**                           improvements.
** 049 SBI 20161005          Applied CA FIRST-CHILD changes: if this window has no window
**                           children, then get its first frame child.
** 050 SVL 20170126          Added support for WINDOW-STATE attribute.
**     HC  20170202          Implemented runtime support for WINDOW-STATE. The implementation is
**                           missing support for WINDOW-DELAYED-MINIMIZE.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
 
/**
 * This class represents the only window widget that exists in a CHUI
 * environment.
 * <p>
 * The window widget differs from all the other widgets also in this: it is 
 * the only widget that is not contained in a frame, but rather contains
 * frames (and other top-level components) itself. 
 */
public class WindowWidget
extends PaneEntity<WindowConfig>
implements CommonWindow,
           CommonHandleTree,
           TopOnlyInterface,
           Hoverable
{
   /** MenuWidget handle owned by this window */
   private MenuWidget menubar;
   
   /**
    * Default constructor.  It is package private because an instance must
    * only be created in a controllable way.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   WindowWidget(boolean dynamic)
   {
      super(dynamic, new WindowConfig());
      
      config.visible = false;

      if (dynamic)
      {
         // Register window
         LogicalTerminal.registerWindow(config, this);
      }
   }

   /**
    * Reports if this widget object is valid for use.
    *
    * @return   The state of the {@link GenericWidget#deleted} flag.
    */
   @Override
   public boolean valid()
   {
      return !deleted;
   }

   /**
    * Gets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @return   The current value of the KEEP-FRAME-Z-ORDER attribute.
    */
   @Override
   public logical isKeepFrameZOrder()
   {
      return new logical(config.keepFrameZOrder);
   }

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @Override
   public void setKeepFrameZOrder(logical keep)
   {
      setKeepFrameZOrder(keep.booleanValue()); 
   }

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @Override
   public void setKeepFrameZOrder(boolean keep)
   {
      config.keepFrameZOrder = keep;
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Gets the FULL-HEIGHT-CHARS writable attribute.
    *
    * @return current value of FULL-HEIGHT-CHARS attribute
    */
   @Override
   public decimal getFullHeightChars()
   {
      if (config.fullHeightChars < 0)
      {
         return new decimal();
      }
      
      return new decimal(config.fullHeightChars);
   }
   
   /**
    * Gets the FULL-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of FULL-HEIGHT-PIXELS attribute
    */
   @Override
   public integer getFullHeightPixels()
   {
      if (config.fullHeightPixels < 0)
      {
         return new integer();
      }
      
      return new integer(config.fullHeightPixels);
   }
   
   /**
    * Gets the FULL-WIDTH-CHARS writable attribute.
    *
    * @return current value of FULL-WIDTH-CHARS attribute
    */
   @Override
   public decimal getFullWidthChars()
   {
      if (config.fullWidthChars < 0)
      {
         return new decimal();
      }

      return new decimal(config.fullWidthChars);
   }
   
   /**
    * Gets the FULL-WIDTH-PIXELS writable attribute.
    *
    * @return current value of FULL-WIDTH-PIXELS attribute
    */
   @Override
   public integer getFullWidthPixels()
   {
      if (config.fullWidthPixels < 0)
      {
         return new integer();
      }

      return new integer(config.fullWidthPixels);
   }
   
   /**
    * Gets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @return current value of MAX-HEIGHT-CHARS attribute
    */
   @Override
   public decimal getMaxHeightChars()
   {
      return new decimal(ConfigHelper.getMaxHeightChars(config));
   }

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   @Override
   public void setMaxHeightChars(NumberType max)
   {
      setMaxHeightChars(max.doubleValue());
   }
   
   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   public void setMaxHeightChars(double max)
   {
      setSizeChars(max, SizeAttribute.MAX_HEIGHT_CHARS);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Gets the MAX-WIDTH-CHARS writable attribute.
    *
    * @return   The current value of MAX-WIDTH-CHARS attribute.
    */
   @Override
   public decimal getMaxWidthChars()
   {
      return new decimal(ConfigHelper.getMaxWidthChars(config));
   }

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   @Override
   public void setMaxWidthChars(NumberType max)
   {
      setMaxWidthChars(max.doubleValue());
   }

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   public void setMaxWidthChars(double max)
   {
      setSizeChars(max, SizeAttribute.MAX_WIDTH_CHARS);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Set maximal height.
    * 
    * @param   maxHeight
    *          New option value.
    */
   @Override
   public void setMaxHeight(NumberType maxHeight)
   {
      this.setMaxHeightChars(maxHeight);
   }

   /**
    * Set maximal width.
    * 
    * @param   maxWidth
    *          New option value.
    */
   @Override
   public void setMaxWidth(NumberType maxWidth)
   {
      this.setMaxWidthChars(maxWidth);
   }

   /**
    * Gets the MIN-HEIGHT-CHARS writable attribute.
    *
    * @return current value of MIN-HEIGHT-CHARS attribute
    */
   @Override
   public decimal getMinHeightChars()
   {
      return new decimal(ConfigHelper.getMinHeightChars(config));
   }
   
   /**
    * Gets the MIN-WIDTH-CHARS writable attribute.
    *
    * @return current value of MIN-WIDTH-CHARS attribute
    */
   @Override
   public decimal getMinWidthChars()
   {
      return new decimal(ConfigHelper.getMinWidthChars(config));
   }

   /**
    * Sets the MIN-HEIGHT-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-HEIGHT-CHARS attribute.
    */
   @Override
   public void setMinHeightChars(NumberType min)
   {
      setSizeChars(min.isUnknown() ? null : min.doubleValue(), SizeAttribute.MIN_HEIGHT_CHARS);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Sets the MIN-WIDTH-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-WIDTH-CHARS attribute.
    */
   @Override
   public void setMinWidthChars(NumberType min)
   {
      setSizeChars(min.isUnknown() ? null : min.doubleValue(), SizeAttribute.MIN_WIDTH_CHARS);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Gets the MIN-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of MIN-HEIGHT-PIXELS attribute
    */
   @Override
   public integer getMinHeightPixels()
   {
      return new integer(ConfigHelper.getMinHeightPixels(config));
   }
   
   /**
    * Gets the MIN-WIDTH-PIXELS writable attribute.
    *
    * @return current value of MIN-WIDTH-PIXELS attribute
    */
   @Override
   public integer getMinWidthPixels()
   {
      return new integer(ConfigHelper.getMinWidthPixels(config));   
   }

   /**
    * Sets the MIN-HEIGHT-PIXELS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-HEIGHT-PIXELS attribute.
    */
   @Override
   public void setMinHeightPixels(integer min)
   {
      setSizePixels(min.isUnknown() ? null : min.intValue(), SizeAttribute.MIN_HEIGHT_PIXELS);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Sets the MIN-WIDTH-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-WIDTH-CHARS attribute.
    */
   @Override
   public void setMinWidthPixels(integer min)
   {
      setSizePixels(min.isUnknown() ? null : min.intValue(), SizeAttribute.MIN_WIDTH_PIXELS);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Enable/disable window resizing.
    * 
    * @param   resize
    *          New option value.
    */
   @Override
   public void setResize(logical resize)
   {
      if (resize.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
                  "**Unable to assign UNKNOWN value to attribute RESIZE on WINDOW widget", 
                  false, false, false);
         return;
      }
      
      if (_isRealized())
      {
         ErrorManager.recordOrShowError(4035, 
                  "**Unable to set RESIZE because the WINDOW widget has been realised.", 
                  false, false, false);
         return;
      }
      
      config.resize = resize.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Enable/disable displaying of scrollbars.
    * 
    * @param   scrollBars
    *          New option value.
    */
   @Override
   public void setScrollBars(logical scrollBars)
   {
      if (_isRealized())
      {
         String msg = "**Unable to set SCROLL-BARS because the WINDOW widget has been realized.";
         ErrorManager.recordOrShowError(4053, msg, true, false, false);
         return;
      }
      
      config.scrollBars = scrollBars.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Enable/disable displaying of status area.
    * 
    * @param   statusArea
    *          New option value.
    */
   @Override
   public void setStatusArea(logical statusArea)
   {
      if (LogicalTerminal.isChui())
      {
         String msg = "**Unable to set attribute STATUS-AREA on WINDOW in the current environment."; 
         ErrorManager.recordOrShowError(4079, msg, false, false, false);
         return;
      }

      if (_isRealized())
      {
         String msg = "**Unable to set STATUS-AREA because the WINDOW widget has been realized.";
         ErrorManager.recordOrShowError(4053, msg, true, false, false);
         return;
      }
      
      config.statusArea = statusArea.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Sets the TITLE writable attribute and pushes the config to the client-side.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   @Override
   public void setTitle(String title)
   {
      config.title = title;
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(boolean s)
   {
      if (config.enabled != s)
      {
         // changing the SENSITIVE attribute does not affect the widget's visibility
         config.enabled = s;
   
         LogicalTerminal.pushWindow(config);
      }
   }

   /**
    * Set (default) background color. 
    */
   public void setBgColor()
   {
      // FIXME what's the default color?
      super.setBgColor(new integer(0));
   }
   
   /**
    * Set (default) foreground color.
    */
   public void setFgColor()
   {
      // FIXME what's the default color?
      super.setFgColor(new integer(0));
   }

   /**
    * Enable/disable KEEP-Z-ORDER option.
    * 
    * @param   keepFrameZOrder
    *          New flag value.
    */
   @Override
   public void setKeepZOrder(logical keepFrameZOrder)
   {
      config.keepFrameZOrder = keepFrameZOrder.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Enable/disable displaying of message area.
    * 
    * @param   messageArea
    *          New flag value.
    */
   @Override
   public void setMessageArea(logical messageArea)
   {
      if (LogicalTerminal.isChui())
      {
         String msg = "**Unable to set attribute MESSAGE-AREA on WINDOW in the current environment."; 
         ErrorManager.recordOrShowError(4079, msg, false, false, false);
         return;
      }

      if (_isRealized())
      {
         String msg = "**Unable to set MESSAGE-AREA because the WINDOW widget has been realized.";
         ErrorManager.recordOrShowError(4053, msg, true, false, false);
         return;
      }
      
      config.messageArea = messageArea.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Get window maximal height.
    * 
    * @return  maximal height.
    */
   @Override
   public decimal getMaxHeight()
   {
      return this.getMaxHeightChars();
   }

   /**
    * Get window maximal width.
    * 
    * @return  maximal width.
    */
   @Override
   public decimal getMaxWidth()
   {
      return this.getMaxWidthChars();
   }

   /**
    * Check if window size can be changed.
    * 
    * @return  <code>true</code> if window size can be changed.
    */
   @Override
   public logical isResize()
   {
      return new logical(config.resize);
   }

   /**
    * Check if window has scrollbars.
    * 
    * @return  <code>true</code> if window has scrollbars.
    */
   @Override
   public logical isScrollBars()
   {
      return new logical(config.scrollBars);
   }

   /**
    * Check if window has status area.
    * 
    * @return  <code>true</code> if window has status area.
    */
   @Override
   public logical isStatusArea()
   {
      return new logical(config.statusArea);
   }

   /**
    * Check if window has KEEP-Z-ORDER option set.
    * 
    * @return <code>true</code> if window has KEEP-Z-ORDER option set.
    */
   @Override
   public logical isKeepZOrder()
   {
      return new logical(config.keepFrameZOrder);
   }

   /**
    * Check if window has message area.
    * 
    * @return <code>true</code> if window has message area.
    */
   @Override
   public logical isMessageArea()
   {
      return new logical(config.messageArea);
   }

   /**
    * Get icon file name previously loaded by {@link #loadIcon(character)}.
    * 
    * @return  icon file name.
    */
   @Override
   public character getIcon()
   {
      return new character(config.iconFileName);
   }

   /**
    * Load icon for window.
    * 
    * @param   iconPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadIcon(character iconPath)
   {
      // call internal version
      return new logical(loadIconInt(iconPath.toStringMessage(), false));
   }

   /**
    * Load icon for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadIcon(String searchPath)
   {
      return loadIcon(new character(searchPath));
   }

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   iconPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadSmallIcon(character iconPath)
   {
      // call internal version
      return new logical(loadIconInt(iconPath.toStringMessage(), true));
   }

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadSmallIcon(String searchPath)
   {
      return loadSmallIcon(new character(searchPath));
   }

   /**
    * Return reference to {@link WindowWidget} as reference to 
    * {@link GenericWidget}.
    * 
    * @return  this.
    */
   public GenericWidget<?> asWidget()
   {
      return this;
   }

   /**
    * Sets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-PIXELS attribute.
    */
   @Override
   public void setMaxHeightPixels(integer max)
   {
      setSizePixels(max.isUnknown() ? null : max.intValue(), SizeAttribute.MAX_HEIGHT_PIXELS);
   }

   /**
    * Sets the MAX-WIDTH-PIXELS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-PIXELS attribute.
    */
   @Override
   public void setMaxWidthPixels(integer max)
   {
      setSizePixels(max.isUnknown() ? null : max.intValue(), SizeAttribute.MAX_WIDTH_PIXELS);
   }
   
   /**
    * Gets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of MAX-HEIGHT-PIXELS attribute
    */
   @Override
   public integer getMaxHeightPixels()
   {
      return new integer(ConfigHelper.getMaxHeightPixels(config));
   }   

   /**
    * Gets the MAX-WIDTH-PIXELS writable attribute.
    *
    * @return current value of MAX-WIDT-PIXELS attribute
    */
   @Override
   public integer getMaxWidthPixels()
   {
      return new integer(ConfigHelper.getMaxWidthPixels(config));
   }
   
   /**
    * Get the message-area's font.
    * 
    * @return   The message-area's font.
    */
   @Override
   public int64 getMessageAreaFont()
   {
      int font = config.messageAreaFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setMessageAreaFont(long fontNum)
   {
      setMessageAreaFont(new int64(fontNum));
   }

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setMessageAreaFont(int64 fontNum)
   {
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized

      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.messageAreaFont = font;
      pushScreenDefinition();
   }
   
   /**
    * Get the status-area's font.
    * 
    * @return   The status-area's font.
    */
   @Override
   public int64 getStatusAreaFont()
   {
      int font = config.statusAreaFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setStatusAreaFont(long fontNum)
   {
      setStatusAreaFont(new int64(fontNum));
   }

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setStatusAreaFont(int64 fontNum)
   {
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized

      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.statusAreaFont = font;
      pushScreenDefinition();
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      boolean delete = super.resourceDelete();

      if (delete)
      {
         LogicalTerminal.deregisterWindow(this);
      }
      
      return delete;
   }
   
   
   /**
    * Perform actual delete of an resource. At the time of this call, it is assumed the resource
    * is valid for deletion (the handle and the resource are both valid).
    * <p>
    * All the direct child windows will be deleted too; and recursively, the entire window tree
    * will be deleted.
    */
   @Override
   public void delete()
   {
      super.delete();

      if (deleted)
      {
         // remove all my children
         List<WindowWidget> children = LogicalTerminal.findDirectChildren(this);
         for (WindowWidget ww : children)
         {
            ww.delete();
         }
      }
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   @Override
   protected boolean validateSize(Number value, SizeAttribute attr)
   {
      boolean isValid = super.validateSize(value, attr);
      
      if (!isValid)
         return false;
      
      if (LogicalTerminal.isChui())
      {
         String msg = String.format(
               "**Unable to set attribute %s on %s in the current environment.", 
               attr.shortName(), type());
         ErrorManager.recordOrShowError(4079, msg, false, false, false);
         return false;
      }
      
      return true;
   }

   /**
    * Sets the value for the attribute WIDTH-CHARS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   @Override
   protected void setWidthCharsWorker(double width)
   {
      super.setWidthCharsWorker(width);
      updateMinMaxSize();
   }

   /**
    * Sets the value for the attribute HEIGHT-CHARS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   @Override
   protected void setHeightCharsWorker(double height)
   {
      super.setHeightCharsWorker(height);
      updateMinMaxSize();
   }

   /**
    * Sets the value for the attribute WIDTH-PIXELS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   @Override
   protected void setWidthPixelsWorker(int width)
   {
      super.setWidthPixelsWorker(width);
      updateMinMaxSize();
   }

   /**
    * Sets the value for the attribute HEIGHT-PIXELS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   @Override
   protected void setHeightPixelsWorker(int height)
   {
      super.setHeightPixelsWorker(height);
      updateMinMaxSize();
      
      config.heightAssigned = true;
   }
   
   /**
    * Sets the value for the attribute MIN-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMinWidthCharsWorker(double width)
   {
      config.minWidthChars = width;
      
      if (width > config.widthChars)
         config.widthChars = width;
      
      if (width > config.virtualWidthChars)
         config.virtualWidthChars = width;
      
      if (width > config.maxWidthChars && config.maxWidthChars != BaseConfig.INV_COORD)
         config.maxWidthChars = width;
   }

   /**
    * Sets the value for the attribute MIN-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMinHeightCharsWorker(double height)
   {
      config.minHeightChars = height;
      
      if (height > config.heightChars)
         config.heightChars = height;
      
      if (height > config.virtualHeightChars)
         config.virtualHeightChars = height;
      
      // when maxHeightPixels is not set (INV_COORD), MAX-HEIGHT is the same as HEIGHT
      if (height > config.maxHeightChars && config.maxHeightChars != BaseConfig.INV_COORD)
         config.maxHeightChars = height;
   }

   /**
    * Sets the value for the attribute MIN-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMinWidthPixelsWorker(int width)
   {
      config.minWidthPixels = width;
      
      if (width > config.widthPixels)
         config.widthPixels = width;
      
      if (width > config.virtualWidthPixels)
         config.virtualWidthPixels = width;
      
      if (width > config.maxWidthPixels && config.maxWidthPixels != BaseConfig.INV_COORD)
         config.maxWidthPixels = width;
   }

   /**
    * Sets the value for the attribute MIN-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMinHeightPixelsWorker(int height)
   {
      config.minHeightPixels = height;
      
      if (height > config.heightPixels)
      {
         config.heightPixels = height;
         config.heightAssigned = true;
      }
      
      if (height > config.virtualHeightPixels)
         config.virtualHeightPixels = height;
      
      // when maxHeightPixels is not set (equals to INV_COORD), MAX-HEIGHT is the same as HEIGHT
      if (height > config.maxHeightPixels && config.maxHeightPixels != BaseConfig.INV_COORD)
      {
         config.maxHeightPixels = height;
         
         if (_isRealized())
            config.unrealizedMaxHeightPixels = BaseConfig.INV_COORD;
      }
   }   
   
   /**
    * Sets the value for the attribute MAX-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMaxWidthCharsWorker(double width)
   {
      config.maxWidthChars = width;
      
      if (width < config.minWidthChars)
         config.minWidthChars = width;
      
      if (width < config.widthChars)
         config.widthChars = width;
      
      if (width < config.virtualWidthChars)
         config.virtualWidthChars = width;
   }

   /**
    * Sets the value for the attribute MAX-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMaxHeightCharsWorker(double height)
   {
      config.maxHeightChars = height;

      if (height < config.minHeightChars)
         config.minHeightChars = height;
      
      if (height < config.heightChars)
         config.heightChars = height;
      
      if (height < config.virtualHeightChars)
         config.virtualHeightChars = height;
   }

   /**
    * Sets the value for the attribute MAX-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMaxWidthPixelsWorker(int width)
   {
      config.maxWidthPixels= width;
      
      if (width < config.minWidthPixels)
         config.minWidthPixels = width;
      
      if (width < config.widthPixels)
         config.widthPixels = width;
      
      if (width < config.virtualWidthPixels)
         config.virtualWidthPixels = width;
   }

   /**
    * Sets the value for the attribute MAX-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMaxHeightPixelsWorker(int height)
   {
      config.maxHeightPixels = height;

      if (height < config.minHeightPixels)
         config.minHeightPixels = height;
      
      if (height < config.heightPixels)
      {
         config.heightPixels = height;
         config.heightAssigned = true;
      }
      
      if (height < config.virtualHeightPixels)
         config.virtualHeightPixels = height;
      
      if (!_isRealized())
      {
         config.unrealizedMaxHeightPixels = height;
      }
      else
      {
         config.unrealizedMaxHeightPixels = BaseConfig.INV_COORD;
      }
   }   

   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthCharsWorker(double width)
   {
      super.setVirtWidthCharsWorker(width);
      
      if (width < config.minWidthChars)
         config.minWidthChars = width;
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightCharsWorker(double height)
   {
      super.setVirtHeightCharsWorker(height);

      if (height < config.minHeightChars)
         config.minHeightChars = height;
   }

   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthPixelsWorker(int width)
   {
      super.setVirtWidthPixelsWorker(width);
      
      if (width < config.minWidthPixels)
         config.minWidthPixels = width;
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightPixelsWorker(int height)
   {
      if (height < config.heightPixels)
         config.heightAssigned = true;
      
      super.setVirtHeightPixelsWorker(height);

      if (height < config.minHeightPixels)
         config.minHeightPixels = height;
      
      
   }

   /**
    * Set the VISIBLE writable attribute.
    *
    * @param    visible
    *           The new value for the VISIBLE attribute.
    */
   @Override
   public void setVisible(boolean visible)
   {
      _setVisible(visible);
      config.hidden = !visible;
      
      if (visible && !LogicalTerminal.isChui())
      {
         if (config.widthPixels > Coordinate.MAX_GUI_WIDTH_PIXELS)
         {
            config.widthPixels = Coordinate.MAX_GUI_WIDTH_PIXELS;
         }
         
         if (config.heightPixels > Coordinate.MAX_GUI_HEIGHT_PIXELS)
         {
            config.heightPixels = Coordinate.MAX_GUI_HEIGHT_PIXELS;
         }
      }
      
      pushScreenDefinition();
   }
   
   /**
    * Get window MENUBAR.
    * 
    * @return  handle.
    */
   @Override
   public handle getMenubar()
   {
      return new handle(this.menubar);
   }

   /**
    * Set window MENUBAR.
    * 
    * @param  menubar
    *         The menu bar handle.
    */
   @Override
   public void setMenubar(handle menubar)
   {
      if (menubar != null && !menubar.isUnknown())
      {
         setMenubar((MenuWidget) menubar.getResource());
      }
      else 
      {
         setMenubar((MenuWidget)null);
      }
   }

   /**
    * Set window MENUBAR.
    * 
    * @param  menubar
    *         The menu widget.
    */
   @Override
   public void setMenubar(MenuWidget menubar)
   {
      if (menubar == null)
      {
         this.config.menubarId = 0;
         this.menubar = null;
      }
      else
      {
         if (menubar.isPopupOnly())
         {
            String msg = "The widget type assigned to WINDOW widget must be a MENUBAR";
            ErrorManager.recordOrShowError(4092, msg, false);
         }

         this.config.menubarId = menubar.getId();
         this.menubar = menubar;
         this.menubar.setOwner(this);
      }
      
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Gets the SCREEN-LINES readable attribute.
    *
    * @return   current value of SCREEN-LINES attribute
    */
   @Override
   public decimal getScreenLines()
   {
      return new decimal(config.screenLines);
   }

   /**
    * For windows, performing a VIEW is the same as setting the VISIBLE attribute to true.
    */
   @Override
   public void view()
   {
      setVisible(true);
   }
   
   /**
    * For windows, performing a HIDe is the same as setting the VISIBLE attribute to false.
    */
   @Override
   public void hide()
   {
      setVisible(false);
   }
   
   /**
    * Accesses the first window in the list of contained windows, having this window as parent.
    *
    * @return   The first widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle firstChild()
   {
      handle child = HandleChain.firstResource(this, LegacyResource.WINDOW);
      if (child.isUnknown())
      {
         child = HandleChain.firstResource(this, LegacyResource.FRAME);
      }

      return child;
   }
   
   /**
    * Accesses the last window in the list of contained windows, having this window as parent.
    *
    * @return   The last widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle lastChild()
   {
      return HandleChain.lastResource(this, LegacyResource.WINDOW);
   }
   
   /**
    * This method implements MOVE-TO-TOP function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveToTop()
   {
      if (LogicalTerminal.isChui())
      {
         ErrorManager.recordOrShowError(4088, 
               String.format(
                  "Progress does not support the %s attribute" + 
                  " for the  %s %s in this display environment", 
                  "MOVE-TO-TOP", type(), widgetName()
               ), 
               false);
         return new logical(false);
      }
      return new logical(LogicalTerminal.moveToTop(config.id.asInt()));
   }

   /**
    * This method implements MOVE-TO-BOTTOM function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveToBottom()
   {
      if (LogicalTerminal.isChui())
      {
         ErrorManager.recordOrShowError(4088, 
               String.format(
                  "Progress does not support the %s attribute" + 
                  " for the  %s %s in this display environment", 
                  "MOVE-TO-BOTTOM", type(), widgetName()
               ), 
               false);
         return new logical(false);
      }
      return new logical(LogicalTerminal.moveToBottom(config.id.asInt()));
   }

   /**
    * Sets the TOP-ONLY option/attribute.
    *
    * @param topOnly
    *        <code>true</code> if the option/attribute is set to ON.
    */
   @Override
   public void setTopOnly(boolean topOnly)
   {
      config.topOnly = topOnly;
      pushScreenDefinition();
   }
   
   /**
    * Sets the TOP-ONLY option/attribute.
    *
    * @param topOnly
    *        <code>true</code> if the attribute is set to ON.
    */
   @Override
   public void setTopOnly(logical topOnly)
   {
      if (topOnly == null || topOnly.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute TOP-ONLY on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setTopOnly(topOnly.booleanValue());
   }
   
   /**
    * Gets the TOP-ONLY attribute's value.
    *
    * @return the value of the attribute
    */
   @Override
   public logical isTopOnly()
   {
      return new logical(config.topOnly);
   }

   /**
    * Gets the WINDOW-STATE attribute.
    *
    * @return  current value of WINDOW-STATE attribute (see <code>WINDOW_STATE_*</code>
    *          constants).
    */
   @Override
   public int64 getWindowState()
   {
      return new int64(config.windowState);
   }

   /**
    * Sets the WINDOW-STATE attribute.
    *
    * @param state
    *        Value of WINDOW-STATE attribute to set (see <code>WINDOW_STATE_*</code>
    *        constants).
    */
   @Override
   public void setWindowState(long state)
   {
      int maxState = config.realized ? WINDOW_STATE_NORMAL : WINDOW_STATE_DELAYED_MINIMIZE;
      if (state < WINDOW_STATE_MAXIMIZED || state > maxState)
      {
         // this is a no-op when value is out of bounds
         return;
      }

      config.windowState = (byte) state;
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Sets the WINDOW-STATE attribute.
    *
    * @param state
    *        Value of WINDOW-STATE attribute to set (see <code>WINDOW_STATE_*</code>
    *        constants).
    */
   @Override
   public void setWindowState(int64 state)
   {
      if (state.isUnknown())
      {
         ErrorManager.recordOrShowError(4083,
            String.format(
               "Unable to assign UNKNOWN value to attribute WINDOW-STATE on %s %s",
               type(), widgetName()),
            true);

         return;
      }

      setWindowState(state.getValue());
   }
   
   /**
    * Indicates an implementation of siblings methods.
    * 
    * @return   Always <code>true</code>. 
    */
   @Override
   protected boolean useHandleChainSiblings()
   {
      return true;
   }

   /**
    * Returns the widget types that can act as parent for this widget.
    *
    * @return   see above.
    */
   @Override
   protected String[] getValidParentTypes()
   {
      // parent must be a window
      return new String[] {LegacyResource.WINDOW};
   }
   
   /**
    * Check if this resource can be reported as chained.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   protected boolean isChained()
   {
      return true;
   }
   
   /**
    * The method returns <code>false</code> always - windows can't be positioned via pixel in Chui.
    * 
    * @return See above.
    */
   protected boolean canSetPixelPositionInChui()
   {
      return false;
   }
   
   /**
    * Sets the window's config {@link WidgetConfig#visible visibility flag}.
    * 
    * @param    visible
    *           The window's visibility state.
    */
   protected void _setVisible(boolean visible)
   {
      config.visible = visible;
      config.assignedVisible = visible;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(WindowConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);
      
      // when window is realized
      if (config.realized && !beforeUpdate.realized)
      {
         // fix min size
         if (config.minHeightChars == BaseConfig.INV_COORD)
            config.minHeightChars = 0;
         
         if (config.minWidthChars == BaseConfig.INV_COORD)
            config.minWidthChars = 0;
         
         // fix max size
         if (config.maxHeightChars == BaseConfig.INV_COORD)
            config.maxHeightChars = config.heightChars;
         
         if (config.maxWidthChars == BaseConfig.INV_COORD)
            config.maxWidthChars = config.widthChars;
      }
   }
   
   /**
    * Load icon for window.  Internal version.
    * 
    * @param    iconSrc
    *           Icon file name.
    * @param    small
    *           Flag indicating usage of the small icon case(16x16 in size).
    *          
    * @return  <code>true</code> if operation was successful.
    */
   private boolean loadIconInt(String iconSrc, boolean small)
   {
      // save current values
      String fileNameSaved = config.iconFileName;
      boolean useSmallIconSaved = config.useSmallIcon; 
      
      // this must be done before sending image bytes because icon size controls icon processing
      // on client side
      config.iconFileName = iconSrc;
      config.useSmallIcon = small;
      LogicalTerminal.pushWindow(config);
   
      // first try to load icon from application jar
      boolean result = LogicalTerminal.setWindowIcon(config.id.asInt(), small, iconSrc);
      
      if (!result)
      {
         // if we are here - there is no icon to load, restore changed values
         config.iconFileName = fileNameSaved;
         config.useSmallIcon = useSmallIconSaved;
         LogicalTerminal.pushWindow(config);
      }
      
      return result;
   }

   /**
    * The method updates min/max size config fields and should be called after
    * window size is changed.
    * <p>
    * Note that min/max size can be indirectly changed only when window size
    * is directly changed, i.e. by assigning WIDTH/HEIGHT attributes for
    * example and not by user input like mouse resize.
    */
   private void updateMinMaxSize()
   {
      // only pixel values are updated, the character equivalents
      // will be recalculated by SyncCoordinatesAspect
      
      if (config.minWidthPixels > config.widthPixels)
      {
         config.minWidthPixels = config.widthPixels;
      }

      if (config.minHeightPixels > config.heightPixels)
      {
         config.minHeightPixels = config.heightPixels;
      }
      
      // when max attribute is never set, its value equals to the width/height
      // (it grows and shrinks with the values of width/height).
      // When set however, the max value grows with width/height but doesn't
      // shrink.
      if (config.maxWidthPixels != BaseConfig.INV_COORD && 
          config.maxWidthPixels < config.widthPixels)
      {
         config.maxWidthPixels = config.widthPixels;
      }
      
      if (config.maxWidthPixels != BaseConfig.INV_COORD &&
          config.maxHeightPixels < config.heightPixels)
      {
         config.maxHeightPixels = config.heightPixels;
      }
      
      if (config.heightPixels > config.unrealizedMaxHeightPixels)
      {
         config.unrealizedMaxHeightPixels = _isRealized() ? 
                                            BaseConfig.INV_COORD : config.heightPixels;
      }
   }
   
   /**
    * In 4GL the widget size (WIDTH/HEIGHT-PIXELS) can be set larger 
    * than the maximum X/Y values. Thus widget area beyond these size limits
    * cannot be addressed. See super implementation for more information.
    * This implementation disables the truncation logic for windows.
    * 
    * @param    size
    *           The value of WIDTH/HEIGHT-PIXELS attributes being assigned.
    *           
    * @return   The truncated size value.   
    */
   @Override
   protected int truncatePixelSize(int size, SizeAttribute attr)
   {
      return size;
   }
   
}