/*
** Module   : AlertBoxImpl.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20111122 Extracted from AlertBox
** 002 SIY 20111221 Fixed constructor visibility.
** 003 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
**                  Replaced size() API with dimension().
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 006 VIG 20150407 Typified Window.getDefaultWindow method call.
** 007 HC  20150612 Implemented GUI ALERT-BOX.
** 008 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 009 HC  20151020 Fixed an abend in GUI ALERT-BOX caused by wrong button initialization.
** 010 IAS 20160315 Support for the DEBUG-ALERT attribute
** 011 VIG 20160315 Removed setUnderscoreOffset() call for button.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.driver.WidgetFactory;
import com.goldencode.p2j.ui.client.event.KeyInput;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * This class is an specific implementation of the AlertBox for CHUI.
 */
public class AlertBoxImpl
extends TitledWindow<ChuiOutputManager>
implements AlertBox<ChuiOutputManager>
{
   /** Width delta - honor the border and a space on each side. */
   private static final int WIDTH_DELTA = 4;

   /** Initial AlertBox width. */
   private static final int INITIAL_WIDTH = 20;

   /** Initial AlertBox width. */
   private static final int INITIAL_HEIGHT = 4;
   
   /** Externalized logic. */
   private AlertBoxCommons<ChuiOutputManager> commons;
   
   /** Cached widget factory. */
   private WidgetFactory<ChuiOutputManager> widgetFactory;
   
   /**
    * Creates a modal dialog with a title and with a frame owner. This
    * constructor creates an AlertBox with a message body which is
    * represented by a content array. The alert box can contain a user
    * defined button set and the specified title. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be strings.
    * In addition, there may be any number of <code>skip</code> elements 
    * which are encoded as a Java <code>Integer</code> instance where the
    * integral value is equivalent to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button (<code>true</code> is the first button, <code>false</code>
    * is the second button and for a 3-button dialog, <code>unknown</code> is 
    * the third button).
    * <p>
    * The result of the user's selection can be read via {@link #getResult}.
    * The mapping of the result value to a button follows the same mapping
    * as used for setting the default button (see above).
    *
    * @param    content
    *           Array of elements which form alert box content.
    * @param    buttonSet
    *           Button set type to be constructed for this AlertBox.
    * @param    var
    *           The variable to be edited.  The initial value sets the 
    *           default button and the resulting value is changed by the
    *           user's press of a button.
    * @param    title
    *           AlertBox title to be rendered at the top of the one line
    *           border.
    * @param    color
    *           Color scheme to use.
    * @param    stack
    *           Progress stack trace
    */
   AlertBoxImpl(Object[] content,
                       int      buttonSet,
                       logical  var,
                       String   title,
                       Color    color,
                       String[] stack)
   {
      super(title, new Insets(0, 0, 0, 0));
      
      widgetFactory = screen().getFactory();
      
      commons = new AlertBoxCommons<ChuiOutputManager>(this,
                                                       this,
                                                       widgetFactory, 
                                                       content, 
                                                       buttonSet, 
                                                       var, 
                                                       color,
                                                       stack);
   }
   
   /**
    * Hide/show widget.
    * <p>
    * If the alert box is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible && parent() == null)
      {
         setParent(WindowManager.<ChuiOutputManager>getDefaultWindow());
      }

      super.setVisible(visible);
   }
   
   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      Dimension scrSize = WindowManager.resolveWindow().getScreenDimension();
      
      double maxWidth = scrSize.width - WIDTH_DELTA; 

      LinkedList<Object> input = new LinkedList<Object>();
      
      double width;
      int height;
      int widthDelta = 0;
      
      boolean hadString   = false;
      boolean hadSkip     = false;
      boolean leadingSkip = false;

      do
      {
         width  = INITIAL_WIDTH;
         height = INITIAL_HEIGHT;

         boolean last = false;
         
         input.clear();
         
         for (Object item : commons.content)
            input.add(item);
         
         ListIterator<Object> iter = input.listIterator();
         Object item               = null;
         ArrayList<Object> list    = new ArrayList<Object>();
         StringBuilder sb          = new StringBuilder();

         // Calculate the width and height of the dialog, concatenate the text
         // elements as needed.
         // Note that the strings may have embedded newline characters which 
         // act as skips.
         
         int lastSkipHeight  = 1;
         int maxlen          = 0;
         
         boolean hadExplicitCR = false;

         while (iter.hasNext())
         {
            // get the next item from the input list
            item = iter.next();

            // Progress has a silly/stupid algorithm for appending a space
            // such that if there is any following element after a string, a
            // space is added (whether that element is a skip or a string)
            if (last)
            {
               if (sb.length() > maxlen)
                  maxlen = sb.length();
               
               if (!hadExplicitCR)
               {
                  // this append must be done only if the message text had no
                  // explicit CR character to split the message into multiple
                  // lines
                  sb.append(" ");
               }
            }
            
            hadExplicitCR = false;

            if (item instanceof String)
            {
               hadString      = true;
               lastSkipHeight = 1;
               
               // first of all, handle embedded newlines as follows:
               // - split the item into 3 pieces: start, newline, rest;
               // - push the rest back into the list first, if not empty;
               // - push the skip(1) back into the list;
               // - handle the start as a string, if not empty, or else loop.
               String sitem = (String)item;
               int nl = sitem.indexOf(10);
               boolean notLast = false;

               // set the flag only if a CR is encountered
               hadExplicitCR = nl != -1;

               if ((nl == -1 || nl > maxWidth) && (sitem.length() > maxWidth))
               {
                  nl = sitem.lastIndexOf(' ', (int) maxWidth + 1);

                  if (nl <= 0)
                     nl = (int) maxWidth + 1;

                  notLast = true;
               }

               if (nl != -1)
               {
                  // handle the rest of the string first
                  if (nl < sitem.length() - 1)
                  {
                     if (nl == maxWidth + 1)
                        iter.add(" " + sitem.substring(nl + 1));
                     else
                        iter.add(sitem.substring(nl + 1));
                     iter.previous();
                  }

                  // handle the newline itself
                  iter.add(new Integer(0));
                  iter.previous();

                  // handle the start of the string
                  if (nl == 0)
                  {
                     sb.setLength(sb.length() > 0 ? sb.length() - 1 : 0);
                     continue;
                  }
                  
                  sitem = sitem.substring(0, nl);
               }

               // if the most recent element was a string, this is essentially
               // a concatenation, else it is adding for the first time
               sb.append(sitem);

               // we only increment the height when this is not a concatenation
               if (!last)
               {
                  height++;
               }

               // remember that the last element was a string
               if (!notLast)
                  last = true;
            }
            else if (item instanceof Integer)
            {
               if (!hadString)
                  leadingSkip = true;
               
               Integer ht = (Integer) item;

               // we have hit a skip, increment the height of the dialog
               
               if (!hadString)
               {
                  int savedSkipHeight = ht.intValue();
                  
                  if (lastSkipHeight > 0 ||
                      (lastSkipHeight == 0 && ht.intValue() > 0))
                  {
                     ht = new Integer(ht.intValue() + 1);
                  }
                  
                  lastSkipHeight = savedSkipHeight;
               }

               height += ht.intValue();
               
               // the last element was not a string
               last = false;

               // save off any buffer that was there
               if (sb.length() > 0)
               {
                  if (!iter.hasNext()       &&
                      width > INITIAL_WIDTH && 
                      sb.length() >= maxlen)
                  {
                     sb.append(" ");
                  }
                  
                  // increment our width if this is the largest string
                  if (sb.length() > width)
                  {
                     width = sb.length();
                  }

                  list.add(sb.toString());
                  sb.setLength(0);
               }

               // remember our skip
               list.add(ht);
               hadSkip = true;
            }
            else
            {
               throw new RuntimeException("Unknown entity of class " + 
                                          item.getClass().getName() + 
                                          " in the message box content array.");
            }
         }

         // save off any final buffer that is there
         if (sb.length() > 0)
         {
            // increment our width if this is the largest string

            if (hadSkip               && 
                width > INITIAL_WIDTH && 
                sb.length() >= maxlen)
            {
               sb.append(" ");
            }

            if (sb.length() > width)
            {
               width = sb.length();
            }

            list.add(sb.toString());
         }
         
         commons.content = list.toArray();
         
         if (leadingSkip && maxlen >= INITIAL_WIDTH)
            widthDelta = 1;
      }
      // safe, comparison of scaled doubles
      while(width > maxWidth + 1);
      
      // convert the modified array back into our contents
      
      // safe, comparison of scaled doubles
      if (width > maxWidth)
         width = maxWidth;
      
      if (width == (INITIAL_WIDTH + 1) && !leadingSkip && !hadSkip)
         width--;
      
      // honor the border and a space on each side
      width += WIDTH_DELTA + widthDelta;
      
      // align left to the whole character
      setLocation((int) (scrSize.width - width) / 2, (int) (scrSize.height - height) / 2);
      setSize(width, height);

      // adjust title size and X position
      if (hasTitle() && title().length() > width)
      {
         // if the title goes outside the dialog width, trim it. also, if 
         // trimming inside a word, remove it completely.
         String part1 = title().substring(0, (int) width);
         String part2 = title().substring((int) width);
         if (part1.lastIndexOf(' ') != -1 &&
             (part2.indexOf(' ') != 0 ||
              part2.indexOf(' ') == -1)   &&
             part2.length() > 0)
         {
            part1 = part1.substring(0, part1.lastIndexOf(' ') + 1);
            commons.forceTitleLeftAlign = true;
         }
         
         setTitle(part1);
      }
      
      setLayout(null);
      
      int row = 0;
      
      for (int i = 0; i < commons.content.length; i++)
      {
         if (commons.content[i] instanceof String)
         {
            String s = (String) commons.content[i];
            
            Label<ChuiOutputManager> ls = widgetFactory.createLabel(WidgetId.nextID(),
                                                                    s, 
                                                                    -1);
            ls.config().dcolor = commons.alertColor;
            ls.setRightAligned(false);
            add(ls);
            
            int pos = ((int) width - s.length())/2;
            
            ls.setLocation(pos, ++row);
         }
         else if (commons.content[i] instanceof Integer)
         {
            row += ((Integer) commons.content[i]).intValue();
         }
      }
      
      List<Button<ChuiOutputManager>> buttons = commons.createButtonSet();
      
      for (Button<ChuiOutputManager> button : buttons)
      {
         if (button != null)
         {
            // TODO: move this to show()
            button.setEnabled(true);
            button.setVisible(true);
         }
      }
      
      // compute the button location
      double btn_width = buttons.get(0).width();
      
      int buttonsNum = buttons.size();
      
      if (buttonsNum >= 2)
         btn_width += buttons.get(1).width() + 1;
      
      if (buttonsNum >= 3)
         btn_width += buttons.get(2).width() + 1;
      
      // place the buttons
      // align left to the whole character
      double xCurr = (int) (width - btn_width) / 2;
      buttons.get(0).setLocation(xCurr, height - 2);
      
      if (buttonsNum >= 2)
      {
         xCurr += buttons.get(0).width() + 1;
         buttons.get(1).setLocation(xCurr, height - 2);
      }
      
      if (buttonsNum >= 3)
      {
         xCurr += buttons.get(1).width() + 1;
         buttons.get(2).setLocation(xCurr, height - 2);
      }
      
      // calculate the default button using the value of the edited variable
      // (true == 0, false == 1 and for a 3-button dialog, unknown == 2)
      int focus = 2;
      
      if (!commons.result.isUnknown() || 
          commons.buttonSet != LogicalTerminal.BTN_YES_NO_CANCEL)
      {
         focus = (commons.result.booleanValue() ? 0 : 1);
      }
      
      // service focus change call; is safe
      buttons.get(focus).parent().setFocus(buttons.get(focus));
   }

   /**
    * The main AlertBox renderer. Draws the contents of the AlertBox and
    * requests the focus to the one of the button within the container.
    */
   @Override
   public void draw()
   {
      // draw the enclosing frame
      Color dc = commons.alertColor;

      clear(location(), dimension(), dc);
      screen().boxToScreen(location(), dimension(), dc);
                          
      // draw the horizontal line
      screen().at(location().offset(2, dimension().height - 3))
              .hLine(dimension().width - 4, dc);

      // draw the title
      if (hasTitle())
      {
         // compute the title's X position, depending if was trimmed or not
         int x = 0;
         
         if (!commons.forceTitleLeftAlign)
            x = (int) (dimension().width - title().length()) / 2;
         
         Point offsetPoint = location().offset(x, 0);
         screen().at(offsetPoint).append(title(), dc);
      }
      
      // draw all the contained components
      Widget<ChuiOutputManager> focus = null;
      
      for (Widget<ChuiOutputManager> widget : widgets())
      {
         if (widget.isVisible())
         {
            if (widget.hasFocus())
               focus = widget;
            else
               widget.draw();
         }
      }
      
      if (focus != null)
         focus.draw();
   }
   
   /**
    * Hide this widget and remove it.
    */
   @Override
   public void hide()
   {
      super.hide();
      
      // this is a special case: if the alert box is hidden, then its no longer available.
      WindowManager.remove(this);
   }
   
   /**
    * Move the input focus to the first widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the last one within the container and
    * we have to simulate the Progress behavior in which the next focused
    * widget will be the first widget which can receive a focus in a
    * container.
    */
   @Override
   public void nextFocus()
   {
      commons.nextFocus();
   }
   
   /**
    * Move the input focus to the last widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the first one within the container and
    * we have to simulate the Progress behavior in which the previous focused
    * widget will be the last widget which can receive a focus in a container.
    */
   @Override
   public void prevFocus()
   {
      commons.prevFocus();
   }
   
   /**
    * Find the button that matches a passed key event.
    * 
    * @param    keyEvent
    *           Event to check.
    * 
    * @return   <code>true</code> if matching button is found and event is
    *           processed.
    */
   @Override
   public boolean findButton(KeyInput keyEvent)
   {
      return commons.findButton(keyEvent);
   }
   
   /**
    * Set the result code, hide the dialog and flag input as complete.
    *
    * @param    rc
    *           The result code for the alert box.
    */
   @Override
   public void finished(logical rc)
   {
      commons.finished(rc);
   }
   
   /**
    * Get status of the input processing.
    * 
    * @return  current status of the input processing.
    */
   @Override
   public boolean inputDone()
   {
      return commons.inputDone();
   }

   /**
    * Helper method to get the edited result when finished.
    * 
    * @return   The result of the editing operation which was set depending
    *           on the pressed button.
    */
   @Override
   public logical getResult()
   {
      return commons.getResult();
   }

   /**
    * Create window for the stack trace
    * 
    * @param title
    *        window title
    * 
    * @return window for the stack trace
    */
   @Override
   public TitledWindow createStackTraceWindow(String title)
   {
      final TitledWindow win = new TitledWindow(title, new Insets(0,0,0,0));
      ThinClient.getInstance().eventBracket(() -> 
         {
            win.setEnabled(true);
         }
      );
      return win; 
   }

   /**
    * Create effectively invisible widget (the gap) for the stack trace dialog
    * to avoid explicit coordinates' usage 
    * TODO: at this moment the layout manager doesn't support this approach.
    *   
    * @param registry 
    *        widget registry
    * 
    * @return the gap widget
    */
   @Override
   public Widget<ChuiOutputManager> createGap(WidgetRegistry<ChuiOutputManager> registry)
   {
      RectangleConfig rcfg = new RectangleConfig(WidgetId.nextID());
      rcfg.clientHeightChars = 1;
      rcfg.clientWidthChars = 1;
      rcfg.filled = false;
      rcfg.edgeChars = 0;
      registry.reconstructWidget(rcfg);
      return registry.getComponent(rcfg.id); 
   }
}