/*
** Module   : BrowseImpl.java
** Abstract :
**
** Copyright (c) 2012-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111124 Extracted CHUI-specific part from Browse.
** 002 CA  20121022 Added support for optional query and column-list
**                  clauses.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 HC  20141021 Implemented window move/position support.
** 008 SVL 20141209 Added support for NO-UNDERLINE and for locked columns.
** 009 SVL 20150212 Browse reworked.
** 010 HC  20150323 Removed business logic from config classes.
** 011 SVL 20150420 Removed invalid REFRESHABLE state check in draw(), added notifyColumnsUpdated.
** 012 SVL 20150508 Added support for rows inserted using insertRow.
** 013 CA  20150722 Refactored so that browse column configurations are transferred via
**                  pushScreenDefinition calls.
** 014 SVL 20150722 Minor refactoring.
** 015 SVL 20150821 Do not draw dynamic browse until it is realized.
** 016 SVL 20151013 Moved most of ChUI-specific functionality from Browse.
** 017 SVL 20151019 Added notifyBrowseScrolled() notification.
** 018 SVL 20151027 Added hasRowMarkers.
** 019 SVL 20151230 Moved startEdit to the parent class.
** 020 SVL 20160325 Added isFirstLastRowCopyQuirk.
** 021 SVL 20160922 Code refactoring because of switching from BrowseConfig.queryId to
**                  BrowseConfig.queryAssigned.
** 022 SVL 20161029 Query model initialization was removed from draw().
** 023 OM  20170320 Exposed hasRowMarkers() as public.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific portion of implementation of Progress BROWSE widget.
 */
@SuppressWarnings("unchecked")
public class BrowseImpl
extends Browse<ChuiOutputManager>
{
   /** Default DOWN value for ChUI browse. */
   private static final int DEFAULT_CHUI_DOWN = 10;

   /** The renderer for columns labels area. */
   private LabelColumnRenderer rndrLabels = null;
   
   /** The renderer for a data area with values. */
   private DataColumnRenderer rndrData = new DataColumnRenderer();
   
   /** The renderer for data lines which are outside data model. */
   private SpaceColumnRenderer rndrSpace = new SpaceColumnRenderer();

   /** Number of character character columns in the displayable page. */
   protected double ilPageWidth = 0;

   /**
    * Number of space columns on screen between the data area and the right
    * border.
    */
   protected double ilRightSpaceWidth = 0;

   /** Separator line length */
   protected double separatorLen = 0;

   /**
    * Constructs the CHUI-specific version of Browse widget.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public BrowseImpl(WidgetId id, BrowseConfig cfg)
   {
      super(id, cfg);
      currRowSelected = true;
   }

   /**
    * Sets position of terminal's cursor. Redirects this positioning to the
    * in-place editor if it is present. Otherwise does nothing because BROWSE
    * itself has no purpose for the hardware cursor.
    */
   @Override
   public void drawCaret()
   {
      FillIn<ChuiOutputManager,?> editor = getActiveEditor();
      
      if (editor != null)
         editor.drawCaret();
   }

   /**
    * Draws the widget: border, header, selection marks and line, internal
    * FillIn etc.
    */
   @Override
   public void draw()
   {
      // Do not draw a browse that hasn't been realized.
      if (config.dynamic && !config.realized)
      {
         return;
      }

      Color clr = UiUtils.getDColor(this);
      colorCommon = clr;
      colorHSep = clr.withUnderline();
      colorCurrent = clr.withReverse();

      // if frame is not alive, do not generate events and clear browse
      // content
      Widget<ChuiOutputManager> frame = (Widget<ChuiOutputManager>) UiUtils.locateFrame(this);
      
      if (frame != null && !frame.isVisible())
         return;
      
      if (frame instanceof Frame && !((Frame<ChuiOutputManager>) frame).isAlive())
      {
         resetRowDisplayState();
         this.dcData = null;
         setCurrentRow(0);
         notifyBrowseScrolled();
         repaint();
      }
      else
      {
         // else, trigger events, before drawing
         triggerRowDisplayEvents();
      }

      // save the row from where drawing was started last time
      oldDisplayRow = getScrollRow();

      Point origin = screenLocation();
      
      // in chui char coordinates are always whole numbers
      int x = (int) origin.x;
      int y = (int) origin.y;

      ChuiOutputManager screen = screen();
      
      if (config.box)
      {
         screen.boxToScreen(origin, dimension(), clr);
         
         // in chui char coordinates are always whole numbers
         int width = (int) width();

         Color tcl = config.titleDColor;

         if (config.title != null && config.title.length() > 0)
         {
            String titleText = (config.title.length() > width) ?
                                     config.title.substring(0, width) : 
                                     config.title;
                                     
            int titleOffset = (width - titleText.length()) / 2;
            screen.at(x + titleOffset, y).append(titleText, tcl);
         }
                           
         ++x;
         ++y;
      }

      if (drawLabels)
      {
         if (rcLabels == null)
         {
            if (rndrLabels == null)
               rndrLabels = new LabelColumnRenderer();
            else
               rndrLabels.clear();

            renderAbstractLine(rndrLabels, -1);
            rndrLabels.printResult();
         }

         // ilLabelsHeight is rounded, the comparison is safe
         for (int i = 0; i < ilLabelsHeight; ++i)
         {
            screen.at(x, y).append(rcLabels[i], colorCommon);
            ++y;
         }

         if (!config.noUnderline)
         {
            int lockedColumnsSeparatorPos = 0;
            if (config.numLockedColumns > 0)
            {
               int availWidth = getAvailableWidth();
               if (availWidth > 0)
                  lockedColumnsSeparatorPos = (int) separatorLen - availWidth - 1;
            }

            if (lockedColumnsSeparatorPos > 0)
            {
               // If there are locked columns, there a space in underline appearing under
               // column labels, which separates locked columns from non-locked.
               screen.at(x, y).
                     append("  ", colorCommon).
                     hLine(lockedColumnsSeparatorPos, colorCommon).
                     append(" ", colorCommon).
                     hLine(separatorLen - lockedColumnsSeparatorPos - 1, colorCommon);
            }
            else
            {
               screen.at(x, y).append("  ", colorCommon).hLine(separatorLen, colorCommon);
            }

            // ilRightSpaceWidth is rounded, the comparison is safe
            if (ilRightSpaceWidth > 0)
            {
               // this cycle usually works only 2 times so it is more effective
               // than to generate special string
               for (int i = 0; i < ilRightSpaceWidth; ++i)
                  screen.append(' ', colorCommon);
            }

            ++y;
         }
      }

      hadDraw = true;
      
      int rowsLeft = getCacheSize();

      FillIn<ChuiOutputManager,?> editor = getActiveEditor();
      boolean readOnly = (editor == null || editor.config().readOnly);
      boolean focused = (hasFocus() && isEnabled()) || pseudoFocus;

      int down = getDown();
      int rowsOnPageData = Math.min(down, rowsLeft);
      config.numIterations = rowsOnPageData;

      int index = 0;
      
      for (; rowsOnPageData > 0; --rowsOnPageData)
      {
         rndrData.newRow(readOnly, focused, getCachedRowNumber(index));
         screen.at(x, y);
         renderAbstractLine(rndrData, index);
         ++y;
         ++index;
      }
      
      for (int i = down - rowsLeft; i > 0; --i)
      {
         screen.at(x, y);
         int row = getCachedRowNumber(index);
         if (row == -1 && index == 0)
         {
            // highlight the first row of an empty browse
            row = 0;
         }
         rndrSpace.newRow(readOnly, focused, row);
         renderAbstractLine(rndrSpace, index);
         ++y;
         ++index;
      }

      if (editor != null)
         editor.draw();

      if (!focused || editor == null)
      {
         screen.at(0, 0);
      }
   }

   /**
    * Lay out the browse.
    */
   @Override
   public void doLayout()
   {
      if (config.performLayout)
      {
         if (config.down <= 0)
         {
            config.down = DEFAULT_CHUI_DOWN;
         }

         super.doLayout();
      }
   }

   /**
    * Sets new widget's size.
    *
    * @param width
    *        The width component of the new size.
    * @param height
    *        The height component of the new size.
    *
    * @throws IllegalArgumentException
    *         When new size is not large enough.
    */
   @Override
   public void setSize(double width, double height)
   {
      originalWidth = width;
      originalHeight = height;

      width = Coordinate.scale(width);
      height = Coordinate.scale(height);

      if (width == 0 || height == 0)
         return;

      // mark columns
      ilPageWidth = width - 2;
      int ilPageRowsCount = (int) height;

      BrowseColumnConfig[] cols = getColumns();

      double colWidth = 0;

      for (int i = 0; i < cols.length; i++)
      {
         colWidth += cols[i].widthChars + 1;
      }

      colWidth = Coordinate.scale(colWidth);

      if (cols.length > 0)
      {
         // the last column has no vertical separator
         colWidth--;
      }

      if (config.box)
      {
         // if no columns specified, use the explicit width set by the user.
         // safe, comparison of scaled doubles
         ilPageWidth = cols.length > 0
               ? Math.min(colWidth, width - 4)
               : width - 4;
         ilPageRowsCount -= 2;
      }

      if (drawLabels)
      {
         ilPageRowsCount -= ilLabelsHeight + (config.noUnderline ? 0 : 1);
      }

      // safe, comparison of scaled doubles
      if (ilPageWidth < 1 || ilPageRowsCount < 2)
      {
         throw new IllegalArgumentException
               ("Too little browse size: " + width + "x" + height +
                     ". It implies the data area size: " + ilPageWidth + "x" +
                     ilPageRowsCount);
      }

      ilRightSpaceWidth = Coordinate.scale(ilPageWidth - colWidth);

      // safe, comparison of scaled doubles
      if (ilRightSpaceWidth <= 0)
      {
         ilRightSpaceWidth = 0;
      }

      ilFirstScrollCol = cols.length;
      double w = -1;
      int availableWidth = getAvailableWidth();

      for (int i = 0; i < cols.length; ++i)
      {
         w += cols[i].widthChars + 1;

         // safe, comparison of scaled doubles
         if ((w = Coordinate.scale(w)) > ilPageWidth)
         {
            if (cols[i].widthChars > availableWidth)
            {
               // there is not enough space to entirely display the first scroll column, so
               // the next column becomes the first scroll column
               ilFirstScrollCol = i + 1;
            }
            else
            {
               ilFirstScrollCol = i;
            }
            break;
         }
      }

      separatorLen = ilPageWidth - (config.box ? 0 : 2);
      config.down = ilPageRowsCount;

      super.setSize(width, height);
   }

   /**
    * The method is called after the configuration associated with
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    *
    * @param   beforeUpdate
    *          Config reference capturing the config state before
    *          the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BrowseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);

      if (config.heightChars == 0 && config.widthChars == 0)
      {
         setDown(config.down);
      }

      setSelectionMultiple(config.multiple);
      setReadOnly(config.readOnly);

      doLayout();

      if (config.queryAssigned)
      {
         initQueryAssigned();
      }
   }

   /**
    * Set number of rows to display data.
    *
    * @param   down
    *          Number of data rows.
    */
   public void setDown(int down)
   {
      if (browseModel == null)
         return;

      // account for the title
      if (config.box)
         down += 2;

      // account for labels
      if (drawLabels)
         down += ilLabelsHeight + (config.noUnderline ? 0 : 1);

      setSize(width(), down);
   }

   /**
    * Notify the browse that some columns have been added to it.
    */
   @Override
   public void notifyColumnsUpdated()
   {
      int newScrollRow = getCachedRowNumber(1);
      if (newScrollRow == -1)
      {
         // actually in this case in 4GL ChUI at this point browse is unusable
         newScrollRow = getCachedRowNumber(0);
      }

      resetColumnsCache();
      rcLabels = null;
      rndrLabels = null;
      updateLabelsHeight();

      WidgetRegistry widgetRegistry = ThinClient.getInstance().registry();
      BrowseColumnConfig[] cols = getColumns();
      for (BrowseColumnConfig col : cols)
      {
         Widget component = widgetRegistry.getComponent(col.id);
         if (component == null)
         {
            widgetRegistry.reconstructWidget(col);
         }
      }

      super.notifyColumnsUpdated(false);

      // This is ChUI-only quirk
      if (newScrollRow != -1)
      {
         boolean editing = getActiveEditor() != null;
         if (editing)
         {
            undoValue(false);
         }

         scrollToRow(newScrollRow, true, false, true);

         for (int i = 0; i < cols.length; ++i)
         {
            if (!cols[i].readOnly)
            {
               currColumn = i;
               break;
            }
         }

         if (editing)
            startEdit();
      }
      else
      {
         refresh(false);
      }
   }

   /**
    * Get width of the browse area available for scrollable columns, i.e. browse page width minus
    * width of locked columns.
    *
    * @return see above.
    */
   private int getAvailableWidth()
   {
      BrowseColumnConfig[] cols = getColumns();

      int lockedWidth = 0;
      for (int i = 0; i < config.numLockedColumns; i++)
         lockedWidth += cols[i].widthChars + 1;

      return Math.max(0, (int) ilPageWidth - lockedWidth);
   }

   /**
    * Initialize editor fill-in config.
    *
    * @param cfg
    *        Editor fill-in config.
    */
   @Override
   protected void initEditorConfig(FillInConfig cfg)
   {
      BrowseColumnConfig[] cols = getColumns();

      // safe, comparison of scaled doubles
      double width = Math.min(cols[currColumn].widthChars, ilPageWidth);
      cfg.widthChars = width;

      double x = 2;
      double y = getViewportRow();

      if (config.box)
      {
         ++x;
         ++y;
      }

      if (drawLabels)
      {
         y += ilLabelsHeight + (config.noUnderline ? 0 : 1);
      }

      if (scrollColumn == 0)
      {
         for (int i = 0; i < currColumn; ++i)
         {
            x += cols[i].widthChars + 1;
         }
      }
      else
      {
         x += ilPageWidth - width;
         for (int i = scrollColumn; i > currColumn; --i)
         {
            x -= cols[i].widthChars + 1;
         }
      }

      cfg.clientColumn = x;
      cfg.clientRow = y;
   }

   /**
    * Add editor fill-in to the browse.
    *
    * @param editor
    *        Editor fill-in.
    */
   @Override
   protected void addEditor(FillIn editor)
   {
      add(editor);
   }
   
   /**
    * If the browse has row markers drawn. Note that the returned value may differ from
    * {@link com.goldencode.p2j.ui.BrowseWidget#hasRowMarkers()}.
    *
    * @return <code>true</code> if the browse has row markers drawn.
    */
   public boolean hasRowMarkers()
   {
      return isEditPossible() && super.hasRowMarkers();
   }
   
   /**
    * Setups scrolling to fully (as possible) show the given column.
    *
    * @param   cellColumn
    *          The cell's column index in model.
    */
   @Override
   protected void columnToView(int cellColumn)
   {
      int newScrollColumn = scrollColumn;

      if (cellColumn == 0)
      {
         newScrollColumn = 0;
      }
      else if (cellColumn != newScrollColumn)
      {
         BrowseColumnConfig[] cols = getColumns();

         if (cellColumn > newScrollColumn)
         {
            newScrollColumn = cellColumn;
         }
         else
         {
            double usedX = cols[newScrollColumn].widthChars;
            for (int i = cellColumn; i < newScrollColumn; ++i)
            {
               usedX += cols[i].widthChars + 1;
            }

            int availableWidth = getAvailableWidth();
            // safe, comparison of scaled doubles
            while ((usedX = Coordinate.scale(usedX)) > availableWidth)
            {
               usedX -= cols[newScrollColumn].widthChars + 1;
               --newScrollColumn;
               if (newScrollColumn == cellColumn)
               {
                  // new scroll column may be wider than all browse
                  break;
               }
            }
         }

         if (newScrollColumn < ilFirstScrollCol)
         {
            newScrollColumn = 0;
         }
      }

      if (newScrollColumn != scrollColumn)
      {
         scrollColumn = newScrollColumn;
         rcLabels = null;
      }
   }

   /**
    * Determine if this browse should have the following quirk: is a new row is inserted before
    * the first row or after the last row, the value of the current column is copied from the
    * first/last row to the new row.
    *
    * @return <code>true</code>
    */
   protected boolean isFirstLastRowCopyQuirk()
   {
      return false;
   }

   /**
    * Renders a line from horizontally scrollable area.
    * 
    * @param renderer
    *        Implementation of various rendering's parts.
    * @param rowIdx
    *        The screen index of the row currently being displayed
    */
   private void renderAbstractLine(ColumnRenderer renderer, int rowIdx)
   {
      renderer.addMarks();
      BrowseColumnConfig[] cols = getColumns();
      // columns may have not been specified, so exit early if so
      if (cols == null || cols.length == 0)
         return;

      // reset columns
      for (int i = 0; i < cols.length; i++)
         cols[i].column = -1;

      int availWidth = getAvailableWidth();

      if (scrollColumn == 0 || availWidth <= 0 || scrollColumn < config.numLockedColumns)
      {
         // no scrolling, the first column is fully visible
         // in chui char coordinates are always whole numbers
         int width = (int) ilPageWidth;
         
         BrowseColumnConfig col = cols[0];
         
         // in chui char coordinates are always whole numbers
         int colWidth = (int) Math.min(col.widthChars, width);

         if (rowIdx >= 0        &&
             colors != null     &&
             rowIdx < colors.length)
         {
            Color color = colors[rowIdx][0];
            col.columnDColor = color == null ? Color.DEFAULT : color;
         }
         renderer.addData(col, 0, colWidth, 0);
         width -= colWidth;
         int i;
         
         for (i = 1; i < cols.length; ++i)
         {
            if (width > 0)
            {
               renderer.addSeparator();
               --width;
            }

            if (width <= 0)
            {
               break;
            }
            
            col = cols[i];
            
            // in chui char coordinates are always whole numbers
            colWidth = (int) col.widthChars;

            if (width < colWidth && i == config.numLockedColumns)
            {
               // Special case: no scrolling, some locked columns and single non-locked
               // column which doesn't entirely fit into available width.
               colWidth = width;
            }

            int scroll = (colWidth > width) ? width - colWidth : 0;

            if (rowIdx >= 0        &&
                colors != null     &&
                rowIdx < colors.length)
            {
               Color color = colors[rowIdx][i];
               col.columnDColor = color == null ? Color.DEFAULT : color;
            }
            renderer.addData(col, i, colWidth, scroll);
            width -= colWidth;
         }
         
         config.numVisibleColumns = i;
      }
      else
      {
         int nCols = 1;

         // Draw locked columns
         if (config.numLockedColumns > 0)
         {
            for (int i = 0; i < config.numLockedColumns; i++)
            {
               BrowseColumnConfig col = cols[i];
               renderer.addData(col, i, (int) col.widthChars, 0);
               renderer.addSeparator();
               nCols++;
            }
         }

         // scrolling active, scroll column is on the right of view area
         BrowseColumnConfig scol = cols[scrollColumn];

         // in chui char coordinates are always whole numbers
         int scolWidth = (int) scol.widthChars;
         
         if (availWidth <= scolWidth)
         {
            // column is at least as wide as visible area
            scolWidth = availWidth;
         }
         else
         { 
            // view area contains also some columns before scroll column
            int fcolWidth = (int) (availWidth - scolWidth - 1);
            int fcolIndex = scrollColumn - 1;
            
            while (true)
            {
               // in chui char coordinates are always whole numbers
               int w = (int) (fcolWidth - cols[fcolIndex].widthChars - 1);
               if (w < 0)
               {
                  break;
               }

               fcolWidth = w;
               --fcolIndex;
            }

            BrowseColumnConfig col;
            
            if (fcolWidth > 0) // it may equal 0 also
            {
               col = cols[fcolIndex];
               int fullWidth = (int) col.widthChars;

               renderer.addData(col, fcolIndex, (int) fullWidth, (int) (fullWidth - fcolWidth));
               nCols++;
            }
            renderer.addSeparator();
            
            for (int i = fcolIndex + 1; i < scrollColumn; ++i)
            {
               col = cols[i];
               
               // in chui char coordinates are always whole numbers
               renderer.addData(col, i, (int) col.widthChars, 0);
               renderer.addSeparator();
               nCols++;
            }
         }

         renderer.addData(scol, scrollColumn, scolWidth, 0);
         config.numVisibleColumns = nCols;
      }
      
      // in chui char coordinates are always whole numbers
      if ((int) ilRightSpaceWidth > 0)
      {
         renderer.addRightSpaces((int) ilRightSpaceWidth);
      }
   }

   /**
    * Outputs several spaces to the terminal.
    *
    * @param   width
    *          Number of spaces.
    * @param   color
    *          Drawing color.
    */
   private void printSpaces(int width, Color color)
   {
      for (int i = width; i > 0; --i)
      {
         screen().append(' ', color);
      }
   }
   
   /**
    * Interface for renderers of horizontally scrollable rows. The place for
    * saving the result is up to implementation. The result may consist from
    * several character rows.
    */
   private interface ColumnRenderer
   {
      /**
       * Renders data from an one column.
       * 
       * @param   col
       *          Column to render.
       * @param   colIndex
       *          Index of the column.
       * @param   colWidth
       *          The full width for formatting.
       * @param   scroll
       *          Scrolling correction. Means the number of characters which
       *          should be erased from the left (if positive) or from the
       *          right (if negative). This scroll differs from the scroll in
       *          <code>DisplayFormat.formatScreenValue()</code> because it
       *          means the scroll of the whole browse column not the scroll
       *          of screen value inside the column.
       */
      public void addData(BrowseColumnConfig col, int colIndex, int colWidth, 
                          int scroll);

      /**
       * Renders an editable line and a selected line marks.
       */
      public void addMarks();

      /**
       * Renders a separator character.
       */
      public void addSeparator();

      /**
       * Renders several spaces to make right space column.
       * 
       * @param   width
       *          The number of spaces.
       */
      public void addRightSpaces(int width);
   }

   /**
    * Renders the labels area. Places the internal result into the internal
    * buffers. It is expected that <code>clear()</code> method is called
    * before rendering and <code>printResult()</code> is called to finish
    * it.
    */
   private class LabelColumnRenderer
   implements ColumnRenderer
   {
      /** Temporary buffers for render result. */
      private StringBuffer[] bufs;

      /**
       * Constructor. Initializes from the current labels. Object should be
       * recreated if labels are changed.
       */
      public LabelColumnRenderer()
      {
         bufs = new StringBuffer[ilLabelsHeight];
         for (int i = 0; i < bufs.length; ++i)
         {
            bufs[i] = new StringBuffer(ilLabelsHeight);
         }
      }

      /**
       * Clears an old content.
       */
      public void clear()
      {
         for (int i = 0; i < bufs.length; ++i)
         {
            bufs[i].setLength(0);
         }
      }

      /**
       * Copies buffers content into <code>rcLabels</code> member.
       */
      public void printResult()
      {
         rcLabels = new String[bufs.length];
         for (int i = 0; i < bufs.length; ++i)
         {
            rcLabels[i] = bufs[i].toString();
         }
      }

      /**
       * @param   colIndex
       *          Is not used.
       * @see  ColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig, int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col, int colIndex, int colWidth, int scroll)
      {
         String[] label = getLabelVar(col);
         
         for (int i = 0; i < label.length; ++i)
         {
            String str = renderCell(label[i], col, colWidth, scroll);
            bufs[i].append(str);
         }

         for (int i = label.length; i < bufs.length; ++i)
         {
            StringBuffer buf = bufs[i];
            for (int j = colWidth - Math.abs(scroll); j > 0; --j)
            {
               buf.append(' ');
            }
         }
      }

      /**
       * @see ColumnRenderer#addMarks()
       */
      @Override
      public void addMarks()
      {
         addRightSpaces(2);
      }

      /**
       * @see ColumnRenderer#addSeparator()
       */
      @Override
      public void addSeparator()
      {
         addRightSpaces(1);
      }

      /**
       * @see ColumnRenderer#addRightSpaces(int)
       */
      @Override
      public void addRightSpaces(int width)
      {
         for (int i = 0; i < bufs.length; ++i)
         {
            for (int j = 0; j < width; ++j)
            {
               bufs[i].append(' ');
            }
         }
      }
   }

   /**
    * Renders a data line which is outside data model. Result is immediately
    * placed to the terminal. No row init necessary.
    */
   private class SpaceColumnRenderer
   implements ColumnRenderer
   {
      /** Indicates read-only mode. */
      protected boolean dcrReadOnly;
      /** Is widget focused. */
      protected boolean dcrFocused;
      /** The index of currently rendered row. */
      protected int dcrRowIndex;

      /** Dummy constructor. */
      public SpaceColumnRenderer(){}

      /**
       * @param   col
       *          Is not used.
       * @param   colIndex
       *          Is not used.
       *          
       * @see  ColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig,
       *       int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col,
                          int colIndex,
                          int colWidth,
                          int scroll)
      {
         printSpaces(colWidth - Math.abs(scroll), getColor(colorCommon));
      }

      /**
       * Calculate row color.
       *
       * @param   color
       *          Base color.
       * @return  Calculated color.
       */
      public Color getColor(Color color)
      {
         return (dcrReadOnly && dcrRowIndex == getCurrentRow() && dcrFocused)
                    ? Color.DEFAULT.withReverse()
                    : (config.separators ? color.withUnderline() : color);
      }

      /**
       * @see  ColumnRenderer#addMarks()
       */
      @Override
      public void addMarks()
      {
         char editMark;
         Color editMarkAttr = colorCommon;
         if (isEditPossible() && hasRowMarkers())
         {
            editMark = '*';
            
            if (dcrReadOnly && 
                dcrRowIndex == getCurrentRow() &&
                isEnabled())
            {
               editMarkAttr = colorCurrent;
            }
         }
         else
         {
            editMark = ' ';
         }

         screen().append(editMark, editMarkAttr);
         screen().append(isRowSelected(dcrRowIndex) ? '>' : ' ', colorCommon);
      }

      /**
       * @see  ColumnRenderer#addSeparator()
       */
      @Override
      public void addSeparator()
      {
         Color clr = config.separators && separatorQuirk ? colorHSep : colorCommon;
         char chr = config.separators ? ChuiOutputManager.UACS_VLINE : ' ';

         screen().append(chr, clr);
      }

      /**
       * @see  ColumnRenderer#addRightSpaces(int)
       */
      @Override
      public void addRightSpaces(int width)
      {
         printSpaces(width, colorCommon);
      }

      /**
       * Begins rendering of a new row.
       * 
       * @param   readOnly
       *          The widget in read-only mode.
       * @param   focused
       *          The widget is focused.
       * @param   rowIndex
       *          Index of the row in the browse model.
       */
      public void newRow(boolean readOnly, boolean focused, int rowIndex)
      {
         dcrFocused = focused;
         dcrReadOnly = readOnly;
         dcrRowIndex = rowIndex;
      }
   }

   /**
    * Renders a data line immediately to the terminal. <code>newRow()</code>
    * method should be called before rendering of each row.
    */
   private class DataColumnRenderer
   extends SpaceColumnRenderer
   {
      /** Data to render. */
      protected BaseDataType[] dcrRow;
      
      /**
       * Reference to deferred updates of current row. Always equals
       * <code>currRowUpdates</code> or <code>null</code>.
       */
      protected BaseDataType[] dcrRowUpdates;
      
      /** Dummy constructor. */
      public DataColumnRenderer(){}

      /**
       * @see  SpaceColumnRenderer#newRow(boolean, boolean, int)
       */
      @Override
      public void newRow(boolean readOnly, boolean focused, int rowIndex)
      {
         super.newRow(readOnly, focused, rowIndex);
         
         dcrRow = dcGetRow(rowIndex);
         dcrRowUpdates = (readOnly || dcrRowIndex != getCurrentRow())
                         ? null
                         : currRowUpdates;
      }

      /**
       * @see SpaceColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig, int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col,
                          int colIndex,
                          int colWidth,
                          int scroll)
      {
         col.column = Coordinate.scale(screen().cursor().x - location().x);

         if (!dcrReadOnly && 
              dcrFocused  && 
              dcrRowIndex == getCurrentRow() &&
              colIndex == currColumn)
         {
            // just cursor moving
            printSpaces(colWidth - Math.abs(scroll), colorCommon);
         }
         else
         {
            BaseDataType var =
               (dcrRowUpdates != null && dcrRowUpdates[colIndex] != null) ?
                  dcrRowUpdates[colIndex] :
                  dcrRow[colIndex];
            
            BrowseColumn<?> wcol = (BrowseColumn<?>) ThinClient.getInstance().getWidget(col.id);
            
            String display;
            try
            {
               String txt;
               if (var == null)
               {
                  // cells of new rows are filled with nulls - empty cells should be displayed
                  txt = "";
               }
               else
               {
                  txt = getCellText(wcol, var);
               }
               
               display = renderCell(txt, col, colWidth, scroll);
            }
            catch (DisplayFormatCheckException ex)
            {
               display = DisplayFormat.generateQuestions(colWidth - 
                                                         Math.abs(scroll));
            }

            Color clr = col.columnDColor;
            screen().append(display, getColor(clr));
         }
      }
   }
  
}