/*
** Module   : ChuiWidgetFactory.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111221 Created initial version
** 002 CA  20130129 Import change.
** 003 MAG 20140410 Add RECTANGLE widget creation.
** 004 EVL 20140610 Added TOGGLE-BOX widget support.
** 005 MAG 20140521 Refactored to implement support for GUI.
** 006 CA  20140711 Added MESSAGE-AREA widget support.
** 007 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 008 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 009 MAG 20141009 Deleted createMessageLine.
** 010 MAG 20141021 Renamed FrameImpl to FrameChuiImpl.
** 011 EVL 20141118 Adding scroll container creation method.
** 012 VIG 20150116 Added Menu classes mapping.
** 013 CA  20150323 Labels are always created using specific APIs; these APIs require the frame to
**                  which the label will belong, initially.
** 014 EVL 20150428 Adding drop-down container creation.
** 015 CA  20150509 Change to allow any widget to be contained by a scroll container.
** 016 HC  20150612 Implemented GUI ALERT-BOX.
** 017 EVL 20150514 Dpop-down widget for ChUI is now separated to it's own class.
** 018 HC  20150831 Scrolling support overhaul.
** 019 IAS 20160227 Added new createScrollPane method as a synonym for the existing one 
** 010 IAS 20160315 Support for the DEBUG-ALERT attribute
** 011 CA  20161101 Added createMnemonic().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import java.util.function.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Text;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.chui.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific widget factory implementation.
 * <p>
 * The return type for the {@code create} APIs is different from the {@link WidgetFactory} 
 * versions because:
 * <ul>
 *    <li>it avoids adding at the signature the {@link ChuiOutputManager} type, i.e. instead
 *        of using {@code Label&lt;ChuiOutputManager&gt;} as return type, {@link LabelImpl}
 *        can be used directly, with the same effect. </li>
 *    <li> adds an additional protection against incorrect usage of the {@link WidgetFactory} 
 *    implementation classes: if the {@link WidgetFactory} implementation class is used 
 *    explicitly, it prevents accidental mixing of ChUI and GUI widgets (i.e. if used from 
 *    incorrect location). </li>
 * </ul> 
 */
public class ChuiWidgetFactory
extends WidgetFactoryAdapter<ChuiOutputManager>
{
   static 
   {
      classMap.put(AlertBox.class, AlertBoxImpl.class);
      classMap.put(Browse.class, BrowseImpl.class);
      classMap.put(BrowseColumn.class, BrowseColumn.class);
      classMap.put(Button.class, ButtonImpl.class);
      classMap.put(ComboBox.class, ComboBoxImpl.class);
      classMap.put(Editor.class, EditorImpl.class);
      classMap.put(FillIn.class, FillInImpl.class);
      classMap.put(Frame.class, FrameChuiImpl.class);
      classMap.put(Menu.class, MenuChuiImpl.class);
      classMap.put(MenuItem.class, MenuItemChuiImpl.class);
      classMap.put(SubMenu.class, SubMenuChuiImpl.class);
      classMap.put(RadioButton.class, RadioButtonImpl.class);
      classMap.put(RadioSet.class, RadioSet.class);
      classMap.put(Rect.class, RectangleImpl.class);
      classMap.put(ScrollableList.class, ScrollableListImpl.class);
      classMap.put(ScrollPane.class, ScrollPaneImpl.class);
      classMap.put(SelectionListBody.class, SelectionListBodyImpl.class);
      classMap.put(SelectionList.class, SelectionListImpl.class);
      classMap.put(Separator.class, SeparatorImpl.class);
      classMap.put(Skip.class, SkipImpl.class);
      classMap.put(Space.class, Space.class);
      classMap.put(StatusLine.class, StatusLineImpl.class);
      classMap.put(Text.class, TextImpl.class);
      classMap.put(ToggleBox.class, ToggleBoxImpl.class);
      classMap.put(Window.class, WindowChuiImpl.class);
   }
   
   /**
    * Create an object accumulating prepared label and mnemonic info.
    * 
    * @return   new {@link MnemonicInfo} instance.
    */
   @Override
   public MnemonicInfo createMnemonic()
   {
      return new MnemonicInfo();
   }
   
   /**
    * Create new {@link Label} instance for given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Label text. 
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    *
    * @return  new {@link Label} instance.
    */
   @Override
   public LabelImpl createLabel(WidgetId id, String text, int frameId)
   {
      return new LabelImpl(id, text, frameId);
   }

   /**
    * Create new button for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Button text.
    *          
    * @return  new {@link Button} instance.
    */
   @Override
   public ButtonImpl createButton(WidgetId id, Container<ChuiOutputManager> container, String text)
   {
      return new ButtonImpl(id, container, text);
   }

   /**
    * Create new toggle-box for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Toggle Box text.
    *          
    * @return  new {@link ToggleBox} instance.
    */
   @Override
   public ToggleBoxImpl createToggleBox(WidgetId id, Container<ChuiOutputManager> container, String text)
   {
      return new ToggleBoxImpl(id, container, text);      
   }

   /**
    * Create new rectangle for given container.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    *          
    * @return  new {@link Rect} instance.
    */
   @Override
   public RectangleImpl createRectangle(WidgetId id, Container<ChuiOutputManager> container)
   {
      return new RectangleImpl(id, container);
   }

   /**
    * Create UI-specific {@link DropDown} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    *          
    * @return  instance of UI-specific implementation of {@link DropDown}.
    */
   @Override
   public DropDownImpl createDropDown(ComboBox<ChuiOutputManager> parent)
   {
      return new DropDownImpl(parent);
   }

   /**
    * Create UI-specific {@link ScrollableList} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    * @param   model
    *          Scrollable list model.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollableList}.
    */
   @Override
   public ScrollableListImpl createScrollableList(ComboBox<ChuiOutputManager> parent, 
                                                  ListModel<String>           model)
   {
      return new ScrollableListImpl(parent, parent.model());
   }

   /**
    * Create UI-specific {@link SensitiveScrollContainer} implementation.
    * 
    * @return  instance of UI-specific implementation of {@link SensitiveScrollContainer}.
    */
   @Override
   public SensitiveScrollContainer<ChuiOutputManager> createScrollContainer()
   {
      return new SensitiveScrollContainer<ChuiOutputManager>();
   }

   /**
    * Create UI-specific {@link ScrollPane} implementation.
    * 
    * @param   widget 
    *          Widget which will be scrolled.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   @Override
   public ScrollPaneImpl createScrollPane(ScrollableWidget<ChuiOutputManager> widget)
   {
      return new ScrollPaneImpl(widget);
   }

   /**
    * Create UI-specific {@link ScrollPane} instance.
    * 
    * @param   widget
    *          Widget which will be scrolled.
    * @param   wnd
    *          TopLevelWindow supplier
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   @Override
   public ScrollPaneImpl createScrollPane(ScrollableWidget<ChuiOutputManager> widget, 
         Supplier<TopLevelWindow<ChuiOutputManager>> wnd)
   {
      return new ScrollPaneImpl(widget);
   }

   /**
    * Create UI-specific {@link Label} instance as a copy of given {@link Label}.
    * 
    * @param   id
    *          The widget's ID.
    * @param   labelInstance
    *          Source {@link Label} instance.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    *          
    * @return  instance of UI-specific implementation of {@link Label}.
    */
   @Override
   public LabelImpl createLabel(WidgetId id, Label<ChuiOutputManager> labelInstance, int frameId)
   {
      return new LabelImpl(id, labelInstance, frameId);
   }

   /**
    * Create UI-specific {@link RadioButton} instance.
    * 
    * @param   config
    *          Parent config. 
    * @param   label
    *          Radio button label. 
    * @param   value
    *          Radio button data type. 
    * 
    * @return  instance of UI-specific implementation of {@link RadioButton}.
    */
   @Override
   public RadioButtonImpl createRadioButton(RadioSetConfig config, 
                                            String         label, 
                                            BaseDataType   value)
   {
      return new RadioButtonImpl(config, label, value);
   }

   /**
    * Create UI-specific {@link Window} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Window}.
    */
   @Override
   public WindowChuiImpl createWindow(WidgetId id)
   {
      return new WindowChuiImpl(id);
   }

   /**
    * Create UI-specific {@link StatusLine} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link StatusLine}.
    */
   @Override
   public StatusLineImpl createStatusLine(WidgetId id)
   {
      return new StatusLineImpl(id);
   }
   
   /**
    * Create UI-specific {@link MessageArea} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   lines
    *          Number of visible lines in message area.         
    * 
    * @return  instance of UI-specific implementation of {@link MessageArea}.
    */
   @Override
   public MessageAreaImpl createMessageArea(WidgetId id, int lines)
   {
      return new MessageAreaImpl(id, lines );
   }
   
   /**
    * Create UI-specific {@link AlertBox} instance.
    * 
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    content
    *           The list of message text components.
    * @param    buttonSet
    *           One of the BTN_* constants specifying used button set.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    title
    *           The dialog title.
    * @param    color
    *           Color scheme to use.
    * @param    stack
    *           Progress stack trace
    *
    * @return   instance of UI-specific implementation of {@link MessageArea}.
    */
   @Override
   public AlertBoxImpl createAlertBox(int      alertType,
                                      Object[] content, 
                                      int      buttonSet, 
                                      logical  var, 
                                      String   title,
                                      Color    color,
                                      String[] stack)
   {
      return new AlertBoxImpl(content, buttonSet, var, title, color, stack);
   }
   
   /**
    * Creates new instance of {@link BorderedPanel} with {@link BorderLayout} as
    * the layout manager and no border.
    * 
    * @return  instance of ChUI-specific implementation of {@link BorderedPanel}.
    */
   @Override
   public BorderedPanel<ChuiOutputManager> createBorderedPanel()
   {
      return new BorderedPanelImpl(new BorderLayout<ChuiOutputManager>(), 
                                   new EmptyBorder<ChuiOutputManager>());
   }
      
   /**
    * Creates new instance of {@link BorderedPanel} with {@link BorderLayout}
    * layout manager and simple border of width 1 measured in native units. 
    * 
    * @return  instance of ChUI-specific implementation of {@link BorderedPanel}.
    */
   @Override
   public BorderedPanel<ChuiOutputManager> createLineBorderedPanel()
   {
      return new BorderedPanelImpl(new BorderLayout<ChuiOutputManager>(), new LineBorder());
   }
   
   /**
    * Creates new instance of {@link BorderedPanel} with a specific layout
    * manager and border.
    * 
    * @param   mgr
    *          Layout manager.
    * @param   border
    *          Border for this widget.
    * 
    * @return  instance of ChUI-specific implementation of {@link BorderedPanel}.
    */
   @Override
   public BorderedPanel<ChuiOutputManager> createBorderedPanel(
      LayoutManager<ChuiOutputManager> mgr, 
      Border<ChuiOutputManager>        border)
   {
      return new BorderedPanelImpl(mgr, border);
   }
   
   /**
    * Creates new instance representing a default line border.
    * 
    * @return  instance of ChUI-specific implementation of {@link Border}.
    */
   @Override
   public LineBorder createLineBorder()
   {
      return new LineBorder();
   }
}