/*
** Module   : BrowseColumn.java
** Abstract : Java standard header
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 SVG 20050803   @21995 Initial version.
** 002 SIY 20050831   @22526 Formatting and spelling cleanups.
** 003 SIY 20050926   @22820 Added a support for the attributes.
** 004 SIY 20060209   @24488 Updated to accommodate changes in xxxConfig classes.
** 005 SIY 20060315   @25120 Changes for complete BROWSE implementation.
** 006 NVS 20060322   @25196 Added dummy isModified() method since now it is part of the 
**                           DataContainer interface.
** 007 SIY 20060324   @25273 Fixed setValue() method.
** 008 SIY 20060824   @28845 Removed getP2JType() method.
** 009 NVS 20080401   @37771 Added isFocusable() method implementation.
** 010 SIY 20100619          Switching to own implementation of character UI.
** 011 SIY 20100804          Changes in imports, some refactorings.
** 012 SIY 20100822          Refactored AttributeContainer support.
** 013 SIY 20111101          Moved from CHUI package. Minor cleanups/refactorings in preparation 
**                           to GUI client.
** 014 SIY 20111202          Updated imports.
** 015 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access 
**                           only to the base APIs provided by the OutputManager.
** 016 HC  20140911          GUI coordinate support - data types refactoring.
** 017 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 018 HC  20150323          Removed business logic from config classes.
** 019 SVL 20151013          Added constructor for markers column.
** 020 SVL 20151019          Moved ChUI-specific column width calculation logic from
**                           BrowseColumnWidget.
** 021 SVL 20151126          Changes in processing since NO-LABELS attribute is now not inherited
**                           form the parent browse.
** 008 IAS 20151214          LOAD-MOUSE-POINTER support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Information about column in BROWSE widget. This is not a real widget, just
 * a container.
 */
public class BrowseColumn<O extends OutputManager<?>>
extends AbstractWidget<O>
implements DataContainer,
           WidgetWithConfig<BrowseColumnConfig>
{
   /** Link to the browse widget. */
   protected Browse<O> browse = null;
   
   /** Component attribute container. */
   protected BrowseColumnConfig config = null;

   /**
    * Forced right alignment. Note that this flag is not transferred via net
    * nor copied when config is copied.
    */
   private boolean rightAligned = false;

   /**
    * Constructor.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public BrowseColumn(WidgetId id, BrowseColumnConfig cfg)
   {
   }

   /**
    * Constructor which creates default config.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param   id
    *          The widget's ID.
    */
   public BrowseColumn(WidgetId id)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, BrowseColumnConfig.class);
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BrowseColumnConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Force right alignment.
    */
   public void forceRightPositioning()
   {
      this.rightAligned = true;
   }
   /**
    * Get current value of the force right alignment flag.
    * 
    * @return  current state of the flag.
    */
   public boolean isRightPositioning()
   {
      return rightAligned;
   }

   /**
    * Link column to owning frame.
    * 
    * @param   browse
    *          Browse instance to link with.
    */
   public void setBrowse(Browse<O> browse)
   {
      this.browse = browse;
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BrowseColumnConfig config()
   {
      return config;
   }

   /**
    * The method is called after the configuration associated with
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    *
    * @param    beforeUpdate
    *           Config reference capturing the config state before
    *           the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BrowseColumnConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);

      if (config.autoResize)
      {
         DisplayFormat colFormat = Browse.getDisplayFormat(config);
         String[] colLabel = Browse.getLabelVar(config);

         double colWidth = colFormat.getScreenWidth();

         boolean browseLabels = true;
         if (browse != null)
         {
            browseLabels = browse.config().labels;
         }

         if ((browseLabels && config.labels) || config.forceLabel)
         {
            for (int i = 0; i < colLabel.length; ++i)
            {
               if (colLabel[i].length() > colWidth)
               {
                  colWidth = colLabel[i].length();
               }
            }
         }

         config.widthChars = Coordinate.scale(Math.max(config.formatWidth, colWidth));
      }
   }

   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }

   //----------------------------------------------------------------
   // DataContainer interface
   //----------------------------------------------------------------

   /**
    * Get text representation of the selected value.
    * 
    * @return  The screen representation of the value.
    */
   @Override
   public String getText()
   {
      if (browse != null)
         return browse.getColumnText(config.ordinal);

      return null;
   }

   /**
    * Returns the BaseDataType instance representing the current value.
    *
    * @return   The value representing the current widget value.
    */
   @Override
   public BaseDataType getValue()
   {
      if (browse != null)
         return browse.getColumnValue(config.ordinal);

      return null;
   }

   /**
    * Getting the modification state of the widget.
    * 
    * @return  always returns <code>false</code> and is never called.
    */
   @Override
   public boolean isModified()
   {
      if (browse != null)
         return browse.isModified(config.ordinal);
      
      return false;
   }

   /**
    * Sets the BaseDataType instance representing the current value.
    *
    * @param    value
    *           The value representing the current widget value.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      if (browse != null)
         browse.setColumnValue(config.ordinal, value);
   }
   
   //----------------------------------------------------------------
   // Component interface
   //----------------------------------------------------------------

   /** 
    * Dummy implementation of the missing method.
    */
   @Override
   public void draw()
   {
   }

   /** 
    * Dummy implementation of the missing method.
    * 
    * @return  0 
    */
   @Override
   public double height()
   {
      return 0;
   }

   /** 
    * Dummy implementation of the missing method.
    * 
    * @return  zero-sized area. 
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(0,0);
   }

   /** 
    * Dummy implementation of the missing method.
    * 
    * @return  0 
    */
   @Override
   public double width()
   {
      return 0;
   }

   /** 
    * Dummy implementation of the missing method.
    * 
    * @return  zero-sized area. 
    */
   @Override
   public Dimension minimumSize()
   {
      return new Dimension(0,0);
   }

   /**
    * Check if widget is visible.
    * 
    * @return  <code>true</code> if widget is visible.
    */
   @Override
   public boolean isVisible()
   {
      return true;
   }
   
   /**
    * Get current hidden state.
    * 
    * @return   hidden state.
    */
   @Override
   public boolean hidden()
   {
      return false;
   }
}