/*
** Module   : BrowseModel.java
** Abstract : data model for Browse widget
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------------------------------
** 001 SVG 20050803   @21995 Initial version.
** 002 SIY 20050831   @22527 Formatting and spelling cleanups. 
** 003 SIY 20050926   @22825 Added isQueryOffEnd() and isNewRow() methods.
** 004 SIY 20051201   @23591 Removed isQueryOffEnd() method.
** 005 SIY 20060315   @25124 Changes for complete BROWSE implementation.
** 006 GES 20060808   @28494 Use int instead of long in the APIs.
** 007 SIY 20061105   @30986 Added getCurrentRow() method.
** 008 SIY 20070504   @33419 Updated getRowCount().
** 009 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 010 SVL 20150212          Modified getRows signature.
** 011 SVL 20150411          Added isRowAvailable.
** 012 SVL 20150503          Added resetColumnsCache.
** 013 SVL 20150608          Added getEditingFillIns.
** 014 SVL 20150625          Added getRowCount().
** 015 CA  20150722          Browse columns have client-side widgets which need to have their own
**                           IDs and configurations.
** 016 SVL 20150722          Updated updateRow signature.
** 017 SVL 20160419          Added moveColumn.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Data model for BROWSE widget.
 */
public interface BrowseModel
{
   /**
    * Get configurations of fill-ins used for in-browse editing.
    *
    * @return   Array of configurations of fill-in widgets which are used for in-browse editing.
    *           An array element is configuration of the fill-in widget for the corresponding
    *           editable column or <code>null</code> if the column is not editable. Return value
    *           is <code>null</code> if the browse is read-only.
    */
   public FillInConfig[] getEditingFillIns();

   /**
    * Get the number of rows iterated by the query (for preselect queries - matches result size,
    * for adaptive queries - matches the size of the part of the result set the query has already
    * iterated over).
    *
    * @return  number of iterated rows in the result set.
    */
   public int getRowCount();

   /**
    * Get number of available rows starting from the given one.
    *
    * @param   startRow
    *          Initial row to start count rows from.
    *
    * @return  Number of available rows.
    */
   public int getRowCount(int startRow);

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    startRowIndex
    *           Index of the first row to retrieve.
    * @param    count
    *           Maximum count of rows to return.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction,
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query specified, an array of size 0 is returned.
    */
   public BaseDataType[][] getRows(int startRowIndex, int count, boolean forward);
   
   /**
    * Updates selected cells in the given row.
    * 
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * @param   onlyNotifyRowReleased
    *          Only notify server that the row has been released. Data is not updated.
    * 
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   public void updateRow(int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues,
                         boolean        onlyNotifyRowReleased)
   throws DataValidationException;

   /**
    * Inform server about current row in the browse.
    * 
    * @param   rowIndex
    *          0-based index of the current row.
    *          
    * @return  New current row.
    */
   public int setCurrentRow(int rowIndex);

   /**
    * Get current row in the model.
    * 
    * @return  The current row index in the model.
    */
   public int getCurrentRow();

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   public boolean isRowAvailable(int rowIndex);

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in the browse widget.
    *
    * @param  src
    *         0-based position of the column to be moved.
    * @param  dest
    *         0-based position to which the column is moved.
    */
   public void moveColumn(int src, int dest);
}