/*
** Module   : Cell.java
** Abstract : container which holds the details for rendering a single row/column in ChUI 
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 SIY 20110112 Created initial version.
** 002 GES 20150414 Moved empty cell definition from another class.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;

/**
 * Single screen cell consisting of a character and a color.
 */
public class Cell
{
   /** Immutable empty cell (space character with normal color). */
   public final static Cell EMPTY_CELL = new Cell(' ', Color.NORMAL);
   
   /** Cell character. */ 
   private final char character;

   /** Cell color. */
   private final Color color;

   /**
    * Constructor.
    *
    * @param   character
    *          Cell character.
    * @param   color
    *          Cell color.
    */
   public Cell(char character, Color color)
   {
      this.character = character;
      this.color = color;
   }

   /**
    * Get cell character.
    *
    * @return cell character
    */
   public char character()
   {
      return character;
   }

   /**
    * Get cell color.
    *
    * @return cell color
    */
   public Color color()
   {
      return color;
   }

   /**
    * Check if cell represents real character and color.
    *
    * @param   cell
    *          Cell instance to check.
    *
    * @return  <code>true</code> if cell can be used to display character.
    */
   public static boolean isValid(Cell cell)
   {
      return cell != null && cell.color != null;
   }

   /**
    * @see java.lang.Object#hashCode()
    */
   @Override
   public int hashCode()
   {
      int result = 31 + character;
      return 31 * result + ((color == null) ? 0 : color.hashCode());
   }

   /**
    * @see java.lang.Object#equals(java.lang.Object)
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;

      if (obj == null || getClass() != obj.getClass())
         return false;

      Cell other = (Cell) obj;

      if (character != other.character)
         return false;

      if (color == null)
      {
         if (other.color != null)
            return false;
      }
      else if (!color.equals(other.color))
         return false;

      return true;
   }

   /**
    * Check if specified cell represents a space character.
    *
    * @param   cell
    *          Cell to check.
    *
    * @return  <code>true</code> if cell holds a space character.
    */
   public static boolean isSpace(Cell cell)
   {
      return (cell != null) ? cell.character() == ' ' : false;
   }
}