/*
** Module   : FrameFocusManager.java
** Abstract :
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20100831          Created initial version
** 002 SIY 20100915          tryFocusChange() moved to UiUtils.
** 003 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 004 SIY 20111202          Updated imports.
** 005 HC  20140911          GUI coordinate support - data types refactoring.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.client.widget.*;

/**
 * Common logic of frame focus transfer.
 * 
 * @author 	SIY
 */
public class FrameFocusTransferManager
{
   /** Owner frame. */
   private final Frame owner;

   /**
    * Constructor.
    * 
    * @param   owner
    *          Owner frame.
    */
   public FrameFocusTransferManager(Frame owner) {
      this.owner = owner;
   }
   
   /**
    * Move the cursor to the component in the nearest row above this widget.
    * <p>
    * The algorithm is dependent upon the array of components returned by
    * the container.  In particular, the order of the widgets in the
    * array MUST be from left to right and top to bottom.
    * <p>
    * The search order (from highest to lowest precedence) is as follows:
    * <p>
    * <ol>
    *    <li> The component that is on the nearest row above AND which has at
    *         least 1 character position exactly over the current cursor
    *         position. In this case, the cursor will be kept in the exact
    *         corresponding column in that widget.
    *    <li> The component that is on the nearest row above AND which has the
    *         same index position as the current widget.  For example, if the
    *         current widget is the 3rd on its row, then the cursor will be
    *         moved to the 3rd widget of the row above.  The cursor will 
    *         always be placed at the start of such a widget.
    *    <li> The component that is on the nearest row above AND which has the
    *         last index position.  For example, if the current widget is the
    *         3rd on its row, and the row above only has 2 widgets then the
    *         cursor will be moved to the 2nd widget of the row above.  The 
    *         cursor will always be placed at the start of such a widget.
    * </ol>
    * <p>
    * The cursor (and focus) will be moved into the component above this one
    * as long as such a component exists.  Otherwise this method has no 
    * affect.
    *
    * @param   wdgt
    *          widget to check
    *
    * @param   coff
    *          cursor offset within the widget
    */
   public void moveAboveWidget(Widget wdgt, int coff)
   {
      // set the new target if needed
      moveToTarget(wdgt, getTarget(wdgt, coff, false));
   }

   /**
    * Move the cursor to the component in the nearest row below this widget.
    * <p>
    * The algorithm is dependent upon the array of components returned by
    * the container.  In particular, the order of the widgets in the
    * array MUST be from left to right and top to bottom.
    * <p>
    * The search order (from highest to lowest precedence) is as follows:
    * <p>
    * <ol>
    *    <li> The component that is on the nearest row below AND which has at
    *         least 1 character position exactly over the current cursor
    *         position. In this case, the cursor will be kept in the exact
    *         corresponding column in that widget.
    *    <li> The component that is on the nearest row below AND which has the 
    *         same index position as the current widget.  For example, if the
    *         current widget is the 3rd on its row, then the cursor will be
    *         moved to the 3rd widget of the row below.  The cursor will 
    *         always be placed at the start of such a widget.
    *    <li> The component that is on the nearest row below AND which has the
    *         last index position.  For example, if the current widget is the
    *         3rd on its row, and the row below only has 2 widgets then the
    *         cursor will be moved to the 2nd widget of the row below.  The 
    *         cursor will always be placed at the start of such a widget.
    * </ol>
    * <p>
    * The cursor (and focus) will be moved into the component below this one
    * as long as such a component exists.  Otherwise this method has no 
    * affect.
    *
    * @param   wdgt
    *          widget to check
    *
    * @param   coff
    *          cursor offset within the widget
    */
   public void moveBelowWidget(Widget wdgt, int coff)
   {
      // set the new target if needed
      moveToTarget(wdgt, getTarget(wdgt, coff, true));
   }
   
   /**
    * Check if the specified widget is the left-most component in the current
    * screen row (not the same as the current record which may wrap).
    * <p>
    * The algorithm is dependent upon the array of components returned by
    * the container.  In particular, the order of the widgets in the
    * array MUST be from left to right and top to bottom.
    *
    * @param   wdgt
    *          widget to check
    * 
    * @return  <code>true</code> if this component is the left-most component
    *          in the row.
    */
   public boolean isLeftMostWidget(Widget wdgt)
   {
      Widget[] list  = widgetList(sortHelper);
      int         num   = 0;
      int         last  = 0;
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i].focusTraversable())
         {
            // increase the number of previous editable widgets
            num++;
            
            if (list[i] == wdgt)
            {
               if (num == 1)
               {
                  // this is the first editable widget
                  return true;
               }
               
               // check the previous editable widget to see if it is in
               // the same row
               Point lloc = list[last].location();
               Point mloc = wdgt.location();
               
               return (lloc.y == mloc.y) ? false : true;
            }
         
            // remember last editable widget that was not the current widget
            last = i;
         }
      }
      
      return false;
   }

   /**
    * Check if the specified widget is the right-most component in the current
    * screen row (not the same as the current record which may wrap).
    * <p>
    * The algorithm is dependent upon the array of components returned by
    * the container.  In particular, the order of the widgets in the
    * array MUST be from left to right and top to bottom.
    * 
    * @param   wdgt
    *          widget to check
    * 
    * @return  <code>true</code> if this component is the right-most component
    *          in the row.
    */
   public boolean isRightMostWidget(Widget wdgt)
   {
      Widget[] list  = widgetList(sortHelper);
      int         num   = 0;
      int         last  = 0;
      
      for (int i = (list.length - 1); i >= 0; i--)
      {
         if (list[i].focusTraversable())
         {
            // increase the number of following editable widgets
            num++;
            
            if (list[i] == wdgt)
            {
               if (num == 1)
               {
                  // this is the last editable widget
                  return true;
               }

               // check the next editable widget to see if it is in the
               // same row
               Point nloc = list[last].location();
               Point mloc = wdgt.location();
               
               return (nloc.y == mloc.y) ? false : true;
            }
         
            // remember last editable widget that was not the current widget
            last = i;
         }
      }
      
      return false;
   }
   
   /**
    * Move to a target widget if needed.  If the given target widget is not
    * <code>null</code>, this will change focus to that target widget and
    * set the cursor position accordingly.
    *
    * @param    source
    *           Source widget.
    * @param    tgt
    *           The target to which to move.  May be <code>null</code> in
    *           which case nothing is changed.
    */
   void moveToTarget(Widget source, Target tgt)
   {
      if (tgt != null && tgt.target != null)
      {
         FocusManager.tryFocusChange(tgt.target);
         
         if (tgt.target instanceof FillIn)
         {
            ((FillIn) tgt.target).silentActivation(source, tgt.pos);
         }
      }
   }

   /**
    * Locate component suitable for focus transfer using cursor movement keys.
    * <p>
    * First of all algorithm locates source component which contains focus
    * now. Then it locates first component which is placed on the closest row
    * below (or above, depending on direction) source component. Then all
    * components which are placed on that row checked and one target component
    * is then returned as a part of result. In most cases algorithm tries to
    * find component which may accept cursor without changing cursors global
    * (in screen coordinates) column. There are exceptions though.
    * <p>
    * Rules which determine the component are following:<br>
    * <ol>
    * <li>Component which has first column greater or equal to first column of
    * source component and has last column less or equal to last column of
    * source component.<br>
    * Such a components have precedence over other components even if there
    * are other component which may accept current cursor global column.</li>
    * <li>Component which may accept current cursor global column.</li>
    * <li>When looking for widget below, choose component closest to right
    * side with start column is less than start column of source component.</li>
    * <li>When looking for widget above, choose component closest to left side
    * among those which have start column greater than start column of source
    * component.</li>
    * </ol>
    * 
    * @param    source
    *           widget to check
    * @param    coff
    *           cursor offset within the widget
    * @param    direction
    *           Direction in which widgets are searched.
    * 
    * @return   Found component and new cursor position.
    */
   private Target getTarget(Widget source, int coff, boolean direction)
   {
      Widget[] list = widgetList((direction) ? sortHelper : reverseSortHelper);
      
      int i;

      // locate source widget
      for (i = 0; i < list.length; i++)
      {
         if (list[i] == source)
            break;
      }

      Widget target = null;

      double row = source.location().y;

      double startX = source.location().x;
      double absPos = startX + coff;

      // locate first widget which is not in the same row
      // as source widget
      for (i++ ; i < list.length; i++)
      {
         Widget tmp = list[i];
         Point loc = tmp.location(); 

         if (tmp.focusTraversable() && loc.y != row)
         {
            target = tmp;
            break;
         }
      }

      boolean contained   = false;
      boolean cursorMatch = false;
      
      // scan all widgets which are in the same row and
      // (almost always) try to find one which is closest 
      // to cursor 
      for (; target != null && i < list.length; i++)
      {
         if (!list[i].focusTraversable())
         {
            // skip non-focusable widgets and widgets in the same row
            continue;
         }
         
         if (list[i].location().y != target.location().y)
            break;

         Point loc = list[i].location(); 
         double targetX = target.location().x;
         
         // yes, it's somewhat unexpected but 4GL prefers such a widgets
         // even if there are better matches 
         if (loc.x >= startX && 
             loc.x + list[i].width() <= startX + source.width())
         {
            // new widget is completely "inside" source widget
            if (contained)
            {
               if (targetX > list[i].location().x)
                  target = list[i];
            }
            else
               target = list[i];
            
            contained = true;
            continue;
         }
         
         // only completely contained widgets may take preference over
         // other contained widgets
         
         if (contained)
            continue;
         
         double tmp = absPos - loc.x;
         
         if (tmp >= 0 && tmp < list[i].width())
         {
            // found match
            target = list[i];
            cursorMatch = true;
         }
         
         if (!cursorMatch)
         {
            // widget below
            if (direction && loc.x > targetX && loc.x <= startX)
            {
               // found candidate
               target = list[i];
            }
            
            // widget above
            if (!direction && loc.x < targetX && targetX > startX)
            {
               // found candidate
               target = list[i];
            }            
         }
      }

      double pos = (target != null) ? (absPos - target.location().x) : 0;
      
      if (pos < 0)
         pos = 0;
      
      if (target != null && pos >= target.width())
         pos = 0;
      
      return new Target(target, (int) pos);
   }

   /**
    * Obtain a list of widgets in this frame which is guaranteed
    * to enforce an ordering in which the each element is left and/or above
    * (positionally) the next array element.  Thus the order of the array
    * is left to right and top to bottom.
    * 
    * @param    helper 
    *           Comparator implementation which determines final order of 
    *           widgets.
    *           
    * @return   The positionally sorted widget list of the frame. 
    */
   private Widget[] widgetList(Comparator<Widget> helper)
   {
      // assumes that we are getting a COPY of the array and not the original
      Widget[] list = owner.getContentPane().widgets();
      Arrays.sort(list, helper);
      
      return list;
   }
   
   /**
    * The container for the target component and cursor position.
    */
   static class Target
   {
      /** Target component. */
      Widget target;
      
      /** Target cursor position. */
      int pos;
      
      /**
       * Construct an instance for the given target component and cursor
       * position.
       * 
       * @param   target
       *          The target component.
       * @param   pos
       *          The cursor position.
       */
      Target(Widget target, int pos)
      {
         this.target = target;
         this.pos = pos;
      }
   }

   /** Sorts widgets based on a left to right and top to bottom ordering. */
   private static Comparator<Widget> sortHelper = new Comparator<Widget>()
   {
      @Override
      public int compare(Widget c1, Widget c2)
      {
         Point p1 = c1.location();
         Point p2 = c2.location();
         
         int result = 0;
         
         if (p1.y < p2.y)
         {
            result = -1;
         }
         else if (p1.y == p2.y)
         {
            if (p1.x < p2.x)
            {
               result = -1;
            }
            else if (p1.x == p2.x)
            {
               result = 0;
            }
            else
            {
               result = 1;
            }
         }
         else
         {
            result = 1;
         }
         
         return result;
      }
   };

   /** Reverse sorting routine. */
   private static Comparator<Widget> reverseSortHelper = new Comparator<Widget>()
   {
      @Override
      public int compare(Widget c1, Widget c2)
      {
         return -sortHelper.compare(c1, c2);
      }
   };
}
