/*
** Module   : OutputManager.java
** Abstract : Provides a generic interface to low level UI primitives.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 GES 20100817 Cleaned up formatting, fixed/added javadoc, removed 
**                  getters in invariant cases.
** 004 SIY 20101108 Make sure that driver will be initialized only once even if OutputManager 
**                  instances are created again and again.
** 005 SIY 20101202 Removed unused method ans slightly changed interaction with the driver in
**                  order to improve decoupling between UI and driver. Reorganized constants.
** 006 SIY 20110112 Refactoring and simplification, changes in color handling.
** 007 GES 20110921 Support for multiple drivers.
** 008 GES 20111012 Moved to context-local instead of singleton. This allows multiple client
**                  sessions to be hosted in the server.
** 009 SIY 20111101 Updated imports, minor cleanups/refactorings.
** 010 GES 20111103 Modifications to allow the screen driver instance to be obtained and reused
**                  upon re-initialization (this happens during CTRL-C processing).
** 011 SIY 20111202 Updated imports.
** 012 SIY 20111221 Added accessor for widget factory.
** 013 EVL 20130724 Replacing CHARVA with our own console driver.
** 014 EVL 20130822 Adding new 6 characters into consideration to handle comb-box and
**                  selection-list widgets drawing.
** 015 MAG 20131114 Add chui web screen driver support. The swing_chui_applet is no longer used.
** 016 EVL 20131104 Batch mode support.
** 017 EVL 20140115 Adding support for background simulation mode.
** 018 HC  20140119 Implemented runtime support for the COLOR protermcap-attribute.
** 019 EVL 20140328 Changing the initErrorWriter() method to be able to get the batch mode option
**                  from the server.
** 020 MAG 20140512 Add support for a GUI screen driver. 
** 021 CA  20140805 Refactored to allow GUI output manager and separate ChUI/GUI specific logic 
**                  into distinct output manager implementation.
** 022 VIG 20140827 Changed processDefaultWindow method to work with specified window Id.
**                  Changed setInvalidate method (added window registering).
** 023 MAG 20140912 Fixed NPE when bitmapReset.
** 024 HC  20140911 GUI coordinate support - data types refactoring.
** 025 CA  20141021 Added translatedOrigin(), which can inform the caller about the real location
**                  of the current origin point.
** 026 HC  20141021 Implemented window move/position support.
** 027 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 028 CA  20141123 Added API to check if a certain key can raise a STOP condition.
** 029 CA  20141125 Added API to conditionally enable/disable OS-level events.
** 030 EVL 20150107 Fix for dynamic windows renderer missing exception.
** 031 HC  20150323 WidgetId.getId() renamed to asInt().
** 032 GES 20150312 Added support for the GUI web client driver. Modified the member names and
**                  type constants to be more consistent.
** 033 CA  20150603 If events are raised by the drawing code, consume and log them.  Events must
**                  never be raised from the widget's draw() API, as they are either lost or 
**                  leaked to the nearest event queue - and at the time this queue is processed,
**                  the source widget may no longer be alive.
** 034 HC  20150612 Implemented support for modal windows.
** 035 GES 20150730 Removed translatedOrigin() since it isn't actually used.
** 036 HC  20150913 Added method withDrawFragments().
** 037 HC  20150917 Fixed wrong usage of ScreenBitmap clipping rectangles.
** 038 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 039 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 040 EVL 20160505 Fix for NPE on ESC key pressing and no current widget active.  When client is
**                  running there is always at least one window on the screen. 
** 041 HC  20160628 Prevented unhandled exceptions when window/frame closed/destroyed.
** 042 HC  20160701 Frame invalidation fix. 
** 043 HC  20160926 Removed disabling of window input events during server processing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.swing.*;
import com.goldencode.p2j.ui.client.gui.driver.web.*;
import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.ui.client.chui.driver.web.*;
import com.goldencode.p2j.ui.client.chui.driver.batch.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Provides a generic interface to low level user interface primitives.
 */
public abstract class OutputManager<P extends OutputPrimitives>
{
   /** CHUI native driver name. */
   public static final String CHUI_DRIVER_CONSOLE = "chui_native";
   
   /** CHUI swing driver name. */
   public static final String CHUI_DRIVER_SWING = "chui_swing";
     
   /** CHUI web driver name. */
   public static final String CHUI_DRIVER_WEB = "chui_web";

   /** CHUI batch mode driver name. */
   public static final String CHUI_DRIVER_BATCH = "chui_batch";
   
   /** GUI swing driver name. */
   public static final String GUI_DRIVER_SWING = "gui_swing";
   
   /** GUI web driver name. */
   public static final String GUI_DRIVER_WEB = "gui_web";
   
   /** Context local instance of the default terminal manager. */
   protected static final ContextLocal<OutputManager<?>> defaultTerminal = new ContextLocal<>();
   
   /** Flag to enable debug messages. */
   protected boolean debug = false;

   /** Default (direct) drawing routines. */
   protected P defaultOutput;

   /** Current drawing routines. */
   protected P output;

   /** The state of invalidation. */
   protected int invalidate = 0;

   /** Reflects drawing in progress to avoid recursion. */
   protected boolean drawing = false;

   /** Active screen driver. */
   protected ScreenDriver<?> driver;
   
   /** Batch mode client override flag. */
   private boolean batchModeOverride = false;

   /** The output manager's coordinate logic implementation **/
   private CoordinatesConversion coordinates = 
         new CoordinatesConversion(new StaticCoordinatesConversionParams(this));
   
   /**
    * Create a new output manager instance, with the specified driver.
    *
    * @param    driver
    *           Implements low level input/output primitives.  Must NOT be <code>null</code>.
    */
   protected OutputManager(ScreenDriver<? extends OutputManager<?>> driver)
   {
      this.driver = driver;
   }
   
   /**
    * Initialize driver. This MUST be the first method called in this class.
    *
    * @param    cfg
    *           The bootstrap configuration file.
    * @param    driver
    *           Implements low level input/output primitives.  This MAY be
    *           <code>null</code> on the first invocation BUT it MUST NOT
    *           be <code>null</code> on re-initializations.
    *           
    * @throws   InterruptedException 
    * @throws   InvocationTargetException 
    */
   @SuppressWarnings({ "rawtypes", "unchecked" })
   public synchronized static void init(BootstrapConfig cfg, ScreenDriver driver)
   throws InterruptedException,
          InvocationTargetException
   {
      OutputManager    mgr    = newInstance(cfg, driver);
      OutputPrimitives output = mgr.driver.getPrimitives();

      mgr.output        = output;
      mgr.defaultOutput = output;
      // Check if the batch mode is forcing in bootstrap config
      mgr.batchModeOverride = isBatchModeForced(cfg);
   }

   /**
    * Force construction of a new instance.
    *
    * @param    cfg
    *           The bootstrap configuration file.
    * @param    driver
    *           Implements low level input/output primitives.  This MAY be
    *           <code>null</code> on the first invocation BUT it MUST NOT
    *           be <code>null</code> on re-initializations.
    * 
    * @return   The newly created instance.
    * 
    * @throws   InterruptedException 
    * @throws   InvocationTargetException 
    */
   public static synchronized OutputManager<? extends OutputPrimitives> 
   newInstance(BootstrapConfig cfg, ScreenDriver<?> driver)
   throws InterruptedException,
          InvocationTargetException
   {
      if (driver == null) 
      {
         String  id    = cfg.getString("client", "driver", "type", CHUI_DRIVER_CONSOLE);
         boolean batch = cfg.getBoolean("client", "mode", "batch", false); 
         
         if (batch || id.equalsIgnoreCase(CHUI_DRIVER_BATCH))
         {
            driver = new BatchDriver(cfg);
         }
         else if (id.equalsIgnoreCase(CHUI_DRIVER_CONSOLE))
         {
            driver = new ConsoleDriver();
         }
         else if (id.equalsIgnoreCase(CHUI_DRIVER_SWING))
         {
            driver = new SwingChuiDriver(cfg);
         }
         else if (id.equalsIgnoreCase(CHUI_DRIVER_WEB))
         {
            driver = new ChuiWebDriver(cfg);
         }
         else if (id.equalsIgnoreCase(GUI_DRIVER_SWING))
         {
            driver = new SwingGuiDriver();
         }
         else if (id.equalsIgnoreCase(GUI_DRIVER_WEB))
         {
            driver = new GuiWebDriver(cfg);
         }
         else
         {
            throw new RuntimeException("Unknown driver " + id);
         }

         driver.init();
      }

      // TODO: when working with streams, a ChUI mgr is needed, too.

      OutputManager<? extends OutputPrimitives> mgr = driver.createOutputManager(cfg);

      defaultTerminal.set(mgr);

      return mgr;
   }

   /**
    * Resolve the widget with the specified ID.
    * 
    * @param    wid
    *           The widget ID.
    *           
    * @return   The widget associated with the specified ID.
    */
   public static Widget<? extends OutputManager<?>> resolveWidget(int wid)
   {
      return instance().getRegistry().getComponent(wid);
   }
   

   /**
    * Get the context-specific instance, if it exists. In a single-client
    * JVM environment, this is equivalent to a singleton.
    * 
    * @return   The instance or <code>null</code> if none exists.
    */
   public static synchronized OutputManager<? extends OutputPrimitives> instance()
   {
      return defaultTerminal.get();
   }
   
   /**
    * Access the current UI driver instance.
    *
    * @return   The current driver or <code>null</code> if the output manager
    *           is not initialized.
    */
   public static synchronized ScreenDriver<?> getDriver()
   {
      OutputManager<?> mgr = instance();
      
      return mgr == null ? null : mgr.driver;
   }
   
   /**
    * Access the instance UI driver.
    *
    * @return   The driver reference or <code>null</code> if the output manager
    *           is not initialized.
    */
   public ScreenDriver<?> getInstanceDriver()
   {
      return driver;
   }

   /**
    * Initializes the proper error writer handler.
    */
   public static synchronized void initErrorWriter()
   {
      initErrorWriter(null);
   }

   /**
    * Initializes the proper error writer handler.
    *
    * @param    session
    *           The server session to get the properties.
    */
   public static synchronized void initErrorWriter(Session session)
   {
      OutputManager<?> mgr = instance();
      
      // This is the pointer to call up the server for properties
      ServerPropertiesInspector spInsp = null;
      
      if (session != null)
      {
         spInsp = (ServerPropertiesInspector) RemoteObject.obtainNetworkInstance(
                                                 ServerPropertiesInspector.class, session);
      }
      
      if ((spInsp != null && spInsp.isBatchMode()) ||
          (mgr != null && mgr.isBatchModeOverride()))
      {
         ErrorManager.initErrorWriter(new ErrorWriterBatch(ThinClient.getInstance()));
      }
      else
      {
         ErrorManager.initErrorWriter(new ErrorWriterInteractive(ThinClient.getInstance()));
      }
   }

   /**
    * Checks batch mode overridding flag.
    * 
    * @return  <code>true</code> batch mode is overridden, <code>false</code> otherwise.
    * 
    */
   public boolean isBatchModeOverride()
   {
      return batchModeOverride;
   }

   /**
    * Checks batch mode background flag.
    * 
    * @return  <code>true</code> batch mode is running in background, <code>false</code>
    *          otherwise.
    * 
    */
   public static boolean isBatchInBackground()
   {
      ScreenDriver<?> sd = getDriver();
      
      return (sd != null) ? sd.inBackgroundMode() : false;
   }
   
   /**
    * Changes the status of the active flag in batch mode if screen drives is in batch mode.
    * 
    * @param    on
    *           The new value of the active flag in batch mode.
    * 
    * @return  <code>true</code> if operation was successful, <code>false</code> otherwise.
    * 
    */
   public static synchronized boolean activateBatch(boolean on)
   {
      ScreenDriver<?> sd = getDriver();
      
      if (sd != null && sd instanceof BatchDriver)
      {
         if (on)
         {
            ((BatchPrimitives)sd.getPrimitives()).activate();
         }
         else
         {
            ((BatchPrimitives)sd.getPrimitives()).deActivate();
         }
         return true;
      }
      else
      {
         return false;
      }
   }

   /**
    * Checks if the client is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public static boolean inBatchMode()
   {
      OutputManager<?> term = instance();
      
      return (term.driver != null) ? term.driver.inBatchMode() : false;
   }
   
   /**
    * Check if the batch mode is overridden via bootstrap config.
    *
    * @param    cfg
    *           The bootstrap configuration file.
    * 
    * @return   <code>true</code> if batch mode overridden, <code>false</code> otherwise.
    */
   private static boolean isBatchModeForced(BootstrapConfig cfg)
   {
      String driverName = cfg.getString("client", "driver", "type", CHUI_DRIVER_CONSOLE);
      
      return cfg.getBoolean("client", "mode", "batch", false) ||
             (driverName == null ? false : driverName.equalsIgnoreCase(CHUI_DRIVER_BATCH));
      
   }
   
   /**
    * Clear area occupied by specified widget.
    * 
    * @param   widget
    *          Reference widget. 
    */
   public abstract void clear(Widget<?> widget);

   /**
    * Access driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   @SuppressWarnings("rawtypes")
   public abstract WidgetFactory getFactory();

   /**
    * Get the widget registry for this output manager.
    * 
    * @return   See above.
    */
   public abstract WidgetRegistry<? extends OutputManager<?>> getRegistry();
   
   /**
    * Check if the given key can raise the STOP condition.
    * 
    * @param    key
    *           The key code to be checked.
    * 
    * @return   <code>true</code> if the key can raise a STOP condition.
    */
   protected abstract boolean raiseStopCondition(int key);

   /**
    * Check if the window being drawn needs to be overridden.
    * 
    * @param    window
    *           The currently active window.
    *           
    * @return   See above.
    */
   protected abstract TitledWindow<?> overrideWindow(TopLevelWindow<?> window);
   
   /**
    * Activates the provided window instance, depending on the output manager type.
    * Implicit, this is a no-op.
    * 
    * @param    window
    *           A window.
    */
   public void activateWindow(TopLevelWindow<?> window)
   {
      // no-op
   }
   
   /**
    * Realize new window, depending on the output manager type.  This is not default window but
    * one is created dynamically.
    * 
    * @param    windowId
    *           A window ID to be realized.
    */
   public void realizeWindow(int windowId)
   {
      // no-op
   }
   
   /**
    * Initialize this output manager.  This is a no-op by default.
    */
   public void init()
   {
      // no-op
   }

   /**
    * Restore the bitmap to a previous state.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   bitmap
    *          The bitmap instance to set as the output primitive's bitmap.
    */
   public OutputManager<?> setBitmap(int windowId, ScreenBitmap bitmap)
   {
      selectWindow(windowId);

      output.setScreenBitmap(bitmap);
      
      releaseWindow();
      
      return this;
   }

   /**
    * Set cursor status (ON/OFF).
    * 
    * @param   on
    *          New cursor state.
    * 
    * @return  <code>true</code> if operation was successful.
    * 
    * @throws  IllegalStateException
    */
   public boolean setCursorStatus(boolean on) throws IllegalStateException
   {
      return driver.setCursorStatus(on);
   }

   /**
    * Clear specified area of the screen.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   origin
    *          Upper left corner of the area to clear.
    * @param   size
    *          Dimensions of the area.
    * @param   color
    *          Color.
    */
   public void clear(int windowId, Point origin, Dimension size, Color color)
   {
      NativePoint norigin = coordinates().pointToNative(origin);
      NativeDimension nsize = coordinates().dimensionToNative(size);
      clear(windowId, norigin, nsize, color);
   }
   
   /**
    * Clear specified area of the screen.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   origin
    *          Upper left corner of the area to clear.
    * @param   size
    *          Dimensions of the area.
    * @param   color
    *          Color.
    */
   public void clear(int windowId, NativePoint origin, NativeDimension size, final Color color)
   {
      selectWindow(windowId);

      NativeRectangle rect = new NativeRectangle(origin, size);
      ScreenBitmap bitmap = screenBitmap(windowId);
      
      if (bitmap.hasCoordinatesOutsideScreen(rect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(rect))
         {
            return;
         }

         if (debug)
         {
            System.err.println("clear: received rectangle with incorrect" +
                               " coordinates: " + rect.toString());
         }

         rect = fixCoordinates(windowId, rect);

         if (rect == null)
         {
            return;
         }
      }

      int left   = rect.left();
      int top    = rect.top();
      int bottom = rect.bottom();
      int right  = rect.right();
      
      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.clearArea(left, top, right, bottom, color);
         return;
      }

      List<NativePoint> frags = null;
      boolean horizontal = size.width >= size.height;
      int len    = horizontal ? size.width : size.height;
      int begin  = horizontal ? top : left;
      int finish = horizontal ? bottom : right;

      for (int i = begin; i <= finish; i++)
      {
         if (horizontal)
         {
            frags = bitmap.getHorizontalFragments(i, left, len);
         }
         else
         {
            frags = bitmap.getVerticalFragments(top, i, len);
         }

         for (NativePoint frag : frags)
         {
            int start = frag.x;
            int end = frag.y;
            int length = end - start + 1;

            int y = 0;
            int x = 0;
            int height = 0;
            int width = 0;

            if (horizontal)
            {
               y = i;
               height = 1;
               x = left + start;
               width = length;
            }
            else
            {
               y = top + start;
               height = length;
               x = i;
               width = 1;
            }
            
            output.clearArea(x, y, x + width - 1, y + height - 1, color);
         }
      }
      
      releaseWindow();
   }
   
   /**
    * Generate beep signal.
    */
   public void beep()
   {
      driver.beep();
   }

   /**
    * Read single keystroke from driver.
    * 
    * @return key code.
    */
   public int readKey()
   {
      return driver.readKey();
   }

   /**
    * Reset current screen mode.
    */
   public void resetMode()
   {
      driver.resetMode();
   }

   /**
    * Suspend output.
    */
   public void suspend()
   {
      driver.suspend();
   }

   /**
    * Resume normal driver operation.
    */
   public void resume()
   {
      driver.resume();
   }

   /**
    * Change terminal type to specified one.
    * 
    * @param    name
    *           New terminal type name.
    */
   public void setTerminalType(String name)
   {
      driver.setTerminalType(name);
   }

   /**
    * Get terminal type.
    * 
    * @return terminal type string.
    */
   public String terminalType()
   {
      return driver.terminalType();
   }

   /**
    * Redirect terminal output to the drawing worker represented by
    * <code>newOps</code> (or back to the interactive terminal if
    * <code>newOps</code> is <code>null</code>) and return the instance
    * representing the current output destination. This may be another
    * <code>null</code> if the interactive terminal is in use.
    * 
    * @param    newOps
    *           The drawing worker or <code>null</code> to remove redirections
    *           and activate the interactive terminal.
    * 
    * @return   The previous drawing worker instance.
    */
   public P switchOutput(P newOps)
   {
      // TODO: output switching doesn't mean just primitive switching... the manager can be 
      //       swiched completely
      P old = output;

      if (newOps == null)
      {
         output = defaultOutput;
      }
      else
      {
         output = newOps;
      }

      return old;
   }

   /**
    * Reports if the terminal is currently redirected to a stream.
    * 
    * @return <code>true</code> if the terminal is redirected to a stream,
    *         <code>false</code> if the terminal is interactive.
    */
   public boolean isRedirected()
   {
      return !output.equals(defaultOutput);
   }

   /**
    * Obtain the interactive output worker.
    * 
    * @return The output worker used by the interactive terminal.
    */
   public P getInteractiveOutput()
   {
      return defaultOutput;
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @param   windowId
    *          The window ID.
    * 
    * @return  The number of columns (in characters).
    */
   public double screenWidth(int windowId)
   {
      selectWindow(windowId);
      
      try
      {
         return coordinates().widthFromNative(output.screenWidth());
      }
      finally
      {
         releaseWindow();
      }
   }

   /**
    * Gets the number of rows on the screen.
    * 
    * @param   windowId
    *          The window ID.
    * 
    * @return  The number of rows (lines of characters).
    */
   public double screenHeight(int windowId)
   {
      selectWindow(windowId);

      try
      {
         return coordinates().heightFromNative(output.screenHeight());
      }
      finally
      {
         releaseWindow();
      }
   }

   /**
    * Get screen dimensions.
    * 
    * @param   windowId
    *          The window ID.
    *
    * @return screen dimension.
    */
   public Dimension screenDimension(int windowId)
   {
      return new Dimension(screenWidth(windowId), screenHeight(windowId));
   }

   /**
    * Reset the screen bitmap.
    * 
    * @param   windowId
    *          The window ID.
    */
   public void resetBitmap(int windowId)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      if (bitmap != null)
      {
         bitmap.resetScreen();
      }
   }

   /**
    * Get a copy of the current screen bitmap and reset clipping entirely. All
    * screen positions become available for output. The previous state can be
    * restored using setBitmap() method.
    * 
    * @param   windowId
    *          The window ID.
    * 
    * @return  a copy of the current screen bitmap.
    */
   public ScreenBitmap getAndResetBitmap(int windowId)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      ScreenBitmap copy = bitmap.getCopy();
      bitmap.addRectangle(null);

      return copy;
   }

   /**
    * Get a copy of the current screen bitmap and clip output further so it is
    * guaranteed not to go outside of the provided rectangle.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   rect
    *          The extra clipping rectangle. <code>null</code> has no effect
    *          on clipping.
    * 
    * @return  a copy of the current screen bitmap.
    */
   public ScreenBitmap getAndSetBitmap(int windowId, Rectangle rect)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      ScreenBitmap copy = bitmap.getCopy();

      if (rect != null)
      {
         bitmap.clipRectangle(coordinates().rectangleToNative(rect));
      }

      return copy;
   }

   /**
    * Get a copy of the current screen bitmap.
    * 
    * @param   windowId
    *          The window ID. 
    * 
    * @return  a copy of the current screen bitmap.
    */
   public ScreenBitmap getBitmapCopy(int windowId)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      return bitmap.getCopy();
   }

   /**
    * Mark a rectangle as needing repainting. Makes sense with the double
    * buffered output.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   rect
    *          rectangle to mark for repainting
    */
   public void refresh(int windowId, Rectangle rect)
   {
      selectWindow(windowId);

      NativeRectangle nrect = coordinates().rectangleToNative(rect);
      output.refresh(nrect);
      
      releaseWindow();
   }
   
   /**
    * Synchronize output with physical terminal.
    * 
    * @param   windowId
    *          The window ID. 
    * 
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager<?> sync(int windowId)
   {
      selectWindow(windowId);

      output.sync();
      
      releaseWindow();
      return this;
   }

   /**
    * Add a new clipping rectangle to the bitmap. If the rectangle is null,
    * fill the entire bitmap with <code>true</code> values. Else, if the
    * bitmap was just reset, fill it with <code>false</code> values and set
    * the clipping region to true.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   rect
    *          The clipping rectangle to be added.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager<?> clipWith(int windowId, Rectangle rect) 
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      // chui - MessageAreaImpl.setMessageTExt. TitledWindow.processEvent
      NativeRectangle nrect = coordinates().rectangleToNative(rect);
      if (nrect != null && bitmap.hasCoordinatesOutsideScreen(nrect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(nrect))
         {
            return this;
         }

         if (debug)
         {
            System.err.println("clipWith: received rectangle with incorrect" +
                               " coordinates: " + nrect.toString());
         }

         nrect = fixCoordinates(windowId, nrect);

         if (nrect == null)
         {
            return this;
         }
      }

      bitmap.addRectangle(nrect);
      return this;
   }

   /**
    * Create a copy of this screen bitmap and count the invalid chars. This
    * initialization is needed to start searching for the frame from where to
    * start drawing.
    * 
    * @param   windowId
    *          The window ID. 
    */
   public void startScreenSearch(int windowId)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      bitmap.startScreenSearch();
   }

   /**
    * Get the total number of available clipped characters, from the search
    * bitmap screen.
    * 
    * @param   windowId
    *          The window ID. 
    * 
    * @return  The total number of invalid characters.
    */
   public int getInvalidChars(int windowId)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      return bitmap.getInvalidChars();
   }
   
   /**
    * Get the list of clipping rectangles intersecting the specified boundary.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   bounds
    *          The boundary.
    *           
    * @return  A list of intersecting rectangles.
    */
   public List<Rectangle> getClippings(int windowId, Rectangle bounds)
   {
      CoordinatesConversion c = coordinates();
      NativeRectangle nbounds = c.rectangleToNative(bounds);
      
      List<NativeRectangle> nrects = getClippings(windowId, nbounds);
      ArrayList<Rectangle> result = new ArrayList<>(nrects.size());
      
      for(NativeRectangle nrect : nrects)
      {
         result.add(c.rectangleFromNative(nrect));
      }
      
      return result;
   }

   /**
    * Get the list of clipping rectangles intersecting the specified boundary.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   nbounds
    *          The boundary.
    *           
    * @return  A list of intersecting rectangles.
    */
   public List<NativeRectangle> getClippings(int windowId, NativeRectangle nbounds)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);
      return bitmap.getClippings(nbounds);
   }
   
   /**
    * Mark a certain area as searched and decrement the total number of
    * clipped characters.
    * 
    * @param   windowId
    *          The window ID. 
    * @param   rect
    *          Area to mark as searched.
    * 
    * @return  <code>true</code> if this area contains a clipped character,
    *          which was not part of an upper widget.
    */
   public boolean markArea(int windowId, Rectangle rect)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      NativeRectangle nrect = coordinates().rectangleToNative(rect);
      if (nrect != null && bitmap.hasCoordinatesOutsideScreen(nrect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(nrect))
         {
            return false;
         }

         if (debug)
         {
            System.err.println("markArea: received rectangle with incorrect" +
                               " coordinates: " + nrect.toString());
         }

         nrect = fixCoordinates(windowId, nrect);

         if (nrect == null)
         {
            return false;
         }
      }

      return bitmap.markArea(nrect);
   }

   /**
    * Check if the given location can draw data.
    *
    * TODO: this is no longer used, determine if it should be removed
    * 
    * @param   windowId
    *          The window ID. 
    * @param   rect
    *          The area of the widget which is checked for drawing.
    * 
    * @return  <code>true</code> if screen is reset or any of the bitmap
    *          locations within the given rectangle are <code>true</code>
    *          <code>false</code> if nothing can be drawn.
    */
   public boolean drawNeeded(int windowId, Rectangle rect)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      NativeRectangle nrect = coordinates().rectangleToNative(rect);
      if (nrect != null && bitmap.hasCoordinatesOutsideScreen(nrect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(nrect))
         {
            return false;
         }

         if (debug)
         {
            System.err.println("drawNeeded: received rectangle with incorrect" +
                               " coordinates: " + nrect.toString());
         }

         nrect = fixCoordinates(windowId, nrect);

         if (nrect == null)
         {
            return false;
         }
      }

      // bypass if there is no clipping
      if (bitmap.isScreenReset())
      {
         return true;
      }

      return bitmap.drawNeeded(nrect);
   }

   /**
    * Sets the invalidation mode.
    * <p>
    * When invalidation mode is set to <code>true</code>, all
    * <code>PaintEvent</code>s are processed differently. Instead of causing
    * immediate drawing, they contribute their respective rectangles to the
    * invalidation rectangle. The moment the invalidation mode is being set to
    * <code>false</code>, the cumulative invalidation rectangle triggers
    * selective redrawing of the affected area.
    * 
    * @param   mode
    *          <code>true</code> is an opening bracket for the invalidation
    *          mode <code>false</code> is a closing bracket which
    *          conditionally flushes the accumulated paint requests
    */
   public void setInvalidate(boolean mode)
   {
      setInvalidate(null, mode, false);
   }

   /**
    * Sets the invalidation mode.
    * <p>
    * When invalidation mode is set to <code>true</code>, all
    * <code>PaintEvent</code>s are processed differently. Instead of causing
    * immediate drawing, they contribute their respective rectangles to the
    * invalidation rectangle. The moment the invalidation mode is being set to
    * <code>false</code>, the cumulative invalidation rectangle triggers
    * selective redrawing of the affected area.
    * 
    * @param   widget
    *          The widget which triggered this painting.
    * @param   mode
    *          <code>true</code> is an opening bracket for the invalidation
    *          mode <code>false</code> is a closing bracket which
    *          conditionally flushes the accumulated paint requests
    * 
    * @throws  NullPointerException
    *          If <code>mode</code> is false and the <code>widget</code> is <code>null</code>.
    */
   public void setInvalidate(Widget<?> widget, boolean mode)
   {
      if (!mode && widget == null)
      {
         throw new NullPointerException("The widget must be specified!");
      }
      
      setInvalidate(widget, mode, false);
   }

   /**
    * Sets the invalidation mode.
    * <p>
    * When invalidation mode is set to <code>true</code>, all
    * <code>PaintEvent</code>s are processed differently. Instead of causing
    * immediate drawing, they contribute their respective rectangles to the
    * invalidation rectangle. The moment the invalidation mode is being set to
    * <code>false</code>, the cumulative invalidation rectangle triggers
    * selective redrawing of the affected area.
    * 
    * @param   widget
    *          The widget which triggered this drawing.  If is <code>null</code>, the window 
    *          as returned by {@link WindowManager#findWindow} will be used; otherwise, the 
    *          widget's window will be used.
    * @param   mode
    *          <code>true</code> is an opening bracket for the invalidation
    *          mode <code>false</code> is a closing bracket which
    *          conditionally flushes the accumulated paint requests
    * @param   flush
    *          only meaningful for mode <code>false</code>. requests
    *          unconditional flushing of all accumulated paint requests
    * 
    */
   public void setInvalidate(Widget<?> widget, boolean mode, boolean flush)
   {
      if (mode)
      {
         invalidate++;
         return;
      }

      invalidate--;

      if (drawing)
         return;
      
      // when a top-level window is being closed we don't want to proceed, its renderer 
      // is not active anyway
      // TODO: is this right? wouldn't it be cleaner not to allow this call when modal 
      //       window is being closed? 
      if (widget instanceof TopLevelWindow && (!widget.isDisplayed() || !widget.isVisible()))
         return;

      @SuppressWarnings("unchecked")
      TopLevelWindow<?> win = widget != null ? widget.parentOrSelf(TopLevelWindow.class) : null;
      
      if (win == null)
      {
         if (isChui())
         {
            win = WindowManager.resolveWindow();
         }
         else if (widget == null)
         {
            win = WindowManager.getDefaultWindow();
         }
         else if (!widget.isDisplayed())
         {
            // the widget is no longer displayed
            return;
         }
         else
         {
            throw new NullPointerException("Could not resolve window!");
         }
      }
      
      activateWindow(win);
      selectWindow(win.getId().asInt());
      
      ScreenBitmap bitmap = screenBitmap(win.getId().asInt());

      if ((invalidate == 0 || flush) && bitmap.hasRectangles())
      {
         drawing = true;
         
         // in ChUI, all drawing is done on the same default window.  When stacked windows are 
         // used (i.e. alert-boxes or other kind of TitledWindow instance), we need to  
         // differentiate between the default-window and the top-level pseudo-window.
         TitledWindow<?> twin = overrideWindow(win);

         // capture all events raised during drawing - these must not be leaked to the current
         // event queue
         try
         {
            EventManager.pushEventQueue();
            twin.draw();
         }
         finally
         {
            Event event = null;
            while ((event = EventManager.getEvent()) != null)
            {
               // consume all events, but log them
               String msg = 
                  "DRAW: Ignoring event %s %d raised while drawing widget %s with id [%d].";
               String evtName = event.getClass().getName();
               int wid = UiUtils.getWidgetIdAsInt(event.source());
               String wname = event.source().getClass().getName();
               
               System.err.println(String.format(msg, evtName, event.id(), wname, wid));
            }
            
            EventManager.popEventQueue();
         }

         sync(win.getId().asInt());

         // the reference might have changed during drawing
         bitmap = screenBitmap(win.getId().asInt());
         bitmap.resetScreen();

         drawing = false;
      }
      
      releaseWindow();
   }

   /**
    * Resets the invalidation mode. This method should be used exclusively as
    * a way of recovering after exceptions in the client before the control is
    * returned to the server. Such a recovery is required to keep the client
    * side drawing balanced.
    * 
    * @param   widget
    *          The widget which triggered this painting.
    */
   public void resetInvalidate(Widget<?> widget)
   {
      // setting the counter to 1 makes the call that follows reset it to 0
      // and draw the whole invalid rectangle
      invalidate = 1;
      setInvalidate(widget, false, true);
   }

   /**
    * Returns and resets the invalidation mode. This method should be used for
    * activating triggers which, by definition, is the source of recursion.
    *
    * @param   widget
    *          The widget which triggered this painting.
    * 
    * @return  current invalidation level
    */
   public int saveInvalidate(Widget<?> widget)
   {
      int save = invalidate;

      resetInvalidate(widget);

      return save;
   }

   /**
    * Restores the invalidation mode. This method should be used for
    * activating triggers which, by definition, is the source of recursion.
    * 
    * @param   savedLevel
    *          saved current invalidation level
    */
   public void restoreInvalidate(int savedLevel)
   {
      invalidate = savedLevel;
   }

   /**
    * Queries the invalidation mode.
    * 
    * @return  <code>true</code> if invalidation is on
    */
   public boolean isInvalidate()
   {
      return invalidate > 0;
   }

   /**
    * Get screen interface type.
    * 
    * @return  <code>true</code> is ChUI <code>false</code> is GUI
    */   
   public boolean isChui()
   {
      return driver.isChui();
   }
   
   /**
    * Limit available the drawing area to the specified boundaries.
    * <p>
    * By default this is a no-op.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(Rectangle bounds)
   {
      // no-op
   }
   
   /**
    * Returns the output manager's coordinates conversion logic implementation.
    * 
    * @return   a reference to the {@link CoordinatesConversion} instance. 
    */
   public CoordinatesConversion coordinates()
   {
      return coordinates;
   }

   /**
    * Fix the coordinates for the given rectangle. If any of the coordinates
    * is outside the screen, fix it. If the top/bottom or left/right are both
    * outside the screen, return <code>null</code>; this means the rectangle
    * should not be fixed, as it refers to an area outside the screen.
    * 
    * @param   windowId
    *          The window ID.
    * @param   rect
    *          The rectangle which needs its coordinates fixed.
    * 
    * @return  a rectangle instance with its coordinates fixed or
    *          <code>null</code> if the rectangle is referring to an area
    *          outside the screen.
    */
   protected NativeRectangle fixCoordinates(int windowId, NativeRectangle rect)
   {
      ScreenBitmap bitmap = screenBitmap(windowId);

      if (rect.left() < 0)
      {
         if (rect.right() >= 0)
         {
            rect = new NativeRectangle(rect.top(), 0, rect.bottom(), rect.right());
         }
         else
         {
            return null;
         }
      }
   
      if (rect.top() < 0)
      {
         if (rect.bottom() >= 0)
         {
            rect = new NativeRectangle(0, rect.left(), rect.bottom(), rect.right());
         }
         else
         {
            return null;
         }
      }
   
      // the bitmap.rows value is EXCLUSIVE (doesn't include the last row);
      // OTOH, the Rectangle._bottom/_top are INCLUSIVE.
      if (rect.bottom() >= bitmap.rows())
      {
         if (rect.top() < bitmap.rows())
         {
            rect = new NativeRectangle(rect.top(), rect.left(), bitmap.rows() - 1,
                                 rect.right());
         }
         else
         {
            return null;
         }
      }
   
      // the bitmap.cols value is EXCLUSIVE (doesn't include the last column);
      // OTOH, the Rectangle._left/_right are INCLUSIVE.
      if (rect.right() >= bitmap.cols())
      {
         if (rect.left() < bitmap.cols())
         {
            rect = new NativeRectangle(rect.top(), rect.left(), rect.bottom(), 
                                       bitmap.cols() - 1);
         }
         else
         {
            return null;
         }
      }
   
      return rect;
   }
   
   /**
    * Determine the screen bitmap associated with the given window.
    * <p>
    * After this call, the {@link #output} will have selected the specified window.
    * <p>
    * WARNING! If the {@link ScreenBitmap} reference is saved and a code is called which in turn
    * ends up changing the {@link ScreenBitmap} reference (via the {@link #setBitmap} API), then
    * ensure the proper reference is used!
    * 
    * @param    windowId
    *           The window ID.
    *           
    * @return   The {@link ScreenBitmap} associated with this window.
    */
   protected ScreenBitmap screenBitmap(int windowId)
   {
      selectWindow(windowId);

      try
      {
         return output.getScreenBitmap();
      }
      finally
      {
         releaseWindow();
      }
   }

   /**
    * On a device which allows multiple windows to be created, this allows selecting an explicit
    * window before using the primitives or before drawing a widget.
    * 
    * @param    windowId
    *           Window ID.
    */
   public void selectWindow(int windowId)
   {
      // no-op
   }
   
   /**
    * On a device which allows multiple windows to be created, this allows releasing the window to
    * allow drawing from other threads.
    */
   public void releaseWindow()
   {
      // no-op
   }
}