/*
** Module   : SelectionList.java
** Abstract : An selection list widget container and scroll pane holder.
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 SIY 20060208   @24454 Created initial version
** 002 SIY 20060210   @24499 Fixed initial coordinates (resolves widget
**                           placement).
** 003 SIY 20060210   @24517 Added getSelection() method, added 
**                           ControlSetComponent interface.
** 004 SIY 20060222   @24642 Added necessary methods to support changes in
**                           DataContainer interface.
** 005 NVS 20060322   @25199 Added isModified() method since now it's part
**                           of the DataContainer interface.
** 006 SIY 20060322   @25214 Added setNoLabels() to match new 
**                           LabeledWidget interface.
** 007 SIY 20060515   @26157 Added getColumnLabelText().
** 008 SIY 20060824   @28858 Removed getP2JType() method.
** 009 NVS 20061214   @31694 Added draw() method override for this class.
** 010 SIY 20070807   @34814 Added isColumnFixed() method.
** 011 SVL 20071221   @36532 Label is shown/hidden together with the
**                           widget. 
** 012 SIY 20080320   @37604 Fixed handling of the SCREEN-VALUE and data
**                           value.
** 013 NVS 20080401   @37781 Added isFocusable() method implementation.
** 014 NVS 20080625   @38920 Added focus related method overrides so that
**                           SelectionList delegates all calls down to
**                           the embedded SelectioListBody. CA:
**                           Override the setEnable() method so the body
**                           of the selection list will have the same
**                           enabled status as the parent container.
** 015 SIY 20080714   @39103 Reordered statements in constructor in order
**                           to avoid call to setEnabled() before list is
**                           constructed.
** 016 SIY 20100619          Switching to own implementation of character UI.
** 017 SIY 20100804          Changes in imports, some refactorings.
** 018 SIY 20100822          Refactored AttributeContainer support.
** 019 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 020 SIY 20111202          Refactored and moved from CHUI package.
** 021 SIY 20111221          Switched to widget factory.
** 022 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 023 HC  20140911          GUI coordinate support - data types refactoring.
** 024 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 025 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 026 HC  20150323          Removed business logic from config classes.
** 027 EVL 20150820          Separating common code from UI specific.  Adding SCROLL-TO-ITEM()
**                           method support. 
** 028 HC  20150831          Scrolling support overhaul.
** 029 IAS 20150924          AUTO-RESIZE attribute support
** 030 HC  20151112          Implemented THREE-D runtime support.
**     HC  20151117          Fixed broken ChUI SELECTION-LIST missing border. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.model.ListSelectionModel;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * This class provides necessary scrolling for the
 * <code>SelectionListBody</code> class which implements remaining
 * functionality. Another purpose of this class is to maintain its size and
 * adjust size of scroll pane and selection list.
 * 
 * @author  SIY
 * @version 1.0
 */
@SuppressWarnings("unchecked")
public abstract class SelectionList<O extends OutputManager<?>>
extends LabeledPanel<O,SelectionListConfig>
implements DataContainer,
           FocusListener,
           ControlSetComponent
{
   /** ScrollPane which wraps the list of items and make scrolling. */
   protected ScrollPane<O> scrollpane;

   /** Component attribute container. */
   protected SelectionListBody<O> list = null;
   
   /** adjustSize() was called */
   protected boolean sizeAdjusted = false;
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   @SuppressWarnings("unchecked")
   public SelectionList(WidgetId id, SelectionListConfig cfg)
   {
      // no-op
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @SuppressWarnings("unchecked")
   @Override
   public void initialize(WidgetId id, SelectionListConfig cfg)
   {
      setLocation(new Point(BaseConfig.INV_COORD, BaseConfig.INV_COORD));

      WidgetFactory<O> factory = screen().getFactory();
      
      config = (SelectionListConfig) ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      
      // TODO: Do not share this.config with the body.
      //       Also remove SelectionListBody.location()/physicalLocation() when this is resolved.
      list = factory.create(SelectionListBody.class, id, config); // the body shares the ID&Config
      scrollpane = factory.createScrollPane(list);
      scrollpane.setBorder(factory.createLineBorder());
      list.initialize(id, config);
      
      add(scrollpane);
      addFocusListener(this);
      
      setEnabled(false);
      
      // this needs to be done last.
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

    /**
     * Returns the focused direct child Component.
     *
     * @return The focused direct child Component.
     */
    @Override
   public Widget<O> currentFocus()
   {
      return list;
   }

   /**
    * This is called when the selection-list takes the input focus. When the
    * help string is defined for the selection-list it should be displayed at
    * the window status line.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent focusEvent)
   {
      // pass this focus event down to the list body
      list.onFocusGained(focusEvent);
   }

   /**
    * This is called when the selection-list looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      // pass this focus event down to the list body
      list.onFocusLost(focusEvent);
   }

   /**
    * Forward call to list container.
    */
   @Override
   public void refreshItems()
   {
      list.refreshItems();
   }

   /**
    * Request indexes of selected items.
    * 
    * @return  array of 0-based indexes of selected items.
    */
   @Override
   public int[] getSelection()
   {
      return list.selection();
   }
   
   /**
    * Forward call to list container.
    * 
    * @param   mode
    *          New selection mode.
    */
   public void setSelectionMode(int mode)
   {
      list.setSelectionMode(mode);
   }
   
   /**
    * Adjust SelectionList size.
    */
   public void adjustSize()
   {
      // do UI specific work
      adjustSizeImpl();

      doLayout();
   }
   
   /**
    * Getting the currently selected item.
    *
    * @return  The BaseDataType object representing the current value of
    *          the selection-list.
    */
   @Override
   public BaseDataType getValue()
   {
      return list.getValue();
   }

   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return list.isModified();
   }
   
   /**
    * Sets the BaseDataType instance representing the current value.
    *
    * @param    value
    *           The value representing the current widget value.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      list.setValue(value);
   }

   /**
    * Get text representation of the currently selected value.
    * 
    * @return  The screen representation of the value.
    */
   @Override
   public String getText()
   {
      return list.getText();
   }
   
   /**
    * Set the enabled flag for this selection list.
    * 
    * @param   flag_
    *          The enabled status for this selection list.
    */
   @Override
   public void setEnabled(boolean flag_)
   {
      super.setEnabled(flag_);
      list.setEnabled(flag_);
   }
   
   /**
    * Draw all the components in this container.
    */
   @Override
   public abstract void draw(); 

   /**
    * Shows or hides the selection list and its label.
    *
    * @param    visible
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (getLabelInstance() != null)
         getLabelInstance().setVisible(visible);
   }

   /**
    * Get screen value.
    * 
    * @return  Current screen value.
    */
   public BaseDataType getScreenValue()
   {
      return list.getScreenValue();
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(SelectionListConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);
      
      int mode = (config().multiple) ? ListSelectionModel.MULTIPLE_INTERVAL_SELECTION
                                     : ListSelectionModel.SINGLE_SELECTION;
      
      setSelectionMode(mode);
      refreshItems();
      // when enabled, the following check is incorrect. 
      // It results in a zero widget size in some situations: it might be related to using the
      // INNER-CHARS/LINES or scrollbar options set at the selection-list definition.
//      if (!sizeAdjusted)
//      {
         adjustSize();
//         sizeAdjusted = true;
//      }
   }

   /**
    * Implements SCROLL-TO-ITEM() method for selection-list widget.
    *
    * @param    index
    *           The 0-based index of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   public boolean scrollToItem(int index)
   {
      return list.scrollToItem(index);
   }

   /**
    * Adjust SelectionList size UI specific part.
    */
   protected abstract void adjustSizeImpl();
}