/* Module   : Space.java
** Abstract : Space artificial widget implementation for CHUI
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050627   @21556 Creation initial version based on abstract
**                           Component class.
** 002 SIY 20050831   @22517 Refactoring (see @@22394-22409) 
** 003 SIY 20050907   @22651 Fixed use of @inheritDoc tag in javadoc
** 004 SIY 20060209   @24444 Updated to accommodate changes in xxxConfig
**                           classes.
** 005 SIY 20060824   @28860 Removed getP2JType() method.
** 006 NVS 20080401   @37783 Added isFocusable() method implementation.
** 007 SIY 20100619          Switching to own implementation of character UI.
** 008 SIY 20100804          Changes in imports, some refactorings.
** 009 SIY 20100822          Refactored AttributeContainer support.
** 010 SIY 20111101          Updated imports.
** 011 SIY 20111202          Refactored and moved from CHUI package.
** 012 HC  20140911          GUI coordinate support - data types refactoring.
** 013 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 014 HC  20150323          Removed business logic from config classes.
** 015 EVL 20160420          Changing the skip length internal to double to have enough presision
**                           for GUI mode.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The virtual invisible widget to represent the space(n) keyword feature for
 * the CHUI frames. Used to simulate the additional empty columns between
 * widgets in a frames or dialog boxes.
 * <p>
 * NOTE: it is expected that widget will be UI-neutral.
 */
public class Space<O extends OutputManager<?>>
extends AbstractWidget<O>
implements WidgetWithConfig<SkipConfig>
{
   /** Component attribute container */
   protected SkipConfig config = null;
   
   /**
    * Constructs the skipping to the next column.
    * 
    * @param   id
    *          The widget's ID.
    */
   public Space(WidgetId id)
   {
      this(id, 1.0);
   }
   
   /**
    * Constructs the space object with the given number of the gap columns.
    * 
    * @param   id
    *          The widget's ID.
    * @param   columnsToSkip
    *          The number of the columns to be skipped.
    */
   public Space(WidgetId id, double columnsToSkip)
   {
      super();
      
      config = ConfigManager.getInstance().resolveWidgetConfig(id, SkipConfig.class);
      config.vertical = false;
      boolean isChui = ThinClient.getInstance().isChui();
      if ((isChui && columnsToSkip > 1.0) || (!isChui && columnsToSkip > 0.0))
      {
         config.length = columnsToSkip;
      }
      
      //config.setVisible(false);
      setVisible(false);
   }
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Space(WidgetId id, SkipConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SkipConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);

      setVisible(false);
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public SkipConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(SkipConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Returns the current width of the object.
    * 
    * @return  The horizontal size of the space object.
    */
   @Override
   public double width()
   {
      return config.length;
   }
   
   /**
    * Returns the current height of the component.
    * 
    * @return   Always zero.
    */
   @Override
   public double height()
   {
      return 0.0;
   }
   
   /**
    * Getting the size of the space object. This is the implementation of the
    * abstract methods for the CHARVA Component.
    * 
    * @return  The size of the space object, zero height.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(config.length, 0.0);
   }
   
   /**
    * The same as the getSize() method. This is the implementation of the
    * abstract methods for the CHARVA Component.
    * 
    * @return  The size of the space object, zero width.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * Draw method, does nothing. This is the implementation of the abstract
    * methods for the CHARVA Component.
    */
   @Override
   public void draw()
   {
   }
   
   /**
    * This component will not receive an input focus when Tab or Shift-Tab is
    * pressed.
    * 
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }

   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }
}