/*
** Module   : StatusLine.java
** Abstract : Status line widget.
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 NVS 20080828   @39620 Created initial version
** 002 SIY 20100619          Switching to own implementation of character UI.
** 003 SIY 20100912          Changed hierarchy in order to avoid need for repaintBackdoor(). 
**                           Renamed class from P2JStatusLabel.
** 004 SIY 20101122          Switching from KeyEvent to KeyInput.
** 005 SIY 20110112          Changes related to new color handling scheme.
** 006 SIY 20111202          Refactored and moved from CHUI package.
** 007 CA  20140711          A new API related to GUI clients.
** 008 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 009 HC  20140911          GUI coordinate support - data types refactoring.
** 010 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 011 HC  20141021          Implemented window move/position support.
** 012 HC  20150323          Removed business logic from config classes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Status line widget.
 * 
 * @author  NVS
 * @version 1.0.0
 */
public abstract class StatusLine<O extends OutputManager<?>>
extends AbstractWidget<O>
implements WidgetWithConfig<BaseConfig>
{
   /** Label color */
   protected Color color = Color.MESSAGES;

   /** Maximal length of the text */
   protected int maxLen = -1;
   
   /** Track virtual null text. */
   protected boolean isNull = true;
   
   /** Label text. */
   protected String labelText;
   
   /** Label width. */
   private double width;

   /** The widget config. */
   protected BaseConfig config;

   /**
    * Construct an empty message line.
    * 
    * @param   id
    *          The widget's ID.
    */
   public StatusLine(WidgetId id)
   {
      labelText = "";
      setLength(0);

      this.config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Attach this widget to the specified window.
    * 
    * @param    window
    *           The parent window.
    */
   public abstract void attach(Window<O> window);
   
   /**
    * Widget drawing method.
    */
   @Override
   public abstract void draw();

   /**
    * Set color specification.
    * 
    * @param   color
    *          Color specification.
    */
   public void setColor(Color color)
   {
      this.color = color;
   }

   /**
    * Return maximal length of the text. Negative value means no limits.
    * 
    * @return  Returns the maximal label length.
    */
   public int getMaxLen()
   {
      return maxLen;
   }

   /**
    * Set maximal label length.
    * 
    * @param   maxLen
    *          The new maximal length.
    */
   public void setMaxLen(int maxLen)
   {
      this.maxLen = maxLen;
   }

   /**
    * Get current color used for displaying.
    * 
    * @return  Current color.
    */
   public Color getColor()
   {
      return color;
   }

   /**
    * Set label text.
    * 
    * @param   text
    *          Label text. 
    */
   public void setText(String text)
   {
      if (text == null)
      {
         isNull = true;
         text = "";
      }
      else
         isNull = false;
   
      labelText = text;
   
      if (text.length() > width)
      {
         width = text.length();
      }
   }

   /**
    * Get label text.
    * 
    * @return  label text. 
    */
   public String getText()
   {
      if (isNull)
         return null;
   
      return getLabelText();
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#minimumSize()
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(this.width(), this.height());
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#width()
    */
   @Override
   public double width()
   {
      return width;
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#height()
    */
   @Override
   public double height()
   {
      return 1;
   }

   /**
    * This component will not receive focus when Tab or Shift-Tab is pressed.
    * 
    * @return  <code>false</code>
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }

   /**
    * This component ignores key events. 
    * 
    * @param   ke 
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
   }

   /**
    * This component never gets the keyboard input focus.
    */
   @Override
   public void requestFocus()
   {
   }

   /**
    * Set component occupied length.
    * 
    * @param   width
    */
   public void setLength(double width)
   {
      this.width = Coordinate.scale(width);
   }

   /**
    * Get label text.
    * 
    * @return  label text.
    */
   public String getLabelText()
   {
      return labelText;
   }

}