/*
** Module   : SubMenu.java
** Abstract : SubMenu abstract client implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150318 Added getFocusedChild() method and overridden focusWorker() method.
**                  Added event processing, repaint, hide/show body methods.
** 004 VIG 20150618 Moved several methods from ChUI implementation.
**                  Fixed hideBody method to be able to calculate menu height correctly.
** 005 VIG 20150906 Added Label field, encapsulating mnemonic logic, fixed drawing.
** 006 VIG 20151105 Overridden ancestor() method, returning owner ancestor for pop-up menu.
** 007 VIG 20151226 Changed focusTraversable(): removed isEnabled() check; 
**                  extracted getFirstFocusableItem() from setInitialFocus(). 
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 VIG 20160222 Label is renamed to MnemonicInfo and createNewLabel() method renamed 
**                  to createLabelInfo().
** 010 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 011 VIG 20160504 Changed getPreparedTitle method for null case.
** 012 VIG 20160627 MnemonicInfo.init() call is renamed to MnemonicInfo.prepare().
** 013 VIG 20160811 parent() is changed to getParentMenuContainer().
*
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>SubMenu</code> client abstract implementation class. 
 */
public abstract class SubMenu<O extends OutputManager<?>>
extends AbstractContainer<O>
implements WidgetWithConfig<SubMenuConfig>,
           MenuElement<O>
{
   /** Menu attribute container. */
   protected SubMenuConfig config = null; 

   /** Label with mnemonic info */
   private MnemonicInfo label;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenu(WidgetId id, SubMenuConfig cfg)
   {
   }
   
   /**
    * Check menu items (first-level) for specified mnemonic. If menu item with such mnemonic found, 
    * than post {@link ItemEvent}.
    * 
    * @param   menu
    *          MENU or SUB-MENU
    * @param   event
    *          Key event with <code>event.keyChar() != KeyInput.CHAR_UNDEFINED</code>
    * 
    * @return  <code>true</code> if corresponding menu item found, otherwise - <code>false</code>
    * 
    * @throws  IllegalArgumentException
    *          If event is generated by other source than character pressing.         
    */
   public static boolean processMnemonicEvent(AbstractContainer<?> menu, KeyInput event)
   {
      if (event.keyChar() == KeyInput.CHAR_UNDEFINED)
      {
         throw new IllegalArgumentException("Specified event is not mnemonic event");
      }
      
      char ch = Character.toLowerCase(event.keyChar());
      for (Widget<?> w : menu.getWidgetsList())
      {
         MnemonicInfo label = w instanceof MenuItem ? 
                      ((MenuItem<?>) w).getPreparedLabel() : 
                      ((SubMenu<?>) w).getPreparedLabel();
                          
         Character mnem = label == null ? null : label.getMnemonic();
                          
         if (mnem != null && Character.toLowerCase(mnem.charValue()) == ch) 
         {
           if (w instanceof SubMenu)
           {
              ((SubMenu<?>) w).menuDropEvents();

              // we are via keys, select initial focus
              ((SubMenu<?>) w).setInitialFocus();
           }
           else
           {
              ((MenuItem<?>) w).selectMenuItem();
           }

           return true;
        }
      }

      return false;
   }
   
   /**
    * Check if specified MENU/SUB-MENU contains SUB-MENU child.
    * 
    * @param    menu
    *           MENU/SUB-MENU container.
    *           
    * @return   <code>true</code> if specified container contains SUB-MENU child.
    */
   public static boolean hasChildSubmenu(Container<?> menu)
   {
      for (Widget<?> w : menu.widgets())
      {
         if (w instanceof SubMenu)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Gets sub-menu title width (with '-&gt;' if sub-menu is in other sub-menu or in popup menu)
    * 
    * @return   width of sub-menu title
    */
   public abstract double titleWidth();
   
   /**
    * Get location of SUB-MENU body, relative to container.
    * 
    * @return  body location.
    */
   protected abstract Point bodyLocation();
   
   /**
    * Calculates body dimension
    * 
    * @return  Dimension
    */
   protected abstract Dimension bodyDimension();
   
   /**
    * Raise the MENU-DROP 4GL events.
    */
   protected abstract void menuDropEvents();
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SubMenuConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public SubMenuConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(SubMenuConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
      
      if (!Objects.equals(config().title, beforeUpdate.title) ||
          (config().title != null && !config().title.isEmpty() && label == null))
      {
         if (label == null)
            label = createLabelInfo();

         label.prepare(config().title);
      }
           
      if (beforeUpdate.title != null && 
          !beforeUpdate.title.isEmpty() && 
          (config().title == null || config().title.isEmpty()))
      {
         label = null;
      }
   }

   /**
    * Get ancestor window for this widget.This method must be called
    * only for widgets that are already attached to Window.
    * 
    * @return  ancestor Window.
    * 
    * @throws  IllegalStateException
    *          When this widget is not attached to window.
    */
   @SuppressWarnings("unchecked")
   @Override
   public TitledWindow<O> ancestor()
   {
      Menu<O> menu = getParentMenu();
      if (menu == null || !menu.config().popupOnly)
      {
         return super.ancestor();
      }
      
      if (menu.config().ownerId == -1)
      {
         int wid = UiUtils.getWidgetIdAsInt(this);
         throw new IllegalStateException("SubMenu " + name() + " " + wid
                                         + " is not attached to a Window instance.");
      }

      Widget<O> owner = ThinClient.getInstance().getWidget(menu.config().ownerId);
      return (TitledWindow<O>) owner.ancestor();
   }

   /**
    * Get MENU-ITEM prepared label.
    * 
    * @return  Label.
    */
   public MnemonicInfo getPreparedLabel()
   {
      return label;
   }
   
   /**
    * Get MENU-ITEM prepared label.
    * 
    * @return  Label.
    */
   public String getPreparedTitle()
   {
      return label == null ? 
             config().title : 
             (label.getPrepLabel() == null ? config().title : label.getPrepLabel());
   }
   
   /**
    * Hides sub-menu body.
    * 
    * @param   recursiveUp
    *          <code>true</code> means hide body for all parent sub-menus.
    */
   public void hideBody(boolean recursiveUp)
   {
      for (Widget<O> w : widgets())
      {
         w.setVisible(false);
      }

      if (recursiveUp)
      {
         AbstractContainer<O> parent = getParentMenuContainer();
         if (parent instanceof SubMenu)
         {
            ((SubMenu<?>) parent).hideBody(true);
         }
         else if (parent instanceof Menu)
         {
            Menu<?> menu = (Menu<?>)parent();
            if (menu.config().popupOnly)
            {
               ((Menu<?>) parent).hide();
            }
            else
            {
               menu.repaint(); // menubar
            }
         }
      }

      Rectangle bodyRect = 
               new Rectangle(bodyLocation(), bodyDimension(), screen().coordinates().baseUnits());
     
      repaint(bodyRect.translate(-1, 0));
      
      // repaint needs to be called for the entire body (as it is displayed), before turning the
      // display flag off; otherwise, the repaint rectangle will not contain the body, and just
      // the title
      repaint();

      config().displayed = false;
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      int actionCode = event.actionCode();
      
      // menu-item choosing events processing
      if (actionCode == Keyboard.SE_VALUE_CHANGED ||
          actionCode == Keyboard.SE_CHOOSE)
      {
         hideBody(true);
         return;
      }

      if (key == Key.VK_ENTER                                                    ||
          key == Keyboard.SE_MENU_DROP                                           ||
          (Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_DOWN) ||
          (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_RIGHT))
      {
         // What if body is already displayed? In this case there will be focused item in 
         // the body, that will process this event.
 
         showBody();
         if (key != Keyboard.SE_MENU_DROP)
         {
            menuDropEvents();
            
            // we are via keys, select initial focus
            setInitialFocus();
         }
         
         parent().repaint();
         event.consume();
         return;
      }
      
      if ((Menu.isMenubarElement(this)  && actionCode == Keyboard.KA_CURSOR_RIGHT) || 
          (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_DOWN))
      {
         if (!Menu.isMenubarElement(this)          && 
             actionCode == Keyboard.KA_CURSOR_DOWN && 
             config().displayed == true)
         {
            setInitialFocus();
         }
         else
         {
            ThinClient.nextFocus(this);
         }
         
         parent().repaint();
         return;
      }
      
      if ((Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_LEFT) || 
          (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_UP))
      {
         if (!Menu.isMenubarElement(this)        && 
             actionCode == Keyboard.KA_CURSOR_UP &&
             config().displayed == true)
         {
            setInitialFocus();
         }
         else
         {
            ThinClient.previousFocus(this);
         }
         
         parent().repaint();
         return;
      }
      
      // hiding sub-menu from menu-item: only current if LEFT
      if (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_LEFT)
      {
         SubMenu<O> parentSub = getParentSubMenu();
         if (parentSub != null)
         {
            parentSub.hideBody(false);

            if (Menu.isMenubarElement(parentSub))
            {
               ThinClient.previousFocus(parentSub);
            }
         } 
         else if (getParentMenu().config().popupOnly) // pop-up menu
         {
            getParentMenu().hide();
         }
         return;
      }
         
      // mnemonic
      if (event.keyChar() != KeyInput.CHAR_UNDEFINED)
      {
         if (this.config().displayed)
         {
            SubMenu.processMnemonicEvent(this, event);
            return;
         }
      }
      
      
      if (event.isRealKey())
      {
         getParentMenuContainer().processKeyEvent(event);
      }
      else
      {
         // Can lead to infinite loop, because AbstractContainer.processKeyEvent()
         // delegates processing to its current focused widget.
         // super.processKeyEvent(event);
      }
   }
   
   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      super.repaint();
      
      if (parent() instanceof SubMenu || parent() instanceof Menu)
      {
         parent().repaint();
      }
   }
   
   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      for (Widget<O> w : widgets())
      {
         w.setVisible(visible);
      }

      super.setVisible(visible);
   }

   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }
   
   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget<O> defaultFocus()
   {
      Widget<O> defFocus = null;

      for (Widget<O> c : widgets())
      {
         if (c.focusTraversable())
         {
            defFocus = c;
            break;
         }
      }

      if (defFocus != null && defFocus instanceof Container)
         return ((Container<O>) defFocus).defaultFocus();

      return defFocus;
   }
   
   /**
    * Check if widget is focus traversable.
    *
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      // The same as AbstractWidget.focusTraversable() except isEnabled().
      // children focusability doesn't affect sub-menu focusability.
      return isVisible() && !hidden();
   }
   
   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }

   /**
    * Creates new MnemonicInfo instance.
    * 
    * @return   MnemonicInfo or its descendant.
    */
   protected MnemonicInfo createLabelInfo()
   {
      return new MnemonicInfo();
   }
   
   /**
    * Get first focused sub-menu child (menu-iter or other sub-menu)
    * 
    * @return   MenuItem or SubMenu
    */
   protected Widget<?> getFocusedChild()
   {
      for(Widget<?> w : widgets())
      {
         if (w.hasFocus())
         {
            return w;
         }
      }
      
      return null;
   }

   /**
    * Search for the next or previous focusable widget in the container and
    * set focus on it.
    *
    * @param   direction
    *          Widget search direction.
    */
   protected void focusWorker(boolean direction)
   {
      cycleFocus(direction);
   }

   /**
    * Shows this sub-menu body.
    */
   protected void showBody() 
   {
      for (Widget<O> w : widgets())
      {
         w.setVisible(true);
      }
      config().displayed = true;
      Point baseUnits = screen().coordinates().baseUnits();
      Rectangle bodyRect = new Rectangle(bodyLocation(), bodyDimension(), baseUnits);
      repaint(bodyRect);
   }

   /**
    * Sets initial focus on first traversable child.
    */
   protected void setInitialFocus()
   {
      Widget<O> firstChild = getFirstFocusableItem();     
      
      FocusManager.setInitialFocus(firstChild);
   }

   /**
    * Returns the first focusable child.
    * 
    * @return  focusable child.
    */
   protected Widget<O> getFirstFocusableItem()
   {
      Widget<O> firstChild = null;
      for (Widget<O> w : widgets())
      {
         if (w.focusTraversable())
         {
            firstChild = w;
            break;
         }
      }
      
      return firstChild;
   }
}