/*
** Module   : UiUtils.java
** Abstract : UI-modality-independent helper methods
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20100821 Created initial version
** 002 SIY 20100831 Common static method containsFrames moved here.
** 003 SIY 20100915 More common static methods moved here.
** 004 SIY 20101015 Static methods from Screen moved here.
** 005 SIY 20110112 Changes related to new color handling scheme.
** 006 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation for GUI. 
** 007 SIY 20111202 Temporary changes caused by splitup of widgets into UI-specific and
**                  UI-independent parts. Will be replaced with widget factory.
** 008 SIY 20111221 Switched to widget factory.
** 009 SVL 20120214 Added locateAlertBox.
** 010 MAG 20140410 Add RECTANGLE widget creation and simple round up method.
** 011 EVL 20140610 Added TOGGLE-BOX widget support.
** 012 MAG 20140627 Add Window class to reconstructComponent method. Changed color attribute
**                  method names to match naming standards.
** 013 CA  20140805 Changes related to eliminating the warnings from adding generics to the 
**                  abstract widgets. 
** 014 EVL 20140812 In new widget ID allocation approach ID for default window is taking from
**                  WidgetId class static constant.
** 015 HC  20140911 GUI coordinate support - data types refactoring.
** 016 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 017 VIG 20141025 Added multi-window support for frame methods.
** 018 CA  20141123 The current focus is always obtained from the top window.
** 019 EVL 20141126 Content pane GUI fix.
** 020 EVL 20150115 Adding image widget support.
** 021 VIG 20150116 Added Menu class support.
** 022 CA  20150323 Labels can not be reconstructed via reconstructComponent, they are managed 
**                  expliclty.
** 023 HC  20150323 Removed business logic from config classes.
** 024 CA  20150429 When creating a rectangle from a location and dimension, use the base units
**                  to determine the coordinates for current client mode.
** 025 GES 20150423 Fixed logging and some coding style cleanup.
** 026 CA  20150725 Fixed getTopLevelComponents: use wnd.getWidgets(), as this is a common API for 
**                  GUI and ChUI widgets.  wnd.getContentPane().widgets() works only in ChUI.
** 027 HC  20150905 Added getWidgetIdAsString method.
** 028 CA  20151006 Changes to allow multi-window focus management in GUI.
** 029 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 030 CA  20151129 Added getVisibleComponents.  Added TODO to hideFrames related to frame 
**                  matching being done via the ID, not frame name.
** 031 CA  20151123 getFramesAbove needs to use the frame's siblings, and not frames directly 
**                  added to window (to support nested frames).
** 032 CA  20151127 Added showWindowForFrame: fixed implicit window show by frame statements, in 
**                  GUI.
** 033 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 034 EVL 20160127 Adding slider widget support.
** 035 IAS 20160217 Added lookupWidget versions to lookup widgets which are direct 4GL counterparts
** 036 HC  20160406 Fixed lookupWidgetId() implementation to conform with the method's contract.
** 037 HC  20160421 Added locateRootFrame() method.
** 038 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 039 HC  20160922 Minor performance improvement to locateFrame().
** 040 CA  20161123 Propagate the needPause flag state to all child frames, in resetNeedPause().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import java.util.stream.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Common static helpers for the client-side UI.
 */
public class UiUtils
{
   /** Logger. */
   protected static final Logger LOG = LogHelper.getLogger(UiUtils.class.getName());

   /**
    * Get the non-hidden components from this frame and all child frames.
    * 
    * @param    frame
    *           The frame.
    * @param    stmt
    *           The current UI statement which uses this frame.
    *           
    * @return   The list of non-hidden widgets.
    */
   public static Widget[] getVisibleComponents(Frame frame, UIStatement stmt)
   {
      Widget[] comps = frame.getContentPane().widgets();
      ArrayList<Widget> visibleComps = new ArrayList<Widget>(comps.length);
      boolean bypassFrame = (stmt == UIStatement.ENABLE && 
                             (frame.hidden() || frame.config().wasVisible));
      
      // remove hidden widgets from list
      for (int i = 0; i < comps.length; i++)
      {
         if (comps[i].hidden())
         {
            continue;
         }

         if (bypassFrame && comps[i] instanceof Frame)
         {
            // comp is a frame - if the parent frame is visible, then do not show it, if the
            // current statement is an ENABLE
            continue;
         }
         
         visibleComps.add(comps[i]);
         
         if (comps[i] instanceof Frame)
         {
            visibleComps.addAll(Arrays.asList(getVisibleComponents((Frame) comps[i], stmt)));
         }
      }

      Widget[] wids = new Widget[visibleComps.size()];
      visibleComps.toArray(wids);
      
      return wids;
   }
   
   /**
    * Check if widget has config.
    * 
    * @param   widget
    *          Widget to check.
    *          
    * @return  <code>true</code> if widget is not null and has config.
    */
   public static boolean hasConfig(Widget<?> widget)
   {
      return getConfig(widget) != null;
   }

   /**
    * Get ID of specified widget.
    * 
    * @param    w
    *           Component to for which ID is requested.
    * 
    * @return   The widget ID or {@code null} if the widget doesn't have a config.
    */
   public static WidgetId getWidgetId(Widget<?> w)
   {
      return w == null ? null : w.getId();
   }

   /**
    * Get ID of specified widget.
    * 
    * @param    widget
    *           Component to for which ID is requested.
    * 
    * @return   The widget ID or -1 if the widget doesn't have a config.
    */
   static public int getWidgetIdAsInt(Widget<?> widget)
   {
      WidgetId wid = getWidgetId(widget);
      
      return wid == null ? -1 : wid.asInt();
   }

   /**
    * Get ID of specified widget formatted to string.
    * 
    * @param    widget
    *           Component to for which ID is requested.
    * @param    ifNull
    *           A value to be returned when the widget has no
    *           assigned id.
    * 
    * @return   The widget ID formatted to string.
    */
   static public String getWidgetIdAsString(Widget<?> widget, String ifNull)
   {
      WidgetId wid = getWidgetId(widget);
      
      if (wid == null)
      {
         return ifNull;
      }
      
      return wid.toString();
   }
   
   /**
    * Returns config reference associated with the passed-in widget.
    * 
    * @param    w
    *           A widget reference or <code>null</code>.
    *           
    * @return   config or <code>null</code> if the widget has
    *           no valid associated widget.
    */
   public static WidgetConfig getConfig(Widget<?> w)
   {
      if (w instanceof ConfigOwner)
      {
         return ((ConfigOwner<?>) w).config();
      }
      
      return null;
   }
   
   /**
    * Returns config reference associated with the passed-in widget.
    * 
    * @param    w
    *           A widget reference or <code>null</code>.
    * @param    clazz
    *           The Java class representing the expected config type.
    *           
    * @return   config or <code>null</code> if the widget has
    *           no valid associated widget of the expected type.
    */
   @SuppressWarnings("unchecked")
   public static <C extends WidgetConfig> C getConfig(Widget<?> w, Class<C> clazz)
   {
      WidgetConfig c = getConfig(w);
      
      if (c == null)
      {
         return null;
      }
      
      if (clazz.isAssignableFrom(c.getClass()))
      {
         return (C) c;
      }
      
      return null;
   }
   
   /**
    * Check if there are other frames in specified area. Frame passed as a
    * parameter is not counted.
    * 
    * @param    frame 
    *           Frame to exclude.
    * @param    r
    *           Rectangle to check.
    * 
    * @return   <code>true</code> if there are other frames in specified area.
    */
   public static boolean containsFrames(Frame<?> frame, Rectangle r)
   {
      ArrayList<Frame<?>> frames = new ArrayList<Frame<?>>();
      Window<?> window = frame.window();
      
      window.findFrames(frames, r, true, frame.isRedirected());
        
      frames.remove(frame);
      
      return frames.isEmpty();
   }

   /**
    * Get ID of the widget which has the input focus.
    *
    * @param   widget
    *          Event source.
    *          
    * @return  ID of the widget or <code>-1</code>
    */
   public static WidgetId lookupWidgetId(Widget<?> widget)
   {
      return lookupWidgetId(widget, false);
   }

   /**
    * Get ID of the 4GL widget which has the input focus.
    *
    * @param   widget
    *          Event source.
    *          
    * @return  ID of the widget or <code>-1</code>
    */
   public static WidgetId lookup4GLWidgetId(Widget<?> widget)
   {
      return lookupWidgetId(widget, true);
   }

   /**
    * Get ID of the widget which has the input focus.
    *
    * @param   widget
    *          Event source.
    *          
    * @return  ID of the widget or <code>-1</code> if the widget doesn't have a config.
    */
   public static int lookupWidgetIdAsInt(Widget<?> widget)
   {
      return lookupWidgetIdAsInt(widget, false);
   }

   /**
    * Get ID of the 4GL widget which has the input focus.
    *
    * @param   widget
    *          Event source.
    *          
    * @return  ID of the widget or <code>-1</code> if the widget doesn't have a config.
    */
   public static int lookup4GLWidgetIdAsInt(Widget<?> widget)
   {
      return lookupWidgetIdAsInt(widget, true);
   }

   /**
    * Get the component which is in focus.
    * 
    * @return   <code>Component</code> instance which is in focus or 
    *           <code>null</code> if there is no focused widget.
    */
   public static Widget<?> getCurrentFocus()
   {
      TitledWindow<?> w = ThinClient.getInstance().isChui() ? WindowManager.getDefaultWindow() 
                                                            : WindowManager.getFocusWindow();
      if (w == null)
      {
         w = WindowManager.topLevelWindow();
      }
      
      return (w == null ? null : w.currentFocus());
   }

   /**
    * Checks the given event list for global events.
    *
    * @param  list
    *         event list to check 
    * @return <code>true</code> if the list contains at least one global event
    */
   public static boolean checkEventList(EventList list)
   {
      // TODO: WindowManager.topLevelWindow() doesn't always point to the active window 
      //       (or ACTIVE-WINDOW) in GUI
      
      // go through the event list and check:
      // 1. if there are events for visible windows
      // 2. if there are events for widgets (part of frames) which are not in the ACTIVE-WINDOW
      // if any of this is true, the we assume we are in "global events" mode
      return list.hasWindows() || 
             list.hasWidgetsOutsideWindow(WindowManager.topLevelWindow());
   }

   /**
    * Return the frame which contains the specified widget. If the given widget 
    * is itself a frame, then it will be returned.
    * 
    * @param    widget
    *           Widget for which to find the frame.
    *
    * @return   The frame which contains specified widget, or <code>null</code>
    *           if <code>null</code> is passed as the parameter or the widget 
    *           is not placed in a frame. 
    */
   public static Frame<?> locateFrame(Widget<?> widget)
   {
      if (widget instanceof Frame)
      {
         return (Frame<?>) widget;
      }

      return getParentFrames(widget).stream().findFirst().orElse(null);
   }

   /**
    * Return the root frame which contains the specified widget. If the given 
    * widget is itself a root frame, then it will be returned.
    * 
    * @param    widget
    *           Widget for which to find the frame.
    *
    * @return   The root frame which contains specified widget, or <code>null</code>
    *           if <code>null</code> is passed as the paramete or the widget is
    *           not placed in a frame.
    */
   public static Frame<?> locateRootFrame(Widget<?> widget)
   {
      return getParentFrames(widget).stream()
               .filter(f -> f.isRootFrame())
               .findFirst()
               .orElse(null);
   }
   
   /**
    * Returns all parent frames up the root frame starting from the provided widget.
    * The frames are sorted from the most inner up to the root frame. If no frames 
    * are found, an empty list is returned. If the given widget is itself a frame, 
    * then it will be included in the resulting list.
    *  
    * @param    widget
    *           A widget. If <code>null</code> an empty list is returned.
    *        
    * @return   See above, if no frames are found, returns an empty list.
    *       
    */
   public static List<Frame<?>> getParentFrames(Widget<?> widget)
   {
      if (widget == null)
      {
         return Collections.emptyList();
      }
         
      List<Frame<?>> result = new ArrayList<>();
      while (widget != null)
      {
         if (widget instanceof DropDown)
         {
            DropDown<?> dd = (DropDown<?>) widget;
            widget = dd.getComboBox();
         }
         
         if (widget instanceof Frame)
         {
            result.add((Frame<?>) widget);
         }
         
         widget = widget.parent();
      }
      
      return result;
   }
   
   /**
    * Returns all parent frames up the root frame starting from the provided widget.
    * The frames are sorted from the most inner up to the root frame. If no frames
    * are found, an empty list is returned. If the given widget is itself a frame,
    * then it will be included in the resulting list.
    *
    * @param    widgetId
    *           A widget id.
    *
    * @return   See above.
    */
   public static List<Integer> getParentFrames(int widgetId)
   {
      Widget widget = ThinClient.getInstance().registry().getComponent(widgetId);
      return getParentFrames(widget).stream()
              .map(f -> f.getId().asInt())
              .collect(Collectors.toList());
   }

   /**
    * Return the alert box which contains the specified widget.
    *
    * @param    widget
    *           Widget for which to find the alert box.
    *
    * @return   The alert box which contains the specified widget or
    *           <code>null</code> if the widget is not contained in an
    *           alert box or <code>null</code> is passed as the parameter.
    */
   public static AlertBox<?> locateAlertBox(Widget<?> widget)
   {
      while (widget != null && !(widget instanceof AlertBox))
      {
         widget = widget.parent();
      }

      return (AlertBox<?>) widget;
   }

   /**
    * Returns <code>true</code> if the whole widget is not even partially
    * obscured by any frame from the specified array.
    *
    * @param   comp
    *          widget to test
    *
    * @param   above
    *          array of visible frames to test against; normally obtained by
    *          Screen.getFramesAbove() helper
    * @return  <code>true</code> if it is not obscured
    */ 
   public static boolean isAllShowing(Widget<?> comp, Frame<?>[] above)
   {
      if (above.length == 0) 
      {
         return true;
      }
      
      // get the widget's rectangle expressed in screen coordinates
      // some service widgets like BrowseColumn do not have parents and do not
      // take screen space and have to be excluded from the check
      if (comp.parent() == null)
      {
         return false;
      }
   
      Point origin = comp.screenLocation();
      Rectangle bounds = new Rectangle(origin,
                                       comp.dimension(), 
                                       comp.screen().coordinates().baseUnits());
   
      for (int i = 0; i < above.length; i ++) 
      {
         // frame rectangle is (practically) in screen coordinates
         Rectangle rect = above[i].bounds();
         if (rect.intersection(bounds) != null)
         {
            return false;
         }
      }
      return true; 
   }

   /**
    * Check if widget is focusable.
    * 
    * @param   comp
    *          Widget to check.
    *          
    * @return  <code>true</code> if widget us focusable.
    */
   public static boolean isFocusable(Widget<?> comp)
   {
      return hasConfig(comp) && comp.isFocusable();
   }

   /**
    * Hide all dialog frames, if not in the specified list. 
    * 
    * @param   wnd
    *          Containing window.
    * @param   except
    *          The list of dialogs which should not be hidden.
    */
   @SuppressWarnings("rawtypes")
   public static void hideDialogs(Window wnd, Collection<Frame> except)
   {
      Widget<?>[] c = UiUtils.getTopLevelComponents(wnd);
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Frame     && 
             ((Frame<?>) c[i]).isDialog() &&
             except.contains(c[i]) == false)
         {
            c[i].setVisible(false);
         }
      }
   }

   /**
    * Get IDs of all frames in the current window.
    * 
    * @param   wnd
    *          Window containing frames, current window.
    * 
    * @return  Array of all frame IDs in their Z-order.
    */
   public static int[] getFrameIDs(TopLevelWindow wnd)
   {
      List<Frame<?>> list = wnd.getFrames();
      int[] res = new int[list.size()];
      int count = 0;
   
      // save IDs of all frames
      for (Frame frame : list)
      {
         int wid = getWidgetIdAsInt(frame);
         
         res[count ++] = wid;
      }
   
      // produce a compact array
      int[] ids = new int[count];
      for (int i = 0; i < count; i++)
      {
         ids[i] = res[i];
      }
      
      return ids;
   }

   /**
    * Instantiate component using type information provided in configuration.
    * <p>
    * WARNING: at this time, <code>Window</code> and <code>AlertBox</code>
    * cannot be reconstructed via this mechanism.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.         
    * 
    * @return  Instantiated component or <code>null</code> in case of
    *          failure.
    */
   @SuppressWarnings("unchecked")
   public static Widget<?> reconstructComponent(WidgetId id, WidgetConfig cfg)
   {
      //TODO<SIY>: case with missing widget type is not handled here
      // instead we let factory to fail. Probably this is not convenient
      // because some important information (requested type) is lost and can't 
      //be logged nor passed to thrown exception.
      
      Class<?> cfgCls = cfg.getClass();
      
      Class<?> clazz = null;
      if (cfgCls == SkipConfig.class)
      {
         clazz = ((SkipConfig) cfg).vertical ? Skip.class : Space.class;
      }
      else if (cfgCls == FrameConfig.class)
      {
         clazz = Frame.class;
      }
      else if (cfgCls == MenuConfig.class)
      {
         clazz = Menu.class;
      }
      else if (cfgCls == MenuItemConfig.class)
      {
         clazz = MenuItem.class;
      }
      else if (cfgCls == SubMenuConfig.class)
      {
         clazz = SubMenu.class;
      }
      else if (cfgCls == ButtonConfig.class)
      {
         clazz = Button.class;
      }
      else if (cfgCls == RectangleConfig.class)
      {
         clazz = Rect.class;
      }
      else if (cfgCls == SelectionListConfig.class)
      {
         clazz = SelectionList.class;
      }
      else if (cfgCls == ComboBoxConfig.class)
      {
         clazz = ComboBox.class;
      }
      else if (cfgCls == RadioSetConfig.class)
      {
         clazz = RadioSet.class;
      }
      else if (cfgCls == FillInConfig.class)
      {
         clazz = FillIn.class;
      }
      else if (cfgCls == TextConfig.class)
      {
         clazz = Text.class;
      }
      else if (cfgCls == EditorConfig.class)
      {
         clazz = Editor.class;
      }
      else if (cfgCls == BrowseConfig.class)
      {
         clazz = Browse.class;
      }
      else if (cfgCls == BrowseColumnConfig.class)
      {
         clazz = BrowseColumn.class;
      }
      else if (cfgCls == ToggleBoxConfig.class)
      {
         clazz = ToggleBox.class;
      }
      else if (cfgCls == WindowConfig.class)
      {
         clazz = Window.class;
      }
      else if (cfgCls == ImageConfig.class)
      {
         clazz = Image.class;
      }
      else if (cfgCls == SliderConfig.class)
      {
         clazz = Slider.class;
      }
   
      return OutputManager.instance().getFactory().create(clazz, id, cfg);
   }
   
   /**
    * Get all frames which are visible, overlap given frame and are above it
    * in Z-order.
    * 
    * @param   frame
    *          Reference frame.
    * 
    * @return  Array of found fames.
    */
   public static Frame<?>[] getFramesAbove(Frame<?> frame)
   {
      Widget<?>[] c = (frame.isRootFrame() ? UiUtils.getTopLevelComponents(frame.window())
                                           : frame.parent().widgets());
      ArrayList<Frame<?>> list = new ArrayList<Frame<?>>();
      Rectangle rect = frame.bounds();
   
      // the components list is ordered by Z-order such that the first entry
      // is at the bottom and the last entry is at the top, so search "down"
      // by walking backwards through the array
      for (int i = c.length - 1; i >= 0; i--)
      {
         if (c[i] instanceof Frame)
         {
            Frame<?> frm = (Frame<?>) c[i];
            
            if (frame == frm)
               break;
            
            if (!frm.isVisible() || frame.isRedirected())
               continue;
            
            if (rect.intersects(frm.bounds()))
               list.add(frm);
         }
      }
      
      return list.toArray(new Frame[list.size()]);
   }

   /**
    * Reset <code>needPause</code> flag for all frames which are visible.
    * 
    * @param   wnd
    *          target window pause.
    */
   public static void resetNeedPause(TopLevelWindow<?> wnd)
   {
      for (Frame<?> frame : wnd.getFrames())
      {
         frame.resetNeedPause();
      }
   }

   /**
    * Check if any visible frame has <code>needPause</code> flag set.
    * 
    * @return  <code>true</code> if any of the frames on the screen has
    *          <code>needPause</code> flag set to <code>true</code>.
    */
   public static boolean isNeedPause()
   {
      Widget<?>[] c = UiUtils.getTopLevelComponents(WindowManager.getDefaultWindow());
   
      for (Widget<?> component : c)
      {
         if (component.isVisible() && component instanceof Frame)
         {
            Frame<?> frm = (Frame<?>) component;
            
            if (frm.isNeedPause())
               return true;
         }
      }
      
      return false;
   }
   
   /**
    * Hide frames which are marked as such that can be hidden.
    * <p>
    * Note that if there is an old instance of given frame (not the same
    * frame, but frame with the same name), then method returns false as if
    * there were no frames hidden at all. Logically this is correct because
    * new instance will replace old one at the screen.
    * 
    * @param    current
    *           The frame for which frame hiding is performed.
    * @param    streamed
    *           <code>true</code> hide streamed frames, <code>false</code> -
    *           hide frames from interactive terminal.
    * 
    * @return   Array of boolean results. <br>
    *           First element of array is <code>true</code> if at least one
    *           frame other than given one was hidden and there were no old
    *           instances of given frame.<br>
    *           Second element of array is <code>true</code> if one of hidden
    *           frames has the same row as current frame.
    */
   public static boolean[] hideFrames(Frame<?> current, boolean streamed)
   {
      if (!current.isRootFrame())
      {
         throw new IllegalStateException("Must not call for nested frames!");
      }
      
      boolean[] res = new boolean[] {false, false};

      Window<?> wnd = current.window();
      if (wnd == null)
      {
         return res;
      }
      
      Widget<?>[] c = UiUtils.getTopLevelComponents(wnd);
      boolean prevInstance = false;
   
      ArrayList<Frame<?>> list = new ArrayList<Frame<?>>();
      boolean needPause = false;
      
      for (Widget<?> component : c)
      {
         if (!component.isVisible() || !(component instanceof Frame))
            continue;
         
         Frame<?> frame = (Frame<?>) component;
         
         if (frame.isCanHide() && 
             !frame.isNoHide() && 
             streamed == frame.isStreamed())
         {
            if (frame != current)
            {
               if (frame.isNeedPause())
                  needPause = true;
   
               list.add(frame);
               
               // TODO: the following assumption is incorrect.  frame names are not unique, and
               // the match should be done via their ID.  But switching to ID checking results in 
               // regressions related to cleaning up PUT SCREEN data.  this is left as is for now.
               
               // In some cases it is possible that frame is instantiated more
               // than once. This happens, for example, with frames which have 
               // NO-HIDE option and which are scoped to procedure which is
               // invoked more than once. There is only one way to check that
               // different Frame instances logically represent the same 4GL 
               // frame - compare names. Fortunately frame names are unique.
               if (frame.name() != null && 
                   current.name() != null &&
                   frame.name().equals(current.name()))
               {
                  prevInstance = true;
               }
            }
         }
      }
   
      try
      {
         if (needPause)
            ThinClient.getInstance().pauseBeforeHide();
      }
      
      finally
      {
         double currentY = (current != null) ? current.screenLocation().y : -1;
         
         for (Frame<?> frame : list)
         {
            if (frame.screenLocation().y == currentY)
               res[1] = true;
               
            frame.setVisible(false);
         }
      }
      
      res[0] = !list.isEmpty() && !prevInstance; 
      
      return res;
   }

   /**
    * Convenience method to retrieve top level frames and other components
    * from the default window.
    * 
    * @return  The array of top level components.
    */
   public static Widget<?>[] getTopLevelComponents(Window wnd)
   {
      return wnd.getWidgets();
   }

   /**
    * Get DCOLOR attribute for specified widget taking into account color
    * inheritance from parent to child widgets.
    *
    * @param   widget
    *          Widget.
    *
    * @return  found color.
    */
   public static Color getDColor(Widget<?> widget) {
      return getDColor(getConfig(widget, BaseConfig.class).dcolor, widget);
   }

   /**
    * Get DCOLOR attribute for specified widget taking into account color
    * inheritance from parent to child widgets. Color inheritance is used only
    * if provided color is <code>null</code>.
    *
    * @param   dcolor
    *          Initial color.
    * @param   widget
    *          Widget.
    *
    * @return  found color.
    */
   public static Color getDColor(Color dcolor, Widget<?> widget)
   {
      Color color = dcolor;

      if (color == null)
      {
         Widget<?> parent = locateFrame(widget);

         if (parent == null || !(parent instanceof Frame) || parent.equals(widget))
            return Color.NORMAL;

         return getDColor(getConfig(parent, BaseConfig.class).dcolor, parent);
      }

      return color;
   }

   /**
    * Gets the data array from input stream.
    * 
    * @param    is
    *           The input stream to be read.
    *
    * @return   The image bytes, or {@code null} in case of error.
    */
   public static byte[] getData(InputStream is)
   {
      byte[] imgData = null;

      try
      {
         ByteArrayOutputStream baos = new ByteArrayOutputStream();
         byte[] b = new byte[512];
         while (is.available() > 0)
         {
            int l = is.read(b);
            baos.write(b, 0, l);
         }
         
         imgData = baos.toByteArray();
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not load image from input stream...", e);
         }
         
         imgData = null;
      }
      finally
      {
         try
         {
            is.close();
         }
         catch (IOException e)
         {
            // ignore
         }
      }
         
      return imgData;
   }
   
   /**
    * Check if the specified window can be implicitly shown by the UI statement targeting the
    * given frame.  If none of the window's parent (or self) is hidden and the window is not 
    * realized or not visible, then, if the frame is a root frame:
    * <p>
    * If the frame is hidden, only the {@link UIStatement#UPDATE}, {@link UIStatement#SET}, 
    * {@link UIStatement#PROMPT_FOR} and {@link UIStatement#VIEW} will show the window.
    * <p>
    * If the frame is not hidden, only the {@link UIStatement#DISPLAY}, {@link UIStatement#UPDATE},
    * {@link UIStatement#SET}, {@link UIStatement#PROMPT_FOR}, {@link UIStatement#VIEW} and 
    * {@link UIStatement#ENABLE} statements will show the window.
    * 
    * @param    window
    *           The window which needs to be shown implicitly.
    * @param    frame
    *           The frame which targets this window.
    * @param    stmt
    *           The UI statement working with this frame.
    *           
    * @return   <code>true</code> if the window needs to be shown implicitly.
    */
   public static boolean showWindowForFrame(TopLevelWindow window, Frame frame, UIStatement stmt)
   {
      boolean allFramesVisible = frame.allFramesVisible(true);

      boolean frameWillShow = false;
      
      if (frame.isRootFrame())
      {
         // check if the frame will be shown on the window (and thus the window will need to be 
         // made visible).  The conditions are these, when a window will be show, with a frame stmt:
         // pristine frame (hidden=visible=no): display, update, set, prompt-for, view, enable
         // hidden=no, visible=yes: display, update, set, prompt-for, view, enable
         // hidden=yes, visible=no: update, set, prompt-for, view
         if (frame.hidden())
         {
            frameWillShow = (stmt == UIStatement.UPDATE     ||
                             stmt == UIStatement.SET        ||
                             stmt == UIStatement.PROMPT_FOR ||
                             stmt == UIStatement.VIEW);
         }
         else
         {
            frameWillShow = (stmt == UIStatement.DISPLAY    ||
                             stmt == UIStatement.UPDATE     ||
                             stmt == UIStatement.SET        ||
                             stmt == UIStatement.PROMPT_FOR ||
                             stmt == UIStatement.VIEW       ||
                             stmt == UIStatement.ENABLE);
         }
      }

      return (allFramesVisible || frameWillShow) &&
             window != null                      &&
             !window.parentOrSelfHidden()        &&
             (!window.isRealized() || !window.isVisible());
   }

   /**
    * Get ID of the 4GL widget which has the input focus.
    *
    * @param   widget
    *          Event source.
    * @param   progressOnly
    *          specifies if only 4GL widgets should be considered.

    * @return  ID of the widget or <code>-1</code>
    */
   private static WidgetId lookupWidgetId(Widget<?> widget, boolean progressOnly)
   {
      WidgetId widgetId = null;
      
      while (widget != null && (widgetId == null || 
            (progressOnly && widgetId.asInt() < -1)))
      {
         widgetId = getWidgetId(widget);
         widget = widget.parent();
      }
      
      if (progressOnly && (widgetId == null || widgetId.asInt() < -1))
      {
         widgetId = new WidgetId();
      }
      
      return widgetId;
   }

   /**
    * Get ID of the 4GL widget which has the input focus.
    *
    * @param   widget
    *          Event source.
    * @param   progressOnly
    *          specifies if only 4GL widgets should be considered.

    * @return  ID of the widget or <code>-1</code> if the widget doesn't have a config.
    */
   private static int lookupWidgetIdAsInt(Widget<?> widget, boolean progressOnly)
   {
      WidgetId widgetId = lookupWidgetId(widget, progressOnly);
      
      return widgetId == null ? -1 : widgetId.asInt();
   }
}