/*
** Module   : WidgetRegistry.java
** Abstract :
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20101015 Widget registry code from Screen moved into separate class. 
** 002 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 003 SIY 20111202 Updated imports.
** 004 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 005 EVL 20140812 In new widget ID allocation approach ID for default window is taking from
**                  WidgetId class static constant.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 VIG 20140919 Changed target window to which a frame is added.
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 HC  20141207 Improved config state client-server synchronization.
** 010 CA  20150105 Changes for SIDE-LABEL-HANDLE runtime support.
** 011 CA  20150114 Fixed support for changing the ROW/COLUMN attrs from server-side, for a 
**                  side-label.
** 012 CA  20150202 If the frame is still attached to the window when is being removed, remove it
**                  from its parent window.
** 013 VIG 20150122 Added addWidgetsToMenu method.
** 014 HC  20150323 Removed business logic from config classes.
** 015 VIG 20150228 Added recursive pushMenuDefinition call.
** 016 HC  20150612 Improved widget registration/deregistration.
** 017 CA  20150722 Browse column widgets need to be processed before any other widgets.
**                  Changes to perform layout only once.
** 018 CA  20150728 Fixed reconstruction of side-labels in a down frame.
** 019 VIG 20150724 Removed menu layouting code from the method pushMenuDescription.
** 020 SVL 20150821 Do not disable dynamic widgets when they are added to the frame.
** 021 HC  20150831 Widgets are automatically detached from their parents when destroyed.
** 022 IAS 20150926 TOOLTIP support.
** 023 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 024 CA  20151026 Added support for nested frames.
** 025 CA  20151123 Replaced root frame check with dedicated API.
**     SBI 20151201 Added findAncestor/findDescendant for a given widget and a target widget type.
** 026 IAS 20160211 Thread-safe access to the widgetList.
** 027 SBI 20160307 Changed findAncestor method to accept a target filter predicate.
** 028 HC  20160628 Prevented unhandled exceptions when window/frame closed/destroyed.
** 029 EVL 20160725 Adding the registry for editable frames.
** 030 IAS 20160817 Minor getComponent() method optimization.
** 031 CA  20160915 Fixed frame delete: child static frames must survive.  Added a predicate check
**                  to removeWidgetsRecursive, so that for i.e. windows the child frames will 
**                  not be removed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import java.util.concurrent.locks.*;
import java.util.function.Predicate;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Widget registry.
 * <p>
 * This class holds all widgets (frames and their inner widgets) which were 
 * pushed from server to client. Also, this class is responsible for building
 * client side widgets from widget definition.  
 * 
 * @author 	SIY
 * @version 1.0
 */
public class WidgetRegistry<O extends OutputManager<?>>
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(WidgetRegistry.class);
   
   /** The list of the widgets in the current screen. */
   private LinkedHashMap<WidgetId, Widget<O>> widgetList = new LinkedHashMap<>();
   
   /** We need to maintain the list of currently editable frames. */
   private Map<Frame<O>, Set<WidgetId>> editableFramesList = new HashMap<>();
   
   /** Guard lock for access to the widgetList */
   private final ReadWriteLock lock = new ReentrantReadWriteLock();
   
   /**
    * Returns a nearest widget's ancestor that satisfies the target filter condition.
    * 
    * @param    w
    *           The target widget that can have an ancestor of a given type.
    * @param    filter
    *           The target filter condition.
    * 
    * @return   The ancestor satisfied the target filter condition or null if there are no
    *           widgets satisfied this condition.
    */
   public static Widget findAncestor(Widget w, Predicate<Widget> filter)
   {
      Widget c = w;
      while (c != null)
      {
         if (filter.test(c))
         {
            return c;
         }
         c = c.parent();
      }
      return null;
   }
   
   /**
    * Returns a nearest widget's descendant having a given type.
    * 
    * @param    w
    *           The target widget that can have a descendant of a given type.
    * @param    type
    *           The specified type condition.
    * 
    * @return   The descendant having the specified type or null if there are no widgets
    *           satisfied this condition.
    */
   public static Widget findDescendant(Widget w, Class type)
   {
      if (w.getClass().equals(type))
      {
         return w;
      }
      else if (w instanceof Container)
      {
         int index = 0;
         Widget c; 
         Widget target = null; 
         for ( ;(c = ((Container) w).get(index)) != null && target == null; index++)
         {
            target = findDescendant(c, type);
         }
         return target;
      }
      return null;
   }
   
   /**
    * Constructor. 
    */
   public WidgetRegistry()
   {
   }

   /**
    * Recreates the widget based on the information can be taken from the
    * source widget snapshot. This snapshot must contain all required
    * information different from the default one which is enough to
    * reconstruct the widget of the given type with the desirable behavior.
    * 
    * @param   config
    *          Widget snapshot taken from the server.
    */
   @SuppressWarnings({ "unchecked" })
   public void reconstructWidget(WidgetConfig config)
   {
      ConfigManager mgr = ConfigManager.getInstance();

      WidgetId cid = config.id;
      Widget<O> comp = getComponent(cid);
      
      // the ID is no longer required - any change made to this config instance must not affect 
      // the associated widget!
      config.id = null;
      
      if (comp != null)
      {
         // existing component, just apply new config values
         Widget<O> frame = (Widget<O>) UiUtils.locateFrame(comp);
         
         if (frame instanceof Frame)
         {
            ((Frame<O>) frame).applyConfig(comp, config);
         }
         else
         {
            mgr.replaceConfig(comp.config(), config);
         }
      }
      else
      {
         // new component
         comp = (Widget<O>) UiUtils.reconstructComponent(cid, config);

         if (comp != null)
         {
            addWidget(cid, comp);
            
            if (comp instanceof WidgetWithConfig)
            {
               ((WidgetWithConfig<WidgetConfig>) comp).initialize(cid, config);
               ThinClient.setupTooltip((WidgetWithConfig<WidgetConfig>) comp);
            }

            WidgetConfig wcfg = comp.config();
            
            // Save the base state, during initialization the widget/config state may change.
            // Without the proper base state the changes would not be sent to the server.
            mgr.addWidgetConfig(wcfg, config);
         }
      }
      
      // restore the id
      config.id = cid;
   }

   /**
    * Add pre-built frame.
    * 
    * @param   frame
    *          Frame reference.
    */
   public void addFrame(Frame<O> frame)
   {
      if (frame == null)
         return;
   
      Widget<O>[] list = frame.getContentPane().widgets();
      
      addWidget(frame.config().id, frame);
   
      for (Widget<O> comp : list)
      {
         if (UiUtils.hasConfig(comp))
         {
            addWidget(comp.getId(), comp);
         }
      }
   }

   /**
    * Get all components defined in the screen.
    * 
    * @return  Array of component.
    */
   @SuppressWarnings("unchecked")
   public Widget<O>[] getComponents()
   {
      lock.readLock().lock();
      try
      {
         return widgetList.values().toArray(new Widget[widgetList.size()]);
      }
      finally
      {
         lock.readLock().unlock();
      }
   }

   /**
    * Remove the given frame and all contained components from the master
    * list of components.
    * 
    * @param    frame
    *           The frame to remove.
    */
   public void removeFrame(Frame<O> frame)
   {
      if (getComponent(frame.config().id) != null)
      {
         // remove the frame from its parent window first
         TitledWindow<O> pwin = frame.ancestor();
         if (pwin == null)
         {
            LOG.log(Level.WARNING, 
                    "Removing frame " + frame.toString() + " with no parent window.");
         }
         Window<O> defw = WindowManager.getDefaultWindow();

         WidgetId base = frame.config().id;
         
         // and remove all contained widgets
         Widget<O>[] w = null;
         if (frame.getContentPane() != null)
         {
            w = frame.getContentPane().widgets();
         }

         // determine widget numbers to remove and protect from NullPointerException
         int wSize = (w != null) ? w.length : 0;

         // the down body is destroyed automatically as the widgets are added to the content pane
         
         for (int i = 0; i < wSize; i++)
         {
            Widget<O> child = w[i];
            WidgetConfig childCfg = child.config();
            
            if (child instanceof Frame && !((FrameConfig) childCfg).dynamic)
            {
               // all static frames must survive
               frame.getContentPane().detach(child);
               child.unrealize();
               frame.config().frameId = childCfg.id.asInt();

               defw.addFrame(child, child.location().y, child.location().x);
               continue;
            }
            
            WidgetId id = UiUtils.getWidgetId(child);
            if (id != null)
            {
               Widget<?> comp = removeWidget(id);
               
               if (comp != null)
               {
                  comp.destroy();
               }
            }
         }

         // remove the frame itself
         removeWidget(base);
         
         // refresh editable frames list cache too
         editableFramesList.remove(frame);
         
         // destroy will detach the frame from its parent
         frame.destroy();
      }
   }

   /**
    * Possibly add the frame with given widget to the editable frames list if it is not already
    * there.
    * 
    * @param    widToCheck
    *           The widget to check.
    */
   public void addEditableFrameForWidget(Widget widToCheck)
   {
      // wirtual widget will not be considered
      if (WidgetId.virtualWidget(widToCheck.getId()))
      {
         return;
      }
      
      // adding the frame for given widget if it is not already in cache
      Frame<O> widFrame = (Frame<O>) UiUtils.locateFrame(widToCheck);
      if (widFrame != null)
      {
         // get ID set if the frame already in the list
         Set<WidgetId> widIdSet = editableFramesList.get(widFrame);
         
         // new frame to add
         if (widIdSet == null)
         {
            widIdSet = new HashSet<WidgetId>();
            editableFramesList.put(widFrame, widIdSet);
         }
         
         // and add the widget ID to the set for respective frame
         widIdSet.add(widToCheck.getId());
      }
   }

   /**
    * Possibly remove the frame with given widget from the editable frames list if there is no
    * more enabled focused widgets in a frame left.
    * 
    * @param    widToCheck
    *           The widget to check.
    */
   public void removeEditableFrameForWidget(Widget widToCheck)
   {
      // wirtual widget will not be considered
      if (WidgetId.virtualWidget(widToCheck.getId()))
      {
         return;
      }
      
      // removing the frame for given widget if it is in cache
      Frame<O> widFrame = (Frame<O>) UiUtils.locateFrame(widToCheck);
      if (widFrame != null)
      {
         // check if the frame is in the list
         Set<WidgetId> widIdSet = editableFramesList.get(widFrame);
         
         // if not - do nothing
         if (widIdSet != null)
         {
            // frame in the list, removing current widget ID from it's set
            widIdSet.remove(widToCheck.getId());
            // check if this was the last widget in a frame
            if (widIdSet.isEmpty())
            {
               // remove the frame from the list
               editableFramesList.remove(widFrame);
            }
         }
      }
   }

   /**
    * Get all currently existed editable frames in the screen.
    * 
    * @return  Array of the editable frames.
    */
   public Frame<O>[] getEditableFrames()
   {
      return editableFramesList.keySet().toArray(new Frame[editableFramesList.size()]);
   }

   /**
    * Return <code>Component</code> instance for the given ID.
    * 
    * @param   id
    *          Component ID.
    * @return  Reference to the <code>Component</code> or <code>null</code>
    *          if no such component exists.
    */
   @SuppressWarnings("unchecked")
   public Widget<O> getComponent(WidgetId id)
   {
      if (id == null)
      {
         return null;
      }

      if (id.equals(WidgetId.DEFAULT_WINDOW_ID))
         return (Widget<O>) WindowManager.getDefaultWindow();

      Widget<O> w = null;
      lock.readLock().lock();
      try
      {
         w = widgetList.get(id);
      }
      finally
      {
         lock.readLock().unlock();
      }
      
      if (w == null)
      {
         w = WindowManager.findWindow(id.asInt());
      }

      return w;
   }

   /**
    * Return <code>Component</code> instance for the given ID.
    * <p>
    * If the widget is part of a frame and the frame's down body is initialized, then return the 
    * widget associated with the current frame line.
    * 
    * @param   id
    *          Component ID.
    * @return  Reference to the <code>Component</code> or <code>null</code>
    *          if no such component exists.
    */
   @SuppressWarnings("unchecked")
   public Widget<O> getComponent(int id)
   {
      if (id == WidgetId.DEFAULT_WINDOW_ID.asInt())
         return (Widget<O>) WindowManager.getDefaultWindow();
      
      Widget<O> w = null;

      if (OutputManager.inBatchMode())
      {
         w = getComponent(new WidgetId(id));
      }
      else 
      {
         int frameID = lookupFrameIdFromWidgetId(id);
         if (frameID != id && frameID != -1)
         {
            Frame<?> frame = (Frame<?>) getComponent(frameID);
            WidgetId wid = Frame.resolveWidgetId(frame, id);
            Widget<O> widget = getComponent(wid);

            // search by ID only if not found... 
            // TODO: confirm this doesn't break anything else!
            w = (widget == null ? getComponent(new WidgetId(id)) : widget);
         }
         else
         {
            w = getComponent(new WidgetId(id));
         }
      }

      if (w == null)
      {
         w = WindowManager.findWindow(id);
      }

      return w;
   }

   /**
    * Retrieve the ID of the frame containing the widget with a given ID.
    *
    * @param    widgetId
    *           The widget ID.
    *
    * @return   The frame ID if widget is found or -1 if not.
    */
   public int lookupFrameIdFromWidgetId(int widgetId)
   {
      int resultId = -1;
      
      Widget<?> wid = getComponent(new WidgetId(widgetId));
      
      if (wid != null)
      {
         Frame<?> widFrame = (Frame<?>) UiUtils.locateFrame(wid);
         
         if (widFrame != null)
         {
            resultId = UiUtils.lookupWidgetIdAsInt(widFrame);
         }
      }
      
      return resultId;
   }

   /**
    * Instantiate screen from the given array of definitions.
    * 
    * @param   sd
    *          Array of data containers.
    */
   public void pushDefinition(ScreenDefinition[] sd)
   {
      for (int i = 0; i < sd.length; i++)
      {
         ScreenDefinition screenDefinition = sd[i];
   
         int frameId = screenDefinition.getRootFrameID();
         Frame<O> frame = (Frame<O>) getComponent(frameId);
         boolean newFrame = (frame == null);

         // if frame already exists, notify it that processing begins
         if (!newFrame)
         {
            frame.preprocess();
         }
         
         // process side-labels first. this requires only update into the client-side config 
         // repository. no Label widget is actually created.
         Iterator<WidgetConfig> itr = screenDefinition.getSideLabelsIterator();
         while (itr.hasNext())
         {
            WidgetConfig cfg = itr.next();
            
            reconstructSideLabel(cfg);
         }

         // process browse columns first
         itr = screenDefinition.getConfigsIterator();
         while (itr.hasNext())
         {
            WidgetConfig cfg = itr.next();
            
            if (cfg instanceof BrowseColumnConfig)
            {
               reconstructWidget(cfg);
            }
         }
         
         // go through the actual widgets
         itr = screenDefinition.getConfigsIterator();
         while (itr.hasNext())
         {
            WidgetConfig cfg = itr.next();
            
            if (cfg instanceof BrowseColumnConfig)
            {
               continue;
            }
            
            reconstructWidget(cfg);
         }
         
         if (newFrame)
         {
            frame = (Frame<O>) getComponent(screenDefinition.getRootFrameID());
         }
         
         if (frame != null)
         {
            int frameID = screenDefinition.getRootFrameID();
            Frame<O> frameAfterProcessing = (Frame<O>) getComponent(frameID);
            
            // the frame may have been destroyed/unregistered when applying configs
            // above
            if (frameAfterProcessing != null)
            {
               addWidgetsToFrame(frameID, screenDefinition);
            }
            
            frame.postprocess(newFrame, screenDefinition.isSuppressRedraw());
         }
      }
   }

   /**
    * Instantiate menus from the given array of descriptions.
    * 
    * @param   md
    *          Array of data containers.
    */
   public void pushMenuDescription(MenuDescription[] md)
   {
      for (int i = 0; i < md.length; i++)
      {
         MenuDescription menuDescr = md[i];
   
         int menuId = menuDescr.getRootMenuID();
         Widget<O> menu = getComponent(menuId);
         
         // go through the actual widgets
         Iterator<WidgetConfig> itr = menuDescr.getConfigsIterator();
         while (itr.hasNext())
         {
            WidgetConfig cfg = itr.next();
            
            reconstructWidget((WidgetConfig) cfg);
         }
         
         if (menu != null)
         {
            addWidgetsToMenu(menuId, menuDescr);
         }
         
         // recursively instantiate children
         pushMenuDescription(menuDescr.getChildDescrArray());
      }
   }
   
   /**
    * Get IDs of all widgets in the current screen belonging to the given frame.
    * 
    * @param   frameId
    *          Frame ID whose widget ID info should be returned.
    * 
    * @return  Array of all IDs.
    */
   public WidgetId[] getIDs(int frameId)
   {
      // check if the given frame is in registry
      Frame<O> frame = (Frame<O>) getComponent(frameId);
      
      if (frame == null)
      {
         return null;
      }
      
      // get the list of widgets from frame
      Widget<O>[] list = frame.getContentPane().widgets();

      // determine widget numbers to remove and protect from NullPointerException
      int wSize = (list != null) ? list.length : 0;
      
      if (wSize == 0)
      {
         return null;
      }
      
      // enumerate all widgets and get ID
      WidgetId[] res = new WidgetId[wSize];
      
      for (int i = 0; i < wSize; i++)
      {
         res[i] = UiUtils.getWidgetId(list[i]);
      }
      
      return res;
   }

   /**
    * Add the specified widget to the registry.
    * 
    * @param    widget
    *           The new widget which needs to be registered.
    */
   public void addWidget(Widget<O> widget)
   {
      addWidget(widget.getId(), widget);
   }

   /**
    * Add the widget with the specified ID to the registry.
    * 
    * @param    id
    *           The widget ID.
    * @param    widget
    *           The new widget which needs to be registered.
    */
   public void addWidget(WidgetId id, Widget<O> widget)
   {
      lock.writeLock().lock();
      try
      {
         this.widgetList.put(id, widget);
      }
      finally
      {
         lock.writeLock().unlock();
      }
   }
   
   /**
    * Remove the specified widget from the registry, immediately.
    * 
    * @param    id
    *           The widget ID.
    */
   public Widget<?> removeWidget(WidgetId id)
   {
      lock.writeLock().lock();
      try
      {
         return this.widgetList.remove(id);
      }
      finally
      {
         lock.writeLock().unlock();
      }
   }

   /**
    * Add the specified widget and all its children recursively to the
    * registry, immediately.
    * 
    * @param    widget
    *           The root widget.
    */
   public void addWidgetsRecursive(Widget<O> widget)
   {
      addWidgetsRecursive(Collections.singletonList(widget));
   }

   /**
    * Add the specified widgets and all the children recursively to the
    * registry, immediately.
    * 
    * @param    widgets
    *           The root widgets.
    */
   public void addWidgetsRecursive(Widget<O>[] widgets)
   {
      if (widgets == null)
         return;
      
      addWidgetsRecursive(Arrays.asList(widgets));
   }
   
   /**
    * Add the specified widgets and all the children recursively to the
    * registry, immediately.
    * 
    * @param    widgets
    *           The root widgets.
    */
   public void addWidgetsRecursive(Iterable<Widget<O>> widgets)
   {
      if (widgets == null)
         return;
      
      for (Widget<O> widget : widgets)
      {
         WidgetId id = widget.getId();
         
         if (id != null)
            addWidget(id, widget);
   
         if (widget instanceof Container)
         {
            Widget<O>[] children = ((Container<O>) widget).widgets();
            addWidgetsRecursive(Arrays.asList(children));
         }
      }
   }
   
   /**
    * Remove the specified widgets and all the children recursively from the
    * registry, immediately.
    * 
    * @param    widgets
    *           The root widgets.
    */
   public void removeWidgetsRecursive(Widget<O>[] widgets)
   {
      if (widgets == null)
         return;
      
      removeWidgetsRecursive(Arrays.asList(widgets));
   }
   
   /**
    * Remove the specified widget and all its children recursively from
    * the registry, immediately.
    * 
    * @param    widget
    *           The root widget.
    */
   public void removeWidgetsRecursive(Widget<O> widget)
   {
      removeWidgetsRecursive(widget, null);
   }
   
   /**
    * Remove the specified widget and all its children recursively from the registry, immediately.
    * <p>
    * The widgets are filtered and removed based on the specified predicate.
    * 
    * @param    widget
    *           The root widget.
    * @param    check
    *           Condition to be satisfied by a widget, so that it will be removed.
    */
   public void removeWidgetsRecursive(Widget<O> widget, Predicate<Widget<O>> check)
   {
      if (check != null && !check.test(widget))
      {
         return;
      }

      WidgetId id = widget.getId();
      if (id != null)
      {
         removeWidget(id);
      }

      if (widget instanceof Container)
      {
         Widget<O>[] widgets = ((Container<O>) widget).widgets();

         for (Widget<O> w : widgets)
         {
            removeWidgetsRecursive(w, check);
         }
      }
   }
   
   /**
    * Remove the specified widgets and all the children recursively from the
    * registry, immediately.
    * 
    * @param    widgets
    *           The root widgets.
    */
   public void removeWidgetsRecursive(Iterable<Widget<O>> widgets)
   {
      if (widgets == null)
         return;
      
      for (Widget<O> widget : widgets)
      {
         if (widget instanceof Container)
         {
            Widget<O>[] children = ((Container<O>) widget).widgets();
            removeWidgetsRecursive(Arrays.asList(children));
         }

         WidgetId id = widget.getId();
         
         if (id != null)
            removeWidget(id);
      }
   }
   
   /**
    * Add widgets to the frame which is identified by the given frame ID.
    * 
    * @param    frameID
    *           The index of the frame to be populated.
    * @param    screen
    *           List of widget configurations.
    */
   private void addWidgetsToFrame(int frameID, ScreenDefinition screen)
   {
      ArrayList<Integer> widgetIndexes = screen.getWidgetsInFrame(frameID);
      Frame<O> frame = (Frame<O>) getComponent(frameID);
   
      for (int i = 0; i < widgetIndexes.size(); i++)
      {
         Integer wid = widgetIndexes.get(i);
         int widgetIndex = wid.intValue();
         
         WidgetConfig config = screen.getConfig(widgetIndex);
         Widget<O> nextWidget = getComponent(widgetIndex);
         
         // associate browse columns with their browse widget which is a
         // container (of sorts) itself
         if (config instanceof BrowseColumnConfig)
         {
            Browse<O> browse = (Browse<O>) getComponent(((BrowseColumnConfig) config).browseId);
            
            if (browse != null)
               ((BrowseColumn<O>) nextWidget).setBrowse(browse);
   
            continue;
         }
         
         // determine if the frame's parent has changed. if so, attach it
         if (config instanceof FrameConfig && !((Frame<O>) nextWidget).isRootFrame())
         {
            Container<O> parent = (Container<O>) UiUtils.locateFrame(nextWidget.parent());
            if (parent == null)
            {
               // not attached to a frame, get its direct parent
               parent = nextWidget.parent();
            }
            
            if (parent != frame)
            {
               if (parent != null)
               {
                  nextWidget.parent().detach(nextWidget);
                  
                  parent.repaint();
               }
               
               frame.addWidget(nextWidget);
            }
         }

         // check if this widget has already had the frame assigned as the
         // parent, if so then we MUST NOT add it to the widget list again
         // since it is already there (the Container._addComponent() method
         // sets the parent reference after unconditionally adding the 
         // component to the components vector
         if (nextWidget.parent() != null)
            continue;
   
         // make widget disabled by default unless it is dynamic
         if (nextWidget.isEnabled())
         {
            WidgetConfig cfg = nextWidget.config();
            if (!(cfg != null &&
                  cfg instanceof BaseConfig &&
                  ((BaseConfig) cfg).dynamic))
            {
               nextWidget.setEnabled(false);
            }
         }
   
         frame.addWidget(nextWidget);
      }
   }
   
   /**
    * Add widgets to the menu which is identified by the given menu ID.
    * 
    * @param    menuID
    *           The index of the menu to be populated.
    * @param    descr
    *           List of widget configurations.
    */
   private void addWidgetsToMenu(int menuID, MenuDescription descr)
   {
      ArrayList<Integer> widgetIndexes = descr.getWidgetsInMenu(menuID);
      AbstractContainer<O> menu = (AbstractContainer<O>) getComponent(menuID);
   
      for (int i = 0; i < widgetIndexes.size(); i++)
      {
         Integer wid = widgetIndexes.get(i);
         int widgetIndex = wid.intValue();
         
         Widget<O> nextWidget = getComponent(widgetIndex);
         
         // check if this widget has already had the menu assigned as the
         // parent, if so then we MUST NOT add it to the widget list again
         // since it is already there (the Container._addComponent() method
         // sets the parent reference after unconditionally adding the 
         // component to the components vector
         if (nextWidget.parent() != null)
            continue;
   
         menu.add(nextWidget);
      }
   }
   
   /**
    * Reconstruct the side-label with the specified config.
    * <p>
    * This will not create a new {@link Label} instance.  It will only register the widget 
    * configuration with the {@link ConfigManager}, so that changes to the label instance can be
    * tracked and sent back to the server.
    * <p>
    * Note that the registered widget configuration must be "linked" to the related label widget.
    * This is because the widget takes care of unregistering the label configuration. The link 
    * is established in {@link Label#linkTo(int, int)}. 
    * 
    * @param   config
    *          Widget snapshot taken from the server.
    */
   private void reconstructSideLabel(WidgetConfig config)
   {
      ConfigManager mgr = ConfigManager.getInstance();

      WidgetId cid = config.id;
      WidgetId saveID = cid;
      
      // the ID is no longer required - any change made to this config instance must not affect 
      // the associated widget!
      config.id = null;

      boolean isNew = mgr.getActiveConfig(cid) == null;
      
      if (!isNew)
      {
         // this may be a side-label in a down body - determine the widget to which it belongs,
         // and from there the row
         Label<?> lbl = (Label<?>) getComponent(cid);
         Widget<?> sideWidget = lbl.getSideWidget();
         if (sideWidget != null && sideWidget instanceof LabeledWidget)
         {
            // determine the frame
            Frame<?> frame = (Frame<?>) UiUtils.locateFrame(sideWidget);
            if (frame != null && frame.isSideLabels())
            {
               sideWidget = frame.getField((Widget) sideWidget);
               if (sideWidget instanceof LabeledWidget)
               {
                  lbl = ((LabeledWidget) sideWidget).getLabelInstance();
                  cid = lbl.getId();
               }
            }
         }
      }
      
      WidgetConfig active = (BaseConfig) mgr.resolveWidgetConfig(cid, config);
      mgr.replaceConfig(active, config);
      
      if (isNew)
      {
         mgr.addWidgetConfig(active);
      }
      
      // restore the id
      config.id = saveID;
   }
}