/* 
** Module   : Window.java
** Abstract : Window widget implementation for CHUI
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 EVL 20050509   @21167 Creation based on CHARVA Window class.
**                           Necessary modification to get compiled.
** 002 EVL 20050514   @21174 Uncommenting the code to set up the frame
**                           location while adding the frame to the
**                           window.
** 003 EVL 20050516   @21181 Adding default Progress termination code on
**                           F4.
** 004 EVL 20050518   @21225 Adding the insert mode flag and code to set
**                           and to get the current state.
** 005 EVL 20050524   @21281 Adding VALIDATE() Progress like method.
** 006 EVL 20050524   @21282 Changing the inner variable names to sync
**                           with the rest of the project.
** 007 EVL 20050525   @21283 Adding the attributes and methods to work
**                           with them.
** 008 EVL 20050602   @21375 Fixing javadoc '.' related issue.
** 009 EVL 20050607   @21435 Syntax for comments checking.
** 010 EVL 20050607   @21438 Removing the visible member. It is completely
**                           duplicated in CHARVA's Component superclass
**                           so we do not need to double this work. Also
**                           the methods to work with the HIDDEN attribute
**                           are modified to reflect this change.
** 011 EVL 20050616   @21509 Adding function to restore the status line to
**                           the default state.
** 012 EVL 20050615   @21511 Adding code to get widget's TAB position
**                           inside the window.
** 013 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 014 SVG 20050828   @22286 Added session attribute data entry return.
** 015 SIY 20050831   @22555 Refactoring (see @@22394-22409)
** 016 SIY 20050907   @22654 Fixed use of @inheritDoc tag in javadoc
** 017 NVS 20050914   @22703 Disabled F4 key processing - hardcoded exit
**                           has to be replaced with a flexible handler.
** 018 EVL 20050916   @22733 Workaround for the problem of existing the UI
**                           without any focusable widgets. Adding the
**                           method to remove the frame from window.
** 019 EVL 20050916   @22747 The more fix for #018. The window always has
**                           at least 3 widgets(non-focusable).
** 020 NVS 20050919   @22754 Added isRunning() method and made constructor
**                           private.
** 021 SIY 20051223   @23744 Updated message() to support null argument.
**                           Fixed isRunning() method for updated CHARVA.
** 022 GES 20060116   @23925 Added a method to check if message text
**                           exists.
** 023 SIY 20060128   @24124 Removed unused methods.
** 024 SIY 20060201   @24212 Deep refactoring.
** 025 SIY 20060209   @24439 Minor cleanups.
** 026 GES 20060213   @24589 Added suspend/resume of normal drawing mode.
** 027 GES 20060223   @24681 Added external control over message/status
**                           control visibility.
** 028 GES 20060224   @24729 Added visibility control over temp text
**                           components.
** 029 GES 20060225   @24735 Save/restore visible state when switching
**                           between interactive and redirected output
**                           modes. In redirected mode, everything on
**                           the interactive screen must be set hidden
**                           so that it doesn't draw to the stream.
** 030 NVS 20060608   @27014 Keeping ID of the session window intact.
** 031 NVS 20060630   @27727 requestFocus() call does not belong to draw()
**                           method but kept there until a permanent fix
**                           is implemented.
** 032 SIY 20060704   @27837 Support for null in message text.
** 033 GES 20060710   @27941 Status line text must be limited to a maximum
**                           of 63 characters.
** 034 NVS 20060713   @28029 draw() sets initial clipping rectangle to
**                           the invalidated rectangle. It also draws the
**                           cursor and does NOT change focus anymore.
** 035 NVS 20060728   @28216 Added new methods getMessageLineRectangle()
**                           and getStatusLineRectangle().
** 036 NVS 20060801   @28285 setMessageText() shouldn't use setLength(0)
**                           before setText(), but constructor should.
**                           status line should sync after direct draw.
** 037 NVS 20060802   @28333 Redraws the status line directly, but defers
**                           syncing till the return from a client export.
**                           No redrawing of insert mode label with every
**                           change of the status line.
** 038 NVS 20060802   @28358 Defers syncing of message lines till the
**                           return from a client export.
** 039 GES 20060803   @28420 Major cleanup and refactoring.
** 040 NVS 20060822   @28734 Window drawing now sets the refresh and
**                           resets the display drawing control flags to
**                           reflect the fact draw() methods are called to
**                           refresh the screen.
** 041 SIY 20060824   @28863 Removed getP2JType() method.
** 042 NVS 20061017   @30469 Implemented ability to restart the session
**                           without exiting the process.
** 043 GES 20061120   @31333 Removed dead code.
** 044 SIY 20061128   @31430 Preserve Z-order between frames and message
**                           lines.
** 045 SIY 20061130   @31501 Cleaned up debug output.
** 046 SIY 20061201   @31527 Refinement for #044: do not manage Z-ordering
**                           when message lines are not empty. It seems is
**                           similar to Progress behavior.
** 047 SIY 20061207   @31610 Added getDefaultFocus() method.
** 048 SIY 20061208   @31627 Added getStatusText() method.
** 049 SIY 20061220   @31722 Cleaned up javadoc and removed unused method.
** 050 NVS 20070404   @32762 Added getMessageColor() and refreshMessages()
**                           methods.
** 051 SIY 20070803   @34809 Number of reserved lines now is a symbolic
**                           constant.
** 052 SIY 20070904   @34939 Fixed status line handling.
** 053 NVS 20070925   @35219 restoreComponentVisibility() and
**                           saveComponentVisibility() are replaced with
**                           showRedirectedFrames() and
**                           hideRedirectedFrames() methods respectively.
** 054 SVL 20071127   @36145 line with color = -1 may be passed to Label
**                           widget. Ref: #36030.
** 055 SIY 20071203   @36146 Rolled back #054.
** 056 SIY 20080109   @36741 Set invalidated rectangle before repainting
**                           frames overridden by message lines.
**                           Ref: #36739
** 057 SIY 20080131   @36938 Prevent cleaning up message area if messages
**                           already set to null. Also, do not paint in
**                           area occupied by second message row if it
**                           is null. This prevents early hiding of
**                           TempText in the message area.
**                           Ref: #36936, #36937
** 058 CA  20080226   @37198 Added "redirected" parameter to method
**                           findFrames: return only the frames with the
**                           same redirection status as the parameter.
**                           Ref: #37196
** 059 EVL 20080303   @27232 Adding the static method to force repainting
**                           of the insert mode to be able to repaint the
**                           indicator even when the insert mode is not
**                           really changing. However there is a one
**                           exception
** 060 NVS 20080320   @37612 No drawing of mode indicator for a disabled
**                           widget in focus.
** 061 NVS 20080401   @37785 Added isFocusable() method implementation.
** 062 NVS 20080826   @39597 Modified draw() method to use invalidation
**                           region technique instead of invalidation
**                           rectangle. The method iterates the region and
**                           draws for every rectangle as a clipping
**                           rectangle. This approach eliminates unwanted
**                           overpainting of the unchanged areas.
** 063 NVS 20080828   @39623 Status line is now instance of P2JStatusLabel
**                           which is allowed to repaint().
** 064 SVL 20080914   @39800 Insert mode reset for non-P2JComponent
**                           widgets removed from the draw().
** 065 SIY 20081022   @40197 Use regular drawing mechanism for repainting
**                           of message area.
**                           Ref: #40196
** 066 SIY 20081023   @40203 Rolled back #065 and made message drawing
**                           completely self-contained (no more
**                           dependencies on external bracketing and/or
**                           Toolkit.sync() calls).
** 067 CA  20081114   @40449 Changes due to new clipping implementation.
** 068 NVS 20081205   @40808 Standardized the use of getAndResetBitmap().
** 069 SIY 20081224   @40988 Rolled back #044, #046 (replaced with ThinClient.java #513).
** 070 NVS 20090213   @41317 Dead code removal.
** 071 SIY 20090226   @41380 Use null instead of empty string for insert
**                           mode indicator text in overpaint mode.
** 072 SVL 20090320   @41660 Use null instead of empty string for insert
**                           mode indicator text in off mode.
** 073 SVL 20090424   @41924 Do not reset insert mode on reset of client-side state.
** 074 SIY 20090428   @42016 Restore frame state when switching to interactive mode.
** 075 SVL 20090515   @42186 Added insertInAmbiguousState flag.
** 076 GES 20090702   @43029 Prevent event brackets from getting out of balance.
** 077 CA  20091111   @44330 Changes to allow non-trimmed status messages.
** 078 SIY 20091218   @44519 Added support method for special frame placement.
** 079 SIY 20100619          Switching to own implementation of character UI.
** 080 SIY 20100804          Changes in imports, some refactorings.
** 081 SIY 20100822          Refactored AttributeContainer support.
** 082 SIY 20100915          Changes related to renamed classes.
** 083 SIY 20110112          Changes related to new color handling scheme.
** 084 GES 20111012          Moved to context-local instead of singleton. This
**                           allows multiple client sessions to be hosted in
**                           the server.
** 085 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 086 SIY 20111202          Refactored and moved from CHUI package.
** 087 SIY 20111221          Switched to widget factory. 
** 088 MAG 20140515          Modifications to support multiple windows (GUI).
** 089 CA  20140711          Enhanced font support; refactoring related to GUI support. 
** 090 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
**                           Refactoring to allow multi-window support and GUI windows.
** 091 VIG 20140828          Added currentWindow, changed resolveWindow methods.
** 092 EVL 20140812          In new widget ID allocation approach ID for default window is taking
**                           from WidgetId class static constant.
** 093 HC  20140911          GUI coordinate support - data types refactoring.
** 094 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 095 MAG 20141009          Moved insert indicator widget to WindowChuiImpl.
** 096 VIG 20140911          Added multi-window status* and message* methods support. 
**                           Some of them are moved from ThinClient.
** 097 CA  20141029          Changes related to keyboard refactoring, to allow windows/linux 
**                           compatibility.
** 098 HC  20141021          Implemented window move/position support.
** 099 HC  20141118          Added screenPhysicalLocation().
** 100 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 101 CA  20141123          When a new window is registered, the top window is preserved (it is 
**                           added just below the current top).  Also, added code to properly set 
**                           the default dimension for this window. 
** 102 EVL 20141126          Content pane GUI fix.
** 103 CA  20141204          Fixed a problem related to client restart in ChUI.
** 104 CA  20141206          Exposed getInsMode to sub-classes.
** 105 EVL 20150114          Adding API to set/load window icon.
** 106 HC  20150214          Fixed message-update layout in case the message is too long.
** 107 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 108 HC  20150323          Removed business logic from config classes.
** 109 CA  20150406          Fixed problems related to frame layout in GUI.
** 110 VIG 20150406          Typified getDefaultWindow method.
** 111 HC  20150612          Implemented support for modal windows.
** 112 HC  20150702          Improved parent-child relationship of top-level windows.
** 113 CA  20150730          The FramePlacementManager is window-specific.
** 114 VIG 20150618          Added menubar field.
** 115 VIG 20150813          Make MENUBAR visible only after warning is shown 
**                           (if no enabled frames are available).
** 116 HC  20150831          Widgets are automatically detached from their parents when destroyed.
** 117 CA  20150921          Unconditional call to drawCaret().
** 118 CA  20150929          Fixed multi-window drawing issues caused by concurrent drawing
**                           access to the GuiDriver: all access to the driver is done
**                           exclusively, each thread taking and releasing ownership for the
**                           duration of the drawing operation or other GUI API invocation which
**                           requires access to the underlying physical window. Changes to allow
**                           multi-window focus management in GUI.  Fixed ACTIVE-WINDOW system
**                           handle processing.
** 119 GES 20151001          Reworked tinyInput() to support GUI. Removed some dead code.
** 120 HC  20151013          Changes to extend GUI multi-window focus management and ACTIVE-
**                           WINDOW system handle processing to modal windows.
** 121 EVL 20151022          Fix for ability to use unknown value as initial for tinyInput in
**                           MESSAGE UPDATE statements.  This can be use for evety data type
**                           except date, datetime and datetime-tz.
**     HC  20151027          Implemented WINDOW MESSAGE-AREA and STATUS-AREA attributes.
**     CA  20151106          A fix for HIDDEN attribute: distinguish between HIDE statement and 
**                           setting the HIDDEN attribute to true.
**     HC  20151112          Implemented THREE-D runtime support.
** 122 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 123 HC  20160418          Do not copy the initial value to tiny input when MESSAGE SET.
** 124 CA  20160627          In case of MESSAGE statement, reactivate the batch mode, if it was
**                           deactivated.
** 125 EVL 20160726          Changing to use screen buffer array for editable frames instead of
**                           single current one.
** 126 HC  20160922          Changed FramePlacementManager to operate on the level of
**                           TopLevelWindow.
** 127 VIG 20160819          Fixed NullPointerException in isMessageTextNull().
** 128 HC  20170201          Fixed MENUBAR made visible or hidden when it is attached or detached
**                           from its window.
** 129 OM  20170320          Replaced anonymous class with lambda.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.FramePlacementManager;
import com.goldencode.p2j.ui.chui.ThinClient; // TODO: remove this when TC is UI independent.
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * A Window is a single CHUI workspace for frames and dialogs.
 * TODO A substantial amount of the insert mode processing may eventually move to the
 * ChUI-specific classes, but remains there until the editing widgets and formatting classes have
 * been reworked.
 */
public abstract class Window<O extends OutputManager<?>>
extends TopLevelWindow<O>
implements WidgetWithConfig<WindowConfig>
{
   /** Help status type. */
   public static final int ST_HELP    = 2;
   
   /** Default status type */
   public static final int ST_DEFAULT = 0;
   
   /** Status type: INPUT */
   public static final int ST_INPUT   = 1;

   /** This is the default value for insert mode, means it is not used. */
   public static final int INSERT_MODE_OFF = 0;
   
   /** The mode when the next char is inserted into current caret position. */
   public static final int INSERT_MODE_INS = 1;
   
   /** Insert mode when the next typed char replaces the current one. */
   public static final int INSERT_MODE_OVR = 2;

   /** 
    * Number of reserved lines at the bottom of the screen.
    * TODO: the usage of this is reversed: the screen dimensions need to be set without
    * the message/status area in mind, as HEIGHT-CHARS attr doesn't need to include these.
    */
   public static final int RESERVED_LINES = 3;

   /** "Insert " text message (with 1 space of padding on right). */
   public static final String INSERT_TEXT = "Insert ";
   
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> wa = new ContextLocal<WorkArea>() 
   {
      @Override
      protected WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /** Inner content pane. */
   protected BorderedPanel<O> contentPane;

   /** Message-area for this window. */ 
   protected MessageArea<O> messageArea = null;
   
   /** Renders status info text. */
   protected StatusLine<O> statusLine = null;
   
   /** Menubar */
   protected Menu<O> menubar;
   
   /** Suspend mode active flag. */
   protected boolean suspend = false;

   /** Synchronization lock for suspend processing. */
   protected Object suspendLock = null;

   /** Window attributes container. */
   protected WindowConfig config = null;
   
   /** <code>true</code> if initialization has been performed. */
   protected boolean isRunning = false;

   /** Global variable to store the current insertion mode. */
   protected int insertMode = 0;

   /**
    * If <code>true</code> then insert mode is ON while insert indicator is 
    * into OFF state.
    */
   protected boolean insertInAmbiguousState = false;
   
   /**
    * Determines whether any output has been performed using PUT SCREEN
    * statement after the last call of HIDE ALL.
    */
   private boolean putScreenPerformed = false;
   
   /** Components that must be visible when interactive mode is restored. */
   protected Set<Widget<O>> vislist = null;
   
   /** Last help text. */
   private String helpText = null;
   
   /** Status message displayed when application is running */
   private String defaultMessage = "";

   /** Status message displayed when input operation is active. */
   private String inputMessage = null;
            
   /** Current status type. */
   private int statusType = ST_DEFAULT;

   /** message line number where tiny input frame is placed */
   private int tinyInputLine = -1;

   /**
    * Flag to force message clearing on the next MESSAGE stmt.
    */
   private boolean clearMessage = false;

   /**
    * <code>true</code> if automatic message clearing occurred during trigger
    * processing.
    */
   private boolean autoCleared = false;

   /**
    * Separate pause before hide flag for messages. <code>true</code> means
    * pause is due.
    */
   private boolean messageNeedPause = false;

   /**
    * Constructs the single Window instance by calling superclass constructor,
    * setting the current layout to null and performing the window
    * initialization with creation of a CHARVA frame without borders.
    * 
    * @param   id
    *          The widget's ID.
    */
   protected Window(WidgetId id)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, WindowConfig.class);

      inputMessage = ThinClient.getInstance().getInputDefaultText();
   }

   /**
    * Construct a window widget based on a specific configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   wcfg
    *          Window configuration.
    */
   protected Window(WidgetId id, WindowConfig wcfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, wcfg);
      
      inputMessage = ThinClient.getInstance().getInputDefaultText();
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WindowConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Get content of message area.
    * 
    * @return  Array of two <code>String</code> instances with message area
    *          content.
    */
   public String[] getMessageText()
   {
      return messageArea.getMessageTexts();
   }

   /**
    * Get colors of message area.
    * 
    * @return  Array of two <code>String</code> instances with message area
    *          content.
    */
   public Color[] getMessageColor()
   {
      return messageArea.getMessageColors();
   }

   /**
    * Determines if the message lines are empty.
    *
    * @return   <code>true</code> if the message line is empty.
    */
   public boolean isMessageNull()
   {
      return isMessageTextNull();
   }   
   
   /**
    * The static function to set up the global insert mode value. The possible
    * values for insert mode:
    * <p>
    * <ul>
    * <li>INSERT_MODE_OFF means the value is not important and turned off.
    * </li>
    * <li>INSERT_MODE_INS the mode when the chars are inserted at the
    *     specified position.
    * </li>
    * <li>INSERT_MODE_OVR the mode of replacing the current char with the new
    *     one
    * </li>
    * </ul>
    * <p>
    * 
    * @param   insertMode
    *          New integer value for insert mode.
    */
   public static void setInsertMode(int insertMode)
   {
      WindowManager.getDefaultWindow().setInsMode(insertMode, false);
   }

   /**
    * Set if clearing needed on next message statement.
    * 
    * @param   clearMessage
    *          <code>true</code> if clearing is needed, <code>false</code> otherwise.
    */
   public void setClearMessage(boolean clearMessage)
   {
      this.clearMessage = clearMessage;
   }

   /**
    * Set if automatic clearing is needed for trigger processing.
    * 
    * @param   autoCleared
    *          <code>true</code> if auto clearing is needed, <code>false</code> otherwise.
    */
   public void setAutoCleared(boolean autoCleared)
   {
      this.autoCleared = autoCleared;
   }
   
   /**
    * Is a pause needed for messages.
    * 
    * @return   See above.
    */
   public boolean isMessageNeedPause()
   {
      return messageNeedPause;
   }

   /**
    * Set true if pause needed for messages.
    * 
    * @param   messageNeedPause
    *          <code>true</code> if pause is needed, <code>false</code> otherwise. 
    */
   public void setMessageNeedPause(boolean messageNeedPause)
   {
      this.messageNeedPause = messageNeedPause;
   }
   
   /**
    * Set true if output has been performed using PUT SCREEN after the last call of HIDE ALL.
    * 
    * @param   putScreenPerformed
    *          <code>true</code> if PUT SCREEN was used, <code>false</code> otherwise. 
    */
   public void setPutScreenPerformed(boolean putScreenPerformed)
   {
      this.putScreenPerformed = putScreenPerformed;
   }

   /**
    * The static function to set up the global insert mode value. The possible
    * values for insert mode:
    * <p>
    * <ul>
    * <li>INSERT_MODE_OFF means the value is not important and turned off.
    * </li>
    * <li>INSERT_MODE_INS the mode when the chars are inserted at the
    *     specified position.
    * </li>
    * <li>INSERT_MODE_OVR the mode of replacing the current char with the new
    *     one
    * </li>
    * </ul>
    * <p>
    * 
    * @param   insertMode
    *          New integer value for insert mode.
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   public static void setInsertMode(int insertMode, boolean force)
   {
      WindowManager.getDefaultWindow().setInsMode(insertMode, force);
   }

   /**
    * The static function to get the global insert mode value.
    * 
    * @return  The integer value for the global insert mode flag.
    * 
    * @see     #setInsertMode for possible return values.
    */
   public static int getInsertMode()
   {
      return WindowManager.getDefaultWindow().getInsMode();
   }

   /**
    * Sets the flag that indicates that insert is in an ambiguous state (i.e.
    * insert mode is ON while insert indicator is into OFF state).
    */
   public static void setInsertInAmbiguousState()
   {
      WindowManager.getDefaultWindow().insertInAmbiguousState = true;
   }
   
   /**
    * Set help text.
    * When the request comes from a widget which is part of an ongoing CHOOSE,
    * the help text is overridden with whatever was supplied for the CHOOSE
    * operation.
    *
    * @param   text
    *          Text for help string or <code>null</code>, which indicates the
    *          default status message has to be restored
    */
   @SuppressWarnings({ "unchecked" })
   public void setHelp(String text)
   {
      String helpText = text;

      Widget<O> comp = currentFocus();
      if (comp != null)
      {
         Frame<O> frame = (Frame<O>) UiUtils.locateFrame(comp);
         if (frame != null && frame.isInChoose(comp))
         {
            helpText = frame.getChooseHelp();
         }
      }

      if (helpText == null || helpText.isEmpty())
      {
         this.helpText = null;
         status(null);
      }
      else
      {
         status(helpText);
         this.helpText = helpText;
         setStatusType(ST_HELP);
      }
   }

   /**
    * Returns <code>true</code> if insert mode is ON while insert indicator is
    * into OFF state).
    *
    * @return See above.
    */
   public static boolean isInsertInAmbiguousState()
   {
      return WindowManager.getDefaultWindow().insertInAmbiguousState;
   }
   
   /**
    * Checks to see whether the windowing environment is up.
    *
    * @return <code>true</code> if the windowing environment is up.
    */
   public static boolean isRunning()
   {
      return WindowManager.getDefaultWindow().isRunning;
   }

   /**
    * Disables the "special components" (status line, insert mode indicator,
    * message lines, temp text components) and stores a list of all currently 
    * visible top-level components (which are then set invisible before this
    * method completes).
    *
    * This is part of the switching of the client to the redirected mode.
    */
   public static void disableSpecial()
   {
      Window<?> win = WindowManager.getDefaultWindow();
      win.setVisibility(false);
      win.showRedirectedFrames();
   }
   
   /**
    * Enables the "special components" (status line, insert mode indicator,
    * message lines, temp text components) and restores visibility to the
    * list of all previously visible top-level components.
    *
    * This is part of the switching of the client to the interactive mode.
    */
   public static void enableSpecial()
   {
      Window<?> win = WindowManager.getDefaultWindow();
      win.setVisibility(true);
      win.hideRedirectedFrames();
   }
   
   /**
    * Suspends drawing except for the status line.  Resume using 
    * {@link #resume} method.
    */
   public static void suspend()
   {
      Window<?> win = WindowManager.getDefaultWindow();
      
      synchronized (win.suspendLock)
      {
         win.suspend = true;
      }
   }
   
   /**
    * Resumes normal drawing (exits suspend mode which was set using
    * {@link #suspend}).
    */
   public static void resume()
   {
      Window<?> win = WindowManager.getDefaultWindow();
      
      synchronized (win.suspendLock)
      {
         win.suspend = false;
      }
   }
   
   /**
    * Merge current settings with new ones for default static window.
    * For dynamic windows dynamic create or update registry.
    * 
    * @param   cfg
    *          New configuration data.
    */
   public static void pushConfig(WindowConfig cfg)
   {
      final WidgetId windId = cfg.id;
      final ConfigManager mgr = ConfigManager.getInstance();
      ThinClient tc = ThinClient.getInstance();
      
      if (windId.equals(WidgetId.DEFAULT_WINDOW_ID))
      {
         Window<?> defWnd = WindowManager.getDefaultWindow();
         
         // update default widow  
         tc.eventDrawingBracket(defWnd, false, true, new Runnable()
         {
            public void run()
            {
               mgr.replaceConfig(defWnd.config(), cfg);
            }
         });
      }
      else
      {
         // check registry
         Window<?> foundWindow = (Window<?>) WindowManager.findWindow(windId.asInt());
         
         if (foundWindow == null)
         {
            // new window
            final Window<?> window = (Window<?>) UiUtils.reconstructComponent(windId, cfg);
            
            // when created register
            if (window != null)
            {
               // the top is preserved
               TitledWindow<?> top = WindowManager.topWindow();
               WindowManager.addWindow(window);
               if (top != null)
               {
                  WindowManager.moveToTop(top);
               }
               
               window.screen().activateWindow(window);

               tc.eventDrawingBracket(window, false, true, new Runnable()
               {
                  public void run()
                  {
                     window.initialize(windId, window.config());

                     // the config instance may have changed after initialize
                     WindowConfig wc = window.config();
                     
                     if (mgr.getActiveConfig(wc.id) == null)
                     {
                        mgr.addWidgetConfig(wc);
                     }
                     
                     ConfigManager.syncConfigChanges(new Runnable() 
                     {
                        @Override
                        public void run()
                        {
                           WindowManager.setInitialDimension(window);
                        }
                     });
                  }
               });
            }
         }
         else
         {
            tc.eventDrawingBracket(foundWindow, false, true, new Runnable()
            {
               public void run()
               {
                  // if already exists just update
                  mgr.replaceConfig(foundWindow.config(), cfg);
               }
            });
         }
      }
   }
   
   /**
    * Save current status type and text in stack.
    *
    * @param    cnt
    *           How many times state should be saved in stack.
    */
   public static void pushStatus(int cnt)
   {
      if (OutputManager.getDriver().inBatchMode())
      {
         // no-op in batch mode
         return;
      }

      WorkArea wrk = wa.get();
      ThinClient tc = ThinClient.getInstance();

      Map<Integer, Integer> wndStatusTypes = new HashMap<Integer, Integer>();
      for (TitledWindow<?> twnd : WindowManager.windowList())
      {
         if (!(twnd instanceof Window))
         {
            continue;
         }
         
         Window<?> wnd = (Window<?>) twnd;
         
         int wndId = UiUtils.getWidgetId(wnd).asInt();
         if (wndId == -1)
         {
            throw new IllegalStateException("Window doesn't have ID");
         }
         else
         {
            wndStatusTypes.put(wndId, wnd.getStatusType());
         }

         if (wnd.getStatusType() != Window.ST_DEFAULT)
         {
            wnd.status(null);
         }
      }
      
      int inputActive = tc.getInputActive();
      for (int i = 0; i < cnt; i++)
      {
         wrk.statusStack.push(new StatusData(wndStatusTypes, inputActive));
      }

      // this needs to be done only once
      tc.setInputActive(0); 
   }

   /**
    * Restore current status type and text from stack.
    *
    * @param    cnt
    *           How many elements should be popped from stack.
    */
   public static void popStatus(int cnt)
   {
      if (OutputManager.getDriver().inBatchMode())
      {
         // no-op in batch mode
         return;
      }

      StatusData storedStatus = null;

      WorkArea wrk = wa.get();
      for (int i = 0; i < cnt; i++)
      {
         if (wrk.statusStack.isEmpty())
            continue;

         storedStatus = (StatusData) wrk.statusStack.pop();
      }

      if (storedStatus == null)
         return;
      
      final StatusData storedStatusFin = storedStatus;

      final ThinClient tc = ThinClient.getInstance();
      for (TitledWindow<?> wnd : WindowManager.windowList())
      {
         final TitledWindow<?> twnd = wnd;
         
         tc.independentEventDrawingBracket(twnd, new Runnable()
         {
            @Override
            public void run()
            {
               if (!(twnd instanceof Window))
               {
                  return;
               }
               
               Window<?> wnd = (Window<?>) twnd;
               
               if (!wnd.isRunning)
               {
                  // the window has not been initialized, do not attempt to restore its status
                  return;
               }

               int wndId = UiUtils.getWidgetId(wnd).asInt();
               final Integer type = storedStatusFin.types.get(wndId);

               // the whole status line will lose PUT SCREENs
               if (wnd.config().statusArea)
               {
                  wnd.refresh(wnd.getStatusLineRectangle());
               }

               String msg = "";

               // Type may be null in case twnd had not existed at the time storedStatusFin 
               // was pushed on the status stack. In this case status text and type should be 
               // set to the default values (an empty and ST_DEFAULT).
               if (type != null)
               {
                  switch (type)
                  {
                     case Window.ST_DEFAULT:
                        msg = wnd.getDefaultMessage();
                        break;
                     case Window.ST_INPUT:
                        msg = wnd.getInputMessage();
                        break;
                     case Window.ST_HELP:
                        msg = wnd.getHelpText();
                        break;
                  }
               }

               if (msg == null)
               {
                  // TODO: Check if this is correct, probably there is more complicated
                  // fallback logic in cases when some status text is null.
                  msg = "";
               }
               
               wnd.status(msg);
               wnd.setStatusType(type == null ? ST_DEFAULT : type);
            }
         });
      }

      tc.setInputActive(storedStatus.input);
   }

   /**
    * Get status area rectangle.
    *
    * @return  Instance of <code>Rectangle</code> which describes the status
    *          area in screen coordinates.
    */
   public abstract Rectangle getStatusLineRectangle();

   /**
    * Get messages area rectangle by message line index.
    *
    * @param   line
    *          message line specification: either 0-based line number for
    *          specific line or -1 for the whole message area.
    * @return  Instance of <code>Rectangle</code> which describes the message
    *          area in screen coordinates.
    */
   public abstract Rectangle getMessageLineRectangle(int line);

   /**
    * Initialize window
    */
   protected void init()
   {
      contentPane = screen().getFactory().createBorderedPanel();
      contentPane.setLayout(null);
   }

   /**
    * Set the contents of the status line. This method is private and
    * called at service level by another static public method within the
    * class.  The status line output is truncated to a maximum of 63
    * characters, only if the <code>noLengthLimit</code> flag is not set.
    * 
    * @param   status
    *          The new string to be displayed inside the status line area.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   protected abstract void setStatusLineText(String status, boolean noLengthLimit);
   
   /**
    * A window's ancestor is itself.
    * 
    * @return   This window instance.
    */
   @Override
   public TitledWindow<O> ancestor()
   {
      return this;
   }
   
   /**
    * A window's "window" is itself.
    * 
    * @return   This window instance.
    */
   @Override
   public Window<O> window()
   {
      return this;
   }

   /**
    * Get the {@link Menu} widget assigned to this window as MENUBAR.
    * 
    * @return   See above.
    */
   public Menu<O> getMenubar()
   {
      return menubar;
   }
   
   /**
    * Get the message area widget associated with this window.
    * 
    * @return   See above.
    */
   public MessageArea<O> getMessageArea()
   {
      return messageArea;
   }
   
   /**
    * Get the last help text.
    * 
    * @return   See above.
    */
   public String getHelpText()
   {
      return helpText;
   }

   /** 
    * Get current status type.
    * 
    *  @return See above.
    */
   public int getStatusType()
   {
      return statusType;
   }

   /** 
    * Set current status type.
    * 
    *  @param    statusType
    *            new statusType - one of 3 constant values.
    */
   public void setStatusType(int statusType)
   {
      this.statusType = statusType;
   }

   /**
    * Get content of status line.
    * 
    * @return  Status line content.
    */
   public String getStatusText()
   {
      return statusLine.getText();
   }
   
   /**
    * Get the default status message.
    * 
    * @return   See above.
    */
   public String getDefaultMessage()
   {
      return defaultMessage;
   }

   /**
    * Get the input status message.
    * 
    * @return   See above.
    */
   public String getInputMessage()
   {
      return inputMessage;
   }

   /**
    * Get the status area widget associated with this window.
    * 
    * @return   See above.
    */
   public StatusLine<O> getStatusArea()
   {
      return statusLine;
   }
   
   /**
    * Renders the specified message into the status line.
    *
    * @param   message
    *          The message body to be rendered.
    */
   public void status(final String message)
   {
      status(message, false);
   }
   
   /**
    * Renders the specified message into the status line. The message
    * is trimmed only if the <code>noLengthLimit</code> flag is not set.
    * <p>
    * The method is a no-op when the status area is not displayed for this window.
    *
    * @param   message
    *          The message body to be rendered.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   public void status(final String message, final boolean noLengthLimit)
   {
      if (!config.statusArea)
      {
         return;
      }
      
      ThinClient tc = ThinClient.getInstance(); 
      int tempTerm = tc.forceInteractive();

      NativePoint   curpos = null;
      boolean curvis = true;

      O screen = screen();
      if (tc.getInputActive() > 0 && screen.isChui())
      {
         // the cursor is active in an edit operation, remember the current
         // cursor position (and make the cursor invisible to avoid flashing)
         curvis = screen.setCursorStatus(false);
         curpos = ((ChuiOutputManager) screen).cursor();
      }

      drawStatusLine(message, noLengthLimit);

      helpText = null;

      if (curpos != null)
      {
         tc.restoreCursor(curpos, curvis);
      }

      tc.restoreRedirection(tempTerm);
   }
   
   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method is used to display a status message, which is trimmed to
    * 63 chars only if the <code>noLengthLimit</code> flag is not set.
    *
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    * @param    noLengthLimit
    *           flag indicating the status message should not be trimmed
    */
   public void statusDefault(String text, boolean noLengthLimit)
   {
      statusType = ST_DEFAULT;
      defaultMessage = (text == null) ? "" : text;
      status(text, noLengthLimit);
   }
   
   /**
    * Specifies a new status message used when the user interface is blocked
    * on input.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT text</code>.
    *
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    */
   public void statusInput(String text)
   {
      if (text != null && text.equals(inputMessage))
         return;

      inputMessage = (text == null) ? "" : text;
      status(null);
   }
   /**
    * Get window content pane.
    * 
    * @return  content pane.
    */
   public Container<O> getContentPane() 
   {
      return contentPane;
   }

   /**
    * Get widgets array from first available window.
    * 
    * @return  array of the widgets belonging to the resolved windows.
    */
   public Widget<O>[] getWidgets()
   {
      return contentPane.widgets();
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public WindowConfig config()
   {
      return config;
   }

   /**
    * Adds the frame to the window at the specified position.
    * 
    * @param    widget
    *           The handle of the frame to be added to the window.
    * @param    row
    *           The row number for the upper-left frame corner.
    * @param    column
    *           The column number for the upper-left frame corner.
    */
   public void addFrame(Widget<O> widget, double row, double column)
   {
      row = Coordinate.scale(row);
      column = Coordinate.scale(column);
      
      if (widget == null && row == 0 && column == 0)
         return;
      
      getContentPane().add(widget);

      FrameConfig cfg = UiUtils.getConfig(widget, FrameConfig.class);
      if (cfg != null)
      {
         cfg.windowID = UiUtils.getWidgetIdAsInt(this);
      }
   }

   /**
    * Draw all the components in this window, and request the keyboard focus.
    * <p>
    * If in suspend mode, only draw the status line.  This mode can be set
    * via {@link #suspend}.
    */
   @Override
   public void draw()
   {
      synchronized (suspendLock)
      {
         if (suspend)
         {
            // suspend path, only draw the status line
            if (config.statusArea)
            {
               statusLine.draw();
            }
         }
         else
         {
            // draw the entire contents of this container
            ThinClient tc = ThinClient.getInstance();
            tc.viewDisplay = false;
            tc.viewRefresh = true;
            getContentPane().draw();
            
            resetBitmap();
            
            drawCaret();
         }
      }
   }
   
   /**
    * Draw status line.
    *
    * @param   message
    *          Message to display.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   public void drawStatusLine(final String message, final boolean noLengthLimit)
   {
      // safely ignore any events that may be produced by processHide() since
      // it only may remove temp texts over the status line which is being
      // redrawn anyway.
      final ThinClient tc = ThinClient.getInstance();
      tc.independentEventDrawingBracket(this, () ->
      {
         String txt = message;
         if (message == null)
         {
            if (tc.getInputActive() > 0)
            {
               if (helpText != null)
               {
                  txt        = helpText;
                  statusType = ST_HELP;
               }
               else
               {
                  txt        = inputMessage;
                  statusType = ST_INPUT;
               }
            }
            else
            {
               txt        = defaultMessage;
               statusType = ST_DEFAULT;
            }
         }
         
         // the whole status line will lose PUT SCREENs
         refresh(getStatusLineRectangle());
         statusInt(txt, noLengthLimit);
      });
   }
   
   /**
    * Shows the new message text.
    * <p>
    * This is a no-op when this window doesn't have the message area displayed.
    * 
    * @param   message
    *          Array of two text message which need to be displayed inside the
    *          message area.
    * @param   cs
    *          Array of two color specifications.
    */
   public void setMessageText(String[] message, Color[] cs)
   {
      if (config.messageArea)
      {
         messageArea.setMessageText(message, cs);   
      }
   }
      

   /**
    * Set message main worker routine.
    * <p>
    * This is a no-op when this window doesn't have the message area displayed.
    * 
    * @param   msg
    *          Message text array.
    * @param   cs
    *          Message color array.
    * @param   force
    *          <code>true</code> means force redisplaying of messages even
    *          if message text is already <code>null</code>.
    */
   public void setMessageText(String[] msg, Color[] cs, boolean force)
   {
      if (config.messageArea)
      {
         messageArea.setMessageText(msg, cs, force);
      }
   }
   
   /**
    * Perform complete input operation: create fill-in widget, process user
    * input and return result.
    *
    * @param   text
    *          The message text.
    * @param   row
    *          Message row in which operation will be performed.
    * @param   var
    *          Source variable.
    * @param   format
    *          Format which will be used for variable.
    * @param   auto
    *          Fill-in AUTO-RETURN flag.
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    *
    * @return  The result of the input operation if the operation finished
    *          successfully, otherwise <code>null</code>.
    */
   @SuppressWarnings({ "rawtypes", "unchecked" })
   public MessageReturnValue tinyInput(String       text,
                                       int          row,
                                       BaseDataType var,
                                       String       format,
                                       boolean      auto,
                                       boolean      set)
   {
      ThinClient tc = ThinClient.getInstance();
      
      // save the currently focused widget so it can be restored later
      // (this must be done before the frame is added otherwise the result
      // will yield the "fin" widget below)
      final Widget oldFocus = UiUtils.getCurrentFocus();
      final Frame saveLastViewed = tc.getLastViewed();

      int id0 = WidgetId.INPUT_FRAME_ID.asInt();
      int id1 = WidgetId.INPUT_FRAME_ID.asInt() + 1; // must remain INPUT_FRAME_ID + 1
      
      int msglen = text.length();
      
      boolean chui = tc.isChui();
      
      final ModalWindow[] win = new ModalWindow[1];
      
      int wndId = UiUtils.getWidgetIdAsInt(this);
      
      // GUI mode is in a modal window instead of inline in the message area
      if (!chui)
      {
         tc.eventBracket(new Runnable()
         {
            @Override
            public void run()
            {
               win[0] = new DialogBoxWindow("Message Update");
               win[0].realize();
               win[0].setEnabled(true);
            }
         });
      }
      
      // in GUI we have a different location AND we use side-labels to display the text
      FrameConfig cfg = new FrameConfig(new WidgetId(id0));
      cfg.box         = false;
      
      // do not set the location for gui so that the created modal window
      // will center itself relative to the owner window
      if (chui)
      {
         cfg.row    = getScreenDimension().height + row;
         cfg.column = msglen + 1;
      }
      
      cfg.overlay     = true;
      cfg.labels      = !chui;   
      cfg.sideLabels  = !chui;
      
      if (!chui)
      {
         wndId = win[0].getId().asInt();
         
         // during Frame.initialize() the frame will be attached to the specified window
         cfg.windowID = wndId;
         
         // set the dialog to true so that the frame window is created
         cfg.dialog = true;
      }
                                                                                
      FillInConfig fcfg = new FillInConfig(new WidgetId(id1));
      fcfg.frameId    = id0;
      fcfg.autoReturn = auto;
      fcfg.dataType   = var.getTypeName();                              // direct access

      if (format != null)
      {
         fcfg.format = format;                                          // direct access
      }
      
      // GUI mode displays the text as a side label, the caller bypassed outputting the
      // text to the message area
      if (!chui)
      {
         fcfg.label = text;
      }

      MessageArea msgArea = getMessageArea();
      
      String[] msg = msgArea.getMessageTexts();
      
      // detect when tinyInput() is nested
      Frame savedTinyInput = (Frame) tc.getWidget(WidgetId.INPUT_FRAME_ID);
      
      String[] messages = null;
      Color[]  colors   = null;
      
      O screen = screen();
      
      if (savedTinyInput != null)          
      {
         messages = msgArea.getMessageTexts();
         colors   = msgArea.getMessageColors();
         screen.getRegistry().removeFrame(savedTinyInput);
      }

      screen.getRegistry().reconstructWidget(cfg);
      screen.getRegistry().reconstructWidget(fcfg);

      Frame frame = (Frame) screen.getRegistry().getComponent(id0);
      FillIn fin = (FillIn) screen.getRegistry().getComponent(id1);
      fin.setForceLeft(true);

      if (chui)
      {
         // TODO: does GUI have any truncation behavior that needs to be supported?
         
         int maxlen = (int) (dimension().width - fin.width() - 1);
   
         if (msglen > maxlen)
         {
            // message is too long to be accommodated with the update FILL-IN
            // and is to be truncated
   
            msg[row] = msg[row].substring(0, maxlen) +
                       StringHelper.repeatChar(' ', msglen - maxlen);
   
            setMessageText(msg, null);
   
            Point loc = frame.location();
            loc.x = maxlen + 1;
            frame.setLocation(loc);
         }
      }

      frame.addWidget(fin);
      // the frame is already attached to the window by Frame.initialize()

      frame.doLayout();

      // Frame adding/removal is performed manually,
      // so we must protect if from automated processing.
      frame.setProtected(true);

      ScreenBuffer sb = new ScreenBuffer(id0, 1);

      // copy initial value in UPDATE mode; unknown value can also be used as initial except for date
      // based variables
      if (!set && !(var.isUnknown() && var instanceof date))
      {
         sb.putWidgetValue(id1, var.duplicate());
      }
      
      screen.setInvalidate(true);

      tc.eventBracket(new Runnable()
      {
         @Override
         public void run()
         {
            if (oldFocus != null)
               EventManager.postEvent(new FocusEvent(oldFocus, EventType.FOCUS_LOST));
         }
      });

      tc.enable(id0, null, new ScreenBuffer[] {sb}, true, wndId);
      tc.clear(id0, true);
      sb.resetChanged();
      screen.setInvalidate(chui ? this : win[0], false);

      tc.setNeedPause(false, this);

      ScreenBuffer[] sbf = null;

      // in ChUI this is a valid message line number, in GUI it just needs to be a positive
      // number that is outside of the range of valid message lines
      tinyInputLine = chui ? row : msg.length + 100;

      tc.setLastViewed(frame);

      // save insert indicator state in order to restore after on leaving
      // waitForWorker()
      int savedInsState = tc.saveInsState();
      String status     = getStatusText();
      int type          = getStatusType();

      try
      {
         EventList list = new EventList(true);
         list.setOperation(set ? EventList.SET : EventList.UPDATE);

         sbf = tc.waitForWorker(list, id1, -1, new ScreenBuffer[] {sb}, false, true, false, 
                                set ? BlockingOperation.MESSAGE_SET 
                                    : BlockingOperation.MESSAGE_UPDATE);
      }

      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         screen.resetInvalidate(chui ? this : win[0]);

         // let the original exception propagate
         throw ce;
      }

      finally
      {
         // cleanup that must occur even if a condition is raised
         tinyInputLine = -1;
         tc.enable(id0, null, new ScreenBuffer[] {sb}, false, wndId);
         tc.hide(id0, true, false);
         Frame tmpFrame = (Frame) screen.getRegistry().getComponent(id0);
         screen.getRegistry().removeFrame(tmpFrame);
         
         // frame removal from registry will trigger frame destroy,
         // that hides and destroys the modal window if in GUI

         if (savedTinyInput != null)
         {
            screen.getRegistry().addFrame(savedTinyInput);
            addFrame(savedTinyInput, 0, 0);
            
            // restore the message lines
            // TODO: does this same behavior occur in GUI?
            setMessageText(messages, colors, true);
         }
         else
         {
            hideMessage(true);
         }

         tc.setNeedPause(false, this);

         setStatusType(type);
         status(status);

         if (oldFocus != null)
         {
            tc.eventDrawingBracket(this, false, false, new Runnable()
            {
               @Override
               public void run()
               {
                  FocusManager.setInitialFocus(oldFocus);
               }
            });
         }

         if (saveLastViewed != null && !saveLastViewed.config().box)
         {
            tc.eventDrawingBracket(this, false, false, new Runnable()
            {
               @Override
               public void run()
               {
                  saveLastViewed.repaint();
               }
            });
         }

         tc.setLastViewed(saveLastViewed);

         tc.restoreInsState(savedInsState, true);
      }

      MessageReturnValue mrv = new MessageReturnValue();

      mrv.keycode = sbf[0].getKeyCode();
      mrv.var     = (BaseDataType) sbf[0].getWidgetValue(id1);

      // no edit occurred
      if (mrv.var == null)
      {
         // return original variable
         mrv.var = var;
      }

      return mrv;
   }
   
   /**
    * Reports if the user is in a tiny input editing session (MESSAGE SET/UPDATE).
    * 
    * @return   <code>true</code> if tiny input is active.
    */
   public boolean isTinyInputActive()
   {
      return tinyInputLine != -1;
   }
   
   /**
    * Obtain the message area line number being used by the currently active MESSAGE SET/UPDATE.
    *
    * @return   The message area line being edited.
    */
   public int getTinyInputLine()
   {
      return tinyInputLine;
   }

   /**
    * Renders the specified message into the message line.
    *
    * @param   message
    *          The message body to be rendered.
    */
   public void message(String message)
   {
      message(message, Color.MESSAGES);
      sync();
   }

   /**
    * Renders the specified message into the message line.
    *
    * @param   message
    *          The message body to be rendered.
    * @param   cs
    *          color specification.
    *
    * @return  Message line number used to display message.
    */
   public int message(String message, Color cs)
   {
      return message(message, cs, false);
   }

   /**
    * Renders the specified message into the message line.
    *
    * @param   message
    *          The message body to be rendered.
    * @param   cs
    *          color specification.
    * @param   clear
    *          If <code>true</code> then the line(s) on which the message will
    *          be displayed, will be cleared first.
    *
    * @return  Message line number used to display message.
    */
   public int message(String message, Color cs, boolean clear)
   {
      if (!isVisible())
      {
         ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
         {
            public void run()
            {
               show();
            }
         });
      }
   
      String original = message;
      ThinClient tc = ThinClient.getInstance();

      // check if the message is too large for 1 message line
      if (dimension().width < message.length())
      {
         return splitMessage(message, cs);
      }
      else
      {
         // special mode that truncates text starting at any newlines (there will always
         // be at least 1 element in the array)
         String[] truncated = message.split("\n|\r\n|\r", 2);
         message = truncated[0];
      }

      // query redirection
      boolean switched = tc.isRedirected();
      int redir = -1;

      // in redirected mode we output to both the terminal and to the stream, but we output
      // the unmodified text, not the text that may have been truncated above
      if (switched)
      {
         redir = tc.sendToStream(original);
      }

      messageClearWorker();

      String[] content = messageArea.getMessageTexts();

      try
      {
         if (content[0] == null)
         {
            // 'clear' flag is no longer used
            content[0] = message;
            displayMessage(content, cs, 0);
            return 0;
         }

         if (content[1] == null)
         {
            // 'clear' flag is no longer used
            content[1] = message;
            displayMessage(content, cs, 1);
            return 1;
         }

         try
         {
            if (messageNeedPause)
               tc.pauseBeforeHide();

            return 0;
         }
         finally
         {
            // 'clear' flag is no longer used
            displayMessage(new String[] {message, null}, cs, -1);
         }
      }
      finally
      {
         // if currently redirected, restore the redirection
         if (switched)
         {
            tc.switchOutput(redir, false);
         }

         messageNeedPause = true;
      }
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    *
    * @param   text
    *          Text to be displayed.
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE,
    *          not used if var parameter is <code>null</code>.
    * @param   var
    *          The variable to set or update.  <code>null</code> means no
    *          set or update will occur.
    * @param   auto
    *          <code>true</code> if the AUTO-RETURN option is specified. This
    *          is only valid if <code>var</code> is not <code>null</code>.
    * @param   format
    *          The format specification to use for the set or update.  This
    *          is only valid if <code>var</code> is not <code>null</code>.
    * @param   cs
    *          The color specification to use for the text.
    *
    * @return  The result of the input operation if input parameter
    *          <code>var</code> is not <code>null</code> and operation
    *          finished successfully and/or if any key was pressed during
    *          editing/pausing.   If no pause occurred and there was no input
    *          (editing) in process, then <code>null</code> will be returned.
    */
   @SuppressWarnings("rawtypes")
   public MessageReturnValue message(String       text,
                                     boolean      set,
                                     BaseDataType var,
                                     boolean      auto,
                                     String       format,
                                     Color        cs)
   {
      // only show the default window
      if (!isVisible() && getId().equals(WidgetId.DEFAULT_WINDOW_ID))
      {
         ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
         {
            public void run()
            {
               show();
            }
         });
      }

      ThinClient tc = ThinClient.getInstance();
      
      boolean reactivateBatch = false;
      
      try
      {
         // special processing for MESSAGE statement in batch mode
         if (tc.isRedirected() && OutputManager.inBatchMode())
         {
            reactivateBatch = OutputManager.activateBatch(false);
         }
         // if not in batch mode == not in batch background
         else if (PlatformHelper.isUnderWindowsFamily() && OutputManager.isBatchInBackground() ||
                  FileChecker.isConsoleRedirected() && OutputManager.inBatchMode())
         {
            reactivateBatch = OutputManager.activateBatch(false);
            
            // in this mode we have only error channel to display something in console
            ErrorManager.displayError(text);
            
            // this is not very good but here we sure do not have return value
            return null;         
         }
   
         MessageReturnValue mrv     = null;
         Widget             focused = UiUtils.getCurrentFocus();
         NativePoint        curpos = null;
         boolean            curvis = false;
         
         if (tc.isChui())
         {
            curpos = ((ChuiOutputManager) screen()).cursor();
            curvis = screen().setCursorStatus(false);
            
            Dimension dim = dimension();
         
            // TODO: does GUI need any counterpart to this?
            if (var != null && text.length() > dim.width)
               text = text.substring(0, (int) dim.width);
         }
   
         // save off the most recent last-key
         int prevKey = tc.getLastKey();
         
         if (tc.isInEditingBlock() && var != null)
         {
            clearMessage = false;
         }
   
         int row = -1;
         
         // the actual message content is only output to the message lines for ChUI
         // if we are in MESSAGE SET/UPDATE mode; if not in MESSAGE SET/UPDATE mode
         // then it should output for GUI or ChUI
         if (var == null || tc.isChui())
         {
            row = message(text, cs, var != null);
         }
   
         // query redirection
         boolean switched = tc.isRedirected();
         int redir = -1;
   
         // switch to interactive mode to allow variable reading
         if (switched)
         {
            redir = tc.switchOutput(-1, false);
         }
   
         // detect if a key was pressed (because of pausing)
         if (prevKey != tc.getLastKey())
         {
            mrv = new MessageReturnValue();
            mrv.keycode = tc.getLastKey();
         }
   
         try
         {
            if (var != null)
            {
               // handle SET/UPDATE modes
   
               mrv = tinyInput(text, row, var, format, auto, set);
   
               clearPutScreenInMessageArea(-1);
   
               // TODO: after the var is read and if the terminal is redirected,
               // then the variable must be output to the stream too.
            }
         }
   
         finally
         {
            if (tc.getTriggerNesting() > 0 && var != null && focused == null)
            {
               // adjust focus to the first focusable widget in the topmost frame
               // only if there is trigger nesting
               Frame frame = getTopVisibleFrame();
               if (frame != null)
               {
                  Widget first = frame.getFirstEnabledWidget();
   
                  if (first != null)
                  {
                     tc.getFocusManager().setFocusOn(first);
                  }
               }
            }
   
            if (tc.isChui())
            {
               tc.restoreCursor(curpos, curvis);   // syncs internally
            }
   
            // if currently redirected, restore the redirection
            if (switched)
            {
               tc.switchOutput(redir, false);
            }
          }
   
         return mrv;
      }
      finally
      {
         if (reactivateBatch)
         {
            OutputManager.activateBatch(true);
         }
      }
   }

   /**
    * Force clearing of any temp-texts inside the message area or on the
    * specific line in the message area.
    *
    * @param line
    *        Message line specification: either 0-based line number for
    *        specific line or -1 for the whole message area.
    */
   public void clearPutScreenInMessageArea(int line)
   {
      Rectangle bounds = getMessageLineRectangle(line);
      refresh(bounds);
      sync();
   }
   
   /**
    * Perform direct screen cleanup.
    *
    * @param   framesToHide
    *          <code>true</code> if there are any frames to hide.
    */
   public void clearScreenDirectly(boolean framesToHide)
   {
      // clear entire screen directly
      WidgetConfig config = config();

      Color color = UiUtils.getDColor(((WindowConfig)config).dcolor, null);
      
      Dimension dim = dimension();
      
      clear(new Point(0, 0), dim, color);
      refresh(new Rectangle(new Point(0, 0), dim, screen().coordinates().baseUnits()));  
      sync();

      if (!framesToHide &&
          !putScreenPerformed &&
          !insertInAmbiguousState)
      {
         // redraw insert indicator
         setInsMode(insertMode, true);
      }
      else if (insertMode == Window.INSERT_MODE_INS)
      {
         // insert is ON and we didn't redraw it
         insertInAmbiguousState = true;
      }

      putScreenPerformed = false;
   }

   /**
    * Hide all frames.
    *
    * @return  <code>true</code> if at least one frame was hidden.
    */
   @SuppressWarnings("rawtypes")
   public boolean hideAllFrames()
   {
      boolean framesToHide = false;

      // set explicitHide flag for all visible frames
      for (Widget comp : UiUtils.getTopLevelComponents(this))
      {
         if (!(comp instanceof Frame))
            continue;

         Frame frame = (Frame) comp;

         // reset frame visibility tracking
         frame.setVisibilityChanged(false);

         if (frame.isVisible())
         {
            frame.setExplicitHide(true);
            frame.resetPlacementQuirk(true);
            framesToHide = true;

            for (Widget widget : frame.getContentPane().widgets())
            {
               if (widget.isVisible())
               {
                  widget.setVisible(false);
               }
            }
         }
      }

      return framesToHide;
   }

   /**
    * Hides all messages displayed in the message area. If pausing is in
    * effect, the standard message "Press space bar to continue" is displayed
    * on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>HIDE MESSAGE [NO_PAUSE] </b>
    *
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public void hideMessage(boolean noPause)
   {
      ThinClient tc = ThinClient.getInstance();
      
      if (isMessageTextNull() || tc.isRedirected() || !config.messageArea)
      {
         // nothing to do, avoids pause (which doesn't happen in Progress)
         // and extra work when the message is already empty
         return;
      }

      NativePoint   curpos = null;
      boolean curvis = true;
      
      O screen = screen();
      if (tc.getInputActive() > 0 && screen.isChui())
      {
         // the cursor is active in an edit operation, remember the current
         // cursor position (and make the cursor invisible to avoid flashing)
         curvis = screen.setCursorStatus(false);
         curpos = ((ChuiOutputManager) screen).cursor();
      }

      try
      {
         if (!noPause && messageNeedPause &&
             (!tc.isInEditingBlock() ||
             (tc.isInEditingBlock() && clearMessage)))
         {
            tc.pauseBeforeHide();
         }
      }
      
      finally
      {
         // force cleanup of message area
         setMessageText(null, null, true);
         clearPutScreenInMessageArea(-1);

         if (curpos != null)
         {
            tc.restoreCursor(curpos, curvis);
         }

         autoCleared = false;
      }
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   @SuppressWarnings("unchecked")
   public void afterConfigUpdate(WindowConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
      boolean layout = false;
      
      if (config.menubarId != 0 && beforeUpdate.menubarId == 0)
      {
         menubar = (Menu<O>) this.screen().getRegistry().getComponent(config.menubarId);
         
         if (menubar != null)
         {
            if (menubar.parent() == null)
            {
               getContentPane().add(menubar);
            }
         }
         
         menubar.setVisible(true);

         layout = true;
      }

      if (config.menubarId == 0 && beforeUpdate.menubarId != 0)
      {
         if(menubar != null)
         {
            getContentPane().detach(menubar);
            menubar = null;
            layout = true;
         }
      }
      
      if (layout)
      {
         doLayout();

         if (isVisible())
         {
            repaint();
         }
      }
   }
   
   /**
    * If the method returns <code>true</code> the widget should be represented with 3D visuals.
    * 
    * @return   see above.
    */
   @Override
   public boolean isThreeD()
   {
      return config().threeD;
   }

   /**
    * Reset Window instance state. This initialization code is common for both ChUI and GUI.
    * This code MUST be called from subclasses. 
    */
   protected void reset()
   {
      @SuppressWarnings("unchecked")
      WidgetFactory<O> factory = screen().getFactory();

      messageArea = factory.createMessageArea(WidgetId.nextID(), RESERVED_LINES - 1);
      statusLine = factory.createStatusLine(WidgetId.nextID());
      
      isRunning = false;
      suspend = false;
      suspendLock = new Object();
      vislist = new HashSet<>();
   }

   /**
    * Enables / Disables the "special components" (status line, insert mode indicator,
    * message lines, temp text components) and restores visibility to the
    * list of all previously visible top-level components.
    *
    * This is part of the switching of the client to the interactive mode.
    * 
    * @param   visibility
    *          <code>true</code> enable <code>false</code> disable.
    */
   protected void setVisibility(boolean visibility)
   {
      if (config.statusArea)
      {
         statusLine.forceVisibility(visibility);
      }
      
      if (config.messageArea)
      {
         messageArea.forceVisibility(visibility);
      }
   }

   /**
    * The private function to set up the global insert mode value.
    * 
    * @param   insMode
    *          New integer value for insert mode.
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   protected abstract void setInsMode(int insMode, boolean force);

   /**
    * Get the global insert mode value.
    * 
    * @return   The integer value for the global insert mode flag.
    */
   protected int getInsMode()
   {
      return insertMode;
   }

   /**
    * Determines if the message lines are empty.
    *
    * @return   <code>true</code> if the message line is empty.
    */
   private boolean isMessageTextNull()
   {
      if (messageArea == null)
      {
         return true;
      }
      
      String[] msgs = messageArea.getMessageTexts();
      
      for (int i = 0; i < msgs.length; i++)
      {
         if (msgs[i] != null)
         {
            return false;
         }
      }
      
      return true;
   }   
   
   /**
    * Message line clear worker.
    */
   private void messageClearWorker()
   {
      // these are the implicit conditions that cause the message lines to
      // be cleared before output (without any pauses)
      ThinClient tc = ThinClient.getInstance();
      
      if (clearMessage                                        ||
          tc.isAfterChoose()                                  ||
          (tc.getTriggerNesting() > 0 && !autoCleared) ||
          (!messageNeedPause && !tc.isInEditingBlock()))
      {
         // if there was something - refresh message area
         if (!isMessageTextNull())
            clearPutScreenInMessageArea(-1);

         setMessageText(null, null);
         autoCleared = true;
         tc.setAfterChoose(false);
         clearMessage = false;
      }
   }

   /**
    * Display specified message lines.
    *
    * @param   content
    *          Message lines.
    * @param   cs
    *          Color specification.
    * @param   ndx
    *          Message index.
    */
   public void displayMessage(String[] content, Color cs, int ndx)
   {
      Color[] colorSpec = (ndx != 1) ? new Color[] {cs, null} : new Color[] {null, cs};

      setMessageText(content, colorSpec);
      clearPutScreenInMessageArea(ndx);
   }

   /**
    * Split long message and display it piece by piece.
    *
    * @param   message
    *          Message to split.
    * @param   cs
    *          Color specification.
    *
    * @return  Index of last displayed message line.
    */
   private int splitMessage(String message, Color cs)
   {
      Dimension screenSize = dimension();
      String[] lines = StringHelper.nolossWrapText(message, (int)screenSize.width);
      int num = -1;

      for (int i = 0; i < lines.length; i++)
      {
         // all lines except the first have leading whitespace removed
         if (i > 0)
            lines[i] = StringHelper.safeTrimLeading(lines[i]);

         // 1-word lines that are longer than the screen must be split
         while (screenSize.width < lines[i].length())
         {
            // output the front of the line
            message(lines[i].substring(0, (int)screenSize.width), cs);

            // shorten our line by the part that was output
            lines[i] = lines[i].substring((int)screenSize.width);
         }

         // use recursion to process each line
         num = message(lines[i], cs);
      }
      return num;
   }


   /**
    * Updates the status line area with the given text. The status line
    * output is truncated to a maximum of 63 characters, only if the 
    * <code>noLengthLimit</code> flag is not set.
    * 
    * @param   status
    *          The new string to be displayed inside the status line area.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   private void statusInt(String status, boolean noLengthLimit)
   {
      setStatusLineText(status, noLengthLimit);
   }

   /**
    * Hides all redirected frames that are visible.
    * All frames being made hidden are stored in a list. This action is part
    * of switching of the terminal back to the interactive mode.
    */
   @SuppressWarnings({ "unchecked", "rawtypes" })
   private void hideRedirectedFrames()
   {
      Widget<?>[] list = (Widget<?>[]) getWidgets();
      
      for (int i = 0; i < list.length; i++)
      {
         if ((list[i] instanceof Frame) && ((Frame<?>)list[i]).isRedirected())
         {
            Frame<?> frame = (Frame<?>) list[i];
            
            if (frame.isVisible())
            {
               frame.forceVisibility(false);
               vislist.add((Widget) list[i]);
            }
            
            if (frame.hasSavedState())
            {
               frame.restore();
               frame.setRedirected(false, -1);
            }
         }
      }
   }

   /**
    * Restores visibility of all redirected frames.
    * This action is part of switching of the terminal to the redirected mode.
    * <p>
    * If the terminal has ever been switched to the redirected mode and then
    * back, all redirected frames have been hidden during the back switch to
    * the interactive mode. Now those frames have to be made visible again.
    */
   private void showRedirectedFrames()
   {
      for (Widget<?> cmp : vislist)
      {
         cmp.forceVisibility(true);
      }
      
      vislist.clear();
   }

   /**
    * Status data container.
    */
   private static class StatusData
   {
      /** Windows status types. The key is a window ID and the value is a status type. */
      private final Map<Integer, Integer> types = new HashMap<Integer, Integer>();
      
      /** Status input. */
      private final int input;

      /**
       * Constructor.
       *
       * @param    types
       *           Status types, per each window.
       * @param    input
       *           Status input.
       */
      public StatusData(Map<Integer, Integer> types, int input)
      {
         this.types.putAll(types);
         this.input = input;
      }
   }

   /**
    * Stores common data for all windows, relating to the state of the current context.
    */
   private static final class WorkArea
   {
      /** Statuses data stack. Entry contains statusTypes and input counter */
      private final Stack<StatusData> statusStack = new Stack<StatusData>();
   }
}