/*
** Module  : ZeroColumnLayout.java
** Purpose : performs all layout manager functions
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050627   @21558 Creation initial version.
** 002 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 003 EVL 20050731   @21925 Fixing the Null pointer exception in the case
**                           when the label is null.
** 004 SIY 20050831   @22556 Formatting and spelling fixes. 
** 005 SIY 20060128   @24125 Heavily rewritten to make layout manager 
**                           behavior identical to Progress.
** 006 SIY 20060203   @24240 Support for the TO and NO-LABELS options.
** 007 SIY 20060203   @24250 Added support for force-label flag.
** 008 SIY 20060207   @24440 Updated to support other widget types.
** 009 SIY 20060316   @25145 Added support for frame headers.
** 010 SIY 20060318   @25146 More fixes and improvements for the header 
**                           support.
** 011 SIY 20060401   @25309 Fixed multiline headers. 
** 012 SIY 20060403   @25355 Fix for right-aligned fill-in with wide top
**                           label.
** 013 SIY 20060410   @25464 Fixed stacking mode when widgets need to be 
**                           wrapped.
** 014 SIY 20060410   @25761 Fixed for class cast exception.
** 015 NVS 20060509   @26033 Widgets in down frames are placed on the same
**                           line using the calculated maximum height of
**                           the top labels.
** 016 NVS 20060510   @26055 Fixed frame size calculation regression after
**                           the fix #26033.
** 017 SIY 20060515   @26160 Fixed layout handling when AT phrase points 
**                           to the edge of the bounding rectangle. 
** 018 SIY 20060516   @26161 Added support for COLUMN-LABEL. 
** 019 NVS 20060522   @26403 Fixed some layout issues when using TO option
**                           of widget placing.
** 020 NVS 20060523   @26555 maxWidth didn't handle properly unassigned
**                           widget locations.
** 021 NVS 20060530   @26752 The case with colon set and labels disabled
**                           was handled incorrectly. Fixed.
** 022 NVS 20060531   @26769 Fixed some layout issues when using COLON
**                           option of widget placing.
** 023 SIY 20060616   @27342 Insert/remove labels into parent container
**                           rather than to one passed as a parameter.
**                           This is necessary to accommodate changes in
**                           layout handling in Frame.
**                           Do not underline header items if there are no
**                           other widgets in the container. 
** 024 SIY 20060616   @27368 Adding/removing labels into container now is
**                           safe again.
** 025 SIY 20060616   @27403 Cleaned up debugging output.
** 026 SIY 20060703   @27750 Fixed incorrect triggering on of stacked
**                           mode.
** 027 SIY 20060710   @27981 Fixed and significantly reworked handling of 
**                           ATTR-SPACE.
** 028 SIY 20060717   @28081 More advanced logic which determines
**                           delimiter width in stacked mode. 
** 029 NVS 20060728   @28217 Using new Container.replace() method to
**                           replace labels in down frames. This new call
**                           maintains the changes in the non-body list of
**                           down frames.
** 030 SIY 20060801   @28279 Fix for ArrayIndexOutOfBoundsException.
** 031 SIY 20060803   @28360 Stacking mode detection improved to better 
**                           match Progress behavior. 
** 032 SIY 20060810   @28537 Do not take into account labels of widgets
**                           which are explicitly marked as NO-LABEL.  
** 033 SIY 20060818   @28671 Handle spacing in containers with only header
**                           widgets as regular containers with no 
**                           headers.
** 034 SIY 20060824   @28835 Fixed support of TO option in stacking mode.
** 035 SIY 20060828   @28923 Fixed stacking mode check.
** 036 SIY 20060908   @29379 Fixed handling of headers in stacking mode.
** 037 NVS 20060918   @29645 Per widget ATTR-SPACE attributes are in
**                           effect when laying out widgets.
** 038 NVS 20060919   @29758 Fixed many bugs in layout processing for
**                           the stacked mode. Some changes are marked TBD
**                           to be reviewed later and for easier watching
**                           for possible regressions.
** 039 NVS 20060919   @29775 Fixed small regression visible as NPE in
**                           labelWidth() method.
** 040 NVS 20060920   @29827 Fixed a bug in frame width calculation that
**                           used to make the frame one character wider.
**                           The bug was there for both the stacked and
**                           regular layout.
** 041 NVS 20060921   @29840 Fixed a regression caused by 038. Headers in
**                           stacked mode were miscalculated.
** 042 NVS 20060922   @29864 Fixed a bug in header placement. Headers
**                           start at offset 0 regardless of ATTR-SPACE.
**                           Fixed delimiter size calculation in stacked
**                           mode.
** 043 NVS 20060925   @29884 Fixed a bug in widget placement when exactly
**                           on a line boundary.
** 044 SIY 20061004   @30171 Fixed stacking mode check for case when 
**                           COLUMN LABEL is present.
** 045 SIY 20061005   @30171 When taking into account container location,
**                           exclude frame border size, otherwise it will
**                           be subtracted twice.
** 046 GES 20061013   @30402 Mark the bogus label used as a delimiter with
**                           a flag to allow it to know that this is its
**                           intended use.  This changes drawing in the
**                           redirected output case.
** 047 SIY 20061018   @30500 Ignore trailing SKIP when detecting stacking 
**                           mode. Fixed handling of COLUMN-LABEL in
**                           stacking mode.
** 048 SIY 20061023   @30603 Fixed handling of empty labels and right 
**                           aligned widgets.
** 049 SIY 20061026   @30698 More fixes in stacking mode. 
** 050 SIY 20061102   @30898 Fixed handling of "headers only" frames and
**                           handling of the spacing in the headers.
** 051 NVS 20061108   @31006 Fixed regressions introduced with the fix 050
**                           in calcLayout() and processStacked() methods.
** 052 NVS 20061108   @31014 Fixed a bug in widget placement for tall
**                           widgets not fitting the current line.
** 053 SIY 20061113   @31099 Fixed double wrapping in some cases of
**                           regular layouts.
** 054 GES 20061114   @31128 Fix for incorrectly using stacked mode when
**                           a row is larger than the current screen width
**                           but smaller than the frame width.  Eliminate
**                           all traces of the delimiter when we are a
**                           frame with no body and are in stacked mode.
** 055 NVS 20061114   @31156 SKIP in headers had no effect on the headers
**                           placement, but affected the regular widgets.
** 056 NVS 20061115   @31174 Widgets in frames with headers and no body
**                           don't need extra spaces.
** 057 SIY 20061221   @31728 Fix for handling COLON and ATTR-SPACE.
** 058 SIY 20070214   @32130 Fixes for stacked mode handling (case when 
**                           some widgets have no labels and delimiter 
**                           maintenance).
** 059 SIY 20070323   @32588 Fixed handling of browse widgets.
** 060 SIY 20070709   @34427 Rolled back #036 since it causes regressions. 
**                           The issue for which #036 was created is fixed 
**                           with other changes.
** 061 SIY 20070807   @34813 Added support for widget positioning outside 
**                           the container size for redirected frames.
** 062 SIY 20070809   @34833 Fixed regression caused by previous update.
** 063 SIY 20070926   @35249 Fixed handling of header widgets.
**                           See: #35247 for more details.
** 064 SIY 20071004   @35367 Fixed widget placement in stacked mode.
** 065 SIY 20071102   @35683 SKIP widgets are placed just like others.
** 066 EVL 20071120   @35841 Stacked mode can not break the header line
**                           separator. Fixed.
** 067 SIY 20071123   @36008 Fixed regression caused by #066.
** 068 SIY 20071126   @36009 Changed placement of SKIP widgets.
**                           Ref: #36006
** 069 SIY 20071126   @36336 Rolled back #067 due to changed approach.
**                           Ref: #36027
** 070 SIY 20071212   @36337 Fixed handling of stacked mode with headers, 
**                           ATTR-SPACE and NO-LABELS. Deep refactoring of
**                           processStacked() method.
**                           Ref: #36338 
** 071 SIY 20071214   @36370 Fixed stacked mode detection for case when 
**                           frame has headers and NO-LABELS attribute set.
**                           Ref: #36371
** 072 SIY 20080107   @36726 Improved #071. Improved and fixed handling of
**                           ATTR-SPACE in stacked mode.
**                           Ref: #36711
** 073 SIY 20080114   @36801 Fixed handling of fixed widget position at the
**                           edge of the visible area in redirected mode.
**                           Ref: #36800
** 074 SIY 20080214   @37111 Use original (provided by application)
**                           frame width, rather than adjusted to match  
**                           screen width, because application can define  
**                           frame which is wider than screen.
**                           Ref: #37107
** 075 SIY 20080222   @37153 Fixed handling of the SPACE when it crosses
**                           maximal width.
**                           Ref: #33737
** 076 SIY 20080228   @37202 Javadoc cleanup.
** 077 SIY 20080307   @37394 Fixed handling of TO option in some cases.
**                           Ref: #37392.
** 078 RK  20080318   @37526 Implemented frame width adjustment behavior.
**                           Ref: #37529.
** 079 SIY 20080521   @38354 Refactored processRegular().
** 080 SIY 20080617   @38797 Fixed in ATTR-SPACE support.
** 081 SIY 20080623   @39038 Many fixes in frame width adjustment handling.
** 082 SIY 20080708   @39074 Fixed regression caused by #081 - vertical
**                           adjustment in some circumstances should be set
**                           to zero in order to achieve correct label 
**                           placement.
** 083 SIY 20080714   @39104 Quirks in label offset appear only when
**                           ATTR-SPACE is active.
** 084 SIY 20081028   @40260 Reimplemented vertical label placement in 
**                           stacking mode. New approach should cover all 
**                           possible situations properly.
**                           Ref: #40261
** 085 SIY 20090120   @41181 Improved detection of stacking mode for cases
**                           when labels are disabled at frame level but 
**                           explicitly enabled in widgets and in cases 
**                           when widget position is set via AT clause 
**                           rather than TO. Fixed label placement in some
**                           cases but there is at least one known issue 
**                           still left.  
** 086 SIY 20090520   @42415 Calculate current header height after checking for
**                           possible wrapping.
** 087 SIY 20090525   @42437 Fixed label placement in stacking mode. 
**                           Related: #30635,#41571.
** 088 GES 20090722   @43327 Converted a log message into a comment since it
**                           was cluttering the logs and we have not yet found
**                           any problem with the code.
** 089 SIY 20090806   @43539 Fixed handling of layout which starts with SKIP.
**                           Also, improved regular mode detections - leading
**                           SKIP should not force stacking mode. Fixed 
**                           placement of SKIP widgets in non-SIDELABELS mode.
** 090 SIY 20090811   @43593 Fixed regression caused by #089 - trailing SKIP 
**                           was placed outside the frame and not displayed
**                           in redirected mode.
** 091 SIY 20091104   @44314 Fixed widget width calculation in stacking mode 
**                           for cases when displaying of labels is disabled.
** 092 SIY 20091106   @44323 Fixed regression caused by #091.
** 093 EVL 20100428   @44836 Fix for incorrect explicit X-defined coordinates 
**                           widgets laying out.
** 094 SIY 20100619          Switching to own implementation of character UI.
** 095 SIY 20100804          Changes in imports, some refactorings.
** 096 SIY 20100822          Refactored AttributeContainer support.
** 097 SIY 20100831          Changes related to refactoring of FrameHelper.
** 098 SIY 20100915          locateFrame() moved to UiUtils.
** 099 EVL 20101222          Changes to implement the possibility to not display
**                           the label of the empty character based Text widget.
**                           This is the fix for harness baseline testcase.
**                           Modifications include:
**                           - adding character class declaration to be used
**                             inside
**                           - adding private method to check the Text widget
**                             if it is empty or not. The method name is:
**                             boolean isTextWidgetEmptyAndInvisible()
**                           - The method getLabel() is modified to have the
**                             empty labelText string in the cases when there
**                             is a headers, widget is Text instance, and
**                             data inside widget is character, empty and not
**                             visible.
** 100 GES 20110422          Fixed NPE in isTextWidgetEmptyAndInvisible() that
**                           occurred when the text widget held a character
**                           var that was unknown value.
** 101 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 102 SIY 20111115          Added support for widget physical location 
**                           calculation. Moved from CHUI package.
** 103 SIY 20111202          Intermediate version in preparation of introduction
**                           of widget factory.
** 104 SIY 20111221          Switched to widget factory.
** 105 EVL 20140528          Fix for displaying the labels for hidden widgets in a frame.  On the
**                           first time the label will not be drawn until the widget become
**                           visible.
** 105 CA  20140711          Font-related changes in preparation for GUI support.
** 106 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access 
**                           only to the base APIs provided by the OutputManager.
** 107 HC  20140911          GUI coordinate support - data types refactoring.
** 108 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 109 HC  20141021          Implemented window move/position support.
** 110 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 111 CA  20141205          Fixed a typo in RegularModeVariables.setNextColumn: it was using 
**                           height instead of width.
** 112 CA  20150114          Fixed support for changing the ROW/COLUMN attrs from server-side, 
**                           for a side-label.
** 113 HC  20150128          Added support for frame widget position changes induced by
**                           setting ROW/COL/X/Y attributes.
** 114 HC  20150214          Improved positioning of side labels.
** 115 CA  20150223          Fixed label's COLUMN and WIDTH-CHARS attributes (in ChUI).
** 116 CA  20150323          Pass the frame ID when creating a label (to properly resolve the font
**                           in GUI).
**                           Removed obsolete code related to widget font.
** 117 HC  20150323          Removed business logic from config classes.
** 118 CA  20150407          Changed to support widget layout in GUI.
** 119 VIG 20150508          Added ToggleBox support to getNumSpaces method.
** 120 CA  20150509          Allow scroll container for any widgets, not just frames.
** 121 CA  20150722          Layout is performed only once - removed code which addressed issues
**                           related to the fact that layout was performed on each 
**                           pushScreenDefinition call.
** 122 CA  20150725          Added FRAME-COL, FRAME-ROW, FRAME-X, FRAME-Y support: for down or top
**                           label frames, an offset from which the widget has been positioned 
**                           (relative to the frame's top label header) is maintained.  These 
**                           offsets are used so that the FRAME- attributes return the location 
**                           from the frame's top-left corner, while the X,Y,ROW,COL attributes
**                           return the location from their parent field-group.
** 123 HC  20150831          Scrolling support overhaul.
** 124 HC  20151013          Changes to support layout of frame in dialog-box.
** 125 CA  20151016          Fixed positioning of dynamic widgets attached to frame (which has 
**                           headers) BEFORE frame realization.  There is still an issue with 
**                           dynamic widgets attached AFTER frame realization.
** 126 IAS 20151021          Added support for the NO-UNDERLINE FRAME option.
** 127 HC  20151103          Fixed height calculation of widget with side label.
** 128 CA  20151123          To preserve z-order, side-labels need to be added just before their
**                           associated widget.
** 129 CA  20160327          Side-labels and top-header labels in the frame must be added to the 
**                           widget registry.
** 130 EVL 20160421          Space pseudo-widgets must increase the container width.  The last
**                           skip in a frame can be ignored.  Colon based alignment forces the
**                           widget to be shifted for length of the ": " string.
** 131 HC  20160519          Improved layout of COLON-ALIGNED widgets in GUI and ChUI. 
** 132 EVL 20160510          Fix for frame height calculation when content pane has multiple frame
**                           widgets inside.  In this case the maxRowHeight must not be set to the
**                           frame height walking through the widget list.  To minimize possible
**                           regressions only GUI mode is involved.
** 133 HC  20161018          Frame layout improvements.
** 134 OM  20161017          Fixed headers height calculation.
** 135 CA  20161125          COLON-ALIGNED option is possible for widgets which do not allow 
**                           side-labels (think TOGGLE-BOX).
** 136 OM  20161216          Fixed headers height calculation. Code maintenance.
** 137 EVL 20170201          Fixed the wrong placement for right aligned widgets.
**     CA  20161101          Some performance improvements: resolve the widget's width/height 
**                           only once, as in GUI these might be expensive.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * This is the LayoutManager which places the widgets into the container side
 * by side into the first available free space for the container's SIDE-LABELS
 * mode.
 */
public class ZeroColumnLayout<O extends OutputManager<?>>
implements LayoutManager<O>
{
   /** This field tracks container's SIDE-LABELS attribute. */
   private boolean sideLabels = false;
   
   /** This field tracks container's NO-LABELS attribute. */
   private boolean labels = false;
   
   /** This field tracks container's ATTR-SPACE attribute. */
   private boolean attr_space = false;
   
   /** This field tracks stacking mode flag. */
   private boolean stacking = false;
   
   /** This field tracks of headers. */
   private boolean hasHeaders = false;
   
   /** This field tracks the need to draw separator line under headers. */
   private boolean drawLine = false;
   
   /** Index of last header widget. */
   private int lastHeader = 0;
   
   /** Array of widgets in the container. */
   private Widget<O>[] c = null;
   
   /** Set to <code>false</code> if container has no components in body. */
   private boolean hasBody = true;
   
   /** Track current state of redirection. */
   private boolean redirected = false;
   
   /** Track explicit setting of container width. */
   private boolean fixedWidth = false;
   
   /** Widget factory */
   private WidgetFactory<O> factory = null;
   
   /** List of dynamic widgets in this frame. */
   private LinkedList<Widget<O>> dynWidgets;
   
   /**
    * Creates the new instance of the LayoutManager.
    */
   public ZeroColumnLayout()
   {
   }
   
   /**
    * Compute the label height in character units.  This is equal to summing the height of each 
    * label component, after splitting it using the new-line character.
    * <p>
    * If the label is underlined, 1 character unit is added. If the label is on top (ie neither
    * NO-LABELS or SIDE-LABELS options were specified) at least one character unit is reserved.
    *
    * @param    widget
    *           The widget to which the label belongs.
    * @param    labelText
    *           The label text.
    * @param    label
    *           The label instance.  May be <code>null</code> if this is not yet resolved.
    * @param    topLabel
    *           Flag identifying that the label is shown above the widget.
    *
    * @return   See above.
    */
   public static double labelHeight(LabeledWidget<?> widget,
                                    String           labelText,
                                    Label<?>         label,
                                    boolean          topLabel)
   {
      if (labelText == null)
      {
         // if the labels are on top, we reserve the current line
         return topLabel ? 1 : 0;
      }
      
      Frame<?> frame = ThinClient.getInstance().getFrame(widget.config().frameId);
      double height = 0;
      
      if (labelText.indexOf('\n') < 0)
      {
         height = frame.getLabelHeight(labelText, (LabeledWidget) widget, topLabel);
      }
      else
      {
         String[] arr = labelText.split("\n");
         for (String tok : arr)
         {
            height += frame.getLabelHeight(tok, (LabeledWidget) widget, topLabel);
         }
      }
      
      
      return height + (label != null && label.isUnderlined() ? 1 :
             ThinClient.getInstance().isChui() ? 0 : 0.5);
   }
   
   /**
    * Compute the label width in character units, depending on the current mode (GUI or ChUI)
    * and the GUI font.
    *
    * @param    widget
    *           The widget to which the label belongs.
    * @param    labelText
    *           The label text.
    *
    * @return   See above.
    */
   public static double labelWidth(LabeledWidget<?> widget, String labelText)
   {
      if (labelText == null)
         return 0;
      
      String res = null;
      
      if (labelText.indexOf('\n') < 0)
      {
         res = labelText;
      }
      else
      {
         String[] arr = labelText.split("\n");
         for (String tok : arr)
         {
            if (res == null || tok.length() > res.length())
            {
               res = tok;
            }
         }
      }
      
      Frame<?> frame = ThinClient.getInstance().getFrame(widget.config().frameId);
      
      double width = frame.getLabelWidth(res, (LabeledWidget) widget);
      
      return width;
   }
   
   /**
    * Makes the widgets coordinates calculation based on the current container
    * size.
    *
    * @param   container
    *          The container to which this LayoutManager should be applied.
    */
   @Override
   public void doLayout(Container<O> container)
   {
      calcLayout(container, container.width(), true);
   }
   
   /**
    * Calculates the minimum size of the container which will be big enough to
    * hold all widgets. For this layout the widgets location should be
    * adjusted to the container size, so this method just return the current
    * container size.
    *
    * @param   container
    *          The container which size should be calculated.
    *
    * @return  The size of the container which is big enough to hold all
    *          widgets.
    */
   @Override
   public Dimension minimumSize(Container<O> container)
   {
      // TODO: only if layout was not already performed
      double containerWidth;
      
      // safe, comparison of scaled doubles
      double cwidth = container.width();
      if (cwidth > 0)
         containerWidth = cwidth;
      else
         containerWidth = maxWidth(container);
      
      return calcLayout(container, containerWidth, false);
   }
   
   /**
    * Parse label text and translate exclamation marks into the new line
    * characters using Progress rules.
    *
    * @param   columnLabelText
    *          Source label text to convert.
    *
    * @return  The result of parsing.
    */
   private static String parseColumnLabel(String columnLabelText)
   {
      if (columnLabelText == null)
         return null;
      
      char[] list = columnLabelText.toCharArray();
      StringBuilder buf = new StringBuilder();
      
      for (int i = 0; i < list.length; i++)
      {
         char ch = list[i];
         
         if (ch == '!')
         {
            if ((i + 1) < list.length && list[i + 1] == '!')
            {
               // skip next exclamation mark to prevent translation
               i++;
            }
            else
               ch = '\n';
         }
         
         buf.append(ch);
      }
      
      return buf.toString();
   }
   
   /**
    * Perform actual formatting or size calculation.
    *
    * @param   container
    *          Container to work with.
    * @param   width
    *          Maximal width of container or screen.
    * @param   doPlacement
    *          Perform actual placement of widgets or just simulate work.
    *
    * @return  Space used by layout.
    */
   private Dimension calcLayout(Container<O> container, double width, boolean doPlacement)
   {
      width = Coordinate.scale(width);
      
      Frame<O> frame = (Frame<O>) UiUtils.locateFrame(container);
      if (frame != null)
      {
         sideLabels = frame.isSideLabels();
         labels = frame.isLabels();
         attr_space = frame.isAttrSpace();
      } // TODO: else we get a NPE ?
      
      c = container.widgets();
      
      // remove the dynamic widgets
      List<Widget<O>> widgets = new LinkedList<>(Arrays.asList(c));
      dynWidgets = new LinkedList<>();
      
      Iterator<Widget<O>> itrw = widgets.iterator();
      while (itrw.hasNext())
      {
         Widget<O> w = itrw.next();
         WidgetConfig wcfg = w.config();
         if (wcfg instanceof BaseConfig)
         {
            BaseConfig bc = (BaseConfig) wcfg;
            if (bc.dynamic)
            {
               dynWidgets.add(w);
               itrw.remove();
            }
         }
      }
      c = widgets.toArray(new Widget[0]);
      
      factory = container.screen().getFactory();
      stacking = false;
      
      hasHeaders = frame.hasHeaders();
      hasBody = true;
      drawLine = !labels || sideLabels;
      lastHeader = 0;
      
      int numNonHeaders = 0;
      double maxlen = 0;
      int ndx = -1;
      
      boolean forcelabel = false;
      
      // determine presence of headers and drawing mode
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Label)
            continue;
         
         if (UiUtils.hasConfig(c[i]))
         {
            if (c[i].isHeader())
            {
               lastHeader = i;
               hasHeaders = true;
            }
            else
            {
               if (c[i] instanceof LabeledWidget)
                  numNonHeaders++;
            }
         }
         
         if (c[i] instanceof FillIn)
         {
            FillIn<O,?> fin = (FillIn<O,?>) c[i];
            
            if (fin.isForceLabel())
            {
               forcelabel = true;
               drawLine = false;
            }
            
            // safe, comparison of scaled doubles
            double finWidth = fin.width();
            if (maxlen < finWidth)
            {
               maxlen = finWidth;
               ndx = i;
            }
         }
      }
      
      if (hasHeaders && numNonHeaders == 0)
         hasBody = false;
      
      int numLabels = determineMode(width, maxWidth(container), forcelabel);
      
      if ((numNonHeaders == 0 && drawLine) || (stacking && !hasHeaders))
         drawLine = false;
      
      if (hasHeaders && hasBody && !stacking)
      {
         int i;
         
         for (i = lastHeader; i >= 0; i--)
         {
            if (!(c[i] instanceof FillIn))
               break;
            
            ((FillIn<O,?>) c[i]).setHeaderParm(drawLine, 0);
         }
         
         if (i >= 0)
         {
            if (ndx >= 0)
            {
               FillIn<O,?> fin = (FillIn<O,?>) c[ndx];
               fin.setHeaderParm(fin.isDrawLine(), (attr_space) ? 2:0);
            }
         }
         else
         {
            for (i = 0; i < lastHeader; i++)
            {
               if (!(c[i] instanceof FillIn))
                  break;
               
               FillIn<O,?> fin = (FillIn<O,?>) c[ndx];
               fin.setHeaderParm(fin.isDrawLine(), 1);
            }
         }
      }
      
      Dimension sz = null;
      
      redirected  = ThinClient.getInstance().isRedirected();
      
      if (stacking && (numLabels > 0 || (hasHeaders && hasBody && !labels)))
         sz = processStacked(container, width, doPlacement);
      else
         sz = processRegular(container, width, doPlacement);
      
      if (hasHeaders && hasBody && doPlacement)
      {
         // safe, comparison of scaled doubles
         double frameWidth = (fixedWidth) ? Math.max(width, sz.width) : sz.width;
         
         // safe, comparison of scaled doubles
         if (c[lastHeader] instanceof FillIn && c[lastHeader].location().x == 0)
         {
            FillIn<O,?> fin = (FillIn<O,?>) c[lastHeader];
            
            // safe, comparison of scaled doubles
            double finWidth = fin.width();
            if (finWidth < frameWidth)
            {
               fin.setHeaderParm(drawLine, 0);
               fin.setHeaderParm(drawLine, frameWidth - finWidth);
            }
         }
      }
      
      return Coordinate.scale(sz);
   }
   
   /**
    * Determine layout mode (stacked or regular). 
    *
    * @param   width
    *          Maximal width of container or screen.
    * @param   maxWidth
    *          Maximal container width which can fit into screen.
    * @param   forcelabel
    *          <code>true</code> if any widgets in container has forced label.
    *
    * @return  number of labels in container.
    */
   private int determineMode(double width, double maxWidth, boolean forcelabel)
   {
      width = Coordinate.scale(width);
      maxWidth = Coordinate.scale(maxWidth);
      
      int numLabels = 0;
      double hsize = 0;
      boolean pendingStacking = (hasHeaders && !labels);
      boolean canBeStacked = (!sideLabels && (labels || forcelabel)) || (hasHeaders && !labels);
      int labeledCount = 0;
      
      // determine mode
      for (int i = 0; canBeStacked && i < c.length; i++)
      {
         // header widgets are not included into mode detection
         if (hasHeaders && i <= (lastHeader + 1))
            continue;
         
         if (c[i] instanceof Label)
            continue;
         
         if (c[i] instanceof Skip)
         {
            hsize = 0;
            
            if (labeledCount > 0)
               pendingStacking = true;
            continue;
         }
         
         if (c[i] instanceof LabeledWidget)
         {
            double cwidth = c[i].width();

            labeledCount++;
            
            if (pendingStacking)
            {
               stacking = true;
               pendingStacking = false;
            }
            
            LabeledWidget<O> lw = (LabeledWidget<O>) c[i];
            
            // safe, comparison of scaled doubles
            if (lw.getTo() != BaseConfig.INV_COORD)
            {
               if (lw.getTo() < hsize)
                  stacking = true;
               
               CoordinatesConversion cc = c[i].screen().coordinates();
               hsize = cc.widthFromNative(cc.widthToNative(lw.getTo())   -
                                          cc.widthToNative(cwidth) + 
                                          1);
            }
            
            BaseConfig bc = c[i].config();
            
            // safe, comparison of scaled doubles
            if (bc.initColumn != BaseConfig.INV_COORD)
            {
               if (bc.initColumn < hsize)
                  stacking = true;
               
               hsize = bc.initColumn;
            }
            
            hsize = Coordinate.scale(hsize + cwidth);
            
            String text = null;
            
            text = getLabel(lw);
            
            if (text != null && (!lw.isNoLabels() || (!labels && lw.isForceLabel())))
            {
               double len = labelWidth(lw, text);
               
               // safe, comparison of scaled doubles
               if (len > 0)
                  numLabels++;
               
               // safe, comparison of scaled doubles
               if (len > cwidth)
                  hsize = Coordinate.scale(hsize + len - cwidth);
            }
            
            if ((i + 1) < c.length)
            {
               BaseConfig bcNext = UiUtils.getConfig(c[i + 1], BaseConfig.class);
               
               if (bcNext != null)
               {
                  double nextColumn = bcNext.initColumn;
                  double nextRow = bcNext.initRow;
                  
                  // safe, comparison of scaled doubles
                  if (nextColumn == BaseConfig.INV_COORD ||
                      (nextRow == BaseConfig.INV_COORD && nextColumn > hsize))
                  {
                     hsize = Coordinate.scale(hsize + getExtraSpacesStacked(i));
                  }
               }
            }
            
            // safe, comparison of scaled doubles
            if (hsize > maxWidth)
               stacking = true;
            
            // safe, comparison of scaled doubles
            if (width > 0 && hsize > width)
               stacking = true;
         }
      }
      return numLabels;
   }
   
   /**
    * Handle "stacked" case of the widget placement. In this case screen is
    * split into two parts: in upper part placed all labels while in bottom -
    * all widgets.
    *
    * @param   container
    *          Container to work with.
    * @param   width
    *          Maximal width of container or screen.
    * @param   doPlacement
    *          Perform actual placement of widgets or just simulate work.
    *
    * @return  Space used by layout.
    */
   private Dimension processStacked(Container<O> container, double width, boolean doPlacement)
   {
      width = Coordinate.scale(width);
      
      double origWidth = width;
      
      Dimension res    = new Dimension(0, 0);
      Dimension top    = new Dimension(0, 0);
      Dimension bottom = new Dimension(0, 0);
      
      StackedModeVariables vars = new StackedModeVariables(width);
      
      // this is used to draw the horizontal line below headers, it is a
      // bogus widget with no text but with underlining
      Label<O> delimiter = null;
      
      // first pass, calculate size
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Label)
         {
            if (((Label<O>) c[i]).isDelimiter())
            {
               delimiter = (Label<O>) c[i];
            }
            
            continue;
         }
         
         double len = 0; // total width = MAX(widget's length, label's length) to be computed
         double wl = c[i].width(); // the widget's length 
         
         if (c[i] instanceof LabeledWidget)
         {
            LabeledWidget<O> lw = (LabeledWidget<O>) c[i];
            
            len = labelWidth(lw, getLabel(lw));
            
            // safe, comparison of scaled doubles
            if (len < wl)
               len = wl; // adjusting the total length, taking into account the widget's length
         }
         
         if (vars.isSkipOrWrap(i, len))
         {
            vars.finalizeHeight();
            
            if (c[i] instanceof Skip)
            {
               vars.processSkip(i);
               continue;
            }
         }
         
         if (c[i] instanceof Space)
         {
            vars.processSpace(i);
            continue;
         }
         
         BaseConfig bc     = c[i].config();
         Point location    = ConfigHelper.getInitialLocation(bc);
         String labelText  = null;
         LabeledWidget<O> fin = null;
         boolean doLabel   = false;
         boolean isHdr     = isHeader(i);
         
         double ll = 0;
         
         if (c[i] instanceof LabeledWidget)
         {
            fin = (LabeledWidget<O>) c[i];
            
            labelText = getLabel(fin);
            
            ll = labelWidth(fin, labelText);
            
            if (isHdr)
            {
               labelText = ((FillIn<O,?>) c[i]).getText();
               ll = labelWidth(fin, labelText);
               doLabel = true;
               
               vars.resetHeaderStart();
            }
            else
            {
               doLabel = labels || fin.isForceLabel();
            }
         }
         
         len = 0;
         int moved = 0;
         
         if (fin != null && fin.getTo() != BaseConfig.INV_COORD)
         {
            // simulate setting column
            CoordinatesConversion cc = c[i].screen().coordinates();
            location.x = cc.widthFromNative(cc.widthToNative(fin.getTo()) -
                                            cc.widthToNative(wl)          +
                                            1);
         }
         
         // if there is a label - account it in the top part
         if (doLabel)
         {
            if (fin.getColon() != BaseConfig.INV_COORD)
            {
               double sideLabelSuffixLength = labelWidth(fin, ": ");
               // simulate setting column
               location.x = fin.getColon() + sideLabelSuffixLength;
            }
            
            vars.processFixedTop(location);
            
            double labelHeight = labelHeight(fin, labelText, fin.getLabelInstance(), !sideLabels);
            if (vars.topCurrentHeight < labelHeight)
            {
               vars.topCurrentHeight = labelHeight;
            }
            
            // len = MAX(ll, wl), safe: comparison of scaled doubles
            len = (ll > wl) ? ll : wl;
            
            if (fin.isRightAligned())
            {
               // safe, comparison of scaled doubles
               if (ll > wl)
                  moved = 1; // adjust also fill-in position
               else
                  len = wl; // NOTE: om: this does not make sense, len = wl already, in this case
            }
            
            vars.processTop(i, len, top);
         }
         else
            len = wl;
         
         // calculate regular widgets and fill-in bodies (bottom part).
         if (!isHdr)
         {
            vars.processFixedBottom(location);
            vars.processBottom(i, len, moved, false, 0);
         }
         
         vars.adjustDimensions(top, bottom); // TODO: om: sync vars.nextTopCol with vars.nextBotomCol (=MAX of them)
      }
      
      // calculate summaries
      vars.finalizeHeight();
      
      vars.adjustDimensions(top, bottom);
      
      // adjust width for ATTR-SPACE
      
      // top is not affected by ATTR-SPACE
      //top.width += (attr_space) ? 1 : 0;
      bottom.width += (attr_space) ? 1 : 0;
      
      // only add an extra row in if we have a delimiter (which only occurs
      // if we have a frame body)
      // this is correct in GUI too, as the delimiter's height is always 1 char tall (regardless 
      // of the frame's font)
      res.height = Coordinate.scale(top.height + bottom.height + (hasBody ? 1 : 0));
      res.width  = top.width;
      
      // safe, comparison of scaled doubles
      if (bottom.width > res.width)
      {
         res.width = bottom.width;
      }
      
      if (!doPlacement)
      {
         return res;
      }
      
      // delimiters are only valid when there is a frame body
      if (hasBody)
      {
         if (delimiter == null)
         {
            int fid = UiUtils.getWidgetIdAsInt(UiUtils.locateFrame(container));
            delimiter = factory.createLabel(WidgetId.nextID(), "", fid);
            delimiter.setUnderlined(true);
            delimiter.setDelimiter(true);
            // force the height to be fixed.
            delimiter.config().fixedHeight = true;
            // the height is correct in GUI too
            delimiter.setHeight(1);
            container.add(delimiter);
            
            OutputManager.instance().getRegistry().addWidget((Widget) delimiter);
         }
         
         // safe, comparison of scaled doubles
         if (maxWidth(container) >= origWidth) // width is set explicitly
            delimiter.setWidth(origWidth);
         else
            delimiter.setWidth(res.width);
         delimiter.setLocation(0, vars.nextTopRow);
         delimiter.setPhysicalLocation(0, vars.nextTopY);
         delimiter.config.widgetPlaced = true;
      }
      
      double delimRow = vars.nextTopRow;
      
      vars.clear();
      
      // second pass, calculate/relocate
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Label)
         {
            continue;
         }
         
         double len = 0;
         double wl = c[i].width();
         
         if (c[i] instanceof LabeledWidget)
         {
            LabeledWidget<O> lw = (LabeledWidget<O>) c[i];
            
            len = labelWidth(lw, getLabel(lw));
            
            // safe, comparison of scaled doubles
            if (len < wl)
               len = wl;
         }
         
         if (vars.isSkipOrWrap(i, len))
         {
            vars.fixLabelsRow();
            vars.finalizeHeight();
            
            if (c[i] instanceof Skip)
            {
               c[i].setLocation(0, vars.nextBottomRow);
               c[i].setPhysicalLocation(0, vars.nextBottomY);
               
               vars.processSkip(i);
               continue;
            }
         }
         
         if (c[i] instanceof Space)
         {
            vars.processSpace(i);
            continue;
         }
         
         BaseConfig bc     = c[i].config();
         Point location    = ConfigHelper.getInitialLocation(bc);
         String labelText  = null;
         LabeledWidget<O> fin = null;
         boolean doLabel   = false;
         double labelLen   = 0;
         
         if (c[i] instanceof LabeledWidget)
         {
            fin = (LabeledWidget<O>) c[i];
            
            if (isHeader(i))
            {
               labelText = ((DataContainer) c[i]).getText();
               labelLen  = labelWidth(fin, labelText);
               doLabel   = true;
               
               vars.resetHeaderStart();
            }
            else
            {
               labelText = getLabel(fin);
               labelLen  = labelWidth(fin, labelText);
               doLabel   = labels || fin.isForceLabel();
               
               // safe, comparison of scaled doubles
               if (!doLabel && len > wl)
                  len = wl;
            }
         }
         
         if (fin != null && fin.getTo() != BaseConfig.INV_COORD)
         {
            // simulate setting column
            CoordinatesConversion cc = c[i].screen().coordinates();
            location.x = cc.widthFromNative(cc.widthToNative(fin.getTo()) -
                                            cc.widthToNative(len)         +
                                            1);
         }
         
         len = 0;
         int moved = 0;
         
         // if there is a label - account it in the top part
         
         if (doLabel)
         {
            // safe, comparison of scaled doubles
            if (fin.getColon() != BaseConfig.INV_COORD)
            {
               double sideLabelSuffixLength = labelWidth(fin, ": ");
               // simulate setting column
               location.x = fin.getColon() + sideLabelSuffixLength;
            }
            
            // len = MAX(labelLen, wl); safe, comparison of scaled doubles
            len = (labelLen > wl) ? labelLen : wl;
            
            vars.processFixed(i, location, len, labelText);
            
            if (fin.isRightAligned())
            {
               // safe, comparison of scaled doubles
               if (labelLen > wl)
                  moved = 1; // adjust also fill-in position
               else
                  len = wl;
            }
            
            if (!isHeader(i))
            {
               // place the label
               Label<O> old = fin.getLabelInstance();
               WidgetId nextId = c[i].config().id.createId(WidgetId.nextID());
               Label<O> label = factory.createLabel(nextId, labelText, c[i].config().frameId);
               
               label.setRightAligned(fin.isRightAligned());
               label.setWidth(len);
               
               label.setLocation(vars.nextTopCol, vars.nextTopRow);
               label.setPhysicalLocation(vars.nextTopX, vars.nextTopY);
               fin.setLabelInstance(label);
               
               if (old == null)
               {
                  container.add(c[i], label);
               }
               else
               {
                  // remove it from the registry
                  OutputManager.instance().getRegistry().removeWidget(old.getId());
                  
                  container.replace(old, label);
               }
               
               // add it to the registry
               OutputManager.instance().getRegistry().addWidget((Widget) label);
               
               vars.addLabel(label, fin);
            }
            else
            {
               c[i].setLocation(vars.nextTopCol, vars.nextTopRow);
               c[i].setPhysicalLocation(vars.nextTopX, vars.nextTopY);
            }
            
            vars.processTop(i, len, top);
         }
         else
            len = wl;
         
         if (!isHeader(i))
         {
            // calculate regular widgets and fill-in bodies (bottom part).
            // safe, comparison of scaled doubles
            if (location.x != BaseConfig.INV_COORD)
            {
               if (!doLabel &&
                   Coordinate.scale(vars.nextBottomCol - vars.extraBottomSpaces) > location.x)
               {
                  vars.incrementNextBottomRow(vars.bottomCurrentHeight);
                  vars.bottomCurrentHeight = 0;
               }
               
               vars.setNextBottomCol(location.x);
            }
            
            // safe, comparison of scaled doubles
            if (location.y != BaseConfig.INV_COORD)
            {
               // delimiter height is always 1 (even in GUI), so this is OK
               vars.incrementNextBottomRow((location.y + delimRow + 1) - vars.nextBottomRow);
            }
            // doPlacement is always true here
            vars.processBottom(i, len, moved, doPlacement, labelLen - wl);
         }
      }
      
      return res;
   }
   
   /**
    * Get per-widget ATTR-SPACE attribute.
    *
    * @param   i
    *          Index of widget in container widget array.
    *
    * @return  The current value of the attribute.
    */
   private boolean isAttrSpace(int i)
   {
      ControlConfig cfg = UiUtils.getConfig(c[i], ControlConfig.class);
      return cfg != null && cfg.attrSpace;
   }
   
   /**
    * Get per-widget HEADER attribute.
    *
    * @param   i
    *          Index of widget in container widget array.
    *
    * @return  The current value of the attribute.
    */
   private boolean isHeader(int i)
   {
      ControlConfig cfg = UiUtils.getConfig(c[i], ControlConfig.class);
      return cfg != null && c[i].isHeader();
   }
   
   /**
    * Calculate number of required spaces between widgets depending on current
    * conditions.
    *
    * @param   i
    *          Index of widget in container widget array.
    *
    * @return  The number of spaces required.
    */
   private double getExtraSpacesStacked(int i)
   {
      return ((attr_space || isAttrSpace(i)) && (labels || hasHeaders)) ? 2 : 1;
   }
   
   /**
    * Check if particular widget has defined column either by specifying
    * column or by setting TO or COLON option.
    *
    * @param   comp
    *          The widget to check.
    *
    * @return  <code>true</code> if widget has defined column.
    */
   private boolean hasDefinedX(Widget<O> comp)
   {
      boolean res = false;
      
      WidgetConfig wc = comp.config();
      if (!(wc instanceof BaseConfig))
      {
         return false;
      }
      
      BaseConfig bc = (BaseConfig) wc;
      
      // safe, comparison of scaled doubles
      if (ConfigHelper.getAssignedColumn(bc) != BaseConfig.INV_COORD)
         res = true;
      
      if (comp instanceof LabeledWidget)
      {
         LabeledWidget<O> fin = (LabeledWidget<O>) comp;
         
         // safe, comparison of scaled doubles
         if (fin.getTo() != BaseConfig.INV_COORD || fin.getColon() != BaseConfig.INV_COORD)
         {
            res = true;
         }
      }
      
      return res;
   }
   
   /**
    * Get label taking into account various conditions and situations.
    *
    * @param    fin
    *           Source widget.
    * @param    sideLabels
    *           Flag indicating if the frame uses side-labels.
    * @param    hasHeaders
    *           Flag indicating if the frame has header widgets.
    *
    * @return   The label assigned to widget.
    */
   public static String getLabel(LabeledWidget<?> fin, boolean sideLabels, boolean hasHeaders)
   {
      if (fin == null)
         return "";
      
      String labelText = null;
      
      if (!sideLabels)
      {
         labelText = parseColumnLabel(fin.getColumnLabelText());
         
         // safe, comparison of scaled doubles
         if (labelText != null && labelWidth(fin, labelText) == 0)
            labelText = null;
      }
      
      if (hasHeaders && fin instanceof Text && isTextWidgetEmptyAndInvisible((Text<?>)fin))
      {
         // headers with empty widget show no label
         
         // TODO: in 4GL, no label is shown if the widget is not empty&invisible, but P2J shows a 
         // label.  maybe "isTextWidgetEmptyAndInvisible" needs to be removed?
         
         labelText = null;
      }
      else if (labelText == null)
      {
         labelText = fin.getLabelText();
      }
      
      // label is either set (when is non-null), or not; empty label is possible in 4GL, so never
      // assume that if there is an empty label string, in fact there is no label. 
      
      if (fin.isNoLabels())
      {
         labelText = null;
      }
      
      return labelText;
   }
   
   /**
    * Get label taking into account various conditions and situations.
    *
    * @param   fin
    *          Source widget.
    *
    * @return  The label assigned to widget.
    */
   private String getLabel(LabeledWidget<O> fin)
   {
      return getLabel(fin, sideLabels, hasHeaders);
   }
   
   /**
    * Check if specified Text widget is BOTH empty AND invisible.
    *
    * @param   textWidget
    *          Widget to check.  Must NOT be <code>null</code>.
    *
    * @return  <code>true</code> if widget is empty and invisible.
    */
   private static boolean isTextWidgetEmptyAndInvisible(Text<?> textWidget)
   {
      BaseDataType val   = textWidget.getValue();
      boolean      empty = false;
      
      if (val != null && val instanceof character)
      {
         character charVal = (character) val;
         
         if (charVal.isUnknown())
         {
            empty = true;
         }
         else
         {
            String txt = charVal.getValue();
            
            if (txt.length() == 0)
               empty = true;
         }
      }
      
      return (val == null || empty) && !textWidget.isVisible();
   }
   
   /**
    * Handle regular case of the widget placement. 
    *
    * @param   container
    *          Container to work with.
    * @param   width
    *          Maximal width of container or screen.
    * @param   doPlacement
    *          Perform actual placement of widgets or just simulate work.
    *
    * @return  Space used by layout.
    */
   private Dimension processRegular(Container<O> container, double width, boolean doPlacement)
   {
      width = Coordinate.scale(width);
      
      RegularModeVariables vars = new RegularModeVariables(width, doPlacement);
      
      vars.calculateHeaderHeight();
      
      for (int i = 0; i < c.length; i++)
      {
         vars.startColumn = 0;
         
         if (c[i] instanceof Label && ((Label) c[i]).isDelimiter())
         {
            container.remove(c[i]);
            continue;
         }
         
         if (hasHeaders && i <= (lastHeader + 1))
         {
            // safe, comparison of scaled doubles
            if (vars.hdrWidth < vars.nextColumn)
               vars.hdrWidth = vars.nextColumn;
         }
         
         if (c[i] instanceof Label)
            continue;
         
         vars.initWidgetLocation(i);
         
         vars.adjustDimensions();
         
         if (vars.isSpaceHandled(i))
         {
            vars.setWidgetLocation(i);
            continue;
         }
         
         vars.setStartColumn(i);
         
         vars.handleInitColumn(i);
         
         vars.handleInitRow(i);
         
         vars.checkForFixup(i);
         
         vars.adjustDimensions();
         
         vars.handleTO(i);
         
         if (isLabeled(i))
         {
            vars.handleLabeledWidget(i, container);
         }
         else
         {
            vars.handleNoLabelWidget(i);
         }
         
         vars.calcQuirksOffset(i);
         
         vars.setWidgetLocation(i);
      }
      
      if (doPlacement && Double.compare(vars.distanceY, 0.0d) != 0)
      {
         // for all dynamic widgets, set the offset (if there is one) and adjust the location
         for (Widget<O> wdyn : dynWidgets)
         {
            vars.setWidgetLocation(wdyn);
         }
      }
      
      Frame<O> frame = (Frame<O>) UiUtils.locateFrame(container);
      if (frame != null)
      {
         FrameConfig fcfg = frame.config();
         fcfg.headerHeightChars = (sideLabels ? 0 : vars.distanceY);
      }
      
      return vars.finalAdjustments();
   }
   
   /**
    * Check if specified widget should be handles as labeled.
    *
    * @param   i
    *          Widget index.
    *
    * @return  <code>true</code> if widget should be handled as labeled.
    */
   private boolean isLabeled(int i)
   {
      LabeledWidget<O> fin = null;
      
      if (c[i] instanceof LabeledWidget)
         fin = (LabeledWidget<O>) c[i];
      else
         return false;
      
      return getLabel(fin) != null && (labels || fin.isForceLabel()) && !fin.isNoLabels();
   }
   
   /**
    * Get number of spaces which need to be inserted before given component.
    *
    * @param   i
    *          Index of component to check.
    *
    * @return  The number of spaces to insert.
    */
   private double getNumSpaces(int i)
   {
      if (hasBody  && i <= lastHeader)
         return 0;
      
      if (i >= c.length)
         return 0;
      
      if (!(c[i] instanceof LabeledWidget || c[i] instanceof Browse || c[i] instanceof ToggleBox))
         return 0;
      
      if (c[i] instanceof LabeledWidget)
      {
         double to = ((LabeledWidget<O>) c[i]).getTo();
         double co = ((LabeledWidget<O>) c[i]).getColon();
         
         // safe, comparison of scaled doubles
         if (to != BaseConfig.INV_COORD || co != BaseConfig.INV_COORD)
            return 0;
      }
      
      boolean wAttrSpace = false;
      if (UiUtils.hasConfig(c[i - 1]))
      {
         ControlConfig cfg = UiUtils.getConfig(c[i - 1], ControlConfig.class);
         if (cfg != null)
         {
            wAttrSpace = cfg.attrSpace;
         }
      }
      
      BaseConfig bc = c[i].config();
      
      // safe, comparison of scaled doubles
      if (ConfigHelper.getAssignedColumn(bc) == BaseConfig.INV_COORD)
      {
         if (isStatic(i - 1) || (i > 0 && !(c[i - 1] instanceof FillIn)))
         {
            return 1;
         }
         
         return (attr_space || wAttrSpace) ? 2 : 1;
      }
      
      return 0;
   }
   
   /**
    * Check if given widget is static.
    *
    * @param   i
    *          Index of the widget to check.
    *
    * @return  <code>true</code> if widget is marked as static.
    */
   private boolean isStatic(int i)
   {
      if (i >= 0 && i < c.length && c[i] instanceof FillIn)
      {
         FillInConfig cfg = UiUtils.getConfig(c[i], FillInConfig.class);
         
         return cfg.flagStatic;
      }
      return false;
   }
   
   /**
    * Calculate maximal container width which can fit into screen.
    *
    * @param   container
    *          Container for which maximal width is calculated.
    *
    * @return  Maximal container width.
    *
    */
   private double maxWidth(Container<O> container)
   {
      Widget<O> comp = (Widget<O>) UiUtils.locateFrame(container);
      
      double maxwidth = container.topLevelWindow().screenWidth();
      
      if (comp instanceof Frame)
      {
         // GUI dialog-box can use the whole display work area
         if (((Frame<?>) comp).isDialog() && !comp.screen().isChui())
         {
            GuiDriver gd = (GuiDriver) comp.screen().getInstanceDriver();
            int topWnd = ((Frame<?>) comp).topLevelWindow().getId().asInt();
            
            // TODO: the area size doesn't seem to be right
            NativeRectangle area = gd.getDisplayWorkArea(topWnd);
            
            maxwidth = comp.screen().coordinates().widthFromNative(area.width());
         }
         
         FrameConfig fc = UiUtils.getConfig(comp, FrameConfig.class);
         Frame<O> frame = (Frame<O>) comp;
         
         fixedWidth  = frame.isFixedWidth();
         
         double borderDelta = frame.calculateWidthDelta();
         
         double fwid = frame.getSavedDimension().width;
         
         // honor the larger of the currently installed "terminal" OR any
         // explicitly set frame width
         // safe, comparison of scaled doubles
         if (fwid < maxwidth)
         {
            if (!fc.centered)
            {
               double loc = Coordinate.scale(container.screenLocation().x -
                                             frame.calculateWidthDelta() / 2);
               
               if (loc > 0)
                  maxwidth = Coordinate.scale(maxwidth - loc);
            }
         }
         else
            maxwidth = fwid;
         
         maxwidth = Coordinate.scale(maxwidth - borderDelta);
      }
      
      return maxwidth;
   }
   
   /**
    * Returns the coordinates conversion implementation.
    *
    * @return   a reference to coordinates conversion implementation.
    */
   private CoordinatesConversion coordinates()
   {
      return OutputManager.instance().coordinates();
   }
   
   /**
    * Regular mode variables bundle.
    */
   class RegularModeVariables
   {
      /** Initial column for new row. */
      double startColumn = 0;
      /** Next free column. */
      double nextColumn = 0;
      /** Header width. */
      double hdrWidth = 0;
      /** Next free row. */
      double nextRow = 0;
      /** Maximal height of widget in current row of widgets. */
      double maxRowHeight = 0;
      /** Maximal height. */
      double maxHeight = 0;
      /** Adjustment for the labels in non-SIDE-LABELS mode. */
      double distanceY = 0;
      /** Index of last leading SKIP widget. */
      int leadingSkips = 0;
      /** Number of spaces to insert between widgets. */
      double extraSpaces = 0;
      /** Maximal layout width. */
      private double width = 0;
      
      /** Local widget ATTR-SPACE. */
      boolean wAttrSpace = false;
      /** Y-coordinate set flag. */
      boolean ySet = false;
      /** X-coordinate set flag. */
      boolean xSet = false;
      /** Perform fixup. */
      boolean doFixup = false;
      /** Perform fixup of next row. */
      boolean fixupNextRow = false;
      /** Flag which controls widget placement. */
      private boolean doPlacement = false;
      
      /** Extra width caused by widget placement quirk. */
      private double deltaWidth = 0;
      
      /** Index of widget which caused adjustment of currentHeight variable. */
      private int maxRowHeightAdjustedByIndex = -1;
      
      /** The location of a widget being currently processed. */
      private Point widgetLocation = null;
      
      /**
       * Resulting dimensions.
       */
      Dimension res = new Dimension(0, 0);
      
      /**
       * Constructor.
       *
       * @param   width
       *          Maximal layout width.
       * @param   doPlacement
       *          Widget placement mode.
       */
      public RegularModeVariables(double width, boolean doPlacement)
      {
         this.width       = Coordinate.scale(width);
         this.doPlacement = doPlacement;
         
         // TBD I don't know how that is affected by per widget ATTR-SPACE 
         if (attr_space && !hasHeaders)
            this.width -= 1;
      }
      
      /**
       * Check and if necessary handle SKIP and SPACE widgets.
       *
       * @param   i
       *          Widget index.
       *
       * @return  <code>true</code> if widget is handled.
       */
      public boolean isSpaceHandled(int i)
      {
         if (c[i] instanceof Skip || c[i] instanceof Space)
         {
            // skip is ignoring for last widget because there is nothing after
            if (c[i] instanceof Skip)
               handleSkip(i);
            else if (c[i] instanceof Space)
               handleSpace(i);
            
            return true;
         }
         
         return false;
      }
      
      /**
       * Calculate headers height.
       */
      public void calculateHeaderHeight()
      {
         boolean widgetsStarted = false;
         
         for (int i = 0; i < c.length; i++)
         {
            double cheight = c[i].height();

            if (!widgetsStarted && c[i] instanceof Skip)
            {
               distanceY += (cheight > 0) ? cheight : 0;
               distanceY = Coordinate.scale(distanceY);
               leadingSkips = i;
            }
            
            if (!(c[i] instanceof LabeledWidget))
            {
               continue;
            }
            
            widgetsStarted = true;
            
            String labelText = null;
            LabeledWidget<O> fin = null;
            
            fin = (LabeledWidget<O>) c[i];
            
            labelText = getLabel(fin);
            
            // safe, comparison of scaled doubles
            if (labelText != null && (labels || fin.isForceLabel()) && !fin.isNoLabels())
            {
               Label<O> label = null;
               Dimension sz = c[i].dimension();
               
               WidgetId nextId = c[i].getId().createId(WidgetId.nextID());
               
               if (sideLabels)
               {
                  labelText += ": ";
                  label =  factory.createLabel(nextId, labelText, c[i].config().frameId);
                  label.setSideWidget(c[i]);
                  sz.width = Coordinate.scale(sz.width +
                                              labelWidth((LabeledWidget<?>) c[i], labelText));
               }
               else
               {
                  // safe, comparison of scaled doubles
                  if (labelWidth(fin, labelText) > sz.width)
                     sz.width = labelWidth(fin, labelText);
                  label = factory.createLabel(nextId, labelText, c[i].config().frameId);
                  label.setUnderlined(isFrameUnderline(c[i].config().frameId));
                  label.setRightAligned(fin.isRightAligned());
                  label.setWidth(sz.width);
               }
               
               // safe, comparison of scaled doubles
               double lblHeight = labelHeight(fin, labelText, label, !sideLabels);
               maxHeight = maxHeight < lblHeight ? lblHeight : maxHeight;
               
               label.destroy();
            }
         }
         
         // final adjustment
         distanceY = Coordinate.scale(maxHeight - distanceY);
      }
      
      /**
       * Check if the frame needs underlining.
       *
       * @param    frameId
       *           The frame ID.
       *
       * @return    <code>true</code> if the frame needs underlining, <code>false</code> if the
       *            frame was defined with the NO-UNDERLINE option.
       */
      boolean isFrameUnderline(int frameId)
      {
         FrameConfig fc = ConfigManager.getInstance().resolveWidgetConfig(
               new WidgetId(frameId), FrameConfig.class);
         
         return fc == null || !fc.noUnderline;
      }
      
      /**
       * Adjust resulting dimensions according to collected information.
       */
      void adjustDimensions()
      {
         // safe, comparison of scaled doubles
         if (res.width < nextColumn - extraSpaces)
            res.width = Coordinate.scale(nextColumn - extraSpaces);
         if (res.height < nextRow)
            res.height = nextRow;
      }
      
      /**
       * Go to thew next row.
       */
      void nextRow()
      {
         nextRow(maxRowHeight, startColumn);
      }
      
      /**
       * Go to thew next row.
       *
       * @param   height
       *          Vertical step.
       */
      void nextRow(double height)
      {
         nextRow(height, 0);
      }
      
      /**
       * Go to the next row and set start column to specified value.
       *
       * @param   height
       *          Vertical step.
       * @param   newCol
       *          New start column.
       */
      void nextRow(double height, double newCol)
      {
         incrementNextRow(height);
         setNextColumn(newCol);
         maxRowHeight = 0;
      }
      
      /**
       * Initialize start column.
       *
       * @param   i
       *          Widget index.
       */
      void setStartColumn(int i)
      {
         wAttrSpace = false;
         
         ControlConfig cfg = UiUtils.getConfig(c[i], ControlConfig.class);
         
         if (cfg != null)
         {
            wAttrSpace = ((ControlConfig) cfg).attrSpace;
         }
         
         startColumn = 0;
         
         if ((attr_space | wAttrSpace) &&
             !(c[i] instanceof Browse) &&
             !sideLabels               &&
             !isStatic(i))
         {
            startColumn = 1;
         }
         
         ySet         = false;
         xSet         = false;
         doFixup      = false;
         fixupNextRow = false;
      }
      
      /**
       * Widgets which are placed outside terminal width in redirected mode
       * need some adjustments. Check if required conditions are met.
       *
       * @param   i
       *          Widget index.
       */
      void checkForFixup(int i)
      {
         if (!(c[i] instanceof LabeledWidget))
            return;
         
         LabeledWidget<O> fin = (LabeledWidget<O>) c[i];
         BaseConfig bc = UiUtils.getConfig(c[i], BaseConfig.class);
         
         boolean assignedCol = ConfigHelper.getAssignedColumn(bc) != BaseConfig.INV_COORD;
         
         if (redirected && bc != null && assignedCol && !fixedWidth)
         {
            doFixup = true;
            
            // safe, comparison of scaled doubles
            if (fin.getTo() != BaseConfig.INV_COORD &&
                fin.getTo() < nextColumn)
            {
               doFixup = false;
            }
            if (fin.getColon() != BaseConfig.INV_COORD && fin.getColon() < nextColumn)
            {
               doFixup = false;
            }
            if (widgetLocation.x != BaseConfig.INV_COORD && widgetLocation.x < nextColumn)
            {
               doFixup = false;
            }
         }
      }
      
      /**
       * Do final adjustments and calculations.
       *
       * @return  calculated dimensions.
       */
      Dimension finalAdjustments()
      {
         incrementNextRow(maxRowHeight);
         
         adjustDimensions();
         
         // safe, comparison of scaled doubles
         if (!hasHeaders || (hdrWidth < res.width))
         {
            // TBD I don't know how that is affected by per widget ATTR-SPACE 
            if (!hasHeaders && attr_space)
               res.width += 1;
            else
               res.width = Coordinate.scale(res.width + startColumn);
         }
         
         res.width = Coordinate.scale(res.width + deltaWidth);
         
         return res;
      }
      
      /**
       * Process widget which has no label for some reason.
       *
       * @param   i
       *          Widget index.
       */
      void handleNoLabelWidget(int i)
      {
         LabeledWidget<O> fin = null;
         double cwidth = c[i].width();
         double cheight = c[i].height();
         
         if (c[i] instanceof LabeledWidget)
            fin = (LabeledWidget<O>) c[i];
         
         // if widget is colon aligned - we need to X-shift widget
         double colonShift = 0;
         
         double colon = BaseConfig.INV_COORD;
         if (fin != null)
         {
            colon = fin.getColon();
         }
         else if (c[i].config() instanceof BaseConfig)
         {
            BaseConfig bc = (BaseConfig) c[i].config();
            
            if (bc.align == ControlEntity.ALIGN_COLON)
            {
               colon = ConfigHelper.getColon(bc);
            }
         }
         // safe, comparison of scaled doubles
         if (colon != BaseConfig.INV_COORD)
         {
            // Progress always shifts the widget by 2 chars; in GUI regardless
            // of the font used when COLON-ALIGNED
            colonShift = 2;
            
            // colon is set, but labels are disabled
            double nc = colon + colonShift;
            
            // safe, comparison of scaled doubles
            if (nc < nextColumn)
            {
               if (!doFixup)
               {
                  if (!ySet)
                  {
                     incrementNextRow(maxRowHeight);
                  }
                  
                  maxRowHeight = 0;
               }
               else
                  fixupNextRow = true;
            }
            // note that attr_space is ignored!
            setNextColumn(nc);
            
            // adjust location to make widget fit into terminal width
            if (doFixup && Coordinate.scale(nextColumn + cwidth) > width)
               setNextColumn(nextColumn - (nextColumn + cwidth - width));
         }
         else if (Coordinate.scale(nextColumn + cwidth) > width)
         {
            if (!doFixup)
            {
               if (!ySet && i != maxRowHeightAdjustedByIndex)
               {
                  incrementNextRow(maxRowHeight);
               }
               maxRowHeight = 0;
            }
            else
               fixupNextRow = true;
            
            setNextColumn(startColumn);
            
            // safe, comparison of scaled doubles
            if (widgetLocation.x != BaseConfig.INV_COORD)
            {
               setNextColumn(nextColumn + widgetLocation.x);
            }
            
            // adjust location to make widget fit into terminal width
            if (doFixup && Coordinate.scale(nextColumn + cwidth) > width)
               setNextColumn(nextColumn - (nextColumn + cwidth - width));
         }
         
         double mh;
         
         // safe, comparison of scaled doubles
         if (!sideLabels &&
             !(hasHeaders && i <= (lastHeader + 1)) &&
             maxRowHeight < (mh = Coordinate.scale(cheight + maxHeight)))
         {
            maxRowHeight = mh;
         }
         
         if (doPlacement)
         {
            if (!sideLabels)
            {
               if (hasHeaders && i <= (lastHeader + 1))
               {
                  widgetLocation = new Point(nextColumn, nextRow);
               }
               else
               {
                  widgetLocation = new Point(nextColumn, nextRow + maxHeight);
               }
            }
            else
            {
               // take into account possible colon align
               widgetLocation = new Point(nextColumn, nextRow);
            }
         }
         
         if (fixupNextRow)
         {
            nextRow();
         }
         
         setNextColumn(nextColumn + cwidth);
         extraSpaces = getNumSpaces(i + 1);
         setNextColumn(nextColumn + extraSpaces);
         
         // safe, comparison of scaled doubles
         if (maxRowHeight < cheight &&
             (ThinClient.getInstance().isChui() || !(c[i] instanceof Frame)))
         {
            maxRowHeight = cheight;
            if (!sideLabels)
               maxRowHeight = Coordinate.scale(maxRowHeight + maxHeight);
         }
      }
      
      /**
       * Calculate 'quirks' offset for labels.
       * <p>
       * Under some circumstances label may be shifted right from its correct
       * placement.
       *
       * @param   i
       *          Widget index.
       */
      private void calcQuirksOffset(int i)
      {
         if (!sideLabels &&
             attr_space  &&
             (c[i] instanceof RadioSet      ||
              c[i] instanceof SelectionList ||
              c[i] instanceof ComboBox      ||
              c[i] instanceof Editor))
         {
            if (i < (c.length - 1) && c[i + 1] instanceof LabeledWidget)
            {
               CoordinatesConversion cc = c[i].screen().coordinates();
               deltaWidth += cc.widthFromNative(1);
            }
         }
      }
      
      /**
       * Handle labeled widget.
       *
       * @param   i
       *          Widget index.
       * @param   container
       *          Container where widget and label are stored.
       */
      void handleLabeledWidget(int i, Container<O> container)
      {
         LabeledWidget<O> fin = (LabeledWidget<O>) c[i];
         
         Dimension sz = c[i].dimension();
         String labelText = getLabel(fin);
         Label<O> removed = fin.getLabelInstance();
         
         Label<O> label = null;
         WidgetId nextId = c[i].getId().createId(WidgetId.nextID());
         
         // side-label space width including colon and space
         double labelSpaceWidth = 0;
         
         // side-label space width excluding colon and space
         double labelWidth = 0;
         
         // the effective column location of the widget's label when 
         // COLON-ALIGNED is used
         double colonColumn = 0;
         
         // create label
         if (sideLabels)
         {
            labelWidth = labelWidth((LabeledWidget<?>) c[i], labelText);
            
            labelText += ": ";
            label = factory.createLabel(nextId, labelText, c[i].config().frameId);
            label.setRightAligned(ThinClient.getInstance().isChui());
            label.setSideWidget(c[i]);
            
            labelSpaceWidth = labelWidth((LabeledWidget<?>) c[i], labelText);
            label.setWidth(labelSpaceWidth);
            
            if (fin.getColon() != BaseConfig.INV_COORD)
            {
               if (ThinClient.getInstance().isChui())
               {
                  colonColumn = fin.getColon() - labelWidth;
               }
               else
               {
                  // Progress always adds 2 chars - colon and space - 
                  // (regardless of the label's font) to the colon value 
                  // when determining the location of the widget's left edge
                  colonColumn = fin.getColon() - labelSpaceWidth + 2;
                  
                  // do not let the column location to negative values
                  if (colonColumn < 0)
                  {
                     colonColumn = 0;
                  }
               }
            }
            
            sz.width = Coordinate.scale(sz.width + labelSpaceWidth);
            
            // the total widget height may be greater due to its label 
            double lheight = label.height();
            if (sz.height < lheight)
            {
               sz.height = lheight;
            }
         }
         else
         {
            // safe, comparison of scaled doubles
            double lblWidth = labelWidth(fin, labelText);
            if (lblWidth > sz.width)
            {
               sz.width = lblWidth;
            }
            
            label = factory.createLabel(nextId, labelText, c[i].config().frameId);
            label.setUnderlined(isFrameUnderline(c[i].config().frameId));
            label.setRightAligned(fin.isRightAligned());
            label.setWidth(sz.width);
            label.setOffset(deltaWidth);
            // fix for UNLESS-HIDDEN option
            if (fin.hidden() && !fin.wasVisible())
            {
               label.setEmptyModeOn();
               label.setEmptyModeOff();
            }
            sz.height = Coordinate.scale(sz.height + maxHeight);
         }
         
         // safe, comparison of scaled doubles
         if (Coordinate.scale(nextColumn + sz.width) > width)
         {
            if (!doFixup)
            {
               nextRow((ySet || i == maxRowHeightAdjustedByIndex) ? 0 : maxRowHeight,
                       startColumn);
            }
            else
            {
               fixupNextRow = true;
            }
            
            // safe, comparison of scaled doubles
            if (widgetLocation.x != BaseConfig.INV_COORD)
            {
               setNextColumn(nextColumn + widgetLocation.x);
               double lblWidth = labelWidth(fin, labelText);
               
               if (sideLabels                          &&
                   fin.getTo() == BaseConfig.INV_COORD &&
                   Coordinate.scale(nextColumn + lblWidth + sz.width) > width)
               {
                  if (Coordinate.scale(lblWidth + sz.width) > width)
                  {
                     // TODO: widget does not fit in frame, report error!!!
                     setNextColumn(0);
                  }
                  else
                     setNextColumn(width - sz.width);
               }
            }
            
            // safe, comparison of scaled doubles
            if (fin.getColon() != BaseConfig.INV_COORD)
            {
               setNextColumn(colonColumn);
               
               if (!sideLabels)
               {
                  double sideLabelSuffixLength = labelWidth(fin, ": ");
                  setNextColumn(nextColumn + sideLabelSuffixLength); // size of the ": "
               }
            }
            
            // adjust location to make widget fit into terminal width
            if (doFixup && Coordinate.scale(nextColumn + sz.width) > width)
               setNextColumn(nextColumn - (nextColumn + sz.width - width));
         }
         else
         {
            // safe, comparison of scaled doubles
            if (fin.getColon() != BaseConfig.INV_COORD)
            {
               if (Coordinate.scale(colonColumn + sz.width) >= width)
               {
                  // check if it does fit at all
                  // safe, comparison of scaled doubles
                  if (sz.width > width)
                  {
                     // does not fit into the given width
                     throw new RuntimeException("**Cannot fit FILL-IN " +
                                                fin.getLabelText() +
                                                " with COLON within FRAME  (4027). " +
                                                "Ignoring position info for COLON (2054)");
                  }
                  
                  // at this point we know that the frame width exceeds the
                  // terminal width, so we adjust the side-label's leading space
                  
                  // check if there other widgets already placed and
                  // if wrapping is necessary
                  
                  // safe, comparison of scaled doubles
                  if (nextColumn > startColumn && Coordinate.scale(nextColumn + sz.width) > width)
                  {
                     // wrap to next row
                     nextRow();
                  }
                  
                  // force widget fit given size
                  setNextColumn(width - sz.width);
               }
               else
               {
                  // safe, comparison of scaled doubles
                  if (nextColumn > Coordinate.scale(colonColumn))
                  {
                     nextRow();
                  }
                  
                  setNextColumn(colonColumn);
                  
                  if (!sideLabels)
                  {
                     double sideLabelSuffixLength = labelWidth(fin, ": ");
                     setNextColumn(nextColumn + sideLabelSuffixLength);
                  }
               }
            }
         }
         
         Point location = new Point(nextColumn, nextRow);
         
         // only place column labels here, side labels are positioned together with their
         // side widgets
         if (doPlacement && !sideLabels)
         {
            double lblHeight = labelHeight((LabeledWidget<?>) c[i],
                                           label.name(),
                                           label,
                                           !sideLabels);
            double deltaY = Coordinate.scale(distanceY - lblHeight);
            setLabelLocation(label, location.x, location.y + deltaY);
         }
         
         if (sideLabels)
         {
            location.x = Coordinate.scale(location.x + labelSpaceWidth);
         }
         else
         {
            double cwidth = c[i].width();
            if (fin.isRightAligned() && sz.width > cwidth)
               location.x = Coordinate.scale(location.x + sz.width - cwidth);      
            
            location.y = Coordinate.scale(location.y + maxHeight);
         }
         
         setNextColumn(nextColumn + sz.width);
         
         if (doPlacement)
         {
            widgetLocation = new Point(location.x, location.y);
         }
         
         // safe, comparison of scaled doubles
         if (maxRowHeight < sz.height)
         {
            maxRowHeight = sz.height;
         }
         
         extraSpaces = getNumSpaces(i + 1);
         setNextColumn(nextColumn + extraSpaces);
         
         if (doPlacement)
         {
            fin.setLabelInstance(label);
            
            // do not discard dirty changes with the destruction of the removed label instance
            if (removed != null)
            {
               // remove it from the registry
               OutputManager.instance().getRegistry().removeWidget(removed.getId());
               
               removed.config().id = null;
               container.replace(removed, label);
            }
            else
            {
               container.add(c[i], label);
            }
            
            // add it from the registry
            OutputManager.instance().getRegistry().addWidget((Widget) label);
         }
         else if (label != null)
         {
            label.destroy();
         }
         
         // safe, comparison of scaled doubles
         if (nextColumn >= width && width > 0)
         {
            // force immediate wrap to the next row
            if (res.width < width)
               res.width = width;
            
            nextRow();
         }
      }
      
      /**
       * Handle SPACE widget.
       *
       * @param   i
       *          Widget index.
       */
      void handleSpace(int i)
      {
         // strangely enough, leading space is not taken into account here...
         double cwidth = c[i].width();
         
         if (ThinClient.getInstance().isChui() &&
             Coordinate.scale(nextColumn + cwidth) > width)
         {
            nextRow(maxRowHeight);
            
            // place it at next row
            if (doPlacement)
            {
               widgetLocation = new Point(0, nextRow);
            }
         }
         else
         {
            // in GUI or case we are inside the effect is to increase final frame width
            setNextColumn(nextColumn + cwidth);
         }
      }
      
      /**
       * Handle SKIP widget.
       *
       * @param   i
       *          Widget index.
       */
      void handleSkip(int i)
      {
         // place SKIP widget
         incrementNextRow(maxRowHeight);
         
         if (doPlacement)
         {
            double row = nextRow;
            
            if (!sideLabels && i <= lastHeader)
            {
               row += maxHeight;
            }
            
            widgetLocation = new Point(0, row);
         }
         
         nextRow(c[i].height());
      }
      
      /**
       * Handle widget TO option.
       *
       * @param   i
       *          Widget index.
       */
      void handleTO(int i)
      {
         WidgetConfig cfg = c[i].config();
         
         if (!(cfg instanceof BaseConfig))
         {
            return;
         }

         double toVal = ConfigHelper.getTo(c[i].config());
         
         // safe, comparison of scaled doubles
         if (toVal == BaseConfig.INV_COORD)
            return;
         
         // add 1 char unit, in both ChUI and GUI
         double newCol = toVal + 1;
         double cwidth = c[i].width();

         if (isLabeled(i))
         {
            LabeledWidget<O> fin = (LabeledWidget<O>) c[i];

            double labelWidth = -1;
            double textWidth = -1;
            Label<O> label = null;
            
            if (sideLabels && (label = fin.getLabelInstance()) != null)
            {
               labelWidth = label.width();
               
               String lblText = label.name();
               if (lblText.length() > 2)
               {
                  lblText = lblText.substring(0, lblText.length() - 2);
               }
               textWidth = labelWidth(fin, lblText);
            }
            
            if (labelWidth < 0)
            {
               String lblText = getLabel(fin);
               labelWidth = labelWidth(fin, lblText + ": ");
               textWidth = labelWidth(fin, lblText);
            }
            
            if (!sideLabels)
               newCol -= Math.max(textWidth, cwidth);
            else
               newCol -= labelWidth + cwidth;
         }
         else
            newCol -= cwidth;
      
         newCol = Coordinate.scale(newCol);
         
         if (newCol < 0)
         {
            System.err.println("## Runtime warning: TO column is too large, ignoring TO");
            
            if (!OutputManager.instance().isRedirected())
            {
               // for interactive frames just increase width
               width = Coordinate.scale(width - newCol);
            }
            setNextColumn(0);
         }
         else
         {
            // safe, comparison of scaled doubles
            if (newCol < nextColumn)
            {
               if (widgetLocation.y == BaseConfig.INV_COORD)
               {
                  incrementNextRow(maxRowHeight);
               }
               
               maxRowHeight = 0;
            }
            
            setNextColumn(newCol);
         }
      }
      
      /**
       * Handle widgets with explicitly set column.
       *
       * @param   i
       *          Widget index.
       */
      void handleInitColumn(int i)
      {
         BaseConfig bc = c[i].config();
         
         // safe, comparison of scaled doubles
         if (ConfigHelper.getAssignedColumn(bc) != BaseConfig.INV_COORD)
         {
            if (nextColumn > widgetLocation.x)
            {
               nextRow(maxRowHeight);
            }
            
            setNextColumn(widgetLocation.x);
            xSet = true;
         }
         else
         {
            if (nextColumn == 0)
               setNextColumn(startColumn);
         }
      }
      
      /**
       * Handle widget with explicitly set initial row.
       *
       * @param   i
       *          widget index.
       */
      void handleInitRow(int i)
      {
         // safe, comparison of scaled doubles
         if (((BaseConfig)c[i].config()).initRow != BaseConfig.INV_COORD)
         {
            setNextRow(widgetLocation.y);
            setNextColumn(xSet ? widgetLocation.x : startColumn);
            ySet = true;
         }
         
         // safe, comparison of scaled doubles
         if (maxRowHeight == 0 &&
             (ThinClient.getInstance().isChui() || !(c[i] instanceof Frame)))
         {
            maxRowHeight = c[i].height();
            maxRowHeightAdjustedByIndex = i;
         }
         else
         {
            maxRowHeightAdjustedByIndex = -1;
         }
      }
      
      /**
       * Set next row for widget placement.
       *
       * @param   nextRow
       *          New value for the nextRow
       */
      void setNextRow(double nextRow)
      {
         incrementNextRow(nextRow - this.nextRow);
      }
      
      /**
       * Increment next row for widget placement by specified delta.
       *
       * @param   delta
       *          Next row increment.
       */
      void incrementNextRow(double delta)
      {
         delta = Coordinate.scale(delta);
         nextRow = Coordinate.scale(nextRow + delta);
      }
      
      /**
       * Set the next column position for widget placement.
       *
       * @param   nextColumn
       *          The nextColumn to set.
       */
      void setNextColumn(double nextColumn)
      {
         nextColumn = Coordinate.scale(nextColumn);
         this.nextColumn = nextColumn;
      }
      
      /**
       * Initializes the instance variable holding the location of a widget 
       * being currently processed.
       *
       * @param   i
       *          Widget index.
       */
      void initWidgetLocation(int i)
      {
         BaseConfig cfg = UiUtils.getConfig(c[i], BaseConfig.class);
         
         if (cfg != null)
         {
            widgetLocation = ConfigHelper.getInitialLocation(cfg);
         }
         else
         {
            widgetLocation = new Point(BaseConfig.INV_COORD, BaseConfig.INV_COORD);
         }
      }
      
      /**
       * Sets widget position from the calculated location or location set
       * by server.
       *
       * @param   i
       *          Widget index.
       */
      void setWidgetLocation(int i)
      {
         if (!doPlacement)
         {
            return;
         }
         
         Widget<O> w = c[i];
         
         setWidgetLocation(w);
      }
      
      /**
       * Sets widget position from the calculated location or location set
       * by server.
       *
       * @param   w
       *          Widget.
       */
      void setWidgetLocation(Widget<O> w)
      {
         if (!doPlacement)
         {
            return;
         }
         
         BaseConfig bc = UiUtils.getConfig(w, BaseConfig.class);
         
         if (bc != null && bc.column != BaseConfig.INV_COORD)
         {
            widgetLocation.x = bc.column;
         }
         
         if (bc != null && bc.row != BaseConfig.INV_COORD)
         {
            widgetLocation.y = bc.row;
         }
         
         w.setLocation(widgetLocation);
         
         NativePoint nloc = coordinates().pointToNative(widgetLocation);
         w.setPhysicalLocation(nloc.x, nloc.y);
         
         if (bc != null)
         {
            // widget column can not be different from the FRAME- attributes, as the 
            // FIELD-GROUP:COLUMN attribute is readonly.  This is left in place in case in the
            // future is found a case where these may be different (maybe nested frames?)
            bc.frameColumnOffset = 0;
            bc.frameXOffset = 0;
            
            // adjust the vertical offset: it includes the height of the frame's top-label header
            bc.frameRowOffset = sideLabels ? 0 : distanceY;
            bc.frameYOffset = w.screen().coordinates().rowToPixels(bc.frameRowOffset);
            
            bc.widgetPlaced = true;
         }
         
         // make sure the side-label is "stuck" to its labeled widget, unless its location
         // is explicitly assigned
         if (sideLabels && w instanceof LabeledWidget)
         {
            Label label = ((LabeledWidget) w).getLabelInstance();
            if (label != null &&
                ConfigHelper.getAssignedColumn(label.config) == BaseConfig.INV_COORD)
            {
               String labelText = getLabel((LabeledWidget<O>) w);
               double labelSpaceWith = labelWidth((LabeledWidget<?>) w, labelText + ": ");
               setLabelLocation(label, widgetLocation.x - labelSpaceWith, widgetLocation.y);
            }
         }
      }
      
      /**
       * Sets label location.
       *
       * @param   label
       *          The label to position.
       * @param   col
       *          Default column calculated by the layout manager. 
       * @param   row
       *          Default row calculated by the layout manager.
       */
      private void setLabelLocation(Label<O> label, double col, double row)
      {
         label.setLocation(col, row);
         label.config.widgetPlaced = true;
      }
   }
   
   /**
    * Stacked mode variables bundle.
    */
   class StackedModeVariables
   {
      /** Initial column for new row. */
      double startColumn         = 0;
      /** Next free row in top part of layout. */
      double nextTopRow          = 0;
      /** Next free row in bottom part of layout. */
      double nextBottomRow       = 0;
      /** Next free column in top part of layout. */
      double nextTopCol          = 0;
      /** Next free column in bottom part of layout. */
      double nextBottomCol       = 0;
      /** Maximal height of widget in current row in top part of layout. */
      double topCurrentHeight    = 0;
      /** Maximal height of widget in current row in bottom part of layout. */
      double bottomCurrentHeight = 0;
      
      /** Extra spaces to insert between widgets for top of layout. */
      double extraTopSpaces    = 0;
      /** Extra spaces to insert between widgets for bottom part of layout. */
      double extraBottomSpaces = 0;
      /** Maximal width of resulting layout. */
      double width;
      
      /** Next widget top part Y coordinate calculated in pixels.*/
      private int nextTopY = 0;
      
      /** Next widget top part Y coordinate calculated in pixels.*/
      private int nextTopX = 0;
      
      /** Next widget bottom part Y coordinate calculated in pixels.*/
      private int nextBottomY = 0;
      
      /** Next widget bottom part Y coordinate calculated in pixels.*/
      private int nextBottomX = 0;
      
      /** List of labels in current row. */
      private List<Label<O>> labelList = new ArrayList<>();
      
      /** Map of side-labels to their associated side-widget. */
      private Map<Label<O>, LabeledWidget<O>> labelMap = new HashMap<>();
      
      /**
       * Construct an variable bundle.
       *
       * @param   width
       *          Container width.
       */
      StackedModeVariables(double width)
      {
         this.width = Coordinate.scale(width);
         nextTopCol    = (attr_space && labels) ? 1 : 0;
         setNextBottomCol((attr_space && labels) ? 1 : 0);
         nextTopX = coordinates().widthToNative(nextTopCol);
         nextBottomX = coordinates().widthToNative(nextBottomCol);
         
         // determine start column
         if (attr_space)
         {
            if (labels)
               startColumn = 1;
            
            if (hasBody && hasHeaders)
               startColumn = 1;
         }
      }
      
      /**
       * Add label into list of current row labels.
       *
       * @param   label
       *          Label to add.
       * @param   widget
       *          The associated side-widget.
       */
      void addLabel(Label<O> label, LabeledWidget<O> widget)
      {
         labelList.add(label);
         labelMap.put(label, widget);
      }
      
      /**
       * Finalize label placement.
       */
      void fixLabelsRow()
      {
         for (Label<O> label : labelList)
         {
            double lblHeight = labelHeight(labelMap.get(label), label.name(), label, !sideLabels);
            //TODO: fix handling of physical row
            label.setLocation(label.location().x,
                              label.location().y + topCurrentHeight - lblHeight);
         }
         
         labelList.clear();
         labelMap.clear();
      }
      
      /**
       * Clear variables and prepare them for real size calculation/relocation.
       */
      void clear()
      {
         nextBottomRow       = nextTopRow + 1;
         nextBottomY         = nextTopY + coordinates().heightToNative(1);
         nextTopRow          = 0;
         nextTopY            = 0;
         setNextTopCol(startColumn);
         setNextBottomCol(startColumn);
         topCurrentHeight    = 0;
         bottomCurrentHeight = 0;
         
         extraTopSpaces      = 0;
         extraBottomSpaces   = 0;
         
         labelList.clear();
         labelMap.clear();
      }
      
      /**
       * Check if next component requires space before it for stacked mode.
       *
       * @param   i
       *          Index of component to check.
       * @param   col
       *          Expected new column for the widget.
       *
       * @return  <code>true</code> if next component requires space before it.
       */
      boolean needSpaceStacked(int i, double col)
      {
         if (i >= c.length || !(c[i] instanceof LabeledWidget))
         {
            return false;
         }
         
         BaseConfig bc = c[i].config();
         Point assignedLoc = ConfigHelper.getInitialLocation(bc);
         
         if (assignedLoc.x == BaseConfig.INV_COORD || assignedLoc.y == BaseConfig.INV_COORD)
         {
            // and it can be relocated
            if (col + c[i].width() < width)
            {
               return true;
            }
         }
         
         return false;
      }
      
      /**
       * Adjust top part dimension according to current variables.
       *
       * @param   top
       *          The {@link Dimension} instance to adjust.
       */
      void adjustTop(Dimension top)
      {
         // safe, comparison of scaled doubles
         if (top.width < nextTopCol - extraTopSpaces)
            top.width = Coordinate.scale(nextTopCol - extraTopSpaces);
         if (top.height < nextTopRow)
            top.height = nextTopRow;
      }
      
      /**
       * Adjust bottom part dimension according to current variables.
       *
       * @param   bottom
       *          The {@link Dimension} instance to adjust.
       */
      void adjustBottom(Dimension bottom)
      {
         // safe, comparison of scaled doubles
         if (bottom.width < nextBottomCol - extraBottomSpaces)
            bottom.width = Coordinate.scale(nextBottomCol - extraBottomSpaces);
         if (bottom.height < nextBottomRow)
            bottom.height = nextBottomRow;
      }
      
      /**
       * Check for possible wrapping.
       *
       * @param   i
       *          Index of component to check.
       * @param   len
       *          Component length.
       *
       * @return  <code>true</code> if wrapping occurs.
       */
      boolean isSkipOrWrap(int i, double len)
      {
         if (c[i] instanceof Skip)
         {
            return true;
         }
         
         return (Coordinate.scale(len + nextBottomCol) > width ||
                 Coordinate.scale(len + nextTopCol) > width)      &&
                !hasDefinedX(c[i]);
      }
      
      /**
       * Finalize height calculations.
       */
      void finalizeHeight()
      {
         incrementNextTopRow(topCurrentHeight);
         incrementNextBottomRow(bottomCurrentHeight);
         setNextTopCol(startColumn);
         setNextBottomCol(startColumn);
         int startX = coordinates().widthToNative(startColumn);
         nextTopX = startX;
         nextBottomX = startX;
         
         topCurrentHeight    = 0;
         bottomCurrentHeight = 0;
      }
      
      /**
       * Increment bottom row calculated in character cells and in pixels.
       *
       * @param   height
       *          Increment height in character cells. 
       */
      private void incrementNextBottomRow(double height)
      {
         height = Coordinate.scale(height);
         nextBottomY += coordinates().heightToNative(height);
         nextBottomRow = Coordinate.scale(nextBottomRow + height);
      }
      
      /**
       * Increment top row calculated in character cells and in pixels.
       *
       * @param   height
       *          Increment height in character cells. 
       */
      private void incrementNextTopRow(double height)
      {
         height = Coordinate.scale(height);
         nextTopY += coordinates().heightToNative(height);
         nextTopRow = Coordinate.scale(nextTopRow + height);
      }
      
      /**
       * Process SKIP widget.
       *
       * @param   i
       *          Index of component to process.
       */
      void processSkip(int i)
      {
         double cheight = c[i].height();

         if (hasHeaders && i < lastHeader)
            incrementNextTopRow(cheight);
         else
            incrementNextBottomRow(cheight);
      }
      
      /**
       * Process SPACE widget.
       *
       * @param   i
       *          Index of component to process.
       */
      void processSpace(int i)
      {
         // strangely enough, leading space is not taken into account here...
         // safe, comparison of scaled doubles
         if (attr_space && labels && nextTopCol == startColumn)
         {
            setNextTopCol(0);
            setNextBottomCol(0);
         }
         
         double cwidth = c[i].width();
         
         setNextTopCol(nextTopCol + cwidth);
         setNextBottomCol(nextBottomCol + cwidth);
      }
      
      /**
       * Process fixed widget coordinates in bottom part of layout.
       *
       * @param   location
       *          The widget location.
       */
      void processFixedBottom(Point location)
      {
         double lx = Coordinate.scale(location.x);
         double ly = Coordinate.scale(location.y);
         
         // safe, comparison of scaled doubles
         if (lx != BaseConfig.INV_COORD)
         {
            if (Coordinate.scale(nextBottomCol - extraBottomSpaces) > lx)
            {
               incrementNextBottomRow(bottomCurrentHeight);
               bottomCurrentHeight = 0;
            }
            setNextBottomCol(lx);
         }
         
         // safe, comparison of scaled doubles
         if (ly != BaseConfig.INV_COORD)
         {
            incrementNextBottomRow(ly - nextBottomRow);
         }
      }
      
      /**
       * Process widget in the bottom part of layout: adjust variables and place widget if
       * requested.
       * <p>
       * As a side-effect, when this method exits, the {@code nextBottomCol} should be incremented
       * at least by {@code len}. If the next component (if it exists) needs extra space,
       * {@code nextBottomCol} will be incremented by 1. If this widget is marked with
       * {@code ATTR-SPACE}, the {@code nextBottomCol} will be additionally incremented by 1. The
       * {@code extraBottomSpaces} will keep the count of the extra spaces added.
       * In any case, the {@code nextBottomCol} and {@code extraBottomSpaces} must match the
       * top-line correspondents (that is {@code nextTopCol} and {@code extraTopSpaces},
       * respectively), in order to keep the layout consistent (the next widget and its label to
       * be aligned at the same column).
       *
       * @param   i
       *          Index of component to process.
       * @param   len
       *          Widget length.
       * @param   sh1
       *          Widget column shift for space calculation.
       * @param   place
       *          Perform widget placement.
       * @param   sh2
       *          Widget column shift for placement.
       */
      void processBottom(int i, double len, int sh1, boolean place, double sh2)
      {
         extraBottomSpaces = needSpaceStacked(i + 1, nextBottomCol)
                                 ? getExtraSpacesStacked(i) : 0;
         
         // safe, comparison of scaled doubles
         if (Coordinate.scale(nextBottomCol + len) == width)
         {
            extraBottomSpaces = 0;
         }
         
         if (Coordinate.scale(nextBottomCol + len + extraBottomSpaces) > width)
         {
            incrementNextBottomRow(bottomCurrentHeight);
            bottomCurrentHeight = 0;
         }
         
         if (sh1 > 0)
         {
            setNextBottomCol(nextBottomCol + sh1);
            sh2 = sh2 - sh1;
         }
         else
         {
            sh2 = 0;
         }
         
         if (place)
         {
            c[i].setLocation(nextBottomCol + sh2, nextBottomRow);
            c[i].setPhysicalLocation(nextBottomX + coordinates().widthToNative(sh2), nextBottomY);
         }
         
         // nextBottomCol already contains the sh1 (or it is null)
         setNextBottomCol(nextBottomCol - sh1 + (len + extraBottomSpaces));
         
         // safe, comparison of scaled doubles
         double cheight = c[i].height();
         if (bottomCurrentHeight < cheight)
            bottomCurrentHeight = cheight;
      }
      
      /**
       * Update top part of layout with specified widget and then adjust passed {@link Dimension}
       * instance.
       * <p>
       * As a side-effect, when this method exits, the {@code nextTopCol} should be incremented at
       * least by {@code len}. If the next component (if it exists) needs extra space,
       * {@code nextTopCol} will be incremented by 1. If this widget is marked with
       * {@code ATTR-SPACE}, the {@code nextTopCol} will be additionally incremented by 1. The
       * {@code extraTopSpaces} will keep the count of the extra spaces added.
       *
       * @param   i
       *          Index of component to process.
       * @param   len
       *          Calculated widget width.
       * @param   top
       *          Instance of {@link Dimension} to adjust.
       */
      void processTop(int i, double len, Dimension top)
      {
         setNextTopCol(nextTopCol + len);
         
         if (needSpaceStacked(i + 1, nextTopCol))
         {
            extraTopSpaces = getExtraSpacesStacked(i);
            setNextTopCol(nextTopCol + extraTopSpaces);
         }
         else
         {
            extraTopSpaces = 0;
         }
         
         adjustTop(top);
      }
      
      /**
       * Process fixed widget coordinates in top part of layout.
       *
       * @param   location
       *          The widget location.
       */
      void processFixedTop(Point location)
      {
         double lx = Coordinate.scale(location.x);
         double ly = Coordinate.scale(location.y);
         
         boolean wrapped = false;
         
         // safe, comparison of scaled doubles
         if (lx != BaseConfig.INV_COORD)
         {
            if (Coordinate.scale(nextTopCol - extraTopSpaces) > lx)
            {
               fixLabelsRow();
               
               incrementNextTopRow(topCurrentHeight);
               topCurrentHeight  = 0;
               wrapped = true;
            }
            setNextTopCol(lx);
         }
         
         // safe, comparison of scaled doubles
         if (ly != BaseConfig.INV_COORD && (!wrapped || ly > nextTopRow))
         {
            incrementNextTopRow(ly - nextTopRow);
         }
      }
      
      /**
       * Process fixed widget coordinates for both, top and bottom parts of
       * layout.
       *
       * @param   i
       *          Widget index.
       * @param   location
       *          Widget location.
       * @param   len
       *          Widget length.
       * @param   text
       *          Label text.
       */
      void processFixed(int i, Point location, double len, String text)
      {
         double lx = Coordinate.scale(location.x);
         double ly = Coordinate.scale(location.y);
         
         boolean wrapped = false;
         
         LabeledWidget<O> lw = (LabeledWidget<O>) c[i];
         
         // safe, comparison of scaled doubles
         if (lx != BaseConfig.INV_COORD)
         {
            // safe, comparison of scaled doubles
            if (Coordinate.scale(nextTopCol - extraTopSpaces) > lx)
            {
               fixLabelsRow();
               
               incrementNextTopRow(topCurrentHeight);
               topCurrentHeight = 0;
               
               incrementNextBottomRow(bottomCurrentHeight);
               bottomCurrentHeight = 0;
               
               setNextBottomCol(startColumn);
               wrapped = true;
            }
            setNextTopCol(lx);
         }
         
         // safe, comparison of scaled doubles
         double labelHeight = labelHeight(lw, text, lw.getLabelInstance(), !sideLabels);
         if (topCurrentHeight < labelHeight)
         {
            topCurrentHeight = labelHeight;
         }
         
         // safe, comparison of scaled doubles
         if (ly == BaseConfig.INV_COORD)
         {
            return;
         }
         
         // safe, comparison of scaled doubles
         if (wrapped && nextTopRow > ly)
         {
            return;
         }
         
         if (!wrapped && i < (c.length - 1))
         {
            Point loc = c[i + 1].location();
            
            // safe, comparison of scaled doubles
            if (loc.y == ly &&
                loc.x < Coordinate.scale(nextTopCol + len))
            {
               incrementNextTopRow((location.y + topCurrentHeight) - nextTopRow);
               return;
            }
         }
         
         incrementNextTopRow(location.y - nextTopRow);
      }
      
      /**
       * Reset start column for header widgets, because they always start at
       * column 0, regardless from the ATTR-SPACE attribute.
       */
      void resetHeaderStart()
      {
         // safe, comparison of scaled doubles
         if (nextTopCol == startColumn && nextTopCol > 0)
            setNextTopCol(0);
      }
      
      /**
       * Adjust provided instances of {@link Dimension} with calculated top
       * and bottom parts of layout.
       *
       * @param   top
       *          Instance of {@link Dimension} to adjust with top part of
       *          layout.
       * @param   bottom
       *          Instance of {@link Dimension} to adjust with bottom part of
       *          layout.
       */
      void adjustDimensions(Dimension top, Dimension bottom)
      {
         adjustTop(top);
         adjustBottom(bottom);
      }
      
      /**
       * Sets the next column to be filled on the top side.
       *
       * @param   nextTopCol
       *          The nextTopCol to set.
       */
      void setNextTopCol(double nextTopCol)
      {
         nextTopCol = Coordinate.scale(nextTopCol);
         
         this.nextTopCol = nextTopCol;
         this.nextTopX = coordinates().widthToNative(nextTopCol);
      }
      
      /**
       * Sets the next column to be filled on the bottom side.
       *
       * @param   nextBottomCol
       *          The nextBottomCol to set.
       */
      void setNextBottomCol(double nextBottomCol)
      {
         nextBottomCol = Coordinate.scale(nextBottomCol);
         
         this.nextBottomCol = nextBottomCol;
         this.nextBottomX = coordinates().widthToNative(nextBottomCol);
      }
   }
}
