/*
** Module   : FrameChuiImpl.java
** Abstract : Frame implementation for CHUI
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111121 Extracted CHUI-specific part from Frame.
** 002 EVL 20140514 Fix for widget underlining for still non-visible frames. This case
**                  requires one time per frame shift for down body content one line up.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
**                  Moved here some ChUI-specific APIs from Frame class.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 MAG 20141017 Move common code to Frame class.
**                  Renamed from FrameImpl and moved to ChUI widgets package.
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 HC  20150323 Removed business logic from config classes.
** 010 CA  20150406 Changes related to computing the label size and layout in GUI.
** 011 CA  20150723 Extracted the down frame drawing code in Frame.drawDownFrame, to allow drawing
**                  of DOWN frames in GUI. 
** 012 CA  20150728 Added frameContentBounds - in GUI, when there is a title, the content has an
**                  Y offset from which it starts being drawn; this needs to be accounted for when 
**                  drawing the down body.
** 013 EVL 20160410 Fix for minimum size of frame calculation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
 
package com.goldencode.p2j.ui.client.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;


/**
 * A <code>Frame</code> is a top-level window with a title and a border. 
 */
@SuppressWarnings("unchecked")
public class FrameChuiImpl
extends Frame<ChuiOutputManager>
{
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public FrameChuiImpl(WidgetId id, FrameConfig cfg)
   {
      super(id, cfg);
   }

   /**
    * Compute the label width in character units.
    * 
    * @param    label
    *           The label text.
    * @param    widget
    *           The widgeth to which the label belongs.
    *           
    * @return   The number of characters in the <code>label</code> string.
    */
   @Override
   public double getLabelWidth(String label, LabeledWidget<ChuiOutputManager> widget)
   {
      return label.length();
   }

   /**
    * Compute the label height in character units.
    * 
    * @param    labelText
    *           The label text.
    * @param    widget
    *           The widget to which the label belongs.
    * @param    topLabel
    *           Flag identifying that the label is shown above the widget.
    *           
    * @return   Always 1.
    */
   @Override
   public double getLabelHeight(String labelText, 
                                LabeledWidget<ChuiOutputManager> widget, boolean topLabel)
   {
      return 1;
   }
   
   /**
    * Get a delta which is added between frames, when their Y coordinate is automatically 
    * computed.
    * 
    * @return   Always <code>0</code>.
    */
   public double getRowDelta()
   {
      return 0;
   }
   
   /**
    * Get clipping rectangle.
    *
    * @return   clipping rectangle which takes into account frame border.
    */
   @Override
   public Rectangle getClipRect()
   {
      Rectangle rect  = super.getClipRect();

      rect = new Rectangle(rect.top()    - getBorderTopChars(),
                           rect.left()   - getBorderLeftChars(),
                           rect.bottom() - getBorderBottomChars(),
                           rect.right()  - getBorderRightChars());

      return rect;
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per frame.
    * 
    * @param   origin
    *          The coordinate of the upper-left corner for peer container in
    *          character units.
    * @param   size
    *          The size of the peer container in character units.
    */
   @Override
   public void draw(Point origin, Dimension size)
   {
      ThinClient tc = ThinClient.getInstance();
      
      boolean drawTitle = false;
      boolean needSync = false;
      boolean box = config().box;
      boolean redir = tc.isRedirected();
      Point z = contentPane.screenLocation();
      z.y += headersHeight;
      
      Color tcolor = config().titleDColor;
      
      // Getting the current rectangle
      Rectangle thisRect = new Rectangle(origin,
                                         new Dimension(size.width + 1,
                                                       size.height + 1),
                                         screen().coordinates().baseUnits());
      origin.x += getBorderLeftChars(); 
      
      if (tc.viewRefresh && !redir)
      {
         clear(origin, size, UiUtils.getDColor(this));
         needSync = true;
      }

      if (frameScroll != null)
      {
         if (!down)
         {
            frameScroll.draw();
            
            if (redir)
            {
               if (!box)
               {
                  // force leading skips
                  // in chui char coordinates are always whole numbers
                  forceSkips((int) thisRect.top()); 
               }
               
               if (trailingSkips)
                  forceSkips((int) thisRect.bottom() - 1);

               Stream stream = tc.getCurrentRedirection();
               RedirectedTerminal rt = 
                  ThinClient.redirectedTerminalForStream(stream);
               
               // in chui char coordinates are always whole numbers
               rt.markHeader((int) headersHeight);
               drawTitle = true;
            }
            
            needSync = true;
         }
         else
         {
            // control frame decorations/headers drawing for interactive and
            // redirected cases
            if (redir)
            {
               Stream stream = tc.getCurrentRedirection();
               
               if (headersOutput == null)
               {
                  headersOutput = new HashSet<Stream>();
               }
               
               // test to see if we have output yet on this page
               if (stream != null && !headersOutput.contains(stream))
               {
                  RedirectedTerminal rt = 
                              ThinClient.redirectedTerminalForStream(stream);
                  
                  // remember that we have output already, this will be
                  // reset based on notifications
                  headersOutput.add(stream);
                  if (stream.rawPageSize() > 0)
                  {
                     stream.registerPageBreakListener(this);
                  }
                  else
                  {
                     stream.registerCloseListener(this);
                  }
                  
                  if (box)
                  {
                     // all frames generate a line at the top of a new page
                     // preserve the syncable status 
                     forceNewLine();
                     rt.flush();
                     
                     // if we have headers, mark those lines as boxed
                     // in chui char coordinates are always whole numbers
                     if ((int) headersHeight > 0)
                        forceBox((int) (z.y - headersHeight), headersHeight);
                  }
                  
                  // draw the headers
                  frameScroll.draw(nonbody);
                  
                  // the following sync call is required so that the next
                  // iteration does not cause adding blank lines between the
                  // header labels and that iteration!
                  sync();
                  needSync = false;

                  // in chui char coordinates are always whole numbers
                  rt.markHeader((int) headersHeight);
                  
                  drawTitle = true;
               }
            }
            else if (tc.viewRefresh)
            {   
               frameScroll.draw(nonbody);
               needSync = true;
            }

            // call drawDownFrame first
            needSync = drawDownFrame(z, redir) || needSync;
         }
      }

      if (((tc.viewRefresh && !redir) || drawTitle) && hasTitle())
      {
         // Draw the title into the enclosing frame.
         // in chui char coordinates are always whole numbers
         drawTitle(origin, (int) size.width, redir, tcolor);
      }
      
      if (needSync)
         sync();
   }

   /**
    * Adjusts the minimum frame width or height value depending on the current min value and
    * the presense of the box within the frame.
    * 
    * @param   minWidthOrHeight
    *          The size value to be adjusted.
    * @param   isHeight
    *          <code>TRUE</code> if the size value is height, <code>FALSE</code> if width.
    * 
    * @return  ChUI dependent new width or height value.
    */
   @Override
   protected double adjustMinSize(double minWidthOrHeight, boolean isHeight)
   {
      // take initial
      double resWidthOrHeight = minWidthOrHeight;

      // consider if we have a boxed frame or not, the box size is 1.0 for each side
      // the same calculation for height and width in ChUI
      if (resWidthOrHeight == 0)
      {
         resWidthOrHeight = hasBox() ? 2.0 : 1.0;
      }
      else
      {               
         resWidthOrHeight += hasBox() ? 1.0 : 0.0;
      }
      
      return resWidthOrHeight;
   }
    
   /**
    * Draw underline for specified widget.
    * 
    * @param   comp
    *          The base component to draw underline for.
    */
   @Override
   protected void drawUnderline(Widget<ChuiOutputManager> comp)
   {
      Point p = comp.screenLocation();
      
      screen().at(p);
  
      Color color = UiUtils.getDColor(this).withoutReverse();

      // underline does not fit into current rectangle
      ScreenBitmap bitmap = getAndResetBitmap();
      
      screen().hLine(comp.width(), color);
      setBitmap(bitmap);
   }
   
   /**
    * Draw frame title.
    * 
    * @param   origin
    *          Frame origin. 
    * @param   width
    *          Frame width.
    * @param   redir
    *          Redirected mode flag.
    * @param   tcolor
    *          Title color.
    */
   private void drawTitle(Point origin, int width, boolean redir, Color tcolor)
   {
      Point p = new Point(origin);

      if (redir)
         p.x += (width - config().title.length() + 1)/2 - 1;
      else
         p.x += (width - config().title.length())/2;

      screen().at(p).append(config().title, tcolor);
   }
   
   /**
    * Forces drawing of an empty line in redirected mode.
    */
   protected void forceNewLine()
   {
      screen().at(0, 0).append(' ', Color.DEFAULT);
      sync();
   }
   
   /**
    * Forces drawing of a box character for a number of consecutive lines
    * in redirected mode.
    *
    * @param    start
    *           line number in screen coordinates where the output starts
    * @param    lines
    *           number of lines to output to
    */
   @Override
   protected void forceBox(double start, double lines)
   {
      screen().box(new Point(0, start), new Dimension(1, lines), Color.DEFAULT);
   }

   /**
    * Forces drawing of skips in redirected mode.
    * 
    * @param    where
    *           Vertical position.
    */
   protected void forceSkips(int where)
   {
      screen().at(0, where);
   }
   
   /**
    * Reset colors.
    * 
    * @param   config
    *          Widget configuration.
    */
   @Override
   protected void resetColors(BaseConfig config)
   {
      config.dcolor = Color.NORMAL;
      config.pfcolor = Color.NORMAL;
   } 
   
   /**
    * Compare colors.  This method must be implemented in subclasses and provide
    * specific implementation as follow:
    * <p>
    * ChUI implementation must compare dcolor and pfcolor.
    * <p>
    * GUI implementation must compare bgcolor and fgcolor.
    * 
    * @param   config
    *          Widget configuration.
    * @param   colors
    *          Colors to be compared.
    *          
    * @return  The colors structure after comparison.         
    */
   @Override
   protected Colors compareColors(BaseConfig config, Colors colors)
   {
      colors.bgEquals = Utils.equals(colors.dcolor, config.dcolor);
      colors.fgEquals = Utils.equals(colors.pfcolor, config.pfcolor);
      return colors;
   }

   /**
    * Get the bounds of the frame's content.
    * 
    * @return   See above.
    */
   @Override
   protected Rectangle frameContentBounds()
   {
      return frameScroll.bounds();
   }
}