/*
** Module   : AbstractChildProcess.java
** Abstract : base implementation for interactive child processes
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150410 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.util.*;

/**
 * Provides a common implementation for an interactive child process which is independent from
 * any specific UI modality. 
 */
public abstract class AbstractChildProcess
implements InteractiveChildProcess
{
   /** Core client helpers. */
   protected ThinClient terminal = null;
   
   /** Batch mode flag. */
   protected boolean batch = false;
   
   /** Artificial terminal flag (<code>true</code> if the client has no native terminal). */
   protected boolean artificial = false;
   
   /**
    * Construct an instance.
    *
    * @param    terminal
    *           Core client helpers.
    * @param    batch
    *           Batch mode flag.
    * @param    artificial
    *           Artificial terminal flag (<code>true</code> if the client has no native
    *           terminal).
    */
   public AbstractChildProcess(ThinClient terminal, boolean batch, boolean artificial)
   {
      this.terminal   = terminal;
      this.batch      = batch;
      this.artificial = artificial;
   }
   
   /**
    * Reports if the client is in batch mode.
    *
    * @return   <code>true</code> if in batch mode.
    */
   public boolean inBatchMode()
   {
      return batch;
   }
   
   /**
    * Reports if the client is missing an already initialized native terminal session.
    *
    * @return   <code>true</code> if this is an artificial terminal (there is no native
    *           terminal).
    */
   public boolean isArtificialTerm()
   {
      return artificial;
   }
   
   /**
    * Wait for the user to press a key.
    *
    * @return   The key code of the key pressed to clear the pause or -1 on any failure.
    */
   public int pauseForUser()
   {
      return terminal.suspendedPause();
   }

   /**
    * Temporarily halts the client UI processing (key reading, normal drawing
    * mode)  After entering suspend mode, the current contents of the
    * terminal are cleared.  Use {@link #resume} to return to normal mode.
    */
   public void suspend()
   {
      terminal.suspend();
   }
   
   /**
    * Re-enables client UI processing (key reading, normal drawing mode).
    * Before resuming the current contents of the terminal are cleared.  This
    * is the mechanism to return to normal processing mode after having
    * called {@link #suspend}.
    */
   public void resume()
   {
      terminal.resume();
   }
}