/*
** Module   : ChuiOutputManager.java
** Abstract : Provides a generic interface to low level ChUI primitives.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20140805 Created initial version - extracted from the OutputManager class.
** 002 HC  20140911 GUI coordinate support - data types refactoring.
** 003 VIG 20141022 TC.screenSize replaced with window.size.
** 004 HC  20141021 Implemented window move/position support.
** 005 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 006 CA  20141123 Added API to check if a certain key can raise a STOP condition.
** 007 CA  20150429 When creating a rectangle from a location and dimension, use the base units
**                  to determine the coordinates for current client mode.
** 008 HC  20150612 Implemented support for modal windows.
** 009 EVK 20150415 Moved cursorInvalid to BasePrimitives.
** 010 HC  20160424 Added activation of the ChUI top-level window. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provides a generic interface to low level character user interface primitives.
 */
public class ChuiOutputManager
extends OutputManager<BasePrimitives>
{
   /** Bottom tee drawing char. */
   public static final char UACS_BTEE = '\u2534';

   /** Top tee drawing char. */
   public static final char UACS_TTEE = '\u252c';

   /** Lower left corner drawing char. */
   public static final char UACS_LLCORNER = '\u2514';

   /** Lower right corner drawing char. */
   public static final char UACS_LRCORNER = '\u2518';

   /** Upper right corner drawing char. */
   public static final char UACS_URCORNER = '\u2510';

   /** Upper left corner drawing char. */
   public static final char UACS_ULCORNER = '\u250c';

   /** Horizontal line drawing char. */
   public static final char UACS_HLINE = '\u2500';

   /** Vertical line drawing char. */
   public static final char UACS_VLINE = '\u2502';
   
   /** Up arrow char for vertical scrollbars. */
   public static final char UACS_UARROW = '\u2191';

   /** Down arrow char for vertical scrollbars. */
   public static final char UACS_DARROW = '\u2193';

   /** Left arrow char for horizontal scrollbars. */
   public static final char UACS_LARROW = '\u2190';

   /** Right arrow char for horizontal scrollbars. */
   public static final char UACS_RARROW = '\u2192';

   /** Checker board char for scrollbars body. */
   public static final char UACS_CKBOARD = '\u2592';

   /** Current selection char for selection lists. */
   public static final char UACS_SELECT = '\u2219'; // This is Unicode bullet char

   /** Direct drawing routines. */
   private BasePrimitives directOutput;

   /** Widget registry. */
   private WidgetRegistry<ChuiOutputManager> registry = new WidgetRegistry<ChuiOutputManager>();
   
   /**
    * Create a new output manager instance, with the specified driver.
    *
    * @param    driver
    *           Implements low level input/output primitives.  Must NOT be <code>null</code>.
    */
   public ChuiOutputManager(ScreenDriver<ChuiOutputManager> driver)
   {
      super(driver);
   }

   /**
    * Activates the provided window instance, depending on the output manager type.
    * <p>
    * For ChUI case, it makes the window visible immediately.
    * 
    * @param    window
    *           A window.
    */
   @Override
   public void activateWindow(TopLevelWindow<?> window)
   {
      // ChUI mode uses a single default window, which can be used by any auth plugins.
      // GUI mode requires the auth plugin to establish its own default window.
      window.setVisible(true);
      WindowManager.windowActivated(window.getId().asInt(), false);
   }
   
   /**
    * Initialize this output manager.
    */
   @Override
   public void init()
   {
      // install direct output driver with shadow buffer
      directOutput = new DirectPrimitives((BasePrimitives) driver.getPrimitives());
   
      // install the double buffered terminal driver
      DoubleBufferedTerminal newOutput = new DoubleBufferedTerminal(directOutput.screenHeight(), 
                                                                    directOutput.screenWidth());
      defaultOutput = newOutput;
      output = newOutput;
   }

   /**
    * Access driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   @Override
   public ChuiWidgetFactory getFactory()
   {
      return (ChuiWidgetFactory) driver.getFactory();
   }
   
   /**
    * Get the widget registry for this output manager.
    * 
    * @return   See above.
    */
   @Override
   public WidgetRegistry<ChuiOutputManager> getRegistry()
   {
      return registry;
   }

   /**
    * Clear area occupied by specified widget.
    * 
    * @param   widget
    *          Reference widget. 
    */
   @Override
   public void clear(Widget<?> widget)
   {
      if (widget == null || isRedirected())
         return;

      clear(WidgetId._DEFAULT_WINDOW_ID, 
            widget.screenLocation(), 
            widget.dimension(), 
            output.widgetColor(widget));
   }

   /**
    * Clear screen.
    */
   public void clearScreen()
   {
      driver.clear();

      directOutput.cursorAt(0, 0);
      directOutput.sync();
   }
   
   /**
    * Append character with specified color at current cursor location and 
    * advance the cursor.
    *  
    * @param   chr
    *          Character to append.
    * @param   color
    *          Color and attribute.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public ChuiOutputManager append(int chr, Color color)
   {
      ScreenBitmap bitmap = screenBitmap();

      if (!output.isCursorInvalid() && bitmap.canDraw(cursor()))
      {
         output.append(chr, color);
      }
      else
      {
         boolean cursorInvalid = !output.cursorAt(output.getX() + 1, output.getY());
         output.setCursorInvalid(cursorInvalid);
      }
      
      return this;
   }

   /**
    * Append string with specified color at current cursor location and 
    * advance the cursor.
    *  
    * @param   text
    *          Character to append.
    * @param   color
    *          Color and attribute.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public ChuiOutputManager append(String text, Color color)
   {
      if (output.isCursorInvalid())
      {
         return this;
      }
   
      ScreenBitmap bitmap = screenBitmap();

      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.append(text, color);
         return this;
      }
   
      NativePoint p = cursor();
   
      for (NativePoint frag : bitmap.getHorizontalFragments(p, text.length()))
      {
         at(p.x + frag.x, p.y);
         output.append(text.substring(frag.x, frag.y + 1), color);
      }
   
      // reposition the cursor to the first position behind the string
      // no matter how the clipping went
      return at(p.x + text.length(), p.y);
   }

   /**
    * Draw horizontal line.
    * 
    * @param   len
    *          Line length.
    * @param   color
    *          Line color.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public ChuiOutputManager hLine(double len, Color color)
   {
      if (output.isCursorInvalid())
      {
         return this;
      }
      
      int nlen = coordinates().widthToNative(len);

      ScreenBitmap bitmap = screenBitmap();

      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.hLine(nlen, color);
         return this;
      }

      NativePoint p = cursor();

      // split the line into drawable fragments and draw them
      List<NativePoint> frags = bitmap.getHorizontalFragments(p, nlen);
      
      for (NativePoint frag : frags)
      {
         int start = frag.x;
         int end = frag.y;
         int length = end - start + 1;

         at(p.x + start, p.y);
         output.hLine(length, color);
      }
      
      return this;
   }

   /**
    * Draw a vertical line of the specified length starting at the current
    * cursor position.
    * 
    * @param   len
    *          Vertical line length.
    * @param   color
    *          Line color.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public ChuiOutputManager vLine(double len, Color color)
   {
      if (output.isCursorInvalid())
      {
         return this;
      }

      int nlen = coordinates().heightToNative(len);
      
      ScreenBitmap bitmap = screenBitmap();

      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.vLine(nlen, color);
         return this;
      }

      NativePoint p = cursor();

      // split the line into drawable fragments and draw them
      List<NativePoint> frags = bitmap.getVerticalFragments(p, nlen);
      
      for (NativePoint frag : frags)
      {
         at(p.x, p.y + frag.x);
         output.vLine(frag.y - frag.x + 1, color);
      }
      
      return this;
   }

   /**
    * Draw box at specified location with provided size and color.
    * 
    * @param   origin
    *          Upper left corner of the box.
    * @param   size
    *          Box dimensions.
    * @param   color
    *          Color and attribute of the box lines.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public ChuiOutputManager box(Point origin, Dimension size, Color color)
   {
      return box(new Rectangle(origin, size, coordinates().baseUnits()), color);
   }

   /**
    * Draw box at interactive terminal.
    * 
    * @param   origin
    *          Upper left corner.
    * @param   size
    *          Box dimensions.
    * @param   color
    *          Box color.
    */
   public void boxToScreen(Point origin, Dimension size, Color color)
   {
      ScreenBitmap old = output.getScreenBitmap();
      
      try
      {
         // set output to screen
         ScreenBitmap bitmap = defaultOutput.getScreenBitmap();
      
         output.setScreenBitmap(bitmap);

         box(origin, size, color);
      }
      finally
      {
         // restore the bitmap
         output.setScreenBitmap(old);
      }
   }

   /**
    * Obtain border color for specified widget.
    * 
    * @param   widget
    *          Widget for which border color should be obtained.
    *          
    * @return  Widget border color.
    */
   public Color borderColor(Widget<ChuiOutputManager> widget)
   {
      return output.borderColor(widget);
   }

   /**
    * Set cursor position.
    * 
    * @param   x
    *          Cursor X.
    * @param   y
    *          Cursor Y.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public ChuiOutputManager at(double x, double y)
   {
      int nx = coordinates().columnToNative(x);
      int ny = coordinates().rowToNative(y);
      boolean cursorInvalid = !output.cursorAt(nx, ny);
      output.setCursorInvalid(cursorInvalid);
      return this;
   }

   /**
    * Set cursor position.
    * 
    * @param   p
    *          New cursor position.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public ChuiOutputManager at(Point p)
   {
      return at(p.x, p.y);
   }

   /**
    * Set cursor position.
    * 
    * @param   p
    *          New cursor position.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public ChuiOutputManager at(NativePoint p)
   {
      boolean cursorInvalid = !output.cursorAt(p.x, p.y);
      output.setCursorInvalid(cursorInvalid);
      return this;
   }
   
   /**
    * Set real cursor position.
    * 
    * @param   p
    *          New cursor position.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public ChuiOutputManager stayAt(Point p)
   {
      int nx = coordinates().columnToNative(p.x);
      int ny = coordinates().rowToNative(p.y);
      output.cursorStay(nx, ny);
      
      return this;
   }

   /**
    * Set real cursor position.
    * 
    * @param   p
    *          New cursor position.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public ChuiOutputManager stayAt(NativePoint p)
   {
      output.cursorStay(p.x, p.y);
      
      return this;
   }
   
   /**
    * Obtain the direct output worker.
    * 
    * @return The output worker used to draw directly to the screen with
    *         ncurses
    */
   public BasePrimitives getDirectOutput()
   {
      return directOutput;
   }

   /**
    * Get current cursor position.
    * 
    * @return  current cursor position.
    */
   public NativePoint cursor()
   {
      return new NativePoint(output.getX(), output.getY());
   }
   
   /**
    * Draw box at specified location with provided size and color.
    * 
    * @param   rect
    *          The box to draw.
    * @param   color
    *          Color and attribute of the box lines.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   private ChuiOutputManager box(Rectangle rect, Color color)
   {
      ScreenBitmap bitmap = screenBitmap();

      NativeRectangle nrect = coordinates().rectangleToNative(rect);
      if (bitmap.hasCoordinatesOutsideScreen(nrect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(nrect))
         {
            return this;
         }
   
         if (debug)
         {
            System.err.println("box: received rectangle with incorrect" +
                               " coordinates: " + nrect.toString());
         }
   
         nrect = fixCoordinates(WidgetId._DEFAULT_WINDOW_ID, nrect);
   
         if (nrect == null)
         {
            return this;
         }
      }
   
      int left   = nrect.left();
      int top    = nrect.top();
      int bottom = nrect.bottom();
      int right  = nrect.right();
   
      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.box(left, top, right, bottom, color);
         return this;
      }
   
      boolean corners = output.hasBoxBorderCorners();
   
      int width = right - left + 1;
      int height = bottom - top + 1;
      int corner = 0;
   
      if (corners)
      {
         width = width - 2;
         height = height - 2;
         corner = 1;
      }
   
      // upper-left corner
      if (corners && bitmap.canDraw(top, left))
      {
         output.drawULCorner(left, top, color);
      }
   
      // upper border
      if (width > 0)
      {
         drawHBorder(left + corner, top, width, color);
      }
   
      // upper-right corner
      if (corners && bitmap.canDraw(top, right))
      {
         output.drawURCorner(right, top, color);
      }
   
      // lower-left corner
      if (corners && bitmap.canDraw(bottom, left))
      {
         output.drawLLCorner(left, bottom, color);
      }
   
      // lower border
      if (width > 0)
      {
         drawHBorder(left + corner, bottom, width, color);
      }
   
      // lower-right corner
      if (corners && bitmap.canDraw(bottom, right))
      {
         output.drawLRCorner(right, bottom, color);
      }
   
      // vertical borders
      if (height > 0)
      {
         drawVBorder(left, top + corner, height, color);
         drawVBorder(right, top + corner, height, color);
      }
      
      return this;
   }

   /**
    * Draw the upper or lower border.
    * 
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   width
    *          The border width.
    * @param   color
    *          The number of the color-pair (foreground+background)
    */
   private void drawHBorder(double col, double row, double width, Color color)
   {
      int ncol = coordinates().columnToNative(col);
      int nrow = coordinates().rowToNative(row);
      int nwidth = coordinates().widthToNative(width);
            
      ScreenBitmap bitmap = screenBitmap();

      // split the line into drawable fragments and draw them
      List<NativePoint> frags = bitmap.getHorizontalFragments(nrow, ncol, nwidth);
      for (NativePoint frag : frags)
      {
         int start = frag.x;
         int end = frag.y;
         int length = end - start + 1;

         output.drawHBorder(ncol + start, nrow, length, color);
      }
   }

   /**
    * Draw the left or right border.
    * 
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   height
    *          The border height.
    * @param   color
    *          The number of the color-pair (foreground+background)
    */
   private void drawVBorder(double col, double row, double height, Color color)
   {
      ScreenBitmap bitmap = screenBitmap();

      int ncol = coordinates().columnToNative(col);
      int nrow = coordinates().rowToNative(row);
      int nheight = coordinates().heightToNative(height);

      // split the line into drawable fragments and draw them
      List<NativePoint> frags = null;
      if (isRedirected() == false)
      {
         frags = bitmap.getVerticalFragments(nrow, ncol, nheight);
      }
      else
      {
         frags = bitmap.getVerticalFragmentsAnyColumn(nrow, nheight);
      }

      for (NativePoint frag : frags)
      {
         int start = frag.x;
         int end = frag.y;
         int length = end - start + 1;

         output.drawVBorder(ncol, nrow + start, length, color);
      }
   }

   /**
    * Get a copy of the current screen bitmap.
    * 
    * @return  a copy of the current screen bitmap.
    */
   public ScreenBitmap getBitmapCopy()
   {
      return super.getBitmapCopy(WidgetId._DEFAULT_WINDOW_ID);
   }

   /**
    * Restore the bitmap to a previous state.
    * 
    * @param   bitmap
    *          The bitmap instance to set as the output primitive's bitmap.
    */
   public void setBitmap(ScreenBitmap bitmap)
   {
      super.setBitmap(WidgetId._DEFAULT_WINDOW_ID, bitmap);
   }
   
   /**
    * Check if the given key can raise the STOP condition.
    * 
    * @param    key
    *           The key code to be checked.
    * 
    * @return   <code>true</code> if the key is the {@link Key#VK_CTRL_C} key.
    */
   @Override
   protected boolean raiseStopCondition(int key)
   {
      return key == Key.VK_CTRL_C;
   }

   /**
    * Determine the screen bitmap associated with the default window.
    * 
    * @return   The {@link ScreenBitmap} associated with this window.
    */
   protected ScreenBitmap screenBitmap()
   {
      return screenBitmap(WidgetId._DEFAULT_WINDOW_ID);
   }

   /**
    * Check if the window being drawn needs to be overridden.
    * <p>
    * In ChUI, the current top window will always be returned.  This might be some other window
    * than default-window.
    *  
    * @param    window
    *           The currently active window.
    *           
    * @return   See above.
    */
   protected TitledWindow<?> overrideWindow(TopLevelWindow<?> window)
   {
      return WindowManager.topWindow();
   }
}