/*
** Module   : SelectionTracker.java
** Abstract : helps to track mouse button events used for selection
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20110921 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver.swing;

import java.awt.event.*;
import javax.swing.event.*;

/**
 * Tracks mouse button events used for selection of a rectangular area of the
 * CHUI simulator screen.
 */
public class SelectionTracker
extends MouseInputAdapter
{
   /** CHUI simulator instance. */
   private ChuiSimulator sim = null;
   
   /** Mouse button that represents the popup trigger event. */
   private int avoid = MouseEvent.NOBUTTON;
   
   /** Mouse button used to initiate the selection tracking. */
   private int active = MouseEvent.NOBUTTON;
   
   /**
    * Construct an instance.
    *
    * @param    sim
    *           The CHUI simulator for which selection tracking is done.
    */
   public SelectionTracker(ChuiSimulator sim)
   {
      this.sim = sim;
   }
   
   /**
    * Process a mouse button pressed event. If the button press is a popup
    * trigger event, it is ignored. Otherwise, the button press starts a
    * selection operation in the CHUI simulator. The button used to start
    * the selection is saved off so that only events on that same button
    * will affect the selection processing. The result of the selection
    * operation is that the selected rectangle of text on the simulator's
    * screen is identified (for copying to the clipboard) and a selection
    * box is drawn around the selection area to give the user feedback on
    * what exactly has been selected.
    *
    * @param    evt
    *           The mouse button event being processed.
    */
   public void mousePressed(MouseEvent evt)
   {
      // don't process if this is supposed to bring up the popup
      if (evt.isPopupTrigger())
      {
         // the first time this is seen, save off the button to avoid in the
         // future
         if (avoid == MouseEvent.NOBUTTON)
         {
            avoid = evt.getButton();
         }
         
         return;
      }
      
      if (!sim.canSelectionTrack())
      {
         return;
      }
      
      int button = evt.getButton();
      
      // the following code will fail to work properly the first time the popup
      // menu is brought up on mouseReleased() using this same button, since
      // this method will start a selection but then in the mouseReleased() we
      // will realize that this is a popup trigger and we will have to cancel
      // the selection
      
      if (button != avoid)
      {
         active = button;
         sim.startSelection(evt.getX(), evt.getY());
      }
   }

   /**
    * Process a mouse dragged event. If the mouse has been moved while
    * the button is still pressed, a dragged event occurs. So long as this
    * dragged event was generated while an active selection is underway,
    * the selection rectangle in the CHUI simulator will be updated by
    * this method.
    *
    * @param    evt
    *           The mouse dragged event being processed.
    */
   public void mouseDragged(MouseEvent evt)
   {
      // the button in the event is always set to NOBUTTON, but we can check
      // if we have an active selection on using internal state
      if (active != MouseEvent.NOBUTTON)
      {
         sim.updateSelection(evt.getX(), evt.getY());
      }
   }

   /**
    * Process a mouse button released event. If the button release is a popup
    * trigger event, it is ignored. Otherwise, so long as a selection operation
    * is in progress and the active selection button is the button that was
    * released, this method will end the selection operation in the CHUI
    * simulator.  The selection at that point is final. The drawn box will not
    * change, unless the selection is canceled by the user or a new selection
    * operation starts.  
    *
    * @param    evt
    *           The mouse button event being processed.
    */
   public void mouseReleased(MouseEvent evt)
   {
      // don't process if this is supposed to bring up the popup
      if (evt.isPopupTrigger())
      {
         // the first time this is seen, save off the button to avoid
         if (avoid == MouseEvent.NOBUTTON)
         {
            avoid = evt.getButton();
         }
         
         // we started a selection in mousePressed() but now we realize that
         // we should not have done so, since the popup menu is going to
         // display, so we need to cancel the selection now
         if (active != MouseEvent.NOBUTTON)
         {
            active = MouseEvent.NOBUTTON;
            sim.cancelSelection();
         }
         
         return;
      }
      
      if (!sim.canSelectionTrack())
      {
         return;
      }
      
      int button = evt.getButton();
      
      if (button != avoid && button == active)
      {
         sim.endSelection(evt.getX(), evt.getY());
         active = MouseEvent.NOBUTTON;
      }
   }
}