/*
** Module   : AbstractDriver.java
** Abstract : simple base class for common screen driver implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150401 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver;

import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Simple base class for common screen driver implementation features. This should not have
 * any features that are specific to ChUI or GUI. Instead, all features must be independent
 * of any UI modality.  In addition, no driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 */
public abstract class AbstractDriver<P extends OutputPrimitives, O extends OutputManager<P>>
implements ScreenDriver<O>
{
   /** This driver-specific widget factory */
   protected WidgetFactory<O> factory = null;
   
   /** Direct to screen output primitives. */
   protected P direct;
   
   /** UI modality flag: <code>true</code> for ChUI, <code>false</code> for GUI. */
   private boolean isChui;
   
   /** Batch mode flag. */
   private boolean batch = false;
   
   /** Background flag (signifies if the client is in the foreground session). */
   private boolean background = false;
   
   /** Terminal type name. */
   private String term;
   
   /**
    * Construct a new instance.
    *
    * @param    factory
    *           The widget factory to use for this driver.
    * @param    direct
    *           The output primitives.
    * @param    isChui
    *           <code>true</code> for ChUI, <code>false</code> for GUI.
    * @param    term
    *           Terminal type string.
    */
   public AbstractDriver(WidgetFactory<O> factory, P direct, boolean isChui, String term)
   {
      this.factory = factory;
      this.direct  = direct;
      this.isChui  = isChui;
      this.term    = term;
   }

   /**
    * Checks if the current driver is running in batch mode.
    * 
    * @return   Always <code>false</code>.
    * 
    */
   public boolean inBatchMode()
   {
      return batch;
   }

   /**
    * Checks if the current driver is running in background mode.
    * 
    * @return   Always <code>false</code>.
    * 
    */
   public boolean inBackgroundMode()
   {
      return background;
   }

   /**
    * Obtain system/implementation specific terminal type string.
    * <p>
    * At present this is used to implement 4GL TERMINAL function and do some 
    * terminal-specific workarounds. At present workarounds exist for "vt320" 
    * terminal type. If implementation requires no special workarounds, then 
    * it should avoid returning "vt320" as terminal type string.  
    * 
    * @return   The current terminal type string.
    */
   public String terminalType()
   {
      return term;
   }

   /**
    * Change terminal type to specified one. The terminal-based implementation
    * should try to change terminal type to specified one. It is assumed that 
    * passed terminal string is specific to particular driver implementation 
    * and recognized by it. 
    * 
    * @param    term
    *           The terminal type.
    */
   public void setTerminalType(String term)
   {
      this.term = term;
   }

   /**
    * Get driver-specific direct output primitives implementation.
    *
    * @return  driver-specific instance of direct screen output primitives.
    */
   public P getPrimitives()
   {
      return direct;
   }

   /**
    * Get driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   public WidgetFactory<O> getFactory()
   {
      return factory;
   }
   
   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
   } 

   /**
    * Get the default WINDOW-SYSTEM used by this driver. See {@link ClientParameters} for the list
    * of possible values.
    * 
    * @return   See above.
    */
   public String getWindowSystem()
   {
      return isChui ? ClientParameters.WINDOW_SYSTEM_TTY
                    : ClientParameters.WINDOW_SYSTEM_MS_WINDOWS;
   }
   
   /**
    * Get screen interface type.
    * 
    * @return  <code>true</code> is ChUI <code>false</code> is GUI
    */
   public boolean isChui()
   {
      return isChui;
   }

   /**
    * Perform driver shutdown.
    */
   public void shutdown()
   {
      // no-op
   }

   /**
    * Assigns the flag determining if the current driver is running in batch mode.
    * 
    * @param    batch
    *           The new state of the batch mode flag.
    * 
    */
   protected void setBatchMode(boolean batch)
   {
      this.batch = batch;
   }

   /**
    * Assigns the flag determining if the current driver is running in background mode.
    * 
    * @param    background
    *           The new state of the background mode flag.
    * 
    */
   protected void setBackgroundMode(boolean background)
   {
      this.background = background;
   }
}