/*
** Module   : ScreenDriver.java
** Abstract : common interface for all output drivers
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20101202 Method getAcsChar() replaced with getGraphics().
** 004 SIY 20110112 Changes related to new color handling scheme.
** 005 SIY 20110927 Updated javadoc.
** 006 GES 20111005 Updated import to match refactoring.
** 007 SIY 20111202 Added explicit import for List.
** 008 SIY 20111221 Added widget factory accessor.
** 009 EVL 20130725 Removed CHARVA package, using javadoc KeyProcessor from console driver.
** 010 CA  20131004 Added API to identify the default window-system used by a certain driver.
** 011 EVL 20131212 Added method to check if the driver is in batch mode.
** 012 EVL 20140115 Added method to check if the driver is in background mode.
** 013 HC  20140119 Implemented runtime support for the COLOR  protermcap-attribute.
** 014 MAG 20140512 Add method to check screen interface type ChUI or GUI. 
** 015 CA  20140805 Added generics to force the abstract widget classes to avoid implementation
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager. Added support for mouse events. Removed
**                  hasColors, as it is not used.
** 016 CA  20141123 Added API to retrieve the available screen size where windows can be created.
** 017 GES 20150401 Minor cleanup.
** 018 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
** 019 IAS 20160217 Added support for LOAD-MOUSE-POINTER and SET-WAIT-STATE.
** 020 SBI 20160412 Added openURL(String path) to open the target page in the browser.
** 021 CA  20170203 Notify the driver that the client is ready (useful for embedded mode, to 
**                  notify the embedded app the client can handle messages).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.DriverPrimitives;
import com.goldencode.p2j.ui.client.chui.driver.console.KeyProcessor;

/**
 * Common interface for all screen drivers.
 * <p>
 * This interface contains all implementation-specific methods which should be 
 * implemented from scratch for each particular driver implementation. The part
 * of the driver which is more hardware independent is listed in 
 * {@link OutputPrimitives} interface and {@link DriverPrimitives} abstract 
 * class.
 * <p>
 * General requirements to the implementation:<br>
 * At present only character based UI is supported. Therefore, all interaction
 * with the driver assumes existence of some kind of character-based terminal. 
 * All coordinates passed are coordinates of the character row and column (Y and 
 * X respectively). All coordinates are 0-based and Y axis has down direction. 
 * Upper left corner has coordinates (0,0) and lower right -  
 * (@link OutputPrimitives#screenWidth()} - 1, 
 * {{@link OutputPrimitives#screenHeight()} - 1). There is no limitations on 
 * actual terminal size, but at present dynamic screen size change is not 
 * supported - once application starts it assumes that terminal has fixes size 
 * which remains stable until application if finished. Particular implementation
 * may return terminal screen size different than phisical screen size, but
 * in this case it is responsible for handling of necessary 
 * scrolling/panning/clipping, because P2J has no support for this funtionality
 * and assumes that entire reported by driver screen area is visible to user.
 * Note that requirement for fixed size of terminal screen applies even if 
 * terminal is suspended and then resumed. Handling size changes (if necessary)
 * is up to the driver implementation.
 */
public interface ScreenDriver<O extends OutputManager<?>>
{
   /**
    * Get the {@link OutputManager} instance associated with this driver.
    * 
    * @param    cfg
    *           The initial configuration for this driver.
    *
    * @return   See above.
    */
   public O createOutputManager(BootstrapConfig cfg);
   
   /**
    * Set cursor status (ON/OFF).
    * <p>
    * The implementation of this method should show/hide actual screen 
    * cursor visible to user. Remaining code assumes that actual cursor status
    * visible to user is properly changed upon exit from this method.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    *
    * @return  <code>true</code> if operation was successful.
    * 
    * @throws  IllegalStateException
    */
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException;

   /**
    * Clear screen.
    * <p>
    * This method should clear the screen and perform necessary flushing of the 
    * system-specific buffers if necessary. Users of this method assume that 
    * actual screen appearance as it is visible to the user is properly changed
    * upon exit from this method.
    */
   public void clear();

   /**
    * Provide an audible feedback to user by generating a beep signal.
    * <p>
    * There is no specific requirements to this method and actual implementation
    * may do nothing if there is no way to perform audible notification to the 
    * user.
    */
   public void beep();

   /**
    * Checks if the current driver is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public boolean inBatchMode(); 

   /**
    * Checks if the current driver is running in background mode.
    * 
    * @return  <code>true</code> for background mode, <code>false</code> otherwise.
    * 
    */
   public boolean inBackgroundMode(); 

   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * <p>
    * The list of codes and detailed description is provided in {@link Keyboard}
    * class. The {@link KeyProcessor} can serve as an example of the key 
    * translation implementation.
    * 
    * @return  key code or -1 if no keys were pressed since last call.
    */
   public int readKey();

   /**
    * Reset current screen mode.
    * <p>
    * Upon exit from this method P2J assumes that screen is in correct state
    * and can be used for interaction with the user.
    */
   public void resetMode();

   /**
    * Suspend all screen I/O operations. If implementation is terminal-based, 
    * then terminal should be released to allow other application executed 
    * from P2J access terminal.
    */
   public void suspend();

   /**
    * Resume normal driver operation. If implementation is terminal-based,
    * they terminal should be acquired again and driver should be ready to 
    * perform regular screen I/O operations.
    */
   public void resume();

   /**
    * Obtain system/implementation specific terminal type string.
    * <p>
    * At present this is used to implement 4GL TERMINAL function and do some 
    * terminal-specific workarounds. At present workarounds exist for "vt320" 
    * terminal type. If implementation requires no special workarounds, then 
    * it should avoid returning "vt320" as terminal type string.  
    * 
    * @return  terminal type string.
    */
   public String terminalType();

   /**
    * Change terminal type to specified one. The terminal-based implementation
    * should try to change terminal type to specified one. It is assumed that 
    * passed terminal string is specific to particular driver implementation 
    * and recognized by it. 
    * 
    * @param   name
    */
   public void setTerminalType(String name);

   /**
    * Get driver-specific direct output primitives implementation.
    *
    * @return  driver-specific instance of direct screen output primitives.
    */
   public OutputPrimitives getPrimitives();

   /**
    * Get driver-specific widget factory.
    * 
    * @return   The reference to the widget factory.
    */
   public WidgetFactory<?> getFactory();
   
   /**
    * Get driver-specific child process factory.
    * 
    * @return   The reference to the factory.
    */
   public ChildProcessFactory getChildProcessFactory();
   
   /**
    * Driver initialization.
    */
   public void init();

   /**
    * Get the default WINDOW-SYSTEM used by this driver. See {@link ClientParameters} for the list
    * of possible values.
    * 
    * @return   See above.
    */
   public String getWindowSystem();
   
   /**
    * Get screen interface type.
    * 
    * @return  <code>true</code> is ChUI <code>false</code> is GUI
    */
   public boolean isChui();

   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   public boolean captureMouseEvents(boolean state);
   
   /**
    * Handle the specified event, in the context of the specified widget.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   public void handleMouseEvent(int sourceID, MouseEvent evt);
   
   /**
    * Perform driver shutdown.
    */
   public void shutdown();

   /**
    * Lock the mouse pointer (suppress widget-specific changes) or unlock it
    * 
    * @param widgetId
    *          the id of the active DropDown which caused the pointer lock or -1
    * @param ptr
    *          the locked mouse pointer or <code>null</code> for unlock
    *           
    */
   public default void lockMousePointer(int widgetId, MousePtrWrapper ptr)
   {
      // do nothing
   }

   /**
    * Push information about the loaded mouse pointer for widget 
    * 
    * @param windowId
    *           to level window id
    * @param wid
    *           widgetId
    * @param ptr
    *           mouse pointer
    */
   public default void propagate(int windowId, int wid, MousePtrWrapper ptr)
   {
      // do nothing
   }
   
   /**
    * Opens the target page or resource in a browser's window.
    * 
    * @param    path
    *           The url path to the target resource.
    */
   public default void openURL(String path)
   {
      // do nothing
   }
   
   /**
    * Notify the driver that the client is ready (useful for embedded mode, to notify the embedded
    * app the client can handle messages).
    */
   public default void clientReady()
   {
      // do nothing
   }
}