/*
** Module   : EmbeddedWebServerImpl.java
** Abstract : embedded web server shared implementation
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150312 Created initial version by moving code from the ChUI web driver and making it
**                  agnostic to web client type.
** 002 CA  20150820 Added waitInitialization, which waits for a notification from the web page
**                  that it has finished initialized.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

// selective import to avoid name conflicts
import java.util.logging.Logger;
import java.util.logging.Level;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;

import org.eclipse.jetty.util.resource.*;
import org.eclipse.jetty.server.handler.*;

/**
 * Web screen driver shared implementation of an embedded web server. 
 */
public class EmbeddedWebServerImpl
implements EmbeddedWebServer
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(EmbeddedWebServerImpl.class.getName());
   
   /** URI format specification. */
   private static final String spec = "https://%s:%d%s";
   
   /** WebSocket target. */
   private static final String AJAX_TARGET = "/ajax";
   
   /** Embedded web server instance. */
   private GenericWebServer server = new GenericWebServer();
   
   /** Driver name to use in logging. */
   private String name = null;
   
   /** WebSocket factory. */
   private GenericWebSocketCreator sfact = null;
   
   /** Classpath resource instance. */
   private Resource cpres = null;
   
   /** Page handlers for our specific targets. */
   private WebPageHandler[] hdlrs = null;
   
   /**
    * Create an instance.
    *
    * @param    name
    *           Driver name to use in logging.
    * @param    sfact
    *           WebSocket factory.
    * @param    cpres
    *           Classpath resource instance.
    * @param    hdlrs
    *           Page handlers for our specific targets.
    */
   public EmbeddedWebServerImpl(String                  name,
                                GenericWebSocketCreator sfact,
                                Resource                cpres,
                                WebPageHandler[]        hdlrs)
   {
      this.name  = name;
      this.sfact = sfact;
      this.cpres = cpres;
      this.hdlrs = hdlrs;
   }
   
   /**
    * Startup the embedded web server using the application server's <code>KeyStore</code> to
    * build a suitable SSL context.
    * 
    * @param    keyStore
    *           The exported server KeyStore. 
    * @param    config
    *           Client configuration.
    * @param    port
    *           An explicit port to start the web server or 0 to let the OS assign one.
    * @param    host
    *           An explicit host to start the web server or <code>null</code> to use the server's
    *           host.
    */
   @Override
   public String startupServer(ServerKeyStore keyStore, BootstrapConfig config, String host, int port)
   {
      String uri = null;
      
      // startup embedded server
      try
      {
         // authorization token
         String token = keyStore.getAuthorizationToken();
         
         // configure and startup server
         server.setKeyStore(keyStore);
         server.setBootstrapConfig(config);
         
         // process each handler
         for (int i = 0; i < hdlrs.length; i++)
         {
            hdlrs[i].setAuthorizationToken(token);
            
            ContextHandler contextHandler = new ContextHandler();
            contextHandler.setHandler(hdlrs[i]);
            
            server.addHandler(contextHandler);
         }
         
         server.addHandler(new DojoToolkitHandler());
         
         ResourceHandler cliRes = new ResourceHandler();
         cliRes.setDirectoriesListed(false);         
         
         ContextHandler cliCtx = new ContextHandler();
         cliCtx.setContextPath("/client");
         cliCtx.setBaseResource(cpres);
         cliCtx.setHandler(cliRes);
         
         ResourceHandler commonRes = new ResourceHandler();
         commonRes.setDirectoriesListed(false);         
         
         ContextHandler commonCtx = new ContextHandler();
         commonCtx.setContextPath("/common");
         commonCtx.setBaseResource(calcResourceBase(EmbeddedWebServerImpl.class, "res"));
         commonCtx.setHandler(commonRes);
         
         ContextHandlerCollection list = new ContextHandlerCollection();
         list.setHandlers(new ContextHandler[] { cliCtx, commonCtx });
         
         server.addHandler(list);
         
         server.addWebSocketHandler(AJAX_TARGET, sfact);
         
         // TODO: should we add a DefaultHandler instance at the end to get 404 processing?
         
         server.startup(host, port);
      }
      
      catch (Exception e)
      {
         LOG.logp(Level.SEVERE,
                  name,
                  "startupServer",
                  LogHelper.generate("Cannot startup server."),
                  e);         
      }
      
      finally
      {
         uri = this.getServerURI();
      }
      
      return uri;
   }

   /**
    * Shutdown the embedded web server.
    */
   @Override
   public void shutdownServer()
   {
      try
      {
         server.shutdown();
      }
      catch (Exception e)
      {
         LOG.logp(Level.SEVERE,
                  name,
                  "shutdownServer",
                  LogHelper.generate("Cannot shutdown server."),
                  e);         
      }
   }

   /**
    * Get the embedded web server's URI.
    * 
    * @return   A string representing the server's URI.
    */
   @Override
   public String getServerURI()
   {
      String uri = null;
      
      if (server.isRunning())
      {
         uri = String.format(spec, server.getHost(), server.getPort(), server.getContextPath());
      }
      
      return uri;
   }
   
   /**
    * Calculate the path in the jar file for the given path as an extention of the origin
    * class' package and return that resource base as a classpath resource. The result can
    * be used in a resource handler as the location for accessing static resources.
    *
    * @param    origin
    *           A class in the package being targeted as the starting path for the resource.
    *           The package will be looked up from this class, any class in that package will
    *           work as the class is not used for anything else.
    * @param    path
    *           The relative sub-directory to add to the package to get the full base path.
    *           This should NOT have either a leading or trailing path separator ('/') included.
    *
    * @return   The calculated resource base.
    */
   public static Resource calcResourceBase(Class<?> origin, String path)
   {
      String pkg = origin.getPackage().getName();
      
      StringBuilder sb = new StringBuilder();
      sb.append(pkg.replace('.', '/')).append('/').append(path).toString();
      
      return Resource.newClassPathResource(sb.toString());
   }

   /**
    * Wait for the page to notify that it has fully initialized its JS code.
    * <p>
    * This implementation is a no-op.
    */
   @Override
   public void waitInitialization()
   {
      // no-op
   }
   
   /**
    * Wait for the page to notify that it has fully initialized its JS code.
    * 
    * @param    webSocket
    *           The socket where to wait for a message.
    */
   public void waitInitialization(WebClientProtocol webSocket)
   {
      webSocket.waitForResult(WebClientMessageTypes.MSG_PAGE_LOADED);
   }
}