/*
** Module   : p2j.clipboard.js
** Abstract : core clipboard implementation.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140701 First version.
** 002 GES 20150324 Moved this class to a common resource base and made it generic (so that it
**                  can be used for both ChUI and GUI web clients. Client-specific portions have
**                  been moved to p2j.clipboard_helpers which will have different versions for
**                  ChUI and GUI.
** 003 SBI 20150731 Applied "use strict" directive, fixed undeclared variables
**                  and end of lines (EOLs) by simulating key "Enter" pressed.
** 004 GES 20150709 Moved mouse listeners, drag handling and most selection processing to ChUI
**                  specific helpers. Removed reference to window.Clipboard since that is
**                  undefined on any browser (it must have been copied from some samples on
**                  the web which had a typo because I can find no evidence it is a real
**                  thing). Added cut event support. Added base implementation to respond to
**                  GUI processing, there are still open issues.
** 005 SBI 20151014 Implemented to set the current editors selection.
** 006 SBI 20160301 The CHUI editor code requires EOL's simulating.
** 007 SBI 20160523 Fixed copy/cut/paste work flow.
**     SBI 20160603 Added Safari COMMAND+C processing.
**     SBI 20160614 Changed to consume copy/cut events if the current selection is invalid.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Core clipboard implementation that calls client-specific callbacks for ChUI or GUI selection
 * and rendering.
 */
p2j.clipboard = (function() 
{
   /** Clipboard object which will be returned from this initialization function. */
   var me = {};
   
   /** Clipboard enabled/disabled flag. */
   me.enabled = false;
   
   /** CTRL-V pressed flag. */
   var ctrlv = false;
   
   /** Hidden input element used on copy/paste operations. */
   var inputText;
   
   /** The current text selection. */
   var userSelection;
   
   /** True value indicates that the copy/cut operations source is the web client. */
   var initiatedByUser = false;
   
   /** The callback function that is called on the copy/cut/paste event.*/
   var onCopyCutPasteAction;
   /**
    * Clipboard paste handler.
    * 
    * @param {type} e event object
    * @returns {undefined}
    */
   function pasteHandler(e)
   {
      if (!me.enabled)
         return;
      
      var text = "";
 
      if (window.clipboardData) 
      {   
         text = window.clipboardData.getData("Text");
      }
      else if (e.clipboardData) 
      {
         text = e.clipboardData.getData("text/plain");
      }
      
      if (onCopyCutPasteAction)
      {
         setClipboardValue(text);
         onCopyCutPasteAction();
         onCopyCutPasteAction = null;
      }
      else
      {
         if (isEmpty(text))
         {
            // if empty string and CTRL-V was pressed
            if (ctrlv)
            {
               // send CTRL-V
               p2j.socket.sendPasteCtrlV();
            }
         }
         else
         {
            // CHUI: replace line ends in order to be handled properly on the thin client
            // as if "Enter" was pressed
            if (!p2j.isGui)
            {
               text = text.replace(/\n/g,"\r");
            }
            setClipboardValue(text);
            p2j.socket.sendPasteText(text); 
         }
      }
      ctrlv = false;
      setTimeout(focusHiddenInputField, 0);
   }
   
   /**
    * Clipboard copy handler.
    * 
    * @param {type} e event object
    * @returns {undefined}
    */
   function copyHandler(e)
   {
      console.debug("copyHandler {");
      if (!me.enabled)
         return;
      
      if (p2j.isGui)
      {
         // consume copy/cut events if the current selection is invalid
         if (getSelection() == null)
         {
            p2j.consumeEvent(e);
            return;
         }
         prepareCopy();
      }
      // restore value to copy
      var text = getClipboardValue();
      console.debug("getClipboardValue: " + text);
      // copy to clipboard
      if (window.clipboardData) 
      {
         window.clipboardData.clearData();
         window.clipboardData.setData("Text", text);
      }
      else if (e.clipboardData) 
      {
         e.clipboardData.clearData();
         e.clipboardData.setData("text/plain", text);
      }
      
      if (onCopyCutPasteAction)
      {
         onCopyCutPasteAction();
         onCopyCutPasteAction = null;
      }
      // clear selection
      me.clearSelection();
      setTimeout(focusHiddenInputField, 0);
      console.debug("copyHandler }");
   }
    
   /**
    * Init module.
    * 
    * @param {type} cfg configurartion object
    */
   me.init = function(cfg)
   {
      me.enabled = cfg.clipboard.enabled;
      
      // input element used on copy/paste operations.
      inputText = document.getElementById(cfg.clipboard.input);
      
      userSelection = " ";
      
      // some browsers need the text entry field to be in focus otherwise they will not copy or
      // paste
      focusHiddenInputField();
      
      p2j.clipboard_helpers.init(cfg);
      
      // the event handlers must have been defined above the init() function
      document.addEventListener("copy", copyHandler);
      document.addEventListener("paste", pasteHandler);
      if (p2j.isIe)
      {
         document.addEventListener("beforepaste", preparePaste);
      }
      
      inputText.addEventListener("input",focusHiddenInputField);
      
      document.addEventListener("mouseup", focusHiddenInputField);
      // ChUI doesn't use CUT for system clipboard integration
      if (p2j.isGui)
      {
         document.addEventListener("cut", copyHandler);
      }
   };
   
   /**
    * Prepare for copy operation.
    * 
    * @returns {undefined}
    */
   function prepareCopy()
   {
      console.debug("prepareCopy {");
      if (!me.needsSelection() || me.haveSelection())
      {
         // get selected text
         // TODO: this is broken for GUI CUT/COPY, this will send a request for the selected test
         //       to the server, but the text will be sent back asynchronously and we have no way
         //       to block this code and wait for the results, thus this text will always be
         //       wrong until we find a way around this issue
         var text = p2j.clipboard_helpers.getSelectedText();

         console.debug("getSelectedText() " + text);

         // copy the current selection to the input hidden field to save this value
         // to the system clipboard if the current selection is valid
         if (text != null)
         {
            setClipboardValue(text);
         }
         
         focusHiddenInputField();
      }
      console.debug("prepareCopy }");
   };
   
   me.prepareCopy = prepareCopy;
   
   /**
    * Prepare for paste operation.
    * 
    * @returns {undefined}
    */
   function preparePaste()
   {
      console.debug("preparePaste {");
      ctrlv = true;
      focusHiddenInputField();
      console.debug("preparePaste }");
   };
   
   me.preparePaste = preparePaste;
   
   /**
    * To focus and to select the hidden input field. It is required for the browser to generate
    * system copy, cut and paste events. These events are generated only if there is a focused
    * selected input field. 
    */
   function focusHiddenInputField()
   {
      inputText.focus();
      inputText.select();
   }
   
   /**
    * Check if this client needs an active selection area for copy to work.
    * 
    * @returns   {boolean} true if the client needs selection processing for copy. 
    */
   me.needsSelection = function()
   {
      return me.enabled && p2j.clipboard_helpers.needsSelection();
   };
   
   /**
    * Clear selection on screen.
    * 
    * @returns   {undefined}
    */
   me.clearSelection = p2j.clipboard_helpers.clearSelection;
   
   /**
    * Check if there is an active selection area.
    * 
    * @returns   {boolean} true if have a selection false if not.
    */
   me.haveSelection = p2j.clipboard_helpers.haveSelection;
   
   /**
    * Process the server's request for the clipboard contents (MSG_READ_CLIPBOARD) and respond
    * with a MSG_CLIPBOARD_CONTENTS.
    */
   me.sendClipboardContents = function()
   {
      console.debug("sendClipboardContents {");

      var text = getClipboardValue();
      
      console.debug("getClipboardValue: " + text);
      
      // send a MSG_CLIPBOARD_CONTENTS back
      p2j.socket.sendClipboardContent(text);
      console.debug("sendClipboardContents }");
   };

   /**
    * Copy the selected text to the clipboard. 
    * 
    * @param   {String} text
    *          The text to copy.
    */
   me.writeClipboard = function(text)
   {
      console.debug("writeClipboard {" + text);
      //in order to get the last selection snapshot we set the value of clipboard to the current selection.
      if (initiatedByUser)
      {
         setClipboardValue(text);
         initiatedByUser = false;
         onCopyCutPasteAction = null;
      }
      else
      {
         // write to the selection field
         setSelection(text);
         setClipboardValue(text);
         var maxZIndex = p2j.getMaxZIndex();
         p2j.createDialog(
               "writeClipboard",
               1000 + maxZIndex,
               document.body,
               p2j.prompts("copy"),
               p2j.hints("copy"),
               function (evt)
               {
                  if ((evt.metaKey && (evt.keyCode === keys.C) && p2j.isSafari) ||
                      (evt.ctrlKey && (evt.keyCode === keys.C || evt.keyCode === keys.INSERT) &&
                            !p2j.isSafari))
                  {
                     p2j.closeDialog();
                     p2j.clipboard.prepareCopy();
                  }
               });
      }
      console.debug("writeClipboard }");
   };
   
   /**
    * Sets the current text selection.
    * 
    * @param   {String} text
    *          The selected text, current selection.
    */
   function setSelection(text)
   {
      userSelection = text;
   }
   
   me.setSelection = setSelection;
   
   /**
    * Stores the current value to be saved in the system clipboard.
    * 
    * @param   {String} text
    *          The value to be stored in the system clipboard.
    */
   function setClipboardValue(text)
   {
      console.debug("setClipboardValue: " + text);
      inputText.value = text;
      focusHiddenInputField();
   }
   
   /**
    * Gets the current selection.
    * 
    * @return   The selected text, current selection.
    */
   function getSelection()
   {
      return userSelection;
   }
   
   me.getSelection = getSelection;
   
   /**
    * Gets the current stored value in the hidden input field. 
    * 
    * @return   The stored text.
    */
   function getClipboardValue()
   {
      return inputText.value;
   }

   /**
    * Check if a string is empty.
    * 
    * @param {str} input string 
    * @returns true if empty false otherwise
    */
   function isEmpty(str) 
   {
      return (!str || 0 === str.length);
   }   
   
   /**
    * Notifies the module that copy/cut action is initiated by the web client.
    * 
    * @param    {Function}
    *           The callback function that must be invoked on the copy/cut/paste event.
    */
   function notifyAboutUserInitiatedAction(onCopyCutPasteUserAction)
   {
      initiatedByUser = true;
      onCopyCutPasteAction = onCopyCutPasteUserAction;
   }
   
   me.notifyAboutUserInitiatedAction = notifyAboutUserInitiatedAction;
   
   return me;   
}());