/*
** Module   : p2j.js
** Abstract : top-level web client java script object 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140204 Prevent back navigation.  Disable browser context menu.
** 003 MAG 20140702 Add tty.clipboard module.
** 004 GES 20150324 Moved this class to a common resource base and made it generic (so that it
**                  can be used for both ChUI and GUI web clients.
** 005 SBI 20150731 Applied "use strict" directive, fixed undeclared variables.
** 006 GES 20150709 Added createCanvas() helper function. Moved clipboard_helpers init to the
**                  common clipboard logic.  Added p2j.fonts.
** 007 SBI 20150901 Modifications to initialization order.
** 008 SBI 20151020 Increased logging buffer size.
** 009 SBI 20151215 Fixed to take into an account that for Chrome an element style object has only
**                  the "getter" access function. Changed to detect Chrome, IE11 and IE12 (Edge).
**     SBI 20151217 Added displayWaitCursor/restoreCursor.
**     SBI 20150104 Changed to reuse DOJO style API.
**     SBI 20160323 Added new methods to save/restore JSON serialized objects in the session and
**                  local storages, added p2j.embedded property, setDesktopBgColor(rgb).
** 010 CA  20160418 Added infrastructure for message passing from converted code to a 
**                  customer-specific application or calls from the customer-specific application 
**                  into the converted code (i.e. invoke external programs, procedures or 
**                  functions).  Can be used only when the P2J client runs embedded in a 
**                  customer-specific application.
** 011 SBI 20160519 Added the Safari detection.
**     SBI 20160609 Added createDialog function to open the modal dialog, added prompts and hints,
**                  added createKeyEvent, fixed Safari detection.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Create a top-level JavaScript object (which acts as a namespace for the web client code)
 * and initialize all submodules.
 * <p>
 * The following order of initialization is required:
 * <p>
 * <ol>
 *    <li> p2j.sound
 *    <li> p2j.screen
 *    <li> p2j.socket
 *    <li> p2j.keyboard
 *    <li> p2j.clipboard
 *    <li> p2j.fonts
 * </ol>
 */
var p2j = (function()
{
   /** The top-level web client object which will be returned from this function. */
   var me = {};

   /** OS detection flags. */
   me.isWindows = navigator.appVersion.indexOf('Win') > -1;
   me.isMac = navigator.appVersion.indexOf('Mac') > -1;
   me.isUnix = navigator.appVersion.indexOf('X11') > -1;
   me.isLinux = navigator.appVersion.indexOf('Linux') > -1;

   /** Browser detection flags. */
   me.isFirefox = navigator.userAgent.toLowerCase().indexOf('firefox') > -1;
   me.isChrome  = navigator.userAgent.toLowerCase().indexOf('chrome') > -1;
   me.isSafari  = (navigator.userAgent.toLowerCase().indexOf('safari/') > -1) && !me.isChrome;
   me.isChromium = !!window.chrome;
   me.isIE11 = navigator.userAgent.indexOf("Trident") > -1
                                                      || navigator.userAgent.indexOf("Edge") > -1;
   me.isIe = !!window.ActiveXObject || me.isIE11;
   
   /** The default container element (if specified), or <code>document.body</code>. */
   var container = null;
   
   /** The instance of the unique copy/cut/paste notification dialog. */
   var dialog;
   
   /** The notification templates to prompt a user to do an action */
   var prompts = {};
   
   /** The hint templates to help a user to perform an action */
   var hints = {};
   
   /** The templates parameters */
   var parameters = {};
   
   /** The notification template for the copy action */
   prompts["copy"] = "The application is requesting a COPY action." +
                     " Press {0} to complete the COPY action.";
   
   /** The hint for the copy action */
   hints["copy"] = "Press {0} to complete the COPY action and to close this dialog.";
   
   /** copy template parameter */
   parameters["copy"] = me.isSafari ? "COMMAND-C" : "CTRL-C";
   
   /** The notification template for the paste action */
   prompts["paste"] = "The application is requesting a PASTE action." +
                      " Press {0} to complete the PASTE action.";
   
   /** The hint for the paste action */
   hints["paste"] = "Press {0} to complete the PASTE action and to close this dialog.";
   
   /** copy template parameter */
   parameters["paste"] = me.isSafari ? "COMMAND-V" : "CTRL-V";
   
   /** The notification template for the cut action */
   prompts["cut"] = "The application is requesting a CUT action." +
                    " Press {0} to complete the CUT action.";
   
   /** The hint for the cut action */
   hints["cut"] = "Press {0} to complete the CUT action and to close this dialog.";
   
   /** cut template parameter */
   parameters["cut"] = me.isSafari ? "COMMAND-X" : "CTRL-X";
   
   /** its properties define if copy, cut and paste commands are supported by this browser */
   var queryCommandSupported = {};
   
   /**
    * Initialize module.
    * 
    * @param {object} cfg  configuration.
    */
   me.init = function(cfg)
   {
      /**
       * Prevent back navigation.
       */
      history.pushState({}, '', cfg.page);

      /**
       * Prevent back navigation.
       * 
       * @param {event} event object
       */
      window.onpopstate = function(event) 
      {
         if (event.type === 'popstate')
            history.pushState({}, '', cfg.page);
      };

      /**
       * Disable browser context menu.
       * 
       * @returns {boolean}   false = disabled
       *                      true  = enabled
       */
      document.oncontextmenu = function()
      {
         return false;
      };
      
      container = document.getElementById(cfg.container) || document.body;
      
      // save off our client mode type
      p2j.isGui = cfg.isGui;
      // set the web client embedded mode and define a "get" access only 
      Object.defineProperty(
                            p2j,
                            "embedded",
                            {
                               get: function ()
                               {
                                  return cfg.embedded;
                               }
                            });
      /** the body html css class name to display wait cursor */
      p2j.waitCursorClass = "wait";
      
      /** the body html css class name to display default cursor */
      p2j.bodyHtmlClass = document.body.className;
      
      /** to display the wait cursor for the document */
      p2j.displayWaitCursor = function ()
      {
         if (document.body.className === p2j.bodyHtmlClass)
         {
            document.body.className += (" " + p2j.waitCursorClass);
         }
      }
      
      /** to restore the cursor for the document */
      p2j.restoreCursor = function ()
      {
         document.body.className = p2j.bodyHtmlClass;
      }
      
      /**
       * Set the desktop background color.
       * 
       * @param    {String} rgb
       *           The CSS color string.
       */
      p2j.setDesktopBgColor = function (rgb)
      {
         if (p2j.embedded)
         {
            setStyleForElement(document.body, {background : rgb});
         }
      }
      
      Object.defineProperty(
            p2j,
            "prompts",
            {
               get: function ()
               {
                  return function(cmd)
                  {
                     var prompt;
                     var template = prompts[cmd];
                     if (template)
                     {
                        prompt = template.replace("{0}", parameters[cmd]);
                     }
                     
                     return prompt;
                  }
               }
            });
      
      Object.defineProperty(
            p2j,
            "hints",
            {
               get: function ()
               {
                  return function(cmd)
                  {
                     var hint;
                     var template = hints[cmd];
                     if (template)
                     {
                        hint = template.replace("{0}", parameters[cmd]);
                     }
                     
                     return hint;
                  }
               }
            });
      
      ["copy", "paste", "cut"].forEach(
         function(cmd)
         {
            try
            {
               queryCommandSupported[cmd] = document.queryCommandSupported(cmd) === true;
            }
            catch (e)
            {
               queryCommandSupported[cmd] = false;
            }
         });
      Object.defineProperty(
         p2j,
         "commandSupported",
         {
            get: function ()
            {
               return function(cmd)
               {
                  return queryCommandSupported[cmd];
               }
            }
         });
      // init modules (don't change the order here!)
      p2j.logger.init(4096, true);
      p2j.sound.init(cfg);
      p2j.fonts.init();
      p2j.screen.init(cfg);
      p2j.socket.init(cfg);
      p2j.keyboard.init(cfg);
      p2j.clipboard.init(cfg);
      if (cfg.embedded)
      {
         p2j.remote.init(window.parent);
      }
   };
   
   /**
    * Copy all "own" properties of the given source object into the target object, using
    * recursion to copy any nested objects. This can recursively copy to an arbitrary depth.
    * <p>
    * This is a deep copy of the object graph but it is NOT a clone.  In particular, any
    * existing properties and functions of target that don't conflict with source, will remain.
    * In addition, the prototype of source will NOT be copied to target. So target will not
    * be considered to be the same type and will not inherit any changes from source.
    *
    * @param    {Object} target
    *           Object into which to deeply copy the source properties.
    * @param    {Object} source
    *           The object from which to copy. Any property that is not found via 
    *           hasOwnProperty() will be excluded. 
    */
   function setOwnPropertiesTo(target, source)
   {
      if (!source)
      {
         return;
      }
      
      for (var prop in source)
      {
         if (source.hasOwnProperty(prop))
         {
            if (typeof source[prop] == "object")
            {
               if (prop === "style")
               {
                  setStyleForElement(target, source[prop]);
               }
               else
               {
                  target[prop] = {};
                  setOwnPropertiesTo(target[prop], source[prop]);
               }
            }
            else
            {
               target[prop] = source[prop];
            }
         }
      }
   }
   
   /**
    * Sets the provided style properties for the given html element.
    * 
    * @param    {HtmlElement} element
    *           The DOM element which will be the parent for the new canvas. If not provided, the
    *           default container element will be used.
    * @param    {Object} styleObj
    *           The custom object that holds the target style properties.
    */
   function setStyleForElement(element, styleObj)
   {
      dojo.style(element, styleObj);
   }
   
   me.setStyleForElement = setStyleForElement;
   
   /**
    * Helper function to crate a new canvas as a child of the given element and with
    * caller-defined optional settings.
    *
    * @param    parentElement
    *           The DOM element which will be the parent for the new canvas. If not provided, the
    *           default container element will be used.
    * @param    settings
    *           Optional caller-defined attributes.
    *
    * @return   The created canvas element.
    */
   me.createCanvas = function(parentElement, settings)
   {
      parentElement = parentElement || container;
      
      var canvas = document.createElement('canvas');
      
      setOwnPropertiesTo(canvas, settings);
      
      parentElement.appendChild(canvas);
      
      return canvas;
   };

   /**
    * Save the target JS object serialized to a string value with the given key.
    * 
    * @param    {String} key
    *           The assigned key.
    * @param    {Object} obj
    *           The target JS object.
    * @param    {Boolean} local
    *           True value indicates that the object is persisted between opened and new sessions,
    *           otherwise the object is stored during the current session.
    */
   function saveObject(key, obj, local)
   {
      var sObj = JSON.stringify(obj);
      var storage;
      if (local)
      {
         storage = localStorage;
      }
      else
      {
         storage = sessionStorage;
      }
      storage.setItem(key, sObj);
      return sObj;
   }
   
   me.saveObject = saveObject;
   
   /**
    * Try to restore a JS object from the retrieved string value for the given key.
    * 
    * @param    {String} key
    *           The provided key.
    * @param    {Boolean} local
    *           True value indicates that the local storage is used, otherwise the session
    *           storage is used.
    * 
    * @return   {Object} obj
    *           The target JS object or null.
    */
   function restoreObject(key, local)
   {
      var storage;
      if (local)
      {
         storage = localStorage;
      }
      else
      {
         storage = sessionStorage;
      }
      var sObj = storage.getItem(key);
      if (sObj)
      {
         var obj = JSON.parse(sObj);
         return obj;
      }
      return null;
   }
   
   me.restoreObject = restoreObject;
   
   /**
    * Delete a JS object by the given key.
    * 
    * @param    {String} key
    *           The provided key.
    * @param    {Boolean} local
    *           True value indicates that the local storage is used, otherwise the session
    *           storage is used.
    */
   function deleteObject(key, local)
   {
      var storage;
      if (local)
      {
         storage = localStorage;
      }
      else
      {
         storage = sessionStorage;
      }
      storage.removeItem(key);
   }
   
   me.deleteObject = deleteObject;
   
   /**
    * Cancels the target event if it is cancelable and stops its further propagation.
    * 
    * @param    {Event} evt
    *           The target event to cancel its default action and to stop its propagation.
    */
   function consumeEvent(evt)
   {
      if (evt.preventDefault)
      {
         evt.preventDefault();
      }
      if (evt.stopPropagation)
      {
         evt.stopPropagation();
      }
      if (evt.stopImmediatePropagation)
      {
         evt.stopImmediatePropagation();
      }
   }
   
   me.consumeEvent = consumeEvent;
   
   /**
    * Create a key event.
    * 
    * @param    {String} type
    *           The event type.
    * @param    {Boolean} bubbles
    *           The bubbled event.
    * @param    {Boolean} cancelable
    *           The cancelable event.
    * @param    {Window} view
    *           The window.
    * @param    {Boolean} ctrlKey
    *           True iff  the Ctrl (left or right) key is on. 
    * @param    {Boolean} altKey
    *           True iff  the Alt (left or right) key is on. 
    * @param    {Boolean} shiftKey
    *           True iff  the Alt (left or right) key is on. 
    * @param    {Boolean} metaKey
    *           True iff  the Meta (left or right) key is on.
    * @param    {Number} keyCode
    *           The key code.
    * @param    {Number} charCode
    *           The char code.
    * 
    * @returns  {Event}
    *           The target keyboard event.
    */
   function createKeyEvent(type, bubbles, cancelable, view, ctrlKey,
         altKey, shiftKey, metaKey, keyCode, charCode)
   {
      var event;
      if (document.createEvent !== undefined)
      {
         try 
         {
            // try to create a key event (Firefox use case)
            event = document.createEvent("KeysEvent");
            
            event.initKeyEvent(type, bubbles, cancelable, view, ctrlKey,
                  altKey, shiftKey, metaKey, keyCode, charCode);
         }
         catch (ex1)
         {
            try
            {
               // try to create a common event (Chrome and Safari use cases)
                event = document.createEvent("Events");
            }
            catch (ex2)
            {
               // // try to create a UIEvent event (Safari use case)
                event = document.createEvent("UIEvents");
            }
            finally
            {
                try
                {
                   event.initEvent(type, bubbles, cancelable);
                   //initialize
                   event.view = view;
                   event.altKey = altKey;
                   event.ctrlKey = ctrlKey;
                   event.shiftKey = shiftKey;
                   event.metaKey = metaKey;
                   event.keyCode = keyCode;
                   event.charCode = charCode;
                }
                catch (ex3)
                {
                   console.error(ex3);
                }
            }
         }
      }
      else if (document.createEventObject !== undefined)
      {
         event = document.createEventObject();
         
         event.bubbles = bubbles;
         event.cancelable = cancelable;
         event.view = view;
         event.ctrlKey = ctrlKey;
         event.altKey = altKey;
         event.shiftKey = shiftKey;
         event.metaKey = metaKey;
         
         event.keyCode = (charCode > 0) ? charCode : keyCode;
         event.type = type;
      }
      
      return event;
   };
   
   me.createKeyEvent = createKeyEvent;
   
   /**
    * Creates and displays the modal dialog.
    * 
    * @param    {String} id
    *           The dialog unique identifier.
    * @param    {Number} zIndex
    *           The maximal z-index calculated for the current page.
    * @param    {HtmlElement} container
    *           The container element.
    * @param    {String} text
    *           The dialog content message.
    * @param    {String} tooltip
    *           The dialog tooltip message.
    * @param    {Function} keyDownHandler
    *           The "keydown" event's handler.
    * 
    * @return   The modal dialog.
    */
   function createDialog(id, zIndex, container, text, tooltip, keyDownHandler)
   {
      if (!dialog)
      {
         dialog = new ModalDialog(id, zIndex, container, text, tooltip);
         
         if (keyDownHandler)
         {
            dialog.addTrigger("keydown", keyDownHandler);
         }
      }
      
      return dialog;
   };

   me.createDialog = createDialog;
   
   /**
    * Closes and disposes the modal dialog.
    */
   function closeDialog()
   {
      if (dialog)
      {
         dialog.close();
         dialog = null;
      };
   };
   
   me.closeDialog = closeDialog;
   
   /**
    * Tests if the modal dialog is active.
    */
   me.isDialogDisplayed = function ()
   {
      return (dialog !== null && dialog !== undefined);
   }
   
   /**
    * Calculates the maximal z-index on the current page.
    * 
    * @return   The maximal z-index on the current page.
    */
   function getMaxZIndex()
   {
      var zIndex;
      var maxZIndex = 1;
      var all  = document.getElementsByTagName('*');
      for (var i = 0, n = all.length; i < n; i++)
      {
          zIndex = window.getComputedStyle(all[i], null).getPropertyValue("z-index");
          zIndex = parseInt(zIndex, 10);
          if (zIndex != null && isFinite(zIndex))
          {
             maxZIndex = Math.max(maxZIndex, zIndex);
          }
      }
      
      return maxZIndex;
   };
   
   me.getMaxZIndex = getMaxZIndex;
   
   /**
    * Creates and displays the modal dialog.
    * 
    * @param    {String} id
    *           The dialog unique identifier.
    * @param    {Number} zIndex
    *           The maximal z-index calculated for the current page.
    * @param    {HtmlElement} container
    *           The container element.
    * @param    {String} text
    *           The dialog content message.
    * @param    {String} tooltip
    *           The dialog tooltip message.
    */
   function ModalDialog (id, zIndex, container, text, tooltip)
   {
      var overlayId = getOverlayId(id);
      var overlay = createOverlayPanel(overlayId, zIndex, container);
      var dialog = document.createElement("div");
      dialog.id = id;
      var textEl = document.createTextNode(text);
      var div = document.createElement("div");
      div.appendChild(textEl);
      div.className = "content centered";
      dialog.appendChild(div);
      dialog.className    = "dialog centered";
      dialog.style.zIndex = zIndex + 1;
      dialog.title = tooltip;
      container.appendChild(dialog);
      // create triggers map
      dialog.triggers = new Map();
      // save the overlay panel instance
      dialog.overlay  = overlay;
      
      overlay.style.display = "block";
      dialog.style.display = "block";

      /**
       * Adds a trigger for target events.
       * 
       * @param    {String} eventType
       *           The event type that is listen by this dialog.
       * @param    {Function} handler
       *           The callback function that is invoked if a trigger event occurs.
       */
      function addTrigger(eventType, handler)
      {
         dialog.triggers.set(eventType, handler);
         document.addEventListener(eventType, handler);
      };
      
      this.addTrigger = addTrigger;
      
      /**
       * Removes a handler of the given event type.
       * 
       * @param    {String} trigger
       *           The event type listen by this dialog.
       */
      function removeTrigger(eventType)
      {
         var handler = dialog.triggers.get(eventType);
         if (handler)
         {
            document.removeEventListener(eventType, handler);
         }
      };
      
      this.removeTrigger = removeTrigger;
      
      /**
       * Closes this modal dialog.
       */
      function close()
      {
         dialog.triggers.forEach(
                                 function(handler, eventType, triggers)
                                 {
                                    removeTrigger(eventType);
                                 });
         dialog.style.display  = "none";
         dialog.parentNode.removeChild(dialog);
         var overlay = dialog.overlay;
         if (overlay)
         {
            overlay.style.display = "none";
            overlay.parentNode.removeChild(overlay);
         }
      };
      
      this.close = close;
      /**
       * Creates the overlay panel that must cover the displayed page.
       * 
       * @param    {String} id
       *           The element unique identifier.
       * @param    {Number} zIndex
       *           The maximal z-index calculated for the current page.
       * @param    {HtmlElement} container
       *           The container element.
       * 
       * @return   The overlay panel.
       */
      function createOverlayPanel(id, zIndex, container)
      {
         var overlayPanel = document.getElementById(id);
         
         if (!overlayPanel)
         {
            overlayPanel    = document.createElement("div");
            overlayPanel.id = id;
            overlayPanel.className = "overlay";
            container.appendChild(overlayPanel);
         }
         
         overlayPanel.style.display = "none";
         overlayPanel.style.zIndex  = zIndex;
         
         return overlayPanel;
      };
      
      /**
       * Gets the modal dialog overlay panel unique identifier.
       * 
       * @param    {String} dialogId
       *           The dialog unique identifier.
       * 
       * @return   The overlay element unique identifier.
       */
      function getOverlayId(dialogId)
      {
         return dialogId + "_overlay";
      }
   };
   
   /* exports */
   return me;
}());