/*
** Module   : p2j.keymap.js
** Abstract : key definitions and mapping
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140228 Add javadoc.
** 003 GES 20150324 Moved this class to a common resource base and made it generic (so that it
**                  can be used for both ChUI and GUI web clients.
** 004 SBI 20150731 Applied "use strict" directive, fixed undeclared variables.
** 005 GES 20150701 Added a clarifying comment.
** 006 SBI 20150922 Rewrite to create separate ChUI and GUI implementations.
** 007 SBI 20151006 Improvements to support better GUI key processing compatibility.
** 008 SBI 20151011 Added SHIFT + DEL key label to return 639 as its key code.
** 009 SBI 20160120 Chrome and IE support.
**     SBI 20160212 Fixes for copy/paste and cut default actions.
** 010 SBI 20160603 Changed to process Safari COMMAND as it would be a CTRL key pressed, added
**                  a virtual key, IME_PROCESSKEY, identified by the 229 code.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Tests is this array has a valid value at the given position.
 * 
 * @param    {Number}
 *           The index of the array element to test.
 * 
 * @return   {Boolean}
 *           True iff the selected element is defined and is not empty.
 */
Array.prototype.isNotEmpty = function(index)
{
   return this[index] !== undefined && this[index] !== null && this[index] !== '';
};

/**
 * Global definitions for common key values.
 */
var keys = {
   BACKSPACE : 8,
   TAB : 9,
   CLEAR : 12,
   ENTER : 13,
   SHIFT : 16,
   CTRL : 17,
   ALT : 18,
   PAUSE : 19,
   CAPS_LOCK : 20,
   ESCAPE : 27,
   SPACE : 32,
   PAGE_UP : 33,
   PAGE_DOWN : 34,
   END : 35,
   HOME : 36,
   LEFT_ARROW : 37,
   UP_ARROW : 38,
   RIGHT_ARROW : 39,
   DOWN_ARROW : 40,
   INSERT : 45,
   DELETE : 46,
   HELP : 47,
   LEFT_WINDOW : 91,
   RIGHT_WINDOW : 92,
   SELECT : 93,
   NUMPAD_0 : 96,
   NUMPAD_1 : 97,
   NUMPAD_2 : 98,
   NUMPAD_3 : 99,
   NUMPAD_4 : 100,
   NUMPAD_5 : 101,
   NUMPAD_6 : 102,
   NUMPAD_7 : 103,
   NUMPAD_8 : 104,
   NUMPAD_9 : 105,
   NUMPAD_MULTIPLY : 106,
   NUMPAD_PLUS : 107,
   NUMPAD_ENTER : 108,
   NUMPAD_MINUS : 109,
   NUMPAD_PERIOD : 110,
   NUMPAD_DIVIDE : 111,
   F1 : 112,
   F2 : 113,
   F3 : 114,
   F4 : 115,
   F5 : 116,
   F6 : 117,
   F7 : 118,
   F8 : 119,
   F9 : 120,
   F10 : 121,
   F11 : 122,
   F12 : 123,
   F13 : 124,
   F14 : 125,
   F15 : 126,
   NUM_LOCK : 144,
   SCROLL_LOCK : 145,
   UP_DPAD : 175,
   DOWN_DPAD : 176,
   LEFT_DPAD : 177,
   RIGHT_DPAD : 178,
   C : 67,
   V : 86,
   X : 88,
   IME_PROCESSKEY : 229
};

var basicKeys = [];

basicKeys[0] = "CTRL-@";
basicKeys[1] = "CTRL-A";
basicKeys[2] = "CTRL-B";
basicKeys[3] = "CTRL-C";
basicKeys[4] = "CTRL-D";
basicKeys[5] = "CTRL-E";
basicKeys[6] = "CTRL-F";
basicKeys[7] = "CTRL-G";
basicKeys[8] = "BACKSPACE";
basicKeys[9] = "TAB";
basicKeys[10] = "CTRL-J";
basicKeys[11] = "CTRL-K";
basicKeys[12] = "CTRL-L";
basicKeys[14] = "CTRL-N";
basicKeys[15] = "CTRL-O";
basicKeys[16] = "CTRL-P";
basicKeys[17] = "CTRL-Q";
basicKeys[18] = "CTRL-R";
basicKeys[19] = "CTRL-S";
basicKeys[20] = "CTRL-T";
basicKeys[21] = "CTRL-U";
basicKeys[22] = "CTRL-V";
basicKeys[23] = "CTRL-W";
basicKeys[24] = "CTRL-X";
basicKeys[25] = "CTRL-Y";
basicKeys[26] = "CTRL-Z";
basicKeys[27] = "ESC";
basicKeys[28] = "CTRL-\\";
basicKeys[29] = "CTRL-]";
basicKeys[30] = "CTRL-^";
basicKeys[31] = "CTRL-_";
basicKeys[127] = "DEL";
basicKeys[300] = "F0";
basicKeys[301] = "F1";
basicKeys[302] = "F2";
basicKeys[303] = "F3";
basicKeys[304] = "F4";
basicKeys[305] = "F5";
basicKeys[306] = "F6";
basicKeys[307] = "F7";
basicKeys[308] = "F8";
basicKeys[309] = "F9";
basicKeys[310] = "F10";
basicKeys[311] = "F11";
basicKeys[312] = "F12";
basicKeys[313] = "F13";
basicKeys[314] = "F14";
basicKeys[315] = "F15";
basicKeys[316] = "F16";
basicKeys[317] = "F17";
basicKeys[318] = "F18";
basicKeys[319] = "F19";
basicKeys[320] = "F20";
basicKeys[321] = "F21";
basicKeys[322] = "F22";
basicKeys[323] = "F23";
basicKeys[324] = "F24";
basicKeys[325] = "F25";
basicKeys[326] = "F26";
basicKeys[327] = "F27";
basicKeys[328] = "F28";
basicKeys[329] = "F29";
basicKeys[330] = "F30";
basicKeys[331] = "F31";
basicKeys[332] = "F32";
basicKeys[333] = "F33";
basicKeys[334] = "F34";
basicKeys[335] = "F35";
basicKeys[336] = "F36";
basicKeys[337] = "F37";
basicKeys[338] = "F38";
basicKeys[339] = "F39";
basicKeys[340] = "F40";
basicKeys[341] = "F41";
basicKeys[342] = "F42";
basicKeys[343] = "F43";
basicKeys[344] = "F44";
basicKeys[345] = "F45";
basicKeys[346] = "F46";
basicKeys[347] = "F47";
basicKeys[348] = "F48";
basicKeys[349] = "F49";
basicKeys[350] = "F50";
basicKeys[351] = "F51";
basicKeys[352] = "F52";
basicKeys[353] = "F53";
basicKeys[354] = "F54";
basicKeys[355] = "F55";
basicKeys[356] = "F56";
basicKeys[357] = "F57";
basicKeys[358] = "F58";
basicKeys[359] = "F59";
basicKeys[360] = "F60";
basicKeys[361] = "F61";
basicKeys[362] = "F62";
basicKeys[363] = "F63";
basicKeys[364] = "F64";
basicKeys[365] = "F65";
basicKeys[366] = "F66";
basicKeys[367] = "F67";
basicKeys[368] = "F68";
basicKeys[369] = "F69";
basicKeys[370] = "F70";
basicKeys[371] = "F71";
basicKeys[372] = "F72";
basicKeys[373] = "F73";
basicKeys[374] = "F74";
basicKeys[375] = "F75";
basicKeys[376] = "F76";
basicKeys[377] = "F77";
basicKeys[378] = "F78";
basicKeys[379] = "F79";
basicKeys[380] = "F80";
basicKeys[381] = "F81";
basicKeys[382] = "F82";
basicKeys[383] = "F83";
basicKeys[384] = "F84";
basicKeys[385] = "F85";
basicKeys[386] = "F86";
basicKeys[387] = "F87";
basicKeys[388] = "F88";
basicKeys[389] = "F89";
basicKeys[390] = "F90";
basicKeys[391] = "F91";
basicKeys[392] = "F92";
basicKeys[393] = "F93";
basicKeys[394] = "F94";
basicKeys[395] = "F95";
basicKeys[396] = "F96";
basicKeys[397] = "F97";
basicKeys[398] = "F98";
basicKeys[399] = "F99";
basicKeys[400] = "PF0";
basicKeys[401] = "PF1";
basicKeys[402] = "PF2";
basicKeys[403] = "PF3";
basicKeys[404] = "PF4";
basicKeys[405] = "PF5";
basicKeys[406] = "PF6";
basicKeys[407] = "PF7";
basicKeys[408] = "PF8";
basicKeys[409] = "PF9";
basicKeys[410] = "PF10";
basicKeys[411] = "PF11";
basicKeys[412] = "PF12";
basicKeys[413] = "PF13";
basicKeys[414] = "PF14";
basicKeys[415] = "PF15";
basicKeys[416] = "PF16";
basicKeys[417] = "PF17";
basicKeys[418] = "PF18";
basicKeys[419] = "PF19";
basicKeys[420] = "PF20";
basicKeys[421] = "PF21";
basicKeys[422] = "PF22";
basicKeys[423] = "PF23";
basicKeys[424] = "PF24";
basicKeys[425] = "PF25";
basicKeys[426] = "PF26";
basicKeys[427] = "PF27";
basicKeys[428] = "PF28";
basicKeys[429] = "PF29";
basicKeys[430] = "PF30";
basicKeys[431] = "PF31";
basicKeys[432] = "PF32";
basicKeys[433] = "PF33";
basicKeys[434] = "PF34";
basicKeys[435] = "PF35";
basicKeys[436] = "PF36";
basicKeys[437] = "PF37";
basicKeys[438] = "PF38";
basicKeys[439] = "PF39";
basicKeys[440] = "PF40";
basicKeys[441] = "PF41";
basicKeys[442] = "PF42";
basicKeys[443] = "PF43";
basicKeys[444] = "PF44";
basicKeys[445] = "PF45";
basicKeys[446] = "PF46";
basicKeys[447] = "PF47";
basicKeys[448] = "PF48";
basicKeys[449] = "PF49";
basicKeys[450] = "PF50";
basicKeys[451] = "PF51";
basicKeys[452] = "PF52";
basicKeys[453] = "PF53";
basicKeys[454] = "PF54";
basicKeys[455] = "PF55";
basicKeys[456] = "PF56";
basicKeys[457] = "PF57";
basicKeys[458] = "PF58";
basicKeys[459] = "PF59";
basicKeys[460] = "PF60";
basicKeys[461] = "PF61";
basicKeys[462] = "PF62";
basicKeys[463] = "PF63";
basicKeys[464] = "PF64";
basicKeys[465] = "PF65";
basicKeys[466] = "PF66";
basicKeys[467] = "PF67";
basicKeys[468] = "PF68";
basicKeys[469] = "PF69";
basicKeys[470] = "PF70";
basicKeys[471] = "PF71";
basicKeys[472] = "PF72";
basicKeys[473] = "PF73";
basicKeys[474] = "PF74";
basicKeys[475] = "PF75";
basicKeys[476] = "PF76";
basicKeys[477] = "PF77";
basicKeys[478] = "PF78";
basicKeys[479] = "PF79";
basicKeys[480] = "PF80";
basicKeys[481] = "PF81";
basicKeys[482] = "PF82";
basicKeys[483] = "PF83";
basicKeys[484] = "PF84";
basicKeys[485] = "PF85";
basicKeys[486] = "PF86";
basicKeys[487] = "PF87";
basicKeys[488] = "PF88";
basicKeys[489] = "PF89";
basicKeys[490] = "PF90";
basicKeys[491] = "PF91";
basicKeys[492] = "PF92";
basicKeys[493] = "PF93";
basicKeys[494] = "PF94";
basicKeys[495] = "PF95";
basicKeys[496] = "PF96";
basicKeys[497] = "PF97";
basicKeys[498] = "PF98";
basicKeys[499] = "PF99";
basicKeys[501] = "CURSOR-UP";
basicKeys[502] = "CURSOR-DOWN";
basicKeys[503] = "CURSOR-RIGHT";
basicKeys[504] = "CURSOR-LEFT";
basicKeys[505] = "HOME";
basicKeys[506] = "END";
basicKeys[507] = "PAGE-UP";
basicKeys[508] = "PAGE-DOWN";
basicKeys[510] = "INS";
basicKeys[511] = "HELP-KEY";
basicKeys[513] = "EXECUTE";
basicKeys[514] = "PAGE";
basicKeys[515] = "FIND";
basicKeys[516] = "INS-LINE";
basicKeys[517] = "DEL-LINE";
basicKeys[518] = "LINE-ERASE";
basicKeys[519] = "PAGE-ERASE";
basicKeys[520] = "BACKSPACE";
basicKeys[521] = "TAB";
basicKeys[522] = "EXIT";
basicKeys[535] = "ERASE";
basicKeys[536] = "WHITE";
basicKeys[537] = "BLUE";
basicKeys[538] = "RED";
basicKeys[539] = "ESC";
basicKeys[541] = "CTRL-BREAK";
basicKeys[551] = "U1";
basicKeys[552] = "U2";
basicKeys[553] = "U3";
basicKeys[554] = "U4";
basicKeys[555] = "U5";
basicKeys[556] = "U6";
basicKeys[557] = "U7";
basicKeys[558] = "U8";
basicKeys[559] = "U9";
basicKeys[560] = "U10";
basicKeys[586] = "DO";
basicKeys[587] = "SELECT";
basicKeys[588] = "REMOVE";
basicKeys[589] = "INSERT-HERE";
basicKeys[609] = "MOUSE-SELECT-UP";
basicKeys[610] = "MOUSE-MOVE-UP";
basicKeys[611] = "MOUSE-MENU-UP";
basicKeys[612] = "MOUSE-EXTEND-UP";
basicKeys[617] = "MOUSE-SELECT-DOWN";
basicKeys[619] = "MOUSE-MENU-DOWN";
basicKeys[620] = "MOUSE-EXTEND-DOWN";
basicKeys[625] = "MOUSE-SELECT-CLICK";
basicKeys[627] = "MOUSE-MENU-CLICK";
basicKeys[628] = "MOUSE-EXTEND-CLICK";
basicKeys[633] = "MOUSE-SELECT-DRAG";
basicKeys[634] = "MOUSE-MOVE-DRAG";
basicKeys[635] = "MOUSE-MENU-DRAG";
basicKeys[636] = "MOUSE-EXTEND-DRAG";
basicKeys[640] = "MOUSE-MOVE";
basicKeys[649] = "MOUSE-SELECT-DBLCLICK";
basicKeys[651] = "MOUSE-MENU-DBLCLICK";
basicKeys[652] = "MOUSE-EXTEND-DBLCLICK";

// completing the table of basic key names with the characters
for (var i = 32; i <= 255; i ++)
{
   if (i !== 127)
   {
      basicKeys[i] = String.fromCharCode(i);
   }
   
};

/**
 * Key codes adjusted to fit P2J needs.
 */
var Progress4GLKeys = {
   VK_DELETE         : 127,
   VK_F1             : 301,
   VK_F2             : 302,
   VK_F3             : 303,
   VK_F4             : 304,
   VK_F5             : 305,
   VK_F6             : 306,
   VK_F7             : 307,
   VK_F8             : 308,
   VK_F9             : 309,
   VK_F10            : 310,
   VK_F11            : 311,
   VK_F12            : 312,
   VK_F13            : 313,
   VK_F14            : 314,
   VK_F15            : 315,
   VK_F16            : 316,
   VK_F17            : 317,
   VK_F18            : 318,
   VK_F19            : 319,
   VK_F20            : 320,
   VK_F21            : 321,
   VK_F22            : 322,
   VK_F23            : 323,
   VK_F24            : 324,
   VK_CURSOR_UP      : 501,
   VK_CURSOR_DOWN    : 502,
   VK_CURSOR_RIGHT   : 503,
   VK_CURSOR_LEFT    : 504,
   VK_HOME           : 505,
   VK_END            : 506,
   VK_PAGE_UP        : 507,
   VK_PAGE_DOWN      : 508,
   VK_BACK_TAB       : 509,
   VK_INSERT         : 510,
   VK_CTRL_BREAK     : 541,
   VK_CTRL_AT        : 0,
   VK_CTRL_A         : 1,
   VK_CTRL_B         : 2,
   VK_CTRL_C         : 3,
   VK_CTRL_D         : 4,
   VK_CTRL_E         : 5,
   VK_CTRL_F         : 6,
   VK_CTRL_G         : 7,
   VK_BELL           : 7,
   VK_BACKSPACE      : 8,
   VK_TAB            : 9,
   VK_CTRL_J         : 10,
   VK_CTRL_K         : 11,
   VK_CTRL_L         : 12,
   VK_ENTER          : 13,
   VK_CTRL_N         : 14,
   VK_CTRL_O         : 15,
   VK_CTRL_P         : 16,
   VK_CTRL_Q         : 17,
   VK_CTRL_R         : 18,
   VK_CTRL_S         : 19,
   VK_CTRL_T         : 20,
   VK_CTRL_U         : 21,
   VK_CTRL_V         : 22,
   VK_CTRL_W         : 23,
   VK_CTRL_X         : 24,
   VK_CTRL_Y         : 25,
   VK_CTRL_Z         : 26,
   VK_ESCAPE         : 27,
   VK_CTRL_BCKSL     : 28,
   VK_CTRL_RBR       : 29,
   VK_CTRL_CARET     : 30,
   VK_CTRL_SUB       : 31,
   VK_SPACE          : 32,

   /*
    * Latin 1
    * (ISO/IEC 8859-1 = Unicode U+0020..U+00FF)
    * Byte 3 = 0
    */

   VK_space :                       0x0020, /* U+0020 SPACE */
   VK_exclam :                      0x0021, /* U+0021 EXCLAMATION MARK */
   VK_quotedbl :                    0x0022, /* U+0022 QUOTATION MARK */
   VK_numbersign :                  0x0023, /* U+0023 NUMBER SIGN */
   VK_dollar :                      0x0024, /* U+0024 DOLLAR SIGN */
   VK_percent :                     0x0025, /* U+0025 PERCENT SIGN */
   VK_ampersand :                   0x0026, /* U+0026 AMPERSAND */
   VK_apostrophe :                  0x0027, /* U+0027 APOSTROPHE */
   VK_quoteright :                  0x0027, /* deprecated */
   VK_parenleft :                   0x0028, /* U+0028 LEFT PARENTHESIS */
   VK_parenright :                  0x0029, /* U+0029 RIGHT PARENTHESIS */
   VK_asterisk :                    0x002a, /* U+002A ASTERISK */
   VK_plus :                        0x002b, /* U+002B PLUS SIGN */
   VK_comma :                       0x002c, /* U+002C COMMA */
   VK_minus :                       0x002d, /* U+002D HYPHEN-MINUS */
   VK_period :                      0x002e, /* U+002E FULL STOP */
   VK_slash :                       0x002f, /* U+002F SOLIDUS */
   VK_0 :                           0x0030, /* U+0030 DIGIT ZERO */
   VK_1 :                           0x0031, /* U+0031 DIGIT ONE */
   VK_2 :                           0x0032, /* U+0032 DIGIT TWO */
   VK_3 :                           0x0033, /* U+0033 DIGIT THREE */
   VK_4 :                           0x0034, /* U+0034 DIGIT FOUR */
   VK_5 :                           0x0035, /* U+0035 DIGIT FIVE */
   VK_6 :                           0x0036, /* U+0036 DIGIT SIX */
   VK_7 :                           0x0037, /* U+0037 DIGIT SEVEN */
   VK_8 :                           0x0038, /* U+0038 DIGIT EIGHT */
   VK_9 :                           0x0039, /* U+0039 DIGIT NINE */
   VK_colon :                       0x003a, /* U+003A COLON */
   VK_semicolon :                   0x003b, /* U+003B SEMICOLON */
   VK_less :                        0x003c, /* U+003C LESS-THAN SIGN */
   VK_equal :                       0x003d, /* U+003D EQUALS SIGN */
   VK_greater :                     0x003e, /* U+003E GREATER-THAN SIGN */
   VK_question :                    0x003f, /* U+003F QUESTION MARK */
   VK_at :                          0x0040, /* U+0040 COMMERCIAL AT */
   VK_A :                           0x0041, /* U+0041 LATIN CAPITAL LETTER A */
   VK_B :                           0x0042, /* U+0042 LATIN CAPITAL LETTER B */
   VK_C :                           0x0043, /* U+0043 LATIN CAPITAL LETTER C */
   VK_D :                           0x0044, /* U+0044 LATIN CAPITAL LETTER D */
   VK_E :                           0x0045, /* U+0045 LATIN CAPITAL LETTER E */
   VK_F :                           0x0046, /* U+0046 LATIN CAPITAL LETTER F */
   VK_G :                           0x0047, /* U+0047 LATIN CAPITAL LETTER G */
   VK_H :                           0x0048, /* U+0048 LATIN CAPITAL LETTER H */
   VK_I :                           0x0049, /* U+0049 LATIN CAPITAL LETTER I */
   VK_J :                           0x004a, /* U+004A LATIN CAPITAL LETTER J */
   VK_K :                           0x004b, /* U+004B LATIN CAPITAL LETTER K */
   VK_L :                           0x004c, /* U+004C LATIN CAPITAL LETTER L */
   VK_M :                           0x004d, /* U+004D LATIN CAPITAL LETTER M */
   VK_N :                           0x004e, /* U+004E LATIN CAPITAL LETTER N */
   VK_O :                           0x004f, /* U+004F LATIN CAPITAL LETTER O */
   VK_P :                           0x0050, /* U+0050 LATIN CAPITAL LETTER P */
   VK_Q :                           0x0051, /* U+0051 LATIN CAPITAL LETTER Q */
   VK_R :                           0x0052, /* U+0052 LATIN CAPITAL LETTER R */
   VK_S :                           0x0053, /* U+0053 LATIN CAPITAL LETTER S */
   VK_T :                           0x0054, /* U+0054 LATIN CAPITAL LETTER T */
   VK_U :                           0x0055, /* U+0055 LATIN CAPITAL LETTER U */
   VK_V :                           0x0056, /* U+0056 LATIN CAPITAL LETTER V */
   VK_W :                           0x0057, /* U+0057 LATIN CAPITAL LETTER W */
   VK_X :                           0x0058, /* U+0058 LATIN CAPITAL LETTER X */
   VK_Y :                           0x0059, /* U+0059 LATIN CAPITAL LETTER Y */
   VK_Z :                           0x005a, /* U+005A LATIN CAPITAL LETTER Z */
   VK_bracketleft :                 0x005b, /* U+005B LEFT SQUARE BRACKET */
   VK_backslash :                   0x005c, /* U+005C REVERSE SOLIDUS */
   VK_bracketright :                0x005d, /* U+005D RIGHT SQUARE BRACKET */
   VK_asciicircum :                 0x005e, /* U+005E CIRCUMFLEX ACCENT */
   VK_underscore :                  0x005f, /* U+005F LOW LINE */
   VK_grave :                       0x0060, /* U+0060 GRAVE ACCENT */
   VK_quoteleft :                   0x0060, /* deprecated */
   VK_a :                           0x0061, /* U+0061 LATIN SMALL LETTER A */
   VK_b :                           0x0062, /* U+0062 LATIN SMALL LETTER B */
   VK_c :                           0x0063, /* U+0063 LATIN SMALL LETTER C */
   VK_d :                           0x0064, /* U+0064 LATIN SMALL LETTER D */
   VK_e :                           0x0065, /* U+0065 LATIN SMALL LETTER E */
   VK_f :                           0x0066, /* U+0066 LATIN SMALL LETTER F */
   VK_g :                           0x0067, /* U+0067 LATIN SMALL LETTER G */
   VK_h :                           0x0068, /* U+0068 LATIN SMALL LETTER H */
   VK_i :                           0x0069, /* U+0069 LATIN SMALL LETTER I */
   VK_j :                           0x006a, /* U+006A LATIN SMALL LETTER J */
   VK_k :                           0x006b, /* U+006B LATIN SMALL LETTER K */
   VK_l :                           0x006c, /* U+006C LATIN SMALL LETTER L */
   VK_m :                           0x006d, /* U+006D LATIN SMALL LETTER M */
   VK_n :                           0x006e, /* U+006E LATIN SMALL LETTER N */
   VK_o :                           0x006f, /* U+006F LATIN SMALL LETTER O */
   VK_p :                           0x0070, /* U+0070 LATIN SMALL LETTER P */
   VK_q :                           0x0071, /* U+0071 LATIN SMALL LETTER Q */
   VK_r :                           0x0072, /* U+0072 LATIN SMALL LETTER R */
   VK_s :                           0x0073, /* U+0073 LATIN SMALL LETTER S */
   VK_t :                           0x0074, /* U+0074 LATIN SMALL LETTER T */
   VK_u :                           0x0075, /* U+0075 LATIN SMALL LETTER U */
   VK_v :                           0x0076, /* U+0076 LATIN SMALL LETTER V */
   VK_w :                           0x0077, /* U+0077 LATIN SMALL LETTER W */
   VK_x :                           0x0078, /* U+0078 LATIN SMALL LETTER X */
   VK_y :                           0x0079, /* U+0079 LATIN SMALL LETTER Y */
   VK_z :                           0x007a, /* U+007A LATIN SMALL LETTER Z */
   VK_braceleft :                   0x007b, /* U+007B LEFT CURLY BRACKET */
   VK_bar :                         0x007c, /* U+007C VERTICAL LINE */
   VK_braceright :                  0x007d, /* U+007D RIGHT CURLY BRACKET */
   VK_asciitilde :                  0x007e, /* U+007E TILDE */

   VK_nobreakspace :                0x00a0, /* U+00A0 NO-BREAK SPACE */
   VK_exclamdown :                  0x00a1, /* U+00A1 INVERTED EXCLAMATION MARK */
   VK_cent :                        0x00a2, /* U+00A2 CENT SIGN */
   VK_sterling :                    0x00a3, /* U+00A3 POUND SIGN */
   VK_currency :                    0x00a4, /* U+00A4 CURRENCY SIGN */
   VK_yen :                         0x00a5, /* U+00A5 YEN SIGN */
   VK_brokenbar :                   0x00a6, /* U+00A6 BROKEN BAR */
   VK_section :                     0x00a7, /* U+00A7 SECTION SIGN */
   VK_diaeresis :                   0x00a8, /* U+00A8 DIAERESIS */
   VK_copyright :                   0x00a9, /* U+00A9 COPYRIGHT SIGN */
   VK_ordfeminine :                 0x00aa, /* U+00AA FEMININE ORDINAL INDICATOR */
   VK_guillemotleft :               0x00ab, /* U+00AB LEFT-POINTING DOUBLE ANGLE QUOTATION MARK */
   VK_notsign :                     0x00ac, /* U+00AC NOT SIGN */
   VK_hyphen :                      0x00ad, /* U+00AD SOFT HYPHEN */
   VK_registered :                  0x00ae, /* U+00AE REGISTERED SIGN */
   VK_macron :                      0x00af, /* U+00AF MACRON */
   VK_degree :                      0x00b0, /* U+00B0 DEGREE SIGN */
   VK_plusminus :                   0x00b1, /* U+00B1 PLUS-MINUS SIGN */
   VK_twosuperior :                 0x00b2, /* U+00B2 SUPERSCRIPT TWO */
   VK_threesuperior :               0x00b3, /* U+00B3 SUPERSCRIPT THREE */
   VK_acute :                       0x00b4, /* U+00B4 ACUTE ACCENT */
   VK_mu :                          0x00b5, /* U+00B5 MICRO SIGN */
   VK_paragraph :                   0x00b6, /* U+00B6 PILCROW SIGN */
   VK_periodcentered :              0x00b7, /* U+00B7 MIDDLE DOT */
   VK_cedilla :                     0x00b8, /* U+00B8 CEDILLA */
   VK_onesuperior :                 0x00b9, /* U+00B9 SUPERSCRIPT ONE */
   VK_masculine :                   0x00ba, /* U+00BA MASCULINE ORDINAL INDICATOR */
   VK_guillemotright :              0x00bb, /* U+00BB RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK */
   VK_onequarter :                  0x00bc, /* U+00BC VULGAR FRACTION ONE QUARTER */
   VK_onehalf :                     0x00bd, /* U+00BD VULGAR FRACTION ONE HALF */
   VK_threequarters :               0x00be, /* U+00BE VULGAR FRACTION THREE QUARTERS */
   VK_questiondown :                0x00bf, /* U+00BF INVERTED QUESTION MARK */
   VK_Agrave :                      0x00c0, /* U+00C0 LATIN CAPITAL LETTER A WITH GRAVE */
   VK_Aacute :                      0x00c1, /* U+00C1 LATIN CAPITAL LETTER A WITH ACUTE */
   VK_Acircumflex :                 0x00c2, /* U+00C2 LATIN CAPITAL LETTER A WITH CIRCUMFLEX */
   VK_Atilde :                      0x00c3, /* U+00C3 LATIN CAPITAL LETTER A WITH TILDE */
   VK_Adiaeresis :                  0x00c4, /* U+00C4 LATIN CAPITAL LETTER A WITH DIAERESIS */
   VK_Aring :                       0x00c5, /* U+00C5 LATIN CAPITAL LETTER A WITH RING ABOVE */
   VK_AE :                          0x00c6, /* U+00C6 LATIN CAPITAL LETTER AE */
   VK_Ccedilla :                    0x00c7, /* U+00C7 LATIN CAPITAL LETTER C WITH CEDILLA */
   VK_Egrave :                      0x00c8, /* U+00C8 LATIN CAPITAL LETTER E WITH GRAVE */
   VK_Eacute :                      0x00c9, /* U+00C9 LATIN CAPITAL LETTER E WITH ACUTE */
   VK_Ecircumflex :                 0x00ca, /* U+00CA LATIN CAPITAL LETTER E WITH CIRCUMFLEX */
   VK_Ediaeresis :                  0x00cb, /* U+00CB LATIN CAPITAL LETTER E WITH DIAERESIS */
   VK_Igrave :                      0x00cc, /* U+00CC LATIN CAPITAL LETTER I WITH GRAVE */
   VK_Iacute :                      0x00cd, /* U+00CD LATIN CAPITAL LETTER I WITH ACUTE */
   VK_Icircumflex :                 0x00ce, /* U+00CE LATIN CAPITAL LETTER I WITH CIRCUMFLEX */
   VK_Idiaeresis :                  0x00cf, /* U+00CF LATIN CAPITAL LETTER I WITH DIAERESIS */
   VK_ETH :                         0x00d0, /* U+00D0 LATIN CAPITAL LETTER ETH */
   VK_Eth :                         0x00d0, /* deprecated */
   VK_Ntilde :                      0x00d1, /* U+00D1 LATIN CAPITAL LETTER N WITH TILDE */
   VK_Ograve :                      0x00d2, /* U+00D2 LATIN CAPITAL LETTER O WITH GRAVE */
   VK_Oacute :                      0x00d3, /* U+00D3 LATIN CAPITAL LETTER O WITH ACUTE */
   VK_Ocircumflex :                 0x00d4, /* U+00D4 LATIN CAPITAL LETTER O WITH CIRCUMFLEX */
   VK_Otilde :                      0x00d5, /* U+00D5 LATIN CAPITAL LETTER O WITH TILDE */
   VK_Odiaeresis :                  0x00d6, /* U+00D6 LATIN CAPITAL LETTER O WITH DIAERESIS */
   VK_multiply :                    0x00d7, /* U+00D7 MULTIPLICATION SIGN */
   VK_Oslash :                      0x00d8, /* U+00D8 LATIN CAPITAL LETTER O WITH STROKE */
   VK_Ooblique :                    0x00d8, /* U+00D8 LATIN CAPITAL LETTER O WITH STROKE */
   VK_Ugrave :                      0x00d9, /* U+00D9 LATIN CAPITAL LETTER U WITH GRAVE */
   VK_Uacute :                      0x00da, /* U+00DA LATIN CAPITAL LETTER U WITH ACUTE */
   VK_Ucircumflex :                 0x00db, /* U+00DB LATIN CAPITAL LETTER U WITH CIRCUMFLEX */
   VK_Udiaeresis :                  0x00dc, /* U+00DC LATIN CAPITAL LETTER U WITH DIAERESIS */
   VK_Yacute :                      0x00dd, /* U+00DD LATIN CAPITAL LETTER Y WITH ACUTE */
   VK_THORN :                       0x00de, /* U+00DE LATIN CAPITAL LETTER THORN */
   VK_Thorn :                       0x00de, /* deprecated */
   VK_ssharp :                      0x00df, /* U+00DF LATIN SMALL LETTER SHARP S */
   VK_agrave :                      0x00e0, /* U+00E0 LATIN SMALL LETTER A WITH GRAVE */
   VK_aacute :                      0x00e1, /* U+00E1 LATIN SMALL LETTER A WITH ACUTE */
   VK_acircumflex :                 0x00e2, /* U+00E2 LATIN SMALL LETTER A WITH CIRCUMFLEX */
   VK_atilde :                      0x00e3, /* U+00E3 LATIN SMALL LETTER A WITH TILDE */
   VK_adiaeresis :                  0x00e4, /* U+00E4 LATIN SMALL LETTER A WITH DIAERESIS */
   VK_aring :                       0x00e5, /* U+00E5 LATIN SMALL LETTER A WITH RING ABOVE */
   VK_ae :                          0x00e6, /* U+00E6 LATIN SMALL LETTER AE */
   VK_ccedilla :                    0x00e7, /* U+00E7 LATIN SMALL LETTER C WITH CEDILLA */
   VK_egrave :                      0x00e8, /* U+00E8 LATIN SMALL LETTER E WITH GRAVE */
   VK_eacute :                      0x00e9, /* U+00E9 LATIN SMALL LETTER E WITH ACUTE */
   VK_ecircumflex :                 0x00ea, /* U+00EA LATIN SMALL LETTER E WITH CIRCUMFLEX */
   VK_ediaeresis :                  0x00eb, /* U+00EB LATIN SMALL LETTER E WITH DIAERESIS */
   VK_igrave :                      0x00ec, /* U+00EC LATIN SMALL LETTER I WITH GRAVE */
   VK_iacute :                      0x00ed, /* U+00ED LATIN SMALL LETTER I WITH ACUTE */
   VK_icircumflex :                 0x00ee, /* U+00EE LATIN SMALL LETTER I WITH CIRCUMFLEX */
   VK_idiaeresis :                  0x00ef, /* U+00EF LATIN SMALL LETTER I WITH DIAERESIS */
   VK_eth :                         0x00f0, /* U+00F0 LATIN SMALL LETTER ETH */
   VK_ntilde :                      0x00f1, /* U+00F1 LATIN SMALL LETTER N WITH TILDE */
   VK_ograve :                      0x00f2, /* U+00F2 LATIN SMALL LETTER O WITH GRAVE */
   VK_oacute :                      0x00f3, /* U+00F3 LATIN SMALL LETTER O WITH ACUTE */
   VK_ocircumflex :                 0x00f4, /* U+00F4 LATIN SMALL LETTER O WITH CIRCUMFLEX */
   VK_otilde :                      0x00f5, /* U+00F5 LATIN SMALL LETTER O WITH TILDE */
   VK_odiaeresis :                  0x00f6, /* U+00F6 LATIN SMALL LETTER O WITH DIAERESIS */
   VK_division :                    0x00f7, /* U+00F7 DIVISION SIGN */
   VK_oslash :                      0x00f8, /* U+00F8 LATIN SMALL LETTER O WITH STROKE */
   VK_ooblique :                    0x00f8, /* U+00F8 LATIN SMALL LETTER O WITH STROKE */
   VK_ugrave :                      0x00f9, /* U+00F9 LATIN SMALL LETTER U WITH GRAVE */
   VK_uacute :                      0x00fa, /* U+00FA LATIN SMALL LETTER U WITH ACUTE */
   VK_ucircumflex :                 0x00fb, /* U+00FB LATIN SMALL LETTER U WITH CIRCUMFLEX */
   VK_udiaeresis :                  0x00fc, /* U+00FC LATIN SMALL LETTER U WITH DIAERESIS */
   VK_yacute :                      0x00fd, /* U+00FD LATIN SMALL LETTER Y WITH ACUTE */
   VK_thorn :                       0x00fe, /* U+00FE LATIN SMALL LETTER THORN */
   VK_ydiaeresis :                  0x00ff, /* U+00FF LATIN SMALL LETTER Y WITH DIAERESIS */
   // GUI navigation in EDITOR
   VK_CTRL_CURSOR_RIGHT  : 2551,
   VK_CTRL_CURSOR_LEFT   : 2552,
   VK_CTRL_HOME          : 2553,
   VK_CTRL_END           : 2554,

   // GUI text selection keys 
   VK_SHIFT_CURSOR_UP    : 1013,
   VK_SHIFT_CURSOR_DOWN  : 1014,
   VK_SHIFT_CURSOR_RIGHT : 1015,
   VK_SHIFT_CURSOR_LEFT  : 1016,
   VK_SHIFT_HOME         : 1017,
   VK_SHIFT_END          : 1018,
   VK_SHIFT_PAGE_UP      : 1019,
   VK_SHIFT_PAGE_DOWN    : 1020,
   
   VK_ALT                 : 1024,
 
   VK_CTRL_SHIFT_CURSOR_RIGHT : 3063,
   VK_CTRL_SHIFT_CURSOR_LEFT  : 3064,
   VK_CTRL_SHIFT_HOME         : 3065,
   VK_CTRL_SHIFT_END          : 3066,

   VK_CTRL_DEL                : 2175,
   VK_CTRL_ENTER              : 2061,
};


/** codes for window events generated by 4GL */
var SE_WINDOW_CLOSE     = 1078;
var SE_WINDOW_MAXIMIZED = 1094;
var SE_WINDOW_MINIMIZED = 1095;
var SE_WINDOW_RESTORED  = 1096;
var SE_WINDOW_RESIZED   = 1120;
/** synthetic event code */
var SE_MENU_DROP = 1079;

var osEvents = [];

osEvents[SE_WINDOW_CLOSE] = true;
osEvents[SE_WINDOW_MAXIMIZED] = true;
osEvents[SE_WINDOW_MINIMIZED] = true;
osEvents[SE_WINDOW_RESIZED] = true;
osEvents[SE_WINDOW_RESTORED] = true;
osEvents[SE_MENU_DROP] = true;

// differences for multi-byte keys
var extendedKeys = [];

/**
 * Returns the basic name for keys in the rage 0-511.
 * 
 * @param    {Number} code
 *           The key code to translate.
 * 
 * @return   {String}
 *           The key label string
 */
function hasBasicName1(code)
{
   var idx = code & 0x01ff;
   
   return basicKeys.isNotEmpty(idx);
}
/**
 * Returns the basic name for keys in the rage 512-655.
 * 
 * @param    {Number} code
 *           The key code to translate.
 * 
 * @return   {String}
 *           The key label string
 */
function hasBasicName2(code)
{
   var idx = code & 0x01ff;
   if (idx > 143)
   {
      return basicKeys.isNotEmpty(idx);
   }
   
   return basicKeys.isNotEmpty(idx + 512);
}

p2j.keymap = (function()
{
   
   /**
    * Process CTRL key.
    * 
    * @param    {KeyboardEvent} evt
    *           The event object.
    * @param   {Number} code
    *           The char code.
    * @param   {Boolean} isChui
    *           The CHUI mode.
    * 
    * @return   {Number}
    *           The mapped key code.
    */
   function processCtrlModifier(evt, code, isChui)
   {
      var ctrlCode = -1;
      
      // process COMMAND key for Safari as CTRL key
      var metaToCtrl = p2j.isSafari && evt.metaKey;
      
      if (evt.ctrlKey || metaToCtrl)
      {         
         if (code === 50 || code === 64) //'2'
         {
            ctrlCode = 0;
         }
         else if (code === 54 || code === 94)  //'6'
         {
            ctrlCode = 30;
         }
         else if (code === 45 || code === 95) //'-'
         {
            ctrlCode = 31;
         }
         else if (code === 92 || code === 124) //'\'
         {           
            ctrlCode = 28;       
         }
         else if (code === 93 || code === 125) //']'
         {           
            ctrlCode = 29;       
         }
         else
         {
            if (code >= 32 && code < 127)
            {
               // clear bit 7 (e.g. this takes CTRL-C which is 0x43/decimal 67 and yields 0x03) 
               ctrlCode = code & 0xffbf;
               
               // VK_ESCAPE is NOT DELIVERED, 4GL doesn't ever return this                        
               if (isChui && ctrlCode === keys.ESCAPE)
               {
                  ctrlCode = -1;
               }
            }
         }
      }
      
      return ctrlCode;
   };
   
   /**
    * Map scan code to ASCII printable,
    * 
    * @param    {KeyboardEvent} evt
    *           The event object.
    * 
    * @return   {Number}
    *           The mapped key code.
    */
   function mapKey(evt)
   {
      var key = evt.keyCode;
      
      if (p2j.isFirefox)
      {
         switch (key)
         {
            case 59:
               key = 186; //; : 
               break;
            case 107:
               key = 187; //= + 
               break;
            case 109:
               key = 189; //- _
               break;
         }
      }
      
      if (evt.shiftKey)
      {
         switch (key)
         {
            case 48:
               return 41; //)
            case 49:
               return 33; //!
            case 50:
               return 64; //@
            case 51:
               return 35; //#
            case 52:
               return 36; //$
            case 53:
               return 37; //%
            case 54:
               return 94; //^
            case 55:
               return 38; //&
            case 56:
               return 42; //*
            case 57:
               return 40; //(
            case 188:
               return 60; //<
            case 190:
               return 62; //>
            case 191:
               return 63; //?
            case 192:
               return 126; //~
            case 219:
               return 123; //{
            case 220:
               return 124; //|
            case 221:
               return 125; //}
            case 222:
               return 34; //"
            case 186:
               return 58; //:
            case 187:
               return 43; //+
            case 173:
            case 189:
               return 95; //_
         }
      }
      else
      {
         // 0...9
         if (key >= 48 && key <= 57)
         {
            return key;
         }

         switch (key)
         {
            case 188:
               return 44; //,
            case 190:
               return 46; //.
            case 191:
               return 47; ///
            case 192:
               return 96; //`
            case 219:
               return 91; //[
            case 220:
               return 92; //\
            case 221:
               return 93; //]
            case 222:
               return 39; //'
            case 186:
               return 59; //;
            case 187:
               return 61; //=
            case 173:
            case 189:
               return 45; //-
            // numpad
            case 96:
            case 97:
            case 98:
            case 99:
            case 100:
            case 101:
            case 102:
            case 103:
            case 104:
            case 105:
               return key - 48; //0...9
            case 106:
               return 42; //*
            case 107:
               return 43; //+
            case 109:
               return 45; //-
            case 110:
               return 46; //.
            case 111:
               return 47; ///
         }
      }
      
      // A...Z  a...z
      if (key >= 65 && key <= 90)
      {
         return evt.shiftKey ? key : (key + 32);
      }
      
      return -1;
   };

   /**
    * Creates browser keys to Progress CHUI keys mapping.
    * 
    * @param    {Array} xlate
    *           The translation table that maps the browser/OS key codes to Progress 4GL codes.
    * @param    {Array} xlateAlt
    *           The translation table that maps the browser/OS key codes to Progress 4GL codes
    *           if the ALT modifier is set.
    */
   function CHUIKeyboardMapping(xlate, xlateAlt)
   {
      // normal keys
      xlate[keys.BACKSPACE] = 8;
      xlate[keys.TAB] = 9;
      xlate[keys.ENTER] = 13;
      xlate[keys.UP_ARROW] = 501;
      xlate[keys.DOWN_ARROW] = 502;
      xlate[keys.RIGHT_ARROW] = 503;
      xlate[keys.LEFT_ARROW] = 504;
      xlate[keys.HOME] = 505;
      xlate[keys.END] = 506;
      xlate[keys.PAGE_UP] = 507;
      xlate[keys.PAGE_DOWN] = 508;
      xlate[keys.INSERT] = 510;
      xlate[keys.DELETE] = 127;     
      xlate[keys.F1] = 301;
      xlate[keys.F2] = 302;
      xlate[keys.F3] = 303;
      xlate[keys.F4] = 304;
      xlate[keys.F5] = 305;
      xlate[keys.F6] = 306;
      xlate[keys.F7] = 307;
      xlate[keys.F8] = 308;
      xlate[keys.F9] = 309;
      xlate[keys.F10] = 310;
      xlate[keys.F11] = 311;
      xlate[keys.F12] = 312;

      // ALT-keys
      xlateAlt[keys.UP_ARROW] = 559;      
      xlateAlt[keys.DOWN_ARROW] = 518;
      xlateAlt[keys.RIGHT_ARROW] = 553;
      xlateAlt[keys.LEFT_ARROW] = 538;
      xlateAlt[keys.PAGE_UP] = 548;
      xlateAlt[keys.PAGE_DOWN] = 543;
      xlateAlt[keys.INSERT] = 533;
      xlateAlt[keys.DELETE] = 127;     
      xlateAlt[keys.HOME] = 528;
      xlateAlt[keys.END] = 523;
      xlateAlt[keys.F1] = 313;
      xlateAlt[keys.F2] = 314;
      xlateAlt[keys.F3] = 315;
      xlateAlt[keys.F4] = 316;
      xlateAlt[keys.F5] = 317;
      xlateAlt[keys.F6] = 318;
      xlateAlt[keys.F7] = 319;
      xlateAlt[keys.F8] = 320;
      xlateAlt[keys.F9] = 321;
      xlateAlt[keys.F10] = 322;
      xlateAlt[keys.F11] = 323;
      xlateAlt[keys.F12] = 324;

      this.mapKey = mapKey;

      this.processCtrl = processCtrl;

      /**
       * Process CTRL key.
       * 
       * @param    {KeyboardEvent} evt
       *           The event object.
       * 
       * @param   {Number} code
       *           The char code.
       * 
       * @return   {Number}
       *           The mapped key code.
       */
      function processCtrl(evt, code)
      {
         return processCtrlModifier(evt, code, true);
      };
   }

   /**
    * Defines GUI key mapping.
    * 
    * @param    {Array} xlate
    *           The translation table that maps the browser/OS key codes to Progress 4GL codes.
    * @param    {Array} xlateAlt
    *           The translation table that maps the browser/OS key codes to Progress 4GL codes
    *           if the ALT modifier is set.
    */
   function GUIKeyboardMapping(xlate, xlateAlt)
   {
      xlate[keys.F1]  = Progress4GLKeys.VK_F1;
      xlate[keys.F2]  = Progress4GLKeys.VK_F2;
      xlate[keys.F3]  = Progress4GLKeys.VK_F3;
      xlate[keys.F4]  = Progress4GLKeys.VK_F4;
      xlate[keys.F5]  = Progress4GLKeys.VK_F5;
      xlate[keys.F6]  = Progress4GLKeys.VK_F6;
      xlate[keys.F7]  = Progress4GLKeys.VK_F7;
      xlate[keys.F8]  = Progress4GLKeys.VK_F8;
      xlate[keys.F9]  = Progress4GLKeys.VK_F9;
      xlate[keys.F10] = Progress4GLKeys.VK_F10;
      xlate[keys.F11] = Progress4GLKeys.VK_F11;
      xlate[keys.F12] = Progress4GLKeys.VK_F12;
      xlate[keys.F13] = Progress4GLKeys.VK_F13;
      xlate[keys.F14] = Progress4GLKeys.VK_F14;
      xlate[keys.F15] = Progress4GLKeys.VK_F15;
      xlate[keys.F16] = Progress4GLKeys.VK_F16;
      xlate[keys.F17] = Progress4GLKeys.VK_F17;
      xlate[keys.F18] = Progress4GLKeys.VK_F18;
      xlate[keys.F19] = Progress4GLKeys.VK_F19;
      xlate[keys.F20] = Progress4GLKeys.VK_F20;
      
      xlate[keys.HOME]   = Progress4GLKeys.VK_HOME;
      xlate[keys.INSERT] = Progress4GLKeys.VK_INSERT;
      xlate[keys.DELETE] = Progress4GLKeys.VK_DELETE;
      xlate[keys.END]        = Progress4GLKeys.VK_END;
      xlate[keys.PAGE_DOWN]   = Progress4GLKeys.VK_PAGE_DOWN;
      xlate[keys.PAGE_UP]     = Progress4GLKeys.VK_PAGE_UP;

      xlate[keys.LEFT_ARROW]  = Progress4GLKeys.VK_CURSOR_LEFT;
      xlate[keys.RIGHT_ARROW] = Progress4GLKeys.VK_CURSOR_RIGHT;
      xlate[keys.UP_ARROW]    = Progress4GLKeys.VK_CURSOR_UP;
      xlate[keys.DOWN_ARROW]  = Progress4GLKeys.VK_CURSOR_DOWN;

      xlate[keys.ENTER]     = Progress4GLKeys.VK_ENTER;
      xlate[keys.TAB]       = Progress4GLKeys.VK_TAB;
      xlate[keys.BACKSPACE] = Progress4GLKeys.VK_BACKSPACE;

      xlate[keys.ESCAPE] = Progress4GLKeys.VK_ESCAPE;
      // ALT is not processed by 4GL... => Processed as menubar activation key.
      xlate[keys.ALT] = Progress4GLKeys.VK_ALT;

      basicKeys[13] = "ENTER";
      basicKeys[509] = "SHIFT-TAB";
      basicKeys[512] = "DELETE";
      basicKeys[525] = "ENTER";
      basicKeys[610] = "MIDDLE-MOUSE-UP";
      basicKeys[618] = "MIDDLE-MOUSE-DOWN";
      basicKeys[626] = "MIDDLE-MOUSE-CLICK";
      basicKeys[650] = "MIDDLE-MOUSE-DBLCLICK";
      
      // differences for multi-byte keys
      extendedKeys[639]  = "SHIFT-DEL";
      extendedKeys[1021] = "SHIFT-SHIFT-TAB";
      extendedKeys[1024] = "ALT";
      extendedKeys[1033] = "ALT-CTRL-I";
      extendedKeys[1037] = "ALT-ENTER";
      extendedKeys[1533] = "ALT-SHIFT-TAB";
      extendedKeys[1536] = "ALT-DELETE";
      extendedKeys[1545] = "SHIFT-ALT-CTRL-I";
      extendedKeys[1549] = "SHIFT-ALT-ENTER";
      extendedKeys[2045] = "SHIFT-ALT-SHIFT-TAB";
      extendedKeys[2061] = "CTRL-ENTER";
      extendedKeys[2557] = "CTRL-SHIFT-TAB";
      extendedKeys[2560] = "CTRL-DELETE";
      extendedKeys[2573] = "CTRL-SHIFT-ENTER";
      extendedKeys[3069] = "CTRL-SHIFT-SHIFT-TAB";
      extendedKeys[3085] = "CTRL-ALT-ENTER";
      extendedKeys[3581] = "CTRL-ALT-SHIFT-TAB";
      extendedKeys[3584] = "CTRL-ALT-DELETE";
      extendedKeys[3597] = "CTRL-SHIFT-ALT-ENTER";
      extendedKeys[4093] = "CTRL-SHIFT-ALT-SHIFT-TAB";

      /**
       * Check if the given key code has a label, defined by this rule 1.
       * <p>
       * Rule 1 means the code is within a range where:
       * <ul>
       *   <li>codes 0-511 index the table of basic key names;
       *   <li>codes outside of 0-511 range produce regular labels with CTRL-,
       *       SHIFT-, ESC- prefixes and their combinations.
       * </ul>
       * 
       * @param    code
       *           The key code to check.
       *           
       * @return   <code>true</code> if rule 1 must be applied to the key code.
       */
      function useRule1(code)
      {
         return (code >= 1013 && code <= 1023) ||
                (code >= 1025 && code <= 1279) ||
                (code >= 0    && code <= 511 ) ||
                (code >= 520  && code <= 521)  ||
                code == 525                    ||
                code == 539                    ||
                (code >= 656  && code <= 767)  ||
                (code >= 812  && code <= 1011) ||
                (code >= 1324 && code <= 1523) ||
                (code >= 1525 && code <= 1535) ||
                (code >= 1544 && code <= 1545) ||
                code == 1549                   ||
                code == 1563                   ||
                (code >= 1663 && code <= 1791) ||
                (code >= 1836 && code <= 2035) ||
                (code >= 2037 && code <= 2303) ||
                (code >= 2348 && code <= 2547) ||
                (code >= 2549 && code <= 2559) ||
                (code >= 2568 && code <= 2569) ||
                code == 2573                   ||
                code == 2587                   ||
                (code >= 2687 && code <= 2815) ||
                (code >= 2860 && code <= 3059) ||
                (code >= 3061 && code <= 3327) ||
                (code >= 3372 && code <= 3571) ||
                (code >= 3573 && code <= 3583) ||
                (code >= 3592 && code <= 3593) ||
                code == 3597                   ||
                code == 3611                   ||
                (code >= 3711 && code <= 3839) ||
                (code >= 3884 && code <= 4083) ||
                (code >= 4085 && code <= 4095);
      };

      /**
       * Tests if the given key code doesn't have a label.
       * 
       * @param    code
       *           The key code to check.
       * 
       * @return   {Boolean}
       *           True if the given key code doesn't have a label.
       */
      function useRule3(code)
      {
         return code == 634 ||
                (code >= 256  && code <= 299)  ||
                code === 500                   ||
                (code >= 523  && code <= 524)  ||
                (code >= 526  && code <= 534)  ||
                code === 540                   ||
                (code >= 542  && code <= 550)  ||
                (code >= 561  && code <= 585)  ||
                (code >= 590  && code <= 608)  ||
                (code >= 613  && code <= 616)  ||
                (code >= 621  && code <= 624)  ||
                (code >= 629  && code <= 632)  ||
                (code >= 637  && code <= 639)  ||
                (code >= 641  && code <= 648)  ||
                (code >= 653  && code <= 655)  ||
                (code >= 768  && code <= 811)  ||
                code === 1012                  ||
                (code >= 1280 && code <= 1323) ||
                code == 1524                   ||
                (code >= 1547 && code <= 1548) ||
                (code >= 1550 && code <= 1558) ||
                code === 1564                  ||
                (code >= 1566 && code <= 1574) ||
                (code >= 1585 && code <= 1609) ||
                (code >= 1614 && code <= 1662) ||
                (code >= 1792 && code <= 1835) ||
                code == 2036                   ||
                (code >= 2304 && code <= 2347) ||
                code === 2548                  ||
                (code >= 2571 && code <= 2572) ||
                (code >= 2574 && code <= 2582) ||
                code === 2588                  ||
                (code >= 2590 && code <= 2598) ||
                (code >= 2609 && code <= 2633) ||
                (code >= 2638 && code <= 2686) ||
                (code >= 2816 && code <= 2859) ||
                code == 3060                   ||
                (code >= 3328 && code <= 3371) ||
                code == 3572                   ||
                (code >= 3595 && code <= 3596) ||
                (code >= 3598 && code <= 3606) ||
                code === 3612                  ||
                (code >= 3614 && code <= 3622) ||
                (code >= 3633 && code <= 3657) ||
                (code >= 3662 && code <= 3710) ||
                (code >= 3840 && code <= 3883) ||
                code === 4084; 
      };

      this.hasLabelForCode = hasLabelForCode;
      
      /**
       * Tests if the given code has a label.
       * 
       * @param    {Number} code
       *           The Progress 4GL key code.
       * 
       * @returns   {Boolean}
       *            True iff the given code has a label.
       */
      function hasLabelForCode(code)
      {
         code = code % 4096;
         if (osEvents.isNotEmpty(code))
         {
            return true;
         }
         if (extendedKeys.isNotEmpty(code))
         {
            return true;
         }
         if (useRule1(code))
         {
            return hasBasicName1(code);
         }
         if (useRule3(code))
         {
            return false;
         }
         return hasBasicName2(code);
      };

      this.shiftModifierExceptions = shiftModifierExceptions;

      /**
       * All codes that have labels and don't satisfy the condition if the shift modifier
       * is applied to the original code has a label not ending on the original label without
       * "CTRL" prefix.
       * 
       * @param    {Number} code
       *           The key code.
       * @return   {Boolean}
       *           True if the shift modifier is applied to this code has a label not ending
       *           on the original code label without "CTRL" prefix.
       */
      function shiftModifierExceptions(code)
      {
         if (code < 97)
         {
            return shiftModifierPattern1(code);
         }
         else if (code < 1024)
         {
            return shiftModifierPattern2(code);
         }
         else if (code < 2048)
         {
            return shiftModifierPattern1(code - 1024);
         }
         else if (code < 3072)
         {
            return shiftModifierPattern1(code - 2048);
         }
         else
         {
            return shiftModifierPattern1(code - 3072);
         }
//         return (0 <= code && code <= 7) || code == 10 || (23 <= code && code <= 26) || code == 29 ||
//         (39 <= code && code <= 48) || (74 <= code && code <= 77) ||
//         
//         (97 <= code && code <= 100) || (105 <= code && code <= 108) ||
//         (113 <= code && code <= 116) || (code == 121) || (123 == code) || (code == 124) || code == 128
//         || (137 <= code && code <= 140) ||
//         
//         (1024 <= code && code <= 1031) || code == 1034 || (1047 <= code && code <= 1050) || code == 1053 ||
//         (1063 <= code && code <= 1072) || (1098 <= code && code <= 1101) ||
//         
//         (2048 <= code && code <= 2055) || code == 2058 || (2071 <= code && code <= 2074) || code == 2077 ||
//         (2087 <= code && code <= 2096) || (2122 <= code && code <= 2125) ||
//         
//         (3072 <= code && code <= 3079) || code == 3082 || (3095 <= code && code <= 3098) || code == 3101 ||
//         (3111 <= code && code <= 3120) || (3146 <= code && code <= 3149);
      }

      /**
       * All codes in [0,96] that have labels and don't satisfy the condition if the shift modifier
       * is applied to the original code has a label not ending on the original label without
       * "CTRL" prefix. It defines a base exception pattern.
       * 
       * @param    {Number} code
       *           The key code.
       * @return   {Boolean}
       *           True iff the shift modifier is applied to this code has a label not ending
       *           on the original code label without "CTRL" prefix.
       */
      function shiftModifierPattern1(code)
      {
         return (0 <= code && code <= 7) || code == 10 || (23 <= code && code <= 26) || code == 29 ||
         (39 <= code && code <= 48) || (74 <= code && code <= 77);
      }

      /**
       * All codes in [97,1023] that have labels and don't satisfy the condition if the shift modifier
       * is applied to the original code has a label not ending on the original label without
       * "CTRL" prefix. It defines a unique exception set.
       * 
       * @param    {Number} code
       *           The key code.
       * @return   {Boolean}
       *           True if the shift modifier is applied to this code has a label not ending
       *           on the original code label without "CTRL" prefix.
       */
      function shiftModifierPattern2(code)
      {
         return (97 <= code && code <= 100) || (105 <= code && code <= 108) ||
         (113 <= code && code <= 116) || (code == 121) || (123 == code) || (code == 124) || code == 128
         || (137 <= code && code <= 140);
      }
      
      this.mapKey = mapKey;
      
      this.processCtrl = processCtrl;

      /**
       * Process CTRL key.
       * 
       * @param    {KeyboardEvent} evt
       *           The event object.
       * 
       * @param   {Number} code
       *           The char code.
       * 
       * @return   {Number}
       *           The mapped key code.
       */
      function processCtrl(evt, code)
      {
         return processCtrlModifier(evt, code, false);
      };

   }
   
   /**
    * Key mapping object which will be returned from this initialization function.
    */
   var me = {};
   
   /* normal key mapping */   
   me.xlate = {};
   
   /* ALT key mapping */
   me.xlateAlt = {};         
   
   /**
    * Init function.
    */
   me.init = function(cfg)
   {
      var keyboardMapping;

      if (cfg.isGui)
      {
         keyboardMapping = new GUIKeyboardMapping(me.xlate, me.xlateAlt);
      }
      else
      {
         keyboardMapping = new CHUIKeyboardMapping(me.xlate, me.xlateAlt);
      }
      me.keyboardMapping = keyboardMapping;
      me.mapKey          = keyboardMapping.mapKey;
      me.processCtrl     = keyboardMapping.processCtrl;
   };

   return me;
}());