/*
** Module   : p2j.socket.js
** Abstract : WebSockets communication management
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140204 Get attributes as a binary message. On quit command go back to login page. 
** 003 MAG 20140211 Implements COLOR.
** 004 MAG 20140214 Implements referrer redirection.
** 005 MAG 20140219 Performance improvements.
** 006 MAG 20140402 Implements interactive child process mode.
** 007 GES 20150324 Moved this class to a common resource base and made it generic (so that it
**                  can be used for both ChUI and GUI web clients.
** 008 SBI 20150731 Applied "use strict" directive, fixed undeclared variables.
** 009 GES 20150709 Added support for MSG_READ_CLIPBOARD (server-driven clipboard request) and
**                  MSG_WRITE_CLIPBOARD, fixed a network order for reading int32
**                  from binaries message. Added infrastructure for font and metrics related
**                  requests.
** 010 GES 20150818 Added more GUI message support including fonts, cursors and z-order.
**                  Once the web socket has connected, notify the Java side, so that it can start
**                  sending requests. Fixed sendInt16/32BinaryMessage - it was sending the bits
**                  in the wrong order. Added mouse support.  Create font returns the font's
**                  index in the local font-table.
** 011 CA  20150925 Added explicit track of widget z-order.
** 012 CA  20151013 Added support for window resize for the GUI web client, added to set
**                  the current editors selection.  Added z-order operations. Fixed incorrect
**                  arguments provided for layoutParagraphWorker. 
** 013 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     SBI 20151026 Fixed mouse processing for modal windows.
** 014 CA  20151126 When waiting for a message result to be received from the JS side, each message
**                  must be uniquely identified; otherwise, if the same message type is sent in
**                  the same batch more than once, it will collide and it will not preserve the
**                  message result.
** 015 IAS 20160217 LOAD-MOUSE-POINTER ans SET-WAIT-STATE support
** 016 SBI 20151217 Added to display a wait cursor if drawing is in process.
**     SBI 20160107 Supported new message with 0xF1 id to test if the target font family is
**                  available for the browser.
**     CA  20160323 Added a drawing cache (and cache management, when the hash was expired by the
**                  client-side).  Added MSG_PARSE_EDITOR_CONTENT, MSG_[GET|SENT]_TEXT_WIDTHS[_BATCH].
**     SBI 20160225 Implemented MSG_PARTIAL, added MSG_PING_PONG and idle/connectivity timers,
**                  added the default "Stay on page or Leave page" dialog if a user tries to close
**                  or to refresh the web application page, added named message types constants,
**                  fixed TypedArray.slice is unsupported by IE.
** 017 HC  20160406 Overhaul of window-activation logic.
** 018 SBI 20160420 Added MSG_OPEN_URL to open the target page in the browser. Fixed the window
**                  location coordinates to have 32 bits size if they are send with
**                  MSG_SET_WINDOW_LOC and MSG_WINDOW_RESIZED. Fixed ControlTimer to start it
**                  again only if the timer has been done.
**     CA  20160418 Added infrastructure for message passing from converted code to a 
**                  customer-specific application or calls from the customer-specific application 
**                  into the converted code (i.e. invoke external programs, procedures or 
**                  functions).  Can be used only when the P2J client runs embedded in a 
**                  customer-specific application.
** 019 SBI 20160519 Safari doesn't support "const" constants declarations in JS.
**     SBI 20160615 Added MSG_INVALIDATE_SELECTION.
** 020 SBI 20160627 Changed to parse PROCESS_MOUSE_WIDGETS message correctly.
** 021 CA  20170104 Added flag to track window's modal state.
** 022 CA  20170203 Notify the driver that the client is ready (useful for embedded mode, to 
**                  notify the embedded app the client can handle messages).
**     CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 023 CA  20170330 In embedded mode, get the real iframe dimension as desktop dimension.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

p2j.socket = (function()
{
   /** Socket object which will be returned from this initialization function. */
   var me = {};
   
   /** Defines messages types. */
   var types = {
                  /** Message that the web page has finished loading/initializing. */
                  MSG_PAGE_LOADED         : 0xFF,
                  
                  /** Progress compatible keyboard message. */
                  MSG_KEY                 : 0x01,
                  
                  /** Virtual VT100 terminal keyboard message. */
                  MSG_KEY_VT100           : 0x02,
                  
                  /** Paste message. */
                  MSG_PASTE               : 0x03,
                  
                  /** Contents from the clipboard. */
                  MSG_CLIPBOARD_CONTENTS  : 0x04,
                  
                  /** Web client is requesting the current selection text. */
                  MSG_CLIPBOARD_PREPARE   : 0x05,
                  
                  /** Web client is sending computed paragraph height. */
                  MSG_SENT_PAR_HEIGHT     : 0x06,
                  
                  /** Web client is sending computed text height. */
                  MSG_SENT_TEXT_HEIGHT    : 0x07,
                  
                  /** Web client is sending computed text width. */
                  MSG_SENT_TEXT_WIDTH     : 0x08,
                  
                  /** Web client is sending computed font height. */
                  MSG_SENT_FONT_HEIGHT    : 0x09,
                  
                  /** Web client is sending computed font widths. */
                  MSG_SENT_FONT_WIDTHS    : 0x0A,
                  
                  /** Web client is informing the font has been created. */
                  MSG_DONE_CREATE_FONT    : 0x0B,
                  
                  /** Web client is informing the font has been derived. */
                  MSG_DONE_DERIVE_FONT    : 0x0C,
                  
                  /** Web client is informing of a mouse event. */
                  MSG_MOUSE_EVENT         : 0x0D,
                  
                  /** Web client is informing of new window location. */
                  MSG_SET_WINDOW_LOC      : 0x0E,
                  
                  /** Web client is informing of new top window. */
                  MSG_WINDOW_ACTIVATED    : 0x0F,
                  
                  /** Web client is informing of new window iconification state. */
                  MSG_WINDOW_ICONIFY      : 0x10,
                  
                  /** Web client is informing of new window location/dimension after resize. */
                  MSG_WINDOW_RESIZED      : 0x11,
                  
                  /** Web client is informing it finished removing the hashes. */
                  MSG_DRAW_HASH_REMOVED   : 0x12,
                  
                  /** Web client is sending computed text widths. */
                  MSG_SENT_TEXT_WIDTHS    : 0x13,
                  
                  /** Web client is sending computed text widths. */
                  MSG_SENT_TEXT_WIDTHS_BATCH : 0x14,
                  
                  /** 
                    * A request to (or response from) a customer-specific application into 4GL
                    * code. 
                    */
                  MSG_P2J_INVOKE             : 0x15,
                  
                  /** A request to (or response from) a message to delete a handle. */
                  MSG_P2J_DELETE_HANDLE      : 0x16,
                  
                  /** 
                   * A request to (or response from) a call from 4GL code into the
                   * customer-specific application.
                   */
                  MSG_REMOTE_CALL            : 0x17,
                  
                  /** FWD Client is ready  */
                  MSG_CLIENT_READY           : 0x18,
                  
                  /** 
                   * A request to (or response from) a customer-specific application to subscribe
                   *  to a named event.
                   */
                  MSG_P2J_SUBSCRIBE          : 0x19,
               
                  /** 
                   * A request to (or response from) a customer-specific application to 
                   * unsubscribe from a named event.
                   */
                  MSG_P2J_UNSUBSCRIBE        : 0x1A,
               
                  /** 
                   * A request for (or response from) a publish event generated by the business
                   * logic and sent to the remote side.
                   */
                  MSG_P2J_PUBLISH            : 0x1B,

                  /** 
                   * A request for (or response from) a call to create a new external resource.
                   */
                  MSG_P2J_CREATE_RESOURCE     : 0x1C,
                  
                  /** Web client is sending desktop dimension. */
                  MSG_SENT_DESKTOP_DIMENSION   : 0x1D,

                  /** Clear screen message. */
                  MSG_CLEAR               : 0x80,
                  
                  /** Draw text on screen. */
                  MSG_DRAW                : 0x81,
                  
                  /** Set cursor position. */
                  MSG_CURSOR_POS          : 0x82,
                  
                  /** Set cursor status. */
                  MSG_CURSOR_STATUS       : 0x83,
                  
                  /** Beep message. */
                  MSG_BEEP                : 0x84,
                  
                  /** Quit message. */
                  MSG_QUIT                : 0x85,
                  
                  /** Switch input processing to virtual VT100 terminal mode. */
                  MSG_VT100               : 0x86,
                  
                  /** Request clipboard contents from web client. */
                  MSG_READ_CLIPBOARD      : 0x87,
                  
                  /** Copy the given text to the web client's clipboard. */
                  MSG_WRITE_CLIPBOARD     : 0x88,
                  
                  /** Create a new top-level window. */
                  MSG_CREATE_WINDOW       : 0x89,
                  
                  /** Create a new child window. */
                  MSG_CREATE_CHILD_WINDOW : 0x8A,
                  
                  /** Destroy the specified window or child window. */
                  MSG_DESTROY_WINDOW      : 0x8B,
                  
                  /** Change the visibility for the specified window or child window. */
                  MSG_WINDOW_VISIBILITY   : 0x8C,
                  
                  /** Ask the client for the paragraph height. */
                  MSG_GET_PAR_HEIGHT      : 0x8D,
                  
                  /** Ask the client for the text height. */
                  MSG_GET_TEXT_HEIGHT     : 0x8E,
                  
                  /** Ask the client for the text width. */
                  MSG_GET_TEXT_WIDTH      : 0x8F,
                  
                  /** Ask the client for the font height. */
                  MSG_GET_FONT_HEIGHT     : 0x90,
                  
                  /** Ask the client for the font widths. */
                  MSG_GET_FONT_WIDTHS     : 0x91,
                  
                  /** Ask the client to create a font. */
                  MSG_CREATE_FONT         : 0x92,
                  
                  /** Ask the client to derive a font. */
                  MSG_DERIVE_FONT         : 0x93,
                  
                  /** Set the cursor style. */
                  MSG_SET_CURSOR_STYLE    : 0x94,
                  
                  /** Restack the z-order of all windows. */
                  MSG_RESTACK_WINDOWS     : 0x95,
                  
                  /** Send the mouse-aware widget details. */
                  PROCESS_MOUSE_WIDGETS   : 0x96,
                  
                  /** Disable/enable mouse event processing. */
                  CAPTURE_MOUSE           : 0x97,
                  
                  /** Disable/enable OS event processing. */
                  ENABLE_OS_EVENTS        : 0x98,
                  
                  /** Window state changed event. */
                  SET_ICONIFICATION_STATE : 0x99,
                  
                  /** Resizeable widget. */
                  SET_RESIZEABLE_WINDOW   : 0x9A,
                  
                  /** Notifies the client about the current selection. */
                  MSG_CURRENT_SELECTION   : 0x9B,
                  
                  /** Requests the client to move a window to the top of the z-order. */
                  MSG_MOVE_TO_TOP         : 0x9C,
                  
                  /** Requests the client to move a window to the bottom of the z-order. */
                  MSG_MOVE_TO_BOTTOM      : 0x9D,
                  
                  /** Change the sensitivity for the specified window or child window. */
                  MSG_WINDOW_SENSITIVITY  : 0x9E,
                  
                  /** Register hoverable widget */
                  MSG_REGISTER_HOVERABLE_WIDGET  : 0x9F,
                  
                  /** Deregister widget */
                  MSG_DEREGISTER_WIDGET  : 0xA0,
                  
                  /** Set mouse pointer */
                  MSG_SET_MOUSE_PTR  : 0xA1,
                  
                  /** Lock mouse pointer */
                  MSG_LOCK_MOUSE_PTR  : 0xA2,
                  
                  /** Lock mouse pointer */
                  MSG_SET_WAIT_STATE  : 0xA3,
                  
                  /** Parse the editor's content. */
                  MSG_PARSE_EDITOR_CONTENT : 0xA4,
                  
                  /** Ask the client for the text's widths. */
                  MSG_GET_TEXT_WIDTHS : 0xA5,
                  
                  /** Ask the client for the texts' widths - in a batch. */
                  MSG_GET_TEXT_WIDTHS_BATCH : 0xA6,

                  /** Ask the client for the desktop dimension. */
                  MSG_GET_DESKTOP_DIMENSION : 0xA7,
                  
                  /** Requests the client if the target font is installed. */
                  MSG_IS_FONT_INSTALLED  : 0xF1,
                  
                  /** Inform the javascript side to remove the list of expired hashes. */
                  MSG_REMOVE_EXPIRED_HASH : 0xF2,
                  
                  /** Open the new page or resource in the browser. */
                  MSG_OPEN_URL : 0xF3,
                  
                  /** Notify that the current selection is invalidated. */
                  MSG_INVALIDATE_SELECTION : 0xF4,
                  
                  /** Wraps the large message from the client to the sequence of this partial type messages. */
                  MSG_PARTIAL : 0xFE,
                  
                  /** Ping/Pong messages. */
                  MSG_PING_PONG : 0xFC,
               };
   
   /** Referrer address. */
   var referrer;

   /** Web socket. */
   var ws;
   
   /** Connected flag. */
   var connected = false;
   
   /** The connection count - to distinguish between a reconnect and the initial connection. */
   var connectNum = 0;
   
   /** The count of the socket operations. */
   var callNo = 0;

   /** The count of the drawing operations. */
   var drawNo = 0;

   /** The count of the enable OS events operations. */
   var enableEventsNo = 0;

   /** The maximal size for binary messages that can be accepted by the websocket server. */ 
   var maxBinaryMessage;
   
   /** The maximal idle time for this websocket connection */
   var maxIdleTime;
   
   /**
    * The maximal period within which to establish new connections with the same websocket server
    * instance is possible.
    */
   var watchdogTimeout;
   
   /** The websocket connection url */
   var socketUrl;
   
   /** The Idle Timer */
   var idleTimer;
   
   /** The Connectivity Timer */
   var connectivityTimer;
   
   /** Watch Ping/Pong */
   var pingPongWatcher;
   
   /**
    * True value indicates the web page can be closed or redirected to the login page without
    * the appearance of the default "stay on page or leave page" warning dialog.
    */
   var exitTheApplication = false;
   
   /**
    * The next unique message id.
    */
   var nextMsgId = 0;

   /** The task executor */
   var executor = new TaskExecutor("https://" + window.location.host);
   
   /**
    * The task executor or fast timer to run tasks in the UI thread synchronously or postponed
    * to the next nearest system scheduled time.
    * 
    * @param    {String} origin
    *           An URL that defines what messages are handled by this timer.
    */
   function TaskExecutor(origin)
   {
      var notification = "fast-timer-notification";
      var targetOrigin = origin;
      
      /**
       * The tasks queue to be performed by the fast timer. A task must have "execute" function
       * as its public member.
       */
      var tasksQueue = [];
      
      /**
       * Schedule a task to be executed by the fast timer.
       * 
       * @param    {Task} task
       *           The task to be executed
       * @param    {Boolean} sync
       *           The true indicates the task should be executed synchronously.
       */
      function executeTask(task, sync)
      {
         if (sync)
         {
            task.execute();
         }
         else
         {
            tasksQueue.push(task);
            window.postMessage(notification, targetOrigin);
         }
      }
      
      this.executeTask = executeTask;
      
      /**
       * Executes the next postponed task from the tasks queue as soon as a notification message
       * is caught.
       * 
       * @param    {Event} event
       *           A notification event that is handled by this fast timer
       */
      function executePostponedTask(event)
      {
          if (event.source == window && event.data == notification)
          {
              event.stopPropagation();
              if (tasksQueue.length > 0)
              {
                  var task = tasksQueue.shift();
                  task.execute();
              }
          }
      }
      
      window.addEventListener("message", executePostponedTask, true);
   };

   /**
    * Sends a large message asynchronously via partial messages as payload data nested by
    * MSG_PARTIAL messages.
    * 
    * @param    {Uint8Array} msg
    *           The large source message that should be send by packets.
    * @param    {Boolean} sync
    *           The true value indicates the message should be send synchronously.
    */
   function sendMessageByPackets(msg, sync)
   {
      var task = {position : 0, capacity : msg.byteLength, msg : msg};
      task.msgId = nextMsgId;
      nextMsgId++;
      task.execute = function ()
      {
         var position = task.position;
         var capacity = task.capacity;
         var remaining = capacity - position;
         
         var payloadLength;
         var isLast;
         if (remaining >=  maxBinaryMessage - 10)
         {
            payloadLength = maxBinaryMessage - 10;
            isLast = false;
         }
         else
         {
            payloadLength = remaining;
            isLast = true;
         }
         var limit = position + payloadLength;
         // send partial message
         sendPartialMessage(task.msg, position, limit, payloadLength, task.msgId, isLast);
         
         position  = limit;
         remaining = capacity - position;
         task.position  = position;
         task.remaining = remaining;
         // if there are bytes to send, then schedule new task
         if (remaining > 0)
         {
            executor.executeTask(task, sync);
         }
      };
      
      executor.executeTask(task, sync);
   }
   
   /**
    * Send data.
    * 
    * @param {array} data
    */
   function send(data)
   {
      if (connected) 
      {
         if (data.byteLength <= maxBinaryMessage)
         {
            ws.send(data);
         }
         else
         {
            sendMessageByPackets(data, true);
         }
      }
      else
      {
         p2j.sound.beep();
      }
   };

   /**
    * Send a binary message with the given message type as the first and only byte.
    * 
    * @param    {byte} type
    *           Message type code.
    */
   function sendNotification(type)
   {
      var msg = new Uint8Array(1);
      msg[0] = type;
      
      send(msg);
   };
   
   /**
    * Check if the web socket is connected.
    *
    * @return   The state of the <code>connected</code> flag.
    */
   me.isConnected = function()
   {
      return connected;
   };
   
   /**
    * Sends notification to prepare the clipboard.
    */
   me.sendPrepareClipboard = function()
   {
      sendNotification(types.MSG_CLIPBOARD_PREPARE);
   };
   
   /**
    * Paste Ctrl + V
    */
   me.sendPasteCtrlV = function()
   {
      var msg = new Uint8Array(3);
      msg[0] = types.MSG_PASTE;
      msg[1] = 0;
      msg[2] = 22;
      send(msg);
   };
   
   /**
    * Paste text
    */
   me.sendPasteText = function(text)
   {
      sendStringBinaryMessage(types.MSG_PASTE, text); 
   };
   
   /**
    * Send the clipboard content.
    */
   me.sendClipboardContent = function(text)
   {
      sendStringBinaryMessage(types.MSG_CLIPBOARD_CONTENTS, text);
   };
   
   /**
    * Sends a compatible Progress 4GL key code message. 
    * 
    *  @param    {number} keyCode
    *            The key code
    */
   me.sendKeyCode = function(keyCode)
   {
      var data = new Uint8Array(3);
      data[0] = types.MSG_KEY;
      var c = keyCode;
      data[2] = c & 0xff;
      c = c >> 8;
      data[1] = c & 0xff;
      
      send(data);
   }

   /**
    * Send key code and character code.
    * 
    * @param {number} keyCode key code
    * @param {number} charCode key character code
    */
   me.sendKeyPair = function(keyCode, charCode)
   {
      var data = new Uint8Array(5);
      data[0] = types.MSG_KEY_VT100;
      var c = keyCode;
      data[2] = c & 0xff;
      c = c >> 8;
      data[1] = c & 0xff;
      c = charCode;
      data[4] = c & 0xff;
      c = c >> 8;
      data[3] = c & 0xff;
      
      send(data);
   };

   // mouse related messages
   
   /**
    * Sends mouse events. A mouse event is a holder for these mouse related properties:
    * windowId : window id, opCode : mouse event operation type(click, move, ...),
    * x : an X-screen coordinate, y : an Y-screen coordinate, clicks : event's clicks number,
    * button : which mouse button is related, bitmask : the modifier keys bit mask,
    * mouseWheel : boolean flag indicating additional wheel event's parameters are present,
    * rotation : a direction of the mouse wheel movement (optional), unit : a rotation unit
    * (optional), wid : an id of the event's source widget(optional).
    * 
    * @param    {Object} mouseEvent
    *           The mouse event that is a holder for mouse related properties.
    */
   me.sendMouseEvent = function(mouseEvent)
   {
      var widLength = (mouseEvent.wid == undefined ? 0 : 4);
      var message = new Uint8Array(21 + widLength + (mouseEvent.wheelEvent ? 2 : 0));
      
      var offset = 0;
      
      // message type
      message[0] = types.MSG_MOUSE_EVENT;
      offset = offset + 1;
      
      // 0. the window ID
      writeInt32BinaryMessage(message, offset, mouseEvent.windowId);
      offset = offset + 4;
      
      // 1. the mouse event type (encoded)
      message[offset] = mouseEvent.opCode;
      offset = offset + 1;
      
      // 2. timestamp: 
      var tstamp = (new Date()).getTime();
      writeInt64BinaryMessage(message, offset, tstamp)
      offset = offset + 8;
      
      // 3. coordinates
      writeInt16BinaryMessage(message, offset, mouseEvent.x);
      offset = offset + 2;
      writeInt16BinaryMessage(message, offset, mouseEvent.y);
      offset = offset + 2;
      
      // 4. click count
      message[offset] = mouseEvent.clicks;
      offset = offset + 1;
      
      // 5. which button: evt.which
      message[offset] = mouseEvent.button;
      offset = offset + 1;
   
      // 6. modifier bitmask
      message[offset] = mouseEvent.bitmask;
      offset = offset + 1;
      
      if (mouseEvent.wheelEvent)
      {
         // 7. wheel rotation (neg/poz, depending on direction)
         message[offset] = mouseEvent.rotation;
         offset = offset + 1;
         // 8. amount unit
         message[offset] = mouseEvent.unit;
         offset = offset + 1;
      }
   
      if (mouseEvent.wid != undefined)
      {
         writeInt32BinaryMessage(message, offset, mouseEvent.wid);
         offset = offset + 4;
      }
      
      send(message);
   };
   
   /**
    * Send the window resized message.
    * 
    * @param    {Number} windowId
    *           The target window
    * @param    {Number} x
    *           X-coordinate of the window's position within the screen.
    * @param    {Number} y
    *           Y-coordinate of the window's position within the screen.
    * @param    {Number} width
    *           The window's width
    * @param    {Number} height
    *           The window's height
    */
   me.sendWindowResized = function(windowId, x, y, width, height)
   {
      // send the window activation to the java side
      var message = new Uint8Array(17);
      var offset = 0;
      
      // message type
      message[0] = types.MSG_WINDOW_RESIZED;
      offset = offset + 1;
      
      // 1. the window ID
      writeInt32BinaryMessage(message, offset, windowId);
      offset = offset + 4;

      // 2. the window location - row
      writeInt32BinaryMessage(message, offset, y);
      offset = offset + 4;
      
      // 3. the window location - column
      writeInt32BinaryMessage(message, offset, x);
      offset = offset + 4;
      
      // 4. the window width
      writeInt16BinaryMessage(message, offset, width);
      offset = offset + 2;

      // 5. the window height
      writeInt16BinaryMessage(message, offset, height);
      offset = offset + 2;

      send(message);
   };
   
   /**
    * Send the window location message.
    * 
    * @param    {Number} windowId
    *           The target window
    * @param    {Number} x
    *           X-coordinate of the window's position within the screen.
    * @param    {Number} y
    *           Y-coordinate of the window's position within the screen.
    */
   me.sendWindowLocation = function(windowId, x, y)
   {
      var message = new Uint8Array(13);
      var offset = 0;
   
      // message type
      message[0] = types.MSG_SET_WINDOW_LOC;
      offset = offset + 1;
      
      // 1. the window ID
      writeInt32BinaryMessage(message, offset, windowId);
      offset = offset + 4;

      // 2. coordinates
      writeInt32BinaryMessage(message, offset, x);
      offset = offset + 4;
      writeInt32BinaryMessage(message, offset, y);
      offset = offset + 4;

      send(message);
   };
   
   /**
    * Send a partial message that wraps the payload data.
    * 
    * @param    {Uint8Array} data
    *           The large source message that should be send by packets.
    * @param    {Number} position
    *           It points to the first byte of the payload data.
    * @param    {Number} limit
    *           The upper barrier index of the payload data.
    * @param    {Number} payloadLength
    *           The number of bytes to send starting from the given position.
    * @param    {Number} msgId
    *           The message id that identifies the large source message.
    * @param    {Boolean} isLast
    *           Indicates the last message in the packets to send.
    */
   function sendPartialMessage(data, position, limit, payloadLength, msgId, isLast)
   {
      var msg = new Uint8Array(payloadLength + 10);
      msg[0] = types.MSG_PARTIAL;
      writeInt32BinaryMessage(msg, 1, msgId);
      msg[5] = isLast ? 0 : 1;
      writeInt32BinaryMessage(msg, 6, payloadLength);
      var i, j;
      for(i = position, j = 10; i < limit; i++, j++)
      {
         msg[j] = data[i];
      }
      
      ws.send(msg);
   };
   
   /**
    * Sends the window active event to the server.
    * 
    * @param    {Number} deactivatingWindowId
    *           Id of the window being deactivated. If -1, there is no deactivating
    *           window, i.e. previous active was a non-P2J window.
    * 
    * @param    {Number} activatingWindowId
    *           Id of the window being activated. If -1, there is no activating
    *           window, i.e. new active is a non-P2J window.
    */
   me.sendWindowActivated = function(deactivatingWindowId, activatingWindowId)
   {
      // send the window activation/deactivation to the java side
      var msg = new Uint8Array(9);
      
      // message type
      msg[0] = types.MSG_WINDOW_ACTIVATED;
      
      // 1. the deactivating window ID
      writeInt32BinaryMessage(msg, 1, deactivatingWindowId);
      
      // 2. the deactivating window ID
      writeInt32BinaryMessage(msg, 5, activatingWindowId);
      
      // send the window active event
      send(msg);
   };
   
   /**
    * Sends the window icon state event to the server.
    * 
    * @param    {Number} wid
    *           The target window id.
    * @param    {Boolean} minimized
    *           The target window has been minimized.
    */
   me.sendWindowIconState = function(wid, minimized)
   {
      var msg = new Uint8Array(6);
      
      // message type
      msg[0] = types.MSG_WINDOW_ICONIFY;
      
      // 0. the window ID
      writeInt32BinaryMessage(msg, 1, wid);
      
      // 1. the iconification state
      msg[5] = minimized ? 1 : 0;
      
      // send the mouse event
      send(msg);
   };
   
   /**
    * Send a MSG_P2J_DELETE_HANDLE request, to delete the handle specified in the message.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    */
   me.sendDeleteHandleRequest = function(msgReq)
   {
      var msgId = msgReq["id"];
      var handleId = msgReq["payload"];
      var noError = (msgReq["noError"] ? true : false);
      
      var msg = new Uint8Array(1 + 4 + 1 + 1 + (handleId ? 2 * handleId.length : 0));
      
      var offset = 0;
      
      // message type
      msg[offset] = types.MSG_P2J_DELETE_HANDLE;
      offset = offset + 1;
      
      // 0. the msg Id
      writeInt32BinaryMessage(msg, offset, msgId);
      offset = offset + 4;
      
      // 1. no-error
      msg[offset] = (noError ? 1 : 0);
      offset = offset + 1;
      
      // 2. the handle ID
      msg[offset] = (handleId ? handleId.length : 0);
      offset = offset + 1;
      
      if (handleId)
      {
         writeStringBinaryMessage(msg, handleId, offset);
         offset = offset + 2 * handleId.length;
      }
      
      // send the message
      send(msg);
   };
   
   /**
    * Send a MSG_P2J_SUBSCRIBE request, to subscribe to a specific named event.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    */
   me.sendSubscribeRequest = function(msgReq)
   {
      sendSubscribeImpl(msgReq, true);
   };
   
   /**
    * Send a MSG_P2J_UNSUBSCRIBE request, to unsubscribe from a specific named event.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    */
   me.sendUnsubscribeRequest = function(msgReq)
   {
      sendSubscribeImpl(msgReq, false);
   };
   
   /**
    * Send a MSG_P2J_PUBLISH request, to publish an event.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    */
   me.sendPublishRequest = function(msgReq)
   {
      var msgId = msgReq["id"];
      var payload = msgReq["payload"];
      
      var event     = payload["event"];
      var publisher = payload["publisher"];
      var args      = payload["args"];
      
      var offset = 0;

      // build the message
      var len = 1 + 4 + 1 + 1 + 
                1 + (event ? 2 * event.length : 0) +
                1 + (publisher ? 2 * publisher.length : 0) +
                1;
      
      var sargs = [];
      for (var i = 0; i < args.length; i++)
      {
         sargs[i] = args[i] == undefined ? "" : args[i].toString();
         len = len + 1 + 2 * args[i].length;
      }
      
      var msg = new Uint8Array(len);
      
      // message type
      msg[offset] = types.MSG_P2J_PUBLISH;
      offset = offset + 1;
      
      // 0. the msg Id
      writeInt32BinaryMessage(msg, offset, msgId);
      offset = offset + 4;
      
      // 1. event
      msg[offset] = (event ? event.length : 0);
      offset = offset + 1;
      writeStringBinaryMessage(msg, event, offset);
      offset = offset + 2 * event.length;
      
      // 2. publisher
      msg[offset] = (publisher ? publisher.length : 0);
      offset = offset + 1;
      writeStringBinaryMessage(msg, publisher, offset);
      offset = offset + 2 * publisher.length;
      
      // 3. arguments
      msg[offset] = sargs.length;
      offset = offset + 1;
      for (var i = 0; i < sargs.length; i++)
      {
         msg[offset] = sargs[i].length;
         offset = offset + 1;
         writeStringBinaryMessage(msg, sargs[i], offset);
         offset = offset + 2 * sargs[i].length;
      }
      
      // send the message
      send(msg);
   };
      
   /**
    * Send a MSG_P2J_CREATE_RESOURCE request, to create a new resource.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    */
   me.sendCreateResourceRequest = function(msgReq)
   {
      var msgId = msgReq["id"];
   
      // build the message
      var len = 1 + 4;

      var msg = new Uint8Array(len);
      
      var offset = 0;
      
      // message type
      msg[offset] = types.MSG_P2J_CREATE_RESOURCE;
      offset = offset + 1;
      
      // 0. the msg Id
      writeInt32BinaryMessage(msg, offset, msgId);
      offset = offset + 4;
      
      // send the message
      send(msg);
   };
   
   /**
    * Send a MSG_P2J_SUBSCRIBE or MSG_P2J_UNSUBSCRIBE request.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    */
   function sendSubscribeImpl(msgReq, subscribe)
   {
      var msgId = msgReq["id"];
      var noError = (msgReq["noError"] ? true : false);
      var payload = msgReq["payload"];
      
      var global     = payload["global"];
      var subscriber = payload["subscriber"];
      var event      = payload["event"];
      var publisher  = payload["publisher"];
      
      var offset = 0;

      // build the message
      var len = 1 + 4 + 1 + 1 + 
                1 + (subscriber ? 2 * subscriber.length : 0) +
                1 + (event ? 2 * event.length : 0) +
                1 + (publisher ? 2 * publisher.length : 0);
      var msg = new Uint8Array(len);
      
      // message type
      msg[offset] = (subscribe ? types.MSG_P2J_SUBSCRIBE : types.MSG_P2J_UNSUBSCRIBE);
      offset = offset + 1;
      
      // 0. the msg Id
      writeInt32BinaryMessage(msg, offset, msgId);
      offset = offset + 4;
      
      // 1. no-error
      msg[offset] = (noError ? 1 : 0);
      offset = offset + 1;
      
      // 2. global
      msg[offset] = (global ? 1 : 0);
      offset = offset + 1;
      
      // 3. subscriber
      msg[offset] = (subscriber ? subscriber.length : 0);
      offset = offset + 1;
      if (subscriber)
      {
         writeStringBinaryMessage(msg, subscriber, offset);
         offset = offset + 2 * subscriber.length;
      }
      
      // 4. event
      msg[offset] = (event ? event.length : 0);
      offset = offset + 1;
      if (event)
      {
         writeStringBinaryMessage(msg, event, offset);
         offset = offset + 2 * event.length;
      }
      
      // 5. publisher
      msg[offset] = (publisher ? publisher.length : 0);
      offset = offset + 1;
      if (publisher)
      {
         writeStringBinaryMessage(msg, publisher, offset);
         offset = offset + 2 * publisher.length;
      }
      
      // send the message
      send(msg);
   };

   /**
    * Send a MSG_REMOTE_CALL response, as result from a previous MSG_REMOTE_CALL request.
    *
    * @param   {object} msgResp
    *          An associative array with the message response.
    */
   me.sendRemoteCallResponse = function(msgResp)
   {
      var msgId = msgResp["id"];
      var payload = msgResp["payload"];
      
      var msg = new Uint8Array(1 + 4 + 4 + 2 * payload.length);
      var offset = 0;
      
      // message type
      msg[offset] = types.MSG_REMOTE_CALL;
      offset = offset + 1;
      
      // 0. the msg Id
      writeInt32BinaryMessage(msg, offset, msgId);
      offset = offset + 4;
      
      // 1. the payload length
      writeInt32BinaryMessage(msg, offset, payload.length);
      offset = offset + 4;
      
      // 2. the payload content
      writeStringBinaryMessage(msg, payload, offset);
      offset = offset + 2 * payload.length;
      
      // send the mouse event
      send(msg);
   };
   
   /**
    * Send a MSG_P2J_INVOKE request, to invoke converted, legacy code, on server-side.  This will
    * be routed by the P2J (java) client-side, via an "invoke-later" event, to the server-side.  
    * The P2J client-side, when executing the request, will block waiting for a response, which 
    * in turn will be routed back to the JS driver; this will notify the customer-specific 
    * application window, via a posted message.
    *
    * @param   {object} msgReq
    *          An associative array with the message request.
    */
   me.sendInvokeRequest = function(msgReq)
   {
      var msgId   = msgReq["id"];
      var payload = msgReq["payload"];
      var name    = payload["name"];
      var type    = payload["type"];
      var args    = payload["arguments"];
      var argNum  = 0;
      var noError = (payload["noError"] ? true : false);
      
      // type + msgId + call type + noError + name.length + 2 * name.length
      var msgLen = 1 + 4 + 1 + 1 + 2 + 2 * name.length;
      if (type == p2j.remote.EXTERNAL_PROGRAM)
      {
         // persistent
         msgLen = msgLen + 1;
      }
      else
      {
         var handleId = payload["handle"];

         // handle
         msgLen = msgLen + 1 + (handleId ? (2 * handleId.length) : 0);
      }
      // arguments
      msgLen = msgLen + 2;
      for (var arg in args)
      {
         var argVal = args[arg];
         argNum = argNum + 1;
         
         msgLen = msgLen + 1 + 2 * arg.length;
         msgLen = msgLen + 2 + 2 * argVal.length;
      }
      
      var msg = new Uint8Array(msgLen);
      
      // message type
      var offset = 0;
      
      // 0. the type: 1
      msg[offset] = types.MSG_P2J_INVOKE;
      offset = offset + 1;
      
      // 1. the msg Id: 4
      writeInt32BinaryMessage(msg, offset, msgId);
      offset = offset + 4;
      
      // 2. the remote call type: 1
      msg[offset] = type;
      offset = offset + 1;
      
      // 3. noError
      msg[offset] = (noError ? 1 : 0);
      offset = offset + 1;
      
      // 4. the name: 2 + 2 * name.length
      writeInt16BinaryMessage(msg, offset, name.length);
      offset = offset + 2;
      writeStringBinaryMessage(msg, name, offset);
      offset = offset + 2 * name.length;
      
      if (type == p2j.remote.EXTERNAL_PROGRAM)
      {
         // 5. persistent: 1
         msg[offset] = payload["persistent"] ? 1 : 0;
         offset = offset + 1;
      }
      else
      {
         // 5. handle: 1 + 2 * hlen
         var handleId = payload["handle"];
         
         msg[offset] = (handleId ? handleId.length : 0);
         offset = offset + 1;
         
         if (handleId)
         {
            writeStringBinaryMessage(msg, handleId, offset);
            offset = offset + 2 * handleId.length;
         }
      }
      
      // 6. arguments
      writeInt16BinaryMessage(msg, offset, argNum);
      offset = offset + 2;
      for (var arg in args)
      {
         msg[offset] = arg.length;
         offset = offset + 1;
         writeStringBinaryMessage(msg, arg, offset);
         offset = offset + 2 * arg.length;
         
         var argVal = args[arg];
         writeInt16BinaryMessage(msg, offset, argVal.length);
         offset = offset + 2;
         writeStringBinaryMessage(msg, argVal, offset);
         offset = offset + 2 * argVal.length;
      }
      
      // send the mouse event
      send(msg);
   };
   
   /**
    * Send a binary message with the given message type and the specified byte array serialized.
    *
    * @param    {byte} type
    *           Message type code.
    * @param    {int} msgId
    *           The message ID.
    * @param    {c} c
    *           The array to send.
    */
   function sendByteArrayBinaryMessage(type, msgId, c)
   {
      var msg = new Uint8Array(1 + 4 + 2 + c.length);

      var offset = 0;
      msg[offset] = type;
      offset = offset + 1;

      writeInt32BinaryMessage(msg, offset, msgId);
      offset = offset + 4;
      
      msg[offset] = c.length & 0xff;
      offset = offset + 1;
      msg[offset] = (c.length >> 8) & 0xff
      offset = offset + 1;
      
      for (var i = 0; i < c.length; i++)
      {
         msg[offset] = c[i] & 0xff;
         offset = offset + 1;
      }
      
      send(msg);
   };
   
   /**
    * Send a binary message with the given message type and the specified byte as the second byte.
    *
    * @param    {byte} type
    *           Message type code.
    * @param    {int} msgId
    *           The message ID.
    * @param    {c} c
    *           The byte to send.
    */
   function sendInt8BinaryMessage(type, msgId, c)
   {
      var msg = new Uint8Array(1 + 4 + 1);

      msg[0] = type;
      writeInt32BinaryMessage(msg, 1, msgId);
      msg[5] = c;
      
      send(msg);
   };

   /**
    * Send a binary message with the given message type and the specified 16bit integer as the 
    * second and third bytes.
    *
    * @param    {byte} type
    *           Message type code.
    * @param    {int} msgId
    *           The message ID.
    * @param    {c} c
    *           The 16bit integer to send.
    */
   function sendInt16BinaryMessage(type, msgId, c)
   {
      var msg = new Uint8Array(1 + 4 + 2);

      msg[0] = type;
      writeInt32BinaryMessage(msg, 1, msgId);
      writeInt16BinaryMessage(msg, 5, c);
      
      send(msg);
   };

   /**
    * Send a binary message with the given message type and the specified 32bit integer as the 
    * second, third, forth and fifth bytes.
    *
    * @param    {byte} type
    *           Message type code.
    * @param    {int} msgId
    *           The message ID.
    * @param    {c} c
    *           The 32bit integer to send.
    */
   function sendInt32BinaryMessage(type, msgId, c)
   {
      var msg = new Uint8Array(1 + 4 + 4);

      msg[0] = type;
      writeInt32BinaryMessage(msg, 1, msgId);
      writeInt32BinaryMessage(msg, 5, c);
      
      send(msg);
   };

   /**
    * Send text as a binary message with the given message type as the first byte.
    *
    * @param    {byte} type
    *           Message type code.
    * @param    {String} text
    *           The text to send.
    */
   function sendStringBinaryMessage(type, text)
   {
      var msg = new Uint8Array(text.length * 2 + 1);
      msg[0] = type;

      for (var i = 0, j = 1; i < text.length; i++, j += 2)
      {
         var c = text.charCodeAt(i);
         msg[j + 1] = c & 0xff;
         c = c >> 8;
         msg[j] = c & 0xff;
      }
      
      send(msg);
   };
   
   /**
    * Read text from a binary message with characters interpreted as 16-bit values.
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    */
   me.readStringBinaryMessage = function(message, offset)
   {
      var text = "";

      for (var i = offset; i < message.length; i += 2)
      {
         var c = String.fromCharCode((message[i] << 8) | message[i + 1]);
         
         text = text + c;
      }
      
      return text;
   };
   
   /**
    * Write the specified text a binary into the message, starting at the specified message.
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {string} text
    *           The text to be written.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    */
   function writeStringBinaryMessage(message, text, offset)
   {
      for (var i = 0, j = offset; i < text.length; i++, j += 2)
      {
         var c = text.charCodeAt(i);
         message[j + 1] = c & 0xff;
         c = c >> 8;
         message[j] = c & 0xff;
      }
   };
   
   /**
    * Read text from a binary message with characters interpreted as 16-bit values, with the
    * specified length
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    * @param    {len} length
    *           The length of the string.
    */
   me.readStringBinaryMessageByLength = function(message, offset, len)
   {
      var text = "";

      for (var i = offset; i < offset + len * 2; i += 2)
      {
         var c = String.fromCharCode((message[i] << 8) | message[i + 1]);
         
         text = text + c;
      }
      
      return text;
   };
   
   /**
    * Read a 32-bit integer from a binary message at a specific offset.
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    *
    * @return   The number that was read.
    */
   me.readInt32BinaryMessage = function(message, offset)
   {
      var num = 0;

      for (var i = 0; i < 4; i++)
      {
         num |= message[offset + i] << (8 * (3 - i));
      }
      
      return num;
   };
   
   /**
    * Read a 64-bit integer from a binary message at a specific offset.
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    *
    * @return   The number that was read.
    */
   me.readInt64BinaryMessage = function(message, offset)
   {
      var num = 0;

      for (var i = 0; i < 8; i++)
      {
         num |= message[offset + i] << (8 * (7 - i));
      }
      
      return num;
   };
   
   /**
    * Write a 64-bit integer to a binary message at a specific offset.
    * 
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    * @param    {int} value
    *           The value to write.
    */
   function writeInt64BinaryMessage(message, offset, value)
   {
      // javascript keeps all numbers as decimal, with:
      // - the fraction in bits from 0 to 51
      // - the exponent in bits from 52 to 62
      // - the sign in bit 63
      
      // for this reason, use the hex representation to transform it into 2 32bit integer values
      // and send the high and low bits separately.
      var hex = value.toString(16);

      // left-pad with zeros
      while (hex.length < 16)
      {
         hex = "0" + hex;
      }

      var hexHigh = hex.substring(0, 8);
      var hexLow = hex.substring(8, 16);

      var ihigh = parseInt(hexHigh, 16);
      var ilow = parseInt(hexLow, 16);

      // clear high bits 20 to 30 (52 to 62, the exponent) 
      for (var i = 20; i <= 30; i++)
      {
         ihigh = ihigh & ~(1 << i);
      }

      writeInt32BinaryMessage(message, offset, ihigh);
      offset = offset + 4;
      writeInt32BinaryMessage(message, offset, ilow);
      offset = offset + 4;
   }

   /**
    * Write a 32-bit integer to a binary message at a specific offset.
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    * @param    {int} value
    *           The value to write.
    */
   function writeInt32BinaryMessage(message, offset, value)
   {
      message[offset]     = ((value >> 24) & 0xFF);
      message[offset + 1] = ((value >> 16) & 0xFF);
      message[offset + 2] = ((value >>  8) & 0xFF);
      message[offset + 3] = ( value        & 0xFF);
   }

   /**
    * Write a 16-bit integer to a binary message at a specific offset.
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    * @param    {int} value
    *           The value to write.
    */
   function writeInt16BinaryMessage(message, offset, value)
   {
      message[offset]     = ((value >> 8)  & 0xFF);
      message[offset + 1] = ( value        & 0xFF);
   }
   
   /**
    * Read a 16-bit integer from a binary message at a specific offset.
    *
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    *
    * @return   The number that was read.
    */
   me.readInt16BinaryMessage = function(message, offset)
   {
      var num = 0;

      for (var i = 0; i < 2; i++)
      {
         num |= message[offset + i] << (8 * (1 - i));
      }
      
      return num;
   };

   /** Web socket message handler */
   var messageHandler = function(message)
   {
      var t1 = (new Date()).getTime();
      callNo = callNo + 1;
      
      // reset the last activity time
      idleTimer.reset(t1);
      
      switch (message[0])
      {
         case types.MSG_CLEAR:
            // clear screen 
            p2j.screen.clear();                        
            break;
         case types.MSG_DRAW:
            drawNo++;
            
            p2j.screen.drawRectangles(message);
            var t2 = (new Date()).getTime();
            
            //console.log(callNo + ":" + drawNo + " draw: " + message.length + " done in " + (t2 - t1));
            return;
         case types.MSG_CURSOR_POS:
            // set cursor position
            p2j.screen.setCursorPosition(message[1], message[2]);                        
            break;
         case types.MSG_CURSOR_STATUS:
            // show cursor
            p2j.screen.setCursorStatus(message[1]);                        
            break;
         case types.MSG_BEEP:
            // message beep
            p2j.sound.beep();                        
            break;
         case types.MSG_QUIT:
            // quit
            doRedirectToLoginPage();
            break;
         case types.MSG_VT100:
            // switch mode p2j/vt100
            p2j.keyboard.vt100 = (message[1] == 0) ? false : true;          
            break;
         case types.MSG_READ_CLIPBOARD:
            // server-driven request for clipboard contents
            p2j.clipboard.sendClipboardContents();          
            break;
         case types.MSG_WRITE_CLIPBOARD:
            // The clipboard is changed.
            var text = me.readStringBinaryMessage(message, 1);
            p2j.clipboard.writeClipboard(text);
            break;
         case types.MSG_CREATE_WINDOW:
            // create a top-level window with the given id
            var id = me.readInt32BinaryMessage(message, 1);
            p2j.screen.createWindow(id);
            break;
         case types.MSG_CREATE_CHILD_WINDOW:
            // create a child window with the given id, owner and title
            var offset = 1;
            var modal = (message[offset++] == 1);
            var resizable = (message[offset++] == 1);
            var isShareActivationWithOwner = (message[offset++] == 1);
            var id    = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var owner = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var title = me.readStringBinaryMessage(message, offset);
            p2j.screen.createChildWindow(id, owner, title, resizable, 
                                         isShareActivationWithOwner, modal);
            break;
         case types.MSG_DESTROY_WINDOW:
            // destroy top-level or child window
            var id = me.readInt32BinaryMessage(message, 1);
            var numberImages = me.readInt32BinaryMessage(message, 5);
            var images = [];
            for (var i = 0; i < numberImages; i++)
            {
               images[i] = me.readInt32BinaryMessage(message, 9 + (i * 4));
            }
            p2j.screen.destroyWindow(id, images);
            break;
         case types.MSG_WINDOW_VISIBILITY:
            // change visibility for top-level or child window
            var id = me.readInt32BinaryMessage(message, 1);
            var visible = message[5] === 0 ? false : true;
            p2j.screen.setWindowVisible(id, visible);
            break;
            
         // font and metrics related requests
         case types.MSG_GET_PAR_HEIGHT:
            // paragraph height
            
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var textLength = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var text  = me.readStringBinaryMessageByLength(message, offset, textLength);
            offset = offset + textLength * 2;
            
            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var maxWidth = me.readInt16BinaryMessage(message, offset);
            offset = offset + 2;

            var pheight = p2j.screen.layoutParagraphWorker(null,
                                                           text,
                                                           font, 
                                                           0,
                                                           0, 
                                                           maxWidth);

            sendInt16BinaryMessage(types.MSG_SENT_PAR_HEIGHT, msgId, pheight);
            break;
         case types.MSG_GET_TEXT_HEIGHT:
            // text height
            
            var offset = 1;

            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var textLength = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var text  = me.readStringBinaryMessageByLength(message, offset, textLength);
            offset = offset + textLength * 2;

            var font = me.readInt32BinaryMessage(message, offset);
            
            var theight = p2j.fonts.getTextHeight(font, text);

            sendInt8BinaryMessage(types.MSG_SENT_TEXT_HEIGHT, msgId, theight);
            break;
         case types.MSG_GET_TEXT_WIDTH:
            // text width

            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var textLength = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var text  = me.readStringBinaryMessageByLength(message, offset, textLength);
            offset = offset + textLength * 2;

            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var twidth = p2j.fonts.getTextWidth(font, text);
            
            sendInt16BinaryMessage(types.MSG_SENT_TEXT_WIDTH, msgId, twidth);
            break;
         case types.MSG_GET_FONT_HEIGHT:
            // font height

            var offset = 1;

            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var fheight = p2j.fonts.getFontHeight(font);
            
            sendInt8BinaryMessage(types.MSG_SENT_FONT_HEIGHT, msgId, fheight);
            break;
         case types.MSG_GET_FONT_WIDTHS:
            // font widths

            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var fwidths = p2j.fonts.getFontWidths(font);
            
            sendByteArrayBinaryMessage(types.MSG_SENT_FONT_WIDTHS, msgId, fwidths);
            break;
         case types.MSG_CREATE_FONT:
            // create font

            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var nameLength = me.readInt16BinaryMessage(message, offset);
            offset = offset + 2;

            var name  = me.readStringBinaryMessageByLength(message, offset, nameLength);
            offset = offset + nameLength * 2;
            
            var size = message[offset];
            offset = offset + 1;
            
            var style = message[offset];
            offset = offset + 1;
            
            var defLength = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var b64font = "";
            if (defLength > 0)
            {
               var binFont = '';
               for (var i = 0; i < defLength; i++)
               {
                  binFont += String.fromCharCode(message[offset]);
                  offset = offset + 1;
               }
               b64font = window.btoa(binFont);
            }

            var fontId = p2j.fonts.createFont(font, name, size, style, b64font);
            
            sendInt32BinaryMessage(types.MSG_DONE_CREATE_FONT, msgId, fontId);
            break;
         case types.MSG_DERIVE_FONT:
            // derive font

            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            p2j.fonts.deriveFont(font);
             
            sendInt8BinaryMessage(types.MSG_DONE_DERIVE_FONT, msgId, 1);
            break;
         case types.MSG_SET_CURSOR_STYLE:
//            console.log("setCursorStyle");
            // set cursor style
            var styleId = me.readInt32BinaryMessage(message, 1);
            var wid = me.readInt32BinaryMessage(message, 5);
            p2j.screen.setCursorStyle(styleId, wid);
            break;
         case types.MSG_RESTACK_WINDOWS:
            // restack windows
            var num = me.readInt32BinaryMessage(message, 1);
            var winids = [];
            for (var i = 0; i < num; i++)
            {
               winids.push(me.readInt32BinaryMessage(message, 5 + (i * 4)));
            }
            p2j.screen.restackZOrderEntries(winids);
            break;
         case types.PROCESS_MOUSE_WIDGETS:
             // register/deregister widgets for mouse actions
             
             var offset = 1;
             
             // the number of windows with new widgets
             var windowNo = me.readInt16BinaryMessage(message, offset);
             offset = offset + 2;
             
             for (var j = 0; j < windowNo; j++)
             {
                // the window ID
                var windowID = me.readInt32BinaryMessage(message, offset);
                offset = offset + 4;
                
                var theWindow = p2j.screen.getWindow(windowID);
                
                // the number of new widgets in this window
                var widgetNo = me.readInt16BinaryMessage(message, offset);
                offset = offset + 2;
                
                for (var k = 0; k < widgetNo; k++)
                {
                   // the widget ID
                   var wid = me.readInt32BinaryMessage(message, offset);
                   offset = offset + 4;

                   // the coordinates
                   var x = me.readInt16BinaryMessage(message, offset);
                   offset = offset + 2;
                   
                   var y = me.readInt16BinaryMessage(message, offset);
                   offset = offset + 2;
   
                   // the size
                   var width = me.readInt16BinaryMessage(message, offset);
                   offset = offset + 2;
   
                   var height = me.readInt16BinaryMessage(message, offset);
                   offset = offset + 2;
   
                   // bit-encoded mouse ops: each bit from 1 to 11, if set, represents a mouse
                   // operation as defined in p2j.screen.mouseOps
                   var actions = me.readInt32BinaryMessage(message, offset);
                   offset = offset + 4;
                   
                   if (theWindow)
                   {
                      theWindow.deregisterMouseWidget(wid, true);
                      theWindow.registerMouseWidget(wid, x, y, width, height, actions);
                   }
                }
                
                var allWNo = me.readInt16BinaryMessage(message, offset);
                offset = offset + 2;
                var zOrder = [];
                
                for (var k = 0; k < allWNo; k++)
                {
                  zOrder[k] = me.readInt32BinaryMessage(message, offset);
                  offset = offset + 4;
                }
                
                if (theWindow)
                {
                   theWindow.setWidgetZOrder(zOrder);
                }
                else
                {
                   console.error("Register new widgets failed for undefined window " + windowID);
                   console.error("Set widgets order failed for undefined window " + windowID);
                }
             }

             // the number of windows with dead widgets
             windowNo = me.readInt16BinaryMessage(message, offset);
             offset = offset + 2;
             
             for (var j = 0; j < windowNo; j++)
             {
                // the window ID
                var windowID = me.readInt32BinaryMessage(message, offset);
                offset = offset + 4;
                
                var theWindow = p2j.screen.getWindow(windowID);
                
                // the number of dead widgets in this window
                var widgetNo = me.readInt16BinaryMessage(message, offset);
                offset = offset + 2;
                
                for (var k = 0; k < widgetNo; k++)
                {
                   // the widget ID
                   var wid = me.readInt32BinaryMessage(message, offset);
                   offset = offset + 4;
                   
                   if (theWindow)
                   {
                      theWindow.deregisterMouseWidget(wid, false);
                   }
                }
                
                if (!theWindow)
                {
                   console.error("Deregister dead widgets failed for undefined window " + windowID);
                }
             }

             // the number of windows with new "any widgets"
             windowNo = me.readInt16BinaryMessage(message, offset);
             offset = offset + 2;
             
             for (var j = 0; j < windowNo; j++)
             {
                // the window ID
                var windowID = me.readInt32BinaryMessage(message, offset);
                offset = offset + 4;
                
                var theWindow = p2j.screen.getWindow(windowID);
                
                // the number of new "any widgets" in this window
                var widgetNo = me.readInt16BinaryMessage(message, offset);
                offset = offset + 2;
                
                for (var k = 0; k < widgetNo; k++)
                {
                   // the widget ID
                   var wid = me.readInt32BinaryMessage(message, offset);
                   offset = offset + 4;
                   
                   if (theWindow)
                   {
                      theWindow.deregisterAnyMouseWidget(wid);
                      theWindow.registerAnyMouseWidget(wid);
                   }
                }
                
                if (!theWindow)
                {
                   console.error("Register any widgets failed for undefined window " + windowID);
                }
             }

             // the number of windows with dead "any widgets"
             windowNo = me.readInt16BinaryMessage(message, offset);
             offset = offset + 2;
             
             for (var j = 0; j < windowNo; j++)
             {
                // the window ID
                var windowID = me.readInt32BinaryMessage(message, offset);
                offset = offset + 4;
                
                var theWindow = p2j.screen.getWindow(windowID);
                
                // the number of dead "any widgets" in this window
                var widgetNo = me.readInt16BinaryMessage(message, offset);
                offset = offset + 2;
                
                for (var k = 0; k < widgetNo; k++)
                {
                   // the widget ID
                   var wid = me.readInt32BinaryMessage(message, offset);
                   offset = offset + 4;
                   
                   if(theWindow)
                   {
                      theWindow.deregisterAnyMouseWidget(wid);
                   }
                }
                
                if (!theWindow)
                {
                   console.error("Deregister dead any widgets failed for undefined window " + windowID);
                }
             }

             break;
         case types.CAPTURE_MOUSE:
            // enable/disable mouse events
            p2j.screen.captureMouseEvents(message[1] == 1);
            break;
         case types.ENABLE_OS_EVENTS:
            // enable/disable OS events
            enableEventsNo++;

            var offset = 1;
            
            var widNo = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var wids = [];
            
            for (var j = 0; j < widNo; j++)
            {
               wids.push(me.readInt32BinaryMessage(message, offset));
               offset = offset + 4;
            }
            
            p2j.screen.enableOsEvents(wids, message[offset] == 1);
            var t2 = (new Date()).getTime();
            //console.log(callNo + ":" + enableEventsNo + " enable calls: " + message.length + " done in " + (t2 - t1));
            return;
         case types.SET_ICONIFICATION_STATE:
            // set window iconification state
            var offset = 1;

            var windowID = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var iconified = (message[offset] == 1);
            offset = offset + 1;
            
            var theWindow = p2j.screen.getWindow(windowID);
            //p2j.logger.log("recieved 0x99: iconified = " + iconified + " windowId=" + windowID);
            if (iconified)
            {
               theWindow.iconify();
            }
            else
            {
               theWindow.deiconify();
            }

            break;
         case types.SET_RESIZEABLE_WINDOW:
            // resizeable window
            var offset = 1;
            
            var windowID = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var theWindow = p2j.screen.getWindow(windowID);

            theWindow.resizeable = (message[offset] == 1);
            offset = offset + 1;
            
            if (theWindow.resizeable)
            {
               theWindow.minWidth = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               theWindow.minHeight = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               
               theWindow.maxWidth = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               theWindow.maxHeight = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
            }
            else
            {
               var rect = theWindow.canvas.getBoundingClientRect();

               var rwidth  = rect.right - rect.left + 1;
               var rheight = rect.bottom - rect.top + 1;
               
               theWindow.minWidth = rwidth;
               theWindow.minHeight = rheight;

               theWindow.maxWidth = rwidth;
               theWindow.maxHeight = rheight;
            }
            
            break;
         case types.MSG_CURRENT_SELECTION:
            // current editors selection is changed
            var text = me.readStringBinaryMessage(message, 1);
            p2j.clipboard.setSelection(text);
            break;
         case types.MSG_INVALIDATE_SELECTION:
            // invalidate the current selection
            p2j.clipboard.setSelection(null);
            break;
         case types.MSG_MOVE_TO_TOP:
            var id = me.readInt32BinaryMessage(message, 1);
            p2j.screen.moveToTop(id);
            break;
         case types.MSG_MOVE_TO_BOTTOM:
            var id = me.readInt32BinaryMessage(message, 1);
            p2j.screen.moveToBottom(id);
            break;
         case types.MSG_WINDOW_SENSITIVITY:
            // change sensitivity for top-level or child window
            var id = me.readInt32BinaryMessage(message, 1);
            var enabled  = message[5] === 0 ? false : true;
            p2j.screen.setWindowEnabled(id, enabled);
            break;
         case types.MSG_REGISTER_HOVERABLE_WIDGET:
//            console.log("registerHoverableWidget");
            var offset = 1;
            var wid = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var id = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var pid = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            // the coordinates
            var x = me.readInt16BinaryMessage(message, offset);
            offset += 2;
            var y = me.readInt16BinaryMessage(message, offset);
            offset += 2;
            // the size
            var width = me.readInt16BinaryMessage(message, offset);
            offset += 2;
            var height = me.readInt16BinaryMessage(message, offset);
            offset += 2;
            // the default ptr is always pre-defined
            var styleId = me.readInt16BinaryMessage(message, offset);
            var defaultPtr = p2j.screen.mouseCursors[styleId];
            offset += 2; 
            var forceDefault = message[offset++] == 1; 
            var ptr = me.readMousePtrBinaryMessage(message, offset);
            var theWindow = p2j.screen.getWindow(wid);
            if (theWindow)
            {
               theWindow.registerHoverableWidget(id, pid, x, y, width, height, 
                                                         ptr, defaultPtr, forceDefault);
            }
            else 
            {
               console.log("unknown window: %s", wid);
            }
            break;
         case types.MSG_DEREGISTER_WIDGET:
//            console.log("deregisterWidget");
            var offset = 1;
            var wid = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var id = me.readInt32BinaryMessage(message, offset);
            var theWindow = p2j.screen.getWindow(wid);
            if (theWindow)
            {
               theWindow.deregisterMouseWidget(id, false);
            }
            else 
            {
               console.log("unknown window: %s", wid);
            }
            break;
         case types.MSG_SET_MOUSE_PTR:
//            console.log("setMousePtr");
            var offset = 1;
            var wid = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var id = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var ptr = me.readMousePtrBinaryMessage(message, offset);
            var theWindow = p2j.screen.getWindow(wid);
            if (theWindow)
            {
               theWindow.setMousePtr(id, ptr);
            }
            else 
            {
               console.log("unknown window: %s", wid);
            }
            break;
         case types.MSG_LOCK_MOUSE_PTR:
//            console.log("lockMousePointer");
            var offset = 1;
            var wid = me.readInt32BinaryMessage(message, offset);
            offset += 4;
            var ptr = me.readMousePtrBinaryMessage(message, offset);
            p2j.screen.lockMousePointer(wid, ptr);
            break;
         case types.MSG_SET_WAIT_STATE:
//            console.log("setWaitState");
            var offset = 1;
            var isWaitState = message[offset] == 1;
            p2j.screen.setWaitState(isWaitState);
            break;
         case types.MSG_IS_FONT_INSTALLED:
            // font is installed
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var fontNameLength = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var fontName  = me.readStringBinaryMessageByLength(message, offset, fontNameLength);
            
            var result = p2j.fonts.isFontInstalled(fontName);
            
            sendInt8BinaryMessage(types.MSG_IS_FONT_INSTALLED, msgId, result ? 1 : 0);
            break;
         case types.MSG_REMOVE_EXPIRED_HASH:
            // remove hashes
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            // window ID
            var windowID = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            // number of hashes
            var hashNo = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            if (windowID == 0)
            {
               for (var i = 0; i < hashNo; i++)
               {
                  // read each hash and remove it
                  var hashBytes = new Uint8Array(message.buffer.slice(offset, offset + 16));
                  var hash = me.createHexString(hashBytes);
                  delete p2j.screen.globalCache[hash];
                  
                  offset += 16;
               }
            }
            else
            {
               var theWindow = p2j.screen.getWindow(windowID);
               
               for (var i = 0; i < hashNo; i++)
               {
                  // read each hash and remove it
                  var hashBytes = new Uint8Array(message.buffer.slice(offset, offset + 16));
                  var hash = me.createHexString(hashBytes);
                  delete theWindow.removeCachedDraw(hash);
                  
                  offset += 16;
               }
            }
            
            sendInt8BinaryMessage(types.MSG_DRAW_HASH_REMOVED, msgId, 1);
            break;
         case types.MSG_PING_PONG:
            // ping/pong server answer
            if (pingPongWatcher)
            {
               clearTimeout(pingPongWatcher);
            }
            break;
         case types.MSG_PARSE_EDITOR_CONTENT:
            var offset = 1;

            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var windowId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var widgetId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var maxWidth = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var start = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var txt = me.readStringBinaryMessage(message, offset);

            var theWindow = p2j.screen.getWindow(windowId);

            var splitInfo = theWindow.parseEditorContent(widgetId, start, font, maxWidth, txt);
            
            // create the response message
            var msg = new Uint8Array(1 + 4 + 4 + splitInfo.length * (4 + 1));
            
            offset = 0;
            msg[offset] = types.MSG_PARSE_EDITOR_CONTENT;
            offset += 1;

            writeInt32BinaryMessage(msg, offset, msgId);
            offset += 4;
            
            writeInt32BinaryMessage(msg, offset, splitInfo.length);
            offset += 4;

            for (var i = 0; i < splitInfo.length; i++)
            {
               var l = splitInfo[i];
               
               var start = l.start;
               var newLine = l.isNewLine();
               
               writeInt32BinaryMessage(msg, offset, start);
               offset += 4;
               
               msg[offset] = (newLine ? 1 : 0);
               offset += 1;
            }
            
            send(msg);
            break;
         case types.MSG_GET_TEXT_WIDTHS:
            // text widths

            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var textLength = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var text  = me.readStringBinaryMessageByLength(message, offset, textLength);
            offset = offset + textLength * 2;

            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var widths = p2j.fonts.getTextWidths(font, text);
            
            // create the response message
            var msg = new Uint8Array(1 + 4 + 4 + widths.length * 2);
            
            offset = 0;
            msg[offset] = types.MSG_SENT_TEXT_WIDTHS;
            offset += 1;

            writeInt32BinaryMessage(msg, offset, msgId);
            offset += 4;
            
            writeInt32BinaryMessage(msg, offset, widths.length);
            offset += 4;

            for (var i = 0; i < widths.length; i++)
            {
               writeInt16BinaryMessage(msg, offset, widths[i]);
               offset += 2;
            }
            
            send(msg);
            break;
         case types.MSG_GET_TEXT_WIDTHS_BATCH:
            // text widths - in batch

            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var font = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var textNo = me.readInt16BinaryMessage(message, offset);
            offset = offset + 2;

            var allWidths = [];
            var sentSize = 1 + 4 + 2;
            
            for (var i = 0; i < textNo; i++)
            {
               var textLength = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;

               var text  = me.readStringBinaryMessageByLength(message, offset, textLength);
               offset = offset + textLength * 2;
               
               var widths = p2j.fonts.getTextWidths(font, text);
               
               allWidths.push(widths);
               
               sentSize = sentSize + 2 + widths.length * 2;
            }
            
            // create the response message
            var msg = new Uint8Array(sentSize);
            
            offset = 0;
            msg[offset] = types.MSG_SENT_TEXT_WIDTHS_BATCH;
            offset += 1;

            writeInt32BinaryMessage(msg, offset, msgId);
            offset += 4;
            
            writeInt16BinaryMessage(msg, offset, allWidths.length);
            offset += 2;

            for (var i = 0; i < allWidths.length; i++)
            {
               writeInt16BinaryMessage(msg, offset, allWidths[i].length);
               offset += 2;
               
               for (var j = 0; j < allWidths[i].length; j++)
               {
                  writeInt16BinaryMessage(msg, offset, allWidths[i][j]);
                  offset += 2;
               }
            }
            
            send(msg);
            break;
         case types.MSG_GET_DESKTOP_DIMENSION:
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var desktopDim = p2j.screen.getDesktopDimension();

            var sentSize = 1 + 4 + 2 + 2;

            // create the response message
            var msg = new Uint8Array(sentSize);
            
            offset = 0;
            msg[offset] = types.MSG_SENT_DESKTOP_DIMENSION;
            offset += 1;
            
            writeInt32BinaryMessage(msg, offset, msgId);
            offset += 4;

            writeInt16BinaryMessage(msg, offset, desktopDim.width);
            offset += 2;

            writeInt16BinaryMessage(msg, offset, desktopDim.height);
            offset += 2;

            send(msg);
            break;
         case types.MSG_OPEN_URL:
            var url = me.readStringBinaryMessage(message, 1);
            window.open(url, "_blank");
            break;
         case types.MSG_P2J_INVOKE:
            // a response from a remote P2J invoke
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var handleIdLen = message[offset];
            offset = offset + 1;
            var handleId = me.readStringBinaryMessageByLength(message, offset, handleIdLen);
            offset = offset + 2 * handleIdLen;

            var retValLen = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            var retVal = me.readStringBinaryMessageByLength(message, offset, retValLen);
            offset = offset + 2 * retValLen;
            
            var isError = (message[offset] == 1);
            offset = offset + 1;
            var errorNumber;
            var errorMessage; 
            if (isError)
            {
               errorNumber = me.readInt32BinaryMessage(message, offset);
               offset = offset + 4;
               
               var errMsgLen = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               
               errorMessage = me.readStringBinaryMessageByLength(message, offset, errMsgLen);
               offset = offset + 2 * errMsgLen;
            }
            var isQuit = (message[offset] == 1);
            offset = offset + 1;
            var isStop = (message[offset] == 1);
            offset = offset + 1;

            var outputArgNum = me.readInt16BinaryMessage(message, offset);
            offset = offset + 2;
            
            var outputArguments = {};
            for (var i = 0; i < outputArgNum; i++)
            {
               var keyLen = message[offset];
               offset = offset + 1;
               var key = me.readStringBinaryMessageByLength(message, offset, keyLen);
               offset = offset + 2 * keyLen;
               
               var valLen = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               var val = me.readStringBinaryMessageByLength(message, offset, valLen);
               offset = offset + 2 * valLen;
               
               outputArguments[key] = val;
            }
            
            var response = 
            {
               "type": p2j.remote.INVOKE, 
               "id"  : msgId,
               "mode": p2j.remote.RESPONSE
            };
            var payload = {};
            
            if (handleId != 0)
            {
               payload["handle"] = handleId;
            }
            payload["return"] = retVal;
            if (isError)
            {
               payload["error"] = true;
               payload["error-num"] = errorNumber;
               payload["error-msg"] = errorMessage;
            }
            if (isQuit)
            {
               payload["quit"] = true;
            }
            if (isStop)
            {
               payload["stop"] = true;
            }
            
            payload["arguments"] = outputArguments;

            response["payload"] = payload;
            
            p2j.remote.send(response);
            
            break;
         case types.MSG_P2J_DELETE_HANDLE:
            // a response from a remote handle delete
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
 
            var isError = (message[offset] == 1);
            offset = offset + 1;
            var errorNumber;
            var errorMessage; 
            if (isError)
            {
               errorNumber = me.readInt32BinaryMessage(message, offset);
               offset = offset + 4;
               
               var errMsgLen = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               
               errorMessage = me.readStringBinaryMessageByLength(message, offset, errMsgLen);
               offset = offset + 2 * errMsgLen;
            }
            
            var response = 
            {
               "type": p2j.remote.DELETE_HANDLE, 
               "id"  : msgId,
               "mode": p2j.remote.RESPONSE
            };
            var payload = {};

            if (isError)
            {
               payload["error"] = true;
               payload["error-num"] = errorNumber;
               payload["error-msg"] = errorMessage;
            }

            response["payload"] = payload;
            
            p2j.remote.send(response);
            
            break;
         case types.MSG_REMOTE_CALL:
            // a request for the remote side
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var payloadLen = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;
            
            var payload = me.readStringBinaryMessageByLength(message, offset, payloadLen);
            offset = offset + 2 * payloadLen;
            
            var request = 
            {
               "type"    : p2j.remote.REMOTE_CALL,
               "mode"    : p2j.remote.REQUEST,
               "id"      : msgId,
               "payload" : payload
            };

            p2j.remote.send(request);
            break;
         case types.MSG_CLIENT_READY:
            if (connectNum == 1)
            {
               p2j.remote.notifyPageLoaded();
            }
            else
            {
               p2j.remote.notifySocketReconnected();
            }

            break;
         case types.MSG_P2J_SUBSCRIBE:
            // a response for a SUBSCRIBE generated by the remote side
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var hlen = message[offset];
            offset = offset + 1;
            
            var hexResId = me.readStringBinaryMessageByLength(message, offset, hlen);
            offset = offset + 2 * hlen;
            
            var isError = (message[offset] == 1);
            offset = offset + 1;
            var errorNumber;
            var errorMessage; 
            if (isError)
            {
               errorNumber = me.readInt32BinaryMessage(message, offset);
               offset = offset + 4;
               
               var errMsgLen = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               
               errorMessage = me.readStringBinaryMessageByLength(message, offset, errMsgLen);
               offset = offset + 2 * errMsgLen;
            }
            
            var payload = { "subscriber-id" : hexResId };
            
            if (isError)
            {
               payload["error"] = true;
               payload["error-num"] = errorNumber;
               payload["error-msg"] = errorMessage;
            }

            var response = 
            {
               "type"    : p2j.remote.SUBSCRIBE,
               "mode"    : p2j.remote.RESPONSE,
               "id"      : msgId,
               "payload" : payload
            };

            p2j.remote.send(response);
            break;
         case types.MSG_P2J_UNSUBSCRIBE:
            // a response for an UNSUBSCRIBE generated by the remote side
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var isError = (message[offset] == 1);
            offset = offset + 1;
            var errorNumber;
            var errorMessage; 
            if (isError)
            {
               errorNumber = me.readInt32BinaryMessage(message, offset);
               offset = offset + 4;
               
               var errMsgLen = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               
               errorMessage = me.readStringBinaryMessageByLength(message, offset, errMsgLen);
               offset = offset + 2 * errMsgLen;
            }

            var payload = {};

            if (isError)
            {
               payload["error"] = true;
               payload["error-num"] = errorNumber;
               payload["error-msg"] = errorMessage;
            }

            var response = 
            {
               "type"    : p2j.remote.UNSUBSCRIBE,
               "mode"    : p2j.remote.RESPONSE,
               "id"      : msgId,
               "payload" : payload
            };

            p2j.remote.send(response);
            break;
         case types.MSG_P2J_PUBLISH:
            // a request for a PUBLISH event generated by the business logic
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            // resource ID
            var resLen = message[offset];
            offset = offset + 1;
            var hexResId = me.readStringBinaryMessageByLength(message, offset, resLen);
            offset = offset + 2 * resLen;

            // event
            var evLen = message[offset];
            offset = offset + 1;
            var event = me.readStringBinaryMessageByLength(message, offset, evLen);
            offset = offset + 2 * evLen;

            // param length
            var pnum = message[offset];
            offset = offset + 1;
            
            var params = [];
            for (var i = 0; i < pnum; i++)
            {
               var valLen = me.readInt16BinaryMessage(message, offset);
               offset = offset + 2;
               var val = me.readStringBinaryMessageByLength(message, offset, valLen);
               offset = offset + 2 * valLen;
               
               params[i] = val;
            }

            var req = {};
            req["event"] = event;
            req["subscriber-id"] = hexResId;
            req["parameters"] = params;

            var request = 
            {
               "type"    : p2j.remote.PUBLISH,
               "mode"    : p2j.remote.REQUEST,
               "id"      : msgId,
               "payload" : req
            };

            p2j.remote.send(request);
            break;
         case types.MSG_P2J_CREATE_RESOURCE:
            // a response for a resource creation request generated by the remote side
            var offset = 1;
            
            var msgId = me.readInt32BinaryMessage(message, offset);
            offset = offset + 4;

            var hlen = message[offset];
            offset = offset + 1;
            
            var hexResId = me.readStringBinaryMessageByLength(message, offset, hlen);
            offset = offset + 2 * hlen;

            var payload = { "resource-id" : hexResId };
            
            var response = 
            {
               "type"    : p2j.remote.CREATE_RESOURCE,
               "mode"    : p2j.remote.RESPONSE,
               "id"      : msgId,
               "payload" : payload
            };

            p2j.remote.send(response);
            break;
      };
      var t2 = (new Date()).getTime();
      console.log(callNo + ": " + message[0].toString(16) + " done in " + (t2 - t1));
   };

   /**
    * It is responsible for executing periodically with the given period the provided callback
    * function if the elapsed time exceeds its period. The elapsed time is calculated using
    * the current time and the last checkpoint time.
    * 
    * @param    {Number} maxLifeTime
    *           The maximal life time of the internal timer measured in milliseconds.
    * @param    {Number} period
    *           The period of the internal timer.
    * @param    {Function} callback
    *           The timer callback function or closure.
    * @param    {Function} completed
    *           The timer cleanup function or closure.
    */
   function ControlTimer(maxLifeTime, period, callback, completed)
   {
      /** The time of the last check */
      var checkpoint;
      
      /**
       * If the elapsed time exceeds its period, then it invokes 
       * the callback function. If the elapsed time exceeds the maximal life time,
       * then the timer is stopped.
       */
      function check()
      {
         var test = (new Date()).getTime();
         var elapsed = test - checkpoint;
         if (elapsed > maxLifeTime)
         {
            kill();
            if (completed)
            {
               completed();
            }
         }
         else if (elapsed > period)
         {
            callback();
         }
      }
      
      /** The timer id */
      var timer;
      
      /** Indicates the timer is running */
      var running = false;
      
      /**
       * Starts a new timer that checks periodicaly with the given rate if the elapsed time
       * exceeds its period and sets the last checkpoint to the current time if it isn't running,
       * otherwise it does nothing.
       */
      function start()
      {
         if (running)
         {
            return;
         }
         reset((new Date()).getTime());
         timer = setInterval(check, period);
         running = true;
      }
      
      this.start = start;
      
      /**
       * Resets the last check time.
       * 
       * @param    {Number} lastActivity
       *           The last activity time in milliseconds.
       */
      function reset(lastActivity)
      {
         checkpoint = lastActivity;
      }
      
      this.reset = reset;
      
      /**
       * Stops and disposes the internal timer. It can be started again.
       */
      function kill()
      {
         if (timer)
         {
            clearInterval(timer);
         }
         running = false;
      }
      
      this.kill = kill;
      
      /**
       * Tests if the timer is running.
       * 
       * @return   True if the timer is running, otherwise false.
       */
      this.isRunning = function()
      {
         return running;
      }
   }
   
   /**
    * Tries to open a new websocket connection to the server.
    * 
    * @param    {String} url
    *           URL connection string
    */
   function createWebSocket(url)
   {
      // check WebSocket support
      if (!('WebSocket' in window))
      {
         p2j.screen.error('WebSockets are NOT supported by your Browser.');
         return;
      }
      
      ws = new WebSocket(url);
      
      // on web socket open
      ws.onopen = function()
      {
         connected = true;
         ws.binaryType = 'arraybuffer';
         
         if (isRequiredToRedirect())
         {
            // reload use case
            doRedirectToLoginPage();
            sendNotification(types.MSG_QUIT);
         }
         else
         {
            // new session or reopen use case
            
            exitTheApplication = false;
            
            connectivityTimer.kill();
            idleTimer.start();
            
            // notify the web socket has opened and the page is loaded
            sendNotification(types.MSG_PAGE_LOADED);
            
            connectNum = connectNum + 1;
         }
      };

      // on web socket message
      ws.onmessage = function(evt)
      {
         var data = evt.data;
         
         if (data instanceof ArrayBuffer)
         {
            // binary messages
            var message = new Uint8Array(data);
            var isDrawing = message[0] === types.MSG_DRAW;
            if (isDrawing)
            {
               p2j.displayWaitCursor();
            }
            setTimeout(function() { messageHandler(message); }, 0);
            if (isDrawing)
            {
               setTimeout(function() { p2j.restoreCursor(); }, 0);
            }
         }
         else
         {
            // text messages
            var pay = JSON.parse(data);
            setTimeout(
                  function()
                  {
                     switch (pay.c)
                     {
                        case 0:
                           // color palette
                           p2j.screen.palette = pay.p;
                           break;
                     };
                  }, 0);
         }
      }

      /**
       * Web socket close events handler
       * 
       * @param    {CloseEvent} evt
       *           The close websocket event
       */
      ws.onclose = function(evt)
      { 
         if (connected)
         {
            p2j.remote.notifySocketDisconnected();
         }
         
         connected = false;
         p2j.screen.error("Connection closed with the returned code " + evt.code); 
         idleTimer.kill();
         // check the normal close reason
         if (evt.code === 1000)
         {
            doRedirectToLoginPage();
         }
         else
         {
            connectivityTimer.start();
         }
      };
      
      /**
       * Web socket errors events handler.
       * 
       * @param    {Event} evt
       *           The websocket error event
       */
      ws.onerror = function(evt)
      { 
         p2j.screen.error("Web socket error " + evt.detail); 

         if (connected)
         {
            p2j.remote.notifySocketDisconnected();
         }

         connected = false;
         idleTimer.kill();
         
         if (exitTheApplication)
         {
            if (!connectivityTimer.isRunning())
            {
               // the reconnect tries are failed the websocket server is down due to the watchdog.
               doRedirectToLoginPage();
            }
         }
         else
         {
            if (!connectivityTimer.isRunning())
            {
               // try to reconnect within watchdog timeout if it is failed, then
               // the websocket server is down
               connectivityTimer.start();
            }
            exitTheApplication = true;
         }
      };
   }
   
   /**
    * Read mouse pointer definition
    * 
    * @param    {byte[]} message
    *           Message bytes.
    * @param    {int} offset
    *           The location in the message at which the characters start.
    *
    * @return   the mouse pointer definition.
    */
   me.readMousePtrBinaryMessage = function(message, offset)
   {
      var ct = message[offset++]; 
      if (ct == 0) 
      {
         return null;
      }
      if (ct == 1) 
      {
         var styleId = this.readInt16BinaryMessage(message, offset);
         var style = p2j.screen.mouseCursors[styleId];
         return style ? style : "default";
      }
      var mlen = this.readInt16BinaryMessage(message, offset);
      offset += 2;
      var mime = this.readStringBinaryMessageByLength(message, offset, mlen);     
      offset += 2 * mlen;
      var dlen = this.readInt16BinaryMessage(message, offset);
      offset += 2;
      var b64 =  btoa(String.fromCharCode.apply(null, 
            new Uint8Array(message.subarray(offset, offset + dlen))));
      return 'url("data:' + mime + ';base64,' + b64 + '"), auto';
   }

   /**
    * Force the web page to be redirected to the login page given by the referrer url.
    */
   function doRedirectToLoginPage()
   {
      exitTheApplication = true;
      if (p2j.embedded)
      {
         p2j.remote.notifyPageClosed();
      }
      else
      {
         setTimeout(function() { window.location.replace(referrer); }, 1000);
      }
   }
   
   /**
    * Checks if it is GUI mode and the current page is reloaded or reopened/new. If the GUI page
    * is reloaded, then it returns true.
    * 
    * @return   True iff the GUI page is reloaded, otherwise false.
    */
   function isRequiredToRedirect()
   {
      if (!p2j.isGui)
      {
         return false;
      }
      
      var exit = p2j.restoreObject("exitTheApplication", false);
      var date = (new Date()).getTime();
      
      // reload use case
      if (exit && (date < exit.date + 1000))
      {
         // reload use case
         return true;
      }
      
      if (exit)
      {
         // reopen use case
         p2j.deleteObject("exitTheApplication", false);
      }
      
      return false;
   }
   
   /**
    * Initialize module.
    * 
    * @param {object} cfg  configuration.
    */
   me.init = function(cfg)
   {
      referrer = cfg.referrer;
      
      maxBinaryMessage = cfg.socket.maxBinaryMessage;
      
      maxIdleTime      = cfg.socket.maxIdleTime;
      
      watchdogTimeout  = cfg.socket.watchdogTimeout;
      
      socketUrl        = cfg.socket.url;
      
      /**
       * Sends ping and pong notifications to the server and starts the ping/pong watcher timer.
       */
      function ping()
      {
         //send Ping/Pong message
         sendNotification(types.MSG_PING_PONG);
         if (pingPongWatcher)
         {
            clearTimeout(pingPongWatcher);
         }
         // start ping/pong watcher
         pingPongWatcher = setTimeout(
               function()
               {
                  // 5000 ms has been elapsed, but the server hasn't replied yet on the ping sent
                  // by the client. If ws is in OPEN state, then send new ping/pong
                  if (ws && ws.readyState == 1)
                  {
                     ping();
                  }
               }, 5000);
         console.debug("Ping/Pong " + (new Date()).getTime());
      };
      
      idleTimer = new ControlTimer(maxIdleTime, maxIdleTime / 3, ping);
      
      /**
       * Connect to the websocket server via the given connection url.
       */
      function connect()
      {
         if (pingPongWatcher)
         {
            clearTimeout(pingPongWatcher);
         }
         // don't reconnect if the websocket state is CONNECTING(0) or OPEN(1) or CLOSING(2)
         if (ws && ws.readyState !== 3)
         {
            console.debug("Connection in process " + (new Date()).getTime());
            return;
         }
         createWebSocket(socketUrl);
         console.debug("Try to connect the server " + (new Date()).getTime());
      };
      
      connectivityTimer = new ControlTimer(Math.max(watchdogTimeout, maxIdleTime), 5000, connect,
            doRedirectToLoginPage);
      
      /**
       * If the user is navigating away from the application webpage.
       * 
       * @param    {PageTransitionEvent} evt
       *           
       */
      window.onpagehide = function(evt) 
      {
         if (!p2j.isGui)
         {
            return;
         }
         if (!exitTheApplication)
         {
            // close or reload cases
            exitTheApplication = true;
            p2j.saveObject(
                  "exitTheApplication",
                  {
                     exitTheApplication : exitTheApplication,
                     date: (new Date()).getTime()
                  },
                  false);
         }
         else
         {
            // after reload
            p2j.deleteObject("exitTheApplication", false);
         }
      };
      
      /**
       * Occurs when the web application page are about to be unloaded.
       * 
       * @param    {Event} evt
       *           The beforeunload event.
       * 
       * @return   {String}
       *           The returned value will be used as a custom message for the default
       *           "Stay on Page or Leave Page" dialog.
       */
      window.onbeforeunload = function(evt) 
      {
         if (p2j.isGui && !exitTheApplication)
         {
            var confirmationMessage = "ANY PENDING CHANGES WILL BE LOST!\n" +
                                          "Please confirm that it is OK to exit the application.";
            
            // Gecko + IE
            (evt || window.event).returnValue = confirmationMessage;
            
            return confirmationMessage;
         }
      };
      
      createWebSocket(socketUrl);
   };

   /**
    * Parse a hex string and return its byte array.
    *
    * @param    {String} str
    *           The hex-encoded string.
    *
    * @return   {Uint8Array} byte-array for the given string.
    */
   me.hexStringToBytes = function(str)
   {
      var bytes = new Uint8Array(str.length / 2);

      var idx = 0;
      for (var c = 0, idx = 0; c < str.length; c += 2, idx++)
      { 
         bytes[idx] = parseInt(str.substr(c, 2), 16);
      }
      
      return bytes;
    }
   
   /**
    * Create a hex string from the given byte array.
    *
    * @param    {Uint8Array}  arr
    *           The byte array.
    *
    * @return   Hex representation of the given byte array.
    */
   me.createHexString = function(arr)
   {
      var result = "";
      for (var i in arr)
      {
         var str = arr[i].toString(16);
         str = str.length == 0 ? "00" :
               str.length == 1 ? "0" + str : 
               str.length == 2 ? str :
               str.substring(str.length - 2, str.length);
               
         result += str;
      }
      
      return result;
   }
   
   /**
    * Remove the sub-array from the given array and save it in the splice parameter.
    *
    * @param    {array}  arr
    *           The byte array from where elements are removed.
    * @param    {int} starting
    *           The starting position (inclusive).
    * @param    {int} deleteCount
    *           The number of elements to delete.
    * @param    {array} splice
    *           The array where to save the elements (must have enough space).
    *
    * @return   The original array with the elements removed.
    */
   function splice(arr, starting, deleteCount, splice)
   {
      if (arguments.length === 1)
      {
         return arr;
      }
      
      starting = Math.max(starting, 0);
      deleteCount = Math.max(deleteCount, 0);

      var newSize = arr.length - deleteCount;
      var splicedArray = new arr.constructor(newSize);

      splicedArray.set(arr.subarray(0, starting));
      splicedArray.set(arr.subarray(starting + deleteCount), starting);
      
      splice.set(arr.subarray(starting, starting + deleteCount));
      
      return splicedArray;
   };

   return me;
}());