/*
** Module   : EventQueue.java
** Abstract : Event queue. 
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 CA  20130827 Added API to check if the event queue is empty.   
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

import java.util.*;

/**
 * Simple event queue which is able to compress (ignore) empty and identical
 * painting events.
 */
public class EventQueue
{
   /** Backing storage for events. */
   private LinkedList<Event> queue = new LinkedList<Event>();
   
   /**
    * Post specified event into queue.
    * 
    * @param   evt
    *          Event to post.
    */
   public synchronized void post(Event evt)
   {
      if (evt instanceof PaintEvent && !queue.isEmpty())
      {
         if (queue.indexOf(evt) >= 0)
         {
            return;
         }
      }
      queue.add(evt);
   }

   /**
    * Retrieve event from the current active input queue.
    * 
    * @return  event from queue or <code>null</code> if no events are available.
    */
   public synchronized Event get()
   {
      if (queue.isEmpty())
         return null;
      return queue.removeFirst();
   }

   /**
    * Check if the event queue is empty.
    * 
    * @return   <code>true</code> if the event queue is empty or <code>false</code> otherwise.
    */
   public synchronized boolean isEmpty()
   {
      return queue.isEmpty();
   }
   
   /**
    * Install new event queue.
    * 
    * @param   other
    *          New event queue to install.
    */
   public void insert(EventQueue other)
   {
      if (other == null || other.queue.isEmpty())
         return;
      
      while(!queue.isEmpty())
         other.post(queue.removeFirst());
      
      queue = other.queue;
      other.queue = new LinkedList<Event>();
   }
}