/*
** Module   : KeyInput.java
** Abstract : Extended version of keyboard input event.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 NVS 20060301   @24810 Created initial version.
** 002 NVS 20060613   @27170 Instances constructed from high level events now have the key
**                           component set to -1.
** 003 NVS 20060613   @27175 Previous fix regressed readKey() method in ThinClient. Fixed.
** 004 NVS 20080502   @38200 Added eventMode variable and isSpecial() getter.
** 005 SIY 20080901   @39640 Added flag for 'real keys' (i.e. keys typed by user).
** 006 CA  20080924   @39933 Fixed H005 regression eventMode must be passed to the c'tor.
** 007 SVL 20090224   @41356 During construction keep the original key code for real keys.
** 008 SIY 20100619          Switching to own implementation of character UI.
** 009 SIY 20100804          Changes in imports, some refactorings.
** 010 SIY 20101122          Moved from CHUI package, merged necessary code from KeyEvent.
** 011 SIY 20110112          Refactoring and simplification.
** 012 GES 20150226          Added flag to note that this event caused AUTO-RETURN.
** 013 EVL 20150401          Adding new constructor to be able for MouseEvt to extend this class.
** 014 IAS 20150801          WIDGET-LEAVE/WIDGER-ENTER support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
* An event which indicates that a keystroke occurred in an object. This keystroke is represented
* in two ways: as the usual KeyEvent and as a Progress-like logical action code.
*/
public class KeyInput
extends Event
{
   /** Progress like action code or KeyInput.CHAR_UNDEFINED */
   private int actionCode = CHAR_UNDEFINED;

   /** A flag which may have varying meaning depending on the event code.
    *  For now, ENTRY event uses this flag to denote an unrelated event, so
    *  that there will be no matching LEAVE event generated.
    */
   private boolean eventMode = false;
   
   /** Mark real keys typed by user. */
   private boolean isKey = false;

   /** If <code>true</code>, the keyboard event was processed and consumed. */
   private boolean consumed = false;

   /** Key associated with the event. */
   private int key;
   
   /** <code>true</code> if this event generated a downstream AUTO-RETURN. */
   private boolean causedAutoReturn = false;

   /** Undefined character constant. */
   public static final char CHAR_UNDEFINED = (char) -1;

   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code.
    * @param   source
    *          widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    */
   public KeyInput(Widget source, int id, int key) 
   {
      this(source, id, key, false);
   }

   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code.
    * @param   source
    *          widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    * @param   eventMode
    *          a boolean flag specific to the event
    */
   public KeyInput(Widget source, int id, int key, boolean eventMode)
   {
      this(source, id, key, eventMode, false);
   }
   

   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code.
    * @param   source
    *          widget associated with the event
    * @param   other
    *          companion widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    * @param   eventMode
    *          a boolean flag specific to the event
    */
   public KeyInput(Widget source, Widget other, int id, int key, boolean eventMode)
   {
      this(source, other, id, key, eventMode, false);
   }

   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code.
    * @param   src
    *          widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    * @param   mode
    *          a boolean flag specific to the event
    * @param   isKey
    *          mark real key typed by user
    */
   public KeyInput(Widget src, int id, int key, boolean mode, boolean isKey) 
   {
      this(src, null, id, key, mode, isKey);
   }

   /**
    * Initializes the action code to the Mapped Progress key action. If it is
    * a valid action code, the original keystroke is translated into a
    * standard, compatible code.
    * @param   src
    *          widget associated with the event
    * @param   other
    *          companion widget associated with the event
    * @param   id
    *          event category
    * @param   key
    *          event code
    * @param   mode
    *          a boolean flag specific to the event
    * @param   isKey
    *          mark real key typed by user
    */
   public KeyInput(Widget src, Widget other, int id, int key, boolean mode, boolean isKey) 
   {
      super(src, other, id);
      
      this.key = key;
      
      this.eventMode = mode;
      this.isKey     = isKey;

      actionCode = Keyboard.keyAction(key);

      if (actionCode == CHAR_UNDEFINED)
      {
         actionCode = key;
      }
      else if (actionCode < CHAR_UNDEFINED)
      {
         // a valid key action is defined; keep the original key code if we
         // have a real key, or try to "guess" a hypothetical key code that
         // would correspond the high level event generated by APPLY
         // statement.
         int standardCode = isKey ? key :
                                    Keyboard.keyCodeForAction(actionCode);
         if (standardCode != CHAR_UNDEFINED)
            setKeyCode(standardCode);
         
         // do not keep key code as action code
         if (key == actionCode && actionCode < 0)
            setKeyCode(CHAR_UNDEFINED);
      }
   }

   /**
    * Constructor for MouseEvt compatibility.
    * 
    * @param   source
    *          Event source.
    * @param   id
    *          Event type ID.
    */
   public KeyInput(Widget source, int id)
   {
      super(source, id);
   }

   /**
    * Returns the integer action code.
    *
    * @return action code
    */
   public int actionCode()
   {
      return actionCode;
   }

   /**
    * Sets the action code.
    *
    * @param code
    *        new action code to be set
    */
   public void setActionCode(int code)
   {
      actionCode = code;
   }

   /**
    * Returns the special mode flag.
    *
    * @return event mode
    */
   public boolean isSpecial()
   {
      return eventMode;
   }

   /**
    * Returns the String representation of this object.
    *
    * @return text representation of this object
    */
   @Override
   public String toString() 
   {
      return ("KeyInput: key=" + keyCode() + " action=" + actionCode +
              " mode=" + (eventMode ? "special" : "regular") + 
              " source=[" + source() + "]");
   }

   /**
    * Return <code>true</code> if event represents key typed by user (as 
    * opposed to some kind of artificially generated event).
    * 
    * @return  <code>true</code> if event was generated by real user input.
    */
   public boolean isRealKey()
   {
      return isKey;
   }

   /**
    * Invoke appropriate listeners' method depending on the event type.
    * 
    * @param   listener
    *          Event listener to which event will be passed.
    */
   public void dispatch(KeyListener listener)
   {
      if (id() == EventType.KEY_PRESSED)
         listener.onKeyPressed(this);
      else if (id() == EventType.KEY_TYPED)
         listener.onKeyTyped(this);
   }

   /**
    * Mark event as consumed.
    */
   public void consume()
   {
      consumed = true;
   }

   /**
    * Get character for this event. If event is generated for non-character 
    * key event, then {@link KeyInput#CHAR_UNDEFINED} is returned.
    * 
    * @return  character.
    */
   public char keyChar()
   {
      return (id() == EventType.KEY_PRESSED) ? CHAR_UNDEFINED : (char) key;
   }

   /**
    * Get key code for this event.
    * 
    * @return  key code.
    */
   public int keyCode()
   {
      return key;
   }

   /**
    * Check if event is already consumed.
    * 
    * @return  <code>true</code> if event is consumed.
    */
   public boolean isConsumed()
   {
      return consumed;
   }

   /**
    * Set key code for this event.
    * 
    * @param   key
    *          Key code to set.
    */
   public void setKeyCode(int key)
   {
      this.key = key;
   }
   
   /**
    * Sets the flag reporting if this event generated an AUTO-RETURN.
    *
    * @param    causedAutoReturn
    *           <code>true</code> if this event generated a downstream AUTO-RETURN.
    */
   public void setCausedAutoReturn(boolean causedAutoReturn)
   {
      this.causedAutoReturn = causedAutoReturn;
   }
   
   /**
    * Reports if this event generated an AUTO-RETURN.
    *
    * @return   <code>true</code> if this event generated a downstream AUTO-RETURN.
    */
   public boolean isCausedAutoReturn()
   {
      return causedAutoReturn;
   }
}