/*
** Module   : WindowActivated.java
** Abstract : Pseudo-event when the simulated OS window gains focus. 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141113 Created initial version.
** 002 CA  20141123 Added new flags to allow focus gained/lost to P2J or OS/native window.
** 003 HC  20150612 Implemented support for modal windows.
** 004 HC  20160104 Implemented toString().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * Pseudo-event generated when the simulated OS window gains focus.
 */
public class WindowActivated
extends Event
{
   /** The window being activated. */
   private boolean activated;

   /** Flag indicating if the opposite window is a OS window. */
   private boolean withNativeWindow;
   
   /**
    * Create a new event.
    * 
    * @param   source
    *          The top-level window being activated.
    * @param   activated
    *          Flag indicating if the window is being activated (gained focus) or deactivated 
    *          (lost focus).
    * @param   withNativeWindow
    *          Flag indicating if the opposite window is a OS window.
    */
   public WindowActivated(TopLevelWindow<?> source, boolean activated, boolean withNativeWindow)
   {
      super(source, -1);
      
      this.activated = activated;
      this.withNativeWindow = withNativeWindow;
   }

   /**
    * Get the window source.
    * 
    * @return   See above.
    */
   @Override
   public TopLevelWindow<?> source()
   {
      return (TopLevelWindow<?>) super.source();
   }
   
   /**
    * Check if the window is being activated (gained focus) or deactivated (lost focus).
    * 
    * @return   The state of the {@link #activated} flag.
    */
   public boolean isActivated()
   {
      return activated;
   }
   
   /**
    * Check if the opposite window for this event is another OS/native window.
    * 
    * @return   The state of the {@link #withNativeWindow} flag.
    */
   public boolean withNativeWindow()
   {
      return withNativeWindow;
   }

   /**
    * Returns a string representation of the object.
    * 
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      return "WindowActivated [source=" + Utils.nonNull(source(), "<null>") + ", " + 
                              "activated=" + activated + ", " + 
                              "withNativeWindow=" + withNativeWindow +
                              "]";
   }
}