/*
** Module   : DisplayFormatFactory.java
** Abstract : factory for display formats.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description---------------------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SIY 20050901   @22535 Fixed formatting.
** 003 GES 20060711   @27975 Cleaned up code to use a common backing method.
** 004 SIY 20090701   @43016 Added format checking method which returns  detailed error
**                           information instead of throwing condition or displaying error
**                           immediately.
** 005 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 006 OM  20130329          Added datetime/datetime-tz and handle formats.
** 007 CA  20150129          Changes related to format validation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.format;

import com.goldencode.p2j.ui.client.format.DisplayFormat.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * Factory for display formats. Because a display format is immutable it is
 * possible to make only one instance per format string value.
 */
public class DisplayFormatFactory
{
   /**
    * Dummy constructor to disable creation of instances of this class.
    */
   private DisplayFormatFactory() {}

   /**
    * Returns the default display format instance accordingly to type of the
    * given variable.
    * 
    * @param    var
    *           The variable to be represented.
    * 
    * @return   The default format for this variable type.
    * 
    * @throws   DisplayFormatParsingException
    *           A format instantiation error or the variable type is not
    *           supported.
    */
   public static DisplayFormat makeDefaultFormat(BaseDataType var)
   throws DisplayFormatParsingException
   {
      return makeFormat(var, var.defaultFormatString());
   }

   /**
    * Returns the display format instance according to the type of the given
    * variable and the format string.
    * 
    * @param    var
    *           The variable to be represented.
    * @param    pformat
    *           Valid format string.
    * 
    * @return   The format for this variable type configured with the given
    *           format string.
    * 
    * @throws   DisplayFormatParsingException
    *           A format instantiation error or the variable type is not
    *           supported.
    */
   public static DisplayFormat makeFormat(BaseDataType var, String pformat)
   throws DisplayFormatParsingException
   {
      DisplayFormat format;
      
      if (var instanceof character)
      {
         format = new StringFormat(pformat);
      }
      else if (var instanceof NumberType)
      {
         format = new NumberFormat(pformat);
      }
      else if (var instanceof datetimetz)
      {
         format = new DatetimeFormat(pformat, true);
      }
      else if (var instanceof datetime)
      {
         format = new DatetimeFormat(pformat, false);
      }
      else if (var instanceof date)
      {
         format = new DateFormat(pformat);
      }
      else if (var instanceof logical)
      {
         format = new BoolFormat(pformat);
      }
      else
      {
         String errmsg = "Unknown display format for variable's type: ";
         throw new DisplayFormatParsingException(errmsg + var.getClass().getName());
      }
      
      return format;
   }
   
   /**
    * Check validness of given value from the point of view of format.
    * 
    * @param    value
    *           Value to check.
    * @param    format
    *           Format to check.
    *           
    * @return   error code and message stored in common container.
    */
   public static ErrorEntry checkFormat(BaseDataType value, String format)
   {
      DisplayFormat fmt = null;
      try
      {
         if (format != null)
            fmt = DisplayFormatFactory.makeFormat(value, format);

      }
      catch (DisplayFormatParsingException e)
      {
         return new ErrorEntry(-1, "Unable to parse format \"" + format + "\"");
      }
      
      if (fmt != null)
      {
         Presentation pres = null;
         
         try
         {
            pres = fmt.fromVar(value);
         }
         catch (DisplayFormatCheckException e)
         {
            return new ErrorEntry(-1, "Unable to check format \"" + format + "\"" +
                                  " for value " + value.toStringMessage());
         }
         
         if (pres != null)
         {
            pres.finishEdit();
            
            ErrorEntry[] res = pres.validateFormat();
            
            return (res == null || res.length == 0 ? null : res[0]);
         }
      }
      
      return null;
   }
}