/*
** Module   : BorderedPanelImpl.java
** Abstract : GUI implementation of bordered panel.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20150831 Created initial version
** 002 HC  20150917 Fixed background drawing. Fixed clipping and location translation during draw. 
**                  When clearing the background, include the border, too. 
** 003 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 004 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 005 VIG 20160209 Added isRMBEntrySupported() method.
** 006 EVL 20160419 Adding method to get additional 3D style for border.  Used in some modes of
**                  the editor widget with scrollbars.
** 007 HC  20160428 GUI window border drawing changed to use the Border interface mechanism.
** 008 HC  20160429 Made the border ID to be selectively assigned. The previous change made the
**                  ID to be assigned to all instances, which caused regressions. The eventing
**                  logic may select widgets with assigned ID as events targets.   
** 009 OM  20170320 Replaced anonymous classes with lambdas.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * GUI implementation of bordered panel.
 */
public class BorderedPanelGuiImpl 
extends BorderedPanel<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver<?, ?> gd;
   
   /** Color resolver. */
   private GuiColorResolver gc;
   
   /** Custom draw runnable. */
   private Runnable customDraw;
   
   /** Additional 3D styled border flag.*/
   private boolean is3DBorder = false; 
   
   /** ID is needed in case panel is registered as resizable for example. */
   private WidgetId id;
   
   /**
    * Creates new instance with no border and no layout manager.
    */
   public BorderedPanelGuiImpl()
   {
      init();
   }

   /**
    * Creates new instance with the specified border.
    * 
    * @param   border
    *          reference to the border implementation, 
    *          may be <code>null</code>
    */
   public BorderedPanelGuiImpl(Border<GuiOutputManager> border)
   {
      super(border);
      init();
   }

   /**
    * Creates new instance with the specified layout manager and border. 
    * 
    * @param   mgr
    *          reference to the layout manager, may be <code>null</code>
    * @param   border
    *          reference to the border implementation, 
    *          may be <code>null</code>
    */
   public BorderedPanelGuiImpl(LayoutManager<GuiOutputManager> mgr, 
                               Border<GuiOutputManager> border)
   {
      super(mgr, border);
      init();
   }
   
   /**
    * Sets a color resolver. The class utilizes {@link GuiColorResolver#bgColor}
    * which is used as the background color for this panel. If no resolver
    * is given, the color {@link SysColor#COLOR_WINDOW} will be used by default.
    * 
    * @param   resolver
    *          color revolver, may be <code>null</code>
    */
   public void setColorResolver(GuiColorResolver resolver)
   {
      if (resolver == null)
      {
         resolver = new GuiColorResolver(gd, new BaseConfig(), 
                                         SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      }
      
      gc = resolver;
   }
   
   /**
    * Sets custom draw routine. By default, when no custom draw routine
    * is registered, the {@link #draw()} method forwards the call to the
    * parent's draw method. When custom draw routine is registered,
    * it is executed instead of the parent.
    * 
    * @param   customDraw
    *          custom draw routine, see above 
    */
   public void setCustomDraw(Runnable customDraw)
   {
      this.customDraw = customDraw;
   }
   
   /**
    * Check if this widget gains ENTRY event when right/middle mouse button is clicked.
    *
    * @return  <code>true</code> if this widget gains ENTRY event
    *          when right/middle mouse button is clicked.
    */
   public boolean isRMBEntrySupported()
   {
      return false;
   }
   
   /**
    * Checks if this widget need to draw additional 3D style border elements.  Used with editor
    * widget and scrollbars.
    *
    * @return  <code>true</code> if additional painting is required, <code>false</code> otherwise.
    */
   public boolean is3DBorder()
   {
      return is3DBorder;
   }
   
   /**
    * Sets the new value for 3D style flag.  Used with editor widget and scrollbars.
    * 
    * @param   is3DBorder
    *          <code>true</code> if additional painting is required, <code>false</code> otherwise.
    */
   public void set3DBorder(boolean is3DBorder)
   {
      this.is3DBorder = is3DBorder;
   }
   
   /** 
    * Draw widget.
    */
   @Override
   public void draw()
   {
      if (!isVisible())
      {
         return;
      }
      
      final NativeDimension d = physicalDimension();
      NativeRectangle r = new NativeRectangle(0, 0, d.height - 1, d.width - 1);
      
      // main drawing bracket, includes potential border and content
      gd.draw(physicalLocation(), r, () ->
      {
         gc.colors();

         if (!isTransparent())
         {
            gd.setColor(gc.bgColor);
            
            // clear the panel area
            int windowId = BorderedPanelGuiImpl.this.topLevelWindow().getId().asInt();
            NativePoint screenOrigin = BorderedPanelGuiImpl.this.screenPhysicalLocation();
            NativeRectangle nrect = new NativeRectangle(screenOrigin, d);
            
            List<NativeRectangle> clippings = screen().getClippings(windowId, nrect);
            
            for (NativeRectangle c : clippings)
            {
               gd.fillRect(c.left() - screenOrigin.x, 
                           c.top() - screenOrigin.y, 
                           c.width(), 
                           c.height());
            }
         }
         
         // draw the border
         if (border != null)
         {
            border.drawBorder(BorderedPanelGuiImpl.this, dimension());
         }
         
         // draw the content
         Insets insets = insets();
         
         if (insets == null)
         {
            insets = Insets.ZERO_INSETS;
         }
         
         CoordinatesConversion c = screen().coordinates();
         NativeInsets ninsets = c.insetsToNative(insets);

         NativePoint p = new NativePoint(ninsets.left, ninsets.top);
         
         int widthDelta = ninsets.left + ninsets.right;
         int heightDelta = ninsets.top + ninsets.bottom;
         NativeRectangle r1 = new NativeRectangle(0, 0, 
                                                 d.height - 1 - heightDelta, 
                                                 d.width - 1 - widthDelta);

         // content drawing bracket
         gd.draw(p, r1, () ->
         {
            if (customDraw != null)
            {
               customDraw.run();
            }
            else
            {
               BorderedPanelGuiImpl.super.draw();
            }
         });
      });
   }
   
   /**
    * Sets widget ID for this widget.
    * <p>
    * The setter allows to selectively assign an ID to borders that do need it.
    * For example a mouse-sensitive border around main window requires an ID
    * to be able to handle mouse events. 
    * 
    * @param   id
    *          A widget id.
    */
   public void setId(WidgetId id)
   {
      this.id = id;
   }
   
   /**
    * Widget ID.
    * 
    * @return  see above.
    */
   @Override
   public WidgetId getId()
   {
      return id;
   }

   /**
    * Instance initializer.
    */
   private void init()
   {
      gd = (GuiDriver<?, ?>) OutputManager.getDriver();
      setColorResolver(null);
   }
}