/*
** Module   : BrowseColumnGuiImpl.java
** Abstract : GUI implementation of browse column.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 SVL 20151013 Initial version.
** 002 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 003 SVL 20151019 Layout procedure refactoring, misc fixes.
** 004 SVL 20151110 Support for multi-line label. Added setWidth.
** 005 SVL 20151126 Fixed label drawing order for multi-line labels.
** 006 SVL 20151126 Use browse.isLabelsVisible.
** 007 SVL 20151206 Update column width if its WIDTH attribute has been updated.
** 008 SVL 20151230 Coloring fix, support for editable browse.
** 009 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 010 IAS 20160229 LOAD-MOUSE-POINTER support.
** 011 SVL 20160407 Support for column dragging.
** 012 SVL 20160412 Updated clipping rectangles. Implemented column resizing.
** 013 SVL 20160719 Updated translate() signature.
** 014 SVL 20160722 Reflect signature change in browse.rowClicked.
** 015 HC  20160716 Implemented runtime support for BROWSE-COLUMN:RESIZABLE attribute.
** 016 EVL 20160921 Fix for column header vertical text position mismatch.
** 017 SBI 20160921 Fixed casual NPE.
** 018 VIG 20160927 Added label update to afterConfigUpdate.
** 019 SVL 20170126 Added support for pressed columns, START-SEARCH/END-SEARCH events and
**                  SORT-ASCENDING/SORT-NUMBER attributes.
** 020 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui;

import static com.goldencode.p2j.ui.client.gui.BrowseGuiImpl.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.awt.event.*;

/**
 * GUI implementation of browse column.
 */
public class BrowseColumnGuiImpl
extends BrowseColumn
implements Hoverable
{
   /** Left offset for text area in a label. */
   private final static int LABEL_OFFSET_LEFT = 2;

   /** Right offset for text area in a label. */
   private final static int LABEL_OFFSET_RIGHT = 2;

   /** Offset of the label text if the column is in pressed state. */
   private final static int LABEL_PRESSED_OFFSET = 1;

   /** Thickness of the shadow inside the column header. */
   public final static int LABEL_SHADOW_THICKNESS = 1;

   /** Maximum offset between sort arrow and label text. */
   public final static int SORT_ARROW_MAX_LEFT_OFFSET = 8;

   /** Left clipping for sort area (sort arrow and number). */
   public final static int SORT_AREA_LEFT_CLIP = 1;

   /** Right clipping for sort area (sort arrow and number). */
   public final static int SORT_AREA_RIGHT_CLIP = 2;

   /** Gap between sort arrow and sort number. */
   public static final int GAP_BETWEEN_SORT_ARROW_AND_NUMBER = 1;

   /** GUI driver. */
   private GuiDriver gd;

   /** Parent browse. */
   private BrowseGuiImpl browse;

   /** Label font. */
   private FontDetails labelFontDetails;

   /** Font for sort number. */
   private FontDetails sortNumberFontDetails;

   /** Label font number. */
   private int labelFontNum;

   /** Column font. */
   private FontDetails columnFontDetails;

   /** Column font number. */
   private int columnFontNum;

   /** Base font number. */
   private int baseFontNum;

   /** Coordinates conversion. */
   private CoordinatesConversion cc;

   /** Column width. */
   private double width;

   /** 0-based column index. */
   private int columnIndex = -1;
   
   /** If it is the markers (leftmost) column. */
   private final boolean markersColumn;

   /** The lines of which the label consist. */
   private String[] labelLines = new String[0];

   /** <code>true</code> if the column is being dragged. */
   private boolean isDragged = false;

   /** <code>true</code> if the column is being resized. */
   private boolean isResized = false;

   /**
    * <code>true</code> if user pressed down mouse button on a column label if
    * ALLOW-COLUMN-SEARCHING is <code>true</code>.
    */
   private boolean startedPress = false;

   /**
    * If column is pressed (column label is pressed if ALLOW-COLUMN-SEARCHING is
    * <code>true</code>).
    */
   private boolean isPressed = false;

   /**
    * Indicates the resize area over which the mouse pointer is hovering now.
    * See {@link ResizeArea} for more information. <code>null</code> if the
    * mouse pointer is not hovering over a resize area.
    */
   private ResizeArea currentResizeArea = null;

   /**
    * Constructor.
    *
    * @param markersColumn
    *        <code>true</code> if it is a markers (leftmost) column.
    */
   public BrowseColumnGuiImpl(boolean markersColumn)
   {
      super(WidgetId.nextID());
      this.markersColumn = markersColumn;
   }

   /**
    * Constructor.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param id
    *        The widget's ID.
    * @param cfg
    *        Configuration data.
    */
   public BrowseColumnGuiImpl(WidgetId id, BrowseColumnConfig cfg)
   {
      this(id, cfg, false);
   }

   /**
    * Constructor.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param id
    *        The widget's ID.
    * @param cfg
    *        Configuration data.
    * @param markersColumn
    *        <code>true</code> if it is a markers (leftmost) column.
    */
   public BrowseColumnGuiImpl(WidgetId id, BrowseColumnConfig cfg, boolean markersColumn)
   {
      super(id, cfg);
      this.markersColumn = markersColumn;
   }

   /**
    * Set column width and update WIDTH attribute accordingly.
    *
    * @param width
    *        Column width in characters.
    */
   void setWidth(double width)
   {
      this.width = width;
      config().widthChars = width - cc.widthFromNative(getColumnPadding());
   }

   /**
    * Initialize the markers (leftmost) column.
    *
    * @param browse
    *        Parent browse.
    */
   public void initMarkersColumn(BrowseGuiImpl browse)
   {
      init(browse, 0, -1);
      OutputManager.instance().getRegistry().addWidget(this);
   }
   
   /**
    * Test whether this is a marker column. 
    * 
    * @return  {@code true} if this a marker column.
    */
   public boolean isMarkersColumn()
   {
      return markersColumn;
   }
   
   /**
    * Initialize the widget.
    *
    * @param browse
    *        Parent browse.
    * @param leftOffset
    *        Column offset in pixels (i.e. width of preceding columns).
    * @param columnIndex
    *        0-based column index.
    */
   public void init(BrowseGuiImpl browse, int leftOffset, int columnIndex)
   {
      this.browse = browse;
      this.columnIndex = columnIndex;

      gd = (GuiDriver) OutputManager.getDriver();
      cc = screen().coordinates();

      setPhysicalLocation(leftOffset, 0);

      width = cc.widthFromNative(calcColumnWidth());
   }
   
   /**
    * Obtain the {@link BrowseGuiImpl} this column is part of.
    * 
    * @return  the parent {@code BrowseGuiImpl} widget.
    */
   public BrowseGuiImpl getBrowse()
   {
      return browse;
   }
   
   /**
    * Calculate column width.
    *
    * @return column width.
    */
   public int calcColumnWidth()
   {
      int width = calcDataWidth();
      width += getColumnPadding();
      return width;
   }

   /**
    * Get column padding, i.e. difference between data width and actual column width.
    *
    * @return column padding in native units.
    */
   public int getColumnPadding()
   {
      int padding = 0;
      if (!markersColumn)
      {
         padding += LABEL_OFFSET_LEFT + LABEL_OFFSET_RIGHT;
         if (isFirstColumn())
         {
            padding += BORDER_THICKNESS;
         }
      }
      return padding;
   }

   /**
    * The method is called after the configuration associated with
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    *
    * @param    beforeUpdate
    *           Config reference capturing the config state before
    *           the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BrowseColumnConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
      if (browse != null)
      {
         resolveFonts();

         if (beforeUpdate.widthChars != config().widthChars && !browse.isPerformLayout())
         {
            width = calcColumnWidth();
         }
         repaint();

         if (!Utils.equals(beforeUpdate.clabel, config.clabel))
         {
            updateLabel();
         }
      }
   }

   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      Container p = parent();
      return new Dimension(width, (p != null) ? p.height() : 0);
   }

   /**
    * Draw the widget.
    */
   @Override
   public void draw()
   {
      final NativePoint p = this.physicalLocation();
      final NativeDimension d = this.physicalDimension();
      final NativeDimension clipDimension = new NativeDimension(d.width, d.height);
      if (browse.config().separators && isLastLockedColumn() && !isLastColumn())
      {
         clipDimension.width += BORDER_THICKNESS; // for separator
      }
      final NativeRectangle clip = clipRectangle(p, clipDimension);

      final int labelsHeight = browse.getLabelsHeight();

      gd.draw(p, clip, () ->
      {
         Theme theme = ThemeManager.getCurrentTheme();
   
         int labelHeight = 0; // including border
         
         if (browse.isLabelsVisible())
         {
            labelHeight += labelsHeight + BORDER_THICKNESS * 2;
            
            theme.drawBrowseColumnLabel(this, gd, 0, 0, d.width, labelHeight, sortNumberFontDetails);
         }
         
         if (markersColumn)
         {
            theme.drawBrowseMarkersColumn(this, gd, 0, labelHeight, d.width, d.height - labelHeight);
         }
         else 
         {
            theme.drawBrowseNormalColumn(this, gd, 0, labelHeight, d.width, d.height - labelHeight);
         }
      });
   }
   
   /**
    * Obtain the current {@code FontDetails} of this label.
    *
    * @return  the current {@code FontDetails} of the label.
    */
   public FontDetails getLabelFontDetails()
   {
      return labelFontDetails;
   }
   
   /**
    * Obtain the current {@code FontDetails} for this column.
    *
    * @return the current {@code FontDetails}.
    */
   public FontDetails getColumnFontDetails()
   {
      return columnFontDetails;
   }
   
   /**
    * Obtain the current font of this label.
    *
    * @return  the current font of the label.
    */
   public int getLabelFontNum()
   {
      return labelFontNum;
   }
   
   /**
    * Obtain the current font for this column.
    *
    * @return  the current font for this column.
    */
   public int getColumnFontNum()
   {
      return columnFontNum;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      NativePoint point = translate(e);
      browse.rowClicked(browse.getRowByCoordinates(point.y), columnIndex, e);
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      NativePoint point = translate(e);

      if (currentResizeArea != null)
      {
         isResized = true;
      }
      else if (browse.config().columnMovable && !markersColumn && browse.isLabelsVisible())
      {
         if (isHeaderArea(point))
         {
            isDragged = true;
         }
      }
      else if (!browse.config().columnMovable &&
               !markersColumn &&
               browse.isLabelsVisible() &&
               browse.config().allowColumnSearching)
      {
         if (isHeaderArea(point))
         {
            browse.resetPressedColumns(true);
            startedPress = true;
            isPressed = true;
            browse.repaint();
         }
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      if (isDragged)
      {
         isDragged = false;
         browse.stopColumnDrag(e);
      }

      if (isResized)
      {
         isResized = false;
         browse.stopColumnResize();
         gd.setCursor(MousePtrWrapper.DEFAULT);
         currentResizeArea = null;
      }

      boolean fireStartSearch = false;

      if (startedPress)
      {
         config.pressed = isHeaderArea(translate(e));
         if (config.pressed)
         {
            fireStartSearch = true;
         }
         startedPress = false;
         isPressed = false;
      }

      // update resize cursor
      mouseMoved(e);

      if (fireStartSearch)
      {
         ThinClient.getInstance().invokeTriggers(browse, Keyboard.SE_START_SEARCH, false);
      }

      browse.repaint();
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (isDragged)
      {
         if (!browse.isColumnDrag())
         {
            int leftDragColumnMouseOffset = translate(e).x;
            browse.startColumnDrag(this, leftDragColumnMouseOffset);
         }

         browse.mouseDragged(e);
      }
      else if (isResized)
      {
         if (!browse.isColumnResize() && currentResizeArea != null)
         {
            int index = 0;
            switch (currentResizeArea)
            {
               case LEFT:
                  index = columnIndex - 1;
                  break;
               case RIGHT:
                  index = columnIndex;
                  break;
               case LAST_LOCKED_RIGHT:
                  index = browse.config().numLockedColumns - 1;
                  break;
            }

            browse.startColumnResize(index);
         }

         browse.mouseDragged(e);
      }
      else if (startedPress)
      {
         NativePoint p = translate(e);
         isPressed = isHeaderArea(p);
         repaint();
      }
   }

   /**
    * Notification of a mouse moved event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      // switch to resize cursor if required
      if ((browse.config().columnResizable || config().resizable) && !markersColumn)
      {
         NativePoint p = translate(e);

         if (getColumnIndex() >= browse.config().numLockedColumns &&
             browse.isLastLockedColumnResizeArea(e))
         {
            gd.setCursor(MousePtrWrapper.E_RESIZE);
            currentResizeArea = ResizeArea.LAST_LOCKED_RIGHT;
         }
         else if (hasLeftResizeArea() && p.x >= 0 && p.x < COLUMN_RESIZE_AREA_WIDTH)
         {
            gd.setCursor(MousePtrWrapper.E_RESIZE);
            currentResizeArea = ResizeArea.LEFT;
         }
         else if (hasRightResizeArea() &&
                  p.x >= physicalDimension().width - COLUMN_RESIZE_AREA_WIDTH &&
                  p.x < physicalDimension().width)
         {
            gd.setCursor(MousePtrWrapper.E_RESIZE);
            currentResizeArea = ResizeArea.RIGHT;
         }
         else
         {
            gd.setCursor(MousePtrWrapper.DEFAULT);
            currentResizeArea = null;
         }
      }
   }

   /**
    * Get ancestor Window for this widget.
    *
    * @return  ancestor Window or <code>null</code> if the widget is not yet attached to a Window.
    */
   @Override
   public Window window()
   {
      return browse.window();
   }

   /**
    * Update the label (it should be parsed to the consisting lines).
    */
   void updateLabel()
   {
      labelLines = getLabelVar(config());
   }

   /**
    * Get the array of labels for this column. You may need to call {@link #updateLabel()} first
    * to have then initialized.
    *
    * @return the lines of which the label consist.
    */
   public String[] getLabelLines()
   {
      return labelLines;
   }

   /**
    * Initialize editor fill-in config.
    *
    * @param cfg
    *        Editor fill-in config.
    */
   void initEditorConfig(FillInConfig cfg)
   {
      cfg.selectionStart = ControlConfig.INVALID_SELECTION;
      cfg.selectionEnd = ControlConfig.INVALID_SELECTION;

      cfg.y = browse.getTopRowOffset(browse.getViewportRow());
      cfg.x = physicalLocation().x;

      cfg.fixedWidth = true;
      cfg.fixedHeight = true;
      cfg.widthPixels = physicalDimension().width;
      cfg.heightPixels = browse.getRowHeight();

      cfg.font = columnFontNum;

      // Background color
      int bgColor = config().columnBgColor;
      if (bgColor < 0)
      {
         bgColor = browse.config().columnBgColor;
      }

      if (bgColor < 0)
      {
         bgColor = browse.config().bgcolor;
      }

      cfg.bgcolor = bgColor;


      // Foreground color
      int fgColor = config().columnFgColor;
      if (fgColor < 0)
      {
         fgColor = browse.config().columnFgColor;
      }

      if (fgColor < 0)
      {
         fgColor = browse.config().fgcolor;
      }

      cfg.fgcolor = fgColor;
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Get the current custom mouse pointer
    * 
    * @return the current custom mouse pointer  
    *        
    */
   @Override
   public MousePtrWrapper getEffectiveMousePointer()
   {
      return MousePtrWrapper.DEFAULT;
   }

   /**
    * Index of the browse column in the set of columns.
    *
    * @return  0-based index of the browse column in the complete set of columns (except the
    *          markers column).
    */
   public int getColumnIndex()
   {
      return columnIndex;
   }

   /**
    * Sub-classes will override this to return a list of mouse actions on which the widget
    * responds.  This is compatible with the AWT {@link MouseEvent} codes and P2J specific
    * {@link com.goldencode.p2j.ui.client.event.MouseEvt} actions.
    *
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED
      };
   }

   /**
    * Returns <code>true</code> if the given point is inside the column header.
    *
    * @param  point
    *         Point to check.
    *
    * @return See above.
    */
   private boolean isHeaderArea(NativePoint point)
   {
      int labelsHeight = browse.getLabelsHeight() + 2 * BORDER_THICKNESS;
      return point.x >= 0 &&
            point.x <= physicalDimension().width &&
            point.y >= 0 &&
            point.y < labelsHeight;
   }

   /**
    * Get width of the label text, in pixels.
    *
    * @return see above.
    */
   public int getColumnLabelTextWidth()
   {
      int res = 0;
      for (String labelLine : labelLines)
      {
         res = Math.max(0, getTextWidth(labelLine, labelFontNum));
      }
      return res;
   }

   /**
    * Get width of the sort area (sort arrow and number), in pixels.
    *
    * @return see above.
    */
   public int getSortAreaWidth()
   {
      if (config.sortAscending == null)
      {
         return 0;
      }

      int res = SORT_ARROW_WIDTH;

      if (config.sortNumber != null && config.sortNumber > 0)
      {
         res += GAP_BETWEEN_SORT_ARROW_AND_NUMBER;

         /*if (sortNumberFontDetails == null)
         {
            sortNumberFontDetails = FontManager.resolveFont(window(),
                                                            FontManager.SORT_NUMBER_FONT,
                                                            gd);
         }*/

         res += getTextWidth(config.sortNumber.toString(), FontManager.SORT_NUMBER_FONT);
      }

      return res;
   }

   /**
    * Returns <code>true</code> if the column is visually in pressed state.
    *
    * @return see above.
    */
   public boolean isDrawnAsPressed()
   {
      if (markersColumn)
      {
         return false;
      }

      return startedPress ? isPressed : isPressed || config.pressed;
   }

   /**
    * Get clipping point for a text inside a label.
    *
    * @return see above.
    */
   public NativePoint getLabelTextClipPoint()
   {
      int x = LABEL_OFFSET_LEFT;
      
      // need to add one pixel more to take into account frame around header
      int y = CELL_OFFSET_TOP + BORDER_THICKNESS + 1;

      if (isDrawnAsPressed())
      {
         x += LABEL_PRESSED_OFFSET;
         y += LABEL_PRESSED_OFFSET;
      }

      return new NativePoint(x, y);
   }

   /**
    * Get clipping rectangle for a text inside a label.
    *
    * @param  clipPoint
    *         Clipping point.
    *
    * @return see above.
    */
   public NativeRectangle getLabelTextClipRectangle(NativePoint clipPoint)
   {
      int clipHeight = browse.getLabelsHeight() - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
      int labelAreaWidth = getLabelAreaWidth();
      int sortAreaWidth = getSortAreaWidth();

      int clipWidth;
      if (sortAreaWidth == 0)
      {
         clipWidth = labelAreaWidth - LABEL_OFFSET_RIGHT - LABEL_OFFSET_LEFT;
      }
      else
      {
         clipWidth = labelAreaWidth - LABEL_OFFSET_LEFT - SORT_AREA_LEFT_CLIP - sortAreaWidth  - SORT_AREA_RIGHT_CLIP;

         int labelTextWidth = getColumnLabelTextWidth();
         if (clipWidth > labelTextWidth)
         {
            clipWidth = Math.max(labelTextWidth, clipWidth - (SORT_ARROW_MAX_LEFT_OFFSET - SORT_AREA_LEFT_CLIP));
         }
      }

      return clipRectangle(clipPoint, new NativeDimension(clipWidth, clipHeight));
   }

   /**
    * Get width of the column header, without borders.
    *
    * @return see above.
    */
   public int getLabelAreaWidth()
   {
      NativeDimension d = physicalDimension();
      return d.width - (isLastColumn() ? BORDER_THICKNESS : 0);
   }

   /**
    * Get top offset for positioning text inside a label.
    *
    * @return see above.
    */
   public int getLabelTextTopOffset()
   {
      return labelFontDetails.legacyHeight + CELL_OFFSET_BOTTOM - 
                FontManager.getBrowseCellBottomTextOffset(labelFontDetails);
   }

   /**
    * Get top offset for positioning text inside a cell.
    *
    * @return see above.
    */
   public int getCellTextTopOffset()
   {
      return columnFontDetails.legacyHeight
            + CELL_OFFSET_BOTTOM - FontManager.getBrowseCellBottomTextOffset(columnFontDetails);
   }

   /**
    * Get clipping point for a text inside a cell relatively to the column top.
    *
    * @param viewportIndex
    *        0-based viewport index of the row the cell belongs to.
    *
    * @return see above.
    */
   public NativePoint getCellTextClipPoint(int viewportIndex)
   {
      return new NativePoint(DATA_OFFSET_LEFT,
                             browse.getTopRowOffset(viewportIndex) + CELL_OFFSET_TOP + 1);
   }

   /**
    * Get clipping rectangle for a text inside a cell.
    *
    * @param  point
    *         Clipping point.
    *
    * @return see above.
    */
   public NativeRectangle getCellTextClipRectangle(NativePoint point)
   {
      final NativeDimension d = physicalDimension();
      int bottom = browse.getRowHeight() - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
      int right = d.width - DATA_OFFSET_RIGHT - DATA_OFFSET_LEFT + 1;
   
      return clipRectangle(point, new NativeDimension(right - 1, bottom - 1));
   }

   /**
    * Calculate width of the data part of the column.
    *
    * @return width of the data part of the column.
    */
   private int calcDataWidth()
   {
      resolveFonts();

      Window window = window();

      if (markersColumn)
      {
         int fontWidth = FontManager.getFontWidth(window, baseFontNum, gd);
         return (int) (fontWidth * 1.6) + 2 * BORDER_THICKNESS; // TODO: is this formula correct?
      }

      if (config().widthChars > 0)
      {
         return cc.widthToNative(config().widthChars);
      }

      int dataWidth = 0;
      String dataType = config().dataType;
      String format = config().format;
      if (format == null)
      {
         format = DisplayFormat.instanceOfType(dataType).defaultFormatString();
      }

      boolean characterFormat = "character".equals(dataType);
      boolean logicalFormat = "logical".equals(dataType);

      if (characterFormat || logicalFormat)
      {
         try
         {
            DisplayFormat formatInstance;
            if (characterFormat)
            {
               formatInstance = new StringFormat(format);
            }
            else
            {
               formatInstance = new BoolFormat(format);
            }

            dataWidth += FontManager.getFontWidth(window, baseFontNum, gd) *
                           formatInstance.getScreenWidth();
         }
         catch (DisplayFormatParsingException e)
         {
            LOG.severe("Cannot parse format for browse column: " + format);
            return 0;
         }
      }
      else
      {
         int len = format.length();
         for (int i = 0; i < len; i++)
         {
            char c = format.charAt(i);
            switch (c)
            {
               case '9':
               case '>':
               case 'z':
               case 'Z':
               case '*':
                  // width of an arbitrary digit (they all have the same widths)
                  dataWidth += getTextWidth("0", baseFontNum);
                  break;
               default:
                  // other symbols like ', . - + /' widths are added as is
                  dataWidth += getTextWidth(String.valueOf(c), baseFontNum);
            }
         }
      }

      updateLabel();
      int labelWidth = 0;
      for (int i = 0; i < labelLines.length; i++)
      {
         labelWidth = Math.max(labelWidth, getTextWidth(labelLines[i], labelFontNum));
      }

      return Math.max(dataWidth, labelWidth);
   }

   /**
    * Resolve browse column fonts.
    */
   private void resolveFonts()
   {
      Window window = window();

      // base font used for calculation of browse elements sizes
      baseFontNum = browse.getBrowseFontResolver().resolveFontNum();

      if (markersColumn)
      {
         return;
      }

      // label font
      int labelFont = config().labelFont;
      if (labelFont != FontManager.UNDEFINED_FONT)
      {
         labelFontDetails = FontManager.resolveFont(window, labelFont, gd);
         labelFontNum = labelFont;
      }
      else
      {
         labelFont = browse.config().labelFont;

         if (labelFont != FontManager.UNDEFINED_FONT)
         {
            labelFontDetails = FontManager.resolveFont(window, labelFont, gd);
            labelFontNum = labelFont;
         }
         else
         {
            GuiFontResolver resolver = browse.getBrowseFontResolver();
            labelFontDetails = resolver.font().font;
            labelFontNum = resolver.resolveFontNum();
         }
      }

      // column font
      int columnFont = config().columnFont;
      if (columnFont != FontManager.UNDEFINED_FONT)
      {
         columnFontDetails = FontManager.resolveFont(window, columnFont, gd);
         columnFontNum = columnFont;
      }
      else
      {
         GuiFontResolver resolver = browse.getBrowseFontResolver();
         columnFontDetails = resolver.font().font;
         columnFontNum = resolver.resolveFontNum();
      }

      sortNumberFontDetails = FontManager.resolveFont(window(), FontManager.SORT_NUMBER_FONT, gd);
   }

   /**
    * Resolve color basing on input parameters.
    *
    * @param  columnValue
    *         Column-specific font number.
    * @param  browseValue
    *         Browse-specific font number.
    * @param  fg
    *         <code>true</code> if it is a foreground color. <code>false</code> if it is a
    *         background color.
    * @param  label
    *         <code>true</code> if it is a label color. <code>false</code> if it is a cell color.
    *
    * @return resolved color.
    */
   public ColorRgb resolveColor(int columnValue, int browseValue, boolean fg, boolean label)
   {
      Window window = window();
      int color = columnValue;
      if (color < 0)
      {
         color = browseValue;
      }

      if (color < 0)
      {
         if (label && !fg)
         {
            return gd.getSysColor(SysColor.COLOR_3DFACE);
         }
         else
         {
            GuiColorResolver browseResolver = browse.getBrowseColorResolver();
            return fg ? browseResolver.fgColor : browseResolver.bgColor;
         }
      }
      else
      {
         return fg ? ColorManager.getFgColor(window, color) :
                     ColorManager.getBgColor(window, color);
      }
   }

   /**
    * Returns <code>true</code> if it is the first column (not including markers column).
    *
    * @return see above.
    */
   public boolean isFirstColumn()
   {
      return columnIndex == 0;
   }

   /**
    * Returns <code>true</code> if it is the last column (not including markers column).
    *
    * @return see above.
    */
   public boolean isLastColumn()
   {
      return !markersColumn && (columnIndex == browse.getColumnCount() - 1);
   }

   /**
    * Determines if this column has left resize area (see {@link ResizeArea} for more
    * information).
    *
    * @return <code>true</code> if this column has left resize area.
    */
   private boolean hasLeftResizeArea()
   {
      if (isFirstColumn())
      {
         return false;
      }
      else if (columnIndex < browse.config().numLockedColumns)
      {
         // non-first locked column
         return true;
      }
      else
      {
         // scrolling set
         return physicalLocation().x + parent().physicalLocation().x >
                  browse.getLockedColumnsWidth();
      }
   }

   /**
    * Determines if this column has right resize area (see {@link ResizeArea} for more
    * information).
    *
    * @return <code>true</code> if this column has right resize area.
    */
   private boolean hasRightResizeArea()
   {
      return physicalLocation().x + parent().physicalLocation().x + physicalDimension().width <
               browse.getViewportDimension().width;
   }

   /**
    * Returns <code>true</code> if it is the last column if the set of locked columns (not
    * including markers column).
    *
    * @return see above.
    */
   public boolean isLastLockedColumn()
   {
      return !markersColumn && (columnIndex == browse.config().numLockedColumns - 1);
   }

   /**
    * Enumeration for different resize areas. Resize area is an area over which the mouse pointer
    * changes to resize pointer and user can resize column by dragging this area. All areas
    * have width of several pixels and height equal to viewport height.
    */
   private enum ResizeArea
   {
      /** Area on the left edge of a column - resizes previous column. */
      LEFT,

      /** Area on the right edge of a column - resizes current column. */
      RIGHT,

      /** Area after the right edge of the last locked column - resizes last locked column. */
      LAST_LOCKED_RIGHT
   }
}