/*
** Module   : ButtonGuiImpl.java
** Abstract : push button implementation for GUI
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20141120 Initial version.
** 002 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.  Do not draw directly from APIs which are   
**                  called from mouse events: instead, rely on repainting, which posts proper event
**                  to the event chain and process the drawing in the parent's context.
** 003 CA  20141206 Improvement of H002 - refresh needs to call repaint() always.
** 004 EVL 20141211 Detaching tooltip code from button widget.  Adding image loading for button.
** 005 EVL 20141226 Using images to draw button implementation.
** 006 EVL 20150115 Rework for image usage implementation.
** 007 EVL 20150121 Fixup image coordinates while loading.  Adding handling of the deferred
**                  images.
** 008 EVL 20150210 Fix for button background while image drawing.
** 009 EVL 20150220 Merge CONVERT-3D-COLORS flag to image widgets used to draw the button.
** 010 EVL 20150224 NO-CONVERT-3D-COLORS option is now processing with CONVERT-3D-COLORS attribute
**                  that is really the same(with boolean inversion).
** 011 EVL 20150313 Adding support for NO-FOCUS, FLAT-BUTTON attributes. Default button can not
**                  display any images, only size adjust.
** 012 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 013 EVL 20150320 Fix for GO event processing.  We need to unlock the update key reader.  Fix
**                  for mouse remains pressed even pointer is leaving the button.
** 014 EVL 20150413 Fixing clipping issues.
** 015 HC  20150323 Removed business logic from config classes.
** 016 EVL 20150428 Adding support for prepackaged images.
** 017 GES 20150425 Eliminated AWT dependencies.
** 018 HC  20150612 Implemented GUI ALERT-BOX.
** 019 EVL 20150629 Fixed the bug for ignoring the DEFAULT button attribute for handling frame's
**                  DEFAULT-BUTTON attribute.
** 020 VIG 20150809 Added super method call to mouseClicked and additional condition
**                  (.. && e.getButton() == MouseEvent.BUTTON1) to simulate CHOOSE event there.
** 021 CA  20150826 Added mouse actions.  Fixed button drawing on mouse events. Fixed 3D effects
**                  and highlight drawing which was 1 pixel too much in the Y dimension. Changed
**                  line stroke to be the small dotted stoke, which matches the Windows Classic
**                  theme which is our default. Future theme work will make this pluggable.
** 022 IAS 20150924 Added AUTO-RESIZE attribute support, changed TOOLTIP support
** 023 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 024 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 025 EVL 20151017 Fix for width/height calculation.  Also the button width/height is now fixed
**                  for 1 pixel mismatch.  Fix for button text pressed and disabled state to use
**                  centered string drawing function.
** 026 HC  20151023 Fixed GUI ALERT-BOX keyboard input processing.
**     CA  20151029 Fixed buttons dimension when is fixed via attributes.  Fixed 'focus lost' 
**                  notification - the button is now repainted.
**     EVL 20151106 Adding grey image creation for disabled state if only UP image is defined.
**                  Improve button paintings.  Fix for several image based buttons drawing, width
**                  and height mismatch.  Some logic reordered and comments cleanup.  The images
**                  used for buttons must have convert-3d-colors flag to be on.
** 027 IAS 20151214 LOAD-MOUSE-POINTER support.
** 028 CA  20160216 Cache the intermediate widget states.
** 029 VIG 20160215 Added labelInfo drawing (mnemonics logic) and isSupportsMnemonic() method.
** 030 VIG 20160510 Fixed NPE with labelInfo.
** 031 SBI 20160706 Changed to reset the button's state if it loses its focus.
** 032 SVL 20160719 Updated translate() signature.
** 033 EVL 20160813 Fix for button drawing state.
** 034 CA  20161014 Fixed misc issues related to FLAT-BUTTONs.  BGCOLOR and FGCOLOR are ignored
**                  when drawing the button.
** 035 HC  20161018 Fixed abends in image processing and drawing when image offset is involved.
**                  Added initial support for rendering of mnemonic labels to all labeled
**                  widgets.
** 036 HC  20161209 Only highlight the button when it really has focus.
** 037 EVL 20170131 Fix for icon images processing.
**     OM  20170201 Fixed icon images alignment.
**     EVL 20170202 Fix for image location regression.
**     CA  20161101 Moved the mnemonic drawing code at the driver.
** 038 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.util.*;

/**
 * GUI implementation of the BUTTON widget. This class should not have anything specific to
 * a specific UI modality (e.g. Swing or the Web).  It should only use the driver interface for
 * all drawing.
 */
public class ButtonGuiImpl
extends GuiButton
implements AutoResizable<ButtonConfig>
{
   /** Dotted line width. */
   public static final float DOT_LINE_WIDTH = 1.0f;
   
   /** Shift in XY to draw images on the button. */
   private static final int IMAGE_SHIFT = 3;
   
   /** Dotted line specific parameter. */
   private static final float DOT_STEP = 2.0f;
   
   /** Screen driver */
   private GuiDriver gd;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Pressed button state flag. Only when using SPACEBAR from keyboard. */
   private boolean keyPressed = false;
   
   /** Image to draw on button in up state. */
   private ImageGuiImpl imgUp = null; 
   
   /** Image to draw on button in pressed state. */
   private ImageGuiImpl imgDown = null; 
   
   /** Image to draw on button in disabled state. */
   private ImageGuiImpl imgDisabled = null;
   
   /**
    * This constructor creates a button with a given text label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this
    *          button.
    * @param   text
    *          The string to be used as the button text.
    */
   ButtonGuiImpl(WidgetId id, Container<GuiOutputManager> owner, String text) 
   {
      super(id, owner, text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ButtonGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ButtonConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ButtonConfig wcfg)
   {
      super.initialize(id, wcfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // screen coordinates conversion support 
      cc = screen().coordinates();
      // font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
      
      screen().getRegistry().addWidget(this);

      // handle pending images to preload
      loadPendingImages();

      // adjust image coordinates and width
      recalcImagesCoords();

      // set up initial dimension
      if (!config.fixedWidth)
      {
         config.widthChars = Coordinate.scale(width()); // direct access
      }
      if (!config.fixedHeight)
      {
         config.heightChars = Coordinate.scale(height()); // direct access
      }
   }
   
   /**
    * Sets the new image for given button widget on the client side.
    * 
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons.
    * @param    image
    *           A new image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setImage(int type, byte[] image)
   {
      boolean result = false;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_UP:
         {
            // get the image name
            String imgName = config().imgUpDef.name != null
                                ? config().imgUpDef.name : config().imgDef.name;
            
            if (imgUp == null)
            {
               boolean isIcon = imgName.toLowerCase().contains(".ico");
               imgUp = new ImageGuiImpl(WidgetId.nextID(), isIcon);
            }
            result = imgUp.setImage(image, false);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgUp, config().imgUpDef.name != null ? config().imgUpDef
                                                                     : config().imgDef);
               // prepackaged images has transparent background
               if (ImageHelper.isPrepackaged(imgName))
               {
                  imgUp.config().transparent = true;
               }
               // if image is used on the top of the button it must have convert-3d-colors on
               imgUp.config().convert3D = !imgUp.isIcon();
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            if (imgDown == null)
            {
               boolean isIcon = config().imgDownDef.name != null &&
                                config().imgDownDef.name.toLowerCase().contains(".ico");
               imgDown = new ImageGuiImpl(WidgetId.nextID(), isIcon);
            }
            result = imgDown.setImage(image, false);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDown, config().imgDownDef);
               // if image is used on the top of the button it must have convert-3d-colors on
               imgDown.config().convert3D = !imgDown.isIcon();
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            if (imgDisabled == null)
            {
               boolean isIcon = config().imgInsDef.name != null &&
                                config().imgInsDef.name.toLowerCase().contains(".ico");
               imgDisabled = new ImageGuiImpl(WidgetId.nextID(), isIcon);
            }
            result = imgDisabled.setImage(image, false);

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDisabled, config().imgInsDef);
               // prepackaged images has transparent background
               if (ImageHelper.isPrepackaged(config().imgInsDef.name))
               {
                  imgDisabled.config().transparent = true;
               }
               // if image is used on the top of the button it must have convert-3d-colors on
               imgDisabled.config().convert3D = !imgDisabled.isIcon();
            }
            break;
         }
      }
      
      return result;
   }
   
   /**
    * Loads the new image for given button widget from the client side filesystem.
    * 
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons.
    * @param    fullName
    *           The fully qualified file name for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean loadImage(int type, String fullName)
   {
      boolean result = false;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_UP:
         {
            if (imgUp == null)
            {
               imgUp = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgUp.loadImage(fullName, false);
            
            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgUp, config().imgUpDef.name != null ? config().imgUpDef
                                                                     : config().imgDef);
               // prepackaged images has transparent background
               if (ImageHelper.isPrepackaged(fullName))
               {
                  imgUp.config().transparent = true;
               }
               // if image is used on the top of the button it must have convert-3d-colors on
               imgUp.config().convert3D = !imgUp.isIcon();
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            if (imgDown == null)
            {
               imgDown = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDown.loadImage(fullName, false); 

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDown, config().imgDownDef);
               // if image is used on the top of the button it must have convert-3d-colors on
               imgDown.config().convert3D = !imgDown.isIcon();
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            if (imgDisabled == null)
            {
               imgDisabled = new ImageGuiImpl(WidgetId.nextID());
            }
            result = imgDisabled.loadImage(fullName, false); 

            // fixup button coordinates
            if (result)
            {
               fixupImageCoord(imgDisabled, config().imgInsDef);
               // prepackaged images has transparent background
               if (ImageHelper.isPrepackaged(fullName))
               {
                  imgDisabled.config().transparent = true;
               }
               // if image is used on the top of the button it must have convert-3d-colors on
               imgDisabled.config().convert3D = !imgDisabled.isIcon();
            }
            break;
         }
      }
    
      // if the image was loaded after button initialization, we need to recalculate new
      // position inside the button
      if (result)
      {
         recalcImagesCoords();
      }
      
      return result;
   }
   
   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * In GUI mode, the drawing needs to be delayed.
    */
   @Override
   protected void refresh()
   {
      repaint();
   }
   
   /**
    * Draw the button. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      // button options need to be refreshed before every painting
      gc = new GuiColorResolver(gd, null, SysColor.COLOR_3DFACE, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // need to update disabled image if there is UP image defined without INSENSITIVE one
      // calling it here instead of initialization stage ensures the proper window emulator
      // has already being loaded at this time
      updateDisabledImage();
      
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
         
      // clippable drawing start 
      gd.draw(this.getId(), p, clip, () ->
            ThemeManager.getCurrentTheme().drawButton(this, gd, gc, gf, 0, 0, d.width, d.height));
   }
   
   /**
    * Creates new {@link MnemonicInfo} instance.
    *
    * @return   MnemonicInfo or its descendant.
    */
   @Override
   public MnemonicInfo createMnemonic()
   {
      return new MnemonicInfoGuiImpl(false);
   }

   /**
    * Sets the disabled image according to UP image if the UP image defined but disabled - not.
    */
   private void updateDisabledImage()
   {
      // create new if none
      if (imgUp != null && imgDisabled == null)
      {
         imgDisabled = new ImageGuiImpl(WidgetId.nextID());
         // create grey version of the UP image
         boolean result = imgDisabled.setGreyImage(imgUp);
         // fixup button coordinates
         if (result)
         {
            // if image loaded we need to recalc coordinates
            recalcImageCoord(imgDisabled, config.widthPixels, config.heightPixels);
            // if image is used on the top of the button it must have convert-3d-colors on
            imgDisabled.config().convert3D = true;
         }
      }
   }
   
   /**
    * Checks if this button works as default one.  The frame DEFAULT-BUTTON attribute should also
    * be considered, not only DEFAULT attribute of the button itself.
    *
    * @return   <code>TRUE</code> if button works as default one <code>FALSE</code> otherwise.
    */
   public boolean isDefaultButton()
   {
      // look frame contains this button
      Frame tfr = UiUtils.locateFrame(this);
      // get the current button config
      ButtonConfig bcfg = config();
      
      // despite of the 4GL documentation declares the button default attribute must also be
      // set to make this work, the real application tests show the button option should be
      // ignored, the only that makes sense is the DEFAULT-BUTTON frame's option.
      return tfr != null && tfr.config().defaultButtonId == bcfg.id.asInt();
   }
   
   /**
    * Sets up initial image location inside the button widget.
    * 
    * @param   imgToFix
    *          Image object to fix the coordinates.
    * @param   def
    *          Image definition object to override the image under fixup.
    */
   private void fixupImageCoord(ImageGuiImpl imgToFix, ButtonImageDefinition def)
   {
      // image position
      if (config().noFocus && imgDown == null)
      {
         imgToFix.setPhysicalLocation(IMAGE_SHIFT, IMAGE_SHIFT);
      }
      // getting reference to image config
      ImageConfig icfg = imgToFix.config();
      // apply new data but only if we changed defaults
      // offsets
      if (def.offsetX != 0)
      {
         icfg.offsetX = def.offsetX;
      }
      else if (def.offsetCol != 0.0)
      {
         icfg.offsetCol = def.offsetCol;
      }
      if (def.offsetY != 0)
      {
         icfg.offsetY = def.offsetY;
      }
      else if (def.offsetRow != 0.0)
      {
         icfg.offsetRow = def.offsetRow;
      }
      // and size too
      if (def.widthPixels != 0)
      {
         icfg.widthPixels = def.widthPixels;
      }
      else if (def.widthChars != 0.0)
      {
         icfg.widthChars = def.widthChars;
      }
      if (def.heightPixels != 0)
      {
         icfg.heightPixels = def.heightPixels;
      }
      else if (def.heightChars != 0.0)
      {
         icfg.heightChars  = def.heightChars;
      }
      // finally merge the convert 3d colors flags for used image
      icfg.convert3D = config().convert3D;
   }
   
   /**
    * Check if the button has any images to be used to draw button.
    * 
    * @return  <code>true</code> there is at least one image to use.
    */
   public boolean hasImages()
   {
      return imgUp != null || imgDown != null || imgDisabled != null;
   }

   /**
    * Check if the button has prepackaged images to be used to draw button.
    * 
    * @return  <code>true</code> there is at least one image to use.
    */
   public boolean hasPrepackagedImages(ButtonConfig bcfg)
   {
      return (bcfg.imgDef.name != null && ImageHelper.isPrepackaged(bcfg.imgDef.name)) ||
             (bcfg.imgUpDef.name != null && ImageHelper.isPrepackaged(bcfg.imgUpDef.name));
   }

   /**
    * Need to recalculate image coordinates after all images have been loaded.
    */
   private void recalcImagesCoords()
   {
      // getting button width and height in pixels
      int bw = screen().coordinates().columnToNative(width());
      int bh = screen().coordinates().rowToNative(height());
      if (imgUp != null)
      {
         recalcImageCoord(imgUp, bw, bh);
      }
      if (imgDown != null)
      {
         recalcImageCoord(imgDown, bw, bh);
      }
      if (imgDisabled != null)
      {
         recalcImageCoord(imgDisabled, bw, bh);
      }
   }
   
   /**
    * Reposition image and change the width according button width and height.
    *
    * @param   imgToFix
    *          Image object to fix the coordinates.
    * @param   bw
    *          Button width.
    * @param   bh
    *          Button height.
    */
   private void recalcImageCoord(ImageGuiImpl imgToFix, int bw, int bh)
   {
      // getting image width and height in pixels
      int iw = imgToFix.getImageWidth();
      int ih = imgToFix.getImageHeight();
      
      // when the button has only UP image - the border around is bold
      boolean boldBorder = !config.noFocus && !config.flatButton && imgDown == null; 
      // effective button size is dependent on button mode
      if (boldBorder)
      {
         bw -= 3;
         bh -= 3;
      }
      // set it to the center of the button, regardless of border width
      imgToFix.setPhysicalLocation(boldBorder ? 1 + (bw - iw)/2 : (bw - iw)/2,
                                   boldBorder ? 1 + (bh - ih)/2 : (bh - ih)/2);
   }
   
   /**
    * Refresh the image objects to be used to draw button in different states with images that was
    * saved before in pending storage area.
    */
   private void loadPendingImages()
   {
      String imgName = null;
      // down image first
      imgName = config().imgDownDef.name;
      if (imgName != null && imgName.length() > 0 && imgDown == null)
      {
         loadPendingImage(ButtonConfig.IMAGE_DOWN);
      }
      // up image next
      imgName = (config().imgUpDef.name != null) ? config().imgUpDef.name : config().imgDef.name;
      if (imgName != null && imgName.length() > 0)
      {
         loadPendingImage(ButtonConfig.IMAGE_UP);
      }
      // image for disabled state
      imgName = config().imgInsDef.name;
      if (imgName != null && imgName.length() > 0)
      {
         loadPendingImage(ButtonConfig.IMAGE_INS);
      }
   }

   /**
    * Refresh the image object to be used to draw button in different states with images that was
    * saved before in pending storage area.
    * 
    * @param   type
    *          Image type to check.  Up, Down or Insensitive.
    */
   private void loadPendingImage(int type)
   {
      int thisId = config().id.asInt();

      // first try to search for images loaded from application jar
      byte[] imgData = ImageHelper.getPendingImageData(thisId, type);
      if (imgData != null)
      {
         setImage(type, imgData);
      }
      else
      {
         String loadName = ImageHelper.getPendingImageName(thisId, type);
         if (loadName != null && loadName.length() > 0)
         {
            loadImage(type, loadName);
         }
      }
   }
   
   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * 
    * @return  See above.
    */
   @Override
   public NativeRectangle physicalBounds()
   {
      // TODO: this is not correct just to verify mouse handling methods 
      return new NativeRectangle(cc.pointToNative(location()),
                                 cc.dimensionToNative(dimension()));
   }
   
   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      // get text mode height
      double hTxt = super.height();
      boolean textModeHeight = (hTxt == 1.0);
      
      if (config.fixedHeight)
      {
         // explicit height setting return it now
         return hTxt;
      }
      // get text mode height
      else if (config.autoResize)
      {
         Window<?> window = WindowManager.resolveWindow(this);
         if (gd == null)
         {
            gd = (GuiDriver) OutputManager.getDriver();
         }
         // get font resolver
         if (gf == null)
         {
            gf = new GuiFontResolver(window, gd, config(), false);
         }
         
         int font = gf.resolveFontNum();
         int fontHeight = FontManager.getFontHeight(window, font, gd);
         // the call below takes the lower even value for font height
         int hNative = 3 * (fontHeight/2) + (FontManager.isThreeD() ? 2 : 0) + 2;
         hTxt = screen().coordinates().heightFromNative(hNative);
         // we have changed the height got from text mode base
         // however this height is still text based and can be too big for small images
         textModeHeight = true;
      }

      // have no images use regular calculation
      if (!hasImages())
      {
         // no images, return text based version
         return hTxt;
      }
      
      // compute existed images heights
      double hUp   = imgUp != null ? imgUp.height() : 0.0;
      double hDown = imgDown != null ? imgDown.height() : 0.0;
      double hDis  = imgDisabled != null ? imgDisabled.height() : 0.0;
      
      // when the text mode height defined only by button text we have to ignore this value
      // from image based button height calculation
      if (textModeHeight)
      {
         // this excludes text base height from Max() computing chain below
         hTxt = 0.0;
      }
      
      // and return max value of them plus image insets
      return config().noFocus && imgDown != null
                       ? Math.max(super.height(), Math.max(hUp, Math.max(hDown, hDis)))
                       : !config().noFocus && imgUp != null && imgDown == null
                              ? Math.max(hTxt, Math.max(hUp, Math.max(hDown, hDis)) +
                                     screen().coordinates().heightFromNative(2*IMAGE_SHIFT + 1))
                              : Math.max(hTxt, Math.max(hUp, Math.max(hDown, hDis)) +
                                     screen().coordinates().heightFromNative(2*IMAGE_SHIFT - 2));
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      double wTxt = super.width();
      // detect if the width is defined by label length
      String lbl = textLabel();
      boolean textModeWidth = wTxt == (lbl != null ? (double)lbl.length() + 2.0 : 0.0);
      
      if (config.fixedWidth)
      {
         return wTxt;
      }
      // get text mode width
      else if (config.autoResize)
      {
         Window<?> window = WindowManager.resolveWindow(this);
         if (gd == null)
         {
            gd = (GuiDriver) OutputManager.getDriver();
         }
         // get font resolver
         if (gf == null)
         {
            gf = new GuiFontResolver(window, gd, config(), false);
         }
         
         int font = gf.resolveFontNum();
         String btnTxt = textLabel();
         // compute pixel based width
         int ntlimit = 10; // 5 pixels from each side is mandatory
         // add current text width in pixels
         ntlimit += btnTxt != null ? getTextWidthNative(btnTxt, font) : 0; 
         // convert to characters
         wTxt = screen().coordinates().widthFromNative(ntlimit);
         // changed width according to label text
         textModeWidth = true;
      }

      // have no images use regular calculation
      if (!hasImages())
      {
         return wTxt;
      }
      
      // compute existed images widths
      double wUp   = imgUp != null ? imgUp.width() : 0.0;
      double wDown = imgDown != null ? imgDown.width() : 0.0;
      double wDis  = imgDisabled != null ? imgDisabled.width() : 0.0;
      
      // when the text mode width defined only by button text we have to ignore this value
      // from image based button width calculation
      if (textModeWidth)
      {
         // this excludes text base width from Max() computing chain below
         wTxt = 0.0;
      }
      
      // and return max value of them plus image insets
      return config().noFocus && imgDown != null
                       ? Math.max(super.width(), Math.max(wUp, Math.max(wDown, wDis)))
                       : !config().noFocus && imgUp != null && imgDown == null 
                              ? Math.max(wTxt, Math.max(wUp, Math.max(wDown, wDis)) +
                                  screen().coordinates().widthFromNative(2*IMAGE_SHIFT + 1))
                              : Math.max(wTxt, Math.max(wUp, Math.max(wDown, wDis)) +
                                  screen().coordinates().widthFromNative(2*IMAGE_SHIFT - 2));
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  Simulate ENTER key typing.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      super.mouseClicked(e);
      
      // simulate enter key
      if (isEnabled() && isMouseOver() && e.getButton() == MouseEvent.BUTTON1)
      {
         // generate CHOOSE on button click
         KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_CHOOSE);
         EventManager.postEvent(ki);
         repaint();
      }
   }
   
   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      if (isFocusable() && !config().noFocus)
      {
         super.requestFocus();
      }
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return super.focusTraversable() && !config().noFocus;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_PRESSED, 
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_DRAGGED
      };
   }
   
   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   protected void handleBasicEvents(KeyInput event)
   {
      // special processing for spacebar key in GUI mode
      int id = event.id();
      
      if (isEnabled())
      {
         // save current state to optimize painting
         boolean oldPressed = keyPressed;

         // this is not a bug, but only way to simulate pressing for spacebar key
         // spacebar is typed enter that generates next pressed SE_CHOOSE and this
         // choose event will depress our button
         if (keyPressed && id == EventType.KEY_PRESSED)
         {
            keyPressed = false;
         }
         else
         {
            int keyCode = event.keyCode();
            // space bar is VK_ENTER with KEY_TYPED id
            if ((keyCode == Key.VK_ENTER || event.actionCode() == Keyboard.SE_CHOOSE) && 
                (id == EventType.KEY_TYPED || id == EventType.KEY_PRESSED))
            {
               // this is spacebar simulation start, the button will be unpressed
               // by next SE_CHOOSE pressed event
               if (event.isRealKey() && keyCode == Key.VK_ENTER && id == EventType.KEY_TYPED)
               {
                  keyPressed = true;
               }
               // simulate enter key by calling button action
               EventManager.postEvent(new ActionEvent(this, getCommand()));
            }
         }
         // redraw the button state if it was changed
         if (oldPressed != keyPressed)
         {
            // this kind of painting is required to properly draw the state changing
            // for several cases like image based button
            ThinClient.getInstance().eventDrawingBracket(this, this::repaint);
         }
      }
      
      // all other should be processed as usual
      super.handleBasicEvents(event);
   }

   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   protected boolean handleCursorKeys(int key, Frame<GuiOutputManager> frame)
   {
      // cursor movement does not work as focus change for GUI button
      if (key != Keyboard.KA_CURSOR_LEFT &&
          key != Keyboard.KA_CURSOR_UP &&
          key != Keyboard.KA_CURSOR_RIGHT &&
          key != Keyboard.KA_CURSOR_DOWN)
      {
         return super.handleCursorKeys(key, frame);
      }
      else
      {
         return false;
      }
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(ButtonConfig beforeUpdate)
   {
      if (!StringHelper.equal(config.label, beforeUpdate.label) ||
          config.font != beforeUpdate.font)
      {
         config.widthChars = width();
         config.heightChars = height();
      }
   }
   
   /**
    * Instructs the widget that its size may have changed and need to recompute the location of
    * the image, if any.
    *
    * @param   beforeUpdate
    *          {@link BaseConfig} reference or <code>null</code>.
    */
   @Override
   protected <C extends WidgetConfig> void updateSizeWorker(C beforeUpdate)
   {
      super.updateSizeWorker(beforeUpdate);
      
      if (hasImages())
      {
         recalcImagesCoords();
      }
      
      // repaint(); TODO: is this necessary ?
   }
   
   /**
    * This is called when the button looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      super.onFocusLost(focusEvent);
      setHighlighted(false);
      if (isEnabled() && keyPressed)
      {
         // the focus is lost, the button's state must be reset!!! 
         keyPressed = false;
      }
      repaint();
   }
   
   /**
    * Prepare event for processing.
    *   
    * @param   ki
    *          Input event.
    */
   protected void setActionCode(KeyInput ki)
   {
      int key = ki.actionCode();

      // alert box processing has no need to generate entry/leave events 
      // so we take a very simple approach
      if (key == Keyboard.KA_CURSOR_UP)
      {
         ki.setActionCode(Keyboard.KA_BACK_TAB);
         ki.setKeyCode(Key.VK_BACK_TAB);
         return;
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         ki.setActionCode(Keyboard.KA_TAB);
         ki.setKeyCode(Key.VK_TAB);
         return;
      }
      
      super.setActionCode(ki);
   }

   /**
    * Check whether this button supports mnemonics. All buttons support mnemonics. 
    * 
    * @return   Allways <code>true</code>.
    */
   public boolean isMnemonicSupported()
   {
      return true;
   }

   /**
    * Get current highlighted state.
    *
    * @return   highlighted state.
    */
   @Override
   public boolean highlighted()
   {
      // TODO: Do not rely on the super inplementation, in some cases it returns true even
      //       though the widget doesn't have focus.
      //       Also reimplement the super implementation by the means of
      //       Container.currentFocus() and WindowManager.isWindowActive() and remove
      //       the AbstractWidget.highlighted field.
      return hasFocus();
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      if (processMnemonicImpl(key, this))
      {
         EventManager.postEvent(new KeyInput(this, EventType.KEY_TYPED, Keyboard.SE_CHOOSE));
         
         // raise a CHOOSE event
         return true;
      }
      
      return false;
   }
   
   /**
    * Obtain the image in 'UP' state, if available.
    *
    * @return  a wrapper with current 'UP' image.
    */
   public ImageGuiImpl getImgUp()
   {
      return imgUp;
   }
   
   /**
    * Obtain the image in 'DOWN' state, if available.
    *
    * @return  a wrapper with current 'DOWN' image.
    */
   public ImageGuiImpl getImgDown()
   {
      return imgDown;
   }
   
   /**
    * Obtain the image in disabled state, if available.
    *
    * @return  a wrapper with current disabled image.
    */
   public ImageGuiImpl getImgDisabled()
   {
      return imgDisabled;
   }
   
   /**
    * Check whether this button is pressed.
    *
    * @return {@code true} if the button is pressed.
    */
   @Override
   public boolean isPressed()
   {
      // this button is pressed when it was pushed either by mouse or keyboard
      return keyPressed || super.isPressed();
   }
   
   /**
    * NO-FOCUS buttons cannot receive focus.
    *
    * @return  always {@code false} for NO-FOCUS buttons.
    */
   @Override
   public boolean isFocusable()
   {
      return !config.noFocus;
   }
   
   @Override
   public String toString()
   {
      return "ButtonGuiImpl{" + String.valueOf(getId()) + ", \"" + textLabel() + "\"}";
   }
}