/*
** Module   : CaptionButton.java
** Abstract : Define caption button associated with a window.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140917 Changed default BGCOLOR.
** 006 CA  20141021 Fixed drawing of caption buttons; allow restore button to show if window is
**                  maximized.
** 007 MAG 20141029 Added clipping rectangle when drawing.
** 008 HC  20150510 Improved drawing of caption buttons - they are now more Windows alike. 
** 009 GES 20150507 Removed AWT dependencies. 
** 010 HC  20150612 Implemented support for modal windows.
** 011 EVL 20150906 Adding loading SVG images for button body drawing.
** 012 CA  20150826 Added mouse actions.
** 013 EVL 20151008 Fix for button border painting.
** 014 CA  20160216 Cache the intermediate widget states.
**     SBI 20160309 Caption buttons aren't focus traversable.
** 015 HC  20160406 Overhaul of window-activation logic.
** 016 VIG 20151229 Added tooltip initialization.
** 017 HC  20160418 Fixed window caption button not to change press state when not enabled. Fixed
**                  ALERT-BOX title bar close button to close the box with correct result.
** 018 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.WindowEvent;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;

/**
 * A caption button associated with a window.
 */
@SuppressWarnings("unchecked")
public class CaptionButton
extends GuiButton
{
   /** The button type. */
   private CaptionButtonType type;

   /** The graphics driver. */
   private GuiDriver gd;

   /** The button's width, in character units. */
   private double width;
   
   /** The button's height, in character units. */
   private double height;
   
   /** Optional click handler. */
   private Runnable clickHandler;
   
   /**
    * Create a new button with the specified type.
    * 
    * @param    id
    *           The widget's ID.
    * @param    type
    *           The button type.
    */
   CaptionButton(WidgetId id, CaptionButtonType type)
   {
      super(id, null, "");

      gd = (GuiDriver) OutputManager.getDriver();

      this.type = type;

      config.tooltip = type.getTooltip();
   }
   
   /**
    * Returns the button type.
    * 
    * @return   The button type.
    */
   public CaptionButtonType getType()
   {
      return type;
   }

   /**
    * Caption buttons are not focus traversable.
    * 
    * @return   false
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }

   /**
    * Draw this caption button, depending on its type.
    */
   @Override
   public void draw()
   {
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      NativeRectangle clip = clipRectangle(p, d); // new NativeRectangle(p, d);
      
      // draw button
      gd.draw(getId(), p, clip, () ->
            ThemeManager.getCurrentTheme().drawCaptionButton(
                  this, gd, 0, 0, d.width, d.height));
   }

   /**
    * Set the button's dimension.
    * 
    * @param    width
    *           The button's width, in character units
    * @param    height
    *           The button's height, in character units
    */
   public void setDimension(double width, double height)
   {
      this.width = width;
      this.height = height;
   }
   
   /**
    * Get the button's {@link #width}.
    * 
    * @return   See above.
    */
   @Override
   public double width()
   {
      return width;
   }
   
   /**
    * Get the button's {@link #height}.
    * 
    * @return   See above.
    */
   @Override
   public double height()
   {
      return height;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      super.mouseClicked(e);
      
      if (clickHandler != null)
      {
         clickHandler.run();
      }
      else
      {
         GuiWindow window = parent(GuiWindow.class);
         
         switch (type)
         {   
             case CLOSE:
                // 4GL doesn't allow closing the app from the close button, 
                // but the WINDOW-CLOSE event is raised
                ThinClient.getInstance().eventBracket(() ->
                {
                   Event evt = new WindowEvent(window, Keyboard.SE_WINDOW_CLOSE);
                   EventManager.postOSEvent(evt);
                });
                break;
             case ICONIFY:
                window.iconify();
                break;
             case MAXIMIZE:
                window.maximize();
                break;
         }
      }
      
      repaint();
   }
   
   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }
   
   /**
    * Sets the optional click handler for the button. 
    * 
    * @param handler
    *        Click handler.
    */
   public void setClickHandler(Runnable handler)
   {
      clickHandler = handler;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      int windowEvt = (type == CaptionButtonType.CLOSE   ? MouseEvt.WINDOW_CLOSED :
                       type == CaptionButtonType.ICONIFY ? MouseEvt.WINDOW_ICONIFIED :
                                                           MouseEvt.WINDOW_MAXIMIZED);
      
      return new int[]
      {
         MouseEvent.MOUSE_MOVED,
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED,
         windowEvt
      };
   }
   
   /**
    * Checks whether this widget can receive focus generally.
    *
    * @return  always {@code false}.
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }
}