/*
** Module   : ComboBoxGuiImpl.java
** Abstract : GUI combo-box implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150421 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
** 003 EVL 20150514 Adding GUI specific keyboard and mouse default exit events.  Turning on
**                  tooltip engine.  Adding consistent cleanup on drop-down exit for GUI mode.
**                  Adding scaling factor to get rid of the hardcoding adjustments.  Moving sets
**                  location of the drop-down button to later stage, just before drawing to
**                  properly support dynamic combo-box.
** 004 EVL 20150719 Fix for issue when initial value displayed in entry field when nothing yet is
**                  selected.  We can use selected flag to detect the difference.  Adding another
**                  scaling option to compute required widget height.  Fix for getting pixel width
**                  and height attribute.  Minor painting details coord issues.  TAB/BACK_TAB
**                  event processing fix.  Need to recalc button location before every painting
**                  because the width can be changed every time with width-* attribute.  Clean up
**                  the scaling compute method.
** 005 EVL 20150814 Adding helper method to provide row height to parts of the combo-box.  If the
**                  width or height was set explicitly - return it immediately.
** 006 EVL 20150907 Reworked drop-down button drawing to use SVG engine to get proper image.
** 007 HC  20150831 Scrolling support overhaul.
** 008 CA  20150910 Added mouse actions.
** 009 IAS 20150924 AUTO-RESIZE and TOOLTIP attributes support
** 010 CA  20150925 Made exitDropDown() public, it needs to be called if the drop-down is
**                  notified that is losing focus.  Fixed multi-window drawing issues caused by 
**                  concurrent drawing access to the GuiDriver: all access to the driver is done
**                  exclusively, each thread taking and releasing ownership for the duration of
**                  the drawing operation or other GUI API invocation which requires access to
**                  the underlying physical window.
** 011 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 012 EVL 20151014 Fix for button background color issue. Fix for incorrect displaying the value
**                  instead of the label om initial start. 
** 013 HC  20151029 Fixed drop-down scroll list width calculation, it must not be smaller than
**                  the parent's width.
**         20151103 Improved GUI COMBO-BOX layout.
**         20151110 GUI combo-box fixes.
**         20151112 Implemented THREE-D runtime support.
**     EVL 20151118 Adding override for getDataEntryReturn().  It should always return FALSE in
**                  GUI mode ignoring SESSION:DATA-ENTRY-RETURN value.  The combo-box does not
**                  react to the ENTER key in GUI.  Key processsing fixes for GUI mode.
** 014 EVL 20151120 Fix for drop-down line spacing issue. And fix for button incorrect drawing
**                  over 3d border.  Drop-down button fix for pressed state drawing.  Pressing
**                  the drop-down button need to be painted accordingly.  When the widget is
**                  loosing focus we need to repaint it.
** 015 EVL 20151204 Restoring SIMPLE and DROP-DOWN modes functionality.
** 016 EVL 20151217 Drop-down approach rework.  Fix for simple mode scrolling regression.
** 017 EVL 20160119 Adding focus move postsprocessing.
** 018 IAS 20160130 LOAD-MOUSE-POINTER support.
** 019 IAS 20160303 Fixed RETURN processing.
** 020 CA  20160216 Cache the intermediate widget states.
** 021 HC  20160406 COMBO-BOX drop-down resources cleanup. Various drop-down related fixes.
** 022 HC  20160413 Fixed COMBO-BOX focus management, this also resolves ENTER not confirming
**                  selected item in the drop-down.
** 023 HC  20160511 Fixed COMBO-BOX focusing issues and related changes.
** 024 SBI 20160708 Changed to reset the active combo box reference and the locked pointer.
** 025 SBI 20161004 Fixed truncated items by using isFormatHonored() and fixed the combo box width
**                  if honorFormat is false, minor height and width differences.
** 026 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
** 027 EVL 20161102 Fix fir SIMPLE mode combo-box issues.
** 028 CA  20170131 Fixed color inheritance when not explicitly set at widget but maybe set at 
**                  frame (and frame might or might not be THREE-D). 
**     CA  20161101 Added support for mnemonic processing.
**     SVL 20170206 Set proper colors for entry field.
** 029 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.ComboBoxConfig.Mode;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Frame;
import com.goldencode.p2j.ui.client.Point;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.ScrollBar.Bars;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Specific GUI implementation of COMBO-BOX widget.
 */
@SuppressWarnings("unchecked")
public class ComboBoxGuiImpl
extends ComboBox<GuiOutputManager>
implements AutoResizable<ComboBoxConfig>
{
   /** Shift in X to draw text inside entry. */
   public static final int TEXT_SHIFT = 4;
   
   /** The space difference between item row height and full combo-box height. */
   private static final int ENTRY_TO_HEIGHT_Y_DIFF = 8;
   
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /** Editable entry field for SIMPLE and DROP-DOWN modes. */
   private EntryFieldGuiImpl entryField = null;
   
   /** The list for SIMPLE combo-box mode. */
   private ScrollableSelectionListGuiImpl simpleList = null;
   
   /** The list for SIMPLE combo-box mode. */
   private ScrollPaneGuiImpl simpleListPane = null;
   
   /** The key event that must be posted after drop-down exit. */
   private KeyInput keyToDefer = null;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ComboBoxGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ComboBoxConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ComboBoxConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      if (gd == null)
      {
         gd = (GuiDriver) OutputManager.getDriver();
      }
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // font resolver
      if (gf == null)
      {
         gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
      }
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param   beforeUpdate
    *          Config reference capturing the config state before 
    *          the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(ComboBoxConfig beforeUpdate)
   {
      NativeDimension beforeDim = this.physicalDimension();

      super.afterConfigUpdate(beforeUpdate);
      
      if (!StringHelper.equal(config.format, beforeUpdate.format) ||
          config.font != beforeUpdate.font)
      {
         updateSize(beforeUpdate);
      }
      
      if (simpleList != null)
      {
         // if location or dimension has changed, then re-layout
         int beforeX = ConfigHelper.getAssignedX(beforeUpdate);
         int beforeY = ConfigHelper.getAssignedY(beforeUpdate);
         
         int afterX = ConfigHelper.getAssignedX(config);
         int afterY = ConfigHelper.getAssignedY(config);
         
         NativeDimension afterDim = this.physicalDimension();
         
         if (beforeX != afterX || beforeY != afterY || !afterDim.equals(beforeDim))
         {
            layoutSimpleList();
         }
      }
   }
   
   /**
    * Draw the combo-box. Called by this ComboBox's parent container.
    */
   @Override
   public void draw()
   {
      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      Theme theme = ThemeManager.getCurrentTheme();
      
      // update entry field for SIMPLE and DROP-DOWN modes
      if (entryField != null)
      {
         int shift = theme.getComboBoxEntryShift();
         entryField.setPhysicalLocation(p.x + shift, p.y + shift);
      }
      
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_WINDOW, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // resolve font
      final FontCache fntCache = gf.font();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(getId(), p, clip, () ->
      {
         theme.drawComboBoxEntryField(this, gd, gc, 0, 0, d.width, d.height,
                                      fntCache.font, pressed, isEnabled(), false);
         // in SIMPLE or DROP-DOWN modes the entry field drawing
         // is happening from the frame the entry field belongs to
         
         // the field border
         if (config().mode == Mode.SIMPLE && simpleListPane != null)
         {
            NativeDimension efDim = new NativeDimension(d.width, getEntryFieldHeight() - 1);
            theme.drawComboBoxFieldBorder(this, gd, 0, 0, efDim.width, efDim.height, 
                                          pressed, isEnabled(), false);
         }
         else
         {
            theme.drawComboBoxFieldBorder(this,
                  gd, 0, 0, d.width, d.height, pressed, isEnabled(), false);
         }
         
         // and drop-down button for modes except simple
         if (config().mode != Mode.SIMPLE)
         {
            theme.drawComboBoxButton(this,
                  gd, 0, 0, d.width, d.height, pressed, isEnabled(), false);
         }
         
         // draw drop-down in SIMPLE mode border
         if (config().mode == Mode.SIMPLE && simpleListPane != null)
         {
            NativeDimension slDim = simpleListPane.physicalDimension();
            // plus extra space for decorations
            slDim.width += 4;
            slDim.height += 3;
            theme.drawComboBoxFieldBorder(this, gd, 0, getEntryFieldHeight(), 
                  slDim.width + 4, slDim.height + 3, pressed, isEnabled(), false);
         }
         // in SIMPLE mode the list drawing is happening from the frame the list belongs to
      });
   }
   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple drivers
      
      // nothing to do in GUI
   }

   /**
    * This is called when the combo-box gains the input focus. In addition to regular action when
    * simple mode is active we need to move input focus to the entry field.
    *
    * @param    event
    *           The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      super.onFocusGained(event);
      
      if (isEnabled())
      {
         // generate the FOCUS_GAINED event
         if (entryField != null)
         {
            entryField.requestFocus();
         }
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  Redraw combo-box in pressed
    * state and request focus.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // push button effect
      if (isEnabled() && !pressed)
      {
         pressed = true;

         // drop-down is opening on mouse press
         if (config.mode != ComboBoxConfig.Mode.SIMPLE)
         {
            // Close another active drop-down here. This solves the case when
            // there is an active drop-down and mouse click causes another drop-down 
            // to activate.
            // We cannot rely on the window-deactivated event here for the drop-down
            // to be closed as the window event may get dispatched after the click event.
            ComboBoxGuiImpl combo = (ComboBoxGuiImpl) ThinClient.getInstance().activeComboBox();
            if (combo != null && combo != this)
            {
               combo.exitDropDown();
            }
            
            // depending on current state activate or dismiss drop-down 
            if (isDropDownActive())
            {
               dropDown.hide();
            }
            else
            {
               // first need to display the button pressing effect 
               ThinClient.getInstance().independentEventDrawingBracket(this, this::repaint);

               // activate drop-down later to avoid activation mess
               ThinClient.getInstance().invokeLater(this::activate);
            }
         }
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  Redraw combo-box in
    * released state.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // release button effect
      if (isEnabled() && pressed)
      {
         pressed = false;
         repaint();
      }
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      return processMnemonicImpl(key, this);
   }
   
   /**
    * Creates new {@link MnemonicInfo} instance.
    *
    * @return   MnemonicInfo or its descendant.
    */
   @Override
   public MnemonicInfo createMnemonic()
   {
      return new MnemonicInfoGuiImpl(false);
   }
   
   /**
    * Process key events that have been generated by this component.
    * 
    * @param    keyEvent
    *           The key event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      if (!isEnabled())
      {
         return;
      }
      
      // this event is generated on combo-box parts to invoke a 4GL triggers
      // we need to consume and ignore it because triggers should already be called
      if (keyEvent.actionCode() == Keyboard.SE_VALUE_CHANGED)
      {
         keyEvent.consume();
         return;
      }
      
      if (!isDropDownActive())
      {
         int key = keyEvent.actionCode();
         
         // consider focus move events first
         if (key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB)
         {
            super.processKeyEvent(keyEvent);
         }
         else if (entryField != null && key != Keyboard.SE_VALUE_CHANGED)
         {
            entryField.processKeyEvent(keyEvent);
         }
         else
         {
            int newIndx = ScrollableList.getFirstCharSelection(keyEvent, model(),
                                                         model().getSelectedIndex());
            if (newIndx >= 0)
            {
               // select the new item
               setNewSelection(newIndx);
               // and consume event
               keyEvent.consume();
            }
            else
            {
               super.processKeyEvent(keyEvent);
            }
         }
      }
      else if (dropDown != null && dropDown.isVisible())
      {
         dropDown.processKeyEvent(keyEvent);
      }
      else
      {
         super.processKeyEvent(keyEvent);
      }
   }
   
   /**
    * Process the action associated with the given key in an independent drawing bracket.  In
    * GUI combo-box the precessing is highly dependend on the current mode.
    * <p>
    * This must be called only in the context of widget-level event processing.
    * 
    * @param    keyEvent
    *           The key to be processed.
    */
   @Override
   public void processSystemKey(final KeyInput keyEvent)
   {
      // get the key code
      int key = keyEvent.actionCode();
   
      // if the key is not in GUI specific key set - call superclass for any mode
      if (key != Keyboard.KA_CURSOR_DOWN && key != Keyboard.KA_CURSOR_UP &&
          key != Keyboard.KA_CURSOR_LEFT && key != Keyboard.KA_CURSOR_RIGHT &&
          key != Keyboard.KA_PAGE_DOWN && key != Keyboard.KA_PAGE_UP &&
          key != Keyboard.KA_HOME && key != Keyboard.KA_END)
      {
         // check if we press GUI combo-box drop-down opener
         int keyCode = keyEvent.keyCode();
         // alternative GUI drop-down opening by keyboard
         if (keyCode == Key.VK_ALT_CURSOR_DOWN || keyCode == Key.VK_ALT_CURSOR_UP)
         {
            activate();
         }
         else
         {
            super.processSystemKey(keyEvent);
         }
      }
      // below is the GUI specific keys processing depending on mode
      else if (config.mode == ComboBoxConfig.Mode.SIMPLE)
      {
         // the mode is simple - route processing to the list for all keys
         simpleList.processKeyEvent(keyEvent);
      }
      // the mode is DROP-DOWN or DROP-DOWN-LIST
      // the keys below change the current selection not opening drop-down part
      else
      {
         // get the currently selected index
         int currIndex = model().getSelectedIndex();
         int newIndexCandidate = currIndex; 

         // change index depending on key types
         if (key == Keyboard.KA_CURSOR_DOWN || key == Keyboard.KA_CURSOR_RIGHT)
         {
            newIndexCandidate++;
         }
         else if (key == Keyboard.KA_CURSOR_UP || key == Keyboard.KA_CURSOR_LEFT)
         {
            newIndexCandidate--;
         }
         else if (key == Keyboard.KA_PAGE_DOWN)
         {
            newIndexCandidate += getInnerLines();
         }
         else if (key == Keyboard.KA_PAGE_UP)
         {
            newIndexCandidate -= getInnerLines();
         }
         else if (key == Keyboard.KA_HOME)
         {
            newIndexCandidate = 0;
         }
         else if (key == Keyboard.KA_END)
         {
            newIndexCandidate = model().size() - 1;
         }
         
         // check the edge conditions
         if (newIndexCandidate < 0)
         {
            newIndexCandidate = 0;
         }
         else if (newIndexCandidate >= model().size())
         {
            newIndexCandidate = model().size() - 1;
         }

         // nothing is changed, return
         if (newIndexCandidate == currIndex)
         {
            return;
         }
         
         // start actual intenal model change
         final int newIndex = newIndexCandidate;
         
         // select the next item
         ThinClient.getInstance().independentEventDrawingBracket(this,
               () -> setNewSelection(newIndex));
      }
   }

   /**
    * Process the action associated with the given focus movement key in an independent drawing
    * bracket.  Currently it is <code>TAB</code> or <code>BACK_TAB</code>.
    * <p>
    * This must be called only in SIMPLE or DROP-DOWN mode either from entry field or from
    * attached selection list.
    * 
    * @param    keyEvent
    *           The key to be processed, <code>TAB</code> or <code>BACK_TAB</code>.
    */
   public void processFocusMoveKey(KeyInput keyEvent)
   {
      // set the current focus to the main combo-box part
      parent().setFocus(this);
      // call superclass to do the rest
      super.processSystemKey(keyEvent);
   }
   
   /**
    * Get widget screen value.
    * 
    * @return  Current widget screen value.
    */
   @Override
   public BaseDataType getScreenValue()
   {
      return entryField != null ? new character(entryField.getScreenValue(true))
                                : super.getScreenValue();
   }

   /**
    * Shows or hides the combo-box and its label.
    *
    * @param    visible_
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible_)
   {
      super.setVisible(visible_);
      
      if (visible_)
      {
         // create fill-in for SIMPLE and DROP_DOWN modes
         if (entryField == null &&
             (config().mode == ComboBoxConfig.Mode.SIMPLE || 
              config().mode == ComboBoxConfig.Mode.DROP_DOWN))
         {
            createEntryField();
         }

         // for simple mode we need to create permanent drop-down
         if (config().mode == ComboBoxConfig.Mode.SIMPLE && simpleListPane == null)
         {
            createSimpleList();
            layoutSimpleList();
         }
      }
      
      if (entryField != null)
      {
         entryField.setVisible(visible_);
      }
      
      if (simpleListPane != null)
      {
         simpleListPane.setVisible(visible_);
      }
   }
   
   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#height()
    */
   @Override
   public double height()
   {
      if (config.clientHeightChars == BaseConfig.INV_COORD)
      {
         updateSize(null);
      }
      
      return config.clientHeightChars != BaseConfig.INV_COORD ?
             config.clientHeightChars : config.heightChars;
   }
   
   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#width()
    */
   @Override
   public double width()
   {
      if (config.clientWidthChars == BaseConfig.INV_COORD)
      {
         updateSize(null);
      }
      
      return config.clientWidthChars != BaseConfig.INV_COORD ?
             config.clientWidthChars : config.widthChars;
   }
   
   /**
    * This method is invoked when selection is changed.
    */
   @Override
   public void changed()
   {
      // call common code first
      super.changed();

      // then GUI specific
      if (config().mode == ComboBoxConfig.Mode.SIMPLE && simpleList != null)
      {
         setNewSelection(simpleList.index());
      }
      else if (isDropDownActive())
      {
         setNewSelection(dropDown.getSelectedIndex());
      }
   }

   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    *          <code>TRUE</code> to make the widget enabled, 
    *          <code>FALSE</code> to disable the widget.
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      // check if we need repaint
      boolean needRepaint = enabled != isEnabled();
      // set flag for main part
      super.setEnabled(enabled);

      // set flag for possible entry filed
      if (entryField != null)
      {
         entryField.setEnabled(enabled);
      }

      // set flag for possible simple list
      if (simpleList != null)
      {
         simpleList.setEnabled(enabled);
         // this ensure proper list repainting
         if (needRepaint)
         {
            ThinClient.getInstance().independentEventDrawingBracket(simpleList,
                  () -> simpleList.repaint());
         }
      }

      if (simpleListPane != null)
      {
         simpleListPane.setEnabled(enabled);
      }

      // and post value changed event
      if (needRepaint)
      {
         repaint();
      }
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      // check the combo-box itself for focus
      boolean cbbFocused = super.hasFocus();
      
      return config().mode == ComboBoxConfig.Mode.SIMPLE &&
                simpleList != null && entryField != null
                   ? cbbFocused || simpleList.hasFocus() || entryField.hasFocus()
                   : config().mode == ComboBoxConfig.Mode.DROP_DOWN && entryField != null 
                      ? cbbFocused || entryField.hasFocus() || isDropDownActive() : cbbFocused;
   }
   
   /**
    * Gets the current row height depending on choosen font.
    * 
    * @return  The row height value in pixels.
    */
   public int getRowHeight()
   {
      // if height is not calculated yet, do it now
      if (config.clientHeightChars == BaseConfig.INV_COORD)
      {
         updateSize(null);
      }

      // subtract extra spacing from client coordinates
      return config.mode == ComboBoxConfig.Mode.SIMPLE
                ? cc.heightToNative(1.0) - ENTRY_TO_HEIGHT_Y_DIFF
                : config.clientHeightPixels - ENTRY_TO_HEIGHT_Y_DIFF;
   }
   
   /**
    * Gets the width of the scrollbar used for drop-down.  If the drop-down is not active or has
    * no vertical scroll-bar - returns 0 width.
    * 
    * @return  The width value in pixels for scrollbar.
    */
   public int getDropDownScrollBarWidth()
   {
      // result value, init to 0
      int wRes = 0;

      // check if drop-down is active and valid
      if (isDropDownActive())
      {
         ScrollPaneGuiImpl sp = (ScrollPaneGuiImpl) dropDown.getScrollPane();
         // if scrollpane is OK, get the bar width
         if (sp != null)
         {
            wRes = sp.barSizeNative(ScrollBar.Orientation.VERTICAL);
         }
      }
      // or we are in simple mode
      else if (config.mode == ComboBoxConfig.Mode.SIMPLE && simpleListPane != null)
      {
         wRes = simpleListPane.barSizeNative(ScrollBar.Orientation.VERTICAL);
      }
      
      return wRes;
   }
   
   /**
    * Destroy this widget.  Cleans up all including widgets before destruction.
    */
   public void destroy()
   {
      // deregister possible SIMPLE list components
      if (simpleListPane != null)
      {
         OutputManager.instance().getRegistry().removeWidgetsRecursive((Widget) simpleListPane);
      }

      // call superclass for main destruction
      super.destroy();
   }

   /**
    * Posts value changed event to the queue.  This should be done in GUI mode after any changes
    * in entry field for SIMLPLE and DROP_DOWN combo-box modes.
    * 
    * @param   newValue
    *          The new value for combo-box.
    */
   public void entryFieldChanged(String newValue)
   {
      // put an ActionEvent onto the system event queue
      EventManager.postEvent(new ActionEvent(this, newValue));
      
      // honor any Progress triggers
      KeyInput vchng = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_VALUE_CHANGED, true);
      EventManager.postEvent(vchng);
      
      ConfigHelper.setModified(config, true);

      // SIMPLE mode must try to scroll the list to the first item match
      if (config.mode == ComboBoxConfig.Mode.SIMPLE && simpleList != null)
      {
         int currTopIndx = simpleList.getTopRow(); 
         int newTopIndx = ScrollableList.getStartStringSelection(newValue, model());
         
         // the list must be in selection off mode until direct selection by mouse or keys
         simpleList.setHighlighted(false);
         if (newTopIndx >= 0 && currTopIndx != newTopIndx)
         {
            // move item on the top
            simpleList.setTopRow(newTopIndx);
         }
         else if (newTopIndx == -1 && currTopIndx != 0)
         {
            // no match found, move first item on the top
            simpleList.setTopRow(0);
         }
      }
   }

   /**
    * Saves the key event to be posted into event queue after drop-down is gone.
    * 
    * @param   ki
    *          The key event to be posted after drop-down become closed.
    */
   public void deferKeyEvent(KeyInput ki)
   {
      keyToDefer = ki;
   }
   
   /**
    * Sometimes it is required to do several action the was deferred during drop-down life cycle.
    */
   @Override
   protected void postprocessDropDown()
   {
      // do any action that was deferred during drop-down life period
      
      // we deferred some key, post it now
      if (keyToDefer != null)
      {
         processFocusMoveKey(keyToDefer);
         keyToDefer = null;
      }
   }
   
   /**
    * Exit the drop-down (the "selection" mode of the combo-box).  GUI specific cleanup.
    */
   @Override
   public void exitDropDown()
   {
      ThinClient.getInstance().resetActiveComboBox();
      
      boolean isConditionRaised = endConditionRaised; 
      endConditionRaised = true;
      
      // call generic cleanup from the superclass
      super.exitDropDown();

      // clean up registry from drop-down widgets added in GUI code
      OutputManager.instance().getRegistry().removeWidgetsRecursive((Widget) dropDown);
      
      if (!isConditionRaised)
      {
         ThinClient.getInstance().requestEndCurrentWaitForEventLoop();
      }
   }
   
   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(ComboBoxConfig beforeUpdate)
   {
      if (gd == null)
      {
         return;
      }
      
      // current combo-box window
      Window<?> window = window();
      if (window == null)
      {
         // the widget is not attached, just return
         return;
      }
      
      CoordinatesConversion cc = screen().coordinates();
      
      // get font resolver
      if (gf == null)
      {
         gf = new GuiFontResolver(window, gd, config(), false);
      }
      
      // get font to use in combo-box
      int font = gf.resolveFontNum();
      
      // in case the height was not explicitly set use 1 value
      // but in simple mode we use the exact value set explicitly
      if (config.mode != ComboBoxConfig.Mode.SIMPLE)
      {
         double inputHeight = config.heightChars > 0.0 && config.heightChars < 1.0
                                 ? config.heightChars : 1.0;
         config.clientHeightPixels = cc.heightToNative(inputHeight, window.getId(), font) +
                                     ENTRY_TO_HEIGHT_Y_DIFF;
      }
      else
      {
         config.clientHeightPixels = config.heightPixels;
      }
      
      if (config.widthChars != 0)
      {
         // the explicitly assigned width is the exact width of the widget,
         // ignore format in this case.
         // the width in this case is converted to physical size using
         // the default font.
         config.clientWidthChars = config.widthChars;
         return;
      }
      
      double charWidth = getFormatLength();
      
      int pixelWidth;
      
      if (!isFormatHonored())
      {
         // the extra 27 pixels cover the button, borders and padding,
         // the value was determined experimentally
         pixelWidth = getItemMaximalWidth(font) + 27;
      }
      else
      {
         // TODO: variable button size
         // the extra 28 pixels cover the button, borders and padding,
         // the value was determined experimentally
         pixelWidth = cc.widthToNative(charWidth, window.getId(), font) + 28;
      }
      config.clientWidthPixels = pixelWidth;
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * This is called when the combo-box loses the input focus.
    *
    * @param    focusEvent
    *           The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      super.onFocusLost(focusEvent);
      // the state changed, need to repaint in GUI
      repaint();
   }
   
   /**
    * Returns the list for SIMPLE combo-box mode.
    *
    * @return   The list instance.
    */
   public Object getList()
   {
      return simpleList;
   }

   /**
    * Refresh items in the combo-box after assignment.
    */
   @Override
   public void refreshItems()
   {
      // in SIMPLE and DROP-DOWN modes we need to ignore format string
      if (config.mode != ComboBoxConfig.Mode.DROP_DOWN_LIST)
      {
         config.honorFormat = false;
      }
      super.refreshItems();
   }
   
   /**
    * Returns simple list height without decorations.
    */
   public int getSimpleListVisibleHeight()
   {
      return config.clientHeightPixels - 4 - getEntryFieldHeight();
   }
   
   /**
    * Registers drop-down frame in current window system.  GUI specific version.
    */
   @Override
   protected void registerDropDownFrame()
   {
      endConditionRaised = false;
      
      ScrollableSelectionListGuiImpl list = (ScrollableSelectionListGuiImpl) dropDown.getList();
      list.setOnItemSelected(() -> dropDown.hide());
      
      dropDown.registerWidgetStateListener(w ->
      {
         if (!endConditionRaised)
         {
            endConditionRaised = true;
            ThinClient.getInstance().requestEndCurrentWaitForEventLoop();
         }
      });
   }
   
   /**
    * Requesting focus to opened drop-down.
    */
   @Override
   protected void dropDownRequestFocus()
   {
      if (dropDown != null)
      {
         dropDown.getList().requestFocus();
      }
   }
   
   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Get the current custom mouse pointer if it is set
    * 
    * @return the current custom mouse pointer if it is set  
    *        
    */
   @Override
   public Optional<MousePtrWrapper> getCustomMousePointer()
   {
      // TODO: override for Browse & BrowseColumn
      return mousePtr.isPresent() ? mousePtr : 
         enclosingFrame().map(Frame::getCustomMousePointer).
         orElse(
                  config().mode == Mode.DROP_DOWN_LIST ? 
                        MousePtrWrapper.DEFAULT_PTR : 
                        MousePtrWrapper.TEXT_PTR 
               );
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[] { MouseEvent.MOUSE_PRESSED, MouseEvent.MOUSE_RELEASED };
   }
   
   /**
    * Posts value changed event to the queue.  This should be done in ChUI and GUI mode after
    * closing drop-down or in GUI mode after pressing UP/DOWN keys with focused combo-box.
    */
   @Override
   protected void valueChangedCommon()
   {
      // set up entry field if exists
      if (entryField != null)
      {
         String newText = model().selected();
         // set new value
         entryField.setValue(new character(newText));
         // set up selection range to the full text selection
         entryField.config().selectionStart = 0;
         // at this time the entry field should have new value
         // we need to use screen representation rather than internal value that can have
         // extra space chars to match the current data format
         entryField.config().selectionEnd = entryField.getText(true).length();
         // update state value for fill-in
         entryField.setState(ScreenBuffer.CHANGED);
      }
      
      // and call superclass
      super.valueChangedCommon();
      
      // clear changed flag to avoid duplications
      valueChanged = false;
      // change the selected flag
      selected = true;
   }
   
   /**
    * Initializes the event set to return.  GUI specifics,
    * 
    * @return  The keyboard event to be used to exit drop-down.
    */
   @Override
   protected Set<String> initEventsSet()
   {
      // call common version
      Set<String> setGuiResult = super.initEventsSet();

      // add GUI specific messages
      setGuiResult.add("TAB");
      setGuiResult.add("BACK-TAB");
      setGuiResult.add("LEAVE");
      
      return setGuiResult;
   }
   
   /**
    * Adding mouse clicks as exit events.  GUI specifics.
    * 
    * @param   listToChange
    *          The event list to modify.
    */
   @Override
   protected void addUISpecificEvents(EventList listToChange)
   {
      // call common version first
      super.addUISpecificEvents(listToChange);

      // mouse click on anywhere dismisses drop-down
      listToChange.addEvent("MOUSE-MENU-DOWN", true);
      listToChange.addEvent("MOUSE-EXTEND-DOWN", true);
      
      // chaining current event list from ThinClient enclosing event loop
      EventList elCurr = ThinClient.getInstance().getCurrentEventList();
      // chain only events for the given combo-box instance
      if (elCurr != null)
      {
         listToChange.setChain(elCurr.getNonExitEventsForWidget(getId().asInt()));
      }

      // enable drop-down painting
      ((DropDownGuiImpl)dropDown).enableDraw();
   }

   /**
    * Clean up the main registry pool from drop-down widgets when it is closing, if the widget is
    * a container, make recursive call.
    */
   private void createEntryField()
   {
      // get ID for new widget
      WidgetId efId = WidgetId.nextID();

      // create config for entry field, we use fill-in config for this widget,
      // meaning no server-side counterpart for this widget
      FillInConfig finCfg = new FillInConfig(efId);

      // inherit some options from combo-box config
      finCfg.labels = false;
      finCfg.frameId = config.frameId;
      finCfg.dataType = config.dataType;
      finCfg.fgcolor = config.fgcolor;
      finCfg.bgcolor = config.bgcolor;
      // format must be ignored if not explicitly defined, compute from item values
      String fmt = config.format != null && !config.format.isEmpty() ? config.format : null;  
      finCfg.setDynamicFormat(fmt);

      // create entry field
      entryField = new EntryFieldGuiImpl(finCfg, this);
      // this is required to properly initialize entry field
      GuiOutputManager.instance().getRegistry().addWidget(efId, entryField);
      // initialize entry field
      entryField.initialize(efId, finCfg);
      // add widget config to config manager
      ConfigManager.getInstance().addWidgetConfig(finCfg);
      
      // add the widget to the frame the combo-box belongs to
      Frame<GuiOutputManager> frCombo = (Frame<GuiOutputManager>) UiUtils.locateFrame(this);
      frCombo.addWidget(entryField);

      // inherit enabled state from combo-box main
      entryField.setEnabled(isEnabled());
      // set up initial value if exists
      BaseDataType initial = getValue();
      if (initial != null && !initial.isUnknown())
      {
          entryField.setValue(new character(getInitialScreenText(initial)));
      }
   }
   
   /**
    * Creates the list for SIMPLE mode of the combo-box.  This part will permanently exist for
    * the combo-box life time.
    */
   private void createSimpleList()
   {
      // create the drop-down box
      simpleList = screen().getFactory().createScrollableList(this, this.model());

      // register it as a new frame
      simpleListPane = screen().getFactory().createScrollPane(simpleList);

      // set height, width and position of the simlpe itself
      // account for the border
      simpleList.setColumns(width() - simpleListPane.getRootPanel().insets().widthDelta());

      // adjust visible rows to list height
      int nativeListHeight = getSimpleListVisibleHeight();
      // ceil to ensure even partial row is visible
      int listDispRows = (int) Math.ceil((double) nativeListHeight / (double) getRowHeight());
      int listVisibleRows = (int) Math.floor((double) nativeListHeight / (double) getRowHeight());
      // update internals
      simpleList.setVisibleRows(listVisibleRows);
      simpleList.setRowsToDisplay(listDispRows);
      // this is just to be in sync, SIMPLE mode ignores this
      config.innerLines = listVisibleRows;
      // force the initial selection
      int idx = model().calcSelection(config.items);
      
      // if initial value was explicitly set, override the defaul initial
      BaseDataType initial = getValue();
      if (initial != null && !initial.isUnknown())
      {
         idx = model().indexOf(getInitialScreenText(initial));
      }
      
      simpleList.select(idx);
      
      // set up initial top row
      simpleList.setTopRow(idx);
     
      // set up notifications
      simpleList.addSelectionListener(this);
      
      // set initial selection mode to off
      simpleList.setHighlighted(false);

      // init scrollbars, visible on demand
      simpleListPane.setScrollBars(Bars.VERTICAL);
         
      // adding to the combo-box parent frame
      Frame<GuiOutputManager> frCombo = (Frame<GuiOutputManager>) UiUtils.locateFrame(this);
      frCombo.addWidget(simpleListPane);

      // adding to the registry, no mouse event will be available if not to register
      OutputManager.instance().getRegistry().addWidgetsRecursive((Widget) simpleListPane);
   }

   /**
    * Returns entry field height.
    */
   private int getEntryFieldHeight()
   {
      return screen().coordinates().heightToNative(1.0) - (FontManager.isThreeD() ? 2 : 0);
   }
   
   /**
    * Layout simple list.
    */
   private void layoutSimpleList()
   {
      double w = width() - screen().coordinates().widthFromNative(4);
      double h = screen().coordinates().heightFromNative(getSimpleListVisibleHeight()); 
      simpleListPane.setSize(w, h);
      
      // set location in character units
      Point origin = screenLocation();
      origin.y += 1;
      simpleListPane.setLocation(origin);
      
      // set up physical location just below the combo-box
      NativePoint slLoc = physicalLocation();
      // the shift to draw the list properly
      slLoc.x += 2;
      slLoc.y += getEntryFieldHeight() + 2;
      simpleListPane.setPhysicalLocation(slLoc.x, slLoc.y);
   }
   
   /**
    * Changes the selected value without opening drop-down.
    * 
    * @param   indexNew
    *          The index of the new item to select.
    */
   private void setNewSelection(int indexNew)
   {
      // new value can not be out of the current range
      if (indexNew < 0 || indexNew >= model().size())
      {
         return;
      }
      
      // set up the model with new selection
      model().setSelectedIndex(indexNew);
      model().setSelectedId(indexNew + 1);
      model().select(model().get(indexNew));

      // and post value changed event
      valueChangedCommon();
   }

   /**
    * Gets the initially setted screen value.
    * 
    * @param   initVal
    *          The number of the list items pair or entry for the list item.
    * 
    * @return  The string representation of the initial value.
    */
   public String getInitialScreenText(BaseDataType initVal)
   {
      // first try to scan by value index
      String ndx = initVal.toStringMessage();
      String res = null;

      // scan the config for items
      for (ControlSetItem itm : config.items)
      {
         if (itm.getValue().toStringMessage().equals(ndx))
         {
            // found the match exiting
            res = itm.getLabel().toStringMessage();
            break;
         }
      }

      // initial value is not index, try to get the value as is
      if (res == null)
      {
         res = initVal.toStringMessage();
      }
      
      return res;
   }
   
   /**
    * Returns the format length of the data type the combo-box represents.
    * 
    * @return   Format length.
    */
   public int getFormatLength()
   {
      String       type = config.dataType;
      BaseDataType bdt  = DisplayFormat.instanceOfType(type);
      
      return bdt.formatLength(config.format); 
   }
   
   /**
    * Calculates the maximal item width in the pixel units for the given font number.
    * 
    * @param    font
    *           The provided font.
    * 
    * @return   The maximal item width.
    */
   private int getItemMaximalWidth(int font)
   {
      int itemWidth = 0;
      ComboBoxModel<String> items = model();
      int itemsSize = items.size();
      for (int i = 0; i < itemsSize; i++)
      {
         itemWidth = Math.max(itemWidth, getTextWidth(items.get(i), font));
      }
      
      return itemWidth;
   }
   
   /**
    * Returns the inner lines value for combo-box.  Used in closed state.
    * 
    * @return   Inner lines value.
    */
   private int getInnerLines()
   {
      // start with config value
      int resIL = config.innerLines;
      
      // second try is to ask the list model
      if (resIL <= 0)
      {
         int mSize = model().size(); 
         resIL = mSize > 0 ? mSize : DropDownGuiImpl.INNER_LINES_MIN;
      }
      
      return resIL;
   }
}