/* 
** Module   : EntryFieldGuiImpl.java
** Abstract : combo-box editable text field implementation, GUI-specific version.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20150605 Created initial version. Adding scaling factor to get rid of the hardcoding
**                  adjustments.
** 002 EVL 20150723 The disabled color should be initialized here, regular fill-in does not have
**                  this drawing mode.  TAB/BACK_TAB event processing fix.
** 003 EVL 20150814 The widget width depends on the main combo-box widget part.
** 004 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 005 HC  20151124 Utilized Keep3DFillinBorder environment variable when drawing FILL-IN border.
** 006 EVL 20151204 Restoring SIMPLE and DROP-DOWN modes combo-box functionality.
** 007 IAS 20151214 LOAD-MOUSE-POINTER support.
** 008 IAS 20160303 Fixed RETURN processing.
** 009 EVL 20161102 Fix for SIMPLE mode combo-box issues.
** 010 CA  20170131 Fixed color inheritance when not explicitly set at widget but maybe set at 
**                  frame (and frame might or might not be THREE-D).
**     SVL 20170206 Added disabledForeFixedColor.
** 011 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;

/**
 * Implementation of Progress entry field inside for GUI combo-box.
 */
@SuppressWarnings("unchecked")
public class EntryFieldGuiImpl
extends FillInGuiImpl
{
   /** Screen driver */
   private GuiDriver gd;

   /** Reference to the possible combo-box object. */
   private ComboBoxGuiImpl comboBoxRef = null;
   
   /** Local focus traversable flag. */
   private boolean focusTraversable = true;

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public EntryFieldGuiImpl(WidgetId id, FillInConfig cfg)
   {
      super(id, cfg);
      disabledForeFixedColor = true;
      
      // init GUI driver
      gd = (GuiDriver) OutputManager.getDriver();
      // set up insets
      this.nativeInsets = new NativeInsets(2, 2, 2, 2);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   cfg
    *          Configuration data.
    * @param   cbbRef
    *          Combo-box reference.
    */
   public EntryFieldGuiImpl(FillInConfig cfg, ComboBoxGuiImpl cbbRef)
   {
      this(cfg.id, cfg);
      comboBoxRef = cbbRef;
      
      cfg.frameId = cbbRef.config().frameId;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, FillInConfig cfg)
   {
      super.initialize(id, cfg);
      
      // disabled text color, entry field differs from fill-in here
      disabledFore  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      // init completed
      setState(comboBoxRef.getState());
   }
   
   /**
    * Process GUI specific keys for combo-box entry field:
    * <ul>
    *    <li><code>SE_VALUE_CHANGED</code> - inform combo-box the value changed.</li>
    * </ul>
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      int key = ke.actionCode();

      // in combo-box mode is active and we have value changed in fill-in
      if (key == Keyboard.SE_VALUE_CHANGED)
      {
         // stop handling this event for fill-in
         ke.consume();
         // and reroute them to the enclosing combo-box
         comboBoxRef.entryFieldChanged(getText());
      }
      else if (key == Keyboard.KA_CURSOR_DOWN || key == Keyboard.KA_CURSOR_UP)
      {
         // up, down keys will be rerouted in the main part of the combo-box
         // depending on the mode currently in use
         comboBoxRef.processSystemKey(ke);
      }
      else if (key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB)
      {
         // tab key will be rerouted in the main part of the combo-box to change the focus
         // we also need to disable focus traversing for the time of the processing this event
         focusTraversable = false;
         comboBoxRef.processFocusMoveKey(ke);
         focusTraversable = true;
      }
      else
      {
         super.processKeyEvent(ke);

         // in GUI delete, backspace generate combo-box change         
         if ((key == Keyboard.KA_BACKSPACE || key == Keyboard.KA_DELETE_CHARACTER) &&
             config.modified)
         {
            KeyInput vchng = new KeyInput(this,
                                          EventType.KEY_PRESSED,
                                          Keyboard.SE_VALUE_CHANGED,
                                          true);
            EventManager.postEvent(vchng);
         }
      }
   }

   /**
    * Convenience method to get access to SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @return  Value of the SESSION:DATA-ENTRY-RETURN attribute.
    */
   protected boolean getDataEntryReturn()
   {
      return false;
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return focusTraversable;
   }

   /**
    * Get widget width.
    * 
    * @return  Widget width.
    */
   @Override
   public double width()
   {
      Theme theme = ThemeManager.getCurrentTheme();
      int subtractWidth = 2 * theme.getComboBoxEntryShift();
      if (comboBoxRef.config().mode == ComboBoxConfig.Mode.DROP_DOWN)
      {
         subtractWidth += theme.getComboBoxButtonSize();
      }
      
      return comboBoxRef.width() - screen().coordinates().widthFromNative(subtractWidth);
   }
   
   /**
    * Get widget height.
    *
    * @return  Widget height.
    */
   @Override
   public double height()
   {
      Theme theme = ThemeManager.getCurrentTheme();
      return comboBoxRef.height() - screen().coordinates().heightFromNative(
            2 * theme.getComboBoxEntryShift());
   }
   
   /**
    * Draws the caret.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple drivers
      if (!comboBoxRef.hasFocus() || !isEnabled())
      {
         return;
      }
      
      // drawing worker
      drawCaretInt();
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // just focus request, no superclass call to avoid click event duplication
      // see FillInGuiImpl for clarification mouse events handling
      requestFocus();
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // call superclass for this
      super.mouseClicked(e);
   }

   /**
    * Get the current effective custom mouse pointer
    * 
    * @return the current effective custom mouse pointer  
    *        
    */
   @Override
   public MousePtrWrapper getEffectiveMousePointer()
   {
      return MousePtrWrapper.TEXT;
   }

   /**
    * Check if the default mouse pointer should be used regardless of
    * the parent settings
    * 
    * @return
    *        <code>true</code> if the default mouse pointer should always be used
    */
   public boolean enforceDefaultMousePtr()
   {
      return true;
   }
   
   /**
    * Determine the portion of the widget's editable part.  This may be different than the widget's
    * width, as determined by the WIDTH-CHARS or WIDTH-PIXELS attribute.
    * <p>
    * For GUI widgets, this is the physical dimension without the insets.
    * 
    * @return   The editable dimension of this widget, in native units.
    */
   @Override
   protected NativeDimension editableNativeDimension()
   {
      NativeDimension dim = physicalDimension();
      
      // combo-box mode needs pure physical dimension based calculation
      dim.width -= 1;
      dim.height-= nativeInsets.top + nativeInsets.bottom;
      
      return dim;
   }

   /**
    * Draws the selected part of the etry field.  This is the entry field version.  Unlike pure
    * fill-in implementation it allows to draw selection even if the entry field does not have
    * the focus itself.
    * 
    * @param   txt
    *          The text to draw.
    * @param   font
    *          The font details to use while text drawing.
    * @param   clipWidth
    *          The width of the clipping rectangle.
    * @param   clipHeight
    *          The height of the clipping rectangle.
    * @param   fgColor
    *          The text color.
    * @param   bgColor
    */
   @Override
   protected void drawSelection(String txt, int font, int clipWidth, int clipHeight, 
                                ColorRgb fgColor, ColorRgb bgColor)
   {
      // we have to check full combo-box widget with all parts to having focus 
      if (!comboBoxRef.hasFocus() || !isEnabled() || !validSelection())
      {
         return;
      }
      
      // call fill-in worker to draw the current selection
      drawSelectionInt(txt, font, clipWidth, clipHeight,
                       gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT),
                       gd.getSysColor(SysColor.COLOR_HIGHLIGHT));
   }
   
   /**
    * Draws the border box around the fill-in.
    * 
    * @param    threeDBorder
    *           Indicates whether 3d or flat border should be drawn.
    */
   @Override
   protected void drawBorder(GuiColorResolver gcd,
                             int x,
                             int y,
                             NativeDimension dim,
                             boolean threeDBorder)
   {
      // no border, the combobox will draw it for us
   }
   
   /**
    * The combo editor is never 3D.
    * 
    * @return   Always false
    */
   @Override
   public boolean isThreeD()
   {
      return false;
   }
}