/*
** Module   : FrameGuiImpl.java
** Abstract : Frame implementation for GUI
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20141010 Created initial version.
** 002 HC  20141021 Implemented window move/position support.
** 003 EVL 20141127 Inserting this container into mouse source chain.
** 004 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.  Replaced saved dimension with real dimension
**                  on resize.
** 005 EVL 20150311 Adding default button processing for GUI.
** 006 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width 
**                  widget.
** 007 EVL 20150414 Fix for boxed frames coordinates 1-pixel shift in resizeScrollPane().
** 008 HC  20150323 Removed business logic from config classes.
** 009 CA  20150406 Added APIs to measure the label's width/height.
** 010 CA  20150507 Fixed frame title drawing.
** 011 CA  20150509 The frame shows the scrollbars on a "as-needed" basis, not always.
** 012 HC  20150612 Improved widget registration/deregistration.
** 013 EVL 20150629 Fixed the bug for ignoring the DEFAULT button attribute for handling frame's
**                  DEFAULT-BUTTON attribute.
** 014 CA  20150615 Added support for EDITOR's RETURN-INSERTED attribute.
** 015 CA  20150723 Fixed drawing of down frames.  The frame box is now being drawn by the frame 
**                  scroller (ScrollPaneGuiImpl), same as ChUI does it.
** 016 CA  20150724 Fixed a regression in H015, when drawing frames with TITLE - the frameScroll
**                  needs to be positioned bellow the body.
** 017 CA  20150728 Fixed more issues related to H015 - drawing down frames with titles and 
**                  side-labels.
** 018 CA  20150731 resizeScrollPane needs to be executed in a eventBracket, to discard paint 
**                  events.
** 019 HC  20150831 Scrolling support overhaul.
** 020 HC  20150920 Fixed the text color of the frame active title.
** 021 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
**                  Fix for inactive frame title painting.
** 022 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 023 IAS 20151012 CANCEL-BUTTON runtime support
** 024 EVL 20151012 Fix for frame focus request with no focusable widgets inside. Frame layout
**                  needs to be done at realization, to be aware of the current window dimension.
**                  Fix for frameScroll size out of sync when config was changed.
** 025 EVL 20151029 Fix for frame title height calculation that cause unwanted scrollbars to
**                  appear.  As soon as we have wrong scrollpane size even for 1 pixel - we have
**                  scrollbars.  The caption height fix calculation.  Having caption automatically
**                  means having box around the frame.  Fix for wrong bottom border drawing.  The
**                  scrollpane size adjust should be performed on frame initialization too.
**     HC  20151112 Implemented THREE-D runtime support.
** 026 CA  20151123 Replaced root frame check with dedicated API.
** 027 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 028 EVL 20151217 Drop-down approach rework.  Make frame capture height API to be public.
** 029 CA  20160105 Added runtime support for BORDER- attributes.
** 030 HC  20160203 Improved runtime support of FRAME:VIRTUAL* and FRAME:SCROLLABLE attributes.
**                  Frame layout improvements.
** 031 IAS 20160217 LOAD-MOUSE-POINTER support.
** 032 HC  20160320 Improved GUI frame layout to respect virtual size of the enclosing frame.
** 033 CA  20160327 Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 034 SBI 20160307 Changed an effective widget id for the frame's hover action.
**     SBI 20160322 Changed to hide a window title bar and a border if the graphics environment
**                  doesn't support window decorations.
** 035 EVL 20160410 Fix for minimum size of frame calculation.
** 036 EVL 20160421 Fix for dialog-box header check.  No frame title in this case.
** 037 EVL 20160513 Adding color support change for dialog-box.  Explicit dialog sizing needs the
**                  internal frame size to be changed accordingly.  Added setInnerSize overriden
**                  implementation to do GUI dialog specifics.  Added method to get the dialog
**                  decoration elements size.
** 038 HC  20160628 Prevented unhandled exceptions when window/frame closed/destroyed.
** 039 EVL 20160622 Adding refresh dialog title after config update.
** 040 SBI 20160713 Fixed the cursor if the mouse pointer over labels.
** 041 SBI 20160818 Fixed the look and feel for frames according to the Progress 4GL GUI minimal
**                  frame size.
** 042 CA  20160915 Fixed placement of unrealized frames (after they were once realized).
** 043 SBI 20160908 Changed to override scroll container to return the scroll dimension that
**                  fits its content if the frame size is greater than its frame's width.
** 044 HC  20160922 Fixed frame scrolling.
** 045 CA  20161020 Fixed a horiz scrollbar 4GL bug, which prevents showing the horiz scrollbar
**                  for the parent frame if the child frame uses a ENABLE or DISPLAY statement 
**                  and the parent frame requires a horiz scrollbar.
** 046 HC  20161018 Fixed resolution of label fonts during frame layout.
** 047 HC  20161206 Fixed frame height calculation.
** 048 CA  20170104 Some improvements in window decoration support for embedded mode (modal 
**                  windows always have decorations).
** 049 CA  20170131 Fixed color inheritance when not explicitly set at widget but maybe set at 
**                  frame (and frame might or might not be THREE-D).
** 050 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.chui.ThinClient.TriggerResult;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.ScrollPaneGuiImpl.ShowBars;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.layout.StackLayout.Direction;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * A <code>Frame</code> is a top-level window with a title and a border. 
 */
@SuppressWarnings("unchecked")
public class FrameGuiImpl
extends Frame<GuiOutputManager>
{
   /**
    * The frame's minimal size given in pixels. This value seems to be hard coded into the 4GL.
    * Based on testing, this value does not change when the default font size is changed.
    */
   private static final int DEFAULT_MINIMAL_SIZE_IN_PIXELS = 18;
   
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;

   /** Title color resolver */
   private GuiColorResolver gct;
   
   /** Font resolver - frame title. */
   private GuiFontResolver gf;
   
   /** Panel for the frame's title. */
   private BorderedPanelGuiImpl frameTitle;
   
   /** Cached line border. */
   private LineBorderGuiImpl lineBorder;
   
   /** Keeping track on explicit virtual width setting. */
   private boolean explicitVirtualWidth = false;
   
   /** Keeping track on explicit virtual height setting. */
   private boolean explicitVirtualHeight = false;
   
   /** The default minimal width of the empty frame. */
   private final double defaultMinimalWidth;
   
   /** The default minimal height of the empty frame. */
   private final double defaultMinimalHeight;
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public FrameGuiImpl(WidgetId id, FrameConfig cfg)
   {
      super(id, cfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // set frame default colors
      gc = new GuiColorResolver(gd, cfg, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      // box and title colors
      gc.boxColor     = gc.sysFgColor;
      gc.titleFgColor = gc.sysBgColor;
      gc.titleBgColor = gd.getSysColor(SysColor.COLOR_ACTIVECAPTION);
      
      // caption color resolver used by title border
      gct = new GuiColorResolver(gd, new BaseConfig(), 
                                 SysColor.COLOR_ACTIVECAPTION, SysColor.COLOR_WINDOWTEXT);
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // calculates the frame's default size from its pixels size
      defaultMinimalHeight = cc.heightFromNative(DEFAULT_MINIMAL_SIZE_IN_PIXELS);
      defaultMinimalWidth  = cc.widthFromNative(DEFAULT_MINIMAL_SIZE_IN_PIXELS);
   }
   
   /**
    * Compute the label width in character units, depending on the current font.
    * 
    * @param    label
    *           The label text.
    * @param    widget
    *           The widgeth to which the label belongs.
    *           
    * @return   See above.
    */
   @Override
   public double getLabelWidth(String label, LabeledWidget<GuiOutputManager> widget)
   {
      int font = resolveLabelFont(widget);
      int nativeWidth = getTextWidth(label, font);
      return screen.coordinates().widthFromNative(nativeWidth);
   }

   /**
    * Compute the label height in character units.
    * 
    * @param    labelText
    *           The label text.
    * @param    widget
    *           The widget to which the label belongs.
    * @param    topLabel
    *           Flag identifying that the label is shown above the widget.
    *           
    * @return   See above.
    */
   @Override
   public double getLabelHeight(String                          labelText, 
                                LabeledWidget<GuiOutputManager> widget, 
                                boolean                         topLabel)
   {
      int font = resolveLabelFont(widget);
      boolean text = true;

      if (topLabel)
      {
         // this is a top-label, get the default-fixed-font's height
         font = FontManager.DEFAULT_FIXED_FONT;
      }
      else
      {
         // if text mode is used, the label height is the actual font height
         text = (widget instanceof com.goldencode.p2j.ui.client.Text);
      }

      Window w = widget.window();
      int legacyHeight = FontManager.getFontHeight(w, font, screen().getInstanceDriver());
      if (!text)
      {
         boolean threeD = FontManager.isThreeD();
         legacyHeight = legacyHeight + (int) (legacyHeight / 2) + (threeD ? 2 : 0);;
      }
      
      return screen.coordinates().heightFromNative(legacyHeight);
   }

   /**
    * Get a delta which is added between frames, when their Y coordinate is automatically 
    * computed.
    * 
    * @return   A row height in char units computed by substracting 1 pixel from the 
    *           {@link FontManager#getPixelsPerRow} value.
    */
   @Override
   public double getRowDelta()
   {
      int ppr = FontManager.getPixelsPerRow(screen().getInstanceDriver());
      return cc.rowFromNative(ppr - 1);
   }
   
   /**
    * Get frame location relative to container.
    * 
    * @return  relative frame location.
    */
   @Override
   public Point location()
   {
      if (isDialog())
      {
         return new Point(0, 0);
      }
      
      return super.location();
   }

   /**
    * Get frame physical location relative to container, expressed in units native to UI type.
    * 
    * @return  relative frame location.
    */
   @Override
   public NativePoint physicalLocation()
   {
      if (isDialog())
      {
         return new NativePoint(0, 0);
      }
      
      return super.physicalLocation();
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, FrameConfig wcfg)
   {
      if (!wcfg.dialog)
      {
         // the window title of the dialog window is used for the frame title 
         frameTitle = new BorderedPanelGuiImpl();
         frameTitle.setColorResolver(gct);
         add(frameTitle);
      }
      
      if (wcfg.dialog)
      {
         // if no dialog window has been created, create it here
         DialogBoxWindow dialogWindow = new DialogBoxWindow(wcfg.title);
            
         // realize early, the window may be needed even before
         // it is shown for the first time
         dialogWindow.realize();
         
         // assign the window id so it is attached in super.initialize()
         wcfg.windowID = dialogWindow.getId().asInt();
      }
      
      // in case of dialogs, this sets the wcfg.windowID to its "parent" window (the window on
      // which the dialog-frame is attached).
      super.initialize(id, wcfg);
      
      // dialog boxes with explicit size assignment needs the internal frame size to be adjusted
      // because the dialog size includes enclosing window title and border values
      if (isDialog() && getOriginalHeightChars() != 0)
      {
         DialogBoxWindow wnd = (DialogBoxWindow) topLevelWindow();
         if (wnd != null)
         {
            Border<?> b = wnd.getBorder();
            WindowTitleBar tbar = wnd.getTitleBar();
            if (b != null && tbar != null)
            {
               Insets ins = b.insets();
               FrameConfig fcfg = config();
               fcfg.clientWidthChars = fcfg.widthChars;
               fcfg.clientHeightChars = fcfg.heightChars;
               fcfg.widthChars  -= ins.left + ins.right;
               fcfg.heightChars -= ins.top + ins.bottom + tbar.height();
               originalHeightChars = config().heightChars;
            }
         }
      }
      
      // the frame's title uses the window title font
      LabelConfig tcfg = new LabelConfig();
      tcfg.font = FontManager.WINDOW_TITLE_FONT;
      gf = new GuiFontResolver(window(), gd, tcfg, false);
      
      configScrollPane();
      
      configContentPanels();
      
      addFocusListener(new FocusListener()
      {
         @Override
         public void onFocusLost(FocusEvent event)
         {
            repaintCaption();
         }
         
         @Override
         public void onFocusGained(FocusEvent event)
         {
            repaintCaption();
         }
      });
      
      // save the frame's top border
      FrameConfig cfg = config();
      
      if (cfg.dialog || cfg.title != null)
      {
         double titleHeight = 0;
         if (cfg.dialog)
         {
            DialogBoxWindow dialogWindow = (DialogBoxWindow) topLevelWindow();

            // can be missing in the web client when in the embedded mode
            if (dialogWindow.hasTitlebar())
            {
               titleHeight = dialogWindow.getTitleBar().height();
            }
         }
         else
         {
            titleHeight = captionHeight();
         }
         
         cfg.titleHeightPixels = screen.coordinates().heightToNative(titleHeight);
      }
   }
   
   /**
    * Handle change of frame visibility.
    * 
    * @param   visible
    *          New visibility state. <code>true</code>- show frame,
    *          <code>false</code>- hide frame.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (isDialog())
      {
         if (visible)
         {
            topLevelWindow().show();
         }
         else
         {
            topLevelWindow().hide();
         }
      }
      
      super.setVisible(visible);
   }
   
   /**
    * Returns the size of the decoration elements for dialog-box.  It includes the window border
    * total size from both sides and the size of the dialog title.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getDialogDecorationCharSize()
   {
      Dimension resDim = new Dimension(0.0, 0.0);
      
      if (isDialog())
      {
         DialogBoxWindow wnd = (DialogBoxWindow) topLevelWindow();
         // frame border first
         if (wnd != null && wnd.hasDecorations())
         {
            Border<?> b = wnd.getBorder();
            Insets ins = b.insets();
            resDim.height = ins.top + ins.bottom;
            resDim.width = ins.left + ins.right;
         }

         // title bar height next
         WindowTitleBar tbar = wnd.getTitleBar();
         if (tbar != null)
         {
            resDim.height += tbar.height();
         }
      }
      
      return resDim;
   }
   
   /**
    * Process frame layout.
    * 
    * @param   suppressDrawing
    *          If <code>true</code> then drawing is suppressed even if frame
    *          is visible.
    */
   @Override
   protected void doLayoutWorker(boolean suppressDrawing)
   {
      super.doLayoutWorker(suppressDrawing);
      
      if (frameTitle != null)
      {
         frameTitle.setHeight(captionHeight());
      }
      updateScrollPaneSize();
      
      new StackLayout<GuiOutputManager>(Direction.VERTICAL).doLayout(this);
   }
   
   /**
    * Realize the frame, only if it's not yet realized. It has been overridden in order to set
    * the default size for an empty frame according to the Progress 4GL GUI look and feel.
    * 
    * @param   realize
    *          Flag indicating if the frame should be realized.
    */
   @Override
   public void realizeFrame(boolean realize)
   {
      boolean horizBarQuirk = false;
      if (realize       &&
          !isRealized() &&
          (currentStatement == UIStatement.ENABLE || currentStatement == UIStatement.DISPLAY) &&
          !isRootFrame())
      {
         horizBarQuirk = true;
      }
      
      super.realizeFrame(realize);
      if (realize)
      {
         if (horizBarQuirk)
         {
            activateHorizontalBarQuirk();
         }
         
         double currentWidth  = width();
         double currentHeight = height();
         if (currentWidth < defaultMinimalWidth || currentHeight < defaultMinimalHeight)
         {
            currentWidth  = Math.max(currentWidth,  defaultMinimalWidth);
            currentHeight = Math.max(currentHeight, defaultMinimalHeight);
            setOuterSize(currentWidth, currentHeight);
         }
      }
   }
   
   /**
    * Get the color used to draw the frame background.
    * 
    * @return   See above.
    */
   ColorRgb getFrameBgColor()
   {
      ScrollPaneGuiImpl scroll = (ScrollPaneGuiImpl) frameScroll;
      scroll.gc.colors();
      
      return scroll.gc.bgColor;
   }

   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per frame.
    * 
    * @param   origin
    *          The coordinate of the upper-left corner for peer container in
    *          character units.
    * @param   size
    *          The size of the peer container in character units.
    */
   @Override
   public void draw(Point origin, Dimension size)
   {
      if (!isVisible())
      {
         return;         
      }
      
      // determine frame colors
      gc.colors();
      
      // setup title
      boolean hasTitle = hasTitle();
      
      if (hasTitle)
      {
         gct.boxColor = gc.sysFgColor;
         
         // frame title colors
         if (hasFocus() && isRootFrame())
         {
            gct.bgColor = gd.getSysColor(SysColor.COLOR_ACTIVECAPTION);
            gct.fgColor = gd.getSysColor(SysColor.COLOR_CAPTIONTEXT);
         }
         else
         {
            gct.bgColor = gd.getSysColor(SysColor.COLOR_INACTIVECAPTION);
            gct.fgColor = gd.getSysColor(SysColor.COLOR_INACTIVECAPTIONTEXT);
         }
         
         gct.titleBgColor = gct.bgColor;
         gct.titleFgColor = gct.fgColor;
      }
      
      if (frameTitle != null)
      {
         frameTitle._setVisible(hasTitle);
      }
      
      // update box, frame in dialog-box never draws its border 
      ScrollPaneGuiImpl scroll = (ScrollPaneGuiImpl) frameScroll;
      scroll.setBorder(hasBox() ? lineBorder : null);
   
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      NativeRectangle clip = clipRectangle(p, d);

      final ThinClient tc = ThinClient.getInstance();
      final Point z = getContentPane().screenLocation();
      z.y += headersHeight;

      gd.draw(p, clip, () ->
      {
         if (frameTitle != null)
         {
            frameTitle.draw();
         }
         
         if (frameScroll != null)
         {
            if (!down)
            {
               frameScroll.draw();
            }
            else
            {
               if (tc.viewRefresh)
               {
                  frameScroll.draw(nonbody);
               }
               
               // GUI frames can't be used in redirected mode!
               drawDownFrame(z, false);
            }
         }
      });
   }
   
   /**
    * Destroy this widget.
    */
   @Override
   public void destroy()
   {
      if (isDialog())
      {
         topLevelWindow().destroy();
      }
      
      super.destroy();
   }
   
   /**
    * Calculate frame height.
    * @param   d
    *          Frame dimensions.
    * @param   delta
    *          Height delta.
    */
   protected void calculateHeight(Dimension d, double delta)
   {
      if (!explicitVirtualHeight)
      {
         super.calculateHeight(d, delta);
         // fix the calculated inner size
         double minimalInnerHeight = defaultMinimalHeight - calculateHeightDelta();
         if (d.height < minimalInnerHeight)
         {
            d.height = minimalInnerHeight;
         }
      }
      else if (config().virtualHeightChars != BaseConfig.INV_COORD)
      {
         d.height = config().virtualHeightChars;
         // explicit virtual size case for implicit dialog definition
         if (getOriginalHeightChars() == 0)
         {
            d.height -= getDialogDecorationCharSize().height;
         }
      }
      else
      {
         super.calculateHeight(d, delta);
      }
   }
   
   /**
    * Adjusts the minimum frame width or height value depending on the current min value and
    * the presense of the box within the frame.
    * 
    * @param   minWidthOrHeight
    *          The size value to be adjusted.
    * @param   isHeight
    *          <code>TRUE</code> if the size value is height, <code>FALSE</code> if width.
    * 
    * @return  GUI dependent new width or height value.
    */
   @Override
   protected double adjustMinSize(double minWidthOrHeight, boolean isHeight)
   {
      // take initial
      double resWidthOrHeight = minWidthOrHeight;

      // calculate possible box size in character units
      // assume 2 pixels for box size in GUI
      double boxSize = isHeight ? screen.coordinates().heightFromNative(2)
                                : screen.coordinates().widthFromNative(2);
      
      // consider if we have a boxed frame or not, the box size is 1.0 for each side
      if (resWidthOrHeight == 0)
      {
         resWidthOrHeight = hasBox() ? 2.0 * boxSize : boxSize;
      }
      else
      {               
         resWidthOrHeight += hasBox() ? boxSize : 0.0;
      }
      
      return resWidthOrHeight;
   }
    
   /**
    * Draw the body of a down frame.
    * 
    * @param    top
    *           The location from where to start drawing.
    * @param    redir
    *           Flag indicating if redirected mode is active.
    *           
    * @return   <code>true</code> if the screen needs to be synchronized.
    */
   @Override
   protected boolean drawDownFrame(final Point top, final boolean redir)
   {
      // get container coordinates
      NativePoint p = frameScroll.viewport().physicalLocation();
      p = p.translate(frameScroll.physicalLocation());
      NativeDimension d = frameScroll.viewport().physicalDimension();
      
      NativeRectangle clip = clipRectangle(p, d);
      
      gd.draw(p, clip, () -> FrameGuiImpl.super.drawDownFrame(top, redir));
      
      return false;
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * Each container should search for own widgets. Because widgets are relative positioned 
    * inside containers the container physical position is adjusted before search.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or {@code null} if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      // the point incoming is related to frame origin, we need to substract the frame title
      NativePoint pCheck = hasTitle() ? p.minus(frameScroll.physicalLocation()) : p;
      
      Widget<GuiOutputManager> foundWidget = frameScroll.findMouseSource(pCheck);
      
      if (foundWidget == null)
      {
         foundWidget = (this.physicalBounds().contains(p) ? this : null);
      }
      
      // if the widget is a label, the "source" will be the frame scroll in order to prevent
      // exit and enter mouse events for the frame if the mouse pointer is over the label.
      if (foundWidget instanceof LabelGuiImpl)
      {
         return frameScroll;
      }
      
      // if the widget is not enabled, then the source is the frame, too
      if (foundWidget != null                              &&
          foundWidget.config() != null                     &&
          !WidgetId.virtualWidget(foundWidget.config().id) &&
          !foundWidget.isEnabled())
      {
         return this;
      }
      
      return foundWidget;
   }

   /**
    * Get a point to represent the LAST-EVENT:X and LAST-EVENT:Y coordinates.  For frames, these
    * represents a point relative to the frame's top-left corner (excluding title).
    * 
    * @param    point
    *           The point (relative to the window's top-left corner) for which the coordinates for
    *           the portable mouse events are needed.
    * @return   A point representing the coordinates of a mouse-portable event or 
    *           <code>null</code> if such an event can't be raised.
    */
   @Override
   public NativePoint getPortableMouseEventCoordinates(NativePoint point)
   {
      Viewport<GuiOutputManager> vp = frameScroll.viewport();
      NativePoint vpLoc = vp.screenPhysicalLocation();
      NativeDimension vpDim = vp.physicalDimension();
      NativeRectangle vpBounds = new NativeRectangle(vpLoc, vpDim);
      
      // exclude scrollbars
      if (!vpBounds.contains(point))
      {
         return null;
      }
      
      NativePoint contentLoc = contentPane.physicalLocation();
      
      // the frame's title does not raise mouse events
      
      if (hasTitle() && frameTitle != null)
      {
         NativePoint np = point.minus(this.screenPhysicalLocation());
         
         NativeDimension nd = frameTitle.physicalDimension();
         nd.height++;
         NativeRectangle titleBounds = new NativeRectangle(new NativePoint(0, 0), nd);
         if (titleBounds.contains(np))
         {
            return null;
         }
         
         point = point.minus(frameTitle.screenPhysicalLocation());
         point.y -= frameTitle.physicalDimension().height;
         
         point = point.minus(contentLoc);
         
         return point;
      }
      
      return point.minus(this.screenPhysicalLocation()).minus(contentLoc);
   }

   /**
    * Processing for default button in a frame.
    *
    * @return <code>TRUE</code> if event be re-routed to default button,
    *         <code>FALSE</code> otherwise.
    */
   @Override
   public boolean processDefaultButton(KeyInput event)
   {
      // scan the event source
      Widget src = event.source();

      if (src != this && config().defaultButtonId != -1 && event.keyCode() == Key.VK_ENTER)
      {
         // check if the source is the field-level widget but not button
         if (src instanceof FillIn            ||
             src instanceof SelectionList     ||
             src instanceof SelectionListBody ||
             src instanceof RadioSet          ||
             src instanceof ComboBox          ||
             src instanceof ToggleBox         ||
             (src instanceof Editor && !(((EditorConfig) ((Editor) src).config()).returnInserted)))
         {
            // calculate new event target
            Widget<?> bTarget =
               OutputManager.resolveWidget(config().defaultButtonId);
   
            if (!(bTarget instanceof Button))
            {
               UnimplementedFeature.todo("Non-button widget as DEFAULT-BUTTON");
            }
   
            // despite of the 4GL documentation declares the button default attribute must also be
            // set to make this work, the real application tests show the button option should be
            // ignored, the only that makes sense is the DEFAULT-BUTTON frame's option.
            if (bTarget != null)
            {
               // route enter key to default button
               bTarget.processKeyEvent(event);
               return true;
            }
         }
      }

      // not field level widget
      return false;
   }
   
   /**
    * Processing for cancel button in a frame.
    *
    * @return <code>TRUE</code> if event be re-routed to cancel button,
    *         <code>FALSE</code> otherwise.
    */
   @Override
   public boolean processCancelButton()
   {
      if (config().cancelButtonId != -1)
      {
         // calculate new event target
         Widget<?> bTarget =
                     OutputManager.resolveWidget(config().cancelButtonId);

         if (!(bTarget instanceof Button))
         {
            UnimplementedFeature.todo("Non-button widget as DEFAULT-BUTTON");
         }

         if (bTarget != null)
         {
            // route CHOOSE event to cancel button
            TriggerResult tr = ThinClient.getInstance().invokeTriggers(
                                    bTarget, Keyboard.eventCode("CHOOSE"), false
                                 );
            if (tr.executed && tr.consume)
            {
               return true;
            }
         }
      }
      return false;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(FrameConfig beforeUpdate)
   {
      FrameConfig fc = config();

      // check if the frame is not realized and it's HIDDEN attribute was just set: if so, 
      // enable the "nested frame scroll 4GL bug": the parent frame will not show a horiz scroll,
      // if the child FRAME:HIDDE = FALSE is executed.
      if (!beforeUpdate.realized && !fc.hidden && fc.hiddenSet && !isRootFrame())
      {
         activateHorizontalBarQuirk();
      }
      
      // we need to sync virtual size with regular size
      if (!explicitVirtualWidth && fc.realized)
      {
         explicitVirtualWidth = beforeUpdate.virtualWidthChars != fc.virtualWidthChars;
      }
      if (!explicitVirtualHeight && fc.realized)
      {
         explicitVirtualHeight = beforeUpdate.virtualHeightChars != fc.virtualHeightChars;
      }
      
      if (fc.virtualWidthChars >= 0.0 && fc.virtualWidthChars < fc.clientWidthChars)
      {
         fc.clientWidthChars = fc.virtualWidthChars;
         if (explicitVirtualWidth)
         {
            Dimension decorDim = getDialogDecorationCharSize();
            fc.widthChars -= decorDim.width;
         }
      }
      if (fc.virtualHeightChars >= 0.0 && fc.virtualHeightChars < fc.clientHeightChars)
      {
         fc.clientHeightChars = fc.virtualHeightChars;
         if (explicitVirtualHeight)
         {
            Dimension decorDim = getDialogDecorationCharSize();
            fc.heightChars -= decorDim.height;
         }
      }
      
      super.afterConfigUpdate(beforeUpdate);
      
      // TODO: The size of frameScroll should be left uninitialized so that 
      //       always inherits the size of its parent (frame). Instead
      //       make sure c.widthChars and c.heightChars are propagated
      //       to the frame.size field.
      // need to sync the frame scroll to in several conditions  
      updateScrollPaneSize();
      
      if (isDialog())
      {
         Point originalLocation = ConfigHelper.getAssignedLocation(beforeUpdate);
         Point location = ConfigHelper.getAssignedLocation(fc);
         
         if (!Objects.equals(originalLocation, location))
         {
            DialogBoxWindow wnd = (DialogBoxWindow) topLevelWindow();
            wnd.onFrameLocationChanged();
         }
      }

      if (frameScroll != null)
      {
         frameScroll.config().sysbgcolor = isThreeD() ? SysColor.COLOR_3DFACE 
                                                      : SysColor.COLOR_WINDOW;
         frameScroll.config().bgcolor = fc.bgcolor;
      }
      
      if (isDialog() && isDisplayed())
      {
         DialogBoxWindow wnd = (DialogBoxWindow) topLevelWindow();
         wnd.refreshColors(fc);
         // refresh dialog title if required
         if (fc.title != null && !fc.title.equals(beforeUpdate.title))
         {
            wnd.setTitle(fc.title);
         }
      }
      
      if (!fc.realized && beforeUpdate.windowID != fc.windowID && getRootFrame() == this)
      {
         // this is a root frame and its window is being set
         placeToWindow(fc.windowID, null, true);
      }
   }
   
   /**
    * Sets the size of the peer container. This is the visible size of the
    * container and can not be the same as the inner size if the container is
    * scrollable.
    * 
    * @param   width
    *          The new width of the container in character units.
    * @param   height
    *          The new height of the container in character units.
    */
   @Override
   public void setOuterSize(double width, double height)
   {
      super.setOuterSize(width, height);
      
      if (isDialog())
      {
         DialogBoxWindow wnd = parent(DialogBoxWindow.class);
         if (wnd != null)
         {
            wnd.onFrameDimensionChanged();
         }
      }
   }
   
   /**
    * Setting the inner size of the container.
    * 
    * @param   innerWidth
    *          The new inner width of the container in character units.
    * @param   innerHeight
    *          The new inner height of the container in character units.
    */
   @Override
   public void setInnerSize(double innerWidth, double innerHeight)
   {
      if (isDialog())
      {
         double w = ConfigHelper.getEffectiveVirtualWidth(config()); 
         double h = ConfigHelper.getEffectiveVirtualHeight(config());
           
         double newVirtWidthChars = innerWidth;
         double newVirtHeightChars = innerHeight;

         // the frame may not be displayed  
         DialogBoxWindow wnd = parentOrSelf(DialogBoxWindow.class);
         
         if (wnd != null)
         {
            if (wnd.hasDecorations())
            {
               // subtract window border size if necessary
               Border<?> b = wnd.getBorder();
               if (b != null)
               {
                  Insets ins = b.insets();
                  if (w != BaseConfig.INV_COORD)
                  {
                     innerWidth -= ins.left + ins.right;
                     if (newVirtWidthChars != w)
                     {
                        newVirtWidthChars += ins.left + ins.right;
                     }
                  }
               
                  if (h != BaseConfig.INV_COORD)
                  {
                     innerHeight -= ins.top + ins.bottom;
                     if (newVirtHeightChars != h)
                     {
                        newVirtHeightChars += ins.top + ins.bottom;
                     }
                  }
               }
            }
   
            // subtract dialog title height
            WindowTitleBar tbar = wnd.getTitleBar();
            if (tbar != null && h != BaseConfig.INV_COORD)
            {
               innerHeight -= tbar.height();
               if (newVirtHeightChars != h)
               {
                  newVirtHeightChars += tbar.height();
               }
            }
         }

         // the virtual size may need to be adjusted for implicit change
         if (newVirtWidthChars != w)
         {
            config().virtualWidthChars =
               newVirtWidthChars + cc.widthFromNative(Frame.EXTRA_DLG_PIXELS);
            // for implictit virtual sizing sync the inner size
            if (!explicitVirtualWidth)
            {
               innerWidth = config().virtualWidthChars;
            }
         }
         if (newVirtHeightChars != h)
         {
            config().virtualHeightChars =
               newVirtHeightChars + cc.heightFromNative(Frame.EXTRA_DLG_PIXELS);
            // for implictit virtual sizing sync the inner size
            if (!explicitVirtualHeight)
            {
               innerHeight = config().virtualHeightChars;
            }
         }

         // another special case to adjust inner dimension when both size and virtual size
         // was explicitly set         
         if (getOriginalHeightChars() != 0 && explicitVirtualWidth)
         {
            innerWidth = config().virtualWidthChars;
         }
         if (getOriginalHeightChars() != 0 && explicitVirtualHeight)
         {
            innerHeight = config().virtualHeightChars;
         }
      }
         
      super.setInnerSize(innerWidth, innerHeight);
   }

   /**
    * The method returns <code>true</code> when the frame should have a box around.
    * Note that the returned value may differ from {@link FrameConfig#box} as 
    * for example dialog boxes do not draw a box even when BOX set to true.
    * 
    * @return  See above.
    */
   @Override
   public boolean hasBox()
   {
      return !isDialog() && super.hasBox();
   }
   
   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Get the current custom mouse pointer if it is set
    * 
    * @return the current custom mouse pointer if it is set  
    *        
    */
   @Override
   public Optional<MousePtrWrapper> getCustomMousePointer()
   {
      return mousePtr.isPresent() ? mousePtr : 
         window().getCustomMousePointer();
   }

   /**
    * Create handler for processing mouse hover events
    * 
    * @return handler for processing mouse hover events
    */
  @Override
   protected MouseHoverAction createMouseHoverAction()
   {
      // frame scroll component is the smallest widget that catch the mouse input and has an id. 
      return new MouseHoverAction(this, this.frameScroll.getId().asInt())
      {
         @Override
         protected MousePtrWrapper parentCursor()
         {
            return window().getEffectiveMousePointer();
         }
      };
   }
   
   /**
    * Adjust frame width.
    * 
    * @param   width
    *          Input width.
    * @param   screenWidth
    *          Screen height. 
    *          
    * @return  adjusted height.
    */
   @Override
   protected double adjustWidth(double width, double screenWidth)
   {
      if (isDialog())
      {
         // do not adjust width when dialog
         return width;
      }
      
      return super.adjustWidth(width, screenWidth);
   }
   
   /**
    * Draw underline for specified widget.
    * 
    * @param   comp
    *          The base component to draw underline for.
    */
   @Override
   protected void drawUnderline(Widget<GuiOutputManager> comp)
   {
      // TODO: implement this
   }

   /**
    * Reset colors.
    * 
    * @param   config
    *          Widget configuration.
    */
   protected void resetColors(BaseConfig config)
   {
      // TODO should be tested
      config.bgcolor = -1;
      config.fgcolor = -1;
   }

   /**
    * Compare colors.  This method must be implemented in subclasses and provide
    * specific implementation as follow:
    * <p>
    * ChUI implementation must compare dcolor and pfcolor.
    * <p>
    * GUI implementation must compare bgcolor and fgcolor.
    * 
    * @param   config
    *          Widget configuration.
    * @param   colors
    *          Colors to be compared.
    *          
    * @return  The colors structure after comparison.         
    */
   @Override
   protected Colors compareColors(BaseConfig config, Colors colors)
   {
      colors.bgEquals = colors.bgcolor == config.bgcolor;
      colors.fgEquals = colors.fgcolor == config.fgcolor;
      return colors;
   }

   /**
    * Calculate frame height delta.
    * 
    * @return height delta.
    */
   @Override
   protected double calculateHeightDelta()
   {
      return super.calculateHeightDelta() + captionHeight();
   }
   
   /**
    * The method returns the maximum frame size according to the virtual size 
    * of the enclosing frame or according to the workspace size of its parent 
    * window.
    *  
    * @return   See above.
    */
   @Override
   protected Dimension getMaxFrameSize()
   {
      Frame<?> f = parent(Frame.class);
      if (f != null)
      {
         FrameConfig fc = f.config();
         double w = ConfigHelper.getEffectiveVirtualWidth(fc); 
         double h = ConfigHelper.getEffectiveVirtualHeight(fc);
                  
         return new Dimension(w, h);
      }
      else
      {
         GuiWindow wnd = parent(GuiWindow.class);
         return wnd.getMaxWorkspaceSize();
      }
   }

   /**
    * Set focus to specified widget.
    * <p>
    * Post a repaint for the frame's title, too, to update the frame focus state.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   @Override
   public void setFocus(Widget<GuiOutputManager> focus)
   {
      super.setFocus(focus);
      
      repaintCaption();
   }
   
   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget<GuiOutputManager> defaultFocus()
   {
      // if the frame has no focusable widgets inside - just return the frame itself
      // we have no (at least now) frames inside the frame
      return hasFocusable(null) ? super.defaultFocus() : this;
   }
   
   /**
    * Get the bounds of the frame's content.
    * 
    * @return   See above.
    */
   protected Rectangle frameContentBounds()
   {
      NativeRectangle bounds = frameScroll.physicalBounds();
      
      int captionHeight = cc.heightToNative(captionHeight());
      bounds = bounds.translate(0, (captionHeight == 0 ? 0 : -captionHeight - 1));
      
      return cc.rectangleFromNative(bounds);
   }
   
   /**
    * Raise a {@link PaintEvent} for the frame's title, as needed.
    */
   private void repaintCaption()
   {
      if (!hasTitle() || gf == null || !paintable || configUpdateActive)
      {
         return;
      }
      
      double height = captionHeight();
      double width = width();
      Point loc = location();
      
      repaint(new Rectangle(loc, new Dimension(width, height),
                            screen().coordinates().baseUnits()));
   }
   
   /***
    * Get the frame's caption height, in character units.
    * 
    * @return   See above.
    */
   @Override
   protected double captionHeight()
   {
      if (!hasTitle() || gf == null || isDialog())
      {
         return 0;
      }
   
      // NOTE: actual height is 20, but the South line is cut (or overlapped by frame content) 
      return screen.coordinates().heightFromNative(20 - 1); 
   }

   /**
    * Creates the frame's scroll container.
    *
    * @return  Returns an instance of {@linkplain SensitiveScrollContainer}.
    */
   protected SensitiveScrollContainer<GuiOutputManager> createScrollContainer()
   {
      return new GuiScrollContainer();
   }

   /**
    * Resolves the effective font for a label of the supplied labeled widget.
    *
    * @param   lwidget
    *          A labeled widget reference.
    *
    * @return  See above.
    */
   private static int resolveLabelFont(LabeledWidget<GuiOutputManager> lwidget)
   {
      Label<GuiOutputManager> lbl = lwidget.getLabelInstance();

      BaseConfig cfg = lbl != null ? lbl.config() : null;

      // if the widget's label has not set the font explicitly use the font assigned
      // to its frame,
      // note that the label doesn't use the font set to its labeled widget or the window
      if (cfg == null)
      {
         Frame f = UiUtils.locateFrame(lwidget);
         if (f != null)
         {
            cfg = f.config();
         }
      }

      int font = FontManager.DEFAULT_FONT;
      if (cfg != null)
      {
         GuiDriver gd = (GuiDriver) lwidget.screen().getInstanceDriver();
         GuiFontResolver gflbl = new GuiFontResolver(lwidget.window(), gd, cfg);

         // this call will attempt to resolve the label's frame font if the label font not
         // assigned
         font = gflbl.resolveFontNum();
      }

      return font;
   }

   /**
    * Configure frame scroll pane
    */
   @SuppressWarnings("rawtypes")
   private void configScrollPane()
   {
      ScrollPaneGuiImpl fs = (ScrollPaneGuiImpl) this.frameScroll;
      fs.setShowBars(ShowBars.AS_NEEDED);
      fs.config().sysbgcolor = isThreeD() ? SysColor.COLOR_3DFACE : SysColor.COLOR_WINDOW;
      fs.config().bgcolor = config().bgcolor;
      fs.setScrollBars(ScrollBar.Bars.BOTH);
      // register scroll pane and its widgets for mouse events
      WidgetRegistry<GuiOutputManager> reg = screen().getRegistry();
      reg.addWidget(fs);
      reg.addWidgetsRecursive(fs.widgets());
      
      // no border for dialogs
      lineBorder = config().dialog ? null : new LineBorderGuiImpl();
   }
   
   /**
    * Configure frame scroll pane
    */
   private void configContentPanels()
   {
      if (frameTitle == null)
      {
         return;
      }
      
      // if frame title exists we must decrease scroll pane height
      updateScrollPaneSize();
      
      // draw title panel
      frameTitle.setCustomDraw(() -> {
         NativeDimension d = frameTitle.physicalDimension();
         ThemeManager.getCurrentTheme().drawFrameTitle(this, gd, gc, gf, 0, 0, d.width, d.height);
      });
      
      double baseY = cc.baseUnits().y;
      double baseX = cc.baseUnits().x;
      Border<GuiOutputManager> border = new LineBorderGuiImpl(new Insets(baseY, baseX, 0, baseX),
                                                              gct);
      frameTitle.setBorder(border);
   }
   
   /**
    * Activate the horizonta scrollbar 4GL bug for all ancestor frames, except the root one.
    */
   private void activateHorizontalBarQuirk()
   {
      FrameGuiImpl parent = (FrameGuiImpl) ThinClient.getInstance().getFrame(config().frameId);
      while (!parent.isRootFrame())
      {
         if (!parent.isRealized())
         {
            ((ScrollPaneGuiImpl) parent.frameScroll).setHorizBarQuirk(true);
         }
         
         parent = (FrameGuiImpl) ThinClient.getInstance().getFrame(parent.config().frameId);
      }
   }

   /**
    * Updates frame scroll pane size with possible frame title.
    */
   private void updateScrollPaneSize()
   {
      FrameConfig c = config();
      // subtract frame title height if frame has size at this time
      if (frameScroll != null && c.widthChars > 0.0 && c.heightChars > 0.0)
      {
         frameScroll.setSize(c.widthChars, c.heightChars - captionHeight());
      }
   }

   /**
    * The implementation of frame's scrollable container.
    */
   protected class GuiScrollContainer
   extends ScrollContainer
   {
      /**
       * Draw widget.
       */
      @Override
      public void draw()
      {
         NativePoint p = physicalLocation();
         NativeDimension d = physicalDimension();
         NativeRectangle clip = clipRectangle(p, d);
         
         // make sure the draw origin is properly translated
         gd.draw(p, clip, super::draw);
      }
   }
}