/*
** Module   : GuiColorResolver.java
** Abstract : GUI widget color resolver
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20141022 Created initial version.
** 002 GES 20150425 Removed AWT dependencies and shifted to ColorRgb.
** 003 SBI 20160324 Added resolved colors listeners.
** 004 CA  20161014 A null config will force to use the default colors.
** 005 CA  20170131 Fixed color inheritance when not explicitly set at widget but maybe set at 
**                  frame (and frame might or might not be THREE-D). 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.beans.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.SysColor;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * Resolve widget color.
 */
public class GuiColorResolver
{
   /** Screen driver. */
   private GuiDriver gd;
   
   /** Widget configuration. */
   private  BaseConfig config;
   
   /** BGCOLOR. */
   public ColorRgb bgColor; 

   /** FGCOLOR. */
   public ColorRgb fgColor; 

   /** Default BGCOLOR. */
   public ColorRgb sysBgColor; 

   /** Default FGCOLOR. */
   public ColorRgb sysFgColor; 

   /** Frame box color. */
   public ColorRgb boxColor; 
      
   /** Title text color. */
   public ColorRgb titleFgColor; 

   /** Title background color. */
   public ColorRgb titleBgColor;

   /** BGCOLOR  selection. */
   public ColorRgb selectionBgColor;
   
   /** FGCOLOR  selection. */
   public ColorRgb selectionFgColor;
   
   /** Thumb button color. */
   public ColorRgb thumbColor;
   
   /** Miscellaneous color. */
   public ColorRgb miscColor;

   /** The resolved colors are cached */
   private final Map<Integer, ColorRgb> resolvedColors;

   /** The listeners are interested in colors change events */
   private final List<PropertyChangeListener> listeners;
   
   /**
    * Constructor.
    * 
    * @param    gd
    *           GUI screen driver.
    * @param    config
    *           Widget configuration.
    * @param    defBgColor
    *           Default system BGCOLOR
    * @param    defFgColor
    *           Default system FGCOLOR
    */
   public GuiColorResolver(GuiDriver gd, BaseConfig config, int defBgColor, int defFgColor)
   {
      this.gd = gd;
      this.config = config;
      
      sysBgColor = gd.getSysColor(defBgColor);
      sysFgColor = gd.getSysColor(defFgColor);
      
      resolvedColors = new HashMap<Integer, ColorRgb>();
      listeners = new LinkedList<PropertyChangeListener>();
   }
      
   /**
    * Resolve widget FGCOLOR and BGCOLOR.
    */
   public void colors()
   {
      // BGCOLOR
      int bgcolor = config == null ? -1 : config.bgcolor;
      
      if (bgcolor == -1)
      {
         bgColor = null;

         // not explicitly defined; inherited from the frame
         if (!(config instanceof FrameConfig) && 
             config != null                   && 
             config.frameId != -1)
         {
            FrameGuiImpl frame = (FrameGuiImpl) ThinClient.getInstance().getWidget(config.frameId);
            if (frame != null && 
                (frame.config().bgcolor != -1 || 
                 (!(config instanceof BrowseConfig) && !config.enabled)))
            {
               // disabled browse does not inherit implicit frame color
               ColorRgb color = frame.getFrameBgColor();
               bgColor = color;
            }
         }
         
         if (bgColor == null)
         {
            int sysbgcolor = config == null ? -1 : config.sysbgcolor;
            bgColor = (sysbgcolor == -1) ? sysBgColor : gd.getSysColor(sysbgcolor);
         }
      }
      else
      {
         bgColor = gd.getBgColor(bgcolor);
      }
      
      // FGCOLOR
      int fgcolor = config == null ? -1 : config.fgcolor;
      
      if (fgcolor == -1)
      {
         int sysfgcolor = config == null ? -1 : config.sysfgcolor;
         fgColor = (sysfgcolor == -1) ? sysFgColor : gd.getSysColor(sysfgcolor);         
      }
      else
      {
         fgColor = gd.getFgColor(fgcolor);
      }
      
      notifyColorsResolved();
   }

   /**
    * Add the registered listener.
    * 
    * @param    listener
    *           The property change listener
    */
   public void addListener(PropertyChangeListener listener)
   {
      int index = listeners.indexOf(listener);
      if (index == -1)
      {
         listeners.add(listener);
      }
   }

   /**
    * Remove the registered listener.
    * 
    * @param    listener
    *           The property change listener
    */
   public void removeListener(PropertyChangeListener listener)
   {
      int index = listeners.indexOf(listener);
      if (index > -1)
      {
         listeners.remove(index);
      }
   }

   /**
    * Notify listeners if the resolved colors are changed.
    */
   private void notifyColorsResolved()
   {
      if (listeners.isEmpty())
      {
         return;
      }
      
      ColorRgb oldBgColor = resolvedColors.get(SysColor.COLOR_BACKGROUND);
      if (bgColor.equals(oldBgColor))
      {
         return;
      }
      
      for (PropertyChangeListener listener : listeners)
      {
         listener.propertyChange(new PropertyChangeEvent(this, "bgcolor", oldBgColor, bgColor));
      }
      
      resolvedColors.put(SysColor.COLOR_BACKGROUND, bgColor);
   }
}