/*
** Module   : GuiKeyboard.java
** Abstract : Keyboard configuration for Windows GUI clients. 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141029 Created initial version.
** 002 CA  20141220 Added mapping for COPY/CTRL-C key function.
** 003 GES 20150708 Added alternative mappings for CUT/COPY/PASTE key functions.
** 004 VIG 20150916 Added keyboard layout support.
** 005 SBI 20151011 Added SHIFT + DEL key label to return 639 as its key code.
** 006 CA  20161101 Added API to check if a key can be used as a mnemonic.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.*;

import com.goldencode.p2j.ui.client.chui.*;

/**
 * This class contains the configurations for a Keyboard compatible with the Windows GUI clients.
 */
public class GuiKeyboard
extends ChuiWindowsKeyboard
{
   /** Windows keyboard layout from driver. */ 
   private KeyboardLayout layout = new USAKeyboardLayout();

   /**
    * Added SHIFT + DEL key
    */
   public GuiKeyboard()
   {
      super();
      extendedKeys.put(639, "SHIFT-DEL"); // since code 127 has DEL as a label
   }

   /**
    * Initialize this keyboard.
    */
   @Override
   public void init()
   {
      super.init();
      
      layout.init();
   }
   
   /**
    * Check if the specified key can behave as a mnemonic activator for widgets (not menus).
    * Only GUI clients can handle label mnemonics - a key is considered a mnemonic activator if 
    * and only if the ALT modifier is present (bit 10 in this key code).
    * 
    * @param    key
    *           The key code to be checked.
    *           
    * @return   <code>true</code> if the key can activate a mnemonic, for widget's labels.
    */
   @Override
   protected boolean mnemonic(int key)
   {
      // bit 10 is set for ALT keys
      return (key & 0b0100_0000_0000) == 0b0100_0000_0000;
   }

   /**
    * Get key name from a keyboard or null if not found.
    *
    * @param   code
    *          Progress key code.
    *         
    * @return  Name of a key button or null if the key name is the same as Progress key label or
    *          specified code is incorrect (not a single key code or invalid value).
    */
   public String keyName(int code)
   {
      if (layout != null)
      {
         return layout.getKeyName(code);
      }
      
      return null;
   }
   
   /**
    * Get the mappings of standard functions to their associated keys.
    * 
    * @return   See above.
    */
   @Override
   protected Map<String, String[]> standardKeyFunctions()
   {
      Map<String, String[]> res = new HashMap<>();
      
      res.put(BACK_TAB,  new String[] {"SHIFT-TAB"});
      res.put(BACKSPACE, new String[] {"BACKSPACE"});
      res.put(BELL,      new String[] {"CTRL-G", "BELL"});

      res.put(CLOSE,        new String[] {"F8"});
      res.put(COMPILE,      new String[] {"SHIFT-F2"});
      res.put(CURSOR_DOWN,  new String[] {"CURSOR-DOWN"});
      res.put(CURSOR_LEFT,  new String[] {"CURSOR-LEFT"});
      res.put(CURSOR_RIGHT, new String[] {"CURSOR-RIGHT"});
      res.put(CURSOR_UP,    new String[] {"CURSOR-UP"});  
      res.put(COPY,         new String[] {"CTRL-C", "CTRL-INS"});
      res.put(CUT,          new String[] {"CTRL-X", "SHIFT-DEL"});
      
      res.put(DELETE_CHARACTER, new String[] {"DEL"});
      
      res.put(END,           new String[] {"END"});
      res.put(END_ERROR,     new String[] {"ESC"});
      res.put(ENTER_MENUBAR, new String[] {"ALT"});
      
      res.put(FIND,          new String[] {"CTRL-F"});
      res.put(FIND_NEXT,     new String[] {"F9"});
      res.put(FIND_PREVIOUS, new String[] {"SHIFT-F9"});
      
      res.put(GET,  new String[] {"F3"});
      res.put(GO,   new String[] {"F2"});
      res.put(GOTO, new String[] {"CTRL-G"});
      
      res.put(HELP, new String[] {"F1"});
      res.put(HOME, new String[] {"HOME"});
      
      res.put(INSERT_MODE, new String[] {"INSERT", "INS"});
      
      res.put(LEFT_END, new String[] {"HOME"});
      
      res.put(NEW,        new String[] {"SHIFT-F3"});
      res.put(NEXT_FRAME, new String[] {"F6"});
      
      res.put(PAGE_DOWN,  new String[] {"PAGE-DOWN", "PGDN", "NEXT-PAGE", "NEXT-SCRN"});
      res.put(PAGE_UP,    new String[] {"PAGE-UP", "PGUP", "PREV-PAGE", "PREV-SCRN"});
      res.put(PASTE,      new String[] {"CTRL-V", "SHIFT-INS"});
      res.put(PREV_FRAME, new String[] {"SHIFT-F6"});
      res.put(PUT,        new String[] {"F6"});
      
      res.put(RETURN, new String[] {"ENTER", "RETURN", "CTRL-M"});
      
      res.put(SAVE_AS, new String[] {"SHIFT-F6"});
      res.put(STOP,    new String[] {"CTRL-BREAK"});
      
      res.put(TAB,     new String[] {"TAB"});
      
      // special entries for key actions
      res.put(ENDKEY, new String[] {});
      res.put(ERROR,  new String[] {});

      return res;
   }
}