/*
** Module   : GuiWidgetFactory.java
** Abstract : a factory for GUI widgets
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140521 Created initial version.
** 002 CA  20140711 Register the new created window with the GUI driver.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added create IMAGE widget.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140917 Added support for scroll bar widget creation.
**                  Added support for scroll pane widget creation.
** 006 MAG 20141009 Deleted createMessageLine. InsertIndicator class has been deleted.
** 007 MAG 20141017 Added GUI frame factory method.
** 008 EVL 20141114 Adding rectangle constructor mapping.  Scroll gui container now creaing here.
** 009 CA  20141123 Removed TODO from createWindow.
** 010 EVL 20141121 Adding button constructor mapping.
** 011 EVL 20141208 Adding image constructor mapping.
** 012 CA  20141220 Added mappings for FILL-IN, LABEL, SKIP, TEXT.
** 013 CA  20150323 Labels are always created using specific APIs; these APIs require the frame to
**                  which the label will belong, initially.
** 014 VIG 20150506 ToggleBoxGuiImpl added to classMap.
** 015 EVL 20150421 Adding combo-box creation.
** 016 CA  20150509 Added EDITOR  creation.  Allow scroll container for any widget, not just frame.
** 017 HC  20150612 Implemented GUI ALERT-BOX.
** 018 CA  20150626 Added support for SPACE in GUI.
** 019 VIG 20150618 Registered Menu classes.
** 020 EVL 20150818 Adding support for selection list widget creation.
** 021 HC  20150831 Scrolling support overhaul.
** 022 EVL 20150923 Adding support for radio-set widget creation.
** 023 SVL 20150828 Added browse classes.
** 024 EVL 20160127 Adding support for slider widget creation.
** 025 IAS 20160217 Added new flavor for the createScrollPane method.
** 026 IAS 20160315 Support for the DEBUG-ALERT attribute
** 027 CA  20161101 Added createMnemonic().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.function.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.chui.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

// explicit import to resolve conflict with com.goldencode.p2j.util.Text
import com.goldencode.p2j.ui.client.Text;

/**
 * GUI-specific widget factory implementation.
 * <p>
 * The return type for the {@code create} APIs is different from the {@link WidgetFactory} 
 * versions because:
 * <ul>
 *    <li>it avoids adding at the signature the {@link GuiOutputManager} type, i.e. instead
 *        of using {@code MessageArea&lt;GuiOutputManager&gt;} as return type, 
 *        {@link MessageAreaImpl} can be used directly, with the same effect. </li>
 *    <li> adds an additional protection against incorrect usage of the {@link WidgetFactory} 
 *    implementation classes: if the {@link WidgetFactory} implementation class is used 
 *    explicitly, it prevents accidental mixing of ChUI and GUI widgets (i.e. if used from 
 *    incorrect location). </li>
 * </ul> 
 */
public class GuiWidgetFactory
extends WidgetFactoryAdapter<GuiOutputManager>
{
   static 
   {
      classMap.put(Browse.class, BrowseGuiImpl.class);
      classMap.put(BrowseColumn.class, BrowseColumnGuiImpl.class);
      classMap.put(Button.class, ButtonGuiImpl.class);
      classMap.put(Editor.class, EditorGuiImpl.class);
      classMap.put(ComboBox.class, ComboBoxGuiImpl.class);
      classMap.put(FillIn.class, FillInGuiImpl.class);
      classMap.put(Frame.class, FrameGuiImpl.class);
      classMap.put(Image.class, ImageGuiImpl.class);      
      classMap.put(RadioButton.class, RadioButtonGuiImpl.class);
      classMap.put(RadioSet.class, RadioSetGuiImpl.class);
      classMap.put(Rect.class, RectangleGuiImpl.class);
      classMap.put(ScrollableList.class, ScrollableSelectionListGuiImpl.class);
      classMap.put(SelectionListBody.class, SelectionListBodyGuiImpl.class);
      classMap.put(SelectionList.class, SelectionListGuiImpl.class);
      classMap.put(Skip.class, SkipGuiImpl.class);
      classMap.put(Slider.class, SliderGuiImpl.class);
      classMap.put(Space.class, Space.class);
      classMap.put(Text.class, TextGuiImpl.class);
      classMap.put(Window.class, WindowGuiImpl.class);
      classMap.put(ToggleBox.class, ToggleBoxGuiImpl.class);
      classMap.put(Menu.class, MenuGuiImpl.class);
      classMap.put(MenuItem.class, MenuItemGuiImpl.class);
      classMap.put(SubMenu.class, SubMenuGuiImpl.class);
   }
   
   /**
    * Create an object accumulating prepared label and mnemonic info.
    * 
    * @return   new {@link MnemonicInfoGuiImpl} instance.
    */
   @Override
   public MnemonicInfo createMnemonic()
   {
      return new MnemonicInfoGuiImpl();
   }

   /**
    * Create UI-specific {@link Window} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  An instance of the UI-specific implementation of {@link Window}.
    */
   @Override
   public WindowGuiImpl createWindow(WidgetId id)
   {
      WindowGuiImpl window = new WindowGuiImpl(id);
      
      return window;
   }
   
   /**
    * Create UI-specific {@link StatusLine} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  An instance of the UI-specific implementation of {@link StatusLine}.
    */
   @Override
   public StatusLineGuiImpl createStatusLine(WidgetId id)
   {
      return new StatusLineGuiImpl(id);
   }
   
   /**
    * Create UI-specific {@link MessageArea} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   lines
    *          Number of visible lines in message area.         
    * 
    * @return  instance of UI-specific implementation of {@link MessageArea}.
    */
   @Override
   public MessageArea<GuiOutputManager> createMessageArea(WidgetId id, int lines)
   {
      return new MessageAreaGuiImpl(id, lines);
   }
   
   /**
    * Create UI-specific {@link Image} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Image}.
    */
   @Override
   public ImageGuiImpl createImage(WidgetId id)
   {
      return new ImageGuiImpl(id);
   }
   
   /**
    * Create UI-specific {@link ScrollBar} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   orientation
    *          Scroll bar orientation horizontal or vertical.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollBarGuiImpl}.
    */
   @Override
   public ScrollBarGuiImpl createScrollBar(WidgetId id, 
                                           ScrollBar.Orientation orientation, 
                                           Viewport<GuiOutputManager> viewport)
   {
      return new ScrollBarGuiImpl(id, orientation, viewport);
   }

   /**
    * Create UI-specific {@link DropDown} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    *          
    * @return  instance of UI-specific implementation of {@link DropDown}.
    */
   @Override
   public DropDownGuiImpl createDropDown(ComboBox<GuiOutputManager> parent)
   {
      return new DropDownGuiImpl(parent);
   }

   /**
    * Create UI-specific {@link ScrollableList} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    * @param   model
    *          Scrollable list model.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollableList}.
    */
   @Override
   public ScrollableSelectionListGuiImpl createScrollableList(ComboBox<GuiOutputManager> parent, 
                                                              ListModel<String>           model)
   {
      return new ScrollableSelectionListGuiImpl(parent, model);
   }

   /**
    * Create UI-specific {@link SensitiveScrollContainer} implementation.
    * 
    * @return  instance of UI-specific implementation of {@link SensitiveScrollContainer}.
    */
   @Override
   public SensitiveScrollContainer<GuiOutputManager> createScrollContainer()
   {
      return new SensitiveScrollContainer<GuiOutputManager>();
   }

   /**
    * Create GUI-specific {@link ScrollPane} instance.
    * 
    * @param   widget
    *          Widget which will be scrolled.
    *          
    * @return  instance of GUI-specific implementation of {@link ScrollPaneGuiImpl}.
    */
   @Override
   public ScrollPaneGuiImpl createScrollPane(ScrollableWidget<GuiOutputManager> widget) 
   {
      return new ScrollPaneGuiImpl(widget);
   }

   /**
    * Create GUI-specific {@link ScrollPane} instance.
    * 
    * @param   widget
    *          Widget which will be scrolled.
    * @param   wnd
    *          TopLevelWindow supplier
    *          
    * @return  instance of GUI-specific implementation of {@link ScrollPaneGuiImpl}.
    */
   @Override
   public ScrollPaneGuiImpl createScrollPane(ScrollableWidget<GuiOutputManager> widget, 
         Supplier<TopLevelWindow<GuiOutputManager>> wnd)
   {
      return new ScrollPaneGuiImpl(widget, wnd);
   }
   
   /**
    * Create UI-specific {@link Label} instance as a copy of given {@link Label}.
    * 
    * @param   id
    *          The widget's ID.
    * @param   labelInstance
    *          Source {@link Label} instance.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    *          
    * @return  instance of UI-specific implementation of {@link Label}.
    */
   @Override
   public LabelGuiImpl createLabel(WidgetId id, Label<GuiOutputManager> labelInstance, int frameId)
   {
      return new LabelGuiImpl(id, labelInstance, frameId);
   }
   
   /**
    * Create new {@link Label} instance for given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Label text. 
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    * 
    * @return  new {@link Label} instance.
    */
   @Override
   public LabelGuiImpl createLabel(WidgetId id, String text, int frameId)
   {
      return new LabelGuiImpl(id, text, frameId);
   }

   /**
    * Create new button for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Button text.
    *          
    * @return  new {@link Button} instance.
    */   
   @Override
   public ButtonGuiImpl createButton(WidgetId id, 
                                     Container<GuiOutputManager> container, 
                                     String text)
   {
      return new ButtonGuiImpl(id, container, text);
   }
   
   /**
    * Create UI-specific {@link AlertBox} instance.
    * 
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    content
    *           The list of message text components.
    * @param    buttonSet
    *           One of the BTN_* constants specifying used button set.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    title
    *           The dialog title.
    * @param    color
    *           Color scheme to use, ignored.
    * @param    stack
    *           Progress stack trace
    *
    * @return   instance of UI-specific implementation of {@link MessageArea}.
    */
   @Override
   public AlertBoxGuiImpl createAlertBox(int      alertType,
                                         Object[] content, 
                                         int      buttonSet, 
                                         logical  var, 
                                         String   title,
                                         Color    color,
                                         String[] stack)
   {
      return new AlertBoxGuiImpl(alertType, content, buttonSet, var, title, stack); 
   }
   
   /**
    * Creates new instance of {@link BorderedPanel} with {@link BorderLayout} as
    * the layout manager and no border.
    * 
    * @return  instance of GUI-specific implementation of {@link BorderedPanel}.
    */
   @Override
   public BorderedPanel<GuiOutputManager> createBorderedPanel()
   {
      return new BorderedPanelGuiImpl(new BorderLayout<GuiOutputManager>(), 
                                      new EmptyBorder<GuiOutputManager>());
   }
   
   /**
    * Creates new instance of {@link BorderedPanel} with {@link BorderLayout}
    * layout manager and simple border of width 1 measured in native units. 
    * 
    * @return  instance of GUI-specific implementation of {@link BorderedPanel}.
    */
   @Override
   public BorderedPanel<GuiOutputManager> createLineBorderedPanel()
   {
      return new BorderedPanelGuiImpl(new BorderLayout<GuiOutputManager>(), 
                                      new LineBorderGuiImpl());
   }
      
   /**
    * Creates new instance of {@link BorderedPanel} with a specific layout
    * manager and border.
    * 
    * @param   mgr
    *          Layout manager.
    * @param   border
    *          Border for this widget.
    * 
    * @return  instance of GUI-specific implementation of {@link BorderedPanel}.
    */
   @Override
   public BorderedPanel<GuiOutputManager> createBorderedPanel(
      LayoutManager<GuiOutputManager> mgr, 
      Border<GuiOutputManager>        border)
   {
      return new BorderedPanelGuiImpl(mgr, border);
   }
   
   /**
    * Creates new instance representing a default line border.
    * 
    * @return  instance of GUI-specific implementation of {@link Border}.
    */
   @Override
   public LineBorderGuiImpl createLineBorder()
   {
      return new LineBorderGuiImpl();
   }

   /**
    * Create GUI-specific {@link RadioButton} instance.
    * 
    * @param   config
    *          Parent config. 
    * @param   label
    *          Radio button label. 
    * @param   value
    *          Radio button data type. 
    * 
    * @return  instance of GUI-specific implementation of {@link RadioButton}.
    */
   @Override
   public RadioButtonGuiImpl createRadioButton(RadioSetConfig config, 
                                               String         label, 
                                               BaseDataType   value)
   {
      return new RadioButtonGuiImpl(config, label, value);
   }
}