/* 
** Module   : GuiWindow.java
** Abstract : Interface for top level GUI windows
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20050509 Initial version.
** 002 HC  20150622 Fixed GUI frame to be properly centered in its parent window.
** 003 HC  20150626 Javadoc fixes.
** 004 HC  20150831 Improved GUI window resize performance.
** 005 CA  20150910 Added message area getter.
** 006 HC  20151013 Added getWorkspaceScreenLocation().
** 007 CA  20151015 Fixed window resize via attributes (they target the workspace, not the full 
**                  window).
** 008 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 009 HC  20160304 Replaced Window.moveBy() with setPhysicalLocation(). Window layout
**                  improvements.
** 010 HC  20160406 Overhaul of window-activation logic.
** 011 HC  20160428 GUI window border drawing changed to use the Border interface mechanism. Also
**                  removed hardcoded assumptions of the border size.
** 012 CA  20170104 Some improvements in window decoration support for embedded mode (modal 
**                  windows always have decorations).
** 013 HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.CoordinatesConversion;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.Widget;

/**
 * The interface provides an API for operations related to top level
 * GUI windows.
 */
public interface GuiWindow
extends Widget<GuiOutputManager>
{
   /**
    * Resolve the font associated with the window's title.
    * 
    * @return   See above.
    */
   int resolveTitleFont();

   /**
    * Get the message area widget associated with this window.
    * 
    * @return   See above.
    */
   MessageAreaGuiImpl getGuiMessageArea();

   /**
    * Iconify the window.
    */
   void iconify();

   /**
    * Iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   void iconify(int sourceId);

   /**
    * Iconify the window.
    *
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    * @param    postEvent
    *           When <true>true</true> the method posts the WINDOW-MINIMIZED legacy event,
    *           <code>false</code> means no event is posted.
    */
   void iconify(int sourceId, boolean postEvent);

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    */
   void restore();

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   void restore(int sourceId);

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    *
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    * @param    postEvent
    *           When <true>true</true> the method posts the WINDOW-RESTORED legacy event,
    *           <code>false</code> means no event is posted.
    */
   void restore(int sourceId, boolean postEvent);

   /**
    * Maximize this window.
    */
   void maximize();

   /**
    * Maximize this window.
    *
    * @param    postEvent
    *           When <true>true</true> the method posts the WINDOW-MAXIMIZED legacy event,
    *           <code>false</code> means no event is posted.
    */
   void maximize(boolean postEvent);

   /**
    * Resize the widget to the specified dimensions.
    * 
    * @param    origin
    *           The top-left window corner new position.
    * @param    size
    *           The new window size.
    */
   void resizeTo(NativePoint origin, NativeDimension size);

   /**
    * Resize the widget to the specified dimensions.
    * 
    * @param    origin
    *           The top-left window corner new position.
    * @param    size
    *           The new window size.
    * @param    raise
    *           Flag indicating if the 4GL WINDOW-RESIZED event needs to be raised.
    */
   void resizeTo(NativePoint origin, NativeDimension size, boolean raise);

   /**
    * Sets the new application icon for given window on the client side.
    * 
    * @param    newIcon
    *           A new icon image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   boolean setWindowIcon(byte[] newIcon);

   /**
    * Loads the new application icon for given window from the client side filesystem.
    * 
    * @param    small
    *           Flag indicating usage of the small icon case(16x16 in size).
    * @param    name
    *           The fully qualified file name for the icon to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   boolean loadWindowIcon(boolean small, String name);

   /**
    * Check if the window is already maximized - it can't be resized.
    * 
    * @return   If maximized returns <code>true</code>, <code>false</code> otherwise.
    */
   boolean isMaximized();
   
  /**
   * Returns the output window's coordinates conversion service.
   * 
   * @return   a reference to the {@link CoordinatesConversion} instance. 
   */
  CoordinatesConversion coordinates();
  
  /**
   * Set window size.
   *
   * @param    width
   *           New width.
   * @param    height
   *           New height.
   */
  void setSize(double width, double height);
  
  /**
   * Determine if this window is being resized.
   * 
   * @return   If being resized returns <code>true</code>, <code>false</code> 
   *           otherwise.
   */
   boolean isResizing();

   /**
    * Determine if this window workspace needs to be resized.
    * 
    * @return   The state of the resizing flag.
    */
   boolean isResizeWorkspace();
  
   /**
     * Determine the resized window size.
    * 
    * @return   The new size.
    */
   NativeDimension getResizedSize();
  
   /**
    * Get the title bar widget associated witn this window.
    * 
    * @return   See above.
    */
   WindowTitleBar getTitleBar();
  
   /**
    * Get the status area widget associated with this window.
    * 
    * @return   See above.
    */
   StatusLine<GuiOutputManager> getStatusArea();
  
   /**
    * Get the screen location of the window workspace area, in native units.
    * 
    * @return   See above.
    */
   NativePoint getWorkspaceScreenLocation();
  
   /**
    * Get the native size of the window workspace area.
    * 
    * @return   See above.
    */
   NativeDimension getWorkspaceNativeSize();

   /**
    * Get the maximum size of the window workspace area.
    * 
    * @return   See above.
    */
   Dimension getMaxWorkspaceSize();
   
   /**
    * Sets the native size of the window workspace area.
    * 
    * @param    width
    *           New workspace width.
    * @param    height
    *           New workspace height.
    */
   void setWorkspaceNativeSize(int width, int height);
   
   /**
    * Sets the window workspace location.
    *  
    * @param    x
    *           The x location. 
    * @param    y
    *           The y location.
    */
   void setWorkspaceLocation(int x, int y);
   
   /**
    * Close this window.
    */
   void close();

   /**
    * Get the message area widget associated with this window.
    * 
    * @return   See above.
    */
   public MessageArea<GuiOutputManager> getMessageArea();
   
   /**
    * Set window location.
    * 
    * The x value corresponds to COLUMN while
    * y value corresponds to ROW 4GL attribute.
    * Note however that the values are expected to be
    * 0-based.
    * 
    * @param   x
    *          Window COLUMN.
    * @param   y
    *          Window ROW.
    */
   public void setLocation(double x, double y);

   /**
    * Set window physical location.
    * 
    * @param   x
    *          Widget physical X.
    * @param   y
    *          Widget physical Y.
    */
   public void setPhysicalLocation(int x, int y);
   
   /**
    * Returns the window minimum size.
    * 
    * @return  See above.
    */
   public NativeDimension getMinSize();
   
   /**
    * This method indicates whether the window shares its activation
    * state with its owner. This method allows multiple windows to be
    * considered active simulteniously. Example use cases are the
    * COMBO-BOX drop-down window or menu window. Note that while at the
    * application level we may recognize multiple active windows
    * simulteniously, at the GUI driver level, there is always one 
    * active window at a time. 
    */
   public default boolean isShareActivationWithOwner()
   {
      return false;
   }
  
   /**
    * Returns the window's border implementation. If <code>null</code>
    * is returned, the window has no border.
    * 
    * @return   A border implementation or <code>null</code>.
    */
   public Border<GuiOutputManager> getBorder();
  
   /**
    * Determine if this window can have decorations.
    * 
    * @return   See above.
    */
   public boolean hasDecorations();
}