/*
** Module   : ImageGuiImpl.java
** Abstract : Implementation of a IMAGE widget for the GUI driver.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.  Not yet fully tested/complete.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 CA  20141021 Fixed image drawing and loading: tested only for the case when is used as a 
**                  window icon.
** 006 MAG 20141029 Added clipping rectangle when drawing. 
** 007 HC  20141021 Implemented window move/position support.
** 008 EVL 20141204 Adding image reloading capability.
** 009 EVL 20141208 Adding image widget initialization code.
** 010 EVL 20150108 Differentiate small icon from large one flag.  Rework for icon/image
**                  implementation.
** 011 EVL 20150123 Handle deferred image loading.
** 012 EVL 20150203 Implementing image offset capability and size phrase.  Change the approach for
**                  getting image coordinates.
** 013 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
** 014 EVL 20150224 Minor change fo using final modifier.
** 015 HC  20150323 Removed business logic from config classes.
** 016 CA  20150511 If the image is not loaded and heigh/width is set, return it.
** 017 GES 20150425 Eliminated AWT dependencies.
** 018 SBI 20150901 Refactored image drawing.
** 019 IAS 20150926 MOVE-TO-TOP()/MOVE-TO_BUTTON() support.
** 020 EVL 20151018 Adding methods to (load|set)Image without conditional repaint.  This is
**                  required to load or set the image without painting by GUI button
**                  implementation.
** 021 HC  20151023 setImage() must not call repaint() when widget not attached to widget tree.
**     EVL 20151106 Adding method to use transformed grey image from given original.
** 022 CA  20160216 Cache the intermediate widget states.
** 023 HC  20160929 Made GUI IMAGE widget trigger mouse events. Fixed abends in image processing
**                  and drawing when image offset is involved.
** 024 EVL 20170131 Fix for icon based image processing.
** 025 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;
import java.io.*;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a IMAGE widget for the GUI driver.
 */
public class ImageGuiImpl
extends Image<GuiOutputManager>
{
   /** logger */
   protected static final Logger LOG = LogHelper.getLogger(ImageGuiImpl.class.getName());

   /** The image object. */
   protected ImageWrapper<?> img;
   
   /** The graphics driver. */
   private GuiDriver gd;
   
   /** To keep track if this is the window icon or not. */
   private boolean isIcon = false;

   /**
    * Create a new image widget.
    * 
    * @param    id
    *           The widget's ID.
    */
   public ImageGuiImpl(WidgetId id)
   {
      super(id);

      this.gd = (GuiDriver) OutputManager.getDriver();
   }

   /**
    * Create a new image widget.
    *
    * @param    id
    *           The widget's ID.
    * @param    isIcon
    *           Flag indicating the image is the window title bar icoon.
    */
   public ImageGuiImpl(WidgetId id, boolean isIcon)
   {
      this(id);

      this.isIcon = isIcon;
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ImageGuiImpl(WidgetId id, ImageConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ImageConfig cfg)
   {
      super.initialize(id, cfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // check if we need to preload image
      String imgSrcFile = cfg.image;
      
      // filename exists, load deferred image
      if (imgSrcFile != null && imgSrcFile.length() > 0 && img == null)
      {
         int thisId = config().id.asInt();
         int type = ButtonConfig.IMAGE_IMG;

         // first try to search for images loaded from application jar
         byte[] imgData = ImageHelper.getPendingImageData(thisId, type);
         
         if (imgData != null)
         {
            setImage(imgData);
         }
         else
         {
            String loadName = ImageHelper.getPendingImageName(thisId, type);
            
            if (loadName != null && loadName.length() > 0)
            {
               loadImage(loadName);
            }
         }
      }
   }

   /**
    * Draw this image
    */
   @Override
   public void draw()
   {
      NativePoint origin = physicalLocation();
      final NativeDimension d = physicalDimension();
      NativeRectangle clip = clipRectangle(origin, d);
      
      // caching works only for rectangular widgets - images with transparent background can not
      // be cached properly, as capturing the clip will include the background.  also, images
      // are cached on the web-side
      gd.draw(origin, clip, () ->
      {
         if (!drawImage())
         {
            gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
            gd.fillRect(0, 0, d.width, d.height);
         }
      });
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      ImageConfig icfg = config();
      
      // here we need the actual image height
      int px = icfg.heightPixels != 0 ? icfg.heightPixels : img == null ? 0 : img.getHeight();
      
      return screen().coordinates().heightFromNative(px);
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      ImageConfig icfg = config();
      
      // here we need the actual image width
      int px = icfg.widthPixels != 0 ? icfg.widthPixels : img == null ? 0 : img.getWidth();
      
      return screen().coordinates().widthFromNative(px);
   }

   /**
    * Sets the new image for given image widget on the client side.
    * 
    * @param    byteImg
    *           A new image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setImage(byte[] byteImg)
   {
      return setImage(byteImg, true);
   }
   
   /**
    * Sets the new image for given image widget on the client side.
    * 
    * @param    byteImg
    *           A new image array data.
    * @param    repaint
    *           Need immediate repaint of the new image.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setImage(byte[] byteImg, boolean repaint)
   {
      // the new image candidate
      ImageConfig icfg = config();
      // check if the image is an icon or not
      boolean isIcoFile = icfg.name != null && icfg.name.toLowerCase().contains(".ico");
      ImageWrapper imgNew = isIcon || isIcoFile ? 
                              gd.createIcon(byteImg, false) : 
                              AbstractGuiDriver.createImage(byteImg);
      
      // image created, replace the old one
      if (imgNew != null)
      {
         img = imgNew;
         
         // repaint changed image if required and only if the widget has been
         // attached to the widget tree
         if (repaint && parent() != null)
         {
            ThinClient.getInstance().eventDrawingBracket(this, false, true, this::repaint);
         }
         return true;
      }
      else
      {
         return false;
      }
   }
   
   /**
    * Sets the new image as result of the grey transformation from given image widget.
    * 
    * @param    imgInit
    *           The image widget with original data to transform.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setGreyImage(ImageGuiImpl imgInit)
   {
      boolean resilt = false;
      
      // check if we have valid source
      if (imgInit.img != null)
      {
         // the new image candidate
         ImageWrapper imgNew = gd.createGreyImage(imgInit.img.getImage());

         if (imgNew != null)
         {
            img = imgNew;
            resilt = true;
         }
      }

      return resilt;
   }

   /**
    * Loads the new image for given image widget from the client side filesystem.
    * 
    * @param    fullName
    *           The fully qualified file name for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean loadImage(String fullName)
   {
      return loadImage(fullName, true);
   }
   
   /**
    * Loads the new image for given image widget from the client side filesystem.
    * 
    * @param    fullName
    *           The fully qualified file name for the image to be loaded.
    * @param    repaint
    *           Need immediate repaint of the new image.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean loadImage(String fullName, boolean repaint)
   {
      InputStream is = null;
      byte[] imageData = null;
         
      // find the image on local file system
      try
      {
         is = new FileInputStream(fullName);
      }
      catch (FileNotFoundException ex)
      {
         is = null;
      }
         
      // the file is found, try to read
      if (is != null)
      {
         imageData = UiUtils.getData(is);
      }
         
      // if having new image data - set it
      if (imageData != null)
      {
         // refresh and remember if this is an icon or not
         if (!isIcon)
         {
            isIcon = fullName.toLowerCase().contains(".ico"); //TODO endswith() better?
         }
         return setImage(imageData, repaint);
      }
      else
      {
         return false;
      }
   }
   
   /**
    * Checks if the image was icon based.
    * 
    * @return <code>TRUE</code> if icon assumed, <code>FALSE</code> otherwide.
    */
   public boolean isIcon()
   {
      return isIcon;
   }
   
   /**
    * Get the Z-Order class
    * 
    * @return Z-Order class of the widget
    */
   @Override
   public ZOrderClass getZOrderClass()
   {
      return ZOrderClass.DECORATION;
   }
   
   /**
    * Get internal image native width in pixels.
    * 
    * @return  internal image width.
    */
   protected int getImageWidth()
   {
      if (img != null)
      {
         ImageConfig icfg = config();

         // compute best match X offset if we have ones
         int xOff = icfg.offsetX > 0 ? icfg.offsetX :
                    icfg.offsetCol > 0.0 ? screen().coordinates().columnToNative(icfg.offsetCol) :
                    0;

         int wid = img.getWidth() - xOff;
                    
         // there are many possible combinations of width and offset
         return icfg.widthPixels > 0 ? 
                   Math.min(wid, icfg.widthPixels) :
                icfg.widthChars > 0.0 ?
                   Math.min(wid, screen().coordinates().columnToNative(icfg.widthChars)) : wid;
      }
      else
      {
         return screen().coordinates().columnToNative(width());
      }
   }
   
   /**
    * Get internal image native height in pixels.
    * 
    * @return  internal image height.
    */
   protected int getImageHeight()
   {
      if (img != null)
      {
         ImageConfig icfg = config();
         
         // compute best match y offset if we have ones
         int yOff = icfg.offsetY > 0 ? icfg.offsetY :
                    icfg.offsetRow > 0.0 ? screen().coordinates().rowToNative(icfg.offsetRow) :
                    0;
                    
         int ht = img.getHeight() - yOff;
                    
         // there are many possible combinations of height and offset
         return icfg.heightPixels > 0 ?
                   Math.min(ht, icfg.heightPixels) :
                icfg.heightChars > 0.0 ?
                   Math.min(ht, screen().coordinates().rowToNative(icfg.heightChars)) : ht;
      }
      else
      {
         return screen().coordinates().rowToNative(height());
      }
   }

   /**
    * Calculates the drawing parameters:
    *     width  =  (Integer) drawingParameters[0];
    *     height =  (Integer) drawingParameters[1];
    *     offsetX = (Integer) drawingParameters[2];
    *     offsetY = (Integer) drawingParameters[3];
    *     transparent  = (Boolean) drawingParameters[4];
    *     stretchToFit = (Boolean) drawingParameters[5];
    *     retainShape  = (Boolean) drawingParameters[6];
    *     convert3D    = (Boolean) drawingParameters[7];
    *
    * @return   The array filled with drawing parameters.
    */
   protected Object[] calculateDrawingParameters()
   {
      if (img == null)
      {
         return null;
      }
      CoordinatesConversion c = screen().coordinates();
      
      // get the current image config
      ImageConfig icfg = config();

      // the width and height in pixels or chars overridden by size phrase has higher priority
      // than natural image size itself
      int w = icfg.widthPixels > 0 ? icfg.widthPixels : 
              icfg.widthChars > 0 ? c.widthToNative(icfg.widthChars) : 
              c.widthToNative(width());
      int h = icfg.heightPixels > 0 ? icfg.heightPixels : 
              icfg.heightChars > 0 ? c.heightToNative(icfg.heightChars) : 
              c.heightToNative(height());

      // compute final image dimensions
      int iw = 0;
      int ih = 0;
      
      if (isStretchToFit())
      {
         // need to fit the image to display area size
         iw = w;
         ih = h;
      }
      else
      {
         // take the final image size as minimum from requested and real image sizes
         iw = Math.min(img.getWidth(), w);
         ih = Math.min(img.getHeight(), h);
      }
      
      // recalculate image offsets that can be changed
      int offsetX = 0;
      int offsetY = 0;
      
      // check if we have the offset variables changed
      if (icfg.offsetX != 0)
      {
         offsetX = icfg.offsetX;
      }
      else if (icfg.offsetCol != 0.0)
      {
         offsetX = screen().coordinates().columnToNative(icfg.offsetCol);
      }
      else if (icfg.widthPixels != 0 || icfg.widthChars != 0.0)
      {
         offsetX = 0;
      }
      // Y coordinate
      if (icfg.offsetY != 0)
      {
         offsetY = icfg.offsetY;
      }
      else if (icfg.offsetRow != 0.0)
      {
         offsetY = screen().coordinates().rowToNative(icfg.offsetRow);
      }
      else if (icfg.heightPixels != 0 || icfg.heightChars != 0.0)
      {
         offsetY = 0;
      }

      return new Object[] { iw, ih, offsetX, offsetY, icfg.transparent,
                            isStretchToFit(), icfg.retainShape, icfg.convert3D};
   }

   /**
    * Draw this image.
    */
   protected boolean drawImage()
   {
      if (img != null)
      {
         Object[] drawingParameters = calculateDrawingParameters();
         int iw = (Integer) drawingParameters[0];
         int ih = (Integer) drawingParameters[1];
         int offsetX = (Integer) drawingParameters[2];
         int offsetY = (Integer) drawingParameters[3];
         boolean transparent  = (Boolean) drawingParameters[4];
         boolean stretchToFit = (Boolean) drawingParameters[5];
         boolean retainShape  = (Boolean) drawingParameters[6];
         boolean convert3D    = (Boolean) drawingParameters[7];
         // render image
         gd.drawImage(img, 0, 0, iw, ih, offsetX, offsetY, transparent,
                      isStretchToFit(), retainShape, convert3D);
      }
      
      return img != null;
   }

   /**
    * Determine the mouse actions processed by this widget.
    *
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED
      };
   }

   /**
    * The method returns a value indicating whether this image is supposed to be stretched-to-fit.
    *
    * @return   See above.
    */
   protected boolean isStretchToFit()
   {
      return config.stretchToFit && !isIcon;
   }
}