/*
** Module   : MenuGuiImpl.java
** Abstract : GUI client implementation of Menu.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-------------------------------------
** 001 VIG 20150525 Created initial version.
** 002 VIG 20150818 Added keyMode flag to indicate key navigation mode.
** 003 CA  20150910 Fixed some issues related to menu acting as the system popup menu.
** 004 CA  20151007 Workaround for mismatch of clipping behavior.
** 005 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 006 CA  20151026 Enclose driver API calls in select/releaseWindow().
**     VIG 20151103 Fixed NPE.
** 007 VIG 20151226 Added key navigation mode flag and focus reset to mouseClicked().
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 IAS 20160331 Fixed potential NPE
** 010 CA  20160216 Cache the intermediate widget states.
** 011 VIG 20160128 If this popup is focused in some container - reset focus after popup hiding.
**                  Added initSystemPopupMenu(), added currentPopup and popupShown().
** 012 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 013 VIG 20160620 Added darkShadow color, fixed popup menu sizes, fixed height/width methods.
** 014 VIG 20160628 Cleaned draw(), added width fix to width() method.
**                  Added getMenubarParentOrSelf() method, added flag mbarPressedState,
**                  fixed width and drawing. Added menubar state static flags.
** 015 VIG 20161105 getParent() call is refactored to getParentMenuContainer().
** 016 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>Menu</code> GUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class MenuGuiImpl
extends Menu<GuiOutputManager>
{
   /** Current pressed MENUBAR item */
   private static MenuElement<GuiOutputManager> mbarPressedItem = null;
   
   /** 
    * Flag, indicating the the state of active MENUBAR item. It can be MENU-ITEM or SUB-MENU.
    * SUB-MENU can be opened or closed. The flag denoting the view of menubar item title when
    * it is active (hovered by mouse): concave (true) or convex (false).   
    */
   private static boolean mbarPressedState = false;
   
   /**
    * The flag indicating the state of menubar sub-menu body. It is used only in key
    * navigation mode. <code>true</code> means the body is displayed. When menubar sub-menu
    * gains the focus in key navigation mode there are 2 possible cases for body: displayed or not.
    */
   private static boolean mbarSmBodyDisplayed = false;
   
   /** Screen driver */
   @SuppressWarnings("rawtypes")
   private GuiDriver gd;
   
   /** Number of MENUBAR lines */
   private int menubarLines = 1;
   
   /** 
    * Keyboard navigation mode: If <code>true</code> =&gt; mnemonics are drawn.
    * This mode is intended only for MENUBAR. It activated using ALT key.
    */
   private boolean keyMode = false;
   
   /** Flag, indicating, that current operation is navigating and it called by key */ 
   private boolean keyNavigation = false;
   
   /** Bracket for menu navigation. */
   private static boolean keepOpenOnBodyClose = false; 
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public MenuGuiImpl(WidgetId id, MenuConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Set the value of the menubar state flags.
    * 
    * @param   mbarPressedItem
    *          MENUBAR item the state is being set for.
    * @param   pressedState
    *          <code>true</code> or <code>false</code>.
    *          
    * @return  <code>true</code> if the state was successfully changed, 
    *          otherwise return <code>false</code>.
    */
   public static boolean setMbarPressedState(MenuElement<GuiOutputManager> mbarPressedItem, 
                                             boolean pressedState)
   {
      if (mbarPressedItem != null && !Menu.isMenubarElement(mbarPressedItem))
      {
         throw new IllegalStateException("Specified item is not element of MENUBAR");
      }
      
      // If new value is <code>false</code> than assign only it if the same item.
      if (!pressedState                        &&
          MenuGuiImpl.mbarPressedItem != null  &&
          mbarPressedItem != null              &&
          !mbarPressedItem.equals(MenuGuiImpl.mbarPressedItem))
      {
         return false;
      }

      MenuGuiImpl.mbarPressedItem = mbarPressedItem;
      MenuGuiImpl.mbarPressedState = pressedState;
      
      return true;
   }

   /**
    * Set the value of the menubar state flags.
    * 
    * @param   mbarPressedItem
    *          MENUBAR item the state is being set for.
    * @param   pressedState
    *          true or false.
    * @param   bodyDisplayed
    *          Body displaying flag value.
    */
   public static void setMbarPressedState(MenuElement<GuiOutputManager> mbarPressedItem, 
                                          boolean pressedState, 
                                          boolean bodyDisplayed)
   {
      if (setMbarPressedState(mbarPressedItem, pressedState))
      {
         mbarSmBodyDisplayed = bodyDisplayed;
      }
   }
   
   /**
    * Get the value of the flag {@link #mbarPressedState}
    * 
    * @return  <code>true</code> or <code>false</code>.
    */
   static boolean isMbarPressedState()
   {
      return mbarPressedState;
   }

   /**
    * Get the value of the flag {@link #mbarSmBodyDisplayed}
    * 
    * @return  <code>true</code> or <code>false</code>.
    */
   static boolean isMbarSmBodyDisplayed()
   {
      return mbarSmBodyDisplayed;
   }
   
   /**
    * Get the value of the flag {@link #mbarPressedState}
    * 
    * @param   item
    *          MENUBAR item to check state state of.
    * 
    * @return  <code>true</code> or <code>false</code>.
    */
   public static boolean isMbarItemPressed(MenuElement<GuiOutputManager> item)
   {
      if (item != null && !Menu.isMenubarElement(item))
      {
         return false;
      }
      
      if (item == null && mbarPressedItem == null)
      {
         return true;
      }
      
      if (item == null || mbarPressedItem == null)
      {
         return false;
      }
      
      return mbarPressedItem.equals(item) && mbarPressedState;
   }
   
   /**
    * Reset MENUBAR state flags.
    */
   public static void resetMbarPressedState()
   {
      mbarPressedItem = null;
      mbarPressedState = false;
      mbarSmBodyDisplayed = false;
   }

   /**
    * A worker to process a focus request for the specified menu widget.  For menus, it is required
    * that a {@link EventType#FOCUS_LOST} is sent to the currently focused menu widget which has
    * the same parent as the focus requester.
    * 
    * @param    requester
    *           The focus requester.
    */
   protected static void requestFocusWorker(MenuElement<?> requester)
   {
      Widget<?> currentFocus = UiUtils.getCurrentFocus();
      Widget<?> thisParent = requester.getParentMenuContainer();

      if (currentFocus != null                &&
          thisParent != currentFocus          &&
         (currentFocus instanceof SubMenuGuiImpl || currentFocus instanceof MenuItemGuiImpl))
      {
         // send FOCUS_LOST to the focused widget which is on the same parent as this one
         Container<?> focusLostParent = ((MenuElement<?>) currentFocus).getParentMenuContainer();
         while (focusLostParent != null && focusLostParent != thisParent)
         {
            if (focusLostParent instanceof MenuElement<?>)
            {
               focusLostParent = ((MenuElement<?>) focusLostParent).getParentMenuContainer();
            }
            else
            {
               focusLostParent = focusLostParent.parent();
            }
         }
         
         Widget<?> lostFocus = null;
         if (focusLostParent != null)
         {
            // we are on the same parent, get the focus directly
            lostFocus = focusLostParent.focus();
         }
         
         if (lostFocus != null)
         {
            EventManager.postEvent(new FocusEvent(lostFocus, EventType.FOCUS_LOST));
         }
      }
   }
   
   /**
    * Find the mouse source in this menu, recursive.
    * 
    * @param    menu
    *           The (sub)menu parent.
    * @param    p
    *           The mouse location, relative to parent.
    * @param    bodyBounds
    *           The bounds of the (sub)menu's body.
    *           
    * @return   The selected sub-menu or menu-item; <code>null</code> if the mouse is not over
    *           a child in the menu tree.
    */
   protected static Widget<GuiOutputManager> findMenuMouseSource(AbstractContainer menu,
                                                                 NativePoint       p, 
                                                                 NativeRectangle   bodyBounds)
   {
      CoordinatesConversion cc = menu.screen().coordinates();
      
      // we have two cases:
      // 1. the mouse is somewhere over one of this sub-menu's direct children.  in this case, 
      // search the "title" dimension
      // 2. the mouse is somewhere over one of this sub-menu's opened hierarchy children (down
      // the tree) - in this case, delegate search to each of the sub-menu
      
      Widget<GuiOutputManager>[] widgets = menu.widgets();
      
      if (bodyBounds != null && bodyBounds.contains(p))
      {
         for (int i = widgets.length - 1; i >= 0; i--)
         {
            Widget<GuiOutputManager> w = widgets[i];
            
            if (!w.isVisible())
            {
               continue;
            }
            
            NativeRectangle wTitleBounds = null;
            
            if (w instanceof MenuItemGuiImpl)
            {
               // menu-items don't have children, so use their bounds as is
               wTitleBounds = w.physicalBounds();
            }
            else if (w instanceof SubMenuGuiImpl)
            {
               // sub-menus can have children, but search the "title" bounds first
               SubMenuGuiImpl sm = (SubMenuGuiImpl) w;
               
               NativePoint smTitleLoc = sm.physicalLocation();
               NativeDimension smTitleBounds = 
                  cc.dimensionToNative(new Dimension(sm.titleWidthWithBorders(), 
                                                     SubMenuGuiImpl.titleHeightWithBorders(sm)));
               
               wTitleBounds = new NativeRectangle(smTitleLoc, smTitleBounds);
            }
            
            if (wTitleBounds != null && wTitleBounds.contains(p))
            {
               return w;
            }
         }
      }
      else
      {
         // search the sub-menus
         for (int i = widgets.length - 1; i >= 0; i--)
         {
            Widget<GuiOutputManager> w = widgets[i];
            
            if (!w.isVisible())
            {
               continue;
            }
            
            if (w instanceof SubMenuGuiImpl)
            {
               NativePoint tp = new NativePoint(p);
               tp = tp.minus(w.physicalLocation());
               
               Widget<GuiOutputManager> found = w.findMouseSource(tp);
               
               if (found != null)
               {
                  return found;
               }
            }
         }
      }
      
      // nothing found, return null; don't delegate to super.findMouseSource()!
      return null;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @SuppressWarnings("rawtypes")
   @Override
   public void initialize(WidgetId id, MenuConfig cfg)
   {
      super.initialize(id, cfg);
      
      gd = (GuiDriver) OutputManager.getDriver();
      
      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent != null ? parent.window() : WindowManager.resolveWindow(this);
   }
   
   /**
    * Get keyboard navigation mode.
    * 
    * @return  <code>true</code> if keyboard navigation mode is on. 
    */
   public boolean isKeyMode()
   {
      return this.keyMode;
   }
   
   /**
    * Set keyboard navigation mode to specified value.
    * 
    * @param   keyMode
    *          <code>true</code> if keyboard navigation mode is on. 
    */
   public void setKeyMode(boolean keyMode)
   {
      this.keyMode = keyMode;
   }

   /**
    * Get keyboard navigation mode.
    * 
    * @return  <code>true</code> if keyboard navigation mode is on. 
    */
   public boolean isKeyNavigation()
   {
      return this.keyNavigation;
   }
   
   /**
    * Set keyboard navigation mode to specified value.
    * 
    * @param   keyNavigation
    *          <code>true</code> if keyboard navigation mode is on. 
    */
   public void setKeyNavigation(boolean keyNavigation)
   {
      this.keyNavigation = keyNavigation;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int [] { MouseEvent.MOUSE_CLICKED };
   }
   
   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget<GuiOutputManager> defaultFocus()
   {
      // We don't set default focus for sub-menus in GUI
      return null;
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      NativePoint p = physicalLocation();
      NativeDimension dim = physicalDimension();
      NativeRectangle clipRect = clipRectangle(p, dim);
      
      final Window<?> wnd = WindowManager.resolveWindow(this);
      if (!config().popupOnly)
      {
         // if there is a non-popup menu, always re-draw the window workspace here... this is a
         // workaround from a "feature" which requires the redrawn area to be the rectangle which 
         // encloses all expanded sub-menus
         ((WindowGuiImpl) wnd).getWorkspace().draw();
         
         ScreenBitmap bitmap = wnd.screen().getBitmapCopy(wnd.getId().asInt());
         NativeRectangle nr = bitmap.getOuterRectangle();
         NativePoint screenLoc = screenPhysicalLocation();
         
         // the clipping rectangle is always with a (0,0) origin: thus we need to translate before
         // intersection
         if (nr != null && nr.intersects(clipRect.translate(screenLoc)))
         {
            // the resulting rectangle will have the same relative origin as your "clip" rectangle
            // so you can intersect safely
            nr = nr.translate(-screenLoc.x, -screenLoc.y); 
            clipRect = clipRect.intersection(nr);
         }
      }
      
      // prepare the background for all menu:
      gd.draw(getId(), p, clipRect, () ->
            ThemeManager.getCurrentTheme().drawMenuBackground(
                  gd, 0, 0, dim.width, dim.height, config().popupOnly));
      
      // then let the super class cal draw() on each components (menu items)
      gd.draw(getId(), p, clipRect, MenuGuiImpl.super::draw);
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#width()
    */
   @Override
   public double width()
   {
      return width(true); 
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#height()
    */
   @Override
   public double height()
   {
      return SubMenuGuiImpl.getMaxSubMenuHeight(this);
   }
   
   /**
    * Lay out container. Set children locations.
    */
   @Override
   public void doLayout()
   {
      TitledWindow<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion c = wnd.screen().coordinates();
      int menuItemHeight = c.heightToNative(SubMenuGuiImpl.titleHeightWithBorders(this));
      
      Widget<GuiOutputManager>[] w = widgets();
      boolean isMenuBar = !config().popupOnly;
      NativeInsets menuInsets = ThemeManager.getCurrentTheme().getMenuInsets();
      int x = isMenuBar ? 0 : menuInsets.left;
      int y = isMenuBar ? 0 : menuInsets.top;
      menubarLines = 1;
      int wndWidth = c.widthToNative(wnd.width());
      
      for (Widget<GuiOutputManager> mi : w)
      {
         // we redo the layout, invalidate the existing sizes
         if (mi instanceof MenuItemGuiImpl)
         {
            ((MenuItemGuiImpl) mi).invalidateCachedSize();
         }
         else if (mi instanceof SubMenuGuiImpl)
         {
            ((SubMenuGuiImpl) mi).invalidateCachedSize();
         }
         
         int itemHeight = c.heightToNative(mi.height());
         
         if (isMenuBar) // menu-bar
         {
            // including window borders
            if (x + c.widthToNative(mi.width()) > wndWidth - 10)
            {
               x = menuInsets.left;
               y += menuItemHeight;
               menubarLines++;
            }
            
            mi.setPhysicalLocation(x, y);
            mi.setLocation(c.widthFromNative(x), c.heightFromNative(y));
            x += c.widthToNative(mi.width());
            if (menuItemHeight < itemHeight)
            {
               // we found a taller menu item. use its height for further calculations.
               menuItemHeight = itemHeight;
            }
         }
         else
         {
            mi.setPhysicalLocation(x, y);
            mi.setLocation(c.widthFromNative(x), c.heightFromNative(y));
            y += itemHeight;
         }
         
         if (mi instanceof SubMenu)
         {
            ((SubMenu<?>) mi).doLayout();
         }
      }
      
      if (isMenuBar)
      {
         setHeight(c.heightFromNative(menubarLines * menuItemHeight));
      }
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (config().popupOnly && config().displayed)
      {
         // determine the real mouse source
         Widget<?> source = topLevelWindow().findMouseSource(new NativePoint(e.getX(), e.getY()));
         if (source != this && !(source instanceof MenuItem || source instanceof SubMenu))
         {
            // call repaint now, to re-draw all widgets bellow it
            repaint();
            setFocusInt(null);
            setVisible(false);
         }
      }
   }
   
   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (gd != null && config().popupOnly)
      {
         int mid = UiUtils.getWidgetIdAsInt(this);

         TopLevelWindow<?> window = parent(TopLevelWindow.class);
         if (window != null && !visible)
         {
            gd.withSelectedWindow(window.getId().asInt(), () ->
            {
               gd.deregisterWidget(mid);
            });
         }
      }
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units),
    * in this menu bar, walking the entire opened menu tree, recursively.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      // search only the "title" bounds
      CoordinatesConversion cc = screen().coordinates();
      
      NativePoint menuLoc = physicalLocation();
      NativeDimension menuDim = cc.dimensionToNative(new Dimension(width(false), height()));
      NativeRectangle menuBounds = new NativeRectangle(menuLoc, menuDim);
      
      NativePoint tp = new NativePoint(p).translate(menuLoc);
      if (menuBounds.contains(tp))
      {
         return super.findMouseSource(p);
      }
      else
      {
         return findMenuMouseSource(this, p, null);
      }
   }
   
   /**
    * Looks up the very first enabled widget.
    * 
    * @return  the first enabled widget or <code>null</code>
    */
   @Override
   public Widget<GuiOutputManager> getFirstEnabledWidget()
   {
      return null;
   }
   
   /**
    * Process key event.
    *
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      int actionCode = event.actionCode();

      if (this.config().popupOnly && currentFocus() == null)
      {
         if (actionCode == Keyboard.KA_CURSOR_DOWN)
         {
            firstFocus();
         }
         else if (actionCode == Keyboard.KA_CURSOR_UP)
         {
            lastFocus();
         }

         Widget<GuiOutputManager> focused = currentFocus();
         if (focused != null)
         {
            EventManager.postEvent(new FocusEvent(focused, EventType.FOCUS_GAINED));
            repaint();
         }
         event.consume();
      }
      else
      {
         super.processKeyEvent(event);
      }
   }
   
   /**
    * Sets the navigation bracket. For more information see {@link #isKeepOpenOnBodyClose()}.
    *
    * @param   flag
    *          {@code true} when entering a menu navigation bracket and {@code false} when
    *          exiting the bracket.
    */
   public static void setKeepOpen(boolean flag)
   {
      if (flag == keepOpenOnBodyClose)
      {
         System.err.println("CRITICAL: Broken menu switching bracketing...");
      }
      keepOpenOnBodyClose = flag;
   }
   
   /**
    * Check if we are in a menu navigation bracket. This is checked when a menu is closing. When
    * this method returns {@code true} the menu bar will stay active. Otherwise, the menu was
    * closed because it was activated or it lost its focus, in which case, the menu bar will
    * no longer be active.
    *
    * @return  {@code true} when in a menu navigation bracket.
    */
   public static boolean isKeepOpenOnBodyClose()
   {
      return keepOpenOnBodyClose;
   }
   
   /**
    * Hide widget.
    */
   public void hide()
   {
      if (!isVisible())
      {
         return;
      }
      
      SubMenuGuiImpl.hideBodies(this, null, false);
      
      setFocusInt(null);
      super.hide();
      
      if (config().popupOnly)
      {
         Widget<?> owner = screen().getRegistry().getComponent(config().ownerId);
         if (owner instanceof AbstractContainer)
         {
            AbstractContainer<?> container = (AbstractContainer<?>) owner;
            if (this.equals(container.getFocusInt()))
               container.setFocusInt(null);
         }
         
         TopLevelWindow<GuiOutputManager> wnd = parentOrSelf(TopLevelWindow.class);
         if (wnd != null)
         {
            // do not let the menu be destroyed
            wnd.getContentPane().detach(this);
            wnd.close();
         }
         
         config().displayed = false;
      }
   }
   
   /**
    * The abstract portion of the logic making pop-up menu shown.
    * 
    * @param   src
    *          Source widget.
    * @param   location
    *          The popup location.
    */
   @Override
   protected void doShowPopup(Widget<GuiOutputManager> src, NativePoint location)
   {
      OverlayWindow wnd = new OverlayWindow(this, src.topLevelWindow());
      
      wnd.setDestroyingListener(() -> 
      {
         GuiDriver<?, ?> gd = (GuiDriver<?, ?>) OutputManager.getDriver();
         int id = wnd.getId().asInt();
         gd.withSelectedWindow(id, () -> gd.deregisterWidget(id));
         hide();
      });
      
      setPhysicalLocation(location.x, location.y);
      doLayout();
      setVisible(true);
      wnd.show();
      
      ThinClient.getInstance().eventDrawingBracket(this, this::repaint);
   }
   
   /**
    * Process accelerator event. If some MENUBAR sub-menu is displayed,
    * than event must be ignored.
    * 
    * @param   event
    *          KeyInput object, related to key event.
    * @param   m
    *          Target accelerator source menu-item.
    *          
    * @return  <code>true</code> means accelerator was processed.
    *          <code>false</code> means accelerator was ignored.
    */
   @Override
   protected boolean processAccelerator(KeyInput event, MenuItem<?> m)
   {   
      for (Widget<?> w : widgets())
      {
         if (w instanceof SubMenu && ((SubMenu<?>) w).config().displayed)
         {
            return false;
         }
      }
      
      return super.processAccelerator(event, m);
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @param   includeNested
    *          <code>true</code> if we should process nested sub-menus
    *          
    * @return  See above.
    */
   private double width(boolean includeNested)
   {
      double width = !config().popupOnly ? // MENUBAR ? 
                      super.width() : 
                      SubMenuGuiImpl.getMaxSubMenuItemWidth(this, includeNested);
      
      NativeInsets insets = ThemeManager.getCurrentTheme().getMenuInsets();
      CoordinatesConversion c = screen().coordinates();
      return width + c.widthFromNative(insets.left + insets.right);
   }
}