/*
** Module   : MenuItemGuiImpl.java
** Abstract : GUI client implementation of MenuItem.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150525 Created initial version.
** 002 VIG 20150818 Changed drawing title for key navigation mode.
** 003 CA  20150918 Added mouse actions.
** 004 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 005 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 006 VIG 20151027 Fixed mouseExit: reset focus when out of the body, added 'pressed' state 
**                  transferring to a neighbor in mouseExit() method, added setPressed().
** 007 VIG 20151226 Added setting of keyNavigation mode flag. 
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 CA  20160216 Cache the intermediate widget states.
** 010 VIG 20160107 Added mouse listeners for items of system pop-up menus.
**                  Added createDynamicItem(), createDynamicRule() and addMouseClickListener().
** 011 VIG 20160222 GuiLabel is renamed to MnemonicInfoGuiImpl and extract to separate class 
**                  and createNewLabel() method is renamed createLabelInfo().
** 012 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 013 VIG 20160410 Added hardcoded font setting. Added title to drawTitle() method, 
**                  added null check and changed modifiers of sizes constants.
**                  Fixed height/width methods and sizes in both cases: menubar and popup.
** 014 SVL 20160719 Updated translate() signature.
** 015 VIG 20160628 window() call is changed to topLevelWindow() in mouseExited().
**                  "pressed" state transferring is refactored, FOCUS_GAINED is moved from
**                  mouseExit() to mouseEntered(), added mbarPressedState flag setting.
** 016 VIG 20161105 getParent() call is refactored to getParentMenuContainer().
** 017 CA  20161101  Moved the mnemonic drawing code at the driver.
**     HC  20170203  Fixed a NPE when menu-item selected. When menu item is selected the menu
**                   body is closed yet before the associated menu item action is triggered.
** 018 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// This imports are explicit, to denote unwanted dependency. 
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Dimension;
import com.goldencode.p2j.ui.client.Menu;
import com.goldencode.p2j.ui.client.MenuItem;
import com.goldencode.p2j.ui.client.Point;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.FocusListener;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.widget.Container;

/**
 * <code>MenuItem</code> GUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class MenuItemGuiImpl
extends MenuItem<GuiOutputManager>
implements FocusListener
{
   /** Font resolver */
   private GuiFontResolver gf;

   /** Screen driver */
   @SuppressWarnings("rawtypes")
   private GuiDriver gd;
   
   /** Tracks mouse movements over this menu item.  */
   private boolean mouseEntered = false;

   /** 
    * TODO: move that to future subclass, intended specially for system pop-ups.
    * 
    * Set of custom mouse listeners by mouse event types like {@link MouseEvent#MOUSE_CLICKED}.
    * This mouse listeners are custom actions (like triggers) but for non-standard cases
    * such as system pop-up menus: editor pop-ups, scroll pop-ups and window title pop-ups.
    * <p>
    * For example, such listener can call clipboard operation when corresponding menu-item
    * was selected.
    */
   private Map<Integer, Set<MouseAdapter>> mouseListeners;

   /** Cached width of this item. If 0 or negative, value is invalidated and need re-computed. */ 
   private double cachedWidth = -1;
   
   /** Cached height of this item. If 0 or negative, value is invalidated and need re-computed. */ 
   private double cachedHeight = -1;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public MenuItemGuiImpl(WidgetId id, MenuItemConfig cfg)
   {
      super(id, cfg);
   }

   /**
    * Create a dynamic item with the specified title and mouse actions
    * for system pop-up menus. Such menus are used for WindowTitleBar;
    * Scrollbar; default popup for FillIn, Editor. 
    * 
    * @param    title
    *           The title of the popup item in the window's system popup menu.
    * @param    enabled
    *           Flag indicating if the item is enabled or not.
    * @param    evt
    *           The event processing.
    *           
    * @return   A {@link MenuItemGuiImpl}.
    */
   static MenuItemGuiImpl createDynamicItem(final String       title, 
                                                  boolean      enabled, 
                                            final MouseAdapter evt)
   {
      final MenuItemGuiImpl item = new MenuItemGuiImpl(null, null);
      final MenuItemConfig micfg = new MenuItemConfig();
      item.initialize(WidgetId.nextID(), micfg);
      ConfigManager.getInstance().addWidgetConfig(item.config());
      item.config().title = title;
      item.afterConfigUpdate(micfg);
      item.setEnabled(enabled);
      item.addMouseClickListener(evt);
      return item;
   }
   
   /**
    * Create a dynamic RULE item. Such menus are used for WindowTitleBar;
    * Scrollbar; default popup for FillIn, Editor.
    *  
    * @return   A {@link MenuItemGuiImpl}.
    */ 
   static MenuItemGuiImpl createDynamicRule()
   {
      final MenuItemGuiImpl item = new MenuItemGuiImpl(null, null);
      final MenuItemConfig micfg = new MenuItemConfig();
      item.initialize(WidgetId.nextID(), micfg);
      ConfigManager.getInstance().addWidgetConfig(item.config());
      item.config().subtype = MenuItemConfig.SBTP_RULE;
      return item;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @SuppressWarnings("rawtypes")
   @Override
   public void initialize(WidgetId id, MenuItemConfig cfg)
   {
      super.initialize(id, cfg);
      
      config().font = FontManager.MENU_FONT;

      gd = (GuiDriver) OutputManager.getDriver();
      
      // the label's window is resolved from the frame... the label is not necessarily attached
      // to the frame at the time of this call, just the frame ID is known

      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent != null ? parent.window() : WindowManager.resolveWindow(this);
      
      gf = new GuiFontResolver(window, gd, config(), false);
      
      addFocusListener(this);
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_RELEASED
      };
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Returns <code>true</code> only if this widget is navigable.
    *
    * @return  <code>true</code> if widget is navigable.
    */
   @Override
   public boolean navigable()
   {
      return focusTraversable();
   }
   
   /**
    * Check if widget is focus traversable. SKIP and RULE aren't traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      if (isNormal())
      {
         // menu-items are focusable even in a closed sub-menu.
         // Implementation is the same as the parents implementation, 
         // except of the parent visibility check instead of its own.
         Container<GuiOutputManager> p = parent();
         return (p != null && p.isVisible() && !hidden());
      }
      
      return  false;
   }

   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      NativePoint p = physicalLocation();
      NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      
      gd.draw(getId(), p, clip, () ->
      {
         Theme theme = ThemeManager.getCurrentTheme();
         // the widget's background is not cleared - is the parent's responsibility to draw the
         // 3D rectangle
         
         // SKIP
         if (isSkip())
         {
            return; 
         }
         else if (isRule())
         {
            // RULE
            theme.drawMenuSeparator(gd, 0, 0, dim.width, dim.height);
            return;
         }
         
         // NORMAL
         Window<?> window = WindowManager.resolveWindow(this);
         int fh = FontManager.getFontHeight(window, gf.resolveFontNum(), gd);
         
         theme.drawMenuItemBackground(this, gd, 0, 0, dim.width, dim.height, mouseEntered);
         theme.drawMenuItemContent(this, gd, gf, 0, 0, dim.width, dim.height, fh, mouseEntered);
      });
   }
   
   /**
    * Invalidate the cached size of this item. This way, the next time the height/width is
    * requested it is automatically re calculated using new environment.   
    */
   public void invalidateCachedSize()
   {
      if (cachedHeight > 0)
      {
         cachedHeight = -cachedHeight;
      }
      
      if (cachedWidth > 0)
      {
         cachedWidth = -cachedWidth;
      }
   }
   
   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      if (isSkip())
      {
         return 0;
      }
      
      if (cachedHeight > 0)
      {
         return cachedHeight;
      }
         
      Theme theme = ThemeManager.getCurrentTheme();
      int natHeight;
      
      if (isRule())
      {
         natHeight = theme.getMenuSeparatorHeight();
      }
      else
      {
         int fh = FontManager.getFontHeight(window(), gf.resolveFontNum(), gd);
         
         if (Menu.isMenubarElement(this))
         {
            natHeight = Math.max(fh + 2 * SubMenuGuiImpl.HEIGHT_PAD,
                                 theme.getMenuBarElementMinSize(false));
         }
         else
         {
            // else, normal popup item
            natHeight = Math.max(fh, theme.getMenuItemHeight(this));
         }
      }
      
      // cache the computed value before returning for fast access next time 
      cachedHeight = screen().coordinates().heightFromNative(natHeight);
      return cachedHeight;
   }

   /**
    * Get widget width. There are two cases:
    * <ul>
    *    <li>if this is a menu bar element, only the width of the text is taken into consideration
    *          (adjusted with padding and theme minimum width);</li>
    *    <li>Otherwise the item is part of a vertical menu/popup so all siblings in the parent
    *          menu are measured. In this case space is reserved, including checkmark/icon,
    *          accelerators and right arrow for submenus.</li>
    * </ul>
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      if (cachedWidth > 0)
      {
         return cachedWidth;
      }
      
      Theme theme = ThemeManager.getCurrentTheme();
      int width;
      
      if (Menu.isMenubarElement(this))
      {
         int fw = FontManager.getTextWidth(getPreparedTitle(), window(), gf.resolveFontNum(), gd);
         width = Math.max(fw + 2 * SubMenuGuiImpl.HEIGHT_PAD,
                          theme.getMenuBarElementMinSize(true));
      }
      else
      {
         AbstractContainer<GuiOutputManager> parent = getParentMenuContainer();
         // set width to max neighbor's width including possible accelerators
         width = theme.getMenuCheckmarkWidth() +
               theme.getMenuLeftPad() +
               SubMenuGuiImpl.getLongestItemTitleWidth(parent, null, gf, gd) +
               theme.getMenuRightPad() +
               SubMenuGuiImpl.getLongestItemAccelWidth(parent, gf, gd) +
               theme.getMenuSmArrowWidth();
      }
      
      // cache the computed value before returning for fast access next time 
      cachedWidth = screen().coordinates().widthFromNative(width);
      return cachedWidth;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      int actionCode = event.actionCode();
      boolean keyNavigation = (actionCode == Keyboard.SE_ENTRY       ||
                               actionCode == Keyboard.KA_CURSOR_DOWN || 
                               actionCode == Keyboard.KA_CURSOR_UP   ||
                               actionCode == Keyboard.KA_CURSOR_LEFT ||
                               actionCode == Keyboard.KA_CURSOR_RIGHT);
                            
      ((MenuGuiImpl) Menu.getMenu(this)).setKeyNavigation(keyNavigation);
      super.processKeyEvent(event);
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
         return;
      
      if (mouseListeners != null && mouseListeners.containsKey(MouseEvent.MOUSE_CLICKED))
      {
         for (MouseAdapter adapter : mouseListeners.get(MouseEvent.MOUSE_CLICKED))
         {
            adapter.mouseClicked(e);
         }
      }
      
      selectMenuItem();
   }
   
   /**
    * Notification of a mouse entered event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      mouseEntered = true;

      if (Menu.isMenubarElement(this))
      {
         MenuGuiImpl.setMbarPressedState(this, MenuGuiImpl.isMbarPressedState());
      }
      
      ((MenuGuiImpl) Menu.getMenu(this)).setKeyNavigation(false);
      
      if (!isRule() && !hasFocus())
      {
         requestFocus();
      }
      
      repaint();
   }

   /**
    * Notification of a mouse exited event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (mouseEntered)
      {
         // reset the focus if it is not set to some neighbor 
         if (parent().focus() != null && parent().focus().getId().asInt() == this.getId().asInt())
         {
            parent().setFocus(null);
         }
         mouseEntered = false;
      }

      EventManager.postEvent(new FocusEvent(this, EventType.FOCUS_LOST));
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      boolean oldVal = MenuGuiImpl.isMbarPressedState(); 
      setPressed(!oldVal);
      
      if (!oldVal)
      {
         repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      boolean oldVal = MenuGuiImpl.isMbarPressedState();
      setPressed(!oldVal);
      
      if (oldVal)
      {
         repaint();
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      boolean wasIn = mouseEntered;
      
      // get coords of the drag event related to button
      NativePoint mp = translate(e);
      // get button location and size
      NativeDimension d = physicalDimension();
      
      mouseEntered = (0 <= mp.x && mp.x <= d.width && 0 <= mp.y && mp.y <= d.height);
      if (wasIn != mouseEntered)
      {
         repaint();
      }
   }
   
   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      Widget<?> currentFocus = UiUtils.getCurrentFocus();
      
      if (currentFocus == this)
         return;
      
      MenuGuiImpl.requestFocusWorker(this);

      super.requestFocus();
   }
   
   /**
    * This method is invoked when focus is gained.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      // hide all other active submenu bodies
      SubMenuGuiImpl.hideBodies(parent(), null, true);
      
      boolean keyNav = ((MenuGuiImpl) Menu.getMenu(this)).isKeyNavigation();
      if (keyNav && Menu.isMenubarElement(this))
      {
         MenuGuiImpl.setMbarPressedState(this, true);
      }
      
      parent().repaint();
   }
   
   /**
    * This method is invoked when focus is lost. If current focus is on in subtree 
    * of this sub-menu, than hide body.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onFocusLost(FocusEvent event)
   {
      mouseEntered = false;
      parent().repaint();
   }
   
   /**
    * Returns a string representation of the object.
    * 
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      return super.toString() + " title=" + config.title;
   }

   /**
    * Set pressed flag.
    * 
    * @param   pressed
    */
   public void setPressed(boolean pressed)
   {
      this.mouseEntered = pressed;
      
      if (Menu.isMenubarElement(this))
      {   
         MenuGuiImpl.setMbarPressedState(this, pressed);
      }
   }
   
   /**
    * Process an event which trigger selection of this menu-item, depending if the item is a 
    * toggle-box or not.
    */
   @Override
   protected void selectMenuItem()
   {
      if (!config.toggleBox)
      {
         // close the parent body yet before the associated menu item action is triggered
         SubMenu<GuiOutputManager> sm = getParentSubMenu();
         if (sm != null)
         {
            // process the events related to the body being closed, this makes sure
            // the body goes away from the screen before the menu item action is triggered
            ThinClient.getInstance().eventBracket(() -> sm.hideBody(true));
         }
      }

      super.selectMenuItem();
      
      MenuGuiImpl menu = (MenuGuiImpl) Menu.getMenu(this);
      if (!menu.config().popupOnly)
      {
         MenuGuiImpl.resetMbarPressedState();
         menu.setKeyMode(false);
      }
   }
   
   /**
    * Creates new MnemonicInfo instance
    * 
    * @return   MnemonicInfo or its descendant.
    */
   @Override
   protected MnemonicInfo createLabelInfo()
   {
      return new MnemonicInfoGuiImpl();
   }
   
   /**
    * Trick, that lets to correct parent location in cases, when it needed,
    * for example for menus.
    * 
    * @param    parent
    *           Parent component - must be a sub-menu.
    *           
    * @return   corrected parent location.
    */
   @Override
   protected Point prepParentLocation(Widget<?> parent)
   {
      if (!(parent instanceof SubMenu))
      {
         return super.prepParentLocation(parent);
      }
      
      Window<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion c = wnd.screen().coordinates();
      NativePoint p = c.pointToNative(parent.location());
      
      if (!Menu.isMenubarElement(this))
      {
         p.x += c.widthToNative(this.width());
      }
      else
      {
         // under the title
         p.y += c.heightToNative(
                  SubMenuGuiImpl.titleHeightWithBorders((AbstractContainer<?>) parent));  
         p.x += 2; // shift
      }
      
      return c.pointFromNative(p); 
   }
   
   /**
    * Adds custom mouse click listener for this item
    * 
    * @param   listener
    *          {@link MouseAdapter} instance.
    */
   private void addMouseClickListener(MouseAdapter listener)
   {
      if (listener == null)
         throw new NullPointerException("listener must not be null");
      
      if (mouseListeners == null)
      {
         mouseListeners = new HashMap<>();
      }
      
      if (!mouseListeners.containsKey(MouseEvent.MOUSE_CLICKED))
      {
         mouseListeners.put(MouseEvent.MOUSE_CLICKED, new HashSet<>());
      }
      
      mouseListeners.get(MouseEvent.MOUSE_CLICKED).add(listener);
   }
}