/*
** Module   : MessageAreaGuiImpl.java
** Abstract : GUI message area implementation. 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 CA  20141021 Removed the repaint() API: all repainting needs to be done via the common
**                  event processing; do no draw the widget immediately.
** 003 MAG 20141017 Set box option. Font management improvements.
** 004 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 005 CA  20141123 Added scroll support via cursor keys.
** 006 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 007 CA  20150507 Fixed height.
** 008 HC  20150508 Improvements to GUI window scrolling support.
** 009 HC  20150612 Layout fix.
** 010 HC  20150831 Scrolling support overhaul.
** 011 CA  20150910 Added mouse actions.
** 012 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 013 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 014 EVL 20151013 Fix for height size issue.
** 015 IAS 20160331 Fixed potential NPE
** 016 HC  20160712 Improved next/prev tab focus in nested frames.
** 017 SBI 20161006 Fixed the message text size if the owner window is resized.
** 018 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.Bars;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.model.*;

/**
 * GUI message area is a scroll pane widget having a vertical scroll bar and a scrollable list
 * which holds messages. 
 */
@SuppressWarnings("unchecked")
public class MessageAreaGuiImpl
extends ScrollPaneGuiImpl
implements MessageArea<GuiOutputManager>
{
   /** Maximum number of lines in message area */
   private static final int MAX_LINES = 50;
   
   /** Ancestor */
   private WindowGuiImpl window;
   
   /** Screen driver */
   private GuiDriver gd;

   /** Font resolver */
   private GuiFontResolver gf;

   /** Messages */
   private ScrollableListGuiImpl list;
      
   /** Message text maximum width in pixels */
   private int maxTextWidth = -1;
   
   /** Message text width and height in character units */
   private Dimension textDimChars;
   
   /** Message area lines */
   private int lines;
   
   /**
    * Constructor.
    * 
    * @param   rows
    *          Number of visible rows.
    */
   public MessageAreaGuiImpl(WidgetId id, int rows)
   {
      super(new ScrollableListGuiImpl(WidgetId.nextID(), new ListModel<>(), false));
      lines = rows;
      gd = (GuiDriver) OutputManager.getDriver();
      // list widget
      list = (ScrollableListGuiImpl) getChild();
      list.config().sysbgcolor = SysColor.COLOR_APPWORKSPACE;
      list.config().font = config().font;
      list.setVisibleRows(lines);
      list.setInheritWidth(true);
      list.setInheritHeight(true);
      // initialize container
      setScrollBars(Bars.VERTICAL);
      setHeight(rows);
      
      // setup panel border and background
      GuiColorResolver gc = new GuiColorResolver(gd,
                                                 list.config(), 
                                                 SysColor.COLOR_3DSHADOW, 
                                                 SysColor.COLOR_WINDOWTEXT);
      setBorder(new LineBorderGuiImpl(null, gc));
      config().sysbgcolor = SysColor.COLOR_APPWORKSPACE;
   }
   
   /**
    * Widget drawing method.
    */   
   @Override
   public void draw()
   {
      if (window != null)
      {
         gf.font();
         // check if font has been changed.
         // when font is changed the height of message area should be adjusted.
         if (gf.fontChanged)
         {
            // TODO: message certainly shouldn't be responsible for laying out the window,
            //       move this call some else
            window.doLayout();
         }
         // draw message area
         super.draw();
      }            
   }

   /**
    * Attach widget to window.
    * 
    * @param   window
    *          parent window.
    */
   public void attach(Window<GuiOutputManager> window)
   {
      this.window = (WindowGuiImpl) window;
      config().font = this.window.resolveMessageAreaFont();
      list.config().font = config().font;
      gf = new GuiFontResolver(window, gd, config(), false);
   }
   
   /**
    * Add a new message text to message area.
    * 
    * @param   message
    *          Message text.
    */
   public void addMessage(String message)
   {
      int max = 0;

      if (message != null)
      {
         String tmp = message;
         while (!tmp.isEmpty())
         {
            // don't try to split one character, this may cause an endless loop
            if (tmp.length() > 1)
            {
               int len = split(tmp);
               max = addItem(tmp.substring(0, len));         
               tmp = tmp.substring(len);
            }
            else
            {
               max = addItem(tmp);
               break;
            }
         }
      }
   }
   
   /**
    * Add new item to message list.  Delete old messages.
    * 
    * @param   message
    *          New message.
    *          
    * @return  First visible row. Could be negative.         
    */
   private int addItem(String message)
   {
      ListModel<String> model = list.model();      
      
      if (model.size() >= MAX_LINES)
      {
         model.removeAt(0);
      }
      
      return list.addItem(message);
   }
   
   /**
    * Clear all messages. This method corresponds to the Progress statement:
    * <pre>
    * HIDE MESSAGE
    * </pre>
    */
   public void clearMessages()
   {
      list.deleteItems();
   }
   
   /**
    * Returns the number of screen lines available for message display.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * MESSAGE-LINES
    * </pre>
    *
    * @return   The number of available lines in the message area.
    */
   public int getMessageLines()
   {
      return lines;
   }
   
   /**
    * Lay out container.
    */
   @Override   
   public void doLayout()
   {
      super.doLayout();
      
      // calc width
      double width = window.workspace.width();
      
      // calc height
      double height = calculateMessageAreaHeight();
      
      setSize(width, height);
      rootPanel.setSize(width, height);
      
      // its window is resized
      maxTextWidth = physicalDimension().width;
   }
   
   /**
    * Calculates the height of the Message Area in the characters units.
    * 
    * @return   The height of the Message Area in the characters units
    */
   public double calculateMessageAreaHeight()
   {
      CoordinatesConversion cc = screen().coordinates();
      int nativeItemHeight = FontManager.getFontHeight(WindowManager.resolveWindow(this),
                                                       gf.resolveFontNum(), gd);
      
      return cc.heightFromNative(nativeItemHeight * lines + 4);
   }
   
   /**
    * Get all the message texts in this message-area.
    * 
    * @return   See above.
    */
   public String[] getMessageTexts()
   {
      String[] msgs = null;

      ListModel<String> model = list.model();
      int size;
      
      if (model != null && (size = model.size()) > 0)
      {
         msgs = new String[size];            
         for (int i = 0;i < size; i++)
         {
            msgs[i] = model.get(i);
         }
      }
      else
      {
         msgs = new String[0];         
      }
      
      return msgs;
   }

   /**
    * Set message main worker routine.
    * 
    * @param   msg
    *          Message text array.
    * @param   cs
    *          Message color array.
    * @param   force
    *          <code>true</code> means force redisplaying of messages even
    *          if message text is already <code>null</code>.
    */
   @Override
   public void setMessageText(final String[] msg, Color[] cs, boolean force)
   {
      ThinClient.getInstance().eventDrawingBracket(this, false, true, () ->
      {
         if (msg == null)
         {
            clearMessages();
         }
         else
         {   
            for (String message : msg)
            {
               if (message != null)
               {
                  addMessage(message);
               }
            }
         }
         
         parent().moveToTop(MessageAreaGuiImpl.this);
         repaint();
      });
   }

   /**
    * Get all the message colors in this message-area.
    * 
    * @return   See above.
    */
   @Override
   public Color[] getMessageColors()
   {
      return null;
   }

   /**
    * Shows the new message text.
    * 
    * @param   message
    *          Array of two text message which need to be displayed inside the
    *          message area.
    * @param   cs
    *          Array of two color specifications.
    */
   @Override
   public void setMessageText(String[] message, Color[] cs)
   {
      setMessageText(message, cs, false);
   }
   
   /**
    * Clear the message content and make sure that the message area displays the cleared
    * space without pausing.
    * <p>
    * In GUI, clearing the messages and repainting should be enough since this will not
    * trigger any pausing.
    */
   @Override
   public void clearMessagesNoPause()
   {
      setMessageText(null, null);
   }
   
   /**
    * This component will not receive focus when Tab or Shift-Tab is pressed.
    * 
    * @return  <code>false</code>
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED
      };
   }

   /**
    * Split message.  If message text is larger than message area width the message is split into
    * multiple lines each one smaller or equal than message area width.  The computation is done 
    * using native units (pixels).  When proportional fonts are used the length in characters 
    * units could be different on each lines.
    *  
    * @param   text
    *          Text to split.
    *          
    * @return  split index inside text.
    */
   private int split(String text)
   {
      int textWidth = getTextWidth(text);      

      if (textWidth <= maxTextWidth)
      {
         return text.length();
      }
      
      int length = text.length() / 2;
      int step = length;
            
      do
      {
         // check text width
         textWidth = getTextWidth(text.substring(0, length));
         if (textWidth == maxTextWidth)
         {
            return length;
         }
         
         // next test
         step = step / 2;                 
         length = length + (textWidth > maxTextWidth ? -step : step);

      }
      while (step != 0);

      // final adjust
      if (textWidth < maxTextWidth) 
      {
         while (textWidth <= maxTextWidth)
         {
            textWidth = getTextWidth(text.substring(0, ++length));
         }
         length--;
      }
      else if (textWidth > maxTextWidth)
      {
         while (textWidth > maxTextWidth)
         {
            textWidth = getTextWidth(text.substring(0, --length));
         }
         length++;
      }

      return length;
   }
   
   /**
    * Get text width.
    * 
    * @param   text
    *          String text.
    *          
    * @return  see above.
    */
   private int getTextWidth(String text)
   {
      return getTextWidthNative(text, gf.resolveFontNum());
   }
}