/* 
** Module   : ModalWindow.java
** Abstract : Base class for modal windows
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20150612 Initial version.
** 002 HC  20150622 Fixed GUI frame to be properly centered in its parent window.
** 003 HC  20150626 Javadoc fixes.
** 004 HC  20150702 Implemented window modality independently on the actual UI driver.
** 005 HC  20150831 Scrolling support overhaul.
** 006 CA  20150826 Added init().  Added message area getter (returns null).
** 007 CA  20150925 MOUSE_CLICKED event must use the same source as MOUSE_PRESSED.
**                  Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 008 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 009 CA  20151015 Fixed window resize via attributes (they target the workspace, not the full 
**                  window).
**     HC  20151020 Made modal windows to draw in batch. Added 3D border to ALERT-BOX and 
*                   DIALOG-BOX.
** 010 HC  20151110 Changed ALERT-BOX owner to be always the DEFAULT-WINDOW.
** 011 CA  20151127 When losing focus, the window must not change the ACTIVE-WINDOW.
** 012 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 013 EVL 20160124 Adding method to check if the window has title bar.
** 014 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes. 
** 015 IAS 20160217 Additional argument in realizeChileWindow() call 
** 016 HC  20160304 Replaced Window.moveBy() with setPhysicalLocation(). Window layout
**                  improvements.
** 017 CA  20160327 When drawing the border, check if the invalidation rectangles registered in
**                  ScreenBitmap intersect it or not. 
**                  Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 018 SBI 20160322 Changed to hide a window title bar and a border if the graphics environment
**                  doesn't support window decorations.
** 019 HC  20160406 Overhaul of window-activation logic.
** 020 VIG 20160307 Added static create* methods for system popups and those deregistration.
** 021 HC  20160413 Fixed top-level windows to always receive focus event when activated, also
**                  deduplicated processing of focus events.
** 022 HC  20160428 GUI window border drawing changed to use the Border interface mechanism.
**                  Changed modal window border to be one pixel thinner.
** 023 HC  20160502 Fixed a regression in findMouseSource().
** 024 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 025 SBI 20160610 Changed to register mouse popupable_right and popapable_left window title
**                  bar's widgets regions to be known by the Web client.
** 026 SBI 20160630 Applied HC changes to fix a next active window.
** 027 SBI 20160811 Corrected the location of the displayed dialog according to Progress 4GL which
**                  makes the whole dialog visible if it can't be centered within its owner,
**                  changed to set the dialog's owner to the current window.
** 028 CA  20170104 Some improvements in window decoration support for embedded mode (modal 
**                  windows always have decorations).  Added flag to track modal state on driver 
**                  side.
** 029 HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The class implements GUI modal top level window. When a modal window
 * is visible it blocks user input to all other application windows. 
 */
public class ModalWindow 
extends TopLevelWindow<GuiOutputManager>
implements GuiWindow
{
   /** Window id. */
   private WidgetId id;
   
   /** Window outer frame. */ 
   protected BorderedPanel<GuiOutputManager> windowPane;

   /** Inner content pane. */
   protected BorderedPanel<GuiOutputManager> contentPane;
   
   /** The title-bar widget for this window. */
   private WindowTitleBar titleBar;
   
   /** Cached GUI driver reference. */  
   protected GuiDriver<?,?> gd;

   /** The last mouse pressed determined for this window. */
   private Widget<GuiOutputManager> mousePressedSource = null;
   
   /** Destroy flag. */
   private boolean destroyInProgress;
   
   /** Cached realized flag. */
   private boolean realized;
   
   /**
    * Default constructor.
    * 
    * @param    title
    *           Window title.
    */
   public ModalWindow(String title)
   {
      this(title, resolveOwner());
   }
   
   /**
    * Default constructor.
    * 
    * @param    title
    *           Window title.
    * 
    * @throws   IllegalArgumentException
    *           when <code>owner</code> is <code>null</code>.
    */
   public ModalWindow(String title, TopLevelWindow<GuiOutputManager> owner)
   {
      super(title);
      
      if (owner == null)
      {
         throw new IllegalArgumentException("owner must not be null!");
      }
      
      id = WidgetId.nextID();
      gd = (GuiDriver<?,?>) OutputManager.getDriver();
      
      setOwner(owner);
      setLayout(new WindowLayout());
      
      GuiWidgetFactory factory = screen().getFactory();
      
      windowPane = factory.createBorderedPanel();
      
      if (hasDecorations())
      {
         titleBar = new WindowTitleBar(WidgetId.nextID(),
                                       false,
                                       new CaptionButtonType[] {CaptionButtonType.CLOSE},
                                       title);
         // window border
         windowPane.setBorder(WindowBorder.createModalWindowBorder());
         windowPane.add(titleBar);
         titleBar.registerWindowTitlePopup();
      }
      
      contentPane = factory.createBorderedPanel();
      
      windowPane.add(contentPane);
      add(windowPane);
      
      screen().getRegistry().addWidgetsRecursive(this);
   }
   
   /**
    * Returns the expected owner of a modal window.
    * 
    * @return   a valid {@link TopLevelWindow} instance.
    */
   public static TopLevelWindow<GuiOutputManager> resolveOwner()
   {
      // by default the modal window is parented to the active top-level window, 
      // not the focus owner
      TopLevelWindow<GuiOutputManager> owner = WindowManager.topLevelWindow();
      if (owner == null)
      {
         owner = WindowManager.getDefaultWindow();
      }
      
      return owner;
   }
   
   /**
    * Determine if this window can have decorations.
    * 
    * @return   Always <code>true</code> - all modal windows have decorations, regardless if 
    *           embedded mode or not. 
    */
   @Override
   public boolean hasDecorations()
   {
      return true;
   }
   
   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      WindowLayout wl = (WindowLayout) getLayout();
      if (wl == null)
      {
         return;
      }

      wl.setDecorated(hasDecorations());
      wl.doLayout(this);
   }
   
   /**
    * Widget ID.
    * 
    * @return  see above.
    */
   @Override
   public WidgetId getId()
   {
      return id;
   }
   
   /**
    * Get window content pane.
    * 
    * @return  content pane.
    */
   @Override
   public Container<GuiOutputManager> getContentPane()
   {
      return contentPane;
   }
   
   /**
    * The method returns <code>true</code> if this represents a modal window. 
    * That is a window that sits on top of other windows and takes exclusive user 
    * input.
    * <p>
    * This implementation always returns <code>true</code>.
    * 
    * @return   See above.
    */
   public boolean isModal()
   {
      return true;
   }
   
   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      // select this window
      Object state = gd.saveDrawingState();
      gd.selectWindow(getId().asInt());
      
      try
      {
         gd.startBatch();
         if (!hasDecorations())
         {
            GuiColorResolver gc = getColorResolver();
            if (gc != null)
            {
               gc.colors();
               gd.setDesktopBgColor(gc.bgColor);
            }
         }
         super.draw();
      }
      finally
      {
         gd.endBatch();
         gd.releaseWindow();
         
         // restore previous window
         gd.restoreDrawingState(state);
      }
   }
      
   /**
    * Returns true when the modal window has been realized.
    * 
    * @return   See above.
    */
   @Override
   public boolean isRealized()
   {
      return realized;
   }

   /**
    * The method realizes this modal window. If the window has
    * been realized, this method is a no-op.
    */
   public void realize()
   {
      if (realized)
      {
         return;
      }
      
      final TopLevelWindow<GuiOutputManager> owner = getOwner();
      if (owner == null)
      {
         throw new IllegalStateException("null owner is unexpected!");
      }
      
      WindowManager.addWindow(this);
      screen().realizeChildWindow(owner.getId().asInt(), id.asInt(), title(), false, 
                                  isShareActivationWithOwner(), true);
      
      realized = true;
      
      registerWidgetStateListener(new WidgetStateListener()
      {
         @Override
         public void notifyWidgetDestroyed(Widget<? extends OutputManager<?>> w)
         {
            realized = false;
            
            WindowManager.remove(ModalWindow.this);
            gd.deregisterWindow(w.getId().asInt());
         }
      });
      
      // can be missing if there are no decorations
      if (hasTitlebar())
      {
         gd.selectWindow(getId().asInt());
         gd.registerMoveableWidget(getTitleBar().getId().asInt());
         gd.releaseWindow();
      }
   }
   
   /**
    * Show window.
    */
   @Override
   public void show()
   {
      if (isVisible())
         return;
      setOwner(WindowManager.getCurrentWindow());
      TopLevelWindow<?> owner = getOwner();
      
      gd.selectWindow(owner.getId().asInt());
      NativePoint ownerLocation = gd.getWindowLocation();
      gd.releaseWindow();
      
      gd.selectWindow(id.asInt());

      // can be missing if there are no decorations
      if (hasTitlebar())
      {
         gd.registerMoveableWidget(getTitleBar().getId().asInt());
      }
      
      realize();
      forceVisibility(true);
      doLayout();
      
      NativeDimension dim = physicalDimension();
      
      // physicalLocation() never returns invalid location
      NativePoint loc = physicalLocation();
      
      // center the window relative to its owner when location not assigned
      if (location.x == BaseConfig.INV_COORD || location.y == BaseConfig.INV_COORD)
      {
         NativeDimension ownerDimension = owner.physicalDimension();
         
         if (location.x == BaseConfig.INV_COORD)
         {
            loc.x = Math.max(ownerLocation.x + (ownerDimension.width - dim.width) / 2, 0);
         }

         if (location.y == BaseConfig.INV_COORD)
         {
            loc.y = Math.max(ownerLocation.y + (ownerDimension.height - dim.height) / 2, 0);
         }
         
         setPhysicalLocation(loc.x, loc.y);
      }
      
      gd.setWindowLocation(loc.x, loc.y);
      
      WindowManager.modalizeWindow(this);
      
      gd.setWindowVisible(true);
      gd.releaseWindow();
      
      // activate us, the driver will send WindowActivated event and the
      // handler of that event will manage the z-order
      WindowManager.forceActivateWindow(this);
   }

   /**
    * Hide window.
    */
   @Override
   public void hide()
   {
      if (!isVisible())
         return;
      
      // The order of super.hide() and GuiWindowCommons.hide(this) is important, otherwise
      // an incorrect window will be activated by GuiWindowCommons.hide(this).
      super.hide();
      GuiWindowCommons.hide(this);
      
      // can be missing if there are no decorations
      if (hasTitlebar())
      {
         gd.selectWindow(id.asInt());
         gd.deregisterWidget(getTitleBar().getId().asInt());
         gd.releaseWindow();
      }
      
      // demodalize and remove the window from window manager
      WindowManager.removeNoDestroy(this);
      
      realized = false;
   }

   /**
    * Get the message area widget associated with this window.
    * 
    * @return   See above.
    */
   @Override
   public MessageAreaGuiImpl getGuiMessageArea()
   {
      return null;
   }

   /**
    * Iconify the window.
    */
   @Override
   public void iconify()
   {
      // modal windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   @Override
   public void iconify(int sourceId)
   {
      // modal windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Iconify the window.
    *
    * @param sourceId  The root ID of the sub-tree which was iconified, and triggered
    *                  iconification of
    *                  this window.
    * @param postEvent When <true>true</true> the method posts the WINDOW-MINIMIZED legacy event,
    */
   @Override
   public void iconify(int sourceId, boolean postEvent)
   {
      // modal windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    */
   @Override
   public void restore()
   {
      // modal windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   @Override
   public void restore(int sourceId)
   {
      // modal windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    *
    * @param sourceId  The root ID of the sub-tree which was iconified, and triggered
    *                  iconification of
    *                  this window.
    * @param postEvent When <true>true</true> the method posts the WINDOW-RESTORED legacy event,
    */
   @Override
   public void restore(int sourceId, boolean postEvent)
   {
      // modal windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Maximize this window.
    */
   @Override
   public void maximize()
   {
      // modal windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Maximize this window.
    *
    * @param postEvent When <true>true</true> the method posts the WINDOW-MAXIMIZED legacy event,
    *                  <code>false</code> means no event is posted.
    */
   @Override
   public void maximize(boolean postEvent)
   {
      // modal windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Resize the widget to the specified dimensions.
    * 
    * @param    origin
    *           The top-left window corner new position.
    * @param    size
    *           The new window size.
    */
   @Override
   public void resizeTo(NativePoint origin, NativeDimension size)
   {
      // modal windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Returns the output window's coordinates conversion service.
    * 
    * @return   a reference to the {@link CoordinatesConversion} instance. 
    */
   @Override
   public CoordinatesConversion coordinates()
   {
      return screen().coordinates();
   }
   
   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      super.setSize(width, height);
      windowPane.setSize(width, height);
   }

   /**
    * Determine if this window is being resized.
    * 
    * @return   This implementation always returns <code>false</code>.
    */
   @Override
   public boolean isResizing()
   {
      return false;
   }

   /**
    * Determine if this window workspace needs to be resized.
    * 
    * @return   This implementation always returns <code>false</code>.
    */
   public boolean isResizeWorkspace()
   {
      return false;
   }

   /**
    * Determine the resized window size.
    * 
    * @return   The new size.
    */
   @Override
   public NativeDimension getResizedSize()
   {
      // modal windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Check if window has title bar.  This should return true for GUI window even if the
    * title length is zero.
    * 
    * @return  <code>true</code> if window has title to repaint.
    */
   @Override
   public boolean hasTitlebar()
   {
      return titleBar != null;
   }
   
   /**
    * Get the title bar widget associated witn this window.
    * 
    * @return   See above.
    */
   @Override
   public WindowTitleBar getTitleBar()
   {
      return titleBar;
   }
   
   /**
    * Get the message area widget associated with this window.
    * 
    * @return   Always <code>null</code>, as modal windows don't have a message area.
    */
   @Override
   public MessageArea<GuiOutputManager> getMessageArea()
   {
      return null;
   }

   /**
    * Get the status area widget associated with this window.
    * 
    * @return   See above.
    */
   @Override
   public StatusLine<GuiOutputManager> getStatusArea()
   {
      return null;
   }
   
   /**
    * Get the size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getWorkspaceSize()
   {
      return contentPane.dimension();
   }

   /**
    * Get the screen location of the window workspace area, in native units.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint getWorkspaceScreenLocation()
   {
      return contentPane.screenPhysicalLocation();
   }
   
   /**
    * Get the native size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWorkspaceNativeSize()
   {
      return contentPane.physicalDimension();
   }

   /**
    * Sets the native size of the window workspace area.
    * 
    * @param    width
    *           New workspace width.
    * @param    height
    *           New workspace height.
    */
   @Override
   public void setWorkspaceNativeSize(int width, int height)
   {
      // modal windows in 4GL can't be resized
      // this is a no-op
   }
   
   /**
    * Sets the window workspace location.
    *  
    * @param    x
    *           The x location. 
    * @param    y
    *           The y location.
    */
   @Override
   public void setWorkspaceLocation(int x, int y)
   {
      contentPane.setPhysicalLocation(x, y);
   }
   
   /**                                     
    * Get dimension of screen area available for the placement of frames.
    * 
    * @return  Instance of <code>Dimension</code> which describes size of the
    *          available screen area.
    */
   @Override
   public Dimension getScreenDimension()
   {
      return getContentPane().dimension();
   }
   
   /**
    * Find the widget to which the given mouse event can be applied.
    * 
    * @param    evt
    *           The mouse event.
    * 
    * @return   A widget over which the mouse is positioned or {@code null} if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(MouseEvt evt)
   {
      Widget<GuiOutputManager> widget = null;
      
      int evtID = evt.getEvent().getID();
      
      if (evtID == MouseEvent.MOUSE_DRAGGED  || 
          evtID == MouseEvent.MOUSE_RELEASED ||
          evtID == MouseEvent.MOUSE_CLICKED)
      {
         return mousePressedSource;
      }
      
      try
      {
         java.awt.Point p = evt.getEvent().getPoint();
         
         // search the other widgets
         widget = windowPane.findMouseSource(new NativePoint(p.x, p.y));
         
         widget = (widget == null ? this : widget);
      }
      finally
      {
         if (widget != null && evtID == MouseEvent.MOUSE_PRESSED)
         {
            mousePressedSource = widget;
         }
         
         if (evtID == MouseEvent.MOUSE_CLICKED)
         {
            mousePressedSource = null;
         }
      }
      
      return widget;
   }

   /**
    * Get a point to represent the LAST-EVENT:X and LAST-EVENT:Y coordinates.  For modal windows,
    * this returns always <code>null</code>.
    * 
    * @param    point
    *           The point (relative to the window's top-left corner) for which the coordinates for
    *           the portable mouse events are needed.
    * @return   Always <code>null</code>.
    */
   @Override
   public NativePoint getPortableMouseEventCoordinates(NativePoint point)
   {
      return null;
   }
   
   /**
    * Destroy this window.
    */
   @Override
   public void destroy()
   {
      // guard against recursive destroy - destroy is called from AbstractContainer.remove() 
      // as well as from WindowManager.remove()
      if (destroyInProgress)
         return;
      
      destroyInProgress = true;
      
      if (isVisible())
      {
         hide();
      }
      
      super.destroy();
   }
   
   /**
    * Get the maximum size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getMaxWorkspaceSize()
   {
      return getWorkspaceSize();
   }
   
   /**
    * Returns the window minimum size.
    * 
    * @return  See above.
    */
   @Override
   public NativeDimension getMinSize()
   {
      return new NativeDimension(0, 0);
   }   
   
   /**
    * Returns the window's border implementation. If <code>null</code>
    * is returned, the window has no border.
    * 
    * @return   A border implementation or <code>null</code>.
    */
   @Override
   public Border<GuiOutputManager> getBorder()
   {
      return windowPane.border(); 
   }
   
   /**
    * Get the color resolver.
    * 
    * @return   <code>null</code> by default.
    */
   protected GuiColorResolver getColorResolver()
   {
      return null;
   }
}