/* 
** Module   : OverlayWindow.java
** Abstract : A kind of window to be used to draw outside the main window
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20151216 Initial version.
** 002 EVL 20160118 Made overlay child implement a generic set of methods.
**         20160121 Improved overlay deactivation code.
** 003 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 004 IAS 20160217 Additional argument in realizeChileWindow() call 
** 005 HC  20160304 Replaced Window.moveBy() with setPhysicalLocation(). Window layout
**                  improvements.
** 006 CA  20160327 Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 007 IAS 20160331 Fixed potential NPE
** 008 HC  20160406 COMBO-BOX drop-down resources cleanup. Various drop-down related fixes.
** 009 VIG 20160401 Added implementations of methods, intended for system popup's creation.
** 010 HC  20160413 Fixed focusing of main window when overlay window closes. Fixed top-level
**                  windows to always receive focus event when activated, also deduplicated
**                  processing of the focus event.
** 011 HC  20160428 GUI window border drawing changed to use the Border interface mechanism. Also
**                  removed hardcoded assumptions of the border size.
** 012 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 013 VIG 20160831 Added isWindowActive check to hide() method.
** 014 CA  20170104 Some improvements in window decoration support for embedded mode (modal 
**                  windows always have decorations).
** 015 HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE.
** 016 OM  20170320 Fixed typos in javadoc.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

//this usage of AWT events is intentional, they are being used separately from the
//AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The class implements GUI top level window that can overlay the main window.  Modality feature
 * is disabled unconditionally for this class.
 */
public class OverlayWindow 
extends TopLevelWindow<GuiOutputManager>
implements GuiWindow
{
   /** Window id. */
   private WidgetId id;
   
   /** Overlay window single content pane. */ 
   protected BorderedPanel<GuiOutputManager> windowPane;

   /** Cached GUI driver reference. */  
   protected GuiDriver<?,?> gd;

   /** The last mouse pressed determined for this window. */
   private Widget<GuiOutputManager> mousePressedSource = null;
   
   /** Cached realized flag. */
   private boolean realized;
   
   /** The callback called when window is being destroyed. */ 
   private Runnable destroyingListener;

   /**
    * Default constructor.
    * 
    * @param    child
    *           Widget to put inside the overlay window.
    */
   public OverlayWindow(Widget<GuiOutputManager> child)
   {
      // by default the overlay window is parented to the active top-level window, 
      // not the focus owner
      this(child, WindowManager.topLevelWindow());
   }
   
   /**
    * Default constructor.
    * 
    * @param    child
    *           Widget to put inside the overlay window.
    * @param    owner
    *           Window that is the owner of the child widget used inside overlay window.
    */
   public OverlayWindow(Widget<GuiOutputManager> child, TopLevelWindow<GuiOutputManager> owner)
   {
      super();
      
      id = WidgetId.nextID();
      gd = (GuiDriver<?,?>) OutputManager.getDriver();
      
      setOwner(owner);
      setLayout(null);
      
      GuiWidgetFactory factory = screen().getFactory();
      
      windowPane = factory.createBorderedPanel();
      windowPane.setBorder(null);
      windowPane.setLayout(null);
      windowPane.add(child);
      add(windowPane);
      
      // this is required to get the mouse events for child widgets inside overlay
      screen().getRegistry().addWidgetsRecursive(this);

      // register the window with renderer
      realize();
   }
   
   /**
    * Determine if this window can have decorations.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   public boolean hasDecorations()
   {
      return true;
   }
   
   /**
    * Widget ID.
    * 
    * @return  see above.
    */
   @Override
   public WidgetId getId()
   {
      return id;
   }
   
   /**
    * Get window content pane.
    * 
    * @return  content pane.
    */
   @Override
   public Container<GuiOutputManager> getContentPane()
   {
      return windowPane;
   }
   
   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      if (!isVisible())
      {
         return;
      }
      
      // select this window
      Object state = gd.saveDrawingState();
      gd.selectWindow(getId().asInt());
      
      try
      {
         gd.startBatch();
         super.draw();
      }
      finally
      {
         gd.endBatch();
         gd.releaseWindow();
         
         // restore previous window
         gd.restoreDrawingState(state);
      }
   }
      
   /**
    * Returns true when the overlay window has been realized.
    * 
    * @return   See above.
    */
   @Override
   public boolean isRealized()
   {
      return realized;
   }

   /**
    * The method realizes this overlay window. If the window has been realized, this method is a
    * no-op.
    */
   public void realize()
   {
      if (realized)
      {
         return;
      }
      
      final TopLevelWindow<GuiOutputManager> owner = getOwner();
      if (owner == null)
      {
         throw new RuntimeException("Null owner of the OverlayWindow");
      }
      int ownerId = owner.getId().asInt();
      
      screen().realizeChildWindow(ownerId, id.asInt(), title(), true,
                                  isShareActivationWithOwner());
      
      realized = true;
      
      WindowManager.addWindow(this);
   }
   
   /**
    * Show window.
    */
   @Override
   public void show()
   {
      if (isVisible())
         return;
      
      TopLevelWindow<GuiOutputManager> owner = getOwner();
      
      gd.selectWindow(owner.getId().asInt());
      NativePoint ownerLocation = gd.getWindowLocation();
      gd.releaseWindow();
      
      realize();
      super.show();
      
      doLayout();
      
      Widget<GuiOutputManager> child = windowPane.get(0);
      
      // set up location and size
      NativePoint npLoc = child.physicalLocation();
      npLoc.x += ownerLocation.x;
      npLoc.y += ownerLocation.y;

      gd.selectWindow(id.asInt());
      gd.setWindowLocation(npLoc.x, npLoc.y);
      
      // now the child is 0,0 based
      child.setPhysicalLocation(0, 0);
      setSize(child.dimension());
      NativeDimension ndSize = child.physicalDimension();
      gd.resizeWindow(ndSize.width, ndSize.height);
      
      gd.setWindowVisible(true);
      gd.releaseWindow();

      WindowManager.activateWindow(this);
   }

   /**
    * Hide window.
    */
   @Override
   public void hide()
   {
      if (!isVisible())
         return;
      
      // do not activate the parent when we are not already an active window,
      // Window Manager may have activated another overlay window (an example use case
      // is another unrelated overlay window shown before this one is requested to be
      // closed/hidden, like opening menu items attached directly to a menu bar)
      if (WindowManager.isWindowActive(this) && getOwner() != null)
      {
         // activation state must be changed before gd.setWindowVisible(false)
         // otherwise Swing doesn't send focus-lost/focus-gained events!
         WindowManager.activateWindow(getOwner());
      }
      
      // hide us now
      gd.selectWindow(getId().asInt());
      gd.setWindowVisible(false);
      gd.releaseWindow();
      
      super.hide();
   }

   /**
    * Get the message area widget associated with this window.
    * 
    * @return   See above.
    */
   @Override
   public MessageAreaGuiImpl getGuiMessageArea()
   {
      return null;
   }

   /**
    * Iconify the window.
    */
   @Override
   public void iconify()
   {
      // overlay windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   @Override
   public void iconify(int sourceId)
   {
      // overlay windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Iconify the window.
    *
    * @param sourceId  The root ID of the sub-tree which was iconified, and triggered
    *                  iconification of
    *                  this window.
    * @param postEvent When <true>true</true> the method posts the WINDOW-MINIMIZED legacy event,
    */
   @Override
   public void iconify(int sourceId, boolean postEvent)
   {
      // overlay windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    */
   @Override
   public void restore()
   {
      // overlay windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   @Override
   public void restore(int sourceId)
   {
      // overlay windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    *
    * @param sourceId  The root ID of the sub-tree which was iconified, and triggered
    *                  iconification of
    *                  this window.
    * @param postEvent When <true>true</true> the method posts the WINDOW-RESTORED legacy event,
    */
   @Override
   public void restore(int sourceId, boolean postEvent)
   {
      // overlay windows in 4GL can't be iconified
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Maximize this window.
    */
   @Override
   public void maximize()
   {
      // overlay windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Maximize this window.
    *
    * @param postEvent When <true>true</true> the method posts the WINDOW-MAXIMIZED legacy event,
    *                  <code>false</code> means no event is posted.
    */
   @Override
   public void maximize(boolean postEvent)
   {
      // overlay windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Resize the widget to the specified dimensions.
    * 
    * @param    origin
    *           The top-left window corner new position.
    * @param    size
    *           The new window size.
    */
   @Override
   public void resizeTo(NativePoint origin, NativeDimension size)
   {
      // overlay windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Returns the output window's coordinates conversion service.
    * 
    * @return   a reference to the {@link CoordinatesConversion} instance. 
    */
   @Override
   public CoordinatesConversion coordinates()
   {
      return screen().coordinates();
   }
   
   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      super.setSize(width, height);
      windowPane.setSize(width, height);
   }

   /**
    * Determine if this window is being resized.
    * 
    * @return   This implementation always returns <code>false</code>.
    */
   @Override
   public boolean isResizing()
   {
      return false;
   }

   /**
    * Determine if this window workspace needs to be resized.
    * 
    * @return   This implementation always returns <code>false</code>.
    */
   public boolean isResizeWorkspace()
   {
      return false;
   }

   /**
    * Determine the resized window size.
    * 
    * @return   The new size.
    */
   @Override
   public NativeDimension getResizedSize()
   {
      // overlay windows in 4GL can't be resized
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Get the title bar widget associated witn this window.
    * 
    * @return   See above.
    */
   @Override
   public WindowTitleBar getTitleBar()
   {
      return null;
   }
   
   /**
    * Get the message area widget associated with this window.
    * 
    * @return   Always <code>null</code>, as overlay windows don't have a message area.
    */
   @Override
   public MessageArea<GuiOutputManager> getMessageArea()
   {
      return null;
   }

   /**
    * Get the status area widget associated with this window.
    * 
    * @return   See above.
    */
   @Override
   public StatusLine<GuiOutputManager> getStatusArea()
   {
      return null;
   }
   
   /**
    * Get the size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getWorkspaceSize()
   {
      return windowPane.dimension();
   }

   /**
    * Get the screen location of the window workspace area, in native units.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint getWorkspaceScreenLocation()
   {
      return windowPane.screenPhysicalLocation();
   }
   
   /**
    * Get the native size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWorkspaceNativeSize()
   {
      return windowPane.physicalDimension();
   }

   /**
    * Sets the native size of the window workspace area.
    * 
    * @param    width
    *           New workspace width.
    * @param    height
    *           New workspace height.
    */
   @Override
   public void setWorkspaceNativeSize(int width, int height)
   {
      // overlay-windows in 4GL can't be resized
      // this is a no-op
   }
   
   /**
    * Sets the window workspace location.
    *  
    * @param    x
    *           The x location. 
    * @param    y
    *           The y location.
    */
   @Override
   public void setWorkspaceLocation(int x, int y)
   {
      windowPane.setPhysicalLocation(x, y);
   }
   
   /**                                     
    * Get dimension of screen area available for the placement of frames.
    * 
    * @return  Instance of <code>Dimension</code> which describes size of the
    *          available screen area.
    */
   @Override
   public Dimension getScreenDimension()
   {
      return getContentPane().dimension();
   }
   
   /**
    * Find the widget to which the given mouse event can be applied.
    * 
    * @param    evt
    *           The mouse event.
    * 
    * @return   A widget over which the mouse is positioned or {@code null} if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(MouseEvt evt)
   {
      Widget<GuiOutputManager> widget = null;
      
      int evtID = evt.getEvent().getID();
      
      if (evtID == MouseEvent.MOUSE_DRAGGED  || 
          evtID == MouseEvent.MOUSE_RELEASED ||
          evtID == MouseEvent.MOUSE_CLICKED)
      {
         return mousePressedSource;
      }
      
      try
      {
         java.awt.Point p = evt.getEvent().getPoint();
         
         // search the other widgets
         widget = findMouseSource(new NativePoint(p.x, p.y));
         
         widget = (widget == null ? this : widget);
      }
      finally
      {
         if (widget != null && evtID == MouseEvent.MOUSE_PRESSED)
         {
            mousePressedSource = widget;
         }
      }
      
      return widget;
   }

   /**
    * Get a point to represent the LAST-EVENT:X and LAST-EVENT:Y coordinates.  For overlay windows,
    * these returns always <code>null</code>.
    * 
    * @param    point
    *           The point (relative to the window's top-left corner) for which the coordinates for
    *           the portable mouse events are needed.
    * @return   Always <code>null</code>.
    */
   @Override
   public NativePoint getPortableMouseEventCoordinates(NativePoint point)
   {
      return null;
   }
   
   /**
    * Destroy this window.
    */
   @Override
   public void destroy()
   {
      if (!realized)
      {
         return;
      }
      
      if (destroyingListener != null)
      {
         destroyingListener.run();
      }
      
      if (isVisible())
      {
         hide();
      }
      
      realized = false;
      
      WindowManager.removeNoDestroy(OverlayWindow.this);
      gd.deregisterWindow(getId().asInt());
      
      super.destroy();
   }
   
   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processEvent(Event event)
   {
      super.processEvent(event);
      
      if (event instanceof WindowActivated &&
          !((WindowActivated)event).isActivated())
      {
         close();
      }
   }
   
   /**
    * Get the maximum size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getMaxWorkspaceSize()
   {
      return getWorkspaceSize();
   }
   
   /**
    * Returns the window minimum size.
    * 
    * @return  See above.
    */
   @Override
   public NativeDimension getMinSize()
   {
      return new NativeDimension(0, 0);
   }   
   
   /**
    * This method indicates whether the window shares its activation state with its owner. This
    * method allows multiple windows to be considered active simultaneously. Example use cases are
    * the COMBO-BOX drop-down window or menu window. Note that while at the application level we
    * may recognize multiple active windows simultaneously, at the GUI driver level, there is
    * always one active window at a time. 
    */
   @Override
   public boolean isShareActivationWithOwner()
   {
      return true;
   }

   /**
    * Returns the window's border implementation. If <code>null</code>
    * is returned, the window has no border.
    * 
    * @return   A border implementation or <code>null</code>.
    */
   @Override
   public Border<GuiOutputManager> getBorder()
   {
      return null;
   }
   
   /**
    * Sets the destroying listener. The passed in runnable is called
    * when the window is being destroyed, i.e. at the beginning of the
    * destroy process.
    * 
    * @param    run
    *           The callback reference.
    */
   public void setDestroyingListener(Runnable run)
   {
      destroyingListener = run;
   }
}