/*
** Module   : RadioButtonGuiImpl.java
** Abstract : GUI specific implementation of the radio-button.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150923 Initial version.
** 002 EVL 20151009 Changing approach to get the native text width.  GuiDriver cal is no more
**                  safe to use.  Shifted to AbstractWidget.getText(Width|Height)Native.
** 003 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 004 CA  20160216 Cache the intermediate widget states.
** 005 CA  20160930 Fixed radio-button position in horizontal and expanded mode.
** 006 EVL 20161021 Small fix to make image Y coordinate more precisely for small button sizes.
** 007 EVL 20161026 Fix for painting button text in disabled state.  The difference can only be
**                  seen inside 3d mode frame.
** 008 CA  20161101 Added support for mnemonic drawing and processing.
** 009 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.util.*;

/**
 * GUI-specific portion of RadioButton.
 */
@SuppressWarnings("unchecked")
public class RadioButtonGuiImpl
extends RadioButton<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /**
    * Use this constructor when you want to set both the label and the value.
    * 
    * @param   config
    *          Reference to config.
    * @param   label
    *          The button's label which can clarify the button purpose.
    * @param   value
    *          The data which is the button's value.
    */
   RadioButtonGuiImpl(RadioSetConfig config, String label, BaseDataType value)
   {
      super(config, label, value);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // output manager
      cc = screen().coordinates();
      // font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
   }
   
   /**
    * Called by this RadioButton's parent container.
    */
   @Override
   public void draw()
   {
      final RadioSetConfig rscfg = config();
      gc = new GuiColorResolver(gd, rscfg, SysColor.COLOR_WINDOW, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(getId(), p, clip, () -> ThemeManager.getCurrentTheme().drawRadioButton(
            this, gd, gc, gf, 0, 0, d.width, d.height, focused, false));
   }
   
   /**
    * Get widget physical location, expressed in units native to UI type.
    *  
    * @return  widget location.
    */
   @Override
   public NativePoint physicalLocation()
   {
      // call the super class first
      NativePoint resLoc = super.physicalLocation();

      // check relative button position in radio-set
      Point pRel = location();
      resLoc.x += cc.columnToNative(pRel.x);
      resLoc.y += cc.rowToNative(pRel.y);

      return resLoc;
   }
   
   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      // call the super class first
      NativeDimension resDim = super.physicalDimension();
      
      // depending on the radio-set orientation we consider the same pixel based width|height
      // for all radio buttons(this is the largest button size) if there are the respective
      // values, this should simplify the radio-button painting
      if (!config.horizontal && config.widthPixels > 0)
      {
         resDim.width = config.widthPixels;
      }
      else if (config.horizontal && config.heightPixels > 0)
      {
         resDim.height = config.heightPixels;
      }
      
      return resDim;
   }

   /**
    * Gets the item height depending on chosen font. Two pixels are reserved for dotted selection. 
    * 
    * @return  The radio button height value in pixels.
    */
   public int getItemHeight()
   {
      return 4 + FontManager.getFontHeight(WindowManager.resolveWindow(this),
                                           gf.resolveFontNum(),
                                           gd);
   }
   
   /**
    * Return the width of the radio button in character units.
    * 
    * @return  The value of the button width including the label text and checkmark image.
    */
   @Override
   public double width()
   {
      double wRes = 0;
      RadioSetGuiImpl radioSet = (RadioSetGuiImpl) parent();
      
      if (width > 0.0)
      {
         wRes = width;
      }
      else
      {
         int widthNative = widthNative();
         widthNative = widthNative + radioSet.getRightPadding();
         
         // choose max value from two ones
         wRes = cc.widthFromNative(widthNative);
      }
      
      return wRes;
   }
   
   /**
    * Compute the native width of this radio-button (without any extra right-padding).
    * 
    * @return   The widget's native width, without padding on the right.
    */
   int widthNative()
   {
      // the width must be calculated based on the label size in pixels
      int textWidthNative = getTextWidthNative(getText(), gf.resolveFontNum());
      // add extra space for image and dotted rectangle
      return textWidthNative + ThemeManager.getCurrentTheme().getRadioButtonImageSpace() + 1;
   }
   
   /**
    * Return the height of the radio button in character units.
    * 
    * @return  The value of the button height.
    */
   @Override
   public double height()
   {
      double hRes = 0.0;
      
      if (config.heightChars > 0.0)
      {
         hRes = config.heightChars / (config.horizontal ? 1 : config.items.length);
      }
      else
      {
         hRes = cc.heightFromNative(getItemHeight());
      }
      
      return hRes;
   }

   /**
    * Draw cursor at the correct position.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Widget#drawCaret()
    */
   @Override
   public void drawCaret()
   {
      // nothing to do in GUI
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      return processMnemonicImpl(key, this);
   }
   
   /**
    * Creates new {@link MnemonicInfoGuiImpl} instance.
    *
    * @return   {@link MnemonicInfoGuiImpl} instance.
    */
   @Override
   public MnemonicInfo createMnemonic()
   {
      return new MnemonicInfoGuiImpl(false);
   }
}