/*
** Module   : RectangleGuiImpl.java
** Abstract : GUI rectangle implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20141114 Initial version.
** 002 EVL 20141120 Fixed origin coords inside painting worker.  Changed double line shaded
**                  rectangle drawing approach.  Also when GROUP-BOX attribute is true the filling
**                  become ignored.
** 003 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.
** 004 EVL 20150409 Turning on clipping tool.
** 005 HC  20150323 Removed business logic from config classes.
** 006 GES 20150425 Eliminated AWT dependencies.
** 007 IAS 20150926 MOVE-TO-TOP()/MOVE-TO_BUTTON() support.
** 008 EVL 20151030 Fix for 1 pixel size issue and incorrect filled rounded rectangle draw.  Fix
**                  for wrong mode with TREE-D defined owning frame.  Added border size capability
**                  for 3D shaded rectangle.
** 009 CA  20160216 Cache the intermediate widget states.
** 010 EVL 20160420 Fix for 3D shaded shape and size, not a simple two rectangles.
** 011 CA  20170130 Fixed border drawing when EDGE-PIXELS > 1.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific GUI implementation of RECTANGLE widget.
 */
public class RectangleGuiImpl
extends Rect<GuiOutputManager>
{
   /** Screen driver. */
   private GuiDriver gd;

   /** Output manager. */
   private GuiOutputManager screen;
   
   /** Coordinate converter. */
   private CoordinatesConversion cc;

   /** Color resolver. */
   private GuiColorResolver gc;
   
   /** Rounding diameter in pixels. */
   private static final int ROUND_DIAMETER = 8;
   
   /** Light shadow box color. */
   private ColorRgb lightShadow;

   /** Dark shadow box color. */
   private ColorRgb darkShadow;
   
   /**
    * This constructor creates a rectangle.
    * 
    * @param    id
    *           The widget's ID.
    * @param    owner
    *           The parent container, frame which will include this
    *           rectangle.
    */
   RectangleGuiImpl(WidgetId id, Container<GuiOutputManager> owner) 
   {
      super(id, owner);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public RectangleGuiImpl(WidgetId id, RectangleConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, RectangleConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // initialize the shadow box colors
      lightShadow = gd.getSysColor(SysColor.COLOR_WINDOW);
      darkShadow  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
   }
   
   /**
    * Draw the rectangle. Called by this rectangle's parent container.
    */
   @Override
   public void draw()
   {
      // rectangle options need to be refreshed before every painting
      final RectangleConfig config = (RectangleConfig) config();
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      
      // determine colors to use
      gc.colors();
   
      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));

      // check the current UI theme in use
      final boolean isXpStyle = gd.getUiTheme().equals("xp");

      // border size, GROUP-BOX is involved, consider Windows XP style or classic
      //    GROUP-BOX == true and Windows XP style -> single line rounded
      //    GROUP-BOX == true and Windows Classic style -> double line "shaded"
      final int borderSize = (config.groupBox && isXpStyle) ? 1 :
                             (config.groupBox && !isXpStyle) ? 2 : (int) config.edgePixels;
                             
      // is rectangle rounded or not
      //    GROUP-BOX == true means ROUNDED value is ignored
      final boolean isRounded = (config.groupBox && isXpStyle) ||
                                (!config.groupBox && config.rounded);
      
      // if owning frame has TREE-D defined and style is classic and EDGE-PIXELS > 1
      // we need to simulate GROUP-BOX mode even if GROUP-BOX is not defined
      Frame<GuiOutputManager> frOwn = (Frame<GuiOutputManager>) UiUtils.locateFrame(this);
      final boolean altGroupBox = config.edgePixels > 1 && !isXpStyle &&
                                  frOwn != null && frOwn.config().threeD && !isRounded;
      
      // clippable drawing start 
      gd.draw(getId(), p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);

            // fill rectangle, GROUP-BOX disables filling
            // should be before border otherwise rectangle will be wrong
            // because fill and border have intersection areas
            if (config.filled && borderSize >= 0 && !config.groupBox)
            {
               fillRectangle(po, d, borderSize, gc.bgColor, isRounded);
            }
            
            // draw rectangle border
            if (borderSize > 0)
            {
               ColorRgb borderColor = gc.fgColor;
               boolean doubleShaded = (config.groupBox && !isXpStyle) || altGroupBox;
               
               if (doubleShaded)
               {
                  borderColor = gd.getSysColor(SysColor.COLOR_3DFACE);
               }
               
               drawRectangleBorder(po, d, borderSize, borderColor, isRounded);

               if (doubleShaded)
               {
                  drawShadedDoubleLineBorder(po, d, lightShadow, darkShadow);
               }
            }
         }
         
      });
   }

   /**
    * Get the Z-Order class
    * 
    * @return Z-Order class of the widget
    */
   @Override
   public ZOrderClass getZOrderClass()
   {
      return ZOrderClass.DECORATION;
   }

   /**
    * Draw rectangle border.
    * 
    * @param    orig
    *           Rectangle origin.
    * @param    dim
    *           Rectangle size.
    * @param    borderSize
    *           Border size in pixels.
    * @param    color
    *           Border color.
    * @param    isRounded
    *           Flag indicating if the border is rounded or not.
    */
   private void drawRectangleBorder(NativePoint     orig, 
                                    NativeDimension dim, 
                                    int             borderSize,
                                    ColorRgb        color,
                                    boolean         isRounded)
   {
      gd.setColor(color);
      // draw rectangle borders from outmost to innermost
      if (isRounded)
      {
         for (int i = 0; i < borderSize; i++)
         {
            gd.drawRoundRect(orig.x + i, orig.y + i, dim.width - 2*i - 1,
                             dim.height - 2*i - 1, ROUND_DIAMETER);
         }
      }
      else
      {
         for (int j = 0; j < borderSize; j++)
         {
            gd.drawRect(orig.x + j, orig.y + j, dim.width - 2*j - 1, dim.height - 2*j - 1);
         }
      }
   }

   /**
    * Draw shaded double line rectangle border.  This means we simulate the embossing by combining
    * dark and light lines with colors provided.
    * 
    * @param    orig
    *           Rectangle origin.
    * @param    dim
    *           Rectangle size.
    * @param    colorLight
    *           The light component of the shade.
    * @param    colorDark
    *           The dark component of the shade.
    */
   private void drawShadedDoubleLineBorder(NativePoint     orig,
                                           NativeDimension dim,
                                           ColorRgb        colorLight,
                                           ColorRgb        colorDark)
   {
      // alternative group box also can have the border size
      int border = config.edgePixels;
      // upper left corner
      int x = orig.x;
      int y = orig.y;
      // width and height
      int w = dim.width - 1;
      int h = dim.height - 1;

      // if border size > 2 the painting is different
      if (border > 2)
      {
         // starting with dark part
         gd.setColor(colorDark);
         // outer
         gd.drawLine(x, y, x, y + h);
         gd.drawLine(x + 1, y, x + w - 1, y);
         // inner
         gd.drawLine(x + w - border + 1, y + border - 1, x + w - border + 1, y + h - border + 1);
         gd.drawLine(x + border, y + h - border + 1, x + w - border, y + h - border + 1);

         // now light part
         gd.setColor(colorLight);
         // outer
         gd.drawLine(x + 1, y + h, x + w, y + h);
         gd.drawLine(x + w, y + h - 1, x + w, y);
         // inner
         gd.drawLine(x + border - 1, y + border - 1, x + border - 1, y + h - border + 2);
         gd.drawLine(x + border, y + border - 1, x + w - border, y + border - 1);
      }
      else
      {
         // starting with dark part
         gd.setColor(colorDark);
         gd.drawRect(x, y, w - 1, h - 1);

         // now light part
         gd.setColor(colorLight);
         gd.drawRect(x + 1, y + 1, w - 1, h - 1);
         gd.drawLine(x + w, y, x + w, y);

         // and dark part again to two remainingdots
         gd.setColor(colorDark);
         gd.drawLine(x + w - 1, y + 1, x + w - 1, y + 1);
         gd.drawLine(x, y + h, x, y + h);
      }
   }
   
   /**
    * Fill rectangle.
    * 
    * @param    orig
    *           Rectangle origin.
    * @param    dim
    *           Rectangle size.
    * @param    borderSize
    *           Border size characters.
    * @param    color
    *           Fill color.
    * @param    isRounded
    *           Flag indicating if the border is rounded or not.
    */
   private void fillRectangle(NativePoint     orig, 
                              NativeDimension dim, 
                              int             borderSize,
                              ColorRgb        color,
                              boolean         isRounded)
   {
      gd.setColor(color);
      // have to include border to filling area, otherwise rounded rectangles are drawing
      // not properly, so the filling must be done before border painting
      int rx = orig.x;
      int ry = orig.y;
      int rw = dim.width - 1;
      int rh = dim.height - 1;
      
      if (isRounded)
      {
         gd.fillRoundRect(rx, ry, rw, rh, ROUND_DIAMETER);
      }
      else
      {
         gd.fillRect(rx, ry, rw, rh);
      }
   }
}