/*
** Module   : ScrollBarGuiButton.java
** Abstract : scroll bar button.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20140917 Created initial version.
** 002 MAG 20141020 Added cliping rectangle.
** 003 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 004 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
** 005 EVL 20150611 Adding handling of the disabled sate of the ScrollBarGuiImpl.
** 006 EVL 20150903 Adding SVG based images for painting the button body.
** 007 HC  20150831 Scrolling support overhaul.
** 008 CA  20150910 Added mouse actions.
** 009 CA  20150925 physicalDimension() must produce the same values as width()/height() - it 
**                  should be overridden only in special circumstances; added width()/height() to
**                  properly report the widget's dimension. 
** 010 EVL 20151010 Fix for thumb drawing issues. Also reworked image loader specifically for
**                  small images.  Optimization for button painting on mouse press, release.  In
**                  these cases we do not need to repaint whole window.
** 011 EVL 20151124 The button pressed state needs repaint() call to properly display pressed mode
**                  when running indide drop-down life cycle.  Improved mouse handling, in 4GL
**                  if to press mouse, move inside button and release - this generates click.  So
**                  we have to track the mouse release, not click.  Adding auto-scroll feature on
**                  press and hold mouse button.  The mouse executor shoule be stopped on widget
**                  destroy to free the resources.  Changing static executor to be instance based.
** 012 CA  20160328 Do not raise legacy mouse events for scroll bar buttons.
** 013 SBI 20151223 Emits only valid auto scroll events.
**     CA  20160216 Cache the intermediate widget states.
** 014 VIG 20160127 Added null check to mousePressed (HC change). Added isRMBEntrySupported() impl, 
**                  added RMB check to mousePressed(), added scroll step type to onButtonClick().
** 015 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;
import java.util.concurrent.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;

/**
 * A scroll bar button.
 */
@SuppressWarnings("unchecked")
class ScrollBarGuiButton
extends GuiButton
{   
   /** Scroll bar container */
   ScrollBar<GuiOutputManager> owner;
   
   /** The graphics driver. */
   private GuiDriver gd;
   
   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Button position */
   private ScrollBar.Position position;
   
   /** Executor for handling mouse input. */
   private ExecutorService mouseInputExecutor = Executors.newSingleThreadExecutor();
   
   /**
    * This constructor creates a scroll bar button.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container.
    * @param   position
    *          Button position within scroll bar.
    */
   ScrollBarGuiButton(WidgetId id, ScrollBar<GuiOutputManager> owner, ScrollBar.Position position)
   {
      super(id, owner, "");
      
      this.owner = owner;
      this.position = position;
      
      _setVisible(true);
      gd = (GuiDriver) OutputManager.getDriver();
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_3DFACE, SysColor.COLOR_3DDKSHADOW);
   }

   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public double width()
   {
      return config.widthChars;
   }

   /**
    * Calculate the height of this widget.
    *
    * @return    The height in rows.
    */
   @Override
   public double height()
   {
      return config.heightChars;
   }
   
   /**
    * Draw this scroll bar button, depending on its type.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible
      if (!isVisible())
      {
         return;
      }

      // resolve colors
      gc.colors();
      
      // get button coordinates
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      // clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      gd.draw(getId(), p, clip, () ->
            ThemeManager.getCurrentTheme().drawScrollBarButton(gd, gc, 
                  0, 0, d.width, d.height, 
                  position, owner.orientation() == Orientation.HORIZONTAL, 
                  owner.isEnabled(), isPressed(), false));
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      super.mousePressed(e); // update isPressed status
      
      // check if the scrollbar widget is enabled
      if (owner.isEnabled() && e.getButton() != getMenuMouse())
      {
         // need to use repaint() here instead of refresh() because for example inside combo-box
         // drop-down standalone event loop the screen synchronization is not happening
         repaint();
         
         // for holding mouse we need to generate scroll events
         mouseInputExecutor.execute(() ->
         {
            ThinClient tc = ThinClient.getInstance(); 
            // finish on mouse release
            while (!isPressed())
            {
               try
               {
                  // wait to see if the mouse is still pressed
                  Thread.sleep(GuiConstants.REPEAT_MILLIS);
                  
                  if (isPressed())
                  {
                     // emit the sequential scroll events
                     ScrollEvent evt =
                        ((ScrollBarGuiImpl)owner).createAutoScrollEvent(position);
                     if (evt != null)
                     {
                        tc.postOSEvent(evt);
                     }
                  }
               }
               catch (InterruptedException ex)
               {
                  // nop
               }
            }
         });
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      super.mouseReleased(e);
      
      // check if the scrollbar widget is enabled
      if (owner.isEnabled())
      {
         repaint();
         // call scroll bar
         owner.onButtonClick(position, ScrollStep.SHORT, new MouseEvt(e, ancestor()));
      }
   }
   
   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }
   
   /**
    * Check if this widget gains ENTRY event when right/middle mouse button is clicked.
    *
    * @return  <code>true</code> if this widget gains ENTRY event
    *          when right/middle mouse button is clicked.
    */
   @Override
   public boolean isRMBEntrySupported()
   {
      return false;
   }
   
   /**
    * Destroy this widget.  Cleans up all including widgets before destruction.
    */
   @Override
   public void destroy()
   {
      // stop the mouse related thread if it is active
      if (!mouseInputExecutor.isShutdown())
      {
         mouseInputExecutor.shutdownNow();
      }

      // call superclass for main destruction
      super.destroy();
   }
   
   /**
    * Refresh the scrollbar button not repainting full window.
    */
   @Override
   protected void refresh()
   {
      // this needs to start drawing relative to parent container, the scroll bar in our case
      final NativePoint origin = owner.screenPhysicalLocation();
      gd.selectWindow(topLevelWindow().getId().asInt());

      gd.drawFromOrigin(origin, null, this::refreshImpl);
      
      gd.releaseWindow();
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED
      };
   }
   
   /**
    * Refresh the scrollbar button not repainting full window.  Internal version.
    */
   protected void refreshImpl()
   {
      super.refresh();
   }
   
   /**
    * This widget can NOT receive focus. The focus always stays with owner.  
    *
    * @return  always {@code false}.
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }
}