/*
** Module   : ScrollPaneGuiImpl.java
** Abstract : scroll panel GUI implementation. 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141016 Implements scroll bars management.
** 003 CA  20141123 The scrollbar visibility is off if they are AS_NEEDED.
** 004 CA  20141205 Do not use the frame's saved dimension, use the real one.
** 005 EVL 20150428 Adding method to set box flag.  Adding mouse handler for combo-box drop-down
**                  child implementation.
** 006 HC  20150508 Improvements to GUI window scrolling support.
** 007 CA  20150509 Added support for always hiding the scroll bars.  Some other misc fixes and
**                  improvements in preparation for GUI EDITOR widget.
** 008 EVL 20150511 Adding method to get the box flag.  Adding flag indicating srolling by mouse.
** 009 CA  20150617 Adjustments to allow editor's pseudo-scrolling needs (where scroll thumbs are 
**                  positioned manually).
** 010 EVL 20150630 Routing mouse move event to the drop-down scrollable list if exists.
** 011 CA  20150723 Implemented drawing of top-frame labels (non-body drawing) in down frames.
** 012 CA  20150728 When drawing the body of a down frame with title, the content has an Y-offset
**                  which needs to be accounted for.  This is done via explicitly setting the 
**                  contentBounds fields, which will require all widgets to be part, otherwise 
**                  they will not be drawn.
**                  When drawing the viewport, translate to its coordinates first.
** 013 CA  20150731 adjustViewportSize and adjustScrollBars need to be executed in a eventBracket,
**                  to discard any repaint events.
** 014 HC  20150831 Scrolling support overhaul.
** 015 CA  20150910 Added mouse actions. Fixed scroll bars eventing and viewport sizing.
**                  Improved scroll layout calculation for the case of native scrolling.
** 016 CA  20150925 Scrollbars must be layout here, but drawn by their parent container.
** 017 SVL 20151013 Added VERTICAL_ALWAYS option and getScrollBar function.
** 018 EVL 20151013 Adding method to get the native scrollbar size.  Fix for visible but disabled
**                  state.  Fix for intersection coordinates calculation.
**     SVL 20151019 Fixed getVisibleDimension. Added manual control for scrollbars.
**     SVL 20151023 Do not automatically update enabled state for manual control mode.
** 019 SVL 20151101 Added support for scroll bar controllers.
** 020 HC  20160203 Improved runtime support of FRAME:VIRTUAL* and FRAME:SCROLLABLE attributes.
**                  Frame layout improvements.
** 021 IAS 20160217 LOAD-MOUSE-POINTER support
** 022 HC  20160313 Fixed frame scroll layout calculation.
** 023 CA  20160323 Added adjustScrollLayout, which needs to be called by i.e. EDITOR each time a
**                  scroll bar is removed/added (before the widget has been realized).
**     SBI 20160328 Changed to provide a color resolver instance to its subclasses.
** 024 EVL 20160419 Sometimes we need more effects to implement for 3d mode.
** 025 SVL 20160412 Fixed calcInitialViewportSize.
** 026 EVL 20160530 Fix for scroll-bars intersection area starting coordinates.
** 027 EVL 20160912 Fix for intersection area fill area coordinates.
** 028 CA  20161020 Fixed a horiz scrollbar 4GL bug, which prevents showing the horiz scrollbar 
**                  for the parent frame if the child frame uses a ENABLE or DISPLAY statement 
**                  and the parent frame requires a horiz scrollbar.
** 029 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;
import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A GUI scroll pane implementation. 
 */
@SuppressWarnings("unchecked")
public class ScrollPaneGuiImpl
extends ScrollPane<GuiOutputManager>
{
   /**
    * Container show bars options:
    * 
    * AS_NEEDED       - Only when needed.
    * ALWAYS          - Always
    * VERTICAL_ALWAYS - Vertical scrollbar is always displayed
    */
   public enum ShowBars
   {
      AS_NEEDED,
      ALWAYS,
      NEVER,
      VERTICAL_ALWAYS
   }
   
   /** Screen driver */
   private GuiDriver gd;
   
   /** Color resolver */
   protected GuiColorResolver gc;

   /** Horizontal scroll bar */
   private ScrollBarGuiImpl hBar;

   /** Vertical scroll bar */
   private ScrollBarGuiImpl vBar;

   /** Scroll bars show policy */
   private ShowBars showBars = ShowBars.AS_NEEDED;
   
   /** True when have both horizontal and vertical scroll bars */
   boolean both;
   
   /** child widget */
   private final Widget widget;
   
   /** top level window supplier*/
   private final Optional<Supplier<TopLevelWindow<GuiOutputManager>>> wnd;
   
   /**
    * Flag which activates a 4GL bug, which will not show the horizontal scrollbar, even if
    * forced.
    */
   private boolean horizBarQuirk = false;
   
   /**
    * Constructor.
    * 
    * @param   widget
    *          Child widget.
    */
   public ScrollPaneGuiImpl(ScrollableWidget<GuiOutputManager> widget)
   {
      this(widget, null);
   }

   /**
    * Constructor.
    * 
    * @param widget
    *           Child widget.
    * @param wnd
    *           TopLevelWindow supplier
    */
   public ScrollPaneGuiImpl(ScrollableWidget<GuiOutputManager> widget,
         Supplier<TopLevelWindow<GuiOutputManager>> wnd)
   {
      super(widget);
      this.widget = widget;
      this.wnd = Optional.ofNullable(wnd);

      // GUI driver
      gd = (GuiDriver) OutputManager.getDriver();
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_3DFACE, SysColor.COLOR_WINDOWTEXT);

      BorderedPanelGuiImpl panel = (BorderedPanelGuiImpl) super.rootPanel;
      panel.setBorder(new EmptyBorder<>());
      panel.setColorResolver(gc);

      _setVisible(true);
   }

   /**
    * Adjust the scroll layout for this pane, as scrollbars may have been added/removed.  All
    * repaint events will be discarded.
    */
   void adjustScrollLayout()
   {
      // do this in an event bracket to discard any repaint events - we are already drawing here
      // resize viewport and adjust scroll bars
      ThinClient.getInstance().eventBracket(true, this::adjustScrollLayoutImpl);
   }
   
   /**
    * Widget drawing method.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible 
      if (!isVisible())
      {
         return;
      }

      // determine frame colors
      gc.colors();
      
      adjustScrollLayout();
      
      // get container coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      final NativeRectangle clip = clipRectangle(p, d);

      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // layout scroll bars
            layoutScrollBars();

            // draw from root
            rootPanel.draw();

            drawIntersection();
         }

         /**
          * Layout container scroll bars.
          */
         private void layoutScrollBars()
         {
            if ((hBar != null) && hBar.isVisible())
            {
               hBar.wOff = 0;
            }
            if ((vBar != null) && vBar.isVisible())
            {
               vBar.hOff = 0;
            }

            if (both)
            {
               // if at least one scroll bar is visible both will be visible
               // used by frames for example
               if (showBars == ShowBars.ALWAYS)
               {
                  boolean visible = hBar.isVisible() || vBar.isVisible();
                  hBar._setVisible(visible);
                  vBar._setVisible(visible);
               }

               // check if both scroll bars are visible
               boolean bothVisible = hBar.isVisible() && vBar.isVisible();
               if (bothVisible)
               {
                  hBar.wOff = vBar.width();
                  vBar.hOff = hBar.height();
                  if (((BorderedPanelGuiImpl) rootPanel).is3DBorder())
                  {
                     int borderOffset = ThemeManager.getCurrentTheme().get3DBorderOffset();
                     hBar.wOff += screen().coordinates().widthFromNative(borderOffset);
                     vBar.hOff += screen().coordinates().heightFromNative(borderOffset);
                  }
               }
            }
         }

         /**
          * Draw bottom right intersection (if there are two scroll bars).
          */
         private void drawIntersection()
         {
            if (both)
            {
               // check if both scroll bars are visible
               boolean bothVisible = hBar.isVisible() && vBar.isVisible();
               if (bothVisible)
               {
                  NativeDimension hb = hBar.physicalDimension();
                  NativeDimension vb = vBar.physicalDimension();
                  // extra pixels to subtract (root panel borer size)
                  NativeInsets ni = screen().coordinates().insetsToNative(rootPanel.insets()); 
                  // fill scroll bar bottom right intersection
                  // taking into account possible non-zero border width of the root panel
                  int x = d.width - vb.width - ni.right;
                  int y = d.height - hb.height - ni.bottom;
                  gd.setColor(gd.getSysColor(SysColor.COLOR_SCROLLBAR));
                  gd.fillRect(x, y, vb.width, hb.height);
               }
            }
         }
      });
   }

   /**
    * Get the vertical or horizontal scroll bar controlled by this pane.
    *
    * @param  orientation
    *         Which scroll bar to get.
    *
    * @return requested scroll bar.
    */
   public ScrollBarGuiImpl getScrollBar(Orientation orientation)
   {
      if (orientation == Orientation.HORIZONTAL)
      {
         return hBar;
      }
      else
      {
         return vBar;
      }
   }

   /**
    * Set custom controller for one of the scroll bars.
    *
    * @param orientation
    *        Orientation of the target scroll bar.
    * @param controller
    *        Custom controller for the scroll bar.
    */
   public void setScrollBarController(ScrollBar.Orientation orientation,
                                      ScrollBarControllerGuiImpl controller)
   {
      if (orientation == Orientation.HORIZONTAL)
      {
         if (hBar != null)
         {
            hBar.clearScrollListeners();
            if (controller == null)
            {
               hBar.addScrollListener(this);
            }
            else
            {
               controller.setScrollBar(hBar);
            }
         }
      }
      else
      {
         if (vBar != null)
         {
            vBar.clearScrollListeners();
            if (controller == null)
            {
               vBar.addScrollListener(this);
            }
            else
            {
               controller.setScrollBar(vBar);
            }
         }
      }
   }

   /**
    * The secondary renderer for this container. Draws the child container. Note that this routine
    * assumes that it is invoked from Frame class.
    * 
    * @param   list
    *          Array of components to draw.
    */
   @Override
   public void draw(final Widget<GuiOutputManager>[] list)
   {
      // get container coordinates
      NativePoint p = viewport().physicalLocation();
      p = p.translate(physicalLocation());
      NativeDimension d = viewport().physicalDimension();
            
      NativeRectangle clip = clipRectangle(p, d);

      // determine frame colors
      gc.colors();

      gd.draw(p, clip, () ->
      {
         Rectangle paneRect = viewport().bounds();
         
         for (int i = 0; i < list.length; i++)
         {
            if (!list[i].isVisible())
               continue;

            // check if component fits the visible area
            if (paneRect.contains(list[i].bounds()))
               list[i].draw();
         }
      });
   }
   
   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      super.setEnabled(enabled);
      
      viewport().getScrollWidget().setEnabled(enabled);
      
      if (vBar != null)
      {
         vBar.setEnabled(enabled);
      }
      
      if (hBar != null)
      {
         hBar.setEnabled(enabled);
      }
   }

   /**
    * Set container scroll bars.
    * 
    * @param   bars
    *          Scroll bars.
    */
   public void setScrollBars(ScrollBar.Bars bars)
   {
      createScrollBars(bars);      
   }
   
   /**
    * Set show bars policy.
    * 
    * @param   showBars
    *          Show bars policy.
    */
   public void setShowBars(ShowBars showBars)
   {
      this.showBars = showBars;
   }
   
   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      if (viewport().getScrollWidget() instanceof ScrollableSelectionListGuiImpl)
      {
         // consume original event
         e.consume();
      }
      else
      {
         super.mousePressed(e);
      }
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (viewport().getScrollWidget() instanceof ScrollableSelectionListGuiImpl)
      {
         // route click to the drop-down internal
         viewport().getScrollWidget().mouseClicked(e);
         // consume original event
         e.consume();
      }
      else
      {
         super.mouseClicked(e);
      }
   }

   /**
    * Notification of a mouse moved event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      if (viewport().getScrollWidget() instanceof ScrollableSelectionListGuiImpl)
      {
         // route move to the drop-down internal
         viewport().getScrollWidget().mouseMoved(e);
         // consume original event
         e.consume();
      }
      else
      {
         super.mouseMoved(e);
      }
   }
   
   /**
    * Get the current effective custom mouse pointer
    * 
    * @return the current effective custom mouse pointer  
    *        
    */
   @Override
   public MousePtrWrapper getEffectiveMousePointer()
   {
      return widget.getEffectiveMousePointer();
   }

   /**
    * Get the Loaded custom mouse pointer if it is set
    * 
    * @return the loaded custom mouse pointer if it is set  
    *        
    */
   @Override
   public Optional<MousePtrWrapper> getLoadedMousePointer()
   {
      return widget.getLoadedMousePointer(); 
   }


   /**
    * Check if the default mouse pointer should be used regardless of
    * the parent settings
    * 
    * @return
    *        <code>true</code> if the default mouse pointer should always be used
    */
   @Override
   public boolean enforceDefaultMousePtr()
   {
      return widget.enforceDefaultMousePtr();
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return widget.supportsCustomMousePointer();
   }

   /**
    * Get ancestor TopLevelWindow for this widget. This method must be called
    * only for widgets that are already attached to a TopLevelWindow.
    * 
    * @return ancestor TopLevelWindow.
    * 
    * @throws    IllegalStateException
    *            When this widget is not attached to a top-level window.
    */
   @Override
   public TopLevelWindow<GuiOutputManager> topLevelWindow()
   {
      return wnd.isPresent() ? wnd.get().get() : super.topLevelWindow();  
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      Widget<?> child = viewport().getScrollWidget();
      
      if (child instanceof ScrollableSelectionListGuiImpl)
      {
         return new int[]
         {
            MouseEvent.MOUSE_CLICKED,
            MouseEvent.MOUSE_PRESSED,
            MouseEvent.MOUSE_RELEASED,
            MouseEvent.MOUSE_MOVED,
            MouseEvent.MOUSE_ENTERED,
            MouseEvent.MOUSE_EXITED
         };
      }
      else
      {
         return new int[]
         {
            MouseEvent.MOUSE_CLICKED,
            MouseEvent.MOUSE_PRESSED,
            MouseEvent.MOUSE_RELEASED
         };
      }
   }
   
   /**
    * This method is invoked when scroll pane is scrolled.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void handleScrollEvent(ScrollEvent event)
   {
      super.handleScrollEvent(event);
      
      // notify scroll bars
      if (event.newX() != null && hBar != null)
      {
         hBar.handleScrollEvent(event);
      }
      
      if (event.newY() != null && vBar != null)
      {
         vBar.handleScrollEvent(event);
      }
   }
   
   /**
    * Calculates and returns initial viewport size. The method doesn't
    * consider scroll bars and always returns the full viewport size.
    
    * @return  Initial viewport dimension.
    */
   @Override
   public Dimension calcInitialViewportSize()
   {
      Point origin = viewport().location();
      
      // calculate an initial viewport size, not considering scroll bars
      double w = Coordinate.scale(width()  - origin.x - rootPanel.insets().widthDelta());
      double h = Coordinate.scale(height() - origin.y - rootPanel.insets().heightDelta());
      return new Dimension(w, h);
   }
   
   /**
    * Get scroll bar size in pixel units.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    *          
    * @return  see above
    */
   public int barSizeNative(Orientation orientation)
   {
      int sizeInPixels = 0;
      
      switch (orientation)
      {
         case HORIZONTAL:
            if (hBar != null && hBar.isVisible())
            {
               sizeInPixels = hBar.physicalDimension().height;
            }
            break;
         case VERTICAL:
            if (vBar != null && vBar.isVisible())
            {
               sizeInPixels = vBar.physicalDimension().width;
            }
            break;
         default:
            break;
            
      }
      
      return sizeInPixels;
   }
         
   /**
    * Adjusts viewport size, scroll bar visibility and positions.
    */
   protected void adjustScrollLayoutImpl()
   {
      ScrollableWidget<GuiOutputManager> scroll = viewport().getScrollWidget();
      
      Point origin = scroll.location();
      
      // scroll area dimension
      NativeDimension scrollDim = scroll.getScrollDimension();
      
      // view area dimension of the scroll area
      NativeDimension visDim = scroll.getVisibleDimension();
      
      int maxH;
      int maxV;
      
      boolean visibleH = false;
      boolean visibleV = false;
      
      // calculate an initial viewport size, not considering scroll bars
      Dimension viewportDim = calcInitialViewportSize();
      
      maxH = scrollDim.width - visDim.width;
      maxV = scrollDim.height - visDim.height;

      if (horizBarQuirk)
      {
         if (hBar != null)
         {
            // if there was previously set a horizontal bar, discard it
            screen().getRegistry().removeWidgetsRecursive(hBar);
            hBar.destroy();
            hBar = null;
            
            both = false;
         }
      }
      
      // do not show scroll bars when viewport dimension is zero
      visibleH = !horizBarQuirk && 
                 ((visDim.width > 0 && maxH > 0 && showBars == ShowBars.AS_NEEDED) || 
                  showBars == ShowBars.ALWAYS);
      visibleV = (visDim.height > 0 && maxV > 0 && showBars == ShowBars.AS_NEEDED) ||
                 showBars == ShowBars.ALWAYS || showBars == ShowBars.VERTICAL_ALWAYS;

      if (visibleH)
      {
         double barSize = barSize(Orientation.HORIZONTAL);
         
         if (showBars == ShowBars.AS_NEEDED)
         {
            visibleV = true;
            maxV += screen().coordinates().heightToNative(barSize);
         }
         
         // for 3d border the scrollbars should be inside the widget
         if (!((BorderedPanelGuiImpl) rootPanel).is3DBorder())
         {
            viewportDim.height -= barSize;
         }
      }

      if (visibleV)
      {
         double barSize = barSize(Orientation.VERTICAL);
         // for 3d border the scrollbars should be inside the widget
         if (!((BorderedPanelGuiImpl) rootPanel).is3DBorder())
         {
            viewportDim.width -= barSize;
         }
      }
      
      viewport().setSize(viewportDim);
      
      Integer newXPosition = null;
      Integer newYPosition = null;
      
      if (hBar != null)
      {
         newXPosition = hBar.getPosition();
         if (hBar.isControlled())
         {
            hBar.adjustControlledScrollbar();
         }
         else
         {
            adjustScrollBar(Orientation.HORIZONTAL, maxH, visibleH);
         }
      }
      
      if (vBar != null)
      {
         newYPosition = vBar.getPosition();
         if (vBar.isControlled())
         {
            vBar.adjustControlledScrollbar();
         }
         else
         {
            adjustScrollBar(Orientation.VERTICAL, maxV, visibleV);
         }
      }
      
      if (hBar != null && newXPosition != hBar.getPosition() && !hBar.isControlled())
      {
         newXPosition = hBar.getPosition();
      }
      else
      {
         // if no scroll position change, clear the value
         newXPosition = null;
      }

      if (vBar != null && newYPosition != vBar.getPosition() && !vBar.isControlled())
      {
         newYPosition = vBar.getPosition();
      }
      else
      {
         // if no scroll position change, clear the value
         newYPosition = null;
      }
      
      // only notify when at least one direction has changed scroll position
      if (newXPosition != null || newYPosition != null)
      {
         viewport().getScrollWidget().scroll(newXPosition, newYPosition);
      }
   }
   
   /**
    * Flag which activates a 4GL bug, which will not show the horizontal scrollbar, even if
    * forced.
    *
    * @param    quirk
    *           The bug's state.
    */
   void setHorizBarQuirk(boolean quirk)
   {
      this.horizBarQuirk = quirk;
   }

   /**
    * Adjust scroll bar max value.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    * @param   max
    *          Max value.
    */
   private void adjustScrollBar(Orientation orientation, int max, boolean visible)
   {
      ScrollBarGuiImpl bar = (orientation == Orientation.HORIZONTAL) ? hBar : vBar;

      if (bar != null)
      {
         bar._setVisible(visible);

         if (max < 0)
         {
            max = 0;
         }

         if (visible)
         {
            bar.setMax(max);
         }
         else
         {
            bar.setMax(0);
         }

         if (bar.getPosition() > max)
         {
            bar.setPosition(max);
         }

         // adjust enable state, some widgets need the scrollbar to be shown even currently
         // there is nothing to scroll, the state is visible but disabled
         bar.setEnabled(max > 0);
      }
   }
   
   /**
    * Get scroll bar size in characters.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    *          
    * @return  see above
    */
   private double barSize(Orientation orientation)
   {
      double size = 0;
      
      switch (orientation)
      {
         case HORIZONTAL:
            if (hBar != null && hBar.isVisible())
            {
               size = hBar.height();
            }
            break;
         case VERTICAL:
            if (vBar != null && vBar.isVisible())
            {
               size = vBar.width();
            }
            break;
         default:
            break;
            
      }
      
      return size;
   }
         
   /**
    * Create and attach container scroll bars.
    * 
    * @param   bars
    *          Scroll bars.
    */
   private void createScrollBars(Bars bars)
   {
      boolean clearV = false;
      boolean clearH = false;
      
      switch (bars)
      {
         case NONE:
            clearV = true;
            clearH = true;
            both = false;
            break;
         case HORIZONTAL:
            if (hBar == null)
            {
               hBar = createScrollBar(Orientation.HORIZONTAL);
            }
            
            clearV = true;
            both = false;
            break;
         case VERTICAL:
            if (vBar == null)
            {
               vBar = createScrollBar(Orientation.VERTICAL);
            }
            
            clearH = true;
            both = false;
            break;
         case BOTH:
            both = true;
            if (hBar == null)
            {
               hBar = createScrollBar(Orientation.HORIZONTAL);
            }
            if (vBar == null)
            {
               vBar = createScrollBar(Orientation.VERTICAL);
            }
            break;
      }

      if (clearV && vBar != null)
      {
         // if there was previously set a vertical bar, discard it
         screen().getRegistry().removeWidgetsRecursive(vBar);
         vBar.destroy();
         vBar = null;
      }
      
      if (clearH && hBar != null)
      {
         // if there was previously set a horizontal bar, discard it
         screen().getRegistry().removeWidgetsRecursive(hBar);
         hBar.destroy();
         hBar = null;
      }
   }

   /**
    * Create scroll bar.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    *          
    * @return  Scroll bar.
    */
   private ScrollBarGuiImpl createScrollBar(Orientation orientation)
   {
      ScrollBarGuiImpl bar = null;

      GuiWidgetFactory factory = screen().getFactory();
      bar = factory.createScrollBar(WidgetId.nextID(), orientation, viewport());
      
      rootPanel.add(bar);
      bar.addScrollListener(this);
      bar.setVisible(showBars == ShowBars.ALWAYS || showBars == ShowBars.AS_NEEDED ||
                     (showBars == ShowBars.VERTICAL_ALWAYS && orientation == Orientation.VERTICAL));
      
      bar.config().font = config().font;
      
      // set step
      NativeDimension step = viewport().getScrollWidget().getScrollStep();
      if (orientation == Orientation.HORIZONTAL)
      {
         bar.setStep(step.width);
      }
      else
      {
         bar.setStep(step.height);
      }

      // register to receive click events
      screen().getRegistry().addWidgetsRecursive(bar);
      
      return bar;
   }
}