/*
** Module   : ScrollPopupGuiImpl.java
** Abstract : GUI client implementation of scroll system popup.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-------------------------------------
** 001 VIG 20160216 Created initial version.
** 002 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * GUI client implementation class of "scroll" system popup. 
 */
public class ScrollPopupGuiImpl 
extends MenuGuiImpl
{
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param    owner
    *           A widget that will own the popup.
    */
   private ScrollPopupGuiImpl(Widget<?> owner)
   {
      super(null, new MenuConfig());
      config().popupOnly = true;
      config().ownerId = owner.getId().asInt();
   }

   /**
    * Create a system scroll popup menu.
    * 
    * @param   owner
    *          The sroll bar owning created system menu.
    * @param   vertical
    *          Orientation of owning scroll bar.
    *   
    * @return   A popup widget.
    */
   public static ScrollPopupGuiImpl create(ScrollBar<?> owner, boolean vertical)
   {
      ScrollPopupGuiImpl menu = new ScrollPopupGuiImpl(owner);
      menu.initialize(WidgetId.nextID(), menu.config());
      
      // Creating items
      
      MenuItemGuiImpl scrollHereOpt = MenuItemGuiImpl.createDynamicItem("Scroll Here", true, 
               new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            NativePoint popupPos = menu.screenPhysicalLocation();

            ScrollBarGuiImpl scrollBar = (ScrollBarGuiImpl) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);

            int travelClick = scrollBar.getButtonSize();
            int travelSize = scrollBar.getTravelSize();

            NativePoint loc = scrollBar.screenPhysicalLocation();

            if (scrollBar.orientation() == ScrollBar.Orientation.HORIZONTAL)
            {
               travelSize -= scrollBar.getThumbDim().width;
               travelClick += loc.x + scrollBar.getThumbDim().width / 2;
               travelClick = popupPos.x - travelClick;
            }
            else
            {
               travelSize -= scrollBar.getThumbDim().height;
               travelClick += loc.y + scrollBar.getThumbDim().height / 2;
               travelClick = popupPos.y - travelClick;
            }

            /* 
             * travelClick is a distance between popup menu point and thumb middle point in it's
             * first possible position in pixels. When half of thumb's size is more than free  
             * scroll pane space there are possible 2 border cases: 
             *   1. travelClick < 0 means the first position
             *   2. resulting step number is more than max - means the last position.
             */

            if (travelClick < 0) 
               travelClick = 0;

            if (travelSize == 0)
               return;

            int resThumbLoc = scrollBar.getMax() * travelClick / travelSize;

            if (resThumbLoc > scrollBar.getMax())
               resThumbLoc = scrollBar.getMax();

            scrollBar.positionUpdated(resThumbLoc, ScrollStep.CUSTOM, true);
            scrollBar.notifyScrollListeners(resThumbLoc);
         }
      });

      MenuItemGuiImpl rule1 = MenuItemGuiImpl.createDynamicRule();

      boolean vert = owner.orientation().equals(ScrollBar.Orientation.VERTICAL);
      MenuItemGuiImpl topOpt = MenuItemGuiImpl.createDynamicItem(vert ? "Top" : "Left Edge", 
                                                                 true, 
                                                                 new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            ScrollBarGuiImpl scrollBar = (ScrollBarGuiImpl) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);

            scrollBar.onButtonClick(Position.LEFT, ScrollStep.MIN, new MouseEvt(e, scrollBar.ancestor()));
         }
      });

      MenuItemGuiImpl bottomOpt = MenuItemGuiImpl.createDynamicItem(vert ? "Bottom" : "Right Edge", 
                                                                    true, 
                                                                    new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            ScrollBarGuiImpl scrollBar = (ScrollBarGuiImpl) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);

            scrollBar.onButtonClick(Position.LEFT, 
                                    ScrollStep.MAX, 
                                    new MouseEvt(e, scrollBar.ancestor()));
         }
      });

      MenuItemGuiImpl rule2 = MenuItemGuiImpl.createDynamicRule();

      String min = vert ? "Up" : "Left";
      String max = vert ? "Down" : "Right";
      
      MenuItemGuiImpl pgUpOpt = MenuItemGuiImpl.createDynamicItem("Page " + min, 
                                                                  true, 
                                                                  new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            ScrollBarGuiImpl scrollBar = (ScrollBarGuiImpl) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);

            scrollBar.onButtonClick(Position.LEFT, 
                                    ScrollStep.LONG, 
                                    new MouseEvt(e, scrollBar.ancestor()));
         }
      });

      MenuItemGuiImpl pgDownOpt = MenuItemGuiImpl.createDynamicItem("Page " + max, 
                                                                    true, 
                                                                    new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            ScrollBarGuiImpl scrollBar = (ScrollBarGuiImpl) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);

            scrollBar.onButtonClick(Position.RIGHT, 
                                    ScrollStep.LONG, 
                                    new MouseEvt(e, scrollBar.ancestor()));
         }
      });

      MenuItemGuiImpl rule3 = MenuItemGuiImpl.createDynamicRule();

      MenuItemGuiImpl scrollUpOpt = MenuItemGuiImpl.createDynamicItem("Scroll " + min, 
                                                                      true, 
                                                                      new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            ScrollBarGuiImpl scrollBar = (ScrollBarGuiImpl) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);

            scrollBar.onButtonClick(Position.LEFT, 
                                    ScrollStep.SHORT, 
                                    new MouseEvt(e, scrollBar.ancestor()));
         }
      });

      MenuItemGuiImpl scrollDownOpt = MenuItemGuiImpl.createDynamicItem("Scroll " + max, 
                                                                        true, 
                                                                        new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            ScrollBarGuiImpl scrollBar = (ScrollBarGuiImpl) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);

            scrollBar.onButtonClick(Position.RIGHT, 
                                    ScrollStep.SHORT, 
                                    new MouseEvt(e, scrollBar.ancestor()));
         }
      });

      menu.add(scrollHereOpt);
      menu.add(rule1);
      menu.add(topOpt);
      menu.add(bottomOpt);
      menu.add(rule2);
      menu.add(pgUpOpt);
      menu.add(pgDownOpt);
      menu.add(rule3);
      menu.add(scrollUpOpt);
      menu.add(scrollDownOpt);
      
      menu.initScrollItems(vertical);
      
      return menu;
   }
   
   /**
    * Initialized items of scroll menu.
    * 
    * @param   vertical
    *          Orientation of owning scroll bar.
    */
   private void initScrollItems(boolean vertical)
   {
      for (Widget<?> w : widgets())
      {
         if (!(w instanceof MenuItem))
            continue;
         
         MenuItem<?> item  = (MenuItem<?>) w;
         String title;
         if (vertical)
         {
            title = item.config().title.replace("Left Edge", "Top");
            title = title.replace("Right Edge", "Bottom");
            title = title.replace("Left", "Up");
            title = title.replace("Right", "Down");
         }
         else
         {
            title = item.config().title.replace("Top", "Left Edge");
            title = title.replace("Bottom", "Right Edge");
            title = title.replace("Up", "Left");
            title = title.replace("Down", "Right");
         }
         
         if (!title.equals(item.config().title))
            item.setTitle(title);
      }
   }
}