/*
** Module   : ScrollableSelectionListGuiImpl.java
** Abstract : scrollable selection list implementation 
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150423 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
** 003 EVL 20150514 Changing scrollbar painting approach.  Several minor UI fixes.  Adding method
**                  to help dismissing of the drop-down on outside mouse press.  Adding mouse
**                  hover tracking to change the current row of the drop-down list items.
** 004 EVL 20150720 Adding protection from getting negative index in draw() method.  Fix for issue
**                  with entry field editing after item select in simple mode of combo-box.  In
**                  this case isLeaveEventRequired() should return false.  Fix for TAB/BACK_TAB
**                  key processing issue.
** 005 EVL 20150814 Inherit row height from owning combo-box.
** 006 HC  20150831 Scrolling support overhaul.
** 007 CA  20150910 Added mouse actions.
** 008 CA  20150925 When the inner list is losing focus, the drop-down must be exited.
** 009 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 010 EVL 20151020 Fix for drop-down drawing issue with scrolling by scrollbar.
** 011 HC  20151029 Fixed drop-down scroll list width calculation, it must not be smaller than
**                  the parent's width.
**         20151103 Improved GUI COMBO-BOX layout.
**         20151110 GUI combo-box fixes.
**     EVL 20151118 Restored initial logic in leaveAndRepostMouse() removing exitDropDown() for
**                  combo-box body from the end of this method.  This exit cal is not required and
**                  can produce the regressions.  All neccessary drop-down exit cleanups are
**                  performing as reaction for ESC key artifitially generated by the
**                  leaveAndRepostMouse() method. This is event order sensitive method, we need
**                  to keep the right events ordering.  Add substitution for LEFT and RIGHT keys
**                  to work the same way as UP and DOWN keys in GUI mode lists.  The ALT+UP and
**                  ALT+DOWN key combinations close the drop-down.
** 012 EVL 20151120 Fix for line spacing and draw items from the list.  Adding dotted line drawing
**                  for selected item.
** 013 EVL 20151207 Adding SIMPLE mode specific for list item drawing.  No dotted rectangle and
**                  different X for text item start.
** 014 EVL 20160119 Adding focus move postsprocessing.
** 015 IAS 20160130 LOAD-MOUSE-POINTER support.
** 016 HC  20160406 COMBO-BOX drop-down resources cleanup. Various drop-down related fixes.
** 017 HC  20160413 Fixed COMBO-BOX focus management, this also resolves ENTER not confirming
**                  selected item in the drop-down.
** 018 HC  20160511 Fixed COMBO-BOX focusing issues and related changes.
** 019 SVL 20160719 Updated translate() signature.
** 020 EVL 20161104 Fix for GUI drawing as part of the SIMPLE mode combo-box.
** 021 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.model.*;

/**
 * Scrollable selection list GUI implementation.  To be used to support combo box and selection
 * list widgets. 
 */
@SuppressWarnings("unchecked")
public class ScrollableSelectionListGuiImpl
extends ScrollableList<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;
   
   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Selected item box color. */
   private ColorRgb selectionBack;

   /** Selected dotted line color. */
   private ColorRgb selectionDots;

   /** Selected item text color. */
   private ColorRgb selectionFore;

   /** Disabled state text color. */
   private ColorRgb disabledFore;
   
   /** The curent top row in the list to draw. */
   private int rowTop;
   
   /** The GUI specific visible rows value including possible partial row. */
   private int rowsToDisplay;

   /** A callback after item selected. */
   private Runnable onItemSelectedAction;
   
   /** Flag indicating the selected item in the list must be highlighted when drawing. */
   private boolean highlighted = true;
   
   /**
    * Create component using given configuration and data model.
    * 
    * @param    box
    *           The instance with which this object must be associated.
    * @param    model
    *           The data model that represents the contents of the list.
    */
   ScrollableSelectionListGuiImpl(ComboBox<GuiOutputManager> box, ListModel<String> model)
   {
      super(box, model);

      // initialize screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // and font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, box.config(), false);
      // color initializing
      selectionBack = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      selectionFore = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      disabledFore  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      // top row number will be initialized later in setTopRow()
      
      // the color below is negative to selectionBack
      selectionDots  = new ColorRgb(255 - selectionBack.getRed(),
                                    255 - selectionBack.getGreen(),
                                    255 - selectionBack.getBlue());

      // init visible row constant
      rowsToDisplay = visibleRows();
   }

   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public void draw()
   {
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = box.config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_WINDOW, SysColor.COLOR_BTNTEXT);

      // determine colors to use
      gc.colors();

      // get actual rows to view
      final int rowsVisible = visibleRows();
      
      // resolve font
      final FontCache fntCache = gf.font();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final int rowHeight = ((ComboBoxGuiImpl)box).getRowHeight();
      
      // TODO: remove this line and rely on the set size, modifying
      //       the size here is very confusing and may lead
      //       to unexpected results
      d.height = box.config().mode == ComboBoxConfig.Mode.SIMPLE
                   ? ((ComboBoxGuiImpl)box).getSimpleListVisibleHeight()
                   : rowsVisible * rowHeight;

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);

      // clippable drawing start 
      gd.draw(p, clip, () ->
      {
         Theme theme = ThemeManager.getCurrentTheme();
         // inside this call widget origin is always (0,0)
         NativePoint po = new NativePoint(0, 0);
         
         theme.drawSelectionListBackground(gd, gc, 0, 0, d.width, d.height);
         
         // draw entry field text
         gd.setGuiFont(fntCache.font);
         gd.setFontStyle(fntCache.font.style);
         
         int x = po.x;
         int y = po.y;
         int fontSize = fntCache.font.pointSize;
         int modelSize = model().size();
         // adjust top row number to have last item from the list as bottom number 
         if ((rowTop + rowsVisible) > modelSize)
         {
            // protect from getting negative index
            rowTop = modelSize > rowsVisible ? modelSize - rowsVisible : 0;
         }
         
         // check if the list is enabled
         boolean enabled = isEnabled();
         
         // draw items from the top to bottom
         int i = rowTop;
         for (; i < modelSize && i < rowsVisible + rowTop; i++, y+= rowHeight)
         {
            if (y < 0)
            {
               continue;
            }
            
            // draw next visible item in the list
            theme.drawSelectionListItem((ComboBoxGuiImpl) box, gd, gc, x, y, d.width, rowHeight,
                  getItemText(i), fontSize, enabled, highlighted, i == currentRow());
         }
         
         // draw possible remaining partial item
         if (rowsToDisplay > rowsVisible && i < modelSize)
         {
            theme.drawSelectionListItem((ComboBoxGuiImpl) box, gd, gc, x, y, d.width, rowHeight,
                  getItemText(i), fontSize, enabled, highlighted, i == currentRow());
         }
      });
   }
   
   @Override
   public void scroll(Integer newXPosition, Integer newYPosition)
   {
      rowTop = newYPosition;
       
      // and repaint the list
      repaint();
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      int rowHeight = ((ComboBoxGuiImpl)box).getRowHeight();
      
      if (model().size() > 0)
      {
         NativePoint n = translate(e);
            
         // calculate clicked line
         int row = n.y / rowHeight + rowTop;

         // in combo-box it is possible to generate selection for already selected item
         // go to selected item
         super.ensureRowVisible(row, false);
         // select new item
         selectItem();
         // turn selection mode on
         if (!highlighted)
         {
            highlighted = true;
         }
            
         if (onItemSelectedAction != null)
         {
            onItemSelectedAction.run();
         }
      }
   }

   /**
    * Notification of a mouse moving event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      if (!isEnabled() || getBox().config().mode == ComboBoxConfig.Mode.SIMPLE)
      {
         return;
      }
      
      int rowHeight = ((ComboBoxGuiImpl)box).getRowHeight();
      
      synchronized (this)
      {
         if (model().size() > 0)
         {
            NativePoint n = translate(e);

            // calculate line under the mouse
            int row = n.y / rowHeight + rowTop;

            // restrict repainting only when we are over the new line
            if (row != currentRow())
            {
               // do not select new item, only make it the current one
               super.ensureRowVisible(row, false);
               
               // no need to call additional repaint here, will be done in other place
            }
         }
      }
   }
   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      
      // nothing to do in GUI mode 
   }
   
   /**
    * Intercept key events and handle some special cases which are not
    * handled by base class.
    * 
    * @param    kvt
    *           Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput kvt)
   {
      ComboBoxGuiImpl cbbRef = (ComboBoxGuiImpl) getBox();
      int key = kvt.actionCode();

      // disable ENTER key itself as event generator to avoid value change duplication
      // this event has already registered as drop-down exit trigger
      if (key == Keyboard.KA_RETURN)
      {
         kvt.consume();
      }
      // in simple mode TAB/BACK_TAB keys should move the focus to next/prev widget
      // comparing to the combo-box this list belongs to
      else if ((key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB) &&
               cbbRef != null && cbbRef.config().mode == ComboBoxConfig.Mode.SIMPLE)
      {
         cbbRef.processFocusMoveKey(kvt);
      }
      else
      {
         // the focus move key should be deferred for opened drop-down
         if ((key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB) &&
               cbbRef != null && cbbRef.config().mode != ComboBoxConfig.Mode.SIMPLE)
         {
            // shold be involed after dop-down is gone
            KeyInput ki = new KeyInput(cbbRef, EventType.KEY_PRESSED, kvt.keyCode());
            cbbRef.deferKeyEvent(ki);
            // disable beep for GUI mode stopping TAB/BACK-TAB processing here
            kvt.consume();
            return;
         }
         
         // get key code
         int keyCode = kvt.keyCode();

         // in GUI mode ALT+UP and ALT+DOWN cause the drop-down to dismiss
         if (keyCode == Key.VK_ALT_CURSOR_DOWN || keyCode == Key.VK_ALT_CURSOR_UP)
         {
            // consume original event
            kvt.consume();
            // emit the new leave event
            KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_LEAVE);
            EventManager.postEvent(ki);
            // and stop processing the current one
            return;
         }
         // in GUI mode RIGHT and LEFT keys works as DOWN and UP, make substitution
         else if (keyCode == Key.VK_CURSOR_RIGHT)
         {
            kvt.setKeyCode(Key.VK_CURSOR_DOWN);
            kvt.setActionCode(Keyboard.KA_CURSOR_DOWN);
         }
         else if (keyCode == Key.VK_CURSOR_LEFT)
         {
            kvt.setKeyCode(Key.VK_CURSOR_UP);
            kvt.setActionCode(Keyboard.KA_CURSOR_UP);
         }
         
         // restore selection mode to on if disabled
         if ((keyCode == Key.VK_CURSOR_RIGHT || keyCode == Key.VK_CURSOR_LEFT ||
              keyCode == Key.VK_CURSOR_UP || keyCode == Key.VK_CURSOR_DOWN) &&
             !highlighted)
         {
            highlighted = true;
         }
            
         super.processKeyEvent(kvt);
      }
   }

   /**
    * Sets the number of rows to display including possible partial row.
    * 
    * @param   dispRows
    *          New value for rows to display.
    */
   public void setRowsToDisplay(int dispRows)
   {
      rowsToDisplay = dispRows;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[] { MouseEvent.MOUSE_CLICKED, MouseEvent.MOUSE_MOVED };
   }
   
   /**
    * Set item number to be displayed on the top of the drop-down list.
    * 
    * @param   index
    *          The index of the top item.
    */
   @Override
   protected void setTopRow(int index)
   {
      ensureRowVisible(index, false);
      rowTop = fixTopRow(index);
      postScrollEvent(null, rowTop);
   }
   
   /**
    * Get item number to be displayed on the top of the list.
    * 
    * @return   See above.
    */
   protected int getTopRow()
   {
      return rowTop;
   }
   
   /**
    * Select row with specified index.
    * 
    * @param   index
    *          Item index.
    */
   @Override
   public void select(int index)
   {
      super.select(index);
      postScrollEvent(null, rowTop);
   }
   
   /**
    * Select specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   @Override
   public void select(int from, int to)
   {
      super.select(from, to);
      postScrollEvent(null, rowTop);
   }
   
   /**
    * The scrollable list's boundary does not keep track of the view size - it posts a repaint 
    * event for the entire list, regardless how small a part is visible.
    * <p>
    * If the list is attached to a parent, then the repaint will be posted for its parent.
    */
   @Override
   public void repaint()
   {
      if (parent() != null)
      {
         parent().repaint();
      }
   }
   
   /**
    * Get the current effective custom mouse pointer
    * 
    * @return the current effective custom mouse pointer  
    *        
    */
   @Override
   public MousePtrWrapper getEffectiveMousePointer()
   {
      return box.isSimpleMode() ? 
                     MousePtrWrapper.DEFAULT : 
                     box.getEffectiveMousePointer();
   }

   /**
    * Get the Loaded custom mouse pointer if it is set
    * 
    * @return the loaded custom mouse pointer if it is set  
    *        
    */
   @Override
   public Optional<MousePtrWrapper> getLoadedMousePointer()
   {
      return box.getLoadedMousePointer(); 
   }

   /**
    * Check if the default mouse pointer should be used regardless of
    * the parent settings
    * 
    * @return
    *        <code>true</code> if the default mouse pointer should always be used
    */
   @Override
   public boolean enforceDefaultMousePtr()
   {
      return box.isSimpleMode();
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }
   
   /** 
    * Set an action to be called when item is selected. The action is called
    * only when the selection also uncoditionally terminates the list.  
    */
   public void setOnItemSelected(Runnable action)
   {
      onItemSelectedAction = action;
   }

   /** 
    * Set the mode if the current row is drawing as selected or not.  
    * 
    * @param   highlighted
    *          If <code>TRUE</code> - the current row is drawing with selected color attributes,
    *          <code>FALSE</code> - the current row is drawing as the rest of the list.
    */
   public void setHighlighted(boolean highlighted)
   {
      this.highlighted = highlighted;
   }

   /**
    * Shifts current row by specified delta.
    * 
    * @param   delta
    *          the number of list items to shift current row by
    * @param   scroll
    *          indicates whether list should be scrolled
    *          to the shift row
    */
   @Override
   protected void shiftCurrentRow(int delta, boolean scroll)
   {
      super.shiftCurrentRow(delta, false);
  
      int origRowTop = rowTop;
      
      // recalc top row changes
      if (currentRow() < rowTop)
      {
         // upper than visible region
         rowTop = currentRow(); 
      }
      else if (currentRow() >= rowTop + visibleRows())
      {
         // lower than visible region
         rowTop = currentRow() - visibleRows() + 1;
      }
      
      // and notify main combo-box about new selection for keyboard event origin case
      if (scroll)
      {
         selectItem();
      }
   }
   
   /**
    * Calculate the scroll Y position based on the current state. 
    * 
    * @return   the new scroll Y position
    */
   @Override
   protected int calculateScrollYPosition()
   {
      return fixTopRow(rowTop);
   }

   /**
    * Draws the next text item in the list inside visible part of the drop-down.
    * 
    * @param    number
    *           The number of the item to draw.
    * @param    x
    *           The X coordinate of the enclosing rectangle to draw item.
    * @param    y
    *           The Y coordinate of the enclosing rectangle to draw item.
    * @param    rw
    *           The pixel width of the enclosing rectangle to draw item.
    * @param    rh
    *           The pixel height of the enclosing rectangle to draw item.
    * @param    fontSize
    *           The vertical size of font to draw item text.
    * @param    enabled
    *           Flag indicating the enabled state of the list.
    */
   private void drawItem(int number, int x, int y, int rw, int rh, int fontSize, boolean enabled)
   {
      // fill the background of the selected item
      if (number == currentRow() && highlighted)
      {
         gd.setColor(selectionBack);
         gd.fillRect(x, y, rw, rh);
      }

      // set the color of the text
      if (enabled && number == currentRow() && highlighted)
      {
         gd.setColor(selectionFore);
      }
      else if (enabled)
      {
         gd.setColor(gc.fgColor);
      }
      else
      {
         gd.setColor(disabledFore);
      }
      
      Theme theme = ThemeManager.getCurrentTheme();
      // draw item text
      gd.drawString(getItemText(number),
                    box != null && box.config().mode == ComboBoxConfig.Mode.SIMPLE
                       ? x + theme.getComboBoxEntryShift() : x + ComboBoxGuiImpl.TEXT_SHIFT,
                    y + 1 + (rh + fontSize)/2);
      
      // draw the dotted line for selected item (not for SIMPLE mode of the combo-box)
      if (enabled && number == currentRow() &&
          !(box == null || box.config().mode != ComboBoxConfig.Mode.DROP_DOWN_LIST))
      {
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         gd.setColor(selectionDots);
         // take into account possible vertical scrollbar width 
         gd.drawRect(x, y, rw - 1 - ((ComboBoxGuiImpl)box).getDropDownScrollBarWidth(), rh - 1);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Converts the passed in value to a valid row unless the
    * value is already valid.
    * 
    * @param    desiredTopRow
    *           The value to check and fix.
    *          
    * @return   valid row value
    */
   private int fixTopRow(int desiredTopRow)
   {
      if (desiredTopRow < 0)
      {
         return 0;
      }
      
      int maxTopRow = model().size() - visibleRows() - 1;
      
      if (maxTopRow < 0)
      {
         maxTopRow = 0;
      }
      
      if (desiredTopRow > maxTopRow)
      {
         return maxTopRow;
      }
      
      return desiredTopRow;
   }
}