/*
** Module   : SelectionListGuiImpl.java
** Abstract : GUI selection-list implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150818 Initial version.  Separating GUI specific code here.
** 002 HC  20150831 Scrolling support overhaul.
** 003 IAS 20150924 AUTO-RESIZE attribute support
** 004 GES 20151008 Removed direct usage of tooltip registration (it is done generically now).
** 005 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 006 EVL 20151013 Restore disabled logic to properly compute the size internal variables. 
** 007 HC  20151112 Implemented THREE-D runtime support.
** 008 IAS 20151214 LOAD-MOUSE-POINTER support.
** 009 CA  20160216 Cache the intermediate widget states.
** 010 SBI 20161005 Fixed the left and bottom borders.
** 011 CA  20170131 Fixed color inheritance when not explicitly set at widget but maybe set at 
**                  frame (and frame might or might not be THREE-D). 
**     CA  20161101 Added support for mnemonic processing.
** 012 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.util.*;
import java.awt.event.*;

/**
 * Specific GUI implementation of SELECTION-LIST widget.
 */
@SuppressWarnings("unchecked")
public class SelectionListGuiImpl
extends SelectionList<GuiOutputManager>
implements AutoResizable<SelectionListConfig>
{
   /** Shift in X to draw text inside list. */
   public static final int TEXT_SHIFT = 3;
   
   /** Total size of the selection-list border 1 pixel on each side. */
   private static final int BORDER_SIZE = 2;
   
   /** Screen driver */
   private GuiDriver gd;

   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Cached border instance. */
   private Border<GuiOutputManager> border;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SelectionListGuiImpl(WidgetId id, SelectionListConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @SuppressWarnings("unchecked")
   @Override
   public void initialize(WidgetId id, SelectionListConfig cfg)
   {
      // call the base code first
      super.initialize(id, cfg);

      // screen driver
      if (gd == null)
      {
         gd = (GuiDriver) OutputManager.getDriver();
      }
      // and font resolver
      if (gf == null)
      {
         gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, cfg, false);
      }
      
      ScrollPaneGuiImpl sl = (ScrollPaneGuiImpl) scrollpane;
      
      // set up the scrollbars, the first part is the scrollpane
      if (config.scrollbarVertical && config.scrollbarHorizontal)
      {
         sl.setScrollBars(Bars.BOTH);
      }
      else if (config.scrollbarVertical)
      {
         sl.setScrollBars(Bars.VERTICAL);
      }
      else if (config.scrollbarHorizontal)
      {
         sl.setScrollBars(Bars.HORIZONTAL);
      }

      // the scrollbars should always be visible if defined
      if (config.scrollbarVertical || config.scrollbarHorizontal)
      {
         sl.setShowBars(ScrollPaneGuiImpl.ShowBars.ALWAYS);
      }

      SelectionListBodyGuiImpl body = (SelectionListBodyGuiImpl) list;
      ((BorderedPanelGuiImpl) sl.getRootPanel()).setColorResolver(body.gc);
      
      // set up some internals with the selection-list body
      list.initialize(id, cfg);

      screen().getRegistry().addWidget(this);
      screen().getRegistry().addWidget(scrollpane);
   }

   /**
    * Draw all the components in this container.
    */
   @Override
   public void draw() 
   {
      if (!isVisible())
         return;

      ThinClient tc = ThinClient.getInstance();

      if (scrollpane.isVisible() && (tc.viewRefresh || tc.viewDisplay))
      {
         // setup the border, do this only once when the widget has been realized
         if (border == null)
         {
            border = isThreeD() ? new ThreeDBorderGuiImpl() : new LineBorderGuiImpl();
            scrollpane.setBorder(border);
            
            int bgColor = isThreeD() && !config.enabled ? 
                     SysColor.COLOR_3DFACE : SysColor.COLOR_WINDOW;
          
            scrollpane.config().sysbgcolor = bgColor;
         }
         
         // get location and size
         NativePoint p = physicalLocation();
         NativeDimension d = physicalDimension();
         
         // set up clipping rectangle
         NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
         
         // clippable drawing start 
         gd.draw(getId(), p, clip, () -> scrollpane.draw());
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      // route mouse event to selection list body
      list.mousePressed(e);
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      // route mouse event to selection list body
      list.mouseReleased(e);
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      // route mouse event to selection list body
      list.mouseClicked(e);
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(SelectionListConfig beforeUpdate)
   {
      if (!StringHelper.equal(config.label, beforeUpdate.label) ||
          config.font != beforeUpdate.font)
      {
         config.widthChars = 0;
         adjustSizeImpl();
      }
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      return processMnemonicImpl(key, this);
   }

   /**
    * Create handler for processing mouse hover events
    * 
    * @return handler for processing mouse hover events
    */
   @Override
   protected MouseHoverAction createMouseHoverAction()
   {
      return new MouseHoverAction(this, scrollpane.getId().asInt());
   }

   /**
    * Adjust SelectionList size ChUI specific part.
    */
   @Override
   protected void adjustSizeImpl()
   {
      Theme theme = ThemeManager.getCurrentTheme();
      
      // native width and height for selection list
      int nativeWidth = 0;
      int nativeHeight = 0;
      // current row height and char width for selection list
      int slRowHeight = ((SelectionListBodyGuiImpl)list).getRowHeight();
      int slCharWidth = ((SelectionListBodyGuiImpl)list).getCharWidth();
      // do we need to update inner chars and lines flag(no need by default)
      boolean explicitSizeSetup = config.widthChars != 0 && config.heightChars != 0;
      // there are two possible ways to set up the selection-list size:
      //  - size phrase when we have to be based on the width/height chars/pixels
      //  - inner-lines and inner-chars options given character units
      // both cases are considered below
      if (explicitSizeSetup)
      {
         // for compatibility reasons the insets size preserved here however looks like
         // the insets are always 0 for GUI selection list
         nativeWidth = config.widthPixels -
                       screen().coordinates().widthToNative(insets().left + insets().right);
         nativeHeight = config.heightPixels -
                       screen().coordinates().heightToNative(insets().top + insets().bottom);
      }
      else
      {
         // in this case we use inner chars and lines option as the base for size calculation
         nativeHeight = config.innerLines * slRowHeight + BORDER_SIZE;
         nativeWidth = config.innerChars * slCharWidth + BORDER_SIZE;
      }
      
      // inner size in pixels(without the border)
      int innerCharsInPixels = nativeWidth - BORDER_SIZE;
      int innerLinesInPixels = nativeHeight - BORDER_SIZE;

      // adjust scrollpane height to the current font size
      //
      // horizontal scrollbar adds additional space
      if (config.scrollbarHorizontal)
      {
         int barHeight = ((ScrollPaneGuiImpl)scrollpane).barSizeNative(Orientation.HORIZONTAL);
         if (explicitSizeSetup)
         {
            innerLinesInPixels -= barHeight != 0 ? barHeight : theme.getScrollBarGaugeWidth();
         }
         else
         {
            nativeHeight += barHeight != 0 ? barHeight : theme.getScrollBarGaugeWidth();
         }
      }
      // the following two doubles are not the same innerLines >= inCharHeight
      double inCharHeight = screen().coordinates().heightFromNative(nativeHeight);
      double innerLines = (double)innerLinesInPixels / (double)slRowHeight;

      // calculate pixel width
      //
      // vertical scrollbar adds additional space
      if (config.scrollbarVertical)
      {
         int barWidth = ((ScrollPaneGuiImpl)scrollpane).barSizeNative(Orientation.VERTICAL);
         if (explicitSizeSetup)
         {
            innerCharsInPixels -= barWidth != 0 ? barWidth : theme.getScrollBarGaugeWidth();
         }
         else
         {
            nativeWidth += barWidth != 0 ? barWidth : theme.getScrollBarGaugeWidth();
         }
      }
      // the following two doubles are not the same innerChars >= inCharWidth
      double inCharWidth = screen().coordinates().widthFromNative(nativeWidth);
      double innerChars = (double)innerCharsInPixels / (double)slCharWidth;
      
      // set new size for GUI selection-list
      setSize(new Dimension(inCharWidth, inCharHeight));

      // and set size related variables for all components inside selection-list
      scrollpane.setWidth(inCharWidth);
      scrollpane.setHeight(inCharHeight);

      // using ceil() guarantees the empty lines/cells missing
      list.setColumns((int)Math.ceil(innerChars));
      list.setVisibleRows((int)Math.ceil(innerLines));

      // sets the exact internal pixel width
      ((SelectionListBodyGuiImpl)list).setInternalWidth(innerCharsInPixels);
      
      // update widget config size if necessary
      if (config.widthChars == 0.0)
      {
         config.widthChars = inCharWidth;
      }
      if (config.heightChars == 0.0)
      {
         config.heightChars = inCharHeight;
      }
      // update widget config inner chars and lines if necessary 
      if (explicitSizeSetup)
      {
         // using floor() guarantees all lines/cells will be completely visible
         config.innerChars = (int)Math.floor(innerChars);
         config.innerLines = (int)Math.floor(innerLines);
      }
   }
}