/*
** Module   : StatusLineGuiImpl.java
** Abstract : window status line implementation 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140522 First version.
** 002 CA  20140711 Enhanced font support; refactoring related to MESSAGE-AREA widget.
** 003 CA  20140805 Added generic's to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Some refactoring to match new drawing techniques.
** 004 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 005 MAG 20140912 Implements GUI status line.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 008 MAG 20140917 Changed background color.
** 009 CA  20141021 Removed the repaint() API: all repainting needs to be done via the common
**                  event processing; do no draw the widget immediately.
** 010 MAG 20141016 Change text vertical position. Eliminate GuiOptions.BORDER_SIZE dependency.
** 011 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 012 CA  20150507 Draw text centered vertically.
** 013 GES 20150425 Eliminated AWT dependencies.
** 014 EVL 20151008 Fix for status line drawing issues.
** 015 CA  20160129 Commented H014 code in draw (gd.flush()) - it leaves the JS side in an 
**                  inconsistent state.
**         20160216 Cache the intermediate widget states.
** 016 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;

/**
 * In GUI implementation the window status line is implemented as widget.
 */
@SuppressWarnings("unchecked")
public class StatusLineGuiImpl
extends StatusLine<GuiOutputManager>
{
   /** Constants used to draw right-bottom corner */
   private static final int G = 1;

   /** Screen driver */
   private GuiDriver gd;
   
   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Parent window of this widget */
   private WindowGuiImpl window;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Widget height in pixels */
   private int widgetHeight;

   /** Widget height in characters */
   private double widgetHeightChars;

   /**
    * Hide base class public constructor.
    * <p>
    * Create a widget representing the window's status line.
    * 
    * @param    id
    *           The widget's ID.
    */
   StatusLineGuiImpl(WidgetId id)
   {
      super(id);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // set default BGCOLOR and FGCOLOR
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_3DFACE, SysColor.COLOR_WINDOWTEXT);
      // color used to draw bottom right corner
      gc.miscColor = gd.getSysColor(SysColor.COLOR_WINDOW);
      // margins
      nativeInsets = new NativeInsets(3, 9, 6, 8);
   }

   /**
    * Widget drawing method.
    */
   @Override
   public void draw()
   {
      if (!isVisible())
      {
         return;
      }
      
      // resolve colors
      gc.colors();
      
      // get status line coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final String text = getLabelText();
            
      gd.draw(getId(), p, null, () ->
      {
         Theme theme = ThemeManager.getCurrentTheme();
         // compute text drawing area 
         NativePoint txtOrig = new NativePoint(nativeInsets.left, nativeInsets.top);
         NativeDimension txtDim = 
               new NativeDimension(d.width - nativeInsets.right - nativeInsets.left,
                                   d.height - nativeInsets.top - nativeInsets.bottom + 3);
         
         theme.drawStatusBar(gd, gc, 0, 0, d.width, d.height,
                             text, gf.font(), txtOrig, txtDim, window.config().resize);
      });
   }

   /**
    * Attach this widget to the specified window.
    * 
    * @param    window
    *           The parent window.
    */
   @Override
   public void attach(Window<GuiOutputManager> window)
   {
      this.window = (WindowGuiImpl) window;
      init();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return widgetHeightChars;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return window.workspace.width();
   }
   
   /**
    * Status bar initialization.
    */
   private void init()
   {
      config().font = window.resolveStatusAreaFont();
      gf = new GuiFontResolver(window, gd, config(), false);
      FontCache cached = gf.font();
      widgetHeight = cached.fontHeight + nativeInsets.top + nativeInsets.bottom;
      widgetHeightChars = screen().coordinates().heightFromNative(widgetHeight);
   }   
}