/*
** Module   : ToolTip.java
** Abstract : GUI tooltip feature implementation.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20141201 Initial version.
** 002 EVL 20141209 Fix for tooltip offset.  Separating tooltip into own thread.
** 003 HC  20150323 Removed business logic from config classes.
** 004 GES 20150425 Eliminated AWT dependencies.
** 005 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 006 EVL 20151012 Fix for tooltip painting approach.
** 007 IAS 20161017 More generic type for the constructor argument.
** 008 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The help string showing when mouse become over the GUI widget.  There is no counterpart in
 * character mode interface.
 */
@SuppressWarnings("unchecked")
public class ToolTip
implements Runnable
{
   /** The distance from mouse pointer coordinate to draw tooltip. */
   private static final int ORIGIN_SHIFT = 16;
   
   /** The delay between mouse enter widget and draw tooltip, ms. */
   private static final int DELAY_MS = 100;
   
   /** The time the tooltip is shown without any action, ms. */
   private static final int DURATION_MS = 5000;
   
   /** Tooltips enabled by default for GUI. */
   private static boolean sessionTooltips = false;

   /** Screen driver. */
   private GuiDriver gd;
   
   /** Font to draw the tooltip. */
   private int fntToolTip;

   /** The referecnce to widget the tooltip is belonging to. */
   private AbstractWidget<GuiOutputManager> owner = null;

   /** Flag indicating the tooltip is on the screen. */
   private boolean toolTipToHide = false;
   
   /** Starting position to draw the tooltip (-1,-1) means unassigned. */
   private NativePoint position = new NativePoint(-1, -1);
   
   /**
    * This constructor creates a tooltip object for given screen driver.
    * 
    * @param   wid
    *          The widget's reference for tooltip.
    * @param   fnt
    *          The font to use to draw the tooltip.
    */
   public ToolTip(Widget<?> wid, int fnt) 
   {
      // initialize tooltip internals
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // font to render text
      fntToolTip = fnt;
      
      // widget owning the tooltip
      owner = (AbstractWidget) wid;
   }
   
   /**
    * Main tooltip life cycle.
    */
   public void run()
   {
      // initial delay before show
      try
      {
         Thread.sleep(DELAY_MS);
      }
      catch (InterruptedException ie)
      {
         // tooltip engine interrupted before text shown
         return;
      }

      // draw the text
      show(position);
      
      // wait the tooltip duration time
      try
      {
         Thread.sleep(DURATION_MS);
      }
      catch (InterruptedException ie)
      {
         // waiting was externally interrupted
      }
      finally
      {
         // hiding the text
         hide();
      }
   }
   
   /**
    * Setting new position for tooltip drawing.
    * 
    * @param   orig
    *          Tooltip origin.
    */
   public void setOrigin(NativePoint orig)
   {
      // need to modify a bit the tooltip position comparing to origin
      if (orig != null)
      {
         position.x = orig.x;
         position.y = orig.y + ORIGIN_SHIFT;
      }
   }
   
   /**
    * Checks if the given tooltip engine should be activated in a curren conditions.  This depends
    * on the session wide TOOLTIPS boolean attribute and on the cuttent text value for given
    * widget.  If TOOLTIPS attribute is <code>false</code> or text does not exist, empty or
    * unknow value the tooltip will not be drawn. 
    *
    * @param   tooltipTxt
    *          Tooltip string to check.
    *
    * @return  <code>true</code> if the tooltip needs to be used,
    *          <code>false</code> otherwise.
    */
   public static boolean isActive(String tooltipTxt)
   {
      // first check the SESSION:TOOLTIPS attribute value
      // second check the tooltip string itself
      // null string or empty string both disable the tooltip engine
      return sessionTooltips && tooltipTxt != null && tooltipTxt.length() > 0;
   }

   /**
    * Sets the SESSION:TOOLTIPS attribute for this session.
    *
    * @param    tooltips
    *           The new value for session related TOOLTIPS attribute.
    */
   public static void setSessionTooltips(boolean tooltips)
   {
      // if nothing changed - do nothing
      if (sessionTooltips == tooltips)
      {
         return;
      }
      
      // accept new value
      sessionTooltips = tooltips;
   }
   
   /**
    * Accesses the SESSION:TOOLTIPS attribute for this session.
    *
    * @return   <code>true</code> if tooltips are available for this session
    *           <code>false</code> otherwise.
    */
   public static boolean getSessionTooltips()
   {
      return sessionTooltips;
   }
   
   /**
    * Draw widget tooltip help string at the specified position.
    * 
    * @param   orig
    *          Tooltip origin.
    */
   private void show(NativePoint orig)
   {
      // simple validity check
      if (owner == null || orig.x == -1 || orig.y == -1)
      {
         return;
      }
      
      // get the current tooltip text, it can be different from call to call
      final String toolTip = UiUtils.getConfig(owner, BaseConfig.class).tooltip;
      
      if (isActive(toolTip))
      {
         // mark we have something to be hidden further
         toolTipToHide = true;
         
         // set up tooltip font
         final FontDetails fnt = FontManager.resolveFont(owner.window(), 
                                                         fntToolTip, 
                                                         owner.screen().getInstanceDriver());
         
         // calculate required rectangle to include the text
         final NativeDimension d =
            new NativeDimension(owner.getTextWidthNative(toolTip, fntToolTip) + 4,
                                owner.getTextHeightNative(toolTip, fntToolTip) + 2); 
         
         // draw the tooltip itself
         ThinClient.getInstance().eventDrawingBracket(owner, false, true,
               () -> ThemeManager.getCurrentTheme().drawToolTip(
                     this, gd, orig.x, orig.y, d.width, d.height, fnt, toolTip));
      }
   }
   
   /**
    * Hiding the widget tooltip string if one exists.
    */
   private void hide()
   {
      if (toolTipToHide)
      {
         toolTipToHide = false;
         
         // repaint everything behind the owner, widget only owner widget is not enough
         ThinClient.getInstance().eventDrawingBracket(owner, false, true, () ->
               WindowManager.resolveWindow(owner).repaint());
      }
   }
}