/*
** Module   : WindowGuiImpl.java
** Abstract : implements window widget drawing on GUI interfaces
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support; refactoring related to MESSAGE-AREA widget.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Reworked the window parts to make use of widgets and generic drawing.
** 004 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 005 MAG 20140912 Set default FGCOLOR and BGCOLOR.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 008 MAG 20140923 Added GUI message area.
** 009 MAG 20141009 Insert indicator widget is not used in GUI.
** 010 CA  20141021 Finished support for the window title bar, window maximize/restore/minimize
**                  and close.  Added support for legacy 4GL window and mouse events.  
**                  Improved window resize.  TODO: support for the Size and Move options.
** 011 MAG 20141020 Added GUI Frame support.
** 012 HC  20141021 Implemented window move/position support.
** 013 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 014 CA  20141123 Fixed ENTRY/LEAVE when window focus get changed.  Fixed default placement of  
**                  dynamic windows.  Fixed window stacking at OS level.
** 015 CA  20141125 Fixed widget deregistration when the window is terminated.  Also, the window 
**                  needs to be placed only after the layout has been calculated (and done only 
**                  once).
** 016 EVL 20141126 Content pane GUI fix.
** 017 CA  20141206 In GUI Window, INSERT mode is always ON (and can't be changed).
** 018 HC  20141207 Implemented runtime support for widget coordinate attributes. Improved window
**                  mouse drag and mouse resize behavior. Implemented window synchronization with
**                  the server coordinate attributes.
** 019 EVL 20150107 Fix for dynamic windows renderer missing exception.
** 020 EVL 20150114 Adding API to set/load window icon.
** 021 EVL 20150311 Filter events for frame's default button and route them to target frame.
** 022 CA  20150408 Added runtime support for WINDOW:PARENT attribute.
** 023 HC  20150323 Removed business logic from config classes.
** 024 HC  20150429 Fixed window positioning.
** 025 VIG 20150407 Fixed getDefaultWindow call (added type).
** 026 CA  20150507 Fixed problems related to window sizing (explicit or implicit).
** 027 CA  20150508 Fixed H026 regression when resizing window via mouse (not attributes).
** 028 EVL 20150507 Adding widget registration when added frame is a drop-down.  Otherwise it is
**                  not possible to route mouse events to drop-down's scrollbar.  Adding method
**                  to get workspace object, used in combo-box drop-down.
** 029 HC  20150508 Improvements to GUI window scrolling support.
** 030 HC  20150509 Fixed calculation of window scrollable size.
** 031 GES 20150425 Match class name change for GuiOptions to GuiConstants.
** 032 CA  20150520 Fixed a NPE in hide on CTRL-C.
** 033 HC  20150612 Implemented support for modal windows.
** 034 HC  20150622 Fixed GUI frame to be properly centered in its parent window.
** 035 HC  20150626 Fixed an unhandled exception when showing a modal window.
** 036 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
** 037 CA  20150615 Moved default button processing code from processEvent to 
**                  FrameGuiImpl.processDefaultButton.
** 038 CA  20150729 Fixed a NPE in getWidgets() - it might be called in cases when the window has
**                  not been initialized yet..
** 039 HC  20150803 Improved runtime support for WINDOW size attributes.
** 040 CA  20150730 Fixed window title drawing.  Also, the window is marked as realized as soon as
**                  it is made visible.
** 041 VIG 20150818 Added ALT key press support for activation key navigation mode of MENUBAR.
** 042 HC  20150831 Scrolling support overhaul.
** 043 CA  20150826 Select the window before registering widgets with the driver.
**                  Added moveToTop/ToBottom/BeforeTabItem: these needs to act at the workspace, 
**                  not window (to be in sync with how the ChUI window behaves in P2J).
**                  Moved the system popup to use the common MenuGuiImpl code.
**                  Fixed border drawing. Temporarily disabled truncateWindowHeightQuirk().
** 044 IAS 20150926 Changed updateSizeWorker argument type.
** 045 CA  20150925 MOUSE_CLICKED event must use the same source as MOUSE_PRESSED. Fixed
**                  iconify() and system menu abends. Fixed frame detach from container.
**                  Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
**                  Changes to allow multi-window focus management in GUI. Added 3D border to the
**                  window.
** 046 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 047 IAS 20151012 DEFAULT-BUTTON/CANCEL-BUTTON runtime support
** 048 CA  20151013 Fixed window resize via attributes (they target the workspace, not the full 
**                  window).
**     HC  20151020 Fixed disappearing 3D border.
** 049 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     HC  20151027 Implemented WINDOW MESSAGE-AREA and STATUS-AREA attributes.
** 050 HC  20151124 Fixed wrong workspace background color when its window is THREE-D.
** 051 CA  20151127 When losing focus, the window must not change the ACTIVE-WINDOW.
** 052 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 053 EVL 20160108 Removing drop-down related dead code.
**     HC  20160113 Default-window is activated when initially shown. The logic was originally in
**                  Swing driver, but moved in here so that the feature is available on all GUI
**                  drivers.
** 054 EVL 20160124 Adding method to check if the window has title bar. 
** 055 HC  20160203 Improved runtime support of WINDOW:SCROLL-BARS attribute.
** 056 IAS 20160217 LOAD-MOUSE-POINTER support.
** 057 HC  20160304 Replaced Window.moveBy() with setPhysicalLocation(). Window layout
**                  improvements. Fixed window size reporting.
** 058 HC  20160308 Removed an optimization that is not needed anymore, a potential regression
**                  cause.
** 059 IAS 20160314 DEBUG-ALERT support.
** 060 CA  20160327 When drawing the border, check if the invalidation rectangles registered in
**                  ScreenBitmap intersect it or not.
**                  Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 061 SBI 20160328 Changed to hide a window title bar and a border iff the graphics environment
**                  doesn't support window decorations, added the window workspace background
**                  changed color listener in order to set the desktop background iff it is
**                  supported.
** 062 HC  20160406 Overhaul of window-activation logic.
** 063 VIG 20160209 Added Window deactivation processing to send ENTRY event to a widget,
**                  that owned previously opened system popup (see processEvent()).
**                  Added 0-check for travelSize in the "ScrollHere" trigger.
**                  Moved window title popup to a separate class and added editor and added create*
**                  methods for system popups.
** 064 HC  20160413 Fixed top-level windows to always receive focus event when activated, also
**                  deduplicated processing of focus events.
** 065 EVL 20160414 Javadoc error fix.
** 066 HC  20160414 Any only top-level window when initially shown must be activated. 
** 067 CA  20160418 Fixed a NPE in getPortableMouseEventCoordinates, when the window doesn't have
**                  the decorations (and the title is missing).
** 068 HC  20160428 Removed redundant override setSize(), contentPane must always be of the same
**                  size as its parent. GUI window border drawing changed to use the Border
**                  interface mechanism, also removed hardcoded assumptions of the border size.
**                  Fixed mouse-resize moving the window when resizing by the left or top edge.
** 069 HC  20160429 Explicitly assigned widget ID to the mouse-sensitive bordered panel. Window
**                  border is not drawn when window decoration is not supported by the driver.
** 070 HC  20160607 Fixed memory leak - window must be removed from WindowManager when destroyed.
**                  GUI menus reworked to be displayed in dedicated top-level windows.
** 071 SBI 20160610 Changed to register mouse popupable_right and popapable_left window title
**                  bar's widgets regions to be known by the Web client.
** 072 HC  20160712 Prevented an NPE when last DIALOG-BOX closed.
** 073 CA  20160915 Fixed window deletion: the child frames/menus must survive.
** 074 VIG 20160815 Added MENUBAR flag reset processing WINDOW_ACTIVATED(false).
**                  Fixed accelerators processing for menubar.
** 075 CA  20170104 Some improvements in window decoration support for embedded mode (modal 
**                  windows always have decorations).
** 076 HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE. Fixed the window activation.
** 077 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
// This imports are explicit, to denote unwanted dependency. 
import java.awt.event.MouseEvent;
import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This class implement WINDOW widget drawing on GUI interfaces.
 */
public class WindowGuiImpl
extends Window<GuiOutputManager> 
implements GuiWindow
{
   /** The default title for GUI windows. */
   public static String DEFAULT_TITLE = "FWD GUI Client";
   
   /** Minimal window width. */
   private static final int MIN_WIDTH_PIXELS = 104;
   
   /** Minimal window height. */
   private static final int MIN_HEIGHT_PIXELS = 100;
   
   /** The title-bar widget for this window. */
   private WindowTitleBar titleBar;

   /** The window's workspace, where all frame drawing is done. */
   WindowWorkSpace workspace;
   
   /** Screen driver */
   GuiDriver gd;

   /** Flag indicating that window resizing is in progress. */
   private boolean resizing;
   
   /** Flag indicating if the resize is for the workspace or for the full window. */
   private boolean resizeWorkspace = false;
   
   /** The physical window width, after resize. */
   private int resizedWidth;
   
   /** The physical window height, after resize. */
   private int resizedHeight;

   /** The last mouse pressed determined for this window. */
   private Widget<GuiOutputManager> mousePressedSource = null;
   
   /** Flag indicating if the window is currently maximized. */
   private boolean maximized = false;
   
   /** The dimension before the window maximization. */
   private Dimension restoreDimension = null;

   /**
    * Default constructor.
    * 
    * @param    id
    *           The widget's ID.
    */
   public WindowGuiImpl(WidgetId id)
   {
      super(id);
      
      // default window colors
      this.setDefaultColors();

      gd = (GuiDriver) OutputManager.getDriver();
      
      this.setLayout(new WindowLayout());

      registerWidgetStateListener(w ->
      {
         FontManager.deregisterFontTable(WindowGuiImpl.this, OutputManager.getDriver());
         WindowManager.removeNoDestroy(WindowGuiImpl.this);
         gd.deregisterWindow(w.getId().asInt());
      });
   }

   /**
    * Construct a window widget based on a specific configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   config
    *          Window configuration.
    */
   public WindowGuiImpl(WidgetId id, WindowConfig config)
   {
      super(id, config);
      
      // default window colors
      this.setDefaultColors();

      gd = (GuiDriver) OutputManager.getDriver();

      this.setLayout(new WindowLayout());

      registerWidgetStateListener(w ->
      {
         FontManager.deregisterFontTable(WindowGuiImpl.this, OutputManager.getDriver());
         WindowManager.removeNoDestroy(WindowGuiImpl.this);
         gd.deregisterWindow(w.getId().asInt());
      });
   }
   
   /**
    * Determine if this window can have decorations.
    * 
    * @return   <code>true</code> if this window supports decorations with the current driver.
    * 
    * @see   GuiDriver#isWindowDecorationSupported()
    */
   @Override
   public boolean hasDecorations()
   {
      return gd.isWindowDecorationSupported();
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WindowConfig cfg)
   {
      // when window is creating at runtime it should register itself with renderer prior to
      // the base init routines
      OutputManager.instance().realizeWindow(id.asInt());
      // cal for base initialization
      super.initialize(id, cfg);
      
      updateFullSizeAttributes();
   }

   /**
    * Hide this window.
    */
   @Override
   public void hide()
   {
      if (!isVisible())
      {
         return;
      }
      
      super.hide();
      
      GuiWindowCommons.hide(this);
   }
   
   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (isVisible() || gd.isIconified(UiUtils.getWidgetIdAsInt(this)))
      {
         return;
      }

      boolean wasRealized = config.realized;

      super.show();

      // realize the window
      config.realized = true;

      // show my parent and all my children
      WindowGuiImpl parent = WindowManager.findWindow(this.config.parentId);
      if (parent != null)
      {
         parent.show();

         // ensure the driver is notified - as only the root parent is actually painted by P2J
         // event drawing bracket
         gd.selectWindow(parent.config().id.asInt());
         gd.setWindowVisible(true);
         gd.releaseWindow();
      }
      
      List<TitledWindow<?>> children = WindowManager.findChildren(this.config.id.asInt(), false);
      for (TitledWindow<?> w : children)
      {
         w.show();

         // ensure the driver is notified - as only the root parent is actually painted by P2J
         // event drawing bracket
         gd.selectWindow(w.config().id.asInt());
         gd.setWindowVisible(true);
         gd.releaseWindow();
      }
      
      List<DialogBoxWindow> dialogs = WindowManager.getDialogsOwnedBy(getId().asInt(), true);
      DialogBoxWindow topDialog = null;
      
      for (DialogBoxWindow d : dialogs)
      {
         FrameGuiImpl frame = d.getFrame();
         if (frame.config().assignedVisible && d.isRealized())
         {
            gd.selectWindow(d.getId().asInt());
            d._setVisible(true);
            gd.setWindowVisible(true);
            gd.releaseWindow();
            
            // set the visibility flag on the frame so that it is reflected on the server
            d.getFrame()._setVisible(true);
            
            topDialog = d;
         }
      }
      
      if (topDialog != null)
      {
         WindowManager.modalizeWindow(topDialog);
      }
      
      // show us now
      gd.selectWindow(config.id.asInt());
      gd.setWindowVisible(true);
      gd.releaseWindow();
      
      // if the OS window is made visible (from hidden), it may be moved to top... 
      // need to preserve the window order!
      WindowManager.restackWindows();
      
      // activate the window when it has just been realized and it is set as the CURRENT-WINDOW
      if (!wasRealized && WindowManager.getCurrentWindow() == (Window) this)
      {
         WindowManager.activateWindow(this);
      }
   }
   
   /**
    * Hide/show widget.
    * <p>
    * 
    * If the window is minimized, it can not be made visible using VIEW or setting VISIBLE 
    * attribute to true.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (gd.isIconified(UiUtils.getWidgetIdAsInt(this)))
      {
         visible = false;
      }
      
      super.setVisible(visible);
   }
   
   /**
    * Get widgets array from first available window.
    * 
    * @return  array of the widgets belonging to the resolved windows.
    */
   @Override
   public Widget<GuiOutputManager>[] getWidgets()
   {
      return workspace == null ? new Widget[0] : workspace.getContentPane().widgets();
   }
   
   /**
    * Detach this widget from the window's workspace.  The widget will not be destroyed.
    * 
    * @param   widget
    *          Widget to detach.
    *          
    * @throws  RuntimeException
    *          when the supplied widget is not attached to this container
    */
   @Override
   public void detach(Widget<GuiOutputManager> widget)
   {
      // the contentPane is attached to the window, not to directly to the workspace... when
      // detaching it, use the super impl
      if (widget == contentPane)
      {
         super.detach(widget);
      }
      else
      {
         workspace.getContentPane().detach(widget);
      }
   }
   
   /**
    * Move specified widget to the top of widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToTop(Widget<GuiOutputManager> widget)
   {
      workspace.getContentPane().moveToTop(widget);
   }
   
   /**
    * Move specified widget to the bottom of widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToBottom(Widget<GuiOutputManager> widget)
   {
      workspace.getContentPane().moveToBottom(widget);
   }
   
   /**
    * Move specified widget before the widget "before".
    * Widgets must be in the same field-group, that is checked on the server-side.
    * 
    * @param   widget
    *          Widget to move.
    * @param   before
    *          The widget before to insert mover.
    */
   @Override
   public void moveBeforeTabItem(Widget<GuiOutputManager> widget, Widget<GuiOutputManager> before)
   {
      workspace.getContentPane().moveBeforeTabItem(widget, before);
   }

   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      boolean placeWindow = !(isRunning || gd == null);
      init();

      if (!isRunning)
      {
         return;
      }
      
      super.doLayout();
      
      if (WindowManager.<GuiOutputManager>getDefaultWindow() == this)
      {
         // can be missing if there are no decorations
         if (hasTitlebar())
         {
            // if this is the default window, set the deltas
            WindowManager.setPlacementDeltas(titleBar.icon.physicalBounds().right(),
                                             titleBar.physicalBounds().bottom());
         }
      }

      if (placeWindow)
      {
         // set the window position
         WindowManager.placeWindow(this);
      }
   }
   
   /**
    * Draw window
    */
   @Override
   public void draw()
   {
      init();

      // Select renderer
      gd.selectWindow(config().id.asInt());
      
      try
      {
         gd.startBatch();
         
         if (!hasDecorations())
         {
            GuiColorResolver gc = workspace.getColorResolver();
            gc.colors();
            gd.setDesktopBgColor(gc.bgColor);
         }

         gd.setDrawingArea(this.clipRectangle());
         
         super.draw();
      }
      finally
      {
         gd.setDrawingArea(null);
         gd.endBatch();
         gd.releaseWindow();
      }
   }

   /**
    * Get dimension of screen area available for the placement of frames.
    * 
    * @return  Instance of <code>Dimension</code> which describes size of the
    *          available screen area.
    */
   @Override
   public Dimension getScreenDimension()
   {
      Dimension sz = dimension();
      return new Dimension(sz.width, sz.height - RESERVED_LINES);
   }

   /**
    * Get the message area widget associated with this window.
    * 
    * @return   See above.
    */
   @Override
   public MessageAreaGuiImpl getGuiMessageArea()
   {
      return (MessageAreaGuiImpl) messageArea;
   }
   
   /**
    * Get the workspace area widget associated with this window.
    * 
    * @return   See above.
    */
   public WindowWorkSpace getWorkspace()
   {
      return workspace;
   }
   
   /**
    * Get messages area rectangle by message line index.
    *
    * @param   line
    *          message line specification: either 0-based line number for
    *          specific line or -1 for the whole message area.
    * @return  Instance of <code>Rectangle</code> which describes the message
    *          area in screen coordinates.
    */
   public com.goldencode.p2j.ui.client.Rectangle getMessageLineRectangle(int line)
   {
      return getGuiMessageArea().bounds();
   }

   /**
    * Get status area rectangle.
    *
    * @return  Instance of <code>Rectangle</code> which describes the status
    *          area in screen coordinates.
    */
   public com.goldencode.p2j.ui.client.Rectangle getStatusLineRectangle()
   {
      return statusLine.bounds();
   }
   
   /**
    * Returns the preferred size of the window workspace.   
    * 
    * @return   See above.
    */
   @Override
   public Dimension getScrollableSize()
   {
      // if window virtual width/height has not been set, use the workspace dimensions
      Dimension d = this.workspace.getScrollableSize();
      return new Dimension(Math.max(d.width, config.virtualWidthChars), 
                           Math.max(d.height, config.virtualHeightChars));
   }

   /**
    * Calculates the effective size of the scrollable area of the window workspace.
    * <p>
    * This implementation accounts all the visible child widgets into the area and
    * simply determines the furthest x and y from the origin of the window workspace.
    * 
    * @return  see above.
    */
   @Override
   public Dimension getEffectiveScrollableSize()
   {
      return workspace.getEffectiveScrollableSize();
   }

   /**
    * Adds the frame to the window at the specified position.
    * 
    * @param    widget
    *           The handle of the frame to be added to the window.
    * @param    row
    *           The row number for the upper-left frame corner.
    * @param    column
    *           The column number for the upper-left frame corner.
    */
   public void addFrame(Widget<GuiOutputManager> widget, double row, double column)
   {
      row = Coordinate.scale(row);
      column = Coordinate.scale(column);
      
      if (widget == null && row == 0 && column == 0)
         return;
      
      workspace.getContentPane().add(widget);
   }
   
   /**
    * Returns the frames contained in this top-level window. If no frames
    * then the result is an empty list.
    * 
    * @return  See above.
    */
   @Override
   public List<Frame<GuiOutputManager>> getFrames()
   {
      if (workspace == null)
      {
         return Collections.emptyList();
      }
      
      Container<GuiOutputManager> content = workspace.getContentPane();
      ArrayList<Frame<GuiOutputManager>> result = new ArrayList<>(content.widgets().length);
      
      for (Widget<GuiOutputManager> w : content.widgets())
      {
         if (w instanceof Frame)
         {
            result.add((Frame<GuiOutputManager>) w);
         }
      }
      
      return result;
   }
   
   /**
    * Find the widget to which the given mouse event can be applied.
    * 
    * @param    evt
    *           The mouse event.
    * 
    * @return   A widget over which the mouse is positioned or {@code null} if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(MouseEvt evt)
   {
      Widget<GuiOutputManager> widget = null;
      
      int evtID = evt.getEvent().getID();
      
      if (evtID == MouseEvent.MOUSE_DRAGGED  || 
          evtID == MouseEvent.MOUSE_RELEASED || 
          evtID == MouseEvent.MOUSE_CLICKED)
      {
         return mousePressedSource;
      }
      
      try
      {
         java.awt.Point p = evt.getEvent().getPoint();
         
         // search the other widgets
         widget = findMouseSource(new NativePoint(p.x, p.y));
         
         widget = (widget == null ? this : widget);
      }
      finally
      {
         if (widget != null && evtID == MouseEvent.MOUSE_PRESSED)
         {
            mousePressedSource = widget;
         }
         
         if (evtID == MouseEvent.MOUSE_CLICKED)
         {
            mousePressedSource = null;
         }
      }
      
      return widget;
   }
   
   /**
    * Iconify the window.
    */
   @Override
   public void iconify()
   {
      iconify(config().id.asInt());
   }

   /**
    * Iconify the window.
    *
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   @Override
   public void iconify(int sourceId)
   {
      iconify(sourceId, true);
   }

   /**
    * Iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    * @param    postEvent
    *           When <true>true</true> the method posts the WINDOW-MINIMIZED legacy event,
    *           <code>false</code> means no event is posted.
    */
   @Override
   public void iconify(int sourceId, boolean postEvent)
   {
      WindowConfig cfg = config();
      cfg.windowState = CommonWindow.WINDOW_STATE_MINIMIZED;

      gd.selectWindow(cfg.id.asInt());

      // the window-minimize event is raised by the OS listener

      gd.iconifyWindow(sourceId, postEvent);
      
      gd.releaseWindow();
   }
   
   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    */
   @Override
   public void restore()
   {
      restore(-1);
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    *
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    */
   @Override
   public void restore(int sourceId)
   {
      restore(sourceId, true);
   }

   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    * @param    postEvent
    *           When <true>true</true> the method posts the WINDOW-RESTORED legacy event,
    *           <code>false</code> means no event is posted.
    */
   @Override
   public void restore(int sourceId, boolean postEvent)
   {
      if (!(maximized || gd.isIconified(UiUtils.getWidgetIdAsInt(this))))
      {
         return;
      }

      if (maximized)
      {
         maximized = false;

         // restore to the previous width/height
         CoordinatesConversion c = screen().coordinates();
         int nwidth = c.widthToNative(restoreDimension.width);
         int nheight = c.heightToNative(restoreDimension.height);
         
         resizeTo(null, new NativeDimension(nwidth, nheight), false);

         if (postEvent)
         {
            // raise the window-restore event
            ThinClient.getInstance().eventBracket(() ->
            {
               Event evt = new WindowEvent(WindowGuiImpl.this, Keyboard.SE_WINDOW_RESTORED);
               EventManager.postOSEvent(evt);
            });
         }
      }
      else
      {
         gd.selectWindow(config().id.asInt());
         gd.deiconifyWindow(sourceId);
         gd.releaseWindow();
      }

      config.windowState = (byte) (maximized ? CommonWindow.WINDOW_STATE_MAXIMIZED:
                                               CommonWindow.WINDOW_STATE_NORMAL);
   }

   /**
    * Maximize this window.
    */
   @Override
   public void maximize()
   {
      maximize(true);
   }

   /**
    * Maximize this window.
    *
    * @param    postEvent
    *           When <true>true</true> the method posts the WINDOW-MAXIMIZED legacy event,
    *           <code>false</code> means no event is posted.
    */
   @Override
   public void maximize(boolean postEvent)
   {
      if (maximized)
      {
         restore();

         return;
      }
      
      gd.selectWindow(config().id.asInt());
      
      Dimension d = this.dimension();
      this.restoreDimension = d;
      
      WindowConfig cfg = config();
      cfg.windowState = CommonWindow.WINDOW_STATE_MAXIMIZED;
      
      double width = ConfigHelper.getMaxWidthChars(cfg);
      double height = ConfigHelper.getMaxHeightChars(cfg);
      
      CoordinatesConversion c = screen().coordinates();

      if (postEvent)
      {
         // raise the window-maximize event first; RETURN NO-APPLY is ignored
         ThinClient.getInstance().eventBracket(() ->
         {
            Event evt = new WindowEvent(WindowGuiImpl.this, Keyboard.SE_WINDOW_MAXIMIZED);
            EventManager.postOSEvent(evt);
         });
      }
      
      // resize, but only if the dimension differs
      if (Double.compare(width, d.width) != 0 || Double.compare(height, d.height) != 0)
      {
         int nwidth = c.widthToNative(width);
         int nheight = c.heightToNative(height);
         
         resizeTo(null, new NativeDimension(nwidth, nheight), false);
      }
      
      this.maximized = true;

      repaint();
      
      gd.releaseWindow();
   }

   /**
    * Resize the widget to the specified dimensions.
    * 
    * @param    origin
    *           The top-left window corner new position.
    * @param    size
    *           The new window size.
    */
   @Override
   public void resizeTo(NativePoint origin, NativeDimension size)
   {
      resizeTo(origin, size, true);
   }
   
   /**
    * Resize the widget to the specified dimensions.
    * 
    * @param    origin
    *           The top-left window corner new position.
    * @param    size
    *           The new window size.
    * @param    raise
    *           Flag indicating if the 4GL WINDOW-RESIZED event needs to be raised.
    */
   @Override
   public void resizeTo(NativePoint origin, NativeDimension size, boolean raise)
   {
      if (maximized)
      {
         return;
      }
      
      WindowConfig c = config();
      
      gd.selectWindow(c.id.asInt());
      
      resizing = true;
      
      resizedWidth = size.width;
      resizedHeight = size.height;
      
      try
      {
         doLayout();
         
         // do not use the size parameter but instead fetch the size from the laid out window,
         // the size may be different (max/min size limits for example)
         NativeDimension dim = physicalDimension();
         
         if (origin != null)
         {
            int dx = resizedWidth - dim.width;
            int dy = resizedHeight - dim.height;
            
            gd.setWindowBounds(origin.x + dx, origin.y + dy, dim.width, dim.height);
         }
         else
         {
            gd.resizeWindow(dim.width, dim.height);
         }
      }
      finally
      {
         this.resizing = false;
         gd.releaseWindow();
      }

      repaint();
      
      if (raise)
      {
         // raise the window-resized event; RETURN NO-APPLY is ignored.
         ThinClient.getInstance().eventBracket(() ->
         {
            // TODO: possible lock in SWING!
            Event evt = new WindowEvent(WindowGuiImpl.this, Keyboard.SE_WINDOW_RESIZED);
            EventManager.postOSEvent(evt);
         });
      }
   }
      
   /**
    * Shows the new message text.
    * <p>
    * Colors are ignored in Windows (GUI)
    * <p>
    * This is a no-op when this window doesn't have the message area displayed.
    * 
    * @param   message
    *          Array of two text message which need to be displayed inside the
    *          message area.
    * @param   cs
    *          Array of two color specifications.
    */
   @Override
   public void setMessageText(String[] message, Color[] cs)
   {
      setMessageText(message, cs, false);
   }

   /**
    * Set message main worker routine.
    * <p>
    * Colors are ignored in Windows (GUI).
    * <p>
    * This is a no-op when this window doesn't have the message area displayed.
    * 
    * @param   msg
    *          Message text array.
    * @param   cs
    *          Message color array.
    * @param   force
    *          <code>true</code> means force redisplaying of messages even
    *          if message text is already <code>null</code>.
    */
   @Override
   public void setMessageText(String[] msg, Color[] cs, boolean force)
   {
      MessageAreaGuiImpl guiMessageArea = getGuiMessageArea();
      if (config.messageArea && guiMessageArea != null)
      {
         guiMessageArea.setMessageText(msg, cs, force);
      }
   }
   
   /**
    * Renders the specified message into the message line.
    *
    * @param   message
    *          The message body to be rendered.
    * @param   cs
    *          color specification.
    * @param   clear
    *          If <code>true</code> then the line(s) on which the message will
    *          be displayed, will be cleared first.
    *
    * @return  Message line number used to display message.
    */
   @Override
   public int message(String message, Color cs, boolean clear)
   {
      ThinClient tc = ThinClient.getInstance();
      
      // only show default window
      if (!isVisible() && getId().equals(WidgetId.DEFAULT_WINDOW_ID))
      {
         ThinClient.getInstance().independentEventDrawingBracket(this, this::show);
      }
      
      if (!isVisible() || !config.messageArea)
      {
         // do not show the non-default, instead display alert-box
         tc.messageBox(new Object[] { message }, 
                       null, 
                       LogicalTerminal.ALERT_MESSAGE, 
                       LogicalTerminal.BTN_OK, 
                       "Message", 
                       cs, 
                       getId().asInt(),
                       null /* TODO: is it always correct? */);
         return 0;
      }
   
      displayMessage(new String[] { message }, cs, -1);

      if (tc.isRedirected())
      {
         tc.sendToStream(message);
      }
      
      return 0;
   }   

   /**
    * Destroy this window.  Notify all listeners about the widget being destroyed.
    */
   @Override
   public void destroy()
   {
      TopLevelWindow<?> wnd = WindowManager.getFocusWindow();

      int wid = config().id.asInt();
      
      gd.selectWindow(wid);
      
      for (Widget<GuiOutputManager> w : widgets())
      {
         WidgetId id = w.getId();
         if (id != null)
         {
            gd.deregisterWidget(id.asInt());
         }
      }
      
      gd.deregisterWidget(wid);
      
      hide();

      super.destroy();
      
      gd.releaseWindow();
      
      if (wnd == this)
      {
         WindowManager.setFocusWindow(WindowManager.getActiveWindow());
      }
   }

   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   event
    *          Event to process.
    *          
    * @throws  IllegalStateException
    *          If window MENUBAR contains wrong widget.
    */
   @Override
   public void processEvent(Event event)
   {
      if (event instanceof KeyInput && ((KeyInput) event).keyCode() == Key.VK_ALT)
      {
         if (menubar != null)
         {
            ThinClient.getInstance().independentEventDrawingBracket(menubar, () ->
            {
               MenuGuiImpl menu = (MenuGuiImpl) menubar;
               
               boolean prevMode = menu.isKeyMode();
               menu.setKeyMode(!prevMode);
               
               if (prevMode)
               {
                  // losing focus
                  Widget<?> focus = menu.focus();
                  if (focus != null)
                  {
                     focus.mouseExited(null);
                     MenuGuiImpl.resetMbarPressedState();
                     menu.setFocusInt(null);
                  }
               }
               else
               {
                  // Set pressed flag to the first MENUBAR widget event if it is disabled.
                  Widget<?>[] widgets = menubar.widgets();
                  if (widgets.length > 0)
                  {
                     MenuGuiImpl.setMbarPressedState((MenuElement<GuiOutputManager>) widgets[0], 
                                                     true);
                     widgets[0].requestFocus();
                  }
               }
               
               menu.repaint();
            });
            
            ((KeyInput) event).consume();
         }
         
         return;
      }

      // update window state when the state change triggered by OS
      switch(event.id())
      {
         case Keyboard.SE_WINDOW_RESTORED:
            config.windowState = (byte) (maximized ? CommonWindow.WINDOW_STATE_MAXIMIZED:
            CommonWindow.WINDOW_STATE_NORMAL);
            break;
         case Keyboard.SE_WINDOW_MINIMIZED:
            config.windowState = CommonWindow.WINDOW_STATE_MINIMIZED;
            break;
         case Keyboard.SE_WINDOW_MAXIMIZED:
            config.windowState = CommonWindow.WINDOW_STATE_MAXIMIZED;
            break;
      }

      // filter the default button related events
      if (event instanceof KeyInput && !(event.source() instanceof Frame))
      {
         Widget<?> src = event.source();
         KeyInput keyEvt = (KeyInput) event; 
         Frame<?> frame = UiUtils.locateFrame(src);
         if (frame != null)
         {
            if ((keyEvt.keyCode() == Key.VK_ESCAPE && frame.processCancelButton()) ||
                (keyEvt.keyCode() == Key.VK_ENTER && frame.processDefaultButton(keyEvt)))
            {
               return;
            }
         }
         // process accelerators event
         else if (menubar != null)
         {
            menubar.processKeyEvent(keyEvt);
            if (keyEvt.isConsumed())
            {
               return;
            }
         }
      }
      else if (event.id() == Keyboard.SE_WINDOW_MINIMIZED &&
               gd.isIconified(UiUtils.getWidgetIdAsInt(this)))
      {
         WindowEvent wevt = (WindowEvent) event;
         
         if (wevt.getOrigin() != -1)
         {
            // if this is not the window who initiated the event cascade, check if there are chained
            // events. if so, raise the next one
            if (wevt.getNextEvent() != null)
            {
               ThinClient.getInstance().postOSEvent(wevt.getNextEvent());
            }
            
            return;
         }

         // the WINDOW-MINIMIZED event must be raised ALWAYS after the OS event for this window
         // that's why we enforce the window to be already iconified here
         
         final List<TitledWindow<?>> children = 
                                        WindowManager.findChildren(this.config.id.asInt(), false);
         final int sourceId = UiUtils.getWidgetIdAsInt(this);

         // the windows which require 4GL events
         final List<TitledWindow<?>> processed = new ArrayList<>();

         // iconify all children hierarchy, only for the event originating from the OS
         
         // no OS events are raised here - these are raised AFTER the window sub-tree is 
         // iconified, using the same rules
         try
         {
            EventManager.pushOSEventQueue();

            for (TitledWindow<?> w : children)
            {
               WindowGuiImpl wgui = (WindowGuiImpl) w;
   
               if (gd.isIconified(UiUtils.getWidgetIdAsInt(wgui)))
               {
                  // skip already iconified windows
                  
                  // but ensure the window is hidden
                  if (wgui.config().iconifiedBy == -1)
                  {
                     wgui.hide();
                     
                     // do not collect it
                  }
   
                  continue;
               }
   
               wgui.iconify(sourceId);
               
               // hide all the children now, regardless if the child's trigger is consumed or not
               wgui.hide();
   
               // save this
               wgui.config().iconifiedBy = sourceId;

               // collect it
               processed.add(wgui);
            }
         }
         finally
         {
            EventManager.popOSEventQueue();
         }

         // create 4GL events for the direct children and chain them
         chainEvents(processed, sourceId, Keyboard.SE_WINDOW_MINIMIZED);
      }
      else if (event.id() == Keyboard.SE_WINDOW_RESTORED &&
               !gd.isIconified(UiUtils.getWidgetIdAsInt(this)))
      {
         WindowEvent wevt = (WindowEvent) event;

         if (wevt.getOrigin() != -1)
         {
            // if this is not the window who initiated the event cascade, check if there are chained
            // events. if so, raise the next one
            if (wevt.getNextEvent() != null)
            {
               ThinClient.getInstance().postOSEvent(wevt.getNextEvent());
            }
            
            return;
         }

         // the WINDOW-RESTORED event (from de-iconification) must be raised always after the OS
         // deiconify event for this window. 
         // that's why we enforce the window to be already de-iconified here

         final List<TitledWindow<?>> children = 
                                        WindowManager.findChildren(this.config.id.asInt(), true);
         final int sourceId = UiUtils.getWidgetIdAsInt(this);

         // the windows which require 4GL events
         final List<TitledWindow<?>> processed = new ArrayList<>();
         
         // show all children hierarchy, but only windows which were iconified by the current
         // one
         
         // only the window with the root event (originating from OS) restores children
         
         // no OS events are raised here - these are raised AFTER the window sub-tree is 
         // deiconified, using the same rules
         try
         {
            EventManager.pushOSEventQueue();

            for (TitledWindow<?> w : children)
            {
               WindowGuiImpl wgui = (WindowGuiImpl) w;
               if (!gd.isIconified(UiUtils.getWidgetIdAsInt(wgui)))
               {
                  // skip already de-iconified windows
                  continue;
               }
   
               if (wgui.config().iconifiedBy == -1)
               {
                  // if they are iconified via OS, show it, do not de-iconify it, but raise a
                  // WINDOW-RESTORED event. this is a 4GL quirk... the WINDOW-RESTORED event will
                  // be raised again, when the window is clicked again and restored
                  
                  // show us now
                  gd.selectWindow(wgui.config.id.asInt());
                  gd.setWindowVisible(true);
                  gd.releaseWindow();
                  
                  // collect it
                  processed.add(wgui);
                  continue;
               }
   
               if (wgui.config().iconifiedBy != sourceId)
               {
                  // skip windows which were not iconified by this one
                  continue;
               }
               
               // show all the children now, regardless if the child's trigger is consumed or not
               wgui.restore(sourceId);
               wgui.show();
               
               // reset this
               wgui.config().iconifiedBy = -1;

               // collect it
               processed.add(wgui);
            }
         }
         finally
         {
            EventManager.popOSEventQueue();
         }

         chainEvents(processed, sourceId, Keyboard.SE_WINDOW_RESTORED);
      }
      
      if (event instanceof WindowActivated && menubar != null)
      {
         if (!((WindowActivated) event).isActivated())
         {
            MenuGuiImpl.resetMbarPressedState();
            ((MenuGuiImpl) menubar).setKeyMode(false);
            menubar.repaint();
         }
      }
      
      super.processEvent(event);
   }

   /**
    * Sets the new application icon for given window on the client side.
    * 
    * @param    newIcon
    *           A new icon image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean setWindowIcon(byte[] newIcon)
   {
      // can be missing if there are no decorations
      if (hasTitlebar())
      {
         // this is GUI version
         return titleBar.icon.setIconData(newIcon);
      }
      
      return true;
   }
   
   /**
    * Loads the new application icon for given window from the client side filesystem.
    * 
    * @param    small
    *           Flag indicating usage of the small icon case(16x16 in size).
    * @param    name
    *           The fully qualified file name for the icon to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean loadWindowIcon(boolean small, String name)
   {
      // can be missing if there are no decorations
      if (hasTitlebar())
      {
         // this is GUI version
         return titleBar.icon.loadIcon(small, name);
      }
      
      return true;
   }
   
   /**
    * The method is called after the configuration associated with the implementor changes. The
    * parameter points to a reference representing the original configuration state. This
    * reference can be used to detect what configuration fields changed and so optimize any
    * processing related to the configuration change.
    * 
    * @param   beforeUpdate
    *          Config reference capturing the config state before the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(WindowConfig beforeUpdate)
   {
      if (beforeUpdate.enabled != isEnabled())
      {
         // push the window:sensitive state to the driver
         gd.selectWindow(getId().asInt());
         gd.setWindowEnabled(config.enabled);
         gd.releaseWindow();
      }

      super.afterConfigUpdate(beforeUpdate);
      
      boolean doLayout = false;
      
      if (!isRunning)
      {
         init();
         doLayout = true;
      }
      else if (config.windowState != beforeUpdate.windowState)
      {
         switch (config.windowState)
         {
            case CommonWindow.WINDOW_STATE_MAXIMIZED:
               maximize(false);
               doLayout = true;
               break;
            case CommonWindow.WINDOW_STATE_MINIMIZED:
               iconify(config.id.asInt(), false);
               break;
            case CommonWindow.WINDOW_STATE_NORMAL:
               restore(config.id.asInt(), false);
               doLayout = true;
               break;
         }
      }

      NativePoint loc = physicalLocation();
      NativePoint newLoc = ConfigHelper.getAssignedPixelLocation(config);
      
      if (newLoc.x == BaseConfig.INV_COORD)
      {
         // x not assigned? use the original location 
         newLoc.x = loc.x;
      }
      
      if (newLoc.y == BaseConfig.INV_COORD)
      {
         // y not assigned? use the original location
         newLoc.y = loc.y;
      }
      
      if (loc.x != newLoc.x || loc.y != newLoc.y)
      {
         setPhysicalLocation(newLoc.x, newLoc.y);
      }

      NativeDimension dim = workspace.physicalDimension();
      
      if (dim.width != config.widthPixels || dim.height != config.heightPixels)
      {
         this.resizeWorkspace = true;
         try
         {
            resizeTo(null, new NativeDimension(config.widthPixels, config.heightPixels), false);
         }
         finally
         {
            this.resizeWorkspace = false;
         }
      }

      int wid = config.id.asInt();

      if (config != null && gd != null && config.title != null)
      {
         gd.selectWindow(wid);
         gd.setTitle(config.title);
         setTitle(config.title);
         gd.releaseWindow();
      }
      
      if (beforeUpdate.minHeightPixels != config.minHeightPixels ||
          beforeUpdate.minWidthPixels  != config.minWidthPixels  ||
          beforeUpdate.maxHeightPixels != config.maxHeightPixels ||
          beforeUpdate.maxWidthPixels  != config.maxWidthPixels  ||
          beforeUpdate.resize != config.resize)
      {
         gd.selectWindow(wid);
         gd.setResizeableWindow(config.resize, 
                                config.minWidthPixels, 
                                config.minHeightPixels, 
                                config.maxWidthPixels, 
                                config.maxHeightPixels);
         gd.releaseWindow();
      }
      
      if (!beforeUpdate.realized && 
          (beforeUpdate.messageArea != config.messageArea || 
           beforeUpdate.statusArea != config.statusArea))
      {
         messageArea.forceVisibility(config.messageArea);
         statusLine.forceVisibility(config.statusArea);
         doLayout = true;
      }
      
      if (doLayout)
      {
         doLayout();
         
         if (isVisible())
         {
            repaint();
         }
      }
      
      if (config.topOnly && config.topOnly != beforeUpdate.topOnly && 
          isRealized() && 
          isVisible())
      {
         WindowManager.moveToTop(this);
      }
   }

   /**
    * Check if the window is already maximized - it can't be resized.
    * 
    * @return   The state of the {@link #maximized} flag.
    */
   @Override
   public boolean isMaximized()
   {
      return maximized;
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Get the current custom mouse pointer if it is set
    * 
    * @return the current custom mouse pointer if it is set  
    *        
    */
   @Override
   public Optional<MousePtrWrapper> getCustomMousePointer()
   {
      return mousePtr; 
   }

   /**
    * Get the current effective custom mouse pointer
    * 
    * @return the current effective custom mouse pointer  
    *        
    */
   @Override
   public MousePtrWrapper getEffectiveMousePointer()
   {
      return ThinClient.lockedMousePointer().orElse(mousePtr.orElse(MousePtrWrapper.DEFAULT));
   }
   
   /**
    * Returns the window minimum size.
    * 
    * @return  See above.
    */
   @Override
   public NativeDimension getMinSize()
   {
      return new NativeDimension(MIN_WIDTH_PIXELS, MIN_HEIGHT_PIXELS);
   }
   
   /**
    * In GUI, the insert mode is always {@link Window#INSERT_MODE_INS}.
    * 
    * @return    Always {@link Window#INSERT_MODE_INS}.
    */
   protected int getInsMode()
   {
      return INSERT_MODE_INS;
   }

   /**
    * Set the contents of the status line. This method is private and
    * called at service level by another static public method within the
    * class.  The status line output is truncated to a maximum of 63
    * characters, only if the <code>noLengthLimit</code> flag is not set.
    * <p>
    * This method is a no-op when status area is not displayed for this window.
    * 
    * @param   status
    *          The new string to be displayed inside the status line area.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   @Override
   protected void setStatusLineText(String status, boolean noLengthLimit)
   {
      // the status area is not displayed, don't do anything
      if (!config.statusArea)
      {
         return;
      }
      
      if (status != null && status.equals(statusLine.getText()))
      {
         // the text has not changed, don't do anything
         return;
      }
      
      getContentPane().moveToTop(statusLine);
      statusLine.setText(status);
      statusLine.repaint();
   }

   /**
    * Reset Window instance state.
    */
   @Override
   protected void reset()
   {
      super.reset();
      if (hasDecorations())
      {
         titleBar = new WindowTitleBar(WidgetId.nextID(), title());
      }
      // TODO: what is the layout mgr for the GUI window (the workspace)?
      ScrollBar.Bars bars = config.scrollBars ? ScrollBar.Bars.BOTH : ScrollBar.Bars.NONE;
      workspace = new WindowWorkSpace(bars);
      
      // set the initial size - this is the WIDTH-CHARS/HEIGHT-CHARS window
      // attributes
      workspace.setSize(config.widthChars, config.heightChars);
   }
         
   /**
    * Initialize this window.  This can be done only after the driver has selected it.
    */
   protected void init()
   {
      if (isRunning || gd == null)
      {
         return;
      }
      
      super.init();
      
      if (hasDecorations())
      {
         contentPane.setBorder(WindowBorder.createWindowBorder());
      }
      
      gd.selectWindow(getId().asInt());

      reset();
      
      isRunning = true;
      
      GuiColorResolver gc = new GuiColorResolver(gd, config(), 
                                                 SysColor.COLOR_3DFACE, 
                                                 SysColor.COLOR_WINDOWTEXT);
      ((BorderedPanelGuiImpl) contentPane).setColorResolver(gc);

      add(contentPane);
      
      // initialize the window's font-table
      FontManager.registerFontTable(this, OutputManager.getDriver());
      
      // can be missing if there are no decorations
      if (hasTitlebar())
      {
         contentPane.add(titleBar);
      }
      
      // workspace
      contentPane.add(workspace);
      
      // message area with variable lines of the text
      messageArea.attach(this);
      
      MessageAreaGuiImpl messageArea = getGuiMessageArea();
      contentPane.add(messageArea);
      
      // status line with default text
      statusLine.attach(this);      
      contentPane.add(statusLine);
      
      WidgetRegistry<GuiOutputManager> reg = screen().getRegistry();
      
      // assign widget ID so that the border can be registered as resizable
      ((BorderedPanelGuiImpl) contentPane).setId(WidgetId.nextID());
      
      // add the widgets to the registry
      reg.addWidgetsRecursive(this);
      
      // TODO: do this conditionally?
      // can be missing if there are no decorations
      if (hasTitlebar())
      {
         gd.registerMoveableWidget(titleBar.getId().asInt());
         titleBar.registerWindowTitlePopup();
      }
      gd.registerResizeableWidget(contentPane.getId().asInt());
      
      gd.setWindowVisible(config.visible);
      
      gd.releaseWindow();

      MouseHoverAction mouseHoverAction = new MouseHoverAction(this, 
            getWorkspace().getId().asInt())
      {
         /**
          * On mouse exit, set the mouse cursor to {@link CursorType#DEFAULT}.
          * 
          * @param    e
          *           The mouse event.
          */
         @Override
         public synchronized void mouseExited(MouseEvent e)
         {
            ThinClient.getInstance().popHover();
         }

         @Override
         protected synchronized void setCursor(MousePtrWrapper ptr)
         {
            try
            {
               gd.selectWindow(window().getId().asInt());
               try
               {
                  gd.setCursor(ThinClient.lockedMousePointer().orElse(ptr));
               }
               finally
               {
                  gd.releaseWindow();
               }
            }
            catch (Exception e)
            {
               System.out.printf("Unexpected exception. Widget: %s\n", widget.getClass().getName());
               e.printStackTrace();
            }
         }
      };
      gd.registerHoverableWidget(mouseHoverAction);
      gd.registerHoverableWidget(new MouseHoverAction(this)
      {
         /**
          * On mouse entered, set the mouse cursor to {@link CursorType#TEXT}.
          * 
          * @param    e
          *           The mouse event.
          */
         @Override
         public synchronized void mouseEntered(MouseEvent e)
         {
            ThinClient.getInstance().pushHover(this);
         }

         /**
          * On mouse exit, set the mouse cursor to {@link CursorType#DEFAULT}.
          * 
          * @param    e
          *           The mouse event.
          */
         @Override
         public synchronized void mouseExited(MouseEvent e)
         {
            ThinClient.getInstance().popHover();
            setCursor(getEffectiveMousePointer());
         }

         /**
          * Set the custom mouse pointer
          *
          * @param ptr
          *        the custom mouse pointer
          */
        @Override
         protected synchronized void setCursor(MousePtrWrapper ptr)
         {
            if (ThinClient.isMousePointerLocked())
            {
               return;
            }
            try
            {
               gd.selectWindow(window().getId().asInt());
               try
               {
                  gd.setCursor(ptr);
               }
               finally
               {
                  gd.releaseWindow();
               }
            }
            catch (Exception e)
            {
               System.out.printf("Unexpected exception. Widget: %s\n", widget.getClass().getName());
               e.printStackTrace();
            }
         }
        
      });
   }
      
   /**
    * The private function to set up the global insert mode value.
    * <p>
    * On GUI no INSERT indicator is used to display insert status. 
    * 
    * @param   insMode
    *          New integer value for insert mode, but is ignored. 
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   @Override
   protected void setInsMode(int insMode, boolean force)
   {  
      insertMode = INSERT_MODE_INS;
   }
   
   /**
    * Instructs the widget that its size may have changed. The widget updates
    * its state from the associated config instance. It is the responsibility
    * of the caller to call this method only after the config size is set with
    * the new values.
    * <p>
    * This implementation ignores the supplied config instance. Instead it relies
    * on {@link #size} and {@link #config()} to detect whether size has changed.
    * 
    * @param    beforeUpdate
    *           {@link BaseConfig} reference or <code>null</code>.
    */
   @Override
   protected void updateSizeWorker(WidgetConfig beforeUpdate)
   {
      truncateWindowHeightQuirk();
   }
   
   /**
    * The method implements a quirk of WINDOW widget height truncation.
    * At certain conditions, the height is silently truncated.
    * 
    * The truncation works as follows.
    * 
    * When 
    *    - MAX-HEIGHT assigned to unrealized window and 
    *    - MAX-HEIGHT never assigned to realized window - even indirectly 
    *      (i.e. through HEIGHT attr) and
    *    - MAX-HEIGHT greater than the work area truncation height (see below) and 
    *    - HEIGHT greater than the work area truncation height and 
    *    - HEIGHT less/equal MAX-HEIGHT and
    *    - condition C1 is not set
    * then
    *    - the assigned HEIGHT value is truncated to the work area truncation height and
    *    - condition C1 is set to true.
    * 
    * The work area truncation height is calculated as the display work area height 
    * less twice window border size less title bar height.
    * 
    * The method should be called every time the window's server state changes and 
    * before the window resizes and updates its layout.
    */
   private void truncateWindowHeightQuirk()
   {
      // TODO: the method has been disabled,
      //       (1) getDisplayWorkArea doesn't return correct display size,
      //           this seems to be related the fact that the method is called
      //           during window init phase.
      //       (2) The method must be called only when height is assigned on the server, 
      //           currently the method is called every time the window config is pushed to 
      //           the client. There used to be a logic on the client side that checked whether 
      //           height really changed, but that code is gone. We need something more explicit 
      //           and reliable. An option could be to move the quirk to the server, the down side
      //           is that we need to know the actual window layout and display work area size.
      if (true)
      {
         return;
      }
      
      // make sure the logic executes only once per client call and when window
      // height is assigned by server
      if (!config.heightAssigned)
      {
         return;
      }
      
      WindowTitleBar titleBar = getTitleBar();
      if (titleBar == null)
      {
         // the method can be called early during the window init phase,
         // just wait for the next call when it is setup
         return;
      }
      
      config.heightAssigned = false;
      
      // unrealizedMaxHeightPixels not equal to BaseConfig.INV_COORD means that
      // max-height has been assigned to unrealized window and max-height never assigned to 
      // realized window
      if (config.unrealizedMaxHeightPixels == BaseConfig.INV_COORD)
      {
         return;
      }
      
      GuiDriver<?,?> drv = (GuiDriver<?,?>) screen().getInstanceDriver();
      NativeRectangle area = drv.getDisplayWorkArea(getId().asInt());
      
      // this is the height value the window will be truncated to when all
      // conditions are met
      Border<?> border = getBorder();
      NativeInsets nborder = screen().coordinates().insetsToNative(border.insets());
      int truncWorkAreaHeight = area.height() - nborder.top - nborder.bottom - 
                                titleBar.physicalDimension().height;

      // max-height must be greater than trunc-work-area
      if (config.unrealizedMaxHeightPixels <= truncWorkAreaHeight)
      {
         return;
      }
      
      // height must be greater than trunc-work-area and 
      // less/equal max-height      
      if (config.heightPixels <= truncWorkAreaHeight || 
          config.heightPixels > config.unrealizedMaxHeightPixels)
      {
         return;
      }
      
      // if the truncation conditions implemented in this method
      // have been met, just unset the condition and return
      if (config.waTruncCondition)
      {
         config.waTruncCondition = false;
         return;
      }
      
      config.heightPixels = truncWorkAreaHeight;
      config.waTruncCondition = true;
   }
   
   /**
    * Resolve the font associated with the status-area.
    * 
    * @return   See above.
    */
   int resolveStatusAreaFont()
   {
      int font = config().statusAreaFont;
      
      // TODO: does it check the window's font first?
      return font == FontManager.UNDEFINED_FONT ? FontManager.DEFAULT_FONT : font;
   }

   /**
    * Resolve the font associated with the message-area.
    * 
    * @return   See above.
    */
   int resolveMessageAreaFont()
   {
      int font = config().messageAreaFont;
      
      // TODO: does it check the window's font first?
      return font == FontManager.UNDEFINED_FONT ? FontManager.DEFAULT_FONT : font;
   }

   /**
    * Determine if this window is being resized.
    * 
    * @return   The state of the {@link #resizing} flag.
    */
   @Override
   public boolean isResizing()
   {
      return resizing;
   }
   
   /**
    * Determine if this window workspace needs to be resized.
    * 
    * @return   The state of the {@link #resizeWorkspace} flag.
    */
   @Override
   public boolean isResizeWorkspace()
   {
      return resizeWorkspace;
   }
   
   /**
    * Set default window BGCOLOR and FGCOLOR. Windows are always use system colors.
    */
   private void setDefaultColors()
   {
      WindowConfig config = config();
      if (config != null)
      {
         // set default BGCOLOR
         config.sysbgcolor = SysColor.COLOR_3DFACE;
         // set default FGCOLOR
         config.sysfgcolor = SysColor.COLOR_WINDOWTEXT;
      }
   }
      
   /**
    * Chain window events of the specified type, using as source the given list of windows.
    * <p>
    * An OS event will be posted only for the first window in the list.  For the others, they
    * will be chained via the {@link WindowEvent#getNextEvent()}.
    * 
    * @param    windows
    *           The list of windows.
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of
    *           this window.
    * @param    type
    *           The type of the window event to be posted.
    */
   private void chainEvents(List<TitledWindow<?>> windows, int sourceId, int type)
   {
      WindowEvent prev = null;
      for (TitledWindow<?> w : windows)
      {
         WindowGuiImpl wgui = (WindowGuiImpl) w;

         WindowEvent evt = new WindowEvent(wgui, type);
         evt.setOrigin(sourceId);
         
         if (prev == null)
         {
            ThinClient.getInstance().postOSEvent(evt);
         }
         else
         {
            prev.setNextEvent(evt);
         }

         prev = evt;
      }
   }

   /**
    * Returns the output window's coordinates conversion service.
    * 
    * @return   a reference to the {@link CoordinatesConversion} instance. 
    */
   @Override
   public CoordinatesConversion coordinates()
   {
      return screen().coordinates();
   }

   /**
    * Check if window has title bar.  This should return true for GUI window even if the
    * title length is zero.
    * 
    * @return  <code>true</code> if window has title to repaint.
    */
   public boolean hasTitlebar()
   {
      return titleBar != null;
   }
   
   /**
    * Get the title bar widget associated witn this window.
    * 
    * @return   See above.
    */
   @Override
   public WindowTitleBar getTitleBar()
   {
      return titleBar;
   }

   /**
    * Get the size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getWorkspaceSize()
   {
      return workspace.dimension();
   }
   
   /**
    * Get the screen location of the window workspace area, in native units.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint getWorkspaceScreenLocation()
   {
      return workspace.screenPhysicalLocation();
   }
   
   /**
    * Get the actual native size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWorkspaceNativeSize()
   {
      return workspace.physicalDimension();
   }
   
   /**
    * Get the maximum size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getMaxWorkspaceSize()
   {
      // is valid virtual size assigned? prefer it over the visible size
      double w = config.virtualWidthChars > config.widthChars ? 
                 config.virtualWidthChars : config.widthChars;
      
      double h = config.virtualHeightChars > config.heightChars ? 
                 config.virtualHeightChars : config.heightChars;
      
      w = Math.max(0, w);
      h = Math.max(0, h);
      
      return new Dimension(w, h);
   }

   /**
    * Sets the native size of the window workspace area.
    * 
    * @param    width
    *           New workspace width.
    * @param    height
    *           New workspace height.
    */
   @Override
   public void setWorkspaceNativeSize(int width, int height)
   {
      // window config holds the workspace size, not the window's outer frame
      BaseConfig wwc = workspace.config();
      wwc.widthPixels = width;
      wwc.heightPixels = height;
      
      workspace.setSize(wwc.widthChars, wwc.heightChars);
      
      // the workspace size is what is reported in the window size attributes
      config.clientWidthPixels = width;
      config.clientHeightPixels = height;
   }

   /**
    * Sets the window workspace location.
    *  
    * @param    x
    *           The x location. 
    * @param    y
    *           The y location.
    */
   @Override
   public void setWorkspaceLocation(int x, int y)
   {
      workspace.setPhysicalLocation(x, y);
   }
   
   /**
    * Determine the resized window size.
    * 
    * @return   The new size.
    */
   @Override
   public NativeDimension getResizedSize()
   {
      return new NativeDimension(resizedWidth, resizedHeight);
   }

   /**
    * Get a point to represent the LAST-EVENT:X and LAST-EVENT:Y coordinates.  For mouse portable
    * events, these represent a point in the window's workspace.  If the mouse event was inside
    * the window's title, border, status or message area, then <code>null</code> is returned.
    * 
    * @param    point
    *           The point (relative to the window's top-left corner) for which the coordinates for
    *           the portable mouse events are needed.
    * @return   A point representing the coordinates of a mouse-portable event or 
    *           <code>null</code> if such an event can't be raised.
    */
   @Override
   public NativePoint getPortableMouseEventCoordinates(NativePoint point)
   {
      if (hasTitlebar() && titleBar.physicalBounds().contains(point))
      {
         return null;
      }
      
      if (GuiWindowCommons.insideBorder(this, point))
      {
         return null;
      }
      
      if (statusLine.physicalBounds().contains(point))
      {
         return null;
      }
      
      if (((MessageAreaGuiImpl) messageArea).physicalBounds().contains(point))
      {
         return null;
      }

      Viewport<GuiOutputManager> vp = workspace.viewport();
      NativePoint vpLoc = vp.screenPhysicalLocation();
      NativeDimension vpDim = vp.physicalDimension();
      NativeRectangle vpBounds = new NativeRectangle(vpLoc, vpDim);
      
      // exclude scrollbars
      if (!vpBounds.contains(point))
      {
         return null;
      }
      
      NativePoint contentLoc = workspace.viewport().getScrollWidget().physicalLocation();
      
      // the point is relative to window's workspace
      return point.minus(workspace.screenPhysicalLocation()).minus(contentLoc);
   }
   
   /**
    * Returns the window's border implementation. If <code>null</code>
    * is returned, the window has no border.
    * 
    * @return   A border implementation or <code>null</code>.
    */
   @Override
   public Border<GuiOutputManager> getBorder()
   {
      return contentPane.border();
   }
   
   /**
    * Destroy all the widgets in this window, except frames and menus - these will be unrealized
    * and detached from the window.
    */
   @Override
   protected void destroyWidgets()
   {
      // when a GUI WINDOW widget gets deleted, the containing frames are detached and unrealized, 
      // not destroyed
      
      Predicate<Widget<GuiOutputManager>> test = (w -> 
      {
         if (w instanceof FrameGuiImpl || 
             ((w instanceof MenuGuiImpl) && !((MenuGuiImpl) w).config().popupOnly))
         {
            // detach the menu/frame from the window, but do not destroy
            detach(w);
            
            if (w instanceof FrameGuiImpl)
            {
               Window defw = WindowManager.getDefaultWindow();
               defw.addFrame(w, w.location().y, w.location().x);
            }

            // unrealize the frame/menu (recursively)
            w.unrealize();
            
            return false;
         }
         
         return true;
      });
      
      screen().getRegistry().removeWidgetsRecursive(this, test);
   }
   
   /**
    * Update {@link WindowConfig#fullWidthChars}, {@link WindowConfig#fullHeightChars},
    * {@link WindowConfig#fullWidthPixels} and {@link WindowConfig#fullHeightPixels}.
    * This method should be called every time the window
    * layout is recalculated.
    * <p>
    * In GUI the window's FULL* attributes are initialized to the current display work 
    * area contrary to 4GL documentation. It says (for FULL-HEIGHT-CHARS): 
    * "The maximum internal height of a window is the height of the screen display minus 
    * the vertical spacing required to display the border, title bar, menu bar, message 
    * area, and status area of the window."  
    */
   private void updateFullSizeAttributes()
   {
      NativeRectangle nrect = gd.getDisplayWorkArea(getId().asInt());
      Rectangle rect = screen().coordinates().rectangleFromNative(nrect);
      
      Point base = screen().coordinates().baseUnits();
      
      WindowConfig cfg = config();
      cfg.fullWidthChars = rect.width(base);
      cfg.fullHeightChars = rect.height(base); 
      cfg.fullWidthPixels = nrect.width();
      cfg.fullHeightPixels = nrect.height();
   }
}