/*
** Module   : BufferedImageDrawHelper.java
** Abstract : Implements the helper function for images processing.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 SBI 20150810 First version which implements the helper function for image processing
**                  as a part of the common functionality used by swing and web clients.
** 002 SBI 20151011 Added new method to calculate the transformed image bounds to support images
**                  caching.
** 003 EVL 20151106 Adding method to convert regular image to disabled grey.  Fix for filter to
**                  convert 3d colors on the fly.
** 004 HC  20160922 Fixed an unhandled exception.
** 005 HC  20161018 Fixed abends in image processing and drawing when image offset was involved.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// the AWT usage is intentional for this class, it is still common code
import java.awt.*;
import java.awt.image.*;

import com.goldencode.p2j.ui.client.*;

/**
 * BufferedImageDrawHelper implements the common functionality used by swing and web clients
 * to transform images.
 */
public class BufferedImageDrawHelper implements ClientImageDrawHelper<BufferedImage>
{
   /**
    * Do the common transformations on the target image.
    * 
    * @param    struct
    *           The holder for images parameters.
    * 
    * @return The wrapped image with its width and its height.
    */
   @Override
   public ImageWrapper<BufferedImage> processImage(EmbeddedImageStructure<BufferedImage> struct)
   {
      // calculate the image size
      Integer[] size = calculateImageBounds(struct);
      int width = size[0];
      int height = size[1];

      BufferedImage img2draw = struct.getWrappedImage();

      int imgWidth = img2draw.getWidth();
      int imgHeight = img2draw.getHeight();

      int xOffset = struct.getXOffset();
      int yOffset = struct.getYOffset();

      // cut the actual image to fit to the calculated size
      int cutWidth = Math.min(width, imgWidth - xOffset);
      int cutHeight = Math.min(height, imgHeight - yOffset);

      if (struct.isStretchToFit())
      {
         cutWidth = imgWidth - xOffset;
         cutHeight = imgHeight - yOffset;
      }

      if (cutWidth != imgWidth || cutHeight != imgHeight)
      {
         img2draw = img2draw.getSubimage(xOffset, yOffset, cutWidth, cutHeight);
      }

      // 3D color conversions
      if (struct.is3D())
      {
         img2draw = get3DColorsConvertedImage(img2draw);
      }

      // transparency
      if (struct.isTransparent())
      {
         int pixTrans = img2draw.getRGB(0, cutHeight - 1);
         img2draw = getTransparentImage(img2draw, pixTrans);
      }

      return new ImageWrapper<>(img2draw, height, width);
   }

   /**
    * Calculate the transformed image bounds.
    * 
    * @param    struct
    *           The holder for images parameters.
    * 
    * @return The 2-elements array filled with new transformed image width and its new height.
    */
   @Override
   public Integer[] calculateImageBounds(EmbeddedImageStructure<BufferedImage> struct)
   {
      int width  = 0;
      int height = 0;
      // scaling factors for vertical and horizontal
      float kHor = 0;
      float kVer = 0;

      // need to recalc the width and height for subimage to cut
      // to not go out of image borders
      int reqOffsetX = struct.getXOffset();
      int reqOffsetY = struct.getYOffset();

      width = struct.getImage().getWidth() - reqOffsetX;
      if (width > struct.getWidth())
      {
         width = struct.getWidth();
      }

      height = struct.getImage().getHeight() - reqOffsetY;
      if (height > struct.getHeight())
      {
         height = struct.getHeight();
      }

      kHor = ((float) struct.getWidth())  / ((float) width);
      kVer = ((float) struct.getHeight()) / ((float) height);

      if (struct.isStretchToFit())
      {
         // resizing
         if (struct.isRetainShape() && kHor != kVer)
         {
            // need to calculate native aspect ratio for image
            // if we have two different scaling factors for x and y we choose only one
            // to keep scaling in both directions, this must be one that is smaller
            if (kHor < kVer)
            {
               width = struct.getWidth();
               height = Math.round(((float) height) * kHor);
            }
            else
            {
               width = Math.round(((float) width) * kVer);
               height = struct.getHeight();
            }
         }
         else
         {
            width = struct.getWidth();
            height = struct.getHeight();
         }
      }

      return new Integer[] {width, height};
   }
   
   /**
    * Transform the regular image to the disabled gray image to draw the controls in disabled
    * state.
    * 
    * @param    img
    *           The initial image to convert.
    * @param    darkColor
    *           The 32-bit RGB color value used for dark parts of the final image.
    * @param    lightColor
    *           The 32-bit RGB color value used for light parts of the final image.
    * 
    * @return   The grey version of the initial imade with 3D effect.
    */
   public BufferedImage createDisabledGreyImage(BufferedImage img, int darkColor, int lightColor)
   {
      // threshold RGB value is greyscale 129.
      final int THRESHOLD_RGB = 0x00818181;
      
      // the first step is to convert initial image to transparent
      int pixTrans = img.getRGB(0, img.getHeight() - 1);
      BufferedImage imgTrans = getTransparentImage(img, pixTrans);
      
      // define custom image producer to create dark version
      ImageProducer ip = new FilteredImageSource(img.getSource(),
                            new ThresholdFilter(THRESHOLD_RGB, darkColor));
      
      // get the dark version of the image;
      BufferedImage imgDark = getBufferedImage(Toolkit.getDefaultToolkit().createImage(ip));
      
      // re-define custom image producer to create light version
      ip = new FilteredImageSource(img.getSource(),
              new ThresholdFilter(THRESHOLD_RGB, lightColor));
      
      // get the light version of the image;
      BufferedImage imgLight = getBufferedImage(Toolkit.getDefaultToolkit().createImage(ip));
      
      // the next step is to move the light image by 1 pixel down-right
      // create empty image to be the target based on the source dimensions
      BufferedImage imgRes = new BufferedImage(img.getWidth(null), img.getHeight(null),
                                               BufferedImage.TYPE_INT_ARGB);

      // if all images are OK do the final thing combining all in proper sequence
      if (imgRes != null && imgLight != null && imgDark != null)
      {
         Graphics2D g2img = imgRes.createGraphics();
         // copy data from light version to newly created image with 1 XY pixel shift
         g2img.drawImage(imgLight, 1, 1, null);
         // put dark version on the top of the light to simulate 3D effect
         g2img.drawImage(imgDark, 0, 0, null);
         // all done commit the changes
         g2img.dispose();
      }
      
      return imgRes;
   }

   /**
    * Converts the image replacing standard 3D colors with currently redefined.
    * 
    * @param   img
    *          Initial image to convert.
    * 
    * @return   The image with replaced 3D colors color.
    */
   private BufferedImage get3DColorsConvertedImage(BufferedImage img)
   {
      // get the driver to identify colors, assuming gd can not be null at this time
      GuiDriver gd = (GuiDriver) OutputManager.getDriver();
      // define custom image producer
      // we take the 3d colors from the color set currently used by GuiDriver
      ImageProducer ip = new FilteredImageSource(img.getSource(),
                new Convert3DColorsFilter(gd.getSysColor(SysColor.COLOR_BTNHIGHLIGHT).getGuiRgb(),
                                          gd.getSysColor(SysColor.COLOR_3DFACE).getGuiRgb(),
                                          gd.getSysColor(SysColor.COLOR_3DSHADOW).getGuiRgb(),
                                          gd.getSysColor(SysColor.COLOR_BTNTEXT).getGuiRgb()));
      // convert image
      return getBufferedImage(Toolkit.getDefaultToolkit().createImage(ip));
   }

   /**
    * Converts the image into buffered image.
    * 
    * @param   imgToConvert
    *          Initial image to convert.
    * 
    * @return   The buffered image for further use.
    */
   private BufferedImage getBufferedImage(java.awt.Image imgToConvert)
   {
      // create empty image
      BufferedImage imgRes = new BufferedImage(imgToConvert.getWidth(null),
                                               imgToConvert.getHeight(null),
                                               BufferedImage.TYPE_INT_ARGB);

      // copy data to newly created image
      Graphics2D g2img = imgRes.createGraphics();
      g2img.drawImage(imgToConvert, 0, 0, null);
      g2img.dispose();
      
      return imgRes;
   }

   /**
    * Custom RGB filter to change image transparency data on the fly.
    */
   static class TransparencyFilter
   extends RGBImageFilter
   {
      /** Alpha channel constant. */
      private static final int TRANSPARENCY_MASK = 0xFF000000;
      
      /** Pixel to use as transparent marker. */
      int rgbCheck;
      
      /**
       * Creates new filter with the given transparent pixel value.
       *
       * @param   pixTrans
       *          The pixel sample we will consider as transparent.
       */
      TransparencyFilter(int pixTrans)
      {
         rgbCheck = pixTrans | TRANSPARENCY_MASK;
      }

      /**
       * This is on the fly converter to turn on transparency for every pixels the sames as
       * one we defined as marker on the filter construction.  This method is calling when image
       * drawing is happening.
       *
       * @param   x
       *          The X pixel coordinate.
       * @param   y
       *          The Y pixel coordinate.
       * @param   pixel
       *          The ARGB pixel value of the original image.
       * 
       * @return   The new ARGB pixel value after transformation to render.
       */
      public int filterRGB(int x, int y, int pixel)
      {
         return (pixel | TRANSPARENCY_MASK) == rgbCheck ? pixel & ~TRANSPARENCY_MASK : pixel; 
      }
   }

   /**
    * Converts the image to one having transparent color.
    * 
    * @param   img
    *          Initial image to convert.
    * @param   transPix
    *          The color to be considered as transparent.
    * 
    * @return   The image with one tarnsparent color.
    */
   private BufferedImage getTransparentImage(BufferedImage img, int transPix)
   {
      // define custom image producer
      ImageProducer ip = new FilteredImageSource(img.getSource(),
                                                 new TransparencyFilter(transPix));
      // convert image
      return getBufferedImage(Toolkit.getDefaultToolkit().createImage(ip));
   }

   /**
    * Custom RGB filter to change image 3D colors on the fly.
    */
   static class Convert3DColorsFilter
   extends RGBImageFilter
   {
      /** Original 3D white color(255, 255, 255). */
      private static final int COLOR_3D_WHITE   = 0x00FFFFFF;
      
      /** Original 3D light gray color(192, 192, 192). */
      private static final int COLOR_3D_LT_GRAY = 0x00C0C0C0;
      
      /** Original 3D dark gray color(128, 128, 128). */
      private static final int COLOR_3D_DK_GRAY = 0x00808080;
      
      /** Original 3D black color(0, 0, 0). */
      private static final int COLOR_3D_BLACK   = 0x00000000;
      
      /** New value for 3D white. */
      int newWhite;

      /** New value for 3D light gray. */
      int newLtGray;
      
      /** New value for 3D dark gray. */
      int newDkGray;
      
      /** New value for 3D black. */
      int newBlack;
      
      /**
       * Creates new filter with the given 3D colors to replace.
       *
       * @param   white
       *          The new value for white color.
       * @param   ltGray
       *          The new value for light gray color.
       * @param   dkGray
       *          The new value for dark gray color.
       * @param   black
       *          The new value for dark gray color.
       */
      Convert3DColorsFilter(int white, int ltGray, int dkGray, int black)
      {
         newWhite  = white;
         newLtGray = ltGray;
         newDkGray = dkGray;
         newBlack  = black;
      }

      /**
       * This is on the fly converter to change the 3D colors with one we passed on the filter
       * construction.  This method is calling when image drawing is happening.
       *
       * @param   x
       *          The X pixel coordinate.
       * @param   y
       *          The Y pixel coordinate.
       * @param   pixel
       *          The ARGB pixel value of the original image.
       * 
       * @return   The new ARGB pixel value after transformation to render.
       */
      public int filterRGB(int x, int y, int pixel)
      {
         // 3d color replacer filter implementation
         // differentiate 4 colors to replace: white(255,255,255), light grey(192,192,192),
         // dark grey(128,128,128) and black(0,0,0) from original pixel and replaces with colors
         // defined on filter construction, the black color search should be a bit different than
         // other because black dows not have any bits in value,
         // the final color keep the tranparency value from the original pixel
         return (pixel & COLOR_3D_WHITE) == COLOR_3D_WHITE ?
                   newWhite | (pixel & TransparencyFilter.TRANSPARENCY_MASK) : 
                (pixel & COLOR_3D_LT_GRAY) == COLOR_3D_LT_GRAY ?
                   newLtGray | (pixel & TransparencyFilter.TRANSPARENCY_MASK) :
                (pixel & COLOR_3D_DK_GRAY) == COLOR_3D_DK_GRAY ?
                   newDkGray | (pixel & TransparencyFilter.TRANSPARENCY_MASK) :
                (pixel & ~TransparencyFilter.TRANSPARENCY_MASK) == COLOR_3D_BLACK ?
                   newBlack | (pixel & TransparencyFilter.TRANSPARENCY_MASK) : pixel; 
      }
   }

   /**
    * Custom RGB filter to apply threshold effect on the fly.
    */
   static class ThresholdFilter
   extends RGBImageFilter
   {
      /** New color for pixels over threshold. */
      private static final int COLOR_BLACK_TRANSPARENT   = 0x00000000;
      
      /** The threshold value. */
      int threshold;

      /** New value for pixel color under threeshold. */
      int replacement;
      
      /**
       * Creates new filter with the threshold value and replacement color.
       *
       * @param   threshold
       *          The color to use as filter threshold.
       * @param   replacement
       *          The color to replace the pixels under threshold.
       */
      ThresholdFilter(int threshold, int replacement)
      {
         // set up internals
         this.threshold  = threshold;
         this.replacement = replacement | TransparencyFilter.TRANSPARENCY_MASK;
         // we are able to handle indexed images
         canFilterIndexColorModel = true;
      }

      /**
       * This is on the fly converter to change the image color by setting all pixels under
       * threshold value to the color defined on filter construction.  The transpatent pixels
       * are set to black transparent color.This method is calling when image drawing is
       * happening.
       *
       * @param   x
       *          The X pixel coordinate.
       * @param   y
       *          The Y pixel coordinate.
       * @param   pixel
       *          The ARGB pixel value of the original image.
       * 
       * @return   The new ARGB pixel value after transformation to render.
       */
      public int filterRGB(int x, int y, int pixel)
      {
         // the threshold filter implementation, the condition for the color of interest:
         // - the pixel must be at least not fully transparent
         // - the RGB value must be less than threshold value
         // if the both conditiona are met - we replace pixel color with predefined constant
         // color, otherwise we replace pixel color with transparent black.
         // the example of the similar work can be seen in GIMP threshold filter
         // the image is not mandatory to be converted to greyscale before transformation
         return (pixel & TransparencyFilter.TRANSPARENCY_MASK) != COLOR_BLACK_TRANSPARENT &&
                (pixel & ~TransparencyFilter.TRANSPARENCY_MASK) < threshold
                   ? replacement : COLOR_BLACK_TRANSPARENT;
      }
   }
}