/*
** Module   : MouseHoverAction.java
** Abstract : Mouse hover event handler.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description--------------------------------------
** 001 IAS 20151224 Created initial version.
** 002 HC  20160511 Fixed COMBO-BOX focusing issues and related changes.
** 003 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui.driver;

import java.awt.event.*;
import java.util.*;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Mouse hover event handler.
 */
public class MouseHoverAction 
extends MouseWidgetAction
{
   /** the id of the sub-widget which is the mouse hover target */
   protected final int effectiveWidgetId;

   /** flag indicationg that handler is registred */
   protected boolean registered = false;
   
   /** the id of thw widget's window */
   protected volatile int windowId = -1; 

   /**
    * Constructor.
    * 
    * @param    widget
    *           The widget which supports custom mouse cursor.
    * @param    effectiveWidgetId
    *           The id of the sub-widget which is the mouse hover target 
    */
   public MouseHoverAction(Widget<?> widget, int effectiveWidgetId)
   {
      super(widget, 
            new int[] { MouseEvent.MOUSE_ENTERED, MouseEvent.MOUSE_EXITED }, 
            MouseEvt.MOUSE_HOVERABLE);
      this.effectiveWidgetId = effectiveWidgetId;
   }

   /**
    * Constructor.
    * 
    * @param    widget
    *           The widget which supports custom mouse cursor.
    */
   public MouseHoverAction(Widget<?> widget)
   {
         this(widget, widget.getId() == null ? -1 : widget.getId().asInt());
   }

   /**
    * On mouse entered, set the mouse cursor to the effective custom cursor.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public synchronized void mouseEntered(MouseEvent e)
   {
      ThinClient.getInstance().pushHover(this);
      setCursor(widget.getEffectiveMousePointer());
   }
   
   /**
    * On mouse exit, set the mouse cursor to the parent widget cursor.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public synchronized void mouseExited(MouseEvent e)
   {
      ThinClient.getInstance().popHover();
      setCursor(parentCursor());
   }

   /**
    * Update the custom cursor shape
    */
   public synchronized void updateCursor()
   {
      setCursor(widget.getEffectiveMousePointer());
   }

   /**
    * Get the custom cursor of the parent widget
    * 
    * @return the custom cursor of the parent widget
    */
   protected MousePtrWrapper parentCursor()
   {
      return widget.enclosingFrame().map(Frame::getEffectiveMousePointer).
            orElse(MousePtrWrapper.DEFAULT);
   }

   /**
    * Get the GUI driver
    * 
    * @return the GUI driver
    */
   public GuiDriver gd()
   {
      return gd;
   }
   
   /**
    * Get the id of the sub-widget which is the mouse hover target
    *
    * @return the id of the sub-widget which is the mouse hover target
    */
   public int wid()
   {
      return effectiveWidgetId;
   }
   
   /**
    * Get the widget for which the action is defined
    *
    * @return the widget for which the action is defined
    */
   public Widget<?> widget()
   {
      return widget;
   }

   /**
    * Get the id of the enclosing window
    *
    * @return the id of the enclosing window
    */
   public int windowId()
   {
      return windowId;
   }
   
   /**
    * Register the handler
    */
   public void register()
   {
      if (!registered)
      {
         gd.registerHoverableWidget(this);
         windowId = widget.topLevelWindow().getId().asInt();
         registered = true;
      }
   }

   /**
    * De-register the handler
    */
   public void deregister()
   {
      if (registered)
      {
         gd.deregisterWidget(widget.getId().asInt());
         registered = false;
      }
   }

   /**
    * Set the custom mouse pointer
    *
    * @param ptr
    *        the custom mouse pointer
    */
   protected void setCursor(MousePtrWrapper ptr)
   {
      if (ThinClient.isMousePointerLocked() && 
          ThinClient.getInstance().activeComboBox().getDropDownList() != widget) 
      {
         return;
      }
      try
      {
         gd.selectWindow(windowId);
         try
         {
            gd.setCursor(ptr);
         }
         finally
         {
            gd.releaseWindow();
         }
      }
      catch (Exception e)
      {
         System.out.printf("Unexpected exception. Widget: %s\n", widget.getClass().getName());
         e.printStackTrace();
      }
   }

   /**
    * Get the String representation of the handler
    *
    * @return the the String representation of the handler
    */
   @Override
   public String toString()
   {
      return String.format("MouseHoverAction[%s, %s]", widget.getClass().getSimpleName(), 
            widget.name());
   }

}