/*
** Module   : SwingEmulatedWindow.java
** Abstract : core GUI renderer for AWT/Swing client
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support - removed fontSize.
** 003 CA  20140806 Added mouse support and other misc APIs related to window state.
** 004 MAG 20140912 Added fill polygon.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140924 Create graphical object when buffered image is created.
** 007 CA  20141021 Added notification of OS window events.  Added support for batch drawing: the 
**                  current AWT repaint processing will wait until the last paint primitive is
**                  received, before returning the control to the AWT event processing thread, to
**                  avoid flickering.  Do not allow window maximize via OS means.
**                  Added implementation for new paint primitives: SET_LOCATION, RESIZE_WINDOW.
** 008 HC  20141031 Implemented support for drawing highlighted areas.
** 009 MAG 20141020 Improve clipping rectangle management. Added mouse wheel listener.
** 010 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 EVL 20141116 Adding round rectangles draw and fill.
** 013 CA  20141123 Fixed processing of window focus gained/lost.  Added APIs to allow movement of
**                  a window to top (at OS level).  When the window is destroyed, first remove the
**                  listeners, so no key/mouse/window events is processed while the window is being
**                  destroyed.
** 014 CA  20141125 Do not accept negative dimensions, when resizing the screen buffers.
**                  Added API to conditionally enable/disable OS-level events.
** 015 EVL 20141125 Adding dotted line draw capability.
** 016 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 017 HC  20141207 Implemented runtime support for several SESSION attributes.
** 018 HC  20150106 Javadoc fixes.
** 019 EVL 20150203 Adding image offset drawing capability.
** 020 EVL 20150218 Adding transparency handling.
** 021 EVL 20150509 Fixing clipping rectangle dimension computing.
** 022 CA  20150408 All window events are treated as OS events.  Allow the title of the OS window
**                  to be set.  Added deiconifyWindow and isIconified.
** 023 CA  20150429 Fixed a problem in setScreenBitmap, when the received bitmap is not yet aware
**                  of the window's dimensions.
** 024 CA  20150507 Allow text to be centered vertically.
**                  Moved from legacy metric scaling to DPI scaling from 72 to OS DPI.
** 025 GES 20150325 Major refactoring to move all generic functionality to a new abstract parent
**                  class or to the new abstract GUI driver. Old JComponent parent is now just
**                  contained as a member.
** 026 HC  20150612 Support for modal windows.
** 027 HC  20150702 Implemented window modality independently on the actual UI driver. Improved
**                  positioning of top-level windows on systems with multiple displays.
** 028 GES 20150715 Cleanup of window resources by calling quit() on the content pane,
**                  moved the common image processing functionalities to the separate unit.
** 029 CA  20150819 Added support for "underline" style fonts.
** 030 HC  20150831 Improved GUI window resize performance.
** 031 GES 20150817 Removed isOnTop() since it is unnecessary. The saveFocusListeners(),
**                  restoreFocusListeners() and moveToTop() are only needed for Swing usage and
**                  are now just local helper methods.  Added custom Swing mouse handler.
**                  Implemented SET_ICON drawing primitive.
** 032 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
**                  SwingUtilities.invokeAndWait must be bracketed in a save/restore bracket state
**                  calls, to allow the AWTEventThread to process the request; otherwise, deadlocks
**                  may occur. 
** 033 SVL 20151013 Handled SET_XOR_COMPOSITE and RESET_COMPOSITE, added new line stroke.
** 034 SBI 20151012 Changed the image helper method signature to be more specific. Added z-order
**                  operations.  Removed save/restore of focus listeners and replaced it with
**                  a global flag.
** 035 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     SBI 20151103 Added the on show component listener to fix the lost keys input.
**     EVL 20151106 Adding method to obtain the BufferedImageDrawHelper instance.
** 036 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 037 EVL 20160104 Overlay window deactivation need to be generated when window is losing focus.
**                  Adding method to control the focus events generation.
**     HC  20160113 Requesting focus in moveToTop() of a non-showing window is performed
**                  asynchronously after the window is shown.
** 038 IAS 20160130 LOAD-MOUSE-POINTER support
** 039 HC  20160304 Window double buffering optimization - image buffer is allocated only for
**                  part of the window area.
** 040 SBI 20160324 Added SET_DESKTOP_BGCOLOR.
** 041 HC  20160410 Overhaul of window-activation logic. Fixed Swing GUI window sometimes not
**                  receiving key input when activated.
** 042 HC  20160502 An attempt to fix Swing JFrame sometime not focusing when
**                  JFrame.requestFocus() called right after the window is made visible.
** 043 SBI 20160729 Fixed repositionScreenBuffers() to run window repaint code in the main thread
**                  and fixed resizeScreenBuffer to remove pack() and to repaint via the main
**                  thread in order to notify the main thread that it leaves its wait state.
** 044 HC  20160823 Another attempt to fix Swing JFrame sometime not focusing when
**                  JFrame.requestFocus() called right after the window is made visible.
** 045 CA  20170106 Avoid posting size/location change events to the driver if these values are 
**                  the same with the current ones.
** 046 CA  20170131 Fixed convertFontStyle - BOLD/ITALIC info was lost.
**     HC  20170202 Fixed potential race condition between GUI and WindowManager.
**     HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE.
**     HC  20170203 Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                  TC.invokeLater() to allow execution of code to be queued on the application
**                  queue.
**     HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
**                  overlapping regions.
** 047 OM  20170214 Replaced base class from JFrame to JDialog which suits better our purpose.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.lang.reflect.*;
import java.text.*;
import java.util.*;
import java.util.List;
import java.util.concurrent.*;
import javax.swing.*;
import sun.java2d.*;
import sun.java2d.loops.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * The core GUI renderer. Because the painting results should be permanent a BufferedImage is
 * used to draw and keep the last screen image. A pool of drawing operations is used to queue 
 * drawing primitives. The operations are enqueued in the pool one at a time and then are drawn
 * on the buffered image. When the pool is empty, the screen is updated with the image from the
 * stored buffered image.  
 */
public class SwingEmulatedWindow 
extends EmulatedWindowState<Font, BufferedImage>
{
   /** The number of milliseconds to delay window activation. */  
   private static final long ACTIVATION_DELAY = 10;

   /** Single threaded executor for delayed window activation. */
   private static ExecutorService activationExecutor = Executors.newSingleThreadExecutor(); 
   
   /** Map of saved event listeners, which can be restored at a later time. */
   private Map<Class<? extends EventListener>, List<EventListener>> savedListeners = null;
   
   /** Content pane for the window. */
   private ContentPane pane = null;

   /** Screen buffer */
   private BufferedImage screenBuffer;
   
   /** Graphical object */
   private Graphics2D g2;

   /** A stack of clipping rectangles. */
   private Deque<Shape> clippings = new ConcurrentLinkedDeque<>();

   /** The X translated location of the drawing area, on the physical canvas. */
   private int translateX = 0;
   
   /** The Y translated location of the drawing area, on the physical canvas. */
   private int translateY = 0;
   
   /** Screen buffer for clearing an active highlight */
   private BufferedImage highlightRestoreBuffer;
   
   /** The left location of the last highlighted widget. */
   private int highlightX;
   
   /** The top location of the last highlighted widget. */
   private int highlightY;

   /** Containing window (may be an applet or application frame/dialog). */
   private Window window = null;
   
   /** Dotted line pattern. */
   private Stroke dotStroke;
   
   /** Small dotted line pattern. */
   private Stroke dotSmallStroke;
   
   /** Small and thin dotted line pattern. */
   private Stroke dotSmallThinStroke;
   
   /** Large dotted line pattern. */
   private Stroke dotLargeStroke;
   
   /** Temporary context stored during <code>paintComponent()</code> operations. */
   private Graphics graphics = null;
   
   /** Helper to generate image wrappers. */
   private final BufferedImageDrawHelper drawHelper;
   
   /** The original window enabled state when enableEvents(false) is called. */
   private boolean prevEnabled = true;
   
   /** The screen buffer offset. **/
   private Point screenBufferOffset = new Point();
   
   /** The flag indicating this window is ready to accept focus requests. */
   private boolean readyToFocus;
   
   /**
    * Set to true when a focus request is issued before this window is ready to accept focus
    * requests.
    */
   private boolean pendingFocus;

   /**
    * Constructor.
    * 
    * @param   window
    *          Application window.
    * @param   keyReader
    *          The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   @SuppressWarnings("serial")
   public SwingEmulatedWindow(Window window, SwingKeyboardReader keyReader)
   {
      super(new SwingMouseHandler());
      
      // initialize dot pattern 
      dotStroke = new BasicStroke(LineStroke.DOT_LINE_WIDTH,
                                  BasicStroke.CAP_BUTT,
                                  BasicStroke.JOIN_BEVEL,
                                  0.0f,
                                  new float[] {LineStroke.DOT_STEP, LineStroke.DOT_STEP},
                                  LineStroke.DOT_STEP);
      
      dotSmallStroke = new BasicStroke(LineStroke.DOT_LINE_WIDTH,
                                       BasicStroke.CAP_BUTT,
                                       BasicStroke.JOIN_BEVEL,
                                       0.0f,
                                       new float[] {LineStroke.DOT_STEP_SMALL,
                                                    LineStroke.DOT_STEP_SMALL},
                                       LineStroke.DOT_STEP_SMALL);
      
      dotSmallThinStroke = new BasicStroke(LineStroke.DOT_LINE_THIN_WIDTH,
                                           BasicStroke.CAP_BUTT,
                                           BasicStroke.JOIN_BEVEL,
                                           0.0f,
                                           new float[] {LineStroke.DOT_STEP_SMALL,
                                                        LineStroke.DOT_STEP_SMALL},
                                           LineStroke.DOT_STEP_SMALL);

      dotLargeStroke     = new BasicStroke(LineStroke.DOT_LINE_WIDTH,
                                           BasicStroke.CAP_BUTT,
                                           BasicStroke.JOIN_MITER,
                                           1.0f,
                                           new float[] {LineStroke.DOT_STEP_LARGE,
                                                        LineStroke.DOT_STEP_LARGE},
                                           0);

      drawHelper = new BufferedImageDrawHelper();
      
      // set the window reference
      this.mouseHandler.setWindow(this);
      
      pane = new ContentPane(window, keyReader)
      {
         public void paintComponent(Graphics graphics)
         {
            // save off graphics context (this is needed because other GUI clients don't use
            // the same graphics code and drawOnScreenBitmap() is common code so it can't be
            // passed as a parameter)
            SwingEmulatedWindow.this.graphics = graphics;
            
            try
            {
               SwingEmulatedWindow.this.drawOnScreenBitmap();
            }
            catch (Exception e)
            {
               // some exception occurred while rendering process, prevent crashing SWING events
               e.printStackTrace();
            }
            
            // clear graphics context
            SwingEmulatedWindow.this.graphics = null;
            
            // at this point the Swing window should be in a sound state to properly handle
            // focus requests
            readyToFocus = true;
            if (pendingFocus)
            {
               pendingFocus = false;
               doMoveToTop(true);
            }
         }
      };
      
      // let the app above handle the decoration itself
      if (window instanceof JFrame)
      {
         ((JFrame) window).setUndecorated(true);
      }
      else if (window instanceof JDialog)
      {
         ((JDialog) window).setUndecorated(true);
      }
      
      // prevent Swing to activate/focus the window when it is brought to front or made visible.
      // This seems to be the only way to properly manage window focus since this will effectively
      // trigger Swing window-focus event only as a result of user or OS action. With 
      // auto-request-focus ON Swing sends window-focus even when window is made visible or
      // brought to front and these cannot be reliably filtered out due to the asynchronous nature
      // of the (Linux) UI backend.
      window.setAutoRequestFocus(false);
      
      this.window = window;
      
      // register listeners for OS window events
      window.addWindowListener(new WindowAdapter()
      {
         @Override
         public void windowIconified(WindowEvent e)
         {
            iconifyWindow(-1, true);
         }
         
         @Override
         public void windowDeiconified(WindowEvent e)
         {
            deiconifyWindow(-1);
         }

         @Override
         public void windowActivated(WindowEvent e)
         {
            // in some cases, when the window is activated, it is not properly focused,
            // listening on key input on the content pane seems to be more reliable
            pane.requestFocusInWindow();
            boolean fromOS = (e.getOppositeWindow() == null);
            ThinClient.getInstance().invokeLaterOS(() ->
                  WindowManager.windowActivated(windowId, fromOS));
         }

         @Override
         public void windowDeactivated(WindowEvent e)
         {
            boolean toOS = (e.getOppositeWindow() == null);
            ThinClient.getInstance().invokeLaterOS(() ->
                  WindowManager.windowDeactivated(windowId, toOS));
         }
      });

      window.addWindowStateListener(e ->
      {
         if ((e.getNewState() & Frame.MAXIMIZED_BOTH) == Frame.MAXIMIZED_BOTH)
         {
            WindowGuiImpl wnd = WindowManager.findWindow(windowId);
            
            wnd.maximize();
         }
      });
      
      window.addComponentListener(new ComponentAdapter() 
      {
         @Override
         public void componentMoved(ComponentEvent e)
         {
            super.componentMoved(e);
            if (window.isShowing())
            {
               repositionScreenBuffers();
            }
         }
      });

      // setup mouse listeners
      pane.addMouseListener(mouseHandler);
      pane.addMouseMotionListener(mouseHandler);
      pane.addMouseWheelListener(mouseHandler);
      
      // TODO: disable the popup of the window actions (min/max/close/etc) via key? 
   }

   /**
    * Cause the outer frame to be disposed, as the window is being terminated.
    */
   @Override
   public void quit()
   {
      safeInvokeAndWait(() ->
      {
         for (MouseListener l : pane.getMouseListeners())
         {
            pane.removeMouseListener(l);
         }
         for (MouseMotionListener l : pane.getMouseMotionListeners())
         {
            pane.removeMouseMotionListener(l);
         }
         for (MouseWheelListener l : pane.getMouseWheelListeners())
         {
            pane.removeMouseWheelListener(l);
         }
         for (WindowListener l : window.getWindowListeners())
         {
            window.removeWindowListener(l);
         }
         for (WindowFocusListener l : window.getWindowFocusListeners())
         {
            window.removeWindowFocusListener(l);
         }
         for (WindowStateListener l : window.getWindowStateListeners())
         {
            window.removeWindowStateListener(l);
         }
         for (KeyListener l : window.getKeyListeners())
         {
            window.removeKeyListener(l);
         }
         for (ComponentListener l : window.getComponentListeners())
         {
            window.removeComponentListener(l);
         }
         
         // cleanup window resources
         pane.quit();
      });
   }
   
   /**
    * Cause repaint processing to occur.
    */
   @Override
   public void repaint()
   {
      SwingUtilities.invokeLater(() -> pane.repaint());
   }

   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   @Override
   public void setWindowVisible(boolean visible)
   {
      SwingUtilities.invokeLater(() -> 
      {
         window.pack();
         window.setVisible(visible);
      });
   }
   
   /**
    * Set window location.
    *
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   @Override
   public void setWindowLocation(int left, int top)
   {
      SwingUtilities.invokeLater(() -> window.setLocation(left, top));
   }
   
   /**
    * Resize window.<br>
    * Because of swing constraints, the effective command is executed on Swing dispatch thread.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height
    */
   @Override
   public void resizeWindow(int width, int height)
   {
      if (getHeight() == height && getWidth() == width)
      {
         return; // no-op
      }
      
      Dimension sz = new Dimension(width, height);
      
      safeInvokeAndWait(() -> {
         window.setSize(sz);
         window.setPreferredSize(sz);
         pane.setSize(sz);
         pane.setPreferredSize(sz);
         pane.setMinimumSize(sz);
         pane.setMaximumSize(sz);
      });
      
      resizeScreenBuffers(width, height);
      super.resizeWindow(width, height); // calls repaint and updates [bitmap] dimensions
   }
   
   /**
    * Set new bounds for this window.<br>
    * Because of swing constraints, the effective command is executed on Swing dispatch thread.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height
    */
   @Override
   public void setWindowBounds(int left, int top, int width, int height)
   {
      NativePoint wloc = getWindowLocation();
      if (getHeight() == height && getWidth() == width && wloc.x == left && wloc.y == top)
      {
         return; // no-op
      }
      
      Dimension sz = new Dimension(width, height);
      
      safeInvokeAndWait(() -> {
         window.setBounds(left, top, width, height);
         window.setPreferredSize(sz);
         pane.setSize(sz);
         pane.setPreferredSize(sz);
         pane.setMinimumSize(sz);
         pane.setMaximumSize(sz);
      });
      
      resizeScreenBuffers(width, height);
      super.resizeWindow(width, height); // calls repaint and updates [bitmap] dimensions
   }
   
   /**
    * Set the window' sensitive state, depending on the {@code enable} flag.
    * 
    * @param    enable
    *           The window sensitive state.
    */
   @Override
   public void setWindowEnabled(boolean enable)
   {
      SwingUtilities.invokeLater(() ->
      {
         // check whether events are enabled by the upper layer
         if (savedListeners == null)
         {
            window.setEnabled(enable);
         }
         else
         {
            // events are disabled, just save the flag, it will be used in enableEventsImpl()
            prevEnabled = enable;
         }
      });
   }

   /**
    * Check if the simulated window is visible.
    * 
    * @return   See above.
    */
   @Override
   public boolean isWindowVisible()
   {
      synchronized(lock())
      {
         return window.isVisible();
      }
   }

   /**
    * Get the window location on screen.
    * 
    * @return   This window's location.
    */
   @Override
   public NativePoint getWindowLocation()
   {
      synchronized(lock())
      {
         if (window.isVisible())
         {
            java.awt.Point p = window.getLocationOnScreen();
            return new NativePoint(p.x, p.y);
         }
         else
         {
            // it seems that getLocation in this case really returns the virtual screen location
            java.awt.Point p = window.getLocation(); 
            
            return new NativePoint(p.x, p.y);
         }
      }
   }

   /**
    * Get the window height.
    * 
    * @return   This window's width in pixels.
    */
   @Override
   public int getHeight()
   {
      synchronized(lock())
      {
         return pane.getHeight();
      }
   }
   
   /**
    * Get the window width.
    * 
    * @return   This window's width in pixels.
    */
   @Override
   public int getWidth()
   {
      synchronized(lock())
      {
         return pane.getWidth();
      }
   }
   
   /**
    * Manage any window state related to iconification.
    * 
    * @param    iconify
    *           <code>true</code> if this window is being iconified, <code>false</code> for
    *           de-iconified.
    */
   @Override
   public void setIconificationState(boolean iconify)
   {
      if (!(window instanceof Frame))
      {
         throw new RuntimeException("setIconificationState is only supported on awt Frames!");
      }
      
      SwingUtilities.invokeLater(() -> 
      {
         super.setIconificationState(iconify);
         
         if (iconify)
         {
            ((Frame) window).setState(Frame.ICONIFIED);
         }
         else
         {
            int extendedState = ((Frame) window).getExtendedState() & ~Frame.ICONIFIED;
            ((Frame) window).setExtendedState(extendedState);
         }
      });
   }

   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    type
    *           The cursor type.
    */
   public void setCursor(CursorType type)
   {
      SwingUtilities.invokeLater(() -> pane.setCursor(Cursor.getPredefinedCursor(type.awtCursor)));
   }

   /**
    * Set the mouse cursor to the given one.
    * 
    * @param    c
    *           The custom cursor.
    */
   public void setCursor(Cursor c)
   {
      SwingUtilities.invokeLater(() -> pane.setCursor(c));
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * The method appends the request on the Swing event dispatching thread
    * and waits for it to be processed before returning. 
    * 
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(boolean capture)
   {
      SwingUtilities.invokeLater(() -> enableEventsImpl(capture));
   }
   
   /**
    * Enable or Disable OS-level events for the specified windows.
    * This is the actual implementation and must be executed on the Swing event
    * dispatching thread.
    * 
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   private void enableEventsImpl(boolean capture)
   {
      if (capture)
      {
         // instead of skipping window state events (and making the app out of sync 
         // with the OS) just change the window's enabled state
         // TODO: do we need to manage the mouse/key listeners when window enabled
         //       state is changed by setEnabled()?
         window.setEnabled(prevEnabled);

         if (savedListeners == null)
         {
            return;
         }
         
         List<EventListener> events;
         
         events = savedListeners.remove(MouseListener.class);
         for (EventListener l : events)
         {
            pane.addMouseListener((MouseListener) l);
         }
         
         events = savedListeners.remove(MouseMotionListener.class);
         for (EventListener l : events)
         {
            pane.addMouseMotionListener((MouseMotionListener) l);
         }
   
         events = savedListeners.remove(MouseWheelListener.class);
         for (EventListener l : events)
         {
            pane.addMouseWheelListener((MouseWheelListener) l);
         }
   
         events = savedListeners.remove(KeyListener.class);
         for (EventListener l : events)
         {
            window.addKeyListener((KeyListener) l);
         }
         
         savedListeners = null;
      }
      else
      {
         if (savedListeners != null)
         {
            return;
         }
         
         prevEnabled = window.isEnabled();
         window.setEnabled(false);

         savedListeners = new HashMap<>();
   
         List<EventListener> events;
         
         events = new ArrayList<>();
         for (MouseListener l : pane.getMouseListeners())
         {
            pane.removeMouseListener(l);
            events.add(l);
         }
         savedListeners.put(MouseListener.class, Collections.unmodifiableList(events));
         
         events = new ArrayList<>();
         for (MouseMotionListener l : pane.getMouseMotionListeners())
         {
            pane.removeMouseMotionListener(l);
            events.add(l);
         }
         savedListeners.put(MouseMotionListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (MouseWheelListener l : pane.getMouseWheelListeners())
         {
            pane.removeMouseWheelListener(l);
            events.add(l);
         }
         savedListeners.put(MouseWheelListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (KeyListener l : window.getKeyListeners())
         {
            window.removeKeyListener(l);
            events.add(l);
         }
         savedListeners.put(KeyListener.class, Collections.unmodifiableList(events));
      }
   }

   /**
    * Get the metrics of the given font.
    *  
    * @param    f
    *           The font for which the font metrics are to be obtained.
    *           
    * @return   The font metrics.
    */
   @Override
   public FontMetricsHelper getFontMetrics(FontDetails<Font> f)
   {
      synchronized (lock())
      {
         return new SwingFontMetrics(getNativeFontMetrics(f.font));
      }
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    f
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Override
   public int getTextHeight(String text, FontDetails<Font> f)
   {
      synchronized (lock())
      {
         FontMetrics fm = getNativeFontMetrics(f.font);
         LineMetrics lm = fm.getLineMetrics(text, g2 == null ? pane.getGraphics() : g2);
         
         return (int) lm.getHeight();
      }
   }
   
   /**
    * The method performs a layout operation on the supplied text and 
    * returns the resulting paragraph height while maintaining the
    * supplied maximum width.
    * <p>
    * Note that the algorithm for text layout is compatible with 
    * {@link PaintPrimitives#DRAW_PARAGRAPH} and so
    * can be used in conjunction with this method. 
    * 
    * @param    text
    *           The text to measure.
    * @param    font
    *           The font to use during layout.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   @Override
   public int getParagraphHeight(String text, FontDetails<Font> font, int maxWidth)
   {
      synchronized (lock())
      {
         Graphics2D g = (Graphics2D) (this.g2 == null ? pane.getGraphics() : this.g2);
         return layoutParagraphWorker(g, text, font, maxWidth, false, 0, 0);
      }
   }   
   
   /**
    * Draw using a paint operation.
    * 
    * @param    ps
    *           Paint operation parameters.
    */
   @Override
   public void draw(final PaintStructure<Font, BufferedImage> ps)
   {
      switch (ps.id)
      {
         case DRAW_STRING:
            drawString(ps);
            break;
         case DRAW_STRING_SCALED:
            drawStringScaled(ps);
            break;
         case DRAW_PARAGRAPH:
            layoutParagraphWorker(g2, ps.text, ps.font, ps.width, true, ps.x, ps.y);
            break;
         case DRAW_LINE:
            g2.drawLine(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_RECT:
            g2.drawRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_ROUND_RECT:
            g2.drawRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter, ps.arcDiameter);
            break;
         case DRAW_IMAGE:
            EmbeddedImageStructure      eis  = EmbeddedImageStructure.create(ps);
            ImageWrapper<BufferedImage> prep = drawHelper.processImage(eis);
            BufferedImage               bi   = prep.getImage();
            
            if (bi != null)
            {
               g2.drawImage(bi, ps.x, ps.y, prep.getWidth(), prep.getHeight(), pane);
            }
            break;
            
         case FILL_RECT:
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
            
         case FILL_ROUND_RECT:
            g2.fillRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter, ps.arcDiameter);
            break;
         case FILL_POLYGON:
            g2.fillPolygon(ps.xPoints, ps.yPoints, ps.nPoints);
            break;
         case DRAW_3D_RECT:
            g2.draw3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case FILL_3D_RECT:
            g2.fill3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case SET_COLOR:
            g2.setColor(SwingHelper.fromColorRgb(ps.color));
            break;
         case SET_XOR_COMPOSITE:
            g2.setComposite(new XORComposite(new Color(0, 0, 0), NullSurfaceData.theInstance));
            break;
         case RESET_COMPOSITE:
            g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
            break;
         case SET_LINE_STROKE:
            // default line stroke
            Stroke stroke = new BasicStroke();
            
            switch (ps.stroke)
            {
               case SPECIFIED_WIDTH:
                  stroke = new BasicStroke(ps.strokeWidth);
                  break;
               case DOTTED:
                  stroke = dotStroke;
                  break;
               case DOTTED_SMALL:
                  stroke = dotSmallStroke;
                  break;
               case DOTTED_SMALL_THIN:
                  stroke = dotSmallThinStroke;
                  break;
               case DOTTED_LARGE:
                  stroke = dotLargeStroke;
                  break;
            }
            
            g2.setStroke(stroke);
            break;
            
         case SET_FONT:
            g2.setFont(ps.font.font);
            break;
            
         case SET_FONT_STYLE:
            Font f = g2.getFont().deriveFont(convertFontStyle(ps.style));
            if (ps.style.isUnderline())
            {
               f = f.deriveFont(SwingGuiDriver.UNDERLINE_FONT);
            }
            g2.setFont(f);
            break;
            
         case TRANSLATE_PUSH:
         case TRANSLATE_POP:
            g2.translate(ps.x, ps.y);
            translateX += ps.x;
            translateY += ps.y;
            break;
            
         case CLEAR_AREA:
         case CLEAR:
            if (ps.color != null)
            {
               g2.setColor(SwingHelper.fromColorRgb(ps.color));
            }
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
            
         case CLIP:
            Shape prevClip = g2.getClip();
            if (prevClip != null)
            {
               clippings.push(prevClip);
            }

            // debugging aid: set debug to true to outline the clipped area
            boolean debug = false;
            if (debug)
            {
               g2.setColor(Color.green);
            }

            Area newArea = new Area();

            // intersect the clip in the paint structure with the currently active clip
            for (NativeRectangle nr : ps.clip)
            {
               newArea.add(new Area(new Rectangle(nr.left(), nr.top(), nr.width(), nr.height())));
            }

            if (prevClip != null)
            {
               if (prevClip instanceof Area)
               {
                  newArea.intersect((Area) prevClip);
               }
               else
               {
                  newArea.intersect(new Area(prevClip));
               }
            }

            if (debug)
            {
               g2.draw(newArea);
            }

            g2.setClip(newArea);
            break;
            
         case NO_CLIP:
            // restore the previous clip if any
            Shape clip = clippings.isEmpty() ? null : clippings.pop();
            g2.setClip(clip);
            break;
            
         case DRAW_HIGHLIGHT:
            drawHighlightWorker(ps.x, ps.y, ps.width, ps.height, ps.color);
            break;
         case CLEAR_HIGHLIGHT:
            clearHighlightWorker();
            break;
         case SET_TITLE:
            if (window instanceof JFrame)
            {
               ((JFrame) window).setTitle(ps.title);
            }
            else if (window instanceof JDialog)
            {
               ((JDialog) window).setTitle(ps.title);
            }
            break;
         case SET_ICON:
            window.setIconImage(ps.img.getImage());
            break;
         case SET_DESKTOP_BGCOLOR:
            // not needed in the Swing driver
            break;
         default:
            break;
      }
   }
   
   /**
    * Returns bounds of a physical screen the top-level window this instance 
    * represents is positioned on.
    * The bounds are in virtual device coordinate system. For more information
    * about the virtual device coordinate system see {@link TopLevelWindow}.
    * <p>
    * For single-display setups the top-left corner of the returned rectangle is always 
    * 0,0. For multiple-display setups it can be 0,0 or positive.
    * 
    * @return   The physical screen bounds.
    */
   @Override
   public NativeRectangle getDisplayBounds()
   {
      synchronized(lock())
      {
         Rectangle displayBounds = window.getGraphicsConfiguration().getBounds(); 
         return new NativeRectangle(new NativePoint(displayBounds.x, displayBounds.y), 
                                    new NativeDimension(displayBounds.width, displayBounds.height));
      }
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * @return   The area in pixels.
    */
   @Override
   public NativeRectangle getDisplayWorkArea()
   {
      synchronized(lock())
      {
         Insets is = window.getToolkit().getScreenInsets(window.getGraphicsConfiguration());
         return new NativeRectangle(is.top, is.left, is.bottom, is.right);
      }
   }
   
   /**
    * Render the given paint operation.  Any graphics context or other resources needed must
    * have already been setup/stored by the caller.
    */
   @Override
   public void flush()
   {
      graphics.drawImage(screenBuffer, screenBufferOffset.x, screenBufferOffset.y, null);
   }
   
   /**
    * Gets the instanve of the helper to work with buffered images.
    * 
    * @return   See above.
    */
   @Override
   public BufferedImageDrawHelper getBufferedImageDrawHelper()
   {
      return drawHelper;
   }
   
   /**
    * Obtain the lock object that can be used to synchronize drawing operations.
    *
    * @return   The shared lock object.
    */
   @Override
   protected Object lock()
   {
      return pane.getTreeLock();
   }
   
   /**
    * Obtain the containing window for the terminal simulator.
    */
   Window getContainingWindow()
   {
      return window;
   }
   
   /**
    * Obtain the content pane for the terminal simulator.
    */
   ContentPane getContentPane()
   {
      return pane;
   }
   
   /**
    * Move this window to top of the z-order, at the OS level. The to-top request is enqueued
    * to be dispatched by Swing's UI thread.
    *
    * @param    focus
    *           <code>true</code> to force the window to request the focus.
    */
   void moveToTop(boolean focus)
   {
      // execute in the Swing UI thread to make the move-to-top requests serialized
      SwingUtilities.invokeLater(() -> doMoveToTop(focus));
   }
   
   /**
    * Move this window to the bottom of the z-order, at the OS level.
    */
   void moveToBottom()
   {
      SwingUtilities.invokeLater(() -> window.toBack());
   }
   
   /**
    * The method returns <code>true</code> if the swing window is active.
    * 
    * @return   See above.
    */
   boolean isWindowActive()
   {
      synchronized(lock())
      {
         return window.isActive();
      }
   }

   /**
    * Move this window to top of the z-order, at the OS level.
    *
    * @param    focus
    *           <code>true</code> to force the window to request the focus.
    */
   private void doMoveToTop(boolean focus)
   {
      if (!readyToFocus && focus)
      {
         pendingFocus = true;
         return;
      }
      
      window.toFront();
      
      if (focus)
      {
         window.requestFocus();
      }
   }
   
   /**
    * Get the metrics of the given font.  If the {@link #g2 graphics} have been obtained, call the
    * {@link #getFontMetrics} and the graphics object.  Otherwise, use the simulator's graphics.
    *  
    * @param    font
    *           The font for which the font metrics are to be obtained.
    *           
    * @return   The font metrics.
    */
   private FontMetrics getNativeFontMetrics(Font font)
   {
      synchronized (lock())
      {
         return (g2 == null) ? pane.getFontMetrics(font) : g2.getFontMetrics(font);
      }
   }

   /**
    * Resize the screen buffers so that they will have the new width/height.
    * <p>
    * This will be a no-op if either the width or the height is not a positive number.
    * 
    * @param   width
    *          The new width.
    * @param   height
    *          The new height.
    */
   private void resizeScreenBuffers(int width, int height)
   {
      if (width <= 0 || height <= 0)
      {
         return;
      }
      
      synchronized (lock())
      {
         Dimension screenSize = getVirtualScreenSize();
         
         // Make the buffer size the maximum of twice the screen size,
         // this overlap allows for smoother window move - the buffer
         // must not be reposition on every move operation when window
         // size is larger than the buffer.
         // Note that repositionScreenBuffers() assumes that the buffer
         // is always larger than the screen.
         int bufW = Math.min(width, 2*screenSize.width);
         int bufH = Math.min(height, 2*screenSize.height);

         BufferedImage resizedImage = new BufferedImage(bufW, bufH, BufferedImage.TYPE_INT_ARGB);
         
         if (screenBuffer != null)
         {
            Graphics2D g = resizedImage.createGraphics();
            g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g.drawImage(screenBuffer, 0, 0, null);
            g.dispose();
         }
         
         screenBuffer = resizedImage;
         
         if (g2 != null)
         {
            g2.dispose();
         }
         
         g2 = screenBuffer.createGraphics();
         g2.translate(translateX, translateY);
         g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
         
         // apply the current clip if any to the graphics device
         if (!clippings.isEmpty())
         {
            g2.clip(clippings.peek());
         }
         
         if (screenBufferOffset != null)
         {
            g2.translate(-screenBufferOffset.x, -screenBufferOffset.y);
         }
   
         lock().notifyAll();
         
         repaintWindow();
      }
   }
   
   /**
    * Repaint a parent window.
    */
   private void repaintWindow()
   {
      ThinClient.getInstance().invokeLaterOS(() -> {
         TopLevelWindow<?> window = (TopLevelWindow<?>) WindowManager.findWindow(windowId);
         if (window != null)
         {
            ThinClient.getInstance().eventDrawingBracket(window, window::repaint);
         }
      });
   }
   
   /**
    * Adjust the screen buffer to fully cover the window's visible area.
    * <p>
    * This method implements an optimization to limit the maximum required
    * screen buffer size to support the theoretical maximum size of 4GL windows.
    * The screen buffer is allocated only for the window's visible area plus
    * some overlap in case the window is bigger than the screen size. 
    * Because the buffer is allocated only for a portion of the window area
    * its origin must be adjusted when the window is moved.
    * The overlap allows the window to be user-moved smoothly, the buffer must be
    * readjusted and window redrawn only when the buffer boundary is hit.
    * <p>
    * Why is this optimization needed? The screen buffer is implemented with 
    * {@link BufferedImage}, the class however has a limit of the maximum buffer 
    * size. The limit is defined as {@code width * height <= Integer.MAX_VALUE}.
    * The second reason, the memory requirement for larger window sizes is
    * huge. For example before this optimization the required Java client heap 
    * size was about 9 GB to resize the window to half the size limit of 
    * {@link BufferedImage}.
    * <p>
    */
   private void repositionScreenBuffers()
   {
      if (screenBuffer == null)
      {
         return;
      }
      
      synchronized (lock())
      {
         Point newBufferOffset = new Point();
         Dimension screenSize = getVirtualScreenSize();
         
         Point wndLoc = window.getLocationOnScreen();
         Rectangle wndRect = new Rectangle(wndLoc, window.getSize());
         
         Dimension bufferDim = new Dimension(screenBuffer.getWidth(), screenBuffer.getHeight());
         Rectangle bufferRect = new Rectangle(screenBufferOffset, bufferDim);
         
         Rectangle viewport = new Rectangle(Math.max(-wndRect.x, 0),
                                            Math.max(-wndRect.y, 0), 
                                            Math.min(wndRect.width, screenSize.width),
                                            Math.min(wndRect.height, screenSize.height));
         
         // check whether current viewport fits fully into the buffer,
         // if so no need to reposition the buffer and return
         if (viewport.union(bufferRect).equals(bufferRect))
         {
            return;
         }
         
         // if buffer width/height is smaller then screen width/height 
         // no overlap is needed - all the screen data will fit into the buffer
         int bufOverlapHoriz = Math.max(0, bufferDim.width - screenSize.width) / 2; 
         int bufOverlapVert = Math.max(0, bufferDim.height - screenSize.height) / 2; 
         
         newBufferOffset.x = Math.max(0, -wndRect.x - bufOverlapHoriz);
         newBufferOffset.y = Math.max(0, -wndRect.y - bufOverlapVert);
         
         // translate back the previous screen buffer offset
         if (screenBufferOffset != null)
         {
            g2.translate(screenBufferOffset.x, screenBufferOffset.y);
         }
         
         // translate the new screen buffer offset
         g2.translate(-newBufferOffset.x, -newBufferOffset.y);
         screenBufferOffset = newBufferOffset;
      }
      
      repaintWindow();
   }

   /**
    * Convert the generic font style to a Swing font style.
    *
    * @param    style
    *           The font style to convert.
    *
    * @return   The Java Swing font style.
    */
   private int convertFontStyle(FontStyle style)
   {
      int result = 0;
      
      switch (style)
      {
         case BOLD:
         case UNDERLINE_BOLD:
            result = Font.BOLD;
            break;
         case ITALIC:
         case UNDERLINE_ITALIC:
            result = Font.ITALIC;
            break;
         case BOLD_ITALIC:
         case UNDERLINE_BOLD_ITALIC:
            result = Font.BOLD | Font.ITALIC;
            break;
         case PLAIN:
            result = Font.PLAIN;
            break;
      }
      
      return result;
   }
   
   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * Once this API is finished, the graphic's font is restored to its state as before the call.
    *  
    * @param    ps
    *           The structure with the drawing details.
    */
   private void drawStringScaled(PaintStructure<Font, BufferedImage> ps)
   {
      Font origFont = g2.getFont();
      
      FontMetrics fm = getNativeFontMetrics(origFont);
      
      int txtWidth = fm.stringWidth(ps.text);
      LineMetrics lm = fm.getLineMetrics(ps.text, g2);
      int txtHeight = (int) lm.getHeight();
      
      if (txtWidth == ps.width && txtHeight == ps.height)
      {
         drawString(ps);
         return;
      }

      double scaleX = (double) ps.width / (double) txtWidth;
      double scaleY = (double) ps.height / (double) txtHeight;

      Font scaled = origFont.deriveFont(AffineTransform.getScaleInstance(scaleX, scaleY));
      g2.setFont(scaled);
      try
      {
         fm = getNativeFontMetrics(scaled);
         lm = fm.getLineMetrics(ps.text, g2);

         drawString(ps);
      }
      finally
      {
         g2.setFont(origFont);
      }
   }

   /**
    * Draw a string using the current font.
    *  
    * @param    ps
    *           The structure with the drawing details.
    */
   private void drawString(PaintStructure<Font, BufferedImage> ps)
   {
      int y = ps.y;
      if (ps.centered)
      {
         FontMetrics fm = g2.getFontMetrics();
         y = ps.y + fm.getAscent() / 2;
      }
      
      g2.drawString(ps.text, ps.x, y);
   }

   /**
    * The method performs a layout operation on the supplied text and 
    * returns the resulting paragraph height while maintaining the
    * supplied maximum width.
    * <p>
    * Note that the algorithm for text layout is compatible with 
    * {@link PaintPrimitives#DRAW_PARAGRAPH} and so
    * can be used in conjunction with this method. 
    * 
    * @param    g2
    *           Graphics2D reference.
    * @param    text
    *           The text to measure.
    * @param    font
    *           The font to use during layout.
    * @param    maxWidth
    *           The maximum paragraph width.
    * @param    draw
    *           When <code>false</code> the method performs
    *           only the layout part and does not render
    *           the text on the output device.
    * @param    x
    *           The drawing x position, ignored when draw set to 
    *           <code>false</code>.
    * @param    y          
    *           The drawing y position, ignored when draw set to 
    *           <code>false</code>.
    */
   private int layoutParagraphWorker(Graphics2D g2, 
                                     String text, 
                                     FontDetails<Font> font, 
                                     int maxWidth, 
                                     boolean draw, 
                                     int x, 
                                     int y)
   {
      // setup font
      FontRenderContext fontContext = g2.getFontRenderContext();
      
      // setup model
      AttributedString aText = new AttributedString(text);
      aText.addAttribute(TextAttribute.FONT, font.font);
      aText.addAttribute(TextAttribute.FOREGROUND, g2.getColor());
      
      AttributedCharacterIterator aIt = aText.getIterator();
      int textEnd = aIt.getEndIndex();
      
      // setup layout
      LineBreakMeasurer lineMeasurer = new LineBreakMeasurer(aIt, fontContext);
      float drawY = 0;
      
      while (lineMeasurer.getPosition() < textEnd)
      {
         TextLayout layout = lineMeasurer.nextLayout(maxWidth);
         
         // TODO: text direction
         float drawX = 0;
         drawY += layout.getAscent();
         
         if (draw)
         {
            layout.draw(g2, x + drawX, y + drawY);
         }
         
         // increment y by height of this layout
         drawY += layout.getDescent() + layout.getLeading();
      }
      
      return (int) drawY;
   }
      
   /**
    * This is the worker method for drawing a highlighted rectangular area of the specified
    * color and bounds. The color is alpha-blended with the content of the screen buffer.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           area width
    * @param    height
    *           area height
    * @param    color
    *           the highlight color
    */
   private void drawHighlightWorker(int x, int y, int width, int height, ColorRgb color)
   {
      // keep the original composite and color to restore later
      Composite prevComposite = g2.getComposite();
      Color     prevColor     = g2.getColor();
      Color     newColor      = SwingHelper.fromColorRgb(color);
      
      try 
      {
         // save the area for later to allow clear
         highlightRestoreBuffer = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
         Graphics2D gh = highlightRestoreBuffer.createGraphics();
         int sx = x - translateX;
         int sy = y - translateY;
         gh.drawImage(screenBuffer, 0, 0, width, height, sx, sy, sx+width, sy+height, pane);
         
         // create and set new composite for the highlight
         AlphaComposite ac = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f);
         g2.setComposite(ac);
   
         // draw the highlight
         g2.setColor(newColor);
         g2.fillRect(x - translateX, y - translateY, width, height);
         highlightX = x;
         highlightY = y;
      }
      finally
      {
         g2.setColor(prevColor);
         g2.setComposite(prevComposite);
      }
   }
   
   /**
    * This is the worker method for restoring the screen area previously highlighted by
    * {@link #drawHighlightWorker(int, int, int, int, ColorRgb)}.
    */
   private void clearHighlightWorker()
   {
      if (highlightRestoreBuffer == null)
      {
         // nothing to restore
         return;
      }
      
      int w = highlightRestoreBuffer.getWidth();
      int h = highlightRestoreBuffer.getHeight();
      
      int dx = highlightX - translateX;
      int dy = highlightY - translateY;
      
      Composite prevComposite = g2.getComposite();
      AlphaComposite ac = AlphaComposite.getInstance(AlphaComposite.SRC, 1f);
      g2.setComposite(ac);
      g2.drawImage(highlightRestoreBuffer, dx, dy, dx+w, dy+h, 0, 0, w, h, pane);
      g2.setComposite(prevComposite);
      
      highlightRestoreBuffer = null;
   }
   
   /**
    * The method invokes the supplied runnable on the Swing event dispatch 
    * thread (EDT) by calling {@link SwingUtilities#invokeAndWait(Runnable)}.
    * If the method is already being executed on the Swing EDT, then
    * the supplied runnable is directly invoked. 
    * 
    * @param   code
    *          A valid {@link Runnable} reference.
    */
   private static void safeInvokeAndWait(Runnable code)
   {
      if (SwingUtilities.isEventDispatchThread())
      {
         code.run();
      }
      else
      {
         try
         {
            SwingUtilities.invokeAndWait(code);
         }
         catch (InvocationTargetException | InterruptedException e)
         {
            throw new RuntimeException(e);
         }
      }
   }
   
   /**
    * Retruns the screen virtual dimension. This is the sum of dimensions of all
    * the individual screens.
    * 
    * @return  See above.
    */
   private Dimension getVirtualScreenSize()
   {
      int width = 0;
      int height = 0;
      
      GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
      GraphicsDevice[] gs = ge.getScreenDevices();
      
      for (GraphicsDevice curr : gs)
      {
        DisplayMode mode = curr.getDisplayMode();
        width += mode.getWidth();
        height += mode.getHeight();
      }
      
      // TODO: cache the dimension
      return new Dimension(width, height);
   }
}