/*
** Module   : GuiWebPageHandler.java
** Abstract : GUI-specific page handler
**
** Copyright (c) 2015-2017 Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150325 Created initial version patterned after the ChuiWebPageHandler.
** 002 GES 20150710 Implemented additional configuration options.
** 003 SBI 20160427 Added graphicsCached parameter to turn off the graphics cache for testing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

import org.eclipse.jetty.server.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.client.driver.web.*;

/**
 * This code primarily exists to parse the HTML text and substitute place holders, but all the
 * rest of the handler functionality is in the parent class where it can be shared. 
 */
public class GuiWebPageHandler
extends WebPageHandler
{
   /** Configuration. */
   private BootstrapConfig config;
   
   /**
    * Constructor.
    *
    * @param    config
    *           Configuration parameters.
    * @param    root
    *           Prefix for all URLs relating to this page.
    * @param    skel
    *           Fully qualified resource name of the HTML file which should be used.
    */
   public GuiWebPageHandler(BootstrapConfig config, String root, String skel)
   {
      super(config, root, skel);
      this.config = config;
   }
   
   /**
    * Process the given string and make any replacements of parameter
    * values as needed. The replacement placeholder follow the JEE 
    * expression language (EL) format for immediate evaluation. 
    * <p>
    * Example: ${expression} 
    *
    * @param    base
    *           The details and state the request.
    * @param    text
    *           The text line that may need modifications.
    *
    * @return   The new text if a modification occurred or the original text
    *           if no modification was needed.
    */
   protected String handleReplacements(Request base, String text)
   {
      text = super.handleReplacements(base, text);
      
      if (text.contains("${client_title}"))
      {
         text = text.replace("${client_title}", "GUI");
      }
      
      if (text.contains("${isGui}"))
      {
         text = text.replace("${isGui}", "true");
      }
      
      if (text.contains("${height_name}"))
      {
         text = text.replace("${height_name}", "height");
      }                           
      
      if (text.contains("${height_val}"))
      {
         int height = config.getInt("client", "gui", "desktopHeight", 768);
         text = text.replace("${height_val}", String.valueOf(height));
      }                           
      
      if (text.contains("${width_name}"))
      {
         text = text.replace("${width_name}", "width");
      }
      
      if (text.contains("${width_val}"))
      {
         int width = config.getInt("client", "gui", "desktopWidth", 1024);
         text = text.replace("${width_val}", String.valueOf(width));
      }
      
      if (text.contains("${font.name}"))
      {
         // TODO: where should we get the font cfg from?
         text = text.replace("${font.name}", "unknown");
      }
      
      if (text.contains("${font.size}"))
      {
         // TODO: where should we get the font cfg from?
         text = text.replace("${font.size}", "12");
      }

      if (text.contains("${font.color}"))
      {
         // TODO: where should we get the font cfg from?
         text = text.replace("${font.color}", htmlColor("#FFA500"));
      }

      if (text.contains("${font.background}"))
      {
         // TODO: where should we get the font cfg from?
         text = text.replace("${font.background}", htmlColor("#000000"));
      }
      
      // the clipboard is always enabled in GUI, we don't honor the directory setting
      if (text.contains("${clipboard.enabled}"))
      {
         text = text.replace("${clipboard.enabled}", "true");
      }
      
      // the taskbar will be used if this option is set true, the default is true
      if (text.contains("${taskbar.enabled}"))
      {
         boolean taskbar = config.getBoolean("client", "gui", "taskbar", true);
         text = text.replace("${taskbar.enabled}", taskbar ? "true" : "false");
      }
      
      if (text.contains("${graphicsCached}"))
      {
         boolean graphicsCached = config.getBoolean("client", "gui", "graphicsCached", true);
         text = text.replace("${graphicsCached}", graphicsCached ? "true" : "false");
      }
      
      return text;
   }
}