/*
** Module   : p2j.canvas_renderer.js
** Abstract : GUI-specific canvas renderer
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI 20150905 Created the canvas renderer in order to enclose all canvas drawing in one class.
** 002 CA  20150925 Fixed clipping: it must intersect the previous posted clip rectangle before 
**                  posting it. 
**                  Fixed putImageData: this is not aware of the current clipped rectangle, so
**                  it must check if the pixel is outside of the clip, before drawing it.
** 003 CA  20151015 Fixed a bug in H002 (clip() was not intersecting correctly).
** 004 SBI 20151125 Move clipping logic from putImageData to drawing primitives: drawImage and
**                  drawLineSegment.
** 005 SBI 20151205 Changed to adjust clipping regions according to a new origin.
**     SBI 20151216 Changed to work around that the Chrome's implementation of getImageData
**                  rounds function parameters.
**     SBI 20160111 Rewrite the copy of the target image taking into account the Chrome's
**                  getImageData implementation.
**     CA  20160129 Removed the "ctx" parameter from the CanvasRenderer APIs.  Added a global 
**                  cache where widget states are saved.
**     SBI 20160213 Reduced putImageData invocations for horizontal and vertical lines.
**     SBI 20160322 Changed to work around if the window canvas has a zero width or height.
** 006 SBI 20160415 Replaced logFormatted(...) by log(...).
** 007 SBI 20160519 Added the Safari to the exception image processing.
** 008 HC  20161018 Fixed drawing of rastered images, right and bottom edges were trimmed.
** 009 CA  20170106 Avoid resizing the canvas if its size hasn't changed.
** 010 HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
**                  overlapping regions.
** 011 OM  20170320 Fixed centered text rendering.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Creates the canvas renderer.
 * 
 * @param    {HtmlCanvasElement} canvas
 *           The html canvas element.
 * @param    {CanvasRenderingContext2D} ctx
 *           The canvas 2D graphics context on which to draw.
 * @param    {LinesStroke} strokesManager
 *           The stroke manager
 * @param    {Object} fontsManager
 *           The p2j fonts manager.
 * @param    {Object} logger
 *           The p2j logger.
 */
function CanvasRenderer(canvas, ctx, strokesManager, fontsManager, logger)
{
   this.canvas = canvas;
   this.ctx = ctx;
   this.strokesManager = strokesManager;
   this.logger = logger;
   this.fontsManager = fontsManager;

   /** Color scaling factor, compatible with Java2D. */
   var SCALE_FACTOR = 0.7;

   /** Minimum color value for darkening/lightening. */
   var MIN_COLOR_VALUE = Math.floor(1.0 / (1.0 - SCALE_FACTOR));

   /** Common data structure used in direct drawing operations. */
   this.pixel = this.ctx.createImageData(1, 1);
   
   /** Internal portion of the pixel data structure used in direct drawing operations. */
   this.pixData = this.pixel.data;
   
   // force alpha to 100% opaque
   this.pixData[3] = 0xFF;
   
   // eliminate anti-aliasing of images
   this.ctx.imageSmoothingEnabled = false;
   
   /** Tracks the current cumulative coordinate translation that is in effect. */
   this.origin = { "x" : 0, "y" : 0 };

   /** Stack of graphics context state, used for managing drawing, clipping and transforms. */
   this.state = new Array();

   /** Current color for direct drawing (in the form [ red, green, blue ]). */
   this.rawColor = new Array(0, 0, 0);

   /** The current window stroke style id. */
   this.strokeStyleId = strokesManager.getDefaultStrokeStyle();
   
   /** The current window stroke width. */
   this.strokeWidth = 1;
   
   /** This stack will contain the drawing operation recordings. */
   this.recordings = new Array();

   //sets the default stroke for the canvas
   this.strokesManager.applyStrokeToCanvas(this.ctx, this.strokeStyleId, this.strokeWidth);
   
   /** True value indicates XOR composite operation will be applied. */
   this.compositeModeOn = false;
   
   /** The screen capture used by drawing primitives */
   this.screenCapture = this.ctx.getImageData(0, 0, 1, 1);
   
   // force all drawing to be inside pixel "cells" instead of "stradling" between two cells
   // which is the default; this eliminates much of the worst of the anti-aliasing and
   // positioning problems inherent in canvas operations; this will also save off the
   // current state of the graphics context so that it can be restored later 
   this.translate(0.5, 0.5);
   
   /**
    * Convert a standard fillStyle or strokeStyle color into an array of [ R, G, B ] values.
    *
    * @param    {String} spec
    *           Input string which must be in the form "#RRGGBB" where RR, GG and BB are 2-digit
    *           hexadecimal numbers.
    *
    * @return   The parsed raw color array of integers in [ R, G, B ] form.
    */
   this.parseColor = function(spec)
   {
      if (typeof spec !== "string")
      {
         p2j.logger.error("parseColor() spec must be a string");
         return null;                                 
      }
      
      if (spec.length !== 7)
      {
         p2j.logger.error("parseColor() spec must have length of 7");
         return null;
      }
      
      if (spec.charAt(0) !== "#")
      {
         p2j.logger.error("parseColor() spec must start with #");
         return null;
      }
      
      var r = parseInt(spec.substr(1, 2), 16);
      var g = parseInt(spec.substr(3, 2), 16);
      var b = parseInt(spec.substr(5, 2), 16);

      if (r === NaN || g === NaN || b === NaN)
      {
         return null;
      }
      
      return [ r, g, b ];
   };

   /**
    * Test if the given argument is an array of 3 integer values.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Boolean} <code>true</code> if the argument is in [ R, G, B ] form.
    */
   this.validateColor = function(color)
   {
      if (!(color instanceof Array))
      {
         return false;
      }
      
      if (color.length !== 3)
      {
         return false;
      }
      
      if (typeof color[0] !== "number" ||
          typeof color[1] !== "number" ||
          typeof color[2] !== "number")
      {
         return false;
      }
      
      if (color[0] < 0 || color[0] > 255 ||
          color[1] < 0 || color[1] > 255 ||
          color[2] < 0 || color[2] > 255)
      {
         return false;
      }
      
      return true;
   };

   /**
    * Convert the given array of 3 integer values into an HTML color string.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {String} A valid HTML color string.
    */
   this.createColorString = function(color)
   {
      if (!this.validateColor(color))
      {
         return null;
      }
                                         
      return "rgb(" + color[0].toString() + ", "
                    + color[1].toString() + ", "
                    + color[2].toString() + ")";
   };

   /**
    * Convert the given array of RGB values into a lighter color in a manner that is compatible 
    * with Java2D.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Number[]} A lighter shade of the same color.
    */
   this.lightenColor = function(color)
   {
      if (!this.validateColor(color))
      {
         return null;
      }
      
      // handle black directly, the result is NOT very much different but this is how it is
      // done in Java
      if (color[0] == 0 && color[1] == 0 && color[2] == 0)
         return [ MIN_COLOR_VALUE, MIN_COLOR_VALUE, MIN_COLOR_VALUE ];
      
      // force inputs to a minimim color value if between 0 and that minimum color value
      // (yes, this is how Java does it)
      var rIn = (color[0] > 0 && color[0] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[0];
      var gIn = (color[1] > 0 && color[1] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[1];
      var bIn = (color[2] > 0 && color[2] < MIN_COLOR_VALUE) ? MIN_COLOR_VALUE : color[2];
      
      // scale it
      var r = Math.min(Math.floor(rIn / SCALE_FACTOR), 255);
      var g = Math.min(Math.floor(gIn / SCALE_FACTOR), 255);
      var b = Math.min(Math.floor(bIn / SCALE_FACTOR), 255);
      
      return [ r, g, b ];
   };

   /**
    * Convert the given array of RGB values into a darker color in a manner that is compatible 
    * with Java2D.
    *
    * @param    {Number[]} color
    *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
    *
    * @return   {Number[]} A darker shade of the same color.
    */
   this.darkenColor = function(color)
   {
      if (!this.validateColor(color))
      {
         return null;
      }
      
      var r = Math.max(0, Math.floor(color[0] * SCALE_FACTOR));
      var g = Math.max(0, Math.floor(color[1] * SCALE_FACTOR));
      var b = Math.max(0, Math.floor(color[2] * SCALE_FACTOR));
      
      return [ r, g, b ];
   };
}

/**
 * Clear the canvas.
 */
CanvasRenderer.prototype.clear = function()
{
   this.ctx.save();
   this.ctx.translate(-this.origin.x, -this.origin.y);
   this.ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
   this.ctx.translate(this.origin.x, this.origin.y);
   this.ctx.restore();
}

/**
 * Mark the start of recording the drawing operations.
 * 
 * @param    x
 *           The X origin at the time recording was started. 
 * @param    y
 *           The Y origin at the time recording was started.
 * @param    width
 *           The width of the resulted image, which will be captured and cached.
 * @param    height
 *           The height of the resulted image, which will be captured and cached.
 * @param    md5
 *           The hash key used to cache the resulted image.
 */
CanvasRenderer.prototype.startRecording = function(x, y, width, height, md5)
{
   this.recordings.push(
   {
      x      : x,
      y      : y,
      width  : width,
      height : height,
      md5    : md5
   });
}

/**
 * Mark the stop of the drawing operations recording.  This will save the resulted image in the
 * p2j.screen.globalCache variable.
 */
CanvasRenderer.prototype.stopRecording = function()
{
   var recording = this.recordings.pop();
   var imgData = this.ctx.getImageData(recording.x, 
                                       recording.y, 
                                       recording.width, 
                                       recording.height);

   p2j.screen.globalCache[recording.md5] = imgData;
}

/**
 * Draw the image with the specified hash from the global cache, at the specified location.
 * 
 * @param    x
 *           The X where to draw the cached image.
 * @param    y
 *           The Y where to draw the cached image.
 */
CanvasRenderer.prototype.drawCached = function(x, y, md5)
{
   var cachedDraw = p2j.screen.globalCache[md5];
   
   if (cachedDraw == undefined)
   {
      p2j.screen.error("Trying to draw unknown cached image for md5 " + md5 + "!");
      console.trace();
      return;
   }
   
   // draw the image at the specified coordinates (these are screen coordinates, not translated)
   this.ctx.save();
   this.ctx.globalAlpha = 1;
   this.ctx.putImageData(cachedDraw, x, y);
   this.ctx.restore();

   console.log("Using global cached draw " + md5 + " posted at coordinates x:" + x + ";y:" +y);
}

/**
 * Fill or stroke the current path on the given context, first using the current path as a
 * clipping region.  This ensures that the drawing of the path will not extend outside of the
 * path by a half-pixel (which it would otherwise). The state of the context is saved on
 * entry and restored after the drawing operation, so there will be no residual clipping
 * region when this function returns.
 *
 * @param    {Boolean} fill
 *           <code>true</code> to fill the drawn rectangle with the given color.
 */
CanvasRenderer.prototype.renderClosedPath = function(fill)
{
   this.ctx.save();
   this.ctx.clip();
   
   if (fill)
   {
      this.ctx.fill();
   }
   else
   {
      this.ctx.stroke();
   }
   
   this.ctx.restore();
};

/**
 * Adjusts the origin for pixel drawing on the canvas and transforms the stacked clipping regions
 * by the given x and y factors in order to be consistent with the new coordinates origin.
 *
 * @param    {Number} x
 *           The translate along X-axis.
 * @param    {Number} y
 *           The translate along Y-axis.
 */
CanvasRenderer.prototype.adjustOrigin = function(x, y)
{
   this.origin.x += x;
   this.origin.y += y;
   
   var last = 0;
   while (last < this.state.length)
   {
      var s = this.state[last];
      
      if (s.clip != undefined)
      {
         for (var i = 0; i < s.clip.length; i++)
         {
            s.clip[i].x -= x;
            s.clip[i].y -= y;
         }
      }
      last = last + 1;
   }
};

/**
 * Directly draw the specified pixel in the given color.
 *
 * @param    {Number} x
 *           X coordinate of the pixel to be drawn.
 * @param    {Number} y
 *           Y coordinate of the pixel to be drawn.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 */
CanvasRenderer.prototype.drawPixel = function(x, y, color)
{
   this.pixData[0] = color[0];
   this.pixData[1] = color[1];
   this.pixData[2] = color[2];
   
   this.putImageData(this.pixel, x, y);
};

/**
 * If the pixel location is outside of the last clipped rectangle return false, otherwise
 * return true.
 *
 * @param    {Number} x
 *           X coordinate of the pixel to be checked.
 * @param    {Number} y
 *           Y coordinate of the pixel to be checked.
 * 
 * @return   {Boolean}
 *           The boolean value indicates the given pixel is in the last clipping region.
 */
CanvasRenderer.prototype.isInClipRegion = function(x, y)
{
   var lastClip = this.lastClippedRegion(true);
   if (lastClip == undefined)
   {
      return true;
   }
   
   for (var i = 0; i < lastClip.length; i++)
   {
      var top = lastClip[i].y;
      var bottom = lastClip[i].y + lastClip[i].height - 1;
      var left = lastClip[i].x;
      var right = lastClip[i].x + lastClip[i].width - 1;

      if (y >= top && y <= bottom && x >= left && x <= right)
      {
         return true;
      }
   }

   return false;
};

/**
 * Calculates the intersection of the input region {left: xa, top: ya, right: xb, bottom: yb}
 * with the last clipping region.
 * 
 * @param    {Number} xa
 *           X coordinate of the left corner of the input region.
 * @param    {Number} ya
 *           Y coordinate of the top corner of the input region.
 * @param    {Number} xb
 *           X coordinate of the right corner of the input region.
 * @param    {Number} yb
 *           Y coordinate of the bottom corner of the input region.
 * 
 * @return   {left: xa, top: ya, right: xb, bottom: yb}
 *           The intersection with the last clipping region or null if it is empty.
 */
CanvasRenderer.prototype.clipRegion = function(xa, ya, xb, yb)
{
   var result = [];
   var lastClip = this.lastClippedRegion(true);
   if (lastClip)
   {
      for (var i = 0; i < lastClip.length; i++)
      {
         var top    = lastClip[i].y;
         var bottom = lastClip[i].y + Math.max(lastClip[i].height - 1, 0);
         var left   = lastClip[i].x;
         var right  = lastClip[i].x + Math.max(lastClip[i].width - 1, 0);

         if (yb < top || ya > bottom || xb < left || xa > right)
         {
            continue;
         }

         result.push(
            {
               left: Math.max(left, xa),
               top: Math.max(top, ya),
               right: Math.min(right, xb),
               bottom: Math.min(bottom, yb)
            });
      }
   }
   else
   {
      result.push({left: xa, top: ya, right: xb, bottom: yb});
   }

   return result.length > 0 ? result : null;
};

/**
 * Renders the given image on the canvas at the point (x,y) given by its coordinates in the
 * canvas associated coordinate system.
 *
 * @param    {ImageData} imageData
 *           The imageData to be drawn.
 * @param    {Number} x
 *           X coordinate of the pixel to be drawn.
 * @param    {Number} y
 *           Y coordinate of the pixel to be drawn.
 */
CanvasRenderer.prototype.putImageData = function(imageData, x, y)
{
   if (this.compositeModeOn)
   {
      this.applyCompositeMode(imageData, x, y);
   }
   try
   {
      this.ctx.putImageData(imageData, x, y);
   }
   catch(ex)
   {
      console.log("putImageData(imageData=" + (imageData ? imageData.width : "") + "x" +
                  (imageData ? imageData.height : "") +
                  " data.byteLength=" + (imageData ? imageData.data.byteLength : "") +
                  ", x=" + x + ", y=" + y + ")" + " compositeModeOn=" + this.compositeModeOn);
      console.log(ex.stack);
   }
};

/**
 * Transforms a given image according to the current composite mode. The input image is changed.
 *
 * @param    {ImageData} imageData
 *           The input image to be transformed.
 * @param    {Number} x
 *           X coordinate of the left image corner.
 * @param    {Number} y
 *           Y coordinate of the top image corner.
 */
CanvasRenderer.prototype.applyCompositeMode = function(imageData, x, y)
{
   if (this.compositeModeOn)
   {
      var w = imageData.width;
      var h = imageData.height;
      var screenShot = this.ctx.getImageData(x, y, w, h);
      var pixelsInBytes = 4 * w * h;
      for (var i = 0; i < pixelsInBytes; i += 4)
      {
         imageData[i]     = imageData[i]     ^ screenShot[i];
         imageData[i + 1] = imageData[i + 1] ^ screenShot[i + 1];
         imageData[i + 2] = imageData[i + 2] ^ screenShot[i + 2];
         imageData[i + 3] = imageData[i + 3] ^ screenShot[i + 3] ^ 255;
      }
   }
};

/**
 * Reset composite mode.
 *
 */
CanvasRenderer.prototype.resetComposite = function()
{
   this.compositeModeOn = false;
   this.ctx.globalCompositeOperation = "source-over";
};

/**
 * Set XOR composite mode.
 */
CanvasRenderer.prototype.setXORComposite = function()
{
   this.compositeModeOn = true;
   this.ctx.globalCompositeOperation = "xor";
};

/**
 * Creates a new ImageData object with the specified width and height. The new object is filled
 * with transparent black pixels.
 * 
 * @param    {Number} width
 *           The image width in pixels.
 * @param    {Number} height
 *           The image height in pixels.
 * 
 * @return   {ImageData}
 *           The ImageData object with the specified width and height.
 */
CanvasRenderer.prototype.createImageData = function(width, height)
{
   return this.ctx.createImageData(width, height);
};

/**
 * Draw a line from (x1, y1) to (x2, y2) inclusive, using the given stroke renderer.
 *
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * @param    {StrokeRenderer} strokeRenderer
 *           The current stroke renderer.
 */
CanvasRenderer.prototype.drawLineSegment = function(x1, y1, x2, y2, strokeRenderer)
{
   x1 = x1 + this.origin.x;
   x2 = x2 + this.origin.x;
   
   y1 = y1 + this.origin.y;
   y2 = y2 + this.origin.y;

   // intersect with clipping region
   // and normalize this region, thus (d[0],d[1]) defines the left top cone point and (d[2],d[3]) 
   // defines the right bottom cone point of the clipping region 
   var d = this.cutLine(x1, y1, x2, y2);
   if (d === null)
   {
      return;
   }
   for (var i = 0; i < d.length; i++)
   {
      x1 = d[i][0];
      y1 = d[i][1];
      x2 = d[i][2];
      y2 = d[i][3];

      // to check if this line  (x1, y1, x2, y2) is vertical, or horizontal, or sloped
      var dx = x2 - x1;
      var dy = y2 - y1;

      // vertical and horizontal lines are handled by the current stroke renderer (optimization)
      // if dx == 0, then a line (x1, y1, x2, y2) is a vertical
      // if dy == 0, then a line (x1, y1, x2, y2) is a horizontal
      if (dx === 0 || dy === 0)
      {
         // to draw a line as an image for the default stroke and widen stroke styles
         // we have x2 >= x1 and y2 >= y1 and dx or dy equals zero.
         strokeRenderer.strokeLine(x1, y1, x2, y2);//optimization
      }
      else
      {
         this.drawSlopedLineSegment(x1, y1, x2, y2, strokeRenderer);
      }
   }
};

/**
 * Draw a sloped line from (x1, y1) to (x2, y2) inclusive, using the given stroke renderer.
 * <p>
 * This is an implementation of the well known 50+ year old Bresenham algorithm.
 *
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * @param    {StrokeRenderer} strokeRenderer
 *           The current stroke renderer.
 */
CanvasRenderer.prototype.drawSlopedLineSegment = function(x1, y1, x2, y2, strokeRenderer)
{
   var dx = Math.abs(x2 - x1);
   var dy = Math.abs(y2 - y1);

   var xIncr = x1 < x2 ? 1 : -1;
   var yIncr = y1 < y2 ? 1 : -1;
   
   dy = -dy;
   
   var err = dx + dy;
   var err2;
   
   var x = x1;
   var y = y1;
   var tooMany = 0;
   
   while (true)
   {
      // use a pixel drawing method to stroke the point
      strokeRenderer.strokePoint(x, y);
      
      tooMany++;
      
      if (x == x2 && y == y2)
      {
         break;
      }
      else
      {
         if (tooMany > 10000)
         {
            this.logger.log("Bresenham is out of control for x1 = %d, y1 = %d,"
                     + " x2 = %d, y2 = %d and current x = %d, y = %d!", x1, y1, x2, y2, x, y);
            console.trace();
            break;
         }
      }
         
      err2 = err * 2;
      
      if (err2 >= dy)
      {
         err += dy;
         x   += xIncr; 
      }
      
      if (err2 <= dx)
      {
         err += dx;
         y   += yIncr;
      }
   }
}

/**
 * 
 * Draw a line in the given color from (x1, y1) to (x2, y2) inclusive, using direct drawing and
 * add its pixels path to the end of the provided pixels path.
 * <p>
 * This is an implementation of the well known 50+ year old Bresenham algorithm.
 *
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 */
CanvasRenderer.prototype.drawLine = function(x1, y1, x2, y2, color)
{
   var strokeRenderer = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, color);
   strokeRenderer.beginStroke();
   this.drawLineSegment(x1, y1, x2, y2, strokeRenderer);
}

/**
 * Stroke a line in the given color from (x1, y1) to (x2, y2) inclusive.
 * <p>
 * This is an implementation of the well known 50+ year old Bresenham algorithm.
 *
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 */
CanvasRenderer.prototype.strokeLineSegment = function(x1, y1, x2, y2, color)
{
   var strokeRenderer = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, color);
   strokeRenderer.beginStroke();
   this.drawLineSegment(x1, y1, x2, y2, strokeRenderer);
}

/**
 * Draw a rectangle in the given color and dimensions with the line drawing being overwritten
 * using direct drawing to eliminate the negative/unwanted effects of anti-aliasing.
 *
 * @param    {Number} x
 *           X coordinate of the top left corner.
 * @param    {Number} y
 *           Y coordinate of the top left corner.
 * @param    {Number} width
 *           Width of the rectangle in pixels.
 * @param    {Number} height
 *           Height of the rectangle in pixels.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 *           This will be used for the border lines and when in fill mode, also the interior.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the drawn rectangle with the given color.
 */
CanvasRenderer.prototype.drawRect = function(x, y, width, height, color, fill)
{
   // filled rectangles draw 1 pixel smaller in 2 dimensions than stroked rectangles
   var inset = 0;
   // use vector operations for the interior of the rectangle
   if (fill)
   {
      inset = 1;
      this.ctx.beginPath();
      this.ctx.rect(x, y, width - inset, height - inset);
      this.renderClosedPath(fill);
   }
   
   // now overdraw the stroked portion to eliminate anti-aliasing, we draw in a
   // clockwise direction (since we are not using paths, this is not strictly necessary)
   var renderer = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, color);
   renderer.beginStroke();
   this.drawLineSegment(x, y, x + width - inset, y, renderer);
   this.drawLineSegment(x + width - inset, y + 1, x + width - inset, y + height - inset, renderer);
   this.drawLineSegment(x + width - 1 - inset, y + height - inset, x + 1, y + height - inset, renderer);
   this.drawLineSegment(x, y + height - inset, x, y + 1, renderer);// close the path
};

/**
 * Stroke a rectangle in the given color and dimensions with the line drawing being overwritten
 * using direct drawing to eliminate the negative/unwanted effects of anti-aliasing.
 *
 * @param    {Number} x
 *           X coordinate of the top left corner.
 * @param    {Number} y
 *           Y coordinate of the top left corner.
 * @param    {Number} width
 *           Width of the rectangle in pixels.
 * @param    {Number} height
 *           Height of the rectangle in pixels.
 * @param    {Number[]} color
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 *           This will be used for the border lines and when in fill mode, also the interior.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the drawn rectangle with the given color.
 */
CanvasRenderer.prototype.strokeRect = function(x, y, width, height, color, fill)
{
   // filled rectangles draw 1 pixel smaller in 2 dimensions than stroked rectangles
   var inset = 0;
   // use vector operations for the interior of the rectangle
   if (fill)
   {
      inset = 1;
      this.ctx.beginPath();
      this.ctx.rect(x, y, width - inset, height - inset);
      this.renderClosedPath(fill);
   }
   
   // now overdraw the stroked portion to eliminate anti-aliasing, we draw in a
   // clockwise direction (since we are not using paths, this is not strictly necessary)
   var renderer = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, color);
   renderer.beginStroke();
   this.drawLineSegment(x, y, x + width - inset, y, renderer);
   this.drawLineSegment(x + width - inset, y + 1, x + width - inset, y + height - inset,
         renderer);
   this.drawLineSegment(x + width - 1 - inset, y + height - inset, x + 1, y + height - inset,
         renderer);
   this.drawLineSegment(x, y + height - inset, x, y + 1, renderer);// close the path
   
};

/**
 * It draws an image on the canvas.
 * 
 * @param    {Number} x
 *           X coordinate of the top left corner.
 * @param    {Number} y
 *           Y coordinate of the top left corner.
 * @param    {Number} width
 *           Width of the image in pixels.
 * @param    {Number} height
 *           Height of the image in pixels.
 * @param    {Array} imgData
 *           The image data array contains the encoded image started from the given offset.
 * @param    {Number} imgDataOffset
 *           The offset of the encoded image data.
 */
CanvasRenderer.prototype.drawImage = function(x, y, width, height, imgData, imgDataOffset)
{
   // the image origin
   var xo = this.origin.x + x;
   var yo = this.origin.y + y;
   
   // the image region : {left: xa, top: ya, right: xb, bottom: yb}
   var xa = xo;
   var ya = yo;
   var xb = xa + width;
   var yb = ya + height;
   
   var clipped = this.clipRegion(xa, ya, xb, yb);
   if (clipped == null)
   {
      return;
   }
   /**
    * Pixels must be returned as non-premultiplied alpha values according to:
    * http://www.w3.org/TR/2dcontext/#pixel-manipulation
    */
   var img = this.ctx.getImageData(xo, yo, width, height);
   var data = img.data;
   if (p2j.isChrome || p2j.isSafari)
   {
      this.copyImage(img, data, imgData, width, height, imgDataOffset);
   }
   else
   {
      // optimized version of copy image
      var pixelsInBytes = width * height * 4;
      var offset = imgDataOffset;
      for (var p = 0; p < pixelsInBytes; p += 4, offset += 4)
      {
         var alpha = imgData[offset + 3];
         var beta = 256 - alpha;
         if (alpha == 255)
         {
            /** server sends non-premultiplied image data */
            data[p]     = imgData[offset];
            data[p + 1] = imgData[offset + 1];
            data[p + 2] = imgData[offset + 2];
            data[p + 3] = imgData[offset + 3];
         }
         else if (alpha > 0)
         {
            /** server sends non-premultiplied image data */
            data[p]     = (imgData[offset]     * alpha + data[p]     * beta) >> 8;
            data[p + 1] = (imgData[offset + 1] * alpha + data[p + 1] * beta) >> 8;
            data[p + 2] = (imgData[offset + 2] * alpha + data[p + 2] * beta) >> 8;
            data[p + 3] = (imgData[offset + 3] * alpha + data[p + 3] * beta) >> 8;
         };
      };
   }
   if (this.compositeModeOn)
   {
      this.applyCompositeMode(img, xo, yo);
   }

   for (var i = 0; i < clipped.length; i++)
   {
      xa = clipped[i].left;
      ya = clipped[i].top;
      xb = clipped[i].right;
      yb = clipped[i].bottom;

      var deltaX = xa - xo;
      var deltaY = ya - yo;

      this.ctx.putImageData(img, xo, yo, deltaX, deltaY, xb - xa + 1, yb - ya + 1);
   }
};

/**
 * Copies a downloaded image encodded as an rgba byte array provided with its width and its height
 * to the screen image given by its own width and height taking into account the coppied image
 * alpha channel and places it at the zero coordinates relative to the screen image.
 * 
 * @param    {ImageData} img
 *           The screen image data
 * @param    {Uint8ClampedArray} data
 *           The screen image rgba byte array. 
 * @param    {Array} imgData
 *           The rgba byte array of the coppied image.
 * @param    {Number} width
 *           The coppied image width.
 * @param    {Number} height
 *           The coppied image height.
 * @param    {Number} imgDataOffset
 *           The coppied image data offset from the 0-index of its data array, imgData.
 * 
 * @return   {Uint8ClampedArray}
 *           Returns the modified screen image data array with the target image.
 *  
 */
CanvasRenderer.prototype.copyImage = function(img, data, imgData, width, height, imgDataOffset)
{
   var awidth  = img.width;
   var aheight = img.height;
   var x = 0;
   var y = 0;
   var pixelsInBytes = awidth * aheight * 4;
   var offset = imgDataOffset;
   for (var p = 0; p < pixelsInBytes; offset += 4)
   {
      var alpha = imgData[offset + 3];
      var beta = 256 - alpha;
      if (alpha == 255)
      {
         /** server sends non-premultiplied image data */
         data[p]     = imgData[offset];
         data[p + 1] = imgData[offset + 1];
         data[p + 2] = imgData[offset + 2];
         data[p + 3] = imgData[offset + 3];
      }
      else if (alpha > 0)
      {
         /** server sends non-premultiplied image data */
         data[p]     = (imgData[offset]     * alpha + data[p]     * beta) >> 8;
         data[p + 1] = (imgData[offset + 1] * alpha + data[p + 1] * beta) >> 8;
         data[p + 2] = (imgData[offset + 2] * alpha + data[p + 2] * beta) >> 8;
         data[p + 3] = (imgData[offset + 3] * alpha + data[p + 3] * beta) >> 8;
      };
      x++;
      // p = (y * awidth + x) * 4
      // p1 - p0 = [(y1 - y0) * awidth + (x1 - x0)] * 4 = 4 or 4 * (awidth - width) + 4 
      if (x == width)
      {
         y++;
         x = 0;
         p += ((awidth - width + 1) * 4);
      }
      else
      {
         p += 4; 
      }
   };
   
   return data;
};

/**
 * Draw a rounded rectangle with the given context, dimensions and fill.
 *
 * @param    {Number} x
 *           X coordinate of the top left of the rectangle (if a square corner was placed
 *           there).
 * @param    {Number} y
 *           Y coordinate of the top left of the rectangle (if a square corner was placed
 *           there).
 * @param    {Number} width
 *           Width of the rectangle.
 * @param    {Number} height
 *           Height of the rectangle.
 * @param    {Number} diameter
 *           Diameter of the arc that defines the rounded corner.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 */
CanvasRenderer.prototype.drawRoundRect = function(x, y, width, height, diameter, color, fill)
{
   var radius = diameter / 2;
   
   // arc control points
   var northWestX = x;
   var northWestY = y;
   var northEastX = x + width;
   var northEastY = y;
   var southWestX = x;
   var southWestY = y + height;
   var southEastX = x + width;
   var southEastY = y + height;
   
   // line end points                                       
   var topLeftX     = northWestX + radius;
   var topLeftY     = northWestY;
   var topRightX    = northEastX - radius;
   var topRightY    = northEastY;
   var rightUpX     = northEastX;
   var rightUpY     = northEastY + radius;
   var rightDownX   = southEastX;
   var rightDownY   = southEastY - radius;
   var bottomRightX = southEastX - radius;
   var bottomRightY = southEastY;
   var bottomLeftX  = southWestX + radius;
   var bottomLeftY  = southWestY;
   var leftDownX    = southWestX;
   var leftDownY    = southWestY - radius;
   var leftUpX      = northWestX;
   var leftUpY      = northWestY + radius;
   
   this.ctx.beginPath();
   this.ctx.moveTo(topRightX, topRightY);
   this.ctx.quadraticCurveTo(northEastX, northEastY, rightUpX, rightUpY);
   this.ctx.lineTo(rightDownX, rightDownY);
   this.ctx.quadraticCurveTo(southEastX, southEastY, bottomRightX, bottomRightY);
   this.ctx.lineTo(bottomLeftX, bottomLeftY);
   this.ctx.quadraticCurveTo(southWestX, southWestY, leftDownX, leftDownY);
   this.ctx.lineTo(leftUpX, leftUpY);
   this.ctx.quadraticCurveTo(northWestX, northWestY, topLeftX, topLeftY);
   this.ctx.closePath();
   
   this.renderClosedPath(fill);
   
   // overdraw the anti-aliased line segments              
   var renderer = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, color);
   renderer.beginStroke();
   this.drawLineSegment(topLeftX, topLeftY, topRightX, topRightY, renderer);
   this.drawLineSegment(rightUpX, rightUpY, rightDownX, rightDownY, renderer);
   this.drawLineSegment(bottomLeftX, bottomLeftY, bottomRightX, bottomRightY, renderer);
   this.drawLineSegment(leftUpX, leftUpY, leftDownX, leftDownY, renderer);
};

/**
 * Draw a slightly shaded "3D" rectangle with the given context, dimensions and fill. In
 * raised mode the rectangle is drawn with a lighter color used for the top and left borders
 * and any fill that is needed, while the bottom and right borders are a darker version of
 * the given color.  If not raised, a darker shade of the given color is used for the top
 * and left borders and any requested fill, while a lighter shade is used for the bottom and
 * right borders.
 *
 * @param    {Number} x
 *           X coordinate of the top left of the rectangle.
 * @param    {Number} y
 *           Y coordinate of the top left of the rectangle.
 * @param    {Number} width
 *           Width of the rectangle in pixels.
 * @param    {Number} height
 *           Height of the rectangle in pixels.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 * @param    {Boolean} raise
 *           <code>true</code> to draw in raised mode.
 */
CanvasRenderer.prototype.draw3DRect = function(x, y, width, height, color, fill, raised)
{
   // save off our colors
   var oriFillColor   = this.ctx.fillStyle;
   var oriStrokeColor = this.ctx.strokeStyle;
   
   var darker  = this.darkenColor(color);
   var lighter = this.lightenColor(color);
   
   // save off our state so that we can clear our clipping region on exit
   this.ctx.save();
   
   // the Java implementation draws 1 pixel larger in each dimension
   width  = width + 1;
   height = height + 1;
   
   // setup our clipping region, this is needed because the line width (1 pixel wide) is
   // drawn half on one side of hte path and half on the other side; this makes paths a
   // half pixel bigger than the caller would expect; by defining the same path as a
   // clipping region, the half pixel width that is outside the path is never output
   this.ctx.beginPath();
   this.ctx.rect(x, y, width, height);
   this.ctx.clip();
   
   var inset = 0;
   
   var current = raised ? color : darker;
   
   // fill the rectangle with the original color (raised mode) or the darker color otherwise
   if (fill)
   {
      inset = 1;
      this.ctx.fillStyle = this.createColorString(current);
      this.ctx.beginPath();
      this.ctx.rect(x, y, width - 2, height - 2);
      this.renderClosedPath(fill);
   }
   
   current = raised ? lighter : darker;
   var renderer1 = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, current);
   renderer1.beginStroke();
   // draw the top and left sides in the contrasting color (raised mode) or the darker color
   // otherwise
   this.drawLineSegment(x, y, x + width - 2 - inset, y, renderer1);
   this.drawLineSegment(x, y + 1, x, y + height - 1 - inset, renderer1);

   // draw the bottom and right sides in the darker color (raised mode) or the contrasting
   // color otherwise
   current = raised ? darker : lighter;
   var renderer2 = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, current);
   renderer2.beginStroke();
   this.drawLineSegment(x + 1, y + height - 1 - inset, x + width - 1 - inset,
         y + height - 1 - inset, renderer2);
   this.drawLineSegment(x + width - 1 - inset, y, x + width - 1 - inset, y + height - 2,
         renderer2);
   
   // clear the clipping region
   this.ctx.restore();
   
   // restore our colors
   this.ctx.fillStyle   = oriFillColor;
   this.ctx.strokeStyle = oriStrokeColor;
};

/**
 * Draw a closed polygon with the given context, vertices and fill.
 *
 * @param    {Number} xPoints
 *           X coordinates of the points.
 * @param    {Number} yPoints
 *           Y coordinates of the points.
 * @param    {Number} num
 *           Number of points to draw.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 */
CanvasRenderer.prototype.drawPolygon = function(xPoints, yPoints, num, color, fill)
{
   var i;

   // use vector operations for the interior of the polygon
   if (fill)
   {
      this.ctx.beginPath();
      this.ctx.moveTo(xPoints[0], yPoints[0]);
      for (i = 1; i < num; i++)
      {
         this.ctx.lineTo(xPoints[i], yPoints[i]);
      }
      this.ctx.closePath();
      this.renderClosedPath(fill);
   }
   var renderer = this.strokesManager.getStrokePathRenderer(this, this.strokeStyleId,
         this.strokeWidth, color);
   renderer.beginStroke();

   // now overdraw the stroked portion
   for (i = 0; i < (num - 1); i++)
   {
      this.drawLineSegment(xPoints[i], yPoints[i], xPoints[i + 1], yPoints[i + 1], renderer);
   }
   // close the path
   this.drawLineSegment(xPoints[num - 1], yPoints[num - 1], xPoints[0], yPoints[0], renderer);
};

/**
 * Stroke a closed polygon with the given context, vertices and fill.
 *
 * @param    {Number} xPoints
 *           X coordinates of the points.
 * @param    {Number} yPoints
 *           Y coordinates of the points.
 * @param    {Number} num
 *           Number of points to draw.
 * @param    {Number[]} color  
 *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
 * @param    {Boolean} fill
 *           <code>true</code> to fill the rectangle with the current color, otherwise just
 *           stroke the rectangle.
 */
CanvasRenderer.prototype.strokePolygon = function(xPoints, yPoints, num, color, fill)
{
   this.drawPolygon(xPoints, yPoints, num, color, fill);
};

/**
 * Set the color for future graphics operations on this window.
 *
 * @param    {Number[]} or {String} color
 *           This must be either a valid array of 3 integers between 0 and 255 inclusive OR it
 *           an HTML color specification in "#RRGGBB" form.
 */
CanvasRenderer.prototype.setColor = function(color)
{
   var colorSpec;

   if (color instanceof Array)
   {
      if (this.validateColor(color))
      {
         colorSpec = this.createColorString(color);
         
         // store the color in a form that can be used for direct drawing
         this.rawColor = [ color[0], color[1], color[2] ];
      }
      else
      {
         this.logger.error("Invalid array form of color specification.");
         return;
      }
   }
   else
   {
      if (typeof color === "string")                  
      {
         colorSpec = color;
         
         // store the color in a form that can be used for direct drawing
         this.rawColor = this.parseColor(color);
      }
      else
      {
         this.logger.error("Invalid string form of color specification.");
         return;
      }
   }
   
   // change the canvas colors
   this.ctx.fillStyle   = colorSpec;
   this.ctx.strokeStyle = colorSpec;
};

/**
 * Sets the stroke style.
 *                              
 * @param    {Number} strokeStyleId
 *           The predefined stroke style.
 */
CanvasRenderer.prototype.setStrokeStyleId = function(strokeStyleId)
{
   this.strokeStyleId = strokeStyleId;
};

/**
 * Sets the stroke line width.
 *                              
 * @param    {Number} width
 *           The stroke line width.
 */
CanvasRenderer.prototype.setStrokeWidth = function(width)
{
   this.strokeWidth = width;    
};


/**
 * Sets the line stroke.
 * 
 * @param    {Number} strokeStyleId
 *           The predefined stroke style id.
 * @param    {Number} width
 *           The stroke line width.
 */
CanvasRenderer.prototype.setLineStroke = function(strokeStyleId, width)
{                               
   this.setStrokeStyleId(strokeStyleId);
   this.setStrokeWidth(width);
   this.strokesManager.applyStrokeToCanvas(this.ctx, strokeStyleId, width);
};

/**
 * Add the given area to the current clipping region.
 * 
 * @param    {Array} area
 *           The array of objects representing multiple rectangular areas.
 *           The object defines properties x, y, width, height.
 */
CanvasRenderer.prototype.clip = function(area)
{
    var newClip = [];

   // find the previous clipping rectangle
   var prevClip = this.lastClippedRegion(false);
   if (prevClip)
   {
      for (var i = 0; i < prevClip.length; i++)
      {
          var prevR =
          {
             "top"    : prevClip[i].y,
             "left"   : prevClip[i].x,
             "bottom" : prevClip[i].y + prevClip[i].height - 1,
             "right"  : prevClip[i].x + prevClip[i].width - 1
          };

          for (var j = 0; j < area.length; j++)
          {
             var newR =
             {
                "top"    : area[j].y,
                "left"   : area[j].x,
                "bottom" : area[j].y + area[j].height - 1,
                "right"  : area[j].x + area[j].width - 1
             };

             var intersect = undefined;
             if (!(prevR.top    > newR.bottom ||
                   prevR.bottom < newR.top    ||
                   prevR.left   > newR.right  ||
                   prevR.right  < newR.left))
             {
                intersect =
                {
                   "top"    : (prevR.top    > newR.top    ? prevR.top    : newR.top),
                   "left"   : (prevR.left   > newR.left   ? prevR.left   : newR.left),
                   "bottom" : (prevR.bottom < newR.bottom ? prevR.bottom : newR.bottom),
                   "right"  : (prevR.right  < newR.right  ? prevR.right  : newR.right),
                };

                newClip.push
                ({
                   x : intersect.left,
                   y : intersect.top,
                   width : intersect.right - intersect.left + 1,
                   height : intersect.bottom - intersect.top + 1
                });
             }
          }
      }
   }
   else
   {
      newClip = area;
   }
   this.save({ clip : newClip });
   this.ctx.beginPath();

   for (var i = 0; i < newClip.length; i++)
   {
      this.ctx.rect(newClip[i].x, newClip[i].y, newClip[i].width, newClip[i].height);
   }

   this.ctx.clip();
   /*
   if (false)
   {
      // enable this code to show the clipped rectangles
      this.ctx.beginPath();
      var f1 = this.ctx.fillStyle;
      var f2 = this.ctx.strokeStyle;
      this.ctx.strokeStyle = "#FF0000";
      for (var i = 0; i < newClip.length; i++)
      {
         this.ctx.rect(newClip[i].x, newClip[i].y, newClip[i].width, newClip[i].height);
      }
      this.ctx.stroke();

      this.ctx.fillStyle = f1;
      this.ctx.strokeStyle = f2;
   }
   */
};

/**
 * Find the last clipping area.
 * <p>
 * If the <code>toScreen</code> flat is true, it will translate the clipping area top-left
 * corner to absolute coordinates.  Otherwise, it will translate it to relative coordinates to
 * the current origin, by applying all the translation operations after the clip area was set.
 * 
 * @param    {boolean} toScreen
 *           Flag indicating if the return clipped rectangle needs to be translated to screen 
 *           coordinates.
 *
 * @return   {Array} The last clip rectangles or undefined if no clipping area exists.
 */
CanvasRenderer.prototype.lastClippedRegion = function(toScreen)
{
   var last = this.state.length - 1;
   var prevClip = undefined;
   while (last >= 0)
   {
      var s = this.state[last];
      if (s.clip != undefined)
      {
         prevClip = [];
         var i = s.clip.length;
         while(i--)
         {
            prevClip.push(
               {
                  "x"     : s.clip[i].x,
                  "y"     : s.clip[i].y,
                  "width" : s.clip[i].width,
                  "height": s.clip[i].height
               });
         }
         break;
      }
      
      last = last - 1;
   }
   
   if (prevClip != undefined)
   {
      if (toScreen)
      {
         for (var i = 0; i < prevClip.length; i++)
         {
             // translate to absolute coordinates
             prevClip[i].x += this.origin.x;
             prevClip[i].y += this.origin.y;
         }
      }
   }

   return prevClip;
};

/**
 * Calculates the intersection of the given line segment with the last clipping region.
 * 
 * @param    {Number} x1
 *           X coordinate of the starting pixel to be drawn.
 * @param    {Number} y1
 *           Y coordinate of the starting pixel to be drawn.
 * @param    {Number} x2
 *           X coordinate of the ending pixel to be drawn.
 * @param    {Number} y2
 *           Y coordinate of the ending pixel to be drawn.
 * 
 * @return   {Array}
 *           The line intersection with the last clipping region in the form of this array 
 *           [new_x1, new_y1, new_x2, new_y2].
 */
CanvasRenderer.prototype.cutLine = function (x1, y1, x2, y2)
{
   //xa, ya xb, yb is normalized region covered this line segment
   var xa;
   var xb;
   var incx = (x1 < x2);
   if (incx)
   {
      xa = x1;
      xb = x2;
   }
   else
   {
      xa = x2;
      xb = x1;
   }
   var ya;
   var yb;
   var incy = (y1 < y2);
   if (incy)
   {
      ya = y1;
      yb = y2;
   }
   else
   {
      ya = y2;
      yb = y1;
   }
   // {left : xa, top : ya, right : xb, bottom : yb}
   var clipped = this.clipRegion(xa, ya, xb, yb);
   if (clipped == null)
   {
      return null;
   }

   var result = [];
   for (var i = 0; i < clipped.length; i++)
   {
      xa = clipped[i].left;
      ya = clipped[i].top;
      xb = clipped[i].right;
      yb = clipped[i].bottom;

      // check that it is vertical or horizontal
      if ((x1 == x2) || (y1 == y2))
      {
         result.push([xa, ya, xb, yb]);
      }
      else
      {
         var fxa;
         var fxb;

         fxa = (y1 * (x2 - xa) / (x2 - x1)) + (y2 * (xa - x1) / (x2 - x1));
         fxb = (y1 * (x2 - xb) / (x2 - x1)) + (y2 * (xb - x1) / (x2 - x1));

         result.push([xa, fxa, xb, fxb]);
      }
   }

   return result.length > 0 ? result : null;
}

/**
 * Remove the most recently added rectangle from the current clipping region. The graphics
 * context state is popped to restore the previously used values.
 */
CanvasRenderer.prototype.unclip = function()
{
   this.restore();
};

/**
 * Save graphics context state to the stack.
 *
 * @param    {Object} additional
 *           Caller provided name/value pairs to add to the state.
 */
CanvasRenderer.prototype.save = function(additional)
{
   var cur = 
   {
      fillstyle     : this.ctx.fillStyle,
      strokeStyle   : this.ctx.strokeStyle,
      strokeStyleId : this.strokeStyleId,
      strokeWidth   : this.strokeWidth,
      compositeModeOn : this.compositeModeOn
   };
   
   if (additional)
   {
      for (var prop in additional)
      {
         if (additional.hasOwnProperty(prop))
         {
            cur[prop] = additional[prop];
         }
      }
   }
   
   this.state.push(cur);
   
   // now force the context to push its state internally (only the context itself can save off
   // its accumulated clip region and its cumulative transforms)
   this.ctx.save();
};

/**
 * Restore the graphics context state to the state as stored on the top of the stack.
 * <p>
 * The state in our own custom state stack is not actually used for the restore, instead the
 * context's internal stack is used (ctx.restore()).  Our stack is used for other purposes
 * and we simply clear it here.
 */
CanvasRenderer.prototype.restore = function()
{
   this.state.pop();
   this.ctx.restore();
};

/**
 * Obtain the graphics context state as stored on the top of the stack.
 *
 * @return   The top of state stack object.
 */
CanvasRenderer.prototype.peek = function()
{
   var last = this.state.length - 1;
   
   return last >= 0 ? this.state[last] : {};
};

/**
 * Replay the saved graphics context state back into a newly reset graphics context. This
 * should only be used after a canvas resize operation.
 */
CanvasRenderer.prototype.replay = function()
{
   var len = this.state.length;
   
   for (var i = 0; i < len; i++)
   {
      var next = this.state[i];
      
      this.ctx.fillstyle   = next.fillStyle;
      this.ctx.strokeStyle = next.strokeStyle;
      this.setLineStroke(next.strokeStyleId, next.strokeWidth);
      this.compositeModeOn = next.compositeModeOn;
      var x;
      var y;
      var width;
      var height;
      
      if (next.translate)
      {
         x = next.translate.x;
         y = next.translate.y;
         
         // don't call the this.translate() method here because it modifies the stack that
         // we are simply trying to replay; we don't want the stack to be modified
         try
         {
            this.ctx.translate(x, y);
         }
         catch(ex)
         {
            console.log("this.ctx.translate(x=" + x + ", y=" + y + ")");
            console.log(ex);
            console.log(ex.stack);
         }
      }
   
      if (next.clip)
      {
         x      = next.clip.x;
         y      = next.clip.y;
         width  = next.clip.width;
         height = next.clip.height;                               
                                                                  
         // don't call the this.translate() method here because it modifies the stack that
         // we are simply trying to replay; we don't want the stack to be modified
         this.ctx.beginPath();
         this.ctx.rect(x, y, width, height);
         this.ctx.clip();
      }
   }
};

/**
 * Untranslate the graphics context by the given X and Y factors. This does a restore on
 * the previously saved context state.  It will error out if some sanity checks don't pass.
 * Make sure that calls to this function are always properly paired with equal and opposite
 * calls to translate().
 *
 * @param    {Number} x
 *           X dimension factor.
 * @param    {Number} y
 *           Y dimension factor.
 */
CanvasRenderer.prototype.untranslate = function(x, y)
{
   var top = this.peek();
   
   if (top.translate)
   {
      if (top.translate.x === -x && top.translate.y === -y)
      {
         // clear our state back to the version that doesn't have this translate
         this.restore();
         
         this.adjustOrigin(x, y);
      }
      else
      {
         this.logger.error("Mismatching translate pop values, expected (" +
                  top.translate.x.toString() + ", " + top.translate.y.toString() +
                  ") and found " + x.toString() + ", " + y.toString() + ").");
      }
   }
   else
   {
      this.logger.error("Out of order translate pop!");
   }
};

/**
 * Resize the canvas to the given width and height.
 *
 * @param    {Number} width
 *           The new canvas width in pixels.
 * @param    {Number} height
 *           The new canvas height in pixels.
 */
CanvasRenderer.prototype.resize = function(width, height)
{
   if (this.canvas.width == width && this.canvas.height == height)
   {
      return;
   }

   // save off the previously drawn content
   var oldPixels;
   if (this.canvas.width > 0 && this.canvas.height > 0)
   {
      oldPixels = this.ctx.getImageData(0, 0, this.canvas.width, this.canvas.height);
   }
   
   // this is expensive, but we really do need to change the overall size of the canvas 
   this.canvas.width  = width;
   this.canvas.height = height;
   
   // all of the current state including colors, line stroke, translation and clipping
   // must be reapplied to the context because changing the size resets everything
   this.replay();
                                
   // re-blit the previously drawn content
   if (oldPixels)
   {
      this.putImageData(oldPixels, 0, 0);
   }
};

/**
 * Translate the graphics context by the given X and Y factors.  This will also save the state
 * of the context before the translate is applied, so that the translate values can be
 * reversed later.
 *
 * @param    {Number} x
 *           X dimension factor.
 * @param    {Number} y
 *           Y dimension factor.
 */
CanvasRenderer.prototype.translate = function(x, y)
{
   this.save({ translate : { "x" : x, "y" : y } });
   
   this.adjustOrigin(x, y);
   try
   {
      this.ctx.translate(x, y);
   }
   catch(ex)
   {
      console.log("this.ctx.translate(x=" + x + ", y=" + y + ")");
      console.log(ex);
      console.log(ex.stack);
   }
};

/**
 * Draws a given text at the given (x,y) position.
 * 
 * @param    {Number} text
 *           The text to draw.
 * @param    {Number} x
 *           The X-axis of the coordinate for the text starting point.
 * @param    {Number} y
 *           The Y-axis of the coordinate for the text starting point or the height
 *           of the text rectangle from the most top screen position.
 * @param    {Boolean} centered
 *           Defines the text base line and the interpretation of y.
 *           If it is true, the text base line is set to 'middle' and interprets y as a height,
 *           otherwise the text base line is set to 'bottom'.
 */
CanvasRenderer.prototype.drawText = function(text, x, y, centered)
{
   if (centered)
   {
      // in this case, the "y" is the height in which it needs to be vertically centered
      this.ctx.textBaseline = 'middle';
      ++y;
   }
   else
   {
      this.ctx.textBaseline = 'alphabetic'; // matches the Swing's baseline
      --y;
   }
   --x;
   
   var cfont = p2j.screen.getCurrentFont();
   /*start debug*/
   //   var txtHeight = p2j.fonts.getTextHeight(cfont, text);
   //   var textWidth = p2j.fonts.getTextWidth(cfont, text);
   //   
   //   if (centered)
   //   {
   //      this.drawRect(x, y - txtHeight / 2, textWidth, txtHeight, new Array(255, 255, 0), false);
   //   }
   //   else 
   //   {
   //      this.drawRect(x, y - 3 * txtHeight / 4, textWidth, txtHeight, new Array(0, 255, 0), false);
   //   }
   /*end debug*/
   
   this.ctx.fillText(text, x, y);
   
   if (p2j.fonts.isUnderlined(cfont))
   {
      // if the style is underline, draw a line under the text
      var txtHeight = p2j.fonts.getTextHeight(cfont, text);
      var textWidth = p2j.fonts.getTextWidth(cfont, text);
      
      var x1 = x;
      var x2 = x + textWidth;
      if (centered)
      {
         y = y + Math.round(txtHeight / 2) - Math.round(txtHeight / 6) + 1;
      }
      
      this.strokeLineSegment(x1, y, x2, y, this.rawColor);
   }
}

/**
 * Draws a given text at the given (x,y) position fitted to the given text rectangle.
 * 
 * @param    {Number} currentFont
 *           The current font ID, in the js font-table.
 * @param    {Number} text
 *           The text to draw.
 * @param    {Number} x
 *           The X-axis of the coordinate for the text starting point.
 * @param    {Number} y
 *           The Y-axis of the coordinate for the text starting point or the height
 *           of the text rectangle from the most top screen position.
 * @param    {Boolean} centered
 *           Defines the text base line and the interpretation of y.
 *           If it is true, the text base line is set to 'middle' and interprets y as a height,
 *           otherwise the text base line is set to 'bottom'.
 * @param    {Number} lwidth
 *           The space width to be used to display the given text.
 *           The text is scaled to fit this width, squeezed or extended
 * @param    {Number} lheight
 *           The space height to be used to display the given text.
 *           The text is scaled to fit this height, squeezed or extended.
 */
CanvasRenderer.prototype.drawScaledText = function(currentFont, text, x, y, centered, lwidth, lheight)
{
   var textWidth = this.fontsManager.getTextWidth(currentFont, text);
   var textHeight = this.fontsManager.getTextHeight(currentFont, text);
   
   var widthScale = lwidth / textWidth;
   var heightScale = lheight / textHeight;
   
   // scale drawing context to computed width
   // this scales the desired structure.x as well so it needs adjusting
   this.ctx.save();
   this.ctx.scale(widthScale, heightScale);
   
   var scaleBackWidth = 1 / widthScale;
   var scaleBackHeight = 1 / heightScale;
   
   if (centered)
   {
      // in this case, the "y" is the height in which it needs to be vertically centered
      y = y / 2;

      this.ctx.textBaseline = 'middle';
   }
   else
   {
      this.ctx.textBaseline = 'bottom';
   }

   this.ctx.fillText(text, x * scaleBackWidth, y * scaleBackHeight);
   this.ctx.restore();
}