/*
** Module   : p2j.virtual_desktop.js
** Abstract : GUI-specific virtual desktop to create task bar and drop zones
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI 20150906 The virtual desktop to create task bar and drop zones.
** 002 SBI 20151020 Fix for disabling old task icons.
** 003 SBI 20151215 Fixed to take into an account that for Chrome an element style object has only
**                  the "getter" access function, to check IE if drag is starting.
**     CA  20160129 Removed the "ctx" parameter from the CanvasRenderer APIs.
** 004 HC  20160406 Overhaul of window-activation logic.
** 005 SBI 20160705 Changed to repaint the task bar after a task icon has been removed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Creates the virtual desktop with dragged task bar and two drop zones: the bottom and the right
 * according to this css style:
 * 
 *.right {
 *   position: fixed;
 *   top: 0px;
 *   right: 0px;
 *   height: 100%;
 *   width: 40px;
 *}
 *.bottom {
 *   position: fixed;
 *   bottom: 0px;
 *   height: 40px;
 *   width: 100%;
 *}
 *.virtualDesktop {
 *   position: fixed;
 *   left: 0px;
 *   top: 0px;
 *}
 *.canvas {
 *   position: relative;
 *   width: 100%;
 *   height: 100%;
 *}
 *.dock {
 *}
 * 
 * @param    {Function} taskIconOnClickCallback
 *           The 1-argument function to send the active window state request to the server.
 * @param    {Array} bg
 *           The background task bar color. It is represented by the 3-elements array
 *           filled with red, green and blue values from 0 to 255.
 * @param    {Array} fg
 *           The foreground task bar color for task icons. It is represented by the 3-elements array
 *           filled with red, green and blue values from 0 to 255.
 * @param    {Object} fontsManager
 *           The p2j fonts manager.
 * @param    {Object} logger
 *           The p2j global application logger.
 */
function VirtualDesktop(taskIconOnClickCallback, bg, fg, fontsManager, logger)
{
   /** task bar id*/
   var TASK_BAR_ID = "taskBar";
   
   /** drop zone class name*/
   var DOCK_CLASS_NAME = "dock";
   
   /** task bar side */
   var TASK_BAR_WIDTH  = 42;

   /** z-index is assigned to the drop zone to which the task bar is attached. */
   var MAX_Z_INDEX = 1000;
   /** z-index is assigned to all others drop zones */
   var MIN_Z_INDEX = 1;

   /** the parent element for the virtual desktop */
   var desktop = document.createElement("div");
   
   p2j.setStyleForElement(desktop, {position : "fixed", top : "0px", left : "0px"});
   
   var body = document.body;
   body.appendChild(desktop);
   
   var bottom = createBottomDropZone(DOCK_CLASS_NAME, TASK_BAR_WIDTH);
   var taskBarPeer = createTaskBarPeer(TASK_BAR_ID);
   
   bottom.appendChild(taskBarPeer);
   desktop.appendChild(bottom);
   desktop.appendChild(createRightDropZone(DOCK_CLASS_NAME, TASK_BAR_WIDTH));
   
   var taskBar = new TaskBar(taskBarPeer, bg, fg, fontsManager, logger);

   /**
    * Creates the bottom drop zone.
    * 
    * @param    {String} dockClassName
    *           The drop zone class.
    * @param    {Number} height
    *           The bottom drop zone height.
    * 
    * @return   {HtmlElement}
    *           The parent element for the bottom drop zone.
    */
   function createBottomDropZone(dockClassName, height)
   {
      var bottom = document.createElement("div");
      bottom.className = dockClassName;
      
      p2j.setStyleForElement(bottom,
                               {
                                  position : "fixed",
                                  left     : "0px",
                                  bottom   : "0px",
                                  height   : height + "px",
                                  width    : "100%",
                                  "z-index"  : MAX_Z_INDEX
                               });
      return bottom;
   };
   
   /**
    * Creates the right drop zone.
    * 
    * @param    {String} dockClassName
    *           The drop zone class.
    * @param    {Number} width
    *           The right drop zone width.
    * 
    * @return   {HtmlElement}
    *           The parent element for the right drop zone.
    */
   function createRightDropZone(dockClassName, width)
   {
      var right = document.createElement("div");
      right.className = dockClassName;
      p2j.setStyleForElement(right,
                              {
                                 position : "fixed",
                                 right    : "0px",
                                 top      : "0px",
                                 height   : "100%",
                                 width    : width + "px",
                                 "z-index"  : MIN_Z_INDEX
                              });
      return right;
   };
   
   /**
    * Creates the html element for the draggable task bar.
    * 
    * @param    {String} id
    *           The task bar peer element id.
    * 
    * @return   {HtmlElement}
    *           The task bar peer element.
    */
   function createTaskBarPeer(id)
   {
      var peer = document.createElement("div");
      peer.id = id;
      peer.tabIndex = 1;
      p2j.setStyleForElement(peer,
            {
               position : "relative",
               height   : "100%",
               width    : "100%",
               "z-index"  : MAX_Z_INDEX
            });
      peer.draggable = true;
      return peer;
   };
   
   /**
    * Gets the task bar.
    * 
    * @return   {TaskBar}
    *           The task bar.
    */
   this.getTaskBar = function()
   {
      return taskBar;
   };
   
   /**
    * Creates the task bar for the given peer element with the provided background and
    * foreground colors.
    * 
    * @param    {HtmlElement} peer
    *           The task bar peer element.
    * @param    {Array} bg
    *           The background task bar color. It is represented by the 3-elements array
    *           filled with red, green and blue values from 0 to 255.
    * @param    {Array} fg
    *           The foreground task bar color for task icons. It is represented by the 3-elements array
    *           filled with red, green and blue values from 0 to 255.
    * @param    {Object} fontsManager
    *           The p2j fonts manager.
    * @param    {Object} logger
    *           The p2j global application logger.
    */
   function TaskBar(peer, bg, fg, fontsManager, logger)
   {
      /** The dragged task bar panel opacity */
      var DRAGGED_OPACITY = 0.3;
      /** Top padding from the task bar border */
      var TOP_PADDING    = 2;
      /** Bottom padding from the task bar border */
      var BOTTOM_PADDING = 2;
      /** Left padding from the task bar border */
      var LEFT_PADDING  = 2;
      /** Right padding from the task bar border */
      var RIGHT_PADDING = 2;
      /** Margin around the icon */
      var ICON_MARGIN  = 1;
      /** Margin around the text */
      var TEXT_MARGIN  = 4;
      /** The task bar title. */
      var TASK_BAR_TITLE = "Task Bar";
      /** The default task icon widget dimension: 40 x32 */
      var TASK_ICON_WIDTH  = 40; 
      var TASK_ICON_HEIGHT = 32;
      /** The space between task icons. */
      var SHIFT = 2;
      
      /**
       * The task bar font id uses negative number in order to be different from fonts
       * created for p2j windows.
       */
      var TASK_BAR_FONT_ID = -1;

      var LIGHTEN_COLOR_INTENSE = 10;
      /** The task bar dragged panel */
      var dragged;
      
      /** active task icon index */
      var activeTaskIconIndex;
      
      var canvas = document.createElement('canvas');
      canvas.className="canvas";
      p2j.setStyleForElement(canvas,
            {
               height   : "100%",
               width    : "100%",
               "z-index"  : MAX_Z_INDEX
            });
      
      peer.appendChild(canvas);
      var ctx = canvas.getContext('2d', {alpha : true});
      var strokesManager = new LineStrokes();
      var canvasRenderer = new CanvasRenderer(canvas, ctx, strokesManager, fontsManager, logger);
      /** visible task icons */
      var taskIcons = [];
      /** hidden task icons */
      var hiddenTaskIcons = [];

      var viewLogButton = createViewLogButton(logger);
      
      var sysButtons = [viewLogButton];

      var foreground  = fg;
      var background  = bg;


      if (peer)
      {
         // events fired on the draggable target 
         peer.addEventListener("dragstart", startDragging);
         peer.addEventListener("dragend", dragEnded);
         //peer.addEventListener("drag", drag);
         setDefaultTooltip();
      }
      
      /**
       * The default task bar tooltip.
       */
      function setDefaultTooltip()
      {
         peer.title = TASK_BAR_TITLE;
      }
      /**
       * The special use case: view log.
       * 
       * @param    {Object} logger
       *           The application logger.
       * @return   {TaskIcon}
       *           The view log button.
       */
      function createViewLogButton(logger)
      {
         var taskIcon = new TaskIcon(1, 3 * TASK_ICON_WIDTH, TASK_ICON_HEIGHT, fg,
               "View Log Output", null, 0, 0);
         taskIcon.setActive(false);
         // Opens new log window.
         taskIcon.doClick = function()
         {
            logger.viewLog();
            // Change the view log button to inactive after 500 ms.
            window.setTimeout(
                  function()
                  {
                     taskIcon.setActive(false);
                     taskIcon.draw();
                  }, 500);
         }
         return taskIcon;
      }
      
      /**
       * Layouts the view log button on the task bar.
       * 
       * @param    {Number} width
       *           The task bar width in pixels.
       * @param    {Number} height
       *           The task bar height in pixels.
       * @param    {Boolean} horizontalLayout
       *           Indicates the task bar layout, true if the task bar is along a horizontal line, otherwise,
       *           it is along a vertical line.
       */
      function doViewLogButtonLayout(width, height, horizontalLayout)
      {
         var x;
         var y;
         if (horizontalLayout)
         {
            x = width - viewLogButton.width - RIGHT_PADDING;
            // to place on the centered horizontal line of the task bar.
            y = (height - viewLogButton.height) / 2;
         }
         else
         {
            // to place on the centered vertical line of the task bar.
            x = (width - viewLogButton.width) / 2;
            y = height - viewLogButton.height - BOTTOM_PADDING;
         }
         viewLogButton.setLocation(x, y);
         viewLogButton.draw();
      }
      
      /**
       * Layouts task icons according to the current task bar position:
       * the horizontal or the vertical.
       * 
       * @param    {Number} width
       *           The task bar width in pixels.
       * @param    {Number} height
       *           The task bar height in pixels.
       */
      function doLayout(width, height)
      {
         canvas.width = width;
         canvas.height = height;
         
         canvasRenderer.draw3DRect(0, 0, width, height, background, true, true);
         
         //check taskIcons is not empty
         if (taskIcons.length === 0)
         {
            return;
         }
         
         var horizontalLayout;
         
         if (width > height)
         {
            horizontalLayout = true;
         }
         else
         {
            horizontalLayout = false;
         }
         
         var unified = calculateTaskIconWidth(width, height, SHIFT, horizontalLayout);
         setUnifiedDimension(unified, TASK_ICON_HEIGHT, horizontalLayout);
        
         //find first visible
         var x;
         var y;
        
         if (horizontalLayout)
         {
            x = LEFT_PADDING;
            // to place on the centered horizontal line of the task bar.
            y = (height - taskIcons[0].height) / 2;
         }
         else
         {
            // to place on the centered vertical line of the task bar.
            x = (width - taskIcons[0].width) / 2;
            y = TOP_PADDING;
         }
         
         for (var i = 0; i < taskIcons.length; i++)
         {
            var taskIcon = taskIcons[i];
            taskIcon.setLocation(x, y);
            taskIcon.draw();
            if (horizontalLayout)
            {
               x += (SHIFT + taskIcon.width);
            }
            else
            {
               y += (SHIFT + taskIcon.height);
            }
         }
         doViewLogButtonLayout(width, height, horizontalLayout);
      };
      
      /**
       * Calculates the maximal pixel width of the task title for the current canvas font.
       * 
       * @param    {Array} widgets
       *           The array of TaskIcon widgets.
       * 
       * @return   {Number}
       *           The maximal pixel width of titles used for task widgets relative to the current
       *           canvas font.
       */
      function calculateMaximalTextWidth(widgets)
      {
         //check that it is not empty
         if (widgets.length === 0)
         {
            return 0;
         }
         var initialValue = fontsManager.getTextWidth(TASK_BAR_FONT_ID, widgets[0].title);
         return widgets.reduce(
               function(previousValue, currentValue, index, array)
               {
                  var textWidth  = fontsManager.getTextWidth(TASK_BAR_FONT_ID, array[index].title);
                  return Math.max(previousValue, textWidth);
               },
               initialValue);
      };
      
      /**
       * Calculates the maximal icon width from all icons for the task bar.
       * 
       * @param    {Array} widgets
       *           The array of TaskIcon widgets.
       * @param    {Boolean} horizontalLayout
       *           Indicates the task bar layout, true if the task bar is along a horizontal line, otherwise,
       *           it is along a vertical line.
       * 
       * @return   {Number}
       *           The maximal icon width from all icons for the task bar.
       */
      function calculateMaximalIconWidth(widgets, horizontalLayout)
      {
         //check that it is not empty
         if (widgets.length === 0)
         {
            return 0;
         }
         var initialValue;
         if (horizontalLayout)
         {
            initialValue = widgets[0].iconWidth;
         }
         else
         {
            initialValue = widgets[0].iconHeight;
         }

         return widgets.reduce(
               function(previousValue, currentValue, index, array)
               {
                  var width;
                  if (horizontalLayout)
                  {
                     width = array[index].iconWidth;
                  }
                  else
                  {
                     width = array[index].iconHeight;
                  }
                  return Math.max(previousValue, width);
               },
               initialValue);
      };
      
      /**
       * Calculates the unified task icon width.
       * 
       * @param    {Number} width
       *           The task bar width in pixels.
       * @param    {Number} height
       *           The task bar height in pixels.
       * @param    {Number} space
       *           The space between widgets in pixels.
       * @param    {Boolean} horizontalLayout
       *           Indicates the task bar layout, true if the task bar is along a horizontal line, otherwise,
       *           it is along a vertical line.
       * 
       * @return   {Number}
       *           The unified task icon width.
       */
      function calculateTaskIconWidth(width, height, space, horizontalLayout)
      {
         var numberOfWidgets = taskIcons.length + sysButtons.length;
         var wasteSpaceWidth = space * (numberOfWidgets - 1);
         var maxWidth;
         if (horizontalLayout)
         {
            // LEFT_PADDING widget1 SPACE widget2 .... SPACE widgetn RIGHT_PADDING
            maxWidth = (width - wasteSpaceWidth - LEFT_PADDING - RIGHT_PADDING) / numberOfWidgets;
         }
         else
         {
            // TOP_PADDING widget1 SPACE widget2 .... SPACE widgetn BOTTOM_PADDING
            maxWidth = (height  - wasteSpaceWidth - TOP_PADDING - BOTTOM_PADDING)/ numberOfWidgets;
         }
         var taskIconsWidth = calculateMaximalIconWidth(taskIcons, horizontalLayout);
         if (taskIconsWidth > 0)
         {
            taskIconsWidth += (2 * ICON_MARGIN);
         }
         taskIconsWidth  += (calculateMaximalTextWidth(taskIcons) + (2 * TEXT_MARGIN));
         var sysButtonsWidth = calculateMaximalIconWidth(sysButtons, horizontalLayout);
         if (sysButtonsWidth > 0)
         {
            sysButtonsWidth += (2 * ICON_MARGIN);
         }
         sysButtonsWidth += (calculateMaximalTextWidth(sysButtons) + (2 * TEXT_MARGIN));
         return Math.min(maxWidth, Math.max(taskIconsWidth, sysButtonsWidth));
      };
      
      /**
       * Sets the unified dimension for all task bar widgets.
       * 
       * @param    {Number} unifiedWidth
       *           The widget width in pixels.
       * @param    {Number} unifiedHeight
       *           The widget height in pixels.
       * @param    {Boolean} horizontalLayout
       *           Indicates the task bar layout, true if the task bar is along a horizontal line, otherwise,
       *           it is along a vertical line.
       */
      function setUnifiedDimension(unifiedWidth, unifiedHeight, horizontalLayout)
      {
         function setDimension(element, index, array)
         {
            if (horizontalLayout)
            {
               array[index].width  = unifiedWidth;
               array[index].height = unifiedHeight;
            }
            else
            {
               array[index].width  = unifiedHeight;
               array[index].height = unifiedWidth;
            }
            array[index].horizontalLayout = horizontalLayout;
         };
         taskIcons.forEach(setDimension);
         sysButtons.forEach(setDimension);
      };

      this.draw = draw;

      /**
       * Draws the task bar on the canvas.
       */
      function draw()
      {
         doLayout(peer.offsetWidth, peer.offsetHeight);
      }

      this.addTaskIcon    = addTaskIcon;

      /**
       * Adds new task to the task bar if there doesn't exist any task icon for the selected
       * window. New created task is hidden. 
       * 
       * @param    {Number} windowId
       *           The id of the window that needs to be represented by this task.
       * 
       * @return   {TaskIcon}
       *           The task icon that represents the target window.
       */
      function addTaskIcon(windowId)
      {
         var taskIcon = findWidget(windowId, taskIcons);
         if (taskIcon === null)
         {
            taskIcon = findWidget(windowId, hiddenTaskIcons);
            if (taskIcon === null)
            {
               taskIcon = new TaskIcon(windowId, TASK_ICON_WIDTH, TASK_ICON_HEIGHT, foreground,
                     "", null, 0, 0);
               hiddenTaskIcons.push(taskIcon);
            }
         }
         return taskIcon;
      };

      this.removeTaskIcon = removeTaskIcon;

      /**
       * Removes the task icon corresponding to the given window.
       * 
       * @param    {Number} windowId
       *           The id of the window.
       * 
       * @return   {TaskIcon}
       *           The removed window task icon.
       */
      function removeTaskIcon(windowId)
      {
         var taskIcon = findWidget(windowId, taskIcons);
         if (taskIcon)
         {
            var index = taskIcon.index;
            //shrunk the task icons by removing it
            taskIcons.splice(index, 1);
         }
         else
         {
            taskIcon = findWidget(windowId, hiddenTaskIcons);
            if (taskIcon)
            {
               var index = taskIcon.index;
               //shrunk the task icons by removing it
               hiddenTaskIcons.splice(index, 1);
            }
         }
         // repaint the task bar
         draw();
         
         return taskIcon;
      };

      this.setTaskIconVisible = setTaskIconVisible;

      /**
       * View/hides the task icon for the given window.
       * 
       * @param    {Number} windowId
       *           The id of the window.
       * @param    {Boolean} visible
       *           The true value indicates that the target task icon becomes
       *           displayed on the task bar.
       * 
       * @return   {TaskIcon}
       *           The visible/hidden window task icon.
       */
      function setTaskIconVisible(windowId, visible)
      {
         var taskIcon;
         var widgetsToAdd;
         var widgetsToRemove;
         if (visible)
         {
            widgetsToRemove = hiddenTaskIcons;
            widgetsToAdd    = taskIcons;
         }
         else
         {
            widgetsToRemove = taskIcons;
            widgetsToAdd    = hiddenTaskIcons;
         }
         taskIcon = findWidget(windowId, widgetsToRemove);
         if (taskIcon)
         {
            var index = taskIcon.index;
            //shrunk the task icons by removing it
            widgetsToRemove.splice(index, 1);
            widgetsToAdd.push(taskIcon);
            if (visible)
            {
               // sets the current position in the visible task icon list
               taskIcon.index = widgetsToAdd.length - 1;
            }
            //if the task icon becomes visible it should be active
            updateTaskBar(taskIcon, visible)
         }
         
         return taskIcon;
      };
      
      this.setIconForTask = setIconForTask;

      /**
       * Sets an icon for the task that represents the target window.
       * 
       * @param    {Number} windowId
       *           The target window id.
       * @param    {Number} iconId
       *           The icon unique identifier.
       * @param    {Array} iconData
       *           It is the binary icon to set. The pixels array with 4-bytes per a pixel that
       *           are red, green, blue and alpha bytes.
       * @param    {Number} iconWidth
       *           The icon width in pixels.
       * @param    {Number} iconHeight
       *           The icon height in pixels.
       * 
       * @return   {TaskIcon}
       *           The task that represents the target window.
       */
      function setIconForTask(windowId, iconId, iconData, iconWidth, iconHeight)
      {
         var taskIcon = findTaskIcon(windowId);
         if (taskIcon)
         {
            if (taskIcon.iconId !== iconId)
            {
               taskIcon.iconId = iconId;
               taskIcon.iconData = iconData;
               taskIcon.iconWidth = iconWidth;
               taskIcon.iconHeight = iconHeight;
               window.setTimeout(draw, 500);
            }
         }
         return taskIcon;
      };
      
      this.setTitleForTask = setTitleForTask;
      
      /**
       * Sets a title for the task that represents the target window.
       * 
       * @param    {Number} windowId
       *           The target window id.
       * @param    {String} title
       *           The title of the task icon for the given window.
       * 
       * @return   {TaskIcon}
       *           The task that represents the target window.
       */
      function setTitleForTask(windowId, title)
      {
         var taskIcon = findTaskIcon(windowId);
         if (taskIcon)
         {
            if (taskIcon.title !== title)
            {
               taskIcon.title = title;
               window.setTimeout(draw, 500);
            }
         }
         return taskIcon;
      };

      this.setTaskIconActive = setTaskIconActive;
      
      /**
       * Sets the active state for the task that represents the target window.
       * 
       * @param    {Number} windowId
       *           The target window id.
       * @param    {Boolean} active
       *           The window state.
       * 
       * @return   {TaskIcon}
       *           The task that represents the target window.
       */
      function setTaskIconActive(windowId, active)
      {
         var taskIcon = findTaskIcon(windowId);
         if (taskIcon)
         {
            updateTaskBar(taskIcon, active);
         }
         return taskIcon;
      };

      /**
       * Updates the given task icon and the active task icon index, then draws the task bar.
       * 
       * @param    {TaskIcon} taskIcon
       *           The task icon that is going to change its state
       * @param    {Boolean} active
       *           The new active state
       */
      function updateTaskBar(taskIcon, active)
      {
         taskIcon.setActive(active);
         if (activeTaskIconIndex === taskIcon.index)
         {
            if (!active)
            {
               activeTaskIconIndex = undefined;
            }
         }
         else
         {
            //new active window
            if (active)
            {
               if (activeTaskIconIndex !== undefined)
               {
                  var oldActiveTaskIcon = taskIcons[activeTaskIconIndex];
                  if (oldActiveTaskIcon !== undefined)
                  {
                     oldActiveTaskIcon.setActive(false);
                  }
               }
               activeTaskIconIndex = taskIcon.index;
            }
         }
         draw();
      }

      this.setFont = setFont;
      
      /**
       * Sets the default font to be used by the task bar.
       * 
       * @param    {Object} fontConfig
       *           The font configuration { name  : 'sans-serif', size  : 8 }.
       */
      function setFont(fontConfig)
      {
         var font = TASK_BAR_FONT_ID;
         var name;
         if (fontConfig.name)
         {
            name = fontConfig.name;
         }
         else
         {
            name = "sans-serif";
         }
         var size;
         if (fontConfig.size)
         {
            size = fontConfig.size;
         }
         else
         {
            size = 8;
         }
         // bold 
         var style = 1;
         var b64Font = "";
         return fontsManager.createFont(font, name, size, style, b64Font);
      }

      /**
       * Find the window task that represents the target window.
       * 
       * @param    {Number} windowId
       *           The target window id.
       * 
       * @return   {TaskIcon}
       *           The task that represents the target window.
       */
      function findTaskIcon(windowId)
      {
         var taskIcon = findWidget(windowId, taskIcons);
         if (taskIcon)
         {
            return taskIcon;
         }
         //try hidden task icons
         return findWidget(windowId, hiddenTaskIcons);
      };
      
      /**
       * Find the window task that represents the target window from the given searchable set.
       * 
       * @param    {Number} windowId
       *           The target window id.
       * @param    {Array} widgets
       *           The searchable array of TaskIcon.
       * 
       * @return   {TaskIcon}
       *           The task that represents the target window.
       */
      function findWidget(windowId, widgets)
      {
         for (var i = 0; i < widgets.length; i++)
         {
            var taskIcon = widgets[i];
            if (taskIcon.id === windowId)
            {
               taskIcon.index = i;
               return taskIcon;
            }
         }
         return null;
      };
      /**
       * Collapses the given window. Sets the corresponding task icon to be inactive.
       * 
       * @param    {Number} windowId
       *           The target window id.
       */
      this.iconify = function(windowId)
      {
         var taskIcon = findTaskIcon(windowId);
         if (taskIcon === null)
         {
            return;
         }
         updateTaskBar(taskIcon, false);
         taskIcon.draw();
      };

      /**
       * Restores the window. Sets the corresponding task icon to be active.
       * 
       * @param    {Number} windowId
       *           The target window id.
       */
      this.deiconify = function(windowId)
      {
         var taskIcon = findTaskIcon(windowId);
         if (taskIcon === null)
         {
            return;
         }
         updateTaskBar(taskIcon, true);
         taskIcon.draw();
      };

      /**
       * It is a "dragstart" event handler that makes the movable task bar to appear half transparent.
       * 
       * @param    {DragEvent} event
       *           The drag event.
       */
      function startDragging(event)
      {
         if (!p2j.isIe)
         {
            event.dataTransfer.setData('text/plain', null);
         }
         dragged = event.target;
         p2j.setStyleForElement(event.target,
               {
                  opacity : DRAGGED_OPACITY
               });
      }

      /**
       * It is a "dragend" event handler that restores the opacity of the movable
       * task bar.
       * 
       * @param    {DragEvent} event
       *           The drag event.
       */
      function dragEnded(event)
      {
         p2j.setStyleForElement(event.target,
               {
                  opacity : 1
               });
      }
      
      /**
       * It is a "drag" event handler.
       * 
       * @param    {DragEvent} event
       *           The drag event.
       */
      function drag( event )
      {
      }

      /**
       * events fired on the drop targets
       */
      document.addEventListener(
            "dragover",
            function(event)
            {
               event.stopPropagation();
               event.preventDefault();
            }, false);

      /**
       * The dragged task panel enters the drop zone
       */
      document.addEventListener("dragenter", function( event )
      {
         var cls = event.target.className;
          if ( cls.indexOf(DOCK_CLASS_NAME) >= 0)
          {
             p2j.setStyleForElement(event.target,
                   {
                      background : getTransparentColor(background, 0.5)
                   });
          }
      }, false);
      
      /**
       * The dragged task panel leaves the drop zone
       */
      document.addEventListener("dragleave", function( event )
      {
         var cls = event.target.className;
         
         if ( cls.indexOf(DOCK_CLASS_NAME) >= 0 )
         {
            p2j.setStyleForElement(event.target,
                  {
                     background : getTransparentColor(background, 0.0)
                  });
         }
      }, false);

      /**
       * The dragged task panel is attached to new drop zone
       */
      document.addEventListener("drop", function( event )
      {
         var cls = event.target.className;
         if ( cls.indexOf(DOCK_CLASS_NAME) >= 0)
         {
            var parent = dragged.parentNode;
            parent.removeChild(dragged);
            try
            {
               event.target.appendChild(dragged);
               p2j.setStyleForElement(event.target,
                     {
                        "z-index" : MAX_Z_INDEX
                     });
               p2j.setStyleForElement(parent,
                     {
                        background : getTransparentColor(background, 0.0),
                        "z-index"    : MIN_Z_INDEX
                     });
            }
            catch(e)
            {
               parent.addChild(dragged);
               p2j.setStyleForElement(parent,
                     {
                        "z-index"    : MAX_Z_INDEX
                     });
            }
            draw();
         }
         event.preventDefault();
      }, false);

      /**
       * Adds the "on click" task bar action. If a task icon widget gets the mouse pointer,
       * then this icon widget sets the active state and sends the request to server in order
       * to change the state of the window represented by this widget.
       */
      canvas.addEventListener("click", function(event)
      {
         //special view log use case.
         if (viewLogButton.testMousePointerInside(event) && !viewLogButton.isActive())
         {
            viewLogButton.setActive(true);
            viewLogButton.doClick();
            viewLogButton.draw();
            return;
         }
         
         if (taskIconOnClickCallback)
         {
            var oldActiveTaskIconId = -1;
            var newActiveTaskIconId = -1;
            
            for (var i = 0; i < taskIcons.length; i++)
            {
               var taskIcon = taskIcons[i];
               if (taskIcon.testMousePointerInside(event) && !taskIcon.isActive())
               {
                  newActiveTaskIconId = taskIcon.id;
               }
               
               if (taskIcon.isActive())
               {
                  oldActiveTaskIconId = taskIcon.id;
               }
               
               if (oldActiveTaskIconId != -1 && newActiveTaskIconId != -1)
               {
                  // found both task icons, no more to do
                  break;   
               }
            }
            
            // only fire the callback when a window is actually activated/deactivated
            if (newActiveTaskIconId != -1 || oldActiveTaskIconId != -1)
            {
               taskIconOnClickCallback(oldActiveTaskIconId, newActiveTaskIconId);
            }
         }
      });

      /**
       * Changes the tooltip for the task bar peer if the mouse pointer is over the task widget.
       * 
       * @param    {MouseEvent} event
       *           Describes the current mouse pointer.
       */
      function changeTooltipOverWidget(event)
      {
         //special view log use case.
         if (viewLogButton.testMousePointerInside(event))
         {
            peer.title = viewLogButton.title;
            return;
         }
         
         for (var i = 0; i < taskIcons.length; i++)
         {
            var taskIcon = taskIcons[i];
            if (taskIcon.testMousePointerInside(event))
            {
               peer.title = (taskIcon.title !== null && taskIcon.title !== undefined)
                                 ? taskIcon.title : " ";
               break;
            }
            
         }
      }

      canvas.addEventListener("mousemove", changeTooltipOverWidget);
      canvas.addEventListener("mouseover", changeTooltipOverWidget);
      canvas.addEventListener("mouseleave", function(event) { setDefaultTooltip(); });

      // add MDN resize throttler
      window.addEventListener("resize", resizeThrottler, false);

      /** The id of the scheduled resize function */
      var resizeTimeout;
      
      /**
       * Redraws this task bar if the browser window is changing its size,
       * with the frequency that doesn' exceed one time per 100 ms
       */
      function resizeThrottler()
      {
         // ignore resize events as long as an actualResizeHandler execution is in the queue
         if (!resizeTimeout)
         {
            resizeTimeout = window.setTimeout(
                  function()
                  {
                     resizeTimeout = null;
                     draw();
                  }, 100);
         }
      };
      /**
       * Returns rgba(r,g,b,a) string representation for the target color
       * with the target transparency.
       * 
       * @param    {Array} c
       *           Array of 3 integer values between 0 and 255 inclusive, representing an RGB color.
       * @param    {Number} transparency
       *           The transparency from 0.0 to 1.0.
       * 
       * @return   {String}
       *           The 'rgba(r,g,b,a)' string representation.
       */
      function getTransparentColor(c, transparency)
      {
         var parameters = [];
         Array.prototype.push.apply(parameters, c);
         parameters.push(transparency);
         return "rgba(" + parameters.join() + ")";
      }
      
      /**
       * Creates the task icon widget that represents the target window on the task bar.
       * 
       * @param    {Number} windowId
       *           The id of the window that needs to be represented by this task icon.
       * @param    {Number} w
       *           The widget width in pixels.
       * @param    {Number} h
       *           The widget height in pixels.
       * @param    {Array} c
       *           The task icon widget color. Array of 3 integer values between 0
       *           and 255 inclusive, representing an RGB color.
       * @param    {String} title
       *           The title of the task icon for the given window.
       * @param    {Array} iconData
       *           It is the binary icon to set. The pixels array with 4-bytes per a pixel that
       *           are red, green, blue and alpha bytes.
       * @param    {Number} iconWidth
       *           The icon width in pixels.
       * @param    {Number} iconHeight
       *           The icon height in pixels.
       */
      function TaskIcon(windowId, w, h, c, title, iconData, iconWidth, iconHeight)
      {
         this.id = windowId;
         this.iconData = iconData;
         this.iconWidth = iconWidth;
         this.iconHeight = iconHeight;
         this.title = title;
         this.width  = w;
         this.height = h;
         this.color  = c;
         var that = this;
         var active = false;
         this.x = 0;
         this.y = 0;
         /**The internal index, if the task icon is on the task bar, it is set
          * to the task icon postion. It holds only a temporary state.*/
         this.index = -1;
         /** Defines the widget's title layout. */
         this.horizontalLayout = true;

         this.setLocation = setLocation;

         /**
          * Sets the widget location.
          */
         function setLocation(x, y)
         {
            this.x = x;
            this.y = y;
         };

         this.setActive = setActive;

         /**
          * Sets an active state.
          */
         function setActive(value)
         {
            active = value;
         };

         this.isActive  = isActive;

         /**
          * Is active?
          */
         function isActive()
         {
            return active;
         };

         this.draw = draw;

         /**
          * Draw the widget on the task bar canvas.
          */
         function draw()
         {
            if ((that.title === "") && !that.iconData)
            {
               return;
            }
            var c;
            if (!active)
            {
               c = canvasRenderer.lightenColor(that.color);
            }
            else
            {
               c = that.color;
            }
            canvasRenderer.draw3DRect(that.x, that.y, that.width, that.height, c, true, !active);
            drawIcon();
            drawTitle();
         };

         /**
          * Draws the icon according to the icon margins and the widget's layout.
          */
         function drawIcon()
         {
            if (!that.iconData)
            {
               return;
            }
            var xIconPos;
            var yIconPos;
            if (that.horizontalLayout)
            {
               xIconPos = that.x + ICON_MARGIN;
               yIconPos = that.y + (that.height - that.iconHeight) / 2;
            }
            else
            {
               xIconPos = that.x + (that.width - that.iconWidth) / 2;
               yIconPos = that.y + ICON_MARGIN;
            }
            canvasRenderer.drawImage(xIconPos, yIconPos, that.iconWidth, that.iconHeight,
                  that.iconData, 0);
         }

         /**
          * Draws the title according to the text and icon margins and the widget's layout
          */
         function drawTitle()
         {
            if (that.title === "")
            {
               return;
            }
               ctx.save();
               var textWidth  = fontsManager.getTextWidth(TASK_BAR_FONT_ID, that.title);
               var textHeight = fontsManager.getTextHeight(TASK_BAR_FONT_ID, that.title);
               ctx.font = fontsManager.getFontName(TASK_BAR_FONT_ID);
               var xTextPos;
               var yTextPos;
               var text;
               if (that.horizontalLayout)
               {
                  var xShift = TEXT_MARGIN;
                  if (that.iconWidth > 0)
                  {
                     xShift += that.iconWidth + 2 * ICON_MARGIN;
                  }
                  if (that.width - TEXT_MARGIN - xShift >= textWidth)
                  {
                     xTextPos = that.x + (that.width - TEXT_MARGIN - textWidth + xShift) / 2;
                     text = that.title;
                  }
                  else
                  {
                     xTextPos = that.x + xShift;
                     text = cutTextToFitWidth(that.title, textWidth, that.width - TEXT_MARGIN - xShift);
                  }
                  
                  yTextPos = that.y + (that.height + textHeight) / 2;
               }
               else
               {
                  var yShift = TEXT_MARGIN;
                  if (that.iconHeight > 0)
                  {
                     yShift += that.iconHeight + 2 * ICON_MARGIN;
                  }
                  xTextPos = that.x + (that.width - textHeight) / 2;
                  if (that.height - TEXT_MARGIN - yShift >= textWidth)
                  {
                     yTextPos = that.y + (that.height - TEXT_MARGIN - textWidth + yShift) / 2;
                     text = that.title;
                  }
                  else
                  {
                     yTextPos = that.y + yShift;
                     text = cutTextToFitWidth(that.title, textWidth, that.height - TEXT_MARGIN - yShift);
                  }
                  ctx.translate(xTextPos, yTextPos);
                  ctx.rotate(Math.PI / 2);
                  xTextPos = 0;
                  yTextPos = 0;
               }
               if (!active)
               {
                  var textRgb = canvasRenderer.parseColor(ctx.fillStyle);
                  var selectedTextRgb = canvasRenderer.lightenColor(textRgb).map(
                        function(channel)
                        {
                           return LIGHTEN_COLOR_INTENSE * channel;
                        });
                  ctx.fillStyle = canvasRenderer.createColorString(selectedTextRgb);
               }
               
               canvasRenderer.drawText(
                     text,
                     xTextPos,
                     yTextPos, false);
               ctx.restore();
         }

         /**
          * Cuts the given text to fit the given width.
          * 
          * @param    {String} text
          *           The text to draw
          * @param    {Number} textWidth
          *           The text width in the current device metric.
          * @param    {Number} widthToFit
          *           The target width to which the trancated text should fit.
          * 
          * @return   {String}
          *           The truncated text that must fit the given width.
          */
         function cutTextToFitWidth(text, textWidth, widthToFit)
         {
            var averageTextLen = Math.floor((widthToFit / (1.0 * textWidth)) * text.length);
            if (averageTextLen > 3)
            {
               return text.substr(0, averageTextLen - 3) + "...";
            }
            else if (averageTextLen > 1)
            {
               return text.substr(0, averageTextLen - 1) + ".";
            }
            else
            {
               return ".";
            }
         }
         
         this.testMousePointerInside = testMousePointerInside;

         /**
          * Tests the mouse pointer inside the widgets bounds.
          */
         function testMousePointerInside(event)
         {
            var rect = peer.getBoundingClientRect();
            var x_rel = event.clientX - rect.left;
            var y_rel = event.clientY - rect.top;
            return isInOrder(that.x, x_rel, that.x + that.width)
                  && isInOrder(that.y, y_rel, that.y + that.height);
         };
         
         /**
          * Returns true iff a <= b <= c
          */
         function isInOrder(a, b, c)
         {
            if (a > b || b > c)
            {
               return false;
            }
            return true;
         };
         
      }
   }
}