/*
** Module   : Theme.java
** Abstract : the interface for all themes supported by FWD. 
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170308 Initial version with declaration of basic renderers and getters.   
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui.theme;

import java.awt.*;
import java.awt.Color;
import java.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Label;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Interface for all themes. It exposes two sets of methods: getters and renderers:
 * <ul>
 *    <li>The getters (their names start with "get"/"is", like beans) are used to compute the
 *          layout when the widget is visually different on different themes.</li>   
 *    <li>The renderers do the painting specific to each theme. Their names start with "draw".
 *          </li>
 * </ul>
 * <p>
 * The interface also contains a set of constants used by teh declared methods. 
 */
public interface Theme
{
   /**
    * The alignment is defined on both direction. Since the constants use different set of bits,
    * they can be combined (using XOR) to have a single integer value coding the alignment on
    * both axes. To extract a specific alignment, two masks are defined.
    * <p>
    * This constant is defined for invalid or uninitialized alignment value.
    */
   public static int ALIGN_NONE = 0;
   
   /** Constant used to extract the horizontal alignment from an alignment value.*/
   public static int HORIZONTAL_ALIGN_MASK = 0x03;
   
   /** Constant for left alignment on horizontal (Ox) axe. */
   public static int ALIGN_LEFT = 1;
   
   /** Constant for right alignment on horizontal (Ox) axe. */
   public static int ALIGN_RIGHT = 2;

   /** Constant for center alignment on horizontal (Ox) axe. */
   public static int ALIGN_CENTER = 3;
   
   /** Constant used to extract the vertical alignment from an alignment value.*/
   public static int VERTICAL_ALIGN_MASK = 0x0C;
   
   /** Constant for top alignment on vertical (Oy) axe. */
   public static int ALIGN_TOP = 4;
   
   /** Constant for bottom alignment on vertical (Oy) axe. */
   public static int ALIGN_BOTTOM = 8;
   
   /** Constant for middle alignment on vertical (Oy) axe. */
   public static int ALIGN_MIDDLE = 12;
   
   /**
    * Get the name of this {@code Theme}. All implementing classes must implement or override
    * this method.
    * 
    * @return  the name of this theme.
    */
   public String getName();
   
   /**
    * Release any held resources when this {@code Theme} is removed. 
    */
   public default void releaseResources() {};
   
   /**
    * Obtain an image resource specific to this {@code Theme}. The image can be cached for
    * performance reasons. It will be automatically released when {@link #releaseResources()} is
    * called, when a new {@code Theme} is loaded and this one is removed.
    * 
    * @param   name
    *          The name of the image as path within jar bundle.
    * @param   width
    *          The desired with of the returned image. This is only a hint. If available, an
    *          image with exact width is returned. One need to check the size of the returned
    *          {@link ImageWrapper} to know the exact value.  
    * @param   height
    *          The desired with of the returned image. This is only a hint. If available, an
    *          image with exact width is returned. One need to check the size of the returned
    *          {@code ImageWrapper} to know the exact value.  
    * @param   antialias
    *          if {@code true} and image is scaled to meet the size hints, an smooth scaling
    *          algorithm will be used, otherwise a faster one will be used but the image quality
    *          will be lower. If image is not scaled the parameter is ignored.
    *          
    * @return  an {@code ImageWrapper} containing the requested image resource if it can be found
    *          and loaded and {@code null} otherwise.
    */
   public ImageWrapper getImage(String name, int width, int height, boolean antialias);
   
   /**
    * Check if the theme is mouse hover sensitive. Mouse hover sensitive themes need to repaint
    * the components when the mouse hovers over them (ie. without clicking them).
    *
    * @return  {@code true} iif the theme is mouse hover sensitive.
    */
   public boolean isMouseHoverSensitive();
   
   /**
    * Obtain the default system palette for this theme.
    * 
    * @return  the default system palette for this theme.
    */
   public Map<Integer, ColorRgb> getDefaultSystemPalette();
   
   /**
    * Obtain the map of standard fonts specific to current {@code Theme}, to which a key defaults,
    * if not overridden in the directory.
    * 
    * @return  the map of standard fonts specific to this {@code Theme}. 
    */
   public Map<String, String> getStandardFontTable();
   
   /**
    * Utility method that returns the size of the prebuilt arrow images.
    *
    * @return  size of prebuilt arrow size.
    */
   public int getPrebuiltArrowImageSize();
   
   /**
    * Returns the height of window's titleBar. It DOES NOT include the upper window border. 
    * 
    * @param   maximized
    *          On some themes, the maximized window have a different title bar height when 
    *          maximized. 
    * 
    * @return  the height of the window's title bar. 
    */
   public int getTitleBarHeight(boolean maximized);
   
   /**
    * Returns the width of window's border. 
    *
    * @param   modal
    *          Usually the modal dialogs/windows have a different width if modal. 
    *
    * @return  the width of window's border. 
    */
   public int getWindowBorderWidth(boolean modal);
   
   /**
    * Returns the height of a caption button on window's title bar. 
    *
    * @param   maximized
    *          On some themes, the maximized window have a different button height when 
    *          maximized. 
    *
    * @return  the height of the caption button. 
    */
   public int getCaptionButtonHeight(boolean maximized);
   
   /**
    * Returns the width of a caption button on window's title bar. 
    * 
    * @param   type
    *          On some specific themes, the width of buttons is not the same.
    * 
    * @param   dialog
    *          {@code true} for dialogs. The dialogs may have specific width for caption buttons.  
    *
    * @return  the width of the requested caption button. 
    */
   public int getCaptionButtonWidth(CaptionButtonType type, boolean dialog);
   
   /**
    * Get the alignment of caption button for this theme. The returned value combines both
    * horizontal and vertical alignment of the buttons.
    * 
    * @return  the alignment of caption button.
    */
   public int getCaptionButtonsAlignment();
   
   /**
    * Get the title alignment for this theme. The returned value combines both horizontal and
    * vertical alignment.
    *
    * @return  the alignment of title.
    */
   public int getWindowTitleAlignment();
   
   /**
    * Get the icon alignment for this theme. The returned value combines both horizontal and
    * vertical alignment.
    *
    * @return  the alignment of icon on titlebar.
    */
   public int getWindowIconAlignment();
   
   /**
    * Return the amount of space to the right of this button. On some themes, different spaces can
    * be found between the caption buttons. Although unusual, negative values can be returned,
    * when buttons overlap for a few pixels. 
    * 
    * @param   type
    *          The caption button type.
    * 
    * @return  the amount of space to the right of this button.
    */
   public int getSpaceBeforeButton(CaptionButtonType type);
   
   /**
    * Do the layout of a titlebar. Each {@code Theme} can define how its components are arranged. 
    * 
    * @param   window
    *          The window whose titlebar is laid out. 
    * @param   icon
    *          The icon widget of the window (if any).
    * @param   title
    *          The {@link Label} that displays the window's title. 
    * @param   captionButtons
    *          An array with the {@link CaptionButton}-s present on the titlebar of this window. 
    * @param   width
    *          The width of the window (in character units).
    * @param   c
    *          the {@link CoordinatesConversion} used for possible units conversions.
    *
    * @return  the computed titlebar height.
    */
   double layoutTitleBar(TopLevelWindow window, 
                         Widget icon, 
                         Label title, 
                         CaptionButton[] captionButtons, 
                         double width, 
                         CoordinatesConversion c);
   
   /**
    * Obtain the {@link Font} used to render the windows title.
    * 
    * @return the {@code Font} used to render the windows title.
    */
   public Font getTitleBarFont();
   
   /**
    * Obtain the {@link Color} used to render the windows title.
    * 
    * @param   active
    *          {@code true} if the rendered window is active. The inactive window is usually fade. 
    *
    * @return  the {@code Color} used to render the windows title.
    */
   public Color getTitleBarTextColor(boolean active);
   
   /**
    * Renders a {@link CaptionButton}.
    * 
    * @param   button
    *          The {@code CaptionButton} to be drawn. 
    * @param   gd
    *          The {@link GuiDriver} used for drawing operations. Note that internal state of 
    *          this object can affect the resulting output.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawCaptionButton(CaptionButton button, 
                                 GuiDriver gd, 
                                 int x, 
                                 int y, 
                                 int width, 
                                 int height);
   
   /**
    * Renders the background of the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   public void drawWindowTitleBar(GuiDriver gd, 
                                  int x, 
                                  int y, 
                                  int width, 
                                  int height, 
                                  boolean active);
   
   /**
    * Renders the text on the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@link GuiFontResolver} for selecting the required font.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   title
    *          The title of the window.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   public void drawWindowTitle(GuiDriver gd, 
                               GuiFontResolver gf, 
                               int x,
                               int y,
                               int width,
                               int height,
                               String title,
                               boolean active);
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border
    *          The width of the border.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   public void drawWindowBorder(GuiDriver gd, 
                                int x, 
                                int y, 
                                int width, 
                                int height, 
                                int border, 
                                boolean active);
   
   /**
    * Renders the status bar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used for color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   text
    *          The text in the status bar.
    * @param   cache
    *          The font cache used for text rendering. 
    * @param   txtOrig
    *          The {@link NativePoint} where the text is drawn.
    * @param   txtDim
    *          The {@link NativeDimension} of the text displayed in status bar.
    * @param   resizable
    *          {@code true} window is resizable. In this case a "grip" is drawn in the lower right
    *          corner.
    */
   public void drawStatusBar(GuiDriver gd, 
                             GuiColorResolver gc,
                             int x, 
                             int y, 
                             int width, 
                             int height, 
                             String text, 
                             GuiFontResolver.FontCache cache, 
                             NativePoint txtOrig, 
                             NativeDimension txtDim, 
                             boolean resizable);
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   button
    *          The {@link ButtonGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   public void drawButton(ButtonGuiImpl button, 
                          GuiDriver gd, 
                          GuiColorResolver gc, 
                          GuiFontResolver gf, 
                          int x, 
                          int y, 
                          int width, 
                          int height);
   
   /**
    * Obtain the padding for alert boxes.
    * 
    * @return  the padding for alert boxes.
    */
   public int getAlertBoxPadding();
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   paragraphWidth
    *          The width of the paragraph.
    * @param   icon
    *          The {@code ImageWrapper} containing the image to be shown in the alert dialog. May
    *          be {@code null}.
    * @param   text
    *          The text to be displayed in this alert box.
    * @param   font
    *          The {@link FontDetails} needed for text layout.
    */
   public void drawAlertBox(GuiDriver gd, 
                            int x, 
                            int y, 
                            int width, 
                            int height, 
                            int paragraphWidth, 
                            ImageWrapper icon, 
                            String text, 
                            FontDetails font);
   
   /**
    * Obtain the distance separating radio buttons in a set.
    *
    * @return  the distance separating radio buttons in a set.
    */
   public int getRadioButtonExtraDist();
   
   /**
    * Obtain the distance before text, including the radio button (text offset).
    * 
    * @return  the distance before text, including the radio button.
    */
   public int getRadioButtonImageSpace();
   
   /**
     * Renders a GUI radio button widget.
     * @param  rb
     *         The {@link RadioButtonGuiImpl} to be rendered.
     * @param  gd
     *         The {@code GuiDriver} used for drawing operations.
     * @param  gc
     *         The {@link GuiColorResolver} to be used for color selection.
     * @param  gf
     *         The {@code GuiFontResolver} to be used.
     * @param  x
     *         The horizontal location (in pixel units) of the widget to be painted/component.
     * @param  y
     *         The vertical location (in pixel units) of the widget to be painted/component.
     * @param  width
     *         The width of the painted widget/component. 
     * @param  height
     *         The height of the painted widget/component.
     * @param  focused
     *         {@code true} if the radio button have the focus
     * @param  hover
     *         {@code true} if mouse is hovering over the checkbox. Some styles are mouse hovering
     */
   public void drawRadioButton(RadioButtonGuiImpl rb,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               GuiFontResolver gf,
                               int x,
                               int y,
                               int width,
                               int height,
                               boolean focused,
                               boolean hover);
   
   /**
    * Obtains the space needed by a checkbox widget.
    * 
    * @return  the space needed by a checkbox widget.
    */
   public int getCheckBoxSpace();
   
   /**
    * Obtain the size of image on a checkbox widget.
    * 
    * @return size of image on a checkbox widget.
    */
   public int getCheckBoxSize();
   
   /**
    * Renders a GUI toggle box widget.
    *
    * @param   checkbox
    *          The {@link ToggleBoxGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the checkbox is pressed (the mouse button is down on the button).
    * @param   hover
    *          {@code true} if mouse is hovering over the checkbox. Some styles are mouse hovering
    */
   public void drawToggleBox(ToggleBoxGuiImpl checkbox,
                             GuiDriver gd,
                             GuiColorResolver gc,
                             GuiFontResolver gf,
                             int x,
                             int y,
                             int width,
                             int height,
                             boolean pressed,
                             boolean hover);
   
   /**
    * Obtain the size of the scroll button on the direction perpendicular to scroll direction.
    * 
    * @return  size of the scroll button on the direction perpendicular to scroll direction.
    */
   public int getScrollBarButtonWidth();
   
   /** 
    * Obtain the thumb/slider width in a scrollbar widget.
    * 
    * @return  the thumb/slider width in a scrollbar widget.
    */
   public int getScrollBarThumbWidth();
   
   /**
    * Obtain the minimum size of the thumb/slider of a scrollbar widget.
    *
    * @return  the minimum size of the thumb/slider of a scrollbar widget.
    */
   public int getScrollBarThumbMinSize();
   
   /**
    * Obtain the width of the gauge of a scrollbar widget.
    *
    * @return  width of the gauge of a scrollbar widget.
    */
   public int getScrollBarGaugeWidth();
   
   /** 
    * Obtain the size of the scroll button on the direction of the scroll.
    * 
    * @return  the size of the scroll button on the direction of the scroll.
    */
   public int getScrollBarButtonHeight();
   
   /**
    * Renders the gauge (background) of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   public void drawScrollBarGauge(GuiDriver gd, 
                                  GuiColorResolver gc, 
                                  int x, 
                                  int y, 
                                  int width, 
                                  int height, 
                                  boolean horizontal);
   
   /**
    * Renders the thumb/slider of the scrollbar widget.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          If the mouse is dragging the thumb of the scrollbar.
    * @param   hover
    *          Is mouse hovering this component ?
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   public void drawScrollBarThumb(GuiDriver gd,
                                  GuiColorResolver gc,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean hover,
                                  boolean horizontal);
   
   /**
    * Obtain the offset in a 3D bordered widget.
    * 
    * @return  the offset in a 3D bordered widget.
    */
   public int get3DBorderOffset();
   
   /**
    * Renders a button of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pos
    *          The position of the button (UP/DOWN, LEFT/RIGHT).
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    * @param   enabled
    *          If the button is enabled.
    * @param   pressed
    *          If the button is down (pressed by mouse).
    * @param   highlighted
    *          If the mouse is hovering over the button.
    */
   public void drawScrollBarButton(GuiDriver gd, 
                                   GuiColorResolver gc,
                                   int x,
                                   int y,
                                   int width,
                                   int height,
                                   ScrollBar.Position pos,
                                   boolean horizontal,
                                   boolean enabled,
                                   boolean pressed,
                                   boolean highlighted);
   
   /** 
    * Obtain the character used to replace text in password fields. The default is "*" (asterisk)
    * but some themes may use other characters like "bullet" char.
    * 
    * @return  the character used to replace individual characters in password fields.
    */
   public char getPasswordCharacter();
   
   /**
    * Renders the border (and possibly the background) of the fill-in widget.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border3D
    *          If fill-in widget has a 3d border. Some flat {@code Theme}s may ignore this.
    * @param   hasFocus
    *          The fill-in widget is focused.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   public void drawFillInBorder(GuiDriver gd,
                                GuiColorResolver gc, 
                                int x, 
                                int y, 
                                int width, 
                                int height, 
                                boolean border3D,
                                boolean hasFocus, 
                                boolean hover);
   
   /**
    * Renders the selection (and possibly its background) of the fill-in widget widget. 
    *
    * @param   fillIn
    *          The {@link FillInGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   clipWidth
    *          The width of the clipping area. 
    * @param   clipHeight
    *          The height of the clipping area.
    * @param   selX
    *          The offset where the selection starts.
    * @param   selWidth
    *          The width of the selection.
    * @param   selHeight
    *          The height of the selection.
    * @param   selTxt
    *          The selected text.
    * @param   font
    *          The font to be used.
    * @param   bgColor
    *          The background color of the selection (selection color).
    * @param   fgColor
    *          The selected text color (foreground color).
    */
   public void drawFillInSelection(FillInGuiImpl fillIn,
                                   GuiDriver gd, 
                                   int clipWidth, 
                                   int clipHeight, 
                                   int selX, 
                                   int selWidth, 
                                   int selHeight, 
                                   String selTxt, 
                                   int font, 
                                   ColorRgb bgColor, 
                                   ColorRgb fgColor);
   
   /**
    * Renders caret of the fill-in widget widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x1
    *          The left extremity of the caret. 
    * @param   y1
    *          The top extremity of the caret.
    * @param   x2
    *          The right extremity of the caret. 
    * @param   y2
    *          The bottom extremity of the caret.
    */
   public void drawFillInCaret(GuiDriver gd, int x1, int y1, int x2, int y2);
   
   /**
    * Obtain the field extra space for combo box layout.
    * 
    * @param   horiz
    *          Direction. {@code true} for horizontal space.
    * @param   is3D
    *          Whether this is widget has a 3D border.
    *
    * @return  the field extra space for combo box layout.
    */
   public int getComboBoxFieldExtraSpace(boolean horiz, boolean is3D);
   
   /**
    * Obtain the shifting distance in a combo box widget.
    * 
    * @return  the shifting distance in a combo box widget.
    */
   public int getComboBoxEntryShift();
   
   /**
    * Obtain the size of the button that opens the combo box.
    * 
    * @return  the size of teh combo box button.
    */
   public int getComboBoxButtonSize();
   
   /**
    * Renders the entry field of a combo box. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose entry field is painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   font
    *          A {@code FontDetails} used for text rendering.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   public void drawComboBoxEntryField(ComboBoxGuiImpl combo,
                                      GuiDriver gd,
                                      GuiColorResolver gc,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      FontDetails<?> font,
                                      boolean pressed,
                                      boolean enabled,
                                      boolean hover);
   
   /**
    * Renders the border (and possibly the background) of the combo box widget. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   public void drawComboBoxFieldBorder(ComboBoxGuiImpl combo, 
                                       GuiDriver gd,
                                       int x,
                                       int y,
                                       int width,
                                       int height,
                                       boolean pressed,
                                       boolean enabled,
                                       boolean hover);
   
   /**
    * Renders the opening button of the combo box widget. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   public void drawComboBoxButton(ComboBoxGuiImpl combo,
                                  GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean enabled,
                                  boolean hover);
   
   /**
    * Renders the background of a selection list. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   public void drawSelectionListBackground(GuiDriver gd,
                                           GuiColorResolver gc,
                                           int x,
                                           int y,
                                           int width,
                                           int height);
   
   /**
    * Renders a single item in a combo box selection list. 
    *
    * @param   box
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   text
    *          The text to be displayed on this item
    * @param   fontSize
    *          The size of teh font to be used.
    * @param   highlighted
    *          The mouse is hovering over the fill-in widget.
    * @param   enabled
    *          Whether the widget has the focus.
    */
   public void drawSelectionListItem(ComboBoxGuiImpl box,
                                     GuiDriver gd,
                                     GuiColorResolver gc,
                                     int x,
                                     int y,
                                     int width,
                                     int height,
                                     String text,
                                     int fontSize,
                                     boolean enabled,
                                     boolean highlighted,
                                     boolean selected);
   
   /**
    * Obtain the vertical distance between the combo box and the scrolling selection list.  
    * 
    * @return the vertical offset for selection list of the combo box.
    */
   public int getComboBoxDropDownOffset();
   
   /**
    * Obtain the left padding of the test in a menu item. 
    * 
    * @return  left padding of the test in a menu item. 
    */
   public int getMenuLeftPad();
   
   /**
    * Obtain the right padding of the test in a menu item. 
    *
    * @return  right padding of the test in a menu item. 
    */
   public int getMenuRightPad();
   
   /**
    * Obtain the insets of a popup menu. They are used for drawing the box/ 3D border around the
    * popup menu.
    *
    * @return  the insets of a popup menu. 
    */
   public NativeInsets getMenuInsets();
   
   /**
    * Obtain the width of the check mark to for togglable menus. 
    * 
    * @return  the width of the check mark in menu items.
    */
   public int getMenuCheckmarkWidth();
   
   /**
    * Obtain the width of the arrow marking menus with submenus (including additional space). 
    *
    * @return  the width of the arrow mark for sub menus.
    */
   public int getMenuSmArrowWidth();
   
   /**
    * Obtain the minimum size of a titlebar element for this Theme, including any border. 
    * 
    * @param   horiz
    *          If {@code true} returns the minimum width, if {@code false} returns minimum height.
    *          
    * @return  The minimum size of a titlebar element on specified direction.
    */
   public int getMenuBarElementMinSize(boolean horiz);
   
   /**
    * Renders the background of a menu widget. 
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   popup
    *          {@code true} for popup menus.
    */
   public void drawMenuBackground(GuiDriver gd, 
                                  int x,
                                  int y, 
                                  int width, 
                                  int height, 
                                  boolean popup);
   
   /**
    * Renders the separator line in a menu widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   public void drawMenuSeparator(GuiDriver gd, int x, int y, int width, int height);
   
   /**
    * Obtain the fixed height of a separator in a popup menu widget.
    * 
    * @return  the height of a separator in a popup menu.
    */
   public int getMenuSeparatorHeight();
   
   /**
    * Obtain the height of a {@link MenuElement} in a menu widget.
    *
    * @return  the height of a normal item in a menu widget.
    */
   public int getMenuItemHeight(MenuElement menuItem);
   
   /**
    * Renders the foreground content of menu item in a menu. 
    *
    * @param   item
    *          The {@link MenuItemGuiImpl} or {@link SubMenuGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   textHeight
    *          The height of the text.
    * @param   focused
    *          {@code true} if this item is highlighted/focused.
    */
   public void drawMenuItemContent(AbstractWidget item, 
                                   GuiDriver gd,
                                   GuiFontResolver gf, 
                                   int x,
                                   int y,
                                   int width,
                                   int height,
                                   int textHeight, 
                                   boolean focused);
   
   /**
    * Renders the background of an item in a menu widget. 
    *
    * @param   item
    *          The {@link MenuItemGuiImpl} or {@link SubMenuGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   hover
    *          {@code true} if this item is highlighted/focused.
    */
   public void drawMenuItemBackground(AbstractWidget item,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      boolean hover);
   
   /**
    * Draws an element of a menu in a menubar.
    * 
    * @param   menu
    *          The {@code SubMenuGuiImpl} widget to be painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          GUI font resolver.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   hover
    *          {@code true} if this item is highlighted/focused.
    */
   public void drawMenuBarElement(SubMenuGuiImpl menu,
                                  GuiDriver gd,
                                  GuiFontResolver gf,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean hover);
   
   /**
    * Renders a label.
    * 
    * @param   label
    *          The {@link LabelGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    * @param   emptyMode
    *          If it is <code>true</code>, empty string is displayed.
    */
   public void drawLabel(LabelGuiImpl label,
                         GuiDriver gd,
                         GuiColorResolver gc,
                         GuiFontResolver gf,
                         int x,
                         int y,
                         int width,
                         int height,
                         boolean emptyMode);
   
   /**
    * Draw the horizontal bar under a label if the label is underlined.
    * 
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    * @param   fgColor
    *          The color attribute for the line.
    */
   public void drawLabelUnderline(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  ColorRgb fgColor);
   
   /**
    * Highlight the currently active row in a {@link BrowseGuiImpl} widget.
    *
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawBrowseRowHighlight(BrowseGuiImpl browse,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width,
                                      int height);
   
   /**
    * Draw animation of a dragged column in a  {@link BrowseGuiImpl} widget.
    *
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawBrowseColumnDrag(BrowseGuiImpl browse,
                                    GuiDriver gd,
                                    int x,
                                    int y,
                                    int width,
                                    int height);
   
   /**
    * Draw animation of a dragged column in a  {@link BrowseGuiImpl} widget.
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   left
    *          The left offset of the resized column.
    * @param   top
    *          The top offset where the columns start.
    * @param   right
    *          The right offset of the resized column. 
    * @param   bottom
    *          The bottom offset where the columns end.
    */
   public void drawBrowseResizingColumn(BrowseGuiImpl browse,
                                        GuiDriver gd,
                                        int left,
                                        int top,
                                        int right,
                                        int bottom);
   
   /**
    * Draw the title panel of a {@link BrowseGuiImpl} widget.
    * 
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawBrowseTitlePanel(BrowseGuiImpl browse,
                                    GuiDriver gd,
                                    GuiColorResolver gc,
                                    GuiFontResolver gf,
                                    int x,
                                    int y,
                                    int width,
                                    int height);
   
   /**
    * Draw the label on a {@link BrowseColumnGuiImpl} widget.
    * 
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    * @param   fontDetails
    *          The font details to be used.
    */
   public void drawBrowseColumnLabel(BrowseColumnGuiImpl column,
                                     GuiDriver gd,
                                     int x,
                                     int y,
                                     int width,
                                     int height,
                                     FontDetails fontDetails);
   
   /**
    * Draw a browse marker column {@link BrowseColumnGuiImpl} widget (without label).
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawBrowseMarkersColumn(BrowseColumnGuiImpl column,
                                       GuiDriver gd,
                                       int x,
                                       int y,
                                       int width,
                                       int height);
   
   /**
    * Draw a normal column {@link BrowseColumnGuiImpl} widget (without label).
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawBrowseNormalColumn(BrowseColumnGuiImpl column,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width, 
                                      int height);
   
   /**
    * Draw a {@link ToolTip} widget.
    * 
    * @param   toolTip
    *          The {@code ToolTip} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    * @param   fnt
    *          The {@code FontDetails} to be used when drawing the text.
    * @param   text
    *          The actual text to be drawn.
    */
   public void drawToolTip(ToolTip toolTip,
                           GuiDriver gd, 
                           int x,
                           int y,
                           int width,
                           int height,
                           FontDetails fnt,
                           String text);
   
   /**
    * Draw the title panel of a {@link BrowseGuiImpl} widget.
    *
    * @param   frame
    *          The {@code FrameGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawFrameTitle(FrameGuiImpl frame,
                              GuiDriver gd,
                              GuiColorResolver gc,
                              GuiFontResolver gf,
                              int x,
                              int y,
                              int width,
                              int height);
   
   /**
    * Draw the line border {@link LineBorderGuiImpl} widget.
    *
    * @param   border
    *          The {@code FrameGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   public void drawLineBorder(LineBorderGuiImpl border,
                              GuiDriver gd,
                              GuiColorResolver gc,
                              int x,
                              int y,
                              int width,
                              int height);
}