/*
** Module   : ComboBoxModel.java
** Abstract : Combo-box data model.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20100629          Created initial version
** 002 SIY 20100804          Changes in imports, some refactorings.
** 003 IAS 20150202          Add selected index.
** 004 HC  20150323          Removed business logic from config classes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.model;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Combo-box data model
 * 
 * @param   <T>
 *          Type of the item stored in the model. 
 */
public class ComboBoxModel<T>
extends ListModel<T>
{
   /** Selected item. */
   private T selected;
   
   /** Selected index. */
   private int selectedIndex = -1;
   
   /** Selected item id. */
   private int selectedId;
   
   /** Selected value. */
   private BaseDataType selectedValue;

   /**
    * Add new item at the end of model.
    * 
    * @param   item
    *          Item to add.
    */
   @Override
   public void add(T item)
   {
      super.add(item);

      if (size() == 1)
         selected = item;
   }

   /**
    * Insert item at specified location.
    *  
    * @param   item
    *          Item to insert.
    * @param   index
    *          Index at which item should be inserted.
    */
   public void insert(T item, int index)
   {
      add(index, item);

      if (size() == 1)
         selected = item;
   }

   /**
    * Remove item with specified value.
    * 
    * @param   item
    *          Item to remove.
    */
   public void remove(T item)
   {
      int index = indexOf(item);
      
      if (index >= 0)
         removeAt(index);
   }

   /**
    * Remove all items.
    */
   @Override
   public void clear()
   {
      super.clear();
      selected = null;
   }

   /**
    * Remove item at specified position.
    * 
    * @param   index
    *          Item index.
    */
   @Override
   public void removeAt(int index)
   {
      if (selected == get(index))
      {
         if (index == 0)
            selected = (size() == 1) ? null : get(1);
         else
            selected = get(index - 1);
      }

      super.removeAt(index);
   }

   /**
    * Get selected item.
    * 
    * @return  selected item.
    */
   public T selected()
   {
      return selected;
   }

   /**
    * Get selected index.
    * 
    * @return  selected index.
    */
   public int getSelectedIndex()
   {
      return selectedIndex;
   }

   /**
    * Set selected index.
    * 
    * @param   index
    *          Selected item index.
    */
   public void setSelectedIndex(int index)
   {
      selectedIndex = index;
   }

   /**
    * Get selected item id.
    * 
    * @return  selected item id.
    */
   public int getSelectedId()
   {
      return selectedId;
   }

   /**
    * Set selected item if.
    * 
    * @param   id
    *          Selected item id.
    */
   public void setSelectedId(int id)
   {
      selectedId = id;
   }

   /**
    * Get selected value.
    * 
    * @return  selected value.
    */
   public BaseDataType getSelectedValue()
   {
      return selectedValue;
   }

   /**
    * Set selected index.
    * 
    * @param   value
    *          Selected value.
    */
   public void setSelectedValue(BaseDataType value)
   {
      selectedValue = value;
   }
   

   /**
    * Select item with specified value.
    * 
    * @param   item
    *          Item value to select.
    */
   public void select(T item)
   {
      selected = item;
   }

     /**
    * Calculate the index position of the current selection.
    *
    * @param    items
    *           The list of items to be used within this drop-down list.
    *
    * @return   The 0-based index position if there is a current selection
    *           in the given model or 0 if there is no selection.
    */
   public int calcSelection(ControlSetItem[] items)
   {
      if (selectedId > 0)
      {
         for (int i = 0; i < items.length; i++) 
         {
            if (items[i].getItemId() == selectedId )
            {
               return selectedIndex >= 0 ? selectedIndex : i;
            }
         }
      }
      int idx = 0;
      if (selectedValue != null)
      {
         String svalue = selectedValue.toStringMessage();
         for (int i = 0; i < items.length; i++)
         {
            if (items[i].getValue() != null && svalue.equalsIgnoreCase(items[i].getValue().toStringMessage()))
            {
               idx = i;
               break;
            }
         }
      }
      return idx;
   }
   
   /**
    * Enforce the current selection.
    *
    * @param    items
    *           The list of items to be used within this drop-down list.
    */

   public void setSelection(ControlSetItem[] items) 
   {
      int idx = calcSelection(items);
      
      if (0 <= idx && idx < items.length) 
      {
         ControlSetItem item = items[idx];
         setSelectedIndex(idx);
         setSelectedId(item.getItemId());
      }
   }
}