/*
** Module   : AbstractContainer.java
** Abstract : Base class for widget containers.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100831 Changes related to refactoring of FrameHelper.
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 SIY 20110425 Fixed next/prev focus handling for deeply nested 
**                  containers.
** 006 GES 20111012 Moved static data to context-local. This allows multiple
**                  client sessions to be hosted in the server.
** 007 EVL 20140528 Fix for displaying the labels for hidden widgets in a frame.  When the hidden
**                  widget become visible the label text should be drawn even later the widget
**                  will be hidden again.
** 008 VIG 20140507 Added tab order management methods and changes.
** 009 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Removed size API, which collided with the dimension API.
**                  First attempt at clipping support (requires decimal char units to work).
**                  First pass at mouse support.  Added widget destroy notifications.
** 010 HC  20140911 GUI coordinate support - data types refactoring.
** 011 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 012 MAG 20140929 Fixed findMouseSource alghoritm.
** 013 CA  20141021 Changed findMouseSource to allow Z-order.  Fixed draw() to allow drawing of
**                  nested containers.
** 014 MAG 20141020 Added virtual size computation. Eliminated GUI drawing area settings in draw()
**                  Added register widgets.
** 015 HC  20141021 Implemented window move/position support.
** 016 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 017 VIG 20150216 Added method getWidgetsList.
** 018 HC  20150221 Improvements to frame tab order.
** 019 HC  20150323 Removed business logic from config classes.
** 020 CA  20150429 When creating a rectangle from a location and dimension, use the base units
**                  to determine the coordinates for current client mode.
** 021 VIG 20150302 SubMenu can gain a focus. Added exclusion.
**                  Fixed focusing methods, extracted prepareWidgets and childrenLocation methods. 
** 022 HC  20150508 Improvements to GUI window scrolling support.
** 023 HC  20150612 Implemented support for modal windows and GUI ALERT-BOX.
** 024 HC  20150702 Improved parent-child relationship of top-level windows.
** 025 HC  20150803 Comments cleanup.
** 026 VIG 20150709 NullPointerException fix.
** 027 HC  20150831 Widgets are automatically detached from their parents when destroyed. Removed
**                  repaint() call from setSize().
** 028 IAS 20150926 AUTO-RESIZE, MOVE-TO-TOP()/MOVE-TO-BOTTOM() support
** 029 HC  20150927 Layout fixes when insets are involved. Changes to allow multi-window focus
**                  management in GUI.
** 030 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 031 CA  20151024 Ignore non-container widgets with no mouse actions, when determining the mouse
**                  source.  Added support for MOVE-AFTER-TAB-ITEM. Fixed moveToTop/BottomInClass.
** 032 CA  20151123 Added APIs to place a widget on a specified location.
**                  In GUI, to determine if a widget can be drawn or not, it requires intersection
**                  with the outer clipping rectangle, not the individual posted rectangles (as
**                  drawing for that widget will always be done using the outer clipping 
**                  rectangle, as Swing doesn't support more than one clipping rectangle).
** 033 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 034 VIG 20151226 Fixed NullPointerException in currentFocus().
** 035 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 036 IAS 20160217 Fixed findMouseSource() method.
** 037 SBI 20160316 Set the focused window for focused widget requests.
** 038 HC  20160406 Prevented an unhandled exception.
** 039 VIG 20160128 Added getFocusInt(), changed visibility of setFocusInt().
** 040 HC  20160413 Fixed focusing of main window when overlay window closes.
** 041 HC  20160428 Fixed findMouseSource() to take into account parents' insets.
** 042 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 043 HC  20160712 Improved next/prev tab focus in nested frames.
** 044 CA  20160915 Added unrealize() method.
** 045 HC  20160922 Added currentFocus(boolean) method.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.util.*;

/**
 * Abstract widget container.
 */
public abstract class AbstractContainer<O extends OutputManager<?>>
extends AbstractWidget<O>
implements Container<O>
{
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(AbstractContainer.class);
   
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work = 
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };

   /** Layout manager. */
   private LayoutManager<O> layoutManager = null;

   /** Current (in focus) widget. */
   private Widget<O> current = null;

   // TODO: remove the field, AbstractWidget already declares insets
   /** Container insets. */
   private Insets insets = new Insets(0, 0, 0, 0);

   /** Container size. */
   private Dimension size = new Dimension(BaseConfig.INV_COORD, BaseConfig.INV_COORD);
   
   /** Widgets stored in the container. */
   private ArrayList<Widget<O>>  widgets = new ArrayList<Widget<O>>();
   
   /** 
    * The set of widgets that are never selected for focus - they are ignored by 
    * {@linkplain #checkWidget(Widget, boolean)}.
    */
   private Set<Widget<?>> ignoreFocusSet; 
   
   /**
    * Constructor.
    */
   public AbstractContainer()
   {
   }

   /**
    * Add widget to container.
    *
    * @param   widget
    *          Widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    * 
    * @throws  IllegalArgumentException
    *          when the argument is null or instance of {@link TopLevelWindow}
    */
   @Override
   public Container<O> add(Widget<O> widget)
   {
      if (widget == null || widget instanceof TopLevelWindow)
         throw new IllegalArgumentException();
         
      widgets.add(widget);

      widget.setParent(this);

      return this;
   }

   /**
    * Add the widget on the specified position.
    * 
    * @param    pos
    *           The widget's position.
    * @param    widget
    *           The child widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    * 
    * @throws  IllegalArgumentException
    *          when the argument is null or instance of {@link TopLevelWindow}
    */
   @Override
   public Container<O> add(int pos, Widget<O> widget)
   {
      if (widget == null || widget instanceof TopLevelWindow)
         throw new IllegalArgumentException();
         
      widgets.add(pos, widget);

      widget.setParent(this);

      return this;
   }

   /**
    * Add the widget before the specified widget.
    * 
    * @param    before
    *           The widget before the new widget will be placed.
    * @param    widget
    *           The child widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    * 
    * @throws  IllegalArgumentException
    *          when the argument is null or instance of {@link TopLevelWindow}
    */
   @Override
   public Container<O> add(Widget<O> before, Widget<O> widget)
   {
      if (widget == null || widget instanceof TopLevelWindow)
         throw new IllegalArgumentException();
     
      int beforePos = widgets.indexOf(before);
      if (beforePos >= 0)
      {
         widgets.add(beforePos, widget);
      }
      else
      {
         widgets.add(widget);
      }

      widget.setParent(this);

      return this;
   }

   /**
    * This method should be overridden in a container that allows clipping.
    *
    * @return  <code>true</code> if container clips its children.
    */
   @Override
   public boolean clipping()
   {
      return false;
   }

   /**
    * Get widget which is in focus.
    *
    * @return  focused widget.
    */
   @Override
   @SuppressWarnings("unchecked")
   public Widget<O> currentFocus()
   {
      return currentFocus(true);
   }

   /**
    * Get widget which is in focus.
    * <p>
    * The method allows to choose whether the focused widget is searched in the nested
    * containers (<code>recurse == true</code>) or whether only this container is searched
    * (<code>recurse == false</code>).
    *
    * @param   recurse
    *          The parameter controls whether the method should recurse to the nested
    *          containers when searching for the focused widget.
    *
    * @return  focused widget.
    */
   public Widget<O> currentFocus(boolean recurse)
   {
      if (current == null)
      {
         return null;
      }

      if (!(current instanceof Container) || !recurse)
      {
         return current;
      }

      if (current instanceof SubMenu)
      {
         SubMenu<?> sm = (SubMenu<?>) current;
         if (!sm.config().displayed || sm.currentFocus() == null)
         {
            return current;
         }
      }

      return ((Container<O>) current).currentFocus();
   }

   /**
    * Get focused widget.
    *
    * @return   focused widget.
    */
   public Widget<O> getFocusInt()
   {
      return current;
   }
   
   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget<O> defaultFocus()
   {
      Widget<O> defFocus = null;

      for (Widget<O> c : widgets)
      {
         if (c.focusTraversable())
         {
            defFocus = c;
            break;
         }
      }

      if (defFocus == null)
      {
         // it is perfectly valid for a container not to have any default
         // focusable widget
         return null;
      }

      if (defFocus instanceof Container)
         return ((Container<O>) defFocus).defaultFocus();

      return defFocus;
   }

   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      if (layoutManager != null)
         layoutManager.doLayout(this);
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * Each container should search for own widgets. Because widgets are relative positioned 
    * inside containers the container physical position is adjusted before search.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<O> findMouseSource(NativePoint p)
   {
      // offset insets of this container if any
      Insets insets = insets();
      if (insets != null && insets != Insets.ZERO_INSETS)
      {
         CoordinatesConversion cc = screen().coordinates();
         NativeInsets ninsets = cc.insetsToNative(insets);
         p = p.offset(-ninsets.left, -ninsets.top);
      }
      
      // to allow z-order awareness, process the widgets in reverse order: the top-most widgets
      // are always last in the widget list
      for (int i = widgets.size() - 1; i >= 0; i--)
      {
         Widget<O> w = widgets.get(i);
         
         // if the widget is not a container and it's mouse actions are not set, then it is ignored
         if (!w.isVisible()                  ||
             !w.physicalBounds().contains(p) ||
             (!(w instanceof Container) && w.getMouseActions().isEmpty()))
         {
            continue;
         }
         
         Widget<O> found = w.findMouseSource(p.minus(w.physicalLocation()));
         
         // MENUBAR can allocate more place than just MENUBAR string itself.
         // If some frame is allocated in the window right under the MENUBAR,
         // than we should properly search mouse event in the controversial area.
         // MENUBAR string and frame cannot intersects, so the algorithm will be the following:
         // MENUBAR will be searched earlier than frames and if it will not find the source child
         // owning the event, than we should search the frames.
         if (found == null && w instanceof Menu && !((Menu<?>)w).config().popupOnly)
         {
            continue;
         }

         return (found == null) ? w : found;
      }

      return null;
   }
   
   /** 
    * Draw widget.
    */
   @SuppressWarnings("unchecked")
   @Override
   public void draw()
   {
      if (!isVisible())
         return;

      O screen = screen();
      Widget<O> parent = parent();
      
      List<Widget<O>> drawableWidgets = widgets;
      if (parent instanceof TitledWindow)
      {
         drawableWidgets = getDrawableWidgets();
      }
      
      int windowId = topLevelWindow().getId().asInt();
      boolean chui = screen().isChui();
      NativeRectangle nouter = OutputManager.getDriver().getPrimitives()
                                            .getScreenBitmap().getOuterRectangle();
      Rectangle outer = nouter == null ? null : screen().coordinates().rectangleFromNative(nouter);
      
      // old method for nested containers
      for (Widget<O> c : prepareWidgets(drawableWidgets))
      {
         if (c.isVisible())
         {
            Rectangle bounds = c.bounds().translate(childrenLocation(c));
            boolean doDraw = false;
            if (chui)
            {
               List<Rectangle> clippings = screen.getClippings(windowId, bounds);
               doDraw = !clippings.isEmpty();
            }
            else
            {
               doDraw = outer == null || bounds.intersects(outer);
            }
            
            if (doDraw)
            {
               c.draw();
            }

            // mark the label widget as "was visible" because we are going to show it
            if (c instanceof LabeledWidget)
            {
               ((LabeledWidget<O>) c).setWasVisible();
            }
         }
      }
   }

   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   public Point childrenLocation(Widget<O> child)
   {
      return child.parent().screenLocation();
   }

   /**
    * Indicates the ending of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    *
    * @param    repaint
    *           If <code>true</code> then the current bounding rectangle and
    *           the rectangle saved during <code>startSizeChange</code> call
    *           are repainted.
    */
   @Override
   public void endSizeChange(boolean repaint)
   {
      WorkArea wa = work.get();
      
      Rectangle savedSize = wa.savedSizes.pop();

      if (repaint)
      {
         if (savedSize != null)
            repaint(savedSize);
         
         repaint();
      }
   }

   /**
    * Returns the focused direct child Component.
    *
    * @return   The focused direct child Component.
    */
   @Override
   @SuppressWarnings("unchecked")
   public Widget<O> focus()
   {
      return current;
   }

   /**
    * Check if widget is focus traversable.
    *
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      if (!super.focusTraversable())
         return false;

      return hasFocusable(null);
   }

   /**
    * Get widget for provided index or <code>null</code> if no such component
    * exists.
    *
    * @param   index
    *          Widget index.
    *
    * @return  widget at specified index or <code>null</code> if index is out
    *          of range.
    */
   @Override
   public Widget<O> get(int index)
   {
      if (index < 0 || index >= widgets.size())
         return null;

      return widgets.get(index);
   }

   /**
    * Get clipping rectangle.
    *
    * @return   clipping rectangle.
    */
   @Override
   public Rectangle getClipRect()
   {
      return new Rectangle(new Point(0, 0), dimension(), screen().coordinates().baseUnits());
   }

   /**
    * Reports if there is at least 1 focusable widget that is not the given
    * widget.
    *
    * @param    exclude
    *           A widget to ignore or <code>null</code> if any focusable
    *           widget will do.
    *
    * @return   <code>true</code> if there is at least one other widget that
    *           can be focused.
    */
   @Override
   public boolean hasFocusable(Widget<O> exclude)
   {
      for (Widget<O> widget : widgets)
      {
         if (widget == exclude)
            continue;

         if (widget.focusTraversable())
            return true;
      }

      return false;
   }

   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      double result = size.height;
      
      // if the size is set, return it
      if (!(result < 0))
         return result;
      
      // if the size of this container has not been set, make it 
      // the size of the parent offset by the relative location
      // and less the insets
      Container<O> p = parent();
      if (p != null)
      {
         result  = p.height();
         if (result >= 0)
         {
            result -= p.insets().heightDelta() + location.y ;
         }
      }
         
      return result < 0 ? 0 : result;
   }

   /**
    * Get container insets.
    *
    * @return  container insets.
    */
   @Override
   public Insets insets()
   {
      return insets;
   }

   /**
    * Get widget minimum size.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      if (layoutManager == null)
         return size;

      return layoutManager.minimumSize(this);
   }

   /**
    * Move specified widget before the widget "before".
    * Widgets must be in the same field-group, that is checked on the server-side.
    * 
    * @param   widget
    *          Widget to move.
    * @param   before
    *          The widget before which to insert mover.
    */
   @Override
   public void moveBeforeTabItem(Widget<O> widget, Widget<O> before)
   {
      if (widgets.size() == 0)
         return;

      if (!widgets.contains(widget))
      {
         throw new RuntimeException("widgets do not contain specified: " + widget);
      }
      
      if (!widgets.contains(before)) 
      {
         throw new RuntimeException("widgets do not contain specified: " + before);
      }

      widgets.remove(widget);
      widgets.add(widgets.indexOf(before), widget);
   }

   /**
    * Move specified widget after the widget "after".
    * Widgets must be in the same field-group, that is checked on the server-side.
    * 
    * @param   widget
    *          Widget to move.
    * @param   after
    *          The widget after which to insert mover.
    */
   @Override
   public void moveAfterTabItem(Widget<O> widget, Widget<O> after)
   {
      if (widgets.size() == 0)
         return;

      if (!widgets.contains(widget))
      {
         throw new RuntimeException("widgets do not contain specified: " + widget);
      }
      
      if (!widgets.contains(after)) 
      {
         throw new RuntimeException("widgets do not contain specified: " + after);
      }

      widgets.remove(widget);
      int idx = widgets.indexOf(after);
      if (idx == widgets.size() - 1)
      {
         widgets.add(widget);
      }
      else
      {
         widgets.add(idx + 1, widget);
      }
   }

   /**
    * Mode specified widget to the bottom of the widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToBottom(Widget<O> widget)
   {
      moveToZpos(widget, 0);
   }

   /**
    * Move specified widget to the top of widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToTop(Widget<O> widget)
   {
      int size = widgets.size();

      if (size == 0)
         return;

      if (!widgets.contains(widget))
         return;

      if (widgets.get(size - 1).equals(widget))
         return;

      widgets.remove(widget);
      widgets.add(widget);
   }

   /**
    * Mode specified widget to the bottom of the widget list considering Z-order class.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public boolean moveToBottomInClass(Widget<O> widget)
   {
      int pos = widgets.indexOf(widget);
      if (pos < 0)
         return false;

      // remove it now, it will always be re-added
      widgets.remove(pos);

      int size = widgets.size();
      int cls = widget.getZOrderClass().ordinal();
      int np = size - 1;
      
      // find first widget in the same z-class and move it in its location (before it);
      // if there are no widgets in the same z-class, then put it on top 
      for (int i = size - 1; i >= 0; i--)
      {
         Widget<?> w = widgets.get(i); 
         int wcls = w.getZOrderClass().ordinal();

         if (wcls == cls)
         {
            np = Math.min(np, i);
         }
      }

      if (np < 0)
      {
         np = 0;
      }

      // move it to the correct location (it may be the same)
      widgets.add(np, widget);
      // this takes care of re-ordering the widgets by their z-class (keeping the order in each
      // specific class)
      normalizeZOrder();
      
      // return true only if the widget's position has moved
      return np != pos;
   }

   /**
    * Move specified widget to the top of widget list considering Z-order class.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public boolean moveToTopInClass(Widget<O> widget)
   {
      int pos = widgets.indexOf(widget);
      if (pos < 0)
      {
         return false;
      }
 
      // remove it now, it will always be re-added
      widgets.remove(pos);

      int size = widgets.size();
      int cls = widget.getZOrderClass().ordinal();
      int np = size - 1;

      // find last widget in the same z-class and move it after it; if there are no widgets
      // in the same z-class, then put it on top
      for (int i = size - 1; i >= 0; i--)
      {
         Widget<?> w = widgets.get(i); 
         int wcls = w.getZOrderClass().ordinal();
         
         if (wcls == cls)
         {
            // found the new class, keep it
            np = Math.max(np, i);
         }
      }

      if (np < 0)
      {
         np = 0;
      }

      np = np + 1;
      // move it to the correct location (it may be the same)
      if (np >= size)
      {
         widgets.add(widget);
      }
      else
      {
         widgets.add(np, widget);
      }
      
      // this takes care of re-ordering the widgets by their z-class (keeping the order in each
      // specific class)
      normalizeZOrder();

      // return true only if the widget's position has moved
      return np != pos;
   }

   /**
    * Move widget to specified position in the widget list. For overlaying
    * widgets this is basically position in widget Z-order.
    *
    * @param   widget
    *          Widget to mover.
    * @param   pos
    *          Position in Z-order.
    */
   @Override
   public void moveToZpos(Widget<O> widget, int pos)
   {
      if (pos < widgets.size() && widgets.contains(widget))
      {
         int ndx = widgets.indexOf(widget);

         if (ndx == pos)
            return;

         if (ndx < pos)
            pos--;

         widgets.remove(widget);
         widgets.add(pos, widget);
      }
   }

   /**
    * Select next focusable widget.
    */
   @Override
   public void nextFocus()
   {
      focusWorker(true);
   }

   /**
    * Select previous focusable widget.
    */
   @Override
   public void prevFocus()
   {
      focusWorker(false);
   }

   /**
    * Process key event.
    *
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
         return;
      
      // TODO process pop-up menu?

      if (current != null && current != this)
         current.processKeyEvent(event);
   }

   /**
    * Remove specified widget from the widget list.
    * <p>
    * The widget will also be destroyed.
    * 
    * @param   widget
    *          Widget to remove.
    */
   @Override
   public void remove(Widget<O> widget)
   {
      // TODO: shouldn't remove only remove from widget list or detach?
      widget.destroy();
   }
   
   /**
    * Detach this widget from this container.  The widget will not be destroyed.
    * 
    * @param   widget
    *          Widget to detach.
    *          
    * @throws  RuntimeException
    *          when the supplied widget is not attached to this container
    */
   @Override
   public void detach(Widget<O> widget)
   {
      boolean removed = widgets.remove(widget);
      
      // fail early here, we don't want to end up with invalid widget tree
      if (!removed)
      {
         String widgetId = widget.getClass().getSimpleName() + ":" + 
                           UiUtils.getWidgetIdAsString(widget, "null");
         String containerId = getClass().getSimpleName() + ":" +
                              UiUtils.getWidgetIdAsString(this, "null");
         
         LOG.log(Level.WARNING, "Cannot detach, widget {0} is not attached to container {1}.", 
                 new Object[] {widgetId, containerId});
      }
         
      // the widget's parent needs to be removed; downstream logic (some of it related to focus
      // processing) is dependent of the widget's parent being set or not.
      widget.setParent(null);

      if (current == widget)
      {
         current = null;
         setFocusInt(currentFocus());
      }
   }
   
   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   public void destroy()
   {
      @SuppressWarnings("unchecked")
      List<Widget<O>> clonedWidgets = (List<Widget<O>>) widgets.clone();
      
      // iterate on a cloned list, the widgets detach themselves 
      for (Widget<O> w : clonedWidgets)
      {
         w.destroy();
      }

      super.destroy();
   }

   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      WorkArea wa = work.get();
      
      if (wa.savedSizes.isEmpty())
         super.repaint();
   }

   /**
    * Replace specified widget with another with preserving widget Z-order.
    *
    * @param   oldWidget
    *          Old widget.
    * @param   newWidget
    *          New widget.
    */
   @Override
   public void replace(Widget<O> oldWidget, Widget<O> newWidget)
   {
      if (oldWidget != null)
      {
         int oldPos = widgets.indexOf(oldWidget);
         
         // the widget will detach during destroy 
         oldWidget.destroy();

         if (current == oldWidget)
            setFocusInt(newWidget);
         
         if (oldPos >= 0)
         {
            add(oldPos, newWidget);
         }
         else
         {
            add(newWidget);
         }
      }
      else
      {
         add(newWidget);
      }
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      Widget<O> focus = currentFocus();

      if (focus == null)
      {
         focus = defaultFocus();
      }
      
      if (focus == this)
      {
         super.requestFocus();
      }
      else 
      {
         if (this instanceof SubMenu)
         {
            super.requestFocus();
         }
         
         if (focus != null)
         {
            focus.requestFocus();
         }
      }
   }

   /**
    * Set focus to specified widget. The method doesn't activate the 
    * <code>focus.topLevelWindow()</code>. If the top-level window needs to be 
    * activated, too, call {@link WindowManager#activateWindow(TitledWindow)}.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   @Override
   public void setFocus(Widget<O> focus)
   {
      setFocusInt(focus);
      Container<O> parent = parent();
      if (parent != null)
      {
         parent.setFocus(this);
      }
   }

   /**
    * Set container height.
    *
    * @param   height
    *          New container height.
    */
   @Override
   public void setHeight(double height)
   {
      setSize(size.width, height);
   }

   /**
    * Set container insets.
    *
    * @param   insets
    *          Insets to set.
    */
   @Override
   public void setInsets(Insets insets)
   {
      this.insets = Coordinate.scale(insets);
   }

   /**
    * Set layout manager for the container.
    *
    * @param   manager
    *          New layout manager.
    */
   @Override
   public void setLayout(LayoutManager<O> manager)
   {
      this.layoutManager = manager;
   }

   /**
    * Get active layout manager for the container.
    *
    * @return  Active layout manager.
    */
   @Override
   public LayoutManager<O> getLayout()
   {
      return layoutManager;
   }

   /**
    * Set container size.
    *
    * @param   size
    *          Size to set.
    */
   @Override
   public void setSize(Dimension size)
   {
      setSize(size.width, size.height);
   }

   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      width = Coordinate.scale(width);
      height = Coordinate.scale(height);
      
      if (isVisible())
      {
         repaint();
      }

      size = new Dimension(width, height);

      if (isVisible())
      {
         repaint(); // post repainting of the new rectangle
      }
   }
   
   /**
    * Set container width.
    *
    * @param   width
    *          Width to set.
    */
   @Override
   public void setWidth(double width)
   {
      setSize(width, size.height);
   }

   /**
    * Indicates the beginning of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    *
    * @param    saveOldRect
    *           If <code>true</code> then the current bounding rectangle is
    *           saved for future repainting.
    */
   @Override
   public void startSizeChange(boolean saveOldRect)
   {
      WorkArea wa = work.get();
      
      if (saveOldRect)
      {
         wa.savedSizes.push(bounds());
      }
      else
      {
         wa.savedSizes.push(null);
      }
   }

   /**
    * Get all widgets stored in the container.
    *
    * @return  array of stored widgets.
    */
   @SuppressWarnings("unchecked")
   @Override
   public Widget<O>[] widgets()
   {
      return widgets.toArray(new Widget[widgets.size()]);
   }

   /**
    * Get all widgets stored in the container.
    *
    * @return  the new list with stored widgets.
    */
   public List<Widget<O>> getWidgetsList()
   {
      return Collections.unmodifiableList(widgets);
   }
   
   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      double result = size.width;
      
      // the size is set, return it
      if (!(result < 0))
         return result;
      
      // if the size of this container has not been set, make it 
      // the size of the parent offset by the relative location
      // and less the insets
      Container<O> p = parent();
      if (p != null)
      {
         result  = p.width();
         if (result >= 0)
         {
            result -= p.insets().widthDelta() + location.x;
         }
      }
         
      return result < 0 ? 0 : result;
   }

   /**
    * Set focus to first focusable widget.
    */
   @Override
   public void firstFocus()
   {
      Widget<O> widget = locateFocusTraversable(widgets);

      if (widget instanceof Container)
         ((Container<O>) widget).firstFocus();

      setFocusInt(widget);
   }

   /**
    * Set focus to last focusable widget.
    */
   @Override
   public void lastFocus()
   {
      Widget<O> widget = locateFocusTraversable(Iterables.reverse(widgets));

      if (widget instanceof Container)
         ((Container<O>) widget).lastFocus();

      setFocusInt(widget);
   }
         
   /**
    * Returns the preferred size of the scrollable area of this container.  
    * <p> 
    * This implementation simply delegates the call to {@link #dimension()}. 
    * 
    * @return  See above.
    */
   public Dimension getScrollableSize()
   {
      return dimension();
   }
   
   /**
    * Calculates the effective size of the scrollable area of this container.
    * <p>
    * This implementation accounts all the visible child widgets into the area and
    * simply determines the furthest x and y from the origin.
    * 
    * @return  see above.
    */
   public Dimension getEffectiveScrollableSize()
   {
      double width = 0, height = 0;

      for (Widget<O> widget : widgets)
      {
         if (widget.isVisible())
         {
            Point origin = widget.location();
            Dimension dimension = widget.dimension();
            
            // virtual width
            double vw = origin.x + dimension.width;
            if (vw > width)
            {
               width = vw;
            }
            // virtual height
            double vh = origin.y + dimension.height;
            if (vh > height)
            {
               height = vh;
            }
         }
      }

      return new Dimension(width, height);
   }
   
   /** 
    * Returns the set of widgets that are never selected for focus.
    * 
    * @return  set of widgets or <code>null</code>. 
    */
   @Override
   public Set<Widget<?>> getIgnoreFocus()
   {
      return Collections.unmodifiableSet(ignoreFocusSet);
   }
   
   /**
    * Sets the set of widgets that are never selected for focus.
    * <code>null</code> may be used to clear the set.
    * 
    * @param   widgets
    *          Set of widgets or <code>null</code>.
    */
   @Override
   public void setIgnoreFocus(Collection<Widget<?>> widgets)
   {
      if (widgets == null || widgets.isEmpty())
      {
         ignoreFocusSet = null;
         return;
      }
      
      ignoreFocusSet = new HashSet<>();
      ignoreFocusSet.addAll(widgets);
   }
   
   /**
    * Reorder widgets according to their ZOrderClass
    * retaining the order within the class
    */
   public void normalizeZOrder()
   {
      Map<ZKey, Widget<O>> m = new TreeMap<ZKey, Widget<O>>();

      int i = 0;
      for (Widget<O> w: widgets)
      {
         m.put(new ZKey(w.getZOrderClass(), i++), w);
      }
      widgets.clear();
      widgets.addAll(m.values());
   }

   /**
    * Set focus to specified widget.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   public void setFocusInt(Widget<O> focus)
   {
      current = focus;
   }
   
   /**
    * Unrealize this container and all its children.
    */
   @Override
   public void unrealize()
   {
      super.unrealize();
      
      for (Widget<O> w : widgets)
      {
         w.unrealize();
      }
   }
   
   /**
    * Prepare widgets before drawing
    * 
    * @param   widgets
    *          Container children.
    *          
    * @return  widgets, prepared for drawing.
    */
   protected List<Widget<O>> prepareWidgets(List<Widget<O>> widgets)
   {
      return Collections.unmodifiableList(widgets);
   }
   
   /**
    * Add widget directly to the list without triggering repaint or other
    * operations.
    *
    * @param   widget
    */
   protected void addDirectly(Widget<O> widget)
   {
      widgets.add(widget);
   }

   /**
    * Check if specified widget belongs to this container.
    *
    * @param   widget
    *          Widget to check.
    *
    * @return  <code>true</code> if widget belongs to container.
    */
   protected boolean belongsTo(Widget<O> widget)
   {
      return widgets.contains(widget);
   }

   /**
    * Build clipped rectangle for given coordinates.
    * 
    * @param   origin
    *          Physical location.
    * @param   dimension
    *          Physical dimension.
    * @param   children
    *          Flag to intersect the rectangle with the to-be-drawn chidren.
    *          
    * @return  see above.
    */
   protected NativeRectangle clipRectangle(NativePoint     origin,
                                           NativeDimension dimension, 
                                           boolean         children)
   {
      NativeRectangle nclip = super.clipRectangle(origin, dimension);
      
      if (children)
      {
         O screen = screen();
         Widget<O> parent = parent();
         
         List<Widget<O>> drawableWidgets = widgets;
         if (parent instanceof TitledWindow)
         {
            drawableWidgets = getDrawableWidgets();
         }
         
         int windowId = parent().topLevelWindow().getId().asInt();
         CoordinatesConversion coords = screen.coordinates();
         
         // this rectangle is computed in relative coordinates.
         Rectangle drawing = null;
         
         // collect the clipping rectangles from all affected children and merge them
         for (Widget<O> c : drawableWidgets)
         {
            if (c.isVisible())
            {
               Point screenLoc = childrenLocation(c);
               Rectangle bounds = c.bounds().translate(screenLoc);
               List<Rectangle> clippings = screen.getClippings(windowId, bounds);
               if (!clippings.isEmpty())
               {
                  // inform the driver to draw only into the specified rectangle
                  Rectangle boundary = null;
                  for (Rectangle clip : clippings)
                  {
                     boundary = (boundary == null ? clip : boundary.union(clip));
                  }

                  boundary = boundary.translate(-screenLoc.x, -screenLoc.y);
                  
                  drawing = (drawing == null ? boundary : drawing.union(boundary));
               }
            }
         }

         if (drawing != null)
         {
            nclip = nclip.intersection(coords.rectangleToNative(drawing));
         }
      }

      return nclip;
   }

   /**
    * Instructs the widget that its size may have changed. The widget updates
    * its state from the associated config instance. It is the responsibility
    * of the caller to call this method only after the config size is set with
    * the new values.
    * <p>
    * This implementation ignores the supplied config instance. Instead it relies
    * on {@link #size} and {@link #config()} to detect whether size has changed.
    * 
    * @param    beforeUpdate
    *           {@link BaseConfig} reference or <code>null</code>.
    */
   @Override
   protected <C extends WidgetConfig> void updateSizeWorker(C beforeUpdate)
   {
      autoResize(beforeUpdate);

      BaseConfig c = UiUtils.getConfig(this, BaseConfig.class);
      
      if (c != null && (size.width != c.widthChars || size.height != c.heightChars))
      {
         setSize(c.widthChars, c.heightChars);
      }
   }
   
   /**
    * Search for the next or previous focusable widget in the container and
    * set focus on it.
    *
    * @param   direction
    *          Widget search direction.
    */
   protected void focusWorker(boolean direction)
   {
      Iterable<Widget<O>> iter = null;
      
      if (current != null)
      {
         iter = (direction) ? Iterables.directFrom(widgets, current) :
                              Iterables.reverseFrom(widgets, current);
      }
      else
      {
         iter = (direction) ? widgets : Iterables.reverse(widgets);
      }

      if (iter == null)
         throw new IllegalStateException("Focused widget does not belong to container");

      for (Widget<O> widget : iter)
      {
         if (checkWidget(widget, direction))
            return;
      }
      
      setFocusInt(null);
      
      // find the nearest focus-traversable parent
      AbstractContainer<O> parent = (AbstractContainer<O>) parent();
      while (parent != null && !parent.focusTraversable())
      {
         parent = (AbstractContainer<O>) parent.parent();
      }
      
      if (parent != null)
      {
         if (direction)
             parent.nextFocus();
         else
             parent.prevFocus();
         return;
      }
      else if (widgets.size() > 0)
      {
         iter = (direction) ? widgets : Iterables.reverse(widgets);

         for (Widget<O> widget : iter)
         {
            if (checkWidget(widget, direction))
               return;
         }
      }
   }

   /**
    * Check potential focus candidate and set focus on it if it matches
    * criteria.
    *
    * @param   widget
    *          Candidate widget.
    * @param   direction
    *          Search direction.
    *
    * @return  <code>true</code> if widget was focused and <code>false</code>
    *          otherwise.
    */
   protected boolean checkWidget(Widget<O> widget, boolean direction)
   {
      if (!widget.focusTraversable())
         return false;

      // check whether an explicit ignore was set for the widget
      if (ignoreFocusSet != null && ignoreFocusSet.contains(widget))
      {
         return false;
      }
      
      if (widget instanceof Container)
      {
         if (direction)
            ((Container<O>) widget).firstFocus();
         else
            ((Container<O>) widget).lastFocus();
      }

      widget.requestFocus();

      return true;
   }
   
   /**
    * Search for the next or previous focusable widget in this container and
    * set focus on it. When the end is reached the focus is moved to the first 
    * widget, similarly when the first widget is reached the focus is moved to
    * the end.
    *
    * @param   direction
    *          Widget search direction.
    */
   protected void cycleFocus(boolean direction)
   {
      List<Widget<O>> widgets = getWidgetsList();
      Iterable<Widget<O>> iter = (direction) ?
                                  Iterables.directFrom(widgets, focus()) :
                                  Iterables.reverseFrom(widgets, focus());
                                     
      for (Widget<O> widget : iter)
      {
         if (checkWidget(widget, direction))
            return;
      }
    
      // cyclic transition to the first or last element
      Widget<O> nextWidget = direction ? widgets.get(0) : widgets.get(widgets.size() - 1);
      checkWidget(nextWidget, direction);
   }

   /**
    * Helper method for traversing widgets in the container.
    *
    * @param   iter
    *          Iterable for the contained widgets.
    *
    * @return  Found widget or <code>null</code> if no matches found.
    */
   private Widget<O> locateFocusTraversable(Iterable<Widget<O>> iter)
   {
      for (Widget<O> widget : iter)
      {
         if (widget.focusTraversable())
            return widget;
      }

      return null;
   }

   /**
    * Get the widgets which can be drawn, from this container.
    * 
    * @return   All the widgets which touch the clipping areas, according to their z-order.
    */
   private List<Widget<O>> getDrawableWidgets()
   {
      Widget<O>[] widgets = widgets();
      int lim = widgets.length;
      int from = 0;
      Widget<O> c = null;
   
      O screen = screen();
      
      int windowId = topLevelWindow().getId().asInt();

      // new method for the top container
      // search down the widgets until all clipped widgets are
      // touched; from those, return the last found frame.
      screen.startScreenSearch(windowId);
      int lastFrom = lim - 1;
      boolean foundContainer = false;
   
      for (from = lim - 1; from >= 0; from--)
      {
         c = widgets[from];
   
         if (!c.isVisible())
         {
            continue;
         }
   
         if (!(c instanceof Container))
         {
            if (screen.markArea(windowId, c.bounds()) && screen.getInvalidChars(windowId) == 0)
            {
               from = lastFrom;
               break;
            }
            continue;
         }
   
         // in redirected mode, only the topmost frame is of interest
         if (screen.isRedirected())
            break;
   
         if (screen.markArea(windowId, c.bounds()))
         {
            lastFrom = from;
            foundContainer = true;
            if (screen.getInvalidChars(windowId) == 0)
            {
               break;
            }
         }
      }
   
      if (from < 0)
      {
         if (foundContainer && screen.getInvalidChars(windowId) > 0)
         {
            // if there are clipped characters left and we found a
            // container, then drawing must start from the last found
            // position
            from = lastFrom;
         }
         else
         {
            from = 0;
         }
      }
   
      List<Widget<O>> res = new ArrayList<>();
      for (int i = from; i < lim; i++)
      {
         res.add(widgets[i]);
      }
      
      return res;
   }

   /**
    * Auxiliary class for use on the normalizaOrder method
    */
   private static class ZKey implements Comparable<ZKey>
   {
      /** Z-order class of the widget */
      private final ZOrderClass cls;
      
      /** initial position of the widget */
      private final int idx;
      
      /**
       * Constructor
       * 
       * @param cls
       *           Z-order class of the widget
       * @param idx
       *           initial position of the widget
       */
      public ZKey(ZOrderClass cls, int idx)
      {
         super();
         this.cls = cls;
         this.idx = idx;
      }
      
      /**
       * The implementation of the {@link java.lang.Comparable#compareTo} method
       */
      @Override
      public int compareTo(ZKey other)
      {
         return this.cls.compareTo(other.cls) == 0 ?
                this.idx - other.idx :
                -this.cls.compareTo(other.cls);
      }
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Container saved sizes stack. */       
      private Stack<Rectangle> savedSizes = new Stack<Rectangle>();
   }
}