/*
** Module   : NativeScrollContainer.java
** Abstract : Scrollable container for native scrolling.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20150831 Created initial version.
** 002 HC  20150920 Improved scroll layout calculation for the case of native scrolling.
** 003 SVL 20151013 Minor fix.
** 004 HC  20160313 Fixed frame scroll layout calculation.
** 005 SVL 20160412 Javadoc fix.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * The class implements a scrollable container capable of
 * so called native scrolling. Native in this case means that 
 * scroll range and scroll position is calculates from the actual 
 * physical dimensions and locations of the parent viewport and 
 * the container itself.
 */
public class NativeScrollContainer<O extends OutputManager<?>>
extends ScrollContainer<O>
{
   /**
    * Returns the scroll dimension of this widget. The scroll dimension is
    * the full size of the widget's content, it may be bigger than widget's
    * visible size.
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getVisibleDimension()} and {@link #getScrollStep()}.
    * 
    * @return   widget's scroll dimension
    */
   @Override
   public NativeDimension getScrollDimension()
   {
      return physicalDimension();
   }

   /**
    * Returns the visible dimension of this widget. The visible dimension is
    * the visible portion of the widget's content (including scrollbars area).
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getScrollDimension()} and {@link #getScrollStep()}.
    *
    * @return   widget's visible dimension
    */
   @Override
   public NativeDimension getVisibleDimension()
   {
      CoordinatesConversion cc = screen().coordinates();
      
      @SuppressWarnings("unchecked")
      ScrollPane<O> pane = parent(ScrollPane.class);

      // do not use viewport's dimension directly, this
      // method is used as part of effective viewport size calculation
      return cc.dimensionToNative(pane.calcInitialViewportSize());
   }

   /**
    * Directs the widget to scroll to the specified X and Y positions.
    * It is the widget's responsibility to update its visuals accordingly.
    * Note that the position values are in the same units as the returned 
    * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
    * and {@link #getScrollStep()}. 
    * 
    * @param    newXPosition
    *           New X scroll position.
    * @param    newYPosition
    *           New Y scroll position.
    */
   @Override
   public void scroll(Integer newXPosition, Integer newYPosition)
   {
      NativePoint p = physicalLocation();
      boolean dirty = false;
      
      if (newXPosition != null && -newXPosition != p.x)
      {
         p.x = -newXPosition;
         dirty = true;
      }
      
      if (newYPosition != null && -newYPosition != p.y)
      {
         p.y = -newYPosition;
         dirty = true;
      }
      
      if (dirty)
      {
         setPhysicalLocation(p.x, p.y);
         repaint();
      }
   }
}