/*
** Module   : SensitiveScrollContainer.java
** Abstract : Scrollable container with support for focusable child widgets.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20150831 Created initial version, most of the focus-handling code is from already 
*                   nonexistent class ScrollContainer.
** 002 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 003 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 004 EVL 20160530 Bottom-right initial scrolling has been disabled for dialog-box in GUI mode.
** 005 HC  20160922 Disabled the container to scroll-to-visible when nested frame focused.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;

/**
 * The class extends native scrolling with support for
 * scrolling focusable child widgets.
 */
public class SensitiveScrollContainer<O extends OutputManager<?>>
extends NativeScrollContainer<O>
{
   /** Track first time when widget is displayed. */
   private boolean first_view = true;
   
   /** Scrollable flag. */
   private boolean scrollable = true;

   /**
    * Hide/show widget.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (first_view)
      {
         Widget<O>[] list = widgets(); 
         
         Widget<O> target = null;
         
         for (Widget<O> comp : list)
         {
            if (target == null)
            {
               target = comp;
               continue;
            }
            
            double y1 = Coordinate.scale(target.location().y + target.height());
            double y2 = Coordinate.scale(comp.location().y + comp.height());

            if (y1 > y2)
            {
               target = comp;
            }
         }
         
         if (target != null)
         {
            ensureVisibility(target);
         }
         
         first_view = false;
      }
      
      super.setVisible(visible);
   }
   
   /**
    * Set the scrollable flag.
    * 
    * @param   scrollable
    *          The new scrollable value.
    */
   public void setScrollable(boolean scrollable)
   {
      this.scrollable = scrollable;
   }

   /**
    * Handling the key event. It is required to intercept this handler to
    * adjust the ScrollContainer contents after moving focus to the next
    * widget which can currently be out of the visible part of the container.
    * In this case the contents is being scrolled to be sure the currently
    * focused component is always visible.
    * 
    * @param   keyEvent
    *          The KeyEvent object to be handled within the ScrollContainer.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      /*
       * First call all KeyListener objects that may have been registered for
       * this component.
       */
      super.processKeyEvent(keyEvent);

      /*
       * Check if any of the KeyListeners consumed the KeyEvent.
       */
      if (keyEvent.isConsumed())
         return;

      int key = keyEvent.actionCode();

      if (!isScrollable())
         return;

      if (key == Keyboard.KA_TAB          || 
          key == Keyboard.KA_BACK_TAB     || 
          key == Keyboard.KA_CURSOR_DOWN  || 
          key == Keyboard.KA_CURSOR_UP    || 
          key == Keyboard.KA_CURSOR_LEFT  || 
          key == Keyboard.KA_CURSOR_RIGHT || 
          key == Keyboard.KA_LEFT_END     || 
          key == Keyboard.KA_RIGHT_END)
      {
         if (ensureVisibility(currentFocus()))
            repaint();
      }
   }

   /**
    * Make sure that given component is visible in the viewport.
    * 
    * @param   comp
    *          The component which should be made visible.
    *          
    * @return  <code>true</code> if there was scrolling.
    * 
    */
   public boolean ensureVisibility(Widget<O> comp)
   {
      // TODO: find the exact rules of ensureVisibility when the focused item is an enclosed frame
      if (!isScrollable() || comp instanceof Frame)
      {
         return false;
      }

      Rectangle r = comp.getVisiblePart();
      
      if (r != null)
      {
         NativeRectangle nr = screen().coordinates().rectangleToNative(r);
         
         // use new logic only for widgets implementing
         // sensible getVisiblePart() method to retain 
         // backward compatibility
         return ensureVisibility(comp.physicalLocation(), nr);
      }
      
      NativePoint      p = comp.physicalLocation();
      NativeDimension  d = comp.physicalDimension();

      // TODO: is this needed?
//      if (comp instanceof Label)
//         d.width = 1;

      Integer scrollX = null;
      Integer scrollY = null;
      
      NativePoint scrollOrigin = physicalLocation();
      NativeDimension scrollSize = physicalDimension();
      NativeDimension viewport = getViewport().physicalDimension();
       
      // scroll up
      if (p.y < -scrollOrigin.y)
      {
         scrollY = p.y;
      }
      
      // scroll left
      if (p.x < -scrollOrigin.x)
      {
         scrollX = p.x;
      }
      
      boolean isGuiDialog = comp != null && UiUtils.locateFrame(comp).isDialog() &&
                            !ThinClient.getInstance().isChui();
      
      // scroll right
      if (p.x + d.width > viewport.width + scrollOrigin.x && !isGuiDialog)
      {
         // increase scroll position so that the widget right is on the right viewport edge,
         // honor the maximum scroll position
         scrollX = Math.min(p.x + d.width - viewport.width, scrollSize.width - viewport.width);
      }
      
      // scroll down
      if (p.y + d.height > viewport.height - scrollOrigin.y && !isGuiDialog)
      {
         // increase scroll position so that the widget bottom is on the bottom viewport edge,
         // honor the maximum scroll position
         scrollY = Math.min(p.y + d.height - viewport.height, scrollSize.height - viewport.height);
      }
      
      if (scrollX != null || scrollY != null)
      {
         EventManager.postEvent(new ScrollEvent(this, scrollX, scrollY));
      }
      
      // scrolled?
      return scrollX != null || scrollY != null;
   }

   /**
    * Set focus to specified widget.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   @Override
   public void setFocus(Widget<O> focus)
   {
      boolean needRepaint = false;
      
      if (belongsTo(focus))
         needRepaint = ensureVisibility(focus);
      
      super.setFocus(focus);
      
      if (needRepaint)
      {
         @SuppressWarnings("unchecked")
         Widget<O> frame = (Widget<O>) UiUtils.locateFrame(this);
         
         if (frame != null)
            frame.repaint();
      }
   }
   
   /**
    * Returns the scroll dimension of this widget. The scroll dimension is
    * the full size of the widget's content, it may be bigger than widget's
    * visible size.
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getVisibleDimension()} and {@link #getScrollStep()}.
    * 
    * @return   widget's scroll dimension
    */
   @Override
   public NativeDimension getScrollDimension()
   {
      return screen().coordinates().dimensionToNative(getEffectiveScrollableSize());
   }

   /**
    * Make sure that given component is visible in the viewport.
    * 
    * @param   p
    *          The component's location.
    * @param   r
    *          The component's sub-rectangle which should be visible as much as possible.
    *          
    * @return  <code>true</code> if there was scrolling.
    * 
    */
   protected boolean ensureVisibility(NativePoint p, NativeRectangle r)
   {
      int x1 = p.x + r.left();
      int y1 = p.y + r.top();
      int x2 = p.x + r.right();
      int y2 = p.y = r.bottom();
      
      NativePoint scrollOrigin = physicalLocation();
      NativeDimension viewport = getViewport().physicalDimension();
      
      int dx = shift4overlap(x1, x2, scrollOrigin.x, scrollOrigin.x + viewport.width);
      int dy = shift4overlap(y1, y2, scrollOrigin.y, scrollOrigin.y + viewport.height);

      if (dx == 0 && dy == 0)
      {
         return false;
      }

      scrollOrigin.x -= dx;
      scrollOrigin.y -= dy;
      setPhysicalLocation(scrollOrigin.x, scrollOrigin.y);
      
      NativePoint scrollPosition = physicalLocation();

      EventManager.postEvent(new ScrollEvent(this, scrollPosition.x, scrollPosition.y));
      return true;
   }

   /**
    * Find min shift offset to achieve max overlap of two line segments  
    *
    * @param x1
    *       start of the first segment
    * @param x2
    *       end of the first segment, x1 &lt; x2
    * @param t1
    *       start of the second segment
    * @param t2
    *       end of the second segment, t1 &lt; 21
    * @return
    *    offset
    */
   private int shift4overlap(int x1, int x2, int t1, int t2)
   {
      return x2 - x1 < t2 - t1? shiftInside(x1, x2, t1, t2) : -shiftInside(t1, t2, x1, x2); 
   }

   /**
    * Find min shift offset to out first line segment inside the second one  
    *
    * @param x1
    *       start of the first segment
    * @param x2
    *       end of the first segment, x1 &lt; x2
    * @param t1
    *       start of the second segment
    * @param t2
    *       end of the second segment, t1 &lt; x2,
    *       
    * @return
    *    offset x2 - x1 &le; t2 - t1
    */
   private int shiftInside(int x1, int x2, int t1, int t2)
   {
      /*
       * t1 <= x1 + dx && x2 + dx <= t2  
       */
      return x1 < t1 ? t1 - x1 : t2 < x2 ? t2 - x2 : 0; 
   }

   /**
    * Check if container can be scrolled.
    * 
    * @return  <code>true</code> if container can be scrolled.
    */
   private boolean isScrollable()
   {
      if (!scrollable)
      {
         return false;
      }
      
      NativeDimension actual  = getScrollDimension();
      NativeDimension visible = getVisibleDimension();
      
      // do not scroll when the visible size is zero, making the
      // widget scrollable would yield unexpected results 
      if (visible.width <= 0 && visible.height <= 0)
      {
         return false;
      }
      
      return actual.width > visible.width || actual.height > visible.height;
   }
}