/*
** Module   : AbstractParameter.java
** Abstract : Common logic for assigns [input-]output parameter values (database fields and normal
**            variables) upon exit from procedure/function
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 OM  20150702 First version based on extracted code form AbstractSimpleParameter and
**                  AbstractExtentParameter for which this is the base class.
** 002 OM  20150723 IdentityHashMap replaces HashMap to keep references to simple parameters.
** 003 OM  20161011 Made getCurrentScope() static. Added Scope.isParameter().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;

/**
 * Base class for all types of parameters. The hierarchy looks like this:
 * {@link AbstractParameter}
 * <ul>
 *    <li>{@link AbstractSimpleParameter}
 *       <ul>
 *          <li>{@link VariableAssigner}
 *          <li>{@link FieldAssigner}
 *       </ul>
 *    <li>{@link AbstractExtentParameter}
 *       <ul>
 *          <li>{@link OutputExtentParameter}
 *             <ul>
 *                <li>{@code ExtentExprN1} (generated-code)
 *                <li>{@link OutputExtentField}
 *             </ul>
 *          <li>{@link InputOutputExtentParameter}
 *             <ul>
 *                <li>{@code ExtentExprN2} (generated-code)
 *                <li>{@link InputOutputExtentField}
 *             </ul>
 *       </ul>
 * </ul>
 * The class does the managements of scopes of for called procedures for both simple and extent
 * parameters, using a {@link ContextLocal} {@code WorkArea}.
 */
abstract public class AbstractParameter
{
   /** Logger */
   protected static final Logger LOG = LogHelper.getLogger(AbstractParameter.class.getName());
   
   /** Object which manages all instances of this class by context. */
   protected static final ContextLocal<WorkArea> context = new ContextLocal<WorkArea>()
   {
      /**
       * Get the weight of this context-local var.
       *
       * @return  Always {@link WeightFactor#LEVEL_5}.
       */
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_5;
      }
      
      /** Retrieve the initial value of this context local variable. */
      @Override
      protected WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /**
    * Retrieve a {@code Scopeable} for the current context. This allows an external object to
    * notify this class' current context of scope open and close events.
    *
    * @return  Context local {@code Scopeable}.
    */
   public static Scopeable getScopeable()
   {
      return context.get();
   }
   
   /**
    * Obtain the current scope from the current context.
    * 
    * @return  current scope from the current context
    */
   protected static Scope getCurrentScope()
   {
      return context.get().getCurrentScope();
   }
   
   /**
    * Check if the {@link BaseDataType} passed as parameter is used for lookup of an 
    * {@code OUTPUT} parameter.
    *
    * @param   key
    *          The variable for lookup in the current scope.
    *
    * @return  {@code true} if the content of {@code key} will be assigned back on return with
    *          success.
    */
   protected static boolean isParameter(BaseDataType key)
   {
      return getCurrentScope().isParameter(key);
   }
   
   /**
    * Get the variable container used as parameter inside the procedure body. It makes sense only
    * for simple parameters.
    *
    * @return  {@code null}. Implementing classes will return the proper field. 
    */
   protected BaseDataType getLocalParam()
   {
      return null;
   }
   
   /**
    * Object which manages the context local state of field assigners. It manages a deque of
    * scopes, each of which contains zero or more {@code OutputParameterAssigner} objects. As the
    * currently executing program's block scopes open and close, these scopes are pushed and
    * popped from the deque.
    */
   protected static class WorkArea
   implements Scopeable
   {
      /** Deque containing field assigners by scope */
      private final Deque<Scope> scopes = new ArrayDeque<>();
      
      /**
       * Default constructor.
       */
      private WorkArea()
      {
         scopes.push(new Scope());
      }
      
      /**
       * Called as each block scope opens, at which time we push a scope onto the deque.
       *
       * @see com.goldencode.p2j.util.Scopeable#scopeStart()
       */
      public void scopeStart()
      {
         scopes.push(new Scope());
      }
      
      /**
       * Called as each block scope closes, at which time we pop a scope from the deque.
       *
       * @see com.goldencode.p2j.util.Scopeable#scopeFinished()
       */
      public void scopeFinished()
      {
         // pop the param list for the inner scope
         scopes.pop();
      }
      
      /** No-op. */
      @Override
      public void scopeDeleted()
      {
         // no-op
      }
      
      /**
       * Get the object which manages field assigners for the current scope.
       *
       * @return  Current scope.
       */
      protected Scope getCurrentScope()
      {
         return scopes.peek();
      }
   }
   
   /**
    * One instance of this class is created for each block scope which the currently executing
    * program enters.  It contains zero or more field assigners specific to that scope.
    * <p>
    * After exiting a function or procedure, the 
    * {@link OutputParameterAssigner#processAssignments(String, boolean)} method is invoked, which
    * causes each field assigner to assign the current value of its parameter value back to the
    * associated DMO property.
    */
   protected static class Scope
   implements com.goldencode.p2j.util.OutputParameterAssigner
   {
      /**
       * Map of output simple parameter assigners for this scope, created lazily. The wrapper
       * outputValue variables are used as keys for quick access.
       */
      private Map<BaseDataType, AbstractSimpleParameter> simpleParameters = null;
      
      /**
       * The list with the wrapped extent parameters, created lazily.
       */
      private List<AbstractExtentParameter> extentParameters = null;
      
      /**
       * Some error occurred. Discard any field assigners.
       */
      @Override
      public void abort()
      {
         clear();
      }
      
      /**
       * Cause each (if any) field assigner in the current scope to assign the current value of
       * its parameter value back to the DMO property with which it is associated.
       *
       * @param   endingRoutine
       *          The legacy name of the routine that is ending when these back assignments are
       *          performed. 
       * @param   function
       *          {@code true} if the called routine is a function.
       */
      @Override
      public void processAssignments(String endingRoutine, boolean function)
      {
         if (simpleParameters != null)
         {
            boolean batch = (simpleParameters.size() > 1);
            
            if (batch)
            {
               RecordBuffer.startBatch();
            }
            
            try
            {
               for (AbstractSimpleParameter param : simpleParameters.values())
               {
                  param.assign();
               }
            }
            finally
            {  
               // TODO: if some exception occur when doing an assign, the next params lose the
               //       assign() call and then the collection is lost ?
               simpleParameters = null;
               
               if (batch)
               {
                  RecordBuffer.endBatch();
               }
            }
         }
         
         if (extentParameters != null)
         {
            // TODO: investigate BATCH
            for (AbstractExtentParameter param : extentParameters)
            {
               param.assign(endingRoutine, function);
            }
            extentParameters = null;
         }
      }
      
      /**
       * Add a parameter assigner to this scope.
       *
       * @param   assigner
       *          Simple field assigner.
       */
      protected void addSimpleParameter(AbstractSimpleParameter assigner)
      {
         if (simpleParameters == null)
         {
            simpleParameters = new IdentityHashMap<>(3);
            
            if (extentParameters == null)
            {
               // register only once, if extentParameters is not null, then we already registered
               TransactionManager.registerOutputParameterAssigner(this);
            }
         }
         
         // use the procedure-local parameter as key for finding and assigning back the value at
         // the end of the procedure to initial Field
         AbstractSimpleParameter old = simpleParameters.put(assigner.getLocalParam(), assigner);
         if (old != null && old != assigner)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING,
                     "Two AbstractOutputParameterAssigner configured with same wrapper variable");
            }
         }
      }
      
      /**
       * Check if the {@link BaseDataType} passed as parameter is used for lookup of an 
       * {@code OUTPUT} parameter.
       * 
       * @param   key
       *          The variable for lookup in the current scope.
       * 
       * @return  {@code true} if the content of {@code key} will be assigned back on return with
       *          success.
       */
      protected boolean isParameter(BaseDataType key)
      {
         return (simpleParameters != null) && simpleParameters.containsKey(key);
      }
      
      /**
       * Update the wrapper variable for a {@code AbstractOutputParameterAssigner}.
       *
       * @param   key
       *          The old key variable. May be the intermediary wrapper for a field or the real
       *          variable passed as *OUTPUT parameter.
       * @param   newVar
       *          The new wrapper variable that will be used inside the calling procedure. At the
       *          end of the call its value will be converted using {@code convCtor} and assigned
       *          to wrapped field or original variable.
       * @param   convCtor
       *          The conversion constructor that will cast the new variable to old variable type
       *          in order to be assigned to field.
       */
      protected void updateSimpleParameter(BaseDataType key,
                                           BaseDataType newVar,
                                           Constructor<?> convCtor)
      {
         AbstractSimpleParameter assigner = null;
         if (simpleParameters == null)
         {
            assigner = new VariableAssigner(newVar, key);
            // side effect: assigners is created an populated with new assigner!
            
            assigner.convCtor = convCtor;
            return; // the side-effect did the trick
         }
         else
         {
            assigner = simpleParameters.remove(key);
            if (assigner == null)
            {
               assigner = new VariableAssigner(newVar, key);
               // side effect: assigner is automatically added to assigners
               
               assigner.convCtor = convCtor;
               return; // the side-effect did the trick
            }
         }
         
         assigner.outputValue = newVar;
         assigner.convCtor = convCtor;
         simpleParameters.put(newVar, assigner);
      }
      
      /**
       * Registers an extent parameter wrapper class to be included in the scope processing.
       *
       * @param   param
       *          Extent parameter wrapper class.
       */
      public void addExtentParameter(AbstractExtentParameter<?> param)
      {
         if (extentParameters == null)
         {
            extentParameters = new LinkedList<>();
            
            if (simpleParameters == null)
            {
               // register only once, if extentParameters is not null, then we already registered
               TransactionManager.registerOutputParameterAssigner(this);
            }
         }
         
         extentParameters.add(param);
      }
      
      /**
       * Clears internal data. All information about simple and extent parameters is removed.
       */
      protected void clear()
      {
         // drop all parameters
         simpleParameters = null;
         extentParameters = null;
      }
   }
}