/*
** Module   : AppServerDefinition.java
** Abstract : The settings of an appserver.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130704 Created initial version.
** 002 CA  20140206 Removed settings which need to be at the associated P2J process, as the
**                  appserver will start using the scheduler and spawner infrastructure.
**                  Moved logging support to use the logging provided by LogHelper.
** 003 OM  20150108 Progress properties read from directory are can be obtained with getProperties
**                  method to be passed to spawned client process environment.
** 004 OM  20150127 Fixed potential NPE when /properties collection is missing from appserver node.
** 005 CA  20150909 Fixed a race condition on DirectoryService usage - as this may be done in a 
**                  separate thread using the server's context, all threads need to sync on this
**                  object, as the DirectoryService is not thread-safe.
** 006 OM  20160425 Project awareness (support for multiple project running in same JVM.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.directory.*;

/**
 * A container of all the settings of an appservers.  Access to the getters and setters is 
 * synchronized, so that in case the setting is changed/refreshed by another thread, concurrent 
 * access is safe.
 */
public class AppServerDefinition
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(AppServerDefinition.class.getName());

   /** A map of all initialized appserver definitions. */
   private static final Map<String, AppServerDefinition> definitions = new HashMap<>();

   /** The appserver's name. */
   private String appServer;

   /** The appserver's operating mode. */
   private AppServerOperatingMode operatingMode;
   
   /**
    * The amount of time (in seconds) to wait for an agent to be available, before attempting
    * to auto-start a new one. 
    */
   private Integer requestTimeout;
   
   /**
    * The number of seconds waiting before auto-trimming the agents, if they exceed the
    * minimum capacity.
    */
   private Integer autoTrimTimeout;
   
   /** Appserver's explicit propath. */
   private String propath;
   
   /** The number of agents to start initially. */
   private Integer initialAgents;
   
   /** The minimum number of agents required for this appserver. */
   private Integer minAgents;
   
   /** The maximum number of agents which can be alive. */
   private Integer maxAgents;
   
   /** The activate procedure; may be null/empty. */
   private String activate;
   
   /** The deactivate procedure; may be null/empty. */
   private String deactivate;
   
   /** The connect procedure; may be null/empty. */
   private String connect;
   
   /** The disconnect procedure; may be null/empty. */
   private String disconnect;
   
   /** The startup procedure; may be null/empty. */
   private String startup;
   
   /** The shutdown procedure; may be null/empty. */
   private String shutdown;
   
   /** The parameter for the startup procedure; may be null/empty. */
   private String startupParameter;
   
   /** The project token. Will be set on the context deserving the app-server. */
   private String projectToken;
   
   /**
    * Environment settings incoming from progress.ini, global per appserver.
    * The spawner will set these into the OS environment directly, before launching the child
    * process.
    */
   private Map<String, String> properties = null;
   
   /**
    * Read the definition for the given appserver, and save it in this instance.
    * 
    * @param    appServer
    *           The application's server name.
    * 
    * @throws   IllegalArgumentException
    *           If no appserver (with the given name) is configured in the directory.
    */
   private AppServerDefinition(String appServer)
   {
      this.appServer = appServer;
      
      refresh();
   }

   /**
    * Get the settings for the specified appserver.  If not yet initialized, it will read the
    * settings for the appserver from the directory.  The created instance will be shared by all
    * parties.
    * <p>
    * The initial call (which creates the instance) must be coming from a thread with its context
    * authorized to access the <code>appservers</code> node for the P2J server.
    * 
    * @param    appServer
    *           The name of the appserver.
    *           
    * @return   The instance containing the settings for the specified appserver.
    */
   public static AppServerDefinition get(String appServer)
   {
      AppServerDefinition def = null;

      synchronized (definitions)
      {
         def = definitions.get(appServer);
         if (def == null)
         {
            def = new AppServerDefinition(appServer);
            definitions.put(appServer, def);
         }
      }
      
      return def;
   }

   /**
    * Remove the definition for the given appserver.
    * 
    * @param    appServer
    *           The appserver name.
    */
   static void remove(String appServer)
   {
      synchronized (definitions)
      {
         definitions.remove(appServer);
      }
   }
   
   /**
    * Refresh this appserver instance from the directory.  Must be called from threads which have
    * their context authorized to access the <code>appservers</code> node for the P2J server.
    */
   public synchronized void refresh()
   {
      // access the directory service
      DirectoryService ds = DirectoryService.getInstance();
      
      synchronized (ds)
      {
         if (!ds.bind())
         {
            throw new RuntimeException("directory bind failed");
         }
         
         String appSrvPath = Utils.findDirectoryNodePath(ds,
                                                         "appservers/" + appServer,
                                                         "appserver",
                                                         false);
         
         if (appSrvPath == null)
         {
            throw new IllegalArgumentException(
               "No appserver named '" + appServer +"' is configured in the directory!");
         }
         
         String mode = ds.getNodeString(appSrvPath, "operating_mode");
         this.operatingMode   = AppServerOperatingMode.fromString(mode);
         
         this.requestTimeout  = ds.getNodeInteger(appSrvPath, "request_timeout");
         this.autoTrimTimeout = ds.getNodeInteger(appSrvPath, "auto_trim_timeout");
         this.propath         = ds.getNodeString(appSrvPath, "propath");
         this.initialAgents   = ds.getNodeInteger(appSrvPath, "initial_agents");
         this.minAgents       = ds.getNodeInteger(appSrvPath, "min_agents");
         this.maxAgents       = ds.getNodeInteger(appSrvPath, "max_agents");
         this.activate        = ds.getNodeString(appSrvPath, "activate");
         this.deactivate      = ds.getNodeString(appSrvPath, "deactivate");
         this.connect         = ds.getNodeString(appSrvPath, "connect");
         this.disconnect      = ds.getNodeString(appSrvPath, "disconnect");
         
         if (operatingMode.isSessionFree() && LOG.isLoggable(Level.WARNING))
         {
            if (connect != null && connect.length() > 0)
            {
               LOG.logp(Level.WARNING, "AppServerDefinition", "refresh", 
                        String.format("Connect procedure specified for appserver %s in " +
                                      "Session-free operating mode", appServer));
            }
            if (disconnect != null && disconnect.length() > 0)
            {
               LOG.logp(Level.WARNING, "AppServerDefinition", "refresh", 
                        String.format("Disconnect procedure specified for appserver %s in " +
                                      "Session-free operating mode", appServer));
            }
         }
         this.startup          = ds.getNodeString(appSrvPath, "startup");
         this.shutdown         = ds.getNodeString(appSrvPath, "shutdown");
         this.startupParameter = ds.getNodeString(appSrvPath, "startup_parameter");
         this.projectToken     = ds.getNodeString(appSrvPath, "project_token");
         
         properties = new HashMap<>();
         String propertiesPath = appSrvPath + "/properties";
         String[] nodes = ds.enumerateNodes(propertiesPath);
         if (nodes != null)
         {
            for (String node : nodes)
            {
               String val = ds.getNodeString(propertiesPath + "/" + node, "value");
               if (val != null)
               {
                  properties.put(node, val);
               }
            }
         }
         
         // last use of the directory, end our session with it
         ds.unbind();
      }
   }
   
   /**
    * Getter for the {@link #appServer} field.
    * 
    * @return   See above.
    */
   public synchronized String getAppServer()
   {
      return appServer;
   }

   /**
    * Setter for the {@link #appServer} field.
    * 
    * @param    appServer
    *           The new value for the {@link #appServer} field.
    */
   public synchronized void setAppServer(String appServer)
   {
      this.appServer = appServer;
   }

   /**
    * Getter for the {@link #operatingMode} field.
    * 
    * @return   See above.
    */
   public synchronized AppServerOperatingMode getOperatingMode()
   {
      return operatingMode;
   }

   /**
    * Setter for the {@link #operatingMode} field.
    * 
    * @param    operatingMode
    *           The new value for the {@link #operatingMode} field.
    */
   public synchronized void setOperatingMode(AppServerOperatingMode operatingMode)
   {
      this.operatingMode = operatingMode;
   }

   /**
    * Getter for the {@link #requestTimeout} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getRequestTimeout()
   {
      return requestTimeout;
   }

  /**
   * Setter for the {@link #requestTimeout} field.
   * 
   * @param    requestTimeout
   *           The new value for the {@link #requestTimeout} field.
   */
   public synchronized void setRequestTimeout(Integer requestTimeout)
   {
      this.requestTimeout = requestTimeout;
   }

   /**
    * Getter for the {@link #autoTrimTimeout} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getAutoTrimTimeout()
   {
      return autoTrimTimeout;
   }

   /**
    * Setter for the {@link #autoTrimTimeout} field.
    * 
    * @param    autoTrimTimeout
    *           The new value for the {@link #autoTrimTimeout} field.
    */
   public synchronized void setAutoTrimTimeout(Integer autoTrimTimeout)
   {
      this.autoTrimTimeout = autoTrimTimeout;
   }

   /**
    * Getter for the {@link #propath} field.
    * 
    * @return   See above.
    */
   public synchronized String getPropath()
   {
      return propath;
   }

   /**
    * Setter for the {@link #propath} field.
    * 
    * @param    propath
    *           The new value for the {@link #propath} field.
    */
   public synchronized void setPropath(String propath)
   {
      this.propath = propath;
   }

   /**
    * Getter for the {@link #initialAgents} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getInitialAgents()
   {
      return initialAgents;
   }

   /**
    * Setter for the {@link #initialAgents} field.
    * 
    * @param    initialAgents
    *           The new value for the {@link #initialAgents} field.
    */
   public synchronized void setInitialAgents(Integer initialAgents)
   {
      this.initialAgents = initialAgents;
   }

   /**
    * Getter for the {@link #minAgents} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getMinAgents()
   {
      return minAgents;
   }

   /**
    * Setter for the {@link #minAgents} field.
    * 
    * @param    minAgents
    *           The new value for the {@link #minAgents} field.
    */
   public synchronized void setMinAgents(Integer minAgents)
   {
      this.minAgents = minAgents;
   }

   /**
    * Getter for the {@link #maxAgents} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getMaxAgents()
   {
      return maxAgents;
   }

   /**
    * Setter for the {@link #maxAgents} field.
    * 
    * @param    maxAgents
    *           The new value for the {@link #maxAgents} field.
    */
   public synchronized void setMaxAgents(Integer maxAgents)
   {
      this.maxAgents = maxAgents;
   }

   /**
    * Getter for the {@link #activate} field.
    * 
    * @return   See above.
    */
   public synchronized String getActivate()
   {
      return activate;
   }

   /**
    * Setter for the {@link #activate} field.
    * 
    * @param    activate
    *           The new value for the {@link #activate} field.
    */
   public synchronized void setActivate(String activate)
   {
      this.activate = activate;
   }

   /**
    * Getter for the {@link #deactivate} field.
    * 
    * @return   See above.
    */
   public synchronized String getDeactivate()
   {
      return deactivate;
   }

   /**
    * Setter for the {@link #deactivate} field.
    * 
    * @param    deactivate
    *           The new value for the {@link #deactivate} field.
    */
   public synchronized void setDeactivate(String deactivate)
   {
      this.deactivate = deactivate;
   }

   /**
    * Getter for the {@link #connect} field.
    * 
    * @return   See above.
    */
   public synchronized String getConnect()
   {
      return connect;
   }

   /**
    * Setter for the {@link #connect} field.
    * 
    * @param    connect
    *           The new value for the {@link #connect} field.
    */
   public synchronized void setConnect(String connect)
   {
      this.connect = connect;
   }

   /**
    * Getter for the {@link #disconnect} field.
    * 
    * @return   See above.
    */
   public synchronized String getDisconnect()
   {
      return disconnect;
   }

   /**
    * Setter for the {@link #disconnect} field.
    * 
    * @param    disconnect
    *           The new value for the {@link #disconnect} field.
    */
   public synchronized void setDisconnect(String disconnect)
   {
      this.disconnect = disconnect;
   }

   /**
    * Getter for the {@link #startup} field.
    * 
    * @return   See above.
    */
   public synchronized String getStartup()
   {
      return startup;
   }

   /**
    * Setter for the {@link #startup} field.
    * 
    * @param    startup
    *           The new value for the {@link #startup} field.
    */
   public synchronized void setStartup(String startup)
   {
      this.startup = startup;
   }

   /**
    * Getter for the {@link #shutdown} field.
    * 
    * @return   See above.
    */
   public synchronized String getShutdown()
   {
      return shutdown;
   }

   /**
    * Setter for the {@link #shutdown} field.
    * 
    * @param    shutdown
    *           The new value for the {@link #shutdown} field.
    */
   public synchronized void setShutdown(String shutdown)
   {
      this.shutdown = shutdown;
   }

   /**
    * Getter for the {@link #startupParameter} field.
    * 
    * @return   See above.
    */
   public synchronized String getStartupParameter()
   {
      return startupParameter;
   }

   /**
    * Setter for the {@link #startupParameter} field.
    * 
    * @param    startupParameter
    *           The new value for the {@link #startupParameter} field.
    */
   public synchronized void setStartupParameter(String startupParameter)
   {
      this.startupParameter = startupParameter;
   }
   
   /**
    * Get the project token that this agent serves. If not set then the default is used.
    * 
    * @return  the {@code prokectToken} the agent serves.
    */
   public String getProjectToken()
   {
      return projectToken;
   }
   
   /**
    * Getter for {@link #properties} properties.
    * 
    * @return  a temporary copy of the private map of environment properties. Returns
    *          <code>null</code> is the map is <code>null</code> or empty.
    */
   public Map<String, String> getProperties()
   {
      if (properties == null || properties.isEmpty())
      {
         return null; // quick out
      }
      
      Map<String, String> ret = new HashMap<>();
      ret.putAll(properties);
      return ret;
   }
}