/*
** Module   : CommonErrorStatus.java
** Abstract : Define all APIs available via the ERROR-STATUS system handle.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130216 Added support for ERROR-OBJECT-DETAIL attribute.
** 003 VMN 20130219 Extended from Errorable interface.
** 004 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 005 EVL 20131008 Adding legacy attributes and methods annotation.
** 006 GES 20141124 Javadoc fix.
** 007 CA  20151017 Added setter for NUM-MESSAGES.
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Exposes the following methods or attributes from the ERROR-STATUS handle:
 * <table summary="">
 * <tr>
 *    <th>ERROR-STATUS attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>NUM-MESSAGES</td>
 *    <td>{@link #numErrors()}</td>
 * </tr>
 * <tr>
 *    <td>GET-MESSAGE(n)</td>
 *    <td>{@link #getErrorText(int)} and {@link #getErrorText(integer)}</td>
 * </tr>
 * <tr>
 *    <td>GET-NUMBER(n)</td>
 *    <td>{@link #getErrorNumber(int)} and {@link #getErrorNumber(integer)}
 *    </td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrapErrorStatus} API, which, depending on the 
 * handle's state, it either returns an {@link ErrorManager#asHandle()} 
 * instance or an dynamic proxy using {@link handle#invalidAttrAccessProxy} 
 */
@LegacyResource(resource = LegacyResource.PSEUDO_WIDGET)
public interface CommonErrorStatus
extends CommonHandle,
        Errorable
{
   /**
    * Returns the Progress compatible error text associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, an empty string "" is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error text at that index or the empty string "" if the
    *           list is empty or if the index is invalid.
    */
   @LegacyMethod(name = "GET-MESSAGE")
   public character getErrorText(integer index);
   
   /**
    * Returns the Progress compatible error text associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, an empty string "" is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error text at that index or the empty string "" if the
    *           list is empty or if the index is invalid.
    */
   @LegacyMethod(name = "GET-MESSAGE")
   public character getErrorText(int index);

   /**
    * Returns the number of the errors (recorded in the error list) which
    * have occurred since the last call to 
    * {@link ErrorManager#silentErrorEnable}.
    *
    * @return    Current number of errors in the error list if silent mode
    *            has been enabled. This will be 0 if no errors have occurred  
    *            since the last <code>silentErrorEnable</code> call OR if  
    *            <code>silentErrorEnable</code> has never been called.
    */
   @LegacyAttribute(name = "NUM-MESSAGES")
   public integer numErrors();

   /**
    * Undocumented feature encountered in customer code. After all, NUM-MESSAGES (KW_NUM_MSG)
    * attribute seems to be writable.
    * 
    * @param   errCount
    *          New value.
    */
   @LegacyAttribute(name = "NUM-MESSAGES", setter = true)
   public void setNumErrors(int64 errCount);
   
   /**
    * Returns the Progress compatible error number associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, 0 is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error number at that index or 0 if the list is empty or
    *           if the index is invalid.
    */
   @LegacyMethod(name = "GET-NUMBER")
   public integer getErrorNumber(integer index);
   
   /**
    * Returns the Progress compatible error number associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, 0 is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error number at that index or 0 if the list is empty or
    *           if the index is invalid.
    */
   @LegacyMethod(name = "GET-NUMBER")
   public integer getErrorNumber(int index);

   /**
    * Returns a handle for the ERROR-OBJECT-DETAIL attribute, set in cases of SOAP-related faults.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "ERROR-OBJECT-DETAIL")
   public handle getErrorObjectDetail();
}