/*
** Module   : CommonLastEvent.java
** Abstract : Define all APIs available via the LAST-EVENT system handle.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130221 Added support for LABEL attribute (not supported attributes call appropriate
**                  API to show error).
** 003 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 004 EVL 20131008 Adding legacy attributes and methods annotation.
** 005 VIG 20131122 Added new method for EVENT-TYPE support.
** 006 GES 20141124 Javadoc fix.
** 007 OM  20150924 Added new read-only specific attributes.
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 CA  20160327 Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Exposes the following methods or attributes from the LAST-EVENT handle:
 * <table summary="">
 * <tr>
 *    <th>LAST-EVENT attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>CODE</td>
 *    <td>{@link #getLastKey()}</td>
 * </tr>
 * <tr>
 *    <td>FUNCTION</td>
 *    <td>{@link #getLastFunction()}</td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrapLastEvent} API, which, depending on the 
 * handle's state, it either returns an {@link KeyReader#asHandle()} 
 * instance or a dynamic proxy using {@link handle#invalidAttrAccessProxy}. 
 */
@LegacyResource(resource = LegacyResource.PSEUDO_WIDGET)
public interface CommonLastEvent
extends CommonHandle,
        XYCoordinates
{
   /**
    * Obtains the last key read from any source (stream or terminal).
    *
    * @return   The last key read.
    */
   @LegacyAttribute(name = "CODE")
   public int getLastKey();

   /**
    * Implementation of the LAST-EVENT system handle's FUNCTION attribute.
    *
    * @return   The key function of the last key event.
    */
   @LegacyAttribute(name = "FUNCTION")
   public character getLastFunction();

   /**
    * Implementation of the LAST-EVENT system handle's EVENT-TYPE attribute.
    *
    * @return   The key function of the last key event.
    */
   @LegacyAttribute(name = "EVENT-TYPE")
   public character getEventType();
   
   /**
    * Implementation of the WIDGET-ENTER system handle's EVENT-TYPE attribute.
    *
    * @return  A handle, in a trigger associated with an ENTRY event or a LEAVE event, to the
    * next widget to receive input focus.
    */
   @LegacyAttribute(name = "WIDGET-ENTER")
   public handle getWidgetEnter();
   
   /**
    * Implementation of the WIDGET-LEAVE system handle's EVENT-TYPE attribute.
    *
    * @return  A handle, in a trigger associated with an ENTRY event or a LEAVE event, to the
    * next widget to receive input focus.
    */
   @LegacyAttribute(name = "WIDGET-LEAVE")
   public handle getWidgetLeave();
   
   /**
    * Implementation of the ON-FRAME-BORDER system handle's EVENT-TYPE attribute.
    *
    * @return  {@code YES} if the last event was a mouse event that occurred on a frame border.
    */
   @LegacyAttribute(name = "ON-FRAME-BORDER")
   public logical isOnFrameBorder();
}