/* 
** Module   : ConnectHelper.java
** Abstract : Helper methods for parsing the options in a connect-like method.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130908 Created initial version.
** 002 GES 20150204 Remove enclosing single quote characters in string values.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.util.*;

/**
 * This is a helper class to bring to central point the options which can be passed to a connect
 * like methods. The options are passed as a string to {@link #parseOptions}; using the list of
 * valid options, this will be able to identify no-value options (which act like flags) and also
 * options which must receive an int value.
 * <p>
 * The <code>-pf file</code> option can be used to read a set of options from a file, and override
 * any existing ones.
 */
public class ConnectHelper
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ConnectHelper.class);

   /**
    * Parse and validate the given options; a map is returned with each parameters's value. If a
    * parameter doesn't have a value, the map will contain <code>null</code>.
    *  
    * @param    options
    *           The options from which the parameters will be extracted.
    * @param    knownOptions
    *           Set of valid options.
    * @param    noValueOptions
    *           Set of options which require no argument, i.e. "flags".
    * @param    intOptions
    *           Set of options which requre an int argument.
    *
    * @return   The parameter-to-value map.
    */
   public static Map<String, String> parseOptions(character            options,
                                                  Set<String>          knownOptions,
                                                  Set<String>          noValueOptions,
                                                  Map<String, Integer> intOptions)
   {
      Map<String, String> parms = new HashMap<String, String>();
      String[] opts = options.toStringMessage().split(" ");
   
      String arg = null;
      
      for (int i = 0; i < opts.length; i++)
      {
         String txt = opts[i];
         
         if (txt.length() == 0)
            continue;
         
         boolean err = false;
         
         // the 4GL allows enclosing single or double quotes that it silently ignores 
         txt = removeEnclosingQuotes(txt);
         
         if (txt.startsWith("-"))
         {
            if (!knownOptions.contains(txt))
            {
               err = true;
            }
            else
            {
               arg = txt;
            }
            
            // some options do not have a corresponding value
            if (noValueOptions.contains(arg))
            {
               parms.put(arg, null);
               arg = null;
            }
         }
         else
         {
            if (arg != null)
            {
               if (intOptions.containsKey(arg))
               {
                  // ensure that int values can be parsed
                  int value = 0;
                  try
                  {
                     value = Integer.parseInt(txt);
                  }
                  catch (NumberFormatException e)
                  {
                     // ignore
                  }
                  int def = intOptions.get(arg);
                  if (value <= def)
                  {
                     final String msg1 = "The %s parameter requires an argument greater than %d";
                     final String msg2 = "Unable to process parameters";
                     ErrorManager.displayError(1404, String.format(msg1, arg, def), false);
                     ErrorManager.displayError(5509, msg2, false);
                     return null;
                  }
               }
   
               parms.put(arg, txt);
            }
            else
            {
               err = true;
            }
            
            arg = null;
         }
         
         if (err)
         {
            final String format = "Could not recognise argument: %s";
            final String msg1 = String.format(format, txt);
            final String msg2 = "Unable to process parameters";
            ErrorManager.displayError(301, msg1);
            ErrorManager.displayError(5509, msg2, false);
            return null;
         }
      }
      
      if (parms.containsKey("-pf"))
      {
         String pf = parms.get("-pf");
         Map<String, String> pfParms = readPfFile(pf);
         if (pfParms == null)
         {
            final String msg = "Unable to open parameter file %s, errno 2";
            ErrorManager.recordOrShowError(1247, String.format(msg, pf), false);
   
            return null;
         }
         
         // the parameters from the file are not validated
         parms.putAll(pfParms);
      }
      
      return parms;
   }

   /**
    * Parse the value for the specified key as an int, or default to the specified value if the
    * key does not exist.
    * 
    * @param    parms
    *           The options map.
    * @param    key
    *           The key to search.
    * @param    def
    *           The default value to return if the key does not exist.
    *           
    * @return   See above.
    */
   public static int getInt(Map<String, String> parms, String key, int def)
   {
      String sval = parms.get(key);
      if (sval == null)
      {
         return def;
      }
   
      int val = 0;
      try
      {
         return Integer.parseInt(sval);
      }
      catch (NumberFormatException e)
      {
         val = 0;
      }
      
      return val;
   }

   /**
    * Get a string from the parameters map, using the specified key and defaulting to the 
    * specified value, if it does not exist.
    * 
    * @param    parms
    *           The options map.
    * @param    key
    *           The key to search.
    * @param    def
    *           The default value to return if the key does not exist.
    *           
    * @return   See above.
    */
   public static String getString(Map<String, String> parms, String key, String def)
   {
      String val = parms.get(key);
      
      return val == null ? def : val;
   }

   /**
    * Read the parameters configured in the given file.
    * 
    * @param    file
    *           The file containing the parameters.
    *           
    * @return   A map containing the value for each parameter or <code>null</code> if there were
    *           problems reading the file.
    */
   public static Map<String, String> readPfFile(String file)
   {
      if (FileSystemOps.searchPath(file).isUnknown())
      {
         // file does not exist.
         return null;
      }
      
      StreamWrapper stream = new StreamWrapper();
      stream.assign(StreamFactory.openFileStream(file, false, false));
      Map<String, String> result = new HashMap<String, String>();
      try
      {
         String line = stream.readLn();
         while (line != null)
         {
            line = line.trim();
            if (line.startsWith("-"))
            {
               int sharpIdx = line.indexOf('#');
               if (sharpIdx >= 0)
               {
                  // drop trailing comment
                  line = line.substring(0, sharpIdx).trim();
               }
               
               // this is a configuration line
               String param = line;
               int spaceIdx = param.indexOf(' ');
               if (spaceIdx >= 0)
               {
                  param = param.substring(0, spaceIdx);
                  // drop the parameter, to leave only the value (if any);
                  line = line.substring(spaceIdx).trim();
               }
               
               String val = (line.length() == 0 ? null : line);
               
               result.put(param,  val);
            }
         }
      }
      catch (EOFException e)
      {
         // ignore
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Problems reading parameter file " + file, e);
         }

         // read error
         return null;
      }
      catch (InterruptedException e)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Problems reading parameter file " + file, e);
         }

         // other error
         return null;
      }

      return result;
   }
   
   /**
    * Remove matched pairs of enclosing single or double quote characters from the given string
    * and return the result.  If there is more than one sequential matched pair of quotes, the
    * extra pairs of quotes will also be removed if and only if those quote characters are the
    * same as the first quote character that was removed. In other words, this will remove
    * multiple matching pairs of single quotes or multiple matching pairs of double quotes but
    * it will NOT remove pairs of double quotes when the first quotes are single quotes (and
    * vice versa).  This is how the 4GL works.  Yes, really.
    *
    * @param    txt
    *           Text to clean up.
    *
    * @return   The text without enclosing quotes.  Mostly (see the mixed quotes issue above).
    */
   private static String removeEnclosingQuotes(String txt)
   {
      if (txt == null || txt.length() == 0)
         return txt;
      
      char rmch = txt.charAt(0);
      
      if (rmch == '\'' || rmch == '\"')
      {
         int len = 0;
         
         do
         {
            txt = StringHelper.stripEnclosing(txt, rmch);
            len = txt.length();
         }
         while (len > 1 && txt.charAt(0) == rmch && txt.charAt(len - 1) == rmch);
      }
   
      return txt;
   }
   
   /**
    * Test the quote removal processing.
    *
    * @param    args
    *           Ignored.
    */
   public static void main(String[] args)
   {
      String[][] tests = new String[][]
      {
         { "'aaa'",                       "aaa"               },
         { "''aaa''",                     "aaa"               },
         { "''aaa'''",                    "aaa'"              },
         { "'''aaa''",                    "'aaa"              },
         { "'''''aaa'''",                 "''aaa"             },
         { "'''aaa'''",                   "aaa"               },
         { "''''''aaa''''''",             "aaa"               },
         { "''\"'''aaa'''\"''",           "\"'''aaa'''\""     },
         { "''\"'''aaa'''''",             "\"'''aaa'''"       },
         { "'aaa\"",                      "'aaa\""            },
         { "'aaa",                        "'aaa"              },
         { "aaa'",                        "aaa'"              },
         { "\"aaa\"",                     "aaa"               },
         { "\"\"aaa\"\"",                 "aaa"               },
         { "\"\"aaa\"\"\"",               "aaa\""             },
         { "\"\"\"aaa\"\"",               "\"aaa"             },
         { "\"\"\"\"\"aaa\"\"\"",         "\"\"aaa"           },
         { "\"\"\"aaa\"\"\"",             "aaa"               },
         { "\"\"\"\"\"\"aaa\"\"\"\"\"\"", "aaa"               },
         { "\"\"'\"\"\"aaa\"\"\"'\"\"",   "'\"\"\"aaa\"\"\"'" },
         { "\"\"'\"\"\"aaa\"\"\"\"\"",    "'\"\"\"aaa\"\"\""  },
         { "\"aaa'",                      "\"aaa'"            },
         { "\"aaa",                       "\"aaa"             },
         { "aaa\"",                       "aaa\""             }
      };
      
      int fails = 0;
      
      for (int i = 0; i < tests.length; i++)
      {
         if (!tests[i][1].equals(removeEnclosingQuotes(tests[i][0])))
         {
            fails++;
            
            System.out.printf("FAIL for test %d; input %s; output %s; expected %s;\n",
                              i,
                              tests[i][0],
                              removeEnclosingQuotes(tests[i][0]),
                              tests[i][1]);
         }
      }
      
      if (fails == 0)
         System.out.printf("All tests passed successfully.\n");
   }
}