/*
** Module   : ErrorConditionException.java
** Abstract : represents a Progress ERROR condition
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 GES 20050811   @22111 Created initial version which represents a
**                           Progress ERROR condition.
** 002 GES 20060505   @25986 Added a flag to differentiate the case where
**                           the cause of this condition was the END-ERROR
**                           key.
** 003 GES 20060623   @27561 Support storing a key code as a cause.
** 004 SVL 20090908   @43848 Added progressErrorCode field. 
** 005 CA  20130612          Set the cause of this condition a NumberedException referring the
**                           error code and associated error message.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.NumberedException;

/**
 * Represents a Progress ERROR condition.
 */
public class ErrorConditionException
extends ConditionException
{
   /** <code>true</code> if the cause of this error was the END-ERROR key. */
   private boolean end_error = false;

   /** Error code used in Progress 4GL. */
   private int progressErrorCode = -1;
   
   /**
    * Constructor which accepts an error message.
    *
    * @param    message
    *           Error message.
    */
   public ErrorConditionException(String message)
   {
      super(message);
   }

   /**
    * Constructor which accepts an error message and an error code.
    *
    * @param    progressErrorCode
    *           Error code (which is used in Progress 4GL).
    * @param    message
    *           Error message.
    */
   public ErrorConditionException(int progressErrorCode, String message)
   {
      super(message);
      this.progressErrorCode = progressErrorCode;
      initCause(new NumberedException(message, progressErrorCode));
   }
    
   /**
    * Constructor which accepts an error message, error code and root cause.
    *
    * @param    progressErrorCode
    *           Error code (which is used in Progress 4GL).
    * @param    message
    *           Error message.
    * @param    cause
    *           A numbered exception identifying the raised condition.
    */
   public ErrorConditionException(int progressErrorCode, String message, NumberedException cause)
   {
      super(message, cause);
      this.progressErrorCode = progressErrorCode;
   }

  /**
    * Constructor which accepts an error message and the key code for the key
    * press that caused this condition.
    *
    * @param   message
    *          Error message.
    * @param   keycode
    *          The key code of the key that caused this condition to be
    *          raised.
    */
   public ErrorConditionException(String message, int keycode)
   {
      super(message, keycode);
   }
    
   /**
    * Constructor which accepts an error message and a root cause. This
    * variant is used when the error condition is detected as the result of
    * catching another exception.
    *
    * @param    message
    *           Error message.
    * @param    rootCause
    *           Exception which has been caught in the course of processing.
    */
   public ErrorConditionException(String message, Throwable rootCause)
   {
      super(message, rootCause);
   }
    
   /**
    * Constructor which accepts and wrappers a root cause. This variant is
    * used when the error condition is detected as the result of catching
    * another exception, and the message or context of the root cause is
    * sufficient to describe the problem.
    *
    * @param    rootCause
    *           Exception which has been caught in the course of processing.
    */
   public ErrorConditionException(Throwable rootCause)
   {
      super(rootCause);
   }
    
   /**
    * Constructor which accepts an error message.
    *
    * @param    message
    *           Error message.
    * @param    end_error
    *           <code>true</code> if the cause of this condition was the
    *           END-ERROR key.
    */
   public ErrorConditionException(String message, boolean end_error)
   {
      super(message);
      this.end_error = end_error;
   }
   
   /**
    * Constructor which accepts an error message and the key code for the key
    * press that caused this condition.
    *
    * @param    message
    *           Error message.
    * @param    keycode
    *           The key code of the key that caused this condition to be
    *           raised.
    * @param    end_error
    *           <code>true</code> if the cause of this condition was the
    *           END-ERROR key.
    */
   public ErrorConditionException(String  message, 
                                  int     keycode,
                                  boolean end_error)
   {
      super(message, keycode);
      this.end_error = end_error;
   }

  /**
    * Detects if the cause of this condition was the END-ERROR key.
    *
    * @return   <code>true</code> if the cause of this condition was the
    *           END-ERROR key.
    */
   public boolean isEndError()
   {
      return end_error;
   }
   
   /**
    * Sets the END-ERROR flag for the cause of this condition.
    *
    * @param    end_error
    *           <code>true</code> if the cause of this condition was the
    *           END-ERROR key.
    */
   public void setEndError(boolean end_error)
   {
      this.end_error = end_error;
   }

   /**
    * Returns the code (which is used in Progress 4GL) of the error that was
    * the cause of this condition.
    *
    * @return   The code of the error that caused this condition or -1 if it
    *           wasn't specified.
    */
   public int getProgressErrorCode()
   {
      return progressErrorCode;
   }
}