/*
** Module   : FileStream.java
** Abstract : represents a file or device for reading/writing
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20050712   @21743 Created initial version which provides input and output semantics
**                           for any file-like file system object (file or device).
** 002 GES 20050801   @21939 New assign() interface for support of shared streams.
** 003 GES 20050824   @22212 Added error condition support.
** 004 GES 20060217   @24671 Added writeByte().
** 005 GES 20060225   @24739 Added call to cleanup() in close().
** 006 GES 20060525   @26517 Avoid setLength() on a device during init.
** 007 GES 20060905   @29227 Fix for readCh() to properly handle EOF.
** 008 NVS 20070725   @34685 Implemented methods isIn() and isOut(), which are newly added to the
**                           base class as abstract.
** 009 SIY 20070928   @35318 Fix for PUT/DISPLAY handling regressions.
**                           Ref: #35314
** 010 GES 20080414   @38015 Support standard character conversion (just the default behavior,
**                           no user-definable source and/or target codepages yet). This
**                           allows other languages (and the upper ASCII range in the latin
**                           codepages) to be properly handled. Added support for the no-convert
**                           option as well, to disable this new support.
** 011 GES 20080724   @39235 Major rewrite to use NIO and memory mapped file support to massively
**                           improve perf.
** 012 ECF 20090526   @42464 Fixed long file defect. An offset >2GB passed to map() was being
**                           truncated when mapped to a MappedByteBuffer.
** 013 SVL 20110403          write(byte[]) was made public.
** 014 EVL 20121204          readLn() has been changed to use indexOfNewLine() to detect the line
**                           separators. This allows to properly handle the input files coming
**                            from OSother than one the P2J runtime is running.
** 015 OM  20130607          Made file and map(long) protected to be accessible in derived classes.
** 028 EVK 20140523          Added unbuffered stream operation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;

/**
 * A stream class supporting input and output semantics for any file-like
 * file system object (file or device).  The backing construct is not really
 * a "stream" because it must provide methods to get and set the current
 * offset (the current read/write position) into the stream.
 *
 * @author    GES
 */
public class FileStream
extends Stream
{
   /** The maximum size of mapped memory "chunks". */
   private static final int MAPPED_BUF_SZ = (256 * 1024);
   
   /** The size of the non-memory mapped write buffer. */
   private static final int WRITE_BUF_SZ = 4096;
   
   /** Charset name to use for encoding or <code>null</code> for default. */
   private static final String override = Utils.getCharsetOverride();
   
   /** Optimized character set conversion helper. */
   private static final CharsetConverter cc = new CharsetConverter(override);
   
   /** <code>true</code> if the file is larger than the maximum chunk size. */
   private boolean chunk = false;
   
   /** <code>true</code> if using memory mapped file support. */
   private boolean mapped = false;
   
   /** <code>true</code> if the file must be writable. */
   private boolean write = false;
   
   /** The backing construct for the "stream". */
   protected FileChannel file = null;
   
   /** Buffer for memory mapped file usage. */
   private MappedByteBuffer mem = null;
   
   /** The starting offset (in bytes) of the memory mapped buffer. */
   private long offset = 0;
   
   /** The size (in bytes) of the memory mapped buffer. */
   private long memsz = 0;
   
   /** Buffer used for non-memory-mapped writes. */
   private byte[] buffer = null;
   
   /** Number of bytes in the write buffer (non-memory-mapped mode). */
   private int pending = 0;
   
   /**
    * Constructs an instance using a filename, this file or device will be
    * opened for reading or writing based on the given <code>write</code>
    * flag.  The current read/write position will be at offset 0 unless in
    * <code>write</code> mode and the <code>append</code> flag is on.
    *
    * @param    filename
    *           The name of the file or device to open.
    * @param    write
    *           If <code>true</code>, open in write mode, otherwise only
    *           open the file for reading.
    * @param    append
    *           In <code>write</code> mode, if <code>true</code>, the current
    *           file pointer is set to the end of the file. Otherwise if in
    *           <code>write</code> mode, the file is truncated to 0 length.
    *           This value is ignored in <code>read</code> mode.
    *
    * @throws   ErrorConditionException
    *           - If the file does not exist.
    *           - If the current user does not have the proper rights to open
    *             the file or device.
    */
   public FileStream(String filename, boolean write, boolean append)
   throws ErrorConditionException
   {
      String errmsg = null;
      int    errnum = 0;
      
      try
      {
         File   target = new File(filename);
         String mode   = (write ? "rw" : "r");
         
         // create a traditional file system resource with the proper mode
         RandomAccessFile raf = new RandomAccessFile(target, mode);
         
         // obtain the backing channel (we no longer need access to the raf)
         file = raf.getChannel();
         
         // remember if the file is read-only or is read-write
         this.write = write;
         
         if (!write)
         {
            // read only files shift directly into memory mapped mode at the start of the file
            map(0);
         }
         else
         {
            // writable files always start writing at the end of the file
            // which means that memory mapped support is not very useful
            // (since memory mapped file regions that are mapping parts of
            // the file that are past the end, cause the file to be extended
            // when the file should in fact not extend)
            mapped = false;
            
            // init the file pointer if as long as the file is not a device like /dev/null
            if (target.isFile())
            {
               if (append)
               {
                  file.position(file.size());
               }
               else
               {
                  file.truncate(0);
               }
            }
         }
      }
      
      catch (FileNotFoundException fnf)
      {
         errmsg = "\"" + filename + "\" was not found";
         errnum = 293;
      }
      
      catch (SecurityException se)
      {
         errmsg = "You do not have permission to access " + filename;
         errnum = 294;
      }
      
      catch (IOException ioe)
      {
         // this error message is missing the " Errno=" portion that would
         // be there in Progress since there is no concept of this in Java
         // --> we also don't know if there are other cases in which this
         //     exception might map to a different error text and error num
         errmsg = "Unable to open file:" + filename + ".";
         errnum = 98;
      }
      
      // if there was an error, honor silent error mode
      if (errmsg != null)
      {
         ErrorManager.recordOrThrowError(errnum, errmsg);
      }
   }

   /**
    * The number of bytes available to be immediately read without blocking.
    *
    * @return   The number of available bytes.
    */
   public long available()
   throws IOException
   {
      long num = 0;
      
      if (mapped)
      {
         num = mem.remaining();
         
         if (num == 0)
         {
            // we must be at the end of the memory region, re-calc if there
            // is anything left
            num = file.size() - (offset + memsz);
         }
      }
      else
      {
         // we must be writing at the end of the file because that is the
         // only non-mapped use case, but check this to confirm
         if (buffer == null)
         {
            // WARNING: this can be expensive
            num = file.size() - file.position();
         }
      }
      
      return num;
   }
   
   /**
    * The 0-based offset into the stream at which the next read or write will 
    * occur.
    *
    * @return   The current position in the stream.
    */
   public long getPos()
   throws IOException
   {
      long pos = 0;
      
      if (mapped)
      {
         pos = offset + mem.position();
      }
      else
      {
         flushBuffered();
         pos = file.position();
      }
      
      return pos;
   }
   
   /**
    * Moves the current read/write position to the specified absolute 0-based
    * offset.  
    *
    * @param    pos
    *           The new read/write position in the stream.
    */
   public void setPos(long pos)
   throws IOException
   {
      long sz = getLen();
      
      if (pos < sz)
      {
         if (mapped)
         {
            if (pos >= offset && pos < (offset + memsz))
            {
               // the target location is inside the current memory region
               mem.position((int) (pos - offset));
            }
            else
            {
               // change our mapping
               map(pos);
            }
         }
         else
         {
            // we can shift into mapped mode
            map(pos);
         }
      }
      else
      {
         if (mapped)
            unmap();
         
         // extend to the end or past the end of the file
         file.position(pos);
      }
   }
   
   /**
    * State of the input side of the stream.
    *
    * @return   <code>true</code> if the input side of the stream is active.
    */
   public boolean isIn()
   {
      return (file != null);
   }
   
   /**
    * State of the output side of the stream.
    *
    * @return   <code>true</code> if the output side of the stream is active.
    */
   public boolean isOut()
   {
      return (file != null);
   }   
   
   /**
    * The length of the stream in bytes.
    *
    * @return   The length of the stream.
    */
   public long getLen()
   throws IOException
   {
      // this may change the file size
      flushBuffered();
      
      return file.size();
   }
   
   /**
    * Truncates or extends the stream to the specified length if this stream
    * supports such an operation. If truncation is requested, all data
    * located after this point in the file is discarded.  If extending the
    * file is requested, the values of the data in the extended portion of
    * the file is undefined. 
    *
    * @param    len
    *           The new length of the file.
    */
   public void setLen(long len)
   throws IOException
   {
      long    sz    = getLen();
      long    pos   = getPos();
      boolean remap = false;
      
      if (len <= sz)
      {
         if (mapped)
         {
            if (len < (offset + memsz))
            {
               // we are truncating the file in a way that affects the mem
               // mapping
               unmap();
               
               remap = true;
            }
            
            // "true up" our current file position so the truncate can
            // manage the position properly
            file.position(pos);
         }
         
         // normal truncation
         file.truncate(len);
      }
      else
      {
         // file extend (Progress has no counterpart to this)
         file.position(len);
         
         if (mapped)
         {
            if (pos < len && pos + MAPPED_BUF_SZ > len)
            {
               remap = true;
            }
         }
      }
      
      if (remap && pos < len)
      {
         // re-establish our mapping
         map(pos);
      }
   }
   
   /**
    * Write the given character to the output stream.
    *
    * @param    ch
    *           The character to be written.
    */
   public void writeCh(char ch)
   throws IOException
   {
      write(convert ? cc.toByte(ch) : (byte) ch);
   }
   
   /**
    * Write the given byte to the output stream.
    *
    * @param    b
    *           The byte to be written.
    */
   public void writeByte(byte b)
   throws IOException
   {
      // a byte is a byte, there should be no need for charset conversion
      write(b);
   }
   
   /**
    * Write the given string to the output stream.
    *
    * @param    data
    *           The data to be written.
    */
   public void write(String data)
   throws IOException
   {
      byte[] bdat = null;
      
      if (!convert)
      {
         int len = data.length();
         bdat = new byte[len];
         
         // disable character conversion
         for (int i = 0; i < len; i++)
         {
            // directly access the character, however it is encoded (it may
            // be unicode or it may have been something else if the data was
            // input via a manner that bypassed encoding), drop the high
            // byte and store the result
            bdat[i] = (byte) data.charAt(i);
         }
      }
      else
      {
         // conversion path
         bdat = (override == null) ? data.getBytes()
                                   : data.getBytes(override);
      }
      
      write(bdat);
   }
   
   /**
    * Read the raw integer value of a character from the current read
    * position in the stream. This method DOES NOT convert the value read
    * from the stream's character set into Unicode.
    * <p>
    * At this time, only a single byte will be read no matter what encoding
    * is used by the underlying stream.  This means that DBCS or Unicode
    * encodings will not be properly handled.
    *
    * @return   The next character read from the stream, -1 on any failure
    *           and -2 upon an <code>EOF</code>.
    */
   public int readCh()
   {
      int ch = 0;
      
      try
      {
         ch = readWorker();
         
         if (ch == -1)
         {
            // signal EOF
            ch = -2;
         }
      }
      
      catch (IOException ioe)
      {
         // non-EOF error thrown during read
         ch = -1;
      }
      
      return ch;
   }
   
   /**
    * Read all characters from the current read position in the stream to the
    * next line separator (as determined by the <code>File.separator</code>
    * or to the <code>EOF</code>. Any line separator character(s) and the
    * <code>EOF</code> character are not returned.
    *
    * @return   The next line read from the stream.
    */
   public String readLn()
   throws EOFException,
          IOException,
          InterruptedException
   {
      StringBuilder sb = new StringBuilder();
      
      while (indexOfNewLine(sb) < 0)
      {
         int ch = read();
         
         if (ch == -1)
         {
            // EOF
            break;
         }
         
         sb.append((char) ch);
      }
      
      return readLineCleanup(sb);
   }
   
   /**
    * Closes the input stream and releases OS resources associated with it.
    */
   public void closeIn()
   {
      close();
   }
   
   /**
    * Closes the output stream and releases OS resources associated with it.
    */
   public void closeOut()
   {
      close();
   }
   
   /**
    * Closes the stream and releases OS resources associated with it.
    */
   public void close()
   {
      try
      {
         endOutput();
         cleanup(true, true);
         
         // force anything pending to be written to the file
         if (mapped)
         {
            unmap();
         }
         else
         {
            flushBuffered();
            
            // de-allocate our buffer in case our reference is kept around
            buffer = null;
         }
         
         if (file != null)
         {
            file.close();
            file = null;
         }
      }
      
      catch (IOException ioe)
      {
         // nothing to do
      }
   }
   
   /**
    * Assigns the internal stream reference to the given reference.
    *
    * @param    stream
    *           The new internal stream reference to use for all operations.
    */
   public void assign(Stream stream)
   throws UnsupportedOperationException
   {
      throw new UnsupportedOperationException("Cannot handle this.");
   }
   
   /**
    * Worker to output a single byte to our file at the current output
    * position.
    *
    * @param    b
    *           The data to write.
    */
   private void write(byte b)
   throws IOException
   {
      if (mapped && !mem.hasRemaining())
      {
         // try to get the next chunk (if it fails we must be at the end of
         // the file and we will shift into unmapped mode)
         map(offset + memsz);
      }
         
      // if we are still mapped at this point, we must have room for a byte
      if (mapped)
      {
         mem.put(b);
      }
      else
      {
         // we must be writing to the end of the file
         writeWorker(b);
      }
   }
   
   /**
    * Write the given byte array to the output stream.
    *
    * @param    data
    *           The data to be written.
    *
    * @throws   IOException
    *           If an I/O error occurs.
    */
   public void write(byte[] data)
   throws IOException
   {
      int start = 0;
      int len   = data.length;
      
      while (mapped && len > 0)
      {
         int num = mem.remaining();
         
         if (num <= 0)
         {
            // no room at all, try to get the next chunk (if it fails we must
            // be at the end of the file and we will shift into unmapped mode)
            map(offset + memsz);
         }
         else
         {
            if (num >= len)
            {
               // there is space and enough of it to output everything left
               mem.put(data, start, len);
               
               // done
               return;
            }
            else
            {
               // partial space left
               mem.put(data, start, num);
               
               start += num;
               len   -= num;
            }
         }
      }
         
      // we may have done partial output above, but if we shifted out of
      // mapped mode or were not in mapped mode, then we must output now
      if (!mapped && len > 0)
      {
         // we must be writing to the end of the file
         writeWorker(data, start, len);
      }
   }
   
   /**
    * Write a byte to the buffer, flushing if the buffer is full.
    *
    * @param    b
    *           The byte to write.
    */
   private void writeWorker(byte b)
   throws IOException
   {
      boolean written = false;
      
      if (pending < WRITE_BUF_SZ)
      {
         if (buffer == null)
            allocate();
         
         // there is room
         buffer[pending] = b;
         pending++;
         written = true;
      }
      
      if (pending == WRITE_BUF_SZ)
      {
         // the buffer is full
         flushBuffered();
         
         if (!written)
         {
            buffer[pending] = b;
            pending++;
         }
      }
   }
   
   /**
    * Worker that provides a mechanism to write a buffer directly to the
    * file channel at the current write position.
    *
    * @param    data
    *           The buffer to write.
    * @param    offset
    *           The starting index to output.
    * @param    len
    *           The number of bytes to output.
    */
   private void writeWorker(byte[] data, int offset, int len)
   throws IOException
   {
      // if the buffer will be filled up or overflowed by this new data OR if
      // there is nothing in the buffer but the new data is larger than or
      // equal to the buffer size, then output everything (buffered and new)
      // immediately
      if ((pending + len) >= WRITE_BUF_SZ)
      {
         if (pending > 0)
         {
            // write from 2 buffers in 1 call
            ByteBuffer[] bufs = new ByteBuffer[2];
            bufs[0] = ByteBuffer.wrap(buffer, 0, pending);
            bufs[1] = ByteBuffer.wrap(data, offset, len);
            
            int[] sz = new int[2];
            sz[0] = pending;
            sz[1] = len;
            
            writeFully(bufs, sz, (pending + len));
            pending = 0;
         }
         else
         {
            // nothing buffered, just make 1 simple call to write
            ByteBuffer buf = ByteBuffer.wrap(data, offset, len);
            writeFully(buf, len);
         }
      }
      else
      {
         // can fit in buffer
         allocate();
         
         int end = offset + len;
         
         // copy the data
         for (int i = offset; i < end; i++)
         {
            buffer[pending] = data[i];
            pending++;
         }
      }
      afterWrite();
   }
   
   /**
    * Worker to read a character and hide any charset conversion that may
    * or may not be needed.
    *
    * @return   The next character read or -1 if EOF is reached.
    */
   private int read()
   throws IOException
   {
      int ch = readWorker();
      
      if (convert && ch >= 0)
      {
         ch = cc.toChar(ch);
      }
      
      return ch;
   }
   
   /**
    * Read a byte from the file at the current file pointer.
    *
    * @return   The next character read or -1 if EOF is reached.
    */
   private int readWorker()
   throws IOException
   {
      int ch = -1;
      
      if (mapped && !mem.hasRemaining())
      {
         // try to get the next chunk (if it fails we must be at the end of
         // the file and we will shift into unmapped mode)
         map(offset + memsz);
      }
         
      // unmapped mode only occurs at the end of the file, so if we are
      // not mapped, then there is nothing to read, protect from this case
      if (mapped)
      {
         // there must be at least 1 byte available to read (it must be
         // widened without sign extension)
         ch = (mem.get() & 0x000000FF);
      }
      
      return ch;
   }
   
   /**
    * Allocate the write buffer if it is not already allocated.
    */
   private void allocate()
   {
      if (buffer == null)
      {
         buffer  = new byte[WRITE_BUF_SZ];
         pending = 0;
      }
   }
   
   /**
    * Output any buffered writes to the file. This only affects writes at
    * the end of the file (non-mapped mode). If called while in mapped mode
    * this is a NOP.
    */
   private void flushBuffered()
   throws IOException
   {
      if (!mapped && buffer != null && pending > 0)
      {
         ByteBuffer buf = ByteBuffer.wrap(buffer, 0, pending);
         writeFully(buf, pending);
         pending = 0;
      }
   }
   
   /**
    * Write <code>len</code> bytes from the given buffer to the file before
    * returning.
    *
    * @param    buf
    *           The buffer to output.
    * @param    len
    *           The size of the buffer in bytes.  This MUST NOT be different
    *           than the amount of data in the buffer.  It is provided here
    *           only for efficiency since the callers of this method already
    *           have this value.
    */
   private void writeFully(ByteBuffer buf, int len)
   throws IOException
   {
      int num = 0;
      
      while (num < len)
      {
         int next = file.write(buf);
         
         num += next;
         
         if (next > 0 && num < len)
         {
            // there must be more in the buffer
            buf.compact();
         }
      }
      afterWrite();
   }
   
   /**
    * Write <code>len</code> bytes from the given buffer to the file before
    * returning. This takes advantage of the NIO scatter/gather functionality
    * to improve performance.
    *
    * @param    buf
    *           The list of buffers to output.
    * @param    sz
    *           The size in bytes of each corresponding buffer. The length of
    *           this array must be the same as that of <code>buf</code>.
    * @param    len
    *           The cumulative size of the data in all buffers in bytes.  This
    *           MUST NOT be different than the amount of data in the buffer.
    *           It is provided here only for efficiency since the callers of
    *           this method already have this value.
    */
   private void writeFully(ByteBuffer[] buf, int[] sz, int len)
   throws IOException
   {
      int num    = 0;
      int start  = 0;
      int amount = buf.length;
      
      while (num < len)
      {
         int next = (int) file.write(buf, start, amount);
         
         num += next;
         
         if (next > 0 && num < len)
         {
            int remain = next;
            
            // there must be more in the buffer, find the place in the list
            // of buffers where we need to restart writing
            for (int i = start; i < amount && remain > 0; i++)
            {
               if (sz[i] <= remain)
               {
                  // this entry was fully output, bypass it on the next write
                  start++;
                  amount--;
                  remain -= sz[i];
               }
               else
               {
                  // this entry was partially written, compact it
                  sz[i] -= remain;
                  buf[i].compact();
                  remain = 0;
               }
            }
         }
      }
      afterWrite();
   }

   /**
    * Flushes this output stream and forces any buffered output bytes to be written out.
    * The general contract of flush is that calling it is an indication that,
    * if any bytes previously written have been buffered by the implementation of the output
    * stream, such bytes should immediately be written to their intended destination.
    *
    * @throws   IOException
    *           if an I/O error occurs.
    */
   @Override
   protected void flushData()
   throws IOException
   {
      if (file != null)
      {
         file.force(true);
      }
   }
   
   /**
    * Setup memory mapping starting at the given offset and extending to
    * the end of the file or to the maximum chunk size, whichever is
    * smaller. If the file size is 0 or the starting offset is greater than
    * or equal to the file size, then shift into unmapped mode.
    *
    * @param    start
    *           The starting offset (0-based) in bytes.
    */
   protected void map(long start)
   throws IOException
   {
      // flush if this is a re-mapping
      if (mapped)
      {
         if (mem != null)
         {
            mem.force();
         }
      }
      else
      {
         flushBuffered();
      }
      
      // this must follow any flushing since the file size can change as
      // as result of flushing buffered writes
      long sz = file.size();
      
      if (sz == 0 || start >= sz)
      {
         if (mapped)
            unmap();
         
         return;
      }
      
      long deferred = -1;
      
      mapped = true;
      
      // by default try to map the whole file, but if the file is bigger than
      // our maximum chunk size, then we shift into chunk mode
      if (sz > MAPPED_BUF_SZ)
      {
         chunk = true;
         sz    = Math.min((sz - start), MAPPED_BUF_SZ);
      }
      else
      {
         // we can map the whole file, make sure that the starting offset
         // is reset to 0
         if (start > 0)
         {
            deferred = start;
            start = 0;
         }
      }
      
      // remember the starting offset and size of the mapped buffer
      offset = start;
      memsz  = sz;
      
      mem = file.map(write ? FileChannel.MapMode.READ_WRITE
                           : FileChannel.MapMode.READ_ONLY,
                     start,
                     sz);
                     
      // force the file "position" to the expected location even though we
      // mapped the whole file
      if (deferred != -1)
      {
         mem.position((int) deferred);
      }
   }
   
   /**
    * Clear/release the memory mapped buffer.
    */
   private void unmap()
   {
      // flush before clearing state
      if (mem != null)
         mem.force();
      
      mapped = false;
      chunk  = false;
      offset = 0;
      memsz  = 0;
      mem    = null;
   }
}