/*
** Module   : FileSystemDaemon.java
** Abstract : worker object that handles delegated file system processing
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20060211   @24539 Created initial version, which completely
**                           implements the FileSystem interface.
** 002 GES 20070914   @35141 Refactored the propath searching code to
**                           use a static worker. This allows the same
**                           routine to be used at conversion time for
**                           searching.
** 003 GES 20080118   @36862 Added NPE protection logic in search().
** 004 GES 20080123   @36903 More NPE protection logic.
** 005 GES 20080321   @37630 Added more data (almost complete) to the
**                           FILE-INFO support.
** 006 GES 20110712          Searching now honors the case-sensitivity of the
**                           legacy system.
** 007 GES 20111025          Modified API export code to capture and use the
**                           authentication token so that multiple registrations
**                           can be done per JVM session. This occurs when the
**                           client code is running in the server process and
**                           there is a CTRL-C that causes the client to
**                           restart.
** 008 GES 20111111          Modified interface export to allow registration
**                           to be exclusively local or remote as needed.
** 009 CS  20121104          Implement PROPATH assignment.
** 010 CS  20130115          Task 1613: finish FILE-INFO system handle support
** 011 CA  20140313          Fixed the regexp for win-separator replacement.
**                           Added delete(File, recursive). 
** 012 ECF 20150715          Replace StringBuffer with StringBuilder.
** 013 EVL 20151026          Adding more criteria to detect file path relative to the current
**                           directory.  For UNIX-like systems added the path beginning not with
**                           File separator.  For Windows systems  the current path must not start
**                           with File separator and root drive letter + ':\\'.  The canonical
**                           name is now used for this case too in getCaseInsensitiveMatch()
**                           filename transformation.  If root or current path search is failed
**                           the PROPATH entries are used as last resort.
** 014 IAS 20160113          Added image files' search with different extensions
** 015 EVL 20160329          Adding the ability to check the environment variable from OS first
**                           and if not found - use defined in JVM or via -D java start option.
** 016 OM  20160428          Added searchAll() method.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.net.*;
import com.goldencode.util.*;

/**
 * Implements file-system utility methods (for query and manipulation) in a
 * manner that is suitable for use on a remote node.  Construction of this
 * object causes these APIs to be exported from the current node.
 */
public class FileSystemDaemon
implements FileSystem
{
   /** Windows path separator */
   private static final String WIN_FILE_SEPARATOR = "\\\\";

   /** Token used to authenticate with the dispatcher when registering APIs. */
   private static Object modToken = null;

   /** Paths to search for relative filenames. */
   private String[] paths = null;
   
   /** The legacy system's filesystem case-sensitivity. */
   private boolean caseSens = true;
   
   /**
    * Create an instance and export its API to the network.
    *
    * @param    single
    *           <code>true</code> to startup within the server process which
    *           must bypass the shared infrastructure initialization. Use
    *           <code>false</code> for the normal client JVM startup.
    */
   public FileSystemDaemon(boolean single)
   {
      synchronized (FileSystemDaemon.class)
      {
         modToken = RemoteObject.registerServer(FileSystem.class, this, modToken, single);
      }
   }
   
   /**
    * Check if the file system is case-sensitive
    * 
    * @return <code>true</code> if the file system is case-sensitive
    */
   public boolean isCaseSensitive()
   {
      return caseSens;
   }
   
   /**
    * Search for the given file in each directory listed in the paths specified. If the file
    * system object exists and is a file, then the filename is returned. If the path list contains
    * an empty string or a '.', then the current directory is searched, otherwise only the listed
    * directories are searched.
    * <p>
    * If the filename is specified as an absolute name, this specific filename is checked and no
    * path searching is done.
    *
    * @param   paths
    *          The list of paths to search. The file separator character used in these paths
    *          should be the same as the current system's file separator character.
    * @param   caseSens
    *          {@code true} if the search should honor a case-sensitive file-system.  When set to
    *          {@code false} and the current filesystem is naturally case-sensitive, a
    *          case-insensitive search will be implemented.
    * @param   filename
    *          The absolute or relative filename for which to search.
    * @param   allowDir
    *          If {@code true} the search will also directories, if {@code false} only file
    *          results will be returned.
    *
    * @return  If the file is found in the current directory, the relative filename is returned.
    *          If the file is absolute or found in any other path, the full absolute filename is
    *          returned.  If the file is not found or the referenced file system object is not a
    *          file, {@code null} is returned.
    */
   public static String search(String[] paths,
                               boolean  caseSens,
                               String   filename,
                               boolean  allowDir)
   {
      // make sure there is something to search for (the code below needs to be protected)
      if (filename == null)
      {
         return null;
      }
      
      // if the the OS is windows replace all / to \ from the search paths
      String fn = PlatformHelper.isUnderWindowsFamily() ? 
            filename.replaceAll("/", WIN_FILE_SEPARATOR) : filename;
      
      String result = getAbsoluteOrCurrent(caseSens, fn, allowDir); 
      if (result != null)
      {
         return result;
      }
      
      // not found in root or current, try PROPATH entries
      if (paths == null)
      {
         return null;
      }
      
      return Utils.tryWhileNull(Arrays.asList(paths), 
            path -> getInPath(caseSens, path, fn, allowDir));
   }
   
   /**
    * Return all files matching the specified {@code filename} from each directory listed in the
    * {@code paths} specified. If the path list contains an empty string or a '.', then the
    * current directory is searched, otherwise only the listed directories are searched.
    * <p>
    * If the filename is specified as an absolute name, this specific filename is checked and no
    * path searching is done and returned as a single element array.
    *
    * @param   paths
    *          The list of paths to search. The file separator character used in these paths
    *          should be the same as the current system's file separator character.
    * @param   caseSens
    *          {@code true} if the search should honor a case-sensitive file-system.  When set to
    *          {@code false} and the current filesystem is naturally case-sensitive, a
    *          case-insensitive search will be implemented.
    * @param   filename
    *          The absolute or relative filename for which to search.
    * @param   allowDir
    *          If {@code true} the search will also directories, if {@code false} only file(s)
    *          results will be returned.
    *          
    * @return  The relative filename for all matching files are returned as {@code String} list.
    *          If the file is absolute or found in any other path, the full absolute filename is
    *          returned as a single element. If the file is not found or the referenced file
    *          system object is not a file, {@code null} is returned.
    */
   public static List<String> searchAll(String[] paths,
                                        boolean  caseSens,
                                        String   filename,
                                        boolean  allowDir)
   {
      // make sure there is something to search for (the code below needs to be protected)
      if (filename == null)
      {
         return null;
      }
      
      // if the the OS is windows replace all / to \ from the search paths
      String fn = PlatformHelper.isUnderWindowsFamily() ? 
            filename.replaceAll("/", WIN_FILE_SEPARATOR) : filename;
      
      String result = getAbsoluteOrCurrent(caseSens, fn, allowDir); 
      if (result != null)
      {
         List<String> ret = new ArrayList<>(1);
         ret.add(result);
         return ret;
      }
      
      // not found in root or current, try PROPATH entries
      if (paths == null)
      {
         return null;
      }
      
      ArrayList<String> ret = new ArrayList<>(paths.length);
      for (String path : paths)
      {
         String fileName = getInPath(caseSens, path, fn, allowDir);
         if (fileName != null)
         {
            ret.add(fileName);
         }
      }
      ret.trimToSize();
      return ret.isEmpty() ? null : ret;
   }
   
   /**
    * Search file in paths with specified extensions 
    * 
    * @param filename
    *        file name
    * @param exts
    *        extensions to be checked
    * 
    * @return filename or <code>null</code> if not found
    */
   public String searchPathWithExts(String filename, List<String> exts)
   {
      // make sure there is something to search for (the code below needs to be protected)
      if (filename == null)
      {
         return null;
      }
      
      // if the the OS is windows replace all / to \ from the search paths
      String fn = PlatformHelper.isUnderWindowsFamily() ? 
            filename.replaceAll("/", WIN_FILE_SEPARATOR) : filename;
      
      String result = Utils.tryWhileNull(exts, 
                              ext -> getAbsoluteOrCurrent(caseSens, fn + ext, false));
      if (result != null)
      {
         return result;
      }
      
      // not found in root or current, try PROPATH entries
      if (paths == null)
      {
         return null;
      }
      
      return Utils.tryWhileNull(Arrays.asList(paths), 
                  path -> Utils.tryWhileNull(exts, 
                        ext -> getInPath(caseSens, path, fn + ext, false))
               );
   }

   /**
    * Find the actual canonical filesystem name that matches the given file or
    * directory, if a case-insensitive match exists.
    *
    * @param    file
    *           The filesystem object to search for.
    * @param    isDir
    *           <code>true</code> if searching for a directory and
    *           <code>false</code> if searching for a file.
    *
    * @return   The filesystem name found or <code>null</code> if no such
    *           object exists.
    */
   public static String getCaseInsensitiveMatch(File file, boolean isDir)
   {
      String result = null;
      
      try
      {
         // if the file is absolute use canonical path, otherwise use regular
         String fileName = file.getCanonicalPath();
         
         // case-insensitive search
         List<File> list = Utils.getCaseInsensitiveFilenameMatches(fileName);
         
         // use the first match that is valid
         for (File next : list)
         {
            if (next.exists())
            {
               if ((!isDir && next.isFile()) || (isDir && next.isDirectory()))
               {
                  result = next.getCanonicalPath();
                  break;
               }
            }
         }
      }
      catch (IOException exc)
      {
         // TODO: add reporting/logging
         
         // file system error, ignore
      }
      
      return result;
   }
   
   /**
    * Deletes the specified file system entity, if necessary using recursion to delete all 
    * contents (if it is a directory) first.  If any error occurs, no exception will be thrown.
    *
    * @param    target
    *           The target file or directory to delete.
    * @param    recursive
    *           Recursively decend through all subdirectories, deleting the entire subtree.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems occur).
    */
   public static int delete(File target, boolean recursive)
   {
      return deleteWorker(target, recursive);
   }
   
   /**
    * Access file system details regarding a specific file.
    *
    * @param    name
    *           The relative or absolute filename for which to obtain 
    *           information.
    *
    * @return   The file details if the file was found and can be read or 
    *           default values otherwise.
    */
   public FileInfo accessFileInfo(String name)
   {
      FileInfo info = new FileInfo();
      
      info.pathname = search(paths, caseSens, name, true);
      
      // did we find a match?
      if (info.pathname != null)
      {
         // initialize the rest of our data
         File local = new File(info.pathname); 
         
         try
         {
            info.size         = local.length();
            info.fullname     = local.getCanonicalPath();
            
            info.modification = local.lastModified();
            info.creation     = FileChecker.getFileCreationTime(info.fullname);
            
            String primeType = FileChecker.getFileType(info.fullname);
            StringBuilder sb = new StringBuilder(primeType);

            if (local.canRead())
               sb.append("R");
            
            try
            {
               if (local.canWrite())
                  sb.append("W");
            }
            catch (SecurityException se)
            {
               // it is OK to not have write access, don't bomb out
            }
            
            info.type = sb.toString();
         }
         catch (Exception exc)
         {
            // we probably aren't allowed to read the file, but there may
            // have been some other failure, reset everything to unknown
            info = new FileInfo();
         }
      }
      
      // the name is always set to the unchanged input value, even if it is
      // invalid
      info.name = name;
      
      return info;
   }
   
   /**
    * Get the comma-separated list of filesystem roots (drive letters on
    * platforms that support the concept, otherwise returns the equivalent
    * of the empty string).
    *
    * @return   The comma-separated list of drive letters or the empty string
    *           if the platform does not support the concept.
    */
   public String getRootList()
   throws IOException
   {
      StringBuilder sb = new StringBuilder("");
      
      File[] drives = File.listRoots();
      
      for (int i = 0; i < drives.length; i++)
      {
         // if this is the 2nd or later iteration, add a comma separator
         if (i > 0)
         {
            sb.append(",");
         }
         
         String name = drives[i].getCanonicalPath();
         int    idx  = name.lastIndexOf(File.separator);
         
         // add the path
         sb.append( idx < 0 ? name : name.substring(0, idx) );
      }
      
      return sb.toString(); 
   }
   
   /**
    * Get the system property (user-defined properties can be specified
    * at the Java command line) associated with the passed key.
    *
    * @return   The system property found or <code>null</code> if the
    *           property does not exist.
    */
   public String getProperty(String key)
   {
      if (key == null || key.length() == 0)
      {
         return null;
      }
      
      // first check the OS for environment variable value
      String res = getEnvironmentValue(key);
      
      return res != null ? res : System.getProperty(key);
   }
   
   /**
    * Get the file separator for the system.
    *
    * @return   The file separator.
    */
   public String getFileSeparator()
   {
      return File.separator;
   }
      
   /**
    * Get the path separator for the system.
    *
    * @return   The path separator.
    */
   public String getPathSeparator()
   {
      return File.pathSeparator;
   }
      
   /**
    * Creates each of the directories requested, if it doesn't exist and all
    * path segments up to the last segment do exist.  If the directory name
    * already exists in the filesystem (even if it is a file), this method
    * just returns (this is the Progress behavior).  If any error occurs, no
    * exception will be thrown, but no subsequent directories in the list
    * will be created.
    *
    * @param    dirs
    *           The list of directories to create.    
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int mkdir(String[] dirs)
   {
      try
      {
         for (int i = 0; i < dirs.length; i++)
         {
            File target = new File(dirs[i]);
            
            if (target.exists())
            {
               // nothing to do (Progress DOES NOT error even when the target is a FILE!)
               continue;
            }
            
            File parent = target.getParentFile();
            
            // is a parent dir specified? (if it is relative, it is not)
            if (parent != null)
            {
               // the parent needs to exist
               if (!parent.exists())
               {
                  return ERR_NO_SUCH_FILE_OR_DIR;
               }
               
               // the parent needs to be a directory
               if (!parent.isDirectory())
               {
                  return ERR_NOT_DIR;               
               }
            }
            
            // make the directory
            boolean result = target.mkdir();
            
            // did a failure occur?
            if (!result)
            {
               // unknown error 
               return ERR_UNKNOWN;
            }
         }
      }
      
      catch (SecurityException se)
      {
         return ERR_PERM_DENIED;
      }
      
      return ERR_NO_ERROR;
   }
         
   /**
    * Renames (and moves if the path is different) the specified source file
    * to the target filename (and path).  If any error occurs, no
    * exception will be thrown.
    *
    * @param    source
    *           The name of the source file or directory.
    * @param    target
    *           The name of the target file or directory.  If this is a
    *           directory name for a directory that already exists and the
    *           <code>source</code> is a file, then the source file name will
    *           be used as the target file in the directory given. 
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int rename(String source, String target)
   {
      if (source == null || target == null)
      {
         return ERR_NO_ERROR;
      }
      
      File src = new File(source);
      File tar = new File(target);
      
      try
      {
         // can't rename a non-existing source file or directory
         if (!src.exists())
         {         
            return ERR_NO_SUCH_FILE_OR_DIR;
         }
         
         if (tar.exists())
         {
            if (tar.isDirectory())
            {
               // append the source file or directory name to the target path
               tar = repathSource(src, target);
            }
            else
            {
               return ERR_FILE_EXISTS;
            }
         }
         
         // try the rename operation
         boolean result = src.renameTo(tar);
         
         // did a failure occur?
         if (!result)
         {
            // unknown error 
            return ERR_UNKNOWN;
         }
      }
      
      catch (SecurityException se)
      {
         return ERR_PERM_DENIED;
      }      
      
      return ERR_NO_ERROR;
   }
   
   /**
    * Initializes the path for subsequent searches by {@link #searchPath}. 
    *
    * @param    paths
    *           The list of paths to search for files specified by relative
    *           filename.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    */
   public void setSearchPath(String[] paths, boolean caseSens)
   {
      this.caseSens = caseSens;
      this.paths = paths;
   }
   
   /**
    * Search for the given file in each directory listed in the path
    * previously specified via {@link #setSearchPath}.  If the file system
    * object exists and is a file, then the filename is returned. If the path
    * contains an empty string or a '.', then the current directory is
    * searched, otherwise only the listed directories are searched.  
    * <p>
    * If the filename is specified as an absolute name, this specific 
    * filename is checked and no path searching is done.
    *
    * @param    filename
    *           The absolute or relative filename for which to search.
    *
    * @return   If the file is found in the current directory, the relative
    *           filename is returned. If the file is absolute or found in any
    *           other path, the full absolute filename is returned.  If the
    *           file is not found or the referenced file system object is not
    *           a file, <code>null</code> is returned.
    */
   public String searchPath(String filename)
   {
      return search(paths, caseSens, filename, false);
   }
   
   /**
    * Copies (or appends) the specified source file to or over the target
    * file. The mode (copy/overwrite or append) is determined by the
    * <code>overwrite</code> parameter.  If any error occurs, no exception
    * will be thrown.
    *
    * @param    source
    *           The name of the source file.  Must NOT be a directory name.
    *           Must be an existing file.
    * @param    target
    *           The name of the target file.  If this is a directory name,
    *           the source file name will be used as the target file in
    *           the directory given. 
    * @param    overwrite
    *           If <code>true</code> this is a copy, otherwise it is an
    *           append.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int copy(String source, String target, boolean overwrite)
   {
      File src = new File(source);
      File tar = new File(target);
      
      try
      {      
         // make sure the source file exists
         if (!src.exists())
         {
            return ERR_NO_SUCH_FILE_OR_DIR;
         }
         
         // the source file can't be a directory
         if (src.isDirectory())
         {
            return ERR_IS_DIR;
         }
         
         // we need read access to the source
         if (!src.canRead())
         {
            return ERR_PERM_DENIED;
         }
         
         // preprocess the target file
         if (tar.exists())
         {
            if (tar.isDirectory())
            {
               // switch to overwrite mode if not already in that mode
               overwrite = true;
               
               // append the source file name to the target path
               tar = repathSource(src, target);
            }
            
            // re-check because directory processing may have reset the
            // target to a file that doesn't exist
            if (tar.exists())
            {         
               if (tar.isFile())
               {
                  // we need write access to the target (this can only be
                  // called if the file exists)
                  if (!tar.canWrite())
                  {
                     return ERR_PERM_DENIED;
                  }                  
                  
                  if (overwrite)
                  {
                     // delete the target file
                     tar.delete();
                  }
               }
               else if (tar.isDirectory())
               {
                  // something is wrong and it can't be corrected
                  return ERR_IS_DIR;    
               }
            }
         }
         else
         {
            // switch to overwrite mode if not already in that mode, since
            // there is no target file
            overwrite = true;
         }
               
         // best efforts attempt to copy
         boolean result = Utils.copyFile(source, target, 0, !overwrite);
         
         // did a failure occur?
         if (!result)
         {
            // on any failure, remove partial results
            if (tar.exists())
            {
               tar.delete();
            }
            
            // unknown error 
            return ERR_UNKNOWN;
         }
      }
      
      catch (SecurityException se)
      {
         return ERR_PERM_DENIED;
      }
      
      return ERR_NO_ERROR;
   }
   
   /**
    * Deletes the specified file system entity, if necessary using recursion
    * to delete all contents (if it is a directory) first.  If any error 
    * occurs, no exception will be thrown.
    *
    * @param    target
    *           The target file or directory to delete.
    * @param    recursive
    *           Recursively decend through all subdirectories, deleting    
    *           the entire subtree.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int delete(String target, boolean recursive)  
   {
      return deleteWorker(new File(target), recursive);
   }
   
   /**
    * Deletes each of the directories or files specified.  If a non-empty
    * directory name is specified, an error will occur unless the
    * <code>recursive</code> flag is <code>true</code>.  If any error occurs,
    * no exception will be thrown, the deletion process just continues. The
    * last error will be the one returned even if there were prior successful
    * deletes or other failures.
    *
    * @param    list
    *           The list of files and directories to delete.
    * @param    recursive
    *           Recursively decend through all subdirectories, deleting
    *           the entire subtree.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int delete(String[] list, boolean recursive)
   {
      // set our result code, on error we keep going and the last error is
      // returned, even if some of the deletes succeed (for this reason, if
      // we set the error code at the beginning, then only if everything
      // succeeds will the rc be ERR_NO_ERROR, otherwise the last rc will
      // be accessible --> this is just as Progress does it)
      int rc = ERR_NO_ERROR;
      
      for (int i = 0; i < list.length; i++)
      {
         rc = deleteWorker(new File(list[i]), recursive);
      }            
      
      return rc;
   }
   
   /**
    * Utility method that checks if a path is in a form that is relative to the current dir
    * 
    * @param  pathName
    *         The path which we want to check.
    *         
    * @return <code>true</code> if the path is in a form that is relative to current dir or
    *         <code>false</code> otherwise
    */
   public static boolean isPathRelativeToCurrentDir(String pathName)
   {
      String separator = File.separator;

      // check if the path is relative to the current dir
      return (pathName.startsWith("." + separator)   ||
              pathName.startsWith(".." + separator)  || 
              pathName.equals(".")                   ||
              pathName.equals("..")                  ||
              (!PlatformHelper.isUnderWindowsFamily() && !pathName.startsWith(separator)) ||
              (PlatformHelper.isUnderWindowsFamily() &&
                 (!pathName.startsWith(separator) && !pathName.startsWith(":" + separator, 1))));
   }
   
   /**
    * Gets the OS environment variable value for given name if devined.
    *
    * @return   The currently defines ins OS environment value or <code>null</code> if variable 
    *           is currentlu not defined for given process.
    */
   public static native String getEnvironmentValue(String vname);
   
   /**
    * Deletes the specified file system entity, if necessary using recursion
    * to delete all contents (if it is a directory) first.  If any error 
    * occurs, no exception will be thrown.
    *
    * @param    target
    *           The target file or directory to delete.
    * @param    recursive
    *           Recursively decend through all subdirectories, deleting    
    *           the entire subtree.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   private static int deleteWorker(File target, boolean recursive)   
   {
      try
      {
         if (!target.exists())
         {
            // we can't delete what isn't there
            return ERR_NO_SUCH_FILE_OR_DIR;
         }
         
         if (target.isDirectory())
         {
            // delete contents recursively if requested
            if (recursive)
            {
               File[] children = target.listFiles();
               
               if (children == null)
               {
                  return ERR_IO;
               }
               
               for (int i = 0; i < children.length; i++)
               {
                  deleteWorker(children[i], recursive);
               }
            }
            
            // we can't delete a non-empty directory (if recursive was true,
            // then some child failed to delete, if false then the caller
            // shouldn't have expected this to work)
            if (target.list().length > 0)
            {
               return ERR_DIR_NOT_EMPTY;
            }
         }
         
         // we are a leaf or we are an empty directory, try to delete target
         // (this must occur after all children are processed in the case of
         // a directory)
         boolean result = target.delete();
         
         if (!result)
         {
            // the deletion failed
            return ERR_UNKNOWN;
         }
      }
      
      catch (SecurityException se)
      {
         return ERR_PERM_DENIED;
      }
      
      return ERR_NO_ERROR;      
   }   
   
   /**
    * Takes the last part of the source filename (removes the path) and
    * appends this to the target directory name.
    *
    * @param    source
    *           The source file or directory.
    * @param    target
    *           The target directory (cannot be a file).
    *
    * @return   The object representing the resulting file name. 
    */
   private File repathSource(File source, String target)
   {
      StringBuilder sb = new StringBuilder(target);
      
      // append the path segment separator if needed
      if (!target.endsWith(File.separator))
      {
         sb.append(File.separator);
      }
      
      // append the source file/directory name (remove the path)
      sb.append(source.getName());
      
      // reset the target
      target = sb.toString();
      
      return new File(target);
   }

   /**
    * Return file name if it is absolute or exists in the current directory
    * 
    * @param caseSens
    *        <code>true</code> if file system is case-sensitive  
    * @param filename
    *        file to check
    * @param allowDir
    *        <code>true</code> if directories are accepted  
    *        
    * @return file name if it is absolute or exists in the current directory
    */
   private static String getAbsoluteOrCurrent(boolean caseSens,
                                              String filename,
                                              boolean allowDir)
   {
      File file = new File(filename);

      if (file.isAbsolute() || isPathRelativeToCurrentDir(filename))
      {
         return getFileName(caseSens, allowDir, file, filename);   
      }
      return null;
   }

   /**
    * Return file name if file is valid in path
    * 
    * @param caseSens
    *        <code>true</code> if file system is case-sensitive  
    * @param path
    *        path to be checked
    * @param allowDir
    *        <code>true</code> if directories are accepted  
    * @param filename
    *        file to check
    *        
    * @return file name if file is valid in path
    */
   private static String getInPath(boolean caseSens,
                                   String path,
                                   String filename,
                                   boolean allowDir) 
   {
      if (path == null)
         return null;

      File file = new File(filename);
      // check if path is empty and filename exists and it's file
      // this doesn't apply for directories.
      if (path.length() != 0 || !file.isFile())
      {
         // if we have to replace separator also do this for PROPATH paths
         file = PlatformHelper.isUnderWindowsFamily() ? 
                                 new File(path.replaceAll("/", WIN_FILE_SEPARATOR), filename) : 
                                 new File(path, filename);
      }
      
      return getFileName(caseSens, allowDir, file, file.getPath());
   }

   /**
    * Return file name if file is valid
    * 
    * @param caseSens
    *        <code>true</code> if file system is case-sensitive  
    * @param allowDir
    *        <code>true</code> if directories are accepted  
    * @param file
    *        file to be checked
    * @param filename
    *        name to return
    *        
    * @return file name if file is valid
    */
   private static String getFileName(boolean caseSens, 
                                     boolean allowDir, 
                                     File file,
                                     String filename)
   {
      return caseSens ?
            file.exists() && (file.isFile() || allowDir) ? filename : null :
               getCaseInsensitiveMatch(file, false);
   }
}