/*
** Module   : NullField.java
** Abstract : outputs null characters to a stream
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description------------------------------
** 001 GES 20060925   @29904 Created initial version which outputs null
**                           characters to a stream.
** 002 SIY 20071001   @35317 Made class suitable for network transfers and
**                           for code reuse. Ref: #35314
** 003 SIY 20071003   @35349 Header cleanup. Added implementation of 
**                           resolveData() method.
**                           Ref: #35347.
** 004 CA  20130308          Added protected setters for the explicit and num fields.
** 005 OM  20130527          Added int64 support. Added 16 bit restriction observed in Progress.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;

/**
 * Outputs null characters to a stream. 
 */
public class NullField
extends FieldEntry
{
   /** Contains an explicit column position. */
   private boolean explicit = false;
   
   /**
    * An explicit column position.
    * From experimental tests one can observe that Progress handles the number of Fields 
    * (CONTROL NULL, SPACE and SKIP) on a signed 16 bit memory space. Although Progress will 
    * accept any 64-bit value it will be automatically truncate it to short. 
    */
   private short num = 0;
   
   /**
    * Default constructor (outputs a single space).
    */
   public NullField()
   {
   }
   
   /**
    * Constructor for an explicit number of null characters.
    *
    * @param    num
    *           The number of null characters to output.
    */
   public NullField(long num)
   {
      this.explicit = true;
      this.num      = (short) num;
   }
   
   /**
    * Renders the contained field definition to the given stream.
    *
    * @param    out
    *           The stream to which the field should be rendered.
    */
   public void output(Stream out)
   throws ErrorConditionException
   {
      if (explicit)
      {
         out.putNull(num);
      }
      else
      {
         out.putNull();
      }
   }

   /**
    * Notify instance that it is about to be transferred so all resolvable
    * data should be resolved and prepared for sending via network.
    */
   public void resolveData()
   {
      // do nothing
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    *
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      explicit = in.readBoolean();
      num      = in.readShort();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    *
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeBoolean(explicit);
      out.writeShort(num);
   }

   /**
    * Get current value of explicit flag.
    * 
    * @return   The current value of flag.
    */
   boolean isExplicit()
   {
      return explicit;
   }

   /**
    * Get explicit count.
    * 
    * @return   The explicit count.
    */
   int getNum()
   {
      return num;
   }

   /**
    * Set explicit count. The value is automatically truncated to 16-bit value.
    * 
    * @param    num
    *           The explicit count.
    */
   protected void setNum(long num)
   {
      this.num = (short) num;
   }
   
   
   /**
    * Set value of explicit flag.
    * 
    * @param    explicit
    *           The value of flag.
    */
   protected void setExplicit(boolean explicit)
   {
      this.explicit = explicit;
   }
}