/*
** Module   : OutputExtentField.java
** Abstract : Data wrapper which associates with a DMO extent property
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130209 Created initial version.
** 002 HC  20140613 Improved extent parameter support.
** 003 OM  20150715 Moved to com.goldencode.p2j.util package.
** 004 ECF 20160226 Change to account for FieldReference API change.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.persist.*;

/**
 * Data wrapper which associates with a DMO extent property. An instance of this class is used to
 * wrap the extent field when passed to an output parameter. It manages the hidden assignment of
 * an output parameter to the database field with which it is associated, upon return from 
 * a procedure or function.
 */
public class OutputExtentField<T extends BaseDataType>
extends OutputExtentParameter<T>
{
   /** The field reference instance for. */
   private final FieldReference fieldRef;

   /**
    * Constructor.
    * 
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    */
   public OutputExtentField(DataModelObject dmo, String property)
   {
      this(dmo, property, false);
   }

   /**
    * Constructor.
    * 
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   initialize
    *          Should this variable be initialized with the DMO's current
    *          value for the target field?
    */
   public OutputExtentField(DataModelObject dmo, String property, boolean initialize)
   {
      // this needs to register with the TM before the FieldAssigner instance
      this(new FieldReference(dmo, property, null));
   }
   
   /**
    * Constructor used internally to help initialize the instance - {@code FieldReference} 
    * instance needs to be passed to the parent as well as stored in {@link #fieldRef}. 
    *  
    * @param   fieldRef
    *          A valid {@link FieldReference} instance.
    */
   @SuppressWarnings("unchecked")
   private OutputExtentField(FieldReference fieldRef)
   {
      super((T[]) fieldRef.getExtentValues());
      this.fieldRef = fieldRef;
   }

   /**
    * Get the current extent reference.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   public T[] getVariable()
   {
      return (T[]) fieldRef.getExtentValues();
   }

   /**
    * Change the caller's reference to the new reference.
    * 
    * @param    reference
    *           The new reference.
    */
   public void setVariable(T[] reference)
   {
      fieldRef.set(reference);
   }
}