/*
** Module   : ProcessOps.java
** Abstract : Progress 4GL compatible process launching interface
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20050629   @21569 First version implementing standard process
**                           execution support.
** 002 GES 20050706   @21666 Integration with the lastError (replacement
**                           for os-error).
** 003 GES 20050712   @21686 Modifications to make silent and wait flags
**                           independent, to make it possible to execute
**                           a shell (with no command) and to add a
**                           version of launch that allows the caller to
**                           provide the in/out/err streams to which the
**                           child process' stdio should be attached.
**                           This is needed for the language statements:
**                           input/input-output/output through/thru.
** 004 GES 20050825   @22246 Modifications to support use of 
**                           StreamWrapper instead of direct references
**                           to ProcessStream.
** 005 GES 20060117   @23957 Moved pause to LogicalTerminal.
** 006 GES 20060207   @24345 New signature for launch (handles new unnamed
**                           streams implementation.
** 007 GES 20060211   @24557 A constant was moved into an interface.
** 008 GES 20060212   @24564 Rewrite to move launching to the remote
**                           client.  This class is now just the Progress
**                           compatible external interface.
** 009 GES 20060310   @24961 Added some launch variants.
** 010 GES 20061020   @30579 Change lastkey if a pause occurred during
**                           interactive launches.
** 011 GES 20111020          Modified per-session initialization to detect
**                           when the client code is running inside the server
**                           process AND to initialize for local redirection
**                           instead of using a remote object.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.util;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.ContextLocal;

/**
 * Provides static utility methods for Progress 4GL compatible process
 * launching and control.  At this time, the following Progress 4GL
 * language statements are handled:
 * <p>
 * <pre>
 * Progress Language Statement    Java Equivalent
 * ---------------------------    ---------------------------
 * BTOS                           {@link #launch(String[], boolean, boolean)}
 * DOS                            {@link #launch(String[], boolean, boolean)}
 * OS-COMMAND                     {@link #launch(String[], boolean, boolean)}
 * OS2                            {@link #launch(String[], boolean, boolean)}
 * UNIX                           {@link #launch(String[], boolean, boolean)}
 * INPUT THROUGH                  {@link #launch(String[], RemoteStream, RemoteStream)}
 * OUTPUT THROUGH                 {@link #launch(String[], RemoteStream, RemoteStream)}
 * INPUT-OUTPUT THROUGH           {@link #launch(String[], RemoteStream, RemoteStream)}
 * </pre>
 * <p>
 *
 * @author    GES
 */
public class ProcessOps
{   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
      
   /**
    * Launches the default shell for this system synchronous operation and
    * displaying a prompt after the child process terminates. The command
    * is always executed as a child process of the current JVM process. 
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the current JVM process using
    * a <code>StreamConnector</code> for each pipe.
    */
   public static void launch()
   {
      launch(null, false, true);
   }
   
   /**
    * Launches the default shell for this system taking into account the
    * flags which specify synchronous or asynchronous operation and whether
    * a prompt is provided after the child process terminates. The command
    * is always executed as a child process of the current JVM process. 
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the current JVM process using
    * a <code>StreamConnector</code> for each pipe.
    *
    * @param    silent
    *           If <code>true</code> the default prompt is suppressed.
    *           Otherwise, <code>ThinClient.pause</code> is called when the
    *           child process exits.
    * @param    wait
    *           If <code>true</code>, synchronous operation is requested
    *           which waits for the termination of the child process on the
    *           current thread.  If <code>false</code>, a thread will be
    *           started to wait for the child process termination and to
    *           subsequently handle cleanup (asynchronous mode).
    */
   public static void launch(boolean silent, boolean wait)
   {
      launch(null, silent, wait);
   }
   
   /**
    * Launches the external executible specified by the <code>cmdlist</code>
    * taking into account the flags which specify synchronous or asynchronous
    * operation and whether a prompt is provided after the child process
    * terminates. The command is always executed as a child process of the 
    * shell which is a child process of the current JVM process. 
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the current JVM process using
    * a <code>StreamConnector</code> for each pipe.
    *
    * @param    cmdlist
    *           The command line to be executed, with the first element
    *           used as the program to run and all other array elements as
    *           the arguments.
    * @param    silent
    *           If <code>true</code> the default prompt is suppressed.
    *           Otherwise, <code>ThinClient.pause</code> is called when the
    *           child process exits.
    * @param    wait
    *           If <code>true</code>, synchronous operation is requested
    *           which waits for the termination of the child process on the
    *           current thread.  If <code>false</code>, a thread will be
    *           started to wait for the child process termination and to
    *           subsequently handle cleanup (asynchronous mode).
    */
   public static void launch(String[] cmdlist, boolean silent, boolean wait)
   {
      // a process launch always sets the last error to 0
      FileSystemOps.setLastError(FileSystem.ERR_NO_ERROR);
      
      int lastkey = -1;
      
      try
      {
         lastkey = work.obtain().rl.launch(cmdlist, silent, wait);
         
         if (lastkey != -1)
         {
            KeyReader.setLastKey(lastkey);
         }
      }
      
      catch (ConditionException ce)
      {
         // make sure that the lastkey is properly maintained
         int kc = ce.getKeyCode();
         
         // anything other than -1 is a valid key press AND even -1 is a
         // valid indicator of CTRL-C if this was a STOP 
         if (kc != -1 || ce instanceof StopConditionException)
         {
            KeyReader.setLastKey(kc);
         }

         // continue the unwinding of the stack
         throw ce;
      }
   }
   
   /**
    * Launches the external executible specified by the <code>cmdlist</code>
    * on an asynchronous basis. The command is always executed as a child
    * process of the shell which is a child process of the current JVM 
    * process. 
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the passed parameters if
    * those parameters are not <code>null</code>. Any <code>null</code>
    * pipes are closed immediately and those pipes connected with the
    * passed parameters must be closed by the caller.  Please see the
    * {@link ProcessStream} for more details.
    *
    * @param    cmdlist
    *           The command line to be executed, with the first element
    *           used as the program to run and all other array elements as
    *           the arguments.
    * @param    sout
    *           A <code>StreamWrapper</code> instance which contains the
    *           <code>ProcessStream</code> to be connected to the child  
    *           process' <code>STDOUT and STDERR</code> or <code>null</code>
    *           for no connection.  Used for reading.
    * @param    sin
    *           A <code>StreamWrapper</code> instance which contains the
    *           <code>ProcessStream</code> to be connected to the child  
    *           process' <code>STDIN</code> or <code>null</code> for no
    *           connection.  Used for writing.
    */
   public static void launch(String[]      cmdlist,
                             StreamWrapper sout,
                             StreamWrapper sin)
   {      
      launch(cmdlist,
             (sout == null ? null : (RemoteStream) sout.dereference()),
             (sin  == null ? null : (RemoteStream) sin.dereference()));
   }

      
   /**
    * Launches the external executible specified by the <code>cmdlist</code>
    * on an asynchronous basis. The command is always executed as a child
    * process of the shell which is a child process of the current JVM 
    * process. 
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the passed parameters if
    * those parameters are not <code>null</code>. Any <code>null</code>
    * pipes are closed immediately and those pipes connected with the
    * passed parameters must be closed by the caller.  Please see the
    * {@link ProcessStream} for more details.
    *
    * @param    cmdlist
    *           The command line to be executed, with the first element
    *           used as the program to run and all other array elements as
    *           the arguments.
    * @param    sout
    *           The <code>RemoteStream</code> to be connected to the child  
    *           process' <code>STDOUT and STDERR</code> or <code>null</code>
    *           for no connection.  Used for reading.
    * @param    sin
    *           The <code>RemoteStream</code> to be connected to the child  
    *           process' <code>STDIN</code> or <code>null</code> for no
    *           connection.  Used for writing.
    */
   public static void launch(String[]     cmdlist,
                             RemoteStream sout,
                             RemoteStream sin)
   {            
      // a process launch always sets the last error to 0
      FileSystemOps.setLastError(FileSystem.ERR_NO_ERROR);
      
      work.obtain().rl.launch(cmdlist,
                              (sout == null ? -1 : sout.getId()), 
                              (sin  == null ? -1 : sin.getId()));
   }      
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Remote instance of our process launching server. */ 
      private Launcher rl = null; 
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return (WorkArea) this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized Object initialValue()
      {
         WorkArea wa = new WorkArea();
         
         // obtain the our proxy, server-side code is left-side by convention
         wa.rl = (Launcher) RemoteObject.obtainInstance(Launcher.class, true);
         
         return wa;
      }   
   }
}