/*
** Module   : RetryUnwindException.java
** Abstract : used to unwind the stack to an arbitrary retry point
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050811 NEW   @22115 Created initial version which is used to 
**                               unwind the stack to an arbitrary retry point.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Used to unwind the stack to an arbitrary retry point.
 * <p>
 * This class is an unchecked exception because the generated code always
 * emits catch blocks regardless of whether the body of the <code>try</code>
 * block can generate these exceptions or not.  The Java compiler fails in 
 * this condition (we have too much exception processing). The other solution
 * to this is to exhaustively analyze all <code>try</code> bodies to
 * determine if any <code>catch</code> blocks must be removed.  This has no
 * runtime benefit and is such a huge/complex effort that it is not
 * warrented.  Instead we use unchecked exceptions.
 */
public class RetryUnwindException
extends RuntimeException
{
   /**
    * Constructor which accepts an error message.
    *
    * @param   message
    *          Error message.
    */
   public RetryUnwindException(String message)
   {
      super(message);
   }
    
   /**
    * Constructor which accepts an error message and a root cause. This
    * variant is used when the error condition is detected as the result of
    * catching another exception.
    *
    * @param   message
    *          Error message.
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public RetryUnwindException(String message, Throwable rootCause)
   {
      super(message, rootCause);
   }
    
   /**
    * Constructor which accepts and wrappers a root cause. This variant is
    * used when the error condition is detected as the result of catching
    * another exception, and the message or context of the root cause is
    * sufficient to describe the problem.
    *
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public RetryUnwindException(Throwable rootCause)
   {
      super(rootCause);
   }
    
}