/*
** Module   : SecurityOps.java
** Abstract : Progress 4GL compatible security interface
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20060331   @25287 First version implementing the USERID builtin function.
** 002 GES 20060331   @25294 Added ENCODE support (not Progress 4GL compatible).
** 003 NVS 20060418   @25585 Added changePassword() convenience method to be called from
**                           converted code.
** 004 GES 20060831   @29063 Make sure encode output only includes chars from the set A-Z and
**                           a-z using a simple table driven base 52 approach.
** 005 GES 20101005          Added stub (non-functional) methods for the USERID(dbname) variant.
** 006 CA  20130129          Added stubs for misc security-related builtins. Implemented
**                           base64-encode/decode.
** 007 EVL 20130225          Adding conversion support for SETUSERID built-in.
** 008 GES 20130523          Minor fixes for the base64-encode/decode.
** 009 VIG 20130123          Implemented base64Decode and base64Encode methods using the own old
**                           Base64.java class. Added new method checkFormat to show an error
**                           while decoding in specific cases like the Progress does.
** 010 OM  20130917          Added _User metadata handling (configured in directory).
** 011 GES 20140117          Added HEX-DECODE and HEX-ENCODE functions.
** 012 GES 20140130          Added missing radix in hex-decode implementation and properly
**                           calculated the byte array size.
** 013 GES 20141031          Added a compatible version of ENCODE, left old version that is used
**                           only when a configuration flag is set. 
** 014 EVL 20160217          Clean up comments from symbols invalid for Solaris to compile.
** 015 ECF 20160315          Stubbed out generateUUID.
** 016 OM  20160706          Disabled condition 12118.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.security.*;
import java.util.logging.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * Provides static utility methods for Progress 4GL compatible security
 * functions.  At this time, the following Progress 4GL functions are handled:
 * <p>
 * <pre>
 * Progress Function              Java Equivalent
 * ---------------------------    ---------------------------
 * USERID                         {@link #getUserId}
 * USERID(dbname)                 {@link #getUserIdFromDB}
 * ENCODE                         {@link #encode}
 * BASE64-ENCODE                  {@link #base64Encode}
 * BASE64-DECODE                  {@link #base64Decode}
 * ENCRYPT                        {@link #encrypt}
 * GENERATE-PBE-KEY               {@link #generatePBEKey}
 * GENERATE-PBE-SALT              {@link #generatePBESalt}
 * GENERATE-UUID                  {@link #generateUUID}
 * SHA1-DIGEST                    {@link #sha1Digest}
 * MDG-DIGEST                     {@link #md5Digest}
 * </pre>
 * <p>
 */
public class SecurityOps
{
   /* Constant for blank user. In 4GL the black user is equivalent of guest user. */
   public static final String BLANK_USER = "";
   
   /** _User table of the database is empty, no access privilege needed */
   public static final int FREE_ACCESS = 0;
   
   /** Database has userids stored in _user table, but do not enforce them. */
   public static final int BLANK_ALLOWED = 1;
   
   /** Only with valid user/password are allowed to access a database. */
   public static final int STRICT = 2;
   
   /** The default classname for _User security handler */
   private static final String DEFAULT_SECURITY_OPS_HANDLER =
      "com.goldencode.p2j.util.SecurityManagerSecurityOps";
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(SecurityOps.class);
   
   /** Table of all alphabetic characters used to limit the encode output. */
   private static final String ALPHAS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" +
                                        "abcdefghijklmnopqrstuvwxyz";
   
   /** Progress error msg, can be shown while base64-encode method error */
   private static final String ERR12118 = "Unable to allocate memory for " +
                                    "result from BASE64-ENCODE function";

   /** Progress error msg, can be shown while base64-decode method error */
   private static final String ERR12119 = "Error converting Base64 to RAW";
   
   /** Base64 PAD service symbol */
   private static final char PAD = '=';

   /** Initial CRC value. */
   private static final int INITIAL_CRC_VALUE = 0x11;
   
   /** Number of bytes in the encoded output. */
   private static final int OUTPUT_ARRAY_SIZE = 16;
   
   /** Number of iterations in the core encoding loop. */
   private static final int CORE_LOOP_PASSES = 5;
   
   /** Reversed polynomial used with the CRC-16 algorithm. */
   private static final int REVERSED_POLYNOMIAL = 0xA001;
   
   /** The current _User security handler for emulation of USERID and SETUSERID P4GL functions */
   private static CustomSecurityOps customSecurityOps = instantiateUserHandler();
   
   /** Flag to enable the old incompatible ENCODE. This is 4GL compatible by default. */
   private static boolean compatible = Utils.getDirectoryNodeBoolean(null,
                                                                     "encode-compatibility",
                                                                     true,
                                                                     false);

   /**
    * Instantiates and returns a SecurityOps handler for emulation of USERID and SETUSERID P4GL
    * functions as configured in the directory.
    *
    * At this moment, only two of implementation are available: SecurityManagerSecurityOps and
    * DatabaseTableSecurityOps. If the directory does not contain an entry for
    * 'userSecurityOpsClass' then the default SecurityManagerSecurityOps will be used.
    *
    * @return  an instance of SecurityOps handler as configured in directory
    */
   private static CustomSecurityOps instantiateUserHandler()
   {
      Directory dir = DirectoryManager.getInstance();
      String handlerName = dir.getString(Directory.ID_RELATIVE,
                                         "userSecurityOpsClass",
                                         DEFAULT_SECURITY_OPS_HANDLER);

      // use the class-loader to get the class into memory
      Class<?> cl = null;
      try
      {
         cl = Class.forName(handlerName);
      }
      catch (Exception exc)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "Can't load user security handler '" + handlerName + "'. Loading default.",
                    exc);
         }
         return getDefaultUserHandler();
      }

      // attempt to instantiate an object
      Object o = null;
      try
      {
         o = cl.newInstance();
      }
      catch (Exception exc)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "User security handler '" + handlerName
                          + "'cannot be instantiated. Loading default.",
                    exc);
         }
         return getDefaultUserHandler();
      }

      // make sure the object is of the desired type
      if (!(o instanceof CustomSecurityOps))
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "Class '" + handlerName
                          + "'cannot be used as security handler. Loading default.");
         }
         return getDefaultUserHandler();
      }

      return (CustomSecurityOps) o;
   }

   /**
    * Instantiate the default _User handler and returns it. This method must not be called
    * multiple times as only one _User handler should exist.
    *
    * @return  An instance of CustomSecurityOps
    */
   private static CustomSecurityOps getDefaultUserHandler()
   {
      return new SecurityManagerSecurityOps();
   }

   /**
    * Returns the userid associated with the current user.
    * <p>
    * Note that in Progress 4GL, this takes an optional string parameter with
    * the name of the logical database in which the userid is set.  There is
    * no corresponding database level authentication concept in P2J so this
    * parameter is not supported.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserId() 
   {
      return new character(customSecurityOps.getUserId());
   }
   
   /**
    * Returns the userid associated with the current user of the given logical database.
    * <p>
    * This will call the current configured implementation of CustomSecurityOps. If the database
    * _User metadata is enabled then the authenticated user is returned, otherwise the default
    * SecurityManager user is returned.
    *
    * @param    dbname
    *           Logical database name.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserIdFromDB(character dbname)
   {
      // call the method with unwrapped-argument signature:
      return getUserIdFromDB(dbname.toStringMessage());
   }
   
   /**
    * Returns the userid associated with the current user of the given logical database.
    * <p>
    * This will call the current configured implementation of CustomSecurityOps. If the database
    * _User metadata is enabled then the authenticated user is returned, otherwise the default
    * SecurityManager user is returned.
    *
    * @param    dbname
    *           Logical database name.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserIdFromDB(String dbname)
   {
      return new character(customSecurityOps.getUserIdFromDB(dbname));
   }

   /**
    * Returns the default userid used for authentication to databases. 
    * If configured with metadata security and the database does not have users defined in its
    * metadata, we will use the <code>getDefaultUserid()</code> user id.
    *
    * @param   ldbname
    *          The logical name of the database.
    *
    * @return  the default userid used for authentication to databases.
    */
   public static String getDefaultUserid(String ldbname)
   {
      if (customSecurityOps.getAuthLevel(ldbname) == SecurityOps.FREE_ACCESS)
      {
         return getDefaultUserid();
      }
      else 
      {
         return BLANK_USER;
      }
   }
   
   /**
    * Returns the default userid used for authentication to databases that do not have _User
    * meta-table populated. Usually this is the username from the underlying OS, or empty string
    * if it cannot be detected for various causes.
    *
    * @return  the default userid used for authentication to databases.
    */
   public static String getDefaultUserid()
   {
      try
      {
         // extract the user name from the OS properties
         String defUserid = System.getProperty("user.name");

         // make sure is empty string instead of null
         return defUserid == null ? BLANK_USER : defUserid;
      }
      catch (SecurityException e)
      {
         // the security manager does not allow extracting the information return empty string, too
         return "";
      }
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, character password, character dbname) 
   {
      return new logical(customSecurityOps.setUserId(userid,
                                                     password.toStringMessage(),
                                                     dbname.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, String password, character dbname) 
   {
      return new logical(customSecurityOps.setUserId(userid.toStringMessage(),
                                                     password,
                                                     dbname.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, String password, character dbname) 
   {
      return new logical(customSecurityOps.setUserId(userid, password, dbname.toStringMessage()));
   }

   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, character password, String dbname)
   {
      return new logical(customSecurityOps.setUserId(userid.toStringMessage(),
                                                     password.toStringMessage(),
                                                     dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, character password, String dbname) 
   {
      return new logical(customSecurityOps.setUserId(userid, password.toStringMessage(), dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, String password, String dbname) 
   {
      return new logical(customSecurityOps.setUserId(userid.toStringMessage(), password, dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, String password, String dbname) 
   {
      return new logical(customSecurityOps.setUserId(userid, password, dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, character password) 
   {
      return new logical(customSecurityOps.setUserId(userid, password.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, String password) 
   {
      return new logical(customSecurityOps.setUserId(userid.toStringMessage(), password));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, String password) 
   {
      return new logical(customSecurityOps.setUserId(userid, password));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, character password) 
   {
      return new logical(customSecurityOps.setUserId(userid.toStringMessage(),
                                                     password.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, character password, character dbname)
   {
      return new logical(customSecurityOps.setUserId(userid.toStringMessage(),
                                                     password.toStringMessage(),
                                                     dbname.toStringMessage()));
   }

   /**
    * Checks if the currently authenticated user has access to a database.
    * <ul>
    *   <li>The user always has access to any database if the <code>SecurityOps</code> is
    *       configured with the default <code>SecurityManagerSecurityOps</code>.
    *   <li>When using <code>MetadataSecurityOps</code>, if the database does not have any users
    *       defined in _user meta table the authentication is not enforced so any user can access
    *       it. Otherwise, the user must have entered a correct combination of userid / password
    *       to get access to respective database.
    * </ul>
    *
    * @param   ldbName
    *          The logical name of the database.
    *          
    * @return  <code>true</code> if access to requested database is permitted for currently
    *          authenticated user.
    */
   public static boolean hasAccessToDatabase(String ldbName)
   {
      return customSecurityOps.hasAccessToDatabase(ldbName);
   }

   /**
    * Check the access level for a database. The response depend on CustomSecurityOps implementation
    * on which this method will forward the message.
    * 
    * @param   ldbname
    *          The logical database name to query. 
    * 
    * @return  The access level to specified database. 
    */
   public static int getAuthLevel(String ldbname)
   {
      return customSecurityOps.getAuthLevel(ldbname);
   }
   
   /**
    * Performs a password change procedure for the currently logged on user.
    * <p>
    * Application has no control over the procedure itself. 
    *
    * @return   failure/success indication
    *           <code>true</code> means 
    */
   public static logical changePassword() 
   {
      logical result = new logical(false);

      SecurityManager sm = SecurityManager.getInstance();
      
      if (sm != null)
      {
         result = new logical(sm.changePassword());
      }
      
      return result;
   }

   /**
    * <code>ENCODE</code> implementation which can be run in compatible mode (the default
    * which uses {@link #_encode} as a worker) or in the incompatible mode using the J2SE
    * MD5 hashing algorithm. The choice of which version to run is made using a flag
    * that is initialized from the directory when the server starts.
    * <p>
    * <b>WARNING: when the incompatible version is used the implementation is NOT compatible
    * with the Progress 4GL algorithm in any way. Any programs relying on comparisons between
    * data encoded by the Progress 4GL ENCODE and the return value from this method will NOT
    * compare as expected.</b>
    * <p>
    * By default, the Progress 4GL compatible <code>ENCODE</code> algorithm is truly compatible
    * and it will operate on stored/migrated values from the original system as expected. The
    * algorithm takes a source byte array (of any size) and generates a 16-byte one-way hash
    * using a 4GL-proprietary approach combined with a variant of CRC-16.  The basic approach is
    * to spread out the source data in an intermediate accumulator array of 16 unsigned bytes
    * and to repeatedly CRC that data and copy the resulting 2-byte CRC into successive elements
    * of this intermediate accumulator, while building the next CRC results on the accumulated
    * CRC value and the recently modified intermediate accumulator.  This set of CRCs is
    * calculated 8 times (since that will generate 16 bytes).  This basic approach is executed 5
    * times in a row. Each byte of the resulting hashed intermediate accumulator array is then
    * translated into one of the 52 possible uppercase or lowercase English alphabetic
    * characters. That translated result is returned.
    * <p>
    * The detailed COMPATIBLE algorithm:
    * <p>
    * <ol>
    *   <li> Initialize storage that can hold at least 2 unsigned bytes to 0x11. This is the
    *        accumulator for the calculated CRC.
    *   <li> Allocate storage that can hold 16 unsigned bytes.  This will be used as an
    *        intermediate output accumulator which will combine the input data and the
    *        calculated CRC data in a manner that will result in each of the 16 bytes having a
    *        value between 0 and 255 inclusive.
    *   <li> Iterate the following steps 5 times:
    *      <ol>
    *         <li>Walk forward through the array of source data provided and XOR each byte into
    *             the intermediate output accumulator (target array), but with the index of the
    *             target matching a backwards walk. Since the source array can be of any length,
    *             the lower index positions in the target may not ever get source data XOR'd.
    *             Source arrays longer than the target array cause wrapping and that means that
    *             the higher index positions in the target array will have data from multiple
    *             source positions XOR'd. See the {@link #sourceToTargetXor} method for details
    *             as well as the significant cryptographic limitations this causes.
    *         <li>Iterate 8 times, doing the following:
    *            <ol>
    *               <li> Call the {@link #crc16} method, passing in both the intermediate output
    *                    accumulator as well as the CRC accumulator.  This is the core CRC
    *                    algorithm and it will calculate the CRC of the current state of the
    *                    intermediate output accumulator while including the initial state of
    *                    the CRC accumulator.  The resulting returned CRC result is assigned
    *                    back to the CRC accumulator.  The CRC-16 algorithm used is not
    *                    cryptographically sound.  It generates a non-uniform distribution of
    *                    hashed data and collisions are not rare.
    *               <li> The least significant byte (byte 0) of the CRC accumulator will be
    *                    directly stored into an index position in the intermediate output
    *                    accumulator. This index position starts at 0 and increments by 2 for
    *                    every iteration of the nearest containing loop.  Since the intermediate
    *                    output accumulator has 16 elements, that is why the containing loop
    *                    iterates only 8 times.  This byte must be treated as an unsigned value.
    *               <li> The next to least significant byte (byte 1) of the CRC accumulator will
    *                    be directly stored into an index position in the intermediate output
    *                    accumulator. This index position starts at 1 and increments by 2 for
    *                    every iteration of the nearest containing loop. This byte must be
    *                    treated as an unsigned value.
    *            </ol>
    *      </ol>
    *   <li> At this point the intermediate output accumulator contains 16 poorly distributed
    *        hashed bytes.  Translate each byte of the intermediate output accumulator into
    *        a byte from the valid output character set.  Valid output characters include all
    *        English alphabetic letters (both uppercase and lowercase), for 52 possible output
    *        characters.  See the {@link #translateToAlpha} method for details on this process
    *        and for the severe limitations of the resulting non-uniform distribution.
    * </ol>
    * <p>
    * The result of this algorithm is NOT cryptographically sound.  It should not be used for
    * secure purposes.  The flaws in this algorithm come from the processing associated with
    * {@link #sourceToTargetXor}, {@link #crc16} and {@link #translateToAlpha}.
    *
    * @param    data
    *           The data to be encoded.  This must not be <code>null</code>.  The string may be
    *           of any length.  The data will be truncated at any contained null characters.
    *
    * @return   A string with 16 encoded bytes where each byte is one of the 52 possible output
    *           characters.  Returns <code>unknown value</code> if the input is
    *           <code>null</code> or <code>unknown value</code>.
    */
   public static character encode(character data)
   {
      if (data == null || data.isUnknown())
      {
         return new character();
      }
      
      return encode(data.toStringMessage());
   }
   
   /**
    * <code>ENCODE</code> implementation which can be run in compatible mode (the default
    * which uses {@link #_encode} as a worker) or in the incompatible mode using the J2SE
    * MD5 hashing algorithm. The choice of which version to run is made using a flag
    * that is initialized from the directory when the server starts.
    * <p>
    * <b>WARNING: when the incompatible version is used the implementation is NOT compatible
    * with the Progress 4GL algorithm in any way. Any programs relying on comparisons between
    * data encoded by the Progress 4GL ENCODE and the return value from this method will NOT
    * compare as expected.</b>
    * <p>
    * By default, the Progress 4GL compatible <code>ENCODE</code> algorithm is truly compatible
    * and it will operate on stored/migrated values from the original system as expected. The
    * algorithm takes a source byte array (of any size) and generates a 16-byte one-way hash
    * using a 4GL-proprietary approach combined with a variant of CRC-16.  The basic approach is
    * to spread out the source data in an intermediate accumulator array of 16 unsigned bytes
    * and to repeatedly CRC that data and copy the resulting 2-byte CRC into successive elements
    * of this intermediate accumulator, while building the next CRC results on the accumulated
    * CRC value and the recently modified intermediate accumulator.  This set of CRCs is
    * calculated 8 times (since that will generate 16 bytes).  This basic approach is executed 5
    * times in a row. Each byte of the resulting hashed intermediate accumulator array is then
    * translated into one of the 52 possible uppercase or lowercase English alphabetic
    * characters. That translated result is returned.
    * <p>
    * The detailed COMPATIBLE algorithm:
    * <p>
    * <ol>
    *   <li> Initialize storage that can hold at least 2 unsigned bytes to 0x11. This is the
    *        accumulator for the calculated CRC.
    *   <li> Allocate storage that can hold 16 unsigned bytes.  This will be used as an
    *        intermediate output accumulator which will combine the input data and the
    *        calculated CRC data in a manner that will result in each of the 16 bytes having a
    *        value between 0 and 255 inclusive.
    *   <li> Iterate the following steps 5 times:
    *      <ol>
    *         <li>Walk forward through the array of source data provided and XOR each byte into
    *             the intermediate output accumulator (target array), but with the index of the
    *             target matching a backwards walk. Since the source array can be of any length,
    *             the lower index positions in the target may not ever get source data XOR'd.
    *             Source arrays longer than the target array cause wrapping and that means that
    *             the higher index positions in the target array will have data from multiple
    *             source positions XOR'd. See the {@link #sourceToTargetXor} method for details
    *             as well as the significant cryptographic limitations this causes.
    *         <li>Iterate 8 times, doing the following:
    *            <ol>
    *               <li> Call the {@link #crc16} method, passing in both the intermediate output
    *                    accumulator as well as the CRC accumulator.  This is the core CRC
    *                    algorithm and it will calculate the CRC of the current state of the
    *                    intermediate output accumulator while including the initial state of
    *                    the CRC accumulator.  The resulting returned CRC result is assigned
    *                    back to the CRC accumulator.  The CRC-16 algorithm used is not
    *                    cryptographically sound.  It generates a non-uniform distribution of
    *                    hashed data and collisions are not rare.
    *               <li> The least significant byte (byte 0) of the CRC accumulator will be
    *                    directly stored into an index position in the intermediate output
    *                    accumulator. This index position starts at 0 and increments by 2 for
    *                    every iteration of the nearest containing loop.  Since the intermediate
    *                    output accumulator has 16 elements, that is why the containing loop
    *                    iterates only 8 times.  This byte must be treated as an unsigned value.
    *               <li> The next to least significant byte (byte 1) of the CRC accumulator will
    *                    be directly stored into an index position in the intermediate output
    *                    accumulator. This index position starts at 1 and increments by 2 for
    *                    every iteration of the nearest containing loop. This byte must be
    *                    treated as an unsigned value.
    *            </ol>
    *      </ol>
    *   <li> At this point the intermediate output accumulator contains 16 poorly distributed
    *        hashed bytes.  Translate each byte of the intermediate output accumulator into
    *        a byte from the valid output character set.  Valid output characters include all
    *        English alphabetic letters (both uppercase and lowercase), for 52 possible output
    *        characters.  See the {@link #translateToAlpha} method for details on this process
    *        and for the severe limitations of the resulting non-uniform distribution.
    * </ol>
    * <p>
    * The result of this algorithm is NOT cryptographically sound.  It should not be used for
    * secure purposes.  The flaws in this algorithm come from the processing associated with
    * {@link #sourceToTargetXor}, {@link #crc16} and {@link #translateToAlpha}.
    *
    * @param    data
    *           The data to be encoded.  This must not be <code>null</code>.  The string may be
    *           of any length.  The data will be truncated at any contained null characters.
    *
    * @return   A string with 16 encoded bytes where each byte is one of the 52 possible output
    *           characters.  Returns <code>unknown value</code> if the input is
    *           <code>null</code>.
    */
   public static character encode(String data)
   {
      if (data == null)
         return new character();
      
      return new character(compatible ? _encode(data) : _incompatible_encode(data));
   }

   /**
    * Progress 4GL compatible <code>ENCODE</code> algorithm which takes a source byte array
    * (of any size) and generates a 16-byte one-way hash using a proprietary approach
    * combined with a variant of CRC-16.  The basic approach is to spread out the source data
    * in an intermediate accumulator array of 16 unsigned bytes and to repeatedly CRC that data
    * and copy the resulting 2-byte CRC into successive elements of this intermediate
    * accumulator, while building the next CRC results on the accumulated CRC value and the
    * recently modified intermediate accumulator.  This set of CRCs is calculated 8 times
    * (since that will generate 16 bytes).  This basic approach is executed 5 times in a row.
    * Each byte of the resulting hashed intermediate accumulator array is then translated into
    * one of the 52 possible uppercase or lowercase English alphabetic characters. That
    * translated result is returned.
    * <p>
    * The detailed algorithm:
    * <p>
    * <ol>
    *   <li> Initialize storage that can hold at least 2 unsigned bytes to 0x11. This is the
    *        accumulator for the calculated CRC.
    *   <li> Allocate storage that can hold 16 unsigned bytes.  This will be used as an
    *        intermediate output accumulator which will combine the input data and the
    *        calculated CRC data in a manner that will result in each of the 16 bytes having a
    *        value between 0 and 255 inclusive.
    *   <li> Iterate the following steps 5 times:
    *      <ol>
    *         <li>Walk forward through the array of source data provided and XOR each byte into
    *             the intermediate output accumulator (target array), but with the index of the
    *             target matching a backwards walk. Since the source array can be of any length,
    *             the lower index positions in the target may not ever get source data XOR'd.
    *             Source arrays longer than the target array cause wrapping and that means that
    *             the higher index positions in the target array will have data from multiple
    *             source positions XOR'd. See the {@link #sourceToTargetXor} method for details
    *             as well as the significant cryptographic limitations this causes.
    *         <li>Iterate 8 times, doing the following:
    *            <ol>
    *               <li> Call the {@link #crc16} method, passing in both the intermediate output
    *                    accumulator as well as the CRC accumulator.  This is the core CRC
    *                    algorithm and it will calculate the CRC of the current state of the
    *                    intermediate output accumulator while including the initial state of
    *                    the CRC accumulator.  The resulting returned CRC result is assigned
    *                    back to the CRC accumulator.  The CRC-16 algorithm used is not
    *                    cryptographically sound.  It generates a non-uniform distribution of
    *                    hashed data and collisions are not rare.
    *               <li> The least significant byte (byte 0) of the CRC accumulator will be
    *                    directly stored into an index position in the intermediate output
    *                    accumulator. This index position starts at 0 and increments by 2 for
    *                    every iteration of the nearest containing loop.  Since the intermediate
    *                    output accumulator has 16 elements, that is why the containing loop
    *                    iterates only 8 times.  This byte must be treated as an unsigned value.
    *               <li> The next to least significant byte (byte 1) of the CRC accumulator will
    *                    be directly stored into an index position in the intermediate output
    *                    accumulator. This index position starts at 1 and increments by 2 for
    *                    every iteration of the nearest containing loop. This byte must be
    *                    treated as an unsigned value.
    *            </ol>
    *      </ol>
    *   <li> At this point the intermediate output accumulator contains 16 poorly distributed
    *        hashed bytes.  Translate each byte of the intermediate output accumulator into
    *        a byte from the valid output character set.  Valid output characters include all
    *        English alphabetic letters (both uppercase and lowercase), for 52 possible output
    *        characters.  See the {@link #translateToAlpha} method for details on this process
    *        and for the severe limitations of the resulting non-uniform distribution.
    * </ol>
    * <p>
    * The result of this algorithm is NOT cryptographically sound.  It should not be used for
    * secure purposes.  The flaws in this algorithm come from the processing associated with
    * {@link #sourceToTargetXor}, {@link #crc16} and {@link #translateToAlpha}.
    *
    * @param    input
    *           The data to be encoded.  This must not be <code>null</code>.  The string may be
    *           of any length.  The data will be truncated at any contained null characters.
    *
    * @return   A string with 16 encoded bytes where each byte is one of the 52 possible output
    *           characters.
    */
   public static String _encode(String input)
   {
      // the input data is truncated at the first encountered null character
      input = Text.javaTruncateNull(input);
      
      byte[] bin  = input.getBytes();
      int[]  data = new int[bin.length];
      
      // copy the input data into a array large enough to handle unsigned characters
      for (int n = 0; n < bin.length; n++)
      {
         data[n] = bin[n];
      }
      
      short[] storage = new short[OUTPUT_ARRAY_SIZE];
      
      int crc    = INITIAL_CRC_VALUE;
      int passes = CORE_LOOP_PASSES;
      
      while (passes != 0)
      {
         sourceToTargetXor(data, storage);
         
         int idx = 0;
         
         while (idx < storage.length)
         {
            crc = crc16(storage, crc);
            
            storage[idx++] = (byte)(crc & 0xFF);
            storage[idx++] = (byte)((crc >> 8) & 0xFF);
         }
         
         passes--;
      }
      
      return new String(translateToAlpha(storage));
   }
   
   /**
    * Converts binary data into a Base64 encoded text.
    *
    * @param    data
    *           The data to be encoded.
    *
    * @return   The encoded string.  <code>unknown</code> will be returned if the given data is
    *           <code>unknown</code> OR if the given data is initialized but the runtime has no
    *           knowledge of the buffer's size.
    */
   public static longchar base64Encode(BinaryData data)
   {
      if (data == null || data.isUnknown())
      {
         return new longchar();
      }

      if (data instanceof memptr)
      {
         return base64Encode((memptr)data);
      }

      return base64Encode(data.getByteArray());
   }
   
   /**
    * Converts binary data into a Base64 encoded text.
    *
    * @param    data
    *           The data to be encoded.
    *
    * @return   The encoded string.  <code>unknown</code> will be returned if the given data is
    *           <code>null</code>.
    */
   public static longchar base64Encode(byte[] data)
   {
      if (data == null)
      {
         return new longchar();
      }

      return new longchar(Base64.byteArrayToBase64(data));
   }

   /**
    * Decodes a base64 encoded string into a binary value.
    *
    * @param    data
    *           The string to be decoded.
    *
    * @return   a memptr with the decoded data.
    */
   public static memptr base64Decode(Text data)
   {
      if (data == null || data.isUnknown()) 
      {
         return new memptr();
      }

      return base64Decode(data.getValue());
   }

   /**
    * Converts a base64 encoded string into a binary value and returns a 
    * memptr with the decoded value. Shows an error 12119 if the source
    * data do not conforms to special format (that Progress checks).
    *
    * @param    data
    *           The string to be decoded.
    *
    * @return   a memptr with the decoded data.
    */
   public static memptr base64Decode(String data)
   {
      if (data == null)
      {
         return new memptr();
      }

      if (!checkFormat(data))
      {
         ErrorManager.recordOrShowError(12119, ERR12119, false, false, false);
         return new memptr();
      }

      // Apache commons Base64.decodeBase64() is incompatible
      return new memptr(Base64.base64ToByteArray(data));
   }
   
   /**
    * Converts a hexadecimal encoded string into a binary value and returns a 
    * raw with the decoded value.  The data must contain an even number of valid
    * hexadecimal digits.
    *
    * @param    data
    *           The text to be decoded.  <code>null</code> will be treated like
    *           unknown value.
    *
    * @return   The decoded binary data if the input is an even number of valid
    *           hexadecimal digits, otherwise it will return unknown value. The
    *           exception is for an empty string, which will return a 0-sized
    *           buffer.
    */
   public static raw hexDecode(Text data)
   {
      if (data == null || data.isUnknown()) 
      {
         return (raw) raw.instantiateUnknownRaw();
      }

      return hexDecode(data.toStringMessage());
   }

   /**
    * Converts a hexadecimal encoded string into a binary value and returns a 
    * raw with the decoded value.  The data must contain an even number of valid
    * hexadecimal digits.
    *
    * @param    data
    *           The string to be decoded.  <code>null</code> will be treated like
    *           unknown value.
    *
    * @return   The decoded binary data if the input is an even number of valid
    *           hexadecimal digits, otherwise it will return unknown value. The
    *           exception is for an empty string, which will return a 0-sized
    *           buffer.
    */
   public static raw hexDecode(String data)
   {
      if (data == null || (data.length() % 2) != 0)
      {
         return (raw) raw.instantiateUnknownRaw();
      }
      
      // TODO: this was written from the Progress docs and has only been lightly
      //       tested; the boundary conditions need to be tested to confirm if this
      //       really is correct

      int    len     = data.length();
      byte[] decoded = new byte[len / 2];
      
      for (int i = 0; i < len; i += 2)
      {
         int next = Integer.parseInt(data.substring(i, i + 2), 16);
         
         decoded[i / 2] = (byte) (next & 0xFF);
      }
      
      return new raw(decoded);
   }
   
   /**
    * Converts a binary value into a hexadecimal encoded string and returns a 
    * the encoded value.  The data will result an even number of valid
    * hexadecimal digits.
    *
    * @param    data
    *           The text to be encoded.  <code>null</code> will be treated like
    *           unknown value.
    *
    * @return   The hexadecimal encoded text if the input is not unknown value, or the
    *           unknown value if that was the input. An empty binary buffer will return
    *           an empty string.
    */
   public static character hexEncode(raw data)
   {
      if (data == null || data.isUnknown()) 
      {
         return new character();
      }
      // TODO: this was written from the Progress docs and has not been tested at
      //       all; also, the boundary conditions need to be tested to confirm if this
      //       really is correct
      
      byte[] input = data.getByteArray();
      
      StringBuilder sb = new StringBuilder();
      
      for (int i = 0; i < input.length; i++)
      {
         sb.append(Integer.toHexString((int) (input[i] & 0xFF)));
      }

      return new character(sb.toString());
   }
   
   /**
    * Encrypts the source and returns a {@link memptr} instance containing 
    * the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    *           
    * @return   See above.
    */
   public static memptr encrypt(String data)
   {
      return encrypt(data, null);
   }
   
   /**
    * Encrypts the source using the given key and returns a {@link memptr}
    * instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(String data, raw key)
   {
      return encrypt(data, key, null);
   }
   
   /**
    * Encrypts the source using the given key and initialization vector, and 
    * returns a {@link memptr} instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(String data, raw key, raw ivValue)
   {
      return encrypt(data, key, ivValue, (String) null);
   }
   
   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(String data,
                                raw    key,
                                raw    ivValue,
                                String algorithm)
   {
      return encrypt(data, 
                     key, 
                     ivValue, 
                     new character(algorithm));
   }
   
   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(String    data,
                                raw       key,
                                raw       ivValue,
                                character algorithm)
   {
      return encrypt(new character(data), key, ivValue, algorithm);
   }
   
   /**
    * Encrypts the source and returns a {@link memptr} instance containing 
    * the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    *           
    * @return   See above.
    */
   public static memptr encrypt(Text data)
   {
      return encrypt(data, null);
   }
   
   /**
    * Encrypts the source using the given key and returns a {@link memptr}
    * instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(Text data, raw key)
   {
      return encrypt(data, key, null);
   }
   
   /**
    * Encrypts the source using the given key and initialization vector, and 
    * returns a {@link memptr} instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(Text data, raw key, raw ivValue)
   {
      return encrypt(data, key, ivValue, (String) null);
   }
   
   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(Text data,
                                raw    key,
                                raw    ivValue,
                                String algorithm)
   {
      return encrypt(data,
                     key,
                     ivValue,
                     new character(algorithm));
   }
   
   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(Text      data,
                                raw       key,
                                raw       ivValue,
                                character algorithm)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(String password)
   {
      return generatePBEKey(new character(password), null);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    * @param    salt
    *           a random series of 8 bytes to use as salt value in generating 
    *           the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(String password, raw salt)
   {
      return generatePBEKey(new character(password), salt);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(Text password)
   {
      return generatePBEKey(password, null);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    * @param    salt
    *           a random series of 8 bytes to use as salt value in generating 
    *           the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(Text password, raw salt)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(BinaryData password)
   {
      return generatePBEKey(password, null);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    * @param    salt
    *           a random series of 8 bytes to use as salt value in generating 
    *           the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(BinaryData password, raw salt)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Generates a random salt value (a series of 8 bytes) to use in generating
    * an encryption key, and returns the salt value as a {@link raw} instance.
    * 
    * @return   See above.
    */
   public static raw generatePBESalt()
   {
      // TODO: IMPLEMENT THIS
      return new raw();
   }
   
   /**
    * Generate a 16-bit UUID.
    * 
    * @return  UUID.
    */
   public static raw generateUUID()
   {
      // TODO: IMPLEMENT THIS
      return new raw();
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data)
   {
      return sha1Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data, String key)
   {
      return sha1Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data, Text key)
   {
      return sha1Digest(new character(data), key);
   }

   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data, BinaryData key)
   {
      return sha1Digest(new character(data), key);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data)
   {
      return sha1Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data, String key)
   {
      return sha1Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data, Text key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data)
   {
      return sha1Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data, String key)
   {
      return sha1Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data, Text key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data)
   {
      return md5Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data, String key)
   {
      return md5Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data, Text key)
   {
      return md5Digest(new character(data), key);
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data, BinaryData key)
   {
      return md5Digest(new character(data), key);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data)
   {
      return md5Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data, String key)
   {
      return md5Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data, Text key)
   {
      // TODO:
      return null;
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data)
   {
      return md5Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data, String key)
   {
      return md5Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data, Text key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Converts memptr into a Base64 encoded text. Also checks memptr bounds.
    * <p>
    * There were reports that the error 12118 was raised (<i>Unable to allocate memory for result
    * from  function</i>) but at this moment a recreate is unavailable. It is possible there is
    * a scenario in which memory is somehow constrained on the runtime system and this error
    * would be generated.  Since there is no known recreate at this time, the corresponding
    * error processing is not implemented.
    *
    * @param   data
    *          The data to be encoded.
    *
    * @return  The encoded string. {@code unknown} will be returned if the given data is
    *          {@code null} or {@code unknown}
    */
   private static longchar base64Encode(memptr data)
   {
      if (data == null || data.isUnknown() || data.isUninitialized())
      {
         return new longchar();
      }
      
      return base64Encode(data.getByteArray());
   }
   
   /**
    * Checks if the input string format conforms to Progress rules:
    * <ol>
    *   <li>All trailing PAD symbols are ignored except the last one.</li>
    *   <li>All characters must be from the base64 alphabet. 
    *   See {@link #isBase64(char)}.
    *   </li>
    *   <li>Lets name l - length of the source string. If l % 4 == 0 than we
    *   only check the rule 2. The last char can be the PAD.
    *   </li>
    *   <li>Lets name r = l % 4. If r != 0 than we get the last r chars to 
    *   check. The r must be only 3. The last char must be the PAD. The first
    *   and second chars of the remainder part must follow rule 2. At the end
    *   if remainder is good check the main part (of the length aliquot to 4)
    *   by the rule 2.
    *   </li>
    * </ol>
    * @param   data
    *          The string to be decoded by Base64
    *          
    * @return  true if data follows Progress rules, false - doesn't follow.
    */
   private static boolean checkFormat(String data)
   {
      char[] chars = data.toCharArray();

      // ignore trailing PAD symbols except only the last one
      int len = chars.length;
      for (; len - 1 > 0; len--)
      {
         if (chars[len - 1] != PAD || chars[len - 2] != PAD)
         {
            break;
         }
      }

      int remainder = len % 4;

      // Check remainder chars first. It's faster.
      if (remainder != 0)
      {
         // common chars num in remainder must be 3 and the last is '='
         if (remainder != 3 || chars[len - 1] != PAD)
         {
            return false;
         }

         // first two symbols must be from the base64 alphabet
         if (!isBase64(chars[len - 2]) || !isBase64(chars[len - 3]))
         {
            return false;
         }
      }

      // Check all chars up to last char with index aliquot to 4
      for (int i = 0; i < len - remainder; i++)
      {
         if (!isBase64(chars[i]))
         {
            boolean lastChar = (remainder == 0 && i == len - remainder - 1);
            if (!lastChar || chars[i] != PAD)
            {
               return false;
            }
         }
      }

      // passed all checks
      return true;
   }

   /**
    * Checks if the symbol from base64 alphabet:
    * "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
    * 
    * @param   c
    *          input character to check
    * 
    * @return  true if the c is in alphabet, false - otherwise
    */
   private static boolean isBase64(char c)
   {
      return (c <= 'Z' && c >= 'A') || 
             (c <= 'z' && c >= 'a') ||
             (c <= '9' && c >= '0') ||
             (c == '+' || c == '/');
   }
   
   /**
    * Hashes an arbitrary input string into a 16-byte output string.  This conversion is one-way
    * (the original data can not be recovered) and the output is always a string of 16-bytes no
    * matter the length of the input string.
    * <p>
    * <b>WARNING: this implementation is NOT compatible with the Progress 4GL algorithm in any
    * way. Any programs relying on comparisons between data encoded by the Progress 4GL ENCODE
    * and the return value from this method will NOT compare as expected.</b>
    * <p>
    * The 16-bytes of output will only include characters in the range <code>A-Z</code> and
    * <code>a-z</code>.
    * <p>
    * This is only here because there is at least 1 customer that has installations that have
    * stored values that were encoded with the incompatible version.  Remove this at such time
    * as it is no longer in use.
    *
    * @param    data
    *           The data to be hashed. Must not be <code>null</code>.
    *
    * @return   The hashed 16-byte string.
    */
   private static String _incompatible_encode(String data)
   {
      StringBuffer sb = new StringBuffer();
      byte[] hashed = null;
      
      try
      {
         MessageDigest md = MessageDigest.getInstance("MD5");
         hashed = md.digest(data.getBytes());
         
         for (int i = 0; i < hashed.length; i++)
         {
            // this is not a fully even distribution but it is close
            // enough for an interrim solution
            int base52 = ((0x000000FF & hashed[i]) % 52); 
            sb.append(ALPHAS.charAt(base52));
         }
      }
      catch (NoSuchAlgorithmException nsa)
      {
         // we will fail below
      }
      
      if (hashed == null || hashed.length != 16)
      {
         throw new RuntimeException("Unexpected hashing result for input '" +
                                    data + "' resulting in output '" +
                                    hashed + "'.");
      }
      
      return sb.toString();
   }
   
   /**
    * Walk forward through the array of source data provided and XOR each byte into the target
    * array, but with the index of the target matching a backwards walk.  The source data can
    * be an array of any length and the target length is not expected to match.  The target
    * indexes are calculated modulo the length of the target array, which means that for a source
    * array longer than the target array, one or more target array elements will store data XOR'd
    * from more than one source source array element (i.e. the XOR processing will wrap around).
    * For a source array smaller than the target array, there will be some low index elements
    * of the target array which will not receive any XOR'd data.  Only in the case where the
    * two arrays are the same length will there be no wrapping and no unmerged elements in the
    * target array.  Only in the case where the source array is modulo the size of the target
    * array will all elements of the target array be modified.
    * <p>
    * From a cryptographic perspective, the algorithm is quite poor.  Source arrays smaller
    * than the target arrays will leave target array bytes unmodified.  Source arrays larger
    * than the target array size will have some (or all) elements modified multiple times.
    * Both cases are causes of concern.  Not modifying elements means that there is less
    * input provided for distribution of results.  Modifying elements more than once can
    * cause issues as well.  Consider the case where the same character appears in the source
    * array in the first byte (index 0) and in the byte modulo the size of the target.  Because
    * of wrapping, this character will be XOR'd twice into the same element of the target
    * array.  XOR is its own inverse.  If you XOR the same data into a byte an even number of
    * times, then the resulting byte will be unchanged. This will have the effect of reducing
    * the distribution of the resulting data for some inputs. This is not suitable for secure
    * purposes.
    *
    * @param    source
    *           Source array to XOR from. Must not be <code>null</code>. May be of any length.
    * @param    target
    *           Target array to XOR into. Must not be <code>null</code>. May be of any length.    
    */
   private static void sourceToTargetXor(int[] source, short[] target)
   {
      int len = target.length;
      int max = len - 1;
      
      for (int i = 0; i < source.length; i++)
      {
         target[max - (i % len)] ^= source[i] & 0xFF;
      }
   }
   
   /**
    * This implements a standard CRC-16 (also known as CRC-16-IBM or CRC-16-IBM) algorithm that
    * uses a reversed polynomial of 0xA001 and swapped byte ordering. Using a reversed polynomial
    * means it processes each byte's least significant bit first (it shifts the binary data to
    * the right). Using swapped byte ordering means that the input data (which is being treated
    * as an arbitrarily large binary number) must be processed from highest element to lowest
    * element, whereas normally one might consider the most significant byte to be in the highest
    * array index position, this algorithm assumes the opposite.
    * <p>
    * If this algorithm generated uniformly distributed hashes, then in a best case scenario the
    * probability of a collision between any 2 items is (1 / 2^16) or .00152588 %.  That seems
    * good until one considers that between any 300 items there is a 50% chance of collisions
    * and between any 430 items there is a 75% chance of a collision!  This is the well known
    * birthday problem (see http://en.wikipedia.org/wiki/Birthday_attack).
    * <p>
    * Please note that the CRC-16 algorithm is not suitable for cryptographic hashing.  It does
    * NOT generate uniformly distributed hashes.  This means that the collision rate is not even
    * as good as the best case scenario.  Even if it did have uniform distribution, the small
    * number of bits means that collisions are not very rare.  CRC is more suitable for error
    * detection.  It should NOT be used for secure purposes.
    *
    * @param    data
    *           The bytes of data to CRC.
    * @param    crc
    *           The initial CRC value into which each element of the data will be XOR'd.
    *
    * @return   The calculated CRC value after factoring in all data bytes and binary dividing
    *           the polynomial.
    */
   private static int crc16(short[] data, int crc)
   {
      // iterate from the top of the input array to the bottom
      for (int idx = (data.length - 1); idx >= 0; idx--)
      {
         // XOR the input data into the crc
         crc ^= (int) (data[idx] & 0xFF);
         
         int bit = 7;
         
         // process each bit
         while (bit >= 0)
         {
            // check if the least significant bit is set (must be done before shifting)
            boolean lsb = ((crc & 0x01) == 1);
            
            // shift the data right by one bit
            crc >>= 1;
            
            if (lsb)
            {
               // the least significant bit was set, XOR the polynomial into the crc 
               crc ^= REVERSED_POLYNOMIAL;
            }
            
            bit--;
         }
      }
      
      return crc;
   }
   
   /**
    * Convert the source array into an array of bytes with the same number of elements, but
    * where each byte may only contain uppercase or lowercase English alphabetic characters
    * (a-z and A-Z).
    * <p>
    * The source array will have each element translated into a corresponding element in the
    * output array.  Only the least significant byte of the source array element is considered
    * in the translation algorithm.  The order of the elements in the output array will be
    * the same order as the source array (e.g. the first source element translates to the
    * first output element and so forth).
    * <p>
    * The following algorithm is used to translate the 256 possible source byte values into
    * the 52 possible output byte values:
    * <p>
    * <ol>
    *   <li> If the least significant 7 bits of the source byte are one of the 52 possible
    *        English alphabetic characters, then that is the resulting byte.  This will
    *        yield a byte with 0x41 ('A') through 0x5A ('Z') or 0x61 ('a') - 0x7A ('z'),
    *        inclusive.
    *   <li> Otherwise, the most significant (upper) nibble of the source byte will be used
    *        to select a character from 'a' through 'q'.  This nibble can only have one of
    *        16 possible values (0 through 15).  This nibble value is used as a direct
    *        index into the 16 possible lowercase output letters.  Thus, a nibble value of
    *        0 (0x0) will yield 'a' (0x61), 1 (0x1) will yield 'b' (0x62) and so on, with
    *        the last possible value of 15 (0xF) yielding 'p' (0x70).
    * </ol>
    * <p>
    * Assuming the source bytes are uniformly distributed, this translation approach is
    * guaranteed to result in a highly UNEVEN distribution of output bytes.  This is
    * cryptographically BAD and should NOT be used for secure purposes. To understand
    * why this occurs:
    * <p>
    * <pre>
    * Source Byte Range    Output Byte                               Distribution Notes
    * -----------------    -----------    ----------------------------------------------------------------------------   
    * 0x00 - 0x40          'a' - 'e'      'a' through 'd' are 16X more likely than 'e'
    * 0x41 - 0x5A          'A' - 'Z'      all equally likely (1 to 1 mapping of input and output)
    * 0x5B - 0x60          'f' - 'g'      'f' 5X more likely than 'g'
    * 0x61 - 0x7A          'a' - 'z'      all equally likely (1 to 1 mapping of input and output)
    * 0x7B - 0xC0          'h' - 'm'      'i' through 'l' are 16X more likely than 'm', 'h' is 5X more likely than 'm'
    * 0xC1 - 0xDA          'A' - 'Z'      all equally likely (1 to 1 mapping of input and output)
    * 0xDB - 0xE0          'n' - 'o'      'n' 5X more likely than 'o'
    * 0xE1 - 0xFA          'a' - 'z'      all equally likely (1 to 1 mapping of input and output)
    * 0xFB - 0xFF          'p'            only 'p' is possible
    * </pre>
    * <p>
    * More graphically, here is the exact distribution that will occur for perfectly distributed input:
    * <p>
    * <pre>
    * Character   Frequency          Histogram
    * ---------   ---------   -----------------------
    * A           2           ++
    * B           2           ++
    * C           2           ++
    * D           2           ++
    * E           2           ++
    * F           2           ++
    * G           2           ++
    * H           2           ++
    * I           2           ++
    * J           2           ++
    * K           2           ++
    * L           2           ++
    * M           2           ++
    * N           2           ++
    * O           2           ++
    * P           2           ++
    * Q           2           ++
    * R           2           ++
    * S           2           ++
    * T           2           ++
    * U           2           ++
    * V           2           ++
    * W           2           ++
    * X           2           ++
    * Y           2           ++
    * Z           2           ++
    * a           18          ++++++++++++++++++
    * b           18          ++++++++++++++++++
    * c           18          ++++++++++++++++++
    * d           18          ++++++++++++++++++
    * e           3           +++
    * f           7           +++++++
    * g           3           +++
    * h           7           +++++++
    * i           18          ++++++++++++++++++
    * j           18          ++++++++++++++++++
    * k           18          ++++++++++++++++++
    * l           18          ++++++++++++++++++
    * m           3           +++
    * n           7           +++++++
    * o           3           +++
    * p           7           +++++++
    * q           2           ++
    * r           2           ++
    * s           2           ++
    * t           2           ++
    * u           2           ++
    * v           2           ++
    * w           2           ++
    * x           2           ++
    * y           2           ++
    * z           2           ++
    * </pre>
    * <p>
    * This non-uniform distribution is purely due to the fallback approach of using the
    * upper nibble of some bytes as a selector into a subset of the possible output values.
    * Since the only a subset of the output values are targeted, it makes those values
    * more likely to occur.  In addition, because there are some upper nibble values that
    * are less likely to be encountered (because they rarely trigger the fallback mechanism
    * in the first place), this causes an additional level of non-uniformity even within
    * the subset that is possible to be selected.
    *
    * @param    source
    *           The bytes to convert.  Must not be <code>null</code>.  Only the least
    *           significant byte of each element will be used.
    *
    * @return   The converted array of bytes, with one element for each corresponding
    *           element of the source array.
    */
   private static byte[] translateToAlpha(short[] source)
   {
      byte[] target = new byte[source.length];
      
      for (int idx = 0; idx < source.length; idx++)
      {
         target[idx] = (byte)(source[idx] & 0x7F);
         
         if ((target[idx] < 'A' || (target[idx] > 'Z' && target[idx] < 'a') || target[idx] > 'z'))
         {
            target[idx] = (byte)(((source[idx] & 0xFF) >> 4) + 'a');
         }
      }
      
      return target;
   }
}