/*
** Module   : SharedVariableManager.java
** Abstract : provides context-local scoped resource pools.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20050729   @21924 Provides context-local scoped variable 
**                           and stream pools.
** 002 GES 20050808   @22069 Moved ContextLocal inner class to a stand-
**                           alone class to promote reuse.
** 003 GES 20050818   @22163 Implemented the Scopeable interface to allow
**                           this class to be automatically notified of
**                           all scope start and end occurrences. This
**                           removes the need for client code to be
**                           instrumented to provide this service. The
**                           use of this class is now tied to the
**                           TransactionManager.
** 004 ECF 20050920   @22887 Implement ScopeableFactory interface. Allows
**                           generic access to singleton instance of this
**                           class from TransactionManager.
** 005 GES 20051004   @22930 Added support for shared buffers and error
**                           manager processing.
** 006 GES 20051101   @23163 Added support for shared frames.
** 007 GES 20060123   @24014 Converted from thread local to context local
**                           and moved initialization code to be based
**                           on the standard directory-based server
**                           initialization routines.
** 008 GES 20061010   @30262 Removed global option from frame support
**                           since this is not allowed in Progress.
** 009 ECF 20070328   @32652 Added shared temp-table support. Added
**                           context container for temp tables and
**                           add/lookup methods for temp tables.
** 010 SIY 20070711   @34430 Changed addFrame() to return existing frame
**                           if one is present in storage.
** 011 GES 20080822   @39520 Slight modification proposed by SVL to allow
**                           the currently registered global var of the
**                           same name to be returned as a result of the
**                           addVariable() method.
** 012 SIY 20080904   @39718 Inferred generics.
** 013 GES 20080918   @39857 Major rewrite to handle scopes differently.
**                           Now each resource supports some or all of
**                           the following: CURRENT, NEXT, GLOBAL. NEXT
**                           is the new feature that allows a resource
**                           add to wait until the next scope is opened
**                           before being added to the dictionary. Add'l
**                           signatures were added to allow a default
**                           scope that differs by resource type.
** 014 GES 20090422   @41914 Converted to standard string formatting.
** 015 CA  20131013          Added no-op scopeDeleted() method, required by the changes in
**                           Scopeable interface.
** 016 OM  20141203          Added double-initialization check for NEW SHARED GLOBAL EXTENT-s.
** 017 VIG 20141215          Added methods for shared menus support.
** 018 OM  20150903          Improved error handling in the case of missing shared temp-tables.
** 019 CA  20151027          When running an internal entry (procedure, function) of a persistent 
**                           procedure, all its shared data (variables, tables, etc) needs to be  
**                           pushed and popped, so that the internal entry (and any other external
**                           program ran from it) has access to it.
** 020 CA  20151222          Refactored registration of global shared resources: first, check 
**                           previous scopes if a resource with the same name exists; if so, use
**                           that; otherwise, create a new resource and register it.  
**                           Fixed error messages when the shared resource can no be found.
**                           When matching a shared resource with an existing one, check if it is
**                           the same (i.e. variables have the same data-type/extent/undo, buffers
**                           are for the same table, etc).
** 021 CA  20160420          Fixed H020 - defined global vars are checked by name in all scopes -
**                           if one is found (regardless if global or not), that is matched and
**                           returned, if matched.
** 022 CA  20160721          Shared vars are registered in the next scope, if they are not defined
**                           as NO-UNDO.
**     CA  20160728          Refactored the shared registries; the var-related registries are 
**                           exposed as copies for each external program, with the global scope
**                           shared among all programs. 
** 023 CA  20160812          Performance improvements for H022: use IdentityHashMap when using
**                           ext prog instances as keys, var ref count is kept via a mutable
**                           AtomicInteger instead of a referent set.
** 024 GES 20160914          Force case-insensitive mode for the scoped dictionaries.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.atomic.*;
import java.util.function.*;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;

/**
 * Provides context-local scoped variable, stream, buffer, frame and temp
 * table pools.
 * <p>
 * All references can be added to the top scope (closest enclosing scope) or
 * to the global scope.  Lookups return the topmost reference by the given
 * name or <code>null</code> if no reference by that name exists.  The backing
 * function is provided by the {@link ScopedSymbolDictionary} and this class
 * exposes the necessary helpers as static methods which de-reference the
 * pool that is localized to the context that is specific to the current
 * authenticated user.  Please see <code>SecurityManager</code> for more
 * details on this facility.
 * <p>
 * The resource dictionaries (pools) are separate since they require
 * different namespaces.  Where possible, the returned reference is of
 * a specific type, but at this time for all non-stream and non-temp table
 * resources, only <code>Object</code> references can be returned.
 * <p>
 * Buffers are persistent object representations of relational database
 * records.
 * <p>
 * Temp tables are actually buffers which point to temp tables, but this is
 * simply an implementation detail.  They represent shared temp table
 * definitions, as opposed to shared buffers.  They are managed separate from
 * regular buffers to ensure a separate namespace is available.
 * <p>
 * Frames are object representations of a screen buffer that is specific to
 * a given user-defined visual layout.
 * <p>
 * Implements the {@link Scopeable} interface to allow this class to be
 * automatically notified of all scope start and end occurrences. This removes
 * the need for client code to be instrumented to provide this service. The
 * use of this class is now tied to the {@link TransactionManager} class. In
 * particular, the singleton instance of this class is used in each context-
 * local list of {@link Scopeable} objects as a dispatcher. This bit
 * of nastiness is necessary since the Java Interface can't be implemented
 * on a static basis but instead requires an instance method implementation.
 * These instance methods then call the static counterparts which is where
 * the content-local versions of the resource pools are accessed.
 * <p>
 * The initialization processing is implemented in the {@link #initialize}
 * method and it is called during server startup based on an entry in the
 * directory until the path "/server/&lt;serverId&gt;/startup" or in
 * "/server/default/startup".
 */
public class SharedVariableManager
implements Scopeable,
           ScopeableFactory
{
   /** Context-local state related to shared variables. */
   private static ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /** Singleton instance of this class (across entire JVM, all contexts). */ 
   private static SharedVariableManager single = null;
   
   /** Flag to detect and avoid attempts to initialize more than once. */
   private static boolean registered = false;
   
   /**
    * Implements the singleton nature of this class, constructs this instance
    * if it hasn't already been constructed.
    *
    * @return   The single instance of this class.
    */
   public static synchronized SharedVariableManager getInstance()
   {
      if (single == null)
      {
         single = new SharedVariableManager();
      }
      
      return single;
   }
   
   /**
    * Initialize the shared variable framework. This registers with the
    * {@link TransactionManager TransactionManager} a factory object which
    * creates a {@link Scopeable} instance.
    *
    * @throws  IllegalStateException
    *          If invoked more than once.
    */
   public static synchronized void initialize()
   {
      if (!registered)
      {
         TransactionManager.registerScopeableFactory(getInstance());
         registered = true;
      }
      else
      {
         throw new IllegalStateException("Initialize called > 1 time!");
      }
   }
   
   /**
    * Adds a variable to the next scope that is opened.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    var
    *           The variable reference.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(String name, T var)
   {
      return addVariable(ScopeLevel.NEXT, name, false, var, null);
   }
   
   /**
    * Adds a variable to the next scope that is opened.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    var
    *           The variable reference.
    * @param    init
    *           If non null, then var is assumed to be an extent variable, and this represents
    *           the default value for the variable's elements.
    *           
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(String name, T var, Supplier<Object> init)
   {
      return addVariable(ScopeLevel.NEXT, name, false, var, init);
   }
   
   /**
    * Adds a variable to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    * <p>
    * Any attempt to add second global variable of the same name, will cause
    * the passed-in instance to be ignored (the currently registered instance
    * will not be replaced).  In addition, the currently registered instance
    * will be returned.
    *
    * @param    <T> 
    *           Variable type.
    * @param    scope
    *           Specifies the target scope.
    * @param    name
    *           The variable name.
    * @param    var
    *           The variable reference.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(ScopeLevel scope, String name, T var)
   {
      return addVariable(scope, name, false, var, null);
   }
   
   
   /**
    * Adds a variable to the next scope that is opened.
    *
    * @param    <T> 
    *           Variable type.
    * @param    noUndo
    *           Flag indicating if the variable was defined as NO-UNDO.
    * @param    name
    *           The variable name.
    * @param    var
    *           The variable reference.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(String name, boolean noUndo, T var)
   {
      return addVariable(ScopeLevel.NEXT, name, noUndo, var, null);
   }
   
   /**
    * Adds a variable to the next scope that is opened.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    noUndo
    *           Flag indicating if the variable was defined as NO-UNDO.
    * @param    var
    *           The variable reference.
    * @param    init
    *           If non null, then var is assumed to be an extent variable, and this represents
    *           the default value for the variable's elements.
    *           
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(String name, boolean noUndo, T var, Supplier<Object> init)
   {
      return addVariable(ScopeLevel.NEXT, name, noUndo, var, init);
   }
   
   /**
    * Adds a variable to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    * <p>
    * Any attempt to add second global variable of the same name, will cause
    * the passed-in instance to be ignored (the currently registered instance
    * will not be replaced).  In addition, the currently registered instance
    * will be returned.
    *
    * @param    <T> 
    *           Variable type.
    * @param    scope
    *           Specifies the target scope.
    * @param    name
    *           The variable name.
    * @param    noUndo
    *           Flag indicating if the variable was defined as NO-UNDO.
    * @param    var
    *           The variable reference.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(ScopeLevel scope, String name, boolean noUndo, T var)
   {
      return addVariable(scope, name, noUndo, var, null);
   }

   /**
    * Adds a variable to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    * <p>
    * Any attempt to add second global variable of the same name, will cause
    * the passed-in instance to be ignored (the currently registered instance
    * will not be replaced).  In addition, the currently registered instance
    * will be returned.
    *
    * @param    <T> 
    *           Variable type.
    * @param    scope
    *           Specifies the target scope.
    * @param    name
    *           The variable name.
    * @param    var
    *           The variable reference.
    * @param    init
    *           If non null, then var is assumed to be an extent variable, and this represents
    *           the default value for the variable's elements.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(ScopeLevel scope, String name, T var, Supplier<Object> init)
   {
      return addVariable(scope, name, false, var, init);
   }
   
   /**
    * Adds a variable to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    * <p>
    * Any attempt to add second global variable of the same name, will cause
    * the passed-in instance to be ignored (the currently registered instance
    * will not be replaced).  In addition, the currently registered instance
    * will be returned.
    *
    * @param    <T> 
    *           Variable type.
    * @param    scope
    *           Specifies the target scope.
    * @param    name
    *           The variable name.
    * @param    noUndo
    *           Flag indicating if the variable was defined as NO-UNDO.
    * @param    var
    *           The variable reference.
    * @param    init
    *           If non null, then var is assumed to be an extent variable, and this represents
    *           the default value for the variable's elements.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(ScopeLevel        scope, 
                                   String            name, 
                                   boolean           noUndo,
                                   T                 var, 
                                   Supplier<Object>  init)
   {
      WorkArea wa = local.get();
      
      VarRegistry vr = wa.getPendingVars(false);
      
      SharedRegistry sr = vr.vars;
      boolean extent = var.getClass().isArray();
      boolean global = (scope == ScopeLevel.GLOBAL);
      
      if (global)
      {
         // when searching for a match, the search is not done only in the global scope. 
         // a DEFINE NEW SHARED can hide a previous DEFINE NEW GLOBAL SHARED, and a subsequent
         // DEFINE NEW GLOBAL SHARED will resolve to the DEFINE NEW SHARED var, not the global one
         Object current = sr.shared.lookupSymbol(name);

         // ignore if there is a previous shared definition (global or not)
         if (current != null)
         {
            checkSharedVariable(name, var, noUndo);

            return (T) current;
         }
      }

      addWorker(sr, scope, name, var);
      
      if (noUndo)
      {
         // save this variable as NO-UNDO
         addWorker(vr.noUndoVars, scope, name, var);
      }
      
      if (extent)
      {
         BaseDataType[] bdvar = (BaseDataType[]) var;

         // assertion check - always pass an uninitialized array...
         for (int i = 0; i < bdvar.length; i++)
         {
            if (bdvar[i] != null)
            {
               throw new IllegalArgumentException(
                 "The array index " + i + " is already set for shared extsent var " + name + "!");
            }
         }
         
         // save the original var, when is first defined.  this will help later on, when the
         // variable is reused via DEFINE NEW GLOBAL SHARED, to check if the extent matches
         addWorker(vr.extentVars, scope, name, var);
         
         Object vinit = (init == null ? null : init.get());
         
         if (vinit == null)
         {
            BaseDataType.initializeDefaultExtent(bdvar);
         }
         else
         {
            if (vinit instanceof BaseDataType[])
            {
               BaseDataType[] bdinit = (BaseDataType[]) vinit;
               if (bdvar.length == 0)
               {
                  // resize it
                  bdvar = ArrayAssigner.resize(bdvar, bdinit.length);
                  var = (T) bdvar;
               }
               
               if (bdinit.length == bdvar.length)
               {
                  ArrayAssigner.assignMulti(bdvar, bdinit);
               }
               else if (bdinit.length < bdvar.length)
               {
                  for (int i = 0; i < bdinit.length; i++)
                  {
                     if (bdvar[i] == null)
                     {
                        bdvar[i] = bdinit[i].duplicate();
                     }
                     else
                     {
                        bdvar[i].assign(bdinit[i]);
                     }
                  }
                  
                  if (bdinit.length > 0)
                  {
                     ArrayAssigner.assignMulti(bdvar, bdinit.length, bdinit[bdinit.length - 1]);
                  }
               }
            }
            else if (init instanceof BaseDataType)
            {
               if (bdvar.length == 0)
               {
                  // resize it
                  bdvar = ArrayAssigner.resize(bdvar, 1);
                  var = (T) bdvar;
               }
               
               ArrayAssigner.assignMulti((BaseDataType[]) var, (BaseDataType) init);
            }
            else
            {
               throw new RuntimeException(
                  "Extent GLOBAL SHARED initializer must be a BaseDataType - received a " +
                  init.getClass()
               );
            }
         }
      }
      
      if (!noUndo)
      {
         if (extent)
         {
            TransactionManager.registerAt(scope, (LazyUndoable[]) var);
         }
         else
         {
            TransactionManager.registerAt(scope, (LazyUndoable) var);
         }
      }
      
      return var;
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first variable with the given 
    * name.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    expected
    *           The defined type of the slave variable (must match the master var type).
    *
    * @return   The variable reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the variable cannot be found or it does not match the master.
    */
   public static <T> T lookupVariable(String name, Class<? extends BaseDataType> expected)
   throws ErrorConditionException
   {
      return lookupVariable(name, expected, null, false);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first variable with the given 
    * name.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    expected
    *           The defined type of the slave variable (must match the master var type).
    * @param    noUndo
    *           Flag indicating if the variable was defined as NO-UNDO.
    *           
    * @return   The variable reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the variable cannot be found or it does not match the master.
    */
   public static <T> T lookupVariable(String                        name, 
                                      Class<? extends BaseDataType> expected, 
                                      boolean                       noUndo)
   throws ErrorConditionException
   {
      return lookupVariable(name, expected, null, noUndo);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first variable with the given 
    * name.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    expected
    *           The defined type of the slave variable (must match the master var type).
    * @param    extent
    *           The found var must be an extent var of the specified extent.
    *           
    * @return   The variable reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the variable cannot be found or it does not match the master.
    */
   public static <T> T lookupVariable(String                        name, 
                                      Class<? extends BaseDataType> expected, 
                                      Integer                       extent)
   throws ErrorConditionException
   {
      return lookupVariable(name, expected, extent, false);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first variable with the given 
    * name.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    expected
    *           The defined type of the slave variable (must match the master var type).
    * @param    extent
    *           The found var must be an extent var of the specified extent.
    * @param    noUndo
    *           Flag indicating if the variable was defined as NO-UNDO.
    *           
    * @return   The variable reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the variable cannot be found or it does not match the master.
    */
   public static <T> T lookupVariable(String                        name,
                                      Class<? extends BaseDataType> expected, 
                                      Integer                       extent, 
                                      boolean                       noUndo)
   throws ErrorConditionException
   {
      Runnable error = () ->
      {
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         String msg = "%s Shared variable %s has not yet been created";
         msg = String.format(msg, pname, name);

         errorHelper(392, msg);
      };

      WorkArea wa = local.get();

      Object found = lookupWorker(wa.getPendingVars(false).vars, name, error);

      if (found == null)
      {
         // fail early if it is not found
         return null;
      }

      Object var = (extent != null ? Array.newInstance(expected, Math.max(0,  extent))
                                   : BaseDataType.generateDefault(expected));

      checkSharedVariable(name, var, noUndo);
      
      return (T) found;
   }
   
   /**
    * Adds a stream to the {@link ScopeLevel#CURRENT} scope.
    *
    * @param    name
    *           The stream name.
    * @param    stream
    *           The stream reference.
    */
   public static void addStream(String name, Stream stream)
   {
      addStream(ScopeLevel.CURRENT, name, stream);
   }
   
   /**
    * Adds a stream to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The stream name.
    * @param    stream
    *           The stream reference.
    */
   public static void addStream(ScopeLevel scope, String name, Stream stream)
   {
      WorkArea wa = local.get();

      addWorker(wa.streams, scope, name, stream);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first
    * stream with the given name.
    *
    * @param    name
    *           The stream name.
    *           
    * @return   The stream reference associated with the name or
    *           <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the stream cannot be found and silent error mode is
    *           disabled.
    */
   public static Stream lookupStream(String name)
   throws ErrorConditionException
   {
      Runnable error = () ->
      {
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         String msg = "%s Shared stream %s not defined yet";
         msg = String.format(msg, pname, name);

         errorHelper(567, msg);
      };
      
      WorkArea wa = local.get();

      return (Stream) lookupWorker(wa.streams, name, error);
   }
   
   /**
    * Adds a buffer to the {@link ScopeLevel#CURRENT} scope.
    *
    * @param    name
    *           The buffer name.
    * @param    buffer
    *           The buffer reference.
    */
   public static <T> void addBuffer(String name, T buffer)
   {
      addBuffer(ScopeLevel.CURRENT, name, buffer);
   }
   
   /**
    * Adds a buffer to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    <T>
    *           Buffer type.
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The buffer name.
    * @param    buffer
    *           The buffer reference.
    */
   public static <T> void addBuffer(ScopeLevel scope, String name, T buffer)
   {
      WorkArea wa = local.get();
      
      addWorker(wa.buffers, scope, name, buffer);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first buffer with the given name.
    *
    * @param   <T> 
    *          Buffer type.
    * @param   name
    *          The buffer name.
    * @param   ttname
    *          The temp-table name.
    * @param   expected
    *          The expected DMO type of the resolved buffer.
    *
    * @return  The buffer reference associated with the name or
    *          <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the buffer cannot be found or it does not match the master.
    */
   public static <T> T lookupBuffer(String name, String   ttname, Class<?> expected)
   throws ErrorConditionException
   {
      Runnable errorLookup = () ->
      {
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         String msg = "%s Unable to find shared buffer for %s";
         msg = String.format(msg, pname, name);

         errorHelper(566, msg);
      };

      WorkArea wa = local.get();
      
      Object found =  (Buffer) lookupWorker(wa.buffers, name, errorLookup);
      
      if (found == null)
      {
         // fail early if it is not found
         return null;
      }
      
      BooleanSupplier pred = () -> (found != null && expected.isAssignableFrom(found.getClass()));
      
      Runnable error = () -> 
      {
         String msg = "Buffer %s for %s in %s does not match shared buffer defined NEW in %s";
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         String tpname = "n/a"; 
         
         Object creator = wa.bufferCreator.get(found);
         if (creator != null)
         {
            // find the external program where the temp-table was originally defined...
            tpname = ProcedureManager.getRelativeName(creator);
         }

         msg = String.format(msg, name, ttname, pname, tpname);
         
         errorHelper(1763, msg);
      };
      
      if (!checkSharedResource(pred, error))
      {
         throw new IllegalStateException(
            "An exception should have been thrown by checkedSharedResource!");
      }
      
      return (T) found;
   }
   
   /**
    * Creates (via {@link TemporaryBuffer#define} and adds a temporary table buffer to the next 
    * scope - the instance will be placed in the list of deferred instances that will be added 
    * the next time a scope is opened.
    *
    * @param    <T> 
    *           Type of constructed instance.
    * @param    name
    *           The temporary table buffer name.
    * @param    dmoBufIface
    *           Interface which defines DMO's public API.  This must be a sub-interface of 
    *           {@link Temporary}.
    * @param    variable
    *           Name of the variable with which the record buffer will be associated in the 
    *           transaction manager.
    * @param    legacyName
    *           Legacy name of the temp-table.
    * @param    undoable
    *           <code>true</code> if changes made to records managed by this buffer can be undone;  
    *           <code>false</code> if such changes are committed immediately, such that they cannot
    *           be undone.
    * 
    * @return   An object which implements the specified interface, but which initially has no 
    *           backing data record.
    */
   public static <T extends Temporary> T addTempTable(String   name, 
                                                      Class<T> dmoBufIface,
                                                      String   variable,
                                                      String   legacyName,
                                                      boolean  undoable)
   {
      return addTempTable(ScopeLevel.NEXT, name, dmoBufIface, variable, legacyName, undoable);
   }

   /**
    * Creates (via {@link TemporaryBuffer#define} and adds a temporary table buffer to the next 
    * scope - the instance will be placed in the list of deferred instances that will be added 
    * the next time a scope is opened.
    *
    * @param    <T> 
    *           Type of constructed instance.
    * @param    name
    *           The temporary table buffer name.
    * @param    dmoBufIface
    *           Interface which defines DMO's public API.  This must be a sub-interface of 
    *           {@link Temporary}.
    * @param    variable
    *           Name of the variable with which the record buffer will be associated in the 
    *           transaction manager.
    * @param    legacyName
    *           Legacy name of the temp-table.
    * 
    * @return   An object which implements the specified interface, but which initially has no 
    *           backing data record.
    */
   public static <T extends Temporary> T addTempTable(String   name, 
                                                      Class<T> dmoBufIface,
                                                      String   variable,
                                                      String   legacyName)
   {
      return addTempTable(ScopeLevel.NEXT, name, dmoBufIface, variable, legacyName, true);
   }
   
   /**
    * Creates (via {@link TemporaryBuffer#define} and adds a temporary table buffer to the 
    * specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list of deferred 
    * instances that will be added the next time a scope is opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most scope on the stack.  
    * Also, first, existing temp-tables with the same name are searched - if one is found in one 
    * of the previous scopes, then that temp-table is returned only if its DMO interface is the 
    * same as or it implements the specified <code>dmoBufIface</code>
    * <p>
    * If current is specified, the instance will be added to the scope that is on the top of the 
    * stack. If there is only one scope on the stack, the global and current scope will be the 
    * same scope.
    *
    * @param    <T> 
    *           Type of constructed instance.
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The temporary table buffer name.
    * @param    dmoBufIface
    *           Interface which defines DMO's public API.  This must be a sub-interface of 
    *           {@link Temporary}.
    * @param    variable
    *           Name of the variable with which the record buffer will be associated in the 
    *           transaction manager.
    * @param    legacyName
    *           Legacy name of the temp-table.
    * 
    * @return   An object which implements the specified interface, but which initially has no 
    *           backing data record.
    */
   public static <T extends Temporary> T addTempTable(ScopeLevel scope, 
                                                      String     name, 
                                                      Class<T>   dmoBufIface,
                                                      String     variable,
                                                      String     legacyName)
   {
      return addTempTable(scope, name, dmoBufIface, variable, legacyName, true);
   }
   
   /**
    * Creates (via {@link TemporaryBuffer#define} and adds a temporary table buffer to the 
    * specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list of deferred 
    * instances that will be added the next time a scope is opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most scope on the stack.  
    * Also, first, existing temp-tables with the same name are searched - if one is found in one 
    * of the previous scopes, then that temp-table is returned only if its DMO interface is the 
    * same as or it implements the specified <code>dmoBufIface</code>
    * <p>
    * If current is specified, the instance will be added to the scope that is on the top of the 
    * stack. If there is only one scope on the stack, the global and current scope will be the 
    * same scope.
    *
    * @param    <T> 
    *           Type of constructed instance.
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The temporary table buffer name.
    * @param    dmoBufIface
    *           Interface which defines DMO's public API.  This must be a sub-interface of 
    *           {@link Temporary}.
    * @param    variable
    *           Name of the variable with which the record buffer will be associated in the 
    *           transaction manager.
    * @param    legacyName
    *           Legacy name of the temp-table.
    * @param    undoable
    *           <code>true</code> if changes made to records managed by this buffer can be undone;  
    *           <code>false</code> if such changes are committed immediately, such that they cannot
    *           be undone.
    * 
    * @return   An object which implements the specified interface, but which initially has no 
    *           backing data record.
    */
   public static <T extends Temporary> T addTempTable(ScopeLevel scope, 
                                                      String     name, 
                                                      Class<T>   dmoBufIface,
                                                      String     variable,
                                                      String     legacyName,
                                                      boolean    undoable)
   {
      WorkArea wa = local.get();
      SharedRegistry sr = wa.tempTables;
      boolean global = (scope == ScopeLevel.GLOBAL);
      
      if (global)
      {
         Object found = sr.shared.lookupSymbol(name);
         
         // ignore repeated global definitions
         if (found != null)
         {
            checkSharedTempTable(found, name, dmoBufIface);

            // create a new buffer for this request (subsequent calls which re-use the temp-table
            // will create a separate slave buffer - do not reuse the master buffer)
            found = TemporaryBuffer.define((T) found, variable, legacyName, false);
            
            return (T) found;
         }
      }

      T var = TemporaryBuffer.define(dmoBufIface, variable, legacyName, global, undoable);
      
      addWorker(sr, scope, name, var);

      return var;
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first temp
    * table with the given name.
    *
    * @param   <T> 
    *          Variable type.
    * @param   name
    *          The temp table buffer name.
    * @param   expected
    *          The expected DMO type of the resolved temp-table.
    *
    * @return  The temp table buffer reference associated with the name or
    *          <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the temp-table cannot be found or it does not match the master.
    */
   public static <T extends Temporary> T lookupTempTable(String name, Class<T> expected)
   throws ErrorConditionException
   {
      Runnable error = () ->
      {
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         String msg = "Unable to locate shared temp-table or work-table definition " + 
                      "for table %s in procedure %s";
         msg = String.format(msg, name, pname);

         errorHelper(1429, msg);
      };

      WorkArea wa = local.get();

      Object found =  lookupWorker(wa.tempTables, name, error);

      if (found == null)
      {
         // fail early if it is not found
         return null;
      }

      checkSharedTempTable(found, name, expected);
      
      return (T) found;
   }
   
   /**
    * Adds a frame to the {@link ScopeLevel#NEXT} scope.
    * <p>
    * The instance will be placed in the list of deferred instances that will
    * be added the next time a scope is opened.
    *
    * @param    name
    *           The frame name.
    * @param    frame
    *           The frame reference.
    *
    * @return   Previous instance with the same name or <code>null</code> if
    *           there were no such instance.  If there is a previous instance,
    *           it will be replaced with this new instance.
    */
   public static <T> T addFrame(String name, T frame)
   {
      return addFrame(ScopeLevel.NEXT, name, frame);
   }
   
   /**
    * Adds a frame to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the next
    * scope on the stack, as there is no such thing as a global frame in
    * the Progress 4GL.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    <T>
    *           Frame type.
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The frame name.
    * @param    frame
    *           The frame reference.
    *
    * @return   Previous instance with the same name or <code>null</code> if
    *           there were no such instance.  If there is a previous instance,
    *           it will be replaced with this new instance.
    */
   public static <T> T addFrame(ScopeLevel scope, String name, T frame)
   {
      if (scope == ScopeLevel.GLOBAL)
      {
         scope = ScopeLevel.NEXT;
      }

      WorkArea wa = local.get();
      
      SharedRegistry sr = wa.frames;
      T result = (T) sr.shared.lookupSymbol(name);
      
      addWorker(sr, scope, name, frame);
      
      return result;
   }
   
   /**
    * Adds a menu definition to the {@link ScopeLevel#NEXT} scope.
    * <p>
    * The instance will be placed in the list of deferred instances that will
    * be added the next time a scope is opened.
    *
    * @param    name
    *           The menu name.
    * @param    menu
    *           The menu definition reference.
    */
   public static <T> void addMenu(String name, T menu)
   {
      addMenu(ScopeLevel.NEXT, name, menu);
   }
   
   /**
    * Adds a menu to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the next
    * scope on the stack, as GLOBAL is not a valid option for shared menus.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    <T>
    *           Menu definition type.
    * @param    scope 
    *           Specifies the target scope. 
    * @param    name
    *           The menu name.
    * @param    menu
    *           The menu reference.
    */
   public static <T> void addMenu(ScopeLevel scope, String name, T menu)
   {
      if (scope == ScopeLevel.GLOBAL)
      {
         scope = ScopeLevel.NEXT;
      }
      
      WorkArea wa = local.get();
      
      addWorker(wa.menus, scope, name, menu);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first frame with the given name.
    *
    * @param   <T> 
    *          Frame type.
    * @param   name
    *          The frame name.
    *
    * @return  The frame reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the frame cannot be found and silent error mode is disabled.
    */
   public static <T> T lookupFrame(String name)
   throws ErrorConditionException
   {
      Runnable error = () ->
      {
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         String msg = "%s Shared frame %s not created yet";
         msg = String.format(msg, pname, name);

         errorHelper(682, msg);
      };

      WorkArea wa = local.get();

      return (T) lookupWorker(wa.frames, name, error);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first menu with the given name.
    *
    * @param   <T> 
    *          menu definition type.
    * @param   name
    *          The menu name.
    *
    * @return  The menu definition reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the menu cannot be found and silent error mode is disabled.
    */
   public static <T> T lookupMenu(String name)
   throws ErrorConditionException
   {
      WorkArea wa = local.get();
      
      return (T) lookupWorker(wa.menus, name, null);
   }
   
   /**
    * Provides the singleton instance of this class as a {@link Scopeable}.
    *
    * @return   The only instance of this class.
    */
   public Scopeable createScopeable()
   {
      return this;
   }
   
   /**
    * Provides a notification that a new scope is about to be entered.
    */
   public void scopeStart()
   {
      // shared resources can be defined only at the external program
      if (!TransactionManager.isExternalBlock() || TransactionManager.isGlobalBlock())
      {
         return;
      }

      Object thisProc = ProcedureManager._thisProcedure();
      WorkArea wa = local.get();
      
      VarRegistry pendingVars = wa.getPendingVars(true);
      wa.pendingVars = null;
      
      // save the registry at THIS-PROCEDURE
      wa.varDictionaries.put(thisProc, pendingVars);

      {
         // process buffer resources first
         SharedRegistry sr = wa.buffers;
         if (sr.pending != null)
         {
            Set<String> keys = sr.pending.keySet();
            
            for (String next : keys)
            {
               Object resource = sr.pending.get(next);
               
               wa.bufferCreator.put(resource, thisProc);
            }
         }
      }
      
      Consumer<SharedRegistry> c = (sr) ->
      {
         sr.shared.addScope(null);
         
         if (sr.pending != null)
         {
            Set<String> keys = sr.pending.keySet();
            
            for (String next : keys)
            {
               Object resource = sr.pending.get(next);

               sr.shared.addSymbol(false, next, resource);
            }
            
            sr.pending = null;
         }
      };

      for (SharedRegistry sr : wa.resources)
      {
         c.accept(sr);
      }
      
      for (SharedRegistry sr : pendingVars.resources)
      {
         c.accept(sr);
      }
      
      SharedRegistry sr = pendingVars.vars;
      // exclude the global scope
      for (int scope = 0; scope <  sr.shared.size() - 1; scope++)
      {
         Map<String, Object> dict = sr.shared.getDictionaryAtScope(scope, false);
         if (dict == null)
         {
            continue;
         }
         
         for (Object var : dict.values())
         {
            AtomicInteger count = wa.varRefs.get(var);
            if (count == null)
            {
               wa.varRefs.put(var, count = new AtomicInteger(1));
            }
            else
            {
               count.incrementAndGet();
            }
         }
      }
   }   
   
   /**
    * Provides a notification that a scope is about to be exited.
    * <p>
    * This is a no-op for the shared variable manager.
    */
   public void scopeFinished()
   {
      // no-op
   }
   
   /**
    * Provides notification that the external procedure scope has been deleted.
    */
   @Override
   public void scopeDeleted()
   {
      if (!TransactionManager.isExternalBlock() || TransactionManager.isGlobalBlock())
      {
         return;
      }

      WorkArea wa = local.get();

      Object referent = ProcedureManager.getProcessedProcedure();
      
      // this needs to be executed only when the procedure gets deleted.
      if (referent != null)
      {
         VarRegistry vars = wa.varDictionaries.remove(referent);
         
         // for all vars exposed via the shared registry of the deleted procedure, decrement their
         // ref count when the ref count reaches zero, all these must be removed from the 
         // undoable list!
         IdentityHashMap<LazyUndoable, Object> removedVars = new IdentityHashMap<>();

         SharedRegistry sr = vars.vars;
         // exclude the global scope
         for (int scope = 0; scope <  sr.shared.size() - 1; scope++)
         {
            Map<String, Object> dict = sr.shared.getDictionaryAtScope(scope, false);
            if (dict == null)
            {
               continue;
            }
            
            for (Object var : dict.values())
            {
               AtomicInteger refs = wa.varRefs.get(var);
               int count = refs.decrementAndGet();
               
               if (count == 0)
               {
                  wa.varRefs.remove(var);
   
                  if (var instanceof BaseDataType)
                  {
                     removedVars.put((BaseDataType) var, null);
                  }
                  else
                  {
                     for (BaseDataType v : (BaseDataType[]) var)
                     {
                        removedVars.put(v, null);
                     }
                  }
               }
            }
         }
         
         if (!removedVars.isEmpty())
         {
            TransactionManager.deregister(null, removedVars);
         }
      }

      // remove the creator for all resources
      Map<String, Object> dict = wa.buffers.shared.getDictionaryAtScope(0, false);
      for (Object res : dict.values())
      {
         wa.bufferCreator.remove(res);
      }

      for (SharedRegistry sr : wa.resources)
      {
         sr.shared.deleteScope();
      }
   }
   
   /**
    * Update the references for the specified array, as a result of a resize.
    * 
    * @param    array
    *           The previous instance.
    * @param    copy
    *           The resized instance.
    */
   static <T extends BaseDataType> void updateExtentVar(T[] array, T[] copy)
   {
      WorkArea wa = local.get();
      
      for (VarRegistry v : wa.varDictionaries.values())
      {
         SharedRegistry sr = v.vars;
         int globalScope = sr.shared.size() - 1;

         // search all scopes and update the variable; do not update the extentVars dictionary - this
         // holds the original variable, to preserve its original extent, at the master definition
         for (int scope = 0; scope < sr.shared.size(); scope++)
         {
            Map<String, Object> dict = sr.shared.getDictionaryAtScope(scope, false);
            if (dict == null)
            {
               continue;
            }
            
            if (dict.values().contains(array))
            {
               // find its key (we don' know it here...)
               String foundKey = null;
               for (String key : dict.keySet())
               {
                  Object val = dict.get(key);
                  if (val == array)
                  {
                     foundKey = key;
                     break;
                  }
               }
               
               dict.put(foundKey, copy);
               
               if (scope == globalScope)
               {
                  for (LazyUndoable undoable : copy)
                  {
                     undoable.setGlobal(true);
                  }
               }
               
               // break on first found, as only one reference must exist
               break;
            }
         }
      }
   }
   
   /**
    * Cleanup the pending shared resources, as external program instantiation failed.
    */
   static void cleanupPending()
   {
      WorkArea wa = local.get();
      
      for (SharedRegistry sr : wa.resources)
      {
         // TODO: is there any resource-specific cleanup which should be done? 
         // i.e. discard temp-tables?

         // VARIABLE: nothing to do, just clean
         // STREAM: nothing to do, just clean
         // BUFFER: ??
         // TEMP-TABLE: ??
         // FRAME: ??
         // MENU: ??

         sr.pending = null;
      }
      
      wa.pendingVars = null;
   }
   
   /**
    * Core add worker that defers adding (using a pending list in the work area) if this is a 
    * <code>ScopeLevel.NEXT</code> request, otherwise the instance will be mapped into the current
    * or global scope immediately.
    *
    * @param    sr
    *           The {@link SharedRegistry} container where to add the var.
    * @param    scope
    *           The scope to target.
    * @param    name
    *           The symbolic name for the shared instance.
    * @param    var
    *           The instance to be shared.
    */
   private static void addWorker(SharedRegistry sr, ScopeLevel scope, String name, Object var)
   {
      if (scope == ScopeLevel.NEXT)
      {
         if (sr.pending == null)
         {
            sr.pending = new HashMap<>();
         }
         
         // we don't have to worry about the global flag as these can't be global, just save off
         // the case-insensitive name and instance
         sr.pending.put(name.toLowerCase(), var);
      }
      else
      {
         sr.shared.addSymbol((scope == ScopeLevel.GLOBAL), name, var);
      }
   }
   
   /**
    * Core lookup worker that returns the first instance found by the
    * given name.  The order of lookup is:
    * <ol>
    *   <li> pending list
    *   <li> each of the scopes starting at the current scope and moving
    *        through the stack to the global scope (if it exists)
    * </ol>
    *
    * @param    sr
    *           The {@link SharedRegistry} container where to add the var.
    * @param    name
    *           The symbolic name for the shared instance.
    * @param    error
    *           If no instance is found, use this to construct and throw the ERROR condition.
    *           
    * @return   The found resource, from the given context-local registry.
    */
   private static Object lookupWorker(SharedRegistry sr, String name, Runnable error)
   throws ErrorConditionException
   {
      Object result = null;
      
      if (sr.pending != null)
      {
         // names are always case-insensitive
         result = sr.pending.get(name.toLowerCase());
      }
      
      if (result == null)
      {
         result = sr.shared.lookupSymbol(name);
      }
      
      if (result == null && error != null)
      {
         error.run();
      }
      
      return result;
   }
   
   /**
    * Simple helper to construct a {@link ErrorConditionException} and throw it, when needed.
    *
    * @param    num
    *           The error number.
    * @param    msg
    *           The error message.
    *
    * @throws   ErrorConditionException
    *           Always, this will be caught by ControlFlowOps and decide to log or re-throw 
    *           the error.
    */
   private static void errorHelper(int num, String msg)
   throws ErrorConditionException
   {
      ErrorConditionException ece = new ErrorConditionException(num, msg);
      throw ece;
   }
   
   /**
    * Check if the shared temp-table DMO class matches the found master temp-table.
    * 
    * @param    found
    *           The found master temp-table.
    * @param    name
    *           The shared temp-table name.
    * @param    expected
    *           The expected temp-table to be defined by the found master.
    */
   private static void checkSharedTempTable(Object                     found, 
                                            String                     name, 
                                            Class<? extends Temporary> expected)
   {
      BooleanSupplier pred = () -> (found != null && expected.isAssignableFrom(found.getClass()));
            
      Runnable error = () ->
      {
         String msg = 
            "In procedure %s, shared temp-table %s has a conflict in field, index or undo status";
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         msg = String.format(msg, pname, name);
         
         errorHelper(2075, msg);
      };
      
      if (!checkSharedResource(pred, error))
      {
         throw new IllegalStateException(
            "An exception should have been thrown by checkedSharedResource!");
      }
   }
   
   /**
    * Check if the defined shared variable matches the existing master variable.
    * 
    * @param    name
    *           The shared variables name.
    * @param    var
    *           The shared var instance (which is attempting to be defined).
    * @param    noUndo
    *           <code>true</code> if this variable has its NO-UNDO option set.
    */
   private static void checkSharedVariable(String  name, Object  var, boolean noUndo)
   {
      BooleanSupplier pred = () -> (variablesMatch(name, var, noUndo));
      
      Runnable error = () ->
      {
         String msg = "%s Conflict in extent, datatype, or undo status for shared %s";
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         msg = String.format(msg, pname, name);
         
         errorHelper(391, msg);
      };
      
      if (!checkSharedResource(pred, error))
      {
         throw new IllegalStateException(
            "An exception should have been thrown by checkedSharedResource!");
      }
   }
   
   /**
    * Check if the candidate slave shared variable matches an existing master variable.
    * 
    * @param    name
    *           The shared var name.
    * @param    var
    *           The slave variable which needs to match the master one.
    * @param    noUndo
    *           Flag indicating if the slave var was defined as NO-UNDO.
    */
   private static boolean variablesMatch(String name, Object var, boolean noUndo)
   {
      WorkArea wa = local.get();

      VarRegistry v = wa.getPendingVars(false);
      SharedRegistry sr = (var.getClass().isArray() ? v.extentVars : v.vars);

      Object shared = sr.shared.lookupSymbol(name);
      
      if (shared == null)
      {
         return false;
      }

      // check if variable matches the registered one
      boolean canAssign = BaseDataType.sameType(shared, var);
      
      if (canAssign)
      {
         // check no-undo
         SharedRegistry noUndoWa = v.noUndoVars;
         Object sharedNoUndoVar = noUndoWa.shared.lookupSymbol(name);
         
         boolean sharedNoUndo = (sharedNoUndoVar != null);
         
         if (sharedNoUndo != noUndo)
         {
            canAssign = false;
         }
      }

      return canAssign;
   }

   /**
    * Check if the specified match condition is <code>true</code>.  If it is not, it will execute
    * the specified error code.
    * 
    * @param    pred
    *           The predicate to evaluate if the resource matches the found master.
    * @param    error
    *           The error code to execute.
    * 
    * @return   <code>true</code> if the resource matches the master.
    */
   private static boolean checkSharedResource(BooleanSupplier pred, Runnable error)
   {
      if (!pred.getAsBoolean())
      {
         error.run();
         return false;
      }
      
      return true;
   }
   
   /** 
    * Stores the state of the current context.
    */
   private static class WorkArea
   {
      /** MAp with each buffer's external program, where it was defined. */
      private Map<Object, Object> bufferCreator = new HashMap<>();

      /** The shared var dictionaries, saved per each external program. */
      private Map<Object, VarRegistry> varDictionaries = new IdentityHashMap<>();
      
      /** The vars being defined in the currently instantiating external program. */
      private VarRegistry pendingVars = null;

      /** Global variable pool which is context-local. Initializes upon first use. */ 
      private VarRegistry globalVars = new VarRegistry();
      
      /**
       * Track references for all shared variables, in all registries (not defined/used) - when
       * the reference count reaches zero, they will be removed from the transaction-related
       * registries.
       */
      private IdentityHashMap<Object, AtomicInteger> varRefs = new IdentityHashMap<>();
      
      /** Pool of streams which is context-local. Initializes upon first use. */ 
      private SharedRegistry streams = new SharedRegistry();
      
      /** Pool of buffers which is context-local. Initializes upon first use. */ 
      private SharedRegistry buffers = new SharedRegistry();
      
      /** Temp-table pool which is context-local. Initializes upon first use. */ 
      private SharedRegistry tempTables = new SharedRegistry();
      
      /** Pool of frames which is context-local. Initializes upon first use. */ 
      private SharedRegistry frames = new SharedRegistry();
      
      /** Pool of menus which is context-local. Initializes upon first use. */ 
      private SharedRegistry menus = new SharedRegistry();

      /** All the shared data. */
      private SharedRegistry[] resources = 
      {
         streams,
         buffers,
         tempTables,
         frames,
         menus
      };
      
      /**
       * Resolve the var registry or return the {@link #pendingVars} if it was already resolved.
       * 
       * @param    source
       *           Flag indicating which procedure to use as master registry: when <code>true</code>,
       *           use SOURCE-PROCEDURE, otherwise THIS-PROCEDURE.
       *           
       * @return   The {@link #pendingVars} registry.
       */
      public VarRegistry getPendingVars(boolean source)
      {
         if (pendingVars != null)
         {
            return pendingVars;
         }
         
         Object referent = (source ? ProcedureManager._sourceProcedure() 
                                   : ProcedureManager._thisProcedure());
         
         VarRegistry vars = null;
         if (referent == null)
         {
            // the startup program doesn't have a SOURCE-PROCEDURE, but in case of appservers,
            // the global var registry must be inherited from one API call to another (thus these 
            // must be kept 'session global', and shared among all procedures, for ease of
            // access).
            vars = globalVars.clone();
         }
         else
         {
            vars = varDictionaries.get(referent);
            vars = vars.clone();
         }
         
         pendingVars = vars;
         
         return pendingVars;
      }
   }

   /**
    * Container for the registries related to shared variables.
    */
   private static class VarRegistry
   implements Cloneable
   {
      /** Pool of extent vars which is context-local (with the state at their definition). */ 
      private SharedRegistry extentVars = new SharedRegistry();
      
      /** Pool of NO-UNDO vars which is context-local (with the state at their definition). */ 
      private SharedRegistry noUndoVars = new SharedRegistry();
      
      /** Variable pool which is context-local. Initializes upon first use. */ 
      private SharedRegistry vars = new SharedRegistry();
      
      /** All the shared data. */
      private SharedRegistry[] resources = 
      {
         vars,
         extentVars,
         noUndoVars
      };

      /**
       * Clone this instance.
       * 
       * @return   A copy of this instance.
       */
      @Override
      public VarRegistry clone()
      {
         VarRegistry var = new VarRegistry();
         
         for (int i = 0; i < resources.length; i++)
         {
            var.resources[i].shared.copyAll(resources[i].shared, true);
         }
         
         return var;
      }
   }
   
   /** 
    * Stores the registry for a specified shared resource.
    */
   private static class SharedRegistry
   {
      /** 
       * Instances mapped by case-insensitive name and organized by scope.  Use this version of
       * the constructor so that a global scope is added during construction. 
       */
      private ScopedSymbolDictionary<Object> shared = new ScopedSymbolDictionary<Object>(null, false);
      
      /** Instances that must be added to the NEXT scope. */
      private Map<String, Object> pending = null;
   }
}