/*
** Module   : SourceNameMapper.java
** Abstract : Progress procedure file name to Java class name run-time 
**            mapping (bidirectional) 
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SIY 20050817   @22156 Created initial version.
** 002 GES 20060124   @24072 Rewrite to move to a directory based 
**                           approach for storage of the mapping data.
**                           In addition, the mapping data used at 
**                           runtime is stored in context-local instances 
**                           to ensure that multiple configurations can
**                           be run simultaneously. Added a reverse
**                           mapping from Java to Progress names.
** 003 GES 20060201   @24185 Added external init option so that the class
**                           can be initialized from a source other than
**                           the directory.
** 004 NVS 20070126   @32009 "class-mapping" node of the directory is 
**                           searched as an account level standard search
**                           item.
** 005 ECF 20070220   @32186 Made map access thread-safe. Synchronized
**                           initialization of mapping data with read and
**                           write requests. This fixes a race condition
**                           which was resulting in an NPE.
** 006 GES 20080722   @39167 Added a fallback for convert() to allow a
**                           parameter which is already an existing
**                           fully qualified class name, to be returned.
**                           Major rewrite to load the name mapping data
**                           from a JAR file resource.  The new resource
**                           has a much simpler format.
** 007 SVL 20090310   @41448 Keys into methodMap for the Progress
**                           proc/function name to Java name mapping
**                           are always lower-cased.
** 008 GES 20090515   @42235 Import change.
** 009 ECF 20101004          Handle unqualified file name references in
**                           convertName method.
** 010 CA  20121220          Reworked the mapping to include the new emitted
**                           info (by collect_names.rules), needed by the
**                           persistent and super procedures implementation.
** 017 CA  20130209          Added dynamic EXTENT and EXTENT parameter support.
** 018 CS  20130304          Added support for checking if a class exists in a
**                           certain package as result of conversion from file
** 019 EVL 20130412          Fixing the regular expressions for windows runtime in
**                           initMappingData() method to properly handle windows file separator.
** 020 EVL 20130529          Replacing getLegacySearchPath() with getSearchPath() to obtain
**                           full effective PROPATH variable value to reflect the changes in
**                           PROPATH approach implementation.
** 021 CA  20130813          Added constants for the parameter modes.
** 022 CA  20130925          Added support for PUBLISHED-EVENTS attribute.
** 023 CA  20131116          Added support for the FUNCTION ... MAP TO clause.
** 024 GES 20131218          Rework to move InternalEntry, Parameter and other logic/data out
**                           into separate classes. This makes the overall code easier to
**                           understand and more extendable.  Refactored some methods to use
**                           common code. Completed native API support.
** 025 CA  20140218          Changes related to PROGRAM-NAME implementation: added 
**                           convertJavaMethod to resolve a Java method name to a legacy name.
**                           Removed convertStackEntry, as implemenetation was moved to 
**                           ProcedureManager.
** 026 EVL 20140310          The canonicalize() and convertNames() methods have been modified to
**                           take into account possible usage of the \ file separator in Windows
**                           for program name.  Currently in run-time case the file separator has
**                           constant value of "/".  All Windows separators will be replaced with
**                           this one.  For Windows the convertNames() method handles the absolute
**                           file names with drive letter included and relative file names.  For
**                           this support the properly configured searchpath value in server's
**                           directory.xml file is required.  But the existence of the 4GL source
**                           code is not expected on the target system.  The canonicalize() method
**                           was removed from the code part where is does nothing useful(because
**                           the input parameters name already "canonicalized").
** 027 ECF 20140913          Reduced memory footprint.
** 028 ECF 20150424          Minor memory tweak.
** 029 ECF 20150601          Minor changes to accommodate InternalEntry changes. Removed some
**                           unnecessary intern()'ing of strings.
** 030 OM  20150710          Added support for obtaining the type of a parameter.
** 031 ECF 20150805          Refactored synchronization to improve throughput. Replaced
**                           StringBuffer with StringBuilder.
** 032 CA  20160418          Added API to resolve an internal-entry (or external program) list of
**                           parameters.
** 033 OM  20160421          Identified 4GL source searched by their legacy absolute path. Added
**                           generics. Javadoc fixes. foreach iterations.
** 034 OM  20160527          SourceNameMapper.convertName() was hidden and exposed two methods that
**                           return converted class-name and legacy file name.
** 035 OM  20160711          Added MEMPTR & RAW support in getTypeForParam() method.
** 036 CA  20161020          Fixed case-insensitive resolution with .r (rcode) ext program names.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import java.net.*;
import org.w3c.dom.*;
import com.goldencode.util.*;
import com.goldencode.p2j.library.*;

/**
 * This class performs a bidirectional run-time mapping of a Progress
 * procedure file name to and from a Java class name.  It also supports
 * run-time translation of Progress internal procedure names into Java
 * method names.
 * <p>
 * <b>The mapping data is loaded from the directory into data structures that
 * are <code>static</code>.  This means that at this time this mapping data
 * is global to the entire server rather than being context-local.  The
 * reason for this is to reduce the amount of memory used for this data
 * since 2+ map entries exist for each original source file (thus these maps
 * each contain thousands of entries for any reasonably sized
 * application).</b>
 * <p>
 * <b>The current implementation assumes a single root path for all Progress
 * procedure file names and single base package path for the generated Java
 * classes.  Despite these assumptions it is possible to translate
 * Progress procedure names incorrectly (to a Java class name which does not
 * belong to the subtree specified by the single root or which does not 
 * belong to the base package path).  These problems are caused by using
 * mismatching absolute paths for Progress names or names which start with
 * a dot ('&#46;') for Java class names.</b> The leading dot is stripped
 * from the returned class name and is not returned as a result of a
 * mapping.
 * <p>
 * The source file name can be in one of the many forms: relative path,
 * absolute path, can contain references to current and to upper level
 * directories. Also, the source file name can use the file naming format of
 * the original operating system used to run the Progress program rather than
 * the current operating system used to run the converted Java application.
 * To this the directory contains values that define the original source
 * system's parameters such as path delimiter, file name delimiter and case 
 * sensitivity.
 */
public class SourceNameMapper
{
   /** Class mappings element. */
   private static final String ELEM_CMAP = "class-mapping";
   
   /** Methods mappings element. */
   private static final String ELEM_MMAP = "method-mapping";
   
   /** Parameter info element. */
   private static final String ELEM_PARAM = "parameter";
   
   /** Progress name attribute. */
   private static final String ATTR_PNAME = "pname";
   
   /** Java name attribute. */
   private static final String ATTR_JNAME = "jname";
   
   /** TYPE attribute. */
   private static final String ATTR_TYPE = "type";
   
   /** LIBNAME attribute. */
   private static final String ATTR_LIBNAME = "libname";
   
   /** MAP-TO attribute. */
   private static final String ATTR_MAP_TO = "map-to";
   
   /** Parameter MODE attribute. */
   private static final String ATTR_PARAM_MODE = "mode";
   
   /** Named events attribute. */
   private static final String PUBLISHED_EVENTS = "published-events";
   
   /** Progress procedure file name to Java class name map table. */
   private static volatile Map<String, ExternalProgram> p2jMap = null;
   
   /** Progress Java class name to procedure file name map table. */
   private static Map<String, ExternalProgram> j2pMap = null;
   
   /** 
    * The 4GL legacy roots for source files. Used for detecting full-paths file names.
    * Each entry identifies a project which is run on the current JVM. 
    */
   private static String[] legacyRoots = null;
   
   /** Base path for all Java packages. */
   private static String pkgroot = null;
   
   /** Custom package to init data for. */
   private static String custpkgroot = null;
   
   /** Base path for Progress 4GL code. */
   private static String proroot = null;
   
   /** 
    * Name element separator used for the Progress file name data.  At runtime this is OS
    * independent value used internally in name_map.xml file to resolve procedure names.
    */
   private static String fileSep = "/"; 
   
   /** Path separator used for the Progress file name data. */
   private static String pathSep = null;
   
   /** Case sensitivity flag for the Progress file name data. */
   private static boolean caseSens;
   
   /**
    * Get the legacy 4gL name for the Java class name. 
    * 
    * @param   jname
    *          The external-program's converted java name.
    * 
    * @return  The legacy program name or <code>null</code> if the given java
    *          program is not registered in name_map.xml.
    */
   public static String convertJavaProg(String jname)
   {
      initMappingData();
      ExternalProgram extp = j2pMap.get(removePkg(jname));
      return extp == null ? null : extp.pname;
   }
   
   /**
    * Get the legacy 4gL name for the Java class name or, if not a legacy program, return the 
    * java class name, without the base package. 
    * 
    * @param   jname
    *          A java class name which may or may not be associated with a legacy 4GL program.
    * 
    * @return  The legacy program name or the java class name, without the base package.
    */
   public static String convertJavaProgSafe(String jname)
   {
      String pname = convertJavaProg(jname);
      if (pname == null)
      {
         pname = removePkg(jname);
      }
      
      return pname;
   }
   
   /**
    * Get the legacy 4gL name for the Java class name. 
    * 
    * @param   jname
    *          The external-program's converted java name.
    * @param   customRoot
    *          Custom root package name.
    * 
    * @return  The legacy program name or <code>null</code> if the given java
    *          program is not registered in name_map.xml.
    */
   public static String convertJavaProg(String customRoot, String jname)
   {
      custpkgroot = customRoot;
      return convertJavaProg(jname);
   }
   
   /**
    * Convert a java method name to a legacy name.
    * 
    * @param    clsName
    *           The class name where the method is defined.
    * @param    methodName
    *           The method's name.
    * @param    function
    *           Flag indicating that this method maps a legacy function (when {@code true)} or 
    *           procedure (when {@code false}).
    *           
    * @return   The legacy method name or {@code null} if not found.
    */
   public static String convertJavaMethod(String clsName, String methodName, boolean function)
   {
      initMappingData();
      ExternalProgram extp = j2pMap.get(removePkg(clsName));
      
      if (extp != null)
      {
         Map<String, InternalEntry> ieMap = (function ? extp.j2pf : extp.j2pp);
         InternalEntry ie = ieMap.get(methodName);
         return (ie == null ? null : ie.pname);
      }
      
      // if not found
      return null;
   }
   
   /**
    * Normalize the given legacy external program name, considering the {@link #caseSens} state.
    * 
    * @param    legacy
    *           The legacy name to be normalized
    *           
    * @return   The normalized legacy name.
    */
   public static String normalizeLegacyName(String legacy)
   {
      if (legacy != null && !caseSens)
      {
         legacy = legacy.toLowerCase();
      }
      
      return legacy;
   }
   
   /**
    * Uses the {@link #convertName(String)} worker for finding the name of the Java class name of
    * a converted P4GL procedure name. 
    * 
    * @param   legacy
    *          Progress file name.
    *
    * @return  the Java class name for the given Progress file name if it is identified as a
    *          converted procedure from legacy sources. If no appropriate mapping is present,
    *          {@code null} is returned.
    */
   public static String convertNameToClass(String legacy)
   {
      String[] paths = convertName(legacy);
      return (paths != null) ? paths[0] : null;
   }
   
   /**
    * Uses the {@link #convertName(String)} worker for detecting the if the {@code legacy} is a
    * name of a converted legacy procedure. In affirmative case, the normalized name of the 
    * procedure filename is returned.
    *
    * @param   legacy
    *          Progress file name.
    *
    * @return  The legacy procedure name. If no appropriate mapping is present {@code null} is
    *          returned.
    */
   public static String lookupLegacyName(String legacy)
   {
      String[] paths = convertName(legacy);
      return (paths != null) ? paths[1] : null;
   }
   
   /**
    * This method loads all mapping data from an external source. DO NOT
    * USE THIS UNLESS YOU KNOW EXACTLY WHAT YOU ARE DOING!!!!!  All data
    * structures are overwritten (including anything read from the
    * directory).  No safety or integrity checking is done at all.
    *
    * @param    p2jMap
    *           Progress procedure file name to Java class name map table.
    * @param    j2pMap
    *           Progress Java class name to procedure file name map table.
    * @param    pkgroot
    *           Base path for all Java packages.
    * @param    proroot
    *           Base path for 4GL code.
    * @param    fileSep
    *           Name element separator used for the Progress file name data.
    * @param    pathSep
    *           Path separator used for the Progress file name data.
    * @param    caseSens
    *           Case sensitivity flag for the Progress file name data.
    */
   public synchronized static void overwriteMappingData(Map<String, ExternalProgram> p2jMap,
                                                        Map<String, ExternalProgram> j2pMap,
                                                        String pkgroot,
                                                        String proroot,
                                                        String fileSep,
                                                        String pathSep,
                                                        boolean caseSens)
   {
      SourceNameMapper.p2jMap   = p2jMap;
      SourceNameMapper.j2pMap   = j2pMap;
      SourceNameMapper.pkgroot  = pkgroot;
      SourceNameMapper.proroot  = proroot;
      SourceNameMapper.fileSep  = fileSep; 
      SourceNameMapper.pathSep  = pathSep;
      SourceNameMapper.caseSens = caseSens;
   }

   /**
    * Get the java name of the passed internal-entry and 4GL program.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    * @param    function
    *           <code>true</code> if the search is for a user-defined 
    *           function.
    *           
    * @return   The Java name of the given internal-entry, if this entry 
    *           exists in the given program. Else, <code>null</code>
    */
   public static String getMethodName(String  pname, String  ieName, boolean function)
   {
      initMappingData();
      
      ExternalProgram extProg = p2jMap.get(pname);
      if (extProg == null)
      {
         return null;
      }
      
      InternalEntry method = extProg.searchp(ieName, function);
      if (method == null)
      {
         return null;
      }
      
      return method.jname;
   }

   /**
    * Get the map-to attribute of the passed internal-entry. The internal entry must be a 
    * function.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    *           
    * @return   The map-to attribute of the passed internal-entry, if is a function. 
    *           <code>null</code> if the {@link #ATTR_MAP_TO} does not exist or the internal
    *           entry does not resolve to a function.
    */
   public static String getMapTo(String pname, String ieName)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(pname);
      if (extProg == null)
      {
         return null;
      }
      
      InternalEntry ie = extProg.search(ieName, true);
      if (ie == null)
      {
         return null;
      }
      
      return ie.getAttribute(ATTR_MAP_TO);
   }

   /**
    * Get the library name attribute of the given internal-entry.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    iename
    *           Legacy internal-entry name.
    *           
    * @return   The library name attribute of the given internal-entry name,
    *           assuming the program exists and the internal-entry is a native
    *           procedure definition. Else, <code>null</code>.
    */
   public static String getLibraryName(String pname, String iename)
   {
      InternalEntry ie = getInternalEntry(pname, iename, false);
      
      if (ie == null)
      {
         return null;
      }
      
      String type = ie.getAttribute(ATTR_TYPE);
      
      if (InternalEntry.Type.DLL_ENTRY.toString().equalsIgnoreCase(type))
      {
         return ie.getAttribute(ATTR_LIBNAME);
      }
      
      return null;
   }

   /**
    * Get the string representation of the parameter modes for this external
    * procedure.  If there are no parameters, return null.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    *           
    * @return   See above.
    */
   public static String getParameterModes(String pname)
   {
      return getParameterModes(pname, null, false);
   }

   /**
    * Get the string representation of the parameter modes for this internal
    * entry. If there are no parameters, return null.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name. If null or the empty string, the
    *           external procedure's parameters are used.
    * @param    function
    *           <code>true</code> if this is a function call.
    *           
    * @return   See above.
    */
   public static String getParameterModes(String pname, String ieName, boolean function)
   {
      initMappingData();
      
      ExternalProgram extProg = p2jMap.get(pname);
      if (extProg == null)
      {
         return null;
      }

      if (ieName == null || "".equals(ieName))
      {
         return extProg.main.getParameterModes();
      }
      
      InternalEntry ie = extProg.search(ieName, function);
      
      return ie == null ? null : ie.getParameterModes();
   }

   /**
    * Get the parameters for this internal entry. If there are no parameters, return an empty list.
    * If the internal entry can't be found, return null.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name. If null or the empty string, the
    *           external procedure's parameters are used.
    * @param    function
    *           <code>true</code> if this is a function call.
    *           
    * @return   See above.
    */
   public static List<Parameter> getParameters(String pname, String ieName, boolean function)
   {
      initMappingData();
      
      ExternalProgram extProg = p2jMap.get(pname);
      if (extProg == null)
      {
         return null;
      }

      if (ieName == null || "".equals(ieName))
      {
         return extProg.main.getParameters();
      }
      
      InternalEntry ie = extProg.search(ieName, function);
      
      return ie == null ? null : ie.getParameters();
   }
   
   /**
    * Get the 4GL-style signature of the given internal-entry.  If the program
    * does not exist or the internal-entry does not belong to this program,
    * return <code>null</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    *           
    * @return   See above.
    */
   public static String getSignature(String pname, String ieName)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(pname);
      if (extProg == null)
      {
         return "";
      }
      
      if (ieName == null || "".equals(ieName))
      {
         return extProg.main.getSignature();
      }
      
      InternalEntry ief = extProg.search(ieName, true);

      // if both func and proc are with the same name, then func has priority
      InternalEntry ie = ief != null ? ief : extProg.search(ieName, false);
      
      return ie == null ? "" : ie.getSignature();
   }

   /**
    * Get the internal-entries attribute of the given external procedure.
    * If the program does not exist, return <code>null</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    prv
    *           Flag to include or not the private members.
    *           
    * @return   See above.
    */
   public static String getInternalEntries(String pname, boolean prv)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(pname);
      
      return (extProg == null) ? null : extProg.getInternalEntries(prv);
   }
   
   /**
    * Get the published named events by the specified procedure.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    *           
    * @return   The comma-separated list of published named events, when the event name is 
    *           hard-coded.
    */
   public static String getPublishedEvents(String pname)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(pname);
      
      return (extProg == null) ? null : extProg.publishedEvents;
   }

   /**
    * Get the extent of the parameter at 0-based position (for the given internal entry).
    * If the parameter is dynamic-extent, then return 0, not -1.
    *  
    * @param    pname
    *           Legacy 4GL program name.
    * @param    iename
    *           Legacy internal-entry name.
    * @param    func
    *           Flag indicating if this is a function or a procedure.
    * @param    param
    *           0-based index of the parameter to be checked.
    *
    * @return   See above.
    */
   public static int getExtentForParam(String pname, String iename, boolean func, int param)
   {
      Parameter p = getParameter(pname, iename, func, param);
      return (p == null) ? -1 : p.getExtent();
   }
   
   /**
    * Get the type/class of the parameter at 0-based position (for the given internal entry).
    *  
    * @param   pname
    *          Legacy 4GL program name.
    * @param   iename
    *          Legacy internal-entry name.
    * @param   func
    *          Flag indicating if this is a function or a procedure.
    * @param   param
    *          0-based index of the parameter to be checked.
    *
    * @return  the class of the requested parameter or {@code null} if parameters do not point
    *          to a valid parameter.
    */
   public static Class getTypeForParam(String pname, String iename, boolean func, int param)
   {
      Parameter p = getParameter(pname, iename, func, param);
      if (p == null) 
      {  
         return null;
      }
      
      switch (p.getType().toUpperCase())
      {
         case "LOGICAL":    return logical.class;
         case "INTEGER":    return integer.class;
         case "INT64":      return int64.class;
         case "DECIMAL":    return decimal.class;
         case "DATE":       return date.class;
         case "DATETIME":   return datetime.class;
         case "DATETIMETZ": return datetimetz.class;
         case "CHARACTER":  return character.class;
         case "MEMPTR":     return memptr.class;
         case "RAW":        return raw.class;
         
         default: return null;
      }
   }
   
   /**
    * Check if the specified internal-entry is defined IN SUPER. If the program
    * does not exist or the internal entry does not belong to the program,
    * return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    iename
    *           Legacy internal-entry name.
    * @param    function
    *           Flag indicating if this is a function or a procedure.
    *
    * @return   See above.
    */
   public static boolean isSuper(String pname, String iename, boolean function)
   {
      InternalEntry ie = getInternalEntry(pname, iename, function);
      
      return (ie == null) ? false : ie.isSuper();
   }

   /**
    * Check if the specified internal-entry is a function. If the program
    * does not exist or the internal entry does not belong to the program,
    * return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    iename
    *           Legacy internal-entry name.
    *
    * @return   See above.
    */
   public static boolean isFunction(String pname, String iename)
   {
      return getInternalEntry(pname, iename, true) != null;
   }

   /**
    * Check if the specified internal-entry is defined PRIVATE. If the program
    * does not exist or the internal entry does not belong to the program,
    * return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    iename
    *           Legacy internal-entry name.
    * @param    function
    *           Flag indicating if this is a function or a procedure.
    *
    * @return   See above.
    */
   public static boolean isPrivate(String pname, String iename, boolean function)
   {
      InternalEntry ie = getInternalEntry(pname, iename, function);
      
      return (ie == null) ? false : ie.isPrivate();
   }

   /**
    * Check if the specified internal-entry is defined IN handle. If the 
    * program does not exist or the internal entry does not belong to the
    * program, return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    iename
    *           Legacy internal-entry name.
    * @param    function
    *           Flag indicating if this is a function or a procedure.
    *
    * @return   See above.
    */
   public static boolean isInHandle(String pname, String iename, boolean function)
   {
      InternalEntry ie = getInternalEntry(pname, iename, function);
      
      return (ie == null) ? false : ie.isInHandle();
   }

   /**
    * Helper function to obtain the Java object that holds the state of a single internal entry.
    *
    * @param    pname
    *           Legacy 4GL program name (the external procedure).
    * @param    iename
    *           Legacy internal-entry name.
    * @param    func
    *           <code>true</code> if the search should be for a function.
    *           
    * @return   The internal entry or <code>null</code> if the external procedure does not exist
    *           or if the external procedure has no such internal entry.
    */
   public static InternalEntry getInternalEntry(String pname, String iename, boolean func)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(pname);
      
      if (extProg == null)
      {
         return null;
      }
      
      return extProg.search(iename, func);
   }
   
   /**
    * Convert a Progress file name into Java class name. The input name can be an absolute file
    * name, a file name relative to current location or a file name relative to some directory
    * listed in the PROPATH variable. Before processing, the input file name is processed by
    * {@link #canonicalize}.
    * <p>
    * A simple lookup is tried first. If this fails direct check for absolute path if attempted
    * using one of the roots from {@code legacyRoots}. In multiple project scenario, the project
    * name is also added as a path token in order o match the structure from {@code name_map.xml}.
    * <p>
    * If these mapping does not exist then each path segment of the PROPATH is tried in order. 
    * The given program name is check to see if it is prefixed with the PROPATH segment, if so
    * that prefix is removed and a lookup is done with this name. The first match found ends the
    * search.
    * <p>
    * Absolute paths are converted into relative paths based on the existence of a matching path
    * in the legacy PROPATH (see {@link EnvironmentOps#getLegacySearchPath}). To handle additional
    * absolute paths, one must add the appropriate path prefix to the legacy PROPATH and it will
    * be removed here.
    * <p>
    * This same PROPATH processing is how names relative to the current directory can be handled.
    * This is the only way to handle that case since the current directory in the source system is
    * runtime information that is not available here.
    * <p>
    * The legacy PROPATH is the key configuration tool by which one specifies to this method how
    * to remove absolute and current directory relative prefixes.  This means that the legacy
    * PROPATH that is used here may need to be different than the PROPATH on the original system.
    * <p>
    * As a fallback, if no mapping can be found and if the given {@code progName} is a fully
    * qualified Java class name which references a valid Java class, then the {@code progName} is
    * returned unchanged.
    * <p>
    * In Windows it can verify the absolute or relative program name is in current PROPATH 
    * directory set.  This allows to properly map the following program names used in
    * PROGRESS RUN statement:
    * "c:/path/to/project/stcases/uast/rundir/absolute/program.p"
    * "rundir/../../uast/rundir/relative/test11.p."
    * "../uast/rundir/relative/test5.p"
    *
    * @param   legacyProgName
    *          Progress file name.
    *
    * @return  A {@code String} array with two components: Java class name for the given Progress
    *          file name as first element and the legacy procedure name in second element of the
    *          array. If no appropriate mapping is present {@code null} is returned.
    */
   private static String[] convertName(String legacyProgName)
   {
      initMappingData();
      
      // get the original propath and split it into pieces
      String searchPath = EnvironmentOps.getSearchPath().toStringMessage();
      
      // get rid of the Windows file separator in working strings
      if (EnvironmentOps.isUnderWindowsFamily())
      {
         legacyProgName = legacyProgName.replaceAll("\\\\", fileSep);
         searchPath = searchPath.replaceAll("\\\\", fileSep);
      }
      
      // internally, we use the canonicalized paths (lowercased and / for path separator)
      // but for the result, we use the original case and path separator (as 4GL does)
      String[] originalPropath = searchPath.split(EnvironmentOps.PROPATH_SEPARATOR);
      String[] propath;
      String   progName;
      
      if (!caseSens)
      {
         propath = new String[originalPropath.length];
         for (int i = 0; i < originalPropath.length; i++)
         {
            propath[i] = originalPropath[i].toLowerCase();
         }
         progName = legacyProgName.toLowerCase();
      }
      else
      {
         propath = originalPropath;
         progName = legacyProgName;
      }
      
      // check the simple form of the name
      String retLegacy = legacyProgName;
      ExternalProgram extProg = p2jMap.get(canonicalize(progName));
      
      // check if the progName represents an absolute filename of a source procedure
      if (extProg == null)
      {
         for (String aRoot : legacyRoots)
         {
            if (progName.startsWith(aRoot))
            {
               String relName = progName.substring(aRoot.length());
               if (relName.startsWith(fileSep))
               {
                  relName = relName.substring(1); // drop the starting fileSep
               }
               extProg = p2jMap.get(relName);
               if (extProg == null)
               {
                  // one more try, because of file collisions, the files from current project were
                  // converted from a 'token' directory of the src-root
                  String token = Utils.getProjectToken();
                  if (token != null)
                  {
                     extProg = p2jMap.get(token + fileSep + relName);
                  }
               }
               if (extProg != null)
               {
                  retLegacy = aRoot + fileSep + relName;
                  break; // stop iteration at first found source file
               }
            }
         }
      }
      
      // if we didn't find a match AND if there is a propath, then remove any paths that are in
      // the propath from the front, this is how we remove an absolute path name, then try a
      // lookup
      if (extProg == null && propath.length != 0)
      {
         // advanced scan with PROPATH involved
         String possible = null;
         
         // check the absolute path in Linux and Windows(excluding possible ".." up elements)
         for (String aPath : propath)
         {
            if (!aPath.contains("..") && progName.startsWith(aPath))
            {
               // remove the prefix
               possible = progName.substring(aPath.length());
               
               // remove any extra file separator
               if (possible.startsWith(fileSep))
               {
                  possible = possible.substring(1);
               }
               
               // check for a match
               extProg = p2jMap.get(canonicalize(possible));
               
               if (extProg != null)
               {
                  retLegacy = legacyProgName; // it should be an absolute path already
                  break; // first match exits the loop
               }
            }
         }
         
         // Check for relative path including up entries ".."
         if (extProg == null && progName.contains(".."))
         {
            for (int i = 0; i < propath.length; i++)
            {
               String aPath = propath[i];
               // clue path and filename suggest
               String relName = aPath.concat(fileSep).concat(progName);
               
               String[] lstRelative = relName.split(fileSep);
               
               LinkedList<String> listRelative = new LinkedList<>();
               // remove ".." values
               for (String aPathToken : lstRelative)
               {
                  if (aPathToken.equals(".."))
                  {
                     if (listRelative.size() > 0)
                     {
                        listRelative.removeLast();
                     }
                     
                     continue;
                  }
                  
                  listRelative.addLast(aPathToken);
               }
               
               // reconstruct new possible absolute name
               StringBuilder bufRelative = new StringBuilder(relName.length());
               
               for (Iterator iter = listRelative.iterator(); iter.hasNext(); )
               {
                  String element = (String) iter.next();
                  
                  bufRelative.append(element);
                  
                  if (iter.hasNext())
                  {
                     bufRelative.append(fileSep);
                  }
               }
               
               // New absolute name based on relative one but without extra ".."
               relName = bufRelative.toString();
               
               // Check if the program is in PROPATH
               if (relName.startsWith(aPath))
               {
                  possible = relName.substring(aPath.length());
                  // check for a match
                  extProg = p2jMap.get(canonicalize(possible));
                  
                  if (extProg != null)
                  {
                     retLegacy = originalPropath[i] + fileSep + legacyProgName;
                     break; // first match exits the loop
                  }
               }
            }
         }
         
         // if we still didn't find a match, strip the canonicalized base path off each
         // canonicalized component of the propath, then prepend that relative path to the name we
         // have and try to look it up
         if (extProg == null)
         {
            try
            {
               String base = Utils.canonicalizePath(proroot, File.separator, fileSep);
               
               // make sure there is a base path (it only exists at conversion)
               if (base.length() > 0)
               {
                  for (int i = 0; i < propath.length; i++)
                  {
                     String next = Utils.canonicalizePath(propath[i], File.separator, fileSep);
                     
                     if (!next.startsWith(base))
                     {
                        continue;
                     }
                     
                     next = next.substring(base.length());
                     possible = canonicalize(next + fileSep + progName);
                     
                     extProg = p2jMap.get(possible);
                     if (extProg != null)
                     {
                        retLegacy = originalPropath[i] + fileSep + legacyProgName;
                        break;
                     }
                  }
               }
            }
            catch (IOException exc)
            {
               // do nothing;  can't go further
            }
         }
      }
      
      // fallback processing
      if (extProg == null)
      {
         // enable support for fully qualified Java class names if they are passed in directly
         try
         {
            Class.forName(progName);
            
            // if we are here, then the progName must have been a fully qualified Java class name
            // to a class that exists so we return it unchanged. The legacy name is unknown and
            // unneeded/inexistent.
            return new String[] { progName, null };
         }
         catch (ClassNotFoundException cnfe)
         {
            // do nothing, the given program can't be found as a Java class
         }
      }
      
      if (extProg == null)
      {
         // search failed. The parameter does not contain a valid procedure name
         return null;
      }
      
      if (EnvironmentOps.isUnderWindowsFamily())
      {
         // 4GL automatically switch all slashes into backslashes on Win OS
         retLegacy = retLegacy.replaceAll(fileSep, "\\\\");
      }
      return new String[] { addPkg(extProg.jname), retLegacy };
   }
   
   /**
    * Private utility method for locating a parameter of a routine.
    *
    * @param   pname
    *          Legacy 4GL program name.
    * @param   iename
    *          Legacy internal-entry name.
    * @param   func
    *          Flag indicating if this is a function or a procedure.
    * @param   param
    *          0-based index of the parameter to be checked.
    *
    * @return  the requested parameter or {@code null} if parameters do not point to a valid
    *          parameter.
    */
   private static Parameter getParameter(String pname, String iename, boolean func, int param)
   {
      InternalEntry ie = getInternalEntry(pname, iename, func);
      
      if (ie == null || param < 0 || param >= ie.getParameterListSize())
      {
         return null;
      }
      
      return ie.getParameter(param);
   }
   
   /**
    * This method loads all mapping data from a URL which is calculated
    * using the <code>pkgroot</code> and the <code>name_map.xml</code> with
    * all "." characters converted to file separators and an extra file
    * separator added before the file name portion. This URL is expected to
    * be found in the application's JAR file. Note that this will silently
    * return if the mapping data is not found.  Any file system configuration
    * data will be read from the directory via the {@link EnvironmentOps}
    * helper.
    */
   private static void initMappingData()
   {
      if (p2jMap != null)
      {
         return;
      }
      
      synchronized (SourceNameMapper.class)
      {
         if (p2jMap != null)
         {
            return;
         }
         
         try
         {
            p2jMap = new HashMap<>(1024);
            j2pMap = new HashMap<>(1024);
            
            // lookup the descriptors for the original file system
            pathSep   = EnvironmentOps.getLegacyPathSeparator();
            caseSens  = EnvironmentOps.getLegacyCaseSensitive();
            pkgroot   = Utils.getDirectoryNodeString(null, "file-system/pkgroot", "");
            // if the pkgroot is null get the custom package and reset it.
            if (pkgroot == null || pkgroot.isEmpty())
            {
               pkgroot = custpkgroot;
               custpkgroot = null;
            }
            
            // this is only a conversion-time configuration value ("basepath")
            proroot = "";
            
            // at runtime we use legacyRoots
            String srcRoots = Utils.getDirectoryNodeString(null, "file-system/4gl-src-root", ".");
            if (!caseSens)
            {
               srcRoots = srcRoots.toLowerCase();
            }
            legacyRoots = srcRoots.split(pathSep);
            
            // build the fully qualified name of the name map XML file
            StringBuilder sb = new StringBuilder();
            sb.append(pkgroot.replaceAll("\\.", StringHelper.fixupRegex(fileSep)));
            
            if (!fileSep.equals(sb.substring(sb.length() - 1)))
            {
               sb.append(fileSep);
            }
            
            sb.append("name_map.xml");
            
            // create a URL to allow the name map to be loaded as a resource
            URL url = ClassLoader.getSystemResource(sb.toString());
            
            // safety first
            if (url == null)
            {
               // TODO: log this
               // no mapping data exists
               return;
            }
            
            Document dom = XmlHelper.parse(url.openStream());
            
            // find the root node
            Element root = dom.getDocumentElement();
            
            if (root == null)
            {
               // TODO: log this
               // no data in the resource
               return;
            }
            
            // enumerate all child nodes (there will be one for each source
            // file that can be mapped)
            NodeList cmaps = root.getElementsByTagName(ELEM_CMAP);
            
            if (cmaps == null)
            {
               // TODO: log this
               // no class mappings
               return;
            }
            
            int len = cmaps.getLength();
            
            for (int i = 0; i < len; i++)
            {
               Element classMap = (Element) cmaps.item(i);
               
               String pname = classMap.getAttribute(ATTR_PNAME);
               String jname = classMap.getAttribute(ATTR_JNAME);
               
               if (pname == null || pname.length() == 0 || jname == null || jname.length() == 0)
               {
                  String errmsg = "Malformed class mapping entry for " + classMap.toString();
                  throw new IllegalStateException(errmsg);
               }
               
               // remove case if required
               if (!caseSens)
               {
                  pname = pname.toLowerCase();
               }
               
               // store bidirectional mappings for the progress relative
               // filename to/from the java relative file name mapping
               ExternalProgram ep = buildExternalProgram(pname, jname, classMap);
               p2jMap.put(ep.pname, ep);
               j2pMap.put(ep.jname, ep);
            }
         }
         catch (Exception e)
         {
            e.printStackTrace();
            throw new RuntimeException("Unable to load name mapping data.");
         }
      }
   }
   
   /**
    * Collect the {@link ExternalProgram} info contained in the <code>classMap
    * </code> node.
    * 
    * @param    pname
    *           The legacy 4GL program name.
    * @param    jname
    *           The converted class name.
    * @param    classMap
    *           The DOM node containing the info for this external program.
    *           
    * @return   An {@link ExternalProgram} instance for this DOM node.
    */
   private static ExternalProgram buildExternalProgram(String pname,
                                                       String jname,
                                                       Element classMap)
   {
      String publishedEvents = classMap.getAttribute(PUBLISHED_EVENTS);
      
      ExternalProgram extProg = new ExternalProgram(pname, jname, publishedEvents);
      
      // enumerate method mapping nodes
      NodeList mmaps = classMap.getElementsByTagName(ELEM_MMAP);
      
      if (mmaps == null)
      {
         return extProg;
      }
      
      int mlen = mmaps.getLength();
      
      if (mlen > 0)
      {
         extProg.ieList = new ArrayDeque<>();
      }
      
      Set<String> defFuncs = new HashSet<>();
      Set<String> forwardFuncs = new HashSet<>();
      
      // process each method mapping
      // 4GL allows same name for a function or procedure, as long as they
      // are not internal. they appear in the internal-entry list, but the
      // get-signature call will see only the function (for this "feature" to
      // work, the procedure must be defined before the function, else 4GL
      // gets compile error).
      for (int j = 0; j < mlen; j++)
      {
         Element methMap = (Element) mmaps.item(j);
         
         String iepname = methMap.getAttribute(ATTR_PNAME);
         String iejname = methMap.getAttribute(ATTR_JNAME);
         
         if (iepname == null || iepname.length() == 0 ||
             iejname == null || iejname.length() == 0)
         {
            String errmsg = "Malformed method mapping entry for " + methMap.toString();
            throw new IllegalStateException(errmsg);
         }
         
         InternalEntry ie = buildMethod(iepname, iejname, methMap);
         
         // the FORWARD functions are emitted in the name_map.xml only to preserve the order.
         if (ie.isFunction())
         {
            String forward = methMap.getAttribute("forward");
            if ("true".equalsIgnoreCase(forward))
            {
               forwardFuncs.add(iepname);
            }
            else
            {
               defFuncs.add(iepname);
               
               // if the function already exists as forward, take the parameter
               // names from here and update the forward declaration
               if (forwardFuncs.contains(iepname))
               {
                  if (ie.getParameterListSize() > 0)
                  {
                     InternalEntry ief = extProg.search(iepname, true);
                     ief.setParameterList(ie.getParameterList());
                  }
                  
                  continue;
               }
            }
         }
         
         // store the bidirectional Progress relative proc/function name from/to Java method name
         // mapping
         extProg.ieList.add(ie);
         
         Map<String, InternalEntry> p2j = null;
         Map<String, InternalEntry> j2p = null;
         if (ie.isFunction())
         {
            if (extProg.p2jf.isEmpty())
            {
               extProg.p2jf = new HashMap<>();
               extProg.j2pf = new HashMap<>();
            }
            p2j = extProg.p2jf;
            j2p = extProg.j2pf;
         }
         else
         {
            if (extProg.p2jp.isEmpty())
            {
               extProg.p2jp = new HashMap<>();
               extProg.j2pp = new HashMap<>();
            }
            p2j = extProg.p2jp;
            j2p = extProg.j2pp;
         }
         p2j.put(iepname.toLowerCase().intern(), ie);
         j2p.put(ie.jname, ie);
      }
      
      // remove all functions defined as forward but without body
      forwardFuncs.removeAll(defFuncs);
      for (String iepname : forwardFuncs)
      {
         InternalEntry ie = extProg.search(iepname, true);
         extProg.ieList.remove(ie);
      }
      
      extProg.main = new InternalEntry("", "", InternalEntry.Type.MAIN);
      buildParameters(classMap, extProg.main);
      extProg.main.initialize();
      
      return extProg;
   }
   
   /**
    * Collect the {@link InternalEntry} info contained in the <code>methMap</code> node.
    * 
    * @param    pname
    *           The legacy 4GL internal-entry name.
    * @param    jname
    *           The converted internal-entry name.
    * @param    methMap
    *           The DOM node containing the info for this internal-entry.
    *           
    * @return   An {@link InternalEntry} instance for this DOM node.
    */
   private static InternalEntry buildMethod(String pname, String jname, Element methMap)
   {
      String type = methMap.getAttribute(ATTR_TYPE);
      
      if (type == null || type.length() == 0)
      {
         String err = "Missing type attribute for method mapping entry " + methMap.toString();
         throw new IllegalStateException(err);
      }
      
      InternalEntry ie = null;
      
      if (InternalEntry.Type.DLL_ENTRY.toString().equals(type))
      {
         // instantiate the subclass that handles native calls
         ie = new NativeAPIEntry(pname, jname);
      }
      else
      {
         // the common case
         ie = new InternalEntry(pname, jname, type);
      }
      
      buildParameters(methMap, ie);
      
      int len = methMap.getAttributes().getLength();
      
      // add the method's attributes
      for (int i = 0; i < len; i++)
      {
         Node n = methMap.getAttributes().item(i);
         
         String name  = n.getNodeName();
         String value = n.getNodeValue();
         
         ie.putAttribute(name, value);
      }
      
      ie.initialize();
      
      return ie;
   }
   
   /**
    * Build the parameters for the specified internal entry.
    * 
    * @param   node
    *          The current node from where the parameters are read.
    * @param   ie
    *          The internal entry where to save the parameters.
    */
   private static void buildParameters(Element node, InternalEntry ie)
   {
      // this enumerates all parameter nodes under the node instance (on any
      // level)
      NodeList params = node.getElementsByTagName(ELEM_PARAM);
      
      if (params == null)
      {
         return;
      }
      
      int plen = params.getLength();
      
      ArrayList<Parameter> list = new ArrayList<>();
      
      // process each method mapping
      for (int j = 0; j < plen; j++)
      {
         Element param = (Element) params.item(j);
         
         // keep only parameters which belong to the given node
         if (param.getParentNode() != node)
         {
            continue;
         }
         
         String ppname = param.getAttribute(ATTR_PNAME);
         String pjname = param.getAttribute(ATTR_JNAME);
         String ptype  = param.getAttribute(ATTR_TYPE);
         String pmode  = param.getAttribute(ATTR_PARAM_MODE);
         
         if (!("TABLE".equalsIgnoreCase(ptype) || 
               "BUFFER".equalsIgnoreCase(ptype)) &&
             (ppname == null || ppname.length() == 0 ||
              pjname == null || pjname.length() == 0))
         {
            String errmsg = "Malformed parameter entry for " +
                            param.toString() + ", in method " + 
                            node.toString();
            throw new IllegalStateException(errmsg);
         }

         Parameter p = new Parameter(ppname, pjname, ptype, pmode);

         // add the parameter's attributes
         int attrLen = param.getAttributes().getLength();
         
         for (int i = 0; i < attrLen; i++)
         {
            Node n = param.getAttributes().item(i);
            String attrName = n.getNodeName();
            String attrValue = n.getNodeValue();

            p.putAttribute(attrName, attrValue);
         }
         
         p.setExtent(p.getAttribute("extent"));

         list.add(p);
      }
      
      if (plen > 0)
      {
         list.trimToSize();
         ie.setParameterList(list);
      }
   }
   
   /**
    * Build the internal-entries 4GL attribute for the given external program.
    * 
    * @param    extProg
    *           The {@link ExternalProgram} instance.
    * @param    prv
    *           Flag to include or not the private members.
    *           
    * @return   See above.
    */
   private static String buildInternalEntries(ExternalProgram extProg, boolean prv)
   {
      // - PRIVATE members are not visible
      // - prototypes (such as IN SUPER or IN handle functions/procs) appear in this list 
      // - FORWARD functions without implementations do not appear
      // - first are procedures, then functions, then IN SUPER procedures are put last (in
      //   reverse order)

      StringBuilder sbSuperProcs = new StringBuilder();
      StringBuilder sbProcs = new StringBuilder();
      StringBuilder sbFuncs = new StringBuilder();
      List<String> superProcs = new LinkedList<>();
      
      for (InternalEntry ie : extProg.ieList)
      {
         StringBuilder sb = null;
         
         if (ie.isPrivate() && !prv)
         {
            continue;
         }

         if (ie.isFunction())
         {
            sb = sbFuncs;
         }
         else
         {
            if (ie.isSuper())
            {
               superProcs.add(0, ie.pname);
               continue;
            }

            sb = sbProcs;
         }
         
         if (sb.length() > 0)
         {
            sb.append(',');
         }
         
         sb.append(ie.pname);
      }

      for (String psuper : superProcs)
      {
         if (sbSuperProcs.length() > 0)
         {
            sbSuperProcs.append(',');
         }
         
         sbSuperProcs.append(psuper);
      }
      
      StringBuilder sb = new StringBuilder();
      
      sb.append(sbProcs);
      
      if (sb.length() > 0 && sbFuncs.length() > 0)
      {
         sb.append(',');
      }
      
      sb.append(sbFuncs);
      
      if (sb.length() > 0 && sbSuperProcs.length() > 0)
      {
         sb.append(',');
      }
      
      sb.append(sbSuperProcs);
      
      return sb.toString();
   }

   /**
    * Convert the given Progress file name into a simple relative path
    * form.  Any leading file separator is removed.  All references
    * to directories <code>&#46;</code> and <code>&#46;&#46;</code> are
    * processed.  Empty path elements are removed.
    * <p>
    * If {@link #caseSens} is false, the program name is lower-cased.
    * 
    * @param    progName
    *           Input name.
    * 
    * @return   The relative name.
    */
   private static String canonicalize(String progName)
   {
      // honor case-insensitivity
      if (!caseSens)
      {
         progName = progName.toLowerCase();
      }
      
      String[] pathTokens = progName.split(fileSep);
      
      LinkedList<String> list = new LinkedList<>();
      
      for (String token : pathTokens)
      {
         if (token.equals("") || token.equals("."))
         {
            continue;
         }
         
         if (token.equals(".."))
         {
            if (list.size() > 0)
            {
               list.removeLast();
            }
            
            continue;
         }
         
         list.addLast(token);
      }
      
      StringBuilder buf = new StringBuilder(progName.length());
      
      for (Iterator iter = list.iterator(); iter.hasNext();)
      {
         String element = (String) iter.next();
         
         buf.append(element);
         
         if (iter.hasNext())
         {
            buf.append(fileSep);
         }
      }
      
      return buf.toString();
   }

   /**
    * Construct a full class name (including packages) for the stored package
    * root and the given class name.
    * 
    * @param    name
    *           Class name.
    * 
    * @return   The fully qualified class name.
    */
   private static String addPkg(String name)
   {
      initMappingData();
      
      if (name == null)
      {
         return null;
      }
      
      if (pkgroot == null)
      {
         return name;
      }
      
      return pkgroot + "." + name; 
   }
   
   /**
    * Remove the package prefix from the given class name.
    * 
    * @param    name
    *           Class name.
    * 
    * @return   The class name with the root package prefix removed.
    */
   private static String removePkg(String name)
   {
      initMappingData();

      if (name == null)
      {
         return null;
      }
      
      if (pkgroot == null)
      {
         return name;
      }
      
      if (name.startsWith(pkgroot))
      {
         name = name.substring(pkgroot.length() + 1);
      }
      
      return name;
   }

   /**
    * Container for all external program information.
    */
   private static class ExternalProgram
   {
      /** Legacy 4GL name for this external program. */
      String pname = null;

      /** Converted Java name for this external program. */
      String jname = null;
      
      /** The named events published by this procedure. */
      String publishedEvents = "";

      /** 
       * List of all internal entries, as they appear in the name_map file.
       * This is needed to provide consistency for the internal-entries 
       * attribute (as a function and a procedure may have the same name,
       * a map can not be used).
       */
      ArrayDeque<InternalEntry> ieList = new ArrayDeque<>();
      
      /**
       * Provide fast access to the java name of a specified legacy function
       * name.
       */
      Map<String, InternalEntry> p2jf = Collections.EMPTY_MAP;

      /**
       * Provide fast access to the java name of a specified legacy procedure
       * name.
       */
      Map<String, InternalEntry> p2jp = Collections.EMPTY_MAP;

      /**
       * Provide fast access to the legacy name of a specified java function
       * name.
       */
      Map<String, InternalEntry> j2pf = Collections.EMPTY_MAP;

      /**
       * Provide fast access to the legacy name of a specified java procedure
       * name.
       */
      Map<String, InternalEntry> j2pp = Collections.EMPTY_MAP;

      /** 
       * The <code>internal-entry</code> attribute for this external program,
       * without private members.
       */
      String internalEntries = null;

      /** 
       * The <code>internal-entry</code> attribute for this external program,
       * with private members.
       */
      String internalEntriesPrv = null;
      
      /** The internal-entry associated with the MAIN (external) procedure. */
      InternalEntry main = null;
      
      /**
       * Basic c'tor.
       * 
       * @param    pname
       *           Legacy 4GL name for this external program.
       * @param    jname
       *           Converted Java name for this external program.
       * @param    publishedEvents
       *           The named events published by this procedure.
       */
      ExternalProgram(String pname, String jname, String publishedEvents) 
      {
         this.pname = pname.intern();
         this.jname = jname.intern();
         this.publishedEvents = (publishedEvents == null ? "" : publishedEvents.intern());
      }
      
      /**
       * Gets the <code>internal-entry</code> attribute for this external
       * program.  If this is the first access, it initializes it.
       * 
       * @param    prv
       *           Flag to include or not the private members.
       *           
       * @return   See above.
       */
      String getInternalEntries(boolean prv)
      {
         if (prv)
         {
            if (internalEntriesPrv == null)
            {
               internalEntriesPrv = buildInternalEntries(this, prv).intern();
            }
            
            return internalEntriesPrv;
         }
         else
         {
            if (internalEntries == null)
            {
               internalEntries = buildInternalEntries(this, prv).intern();
            }
            
            return internalEntries;
         }
      }

      /**
       * Search the {@link #ieList internal-entry list} for the specified 
       * name, using the <code>function</code> flag to search only for 
       * functions or only for procedures.
       * 
       * @param    pname
       *           Legacy 4GL name for the search internal-entry.
       * @param    function
       *           <code>true</code> if search for a function
       * 
       * @return   The {@link InternalEntry} instance, or <code>null</code> if
       *           it was not found.
       */
      InternalEntry search(String pname, boolean function)
      {
         for (InternalEntry ie : ieList)
         {
            if (function == ie.isFunction() && ie.pname.equalsIgnoreCase(pname))
            {
               return ie;
            }
         }
         
         return null;
      }

      /**
       * Search the fast-access p2j maps for a function or procedure, using 
       * the <code>function</code> flag.
       * 
       * @param    pname
       *           Legacy 4GL name for the search internal-entry.
       * @param    function
       *           <code>true</code> if search for a function
       * 
       * @return   The {@link InternalEntry} instance, or <code>null</code> if
       *           it was not found.
       */
      InternalEntry searchp(String pname, boolean function)
      {
         Map<String, InternalEntry> p2j = function ? p2jf : p2jp;
         
         return p2j.get(pname.toLowerCase());
      }

      /**
       * Search the fast-access j2p maps for a function or procedure. This
       * ignores the IN SUPER and IN HANDLE entries, as it is expected to
       * always find a method with its body in this {@link ExternalProgram}.
       * 
       * @param    jname
       *           Java method name.
       * 
       * @return   The {@link InternalEntry} instance, or <code>null</code> if
       *           it was not found.
       */
      InternalEntry searchLocalj(String jname)
      {
         // in this case, we need only internal entries which body resides
         // in this procedure. so, IN SUPER and IN HANDLE entries are ignored.
         
         InternalEntry ie = j2pf.get(jname);
         if (ie != null && !ie.isInHandle() && !ie.isSuper())
         {
            return ie;
         }
         
         ie = j2pp.get(jname);
         if (ie != null && !ie.isInHandle() && !ie.isSuper())
         {
            return ie;
         }

         return null;
      }
   }
}