/*
** Module   : StopConditionException.java
** Abstract : represents a Progress STOP condition
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20050811   @22113 Created initial version which represents a Progress STOP condition.
** 002 NVS 20060222   @24625 Constructors which take a Throwable now assume the original
**                           condition was caused by some kind of interrupt and call
**                           Control.clearInterrupted method to avoid duplicate exceptions.
** 003 GES 20060623   @27562 Support storing a key code as a cause.
** 004 GES 20150109          When investigating newly converted code it is common for simple
**                           configuration mistakes to cause missing procedures or other problems
**                           which manifest as stop conditions.  Normally such conditions are
**                           not logged and can drop the user back to a login screen. Logging
**                           has been added here to make such cases easier to diagnose.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.logging.*;

import com.goldencode.p2j.net.Control;

/**
 * Represents a Progress STOP condition.
 */
public class StopConditionException
extends ConditionException
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(StopConditionException.class.getName());
   
   /** To log or not to log. WARNING: not compatible with dynamic logging configurations. */
   private static final boolean logit = LOG.isLoggable(Level.FINEST);
   
   /**
    * Constructor which accepts an error message.
    *
    * @param   message
    *          Error message.
    */
   public StopConditionException(String message)
   {
      super(message);
      tryLog(message, this);
   }
    
   /**
    * Constructor which accepts an error message and the key code for the key
    * press that caused this condition.
    *
    * @param   message
    *          Error message.
    * @param   keycode
    *          The key code of the key that caused this condition to be
    *          raised.
    */
   public StopConditionException(String message, int keycode)
   {
      super(message, keycode);
      tryLog(message, this);
   }
    
   /**
    * Constructor which accepts an error message and a root cause. This
    * variant is used when the error condition is detected as the result of
    * catching another exception.
    *
    * @param   message
    *          Error message.
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public StopConditionException(String message, Throwable rootCause)
   {
      super(message, rootCause);
      Control.clearInterrupted();
      tryLog(message, rootCause);
   }
    
   /**
    * Constructor which accepts and wrappers a root cause. This variant is
    * used when the error condition is detected as the result of catching
    * another exception, and the message or context of the root cause is
    * sufficient to describe the problem.
    *
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public StopConditionException(Throwable rootCause)
   {
      super(rootCause);
      Control.clearInterrupted();
      tryLog("", rootCause);
   }
   
   /**
    * If logging is enabled, emit a log entry for this instance.
    *
    * @param   message
    *          Error message.
    * @param   cause
    *          Exception which has been caught in the course of processing.
    */
   private void tryLog(String message, Throwable cause)
   {
      if (logit)
      {
         LOG.logp(Level.FINEST, "StopConditionException", "", message, cause);
      }
   }
}