/*
** Module   : TransparentWrapper.java
** Abstract : simple container that implements WrappedResource for any object
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description---------------------------
** 001 GES 20070221   @32199 Created initial version which represents a
**                           simple container that implements
**                           WrappedResource for any object.
** 002 CA  20090227   @41382 Added WrappedResources.isUnknown method; this
**                           implementation returns false, if referent is
**                           not an instance of WrappedResource; else,
**                           it returns referent.isUnknown() value.
** 003 CA  20130112          Made the referent field protected. Fixed equals.
** 004 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 005 CA  20150502          The resource ID needs to be kept at the resource, so it can be 
**                           accessed after the resource is deleted.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Simple container that stores an object instance and implements the
 * {@link WrappedResource} interface. Since {@link #hashCode} and 
 * {@link #equals} are redirected to the contained reference, this is as
 * suitable for use in a collection as the contained reference.
 */
public class TransparentWrapper
implements WrappedResource
{
   /** The object being contained. */
   protected Object referent = null;
   
   /**
    * The resource's ID.  Remains <code>null</code> until the ID is requested via 
    * {@link handle#resourceId}
    */
   private Long id = null;
   
   /**
    * Construct an instance given the object reference to contain.
    *
    * @param    referent
    *           The object to contain.
    */
   public TransparentWrapper(Object referent)
   {
      this.referent = referent;
   }
   
   /**
    * Get this resource's ID, if is already set.
    * 
    * @return   The resource's ID or <code>null</code> if not set.
    */
   @Override
   public Long id()
   {
      return id;
   }
   
   /**
    * Set this resource's ID.
    *  
    * @param    id
    *           The resource's ID.
    */
   @Override
   public void id(long id)
   {
      this.id = id;
   }

   /**
    * Returns the referent of this instance as an <code>Object</code>.
    *
    * @return   The value as an <code>Object</code> or <code>null</code> if
    *           there is no contained object.
    */
   public Object get()
   {
      return referent;
   }   
   
   /**
    * Set the referent of this instance.
    *
    * @param    referent
    *           The value to be represented in this instance.
    */
   public void set(Object referent)
   {
      this.referent = referent;
   }
   
   /**
    * Reports if this object is valid for use.  If the contained object  
    * reference itself implements {@link WrappedResource}, then this
    * call will be redirected to the contained object.  Otherwise, if
    * the contained object is <code>null</code>, then it is considered
    * invalid and not <code>null</code> is valid.
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      if (referent != null && referent instanceof WrappedResource)
      {
         // pass through the validity check
         WrappedResource res = (WrappedResource) referent; 
         return res.valid();
      }
      
      // otherwise non-null means valid
      return (referent != null); 
   }
   
   /**
    * Reports if this object is unknown.<p>
    * If {@link #referent} is not an instance of {@link WrappedResource}, it
    * returns false. Else, it returns {@link WrappedResource#unknown()} 
    * result.
    *
    * @return   <code>true</code> if object is unknown.
    */
   public boolean unknown()
   {
      if (referent != null && referent instanceof WrappedResource)
      {
         WrappedResource res = (WrappedResource) referent; 
         return res.unknown();
      }

      return false;
   }

   /**
    * Hash code implementation which is redirected to that of the 
    * contained object so long as that object is not <code>null</code>.
    *
    * @return   Hash code value for this object instance.
    */
   public int hashCode()
   {  
      return referent == null ? ((Object) this).hashCode() 
                              : referent.hashCode();
   }
   
   /**
    * Determines if this instance and the passed-in instance are equivalent.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   <code>true</code> if the objects are the same.
    */
   public boolean equals(Object obj)   
   {  
      if (obj != null && obj instanceof TransparentWrapper)
      {
         obj = ((TransparentWrapper) obj).referent;
      }

      return (referent == null) ? ((Object) this).equals(obj)
                                : referent.equals(obj);
   }
}