/*
** Module   : TriggerDefinition.java
** Abstract : container for trigger metadata that allows a new instance to
**            be manufactured
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20081217   @40928 Created initial version as the container for trigger metadata that 
**                           allows a new instance to be manufactured.
** 002 CA  20140218          Expose trigger's external program instance.
** 003 CA  20151030          Report if the trigger definition was created in an external program.
** 004 GES 20161004          Enabled an alternate path that uses a single pre-instantiated
**                           Trigger instance (using lambdas).  Inner class mode is still
**                           supported and is needed when there are local variables defined for
**                           the trigger.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.logging.*;
import static com.goldencode.p2j.util.BlockManager.*;

/**
 * Container for trigger metadata that allows a new instance to be
 * manufactured.
 * <p>
 * Triggers are non-static inner classes that may have data members. The
 * initialization of those data members naturally occurs during construction.
 * This means that only after construction can one be assured that the value
 * of a trigger's data members is set to the expected default value. This
 * means that every time a trigger is executed, that execution must occur on
 * a newly constructed instance, so that the data member's default values
 * will be correct.
 * <p>
 * This class stores the data needed to delegate the construction of a
 * {@link Trigger} subclass to code that is executing that trigger. This is
 * complicated by the fact that triggers are implemented as non-static inner
 * classes of business logic classes. To instantiate a non-static inner class
 * it is required that the containing instance be available as an implicit
 * first parameter to the constructor.
 * <p>
 * The complexity of the instantiation is hidden in this class. The logic
 * can be found in {@link #getTrigger}.
 */
public class TriggerDefinition
{
   /** Logger (this is JVM-wide rather than being context-local). */
   private static final Logger LOG = 
      LogHelper.getLogger(TransactionManager.class.getName());
   
   /** Common case where there is a single trigger instance. */
   private Trigger trigger = null;
   
   /** Specific subclass of {@link Trigger} which implements the trigger. */
   private Class<?> cls = null;
   
   /** Containing business logic instance (triggers are inner classes). */
   private Object outer = null;
   
   /** The external program instance where the trigger was defined. */
   private Object externalProgram = null;
   
   /** Determines if this trigger should start a full transaction. */
   private boolean fullTrans = false;
   
   /** Flag indicating the trigger definition was created in an external program. */
   private final boolean fromExternalProgram;
   
   /**
    * Constructor for inner class mode.
    *
    * @param    cls
    *           The specific {@link Trigger} subclass that implements the
    *           trigger logic.
    * @param    outer
    *           The instance of the containing class for the trigger. This
    *           is needed since all triggers are implemented as non-static
    *           inner classes.
    * @param    fullTrans
    *           <code>true</code> if the trigger should start a full
    *           transaction, <code>false</code> if the trigger should
    *           create a sub-transaction.
    */
   public TriggerDefinition(Class<?> cls, Object outer, boolean fullTrans)
   {
      this.cls       = cls;
      this.outer     = outer;
      this.fullTrans = fullTrans;
      this.externalProgram = ProcedureManager.findRootEnclosingInstance(outer);
      this.fromExternalProgram = TransactionManager.isExternalBlock();
      
      if (externalProgram == null)
      {
         externalProgram = outer;
      }
   }
   
   /**
    * Constructor for lambda mode.
    *
    * @param    trigger
    *           The specific {@link Trigger} instance that implements the
    *           trigger logic.
    * @param    outer
    *           The instance of the containing class for the trigger.
    * @param    fullTrans
    *           <code>true</code> if the trigger should start a full
    *           transaction, <code>false</code> if the trigger should
    *           create a sub-transaction.
    */
   public TriggerDefinition(Trigger trigger, Object outer, boolean fullTrans)
   {
      this.trigger   = trigger;
      this.fullTrans = fullTrans;
      this.externalProgram = outer;
      this.fromExternalProgram = TransactionManager.isExternalBlock();
      
      if (fullTrans)
      {
         trigger.setLevel(TransactionType.FULL);
      }
   }
   
   /**
    * Construct a new instance of the trigger class that is defined by the
    * contained metadata.
    *
    * @return   The newly instantiated trigger instance. If any failure
    *           occurs, this will return <code>null</code>.
    */
   public Trigger getTrigger()
   {
      Trigger instance = trigger;
      
      if (instance == null)
      {
         // inner class mode
         try
         {
            Class<?>[]     parms  = new Class[] { outer.getClass() };
            Constructor<?> constr = cls.getConstructor(parms);
            
            instance = (Trigger) constr.newInstance(new Object[] { outer });
            
            if (fullTrans)
            {
               instance.setLevel(TransactionType.FULL);
            }
         }
         
         catch (Exception excp)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               String err = 
                  LogHelper.generate("Cannot create a new Trigger instance.");
                  
               LOG.logp(Level.WARNING, "getTrigger", "", err, excp);
            }
         }
      }
      
      return instance;
   }
   
   /**
    * Get the external program instance where this trigger was defined.
    * 
    * @return    See above.
    */
   public Object getExternalProgram()
   {
      return externalProgram;
   }
   
   /**
    * Check if this trigger was created by an external program.
    * 
    * @return   The state of the {@link #fromExternalProgram} flag.
    */
   public boolean isFromExternalProgram()
   {
      return fromExternalProgram;
   }
}