/*
** Module   : rowid.java
** Abstract : Progress 4GL compatible rowid object
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20080721   @39290 Created initial version. Represents the
**                           Progress rowid data type. Much of the
**                           implementation was borrowed from the handle
**                           type.
** 002 ECF 20090212   @41302 Implemented a faster equals(). Handles most
**                           common cases without having to invoke the
**                           superclass' implementation.
** 003 CA  20130123          Added c'tor which builds a new instance from a
**                           BaseDataType value, if possible. This should be
**                           used only when constructing an instance using the
**                           value returned by a DYNAMIC-FUNCTION call; 4GL
**                           does some special casting in this case, which is
**                           not implemented yet.
** 004 OM  20121023          Added/fixed support for TO-ROWID(STRING(ROWID)).
** 005 CS  20130315          Made parseRowidString public.
** 006 GES 20130322          Fixed the BDT c'tor and modified the assign(BDT,boolean) to be the
**                           common location for the _POLY morphing. This is safe because all
**                           current usage would have generated a ClassCastException if that
**                           method was ever used with an invalid type. Now, instead of the CCE,
**                           the code will morph the value as is possible. The morphing code
**                           still needs to be implemented.
** 007 OM  20130416          Cosmetic change: removed '.' at the end of error message.
** 008 MAG 20140602          Fixed the BDT c'tor and modified the assign(BDT,boolean) on _POLY morphing.
**                           Add BDT types on compareTo(Object).
** 009 ECF 20151209          Removed bitwise parity check and converted characters to lowercase
**                           when parsing rowid strings.
** 010 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 011 GES 20160617          Added unknown literal processing for the polymorphic assign case.
** 012 CA  20160627          Reworked undoable support - the undoables register themselves with
**                           all blocks up the stack, until either 1. the tx block which created
**                           it is reached or 2. the last tx block where it was saved is reached.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;

/**
 * A class that represents a Progress 4GL compatible rowid data type.
 * Internally, the row identifier is stored as a 64-bit integer.
 */
public final class rowid
extends BaseDataType
{
   /** Stores the integral row identifier, unique within a database */
   private Long value = null;
   
   /**
    * Worker method that parses and validates a <code>java.lang.String</code>
    * according to rowid Progress data type constraints:
    * <ul>
    *    <li>must start with "0x" prefix
    *    <li>must only contain lower-case hexadecimal digits
    *    <li>the length of string must be even (not mandatory 16 digits)
    * </ul>
    * <p>
    * Since this implementation of <code>rowid</code> is not compatible with the
    * internal representations of Progress <code>rowid</code> values, the only one
    * of the above constraints which we enforce here is the "0x" prefix.
    * <p>
    * The actual internal implementation warps around values above
    * <code>Long.MAX_VALUE</code> and uses the negative interval of java long
    * type to store using all 64 bit available. Because of this rowids should
    * not be compared using &lt;, &gt;, &le; or &ge;.
    * <p>
    * The method will throw <code>ErrorConditionException</code> if parsing
    * fails because the string exceeds the 64 bit bound of <code>Long</code>
    * data type (ie more than 16 hexadecimal digits).
    * <p>
    * Java language supports only signed primitive data type, so the
    * parsing utilities (<code>Long.parseLong()</code>) are implemented that
    * way. Special code is needed to parse string values larger than
    * <code>Long.MAX_VALUE</code> and to store within negative interval.
    *
    * @param   val
    *          A hex string in the format 0xhhhhhhhhhhhhhhhh.
    *          Any other string will, including <code>null</code> will
    *          indicate the unknown value.
    *
    * @return  A <code>Long</code> representing the internal value for a
    *          rowid or <code>null</code> if parsing fails.
    */
   public static Long parseRowidString(String val)
   {
      // trivial case:
      if (val == null)
      {
         return null;
      }
      
      // check for 0x prefix
      if (!val.startsWith("0x"))
      {
         return null;
      }
      
      // remove 0x prefix:
      val = val.substring(2);
      
      // length of hex digits only:
      int len = val.length();
      
      // must have at least 1 digit (2 digits with parity)
      if (len < 2)
      {
         return null;
      }
      
      // java long overflow check:
      if (len > 16)
      {
         ErrorManager.recordOrThrowError(-1, "P2J rowid overflow");
      }
      
      // check hex digits and lower-case at once:
      for (int i = 0; i < len; i++)
      {
         char ch = Character.toLowerCase(val.charAt(i));
         
         // ASCII char map: ... 0 *** 9 ... a *** f ...
         // we are only interested in the stared sub-intervals
         // chars in the outer interval are illegal hex-digits:
         if (ch < '0' || ch > 'f')
         {
            return null;
         }
         
         // chars in the inner interval are also illegal hex-digits:
         if (ch > '9' && ch < 'a')
         {
            return null;
         }
      }
      
      // now check for a very special case: 1st bit of a 16-digit case
      // treat the MSB as the sign of java long
      boolean isNegative = (len == 16 && val.charAt(0) > '7');
      
      // clear the 'sign' bit if is negative:
      if (isNegative)
      {
         int firstDigit = Character.digit(val.charAt(0), 16);
         firstDigit -= 0x08;
         // replace the first digit with the newly computed value:
         val = Character.forDigit(firstDigit, 16) + val.substring(1);
      }
      
      // assume all in upper case, it does not matter here in fact if
      // there are lower-case digits, the conversion should be successful
      try
      {
         long parsedValue =  Long.parseLong(val, 16);
         
         // returning with respect to the sign computed above:
         return isNegative ? (Long.MIN_VALUE + parsedValue) : parsedValue;
      }
      catch (NumberFormatException nfe) // invalid hex number ?
      {
         // after all previous checks this shouldn't be reachable
         return null;
      }
   }
   
   /**
    * Default constructor, creates an instance that represents the unknown
    * value.
    */
   public rowid()
   {
   }
   
   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType,boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public rowid(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         assign(value, false);
      }
   }
   
   /**
    * Constructs an instance based upon an existing instance.
    *
    * @param   value
    *          The instance to copy.
    */
   public rowid(rowid value)
   {
      assign(value, false);
   }
   
   /**
    * Constructs an instance given a 64-bit integer value.
    * 
    * @param   value
    *          64-bit integer value;  should be unique across all records in a
    *          database.  May be <code>null</code> to indicate unknown value.
    */
   public rowid(Long value)
   {
      this.value = value;
   }
   
   /**
    * Constructs an instance given a character encoding (hex string in the
    * format 0xhhhhhhhhhhhhhhhh, where each h is an lower-case, hexadecimal 
    * digit, limited to maximum 16 digits because that actual internal 
    * implementation uses a java 64-bit Long).
    *
    * @param   val
    *          A hex string in the format 0xhhhhhhhhhhhhhhhh.
    *          Any other string will, including <code>null</code> will
    *          indicate the unknown value.
    */
   public rowid(character val)
   {
      // trivial cases:
      if (val == null || val.isUnknown())
      {
         this.value = null;
         return;
      }
      
      this.value = parseRowidString(val.toJavaType());
   }
   
   /**
    * Constructs an instance given a character encoding (hex string in the
    * format 0xhhhhhhhhhhhhhhhh, where each h is an lower-case, hexadecimal 
    * digit, limited to maximum 16 digits because that actual internal 
    * implementation uses a java 64-bit Long).
    *
    * @param   val
    *          A hex string in the format 0xhhhhhhhhhhhhhhhh.
    *          Any other string will, including <code>null</code> will
    *          indicate the unknown value.
    */
   public rowid(String val)
   {
      this.value = parseRowidString(val);
   }
   
   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return  An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()
   {
      return new rowid();
   }
   
   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return  A clone of this instance.
    */
   public BaseDataType duplicate()
   {
      return (new rowid(this));
   }
   
   /**
    * An equality test which handles the most common cases with the least
    * amount of overhead.  The superclass implementation is invoked otherwise.
    * <p>
    * This implementation is consistent with {@link #hashCode()}.
    * 
    * @see com.goldencode.p2j.util.BaseDataType#equals(java.lang.Object)
    */
   public boolean equals(Object o)
   {
      if (this == o)
      {
         return true;
      }
      
      if (o instanceof rowid)
      {
         rowid that = (rowid) o;
         
         boolean u1 = (this.value == null);
         boolean u2 = (that.value == null);
         
         if (u1 && u2)
         {
            return true;
         }
         
         if (u1 || u2)
         {
            return false;
         }
         
         return (this.value.equals(that.value));
      }
      
      return super.equals(o);
   }
   
   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance. All instances of
    *          this class return the same hash code value, since all
    *          instances are inherently equal.
    */
   public int hashCode()
   {
      int result = 17;
      
      if (value != null)
      {
         result = 37 * result + value.hashCode();
      }
      
      return result;
   }
   
   /**
    * Returns the value of this instance as an <code>Long</code>.
    * <p>
    * If this instance represents the <code>unknown value</code>, a
    * <code>null</code> will be returned.
    *
    * @return  The value as a <code>Long</code> or <code>null</code> if
    *          this is <code>unknown</code>.
    */
   public Long getValue()
   {
      return value;
   }
   
   /**
    * Determines if this instance represents the <code>unknown value</code>.
    *
    * @return  <code>true</code> if this instance is set to the
    *          <code>unknown value</code>.
    */
   public boolean isUnknown()
   {
      return (value == null);
   }
   
   /**
    * Sets this instance to represent the unknown value.
    * <p>
    * Any existing value is discarded.
    */
   public void setUnknown()
   {
      if (value != null)
      {
         checkUndoable(true);
      }
      
      value = null;
   }
   
   /**
    * Sets the data of this instance based on the given instance.
    * <p>
    * If the value is not of type <code>rowid</code>, the following automatic type
    * conversion will occur:
    * <p>
    * <pre>
    * This type support only rowid and unknown types (including unknown literal).
    * </pre>
    * <p>
    * This variant is meant to handle the cases of built-in functions and methods in the 4GL
    * which have polymorphic return types (e.g. DYNAMIC-FUNCTION()). This should NOT be used for
    * non-polymorphic assignments.
    *
    * @param   value
    *          The instance from which to copy state.
    * @param   force
    *          Flag to indicate if a pending error should be ignored (see 
    *          {@link ErrorManager#isPendingError} and the assignment
    *          forced regardless.  Note that normal assignment silently
    *          exits with no change if there is a pending error. If this
    *          flag is <code>true</code>, that check is bypassed and the
    *          assignment will always be done.
    * 
    * @throws  ClassCastException
    *          if <code>value</code> is not an instance of this class.
    */
   public void assign(BaseDataType value, boolean force)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else if (value instanceof rowid)
      {
         assign((rowid) value, force);
      }
      else
      {
         // conversion impossible
         incompatibleTypesOnConversion();
      }
   }
   
   /**
    * Sets the data of this instance based on the given instance.
    *
    * @param   value
    *          The instance from which to copy state.
    * @param   force
    *          Flag to indicate if a pending error should be ignored (see 
    *          {@link ErrorManager#isPendingError} and the assignment
    *          forced regardless.  Note that normal assignment silently
    *          exits with no change if there is a pending error. If this
    *          flag is <code>true</code>, that check is bypassed and the
    *          assignment will always be done.
    */
   public void assign(rowid value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         checkUndoable(value);

         this.value = value.value;
      }
   }
   
   /**
    * Sets the data of this instance based on the given instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param   value
    *          The instance from which to copy state.
    */
   public void assign(rowid value)
   {
      assign(value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param   value
    *          The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((rowid) value, false);
   }
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The rowids' internal, 64-bit integer values are compared.
    *
    * @param   obj
    *          The instance to compare against.
    *
    * @return   0 or 1 depending on if the passed instance is or is not
    *           <code>unknown</code>.
    */
   public int compareTo(Object obj)
   {
      rowid that = null;
      
      if (obj instanceof rowid)
      {
         that = (rowid) obj;
      }
      else if (obj instanceof BaseDataType)
      {
         that = new rowid((BaseDataType) obj);
      }
      else
      {
         return -1;
      }
      
      Long l1 = this.value;
      Long l2 = that.value;
      
      if (l1 == l2)
      {
         return 0;
      }
      else if (l1 == null)
      {
         return 1;
      }
      else if (l2 == null)
      {
         return -1;
      }
      
      return l1.compareTo(l2);
   }
   
   /**
    * Creates a string representation of the instance data as a hex string
    * with a leading "0x" prefix.
    * <p>
    * If the instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return  The formatted string.
    */
   public String toString()
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data as a hex string
    * with a leading "0x" prefix.
    * <p>
    * If the instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @param   fmt
    *          The Progress 4GL format string or <code>null</code> if the
    *          default format is to be used.  This parameter is ignored.
    *
    * @return  The formatted string.
    */
   public String toString(String fmt)
   {
      if (isUnknown())
      {
         return "?";
      }
      
      StringBuilder buf = new StringBuilder("0x");
      buf.append(Long.toHexString(value).toUpperCase());
      
      return buf.toString();
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return  The 'message' formatted string.
    */
   public String toStringMessage()
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  If the instance represents the <code>unknown value</code>, a
    * '?' will be returned.
    *
    * @return  The 'export' formatted string.
    */
   public String toStringExport()
   {
      return toString(null);
   }
   
   /**
    * Return the default display format string for this type.  This
    * implementation returns <code>null</code>, because there is no formatted
    * representation of a rowid:  it must be converted to a string first using
    * the Progress STRING() function.
    *
    * @return  <code>null</code>.
    */
   public String defaultFormatString()
   {
      return null;
   }
   
   /**
    * Calculate the length of a formatted value (the text form of the value)
    * using this format string.
    *
    * @param    fmt
    *           The format string.
    *
    * @return   The length in characters of the any resulting formatted
    *           value.
    */
   public int formatLength(String fmt)
   {
      return 18;  // an 8-byte hex string + leading '0x' (0x0123456789ABCDEF)
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param   in
    *          The input source from which fields will be restored.
    *
    * @throws  IOException
    *          in case of I/O errors.
    * @throws  ClassNotFoundException
    *          if payload can't be instantiated.
    * @throws  NotSerializableException
    *          always.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      throw new NotSerializableException("rowid");
   }
   
   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param   out
    *          The output destination to which fields will be saved.
    *
    * @throws  IOException
    *          in case of I/O errors.
    * @throws  NotSerializableException
    *          always.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      throw new NotSerializableException("rowid");
   }
}