/*
** Module   : XmlTokenTypes.java
** Abstract : Token type constants used wit XML-specific ASTs.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20050421 ADD   @20840 Created initial version.
** 002 EVL 20160224              Javadoc fixes to make compatible with Oracle Java 8 for
**                               Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.xml;

import org.w3c.dom.*;

/**
 * Token types necessary to represent an XML Document Object Model in AST
 * form. All constants whose names end in "<code>_NODE</code>" are based
 * on the node types defined in <code>org.w3c.dom.Node</code>. They are
 * redefined here to provide a unified interface for all XML AST token types,
 * but their names and values are the same as in the <code>Node</code>
 * interface (except that they are promoted to <code>int</code>s). Several
 * additional types are necessary to represent some DOM node data in AST
 * form, as noted in the following table:
 * <table border="1" cellpadding="2" cellspacing="0" summary="">
 * <tr>
 * <th><strong>TOKEN TYPE</strong></th>
 * <th><strong>PURPOSE</strong></th>
 * </tr>
 * <tr>
 * <td>PUBLIC_ID</td>
 * <td>DTD public identifier. This is not a node in DOM, so there is no
 *     node type defined for it.</td>
 * </tr>
 * <tr>
 * <td>SYSTEM_ID</td>
 * <td>DTD system identifier. This is not a node in DOM, so there is no
 *     node type defined for it.</td>
 * </tr>
 * <tr>
 * <td>INTERNAL_SUBSET</td>
 * <td>DTD internal subset. This is not a node in DOM, so there is no
 *     node type defined for it.</td>
 * </tr>
 * <tr>
 * <td>ATTR_SET</td>
 * <td>Artificial AST to hold an element's attributes</td>
 * </tr>
 * <tr>
 * <td>CONTENT</td>
 * <td>Holds the value text of attribute, CDATA section, comment, processing
 *     instruction, and text nodes. Will always be a child of one of these
 *     node types. This extra AST level is necessary because DOM nodes
 *     have two text slots (one for name, one for value), while ASTs have
 *     only one.</td>
 * </tr>
 * </table>
 * <p>
 * These extra constants are defined above the range of Node types, to
 * prevent collisions.
 */
public interface XmlTokenTypes
{
   /** Element attribute */
   public static final int ATTRIBUTE_NODE = (0xFFFF & Node.ATTRIBUTE_NODE);
   
   /** CDATA section node */
   public static final int CDATA_SECTION_NODE =
      (0xFFFF & Node.CDATA_SECTION_NODE);
   
   /** Comment node */
   public static final int COMMENT_NODE = (0xFFFF & Node.COMMENT_NODE);
   
   /** Document fragment node */
   public static final int DOCUMENT_FRAGMENT_NODE =
      (0xFFFF & Node.DOCUMENT_FRAGMENT_NODE);
   
   /** Document "container" */
   public static final int DOCUMENT_NODE = (0xFFFF & Node.DOCUMENT_NODE);
   
   /** Document type declaration section */
   public static final int DOCUMENT_TYPE_NODE =
      (0xFFFF & Node.DOCUMENT_TYPE_NODE);
   
   /** Element node */
   public static final int ELEMENT_NODE = (0xFFFF & Node.ELEMENT_NODE);
   
   /** Entity node */
   public static final int ENTITY_NODE = (0xFFFF & Node.ENTITY_NODE);
   
   /** Entity reference node */
   public static final int ENTITY_REFERENCE_NODE =
      (0xFFFF & Node.ENTITY_REFERENCE_NODE);
   
   /** Notation node */
   public static final int NOTATION_NODE = (0xFFFF & Node.NOTATION_NODE);
   
   /** Processing instruction node */
   public static final int PROCESSING_INSTRUCTION_NODE =
      (0xFFFF & Node.PROCESSING_INSTRUCTION_NODE);
   
   /** Text node */
   public static final int TEXT_NODE = (0xFFFF & Node.TEXT_NODE);
   
   /** Public identifier within DTD section */
   public static final int PUBLIC_ID = Short.MAX_VALUE + 1;
   
   /** System identifier within DTD section */
   public static final int SYSTEM_ID = Short.MAX_VALUE + 2;
   
   /** Internal subset within DTD section */
   public static final int INTERNAL_SUBSET = Short.MAX_VALUE + 3;
   
   /** Bundle of attributes for an element */
   public static final int ATTR_SET = Short.MAX_VALUE + 4;
   
   /** Content of node's value string */
   public static final int CONTENT = Short.MAX_VALUE + 5;
}