/*
** Module   : VT100AbstractRenderer.java
** Abstract : Generic VT-100 renderer implementation.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 MAG 20140318 First version.
** 002 MAG 20140507 Make auto wrap default true.
** 003 GES 20150414 Reduced direct member usage in subclasses by adding a constructor.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.terminal;

/**
 * Implements a generic abstract VT-100 renderer.
 */
public abstract class VT100AbstractRenderer<T> 
implements VT100Renderer
{
   /** Number of rows in the screen. */
   protected int rows = 24;

   /** Number of columns in the screen. */
   protected int cols = 80;
   
   /** Scroll window, starting position. */
   protected int begScroll = 0;

   /** Scroll window, ending position. */
   protected int endScroll = 23;
   
   /** Cursor row position. */
   protected int row = 0;

   /** Cursor column position. */
   protected int col = 0;

   /** Screen contents. */
   protected T[][] data = null;
   
   /** Auto wrap mode */
   private boolean autoWrap = true;
   
   /**
    * Constructor.
    * 
    * @param    data
    *           Video buffer.
    * @param    rows
    *           Number of rows on the screen.
    * @param    cols
    *           Number of columns on the screen.
    * @param    begScroll
    *           Scroll window, starting position.
    * @param    endScroll
    *           Scroll window, ending position.
    */
   public VT100AbstractRenderer(T[][]data, int rows, int cols, int begScroll, int endScroll)   
   {
      this.data      = data;
      this.rows      = rows;
      this.cols      = cols;
      this.begScroll = begScroll;
      this.endScroll = endScroll;
   }
   
   /**
    * Set cursor position.
    *  
    * @param   row
    *          Line number.
    * @param   col
    *          Column number.
    */   
   @Override
   public void setCursorPos(int row, int col)
   {
      synchronized (this)
      {
         this.row = (row == 0) ? 0 : row - 1;
         this.col = (col == 0) ? 0 : col - 1;
         
         // fix row position
         if (row < begScroll)
         {
            row = begScroll;
         }
         else if (row > endScroll)
         {
            row = endScroll;
         }      
         
         updateScreen();
      }
   }

   /**
    * Move cursor position with the amount of rows and columns.
    *  
    * @param   nRows
    *          Number of rows.
    * @param   nCols
    *          Number of columns.
    */   
   @Override
   public void moveCursorPos(int nRows, int nCols)
   {
      synchronized (this)
      {
         // row
         row += nRows;
         if (row < begScroll)
         {
            row = begScroll;
         } else if (row > endScroll)
         {
            row = endScroll;
         }
   
         // column
         col += nCols;
         if (col < 0)
         {
            col = 0;
         } else if (col >= cols)
         {
            col = cols - 1;
         }
         
         updateScreen();
      }
   }


   /**
    * Clear lines on screen.
    * 
    * @param   mode
    *          An integer value representing a screen region:
    *          0 - Erase from the active position to the end of the screen, inclusive (default)
    *          1 - Erase from start of the screen to the active position, inclusive
    *          2 - Erase all of the display -- all lines are erased, changed to single-width, 
    *              and the cursor does not move.
    */
   @Override
   public void eraseInDisplay(int mode)
   {
      synchronized (this)
      {
         switch (mode)
         {
         case 0:
            eraseInLine(0);
            clearScreen(row + 1, rows - 1);
            break;
         case 1:
            clearScreen(0, row - 1);
            eraseInLine(1);
            break;
         case 2:
            clearScreen(0, rows - 1);
            break;
         }
         
         updateScreen();
      }
   }
   
   /**
    * Clear characters in current line.
    * 
    * @param   mode
    *          An integer value representing a line region:
    *          0 - Erase from the active position to the end of the line, inclusive (default)
    *          1 - Erase from the start of the line to the active position, inclusive
    *          2 - Erase all of the line, inclusive
    */
   @Override
   public  void eraseInLine(int mode)
   {
      synchronized (this)
      {
         switch (mode)
         {
         case 0:
            clearRow(row, col, cols - 1);
            break;
         case 1:
            clearRow(row, 0, col);
            break;
         case 2:
            clearRow(row, 0, cols -1);
            break;
         }
         
         updateScreen();
      }
   }

   /**
    * Set cell text attributes and color.
    * 
    * @param   text
    *          Character that should be displayed.
    * @param   attribute
    *          Cell attributes.
    * @param   color
    *          Call color FG/BG
    */   
   @Override
   public void setCellText(char text, int attribute, int color)
   {
      synchronized (this)
      {         
         // Set cell properties
         data[row][col] = makeCell(text, attribute, color);
         moveCursorNext();         

         updateScreen();
      }
   }
   
   /**
    * Play a message beep.
    */
   @Override
   public abstract void beep();

   /**
    * Set cursor at column.
    * 
    * @param   column
    *          Column number.
    */   
   @Override
   public void setCursorCol(int column)
   {
      synchronized (this)
      {
         if (column >= 0 && column < cols)
         {
            col = column;
         }         

         updateScreen();
      }
   }

   /**
    * NEL -- Next Line ESC E
    * 
    * This sequence causes the active position to move to the first position on the next line
    * downward. If the active position is at the bottom margin, a scroll up is performed.
    */
   @Override
   public void nextLine()
   {
      synchronized (this)
      {
         col = 0;
         index();
         
         updateScreen();
      }
   }

   /**
    * IND -- Index ESC D
    * 
    * This sequence causes the active position to move downward one line without changing the 
    * column position. If the active position is at the bottom margin, a scroll up is performed.    
    */
   @Override
   public void index()
   {
      synchronized (this)
      {
         if (++row > endScroll)
         {
            row = endScroll;
            scrollUp();
            clearRow(row, 0, cols - 1);
         }

         updateScreen();
      }
   }

   /**
    * RI -- Reverse Index ESC M
    * 
    * Move the active position to the same horizontal position on the preceding line.
    * If the active position is at the top margin, a scroll down is performed.
    */   
   @Override
   public void reverseIndex()
   {
      synchronized (this)
      {
         if (--row < begScroll)
         {
            row = begScroll;
            scrollDown();
            clearRow(row, 0, cols - 1);            
         }
         
         updateScreen();
      }
   }
   
   /**
    * Set scroll window.
    * 
    * @param   begLine
    *          Start line.
    * @param   endLine
    *          End line.
    */   
   @Override
   public void setScrollRegion(int begLine, int endLine)
   {
      begScroll = (begLine == 0) ? 0 : begLine - 1;
      endScroll = (endLine == 0) ? rows - 1 : endLine - 1;
   }

   /**
    * Set cursor visibility.
    * 
    * @param   showCursor
    * 
    */
   @Override
   public abstract void setCursorVisible(boolean showCursor);

   /**
    * Set auto wrap mode.
    * 
    * @param   autoWrap
    *          
    */   
   @Override
   public void setAutoWrap(boolean autoWrap)
   {
      this.autoWrap = autoWrap;      
   }
   
   /**
    * Get cursor current row position.
    *  
    * @return  row number
    */
   @Override
   public int getCursorRow()
   {
      return row;
   }
   
   /**
    * Get cursor current column position.
    * 
    * @return  column number
    */
   @Override
   public int getCursorCol()
   {
      return col;
   }
   
   /**
    * Get number of columns.
    * 
    * @return  Number of columns
    */
   @Override
   public int getColumns()
   {
      return cols;
   }   
   
   /**
    * Get attribute mask.
    * 
    * @param   id
    *          Attribute id.
    * <p>
    *          0 - NORMAL 
    *          1 - BLINK 
    *          2 - BOLD 
    *          3 - UNDERLINE 
    *          4 - REVERSE
    *          
    * @return  attribute mask.
    */
   public abstract int attributeMask(int id);   
      
   /**
    * Create an empty cell.
    * 
    * @return  An empty video memory cell.
    */
   protected abstract T emptyCell();
   
   /**
    * Build a new video memory cell.
    * 
    * @param   text
    *          Cell character.
    * @param   attribute
    *          Color attributes
    * @param   color
    *          Color pair FG/BG
    * 
    * @return  A new video memory cell.
    */
   protected abstract T makeCell(char text, int attribute, int color);
   
   /**
    * Update screen.
    */
   protected abstract void updateScreen();

   /**
    * Move cursor to the next position.
    */
   private void moveCursorNext()
   {
      // column
      if (++col >= cols)
      {
         if (autoWrap)
         {
            nextLine();
         }
         else
         {
            col = cols - 1;
         }
      }
   }

   /**
    * 
    * @param   rowIndex
    *          Row number.
    * @param   begCol
    *          Start column position.
    * @param   endCol
    *          End column position.
    */
   private void clearRow(int rowIndex, int begCol, int endCol)
   {
      for (int j = begCol;j <= endCol;j++)
      {
         data[rowIndex][j] = emptyCell();         
      }      
   }
   
   /**
    * Scroll UP one line in scroll window.
    */
   private void scrollUp()
   {
      for (int i = begScroll + 1;i <= endScroll;i++)
      {
         for (int j = 0;j < cols;j++)
         {
            data[i - 1][j] = data[i][j];
         }
      }      
   }

   /**
    * Scroll DOWN one line in scroll window.
    */
   private void scrollDown()
   {
      for (int i = endScroll - 1;i >= begScroll;i--)
      {
         for (int j = 0;j < cols;j++)
         {
            data[i + 1][j] = data[i][j];
         }
      }      
   }

   /**
    * Clear screen.
    *  
    * @param   begRow
    *          Start row number.
    * @param   endRow
    *          End row number.
    */
   private void clearScreen(int begRow, int endRow)
   {
      for (int i = begRow;i <= endRow;i++)
      {
         clearRow(i, 0, cols -1);
      }  
   }   
}