/*
** Module   : Engine.java
** Abstract : drives the multi-threaded processing of a list of trees 
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20090525   @42534 Drives the multi-threaded processing of a list of
**                           trees.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.trpl; 
 
import java.util.*;
import java.util.concurrent.*;
import com.goldencode.ast.*;

/**
 * Drives the multi-threaded processing of a list of trees.
 */
public class Engine
{
   /** Provides core AST services. */
   private final AstManager mgr;
   
   /**
    * Construct an instance.
    *
    * @param    mgr
    *           Provides core AST services.
    */
   public Engine(AstManager mgr)
   {
      if (mgr == null)
         throw new NullPointerException("AstManager cannot be null!");
      
      this.mgr = mgr;
   }
   
   /**
    * Iterate through the list of tree names to be processed and for each one,
    * traverse the tree and invoke the hooks of each tree walker in the
    * pipeline.
    *
    * @param    asts
    *           List of tree names. Each one will be loaded in turn and
    *           optionally saved at the end (depending on the pipeline
    *           configuration).
    * @param    pipeline
    *           The tree walkers to be invoked and other configuration values.
    */
   public void process(List<String> asts, Pipeline pipeline)
   {
      // common state setup
      BaseState base = new BaseState(mgr, this);
      
      // global init processing
      pipeline.init(base);
      
      // create the queue which allows multiple threads to obtain tree names
      // to process
      ConcurrentLinkedQueue<String> queue =
         new ConcurrentLinkedQueue<String>(asts);
      
      int num = pipeline.getThreads();
      
      // safety first (non-positive numbers will be set to 1)
      if (num < 1)
         num = 1;
         
      // create latch to wait until all threads are complete
      CountDownLatch latch = new CountDownLatch(num);
      
      // create and start threads
      for (int i = 0; i < num; i++)
      {
         Driver driver = new Driver(queue, pipeline, base, latch);
         driver.setName(String.format("trpl%d", i));
         driver.setDaemon(true);
         driver.start();
      }
      
      // wait for the threads to finish
      try
      {
         latch.await();
      }
      
      catch (InterruptedException ie)
      {
         // TODO: log this
      }
      
      // TODO: common state reset/setup?
      
      // global term processing
      pipeline.term(base);
      
      // persist the AST metadata if needed
      if (!pipeline.isReadOnly())
      {
         mgr.save();
      }
   }
}