/*
** Module   : Pipeline.java
** Abstract : contains the list of tree walkers and associated configuration 
**            which defines a run of the tree processing engine
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20090526   @42537 Contains the list of tree walkers and associated 
**                           configuration which defines a run of the tree
**                           processing engine.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.trpl;

import java.util.*;
 
/**
 * Contains the list of tree walkers and associated configuration which defines
 * a run of the tree processing engine.
 */
public class Pipeline
{
   /** Number of threads in the pool to process trees. */
   private int threads = 1;
   
   /** Flag to force all loaded trees to save changes after {@link #term}. */
   private boolean autoSave = false;
   
   /** Trees will be loaded for read-only access. No changes will be saved. */
   private boolean readOnly = false;
   
   /** List of walkers to process as the pipeline. */
   protected List<WalkElement> walkers;
   
   /**
    * Default constructor. This instance will have no configured pipeline.
    */
   public Pipeline()
   {
   }
   
   /**
    * Construct an instance that contains the given list of walkers as the
    * pipeline to execute.
    *
    * @param    walkers
    *           List of walkers to process as the pipeline.
    */
   public Pipeline(List<WalkElement> walkers)
   {
      this.walkers = walkers;
   }
   
   /**
    * Called once per pipeline run, before any tree walkers are processed. At
    * the time this is called, no tree node is in scope nor is any tree being
    * processed.
    * <p>
    * This is useful for allocating resources or calculating/initializing any
    * values that should be handled once per pipeline.
    *
    * @param    state
    *           Provides the processing state of the current run. However it
    *           is important to note that there is no tree nor current node
    *           defined when this is called.
    */
   public void init(BaseState state)
   {
      // base implementation does nothing
   }
   
   /**
    * Called once per pipeline run, after all tree walkers are processed. At
    * the time this is called, no tree node is in scope nor is any tree being
    * processed.
    * <p>
    * This is useful for cleanup or termination processing. For example, this
    * is the common location for clearing or saving any values or state that
    * should be handled once per pipeline.
    *
    * @param    state
    *           Provides the processing state of the current run. However it
    *           is important to note that there is no tree nor current node
    *           defined when this is called.
    */
   public void term(BaseState state)
   {
      // base implementation does nothing
   }
   
   /**
    * Provides an independent list of the walkers that that define the pipeline.
    * The instance provided is unique and is a deep copy of the walker instances
    * contained in the pipeline. This may be called multiple times so that each
    * thread processing trees can have an independent list of unique instances.
    *
    * @return   The pipeline to execute or <code>null</code> if no walk
    *           elements are configured.
    */
   public List<WalkElement> getWalkers()
   {
      List<WalkElement> list = null;
      
      if (walkers != null)
      {
         list = new ArrayList<WalkElement>(1);
         
         for (WalkElement elem : walkers)
         {
            list.add(elem.duplicate());
         }
      }
      
      return list;
   }
   
   /**
    * Get the number of threads in the pool to process trees..
    *
    * @return   The number of threads.
    */
   public int getThreads()
   {
      return threads;
   }
   
   /**
    * Set the number of threads in the pool to process trees..
    *
    * @param    threads
    *           The number of threads.
    */
   public void setThreads(int threads)
   {
      this.threads = threads;
   }
   
   /**
    * Get the flag that determines if all loaded trees should be forced to save
    * changes after the pipeline completes processing (after {@link #term}).
    *
    * @return   The auto-save flag.
    */
   public boolean isAutoSave()
   {
      return autoSave;
   }
   
   /**
    * Set the flag that determines if all loaded trees should be forced to save
    * changes after the pipeline completes processing (after {@link #term}).
    *
    * @param    autoSave
    *           The auto-save flag.
    */
   public void setAutoSave(boolean autoSave)
   {
      this.autoSave = autoSave;
   }
   
   /**
    * Get the flag that determines if trees will be loaded for read-only
    * access. if <code>true</code>, no changes will be saved.
    *
    * @return   The read-only flag.
    */
   public boolean isReadOnly()
   {
      return readOnly;
   }
   
   /**
    * Set the flag that determines if trees will be loaded for read-only
    * access. if <code>true</code>, no changes will be saved.
    *
    * @param    readOnly
    *           The read-only flag.
    */
   public void setReadOnly(boolean readOnly)
   {
      this.readOnly = readOnly;
   }
}