/*
** Module   : Resolver.java
** Abstract : provides basic services to assist tree walkers 
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20090529   @42538 Provides basic services to assist tree walkers.
**                           This code is largely a subset of the code in
**                           the original pattern engine's AstSymbolResolver.
**                           One major difference is that this class is designed
**                           to allow multiple, independent instances operating
**                           simultaneously (the original code was a singleton).
**                           This design change enables highly multithreaded
**                           tree processing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.trpl; 
 
import java.util.*;
import com.goldencode.ast.*;

/**
 * Provides basic services to assist tree walkers.
 */
public class Resolver
{
   /** Map of ids to tree nodes. */
   private Map<Long, Aast> astMap = null;
   
   /** Set of filtered results which define a view of the original tree. */
   private Set<Aast> view = new LinkedHashSet<Aast>();
   
   /**
    * Construct an instance, optionally providing a pre-initialized ID to tree
    * map.
    *
    * @param    astMap
    *           Map of IDs to tree nodes or <code>null</code>.
    */
   public Resolver(Map<Long, Aast> astMap)
   {
      if (astMap != null)
      {
         this.astMap = astMap;
      }
      else
      {
         astMap = new HashMap<Long, Aast>();
      }
   }
   
   /**
    * Clear the contents of the view.
    */
   public void reset()
   {
      view = new LinkedHashSet<Aast>();
   }
   
   /**
    * Register an AST, so that it is accessible by its unique ID with the
    * {@link #getAst} method.
    *
    * @param    ast
    *           AST to be registered.
    *           
    * @throws   NullPointerException
    *           If <code>ast</code> has not yet been assigned a unique ID.
    */
   public void registerAst(Aast ast)
   {
      Long id = ast.getId();
      
      if (id == null)
      {
         throw new NullPointerException(ast.getSymbolicTokenType() +
                                        " (" + ast.getText() + ") has no ID");
      }
      
      astMap.put(id, ast);
   }
   
   /**
    * Helper to add an entire subtree (rooted at the specified AST node) to
    * the AST registry.
    *
    * @param    root
    *           Root node of the tree to be added.
    */
   public void registerTree(Aast root)
   {
      Iterator iter = root.iterator();
      
      while (iter.hasNext())
      {
         registerAst((Aast) iter.next());
      }
   }
   
   /**
    * Get the AST with the specified ID. The resulting AST may be a node in
    * either the source or target tree.
    *
    * @param    id
    *           ID which uniquely identifies the desired AST.
    *           
    * @return   The AST associated with <code>id</code>, or <code>null</code>
    *           if no node is found for <code>id</code>.
    */
   public Aast getAst(Long id)
   {
      return (Aast) astMap.get(id);
   }
   
   /**
    * Add an arbitrary AST node to the filtered view of AST nodes associated
    * with the current ruleset run. This view is made available to the next
    * walker in the pipeline.  Please note that the AST MUST exist in one
    * of the currently loaded ASTs (copy or target) and it must be registered
    * in the ID map.
    *
    * @return   <code>true</code> if the AST is valid (and thus was added),
    *           <code>false</code> otherwise.
    */
   public boolean addToView(Long id)
   {
      Aast ast = getAst(id);
      
      if (ast == null)
         return false;
         
      view.add(ast);
      
      return true;
   }
      
   /**
    * Add the all nodes of the given sub-tree to the view, descending at most
    * <code>maxLevels</code> generations deep.
    *
    * @param    root
    *           Node that is the root of the sub-tree to add to the view.
    * @param    maxLevels
    *           Number of generations to be added to the result set,
    *           beginning with the current AST. To add only the current
    *           AST, specify <code>1</code> (or better yet, use {@link
    *           #addToView}).
    */
   public void addTreeToView(Aast root, int maxLevels)
   {
      Iterator iter = root.iterator(maxLevels, null);
      
      while (iter.hasNext())
      {
         view.add((Aast) iter.next());
      }
   }
   
   /**
    * Add to the filtered view of ASTs all of the <em>copy</em> nodes
    * which form the path from an ancestor node of type <code>type</code>
    * to the given AST. The nodes are added in root to leaf order.
    * If an ancestor node of type <code>type</code> does not exist, no nodes
    * are added (not even the current one).
    *
    * @param    node
    *           Node from which to start the search.
    * @param    type
    *           Ancestor token type which marks the beginning of the path.
    *
    * @return   <code>true</code> if an ancestor of type <code>type</code>
    *           was found and the path of nodes was successfully added to the
    *           filtered view, else <code>false</code>.
    */
   public boolean addPathToViewFrom(Aast node, long type)
   {
      Aast next = node;
      
      Stack<Aast> stack = new Stack<Aast>();
      
      while (next != null)
      {
         stack.push(next);
         
         if (next.getType() == type)
         {
            break;
         }
         
         next = next.getParent();
      }
      
      // If we have walked all the way back to the root AST but have not
      // found an ancestor which matches our boundary condition, do not
      // add anything to the results set.
      if (next == null)
      {
         return false;
      }
      
      // Add each result in order from root to leaf.
      while (!stack.isEmpty())
      {
         view.add(stack.pop());
      }
      
      return true;
   }
   
   /**
    * Return an iterator on the filtered view of ASTs currently associated
    * with this resolver. Actions may add copy AST nodes to this view.
    * <p>
    * This method provides the mechanism which allows the next walker in a
    * pipeline to iterate over the filtered results of the previous walker's
    * processing of the tree hierarchy.
    * <p>
    * The collection underlying this view is replaced with an empty
    * collection before each walker in the pipeline is processed. However,
    * it is safe to iterate over the view returned by this method after this
    * occurs, because a new collection is used for each pass, and the
    * <code>Iterator</code> instance returned by this method continues to be
    * associated with the results collection which was in use at the time
    * it was called.
    *
    * @return   An iterator over the filtered view which results from the
    *           application of a ruleset against an AST.
    */
   Iterator view()
   {
      return view.iterator();
   }
}