/*
** Module   : WalkElement.java
** Abstract : stores configuration values for a tree walker in a pipeline
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20090526   @42540 Stores configuration values for a tree walker in
**                           a pipeline.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.trpl;

/**
 * Stores configuration values for a tree walker in a pipeline.
 */
public class WalkElement
{
   /** Input to this walker is the entire tree or the current view. */
   private InputMode mode = InputMode.TREE;
   
   /** The maximum levels from the root to walk, use 0 for no limit. */
   private int maxDepth = 0;
   
   /** <code>false</code> to bypass nodes marked as hidden. */
   private boolean walkHidden = true;
   
   /** The walker to execute. */
   private TreeWalker walker = null;
   
   /**
    * Constructor to create an instance which contains the given walker.
    *
    * @param    walker
    *           The tree walker to execute.
    */
   public WalkElement(TreeWalker walker)
   {
      this.walker = walker;
   }
   
   /**
    * Default constructor with no configured walker.
    */
   private WalkElement()
   {
   }
   
   /**
    * Get the flag which determines the input to the walk.
    *
    * @return   The source of AST nodes for the walk.
    */
   public InputMode getMode()
   {
      return mode;
   }
   
   /**
    * Set the flag which determines the input to the walk.
    *
    * @param    mode
    *           The source of AST nodes for the walk.
    */
   public void setMode(InputMode mode)
   {
      this.mode = mode;
   }
   
   /**
    * Get the maximum depth of the walk in levels downward from the root node.
    *
    * @return   The maximum levels down from the root to walk.
    */
   public int getMaxDepth()
   {
      return maxDepth;
   }
   
   /**
    * Set the maximum depth of the walk in levels downward from the root node.
    *
    * @param    maxDepth
    *           The maximum levels down from the root to walk.
    */
   public void setMaxDepth(int maxDepth)
   {
      this.maxDepth = maxDepth;
   }
   
   /**
    * Get the flag that determines if this walk should invoke hooks for hidden
    * nodes.
    *
    * @return   <code>false</code> to bypass hidden nodes in the walk.
    *           <code>true</code> to include hidden nodes in the walk.
    */
   public boolean isWalkHidden()
   {
      return walkHidden;
   }
   
   /**
    * Set the flag that determines if this walk should invoke hooks for
    * hidden nodes.
    *
    * @param    walkHidden
    *           <code>false</code> to bypass hidden nodes in the walk.
    *           <code>true</code> to include hidden nodes in the walk.
    */
   public void setWalkHidden(boolean walkHidden)
   {
      this.walkHidden = walkHidden;
   }
   
   /**
    * Obtain the walker to be hooked for this stage of the pipeline.
    *
    * @return   The callbacks to be invoked.
    */
   public TreeWalker getWalker()
   {
      return walker;
   }
   
   /**
    * Make a deep copy of this instance and return that new instance.
    *
    * @return   The copy of this instance.
    */
   public WalkElement duplicate()
   {
      WalkElement copy = new WalkElement();
      
      copy.mode       = mode;
      copy.maxDepth   = maxDepth;
      copy.walkHidden = walkHidden;
      copy.walker     = walker.duplicate();
      
      return copy;
   }
}