/*
** Module   : library_win.c
** Abstract : implementation of operating-system specific helpers for library management
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20131024 First version.
** 002 EVL 20140127 Cycle variable needs to be declared outside the loop.
** 003 GES 20140212 Added flag to let the native layer conditionally process for loading
**                  persistent libraries. It is ignored on Windows because there is no
**                  differential behavior to implement.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include "library.h"
#include <windows.h>
#include <string.h>

// local function declarations
HMODULE int64ToHandle(int64_t modhandle);
int64_t handleToInt64(HMODULE handle);
int64_t pointerToInt64(FARPROC ptr);
                                          
/**
 * Attempts to load the named library.  This will use the operating-system runtime library
 * loading mechanism.
 *
 * @param    libname
 *           This may be any valid library filename for the current system, including names
 *           that contain path separators (both relative and absolute). This should not be
 *           <code>null</code> or the empty string.
 * @param    persist
 *           If <code>true</code>, the library should not be unloaded at the end of the
 *           function invocation, which means that this method should cache the module
 *           handle if it is successfully loaded.
 *
 * @return   An operating-system specific "handle" to the library. Under Windows this is
 *           the "module handle" of type HMODULE and under Linux it is just referred to as
 *           a "handle" that is returned as a void pointer.  In either case, if
 *           <code>NULL</code> (0) is returned, then the call failed.
 */ 
int64_t load_library_native(const char* libname, int persist)
{
   if (libname == NULL || strlen(libname) == 0)
      return (int64_t) 0;
   
   char* modname = strdup(libname);
   
   // out of memory, just return a failure
   if (modname == NULL)
      return (int64_t) 0;
   
   int len = strlen(modname);
   
   // on Windows (and only Windows), the 4GL implements a conversion of any forward slash
   // chars to be backslashes, for the purposes of this lookup (but not for any caching that
   // may be done of the returned handle)
   int i;
   for (i = 0; i < len; i++)
   {
      if (modname[i] == '/')
         modname[i] = '\\';
   }
   
   // at this time we have no known requirements that would force use of LoadLibraryEx
   HMODULE handle = LoadLibrary(TEXT(modname));
   
   // clean up after ourselves
   free(modname);
   
   return handleToInt64(handle);
}

/**
 * Attempts to unload the library identified by the "handle" which was returned from a
 * previous call to load_library_native.  This will use the operating-system runtime library
 * unloading mechanism. Although this call may fail, Progress seems to always silently ignore
 * any failures, so this method has no return value.
 *
 * @param    modhandle
 *           This must be a valid handle to the library.
 */ 
void unload_library_native(int64_t modhandle)
{
   FreeLibrary(int64ToHandle(modhandle));
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given symbol in the library.  The symbol name must match exactly, including the case
 * and any decorations that exist as exported in the library.
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    funcname
 *           This is the exact symbol to lookup in the library.
 *
 * @return   The memory address of the function or <code>NULL</code> (0) if the function cannot
 *           be found by that name in the given library.
 */ 
int64_t find_by_name_native(int64_t modhandle, const char* funcname)
{
   FARPROC addr = GetProcAddress(int64ToHandle(modhandle), funcname);
   return pointerToInt64(addr);
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given ordinal in the library.  Only some operating systems allow exporting functions by
 * ordinal number.  Those that do allow it, will typically have a table of contiguous ordinal
 * values in the library, each of which represents a function that may or may not be also
 * exported by name (as a symbol).
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    ordinal
 *           The ordinal to lookup in the library.  Only the lower word is honored.
 *
 * @return   Linux/UNIX systems don't support ordinal exports. This always returns 0.
 */ 
int64_t find_by_ordinal_native(int64_t modhandle, uint16_t ordinal)
{
   #ifdef WORD_SIZE_64
   const char* ptr = (char*) ((int64_t) ordinal);
   #else
   const char* ptr = (char*) ((int32_t) ordinal);
   #endif
   
   // yes, windows really does "re-use" the character pointer as the ordinal, if the lower
   // word is a valid index into the ordinal table (between the starting and ending
   // ordinals, inclusive); it is ridiculous, but well, its windows
   return find_by_name_native(modhandle, ptr);
}

/**
 * Convert the given integer value into a void pointer in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 *
 * @return   The module handle as a void pointer.
 */
HMODULE int64ToHandle(int64_t modhandle)
{
   #ifdef WORD_SIZE_64
   HMODULE handle = (HMODULE) modhandle;
   #else
   HMODULE handle = (HMODULE) ((int32_t) modhandle);
   #endif
   
   return handle;
}

/**
 * Convert the given handle into an integer value in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    handle
 *           This must be a handle.
 *
 * @return   The module handle as an integer.
 */
int64_t handleToInt64(HMODULE handle)
{
   #ifdef WORD_SIZE_64
   int64_t result = (int64_t) handle;
   #else
   int32_t result = (int32_t) handle;
   #endif
   
   return (int64_t) result;
}

/**
 * Convert the given void pointer into an integer value in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    ptr
 *           This must be a pointer.
 *
 * @return   The module handle as an integer.
 */
int64_t pointerToInt64(FARPROC ptr)
{
   #ifdef WORD_SIZE_64
   int64_t result = (int64_t) ptr;
   #else
   int32_t result = (int32_t) ptr;
   #endif
   
   return (int64_t) result;
}