/*
** Module   : shell.c
** Abstract : Run OS shell and capture input and output.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140318 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include <stdlib.h>
#include <string.h>
#include "shell.h"
#include "com_goldencode_terminal_NativePty.h"

/**
 * Start an OS Shell session. On Linux/Unix The master PTY file descriptor is used
 * to capture the INPUT and OUTOPUT streams. On Windows pipe handles are used.
 * 
 * @param   env
 *          Pointer to a structure that provides access to JNI environment
 * @param   thisObj
 *          Pointer to Java object instance having which implements native artefacts.
 * @param   term
 *          A string to set terminal type environment.
 *          Example: "TERM=vt100"  
 * @param   cmdline
 *          Command line arguments.
 * @param   overlay
 *          Overlapping mode.
 *
 * @return  On success return PID and set the "mfd_inp" and "mfd_out" fields on the native class
 *          with the master PTY file descriptor. On Linux mfd_inp and mfd_out are the same.
 *          For Windows mfd_inp and mfd_out are pipe handles.
 *          On error, -1 is returned, and errno is set appropriately.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_terminal_NativePty_shell(JNIEnv* env, 
                                                                    jobject thisObj, 
                                                                    jstring term,
                                                                    jobjectArray cmdline,
                                                                    jint overlay)
{
   // Master PTY descriptors
   int mfd_inp = -1, mfd_out = -1;
   
   // Process ID
   int pid = -1;
   
   // Number of command line arguments
   int len = 0;

   // Command args  
   char** argv = NULL;
   
   int i;
   if (cmdline != NULL)
   { 
      len = (int) (*env)->GetArrayLength(env, cmdline);
      argv = malloc(sizeof(char*) * (len + 1)); 
      
      char** current = argv;
      
      // copy the strings into our own array
      for (i = 0; i < len; i++, current++)
      {
         jobject next = (*env)->GetObjectArrayElement(env, cmdline, i); 
         char*   arg  = (char*) (*env)->GetStringUTFChars(env, (jstring) next, NULL);
         
         if (arg == NULL)
         {
            // out of memory!
            return -1;
         }
         
         *current = strdup(arg);
         
         (*env)->ReleaseStringUTFChars(env, (jstring) next, arg);
      }
      
      // terminate the argv[] with a null pointer
      *current = NULL;
   }
   
   // Get terminal type
   const char* terminal = (*env)->GetStringUTFChars(env, term, NULL);   

   // Run shell
   pid = os_shell((char *)terminal, argv, &mfd_inp, &mfd_out, overlay);
   
   (*env)->ReleaseStringUTFChars(env, term, terminal);

   if (argv)
   {
      // Clean up
      for (i = 0; i < len; i++)
      {
         free(argv[i]);
      }
      free(argv);
   }
   
   if (pid > 0)
   {
      // Get a reference to this object's class
      jclass thisClass = (*env)->GetObjectClass(env, thisObj);

      // Get the Field ID of the instance variables "mfd_inp"
      jfieldID fidNumber = (*env)->GetFieldID(env, thisClass, "mfd_inp", "I");
      if (fidNumber)
      {
         (*env)->SetIntField(env, thisObj, fidNumber, mfd_inp);
      }

      // Get the Field ID of the instance variable "mfd_out"
      fidNumber = (*env)->GetFieldID(env, thisClass, "mfd_out", "I");
      if (fidNumber)
      {
         (*env)->SetIntField(env, thisObj, fidNumber, mfd_out);
      }      
   }
   
   // return PID
   return pid;
}

/**
 * Wait for a child process to exit.
 * 
 * @param   env      
 *          Pointer to a structure that provides access to JNI environment
 * @param   thisObj
 *          Pointer to Java object instance having which implements native artefacts.
 * @param   pid
 *          Child process pid.
 *
 * @return  Process status.
 *          On error, -1 is returned, and errno is set appropriately.
 */
JNIEXPORT jint JNICALL Java_com_goldencode_terminal_NativePty_waitPid(JNIEnv* env, 
                                                                      jobject obj, 
                                                                      jint pid)
{
   return wait_pid(pid);
}

/**
 * Read from master PTY file descriptor.
 * 
 * @param   env      
 *          Pointer to a structure that provides access to JNI environment
 * @param   thisObj
 *          Pointer to Java object instance having which implements native artefacts.
 * @param   mfd
 *          Master PTY file descriptor.
 *
 * @return  A byte array read from master PTY file descriptor. On error return an empty array.
 */
JNIEXPORT jbyteArray JNICALL Java_com_goldencode_terminal_NativePty_read(JNIEnv *env, 
                                                                         jobject obj, 
                                                                         jint mfd) 
{
   char buffer[BUFFER_SIZE];
   int size = read_pty(mfd, buffer, sizeof(buffer));

   if (size == -1) 
   {
      return (*env)->NewByteArray(env, 0);
   }    
   
   jbyteArray array = (*env)->NewByteArray(env, size);
   (*env)->SetByteArrayRegion (env, array, 0, size, (jbyte*)buffer);
   return array;
}

/**
 * Write to master PTY file.
 * 
 * @param   env      
 *          Pointer to a structure that provides access to JNI environment
 * @param   thisObj
 *          Pointer to Java object instance having which implements native artifacts.
 * @param   mfd
 *          Master PTY file descriptor.
 * @param   array
 *          A byte array to write to master PTY file.
 *
 * @return  Number of bytes written. 
 */
JNIEXPORT jint JNICALL Java_com_goldencode_terminal_NativePty_write(JNIEnv *env, 
                                                                    jobject obj, 
                                                                    jint mfd, 
                                                                    jbyteArray array) 
{
   jbyte* buffer = (*env)->GetByteArrayElements(env, array, NULL);
   jsize size = (*env)->GetArrayLength(env, array);
   // write buffer
   jint result = write_pty(mfd, (void*) buffer, (size_t) size);   
   (*env)->ReleaseByteArrayElements(env, array, buffer, 0);
   return result;
}