/*
** Module   : signals_win.c
** Abstract : Windows implementation for initialize calls to prepare native layer to work
**            properly.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20140117 Created initial version to implement init steps.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include "signals.h"
#include "init.h"
#include "terminal.h"

#include <sys/types.h>
#include <windows.h>
#include <fcntl.h>

// Internal Windows specific functions and variables

// Event semaphore object handle
HANDLE hSignalEvent;

/**
 * sigintHandler - The new handler for SIGINT incoming events.
 */ 
void sigintHandler(int sig)
{
   // When the handler is called the signals become available to got usual way
   // Ignore the signals for other SIGINT's
   signal(SIGINT,  SIG_IGN);
   // The SIGINT received, check if we are not in the middle of the previous event handling
   if (WaitForSingleObject(hSignalEvent, 0) != WAIT_OBJECT_0) // No previous event
   {
      SetEvent(hSignalEvent); // Let's go Java notifier
   }
}

// JNI calls code starts here

/**
 * initSignalsProcessing - Set up signal handling to ignore job control signals and redefine
 *                         CTRL-C handlers(reflecteds in the child process as well).
 *                         Windows version.
 */ 
JNIEXPORT void JNICALL initSignalsProcessing(JNIEnv* env)
{
   // TODO: Tink what we can do here. Just do nothing for now.
   // The function sigemptyset() is not supported in Windows due to Windows signal paradigm does
   // not have a mask and what is more implortant we do not need to prepare structure used in
   // this function under Linux in Windows version if the init call.

   // Creating event object
   hSignalEvent = CreateEvent(NULL,  // Default security attributes
                              TRUE,  // Manual reset flag, TRUE meaning need to control the state 
                              FALSE, // Initial state is not set
                              NULL); // No name required for event

   // Error while event creating
   if (hSignalEvent == NULL)
   {
      // Unable to create semaphore, terminating
      throwException(env, "java/lang/IllegalStateException", "Event semaphore creation failure");
      return;
   }
   // Setting up signal handlers
   // The signals SIGTSTP, SIGTTIN, SIGTTOU are missing in Windows meaning ignore by default
   // This means in turn we do not need special calls to make them ignored.
   signal(SIGINT,  sigintHandler);
}

/**
 * waitSigint - Intercepts external SIGING signal to inform the caller the special processing is
 *              required to handle.
 */ 
JNIEXPORT jint JNICALL waitSigint()
{
   jint result = -1; // This is unexpected return code
   
   // Waiting until event arriving
   if (WaitForSingleObject(hSignalEvent, INFINITE) == WAIT_OBJECT_0)
   {
      result = 0;
   }
   
   return result;
}

/**
 * unlockSigint - Enable SIGINT handler again.
 */ 
JNIEXPORT void JNICALL unlockSigint(JNIEnv* env)
{
   // Need to watch for external INT signals to react as for CTRL-C 
   signal(SIGINT,  sigintHandler);
   // Resetting event to be able infinite wait in waitSigint()
   if (!ResetEvent(hSignalEvent))
   {
      // Unable to reset semaphore, terminating
      throwException(env, "java/lang/IllegalStateException", "Unable to reset event semaphore");
      return;
   }
}

/**
 * stopWaitingSigint - Disable block waitings for SIGINT.
 */ 
JNIEXPORT void JNICALL stopWaitingSigint()
{
   // Ignore the signals for other SIGINT's
   signal(SIGINT,  SIG_IGN);
   // Unbock possible waiters
   SetEvent(hSignalEvent);
}