/*
** Module   : ThinClient.java
** Abstract : Thin Client class for generic UI driver implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description------------------------------
** 001 EVL 20050710   @21707 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050822   @22196 Adding code to have the client and server on
**                           the same machine.
** 003 EVL 20050824   @22239 Adding the events/triggers processing.
** 004 EVL 20050825   @22275 Adding GO and CHOOSE button handlers.
** 005 EVL 20050827   @22375 Adding usage of the ScreenBuffer to handle
**                           the events.
** 006 EVL 20050830   @22416 Adding ENTRY and LEAVE events handling.
** 007 SIY 20050831   @22519 Refactoring (see @@22394-22409)
** 008 EVL 20050903   @22457 Adding KEY event handler. (Merged changes
**                           from EVL)
** 009 EVL 20050907   @22598 Getting the CHARVA key code while
**                           interception of the key events.
** 010 EVL 20050908   @22617 Using single handler callback for all events.
**                           The key code is stored in ScreenBuffer for
**                           key events.
** 011 EVL 20050912   @22682 Adding the validation list and the validation
**                           method.
** 012 EVL 20050913   @22701 Adding debug info message for the case when
**                           it is not possible to identify the widget ID.
**                           Fixing focus event handling for radio-set.
** 013 EVL 20050916   @22740 Changing to use BaseDataType class as the
**                           value of the widget to be placed in screen
**                           buffer.
** 014 EVL 20050919   @22751 Modification of the key handler to be able to
**                           intercept the events from radio set,
**                           selection list and combo box.
** 015 EVL 20050920   @22763 Adding code to check whether the value of the
**                           widget to be validated is really changed or
**                           not and the validation is not necessary.
** 016 EVL 20050921   @22768 Adding the usage of the undoValue() method to
**                           roll back the widget change for failed
**                           validation.
** 017 EVL 20050922   @22815 Adding the interceptor for editing block
**                           mode.
** 018 EVL 20050926   @22856 Including the value into the screen buffer
**                           via EDITING block handling.
** 019 NVS 20051204   @24062 Added trigger and validate methods to this
**                           class that call appropriate server exports to
**                           execute triggers and validators.
** 020 SIY 20051204   @24063 Added stubs for ClientExports methods and
**                           many other changes.
** 021 NVS 20051229   @24064 Added consumeEvent() implementation.
** 022 NVS 20060111   @24065 apply() now returns ScreenBuffer to server.
** 023 GES 20060113   @23901 Improvements to message (it still doesn't
**                           support any options).  Now it handles arrays
**                           of different object types including
**                           BaseDataType.
** 024 NVS 20060114   @24066 hide() export hides widgets as well as frames
** 025 GES 20060116   @23924 Safety code in some widget/frame lookup
**                           methods.
** 026 GES 20060116   @23927 Rework and additions to putScreen() to make
**                           it functional.
** 027 GES 20060116   @23928 Fixed readKey().  It never broke out of the
**                           timing loop when it got a key, it just kept
**                           looping.
** 028 GES 20060116   @23932 view() needed to request focus otherwise the
**                           previous frame (even if hidden) still had
**                           focus.
** 029 GES 20060116   @23948 Fixed pause(0) and cleaned up signature.
** 030 NVS 20060119   @23979 Implemented correct Pause Before Hide logic
**                           for view, clear and hide operations. Other
**                           situations causing frame hide like Z-order
**                           change still need some work.
** 031 GES 20060121   @24011 Moved core apply logic to a worker that
**                           doesn't return a screen buffer since the
**                           vast majority of times that apply() is
**                           called it is from the waitFor loop which
**                           doesn't need the screen buffer returned.
**                           This way the only case where we take the
**                           hit of building the screen-buffer is the
**                           case where the server calls apply().
** 032 NVS 20060125   @24067 Implemented iterateFrame(), down(), up() and
**                           scroll() methods that manage the cursor and
**                           contents of down frame.
** 033 SIY 20060125   @24123 Support code for the frame manager.
** 034 NVS 20060130   @24147 All client and server exports returning
**                           screen buffer, now take screen buffer as
**                           well. These are: waitFor(), apply(),
**                           trigger() and validate.
** 035 SIY 20060130   @24152 Added pauseBeforeEnd() exported method.
** 036 SIY 20060131   @24191 Extended MESSAGE support and merged
**                           status/help handling.
** 037 NVS 20060131   @24192 view() call should bring the frame on top.
** 038 NVS 20060201   @24217 Hide details of TempText creation in putScreen
**                           inside the TempText class.
** 039 NVS 20060202   @24218 getScreenLines() returned screen height, not
**                           the height of frame display area.
** 040 SIY 20060203   @24242 Fixes for hiding/placing frames and pause
**                           management.
** 041 NVS 20060206   @24307 Using TypeAhead class for reading events,
**                           clearing the type ahead buffer and catching
**                           Ctrl-C (STOP).
** 042 NVS 20060209   @24371 Using Control.interrupt() to deliver Ctrl-C.
** 043 SIY 20060209   @24436 Fixes to make ComboBox, SelectionList,
**                           RadioSet controls handle global keys.
** 044 GES 20060210   @24507 Simplified exporting of methods.
** 045 SIY 20060210   @24521 Added getSelection() method; fixed trap if
**                           typing is started before first window does
**                           appear on the screen.
** 046 GES 20060210   @24532 frame-field attribute must return empty
**                           string when null.
** 047 GES 20060211   @24541 Added initialization for remote file system
**                           and remote stream services on the client.
** 048 GES 20060212   @24559 Added initialization for process daemon.
** 049 SIY 20060213   @24578 Support for DialogBox; fixes for the frame
**                           hiding.
** 050 NVS 20060213   @24585 Added installation of ProgressKeyboardMapper.
** 051 GES 20060212   @24588 Exposed terminal suspend/resume and pseudo-
**                           terminal process launching mechanism.
** 052 SIY 20060215   @24598 Implemented messageBox(). Some performance
**                           improvements.
** 053 NVS 20060221   @24626 STOP key function raises STOP condition. HELP
**                           key function executes Help Application on the
**                           server.
** 054 SIY 20060220   @24645 Implemented choose() support.
** 055 NVS 20060222   @24667 ENDKEY key action should be used for this
**                           event, not the END key.
** 056 GES 20060214   @24686 Stream redirection and some general cleanup.
** 057 NVS 20060224   @24728 ThinClient tracks the high level I/O
**                           operations for proper END-ERROR handling.
**                           Added new field "interactions". It is set to
**                           0 in blockStart() and blockIterate() and
**                           incremented by 1 on entry to waitFor(). END
**                           condition is generated if interactions equals
**                           1.
** 058 GES 20060225   @24734 Save/restore visible state when switching
**                           between interactive and redirected output
**                           modes. In redirected mode, everything on
**                           the interactive screen must be set hidden
**                           so that it doesn't draw to the stream. Also
**                           added more APIs into forced interactive mode.
** 059 GES 20060226   @24756 Header and footer support.
** 060 SIY 20060226   @24765 Fixed handling of SET/UPDATE in message().
**                           Fixed focus management in messageBox().
** 061 GES 20060227   @24774 Final header and footer support.
** 062 NVS 20060302   @24848 processEvents() export now takes and returns
**                           the ScreenBuffer.
** 063 NVS 20060302   @24872 Keyboard events are translated into
**                           respective key actions whenever possible,
**                           then actions are recognized by their codes.
** 064 NVS 20060306   @24902 Added call to ErrorManager.initClient.
** 065 NVS 20060306   @24907 WaitList does not need default events, they
**                           are recognized by ThinClient anyway.
** 066 NVS 20060308   @24969 Using new ScreenBuffer which keeps state and
**                           allows for differential exchange.
** 067 NVS 20060314   @25072 Fixed setScreenBuffer() to transfer the state
**                           to widgets for all widgets.
** 068 SIY 20060315   @25067 Yet another try to fix frame hiding.
** 069 SIY 20060316   @25142 Added support for headers.
** 070 SIY 20060228   @25143 Changed signature for the getColumns() and
**                           other methods.
** 071 NVS 20060320   @25159 Simplified validation protocol by using new
**                           class ValidationResult.
** 072 SIY 20060320   @25167 Fixed clear() on headers.
** 073 GES 20060322   @25186 Added terminal type processing methods.
** 074 NVS 20060322   @25200 Added refocusing to the specified widget in
**                           waitFor().
** 075 NVS 20060322   @25211 Implemented widget value validation during
**                           all kinds of blocking input.
** 076 SIY 20060323   @25215 Changes to support new down frame
**                           implementation. As a side effect refactored
**                           some encapsulation violations.
** 077 NVS 20060324   @25236 Fixed the event processing for the focus
**                           change related events. Now they are properly
**                           routed and handled. This fix enabled LEAVE
**                           and ENTER events generation and, therefore,
**                           the changed widget validation.
** 078 SIY 20060327   @25279 Pass focus events to widgets.
** 079 SIY 20060401   @25308 Access header flag directly.
** 080 SIY 20060412   @25598 Don't hide messages when frame iterates.
** 081 SIY 20060424   @25688 Fixed pausing logic and messages are set
**                           hidden where appropriate, without breaking
**                           pausing logic.
** 082 GES 20060501   @25907 Don't return a frame marked for destruction
**                           as an active frame as this causes problems
**                           on the server.
** 083 GES 20060502   @25928 Safer pausing logic.
** 084 SIY 20060502   @25930 Mark frames as such that had input in all
**                           cases when there was a waiting for a key when
**                           this frame was a top Z-order frame.
** 085 GES 20060505   @25984 Scoped interactions member to a given block
**                           such that enclosing block's interactions
**                           value is restored when more deeply nested
**                           blocks end. Added a mechanism to raise the
**                           error condition when in an UPDATE and the
**                           END-ERROR key is pressed.  Previously this
**                           was hard coded to the ENDKEY condition but
**                           while the UPDATE should exit (normally it
**                           retries on error) it must also sometimes be
**                           properly delivered as an ERROR condition for
**                           the containing logic to be correct.
** 086 SIY 20060504   @25987 Another attempt to handle frame hiding
**                           properly.
** 087 SIY 20060505   @25990 Check for null frame in setScreenBuffer().
** 088 SIY 20060509   @25990 Fixed hideAll().
** 089 NVS 20060509   @26035 Partial implementation of underline() export.
**                           For regular frames, it clears the frame as it
**                           should. For down frames, it moves the cursor
**                           down without causing underlining to appear.
** 090 GES 20060510   @26047 hideAll() needs to clear all temp text
**                           widgets (put screen output is all removed).
** 091 GES 20060510   @26059 New flag passed on blockIterate().
** 092 SIY 20060510   @26063 More fixes in hiding/pausing.
** 093 SIY 20060511   @26075 Fixed hiding/pausing, hide() now clears
**                           messages as it does in Progress.
** 094 NVS 20060511   @26087 Reimplemented conditional down() and up()
**                           calls in GenericFrame. Now block iteration
**                           issues a conditional down() on the server and
**                           that call is entirely handled there.
** 095 SIY 20060516   @26159 hideAll() explicitly sets needPause to true.
** 096 GES 20060517   @26214 tinyInput always must have left aligned data.
** 097 SIY 20060517   @26217 Commented out redundant call to down() in
**                           iterateFrame() because it causes layout
**                           issues now.
** 098 SIY 20060517   @26233 Added handling for block retry.
** 099 GES 20060522   @26431 Ensure that any exit from waitFor() will
**                           still trip cleanup, including via exception.
** 100 GES 20060523   @26441 Refactored logic in continueEditing() to
**                           move cleanup processing into a common method
**                           stopEditingMode() which is also exposed as
**                           a client export (so it can be called directly
**                           from the server).
** 101 GES 20060523   @26448 Limited validation processing to only those
**                           widgets that are enabled.
** 102 SIY 20060524   @26458 Fixed getFrameLine() method.
** 103 SIY 20060525   @26537 Another attempt to fix pause before exit
**                           handling.
** 104 SIY 20060526   @26547 More careful tracking of differences in
**                           handling of F1/F4 exits in CHOOSE and
***                          WAIT-FOR.
** 105 NVS 20060526   @26584 First attempt to display frame completes its
**                           realization. Realized frame won't be subject
**                           to layout or screen placement changes.
** 106 GES 20060529   @26677 Added support for condition processing in
**                           pause processing (e.g. END-ERROR).
** 107 NVS 20060531   @26827 Fixed getFrameLine() method to return a right
**                           1-based value.
** 108 GES 20060531   @26810 Changes to make enable() maintain tab order
**                           and handle focus changes as needed.
** 109 NVS 20060601   @26896 Another fix to getFrameLine() to properly
**                           report 0 as the frame line before any output
**                           has occurred on the down frame.
** 110 SIY 20060601   @26908 Redesign of pause handling, many other minor
**                           updates and fixes.
** 111 NVS 20060601   @26939 Fixed a regression in pause() caused by some
**                           new code ignoring the fact waitForKey() can
**                           return null.
** 112 NVS 20060608   @27018 Fixed a bug in button LEAVE event handling
**                           causing a class cast exception.
** 113 SIY 20060607   @27100 Changed blockIterate/blockStart to support
**                           reliable handling of exceptions thrown in
**                           pause processing.
** 114 SIY 20060609   @27101 Fixed another hiding issue.
** 115 NVS 20060609   @27113 waitFor processing now differentiates the
**                           events in the WaitList as triggers vs exit
**                           events.
** 116 SIY 20060612   @27119 hide() should not hide messages, only
**                           pause() does that, if necessary.
** 117 NVS 20060612   @27142 Fixed many NullPointerExceptions in trigger
**                           handling code for ANYWHERE triggers with no
**                           frame involved.
** 118 GES 20060613   @27160 Reset interactions on block retry to ensure
**                           F4 processing is correct (otherwise it may
**                           generate an error where an end should occur).
** 119 NVS 20060613   @27172 Reimplemented getActiveFrame() based on a new
**                           method in Screen returning the top frame.
**                           Fixed waitFor loop checking for events (now
**                           both the key codes and events are checked).
** 120 GES 20060614   @27191 Converted the status related signatures to
**                           String parameters instead of character. Fixed
**                           all status*() forms to output the latest
**                           status message to the terminal at the time
**                           of the call.
** 121 GES 20060614   @27197 Apply processing must temporarily create a
**                           default wait list if one is not already
**                           active.
** 122 SIY 20060616   @27345 Do not set pause flag if frame is not visible.
** 123 SIY 20060616   @27399 Another round of pausing/hiding fixes.
**                           All calls to pauseBeforeHide() made exception
**                           safe.
** 124 SIY 20060617   @27421 Fixed typo in comments.
** 125 NVS 20060616   @27431 CHOOSE event generation off of SELECTED state
**                           of ItemEvent.
** 126 GES 20060619   @27453 All screen buffers that are passed to the
**                           server (by a return value from a method call
**                           from the server or via a parameter when
**                           calling the server) must be properly matched
**                           up with the current widget being processed.
**                           The previous implementation of the
**                           getScreenBuffer() method returned the screen
**                           buffer associated with the used the currently
**                           focused widget.  This almost always worked
**                           but there were cases where apply() could be
**                           operating on a different frame than would be
**                           detected via the current focus.  This must
**                           never be allowed to happen as it corrupts
**                           the server's resulting screen buffer when
**                           the contents are merged on the server side.
** 127 NVS 20060621   @27511 Rewritten the processProgressEvent() method
**                           to straighten up the logic and implement some
**                           missing parts like CURRENT-WINDOW related
**                           triggers and exit conditions. New code makes
**                           use of the new optimized WaitListClient class
**                           methods.
** 128 NVS 20060622   @27527 Fixed a NullPointerException regression in
**                           processProgressEvent() caused by a too broad
**                           check when there is no enabled widget on the
**                           screen.
** 129 SIY 20060621   @27532 More changes in frame hiding algorithms.
** 130 GES 20060623   @27558 Return lastKey for message (if it paused),
**                           choose and pause.  Added safety code to
**                           ensure that callers of waitForWorker that
**                           need post processing use try/finally to
**                           ensure it occurs.
** 131 NVS 20060623   @27567 Fixed the focus change related event handling
**                           and provided a synchronous LEAVE event that
**                           activates a trigger and can be consumed by
**                           RETURN NO-APPLY which cancels the focus
**                           change.
** 132 NVS 20060626   @27570 Rewritten nextPrompt() method. Now it saves
**                           the widget ID for later use in waitFor(). The
**                           use is one time and only if the waitFor()
**                           does not have a specific widget to focus.
** 133 NVS 20060627   @27591 enable() call changes the focus immediately
**                           if it's disabling the widget in focus. All
**                           enabling calls calculate the new focus widget
**                           and store its ID for later use in waitFor()
**                           instead of immediate focus change. view()
**                           calls don't change focus either. Only waitFor
**                           method changes the focus for all delayed
**                           requests, including those made with calls to
**                           nextPrompt().
** 134 SIY 20060626   @27586 Shifted frame hiding from blockIterate() to
**                           placeFrameInt() to enable more granular
**                           condition control.
** 135 GES 20060627   @27606 Signature change to pass a screen buffer
**                           to enable() to allow truly dynamic down
**                           frame sizing.
** 136 SIY 20060628   @27662 Support for dynamic down support.
** 137 NVS 20060629   @27705 Backed out the change #131 as having too much
**                           trouble with lost synchronism. Another way
**                           of controlling the focus change centrally
**                           will be implemented.
** 138 GES 20060629   @27709 Now that enable() takes a screen buffer
**                           parameter it also needed to honor the changes
**                           that were sent otherwise they would not be
**                           displayed as expected (since ENABLE stmt
**                           is supposed to be a superset of VIEW). A
**                           previous bug on the server side sent the
**                           changes over multiple times, masking the
**                           problem.  When that was fixed, the bug here
**                           became visible.
** 139 GES 20060630   @27738 choose() must process a screen buffer on
**                           both input and output. Also the frameIndex
**                           and frameValue processing was made a bit
**                           safer (though they are still not handled
**                           properly in non-CHOOSE situations).
** 140 NVS 20060629   @27769 Added static tryFocusChange() method which is
**                           a correct way of changing focus synchronously
**                           between widgets of the active frame while
**                           maintaining the Progress semantics of LEAVE
**                           and ENTER events and the ability of triggers
**                           to cancel the focus change by returning
**                           NO-APPLY. Inspected the existing requestFocus
**                           calls within this class. See details in 27728
** 141 NVS 20060703   @27804 Added static nextFocus() and previousFocus()
**                           methods which are variations of the focus
**                           change methods providing Progress semantics
**                           through LEAVE and ENTRY events. New methods
**                           are used throughout P2J widgets code.
** 142 NVS 20060705   @27810 Fixed: initial ENTRY event handling; null
**                           ScreenBuffer in trigger() that may be returned
**                           if trigger was not recognized on the server;
**                           getScreenBuffer() method wrong buffer size.
**                           ENABLE now maintains its own last focused
**                           widget in a variable.
** 143 NVS 20060705   @27831 Fixed a regression in waitForWorker() caused
**                           by the fix #142.
** 144 SIY 20060704   @27836 Use null instead of empty string to mark
**                           unused message line.
**                           Added checking of keys before passing them
**                           down to widgets.
**                           Added clearing of the messages before pausing
**                           new messages if needPause is set to false;
**                           Now CHOOSE does not change status line (match
**                           Progress behavior).
** 145 GES 20060706   @27863 getFrameRow() now returns a 1-based index.
** 146 GES 20060706   @27870 Explicitly disable mouse processing in
**                           CHARVA.
** 147 NVS 20060706   @27875 Fixed choose() method by defining a unique
**                           operation code for waitForWorker. Termination
**                           of CHOOSE now works properly paying attention
**                           to supplied GO-ON events.
** 148 GES 20060707   @27887 Support the implicit hide message that occurs
**                           during message processing while inside a
**                           trigger or some code invoked by a trigger.
**                           To accommodate this, we now track the nesting
**                           (recursion) level of triggers.
** 149 GES 20060707   @27908 Honor dynamic title expressions passed via
**                           a screen buffer.  Set the title only while
**                           the frame is hidden.
** 150 GES 20060710   @27931 Fix for hide processing during triggers.
** 151 SIY 20060707   @27930 Reset cursor position on view()/enable().
** 152 GES 20060711   @27986 Clear the current focus on disable.
** 153 NVS 20060713   @28028 All client exports that cause any drawing in
**                           UI now provide invalidation rectangle control
**                           in order to optimize drawing. On entry,
**                           invalidation is turned on. On exit, it is
**                           turned off and invalidation rectangle flushed
**                           although the invalidation mode may still be
**                           on (recursion in triggers).
** 154 GES 20060714   @28059 Apply processing must properly handle a null
**                           screen buffer for certain "current-window"
**                           operations to work properly.
** 155 SIY 20060719   @28146 Simulate input status in choose to fix help
**                           processing in CHOOSE. Added support for
**                           returning frame-value and other relevant data
**                           in waitFor().
** 156 GES 20060726   @28204 Moved interactions counter maintenance to
**                           the server.  The state sync processing
**                           ensures that the latest version is always
**                           available here.
** 157 NVS 20060724   @28215 Added popTrash() method that trashes all
**                           events in the current queue and then pops it
**                           off the stack. Applied popTrash() to the
**                           hideAll() export which now clears the screen
**                           directly. Added display control flags and
**                           used them in view(), scroll() and down().
**                           Flags control the frame decorations drawing
**                           and data iteration display for down frames in
**                           both interactive and redirected mode.
** 158 GES 20060728   @28221 Moved conditional up/down processing to be
**                           driven by the state synchronization process
**                           to avoid an extra trip to the client.
** 159 NVS 20060728   @28222 tinyInput used to leave focus control vars
**                           in a bad state. Fixed by resetting those vars
**                           inside waitForWorker instead of waitFor.
** 160 SIY 20060726   @28227 Complete rewrite of frame hiding algorithm.
**                           Many other minor changes.
** 161 GES 20060731   @28242 Deferred processing of conditional up/down
**                           until the piggybacked client export is
**                           invoked.  Otherwise any pausing that
**                           generates a condition exception will cause
**                           the client to abend instead of forwarding
**                           the exception back to the server. Added
**                           state sync for markFrames() and
**                           destroyFrame().
** 162 SIY 20060731   @28251 Frame cleanup changes.
** 163 NVS 20060731   @28261 Removed some unnecessary repainting related
**                           code from status(), which draws directly.
** 164 NVS 20060801   @28286 Removed even more unnecessary code from
**                           status().
** 165 NVS 20060801   @28292 Removed some unnecessary code around the call
**                           to leaveChoose().
** 166 NVS 20060802   @28332 statusInput() and statusDefault() exports
**                           use deferred sync to refresh the status line
**                           thus avoiding otherwise possible partial
**                           screen updates and flashing.
** 167 SIY 20060802   @28335 More careful resetting of frame in certain
**                           cases.
** 168 NVS 20060802   @28357 putScreen() is drawn directly now instead of
**                           causing the whole tree to redraw.
** 169 EVL 20060804   @28422 The method to hide/show cursor is completely
**                           implemented. The member cursorVisible is now
**                           added to remember the cursor status. Also
**                           when we need to explicitly put the cursor at
**                           some row,column position the cursor must be
**                           turned on if it not the case.
** 170 GES 20060807   @28447 Make the cursor visible during editing
**                           operations.
** 171 GES 20060807   @28450 Moved message list content processing to
**                           server to reduce traffic over the protocol.
** 172 SIY 20060807   @28452 Updated frame hiding algorithm and other
**                           changes.
** 173 GES 20060808   @28488 Shifted APIs to use Java primitives instead
**                           of the Progress compatible wrappers, where
**                           possible.  Moved the wrapper processing to
**                           the server from the client. Major fixes to
**                           messageBox() to properly handle initial focus
**                           and returning unknown value. Also added
**                           condition processing to the event handling.
** 174 GES 20060808   @28496 Use int instead of long for the backing
**                           methods that support the server exports.
** 175 SIY 20060810   @28520 Added handling for explicit setting/resetting
**                           of global canHide flag.
** 176 NVS 20060811   @28554 Enabling already visible frame avoids drawing
**                           the whole frame and limits drawing to the
**                           affected widgets.
** 177 NVS 20060814   @28569 Widgets losing focus now receive a draw()
**                           call right after FOCUS_LOST event to be able
**                           to redraw their state correctly.
** 178 NVS 20060816   @28615 Fixed doubled output in redirected mode by
**                           eliminating paint event brackets in
**                           placeFrameInt() method.
** 179 NVS 20060817   @28668 Fixed headers display in down frames. Headers
**                           are drawn directly now.
** 180 GES 20060818   @28669 Editing blocks must set the focus to the
**                           widget which was deferred by the previous
**                           enable.
** 181 NVS 20060817   @28670 Headers in down frames are redrawn only when
**                           they are changed.
** 182 GES 20060821   @28673 putScreen during editing mode must leave the
**                           cursor where it was previously.
** 183 GES 20060821   @28695 Honor a special case where the apply stmt
**                           raises a stop condition. Added support for
**                           overlong message lines.
** 184 NVS 20060822   @28733 Relocated default Window creation call from
**                           the constructor to the getInstance() to avoid
**                           a circular dependency after fixes to Window.
** 185 GES 20060823   @28755 status(), message() and hideMessage() during
**                           editing mode must leave the cursor where it
**                           was previously.
** 186 GES 20060824   @28875 Make apply() safe to call with a -1 widget
**                           ID when there is no frame and/or no focused
**                           widget.
** 187 GES 20060825   @28917 Honor GO in alert box event processing. Other
**                           conditions raised will not throw an exception
**                           but instead will cause a cancel return.
** 188 NVS 20060828   @28933 Implemented direct drawing in view() when it
**                           performs display (not refresh) function.
**                           Eliminated unnecessary drawing of specific
**                           widgets in case when the whole frame is
**                           redrawn due to layout change caused by widget
**                           width change caused by new data size.
** 189 NVS 20060829   @28955 Fixed latent bugs (lack of sync) in message()
**                           and status() methods which have been masked
**                           by the keyboard thread syncing.
** 190 NVS 20060830   @29003 Fixed dialog frame painting issue caused by
**                           the dialog's inability to move to the top of
**                           Z-order.
** 191 GES 20060830   @29004 More correct implementation of H183 which
**                           should have been in readKey instead of apply.
**                           Improved resetTerminal to implement a
**                           hideAll to clear the screen.  hideAll is
**                           what sets the hadReset flag (even when a
**                           resetTerminal is not called, but the hideAll
**                           is called directly by the business logic).
**                           Apply is a NOP now in editing mode when a
**                           reset has occurred.
** 192 GES 20060830   @29012 Fixed cursor save/restore processing to be
**                           more reliable by forcing the physical cursor
**                           to be updated before changing visibility.
** 193 NVS 20060830   @29023 enable() client export incorporates view()
**                           export functionality internally. Now it's
**                           true literally, since both methods call the
**                           shared viewWorker() method.
** 194 NVS 20060831   @29076 direct drawing of frame widgets performed in
**                           the view() method and frame headers performed
**                           in setScreenBuffer() method use new methods
**                           to control clipping rectangle. These new
**                           methods take into account the scroll viewport
** 195 NVS 20060901   @29163 setScreenBuffer method filters the character
**                           data coming from the server before assigning
**                           them to widgets. Filtering consists of looking
**                           for non-displayable characters with their
**                           replacement with spaces. Currently, all
**                           characters above \u007f and below \o0020,
**                           except '\t' and '\n', are considered
**                           non-displayable.
** 196 SIY 20060901   @29204 Support for entering data in down frames.
** 197 SIY 20060904   @29212 Fixed focus setting after application try to
**                           set focus to inactive widget.
** 198 NVS 20060905   @29231 MESSAGE UPDATE now truncates the message text
**                           if it's too long to be accommodated on a
**                           single message line together with a spacer
**                           and a FILL-IN. Modified methods are message()
**                           and tinyInput().
** 199 NVS 20060911   @29444 pushScreenDefinition() method needed brackets
**                           of setInvalidate() in order to handle the
**                           dynamic widget color changes properly.
** 200 GES 20060915   @29621 Honor a screen buffer value that is passed
**                           with an associated format string.
** 201 GES 20060921   @29923 Fix for cases where on GO, a widget fails
**                           validation and the focus was not properly
**                           reset into the field.
** 202 NVS 20060926   @29931 Added a variation of honorConditionalUpDown()
**                           method called from view() and underline().
**                           It reacts to and clears the ALL CLEAR flag
**                           of the frame. This flag controls whether the
**                           conditional UPs and DOWNs are to be honored.
** 203 NVS 20060927   @29995 pause(0) calls may be issued by the server
**                           and they are now handled properly (pending
**                           pause gets cancelled).
** 204 NVS 20060928   @29999 Fixed a regression caused by #202. Under some
**                           conditions allClear flag stayed set where it
**                           should have been cleared.
** 205 SIY 20060928   @30024 nextFocus/previousFocus should take into
**                           account TEXT group mode.
** 206 NVS 20061002   @30063 Fixed setScreenBuffer() which now bypasses
**                           static text widgets if setting UNINITIALIZED.
**                           This change fixes disappearing labels.
** 207 NVS 20061005   @30162 Refactored some common code into methods and
**                           added implementation of refreshFrameWidget().
** 208 GES 20061012   @30366 Fixed small performance regression caused by
**                           H207.
** 209 GES 20061013   @30395 Eliminated extra drawing in redirected mode
**                           during setScreenBuffer().
** 210 GES 20061017   @30561 Modifications to allow GO, ENDKEY, END-ERROR,
**                           ERROR and STOP to be honored as triggers.
** 211 NVS 20061017   @30468 Implemented ability to restart the session
**                           without exiting the process.
** 212 GES 20061020   @30560 Honor CTRL-C in suspendedPause().
** 213 NVS 20061020   @30586 Implemented recovery of frame's rectangle
**                           affected by overlay frames as well as hiding
**                           those overlay frames which are no longer in
**                           use.
** 214 GES 20061027   @30781 Added additional "marking" of a frame as
**                           no longer in scope during destroyFrame().
** 215 SIY 20061030   @30784 Added pausing after DOWN on non-DOWN frames.
** 216 GES 20061030   @30786 consume (NO-APPLY) and isGoPending flags
**                           are now safe for recursive processing.
** 217 GES 20061030   @30798 Save and restore the currently focused
**                           widget in tinyInput().  This is more
**                           correct in recursive processing. WARNING:
**                           reentrant calls to waitFor() are still
**                           broken.  An attempt to fix this was made by
**                           saving/restoring in enable/disable but this
**                           did not work properly so it was backed out.
** 218 GES 20061031   @30810 hideMessage() must be a NOP if the terminal
**                           is redirected.
** 219 SIY 20061031   @30826 Do not pause on hidden frames.
** 220 SIY 20061101   @30880 Do not apply DOWN to hidden 1-DOWN frames.
** 221 GES 20061102   @30894 Added forceDeregisterPageElement() and
**                           page element cleanup processing that does
**                           not depend upon being called while the
**                           terminal is redirected.
** 222 SIY 20061105   @30980 Changed signatures of Browse related methods.
**                           Added refreshStartRow() method.
** 223 GES 20061107   @31005 Fix to allow the drop-down box of a combo-box
**                           to ignore normal conditions/go events and to
**                           simply close the drop-down in END-ERROR.
** 224 GES 20061109   @31056 Added safety code.
** 225 SIY 20061110   @31066 Reset cursor in down frame even if no widgets
**                           are updated.
** 226 GES 20061113   @31103 Only filter unprintable chars when the
**                           terminal is interactive (not redirected).
** 227 GES 20061113   @31120 Fix for null pointer exception during page
**                           element deregistration.
** 228 GES 20061114   @31125 Enable apply to update the list of the
**                           currently active triggers.  This is
**                           especially important in the case where apply
**                           is used outside of a wait-for.  In addition,
**                           explicit widget IDs should take precedence
**                           over the session ID in checkTriggers().
** 229 GES 20061115   @31173 Provide direct access to the stream to which
**                           the terminal is redirected.  Reworked view
**                           to generate output multiple times if we are
**                           redirected and the stream had to reject the
**                           output due to crossing a page boundary.
** 230 NVS 20061115   @31200 setScreenBuffer() shouldn't attempt direct
**                           drawing on a hidden frame.
** 231 NVS 20061116   @31221 messageBox() had problems with changing focus
**                           to and from it and used to leave the cursor
**                           invisible.
** 232 NVS 20061117   @31255 Fixed NPE (regression after H231) in focus
**                           switch in messageBox(). Don't try to switch
**                           focus where there is no focused widget!
** 233 GES 20061117   @31265 Deregistration of page elements is safe to
**                           call at any time now (a requirement since it
**                           is called from hide processing which must
**                           also act this way).
** 234 NVS 20061117   @31266 Applied a better focus change technique so it
**                           makes the H231 fix better and also eliminates
**                           the problem with H232.
** 235 SIY 20061117   @31291 Updated for support of dynamic row/column.
** 236 GES 20061121   @31332 Added processing of dead frames during frame
**                           placement.  This eliminates a potential
**                           memory leak (frames are not otherwise
**                           removed from the client).
** 237 GES 20061121   @31342 Migrated recent changes to checkTriggers()
**                           into checkBreak() which needed to operate
**                           in the same manner.
** 238 GES 20061127   @31412 Don't mark a frame as out of scope if it is
**                           a shared frame that is exiting from a more
**                           deeply nested scope. To do this a shared
**                           flag was added to destroyFrame().
** 239 NVS 20061129   @31476 Reset down frame contents and cursor on OPEN
**                           SCOPE notification.
** 240 NVS 20061130   @31502 Full frame refresh was not always performed
**                           when required.
** 241 NVS 20061130   @31511 stopEditingMode() didn't reset currentWait-
**                           List variable causing malfunctioning of some
**                           triggers after editing block completion.
** 242 NVS 20061130   @31512 Fixed a painting regression caused by the fix
**                           H240.
** 243 NVS 20061204   @31544 Reimplemented LEAVE and ENTRY event handling
**                           model to make APPLY ENTRY work correctly as
**                           well as any triggers associated with LEAVE or
**                           ENTRY regardless of the event origin.
** 244 GES 20061206   @31597 Added promptFor() helper to provide a higher
**                           performance interface than multiple
**                           independent calls (enable then wait-for...)
**                           could provide.
** 245 SIY 20061207   @31611 Fixed handling of the nested waitFor() calls.
** 246 SIY 20061208   @31626 Deeper preservation of state during trigger
**                           invocation.
** 247 SIY 20061208   @31630 Retain cursor state during trigger
**                           invocation.
** 248 GES 20061208   @31637 Made the cursor visibility change and restore
**                           unconditional on every entry and exit from
**                           input mode (as opposed to only doing this on
**                           the "top" level input mode). In the case of
**                           multiple nested input mode entries, the
**                           original cursor visibility is stored in a
**                           stack, allowing the proper restore at every
**                           exit from a nested input mode, including at
**                           the top level.
** 249 GES 20061211   @31662 Added versions of client exports which take
**                           a stream ID and which implement an inline
**                           switch of the output destination to that ID
**                           and then a restoration of the previous
**                           destination back before returning.  This
**                           eliminates a pair of bracketed switchOutput()
**                           calls from the server in these cases. The
**                           objective is to improve performance of
**                           redirected output.
** 250 SIY 20061214   @31679 Force cursor position before waiting for user
**                           input.
** 251 NVS 20061214   @31695 Fixed viewWorker() to properly draw
**                           all uninitialized widgets of the current
**                           iteration of the down frame. Fixed
**                           setScreenBuffer() that used to reset
**                           uninitialized state of widgets excessively.
** 252 NVS 20070111   @31789 Fixed viewWorker() to show the frame if
**                           it's hidden by the time the method is called
**                           to display a widget in it.
** 253 GES 20070110   @31805 Eliminated unnecessary interface, simplified
**                           initialization and moved UI-specific init
**                           here from non-UI packages.
** 254 SIY 20070116   @31841 Triggering removal of already hidden out of
**                           scope frames.
** 255 SIY 20070176   @31847 Remove frame from the component list to
**                           prevent ID conflicts.
** 256 NVS 20070118   @31891 Using new trigger registry management based
**                           on EventList class. WaitList class is no
**                           longer in use.
** 257 SIY 20070119   @31898 Removed separate handling of shared frames.
** 258 SIY 20070122   @31908 Call setScreenBuffer() AFTER placing frame on
**                           the screen.
** 259 SIY 20070125   @31970 Removed unused removeFrame() method.
** 260 SIY 20070126   @32018 Fixed drawing of frame title.
** 261 SIY 20070202   @32068 Fixed access to missing frame in
**                           honorConditionalUpDown(boolean).
** 262 SIY 20070215   @32132 Prevent accessing non-existent frame.
** 263 SIY 20070221   @32200 Fixed handling of pause in hide().
** 264 SIY 20070301   @32280 Changed handling of frame removal.
** 265 NVS 20070308   @32293 Added detailed description of the process
**                           launch method since its semantics have
**                           changed.
** 266 SIY 20070309   @32383 Added special handling of frame hiding in
**                           promptFor. Removed unnecessary call to
**                           cleanupDeadFrames.
** 267 SIY 20070314   @32393 Added support for frame activeness tracking.
** 268 NVS 20070328   @32633 help() request from the client should report
**                           the lastKey variable.
** 269 SIY 20070328   @32643 Fixed handling of focus/frame activation in
**                           triggers. Pause flag in frames now is reset
**                           after each input key in order to properly
**                           handle new frames activated inside waitFor()
**                           loop from triggers.
** 270 SIY 20070329   @32657 Removed temporary workaround.
** 271 NVS 20070329   @32667 Blocking recursive HELP calls. The terminal
**                           beeps instead.
** 272 NVS 20070404   @32763 Rewrote help() method implementation. Now it
**                           preserves frames visibility, hides frames
**                           created by help code and clears the messages
**                           (excluded the message that is part of INPUT).
** 273 NVS 20070417   @33093 Universal event recognition logic moved to
**                           EventList class.
** 274 NVS 20070417   @33257 Fixed tab order processing in enable() method
**                           ignoring KEEP-TAB-ORDER attribute of frames.
**                           Fixed message cleanup in recursive triggers.
**                           Fixed focus restoration after triggers which
**                           create frames.
**                           Added new exports: a variation of enable()
**                           and isEnabled() in support of SENSITIVE
**                           attribute, nextTabItem().
** 275 SIY 20070430   @33295 Added handling for some special cases for
**                           getActiveFrame().
** 276 SIY 20070504   @33421 Updated getRowCount().
** 277 NVS 20070523   @33725 Fixed redirected output of DOWNs. down()
**                           method maintains a counter and data iteration
**                           display produces the required number of empty
**                           lines. Conditional DOWNs don't count.
** 278 SIY 20070517   @33748 New focus handling approach.
**                           Fixed down frame painting after scroll().
** 279 SIY 20070527   @33781 Post FOCUS_GAINED when widget is forced to
**                           be focused. Fixed case when first enabled
**                           widget is not focus traversable.
** 280 SIY 20070528   @33784 Fixed handling of status line and message
**                           clearing after choose processing.
** 281 SIY 20070531   @33883 More focus and messagBox() fixes.
** 282 EVL 20070609   @34005 Adding explicit import of the class
**                           com.goldencode.p2j.net.Queue to eliminate
**                           conflict with the same class from java.util
**                           package to be able to compile for Java 6
** 283 SIY 20070611   @34007 Rolled back overlay frame handling in frame
**                           hiding. It does cause regression.
** 284 NVS 20070611   @34009 Fixed deregistration of page elements. Fixed
**                           the page headers output, when it's caused by
**                           drawing a frame having no visible widgets.
** 285 NVS 20070614   @34034 Fixed output of page top frames produced by
**                           DISPLAY statement. They produced duplicates
**                           every time the frame was the first one on the
**                           page.
** 286 SIY 20070614   @34038 Fixed focus handling for ComboBox.
** 287 NVS 20070618   @34082 Standardized event processing in
**                           refreshStartRow() export.
** 288 SIY 20070619   @34097 Fixed pausing in rare case of iterated single
**                           frame display without explicit enter/pause.
** 289 SIY 20070622   @34203 Separate flag for message pausing handling.
** 290 SIY 20070706   @34389 Fixed restoring focus after showing frame.
**                           Fixed processing of nested event lists.
** 291 SIY 20070712   @34460 Fixed CLEAR/CLEAR ALL for down frames.
** 292 NVS 20070716   @34541 Massive changes to the redirected output mode
**                           applied to multiple methods, dictated by the
**                           need to support partial displays and other
**                           features.
** 293 SIY 20070717   @34560 Finally cleaned up #283. Do not print stack
**                           trace for IllegalComponentStateException.
** 294 NVS 20070724   @34674 Another take on redirected output mode. All
**                           output remains buffered until the target
**                           frame changes. At this moment the buffer is
**                           flushed.
** 295 NVS 20070730   @34735 Fixed down frame headers output in redirected
**                           mode and blocked extra output for page top
**                           frames caused by their false visibility
**                           status.
** 296 NVS 20070801   @34749 Added notifyAdvance() method. It gets called
**                           whenever page advance is performed on a
**                           stream. The purpose is to flush the terminal
**                           buffer.
** 297 SIY 20070802   @34770 Fixed SET STATUS DEFAULT behavior inside input
**                           loop. Ref: 34768
** 298 NVS 20070802   @34779 Added notifyFlush() method. It gets called
**                           whenever flush operation is performed on a
**                           stream. The purpose is to flush the terminal
**                           buffer.
** 299 NVS 20070807   @34798 DOWN for a non-down frame in redirected mode
**                           is a meaningful operation. It flushes the
**                           terminal buffer and inserts the required
**                           number of new lines into the stream.
** 300 SIY 20070807   @34812 Handling for width of redirected output
**                           terminal. Some refactoring and cleanup.
** 301 NVS 20070808   @34819 Temporary redirection is provided when needed
**                           in notifyClose(), notifyAdvance() and
**                           notifyFlush() methods.
** 302 NVS 20070808   @34820 Fixed double display of page element frames
**                           that are DISPLAYed vs VIEWed by implementing
**                           deferred registration of the frame as page
**                           element. Registration should not happen until
**                           the DISPLAY output is flushed.
** 303 NVS 20070808   @34821 Deferred registration should be done for page
**                           top elements only.
** 304 NVS 20070809   @34830 Fixed forced deregistration of page elements
**                           for frames going out of scope.
** 305 NVS 20070810   @34839 Fixed pause() method. In redirected mode,
**                           PAUSE MESSAGE should temporarily switch to
**                           the interactive terminal to display the
**                           message.
** 306 SIY 20070814   @34845 Track frame redirection.
** 307 NVS 20070815   @34851 Frame visibility in redirected mode is
**                           controlled by setVisibilityNoPaint() method.
**                           Frames are set to visible only for the time
**                           of the view() call. Only the first view()
**                           call in a chain of calls for the same frame
**                           makes the frame visible in the traditional
**                           way, which is required for the frame to draw
**                           decorations and be placed properly.
** 308 NVS 20070816   @34852 Page advance listener should flush redirected
**                           terminal *and* reset the owning frame ID.
** 309 NVS 20070817   @34881 Quick return for no-op call of redirectOutput
** 310 NVS 20070829   @34929 Fixes to the conditional down() processing.
** 311 NVS 20070830   @34977 down() calls to the client now pass the flag
**                           that tells conditional downs. Down exports
**                           reimplemented.
**                           Numerous changes to the view() and down()
**                           exports to support stream switching for
**                           frames.
** 312 SIY 20070904   @34940 Fixed status message drawing. Ref: #34941.
** 313 NVS 20070906   @34993 switchStreams() should bypass flushing to a
**                           stream which is no longer around.
** 314 NVS 20070907   @35006 switchStreams() should be called from the
**                           promptFor() export as well.
** 315 NVS 20070907   @35013 Clearing of streamed frames switching back to
**                           interactive terminal is done via a new method
**                           clearCurrent() in FrameHelper.
** 316 NVS 20070925   @35217 Redirected frames are marked as visible while
**                           terminal is redirected to avoid duplicates of
**                           frame headers or partially drawn headers.
**                           Hiding and showing of the redirected frames
**                           is done when switching to and from the
**                           redirected mode in Window.java
** 317 ECF 20070928   @35271 Added browse-specific event dispatching. We
**                           must give Browse an opportunity to handle
**                           (and consume) END-ERROR event under certain
**                           conditions. Browse has an internal undo mode
**                           for edits, which is triggered by END-ERROR.
** 318 NVS 20071001   @35310 Down may cause page break. The corresponding
**                           listener makes the frame invisible so that
**                           the very first use of the frame on the new
**                           page redraws the frame decorations. But the
**                           same event caused also the current iteration
**                           of the down frame to be redrawn since the
**                           frame wasn't cleared yet. Now the frame gets
**                           cleared before flushing the buffers.
** 319 NVS 20071011   @35431 Pausing of streamed frames in interactive
**                           mode is implemented. Minor fix to downWorker.
** 320 NVS 20071025   @35522 Added new method getFrameIdForStream() which
**                           searches the list of frames for one connected
**                           to the specified stream. Added new method
**                           drawFrameHeaders() which outputs frame
**                           headers for the last frame used with the
**                           specified stream.
** 321 EVL 20071025   @35514 Adding switchOutput(null) to restore output
**                           device to be the real terminal for the case
**                           we need restore from Ctrl-C sequence (#35515).
** 322 SIY 20071105   @35712 Support for correct SKIP handling in
**                           redirected mode.
** 323 NVS 20071107   @35725 HELP event processing should not take any
**                           precedence over exit conditions or GO-ON
**                           events.
** 324 CA  20071109   @35715 Removed the status restoration in method
**                           invokeTrigger.
** 325 NVS 20071109   @35765 Fixed a NPE in the view() method that could
**                           be raised in redirected mode when the stream
**                           gets finalized.
** 326 NVS 20071113   @35801 Prevented drawing of the current row when
**                           drawing the frame headers after page break.
** 327 SIY 20071116   @35822 Explicit deregistration of page element
**                           must be more thorough to avoid reappearing of
**                           frame in the output. Some minor refactoring
**                           and cleanup.
**                           Ref: #35800.
** 328 NVS 20071116   @35830 drawFrameHeaders() bypasses frames that are
**                           registered as page elements.
** 329 SVL 20071120   @35844 Previously displayed status messages hide
**                           even if PAUSE 0 was called.
** 330 SVL 20071120   @35846 The information about last focused element
**                           is saved on frame deactivation in focusId.
** 331 ECF 20071120   @35913 Change to match net package refactoring.
**                           Replaced uses of Queue with Session.
** 332 SIY 20071120   @35917 Do not redraw regular frames on page break.
**                           Ref: #35824
** 333 NVS 20071121   @35925 Merging in multiple parallel updates.
** 334 SIY 20071122   @35958 Fixed redundant pause on frame down() when
**                           redirection is not active anymore.
**                           Ref: #35957
** 335 CA  20071126   @36016 Down statements with down frames redirected
**                           and paged which land outside current page
**                           will add NL's to fill this page and only one
**                           NL in the next page, after the frame header.
** 336 SIY 20071126   @36023 Less aggressive resetting pause in hide().
**                           Ref: #36022
** 337 CA  20071127   @36032 Down statements with redirected, paged,
**                           non-boxed and down frames which land outside
**                           current page will add only one NL to the
**                           second page, after the header. Ref: #36031
** 338 SIY 20071126   @36036 Added support for frame DOWN maintenance
**                           using StateSynchronizer.
**                           Ref: #36034
** 339 SVL 20071128   @36044 When frame is enabled, the focusId cannot be
**                           set to a widget that is missing in the
**                           visible widgets list.
** 340 SIY 20071128   @36057 Restored lost #322 fix.
** 341 SVL 20071206   @36228 screenCurrentInitial tab order restored when
**                           ENABLE ALL is called for frame with no
**                           keep-tab-order specified.
** 342 GES 20071207   @36261 Added safety code to getChanges() to allow
**                           it to be called when the UI has never been
**                           used (may be the case with batch processes).
** 343 SVL 20071207   @36255 Corrections to #339 fix - all enable widgets
**                           on frame are considered.
** 344 NVS 20071212   @36318 view() call for DISPLAY now tries to freeze
**                           the list of page headers for the target
**                           stream in redirected mode.
** 345 GES 20071213   @36331 Copy the frame line into the screen buffer
**                           whenever a screen buffer is returned or
**                           passed to the server. This allows frame line
**                           requests on the server to be serviced without
**                           a call to the client (improves performance).
**                           It also allows the screen buffer caching
**                           implementation to be kept in sync with
**                           iteration changes that happen only on the
**                           client side (e.g. during choose).
** 346 GES 20071213   @36340 During CHOOSE, bypass default processing of
**                           an event when a match to a GO-ON event is
**                           found.
** 347 SVL 20071214   @36356 focus events in waitForWorker() are processed
**                           immediately
** 348 GES 20071214   @36367 Only send frame-line updates in the screen
**                           buffer if the frame is a down frame and it
**                           is not currently redirected. Frame-line in
**                           these non-down or redirected cases is a
**                           constant and doesn't need to be sent.
** 349 GES 20071220   @36507 Safety code to rebalance invalidation state
**                           when down worker causes a pause and the
**                           user presses end-error.
** 350 GES 20071227   @36589 Focus changes caused by triggers are honored
**                           just after the trigger returns to ensure
**                           that subsequent event processing (for the
**                           event that caused the trigger) will "see"
**                           the focus changes.
** 351 SVL 20071225   @36568 focusable variable in enable() is now
**                           assigned correctly in case widgetId==null
** 352 SVL 20071221   @36538 focusChange() and tryFocusChange() functions
**                           do not process Progress leave/entry events
**                           if focus remains on the same widget.
** 353 CA  20071219   @36544 The downWorker method must freeze the
**                           headers before making the frame visible.
**                           The notifyClose method must freeze the
**                           headers before sending the data to the stream
**                           The viewWorker must freeze the headers if the
**                           frames have switched.
** 354 NVS 20071229   @36614 view_() method implements new redirected mode
**                           logic as described in the method javadoc.
**                           Removed some obsolete code: drawFrameHeaders
**                           and getFrameIdForStream methods.
** 355 GES 20080103   @36661 Eliminated use of freezeHeaders(). This was
**                           an invalid approach to processing page
**                           headers. It has been replaced with code in
**                           the redirected terminal to force page headers
**                           on first output.
** 356 SIY 20080107   @36701 Fixed frame displaying when its size is
**                           changed by form headers.
**                           Ref: #36700
** 357 NVS 20080110   @36770 Fixed viewWorker() method. Down frames getting
**                           continuing output are now drawn properly and
**                           optimally in redirected mode, using
**                           repaintCurrent() helper method in FrameHelper
** 358 NVS 20080114   @36805 Delayed registration of page elements made by
**                           DISPLAY statement is no longer needed with
**                           the new implementation of the redirected mode
**                           output. Registration is done right in the
**                           view() export for both page top and page
**                           bottom elements.
** 359 SIY 20080115   @36813 Missing pause on hideAll()
** 360 SIY 20080117   @36820 Fixed centering in redirected mode when
**                           terminal width is other than 80 columns.
** 361 GES 20080116   @36828 downWorker in redirected mode was completely
**                           reworked to cleanup problems and to properly
**                           handle cases where the down crosses a page
**                           boundary. view_() has bugs fixed in cases
**                           dealing with page boundaries. Stream flushing
**                           now bypasses page header processing when
**                           the flushed output is from the redirected
**                           terminal. This avoids duplicate page headers
**                           since the first write to a fresh redirected
**                           terminal already handles page headers.
**                           Backed out H350 since this causes a stack
**                           overflow in certain cases.
** 362 SIY 20080121   @36874 Support for postponed event processing.
**                           Ref: #36871
** 363 CA  20080123   @36898 For shared frames, the nesting level is
**                           maintained on client too. When the frame
**                           closes it's root nesting level or performs a
**                           down on its root nesting level the flag
**                           delayedRepaint will be set to true. On the
**                           next display/view stmt, viewWorker() will
**                           trigger a repaint if delayedRepaint is true.
**                           Ref: #36895
** 364 SIY 20080121   @36933 Do not clear redirected terminal to prevent
**                           unexpected clearing of terminal. Otherwise
**                           after launch of application first operation
**                           which updates part of screen will trigger
**                           wiping out of rest of screen.
**                           Ref: #36932
** 365 NVS 20080212   @37034 Fixed launching of interactive processes.
**                           Now NCURSES is properly suspended and resumed
** 366 ECF 20080212   @37036 Added getSelectedRowIndex(). Gets the row
**                           index of a selected row in a browse widget.
**                           Delegates to Browse.
** 367 SIY 20080207   @36971 Paging should not be triggered if there are
**                           still some space at the page and non-down
**                           frame is displayed unless frame is larger
**                           than the page itself.
**                           Ref: #36968
** 368 SIY 20080211   @36988 Repaint terminal after resume() if
**                           redirection is not active.
**                           Ref: #36987
** 369 SIY 20080218   @37114 Simplified redirected mode frame layout and
**                           size logic.
**                           Ref: #37107
** 370 SIY 20080218   @37139 Rolled back #367.
** 371 NVS 20080221   @37143  hideAll() export should switch temporarily to
**                           interactive mode if redirected, since it acts
**                           on the contents of the interactive terminal
**                           only.
** 372 CA  20080303   @37234 If redirected to an unnamed stream,
**                           message stmt sends the data to the stream too;
**                           else, the message is sent only to the terminal.
**                           Ref: #37225
** 373 SIY 20080304   @37292 Restore Z-order after displaying page headers.
**                           Ref: #37288
** 374 EVL 20080306   @37312 The insert mode indicator should be repainted
**                           as a part of update - promptFor execution.
** 375 SIY 20080305   @37317 Reset start row and trigger header processing
**                           at the beginning of the page.
**                           Ref: #37313
** 376 NVS 20080307   @37412 Added setLocation() for dynamic ROW and COL
**                           widget attributes support. For now this is a
**                           placeholder. The implementation will be added
**                           when an application shows the need for it.
** 377 SIY 20080310   @37422 Tell Screen.hideFrames() the mode (redirected
**                           or interactive) to let it choose frames for
**                           hiding.
**                           Ref: #37420
** 378 SIY 20080317   @37458 Added helper methods for terminal stream.
**                           Ref: #37453
** 379 SIY 20080320   @37607 Fixed handling of the SCREEN-VALUE and data
**                           value.
** 380 NVS 20080320   @37622 A major overhauling of the focus management
**                           with changes to enable/disable, next-prompt,
**                           wait-for and choose implementations. New
**                           implementation is based on the findings
**                           described in the "Progress Focus Management"
**                           document.
** 381 SIY 20080321   @37625 Postpone generation of conditions until all
**                           event lists are checked.
** 382 SIY 20080326   @37682 Fixed handling of pausing in messages.
** 383 NVS 20080326   @37688 Finished overhauling of the focus management
**                           for editing blocks. Allows indirect recursion
**                           of editing blocks through WAIT-FOR.
** 384 NVS 20080327   @37694 Target widget for keystrokes made when there
**                           is no valid focus is CURRENT-WINDOW now.
**                           This fixes the trigger lookup issue
** 385 GES 20080327   @37700 Removed filtering code from setChangedValue
**                           method. This was needed when the CHARVA layer
**                           did not properly handle the extended ASCII
**                           range.  Now that that range is properly
**                           handled, the filtering logic is actually
**                           harmful since it removes any I18N characters
**                           that use the range above 7-bit ASCII.
** 386 SIY 20080328   @37723 Send LEAVE message at the end of WAIT-FOR.
** 387 NVS 20080328   @37727 List of displayed frames is no longer
**                           maintained.
** 388 SIY 20080331   @37735 Improved #386 - event should be generated only
**                           in limited number of cases.
**                           Ref: #37734
** 389 NVS 20080401   @37786 Fixed multiple inconsistencies when looking
**                           up frames for focusable widgets. Added widget
**                           filtering by focusability in enable() export.
** 390 NVS 20080402   @37790 Setting initial focus in waitForWorker() now
**                           takes into account dialog box option.
** 391 SIY 20080403   @37801 Pass ENTRY event to the widget which receives
**                           focus unless it is consumed by trigger.
** 392 NVS 20080403   @37827 Fixed sending the final LEAVE event on exit
**                           from wait-for. The event list was restored
**                           too early.
** 393 SIY 20080404   @37844 Fixed initialization of editing mode for
**                           down frames.
** 394 SIY 20080408   @37870 Flush stream when switching from non-terminal
**                           to terminal stream.
**                           Ref: #37868
** 395 SIY 20080409   @37885 Fixed NPE when widget which should receive
**                           focus has ho parent.
**                           Ref: #37887
** 396 SIY 20080410   @37900 Fixed "sticky focus" behavior to limit it to
**                           cases of source FILL-IN widget with events
**                           other than VALUE-CHANGED.
**                           Ref: #37882
** 397 SIY 20080414   @37975 Added complete ENTRY generation/processing
**                           on repeated activation of the same widget.
**                           Ref: #37973
** 398 NVS 20080417   @38024 Eliminated excessive drawing when disabling
**                           displayed widgets.
** 399 GES 20080421   @38069 Added setFormat() client export.
** 400 NVS 20080423   @38099 Based on SIY's submission, when exiting event
**                           processing loop with a widget in focus, that
**                           widget is sent LEAVE event. Thus, LEAVE
**                           trigger, if any, gets its chance to run.
**                           NO-APPLY return is honored by consuming GO
**                           event.
**                           Ref: #38083
** 401 NVS 20080423   @38104 Major reimplementation of CHOOSE in field
**                           mode to support triggers and regular focus
**                           change.
** 402 NVS 20080428   @38126 Fixed few cursor positioning issues with the
**                           WaitForKey() method, which were responsible
**                           for a few recent regressions, including extra
**                           blank lines in redirected reports.
** 403 GES 20080501   @38189 Merged endOutput calls into other client
**                           calls to avoid the extra trip.
** 404 NVS 20080501   @38192 Fixed few not so recent regressions with UI
**                           found in customer application code.
**                           Fixed a focus restoration bug in
**                           waitForWorker().
** 405 NVS 20080502   @38197 Fixed NPE in WaitForWorker() when dealing
**                           with null ScreenBuffer.
** 406 NVS 20080502   @38198 Eliminated duplicate LEAVE events in both
**                           known cases: when setting initial focus in
**                           new wait-for and when GO generates LEAVE.
** 407 NVS 20080503   @38201 Based on the update submitted by SIY.
**                           Additional code cleanup for GO event handling
**                           Do not generate LEAVE on GO for WAIT-FOR.
**                           Moved code which generates LEAVE on GO before
**                           all calls to invokeTriggers() in order to
**                           preserve correct order of events as they are
**                           seen by triggers (must be LEAVE then GO, not
**                           GO then LEAVE). Also, if trigger on LEAVE
**                           returns NO-APPLY, there must be no GO event.
**                           GO event does not cause exit from WAIT-FOR if
**                           it is not in the list of events.
** 408 NVS 20080503   @38214 Based on the update submitted by CA. Extended
**                           LEAVE event generation on exit for any kind
**                           of exit condition, not simply GO.
** 409 CA  20080506   @38215 Fixed regression introduced by H403 (#38189):
**                           ThinClient.endOutput() must append a NL to
**                           the stream, if needed (i.e. must call stream.
**                           endOutput(true)).
** 410 CA  20080507   @38227 screenCurrent is initialized when the class
*                            is initialized. Ref: #38226
** 411 NVS 20080508   @38239 Removed checks like screenCurrent == null
**                           They are redundant since the H410 and should
**                           have been removed as part of H410 indeed.
** 412 SIY 20080509   @38248 Enable frame before staring event processing
**                           loop in choose() and restore state when
**                           processing is finished.
**                           Ref: #38247
** 413 SIY 20080509   @38249 Cleanup message area in hideMessage() if there
**                           was some content.
** 414 SIY 20080512   @38256 Allow Browse receive 4GL events even in
**                           disabled state.
** 415 NVS 20080512   @38261 Message box should not interfere with the
**                           focus management. Now current focus is
**                           preserved across the message box activity.
** 416 NVS 20080512   @38262 Help text should be saved and restored around
**                           trigger invocation.
** 417 SIY 20080516   @38288 If tinyInput() detects that message is too
**                           long, it should replace redundant characters
**                           with spaces, otherwise garbage left from full
**                           message may remain on screen.
** 418 SVL 20080520   @38333 hideAll() does nothing if the terminal is
**                           redirected.
** 419 SIY 20080522   @38369 Frame scope opening resets regular frames too.
**                           Ref: #38365
** 420 SIY 20080522   @38372 Fixed regression caused by #419.
** 421 NVS 20080528   @38428 Frame as a focus target of wait-for should be
**                           simply ignored and handled as if there were
**                           no focus clause.
** 422 SIY 20080527   @38404 Fixed missing pause and redundant message
**                           clearing when MESSAGE is invoked from two
**                           sequential triggers with no other intermediate
**                           interactive statements.
** 423 SIY 20080530   @38461 Enable/disable special handling of DELETE key
**                           for VT220 terminal.
**                           Ref: #38459
** 424 SIY 20080530   @38465 Call to AlertBox.finished() should be
**                           bracketed with push()/pop(), otherwise hiding
**                           is not handled properly.
**                           Ref: #38464
** 425 NVS 20080530   @38467 STATUS INPUT issued from an editing block
**                           should not override the field's help text.
** 426 SIY 20080530   @38470 Prepare ScreenBuffered before focus state is
**                           restored to get correct value for FRAME-FIELD
**                           and FRAME-VALUE (see Frame.getFrameValue()).
** 427 NVS 20080603   @38506 Eliminated unnecessary drawing causing PUT
**                           SCREEN output erasure in pushScreenDefinition.
** 428 NVS 20080605   @38586 Fix 427 should allow drawing for hidden frame
**                           that is being made visible through config
**                           change.
** 429 NVS 20080606   @38611 Fixed setHelp() method which now properly
**                           reverts the status text to the default text
**                           when called for a widget having no own help
**                           text.
** 430 SIY 20080609   @38651 Repaint areas which belong to frames whose
**                           Z-order is about to be changed.
** 431 SIY 20080610   @38674 Make #430 a bit less aggressive - repaint only
**                           frames which overlap current. Also, when there
**                           are overlapping frames, force frame placement
**                           since this affects pausing.
** 432 NVS 20080611   @38689 message() worker saves and restores cursor
**                           position always as it can be called for error
**                           message display internally.
**                           waitForWorker() undefines focus for high
**                           level operations if it was undefined on entry
** 433 NVS 20080611   @38711 Fixed focus issues after END-ERROR issued in
**                           previous frame.
** 434 NVS 20080613   @38763 Focus should be restored in cases of CHOOSE
**                           nested within WAIT-FOR or editing blocks.
**                           Added missing type-ahead buffer reset to
**                           pause().
** 435 NVS 20080527   @38403 Implemented conditional targets for keystroke
**                           which may be widget, frame or current window.
**                           Implemented LEAVE event for frames received
**                           ENTRY.
** 436 NVS 20080616   @38790 NEXT PROMPT should be ignored if it points
**                           outside of the widget list for current op.
** 437 NVS 20080619   @38873 Fixed help text display for CHOOSE in row
**                           mode. Fixed PUT SCREEN quirk for the status
**                           line: no output when it would touch either of
**                           the two rightmost character position (depends
**                           on the terminal window width).
** 438 GES 20080613   @38881 Shifted to using ClientState for state sync
**                           and added a hadPause flag to notify the
**                           server whenever an explicit or implicit
**                           pause occurs.  It is used to conditionally
**                           control infinite loop control in certain
**                           cases.
** 439 NVS 20080625   @38921 Fixed status line display in WAIT-FOR when
**                           there is no focused widget.
**                           CA: fixed a potential NPE in waitForWorker().
** 440 NVS 20080626   @38947 Backed out H416. Now waitForWorker() does a
**                           better job of managing focus and H416 causes
**                           a regression.
** 441 SIY 20080630   @38989 Protect existing frames from removal when help
**                           application is invoked.
**                           Ref: #38987
** 442 SIY 20080711   @39101 Use event source instead of current focus when
**                           making decision about consuming event.
**                           Otherwise trigger which activated/deactivated
**                           some other widget may prevent event from
**                           passing to source widget.
** 443 SIY 20080731   @39260 Added handling of pause for frames which are
**                           cleared because of repeated entering/leaving
**                           frame scope.
** 444 SIY 20080808   @39345 Added handling of forced frame refreshing
**                           in VIEW after interrupted pause on cursor move
**                           in DOWN frames.
** 445 NVS 20080812   @39372 Frames with editors in redirected mode are
**                           subjects of widget rearrangement before
**                           drawing and restoration right after it. This
**                           is required to allow editors output all of
**                           their contents regardless of the visible
**                           widget's height.
** 446 NVS 20080818   @39437 view() call may cause pausing due to frame
**                           hiding and F4 pressed while in that pause
**                           should be properly propagated to the business
**                           logic interrupting frame display. Fixed
**                           placeFrameInt so it bypasses frame display
**                           in case of END_ERROR_CONDITION and viewWorker
**                           to propagate the condition up. Based on CA's
**                           findings.
** 447 NVS 20080820   @39440 waitForWorker() saves and restores leaveSent
**                           flag so that it's preserved across nested
**                           waits. This guarantees that LEAVE event is
**                           sent no more than once per WAIT-FOR, even if
**                           caused by APPLY LEAVE from a business logic
**                           of a trigger.
** 448 SIY 20080826   @39573 Processing of message box should reset pause
**                           just like other kinds of pauses do.
** 449 NVS 20080826   @39598 Modified refreshScreen() method to use
**                           invalidation region technique instead of
**                           invalidation rectangle.
** 450 SIY 20080901   @39642 Added ANY-KEY event support.
** 451 SIY 20080903   @39697 Fixed order of checks in ANY-KEY support.
** 452 CA  20080904   @39749 WAIT-FOR will keep a list of frames which are
**                           waiting for events. When another frame is
**                           viewed, it will not hide the dialogs which
**                           are still waiting for evens. Added the getter
**                           for FRAME:VISIBLE attribute. If the frame was
**                           moved, we must hide the frame before applying
**                           the new configuration in pushScreenDefinition.
** 453 ECF 20080911   @39795 Partially rolled back #452 (@39749). The loop
**                           at the top of the finally block in the
**                           viewWorker() method was commented out in
**                           response to a ConcurrentModificationException
**                           in Window.draw(). Further investigation is
**                           needed.
** 454 SIY 20080915   @39810 Rethrow exception in viewWorker() regardless
**                           from the redrawing/refreshing modes.
** 455 CA  20080909   @39750 getFrameLine was changed to consider the
**                           pendingUpDown value.
** 456 SVL 20080914   @39801 Added lastActionCode field. Some changes in
**                           the insert mode logic.
** 457 NVS 20080917   @39842 While in editing block, readKey() should NOT
**                           reset "message needs pause" flag.
** 458 SVL 20080919   @39860 Auto-zap is reseted when the focus moves
**                           between fill-ins (in some cases).
** 459 NVS 20080922   @39902 hideFrame() for dialog frame checks if there
**                           is another dialog and reactivates it
**                           immediately. viewWorker() brings frames on
**                           top conditionally - dialog frame comes on top
**                           always whereas regular frame is placed below
**                           first visible dialog, if any.
** 460 SIY 20080918   @39900 Draw labels in side-labeled DOWN frame.
** 461 CA  20080926   @39953 Avoid duplicate LEAVE events to be sent by
**                           waitForWorker.
** 462 NVS 20080922   @39997 Fixed handling of interactive conditional
**                           DOWNs when previous output was redirected.
**                           Fixed unconditional DOWN N in redirected mode
**                           to produce the specified number of iterations
**                           not just the blank lines.
**                           Fixed forced page break to produce iteration
**                           not just a single blank line.
** 463 SVL 20081001   @39978 Call to frame.draw() removed from scroll(..).
** 464 SIY 20081003   @40018 Added support for MarkEntry.UNHIDE.
** 465 NVS 20081008   @40072 Partially backed out #462 due to regressions.
**                           Redone iteration based calculations for the
**                           number of blank lines in downWorker().
** 466 SIY 20081013   @40084 Do not sent LEAVE on exit from WAIT-FOR if
**                           exit was performed during processing of
**                           consumed condition (the condition was consumed
**                           but other (possibly artificial) event caused
**                           end of WAIT-FOR loop).
** 467 CA  20080930   @39981 Added accessors for FRAME:HEIGHT/WIDTH-CHARS
**                           FRAME:COL/ROW attributes.
** 468 NVS 20081015   @40093 Reimplemented conditional vs unconditional
**                           DOWN processing for redirected mode.
**                           Interactive conditional DOWN causes flushing
**                           of the previous stream and that affects the
**                           logic of the unconditional DOWN to the stream
**                           which may follow.
** 469 SIY 20081016   @40100 Slightly changed clearWorker() signature and
**                           allowed it pass through value returned by
**                           FrameHelper.clearWorker().
** 470 NVS 20081016   @40101 Fixed a regression caused by #468
**                           and the last known regression in redirected
**                           reports suite.
** 471 SIY 20081017   @40157 Fixed missing pause after breaking pause on
**                           frame clearing.
** 472 SIY 20081020   @40171 Improved #471 and now it also hides frames
**                           which can be hidden.
** 473 SIY 20081022   @40198 Do not clear messageNeedPause when hiding
**                           frame without pauses.
**                           Ref: #40196
** 474 SIY 20081023   @40204 Adopted changes in message text handling API,
**                           removed redundant bracketing.
** 475 CA  20081023   @40163 Fixed cursor reset when NEXT-PROMPT is used
**                           and message handling within a EDITING block.
**                           Modified by NVS to use continueEditing()
**                           export, which already existed.
** 476 CA  20081024   @40222 On a new client instance, clear the toolkit
**                           window list.
** 477 SIY 20081007   @40234 Added support for postponed cursor wrapping
**                           for DOWN frames.
**                           Ref: #40235
** 478 SIY 20081029   @40264 Do not clear status line when WAIT-FOR loop is
**                           interrupted by condition exception.
**                           Ref: #40263
** 479 SIY 20081031   @40283 Reworked #478 to avoid regressions.
** 480 SIY 20081031   @40285 In some cases, when focus target is frame with
**                           browse, we should choose browse as target.
**                           Ref: #40284
** 481 GES 20081106   @40360 Avoid treating editing blocks as a wait-for
**                           when processing END-ERROR even though the
**                           opcode is set as WAIT_FOR.
** 482 SIY 20081107   @40372 Replaced #480 with generic fix.
**                           Ref: #40367
** 483 SVL 20081107   @40391 getFrameLine() does not return 0 for hidden
**                           and redirected frames any more. Current
**                           iteration is cleared for non-down frames
**                           into downWorker().
** 484 CA  20081114   @40447 Changes due to new clipping implementation.
** 485 SIY 20081117   @40507 Do not redraw hidden widgets in enable().
**                           Hide overlay frames which intersects with one
**                           which starts the input loop.
** 486 SIY 20081120   @40598 Fixed regression caused by #457. Fixed issue
**                           for which #457 was introduced (actual cause -
**                           difference in message auto-clearing behavior
**                           inside EDITING block).
**                           Ref: #40597
** 487 CA  20081124   @40684 Only frames marked with canHide=true can hide
**                           other frames, on a VIEW/DISPLAY call.
** 488 SIY 20081125   @40693 Added missing drawing bracket on early exit in
**                           waitForWorker().
** 489 NVS 20081126   @40698 Added a state flag that tracks the duration
**                           of the viewWorker() method activity. Required
**                           for fixing temp text deletion when frame
**                           hiding is caused by view.
** 490 SIY 20081127   @40709 Fixed more focused frame Z-order issues
**                           (related to second part of #485), common code
**                           is refactored in separate method.
** 491 CA  20081127   @40706 Modified H487 to use the 'canHideOthers' flag;
**                           this flags tracks if a certain frame can hide
**                           other frames, on its first DISPLAY/VIEW call.
** 492 SVL 20081127   @40713 lastActionCode in not modified if an alert
**                           box is displayed.
** 493 CA  20081128   @40717 After a WAIT-FOR time-out, no leave event will
**                           be sent.
** 494 SIY 20081201   @40755 Fixed pausing regressions caused by changed
**                           'can hide' flag maintenance and frame hiding
**                           behavior.
** 495 CA  20081203   @40758 Browse widgets will clear the temp-texts from
**                           the message area when they are viewed or
**                           repositioned.
** 496 SIY 20081203   @40771 Set widgets values early if possible. This is
**                           necessary for correct frame size calculation.
**                           Ref: #40772.
** 497 NVS 20081204   @40789 CHOOSE processing shouldn't try and restore
**                           focus when exiting event processing loop in
**                           waitForWorker().
**                           When iterating editing blocks, focus change
**                           may cause needed events, so now it's enclosed
**                           in push()/pop(). It restores the HELP string
**                           after triggers correctly.
** 498 CA  20081205   @40798 Fixed regression caused by H495: frames which
**                           intersect the message area need to be redrawn,
**                           if temp-text was hidden.
** 499 NVS 20081205   @40809 Standardized the use of getAndResetBitmap()
**                           and getAndSetBitmap().
** 500 SIY 20081205   @40804 Partially backed out #494.
** 501 CA  20081208   @40822 Changes needed for the ROW-DISPLAY event.
** 502 SVL 20081208   @40827 Operation code is set to UPDATE or SET during
**                           "message ... update/set ..." actions.
** 503 SVL 20081209   @40849 Added code for the exceptional conditions
**                           handling into processProgressEvent().
**                           isGoPending changed to int in order to count
**                           the number of conditions which cause GO
**                           event. Added immediateException concept.
** 504 SIY 20081209   @40855 Reworked frame hiding implementation.
** 505 CA  20081210   @40846 Implemented CLOSE QUERY statement - on close,
**                           the browse widget will be notified to reset
**                           its state.
** 506 SVL 20081211   @40862 immediateException is thrown into
**                           processEventsWorker() only if we are in the
**                           WAIT-FOR statement.
** 507 NVS 20081211   @40882 Backed out the problematic fixes ##503,506.
** 508 NVS 20081215   @40911 APPLY adjusts focus instead of applying event
**                           if the focus was not valid at the time. This
**                           is a partial fix as no event filtering occurs
**                           yet.
** 509 SVL 20081216   @40920 Reapplied #503, added triggerNestingByWaitFor
**                           and lastAppliedCode.
** 510 NVS 20081218   @40941 Backed out #508 due to regressions.
** 511 SIY 20081218   @40957 Added static accessor for Screen instance.
** 512 SVL 20081224   @40981 Temp text is cleared in the message area
**                           before (for the specific line) and after the
**                           MESSAGE ... SET/UPDATE ... action.
** 513 SIY 20081224   @40989 Added hideWindow() method. Added fix for
**                           frame Z-order in promptFor() (replacement for
**                           Window.java #044 and #046).
** 514 SIY 20081225   @40994 Hide dialog immediately when it leaves scope.
** 515 NVS 20081226   @40998 Added private method highlight() and used it
**                           in waitForWorker() to control highlighted
**                           state of the focused widget. Added highlight
**                           control when sending ENTRY and LEAVE events.
** 516 CA  20090108   @41024 Re-display frame if it was hidden during
**                           trigger invocation and it was not an explicit
**                           HIDE call. Added more highlight state fixes.
**                           ENABLE calls will make all targeted hidden
**                           widgets visible. Fixed VIEW FRAME calls - all
**                           hidden widgets remain hidden.
** 517 CA  20090109   @41071 On event processing, any ConditionException's
**                           must be raised and re-thrown. Ref @41070
** 518 SIY 20090112   @41094 hideAll() disables placement quirk mode just
**                           like explicit frame hiding.
** 519 CA  20090114   @41124 Fix focus change when current focus is
**                           invalidated during trigger invocation. The
**                           cause for this invalidation is focus parent
**                           frame visibility change from not-visible to
**                           visible, via an explicit VIEW.
** 520 NVS 20090116   @41150 Fixed direct drawing issue caused by enable()
**                           called from promptFor() for a visible frame
**                           which was obscured by an overlay frame.
**                           Just the frame and widget's visibility check
**                           is not enough: widget should not be even
**                           partially obscured by overlay frames to be
**                           eligible for direct drawing.
**                           Added isAllShowing() static helper method.
**                           TO-DO: check whether all other cases of the
**                           direct drawing are susceptible to the same
**                           problem.
** 521 SVL 20090120   @41182 Added inEnable flag.
** 522 CA  20090121   @41185 Fixed CLEAR FRAME stmt for down frames -
**                           clearWorker(boolean) should handle this stmt
**                           by calling FrameHelper.clearCurrent instead
**                           of trying to clear the entire frame (calling
**                           FrameHelper.clearWorker(boolean)).
** 523 NVS 20090122   @41195 Fixed a regression caused by #520: ENABLE or
**                           DISABLE referencing BrowseColumn would throw
**                           an exception. Optimized enable() to bypass
**                           shadowed state check if no frames are above.
** 524 SIY 20090128   @41233 Fixed order of setEnabled()/view() calls which
**                           may affect visible appearance of the widgets
**                           if enabled widget belongs to iterated frame.
** 525 SIY 20090130   @41246 Remove highlighting from widget which looses
**                           focus.
** 526 SIY 20090203   @41255 Trigger call should transfer highlighted state
**                           from widget which has focus before trigger
**                           invocation to widget which receives focus
**                           after trigger invocation.
** 527 NVS 20090209   @41283 Installed double buffered output driver as
**                           a standard interactive output driver.
**                           Reimplemented putScreen() and putCursor()
**                           exports using direct output driver.
** 528 SIY 20090212   @41293 Refresh wiped out area in hideAll().
** 529 SIY 20090213   @41310 Added 'force' parameter into refreshStartRow
**                           signature.
** 530 NVS 20090213   @41318 Fixed few regressions in message and status
**                           processing caused by transition to the double
**                           buffering output driver. Code cleanup.
** 531 NVS 20090225   @41363 Added static getScreenSize() method.
** 532 SIY 20090225   @41378 Refresh message area during implicit clear
**                           if there was a message.
** 533 CA  20090304   @41410 ENTRY event must not adjust focus before its
**                           trigger is fired; reworked focus handling when
**                           explicit or implicit focus or first enabled
**                           widget on screen get disabled/hidden during
**                           trigger invocation; consume'd events do not
**                           specify that the trigger returned no-apply -
**                           the last trigger result for the current level
**                           will be kept on a stack.
** 534 CA  20090306   @41424 Fixed a focus related regression introduced
**                           by H533. Fixed editing block behavior: send
**                           an ENTRY to the requested widget, when
**                           editing mode starts. Fixed a issue related to
**                           UPDATE: when a widget list is used, first
**                           focusable widget from the list must be used.
** 535 SVL 20090305   @41392 Insert indicator states are saved on entering
**                           and restored on leaving of prompting blocks.
** 536 SVL 20090311   @41504 Enable auto-zap and reset insert mode for
**                           fill-in which validation has failed when
**                           it is currently focused.
** 537 CA  20090310   @41455 Implemented focus handling when implicit or
**                           explicit focus are hidden/disabled before
**                           trigger invocation. A MSG ... UPDATE stmt
**                           will adjust focus after execution, if it is
**                           previously undefined. Fixed READKEY behavior
**                           related to CHOOSE event.
** 538 CA  20090312   @41516 Implemented focus handling rules when in not
**                           in nested trigger mode.
** 539 SIY 20090311   @41545 Added widget drawing brackets for conditional
**                           widget refresh.
** 540 CA  20090317   @41623 Fixed issues related to out-of-scope pending
**                           focus. Fixed a regression related to APPLY,
**                           introduced by H538.
** 541 SIY 20090319   @41647 Added conditionalRefresh(Rectangle) method.
** 542 SVL 20090320   @41658 If leave action for a widget has returned
**                           no-apply, then insert mode will be reseted
**                           and zap will be enabled for the focused
**                           fill-in.
** 543 CA  20090323   @41663 Added missed rule about READKEY and CHOOSE
**                           trigger: if a GO event was trigger, no more
**                           key is read by current READKEY op. Fixed more
**                           issues related to APPLY, ENTRY/LEAVE events
**                           and focus search. Added missed rule about
**                           focused undef explicit for wait-for.
** 544 CA  20090324   @41669 Fixed focus restoration for nested UPDATE.
**                           Focus must not be adjusted if a trigger
**                           generated a GO event. In focus undef explicit
**                           case, operation must end if no widget accepts
**                           entry and the explicit focus is disabled or
**                           hidden.
** 545 CA  20090324   @41671 Fixed issues related to APPLY stmt. Fixed
**                           possible NPE when combo widget is used.
** 546 CA  20090327   @41694 Fixed NPE in processEventsWorkers caused by
**                           H545: locateFrame may return NULL if a window
**                           component is passed as a parameter.
** 547 CA  20090330   @41695 Fixed regression introduced by H546, related
**                           to APPLY stmt, when target is invalid.
** 548 SVL 20090402   @41716 Insert mode is saved/restored on triggers.
**                           Insert mode saving/restoration refactored.
**                           Fixed: unnecessary insert indicator clearing
**                           with HIDE ALL, zap on WAIT-FOR PAUSE, insert
**                           mode on focus restoration after WAIT-FOR,
**                           insert mode accepting if there are no focused
**                           widgets.
** 549 SIY 20090326   @41710 Implemented status saving/restoring logic.
** 550 CA  20090402   @41722 On frame destroy, all widgets get disabled,
**                           if frame is still visible.
** 551 CA  20090403   @41726 Fixed regression introduced by H550 - disable
**                           frame before removing it.
** 552 CA  20090402   @41725 Reworked APPLY focus handling (before and
**                           after trigger execution).
** 553 SIY 20090407   @41729 Do not disable widgets in frames which don't
**                           belong to interactive terminal.
** 554 SIY 20090407   @41731 Changed approach of disabling widgets during
**                           frame destroying in order to avoid regressions
**                           in UNDERLINE handling.
** 555 CA  20090408   @41735 On LEAVE, fill-in cursor must be set on first
**                           position. No trigger should be invoked when
**                           a combo's drop-down list is in focus. Fixed
**                           focus adjustment after wait-for iteration,
**                           when focus got disabled. Fixed frame:sensitive
**                           flag. Fixed issues related to WINDOW and FRAME
**                           events. WAIT-FOR will generate a STOP condition
**                           if it doesn't wait for global events and there
**                           no enabled widgets between ones registered for
**                           exit events.
** 556 CA  20090413   @41782 Fixed a couple of focus issues: adjustFocus
**                           must not skip focus adjustment if focus is
**                           disabled and there was a wait-for loop
**                           executed by the previous trigger; after a
**                           wait-for iteration, focus must be adjusted if
**                           current state is undefined. Fixed SENSITIVE/
**                           VISIBLE/HIDDEN flags behavior for widgets and
**                           frames. Widgets which had their HIDDEN or
**                           VISIBLE flag changes will send the new state
**                           back to the server, for synchronization.
** 557 SIY 20090414   @41786 Prevent NPE in popStatus().
** 558 CA  20090414   @41785 Made collectFlag package private, to collect
**                           flags for frame header widgets, when a frame
**                           has its VISIBLE flag changed.
** 559 CA  20090414   @41790 Focus must be restored after an wait-for loop
**                           when executed inside an editing block. If it
**                           needs adjustment, will be adjusted regardless
**                           of operation type. When executing first
**                           wait-for iteration, frame is not used as
**                           target if focus exists and is not visible.
** 560 CA  20090415   @41795 Fixed another regression related to focus do
**                           not use frame initial target, if focus does
**                           not exist; fixed a regression related to an
**                           APPLY stmt executed inside an EDITING block.
** 561 CA  20090416   @41804 Fix for an abend generated by UPDATE calls
**                           for down frames - related to initial focus.
** 562 CA  20090416   @41805 UPDATE/SET/PROMPT-FOR with EDITING block must
**                           restore focus to the widget in focus prior to
**                           this call (the same way simple UPDATE statements
**                           work).
** 563 CA  20090416   @41806 Fixed a case about initial hidden focus frame
**                           when wait-for has explicit target.
** 564 CA  20090417   @41808 Fix abend caused by H562 - Dequeue does not
**                           allow null values on stack, use Stack instead
**                           as type for the editingFocus field.
** 565 CA  20090421   @41820 No triggers should be invoked when for an APPLY
**                           with implicit target there is no focused widget.
**                           CHOOSE in row mode must use frame as target,
**                           even if P2J doesn't set it in focus.
** 566 SIY 20090421   @41834 Wipe screen when frame can hide others.
** 567 CA  20090422   @41836 The implicitApplyNoTrigger flag must be restored
**                           in TC.apply in any situation, even if a condition
**                           is raised (due to F4) - this fixes a regression
**                           introduced by H565.
** 568 SIY 20090422   @41844 Limited wiping screen to only specially marked
**                           frames.
** 569 CA  20090422   @41840 Added handling of NEXT-FRAME and PREV-FRAME
**                           events.
** 570 CA  20090423   @41903 Fixed a special case about read-only EDITORs: if
**                           a trigger had a DISPLAY with editor widgets as
**                           target, then caret (not focus) is moved to last
**                           targeted editor, in tab-order.
** 571 SVL 20090424   @41922 Addition to #542 - do not reset insert mode and
**                           enable zap if focus has remained on the same
**                           fill-in. Skip insert mode restoration if an
**                           editor is in focus.
** 572 GES 20090424   @41967 Import change.
** 573 SIY 20090429   @42073 Save/restore status when invoking message box.
** 574 SIY 20090430   @42085 Status saving/restoring in PROMPT-FOR and
**                           MESSAGE ... UPDATE... .
** 575 SIY 20090501   @42086 Rolled back #573 because it causes regressions
**                           and relevant cases are fixed by #574.
** 576 SIY 20090501   @42089 Added notification of frames about stream end of
**                           life.
** 577 SIY 20090519   @42263 Fixed widget zap state on widget activation when
**                           ENTRY event is not sent.
** 578 SVL 20090515   @42185 Removed unnecessary insert mode adjustment into
**                           waitForWorker(). Do not redraw insert indicator
**                           after HIDE ALL if it is into ambiguous state.
** 579 SVL 20090521   @42422 Empty help string allows default input message to
**                           be used.
** 580 SVL 20090527   @42472 For the APPLY "ENTRY" statement, we should enable
**                           zap and reset insert mode for the ENTRY target
**                           (fill-in), or to the currently focused fill-in if
**                           ENTRY has returned NO-APPLY. When the focus is
**                           restored after a trigger call, zap should be
**                           enabled for the focus target. Rolled back #542.
** 581 SIY 20090528   @42477 Avoid redundant repainting in redirected mode if
**                           entire iteration of DOWN frame is displayed.
**                           Ref: #42476.
** 582 SIY 20090529   @42500 Limit #581 to redirected mode only.
** 583 SVL 20090529   @42487 After external procedures restore only type of
**                           the status rather than statuses themselves.
** 584 SIY 20090601   @42506 Removed unused inner class.
** 585 SIY 20090602   @42513 Fixed NPE in popStatus().
** 586 SIY 20090609   @42619 inputActive should be reset on new procedure entry
**                           and restored upon exit.
** 587 SIY 20090610   @42637 Added support for authentication operation mode.
**                           In this mode Ctrl-C is processed as regular key
**                           which triggers StopConditionException.
** 588 SIY 20090611   @42658 Reset all pause flags if Ctrl-C is observed.
** 589 SIY 20090612   @42668 inputActive is increased for waitFor() with
**                           explicit focus.
** 590 SVL 20090612   @42681 Fixed NPE in message(). Notify BufferSizeManager
**                           about blocking operations. "Attempt to write
**                           to closed stream" may be raised into view_().
**                           Added resetBufferSizeManager() method. Raise STOP
**                           instead of ERROR into switchStreams(),
**                           flushStream() and flushRedirected().
** 591 SIY 20090623   @42782 Clear keyboard buffer in pause() only if non-zero
**                           time interval is passed.
** 592 GES 20090623   @42792 Any implicit or explicit pause has a special
**                           behavior in processCondition() that it does not
**                           change the interactions counter BUT if it executes
**                           when the interactions counter is 1 (but not when it
**                           is 0), then an END-ERROR will be translated into
**                           ERROR (as if the real interactions counter was 2
**                           or more) instead of into an END.
** 593 SIY 20090623   @42794 Added protection and logging for attempts to pop
**                           from empty cursor stack.
** 594 SIY 20090624   @42886 Rolled back #593 and fixed trap during redundant
**                           stopEditingMode() call.
** 595 SIY 20090701   @43022 Added client side part of format checking code.
** 596 SIY 20090707   @43069 Do not force frame z-order if it is displayed
**                           inside editing block and has no enabled widgets.
** 597 SIY 20090708   @43105 Added one more case of wiping screen from
**                           PUT SCREEN artifacts.
** 598 GES 20090701   @43028 Prevent event brackets from getting out of balance.
**                           Also added more complete state clearing on CTRL-C.
** 599 SIY 20090709   @43124 Less aggressive screen wiping (fix for #597).
** 600 NVS 20090709   @43130 The fix #598 has left an unbalanced setInvalidate
**                           call in popStatus() method. Fixed.
** 601 SIY 20090713   @43147 Fixed handling of focus restoring after trigger
**                           when source of the event is Button.
** 602 GES 20090720   @43249 NPE protection for highlight().
** 603 SIY 20090723   @43348 Added event/drawing brackets for focus restoring.
** 604 SIY 20090724   @43388 Fixed terminal width adjustment in redirected mode
**                           for frames which have initial column specified.
** 605 SIY 20090804   @43510 Added support for NO-PAUSE in handling of
**                           conditional UP/DOWN.
** 606 CA  20090805   @43520 Ensure the event queue is popped in pop(), even
**                           if a condition is raised during event processing.
**                           This will solve the client-side OOME caused by
**                           event queue leaks.
** 607 SIY 20090806   @43541 Fixed paging of DOWN frames with iterations which
**                           have height > 1 but not end with SKIP widgets.
** 608 SIY 20090810   @43580 Fixed message hiding in some special cases (when
**                           placed frame hides other frame at the same row).
** 609 SIY 20090811   @43595 Made more narrow conditions which trigger message
**                           hiding in #608.
** 610 SIY 20090813   @43611 Even more narrow conditions which trigger message
**                           hiding in #608.
** 611 SIY 20090814   @43617 Reworked handling of DOWN paging for frames which
**                           have iteration height greater than 1. Fixed
**                           handling of leading/trailing SKIP widgets in
**                           iteration.
** 612 SIY 20090820   @43697 Fixed handling of Ctrl-C during pause.
** 613 SIY 20090821   @43715 Fixed handling focus in cases when frame is
**                           completely disabled in trigger.
** 614 SIY 20090824   @43731 Use hasAnyEnabled() instead of hasEnabled() in
**                           hideFramesInt(). Fixes regressions caused by
**                           recent changes in hasEnabled().
** 615 SIY 20090824   @43792 Added proper handling for 'forced message hiding'
**                           mode which can be triggered in some circumstances
**                           during PROMPT-FOR initialization.
** 616 SIY 20090901   @43807 Added event bracket for widget hiding loop
**                           in hideAll() method.
** 617 SIY 20090902   @43826 Fixed handling focus in certain scenarios of
**                           button-browse-trigger interactions.
** 618 SIY 20090906   @43830 Do not process events from components in removed
**                           frames.
** 619 SIY 20090907   @43835 Fixed restoring of redirection in message() in
**                           case when user presses END-KEY during pause.
** 620 CA  20090902   @43823 Fixed a deadlock in initializePrep - if CTRL-C is
**                           pressed at a certain time, the watchCtrl call
**                           will block, as it wants to acquire class-level
**                           lock both on TC and Window classes.
** 621 SIY 20090903   @43838 Rolled back #618.
** 622 SIY 20090904   @43839 Partially rolled back #617.
** 623 SIY 20090909   @43850 Protection from NPE in iterateEditingMode().
** 624 SIY 20090910   @43856 getFocus() should not return widget in hidden
**                           frame as current focus.
** 625 SIY 20090924   @44008 Skip tinyInput() frame when preparing batch for
**                           server.
** 626 SIY 20091015   @44136 Refresh frame, if it is about to be edited and
**                           there are other frames at the screen.
** 627 SIY 20091025   @44161 Fix for regressions caused by #626. It should be
**                           activated only if last frame which was made
**                           visible is not the same as one which is about to
**                           be edited.
** 628 SIY 20091028   @44274 Disable automated removal of frame created by
**                           tinyInput.
** 629 SIY 20091118   @44429 Made tinyInput() re-entrant.
** 630 CA  20091111   @44329 Added a statusDefault(String, boolean) and
**                           status(String, boolean) which disable 63-char
**                           trimming, if the noLengthLimit flag is set.
** 631 SIY 20100111   @44524 Fixed a rare case when drawing of empty space
**                           instead of label is necessary.
** 632 SIY 20100115   @44530 Added conditionalLineRefresh() method.
** 633 SIY 20100122   @44547 Push/pop event list during start/stop of editing
**                           mode.
** 634 SIY 20100128   @44552 Narrowed set of cases when #631 fix is triggered.
** 635 SIY 20100128   @44554 Even more narrow conditions for #631.
** 636 SIY 20100129   @44564 Added method to access refresh quirk state var.
** 637 SIY 20100316   @44751 Conditional DOWN after regular DOWN 1 which
**                           flushed output should be handled as regular DOWN.
** 638 SIY 20100329   @44785 Disable immediate exceptions while processing
**                           applyChanges().
** 639 CA  20100507   @44850 Fixed a CTRL-C case, when its pressed while the
**                           combo's drop-down list is active - it is removed
**                           "manually" in the waitForWorker's finally clause,
**                           else it will always remain attached to the screen
**                           Note that behavior is not confirmed in Progress
**                           environment.
** 640 EVL 20100513   @44853 Add explicit focus setting to fix the problem
**                    @44875 after invoking search tool in browser choose
**                           frame (the second approach).
** 641 EVL 20100810          The new implementation of the fix for labels
**                           hiding. We trying to take into account only the
**                           only bottom line. And only for the case when the
**                           frame to be hidden is two chars less than the
**                           underlying frame.
** 642 EVL 20100812          Fix for PUT SCREEN artifact when the frame clears
**                           the output result incorrectly. The refresh()
**                           method from Toolkit class is now called a bit less
**                           frequently in promptFor() method. And only for
**                           boxed frames with overlapping areas with the
**                           previously rendered frame. Regression fixed.
** 643 SIY 20100619          Switching to own implementation of character UI.
** 644 SIY 20100804          Changes in imports, some refactorings.
** 645 EVL 20100823          Adding code to repaint labels in original state
**                           after making the call to turn on the empty mode.
**                           Also the finishing working with tinyInput() must
**                           force previous frame repainting to display
**                           the labels in its original state.
** 646 SIY 20100822          Refactored AttributeContainer support.
** 647 SIY 20100831          Changes related to refactoring of FrameHelper.
** 648 SIY 20100915          First part of deep refactoring.
** 649 SIY 20101015          More refactoring, changes related to refactoring
**                           of the Screen.
** 650 SIY 20101108          Fixed NPE during initializePrep().
** 651 SIY 20101122          Switching from KeyEvent to KeyInput.
** 652 SIY 20110504          Fixed focus handling for ENABLE inside trigger.
** 653 SIY 20110506          Narroved conditions for ENABLE focus handling fix.
** 654 SIY 20110510          Another attempt to fix focus handling fix.
** 655 SIY 20110511          Enabled AUTO-ZAP mode in fill-in widget after
**                           focus change during ENABLE processing.
** 656 SIY 20101202          Minor refactoring.
** 657 SIY 20110112          Changes related to new color handling scheme.
** 658 GES 20110921          Enable multiple drivers. The runtime driver is
**                           determined at startup from the bootstrap cfg.
** 659 GES 20111006          Eliminated pseudoTerminal methods (users can call
**                           the LaunchManager directly.
** 660 GES 20111012          Moved to context-local instead of singleton. This
**                           allows multiple client sessions to be hosted in
**                           the server.
** 661 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 662 GES 20111028          Modified initialization, termination and CTRL-C
**                           processing to allow for running inside the server
**                           process.
** 663 GES 20111111          Modified interface export to allow registration
**                           to be exclusively local or remote as needed.
** 664 SIY 20111202          Changes related to split up of widgets into two 
**                           parts (UI-independent and UI-specific).
** 665 SVL 20120214          Repaint button which loses focus in an alert box.
** 666 SVL 20120821          Fixed getSelection().
** 667 CS  20121031          Removed client specific references.
** 668 GES 20130523          Added support for initializing the memory daemon.
** 669 OM  20135021          Added SessionExpport registration for Session attributes.
** 670 EVL 20130606          Added support for starting environment handlind daemon.
** 671 CA  20130822          Added support of appserver async requests: posting and executing of
**                           server-side events on the client side.
** 672 EVL 20130903          In Windows we need to hide the cursor when pausing with key prompt.
** 673 CA  20130908          Added support for socket event runtime.
**                           Initialize socket-related exports at client launch.
** 674 CA  20131001          Added API to change the numeric separators on client-side.
** 675 HC  20131031          Changed resource ID data type to Long, see issue #2183.
** 676 CA  20131013          Added implementation for the DELETED_SCOPE signal. The hidden/visible
**                           state changes must be collected when the widget's attribute is 
**                           actually set.
** 677 EVL 20131127          Adding support for batch mode.
** 678 VIG 20131225          Added function key support (LAST-EVENT:FUNCTION)
** 679 GES 20131124          Initialize native library support at client launch.
** 680 EVL 20140114          Fix for key reader in batch mode. We need to return -2 bypassing
**                           the key reader thread. Also added separation for background and
**                           foreground modes simulation.
** 681 EVL 20140120          Adding signal waiter thread starting.
** 682 HC  20140119          Implemented runtime support for the COLOR 
**                           protermcap-attribute.
** 683 EVL 20140210          Fix for background batch mode in MESSAGE and processing including
**                           generating message box.
** 684 EVL 20140221          The message() and messageBox() methods modified to take into account
**                           redirection of the STDOUT from on the command line level.
** 685 GES 20140228          Fixed message statement processing with SKIP or embedded newlines.
** 686 EVL 20140326          Adding OsPropertiesDaemon to pass the client OS properties to the
**                           server side.  Also the EnvirinmentDaemon creation has been moved
**                           to the point where the the connection to the server nas beed created.
** 687 EVL 20140514          The underline call needs special processing when frame holding the
**                           widget to be underlining is not yet visible.
** 688 VIG 20140601          Adding moveBeforeTabItem, getFirstTabItem, setFirstTabItem, 
**                           getNextTabItem, extractWidgetIDs tab order management methods.
** 689 VIG 21040612          Fixing getFirstTabItem and moveBeforeTabItem with new signatures.
** 690 EVL 20140617          Fix for waitForWorker() method forcing the cursor to be on. This
**                           is OK only when widget under focus requires the cursor to be on.
** 691 MAG 20140512          Small changes to support the early GUI screen driver.
** 692 CA  20140711          Enhanced font support for GUI clients, add MESSAGE-AREA widget 
**                           support.
** 693 MAG 20140801          Check if frames still exists when restore frames visibility in requestHelp.
** 694 VIG 20140815          Changed pause and pauseBeforeHide methods signatures: added windowId
** 695 CA  20140806          Moved widget registry at the output manager. Added mouse events.
**                           Isolated some ChUI sepecific from the GUI code.
** 696 MAG 20140812          Implements COLOR-TABLE system handle.          
** 697 VIG 20140829          Added status methods with window Id argument.
**                           Changed default window ID from the constant value to dynamic. 
** 698 MAG 20140910          Implements environment API.          
** 699 EVL 20140808          New widget ID approach, the client code now extract frame ID from
**                           widget ID using widget implementation where the frame ID is stored.
** 700 HC  20140911          GUI coordinate support - data types refactoring.
** 701 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 702 MAG 20141007          Implements GUI message area.
** 703 CA  20141021          Added support for MOUSE and WINDOW events required by GUI mode.
** 704 EVL 20141024          Client side special handling for dynamic frames. OPEN_SCOPE,
**                           CLOSE_SCOPE leave untouched, executing the same way for static and
**                           dynamic frames.
** 705 VIG 20140911          Added status* and message* methods for cases when IN WINDOW clause 
**                           is used and moved all status* and message* methods to Window class.
** 706 CA  20141029          Refactored to allow key events for ChUI (linux and windows) and GUI 
**                           (windows) modes.  Fixed problems related to setting the LAST-EVENT
**                           attrs when a key is read via something other than READKEY. 
** 707 HC  20141021          Implemented window move/position support.
** 708 CA  20141106          Added window ID to promptFor.  Also, the window ID is set by the 
**                           client-side, in viewWorker - if the window has changed, the frame 
**                           needs to be moved to the new window.
** 709 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 710 CA  20141123          Fixed placement of default window.  Other fixes related to scrollable
**                           support and general focus management in GUI.  First pass at GUI client
**                           restart.
** 711 CA  20141125          Fixed CTRL-BREAK on GUI client.  Read the USE-3D-SIZE from the client
**                           config or environment (in this order), otherwise default to yes.
**                           The GUI Window is aware of events ONLY if the control is on client-
**                           side; while server-side does its processing, no events are captured
**                           or processed by the client-side.
** 712 CA  20141204          Fixed a problem in H711, related to client restart in ChUI.
** 713 CA  20141205          An initial attempt at fixing drawing inside widgetDrawingOn/Off 
**                           bracket: this should translate to frame's origin.
** 714 SVL 20141209          Implemented some browse-related functions.
** 715 CA  20141220          Fixed a mouse-event related NPE; in GUI, the default window height
**                           (including message and status areas) is 25 rows.
** 716 HC  20141207          Implemented runtime support for several SESSION attributes. Improved
**                           server-client state transfer of the default window.
** 717 EVL 20141216          Adding SESSION:TOOLTIPS attribute handling on the client side.
** 718 GES 20150102          Fix for abend when DISABLE/ENABLE operate on a frame that was never
**                           loaded. Added deleteEditorLine().
** 719 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 720 HC  20150111          Fixed widget position change in a frame.
** 721 CA  20150114          Fixed support for changing the ROW/COLUMN attrs from server-side, 
**                           for a side-label.
** 722 CA  20150115          Added logging message for missing feature: moving widgets in a down 
**                           frame is not yet implemented.
** 723 EVL 20150114          Adding API to set/load window icon and images.
** 724 CA  20150122          Added runtime support for APPLY ... TO non-widget resource.
** 725 GES 20150129          Rework to pass the actual matched event to the server side trigger.
**                           This can be different from the input event for ANY-PRINTABLE and
**                           ANY-KEY. The change is needed to allow the server side to properly
**                           avoid recursion of the same widget + event combo.
** 726 HC  20150128          Added support for frame widget position changes induced by
**                           setting ROW/COL/X/Y attributes.
** 727 CA  20150130          OVERLAY frames are always moved to top, regardless if EDITING mode is 
**                           on or not.
**                           For EDITING phrase, the screen buffer needs to contain the frame and 
**                           focus widget.
** 728 CA  20150202          For an EDITOR, if it loses focus after an RETURN/ENTER is pressed, it
**                           will consume the event (the key will no be applied to the EDITOR).
**                           Fixed a problem with resolving the target window in viewWorker.
** 729 EVL 20150122          Adding pending loading for images support.
** 730 GES 20150204          Clear GO-PENDING when a NEXT statement is used in an EDITING block.
** 731 GES 20150206          Apply processing must propagate the current event list's operation
**                           when chaining.  Editing blocks now clear the leaveSent flag on
**                           starting the edit session.
** 732 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 733 GES 20150217          Added safety code to restore readkey blocking operation in finally.
** 734 CA  20150218          On trigger invocation, the screen buffer is retrieved from the focus,
**                           if not null; otherwise, is retrieved from the trigger source.
**                           An APPLY GO can terminate the current WAIT-FOR loop, if it targets a
**                           widget in a frame for which there is a UPDATE/SET/PROMPT-FOR loop on 
**                           the stack.
** 735 VIG 20150128          Added pushMenuDescription, destroyMenu methods and 
**                           DEFAULT-POP-UP event processing.
** 736 CA  20150220          Fixed the FRAME-VALUE statement: if HELP is active, it references the
**                           FOCUSed frame/widget as before the HELP is invoked.
** 737 HC  20150221          Improvements to frame tab order.
** 738 HC  20150223          Focus management fix.
** 739 GES 20150226          Fixes for AUTO-RETURN.
** 740 CA  20150228          ENABLE must make the widget's side-labels visible, too.
** 741 SVL 20150212          Modified getRows and refreshStartRow signatures.
** 742 IAS 20150305          Prevent NPE in positionCursor() for an empty RADIO-SET .
** 743 CA  20150308          Fixed problems in synchronizing the frame's screen-buffer with the
**                           server-side, after DISPLAY/UPDATE/... EDITING stmts are executed.
** 744 SVL 20150305          Updated refresh() signature.
** 745 CA  20150308          PAUSE ... NO-MESSAGE must clear all messages.
**                           CLEAR and CLEAR ALL APIs must update the frame's screen buffer.
** 746 CA  20150323          When setting the widget's format, notify the widge to update internal
**                           state.
** 747 EVK 20150324          Added SAVE-FILE and READ-FILE support for editor. 
** 748 SVL 20150331          Added some browse-related functions. refreshStartRow become
**                           refreshScrollRow with more options. Browse column values are added
**                           to the screen buffer.
** 749 EVL 20150320          Fix for incoming GO event from GUI widgets during update statement.
**                           In this case the update key reading loop should be run in mode when
**                           mouse click to be considered as potential significant event source
**                           because it can cause GO, exit trigger, etc.
** 750 IAS 20150401          Added support for new EDITOR attributes/methods
** 751 CA  20150408          All window events need to be treated as OS events.
**                           Fixed OS event processing when the OS event is consumed.
** 752 SVL 20150411          Added select{Next|Previous}Row, isRowAvailable.
** 753 HC  20150323          Removed business logic from config classes.
** 754 CA  20150429          Fixed widgetDrawingOff/On in GUI.  When creating a rectangle from a 
**                           location and size, pass the base units to the c'tor.
** 755 CA  20150417          Fixed setScreenBuffer, which was computing widget IDs from the base 
**                           ID.
** 756 VIG 20150226          Added null check.
** 757 EVL 20150415          Adding support for prepackaged button widget images.
** 758 SVL 20150420          Added executePendingRefresh, notifyColumnsUpdated.
** 759 CA  20150507          Fixed a ghosting problem when widget is losing focus (i.e. message 
**                           area).
** 760 HC  20150508          Improvements to GUI window scrolling support.
** 761 GES 20150506          Changes to make process launching less dependent on the UI packages.
** 762 CA  20150520          Fixed FontManager initialization.
** 763 CA  20150603          Protect processEventsWorker against leaked sync events for which the 
**                           widget is no longer alive.
** 764 SVL 20150508          Added some browse-related functions.
** 765 HC  20150612          Implemented support for modal windows and GUI ALERT-BOX.
** 766 HC  20150626          Fixed an unhandled exception when showing a modal window.
** 767 SVL 20150625          Added getRowCount().
** 768 HC  20150702          Improved positioning of top-level windows on systems with multiple
**                           displays.
** 769 IAS 20150703          Added 'debug' method.
** 770 CA  20150615          For frames with default-button, a RETURN in a EDITOR widget (with
**                           RETURN-INSERTEd set to false) must send the RETURN to the default 
**                           button.
** 771 EVL 20150518          Adding LEAVE generation when focus is moving from drop-down to
**                           another widget in GUI mode by mouse click.  Adding API to get the
**                           currently active event list.  Fix for possible NPE generation and
**                           strange logic in method fixFocusedFrameZOrder().  Moving mouse click
**                           leaving to happen early enough to get real focused widget and be
**                           able to react on mouse press instead of the click.
** 772 CA  20150722          Refactored so that browse column configurations are transfered via 
**                           pushScreenDefinition calls.
**                           In case of APIs which change the frame row in a down frame, they need
**                           to transfer to the server-side the configurations for the widgets in
**                           this frame row: pushScreenDefinition(), clearAll() and down() APIs
**                           are changed.
** 773 SVL 20150722          Updated browse-related API.
** 774 CA  20150725          ClientExports.view APIs now return the frame's down configs, too.
** 775 HC  20150726          PAUSE must not clear WINDOW messages in GUI.
** 776 CA  20150731          Made eventBracket(boolean, Runnable) public.
** 777 CA  20150730          The FramePlacementManager is window-specific.
**                           When switching windows in GUI, the frame needs to be repainted fully.
**                           When calling draw() code outside of drawing brackets, the window 
**                           needs to be explicitly selected.
** 778 VIG 20150801          Added warning about absence of enabled frames in a window with menubar. 
** 779 GES 20150730          Modified GUI driver translate() usage to match API changes. 
** 780 IAS 20150801          WIDGET-LEAVE/WIDGET-ENTRY support.
** 781 VIG 20150827          Added ALT ignoring for PAUSE.
**                           Added exclusion of menu widgets to be visible for wait-for checks.
** 782 EVL 20150826          Adding support for SCROLL-TO-ITEM() method for selection-list.
** 783 HC  20150831          Scrolling support overhaul.
** 784 CA  20150827          Focus improvements with mouse events.
** 785 IAS 20150922          Added support for the MOVE-TO-TOP()/MOVE-TO-BOTTOM() methods
**                           and TOOLTIP attribute.
** 786 CA  20150929          Fixed multi-window drawing issues caused by concurrent drawing
**                           access to the GuiDriver: all access to the driver is done
**                           exclusively, each thread taking and releasing ownership for the
**                           duration of the drawing operation or other GUI API invocation which
**                           requires access to the underlying physical window.
**                           Changes to allow multi-window focus management and frame placement in 
**                           GUI.
** 787 VIG 20151011          Set lastkey for mouse CLICK/UP/DOWN events.
** 788 HC  20151013          Changes to extend GUI multi-window focus management and ACTIVE-
**                           WINDOW system handle processing to modal windows.
** 789 EVL 20151012          More tooltip setup restrictions.  Frame, window, alert-box and
**                           dialog-box can not have a tooltip.
**                           ENABLE/SENSITIVE attribute do not switch focus if the current focus 
**                           is valid.
**                           Invoking HELP in GUI must not re-place frame.
** 790 IAS 20151027          Support for NO-(AUTO-)VALIDATE options,
** 791 CA  20151024          Fixed frame placement when CURRENT-WINDOW is not DEFAULT-WINDOW.
**                           Added support for nested frames. 
**                           Added support for MOVE-AFTER-TAB-ITEM.
**                           Triggers created in a persistent procedure while a event processing
**                           loop is active need to be pushed to the event list for the nearest
**                           event processing loop, so they will be available in the UI.
**     HC  20151024          Added support for invoking code on event thread.
**         20151027          Implemented WINDOW MESSAGE-AREA and STATUS-AREA attributes.
**     VIG 20151027          Commented out eventDrawingBrackets call in pushMenuDescription.
**     SVL 20151029          Added queryRepositioned.
**     EVL 20151103          Adding handing image names without extensions while loading from
**                           client file system.
**     SBI 20151103          Changed terminate() to reset windows for gui as for chui.
**     CA  20151106          A fix for HIDDEN attribute: distinguish between HIDE statement and 
**                           setting the HIDDEN attribute to true.  Also, if the frame's window is 
**                           marked as HIDDEN, do not show the window.
**                           More fixes related to working with HIDDEN frame/widgets and resolving  
**                           the window when nested frames are viewed directly.
** 792 CA  20151123          More fixes related to nested frame support.
**                           A DISPLAY with a hidden frame must not place it on window.
**                           Replace direct drawing calls (via widget's draw() ) with event 
**                           drawing bracket, as this causes issues when the widget is covered
**                           by other widgets (which can be other frames), in z-order.
**                           placeFrameInt must clear the HIDDEN flag, if it makes the frame 
**                           VISIBLE.
**                           Fix for ACTIVE-WINDOW handle - LEAVE event for window must not 
**                           consume the event.
**                           Fixed implicit window show by frame statements, in GUI. 
**     HC  20151124          Utilized Keep3DFillinBorder environment variable when drawing FILL-
**                           IN border.
**     VIG 20151118          Fixed MENUBAR repaint for dynamic case.
** 793 SVL 20151230          FOCUS for a editable browse should return fill-in reference.
** 794 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 795 EVL 20151223          Overlay window support related changes.
** 796 VIG 20151226          isEnabled() is replaced with navigable() in computeFocus(),
**                           processProgressEvent(), processEventsWorker(). 
**                           Fixed NullPointerException when using WindowManager.topLevelWindow().
** 797 HC  20160203          Improved runtime support of KEEP-FRAME-Z-ORDER and TOP-ONLY
**                           attributes.
** 798 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 799 IAS 20160229          LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 800 HC  20160304          Replaced Window.moveBy() with setPhysicalLocation().
** 801 CA  20160304          Added APIs to deregister widgets which are part only of a shared 
**                           frame (they are not inherited from the master).
** 802 IAS 20160305          Support for the APPL-ALERT-BOXES, DATA-ENTRY-RETURN, DEBUG-ALERT
**                           and IMMEDIATE-DISPLAY SESSION attributes
** 803 IAS 20160328          Client log support
** 804 CA  20160327          Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 805 IAS 20160331          Fixed potential NPE
** 806 OM  20160323          Exposed WebServices interface for remote access from server. Renamed 
**                           getLocalImageName() to getLocalFileName().
** 807 CA  20160129          move-to-top/bottom need to repaint the widget, not the entire window.
** 808 SVL 20160406          Added moveColumn.
** 809 HC  20160406          Prevent event triggering for virtual widgets. Prevent processing
**                           events for non-displayed windows.
** 810 VIG 20160209          Added check for RMB ENTRY event to processProgressEvent() method. 
** 811 SBI 20160412          Added openURL(String path) to open the target page in the browser.
**     CA  20160418          Added infrastructure for message passing from converted code to a 
**                           customer-specific application or calls from the customer-specific 
**                           application into the converted code (i.e. invoke external programs,
**                           procedures or functions).  Can be used only when the P2J client runs 
**                           embedded in a customer-specific application.
** 812 SVL 20160419          Added moveColumnServer.
** 813 GES 20160311          Removed dead code (setFrameValue() and getValue()).  FRAME-VALUE
**                           is implemented on the server-side only (along with some cooperation
**                           with the screen-buffer processing.
** 814 HC  20160511          Fixed COMBO-BOX focusing issues and related changes.
** 815 HC  20160607          GUI menus reworked to be displayed in dedicated top-level windows.
** 816 SBI 20160516          Exposed the client's system clipboard to the server. 
** 817 EVL 20160511          Fix for wrong main window show when the frame is a part of the dialog
**                           box.  Need to find the root frame to check for dialog.  The nested
**                           frames can not have the dialog flag.  Added fix for dialog-box size
**                           getting.
** 818 HC  20160628          Prevented unhandled exceptions when window/frame closed/destroyed.
** 819 CA  20160627          In case of MESSAGE ... VIEW-AS, PUT SCREEN, PUASE and output 
**                           redirection to terminal, reactivate the batch mode, if it was
**                           deactivated.
** 820 SBI 20160708          Added resetActiveComboBox(). 
** 821 SBI 20160720          Fixed NPE using that APPLY "ENTRY" to window doesn't need to be
**                           propagated to contained frames and widgets.
** 822 EVL 20160721          Rework for using screen buffers in client-server interactions.
**                           Adding helper method to process array of screen buffers to separate
**                           common code.
** 823 IAS 20160806          Made int getFrameLine(Frame, int) method public
** 824 IAS 20160817          Minor optimization of the DOWN processing in the redirected mode
** 825 EVL 20160805          Fix for GO event processing inside GUI dialog-boxes and frames.
** 826 EVL 20160812          Adding NPE protection in getScreenBuffer() method for the cases when
**                           the frame from editable frame list does not yet exist in general
**                           widget registry.  It can be taken a bit later when or if it comes up
**                           in general registry.
** 827 IAS 20160829          displayAndDown() operation support
** 828 HC  20160901          Error messages in GUI are always displayed in alert boxes.
** 829 CA  20160915          Fixed focus handling when the focus window/frame has been deleted.
** 830 SBI 20160921          Removed unused import.
** 831 HC  20160922          Improvements to mouse input and related focus processing.
** 832 HC  20160926          Removed disabling of window input events during server processing.
** 833 OM  20160919          Fixed infinite pause in batch mode (client-side).
** 834 CA  20161014          viewWorker() must make visible any non-hidden and not-yet-realized 
**                           (i.e. dynamic widgets) added after the frame was made visible.
**                           Commented out a code in placeFrameInt related to nested frames.
** 835 CA  20161018          Any client-to-server code invocations must maintain the trigger 
**                           registries, not just explicit trigger invocation.
** 836 CA  20161020          Fixed a horiz scrollbar 4GL bug, which prevents showing the horiz  
**                           scrollbar for the parent frame if the child frame uses a ENABLE or 
**                           DISPLAY statement and the parent frame requires a horiz scrollbar.
** 837 VIG 20160831          Added MOUSE_CLICKED processing to resetting menubar state.
** 838 CA  20161123          If a child frame is made visible implicitly by its parent, then its
**                           "viewed" flag must not be set; also, this affects clearWorker for the
**                           child frame, which must not pause if the "viewed" flag is not set.
**                           Fixed destroyFrame() for overlay frames (they must be hidden and
**                           destroyed when their scope ends)/
** 839 HC  20170102          Added break of current WAIT-FOR in case a Window or Mouse event
**                           trigger causes an END-condition.
** 840 OM  20161209          Fixed unconditionalRedirectedDown().
** 841 HC  20170106          Fixed messageBox() making invalidation mode unbalanced.
**     HC  20170201          Fixed MENUBAR made visible or hidden when it is attached or detached
**                           from its window.
**     CA  20161101          Added support for mnemonic processing.
**                           Browse is able to handle HOME/END in invisible state.
**     HC  20170203          Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                           TC.invokeLater() to allow execution of code to be queued on the
**                           application queue.
**     HC  20170215          PROMPT-FOR causes the related window to be brought to the foreground.
**     CA  20170228          Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and
**                           for external applications.
** 842 OM  20170320          Added optional parameter to help identify the correct window in
**                           setHelp(). Balanced eventDrawingBracket window selection in 
**                           eventDrawingBracket().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.function.*;
import java.util.logging.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.library.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * Thin Client generic CHUI driver.
 */
public class ThinClient
implements ClientExports,
           StateSynchronizer,
           Watcher,
           CloseListener,
           AdvanceListener,
           FlushListener
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ThinClient.class.getName());

   /** Default pause before hide message text. */
   private static final String PBH_DEFAULT_TEXT = "Press space bar to continue.";

   /** Default pause before end message text. */
   private static final String PBE_DEFAULT_TEXT = "Procedure complete. " +
                                                  PBH_DEFAULT_TEXT;

   /** Exception message used in END-ERROR key processing. */
   private static final String END_ERR_MSG = "END-ERROR key function";

   /** Context local instance of this class. */
   private static final ContextLocal<ThinClient> local =
      new ContextLocal<ThinClient>();
   
   /** Token used to authenticate with the dispatcher when registering APIs. */
   private static Object modToken = null;
   
   /** Postponed VALUE-CHANGED event from BROWSE. */
   private KeyInput postponedValueChanged = null;

   /** This view refreshes the frame including decorations */
   public boolean viewRefresh = false;

   /** This view displays frame data */
   public boolean viewDisplay = false;

   /** This flag is on for the duration of view operation */
   boolean viewInProgress = false;

   /**
    *  <code>false</code> has no effect.
    *  <code>true</code> forces full frame redrawing, including decorations.
    */
   boolean nextViewRefresh = false;

   /**
    *  <code>false</code> has no effect.
    *  <code>true</code> forces data only output.
    *  Normally, only one of {nextViewRefresh, nextViewDisplay} flags should be
    *  set to <code>true</code> or else the requirements are in contradiction.
    *  If that happens nonetheless, nextViewRefresh takes precedence.
    */
   boolean nextViewDisplay = false;

   /** Reference to <code>Toolkit</code> instance. */
   /* 
    * TODO: there must be a "main" tk which is used to resolve the factory, so widgets can be
    * created while the secondory (i.e. ChUI mgr) is active.
    */
   private OutputManager tk = null;
   
   /** When running in the server process, this is the user's thread. */
   private Thread dedicatedThread = null;

   /** Server exports */
   private ServerExports server = null;

   /** Validation message from last failed validation operation. */
   private String validationError = null;

   /** Pause before hide state flag. <code>true</code> means pause is due. */
   private boolean needPause = false;

   /** Pause before hide time for automatic frame hiding. */
   private int timePBH = -1;

   /** Flag that determines if pause before hide mode is active. */
   private boolean honorPBH = true;

   /** Text displayed while processing pause before hide. */
   private String textPBH = PBH_DEFAULT_TEXT;

   /**
    * Default status message displayed when input operation is active.
    * Can't be initialized here because the client's type is not available (GUI or ChUI). 
    */
   private String inputDefaultText = null;

   /** Current active wait list. */
   private EventList currentEventList = null;

   /** Current registered ROW-DISPLAY events. */
   private Map<Integer, Integer> rowDisplayEvents = null;

   /** Flag indicating if a ROW-DISPLAY event is being executed. */
   private boolean inRowDisplayTrigger = false;

   /** Number of times when we've found that the GO event is pending. */
   private int isGoPending = 0;

   /**
    * An exception which causes all trigger processing recursions to be
    * terminated.
    */
   private ConditionException immediateException = null;

   /**
    * Contains levels of trigger nesting relative to active WAIT-FOR cycles.
    */
   private Stack<Integer> triggerNestingByWaitFor = new Stack<Integer>();

   /**
    * Contains true if the last trigger result (on this level) returned
    * no-apply.
    */
   private Stack<Boolean> triggerResultNoApply = new Stack<Boolean>();

   /**
    * Contains true if the last trigger (on this level) did execute a nested
    * wait-for loop.
    */
   private Stack<Boolean> triggerHadWaitFor = new Stack<Boolean>();

   /** The action code currently being applied by the apply() method. */
   private int lastAppliedCode = KeyInput.CHAR_UNDEFINED;

   /** Last key received from the keyboard. */
   private int lastKey = -1;

   /** File system daemon. */
   @SuppressWarnings("unused")
   private FileSystemDaemon fsd = null;

   /** Stream daemon (provides access to streams by id). */
   private StreamDaemon sd = null;

   /** Process daemon. */
   @SuppressWarnings("unused")
   private ProcessDaemon pd = null;
   
   /** Memory manager daemon. */
   @SuppressWarnings("unused")
   private MemoryDaemon mmd = null;
   
   /** Environment handling daemon. */
   @SuppressWarnings("unused")
   private EnvironmentDaemon ed = null;
   
   /** Library manager daemon. */
   @SuppressWarnings("unused")
   private LibraryDaemon libd = null;
   
   /** OS properties daemon. */
   @SuppressWarnings("unused")
   private OsPropertiesDaemon ospd = null;
   
   /** Tracks pseudo-terminals used for redirecting output to a stream. */
   private Map<Integer, RedirectedTerminal> redirections =
      new HashMap<Integer, RedirectedTerminal>();

   /** Tracks pseudo-terminals which need deferred cleanup processing. */
   private Set<Integer> deferredCleanup = new HashSet<Integer>();

   /** Tracks the id of the current output stream (-1 for interactive). */
   private int currentStream = -1;

   /** Tracks the number of I/O interactions within the current block */
   private int interactions = 0;

   /** Tracks all page elements which are active. */
   private Map<Integer,PageElement> pageElements =
      new HashMap<Integer, PageElement>();

   /** Tracks the streams in which a given page element is registered. */
   private Map<PageElement, Set<Stream>> streamMap =
      new HashMap<PageElement, Set<Stream>>();

   /** Level of nesting (recursion) that exists with trigger processing. */
   private int triggerNesting = 0;

   /** Tracks any conditional up or down that must be applied. */
   private int conditionalUpDown = 0;

   /** Tracks the frame ID associated with the conditional up or down. */
   private int conditionalFrameId = 0;

   /** Cache of the cursor visibility. */
   private boolean cursorVisible = true;

   /** Stores the state of the cursor at the time input mode was activated. */
   private Stack<Boolean> oldcursor = new Stack<Boolean>();

   /** Track editing block usage. */
   private boolean inEditingBlock = false;

   /** Track APPLY with implicit target usage and there is no focus. */
   private boolean implicitApplyNoTrigger = false;

   /** Track resetTerminal usage in editing blocks. */
   private boolean hadReset = false;

   /** List of frames which should be deleted at server. */
   private ArrayList<Integer> removedFrames = new ArrayList<Integer>();

   /** List of DOWN frames which were wrapped. */
   private ArrayList<Integer> wrappedFrames = new ArrayList<Integer>();

   /** Track HELP function activity */
   private boolean helpActive = false;

   /** Last frame which was displayed. */
   private Frame lastViewed = null;

   /** Counter of nested (just in case) alert boxes. */
   private int alertBoxActive = 0;

   /** Force automatic clearing. */
   private boolean afterChoose = false;

   /** Saving the last focused widget before choose() */
   private FillIn focusToRestoreFillIn = null;

   /** Tracks the LEAVE event sent status */
   private boolean leaveSent = false;

   /** Flags if the current operation had an implicit or explicit pause. */
   private boolean hadPause = false;

   /** Stack of frames targeted by WAIT-FOR statements. */
   private Deque<WaitForOp> waitForFrames = new ArrayDeque<>();

   /** Stack of frames targeted by editing blocks. */
   private Deque<Frame> editingFrames = new ArrayDeque<Frame>();

   /** Stack of event lists saved between editing blocks. */
   private Stack<EventList> eventLists = new Stack<EventList>();

   /** Stack of prior focus before an editing block started. */
   private Stack<Widget> editingFocus = new Stack<Widget>();

   /** Contains the action code of the last key input event. */
   private int lastActionCode = 0;
   
   /** Contains the action code of the last key input event. */
   private int functionKey = -1; // TODO: check if this field acts the same as last action code
   
   /** The value of the LAST-EVENT:X attribute. */
   private Integer lastEventX = null;
   
   /** The value of the LAST-EVENT:Y attribute. */
   private Integer lastEventY = null;

   /** The ID of the last non-key event raised on client-side. */
   private Integer lastEvent = null;
   
   /** Set to <code>true</code> when condition exception may be thrown. */
   private boolean condExit = false;

   /**
    * If <code>true</code> then <code>enable(..)</code> function is being run.
    */
   private boolean inEnable = false;

   /** Track input mode. */
   private int inputActive = 0;
   
   /** Custom mouse pointers by name */
   private final Map<String, MousePtrWrapper> mousePointers = new HashMap<>();

   /** Map pre-defined pointers to corresponding wrappers */
   private final Map<MousePointer, MousePtrWrapper> wrappers =
         new EnumMap<MousePointer, MousePtrWrapper>(MousePointer.class);

   /**
    * The stack of saved insert indicator states. -1 value indicates that the
    * saved state has been cleared (mostly because it has been applied to a
    * fill-in). The current states are pushed at the beginning of SET/UPDATE/
    * PROMPT-FOR/MESSSAGE UPDATE/MESSAGE SET (including EDITING) blocks, and
    * popped and their end. Note that this stack always keeps at least one
    * value (in order to restore insert indicator state after leaving
    * first-level prompting block).
    */
   private Stack<Integer> savedInsStates = new Stack<Integer>();

   /**
    * The stack of saved insert indicator states. Unlike savedInsState, saved
    * values cannot be cleared, and current states are pushed/popped only
    * for EDITING blocks.
    */
   private Stack<Integer> savedInsStateInitials = new Stack<Integer>();

   /**
    * Save pending focus (to which will be adjusted to on the next wait-for
    * loop.
    */
   private Widget pendingFocus = null;

   /** External procedure call nesting. */
   private int nesting = 0;

   /**
    * Save pending focus, when focus is disabled or hidden. In some case, it
    * will be used by the APPLY statement.
    */
   private Widget pendingFocusApply = null;

   /** Track the last displayed and enabled editor widget. */
   private Editor lastEditorDisplayed = null;

   /**
    * Flag indicating if caret draw is skipped on next cursor positioning,
    * after a wait-for loop iteration.
    */
   private boolean skipCaretDraw = false;

   /** Authentication mode flag. */
   private boolean authMode = false;

   /** Control editing mode processing. */
   private boolean ignoreNextStopEditing = false;

   /**
    * Force wiping of the screen from PUT SCREEN artifacts on next frame
    * hiding.
    */
   private boolean forceWipeScreen = false;

   /** Flag which controls behavior of message clear quirk simulation. */
   protected boolean messageClearQuirk = false;

   /** Synchronization object for pause/Ctrl-C handling. */
   private Object cleanupSynchronizer = new Object();
   /** Pause cleanup block. */
   private Runnable cleanupBlock = null;

   /** Last frame, which was made visible */
   private Frame lastExposedFrame = null;
   
   /** Track the type of the current blocking operation. */
   private BlockingOperation currentBlockingOp = BlockingOperation.NO_BLOCKING_OPERATION;
   
   /** Control behavior of some wipe-out quirks. */
   private boolean refreshQuirk = false;

   /** {@link FocusManager} instance. */
   private FocusManager focusManager = new FocusManager(this);
   
   /** Key reader and type ahead buffer processor. */
   private TypeAhead typeAhead = null;

   /** Waiter thread for SIGINT incoming. */
   private SigintWaiter sigintThread = null;
   
   /** The coordinates conversion implementation used when the output is redirected. */ 
   private CoordinatesConversion redirectedConversion = CoordinatesConversion.chuiInstance();
   
   /** Flag indicating if the client is running in windows compatibility mode. */
   private boolean underWindows;
   
   /** THREE-D value for this session. */
   private boolean is3D;
   
   /** The effective value of the keep3DFillinBorder environment variable. */
   private boolean keep3DFillinBorder;
   
   /** Active hovers stack */
   private final Stack<MouseHoverAction> hovers = new Stack<>();

   /** Most recent hovers action */
   private volatile MouseHoverAction recentlyHovered = null;
   
   /** Locked mouse pointer */
   private volatile Optional<MousePtrWrapper> lockedMousePointer = Optional.empty();

   /** Combo-box with active drop-down. */
   private ComboBox activeComboBox = null;

   /** wait-state flag */
   private boolean  isWaitState = false;

   /** The state of SESSION:IMMEDIATE-DISPLAY attribute */
   private boolean isImmediateDisplay = false;

   /** The state of SESSION:DATA-ENTRY-RETURN attribute */
   private Boolean isDataEntryReturn = null;
   
   /** The client log file */
   private FileHandler clientLogFile = null;
   
   /** The client log */
   private Logger clientLog = null;

   /**
    * Constructor.
    *
    * @param    single
    *           <code>true</code> to startup within the server process which
    *           must bypass the shared infrastructure initialization. Use
    *           <code>false</code> for the normal client JVM startup.
    */
   private ThinClient(boolean single)
   {
      triggerResultNoApply.push(false);
      triggerHadWaitFor.push(false);
      triggerNestingByWaitFor.push(0);
      savedInsStates.push(-1);

      tk = OutputManager.instance();

      String type = tk.terminalType();

      if (type != null && type.equalsIgnoreCase("vt320"))
      {
         refreshQuirk = true;
      }

      // Start external signal waiter thread
      sigintThread = new SigintWaiter(this);

      // force export of other client-side services
      fsd  = new FileSystemDaemon(single);
      sd   = new StreamDaemon(this, single);
      pd   = new ProcessDaemon(tk.getDriver().getChildProcessFactory(), single, sd);
      mmd  = new MemoryDaemon(single);
      libd = new LibraryDaemon(single);
      ospd = new OsPropertiesDaemon(single);
      
      // export our interfaces
      synchronized (ThinClient.class)
      {
         modToken = RemoteObject.registerServer(ClientExports.class, this, modToken, single);
         Class<?>[] ifaceList = new Class<?>[] { SessionExports.class };
         RemoteObject.registerStaticNetworkServer(ifaceList, SessionUtils.class);
      }

      tk.init();
   }

   /**
    * Get the context-local instance of this class. When run in a single
    * client JVM environment, this is effectively a singleton instance. Using
    * context local support allows multiple instances to co-reside in the same
    * JVM (e.g. used when the client code is hosted inside the application
    * server).
    *
    * @return   <code>null</code> if {@link #initializePrep} has not yet
    *           been called. Otherwise, the singleton instance of this class
    *           will be returned.
    */
   public static synchronized ThinClient getInstance()
   {
      return local.get();
   }
   
   /**
    * Set help text.
    * When the request comes from a widget which is part of an ongoing CHOOSE, the help text
    * is overridden with whatever was supplied for the CHOOSE operation.
    *
    * @param   text
    *          Text for help string or {@code null}, which indicates the default status message
    *          has to be restored.
    * @param   widget
    *          Optional widget for window detection. If {@code null}, the default window is
    *          selected for displaying the help message.
    */
   public static void setHelp(String text, AbstractWidget widget)
   {
      Window<?> wnd = null;
      if (widget != null)
      {
         wnd = widget.window(); 
      }
      
      if (widget == null)
      {
         wnd = WindowManager.resolveWindow();
      }
      
      wnd.setHelp(text);
   }
   
   /** Initialize client log 
    *
    * @param clientlog
    *        log file name
    */
   public static void initClientLog(String clientlog)
   {
      Logger logger = Logger.getAnonymousLogger();
      logger.setUseParentHandlers(false);
      try
      {
         FileHandler handler = new FileHandler(clientlog, 0, 1, true);
         handler.setFormatter(new LogHelper.CleanFormatter());
         logger.addHandler(handler);
         logger.setLevel(Level.INFO);
         getInstance().clientLogFile = handler;
         getInstance().clientLog = logger;
      }
      catch (Exception e)
      {
         LOG.log(Level.WARNING, "Failed to initialize client log", e);
      }
   }

   /**
    * Initialize core UI services and create or re-create the singleton
    * instance of this class.  This MUST only be called AFTER the security
    * manager and dispatcher have been initialized BUT BEFORE any call to
    * {@link #getInstance}.
    *
    * @param    cfg
    *           The bootstrap configuration file.
    * @param    driver
    *           If non-null, this is not the first init and the output
    *           manager must re-use this instance as its driver.
    * @param    single
    *           <code>true</code> to startup within the server process which
    *           must bypass the shared infrastructure initialization. Use
    *           <code>false</code> for the normal client JVM startup.
    */
   public static synchronized void initializePrep(BootstrapConfig cfg,
                                                  ScreenDriver    driver,
                                                  boolean         single)
   throws InterruptedException,
          InvocationTargetException
   {
      /* 
       * this must be synchronized on the Window.class object, as CTRL-C may
       * appear during initialization (right after the thin client was
       * constructed but before the Window was initialized), resulting in a
       * deadlock
       */ 
      synchronized (Window.class)
      {
         // initialize the driver
         OutputManager.init(cfg, driver);
         
         ThinClient instance = local.get();
         
         // initialize the thin client and export its APIs
         if (instance == null)
         {
            // the process is just starting up
            ThinClient tc = new ThinClient(single);
            local.set(tc);

            // typeahead needs to be initialized only after the window has been created/selected
            tc.typeAhead = TypeAhead.init(tc);
         }
         else
         {
            // reset our state because we are going to initialize again
            OutputManager<?> tk = OutputManager.instance();

            if (tk.isChui())
            {
               ((ChuiOutputManager) tk).setBitmap(null);
            }

            // reset other state
            ErrorManager.resetClient();
            
            ThinClient tc = new ThinClient(single);
            local.set(tc);

            // typeahead needs to be initialized only after the window has been created/selected
            tc.typeAhead = TypeAhead.init(tc);
            
            // Reset the error writer
            OutputManager.initErrorWriter();
         }
      }
   }

   /**
    * Initialize network services (must only be called after the user has been authenticated AND
    * after {@link #initializePrep} has been called).
    *
    * @param   cfg
    *          The configuration to use for setup of the client.
    * @param   session
    *          The connection to the server. {@code null} if this is running on the server side
    *          (and thus there is no network session).
    * @param   context
    *          The security context (if non-null). This is only used when run on the server. This
    *          context must be propagated to other threads which need to access context-specific
    *          state.
    */
   public static void initializePost(BootstrapConfig cfg, Session session, Object context)
   {
      ThinClient tc = getInstance();

      boolean single = (session == null);
      
      // initialize the exposed socket services
      LowLevelSocketListenerImpl.initialize(cfg, session, single);
      LowLevelSocketImpl.initialize(cfg, session, single);
      // initialize the exposed WebServices interface
      WebServiceImpl.initialize(session, single);

      if (session != null)
      {
         ServerPropertiesInspector spInsp = null;
         spInsp = (ServerPropertiesInspector) RemoteObject.obtainNetworkInstance(
            ServerPropertiesInspector.class, session);

         tc.underWindows = spInsp.isUnderWindowsFamily();

         // standalone/remote client implementation
         
         tc.server = (ServerExports)
            RemoteObject.obtainNetworkInstance(ServerExports.class, session);
   
         RemoteErrorData errorServer = (RemoteErrorData)
            RemoteObject.obtainNetworkInstance(RemoteErrorData.class, session);
   
         ErrorManager.initClient(errorServer);
         
         // This call must be here, after connection to the server is ready
         // Only initialize the environment subsystem when running on Windows
         // Now we can use OsPropertiesDaemon to check this
         
         // allow the daemon to process stanza ini files regardless of current OS 
         tc.ed = new EnvironmentDaemon(session, 
                                       single,
                                       tc.ospd.getOperatingSystem().contains("win"));
         
         if (!OutputManager.instance().isChui())
         {
            // initialize default window when we have server connection, for example font manager 
            // is used during this call
            WindowManager.resetWindow();
            WindowManager.getDefaultWindow();
         }
         
         session.registerSynchronizer(tc);
      }
      else
      {
         // server-side implementation
         
         tc.underWindows = EnvironmentOps.isUnderWindowsFamily();
         
         // force the key reader thread to the proper context
         tc.typeAhead.setPendingContextSwitch(context);
         
         // client code is "right" side by convention
         RemoteObject.registerRightSynchronizer(tc);
         
         tc.server = (ServerExports) RemoteObject.obtainLocalInstance(ServerExports.class, false);
            
         // remember that this is single mode by storing a reference to the
         // dedicated thread for this user (same as the current thread at the
         // time of this method call)
         tc.dedicatedThread = Thread.currentThread();
      }
      
      // register the config classes
      ConfigManager.<ClientConfigManager>getInstance().pushConfigDefs(tc.server);
      
      // negotiate the UI theme (before requesting fonts and colors from server)
      ThemeManager.init(tc.server, cfg.getString("client", "driver", "theme", null));
      
      // initialize the font manager
      FontManager.init(tc.server);
      
      // initialize the client side widget ID worker
      WidgetIdHelper.initWorker(new ClientIdHelper());
   }
   
   /**
    * The method is called when the environment ({@link EnvironmentDaemon}) has
    * been fully initialized.
    * 
    * @param    cfg
    *           A valid bootstrap configuration.
    */
   public static void loadEnvironment(BootstrapConfig cfg)
   {
      ThinClient tc = getInstance();
      
      if (!OutputManager.instance().isChui())
      {
         Window<?> defw = WindowManager.getDefaultWindow();

         // set the initial configuration of the default window
         
         EnvironmentDaemon ed = getInstance().getEnvironmentDaemon();

         // default settings from the environment
         String srows = (ed == null ? null : ed.getKeyValue("Default Window", "rows"));
         String scols = (ed == null ? null : ed.getKeyValue("Default Window", "columns"));
         String sx    = (ed == null ? null : ed.getKeyValue("Default Window", "x"));
         String sy    = (ed == null ? null : ed.getKeyValue("Default Window", "y"));
         
         // client override settings
         String crows = cfg.getString("client", "gui", "rows",    null);
         String ccols = cfg.getString("client", "gui", "columns", null);
         String cx    = cfg.getString("client", "gui", "x",       null);
         String cy    = cfg.getString("client", "gui", "y",       null);
         
         // client-side settings have priority over the environment
         int rows = (crows == null || crows.length() == 0 
                       ? srows == null || srows.length() == 0 ? 25
                                                              : Integer.parseInt(srows.trim())
                       : Integer.parseInt(crows.trim()));
         int cols = (ccols == null || ccols.length() == 0
                       ? scols == null || scols.length() == 0 ? 80
                                                              : Integer.parseInt(scols.trim())
                       : Integer.parseInt(ccols.trim()));
         int x = (cx == null || cx.length() == 0
                    ? sx == null || sx.length() == 0 ? 0
                                                     : Integer.parseInt(sx.trim())
                    : Integer.parseInt(cx.trim()));
         int y = (cy == null || cy.length() == 0
                    ? sy == null || sy.length() == 0 ? 0
                                                     : Integer.parseInt(sy.trim())
                    : Integer.parseInt(cy.trim()));
         
         WindowManager.setDefaultDimension(rows - Window.RESERVED_LINES, cols);
         WindowManager.setInitialDimension(defw);

         defw.setPhysicalLocation(x, y);
         
         // get the default value for the "use-3d-size" setting - if a custom env exists, that
         // is the value which will be used
         boolean use3dSize = true;
         String cuse3dSize = cfg.getString("client", "gui", "use-3d-size", null);
         if (cuse3dSize != null && !cuse3dSize.isEmpty())
         {
            use3dSize = Boolean.parseBoolean(cuse3dSize);
         }
         else
         {
            use3dSize = ed.getBooleanKeyValue("Startup", "use-3d-size", use3dSize);
         }
         FontManager.setThreeD(use3dSize);

         // get the value for the "Keep3DFillinBorder" environment variable
         boolean keep3DFillinBorder = ed != null &&
               ed.getBooleanKeyValue("Startup", "Keep3DFillinBorder", false);
         tc.setKeep3DFillinBorder(keep3DFillinBorder);
      }
   }   
   
   /**
    * Initialize initial tooltip settings
    */
   public static void setSessionTooltips()
   {
      ToolTip.setSessionTooltips(getInstance().server.getSessionTooltips());
   }

   /**
    * Clears all content from the terminal.
    * This is a direct output operation that clears the screen through
    * ncurses.
    */
   private void clearTerminal()
   {
      if (tk.isRedirected() || !isChui())
         return;

      // these are direct output calls, only for ChUI
      ((ChuiOutputManager) tk).clearScreen();
   }
   
   /**
    * When run in-process on the server, the client has threads with security
    * context that must be shutdown. This method synchronously causes the
    * threads to exit and returns when they are gone.
    */
   public static void terminate()
   {
      ThinClient tc = getInstance();
      
      tc.typeAhead.closeReader();
      tc.sigintThread.shutdown();

      WindowManager.clearWindowList();
      
      ClientConfigManager mgr = ConfigManager.getInstance();
      mgr.deactivateConfigUpdates();
      
      WindowManager.resetWindow();
      
      if (tc.clientLogFile != null)
      {
         tc.clientLogFile.close();
         tc.clientLog.removeHandler(tc.clientLogFile);
         tc.clientLogFile = null;
         tc.clientLog = null;
      }
   }

   /**
    * Temporarily halts the client UI processing (key reading, normal drawing
    * mode)  After entering suspend mode, the current contents of the
    * terminal are cleared.  Use {@link #resume} to return to normal mode.
    */
   public void suspend()
   {
      // deactivate batch mode processing if we are in the batch mode
      OutputManager.activateBatch(false);
      
      typeAhead.suspend();
      Window.suspend();
      clearTerminal();

      tk.suspend();
   }

   /**
    * Check if wait state is active
    * 
    * @return <code>true</code> if wait state is active
    */
   public static boolean isWaitState()
   {
      return getInstance().isWaitState;
   }
   
   /**
    * (De)register combo-box with active drop down.
    * 
    * @param activeComboBox
    *        COMBO-BOX with active drop down or <code>null</code> when deacticated
    * @param widgetId
    *        the ID of the widget which caused locking
    * @param ptr
    *        locked mouse pointer
    */
   public void setActiveComboBox(ComboBox activeComboBox, int widgetId, MousePtrWrapper ptr)
   {
      this.activeComboBox = activeComboBox;
      lockMousePointer(widgetId, ptr);
   }

   /**
    * Reset the active combo box global reference and the locked mouse pointer that has been
    * used by the active drop down.
    */
   public void resetActiveComboBox()
   {
      setActiveComboBox(null, -1, null);
      refreshHover();
   }

   /**
    * Lock mouse pointer
    * 
    * @param widgetId
    *        the ID of the widget which caused locking
    * @param ptr
    *        locked mouse pointer
    */
   public void lockMousePointer(int widgetId, MousePtrWrapper ptr)
   {
      if (!isWaitState || (ptr != null))
      {
         tk.getDriver().lockMousePointer(widgetId, ptr);
         lockedMousePointer = Optional.ofNullable(ptr);
      }
   }

   /**
    * Check if mouse pointer is locked
    * 
    * @return <code>true</code> if mouse pointer is locked
    */
   public static boolean isMousePointerLocked()
   {
      return getInstance().lockedMousePointer.isPresent();
   }

   /**
    * Get locked mouse pointer if it is locked
    * 
    * @return locked mouse pointer if it is locked
    */
   public static Optional<MousePtrWrapper> lockedMousePointer()
   {
      return getInstance().lockedMousePointer;
   }

   /**
    * Get the combo-box with active drop-down.
    * 
    * @return the active combo-box.
    */
   public ComboBox<?> activeComboBox()
   {
      return activeComboBox;
   }
   
   /** 
    * Push action of the currently hovered widget on the stack
    * 
    * @param action
    *        action of the currently hovered widget on the stack
    */
   public synchronized void pushHover(MouseHoverAction action)
   {
      hovers.push(action);
      recentlyHovered = action;
   }

   /** 
    * Pop action of the recently hovered widget off the stack
    *
    * @return
    *        action of the recently hovered widget or <code>null</code> if the stack is empty
    */
   public synchronized MouseHoverAction popHover()
   {
      if (!hovers.isEmpty())
      {
         recentlyHovered = hovers.pop();
      }
      else
      {
         recentlyHovered = null;
      }
      return recentlyHovered;
   }
   
   /**
    * Refresh mouse cursor shape for the currently hovered widget
    */
   public synchronized void refreshHover()
   {
      Window<?> wnd = WindowManager.getActiveWindow();
      
      if (recentlyHovered != null && recentlyHovered.windowId() == wnd.getId().asInt())
      {
         recentlyHovered.updateCursor();
      }
      else
      {
         if (wnd instanceof WindowGuiImpl)
         {
            try
            {
               GuiDriver gd = (GuiDriver) OutputManager.getDriver();
               gd.selectWindow(wnd.getId().asInt());
               try
               {
                  gd.setCursor(wnd.getEffectiveMousePointer());
               }
               finally
               {
                  gd.releaseWindow();
               }
            }
            catch (Exception e)
            {
               LOG.log(Level.WARNING, "Unexpected exception", e);
            }
         }
      }
   }

   /**
    * Re-enables client UI processing (key reading, normal drawing mode).
    * Before resuming the current contents of the terminal are cleared.  This
    * is the mechanism to return to normal processing mode after having
    * called {@link #suspend}.
    */
   public void resume()
   {
      tk.resume();
      Window.resume();
      typeAhead.resume();
      refreshScreen();
      
      // reactivate batch mode if was disabled before
      OutputManager.activateBatch(true);
   }

   /**
    * Redraw everything after suspending terminal or after use of terminal
    * stream.
    */
   public void refreshScreen()
   {
      // if terminal is not redirected, redraw everything because
      // we have cleared terminal in suspend().
      // WARNING!!!
      // PUT SCREENs will not survive this redrawing!
      // whether this constitutes a problem or a solution remains to be seen
      // with a testcase.
      if (!tk.isRedirected())
      {
         Window<?> wnd = WindowManager.getDefaultWindow();
         wnd.refresh(wnd.bounds());
         wnd.sync();
      }
   }

   /**
    * Indefinitely pause and display the default message. This method works
    * when in suspend mode and it will not trigger a full redraw of the
    * entire screen.
    *
    * @return   The key code of the key pressed to clear the pause or -1
    *           on any failure.
    */
   public int suspendedPause()
   {
      tk.resetMode();

      setNeedPause(false, WindowManager.resolveWindow());

      // honor pause before-hide setting
      if (!honorPBH || timePBH == 0)
      {
         return lastKey;
      }

      KeyInput key = null;

      // the pause itself (reads keys directly)
      do
      {
         key = EventManager.eventFromKey(tk.readKey());
      }
      while (key == null);

      // maintain lastKey properly
      lastKey = key.keyCode();

      try
      {
         // honor CTRL-C processing
         if (lastKey == Key.VK_CTRL_C)
         {
            lastKey = -1;
            throw new StopConditionException("CTRL-C");
         }

         // honor the STOP, ENDKEY and END-ERROR actions
         processCondition(key.actionCode(), -1, true);
      }
      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate(WindowManager.resolveWindow());

         // let the original exception propagate
         throw ce;
      }

      return lastKey;
   }

   /**
    * Send message to stream.                             
    *
    * @param   message
    *          Message to send.
    *
    * @return  redirection status.
    */
   public int sendToStream(String message)
   {
      int redir;
      Stream out = streamFromId(currentStream);

      // if the terminal is redirected, then send this message to the
      // stream
      if (out != null && out.isUnnamed())
      {
         // assumes that we are in a reset state
         ((ChuiOutputManager) tk).append(message, Color.DEFAULT);
         flushStream(out);
      }

      // the message is displayed on the terminal only if the input is not
      // read from a null device, like /dev/null

      // switch back to interactive mode temporarily, to allow the message
      // to be displayed on the terminal
      redir = switchOutput(-1, false);
      return redir;
   }

   /**
    * Process pause before hide as requested.
    */
   public void pauseBeforeHide()
   {
      if (honorPBH)
         pause(timePBH, textPBH, false, WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Process pause after page for terminal stream.
    */
   public void pauseAfterPage()
   {
      pause(timePBH, textPBH, false, WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Applies the encoded event to the widget currently in focus in the given
    * frame, or directly to the specified widget.
    *
    * @param    widgetId
    *           ID of frame or a contained widget.
    * @param    eventCode
    *           Progress-compatible event code.
    * @param    inbuf
    *           The screen buffer for the related frame if the widgetId is
    *           from a widget that is a frame or is contained within a frame.
    *           May be <code>null</code> if the widget is not associated with
    *           any frame (e.g. the session window).
    * @param    eventList
    *           The list of the latest triggers (which may be more recent
    *           than that used in any previous <code>WAIT-FOR</code>).
    * @param    explicit
    *           <code>true</code> if this APPLY call has an explicit target.
    *
    * @return   Updated data for the associated frame representing the current
    *           screen values of all widgets in that frame.  May be
    *           <code>null</code> if the frame has no data containing widgets
    *           or if the input screen buffer is <code>null</code> or is
    *           otherwise invalid.
    */
   @Override
   public ScreenBuffer[] apply(int            widgetId,
                               int            eventCode,
                               ScreenBuffer[] inbuf,
                               EventList      eventList,
                               boolean        explicit)
   {
      // when a reset occurs inside an editing block, apply is a NOP
      if (inEditingBlock && hadReset)
         return null;

      Widget saveFocus  = UiUtils.getCurrentFocus();
      Widget focusFrame = UiUtils.locateFrame(saveFocus);
      int focusState    = FocusManager.focusState(saveFocus);

      if (!(focusFrame instanceof Frame))
      {
         focusFrame = null;
      }

      // check if this is the NEXT-FRAME or PREV-FRAME event
      boolean frameChangeEvent = eventCode == Keyboard.SE_NEXT_FRAME ||
                                 eventCode == Keyboard.SE_PREV_FRAME;

      // determine pending focus - pending from hidden/disabled widget has
      // priority over pending from hidden frame
      Widget pending = pendingFocus == null ? pendingFocusApply : pendingFocus;

      boolean isFocusFrameHidden = focusFrame != null &&
                                   !focusFrame.isVisible();
      boolean isFocusInvalid     = focusState == FocusManager.FOCUS_INVALID &&
                                   !isFocusFrameHidden;
      boolean isFocusHidden  = focusState == FocusManager.FOCUS_HIDDEN || isFocusFrameHidden;
      boolean isFocusValid   = !isFocusHidden && !isFocusInvalid;
      boolean isPendingValid = adjustFocusBeforeProcessing(explicit,
                                                           saveFocus,
                                                           focusFrame,
                                                           focusState,
                                                           frameChangeEvent);

      // null input buffer just silently returns here
      setScreenBuffers(inbuf);

      Widget inFocus = locateInFocus(widgetId);
      int inFocusState = FocusManager.focusState(inFocus);
      Widget tmpFrame = UiUtils.locateFrame(inFocus);

      if (tmpFrame instanceof Frame)
         inFocus = ((Frame) tmpFrame).getField(inFocus);

      if (triggerNesting > 0              &&
          (inFocus instanceof FixedSizeContainer || inFocus instanceof Button) &&
          inFocus.isVisible()             &&
          inFocus.isEnabled()             &&
          (tmpFrame == null ||
           tmpFrame.isVisible()))
      {
         // no pending is used after trigger invocation if target is valid
         // before trigger invocation
         pending = null;
         isPendingValid = false;
      }

      KeyInput event = new KeyInput(inFocus, EventType.KEY_TYPED, eventCode);

      // setup event list chain
      EventList save = currentEventList;
      eventList.setChain(currentEventList);
      
      // apply has no operation of its own, make sure to propagate the active operation type
      if (currentEventList != null)
         eventList.setOperation(currentEventList.getOperation());

      currentEventList = eventList;
      lastAppliedCode = eventCode;

      // if this is an APPLY with implicit target and target is not a valid
      // widget, then no trigger should be invoked
      boolean saveImplicit = implicitApplyNoTrigger;

      try
      {
         implicitApplyNoTrigger = !explicit && widgetId == -1;

         tk.setInvalidate(true);
         applyWorker(event);
         tk.setInvalidate(event.source(), false, true);

         if (triggerNesting > 0 && !frameChangeEvent)
         {
            boolean targetValid = (inFocus instanceof FixedSizeContainer ||
                                   inFocus instanceof Button) &&
                                  inFocus.isVisible()         &&
                                  inFocus.isEnabled()         &&
                                  (tmpFrame == null || tmpFrame.isVisible());
                                  
            if (!explicit)
            {
               Widget currentFocus = UiUtils.getCurrentFocus();
               
               // this is being removed from refocusing because cases have been found that
               // require the focus change to be left alone (e.g. AUTO-RETURN caused by
               // a non-explicit APPLY)
               if (targetValid && currentFocus != null)
               {
                  int cfid = currentFocus.getId().asInt();
                  
                  if (cfid != widgetId)
                  {
                     // in the 4GL, there is a case where AUTO-RETURN that fails and forces
                     // the focus back to the original widget (if the AUTO-RETURN was generated
                     // by an APPLY from inside a nested trigger AND that trigger does NOT
                     // consume the original key, in the 4GL this leads to a weird state where
                     // the focus is forced back to the widget BUT the normal activation/zapping
                     // is suppressed
                     if (inFocus instanceof FillIn && event.isCausedAutoReturn())
                     {
                        ((FillIn) inFocus).setInAutoReturn(true);
                     }
                     
                     // the following can be used to find the impact of removing the test for 
                     // currentFocus.config().id.getId() != widgetId from the refocusing section
                     // below; in testing, no negative effects have been seen
                     // String spec  = "NON-EXPLICIT APPLY FOCUS CHANGE %d to %d in frame %s\n";
                     // String fname = tmpFrame == null ? "none" : tmpFrame.toString();
                     // System.err.printf(spec, widgetId, cfid, fname);
                     // Thread.currentThread().dumpStack();
                  }
               }
               
               // TODO: this code is probably no longer needed, the logging should help determine
               //       if it is needed
               if (targetValid && currentFocus == null)
               {
                  String spec  = "NON-EXPLICIT APPLY REFOCUS current null to %d in frame %s\n";
                  String fname = tmpFrame == null ? "none" : tmpFrame.toString();
                  System.err.printf(spec, widgetId, fname);
                  
                  refocusTarget(widgetId, inFocus);
               }
            }
            else
            {
               if (!targetValid)
               {
                  adjustFocusForTriggerMode(pending, isPendingValid);
               }
               else
               {
                  // else, if target is valid, adjust focus only if focus was
                  // not valid before trigger invocation
                  if (!isFocusValid)
                  {
                     if (tmpFrame != focusFrame && !isFocusHidden)
                     {
                        // adjust to target only if is not in the same frame
                        // as the previous focus and the focus was not hidden
                        // before trigger invocation
                        focusManager.setFocusOn(inFocus);
                     }
                     else
                     {
                        // else, restore focus to previous one, only if this
                        // is not the ENTRY event
                        if (event.actionCode() != Keyboard.SE_ENTRY)
                        {
                           focusManager.setFocusOn(saveFocus);
                        }
                     }
                  }
               }
            }
         }
      }
      finally
      {
         // restore flag, even if a condition is raised
         implicitApplyNoTrigger = saveImplicit;
      }

      handleFillInEntry(eventCode, inFocus);

      lastAppliedCode = KeyInput.CHAR_UNDEFINED;
      // restore event list chain to its previous state
      currentEventList = save;

      return inbuf == null ? null
                           : getEditableClientScreenBuffers(
                                getScreenBuffer(inbuf[0].getBaseId(), inbuf[0].size(), true));
   }
   
   /**
    * Apply this event to a non-widget resource.
    * 
    * @param    resId
    *           The resource ID to which this events needs to be applied.
    * @param    eventCode
    *           Progress-compatible event code.
    * @param    eventList
    *           The list of the latest triggers (which may be more recent than that used in any 
    *           previous <code>WAIT-FOR</code>).
    */
   @Override
   public void apply(long resId, int eventCode, EventList eventList)
   {
      // when a reset occurs inside an editing block, apply is a NOP
      if (inEditingBlock && hadReset)
         return;

      // setup event list chain
      EventList save = currentEventList;
      eventList.setChain(currentEventList);

      currentEventList = eventList;
      lastAppliedCode = eventCode;

      // if this is an APPLY with implicit target and target is not a valid
      // widget, then no trigger should be invoked
      boolean saveImplicit = implicitApplyNoTrigger;

      try
      {
         ResourceEvent event = new ResourceEvent(WindowManager.getCurrentWindow(), resId, eventCode);
         
         tk.setInvalidate(true);
         applyWorker(event);
         tk.setInvalidate(event.source(), false, true);
      }
      finally
      {
         // restore flag, even if a condition is raised
         implicitApplyNoTrigger = saveImplicit;
      }

      lastAppliedCode = KeyInput.CHAR_UNDEFINED;
      // restore event list chain to its previous state
      currentEventList = save;
   }
   

   /**
    * Locate widget in focus.
    *
    * @param   widgetId
    *          Widget ID to start search from.
    * @return  reference to widget in focus.
    */
   private Widget locateInFocus(int widgetId)
   {
      Widget inFocus = getWidget(widgetId);

      if (inFocus == null && widgetId != -1)
      {
         inFocus = getFrame(widgetId);
      }

      if (inFocus == null)
      {
         // the source of the event can be one of the dynamically
         // generated widgets
         inFocus = UiUtils.getCurrentFocus();
      }

      if (inFocus == null)
      {
         // OK, we have no frames and/or no focused widgets so use this as
         // the last resort (it will always exist), otherwise the super
         // constructor of KeyInput() will fail badly (on null)
         inFocus = WindowManager.getDefaultWindow();
      }
      return inFocus;
   }

   /**
    * Adjust focus before processing event.
    *
    * @param   explicit
    *          Explicit event flag.
    * @param   saveFocus
    *          Saved old focused widget.
    * @param   focusFrame
    *          Frame which holds focused widget.
    * @param   focusState
    *          Focus state.
    * @param   frameChangeEvent
    *          <code>true</code> indicates frame change event.
    *
    * @return  <code>true</code> if pending focus is valid.
    */
   private boolean adjustFocusBeforeProcessing(boolean explicit,
                                               Widget saveFocus,
                                               Widget focusFrame,
                                               int focusState,
                                               boolean frameChangeEvent)
   {
      boolean isFocusFrameHidden = focusFrame != null &&  !focusFrame.isVisible();
      boolean isFocusInvalid     = focusState == FocusManager.FOCUS_INVALID && !isFocusFrameHidden;
      boolean isFocusHidden  = focusState == FocusManager.FOCUS_HIDDEN || isFocusFrameHidden;

      boolean isPendingValid = false;

      Widget pending = pendingFocus == null ? pendingFocusApply : pendingFocus;

      // NEXT-FRAME and PREV-FRAME have a special behavior - no focus adjustment
      // takes place prior to event processing
      if (triggerNesting > 0 && !frameChangeEvent)
      {
         if (pending != null)
         {
            Widget pendingFrame = UiUtils.locateFrame(pending);

            if (pendingFrame instanceof Frame && ((Frame) pendingFrame).isAlive())
            {
               // determine if pending focus is valid - only if its enabled
               // and parent frame is visible. hidden widget state is OK
               isPendingValid = pending.isEnabled() && pendingFrame.isVisible();
            }
         }

         if (!explicit && saveFocus != null && focusFrame != null)
         {
            // must adjust focus if focus is disabled and implicit target is
            // used
            if (isFocusInvalid)
            {
               // if current focus is invalid, search for focus
               focusManager.searchForFocusWithoutEntry(saveFocus);
            }
            else
            {
               // if pending focus is valid and focus is hidden, set focus on
               // found pending
               if (isPendingValid && isFocusHidden)
               {
                  focusManager.setFocusOn(pending);
                  // reset all pending
                  pendingFocus = pendingFocusApply = null;
                  isPendingValid = false;
               }
            }
         }
      }
      return isPendingValid;
   }

   /**
    * Set focus to target.
    *
    * @param   widgetId
    *          Widget ID.
    * @param   inFocus
    *          Focus target.
    */
   private void refocusTarget(int widgetId, Widget inFocus)
   {
      // implicit case, refocus target, except when executed inside an editing block
      if (!inEditingBlock && widgetId != -1)
      {
         focusManager.setFocusOn(inFocus);
      }
   }

   /**
    * Adjust focus when processing nested triggers.
    *
    * @param   pending
    *          Pending focus target.
    * @param   isPendingValid
    *          flag indicating valid pending focus target.
    */
   private void adjustFocusForTriggerMode(Widget pending, boolean isPendingValid)
   {
      // focus must be adjusted if we are in trigger nesting mode,
      // and:
      // 1. target is not valid
      // 2. there was a valid pending focus before trigger
      if (isPendingValid)
      {
         focusManager.setFocusOn(pending);
         // reset all pending
         pendingFocus = pendingFocusApply = null;

         // if no valid pending, leave focus where it is
      }
   }

   /**
    * Process event for {@link FillIn} widgets.
    *
    * @param   eventCode
    *          Event code.
    * @param   inFocus
    *          Widget in focus.
    */
   private void handleFillInEntry(int eventCode, Widget inFocus)
   {
      if (eventCode == Keyboard.SE_ENTRY)
      {
         // For the APPLY "ENTRY" statement, we should enable zap and reset
         // insert mode for the ENTRY target (fill-in), or to the currently
         // focused fill-in if ENTRY has returned NO-APPLY.

         FillIn fin = null;
         Widget focus = UiUtils.getCurrentFocus();

         if (triggerResultNoApply.peek() && focus != null)
         {
            if (focus instanceof FillIn)
            {
               fin = (FillIn) focus;
            }
         }
         else if (inFocus != null && inFocus instanceof FillIn)
         {
            fin = (FillIn) inFocus;
         }

         enableFillInAutoZap(fin);
      }
   }

   /**
    * Rings a bell on the terminal.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * BELL
    * </pre>
    */
   @Override
   public void bell()
   {
      if (!isRedirected())
         tk.beep();
   }

   /**
    * Performs CHOOSE processing.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CHOOSE</b>
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    c
    *           CHOOSE statement options
    * @param    wl
    *           Event list.
    *
    * @return   result of CHOOSE processing.
    */
   @Override
   public ChooseData choose(int frameId, final ChooseData c, EventList wl)
   {
      // must be called first!
      honorConditionalUpDown(false);

      final Frame frame = getFrame(frameId);

      final Integer[] focusId = new Integer[1];

      final Widget[] comps = prepareChooseWidgetList(c, frame, focusId);

      enterChoose(c, frame, focusId, comps);

      int toFocusId = focusId[0];

      setNeedPause(false, frame.window());

      ChooseData cdret = chooseWorker(c, wl, frame, comps, toFocusId);

      // force automatic clearing of messages in any case.
      afterChoose = true;

      return cdret;
   }

   /**
    * CHOOSE processor.
    *
    * @param   c
    *          Input CHOOSE parameters.
    * @param   wl
    *          Event list.
    * @param   frame
    *          Target frame.
    * @param   comps
    *          Widget list.
    * @param   toFocusId
    *          Target focus ID.
    *
    * @return  result of CHOOSE processing.
    */
   private ChooseData chooseWorker(final ChooseData c, EventList wl,
                                   final Frame frame, final Widget[] comps,
                                   int toFocusId)
   {
      ChooseData cdret = null;

      ScreenBuffer[] sb  = c.getScreenBuffers();
      ScreenBuffer[] sbo = null;

      lastViewed = frame;

      boolean saveEnabled = frame.isEnabled();

      try
      {
         // enable frame if necessary
         if (!saveEnabled)
            frame.setEnabled(true);

         BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.CHOOSE);

         // note that in CHOOSE WAIT-FOR does not set status line
         sbo = waitForWorker(wl,
                             toFocusId,
                             c.getPause(),
                             sb,
                             true,
                             false,
                             false,
                             BlockingOperation.CHOOSE);
      }

      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate(frame);

         // let the original exception propagate
         throw ce;
      }

      finally
      {
         cdret = restoreEnabledAfterChoose(c, frame, comps, sbo, saveEnabled);
      }

      if (frame.isInternalGo())
         setNeedPause(true, frame.topLevelWindow());
      return cdret;
   }

   /**
    * Restore enabled widget after CHOOSE processing.
    * @param   c
    *          Input CHOOSE data.
    * @param   frame
    *          Target frame.
    * @param   comps
    *          Array of widgets to process.
    * @param   sbo
    *          Output screen buffers array.
    * @param   saveEnabled
    *          Saved frame enabled flag.
    *
    * @return  Updated CHOOSE data.
    */
   private ChooseData restoreEnabledAfterChoose(final ChooseData c, final Frame frame,
                                                final Widget[] comps, ScreenBuffer[] sbo,
                                                boolean saveEnabled)
   {
      ChooseData cdret;
      // restore enabled state
      if (saveEnabled != frame.isEnabled())
         frame.setEnabled(saveEnabled);

      if (c.isFieldMode())
      {
         // disable all involved fields unconditionally
         eventBracket(true, new Runnable()
         {
            @Override
            public void run()
            {
               for (int i = 0; i < comps.length; i++)
               {
                  comps[i].setEnabled(false);
               }
            }
         });
      }

      cdret = frame.getChooseData();
      cdret.setScreenBuffers(sbo);
      lastViewed = frame;
      return cdret;
   }

   /**
    * Start CHOOSE processing.
    *
    * @param   c
    *          CHOOSE parameters.
    * @param   frame
    *          Target frame.
    * @param   focusId
    *          List of focus IDs.
    * @param   comps
    *          Widget list.
    */
   private void enterChoose(final ChooseData c, final Frame frame,
                          final Integer[] focusId, final Widget[] comps)
   {
      if (focusToRestoreFillIn == null &&
          pendingFocus != null &&
          pendingFocus instanceof FillIn)
      {
         focusToRestoreFillIn = (FillIn) pendingFocus;
      }

      eventDrawingBracket(frame, new Runnable()
      {
         @Override
         public void run()
         {
            focusId[0] = frame.enterChoose(comps, c, focusId[0]);
         }
      });
   }

   /**
    * Prepare list of widgets involved in CHOOSE.
    *
    * @param   c
    *          CHOOSE parameters.
    * @param   frame
    *          Target frame.
    * @param   focusId
    *          List of focused widget IDs.
    *
    * @return  List of widgets.
    */
   private Widget[] prepareChooseWidgetList(final ChooseData c, final Frame frame,
                                            final Integer[] focusId)
   {
      // honor the outstanding next-prompt request if any
      int toFocusId = frame.getFocusId();
      frame.setFocusId(-1);

      Widget[] list = null;

      if (c.isFieldMode())
      {
         int[] widgetIds = c.getFields();
         list = new Widget[widgetIds.length];

         enableChooseWidgets(c, list);

         // honor the outstanding next-prompt request if any
         toFocusId = handleOutstandingNextPrompt(toFocusId, widgetIds);
      }
      else
      {
         list = new Widget[1];

         list[0] = getWidget(c.getRowId());

         // set row mode help text
         setHelp(c.getRowHelp(), frame);
      }

      focusId[0] = toFocusId;
      return list;
   }

   /**
    * Enable widgets involved in CHOOSE.
    *
    * @param   c
    *          Input CHOOSE data.
    * @param   list
    *          List of widgets to enable.
    */
   private void enableChooseWidgets(final ChooseData c, Widget[] list)
   {
      final Widget[] comps = list;

      // access and enable all involved fields unconditionally
      eventBracket(true, new Runnable()
      {
         @Override
         public void run()
         {
            for (int i = 0; i < comps.length; i++)
            {
               comps[i] = getWidget(c.getField(i));
               comps[i].setEnabled(true);
            }
         }
      });
   }

   /**
    * Handle outstanding NEXT-PROMPT clause.
    *
    * @param   toFocusId
    *          ID of pending widget.
    * @param   widgetIds
    *          Widget ID list.
    *
    * @return  final target widget ID
    */
   private int handleOutstandingNextPrompt(int toFocusId, int[] widgetIds)
   {
      if (toFocusId != -1)
      {
         // validate the widget
         toFocusId = validateWidget(toFocusId, widgetIds);
         Widget widget = getWidget(toFocusId);

         if (widget instanceof FillIn)
         {
            ((FillIn) widget).resetCaret();
         }
      }
      return toFocusId;
   }

   /**
    * Clears the data and colors for all fill-in fields of the frame. For
    * down frames, also clears side labels.
    * <p>
    * This method corresponds to this Progress statement:<br>
    * <b>CLEAR</b>
    *
    * @param    frameId
    *           The ID of the frame to clear.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used (no pause will
    *           occur even if the client detects that the user has had no
    *           chance to review the screen's contents).
    *
    * @return   The screen buffer array for all editable frames.
    */
   @Override
   public ScreenBuffer[] clear(int frameId, boolean noPause)
   {
      return getEditableClientScreenBuffers(
                clearWorker(getFrame(frameId), noPause, false, true));
   }

   /**
    * Clears all occurrences for the down frame and resets the current display
    * position to the top of the frame.
    * <p>
    * This method corresponds to this Progress statement:<br>
    * <b>CLEAR ALL</b>
    *
    * @param    frameId
    *           The ID of the frame to clear.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used (no pause will
    *           occur even if the client detects that the user has had no
    *           chance to review the screen's contents).
    *
    * @return   The screen buffer array for all editable frames.
    */
   @Override
   public ScreenBuffer[] clearAll(int frameId, boolean noPause)
   {
      return getEditableClientScreenBuffers(clearAll(frameId, noPause, true));
   }

   /**
    * Clears all occurrences for the down frame and resets the current display
    * position to the top of the frame.
    * <p>
    * This method corresponds to this Progress statement:<br>
    * <b>CLEAR ALL</b>
    *
    * @param    frameId
    *           The ID of the frame to clear.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used (no pause will
    *           occur even if the client detects that the user has had no
    *           chance to review the screen's contents).
    * @param   resolveBuffer
    *          When <code>true</code>, resolve the frame's screen buffer and return it.
    * 
    * @return   The frame's screen buffer.
    */
   private ScreenBuffer clearAll(int frameId, boolean noPause, boolean resolveBuffer)
   {
      Frame frame = getFrame(frameId);
      
      tk.setInvalidate(true);
      ScreenBuffer sb = clearWorker(frame, noPause, true, resolveBuffer);
      frame.clear();
      tk.setInvalidate(frame, false, true);
      frame.setAllClear(true);
      
      if (sb == null)
      {
         sb = new ScreenBuffer();
      }
      sb.setDownConfigs(frame.getDownConfigs());
      
      return sb;
   }

   /**
    * Implements the <b>DESELECT-ROWS() </b> <b>BROWSE </b> widget method.
    *
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   @Override
   public void deselectRows(final int browseId)
   {
      final Browse<?> browse = (Browse<?>) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.deselectAll();
            }
         });
      }
   }

   /**
    * Notify the client that columns have been added to the browse.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    */
   @Override
   public void notifyColumnsUpdated(final int browseId)
   {
      final Browse<?> browse = (Browse<?>) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.notifyColumnsUpdated();
            }
         });
      }
   }

   /**
    * Delete the line of text in which the targeted editor's cursor currently resides.  Leave
    * the cursor at the leftmost position of the same line number (which will be the line
    * previously below this one if the cursor is not already on the last line) or at the
    * leftmost position of the last line (if there are no lines below this one).
    *
    * @param    id
    *           Widget id of the target editor.
    *
    * @return   <code>true</code> if the line was deleted.
    */
   @Override
   public Boolean deleteEditorLine(int id)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);
      return editor == null ? null : Boolean.valueOf(editor.deleteLine());
   }
   
   /**
    * Selects text in the targeted editor between start and end points.
    * 
    * @param    id
    *           Widget id of the target editor.
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public Boolean setEditorSelection(int id, int start, int end)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);
      return editor == null ? null : Boolean.valueOf(editor.setSelection(start, end));
   }

   /**
    * Sets the cursor line of the targeted editor
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    line
    *           the new cursor line value
    *           
    * @return   <code>true</code> if the cursor line is valid and set, 
    *           null if no widget with given id is found
    */
   @Override
   public Boolean setEditorCursorLine(int id, int line)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);
      return editor == null ? null : Boolean.valueOf(editor.setCursorLine(line));
   }
      
   /**
    * Sets the cursor position in the current line of the targeted editor
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    pos
    *           the new cursor position value
    *           
    * @return   <code>true</code> if the cursor position is valid and set, 
    *           null if no widget with given id is found
    */
   @Override
   public Boolean setEditorCursorChar(int id, int pos)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);
      return editor == null ? null : Boolean.valueOf(editor.setCursorChar(pos));
   }
      
   /**
    * Utility method for debugging, can be used to include information from the server-side
    * to the client-side log
    *
    * @param id
    *        widget id
    * @param message
    *       String parameter
    */
   @Override
   public void debug(int id, String message)
   {
      AbstractWidget<?> widget = (AbstractWidget<?>)getWidget(id);
      if (widget != null)
      {
         widget.debug(message);
      }
   }
   /**
    * Sets the cursor offset of the targeted editor
    *
    * @param    id
    *           Widget id of the target editor.
    *
    * @param    offset
    *           the new offset value
    * @return   <code>true</code> if the cursor offset is valid and set, 
    *           null if no widget with given id is found
    */
   @Override
   public Boolean setEditorCursorOffset(int id, int offset)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);
      return editor == null ? null : Boolean.valueOf(editor.setCursorOffset(offset));
   }
      

   /**
    * Inserts string at the current cursor position of the targeted editor.
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    txt
    *           the string to be inserted
    *
    * @return   <code>true</code> if the string is inserted, 
    *           null if no widget with given id is found
    */
   @Override
   public Boolean insertEditorString(int id, String txt)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);

      return editor == null ? null : editor.insertString(txt);
   }

   /**
    * Replace selection of the targeted editor with a given string.
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    txt
    *           the replacement string
    *
    * @return   <code>true</code> if the selection is replaced, 
    *           null if no widget with given id is found
    */
   @Override
   public Boolean replaceEditorSelectionText(int id, String txt)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);

      return editor == null ? null : editor.replaceSelectionText(txt);
   }

   /**
    * Replace given substring of the targeted editor with a given string
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    from
    *           the string to be replaced
    * @param    to
    *           the replacement string
    * @param    flag
    *           the replacement mode flag as described in 4GL EDITOR:REPLACE documentation 
    * 
    * @return   <code>true</code> if the replacement was performed, 
    *           null if no widget with given id is found
    */
   @Override
   public Boolean replaceEditorString(int id, String from, String to, int flag)
   {
      Editor<?> editor = (Editor<?>) getWidget(id);
      return editor == null ? null : editor.replaceString(from, to, flag);
   }

   /**
    * Enables/disables input for all specified field-level widgets of the
    * frame.
    * If the enabled status changes, the widgets redraw themselves with the
    * appropriate visual cue added or removed. This redrawing is limited to
    * the widget rectangle, however, and should preserve any PUT SCREEN output
    * that may have appeared anywhere else on the frame.
    * <p>
    * This method maintains the tab order of the specified frame. Every widget
    * listed for enabling is moved to the end of the tab order
    * retaining the same order as specified in the list.  If a
    * frame with x, y and z widgets gets an enable call for y and
    * x (in that order), then the tab order will change from the
    * default <code>x y z</code> to <code>z y x</code>. In the case
    * where all widgets are enabled, no change to the tab order
    * occurs.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DISABLE ALL</b> <br>
    * <b>DISABLE ALL EXCEPT</b> <br>
    * <b>ENABLE ALL</b> <br>
    * <b>ENABLE ALL EXCEPT</b>
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    widgetId
    *           array of field level widget IDs to be enabled or
    *           <code>null</code>, which means all of them.
    * @param    sb
    *           The currently editable frames screen buffer array.
    * @param    enabled
    *           New state for widgets.
    * @param    windowId
    *           Id of target window
    *           
    * @return   <code>ScreenBuffer</code> array with the current values.
    */
   @Override
   public ScreenBuffer[] enable(final int            frameId,
                                final int[]          widgetId,
                                final ScreenBuffer[] sb,
                                final boolean        enabled,
                                final int            windowId)
   {
      return enable(frameId, widgetId, sb, enabled, windowId, enabled);
   }
   
   /**
    * Enables/disables input for all specified field-level widgets of the
    * frame.
    * If the enabled status changes, the widgets redraw themselves with the
    * appropriate visual cue added or removed. This redrawing is limited to
    * the widget rectangle, however, and should preserve any PUT SCREEN output
    * that may have appeared anywhere else on the frame.
    * <p>
    * This method maintains the tab order of the specified frame. Every widget
    * listed for enabling is moved to the end of the tab order
    * retaining the same order as specified in the list.  If a
    * frame with x, y and z widgets gets an enable call for y and
    * x (in that order), then the tab order will change from the
    * default <code>x y z</code> to <code>z y x</code>. In the case
    * where all widgets are enabled, no change to the tab order
    * occurs.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DISABLE ALL</b> <br>
    * <b>DISABLE ALL EXCEPT</b> <br>
    * <b>ENABLE ALL</b> <br>
    * <b>ENABLE ALL EXCEPT</b>
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    widgetId
    *           array of field level widget IDs to be enabled or
    *           <code>null</code>, which means all of them.
    * @param    sb
    *           The currently editable frames screen buffer array.
    * @param    enabled
    *           New state for widgets.
    * @param    windowId
    *           Id of target window
    * @param    resolveBuffer
    *           Flag indicating if the screen buffer needs to be resolved and returned.
    * 
    * @return   The frames screen buffer array, only if <code>resolveBuffer</code> flag is set to
    *           <code>true</code>.
    */
   public ScreenBuffer[] enable(final int            frameId,
                                final int[]          widgetId,
                                final ScreenBuffer[] sb,
                                final boolean      enabled,
                                      int          windowId,
                                      boolean      resolveBuffer)
   {
      inEnable = true;
      Widget focus = UiUtils.getCurrentFocus();
      Widget pendingFocus = getEnablePendingFocus(focus, widgetId, enabled);

      // must be called first (should only be needed in prompt-for, set and
      // update situations but not in a normal enable language stmt)!
      honorConditionalUpDown(false);

      int tempTerm = forceInteractive();

      // use getWidget() directly here instead of getFrame() because it is OK to have a null
      // returned
      final Frame frame = (Frame) getWidget(frameId);

      try
      {
         // both ENABLE and DISABLE are NOPs when operating against a frame that has never
         // been otherwise loaded (via VIEW, DISPLAY...), 
         if (frame == null)
            return null;

         // quick exit if this a FRAME:SENSITIVE call
         if (widgetId != null     &&
             widgetId.length == 1 &&
             frameId == widgetId[0])
         {
            frame.setEnabled(enabled);
            return null;
         }

         UIStatement stmt = (sb == null || sb[0] == null ? null : sb[0].getCurrentStatement());
         
         frame.setCurrentStatement(stmt);

         // check if frame will be reset or paused during view()
         boolean preview = (frame.isNeedReset() && !frame.isDown());

         if (enabled && preview)
         {
            // this is a view call
            viewWorker(frame, sb, widgetId, true, windowId);
         }

         // realize the frame now, but don't make it visible.  this is needed because when 
         // enabling widgets in the frame, their tab order gets adjusted.  doing this later, after
         // the tab adjustment, the widget's order no longer matches their setup order.
         windowId = frame.placeToWindow(windowId, stmt);
         frame.realizeFrame(true);
         
         startWidgetOutput();
         tk.setInvalidate(true);

         try
         {
            enableWorker(widgetId, sb, enabled, focus, frame, preview, windowId);

            if (FocusManager.tryFocusChange(pendingFocus))
               enableFillInAutoZap(pendingFocus);
         }

         finally
         {
            tk.setInvalidate(frame, false, true);
            endWidgetOutput();
         }
      }

      finally
      {
         restoreRedirection(tempTerm);
         inEnable = false;

         frame.setCurrentStatement(null);
      }

      return resolveBuffer ? getEditableClientScreenBuffers(getScreenBuffer(frameId, -1, true))
                           : null;
   }

   /**
    * Get widget which may receive focus after ENABLE.
    *
    * @param   focus
    *          Frame ID.
    * @param   widgetId
    *          Array of enabled widget IDs.
    * @param   enabled
    *          <code>true</code> if call enables widgets.
    *
    * @return  reference to widget which may receive focus or <code>null</code>
    *          if no focus management should be performed.
    */
   private Widget getEnablePendingFocus(Widget focus, int[] widgetId,
                                        boolean enabled)
   {
      if (!enabled || triggerNestingByWaitFor.peek() != 1 || isEmpty(widgetId))
      {
         // ignore DISABLE calls, calls outside triggers and
         // calls when there is no active focus yet
         return null;
      }

      // do not switch focus if the current focus is valid
      if (FocusManager.focusState(focus) == FocusManager.FOCUS_VALID)
      {
         return null;
      }
      
      Widget pending = getWidget(widgetId[0]);

      if (UiUtils.locateFrame(focus) != UiUtils.locateFrame(pending)
          ||  pending == focus
          || !(focus instanceof Button))
      {
         return null;
      }

      return pending;
   }

   /**
    * Check if array of widget IDs is empty.
    *
    * @param   widgetId
    *          Array of widget IDs.
    *
    * @return  <code>true</code> if array is <code>null</code> or contains no
    *          elements.
    */
   private boolean isEmpty(int[] widgetId)
   {
      return widgetId == null || widgetId.length < 1;
   }

   /**
    * Widget enabling worker.
    *
    * @param   widgetId
    *          List of widget IDs.
    * @param   sb
    *          Screen buffer array.
    * @param   enabled
    *          New enabled state.
    * @param   focus
    *          Focused widget.
    * @param   frame
    *          Focused frame.
    * @param   preview
    *          Preview flag.
    * @param   windowId
    *          target window id.
    */
   private void enableWorker(final int[] widgetId, final ScreenBuffer[] sb,
                          final boolean enabled, Widget focus,
                          final Frame frame, boolean preview, int windowId)
   {
      // list of affected widgets
      Widget[] affw = null;

      push();

      try
      {
         affw = processEnable(frame, enabled, widgetId, sb, focus);
      }

      finally
      {
         if (frame.isVisible())
         {
            popTrash();
         }
         else
         {
            pop();
         }
      }

      if (frame.isVisible())
      {
         redrawEnabled(frame, affw);
      }

      // ENABLE calls for widgets within a hidden frame will not make
      // any widgets visible
      if (enabled && !preview && 
          (!frame.hidden() || 
           (frame.isRootFrame() && 
            sb[0].getCurrentStatement() != UIStatement.ENABLE)))
      {
         // if frame is not visible, then force to view all widgets
         int[] wids = !frame.isVisible() ? null : widgetId;
         // this is a view call
         viewWorker(frame, sb, wids, true, windowId);

         if (frame.hidden())
         {
            frame._setVisible(true);
            frame._setHidden(false);
         }
      }
   }

   /**
    * Redraw enabled widgets.
    *
    * @param   frame
    *          Target frame.
    * @param   affw
    *          List of widgets to redraw.
    */
   private void redrawEnabled(final Frame frame, Widget[] affw)
   {
      final Widget[] affWidgets = affw;

      // draw widgets directly
      independentEventDrawingBracket(frame, new Runnable()
      {
         @Override
         public void run()
         {
            repaintAffectedWidgets(UiUtils.getFramesAbove(frame), Arrays.asList(affWidgets));
         }
      });
   }

   /**
    * Process enable.
    *
    * @param   frame
    *          Target frame.
    * @param   enabled
    *          New enabled state.
    * @param   widgetId
    *          List of widget IDs.
    * @param   sb
    *          Screen buffer array.
    * @param   focus
    *          Focused widgets.
    *
    * @return  List of widgets.
    */
   private Widget[] processEnable(Frame frame, boolean enabled,
                                  int[] widgetId, ScreenBuffer[] sb,
                                  Widget focus)
   {
      int naffw;
      FrameConfig cfg = frame.config();
      Widget[] affw = null;

      boolean seen = frame.isVisible() && enabled;

      if (!enabled && sb != null)
      {
         // honor any changes in our data
         setScreenBuffers(sb);

         // adjust DOWN value
         if (sb[0] != null)
         {
            frame.adjustDown(sb[0].getDown());
            frame.setDynRow(sb[0].getRow());
            frame.setDynCol(sb[0].getColumn());
         }
      }

      // core enable/disable processing
      if (widgetId == null)
      {
         // no widget list specified
         affw = frame.enableAllWidgets(focus, enabled);
         naffw = affw.length;

         // maintain tab order
         if (enabled && !cfg.keepTabOrder)
         {
            frame.restoreInitialTabOrder();
         }
      }
      else
      {
         // a widget list specified
         naffw = widgetId.length;
         if (naffw > 0)
         {
            affw = new Widget[naffw];
         }
         // enable/disable all widgets in the given list
         frame.enabledWidgetsByList(affw, focus, widgetId, enabled);
      }

      if (enabled)
         frame.needClearRow();

      if (seen)
         frame.setPaintable(false);

      // FRAME:SENSITIVE flag can only be set to ON by an ENABLE call
      // or by explicit FRAME:SENSITIVE = true
      if (enabled)
      {
         frame.setEnabled(enabled);
      
         if (frame.isSideLabels())
         {
            // add the widget labels too, as they are affected (they must be brought back to 
            // visible state, if they were hidden previously)

            List<Widget<?>> lw = new ArrayList<>();
            for (Widget<?> w : affw)
            {
               lw.add(w);
               
               if (w instanceof LabeledWidget<?>)
               {
                  LabeledWidget<?> lblw = (LabeledWidget<?>) w;
                  
                  Label<?> lbl = lblw.getLabelInstance();
                  
                  if (lbl != null)
                  {
                     lw.add(lbl);
                  }
               }
            }
            
            affw = lw.toArray(new Widget[0]);
            naffw = affw.length;
         }
      }

      if (seen)
         frame.setPaintable(true);
      return affw;
   }
   
   /**
    * Move specified widget after the widget "after".
    * 
    * @param   frameId
    *          ID of the frame that contains widgets being moved.
    * @param   widget
    *          ID of the widget to move.
    * @param   after
    *          ID of the widget after which the widget should be moved.
    *          
    * @return  TRUE - the widget was moved successfully, FALSE - otherwise.
    */
   public logical moveAfterTabItem(int frameId, int widget, int after)
   {
      Frame frame = getFrame(frameId);
      if (frame == null)
      {
         throw new IllegalArgumentException("Invalid frame ID " + frameId);
      }
      
      Widget currentWidget = null;
      Widget afterWidget = null;
      for (Widget w : frame.getContentPane().widgets())
      {
         int wId = UiUtils.lookupWidgetIdAsInt(w);
         
         if (wId == widget)
         {
            currentWidget = w;
         }
         if (wId == after)
         {
            afterWidget = w;
         }
      }
      
      if (afterWidget == null)
      {
         throw new IllegalArgumentException("Invalid afterWidget Id=" + after);
      }
      
      if (currentWidget == null)
      {
         throw new IllegalArgumentException("Invalid currentWidget Id=" + widget);
      }
      
      frame.getContentPane().moveAfterTabItem(currentWidget, afterWidget);
      
      return new logical(true);
   }
   
   /**
    * Move specified widget before the widget "before" in the tab order.
    * Widgets must be in the same field-group, that is checked on the server-side.
    * 
    * @param   frameId
    *          ID of the frame that contains widgets being moved.
    * @param   widget
    *          ID of the widget to move.
    * @param   before            
    *          ID of the widget before which the widget should be moved.
    *          
    * @return  TRUE - the widget was moved successfully, FALSE - otherwise.         
    */
   @Override
   public logical moveBeforeTabItem(int frameId, int widget, int before)
   {
      Frame frame = getFrame(frameId);
      if (frame == null)
      {
         throw new IllegalArgumentException("Invalid frame ID " + frameId);
      }
      
      Widget currentWidget = null, beforeWidget = null;
      for (Widget w : frame.getContentPane().widgets())
      {
         int wId = UiUtils.lookupWidgetIdAsInt(w);
         
         if (wId == widget)
         {
            currentWidget = w;
         }
         if (wId == before)
         {
            beforeWidget = w;
         }
      }
      
      if (beforeWidget == null)
      {
         throw new IllegalArgumentException("Invalid beforeWidget Id=" + before);
      }
      
      if (currentWidget == null)
      {
         throw new IllegalArgumentException("Invalid currentWidget Id=" + widget);
      }
      
      frame.getContentPane().moveBeforeTabItem(currentWidget, beforeWidget);
      
      return new logical(true);
   }


   /**
    * Set {@link #pendingFocusApply} reference.
    *
    * @param   widget
    *          Widget reference.
    */
   public void setPendingFocusApply(Widget widget)
   {
      pendingFocusApply = widget;
   }

   /**
    * Enables/disables input for the specified field-level widget. This method
    * preserves the current tab order of the frame.
    * <p>
    * This method corresponds to these Progress statement: <br>
    * <b>widget:SENSITIVE=true/false</b>
    *
    * @param   frameId
    *          ID of frame this request came for
    * @param   widgetId
    *          widget ID to be enabled or disabled
    * @param   sb
    *          The currently editable frames screen buffer array.
    * @param   enabled
    *          New state of the widget
    * @param   windowId
    *          Id of target window
    *           
    * @return   The screen buffer array for editable frames.
    */
   @Override
   public ScreenBuffer[] enable(int            frameId,
                                int            widgetId,
                                ScreenBuffer[] sb,
                                boolean        enabled,
                                int            windowId)
   {
      Frame frame = getFrame(frameId);
      FrameConfig cfg = frame.config();

      boolean save_kto = cfg.keepTabOrder;
      cfg.keepTabOrder = true;

      try
      {
         return enable(frameId, new int[] {widgetId}, sb, enabled, windowId);
      }
      finally
      {
         if (!save_kto)
         {
            cfg.keepTabOrder = false;
         }
      }
   }

   /**
    * Underlines a set of widgets of the specified frame. If the target
    * frame is not a down frame, the frame is simply cleared.
    *
    * @param    frameId
    *           The frame ID this request came for.
    * @param    widgetIds
    *           The array of field level widget IDs to be underlined.
    */
   @Override
   public void underline(final int frameId, final int[] widgetIds)
   {
      // must be called first!
      honorConditionalUpDown();

      if (widgetIds == null)
         return;

      final Frame frame = getFrame(frameId);

      eventDrawingBracket(frame, new Runnable()
      {
         @Override
         public void run()
         {

            if (!frame.isDown())
            {
               // clear the regular frame
               frame.clearWorker(true);
               frame.repaint();
            }
            else
           {
               Widget[] list = new Widget[widgetIds.length];
               for (int i = 0; i < widgetIds.length; i ++)
               {
                  list[i] = getWidget(widgetIds[i]);
               }
               frame.underline(list);
               
               // if frame was not visible at the moment of underline call make special mark
               // but limit this mark only for display output
               if (!frame.isVisible() && !isRedirected())
               {
                  frame.setPendingUnderline(true);
               }
            }
         }
      });
   }

   /**
    * Positions the cursor on a new line in a down frame.  When
    * <code>lines</code> is negative, this operates as a move of the cursor
    * <code>up</code>.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DOWN n </b>
    * <b>UP n </b>
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    lines
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    sid
    *           The stream for which pending PUT or EXPORT output must be
    *           ended and written to the stream. -1 means that this is not
    *           necessary.
    *           
    * @return   If this is a down frame, the configs for the widgets in the new row.  Otherwise
    *           <code>null</code> is returned.
    */
   @Override
   public WidgetConfig[] down(int frameId, int lines, boolean cond, int sid)
   {
      // do this regardless of anything else (even if we will be returning
      // immediately)
      endOutput(sid);

      if (lines == 0)
         return null;

      Frame frame = (Frame) getWidget(frameId);

      if (frame == null)
         return null;

      // check the frame for possible stream switching
      boolean newStream = switchStreams(frame);

      downWorker(frameId, lines, cond, newStream, false);
      
      return frame.getDownConfigs();
   }
   
   /**
    * Positions the cursor on a new line in a down frame.  When {@code lines} is negative, this
    * operates as a move of the cursor {@code up}.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DOWN n </b>
    * <b>UP n </b>
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    steps
    *           Number and direction of iterations of the data to move. If positive, the move is
    *           {@code down}, if negative the move is {@code up}.  0 is ignored.
    * @param    conditional
    *           {@code true} when this call is for conditional up/down.
    * @param    newStream
    *           {@code true} when stream switch has just occurred
    * @param    noPause
    *           if {@code true} then no pause will be generated on cursor wrap.
    *
    * @return   {@code true} if wrap is occurred.
    */
   private boolean downWorker(int frameId,
                              int steps,
                              boolean conditional,
                              boolean newStream,
                              final boolean noPause)
   {
      if (steps == 0)
      {
         return false;
      }
      
      final Frame frame = (Frame) getWidget(frameId);
      if (frame == null)
      {
         return false;
      }
      
      boolean down = frame.isDown();
      boolean redir = isRedirected();
      boolean force = redir && conditional && frame.isHonorNextConditionalDown();
      
      frame.setHonorNextConditionalDown(false);
      
      // process redirected down, conditional or not
      if (redir)
      {
         RedirectedTerminal rt = redirections.get(currentStream);
         Stream stream = sd.streamFromId(currentStream);
         boolean reducedDown = isReducedDown(newStream, frame, down, stream);
         
         int prevLineNum = -1;
         boolean flushed = rt.flush();
         
         if (flushed)
         {
            // if there was pending output in the redirected terminal, we flush now
            prevLineNum = flushNow(steps, stream);
         }
         
         clearFrame(frameId, conditional, frame, down);
         
         // end of conditional down processing
         if (conditional && !force)
         {
            return false;
         }
         
         unconditionalRedirectedDown(frameId, steps, frame, rt, stream,
                                     reducedDown, prevLineNum, flushed);
         
         // end of redirected down processing
         return false;
      }
      
      // process interactive down, conditional or not
      if (frame.isStreamed())
      {
         // clearing of a streamed frame in interactive mode is required
         // this action should cause pausing on the terminal if redirection
         // is still active
         if (redirections.get(frame.getStreamId()) != null)
         {
            pause(-1, null, false, frame.topLevelWindow());
         }
         
         frame.clearCurrent();
         // no cursor advance needed in this case; we're done
         return false;
      }
      
      // advance the frame cursor
      if (down)
      {
         return advanceCursor(steps, noPause, frame);
      }
      
      frame.clearCurrent();
      
      return false;
   }

   /**
    * Perform immediate stream flush.
    *
    * @param   steps
    *          DOWN size.
    * @param   stream
    *          Output stream.
    *
    * @return  The next line number AFTER all buffered lines are output.
    */
   private int flushNow(int steps, Stream stream)
   {
      // DOWN X handles page breaks with special logic below so the
      // stream must not do any of the default page break processing
      if (steps > 1)
      {
         stream.setContiguous(true);
      }

      return flushRedirected(stream, true);
   }

   /**
    * Advance DOWN cursor.
    *
    * @param   steps
    *          UP/DOWN size.
    * @param   noPause
    *          Disable pause processing/
    * @param   frame
    *          Target frame.
    *
    * @return  <code>true</code> if cursor wrap has occurred
    */
   private boolean advanceCursor(int steps, final boolean noPause, final Frame frame)
   {
      final int num  = steps;
      final boolean[] wrap = new boolean[1];

      // interactive down is meaningful for the down frames only
      try
      {
         eventDrawingBracket(frame, new Runnable()
         {
            @Override
            public void run()
            {
               wrap[0] = false;

               try
               {
                  // This call can trigger pause if we are moving cursor
                  // over the edge of the frame. During the pause a
                  // condition can be raised which will cause this block
                  // to exit abnormally.
                  wrap[0] = frame.moveCursor(num, noPause);
               }

               finally
               {
                  if (wrap[0])
                  {
                     viewRefresh = true;
                     viewDisplay = false;
                  }
               }
            }
         });
      }

      finally
      {
         if (wrap[0])
            nextViewRefresh = true;
      }

      return wrap[0];
   }

   /**
    * Clear frame.
    *
    * @param   frameId
    *          Frame ID.
    * @param   conditional
    *          <code>true</code> if DOWN processing is conditional/
    * @param   frame
    *          Target frame.
    * @param   down
    *          Direction flag.
    */
   private void clearFrame(int frameId, boolean conditional,
                          final Frame frame, boolean down)
   {
      // clear the frame buffer early enough so that any views of the
      // frame (which may occur in the down processing below) won't
      // redraw the current iteration but rather one that is uninitialized
      if (down)
      {
         clearAll(frameId, true, false);
      }
      else
      {
         if (conditional)
         {
            eventBracket(new Runnable()
            {
               @Override
               public void run()
               {
                  frame.clearWorker(true);
               }
            });
         }
      }
   }

   /**
    * Determine if DOWN is a reduced one.
    *
    * @param   newStream
    *          Stream.
    * @param   frame
    *          Target frame.
    * @param   down
    *          DOWN direction.
    * @param   stream
    *          Old stream.
    *
    * @return  <code>true</code> if DOWN is a reduced one.
    */
   private boolean isReducedDown(boolean newStream, final Frame frame,
                                 boolean down, Stream stream)
   {
      boolean reducedDown = false;
      
      if (stream == null)
      {
         throw new IllegalStateException("Invalid stream id " + currentStream + ".");
      }
      
      // for redirected down frames, switching streams has flushed the stream by now, 
      // which is equivalent to DOWN 1 detect and save this condition
      if (down      &&                 // it's a down frame
          newStream &&                 // switching streams has flushed
          !frame.isCurrentTouched())   // no output since last down
      {
         reducedDown = true;
      }
      return reducedDown;
   }

   /**
    * Handle unconditional DOWN processing in redirected mode.
    *
    * @param   frameId
    *          ID of frame this request came for.
    * @param   steps
    *          Number and direction of iterations of the data to move. If positive, the move is
    *          {@code down}, if negative the move is {@code up}. 0 is ignored.
    * @param   frame
    *          The processed frame.
    * @param   rt
    *          The {@link RedirectedTerminal}.
    * @param   stream
    *          The redirected stream.
    * @param   reducedDown
    *          {@code true} if this down is reduced.
    * @param   prevLineNum
    *          The next line number AFTER all buffered lines were flushed or -1 if the
    *          {@code RedirectedTerminal} was not flushed.
    * @param   flushed
    *          {@code true} if the {@code RedirectedTerminal} was flushed.
    */
   private void unconditionalRedirectedDown(int frameId,
                                            int steps, 
                                            final Frame frame,
                                            RedirectedTerminal rt,
                                            Stream stream,
                                            boolean reducedDown,
                                            int prevLineNum,
                                            boolean flushed)
   {
      boolean deferView = false;
      boolean down = frame.isDown();
      int lines = (int) (steps * frame.getIterationHeight());

      // if an unconditional down is the first output for a given frame on
      // a page, it needs to draw the frame (headers etc...), this must
      // occur BEFORE down output for DOWN frames and AFTER the output
      // for non-DOWN frames
      if (!frame.isWithHeaders(stream))
      {
         if (down)
         {
            forceFrameDraw(rt, frameId, currentStream);
         }
         else
         {
            deferView = !flushed;
         }
      }
      
      boolean forceNextPage = false;
      boolean forceNewLine  = false;
      
      int rawSize     = stream.rawPageSize();
      int pageSize    = stream.getEffectivePageSize();
      int nextLineNum = stream.rawNextLineNum();
      
      // output was flushed
      boolean prevFlush = prevLineNum != -1 && prevLineNum != nextLineNum;
      
      // check here for possible page spill with one full iteration
      if (steps == 1 && lines > 1 && /* fh.isTrailingSkips() && */
          (nextLineNum + frame.getIterationHeight() > pageSize))
      {
         double delta = frame.getLeadingSkipSize() + frame.getTrailingSkipSize();
         
         delta += frame.getIterationHeight() - delta - 1;
         
         if (frame.getLeadingSkipSize() == -1 ||
             delta > (pageSize - nextLineNum))
         {
            // the page would spill, so: force page eject, bypass regular
            // down processing and force full iteration of blank lines
            steps = 0;
            forceNextPage = true;
         }
      }

      if (!flushed && !reducedDown)
      {
         stream.writeControlled("");
      }
      else if (steps == 1)
      {
         frame.setHonorNextConditionalDown(true);
      }

      // DOWN 1
      if (steps == 1)
      {
         // flush the pending new line that we just added
         if (!flushed && !reducedDown)
         {
            // the rest of the full iteration
            for(int i = 1; i < lines; i++)
            {
               stream.writeControlled("");
            }

            nextLineNum = flushRedirected(stream, false);
         }

         // only paged streams get additional processing below
         if (rawSize > 0)
         {
            // the previous flushing at the top of this worker may have
            // caused a page break, we detect that case here
            if (prevFlush)
            {
               // if the previous line number was greater than the end
               // of the writable portion of the page
               if (prevLineNum > pageSize)
               {
                  // output a single new line on the next page
                  forceNewLine = true;
               }
            }
            else
            {
               // no page break was caused by the previous flush, but
               // would one be caused by the our next flush?
               if (nextLineNum > pageSize)
               {
                  // force a page break and then output a new line on the next page
                  forceNextPage = true;
               }
            }
         }
      }

      // DOWN X
      if (steps > 1)
      {
         // by default output our lines
         int numNL = lines - 1;

         // TODO: the DOWN reduction below is not well tested
         if (reducedDown)
         {
            numNL -= frame.getIterationHeight();
         }

         // paged streams get special/bizarre behavior with overflows
         if (rawSize > 0)
         {
            // if we are starting PAST the end of the writable page
            if (nextLineNum > pageSize)
            {
               // only output 1 line on this page
               // TO-DO: line or iteration?
               numNL = 1;

               // then force the page break and output one more line
               forceNextPage = true;
            }
            else
            {
               // we are starting BEFORE or AT the writable limit of the
               // page, calculate where we are going to end our output
               int estEnd = nextLineNum + lines - 1;

               // check if we are crossing the page boundary
               if (estEnd > pageSize)
               {
                  // calculate how many new lines can fit "legally"
                  int fin = pageSize - nextLineNum;

                  // calculate how many extra newlines we have been told
                  // to output (these are "illegal" ones)
                  int extra = estEnd - pageSize;

                  // only a maximum of 2 of the illegal newlines will be
                  // output
                  extra = extra <= 2 ? extra : 2;

                  // calculate the actual number of legal + illegal
                  // newlines to be output
                  numNL = fin + extra;

                  // force the page break and one more line on the next
                  // page
                  forceNextPage = true;
               }
            }

            // explicitly forcing the page break means that the output
            // of our NL chars must not trigger the page break processing
            if (forceNextPage)
            {
               stream.setContiguous(true);
            }
         }

         // output our new lines
         for(int i = 0; i < numNL; i++)
         {
            stream.writeControlled("");
         }

         flushRedirected(stream, false);
      }

      // manually force the page break if deferred above
      if (forceNextPage)
      {
         stream.forcePageBreak(false);
      }

      // output the frame and then a NL char if deferred above
      if (forceNextPage || forceNewLine || deferView)
      {
         // don't duplicate the footer/header processing since on a small
         // page (that is smaller than the size of headers + footers) this
         // processing can cause inappropriate recursion
         stream.setContiguous(true);

         // draw the frame
         forceFrameDraw(rt, frameId, currentStream);

         // push the output to the stream
         boolean bypass = rt.flush();

         // add single empty iteration and flush
         for(int i = 0; i < frame.getIterationHeight(); i++)
         {
            stream.writeControlled("");
         }

         flushRedirected(stream, bypass);
      }
   }

   /**
    * Positions the cursor on a new line in a down frame.  When
    * <code>lines</code> is negative, this operates as a move of the cursor
    * <code>up</code>.
    * <p>
    * The terminal will be temporarily redirected to the given stream for the
    * duration of this method call.  This allows the caller to avoid
    * bracketing this call with {@link #switchOutput} invocations, as the
    * same result is achieved internal to this method's operation.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DOWN n </b>
    * <b>UP n </b>
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    lines
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call).
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    endOut
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    *           
    * @return   If this is a down frame, the configs for the widgets in the new row.  Otherwise
    *           <code>null</code> is returned.
    */
   @Override
   public WidgetConfig[] down(int frameId, int lines, int streamId, boolean cond, boolean endOut)
   {
      // do this regardless of whether we are about to return immediately
      // or not
      if (endOut)
      {
         endOutput(streamId);
      }

      if (lines == 0)
         return null;

      boolean switched = currentStream == streamId;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(streamId, true);
      }

      // check the frame for possible stream switching
      Frame frame = (Frame) getWidget(frameId);

      if (frame == null)
         return null;

      boolean newStream = switchStreams(frame);

      try
      {
         downWorker(frameId, lines, cond, newStream, false);
      }

      finally
      {
         if (!switched)
         {
            switchOutput(tempOut, false);
         }
      }

      return frame.getDownConfigs();
   }

   /**
    * Get the focus manager.
    *
    * @return   current focus manager.
    */
   public FocusManager getFocusManager()
   {
      return focusManager;
   }

   /**
    * Get ID of the widget which has the input focus.
    *
    * @return   ID of the focused widget.
    */
   @Override
   public int getFocus()
   {
      Widget focus = UiUtils.getCurrentFocus();
      Widget frame = UiUtils.locateFrame(focus);

      if (triggerNesting == 0 && frame instanceof Frame && !frame.isVisible())
         return -1;

      return UiUtils.getWidgetIdAsInt(focus);
   }

   /**
    * Returns pause before hide time for automatic frame hiding.
    * 
    * @return   See above.
    */
   public int getTimePBH()
   {
      return timePBH;
   }

   /**
    * Returns true if pause before hide mode is active.
    *
    * @return   See above.
    */
   public boolean isHonorPBH()
   {
      return honorPBH;
   }

   /**
    * Returns text displayed while processing pbh.
    * 
    * @return   See above.
    */
   public String getTextPBH()
   {
      return textPBH;
   }

   /**
    * Post the server event to the client side. The event will be executed only on the next 
    * PROCESS-EVENTS or UI-blocking statement.
    * 
    * @param    event
    *           The server event.
    */
   public void postServerEvent(ServerEvent event)
   {
      EventManager.postServerEvent(event);
      
      // notify the type-ahead that a server-side event has arrived, and can stop waiting for keys
      typeAhead.notifyServerEvent();
   }
   
   /**
    * This is a special API which is used to delegate a call from converted 4GL code 
    * (via _P2J_REMOTE_CALL_ functions) to the remote side.  Usually, this is used when P2J client
    * runs in embedded mode, i.e. the P2J screens are embedded in another customer-specific 
    * application.  The call will be remoted into the customer-specific application.
    * <p>
    * This API is expected to block waiting for a response.
    * 
    * @param    request
    *           The request data incoming from 4GL.
    * 
    * @return   The response from the customer-specific application, which will be sent back to 
    *           the converted 4GL code.
    */
   @Override
   public UnformattedPayload remoteCall(UnformattedPayload request)
   {
      if (!isEmbedded())
      {
         String msg = "Attempted to execute a remote call request in a non-embedded client.";
         throw new SilentUnwindException(new IllegalStateException(msg));
      }

      EmbeddedClient ec = (EmbeddedClient) OutputManager.getDriver();
      return ec.remoteCall(request);
   }

   /**
    * Send a PUBLISH notification to a remote application. 
    * 
    * @param    subId
    *           The subscriber's ID, known by the remote application.
    * @param    event
    *           The published event name.
    * @param    param
    *           The parameters for this event.
    * @param    delayed
    *           Flag indicating if the PUBLISH must be executed immediately (when <code>true</code>)
    *           or posted to the client's event queue, otherwise.
    */
   public void publish(long subId, String event, String[] param, boolean delayed)
   {
      if (!isEmbedded())
      {
         String msg = "Attempted to execute a remote call request in a non-embedded client.";
         throw new SilentUnwindException(new IllegalStateException(msg));
      }

      EmbeddedClient ec = (EmbeddedClient) OutputManager.getDriver();

      if (delayed)
      {
         // TODO: invokeLater can't be used because why?
         // TODO: if OS-queue is used, PROCESS-EVENTS will not work (and can't add OS-queue
         // to that because it will include other mouse/windiow events).
         invokeLaterOS(() -> { ec.publish(subId, event, param); });
      }
      else
      {
         ec.publish(subId, event, param);
      }
   }
   
   /**
    * Post the specified mouse event to the {@link EventManager}, which will save it in a 
    * dedicated queue.
    * <p>
    * This usage of an AWT event class is intentional.  It is being used separately from the
    * AWT/Swing event model.  No other AWT/Swing usage is occurring in this class.
    * 
    * @param    event
    *           The mouse event.
    * @param    windowId
    *           The window where the event was captured.
    */
   public void postMouseEvent(java.awt.event.MouseEvent event, int windowId)
   {
      if (EventManager.postMouseEvent(event, windowId))
      {
         // notify the type-ahead that a mouse event has arrived
         typeAhead.notifyOSEvent();
      }
   }

   /**
    * Post the specified mouse event to the {@link EventManager}, which will save it in a 
    * dedicated queue.
    * <p>
    * This usage of an AWT event class is intentional.  It is being used separately from the
    * AWT/Swing event model.  No other AWT/Swing usage is occurring in this class.
    * 
    * @param    event
    *           The mouse event.
    * @param    windowId
    *           The window where the event was captured.
    * @param    sourceId
    *           The explicit source of this mouse event.
    */
   public void postMouseEvent(java.awt.event.MouseEvent event, int windowId, int sourceId)
   {
      if (EventManager.postMouseEvent(event, windowId, sourceId))
      {
         // notify the type-ahead that a mouse event has arrived
         typeAhead.notifyOSEvent();
      }
   }
   
   /**
    * Post an OS event (which needs to be processed by the 4GL drawing layer, but not 
    * necessarily as a legacy 4GL event) to the event queue.
    * 
    * @param    event
    *           The event to be posted.
    */
   public void postOSEvent(Event event)
   {
      if (EventManager.postOSEvent(event))
      {
         typeAhead.notifyOSEvent();
      }
   }
   
   /**
    * Remove from the server event queue all the events in the given list, identified by their id.
    * 
    * @param    eventIds
    *           The list of event IDs to be removed.
    */
   public void removeServerEvents(int[] eventIds)
   {
      EventManager.removeServerEvents(eventIds);
   }
   
   /**
    * Invoke the specified server-side code, while maintaining any changes in the trigger
    * registry.
    * 
    * @param    code
    *           The code to be executed.
    * 
    * @return   A custom payload returned by the server-side.
    */
   public Object invokeOnServer(Supplier<ServerReturnValue> code)
   {
      EventList save = currentEventList;
      ServerReturnValue srv = null;
      
      try
      {
         srv = code.get();
      }
      finally
      {
         if (srv != null)
         {
            if (save != null)
            {
               // remove deleted triggers
               if (srv.deadTriggers != null)
               {
                  save.remove(srv.deadTriggers);
               }
               
               // merge new triggers
               if (srv.newTriggers != null)
               {
                  save.merge(srv.newTriggers, true);
               }
            }
         }

         currentEventList = save;
      }
      
      return srv.payload;
   }
   
   /**
    * Asynchronously executes the supplied runnable reference on the event thread.
    * A pseudo event is created and enqueued in the OS event queue and processed/executed
    * in-order. The method returns right after the runnable is enqueued in the event queue.
    * <p>
    * Use this method for processing the OS input.
    *
    * @param    runnable
    *           A valid {@link Runnable} reference.
    *
    * @throws   IllegalArgumentException
    *           when runnable is <code>null</code>.
    */
   public void invokeLaterOS(Runnable runnable)
   {
      // post to OS event queue, we need guaranteed invocation
      postOSEvent(new InvocationEvent(runnable));
   }

   /**
    * Asynchronously executes the supplied runnable reference on the event thread.
    * A pseudo event is created and enqueued in the application event queue and processed/executed
    * in-order. The method returns right after the runnable is enqueued in the event queue.
    * <p>
    * Use this method for the widget-related application logic.
    *
    * @param    runnable
    *           A valid {@link Runnable} reference.
    *
    * @throws   IllegalArgumentException
    *           when runnable is <code>null</code>.
    */
   public void invokeLater(Runnable runnable)
   {
      EventManager.postEvent(new InvocationEvent(runnable));
   }

   /**
    * Register the frame as a header or footer from the stream that is the
    * current redirected terminal output destination.  The terminal must be
    * non-interactive for this method to process properly.  Subsequent
    * pagination operations on the stream will cause the header or footer to
    * render as needed, without any further explicit request from the server.
    *
    * @param   frameId
    *          The id of the frame for this request.
    * @param   header
    *          <code>true</code> to register as a header, <code>false</code>
    *          to register as a footer.
    */
   @Override
   public void registerPageElement(int frameId, boolean header)
   {
      registerPageElement(frameId, header, currentStream, false);
   }

   /**
    * Register the frame as a header or footer with the given stream as the
    * redirected terminal output destination.  Subsequent pagination
    * operations on the stream will cause the header or footer to render
    * as needed, without any further explicit request from the server.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    */
   @Override
   public void registerPageElement(int frameId, boolean header, int streamId)
   {
      registerPageElement(frameId, header, streamId, true);
   }

   /**
    * Register the frame as a header or footer with the given stream as the
    * redirected terminal output destination.  Subsequent pagination
    * operations on the stream will cause the header or footer to render
    * as needed, without any further explicit request from the server.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    * @param    endOut
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    */
   public void registerPageElement(int     frameId,
                                   boolean header,
                                   int     streamId,
                                   boolean endOut)
   {
      if (endOut)
         endOutput(streamId);

      PageElement pe  = null;
      Integer     key = Integer.valueOf(frameId);

      if (streamId == -1)
      {
         throw new IllegalStateException("The given stream is the " +
                                         "interactive terminal!");
      }

      if (pageElements.containsKey(key))
      {
         pe = pageElements.get(key);
      }
      else
      {
         pe = new PageElement(this, frameId);
         pageElements.put(key, pe);
      }

      Stream s = sd.streamFromId(streamId);

      if (s == null)
      {
         throw new IllegalStateException("Invalid stream ID.");
      }

      // save off the mapping to ensure that a forced deregistration
      // can occur later
      if (streamMap.containsKey(pe))
      {
         streamMap.get(pe).add(s);
      }
      else
      {
         Set<Stream> streams = new HashSet<Stream>();
         streams.add(s);
         streamMap.put(pe, streams);
      }

      if (header)
      {
         s.registerHeader(pe);
      }
      else
      {
         s.registerFooter(pe);
      }
   }

   /**
    * Deregister (remove) the frame as a header or footer from the stream
    * that is the current redirected terminal output destination.  The
    * terminal must be non-interactive for this method to process properly.
    * <p>
    * This is called (among other places) during hide processing on the
    * server, which must be safe to be called at any time.  This is true,
    * even if we aren't redirected, if the frame is not visible and/or if
    * there was no page element ever registered for this frame (because it
    * was never viewed).
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    */
   @Override
   public void deregisterPageElement(int frameId, boolean header)
   {
      deregisterPageElement(frameId, header, currentStream);
   }

   /**
    * Deregister (remove) the frame as a header or footer from the given
    * stream.
    * <p>
    * This is called (among other places) during hide processing on the
    * server, which must be safe to be called at any time.  This is true,
    * even if we aren't redirected, if the frame is not visible and/or if
    * there was no page element ever registered for this frame (because it
    * was never viewed).
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    */
   @Override
   public void deregisterPageElement(int     frameId,
                                     boolean header,
                                     int     streamId)
   {
      // Clear references to frame from terminal and from frame to stream.
      Frame frame = (Frame) getWidget(frameId);

      if (frame != null)
      {
         RedirectedTerminal rt = redirections.get(frame.getStreamId());

         if (rt != null)
            rt.resetLastId(frameId);

         frame.setRedirected(false, -1);
      }

      // deregister page elements
      Integer     key = Integer.valueOf(frameId);
      PageElement pe  = pageElements.get(key);

      if (pe == null)
      {
         return;
      }

      if (streamId == -1)
      {
         return;
      }

      pageElements.remove(key);

      Stream s = sd.streamFromId(streamId);

      if (s == null)
      {
         throw new IllegalStateException("Invalid stream ID.");
      }

      // remove this stream from our list of mapped streams
      Set<Stream> streams = streamMap.get(pe);

      if (streams != null)
      {
         streams.remove(s);

         // remove the entry completely when no longer needed
         if (streams.isEmpty())
         {
            streamMap.remove(pe);
         }
      }

      if (header)
      {
         s.deregisterHeader(pe);
      }
      else
      {
         s.deregisterFooter(pe);
      }
   }

   /**
    * Deregister (remove) the frame as a header or footer from all streams
    * in which it is registered.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    */
   @Override
   public void forceDeregisterPageElement(int frameId, boolean header)
   {
      Integer     key = Integer.valueOf(frameId);
      PageElement pe  = pageElements.get(key);

      if (pe == null)
      {
         // we must never have been registered or we were cleanly
         // deregistered so we are done
         return;
      }

      pageElements.remove(key);

      Set<Stream> streams = streamMap.get(pe);

      if (streams != null)
      {
         for (Stream s : streams)
         {
            if (header)
               s.deregisterHeader(pe);
            else
               s.deregisterFooter(pe);
         }

         // cleanup
         streamMap.remove(pe);
      }
   }

   /**
    * Deregister (remove) the frame either as a header or footer from all
    * streams in which it is registered.
    *
    * @param    frameId
    *           The id of the frame for this request.
    */
   public void forceDeregisterPageElement(int frameId)
   {
      Integer     key = Integer.valueOf(frameId);
      PageElement pe  = pageElements.get(key);

      if (pe == null)
      {
         // we must never have been registered or we were cleanly
         // deregistered so we are done
         return;
      }

      Set<Stream> streams = streamMap.get(pe);

      if (streams != null)
      {
         for (Stream s : streams)
         {
            s.deregisterHeader(pe);
            s.deregisterFooter(pe);
         }

         // cleanup
         streamMap.remove(pe);
      }
   }

   /**
    * Force the given frame to output to the given stream in redirected mode.
    *
    * @param    rt
    *           The current redirected terminal for the stream.
    * @param    frameId
    *           The frame to output.
    * @param    streamId
    *           The output destination.
    */
   private void forceFrameDraw(RedirectedTerminal rt,
                               int                frameId,
                               int                streamId)
   {
      rt.setSyncable(true);
      view(frameId,
           null,                  //   frame buffer
           new int[] {frameId},   //   list of widgets
           streamId,              //   stream id
           false,                 //   register
           false,                 //   page top
           true,                  //   is view
           WindowConfig.RESOLVE_WINDOW);
      rt.setSyncable(false);
   }

   /**
    * Get input active counter.
    *
    * @return current value of the counter.
    */
   public int getInputActive()
   {
      return inputActive;
   }
   
   /**
    * Set input active counter.
    *
    * @param    inputActive 
    *           value of the counter.
    */
   public void setInputActive(int inputActive)
   {
      this.inputActive = inputActive;
   }
   
   /**
    * Get input active counter.
    *
    * @return current value of the counter.
    */
   public OutputManager getOutputManager()
   {
      return tk;
   }
   
   /**
    * Returns the column position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-COL
    * </pre>
    *
    * @param    frameId
    *           The ID of the frame for this request.
    *
    * @return   The column position for the upper-left corner of the frame.
    */
   @Override
   public double getFrameCol(int frameId)
   {
      Frame frame = getFrame(frameId);
      double result = 0;

      if (frame != null && frame.isVisible() && !frame.isRedirected())
      {
         result = ConfigHelper.getAssignedColumn(frame.config());
      }
      
      if (result == BaseConfig.INV_COORD)
      {
         result = 0;
      }

      return ++result;
   }
   
   /**
    * Get the PIXELS-PER-COLUMN value for this client session.
    * 
    * @return   See above.
    */
   public int getPixelsPerColumn()
   {
      return FontManager.getPixelsPerColumn(OutputManager.instance().getInstanceDriver());
   }
   
   /**
    * Get the PIXELS-PER-ROW value for this client session.
    * 
    * @return   See above.
    */
   public int getPixelsPerRow()
   {
      return FontManager.getPixelsPerRow(OutputManager.instance().getInstanceDriver());
   }

   /**
    * Change the specified font, in the user's environment.
    * This implements the <code>SYSTEM-DIALOG FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be changed.
    * @param    windowId
    *           The window ID to which the edited font-table belongs.
    */
   public void chooseFont(int fontNum, int windowId)
   {
      // TODO: implement me
      
      // TODO: all widgets (part of the specified window) having this font are immediately redrawn.
      // TODO: widgets are resized only if the auto-resize phrase is set.
   }

   /**
    * Set the number of entries in the font-table.
    * 
    * @param    numEntries
    *           The new size of the font-table. 
    */
   public void adjustFontTableSize(int numEntries)
   {
      // TODO: which window to use?
      Window window = WindowManager.getDefaultWindow();
      
      FontManager.adjustFontTableSize(window, numEntries);
   }

   /**
    * Get the number of entries in the font-table.
    * 
    * @return   See above.
    */
   public int fontTableSize()
   {
      // TODO: which window to use?
      Window window = WindowManager.getDefaultWindow();
      
      return FontManager.fontTableSize(window);
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public int getTextWidthPixels(String txt, int fontNum)
   {
      // TODO: which window to use?
      Window window = WindowManager.getDefaultWindow();
      
      return FontManager.getTextWidth(txt, window, fontNum, 
                                      OutputManager.instance().getInstanceDriver());
   }

   /**
    * Get the height in pixels, of the specified font.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public int getFontHeightPixels(int fontNum)
   {
      // TODO: which window to use?
      Window window = WindowManager.getDefaultWindow();
      
      return FontManager.getFontHeight(window, fontNum, 
                                       OutputManager.instance().getInstanceDriver());
   }
   
   /**
    * Resolve the font key associated with the given font, in the current window's font-table.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public String getFontKey(int fontNum)
   {
      // TODO: which window to use?
      Window window = WindowManager.getDefaultWindow();
      
      return FontManager.getFontKey(window, fontNum, 
                                    OutputManager.instance().getInstanceDriver());
   }
     
   /**
    * Change the specified color, in the user's environment.
    * This implements the <code>SYSTEM-DIALOG COLOR color-num</code> statement.
    * 
    * @param    colorNum
    *           The color-table entry to be changed.
    * @param    windowId
    *           The window ID to which the edited font-table belongs.
    */
   public void chooseColor(int colorNum, int windowId)
   {
      // TODO: implement me
      
      // TODO: all widgets (part of the specfied window) having this color are immediately redrawn.
   }
   
   /**
    * Inform the client-side that it can start sending the config updates.
    */
   @Override
   public void activateConfigUpdates()
   {
      ConfigManager.<ClientConfigManager>getInstance().activateConfigUpdates();
   }
   
   /**
    * Get the {@link EnvironmentDaemon} for this client.
    * 
    * @return   See above.
    */
   public EnvironmentDaemon getEnvironmentDaemon()
   {
      return ed;
   }
   
   /**
    * Is to force automatic clearing.
    * 
    * @return   See above.
    */
   public boolean isAfterChoose()
   {
      return afterChoose;
   }

   /**
    * Force or reset forcing of automatic clearing.
    * 
    * @param   afterChoose
    *          <code>true</code> force, <code>false</code> otherwise.
    */
   public void setAfterChoose(boolean afterChoose)
   {
      this.afterChoose = afterChoose;
   }

   /**
    * Get the last key, received form keyboard.
    * 
    * @return   See above.
    */
   public int getLastKey()
   {
      return lastKey;
   }
   
   /**
    * Set the last key, received form keyboard.
    * 
    * @param   lastKey
    *          Code of the last key.
    */
   public void setLastKey(int lastKey)
   {
      this.lastKey = lastKey;
   }

   /**
    * Returns the width of the specified frame (the <code>FRAME:WIDTH-CHARS
    * </code> attribute).
    *
    * @param    frameId
    *           The ID of the frame for this request.
    *
    * @return   The width of the frame.
    */
   @Override
   public double getWidth(int frameId)
   {
      Frame frame = getFrame(frameId);
      double width = 0;

      if (frame != null)
      {
         // the dialog-box can use client width version first
         width = frame.config().clientWidthChars != BaseConfig.INV_COORD
                    ? frame.config().clientWidthChars : frame.config().widthChars; 

         // safe, comparison of scaled doubles
         if (width < 0)
         {
            width = 0;
         }
      }

      return width;
   }

   /**
    * Returns the height of the specified frame (the <code>FRAME:HEIGHT-CHARS
    * </code> attribute).
    *
    * @param    frameId
    *           The ID of the frame for this request.
    *
    * @return   The height of the frame.
    */
   @Override
   public double getHeight(int frameId)
   {
      Frame frame = getFrame(frameId);
      double height = 0;

      if (frame != null)
      {
         // the dialog-box can use client height version first
         height = frame.config().clientHeightChars != BaseConfig.INV_COORD
                     ? frame.config().clientHeightChars : frame.config().heightChars; 

         // safe, comparison of scaled doubles
         if (height < 0)
         {
            height = 0;
         }
      }

      return height;
   }

   /**
    * Returns the number of iterations in a down frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-DOWN
    * </pre>
    *
    * @param    frameId
    *           The ID of the frame for this request.
    *
    * @return   The number of data iterations in the frame.
    */
   @Override
   public int getFrameDown(int frameId)
   {
      return getFrame(frameId).getDown();
   }

   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-LINE
    * </pre>
    *
    * @param    frameId
    *           The ID of the frame for this request.
    * @param    pendingUpDown
    *           The pending down value for this frame.
    *
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is not a down frame then 1 will be returned.
    */
   @Override
   public int getFrameLine(int frameId, int pendingUpDown)
   {
      return getFrameLine(getFrame(frameId), pendingUpDown);
   }

   /**
    * Returns the row position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-ROW
    * </pre>
    *
    * @param    frameId
    *           The ID of the frame for this request.
    *
    * @return   The row position for the upper-left corner of the frame.
    */
   @Override
   public double getFrameRow(int frameId)
   {
      Frame frame = getFrame(frameId);
      double result = 1;

      if (frame != null && frame.isVisible())
      {
         result = frame.location().y + 1;
      }

      return result;
   }

   /**
    * Gets the first widget in the tab order from the widgets of specified list.
    * Widgets from the list are assumed from the field-group for which this method is called. 
    * 
    * @param    frameId
    *           ID of the containing frame.
    * @param    widgetIDs 
    *           IDs of widgets in the target field-group.
    *           
    * @return   ID of the first widget in tab order.
    */
   @Override
   public int getFirstTabItem(int frameId, List<Integer> widgetIDs)
   {
      Frame frame = getFrame(frameId);

      if (frame == null)
         throw new IllegalArgumentException("Invalid frame ID " + frameId);
      
      if (widgetIDs == null)
         throw new NullPointerException("Specified field-group widgets IDs are null");
      
      // TODO needed for debug purpose. Delete in future.
      /*Widget[] list  = frame.getContentPane().widgets();
      for (Widget w : list)
      {
         System.out.println(w.config().id + ": " + w);
      }*/
      
      /*
       * TODO fix the widget IDs problem. It consists of the fact that any down-frame will contain
       * equal widget IDs set for widgets in each row. At the same time widgetIDs is the list of  
       * IDs in concrete field-group. There are separate field-group for each row in down frames.
       * Therefore at this moment there are no possibility to identify widgets in specified frame,
       * that are comprised in specified field-group widgets IDs list. There will be IDs from each 
       * row (field-group) that are comprised in widgetsIDs.
       */

      return Frame.getFirstFocusableId(frame, widgetIDs);
   }
   
   /**
    * Move specified widget to the first position in the tab order.
    * 
    * @param   frameId
    *          ID of the frame that contains widgets being moved.
    * @param   fgWidgetIds
    *          Target field-group widgets.
    * @param   widgetId
    *          ID of the widget to move.
    */
   @Override
   public void setFirstTabItem(int frameId, List<Integer> fgWidgetIds, int widgetId)
   {
      if (fgWidgetIds == null || fgWidgetIds.isEmpty())
         throw new IllegalArgumentException("Invalid field-group widgets list: emtpy ");
      
      Frame frame = getFrame(frameId);
      if (frame == null)
         throw new IllegalArgumentException("Invalid frame ID " + frameId);
      
      Widget widget = getWidget(widgetId);
      if (widget == null)
         throw new IllegalArgumentException("Invalid frame ID " + widgetId);

      Widget[] list  = frame.getContentPane().widgets();
      
      Widget firstWidget = null;
      
      for (int i = 0; i < list.length; i++)
      {
         int listWidgetId = UiUtils.lookupWidgetIdAsInt(list[i]);
         if (!fgWidgetIds.contains(listWidgetId))
         {
            continue;
         }
         
         if (UiUtils.getConfig(list[i]) == null)
         {
            continue;
         }
   
         if (!list[i].isEnabled())
         {
            continue;
         }
   
         if (!list[i].focusTraversable())
         {
            continue;
         }
   
         // insert before the first focusable.
         firstWidget = list[i];
         break;
      }
      
      if (firstWidget != null)
      {
         frame.getContentPane().moveBeforeTabItem(widget, firstWidget);
      }
   }
   
   /**
    * Returns the number of screen lines available for message display.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * MESSAGE-LINES
    * </pre>
    *
    * @return   The number of available lines in the message area.
    */
   @Override
   public int getMessageLines()
   {
      return Window.RESERVED_LINES - 1;
   }

   /**
    * Getter for the <b>NUM-SELECTED-ROWS </b> attribute.
    *
    * @param   browseId
    *          ID of the browse widget this request came for
    *
    * @return  current value of NUM-SELECTED-ROWS attribute
    */
   @Override
   public int getNumSelectedRows(int browseId)
   {
      Browse browse = (Browse) getWidget(browseId);

      if (browse != null)
         return browse.getSelectedRowsCount();

      return 0;
   }
   
   /**
    * Gets the widget that is the next from specified in the tab order.
    * 
    * @param    frameId
    *           ID of the containing frame.
    * @param    widgetId
    *           The widget to get the next widget in the tab order.
    * 
    * @return   ID of the widget that is the next from specified in the tab order.
    */
   @Override
   public int getNextTabItem(int frameId, int widgetId)
   {
      Widget comp = getWidget(widgetId);
      if (comp == null)
      {
         throw new IllegalArgumentException("Widget with id=" + widgetId + " does not exist.");
      }
      
      Widget nextEnabledWidget = getFrame(frameId).getNextEnabledWidget(comp, false);
      return UiUtils.getWidgetIdAsInt(nextEnabledWidget);
   }

   /**
    * Retrieve the 0-based row index of the <i>n</i>th selected row in the
    * given browse widget.
    *
    * @param   browseId
    *          ID of the browse widget associated with the request.
    * @param   n
    *          0-based position of the row in the list of selected rows. <code>-1</code> for the
    *          last selected row.
    *
    * @return  0-based row index associated with the <i>n</i>th selected row in the
    *          browse, or -1 if the associated Browse widget could not be
    *          identified.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   @Override
   public int getSelectedRowIndex(int browseId, int n)
   {
      Browse browse = (Browse) getWidget(browseId);

      if (browse != null)
         return browse.getSelectedRowIndex(n, false, true);

      return -1;
   }

   /**
    * Retrieve the row indexes of the selected rows in the given browse widget.
    *
    ** @param  browseId
    *          ID of the browse widget this request came for.
    *
    * @return  Array of 0-based row indexes of the selected rows in the browse.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   @Override
   public int[] getSelectedRowIndexes(int browseId)
   {
      Browse browse = (Browse) getWidget(browseId);

      if (browse != null)
         return browse.getSelectedRowIndexes();

      return null;
   }

   /**
    * Returns the number of screen lines available for frame display (does
    * not include the message or status lines).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * SCREEN-LINES
    * </pre>
    *
    * @return   The number of available screen lines.
    */
   @Override
   public double getScreenLines()
   {
      return WindowManager.resolveWindow().getScreenDimension().height;
   }

   /**
    * Returns the terminal type.  This usually corresponds to the value
    * set as the <code>TERM</code> environment variable on UNIX or Linux
    * unless set via {@link #setTerminalType}.  On Windows it depends upon
    * whether the terminal is in graphical or character mode.  If it is
    * in character mode, the result depends on the monitor type.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * TERMINAL
    * </pre>
    *
    * @return   Terminal type for in this session.
    */
   @Override
   public String getTerminalType()
   {
      return tk.terminalType();
   }

   /**
    * Makes the specified frame or widget invisible.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>HIDE widget </b>
    *
    * @param   widgetId
    *          ID of frame or widget this request came for
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    * @param   hidden
    *          If <code>true</code>, set the HIDDEN attribute.
    *          
    * @return  <code>true</code> if the widget has been hidden,
    *          <code>false</code> otherwise. 
    */
   @Override
   public boolean hide(final int widgetId, final boolean noPause, boolean hidden)
   {
      HidingProcessor hp = new HidingProcessor(widgetId, noPause, hidden);
      Widget widget = getWidget(widgetId);
      
      if (!hp.validateHide(widget))
      {
         return false;
      }
      
      eventDrawingBracket(widget, hp);
      return true;
   }

   /**
    * Hides all widgets and messages displayed in the message area. If pausing
    * is in effect, the standard message "Press space bar to continue" is
    * displayed on the status line and the user is required to press a key.
    * <p>
    * This method operates on the contents of the interactive terminal no
    * matter what redirection stream is in effect.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>HIDE ALL [NO_PAUSE] </b>
    *
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding.
    * @param   windowId
    *          Id of target window or -1 for default window
    */
   @Override
   public void hideAll(final boolean noPause, int windowId)
   {
      final Window wnd = targetWindowWorker(windowId);
      
      // do nothing if the terminal is redirected
      if (isRedirected())
      {
         return;
      }

      if (inEditingBlock)
         hadReset = true;

      final boolean[] framesToHide = {false};

      eventBracket(true, new Runnable()
      {
         @Override
         public void run()
         {
            framesToHide[0] = wnd.hideAllFrames();
         }
      });

      try
      {
         wnd.hideMessage(noPause);
      }

      finally
      {
         wnd.status(null);

         try
         {
            eventBracket(true, new Runnable()
            {
               @Override
               public void run()
               {
                  FramePlacementManager placementManager = wnd.getPlacementManager();
                  placementManager.hideAllFrames(!noPause); 
               }
            });
         }

         finally
         {
            wnd.clearScreenDirectly(framesToHide[0]);
         }
      }
   }

   /**
      
    * Hides all messages displayed in the message area. If pausing is in
    * effect, the standard message "Press space bar to continue" is displayed
    * on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>HIDE MESSAGE [NO_PAUSE] </b>
    *
    * @param   windowId
    *          Id of target window or -1 for default window
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   @Override
   public void hideMessage(boolean noPause, int windowId)
   {
      targetWindowWorker(windowId).hideMessage(noPause);
   }

   /**
    * Clears the type-ahead buffer if the input source is the keyboard.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>INPUT CLEAR </b>
    */
   @Override
   public void inputClear()
   {
      typeAhead.clear();
   }

   /**
    * Checks if the client is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public boolean isInBatchMode()
   {
      return (tk != null) ? tk.isBatchModeOverride() : false; 
   }
   
   /**
    * Checks if the client is running in batch background mode.
    * 
    * @return  <code>true</code> for background mode, <code>false</code> otherwise.
    * 
    */
   public boolean isBatchInBackground()
   {
      return OutputManager.isBatchInBackground();
   }
   
   /**
    * Stops the associated client session by generating UnstoppableExitException.
    */
   public void stopBatchSession()
   throws UnstoppableExitException
   {
      OutputManager.activateBatch(false);
      throw new UnstoppableExitException("");
   }

   /**
    * Changes the activation flag for batch mode driver if loaded.
    *
    * @param   on
    *          The new satus of the active flag for batch driver.
    */
   public void activateBatchMode(boolean on)
   {
      OutputManager.activateBatch(on);
   }
   
   /**
    * Displays the error message via current client side error writer.  This does not use
    * the client UI code tools, ususally meaning standard outpot channel anf further client
    * termination.
    *
    * @param   errmsg
    *          The error message to display.
    */
   public void displayError(String errmsg)
   {
      ErrorManager.displayError(errmsg);
   }
   
   /**
    * Displays the error message via current client side error writer.  This does not use
    * the client UI code tools, ususally meaning standard outpot channel anf further client
    * termination.  This is the redirected version of the displayError.  If the output is
    * redirected to file the message is writing to the file.
    *
    * @param   errmsg
    *          The error message to display.
    */
   public void displayErrorRedirected(String errmsg)
   {
      ErrorManager.displayErrorRedirected(errmsg);
   }
   
   /**
    * Renders the specified message into the message line.
    *
    * @param   message
    *          The message body to be rendered.
    */
   public void message(String message)
   {
      WindowManager.resolveWindow().message(message, Color.MESSAGES);
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    *
    * @param   text
    *          Text to be displayed.
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE,
    *          not used if var parameter is <code>null</code>.
    * @param   var
    *          The variable to set or update.  <code>null</code> means no
    *          set or update will occur.
    * @param   auto
    *          <code>true</code> if the AUTO-RETURN option is specified. This
    *          is only valid if <code>var</code> is not <code>null</code>.
    * @param   format
    *          The format specification to use for the set or update.  This
    *          is only valid if <code>var</code> is not <code>null</code>.
    * @param   cs
    *          The color specification to use for the text.
    * @param   windowId
    *          Id of target window
    *
    * @return  The result of the input operation if input parameter
    *          <code>var</code> is not <code>null</code> and operation
    *          finished successfully and/or if any key was pressed during
    *          editing/pausing.   If no pause occurred and there was no input
    *          (editing) in process, then <code>null</code> will be returned.
    */
   @Override
   public MessageReturnValue message(String       text,
                                     boolean      set,
                                     BaseDataType var,
                                     boolean      auto,
                                     String       format,
                                     Color        cs,
                                     int          windowId)
   {
      return targetWindowWorker(windowId).message(text, set, var, auto, format, cs);
   }

   /**
    * Displays an error message. Depending on the UI mode (ChUI vs GUI) the message is rendered
    * as a regular window message or as a GUI ALERT-BOX.
    * <p>
    * The method takes care of rendering the error message directly and does not interact with
    * an error writer.
    *
    * @param   text
    *          Error text to be displayed.
    */
   @Override
   public void displayErrorMessage(String text)
   {
      if (isChui())
      {
         ColorSpec cs = new ColorSpec(ColorSpec.COLOR_MESSAGES);
         targetWindowWorker(WindowConfig.RESOLVE_WINDOW).message(text, false, null, false, null,
                                                                 cs.convert());
      }
      else
      {
         // note that the error in GUI is displayed in an alert box regardless of the
         // SESSION:APPL-ALERT-BOXES value
         messageBox(new Object[] {text},
                    null,
                    LogicalTerminal.ALERT_ERROR,
                    LogicalTerminal.BTN_OK,
                    "Error",
                    null,
                    WindowConfig.RESOLVE_WINDOW,
                    null);
      }
   }

   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX
    * MESSAGE list VIEW-AS ALERT-BOX SET/UPDATE var
    * ... and all other variants that include options such as title, buttons
    * and colors
    * </pre>
    *
    * @param    list
    *           The list of message text components.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    color
    *           Color scheme to use.
    * @param    windowId
    *           Id of target window
    * @param    stack
    *           Progress stack trace
    *
    * @return   The edited result which can be mapped into a specified
    *           button.
    */
   @Override
   public logical messageBox(final Object[] list,
                                   logical  var,
                             final int      alertType,
                             final int      buttons,
                                   String   title,
                             final Color    color, 
                                   int      windowId,
                             final String[] stack)
   {
      // Special processing for ALERT-BOX message.
      if (((isRedirected() ||
                   FileChecker.isConsoleRedirected()) && OutputManager.inBatchMode()) ||
          (PlatformHelper.isUnderWindowsFamily() && OutputManager.isBatchInBackground()))
      {
         boolean reactivate = OutputManager.activateBatch(false);

         try
         {
            // Decode message box body
            StringBuilder sb = new StringBuilder();
            for (Object msgItem : list)
            {
               sb.append(msgItem);
            }
            // Here we have to separate Linux & Winodws redirected mode from background
            // mode in Winodws because the reaction is different
            if (isRedirected() || FileChecker.isConsoleRedirected())
            {
               // In this mode we have only error channel to display something in console
               ErrorManager.displayErrorRedirected(sb.toString());
            }
            else if (PlatformHelper.isUnderWindowsFamily() && OutputManager.isBatchInBackground())
            {
               // In this mode we have only error channel to display something in console
               ErrorManager.displayError(sb.toString());
            }
         }
         finally
         {
            if (reactivate)
            {
               OutputManager.activateBatch(true);
            }
         }
         
         // Return true value bypassing the rest of this method
         return new logical(true);
      }

      final AlertBox[] alrtBox = new AlertBox[1];

      if (var == null)
         var = new logical(true);

      if (title == null)
      {
         title = selectPredefinedTitle(alertType);
      }

      // save and check focus
      final Widget saveFocus  = UiUtils.getCurrentFocus();
      final int    focusState = FocusManager.focusState(saveFocus);

      EventList save = currentEventList;
      currentEventList = null;

      final String  ttle = title + (stack == null ? "" : " (Press HELP to view stack trace)");
      final logical bool = var;

      eventBracket(new Runnable()
      {
         @Override
         public void run()
         {
            WidgetFactory wf = tk.getFactory(); 
            alrtBox[0] = wf.createAlertBox(alertType, list, buttons, bool, ttle, color, stack);
            alrtBox[0].setEnabled(true);
            alrtBox[0].setVisible(true);
         }
      });

      Window wnd = targetWindowWorker(windowId);
      final Container pane = wnd.getContentPane();

      eventDrawingBracket(alrtBox[0], new Runnable()
      {
         @Override
         public void run()
         {
            // alert box can have two forms - an embeddable window or
            // top-level window
            if (!(alrtBox[0] instanceof TopLevelWindow))
               pane.add(alrtBox[0]);
            
            alrtBox[0].setFocus(alrtBox[0].focus());
            EventManager.postEvent(new FocusEvent(alrtBox[0].focus(),
                                                  EventType.FOCUS_GAINED));
            alrtBox[0].repaint();
         }
      });

      try
      {
         alertBoxActive++;

         messageBoxEventWorker(alrtBox[0]);
      }

      finally
      {
         alertBoxActive--;
         
         // this is needed when stop-condition is raised
         if (alrtBox[0].isVisible())
            alrtBox[0].setVisible(false);         

         switch (focusState)
         {
            case FocusManager.FOCUS_UNDEF:
               pane.setFocus(null);
               break;

            case FocusManager.FOCUS_VALID:
            case FocusManager.FOCUS_INVALID:
               // focus has to be restored no matter enabled or not
               focusManager.setFocusOn(saveFocus);
         }
         
         tk.setCursorStatus(true);

         currentEventList = save;
      }

      return alrtBox[0].getResult();
   }

   /**
    * Select one of the standard titles.
    *
    * @param   alertType
    *          Type of Alert.
    *
    * @return  selected title.
    */
   private String selectPredefinedTitle(int alertType)
   {
      String title = null;

      switch (alertType)
      {
         case LogicalTerminal.ALERT_QUESTION:
            title = " Question ";
            break;
         case LogicalTerminal.ALERT_INFO:
            title = " Information ";
            break;
         case LogicalTerminal.ALERT_ERROR:
            title = " Error ";
            break;
         case LogicalTerminal.ALERT_MESSAGE:
         case LogicalTerminal.APPL_ALERT:
            title = " Message ";
            break;
         case LogicalTerminal.ALERT_WARNING:
            title = " Warning ";
            break;
      }
      return title;
   }

   /**
    * Event worker for message box.
    *
    * @param   alrtBox
    *          Reference to AlertBox.
    */
   private void messageBoxEventWorker(final AlertBox alrtBox)
   {
      boolean isChui = isChui();
      
      while (!alrtBox.inputDone())
      {
         Event evt = null;

         // Progress does not maintain LASTKEY in a VIEW-AS ALERT-BOX so
         // this "direct" keyboard reading is safe (otherwise we would
         // need to call waitForKey() which does maintain LASTKEY)
         while (evt == null)
         {
            Object state = null;
            if (!isChui)
            {
               state = ((GuiDriver) tk.getInstanceDriver()).saveDrawingState();
            }

            evt = typeAhead.getKeystroke(1000, false);

            if (!isChui)
            {
               ((GuiDriver) tk.getInstanceDriver()).restoreDrawingState(state);
            }

            // handle mouse events etc.
            if (EventManager.hasOSEvents())
            {
               evt = EventManager.getOSEvent();
            }
         }
         
         if (evt instanceof InvocationEvent)
         {
            ((InvocationEvent) evt).getRunnable().run();
            continue;
         }

         if (evt instanceof KeyInput)
         {
            /* TODO set correct window */
            setNeedPause(false, WindowManager.resolveWindow());
            KeyInput key = (KeyInput) evt;
   
            int action = key.actionCode();
   
            // honor the GO, STOP, ERROR, ENDKEY and END-ERROR actions
            // (they DO NOT raise the associated condition but rather they
            // have alert-box specific functions)
   
            if (action == Keyboard.KA_GO)
            {
               // dismiss the alert box with by selecting the current
               // button
               key.setKeyCode(Key.VK_ENTER);
            }
            else if (action == Keyboard.KA_STOP)
            {
               tk.beep();
               continue;
            }
            else if (action == Keyboard.KA_ENDKEY ||
                     action == Keyboard.KA_ERROR  ||
                     action == Keyboard.KA_END_ERROR)
            {
               // dismiss the alert box with an "unknown" result
               eventDrawingBracket(alrtBox, new Runnable()
               {
                  @Override
                  public void run()
                  {
                     alrtBox.finished(new logical());
                  }
               });
               
               // skip further processing of the event, the alert-box is destroyed
               continue;
            }
         }

         final Event nextEvt = evt;

         // let the button and alert box handle the event
         eventDrawingBracket(alrtBox, new Runnable()
         {
            @Override
            public void run()
            {
               EventManager.postEvent(nextEvt);
            }
         });
      }
   }

   /**
    * Specifies the field to be put in focus during the next input operation
    * which involves this field of the frame.
    * <p>
    * If the very next operation does not involve the field, this request will
    * be disregarded.
    * <p>
    * The next high level input operation clears this information for the
    * involved frame only. WAIT-FOR does not clear it, however.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>NEXT-PROMPT </b>
    *
    * @param   frameId
    *          ID of frame this request came for
    * @param   widgetId
    *          widget to be put in focus
    */
   @Override
   public void nextPrompt(int frameId, int widgetId)
   {
      Frame frame = getFrame(frameId);

      if (frame != null)
      {
         frame.setFocusId(widgetId);
      }
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PAUSE n MESSAGE/NO-MESSAGE </b>
    *
    * @param   seconds
    *          Pause duration in seconds (0 means immediate return, -1 means
    *          infinite wait).
    * @param   text
    *          Either <code>null</code> if NO-MESSAGE is in effect, or the
    *          message text.
    * @param   noMessage
    *          <code>true</code> if NO-MESSAGE is in effect. This suppresses
    *          the message display, even if an explicit message text was
    *          passed.
    * @param   windowId
    *          Id of target window
    *
    * @return   The key code of the key pressed to clear the pause or -1
    *           if a timeout occurred (no key was pressed) or -2 if this
    *           was an immediate return due to a timeout of 0.
    */
   @Override
   public int pause(int seconds, String text, boolean noMessage, int windowId)
   {
      return pause(seconds, text, noMessage, targetWindowWorker(windowId));
   }
   
   /**
    * Suspends processing for the specified number of seconds or until the user presses any key,
    * whichever comes first. The type-ahead buffer is cleared. The specified (or no) message
    * appears on the status line.
    * <p>
    * Behaviour in case of batch processing: if the timeout is limited, it is honored, if infinite
    * ({@code seconds == -1}) return immediately.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PAUSE n MESSAGE/NO-MESSAGE </b>
    *
    * @param   seconds
    *          Pause duration in seconds (0 means immediate return, -1 means infinite wait).
    * @param   text
    *          Either {@code null} if NO-MESSAGE is in effect, or the message text.
    * @param   noMessage
    *          {@code true} if NO-MESSAGE is in effect. This suppresses the message display, even
    *          if an explicit message text was passed.
    * @param   wnd
    *          Not {@code null} target window. If {@code null} - an NPE will be thrown.
    *
    * @return  The key code of the key pressed to clear the pause or -1 if a timeout occurred (no
    *          key was pressed) or -2 if this was an immediate return due to a timeout of 0.
    */
   public int pause(int seconds, String text, boolean noMessage, TopLevelWindow wnd)
   {
      if (wnd == null)
      {
         throw new NullPointerException("Target window is not specified");
      }
      
      // if we are under Windows we have to hide the cursor unconditionally
      boolean cursorRestore = PlatformHelper.isUnderWindowsFamily();
      boolean cursorOn = false;
      
      boolean reactivateBatch = OutputManager.activateBatch(false);
      
      try
      {
         // special processing for PAUSE statement
         if (reactivateBatch)
         {
            if (seconds == -1)
            {
               // in batch mode, an infinite pause returns immediately
               seconds = 0;
            }
            noMessage = true;
         }
         
         if (cursorRestore)
         {
            cursorOn = tk.setCursorStatus(false);
         }
         // reset need pause condition
         setNeedPause(false, wnd);
         
         // pause immediately returns when called with a 0 timeout
         if (seconds == 0)
         {
            if (isChui())
            {
               // hide previously displayed messages even if PAUSE 0 called
               wnd.window().setMessageText(null, null);
            }
   
            // in this case the lastkey should not be changed
            return -2;
         }
   
         // clear type-ahead buffer
         typeAhead.clear();
   
         if (isRedirected() && !noMessage)
         {
            // if MESSAGE mode is in effect but the terminal is redirected,
            // switch output back to the terminal temporarily
   
            synchronized (cleanupSynchronizer)
            {
               cleanupBlock = new Runnable()
               {
                  private int tempTerm = forceInteractive();
   
                  @Override
                  public void run()
                  {
                     // take down temporary interactive terminal, restore redirection
                     switchOutput(tempTerm, false);
                  }
               };
            }
         }
   
         // display temporary status message
         if (!noMessage)
         {
            wnd.window().status((text == null) ? makePauseText(seconds) : text);
         }
   
         // track that a pause did occur
         hadPause = true;
   
         BlockingOperation saveOp = currentBlockingOp;
         currentBlockingOp = BlockingOperation.PAUSE;
         BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.PAUSE);
   
         // pause (this will maintain LASTKEY too, which is required!)
         // target widget is not important
         KeyInput key = (KeyInput) waitForEvent(seconds, false, false);
   
         try
         {
            // honor the STOP, ENDKEY and END-ERROR actions
            if (key != null)
               processCondition(key.actionCode(), -1, true);
         }
         finally
         {
            // erase status message, only in ChUI
            if (!noMessage && isChui())
            {
               wnd.window().status(null);
            }
   
            if (isChui())
            {
               // hide messages
               wnd.window().setMessageText(null, null);
            }
   
            pauseCleanup();
            // Show the cursor again if it was on under Windows
            if (cursorRestore && cursorOn)
            {
               tk.setCursorStatus(true);
            }
            
            currentBlockingOp = saveOp;
         }
   
         return lastKey;
      }
      finally
      {
         if (reactivateBatch)
         {
            OutputManager.activateBatch(true);
         }
      }
   }
   
   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The specified (or no)
    * message appears on the status line.
    * <p>
    * This method corresponds to Progress statements:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE MESSAGE text
    * PAUSE seconds BEFORE-HIDE NO-MESSAGE
    * </pre>
    *
    * @param    seconds
    *           Pause duration in seconds.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    * @param    windowId
    *           Id of target window
    */
   @Override
   public void pauseBeforeHide(int seconds, String text, int windowId)
   {
      // we must store the new values even if honorPBH is false!
      timePBH = seconds;
      textPBH = (text == null) ? PBH_DEFAULT_TEXT : text;
      
      // TODO call pause method with specified window?
   }

   /**
    * Processes all outstanding events without blocking for user input.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PROCESS EVENTS </b>
    *
    * @param   inbuf
    *          current screen buffer array for editable frames known to the server
    * 
    * @return  <code>ScreenBuffer</code> array with the current values
    */
   @Override
   public ScreenBuffer[] processEvents(ScreenBuffer[] inbuf)
   {
      // setup
      setScreenBuffers(inbuf);

      EventList save = currentEventList;
      currentEventList = new EventList(true);

      // TODO: can this work with multiple windows??
      tk.setInvalidate(true);
      processEventsWorker();
      tk.setInvalidate(WindowManager.resolveWindow(), false, true);

      // cleanup
      currentEventList = save;
      return getEditableClientScreenBuffers(
                getScreenBuffer(inbuf[0].getBaseId(), inbuf[0].size()));
   }

   /**
    * Register the specified ROW-DISPLAY events.
    *
    * @param   wids
    *          The id of the browse widget for which there are ROW-DISPLAY
    *          triggers.
    * @param   triggerIds
    *          The corresponding trigger ID for each widget.
    */
   @Override
   public void registerRowDisplayEvents(int[] wids, int[] triggerIds)
   {
      if (rowDisplayEvents == null)
      {
         rowDisplayEvents = new HashMap<Integer, Integer>(wids.length);
      }
      for (int i = 0; i < wids.length; i++)
      {
         rowDisplayEvents.put(wids[i], triggerIds[i]);
      }
   }

   /**
    * Remove the ROW-DISPLAY triggers from the specified list.
    *
    * @param   triggers
    *          The triggers to be removed, if they are associated to a
    *          ROW-DISPLAY event.
    */
   @Override
   public void removeRowDisplayEvents(int[] triggers)
   {
      if (rowDisplayEvents == null)
      {
         return;
      }

      Collection<Integer> values = rowDisplayEvents.values();
      Set<Integer> keys = rowDisplayEvents.keySet();

      for (int i = 0; i < triggers.length; i++)
      {
         if (values.contains(triggers[i]))
         {
            for (int wid : keys)
            {
               if (rowDisplayEvents.get(wid) == triggers[i])
               {
                  keys.remove(wid);
               }
            }
         }
      }
   }

   /**
    * Send array of menu element descriptions to the client.
    * 
    * @param   md
    *          Array of menu descriptions.
    */
   @Override
   public void pushMenuDescription(final MenuDescription[] md)
   {
      for (final MenuDescription definition : md)
      {
         final WidgetConfig cc = definition.getConfig(definition.getRootMenuID());
         final Widget menu = getWidget(definition.getRootMenuID());
         boolean ignorePaint = false;

         if (menu == null || cc == null)
         {
            ignorePaint = true;
         }
         else
         {
            ignorePaint = !menu.isVisible() && !cc.visible;
         }

         registry().pushMenuDescription(new MenuDescription[] { definition });
         
         /*
         TODO find out why it leads to error: menu is not attached to a window: from topLevelWindow
         eventDrawingBracket(menu, ignorePaint, true, new Runnable()
         {
            @Override
            public void run()
            {
               registry().pushMenuDescription(new MenuDescription[] { definition });
            }
         });
         */
      }
   }
   
   /**
    * Process array of frame/widget definitions provided by server.
    *
    * @param   sd
    *          Array of screen definitions.
    *          
    * @return  For each down frame, the configs for the widgets in the new row.  Otherwise
    *          <code>null</code> is returned.
    */
   @Override
   public WidgetConfig[] pushScreenDefinition(final ScreenDefinition[] sd)
   {
      // must be called before the main processing, should only be needed
      // during color stmt processing, otherwise this will be a NOP
      honorConditionalUpDown(false);

      List<WidgetConfig> wcfgs = new ArrayList<>();

      for (ScreenDefinition definition : sd)
      {
         Frame frame = (Frame) getWidget(definition.getRootFrameID());

         pushOneDef(definition);

         if (frame != null && frame.isDown())
         {
            // avoid new frames...
            WidgetConfig[] fwcfgs = frame.getDownConfigs();
            if (fwcfgs == null)
            {
               continue;
            }

            for (WidgetConfig wcfg : fwcfgs)
            {
               if (wcfg == null)
               {
                  continue;
               }
               
               if (WidgetId.virtualWidget(wcfg.id))
               {
                  continue;
               }
              
               wcfgs.add(wcfg);
            }
         }
      }
      
      return wcfgs.toArray(new WidgetConfig[0]);
   }

   /**
    * Push one screen definition.
    *
    * @param   definition
    *          Definition to push.
    */
   private void pushOneDef(final ScreenDefinition definition)
   {
      final WidgetConfig cc = definition.getConfig(definition.getRootFrameID());
      final Frame frame = (Frame) getWidget(definition.getRootFrameID());
      boolean ignorePaint = false;

      if (frame == null || cc == null || inRowDisplayTrigger)
      {
         ignorePaint = true;
      }
      else
      {
         ignorePaint = !frame.isVisible() && !cc.visible;
      }

      eventDrawingBracket(frame, ignorePaint, true, new Runnable()
      {
         @Override
         public void run()
         {
            // if the position was changed, before doing anything else,
            // we must hide the frame from the old position.
            if (frame != null && frame.isVisible())
            {
               preprocessFrameDef(frame, cc);
            }

            registry().pushDefinition(new ScreenDefinition[] { definition });
         }
      });
   }

   /**
    * Process window configuration provided by server.
    *
    * @param   config
    *          instance of <code>WindowConfig</code>
    */
   @Override
   public void pushWindow(final WindowConfig config)
   {  
      // registering the config here is needed. if the related window 
      // widget is not yet created, pushConfig call in the bracket would
      // otherwise replace all config changes performed as part of
      // the window initialization. the window widget initialization
      // will work with the registered config.
      ConfigManager cm  = ConfigManager.getInstance();
      if (cm.getActiveConfig(config.id) == null)
      {
         cm.addWidgetConfig(config);
      }
      
      Window.pushConfig(config);
   }
   
   /**
    * Remove the window instance from the client.
    * 
    * @param   window
    *          The window ID.
    */
   public void removeWindow(final int window)
   {
      eventBracket(new Runnable()
      {
         @Override
         public void run()
         {            
            WindowManager.removeWindow(window);
         }
      });
   }
   
   /**
    * Returns the ID of the currently active frame.
    *
    * @return   The frame ID or -1 if there is no active frame (or if the
                active frame has been marked for destruction).
    */
   @Override
   public int getActiveFrame()
   {
      Frame frame = getActiveFrameInt();

      if (frame == null)
         return -1;

      return frame.config().id.asInt();
   }

   /**
    * Turns cursor visibility on or off.
    * <p>
    * This method corresponds to the Progress statements <br>
    * <b>PUT CURSOR </b> <br>
    * <b>PUT CURSOR OFF </b>
    *
    * @param    onOff
    *           <code>true</code> if the cursor display is being turned on.
    */
   @Override
   public void putCursor(boolean onOff)
   {
      try
      {
         tk.setCursorStatus(onOff);
         cursorVisible = onOff;
      }
      catch (IllegalStateException ise)
      {
         // TODO: log this
         ise.printStackTrace(System.err);
      }
   }

   /**
    * Puts cursor to the specified screen position.  If the cursor visibility
    * is OFF, this method will first make the cursor visible before moving
    * it.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PUT CURSOR ROW COLUMN</b>
    *
    * @param    row
    *           1-based screen row number
    * @param    column
    *           1-based screen column number
    */
   @Override
   public void putCursor(double row, double column)
   {
      row = Coordinate.scale(row);
      column = Coordinate.scale(column);
      
      int tempTerm = forceInteractive();

      if(!cursorVisible)
      {
         putCursor(true);
      }

      double x = column;
      double y = row;

      Dimension screenSize = WindowManager.getDefaultWindow().dimension();
      
      // safe, comparison of scaled doubles
      x = x < 1 || x > screenSize.width  ? screenSize.width  - 2 : x - 1;
      y = y < 1 || y > screenSize.height ? screenSize.height - 1 : y - 1;

      ((ChuiOutputManager) tk).stayAt(new Point(x, y));

      restoreRedirection(tempTerm);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN ROW y COLUMN x
    * </pre>
    *
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    color
    *           The color specification or -1 for the default color.
    */
   @Override
   public void putScreen(String text, double row, double column, Color color)
   {
      row = Coordinate.scale(row);
      column = Coordinate.scale(column);
      
      // Special processing for PUT-SCREEN statement in batch mode if we have this active.
      boolean reactivateBatch = OutputManager.activateBatch(false);

      try
      {
         WindowManager.resolveWindow().setPutScreenPerformed(true);
   
         // trivial case first
         int textLen = text.length();
         if (textLen == 0)
         {
            return;
         }
   
         // isolate the peculiar case of putting text on the status line:
         // any output that would touch either of the two rightmost positions
         // is silently rejected
         Window<?> wnd = WindowManager.resolveWindow();
         if (wnd.config().statusArea)
         {
            Rectangle stat = wnd.getStatusLineRectangle();
            double statLine = stat.top();
            double screenChars  = stat.right() + 1;
            double effRow = row - 1;
            double effCol = column - 1;
      
            // safe, comparison of scaled doubles
            if (effRow == statLine)
            {
               // safe, comparison of scaled doubles
               if (effCol + textLen > screenChars - 2)
               {
                  return;
               }
            }
         }
   
         // OK, there will be some output here
         putScreenDirect(text, row, column, color);
      }
      finally
      {
         if (reactivateBatch)
         {
            OutputManager.activateBatch(true);
         }
      }
   }

   /**
    * Reads and returns one keystroke from the keyboard type-ahead buffer.
    * <p>
    * This method corresponds to the Progress statements <br>
    * <b>READKEY PAUSE n READKEY </b>
    *
    * @param   seconds
    *          -1 means indefinite wait, 0 means no wait and positive values
    *          mean the number of seconds to wait
    * @return  -1 if timed out, otherwise the event code
    */
   @Override
   public int readKey(int seconds)
   {
      int retKey = -1;

      // This is the special direct call to key handler for batch mode, no need to block
      // for key press waiting.
      if (OutputManager.inBatchMode())
      {
         return OutputManager.getDriver().readKey();
      }

      BlockingOperation saveOp = currentBlockingOp;
      currentBlockingOp = BlockingOperation.READKEY;
      
      try
      {
         do
         {
            Window wnd = WindowManager.resolveWindow();
            
            // reset need pause condition
            setNeedPause(false, wnd);
      
            // detect special case where we must raise a stop condition
            if (inEditingBlock && hadReset)
            {
               try
               {
                  stopNoAvailableWidgets();
               }
      
               finally
               {
                  hadReset = false;
               }
               
               // exit early
               break;
            }
      
            // TO-DO: prove 100% it's the right place for this logic!
            // continueEditing() is a candidate for this action
            if (inEditingBlock)
            {
               // if a message is output after a READKEY is executed, all messages
               // must be cleared
               wnd.setClearMessage(true);
            }
      
            BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.READKEY);
      
            // this will maintain LASTKEY too, which is required!
            // target widget is not important
            KeyInput key = (KeyInput) waitForEvent(seconds, false, false);
      
            // determine how this readkey is treated: as a system event, 4gl-level event, etc.
            retKey = (key == null ? -1 : key.keyCode());
            
            if (inEditingBlock && key != null && checkForSystemEvent(key))
            {
               // read more keys
               continue;
            }
            
            break;
         }
         while (true);
      }
      
      finally
      {
         currentBlockingOp = saveOp;
      }
      
      return retKey;
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    *
    * @param   browseId
    *          ID of the browse widget this request came for
    * @param   deleteCurrentRowMode
    *          <code>true</code> if this function is used for update on DELETE-CURRENT-ROW.
    *          <code>false</code> if this function is used for update on REFRESH.
    */
   @Override
   public void refresh(int browseId, boolean deleteCurrentRowMode)
   {
      Browse browse = (Browse) getWidget(browseId);

      tk.setInvalidate(true);

      if (browse != null)
         browse.refresh(deleteCurrentRowMode);
      cleanupQueue();

      tk.setInvalidate(browse, false, true);
   }

   /**
    * Refresh the browse after some of its rows were deleted.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   rowIndex
    *          0-based index of the deleted row. <code>-1</code> if all of the selected rows were
    *          deleted.
    */
   @Override
   public void refreshAfterDeletion(int browseId, final int rowIndex)
   {
      final Browse browse = (Browse) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {

               browse.refreshAfterDeletion(rowIndex);
            }
         });
      }
   }

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    *
    * @return  <code>true</code> on success.
    */
   @Override
   public logical selectNextRow(int browseId)
   {
      final logical res = new logical(false);
      final Browse browse = (Browse) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res.assign(browse.selectNextRow());
            }
         });
      }

      return res;
   }

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    *
    * @return  <code>true</code> on success.
    */
   @Override
   public logical selectPreviousRow(int browseId)
   {
      final logical res = new logical(false);
      final Browse browse = (Browse) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res.assign(browse.selectPreviousRow());
            }
         });
      }

      return res;
   }

   /**
    * Defines a specific action for the key.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>ON key-label key-function </b>
    *
    * @param   keyLabel
    *          the label of the key
    * @param   keyFunction
    *          the action to take when the user presses the key
    */
   @Override
   public void remapKey(String keyLabel, String keyFunction)
   {
      Keyboard.remapKey(keyLabel, keyFunction);
   }

   /**
    * Scrolls the lines of the down frame up or down.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>SCROLL </b>
    *
    * @param   frameId
    *          ID of frame this request came for
    * @param   fromCurrent
    *          <code>true</code> if FROM-CURRENT option was used
    * @param   up
    *          <code>true</code> if scrolling up
    */
   @Override
   public void scroll(final int     frameId,
                      final boolean fromCurrent,
                      final boolean up)
   {
      // must be called first!
      honorConditionalUpDown(false);

      // reset view flags
      viewRefresh = true;
      viewDisplay = false;

      final Frame frame = getFrame(frameId);
      
      eventDrawingBracket(frame, new Runnable()
      {
         @Override
         public void run()
         {
            frame.scroll(fromCurrent, up);
         }
      });
   }

   /**
    * Implements the <b>SELECT-ALL() </b> <b>BROWSE </b> widget method.
    *
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   @Override
   public void selectAll(int browseId)
   {
      Browse browse = (Browse) getWidget(browseId);

      tk.setInvalidate(true);

      if (browse != null)
         browse.selectAll();
      cleanupQueue();

      tk.setInvalidate(browse, false, true);
   }

   /**
    * Implements the <b>SELECT-FOCUSED-ROW() </b> <b>BROWSE </b> widget
    * method.
    *
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   @Override
   public void selectFocusedRow(int browseId)
   {
      Browse browse = (Browse) getWidget(browseId);

      tk.setInvalidate(true);

      if (browse != null)
         browse.selectFocusedRow();
      cleanupQueue();

      tk.setInvalidate(browse, false, true);
   }

   /**
    * Implements the <b>DESELECT-FOCUSED-ROW() </b> <b>BROWSE </b> widget
    * method.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    */
   @Override
   public void deselectFocusedRow(int browseId)
   {
      final Browse browse = (Browse) getWidget(browseId);

      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.deselectFocusedRow();
            }
         });
      }
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p> Deselects the <i>n</i>th selected
    * row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   rowIndex
    *          1-based index of the row (in the list of selected rows) to deselect.
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   @Override
   public boolean deselectSelectedRow(int browseId, final int rowIndex)
   {
      final Browse browse = (Browse) getWidget(browseId);
      final boolean[] res = {false};

      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res[0] = browse.deselectSelectedRow(rowIndex - 1);
            }
         });
      }

      return res[0];
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   rowIndex
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   @Override
   public boolean selectRow(int browseId, final int rowIndex)
   {
      final Browse browse = (Browse) getWidget(browseId);
      final boolean[] res = {false};

      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res[0] = browse.selectRow(rowIndex - 1);
            }
         });
      }

      return res[0];
   }

   /**
    * Implements the <b>IS-ROW-SELECTED()</b> <b>BROWSE </b> widget method.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   rowIndex
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   @Override
   public boolean isRowSelected(int browseId, final int rowIndex)
   {
      final Browse browse = (Browse) getWidget(browseId);
      final boolean[] res = {false};

      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res[0] = browse.isRowSelected(rowIndex - 1);
            }
         });
      }

      return res[0];
   }

   /**
    * Implements the client side of the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  srcIndex
    *         0-based position of the column to be moved.
    * @param  destIndex
    *         0-based position to which the column is moved.
    */
   public void moveColumnClient(int browseId, int srcIndex, int destIndex)
   {
      final Browse browse = (Browse) getWidget(browseId);

      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.moveColumn(srcIndex, destIndex);
            }
         });
      }
   }

   /**
    * Implements the server-side MOVE-COLUMN() widget method.<p>
    * Repositions a column in the browse widget.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  src
    *         0-based position of the column to be moved.
    * @param  dest
    *         0-based position to which the column is moved.
    */
   public void moveColumnServer(int browseId, int src, int dest)
   {
      server.moveColumnServer(browseId, src, dest);
   }

   /**
    * Implements the FOCUSED-ROW attribute getter.<p>
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    *
    * @return  1-based index of the focused row in the viewport or <code>-1</code> if browse
    *          doesn't exist.
    */
   public int getFocusedRow(int browseId)
   {
      final Browse browse = (Browse) getWidget(browseId);
      final int[] res = {-1};

      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res[0] = browse.getFocusedRow();
            }
         });
      }

      return res[0];
   }

   /**
    * Get number of rows in the browse viewport set either by DOWN or SIZE options.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    *
    * @return  number of rows in the browse viewport.
    */
   @Override
   public int getBrowsePageRowCount(int browseId)
   {
      final Browse browse = (Browse) getWidget(browseId);
      final int[] res = {0};

      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res[0] = browse.getDown();
            }
         });
      }

      return res[0];
   }

   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   leaveOnRowInView
    *          If the current row is in the viewport, do not update the scroll row.
    * @param   preserveViewportRow
    *          Try to preserve the current viewport row (i.e. the row at the same position in
    *          the viewport becomes the current row after refresh).
    * @param   repositioned
    *          If the browse is refreshed because the query was repositioned.
    * @param   repositionError
    *          For reposition mode - if an error has happened during reposition.
    * @param   targetRepositionRow
    *          For reposition mode - row used as the target for reposition. If it was deleted,
    *          reposition with fetching may end up on a different row.
    */
   @Override
   public void refreshScrollRow(int browseId,
                                final boolean leaveOnRowInView,
                                final boolean preserveViewportRow,
                                final boolean repositioned,
                                final boolean repositionError,
                                final int targetRepositionRow)
   {
      final Browse browse = (Browse) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.refreshScrollRow(leaveOnRowInView,
                                       preserveViewportRow,
                                       repositioned,
                                       repositionError,
                                       targetRepositionRow);
               browse.window().clearPutScreenInMessageArea(-1);
            }
         });
      }
   }

   /**
    * Execute pending refresh when REFRESHABLE is turned on. Includes pending query repositions
    * and reopenings.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   queryReopened
    *          <code>true</code> if there is a pending query reopening.
    * @param   repositionBeforeReopen
    *          <code>true</code> if the last pending reposition was <i>before</i> the last pending
    *          reopening.
    * @param   repositionedRow
    *          0-based index of the last row the query was repositioned to. <code>-1</code> if
    *          there is no repositioned row.
    */
   @Override
   public void executePendingRefresh(int browseId,
                                     final boolean queryReopened,
                                     final boolean repositionBeforeReopen,
                                     final int repositionedRow)
   {
      final Browse browse = (Browse) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.executePendingRefresh(queryReopened,
                                            repositionBeforeReopen,
                                            repositionedRow);
            }
         });
      }
   }

   /**
    * Notify the browse that the underlying query has been repositioned..
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   repositionError
    *          For reposition mode - if an error has happened during reposition.
    * @param   targetRepositionRow
    *          For reposition mode - row used as the target for reposition. If it was deleted,
    *          reposition with fetching may end up on a different row.
    */
   @Override
   public void queryRepositioned(int browseId,
                                 final boolean repositionError,
                                 final int targetRepositionRow)
   {
      final Browse browse = (Browse) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.queryRepositioned(repositionError, targetRepositionRow);
               browse.window().clearPutScreenInMessageArea(-1);
            }
         });
      }
   }

   /**
    * Notify browse widget at client side that query was closed.
    *
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   @Override
   public void queryClosed(final int browseId)
   {
      final Browse browse = (Browse) getWidget(browseId);
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               browse.queryClosed();
               browse.window().clearPutScreenInMessageArea(-1);
            }
         });
      }
   }

   /**
    * Returns <code>true</code> if the current row is a new row and is not present in the
    * database.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    *
    * @return  <code>true</code> if the current row is a new row and <code>false</code> otherwise.
    */
   @Override
   public boolean isNewRow(final int browseId)
   {
      Browse browse = (Browse) getWidget(browseId);
      return browse != null && browse.isNewRow();
   }

   /**
    * Returns <code>true</code> if the current browse row is modified.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    *
    * @return  <code>true</code> if the current browse row is modified.
    */
   @Override
   public boolean isCurrentRowModified(final int browseId)
   {
      Browse browse = (Browse) getWidget(browseId);
      return browse != null && browse.isCurrentRowModified();
   }

   /**
    * Inserts a blank row in an updateable browse before or after the last selected row.
    * The blank row is a placeholder for a new record to be added through the browse.
    *
    * @param   browseId
    *          ID of the browse widget this request came for.
    * @param   after
    *          <code>true</code> if the row is added after the last selected row.
    *          <code>false</code> if the row is added before the last selected row.
    *
    * @return  <code>true</code> if the row was successfully added.
    */
   @Override
   public boolean insertRow(final int browseId, final boolean after)
   {
      final Browse browse = (Browse) getWidget(browseId);
      final boolean res[] = new boolean[] {false};
      if (browse != null)
      {
         eventDrawingBracket(browse, new Runnable()
         {
            @Override
            public void run()
            {
               res[0] = browse.insertRow(after);
            }
         });
      }

      return res[0];
   }

   /**
    * Invokes the SCROLL-TO-ITEM() method on the specified selection-list widget.
    *
    * @param    slId
    *           Widget ID of the selection list to scroll.
    * @param    index
    *           The 0-based index of the item to which to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   @Override
   public boolean scrollToItem(int slId, int index)
   {
      SelectionList sl = (SelectionList) getWidget(slId);
      
      return sl != null ? sl.scrollToItem(index) : false;
   }

   /**
    * Sets the terminal type.  This will cause the terminal to potentially
    * change its configuration if the new type is different from the old
    * type.
    * <p>
    * This processing also clears the input buffer and resets any key
    * mappings to their default values.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>TERMINAL</b>
    *
    * @param   term
    *          Terminal type name.  If <code>null</code>, this resets the
    *          terminal without changing the type.
    */
   @Override
   public void setTerminalType(String term)
   {
      // restore default key mappings (must be done on both server and client)
      Keyboard.resetKeyboard();

      // clear typeahead
      inputClear();

      if (term != null)
      {
         tk.setTerminalType(term);
      }
      else
      {
         // the extra hide message here is needed because this one is
         // supposed to pause (but the one in hideAll won't)
         Window wnd = WindowManager.getDefaultWindow(); 
         wnd.hideMessage(false);
         hideAll(true, UiUtils.getWidgetIdAsInt(wnd));

         // calling resetTermType causes catastrophic behavior so let's avoid
         // it (in Progress it seems to make things bad too but in ncurses
         // things just stop working altogether)
         // Toolkit.resetTermType();
      }
   }

   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method is used to display a status message, which is trimmed to
    * 63 chars only if the <code>noLengthLimit</code> flag is not set.
    *
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    * @param    noLengthLimit
    *           flag indicating the status message should not be trimmed
    * @param    windowId
    *           Id of target window
    */
   @Override
   public void statusDefault(String text, boolean noLengthLimit, int windowId)
   {
      Window wnd = targetWindowWorker(windowId);
      wnd.statusDefault(text, noLengthLimit);
   }

   /**
    * Specifies a new status message used when the user interface is blocked
    * on input.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT text</code>.
    *
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    * @param    windowId
    *           Id of target window
    */
   @Override
   public void statusInput(String text, int windowId)
   {
      Window wnd = targetWindowWorker(windowId);
      wnd.statusInput(text);
   }

   /**
    * Reverts the input status message to the default.
    * 
    * @param    windowId
    *           Id of target window
    */
   @Override
   public void statusInputRevert(int windowId)
   {
      statusInput(getInputDefaultText(), windowId);
   }

   /**
    * Disables display of the input status message.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT OFF IN WINDOW wnd</code>.
    * 
    * @param   windowId
    *          Id of target window
    */
   @Override
   public void statusInputOff(int windowId)
   {
      statusInput(null, windowId);
   }

   /**
    * Shows all fields of the specified frame on screen.
    *
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer with current values of widgets.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    isView
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    * @param    windowId
    *           target window id.
    *           
    * @return   The frames screen buffer array.
    */
   @Override
   public ScreenBuffer[] view(int            widgetId,
                              ScreenBuffer[] frameBuf,
                              int[]          widgetIds,
                              boolean        isView,
                              int            windowId)
   {
      view_(widgetId, frameBuf, widgetIds, isView, true, windowId);
      
      Widget widget = getWidget(widgetId);

      // nothing to do?
      if (widget == null)
      {
         return null;
      }

      int frameId = registry().lookupFrameIdFromWidgetId(widgetId);

      ScreenBuffer[] sb = (frameId == -1 ? null : getEditableClientScreenBuffers(
                                                     getScreenBuffer(frameId, -1, true)));

      if (sb != null)
      {
         Frame frame = (Frame) getWidget(frameId);
         sb[0].setDownConfigs(frame.getDownConfigs());
      }
      
      return sb;
   }

   /**
    * Write message and stack trace to the client log
    * 
    * @param message
    *        the message
    * @param stack
    *        the Progress stack trace
    */
   @Override
   public void clientLog(String message, String[] stack)
   {
      if (clientLog == null)
      {
         return;
      }
      clientLog.log(Level.INFO, message);

      if (stack == null)
      {
         return;
      }

      clientLog.log(Level.INFO, "** ABL Debug-Alert Stack Trace **");
      String prefix = "--> ";
      for (String s: stack)
      {
         clientLog.log(Level.INFO, prefix + s);
         prefix = "    ";
      }
   }

   /**
    * Write message and stack trace to the client log
    * 
    * @param message
    *        the message
    */
   @Override
   public void logStackTrace(String message)
   {
      clientLog(message, server.getProgressStackTrace());
   }

   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * Notes on the logic of view/display calls in redirected mode:
    * <p>
    * Listed below are steps taken in redirected mode before the common worker
    * viewWorker() is called.
    * <ul>
    *  <li>
    *     Check the redirected terminal (RT) ownership. If it's owned by
    *     another frame, do frame switching and set SWITCHED flag.
    *     otherwise clear SWITCHED flag and proceed with the next step.<br>
    *     Frame switching flushes the RT buffer to the stream S and sets
    *     the RT ownership to F. The RT buffer becomes fresh.
    *  <li>
    *     Check if the RT buffer is fresh. If it is not, all the remaining
    *     steps are skipped.
    *  <li>
    *     The RT buffer is fresh and owned by the frame F. This is the only
    *     moment when we can determine whether:
    *     <ul>
    *       <li> the frame has to be drawn with or without headers;
    *       <li> the current page needs to be ejected;
    *       <li> the current page needs to be contiguous.
    *     </ul>
    *     The following steps do these determinations.
    *  <li>
    *     Query the stream S and obtain the following values:
    *     <ul>
    *        <li> a flag FRESH-PAGE, telling whether the page is fresh
    *             or contains some output (page headers are not considered
    *             as output here)
    *        <li> page size PS;
    *        <li> remaining lines on the page RL. This value accounts for
    *             the size of the current output within S and all defined
    *             page footers.
    *     </ul>
    *  <li>
    *     Query the frame F and obtain the following values:
    *     <ul>
    *        <li> a flag DOWN-FRAME which is true for the down frame;
    *        <li> a flag PAGE-LIST which is true if DOWN-FRAME is true and
    *             when F is not in pagedList for the stream S (pagedList gets
    *             reset with page break notifications as well as during frame
    *             switching and reflects the fact that this down frame has
    *             contributed some output to the current page with frame
    *             headers already in place)
    *        <li> size of frame headers HS;
    *        <li> size of frame body BS
    *     </ul>
    *  <li>
    *     Determine whether the frame F needs headers and set NEED-HDR flag
    *     accordingly.<br>
    *     Headers are needed if any condition below is true:
    *     <ul>
    *        <li> frame switch has happened, SWITCHED flag is set;
    *        <li> fresh page flag FRESH-PAGE is set;
    *        <li> this is a 1-down frame (DOWN-FRAME flag is reset);
    *        <li> PAGE-LIST flag is set.
    *     </ul>
    *  <li>
    *     Calculate the number of lines the frame F needs on the page as<br>
    *     <code>FS = (NH ? HS : 0) + BS</code>
    *     <br>
    *  <li>
    *     Handle the large output case. The large output means the frame size
    *     FS is greater than the page size PS and output is paged. If this is
    *     the case:
    *     <ul>
    *        <li> check whether we are on the fresh page, and if we are not,
    *             call the stream S and force page break;
    *        <li> now we are on the fresh page; call the stream S and set
    *             contiguous mode.
    *     </ul>
    *     The contiguous mode gets reset with the next flush and page break
    *     operation.<br>
    *     In contiguous mode the rest of steps is skipped.
    *  <li>
    *     Determine whether the page break is required. If the output is paged
    *     and the frame size FS is greater than the remaining lines RL, call
    *     the stream S and force page break.<br>
    *     Page break outputs Form Feed character, resets page state variables
    *     and outputs page headers. During page headers drawing, the RT will
    *     be used for drawing, but this is safe since at this time the RT is
    *     fresh and the page break processing will return it back fresh.
    * </ul>
    * Now the current view/display operation can proceed with drawing into
    * the RT buffer with or without frame headers according to
    * the NEED-HDR flag.
    *
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer array with current values of widgets in editable frames.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    isView
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    * @param    markTarget
    *           <code>true</code> if the call is for regular frame,
    *            <code>false</code> for page elements rendering
    * @param    windowId
    *           target window id.
    */
   @SuppressWarnings("null")
   private void view_(int            widgetId,
                      ScreenBuffer[] frameBuf,
                      int[]          widgetIds,
                      boolean        isView,
                      boolean        markTarget,
                      int            windowId)
   {
      Widget widget = getWidget(widgetId);

      preprocessView(frameBuf, widget, widgetIds);

      // nothing to do?
      if (widget == null)
         return;

      // check the frame for possible stream switching
      int   frameId = registry().lookupFrameIdFromWidgetId(widgetId);
      Frame frame   = null;
      if (frameId != -1)
      {
         frame = (Frame) getWidget(frameId);
      }

      if (frame != null)
      {
         switchStreams(frame);
      }

      Frame wframe = (Frame) (widget instanceof Frame ? widget : getWidget(frameId));
      UIStatement stmt = (frameBuf == null || frameBuf[0] == null 
                           ? null 
                           : frameBuf[0].getCurrentStatement());

      if (!isRedirected())
      {
         if (frame != null && frame.isDown())
            frame.prepareIteration(widget, frameBuf[0], widgetIds, isView);

         // must be called first!
         honorConditionalUpDown();

         try
         {
            if (wframe != null)
            {
               wframe.setCurrentStatement(stmt);
            }
            
            // do the drawing
            viewWorker(widget, frameBuf, widgetIds, isView, windowId);
         }
         finally
         {
            if (wframe != null)
            {
               wframe.setCurrentStatement(null);
            }
         }
         
         return;
      }

      // the rest of the code is for redirected mode only

      Stream s = sd.streamFromId(currentStream);

      if (s == null)
      {
         ErrorManager.recordOrThrowError(1387, "Attempt to write to closed stream");
         return;
      }

      RedirectedTerminal rt = redirections.get(currentStream);

      // filter out the case of utility call to render a page element
      if (markTarget)
      {
         // check the redirected terminal's ownership and
         // notify the redirected terminal about the new target frame
         boolean freshRT     = false;
         boolean switched    = switchFrames(currentStream, frameId);
         boolean needHeaders = false;

         // figure out the redirected terminal's buffer state
         if (switched)
         {
            freshRT = true;
         }
         else
         {
            freshRT = rt.isFresh();
         }

         // check to see if the redirected terminal's buffer is fresh
         if (freshRT)
         {
            // this is the moment for establishing the place of the frame
            // with the stream

            // query the stream and collect all state variables we may need
            boolean freshPage = s.isFreshPage();
            int     pageSize  = s.getPageSize().intValue();
            int     remLines  = s.getAvailableLines();

            // query the frame and collect all state variables we may need
            boolean     downFrame = frame.isDown();
            double      headSize  = frame.getFixedHeight();
            boolean     noHeaders = !frame.isWithHeaders(s);
            boolean     box       = frame.config().box;
            double      bodySize  = frame.getIterationHeight();

            // determine whether this frame needs to be drawn with headers
            needHeaders = (switched | freshPage | !downFrame | noHeaders);

            // the remaining steps are needed for the paged output only
            if (pageSize > 0)
            {
               // calculate the frame size on the output
               double frameSize = (needHeaders ? headSize : 0) + bodySize;
               frameSize = Coordinate.scale(frameSize);

               // boxed frames that have to draw their headers must account
               // for the top border line (in redirected mode only the top
               // border draws)
               if (box && needHeaders)
                  frameSize++;

               // handle the large output case where the frame is bigger than
               // there is available room
               if (frameSize > remLines)
               {
                  if (!freshPage)
                  {
                     // break the page now (early)
                     s.forcePageBreak();

                     // ensure that the full frame is redrawn
                     needHeaders = true;
                  }

                  // ensure uninterruptible output for this whole frame
                  s.setContiguous(true);
               }
            }
         }

         // register event listeners
         s.registerCloseListener(this);
         s.registerAdvanceListener(this);
         s.registerFlushListener(this);

         // the final preparation for the frame drawing
         if (needHeaders)
         {
            // force drawing of headers and iteration
            nextViewRefresh = true;
            nextViewDisplay = false;
         }
         else
         {
            // force drawing of iteration only
            nextViewRefresh = false;
            nextViewDisplay = true;
         }
      }

      try
      {
         if (wframe != null)
         {
            wframe.setCurrentStatement(stmt);
         }
         
         // do the drawing
         viewWorker(widget, frameBuf, widgetIds, isView, windowId);
      }
      finally
      {
         if (wframe != null)
         {
            wframe.setCurrentStatement(null);
         }
      }
   }

   /**
    * Retrieve the ID of the frame containing the widget with a given ID.
    *
    * @param    widgetId
    *           The widget ID.
    *
    * @return   The frame ID if widget is found or -1 if not.
    */
   public static int frameIdFromWidgetId(int widgetId)
   {
      return getInstance().registry().lookupFrameIdFromWidgetId(widgetId);
   }

   /**
    * Support for the DISPLAY followed by DOWN combination.
    * 
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer array with current values of widgets for editable frames.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call).
    * @param    register
    *           <code>true</code> if the frame associated with this request
    *           must be registered as a page element with the given stream. 
    * @param    pageTop
    *           <code>true</code> if the page element registration is for
    *           a page top element, <code>false</code> for a page bottom
    *           element.  This parameter is ignored if <code>register</code>
    *           is <code>false</code>.
    * @param    windowId
    *           Id of target window or {@link WindowConfig#RESOLVE_WINDOW} for 
    *           current or default window.
    */
   public void displayAndDown(int widgetId, 
                              ScreenBuffer[] frameBuf, 
                              int[] widgetIds, 
                              int streamId,
                              boolean register,
                              boolean pageTop,
                              int windowId) 
   {
      view(widgetId, frameBuf, widgetIds, streamId, register, pageTop, false, windowId);
      down(widgetId, 1, streamId, true, false);
   };

   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * The terminal will be temporarily redirected to the given stream for the
    * duration of this method call.  This allows the caller to avoid
    * bracketing this call with {@link #switchOutput} invocations, as the
    * same result is achieved internal to this method's operation.
    *
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer array with current values of widgets for editable frames.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call). If this is not -1,
    *           then the any pending PUT or EXPORT output on this stream will
    *           be ended/forced out.
    * @param    register
    *           <code>true</code> if the frame associated with this request
    *           must be registered as a page element with the given stream.
    * @param    pageTop
    *           <code>true</code> if the page element registration is for
    *           a page top element, <code>false</code> for a page bottom
    *           element.  This parameter is ignored if <code>register</code>
    *           is <code>false</code>.
    * @param    isView
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    *           
    * @return   The frame's widget config array.
    */
   @Override
   public WidgetConfig[] view(int            widgetId,
                              ScreenBuffer[] frameBuf,
                              int[]          widgetIds,
                              int            streamId,
                              boolean        register,
                              boolean        pageTop,
                              boolean        isView,
                              int            windowId)
   {
      // end any PUT or EXPORT output that is pending and write it to the
      // stream
      endOutput(streamId);

      boolean switched = currentStream == streamId;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(streamId, true);
      }

      if (register)
      {
         // this deregistration removes the current registration for this
         // page element for the following reasons:
         // - the view_ call produces a single copy of frame anyway
         // - the registration that follows keeps the right order of elements
         deregisterPageElement(frameBuf[0].getBaseId(), pageTop, streamId);
      }

      try
      {
         view_(widgetId, frameBuf, widgetIds, isView, true, windowId);
      }

      finally
      {
         // (re-)register page element now
         if (register)
         {
            registerPageElement(frameBuf[0].getBaseId(), pageTop, streamId);
         }

         if (!switched)
         {
            switchOutput(tempOut, false);
         }
      }
      
      Widget widget = getWidget(widgetId);

      // nothing to do?
      if (widget == null)
      {
         return null;
      }

      int frameId = registry().lookupFrameIdFromWidgetId(widgetId);

      if (frameId == -1)
      {
         return null;
      }

      Frame frame = (Frame) getWidget(frameId);
      return frame == null ? null : frame.getDownConfigs();
   }

   /**
    * Provides a composite helper to implement a single call to the client
    * which does an enable, wait-for, disable and hide message (in that
    * order).  This single call is more efficient than executing the calls
    * separately since the client has all the state needed AND since it
    * avoids extra round trips.
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    widgetIds
    *           array of field level widget IDs to be enabled or
    *           <code>null</code>, which means all of them.
    * @param    sb
    *           The currently editable frames screen buffer array.
    * @param    seconds
    *           Time to wait for event.
    * @param    el
    *           List of events to wait for.
    * @param    doIO
    *           tells to perform the requested operation if <code>true</code>.
    *           otherwise prepares the frame for editing session only.
    * @param    windowId
    *           Id of target window or {@link WindowConfig#RESOLVE_WINDOW} for 
    *           current or default window.
    *
    * @return   <code>ScreenBuffer</code> array with the current values.
    */
   @Override
   public ScreenBuffer[] promptFor(int            frameId,
                                   int[]          widgetIds,
                                   ScreenBuffer[] sb,
                                   int            seconds,
                                   EventList      el,
                                   boolean        doIO,
                                   int            windowId)
   {
      // TO-DO (eventually):
      // In Progress, the active editing block can't be reentered with any
      // other IO operation like SET, PROMPT-FOR or UPDATE issued against
      // the same frame. WAIT-FOR is allowed, however. Currently, we don't
      // check this limitation and simply stack the state for recursion.

      // save and reset editing mode flag for possible recursion
      boolean saveInEditingBlock = inEditingBlock;
      inEditingBlock = false;

      int savedInsState = saveInsState();
      
      // starting a new interactive session, we must clear this state if set (TODO: not sure
      // why leaveSent would already be set true on entry to this method, but it has been seen
      // in editing blocks and will break proper leave processing if we don't clear it; we
      // probably need to find out where we are not clearing the flag, but this seems safe
      // since we are starting a new edit with the user; in editing blocks we won't drop into
      // the wait-for portion (looping on the event processing), so this method will actually
      // return and the event processing is driven from a server-side loop that does a readkey
      // and apply; for this reason we cannot save/restore the leaveSent as is done elsewhere
      // but instead this must be unconditional)
      if (!doIO)
      {
         leaveSent = false;
      }

      // this call guarantees frame is not null
      Frame frame = getFrame(frameId);
      Window wnd = frame.window();

      switchStreams(frame);

      boolean exception = true;
      boolean saveCHO   = frame.isCanHideOthers();
      ignoreNextStopEditing = false;

      try
      {
         enable(frameId, widgetIds, sb, true, windowId, false);
         
         Dimension screenSize = wnd.dimension();

         //TODO<SIY>: probably calculations below should be revised for 
         // GUI, but this should be checked deeper 
         if (lastExposedFrame != frame &&
             (lastExposedFrame == null ||
             !frame.config().box       ||
              frame.height() < screenSize.height - 4 ||
              frame.calculateFrameRect(false).intersects(
                   new Rectangle(lastExposedFrame.location(),
                                 lastExposedFrame.dimension(),
                                 lastExposedFrame.screen().coordinates().baseUnits()))) &&
             !UiUtils.containsFrames(frame, new Rectangle(new Point(0, 0),
                                                          screenSize,
                                                          frame.screen().coordinates().baseUnits())))
         {
            frame.refresh(frame.calculateFrameRect(false));
         }

         exception = false;
      }
      finally
      {
         if (!doIO && exception)
         {
            // We were about to enter editing mode but user had pressed F4.
            // Cleanup code at server is not aware about real reason and
            // should not make assumptions about clients' state, so it will
            // invoke stopEditinMode() anyway. Since we haven't actually
            // entered editing mode, we should ignore next stopEditingMode call.
            ignoreNextStopEditing = true;
            frame.setForceWipeScreen(saveCHO);
            frame.setCanHideOthers(saveCHO);
         }
      }

      if (!doIO)
      {
         // save insert indicator state in order to restore it after leaving
         // editing block
         savedInsStateInitials.push(savedInsState);
      }

      sb[0].resetChanged();

      ScreenBuffer[] result = null;

      // honor the outstanding next-prompt request if any
      int focusWidgetId = frame.getFocusId();
      frame.setFocusId(-1);

      focusWidgetId = handleOutstandingNextPrompt(focusWidgetId, widgetIds);

      // if no next-prompt, use the first widget from the list or the first
      // enabled widget of the frame
      if (focusWidgetId == -1)
      {
         if (widgetIds != null && widgetIds.length != 0)
         {
            // here the first focusable widget must be used, from all ids
            for (int i = 0; i < widgetIds.length; i++)
            {
               Widget widget = getWidget(widgetIds[i]);
               // get the actual field, in a down frame
               widget = frame.getField(widget);

               if (widget.focusTraversable())
               {
                  focusWidgetId = widgetIds[i];
                  break;
               }
            }
         }
         else
         {
            focusWidgetId = Frame.getFirstFocusableId(frame);
         }
      }

      // check if it is editing block preparation only
      if (!doIO)
      {
         // this call sets the editing mode flag so we leave it as is
         startEditingMode(focusWidgetId, el);
         
         // editing phrase needs to know about the frame and the focus...
         result = getEditableClientScreenBuffers(getScreenBuffer(frameId, -1, true));
         result[0].putKeyCode(lastKey);
         FrameData fd = frame.getFrameValue(null);
         result[0].setFrameValue(fd.getValue());
         result[0].setValueWidgetId(fd.getId());
         return result;
      }

      final Widget focus = UiUtils.getCurrentFocus();
      
      boolean condition = false;
      String status     = wnd.getStatusText();
      int type          = wnd.getStatusType();
      forceWipeScreen   = false;

      try
      {
         if (focus != null)
         {
            highlight(focus, false);
         }

         // if in trigger nesting mode, save previous focus
         if (triggerNesting > 0)
         {
            pendingFocus = focus;
         }

         BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.PROMPT_FOR);

         // old focus has to be undefined
         wnd.setFocus(null);

         if (!isChui())
         {
            // PROMPT-FOR statement in native 4GL causes the window to be moved to the foreground
            WindowManager.activateWindow(wnd);
         }

         result = waitFor(el, focusWidgetId, seconds, sb, BlockingOperation.PROMPT_FOR, true);
      }

      catch (ConditionException ce)
      {
         condition = true;

         throw ce;
      }

      finally
      {
         wnd.setStatusType(type);
         wnd.status(status);

         if (condition && focus != null)
         {
            // restore focus and highlight it, only if a condition was raised
            // focus is restored only if within a trigger
            if (triggerNesting > 0)
            {
               eventDrawingBracket(wnd, false, false, new Runnable()
               {
                  @Override
                  public void run()
                  {
                     focus.requestFocus();
                  }
               });
            }
            highlight(focus, true);
         }

         // reset the pending focus
         pendingFocus = null;

         wnd.hideMessage(false);
         // fix Z-order
         frame.moveToTop();
         // make sure we disable, even on an ERROR/END condition
         enable(frameId, widgetIds, null, false, windowId);

         restoreInsState(savedInsState, true);

         forceWipeScreen = true;

         // restore editing mode flag
         inEditingBlock = saveInEditingBlock;
      }

      return result;
   }

   /**
    * Refreshes the specified frame and widget with a new screen value.
    * <p>
    * The frame is not necessarily on top of Z-order and the widget is not
    * necessarily fully visible. This operation does not change the Z-order.
    *
    * @param   fid
    *          frame to refresh
    * @param   wid
    *          widget to refresh
    * @param   value
    *          new widget value
    * @return  string representation of the value
    */
   @Override
   public String refreshFrameWidget(int fid, int wid, BaseDataType value)
   {
      // must be called first!
      honorConditionalUpDown();

      Widget widget = getWidget(fid);

      // nothing to do?
      if (widget == null || !(widget instanceof Frame))
         return null;

      Frame frame = (Frame) widget;
      widget = getWidget(wid);

      if (widget == null)
         return null;

      // update the screen value
      WidgetConfig config = widget.config();
      setChangedValue(frame, widget, config, value, null);

      final Widget widgetToPaint = widget;

      // repaint the affected rectangle
      eventDrawingBracket(widget, new Runnable()
      {
         @Override
         public void run()
         {
            widgetToPaint.repaint();
         }
      });

      // return the string representation of the widget value
      FrameData dt = frame.getFrameValue(widget);

      if (dt != null && dt.getValue() != null)
         return dt.getValue();

      return "";
   }

   /**
    * This method is called on each start of an EDITING block iteration.
    */
   private void iterateEditingBlock()
   {
      Frame frame = editingFrames.peek();

      // Stack can be empty if there are no active widgets in the widget list
      // (see second parameter in BlockManager.updateEditing() methods) and
      // no frame was pushed in the stack (and no frames will be popped).
      if (frame == null)
         return;

      // honor the outstanding next-prompt request if any
      int focusWidgetId = frame.getFocusId();
      frame.setFocusId(-1);

      if (focusWidgetId != -1)
      {
         final Widget widget = getWidget(focusWidgetId);

         if (widget instanceof FillIn)
         {
            ((FillIn) widget).resetCaret();
         }

         int currentFocus = getFocus();

         if (focusWidgetId != currentFocus)
         {
            // change the focus; this action may trigger focus gained event
            // that would notify listeners and change the status line with
            // the correct help string
            eventBracket(new Runnable()
            {
               @Override
               public void run()
               {
                  widget.requestFocus();
               }
            });
         }
      }
   }

   /**
    * Externally accessible static wrapper for
    * {@link #viewWorker(Widget, ScreenBuffer[], int[], boolean, int)}.
    *
    * @param   widget
    *          The frame or widget for this request.
    * @param   frameBuf
    *          The current values of widgets in editable frames.
    * @param   widgetIds
    *          Array of field level widget IDs to be shown or
    *          <code>null</code>, which means show all contained widgets
    *          (for frame only).
    * @param   isView
    *          <code>true</code> if the call is for VIEW, <code>false</code>
    *          for DISPLAY.
    */
   public static void view(Widget         widget,
                           ScreenBuffer[] frameBuf,
                           int[]          widgetIds,
                           boolean        isView)
   {
      getInstance().viewWorker(widget, frameBuf, widgetIds, isView, WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Force the visibility and drawing of all specified fields of the
    * given frame.  This method may be safely called while processing in
    * the {@link #view} method.
    *
    * @param   widget
    *          The frame or widget for this request.
    * @param   frameBuf
    *          The current values of widgets in editable frames.
    * @param   widgetIds
    *          Array of field level widget IDs to be shown or
    *          <code>null</code>, which means show all contained
    *          widgets (for frame only).
    * @param   isView
    *          <code>true</code> if the call is for VIEW, <code>false</code>
    *          for DISPLAY.
    * @param   windowId
    *          target window id.
    */
   private void viewWorker(final Widget         widget,
                           final ScreenBuffer[] frameBuf,
                           final int[]          widgetIds,
                           final boolean        isView,
                                 int            windowId)
   {
      // mark the view operation in progress
      boolean saveViewInProgress = viewInProgress;
      viewInProgress = true;

      // inspect and reset forced view flags
      boolean forcedRefresh = nextViewRefresh;
      boolean forcedDisplay = nextViewDisplay;

      UIStatement stmt = (frameBuf == null || frameBuf[0] == null
                             ? null : frameBuf[0].getCurrentStatement());

      if (forcedRefresh)
      {
         forcedDisplay = false;
      }

      WidgetConfig c = widget.config();
      Frame wframe = (Frame) (widget instanceof Frame ? widget 
                                                       : getWidget(c.frameId));
      windowId = wframe.placeToWindow(windowId, stmt);

      final int _windowId = windowId;
      
      // check if the window is hidden; if so, make it visible;
      // when dialog the window points to the owner and not the 
      // window the frame is attached to
      // excluding the case when the frame is a part of the dialog-box
      if (!wframe.getRootFrame().isDialog())
      {
         final TopLevelWindow window = WindowManager.findWindow(windowId);
         if (UiUtils.showWindowForFrame(window, wframe, stmt))
         {
            independentEventDrawingBracket(window, new Runnable()
            {
               public void run()
               {
                  window.show();
               }
            });
         }
      }
      
      nextViewRefresh = false;
      nextViewDisplay = false;

      viewRefresh = forcedRefresh;
      viewDisplay = forcedDisplay;

      if (!(widget instanceof Frame))
      {
         eventDrawingBracket(widget, new Runnable()
         {
            @SuppressWarnings("null")
            @Override
            public void run()
            {
               // update our backing data to the latest version
               if (frameBuf != null)
               {
                  setScreenBuffers(frameBuf, true);
               }

               // handle widget view
               viewDisplay = true;         // always?
               widget.setVisible(true);

               // change the hidden flag for widget
               widget.setHidden(false);

               // ensure the frame is visible
               if (frameBuf != null && frameBuf[0] != null)
               {
                  Frame frame = (Frame) getWidget(frameBuf[0].getBaseId());
   
                  if (frame != null && !frame.isVisible())
                  {
                     viewWorker(frame, null, null, isView, _windowId);
                  }
               }
            }
         });

         viewInProgress = saveViewInProgress;

         // at this time, the frame's explicitHide flag is already reset by
         // the above viewWorker recursive call, so nothing more is
         // needed
         return;
      }

      // access the frame. TODO if the frame is (or is not) realized, the frame needs to be 
      // attached to the window (or the frame's window needs to be used).
      final Frame   frame         = (Frame) widget;
      final FrameConfig fconfig   = frame.config();
      final boolean down          = frame.isDown();
      final boolean toBeDisplayed = !frame.isVisible();
      final boolean redir         = isRedirected();
      final boolean editors       = frame.hasEditors();

      frame.setHonorNextConditionalDown(false);

      if (!frame.isVisible())
      {
         // if the frame is not visible when this explicit VIEW is called, it
         // must be marked.
         // if this VIEW call was executed by a trigger (called via an APPLY
         // statement), then focus will be set to the APPLY target widget, if
         // the focused widget's parent frame change its visibility from
         // not-visible to visible.
         frame.setVisibilityChanged(true);

         if (!redir || isView)
            lastExposedFrame  = frame;
      }

      if (redir && !isView)
      {
         frame.markAsStreamed();
      }

      lastViewed = frame;

      // create list of widgets to set visible
      Widget[] wids = null;

      if (widgetIds == null ||
          (widgetIds.length == 1 && getWidget(widgetIds[0]) == widget))
      {
         // this is a view frame call
         if (!forcedDisplay)
         {
            viewRefresh = true;
         }

         // all widgets
         wids = UiUtils.getVisibleComponents(frame, stmt);
      }
      else
      {
         // this is a data display call
         if (!forcedRefresh)
         {
            viewDisplay = true;
         }

         frame.setAllClear(false);

         HashSet<Widget> visibleComps = new HashSet<Widget>();

         // explicit widget list

         if (!frame.isVisible())
         {
            // collect all non-hidden widgets, if frame is not visible
            Widget[] comps = frame.getContentPane().widgets();

            for (int i = 0; i < comps.length; i++)
            {
               if (!comps[i].hidden())
               {
                  if (down)
                  {
                     comps[i] = frame.getField(comps[i]);
                  }

                  visibleComps.add(comps[i]);

                  if (comps[i] instanceof Frame)
                  {
                     visibleComps.addAll(
                        Arrays.asList(UiUtils.getVisibleComponents((Frame) comps[i], stmt)));
                  }
               }
            }

            // add the explicit components, regardless of their state
            for (int i = 0; i < widgetIds.length; i++)
            {
               Widget comp = getWidget(widgetIds[i]);
               if (down)
               {
                  comp = frame.getField(comp);
               }

               visibleComps.add(comp);

               if (comp instanceof Frame)
               {
                  visibleComps.addAll(
                     Arrays.asList(UiUtils.getVisibleComponents((Frame) comp, stmt)));
               }
            }

         }
         else
         {
            // frame is visible, use only explicit widgets or non-hidden, not-yet-realized widgets
            Widget[] comps = frame.getContentPane().widgets();

            for (int i = 0; i < comps.length; i++)
            {
               if (!comps[i].hidden() && !comps[i].isRealized())
               {
                  if (down)
                  {
                     comps[i] = frame.getField(comps[i]);
                  }

                  visibleComps.add(comps[i]);

                  if (comps[i] instanceof Frame)
                  {
                     visibleComps.addAll(
                        Arrays.asList(UiUtils.getVisibleComponents((Frame) comps[i], stmt)));
                  }
               }
            }
            
            for (int i = 0; i < widgetIds.length; i++)
            {
               Widget comp = getWidget(widgetIds[i]);

               if (down)
               {
                  comp = frame.getField(comp);
               }

               visibleComps.add(comp);

               if (comp instanceof Frame)
               {
                  visibleComps.addAll(
                     Arrays.asList(UiUtils.getVisibleComponents((Frame) comp, stmt)));
               }
            }
         }

         wids = new Widget[visibleComps.size()];
         visibleComps.toArray(wids);
      }

      final Widget[] list = wids;

      // handle frame view
      boolean copyRefresh = viewRefresh;
      boolean copyDisplay = viewDisplay;

      startWidgetOutput();

      boolean frameRepaint = false;

      // runtime exception saver
      final RuntimeException[] rtx = new RuntimeException[1];

      final Boolean[] allDisplayed = new Boolean[1];
      allDisplayed[0] = false;

      try
      {
         frameRepaint = eventDrawingBracket(frame, new Runnable()
         {
            @Override
            public void run()
            {
               if (frame.isForcedRefresh())
               {
                  frame.repaint();
                  frame.resetForcedRefresh();
               }

               if (frameBuf != null && frameBuf[0] != null && down && redir &&
                   (!frame.isSideLabels() || !frame.isLabels()))
               {
                  Set<Integer> iteration = frame.getIterationIds();

                  Iterator<Integer>  iter = frameBuf[0].getWidgetIDIterator();

                  while (iter.hasNext())
                  {
                     iteration.remove(iter.next());
                  }

                  allDisplayed[0] = (iteration.size() == 0);
               }

               if (frame.isNeedReset() && !down && frame.isRootFrame())
               {
                  try
                  {
                     if (frame.isVisible())
                     {
                        // force immediate frame hiding
                        if (!hideFramesInt(frame))
                        {
                           if (frame.isNeedPause())
                              pauseBeforeHide();
                        }
                        frame.setCanHide(false);
                     }
                  }

                  finally
                  {
                     frame.clearWorker(true);
                     frame.setNeedReset(false);
                  }
               }
               else if (down && frame.isForcePause())
               {
                  boolean normalExit = false;

                  try
                  {
                     frame.setForcePause(false);
                     pauseBeforeHide();
                     normalExit = true;
                  }

                  finally
                  {
                     if (!normalExit && !redir && frame.isRootFrame())
                     {
                        frame.clearWorker(true);
                        frame.setCanHideOthers(true);
                        hideFramesInt(frame);
                        frame.repaint();
                        frame.setNeedReset(false);
                        frame.setNeedPause(true);
                     }
                  }
               }

               if (frame.isDelayedRepaint() && frame.isVisible())
               {
                  frame.clearWorker(true);
               }

               if (frame.isCanHide() && frame.isVisible() && frame.isNeedPause())
               {
                  if (needPause)
                  {
                     pauseBeforeHide();
                  }
               }

               // update dynamic properties, such as headers, dynamic down, row, etc.

               ArrayList<Integer> updatedHeaders = null;

               if (frameBuf != null && frameBuf[0] != null)
               {
                  // adjust DOWN value
                  if (down)
                     frame.adjustDown(frameBuf[0].getDown());

                  frame.setDynRow(frameBuf[0].getRow());
                  frame.setDynCol(frameBuf[0].getColumn());

                  // update headers
                  updatedHeaders = updateHeaders(frameBuf[0]);
               }

               // display of a regular frame hides all dialogs, note that in GUI
               // dialogs must stay visible
               if (!frame.isDialog() && isChui())
               {
                  UiUtils.hideDialogs(targetWindowWorker(_windowId), extractFrames(waitForFrames));
               }

               if (!viewRefresh)
               {
                  viewRefresh = toBeDisplayed;
               }

               if (redir)
               {
                  RedirectedTerminal rt = redirections.get(currentStream);

                  if (rt.isFresh() && rt.needsHeadersOutput())
                  {
                     // TODO: in GUI and redirected mode, does this needs its own pseudo-window?
                     FramePlacementManager placementManager = 
                        WindowManager.getDefaultWindow().getPlacementManager();
                     placementManager.resetStartRow();
                     rt.processHeaders();
                  }
               }

               boolean bufferSet = false;

               // try set buffer early to be sure that
               // all widgets reach their real size
               if (frameBuf != null && !down)
               {
                  setScreenBuffers(frameBuf, !isRedirected());
                  bufferSet = true;
               }

               if (!frame.config().hidden || stmt != UIStatement.DISPLAY)
               {
                  placeFrameInt(frame, false);
               }

               if (down && redir)
               {
                  frame.arrangeToStream();
               }

               // assign values AFTER placing frame on the screen
               if (frameBuf != null)
               {
                  // update our backing data to the latest version (only force
                  // direct drawing/refresh if we are not redirected)
                  if (!bufferSet)
                     setScreenBuffers(frameBuf, !isRedirected());

                  if (redir && editors)
                  {
                     frame.rearrangeEditors();
                  }

                  // redraw updated headers now
                  if (!isRedirected() &&
                      updatedHeaders != null &&
                      !updatedHeaders.isEmpty())
                  {
                     for (Integer hid : updatedHeaders)
                        registry().getComponent(hid.intValue()).repaint();
                  }

                  needPause = true;
               }

               if (frame.isDelayedRepaint())
               {
                  frame.repaint();
                  frame.setDelayedRepaint(false);
               }

               if (toBeDisplayed || stmt == UIStatement.DISPLAY)
               {
                  for (int i = 0; i < list.length; i++)
                  {
                     if (list[i] != null)
                     {
                        boolean viewed = false;
                        if (list[i] instanceof Frame)
                        {
                           viewed = ((Frame) list[i]).isViewed();
                        }
                        
                        list[i].setVisible(true);
                        list[i].setHidden(false);
                        
                        if (list[i] instanceof Frame)
                        {
                           ((Frame) list[i]).setViewed(viewed);
                        }
                     }
                  }
               }
            }
         });
         
         if (!redir)
         {
            fconfig.assignedVisible = fconfig.visible;
         }
      }

      catch (RuntimeException xc)
      {
         rtx[0] = xc;
      }

      finally
      {
         if (down && redir)
         {
            frame.arrangeToInteract();
         }

         if (isRedirected() && frameRepaint && frame.height() == 0)
         {
            Stream s = sd.streamFromId(currentStream);

            if (s == null)
            {
               viewInProgress = saveViewInProgress;

               throw new IllegalStateException("Invalid stream id " +
                                               currentStream + ".");
            }

            s.pulseOutput();
         }

         if (rtx[0] == null)
         {
            // check if any of the widgets is a browse; if yes, hide any
            // temp texts inside message area
            for (int i = 0; i < list.length; i++)
            {
               if (list[i] instanceof Browse)
               {
                  // if found, clear the message area
                  targetWindowWorker(windowId).clearPutScreenInMessageArea(-1);
                  break;
               }
            }
         }

         if (toBeDisplayed || frameRepaint || allDisplayed[0])
         {
             viewRefresh = copyRefresh;
             viewDisplay = copyDisplay;

             if (redir && editors)
             {
                frame.restoreEditors();
             }

             if (rtx[0] == null)
             {
                viewInProgress = saveViewInProgress;
                return;
             }
         }

         // rethrow exception regardless from the mode
         if (rtx[0] != null)
         {
            viewInProgress = saveViewInProgress;
            throw rtx[0];
         }
      }

      // down frames receiving continuous output aren't redrawn by the code
      // above since they are visible before the call
      // this extra drawing cycle below solves this by explicitly repainting
      // the current iteration.

      if (redir && down)
      {
         eventDrawingBracket(frame, new Runnable()
         {
            @Override
            public void run()
            {
               frame.repaintCurrent();
            }
         });

         viewRefresh = copyRefresh;
         viewDisplay = copyDisplay;

         if (editors)
         {
            frame.restoreEditors();
         }

         viewInProgress = saveViewInProgress;
         return;
      }

      viewRefresh = copyRefresh;
      viewDisplay = copyDisplay;

      independentEventDrawingBracket(frame, new Runnable()
      {
         public void run()
         {
            for (int i = 0; i < list.length; i++)
            {
               if (list[i] != null)
               {
                  if (!list[i].isVisible())
                  {
                     final int idx = i;

                     eventBracket(true, new Runnable()
                     {
                        @Override
                        public void run()
                        {
                           list[idx].setVisible(true);

                           // change hidden flag too
                           list[idx].setHidden(false);
                        }
                     });
                  }
                  list[i].repaint();

                  // redraw side labels if any
                  if (down && frame.isSideLabels())
                  {
                     if (list[i] instanceof LabeledWidget)
                     {
                        Label label = ((LabeledWidget) list[i]).getLabelInstance();

                        if (label != null)
                           label.repaint();
                     }
                  }
               }
            }

            if (down)
            {
               frame.repaintUninitialized();
            }
         }
      });

      endWidgetOutput();

      if (redir && editors)
      {
         frame.restoreEditors();
      }

      // find the last enabled editor, in tab order, targeted by this display
      // call
      if (frameBuf != null && frameBuf[0] != null)
      {
         Set<Editor> edset = new HashSet<Editor>(3);
         for (int i = 0; i < list.length; i++)
         {
            if (list[i].isEnabled() && list[i] instanceof Editor)
            {
               // the editor must have at least some text, too
               Editor editor = (Editor) list[i];
               int eid = editor.config().id.asInt();
               if (frameBuf[0].getState(eid) == ScreenBuffer.CHANGED &&
                   editor.getText() != null                       &&
                   editor.getText().length() > 0)
               {
                  edset.add(editor);
               }
            }
         }

         // get all widgets of this frame and search for last targeted editor,
         // in tab order
         if (edset.size() > 0)
         {
            Widget[] widgets = frame.getContentPane().widgets();
            for (int i = widgets.length - 1; i >= 0; i--)
            {
               if (widgets[i] instanceof Editor && edset.contains(widgets[i]))
               {
                  // found the last editor, in tab order
                  lastEditorDisplayed = (Editor) widgets[i];
                  break;
               }
            }
         }
      }

      viewInProgress = saveViewInProgress;
   }

   /**
    * Mark end of widget output.
    */
   void endWidgetOutput()
   {
      OutputPrimitives output = tk.getInteractiveOutput();
      if (output instanceof DoubleBufferedTerminal)
         ((DoubleBufferedTerminal) output).conditionalRefresh();
   }

   /**
    * Mark start of widget drawing.
    */
   void startWidgetOutput()
   {
      OutputPrimitives output = tk.getInteractiveOutput();

      if (output instanceof DoubleBufferedTerminal)
         ((DoubleBufferedTerminal) output).reset();
   }

   /**
    * Conditionally refresh specified rectangle not occupied by frames.
    *
    * @param    rect
    *           Rectangle to refresh.
    */
   public void conditionalRefresh(Rectangle rect)
   {
      OutputPrimitives output = tk.getInteractiveOutput();

      if (output instanceof DoubleBufferedTerminal)
         ((DoubleBufferedTerminal) output).conditionalRefresh(rect);
   }

   /**
    * Conditionally refresh specified horizontal lice described by provided
    * rectangle.
    *
    * @param    rect
    *           Rectangle to refresh.
    */
   public void conditionalLineRefresh(Rectangle rect)
   {
      OutputPrimitives output = tk.getInteractiveOutput();

      if (output instanceof DoubleBufferedTerminal)
         ((DoubleBufferedTerminal) output).conditionalLineRefresh(rect);
   }

   /**
    * Hide frames if necessary to prepare space for specified frame and then
    * place and realize current frame.
    *
    * @param   frame
    *          Frame which will be displayed.
    * @param   force
    *          <code>true</code> means forced frame placement even if
    *          the frame is currently visible; useful for recovering partially
    *          hidden frames.
    */
   private void placeFrameInt(Frame frame, boolean force)
   {
      if (!frame.isRootFrame())
      {
         // TODO: this code was added to fix some cases in 
         // testcases/uast/frame-z-order/nested-frames.p, but the logic is not enough.
         // when placing the frame, we need to find which selection of UI statements brings
         // it to top
         /*
         if (!frame.isTopFrame())
         {
            frame.moveToTop();
            frame.repaint();
         }
         */

         frame.setVisible(true);
         frame.setHidden(false);
         return;
      }
      
      hideFramesInt(frame);

      if (frame.isVisible() && !isRedirected() && !frame.isRedirected())
      {
         Frame[] frames = UiUtils.getFramesAbove(frame);

         if (frames.length > 0)
         {
            force = true;
            // Code below has following effect: all frames above current
            // are not repainted, but invalidated rectangle is calculated
            // instead. Then current frame is then moved to top (see code
            // inside 'finally' block) and later calculated invalidated
            // rectangle is repainted.
            // Improved: only frames which overlap current are taken into
            // account.

            for (Frame frm : frames)
            {
               frm.repaint();
            }
         }
      }

      if (!frame.isVisible() || force)
      {
         if (isRedirected())
         {
            // set terminal logical width to let frame fit terminal
            double windowWidth = frame.window().getScreenDimension().width;
            int width          = redirectedConversion.widthToNative(windowWidth);
            double dFrameWidth  = frame.getSavedDimension().width;
            int frameWidth     = redirectedConversion.widthToNative(dFrameWidth);

            if (frame.location().x != BaseConfig.INV_COORD)
               frameWidth += frame.location().x;

            RedirectedTerminal rt = redirections.get(currentStream);

            rt.setLogicalWidth(Math.max(frameWidth, width));

            // relayout frame because it can be formatted for wrong width
            frame.relayout();
         }

         FramePlacementManager placementManager = frame.window().getPlacementManager();
         placementManager.placeFrame(frame);
         
         frame.setVisible(true);
         frame.setHidden(false);

         // do not fix frame's focus at this point, it will be eventually resolved during a
         // wait-for
      }

      if (!frame.window().config().keepFrameZOrder &&
          (editingFrames.isEmpty()            ||
           editingFrames.peekFirst() == frame ||
           frame.hasEnabled()                 ||
           frame.config().overlay))
      {
         frame.placeAtTop();
      }

      frame.setCanHide(false);
      frame.setCanWipeScreen(false);
      frame.setCanHideOthers(false);
      frame.setAlive(true);
      frame.setNeedPause(true);
      frame.placed();
   }

   /**
    * Internal method which hides frames if necessary and maintains global
    * flag.
    *
    * @param   frame
    *          Frame which is about to be displayed.
    *
    * @return  <code>true</code> if pause was triggered.
    */
   private boolean hideFramesInt(Frame frame)
   {
      FrameConfig cfg = frame.config();
      boolean[] res = new boolean[] {false, false};

      if ((!cfg.id.equals(WidgetId.INPUT_FRAME_ID) || helpActive) &&
          frame.isCanHideOthers())
      {
         res = UiUtils.hideFrames(frame, (currentStream != -1));
         frame.setCanHideOthers(false);

         // trigger screen refresh in following cases:
         // 1. Frame is explicitly marked as such that can wipe screen and
         //    there was at least one other frame hidden.
         // 2. Frame was hidden and has no enabled widgets and forceWipeScreen
         //    flag is set.
         // 3. There was at least one other frame hidden and among hidden
         //    frames there is a frame which has the same row as given one.
         
         Window wnd = frame.window();

         if ((!tk.isRedirected() && frame.isForceWipeScreen()) ||
             (!tk.isRedirected() && frame.isCanWipeScreen() && res[0]) ||
             (forceWipeScreen && !frame.isVisible() && !frame.hasAnyEnabled()) ||
             (res[0] == true && res[1] == true && messageClearQuirk))
         {
            wnd.refresh(new Rectangle(new Point(0, 0), 
                                      wnd.dimension(), 
                                      wnd.screen().coordinates().baseUnits()));

            if (res[0] == true && res[1] == true && messageClearQuirk)
            {
               // Replace messages with spaces.
               // Note that hiding of messages is not adequate, because
               // affects pausing.
               String[] msg = wnd.getMessageArea().getMessageTexts();
               Color[] cs = new Color[] {null, null};

               for (int i = 0; i < msg.length; i++)
               {
                  if (msg[i] != null)
                  {
                     msg[i] = StringHelper.repeatChar(' ', msg[i].length());
                     cs[i] = Color.NORMAL;
                  }
               }

               wnd.setMessageText(msg, cs);
            }

            if (frame.isForceWipeScreen())
            {
               wnd.setMessageText(null, null);
               wnd.setMessageNeedPause(false);
            }
         }

         forceWipeScreen = false;
         frame.setForceWipeScreen(false);
      }

      messageClearQuirk = false;

      return res[0];
   }

   /**
    * Perform operation on array of frames or widgets.
    *
    * @param   me
    *          <code>MarkEntry</code> container with operation and array of
    *          frames to process.
    */
   public void markFrames(MarkEntry me)
   {
      for (int i = 0; i < me.frames.length; i++)
      {
         if (me.frames[i]  == -1)
            continue;
         
         if (me.mode == MarkEntry.DESTROY_WIDGETS)
         {
            Widget<?> widget = getWidget(me.frames[i]);
            removeWidget(widget);
            continue;
         }

         Frame frame = (Frame) getWidget(me.frames[i]);

         if (frame == null)
            continue;

         switch(me.mode)
         {
            case MarkEntry.RESET_FRAME:
               frame.setNeedReset(true);
               break;

            case MarkEntry.DESTROY:
               destroyFrame(me.frames[i]);
               break;

            case MarkEntry.HIDE_NO_PAUSE:
               if (frame.isVisible())
               {
                  frame.setNeedPause(false);
                  frame.setCanHide(true);
               }
               break;

            case MarkEntry.CAN_HIDE_OTHERS:
            case MarkEntry.CAN_HIDE_OTHERS_EXT:
               {
                  frame.setCanHideOthers(true);

                  if (me.mode == MarkEntry.CAN_HIDE_OTHERS_EXT)
                     frame.setCanWipeScreen(true);
               }
               break;

            case MarkEntry.HIDE:
               if (frame.isVisible() && !frame.isNoHide())
               {
                  frame.setCanHide(true);
               }
               break;

            case MarkEntry.UNHIDE:
               if (frame.isVisible())
               {
                  frame.setCanHide(false);
               }
               break;

            case MarkEntry.OPEN_SCOPE:
               int scopeNesting = frame.getScopeNesting();
               frame.setScopeNesting(scopeNesting + 1);

               if (scopeNesting == 0)
               {
                  if (frame.isDown() && !frame.config().dynamic)
                  {
                     if (frame.getCursor() != -1 &&
                         frame.isNeedPause())
                     {
                        frame.setForcePause(true);
                     }

                     frame.setNeedReset(true);
                     frame.reset();
                  }
                  else
                  {
                     frame.resetAll();
                  }
                  frame.setInScope(true);
               }
               break;

            case MarkEntry.CLOSE_SCOPE:
               scopeNesting = frame.getScopeNesting() - 1;
               frame.setScopeNesting(scopeNesting);

               if (scopeNesting == 0)
               {
                  if (lastViewed == frame)
                     lastViewed = null;

                  // the repaint will be postponed until the next display/view
                  // statement is executed with this frame (when the frame is
                  // inside a loop)
                  // for shared frames, the repaint will be forced only if
                  // this is the root nested scope
                  frame.setDelayedRepaint(true);

                  if (frame.isDialog())
                  {
                     tk.setInvalidate(true);
                     frame.setVisible(false);
                     tk.setInvalidate(frame, false, false);
                  }
               }
               
               frame.setInScope(false);
               break;
               
            case MarkEntry.DELETED_SCOPE:
               forceDeregisterPageElement(me.frames[i]);
               break;
         }
      }
   }

   /**
    * Wait for specified list of events. <br>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    *
    * @param   list
    *          List of events to wait for.
    * @param   focusWidgetId
    *          Widget ID of a widget receiving focus first, or -1.
    * @param   seconds
    *          Time to wait for event.
    * @param   inbuf
    *          currently editable frames screen buffer known to the server.
    * 
    * @return  <code>ScreenBuffer</code> array with the current values.
    */
   @Override
   public ScreenBuffer[] waitFor(EventList      list,
                                 int            focusWidgetId,
                                 int            seconds,
                                 ScreenBuffer[] inbuf)
   {
      return waitFor(list, focusWidgetId, seconds, inbuf, BlockingOperation.WAIT_FOR, false);
   }

   /**
    * Wait for specified list of events. <br>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    *
    * @param   list
    *          List of events to wait for.
    * @param   focusWidgetId
    *          Widget ID of a widget receiving focus first, or -1.
    * @param   seconds
    *          Time to wait for event.
    * @param   inbuf
    *          currently editable frames screen buffer known to the server.
    * @param   op
    *          The type of blocking operation.
    * @param   force
    *          If <code>true</code>, on output, copy the current data into
    *          the returned screen buffer without regard to the changed
    *          flag.
    *          
    * @return  <code>ScreenBuffer</code> array with the current values.
    */
   public ScreenBuffer[] waitFor(EventList         list,
                                 int               focusWidgetId,
                                 int               seconds,
                                 ScreenBuffer[]    inbuf,
                                 BlockingOperation op,
                                 boolean           force)
   {
      // must be called first!
      honorConditionalUpDown(false);

      // save and reset editing mode flag for possible recursion
      boolean saveInEditingBlock = inEditingBlock;
      inEditingBlock = false;

      if (focusWidgetId != -1)
         inputActive++;

      ScreenBuffer[] result = null;
      try
      {
         // do the work
         result = waitForWorker(list, focusWidgetId, seconds, inbuf, force, true, true, op);
      }
      finally
      {
         // restore editing mode flag
         inEditingBlock = saveInEditingBlock;

         if (focusWidgetId != -1)
            inputActive--;
      }

      return result;
   }

   /**
    * Main WAIT-FOR worker routine.
    * <p>
    * It manages the focus scopes, sets the initial focus and provides a
    * message processing loop. The focus validity is constantly monitored
    * and when it needs adjustment and the conditions are appropriate, the
    * focus gets adjusted.
    *
    * @param    list
    *           List of events to wait for.
    * @param    focusWidgetId
    *           Widget ID of a widget receiving focus first, or -1.
    * @param    seconds
    *           The time to wait for a given key event.
    * @param    inbuf
    *           The currently editable frames screen buffer known to the server.
    * @param    force
    *           If <code>true</code>, on output, copy the current data into
    *           the returned screen buffer without regard to the changed
    *           flag.
    * @param    status
    *           If <code>true</code> method will change status line.
    * @param    entryWindow
    *           <code>true</code> if ENTRY for window should be sent.
    * @param    operation
    *           The {@link BlockingOperation operation} for which this event loop is executed.
    *           
    * @return   <code>ScreenBuffer</code> array with the current values.
    */
   public ScreenBuffer[] waitForWorker(EventList            list,
                                       int                  focusWidgetId,
                                       int                  seconds,
                                       final ScreenBuffer[] inbuf,
                                       boolean              force,
                                       boolean              status,
                                       boolean              entryWindow,
                                       BlockingOperation    operation)
   {
      return waitForWorker(list, focusWidgetId, seconds, inbuf, force, status, entryWindow, 
                           operation, true);
   }
   
   /**
    * Main WAIT-FOR worker routine.
    * <p>
    * It manages the focus scopes, sets the initial focus and provides a
    * message processing loop. The focus validity is constantly monitored
    * and when it needs adjustment and the conditions are appropriate, the
    * focus gets adjusted.
    *
    * @param    list
    *           List of events to wait for.
    * @param    focusWidgetId
    *           Widget ID of a widget receiving focus first, or -1.
    * @param    seconds
    *           The time to wait for a given key event.
    * @param    inbuf
    *           The currently editable frames screen buffer known to the server.
    * @param    force
    *           If <code>true</code>, on output, copy the current data into
    *           the returned screen buffer without regard to the changed
    *           flag.
    * @param    status
    *           If <code>true</code> method will change status line.
    * @param    entryWindow
    *           <code>true</code> if ENTRY for window should be sent.
    * @param    operation
    *           The {@link BlockingOperation operation} for which this event loop is executed.
    * @param    restoreFocus
    *           When <code>true</code> currenty focused widget will be kept before the event loop
    *           is started and eventually restored when condition is raised.
    *           When <code>false</code> the method will not attempt to restore focus.
    *           
    * @return   <code>ScreenBuffer</code> array with the current values.
    */
   public ScreenBuffer[] waitForWorker(EventList            list,
                                       int                  focusWidgetId,
                                       int                  seconds,
                                       final ScreenBuffer[] inbuf,
                                       boolean              force,
                                       boolean              status,
                                       boolean              entryWindow,
                                       BlockingOperation    operation,
                                       boolean              restoreFocus)
   {
      // save cursor state
      boolean curvis = tk.setCursorStatus(true);

      // save and check focus
      Widget saveFocus = UiUtils.getCurrentFocus();
      int focusState = FocusManager.focusState(saveFocus);

      // save pending focus - UPDATE operation may restore focus to it
      Widget savePending = pendingFocus;
      int pendingState = FocusManager.focusState(pendingFocus);

      // save the global events flag and the widget ids registered for
      // exit events
      boolean hasGlobalEvents = UiUtils.checkEventList(list);
      int[] exitWidgets = list.getExitWidgetIds();
      long[] exitResources = list.getExitResourceIds();

      // save off the previous value of go pending and reset to false
      // this should be done early enough to allow normal event processing
      // for LEAVE and ENTRY that may be generated on entry
      int oldGoPending = isGoPending;
      isGoPending = 0;
      immediateException = null;

      // determine the initial frame and focused widget
      Widget startWidget = null;
      boolean explicit = false;
      boolean pendingExplicit = false;

      if (focusWidgetId == -1     &&
          triggerNesting == 0     &&
          pendingFocus != null    &&
          pendingFocus.focusTraversable())
      {
         // if there is no explicit widget and we are on root level, then
         // force explicit widget to be the previous widget which had ENTRY
         // this is usable only for wait-for, as UPDATE always has explicit
         // focus
         Frame pendingFrame = (Frame) UiUtils.locateFrame(pendingFocus);
         if (pendingFrame != null && pendingFrame.isAlive())
         {
            // force explicit focus only if the frame is not out of scope
            pendingExplicit = true;
            focusWidgetId = UiUtils.lookupWidgetIdAsInt(pendingFocus);
         }
         else
         {
            pendingFocus = null;
         }
      }

      if (focusWidgetId != -1)
      {
         Widget target = getWidget(focusWidgetId);

         Frame sFrame = (Frame) UiUtils.locateFrame(saveFocus);

         // in some cases frame specified as a target may attract focus
         // and break regular focus handling rules
         if (target instanceof Frame && sFrame != null)
         {
            Frame tFrame = (Frame) target;

            if (inputActive > 0 &&
                sFrame.bounds().intersects(tFrame.bounds()) &&
                tFrame.isFocusAttractor())
            {
               Widget first = tFrame.getFirstEnabledWidget();

               if (first != null)
                  target = first;
            }

            // reset the flag
            tFrame.setFocusAttractor(false);
         }

         // ignore frame as focus target
         if (target == null || !(target instanceof Frame))
         {
            explicit   = true;
            startWidget = target;
         }
         else
         {
            focusWidgetId = -1;
         }
      }
      else
      {
         startWidget = saveFocus;
      }

      // set the screen buffer and refresh the visual appearance if needed
      tk.setInvalidate(true);

      eventBracket(new Runnable()
      {
         @Override
         public void run()
         {
            setScreenBuffers(inbuf);
         }
      });

      // save off old wait list and override with the specified list
      boolean saveLeaveSent = leaveSent;
      leaveSent = false;
      EventList save = currentEventList;
      if (operation == BlockingOperation.NO_BLOCKING_OPERATION)
      {
         currentEventList = new EventList().merge(list, false).merge(currentEventList, false);
      }
      else
      {   
         currentEventList = list;
      }

      // set the initial focus
      boolean frameAsTarget = true;
      boolean immedExit = false;
      int cop = currentEventList.getOperation();

      // a dialog box needs special attention
      boolean dialog = false;
      TopLevelWindow topLevelWnd = WindowManager.topLevelWindow();
      Frame framed = topLevelWnd == null ? null : topLevelWnd.getFirstVisibleFrame();

      if (framed != null && framed.isDialog())
      {
         dialog = true;
      }

      // row mode CHOOSE needs special attention
      boolean rowChoose = false;

      if (cop == EventList.CHOOSE)
      {
         Frame chFrame = (Frame) getWidget(inbuf[0].getBaseId());
         if (!chFrame.getChooseData().isFieldMode())
         {
            rowChoose = true;

            // no focus for row mode CHOOSE
            chFrame.setFocus(null);
         }
      }

      if (!rowChoose)
      {
         if (!explicit)
         {
            switch (focusState)
            {
               case FocusManager.FOCUS_VALID:
                  if (dialog)
                  {
                     // treat as undefined, implicit
                     immedExit = !focusManager.focusUndefImplicit(list, false);
                  }
                  else
                  {
                     // valid, implicit
                     immedExit = !focusManager.focusValidImplicit(saveFocus);
                  }
                  break;

               case FocusManager.FOCUS_UNDEF:
                  // undefined, implicit
                  immedExit = !focusManager.focusUndefImplicit(list, false);
                  break;

               case FocusManager.FOCUS_INVALID:
                  // invalid, implicit
                  immedExit = !focusManager.focusInvalidImplicit(saveFocus);
                  break;

               case FocusManager.FOCUS_HIDDEN:
                  // hidden, implicit
                  immedExit = !focusManager.focusUndefImplicit(list, true);
                  break;
            }
         }
         else
         {
            switch(focusState)
            {
               case FocusManager.FOCUS_VALID:
                  // valid, explicit
                  immedExit = !focusManager.focusValidExplicit(saveFocus,
                                                  startWidget,
                                                  list,
                                                  cop,
                                                  false);
                  break;

               case FocusManager.FOCUS_UNDEF:
                  // undefined, explicit
                  immedExit = !focusManager.focusUndefExplicit(startWidget,
                                                  list,
                                                  cop,
                                                  !pendingExplicit);
                  break;

               case FocusManager.FOCUS_INVALID:
                  // invalid, explicit
                  immedExit = !focusManager.focusInvalidExplicit(saveFocus, startWidget, list);
                  break;

               case FocusManager.FOCUS_HIDDEN:
                  // hidden, explicit
                  immedExit = !focusManager.focusValidExplicit(saveFocus,
                                                  startWidget,
                                                  list,
                                                  cop,
                                                  true);
                  break;
            }
         }
      }

      final Window wnd = WindowManager.resolveWindow();
      
      // immedExit is true if menubar.visible = false, i.e. no available widgets.
      if (immedExit && wnd.getMenubar() != null)
      {
         immedExit = false;
      }

      // TODO: The menu bar *must* be shown yet before WAIT-FOR is executed.
      //       See the simple test case:
      //       DEFINE SUB-MENU sb.
      //       DEFINE MENU m MENUBAR SUB-MENU sb.
      //       DEFAULT-WINDOW:MENUBAR = MENU m:HANDLE.
      //       PAUSE. /* menu bar is visible when the program pauses */
      //       WAIT-FOR CLOSE OF THIS-PROCEDURE.
      // If no enabled frames when MENUBAR is assigned - show the warning.
//      if (!isChui() && wnd.getMenubar() != null)
//      {
//         if(!hasEnabledFrame())
//         {
//            warnNoEnabledFrames(wnd);
//         }
//
//         wnd.getMenubar().setVisible(true);
//         wnd.doLayout();
//
//         eventDrawingBracket(wnd, new Runnable()
//         {
//            @Override
//            public void run()
//            {
//               wnd.repaint();
//            }
//         });
//      }
      
      // even if no focused widget, is possible to start the listening loop if there are 
      // registered resources
      if (immedExit && exitResources == null)
      {
         // close drawing bracket
         tk.setInvalidate(WindowManager.getCurrentWindow(), false, false);

         // immediate return
         currentEventList = save;
         leaveSent = saveLeaveSent;
         return inbuf;
      }

      // prepare to the event processing loop
      boolean canAdjustFocus = !focusManager.isNoFocusAdjustment();
      Widget nowFocused = UiUtils.getCurrentFocus();
      int focusStateNow = FocusManager.focusState(nowFocused);

      if (rowChoose)
      {
         canAdjustFocus = false;
         focusStateNow = FocusManager.FOCUS_UNDEF;
      }

      // if no focus, park cursor right after the insert mode indicator
      // and disallow frame as event target
      boolean statusSet = false;

      if (focusStateNow != FocusManager.FOCUS_VALID)
      {
         Window.setInsertMode(Window.INSERT_MODE_OFF, true);
         // if focus is invalid, then frame is used as target.
         // the flag will be adjusted on the next wait-for iteration.
         // when in row-choose mode, frame must be used as target so the
         // frame will accept events (even P2J doesn't set it in focus)
         if (!rowChoose &&
             (focusState == FocusManager.FOCUS_HIDDEN ||
              focusState == FocusManager.FOCUS_UNDEF))
         {
            frameAsTarget = false;
         }
      }
      // or else generate frame ENTRY event, only if previous focus is not
      // part of this frame
      else
      {
         Frame frame = (Frame)UiUtils.locateFrame(UiUtils.getCurrentFocus());
         if (saveFocus == null || frame != UiUtils.locateFrame(saveFocus))
         {
            if (frame != null)
               sendEntry(frame);

            if (triggerNesting == 0 && entryWindow)
            {
               // send entry to the window
               sendEntry(WindowManager.getDefaultWindow());
            }
         }

         // set status line if there is a focused widget
         if (status)
         {
            Widget widOnFocus = UiUtils.getCurrentFocus();
            
            startInputStatus((widOnFocus == null) || widOnFocus.forceCursorOnFocus());
            statusSet = true;
         }

         // highlight the currently focused widget
         highlight(UiUtils.getCurrentFocus(), true);
      }

      fixFocusedFrameZOrder();

      tk.setInvalidate(WindowManager.getCurrentWindow(), false, false);

      // clear GO pending flag just in case some event triggered it again
      isGoPending = 0;
      immediateException = null;

      // focus adjustments has been completed, so we can reset saved insert
      // indicator state
      clearSavedInsState();

      // if we have a WAIT-FOR PAUSE cycle and a fill-in receives the focus
      // upon initial focus adjustment, then auto-zap state becomes OFF for
      // this fill-in
      if (seconds >= 0)
      {
         Widget focus = UiUtils.getCurrentFocus();
         if (focus != null &&
             focus instanceof FillIn &&
             focus.isEnabled() &&
             focus.isVisible())
         {
            ((FillIn) focus).setPendingNoZap();
         }
      }

      triggerNestingByWaitFor.push(0);

      boolean saveCondExit = condExit;
      boolean condition = false;
      BlockingOperation saveOp = currentBlockingOp;
      currentBlockingOp = operation;
      
      // event processing loop
      try
      {
         boolean fixFocus = false;
         boolean saveFrameAsTarget = frameAsTarget;
         Event event = null;
         do
         {
            processPostponedValueChanged();
            Object source = null;
            
            try
            {
               waitForFrames.push(new WaitForOp(null, operation));

               BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.WAIT_FOR);

               // this will maintain LASTKEY too, which is required!
               event = waitForEvent(seconds, frameAsTarget, true, true);
               KeyInput key = (event instanceof KeyInput ? (KeyInput) event : null);
               condExit = false;
               
               if (event != null)
               {
                  source = (key == null ? null : key.source());

                  if (source instanceof Editor)
                  {
                     lastEditorDisplayed = (Editor) source;
                  }

                  Widget prevFocus = UiUtils.getCurrentFocus();
                  int prevState = FocusManager.focusState(prevFocus);

                  // for GUI dialog boxes the frame is inside widget, not outside
                  // otherwise we will have null as operation source
                  if (source instanceof DialogBoxWindow)
                  {
                     source = ((DialogBoxWindow)source).getFrame();
                  }
                  else if (source instanceof WindowGuiImpl && event instanceof MouseEvt)
                  {
                     source = UiUtils.locateFrame(((WindowGuiImpl)source).findMouseSource(
                                                                               (MouseEvt)event));
                     Frame<?> rootFrame = source != null ? ((Frame)source).getRootFrame() : null;
                     if (rootFrame != null)
                     {
                        source = rootFrame;
                     }
                  }
                  else if (source instanceof Widget)
                  {
                     source = UiUtils.locateFrame((Widget) source);
                  }

                  // replace the frame with the current frame
                  waitForFrames.pop();
                  waitForFrames.push(new WaitForOp((Frame) source, operation));

                  /* TODO set correct window */
                  setNeedPause(false, WindowManager.resolveWindow());

                  int action = (key == null ? KeyInput.CHAR_UNDEFINED : key.actionCode());
                  Widget src = (key == null ? null : key.source());

                  if ((action == Keyboard.KA_STOP   ||
                      action == Keyboard.KA_ENDKEY ||
                      action == Keyboard.KA_ERROR  ||
                      action == Keyboard.KA_END_ERROR) &&
                     checkBreak(src, action, false) &&
                     checkBreak(src, Keyboard.SE_LEAVE, false))
                  {
                     // The original event constitutes an exceptional exit
                     // condition which is used in the wait-for list as
                     // a regular exit condition and the LEAVE event is also
                     // specified in the wait list. No regular event
                     // application is taken (no call to applyWorker()), but
                     // a special event processing is performed.

                     Widget cur = prevFocus;
                     if (cur != null && !cur.isEnabled())
                     {
                        cur = null;
                     }

                     if (src instanceof Frame || src instanceof Window)
                     {
                        cur = src;
                     }

                     // Immediately consume the original event.
                     key.consume();

                     // Generate and apply LEAVE event.
                     if (cur != null)
                     {
                        if (sendLeave(cur))
                        {
                           // Exit the processing loop.
                           isGoPending++;
                        }
                     }
                  }
                  else
                  {
                     try
                     {
                        // event application
                        applyWorker(event);
                     }

                     finally
                     {
                        if (source != null)
                        {
                           Frame frame = (Frame) source;

                           if (frame.isRealized())
                           {
                              // TODO: is this needed for GUI dialog?
                              // if the frame was not hidden via an explicit
                              // HIDE call, re-display it
                              if ((isChui() || !frame.isDialog()) && 
                                  !frame.isVisible() && !frame.isExplicitHide())
                              {
                                 // restore focus
                                 if (prevFocus != null)
                                 {
                                    prevFocus.requestFocus();
                                    prevFocus.setHighlighted(true);
                                 }
   
                                 // repaint this frame without the hidden widgets
                                 view(frame.config().id.asInt(), inbuf, null, true,
                                      WindowConfig.RESOLVE_WINDOW);
                              }
   
                              frame.setExplicitHide(false);
                           }
                        }
                     }
                  }

                  // determine the state of the focus after event application
                  // if either undefined or invalid, focus needs adjustment
                  Widget focus = UiUtils.getCurrentFocus();
                  int state = FocusManager.focusState(focus);
                  fixFocus = state != FocusManager.FOCUS_VALID &&
                             (focus != prevFocus ||
                              state != prevState ||
                              state == FocusManager.FOCUS_UNDEF);
                  // restore flag to value as before wait-for loop, only if
                  // there is a focused widget
                  if (state != FocusManager.FOCUS_UNDEF)
                  {
                     frameAsTarget = saveFrameAsTarget;
                  }

                  // in case a GO event was generated by this key, then no
                  // focus adjustment is needed
                  if (isGoPending == 0 && canAdjustFocus && fixFocus)
                  {
                     focusManager.adjustFocus();
                     // if no focus, then frame as target must not be used
                     frameAsTarget = UiUtils.getCurrentFocus() != null;
                  }

                  // if no GO was generated
                  if (isGoPending == 0 && !hasGlobalEvents)
                  {
                     // if the wait-for is not waiting for window events, check if at least one 
                     // widget is still enabled and visible
                     if (exitWidgets != null)
                     {
                        boolean noWidgets = true;
                        for (int ew : exitWidgets)
                        {
                           Widget widget = getWidget(ew);
                           if (widget == null)
                           {
                              // the widget can be already destroyed
                              // e.g. ComboBoxGuiImpl deactivated on SET-WAIT-STATE
                              noWidgets = ew != (focusWidgetId);
                              continue;
                           }
                           boolean visible = widget.isVisible();
                           
                           // WAIT-FOR must be able to wait for MENU-ITEM in a closed SUB-MENU.
                           if (widget instanceof MenuItem || widget instanceof SubMenu)
                              visible = true;
                           
                           if (widget != null && widget.isEnabled() && visible)
                           {
                              noWidgets = false;
                              break;
                           }
                        }

                        if (noWidgets)
                        {
                           // if no enabled or visible widgets exist, then
                           // processing must be stopped
                           stopNoAvailableWidgets();
                        }
                     }

                     // check resource states too
                     if (exitResources != null)
                     {
                        boolean noResources = true;
                        for (int i = 0; i < exitResources.length; i++)
                        {
                           if (server.validResource(exitResources[i]))
                           {
                              noResources = false;
                              break;
                           }
                        }

                        if (noResources)
                        {
                           // if no valid and sensitive resources exit, stop processing
                           stopNoAvailableWidgets();
                        }
                     }
                  }

                  // a special case for read-only editor widget
                  focus = UiUtils.getCurrentFocus();
                  if (lastEditorDisplayed != null &&
                      focus instanceof Editor     &&
                      ((Editor) focus).config().readOnly)
                  {
                     // if focus is a read-only editor, then check the last
                     // editor which had buffer update. if found, set
                     // cursor on it
                     lastEditorDisplayed.drawCaretUnconditionally();

                     if (isChui())
                     {
                        ((ChuiOutputManager) tk).stayAt(((ChuiOutputManager) tk).cursor());
                     }

                     // do not re-draw caret on next key wait
                     skipCaretDraw = true;
                  }
                  // clear the saved editor
                  lastEditorDisplayed = null;
               }
            }

            finally
            {
               waitForFrames.pop();

               fixFocusedFrameZOrder();
            }
         }
         while (isGoPending == 0 && event != null);

         // Usually when condition is thrown we just don't reach this point
         // and down't send LEAVE and this is consistent with 4GL. But in case
         // when condition was intercepted by trigger and it exits with
         // NO-APPLY then we also should not send LEAVE to be consistent with
         // 4GL behavior.
         // if no key was pressed, it means the wait-for timed-out; so, no
         // leave event will be sent.
         if (!condExit && event != null && (event instanceof KeyInput))
         {
            // generate LEAVE event for the last focused widget
            if (!leaveSent && cop != EventList.CHOOSE)
            {
               currentEventList = list;
               sendLeave(UiUtils.getCurrentFocus());
               currentEventList = save;
            }
         }
         else
         {
            // remove the highlight from the widget
            highlight(UiUtils.getCurrentFocus(), false);
         }
      }

      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate(WindowManager.getDefaultWindow());

         // set condition flag
         condition = true;

         // let the original exception propagate
         throw ce;
      }

      finally
      {
         // cleanup
         currentEventList = save;
         isGoPending      = oldGoPending;
         leaveSent        = saveLeaveSent;
         condExit         = saveCondExit;
         immediateException = null;
         currentBlockingOp   = saveOp;
         
         Widget focused = UiUtils.getCurrentFocus();
         if (focused != null && condition && isChui())
         {
            Frame frame = (Frame) UiUtils.locateFrame(focused);
            if (frame != null)
            {
               if (focused.isVisible() && frame.isVisible())
               {
                  highlight(focused, false);
               }
               else
               {
                  focused.setHighlighted(false);
               }
            }
         }

         triggerNestingByWaitFor.pop();

         if (statusSet)
         {
            stopInputStatus(false);
         }

         /* TODO set correct window */
         UiUtils.resetNeedPause(WindowManager.resolveWindow());

         // leave choose processing is done here because it needs to occur
         // *before* the getScreenBuffer() below, that is why it is here
         // instead of in the more logical place in choose()
         if (list.getOperation() == EventList.CHOOSE)
         {
            final Frame frame = (Frame) getWidget(inbuf[0].getBaseId());

            if (frame != null)
            {
               // leaveChoose should not produce any useful events
               eventBracket(true, new Runnable()
               {
                  @Override
                  public void run()
                  {
                     frame.leaveChoose();
                  }
               });
            }
         }

         tk.setCursorStatus(curvis);

         // set this trigger as it had wait-for
         triggerHadWaitFor.pop();
         triggerHadWaitFor.push(true);
         
         if (restoreFocus && condition && saveFocus != null && saveFocus.ancestor() != null)
         {
            fixWaitForFocus(cop, saveFocus, savePending, focusState, pendingState);
         }
      }

      // grab screen buffer before focus will be restored, because
      // focus state is relevant for FRAME-FIELD/FRAME-VALUE
      ScreenBuffer[] result = null;

      // where there is an WAIT-FOR ... ANYWHERE , there might be no active
      // frame
      if (inbuf != null)
      {
         result = getEditableClientScreenBuffers(
                     getScreenBuffer(inbuf[0].getBaseId(), inbuf[0].size(), force));
      }

      if (restoreFocus)
      {
         // restore focus in the previous scope, if inside an editing block or
         // inside a trigger
         if (triggerNesting == 0 && !inEditingBlock)
         {
            // always set focus to undefined, if on root nesting level
            WindowManager.getDefaultWindow().setFocus(null);
         }
         else
         {
            fixWaitForFocus(cop, saveFocus, savePending, focusState, pendingState);
         }
      }
      
      // return the final screen buffer
      return result;
   }
   
   /**
    * Fix the focus after a event loop has finished, regradless if a condition was raised or not.
    * 
    * @param    cop
    *           The current operation.
    * @param    saveFocus
    *           The focus before the event loop started.
    * @param    savePending
    *           The pending focus set via {@link #pendingFocus}, before the event loop started.
    * @param    focusState
    *           The focus state for the <code>saveFocus</code>.
    * @param    pendingState
    *           The focus state for the <code>savePending</code>.
    */
   private void fixWaitForFocus(int       cop, 
                                Widget<?> saveFocus, 
                                Widget<?> savePending, 
                                int       focusState, 
                                int       pendingState)
   {
      // for UPDATE operation, focus must be restored to previous focus -
      // get it from the saved pending focus
      if (triggerNesting > 0 &&
          (cop == EventList.UPDATE ||
           cop == EventList.SET    ||
           cop == EventList.PROMPT_FOR))
      {
         focusState = pendingState;
         saveFocus = savePending;
      }

      switch(focusState)
      {
         case FocusManager.FOCUS_UNDEF:
            if (cop != EventList.WAIT_FOR)
            {
               // focus has to be undefined
               WindowManager.getDefaultWindow().setFocus(null);
            }
            break;

         case FocusManager.FOCUS_VALID:
         case FocusManager.FOCUS_INVALID:
            // focus has to be restored no matter enabled or not, and if
            // a fill-in receives it, we should use the current insert
            // mode for it rather than a stored one
            if (saveFocus != null && saveFocus instanceof FillIn)
            {
               ((FillIn) saveFocus).setInsertMode(
                  Window.getInsertMode() == Window.INSERT_MODE_INS);
            }
            focusManager.setFocusOn(saveFocus);
            
            // saveFocus may be located in an inactive top-level window
            WindowManager.activateWindow(saveFocus.topLevelWindow());
      }
   }

   /**
    * Fix possible Z-ordering issues.
    */
   private void fixFocusedFrameZOrder()
   {
      if (UiUtils.getCurrentFocus() == null)
         return;

      final Frame frame = (Frame) UiUtils.locateFrame(UiUtils.getCurrentFocus());

      if (frame == null || !frame.hasEnabled() || frame.isDialog() || !frame.isRootFrame())
         return;

      final boolean keepFrameZOrder = !isChui()              &&
                                      frame.window() != null && 
                                      frame.window().config().keepFrameZOrder;
      Frame[] flist = UiUtils.getFramesAbove(frame);

      // check for empty frame list
      if (flist == null || flist.length == 0)
      {
         return;
      }
      
      for (Frame ovl : flist)
      {
         FrameConfig config = ovl.config();

         if (!config.overlay)
            continue;

         eventDrawingBracket(ovl, false, false, new Runnable()
         {
            @Override
            public void run()
            {
               if (!keepFrameZOrder)
               {
                  frame.placeAtTop();
               }
               frame.repaint();
            }
         });

      }
   }

   /**
    * Sends ENTRY event to the specified widget and returns a boolean result
    * of the possible trigger invocation.
    *
    * @param   widget
    *          target widget for sending ENTRY event
    * @return  <code>true</code> if no trigger or the trigger has returned
    *                            normally<br>
    *          <code>false</code> if the trigger has returned no-apply
    */
   public boolean sendEntry(Widget widget)
   {
      return sendEntry(widget, false);
   }
   
   /**
    * Sends ENTRY event to the specified widget and returns a boolean result
    * of the possible trigger invocation.
    *
    * @param   widget
    *          target widget for sending ENTRY event
    * @param   mouse
    *          Flag indicating this ENTRY is posted from a mouse event.  In this case, drawing is
    *          bypassed.
    *          
    * @return  <code>true</code> if no trigger or the trigger has returned
    *                            normally<br>
    *          <code>false</code> if the trigger has returned no-apply
    */
   public boolean sendEntry(Widget widget, boolean mouse)
   {
      // create an ENTRY event which does not require matching LEAVE
      KeyInput event = new KeyInput(widget, EventType.KEY_TYPED,
                                    Keyboard.SE_ENTRY, true);

      tk.setInvalidate(true);
      try
      {
         applyWorker(event);
      }
      finally
      {
         tk.setInvalidate(widget, false, !mouse);
      }

      return triggerResultNoApply.peek() == false;
   }

   /**
    * Sends LEAVE event to the specified widget.
    *
    * @param   widget
    *          target widget for sending LEAVE event
    *
    * @return  <code>true</code> if no trigger or the trigger has returned
    *                            normally<br>
    *          <code>false</code> if the trigger has returned no-apply
    */
   public boolean sendLeave(Widget widget)
   {
      return sendLeave(widget, false);
   }
   
   /**
    * Sends LEAVE event to the specified widget.
    *
    * @param   widget
    *          target widget for sending LEAVE event
    * @param   mouse
    *          Flag indicating this LEAVE is posted from a mouse event.  In this case, drawing is
    *          bypassed.
    *          
    * @return  <code>true</code> if no trigger or the trigger has returned
    *                            normally<br>
    *          <code>false</code> if the trigger has returned no-apply
    */
   public boolean sendLeave(Widget widget, boolean mouse)
   {
      leaveSent=true;

      if (widget == null || (widget == WindowManager.getDefaultWindow() && isChui()))
      {
         // no leave is sent if no widget or the widget is the default window
         return true;
      }

      KeyInput event = new KeyInput(widget, EventType.KEY_TYPED,
                                    Keyboard.SE_LEAVE);

      if (isChui())
      {
         tk.setInvalidate(true);
         try
         {
            applyWorker(event);
         }
         finally
         {
            tk.setInvalidate(widget, false, !mouse);
         }
      }
      else
      {
         independentEventDrawingBracket(widget, new Runnable()
         {
            public void run()
            {
               EventManager.postEvent(event);
            }
         });
      }

      if (widget instanceof FillIn)
      {
         // fill-in's always reset the cursor on the first position, after the
         // widget is left
         ((FillIn) widget).setCursorOffset(0);
      }

      return !event.isConsumed();
   }

   /**
    * Check if trigger processing was finished with NO-APPLY and there was no
    * WAIT-FOR during trigger processing.
    *
    * @return  <code>true</code> if both conditions are met.
    */
   public boolean isTriggerWasAppliedWithoutWaitFor()
   {
      return !triggerResultNoApply.peek() && !triggerHadWaitFor.peek();
   }

   /**
    * Handle postponed VALUE-CHANGED if any.
    */
   private void processPostponedValueChanged()
   {
      while (postponedValueChanged != null)
      {
         KeyInput key = postponedValueChanged;
         postponedValueChanged = null;

         Widget frame = UiUtils.locateFrame(key.source());

         if (frame != null && frame.parent() != null)
            applyWorker(key);
      }
   }

   /**
    * Activate input mode.
    *
    * @param    cursor
    *           New state of the cursor. <code>true</code> for the cursor
    *           to be forced into visibility and <code>false</code> to hide
    *           the cursor.
    */
   private void startInputStatus(boolean cursor)
   {
      // force the new cursor mode and save the original mode
      oldcursor.push(new Boolean(tk.setCursorStatus(cursor)));
      inputActive++;

      // TODO/WARNING resolved window might be different than in stopInputStatus method.
      Window wnd = WindowManager.resolveWindow();
      
      wnd.setStatusType(Window.ST_HELP);
      wnd.status(wnd.getHelpText());
   }

   /**
    * Deactivate input mode.
    */
   private void stopInputStatus()
   {
      stopInputStatus(true);
   }

   /**
    * Deactivate input mode.
    *
    * @param   changeStatus
    *          Allow changing status line.
    */
   private void stopInputStatus(boolean changeStatus)
   {
      // reset cursor to previous state
      tk.setCursorStatus(oldcursor.pop().booleanValue());

      inputActive--;

      if (inputActive == 0 && changeStatus)
      {
         // reset (blank) the status line
         // TODO/WARNING resolved window might be different than in startInputStatus method.
         WindowManager.resolveWindow().status(null);
      }
   }

   /**
    * Signals a trigger to the application.  If a trigger is registered for
    * this event and source widget combination, then it will be executed.
    *
    * @param    triggerId
    *           The trigger to be called.
    * @param    eventId
    *           The event which was generated.
    * @param    sourceId
    *           The widget which generated the event.
    * @param    otherId
    *           The companion widget id (for ENTER/LEAVE).
    * @param    resourceId
    *           The resource which generated the event.
    * @param    sb
    *           The associated frame's current <code>ScreenBuffer</code>
    *
    * @return   <code>true</code> if the event must be consumed (this is
    *           equivalent to <code>RETURN NO-APPLY</code>.
    */
   private boolean trigger(int          triggerId,
                           int          eventId,
                           int          sourceId,
                           int          otherId,
                           long         resourceId,
                           ScreenBuffer sb,
                           EventList    previousEventList)
   {
      boolean saveLeaveSent = leaveSent;
      int saveGoPending = isGoPending;
      Widget savePending = pendingFocus;
      leaveSent = false;
      pendingFocus = null;

      // a special case for EDITOR - clear saved editor
      lastEditorDisplayed = null;

      // save the frame for the current focus
      Widget focus = UiUtils.getCurrentFocus();
      Frame frame = null;
      boolean saveVisibilityChanged = false;
      boolean saveHighlighted = false;

      if (focus != null)
      {
         frame = (Frame) UiUtils.locateFrame(focus);
         // save visibility-changed flag, so it can be restored after trigger invocation
         if (frame != null)
         {
            saveVisibilityChanged = frame.isVisibilityChanged();
            frame.setVisibilityChanged(false);
            saveHighlighted = focus.highlighted();
         }
      }

      int savedInsState = saveInsState();

      int saveLevel = tk.saveInvalidate(getWidget(sourceId));

      int val = triggerNestingByWaitFor.pop();
      triggerNestingByWaitFor.push(val + 1);

      setTriggerNesting(triggerNesting + 1);

      ServerReturnValue srv = null;
      TriggerReturnValue trv = null;

      try
      {
         srv = server.trigger(triggerId, eventId, sourceId, otherId, resourceId,
                              getEditableClientScreenBuffers(sb));
         trv = srv == null ? null : (TriggerReturnValue) srv.payload;
      }
      finally
      {
         setTriggerNesting(triggerNesting - 1);

         val = triggerNestingByWaitFor.pop();
         triggerNestingByWaitFor.push(val - 1);

         tk.restoreInvalidate(saveLevel);
         pendingFocus = savePending;

         if (trv != null)
         {
            if (!trv.consume)
            {
               restoreInsState(savedInsState, false);
   
               Widget current = UiUtils.getCurrentFocus();
               
               // fix for the case when trigger starts and ends up into the same
               // fill-in
               if (focus != null)
               {
                  if (focus == current && focus instanceof FillIn)
                  {
                     ((FillIn) focus).setInsertMode(savedInsState == Window.INSERT_MODE_INS);
                     clearSavedInsState();
                  }
                  else if (focus != current        && 
                           focus instanceof Editor && 
                           (eventId == Keyboard.KA_RETURN || eventId == 13))
                  {
                     // in case a RETURN/ENTER trigger moves focus from the editor, consume the key
                     // any other key is still applied
                     trv.consume = true;
                  }
               }
            }
            
            if (previousEventList != null)
            {
               // remove deleted triggers
               if (srv.deadTriggers != null)
               {
                  previousEventList.remove(srv.deadTriggers);
               }
               
               // merge new triggers
               if (srv.newTriggers != null)
               {
                  previousEventList.merge(srv.newTriggers, true);
               }
            }
         }
         else
         {
            // trigger returned NO-APPLY, leave current insert mode
            restoreInsState(-1, false);
         }

         if (isGoPending == saveGoPending)
         {
            // restore flag only if no GO was generated during trigger - this
            // will block duplicate LEAVE events
            leaveSent = saveLeaveSent;
         }

         // check if focused frame changed its visibility (from not-visible to
         // visible)
         if (frame != null)
         {
            // move focus to the targeted widget
            if (frame.isVisibilityChanged() && !frame.isFocusReset())
            {
               // old focus loses highlight
               if (focus != null)
                  focus.setHighlighted(false);

               Widget comp = getWidget(sourceId);
               // the target widget receives focus
               comp.requestFocus();

               highlight(comp, saveHighlighted);
            }

            // restore the flag
            frame.setVisibilityChanged(saveVisibilityChanged);
         }
      }

      if (trv != null)
      {
         if (trv.buffer != null)
         {
            // handle screen buffers array taken from server
            setScreenBuffers(trv.buffer, true);
         }
         
         return trv.consume;
      }

      return false;
   }

   /**
    * Signals a validation event to the application.
    * <p>
    * This signal is used to trigger validation execution on the server.
    *
    * @param   sourceId
    *          widget ID of the widget generated the trigger.
    * @param   sb
    *          current <code>ScreenBuffer</code>
    *
    * @return  result of validation check.
    *
    */
   public boolean validate(int sourceId, ScreenBuffer sb)
   {
      ValidationResult vr = server.validate(sourceId, getEditableClientScreenBuffers(sb));
      // handle screen buffers array taken from server
      setScreenBuffers(vr.getScreenBuffers());

      boolean res = vr.isValidated();
      if (res == false)
      {
         // retrieve validation error message
         validationError = vr.getError();
      }

      return res;
   }

   /**
    * Prepares the array of the screen buffers to pass to the server when calling the server
    * export methods to be able the screen buffer synchronization.
    *
    * @param   sbDefault
    *          The default screen buffer for the exact frame in processing.
    *
    * @return  The array of the screen buffers to sync.
    *
    */
   private ScreenBuffer[] getEditableClientScreenBuffers(ScreenBuffer sbDefault)
   {
      ArrayList<ScreenBuffer> sbArrayList = new ArrayList<ScreenBuffer>();
      // add default one if not null as the first entry
      if (sbDefault != null)
      {
         sbArrayList.add(sbDefault);
      }
      // check the registry for more current editable frames to make additional screen buffers
      Frame<?>[] frList = registry().getEditableFrames();
      for (Frame<?> nextFrame : frList)
      {
         if (nextFrame != null &&
             (sbDefault == null || nextFrame.getId().asInt() != sbDefault.getBaseId()))
         {
            // create screen buffer to the given frame
            ScreenBuffer moreScreenBuff = getScreenBuffer(nextFrame.getId().asInt(), -1, false);
            // and add it to the list for further processing
            if (moreScreenBuff != null)
            {
               sbArrayList.add(moreScreenBuff);
            }
         }
      }
      // convert array list to array and return to the caller
      return sbArrayList.isEmpty() ? null
                                   : sbArrayList.toArray(new ScreenBuffer[sbArrayList.size()]);
   }

   /**
    * Returns the last validation error message.
    * <p>
    *
    * @return  error message generated during the last validation call or
    *          <code>null</code> if the last validation succeeded.
    */
   public String getValidationError()
   {
      return validationError;
   }

   /**
    * Get configurations of fill-ins used for in-browse editing.
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    *
    * @return   Array of configurations of fill-in widgets which are used for in-browse editing.
    *           An array element is configuration of the fill-in widget for the corresponding
    *           editable column or <code>null</code> if the column is not editable. Return value
    *           is <code>null</code> if the browse is read-only.
    */
   public FillInConfig[] getEditingFillIns(int browseId)
   {
      return server.getEditingFillIns(browseId);
   }

   /**
    * Get the number of rows iterated by the query (for preselect queries - matches result size,
    * for adaptive queries - matches the size of the part of the result set the query has already
    * iterated over).
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *
    * @return  number of iterated rows in the result set.
    */
   public int getRowCount(int browseId)
   {
      return server.getRowCount(browseId);
   }

   /**
    * Get number of available rows starting from the given one.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *
    * @param   startRow
    *          Initial row to start count rows from.
    *
    * @return  Number of rows in browse.
    */
   public int getRowCount(int browseId, int startRow)
   {
      return server.getRowCount(browseId, startRow);
   }

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    * @param    startRowIndex
    *           Index of the first row to retrieve.
    * @param    count
    *           Maximum count of rows to return.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction,
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query specified, an array of size 0 is returned.
    */
   public BaseDataType[][] getRows(int browseId, int startRowIndex, int count, boolean forward)
   {
      return server.getRows(browseId, startRowIndex, count, forward);
   }

   /**
    * Updates selected cells in the given row.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * @param   onlyNotifyRowReleased
    *          Only notify server that the row has been released. Data is not updated.
    *
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   public void updateRow(int            browseId,
                         int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues,
                         boolean        onlyNotifyRowReleased)
   throws DataValidationException
   {
      try
      {
         server.updateRow(browseId, rowIndex, cellIndices, cellValues, onlyNotifyRowReleased);
      }
      catch (DataValidationException e)
      {
         validationError = e.getMessage();
         throw e;
      }
      cleanupQueue();
   }

   /**
    * Inform server about current row in the browse.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          0-based index of the current row.
    *
    * @return  New current row.
    */
   public int setCurrentRow(int browseId, int rowIndex)
   {
      return server.setCurrentRow(browseId, rowIndex);
   }

   /**
    * Get current row in the query.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *
    * @return  The current row index.
    */
   public int getCurrentRow(int browseId)
   {
      return server.getCurrentRow(browseId);
   }

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   public boolean isRowAvailable(int browseId, int rowIndex)
   {
      return server.isRowAvailable(browseId, rowIndex);
   }

   /**
    * Checks to see whether a pending GO event is about to end the editing block.
    *
    * @param    hadNext
    *           <code>true</code> if the executing business logic executed a <code>NEXT</code>
    *           statement causing the editing block to iterate. This will result in clearing
    *           any current <code>GO-PENDING</code> condition.  This is passed as a parameter
    *           here so that a separate trip down to the client is not needed to process the
    *           editing block iteration state update.
    *
    * @return   <code>true</code> if there is no <code>GO-PENDING</code> and the editing mode
    *           continues.
    */
   @Override
   public boolean continueEditing(boolean hadNext)
   {
      putCursor(true);
      
      boolean stop = false;
      
      if (hadNext)
      {
         // reset GO-PENDING state
         isGoPending = 0;
      }
      else
      {
         stop = isGoPending > 0;
      }

      if (stop)
      {
         stopEditingMode();
      }
      else
      {
         iterateEditingBlock();
      }

      return !stop;
   }

   /**
    * Marks the beginning of the editing block and passes a customized list of
    * GO-ON events that define termination condition for this I/O operation.
    * <p>
    * This call is responsible for setting the initial focus to the first
    * enabled widget or NEXT-PROMPT hint.
    *
    * @param   focusWidgetId
    *          ID of the target frame for this editing session
    * @param   eventList
    *          Information about events to wait for.
    */
   private void startEditingMode(int focusWidgetId, EventList eventList)
   {
      eventLists.push(currentEventList);
      currentEventList = eventList;
      inEditingBlock  = true;
      isGoPending     = 0;
      hadReset        = false;
      forceWipeScreen = false;
      Widget comp =  getWidget(focusWidgetId);
      Widget frame = UiUtils.locateFrame(comp);
      // translate widget
      if (frame instanceof Frame)
         comp = ((Frame) frame).getField(comp);

      // this should be before the focus change to leave the cursor in
      // the right place
      startInputStatus(true);

      // save current focus, to be restored when editing block finishes
      editingFocus.push(UiUtils.getCurrentFocus());

      // ENTRY must be sent to the requested widget before the EDITING loop
      // starts executing
      focusManager.focusUndefExplicit(comp, eventList, eventList.getOperation(), true);

      editingFrames.push((Frame)frame);

      waitForFrames.push(new WaitForOp((Frame)frame, BlockingOperation.PROMPT_FOR));
   }

   /**
    * Marks the end of the editing block when triggered by any abnormal exit
    * from the block on the server.  This method does NOT need to be called
    * if the exit is naturally triggered via {@link #continueEditing}. This
    * method allows the client's state to be cleaned up in the abnormal
    * termination case.
    *           
    * @return   The screen buffer array for all editable frames.
    */
   @Override
   public ScreenBuffer[] stopEditingMode()
   {
      if (ignoreNextStopEditing)
      {
         ignoreNextStopEditing = false;
         return null;
      }

      stopInputStatus();
      isGoPending      = 0;
      inEditingBlock   = false;
      hadReset         = false;
      currentEventList = eventLists.pop();
      editingFrames.pop();
      // restore focus to the widget which had focus before editing mode
      // started, if in nested trigger mode
      Widget focus = editingFocus.pop();
      if (triggerNesting > 0 && focus != null)
      {
         focusManager.setFocusOn(focus);
      }
      else
      {
         WindowManager.getDefaultWindow().setFocus(null);
      }
      
      // TODO: should we clear leaveSent here?

      // restore insert indicator state
      restoreInsState(savedInsStateInitials.pop(), true);
      forceWipeScreen = true;
      
      WaitForOp wop = waitForFrames.pop();
      
      try
      {
         // re-enable editing mode for the duration of this call
         inEditingBlock  = true;
         return getEditableClientScreenBuffers(getScreenBuffer(wop.frame));
      }
      finally
      {
         inEditingBlock = false;
      }
   }

   /**
    * Conditionally blocks until a next UI event is available.
    */
   @Override
   public void waitForNextKey()
   {
      typeAhead.waitForInput();
   }

   /**
    * Called by the protocol driver to obtain the latest state that should
    * be sent to the other side of the connection.  This is called just
    * before a "real" message is sent and any changes provided will be
    * piggybacked into that message.
    * <p>
    * To enable incremental changes, it is guaranteed that this method will
    * be called every time any message is sent to the other side of the
    * connection.  In addition, this will only be called once per message.
    * <p>
    * In a system in which control flow is "sent" to the other side of the
    * connection (a type of "conversation" or single-threaded synchronous
    * mode), it is further guaranteed that any changes that are made at the
    * other side after this is called will result in a call to this side's
    * {@link #applyChanges} method before this method is ever called again
    * (when control flow returns to this side of the connection). This allows
    * one to ensure that on the call to this method, any "dirty" flags can be
    * safely cleared and the state of this side will be updated with
    * the results of the other side's changes before this method must ever
    * again check to see if there are changes to send.
    * <p>
    * Note that any user input events received by the application while the
    * thread is waiting for the server's response are enqueued in the event
    * queues and processed once server response is received and the thread
    * resumes processing. During this time the client app appears as locked
    * for any input.
    *
    * @return   The payload of state changes to be sent to the other side or
    *           <code>null</code> if no changes need to be sent.
    */
   @Override
   public Serializable getChanges()
   {
      ClientState cs = new ClientState();

      if (removedFrames.size() > 0)
      {
         cs.deregister = Utils.integerCollectionToPrimitive(removedFrames);
         removedFrames.clear();
      }

      if (wrappedFrames.size() > 0)
      {
         cs.wrapped = Utils.integerCollectionToPrimitive(wrappedFrames);
         wrappedFrames.clear();
      }

      ArrayList<Integer> ids  = new ArrayList<Integer>();
      ArrayList<Integer> down = new ArrayList<Integer>();

      for (TitledWindow wnd : WindowManager.windowList())
      {
         if (!(wnd instanceof TopLevelWindow))
         {
            continue;
         }

         int[] list = UiUtils.getFrameIDs((TopLevelWindow) wnd);
         for (int i : list)
         {
            // skip tinyInput frame
            if (i == WidgetId.INPUT_FRAME_ID.asInt())
               continue;

            Frame frame = (Frame) getWidget(new WidgetId(i));

            if ((frame ==  null) || !frame.isDown() || frame.getDown() == 0 ||
                 frame.isRedirected())
               continue;

            ids.add(Integer.valueOf(i));
            down.add(Integer.valueOf(frame.getDown()));
         }
      }

      cs.downIDs = Utils.integerCollectionToPrimitive(ids);
      cs.down    = Utils.integerCollectionToPrimitive(down);

      // was there a pause in this last operation?
      if (hadPause)
      {
         // notify the server
         cs.hadPause = true;

         // clear our state
         hadPause = false;
      }
      
      if (functionKey != KeyInput.CHAR_UNDEFINED)
      {
         cs.functionKey = functionKey;
         functionKey = KeyInput.CHAR_UNDEFINED;
      }
      
      cs.eventX = lastEventX;
      cs.eventY = lastEventY;
      
      if (lastEvent != null)
      {
         cs.virtualLastKey = lastEvent;
         lastEvent = null;
      }
      
      cs.configUpdates = ConfigManager.getInstance().getConfigUpdates();
      
      Window<?> active = WindowManager.getActiveWindow();
      cs.activeWindowId = active.getId().asInt();
      return cs;
   }

   /**
    * Called by the protocol driver upon reading a message that contains
    * state changes that need to be applied to this side of the connection.
    * This is called just after a new message is read from the connection
    * and only if that message had piggybacked state changes attached. This
    * is guaranteed to be called before the containing message is enqueued
    * for further "real" processing on this node.
    *
    * @param    changes
    *           The payload of state changes that must be locally applied.
    */
   @Override
   public void applyChanges(Serializable changes)
   {
      // WARNING: DO NOT call any processing from here which may generate
      //          an exception (especially any of the ConditionExceptions).
      //          Any pausing logic or other event processing must be
      //          avoided.  This is because this method is called during
      //          protocol-driver processing, before the proper environment
      //          is setup to deliver any condition exceptions back to the
      //          server as a valid response.  Anything thrown at this time
      //          will cause the client to abend!

      ConditionException saveExc = immediateException;
      immediateException = null;

      try
      {
         ServerState sstate = (ServerState) changes;
         interactions = sstate.interactions;

         if (sstate.nesting > nesting)
         {
            Window.pushStatus(sstate.nesting - nesting);
         }
         else if (sstate.nesting < nesting)
         {
            Window.popStatus(nesting - sstate.nesting);
         }

         nesting = sstate.nesting;

         // cache any conditional up/down value for the specified frame (it will
         // be honored in the actual ClientExports call that follows)
         if (sstate.conditionalFrameId != 0)
         {
            this.conditionalUpDown = sstate.conditionalUpDown;
            this.conditionalFrameId = sstate.conditionalFrameId;

            Widget frame = getWidget(conditionalFrameId);

            if (frame instanceof Frame && !((Frame) frame).isDown() && frame.isVisible())
            {
               Frame f = (Frame)frame;

               f.setNeedReset(true);

               // only frames on the root nesting level are
               // forced a repaint
               if (f.getScopeNesting() == 1)
               {
                  f.setDelayedRepaint(true);
               }
            }
         }

         if (sstate.formatCheckId > 0)
         {
            Widget frame = getWidget(sstate.formatCheckId);

            if (frame instanceof Frame && frame.isVisible())
            {
               ((Frame) frame).setFormatCheck(true);
            }
         }
         
         if (sstate.currentWindow != WindowConfig.RESOLVE_WINDOW)
         {
            Window<?> wnd = (Window<?>) tk.getRegistry().getComponent(sstate.currentWindow);
            
            // TODO: not sure how to process null here.
            WindowManager.setCurrentWindow(wnd);
         }

         // implements deferred hide/pause MARKING (only state changes, no
         // actual hiding or pausing occurs)
         if (sstate.markList != null)
         {
            for (MarkEntry item : sstate.markList)
            {
               markFrames(item);
            }
         }
         
         // update envrionments collection
         if (sstate.environments != null && !sstate.environments.isEmpty())
         {
            ColorManager.update(sstate.environments);
         }

         if (sstate.deadFrames != null)
         {
            for (int frameId : sstate.deadFrames)
            {
               Frame frame = (Frame) getWidget(frameId);
               if (frame != null)
               {
                  frame.setAlive(false);
               }
            }
         }
         
         // no config updates are incoming from the server-side: all changes are sent via
         // pushScreenDefinition API calls
      }
      finally
      {
         immediateException = saveExc;
      }
   }

   /**
    * Checks whether the CHUI is up and running on the client.
    *
    * @return  <code>true</code> if ChUI driver <code>false</code> if GUI driver.
    */
   @Override
   public boolean isChui()
   {
      return tk.isChui();
   }
   
   /**
    * Check the client type: embedded or not.
    * 
    * @return   <code>TRUE</code> if the client is running in embedded mode, 
    *           <code>FALSE</code> otherwise 
    */
   @Override
   public boolean isEmbedded() 
   {
      ScreenDriver<?> driver = OutputManager.getDriver();
      if (driver instanceof EmbeddedClient)
      {
         return ((EmbeddedClient) driver).isEmbeddedClient();
      }
      
      return false;
   }

   /**
    * Check if the client is running in windows compatibility mode.
    * 
    * @return   See above.
    */
   public boolean isUnderWindows()
   {
      return underWindows;
   }

   /**
    * Sets the SESSION:TOOLTIPS attribute for this session.
    *
    * @param    tooltips
    *           The new value for session related TOOLTIPS attribute.
    */
   @Override
   public void setSessionTooltips(boolean tooltips)
   {
      // default or ChUI do nothing
      if (isChui() || ToolTip.getSessionTooltips() == tooltips)
      {
         return;
      }
      
      // this is GUI code
      ToolTip.setSessionTooltips(tooltips);
      // activate all pending tootip workers
      if (tooltips)
      {
         ((GuiDriver) tk.getInstanceDriver()).registerPendingTooltipWorkers(0);
      }
      else
      {
         ((GuiDriver) tk.getInstanceDriver()).deregisterTooltipWorkers(0);
      }
   }
   
   /**
    * Activates tootip feature for particular widget.
    *
    * @param   wId
    *          The widget ID to activate tooltip.
    */
   @Override
   public void activateTooltipForWidget(int wId)
   {
      // protect from calling in CHUI mode
      if (isChui() || !ToolTip.getSessionTooltips())
      {
         return;
      }
      
      // activate pending tooltip worker for particular widget
      GuiDriver gd = (GuiDriver) tk.getInstanceDriver();
      Widget<?> widget = getWidget(wId);
      TopLevelWindow<?> window = widget.topLevelWindow();
      int winID = UiUtils.getWidgetIdAsInt(window);

      gd.selectWindow(winID);
      gd.registerPendingTooltipWorkers(wId);
      gd.releaseWindow();
   }
   
   /**
    * Deactivates tootip feature for particular widget.
    *
    * @param   wId
    *          The widget ID to deactivate tooltip.
    */
   @Override
   public void deactivateTooltipForWidget(int wId)
   {
      // protect from calling in CHUI mode
      if (isChui())
      {
         return;
      }
      
      // activate pending tootip worker for particular widget
      ((GuiDriver) tk.getInstanceDriver()).deregisterTooltipWorkers(wId);
   }
   
   /**
    * Destroy the frame and free associated resources as needed. If the frame
    * is a dialog, it will be hidden immediately, otherwise it will be marked
    * as out of scope and valid for being hidden and resources will be
    * released when frame will be hidden.
    *
    * @param    frameId
    *           The frame requiring cleanup.
    */
   @Override
   public void destroyFrame(int frameId)
   {
      final Frame frame = (Frame) getWidget(frameId);

      if (frame == null)
         return;

      // cleanup any traces of this frame in our page elements list
      forceDeregisterPageElement(frameId);

      // mark the frame as one that can be destroyed (it is inactive)
      frame.setAlive(false);

      // dialog boxes MUST actually hide right now
      if ((frame.isDialog() || frame.config().overlay) && 
          (frame.isVisible() || frame.isDelayedRepaint()))
      {
         eventDrawingBracket(frame, new Runnable()
         {
            @Override
            public void run()
            {
               if (frame.isDelayedRepaint())
               {
                  frame.repaint();
               }
               else
               {
                  // this also triggers frame removal
                  frame.setVisible(false);
               }
            }
         });
      }
      else
      {
         // hidden frames removed immediately.
         if (!frame.isVisible())
         {
            removeFrame(frame);
         }
         else
         {
            // redirected visible frames should be removed immediately
            // safety net, should not happen normally
            if (frame.isRedirected() || frame.isPlacedForRedirected())
            {
               removeFrame(frame);
            }
            else
            {
               // at this time, all enabled widgets must be disabled
               // but only for frames which were nor redirected nor
               // placed for redirected mode (i.e. this is only for
               // interactive mode frames!)

               selectiveDisable(frame);
            }

            // regular visible frames will be removed when setVisible(false)
            // is called on them.
         }
      }
   }

   /**
    * Destroy the menu and free associated resources as needed. If the menu
    * is popup-only, it will be hidden immediately, otherwise it will be marked
    * as out of scope and valid for being hidden and resources will be
    * released when menu will be hidden.
    *
    * @param    menuId
    *           The menu requiring cleanup.
    */
   @Override
   public void destroyMenu(int menuId)
   {
      // TODO implement
   }
   
   /**
    * Selectively disable and redraw widgets.
    *
    * @param    frame
    *           Frame where widgets should be disabled.
    */
   private void selectiveDisable(final Frame frame)
   {
      if (!frame.isDisplayed())
      {
         return;
      }

      final ArrayList<Widget> widgets = new ArrayList<Widget>();

      final Frame[] above = UiUtils.getFramesAbove(frame);

      independentEventDrawingBracket(frame, new Runnable()
      {
         @Override
         public void run()
         {
            Widget[] list = frame.getContentPane().widgets();

            // collect enabled widgets
            for (Widget comp : list)
            {
               if (comp.isEnabled() && UiUtils.hasConfig(comp))
               {
                  comp.setEnabled(false);
                  widgets.add(comp);
               }
            }

            repaintAffectedWidgets(above, widgets);
         }
      });
   }

   /**
    * Checks whether a GO event is pending for this editing operation
    * (either an <code>editing</code> block or a <code>wait-for</code>).
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>GO-PENDING </b>
    *
    * @return   <code>true</code> if an event has been applied that caused
    *           pending GO for the current editing operation.
    */
   @Override
   public boolean isGoPending()
   {
      return isGoPending > 0;
   }

   /**
    * Get frame widget from the given frame ID.
    *
    * @param   frameId
    *          Frame ID.
    * @return  Reference to Frame instance.
    */
   public Frame getFrame(int frameId)
   {
      Frame frame = (Frame) getWidget(frameId);

      if (frame == null)
         throw new RuntimeException("Invalid frame ID " + frameId);

      return frame;
   }

   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-LINE
    * </pre>
    *
    * @param    frame
    *           The frame for this request.  Must NOT be <code>null</code>.
    * @param    pendingUpDown
    *           The pending down value for this frame.
    *
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is not a down frame then 1 will be returned.
    */
   public int getFrameLine(Frame frame, int pendingUpDown)
   {
      return frame.getCursor() + pendingUpDown + 1;
   }

   /**
    * Process events internally if there are no interactive requests from the
    * server.
    */
   private void cleanupQueue()
   {
      if (currentEventList == null)
         processEventsWorker();
   }

   /**
    * Force the cursor to the specified location and make it visible again.
    *
    * @param    curpos
    *           The 0-based location to which the cursor must be forced.
    * @param    curvis
    *           The new cursor visibility state (<code>true</code> means
    *           set the cursor visible).
    *
    * @return   The previous cursor visibility state.
    */
   public boolean restoreCursor(NativePoint curpos, boolean curvis)
   {
      // move the cursor
      ((ChuiOutputManager) tk).at(curpos);

      // the physical cursor position on the terminal is not moved until a
      // sync, until then the cursor is only moved in a logical operation
      // and setCursorStatus will re-display the current physical cursor
      // without a sync so the cursor will appear at an unexpected location
      // if we don't call sync here
      tk.sync(WidgetId._DEFAULT_WINDOW_ID);

      // set the cursor visibility
      return tk.setCursorStatus(curvis);
   }

   /**
    * Returns if processing is in editing block.
    *
    * @return <code>true</code> if in editing block, <code>false</code> - otherwise. 
    */
   public boolean isInEditingBlock()
   {
      return inEditingBlock;
   }

   /**
    * Get the last viewed frame.
    * 
    * @return   See above.
    */
   public Frame getLastViewed()
   {
      return lastViewed;
   }

   /**
    * Set last viewed frame reference.
    *
    * @param   frame
    *          Last viewed frame.
    */
   public void setLastViewed(Frame frame)
   {
      lastViewed = frame;
   }

   /**
    * Honor the deferred processing for a conditional up or down.
    *
    * @param    noPause
    *           Make no pause if frame will need it during DOWN processing.
    *
    * @return   <code>true</code> if down frame was wrapped during processing of
    *           conditional UP/DOWN.
    */
   private boolean honorConditionalUpDown(boolean noPause)
   {
      boolean res = false;

      if (conditionalFrameId != 0)
      {
         int tmp = conditionalFrameId;
         conditionalFrameId = 0;
         res = downWorker(tmp, conditionalUpDown, true, false, noPause);
         conditionalUpDown  = 0;
      }

      return res;
   }

   /**
    * Honor the deferred processing for a conditional up or down.
    */
   private void honorConditionalUpDown()
   {
      if (conditionalFrameId != 0)
      {
         Frame frame = (Frame) getWidget(conditionalFrameId);

         if (frame != null)
         {
            if (frame.isAllClear())
            {
               frame.setAllClear(false);
               conditionalUpDown  = 0;
               conditionalFrameId = 0;
               return;
            }
            int tmp = conditionalFrameId;
            conditionalFrameId = 0;
            downWorker(tmp, conditionalUpDown, true, false, false);
         }
         conditionalUpDown  = 0;
         conditionalFrameId = 0;
      }
   }

   /**
    * Get <code>Component</code> instance for a given ID.
    *
    * @param   componentId
    *          ID of the requested <code>Component</code> instance.
    * @return  reference to <code>Component</code> instance with given ID.
    */
   public Widget getWidget(WidgetId componentId)
   {
      Widget<?> w = registry().getComponent(componentId);
      
      return w;
   }

   /**
    * Get <code>Component</code> instance for a given ID.
    *
    * @param   componentId
    *          ID of the requested <code>Component</code> instance.
    * @return  reference to <code>Component</code> instance with given ID.
    */
   public Widget getWidget(int componentId)
   {
      Widget<?> w = registry().getComponent(componentId);
      
      return w;
   }

   /**
    * Sets the format specification for the specified widget and redraws the
    * widget.
    *
    * @param    widget
    *           The widget to modify.
    * @param    format
    *           The format specification.
    */
   @Override
   public void setFormat(final int widget, final String format)
   {
      final Widget wcomp = getWidget(widget);
      if (wcomp == null)
      {
         return;
      }

      eventDrawingBracket(wcomp, new Runnable()
      {
         @Override
         public void run()
         {
            Widget comp = wcomp;

            int   fid   = registry().lookupFrameIdFromWidgetId(widget);
            Frame frame = getFrame(fid);

            if (frame.isDown())
            {
               comp = frame.getField(comp);
            }

            // TODO: this really should raise an error condition if the
            //       new format string specifies a larger size than can
            //       already fit in the widget (the frame should NOT relayout
            //       but it does in our implementation)
            final WidgetConfig c = comp.config();
            if (c != null)
            {
               ConfigManager.syncConfigChanges(new Runnable()
               {
                  @Override
                  public void run()
                  {
                     c.setDynamicFormat(format);
                  }
               });
            }
         }
      });
   }

   /**
    * Set the new decimal and group separators.
    * 
    * @param    decSep
    *           The new decimal separator.
    * @param    groupSep
    *           The new group separator.
    */
   public void setNumericFormat(char decSep, char groupSep)
   {
      NumberType.setNumericFormat(decSep, groupSep);
   }
   
   /**
    * Returns the action code of the last key input event.
    *
    * @return The action code of the last key input event.
    */
   public int getLastActionCode()
   {
      return lastActionCode;
   }

   /**
    * Gets the current event list.
    *
    * @return The currently active event list.
    */
   public EventList getCurrentEventList()
   {
      return currentEventList;
   }

   /**
    * Prepare message text for specified time to wait.
    *
    * @param   seconds
    *          Seconds to wait for an event. Negative value means wait
    *          indefinitely and number of seconds is not displayed.
    *
    * @return  prepared text message.
    */
   private String makePauseText(int seconds)
   {
      StringBuffer buf = new StringBuffer();

      if (seconds > 0)
      {
         buf.append("Pausing for ");
         buf.append(seconds);
         buf.append(" seconds. ");
      }

      buf.append(PBH_DEFAULT_TEXT);

      return buf.toString();
   }

   /**
    * Compute the next focus in a wait-for loop.
    * 
    * @param   frameAsTarget
    *          if <code>true</code>, then the active frame is the target of
    *          events if nothing is enabled and in focus, otherwise
    *          the current window
    *          
    * @return  The currently focused widget, or <code>null</code> if no focus can be determind.
    */
   private Widget computeFocus(boolean frameAsTarget)
   {
      // get and check the focus - the widget in focus may be disabled
      Widget comp = UiUtils.getCurrentFocus();
      if (comp != null && !comp.navigable())
      {
         comp = null;
      }

      // reposition the cursor before blocking (interactive mode only)
      positionCursor();

      // determine the target for this keystroke
      // If there is no focused and enabled widget, the target should be
      // either the last used frame or the current window.
      // This implementation leaves the decision to the caller. Check the
      // waitForWorker() method, where this decision is made. That logic
      // should allow CHOOSE to always direct the keystrokes to the frame

      if (comp == null || comp instanceof TopLevelWindow)
      {
         Frame frame = getActiveFrameInt();

         if (frameAsTarget && frame != null)
         {
            comp = frame;
         }
         else
         {
            comp = WindowManager.getDefaultWindow();
         }
      }
      
      return comp;
   }
   
   /**
    * Wait for key press indefinitely or for specified number of seconds.
    *
    * @param   seconds
    *          -1 means indefinite wait, 0 means no wait and positive values
    *          mean the number of seconds to wait
    * @param   frameAsTarget
    *          if <code>true</code>, then the active frame is the target of
    *          events if nothing is enabled and in focus, otherwise
    *          the current window
    * @param   honorServerEvent
    *          flag indicating if server events should be honored or not.
    * 
    * @return  key event code or null if no key was received during specified
    *          time.
    */
   private Event waitForEvent(int seconds, boolean frameAsTarget, boolean honorServerEvent)
   {
      return waitForEvent(seconds, frameAsTarget, honorServerEvent, false);
   }
   
   /**
    * Wait for key press indefinitely or for specified number of seconds.
    *
    * @param   seconds
    *          -1 means indefinite wait, 0 means no wait and positive values
    *          mean the number of seconds to wait
    * @param   frameAsTarget
    *          if <code>true</code>, then the active frame is the target of
    *          events if nothing is enabled and in focus, otherwise
    *          the current window
    * @param   honorServerEvent
    *          flag indicating if server events should be honored or not.
    * @param   includeOSEvent
    *          flag indicating if OS events (mouse, window) should be considered the as action 
    *          source or not.
    * 
    * @return  key event code or null if no key was received during specified time.
    */
   private Event waitForEvent(int     seconds,
                              boolean frameAsTarget,
                              boolean honorServerEvent,
                              boolean includeOSEvent)
   {
      // the event to be returned
      Event evt = null;
      boolean isChui = isChui();
      
      // reset the LAST-EVENT:X/Y attributes
      lastEventX = null;
      lastEventY = null;
      
      // release the window and allow it lock it again after the event was received
      do
      {
         Widget comp = computeFocus(frameAsTarget);
         Object state = null;
         if (!isChui)
         {
            state = ((GuiDriver) tk.getInstanceDriver()).saveDrawingState();
         }
         
         evt = typeAhead.getKeystroke(seconds * 1000, honorServerEvent);

         if (!isChui)
         {
            ((GuiDriver) tk.getInstanceDriver()).restoreDrawingState(state);
         }

         if (evt == null && EventManager.hasOSEvents())
         {
            // get possible OS event
            Event osEvent = EventManager.getOSEvent();
            if (osEvent == null)
            {
               continue;
            }

            if (osEvent instanceof InvocationEvent)
            {
               ((InvocationEvent) osEvent).getRunnable().run();
               continue;
            }
            
            // in the case we request mouse event to be considered and have mouse click, press
            // or release pass the event to the key handling route
            // the press and release events generate click
            if (includeOSEvent && 
                ((osEvent instanceof MouseEvt &&
                 (((MouseEvt) osEvent).getEvent().getID() == MouseEvt.MOUSE_CLICKED ||
                  ((MouseEvt) osEvent).getEvent().getID() == MouseEvt.MOUSE_PRESSED ||
                  ((MouseEvt)osEvent).getEvent().getID() == MouseEvt.MOUSE_RELEASED)) ||
                 (osEvent.id() == Keyboard.SE_WINDOW_RESTORED  ||
                  osEvent.id() == Keyboard.SE_WINDOW_CLOSE     ||
                  osEvent.id() == Keyboard.SE_WINDOW_MINIMIZED ||
                  osEvent.id() == Keyboard.SE_WINDOW_MAXIMIZED ||
                  osEvent.id() == Keyboard.SE_WINDOW_RESIZED)))
            {
               // however sometimes we have triggers or exit event attached to 4GL mouse/windows
               // event in this case we need to exit the key reading loop here
               
               if (osEvent instanceof MouseEvt)
               {
                  int mouseEvId = ((MouseEvt) osEvent).getEvent().getID();
                  
                  if (mouseEvId == MouseEvt.MOUSE_CLICKED || mouseEvId == MouseEvt.MOUSE_RELEASED)
                  {
                     lastKey = Keyboard.keyCode(Keyboard.MOUSE_SELECT_UP);
                  }
                  else if (mouseEvId == MouseEvt.MOUSE_PRESSED)
                  {
                     lastKey = Keyboard.keyCode(Keyboard.MOUSE_SELECT_DOWN);
                  }
                  
                  if (mouseEvId == MouseEvt.MOUSE_CLICKED)
                  {
                     Widget w = osEvent.source();                     
                     if (w instanceof Window)
                     {
                        w = ((Window) w).findMouseSource((MouseEvt) osEvent);
                     }
                     
                     Window<?> wnd = w == null ? null : (Window<?>) w.parentOrSelf(Window.class);
                     
                     MenuGuiImpl menubar = wnd == null ? null : (MenuGuiImpl) wnd.getMenubar();
                     
                     if (wnd != null && menubar != null && !(w instanceof MenuElement))
                     {
                        MenuGuiImpl.resetMbarPressedState();
                        menubar.setKeyMode(false);
                        getInstance().independentEventDrawingBracket(menubar, () ->  
                        {
                           wnd.getMenubar().repaint();               
                        });
                     }
                  }
               }
               
               return osEvent;
            }
            else
            {
               // OS events will be treated first
               applyWorker(osEvent);

               // recompute the focus (might have changed)
               comp = computeFocus(frameAsTarget);

               // force waiting for a key
               continue;
            }
         }
         
         if (evt != null)
         {
            KeyInput keyEvt = (KeyInput) evt;
            
            // ALT doesn't terminate PAUSE
            if (keyEvt.keyCode() == Key.VK_ALT && currentBlockingOp == BlockingOperation.PAUSE)
            {
               continue;
            }

            if (comp == null)
               comp = keyEvt.source();

            // filter the default.cancel button related events
            Frame<?> frame = (Frame<?>) UiUtils.locateFrame(comp);
            if (frame != null)
            {
               if (keyEvt.keyCode() == Key.VK_ESCAPE && frame.processCancelButton())
               {
                  continue;
               }
               if (keyEvt.keyCode() == Key.VK_ENTER && frame.processDefaultButton(keyEvt))
               {
                  break;
               }
            }

            boolean real = keyEvt.isRealKey();

            evt = new KeyInput(comp, keyEvt.id(), keyEvt.keyCode(), false, real);
            lastKey = keyEvt.keyCode();
            
            // only if the event loop is a WAIT-FOR
            if (!processMnemonic((KeyInput) evt, keyEvt.keyCode(), comp) &&
                (currentBlockingOp == BlockingOperation.WAIT_FOR ||
                 currentBlockingOp == BlockingOperation.PROMPT_FOR)) 
            {
               // get the root frame for GUI mode
               if (!isChui() && frame != null)
               {
                  frame = frame.getRootFrame();
               }
               // store the operation for loop to exit inside mext call
               waitForFrames.push(new WaitForOp(frame, currentBlockingOp));
               // check for system event
               boolean res = checkForSystemEvent((KeyInput) evt);
               // get the wait for frames stack state back
               waitForFrames.pop();
               // go and read another key if it was processed via a system event
               if (res)
               {
                  continue;
               }
            }
         }
         else
         {
            lastKey = -1;
   
            // check if a server or mouse event is on queue
            if (honorServerEvent && EventManager.hasServerEvents())
            {
               evt = EventManager.getServerEvent();
            }
            else if (EventManager.hasOSEvents())
            {
               continue;
            }
         }
         
         break;
      }
      while (true);

      return evt;
   }

   /**
    * Check if the given event can be treated by the focused widget on its own.  Some widgets
    * allow the event to be "consumed" and a new key read, in cases when a READKEY within a 
    * EDITING block is used: in these cases, the READKEY does not return until AFTER a key which
    * has an associated trigger has be pressed IOR a key which is treated as a global system 
    * event.
    *  
    * @param    key
    *           The key to be checked.
    *           
    * @return   <code>true</code> if more keys need to be read.
    */
   private boolean checkForSystemEvent(final KeyInput key)
   {
      if (key == null)
      {
         return false;
      }
      
      // only widgets which can handle system events are treated
      final AbstractWidget focused = (AbstractWidget) UiUtils.getCurrentFocus();
      boolean moreKeys = false;

      if (focused != null            &&
          focused.manageSystemActions() &&
          !focused.ignoreAction(key.actionCode()))
      {
         // assume no trigger is found
         moreKeys = true;

         Frame frame = (Frame) UiUtils.locateFrame(focused);
         int focusId = UiUtils.lookupWidgetIdAsInt(focused);
         int frameId = UiUtils.lookupWidgetIdAsInt(frame);
         
         TriggerMatch result = new TriggerMatch(key.keyCode());

         // check if there is a trigger for this specific key
         currentEventList.lookup(frameId, focusId, true, true, result);
         
         if (result.triggerId == -1)
         {
            // no trigger found, check for a default event for this key
            int defaultEvent = focused.getDefaultEvent(key, inEditingBlock);
            
            if (defaultEvent != -1)
            {
               result = new TriggerMatch(defaultEvent);
               
               // check if there is a trigger for this default event
               currentEventList.lookup(frameId, focusId, true, true, result);
               
               // only if there is no other more specific match AND there is a match to 
               // a system event, do we call invoke
               if (result.triggerId != -1)
               {
                  // invoke trigger
                  invokeTriggers(focused, defaultEvent, true);
                  
                  this.functionKey = defaultEvent;

                  if (isGoPending())
                  {
                     moreKeys = false;
                  }
               }
            }
            
            if (moreKeys)
            {
               // if the key is ALT, send it to the window
               if (key.keyCode() == Key.VK_ALT && !isChui())
               {
                  focused.ancestor().processEvent(key);
                  moreKeys = false;
               }
               else if (currentEventList.isGoEvent(Keyboard.eventName(key.keyCode())) ||
                        currentEventList.isGoEvent(Keyboard.eventName(key.actionCode())))
               {
                  moreKeys  = false;
               }
               else
               {
                  // let the widget process this key; more keys will be read after this
                  focused.processSystemKey(key);
                  key.consume();
                  positionCursor();
   
                  if (isGoPending())
                  {
                     moreKeys = false;
                  }
                  else
                  {
                     ScreenBuffer after = getScreenBuffer(focused);
                     if (after != null)
                     {
                        server.refreshBuffers(getEditableClientScreenBuffers(after));
                     }
                  }
               }
            }
         }
         else
         {
            // a trigger was found, stop reading keys
            moreKeys = false;
         }
      }
      
      return moreKeys;
   }

   /**
    * Run pause cleanup block if any.
    */
   private void pauseCleanup()
   {
      synchronized (cleanupSynchronizer)
      {
         if (cleanupBlock != null)
            cleanupBlock.run();
         cleanupBlock = null;
      }
   }

   /**
    * Generate a copy of the <code>ScreenBuffer</code> instance for the given
    * frame or for the frame in which the given widget is contained.
    *
    * @param    widget
    *           A frame or a widget in the frame whose screen buffer must be
    *           returned.
    *
    * @return   <code>ScreenBuffer</code> instance.
    */
   private ScreenBuffer getScreenBuffer(Widget widget)
   {
      // find the associated frame
      Widget frame = UiUtils.locateFrame(widget);

      if (frame == null)
      {
         // nothing to do
         return null;
      }

      int frameId = frame.getId().asInt();

      return getScreenBuffer(frameId, -1, false);
   }

   /**
    * Generate a copy of the <code>ScreenBuffer</code> instance for the given
    * frame.
    *
    * @param    frameId
    *           The frame whose screen buffer must be returned.
    * @param    size
    *           The number of widgets in the screen buffer in total or -1
    *           to allow this value to be calculated.
    *
    * @return   <code>ScreenBuffer</code> instance.
    */
   private ScreenBuffer getScreenBuffer(int frameId, int size)
   {
      return getScreenBuffer(frameId, size, false);
   }

   /**
    * Generate a copy of the <code>ScreenBuffer</code> instance for the given
    * frame.
    *
    * @param    frameId
    *           The frame whose screen buffer must be returned.
    * @param    size
    *           The number of widgets in the screen buffer in total or -1
    *           to allow this value to be calculated.
    * @param    force
    *           <code>true</code> to copy the current data into the screen
    *           buffer without regard to the changed flag.
    *
    * @return   <code>ScreenBuffer</code> instance.
    */
   private ScreenBuffer getScreenBuffer(int frameId, int size, boolean force)
   {
      if (frameId == -1)
      {
         // nothing to do
         return null;
      }

      Frame        frame  = (Frame) getWidget(frameId);
      // another NPE protection
      if (frame == null)
      {
         // this conidition is pretty rare, one known case is when the frame from editable frame
         // list does not exist in general widget registry, can be a kind if race or timing
         LOG.log(Level.WARNING, "Unable to get the screen buffer for frame ID: "+frameId+
                                ". The frame with given ID is missing in widget registry.");
         return null;
      }

      // get the list of all widgets
      WidgetId[] list = registry().getIDs(frameId);
      int lSize = (list != null) ? list.length : 0;

      // find out the maximum number of widgets, if needed
      if (size == -1)
      {
         size = 0;
         // do not process any widgets which are client-side only (negative IDs)
         for (int i = 0; i < lSize; i++)
         {
            if (!WidgetId.virtualWidget(list[i]))
            {
               size ++;
            }
         }
      }

      // create the ScreenBuffer that big
      Widget       comp   = null;
      WidgetConfig config = null;

      ScreenBuffer buffer = new ScreenBuffer(frameId, size);

      for (int i = 0; i < lSize; i++)
      {
         // skip labels or other widgets widhout server-side counter part
         if (WidgetId.virtualWidget(list[i]))
            continue;

         int id = list[i].asInt();
         comp = registry().getComponent(list[i]);

         // WARNING: short-cut taken here! Since the only caller that uses
         //          force flag today is CHOOSE, we use this flag for a
         //          2nd purpose. If more users are made, then this would
         //          need to change at that time.

         if (force && frame.isChooseActive())
         {
            // CHOOSE does not maintain the down frame cursor in ROW mode
            // so we need to find the actual widget that is at the selected
            // index position using the base widget to identify the column.
            comp = frame.getSelectedField(comp);
         }
         else
            comp = frame.getField(comp);

         if (comp instanceof DataContainer)
         {
            boolean changed = (comp.getState() == ScreenBuffer.CHANGED);

            if (force || changed)
            {
               boolean uninit = comp.getState() == ScreenBuffer.UNINITIALIZED;
               
               if (!force || !uninit)
               {
                  // all validation errors are ignored here

                  // TODO: if there is a validation error, then the widget value defaults to the 
                  // screen-value
                  RemoteErrorData old = 
                     ErrorManager.switchErrorStorage(ErrorManager.getLocalAccess());
                  Object val = null;
                  
                  try
                  {
                     ErrorManager.setIgnore(true);
                     
                     val = ((DataContainer) comp).getValue();
                  }
                  finally
                  {
                     ErrorManager.setIgnore(false);

                     ErrorManager.switchErrorStorage(old);
                  }

                  if (val != null)
                  {
                     buffer.putWidgetValue(id, val, !changed);
                  }
                  
                  if (comp instanceof ComboBox)
                  {
                     ComboBox<?> cb = (ComboBox<?>) comp;
   
                     buffer.putScreenValue(id, cb.getScreenValue());
                     
                     if (!inEditingBlock && !cb.isSelected())
                     {
                        buffer.putWidgetValue(id, ((ComboBox) comp).getScreenValue(), !changed);
                     }
                  }
                  else if (comp instanceof SelectionList)
                  {
                     buffer.putScreenValue(id, ((SelectionList) comp).getScreenValue());
                  }
               }
            }
         }
         else if (comp instanceof Browse)
         {
            if (force) // assuming that browse column screen value is accessed in a trigger
            {
               Browse browse = (Browse) comp;
               int[] ids = browse.getIdsOfColumnWidgets();

               for (int columId : ids)
               {
                  BrowseColumn col = (BrowseColumn) registry().getComponent(columId);
                  buffer.putScreenValue(columId, col.getValue());
               }
            }
         }
         
         processSideLabel(buffer, comp);
      }

      // retrieve frame value/widget ID.

      FrameData dt = frame.getFrameValue(null);

      if (dt != null && dt.getValue() != null)
      {
         buffer.setFrameValue(dt.getValue());
         buffer.setValueWidgetId(dt.getId());
      }

      resetChanged(frameId);

      // non-redirected down frames need to switch screen buffers on
      // the server if the frame-line has changed during our execution
      // on the client
      if (frame.isDown() && !frame.isRedirected())
      {
         buffer.putLine(getFrameLine(frame, 0));
      }

      buffer.putKeyCode(lastKey);

      return buffer;
   }

   /**
    * Process the side-label associated with the given widget.
    * <p>
    * As side-labels are detached from normal frame widgets, the frame's buffer needs to be 
    * manually updated.
    * 
    * @param    buffer
    *           The screen buffer.
    * @param    comp
    *           The widget for which the side-label needs processing.
    */
   private void processSideLabel(ScreenBuffer buffer, Widget<?> comp)
   {
      ControlConfig cc = UiUtils.getConfig(comp, ControlConfig.class);
      if (cc == null)
      {
         return;
      }
      
      if (cc.sideLabelId == -1 || !cc.realized)
      {
         return;
      }
      
      Label<?> lbl = (Label<?>) ThinClient.getInstance().getWidget(cc.sideLabelId);
      
      // the screen-value is actually the label's name (the contained text)
      buffer.putScreenValue(cc.sideLabelId, lbl.name());
   }

   /**
    * Parses <code>ScreenBuffer</code> array for current set of frames and assigns new widget
    * values.  No checking for frame existence.
    *
    * @param    frameBuf
    *           <code>ScreenBuffer</code> array to apply.
    */
   private void setScreenBuffers(ScreenBuffer[] frameBuf)
   {
      setScreenBuffers(frameBuf, false);
   }
   
   /**
    * Parses <code>ScreenBuffer</code> array for current set of frames and assigns new widget
    * values.
    *
    * @param    frameBuf
    *           <code>ScreenBuffer</code> array to apply.
    * @param    checkFrameExists
    *           Flag indicating checking frame for existence before applying the buffer.
    */
   private void setScreenBuffers(ScreenBuffer[] frameBuf, boolean checkFrameExists)
   {
      // simple NPE check
      if (frameBuf == null)
      {
         return;
      }
      
      // handle every screen buffer from array
      for (ScreenBuffer buff : frameBuf)
      {
         // check for frame existence if required
         if (buff != null && (!checkFrameExists || getWidget(buff.getBaseId()) != null))
         {
            setScreenBuffer(buff);
         }
      }
   }

   /**
    * Parses <code>ScreenBuffer</code> instance for current set of frames
    * and assigns new widget values.
    *
    * @param    frameBuf
    *           <code>ScreenBuffer</code> instance.
    */
   private void setScreenBuffer(ScreenBuffer frameBuf)
   {
      setScreenBuffer(frameBuf, false);
   }

   /**
    * Parses <code>ScreenBuffer</code> instance for current set of frames
    * and assigns new widget values.
    * <p>
    * Allows direct drawing of headers.
    *
    * @param    frameBuf
    *           <code>ScreenBuffer</code> instance.
    * @param    direct
    *           if set to <code>true</code>, header widgets draw directly
    *           instead of posting <code>PaintEvent</code>.
    */
   private void setScreenBuffer(final ScreenBuffer frameBuf, boolean direct)
   {
      if (frameBuf == null)
      {
         return;
      }

      int  frameId = frameBuf.getBaseId();
      byte state   = -1;

      final Frame     frame  = getFrame(frameId);
      BaseDataType    value  = null;
      WidgetConfig config = null;

      if(frame.isDown())
      {
         if (frame.getCursor() < 0)
            frame.clear();

         frame.touchRow();
      }
      else
         frame.touchBody();

      // do not do direct drawing if frame is hidden
      if (!frame.isVisible())
         direct = false;

      // honor dynamic title if the frame is HIDDEN and a dynamic title was
      // specified (it is not null)
      String dynTitle = frameBuf.getTitle();

      if (dynTitle != null)
      {
         frame.config().title = dynTitle;
      }

      // iterate through all widgets and push new values and state as needed, but only for the
      // current row...
      WidgetId[] list = registry().getIDs(frameId);
      
      if (list != null)
      {
         for (WidgetId wid : list)
         {
            if (frame.isDown())
            {
               WidgetId resolved = Frame.resolveWidgetId(frame, wid.asInt());
               
               if (resolved.getClass() != wid.getClass() || !resolved.equals(wid))
               {
                  continue;
               }
            }
            
            int id = wid.asInt();
            state = frameBuf.getState(id);
            
            if (state == -1)
            {
               // widget not tracked by the screen-buffer, ignore it
               continue;
            }
   
            // get the widget
            Widget comp = registry().getComponent(id);
            
            if (comp == null)
            {
               continue;
            }
   
            if (frame.isDown())
               comp = frame.getField(comp);
   
            config = comp.config();
   
            byte oldStat = comp.getState();
            if (oldStat == state && state != ScreenBuffer.CHANGED)
               continue;
   
            boolean enabled = comp.isEnabled();
   
            switch (state)
            {
               case ScreenBuffer.UNCHANGED:
                  if (oldStat == ScreenBuffer.UNINITIALIZED && !enabled)
                     continue;
                  // fall through
   
               case ScreenBuffer.UNINITIALIZED:
   
                  if ((config instanceof ControlTextConfig)  &&
                      ((ControlTextConfig)config).flagStatic &&
                      state == ScreenBuffer.UNINITIALIZED)
                     break;
   
                  frame.setState(comp, state);
                  //TODO: no separate processing is necessary because frame can't
                  // be null here?
   //               if (config instanceof RadioSetConfig && frame == null)
   //                  ((RadioSet) comp).reset();
   //               else
   //                  if (frame != null)
                  frame.resetWidget(comp);
   
                  break;
   
               case ScreenBuffer.CHANGED:
                  if (comp instanceof DataContainer)
                  {
                     String fmt = null;
                     Object wv  = frameBuf.getWidgetValue(Integer.valueOf(id));
   
                     if (wv instanceof FormattedValue)
                     {
                        FormattedValue fv = (FormattedValue) wv;
                        value = fv.getValue();
                        fmt   = fv.getFormat();
                     }
                     else
                     {
                        value = (BaseDataType) wv;
                     }
   
                     setChangedValue(frame, comp, config, value, fmt);
                  }
                  else
                  {
                     System.err.println("Unexpected type (not DataContainer) " +
                                        "in setScreenBuffer: " + comp);
                  }
                  break;
   
               default:;
            }
         }
      }

      // refresh frame headers
      if (direct)
      {
         independentEventDrawingBracket(frame, new Runnable()
         {
            @Override
            public void run()
            {
               updateHeaders(frameBuf);
            }
         });
      }
      else
      {
         updateHeaders(frameBuf);
      }

      if (frame.isDown() &&
          frame.getCursor() < 0)
      {
         // reset cursor to 0
         frame.clear();
      }
   }

   /**
    * Update form headers from given {@link ScreenBuffer} instance.
    *
    * @param   frameBuf
    *          The {@link ScreenBuffer} instance to update headers from.
    *
    * @return  List of IDs of updated header widgets.
    */
   private ArrayList<Integer> updateHeaders(ScreenBuffer frameBuf)
   {
      Iterator<Integer> iter = frameBuf.getHeaderIDIterator();
      BaseDataType value;
      ArrayList<Integer> updated = new ArrayList<Integer>();

      while (iter.hasNext())
      {
         Integer hid = iter.next();

         Widget comp = registry().getComponent(hid.intValue());

         if (comp instanceof DataContainer)
         {
            value = (BaseDataType) frameBuf.getHeaderValue(hid);

            // value is set regardless if changed or not
            ((DataContainer) comp).setValue(value);

            if (!value.equals(((DataContainer) comp).getValue()))
            {
               updated.add(hid);
            }
         }
      }

      return updated;
   }

   /**
    * Applies the encoded event to the widget currently in focus in the given
    * frame, or directly to the specified widget.
    *
    * @param   event
    *          Progress event
    */
   private void applyWorker(final Event event)
   {
      eventDrawingBracket(event.source(), false, false, new Runnable()
      {
         @Override
         public void run()
         {
            EventManager.postEvent(event);
         }
      });
   }

   /**
    * Clear specified frame widgets.
    *
    * @param   frame
    *          Frame to clear.
    * @param   noPause
    *          Do not generate pause.
    * @param   clearAll
    *          If <code>true</code> then all widgets are cleared (the
    *          <b>CLEAR ALL</b> behavior).
    * @param   resolveBuffer
    *          When <code>true</code>, resolve the frame's screen buffer and return it.
    *          
    * @return   The frame's screen buffer, or <code>null</code> if the <code>resovelBuffer</code>
    *           flag is not set.
    */
   ScreenBuffer clearWorker(final Frame   frame, 
                                  boolean noPause, 
                            final boolean clearAll, 
                                  boolean resolveBuffer)
   {
      // must be called first!
      boolean res = honorConditionalUpDown(noPause);

      if (frame == null)
         return null;

      // DOWN frame is wrapped, nothing to do, it is already
      // paused/wrapped as necessary.
      if (frame.isDown() && res)
         return null;

      try
      {
         if (!noPause            &&
             frame.isNeedPause() &&
             frame.isVisible()   &&
             (frame.isRootFrame() || frame.isViewed()))
            pauseBeforeHide();
      }

      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate(frame);

         // let the original exception propagate
         throw ce;
      }

      finally
      {
         eventDrawingBracket(frame, new Runnable()
         {
            @Override
            public void run()
            {
               if (clearAll)
               {
                  // this represents a CLEAR FRAME ALL stmt call - in this case,
                  // we must clear the entire frame, not just the current iteration.
                  frame.clearWorker(clearAll);
               }
               else
               {
                  // this is the plain CLEAR FRAME stmt call - in this case, only
                  // the current iteration of a DOWN frame must be cleared
                  frame.clearCurrent();
               }

               if (!frame.isRedirected())
                  frame.repaint();
            }
         });
      }

      return resolveBuffer && frame.isRealized()
               ? getScreenBuffer(UiUtils.getWidgetIdAsInt(frame), -1, true) 
               : null;
   }

   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} and finally disables invalidation mode. The closing of the
    * event bracket and the disabling of invalidation mode (which can cause
    * drawing) occur even if the executed code throws an exception. This ensures
    * that the event processing and drawing will never get out of balance.
    * <p>
    * This version forces flushing when the invalidation mode is disabled. In
    * addition, the invalidation counter is saved at the beginning and restored
    * at the end of this method.
    *
    * @param    widget
    *           The widget which triggered this drawing.  If is <code>null</code>, the window 
    *           as returned by {@link WindowManager#findWindow} will be used; otherwise, the 
    *           widget's window will be used.
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>.
    */
   public boolean independentEventDrawingBracket(Widget widget, Runnable code)
   {
      boolean result = false;
      int     save   = tk.saveInvalidate(widget);

      try
      {
         result = eventDrawingBracket(widget, false, true, code);
      }

      finally
      {
         tk.restoreInvalidate(save);
      }

      return result;
   }
   
   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} and finally disables invalidation mode. The closing of the
    * event bracket and the disabling of invalidation mode (which can cause
    * drawing) occur even if the executed code throws an exception. This ensures
    * that the event processing and drawing will never get out of balance.
    * <p>
    * This version forces flushing when the invalidation mode is disabled.
    *
    * @param    widget
    *           The widget which triggered this drawing.  If is <code>null</code>, the window 
    *           as returned by {@link WindowManager#findWindow} will be used; otherwise, the 
    *           widget's window will be used.
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>.
    */
   public boolean eventDrawingBracket(Widget<?> widget, Runnable code)
   {
      return eventDrawingBracket(widget, false, true, code);
   }

   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} (or {@link #popTrash}) and finally disables invalidation
    * mode. The closing of the event bracket and the disabling of invalidation
    * mode (which can cause drawing) occur even if the executed code throws an
    * exception. This ensures that the event processing and drawing will never
    * get out of balance.
    * <p>
    * This version forces flushing when the invalidation mode is disabled.
    *
    * @param    widget
    *           The widget which triggered this drawing.  If is <code>null</code>, the window 
    *           as returned by {@link WindowManager#findWindow} will be used; otherwise, the 
    *           widget's window will be used.
    * @param    discard
    *           <code>true</code> to use <code>popTrash</code>,
    *           <code>false</code> to use <code>pop</code>.
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>. If
    *           <code>popTrash</code> is used, the result will always be
    *           <code>false</code>.
    */
   boolean eventDrawingBracket(Widget<?> widget, boolean discard, Runnable code)
   {
      return eventDrawingBracket(widget, discard, true, code);
   }

   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} (or {@link #popTrash}) and finally disables invalidation
    * mode. The closing of the event bracket and the disabling of invalidation
    * mode (which can cause drawing) occur even if the executed code throws an
    * exception. This ensures that the event processing and drawing will never
    * get out of balance.
    *
    * @param   widget
    *          The widget which triggered this drawing.  If is {@code null}, the window 
    *          as returned by {@link WindowManager#findWindow} will be used; otherwise, the 
    *          widget's window will be used.
    * @param   discard
    *          {@code true} to use {@code popTrash}, {@code false} to use {@code pop}.
    * @param   flush
    *          Force flushing when invalidation mode is disabled.
    * @param   code
    *          Logic to safely execute inside the event bracket.
    *
    * @return  The result as returned by {@code pop}. If {@code popTrash} is used, the result
    *          will always be {@code false}.
    */
   public boolean eventDrawingBracket(Widget<?> widget,
                                      boolean discard,
                                      boolean flush,
                                      Runnable code)
   {
      if (!discard)
      {
         tk.setInvalidate(true);
      }
      
      boolean selectWindow = !discard && widget != null; // intention
      boolean windowSelected = false;                    // truly selected ?
      Widget<?> window = null;
      
      if (selectWindow)
      {
         window = widget.parentOrSelf(TopLevelWindow.class);
         selectWindow = window != null;
         
         if (selectWindow)
         {
            tk.selectWindow(window.getId().asInt());
            windowSelected = true;
         }
         else
         {
            System.err.println(
               "INFO: Drawing with widget not attached to a window instance: " + widget);
         }
      }
      else if (!discard)
      {
         System.err.println("INFO: Drawing with null widget");
      }
      
      try
      {
         return eventBracket(discard, code);
      }
      finally
      {
         if (!discard)
         {
            tk.setInvalidate((selectWindow ? window : widget), false, flush);
            
            if (windowSelected)
            {
               tk.releaseWindow();
            }
         }
      }
   }

   /**
    * Execute the given code inside a new bracket of event processing which
    * starts with a {@link #push} and ends with a {@link #pop} even if the
    * executed code throws an exception. This ensures that the event processing
    * will never get out of balance.
    *
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>.
    */
   public boolean eventBracket(Runnable code)
   {
      return eventBracket(false, code);
   }

   /**
    * Execute the given code inside a new bracket of event processing which
    * starts with a {@link #push} and ends with a {@link #pop} or
    * {@link #popTrash} even if the executed code throws an exception. This
    * ensures that the event processing will never get out of balance.
    *
    * @param    discard
    *           <code>true</code> to use <code>popTrash</code>,
    *           <code>false</code> to use <code>pop</code>.
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>. If
    *           <code>popTrash</code> is used, the result will always be
    *           <code>false</code>.
    */
   public boolean eventBracket(boolean discard, Runnable code)
   {
      boolean result = false;

      push();

      try
      {
         code.run();
      }
      catch (RuntimeException re)
      {
         // allow these to propagate
         throw re;
      }
      catch (Exception exc)
      {
         // log and ignore these
         exc.printStackTrace(System.err);
      }
      finally
      {
         if (discard)
         {
            popTrash();
         }
         else
         {
            result = pop();
         }
      }

      return result;
   }

   /**
    * Process events and restore previous event queue.
    *
    * @return <code>true</code> if at least one frame related
    *         <code>PaintEvent</code> has been processed
    */
   private boolean pop()
   {
      try
      {
         return processEventsWorker();
      }
      finally
      {
         EventManager.popEventQueue();
      }
   }

   /**
    * Trash all events and restore previous event queue.
    */
   private void popTrash()
   {
      while (EventManager.getEvent() != null)
      {
         // eat the events
      }
      EventManager.popEventQueue();
   }

   /**
    * Save current event queue.
    */
   private void push()
   {
      EventManager.pushEventQueue();
   }

   /**
    * Get trigger nesting.
    *
    * @return  current trigger nesting level.
    */
   public int getTriggerNesting()
   {
      return triggerNesting;
   }

   /**
    * Get current state of the needPause flag. It tracks recent input and
    * other operations to tell application that pause is required.
    *
    * @return  Current state of the flag.
    */
   public static boolean needPause()
   {
      return getInstance().needPause;
   }

   /**
    * Perform pause before end of application.
    */
   @Override
   public void pauseBeforeEnd()
   {
      Window wnd = WindowManager.getDefaultWindow();
      if (isNeedPause() || wnd.isMessageNeedPause() || wnd.isVisible())
      {
         pause(timePBH, PBE_DEFAULT_TEXT, false, wnd);
      }
      else
      {
         wnd.hideMessage(false);
      }
   }

   /**
    * Callback method that notifies the application about Ctrl-C keystroke
    * and returns the decision the application has made about the disposition
    * of the Ctrl-C keystroke.
    * <p>
    * This method is called in the context of the key reader thread.
    * <p>
    * This implementation's goal is to properly generate the STOP condition.
    *
    * @return Ctrl-C keystroke disposition value as one of the predefined
    *         action code constants.
    */
   @Override
   public int watchCtrlC()
   {
      if (authMode)
         return Watcher.CLEAR_AND_QUEUE;

      pauseCleanup();
      /* TODO set correct window */
      setNeedPause(false, WindowManager.resolveWindow());

      ThinClient tc = getInstance();
      
      if (tc.dedicatedThread != null)
      {
         // running in the server, directly interrupt the user's thread
         CriticalSectionManager.interrupt(tc.dedicatedThread);
      }
      else
      {
         // standard client (running remotely from the server in its own JVM)
         // the Control instance is used to handle the interruption
         // asynchronously (including remoting the interruption to the server
         // side when needed)
         Control.interrupt();
      }
      
      return Watcher.IGNORE;
   }

   /**
    * Set authentication mode flag. When authentication mode is enabled,
    * Ctrl-C keystroke is not passed to {@link Control} but processed as a
    * normal sequence.
    *
    * @param    auth
    *           New state of the flag.
    */
   public void setAuthMode(boolean auth)
   {
      authMode = auth;
   }

   /**
    * Looks up the redirected terminal instance for a given stream.
    *
    * @param    out
    *           The stream being queried.
    * @return   instance of <code>RedirectedTerminal</code> associated with
    *           the stream.
    */
   public static RedirectedTerminal redirectedTerminalForStream(Stream out)
   {
      ThinClient tc = ThinClient.getInstance();
      int sid = tc.sd.idFromStream(out);
      Integer key = Integer.valueOf(sid);
      RedirectedTerminal rt = tc.redirections.get(key);
      return rt;
   }

   /**
    * Notify the registered object that the given stream is being closed.
    * <p>
    * This listener is used to catch the event of the stream closure to flush
    * any output that may remain sitting in the redirected terminal buffer.
    *
    * @param    out
    *           The stream being closed.
    */
   @Override
   public void notifyClose(Stream out)
   {
      flushStream(out);
   }

   /**
    * Notify registered object that given stream reached end of life cycle and
    * all references to it can be safely discarded.
    * <p>
    * This notification is used to unbind frames which were switched from
    * interactive mode to redirected mode and restore their initial state.
    *
    * @param    out
    *           The stream being closed.
    */
   @Override
   public void notifyUnbind(Stream out)
   {
      // TODO check exact window
      Widget[] c = UiUtils.getTopLevelComponents(WindowManager.getDefaultWindow());

      int id = idFromStream(out);

      for (Widget comp : c)
      {
         if (comp instanceof Frame)
         {
            ((Frame) comp).unbind(id);
         }
      }
   }

   /**
    * Notify the registered object that the given stream is advancing page.
    * <p>
    * This listener is used to catch the event to flush
    * any output that may remain sitting in the redirected terminal buffer.
    *
    * @param    out
    *           The stream being paged
    */
   @Override
   public void notifyAdvance(Stream out)
   {
      int sid = sd.idFromStream(out);

      boolean switched = currentStream == sid;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(sid, true);
      }

      switchFrames(sid, -1);

      if (!switched)
      {
         switchOutput(tempOut, false);
      }
   }

   /**
    * Notify the registered object that the given stream is flushing.
    * <p>
    * This listener is used to catch the event to flush
    * any output that may remain sitting in the redirected terminal buffer.
    *
    * @param    out
    *           The stream being flushed.
    */
   @Override
   public void notifyFlush(Stream out)
   {
      flushStream(out);
   }

   /**
    * Returns the stream ID associated with a given stream.
    *
    * @param    s
    *           stream to operate on
    *
    * @return   The stream ID or <code>-1</code> if no stream ID is associated
    *           with the given stream.
    */
   public int idFromStream(Stream s)
   {
      return sd.idFromStream(s);
   }

   /**
    * Returns the stream associated with a given stream.ID.
    *
    * @param    sid
    *           stream ID
    *
    * @return   The stream instance or <code>null</code> if stream ID unknown.
    */
   public Stream streamFromId(int sid)
   {
      return sd.streamFromId(sid);
   }

   /**
    * Request indexes of selected items in RadioSet, ComboBox or SelectionList
    * widgets.
    *
    * @param   widgetId
    *          Widget ID.
    *
    * @return  array of 0-based indexes of selected items or <code>null</code>
    *          if specified widget does not exists or does not support such a
    *          request.
    */
   @Override
   public int[] getSelection(int widgetId)
   {
      Widget widget = getWidget(widgetId);

      if (!(widget instanceof ControlSetComponent))
         return null;

      return ((ControlSetComponent) widget).getSelection();
   }

   /**
    * Request HELP application execution on the server.
    */
   private void requestHelp()
   {
      Window wnd = WindowManager.resolveWindow();
      
      // not initialized
      if (server == null)
      {
         return;
      }

      // recursive help
      if (helpActive)
      {
         tk.beep();
         return;
      }

      // flag recursion
      helpActive = true;

      // the following code arranges the safe help application invocation
      // in a way that is similar to triggers

      Widget inFocus = UiUtils.getCurrentFocus();

      // save state variables
      Frame focusedFrame  = (Frame) UiUtils.locateFrame(inFocus);
      int focusedFrameId = (focusedFrame == null ? -1 : UiUtils.getWidgetIdAsInt(focusedFrame));
      EventList save = currentEventList;

      String status = wnd.getStatusText();

      boolean needPause = false;
      boolean canHide   = false;
      boolean inScope   = false;

      boolean curvis = true;

      curvis = tk.setCursorStatus(true);

      if (focusedFrame != null)
      {
         needPause = focusedFrame.isNeedPause();
         canHide   = focusedFrame.isCanHide();
         inScope   = focusedFrame.isAlive();
      }

      // save visibility state of all existing frames
      int[] frames = UiUtils.getFrameIDs(wnd);
      boolean[] framesV = new boolean[frames.length];

      for (int i = 0; i < frames.length; i ++)
      {
         Frame f = getFrame(frames[i]);
         framesV[i] = f.isVisible();
         f.setProtected(true);
      }

      // save message lines
      // TODO: When not in tinyInput(), this refreshes the message area (redraws whatever
      //       content was there when this was called).  Is that correct behavior?
      // TODO: This implementation is too hard coded to the message area internals. Reduce
      //       this dependency and delegate more processing to the message area itself.
      MessageArea msgArea  = wnd.getMessageArea();
      String[]    messages = msgArea.getMessageTexts();
      Color[]     colors   = msgArea.getMessageColors();

      // clear messages and leave only the message input text
      if (wnd.isTinyInputActive())
      {
         // When HELP is called while in tinyInput() on ChUI this clears all messages
         // except for the one being used for tinyInput(). When in tinyInput() in GUI,
         // this clears all messages since in GUI mode the editing is done in a modal
         // popup window and the text is not output to the message area.
         // TODO: the GUI behavior has not been tested for compatibility.
         
         // in ChUI, this will match an index in the messages/colors arrays, but in GUI
         // this will be some large number that is outside of the array length  
         int til = wnd.getTinyInputLine();
         
         int nlines = messages.length;

         for (int i = 0; i < nlines; i ++)
         {
            if (i != til)
            {
               messages[i] = null;
               colors[i]   = null;
            }
         }
      }

      // call the help application
      int saveLevel = tk.saveInvalidate(inFocus);

      int val = triggerNestingByWaitFor.pop();
      triggerNestingByWaitFor.push(val + 1);

      setTriggerNesting(triggerNesting + 1);

      server.help(lastKey, getEditableClientScreenBuffers(getScreenBuffer(focusedFrameId, -1)));

      setTriggerNesting(triggerNesting - 1);

      val = triggerNestingByWaitFor.pop();
      triggerNestingByWaitFor.push(val - 1);

      if (triggerNesting == 0)
         wnd.setAutoCleared(false);

      tk.restoreInvalidate(saveLevel);

      int[] framesNow = UiUtils.getFrameIDs(wnd);

      // restore cursor visibility
      tk.setCursorStatus(curvis);

      // mark all HELP related frames for hiding
      int[] fresh = Utils.diffIntArrays(framesNow, frames);

      for (int i = 0; i < fresh.length; i ++)
      {
         Frame x = getFrame(fresh[i]);
         x.setCanHide(true);
         x.setAlive(false);
      }

      // restore existing frames visibility
      tk.setInvalidate(true);

      // prepare for binary search
      Arrays.sort(framesNow);
      
      for (int i = 0; i < frames.length; i ++)
      {
         // check if frame still exists. 
         if (Arrays.binarySearch(framesNow, frames[i]) < 0)
         {
            // if not skip.
            continue;
         }
         
         Frame f = getFrame(frames[i]);

         f.setProtected(false);

         if (f.isVisible() == false && framesV[i] == true)
            viewWorker(f, null, null, true, UiUtils.getWidgetIdAsInt(wnd));
      }

      // restore message lines
      msgArea.setMessageText(messages, colors);

      Frame currFrame = (Frame) UiUtils.locateFrame(UiUtils.getCurrentFocus());

      // restore state variables, only in ChUI - in GUI, help is in a separate window
      if (isChui() && focusedFrame != null)
      {
         if (focusedFrame.isVisible())
         {
            if (currFrame == null || !currFrame.config().overlay)
            {
               placeFrameInt(focusedFrame, true);
            }
         }

         focusedFrame.setNeedPause(needPause);
         focusedFrame.setCanHide(canHide);
         focusedFrame.setAlive(inScope);
      }

      currentEventList = save;

      // restore the focus
      if (inFocus != UiUtils.getCurrentFocus() && inFocus != null)
         inFocus.requestFocus();

      tk.setInvalidate(WindowManager.getDefaultWindow(), false, false);

      // restore the status line
      wnd.status(status);

      // unflag recursion
      helpActive = false;
   }

   /**
    * Reports if the terminal is currently redirected to a stream.
    *
    * @return   <code>true</code> if the terminal is redirected to a stream,
    *           <code>false</code> if the terminal is interactive.
    */
   public boolean isRedirected()
   {
      return (currentStream != -1);
   }

   /**
    * Reports if the terminal is currently redirected to the specified stream.
    *
    * @param    s
    *           a stream to to check the redirection for
    * @return   <code>true</code> if the terminal is redirected to the stream,
    *           <code>false</code> if the terminal is interactive.or
    *           redirected to a different stream
    */
   public boolean isRedirected(Stream s)
   {
      int sid = sd.idFromStream(s);

      return (sid != -1 && currentStream == sid);
   }

   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and close/
    * cleanup resources for the redirected terminal associated with
    * <code>oldId</code> (unless <code>oldId</code> is -1 in which case
    * nothing needs to be closed).
    *
    * @param   oldId
    *          The previously redirected stream for which the redirected
    *          terminal is to be closed or -1 for no close/cleanup action.
    * @param   newId
    *          The stream for which a redirected terminal is to be setup for
    *          redirection or -1 to remove redirections and activate the
    *          interactive terminal.
    */
   @Override
   public void redirectOutput(int oldId, int newId)
   {
      // If batch mode is active the redirected output deactivate error generation for this mode.
      if (OutputManager.inBatchMode())
      {
         OutputManager.activateBatch(false);
      }

      if (newId == oldId)
      {
         return;
      }

      // flush stream if switching from non-terminal stream to terminal
      if (oldId != -1 && newId == -1)
      {
         Stream out = sd.streamFromId(oldId);

         if (out != null)
            flushStream(out);
      }

      int old = switchWorker(newId);

      // all we need to do is delete it from the map
      if (oldId != -1)
      {
         if (oldId != old)
         {
            throw new IllegalArgumentException("Attempt to close non-active" +
                                               " terminal redirection.");
         }

         redirections.remove(Integer.valueOf(oldId));
      }
      
      if (newId == -1 && OutputManager.inBatchMode())
      {
         OutputManager.activateBatch(true);
      }
   }

   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * redirected terminal or it may be -1 if the interactive terminal is in
    * use.
    * <p>
    * If redirection has not yet been established for the given stream, it
    * will be setup.
    *
    * @param    newId
    *           The stream for which a redirected terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    * @param    clean
    *           <code>true</code> if this stream will need deferred cleanup.
    *
    * @return   The id of the previous terminal or -1 if the previous
    *           terminal was interactive.
    */
   public int switchOutput(int newId, boolean clean)
   {
      return switchOutput(newId, clean, false);
   }

   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * redirected terminal or it may be -1 if the interactive terminal is in
    * use.
    * <p>
    * If redirection has not yet been established for the given stream, it
    * will be setup.
    *
    * @param    newId
    *           The stream for which a redirected terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    * @param    clean
    *           <code>true</code> if this stream will need deferred cleanup.
    * @param    endOut
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    *
    * @return   The id of the previous terminal or -1 if the previous
    *           terminal was interactive.
    */
   @Override
   public int switchOutput(int newId, boolean clean, boolean endOut)
   {
      if (endOut)
         endOutput(newId);

      if (newId != -1 && clean)
      {
         // remember that this stream may need deferred cleanup
         deferredCleanup.add(newId);
      }

      return switchWorker(newId);
   }

   /**
    * Called when a stream is closed to ensure that any terminal resources
    * associated with that id are cleaned up.  While this will be called
    * for any output stream close, actual cleanup processing is only
    * triggered when the given stream id was used at least once via the
    * {@link #switchOutput} interface.
    * <p>
    * This should never be called when the redirection is currently active.
    *
    * @param    id
    *           The stream needing cleanup.
    */
   @Override
   public void cleanupPseudoTerminal(int id)
   {
      Integer key = Integer.valueOf(id);

      if (deferredCleanup.remove(key))
      {
         if (id == currentStream)
         {
            throw new IllegalArgumentException("Attempt to close currently " +
                                               "active terminal redirection.");
         }
         // we need cleanup for this id
         redirections.remove(key);
      }
   }

   /**
    * Obtains the next widget in the frame's tab order from the client.
    *
    * @param    frameId
    *           frame where the widget is placed
    *
    * @param    widgetId
    *           target widget which is being queried
    *
    * @return   id of the widget or<code>-1</code>.
    */
   @Override
   public int nextTabItem(int frameId, int widgetId)
   {
      Widget widget = getWidget(widgetId);
      if (widget == null)
      {
         return -1;
      }

      Widget[] list = widget.parent().widgets();
      WidgetId wid = widget.getId();
      
      for (int i = 0; i < list.length; i ++)
      {
         if (UiUtils.lookupWidgetId(list[i]).equals(wid))
         {
            if (i == list.length - 1)
            {
               return -1;
            }

            // TO-DO: doesn't it need a check for isFocusTraversable()?
            return UiUtils.lookupWidgetIdAsInt(list[i + 1]);
         }
      }

      return -1;
   }

   /**
    * Reset buffer size manager to its initial state.
    */
   @Override
   public void resetBufferSizeManager()
   {
      BufferSizeManager.getInstance().reset();
   }

   /**
    * Assign VALUE CHANGED event.
    *
    * @param   input
    *          The event to assign.
    */
   public static void setValueChangedEvent(KeyInput input)
   {
      getInstance().postponedValueChanged  = input;
   }

   /**
    * Remove frame from the registry(). Here removed all local references to the
    * frame and frame is registered in the list of frames which should be
    * removed at server side. Actual removal of frames at server side is
    * performed when processing of server request is done and result of
    * processing is returned to server.
    *
    * @param   frame
    *          The frame to remove.
    */
   public static void removeFrame(Frame frame)
   {
      if (frame == null)
         return;
      
      ThinClient tc = getInstance();

      if (UiUtils.hasConfig(frame) && tc.server != null)
         tc.removedFrames.add(frame.config().id.asInt());

      tc.registry().removeFrame(frame);

      if (tc.lastExposedFrame == frame)
         tc.lastExposedFrame = null;
   }
   
   /**
    * Remove the specified widget from the registry, and from its parent frame.
    * 
    * @param    widget
    *           The widget to be removed.
    */
   public static void removeWidget(Widget<?> widget)
   {
      if (widget == null)
      {
         return;
      }
      
      ThinClient tc = getInstance();
      widget.destroy();
      tc.registry().removeWidget(widget.config().id);
   }

   /**
    * Remove frame from the registry. Here removed all local references to the
    * frame and frame is registered in the list of frames which should be
    * removed at server side. Actual removal of frames at server side is
    * performed when processing of server request is done and result of
    * processing is returned to server.
    *
    * @param   frame
    *          The frame to remove.
    */
   public static void wrapFrame(Frame frame)
   {
      if (frame == null)
         return;
      
      ThinClient tc = getInstance();

      if (UiUtils.hasConfig(frame) && tc.server != null)
         tc.wrappedFrames.add(frame.config().id.asInt());
   }

   /**
    * Get status of refresh quirk (odd refreshing behavior in some terminal
    * types).
    *
    * @return   <code>true</code> if refreshing quirk is active.
    */
   public static boolean isRefreshQuirkActive()
   {
      return getInstance().refreshQuirk;
   }
   
   /**
    * Setup tooltip for widget
    * 
    * @param widget
    *        the widget the tooltip is configured for
    */
   public static void setupTooltip(WidgetWithConfig<? extends WidgetConfig> widget)
   {
      // frame, window, alert-box and dialog box can not have a tooltip
      if (!(widget instanceof AbstractWidget<?>) ||
          widget instanceof Frame ||
          widget instanceof TitledWindow)
      {
         return;
      }
      
      AbstractWidget<?> awidget = (AbstractWidget<?>) widget; 
      WidgetConfig wcfg = widget.config(); 
      if (!(wcfg instanceof BaseConfig))
      {
         return;
      }
      
      BaseConfig cfg = (BaseConfig) wcfg;
      ScreenDriver<?> sd = OutputManager.getDriver();
      if (!(sd instanceof GuiDriver))
      {
         return;
      }
      
      GuiDriver gd = (GuiDriver) sd;
      Window<?> wnd = WindowManager.resolveWindow(awidget);
      GuiFontResolver gf = new GuiFontResolver(wnd, gd, cfg, false);
      
      // register tooltip action listener
      int winID = UiUtils.getWidgetIdAsInt(wnd);
      gd.selectWindow(winID);
      gd.registerTooltipWorker(awidget, gf.resolveFontNum(), ToolTip.isActive(cfg.tooltip));
      gd.releaseWindow();
   }

   /**
    * Resets the changed state for all widgets of the frame into unchanged.
    *
    * @param   frameId
    *          ID of frame this request came for
    */
   private void resetChanged(int frameId)
   {
      getFrame(frameId).resetChanged();
   }

   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * redirected terminal or it may be -1 if the interactive terminal is in
    * use.
    * <p>
    * If redirection has not yet been established for the given stream, it
    * will be setup.
    *
    * @param    newId
    *           The stream for which a redirected terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    *
    * @return   The id of the previous terminal or -1 if the previous
    *           terminal was interactive.
    */
   private int switchWorker(int newId)
   {
      Integer key = Integer.valueOf(newId);

      if (newId == -1)
      {
         Window.enableSpecial();

         // switch to interactive mode
         tk.switchOutput(null);

         honorPBH = true;
      }
      else if (redirections.containsKey(key))
      {
         Window.disableSpecial();

         // pre-existing redirection target
         RedirectedTerminal rt = redirections.get(key);
         tk.switchOutput(rt);

         honorPBH = false;
      }
      else
      {
         Window.disableSpecial();

         // first time redirection
         Stream out = sd.streamFromId(newId);
         RedirectedTerminal rt = new RedirectedTerminal(out, 512, 128);

         redirections.put(key, rt);
         tk.switchOutput(rt);

         honorPBH = false;
      }

      // remember the current stream id and return the old id
      int oldId = currentStream;
      currentStream = newId;

      return oldId;
   }

   /**
    * Switch the terminal into interactive mode if it is redirected. This is
    * just a convenience function.
    *
    * @return   The terminal id of the redirected terminal or -1 if the
    *           terminal is already interactive.  Used with
    *           {@link #restoreRedirection} to restore the terminal when
    *           done with interactive mode.
    */
   public int forceInteractive()
   {
      int tempTerm = -1;

      if (isRedirected())
      {
         // set up temporary interactive terminal
         tempTerm = switchOutput(-1, false);
      }

      return tempTerm;
   }

   /**
    * Restore the terminal to redirected mode from interactive mode, if the
    * terminal to switch to is not -1.  This is just a convenience function.
    *
    * @param    tempTerm
    *           The terminal id of the redirected terminal that was previously
    *           removed using {@link #forceInteractive} or -1 if there was no
    *           previous redirected terminal.
    */
   public void restoreRedirection(int tempTerm)
   {
      if (tempTerm != -1)
      {
         // take down temporary interactive terminal, restore redirection
         switchOutput(tempTerm, false);
      }
   }

   /**
    * Processes all outstanding events without blocking for user input.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PROCESS EVENTS </b>
    *
    * @return <code>true</code> if at least one frame related
    *         <code>PaintEvent</code> has been processed
    */
   private boolean processEventsWorker()
   {
      Event event;

      Set<TopLevelWindow> syncPending = new HashSet<>();
      boolean losingFocus = false;
      boolean framePaint  = false;
      boolean gainingFocus = false;

      while ((event = EventManager.getEvent()) != null)
      {
         // synchronization service events
         if (event instanceof SyncEvent)
         {
            Widget<?> source = event.source();
            int wid = UiUtils.getWidgetIdAsInt(source);
            Widget<?> ancestor = source.ancestor();

            if (ancestor == null)
            {
               // the widget is detached!
               String msg = "SYNC: Widget %s with id [%d] has been removed!";
               System.err.println(String.format(msg, source.getClass().getName(), wid));
               continue;
            }

            if (!(ancestor instanceof TopLevelWindow))
            {
               int aid = UiUtils.getWidgetIdAsInt(ancestor);
               // the widget is not attached to a window!

               String msg = "SYNC: Widget %s with id [%d] is attached to a %s with id [%d] " +
                            "instead of a window!";
               System.err.println(String.format(msg, 
                                                source.getClass().getName(), wid,
                                                ancestor.getClass().getName(), aid));
               continue;
            }
            
            syncPending.add((TopLevelWindow) ancestor);
            continue;
         }
         
         if ((event instanceof ServerEvent || event instanceof ResourceEvent) && 
             currentEventList != null)
         {
            processProgressEvent(event);
            continue;
         }
         
         if (event instanceof MouseEvt           ||
             event instanceof WindowActivated)
         {
            processProgressEvent(event);

            // processProgressEvent() already dispatches the event to the corresponding widget
            // so break this iteration early
            continue;
         }

         if (event instanceof PortableMouseEvent || event instanceof WindowEvent)
         {
            processProgressEvent(event);

            // portable mouse triggers and window triggers may cause the immediateException
            // to be set
            if (immediateException != null)
            {
               throw immediateException;
            }

            // processProgressEvent() already dispatches the event to the corresponding widget
            // so break this iteration early
            continue;
         }

         if (event instanceof InvocationEvent)
         {
            ((InvocationEvent) event).getRunnable().run();
            continue;
         }

         Widget src = event.source();

         // keyboard events
         if (event instanceof KeyInput && currentEventList != null)
         {
            KeyInput ki = (KeyInput) event;

            processProgressEvent(ki);

            // block default processing for this event if
            // the widget in focus is disabled
            // TO-DO: why not the event source???
            if (!ki.isConsumed())
            {
               Widget inf = ki.source();
               Widget infFrame = UiUtils.locateFrame(inf);

               // the source component may be the window, so null frame is
               // allowed
               if (inf != null &&
                   !(inf instanceof TopLevelWindow) &&
                   !(inf.navigable() &&
                     inf.isVisible() &&
                     (infFrame == null ||
                      infFrame.isVisible())))
               {
                  // Do not consume some of event targeted to BROWSE,
                  // it receives them regardless from the state.
                  // TODO browse is able to react to the *most* of events, so the list
                  // below should be significantly widened.
                  // Note that this has no side effects visible by application
                  // since these events are used by browse internally.

                  if (!(src instanceof Browse &&
                        (ki.actionCode() == Keyboard.SE_VALUE_CHANGED ||
                         ki.actionCode() == Keyboard.SE_ENTRY ||
                         ki.actionCode() == Keyboard.KA_END ||
                         ki.actionCode() == Keyboard.KA_HOME)))
                  {
                     // this consume doesn't mean the trigger returned
                     // no-apply - it just stops further processing of the
                     // event
                     ki.consume();
                  }
               }

               if (ki.actionCode() == Keyboard.SE_VALUE_CHANGED &&
                   !ki.isSpecial())
               {
                  Frame frame = (Frame) UiUtils.locateFrame(src);

                  if (frame != null)
                     frame.setFocusAttractor(true);
               }
            }

            if (ki.actionCode() == Keyboard.KA_GO && !triggerResultNoApply.peek())
            {
               Widget inf = ki.source();
               Widget frame = UiUtils.locateFrame(inf);

               if (frame instanceof Frame && frame != ki.source())
               {
                  // the following widgets will trigger GO for frame, even if
                  // the event was consumed
                  boolean override = inf instanceof SelectionList     ||
                                     inf instanceof SelectionListBody ||
                                     inf instanceof RadioSet          ||
                                     inf instanceof Button            ||
                                     inf instanceof ComboBox          ||
                                     inf instanceof Editor            ||
                                     inf instanceof Browse;

                  if (!ki.isConsumed() || override)
                  {
                     KeyInput res = new KeyInput(frame,
                                                 ki.actionCode(),
                                                 ki.keyCode(),
                                                 true);
                     EventManager.postEvent(res);
                  }
               }
            }

            // if this is a NEXT/PREV-FRAME event, generate a new event and
            // send it to the parent frame, if exists
            if (!triggerResultNoApply.peek() &&
                (ki.actionCode() == Keyboard.SE_NEXT_FRAME ||
                 ki.actionCode() == Keyboard.SE_PREV_FRAME))
            {
               Widget inf = ki.source();
               Widget frame = UiUtils.locateFrame(inf);

               if (frame instanceof Frame && frame != ki.source())
               {
                  KeyInput res = new KeyInput(frame,
                                              ki.actionCode(),
                                              ki.keyCode(),
                                              true);
                  EventManager.postEvent(res);
               }
            }

            // no default processing for consumed events
            if (ki.isConsumed())
            {
               continue;
            }
         }

         // generate CHOOSE event
         if (event instanceof ItemEvent)
         {
            ItemEvent evt = (ItemEvent) event;
            KeyInput res = null;

            if (evt.selected())
            {
               res = new KeyInput(evt.source(), EventType.KEY_PRESSED,
                                  Keyboard.SE_CHOOSE);
               EventManager.postEvent(res);
            }
         }

         // pre-processing of FOCUS_LOST and FOCUS_GAINED events
         if (event instanceof FocusEvent)
         {
            switch (event.id())
            {
               case EventType.FOCUS_LOST:
                  losingFocus = true;
                  break;
               case EventType.FOCUS_GAINED:
                  gainingFocus = true;
                  break;
            }
         }

         // monitoring Paint events
         if (event instanceof PaintEvent)
         {
            if (src instanceof Frame)
               framePaint = true;
         }

         if (alertBoxActive == 0)
         {
            if (event instanceof KeyInput)
            {
               int actionCode = ((KeyInput) event).actionCode();
               if (actionCode != Keyboard.SE_LEAVE &&
                   actionCode != Keyboard.SE_ENTRY)
               {
                  lastActionCode = actionCode;
               }
            }
            else if (!(event instanceof PaintEvent) &&
                     !(event instanceof FocusEvent))
            {
               lastActionCode = 0;
            }
         }

         // processing of all events by CHARVA
         try
         {
            TitledWindow<?> window = null;
            if (src.parent() == null)
            {
               if (src instanceof TopLevelWindow)
               {
                  window = (TopLevelWindow<?>) src;
               }
               else
               {
                  int wid = OutputManager.getDriver().getPrimitives().getWindowId();
                  if (wid != -1)
                  {
                     window = WindowManager.findWindow(wid);
                  }
               }
            }
            else
            {
               window = (TopLevelWindow<?>) src.parentOrSelf(TopLevelWindow.class);
            }
            
            // The window may have been already gone. 
            // For example when GUI alert-box button is pressed, 
            // button paint event may arrive when the top-level
            // window is gone.
            if (window != null && window.isDisplayed())
            {
               window.processEvent(event);
            }
         }

         catch(ConditionException e)
         {
            // condition exceptions generated during event processing must
            // be handled
            throw e;
         }

         catch (NullPointerException npe)
         {
            throw npe;
         }

         catch (Exception e)
         {
            e.printStackTrace(System.err);
            continue;
            // throw new RuntimeException("")
         }

         // post-processing of FOCUS_LOST event
         if (losingFocus)
         {
            losingFocus = false;

            // redraw the widget just lost the focus

            if (src.isVisible())
            {
               src.setHighlighted(false);
               //src.repaint();

               Frame srcFrame = (Frame) UiUtils.locateFrame(src);
               if (srcFrame != null)
               {
                  srcFrame = srcFrame.getRootFrame();
               }
               
               boolean repaint = src.parent() != null && srcFrame == src.window().getTopVisibleFrame();

               if (!repaint)
               {
                  AlertBox box = UiUtils.locateAlertBox(src);
                  if (box != null && box.isVisible())
                     repaint = true;
               }

               if (repaint)
               {
                  independentEventDrawingBracket(src, new Runnable()
                  {
                     @Override
                     public void run()
                     {
                        src.repaint();
                     }
                  });
               }
            }
         }

         if (gainingFocus)
         {
            gainingFocus = false;

            if (src.isVisible())
            {
               src.setHighlighted(true);
            }
         }

         if (immediateException != null)
         {
            throw immediateException;
         }
      }

      for (TopLevelWindow w : syncPending)
      {
         w.processEvent(new SyncEvent(w));
      }

      return framePaint;
   }

   /**
    * Worker to process the given key as a mnemonic (ALT or simple key), for normal widgets 
    * (not menus).
    * 
    * @param    evt
    *           The event to process.
    * @param    key
    *           The associated key.
    * @param    src
    *           The event's source.
    *           
    * @return   <code>true</code> if the key was processed as a mnemonic by a widget.
    */
   public boolean processMnemonic(KeyInput evt, int key, Widget<?> src)
   {
      if (src instanceof SelectionListBody)
      {
         src = ((SelectionListBody) src).parent(SelectionList.class);
      }
      
      if (!isChui()                  &&
          evt != null                &&
          evt.isRealKey()            &&
          src != null                &&
          src.getId() != null        &&
          (Keyboard.isMnemonic(evt.keyCode()) ||
           !(src instanceof FillIn        ||
             src instanceof Editor        ||
             src instanceof SelectionList ||
             src instanceof ComboBox      ||
             src instanceof Browse)) &&
          src.enclosingFrame().isPresent())
      {
         // a. ALT-<key> mnemonics are available only in GUI; also, their default behaviour is
         // available only if no trigger (of any kind) can intercept them
         // b. <key> mnemonics will fire only if:
         //   1. the widget can't handle the key on its own (FILL-IN, EDITOR, SELECTION-LIST,
         //      COMBO-BOX, BROWSE)
         //   2. there is no trigger for this key

         Frame frm = src.enclosingFrame().get();
         int frameId = frm.getId().asInt();
         int widgetId = src.getId().asInt();
         TriggerMatch match = new TriggerMatch(key);
         currentEventList.lookup(frameId, widgetId, true, true, match);

         if (match.triggerId == -1)
         {
            Widget[] wt = new Widget[1];
            Widget wsrc = src;
            
            eventDrawingBracket(src, () -> { wt[0] = frm.processFrameMnemonic(key, wsrc); });
            
            // search starts from the root frame
            Widget target = wt[0];

            // following widgets do not switch focus on mnemonic: BUTTON
            if (target != null && !target.hasFocus() && !(target instanceof Button))
            {
               Frame newFrm = (Frame) target.enclosingFrame().get();
      
               // order of events
               // 1. FOCUS:LEAVE - if NO-APPLY, then abort
               // 2. FOCUS:FRAME:LEAVE - if NO-APPLY, then don't send entry to target:FRAME or
               //    target, but set focus to target
               // 3. target:FRAME:ENTRY - if NO-APPLY,  then don't send entry to target, 
               //    but set focus to target
               // 4. target:ENTRY - if NO-APPLY, then set focus to next focusable component. 
               //    this includes nested frames and tab order
               
               if (sendLeave(src, true))
               {
                  boolean focusTarget = false;
                  
                  if (sendLeave(frm, true))
                  {
                     if (sendEntry(newFrm, true))
                     {
                        if (sendEntry(target, true))
                        {
                           focusTarget = true;
                        }
                        else
                        {
                           // TODO: next focusable component, including nested frames and tab order
                        }
                     }
                     else
                     {
                        focusTarget = true;
                     }
                  }
                  else
                  {
                     focusTarget = true;
                  }
                  
                  if (focusTarget)
                  {
                     target.requestFocus();
                  }
               }
               else
               {
                  // focus remains on source
               }
            }
            
            if (target != null)
            {
               // no more processing for this event
               evt.consume();
               return true;
            }
         }
      }
      
      return false;
   }
   
   /**
    * Handle event as Progress event.
    * <p>
    * Precedence order for honoring Progress key events:
    * <ol>
    *    <li> In the accelerator list for the current window.
    *    <li> In the event list for an active user interface trigger.
    *    <li> An active pop-up menu key.
    *    <li> An active menu mnemonic key.
    *    <li> An active field-level (button) mnemonic key (Windows only).
    *    <li> A special Progress internal key (built-in key function, like
    *         GO).
    *    <li> A key associated with a key function in the PROTERMCAP file
    *         (UNIX only).
    *    <li> A standard input key (A, B and so on).
    * </ol>
    * <p>
    * If any exit condition is found, the <code>isGoPending</code> flag is
    * set.
    *
    * @param    event
    *           Event to handle.
    */
   public void processProgressEvent(Event event)
   {
      if (event instanceof ResourceEvent)
      {
         ResourceEvent revt = (ResourceEvent) event;
         
         int evtId = revt.id();
         long resourceId = revt.resourceId();

         if (checkBreak(resourceId, evtId))
         {
            isGoPending++;
         }

         // invoke the trigger directly; RETURN NO-APPLY is ignored
         invokeTriggers(event.source(), resourceId, evtId, false);
         
         return;
      }
      
      if (event instanceof ServerEvent)
      {
         ServerEvent srv = (ServerEvent) event;
         
         int evtCode = Keyboard.eventCode(srv.getName());
         long resourceID = srv.resourceId();
         
         if (checkBreak(resourceID, evtCode))
         {
            isGoPending++;
         }
         
         // invoke event
         invokeOnServer(() -> server.invokeServerEvent(srv.id()));
         
         return;
      }
      
      if (event instanceof MouseEvt)
      {
         MouseEvt evt = (MouseEvt) event;
         
         // apply the event
         TopLevelWindow window = (TopLevelWindow) evt.source();
         Widget<?> mouseSource = evt.getExplicitSource();

         if (mouseSource == null)
         {
            mouseSource = window.findMouseSource(evt);
         }

         if (mouseSource == null)
         {
            mouseSource = window;
         }

         Widget<?> legacyMouseSource;

         if (mouseSource instanceof Label)
         {
            Frame temp = UiUtils.locateFrame(mouseSource);
            legacyMouseSource = temp != null ? temp : mouseSource.getLegacyWidget();
         }
         else
         {
            legacyMouseSource = mouseSource.getLegacyWidget();
         }

         // find the actual focus target,
         // label widget is interesting, when it is the mouse source the focus target is its
         // frame and not its associated widget
         Widget newFocus = legacyMouseSource;

         // focus the frame when the focus target itself is not focusable
         if (newFocus != null && !(newFocus instanceof Frame) && !newFocus.focusTraversable())
         {
            newFocus = UiUtils.locateFrame(legacyMouseSource);
         }

         // fail over to the window
         if (newFocus == null)
         {
            newFocus = legacyMouseSource != null ?
                       legacyMouseSource.parentOrSelf(TopLevelWindow.class) :
                       mouseSource.parentOrSelf(TopLevelWindow.class);
         }

         // check for leaving the drop-down, it should be done early enough to get
         // proper currently focused widget
         Widget currFocus = UiUtils.getCurrentFocus();
         Widget oldFocus = currFocus != null ? currFocus.getLegacyWidget() : null;
         Frame oldFrame = oldFocus != null ? UiUtils.locateFrame(oldFocus) : null;

         boolean maySwitchFocus = false;
         
         // get the mouse event type
         int evtID = evt.getEvent().getID();
         // exclude mouse move events from consideration
         if (evtID == MouseEvt.MOUSE_PRESSED ||
             evtID == MouseEvt.MOUSE_RELEASED ||
             evtID == MouseEvt.MOUSE_CLICKED)
         {
            maySwitchFocus = (mouseSource != null && evtID == MouseEvt.MOUSE_PRESSED);
            
            if (mouseSource != null && !mouseSource.isRMBEntrySupported())
            {
               maySwitchFocus = maySwitchFocus && evt.isLeftButton();
            }
         }

         window.processEvent(evt);

         List<PortableMouseEvent> mouseEvents = Collections.emptyList();
         if (legacyMouseSource != null && legacyMouseSource.raiseLegacyMouseEvents())
         {
            mouseEvents = evt.legacyEvents(legacyMouseSource);
         }
         
         // post mouse down events before focus is transferred and entry/leave events sent
         Iterator<PortableMouseEvent> mouseEventsIt = mouseEvents.iterator();
         while (mouseEventsIt.hasNext())
         {
            PortableMouseEvent mouseEvent = mouseEventsIt.next();
            int eventId = mouseEvent.id();
            if (eventId == Keyboard.keyCode(Keyboard.MOUSE_SELECT_DOWN) ||
                eventId == Keyboard.keyCode(Keyboard.MOUSE_EXTEND_DOWN))
            {
               if (maySwitchFocus &&
                   currentEventList != null &&
                   !(legacyMouseSource instanceof Frame) &&
                   legacyMouseSource.focusTraversable())
               {
                  TriggerMatch match = new TriggerMatch(eventId);
                  currentEventList.lookup(-1, legacyMouseSource.getId().asInt(), true, false,
                                          true, match);

                  // for mouse down events, when the target is focusable and it explicitly
                  // defines the mouse down event and it is itself not a frame do not switch
                  // focus to it (or its frame), this is how 4GL works
                  maySwitchFocus = match.triggerId == -1;
               }

               // invoke the trigger directly; RETURN NO-APPLY is ignored
               // process the event right away, it must be triggered before focus is handled below
               eventBracket(() -> EventManager.postEvent(mouseEvent));

               // event triggered, remove from the list
               mouseEventsIt.remove();
            }
         }

         if (maySwitchFocus)
         {
            // use frame's current or default widget when focusing frame
            if (newFocus instanceof Frame && ((Frame<?>) newFocus).hasFocusable(null))
            {
               Frame<?> newFrame = (Frame<?>) newFocus;

               Widget<?> focus = newFrame.getContentPane().currentFocus(false);

               // do not recurse to inner frames, even when there are no focusable items in
               // this frame
               if (focus == null || focus instanceof Frame)
               {
                  focus = newFrame.defaultFocus();
               }

               if (focus != null && focus.focusTraversable())
               {
                  newFocus = focus;
               }
            }
            
            // if the new focus is a window which to which the old focus belongs, then do nothing
            if (newFocus instanceof TopLevelWindow && 
                oldFocus != null                   &&
                newFocus == oldFocus.topLevelWindow())
            {
               newFocus = oldFocus;
            }
            
            if (newFocus != oldFocus && newFocus.isFocusable() && newFocus.isEnabled())
            {
               // Send LEAVE to oldFocus.  If trigger returns NO-APPLY, then focus is switched
               // back to oldFocus. Note that 4GL first GIVES focus to newFocus (as the mouse
               // event is an OS event) and if the old focus doesn't accept LEAVE, it moves focus
               // back to the old one.  If old focus accepts LEAVE, ENTRY is sent to new focus.
               
               if (oldFocus != null && sendLeave(oldFocus, true))
               {
                  // if new and old focus are in different frames, send LEAVE/ENTRY to them
                  Widget newFrame = UiUtils.locateFrame(newFocus);
                  boolean sendFocusLost = false;
                  boolean sendFocusGained = false;
                  if (oldFrame != newFrame && oldFrame != null)
                  {
                     // prevent double-posting of LEAVE event if FRAME is the focused widget
                     if (oldFrame != oldFocus)
                     {
                        sendLeave(oldFrame, true);
                     }

                     sendFocusLost = true;
                  }
                  
                  boolean newFocusEntry = true;
                  if (oldFrame != newFrame && newFrame != null)
                  {
                     newFocusEntry = sendEntry(newFrame, true);
                     sendFocusGained = true;
                  }
                  
                  if (newFocusEntry)
                  {
                     // prevent double-posting of SEND event when FRAME is the target widget
                     if (newFrame != newFocus)
                     {
                        sendEntry(newFocus, true);
                     }

                     if (sendFocusLost && oldFrame != null)
                     {
                        EventManager.postEvent(new FocusEvent(oldFrame, EventType.FOCUS_LOST));
                     }
                     if (sendFocusGained)
                     {
                        EventManager.postEvent(new FocusEvent(newFrame, EventType.FOCUS_GAINED));
                     }
                     newFocus.requestFocus();
                  }
               }
               else if (oldFocus != null)
               {
                  // no ENTRY is generated when switching focus back to it
                  oldFocus.requestFocus();
               }
            }
         }

         // post the other mouse events after focus is transferred and entry/leave events sent
         mouseEventsIt = mouseEvents.iterator();
         while (mouseEventsIt.hasNext())
         {
            // invoke the trigger directly; RETURN NO-APPLY is ignored
            // the event determines which 4GL events are raised

            // do not make recursion here, just put next message into queue to handle
            // messages sequentially
            EventManager.postEvent(mouseEventsIt.next());
         }
         
         return;
      }
      
      if (event instanceof PortableMouseEvent || event instanceof WindowEvent)
      {
         // set the event details
         int evtId = event.id();
         
         functionKey = event instanceof WindowEvent ? ((WindowEvent) event).getAction()
                                                    : ((PortableMouseEvent) event).getAction();
         lastEvent = evtId;
         
         Widget<?> src = event.source();
         
         WidgetId id = src.getId();
         if (src != null && id != null && WidgetId.virtualWidget(id))
         {
            // virtual widgets do not call mouse triggers
         }
         else
         {
            Integer eventX = null;
            Integer eventY = null;
            
            if (event instanceof PortableMouseEvent)
            {
               eventX = ((PortableMouseEvent) event).getX();
               eventY = ((PortableMouseEvent) event).getY();
            }
            
            // invoke the trigger directly; RETURN NO-APPLY is ignored

            // some events has the second event embedded, the example is 4GL MOUSE_SELECT_UP
            // with integrated MOUSE_SELECT_CLICK, for both of them we can have separate triggers
            TriggerResult tr = invokeTriggers(event.source(), evtId, false, eventX, eventY);
            boolean consumed = tr.consume;
            if (functionKey != -1 && functionKey != evtId)
            {
               tr = invokeTriggers(event.source(), functionKey, false, eventX, eventY);
               consumed = consumed || tr.consume;
            }
            
            if (consumed)
            {
               // if the event is consumed by the trigger, the wait-for loop can't be interrupted
               return;
            }

            // check if we have mouse event attached exit conditions
            if (checkBreak(event.source(), evtId, false) ||
                (functionKey != -1 && checkBreak(event.source(), functionKey, false)))
            {
               isGoPending++;
            }
         }
         
         // if the event was not consumed, process it at the widget level
         if (src != null)
         {
            src.processEvent(event);
         }

         return;
      }
      
      if (event instanceof WindowActivated)
      {
         event.source().processEvent(event);
         return;
      }
      
      KeyInput evt    = (event instanceof KeyInput ? (KeyInput) event : null);
      Widget   src    = event.source();
      int      key    = (evt == null ? KeyInput.CHAR_UNDEFINED : evt.keyCode());
      int      action = (evt == null ? KeyInput.CHAR_UNDEFINED  : evt.actionCode());

      // skip developer events U1-U10
      if (!Keyboard.isDevEvent(action)) 
      {
         String evName = Keyboard.eventName(action);
         this.functionKey = action;
         if (key != KeyInput.CHAR_UNDEFINED)
         {
            this.lastEvent = key;
         }
      }
      
      if (currentEventList == null)
         return;

      int cop = currentEventList.getOperation();

      boolean foundCondExit = false;

      if(action == Keyboard.KA_STOP   ||
         action == Keyboard.KA_ENDKEY ||
         action == Keyboard.KA_ERROR  ||
         action == Keyboard.KA_END_ERROR)
      {
         condExit = true;
         foundCondExit = true;
      }

      if (authMode && key == Key.VK_CTRL_C)
      {
         condExit = true;
         throw new StopConditionException("STOP key function executed", key);
      }

      TriggerResult tr = null;
      Widget cur = UiUtils.getCurrentFocus();
      if (cur != null && !cur.navigable())
      {
         cur = null;
      }

      if (src instanceof Frame || src instanceof TopLevelWindow)
      {
         cur = src;
      }

      boolean foundGo = false;
      boolean exceptionalExit = false;

      // detect exit condition early
      if ((evt != null && checkBreak(src, key, evt.isRealKey())) ||
          checkBreak(src, action, false))
      {
         isGoPending++;
         foundGo = true;
      }
      else if (foundCondExit)
      {
         exceptionalExit = true;
      }
      else if (!waitForFrames.isEmpty() && 
               action == Keyboard.KA_GO && 
               currentBlockingOp == BlockingOperation.WAIT_FOR)
      {
         // if:
         // 0. this is the GO event
         //    AND
         // 1. we are in a WAIT-FOR event loop
         //    AND 
         // 2. the event's source is not in the current frame
         //    AND
         // 3. the event's source frame is part of a UPDATE/SET/PROMPT-FOR event loop
         //    THEN
         // terminate the current WAIT-FOR loop
         //    OR
         // 2b. the GO happening inside GUI dialog-box when event source can be the frame inside
         //     main dialog frame
         //    OR
         // 2c. the GO happening inside GUI frame when event source can be the nesting frame
         //
         // then we also terminate the current WAIT-FOR loop

         WaitForOp lastOp = waitForFrames.peek();
         Widget<?> srcFrame = UiUtils.locateFrame(src);
         
         // check for GUI mode the root dialog frame or nested frames 
         Frame<?> rootFrame = srcFrame != null ? ((Frame)srcFrame).getRootFrame() : null;
         boolean guiInsideDialogOrNested = false;
         if (rootFrame != null &&
             (rootFrame.isDialog() || rootFrame.getId().asInt() != srcFrame.getId().asInt()) &&
             !isChui())
         {
            srcFrame = rootFrame;
            guiInsideDialogOrNested = true;
         }
         
         if (srcFrame != null && (srcFrame != lastOp.frame || guiInsideDialogOrNested))
         {
            BlockingOperation srcFrameOp = null;
            for (WaitForOp op : waitForFrames)
            {
               if (op.frame == srcFrame)
               {
                  srcFrameOp = op.op;
                  break;
               }
            }
            
            if (srcFrameOp != null &&
                (srcFrameOp == BlockingOperation.PROMPT_FOR || guiInsideDialogOrNested))
            {
               isGoPending++;
               foundGo = true;
            }
         }
      }

      // low level Progress events:

      // in the accelerator list for the current window

      // implied LEAVE on exit
      if (isGoPending > 0 && !leaveSent && cur != null)
      {
         // clear pending GO temporarily for possible NO-APPLY and recursion
         int save = isGoPending;
         isGoPending = 0;

         // mark LEAVE as sent so there is no duplicate sent from WAIT-FOR
         leaveSent = true;

         // generate LEAVE and check if accepted, but only if the current
         // focus is valid
         if (FocusManager.focusState(cur) == FocusManager.FOCUS_VALID)
         {
            if (!sendLeave(cur))
            {
               // suppress further processing of this event
               evt.consume();

               // if LEAVE was not accepted, then the flag must be reset to
               // allow LEAVE to be sent again
               leaveSent = false;
               return;
            }

            if (checkBreak(cur, Keyboard.SE_LEAVE, false))
            {
               isGoPending ++;
               return;
            }
         }

         // LEAVE accepted; resume exit processing
         isGoPending = save;
      }

      // in the event list for an active user interface trigger
      if (key > 0 && key != KeyInput.CHAR_UNDEFINED)
      {
         tr = invokeTriggers(src, key, false);

         if (tr.consume)
         {
            evt.consume();

            if (foundGo)
            {
               isGoPending--;
               leaveSent = false;
            }
            return;
         }
      }

      // other high level events
      if ((tr == null || !tr.executed) && key != action &&
          action != Keyboard.SE_ENTRY)
      {
         tr = invokeTriggers(src, event.other(), action, false);

         // let widget perform "hard" check if necessary
         if (action == Keyboard.SE_LEAVE && src instanceof FillIn)
         {
            src.processKeyEvent(evt);

            if (evt.isConsumed())
               tr.consume = true;
         }

         if (tr.consume)
         {
            evt.consume();

            if (foundGo)
            {
               isGoPending--;
               leaveSent = false;
            }
            return;
         }
      }

      // check for possible ANY-KEY
      if ((tr == null || !tr.executed) && evt != null && evt.isRealKey())
      {
         tr = invokeTriggers(src, key, true);

         if (tr.consume)
         {
            evt.consume();

            if (foundGo)
            {
               isGoPending--;
               leaveSent = false;
            }
            return;
         }
      }

      // disable all following logic if this is a drop-down portion of a
      // combo-box
      if (src instanceof ScrollableList)
      {
         // STOP, ENDKEY and END-ERROR actions
         if (isGoPending == 0 || exceptionalExit)
         {
            processCondition(action, cop, false);
         }

         return;
      }

      // an active pop-up menu key
      if (action == Keyboard.KA_DEFAULT_POP_UP && (tr == null || !tr.executed) && src != null)
      {
         int menuId = src.config().popupMenuId;
         if (menuId != -1)
         {
            final Menu<?> menu = (Menu<?>) tk.getRegistry().getComponent(menuId);
            if (menu != null)
            {
               final Window wnd = (Window) src.ancestor();
               eventDrawingBracket(src, new Runnable()
               {
                  @Override
                  public void run()
                  {
                     if (menu.parent() == null)
                     {
                        wnd.getContentPane().add(menu);
                     }

                     menu.setVisible(true);
                     menu.repaint();
                     menu.requestSync();
                  }
               });
            }
         }
      }
      
      // an active menu mnemonic key

      // an active field-level (button) mnemonic key (Windows only)

      // translate not consumed CHOOSE to ENDKEY/GO for some buttons
      if (action == Keyboard.SE_CHOOSE &&  src instanceof Button)
      {
         ButtonConfig btc = (ButtonConfig) ((Button) src).config();

         int code = -1;

         // AUTO-ENDKEY has precedence in ordering if both options are defined as true
         if (btc.autoEndKey)
         {
            code = Keyboard.KA_ENDKEY;
         }
         else if (btc.autoGo)
         {
            code = Keyboard.KA_GO;
         }

         if (code != -1)
         {
            Frame<?> frame = UiUtils.locateFrame(src);
            
            // TODO: No frame means the widget has been destroyed before the
            //       event is processed. Is it OK to just abort processing 
            //       of the event here?
            if (frame != null)
            {
               KeyInput ki = new KeyInput(frame, EventType.KEY_PRESSED, code);
               EventManager.postEvent(ki);
            }
            
            return;
         }
      }

      // BELL action
      if (action == Keyboard.KA_BELL)
      {
         tk.beep();
      }

      // Internal browse events (currently just END-ERROR under certain
      // circumstances)
      if (processBrowseEvent(evt))
      {
         return;
      }

      // LEAVE event: do widget-level validation if defined and the widget is the explicit
      //              target of the leave (e.g. the widget is in focus AND the event is
      //              for the widget); DON'T execute this for a frame-level leave event
      //              generated by a GO (that case is handled further down this method) 
      if (action == Keyboard.SE_LEAVE)
      {
         if (src instanceof DataContainer)
         {
            DataContainer dc = (DataContainer) src;

            if (dc.isModified() && !validateWidget(src))
            {
               evt.consume();
               return;
            }
         }

         // remove the highlight from the widget
         src.setHighlighted(false);
      }

      // ENTRY event: do widget validation if defined
      if (action == Keyboard.SE_ENTRY)
      {
         Widget inFocus = UiUtils.getCurrentFocus();
         boolean fire = inFocus != src && !(src instanceof Frame);
         Frame frame = (Frame) UiUtils.locateFrame(src);

         boolean pushed = false;
         boolean failed = false;

         try
         {
            if (fire)
            {
               if (!src.isVisible() ||
                   !src.navigable() ||
                   (frame != null &&
                    !frame.isVisible()))
               {
                  // reject entry, source is not valid to receive such an event
                  evt.consume();
                  return;
               }

               // since the attempted focus change produces related CHARVA events
               // we have to shield ourselves with another event queue
               push();

               // force pop()/popTrash() cleanup in finally block
               pushed = true;

               if (!FocusManager.tryFocusChange(src, !evt.isSpecial(), false))
               {
                  // a trigger blocked LEAVE event
                  // trash the focus change events and rebalance the queue
                  failed = true;

                  // no go
                  evt.consume();
                  return;
               }

               if (!evt.isSpecial())
               {
                  // this must be done only if we are in an APPLY call
                  // if focus and target frame are different, send ENTRY to
                  // frame first
                  if (UiUtils.locateFrame(inFocus) != frame)
                  {
                     // if NO-APPLY, focus is undefined and no trigger is
                     // invoked
                     if (frame != null && (sendEntry(frame) == false))
                     {
                        // TODO: we use popTrash() here but is that correct?
                        failed = true;

                        WindowManager.getDefaultWindow().setFocus(null);
                        return;
                     }
                  }
               }
            }

            // delayed trigger processing for the ENTRY event
            tr = invokeTriggers(src, event.other(), action, false);

            if (tr.consume)
            {
               if (fire)
               {
                  // disable deferred cleanup because we will cleanup here
                  pushed = false;
                  popTrash();
               }

               // a trigger for ENTRY returned NO-APPLY: must adjust focus to
               // another widget in frame
               evt.consume();

               if (evt.isSpecial() == false && frame != null)
               {
                  // this must be done only if we are in an APPLY call
                  Widget next = frame.getNextEnabledWidget(src, FocusManager.FOCUS_NOWRAP);
                  if (next != null)
                  {
                     applyWorker(new KeyInput(next,
                                              evt.id(),
                                              evt.isRealKey() ? key : action,
                                              evt.isSpecial(),
                                              evt.isRealKey()));
                  }
               }

               return;
            }

            // allow the focus change events and rebalance the queue
            if (fire)
            {
               if (triggerNesting == 0)
               {
                  // save target widget, so it can be used in subsequent root
                  // level wait-for loops
                  pendingFocus = src;
               }
               else
               {
                  pendingFocus = null;
               }

               // ENTRY was accepted, set focus to source only if the widget
               // is still enabled, visible and within a nested trigger
               if (src.navigable() &&
                   src.isVisible() &&
                   (frame == null ||
                    frame.isVisible()))
               {
                  if (evt.isSpecial() || triggerNesting > 0)
                  {
                     src.requestFocus();

                     // if a CHOOSE is active, notify the frame
                     if (frame != null && frame.isChooseActive())
                     {
                        // CHOOSE will have no highlight if src is not part of it
                        frame.setChooseAt(src);
                     }
                  }
               }
               else
               {
                  // if focus was accepted and source is disabled or hidden,
                  // focus must not be changed -> but this doesn't mean the
                  // ENTRY was not accepted
                  return;
               }
            }
            else
            {
               // highlight the widget
               if (inputActive > 0)
               {
                  src.setHighlighted(true);
               }
            }
         }

         finally
         {
            if (pushed)
            {
               if (failed)
               {
                  popTrash();
               }
               else
               {
                  pop();
               }
            }
         }
      }

      // check additional exit conditions
      if (isGoPending == 0 && cop != EventList.WAIT_FOR)
      {
         if (currentEventList.isGoEvent(Keyboard.eventName(key)) ||
             currentEventList.isGoEvent(Keyboard.eventName(action)))
             isGoPending ++;
      }

      // GO action
      if ((isGoPending > 0 || action == Keyboard.KA_GO) && cur != null &&
         cop != EventList.WAIT_FOR)
      {
         Frame frame = null;

         if (src instanceof Frame)
         {
            frame = (Frame)src;
         }
         else
         {
            frame = (Frame)UiUtils.locateFrame(src);
         }

         // validate all validateable widgets before accepting GO
         Widget[] list = null;

         if (frame == null)
         {
            list = new Widget[0];
         }
         else
         {
            list = frame.getContentPane().widgets();
         }

         for (int i = 0; i < list.length; i++)
         {
            // no need to validate headers
            if (UiUtils.hasConfig(list[i]) && list[i].isHeader())
            {
               continue;
            }

            // only certain cases get validated
            if (list[i] instanceof DataContainer && list[i].isEnabled())
            {
               // validate
               if (!validateWidget(list[i]))
               {
                  evt.consume();
                  isGoPending --;
                  return;
               }
            }
         }

         isGoPending ++;
      }

      // exit condition for current-window
      if (isGoPending == 0 && cop == EventList.WAIT_FOR)
      {
         src = WindowManager.getDefaultWindow();

         if ((evt != null && checkBreak(src, key, evt.isRealKey())) ||
             checkBreak(src, action, false))
         {
            isGoPending ++;
         }
      }

      // check if we are exiting a choose
      if (isGoPending > 0 && cop == EventList.CHOOSE)
      {
         // choose does no further processing once a GO-ON has been reached
         evt.consume();
         return;
      }

      // a key associated with a key function in the PROTERMCAP file
      // (UNIX only)

      // a standard input key (A, B, and so on)

      // HELP action
      if (isGoPending == 0 && action == Keyboard.KA_HELP)
      {
         requestHelp();
      }

      // STOP, ENDKEY and END-ERROR actions
      if (isGoPending == 0 || exceptionalExit)
      {
         processCondition(action, cop, false);
      }
   }

   /**
    * Give a browse widget, if it is in this event's component hierarchy, a
    * chance to process and consume the END-ERROR event.  This is only
    * necessary if the browse widget is in edit mode and is currently set to
    * handle an internal, row-level undo.  Otherwise the event is not handled
    * or consumed.
    *
    * @param   evt
    *          Current event.
    *
    * @return  <code>true</code> if the event was an END-ERROR event and was
    *          handled and consumed by a browse widget in the event's
    *          component hierarchy;  else <code>false</code>.
    */
   private boolean processBrowseEvent(KeyInput evt)
   {
      if (evt == null || evt.actionCode() != Keyboard.KA_END_ERROR)
         return false;

      // Give containing browse a chance to consume END-ERROR.
      Widget comp = evt.source();
      do
      {
         if (comp instanceof Browse)
         {
            ((Browse) comp).onKeyPressed(evt);

            if (evt.isConsumed())
               return true;

            break;
         }
         comp = comp.parent();
      } while (comp != null);

      return false;
   }

   /**
    * Check the {@link #rowDisplayEvents} map for a ROW-DISPLAY trigger
    * associated with this widget. If this is a browse widget and such a
    * trigger is found, it will be executed.
    *
    * @param   widgetId
    *          event source.
    *
    * @return  Execution and event consume flags for an executed trigger.
    */
   public TriggerResult invokeRowDisplayTrigger(int widgetId)
   {
      TriggerResult tr = new TriggerResult();
      if (rowDisplayEvents == null)
      {
         return tr;
      }

      Widget widget = getWidget(widgetId);
      // allow only BROWSE widgets to trigger ROW-DISPLAY events
      if (widget instanceof Browse == false)
      {
         return tr;
      }

      // search for triggers
      int triggerId = -1;

      if (rowDisplayEvents.containsKey(widgetId))
      {
         triggerId = rowDisplayEvents.get(widgetId);
      }
      else
      {
         // search for ANYWHERE trigger
         if (rowDisplayEvents.containsKey(-1))
         {
            triggerId = rowDisplayEvents.get(-1);
         }
      }

      // activate trigger
      if (triggerId != -1)
      {
         // obtain its home frame ID
         int frameId  = -1;
         Widget frame = UiUtils.locateFrame(widget);
         int eventCode = Keyboard.eventCode(Keyboard.ROW_DISPLAY);

         frameId = UiUtils.getWidgetIdAsInt(frame);

         tr.executed = true;

         // save state variables
         EventList save = currentEventList;
         boolean state = inRowDisplayTrigger;

         inRowDisplayTrigger = true;

         // call the trigger
         try
         {
            tr.consume = trigger(triggerId,
                                 eventCode,
                                 widgetId,
                                 -1,
                                 -1,
                                 getScreenBuffer(frameId, -1),
                                 save);
         }

         finally
         {
            // restore state variables
            currentEventList = save;
            inRowDisplayTrigger = state;
         }
      }

      return tr;
   }

   /**
    * Check active <code>EventList</code> instance for triggers defined
    * for this event and widget pair and activate the first one found.
    * <p>
    * Triggers are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>trigger for the specified widget and event;
    *  <li>trigger for the specified event anywhere within the frame;
    *  <li>trigger for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * <p>
    * The first found trigger, if any, gets activated.
    *
    * @param   widget
    *          event source.
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          This flags marks reals keys typed by user.
    *
    * @return  Execution and event consume flags for an executed trigger.
    */
   public TriggerResult invokeTriggers(Widget widget, int evtCode, boolean isKey)
   {
      return invokeTriggers(widget, -1, evtCode, isKey);
   }

   /**
    * Check active <code>EventList</code> instance for triggers defined
    * for this event and widget pair and activate the first one found.
    * <p>
    * Triggers are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>trigger for the specified widget and event;
    *  <li>trigger for the specified event anywhere within the frame;
    *  <li>trigger for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * <p>
    * The first found trigger, if any, gets activated.
    *
    * @param   widget
    *          event source.
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          This flags marks reals keys typed by user.
    * @param   eventX
    *          The value of LAST-EVENT:X attribute; <code>null</code> if not used.
    * @param   eventY
    *          The value of LAST-EVENT:Y attribute; <code>null</code> if not used.
    *          
    * @return  Execution and event consume flags for an executed trigger.
    */
   public TriggerResult invokeTriggers(Widget  widget, 
                                       int     evtCode, 
                                       boolean isKey, 
                                       Integer eventX, 
                                       Integer eventY)
   {
      return invokeTriggers(widget, (Widget) null, -1, evtCode, isKey, eventX, eventY);
   }

   /**
    * Check active <code>EventList</code> instance for triggers defined
    * for this event and widget pair and activate the first one found.
    * <p>
    * Triggers are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>trigger for the specified widget and event;
    *  <li>trigger for the specified event anywhere within the frame;
    *  <li>trigger for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * <p>
    * The first found trigger, if any, gets activated.
    *
    * @param   widget
    *          event source.
    * @param   other
    *          a companion widget (for ENTER/LEAVE events).
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          This flags marks reals keys typed by user.
    *
    * @return  Execution and event consume flags for an executed trigger.
    */
   public TriggerResult invokeTriggers(Widget widget, Widget other, int evtCode, boolean isKey)
   {
      return invokeTriggers(widget, other, -1, evtCode, isKey, null, null);
   }
   
   /**
    * Check active <code>EventList</code> instance for triggers defined
    * for this event and widget pair and activate the first one found.
    * <p>
    * Triggers are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>trigger for the specified widget and event;
    *  <li>trigger for the specified event anywhere within the frame;
    *  <li>trigger for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * <p>
    * The first found trigger, if any, gets activated.
    *
    * @param   widget
    *          event source.
    * @param   resourceId
    *          When not set to -1, is a valid resource ID (for non-widget events).
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          This flags marks reals keys typed by user.
    *
    * @return  Execution and event consume flags for an executed trigger.
    */
   private TriggerResult invokeTriggers(Widget    widget,
                                        long      resourceId,
                                        int       evtCode,
                                        boolean   isKey)
   {
      return invokeTriggers(widget, null, resourceId, evtCode, isKey, null, null);
   }

   /**
    * Check active <code>EventList</code> instance for triggers defined
    * for this event and widget pair and activate the first one found.
    * <p>
    * Triggers are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>trigger for the specified widget and event;
    *  <li>trigger for the specified event anywhere within the frame;
    *  <li>trigger for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * <p>
    * The first found trigger, if any, gets activated.
    *
    * @param   widget
    *          event source.
    * @param   other
    *          a companion widget (for ENTER/LEAVE events).
    * @param   resourceId
    *          When not set to -1, is a valid resource ID (for non-widget events).
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          This flags marks reals keys typed by user.
    *
    * @return  Execution and event consume flags for an executed trigger.
    */
   private TriggerResult invokeTriggers(Widget    widget,
                                        Widget    other,
                                        long      resourceId,
                                        int       evtCode,
                                        boolean   isKey)
   {
      return invokeTriggers(widget, other, resourceId, evtCode, isKey, null, null);
   }

   /**
    * Check active <code>EventList</code> instance for triggers defined
    * for this event and widget pair and activate the first one found.
    * <p>
    * Triggers are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>trigger for the specified widget and event;
    *  <li>trigger for the specified event anywhere within the frame;
    *  <li>trigger for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * <p>
    * The first found trigger, if any, gets activated.
    *
    * @param   widget
    *          event source.
    * @param   other
    *          a companion widget (for ENTER/LEAVE events).
    * @param   resourceId
    *          When not set to -1, is a valid resource ID (for non-widget events).
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          This flags marks reals keys typed by user.
    * @param   eventX
    *          The value of LAST-EVENT:X attribute; <code>null</code> if not used.
    * @param   eventY
    *          The value of LAST-EVENT:Y attribute; <code>null</code> if not used.
    *
    * @return  Execution and event consume flags for an executed trigger.
    */
   private TriggerResult invokeTriggers(Widget    widget,
                                        Widget    other,
                                        long      resourceId,
                                        int       evtCode,
                                        boolean   isKey,
                                        Integer   eventX,
                                        Integer   eventY)
   {
      // this method must not be called for mouse events on virtual widgets; for key presses, 
      // this is allowed because ANY-KEY triggers need to be reached.

      TriggerResult tr = new TriggerResult();

      WidgetId id = widget.getId();
      if (widget != null          && 
          widget.parent() != null &&
          WidgetId.virtualWidget(id))
      {
         // all keys from virtual sources (still attached to a window) are sent directly to the 
         // associated window...
         widget = widget.topLevelWindow();
      }
      
      // no wait list, this is a combo drop-down list widget or this is an
      // APPLY with implicit target, and no focus exists - nothing to do
      if (currentEventList == null || implicitApplyNoTrigger)
         return tr;

      // obtain the widget ID and its home frame ID
      int frameId  = -1;
      int widgetId = UiUtils.lookup4GLWidgetIdAsInt(widget);
      Widget frame = UiUtils.locateFrame(widget);

      // Progress watches the focus during event processing and restores it
      // whenever triggers move it away but then allow the default event
      // processing to occur. I assume here that it should only happen if
      // the event source is also the widget in focus, although it may leave
      // other cases out

      Widget inFocus = UiUtils.getCurrentFocus();
      
      // TODO: is this correct?
      if (frame == null && inFocus != null)
      {
         // if frame has not been resolved (for example widget is a top-level window)
         // try to locate it from the focused widget
         frame = UiUtils.locateFrame(inFocus);
      }

      // use the session window if no specific widget was specified AND there
      // is nothing in focus
      if (widgetId == -1 && inFocus == null)
      {
         widgetId = WidgetId._DEFAULT_WINDOW_ID;
      }
      else if (frame != null)
      {
         // otherwise try to initialize the frame ID
         frameId = UiUtils.getWidgetIdAsInt(frame);
      }
      
      TriggerMatch result = new TriggerMatch(evtCode);

      // do the search for triggers
      if (resourceId != -1)
      {
         currentEventList.lookup(resourceId, true, result);
      }
      else if (!WidgetId.virtualWidget(widgetId))
      {
         currentEventList.lookup(frameId, widgetId, true, isKey, result);
      }

      // activate trigger
      if (result.triggerId != -1)
      {
         tr.executed = true;

         // save state variables
         EventList save = currentEventList;
         Object drawState = null;
         
         // call the trigger
         
         try
         {
            if (!isChui())
            {
               drawState = ((GuiDriver) tk.getInstanceDriver()).saveDrawingState();
            }
            
            triggerResultNoApply.push(false);
            triggerHadWaitFor.push(false);
            
            // TODO: should we move the recursion protection code here from the
            //       LT.trigger() to avoid the server up-call AND avoid the extra
            //       processing in that method that occurs before the recursion is
            //       detected?

            int focusFrameId = frameId;
            Widget<?> focus = UiUtils.getCurrentFocus();
            if (focus != null)
            {
               Widget<?> focusFrame = UiUtils.locateFrame(focus);
               if (focusFrame != null)
               {
                  focusFrameId = UiUtils.getWidgetIdAsInt(focusFrame);
               }
            }
         
            lastEventX = eventX;
            lastEventY = eventY;
            
            // note that the event id may be different from the original event code for
            // "wildcard" events like ANY-KEY or ANY-PRINTABLE, so we must pass the
            // modified event id so that we avoid recursion
            tr.consume = trigger(result.triggerId,
                                 result.eventId,
                                 widgetId,
                                 UiUtils.lookupWidgetIdAsInt(other),
                                 resourceId,
                                 getScreenBuffer(focusFrameId, -1, true),
                                 save);
         }

         finally
         {
            if (!isChui())
            {
               ((GuiDriver) tk.getInstanceDriver()).restoreDrawingState(drawState);
            }
            
            triggerResultNoApply.pop();
            boolean hadWaitFor = triggerHadWaitFor.pop();

            // save last result, on this level
            triggerResultNoApply.pop();
            triggerResultNoApply.push(tr.consume);

            triggerHadWaitFor.pop();
            triggerHadWaitFor.push(hadWaitFor);

            // restore state variables
            currentEventList = save;

            // check if the previous focus is still attached to a window
            // also, check if the frame is still valid
            boolean hadFrame = frame != null;
            if (frame != null)
            {
               // if the wait-for loop started with a valid frame, this might have been deleted
               // by a trigger; this check will verify if the frame still exits
               frame = getWidget(frame.getId().asInt());
            }
            boolean hadInFocus = inFocus != null;
            if (inFocus != null)
            {
               // always perform the check on the related legacy widget
               Widget legacyInFocus = inFocus.getLegacyWidget();

               if (legacyInFocus != null)
               {
                  // if the wait-for loop started with a valid focus, this might have been deleted
                  // by a trigger; this check will verify if the focus still exits
                  legacyInFocus = getWidget(legacyInFocus.getId().asInt());
               }

               if (legacyInFocus == null || !legacyInFocus.isRealized())
               {
                  inFocus = null;
               }
            }
            
            if ((hadFrame && frame == null) || 
                (hadInFocus && (inFocus == null || inFocus.ancestor() == null)))
            {
               // the frame or the focused widget has 'dissapeared'
            }
            else
            {
               Widget focusNow = UiUtils.getCurrentFocus();

               // restore focus under certain circumstances
               // the exact circumstances are to be determined (work in progress)
               // - event is not consumed
               // - event is not VALUE-CHANGED
               // - widget is FILL-IN
               if (!tr.consume              &&
                   focusNow != inFocus      &&
                   inFocus != null          &&
                   widget == inFocus        &&
                   (widget instanceof FillIn || widget instanceof ComboBox) &&
                   evtCode != Keyboard.SE_VALUE_CHANGED)
               {
                  setFocusSilent(inFocus);
               }
               else if (tr.consume          &&
                        focusNow != inFocus &&
                        inFocus != null     &&
                        widget == inFocus   &&
                        widget instanceof Button)
               {
                  setFocusSilent(inFocus);
                  focusNow = inFocus;
               }
               else if (frame != null && ((Frame) frame).isFocusReset())
               {
                  focusNow = ((Frame) frame).getFirstEnabledWidget();

                  if (focusNow != null)
                     setFocusSilent(focusNow);
               }

               if (focusNow != null && focusNow == inFocus)
               {
                  checkHelpText(focusNow);
               }
            }
         }
      }
      else
      {
         // no trigger, reset the result to false
         triggerResultNoApply.pop();
         triggerResultNoApply.push(false);
         triggerHadWaitFor.pop();
         triggerHadWaitFor.push(false);
      }
      
      return tr;
   }

   /**
    * Silently restore and redraw the focus. No LEAVE/ENTRY events should be
    * generated here.
    *
    * @param    inFocus
    *           Widget to focus.
    */
   private void setFocusSilent(final Widget inFocus)
   {
      eventBracket(new Runnable()
      {
         @Override
         public void run()
         {
            inFocus.requestFocus();

            if (inFocus instanceof FillIn)
            {
               ((FillIn) inFocus).setZapMode(true);
            }
         }
      });
   }

   /**
    * Check help text and restore it if necessary.
    *
    * @param    focus
    *           Widget in focus.
    */
   private void checkHelpText(Widget focus)
   {
      WidgetConfig cfg = focus.config();
      
      if (cfg == null)
      {
         return;
      }
      
      String help = cfg.getDynamicHelp();

      // TODO: what window to use when the parent main window has its status area hidden?
      Window wnd = focus.window();
      
      if (wnd == null) 
      {
         throw new IllegalStateException("Widget " + focus.name() + " " + 
                  UiUtils.getWidgetIdAsInt(focus) + " is not attached to a Window instance.");
      }
      
      String status = wnd.getStatusText();

      if (help == null || help.equals(status))
         return;

      wnd.setHelp(help);
   }

   /**
    * Check active <code>EventList</code> instance for exit conditions
    * for this event and widget pair.
    * <p>
    * Events are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>exact match for the specified widget and event;
    *  <li>match for the specified event anywhere within the frame;
    *  <li>match for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    *
    * @param   widget
    *          event source.
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          Mark real keys typed by user.
    *
    * @return  <code>true</code> if a matching event exists
    */
   private boolean checkBreak(Widget widget, int evtCode, boolean isKey)
   {
      // no wait list; nothing to do
      if (currentEventList == null)
         return false;

      // obtain the widget ID and its home frame ID
      int frameId  = -1;
      int widgetId = UiUtils.lookupWidgetIdAsInt(widget);
      Widget frame = UiUtils.locateFrame(widget);
      Widget inFocus = UiUtils.getCurrentFocus();
      
      // TODO: is this correct?
      if (frame == null && inFocus != null)
      {
         // if frame has not been resolved (for example widget is a top-level window)
         // try to locate it from the focused widget
         frame = UiUtils.locateFrame(inFocus);
      }

      // use the session window if no specific widget was specified AND there
      // is nothing in focus
      if (widgetId == -1 && inFocus == null)
      {
         widgetId = WidgetId._DEFAULT_WINDOW_ID;
      }
      else if (frame != null)
      {
         // otherwise try to initialize the frame ID
         frameId = UiUtils.getWidgetIdAsInt(frame);
      }
      
      TriggerMatch result = new TriggerMatch(evtCode);

      // do the search for exit condition
      currentEventList.lookup(frameId, widgetId, false, isKey, result);

      return result.triggerId != -1;
   }

   /**
    * Check active <code>EventList</code> instance for exit conditions for this event and resource pair.
    * <p>
    * The specific search is done by exact match for the specified resource and event
    *
    * @param   resourceId
    *          id of the resource generating the event.
    * @param   evtCode
    *          event code.
    *
    * @return  <code>true</code> if a matching event exists
    */
   private boolean checkBreak(long resourceId, int evtCode)
   {
      // no wait list; nothing to do
      if (currentEventList == null)
         return false;
      
      TriggerMatch result = new TriggerMatch(evtCode);

      // do the search for exit condition
      currentEventList.lookup(resourceId, false, result);

      return result.triggerId != -1;
   }

   /**
    * Handles processing of ENDKEY, END-ERROR and STOP conditions in a
    * uniform manner.  The given action is checked for a match with the
    * list of conditions, if a match exists then the associated condition
    * is raised (an exception is thrown).
    *
    * @param    action
    *           The key action to check against the list of possible
    *           conditions.
    * @param    opcode
    *           Specifies which operation is in process (for example, a
    *           {@link EventList#WAIT_FOR}).
    * @param    inPause
    *           Specifies whether the method is called from a pausing
    *           function.
    *
    * @throws   StopConditionException
    *           If action is {@link Keyboard#KA_STOP}.
    * @throws   EndConditionException
    *           If action is {@link Keyboard#KA_ENDKEY} or on
    *           {@link Keyboard#KA_END_ERROR} on the first interaction in the
    *           current block or when opcode is
    *           {@link EventList#WAIT_FOR}.
    * @throws   ErrorConditionException
    *           If action is {@link Keyboard#KA_END_ERROR} and one of the
    *           above conditions does not hold.
    */
   private void processCondition(int action, int opcode, boolean inPause)
   throws StopConditionException,
          EndConditionException,
          ErrorConditionException
   {
      // Note that the opcode == EventList.UPDATE, pause(), message(),
      // choose() and other cases on the server need to know if the cause of
      // a condition was a specific key (e.g. END-ERROR).  For this reason
      // there is an extra flag passed to the constructor for all of these
      // cases.

      boolean store = (lastAppliedCode == Keyboard.KA_STOP ||
                       lastAppliedCode == Keyboard.KA_END_ERROR ||
                       inPause &&
                       triggerNestingByWaitFor.peek() > 1) &&
                      currentEventList != null &&
                      currentEventList.getOperation() == EventList.WAIT_FOR &&
                      !inEditingBlock;

      // STOP action
      if (action == Keyboard.KA_STOP)
      {
         ConditionException exception =
            new StopConditionException("STOP key function executed", lastKey);

         if (store)
            immediateException = exception;
         throw exception;
      }

      // ENDKEY action
      if (action == Keyboard.KA_ENDKEY)
         throw new EndConditionException("END key function executed",
                                         lastKey);

      // ERROR action
      if (action == Keyboard.KA_ERROR)
         throw new ErrorConditionException("ERROR key function executed",
                                           lastKey);

      // END-ERROR action
      if (action == Keyboard.KA_END_ERROR)
      {
         int effective = interactions;

         // pause has a special behavior in that it does not change the
         // interactions counter BUT if it executes when the interactions
         // counter is 1 (but not when it is 0), then an END-ERROR will be
         // translated into ERROR (as if the real interactions counter was
         // 2 or more)
         if (inPause && effective == 1)
         {
            effective = 2;
         }

         // no interactions in this block iteration, 1st interaction in this
         // block iteration or a wait-for both always get an ENDKEY instead
         // of an error
         if (effective < 2 ||
             (opcode == EventList.WAIT_FOR && !inEditingBlock))
         {
            ConditionException exception =
               new EndConditionException(END_ERR_MSG, lastKey);

            if (store)
               immediateException = exception;
            throw exception;
         }

         // the opcode == EventList.UPDATE on the server needs even one
         // more flag... to note that this was an END-ERROR special case
         ConditionException exception =
            new ErrorConditionException(END_ERR_MSG, lastKey, true);

         if (store)
            immediateException = exception;
         throw exception;
      }
   }

   /**
    * Validates the widget's current value and displays the validation error
    * message if any.
    *
    * @param   widget
    *          widget to validate
    * @return  The result of the validation.
    */
   private boolean validateWidget(Widget widget)
   {
      int wid = UiUtils.lookupWidgetIdAsInt(widget);

      if (!currentEventList.isValidatable(wid))
         return true;

      // hide previously displayed error message, if any
      Widget ancestor = widget.ancestor();
      
      if (!(ancestor instanceof Window))
      {
         throw new IllegalStateException("Frame " + widget.name() + " " + 
                  UiUtils.getWidgetIdAsInt(widget) + " is not attached to a Window instance.");
      }
      
      Window wnd = (Window) ancestor;
      
      if (validationError != null)
      {
         wnd.hideMessage(true);
      }
 
      // check if validation is required
      boolean validate = true;
      Frame<?> frame = (Frame<?>) UiUtils.locateFrame(widget);
      if (frame != null)
      {
         FrameConfig fc = frame.config();
         validate = !fc.noAutoValidate;
      }
      // call the validation expression evaluator
      if (validate && !validate(wid, getScreenBuffer(widget)))
      {
         // validation failed: display message, beep and stay in
         wnd.message(validationError);

         // what widget is in focus right now?
         Widget comp = UiUtils.getCurrentFocus();

         if (widget == comp)
         {
            // enable auto-zap and reset insert mode for fill-ins
            enableFillInAutoZap(widget);

            // request focus will silently fail here, but we do need to
            // reset the cursor and state as if we just entered the field
            // so force activation here
            EventManager.postEvent(new FocusEvent(comp, EventType.FOCUS_GAINED));
            comp.repaint();
         }
         else
         {
            // focus had already changed, change the focus back
            widget.requestFocus();
         }

         bell();
         return false;
      }

      return true;
   }

   /**
    * Activate AUTO-ZAP mode for widget if it a FillIn instance.
    *
    * @param   widget
    *          Widget to process.
    */
   private void enableFillInAutoZap(Widget widget)
   {
      if (widget instanceof FillIn)
      {
         FillIn fin = (FillIn) widget;
         fin.setZapMode(true);
         fin.setInsertMode(false);
      }
   }


   /**
    * Validates the widget's ID by comparing with array of allowed IDs.
    *
    * @param   id
    *          widget id to validate
    * @param   ids
    *          array of allowed ids or <code>null</code>
    * @return  equals the input id or <code>-1</code>
    */
   private int validateWidget(int id, int[] ids)
   {
      if (ids == null)
      {
         return id;
      }

      for (int i = 0; i < ids.length; i ++)
      {
         if (id == ids[i])
         {
            return id;
         }
      }

      return -1;
   }

   /**
    * Changes the highlighted state of the widget and repaints it.
    *
    * @param    widget
    *           The widget to modify.
    * @param    state
    *           The new state of the widget's highlight.
    */
   public void highlight(final Widget widget, final boolean state)
   {
      if (widget != null)
      {
         // change the highlighted state of the widget
         eventDrawingBracket(widget, false, false, new Runnable()
         {
            @Override
            public void run()
            {
               widget.setHighlighted(state);
               Widget<?> focus = UiUtils.getCurrentFocus();
               if (focus != null)
               {
                  focus.repaint();
               }
            }
         });
      }
   }

   /**
    * Save insert mode before using a prompting function or a trigger.
    *
    * @return  Insert mode that was saved.
    */
   public int saveInsState()
   {
      int savedInsState = Window.getInsertMode();

      // save insert mode in order to make the next focused fill-in accept it
      savedInsStates.push(savedInsState);

      return savedInsState;
   }

   /**
    * Restore insert mode after a prompting function or a trigger.
    *
    * @param   savedInsState
    *          Insert mode that was before the call of the corresponding
    *          prompting function or trigger.
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   public void restoreInsState(int savedInsState, boolean force)
   {
      // restore insert indicator state
      if (savedInsState != -1)
      {
         // skip insert mode adjustment for editors
         Widget focus = UiUtils.getCurrentFocus();
         if (focus == null || !(focus instanceof Editor))
         {
            Window.setInsertMode(savedInsState, force);
         }
      }
      savedInsStates.pop();

      // replace the top stack value with the restored insert mode in order to
      // make the next focused fill-in accept it
      savedInsStates.pop();
      savedInsStates.push(savedInsState);
   }

   /**
    * Set new value for the <code>needPause</code> flag.
    *
    * @param   needPause
    *          New value for the flag.
    * @param   wnd
    *          Not null target window. If null - an NPE will be thrown.
    */
   public void setNeedPause(boolean needPause, TopLevelWindow wnd)
   {
      if (needPause == false)
      {
         UiUtils.resetNeedPause(wnd);
         wnd.window().setMessageNeedPause(false);
      }

      this.needPause = needPause;
   }

   /**
    * Get current value of the <code>needPause</code> flag.
    *
    * @return  Returns the needPause.
    */
   boolean isNeedPause()
   {
      return needPause || UiUtils.isNeedPause();
   }

   /**
    * Returns <code>true</code> if <code>enable(..)</code> function is being
    * run.
    *
    * @return  See above.
    */
   public boolean isInEnable()
   {
      return inEnable;
   }

   /**
    * Returns saved insert indicator state that should be applied to the next
    * focused fill-in, or <code>-1</code> if there is no state to apply.
    *
    * @return  See above.
    */
   public int getSavedInsState()
   {
      return savedInsStates.peek();
   }

   /**
    * Clears saved insert mode.
    */
   public void clearSavedInsState()
   {
      savedInsStates.pop();
      savedInsStates.push(-1);
   }

   /**
    * Get the stream to which the terminal is redirected or <code>null</code>
    * if no redirection is currently in effect.
    *
    * @return   The current redirection or <code>null</code> if not
    *           redirected.
    */
   public Stream getCurrentRedirection()
   {
      if (!isRedirected())
         return null;

      return sd.streamFromId(currentStream);
   }

   /**
    * Display frame content in redirected mode by page element.
    *
    * @param   frameId
    *          Frame ID.
    */
   public void view(int frameId)
   {
      Frame frame = (Frame) getWidget(frameId);
      if (frame == null)
         return;

      ScreenBuffer[] frameBuf = server.refresh(frameId);

      Frame topFrameIn = frame.window().getTopFrame(true);

      // this is an equivalent of view call for page headers/footers
      view_(frameId, frameBuf, null, true, false, UiUtils.getWidgetIdAsInt(frame.window()));

      Frame topFrameOut = frame.window().getTopFrame(true);

      if (topFrameIn != topFrameOut)
         topFrameOut.moveToBottom();
   }

   /**
    * Execute the NEXT-FRAME or PREV-FRAME event, depending on flag. Frame
    * search starts from the specified frame.
    *
    * @param   frame
    *          The frame from which search will start.
    * @param   nextFrame
    *          <code>true</code> if this is the NEXT-FRAME event, <code>false</code>
    *          if this is the PREV-FRAME event.
    */
   public void handleNextPrevFrameEvents(Frame frame, boolean nextFrame)
   {
      transferFocusToFrame(frame, nextFrame ? nextFrame(frame) : prevFrame(frame));
   }

   /**
    * Get next frame.
    *
    * @param   frame
    *          Current frame.
    *
    * @return  Next frame or <code>null</code> if no such frame available.
    */
   private Frame nextFrame(Frame frame)
   {
      Window window = frame.window();
      
      Frame newFrame = window.getNextEnabledFrame(frame);

      // if no other enabled frames in this direction, get first
      // enabled and visible frame, in z-order
      return newFrame == null ? window.getNextEnabledFrame(null)
                              : newFrame;

   }

   /**
    * Get previous frame.
    *
    * @param   frame
    *          Current frame.
    *
    * @return  Previous frame or <code>null</code> if no such frame available.
    */
   private Frame prevFrame(Frame frame)
   {
      Window window = frame.window();
      
      Frame newFrame = window.getPrevEnabledFrame(frame);

      // if no other enabled frames in this direction, get last
      // enabled and visible frame, in z-order
      return newFrame == null ? window.getPrevEnabledFrame(null)
                              : newFrame;
   }

   /**
    * Transfer focus between frames.
    *
    * @param   frame
    *          Frame which holds focus now.
    * @param   newFrame
    *          Frame which should receive focus.
    */
   private void transferFocusToFrame(Frame frame, Frame newFrame)
   {
      if (newFrame == null || newFrame == frame)
         return;

      // first enabled widget in new frame
      Widget comp  = getWidget(Frame.getFirstFocusableId(newFrame));
      // get current focus
      Widget focus = UiUtils.getCurrentFocus();
      // get next enabled widget in frame
      Widget next  = frame.getNextEnabledWidget(focus, FocusManager.FOCUS_WRAP);

      if (!focus.isEnabled())
      {
         // if focus is disabled and there is another enabled widget in
         // its parent frame, set focus on it; else, focus first enabled
         // widget in new frame
         if (next != null)
         {
            focusManager.setFocusOn(next);
         }
         else
         {
            focusManager.setFocusOn(comp);
         }

         focus = next;
      }

      if (focus != null)
      {
         // current focus receives leave first
         sendLeave(focus);

         // send LEAVE to frame and stop processing, if it doesn't accept
         // LEAVE
         if (frame != focus)
         {
            sendLeave(frame);
            // if NO-APPLY, stop further processing
            if (triggerResultNoApply.peek())
            {
               return;
            }
         }
      }

      // order of events is: ENTRY to frame first, ENTRY to widget second
      sendEntry(UiUtils.locateFrame(comp));
      sendEntry(comp);
   }

   /**
    * Checks whether the frame needs switching streams and does switching.
    * The last used stream for the frame is compared with the current stream
    * in effect. If these are different streams, the switching will occur.
    *
    * @param   frame
    *          the target frame for the operation
    * @return  <code>true</code> if the stream switch has happened and it was
    *          not the very first switch for the frame
    */
   @SuppressWarnings("null")
   private boolean switchStreams(Frame frame)
   {
      int oldid = -1;
      int newid = -1;
      Stream olds = null;

      if (isRedirected())
      {
         newid = currentStream;
      }

      if (frame.isRedirected())
      {
         oldid = frame.getStreamId();
         olds = sd.streamFromId(oldid);
      }

      if (oldid == newid)
      {
         return false;
      }

      // flush the old stream
      if (oldid != -1)
      {
         Integer key = Integer.valueOf(oldid);
         RedirectedTerminal rt = redirections.get(key);

         // is the stream still around?
         if (rt != null)
         {
            // flush the terminal and stream buffers
            if (rt.flush())
            {
               try
               {
                  // only done if there was pending RT output so make sure
                  // that headers are bypassed
                  olds.flushControlled(true);
               }

               catch (IOException ioe)
               {
                  ErrorManager.showErrorAndAbend(140,
                                                 Stream.PIPE_BROKEN_MESSAGE);
               }
            }
         }
      }

      // note the new stream
      if (newid == -1)
      {
         frame.setRedirected(false, -1);
      }
      else
      {
         frame.setRedirected(true, newid);
      }

      // normally, at this point, the method should return true meaning
      // there was a real stream switch but the very first switch for the
      // frame should be excluded and this what the use of the streamed flag
      // provides
      return frame.isStreamed();
   }

   /**
    * Checks whether the stream and its redirected terminal need to switch to
    * another frame and does switching if so.
    * <p>
    * Switching the stream between frames causes flushing the buffers of the
    * redirected terminal to the stream. Otherwise the redirected terminal
    * receives continuous output from the same frame.
    *
    * @param   streamId
    *          ID of the output stream the current output is redirected to.
    * @param   frameId
    *          ID of the frame about to produce some output
    * @return  <code>true</code> when the frame switch has happened or
    *          <code>false</code> for the case of continuous output.
    */
   private boolean switchFrames(int streamId, int frameId)
   {
      Stream             stream = sd.streamFromId(streamId);
      RedirectedTerminal rt     = redirections.get(streamId);
      Frame              frame  = (Frame) getWidget(frameId);

      if (stream == null || rt == null || (frame == null && frameId != -1))
      {
         return false;
      }

      // save the fact of whether there is something to flush or not since
      // the targetFrame() call will cause internal flushing and does not
      // return an indication of whether flushing was needed or not;
      boolean bypass = !rt.isFresh();

      int flushId = rt.targetFrame(frameId);

      boolean flushed = flushId != frameId && flushId != -1;

      if (flushed)
      {
         // redirected terminal has changed ownership and flushed buffers
         // flush the stream now
         flushRedirected(null, bypass);

         // the previous owner of the redirected terminal needs to be notified
         // this notification removes the stream from the list of streams that
         // have frame headers output and also hides the frame.
         // these are exactly the actions that are taken when the page break
         // occurs and that't why the notifyPageBreak() is reused here
         Frame gone = (Frame) getWidget(flushId);

         if (gone != null)
         {
            FrameConfig cfg = gone.config();

            boolean top = cfg.pageTop;
            boolean bot = cfg.pageBottom;

            if (!top && !bot)
            {
               gone.notifyPageBreak(stream);
            }
         }
      }
      else if (flushId == frameId)
      {
         return false;
      }

      return true;
   }

   /**
    * Push any pending output from the redirected terminal to the stream and
    * then flush the stream.
    *
    * @param    out
    *           The stream being flushed.
    */
   public void flushStream(Stream out)
   {
      int sid = sd.idFromStream(out);

      boolean switched = currentStream == sid;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(sid, true);
      }

      RedirectedTerminal rt = redirectedTerminalForStream(out);

      // flush the redirected terminal buffer to the stream
      boolean bypass = rt.flush();

      // flush the stream buffer
      try
      {
         out.flushControlled(bypass);
      }

      catch (IOException ioe)
      {
         ErrorManager.showErrorAndAbend(140, Stream.PIPE_BROKEN_MESSAGE);
      }

      finally
      {
         if (!switched)
         {
            switchOutput(tempOut, false);
         }
      }
   }

   /**
    * Check if outermost trigger processing result was 'NO-APPLY'.
    *
    * @return  outermost trigger processing result.
    */
   public Boolean isTriggerResultNoApply()
   {
      return triggerResultNoApply.peek();
   }

   /**
    * Force the given stream to output any buffered PUT or EXPORT content, so
    * long as the stream is not the terminal.
    *
    * @param    sid
    *           The stream on which to call <code>endOutput</code>.
    */
   private void endOutput(int sid)
   {
      if (sid != -1)
      {
         Stream stream = sd.streamFromId(sid);

         if (stream != null)
         {
            // when ending output, a NL must be added for the PUT statement
            // the EXPORT statement has this NL added after its execution
            stream.endOutput(true);
         }
      }
   }

   /**
    * Flushes all buffered output to the redirected output stream (does
    * nothing if output is not redirected).  This call may cause recursive
    * calls to {@link #view} in order to write page headers or footers.
    *
    * @param    stream
    *           The stream to flush or <code>null</code> to flush the
    *           current redirection destination.
    * @param    bypass
    *           <code>true</code> if page header processing must be bypassed
    *           when flushing the output. This should be <code>true</code>
    *           if the caller has pushed pending redirected terminal output
    *           into the stream since this will have already processed the
    *           page headers and this will avoid duplicates.
    *
    * @return   The next line number AFTER all buffered lines are output but
    *           BEFORE any final page break processing has occurred. This can
    *           be compared with the next line number as queried after this
    *           method returns to determine if a page break was output or
    *           not (among other uses). If there was no output or if the
    *           terminal is not redirected, then -1 will be returned.
    */
   private int flushRedirected(Stream stream, boolean bypass)
   {
      int rc = -1;

      if (isRedirected())
      {
         if (stream == null)
         {
            stream = sd.streamFromId(currentStream);
         }
         
         if (stream == null)
         {
            throw new IllegalStateException("Invalid stream id " + currentStream + ".");
         }
         
         try
         {
            rc = stream.flushControlled(bypass);
         }
         catch (IOException ioe)
         {
            ErrorManager.showErrorAndAbend(140, Stream.PIPE_BROKEN_MESSAGE);
         }
      }

      return rc;
   }

   /**
    * Changes the value of a widget in a frame.
    *
    * @param   frame
    *          <code>Frame</code> where the widget belongs
    * @param   comp
    *          <code>Component</code> for the widget
    * @param   config
    *          <code>ComponentConfig</code> for the widget
    * @param   value
    *          new value
    * @param   fmt
    *          format if formatted value
    */
   private void setChangedValue(Frame        frame,
                                Widget       comp,
                                WidgetConfig config,
                                BaseDataType value,
                                String       fmt)
   {
      // force the new value into the widget
      if (frame != null)
      {
         // handles any down frame processing as needed
         frame.setValue(comp, value, fmt);
         // Force focus transfer to the changed fill-in after choose
         if (comp.focusTraversable()      &&
             comp instanceof FillIn       &&
             focusToRestoreFillIn != null &&
             ((FillIn)comp).config().id.asInt() == focusToRestoreFillIn.config().id.asInt())
         {
            focusToRestoreFillIn.requestFocus();
            focusToRestoreFillIn = null;
         }
      }
      else
      {
         if (comp instanceof FillIn)
         {
            // honor any format string override
            FillIn fi = (FillIn) comp;
            fi.setValue(value, fmt);
         }
         else
         {
            ((DataContainer) comp).setValue(value);
         }

         comp.setState(ScreenBuffer.UNCHANGED);
      }
   }

   /**
    * Get reference to active frame.
    *
    * @return  Reference to active frame or <code>null</code> if there is no
    *          such frame.
    */
   public Frame getActiveFrameInt()
   {
      TopLevelWindow wnd = WindowManager.getActiveWindow();
      Frame frame = wnd != null ? wnd.getTopFrame(false) : null;

      if (frame != null)
         return frame;

      if (lastViewed == null      ||
          !lastViewed.isVisible() ||
          !lastViewed.isAlive())
      {
         return null;
      }

      return lastViewed;
   }

   /**
    * Stop the WAIT-FOR loop, as it determined there are no available widgets,
    * in the following cases:
    * <ul>
    *    <li>when the terminal is reset, inside an editing block</li>
    *    <li>
    *       when the WAIT-FOR loop is NOT waiting for global events and there
    *       are no enabled components registered for exit events
    *     </li>
    * </ul>
    *
    * @throws  StopConditionException
    *          always thrown - stop all other processing
    */
   private void stopNoAvailableWidgets()
   {
      String err = "None of the widgets used in the WAIT-FOR " +
                   "statement are in a state (such as SENSITIVE) " +
                   "such that the specified event can occur.  " +
                   "WAIT-FOR terminated. (4123)";
      Window wnd = WindowManager.resolveWindow();
      wnd.setMessageNeedPause(false);
      wnd.message(err);

      try
      {
         pause(-1, textPBH, false, wnd);
      }
      catch (Throwable t)
      {
         // do nothing
      }

      // all other processing fails
      throw new StopConditionException(err);
   }

   /**
    * Outputs text as the result of PUT SCREEN operation.
    *
    * @param   text
    *          Text to out.
    * @param   row
    *          Row.
    * @param   column
    *          Column.
    * @param   color
    *          Color.
    *
    */
   private void putScreenDirect(String text, double row, double column,
                                Color color)
   {
      BasePrimitives directOutput = ((ChuiOutputManager) tk).getDirectOutput();
      
      int effRow = (int)(row - 1);
      int effCol = (int)(column - 1);

      boolean curvis = true;
      int cpx = 0;
      int cpy = 0;

      if (inputActive > 0)
      {
         // the cursor is active in an edit operation, remember the current
         // cursor position (and make the cursor invisible to avoid flashing)
         cpx = ((BasePrimitives) directOutput).getX();
         cpy = ((BasePrimitives) directOutput).getY();
         curvis = tk.setCursorStatus(false);
      }

      // honor the default color
      color = (color == null) ? Color.NORMAL : color;

      boolean res = ((BasePrimitives) directOutput).cursorAt(effCol, effRow);

      if (res)
      {
         ((BasePrimitives) directOutput).append(text, color);
      }

      if (inputActive > 0)
      {
         ((BasePrimitives) directOutput).cursorAt(cpx, cpy);
      }

      directOutput.sync();

      // set the cursor visibility
      tk.setCursorStatus(curvis);
   }

   /**
    * Positions the real cursor before pausing according to the focused widget
    * "directions".
    */
   private void positionCursor()
   {
      if (isRedirected() || !isChui())
      {
         return;
      }

      Dimension screenSize = WindowManager.getDefaultWindow().dimension();
      
      // cursor parking position
      NativePoint p = tk.coordinates().pointToNative(new Point(screenSize.width - 2, screenSize.height - 1));

      Widget comp = UiUtils.getCurrentFocus();
      if (comp != null && !comp.isEnabled())
      {
         comp = null;
      }

      // reposition the cursor before blocking (interactive mode only)
      if (comp != null)
      {
         Frame frame = (Frame)UiUtils.locateFrame(comp);

         if ((frame != null && !frame.isInChoose(comp)) ||
             comp instanceof ScrollableList)
         {
            if (!skipCaretDraw)
            {
               if (comp instanceof FixedSizeContainer)
               {
                  ((FixedSizeContainer)comp).drawCaret();
               }
               else if (comp instanceof ComboBox)
               {
                  ((ComboBox)comp).drawCaret();
               }
               else if (comp instanceof RadioButton)
               {
                  ((RadioButton)comp).drawCaret();
               }
               else if (comp instanceof RadioSet)
               {
                  RadioButton<OutputManager<?>> selected = ((RadioSet) comp).getSelected();
                  if (selected != null)
                  {
                     selected.drawCaret();
                  }
               }
               else if (comp instanceof SelectionList)
               {
                  ((SelectionList) comp).drawCaret();
               }
               else if (comp instanceof ScrollableList)
               {
                  ((ScrollableList)comp).drawCaret();
               }
               else
               {
                  return;
               }
            }
            else
            {
               // reset the flag
               skipCaretDraw = false;
            }

            // get the last cursor position as drawCaret() dispatched it
            p = ((ChuiOutputManager) tk).cursor();
         }
      }

      // make the cursor stay at this position
      ((ChuiOutputManager) tk).stayAt(p);
   }

   /**
    * Repaint widgets affected during ENABLE/DISABLE processing.
    *
    * @param   frames
    *          List of frames above current frame.
    * @param   widgets
    *          Array of affected widgets.
    */
   private void repaintAffectedWidgets(Frame[] frames, Iterable<Widget> widgets)
   {
      for (Widget widget : widgets)
      {
         if (widget.isVisible() && UiUtils.isAllShowing(widget, frames))
         {
            widget.repaint();
         }
      }
   }

   /**
    * Set {@link #pendingFocus} widget.
    *
    * @param   pendingFocus
    *          New value for {@link #pendingFocus}.
    */
   public void setPendingFocus(Widget pendingFocus)
   {
      this.pendingFocus = pendingFocus;
   }

   /**
    * Get {@link #pendingFocus} widget.
    *
    * @return  current {@link #pendingFocus} widget.
    */
   public Widget getPendingFocus()
   {
      return pendingFocus;
   }

   /**
    * Set value of {@link #triggerNesting} counter.
    *
    * @param   triggerNesting
    *          New value of {@link #triggerNesting} counter.
    */
   public void setTriggerNesting(int triggerNesting)
   {
      this.triggerNesting = triggerNesting;
   }

   /**
    * If this is a nested frame, when performing a VIEW FRAME, in some cases all the ancestors 
    * need to be displayed, if not already visible.
    *  
    * @param    frameBuf
    *           The frames buffer array for editable frames.
    * @param    widget
    *           The widget targeted by VIEW (may be a FRAME or another widget).
    * @param    widgetIds
    *           The widget list.
    */
   private void preprocessView(ScreenBuffer[] frameBuf, Widget widget, int[] widgetIds)
   {
      if (!(widget instanceof Frame) && 
          (widgetIds == null || !(widgetIds.length == 1 && getWidget(widgetIds[0]) == widget)))
      {
         // an explicit non-frame widget is used, do not recurse the VIEW
         return;
      }
      
      if (frameBuf == null || frameBuf[0] == null ||
          frameBuf[0].getCurrentStatement() != UIStatement.VIEW)
      {
         return;
      }
      
      Frame frame = getFrame(frameBuf[0].getBaseId());
      if (frame == null || frame.isRootFrame())
      {
         return;
      }
      
      // go up the hierarchy and force a VIEW for each parent frame
      frame = getFrame(frame.config().frameId);
      if (frame == null)
      {
         return;
      }

      int frameId = UiUtils.getWidgetIdAsInt(frame);
      ScreenBuffer[] buff = server.refresh(frameId);
      buff[0].setCurrentStatement(UIStatement.VIEW);

      if (frame.isVisible())
      {
         preprocessView(buff, frame, null);
      }
      else
      {
         view_(frameId, buff, null, true, true, frame.config().windowID);
         
         frame.config().visible = true;
         frame.config().assignedVisible = true;
         frame.config().hidden = false;
      }
   }
   
   /**
    * Process frame definition.
    *
    * @param   frame
    *          Frame to process.
    * @param   cc
    *          New widget config.
    */
   private void preprocessFrameDef(final Frame frame, final WidgetConfig cc)
   {
      FrameConfig fc = (FrameConfig) cc;
      if (!frame.isRedirected())
      {
         if (fc.positionChanged || fc.sizeChanged)
         {
            frame.setVisible(false);
            frame.repaint();
            frame.setVisible(true);
         }
      }
      else
      {
         double toCol   = ConfigHelper.getEffectiveColumn(fc);
         double fromCol = ConfigHelper.getEffectiveColumn(frame.config());

         // safe, comparison of scaled doubles
         if (fc.positionChanged && toCol != fromCol && toCol >= 0)
         {
            // all the previous output must be moved to the new column
            // if the column was modified
            int sid = frame.getStreamId();
            RedirectedTerminal term = redirections.get(sid);
            
            int nfromCol = redirectedConversion.columnToNative(fromCol);
            int ntoCol = redirectedConversion.columnToNative(toCol);
            term.moveOutputToColumn(nfromCol, ntoCol);
            
         }
      }
   }

   /**
    * Get the {@link ServerExports} instance.
    * 
    * @return   See above.
    */
   public ServerExports getServer()
   {
      return this.server;
   }
   
   /**
    * Get the {@link LowLevelBuffer} implementation.
    * 
    * @return   See above.
    */
   public LowLevelBuffer getMemoryDaemon()
   {
      return mmd;
   }
   
   /**
    * Performs conditional focus change procedure to the next widget.
    * The focus change can be cancelled by a trigger.
    *
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    *
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean nextFocus(Widget widget)
   {
      return getInstance().focusManager.focusChange(widget, false, true);
   }

   /**
    * Performs conditional focus change procedure to the previous widget.
    * The focus change can be cancelled by a trigger.
    *
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    *
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean previousFocus(Widget widget)
   {
      return getInstance().focusManager.focusChange(widget, false, false);
   }

   /**
    * Performs conditional focus change procedure to the next tab stop widget.
    * The focus change can be cancelled by a trigger.
    *
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    *
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean nextTabStop(Widget widget)
   {
      return getInstance().focusManager.focusChange(widget, true, true);
   }

   /**
    * Performs conditional focus change procedure to the previous tab stop widget.
    * The focus change can be cancelled by a trigger.
    *
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    *
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean previousTabStop(Widget widget)
   {
      return getInstance().focusManager.focusChange(widget, true, false);
   }

   /**
    * Resolve the widget registry being in use.
    */
   public WidgetRegistry registry()
   {
      return tk.getRegistry();
   }
   
   /**
    * Resolves client-side window with specified id.
    * 
    * @param   windowId
    *          Id of target window or -1 for current or default window
    *          
    * @return  target Window with specified Id or "default" window. 
    *          See {@link WindowManager#resolveWindow} for details.
    */
   private Window<?> targetWindowWorker(int windowId)
   {
      Window<?> wnd;
      if (windowId == WindowConfig.RESOLVE_WINDOW)
      {
         wnd = WindowManager.resolveWindow();
      }
      else
      {
         wnd = (Window<?>) tk.getRegistry().getComponent(windowId);
         if (wnd == null)
         {
            throw new IllegalArgumentException("There are no Window with this id=" + windowId);
         }
      }
      return wnd;
   }

   /**
    * Resolve the input default text (if not already resolved) and return it.
    * 
    * @return   See above.
    */
   public String getInputDefaultText()
   {
      if (inputDefaultText == null)
      {
         inputDefaultText = "Enter data or press " + Keyboard.kbLabel("END-ERROR") + " to end.";
      }
      
      return inputDefaultText;
   }
   
   /**
    * Returns display width in native units.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The width in native units.
    */
   @Override
   public int getDisplayWidth(int windowId)
   {
      return getDisplayBounds(windowId).width();
   }

   /**
    * Returns display height in native.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The height in native.
    */
   @Override
   public int getDisplayHeight(int windowId)
   {
      return getDisplayBounds(windowId).height();   
   }

   /**
    * Returns the X coordinate of the application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The value in native units.
    */
   @Override
   public int getDisplayWorkAreaX(int windowId)
   {
      return getDisplayWorkArea(windowId).left();
   }

   /**
    * Returns the Y coordinate of the application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The value in native units.
    */
   @Override
   public int getDisplayWorkAreaY(int windowId)
   {
      return getDisplayWorkArea(windowId).top();
   }

   /**
    * Returns the width of the application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The value in native units.
    */
   @Override
   public int getDisplayWorkAreaWidth(int windowId)
   {
      NativeRectangle dim = getDisplayWorkArea(windowId);
      return dim.right() - dim.left() + 1;
   }

   /**
    * Returns the height of the application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The value in native units.
    */
   @Override
   public int getDisplayWorkAreaHeight(int windowId)
   {
      NativeRectangle dim = getDisplayWorkArea(windowId);
      return dim.bottom() - dim.top() + 1;
   }

   /**
    * Sets the new application icon for given window on the client side.
    * 
    * @param    windowId
    *           A valid window id.
    * @param    icon
    *           A new icon image array date.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean setWindowIcon(final int windowId, final byte[] icon)
   {
      // route call to respective Window implementation
      final boolean[] res = { false };
      
      independentEventDrawingBracket(getWidget(windowId), new Runnable()
      {
         @Override
         public void run()
         {
            res[0] = targetWindowWorker(windowId).setWindowIcon(icon);
         }
      });
      
      return res[0];
   }
   
   /**
    * Loads the new application icon for given window from the client side filesystem.
    * 
    * @param    windowId
    *           A valid window id.
    * @param    small
    *           Flag indicating usage of the small icon case(16x16 in size).
    * @param    name
    *           The fully qualified file name for the icon to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean loadWindowIcon(final int windowId, final boolean small, final String name)
   {
      // route call to respective Window implementation
      final boolean[] res = { false };
      
      independentEventDrawingBracket(getWidget(windowId), new Runnable()
      {
         @Override
         public void run()
         {
            // at this time we do not have image found, next step - search in PROPATH
            String fileNameToLoad = fsd.searchPath(name);

            res[0] = (fileNameToLoad == null ? false :
               targetWindowWorker(windowId).loadWindowIcon(small, fileNameToLoad));
         }
      });
      
      return res[0];
   }

   /**
    * Sets the default window icon for all new windows on the client side.
    * 
    * @param    type
    *           Data type for second parameter.
    * @param    data
    *           Data to create new icon on client side.  Can be the byte array, filename string
    *           or nothing for predefined icon from p2j.jar client side file. 
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean setWindowDefaultIcon(Window.IconDataType type, Object data)
   {
      if (type == Window.IconDataType.ICON_CUSTOM_CLIENT_FILENAME)
      {
         String fileNameToLoad = fsd.searchPath((String) data);
         // search the icon on client file system or load the default one from p2j.jar
         // if not found
         if (fileNameToLoad != null)
         {
            return Window.setWindowDefaultIcon(type, fileNameToLoad);
         }
         else
         {
            return Window.setWindowDefaultIcon(Window.IconDataType.ICON_DEFAULT, null);
         }
      }
      else
      {
         return Window.setWindowDefaultIcon(type, data);
      }
   }

   /**
    * Sets the new image for given button or image widgets on the client side.
    * 
    * @param    wId
    *           A valid wdget id.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons, -1 for image.
    * @param    image
    *           A new image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean setImage(int wId, int type, byte[] image)
   {
      Widget w = getWidget(wId);
      
      // this is pending load case, store data in the session cache
      if (w == null)
      {
         return ImageHelper.storePendingImageData(wId, type, image);
      }
      
      if (w instanceof ButtonGuiImpl)
      {
         return ((ButtonGuiImpl) w).setImage(type, image);
      }
      else if (w instanceof ImageGuiImpl)
      {
         return ((ImageGuiImpl) w).setImage(image);
      }
      else
      {
         return false;
      }
   }

   /**
    * Loads the new image for given button or image widgets from the client side filesystem.
    * 
    * @param    wId
    *           A valid wdget id.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons, -1 for image.
    * @param    name
    *           The fully qualified file name for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean loadImage(int wId, int type, String name)
   {
      // find out if requested image in prebuilt in p2j.jar
      if (ImageHelper.isPrepackaged(name))
      {
         // load prebuilt image from p2j.jar
         byte[] image = ImageHelper.getPrepackagedImageData(name);
         byte[] imageIns = ImageHelper.getPrepackagedImageData(
                              name.concat(ButtonConfig.PREPACKAGED_INS_SUFFIX));
         // load OK, set the image, otherwise do nothing, no search file system for alternative
         return image != null && imageIns != null
                ? setImage(wId, type, image) && setImage(wId, ButtonConfig.IMAGE_INS, imageIns)
                : false; 
      }

      Widget w = getWidget(wId);

      String fileNameToLoad = getImageFileName(name);

      if (fileNameToLoad == null)
      {
         return false;
      }
      
      // this is pending load case, store data in the session cache
      if (w == null)
      {
         return ImageHelper.storePendingImageName(wId, type, fileNameToLoad);
      }
      
      if (w instanceof ButtonGuiImpl)
      {
         return ((ButtonGuiImpl) w).loadImage(type, fileNameToLoad);
      }
      else if (w instanceof ImageGuiImpl)
      {
         return ((ImageGuiImpl) w).loadImage(fileNameToLoad);
      }
      else
      {
         return false;
      }
   }

   /**
    * Find the name of the image file in PROPATH appending possible exentions if needed
    * 
    * @param imageName
    *        the name of the image
    * 
    * @return
    *        the name of the image file or <code>null</code> of not found
    */
   public String getImageFileName(String imageName)
   {
      if (!"".equals(Utils.getExt(imageName)))
      {
         return fsd.searchPath(imageName);
      }
      return fsd.searchPathWithExts(imageName, 
            fsd.isCaseSensitive() ? ImageExt.EXT_VARIANTS : ImageExt.EXT_BASES);
   }
   
   /**
    * Save the content in the widgets in the client side filesystem.
    *
    * @param    wId
    *           A valid editor widget id.
    * @param    name
    *           The file name for the file to be written.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean editorSaveFile(int wId, String name)
   {
      Widget w = getWidget(wId);
      if (!(w instanceof Editor))
      {
         return false;
      }
      return ((Editor)w).editorSaveFile(name);
   }

   /**
    * Loads the file from the client side filesystem.
    *
    * @param    wId
    *           A valid editor widget id.
    * @param    name
    *           The file name for the file to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   @Override
   public boolean editorReadFile(int wId, String name)
   {
      Widget w = getWidget(wId);
      if (!(w instanceof Editor))
      {
         return false;
      }

      return ((Editor)w).editorReadFile(name);
   }

   /**
    * Locates a file name to load from local file system. The image file must be somewhere in
    * the current {@code PROPATH} directories. The method will also adjust the case-sensitivity
    * so that in case of porting from a Windows installation to Linux files can be open even if
    * {@code name} does not match the exact cases required by Linux filesystem. 
    * 
    * @param   name
    *          The name of the file to be located in the client file system.
    *
    * @return  The file name from {@code PROPATH} directories with adjusted casings or 
    *          {@code null} if the specified file cannot be found.
    */
   public static String getLocalFileName(String name)
   {
      if (name == null)
      {
         return null;
      }
      
      ThinClient tc = getInstance();
      
      return tc.fsd.searchPath(name);
   }

   /**
    * Get the THREE-D value for this client session.
    * 
    * @return   See above.
    */
   @Override
   public boolean is3D()
   {
      return is3D;
   }

   /**
    * Set the THREE-D value for this client session.
    * 
    * @param    value
    *           The value to set.
    */
   @Override
   public void set3D(boolean value)
   {
      is3D = value;
   }

   /**
    * Get the Keep3DFillinBorder environment value for this client session.
    * 
    * @return   See above.
    */
   public boolean isKeep3DFillinBorder()
   {
      return keep3DFillinBorder;
   }

   /**
    * Set the Keep3DFillinBorder environment value for this client session.
    * 
    * @param    value
    *           The value to set.
    */
   public void setKeep3DFillinBorder(boolean value)
   {
      keep3DFillinBorder = value;
   }
   
   /**
    * Move the widget to the top in z-order
    * 
    * @param id
    *          widget id
    *          
    * @return <code>Boolean.TRUE</code> if operation was successful
    *         null if no widget with given id is found
    */
   @Override
   public Boolean moveToTop(int id)
   {
      Widget<?> w = getWidget(id);
      if (w instanceof TopLevelWindow)
      {
         WindowManager.moveToTop((TopLevelWindow)w);
         return Boolean.TRUE;
      }
      if (w == null)
      {
         return null;
      }
      
      @SuppressWarnings("unchecked")
      Frame<?> frame = (Frame<?>) UiUtils.locateFrame(w);
      if (frame == null)
      {
         return Boolean.FALSE;
      }
      
      final TopLevelWindow wnd = frame.topLevelWindow();
      if (wnd == null)
      {
         return Boolean.FALSE; 
      }
      
      boolean moved = w.moveToTop();
      independentEventDrawingBracket(w, new Runnable() 
      {
         @Override
         public void run()
         {
            w.repaint();
         }
      });
      
      return Boolean.valueOf(moved);
   }

   /**
    * Move the widget to the bottom in z-order
    * 
    * @param id
    *          widget id
    *          
    * @return <code>Boolean.TRUE</code> if operation was successful
    *         null if no widget with given id is found
    */
   @Override
   public Boolean moveToBottom(int id)
   {
      Widget<?> w = getWidget(id);
      if (w == null)
      {
         return null;
      }
      
      if (w instanceof TopLevelWindow)
      {
         WindowManager.moveToBottom((TopLevelWindow)w);
         return Boolean.TRUE;
      }
      
      @SuppressWarnings("unchecked")
      Frame<?> frame = (Frame<?>) UiUtils.locateFrame(w);
      if (frame == null)
      {
         return Boolean.FALSE;
      }
      
      final TopLevelWindow wnd = frame.topLevelWindow();
      if (wnd == null)
      {
         return Boolean.FALSE; 
      }
      
      boolean moved = w.moveToBottom();
      independentEventDrawingBracket(w, new Runnable()
      {
         @Override
         public void run()
         {
            w.repaint();
         }
      });
      
      return Boolean.valueOf(moved);
   }

   /**
    * Set the custom mouse pointer for a widget or reset to the default one
    * 
    * @param id
    *        the widget id
    * @param data
    *        the pointer data
    * @param ptrName
    *        the pointer name
    *               
    * @return <code>true</code> if operation was successful       
    */
   public boolean setMousePointer(int id, byte[] data, String ptrName)
   {
      MousePtrWrapper ptr = createMousePointer(data, ptrName);
      if (ptr == null)
      {
         return false;
      }
      return setMousePtr(id, ptr);
   }

   /**
    * Set the the pre-defined custom mouse pointer for a widget or reset to the default one
    * 
    * @param id
    *        the widget id
    * @param ptr
    *        the new custom mouse pointer or <code>null</code>
    *               
    * @return <code>true</code> if operation was successful       
    */
   @Override
   public Boolean setMousePointer(int id, MousePointer ptr)
   {
      MousePtrWrapper ptrWrapper = null;
      if (ptr != null)
      {
         ptrWrapper = createPtrWrapper(ptr); 
         if (ptrWrapper == null)
         {
            return false;
         }
      }
      return setMousePtr(id, ptrWrapper);
   }

   /**
    * Load custom mouse pointer from a file
    * 
    * @param id
    *        widget id
    * @param fileName
    *        the file name of the pointer data
    * 
    * @return <code>true</code> if successful.
    */
   @Override
   public boolean loadMousePointer(int id, String fileName)
   {
      String ptrFileName = getImageFileName(fileName);
      ImageExt ext = ImageExt.fromExt(Utils.getExt(ptrFileName));
      if (ext == ImageExt.NONE ||ext == ImageExt.BMP)
      {
         return false;
      }
      MousePtrWrapper ptrWrapper = loadMousePointer(ptrFileName);
      if (ptrWrapper != null)
      {
         return setMousePtr(id, ptrWrapper);
      }
      return false;
   }

   /**
    * Set the state of SESSION:IMMEDIATE-DISPLAY attribute.
    * 
    * @param   on
    *          The new state for SESSION:IMMEDIATE-DISPLAY attribute.
    */
   @Override
   public void setImmediateDisplay(boolean on)
   {
      this.isImmediateDisplay = on;
   }

   /**
    * Set the state of SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @param   on
    *          The new state for SESSION:DATA-ENTRY-RETURN attribute.
    */
   @Override
   public void setDataEntryReturn(boolean on)
   {
      this.isDataEntryReturn = on;
   }

   /**
    * Get the state of SESSION:DATA-ENTRY-RETURN attribute. 
    * 
    * @return  the state of SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static boolean isDataEntryReturn()
   {
      Boolean on = getInstance().isDataEntryReturn; 
      return on == null ? getInstance().isChui() : on.booleanValue();
   }
   /**
    * Set the custom mouse pointer for a widget
    * 
    * @param id
    *        the widget id
    * @param ptrWrapper
    *        the mouse pointer
    *        
    * @return <code>true</code> on success
    */
   protected boolean setMousePtr(int id, MousePtrWrapper ptrWrapper)
   {
      Widget<?> w = getWidget(id);
      if (w == null)
      {
         return false;
      }
      ((GuiDriver)tk.getDriver()).selectWindow(w.window().getId().asInt());
      try
      {
         w.setMousePointer(ptrWrapper);
      }
      finally
      {
         ((GuiDriver)tk.getDriver()).releaseWindow();
      }
      return true;
   }


   /**
    * Initiate/cancel wait state
    * 
    * @param ptr 
    *        new mouse pointer or <code>null</code> for cancel 
    * 
    * @return <code>true</code> if operation was successful       
    */
   @Override
   public boolean setWaitState(MousePointer ptr)
   {
      LOG.info("setWaitState:" + ptr);
      return setWaitState(ptr == null ? null : createPtrWrapper(ptr));
   }

   /**
    * Initiate/cancel wait state
    * 
    * @param data 
    *        new mouse pointer data 
    * 
    * @return <code>true</code> if operation was successful       
    */
   @Override
   public boolean setWaitState(byte[] data)
   {
      LOG.info("setWaitState: (binary)");
      MousePtrWrapper ptr = createMousePointer(data, "WAIT-STATE");
      return ptr == null ? false : setWaitState(ptr);
   }

   /**
    * Initiate wait state with a custom cursor from file
    * 
    * @param ptrName 
    *        the name of the file with new mouse pointer data 
    * 
    * @return <code>true</code> if operation was successful       
    */
   public boolean setWaitState(String ptrName)
   {
      LOG.info("setWaitState: " + ptrName);
      MousePtrWrapper ptr = loadMousePointer(ptrName);
      return ptr == null ? false : setWaitState(ptr);
   }

   /**
    * Push information about the loaded mouse pointer for widget 
    * 
    * @param windowId
    *           to level window id
    * @param wid
    *           widgetId
    * @param ptr
    *           mouse pointer
    */
   public static void propagate(int windowId, int wid, MousePtrWrapper ptr)
   {
      getInstance().tk.getDriver().propagate(windowId, wid, ptr);
   }

   /**
    * Returns display size.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The dimensions in native units.
    */
   public NativeRectangle getDisplayBounds(int windowId)
   {
      ScreenDriver<?> driver = tk.getInstanceDriver();
      if (driver instanceof GuiDriver)
      {
         return ((GuiDriver) driver).getDisplayBounds(windowId);
      }
      else
      {
         Dimension dim = tk.screenDimension(windowId);
         return new NativeRectangle(new NativePoint(0, 0), 
                                    new NativeDimension((int) dim.width, (int) dim.height));
      }
   }
   
   /**
    * Requests the current wait-for event loop to be interrupted. The request 
    * is not processed right away but it is enqueued in the OS event queue
    * and dispatched later.
    */
   public void requestEndCurrentWaitForEventLoop()
   {
      invokeLaterOS(() -> { throw new EndConditionException((Throwable) null); });
   }
   
   /**
    * Load custom mouse pointer from a file
    * 
    * @param fileName
    *        the file name of the pointer data
    * 
    * @return mouse pointer.
    */
   private synchronized MousePtrWrapper loadMousePointer(String fileName)
   {
      MousePtrWrapper ptr = mousePointers.get(fileName);
      if (ptr != null)
      {
         return ptr;
      }
      
      File file = new File(fileName); 
      if (!file.exists() || !file.isFile() || !file.canRead())
      {
         return null;
      }
      FileInputStream fis;
      try
      {
         fis = new FileInputStream(file);
         ptr = createMousePointer(UiUtils.getData(fis), fileName);
         if (ptr != null)
         {
            mousePointers.put(fileName, ptr);
         }
      }
      catch (Exception e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not load cursor from a file " + file.getAbsolutePath(),
                  e);
         }

      }
      return ptr;
   }

  /**
    * Create mouse pointer from the binary data
    * 
    * @param data
    *        cusrsor data
    * @param ptrName
    *        pointer name
    *        
    * @return
    *        pointer wrapper or <code>null</code>
    */
   private MousePtrWrapper createMousePointer(byte[] data, String ptrName)
   {
      if (data == null)
      {
         return null;
      }
      ScreenDriver<?> driver = tk.getInstanceDriver();
      return (driver instanceof GuiDriver) ? 
                     ((GuiDriver) driver).createMousePointer(data, ptrName) : null; 
   }

   /**
    * Create wrapper for the pre-defined cursor
    * 
    * @param ptr
    *        source mouse pointer
    * 
    * @return wrapper for a pointer and a current GUI driver
    */
   public synchronized MousePtrWrapper createPtrWrapper(MousePointer ptr)
   {
      MousePtrWrapper ptrWrapper = wrappers.get(ptr);
      if (ptrWrapper != null)
      {
         return ptrWrapper;
      }

      ScreenDriver driver = OutputManager.instance().getInstanceDriver(); 
      if (!(driver instanceof GuiDriver))
      {
         return null;
      }

      GuiDriver gd = (GuiDriver)driver;
      
      CursorType type = ImageHelper.cursorType(ptr);
      if (type != null)
      {
         return MousePtrWrapper.createPreDefinedInstance(type);
      }
      
      String fileName = ImageHelper.cursorFile(ptr);
      if (fileName != null)
      {
         // loading image embedded into p2j.jar
         InputStream is = ImageHelper.class.getResourceAsStream(fileName);
         if (is == null)
         {
            return null;
         }
         
         byte[] imgData = UiUtils.getData(is);
         if (imgData == null)
         {
            return null;
         }
         ptrWrapper = gd.createMousePointer(imgData, ptr.progressName);
         if (ptrWrapper != null)
         {
            wrappers.put(ptr, ptrWrapper);
         }
         return ptrWrapper;
      }
      
      return null;
   }

   /**
    * Initiate/cancel wait state
    * 
    * @param ptr 
    *        new mouse pointer or <code>null</code> for cancel 
    * 
    * @return <code>true</code> if operation was successful       
    */
   private boolean setWaitState(MousePtrWrapper ptr)
   {
      ScreenDriver driver = OutputManager.getDriver();
      if (!(driver instanceof GuiDriver))
      {
         return false;
      }
      GuiDriver gd = (GuiDriver)driver;
      
      isWaitState = (ptr != null);
      gd.propagateWaitState(isWaitState);
      lockMousePointer(-1, ptr);
      if (!isWaitState)
      {
         refreshHover();
      }
      else
      {
         if (activeComboBox != null)
         {
            independentEventDrawingBracket(activeComboBox, activeComboBox.deactivateCode());
         }
         
         Window wnd = WindowManager.getActiveWindow();
         if (wnd != null)
         {
            try
            {
               gd.selectWindow(wnd.getId().asInt());
               try
               {
                  gd.setCursor(lockedMousePointer().get());
               }
               finally
               {
                  gd.releaseWindow();
               }
            }
            catch (Exception e)
            {
               LOG.log(Level.WARNING, "Unexpected exception", e);
               return false;
            }
         }
      }
      return true;
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The area in native units.
    */
   private NativeRectangle getDisplayWorkArea(int windowId)
   {
      ScreenDriver<?> driver = tk.getInstanceDriver();
      if (driver instanceof GuiDriver)
      {
         return ((GuiDriver) driver).getDisplayWorkArea(windowId);
      }
      else
      {
         Dimension dim = tk.screenDimension(windowId);
         return new NativeRectangle(0, 0, (int) dim.height, (int) dim.width);
      }
   }
   
   /**
    * Extract the frames from the stack of wait-for event loops.
    * 
    * @param    waitForOps
    *           The current operations.
    *           
    * @return   The list of frames.
    */
   private List<Frame> extractFrames(Deque<WaitForOp> waitForOps)
   {
      List<Frame> res = new ArrayList<>(waitForOps.size());
      for (WaitForOp op : waitForOps)
      {
         if (op.frame != null)
         {
            res.add(op.frame);
         }
      }
      
      return res;
   }

   /**
    * Returns if the current window has at least one enabled frame.
    * (The frame with enabled widgets).
    * 
    * @return   <code>true</code> if there is a frame exists.
    */
   private boolean hasEnabledFrame()
   {
      Widget first = focusManager.getFirstEnabledWidget();
      
      // not MENUBAR
      return first != null && !(first instanceof Menu) && !(first instanceof SubMenu);
   }
   
   /**
    * Shows a warning in the case when menubar is assigned to a window,
    * that has no enabled frames.
    * 
    * @param   wnd
    *          Target window to search frames.
    */
   private void warnNoEnabledFrames(Window wnd)
   {
      if (!isChui())
      {
         UnimplementedFeature.todo("Line number and program name resolution is not implemented yet.");

         String err = "Run-time error in WAIT-FOR at line <line_num> (relative to expanded " + 
                      "source) of <program_name>. (3269)"; 
         
         wnd.setMessageNeedPause(false);
         wnd.message(err);

         try
         {
            pause(-1, textPBH, false, wnd);

            wnd.status(null);
            wnd.setMessageText(null, null);
         }
         catch (Throwable t)
         {
            // do nothing
         }
      }
   }
   
   /**
    * Open the target page or resource in the browser's window.
    * 
    * @param    path
    *           The URL path to the target page or resource.
    */
   @Override
   public void openURL(String path)
   {
      OutputManager.getDriver().openURL(path);
   }

   /**
    * Exposes the client's system clipboard to the server.
    * 
    * @return   The clipboard content.
    */
   public character getClipboardValue()
   {
      String value = "";
      if (!getOutputManager().isChui())
      {
         GuiDriver gd = (GuiDriver) OutputManager.getDriver();
         value = gd.clipboardContents();
      }
      else
      {
         throw new UnsupportedOperationException();
      }
      
      return new character(value);
   }
   
   /**
    * Copies a string value to the client's system clipboard.
    * 
    * @param    value
    *           The string to be stored in the client's system clipboard.
    */
   public void setClipboardValue(String value)
   {
      if (!getOutputManager().isChui())
      {
         GuiDriver gd = (GuiDriver) OutputManager.getDriver();
         gd.copyToClipboard(value);
      }
      else
      {
         throw new UnsupportedOperationException();
      }
   }
   
   /**
    * Obtain the default system palette for current theme from client.
    *
    * @return  the default system palette for current theme from client. 
    */
   public Map<Integer, ColorRgb> getThemeSystemPalette()
   {
      return ThemeManager.getCurrentTheme().getDefaultSystemPalette();
   }
   
   /**
    * Container for the flags related to trigger processing.
    */
   public static class TriggerResult
   {
      /** <code>true</code> if a trigger was found and executed. */
      public boolean executed = false;

      /** <code>true</code> sets <code>NO-APPLY</code> for the trigger. */
      public boolean consume = false;
   }

   /**
    * Widget/frame hiding processor
    *
    * @author  SIY
    * @version 1.0
    */
   private final class HidingProcessor
   implements Runnable
   {
      /** Widget ID. */
      private final int widgetId;

      /** NO-PAUSE flag */
      private final boolean noPause;

      /** The state of the HIDDEN attribute. */
      private boolean hidden;
      
      /**
       * Constructor.
       *
       * @param   widgetId
       *          Widget ID.
       * @param   noPause
       *          NO-PAUSE flag.
       * @param   hidden
       *          If <code>true</code>, set the HIDDEN attribute.
       */
      private HidingProcessor(int widgetId, boolean noPause, boolean hidden)
      {
         this.widgetId = widgetId;
         this.noPause = noPause;
         this.hidden = hidden;
      }

      /**
       * @see java.lang.Runnable#run()
       */
      @Override
      public void run()
      {
         try
         {
            hideWorker(widgetId, noPause);
         }
         finally
         {
            if (noPause)
            {
               needPause = false;
            }
         }
      }
      
      /**
       * Checks whether the supplied <code>widget</code> is allowed
       * to be hidden. If so, returns <code>true</code> or <code>false</code>
       * otherwise.
       * 
       * @param   widget
       *          A widget reference.
       *          
       * @return  See above.
       */
      private boolean validateHide(Widget widget)
      {
         Frame frame = widget instanceof Frame ? ((Frame) widget) : null;
         
         if (!isChui() && frame != null && frame.isVisible() && frame.isDialog())
         {
            DialogBoxWindow box = (DialogBoxWindow) frame.parent(DialogBoxWindow.class);
            TopLevelWindow<?> owner = box.getOwner();
            int ownerId = owner.getId().asInt();

            // get all visible dialog boxes with the same owner as this frame's 
            // dialog box owner; the list is in z-order
            List<DialogBoxWindow> dialogs = WindowManager.getDialogsOwnedBy(ownerId, false);

            // raise an error if this frame's owner is not the top dialog 
            // and do not proceed with the hide
            DialogBoxWindow top = dialogs.isEmpty() ? null : dialogs.get(dialogs.size() - 1);
            if (top != null && top != box)
            {
               String msg = "Cannot hide DIALOG-BOX %s while DIALOG-BOX %s is currently active";
               msg = String.format(msg, frame.shortName(), top.getFrame().shortName());
               ErrorManager.recordOrShowError(4113, msg, true, false, true);
               return false;
            }
            
            // hiding DIALOG-BOX when ALERT-BOX is active is not a valid case,
            // it is assumed that no 4gl-app logic can be trigered while an
            // ALERT-BOX active
         }
         
         return true;
      }

      /**
       * Hiding worker.
       *
       * @param   widgetId
       *          Widget ID
       * @param   noPause
       *          NO-PAUSE flag
       */
      private void hideWorker(final int widgetId, final boolean noPause)
      {
         Widget comp = getWidget(widgetId);

         if (comp instanceof Frame)
         {
            hideFrame(comp, noPause);
         }
         else if (UiUtils.hasConfig(comp))
         {
            hideWidget(comp);
         }
      }

      /**
       * Hide regular widget.
       *
       * @param   comp
       *          Widget to hide.
       */
      private void hideWidget(Widget comp)
      {
         if (comp.isFocusable())
         {
            Frame frame = (Frame) UiUtils.locateFrame(comp);
            
            if (frame != null)
            {
               // set as pending focus the next enabled widget
               pendingFocusApply = frame.getNextEnabledWidget(comp,
                                                              FocusManager.FOCUS_WRAP);
            }
         }
         comp.setVisible(false);

         comp.setHidden(true);

         // once hidden, its highlight state must be lost
         comp.setHighlighted(false);
      }

      /**
       * Hide Frame widget.
       * @param   comp
       *          Frame to hide.
       * @param   noPause
       *          NO-PAUSE flag.
       */
      private void hideFrame(Widget comp, final boolean noPause)
      {
         Frame frame = (Frame) comp;

         boolean visible = frame.isVisible();

         if (visible      &&
             refreshQuirk &&
             noPause      &&
             frame.config().box)
         {
            switchLabelsEmptyMode(frame);
         }

         // reset frame visibility tracking
         frame.setVisibilityChanged(false);

         if (!frame.isDialog() &&
             frame.isNeedPause() &&
             !frame.isDown())
         {
            messageClearQuirk  = true;
         }

         // hide frame first, this will suppress
         // repainting on hiding of other components
         FramePlacementManager placementManager = frame.window().getPlacementManager();
         placementManager.hideFrame(frame, (noPause == false));

         if (visible)
         {
            hideWidgetsInFrame(frame);
         }

         if (frame.isDialog())
         {
            hideDialog(frame);
         }
         
         if (hidden)
         {
            frame.setVisible(false);
            frame.setHidden(true);
         }
      }

      /**
       * Hide dialog frame.
       *
       * @param   frame
       *          Dialog to hide.
       */
      private void hideDialog(Frame frame)
      {
         // check to see whether another dialog has to be
         // reactivated
         Container parent = frame.parent();
         Widget lst[]  = parent.widgets();

         int nlst = lst.length;
         int i = 0;

         for (i = nlst - 1; i >= 0; i--)
         {
            if (lst[i].isVisible() && lst[i] instanceof Frame &&
               (((Frame)lst[i]).isDialog()))
            {
               ((Frame)lst[i]).moveToTop();
               lst[i].repaint();
               break;
            }
         }
      }

      /**
       * Hide inner widgets in frame.
       *
       * @param   frame
       *          Input frame.
       */
      private void hideWidgetsInFrame(Frame frame)
      {
         // for visible frames, set the explicit hide flag
         frame.setExplicitHide(true);

         if (pendingFocus == null &&
             triggerNesting > 0   &&
             UiUtils.locateFrame(UiUtils.getCurrentFocus()) == frame)
         {
            // if the focus frame is hidden, then first enabled
            // widget on screen must be saved as pending focus;
            // this is only for nested trigger mode
            pendingFocus = focusManager.getFirstEnabledWidget();
         }

         // reset VISIBLE flag for all widgets
         Widget[] comps = frame.getContentPane().widgets();

         for (Widget component : comps)
         {
            if (component.isVisible())
            {
               component.setVisible(false);
               component.setHighlighted(false);
            }
         }
      }

      /**
       * Switch special label handling flag.
       *
       * @param   frame
       *          Frame to process.
       */
      private void switchLabelsEmptyMode(Frame frame)
      {
         Rectangle rect = frame.calculateFrameRect(true);

         rect = new Rectangle(rect.bottom(),
                              rect.left(),
                              rect.bottom(),
                              rect.right());

         List<Frame<?>> list = new ArrayList<Frame<?>>();

         frame.window().findFrames(list, rect, false, frame.isRedirected());

         if (list.size() == 1 &&
             !list.get(0).config().box &&
             (list.get(0).calculateFrameRect(true).right() ==
              rect.right()+2) &&
             (list.get(0).calculateFrameRect(true).bottom() ==
              rect.bottom()))
         {
            list.get(0).switchLabelsEmptyMode(rect);
         }
      }
   }

   /**
    * Status data container.
    */
   private static class StatusData
   {
      /** Status type. */
      final int type;
      
      /** Status input. */
      final int input;

      /**
       * Constructor.
       *
       * @param    type
       *           Status type.
       * @param    input
       *           Status input.
       */
      public StatusData(int type, int input)
      {
         this.type  = type;
         
         this.input = input;
      }
   }
   
   /**
    * Container for data related to wait-for loops.
    */
   private static class WaitForOp
   {
      /** The frame targeted by a wait-for event loop. */
      final Frame frame;
      
      /** The blocking operation for this wait-for loop. */
      final BlockingOperation op;
      
      /**
       * Constructor.
       *
       * @param    frame
       *           The frame.
       * @param    op
       *           The blocking operation.
       */
      public WaitForOp(Frame frame, BlockingOperation op)
      {
         this.frame = frame;
         this.op = op;
      }
   }
}